#include "tree_sitter/parser.h"

#if defined(__GNUC__) || defined(__clang__)
#pragma GCC diagnostic ignored "-Wmissing-field-initializers"
#endif

#ifdef _MSC_VER
#pragma optimize("", off)
#elif defined(__clang__)
#pragma clang optimize off
#elif defined(__GNUC__)
#pragma GCC optimize ("O0")
#endif

#define LANGUAGE_VERSION 14
#define STATE_COUNT 7671
#define LARGE_STATE_COUNT 528
#define SYMBOL_COUNT 280
#define ALIAS_COUNT 0
#define TOKEN_COUNT 175
#define EXTERNAL_TOKEN_COUNT 29
#define FIELD_COUNT 19
#define MAX_ALIAS_SEQUENCE_LENGTH 8
#define PRODUCTION_ID_COUNT 159

enum ts_symbol_identifiers {
  sym_word = 1,
  anon_sym_for = 2,
  anon_sym_select = 3,
  anon_sym_in = 4,
  anon_sym_LPAREN_LPAREN = 5,
  anon_sym_RPAREN_RPAREN = 6,
  anon_sym_SEMI = 7,
  anon_sym_COMMA = 8,
  anon_sym_EQ = 9,
  anon_sym_PLUS_PLUS = 10,
  anon_sym_DASH_DASH = 11,
  anon_sym_PLUS_EQ = 12,
  anon_sym_DASH_EQ = 13,
  anon_sym_STAR_EQ = 14,
  anon_sym_SLASH_EQ = 15,
  anon_sym_PERCENT_EQ = 16,
  anon_sym_STAR_STAR_EQ = 17,
  anon_sym_LT_LT_EQ = 18,
  anon_sym_GT_GT_EQ = 19,
  anon_sym_AMP_EQ = 20,
  anon_sym_CARET_EQ = 21,
  anon_sym_PIPE_EQ = 22,
  anon_sym_PIPE_PIPE = 23,
  anon_sym_DASHo = 24,
  anon_sym_AMP_AMP = 25,
  anon_sym_DASHa = 26,
  anon_sym_PIPE = 27,
  anon_sym_CARET = 28,
  anon_sym_AMP = 29,
  anon_sym_EQ_EQ = 30,
  anon_sym_BANG_EQ = 31,
  anon_sym_LT = 32,
  anon_sym_GT = 33,
  anon_sym_LT_EQ = 34,
  anon_sym_GT_EQ = 35,
  anon_sym_LT_LT = 36,
  anon_sym_GT_GT = 37,
  anon_sym_PLUS = 38,
  anon_sym_DASH = 39,
  anon_sym_STAR = 40,
  anon_sym_SLASH = 41,
  anon_sym_PERCENT = 42,
  anon_sym_STAR_STAR = 43,
  anon_sym_LPAREN = 44,
  anon_sym_RPAREN = 45,
  aux_sym__c_word_token1 = 46,
  anon_sym_while = 47,
  anon_sym_until = 48,
  anon_sym_do = 49,
  anon_sym_done = 50,
  anon_sym_if = 51,
  anon_sym_then = 52,
  anon_sym_fi = 53,
  anon_sym_elif = 54,
  anon_sym_else = 55,
  anon_sym_case = 56,
  anon_sym_esac = 57,
  anon_sym_SEMI_SEMI = 58,
  anon_sym_SEMI_AMP = 59,
  anon_sym_SEMI_SEMI_AMP = 60,
  anon_sym_function = 61,
  anon_sym_LBRACE = 62,
  anon_sym_RBRACE = 63,
  anon_sym_PIPE_AMP = 64,
  anon_sym_BANG = 65,
  anon_sym_LBRACK = 66,
  anon_sym_RBRACK = 67,
  anon_sym_LBRACK_LBRACK = 68,
  anon_sym_RBRACK_RBRACK = 69,
  anon_sym_declare = 70,
  anon_sym_typeset = 71,
  anon_sym_export = 72,
  anon_sym_readonly = 73,
  anon_sym_local = 74,
  anon_sym_unset = 75,
  anon_sym_unsetenv = 76,
  anon_sym_EQ_TILDE = 77,
  anon_sym_AMP_GT = 78,
  anon_sym_AMP_GT_GT = 79,
  anon_sym_LT_AMP = 80,
  anon_sym_GT_AMP = 81,
  anon_sym_GT_PIPE = 82,
  anon_sym_LT_AMP_DASH = 83,
  anon_sym_GT_AMP_DASH = 84,
  anon_sym_LT_LT_DASH = 85,
  aux_sym_heredoc_redirect_token1 = 86,
  anon_sym_LT_LT_LT = 87,
  anon_sym_QMARK = 88,
  anon_sym_COLON = 89,
  anon_sym_PLUS_PLUS2 = 90,
  anon_sym_DASH_DASH2 = 91,
  anon_sym_DASH2 = 92,
  anon_sym_PLUS2 = 93,
  anon_sym_TILDE = 94,
  anon_sym_DOLLAR_LPAREN_LPAREN = 95,
  anon_sym_DOLLAR_LBRACK = 96,
  aux_sym_brace_expression_token1 = 97,
  anon_sym_DOT_DOT = 98,
  anon_sym_RBRACE2 = 99,
  aux_sym_concatenation_token1 = 100,
  anon_sym_DOLLAR = 101,
  sym__special_character = 102,
  anon_sym_DQUOTE = 103,
  sym_string_content = 104,
  sym_raw_string = 105,
  sym_ansi_c_string = 106,
  aux_sym_number_token1 = 107,
  aux_sym_number_token2 = 108,
  anon_sym_POUND = 109,
  anon_sym_DOLLAR_LBRACE = 110,
  anon_sym_RBRACE3 = 111,
  anon_sym_BANG2 = 112,
  anon_sym_AT = 113,
  anon_sym_STAR2 = 114,
  anon_sym_POUND2 = 115,
  anon_sym_EQ2 = 116,
  anon_sym_COLON_EQ = 117,
  anon_sym_DASH3 = 118,
  anon_sym_COLON_DASH = 119,
  anon_sym_PLUS3 = 120,
  anon_sym_COLON_PLUS = 121,
  anon_sym_QMARK2 = 122,
  anon_sym_COLON_QMARK = 123,
  anon_sym_PERCENT_PERCENT = 124,
  aux_sym__expansion_regex_token1 = 125,
  anon_sym_SLASH_SLASH = 126,
  anon_sym_SLASH_POUND = 127,
  anon_sym_SLASH_PERCENT = 128,
  anon_sym_COMMA_COMMA = 129,
  anon_sym_CARET_CARET = 130,
  anon_sym_U = 131,
  anon_sym_u = 132,
  anon_sym_L = 133,
  anon_sym_Q = 134,
  anon_sym_E = 135,
  anon_sym_P = 136,
  anon_sym_A = 137,
  anon_sym_K = 138,
  anon_sym_a = 139,
  anon_sym_k = 140,
  anon_sym_DOLLAR_LPAREN = 141,
  anon_sym_BQUOTE = 142,
  anon_sym_DOLLAR_BQUOTE = 143,
  anon_sym_LT_LPAREN = 144,
  anon_sym_GT_LPAREN = 145,
  sym_comment = 146,
  sym__comment_word = 147,
  aux_sym__simple_variable_name_token1 = 148,
  aux_sym__multiline_variable_name_token1 = 149,
  anon_sym_AT2 = 150,
  anon_sym_0 = 151,
  anon_sym__ = 152,
  sym_heredoc_start = 153,
  sym_simple_heredoc_body = 154,
  sym__heredoc_body_beginning = 155,
  sym_heredoc_content = 156,
  sym_heredoc_end = 157,
  sym_file_descriptor = 158,
  sym__empty_value = 159,
  sym__concat = 160,
  sym_variable_name = 161,
  sym_test_operator = 162,
  sym_regex = 163,
  sym__regex_no_slash = 164,
  sym__regex_no_space = 165,
  sym__expansion_word = 166,
  sym_extglob_pattern = 167,
  sym__bare_dollar = 168,
  sym__brace_start = 169,
  sym__immediate_double_hash = 170,
  sym__external_expansion_sym_hash = 171,
  sym__external_expansion_sym_bang = 172,
  sym__external_expansion_sym_equal = 173,
  sym___error_recovery = 174,
  sym_program = 175,
  sym__statements = 176,
  aux_sym__terminated_statement = 177,
  sym__statement_not_pipeline = 178,
  sym_redirected_statement = 179,
  sym_for_statement = 180,
  sym_c_style_for_statement = 181,
  sym__for_body = 182,
  sym__c_expression = 183,
  sym__c_expression_not_assignment = 184,
  sym__c_variable_assignment = 185,
  sym__c_unary_expression = 186,
  sym__c_binary_expression = 187,
  sym__c_postfix_expression = 188,
  sym__c_parenthesized_expression = 189,
  sym_while_statement = 190,
  sym_do_group = 191,
  sym_if_statement = 192,
  sym_elif_clause = 193,
  sym_else_clause = 194,
  sym_case_statement = 195,
  sym_case_item = 196,
  sym_last_case_item = 197,
  sym_function_definition = 198,
  sym_compound_statement = 199,
  sym_subshell = 200,
  sym_pipeline = 201,
  sym_list = 202,
  sym_negated_command = 203,
  sym_test_command = 204,
  sym__test_command_binary_expression = 205,
  sym_declaration_command = 206,
  sym_unset_command = 207,
  sym_command = 208,
  sym_command_name = 209,
  sym_variable_assignment = 210,
  sym_variable_assignments = 211,
  sym_subscript = 212,
  sym_file_redirect = 213,
  sym_heredoc_redirect = 214,
  sym__heredoc_pipeline = 215,
  sym__heredoc_expression = 216,
  aux_sym__heredoc_command = 217,
  sym__heredoc_body = 218,
  sym_heredoc_body = 219,
  sym__simple_heredoc_body = 220,
  sym_herestring_redirect = 221,
  sym__expression = 222,
  sym_binary_expression = 223,
  sym_ternary_expression = 224,
  sym_unary_expression = 225,
  sym_postfix_expression = 226,
  sym_parenthesized_expression = 227,
  sym_arithmetic_expansion = 228,
  sym_brace_expression = 229,
  sym__arithmetic_expression = 230,
  sym__arithmetic_literal = 231,
  sym__arithmetic_binary_expression = 232,
  sym__arithmetic_ternary_expression = 233,
  sym__arithmetic_unary_expression = 234,
  sym__arithmetic_postfix_expression = 235,
  sym__arithmetic_parenthesized_expression = 236,
  sym_concatenation = 237,
  sym_string = 238,
  sym_translated_string = 239,
  sym_array = 240,
  sym_number = 241,
  sym_simple_expansion = 242,
  sym_expansion = 243,
  sym__expansion_body = 244,
  sym__expansion_expression = 245,
  sym__expansion_regex = 246,
  sym__expansion_regex_replacement = 247,
  sym__expansion_regex_removal = 248,
  sym__expansion_max_length = 249,
  sym__expansion_max_length_expression = 250,
  sym__expansion_max_length_binary_expression = 251,
  sym__expansion_operator = 252,
  sym__concatenation_in_expansion = 253,
  sym_command_substitution = 254,
  sym_process_substitution = 255,
  sym__extglob_blob = 256,
  sym__c_terminator = 257,
  aux_sym__statements_repeat1 = 258,
  aux_sym_redirected_statement_repeat1 = 259,
  aux_sym_redirected_statement_repeat2 = 260,
  aux_sym_for_statement_repeat1 = 261,
  aux_sym__for_body_repeat1 = 262,
  aux_sym_if_statement_repeat1 = 263,
  aux_sym_case_statement_repeat1 = 264,
  aux_sym_case_item_repeat1 = 265,
  aux_sym_pipeline_repeat1 = 266,
  aux_sym_declaration_command_repeat1 = 267,
  aux_sym_unset_command_repeat1 = 268,
  aux_sym_command_repeat1 = 269,
  aux_sym_command_repeat2 = 270,
  aux_sym_variable_assignments_repeat1 = 271,
  aux_sym_heredoc_body_repeat1 = 272,
  aux_sym__literal_repeat1 = 273,
  aux_sym_arithmetic_expansion_repeat1 = 274,
  aux_sym_concatenation_repeat1 = 275,
  aux_sym_string_repeat1 = 276,
  aux_sym__expansion_body_repeat1 = 277,
  aux_sym__expansion_regex_repeat1 = 278,
  aux_sym__concatenation_in_expansion_repeat1 = 279,
};

static const char * const ts_symbol_names[] = {
  [ts_builtin_sym_end] = "end",
  [sym_word] = "word",
  [anon_sym_for] = "for",
  [anon_sym_select] = "select",
  [anon_sym_in] = "in",
  [anon_sym_LPAREN_LPAREN] = "((",
  [anon_sym_RPAREN_RPAREN] = "))",
  [anon_sym_SEMI] = ";",
  [anon_sym_COMMA] = ",",
  [anon_sym_EQ] = "=",
  [anon_sym_PLUS_PLUS] = "++",
  [anon_sym_DASH_DASH] = "--",
  [anon_sym_PLUS_EQ] = "+=",
  [anon_sym_DASH_EQ] = "-=",
  [anon_sym_STAR_EQ] = "*=",
  [anon_sym_SLASH_EQ] = "/=",
  [anon_sym_PERCENT_EQ] = "%=",
  [anon_sym_STAR_STAR_EQ] = "**=",
  [anon_sym_LT_LT_EQ] = "<<=",
  [anon_sym_GT_GT_EQ] = ">>=",
  [anon_sym_AMP_EQ] = "&=",
  [anon_sym_CARET_EQ] = "^=",
  [anon_sym_PIPE_EQ] = "|=",
  [anon_sym_PIPE_PIPE] = "||",
  [anon_sym_DASHo] = "-o",
  [anon_sym_AMP_AMP] = "&&",
  [anon_sym_DASHa] = "-a",
  [anon_sym_PIPE] = "|",
  [anon_sym_CARET] = "^",
  [anon_sym_AMP] = "&",
  [anon_sym_EQ_EQ] = "==",
  [anon_sym_BANG_EQ] = "!=",
  [anon_sym_LT] = "<",
  [anon_sym_GT] = ">",
  [anon_sym_LT_EQ] = "<=",
  [anon_sym_GT_EQ] = ">=",
  [anon_sym_LT_LT] = "<<",
  [anon_sym_GT_GT] = ">>",
  [anon_sym_PLUS] = "+",
  [anon_sym_DASH] = "-",
  [anon_sym_STAR] = "*",
  [anon_sym_SLASH] = "/",
  [anon_sym_PERCENT] = "%",
  [anon_sym_STAR_STAR] = "**",
  [anon_sym_LPAREN] = "(",
  [anon_sym_RPAREN] = ")",
  [aux_sym__c_word_token1] = "word",
  [anon_sym_while] = "while",
  [anon_sym_until] = "until",
  [anon_sym_do] = "do",
  [anon_sym_done] = "done",
  [anon_sym_if] = "if",
  [anon_sym_then] = "then",
  [anon_sym_fi] = "fi",
  [anon_sym_elif] = "elif",
  [anon_sym_else] = "else",
  [anon_sym_case] = "case",
  [anon_sym_esac] = "esac",
  [anon_sym_SEMI_SEMI] = ";;",
  [anon_sym_SEMI_AMP] = ";&",
  [anon_sym_SEMI_SEMI_AMP] = ";;&",
  [anon_sym_function] = "function",
  [anon_sym_LBRACE] = "{",
  [anon_sym_RBRACE] = "}",
  [anon_sym_PIPE_AMP] = "|&",
  [anon_sym_BANG] = "!",
  [anon_sym_LBRACK] = "[",
  [anon_sym_RBRACK] = "]",
  [anon_sym_LBRACK_LBRACK] = "[[",
  [anon_sym_RBRACK_RBRACK] = "]]",
  [anon_sym_declare] = "declare",
  [anon_sym_typeset] = "typeset",
  [anon_sym_export] = "export",
  [anon_sym_readonly] = "readonly",
  [anon_sym_local] = "local",
  [anon_sym_unset] = "unset",
  [anon_sym_unsetenv] = "unsetenv",
  [anon_sym_EQ_TILDE] = "=~",
  [anon_sym_AMP_GT] = "&>",
  [anon_sym_AMP_GT_GT] = "&>>",
  [anon_sym_LT_AMP] = "<&",
  [anon_sym_GT_AMP] = ">&",
  [anon_sym_GT_PIPE] = ">|",
  [anon_sym_LT_AMP_DASH] = "<&-",
  [anon_sym_GT_AMP_DASH] = ">&-",
  [anon_sym_LT_LT_DASH] = "<<-",
  [aux_sym_heredoc_redirect_token1] = "heredoc_redirect_token1",
  [anon_sym_LT_LT_LT] = "<<<",
  [anon_sym_QMARK] = "\?",
  [anon_sym_COLON] = ":",
  [anon_sym_PLUS_PLUS2] = "++",
  [anon_sym_DASH_DASH2] = "--",
  [anon_sym_DASH2] = "-",
  [anon_sym_PLUS2] = "+",
  [anon_sym_TILDE] = "~",
  [anon_sym_DOLLAR_LPAREN_LPAREN] = "$((",
  [anon_sym_DOLLAR_LBRACK] = "$[",
  [aux_sym_brace_expression_token1] = "number",
  [anon_sym_DOT_DOT] = "..",
  [anon_sym_RBRACE2] = "}",
  [aux_sym_concatenation_token1] = "``",
  [anon_sym_DOLLAR] = "$",
  [sym__special_character] = "_special_character",
  [anon_sym_DQUOTE] = "\"",
  [sym_string_content] = "string_content",
  [sym_raw_string] = "raw_string",
  [sym_ansi_c_string] = "ansi_c_string",
  [aux_sym_number_token1] = "number_token1",
  [aux_sym_number_token2] = "number_token2",
  [anon_sym_POUND] = "#",
  [anon_sym_DOLLAR_LBRACE] = "${",
  [anon_sym_RBRACE3] = "}",
  [anon_sym_BANG2] = "!",
  [anon_sym_AT] = "@",
  [anon_sym_STAR2] = "*",
  [anon_sym_POUND2] = "#",
  [anon_sym_EQ2] = "=",
  [anon_sym_COLON_EQ] = ":=",
  [anon_sym_DASH3] = "-",
  [anon_sym_COLON_DASH] = ":-",
  [anon_sym_PLUS3] = "+",
  [anon_sym_COLON_PLUS] = ":+",
  [anon_sym_QMARK2] = "\?",
  [anon_sym_COLON_QMARK] = ":\?",
  [anon_sym_PERCENT_PERCENT] = "%%",
  [aux_sym__expansion_regex_token1] = "regex",
  [anon_sym_SLASH_SLASH] = "//",
  [anon_sym_SLASH_POUND] = "/#",
  [anon_sym_SLASH_PERCENT] = "/%",
  [anon_sym_COMMA_COMMA] = ",,",
  [anon_sym_CARET_CARET] = "^^",
  [anon_sym_U] = "U",
  [anon_sym_u] = "u",
  [anon_sym_L] = "L",
  [anon_sym_Q] = "Q",
  [anon_sym_E] = "E",
  [anon_sym_P] = "P",
  [anon_sym_A] = "A",
  [anon_sym_K] = "K",
  [anon_sym_a] = "a",
  [anon_sym_k] = "k",
  [anon_sym_DOLLAR_LPAREN] = "$(",
  [anon_sym_BQUOTE] = "`",
  [anon_sym_DOLLAR_BQUOTE] = "$`",
  [anon_sym_LT_LPAREN] = "<(",
  [anon_sym_GT_LPAREN] = ">(",
  [sym_comment] = "comment",
  [sym__comment_word] = "word",
  [aux_sym__simple_variable_name_token1] = "variable_name",
  [aux_sym__multiline_variable_name_token1] = "variable_name",
  [anon_sym_AT2] = "special_variable_name",
  [anon_sym_0] = "special_variable_name",
  [anon_sym__] = "special_variable_name",
  [sym_heredoc_start] = "heredoc_start",
  [sym_simple_heredoc_body] = "heredoc_body",
  [sym__heredoc_body_beginning] = "_heredoc_body_beginning",
  [sym_heredoc_content] = "heredoc_content",
  [sym_heredoc_end] = "heredoc_end",
  [sym_file_descriptor] = "file_descriptor",
  [sym__empty_value] = "_empty_value",
  [sym__concat] = "_concat",
  [sym_variable_name] = "variable_name",
  [sym_test_operator] = "test_operator",
  [sym_regex] = "regex",
  [sym__regex_no_slash] = "regex",
  [sym__regex_no_space] = "regex",
  [sym__expansion_word] = "word",
  [sym_extglob_pattern] = "extglob_pattern",
  [sym__bare_dollar] = "$",
  [sym__brace_start] = "{",
  [sym__immediate_double_hash] = "##",
  [sym__external_expansion_sym_hash] = "#",
  [sym__external_expansion_sym_bang] = "!",
  [sym__external_expansion_sym_equal] = "=",
  [sym___error_recovery] = "__error_recovery",
  [sym_program] = "program",
  [sym__statements] = "_statements",
  [aux_sym__terminated_statement] = "_terminated_statement",
  [sym__statement_not_pipeline] = "_statement_not_pipeline",
  [sym_redirected_statement] = "redirected_statement",
  [sym_for_statement] = "for_statement",
  [sym_c_style_for_statement] = "c_style_for_statement",
  [sym__for_body] = "_for_body",
  [sym__c_expression] = "_c_expression",
  [sym__c_expression_not_assignment] = "_c_expression_not_assignment",
  [sym__c_variable_assignment] = "variable_assignment",
  [sym__c_unary_expression] = "unary_expression",
  [sym__c_binary_expression] = "binary_expression",
  [sym__c_postfix_expression] = "postfix_expression",
  [sym__c_parenthesized_expression] = "parenthesized_expression",
  [sym_while_statement] = "while_statement",
  [sym_do_group] = "do_group",
  [sym_if_statement] = "if_statement",
  [sym_elif_clause] = "elif_clause",
  [sym_else_clause] = "else_clause",
  [sym_case_statement] = "case_statement",
  [sym_case_item] = "case_item",
  [sym_last_case_item] = "case_item",
  [sym_function_definition] = "function_definition",
  [sym_compound_statement] = "compound_statement",
  [sym_subshell] = "subshell",
  [sym_pipeline] = "pipeline",
  [sym_list] = "list",
  [sym_negated_command] = "negated_command",
  [sym_test_command] = "test_command",
  [sym__test_command_binary_expression] = "binary_expression",
  [sym_declaration_command] = "declaration_command",
  [sym_unset_command] = "unset_command",
  [sym_command] = "command",
  [sym_command_name] = "command_name",
  [sym_variable_assignment] = "variable_assignment",
  [sym_variable_assignments] = "variable_assignments",
  [sym_subscript] = "subscript",
  [sym_file_redirect] = "file_redirect",
  [sym_heredoc_redirect] = "heredoc_redirect",
  [sym__heredoc_pipeline] = "pipeline",
  [sym__heredoc_expression] = "_heredoc_expression",
  [aux_sym__heredoc_command] = "_heredoc_command",
  [sym__heredoc_body] = "_heredoc_body",
  [sym_heredoc_body] = "heredoc_body",
  [sym__simple_heredoc_body] = "_simple_heredoc_body",
  [sym_herestring_redirect] = "herestring_redirect",
  [sym__expression] = "_expression",
  [sym_binary_expression] = "binary_expression",
  [sym_ternary_expression] = "ternary_expression",
  [sym_unary_expression] = "unary_expression",
  [sym_postfix_expression] = "postfix_expression",
  [sym_parenthesized_expression] = "parenthesized_expression",
  [sym_arithmetic_expansion] = "arithmetic_expansion",
  [sym_brace_expression] = "brace_expression",
  [sym__arithmetic_expression] = "_arithmetic_expression",
  [sym__arithmetic_literal] = "_arithmetic_literal",
  [sym__arithmetic_binary_expression] = "binary_expression",
  [sym__arithmetic_ternary_expression] = "ternary_expression",
  [sym__arithmetic_unary_expression] = "unary_expression",
  [sym__arithmetic_postfix_expression] = "postfix_expression",
  [sym__arithmetic_parenthesized_expression] = "parenthesized_expression",
  [sym_concatenation] = "concatenation",
  [sym_string] = "string",
  [sym_translated_string] = "translated_string",
  [sym_array] = "array",
  [sym_number] = "number",
  [sym_simple_expansion] = "simple_expansion",
  [sym_expansion] = "expansion",
  [sym__expansion_body] = "_expansion_body",
  [sym__expansion_expression] = "_expansion_expression",
  [sym__expansion_regex] = "_expansion_regex",
  [sym__expansion_regex_replacement] = "_expansion_regex_replacement",
  [sym__expansion_regex_removal] = "_expansion_regex_removal",
  [sym__expansion_max_length] = "_expansion_max_length",
  [sym__expansion_max_length_expression] = "_expansion_max_length_expression",
  [sym__expansion_max_length_binary_expression] = "binary_expression",
  [sym__expansion_operator] = "_expansion_operator",
  [sym__concatenation_in_expansion] = "concatenation",
  [sym_command_substitution] = "command_substitution",
  [sym_process_substitution] = "process_substitution",
  [sym__extglob_blob] = "_extglob_blob",
  [sym__c_terminator] = "_c_terminator",
  [aux_sym__statements_repeat1] = "_statements_repeat1",
  [aux_sym_redirected_statement_repeat1] = "redirected_statement_repeat1",
  [aux_sym_redirected_statement_repeat2] = "redirected_statement_repeat2",
  [aux_sym_for_statement_repeat1] = "for_statement_repeat1",
  [aux_sym__for_body_repeat1] = "_for_body_repeat1",
  [aux_sym_if_statement_repeat1] = "if_statement_repeat1",
  [aux_sym_case_statement_repeat1] = "case_statement_repeat1",
  [aux_sym_case_item_repeat1] = "case_item_repeat1",
  [aux_sym_pipeline_repeat1] = "pipeline_repeat1",
  [aux_sym_declaration_command_repeat1] = "declaration_command_repeat1",
  [aux_sym_unset_command_repeat1] = "unset_command_repeat1",
  [aux_sym_command_repeat1] = "command_repeat1",
  [aux_sym_command_repeat2] = "command_repeat2",
  [aux_sym_variable_assignments_repeat1] = "variable_assignments_repeat1",
  [aux_sym_heredoc_body_repeat1] = "heredoc_body_repeat1",
  [aux_sym__literal_repeat1] = "_literal_repeat1",
  [aux_sym_arithmetic_expansion_repeat1] = "arithmetic_expansion_repeat1",
  [aux_sym_concatenation_repeat1] = "concatenation_repeat1",
  [aux_sym_string_repeat1] = "string_repeat1",
  [aux_sym__expansion_body_repeat1] = "_expansion_body_repeat1",
  [aux_sym__expansion_regex_repeat1] = "_expansion_regex_repeat1",
  [aux_sym__concatenation_in_expansion_repeat1] = "_concatenation_in_expansion_repeat1",
};

static const TSSymbol ts_symbol_map[] = {
  [ts_builtin_sym_end] = ts_builtin_sym_end,
  [sym_word] = sym_word,
  [anon_sym_for] = anon_sym_for,
  [anon_sym_select] = anon_sym_select,
  [anon_sym_in] = anon_sym_in,
  [anon_sym_LPAREN_LPAREN] = anon_sym_LPAREN_LPAREN,
  [anon_sym_RPAREN_RPAREN] = anon_sym_RPAREN_RPAREN,
  [anon_sym_SEMI] = anon_sym_SEMI,
  [anon_sym_COMMA] = anon_sym_COMMA,
  [anon_sym_EQ] = anon_sym_EQ,
  [anon_sym_PLUS_PLUS] = anon_sym_PLUS_PLUS,
  [anon_sym_DASH_DASH] = anon_sym_DASH_DASH,
  [anon_sym_PLUS_EQ] = anon_sym_PLUS_EQ,
  [anon_sym_DASH_EQ] = anon_sym_DASH_EQ,
  [anon_sym_STAR_EQ] = anon_sym_STAR_EQ,
  [anon_sym_SLASH_EQ] = anon_sym_SLASH_EQ,
  [anon_sym_PERCENT_EQ] = anon_sym_PERCENT_EQ,
  [anon_sym_STAR_STAR_EQ] = anon_sym_STAR_STAR_EQ,
  [anon_sym_LT_LT_EQ] = anon_sym_LT_LT_EQ,
  [anon_sym_GT_GT_EQ] = anon_sym_GT_GT_EQ,
  [anon_sym_AMP_EQ] = anon_sym_AMP_EQ,
  [anon_sym_CARET_EQ] = anon_sym_CARET_EQ,
  [anon_sym_PIPE_EQ] = anon_sym_PIPE_EQ,
  [anon_sym_PIPE_PIPE] = anon_sym_PIPE_PIPE,
  [anon_sym_DASHo] = anon_sym_DASHo,
  [anon_sym_AMP_AMP] = anon_sym_AMP_AMP,
  [anon_sym_DASHa] = anon_sym_DASHa,
  [anon_sym_PIPE] = anon_sym_PIPE,
  [anon_sym_CARET] = anon_sym_CARET,
  [anon_sym_AMP] = anon_sym_AMP,
  [anon_sym_EQ_EQ] = anon_sym_EQ_EQ,
  [anon_sym_BANG_EQ] = anon_sym_BANG_EQ,
  [anon_sym_LT] = anon_sym_LT,
  [anon_sym_GT] = anon_sym_GT,
  [anon_sym_LT_EQ] = anon_sym_LT_EQ,
  [anon_sym_GT_EQ] = anon_sym_GT_EQ,
  [anon_sym_LT_LT] = anon_sym_LT_LT,
  [anon_sym_GT_GT] = anon_sym_GT_GT,
  [anon_sym_PLUS] = anon_sym_PLUS,
  [anon_sym_DASH] = anon_sym_DASH,
  [anon_sym_STAR] = anon_sym_STAR,
  [anon_sym_SLASH] = anon_sym_SLASH,
  [anon_sym_PERCENT] = anon_sym_PERCENT,
  [anon_sym_STAR_STAR] = anon_sym_STAR_STAR,
  [anon_sym_LPAREN] = anon_sym_LPAREN,
  [anon_sym_RPAREN] = anon_sym_RPAREN,
  [aux_sym__c_word_token1] = sym_word,
  [anon_sym_while] = anon_sym_while,
  [anon_sym_until] = anon_sym_until,
  [anon_sym_do] = anon_sym_do,
  [anon_sym_done] = anon_sym_done,
  [anon_sym_if] = anon_sym_if,
  [anon_sym_then] = anon_sym_then,
  [anon_sym_fi] = anon_sym_fi,
  [anon_sym_elif] = anon_sym_elif,
  [anon_sym_else] = anon_sym_else,
  [anon_sym_case] = anon_sym_case,
  [anon_sym_esac] = anon_sym_esac,
  [anon_sym_SEMI_SEMI] = anon_sym_SEMI_SEMI,
  [anon_sym_SEMI_AMP] = anon_sym_SEMI_AMP,
  [anon_sym_SEMI_SEMI_AMP] = anon_sym_SEMI_SEMI_AMP,
  [anon_sym_function] = anon_sym_function,
  [anon_sym_LBRACE] = anon_sym_LBRACE,
  [anon_sym_RBRACE] = anon_sym_RBRACE,
  [anon_sym_PIPE_AMP] = anon_sym_PIPE_AMP,
  [anon_sym_BANG] = anon_sym_BANG,
  [anon_sym_LBRACK] = anon_sym_LBRACK,
  [anon_sym_RBRACK] = anon_sym_RBRACK,
  [anon_sym_LBRACK_LBRACK] = anon_sym_LBRACK_LBRACK,
  [anon_sym_RBRACK_RBRACK] = anon_sym_RBRACK_RBRACK,
  [anon_sym_declare] = anon_sym_declare,
  [anon_sym_typeset] = anon_sym_typeset,
  [anon_sym_export] = anon_sym_export,
  [anon_sym_readonly] = anon_sym_readonly,
  [anon_sym_local] = anon_sym_local,
  [anon_sym_unset] = anon_sym_unset,
  [anon_sym_unsetenv] = anon_sym_unsetenv,
  [anon_sym_EQ_TILDE] = anon_sym_EQ_TILDE,
  [anon_sym_AMP_GT] = anon_sym_AMP_GT,
  [anon_sym_AMP_GT_GT] = anon_sym_AMP_GT_GT,
  [anon_sym_LT_AMP] = anon_sym_LT_AMP,
  [anon_sym_GT_AMP] = anon_sym_GT_AMP,
  [anon_sym_GT_PIPE] = anon_sym_GT_PIPE,
  [anon_sym_LT_AMP_DASH] = anon_sym_LT_AMP_DASH,
  [anon_sym_GT_AMP_DASH] = anon_sym_GT_AMP_DASH,
  [anon_sym_LT_LT_DASH] = anon_sym_LT_LT_DASH,
  [aux_sym_heredoc_redirect_token1] = aux_sym_heredoc_redirect_token1,
  [anon_sym_LT_LT_LT] = anon_sym_LT_LT_LT,
  [anon_sym_QMARK] = anon_sym_QMARK,
  [anon_sym_COLON] = anon_sym_COLON,
  [anon_sym_PLUS_PLUS2] = anon_sym_PLUS_PLUS,
  [anon_sym_DASH_DASH2] = anon_sym_DASH_DASH,
  [anon_sym_DASH2] = anon_sym_DASH,
  [anon_sym_PLUS2] = anon_sym_PLUS,
  [anon_sym_TILDE] = anon_sym_TILDE,
  [anon_sym_DOLLAR_LPAREN_LPAREN] = anon_sym_DOLLAR_LPAREN_LPAREN,
  [anon_sym_DOLLAR_LBRACK] = anon_sym_DOLLAR_LBRACK,
  [aux_sym_brace_expression_token1] = sym_number,
  [anon_sym_DOT_DOT] = anon_sym_DOT_DOT,
  [anon_sym_RBRACE2] = anon_sym_RBRACE,
  [aux_sym_concatenation_token1] = aux_sym_concatenation_token1,
  [anon_sym_DOLLAR] = anon_sym_DOLLAR,
  [sym__special_character] = sym__special_character,
  [anon_sym_DQUOTE] = anon_sym_DQUOTE,
  [sym_string_content] = sym_string_content,
  [sym_raw_string] = sym_raw_string,
  [sym_ansi_c_string] = sym_ansi_c_string,
  [aux_sym_number_token1] = aux_sym_number_token1,
  [aux_sym_number_token2] = aux_sym_number_token2,
  [anon_sym_POUND] = anon_sym_POUND,
  [anon_sym_DOLLAR_LBRACE] = anon_sym_DOLLAR_LBRACE,
  [anon_sym_RBRACE3] = anon_sym_RBRACE,
  [anon_sym_BANG2] = anon_sym_BANG,
  [anon_sym_AT] = anon_sym_AT,
  [anon_sym_STAR2] = anon_sym_STAR,
  [anon_sym_POUND2] = anon_sym_POUND,
  [anon_sym_EQ2] = anon_sym_EQ,
  [anon_sym_COLON_EQ] = anon_sym_COLON_EQ,
  [anon_sym_DASH3] = anon_sym_DASH,
  [anon_sym_COLON_DASH] = anon_sym_COLON_DASH,
  [anon_sym_PLUS3] = anon_sym_PLUS,
  [anon_sym_COLON_PLUS] = anon_sym_COLON_PLUS,
  [anon_sym_QMARK2] = anon_sym_QMARK,
  [anon_sym_COLON_QMARK] = anon_sym_COLON_QMARK,
  [anon_sym_PERCENT_PERCENT] = anon_sym_PERCENT_PERCENT,
  [aux_sym__expansion_regex_token1] = sym_regex,
  [anon_sym_SLASH_SLASH] = anon_sym_SLASH_SLASH,
  [anon_sym_SLASH_POUND] = anon_sym_SLASH_POUND,
  [anon_sym_SLASH_PERCENT] = anon_sym_SLASH_PERCENT,
  [anon_sym_COMMA_COMMA] = anon_sym_COMMA_COMMA,
  [anon_sym_CARET_CARET] = anon_sym_CARET_CARET,
  [anon_sym_U] = anon_sym_U,
  [anon_sym_u] = anon_sym_u,
  [anon_sym_L] = anon_sym_L,
  [anon_sym_Q] = anon_sym_Q,
  [anon_sym_E] = anon_sym_E,
  [anon_sym_P] = anon_sym_P,
  [anon_sym_A] = anon_sym_A,
  [anon_sym_K] = anon_sym_K,
  [anon_sym_a] = anon_sym_a,
  [anon_sym_k] = anon_sym_k,
  [anon_sym_DOLLAR_LPAREN] = anon_sym_DOLLAR_LPAREN,
  [anon_sym_BQUOTE] = anon_sym_BQUOTE,
  [anon_sym_DOLLAR_BQUOTE] = anon_sym_DOLLAR_BQUOTE,
  [anon_sym_LT_LPAREN] = anon_sym_LT_LPAREN,
  [anon_sym_GT_LPAREN] = anon_sym_GT_LPAREN,
  [sym_comment] = sym_comment,
  [sym__comment_word] = sym_word,
  [aux_sym__simple_variable_name_token1] = sym_variable_name,
  [aux_sym__multiline_variable_name_token1] = sym_variable_name,
  [anon_sym_AT2] = anon_sym_AT2,
  [anon_sym_0] = anon_sym_AT2,
  [anon_sym__] = anon_sym_AT2,
  [sym_heredoc_start] = sym_heredoc_start,
  [sym_simple_heredoc_body] = sym_heredoc_body,
  [sym__heredoc_body_beginning] = sym__heredoc_body_beginning,
  [sym_heredoc_content] = sym_heredoc_content,
  [sym_heredoc_end] = sym_heredoc_end,
  [sym_file_descriptor] = sym_file_descriptor,
  [sym__empty_value] = sym__empty_value,
  [sym__concat] = sym__concat,
  [sym_variable_name] = sym_variable_name,
  [sym_test_operator] = sym_test_operator,
  [sym_regex] = sym_regex,
  [sym__regex_no_slash] = sym_regex,
  [sym__regex_no_space] = sym_regex,
  [sym__expansion_word] = sym_word,
  [sym_extglob_pattern] = sym_extglob_pattern,
  [sym__bare_dollar] = anon_sym_DOLLAR,
  [sym__brace_start] = anon_sym_LBRACE,
  [sym__immediate_double_hash] = sym__immediate_double_hash,
  [sym__external_expansion_sym_hash] = anon_sym_POUND,
  [sym__external_expansion_sym_bang] = anon_sym_BANG,
  [sym__external_expansion_sym_equal] = anon_sym_EQ,
  [sym___error_recovery] = sym___error_recovery,
  [sym_program] = sym_program,
  [sym__statements] = sym__statements,
  [aux_sym__terminated_statement] = aux_sym__terminated_statement,
  [sym__statement_not_pipeline] = sym__statement_not_pipeline,
  [sym_redirected_statement] = sym_redirected_statement,
  [sym_for_statement] = sym_for_statement,
  [sym_c_style_for_statement] = sym_c_style_for_statement,
  [sym__for_body] = sym__for_body,
  [sym__c_expression] = sym__c_expression,
  [sym__c_expression_not_assignment] = sym__c_expression_not_assignment,
  [sym__c_variable_assignment] = sym_variable_assignment,
  [sym__c_unary_expression] = sym_unary_expression,
  [sym__c_binary_expression] = sym_binary_expression,
  [sym__c_postfix_expression] = sym_postfix_expression,
  [sym__c_parenthesized_expression] = sym_parenthesized_expression,
  [sym_while_statement] = sym_while_statement,
  [sym_do_group] = sym_do_group,
  [sym_if_statement] = sym_if_statement,
  [sym_elif_clause] = sym_elif_clause,
  [sym_else_clause] = sym_else_clause,
  [sym_case_statement] = sym_case_statement,
  [sym_case_item] = sym_case_item,
  [sym_last_case_item] = sym_case_item,
  [sym_function_definition] = sym_function_definition,
  [sym_compound_statement] = sym_compound_statement,
  [sym_subshell] = sym_subshell,
  [sym_pipeline] = sym_pipeline,
  [sym_list] = sym_list,
  [sym_negated_command] = sym_negated_command,
  [sym_test_command] = sym_test_command,
  [sym__test_command_binary_expression] = sym_binary_expression,
  [sym_declaration_command] = sym_declaration_command,
  [sym_unset_command] = sym_unset_command,
  [sym_command] = sym_command,
  [sym_command_name] = sym_command_name,
  [sym_variable_assignment] = sym_variable_assignment,
  [sym_variable_assignments] = sym_variable_assignments,
  [sym_subscript] = sym_subscript,
  [sym_file_redirect] = sym_file_redirect,
  [sym_heredoc_redirect] = sym_heredoc_redirect,
  [sym__heredoc_pipeline] = sym_pipeline,
  [sym__heredoc_expression] = sym__heredoc_expression,
  [aux_sym__heredoc_command] = aux_sym__heredoc_command,
  [sym__heredoc_body] = sym__heredoc_body,
  [sym_heredoc_body] = sym_heredoc_body,
  [sym__simple_heredoc_body] = sym__simple_heredoc_body,
  [sym_herestring_redirect] = sym_herestring_redirect,
  [sym__expression] = sym__expression,
  [sym_binary_expression] = sym_binary_expression,
  [sym_ternary_expression] = sym_ternary_expression,
  [sym_unary_expression] = sym_unary_expression,
  [sym_postfix_expression] = sym_postfix_expression,
  [sym_parenthesized_expression] = sym_parenthesized_expression,
  [sym_arithmetic_expansion] = sym_arithmetic_expansion,
  [sym_brace_expression] = sym_brace_expression,
  [sym__arithmetic_expression] = sym__arithmetic_expression,
  [sym__arithmetic_literal] = sym__arithmetic_literal,
  [sym__arithmetic_binary_expression] = sym_binary_expression,
  [sym__arithmetic_ternary_expression] = sym_ternary_expression,
  [sym__arithmetic_unary_expression] = sym_unary_expression,
  [sym__arithmetic_postfix_expression] = sym_postfix_expression,
  [sym__arithmetic_parenthesized_expression] = sym_parenthesized_expression,
  [sym_concatenation] = sym_concatenation,
  [sym_string] = sym_string,
  [sym_translated_string] = sym_translated_string,
  [sym_array] = sym_array,
  [sym_number] = sym_number,
  [sym_simple_expansion] = sym_simple_expansion,
  [sym_expansion] = sym_expansion,
  [sym__expansion_body] = sym__expansion_body,
  [sym__expansion_expression] = sym__expansion_expression,
  [sym__expansion_regex] = sym__expansion_regex,
  [sym__expansion_regex_replacement] = sym__expansion_regex_replacement,
  [sym__expansion_regex_removal] = sym__expansion_regex_removal,
  [sym__expansion_max_length] = sym__expansion_max_length,
  [sym__expansion_max_length_expression] = sym__expansion_max_length_expression,
  [sym__expansion_max_length_binary_expression] = sym_binary_expression,
  [sym__expansion_operator] = sym__expansion_operator,
  [sym__concatenation_in_expansion] = sym_concatenation,
  [sym_command_substitution] = sym_command_substitution,
  [sym_process_substitution] = sym_process_substitution,
  [sym__extglob_blob] = sym__extglob_blob,
  [sym__c_terminator] = sym__c_terminator,
  [aux_sym__statements_repeat1] = aux_sym__statements_repeat1,
  [aux_sym_redirected_statement_repeat1] = aux_sym_redirected_statement_repeat1,
  [aux_sym_redirected_statement_repeat2] = aux_sym_redirected_statement_repeat2,
  [aux_sym_for_statement_repeat1] = aux_sym_for_statement_repeat1,
  [aux_sym__for_body_repeat1] = aux_sym__for_body_repeat1,
  [aux_sym_if_statement_repeat1] = aux_sym_if_statement_repeat1,
  [aux_sym_case_statement_repeat1] = aux_sym_case_statement_repeat1,
  [aux_sym_case_item_repeat1] = aux_sym_case_item_repeat1,
  [aux_sym_pipeline_repeat1] = aux_sym_pipeline_repeat1,
  [aux_sym_declaration_command_repeat1] = aux_sym_declaration_command_repeat1,
  [aux_sym_unset_command_repeat1] = aux_sym_unset_command_repeat1,
  [aux_sym_command_repeat1] = aux_sym_command_repeat1,
  [aux_sym_command_repeat2] = aux_sym_command_repeat2,
  [aux_sym_variable_assignments_repeat1] = aux_sym_variable_assignments_repeat1,
  [aux_sym_heredoc_body_repeat1] = aux_sym_heredoc_body_repeat1,
  [aux_sym__literal_repeat1] = aux_sym__literal_repeat1,
  [aux_sym_arithmetic_expansion_repeat1] = aux_sym_arithmetic_expansion_repeat1,
  [aux_sym_concatenation_repeat1] = aux_sym_concatenation_repeat1,
  [aux_sym_string_repeat1] = aux_sym_string_repeat1,
  [aux_sym__expansion_body_repeat1] = aux_sym__expansion_body_repeat1,
  [aux_sym__expansion_regex_repeat1] = aux_sym__expansion_regex_repeat1,
  [aux_sym__concatenation_in_expansion_repeat1] = aux_sym__concatenation_in_expansion_repeat1,
};

static const TSSymbolMetadata ts_symbol_metadata[] = {
  [ts_builtin_sym_end] = {
    .visible = false,
    .named = true,
  },
  [sym_word] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_for] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_select] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_in] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LPAREN_LPAREN] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_RPAREN_RPAREN] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_SEMI] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_COMMA] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PLUS_PLUS] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DASH_DASH] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PLUS_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DASH_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_STAR_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_SLASH_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PERCENT_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_STAR_STAR_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LT_LT_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_GT_GT_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_AMP_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_CARET_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PIPE_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PIPE_PIPE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DASHo] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_AMP_AMP] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DASHa] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PIPE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_CARET] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_AMP] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_EQ_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_BANG_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_GT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LT_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_GT_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LT_LT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_GT_GT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PLUS] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DASH] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_STAR] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_SLASH] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PERCENT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_STAR_STAR] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LPAREN] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_RPAREN] = {
    .visible = true,
    .named = false,
  },
  [aux_sym__c_word_token1] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_while] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_until] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_do] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_done] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_if] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_then] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_fi] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_elif] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_else] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_case] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_esac] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_SEMI_SEMI] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_SEMI_AMP] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_SEMI_SEMI_AMP] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_function] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LBRACE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_RBRACE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PIPE_AMP] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_BANG] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LBRACK] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_RBRACK] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LBRACK_LBRACK] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_RBRACK_RBRACK] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_declare] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_typeset] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_export] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_readonly] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_local] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_unset] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_unsetenv] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_EQ_TILDE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_AMP_GT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_AMP_GT_GT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LT_AMP] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_GT_AMP] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_GT_PIPE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LT_AMP_DASH] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_GT_AMP_DASH] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LT_LT_DASH] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_heredoc_redirect_token1] = {
    .visible = false,
    .named = false,
  },
  [anon_sym_LT_LT_LT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_QMARK] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_COLON] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PLUS_PLUS2] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DASH_DASH2] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DASH2] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PLUS2] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_TILDE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DOLLAR_LPAREN_LPAREN] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DOLLAR_LBRACK] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_brace_expression_token1] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_DOT_DOT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_RBRACE2] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_concatenation_token1] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DOLLAR] = {
    .visible = true,
    .named = false,
  },
  [sym__special_character] = {
    .visible = false,
    .named = true,
  },
  [anon_sym_DQUOTE] = {
    .visible = true,
    .named = false,
  },
  [sym_string_content] = {
    .visible = true,
    .named = true,
  },
  [sym_raw_string] = {
    .visible = true,
    .named = true,
  },
  [sym_ansi_c_string] = {
    .visible = true,
    .named = true,
  },
  [aux_sym_number_token1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_number_token2] = {
    .visible = false,
    .named = false,
  },
  [anon_sym_POUND] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DOLLAR_LBRACE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_RBRACE3] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_BANG2] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_AT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_STAR2] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_POUND2] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_EQ2] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_COLON_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DASH3] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_COLON_DASH] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PLUS3] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_COLON_PLUS] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_QMARK2] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_COLON_QMARK] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PERCENT_PERCENT] = {
    .visible = true,
    .named = false,
  },
  [aux_sym__expansion_regex_token1] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_SLASH_SLASH] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_SLASH_POUND] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_SLASH_PERCENT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_COMMA_COMMA] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_CARET_CARET] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_U] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_u] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_L] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_Q] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_E] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_P] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_A] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_K] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_a] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_k] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DOLLAR_LPAREN] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_BQUOTE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DOLLAR_BQUOTE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LT_LPAREN] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_GT_LPAREN] = {
    .visible = true,
    .named = false,
  },
  [sym_comment] = {
    .visible = true,
    .named = true,
  },
  [sym__comment_word] = {
    .visible = true,
    .named = true,
  },
  [aux_sym__simple_variable_name_token1] = {
    .visible = true,
    .named = true,
  },
  [aux_sym__multiline_variable_name_token1] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_AT2] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_0] = {
    .visible = true,
    .named = true,
  },
  [anon_sym__] = {
    .visible = true,
    .named = true,
  },
  [sym_heredoc_start] = {
    .visible = true,
    .named = true,
  },
  [sym_simple_heredoc_body] = {
    .visible = true,
    .named = true,
  },
  [sym__heredoc_body_beginning] = {
    .visible = false,
    .named = true,
  },
  [sym_heredoc_content] = {
    .visible = true,
    .named = true,
  },
  [sym_heredoc_end] = {
    .visible = true,
    .named = true,
  },
  [sym_file_descriptor] = {
    .visible = true,
    .named = true,
  },
  [sym__empty_value] = {
    .visible = false,
    .named = true,
  },
  [sym__concat] = {
    .visible = false,
    .named = true,
  },
  [sym_variable_name] = {
    .visible = true,
    .named = true,
  },
  [sym_test_operator] = {
    .visible = true,
    .named = true,
  },
  [sym_regex] = {
    .visible = true,
    .named = true,
  },
  [sym__regex_no_slash] = {
    .visible = true,
    .named = true,
  },
  [sym__regex_no_space] = {
    .visible = true,
    .named = true,
  },
  [sym__expansion_word] = {
    .visible = true,
    .named = true,
  },
  [sym_extglob_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym__bare_dollar] = {
    .visible = true,
    .named = false,
  },
  [sym__brace_start] = {
    .visible = true,
    .named = false,
  },
  [sym__immediate_double_hash] = {
    .visible = true,
    .named = false,
  },
  [sym__external_expansion_sym_hash] = {
    .visible = true,
    .named = false,
  },
  [sym__external_expansion_sym_bang] = {
    .visible = true,
    .named = false,
  },
  [sym__external_expansion_sym_equal] = {
    .visible = true,
    .named = false,
  },
  [sym___error_recovery] = {
    .visible = false,
    .named = true,
  },
  [sym_program] = {
    .visible = true,
    .named = true,
  },
  [sym__statements] = {
    .visible = false,
    .named = true,
  },
  [aux_sym__terminated_statement] = {
    .visible = false,
    .named = false,
  },
  [sym__statement_not_pipeline] = {
    .visible = false,
    .named = true,
  },
  [sym_redirected_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_for_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_c_style_for_statement] = {
    .visible = true,
    .named = true,
  },
  [sym__for_body] = {
    .visible = false,
    .named = true,
  },
  [sym__c_expression] = {
    .visible = false,
    .named = true,
  },
  [sym__c_expression_not_assignment] = {
    .visible = false,
    .named = true,
  },
  [sym__c_variable_assignment] = {
    .visible = true,
    .named = true,
  },
  [sym__c_unary_expression] = {
    .visible = true,
    .named = true,
  },
  [sym__c_binary_expression] = {
    .visible = true,
    .named = true,
  },
  [sym__c_postfix_expression] = {
    .visible = true,
    .named = true,
  },
  [sym__c_parenthesized_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_while_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_do_group] = {
    .visible = true,
    .named = true,
  },
  [sym_if_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_elif_clause] = {
    .visible = true,
    .named = true,
  },
  [sym_else_clause] = {
    .visible = true,
    .named = true,
  },
  [sym_case_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_case_item] = {
    .visible = true,
    .named = true,
  },
  [sym_last_case_item] = {
    .visible = true,
    .named = true,
  },
  [sym_function_definition] = {
    .visible = true,
    .named = true,
  },
  [sym_compound_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_subshell] = {
    .visible = true,
    .named = true,
  },
  [sym_pipeline] = {
    .visible = true,
    .named = true,
  },
  [sym_list] = {
    .visible = true,
    .named = true,
  },
  [sym_negated_command] = {
    .visible = true,
    .named = true,
  },
  [sym_test_command] = {
    .visible = true,
    .named = true,
  },
  [sym__test_command_binary_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_declaration_command] = {
    .visible = true,
    .named = true,
  },
  [sym_unset_command] = {
    .visible = true,
    .named = true,
  },
  [sym_command] = {
    .visible = true,
    .named = true,
  },
  [sym_command_name] = {
    .visible = true,
    .named = true,
  },
  [sym_variable_assignment] = {
    .visible = true,
    .named = true,
  },
  [sym_variable_assignments] = {
    .visible = true,
    .named = true,
  },
  [sym_subscript] = {
    .visible = true,
    .named = true,
  },
  [sym_file_redirect] = {
    .visible = true,
    .named = true,
  },
  [sym_heredoc_redirect] = {
    .visible = true,
    .named = true,
  },
  [sym__heredoc_pipeline] = {
    .visible = true,
    .named = true,
  },
  [sym__heredoc_expression] = {
    .visible = false,
    .named = true,
  },
  [aux_sym__heredoc_command] = {
    .visible = false,
    .named = false,
  },
  [sym__heredoc_body] = {
    .visible = false,
    .named = true,
  },
  [sym_heredoc_body] = {
    .visible = true,
    .named = true,
  },
  [sym__simple_heredoc_body] = {
    .visible = false,
    .named = true,
  },
  [sym_herestring_redirect] = {
    .visible = true,
    .named = true,
  },
  [sym__expression] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym_binary_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_ternary_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_unary_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_postfix_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_parenthesized_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_arithmetic_expansion] = {
    .visible = true,
    .named = true,
  },
  [sym_brace_expression] = {
    .visible = true,
    .named = true,
  },
  [sym__arithmetic_expression] = {
    .visible = false,
    .named = true,
  },
  [sym__arithmetic_literal] = {
    .visible = false,
    .named = true,
  },
  [sym__arithmetic_binary_expression] = {
    .visible = true,
    .named = true,
  },
  [sym__arithmetic_ternary_expression] = {
    .visible = true,
    .named = true,
  },
  [sym__arithmetic_unary_expression] = {
    .visible = true,
    .named = true,
  },
  [sym__arithmetic_postfix_expression] = {
    .visible = true,
    .named = true,
  },
  [sym__arithmetic_parenthesized_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_concatenation] = {
    .visible = true,
    .named = true,
  },
  [sym_string] = {
    .visible = true,
    .named = true,
  },
  [sym_translated_string] = {
    .visible = true,
    .named = true,
  },
  [sym_array] = {
    .visible = true,
    .named = true,
  },
  [sym_number] = {
    .visible = true,
    .named = true,
  },
  [sym_simple_expansion] = {
    .visible = true,
    .named = true,
  },
  [sym_expansion] = {
    .visible = true,
    .named = true,
  },
  [sym__expansion_body] = {
    .visible = false,
    .named = true,
  },
  [sym__expansion_expression] = {
    .visible = false,
    .named = true,
  },
  [sym__expansion_regex] = {
    .visible = false,
    .named = true,
  },
  [sym__expansion_regex_replacement] = {
    .visible = false,
    .named = true,
  },
  [sym__expansion_regex_removal] = {
    .visible = false,
    .named = true,
  },
  [sym__expansion_max_length] = {
    .visible = false,
    .named = true,
  },
  [sym__expansion_max_length_expression] = {
    .visible = false,
    .named = true,
  },
  [sym__expansion_max_length_binary_expression] = {
    .visible = true,
    .named = true,
  },
  [sym__expansion_operator] = {
    .visible = false,
    .named = true,
  },
  [sym__concatenation_in_expansion] = {
    .visible = true,
    .named = true,
  },
  [sym_command_substitution] = {
    .visible = true,
    .named = true,
  },
  [sym_process_substitution] = {
    .visible = true,
    .named = true,
  },
  [sym__extglob_blob] = {
    .visible = false,
    .named = true,
  },
  [sym__c_terminator] = {
    .visible = false,
    .named = true,
  },
  [aux_sym__statements_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_redirected_statement_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_redirected_statement_repeat2] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_for_statement_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__for_body_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_if_statement_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_case_statement_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_case_item_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_pipeline_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_declaration_command_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_unset_command_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_command_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_command_repeat2] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_variable_assignments_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_heredoc_body_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__literal_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_arithmetic_expansion_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_concatenation_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_string_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__expansion_body_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__expansion_regex_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__concatenation_in_expansion_repeat1] = {
    .visible = false,
    .named = false,
  },
};

enum ts_field_identifiers {
  field_alternative = 1,
  field_argument = 2,
  field_body = 3,
  field_condition = 4,
  field_consequence = 5,
  field_descriptor = 6,
  field_destination = 7,
  field_fallthrough = 8,
  field_index = 9,
  field_initializer = 10,
  field_left = 11,
  field_name = 12,
  field_operator = 13,
  field_redirect = 14,
  field_right = 15,
  field_termination = 16,
  field_update = 17,
  field_value = 18,
  field_variable = 19,
};

static const char * const ts_field_names[] = {
  [0] = NULL,
  [field_alternative] = "alternative",
  [field_argument] = "argument",
  [field_body] = "body",
  [field_condition] = "condition",
  [field_consequence] = "consequence",
  [field_descriptor] = "descriptor",
  [field_destination] = "destination",
  [field_fallthrough] = "fallthrough",
  [field_index] = "index",
  [field_initializer] = "initializer",
  [field_left] = "left",
  [field_name] = "name",
  [field_operator] = "operator",
  [field_redirect] = "redirect",
  [field_right] = "right",
  [field_termination] = "termination",
  [field_update] = "update",
  [field_value] = "value",
  [field_variable] = "variable",
};

static const TSFieldMapSlice ts_field_map_slices[PRODUCTION_ID_COUNT] = {
  [2] = {.index = 0, .length = 1},
  [3] = {.index = 1, .length = 1},
  [4] = {.index = 2, .length = 1},
  [5] = {.index = 3, .length = 3},
  [6] = {.index = 6, .length = 3},
  [7] = {.index = 9, .length = 1},
  [8] = {.index = 10, .length = 1},
  [9] = {.index = 10, .length = 1},
  [12] = {.index = 11, .length = 1},
  [14] = {.index = 12, .length = 2},
  [15] = {.index = 14, .length = 1},
  [16] = {.index = 15, .length = 1},
  [17] = {.index = 15, .length = 1},
  [18] = {.index = 16, .length = 3},
  [19] = {.index = 19, .length = 2},
  [20] = {.index = 21, .length = 2},
  [21] = {.index = 23, .length = 2},
  [22] = {.index = 23, .length = 2},
  [23] = {.index = 2, .length = 1},
  [24] = {.index = 25, .length = 2},
  [25] = {.index = 25, .length = 2},
  [26] = {.index = 27, .length = 2},
  [27] = {.index = 29, .length = 1},
  [28] = {.index = 30, .length = 2},
  [29] = {.index = 32, .length = 2},
  [30] = {.index = 34, .length = 3},
  [31] = {.index = 37, .length = 1},
  [32] = {.index = 37, .length = 1},
  [33] = {.index = 11, .length = 1},
  [34] = {.index = 38, .length = 2},
  [35] = {.index = 40, .length = 1},
  [36] = {.index = 41, .length = 2},
  [37] = {.index = 41, .length = 2},
  [38] = {.index = 43, .length = 4},
  [39] = {.index = 47, .length = 4},
  [40] = {.index = 51, .length = 2},
  [41] = {.index = 51, .length = 2},
  [42] = {.index = 53, .length = 2},
  [43] = {.index = 55, .length = 3},
  [44] = {.index = 58, .length = 1},
  [45] = {.index = 59, .length = 1},
  [46] = {.index = 59, .length = 1},
  [47] = {.index = 60, .length = 3},
  [49] = {.index = 63, .length = 2},
  [50] = {.index = 65, .length = 2},
  [51] = {.index = 67, .length = 2},
  [52] = {.index = 69, .length = 2},
  [53] = {.index = 69, .length = 2},
  [54] = {.index = 71, .length = 2},
  [55] = {.index = 25, .length = 2},
  [56] = {.index = 73, .length = 4},
  [57] = {.index = 77, .length = 1},
  [58] = {.index = 78, .length = 1},
  [59] = {.index = 79, .length = 1},
  [60] = {.index = 79, .length = 1},
  [61] = {.index = 80, .length = 2},
  [62] = {.index = 63, .length = 2},
  [63] = {.index = 82, .length = 3},
  [64] = {.index = 85, .length = 3},
  [65] = {.index = 88, .length = 3},
  [66] = {.index = 91, .length = 2},
  [67] = {.index = 93, .length = 2},
  [68] = {.index = 95, .length = 4},
  [69] = {.index = 99, .length = 2},
  [70] = {.index = 101, .length = 2},
  [71] = {.index = 103, .length = 2},
  [72] = {.index = 105, .length = 2},
  [73] = {.index = 107, .length = 2},
  [74] = {.index = 109, .length = 2},
  [75] = {.index = 111, .length = 3},
  [76] = {.index = 114, .length = 3},
  [77] = {.index = 117, .length = 2},
  [78] = {.index = 119, .length = 2},
  [79] = {.index = 121, .length = 2},
  [80] = {.index = 123, .length = 2},
  [81] = {.index = 117, .length = 2},
  [82] = {.index = 119, .length = 2},
  [83] = {.index = 121, .length = 2},
  [84] = {.index = 125, .length = 3},
  [85] = {.index = 67, .length = 2},
  [86] = {.index = 128, .length = 3},
  [87] = {.index = 128, .length = 3},
  [88] = {.index = 63, .length = 2},
  [89] = {.index = 131, .length = 2},
  [90] = {.index = 133, .length = 3},
  [91] = {.index = 136, .length = 3},
  [92] = {.index = 139, .length = 4},
  [93] = {.index = 143, .length = 2},
  [94] = {.index = 145, .length = 1},
  [95] = {.index = 146, .length = 1},
  [96] = {.index = 147, .length = 3},
  [97] = {.index = 150, .length = 3},
  [98] = {.index = 153, .length = 3},
  [99] = {.index = 156, .length = 3},
  [100] = {.index = 159, .length = 3},
  [101] = {.index = 162, .length = 3},
  [102] = {.index = 165, .length = 3},
  [103] = {.index = 168, .length = 2},
  [104] = {.index = 170, .length = 2},
  [105] = {.index = 172, .length = 3},
  [106] = {.index = 175, .length = 3},
  [107] = {.index = 178, .length = 2},
  [108] = {.index = 180, .length = 2},
  [109] = {.index = 182, .length = 2},
  [110] = {.index = 178, .length = 2},
  [111] = {.index = 180, .length = 2},
  [112] = {.index = 182, .length = 2},
  [113] = {.index = 168, .length = 2},
  [114] = {.index = 170, .length = 2},
  [115] = {.index = 172, .length = 3},
  [116] = {.index = 175, .length = 3},
  [117] = {.index = 184, .length = 3},
  [118] = {.index = 184, .length = 3},
  [119] = {.index = 67, .length = 2},
  [120] = {.index = 187, .length = 3},
  [121] = {.index = 131, .length = 2},
  [122] = {.index = 190, .length = 3},
  [123] = {.index = 193, .length = 4},
  [124] = {.index = 197, .length = 3},
  [125] = {.index = 200, .length = 2},
  [126] = {.index = 202, .length = 2},
  [127] = {.index = 204, .length = 3},
  [128] = {.index = 207, .length = 4},
  [129] = {.index = 211, .length = 4},
  [130] = {.index = 215, .length = 4},
  [131] = {.index = 219, .length = 4},
  [132] = {.index = 223, .length = 4},
  [133] = {.index = 227, .length = 4},
  [134] = {.index = 231, .length = 3},
  [135] = {.index = 234, .length = 3},
  [136] = {.index = 237, .length = 2},
  [137] = {.index = 239, .length = 2},
  [138] = {.index = 241, .length = 3},
  [139] = {.index = 244, .length = 3},
  [140] = {.index = 237, .length = 2},
  [141] = {.index = 239, .length = 2},
  [142] = {.index = 241, .length = 3},
  [143] = {.index = 244, .length = 3},
  [144] = {.index = 231, .length = 3},
  [145] = {.index = 234, .length = 3},
  [146] = {.index = 247, .length = 3},
  [147] = {.index = 250, .length = 3},
  [148] = {.index = 250, .length = 3},
  [149] = {.index = 253, .length = 4},
  [150] = {.index = 257, .length = 5},
  [151] = {.index = 262, .length = 5},
  [152] = {.index = 267, .length = 5},
  [153] = {.index = 272, .length = 3},
  [154] = {.index = 275, .length = 3},
  [155] = {.index = 272, .length = 3},
  [156] = {.index = 275, .length = 3},
  [157] = {.index = 278, .length = 3},
  [158] = {.index = 281, .length = 6},
};

static const TSFieldMapEntry ts_field_map_entries[] = {
  [0] =
    {field_name, 0},
  [1] =
    {field_redirect, 0},
  [2] =
    {field_descriptor, 0},
  [3] =
    {field_left, 0, .inherited = true},
    {field_operator, 0, .inherited = true},
    {field_right, 0, .inherited = true},
  [6] =
    {field_alternative, 0, .inherited = true},
    {field_condition, 0, .inherited = true},
    {field_consequence, 0, .inherited = true},
  [9] =
    {field_operator, 0, .inherited = true},
  [10] =
    {field_destination, 1},
  [11] =
    {field_operator, 0},
  [12] =
    {field_body, 0},
    {field_redirect, 1},
  [14] =
    {field_body, 0},
  [15] =
    {field_argument, 0},
  [16] =
    {field_argument, 1, .inherited = true},
    {field_name, 0},
    {field_redirect, 1, .inherited = true},
  [19] =
    {field_name, 1},
    {field_redirect, 0, .inherited = true},
  [21] =
    {field_redirect, 0, .inherited = true},
    {field_redirect, 1, .inherited = true},
  [23] =
    {field_descriptor, 0},
    {field_destination, 2},
  [25] =
    {field_name, 0},
    {field_value, 2},
  [27] =
    {field_name, 0, .inherited = true},
    {field_value, 0, .inherited = true},
  [29] =
    {field_operator, 1},
  [30] =
    {field_body, 2},
    {field_condition, 1},
  [32] =
    {field_body, 2},
    {field_name, 1},
  [34] =
    {field_left, 1, .inherited = true},
    {field_operator, 1, .inherited = true},
    {field_right, 1, .inherited = true},
  [37] =
    {field_operator, 1, .inherited = true},
  [38] =
    {field_operator, 0, .inherited = true},
    {field_operator, 1, .inherited = true},
  [40] =
    {field_redirect, 1},
  [41] =
    {field_argument, 0},
    {field_argument, 1},
  [43] =
    {field_argument, 0, .inherited = true},
    {field_argument, 1, .inherited = true},
    {field_redirect, 0, .inherited = true},
    {field_redirect, 1, .inherited = true},
  [47] =
    {field_argument, 2, .inherited = true},
    {field_name, 1},
    {field_redirect, 0, .inherited = true},
    {field_redirect, 2, .inherited = true},
  [51] =
    {field_index, 2},
    {field_name, 0},
  [53] =
    {field_body, 3},
    {field_variable, 1},
  [55] =
    {field_left, 0},
    {field_operator, 1},
    {field_right, 2},
  [58] =
    {field_condition, 1},
  [59] =
    {field_value, 1},
  [60] =
    {field_body, 2},
    {field_name, 1},
    {field_redirect, 3},
  [63] =
    {field_operator, 0},
    {field_operator, 1},
  [65] =
    {field_operator, 0},
    {field_operator, 1, .inherited = true},
  [67] =
    {field_operator, 0},
    {field_operator, 2},
  [69] =
    {field_operator, 0},
    {field_operator, 2, .inherited = true},
  [71] =
    {field_body, 3},
    {field_name, 0},
  [73] =
    {field_body, 4},
    {field_condition, 2, .inherited = true},
    {field_initializer, 2, .inherited = true},
    {field_update, 2, .inherited = true},
  [77] =
    {field_initializer, 0},
  [78] =
    {field_update, 2},
  [79] =
    {field_value, 0},
  [80] =
    {field_body, 4},
    {field_name, 1},
  [82] =
    {field_operator, 0},
    {field_operator, 1},
    {field_operator, 2, .inherited = true},
  [85] =
    {field_operator, 0},
    {field_operator, 1, .inherited = true},
    {field_operator, 2},
  [88] =
    {field_body, 3},
    {field_name, 0},
    {field_redirect, 4},
  [91] =
    {field_operator, 0},
    {field_right, 1},
  [93] =
    {field_argument, 0, .inherited = true},
    {field_argument, 1, .inherited = true},
  [95] =
    {field_body, 5},
    {field_condition, 2, .inherited = true},
    {field_initializer, 2, .inherited = true},
    {field_update, 2, .inherited = true},
  [99] =
    {field_condition, 2},
    {field_initializer, 0},
  [101] =
    {field_initializer, 0},
    {field_update, 3},
  [103] =
    {field_initializer, 0},
    {field_initializer, 1},
  [105] =
    {field_condition, 1},
    {field_update, 3},
  [107] =
    {field_condition, 1},
    {field_condition, 2},
  [109] =
    {field_update, 2},
    {field_update, 3},
  [111] =
    {field_body, 5},
    {field_value, 3},
    {field_variable, 1},
  [114] =
    {field_alternative, 4},
    {field_condition, 0},
    {field_consequence, 2},
  [117] =
    {field_termination, 2},
    {field_value, 0},
  [119] =
    {field_fallthrough, 2},
    {field_value, 0},
  [121] =
    {field_value, 0},
    {field_value, 1, .inherited = true},
  [123] =
    {field_value, 0, .inherited = true},
    {field_value, 1, .inherited = true},
  [125] =
    {field_body, 4},
    {field_name, 1},
    {field_redirect, 5},
  [128] =
    {field_operator, 0},
    {field_operator, 1},
    {field_operator, 3},
  [131] =
    {field_operator, 0},
    {field_operator, 3},
  [133] =
    {field_operator, 0},
    {field_operator, 2},
    {field_operator, 3, .inherited = true},
  [136] =
    {field_operator, 0},
    {field_operator, 1},
    {field_operator, 3, .inherited = true},
  [139] =
    {field_operator, 0},
    {field_operator, 1, .inherited = true},
    {field_operator, 2},
    {field_operator, 3, .inherited = true},
  [143] =
    {field_operator, 2, .inherited = true},
    {field_right, 2, .inherited = true},
  [145] =
    {field_argument, 2, .inherited = true},
  [146] =
    {field_redirect, 2},
  [147] =
    {field_condition, 2},
    {field_initializer, 0},
    {field_update, 4},
  [150] =
    {field_condition, 2},
    {field_condition, 3},
    {field_initializer, 0},
  [153] =
    {field_initializer, 0},
    {field_update, 3},
    {field_update, 4},
  [156] =
    {field_condition, 3},
    {field_initializer, 0},
    {field_initializer, 1},
  [159] =
    {field_initializer, 0},
    {field_initializer, 1},
    {field_update, 4},
  [162] =
    {field_condition, 1},
    {field_update, 3},
    {field_update, 4},
  [165] =
    {field_condition, 1},
    {field_condition, 2},
    {field_update, 4},
  [168] =
    {field_termination, 3},
    {field_value, 0},
  [170] =
    {field_fallthrough, 3},
    {field_value, 0},
  [172] =
    {field_termination, 3},
    {field_value, 0},
    {field_value, 1, .inherited = true},
  [175] =
    {field_fallthrough, 3},
    {field_value, 0},
    {field_value, 1, .inherited = true},
  [178] =
    {field_termination, 3},
    {field_value, 1},
  [180] =
    {field_fallthrough, 3},
    {field_value, 1},
  [182] =
    {field_value, 1},
    {field_value, 2, .inherited = true},
  [184] =
    {field_operator, 0},
    {field_operator, 2},
    {field_operator, 4},
  [187] =
    {field_operator, 0},
    {field_operator, 1},
    {field_operator, 4},
  [190] =
    {field_operator, 0},
    {field_operator, 2},
    {field_operator, 4, .inherited = true},
  [193] =
    {field_operator, 0},
    {field_operator, 1, .inherited = true},
    {field_operator, 2},
    {field_operator, 4, .inherited = true},
  [197] =
    {field_descriptor, 0},
    {field_operator, 3, .inherited = true},
    {field_right, 3, .inherited = true},
  [200] =
    {field_argument, 3, .inherited = true},
    {field_descriptor, 0},
  [202] =
    {field_descriptor, 0},
    {field_redirect, 3},
  [204] =
    {field_operator, 3, .inherited = true},
    {field_redirect, 2},
    {field_right, 3, .inherited = true},
  [207] =
    {field_condition, 2},
    {field_initializer, 0},
    {field_update, 4},
    {field_update, 5},
  [211] =
    {field_condition, 2},
    {field_condition, 3},
    {field_initializer, 0},
    {field_update, 5},
  [215] =
    {field_condition, 3},
    {field_initializer, 0},
    {field_initializer, 1},
    {field_update, 5},
  [219] =
    {field_condition, 3},
    {field_condition, 4},
    {field_initializer, 0},
    {field_initializer, 1},
  [223] =
    {field_initializer, 0},
    {field_initializer, 1},
    {field_update, 4},
    {field_update, 5},
  [227] =
    {field_condition, 1},
    {field_condition, 2},
    {field_update, 4},
    {field_update, 5},
  [231] =
    {field_termination, 4},
    {field_value, 0},
    {field_value, 1, .inherited = true},
  [234] =
    {field_fallthrough, 4},
    {field_value, 0},
    {field_value, 1, .inherited = true},
  [237] =
    {field_termination, 4},
    {field_value, 1},
  [239] =
    {field_fallthrough, 4},
    {field_value, 1},
  [241] =
    {field_termination, 4},
    {field_value, 1},
    {field_value, 2, .inherited = true},
  [244] =
    {field_fallthrough, 4},
    {field_value, 1},
    {field_value, 2, .inherited = true},
  [247] =
    {field_operator, 0},
    {field_operator, 2},
    {field_operator, 5},
  [250] =
    {field_operator, 0},
    {field_operator, 3},
    {field_operator, 5},
  [253] =
    {field_descriptor, 0},
    {field_operator, 4, .inherited = true},
    {field_redirect, 3},
    {field_right, 4, .inherited = true},
  [257] =
    {field_condition, 2},
    {field_condition, 3},
    {field_initializer, 0},
    {field_update, 5},
    {field_update, 6},
  [262] =
    {field_condition, 3},
    {field_initializer, 0},
    {field_initializer, 1},
    {field_update, 5},
    {field_update, 6},
  [267] =
    {field_condition, 3},
    {field_condition, 4},
    {field_initializer, 0},
    {field_initializer, 1},
    {field_update, 6},
  [272] =
    {field_termination, 5},
    {field_value, 1},
    {field_value, 2, .inherited = true},
  [275] =
    {field_fallthrough, 5},
    {field_value, 1},
    {field_value, 2, .inherited = true},
  [278] =
    {field_operator, 0},
    {field_operator, 3},
    {field_operator, 6},
  [281] =
    {field_condition, 3},
    {field_condition, 4},
    {field_initializer, 0},
    {field_initializer, 1},
    {field_update, 6},
    {field_update, 7},
};

static const TSSymbol ts_alias_sequences[PRODUCTION_ID_COUNT][MAX_ALIAS_SEQUENCE_LENGTH] = {
  [0] = {0},
  [1] = {
    [0] = sym_word,
  },
  [9] = {
    [1] = sym_word,
  },
  [10] = {
    [1] = sym_word,
  },
  [11] = {
    [1] = anon_sym_AT2,
  },
  [13] = {
    [0] = anon_sym_AT2,
  },
  [16] = {
    [0] = sym_word,
  },
  [22] = {
    [2] = sym_word,
  },
  [23] = {
    [2] = sym_word,
  },
  [25] = {
    [2] = sym_word,
  },
  [32] = {
    [0] = anon_sym_AT2,
  },
  [33] = {
    [1] = anon_sym_AT2,
  },
  [36] = {
    [1] = sym_word,
  },
  [40] = {
    [2] = sym_word,
  },
  [45] = {
    [1] = sym_word,
  },
  [48] = {
    [0] = sym_regex,
  },
  [53] = {
    [1] = anon_sym_AT2,
  },
  [55] = {
    [0] = sym_variable_name,
  },
  [59] = {
    [0] = sym_word,
  },
  [62] = {
    [2] = sym_word,
  },
  [77] = {
    [0] = sym_word,
  },
  [78] = {
    [0] = sym_word,
  },
  [79] = {
    [0] = sym_word,
  },
  [85] = {
    [3] = sym_word,
  },
  [86] = {
    [2] = sym_word,
  },
  [88] = {
    [3] = sym_word,
  },
  [103] = {
    [0] = sym_word,
  },
  [104] = {
    [0] = sym_word,
  },
  [105] = {
    [0] = sym_word,
  },
  [106] = {
    [0] = sym_word,
  },
  [107] = {
    [1] = sym_word,
  },
  [108] = {
    [1] = sym_word,
  },
  [109] = {
    [1] = sym_word,
  },
  [117] = {
    [3] = sym_word,
  },
  [119] = {
    [4] = sym_word,
  },
  [120] = {
    [3] = sym_word,
  },
  [121] = {
    [4] = sym_word,
  },
  [134] = {
    [0] = sym_word,
  },
  [135] = {
    [0] = sym_word,
  },
  [136] = {
    [1] = sym_word,
  },
  [137] = {
    [1] = sym_word,
  },
  [138] = {
    [1] = sym_word,
  },
  [139] = {
    [1] = sym_word,
  },
  [146] = {
    [4] = sym_word,
  },
  [147] = {
    [4] = sym_word,
  },
  [153] = {
    [1] = sym_word,
  },
  [154] = {
    [1] = sym_word,
  },
};

static const uint16_t ts_non_terminal_alias_map[] = {
  aux_sym__literal_repeat1, 2,
    aux_sym__literal_repeat1,
    sym_word,
  0,
};

static const TSStateId ts_primary_state_ids[STATE_COUNT] = {
  [0] = 0,
  [1] = 1,
  [2] = 2,
  [3] = 2,
  [4] = 4,
  [5] = 4,
  [6] = 4,
  [7] = 4,
  [8] = 4,
  [9] = 4,
  [10] = 4,
  [11] = 4,
  [12] = 12,
  [13] = 13,
  [14] = 13,
  [15] = 13,
  [16] = 12,
  [17] = 17,
  [18] = 17,
  [19] = 17,
  [20] = 13,
  [21] = 17,
  [22] = 22,
  [23] = 23,
  [24] = 24,
  [25] = 25,
  [26] = 26,
  [27] = 27,
  [28] = 28,
  [29] = 29,
  [30] = 30,
  [31] = 31,
  [32] = 32,
  [33] = 33,
  [34] = 34,
  [35] = 35,
  [36] = 36,
  [37] = 37,
  [38] = 38,
  [39] = 39,
  [40] = 40,
  [41] = 41,
  [42] = 42,
  [43] = 43,
  [44] = 44,
  [45] = 45,
  [46] = 46,
  [47] = 47,
  [48] = 48,
  [49] = 49,
  [50] = 50,
  [51] = 50,
  [52] = 49,
  [53] = 49,
  [54] = 50,
  [55] = 49,
  [56] = 49,
  [57] = 49,
  [58] = 49,
  [59] = 50,
  [60] = 49,
  [61] = 61,
  [62] = 34,
  [63] = 63,
  [64] = 64,
  [65] = 61,
  [66] = 66,
  [67] = 67,
  [68] = 68,
  [69] = 64,
  [70] = 70,
  [71] = 71,
  [72] = 63,
  [73] = 61,
  [74] = 68,
  [75] = 66,
  [76] = 70,
  [77] = 67,
  [78] = 71,
  [79] = 64,
  [80] = 61,
  [81] = 63,
  [82] = 68,
  [83] = 66,
  [84] = 70,
  [85] = 71,
  [86] = 63,
  [87] = 68,
  [88] = 67,
  [89] = 64,
  [90] = 70,
  [91] = 71,
  [92] = 63,
  [93] = 68,
  [94] = 70,
  [95] = 71,
  [96] = 63,
  [97] = 61,
  [98] = 66,
  [99] = 68,
  [100] = 67,
  [101] = 64,
  [102] = 102,
  [103] = 70,
  [104] = 71,
  [105] = 34,
  [106] = 61,
  [107] = 61,
  [108] = 66,
  [109] = 67,
  [110] = 64,
  [111] = 61,
  [112] = 112,
  [113] = 66,
  [114] = 64,
  [115] = 67,
  [116] = 64,
  [117] = 61,
  [118] = 66,
  [119] = 61,
  [120] = 67,
  [121] = 66,
  [122] = 102,
  [123] = 67,
  [124] = 64,
  [125] = 112,
  [126] = 61,
  [127] = 66,
  [128] = 64,
  [129] = 61,
  [130] = 66,
  [131] = 67,
  [132] = 67,
  [133] = 102,
  [134] = 64,
  [135] = 61,
  [136] = 61,
  [137] = 71,
  [138] = 66,
  [139] = 67,
  [140] = 64,
  [141] = 61,
  [142] = 66,
  [143] = 70,
  [144] = 67,
  [145] = 112,
  [146] = 66,
  [147] = 64,
  [148] = 64,
  [149] = 61,
  [150] = 61,
  [151] = 66,
  [152] = 67,
  [153] = 102,
  [154] = 63,
  [155] = 112,
  [156] = 61,
  [157] = 66,
  [158] = 64,
  [159] = 61,
  [160] = 66,
  [161] = 67,
  [162] = 67,
  [163] = 112,
  [164] = 64,
  [165] = 64,
  [166] = 64,
  [167] = 61,
  [168] = 61,
  [169] = 66,
  [170] = 66,
  [171] = 67,
  [172] = 67,
  [173] = 112,
  [174] = 64,
  [175] = 61,
  [176] = 66,
  [177] = 67,
  [178] = 64,
  [179] = 112,
  [180] = 64,
  [181] = 61,
  [182] = 67,
  [183] = 61,
  [184] = 66,
  [185] = 67,
  [186] = 64,
  [187] = 61,
  [188] = 66,
  [189] = 67,
  [190] = 64,
  [191] = 66,
  [192] = 61,
  [193] = 66,
  [194] = 67,
  [195] = 67,
  [196] = 64,
  [197] = 61,
  [198] = 66,
  [199] = 64,
  [200] = 67,
  [201] = 64,
  [202] = 61,
  [203] = 66,
  [204] = 67,
  [205] = 64,
  [206] = 61,
  [207] = 66,
  [208] = 61,
  [209] = 66,
  [210] = 67,
  [211] = 64,
  [212] = 61,
  [213] = 66,
  [214] = 67,
  [215] = 64,
  [216] = 61,
  [217] = 67,
  [218] = 66,
  [219] = 67,
  [220] = 64,
  [221] = 61,
  [222] = 66,
  [223] = 67,
  [224] = 64,
  [225] = 61,
  [226] = 226,
  [227] = 66,
  [228] = 67,
  [229] = 64,
  [230] = 61,
  [231] = 66,
  [232] = 232,
  [233] = 64,
  [234] = 234,
  [235] = 67,
  [236] = 64,
  [237] = 71,
  [238] = 61,
  [239] = 61,
  [240] = 112,
  [241] = 66,
  [242] = 67,
  [243] = 64,
  [244] = 61,
  [245] = 66,
  [246] = 66,
  [247] = 66,
  [248] = 67,
  [249] = 64,
  [250] = 67,
  [251] = 61,
  [252] = 64,
  [253] = 66,
  [254] = 67,
  [255] = 64,
  [256] = 61,
  [257] = 66,
  [258] = 67,
  [259] = 61,
  [260] = 66,
  [261] = 64,
  [262] = 61,
  [263] = 66,
  [264] = 67,
  [265] = 67,
  [266] = 64,
  [267] = 61,
  [268] = 64,
  [269] = 61,
  [270] = 66,
  [271] = 66,
  [272] = 67,
  [273] = 67,
  [274] = 64,
  [275] = 66,
  [276] = 61,
  [277] = 64,
  [278] = 64,
  [279] = 66,
  [280] = 70,
  [281] = 61,
  [282] = 68,
  [283] = 66,
  [284] = 63,
  [285] = 67,
  [286] = 64,
  [287] = 64,
  [288] = 66,
  [289] = 66,
  [290] = 68,
  [291] = 61,
  [292] = 64,
  [293] = 66,
  [294] = 67,
  [295] = 61,
  [296] = 64,
  [297] = 61,
  [298] = 61,
  [299] = 64,
  [300] = 66,
  [301] = 61,
  [302] = 64,
  [303] = 66,
  [304] = 61,
  [305] = 64,
  [306] = 66,
  [307] = 61,
  [308] = 64,
  [309] = 66,
  [310] = 61,
  [311] = 64,
  [312] = 66,
  [313] = 61,
  [314] = 64,
  [315] = 66,
  [316] = 61,
  [317] = 64,
  [318] = 66,
  [319] = 61,
  [320] = 64,
  [321] = 66,
  [322] = 61,
  [323] = 64,
  [324] = 66,
  [325] = 61,
  [326] = 64,
  [327] = 66,
  [328] = 61,
  [329] = 64,
  [330] = 66,
  [331] = 61,
  [332] = 64,
  [333] = 66,
  [334] = 61,
  [335] = 64,
  [336] = 66,
  [337] = 67,
  [338] = 66,
  [339] = 67,
  [340] = 64,
  [341] = 61,
  [342] = 66,
  [343] = 61,
  [344] = 64,
  [345] = 67,
  [346] = 66,
  [347] = 61,
  [348] = 64,
  [349] = 67,
  [350] = 67,
  [351] = 64,
  [352] = 66,
  [353] = 61,
  [354] = 66,
  [355] = 67,
  [356] = 66,
  [357] = 67,
  [358] = 61,
  [359] = 64,
  [360] = 64,
  [361] = 67,
  [362] = 66,
  [363] = 61,
  [364] = 61,
  [365] = 66,
  [366] = 67,
  [367] = 64,
  [368] = 67,
  [369] = 66,
  [370] = 61,
  [371] = 64,
  [372] = 67,
  [373] = 66,
  [374] = 64,
  [375] = 67,
  [376] = 64,
  [377] = 67,
  [378] = 66,
  [379] = 61,
  [380] = 64,
  [381] = 67,
  [382] = 66,
  [383] = 61,
  [384] = 64,
  [385] = 67,
  [386] = 66,
  [387] = 61,
  [388] = 64,
  [389] = 67,
  [390] = 66,
  [391] = 61,
  [392] = 64,
  [393] = 67,
  [394] = 66,
  [395] = 61,
  [396] = 64,
  [397] = 67,
  [398] = 66,
  [399] = 61,
  [400] = 64,
  [401] = 67,
  [402] = 66,
  [403] = 61,
  [404] = 64,
  [405] = 67,
  [406] = 66,
  [407] = 61,
  [408] = 64,
  [409] = 67,
  [410] = 61,
  [411] = 66,
  [412] = 61,
  [413] = 64,
  [414] = 67,
  [415] = 66,
  [416] = 61,
  [417] = 64,
  [418] = 418,
  [419] = 419,
  [420] = 420,
  [421] = 421,
  [422] = 421,
  [423] = 423,
  [424] = 420,
  [425] = 421,
  [426] = 423,
  [427] = 423,
  [428] = 421,
  [429] = 423,
  [430] = 420,
  [431] = 423,
  [432] = 420,
  [433] = 433,
  [434] = 434,
  [435] = 435,
  [436] = 433,
  [437] = 435,
  [438] = 435,
  [439] = 435,
  [440] = 440,
  [441] = 433,
  [442] = 435,
  [443] = 435,
  [444] = 433,
  [445] = 433,
  [446] = 435,
  [447] = 433,
  [448] = 435,
  [449] = 433,
  [450] = 433,
  [451] = 451,
  [452] = 452,
  [453] = 452,
  [454] = 451,
  [455] = 452,
  [456] = 451,
  [457] = 452,
  [458] = 451,
  [459] = 451,
  [460] = 451,
  [461] = 452,
  [462] = 452,
  [463] = 451,
  [464] = 464,
  [465] = 465,
  [466] = 464,
  [467] = 465,
  [468] = 468,
  [469] = 469,
  [470] = 470,
  [471] = 471,
  [472] = 472,
  [473] = 473,
  [474] = 474,
  [475] = 475,
  [476] = 476,
  [477] = 477,
  [478] = 478,
  [479] = 479,
  [480] = 480,
  [481] = 481,
  [482] = 482,
  [483] = 483,
  [484] = 484,
  [485] = 485,
  [486] = 486,
  [487] = 487,
  [488] = 488,
  [489] = 489,
  [490] = 490,
  [491] = 491,
  [492] = 492,
  [493] = 493,
  [494] = 494,
  [495] = 495,
  [496] = 472,
  [497] = 471,
  [498] = 473,
  [499] = 470,
  [500] = 469,
  [501] = 468,
  [502] = 502,
  [503] = 489,
  [504] = 490,
  [505] = 484,
  [506] = 488,
  [507] = 492,
  [508] = 476,
  [509] = 481,
  [510] = 480,
  [511] = 486,
  [512] = 483,
  [513] = 482,
  [514] = 485,
  [515] = 474,
  [516] = 475,
  [517] = 491,
  [518] = 477,
  [519] = 478,
  [520] = 479,
  [521] = 487,
  [522] = 493,
  [523] = 494,
  [524] = 502,
  [525] = 495,
  [526] = 526,
  [527] = 526,
  [528] = 528,
  [529] = 529,
  [530] = 528,
  [531] = 529,
  [532] = 532,
  [533] = 528,
  [534] = 528,
  [535] = 535,
  [536] = 529,
  [537] = 529,
  [538] = 465,
  [539] = 464,
  [540] = 540,
  [541] = 532,
  [542] = 540,
  [543] = 529,
  [544] = 528,
  [545] = 529,
  [546] = 546,
  [547] = 535,
  [548] = 548,
  [549] = 465,
  [550] = 464,
  [551] = 551,
  [552] = 465,
  [553] = 464,
  [554] = 528,
  [555] = 464,
  [556] = 465,
  [557] = 540,
  [558] = 464,
  [559] = 551,
  [560] = 535,
  [561] = 535,
  [562] = 532,
  [563] = 548,
  [564] = 465,
  [565] = 546,
  [566] = 540,
  [567] = 464,
  [568] = 465,
  [569] = 532,
  [570] = 551,
  [571] = 535,
  [572] = 464,
  [573] = 573,
  [574] = 532,
  [575] = 575,
  [576] = 548,
  [577] = 546,
  [578] = 465,
  [579] = 465,
  [580] = 464,
  [581] = 546,
  [582] = 540,
  [583] = 583,
  [584] = 540,
  [585] = 551,
  [586] = 548,
  [587] = 535,
  [588] = 464,
  [589] = 465,
  [590] = 590,
  [591] = 575,
  [592] = 548,
  [593] = 583,
  [594] = 551,
  [595] = 573,
  [596] = 528,
  [597] = 464,
  [598] = 598,
  [599] = 599,
  [600] = 529,
  [601] = 575,
  [602] = 548,
  [603] = 465,
  [604] = 604,
  [605] = 464,
  [606] = 464,
  [607] = 607,
  [608] = 465,
  [609] = 583,
  [610] = 551,
  [611] = 465,
  [612] = 546,
  [613] = 613,
  [614] = 573,
  [615] = 573,
  [616] = 573,
  [617] = 529,
  [618] = 618,
  [619] = 618,
  [620] = 528,
  [621] = 583,
  [622] = 622,
  [623] = 618,
  [624] = 622,
  [625] = 625,
  [626] = 618,
  [627] = 583,
  [628] = 465,
  [629] = 613,
  [630] = 583,
  [631] = 604,
  [632] = 590,
  [633] = 598,
  [634] = 464,
  [635] = 607,
  [636] = 618,
  [637] = 464,
  [638] = 465,
  [639] = 464,
  [640] = 465,
  [641] = 464,
  [642] = 575,
  [643] = 465,
  [644] = 618,
  [645] = 607,
  [646] = 575,
  [647] = 465,
  [648] = 604,
  [649] = 573,
  [650] = 618,
  [651] = 583,
  [652] = 464,
  [653] = 575,
  [654] = 573,
  [655] = 625,
  [656] = 618,
  [657] = 575,
  [658] = 604,
  [659] = 659,
  [660] = 583,
  [661] = 659,
  [662] = 464,
  [663] = 464,
  [664] = 598,
  [665] = 607,
  [666] = 465,
  [667] = 613,
  [668] = 465,
  [669] = 465,
  [670] = 465,
  [671] = 659,
  [672] = 583,
  [673] = 573,
  [674] = 464,
  [675] = 535,
  [676] = 465,
  [677] = 464,
  [678] = 583,
  [679] = 464,
  [680] = 573,
  [681] = 604,
  [682] = 604,
  [683] = 464,
  [684] = 573,
  [685] = 659,
  [686] = 575,
  [687] = 464,
  [688] = 540,
  [689] = 607,
  [690] = 590,
  [691] = 575,
  [692] = 465,
  [693] = 590,
  [694] = 604,
  [695] = 613,
  [696] = 465,
  [697] = 659,
  [698] = 659,
  [699] = 464,
  [700] = 465,
  [701] = 659,
  [702] = 573,
  [703] = 532,
  [704] = 583,
  [705] = 575,
  [706] = 607,
  [707] = 575,
  [708] = 607,
  [709] = 598,
  [710] = 464,
  [711] = 711,
  [712] = 712,
  [713] = 711,
  [714] = 573,
  [715] = 583,
  [716] = 464,
  [717] = 464,
  [718] = 590,
  [719] = 573,
  [720] = 575,
  [721] = 465,
  [722] = 607,
  [723] = 464,
  [724] = 604,
  [725] = 465,
  [726] = 613,
  [727] = 540,
  [728] = 728,
  [729] = 729,
  [730] = 730,
  [731] = 598,
  [732] = 464,
  [733] = 604,
  [734] = 734,
  [735] = 465,
  [736] = 728,
  [737] = 465,
  [738] = 607,
  [739] = 604,
  [740] = 464,
  [741] = 583,
  [742] = 575,
  [743] = 613,
  [744] = 729,
  [745] = 728,
  [746] = 546,
  [747] = 583,
  [748] = 548,
  [749] = 464,
  [750] = 465,
  [751] = 532,
  [752] = 535,
  [753] = 728,
  [754] = 573,
  [755] = 575,
  [756] = 465,
  [757] = 573,
  [758] = 729,
  [759] = 604,
  [760] = 734,
  [761] = 590,
  [762] = 712,
  [763] = 730,
  [764] = 607,
  [765] = 465,
  [766] = 464,
  [767] = 464,
  [768] = 465,
  [769] = 729,
  [770] = 728,
  [771] = 729,
  [772] = 607,
  [773] = 575,
  [774] = 465,
  [775] = 465,
  [776] = 464,
  [777] = 551,
  [778] = 607,
  [779] = 604,
  [780] = 599,
  [781] = 583,
  [782] = 464,
  [783] = 604,
  [784] = 464,
  [785] = 604,
  [786] = 573,
  [787] = 465,
  [788] = 573,
  [789] = 575,
  [790] = 604,
  [791] = 575,
  [792] = 607,
  [793] = 465,
  [794] = 625,
  [795] = 464,
  [796] = 464,
  [797] = 551,
  [798] = 607,
  [799] = 548,
  [800] = 622,
  [801] = 465,
  [802] = 599,
  [803] = 546,
  [804] = 607,
  [805] = 465,
  [806] = 573,
  [807] = 807,
  [808] = 470,
  [809] = 809,
  [810] = 625,
  [811] = 583,
  [812] = 622,
  [813] = 813,
  [814] = 473,
  [815] = 815,
  [816] = 575,
  [817] = 583,
  [818] = 818,
  [819] = 819,
  [820] = 468,
  [821] = 575,
  [822] = 822,
  [823] = 573,
  [824] = 469,
  [825] = 604,
  [826] = 607,
  [827] = 827,
  [828] = 607,
  [829] = 604,
  [830] = 604,
  [831] = 476,
  [832] = 832,
  [833] = 607,
  [834] = 469,
  [835] = 818,
  [836] = 604,
  [837] = 470,
  [838] = 473,
  [839] = 809,
  [840] = 464,
  [841] = 468,
  [842] = 465,
  [843] = 481,
  [844] = 486,
  [845] = 487,
  [846] = 479,
  [847] = 822,
  [848] = 490,
  [849] = 807,
  [850] = 469,
  [851] = 583,
  [852] = 815,
  [853] = 819,
  [854] = 854,
  [855] = 478,
  [856] = 477,
  [857] = 491,
  [858] = 573,
  [859] = 859,
  [860] = 583,
  [861] = 861,
  [862] = 573,
  [863] = 575,
  [864] = 483,
  [865] = 613,
  [866] = 482,
  [867] = 573,
  [868] = 489,
  [869] = 475,
  [870] = 575,
  [871] = 485,
  [872] = 474,
  [873] = 464,
  [874] = 874,
  [875] = 598,
  [876] = 876,
  [877] = 465,
  [878] = 590,
  [879] = 575,
  [880] = 880,
  [881] = 465,
  [882] = 607,
  [883] = 464,
  [884] = 480,
  [885] = 885,
  [886] = 813,
  [887] = 484,
  [888] = 488,
  [889] = 492,
  [890] = 465,
  [891] = 470,
  [892] = 464,
  [893] = 473,
  [894] = 468,
  [895] = 583,
  [896] = 464,
  [897] = 573,
  [898] = 478,
  [899] = 465,
  [900] = 477,
  [901] = 491,
  [902] = 465,
  [903] = 575,
  [904] = 464,
  [905] = 604,
  [906] = 465,
  [907] = 813,
  [908] = 604,
  [909] = 607,
  [910] = 482,
  [911] = 464,
  [912] = 483,
  [913] = 807,
  [914] = 481,
  [915] = 486,
  [916] = 487,
  [917] = 490,
  [918] = 479,
  [919] = 487,
  [920] = 469,
  [921] = 479,
  [922] = 489,
  [923] = 490,
  [924] = 486,
  [925] = 925,
  [926] = 926,
  [927] = 478,
  [928] = 477,
  [929] = 491,
  [930] = 481,
  [931] = 488,
  [932] = 819,
  [933] = 925,
  [934] = 492,
  [935] = 470,
  [936] = 583,
  [937] = 925,
  [938] = 815,
  [939] = 607,
  [940] = 813,
  [941] = 465,
  [942] = 464,
  [943] = 464,
  [944] = 604,
  [945] = 807,
  [946] = 465,
  [947] = 473,
  [948] = 809,
  [949] = 607,
  [950] = 590,
  [951] = 925,
  [952] = 494,
  [953] = 464,
  [954] = 465,
  [955] = 925,
  [956] = 464,
  [957] = 815,
  [958] = 925,
  [959] = 470,
  [960] = 465,
  [961] = 822,
  [962] = 613,
  [963] = 475,
  [964] = 484,
  [965] = 468,
  [966] = 469,
  [967] = 583,
  [968] = 480,
  [969] = 469,
  [970] = 468,
  [971] = 573,
  [972] = 575,
  [973] = 573,
  [974] = 473,
  [975] = 488,
  [976] = 583,
  [977] = 492,
  [978] = 470,
  [979] = 598,
  [980] = 474,
  [981] = 575,
  [982] = 485,
  [983] = 473,
  [984] = 484,
  [985] = 874,
  [986] = 822,
  [987] = 480,
  [988] = 476,
  [989] = 925,
  [990] = 475,
  [991] = 818,
  [992] = 859,
  [993] = 832,
  [994] = 885,
  [995] = 880,
  [996] = 468,
  [997] = 925,
  [998] = 489,
  [999] = 482,
  [1000] = 854,
  [1001] = 861,
  [1002] = 818,
  [1003] = 876,
  [1004] = 483,
  [1005] = 474,
  [1006] = 809,
  [1007] = 819,
  [1008] = 485,
  [1009] = 476,
  [1010] = 465,
  [1011] = 1011,
  [1012] = 813,
  [1013] = 476,
  [1014] = 1014,
  [1015] = 474,
  [1016] = 475,
  [1017] = 491,
  [1018] = 477,
  [1019] = 478,
  [1020] = 479,
  [1021] = 876,
  [1022] = 487,
  [1023] = 874,
  [1024] = 464,
  [1025] = 465,
  [1026] = 926,
  [1027] = 485,
  [1028] = 822,
  [1029] = 476,
  [1030] = 1030,
  [1031] = 490,
  [1032] = 470,
  [1033] = 474,
  [1034] = 475,
  [1035] = 575,
  [1036] = 1036,
  [1037] = 482,
  [1038] = 483,
  [1039] = 854,
  [1040] = 473,
  [1041] = 464,
  [1042] = 861,
  [1043] = 491,
  [1044] = 477,
  [1045] = 1045,
  [1046] = 478,
  [1047] = 479,
  [1048] = 464,
  [1049] = 487,
  [1050] = 465,
  [1051] = 464,
  [1052] = 495,
  [1053] = 465,
  [1054] = 473,
  [1055] = 470,
  [1056] = 573,
  [1057] = 486,
  [1058] = 484,
  [1059] = 809,
  [1060] = 481,
  [1061] = 489,
  [1062] = 488,
  [1063] = 492,
  [1064] = 485,
  [1065] = 468,
  [1066] = 859,
  [1067] = 1067,
  [1068] = 1068,
  [1069] = 832,
  [1070] = 464,
  [1071] = 465,
  [1072] = 885,
  [1073] = 470,
  [1074] = 473,
  [1075] = 583,
  [1076] = 880,
  [1077] = 819,
  [1078] = 815,
  [1079] = 468,
  [1080] = 813,
  [1081] = 1081,
  [1082] = 1082,
  [1083] = 482,
  [1084] = 483,
  [1085] = 480,
  [1086] = 470,
  [1087] = 473,
  [1088] = 468,
  [1089] = 469,
  [1090] = 604,
  [1091] = 469,
  [1092] = 469,
  [1093] = 874,
  [1094] = 876,
  [1095] = 494,
  [1096] = 809,
  [1097] = 494,
  [1098] = 822,
  [1099] = 491,
  [1100] = 1100,
  [1101] = 477,
  [1102] = 607,
  [1103] = 478,
  [1104] = 480,
  [1105] = 469,
  [1106] = 859,
  [1107] = 832,
  [1108] = 885,
  [1109] = 880,
  [1110] = 607,
  [1111] = 604,
  [1112] = 486,
  [1113] = 484,
  [1114] = 465,
  [1115] = 464,
  [1116] = 481,
  [1117] = 604,
  [1118] = 607,
  [1119] = 483,
  [1120] = 482,
  [1121] = 859,
  [1122] = 861,
  [1123] = 470,
  [1124] = 854,
  [1125] = 489,
  [1126] = 488,
  [1127] = 832,
  [1128] = 492,
  [1129] = 490,
  [1130] = 489,
  [1131] = 473,
  [1132] = 818,
  [1133] = 475,
  [1134] = 465,
  [1135] = 465,
  [1136] = 479,
  [1137] = 487,
  [1138] = 486,
  [1139] = 481,
  [1140] = 468,
  [1141] = 464,
  [1142] = 476,
  [1143] = 485,
  [1144] = 469,
  [1145] = 474,
  [1146] = 1146,
  [1147] = 1011,
  [1148] = 1148,
  [1149] = 1149,
  [1150] = 807,
  [1151] = 1030,
  [1152] = 490,
  [1153] = 1030,
  [1154] = 464,
  [1155] = 480,
  [1156] = 1148,
  [1157] = 484,
  [1158] = 1158,
  [1159] = 488,
  [1160] = 885,
  [1161] = 492,
  [1162] = 880,
  [1163] = 1163,
  [1164] = 465,
  [1165] = 487,
  [1166] = 1148,
  [1167] = 1011,
  [1168] = 1168,
  [1169] = 1082,
  [1170] = 1068,
  [1171] = 1067,
  [1172] = 607,
  [1173] = 1173,
  [1174] = 469,
  [1175] = 1045,
  [1176] = 489,
  [1177] = 470,
  [1178] = 604,
  [1179] = 494,
  [1180] = 1014,
  [1181] = 1181,
  [1182] = 1182,
  [1183] = 491,
  [1184] = 1081,
  [1185] = 477,
  [1186] = 486,
  [1187] = 478,
  [1188] = 483,
  [1189] = 482,
  [1190] = 1190,
  [1191] = 1191,
  [1192] = 1148,
  [1193] = 1011,
  [1194] = 488,
  [1195] = 492,
  [1196] = 475,
  [1197] = 464,
  [1198] = 483,
  [1199] = 482,
  [1200] = 485,
  [1201] = 1146,
  [1202] = 1149,
  [1203] = 876,
  [1204] = 489,
  [1205] = 476,
  [1206] = 483,
  [1207] = 482,
  [1208] = 475,
  [1209] = 494,
  [1210] = 476,
  [1211] = 470,
  [1212] = 926,
  [1213] = 489,
  [1214] = 473,
  [1215] = 473,
  [1216] = 474,
  [1217] = 468,
  [1218] = 485,
  [1219] = 874,
  [1220] = 475,
  [1221] = 474,
  [1222] = 476,
  [1223] = 490,
  [1224] = 485,
  [1225] = 474,
  [1226] = 469,
  [1227] = 470,
  [1228] = 473,
  [1229] = 495,
  [1230] = 468,
  [1231] = 487,
  [1232] = 480,
  [1233] = 484,
  [1234] = 469,
  [1235] = 480,
  [1236] = 484,
  [1237] = 491,
  [1238] = 477,
  [1239] = 478,
  [1240] = 491,
  [1241] = 477,
  [1242] = 478,
  [1243] = 861,
  [1244] = 854,
  [1245] = 470,
  [1246] = 492,
  [1247] = 488,
  [1248] = 484,
  [1249] = 473,
  [1250] = 468,
  [1251] = 470,
  [1252] = 492,
  [1253] = 488,
  [1254] = 491,
  [1255] = 477,
  [1256] = 478,
  [1257] = 473,
  [1258] = 880,
  [1259] = 481,
  [1260] = 495,
  [1261] = 885,
  [1262] = 468,
  [1263] = 832,
  [1264] = 469,
  [1265] = 859,
  [1266] = 469,
  [1267] = 926,
  [1268] = 480,
  [1269] = 1158,
  [1270] = 1163,
  [1271] = 859,
  [1272] = 480,
  [1273] = 832,
  [1274] = 885,
  [1275] = 1100,
  [1276] = 880,
  [1277] = 479,
  [1278] = 490,
  [1279] = 474,
  [1280] = 479,
  [1281] = 1281,
  [1282] = 489,
  [1283] = 485,
  [1284] = 481,
  [1285] = 486,
  [1286] = 479,
  [1287] = 488,
  [1288] = 487,
  [1289] = 492,
  [1290] = 1158,
  [1291] = 1163,
  [1292] = 483,
  [1293] = 482,
  [1294] = 490,
  [1295] = 494,
  [1296] = 475,
  [1297] = 486,
  [1298] = 490,
  [1299] = 481,
  [1300] = 484,
  [1301] = 1301,
  [1302] = 486,
  [1303] = 476,
  [1304] = 487,
  [1305] = 481,
  [1306] = 479,
  [1307] = 1163,
  [1308] = 469,
  [1309] = 488,
  [1310] = 492,
  [1311] = 470,
  [1312] = 1173,
  [1313] = 494,
  [1314] = 583,
  [1315] = 487,
  [1316] = 486,
  [1317] = 573,
  [1318] = 1068,
  [1319] = 481,
  [1320] = 1100,
  [1321] = 1045,
  [1322] = 1067,
  [1323] = 478,
  [1324] = 477,
  [1325] = 1191,
  [1326] = 1190,
  [1327] = 490,
  [1328] = 1182,
  [1329] = 1181,
  [1330] = 832,
  [1331] = 859,
  [1332] = 575,
  [1333] = 473,
  [1334] = 479,
  [1335] = 1014,
  [1336] = 1081,
  [1337] = 1148,
  [1338] = 495,
  [1339] = 859,
  [1340] = 491,
  [1341] = 1011,
  [1342] = 832,
  [1343] = 885,
  [1344] = 880,
  [1345] = 487,
  [1346] = 1168,
  [1347] = 468,
  [1348] = 486,
  [1349] = 1149,
  [1350] = 478,
  [1351] = 495,
  [1352] = 481,
  [1353] = 477,
  [1354] = 1011,
  [1355] = 1148,
  [1356] = 469,
  [1357] = 470,
  [1358] = 494,
  [1359] = 491,
  [1360] = 1067,
  [1361] = 1361,
  [1362] = 1068,
  [1363] = 468,
  [1364] = 491,
  [1365] = 477,
  [1366] = 478,
  [1367] = 473,
  [1368] = 1368,
  [1369] = 469,
  [1370] = 1158,
  [1371] = 468,
  [1372] = 1163,
  [1373] = 468,
  [1374] = 1149,
  [1375] = 474,
  [1376] = 1082,
  [1377] = 1146,
  [1378] = 485,
  [1379] = 1014,
  [1380] = 1146,
  [1381] = 1158,
  [1382] = 1382,
  [1383] = 473,
  [1384] = 484,
  [1385] = 490,
  [1386] = 1082,
  [1387] = 1301,
  [1388] = 1163,
  [1389] = 479,
  [1390] = 1281,
  [1391] = 1148,
  [1392] = 1011,
  [1393] = 480,
  [1394] = 1158,
  [1395] = 880,
  [1396] = 885,
  [1397] = 1081,
  [1398] = 470,
  [1399] = 1158,
  [1400] = 479,
  [1401] = 1146,
  [1402] = 487,
  [1403] = 486,
  [1404] = 495,
  [1405] = 469,
  [1406] = 480,
  [1407] = 1163,
  [1408] = 476,
  [1409] = 474,
  [1410] = 475,
  [1411] = 485,
  [1412] = 468,
  [1413] = 481,
  [1414] = 488,
  [1415] = 473,
  [1416] = 489,
  [1417] = 488,
  [1418] = 492,
  [1419] = 492,
  [1420] = 470,
  [1421] = 470,
  [1422] = 483,
  [1423] = 482,
  [1424] = 476,
  [1425] = 464,
  [1426] = 465,
  [1427] = 488,
  [1428] = 492,
  [1429] = 1429,
  [1430] = 475,
  [1431] = 1045,
  [1432] = 489,
  [1433] = 494,
  [1434] = 475,
  [1435] = 491,
  [1436] = 1149,
  [1437] = 476,
  [1438] = 477,
  [1439] = 481,
  [1440] = 482,
  [1441] = 470,
  [1442] = 483,
  [1443] = 473,
  [1444] = 473,
  [1445] = 469,
  [1446] = 1100,
  [1447] = 470,
  [1448] = 494,
  [1449] = 484,
  [1450] = 486,
  [1451] = 468,
  [1452] = 485,
  [1453] = 487,
  [1454] = 489,
  [1455] = 474,
  [1456] = 479,
  [1457] = 473,
  [1458] = 490,
  [1459] = 478,
  [1460] = 469,
  [1461] = 483,
  [1462] = 482,
  [1463] = 489,
  [1464] = 490,
  [1465] = 482,
  [1466] = 483,
  [1467] = 1011,
  [1468] = 469,
  [1469] = 480,
  [1470] = 1148,
  [1471] = 475,
  [1472] = 484,
  [1473] = 476,
  [1474] = 1474,
  [1475] = 468,
  [1476] = 485,
  [1477] = 474,
  [1478] = 480,
  [1479] = 484,
  [1480] = 1361,
  [1481] = 1068,
  [1482] = 488,
  [1483] = 487,
  [1484] = 488,
  [1485] = 474,
  [1486] = 479,
  [1487] = 478,
  [1488] = 477,
  [1489] = 490,
  [1490] = 485,
  [1491] = 491,
  [1492] = 492,
  [1493] = 482,
  [1494] = 483,
  [1495] = 475,
  [1496] = 480,
  [1497] = 485,
  [1498] = 1498,
  [1499] = 486,
  [1500] = 474,
  [1501] = 481,
  [1502] = 490,
  [1503] = 484,
  [1504] = 474,
  [1505] = 474,
  [1506] = 485,
  [1507] = 486,
  [1508] = 476,
  [1509] = 1146,
  [1510] = 486,
  [1511] = 490,
  [1512] = 1148,
  [1513] = 487,
  [1514] = 479,
  [1515] = 480,
  [1516] = 475,
  [1517] = 476,
  [1518] = 487,
  [1519] = 476,
  [1520] = 475,
  [1521] = 1182,
  [1522] = 494,
  [1523] = 1011,
  [1524] = 1163,
  [1525] = 489,
  [1526] = 1158,
  [1527] = 479,
  [1528] = 470,
  [1529] = 880,
  [1530] = 484,
  [1531] = 479,
  [1532] = 482,
  [1533] = 483,
  [1534] = 1534,
  [1535] = 495,
  [1536] = 1011,
  [1537] = 486,
  [1538] = 484,
  [1539] = 1149,
  [1540] = 481,
  [1541] = 476,
  [1542] = 1542,
  [1543] = 1281,
  [1544] = 475,
  [1545] = 491,
  [1546] = 490,
  [1547] = 489,
  [1548] = 488,
  [1549] = 492,
  [1550] = 477,
  [1551] = 495,
  [1552] = 1181,
  [1553] = 478,
  [1554] = 479,
  [1555] = 487,
  [1556] = 1556,
  [1557] = 473,
  [1558] = 1148,
  [1559] = 885,
  [1560] = 1081,
  [1561] = 1542,
  [1562] = 809,
  [1563] = 490,
  [1564] = 478,
  [1565] = 477,
  [1566] = 491,
  [1567] = 1474,
  [1568] = 483,
  [1569] = 482,
  [1570] = 485,
  [1571] = 480,
  [1572] = 469,
  [1573] = 1301,
  [1574] = 1556,
  [1575] = 1163,
  [1576] = 1576,
  [1577] = 490,
  [1578] = 478,
  [1579] = 477,
  [1580] = 480,
  [1581] = 491,
  [1582] = 468,
  [1583] = 479,
  [1584] = 478,
  [1585] = 489,
  [1586] = 492,
  [1587] = 475,
  [1588] = 491,
  [1589] = 487,
  [1590] = 482,
  [1591] = 477,
  [1592] = 477,
  [1593] = 478,
  [1594] = 491,
  [1595] = 489,
  [1596] = 832,
  [1597] = 469,
  [1598] = 474,
  [1599] = 859,
  [1600] = 489,
  [1601] = 1100,
  [1602] = 1067,
  [1603] = 1173,
  [1604] = 494,
  [1605] = 480,
  [1606] = 1429,
  [1607] = 1607,
  [1608] = 494,
  [1609] = 492,
  [1610] = 488,
  [1611] = 1168,
  [1612] = 475,
  [1613] = 1301,
  [1614] = 487,
  [1615] = 1576,
  [1616] = 480,
  [1617] = 484,
  [1618] = 1361,
  [1619] = 483,
  [1620] = 1576,
  [1621] = 483,
  [1622] = 1281,
  [1623] = 482,
  [1624] = 476,
  [1625] = 482,
  [1626] = 483,
  [1627] = 822,
  [1628] = 489,
  [1629] = 1301,
  [1630] = 1158,
  [1631] = 1631,
  [1632] = 1163,
  [1633] = 485,
  [1634] = 1634,
  [1635] = 1635,
  [1636] = 1576,
  [1637] = 470,
  [1638] = 473,
  [1639] = 1281,
  [1640] = 470,
  [1641] = 492,
  [1642] = 488,
  [1643] = 1382,
  [1644] = 475,
  [1645] = 473,
  [1646] = 494,
  [1647] = 1191,
  [1648] = 1173,
  [1649] = 468,
  [1650] = 1190,
  [1651] = 474,
  [1652] = 476,
  [1653] = 819,
  [1654] = 604,
  [1655] = 481,
  [1656] = 495,
  [1657] = 486,
  [1658] = 1474,
  [1659] = 468,
  [1660] = 470,
  [1661] = 473,
  [1662] = 469,
  [1663] = 470,
  [1664] = 473,
  [1665] = 469,
  [1666] = 485,
  [1667] = 1498,
  [1668] = 484,
  [1669] = 482,
  [1670] = 607,
  [1671] = 474,
  [1672] = 470,
  [1673] = 1168,
  [1674] = 476,
  [1675] = 469,
  [1676] = 481,
  [1677] = 490,
  [1678] = 483,
  [1679] = 1576,
  [1680] = 481,
  [1681] = 1190,
  [1682] = 486,
  [1683] = 484,
  [1684] = 481,
  [1685] = 1576,
  [1686] = 489,
  [1687] = 486,
  [1688] = 1158,
  [1689] = 484,
  [1690] = 492,
  [1691] = 481,
  [1692] = 491,
  [1693] = 477,
  [1694] = 470,
  [1695] = 473,
  [1696] = 1191,
  [1697] = 478,
  [1698] = 1576,
  [1699] = 1368,
  [1700] = 488,
  [1701] = 492,
  [1702] = 480,
  [1703] = 1703,
  [1704] = 468,
  [1705] = 469,
  [1706] = 1706,
  [1707] = 1707,
  [1708] = 1163,
  [1709] = 488,
  [1710] = 469,
  [1711] = 1158,
  [1712] = 1181,
  [1713] = 479,
  [1714] = 473,
  [1715] = 470,
  [1716] = 1182,
  [1717] = 815,
  [1718] = 487,
  [1719] = 813,
  [1720] = 1361,
  [1721] = 473,
  [1722] = 468,
  [1723] = 485,
  [1724] = 495,
  [1725] = 469,
  [1726] = 480,
  [1727] = 473,
  [1728] = 476,
  [1729] = 1542,
  [1730] = 1730,
  [1731] = 1556,
  [1732] = 475,
  [1733] = 482,
  [1734] = 483,
  [1735] = 1498,
  [1736] = 494,
  [1737] = 489,
  [1738] = 1149,
  [1739] = 1429,
  [1740] = 1382,
  [1741] = 1146,
  [1742] = 1361,
  [1743] = 1368,
  [1744] = 1474,
  [1745] = 880,
  [1746] = 485,
  [1747] = 1498,
  [1748] = 1556,
  [1749] = 1749,
  [1750] = 885,
  [1751] = 494,
  [1752] = 832,
  [1753] = 474,
  [1754] = 491,
  [1755] = 477,
  [1756] = 478,
  [1757] = 1757,
  [1758] = 859,
  [1759] = 1542,
  [1760] = 491,
  [1761] = 477,
  [1762] = 478,
  [1763] = 490,
  [1764] = 479,
  [1765] = 487,
  [1766] = 479,
  [1767] = 1534,
  [1768] = 486,
  [1769] = 488,
  [1770] = 486,
  [1771] = 492,
  [1772] = 481,
  [1773] = 488,
  [1774] = 490,
  [1775] = 492,
  [1776] = 1607,
  [1777] = 487,
  [1778] = 1749,
  [1779] = 470,
  [1780] = 473,
  [1781] = 484,
  [1782] = 480,
  [1783] = 484,
  [1784] = 854,
  [1785] = 481,
  [1786] = 469,
  [1787] = 861,
  [1788] = 480,
  [1789] = 494,
  [1790] = 480,
  [1791] = 1757,
  [1792] = 481,
  [1793] = 484,
  [1794] = 874,
  [1795] = 876,
  [1796] = 487,
  [1797] = 819,
  [1798] = 490,
  [1799] = 815,
  [1800] = 813,
  [1801] = 1474,
  [1802] = 474,
  [1803] = 1474,
  [1804] = 492,
  [1805] = 488,
  [1806] = 1757,
  [1807] = 485,
  [1808] = 575,
  [1809] = 476,
  [1810] = 486,
  [1811] = 822,
  [1812] = 475,
  [1813] = 479,
  [1814] = 494,
  [1815] = 474,
  [1816] = 1368,
  [1817] = 491,
  [1818] = 477,
  [1819] = 478,
  [1820] = 485,
  [1821] = 470,
  [1822] = 473,
  [1823] = 469,
  [1824] = 1730,
  [1825] = 489,
  [1826] = 1429,
  [1827] = 482,
  [1828] = 1149,
  [1829] = 494,
  [1830] = 1730,
  [1831] = 573,
  [1832] = 469,
  [1833] = 476,
  [1834] = 483,
  [1835] = 495,
  [1836] = 478,
  [1837] = 477,
  [1838] = 491,
  [1839] = 1158,
  [1840] = 475,
  [1841] = 479,
  [1842] = 1730,
  [1843] = 1163,
  [1844] = 495,
  [1845] = 1163,
  [1846] = 1158,
  [1847] = 476,
  [1848] = 490,
  [1849] = 474,
  [1850] = 1749,
  [1851] = 482,
  [1852] = 475,
  [1853] = 491,
  [1854] = 477,
  [1855] = 478,
  [1856] = 479,
  [1857] = 487,
  [1858] = 485,
  [1859] = 482,
  [1860] = 1631,
  [1861] = 483,
  [1862] = 486,
  [1863] = 483,
  [1864] = 489,
  [1865] = 488,
  [1866] = 492,
  [1867] = 1749,
  [1868] = 583,
  [1869] = 1730,
  [1870] = 490,
  [1871] = 1730,
  [1872] = 1281,
  [1873] = 494,
  [1874] = 486,
  [1875] = 484,
  [1876] = 481,
  [1877] = 487,
  [1878] = 489,
  [1879] = 468,
  [1880] = 488,
  [1881] = 492,
  [1882] = 809,
  [1883] = 1301,
  [1884] = 1730,
  [1885] = 495,
  [1886] = 494,
  [1887] = 1361,
  [1888] = 484,
  [1889] = 473,
  [1890] = 495,
  [1891] = 481,
  [1892] = 1634,
  [1893] = 1635,
  [1894] = 470,
  [1895] = 1382,
  [1896] = 480,
  [1897] = 470,
  [1898] = 1146,
  [1899] = 1474,
  [1900] = 1361,
  [1901] = 473,
  [1902] = 1190,
  [1903] = 1191,
  [1904] = 468,
  [1905] = 1730,
  [1906] = 1749,
  [1907] = 822,
  [1908] = 470,
  [1909] = 1703,
  [1910] = 473,
  [1911] = 468,
  [1912] = 469,
  [1913] = 1757,
  [1914] = 470,
  [1915] = 469,
  [1916] = 469,
  [1917] = 474,
  [1918] = 1158,
  [1919] = 489,
  [1920] = 476,
  [1921] = 482,
  [1922] = 1163,
  [1923] = 485,
  [1924] = 483,
  [1925] = 1749,
  [1926] = 1926,
  [1927] = 1749,
  [1928] = 475,
  [1929] = 1361,
  [1930] = 1930,
  [1931] = 1930,
  [1932] = 1932,
  [1933] = 1933,
  [1934] = 1934,
  [1935] = 1935,
  [1936] = 1936,
  [1937] = 1937,
  [1938] = 1938,
  [1939] = 1930,
  [1940] = 1940,
  [1941] = 1941,
  [1942] = 1942,
  [1943] = 1943,
  [1944] = 1944,
  [1945] = 481,
  [1946] = 480,
  [1947] = 1498,
  [1948] = 1948,
  [1949] = 1949,
  [1950] = 1950,
  [1951] = 495,
  [1952] = 1281,
  [1953] = 1534,
  [1954] = 479,
  [1955] = 486,
  [1956] = 488,
  [1957] = 1607,
  [1958] = 1301,
  [1959] = 1498,
  [1960] = 492,
  [1961] = 1556,
  [1962] = 490,
  [1963] = 1942,
  [1964] = 494,
  [1965] = 487,
  [1966] = 1631,
  [1967] = 1948,
  [1968] = 1158,
  [1969] = 1949,
  [1970] = 1950,
  [1971] = 484,
  [1972] = 1163,
  [1973] = 481,
  [1974] = 1703,
  [1975] = 1975,
  [1976] = 1932,
  [1977] = 480,
  [1978] = 1933,
  [1979] = 474,
  [1980] = 1934,
  [1981] = 1935,
  [1982] = 1982,
  [1983] = 1936,
  [1984] = 1937,
  [1985] = 1938,
  [1986] = 485,
  [1987] = 1940,
  [1988] = 1941,
  [1989] = 1942,
  [1990] = 1943,
  [1991] = 1944,
  [1992] = 491,
  [1993] = 476,
  [1994] = 495,
  [1995] = 474,
  [1996] = 491,
  [1997] = 475,
  [1998] = 1975,
  [1999] = 1932,
  [2000] = 1934,
  [2001] = 1948,
  [2002] = 485,
  [2003] = 1975,
  [2004] = 482,
  [2005] = 483,
  [2006] = 477,
  [2007] = 1975,
  [2008] = 1935,
  [2009] = 1936,
  [2010] = 1937,
  [2011] = 1938,
  [2012] = 478,
  [2013] = 1940,
  [2014] = 1932,
  [2015] = 1542,
  [2016] = 1941,
  [2017] = 1634,
  [2018] = 1942,
  [2019] = 1635,
  [2020] = 1556,
  [2021] = 1934,
  [2022] = 1935,
  [2023] = 2023,
  [2024] = 1474,
  [2025] = 489,
  [2026] = 494,
  [2027] = 1936,
  [2028] = 1944,
  [2029] = 1943,
  [2030] = 1937,
  [2031] = 1938,
  [2032] = 487,
  [2033] = 469,
  [2034] = 1940,
  [2035] = 2023,
  [2036] = 1948,
  [2037] = 478,
  [2038] = 1942,
  [2039] = 1941,
  [2040] = 1940,
  [2041] = 2041,
  [2042] = 1938,
  [2043] = 1937,
  [2044] = 1936,
  [2045] = 1935,
  [2046] = 476,
  [2047] = 1934,
  [2048] = 1933,
  [2049] = 1949,
  [2050] = 1943,
  [2051] = 1932,
  [2052] = 1944,
  [2053] = 1975,
  [2054] = 1941,
  [2055] = 1635,
  [2056] = 1950,
  [2057] = 475,
  [2058] = 479,
  [2059] = 2023,
  [2060] = 1930,
  [2061] = 2061,
  [2062] = 1943,
  [2063] = 1944,
  [2064] = 1933,
  [2065] = 477,
  [2066] = 1982,
  [2067] = 470,
  [2068] = 473,
  [2069] = 1948,
  [2070] = 1930,
  [2071] = 1542,
  [2072] = 1933,
  [2073] = 486,
  [2074] = 469,
  [2075] = 482,
  [2076] = 1930,
  [2077] = 483,
  [2078] = 1930,
  [2079] = 1930,
  [2080] = 490,
  [2081] = 488,
  [2082] = 1930,
  [2083] = 1930,
  [2084] = 1930,
  [2085] = 1930,
  [2086] = 492,
  [2087] = 1950,
  [2088] = 1930,
  [2089] = 1949,
  [2090] = 2061,
  [2091] = 492,
  [2092] = 1982,
  [2093] = 489,
  [2094] = 495,
  [2095] = 488,
  [2096] = 1930,
  [2097] = 1930,
  [2098] = 1930,
  [2099] = 1930,
  [2100] = 486,
  [2101] = 1930,
  [2102] = 1930,
  [2103] = 1498,
  [2104] = 1930,
  [2105] = 479,
  [2106] = 1556,
  [2107] = 1301,
  [2108] = 490,
  [2109] = 487,
  [2110] = 1930,
  [2111] = 484,
  [2112] = 1930,
  [2113] = 464,
  [2114] = 481,
  [2115] = 465,
  [2116] = 480,
  [2117] = 1607,
  [2118] = 494,
  [2119] = 1982,
  [2120] = 1930,
  [2121] = 489,
  [2122] = 1703,
  [2123] = 607,
  [2124] = 494,
  [2125] = 880,
  [2126] = 885,
  [2127] = 468,
  [2128] = 1930,
  [2129] = 1950,
  [2130] = 1949,
  [2131] = 1950,
  [2132] = 1949,
  [2133] = 1930,
  [2134] = 1982,
  [2135] = 473,
  [2136] = 1930,
  [2137] = 1948,
  [2138] = 604,
  [2139] = 2041,
  [2140] = 1930,
  [2141] = 1930,
  [2142] = 1930,
  [2143] = 1930,
  [2144] = 1930,
  [2145] = 1982,
  [2146] = 1930,
  [2147] = 1474,
  [2148] = 1930,
  [2149] = 1930,
  [2150] = 1930,
  [2151] = 1930,
  [2152] = 470,
  [2153] = 1930,
  [2154] = 832,
  [2155] = 859,
  [2156] = 1634,
  [2157] = 1930,
  [2158] = 1498,
  [2159] = 1930,
  [2160] = 495,
  [2161] = 1474,
  [2162] = 483,
  [2163] = 1930,
  [2164] = 1930,
  [2165] = 1930,
  [2166] = 2061,
  [2167] = 1930,
  [2168] = 1930,
  [2169] = 1930,
  [2170] = 1930,
  [2171] = 1930,
  [2172] = 1930,
  [2173] = 1542,
  [2174] = 1975,
  [2175] = 1930,
  [2176] = 1930,
  [2177] = 1930,
  [2178] = 1930,
  [2179] = 1930,
  [2180] = 1930,
  [2181] = 1930,
  [2182] = 495,
  [2183] = 1930,
  [2184] = 1950,
  [2185] = 1930,
  [2186] = 1930,
  [2187] = 1930,
  [2188] = 1949,
  [2189] = 1930,
  [2190] = 1631,
  [2191] = 1930,
  [2192] = 1932,
  [2193] = 1930,
  [2194] = 1930,
  [2195] = 1930,
  [2196] = 1930,
  [2197] = 1938,
  [2198] = 1948,
  [2199] = 854,
  [2200] = 1930,
  [2201] = 1534,
  [2202] = 2041,
  [2203] = 876,
  [2204] = 861,
  [2205] = 1941,
  [2206] = 1930,
  [2207] = 1930,
  [2208] = 1934,
  [2209] = 1935,
  [2210] = 482,
  [2211] = 874,
  [2212] = 1930,
  [2213] = 1982,
  [2214] = 478,
  [2215] = 495,
  [2216] = 477,
  [2217] = 1930,
  [2218] = 491,
  [2219] = 1930,
  [2220] = 1930,
  [2221] = 1930,
  [2222] = 475,
  [2223] = 1930,
  [2224] = 476,
  [2225] = 1944,
  [2226] = 1943,
  [2227] = 1281,
  [2228] = 1930,
  [2229] = 1942,
  [2230] = 1936,
  [2231] = 1937,
  [2232] = 1941,
  [2233] = 1940,
  [2234] = 2023,
  [2235] = 2023,
  [2236] = 1930,
  [2237] = 495,
  [2238] = 1937,
  [2239] = 1936,
  [2240] = 1938,
  [2241] = 2061,
  [2242] = 1940,
  [2243] = 2041,
  [2244] = 2061,
  [2245] = 1942,
  [2246] = 1943,
  [2247] = 1556,
  [2248] = 1935,
  [2249] = 1944,
  [2250] = 1934,
  [2251] = 2061,
  [2252] = 1933,
  [2253] = 1542,
  [2254] = 494,
  [2255] = 1930,
  [2256] = 1932,
  [2257] = 484,
  [2258] = 1975,
  [2259] = 1930,
  [2260] = 474,
  [2261] = 2041,
  [2262] = 1930,
  [2263] = 485,
  [2264] = 1930,
  [2265] = 1930,
  [2266] = 2023,
  [2267] = 2023,
  [2268] = 480,
  [2269] = 485,
  [2270] = 1474,
  [2271] = 1158,
  [2272] = 1634,
  [2273] = 1635,
  [2274] = 481,
  [2275] = 490,
  [2276] = 484,
  [2277] = 486,
  [2278] = 880,
  [2279] = 1148,
  [2280] = 495,
  [2281] = 1474,
  [2282] = 1011,
  [2283] = 880,
  [2284] = 885,
  [2285] = 470,
  [2286] = 1703,
  [2287] = 491,
  [2288] = 885,
  [2289] = 474,
  [2290] = 470,
  [2291] = 1081,
  [2292] = 480,
  [2293] = 473,
  [2294] = 476,
  [2295] = 477,
  [2296] = 468,
  [2297] = 473,
  [2298] = 469,
  [2299] = 478,
  [2300] = 1081,
  [2301] = 475,
  [2302] = 469,
  [2303] = 1100,
  [2304] = 488,
  [2305] = 495,
  [2306] = 495,
  [2307] = 492,
  [2308] = 859,
  [2309] = 1163,
  [2310] = 495,
  [2311] = 494,
  [2312] = 468,
  [2313] = 479,
  [2314] = 832,
  [2315] = 491,
  [2316] = 494,
  [2317] = 477,
  [2318] = 486,
  [2319] = 478,
  [2320] = 488,
  [2321] = 479,
  [2322] = 487,
  [2323] = 1011,
  [2324] = 1148,
  [2325] = 832,
  [2326] = 492,
  [2327] = 859,
  [2328] = 1100,
  [2329] = 470,
  [2330] = 1067,
  [2331] = 1068,
  [2332] = 926,
  [2333] = 484,
  [2334] = 489,
  [2335] = 490,
  [2336] = 473,
  [2337] = 1158,
  [2338] = 495,
  [2339] = 487,
  [2340] = 1163,
  [2341] = 483,
  [2342] = 469,
  [2343] = 481,
  [2344] = 482,
  [2345] = 494,
  [2346] = 479,
  [2347] = 469,
  [2348] = 486,
  [2349] = 2349,
  [2350] = 473,
  [2351] = 495,
  [2352] = 1146,
  [2353] = 2353,
  [2354] = 476,
  [2355] = 485,
  [2356] = 469,
  [2357] = 1100,
  [2358] = 2358,
  [2359] = 468,
  [2360] = 2349,
  [2361] = 2353,
  [2362] = 488,
  [2363] = 492,
  [2364] = 2364,
  [2365] = 1191,
  [2366] = 494,
  [2367] = 1190,
  [2368] = 1082,
  [2369] = 1158,
  [2370] = 1149,
  [2371] = 1163,
  [2372] = 2358,
  [2373] = 469,
  [2374] = 469,
  [2375] = 485,
  [2376] = 2349,
  [2377] = 2364,
  [2378] = 2364,
  [2379] = 2353,
  [2380] = 490,
  [2381] = 495,
  [2382] = 469,
  [2383] = 475,
  [2384] = 476,
  [2385] = 470,
  [2386] = 469,
  [2387] = 2387,
  [2388] = 475,
  [2389] = 469,
  [2390] = 487,
  [2391] = 473,
  [2392] = 482,
  [2393] = 483,
  [2394] = 1014,
  [2395] = 1068,
  [2396] = 469,
  [2397] = 2397,
  [2398] = 474,
  [2399] = 484,
  [2400] = 481,
  [2401] = 482,
  [2402] = 483,
  [2403] = 473,
  [2404] = 470,
  [2405] = 1011,
  [2406] = 1148,
  [2407] = 470,
  [2408] = 469,
  [2409] = 469,
  [2410] = 2364,
  [2411] = 2358,
  [2412] = 489,
  [2413] = 880,
  [2414] = 2349,
  [2415] = 885,
  [2416] = 2353,
  [2417] = 1045,
  [2418] = 495,
  [2419] = 1067,
  [2420] = 480,
  [2421] = 491,
  [2422] = 1158,
  [2423] = 477,
  [2424] = 470,
  [2425] = 468,
  [2426] = 473,
  [2427] = 1081,
  [2428] = 2358,
  [2429] = 470,
  [2430] = 478,
  [2431] = 1163,
  [2432] = 859,
  [2433] = 469,
  [2434] = 1011,
  [2435] = 468,
  [2436] = 832,
  [2437] = 1163,
  [2438] = 1158,
  [2439] = 474,
  [2440] = 1148,
  [2441] = 473,
  [2442] = 489,
  [2443] = 479,
  [2444] = 484,
  [2445] = 2445,
  [2446] = 478,
  [2447] = 489,
  [2448] = 477,
  [2449] = 483,
  [2450] = 482,
  [2451] = 2451,
  [2452] = 487,
  [2453] = 469,
  [2454] = 481,
  [2455] = 2455,
  [2456] = 2451,
  [2457] = 1474,
  [2458] = 469,
  [2459] = 470,
  [2460] = 469,
  [2461] = 2455,
  [2462] = 475,
  [2463] = 2445,
  [2464] = 468,
  [2465] = 490,
  [2466] = 470,
  [2467] = 492,
  [2468] = 488,
  [2469] = 480,
  [2470] = 470,
  [2471] = 473,
  [2472] = 2472,
  [2473] = 473,
  [2474] = 494,
  [2475] = 473,
  [2476] = 1429,
  [2477] = 486,
  [2478] = 476,
  [2479] = 474,
  [2480] = 469,
  [2481] = 485,
  [2482] = 485,
  [2483] = 468,
  [2484] = 2472,
  [2485] = 481,
  [2486] = 468,
  [2487] = 1158,
  [2488] = 473,
  [2489] = 1158,
  [2490] = 1163,
  [2491] = 470,
  [2492] = 495,
  [2493] = 1163,
  [2494] = 470,
  [2495] = 470,
  [2496] = 479,
  [2497] = 2455,
  [2498] = 476,
  [2499] = 479,
  [2500] = 2451,
  [2501] = 2445,
  [2502] = 475,
  [2503] = 2451,
  [2504] = 474,
  [2505] = 486,
  [2506] = 2397,
  [2507] = 480,
  [2508] = 481,
  [2509] = 2387,
  [2510] = 484,
  [2511] = 468,
  [2512] = 487,
  [2513] = 490,
  [2514] = 482,
  [2515] = 492,
  [2516] = 488,
  [2517] = 486,
  [2518] = 483,
  [2519] = 487,
  [2520] = 479,
  [2521] = 2451,
  [2522] = 1301,
  [2523] = 486,
  [2524] = 490,
  [2525] = 489,
  [2526] = 478,
  [2527] = 477,
  [2528] = 491,
  [2529] = 478,
  [2530] = 477,
  [2531] = 491,
  [2532] = 1149,
  [2533] = 473,
  [2534] = 1281,
  [2535] = 478,
  [2536] = 492,
  [2537] = 469,
  [2538] = 1191,
  [2539] = 1190,
  [2540] = 470,
  [2541] = 1382,
  [2542] = 477,
  [2543] = 473,
  [2544] = 474,
  [2545] = 485,
  [2546] = 2455,
  [2547] = 2445,
  [2548] = 468,
  [2549] = 2455,
  [2550] = 470,
  [2551] = 473,
  [2552] = 2445,
  [2553] = 473,
  [2554] = 491,
  [2555] = 470,
  [2556] = 1168,
  [2557] = 494,
  [2558] = 1182,
  [2559] = 2559,
  [2560] = 480,
  [2561] = 2445,
  [2562] = 1181,
  [2563] = 2455,
  [2564] = 470,
  [2565] = 476,
  [2566] = 489,
  [2567] = 488,
  [2568] = 494,
  [2569] = 1146,
  [2570] = 491,
  [2571] = 481,
  [2572] = 1158,
  [2573] = 2451,
  [2574] = 475,
  [2575] = 473,
  [2576] = 484,
  [2577] = 487,
  [2578] = 1163,
  [2579] = 483,
  [2580] = 482,
  [2581] = 490,
  [2582] = 473,
  [2583] = 484,
  [2584] = 495,
  [2585] = 489,
  [2586] = 484,
  [2587] = 483,
  [2588] = 2451,
  [2589] = 2589,
  [2590] = 482,
  [2591] = 475,
  [2592] = 476,
  [2593] = 487,
  [2594] = 2594,
  [2595] = 494,
  [2596] = 478,
  [2597] = 477,
  [2598] = 491,
  [2599] = 485,
  [2600] = 474,
  [2601] = 2559,
  [2602] = 494,
  [2603] = 490,
  [2604] = 480,
  [2605] = 481,
  [2606] = 484,
  [2607] = 487,
  [2608] = 490,
  [2609] = 492,
  [2610] = 488,
  [2611] = 486,
  [2612] = 1163,
  [2613] = 1158,
  [2614] = 479,
  [2615] = 478,
  [2616] = 477,
  [2617] = 491,
  [2618] = 470,
  [2619] = 473,
  [2620] = 468,
  [2621] = 2621,
  [2622] = 1607,
  [2623] = 1301,
  [2624] = 492,
  [2625] = 488,
  [2626] = 491,
  [2627] = 2627,
  [2628] = 477,
  [2629] = 478,
  [2630] = 486,
  [2631] = 480,
  [2632] = 469,
  [2633] = 483,
  [2634] = 482,
  [2635] = 1281,
  [2636] = 2636,
  [2637] = 1534,
  [2638] = 489,
  [2639] = 475,
  [2640] = 476,
  [2641] = 2472,
  [2642] = 479,
  [2643] = 485,
  [2644] = 474,
  [2645] = 486,
  [2646] = 488,
  [2647] = 479,
  [2648] = 2648,
  [2649] = 492,
  [2650] = 490,
  [2651] = 480,
  [2652] = 487,
  [2653] = 492,
  [2654] = 479,
  [2655] = 488,
  [2656] = 481,
  [2657] = 486,
  [2658] = 487,
  [2659] = 2659,
  [2660] = 479,
  [2661] = 486,
  [2662] = 490,
  [2663] = 484,
  [2664] = 488,
  [2665] = 492,
  [2666] = 481,
  [2667] = 480,
  [2668] = 478,
  [2669] = 477,
  [2670] = 491,
  [2671] = 1474,
  [2672] = 481,
  [2673] = 490,
  [2674] = 486,
  [2675] = 474,
  [2676] = 487,
  [2677] = 479,
  [2678] = 1361,
  [2679] = 480,
  [2680] = 487,
  [2681] = 485,
  [2682] = 490,
  [2683] = 474,
  [2684] = 478,
  [2685] = 485,
  [2686] = 477,
  [2687] = 491,
  [2688] = 1368,
  [2689] = 1146,
  [2690] = 2472,
  [2691] = 1634,
  [2692] = 494,
  [2693] = 1635,
  [2694] = 2694,
  [2695] = 494,
  [2696] = 2696,
  [2697] = 484,
  [2698] = 481,
  [2699] = 494,
  [2700] = 476,
  [2701] = 475,
  [2702] = 2702,
  [2703] = 1149,
  [2704] = 2704,
  [2705] = 465,
  [2706] = 464,
  [2707] = 482,
  [2708] = 478,
  [2709] = 477,
  [2710] = 491,
  [2711] = 483,
  [2712] = 489,
  [2713] = 495,
  [2714] = 1361,
  [2715] = 1474,
  [2716] = 489,
  [2717] = 2472,
  [2718] = 1429,
  [2719] = 476,
  [2720] = 473,
  [2721] = 474,
  [2722] = 475,
  [2723] = 470,
  [2724] = 2724,
  [2725] = 495,
  [2726] = 1382,
  [2727] = 494,
  [2728] = 481,
  [2729] = 483,
  [2730] = 482,
  [2731] = 475,
  [2732] = 482,
  [2733] = 483,
  [2734] = 489,
  [2735] = 476,
  [2736] = 485,
  [2737] = 486,
  [2738] = 2738,
  [2739] = 2739,
  [2740] = 2739,
  [2741] = 2739,
  [2742] = 2739,
  [2743] = 2739,
  [2744] = 2739,
  [2745] = 2739,
  [2746] = 2739,
  [2747] = 2739,
  [2748] = 2739,
  [2749] = 2739,
  [2750] = 2739,
  [2751] = 2739,
  [2752] = 2739,
  [2753] = 2739,
  [2754] = 2739,
  [2755] = 2739,
  [2756] = 2739,
  [2757] = 2739,
  [2758] = 2739,
  [2759] = 2739,
  [2760] = 2739,
  [2761] = 2739,
  [2762] = 2739,
  [2763] = 2739,
  [2764] = 2739,
  [2765] = 2739,
  [2766] = 2739,
  [2767] = 2739,
  [2768] = 2739,
  [2769] = 2739,
  [2770] = 2739,
  [2771] = 2739,
  [2772] = 2739,
  [2773] = 2739,
  [2774] = 2739,
  [2775] = 2739,
  [2776] = 1474,
  [2777] = 2739,
  [2778] = 2739,
  [2779] = 2739,
  [2780] = 1635,
  [2781] = 2739,
  [2782] = 2739,
  [2783] = 2739,
  [2784] = 2739,
  [2785] = 2739,
  [2786] = 2739,
  [2787] = 2739,
  [2788] = 2739,
  [2789] = 2739,
  [2790] = 2739,
  [2791] = 2739,
  [2792] = 2739,
  [2793] = 2739,
  [2794] = 2739,
  [2795] = 2739,
  [2796] = 495,
  [2797] = 2739,
  [2798] = 495,
  [2799] = 2739,
  [2800] = 2739,
  [2801] = 2801,
  [2802] = 2802,
  [2803] = 2803,
  [2804] = 2739,
  [2805] = 2724,
  [2806] = 494,
  [2807] = 1556,
  [2808] = 2808,
  [2809] = 2809,
  [2810] = 2810,
  [2811] = 2739,
  [2812] = 2812,
  [2813] = 2813,
  [2814] = 2814,
  [2815] = 2808,
  [2816] = 2816,
  [2817] = 1542,
  [2818] = 490,
  [2819] = 495,
  [2820] = 2739,
  [2821] = 491,
  [2822] = 477,
  [2823] = 478,
  [2824] = 479,
  [2825] = 2825,
  [2826] = 487,
  [2827] = 2827,
  [2828] = 2828,
  [2829] = 2829,
  [2830] = 2830,
  [2831] = 2831,
  [2832] = 2832,
  [2833] = 2808,
  [2834] = 2808,
  [2835] = 2835,
  [2836] = 2739,
  [2837] = 2837,
  [2838] = 2838,
  [2839] = 2627,
  [2840] = 2840,
  [2841] = 483,
  [2842] = 482,
  [2843] = 2843,
  [2844] = 2739,
  [2845] = 2845,
  [2846] = 1607,
  [2847] = 486,
  [2848] = 2848,
  [2849] = 2849,
  [2850] = 489,
  [2851] = 2808,
  [2852] = 484,
  [2853] = 481,
  [2854] = 475,
  [2855] = 488,
  [2856] = 492,
  [2857] = 476,
  [2858] = 2739,
  [2859] = 485,
  [2860] = 2860,
  [2861] = 474,
  [2862] = 2559,
  [2863] = 2594,
  [2864] = 1534,
  [2865] = 2559,
  [2866] = 480,
  [2867] = 494,
  [2868] = 2868,
  [2869] = 480,
  [2870] = 484,
  [2871] = 481,
  [2872] = 486,
  [2873] = 487,
  [2874] = 2808,
  [2875] = 479,
  [2876] = 490,
  [2877] = 492,
  [2878] = 1498,
  [2879] = 488,
  [2880] = 2739,
  [2881] = 478,
  [2882] = 477,
  [2883] = 491,
  [2884] = 1542,
  [2885] = 481,
  [2886] = 1634,
  [2887] = 464,
  [2888] = 465,
  [2889] = 2659,
  [2890] = 2890,
  [2891] = 487,
  [2892] = 479,
  [2893] = 490,
  [2894] = 2894,
  [2895] = 464,
  [2896] = 465,
  [2897] = 2648,
  [2898] = 2898,
  [2899] = 2899,
  [2900] = 2636,
  [2901] = 481,
  [2902] = 464,
  [2903] = 465,
  [2904] = 486,
  [2905] = 1281,
  [2906] = 487,
  [2907] = 479,
  [2908] = 478,
  [2909] = 477,
  [2910] = 491,
  [2911] = 490,
  [2912] = 494,
  [2913] = 1556,
  [2914] = 1474,
  [2915] = 2915,
  [2916] = 2916,
  [2917] = 2917,
  [2918] = 478,
  [2919] = 477,
  [2920] = 491,
  [2921] = 2921,
  [2922] = 2922,
  [2923] = 2923,
  [2924] = 2924,
  [2925] = 2621,
  [2926] = 2926,
  [2927] = 495,
  [2928] = 495,
  [2929] = 2929,
  [2930] = 2930,
  [2931] = 2931,
  [2932] = 2739,
  [2933] = 2933,
  [2934] = 1498,
  [2935] = 495,
  [2936] = 2936,
  [2937] = 2937,
  [2938] = 2938,
  [2939] = 1631,
  [2940] = 1301,
  [2941] = 1703,
  [2942] = 2942,
  [2943] = 2828,
  [2944] = 2944,
  [2945] = 2589,
  [2946] = 2946,
  [2947] = 2947,
  [2948] = 2948,
  [2949] = 2838,
  [2950] = 2648,
  [2951] = 2868,
  [2952] = 2952,
  [2953] = 2659,
  [2954] = 2627,
  [2955] = 2868,
  [2956] = 2956,
  [2957] = 2810,
  [2958] = 2812,
  [2959] = 2813,
  [2960] = 2814,
  [2961] = 2816,
  [2962] = 2837,
  [2963] = 2827,
  [2964] = 2828,
  [2965] = 2965,
  [2966] = 2829,
  [2967] = 2810,
  [2968] = 2812,
  [2969] = 2813,
  [2970] = 2814,
  [2971] = 2816,
  [2972] = 2827,
  [2973] = 2828,
  [2974] = 2829,
  [2975] = 2830,
  [2976] = 2802,
  [2977] = 2831,
  [2978] = 2832,
  [2979] = 2835,
  [2980] = 495,
  [2981] = 2830,
  [2982] = 2944,
  [2983] = 2704,
  [2984] = 2702,
  [2985] = 2802,
  [2986] = 2831,
  [2987] = 2813,
  [2988] = 2832,
  [2989] = 2835,
  [2990] = 2636,
  [2991] = 2991,
  [2992] = 2992,
  [2993] = 2993,
  [2994] = 2837,
  [2995] = 2835,
  [2996] = 2832,
  [2997] = 480,
  [2998] = 2627,
  [2999] = 2999,
  [3000] = 2802,
  [3001] = 2830,
  [3002] = 2829,
  [3003] = 2828,
  [3004] = 2827,
  [3005] = 3005,
  [3006] = 2924,
  [3007] = 2816,
  [3008] = 2814,
  [3009] = 2621,
  [3010] = 2813,
  [3011] = 2738,
  [3012] = 2933,
  [3013] = 2812,
  [3014] = 2810,
  [3015] = 465,
  [3016] = 464,
  [3017] = 3017,
  [3018] = 465,
  [3019] = 3019,
  [3020] = 481,
  [3021] = 484,
  [3022] = 2942,
  [3023] = 2810,
  [3024] = 491,
  [3025] = 477,
  [3026] = 478,
  [3027] = 464,
  [3028] = 2696,
  [3029] = 2694,
  [3030] = 2812,
  [3031] = 2838,
  [3032] = 3032,
  [3033] = 3033,
  [3034] = 2831,
  [3035] = 2589,
  [3036] = 3036,
  [3037] = 2942,
  [3038] = 3038,
  [3039] = 2814,
  [3040] = 3040,
  [3041] = 2933,
  [3042] = 3042,
  [3043] = 2738,
  [3044] = 464,
  [3045] = 487,
  [3046] = 2816,
  [3047] = 465,
  [3048] = 2924,
  [3049] = 2724,
  [3050] = 2868,
  [3051] = 2827,
  [3052] = 3052,
  [3053] = 2594,
  [3054] = 2837,
  [3055] = 495,
  [3056] = 3056,
  [3057] = 3038,
  [3058] = 2724,
  [3059] = 2829,
  [3060] = 2868,
  [3061] = 2830,
  [3062] = 2802,
  [3063] = 2831,
  [3064] = 2832,
  [3065] = 494,
  [3066] = 3066,
  [3067] = 2835,
  [3068] = 2837,
  [3069] = 2694,
  [3070] = 3032,
  [3071] = 2594,
  [3072] = 2696,
  [3073] = 3033,
  [3074] = 3074,
  [3075] = 3075,
  [3076] = 490,
  [3077] = 2810,
  [3078] = 2812,
  [3079] = 2813,
  [3080] = 2814,
  [3081] = 2816,
  [3082] = 2827,
  [3083] = 2828,
  [3084] = 2829,
  [3085] = 3052,
  [3086] = 2830,
  [3087] = 3040,
  [3088] = 2802,
  [3089] = 2933,
  [3090] = 2831,
  [3091] = 3042,
  [3092] = 1163,
  [3093] = 1158,
  [3094] = 2832,
  [3095] = 2835,
  [3096] = 2738,
  [3097] = 2837,
  [3098] = 2933,
  [3099] = 2702,
  [3100] = 2738,
  [3101] = 2924,
  [3102] = 2924,
  [3103] = 2704,
  [3104] = 2942,
  [3105] = 2944,
  [3106] = 3038,
  [3107] = 2924,
  [3108] = 2738,
  [3109] = 3066,
  [3110] = 469,
  [3111] = 3042,
  [3112] = 2933,
  [3113] = 3113,
  [3114] = 1474,
  [3115] = 468,
  [3116] = 3116,
  [3117] = 473,
  [3118] = 3118,
  [3119] = 3119,
  [3120] = 470,
  [3121] = 2636,
  [3122] = 3032,
  [3123] = 3033,
  [3124] = 3052,
  [3125] = 3040,
  [3126] = 3042,
  [3127] = 3127,
  [3128] = 3040,
  [3129] = 2942,
  [3130] = 2621,
  [3131] = 3052,
  [3132] = 464,
  [3133] = 2944,
  [3134] = 465,
  [3135] = 3038,
  [3136] = 3066,
  [3137] = 3032,
  [3138] = 2648,
  [3139] = 3033,
  [3140] = 3052,
  [3141] = 3040,
  [3142] = 3042,
  [3143] = 2942,
  [3144] = 3033,
  [3145] = 3052,
  [3146] = 2659,
  [3147] = 3040,
  [3148] = 2942,
  [3149] = 495,
  [3150] = 3033,
  [3151] = 2868,
  [3152] = 3052,
  [3153] = 3040,
  [3154] = 3033,
  [3155] = 3052,
  [3156] = 3040,
  [3157] = 479,
  [3158] = 3033,
  [3159] = 488,
  [3160] = 492,
  [3161] = 486,
  [3162] = 464,
  [3163] = 465,
  [3164] = 3066,
  [3165] = 3165,
  [3166] = 3165,
  [3167] = 487,
  [3168] = 2931,
  [3169] = 2930,
  [3170] = 479,
  [3171] = 478,
  [3172] = 477,
  [3173] = 491,
  [3174] = 486,
  [3175] = 484,
  [3176] = 481,
  [3177] = 492,
  [3178] = 478,
  [3179] = 477,
  [3180] = 491,
  [3181] = 488,
  [3182] = 3165,
  [3183] = 2924,
  [3184] = 480,
  [3185] = 3165,
  [3186] = 465,
  [3187] = 464,
  [3188] = 2845,
  [3189] = 3056,
  [3190] = 2931,
  [3191] = 477,
  [3192] = 3165,
  [3193] = 2738,
  [3194] = 2930,
  [3195] = 3195,
  [3196] = 3005,
  [3197] = 2933,
  [3198] = 2999,
  [3199] = 2993,
  [3200] = 2992,
  [3201] = 2991,
  [3202] = 3019,
  [3203] = 487,
  [3204] = 3165,
  [3205] = 2965,
  [3206] = 495,
  [3207] = 2956,
  [3208] = 2929,
  [3209] = 3165,
  [3210] = 488,
  [3211] = 486,
  [3212] = 3165,
  [3213] = 492,
  [3214] = 2952,
  [3215] = 484,
  [3216] = 2948,
  [3217] = 3036,
  [3218] = 2947,
  [3219] = 2926,
  [3220] = 481,
  [3221] = 486,
  [3222] = 479,
  [3223] = 488,
  [3224] = 3165,
  [3225] = 487,
  [3226] = 479,
  [3227] = 465,
  [3228] = 2946,
  [3229] = 490,
  [3230] = 464,
  [3231] = 2929,
  [3232] = 2926,
  [3233] = 3233,
  [3234] = 478,
  [3235] = 3165,
  [3236] = 3236,
  [3237] = 465,
  [3238] = 464,
  [3239] = 3165,
  [3240] = 465,
  [3241] = 492,
  [3242] = 2938,
  [3243] = 464,
  [3244] = 3165,
  [3245] = 484,
  [3246] = 2937,
  [3247] = 2936,
  [3248] = 2923,
  [3249] = 2922,
  [3250] = 478,
  [3251] = 491,
  [3252] = 3165,
  [3253] = 477,
  [3254] = 480,
  [3255] = 479,
  [3256] = 491,
  [3257] = 3165,
  [3258] = 2921,
  [3259] = 2917,
  [3260] = 3165,
  [3261] = 2916,
  [3262] = 2915,
  [3263] = 2899,
  [3264] = 2898,
  [3265] = 2894,
  [3266] = 2890,
  [3267] = 480,
  [3268] = 3165,
  [3269] = 465,
  [3270] = 464,
  [3271] = 2860,
  [3272] = 3074,
  [3273] = 3165,
  [3274] = 3165,
  [3275] = 480,
  [3276] = 3075,
  [3277] = 2849,
  [3278] = 2848,
  [3279] = 481,
  [3280] = 481,
  [3281] = 2843,
  [3282] = 484,
  [3283] = 2840,
  [3284] = 3165,
  [3285] = 3113,
  [3286] = 3116,
  [3287] = 480,
  [3288] = 490,
  [3289] = 3165,
  [3290] = 3118,
  [3291] = 487,
  [3292] = 3119,
  [3293] = 3165,
  [3294] = 481,
  [3295] = 484,
  [3296] = 3165,
  [3297] = 490,
  [3298] = 464,
  [3299] = 3165,
  [3300] = 490,
  [3301] = 3165,
  [3302] = 3165,
  [3303] = 3165,
  [3304] = 3165,
  [3305] = 2837,
  [3306] = 492,
  [3307] = 488,
  [3308] = 3165,
  [3309] = 3165,
  [3310] = 3310,
  [3311] = 474,
  [3312] = 486,
  [3313] = 465,
  [3314] = 2835,
  [3315] = 3165,
  [3316] = 3165,
  [3317] = 3165,
  [3318] = 2832,
  [3319] = 2831,
  [3320] = 3165,
  [3321] = 3165,
  [3322] = 479,
  [3323] = 2802,
  [3324] = 3165,
  [3325] = 3165,
  [3326] = 3165,
  [3327] = 2830,
  [3328] = 3165,
  [3329] = 3165,
  [3330] = 2829,
  [3331] = 485,
  [3332] = 2828,
  [3333] = 2827,
  [3334] = 3165,
  [3335] = 2816,
  [3336] = 3165,
  [3337] = 3165,
  [3338] = 3165,
  [3339] = 3165,
  [3340] = 3165,
  [3341] = 3165,
  [3342] = 3165,
  [3343] = 3165,
  [3344] = 3165,
  [3345] = 2814,
  [3346] = 3165,
  [3347] = 2813,
  [3348] = 3165,
  [3349] = 3165,
  [3350] = 3195,
  [3351] = 2812,
  [3352] = 478,
  [3353] = 477,
  [3354] = 491,
  [3355] = 3165,
  [3356] = 2810,
  [3357] = 3165,
  [3358] = 2868,
  [3359] = 2825,
  [3360] = 476,
  [3361] = 2809,
  [3362] = 3165,
  [3363] = 3165,
  [3364] = 3165,
  [3365] = 3365,
  [3366] = 3165,
  [3367] = 2803,
  [3368] = 3365,
  [3369] = 3165,
  [3370] = 3165,
  [3371] = 3165,
  [3372] = 3165,
  [3373] = 3165,
  [3374] = 3165,
  [3375] = 2801,
  [3376] = 475,
  [3377] = 3310,
  [3378] = 3378,
  [3379] = 2845,
  [3380] = 3056,
  [3381] = 3005,
  [3382] = 3233,
  [3383] = 3236,
  [3384] = 2999,
  [3385] = 3310,
  [3386] = 3378,
  [3387] = 2993,
  [3388] = 3388,
  [3389] = 3389,
  [3390] = 2992,
  [3391] = 2801,
  [3392] = 2991,
  [3393] = 489,
  [3394] = 3019,
  [3395] = 2965,
  [3396] = 2803,
  [3397] = 3397,
  [3398] = 2809,
  [3399] = 2956,
  [3400] = 2825,
  [3401] = 3195,
  [3402] = 2952,
  [3403] = 2948,
  [3404] = 3036,
  [3405] = 2947,
  [3406] = 2946,
  [3407] = 3389,
  [3408] = 3365,
  [3409] = 2938,
  [3410] = 2937,
  [3411] = 3233,
  [3412] = 2936,
  [3413] = 2923,
  [3414] = 465,
  [3415] = 482,
  [3416] = 3236,
  [3417] = 464,
  [3418] = 483,
  [3419] = 3310,
  [3420] = 3378,
  [3421] = 2922,
  [3422] = 3388,
  [3423] = 3389,
  [3424] = 3388,
  [3425] = 2921,
  [3426] = 3165,
  [3427] = 487,
  [3428] = 2917,
  [3429] = 2916,
  [3430] = 2915,
  [3431] = 2899,
  [3432] = 2898,
  [3433] = 2894,
  [3434] = 2890,
  [3435] = 3165,
  [3436] = 2860,
  [3437] = 3389,
  [3438] = 3388,
  [3439] = 486,
  [3440] = 3378,
  [3441] = 3119,
  [3442] = 3118,
  [3443] = 3116,
  [3444] = 3113,
  [3445] = 488,
  [3446] = 2840,
  [3447] = 2843,
  [3448] = 492,
  [3449] = 2848,
  [3450] = 2849,
  [3451] = 3233,
  [3452] = 3236,
  [3453] = 490,
  [3454] = 3075,
  [3455] = 3074,
  [3456] = 832,
  [3457] = 465,
  [3458] = 3458,
  [3459] = 470,
  [3460] = 3458,
  [3461] = 473,
  [3462] = 3462,
  [3463] = 465,
  [3464] = 464,
  [3465] = 1011,
  [3466] = 1148,
  [3467] = 880,
  [3468] = 468,
  [3469] = 1474,
  [3470] = 3470,
  [3471] = 1100,
  [3472] = 885,
  [3473] = 494,
  [3474] = 464,
  [3475] = 3458,
  [3476] = 469,
  [3477] = 3477,
  [3478] = 583,
  [3479] = 1163,
  [3480] = 1158,
  [3481] = 1081,
  [3482] = 465,
  [3483] = 859,
  [3484] = 3458,
  [3485] = 464,
  [3486] = 3486,
  [3487] = 3487,
  [3488] = 481,
  [3489] = 3487,
  [3490] = 479,
  [3491] = 3491,
  [3492] = 486,
  [3493] = 3487,
  [3494] = 488,
  [3495] = 3487,
  [3496] = 3487,
  [3497] = 3487,
  [3498] = 492,
  [3499] = 3487,
  [3500] = 3487,
  [3501] = 476,
  [3502] = 3487,
  [3503] = 3487,
  [3504] = 491,
  [3505] = 477,
  [3506] = 478,
  [3507] = 3487,
  [3508] = 483,
  [3509] = 3487,
  [3510] = 482,
  [3511] = 495,
  [3512] = 475,
  [3513] = 474,
  [3514] = 3487,
  [3515] = 485,
  [3516] = 480,
  [3517] = 3487,
  [3518] = 3487,
  [3519] = 3487,
  [3520] = 3487,
  [3521] = 3487,
  [3522] = 490,
  [3523] = 3487,
  [3524] = 3487,
  [3525] = 489,
  [3526] = 3487,
  [3527] = 3487,
  [3528] = 487,
  [3529] = 484,
  [3530] = 3487,
  [3531] = 3531,
  [3532] = 3532,
  [3533] = 3533,
  [3534] = 3534,
  [3535] = 464,
  [3536] = 1361,
  [3537] = 465,
  [3538] = 464,
  [3539] = 1146,
  [3540] = 465,
  [3541] = 1382,
  [3542] = 1149,
  [3543] = 1474,
  [3544] = 494,
  [3545] = 1429,
  [3546] = 3546,
  [3547] = 1301,
  [3548] = 3548,
  [3549] = 3546,
  [3550] = 3546,
  [3551] = 3548,
  [3552] = 495,
  [3553] = 3548,
  [3554] = 3546,
  [3555] = 3555,
  [3556] = 1534,
  [3557] = 3548,
  [3558] = 3548,
  [3559] = 464,
  [3560] = 465,
  [3561] = 3546,
  [3562] = 465,
  [3563] = 3548,
  [3564] = 3546,
  [3565] = 3565,
  [3566] = 3548,
  [3567] = 3567,
  [3568] = 3546,
  [3569] = 1281,
  [3570] = 3565,
  [3571] = 3571,
  [3572] = 464,
  [3573] = 1607,
  [3574] = 3574,
  [3575] = 3548,
  [3576] = 3546,
  [3577] = 3546,
  [3578] = 1542,
  [3579] = 464,
  [3580] = 465,
  [3581] = 3581,
  [3582] = 464,
  [3583] = 3567,
  [3584] = 464,
  [3585] = 3565,
  [3586] = 3548,
  [3587] = 3546,
  [3588] = 3546,
  [3589] = 3548,
  [3590] = 3546,
  [3591] = 3567,
  [3592] = 3548,
  [3593] = 3548,
  [3594] = 3548,
  [3595] = 3546,
  [3596] = 3567,
  [3597] = 1556,
  [3598] = 465,
  [3599] = 3548,
  [3600] = 465,
  [3601] = 1498,
  [3602] = 3548,
  [3603] = 3565,
  [3604] = 583,
  [3605] = 3567,
  [3606] = 3546,
  [3607] = 465,
  [3608] = 3565,
  [3609] = 3609,
  [3610] = 3546,
  [3611] = 3548,
  [3612] = 464,
  [3613] = 3546,
  [3614] = 3614,
  [3615] = 3615,
  [3616] = 3616,
  [3617] = 3614,
  [3618] = 3618,
  [3619] = 465,
  [3620] = 3620,
  [3621] = 3621,
  [3622] = 3618,
  [3623] = 3621,
  [3624] = 3618,
  [3625] = 3618,
  [3626] = 3621,
  [3627] = 3621,
  [3628] = 3628,
  [3629] = 3618,
  [3630] = 3621,
  [3631] = 3621,
  [3632] = 3621,
  [3633] = 3618,
  [3634] = 3618,
  [3635] = 3618,
  [3636] = 464,
  [3637] = 3618,
  [3638] = 3621,
  [3639] = 3621,
  [3640] = 3618,
  [3641] = 3621,
  [3642] = 3618,
  [3643] = 3615,
  [3644] = 3616,
  [3645] = 3618,
  [3646] = 3621,
  [3647] = 3618,
  [3648] = 3621,
  [3649] = 3618,
  [3650] = 3621,
  [3651] = 3618,
  [3652] = 3621,
  [3653] = 3618,
  [3654] = 3618,
  [3655] = 3621,
  [3656] = 3616,
  [3657] = 3615,
  [3658] = 3615,
  [3659] = 3621,
  [3660] = 3620,
  [3661] = 3618,
  [3662] = 3621,
  [3663] = 3616,
  [3664] = 465,
  [3665] = 3618,
  [3666] = 464,
  [3667] = 3621,
  [3668] = 3616,
  [3669] = 3618,
  [3670] = 3618,
  [3671] = 3615,
  [3672] = 3620,
  [3673] = 3621,
  [3674] = 3621,
  [3675] = 3621,
  [3676] = 3614,
  [3677] = 3618,
  [3678] = 3618,
  [3679] = 3621,
  [3680] = 3621,
  [3681] = 3621,
  [3682] = 3620,
  [3683] = 3616,
  [3684] = 3618,
  [3685] = 3615,
  [3686] = 3620,
  [3687] = 3621,
  [3688] = 3618,
  [3689] = 3618,
  [3690] = 3618,
  [3691] = 3621,
  [3692] = 3614,
  [3693] = 3621,
  [3694] = 3621,
  [3695] = 3620,
  [3696] = 3618,
  [3697] = 3620,
  [3698] = 3698,
  [3699] = 3699,
  [3700] = 3700,
  [3701] = 3700,
  [3702] = 3702,
  [3703] = 3703,
  [3704] = 3702,
  [3705] = 3705,
  [3706] = 3706,
  [3707] = 3702,
  [3708] = 3699,
  [3709] = 3705,
  [3710] = 3699,
  [3711] = 3698,
  [3712] = 3700,
  [3713] = 3705,
  [3714] = 3699,
  [3715] = 3703,
  [3716] = 3703,
  [3717] = 3702,
  [3718] = 3706,
  [3719] = 3698,
  [3720] = 3706,
  [3721] = 3706,
  [3722] = 3705,
  [3723] = 3698,
  [3724] = 3699,
  [3725] = 3703,
  [3726] = 3705,
  [3727] = 3700,
  [3728] = 3699,
  [3729] = 3699,
  [3730] = 3705,
  [3731] = 3731,
  [3732] = 3702,
  [3733] = 3702,
  [3734] = 3703,
  [3735] = 3698,
  [3736] = 3706,
  [3737] = 3700,
  [3738] = 3702,
  [3739] = 3703,
  [3740] = 3706,
  [3741] = 3699,
  [3742] = 3705,
  [3743] = 3705,
  [3744] = 3699,
  [3745] = 3698,
  [3746] = 3700,
  [3747] = 3698,
  [3748] = 3705,
  [3749] = 3699,
  [3750] = 3700,
  [3751] = 3705,
  [3752] = 3698,
  [3753] = 3699,
  [3754] = 3705,
  [3755] = 3699,
  [3756] = 3700,
  [3757] = 3700,
  [3758] = 3702,
  [3759] = 3705,
  [3760] = 3698,
  [3761] = 3706,
  [3762] = 3703,
  [3763] = 3703,
  [3764] = 3699,
  [3765] = 3706,
  [3766] = 3702,
  [3767] = 3698,
  [3768] = 3702,
  [3769] = 3703,
  [3770] = 3698,
  [3771] = 3698,
  [3772] = 3706,
  [3773] = 3700,
  [3774] = 3702,
  [3775] = 3703,
  [3776] = 3702,
  [3777] = 3706,
  [3778] = 3702,
  [3779] = 3706,
  [3780] = 3703,
  [3781] = 3698,
  [3782] = 3706,
  [3783] = 3705,
  [3784] = 3699,
  [3785] = 3700,
  [3786] = 3699,
  [3787] = 3703,
  [3788] = 3702,
  [3789] = 3705,
  [3790] = 3700,
  [3791] = 3698,
  [3792] = 3700,
  [3793] = 3705,
  [3794] = 3698,
  [3795] = 3700,
  [3796] = 3698,
  [3797] = 3699,
  [3798] = 3702,
  [3799] = 3703,
  [3800] = 3800,
  [3801] = 3706,
  [3802] = 3702,
  [3803] = 3698,
  [3804] = 3705,
  [3805] = 3703,
  [3806] = 3700,
  [3807] = 3706,
  [3808] = 3703,
  [3809] = 3699,
  [3810] = 3698,
  [3811] = 3698,
  [3812] = 3705,
  [3813] = 3700,
  [3814] = 3700,
  [3815] = 3702,
  [3816] = 3699,
  [3817] = 3699,
  [3818] = 3705,
  [3819] = 3700,
  [3820] = 3705,
  [3821] = 3703,
  [3822] = 3706,
  [3823] = 3702,
  [3824] = 3699,
  [3825] = 3705,
  [3826] = 3698,
  [3827] = 3827,
  [3828] = 3706,
  [3829] = 3702,
  [3830] = 3699,
  [3831] = 3705,
  [3832] = 3703,
  [3833] = 3706,
  [3834] = 3700,
  [3835] = 3699,
  [3836] = 3706,
  [3837] = 3702,
  [3838] = 3703,
  [3839] = 3698,
  [3840] = 3703,
  [3841] = 3698,
  [3842] = 3842,
  [3843] = 3700,
  [3844] = 3700,
  [3845] = 3698,
  [3846] = 3700,
  [3847] = 3698,
  [3848] = 3700,
  [3849] = 3699,
  [3850] = 3705,
  [3851] = 3698,
  [3852] = 3700,
  [3853] = 3702,
  [3854] = 3700,
  [3855] = 3705,
  [3856] = 3700,
  [3857] = 3699,
  [3858] = 3698,
  [3859] = 3706,
  [3860] = 3702,
  [3861] = 3700,
  [3862] = 3705,
  [3863] = 3699,
  [3864] = 3702,
  [3865] = 3702,
  [3866] = 3702,
  [3867] = 3698,
  [3868] = 3698,
  [3869] = 3702,
  [3870] = 3702,
  [3871] = 3698,
  [3872] = 3700,
  [3873] = 3702,
  [3874] = 3705,
  [3875] = 3699,
  [3876] = 3705,
  [3877] = 3702,
  [3878] = 3702,
  [3879] = 3699,
  [3880] = 3702,
  [3881] = 3702,
  [3882] = 3702,
  [3883] = 3698,
  [3884] = 3702,
  [3885] = 3702,
  [3886] = 3698,
  [3887] = 3700,
  [3888] = 3699,
  [3889] = 3700,
  [3890] = 3702,
  [3891] = 3702,
  [3892] = 3705,
  [3893] = 3705,
  [3894] = 3702,
  [3895] = 3702,
  [3896] = 3702,
  [3897] = 3698,
  [3898] = 3706,
  [3899] = 3702,
  [3900] = 3702,
  [3901] = 3700,
  [3902] = 3703,
  [3903] = 3700,
  [3904] = 3698,
  [3905] = 3702,
  [3906] = 3698,
  [3907] = 3703,
  [3908] = 3700,
  [3909] = 3698,
  [3910] = 3700,
  [3911] = 3698,
  [3912] = 3702,
  [3913] = 3700,
  [3914] = 3700,
  [3915] = 3699,
  [3916] = 3705,
  [3917] = 3698,
  [3918] = 3700,
  [3919] = 3699,
  [3920] = 3705,
  [3921] = 3702,
  [3922] = 3698,
  [3923] = 3700,
  [3924] = 3700,
  [3925] = 3698,
  [3926] = 3698,
  [3927] = 3700,
  [3928] = 3698,
  [3929] = 3702,
  [3930] = 3700,
  [3931] = 3700,
  [3932] = 3699,
  [3933] = 3705,
  [3934] = 3699,
  [3935] = 3698,
  [3936] = 3700,
  [3937] = 3700,
  [3938] = 3698,
  [3939] = 3698,
  [3940] = 3702,
  [3941] = 3702,
  [3942] = 3702,
  [3943] = 3702,
  [3944] = 3705,
  [3945] = 3699,
  [3946] = 3702,
  [3947] = 3702,
  [3948] = 3702,
  [3949] = 3698,
  [3950] = 3702,
  [3951] = 3702,
  [3952] = 3702,
  [3953] = 3702,
  [3954] = 3698,
  [3955] = 3698,
  [3956] = 3698,
  [3957] = 3700,
  [3958] = 3698,
  [3959] = 3700,
  [3960] = 3960,
  [3961] = 3700,
  [3962] = 3700,
  [3963] = 3706,
  [3964] = 3698,
  [3965] = 3703,
  [3966] = 3699,
  [3967] = 3705,
  [3968] = 3700,
  [3969] = 3698,
  [3970] = 3970,
  [3971] = 3700,
  [3972] = 3698,
  [3973] = 3705,
  [3974] = 3699,
  [3975] = 3700,
  [3976] = 3698,
  [3977] = 3705,
  [3978] = 3699,
  [3979] = 3700,
  [3980] = 3698,
  [3981] = 3981,
  [3982] = 3705,
  [3983] = 3699,
  [3984] = 3698,
  [3985] = 3705,
  [3986] = 3699,
  [3987] = 3842,
  [3988] = 3700,
  [3989] = 3700,
  [3990] = 3698,
  [3991] = 3700,
  [3992] = 3698,
  [3993] = 3700,
  [3994] = 3705,
  [3995] = 3699,
  [3996] = 3698,
  [3997] = 3705,
  [3998] = 3700,
  [3999] = 3698,
  [4000] = 3699,
  [4001] = 4001,
  [4002] = 3700,
  [4003] = 3700,
  [4004] = 3698,
  [4005] = 3705,
  [4006] = 3699,
  [4007] = 3700,
  [4008] = 3705,
  [4009] = 3698,
  [4010] = 3700,
  [4011] = 3698,
  [4012] = 3699,
  [4013] = 3698,
  [4014] = 3700,
  [4015] = 3698,
  [4016] = 3702,
  [4017] = 3700,
  [4018] = 3827,
  [4019] = 3700,
  [4020] = 3698,
  [4021] = 3705,
  [4022] = 3700,
  [4023] = 3700,
  [4024] = 3706,
  [4025] = 3699,
  [4026] = 3700,
  [4027] = 3702,
  [4028] = 3702,
  [4029] = 3698,
  [4030] = 3703,
  [4031] = 3706,
  [4032] = 4001,
  [4033] = 4033,
  [4034] = 3702,
  [4035] = 4035,
  [4036] = 3703,
  [4037] = 3705,
  [4038] = 3700,
  [4039] = 3706,
  [4040] = 4040,
  [4041] = 4041,
  [4042] = 3698,
  [4043] = 4043,
  [4044] = 4044,
  [4045] = 3703,
  [4046] = 3700,
  [4047] = 3699,
  [4048] = 3705,
  [4049] = 3699,
  [4050] = 3706,
  [4051] = 4051,
  [4052] = 3703,
  [4053] = 3706,
  [4054] = 3705,
  [4055] = 3698,
  [4056] = 4056,
  [4057] = 3970,
  [4058] = 4058,
  [4059] = 3698,
  [4060] = 3731,
  [4061] = 3699,
  [4062] = 3699,
  [4063] = 3705,
  [4064] = 4033,
  [4065] = 3702,
  [4066] = 3981,
  [4067] = 3842,
  [4068] = 4035,
  [4069] = 4069,
  [4070] = 4040,
  [4071] = 3698,
  [4072] = 4041,
  [4073] = 4043,
  [4074] = 3705,
  [4075] = 3700,
  [4076] = 3699,
  [4077] = 3700,
  [4078] = 3827,
  [4079] = 3705,
  [4080] = 3698,
  [4081] = 3700,
  [4082] = 3699,
  [4083] = 4083,
  [4084] = 4044,
  [4085] = 4085,
  [4086] = 3698,
  [4087] = 3698,
  [4088] = 3700,
  [4089] = 3700,
  [4090] = 4051,
  [4091] = 3960,
  [4092] = 3698,
  [4093] = 4056,
  [4094] = 3698,
  [4095] = 3698,
  [4096] = 4069,
  [4097] = 3700,
  [4098] = 4058,
  [4099] = 3698,
  [4100] = 3698,
  [4101] = 3705,
  [4102] = 4085,
  [4103] = 3699,
  [4104] = 3700,
  [4105] = 3700,
  [4106] = 3698,
  [4107] = 3703,
  [4108] = 4108,
  [4109] = 3706,
  [4110] = 4083,
  [4111] = 3706,
  [4112] = 3705,
  [4113] = 3698,
  [4114] = 3699,
  [4115] = 3703,
  [4116] = 3698,
  [4117] = 3700,
  [4118] = 3703,
  [4119] = 3706,
  [4120] = 4085,
  [4121] = 4083,
  [4122] = 3827,
  [4123] = 4069,
  [4124] = 3981,
  [4125] = 3703,
  [4126] = 3706,
  [4127] = 3702,
  [4128] = 3698,
  [4129] = 3698,
  [4130] = 3731,
  [4131] = 4058,
  [4132] = 3970,
  [4133] = 3700,
  [4134] = 4056,
  [4135] = 3705,
  [4136] = 3699,
  [4137] = 3800,
  [4138] = 3703,
  [4139] = 3698,
  [4140] = 3698,
  [4141] = 3700,
  [4142] = 3960,
  [4143] = 4051,
  [4144] = 3705,
  [4145] = 4044,
  [4146] = 4043,
  [4147] = 4041,
  [4148] = 4040,
  [4149] = 4035,
  [4150] = 4033,
  [4151] = 4001,
  [4152] = 3706,
  [4153] = 3700,
  [4154] = 3699,
  [4155] = 3705,
  [4156] = 3700,
  [4157] = 3698,
  [4158] = 3698,
  [4159] = 3700,
  [4160] = 3700,
  [4161] = 3700,
  [4162] = 3700,
  [4163] = 3698,
  [4164] = 3699,
  [4165] = 3703,
  [4166] = 3705,
  [4167] = 3706,
  [4168] = 3702,
  [4169] = 3699,
  [4170] = 3703,
  [4171] = 3700,
  [4172] = 3698,
  [4173] = 3700,
  [4174] = 3706,
  [4175] = 3699,
  [4176] = 3705,
  [4177] = 3700,
  [4178] = 3700,
  [4179] = 3698,
  [4180] = 3700,
  [4181] = 3698,
  [4182] = 3703,
  [4183] = 3705,
  [4184] = 3699,
  [4185] = 3800,
  [4186] = 3706,
  [4187] = 3698,
  [4188] = 3703,
  [4189] = 3706,
  [4190] = 3699,
  [4191] = 3700,
  [4192] = 3705,
  [4193] = 3699,
  [4194] = 3705,
  [4195] = 3705,
  [4196] = 3699,
  [4197] = 3700,
  [4198] = 3698,
  [4199] = 3698,
  [4200] = 3699,
  [4201] = 3705,
  [4202] = 3699,
  [4203] = 3700,
  [4204] = 3705,
  [4205] = 3699,
  [4206] = 3705,
  [4207] = 3699,
  [4208] = 3705,
  [4209] = 3705,
  [4210] = 3699,
  [4211] = 3699,
  [4212] = 3705,
  [4213] = 3700,
  [4214] = 3698,
  [4215] = 3705,
  [4216] = 464,
  [4217] = 4217,
  [4218] = 469,
  [4219] = 4217,
  [4220] = 4217,
  [4221] = 4221,
  [4222] = 469,
  [4223] = 4223,
  [4224] = 4224,
  [4225] = 4225,
  [4226] = 4217,
  [4227] = 4227,
  [4228] = 465,
  [4229] = 1011,
  [4230] = 4230,
  [4231] = 4231,
  [4232] = 468,
  [4233] = 469,
  [4234] = 4227,
  [4235] = 4235,
  [4236] = 4236,
  [4237] = 4237,
  [4238] = 1148,
  [4239] = 4225,
  [4240] = 1100,
  [4241] = 4241,
  [4242] = 832,
  [4243] = 4243,
  [4244] = 468,
  [4245] = 473,
  [4246] = 859,
  [4247] = 1081,
  [4248] = 470,
  [4249] = 4223,
  [4250] = 4250,
  [4251] = 470,
  [4252] = 4252,
  [4253] = 468,
  [4254] = 4224,
  [4255] = 469,
  [4256] = 468,
  [4257] = 885,
  [4258] = 473,
  [4259] = 469,
  [4260] = 880,
  [4261] = 494,
  [4262] = 1429,
  [4263] = 489,
  [4264] = 491,
  [4265] = 475,
  [4266] = 468,
  [4267] = 476,
  [4268] = 468,
  [4269] = 4227,
  [4270] = 880,
  [4271] = 1100,
  [4272] = 885,
  [4273] = 4273,
  [4274] = 832,
  [4275] = 478,
  [4276] = 859,
  [4277] = 1429,
  [4278] = 4278,
  [4279] = 4235,
  [4280] = 474,
  [4281] = 4224,
  [4282] = 4230,
  [4283] = 4283,
  [4284] = 473,
  [4285] = 482,
  [4286] = 468,
  [4287] = 1361,
  [4288] = 474,
  [4289] = 4224,
  [4290] = 1146,
  [4291] = 2704,
  [4292] = 832,
  [4293] = 2702,
  [4294] = 1149,
  [4295] = 485,
  [4296] = 859,
  [4297] = 473,
  [4298] = 481,
  [4299] = 4252,
  [4300] = 486,
  [4301] = 487,
  [4302] = 479,
  [4303] = 483,
  [4304] = 470,
  [4305] = 1382,
  [4306] = 490,
  [4307] = 488,
  [4308] = 492,
  [4309] = 470,
  [4310] = 1100,
  [4311] = 479,
  [4312] = 4223,
  [4313] = 485,
  [4314] = 4225,
  [4315] = 469,
  [4316] = 481,
  [4317] = 474,
  [4318] = 486,
  [4319] = 4231,
  [4320] = 484,
  [4321] = 4225,
  [4322] = 489,
  [4323] = 481,
  [4324] = 494,
  [4325] = 486,
  [4326] = 487,
  [4327] = 479,
  [4328] = 478,
  [4329] = 477,
  [4330] = 491,
  [4331] = 1081,
  [4332] = 490,
  [4333] = 487,
  [4334] = 479,
  [4335] = 475,
  [4336] = 476,
  [4337] = 482,
  [4338] = 483,
  [4339] = 4339,
  [4340] = 485,
  [4341] = 4341,
  [4342] = 4342,
  [4343] = 474,
  [4344] = 2694,
  [4345] = 469,
  [4346] = 4346,
  [4347] = 1146,
  [4348] = 486,
  [4349] = 4349,
  [4350] = 469,
  [4351] = 1382,
  [4352] = 885,
  [4353] = 4227,
  [4354] = 478,
  [4355] = 477,
  [4356] = 491,
  [4357] = 482,
  [4358] = 4237,
  [4359] = 469,
  [4360] = 492,
  [4361] = 469,
  [4362] = 469,
  [4363] = 2696,
  [4364] = 4231,
  [4365] = 478,
  [4366] = 488,
  [4367] = 492,
  [4368] = 494,
  [4369] = 475,
  [4370] = 1081,
  [4371] = 4250,
  [4372] = 490,
  [4373] = 488,
  [4374] = 491,
  [4375] = 880,
  [4376] = 4224,
  [4377] = 480,
  [4378] = 490,
  [4379] = 487,
  [4380] = 470,
  [4381] = 4227,
  [4382] = 1149,
  [4383] = 1361,
  [4384] = 488,
  [4385] = 492,
  [4386] = 484,
  [4387] = 4387,
  [4388] = 484,
  [4389] = 489,
  [4390] = 1011,
  [4391] = 481,
  [4392] = 485,
  [4393] = 489,
  [4394] = 494,
  [4395] = 480,
  [4396] = 4223,
  [4397] = 4223,
  [4398] = 483,
  [4399] = 480,
  [4400] = 4225,
  [4401] = 469,
  [4402] = 477,
  [4403] = 476,
  [4404] = 482,
  [4405] = 484,
  [4406] = 4252,
  [4407] = 475,
  [4408] = 476,
  [4409] = 483,
  [4410] = 1148,
  [4411] = 477,
  [4412] = 480,
  [4413] = 468,
  [4414] = 473,
  [4415] = 4415,
  [4416] = 4349,
  [4417] = 476,
  [4418] = 859,
  [4419] = 483,
  [4420] = 4387,
  [4421] = 1146,
  [4422] = 470,
  [4423] = 495,
  [4424] = 481,
  [4425] = 486,
  [4426] = 487,
  [4427] = 479,
  [4428] = 4339,
  [4429] = 483,
  [4430] = 880,
  [4431] = 1361,
  [4432] = 490,
  [4433] = 4341,
  [4434] = 481,
  [4435] = 486,
  [4436] = 487,
  [4437] = 479,
  [4438] = 475,
  [4439] = 4230,
  [4440] = 832,
  [4441] = 490,
  [4442] = 475,
  [4443] = 478,
  [4444] = 477,
  [4445] = 491,
  [4446] = 489,
  [4447] = 478,
  [4448] = 477,
  [4449] = 491,
  [4450] = 885,
  [4451] = 480,
  [4452] = 880,
  [4453] = 468,
  [4454] = 476,
  [4455] = 4283,
  [4456] = 479,
  [4457] = 489,
  [4458] = 1163,
  [4459] = 885,
  [4460] = 1158,
  [4461] = 485,
  [4462] = 1146,
  [4463] = 832,
  [4464] = 485,
  [4465] = 4278,
  [4466] = 859,
  [4467] = 859,
  [4468] = 473,
  [4469] = 488,
  [4470] = 492,
  [4471] = 470,
  [4472] = 1081,
  [4473] = 485,
  [4474] = 474,
  [4475] = 481,
  [4476] = 4231,
  [4477] = 832,
  [4478] = 475,
  [4479] = 1100,
  [4480] = 4250,
  [4481] = 4231,
  [4482] = 474,
  [4483] = 494,
  [4484] = 4252,
  [4485] = 494,
  [4486] = 1100,
  [4487] = 469,
  [4488] = 4252,
  [4489] = 4489,
  [4490] = 476,
  [4491] = 469,
  [4492] = 1149,
  [4493] = 482,
  [4494] = 483,
  [4495] = 469,
  [4496] = 481,
  [4497] = 486,
  [4498] = 494,
  [4499] = 485,
  [4500] = 1149,
  [4501] = 487,
  [4502] = 479,
  [4503] = 495,
  [4504] = 1081,
  [4505] = 473,
  [4506] = 4231,
  [4507] = 1429,
  [4508] = 474,
  [4509] = 2926,
  [4510] = 473,
  [4511] = 470,
  [4512] = 489,
  [4513] = 880,
  [4514] = 1011,
  [4515] = 486,
  [4516] = 490,
  [4517] = 4235,
  [4518] = 1081,
  [4519] = 4230,
  [4520] = 468,
  [4521] = 1429,
  [4522] = 1100,
  [4523] = 470,
  [4524] = 469,
  [4525] = 1148,
  [4526] = 482,
  [4527] = 1146,
  [4528] = 4342,
  [4529] = 4252,
  [4530] = 4235,
  [4531] = 1148,
  [4532] = 482,
  [4533] = 494,
  [4534] = 4237,
  [4535] = 1011,
  [4536] = 487,
  [4537] = 488,
  [4538] = 1149,
  [4539] = 483,
  [4540] = 492,
  [4541] = 2931,
  [4542] = 2930,
  [4543] = 478,
  [4544] = 477,
  [4545] = 491,
  [4546] = 1382,
  [4547] = 4235,
  [4548] = 476,
  [4549] = 4346,
  [4550] = 1148,
  [4551] = 4489,
  [4552] = 478,
  [4553] = 477,
  [4554] = 1382,
  [4555] = 1100,
  [4556] = 488,
  [4557] = 492,
  [4558] = 473,
  [4559] = 1011,
  [4560] = 491,
  [4561] = 484,
  [4562] = 4252,
  [4563] = 495,
  [4564] = 490,
  [4565] = 4227,
  [4566] = 482,
  [4567] = 4489,
  [4568] = 4237,
  [4569] = 475,
  [4570] = 4224,
  [4571] = 4250,
  [4572] = 484,
  [4573] = 4250,
  [4574] = 4237,
  [4575] = 480,
  [4576] = 484,
  [4577] = 495,
  [4578] = 469,
  [4579] = 832,
  [4580] = 4273,
  [4581] = 4230,
  [4582] = 1081,
  [4583] = 470,
  [4584] = 473,
  [4585] = 473,
  [4586] = 470,
  [4587] = 480,
  [4588] = 488,
  [4589] = 859,
  [4590] = 492,
  [4591] = 473,
  [4592] = 470,
  [4593] = 469,
  [4594] = 470,
  [4595] = 484,
  [4596] = 1382,
  [4597] = 1429,
  [4598] = 474,
  [4599] = 4231,
  [4600] = 480,
  [4601] = 885,
  [4602] = 885,
  [4603] = 473,
  [4604] = 880,
  [4605] = 2929,
  [4606] = 489,
  [4607] = 4607,
  [4608] = 4339,
  [4609] = 4609,
  [4610] = 4610,
  [4611] = 473,
  [4612] = 4607,
  [4613] = 4610,
  [4614] = 1081,
  [4615] = 4615,
  [4616] = 4616,
  [4617] = 1301,
  [4618] = 475,
  [4619] = 859,
  [4620] = 4620,
  [4621] = 4252,
  [4622] = 832,
  [4623] = 4623,
  [4624] = 4624,
  [4625] = 4273,
  [4626] = 4231,
  [4627] = 1100,
  [4628] = 491,
  [4629] = 4624,
  [4630] = 470,
  [4631] = 477,
  [4632] = 4632,
  [4633] = 4252,
  [4634] = 470,
  [4635] = 473,
  [4636] = 1556,
  [4637] = 488,
  [4638] = 492,
  [4639] = 1498,
  [4640] = 1081,
  [4641] = 4624,
  [4642] = 4642,
  [4643] = 4609,
  [4644] = 4250,
  [4645] = 1146,
  [4646] = 4646,
  [4647] = 4642,
  [4648] = 4648,
  [4649] = 4624,
  [4650] = 4283,
  [4651] = 4651,
  [4652] = 4624,
  [4653] = 478,
  [4654] = 4624,
  [4655] = 4655,
  [4656] = 476,
  [4657] = 4624,
  [4658] = 4624,
  [4659] = 4346,
  [4660] = 4624,
  [4661] = 4624,
  [4662] = 4349,
  [4663] = 4341,
  [4664] = 4664,
  [4665] = 4235,
  [4666] = 4632,
  [4667] = 4387,
  [4668] = 4624,
  [4669] = 473,
  [4670] = 476,
  [4671] = 4624,
  [4672] = 4624,
  [4673] = 880,
  [4674] = 470,
  [4675] = 4624,
  [4676] = 4624,
  [4677] = 4624,
  [4678] = 4624,
  [4679] = 4624,
  [4680] = 4624,
  [4681] = 4624,
  [4682] = 4624,
  [4683] = 859,
  [4684] = 880,
  [4685] = 4273,
  [4686] = 4624,
  [4687] = 832,
  [4688] = 1542,
  [4689] = 4231,
  [4690] = 4624,
  [4691] = 4624,
  [4692] = 1081,
  [4693] = 4624,
  [4694] = 4624,
  [4695] = 1429,
  [4696] = 4624,
  [4697] = 4283,
  [4698] = 885,
  [4699] = 880,
  [4700] = 4700,
  [4701] = 484,
  [4702] = 4648,
  [4703] = 4624,
  [4704] = 1149,
  [4705] = 1100,
  [4706] = 480,
  [4707] = 4624,
  [4708] = 475,
  [4709] = 1301,
  [4710] = 4278,
  [4711] = 885,
  [4712] = 4712,
  [4713] = 4624,
  [4714] = 4624,
  [4715] = 4624,
  [4716] = 4339,
  [4717] = 4342,
  [4718] = 4624,
  [4719] = 4387,
  [4720] = 4624,
  [4721] = 4721,
  [4722] = 4342,
  [4723] = 4387,
  [4724] = 4339,
  [4725] = 4278,
  [4726] = 495,
  [4727] = 4727,
  [4728] = 4624,
  [4729] = 4624,
  [4730] = 4624,
  [4731] = 4624,
  [4732] = 4624,
  [4733] = 4252,
  [4734] = 4278,
  [4735] = 479,
  [4736] = 1607,
  [4737] = 4624,
  [4738] = 4624,
  [4739] = 490,
  [4740] = 4624,
  [4741] = 4741,
  [4742] = 4742,
  [4743] = 4743,
  [4744] = 4624,
  [4745] = 4342,
  [4746] = 4283,
  [4747] = 4273,
  [4748] = 859,
  [4749] = 4624,
  [4750] = 880,
  [4751] = 4624,
  [4752] = 4624,
  [4753] = 481,
  [4754] = 885,
  [4755] = 4755,
  [4756] = 4624,
  [4757] = 4757,
  [4758] = 4346,
  [4759] = 1498,
  [4760] = 832,
  [4761] = 4624,
  [4762] = 486,
  [4763] = 4624,
  [4764] = 4624,
  [4765] = 4624,
  [4766] = 486,
  [4767] = 1382,
  [4768] = 4624,
  [4769] = 4755,
  [4770] = 487,
  [4771] = 885,
  [4772] = 495,
  [4773] = 4624,
  [4774] = 469,
  [4775] = 1011,
  [4776] = 470,
  [4777] = 473,
  [4778] = 4341,
  [4779] = 4624,
  [4780] = 479,
  [4781] = 470,
  [4782] = 1534,
  [4783] = 473,
  [4784] = 4230,
  [4785] = 4785,
  [4786] = 4727,
  [4787] = 4237,
  [4788] = 495,
  [4789] = 1607,
  [4790] = 485,
  [4791] = 4624,
  [4792] = 4792,
  [4793] = 4341,
  [4794] = 4616,
  [4795] = 4231,
  [4796] = 474,
  [4797] = 4349,
  [4798] = 4624,
  [4799] = 4624,
  [4800] = 4624,
  [4801] = 474,
  [4802] = 1148,
  [4803] = 4803,
  [4804] = 1542,
  [4805] = 483,
  [4806] = 832,
  [4807] = 488,
  [4808] = 489,
  [4809] = 4624,
  [4810] = 482,
  [4811] = 483,
  [4812] = 492,
  [4813] = 4624,
  [4814] = 4624,
  [4815] = 4624,
  [4816] = 4252,
  [4817] = 4624,
  [4818] = 4623,
  [4819] = 4624,
  [4820] = 1534,
  [4821] = 4655,
  [4822] = 859,
  [4823] = 4624,
  [4824] = 4824,
  [4825] = 490,
  [4826] = 4624,
  [4827] = 1281,
  [4828] = 1100,
  [4829] = 4346,
  [4830] = 4624,
  [4831] = 482,
  [4832] = 4624,
  [4833] = 485,
  [4834] = 1100,
  [4835] = 4231,
  [4836] = 4624,
  [4837] = 480,
  [4838] = 1081,
  [4839] = 4624,
  [4840] = 1556,
  [4841] = 4624,
  [4842] = 487,
  [4843] = 4624,
  [4844] = 4615,
  [4845] = 4624,
  [4846] = 495,
  [4847] = 481,
  [4848] = 4624,
  [4849] = 491,
  [4850] = 4624,
  [4851] = 4624,
  [4852] = 477,
  [4853] = 478,
  [4854] = 4624,
  [4855] = 494,
  [4856] = 4712,
  [4857] = 4624,
  [4858] = 4785,
  [4859] = 489,
  [4860] = 4349,
  [4861] = 1281,
  [4862] = 484,
  [4863] = 4624,
  [4864] = 4712,
  [4865] = 4648,
  [4866] = 4866,
  [4867] = 4867,
  [4868] = 4868,
  [4869] = 1301,
  [4870] = 1542,
  [4871] = 4871,
  [4872] = 4872,
  [4873] = 4873,
  [4874] = 4874,
  [4875] = 4712,
  [4876] = 4876,
  [4877] = 4877,
  [4878] = 4878,
  [4879] = 1534,
  [4880] = 4880,
  [4881] = 4881,
  [4882] = 4882,
  [4883] = 4883,
  [4884] = 4884,
  [4885] = 1301,
  [4886] = 4886,
  [4887] = 4887,
  [4888] = 1301,
  [4889] = 4877,
  [4890] = 4890,
  [4891] = 4891,
  [4892] = 4892,
  [4893] = 1281,
  [4894] = 4873,
  [4895] = 4895,
  [4896] = 880,
  [4897] = 885,
  [4898] = 4898,
  [4899] = 4871,
  [4900] = 4900,
  [4901] = 1081,
  [4902] = 4655,
  [4903] = 4884,
  [4904] = 4824,
  [4905] = 4615,
  [4906] = 4231,
  [4907] = 1607,
  [4908] = 4908,
  [4909] = 4883,
  [4910] = 4620,
  [4911] = 4911,
  [4912] = 4866,
  [4913] = 495,
  [4914] = 1607,
  [4915] = 4915,
  [4916] = 4916,
  [4917] = 4755,
  [4918] = 4918,
  [4919] = 859,
  [4920] = 4920,
  [4921] = 1100,
  [4922] = 4922,
  [4923] = 4900,
  [4924] = 4924,
  [4925] = 1498,
  [4926] = 4920,
  [4927] = 4927,
  [4928] = 4928,
  [4929] = 4929,
  [4930] = 1281,
  [4931] = 4931,
  [4932] = 4915,
  [4933] = 4892,
  [4934] = 4924,
  [4935] = 4929,
  [4936] = 4936,
  [4937] = 4937,
  [4938] = 4938,
  [4939] = 4792,
  [4940] = 4940,
  [4941] = 4252,
  [4942] = 4616,
  [4943] = 4943,
  [4944] = 4944,
  [4945] = 4945,
  [4946] = 4632,
  [4947] = 4947,
  [4948] = 4936,
  [4949] = 4949,
  [4950] = 4610,
  [4951] = 4951,
  [4952] = 4952,
  [4953] = 4615,
  [4954] = 4607,
  [4955] = 4607,
  [4956] = 4886,
  [4957] = 4887,
  [4958] = 4609,
  [4959] = 4609,
  [4960] = 4960,
  [4961] = 4922,
  [4962] = 4727,
  [4963] = 4895,
  [4964] = 1534,
  [4965] = 4642,
  [4966] = 4898,
  [4967] = 4891,
  [4968] = 4890,
  [4969] = 4960,
  [4970] = 4881,
  [4971] = 4664,
  [4972] = 4918,
  [4973] = 4872,
  [4974] = 4924,
  [4975] = 1474,
  [4976] = 4615,
  [4977] = 4952,
  [4978] = 4642,
  [4979] = 4908,
  [4980] = 4876,
  [4981] = 4874,
  [4982] = 4785,
  [4983] = 4867,
  [4984] = 4648,
  [4985] = 4868,
  [4986] = 4283,
  [4987] = 4878,
  [4988] = 4880,
  [4989] = 4882,
  [4990] = 1556,
  [4991] = 4911,
  [4992] = 832,
  [4993] = 4951,
  [4994] = 4727,
  [4995] = 4387,
  [4996] = 4342,
  [4997] = 4339,
  [4998] = 4278,
  [4999] = 4918,
  [5000] = 4273,
  [5001] = 4346,
  [5002] = 4944,
  [5003] = 4931,
  [5004] = 1607,
  [5005] = 4916,
  [5006] = 4938,
  [5007] = 4951,
  [5008] = 4712,
  [5009] = 4927,
  [5010] = 4785,
  [5011] = 4947,
  [5012] = 4727,
  [5013] = 4940,
  [5014] = 4943,
  [5015] = 4891,
  [5016] = 4785,
  [5017] = 4949,
  [5018] = 4607,
  [5019] = 4609,
  [5020] = 4642,
  [5021] = 494,
  [5022] = 4937,
  [5023] = 1534,
  [5024] = 4623,
  [5025] = 4648,
  [5026] = 4945,
  [5027] = 5027,
  [5028] = 5028,
  [5029] = 1281,
  [5030] = 5028,
  [5031] = 470,
  [5032] = 5027,
  [5033] = 473,
  [5034] = 469,
  [5035] = 4882,
  [5036] = 4824,
  [5037] = 4884,
  [5038] = 468,
  [5039] = 4883,
  [5040] = 4792,
  [5041] = 1163,
  [5042] = 4615,
  [5043] = 4877,
  [5044] = 495,
  [5045] = 1158,
  [5046] = 4945,
  [5047] = 4880,
  [5048] = 5048,
  [5049] = 4712,
  [5050] = 4727,
  [5051] = 4792,
  [5052] = 4947,
  [5053] = 4642,
  [5054] = 4620,
  [5055] = 473,
  [5056] = 4224,
  [5057] = 470,
  [5058] = 4227,
  [5059] = 473,
  [5060] = 4878,
  [5061] = 4916,
  [5062] = 5062,
  [5063] = 4927,
  [5064] = 469,
  [5065] = 470,
  [5066] = 4915,
  [5067] = 4867,
  [5068] = 4908,
  [5069] = 4866,
  [5070] = 4620,
  [5071] = 4943,
  [5072] = 4872,
  [5073] = 4873,
  [5074] = 4785,
  [5075] = 4871,
  [5076] = 4876,
  [5077] = 4881,
  [5078] = 4874,
  [5079] = 4620,
  [5080] = 4890,
  [5081] = 4664,
  [5082] = 4664,
  [5083] = 4922,
  [5084] = 4868,
  [5085] = 4931,
  [5086] = 1301,
  [5087] = 4900,
  [5088] = 4938,
  [5089] = 4940,
  [5090] = 4895,
  [5091] = 4607,
  [5092] = 4609,
  [5093] = 4960,
  [5094] = 4648,
  [5095] = 4824,
  [5096] = 1534,
  [5097] = 469,
  [5098] = 4920,
  [5099] = 4929,
  [5100] = 4936,
  [5101] = 4892,
  [5102] = 4824,
  [5103] = 4937,
  [5104] = 4898,
  [5105] = 4944,
  [5106] = 1281,
  [5107] = 5027,
  [5108] = 4792,
  [5109] = 469,
  [5110] = 469,
  [5111] = 4952,
  [5112] = 4887,
  [5113] = 4949,
  [5114] = 1607,
  [5115] = 4664,
  [5116] = 4911,
  [5117] = 5117,
  [5118] = 5118,
  [5119] = 5028,
  [5120] = 4886,
  [5121] = 5121,
  [5122] = 5122,
  [5123] = 474,
  [5124] = 859,
  [5125] = 5125,
  [5126] = 5126,
  [5127] = 885,
  [5128] = 880,
  [5129] = 885,
  [5130] = 832,
  [5131] = 859,
  [5132] = 5132,
  [5133] = 1081,
  [5134] = 4231,
  [5135] = 1100,
  [5136] = 4252,
  [5137] = 1148,
  [5138] = 1011,
  [5139] = 5139,
  [5140] = 880,
  [5141] = 5141,
  [5142] = 5142,
  [5143] = 5143,
  [5144] = 5144,
  [5145] = 5145,
  [5146] = 5146,
  [5147] = 4230,
  [5148] = 4237,
  [5149] = 480,
  [5150] = 481,
  [5151] = 484,
  [5152] = 487,
  [5153] = 490,
  [5154] = 492,
  [5155] = 485,
  [5156] = 486,
  [5157] = 479,
  [5158] = 488,
  [5159] = 478,
  [5160] = 477,
  [5161] = 491,
  [5162] = 5162,
  [5163] = 5122,
  [5164] = 5164,
  [5165] = 476,
  [5166] = 5166,
  [5167] = 4250,
  [5168] = 5142,
  [5169] = 5169,
  [5170] = 4252,
  [5171] = 475,
  [5172] = 5172,
  [5173] = 832,
  [5174] = 469,
  [5175] = 470,
  [5176] = 4250,
  [5177] = 5177,
  [5178] = 1081,
  [5179] = 4231,
  [5180] = 473,
  [5181] = 4235,
  [5182] = 1100,
  [5183] = 5183,
  [5184] = 4224,
  [5185] = 5185,
  [5186] = 5186,
  [5187] = 5187,
  [5188] = 468,
  [5189] = 5122,
  [5190] = 5190,
  [5191] = 5191,
  [5192] = 5192,
  [5193] = 5193,
  [5194] = 5194,
  [5195] = 4227,
  [5196] = 5196,
  [5197] = 5197,
  [5198] = 5198,
  [5199] = 482,
  [5200] = 483,
  [5201] = 5201,
  [5202] = 5202,
  [5203] = 5203,
  [5204] = 489,
  [5205] = 4235,
  [5206] = 469,
  [5207] = 470,
  [5208] = 473,
  [5209] = 470,
  [5210] = 473,
  [5211] = 5211,
  [5212] = 5212,
  [5213] = 4227,
  [5214] = 5186,
  [5215] = 5215,
  [5216] = 4664,
  [5217] = 5185,
  [5218] = 5187,
  [5219] = 5190,
  [5220] = 5191,
  [5221] = 469,
  [5222] = 5192,
  [5223] = 5194,
  [5224] = 4824,
  [5225] = 5196,
  [5226] = 5142,
  [5227] = 5227,
  [5228] = 5197,
  [5229] = 5198,
  [5230] = 5202,
  [5231] = 5203,
  [5232] = 5232,
  [5233] = 5187,
  [5234] = 5203,
  [5235] = 5202,
  [5236] = 5198,
  [5237] = 5197,
  [5238] = 5196,
  [5239] = 5194,
  [5240] = 469,
  [5241] = 5192,
  [5242] = 5191,
  [5243] = 5190,
  [5244] = 5186,
  [5245] = 5245,
  [5246] = 5185,
  [5247] = 5247,
  [5248] = 5248,
  [5249] = 5249,
  [5250] = 5250,
  [5251] = 5251,
  [5252] = 5252,
  [5253] = 1148,
  [5254] = 5254,
  [5255] = 486,
  [5256] = 1011,
  [5257] = 5257,
  [5258] = 1149,
  [5259] = 5259,
  [5260] = 5260,
  [5261] = 5261,
  [5262] = 478,
  [5263] = 491,
  [5264] = 5264,
  [5265] = 5265,
  [5266] = 5266,
  [5267] = 469,
  [5268] = 5268,
  [5269] = 488,
  [5270] = 470,
  [5271] = 473,
  [5272] = 5272,
  [5273] = 5273,
  [5274] = 494,
  [5275] = 5275,
  [5276] = 468,
  [5277] = 5277,
  [5278] = 880,
  [5279] = 492,
  [5280] = 5280,
  [5281] = 5281,
  [5282] = 5282,
  [5283] = 5283,
  [5284] = 477,
  [5285] = 5285,
  [5286] = 880,
  [5287] = 885,
  [5288] = 490,
  [5289] = 5266,
  [5290] = 885,
  [5291] = 5291,
  [5292] = 1146,
  [5293] = 1081,
  [5294] = 487,
  [5295] = 5295,
  [5296] = 5296,
  [5297] = 470,
  [5298] = 481,
  [5299] = 5299,
  [5300] = 832,
  [5301] = 885,
  [5302] = 5302,
  [5303] = 880,
  [5304] = 5304,
  [5305] = 4231,
  [5306] = 832,
  [5307] = 5307,
  [5308] = 5308,
  [5309] = 5309,
  [5310] = 5310,
  [5311] = 5311,
  [5312] = 859,
  [5313] = 5313,
  [5314] = 5314,
  [5315] = 5266,
  [5316] = 5316,
  [5317] = 5266,
  [5318] = 4252,
  [5319] = 5319,
  [5320] = 5320,
  [5321] = 5321,
  [5322] = 470,
  [5323] = 859,
  [5324] = 5324,
  [5325] = 5325,
  [5326] = 5326,
  [5327] = 480,
  [5328] = 473,
  [5329] = 5329,
  [5330] = 5330,
  [5331] = 1081,
  [5332] = 5332,
  [5333] = 5333,
  [5334] = 5334,
  [5335] = 5335,
  [5336] = 5336,
  [5337] = 479,
  [5338] = 5338,
  [5339] = 5339,
  [5340] = 5340,
  [5341] = 5341,
  [5342] = 5342,
  [5343] = 5343,
  [5344] = 5344,
  [5345] = 5345,
  [5346] = 5346,
  [5347] = 832,
  [5348] = 859,
  [5349] = 468,
  [5350] = 4283,
  [5351] = 5351,
  [5352] = 1081,
  [5353] = 5353,
  [5354] = 5354,
  [5355] = 5355,
  [5356] = 5266,
  [5357] = 1382,
  [5358] = 5266,
  [5359] = 4231,
  [5360] = 4237,
  [5361] = 5266,
  [5362] = 4250,
  [5363] = 4235,
  [5364] = 489,
  [5365] = 1474,
  [5366] = 4283,
  [5367] = 4387,
  [5368] = 4342,
  [5369] = 4339,
  [5370] = 4278,
  [5371] = 4252,
  [5372] = 4273,
  [5373] = 4252,
  [5374] = 468,
  [5375] = 4346,
  [5376] = 1429,
  [5377] = 4231,
  [5378] = 1100,
  [5379] = 1100,
  [5380] = 483,
  [5381] = 5381,
  [5382] = 4278,
  [5383] = 5383,
  [5384] = 5384,
  [5385] = 4230,
  [5386] = 5386,
  [5387] = 482,
  [5388] = 494,
  [5389] = 475,
  [5390] = 473,
  [5391] = 5391,
  [5392] = 473,
  [5393] = 470,
  [5394] = 5266,
  [5395] = 5395,
  [5396] = 5396,
  [5397] = 5397,
  [5398] = 1100,
  [5399] = 484,
  [5400] = 474,
  [5401] = 5401,
  [5402] = 485,
  [5403] = 476,
  [5404] = 485,
  [5405] = 4346,
  [5406] = 5406,
  [5407] = 473,
  [5408] = 1361,
  [5409] = 489,
  [5410] = 495,
  [5411] = 468,
  [5412] = 832,
  [5413] = 483,
  [5414] = 474,
  [5415] = 880,
  [5416] = 495,
  [5417] = 4231,
  [5418] = 1081,
  [5419] = 4339,
  [5420] = 859,
  [5421] = 4342,
  [5422] = 480,
  [5423] = 4387,
  [5424] = 1429,
  [5425] = 5425,
  [5426] = 494,
  [5427] = 1100,
  [5428] = 482,
  [5429] = 5429,
  [5430] = 487,
  [5431] = 475,
  [5432] = 491,
  [5433] = 477,
  [5434] = 478,
  [5435] = 484,
  [5436] = 4283,
  [5437] = 1429,
  [5438] = 483,
  [5439] = 490,
  [5440] = 482,
  [5441] = 474,
  [5442] = 492,
  [5443] = 488,
  [5444] = 489,
  [5445] = 479,
  [5446] = 476,
  [5447] = 1146,
  [5448] = 485,
  [5449] = 486,
  [5450] = 476,
  [5451] = 475,
  [5452] = 481,
  [5453] = 475,
  [5454] = 491,
  [5455] = 477,
  [5456] = 5456,
  [5457] = 476,
  [5458] = 5458,
  [5459] = 478,
  [5460] = 5460,
  [5461] = 482,
  [5462] = 1149,
  [5463] = 480,
  [5464] = 490,
  [5465] = 4278,
  [5466] = 1146,
  [5467] = 474,
  [5468] = 481,
  [5469] = 494,
  [5470] = 484,
  [5471] = 470,
  [5472] = 481,
  [5473] = 479,
  [5474] = 885,
  [5475] = 4273,
  [5476] = 4252,
  [5477] = 488,
  [5478] = 486,
  [5479] = 494,
  [5480] = 483,
  [5481] = 485,
  [5482] = 487,
  [5483] = 492,
  [5484] = 478,
  [5485] = 488,
  [5486] = 492,
  [5487] = 489,
  [5488] = 486,
  [5489] = 484,
  [5490] = 487,
  [5491] = 479,
  [5492] = 480,
  [5493] = 491,
  [5494] = 1382,
  [5495] = 477,
  [5496] = 1382,
  [5497] = 490,
  [5498] = 1149,
  [5499] = 4824,
  [5500] = 1146,
  [5501] = 492,
  [5502] = 488,
  [5503] = 465,
  [5504] = 5504,
  [5505] = 5505,
  [5506] = 495,
  [5507] = 481,
  [5508] = 474,
  [5509] = 480,
  [5510] = 487,
  [5511] = 479,
  [5512] = 495,
  [5513] = 490,
  [5514] = 464,
  [5515] = 5504,
  [5516] = 5505,
  [5517] = 5504,
  [5518] = 469,
  [5519] = 1429,
  [5520] = 494,
  [5521] = 484,
  [5522] = 478,
  [5523] = 486,
  [5524] = 477,
  [5525] = 485,
  [5526] = 1301,
  [5527] = 491,
  [5528] = 4664,
  [5529] = 495,
  [5530] = 1534,
  [5531] = 5505,
  [5532] = 1382,
  [5533] = 5533,
  [5534] = 476,
  [5535] = 5535,
  [5536] = 475,
  [5537] = 489,
  [5538] = 5504,
  [5539] = 1607,
  [5540] = 5505,
  [5541] = 482,
  [5542] = 483,
  [5543] = 1281,
  [5544] = 1149,
  [5545] = 5505,
  [5546] = 5504,
  [5547] = 4615,
  [5548] = 859,
  [5549] = 1301,
  [5550] = 5550,
  [5551] = 5550,
  [5552] = 1542,
  [5553] = 1081,
  [5554] = 5550,
  [5555] = 1281,
  [5556] = 4866,
  [5557] = 5550,
  [5558] = 5550,
  [5559] = 5550,
  [5560] = 4632,
  [5561] = 4616,
  [5562] = 4252,
  [5563] = 1607,
  [5564] = 4623,
  [5565] = 4900,
  [5566] = 1498,
  [5567] = 4610,
  [5568] = 4892,
  [5569] = 5550,
  [5570] = 4755,
  [5571] = 4882,
  [5572] = 4880,
  [5573] = 4878,
  [5574] = 4915,
  [5575] = 1534,
  [5576] = 4908,
  [5577] = 470,
  [5578] = 4872,
  [5579] = 4881,
  [5580] = 473,
  [5581] = 4890,
  [5582] = 4655,
  [5583] = 468,
  [5584] = 4231,
  [5585] = 4922,
  [5586] = 4936,
  [5587] = 1281,
  [5588] = 4607,
  [5589] = 4920,
  [5590] = 4929,
  [5591] = 4609,
  [5592] = 4642,
  [5593] = 5550,
  [5594] = 4648,
  [5595] = 4937,
  [5596] = 4944,
  [5597] = 4947,
  [5598] = 4886,
  [5599] = 4887,
  [5600] = 4867,
  [5601] = 1607,
  [5602] = 4824,
  [5603] = 4727,
  [5604] = 1534,
  [5605] = 880,
  [5606] = 885,
  [5607] = 495,
  [5608] = 1100,
  [5609] = 4235,
  [5610] = 1556,
  [5611] = 4664,
  [5612] = 1301,
  [5613] = 4785,
  [5614] = 5550,
  [5615] = 4712,
  [5616] = 832,
  [5617] = 4920,
  [5618] = 481,
  [5619] = 1534,
  [5620] = 4727,
  [5621] = 4785,
  [5622] = 475,
  [5623] = 4943,
  [5624] = 4945,
  [5625] = 480,
  [5626] = 4887,
  [5627] = 4886,
  [5628] = 484,
  [5629] = 1281,
  [5630] = 4947,
  [5631] = 4944,
  [5632] = 1382,
  [5633] = 4937,
  [5634] = 4911,
  [5635] = 4936,
  [5636] = 4929,
  [5637] = 4648,
  [5638] = 4900,
  [5639] = 4922,
  [5640] = 4890,
  [5641] = 4881,
  [5642] = 4872,
  [5643] = 1146,
  [5644] = 4642,
  [5645] = 4908,
  [5646] = 4867,
  [5647] = 4609,
  [5648] = 4607,
  [5649] = 4915,
  [5650] = 4878,
  [5651] = 1149,
  [5652] = 4712,
  [5653] = 4882,
  [5654] = 492,
  [5655] = 488,
  [5656] = 4952,
  [5657] = 494,
  [5658] = 4892,
  [5659] = 4880,
  [5660] = 483,
  [5661] = 4898,
  [5662] = 486,
  [5663] = 487,
  [5664] = 479,
  [5665] = 4884,
  [5666] = 490,
  [5667] = 4883,
  [5668] = 478,
  [5669] = 477,
  [5670] = 491,
  [5671] = 1301,
  [5672] = 4877,
  [5673] = 4916,
  [5674] = 1429,
  [5675] = 4927,
  [5676] = 4615,
  [5677] = 4876,
  [5678] = 4874,
  [5679] = 4866,
  [5680] = 482,
  [5681] = 4868,
  [5682] = 4960,
  [5683] = 474,
  [5684] = 4895,
  [5685] = 4873,
  [5686] = 4871,
  [5687] = 4931,
  [5688] = 485,
  [5689] = 4938,
  [5690] = 4940,
  [5691] = 5028,
  [5692] = 1607,
  [5693] = 476,
  [5694] = 5027,
  [5695] = 4949,
  [5696] = 489,
  [5697] = 5697,
  [5698] = 465,
  [5699] = 464,
  [5700] = 5700,
  [5701] = 495,
  [5702] = 5702,
  [5703] = 5702,
  [5704] = 5702,
  [5705] = 5705,
  [5706] = 5705,
  [5707] = 5702,
  [5708] = 5705,
  [5709] = 5702,
  [5710] = 5702,
  [5711] = 5702,
  [5712] = 5702,
  [5713] = 5702,
  [5714] = 5702,
  [5715] = 5702,
  [5716] = 5705,
  [5717] = 5702,
  [5718] = 5705,
  [5719] = 5702,
  [5720] = 5705,
  [5721] = 5705,
  [5722] = 5705,
  [5723] = 5705,
  [5724] = 5705,
  [5725] = 5705,
  [5726] = 5702,
  [5727] = 5705,
  [5728] = 5702,
  [5729] = 5705,
  [5730] = 5705,
  [5731] = 5702,
  [5732] = 5702,
  [5733] = 5705,
  [5734] = 5705,
  [5735] = 5705,
  [5736] = 5705,
  [5737] = 5702,
  [5738] = 1534,
  [5739] = 5702,
  [5740] = 5702,
  [5741] = 5705,
  [5742] = 5702,
  [5743] = 5705,
  [5744] = 1281,
  [5745] = 5702,
  [5746] = 5705,
  [5747] = 1301,
  [5748] = 5702,
  [5749] = 5702,
  [5750] = 5705,
  [5751] = 5702,
  [5752] = 5702,
  [5753] = 5705,
  [5754] = 5705,
  [5755] = 5705,
  [5756] = 5702,
  [5757] = 5705,
  [5758] = 5705,
  [5759] = 5702,
  [5760] = 5702,
  [5761] = 5705,
  [5762] = 5705,
  [5763] = 5702,
  [5764] = 5705,
  [5765] = 5702,
  [5766] = 5702,
  [5767] = 5702,
  [5768] = 5768,
  [5769] = 5705,
  [5770] = 5702,
  [5771] = 5702,
  [5772] = 5705,
  [5773] = 5705,
  [5774] = 5702,
  [5775] = 5705,
  [5776] = 5702,
  [5777] = 5705,
  [5778] = 5702,
  [5779] = 5705,
  [5780] = 5705,
  [5781] = 5702,
  [5782] = 5705,
  [5783] = 5705,
  [5784] = 5705,
  [5785] = 5702,
  [5786] = 5702,
  [5787] = 5705,
  [5788] = 5705,
  [5789] = 1607,
  [5790] = 5702,
  [5791] = 5705,
  [5792] = 5702,
  [5793] = 5705,
  [5794] = 5702,
  [5795] = 5702,
  [5796] = 5705,
  [5797] = 5705,
  [5798] = 5702,
  [5799] = 5705,
  [5800] = 5705,
  [5801] = 5702,
  [5802] = 5702,
  [5803] = 5702,
  [5804] = 5705,
  [5805] = 5702,
  [5806] = 5705,
  [5807] = 5705,
  [5808] = 5702,
  [5809] = 5705,
  [5810] = 5702,
  [5811] = 5705,
  [5812] = 5702,
  [5813] = 5702,
  [5814] = 5705,
  [5815] = 5705,
  [5816] = 5702,
  [5817] = 5705,
  [5818] = 5702,
  [5819] = 5705,
  [5820] = 5702,
  [5821] = 5702,
  [5822] = 5705,
  [5823] = 5705,
  [5824] = 5702,
  [5825] = 5705,
  [5826] = 5702,
  [5827] = 5705,
  [5828] = 5702,
  [5829] = 5702,
  [5830] = 5705,
  [5831] = 5702,
  [5832] = 5705,
  [5833] = 5705,
  [5834] = 5702,
  [5835] = 5705,
  [5836] = 5702,
  [5837] = 5705,
  [5838] = 5702,
  [5839] = 5702,
  [5840] = 5705,
  [5841] = 5702,
  [5842] = 5705,
  [5843] = 5702,
  [5844] = 5705,
  [5845] = 5702,
  [5846] = 5705,
  [5847] = 5702,
  [5848] = 5702,
  [5849] = 5705,
  [5850] = 5705,
  [5851] = 5705,
  [5852] = 5705,
  [5853] = 5702,
  [5854] = 5705,
  [5855] = 5702,
  [5856] = 5705,
  [5857] = 5702,
  [5858] = 5705,
  [5859] = 5702,
  [5860] = 5702,
  [5861] = 5705,
  [5862] = 5702,
  [5863] = 5705,
  [5864] = 5702,
  [5865] = 5865,
  [5866] = 5866,
  [5867] = 5866,
  [5868] = 5866,
  [5869] = 5866,
  [5870] = 5865,
  [5871] = 5871,
  [5872] = 5871,
  [5873] = 5871,
  [5874] = 5871,
  [5875] = 5865,
  [5876] = 5865,
  [5877] = 5871,
  [5878] = 5866,
  [5879] = 5865,
  [5880] = 5866,
  [5881] = 5866,
  [5882] = 5866,
  [5883] = 5865,
  [5884] = 5865,
  [5885] = 5871,
  [5886] = 5866,
  [5887] = 5871,
  [5888] = 5871,
  [5889] = 5871,
  [5890] = 5866,
  [5891] = 5866,
  [5892] = 5865,
  [5893] = 5871,
  [5894] = 5865,
  [5895] = 5865,
  [5896] = 5871,
  [5897] = 5866,
  [5898] = 5866,
  [5899] = 5866,
  [5900] = 5865,
  [5901] = 5871,
  [5902] = 5866,
  [5903] = 5865,
  [5904] = 5871,
  [5905] = 5871,
  [5906] = 5866,
  [5907] = 5865,
  [5908] = 5865,
  [5909] = 5871,
  [5910] = 5871,
  [5911] = 5871,
  [5912] = 5866,
  [5913] = 5866,
  [5914] = 5866,
  [5915] = 5865,
  [5916] = 5871,
  [5917] = 5866,
  [5918] = 5871,
  [5919] = 5865,
  [5920] = 5865,
  [5921] = 5871,
  [5922] = 5871,
  [5923] = 5865,
  [5924] = 5865,
  [5925] = 5865,
  [5926] = 5865,
  [5927] = 5866,
  [5928] = 5871,
  [5929] = 5866,
  [5930] = 5865,
  [5931] = 5871,
  [5932] = 5866,
  [5933] = 5871,
  [5934] = 5866,
  [5935] = 5865,
  [5936] = 5865,
  [5937] = 5871,
  [5938] = 5866,
  [5939] = 5866,
  [5940] = 5865,
  [5941] = 5871,
  [5942] = 5871,
  [5943] = 5866,
  [5944] = 5866,
  [5945] = 5871,
  [5946] = 5866,
  [5947] = 5866,
  [5948] = 5871,
  [5949] = 5865,
  [5950] = 5865,
  [5951] = 5866,
  [5952] = 5865,
  [5953] = 5871,
  [5954] = 5871,
  [5955] = 5871,
  [5956] = 5865,
  [5957] = 5866,
  [5958] = 5865,
  [5959] = 5866,
  [5960] = 5871,
  [5961] = 5866,
  [5962] = 5866,
  [5963] = 5866,
  [5964] = 5866,
  [5965] = 5866,
  [5966] = 5866,
  [5967] = 5871,
  [5968] = 5865,
  [5969] = 5865,
  [5970] = 5871,
  [5971] = 5865,
  [5972] = 5866,
  [5973] = 5871,
  [5974] = 5865,
  [5975] = 5871,
  [5976] = 5865,
  [5977] = 5871,
  [5978] = 5871,
  [5979] = 5865,
  [5980] = 5866,
  [5981] = 5865,
  [5982] = 5866,
  [5983] = 5871,
  [5984] = 5866,
  [5985] = 5871,
  [5986] = 5866,
  [5987] = 5866,
  [5988] = 5871,
  [5989] = 5871,
  [5990] = 5866,
  [5991] = 5871,
  [5992] = 5865,
  [5993] = 5866,
  [5994] = 5865,
  [5995] = 5871,
  [5996] = 5871,
  [5997] = 5865,
  [5998] = 5866,
  [5999] = 5866,
  [6000] = 5866,
  [6001] = 5871,
  [6002] = 5871,
  [6003] = 5871,
  [6004] = 5871,
  [6005] = 5865,
  [6006] = 5866,
  [6007] = 5865,
  [6008] = 5871,
  [6009] = 5871,
  [6010] = 5866,
  [6011] = 5871,
  [6012] = 5865,
  [6013] = 5871,
  [6014] = 5865,
  [6015] = 5866,
  [6016] = 5871,
  [6017] = 5871,
  [6018] = 5866,
  [6019] = 5866,
  [6020] = 5866,
  [6021] = 5871,
  [6022] = 5871,
  [6023] = 5866,
  [6024] = 5865,
  [6025] = 5866,
  [6026] = 5866,
  [6027] = 5871,
  [6028] = 5865,
  [6029] = 5871,
  [6030] = 5865,
  [6031] = 5871,
  [6032] = 5866,
  [6033] = 5871,
  [6034] = 5865,
  [6035] = 5865,
  [6036] = 5866,
  [6037] = 5871,
  [6038] = 5865,
  [6039] = 5865,
  [6040] = 5871,
  [6041] = 5866,
  [6042] = 5871,
  [6043] = 5866,
  [6044] = 5865,
  [6045] = 5871,
  [6046] = 5871,
  [6047] = 5865,
  [6048] = 5865,
  [6049] = 5871,
  [6050] = 5866,
  [6051] = 5865,
  [6052] = 5871,
  [6053] = 5865,
  [6054] = 5866,
  [6055] = 5871,
  [6056] = 5866,
  [6057] = 5866,
  [6058] = 5866,
  [6059] = 5866,
  [6060] = 5865,
  [6061] = 5865,
  [6062] = 5871,
  [6063] = 5871,
  [6064] = 5866,
  [6065] = 5871,
  [6066] = 5865,
  [6067] = 5866,
  [6068] = 5866,
  [6069] = 5866,
  [6070] = 5865,
  [6071] = 5865,
  [6072] = 5866,
  [6073] = 5871,
  [6074] = 5871,
  [6075] = 5866,
  [6076] = 5866,
  [6077] = 5871,
  [6078] = 5865,
  [6079] = 5866,
  [6080] = 5871,
  [6081] = 5871,
  [6082] = 5865,
  [6083] = 5866,
  [6084] = 5865,
  [6085] = 5865,
  [6086] = 5871,
  [6087] = 5866,
  [6088] = 6088,
  [6089] = 6088,
  [6090] = 6088,
  [6091] = 6088,
  [6092] = 6088,
  [6093] = 6088,
  [6094] = 6088,
  [6095] = 488,
  [6096] = 492,
  [6097] = 6088,
  [6098] = 6098,
  [6099] = 6099,
  [6100] = 6100,
  [6101] = 6099,
  [6102] = 6099,
  [6103] = 6099,
  [6104] = 6100,
  [6105] = 6100,
  [6106] = 6100,
  [6107] = 6100,
  [6108] = 6099,
  [6109] = 6100,
  [6110] = 6099,
  [6111] = 6099,
  [6112] = 6099,
  [6113] = 6100,
  [6114] = 6099,
  [6115] = 6100,
  [6116] = 6100,
  [6117] = 6117,
  [6118] = 6117,
  [6119] = 6117,
  [6120] = 6117,
  [6121] = 6121,
  [6122] = 479,
  [6123] = 478,
  [6124] = 486,
  [6125] = 6125,
  [6126] = 6117,
  [6127] = 6127,
  [6128] = 6128,
  [6129] = 6117,
  [6130] = 6117,
  [6131] = 6131,
  [6132] = 6117,
  [6133] = 6117,
  [6134] = 6117,
  [6135] = 6117,
  [6136] = 6117,
  [6137] = 6137,
  [6138] = 474,
  [6139] = 6117,
  [6140] = 492,
  [6141] = 485,
  [6142] = 6117,
  [6143] = 488,
  [6144] = 6117,
  [6145] = 6117,
  [6146] = 6117,
  [6147] = 477,
  [6148] = 6117,
  [6149] = 6117,
  [6150] = 6117,
  [6151] = 6117,
  [6152] = 6117,
  [6153] = 6117,
  [6154] = 491,
  [6155] = 6117,
  [6156] = 6117,
  [6157] = 6117,
  [6158] = 6117,
  [6159] = 6159,
  [6160] = 6160,
  [6161] = 6161,
  [6162] = 6161,
  [6163] = 6161,
  [6164] = 1173,
  [6165] = 6161,
  [6166] = 6161,
  [6167] = 6160,
  [6168] = 6159,
  [6169] = 6159,
  [6170] = 6161,
  [6171] = 6160,
  [6172] = 6160,
  [6173] = 6159,
  [6174] = 6160,
  [6175] = 6160,
  [6176] = 6176,
  [6177] = 6159,
  [6178] = 6161,
  [6179] = 6179,
  [6180] = 6159,
  [6181] = 6159,
  [6182] = 6161,
  [6183] = 6183,
  [6184] = 6159,
  [6185] = 6160,
  [6186] = 6160,
  [6187] = 6187,
  [6188] = 469,
  [6189] = 6189,
  [6190] = 6190,
  [6191] = 6189,
  [6192] = 6192,
  [6193] = 473,
  [6194] = 6194,
  [6195] = 6195,
  [6196] = 6196,
  [6197] = 6197,
  [6198] = 6192,
  [6199] = 6192,
  [6200] = 6189,
  [6201] = 470,
  [6202] = 6192,
  [6203] = 6189,
  [6204] = 468,
  [6205] = 483,
  [6206] = 489,
  [6207] = 474,
  [6208] = 6208,
  [6209] = 481,
  [6210] = 476,
  [6211] = 492,
  [6212] = 6212,
  [6213] = 6213,
  [6214] = 488,
  [6215] = 6215,
  [6216] = 486,
  [6217] = 486,
  [6218] = 6218,
  [6219] = 6213,
  [6220] = 6213,
  [6221] = 479,
  [6222] = 492,
  [6223] = 6213,
  [6224] = 478,
  [6225] = 482,
  [6226] = 477,
  [6227] = 6227,
  [6228] = 484,
  [6229] = 480,
  [6230] = 475,
  [6231] = 485,
  [6232] = 478,
  [6233] = 491,
  [6234] = 479,
  [6235] = 488,
  [6236] = 484,
  [6237] = 479,
  [6238] = 480,
  [6239] = 486,
  [6240] = 494,
  [6241] = 491,
  [6242] = 477,
  [6243] = 487,
  [6244] = 490,
  [6245] = 478,
  [6246] = 477,
  [6247] = 491,
  [6248] = 6248,
  [6249] = 6248,
  [6250] = 6248,
  [6251] = 6248,
  [6252] = 6248,
  [6253] = 6248,
  [6254] = 6248,
  [6255] = 6248,
  [6256] = 6256,
  [6257] = 6248,
  [6258] = 6258,
  [6259] = 6259,
  [6260] = 6260,
  [6261] = 6261,
  [6262] = 6262,
  [6263] = 6248,
  [6264] = 6264,
  [6265] = 6265,
  [6266] = 6266,
  [6267] = 6267,
  [6268] = 6268,
  [6269] = 6248,
  [6270] = 6270,
  [6271] = 6248,
  [6272] = 6248,
  [6273] = 6248,
  [6274] = 6248,
  [6275] = 6275,
  [6276] = 6248,
  [6277] = 6248,
  [6278] = 6248,
  [6279] = 6279,
  [6280] = 6280,
  [6281] = 6248,
  [6282] = 6248,
  [6283] = 6248,
  [6284] = 6248,
  [6285] = 6248,
  [6286] = 6286,
  [6287] = 6287,
  [6288] = 6248,
  [6289] = 6265,
  [6290] = 6248,
  [6291] = 6248,
  [6292] = 6248,
  [6293] = 490,
  [6294] = 6248,
  [6295] = 6248,
  [6296] = 6248,
  [6297] = 6248,
  [6298] = 6248,
  [6299] = 6299,
  [6300] = 6300,
  [6301] = 6301,
  [6302] = 6302,
  [6303] = 6303,
  [6304] = 495,
  [6305] = 6248,
  [6306] = 487,
  [6307] = 6248,
  [6308] = 6248,
  [6309] = 6248,
  [6310] = 6310,
  [6311] = 6248,
  [6312] = 6312,
  [6313] = 6248,
  [6314] = 6314,
  [6315] = 6248,
  [6316] = 6248,
  [6317] = 481,
  [6318] = 6318,
  [6319] = 6319,
  [6320] = 6248,
  [6321] = 6248,
  [6322] = 6248,
  [6323] = 6248,
  [6324] = 6248,
  [6325] = 6248,
  [6326] = 469,
  [6327] = 6265,
  [6328] = 6328,
  [6329] = 6248,
  [6330] = 6330,
  [6331] = 6265,
  [6332] = 6248,
  [6333] = 6248,
  [6334] = 6334,
  [6335] = 6248,
  [6336] = 6248,
  [6337] = 6337,
  [6338] = 6338,
  [6339] = 6337,
  [6340] = 6337,
  [6341] = 6248,
  [6342] = 6248,
  [6343] = 6248,
  [6344] = 6344,
  [6345] = 6248,
  [6346] = 6346,
  [6347] = 6248,
  [6348] = 6248,
  [6349] = 6248,
  [6350] = 6248,
  [6351] = 6248,
  [6352] = 6248,
  [6353] = 6353,
  [6354] = 6248,
  [6355] = 6248,
  [6356] = 6248,
  [6357] = 6248,
  [6358] = 6248,
  [6359] = 6248,
  [6360] = 6248,
  [6361] = 6361,
  [6362] = 6248,
  [6363] = 6337,
  [6364] = 6364,
  [6365] = 6365,
  [6366] = 6248,
  [6367] = 6248,
  [6368] = 6368,
  [6369] = 6369,
  [6370] = 2702,
  [6371] = 6371,
  [6372] = 6372,
  [6373] = 470,
  [6374] = 6374,
  [6375] = 6375,
  [6376] = 6374,
  [6377] = 6377,
  [6378] = 6371,
  [6379] = 6379,
  [6380] = 6380,
  [6381] = 6381,
  [6382] = 6382,
  [6383] = 6383,
  [6384] = 6380,
  [6385] = 6377,
  [6386] = 6386,
  [6387] = 6369,
  [6388] = 2704,
  [6389] = 6389,
  [6390] = 6390,
  [6391] = 6391,
  [6392] = 6371,
  [6393] = 6393,
  [6394] = 6394,
  [6395] = 6375,
  [6396] = 6386,
  [6397] = 6397,
  [6398] = 6369,
  [6399] = 6382,
  [6400] = 6400,
  [6401] = 6368,
  [6402] = 6402,
  [6403] = 2696,
  [6404] = 6382,
  [6405] = 6405,
  [6406] = 6375,
  [6407] = 2694,
  [6408] = 6381,
  [6409] = 6409,
  [6410] = 6410,
  [6411] = 6397,
  [6412] = 6377,
  [6413] = 6374,
  [6414] = 6414,
  [6415] = 6415,
  [6416] = 6397,
  [6417] = 6386,
  [6418] = 6391,
  [6419] = 468,
  [6420] = 6420,
  [6421] = 6368,
  [6422] = 6369,
  [6423] = 6391,
  [6424] = 6424,
  [6425] = 6425,
  [6426] = 6426,
  [6427] = 6427,
  [6428] = 6428,
  [6429] = 6391,
  [6430] = 6430,
  [6431] = 6430,
  [6432] = 6377,
  [6433] = 6371,
  [6434] = 6414,
  [6435] = 478,
  [6436] = 477,
  [6437] = 6371,
  [6438] = 491,
  [6439] = 473,
  [6440] = 6377,
  [6441] = 6386,
  [6442] = 6397,
  [6443] = 6381,
  [6444] = 6383,
  [6445] = 6380,
  [6446] = 6383,
  [6447] = 6375,
  [6448] = 6448,
  [6449] = 6449,
  [6450] = 6414,
  [6451] = 6380,
  [6452] = 6397,
  [6453] = 6380,
  [6454] = 6414,
  [6455] = 6382,
  [6456] = 6368,
  [6457] = 6430,
  [6458] = 6414,
  [6459] = 6430,
  [6460] = 6460,
  [6461] = 6368,
  [6462] = 6430,
  [6463] = 6463,
  [6464] = 6383,
  [6465] = 6465,
  [6466] = 6375,
  [6467] = 6382,
  [6468] = 6369,
  [6469] = 6374,
  [6470] = 6470,
  [6471] = 6381,
  [6472] = 6472,
  [6473] = 6391,
  [6474] = 6474,
  [6475] = 6475,
  [6476] = 6476,
  [6477] = 6477,
  [6478] = 6478,
  [6479] = 6383,
  [6480] = 486,
  [6481] = 6481,
  [6482] = 6482,
  [6483] = 478,
  [6484] = 477,
  [6485] = 491,
  [6486] = 483,
  [6487] = 6487,
  [6488] = 6488,
  [6489] = 6489,
  [6490] = 475,
  [6491] = 6491,
  [6492] = 6492,
  [6493] = 6493,
  [6494] = 482,
  [6495] = 6495,
  [6496] = 6496,
  [6497] = 479,
  [6498] = 6498,
  [6499] = 478,
  [6500] = 494,
  [6501] = 6501,
  [6502] = 486,
  [6503] = 488,
  [6504] = 492,
  [6505] = 1556,
  [6506] = 492,
  [6507] = 488,
  [6508] = 6495,
  [6509] = 490,
  [6510] = 487,
  [6511] = 6511,
  [6512] = 6495,
  [6513] = 489,
  [6514] = 481,
  [6515] = 487,
  [6516] = 479,
  [6517] = 1542,
  [6518] = 490,
  [6519] = 484,
  [6520] = 481,
  [6521] = 6521,
  [6522] = 475,
  [6523] = 480,
  [6524] = 6524,
  [6525] = 491,
  [6526] = 485,
  [6527] = 2930,
  [6528] = 2929,
  [6529] = 2931,
  [6530] = 6530,
  [6531] = 6531,
  [6532] = 477,
  [6533] = 6533,
  [6534] = 476,
  [6535] = 474,
  [6536] = 6536,
  [6537] = 478,
  [6538] = 477,
  [6539] = 6539,
  [6540] = 491,
  [6541] = 6541,
  [6542] = 6542,
  [6543] = 6543,
  [6544] = 6495,
  [6545] = 494,
  [6546] = 2926,
  [6547] = 6547,
  [6548] = 6548,
  [6549] = 6547,
  [6550] = 6463,
  [6551] = 6551,
  [6552] = 6548,
  [6553] = 6553,
  [6554] = 491,
  [6555] = 477,
  [6556] = 478,
  [6557] = 490,
  [6558] = 6558,
  [6559] = 6548,
  [6560] = 6547,
  [6561] = 485,
  [6562] = 6548,
  [6563] = 479,
  [6564] = 487,
  [6565] = 6548,
  [6566] = 6558,
  [6567] = 6567,
  [6568] = 6558,
  [6569] = 6569,
  [6570] = 6570,
  [6571] = 486,
  [6572] = 6558,
  [6573] = 481,
  [6574] = 6548,
  [6575] = 6547,
  [6576] = 6548,
  [6577] = 6548,
  [6578] = 6547,
  [6579] = 6548,
  [6580] = 6548,
  [6581] = 495,
  [6582] = 6582,
  [6583] = 2696,
  [6584] = 6558,
  [6585] = 2694,
  [6586] = 6548,
  [6587] = 6587,
  [6588] = 6548,
  [6589] = 6589,
  [6590] = 6590,
  [6591] = 6591,
  [6592] = 6548,
  [6593] = 6548,
  [6594] = 6548,
  [6595] = 6595,
  [6596] = 6596,
  [6597] = 6597,
  [6598] = 6598,
  [6599] = 6599,
  [6600] = 6548,
  [6601] = 6548,
  [6602] = 6548,
  [6603] = 6547,
  [6604] = 6547,
  [6605] = 6551,
  [6606] = 6548,
  [6607] = 6548,
  [6608] = 6548,
  [6609] = 2702,
  [6610] = 2704,
  [6611] = 6611,
  [6612] = 6548,
  [6613] = 6548,
  [6614] = 6558,
  [6615] = 474,
  [6616] = 6547,
  [6617] = 6548,
  [6618] = 6618,
  [6619] = 6547,
  [6620] = 6620,
  [6621] = 6548,
  [6622] = 6548,
  [6623] = 6548,
  [6624] = 6548,
  [6625] = 6548,
  [6626] = 6626,
  [6627] = 6627,
  [6628] = 6547,
  [6629] = 6548,
  [6630] = 6548,
  [6631] = 6631,
  [6632] = 6632,
  [6633] = 6548,
  [6634] = 6547,
  [6635] = 490,
  [6636] = 6547,
  [6637] = 6637,
  [6638] = 6547,
  [6639] = 6548,
  [6640] = 6589,
  [6641] = 6547,
  [6642] = 6548,
  [6643] = 6548,
  [6644] = 6548,
  [6645] = 487,
  [6646] = 6646,
  [6647] = 6548,
  [6648] = 6648,
  [6649] = 6649,
  [6650] = 6548,
  [6651] = 495,
  [6652] = 6652,
  [6653] = 6548,
  [6654] = 6654,
  [6655] = 6548,
  [6656] = 6548,
  [6657] = 6548,
  [6658] = 6658,
  [6659] = 6589,
  [6660] = 6548,
  [6661] = 6548,
  [6662] = 6547,
  [6663] = 481,
  [6664] = 6547,
  [6665] = 6665,
  [6666] = 6666,
  [6667] = 6667,
  [6668] = 6547,
  [6669] = 6548,
  [6670] = 6670,
  [6671] = 6551,
  [6672] = 6672,
  [6673] = 6547,
  [6674] = 6558,
  [6675] = 6675,
  [6676] = 6676,
  [6677] = 2559,
  [6678] = 6678,
  [6679] = 6463,
  [6680] = 6548,
  [6681] = 6681,
  [6682] = 6548,
  [6683] = 6548,
  [6684] = 6548,
  [6685] = 6548,
  [6686] = 6547,
  [6687] = 6548,
  [6688] = 6548,
  [6689] = 6689,
  [6690] = 6690,
  [6691] = 6548,
  [6692] = 6548,
  [6693] = 6548,
  [6694] = 6694,
  [6695] = 6695,
  [6696] = 6547,
  [6697] = 6548,
  [6698] = 6548,
  [6699] = 6548,
  [6700] = 6700,
  [6701] = 6548,
  [6702] = 6702,
  [6703] = 6703,
  [6704] = 6547,
  [6705] = 6547,
  [6706] = 6706,
  [6707] = 6707,
  [6708] = 6548,
  [6709] = 6709,
  [6710] = 6710,
  [6711] = 6711,
  [6712] = 6548,
  [6713] = 6713,
  [6714] = 6714,
  [6715] = 6548,
  [6716] = 6716,
  [6717] = 6547,
  [6718] = 6558,
  [6719] = 6719,
  [6720] = 6548,
  [6721] = 6548,
  [6722] = 6722,
  [6723] = 6548,
  [6724] = 6548,
  [6725] = 6725,
  [6726] = 6726,
  [6727] = 6726,
  [6728] = 6531,
  [6729] = 6729,
  [6730] = 6730,
  [6731] = 6726,
  [6732] = 6730,
  [6733] = 2930,
  [6734] = 2931,
  [6735] = 2636,
  [6736] = 6726,
  [6737] = 6726,
  [6738] = 6531,
  [6739] = 6726,
  [6740] = 476,
  [6741] = 6481,
  [6742] = 6481,
  [6743] = 6743,
  [6744] = 6743,
  [6745] = 6730,
  [6746] = 6726,
  [6747] = 6747,
  [6748] = 6748,
  [6749] = 6749,
  [6750] = 6726,
  [6751] = 6726,
  [6752] = 6729,
  [6753] = 490,
  [6754] = 6726,
  [6755] = 6755,
  [6756] = 6726,
  [6757] = 6726,
  [6758] = 6726,
  [6759] = 6726,
  [6760] = 6726,
  [6761] = 489,
  [6762] = 6493,
  [6763] = 6743,
  [6764] = 6764,
  [6765] = 481,
  [6766] = 2724,
  [6767] = 487,
  [6768] = 6493,
  [6769] = 6726,
  [6770] = 6743,
  [6771] = 6726,
  [6772] = 6726,
  [6773] = 6726,
  [6774] = 6726,
  [6775] = 6775,
  [6776] = 2929,
  [6777] = 6729,
  [6778] = 2926,
  [6779] = 6779,
  [6780] = 6780,
  [6781] = 6726,
  [6782] = 6782,
  [6783] = 6730,
  [6784] = 6784,
  [6785] = 6729,
  [6786] = 6726,
  [6787] = 6787,
  [6788] = 6788,
  [6789] = 6789,
  [6790] = 6726,
  [6791] = 6791,
  [6792] = 6792,
  [6793] = 6793,
  [6794] = 6794,
  [6795] = 6795,
  [6796] = 6796,
  [6797] = 6793,
  [6798] = 6798,
  [6799] = 6799,
  [6800] = 6800,
  [6801] = 6801,
  [6802] = 6793,
  [6803] = 6793,
  [6804] = 6801,
  [6805] = 6805,
  [6806] = 6806,
  [6807] = 6800,
  [6808] = 6795,
  [6809] = 6799,
  [6810] = 6800,
  [6811] = 6793,
  [6812] = 6793,
  [6813] = 6805,
  [6814] = 6814,
  [6815] = 6799,
  [6816] = 6794,
  [6817] = 6814,
  [6818] = 6801,
  [6819] = 6795,
  [6820] = 6793,
  [6821] = 6801,
  [6822] = 6795,
  [6823] = 6794,
  [6824] = 6824,
  [6825] = 6825,
  [6826] = 6826,
  [6827] = 6794,
  [6828] = 6814,
  [6829] = 6793,
  [6830] = 6814,
  [6831] = 6799,
  [6832] = 6800,
  [6833] = 6833,
  [6834] = 6834,
  [6835] = 6835,
  [6836] = 6836,
  [6837] = 6792,
  [6838] = 6793,
  [6839] = 6806,
  [6840] = 6795,
  [6841] = 6800,
  [6842] = 6800,
  [6843] = 6843,
  [6844] = 6844,
  [6845] = 6796,
  [6846] = 6793,
  [6847] = 6793,
  [6848] = 6805,
  [6849] = 6799,
  [6850] = 6795,
  [6851] = 6851,
  [6852] = 6814,
  [6853] = 6836,
  [6854] = 6794,
  [6855] = 6795,
  [6856] = 6793,
  [6857] = 6801,
  [6858] = 6814,
  [6859] = 6801,
  [6860] = 6800,
  [6861] = 6799,
  [6862] = 6862,
  [6863] = 6794,
  [6864] = 6795,
  [6865] = 6793,
  [6866] = 6835,
  [6867] = 6801,
  [6868] = 6836,
  [6869] = 6814,
  [6870] = 6792,
  [6871] = 6806,
  [6872] = 6801,
  [6873] = 6843,
  [6874] = 6793,
  [6875] = 6844,
  [6876] = 6876,
  [6877] = 6795,
  [6878] = 6794,
  [6879] = 6799,
  [6880] = 6795,
  [6881] = 6794,
  [6882] = 6882,
  [6883] = 6793,
  [6884] = 6884,
  [6885] = 6814,
  [6886] = 6886,
  [6887] = 6793,
  [6888] = 6799,
  [6889] = 6805,
  [6890] = 6890,
  [6891] = 6800,
  [6892] = 6793,
  [6893] = 6801,
  [6894] = 6894,
  [6895] = 6800,
  [6896] = 6801,
  [6897] = 6794,
  [6898] = 6898,
  [6899] = 6795,
  [6900] = 6800,
  [6901] = 6793,
  [6902] = 6799,
  [6903] = 6794,
  [6904] = 6800,
  [6905] = 6794,
  [6906] = 6799,
  [6907] = 6795,
  [6908] = 6800,
  [6909] = 6801,
  [6910] = 6793,
  [6911] = 6795,
  [6912] = 6814,
  [6913] = 6913,
  [6914] = 6914,
  [6915] = 6801,
  [6916] = 6814,
  [6917] = 6914,
  [6918] = 6799,
  [6919] = 6793,
  [6920] = 6913,
  [6921] = 6921,
  [6922] = 6800,
  [6923] = 6923,
  [6924] = 6799,
  [6925] = 6800,
  [6926] = 6793,
  [6927] = 6799,
  [6928] = 6835,
  [6929] = 6794,
  [6930] = 6794,
  [6931] = 6836,
  [6932] = 6795,
  [6933] = 6793,
  [6934] = 6795,
  [6935] = 6792,
  [6936] = 6806,
  [6937] = 6801,
  [6938] = 6801,
  [6939] = 6843,
  [6940] = 6793,
  [6941] = 6844,
  [6942] = 6942,
  [6943] = 6794,
  [6944] = 6843,
  [6945] = 6844,
  [6946] = 6946,
  [6947] = 6793,
  [6948] = 6882,
  [6949] = 6814,
  [6950] = 6884,
  [6951] = 6799,
  [6952] = 6886,
  [6953] = 6793,
  [6954] = 6793,
  [6955] = 6805,
  [6956] = 6890,
  [6957] = 6957,
  [6958] = 6958,
  [6959] = 6959,
  [6960] = 6960,
  [6961] = 6793,
  [6962] = 6834,
  [6963] = 6801,
  [6964] = 6964,
  [6965] = 6965,
  [6966] = 6966,
  [6967] = 6967,
  [6968] = 6793,
  [6969] = 6833,
  [6970] = 6894,
  [6971] = 6814,
  [6972] = 6972,
  [6973] = 6973,
  [6974] = 6974,
  [6975] = 6793,
  [6976] = 6794,
  [6977] = 6977,
  [6978] = 6978,
  [6979] = 6979,
  [6980] = 6980,
  [6981] = 6981,
  [6982] = 6793,
  [6983] = 6799,
  [6984] = 6800,
  [6985] = 6985,
  [6986] = 6986,
  [6987] = 6987,
  [6988] = 6988,
  [6989] = 6793,
  [6990] = 6834,
  [6991] = 6991,
  [6992] = 6992,
  [6993] = 6993,
  [6994] = 6994,
  [6995] = 6995,
  [6996] = 6793,
  [6997] = 6997,
  [6998] = 6836,
  [6999] = 6999,
  [7000] = 7000,
  [7001] = 7001,
  [7002] = 7002,
  [7003] = 6793,
  [7004] = 6862,
  [7005] = 6851,
  [7006] = 7006,
  [7007] = 6824,
  [7008] = 7008,
  [7009] = 6826,
  [7010] = 6793,
  [7011] = 6796,
  [7012] = 6921,
  [7013] = 6923,
  [7014] = 7014,
  [7015] = 6814,
  [7016] = 6942,
  [7017] = 6793,
  [7018] = 6833,
  [7019] = 6814,
  [7020] = 7020,
  [7021] = 6898,
  [7022] = 7022,
  [7023] = 7023,
  [7024] = 6793,
  [7025] = 7025,
  [7026] = 7026,
  [7027] = 7027,
  [7028] = 6835,
  [7029] = 7029,
  [7030] = 6835,
  [7031] = 6793,
  [7032] = 6801,
  [7033] = 6795,
  [7034] = 6795,
  [7035] = 6898,
  [7036] = 6792,
  [7037] = 7037,
  [7038] = 6793,
  [7039] = 6806,
  [7040] = 6794,
  [7041] = 6843,
  [7042] = 6844,
  [7043] = 6794,
  [7044] = 6997,
  [7045] = 6793,
  [7046] = 6800,
  [7047] = 6799,
  [7048] = 6799,
  [7049] = 7020,
  [7050] = 6794,
  [7051] = 6800,
  [7052] = 6793,
  [7053] = 6914,
  [7054] = 6814,
  [7055] = 6882,
  [7056] = 6991,
  [7057] = 6884,
  [7058] = 6795,
  [7059] = 6793,
  [7060] = 6801,
  [7061] = 6886,
  [7062] = 6793,
  [7063] = 6805,
  [7064] = 6890,
  [7065] = 7065,
  [7066] = 6793,
  [7067] = 6801,
  [7068] = 6894,
  [7069] = 6981,
  [7070] = 6795,
  [7071] = 6814,
  [7072] = 6795,
  [7073] = 6793,
  [7074] = 6814,
  [7075] = 6794,
  [7076] = 6814,
  [7077] = 6794,
  [7078] = 6799,
  [7079] = 6795,
  [7080] = 6793,
  [7081] = 6799,
  [7082] = 6800,
  [7083] = 6794,
  [7084] = 6799,
  [7085] = 6795,
  [7086] = 6794,
  [7087] = 6793,
  [7088] = 6799,
  [7089] = 6800,
  [7090] = 6814,
  [7091] = 6795,
  [7092] = 6801,
  [7093] = 6795,
  [7094] = 6793,
  [7095] = 6795,
  [7096] = 6794,
  [7097] = 6799,
  [7098] = 6801,
  [7099] = 6795,
  [7100] = 6794,
  [7101] = 6793,
  [7102] = 7102,
  [7103] = 6799,
  [7104] = 6800,
  [7105] = 6800,
  [7106] = 6814,
  [7107] = 6795,
  [7108] = 6793,
  [7109] = 6794,
  [7110] = 6794,
  [7111] = 6799,
  [7112] = 6800,
  [7113] = 6800,
  [7114] = 6981,
  [7115] = 6793,
  [7116] = 6814,
  [7117] = 6914,
  [7118] = 6795,
  [7119] = 6794,
  [7120] = 6801,
  [7121] = 6913,
  [7122] = 6793,
  [7123] = 6795,
  [7124] = 6794,
  [7125] = 6799,
  [7126] = 6800,
  [7127] = 6799,
  [7128] = 6800,
  [7129] = 6793,
  [7130] = 6795,
  [7131] = 6800,
  [7132] = 6794,
  [7133] = 6799,
  [7134] = 6991,
  [7135] = 6800,
  [7136] = 6793,
  [7137] = 6800,
  [7138] = 6795,
  [7139] = 6794,
  [7140] = 6799,
  [7141] = 6800,
  [7142] = 6826,
  [7143] = 6793,
  [7144] = 6799,
  [7145] = 6814,
  [7146] = 6800,
  [7147] = 6799,
  [7148] = 6799,
  [7149] = 6814,
  [7150] = 6793,
  [7151] = 6795,
  [7152] = 6794,
  [7153] = 6794,
  [7154] = 6801,
  [7155] = 6795,
  [7156] = 6794,
  [7157] = 6793,
  [7158] = 6795,
  [7159] = 6913,
  [7160] = 6799,
  [7161] = 6799,
  [7162] = 6800,
  [7163] = 6800,
  [7164] = 6793,
  [7165] = 6801,
  [7166] = 6795,
  [7167] = 6801,
  [7168] = 6794,
  [7169] = 6799,
  [7170] = 6800,
  [7171] = 6793,
  [7172] = 6814,
  [7173] = 6795,
  [7174] = 6794,
  [7175] = 6799,
  [7176] = 6814,
  [7177] = 6833,
  [7178] = 6793,
  [7179] = 6800,
  [7180] = 6795,
  [7181] = 6834,
  [7182] = 6794,
  [7183] = 6801,
  [7184] = 6800,
  [7185] = 6793,
  [7186] = 6799,
  [7187] = 6800,
  [7188] = 6913,
  [7189] = 6795,
  [7190] = 6794,
  [7191] = 6814,
  [7192] = 6793,
  [7193] = 6834,
  [7194] = 6799,
  [7195] = 6799,
  [7196] = 6800,
  [7197] = 6800,
  [7198] = 6833,
  [7199] = 6793,
  [7200] = 6799,
  [7201] = 6795,
  [7202] = 6794,
  [7203] = 6794,
  [7204] = 6799,
  [7205] = 7205,
  [7206] = 6793,
  [7207] = 6800,
  [7208] = 6794,
  [7209] = 6794,
  [7210] = 6801,
  [7211] = 6795,
  [7212] = 6795,
  [7213] = 6793,
  [7214] = 6794,
  [7215] = 6795,
  [7216] = 6795,
  [7217] = 6799,
  [7218] = 7218,
  [7219] = 6800,
  [7220] = 7220,
  [7221] = 7221,
  [7222] = 7222,
  [7223] = 7223,
  [7224] = 6801,
  [7225] = 7225,
  [7226] = 6814,
  [7227] = 6794,
  [7228] = 6795,
  [7229] = 6794,
  [7230] = 6801,
  [7231] = 6799,
  [7232] = 7232,
  [7233] = 7233,
  [7234] = 7234,
  [7235] = 7235,
  [7236] = 7236,
  [7237] = 7237,
  [7238] = 7238,
  [7239] = 7239,
  [7240] = 7240,
  [7241] = 7241,
  [7242] = 7242,
  [7243] = 7243,
  [7244] = 6814,
  [7245] = 6800,
  [7246] = 6799,
  [7247] = 6814,
  [7248] = 6814,
  [7249] = 6801,
  [7250] = 6794,
  [7251] = 7251,
  [7252] = 6795,
  [7253] = 6794,
  [7254] = 6799,
  [7255] = 6800,
  [7256] = 6799,
  [7257] = 6800,
  [7258] = 6882,
  [7259] = 6884,
  [7260] = 6800,
  [7261] = 6800,
  [7262] = 6799,
  [7263] = 6814,
  [7264] = 6801,
  [7265] = 7265,
  [7266] = 6795,
  [7267] = 6794,
  [7268] = 6894,
  [7269] = 7269,
  [7270] = 6799,
  [7271] = 6800,
  [7272] = 7272,
  [7273] = 6890,
  [7274] = 7274,
  [7275] = 6997,
  [7276] = 6814,
  [7277] = 6794,
  [7278] = 6814,
  [7279] = 6801,
  [7280] = 6795,
  [7281] = 6814,
  [7282] = 6794,
  [7283] = 6799,
  [7284] = 6800,
  [7285] = 7285,
  [7286] = 6801,
  [7287] = 6814,
  [7288] = 6801,
  [7289] = 6795,
  [7290] = 6794,
  [7291] = 6923,
  [7292] = 7205,
  [7293] = 6914,
  [7294] = 6801,
  [7295] = 6799,
  [7296] = 6800,
  [7297] = 6814,
  [7298] = 6800,
  [7299] = 6799,
  [7300] = 6795,
  [7301] = 6794,
  [7302] = 7302,
  [7303] = 7303,
  [7304] = 6801,
  [7305] = 6795,
  [7306] = 6799,
  [7307] = 6794,
  [7308] = 6800,
  [7309] = 6814,
  [7310] = 6801,
  [7311] = 6795,
  [7312] = 6799,
  [7313] = 6800,
  [7314] = 6794,
  [7315] = 6942,
  [7316] = 6799,
  [7317] = 6800,
  [7318] = 6913,
  [7319] = 6794,
  [7320] = 6921,
  [7321] = 6795,
  [7322] = 6814,
  [7323] = 6795,
  [7324] = 6814,
  [7325] = 6801,
  [7326] = 6886,
  [7327] = 6795,
  [7328] = 6794,
  [7329] = 6799,
  [7330] = 6793,
  [7331] = 6800,
  [7332] = 6800,
  [7333] = 6801,
  [7334] = 6801,
  [7335] = 6795,
  [7336] = 6799,
  [7337] = 6814,
  [7338] = 6801,
  [7339] = 6795,
  [7340] = 6794,
  [7341] = 6799,
  [7342] = 6794,
  [7343] = 6814,
  [7344] = 6805,
  [7345] = 6800,
  [7346] = 6814,
  [7347] = 6795,
  [7348] = 6800,
  [7349] = 7274,
  [7350] = 6914,
  [7351] = 6814,
  [7352] = 6890,
  [7353] = 6801,
  [7354] = 6795,
  [7355] = 6794,
  [7356] = 6799,
  [7357] = 6801,
  [7358] = 6795,
  [7359] = 6794,
  [7360] = 6799,
  [7361] = 6800,
  [7362] = 6834,
  [7363] = 6833,
  [7364] = 6801,
  [7365] = 6814,
  [7366] = 6801,
  [7367] = 6795,
  [7368] = 6794,
  [7369] = 6801,
  [7370] = 6795,
  [7371] = 6794,
  [7372] = 6799,
  [7373] = 6800,
  [7374] = 6800,
  [7375] = 6799,
  [7376] = 6814,
  [7377] = 6800,
  [7378] = 6799,
  [7379] = 6799,
  [7380] = 6800,
  [7381] = 6814,
  [7382] = 6801,
  [7383] = 6795,
  [7384] = 6794,
  [7385] = 6799,
  [7386] = 6800,
  [7387] = 6814,
  [7388] = 6801,
  [7389] = 7265,
  [7390] = 6795,
  [7391] = 6794,
  [7392] = 6894,
  [7393] = 7269,
  [7394] = 6799,
  [7395] = 6800,
  [7396] = 6890,
  [7397] = 7274,
  [7398] = 6800,
  [7399] = 6794,
  [7400] = 6795,
  [7401] = 7205,
  [7402] = 6991,
  [7403] = 6799,
  [7404] = 6814,
  [7405] = 6801,
  [7406] = 6814,
  [7407] = 6801,
  [7408] = 6913,
  [7409] = 6795,
  [7410] = 6794,
  [7411] = 6799,
  [7412] = 6800,
  [7413] = 6997,
  [7414] = 6794,
  [7415] = 6814,
  [7416] = 6862,
  [7417] = 6862,
  [7418] = 6801,
  [7419] = 6795,
  [7420] = 6794,
  [7421] = 6799,
  [7422] = 6800,
  [7423] = 6795,
  [7424] = 6801,
  [7425] = 6814,
  [7426] = 6801,
  [7427] = 6914,
  [7428] = 6801,
  [7429] = 6795,
  [7430] = 6794,
  [7431] = 6795,
  [7432] = 6799,
  [7433] = 6800,
  [7434] = 6794,
  [7435] = 6814,
  [7436] = 6799,
  [7437] = 6800,
  [7438] = 6801,
  [7439] = 7269,
  [7440] = 6851,
  [7441] = 6914,
  [7442] = 6814,
  [7443] = 6814,
  [7444] = 6913,
  [7445] = 6801,
  [7446] = 6795,
  [7447] = 6794,
  [7448] = 6799,
  [7449] = 6800,
  [7450] = 6814,
  [7451] = 6824,
  [7452] = 6814,
  [7453] = 6814,
  [7454] = 6801,
  [7455] = 6795,
  [7456] = 6801,
  [7457] = 6795,
  [7458] = 6794,
  [7459] = 6794,
  [7460] = 6799,
  [7461] = 6800,
  [7462] = 6801,
  [7463] = 6799,
  [7464] = 6800,
  [7465] = 6826,
  [7466] = 6796,
  [7467] = 6795,
  [7468] = 6814,
  [7469] = 7469,
  [7470] = 7470,
  [7471] = 6801,
  [7472] = 6795,
  [7473] = 6794,
  [7474] = 6794,
  [7475] = 6799,
  [7476] = 6800,
  [7477] = 7477,
  [7478] = 6981,
  [7479] = 7265,
  [7480] = 6921,
  [7481] = 6814,
  [7482] = 6894,
  [7483] = 7269,
  [7484] = 6801,
  [7485] = 6795,
  [7486] = 6890,
  [7487] = 7274,
  [7488] = 6794,
  [7489] = 7020,
  [7490] = 6814,
  [7491] = 7205,
  [7492] = 6799,
  [7493] = 6800,
  [7494] = 6923,
  [7495] = 6942,
  [7496] = 6814,
  [7497] = 6801,
  [7498] = 6795,
  [7499] = 6794,
  [7500] = 6799,
  [7501] = 6800,
  [7502] = 6801,
  [7503] = 7020,
  [7504] = 6898,
  [7505] = 6814,
  [7506] = 6801,
  [7507] = 6795,
  [7508] = 6794,
  [7509] = 6799,
  [7510] = 6800,
  [7511] = 6894,
  [7512] = 7269,
  [7513] = 6890,
  [7514] = 7274,
  [7515] = 6795,
  [7516] = 6894,
  [7517] = 7205,
  [7518] = 6794,
  [7519] = 6796,
  [7520] = 6799,
  [7521] = 6800,
  [7522] = 6799,
  [7523] = 6826,
  [7524] = 6800,
  [7525] = 6814,
  [7526] = 6801,
  [7527] = 6795,
  [7528] = 6851,
  [7529] = 6794,
  [7530] = 7530,
  [7531] = 6799,
  [7532] = 6814,
  [7533] = 6800,
  [7534] = 6814,
  [7535] = 6801,
  [7536] = 6795,
  [7537] = 7269,
  [7538] = 7269,
  [7539] = 7269,
  [7540] = 7269,
  [7541] = 7269,
  [7542] = 7269,
  [7543] = 7269,
  [7544] = 7269,
  [7545] = 7269,
  [7546] = 7269,
  [7547] = 7269,
  [7548] = 7269,
  [7549] = 7269,
  [7550] = 7269,
  [7551] = 7269,
  [7552] = 7269,
  [7553] = 7269,
  [7554] = 7269,
  [7555] = 7269,
  [7556] = 7269,
  [7557] = 7269,
  [7558] = 7269,
  [7559] = 7269,
  [7560] = 7269,
  [7561] = 7269,
  [7562] = 7269,
  [7563] = 7269,
  [7564] = 7269,
  [7565] = 7269,
  [7566] = 7269,
  [7567] = 7269,
  [7568] = 7269,
  [7569] = 7269,
  [7570] = 7269,
  [7571] = 7269,
  [7572] = 7269,
  [7573] = 7269,
  [7574] = 7269,
  [7575] = 7269,
  [7576] = 7269,
  [7577] = 7269,
  [7578] = 7269,
  [7579] = 7269,
  [7580] = 7269,
  [7581] = 7269,
  [7582] = 7269,
  [7583] = 7269,
  [7584] = 7269,
  [7585] = 7269,
  [7586] = 7269,
  [7587] = 7269,
  [7588] = 7269,
  [7589] = 7269,
  [7590] = 7269,
  [7591] = 7269,
  [7592] = 7269,
  [7593] = 7269,
  [7594] = 7269,
  [7595] = 7269,
  [7596] = 7269,
  [7597] = 7597,
  [7598] = 6794,
  [7599] = 6799,
  [7600] = 6799,
  [7601] = 6800,
  [7602] = 6795,
  [7603] = 6824,
  [7604] = 7265,
  [7605] = 7597,
  [7606] = 6794,
  [7607] = 7597,
  [7608] = 7597,
  [7609] = 7597,
  [7610] = 7597,
  [7611] = 7597,
  [7612] = 7597,
  [7613] = 7597,
  [7614] = 7597,
  [7615] = 7597,
  [7616] = 7597,
  [7617] = 7597,
  [7618] = 7597,
  [7619] = 7597,
  [7620] = 7597,
  [7621] = 7597,
  [7622] = 7597,
  [7623] = 7597,
  [7624] = 7597,
  [7625] = 7597,
  [7626] = 7597,
  [7627] = 7597,
  [7628] = 7597,
  [7629] = 7597,
  [7630] = 7597,
  [7631] = 7597,
  [7632] = 7597,
  [7633] = 7597,
  [7634] = 7597,
  [7635] = 7597,
  [7636] = 7597,
  [7637] = 7597,
  [7638] = 7597,
  [7639] = 7597,
  [7640] = 7597,
  [7641] = 7597,
  [7642] = 7597,
  [7643] = 7597,
  [7644] = 7597,
  [7645] = 7597,
  [7646] = 7597,
  [7647] = 7597,
  [7648] = 7597,
  [7649] = 7597,
  [7650] = 7597,
  [7651] = 7597,
  [7652] = 7597,
  [7653] = 7597,
  [7654] = 7597,
  [7655] = 7597,
  [7656] = 7597,
  [7657] = 7597,
  [7658] = 7597,
  [7659] = 7597,
  [7660] = 7597,
  [7661] = 7597,
  [7662] = 7597,
  [7663] = 7597,
  [7664] = 7597,
  [7665] = 7597,
  [7666] = 7597,
  [7667] = 7597,
  [7668] = 7597,
  [7669] = 7597,
  [7670] = 7670,
};

static inline bool sym_word_character_set_1(int32_t c) {
  return (c < '&'
    ? (c < ' '
      ? (c < '\t'
        ? c == 0
        : c <= '\r')
      : (c <= ' ' || (c < '$'
        ? c == '"'
        : c <= '$')))
    : (c <= ')' || (c < '['
      ? (c < '>'
        ? (c >= ';' && c <= '<')
        : c <= '>')
      : (c <= ']' || (c < '{'
        ? c == '`'
        : c <= '}')))));
}

static inline bool sym_word_character_set_2(int32_t c) {
  return (c < '&'
    ? (c < ' '
      ? (c < '\t'
        ? c == 0
        : c <= '\r')
      : (c <= ' ' || (c < '$'
        ? c == '"'
        : c <= '$')))
    : (c <= ')' || (c < '`'
      ? (c < '['
        ? (c >= ';' && c <= '>')
        : c <= ']')
      : (c <= '`' || (c >= '{' && c <= '}')))));
}

static inline bool sym_word_character_set_3(int32_t c) {
  return (c < '$'
    ? (c < ' '
      ? (c < '\t'
        ? c == 0
        : c <= '\r')
      : (c <= ' ' || c == '"'))
    : (c <= ')' || (c < '`'
      ? (c < '['
        ? (c >= ';' && c <= '>')
        : c <= ']')
      : (c <= '`' || (c >= '{' && c <= '}')))));
}

static inline bool sym_word_character_set_4(int32_t c) {
  return (c < '&'
    ? (c < ' '
      ? (c < '\t'
        ? c == 0
        : c <= '\r')
      : (c <= ' ' || (c < '$'
        ? c == '"'
        : c <= '$')))
    : (c <= ')' || (c < '['
      ? (c < '>'
        ? (c >= ';' && c <= '<')
        : c <= '>')
      : (c <= ']' || (c >= '`' && c <= '}')))));
}

static inline bool sym_word_character_set_5(int32_t c) {
  return (c < ';'
    ? (c < ' '
      ? (c < '\t'
        ? c == 0
        : c <= '\r')
      : (c <= ' ' || (c < '&'
        ? (c >= '"' && c <= '$')
        : c <= ')')))
    : (c <= '<' || (c < '`'
      ? (c < '['
        ? c == '>'
        : c <= ']')
      : (c <= '`' || (c >= '{' && c <= '}')))));
}

static inline bool sym_word_character_set_6(int32_t c) {
  return (c < '&'
    ? (c < ' '
      ? (c < '\t'
        ? c == 0
        : c <= '\r')
      : (c <= ' ' || (c >= '"' && c <= '$')))
    : (c <= ')' || (c < '['
      ? (c < '>'
        ? (c >= ';' && c <= '<')
        : c <= '>')
      : (c <= ']' || (c >= '`' && c <= '}')))));
}

static bool ts_lex(TSLexer *lexer, TSStateId state) {
  START_LEXER();
  eof = lexer->eof(lexer);
  switch (state) {
    case 0:
      if (eof) ADVANCE(547);
      if (lookahead == '!') ADVANCE(820);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(825);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '%') ADVANCE(647);
      if (lookahead == '&') ADVANCE(601);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(654);
      if (lookahead == ')') ADVANCE(655);
      if (lookahead == '*') ADVANCE(824);
      if (lookahead == '+') ADVANCE(770);
      if (lookahead == ',') ADVANCE(557);
      if (lookahead == '-') ADVANCE(768);
      if (lookahead == '.') ADVANCE(949);
      if (lookahead == '/') ADVANCE(642);
      if (lookahead == '0') ADVANCE(944);
      if (lookahead == ':') ADVANCE(763);
      if (lookahead == ';') ADVANCE(554);
      if (lookahead == '<') ADVANCE(607);
      if (lookahead == '=') ADVANCE(827);
      if (lookahead == '>') ADVANCE(616);
      if (lookahead == '?') ADVANCE(837);
      if (lookahead == '@') ADVANCE(822);
      if (lookahead == '[') ADVANCE(674);
      if (lookahead == '\\') ADVANCE(417);
      if (lookahead == ']') ADVANCE(675);
      if (lookahead == '^') ADVANCE(595);
      if (lookahead == '_') ADVANCE(946);
      if (lookahead == '`') ADVANCE(856);
      if (lookahead == 'e') ADVANCE(956);
      if (lookahead == 'i') ADVANCE(955);
      if (lookahead == '{') ADVANCE(666);
      if (lookahead == '|') ADVANCE(592);
      if (lookahead == '}') ADVANCE(778);
      if (lookahead == '~') ADVANCE(772);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(541)
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 1:
      if (lookahead == '\n') SKIP(239)
      END_STATE();
    case 2:
      if (lookahead == '\n') SKIP(255)
      END_STATE();
    case 3:
      if (lookahead == '\n') SKIP(256)
      END_STATE();
    case 4:
      if (lookahead == '\n') SKIP(257)
      END_STATE();
    case 5:
      if (lookahead == '\n') SKIP(6)
      END_STATE();
    case 6:
      if (lookahead == '\n') ADVANCE(688);
      if (lookahead == '!') ADVANCE(670);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(862);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '%') ADVANCE(649);
      if (lookahead == '&') ADVANCE(601);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(654);
      if (lookahead == ')') ADVANCE(655);
      if (lookahead == '*') ADVANCE(638);
      if (lookahead == '+') ADVANCE(771);
      if (lookahead == '-') ADVANCE(769);
      if (lookahead == '/') ADVANCE(644);
      if (lookahead == '0') ADVANCE(803);
      if (lookahead == ';') ADVANCE(555);
      if (lookahead == '<') ADVANCE(607);
      if (lookahead == '=') ADVANCE(560);
      if (lookahead == '>') ADVANCE(616);
      if (lookahead == '?') ADVANCE(761);
      if (lookahead == '[' ||
          lookahead == ']' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') ADVANCE(421);
      if (lookahead == '^') ADVANCE(596);
      if (lookahead == '`') ADVANCE(856);
      if (lookahead == '|') ADVANCE(592);
      if (lookahead == '~') ADVANCE(772);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(6)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(805);
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 7:
      if (lookahead == '\n') SKIP(241)
      END_STATE();
    case 8:
      if (lookahead == '\n') SKIP(259)
      END_STATE();
    case 9:
      if (lookahead == '\n') SKIP(260)
      END_STATE();
    case 10:
      if (lookahead == '\n') SKIP(242)
      END_STATE();
    case 11:
      if (lookahead == '\n') SKIP(244)
      END_STATE();
    case 12:
      if (lookahead == '\n') ADVANCE(689);
      if (lookahead == '!') ADVANCE(670);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '%') ADVANCE(649);
      if (lookahead == '&') ADVANCE(601);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(654);
      if (lookahead == ')') ADVANCE(655);
      if (lookahead == '*') ADVANCE(638);
      if (lookahead == '+') ADVANCE(630);
      if (lookahead == '-') ADVANCE(633);
      if (lookahead == '/') ADVANCE(644);
      if (lookahead == '0') ADVANCE(943);
      if (lookahead == ';') ADVANCE(555);
      if (lookahead == '<') ADVANCE(607);
      if (lookahead == '=') ADVANCE(560);
      if (lookahead == '>') ADVANCE(616);
      if (lookahead == '?') ADVANCE(761);
      if (lookahead == '@') ADVANCE(942);
      if (lookahead == '[' ||
          lookahead == ']' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') ADVANCE(355);
      if (lookahead == '^') ADVANCE(596);
      if (lookahead == '_') ADVANCE(947);
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == '|') ADVANCE(592);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(12)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(807);
      if (('A' <= lookahead && lookahead <= 'z')) ADVANCE(873);
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 13:
      if (lookahead == '\n') ADVANCE(880);
      END_STATE();
    case 14:
      if (lookahead == '\n') ADVANCE(940);
      END_STATE();
    case 15:
      if (lookahead == '\n') ADVANCE(940);
      if (lookahead == '\r') ADVANCE(14);
      END_STATE();
    case 16:
      if (lookahead == '\n') ADVANCE(940);
      if (lookahead == '\r') ADVANCE(14);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\f' < lookahead)) ADVANCE(958);
      END_STATE();
    case 17:
      if (lookahead == '\n') ADVANCE(918);
      END_STATE();
    case 18:
      if (lookahead == '\n') SKIP(112)
      END_STATE();
    case 19:
      if (lookahead == '\n') SKIP(263)
      END_STATE();
    case 20:
      if (lookahead == '\n') SKIP(258)
      END_STATE();
    case 21:
      if (lookahead == '\n') SKIP(177)
      END_STATE();
    case 22:
      if (lookahead == '\n') SKIP(180)
      END_STATE();
    case 23:
      if (lookahead == '\n') SKIP(183)
      END_STATE();
    case 24:
      if (lookahead == '\n') ADVANCE(882);
      END_STATE();
    case 25:
      if (lookahead == '\n') SKIP(186)
      END_STATE();
    case 26:
      if (lookahead == '\n') ADVANCE(884);
      END_STATE();
    case 27:
      if (lookahead == '\n') ADVANCE(888);
      END_STATE();
    case 28:
      if (lookahead == '\n') SKIP(189)
      END_STATE();
    case 29:
      if (lookahead == '\n') SKIP(192)
      END_STATE();
    case 30:
      if (lookahead == '\n') SKIP(195)
      END_STATE();
    case 31:
      if (lookahead == '\n') ADVANCE(898);
      END_STATE();
    case 32:
      if (lookahead == '\n') SKIP(198)
      END_STATE();
    case 33:
      if (lookahead == '\n') SKIP(299)
      END_STATE();
    case 34:
      if (lookahead == '\n') SKIP(295)
      END_STATE();
    case 35:
      if (lookahead == '\n') SKIP(204)
      END_STATE();
    case 36:
      if (lookahead == '\n') ADVANCE(902);
      END_STATE();
    case 37:
      if (lookahead == '\n') SKIP(262)
      END_STATE();
    case 38:
      if (lookahead == '\n') ADVANCE(904);
      END_STATE();
    case 39:
      if (lookahead == '\n') ADVANCE(909);
      END_STATE();
    case 40:
      if (lookahead == '\n') SKIP(213)
      END_STATE();
    case 41:
      if (lookahead == '\n') ADVANCE(920);
      END_STATE();
    case 42:
      if (lookahead == '\n') SKIP(308)
      END_STATE();
    case 43:
      if (lookahead == '\n') SKIP(302)
      END_STATE();
    case 44:
      if (lookahead == '\n') ADVANCE(922);
      END_STATE();
    case 45:
      if (lookahead == '\n') SKIP(298)
      END_STATE();
    case 46:
      if (lookahead == '\n') ADVANCE(913);
      END_STATE();
    case 47:
      if (lookahead == '\n') ADVANCE(930);
      END_STATE();
    case 48:
      if (lookahead == '\n') ADVANCE(923);
      END_STATE();
    case 49:
      if (lookahead == '\n') ADVANCE(929);
      END_STATE();
    case 50:
      if (lookahead == '\n') SKIP(306)
      END_STATE();
    case 51:
      if (lookahead == '\n') ADVANCE(927);
      END_STATE();
    case 52:
      if (lookahead == '\n') ADVANCE(932);
      END_STATE();
    case 53:
      if (lookahead == '\n') SKIP(272)
      END_STATE();
    case 54:
      if (lookahead == '\n') SKIP(273)
      END_STATE();
    case 55:
      if (lookahead == '\n') SKIP(261)
      END_STATE();
    case 56:
      if (lookahead == '\n') ADVANCE(919);
      END_STATE();
    case 57:
      if (lookahead == '\n') SKIP(278)
      END_STATE();
    case 58:
      if (lookahead == '\n') SKIP(235)
      END_STATE();
    case 59:
      if (lookahead == '\n') SKIP(276)
      END_STATE();
    case 60:
      if (lookahead == '\n') SKIP(281)
      END_STATE();
    case 61:
      if (lookahead == '\n') SKIP(277)
      END_STATE();
    case 62:
      if (lookahead == '\n') ADVANCE(881);
      END_STATE();
    case 63:
      if (lookahead == '\n') SKIP(64)
      END_STATE();
    case 64:
      if (lookahead == '\n') ADVANCE(690);
      if (lookahead == '!') ADVANCE(344);
      if (lookahead == '#') ADVANCE(862);
      if (lookahead == '%') ADVANCE(650);
      if (lookahead == '&') ADVANCE(600);
      if (lookahead == '*') ADVANCE(639);
      if (lookahead == '+') ADVANCE(631);
      if (lookahead == ',') ADVANCE(556);
      if (lookahead == '-') ADVANCE(635);
      if (lookahead == '/') ADVANCE(645);
      if (lookahead == ';') ADVANCE(553);
      if (lookahead == '<') ADVANCE(615);
      if (lookahead == '=') ADVANCE(561);
      if (lookahead == '>') ADVANCE(620);
      if (lookahead == '\\') SKIP(455)
      if (lookahead == '^') ADVANCE(597);
      if (lookahead == '|') ADVANCE(594);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(64)
      END_STATE();
    case 65:
      if (lookahead == '\n') ADVANCE(690);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(862);
      if (lookahead == '$') ADVANCE(785);
      if (lookahead == '&') ADVANCE(599);
      if (lookahead == '(') ADVANCE(653);
      if (lookahead == '+') ADVANCE(338);
      if (lookahead == ',') ADVANCE(556);
      if (lookahead == '-') ADVANCE(340);
      if (lookahead == '0') ADVANCE(809);
      if (lookahead == ';') ADVANCE(553);
      if (lookahead == '\\') SKIP(465)
      if (lookahead == '`') ADVANCE(855);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(65)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(811);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(656);
      END_STATE();
    case 66:
      if (lookahead == '\n') ADVANCE(690);
      if (lookahead == '#') ADVANCE(862);
      if (lookahead == '$') ADVANCE(783);
      if (lookahead == '&') ADVANCE(327);
      if (lookahead == '(') ADVANCE(654);
      if (lookahead == '-') ADVANCE(342);
      if (lookahead == '0') ADVANCE(808);
      if (lookahead == ':') ADVANCE(762);
      if (lookahead == '<') ADVANCE(611);
      if (lookahead == '>') ADVANCE(618);
      if (lookahead == '\\') SKIP(471)
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == '|') ADVANCE(354);
      if (lookahead == '}') ADVANCE(818);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(66)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(810);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(879);
      END_STATE();
    case 67:
      if (lookahead == '\n') ADVANCE(690);
      if (lookahead == '#') ADVANCE(862);
      if (lookahead == '&') ADVANCE(602);
      if (lookahead == ';') ADVANCE(554);
      if (lookahead == '<') ADVANCE(612);
      if (lookahead == '>') ADVANCE(618);
      if (lookahead == '[' ||
          lookahead == ']' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') SKIP(466)
      if (lookahead == '`') ADVANCE(348);
      if (lookahead == 'e') ADVANCE(352);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(67)
      END_STATE();
    case 68:
      if (lookahead == '\n') ADVANCE(690);
      if (lookahead == '#') ADVANCE(862);
      if (lookahead == '&') ADVANCE(602);
      if (lookahead == ';') ADVANCE(554);
      if (lookahead == '<') ADVANCE(613);
      if (lookahead == '>') ADVANCE(618);
      if (lookahead == '[' ||
          lookahead == ']' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') SKIP(504)
      if (lookahead == '`') ADVANCE(348);
      if (lookahead == 'e') ADVANCE(352);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(68)
      END_STATE();
    case 69:
      if (lookahead == '\n') ADVANCE(690);
      if (lookahead == '#') ADVANCE(862);
      if (lookahead == '&') ADVANCE(327);
      if (lookahead == '<') ADVANCE(611);
      if (lookahead == '>') ADVANCE(618);
      if (lookahead == '[' ||
          lookahead == ']' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') SKIP(475)
      if (lookahead == '`') ADVANCE(348);
      if (lookahead == '|') ADVANCE(354);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(69)
      END_STATE();
    case 70:
      if (lookahead == '\n') ADVANCE(690);
      if (lookahead == '#') ADVANCE(862);
      if (lookahead == '&') ADVANCE(599);
      if (lookahead == ';') ADVANCE(555);
      if (lookahead == '[' ||
          lookahead == ']' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') SKIP(480)
      if (lookahead == '`') ADVANCE(348);
      if (lookahead == 'i') ADVANCE(351);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(70)
      END_STATE();
    case 71:
      if (lookahead == '\n') ADVANCE(883);
      END_STATE();
    case 72:
      if (lookahead == '\n') SKIP(311)
      END_STATE();
    case 73:
      if (lookahead == '\n') SKIP(275)
      END_STATE();
    case 74:
      if (lookahead == '\n') SKIP(280)
      END_STATE();
    case 75:
      if (lookahead == '\n') ADVANCE(889);
      END_STATE();
    case 76:
      if (lookahead == '\n') ADVANCE(891);
      END_STATE();
    case 77:
      if (lookahead == '\n') ADVANCE(916);
      END_STATE();
    case 78:
      if (lookahead == '\n') SKIP(237)
      END_STATE();
    case 79:
      if (lookahead == '\n') SKIP(310)
      END_STATE();
    case 80:
      if (lookahead == '\n') SKIP(316)
      END_STATE();
    case 81:
      if (lookahead == '\n') SKIP(320)
      END_STATE();
    case 82:
      if (lookahead == '\n') ADVANCE(901);
      END_STATE();
    case 83:
      if (lookahead == '\n') ADVANCE(933);
      END_STATE();
    case 84:
      if (lookahead == '\n') ADVANCE(926);
      END_STATE();
    case 85:
      if (lookahead == '\n') ADVANCE(917);
      END_STATE();
    case 86:
      if (lookahead == '\n') ADVANCE(934);
      END_STATE();
    case 87:
      if (lookahead == '\n') SKIP(317)
      END_STATE();
    case 88:
      if (lookahead == '\n') SKIP(289)
      END_STATE();
    case 89:
      if (lookahead == '\n') SKIP(65)
      END_STATE();
    case 90:
      if (lookahead == '\n') SKIP(67)
      END_STATE();
    case 91:
      if (lookahead == '\n') SKIP(292)
      END_STATE();
    case 92:
      if (lookahead == '\n') SKIP(312)
      END_STATE();
    case 93:
      if (lookahead == '\n') SKIP(324)
      END_STATE();
    case 94:
      if (lookahead == '\n') SKIP(321)
      END_STATE();
    case 95:
      if (lookahead == '\n') SKIP(66)
      END_STATE();
    case 96:
      if (lookahead == '\n') SKIP(322)
      END_STATE();
    case 97:
      if (lookahead == '\n') SKIP(325)
      END_STATE();
    case 98:
      if (lookahead == '\n') SKIP(313)
      END_STATE();
    case 99:
      if (lookahead == '\n') ADVANCE(907);
      END_STATE();
    case 100:
      if (lookahead == '\n') SKIP(69)
      END_STATE();
    case 101:
      if (lookahead == '\n') SKIP(293)
      END_STATE();
    case 102:
      if (lookahead == '\n') ADVANCE(937);
      END_STATE();
    case 103:
      if (lookahead == '\n') ADVANCE(938);
      END_STATE();
    case 104:
      if (lookahead == '\n') ADVANCE(939);
      END_STATE();
    case 105:
      if (lookahead == '\n') SKIP(323)
      END_STATE();
    case 106:
      if (lookahead == '\n') SKIP(314)
      END_STATE();
    case 107:
      if (lookahead == '\n') SKIP(70)
      END_STATE();
    case 108:
      if (lookahead == '\n') SKIP(315)
      END_STATE();
    case 109:
      if (lookahead == '\n') SKIP(326)
      END_STATE();
    case 110:
      if (lookahead == '\n') SKIP(245)
      END_STATE();
    case 111:
      if (lookahead == '\n') SKIP(243)
      END_STATE();
    case 112:
      if (lookahead == '\n') ADVANCE(691);
      if (lookahead == '!') ADVANCE(951);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(862);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '%') ADVANCE(649);
      if (lookahead == '&') ADVANCE(601);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(654);
      if (lookahead == ')') ADVANCE(655);
      if (lookahead == '*') ADVANCE(638);
      if (lookahead == '+') ADVANCE(630);
      if (lookahead == '-') ADVANCE(633);
      if (lookahead == '/') ADVANCE(644);
      if (lookahead == '0') ADVANCE(803);
      if (lookahead == ';') ADVANCE(555);
      if (lookahead == '<') ADVANCE(607);
      if (lookahead == '=') ADVANCE(560);
      if (lookahead == '>') ADVANCE(616);
      if (lookahead == '?') ADVANCE(761);
      if (lookahead == '[' ||
          lookahead == ']' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') ADVANCE(427);
      if (lookahead == '^') ADVANCE(596);
      if (lookahead == '`') ADVANCE(856);
      if (lookahead == '|') ADVANCE(592);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(112)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(805);
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 113:
      if (lookahead == '\n') SKIP(174)
      END_STATE();
    case 114:
      if (lookahead == '\n') SKIP(264)
      END_STATE();
    case 115:
      if (lookahead == '\n') SKIP(215)
      END_STATE();
    case 116:
      if (lookahead == '\n') SKIP(217)
      END_STATE();
    case 117:
      if (lookahead == '\n') SKIP(219)
      END_STATE();
    case 118:
      if (lookahead == '\n') ADVANCE(886);
      END_STATE();
    case 119:
      if (lookahead == '\n') SKIP(221)
      END_STATE();
    case 120:
      if (lookahead == '\n') ADVANCE(890);
      END_STATE();
    case 121:
      if (lookahead == '\n') ADVANCE(894);
      END_STATE();
    case 122:
      if (lookahead == '\n') SKIP(225)
      END_STATE();
    case 123:
      if (lookahead == '\n') SKIP(231)
      END_STATE();
    case 124:
      if (lookahead == '\n') SKIP(223)
      END_STATE();
    case 125:
      if (lookahead == '\n') SKIP(227)
      END_STATE();
    case 126:
      if (lookahead == '\n') SKIP(300)
      END_STATE();
    case 127:
      if (lookahead == '\n') SKIP(232)
      END_STATE();
    case 128:
      if (lookahead == '\n') ADVANCE(908);
      END_STATE();
    case 129:
      if (lookahead == '\n') ADVANCE(912);
      END_STATE();
    case 130:
      if (lookahead == '\n') SKIP(236)
      END_STATE();
    case 131:
      if (lookahead == '\n') ADVANCE(921);
      END_STATE();
    case 132:
      if (lookahead == '\n') SKIP(309)
      END_STATE();
    case 133:
      if (lookahead == '\n') SKIP(303)
      END_STATE();
    case 134:
      if (lookahead == '\n') SKIP(296)
      END_STATE();
    case 135:
      if (lookahead == '\n') ADVANCE(915);
      END_STATE();
    case 136:
      if (lookahead == '\n') ADVANCE(931);
      END_STATE();
    case 137:
      if (lookahead == '\n') SKIP(307)
      END_STATE();
    case 138:
      if (lookahead == '\n') ADVANCE(928);
      END_STATE();
    case 139:
      if (lookahead == '\n') ADVANCE(885);
      END_STATE();
    case 140:
      if (lookahead == '\n') ADVANCE(887);
      END_STATE();
    case 141:
      if (lookahead == '\n') SKIP(274)
      END_STATE();
    case 142:
      if (lookahead == '\n') ADVANCE(895);
      END_STATE();
    case 143:
      if (lookahead == '\n') ADVANCE(899);
      END_STATE();
    case 144:
      if (lookahead == '\n') SKIP(238)
      END_STATE();
    case 145:
      if (lookahead == '\n') ADVANCE(903);
      END_STATE();
    case 146:
      if (lookahead == '\n') ADVANCE(935);
      END_STATE();
    case 147:
      if (lookahead == '\n') ADVANCE(936);
      END_STATE();
    case 148:
      if (lookahead == '\n') SKIP(68)
      END_STATE();
    case 149:
      if (lookahead == '\n') SKIP(246)
      END_STATE();
    case 150:
      if (lookahead == '\n') ADVANCE(692);
      if (lookahead == '!') ADVANCE(669);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(780);
      if (lookahead == '&') ADVANCE(602);
      if (lookahead == '*') ADVANCE(637);
      if (lookahead == '-') ADVANCE(632);
      if (lookahead == '0') ADVANCE(945);
      if (lookahead == ';') ADVANCE(554);
      if (lookahead == '<') ADVANCE(612);
      if (lookahead == '>') ADVANCE(618);
      if (lookahead == '?') ADVANCE(760);
      if (lookahead == '@') ADVANCE(941);
      if (lookahead == '\\') ADVANCE(373);
      if (lookahead == '_') ADVANCE(948);
      if (lookahead == 'e') ADVANCE(877);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(150)
      if (('1' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(879);
      END_STATE();
    case 151:
      if (lookahead == '\n') ADVANCE(892);
      END_STATE();
    case 152:
      if (lookahead == '\n') SKIP(229)
      END_STATE();
    case 153:
      if (lookahead == '\n') SKIP(233)
      END_STATE();
    case 154:
      if (lookahead == '\n') SKIP(234)
      END_STATE();
    case 155:
      if (lookahead == '\n') ADVANCE(911);
      END_STATE();
    case 156:
      if (lookahead == '\n') SKIP(304)
      END_STATE();
    case 157:
      if (lookahead == '\n') SKIP(301)
      END_STATE();
    case 158:
      if (lookahead == '\n') ADVANCE(893);
      END_STATE();
    case 159:
      if (lookahead == '\n') ADVANCE(905);
      END_STATE();
    case 160:
      if (lookahead == '\n') ADVANCE(896);
      END_STATE();
    case 161:
      if (lookahead == '\n') SKIP(297)
      END_STATE();
    case 162:
      if (lookahead == '\n') ADVANCE(897);
      END_STATE();
    case 163:
      if (lookahead == '\n') ADVANCE(693);
      if (lookahead == '!') ADVANCE(672);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(602);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(654);
      if (lookahead == '*') ADVANCE(640);
      if (lookahead == '-') ADVANCE(636);
      if (lookahead == '0') ADVANCE(943);
      if (lookahead == ';') ADVANCE(554);
      if (lookahead == '<') ADVANCE(609);
      if (lookahead == '=') ADVANCE(952);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '?') ADVANCE(761);
      if (lookahead == '@') ADVANCE(942);
      if (lookahead == '[' ||
          lookahead == ']' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') ADVANCE(357);
      if (lookahead == '_') ADVANCE(947);
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == 'e') ADVANCE(871);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(163)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(807);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(873);
      if (lookahead != 0 &&
          lookahead != ')') ADVANCE(958);
      END_STATE();
    case 164:
      if (lookahead == '\n') SKIP(201)
      END_STATE();
    case 165:
      if (lookahead == '\n') ADVANCE(900);
      END_STATE();
    case 166:
      if (lookahead == '\n') SKIP(207)
      END_STATE();
    case 167:
      if (lookahead == '\n') SKIP(210)
      END_STATE();
    case 168:
      if (lookahead == '\n') ADVANCE(906);
      END_STATE();
    case 169:
      if (lookahead == '\n') ADVANCE(910);
      END_STATE();
    case 170:
      if (lookahead == '\n') SKIP(305)
      END_STATE();
    case 171:
      if (lookahead == '\n') ADVANCE(924);
      END_STATE();
    case 172:
      if (lookahead == '\n') ADVANCE(914);
      END_STATE();
    case 173:
      if (lookahead == '\n') ADVANCE(925);
      END_STATE();
    case 174:
      if (lookahead == '\n') ADVANCE(694);
      if (lookahead == '!') ADVANCE(951);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(862);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '%') ADVANCE(649);
      if (lookahead == '&') ADVANCE(601);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(654);
      if (lookahead == ')') ADVANCE(655);
      if (lookahead == '*') ADVANCE(638);
      if (lookahead == '+') ADVANCE(630);
      if (lookahead == '-') ADVANCE(633);
      if (lookahead == '/') ADVANCE(644);
      if (lookahead == '0') ADVANCE(803);
      if (lookahead == ';') ADVANCE(555);
      if (lookahead == '<') ADVANCE(607);
      if (lookahead == '=') ADVANCE(560);
      if (lookahead == '>') ADVANCE(616);
      if (lookahead == '?') ADVANCE(761);
      if (lookahead == '[' ||
          lookahead == ']' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') ADVANCE(485);
      if (lookahead == '^') ADVANCE(596);
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == '|') ADVANCE(592);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(174)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(805);
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 175:
      if (lookahead == '\n') ADVANCE(695);
      if (lookahead == '!') ADVANCE(669);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(780);
      if (lookahead == '&') ADVANCE(602);
      if (lookahead == '*') ADVANCE(637);
      if (lookahead == '-') ADVANCE(632);
      if (lookahead == '0') ADVANCE(945);
      if (lookahead == ';') ADVANCE(554);
      if (lookahead == '<') ADVANCE(612);
      if (lookahead == '>') ADVANCE(618);
      if (lookahead == '?') ADVANCE(760);
      if (lookahead == '@') ADVANCE(941);
      if (lookahead == '\\') ADVANCE(374);
      if (lookahead == '_') ADVANCE(948);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(175)
      if (('1' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(879);
      END_STATE();
    case 176:
      if (lookahead == '\n') ADVANCE(696);
      if (lookahead == '!') ADVANCE(672);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(602);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(654);
      if (lookahead == '*') ADVANCE(640);
      if (lookahead == '-') ADVANCE(636);
      if (lookahead == '0') ADVANCE(943);
      if (lookahead == ';') ADVANCE(554);
      if (lookahead == '<') ADVANCE(609);
      if (lookahead == '=') ADVANCE(952);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '?') ADVANCE(761);
      if (lookahead == '@') ADVANCE(942);
      if (lookahead == '[' ||
          lookahead == ']' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') ADVANCE(358);
      if (lookahead == '_') ADVANCE(947);
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(176)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(807);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(873);
      if (lookahead != 0 &&
          lookahead != ')') ADVANCE(958);
      END_STATE();
    case 177:
      if (lookahead == '\n') ADVANCE(697);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(862);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(602);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(654);
      if (lookahead == '-') ADVANCE(950);
      if (lookahead == '0') ADVANCE(803);
      if (lookahead == ';') ADVANCE(554);
      if (lookahead == '<') ADVANCE(609);
      if (lookahead == '=') ADVANCE(952);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '[' ||
          lookahead == ']' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') ADVANCE(430);
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == 'e') ADVANCE(956);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(177)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(805);
      if (lookahead != 0 &&
          lookahead != ')') ADVANCE(958);
      END_STATE();
    case 178:
      if (lookahead == '\n') ADVANCE(698);
      if (lookahead == '!') ADVANCE(669);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(780);
      if (lookahead == '&') ADVANCE(602);
      if (lookahead == '*') ADVANCE(637);
      if (lookahead == '-') ADVANCE(632);
      if (lookahead == '0') ADVANCE(945);
      if (lookahead == ';') ADVANCE(554);
      if (lookahead == '<') ADVANCE(613);
      if (lookahead == '>') ADVANCE(618);
      if (lookahead == '?') ADVANCE(760);
      if (lookahead == '@') ADVANCE(941);
      if (lookahead == '\\') ADVANCE(416);
      if (lookahead == '_') ADVANCE(948);
      if (lookahead == 'e') ADVANCE(877);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(178)
      if (('1' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(879);
      END_STATE();
    case 179:
      if (lookahead == '\n') ADVANCE(699);
      if (lookahead == '!') ADVANCE(672);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(602);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(332);
      if (lookahead == '*') ADVANCE(640);
      if (lookahead == '-') ADVANCE(636);
      if (lookahead == '0') ADVANCE(943);
      if (lookahead == ';') ADVANCE(554);
      if (lookahead == '<') ADVANCE(609);
      if (lookahead == '=') ADVANCE(952);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '?') ADVANCE(761);
      if (lookahead == '@') ADVANCE(942);
      if (lookahead == '[' ||
          lookahead == ']' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') ADVANCE(389);
      if (lookahead == '_') ADVANCE(947);
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == 'e') ADVANCE(871);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(179)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(807);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(873);
      if (lookahead != 0 &&
          lookahead != ')') ADVANCE(958);
      END_STATE();
    case 180:
      if (lookahead == '\n') ADVANCE(700);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(862);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(602);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(654);
      if (lookahead == '-') ADVANCE(950);
      if (lookahead == '0') ADVANCE(803);
      if (lookahead == ';') ADVANCE(554);
      if (lookahead == '<') ADVANCE(609);
      if (lookahead == '=') ADVANCE(952);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '[' ||
          lookahead == ']' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') ADVANCE(431);
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(180)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(805);
      if (lookahead != 0 &&
          lookahead != ')') ADVANCE(958);
      END_STATE();
    case 181:
      if (lookahead == '\n') ADVANCE(701);
      if (lookahead == '!') ADVANCE(669);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(780);
      if (lookahead == '&') ADVANCE(602);
      if (lookahead == '*') ADVANCE(637);
      if (lookahead == '-') ADVANCE(632);
      if (lookahead == '0') ADVANCE(945);
      if (lookahead == ';') ADVANCE(554);
      if (lookahead == '<') ADVANCE(613);
      if (lookahead == '>') ADVANCE(618);
      if (lookahead == '?') ADVANCE(760);
      if (lookahead == '@') ADVANCE(941);
      if (lookahead == '\\') ADVANCE(398);
      if (lookahead == '_') ADVANCE(948);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(181)
      if (('1' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(879);
      END_STATE();
    case 182:
      if (lookahead == '\n') ADVANCE(702);
      if (lookahead == '!') ADVANCE(672);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(602);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(654);
      if (lookahead == ')') ADVANCE(655);
      if (lookahead == '*') ADVANCE(640);
      if (lookahead == '-') ADVANCE(636);
      if (lookahead == '0') ADVANCE(943);
      if (lookahead == ';') ADVANCE(555);
      if (lookahead == '<') ADVANCE(609);
      if (lookahead == '=') ADVANCE(952);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '?') ADVANCE(761);
      if (lookahead == '@') ADVANCE(942);
      if (lookahead == '[' ||
          lookahead == ']' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') ADVANCE(359);
      if (lookahead == '_') ADVANCE(947);
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(182)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(807);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(873);
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 183:
      if (lookahead == '\n') ADVANCE(703);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(862);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(602);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(654);
      if (lookahead == ')') ADVANCE(655);
      if (lookahead == '-') ADVANCE(950);
      if (lookahead == '0') ADVANCE(803);
      if (lookahead == ';') ADVANCE(555);
      if (lookahead == '<') ADVANCE(609);
      if (lookahead == '=') ADVANCE(952);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '[' ||
          lookahead == ']' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') ADVANCE(432);
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(183)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(805);
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 184:
      if (lookahead == '\n') ADVANCE(704);
      if (lookahead == '!') ADVANCE(669);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(780);
      if (lookahead == '&') ADVANCE(602);
      if (lookahead == '*') ADVANCE(637);
      if (lookahead == '-') ADVANCE(632);
      if (lookahead == '0') ADVANCE(945);
      if (lookahead == ';') ADVANCE(555);
      if (lookahead == '<') ADVANCE(612);
      if (lookahead == '>') ADVANCE(618);
      if (lookahead == '?') ADVANCE(760);
      if (lookahead == '@') ADVANCE(941);
      if (lookahead == '\\') ADVANCE(375);
      if (lookahead == '_') ADVANCE(948);
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(184)
      if (('1' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(879);
      END_STATE();
    case 185:
      if (lookahead == '\n') ADVANCE(705);
      if (lookahead == '!') ADVANCE(672);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(602);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(654);
      if (lookahead == '*') ADVANCE(640);
      if (lookahead == '-') ADVANCE(636);
      if (lookahead == '0') ADVANCE(943);
      if (lookahead == ';') ADVANCE(555);
      if (lookahead == '<') ADVANCE(609);
      if (lookahead == '=') ADVANCE(952);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '?') ADVANCE(761);
      if (lookahead == '@') ADVANCE(942);
      if (lookahead == '[' ||
          lookahead == ']' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') ADVANCE(390);
      if (lookahead == '_') ADVANCE(947);
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(185)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(807);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(873);
      if (lookahead != 0 &&
          lookahead != ')') ADVANCE(958);
      END_STATE();
    case 186:
      if (lookahead == '\n') ADVANCE(706);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(862);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(602);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(332);
      if (lookahead == '-') ADVANCE(950);
      if (lookahead == '0') ADVANCE(804);
      if (lookahead == ';') ADVANCE(554);
      if (lookahead == '<') ADVANCE(610);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '[' ||
          lookahead == ']' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') ADVANCE(433);
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == 'e') ADVANCE(866);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(186)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(806);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(868);
      if (lookahead != 0 &&
          lookahead != ')') ADVANCE(958);
      END_STATE();
    case 187:
      if (lookahead == '\n') ADVANCE(707);
      if (lookahead == '!') ADVANCE(669);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(780);
      if (lookahead == '&') ADVANCE(602);
      if (lookahead == ')') ADVANCE(655);
      if (lookahead == '*') ADVANCE(637);
      if (lookahead == '-') ADVANCE(632);
      if (lookahead == '0') ADVANCE(945);
      if (lookahead == ';') ADVANCE(555);
      if (lookahead == '<') ADVANCE(612);
      if (lookahead == '>') ADVANCE(618);
      if (lookahead == '?') ADVANCE(760);
      if (lookahead == '@') ADVANCE(941);
      if (lookahead == '\\') ADVANCE(376);
      if (lookahead == '_') ADVANCE(948);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(187)
      if (('1' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(879);
      END_STATE();
    case 188:
      if (lookahead == '\n') ADVANCE(708);
      if (lookahead == '!') ADVANCE(672);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(602);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(332);
      if (lookahead == '*') ADVANCE(640);
      if (lookahead == '-') ADVANCE(636);
      if (lookahead == '0') ADVANCE(943);
      if (lookahead == ';') ADVANCE(554);
      if (lookahead == '<') ADVANCE(609);
      if (lookahead == '=') ADVANCE(952);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '?') ADVANCE(761);
      if (lookahead == '@') ADVANCE(942);
      if (lookahead == '[' ||
          lookahead == ']' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') ADVANCE(404);
      if (lookahead == '_') ADVANCE(947);
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(188)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(807);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(873);
      if (lookahead != 0 &&
          lookahead != ')') ADVANCE(958);
      END_STATE();
    case 189:
      if (lookahead == '\n') ADVANCE(709);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(862);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(602);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(332);
      if (lookahead == '-') ADVANCE(950);
      if (lookahead == '0') ADVANCE(804);
      if (lookahead == ';') ADVANCE(554);
      if (lookahead == '<') ADVANCE(610);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '[' ||
          lookahead == ']' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') ADVANCE(434);
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(189)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(806);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(868);
      if (lookahead != 0 &&
          lookahead != ')') ADVANCE(958);
      END_STATE();
    case 190:
      if (lookahead == '\n') ADVANCE(710);
      if (lookahead == '!') ADVANCE(669);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(780);
      if (lookahead == '&') ADVANCE(602);
      if (lookahead == '*') ADVANCE(637);
      if (lookahead == '-') ADVANCE(632);
      if (lookahead == '0') ADVANCE(945);
      if (lookahead == ';') ADVANCE(555);
      if (lookahead == '<') ADVANCE(612);
      if (lookahead == '>') ADVANCE(618);
      if (lookahead == '?') ADVANCE(760);
      if (lookahead == '@') ADVANCE(941);
      if (lookahead == '\\') ADVANCE(406);
      if (lookahead == '_') ADVANCE(948);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(190)
      if (('1' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(879);
      END_STATE();
    case 191:
      if (lookahead == '\n') ADVANCE(711);
      if (lookahead == '!') ADVANCE(672);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(602);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(332);
      if (lookahead == ')') ADVANCE(655);
      if (lookahead == '*') ADVANCE(640);
      if (lookahead == '-') ADVANCE(636);
      if (lookahead == '0') ADVANCE(943);
      if (lookahead == ';') ADVANCE(555);
      if (lookahead == '<') ADVANCE(609);
      if (lookahead == '=') ADVANCE(952);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '?') ADVANCE(761);
      if (lookahead == '@') ADVANCE(942);
      if (lookahead == '[' ||
          lookahead == ']' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') ADVANCE(391);
      if (lookahead == '_') ADVANCE(947);
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(191)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(807);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(873);
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 192:
      if (lookahead == '\n') ADVANCE(712);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(862);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(602);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(332);
      if (lookahead == ')') ADVANCE(655);
      if (lookahead == '-') ADVANCE(950);
      if (lookahead == '0') ADVANCE(804);
      if (lookahead == ';') ADVANCE(555);
      if (lookahead == '<') ADVANCE(610);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '[' ||
          lookahead == ']' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') ADVANCE(435);
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(192)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(806);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(868);
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 193:
      if (lookahead == '\n') ADVANCE(713);
      if (lookahead == '!') ADVANCE(669);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(780);
      if (lookahead == '&') ADVANCE(602);
      if (lookahead == '*') ADVANCE(637);
      if (lookahead == '-') ADVANCE(632);
      if (lookahead == '0') ADVANCE(945);
      if (lookahead == ';') ADVANCE(555);
      if (lookahead == '<') ADVANCE(613);
      if (lookahead == '>') ADVANCE(618);
      if (lookahead == '?') ADVANCE(760);
      if (lookahead == '@') ADVANCE(941);
      if (lookahead == '\\') ADVANCE(399);
      if (lookahead == '_') ADVANCE(948);
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(193)
      if (('1' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(879);
      END_STATE();
    case 194:
      if (lookahead == '\n') ADVANCE(714);
      if (lookahead == '!') ADVANCE(672);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(602);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(332);
      if (lookahead == '*') ADVANCE(640);
      if (lookahead == '-') ADVANCE(636);
      if (lookahead == '0') ADVANCE(943);
      if (lookahead == ';') ADVANCE(555);
      if (lookahead == '<') ADVANCE(609);
      if (lookahead == '=') ADVANCE(952);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '?') ADVANCE(761);
      if (lookahead == '@') ADVANCE(942);
      if (lookahead == '[' ||
          lookahead == ']' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') ADVANCE(408);
      if (lookahead == '_') ADVANCE(947);
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(194)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(807);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(873);
      if (lookahead != 0 &&
          lookahead != ')') ADVANCE(958);
      END_STATE();
    case 195:
      if (lookahead == '\n') ADVANCE(715);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(862);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(602);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(332);
      if (lookahead == '-') ADVANCE(950);
      if (lookahead == '0') ADVANCE(803);
      if (lookahead == ';') ADVANCE(554);
      if (lookahead == '<') ADVANCE(609);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '[' ||
          lookahead == ']' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') ADVANCE(436);
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == 'e') ADVANCE(956);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(195)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(805);
      if (lookahead != 0 &&
          lookahead != ')') ADVANCE(958);
      END_STATE();
    case 196:
      if (lookahead == '\n') ADVANCE(716);
      if (lookahead == '!') ADVANCE(669);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(780);
      if (lookahead == '&') ADVANCE(602);
      if (lookahead == '*') ADVANCE(637);
      if (lookahead == '-') ADVANCE(632);
      if (lookahead == '0') ADVANCE(945);
      if (lookahead == ';') ADVANCE(555);
      if (lookahead == '<') ADVANCE(613);
      if (lookahead == '>') ADVANCE(618);
      if (lookahead == '?') ADVANCE(760);
      if (lookahead == '@') ADVANCE(941);
      if (lookahead == '\\') ADVANCE(409);
      if (lookahead == '_') ADVANCE(948);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(196)
      if (('1' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(879);
      END_STATE();
    case 197:
      if (lookahead == '\n') ADVANCE(717);
      if (lookahead == '!') ADVANCE(672);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(602);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(332);
      if (lookahead == '*') ADVANCE(640);
      if (lookahead == '-') ADVANCE(636);
      if (lookahead == '0') ADVANCE(943);
      if (lookahead == ';') ADVANCE(554);
      if (lookahead == '<') ADVANCE(609);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '?') ADVANCE(761);
      if (lookahead == '@') ADVANCE(942);
      if (lookahead == '[' ||
          lookahead == ']' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') ADVANCE(360);
      if (lookahead == '_') ADVANCE(947);
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == 'e') ADVANCE(871);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(197)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(807);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(873);
      if (lookahead != 0 &&
          lookahead != ')') ADVANCE(958);
      END_STATE();
    case 198:
      if (lookahead == '\n') ADVANCE(718);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(862);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(602);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(332);
      if (lookahead == '-') ADVANCE(950);
      if (lookahead == '0') ADVANCE(803);
      if (lookahead == ';') ADVANCE(554);
      if (lookahead == '<') ADVANCE(609);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '[' ||
          lookahead == ']' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') ADVANCE(437);
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(198)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(805);
      if (lookahead != 0 &&
          lookahead != ')') ADVANCE(958);
      END_STATE();
    case 199:
      if (lookahead == '\n') ADVANCE(719);
      if (lookahead == '!') ADVANCE(669);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(780);
      if (lookahead == '&') ADVANCE(602);
      if (lookahead == ')') ADVANCE(655);
      if (lookahead == '*') ADVANCE(637);
      if (lookahead == '-') ADVANCE(632);
      if (lookahead == '0') ADVANCE(945);
      if (lookahead == ';') ADVANCE(555);
      if (lookahead == '<') ADVANCE(613);
      if (lookahead == '>') ADVANCE(618);
      if (lookahead == '?') ADVANCE(760);
      if (lookahead == '@') ADVANCE(941);
      if (lookahead == '\\') ADVANCE(400);
      if (lookahead == '_') ADVANCE(948);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(199)
      if (('1' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(879);
      END_STATE();
    case 200:
      if (lookahead == '\n') ADVANCE(720);
      if (lookahead == '!') ADVANCE(672);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(602);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(332);
      if (lookahead == '*') ADVANCE(640);
      if (lookahead == '-') ADVANCE(636);
      if (lookahead == '0') ADVANCE(943);
      if (lookahead == ';') ADVANCE(554);
      if (lookahead == '<') ADVANCE(610);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '?') ADVANCE(761);
      if (lookahead == '@') ADVANCE(942);
      if (lookahead == '[' ||
          lookahead == ']' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') ADVANCE(410);
      if (lookahead == '_') ADVANCE(947);
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == 'e') ADVANCE(871);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(200)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(807);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(873);
      if (lookahead != 0 &&
          lookahead != ')') ADVANCE(958);
      END_STATE();
    case 201:
      if (lookahead == '\n') ADVANCE(721);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(862);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(602);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(332);
      if (lookahead == '-') ADVANCE(950);
      if (lookahead == '0') ADVANCE(803);
      if (lookahead == ';') ADVANCE(554);
      if (lookahead == '<') ADVANCE(610);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '[' ||
          lookahead == ']' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') ADVANCE(512);
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == 'e') ADVANCE(956);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(201)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(805);
      if (lookahead != 0 &&
          lookahead != ')') ADVANCE(958);
      END_STATE();
    case 202:
      if (lookahead == '\n') ADVANCE(722);
      if (lookahead == '!') ADVANCE(669);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(780);
      if (lookahead == '&') ADVANCE(327);
      if (lookahead == '*') ADVANCE(637);
      if (lookahead == '-') ADVANCE(632);
      if (lookahead == '0') ADVANCE(945);
      if (lookahead == '<') ADVANCE(612);
      if (lookahead == '>') ADVANCE(618);
      if (lookahead == '?') ADVANCE(760);
      if (lookahead == '@') ADVANCE(941);
      if (lookahead == '\\') ADVANCE(378);
      if (lookahead == '_') ADVANCE(948);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(202)
      if (('1' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(879);
      END_STATE();
    case 203:
      if (lookahead == '\n') ADVANCE(723);
      if (lookahead == '!') ADVANCE(672);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(602);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(332);
      if (lookahead == '*') ADVANCE(640);
      if (lookahead == '-') ADVANCE(636);
      if (lookahead == '0') ADVANCE(943);
      if (lookahead == ';') ADVANCE(554);
      if (lookahead == '<') ADVANCE(609);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '?') ADVANCE(761);
      if (lookahead == '@') ADVANCE(942);
      if (lookahead == '[' ||
          lookahead == ']' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') ADVANCE(361);
      if (lookahead == '_') ADVANCE(947);
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(203)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(807);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(873);
      if (lookahead != 0 &&
          lookahead != ')') ADVANCE(958);
      END_STATE();
    case 204:
      if (lookahead == '\n') ADVANCE(724);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(862);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(602);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(332);
      if (lookahead == ')') ADVANCE(655);
      if (lookahead == '-') ADVANCE(950);
      if (lookahead == '0') ADVANCE(803);
      if (lookahead == ';') ADVANCE(555);
      if (lookahead == '<') ADVANCE(609);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '[' ||
          lookahead == ']' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') ADVANCE(440);
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(204)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(805);
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 205:
      if (lookahead == '\n') ADVANCE(725);
      if (lookahead == '!') ADVANCE(669);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(780);
      if (lookahead == '&') ADVANCE(327);
      if (lookahead == '*') ADVANCE(637);
      if (lookahead == '-') ADVANCE(632);
      if (lookahead == '0') ADVANCE(945);
      if (lookahead == '<') ADVANCE(613);
      if (lookahead == '>') ADVANCE(618);
      if (lookahead == '?') ADVANCE(760);
      if (lookahead == '@') ADVANCE(941);
      if (lookahead == '\\') ADVANCE(401);
      if (lookahead == '_') ADVANCE(948);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(205)
      if (('1' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(879);
      END_STATE();
    case 206:
      if (lookahead == '\n') ADVANCE(726);
      if (lookahead == '!') ADVANCE(672);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(602);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(332);
      if (lookahead == ')') ADVANCE(655);
      if (lookahead == '*') ADVANCE(640);
      if (lookahead == '-') ADVANCE(636);
      if (lookahead == '0') ADVANCE(943);
      if (lookahead == ';') ADVANCE(555);
      if (lookahead == '<') ADVANCE(609);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '?') ADVANCE(761);
      if (lookahead == '@') ADVANCE(942);
      if (lookahead == '[' ||
          lookahead == ']' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') ADVANCE(362);
      if (lookahead == '_') ADVANCE(947);
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(206)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(807);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(873);
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 207:
      if (lookahead == '\n') ADVANCE(727);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(862);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(602);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(332);
      if (lookahead == '-') ADVANCE(950);
      if (lookahead == '0') ADVANCE(803);
      if (lookahead == ';') ADVANCE(554);
      if (lookahead == '<') ADVANCE(610);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '[' ||
          lookahead == ']' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') ADVANCE(513);
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(207)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(805);
      if (lookahead != 0 &&
          lookahead != ')') ADVANCE(958);
      END_STATE();
    case 208:
      if (lookahead == '\n') ADVANCE(728);
      if (lookahead == '!') ADVANCE(669);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(780);
      if (lookahead == '&') ADVANCE(327);
      if (lookahead == '*') ADVANCE(637);
      if (lookahead == '-') ADVANCE(632);
      if (lookahead == '0') ADVANCE(945);
      if (lookahead == '<') ADVANCE(611);
      if (lookahead == '>') ADVANCE(618);
      if (lookahead == '?') ADVANCE(760);
      if (lookahead == '@') ADVANCE(941);
      if (lookahead == '\\') ADVANCE(407);
      if (lookahead == '_') ADVANCE(948);
      if (lookahead == '|') ADVANCE(354);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(208)
      if (('1' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(879);
      END_STATE();
    case 209:
      if (lookahead == '\n') ADVANCE(729);
      if (lookahead == '!') ADVANCE(672);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(602);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(332);
      if (lookahead == '*') ADVANCE(640);
      if (lookahead == '-') ADVANCE(636);
      if (lookahead == '0') ADVANCE(943);
      if (lookahead == ';') ADVANCE(554);
      if (lookahead == '<') ADVANCE(610);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '?') ADVANCE(761);
      if (lookahead == '@') ADVANCE(942);
      if (lookahead == '[' ||
          lookahead == ']' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') ADVANCE(411);
      if (lookahead == '_') ADVANCE(947);
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(209)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(807);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(873);
      if (lookahead != 0 &&
          lookahead != ')') ADVANCE(958);
      END_STATE();
    case 210:
      if (lookahead == '\n') ADVANCE(730);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(862);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(602);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(332);
      if (lookahead == ')') ADVANCE(655);
      if (lookahead == '-') ADVANCE(950);
      if (lookahead == '0') ADVANCE(803);
      if (lookahead == ';') ADVANCE(555);
      if (lookahead == '<') ADVANCE(610);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '[' ||
          lookahead == ']' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') ADVANCE(514);
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(210)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(805);
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 211:
      if (lookahead == '\n') ADVANCE(731);
      if (lookahead == '!') ADVANCE(669);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(780);
      if (lookahead == '&') ADVANCE(599);
      if (lookahead == '*') ADVANCE(637);
      if (lookahead == '-') ADVANCE(632);
      if (lookahead == '0') ADVANCE(945);
      if (lookahead == ';') ADVANCE(555);
      if (lookahead == '?') ADVANCE(760);
      if (lookahead == '@') ADVANCE(941);
      if (lookahead == '\\') ADVANCE(383);
      if (lookahead == '_') ADVANCE(948);
      if (lookahead == 'i') ADVANCE(876);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(211)
      if (('1' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(879);
      END_STATE();
    case 212:
      if (lookahead == '\n') ADVANCE(732);
      if (lookahead == '!') ADVANCE(672);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(602);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(332);
      if (lookahead == '*') ADVANCE(640);
      if (lookahead == '-') ADVANCE(636);
      if (lookahead == '0') ADVANCE(943);
      if (lookahead == ';') ADVANCE(555);
      if (lookahead == '<') ADVANCE(609);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '?') ADVANCE(761);
      if (lookahead == '@') ADVANCE(942);
      if (lookahead == '[' ||
          lookahead == ']' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') ADVANCE(392);
      if (lookahead == '_') ADVANCE(947);
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(212)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(807);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(873);
      if (lookahead != 0 &&
          lookahead != ')') ADVANCE(958);
      END_STATE();
    case 213:
      if (lookahead == '\n') ADVANCE(733);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(862);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(327);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(332);
      if (lookahead == '-') ADVANCE(950);
      if (lookahead == '0') ADVANCE(803);
      if (lookahead == '<') ADVANCE(608);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '[' ||
          lookahead == ']' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') ADVANCE(442);
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(213)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(805);
      if (lookahead != 0 &&
          lookahead != ')' &&
          lookahead != ';') ADVANCE(958);
      END_STATE();
    case 214:
      if (lookahead == '\n') ADVANCE(734);
      if (lookahead == '!') ADVANCE(672);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(327);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(654);
      if (lookahead == '*') ADVANCE(640);
      if (lookahead == '-') ADVANCE(636);
      if (lookahead == '0') ADVANCE(943);
      if (lookahead == '<') ADVANCE(609);
      if (lookahead == '=') ADVANCE(952);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '?') ADVANCE(761);
      if (lookahead == '@') ADVANCE(942);
      if (lookahead == '[' ||
          lookahead == ']' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') ADVANCE(363);
      if (lookahead == '_') ADVANCE(947);
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(214)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(807);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(873);
      if (lookahead != 0 &&
          lookahead != ')' &&
          lookahead != ';') ADVANCE(958);
      END_STATE();
    case 215:
      if (lookahead == '\n') ADVANCE(735);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(862);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(602);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(654);
      if (lookahead == '-') ADVANCE(950);
      if (lookahead == '0') ADVANCE(803);
      if (lookahead == ';') ADVANCE(554);
      if (lookahead == '<') ADVANCE(609);
      if (lookahead == '=') ADVANCE(952);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '[' ||
          lookahead == ']' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') ADVANCE(487);
      if (lookahead == '`') ADVANCE(856);
      if (lookahead == 'e') ADVANCE(956);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(215)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(805);
      if (lookahead != 0 &&
          lookahead != ')') ADVANCE(958);
      END_STATE();
    case 216:
      if (lookahead == '\n') ADVANCE(736);
      if (lookahead == '!') ADVANCE(672);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(602);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(332);
      if (lookahead == ')') ADVANCE(655);
      if (lookahead == '*') ADVANCE(640);
      if (lookahead == '-') ADVANCE(636);
      if (lookahead == '0') ADVANCE(943);
      if (lookahead == ';') ADVANCE(555);
      if (lookahead == '<') ADVANCE(610);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '?') ADVANCE(761);
      if (lookahead == '@') ADVANCE(942);
      if (lookahead == '[' ||
          lookahead == ']' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') ADVANCE(412);
      if (lookahead == '_') ADVANCE(947);
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(216)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(807);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(873);
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 217:
      if (lookahead == '\n') ADVANCE(737);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(862);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(602);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(654);
      if (lookahead == '-') ADVANCE(950);
      if (lookahead == '0') ADVANCE(803);
      if (lookahead == ';') ADVANCE(554);
      if (lookahead == '<') ADVANCE(609);
      if (lookahead == '=') ADVANCE(952);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '[' ||
          lookahead == ']' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') ADVANCE(488);
      if (lookahead == '`') ADVANCE(856);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(217)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(805);
      if (lookahead != 0 &&
          lookahead != ')') ADVANCE(958);
      END_STATE();
    case 218:
      if (lookahead == '\n') ADVANCE(738);
      if (lookahead == '!') ADVANCE(672);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(602);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(332);
      if (lookahead == '*') ADVANCE(640);
      if (lookahead == '-') ADVANCE(636);
      if (lookahead == '0') ADVANCE(943);
      if (lookahead == ';') ADVANCE(555);
      if (lookahead == '<') ADVANCE(610);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '?') ADVANCE(761);
      if (lookahead == '@') ADVANCE(942);
      if (lookahead == '[' ||
          lookahead == ']' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') ADVANCE(405);
      if (lookahead == '_') ADVANCE(947);
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(218)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(807);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(873);
      if (lookahead != 0 &&
          lookahead != ')') ADVANCE(958);
      END_STATE();
    case 219:
      if (lookahead == '\n') ADVANCE(739);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(862);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(602);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(654);
      if (lookahead == ')') ADVANCE(655);
      if (lookahead == '-') ADVANCE(950);
      if (lookahead == '0') ADVANCE(803);
      if (lookahead == ';') ADVANCE(555);
      if (lookahead == '<') ADVANCE(609);
      if (lookahead == '=') ADVANCE(952);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '[' ||
          lookahead == ']' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') ADVANCE(489);
      if (lookahead == '`') ADVANCE(856);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(219)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(805);
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 220:
      if (lookahead == '\n') ADVANCE(740);
      if (lookahead == '!') ADVANCE(672);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(327);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(332);
      if (lookahead == '*') ADVANCE(640);
      if (lookahead == '-') ADVANCE(636);
      if (lookahead == '0') ADVANCE(943);
      if (lookahead == '<') ADVANCE(609);
      if (lookahead == '=') ADVANCE(952);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '?') ADVANCE(761);
      if (lookahead == '@') ADVANCE(942);
      if (lookahead == '[' ||
          lookahead == ']' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') ADVANCE(393);
      if (lookahead == '_') ADVANCE(947);
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(220)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(807);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(873);
      if (lookahead != 0 &&
          lookahead != ')' &&
          lookahead != ';') ADVANCE(958);
      END_STATE();
    case 221:
      if (lookahead == '\n') ADVANCE(741);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(862);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(602);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(332);
      if (lookahead == '-') ADVANCE(950);
      if (lookahead == '0') ADVANCE(804);
      if (lookahead == ';') ADVANCE(554);
      if (lookahead == '<') ADVANCE(610);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '[' ||
          lookahead == ']' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') ADVANCE(490);
      if (lookahead == '`') ADVANCE(856);
      if (lookahead == 'e') ADVANCE(866);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(221)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(806);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(868);
      if (lookahead != 0 &&
          lookahead != ')') ADVANCE(958);
      END_STATE();
    case 222:
      if (lookahead == '\n') ADVANCE(742);
      if (lookahead == '!') ADVANCE(672);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(327);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(332);
      if (lookahead == '*') ADVANCE(640);
      if (lookahead == '-') ADVANCE(636);
      if (lookahead == '0') ADVANCE(943);
      if (lookahead == '<') ADVANCE(609);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '?') ADVANCE(761);
      if (lookahead == '@') ADVANCE(942);
      if (lookahead == '[' ||
          lookahead == ']' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') ADVANCE(366);
      if (lookahead == '_') ADVANCE(947);
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(222)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(807);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(873);
      if (lookahead != 0 &&
          lookahead != ')' &&
          lookahead != ';') ADVANCE(958);
      END_STATE();
    case 223:
      if (lookahead == '\n') ADVANCE(743);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(862);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(602);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(332);
      if (lookahead == '-') ADVANCE(950);
      if (lookahead == '0') ADVANCE(803);
      if (lookahead == ';') ADVANCE(554);
      if (lookahead == '<') ADVANCE(609);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '[' ||
          lookahead == ']' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') ADVANCE(493);
      if (lookahead == '`') ADVANCE(856);
      if (lookahead == 'e') ADVANCE(956);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(223)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(805);
      if (lookahead != 0 &&
          lookahead != ')') ADVANCE(958);
      END_STATE();
    case 224:
      if (lookahead == '\n') ADVANCE(744);
      if (lookahead == '!') ADVANCE(672);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(327);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(332);
      if (lookahead == '*') ADVANCE(640);
      if (lookahead == '-') ADVANCE(636);
      if (lookahead == '0') ADVANCE(943);
      if (lookahead == '<') ADVANCE(610);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '?') ADVANCE(761);
      if (lookahead == '@') ADVANCE(942);
      if (lookahead == '[' ||
          lookahead == ']' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') ADVANCE(414);
      if (lookahead == '_') ADVANCE(947);
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(224)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(807);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(873);
      if (lookahead != 0 &&
          lookahead != ')' &&
          lookahead != ';') ADVANCE(958);
      END_STATE();
    case 225:
      if (lookahead == '\n') ADVANCE(745);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(862);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(602);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(332);
      if (lookahead == '-') ADVANCE(950);
      if (lookahead == '0') ADVANCE(804);
      if (lookahead == ';') ADVANCE(554);
      if (lookahead == '<') ADVANCE(610);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '[' ||
          lookahead == ']' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') ADVANCE(491);
      if (lookahead == '`') ADVANCE(856);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(225)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(806);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(868);
      if (lookahead != 0 &&
          lookahead != ')') ADVANCE(958);
      END_STATE();
    case 226:
      if (lookahead == '\n') ADVANCE(746);
      if (lookahead == '!') ADVANCE(672);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(327);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(332);
      if (lookahead == '*') ADVANCE(640);
      if (lookahead == '-') ADVANCE(636);
      if (lookahead == '0') ADVANCE(943);
      if (lookahead == '<') ADVANCE(608);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '?') ADVANCE(761);
      if (lookahead == '@') ADVANCE(942);
      if (lookahead == '[' ||
          lookahead == ']' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') ADVANCE(395);
      if (lookahead == '_') ADVANCE(947);
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == '|') ADVANCE(354);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(226)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(807);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(873);
      if (lookahead != 0 &&
          lookahead != ')' &&
          lookahead != ';') ADVANCE(958);
      END_STATE();
    case 227:
      if (lookahead == '\n') ADVANCE(747);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(862);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(602);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(332);
      if (lookahead == '-') ADVANCE(950);
      if (lookahead == '0') ADVANCE(803);
      if (lookahead == ';') ADVANCE(554);
      if (lookahead == '<') ADVANCE(609);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '[' ||
          lookahead == ']' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') ADVANCE(494);
      if (lookahead == '`') ADVANCE(856);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(227)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(805);
      if (lookahead != 0 &&
          lookahead != ')') ADVANCE(958);
      END_STATE();
    case 228:
      if (lookahead == '\n') ADVANCE(748);
      if (lookahead == '!') ADVANCE(672);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(599);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(332);
      if (lookahead == '*') ADVANCE(640);
      if (lookahead == '-') ADVANCE(636);
      if (lookahead == '0') ADVANCE(943);
      if (lookahead == ';') ADVANCE(555);
      if (lookahead == '<') ADVANCE(333);
      if (lookahead == '>') ADVANCE(334);
      if (lookahead == '?') ADVANCE(761);
      if (lookahead == '@') ADVANCE(942);
      if (lookahead == '[' ||
          lookahead == ']' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') ADVANCE(377);
      if (lookahead == '_') ADVANCE(947);
      if (lookahead == '`') ADVANCE(855);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(228)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(807);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(873);
      if (lookahead != 0 &&
          lookahead != ')' &&
          lookahead != '|') ADVANCE(958);
      END_STATE();
    case 229:
      if (lookahead == '\n') ADVANCE(749);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(862);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(602);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(332);
      if (lookahead == '-') ADVANCE(950);
      if (lookahead == '0') ADVANCE(803);
      if (lookahead == ';') ADVANCE(554);
      if (lookahead == '<') ADVANCE(610);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '[' ||
          lookahead == ']' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') ADVANCE(506);
      if (lookahead == '`') ADVANCE(856);
      if (lookahead == 'e') ADVANCE(956);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(229)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(805);
      if (lookahead != 0 &&
          lookahead != ')') ADVANCE(958);
      END_STATE();
    case 230:
      if (lookahead == '\n') ADVANCE(750);
      if (lookahead == '!') ADVANCE(672);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(332);
      if (lookahead == '*') ADVANCE(640);
      if (lookahead == '-') ADVANCE(636);
      if (lookahead == '0') ADVANCE(943);
      if (lookahead == '<') ADVANCE(333);
      if (lookahead == '>') ADVANCE(334);
      if (lookahead == '?') ADVANCE(761);
      if (lookahead == '@') ADVANCE(942);
      if (lookahead == '[' ||
          lookahead == ']' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') ADVANCE(381);
      if (lookahead == '_') ADVANCE(947);
      if (lookahead == '`') ADVANCE(855);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(230)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(807);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(873);
      if (lookahead != 0 &&
          (lookahead < '&' || ')' < lookahead) &&
          lookahead != ';' &&
          lookahead != '|') ADVANCE(958);
      END_STATE();
    case 231:
      if (lookahead == '\n') ADVANCE(751);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(862);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(602);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(332);
      if (lookahead == ')') ADVANCE(655);
      if (lookahead == '-') ADVANCE(950);
      if (lookahead == '0') ADVANCE(804);
      if (lookahead == ';') ADVANCE(555);
      if (lookahead == '<') ADVANCE(610);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '[' ||
          lookahead == ']' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') ADVANCE(492);
      if (lookahead == '`') ADVANCE(856);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(231)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(806);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(868);
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 232:
      if (lookahead == '\n') ADVANCE(752);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(862);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(602);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(332);
      if (lookahead == ')') ADVANCE(655);
      if (lookahead == '-') ADVANCE(950);
      if (lookahead == '0') ADVANCE(803);
      if (lookahead == ';') ADVANCE(555);
      if (lookahead == '<') ADVANCE(609);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '[' ||
          lookahead == ']' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') ADVANCE(496);
      if (lookahead == '`') ADVANCE(856);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(232)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(805);
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 233:
      if (lookahead == '\n') ADVANCE(753);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(862);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(602);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(332);
      if (lookahead == '-') ADVANCE(950);
      if (lookahead == '0') ADVANCE(803);
      if (lookahead == ';') ADVANCE(554);
      if (lookahead == '<') ADVANCE(610);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '[' ||
          lookahead == ']' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') ADVANCE(507);
      if (lookahead == '`') ADVANCE(856);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(233)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(805);
      if (lookahead != 0 &&
          lookahead != ')') ADVANCE(958);
      END_STATE();
    case 234:
      if (lookahead == '\n') ADVANCE(754);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(862);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(602);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(332);
      if (lookahead == ')') ADVANCE(655);
      if (lookahead == '-') ADVANCE(950);
      if (lookahead == '0') ADVANCE(803);
      if (lookahead == ';') ADVANCE(555);
      if (lookahead == '<') ADVANCE(610);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '[' ||
          lookahead == ']' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') ADVANCE(508);
      if (lookahead == '`') ADVANCE(856);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(234)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(805);
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 235:
      if (lookahead == '\n') ADVANCE(755);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(862);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(599);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(654);
      if (lookahead == '-') ADVANCE(950);
      if (lookahead == '0') ADVANCE(803);
      if (lookahead == ';') ADVANCE(555);
      if (lookahead == '<') ADVANCE(333);
      if (lookahead == '>') ADVANCE(334);
      if (lookahead == '[' ||
          lookahead == ']' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') ADVANCE(451);
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == 'e') ADVANCE(956);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(235)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(805);
      if (lookahead != 0 &&
          lookahead != ')' &&
          lookahead != '|') ADVANCE(958);
      END_STATE();
    case 236:
      if (lookahead == '\n') ADVANCE(756);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(862);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(327);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(332);
      if (lookahead == '-') ADVANCE(950);
      if (lookahead == '0') ADVANCE(803);
      if (lookahead == '<') ADVANCE(608);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '[' ||
          lookahead == ']' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') ADVANCE(497);
      if (lookahead == '`') ADVANCE(856);
      if (lookahead == '|') ADVANCE(354);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(236)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(805);
      if (lookahead != 0 &&
          lookahead != ')' &&
          lookahead != ';') ADVANCE(958);
      END_STATE();
    case 237:
      if (lookahead == '\n') ADVANCE(757);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(862);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(599);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(332);
      if (lookahead == '-') ADVANCE(950);
      if (lookahead == '0') ADVANCE(803);
      if (lookahead == ';') ADVANCE(555);
      if (lookahead == '<') ADVANCE(333);
      if (lookahead == '>') ADVANCE(334);
      if (lookahead == '[' ||
          lookahead == ']' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') ADVANCE(459);
      if (lookahead == '`') ADVANCE(855);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(237)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(805);
      if (lookahead != 0 &&
          lookahead != ')' &&
          lookahead != '|') ADVANCE(958);
      END_STATE();
    case 238:
      if (lookahead == '\n') ADVANCE(758);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(862);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(599);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(332);
      if (lookahead == '-') ADVANCE(950);
      if (lookahead == '0') ADVANCE(803);
      if (lookahead == ';') ADVANCE(555);
      if (lookahead == '<') ADVANCE(333);
      if (lookahead == '>') ADVANCE(334);
      if (lookahead == '[' ||
          lookahead == ']' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') ADVANCE(503);
      if (lookahead == '`') ADVANCE(856);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(238)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(805);
      if (lookahead != 0 &&
          lookahead != ')' &&
          lookahead != '|') ADVANCE(958);
      END_STATE();
    case 239:
      if (lookahead == '!') ADVANCE(670);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '%') ADVANCE(647);
      if (lookahead == '&') ADVANCE(601);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(654);
      if (lookahead == ')') ADVANCE(655);
      if (lookahead == '*') ADVANCE(638);
      if (lookahead == '+') ADVANCE(770);
      if (lookahead == ',') ADVANCE(557);
      if (lookahead == '-') ADVANCE(768);
      if (lookahead == '/') ADVANCE(642);
      if (lookahead == '0') ADVANCE(944);
      if (lookahead == ':') ADVANCE(765);
      if (lookahead == ';') ADVANCE(554);
      if (lookahead == '<') ADVANCE(607);
      if (lookahead == '=') ADVANCE(563);
      if (lookahead == '>') ADVANCE(616);
      if (lookahead == '?') ADVANCE(761);
      if (lookahead == '@') ADVANCE(942);
      if (lookahead == '[') ADVANCE(674);
      if (lookahead == '\\') ADVANCE(417);
      if (lookahead == ']') ADVANCE(675);
      if (lookahead == '^') ADVANCE(595);
      if (lookahead == '_') ADVANCE(946);
      if (lookahead == '`') ADVANCE(856);
      if (lookahead == 'e') ADVANCE(956);
      if (lookahead == 'i') ADVANCE(955);
      if (lookahead == '{') ADVANCE(666);
      if (lookahead == '|') ADVANCE(592);
      if (lookahead == '}') ADVANCE(818);
      if (lookahead == '~') ADVANCE(772);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(239)
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 240:
      if (lookahead == '!') ADVANCE(670);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '%') ADVANCE(649);
      if (lookahead == '&') ADVANCE(601);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(654);
      if (lookahead == '*') ADVANCE(638);
      if (lookahead == '+') ADVANCE(630);
      if (lookahead == '-') ADVANCE(633);
      if (lookahead == '/') ADVANCE(644);
      if (lookahead == '0') ADVANCE(943);
      if (lookahead == '<') ADVANCE(607);
      if (lookahead == '=') ADVANCE(560);
      if (lookahead == '>') ADVANCE(616);
      if (lookahead == '?') ADVANCE(761);
      if (lookahead == '@') ADVANCE(942);
      if (lookahead == '[' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') ADVANCE(356);
      if (lookahead == ']') ADVANCE(675);
      if (lookahead == '^') ADVANCE(596);
      if (lookahead == '_') ADVANCE(947);
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == '|') ADVANCE(592);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(240)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(807);
      if (('A' <= lookahead && lookahead <= 'z')) ADVANCE(873);
      if (lookahead != 0 &&
          lookahead != ')' &&
          lookahead != ';') ADVANCE(958);
      END_STATE();
    case 241:
      if (lookahead == '!') ADVANCE(670);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(862);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '%') ADVANCE(649);
      if (lookahead == '&') ADVANCE(601);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(654);
      if (lookahead == '*') ADVANCE(638);
      if (lookahead == '+') ADVANCE(771);
      if (lookahead == '-') ADVANCE(769);
      if (lookahead == '/') ADVANCE(644);
      if (lookahead == '0') ADVANCE(803);
      if (lookahead == '<') ADVANCE(607);
      if (lookahead == '=') ADVANCE(560);
      if (lookahead == '>') ADVANCE(616);
      if (lookahead == '?') ADVANCE(761);
      if (lookahead == '[' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') ADVANCE(422);
      if (lookahead == ']') ADVANCE(675);
      if (lookahead == '^') ADVANCE(596);
      if (lookahead == '`') ADVANCE(856);
      if (lookahead == '|') ADVANCE(592);
      if (lookahead == '~') ADVANCE(772);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(241)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(805);
      if (lookahead != 0 &&
          lookahead != ')' &&
          lookahead != ';') ADVANCE(958);
      END_STATE();
    case 242:
      if (lookahead == '!') ADVANCE(670);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(862);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '%') ADVANCE(649);
      if (lookahead == '&') ADVANCE(600);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(654);
      if (lookahead == ')') ADVANCE(655);
      if (lookahead == '*') ADVANCE(638);
      if (lookahead == '+') ADVANCE(771);
      if (lookahead == '-') ADVANCE(769);
      if (lookahead == '/') ADVANCE(644);
      if (lookahead == '0') ADVANCE(803);
      if (lookahead == '<') ADVANCE(614);
      if (lookahead == '=') ADVANCE(560);
      if (lookahead == '>') ADVANCE(619);
      if (lookahead == '?') ADVANCE(761);
      if (lookahead == '[' ||
          lookahead == ']' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') ADVANCE(425);
      if (lookahead == '^') ADVANCE(596);
      if (lookahead == '`') ADVANCE(856);
      if (lookahead == '|') ADVANCE(594);
      if (lookahead == '~') ADVANCE(772);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(242)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(805);
      if (lookahead != 0 &&
          lookahead != ';') ADVANCE(958);
      END_STATE();
    case 243:
      if (lookahead == '!') ADVANCE(670);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(862);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '%') ADVANCE(649);
      if (lookahead == '&') ADVANCE(600);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(654);
      if (lookahead == ')') ADVANCE(337);
      if (lookahead == '*') ADVANCE(638);
      if (lookahead == '+') ADVANCE(771);
      if (lookahead == '-') ADVANCE(769);
      if (lookahead == '/') ADVANCE(644);
      if (lookahead == '0') ADVANCE(803);
      if (lookahead == '<') ADVANCE(614);
      if (lookahead == '=') ADVANCE(560);
      if (lookahead == '>') ADVANCE(619);
      if (lookahead == '?') ADVANCE(761);
      if (lookahead == '[' ||
          lookahead == ']' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') ADVANCE(484);
      if (lookahead == '^') ADVANCE(596);
      if (lookahead == '`') ADVANCE(856);
      if (lookahead == '|') ADVANCE(594);
      if (lookahead == '~') ADVANCE(772);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(243)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(805);
      if (lookahead != 0 &&
          lookahead != ';') ADVANCE(958);
      END_STATE();
    case 244:
      if (lookahead == '!') ADVANCE(670);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(862);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '%') ADVANCE(649);
      if (lookahead == '&') ADVANCE(600);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(654);
      if (lookahead == '*') ADVANCE(638);
      if (lookahead == '+') ADVANCE(771);
      if (lookahead == '-') ADVANCE(769);
      if (lookahead == '/') ADVANCE(644);
      if (lookahead == '0') ADVANCE(803);
      if (lookahead == ':') ADVANCE(765);
      if (lookahead == '<') ADVANCE(614);
      if (lookahead == '=') ADVANCE(560);
      if (lookahead == '>') ADVANCE(619);
      if (lookahead == '?') ADVANCE(761);
      if (lookahead == '[' ||
          lookahead == ']' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') ADVANCE(426);
      if (lookahead == '^') ADVANCE(596);
      if (lookahead == '`') ADVANCE(856);
      if (lookahead == '|') ADVANCE(594);
      if (lookahead == '~') ADVANCE(772);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(244)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(805);
      if (lookahead != 0 &&
          lookahead != ')' &&
          lookahead != ';') ADVANCE(958);
      END_STATE();
    case 245:
      if (lookahead == '!') ADVANCE(670);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(862);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '%') ADVANCE(649);
      if (lookahead == '&') ADVANCE(600);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(654);
      if (lookahead == '*') ADVANCE(638);
      if (lookahead == '+') ADVANCE(771);
      if (lookahead == '-') ADVANCE(769);
      if (lookahead == '/') ADVANCE(644);
      if (lookahead == '0') ADVANCE(803);
      if (lookahead == '<') ADVANCE(614);
      if (lookahead == '=') ADVANCE(560);
      if (lookahead == '>') ADVANCE(619);
      if (lookahead == '?') ADVANCE(761);
      if (lookahead == '[' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') ADVANCE(483);
      if (lookahead == ']') ADVANCE(675);
      if (lookahead == '^') ADVANCE(596);
      if (lookahead == '`') ADVANCE(856);
      if (lookahead == '|') ADVANCE(594);
      if (lookahead == '~') ADVANCE(772);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(245)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(805);
      if (lookahead != 0 &&
          lookahead != ')' &&
          lookahead != ';') ADVANCE(958);
      END_STATE();
    case 246:
      if (lookahead == '!') ADVANCE(670);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(862);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '%') ADVANCE(649);
      if (lookahead == '&') ADVANCE(600);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(654);
      if (lookahead == '*') ADVANCE(638);
      if (lookahead == '+') ADVANCE(771);
      if (lookahead == '-') ADVANCE(769);
      if (lookahead == '/') ADVANCE(644);
      if (lookahead == '0') ADVANCE(803);
      if (lookahead == '<') ADVANCE(614);
      if (lookahead == '=') ADVANCE(560);
      if (lookahead == '>') ADVANCE(619);
      if (lookahead == '?') ADVANCE(761);
      if (lookahead == '[' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') ADVANCE(505);
      if (lookahead == ']') ADVANCE(787);
      if (lookahead == '^') ADVANCE(596);
      if (lookahead == '`') ADVANCE(856);
      if (lookahead == '|') ADVANCE(594);
      if (lookahead == '~') ADVANCE(772);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(246)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(805);
      if (lookahead != 0 &&
          lookahead != ')' &&
          lookahead != ';') ADVANCE(958);
      END_STATE();
    case 247:
      if (lookahead == '!') ADVANCE(672);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(346);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(332);
      if (lookahead == '*') ADVANCE(640);
      if (lookahead == '-') ADVANCE(636);
      if (lookahead == '0') ADVANCE(943);
      if (lookahead == '<') ADVANCE(608);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '?') ADVANCE(761);
      if (lookahead == '@') ADVANCE(942);
      if (lookahead == '[' ||
          lookahead == ']' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') ADVANCE(372);
      if (lookahead == '_') ADVANCE(947);
      if (lookahead == '`') ADVANCE(855);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(247)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(807);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(873);
      if (lookahead != 0 &&
          lookahead != ')' &&
          lookahead != ';' &&
          lookahead != '|') ADVANCE(958);
      END_STATE();
    case 248:
      if (lookahead == '!') ADVANCE(672);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(327);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(654);
      if (lookahead == '*') ADVANCE(640);
      if (lookahead == '-') ADVANCE(636);
      if (lookahead == '0') ADVANCE(943);
      if (lookahead == '<') ADVANCE(609);
      if (lookahead == '=') ADVANCE(952);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '?') ADVANCE(761);
      if (lookahead == '@') ADVANCE(942);
      if (lookahead == '[' ||
          lookahead == ']' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') ADVANCE(364);
      if (lookahead == '_') ADVANCE(947);
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(248)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(807);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(873);
      if (lookahead != 0 &&
          lookahead != ')' &&
          lookahead != ';') ADVANCE(958);
      END_STATE();
    case 249:
      if (lookahead == '!') ADVANCE(672);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(327);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(332);
      if (lookahead == '*') ADVANCE(640);
      if (lookahead == '-') ADVANCE(636);
      if (lookahead == '0') ADVANCE(943);
      if (lookahead == '<') ADVANCE(609);
      if (lookahead == '=') ADVANCE(952);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '?') ADVANCE(761);
      if (lookahead == '@') ADVANCE(942);
      if (lookahead == '[' ||
          lookahead == ']' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') ADVANCE(394);
      if (lookahead == '_') ADVANCE(947);
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(249)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(807);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(873);
      if (lookahead != 0 &&
          lookahead != ')' &&
          lookahead != ';') ADVANCE(958);
      END_STATE();
    case 250:
      if (lookahead == '!') ADVANCE(672);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(327);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(332);
      if (lookahead == '*') ADVANCE(640);
      if (lookahead == '-') ADVANCE(636);
      if (lookahead == '0') ADVANCE(943);
      if (lookahead == '<') ADVANCE(609);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '?') ADVANCE(761);
      if (lookahead == '@') ADVANCE(942);
      if (lookahead == '[' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') ADVANCE(365);
      if (lookahead == ']') ADVANCE(675);
      if (lookahead == '_') ADVANCE(947);
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(250)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(807);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(873);
      if (lookahead != 0 &&
          lookahead != ')' &&
          lookahead != ';') ADVANCE(958);
      END_STATE();
    case 251:
      if (lookahead == '!') ADVANCE(672);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(327);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(332);
      if (lookahead == '*') ADVANCE(640);
      if (lookahead == '-') ADVANCE(636);
      if (lookahead == '0') ADVANCE(943);
      if (lookahead == '<') ADVANCE(609);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '?') ADVANCE(761);
      if (lookahead == '@') ADVANCE(942);
      if (lookahead == '[' ||
          lookahead == ']' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') ADVANCE(368);
      if (lookahead == '_') ADVANCE(947);
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(251)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(807);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(873);
      if (lookahead != 0 &&
          lookahead != ')' &&
          lookahead != ';') ADVANCE(958);
      END_STATE();
    case 252:
      if (lookahead == '!') ADVANCE(672);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(327);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(332);
      if (lookahead == '*') ADVANCE(640);
      if (lookahead == '-') ADVANCE(636);
      if (lookahead == '0') ADVANCE(943);
      if (lookahead == '<') ADVANCE(610);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '?') ADVANCE(761);
      if (lookahead == '@') ADVANCE(942);
      if (lookahead == '[' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') ADVANCE(413);
      if (lookahead == ']') ADVANCE(675);
      if (lookahead == '_') ADVANCE(947);
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(252)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(807);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(873);
      if (lookahead != 0 &&
          lookahead != ')' &&
          lookahead != ';') ADVANCE(958);
      END_STATE();
    case 253:
      if (lookahead == '!') ADVANCE(672);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(327);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(332);
      if (lookahead == '*') ADVANCE(640);
      if (lookahead == '-') ADVANCE(636);
      if (lookahead == '0') ADVANCE(943);
      if (lookahead == '<') ADVANCE(610);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '?') ADVANCE(761);
      if (lookahead == '@') ADVANCE(942);
      if (lookahead == '[' ||
          lookahead == ']' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') ADVANCE(415);
      if (lookahead == '_') ADVANCE(947);
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(253)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(807);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(873);
      if (lookahead != 0 &&
          lookahead != ')' &&
          lookahead != ';') ADVANCE(958);
      END_STATE();
    case 254:
      if (lookahead == '!') ADVANCE(672);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(332);
      if (lookahead == ')') ADVANCE(655);
      if (lookahead == '*') ADVANCE(640);
      if (lookahead == '-') ADVANCE(636);
      if (lookahead == '0') ADVANCE(943);
      if (lookahead == '<') ADVANCE(333);
      if (lookahead == '>') ADVANCE(334);
      if (lookahead == '?') ADVANCE(761);
      if (lookahead == '@') ADVANCE(942);
      if (lookahead == '[' ||
          lookahead == ']' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') ADVANCE(380);
      if (lookahead == '_') ADVANCE(947);
      if (lookahead == '`') ADVANCE(855);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(254)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(807);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(873);
      if (lookahead != 0 &&
          lookahead != '&' &&
          lookahead != ';' &&
          lookahead != '|') ADVANCE(958);
      END_STATE();
    case 255:
      if (lookahead == '!') ADVANCE(672);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(862);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(346);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(654);
      if (lookahead == ')') ADVANCE(655);
      if (lookahead == '-') ADVANCE(950);
      if (lookahead == '0') ADVANCE(803);
      if (lookahead == ';') ADVANCE(328);
      if (lookahead == '<') ADVANCE(608);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '[') ADVANCE(674);
      if (lookahead == '\\') ADVANCE(418);
      if (lookahead == ']' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == '{') ADVANCE(666);
      if (lookahead == '|') ADVANCE(590);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(255)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(805);
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 256:
      if (lookahead == '!') ADVANCE(672);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(862);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(346);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(654);
      if (lookahead == ')') ADVANCE(337);
      if (lookahead == '+') ADVANCE(771);
      if (lookahead == '-') ADVANCE(769);
      if (lookahead == '0') ADVANCE(803);
      if (lookahead == '<') ADVANCE(608);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '[') ADVANCE(674);
      if (lookahead == '\\') ADVANCE(419);
      if (lookahead == ']' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == '{') ADVANCE(666);
      if (lookahead == '|') ADVANCE(591);
      if (lookahead == '~') ADVANCE(772);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(256)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(805);
      if (lookahead != 0 &&
          lookahead != ';') ADVANCE(958);
      END_STATE();
    case 257:
      if (lookahead == '!') ADVANCE(672);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(862);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(346);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(654);
      if (lookahead == '+') ADVANCE(771);
      if (lookahead == '-') ADVANCE(769);
      if (lookahead == '0') ADVANCE(803);
      if (lookahead == '<') ADVANCE(608);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '[') ADVANCE(674);
      if (lookahead == '\\') ADVANCE(420);
      if (lookahead == ']') ADVANCE(675);
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == '{') ADVANCE(666);
      if (lookahead == '}') ADVANCE(786);
      if (lookahead == '~') ADVANCE(772);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(257)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(805);
      if (lookahead != 0 &&
          lookahead != ')' &&
          lookahead != ';' &&
          lookahead != '|') ADVANCE(958);
      END_STATE();
    case 258:
      if (lookahead == '!') ADVANCE(672);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(862);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(346);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(654);
      if (lookahead == '+') ADVANCE(771);
      if (lookahead == '-') ADVANCE(769);
      if (lookahead == '0') ADVANCE(803);
      if (lookahead == '<') ADVANCE(608);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '[') ADVANCE(674);
      if (lookahead == '\\') ADVANCE(429);
      if (lookahead == ']' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == '~') ADVANCE(772);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(258)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(805);
      if (lookahead != 0 &&
          lookahead != ')' &&
          lookahead != ';' &&
          lookahead != '|') ADVANCE(958);
      END_STATE();
    case 259:
      if (lookahead == '!') ADVANCE(672);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(862);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(346);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(654);
      if (lookahead == '-') ADVANCE(950);
      if (lookahead == '0') ADVANCE(803);
      if (lookahead == ';') ADVANCE(328);
      if (lookahead == '<') ADVANCE(608);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '[') ADVANCE(674);
      if (lookahead == '\\') ADVANCE(423);
      if (lookahead == ']' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == 'e') ADVANCE(956);
      if (lookahead == '{') ADVANCE(666);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(259)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(805);
      if (lookahead != 0 &&
          lookahead != ')' &&
          lookahead != '|') ADVANCE(958);
      END_STATE();
    case 260:
      if (lookahead == '!') ADVANCE(672);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(862);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(346);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(654);
      if (lookahead == '-') ADVANCE(950);
      if (lookahead == '0') ADVANCE(803);
      if (lookahead == '<') ADVANCE(608);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '[') ADVANCE(674);
      if (lookahead == '\\') ADVANCE(424);
      if (lookahead == ']') ADVANCE(786);
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == '{') ADVANCE(666);
      if (lookahead == '}') ADVANCE(667);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(260)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(805);
      if (lookahead != 0 &&
          lookahead != ')' &&
          lookahead != ';' &&
          lookahead != '|') ADVANCE(958);
      END_STATE();
    case 261:
      if (lookahead == '!') ADVANCE(672);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(862);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(654);
      if (lookahead == ')') ADVANCE(337);
      if (lookahead == '+') ADVANCE(771);
      if (lookahead == '-') ADVANCE(769);
      if (lookahead == '0') ADVANCE(803);
      if (lookahead == '<') ADVANCE(333);
      if (lookahead == '>') ADVANCE(334);
      if (lookahead == '[' ||
          lookahead == ']' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') ADVANCE(449);
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == '~') ADVANCE(772);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(261)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(805);
      if (lookahead != 0 &&
          lookahead != '&' &&
          lookahead != ';' &&
          lookahead != '|') ADVANCE(958);
      END_STATE();
    case 262:
      if (lookahead == '!') ADVANCE(672);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(862);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(654);
      if (lookahead == ')') ADVANCE(337);
      if (lookahead == '+') ADVANCE(771);
      if (lookahead == '-') ADVANCE(769);
      if (lookahead == '0') ADVANCE(804);
      if (lookahead == '<') ADVANCE(333);
      if (lookahead == '>') ADVANCE(334);
      if (lookahead == '[' ||
          lookahead == ']' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') ADVANCE(441);
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == '~') ADVANCE(772);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(262)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(806);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(868);
      if (lookahead != 0 &&
          lookahead != '&' &&
          lookahead != ';' &&
          lookahead != '|') ADVANCE(958);
      END_STATE();
    case 263:
      if (lookahead == '!') ADVANCE(951);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(862);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '%') ADVANCE(649);
      if (lookahead == '&') ADVANCE(601);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(654);
      if (lookahead == '*') ADVANCE(638);
      if (lookahead == '+') ADVANCE(630);
      if (lookahead == '-') ADVANCE(633);
      if (lookahead == '/') ADVANCE(644);
      if (lookahead == '0') ADVANCE(803);
      if (lookahead == '<') ADVANCE(607);
      if (lookahead == '=') ADVANCE(560);
      if (lookahead == '>') ADVANCE(616);
      if (lookahead == '?') ADVANCE(761);
      if (lookahead == '[' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') ADVANCE(428);
      if (lookahead == ']') ADVANCE(675);
      if (lookahead == '^') ADVANCE(596);
      if (lookahead == '`') ADVANCE(856);
      if (lookahead == '|') ADVANCE(592);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(263)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(805);
      if (lookahead != 0 &&
          lookahead != ')' &&
          lookahead != ';') ADVANCE(958);
      END_STATE();
    case 264:
      if (lookahead == '!') ADVANCE(951);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(862);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '%') ADVANCE(649);
      if (lookahead == '&') ADVANCE(601);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(654);
      if (lookahead == '*') ADVANCE(638);
      if (lookahead == '+') ADVANCE(630);
      if (lookahead == '-') ADVANCE(633);
      if (lookahead == '/') ADVANCE(644);
      if (lookahead == '0') ADVANCE(803);
      if (lookahead == '<') ADVANCE(607);
      if (lookahead == '=') ADVANCE(560);
      if (lookahead == '>') ADVANCE(616);
      if (lookahead == '?') ADVANCE(761);
      if (lookahead == '[' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') ADVANCE(486);
      if (lookahead == ']') ADVANCE(675);
      if (lookahead == '^') ADVANCE(596);
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == '|') ADVANCE(592);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(264)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(805);
      if (lookahead != 0 &&
          lookahead != ')' &&
          lookahead != ';') ADVANCE(958);
      END_STATE();
    case 265:
      if (lookahead == '!') ADVANCE(671);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(780);
      if (lookahead == '%') ADVANCE(650);
      if (lookahead == '&') ADVANCE(600);
      if (lookahead == ')') ADVANCE(655);
      if (lookahead == '*') ADVANCE(639);
      if (lookahead == '+') ADVANCE(631);
      if (lookahead == '-') ADVANCE(634);
      if (lookahead == '/') ADVANCE(645);
      if (lookahead == '0') ADVANCE(945);
      if (lookahead == '<') ADVANCE(615);
      if (lookahead == '=') ADVANCE(562);
      if (lookahead == '>') ADVANCE(620);
      if (lookahead == '?') ADVANCE(760);
      if (lookahead == '@') ADVANCE(941);
      if (lookahead == '\\') ADVANCE(370);
      if (lookahead == '^') ADVANCE(597);
      if (lookahead == '_') ADVANCE(948);
      if (lookahead == '|') ADVANCE(594);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(265)
      if (('1' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(879);
      END_STATE();
    case 266:
      if (lookahead == '!') ADVANCE(671);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(780);
      if (lookahead == '%') ADVANCE(650);
      if (lookahead == '&') ADVANCE(600);
      if (lookahead == ')') ADVANCE(337);
      if (lookahead == '*') ADVANCE(639);
      if (lookahead == '+') ADVANCE(631);
      if (lookahead == '-') ADVANCE(634);
      if (lookahead == '/') ADVANCE(645);
      if (lookahead == '0') ADVANCE(945);
      if (lookahead == '<') ADVANCE(615);
      if (lookahead == '=') ADVANCE(562);
      if (lookahead == '>') ADVANCE(620);
      if (lookahead == '?') ADVANCE(760);
      if (lookahead == '@') ADVANCE(941);
      if (lookahead == '\\') ADVANCE(397);
      if (lookahead == '^') ADVANCE(597);
      if (lookahead == '_') ADVANCE(948);
      if (lookahead == '|') ADVANCE(594);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(266)
      if (('1' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(879);
      END_STATE();
    case 267:
      if (lookahead == '!') ADVANCE(671);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(780);
      if (lookahead == '%') ADVANCE(650);
      if (lookahead == '&') ADVANCE(600);
      if (lookahead == '*') ADVANCE(639);
      if (lookahead == '+') ADVANCE(631);
      if (lookahead == '-') ADVANCE(634);
      if (lookahead == '/') ADVANCE(645);
      if (lookahead == '0') ADVANCE(945);
      if (lookahead == ':') ADVANCE(762);
      if (lookahead == '<') ADVANCE(615);
      if (lookahead == '=') ADVANCE(562);
      if (lookahead == '>') ADVANCE(620);
      if (lookahead == '?') ADVANCE(760);
      if (lookahead == '@') ADVANCE(941);
      if (lookahead == '\\') ADVANCE(369);
      if (lookahead == '^') ADVANCE(597);
      if (lookahead == '_') ADVANCE(948);
      if (lookahead == '|') ADVANCE(594);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(267)
      if (('1' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(879);
      END_STATE();
    case 268:
      if (lookahead == '!') ADVANCE(671);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(780);
      if (lookahead == '%') ADVANCE(650);
      if (lookahead == '&') ADVANCE(600);
      if (lookahead == '*') ADVANCE(639);
      if (lookahead == '+') ADVANCE(631);
      if (lookahead == '-') ADVANCE(634);
      if (lookahead == '/') ADVANCE(645);
      if (lookahead == '0') ADVANCE(945);
      if (lookahead == '<') ADVANCE(615);
      if (lookahead == '=') ADVANCE(562);
      if (lookahead == '>') ADVANCE(620);
      if (lookahead == '?') ADVANCE(760);
      if (lookahead == '@') ADVANCE(941);
      if (lookahead == '\\') ADVANCE(367);
      if (lookahead == ']') ADVANCE(675);
      if (lookahead == '^') ADVANCE(597);
      if (lookahead == '_') ADVANCE(948);
      if (lookahead == '|') ADVANCE(594);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(268)
      if (('1' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(879);
      END_STATE();
    case 269:
      if (lookahead == '!') ADVANCE(671);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(780);
      if (lookahead == '%') ADVANCE(650);
      if (lookahead == '&') ADVANCE(600);
      if (lookahead == '*') ADVANCE(639);
      if (lookahead == '+') ADVANCE(631);
      if (lookahead == '-') ADVANCE(634);
      if (lookahead == '/') ADVANCE(645);
      if (lookahead == '0') ADVANCE(945);
      if (lookahead == '<') ADVANCE(615);
      if (lookahead == '=') ADVANCE(562);
      if (lookahead == '>') ADVANCE(620);
      if (lookahead == '?') ADVANCE(760);
      if (lookahead == '@') ADVANCE(941);
      if (lookahead == '\\') ADVANCE(396);
      if (lookahead == ']') ADVANCE(347);
      if (lookahead == '^') ADVANCE(597);
      if (lookahead == '_') ADVANCE(948);
      if (lookahead == '|') ADVANCE(594);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(269)
      if (('1' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(879);
      END_STATE();
    case 270:
      if (lookahead == '!') ADVANCE(671);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(780);
      if (lookahead == '%') ADVANCE(650);
      if (lookahead == '&') ADVANCE(600);
      if (lookahead == '*') ADVANCE(639);
      if (lookahead == '+') ADVANCE(631);
      if (lookahead == '-') ADVANCE(634);
      if (lookahead == '/') ADVANCE(645);
      if (lookahead == '0') ADVANCE(945);
      if (lookahead == '<') ADVANCE(615);
      if (lookahead == '=') ADVANCE(562);
      if (lookahead == '>') ADVANCE(620);
      if (lookahead == '?') ADVANCE(760);
      if (lookahead == '@') ADVANCE(941);
      if (lookahead == '\\') ADVANCE(371);
      if (lookahead == '^') ADVANCE(597);
      if (lookahead == '_') ADVANCE(948);
      if (lookahead == '|') ADVANCE(594);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(270)
      if (('1' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(879);
      END_STATE();
    case 271:
      if (lookahead == '!') ADVANCE(344);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(862);
      if (lookahead == '$') ADVANCE(785);
      if (lookahead == '%') ADVANCE(650);
      if (lookahead == '&') ADVANCE(600);
      if (lookahead == ')') ADVANCE(337);
      if (lookahead == '*') ADVANCE(639);
      if (lookahead == '+') ADVANCE(631);
      if (lookahead == ',') ADVANCE(556);
      if (lookahead == '-') ADVANCE(634);
      if (lookahead == '.') ADVANCE(341);
      if (lookahead == '/') ADVANCE(645);
      if (lookahead == ':') ADVANCE(762);
      if (lookahead == ';') ADVANCE(328);
      if (lookahead == '<') ADVANCE(615);
      if (lookahead == '=') ADVANCE(562);
      if (lookahead == '>') ADVANCE(620);
      if (lookahead == '?') ADVANCE(760);
      if (lookahead == '[') ADVANCE(673);
      if (lookahead == '\\') SKIP(447)
      if (lookahead == ']') ADVANCE(347);
      if (lookahead == '^') ADVANCE(597);
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == 'e') ADVANCE(352);
      if (lookahead == 'i') ADVANCE(351);
      if (lookahead == '|') ADVANCE(594);
      if (lookahead == '}') ADVANCE(818);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(272)
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(775);
      END_STATE();
    case 272:
      if (lookahead == '!') ADVANCE(344);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(862);
      if (lookahead == '$') ADVANCE(785);
      if (lookahead == '%') ADVANCE(650);
      if (lookahead == '&') ADVANCE(600);
      if (lookahead == ')') ADVANCE(337);
      if (lookahead == '*') ADVANCE(639);
      if (lookahead == '+') ADVANCE(631);
      if (lookahead == ',') ADVANCE(556);
      if (lookahead == '-') ADVANCE(634);
      if (lookahead == '/') ADVANCE(645);
      if (lookahead == ':') ADVANCE(762);
      if (lookahead == ';') ADVANCE(328);
      if (lookahead == '<') ADVANCE(615);
      if (lookahead == '=') ADVANCE(562);
      if (lookahead == '>') ADVANCE(620);
      if (lookahead == '?') ADVANCE(760);
      if (lookahead == '[') ADVANCE(673);
      if (lookahead == '\\') SKIP(447)
      if (lookahead == ']') ADVANCE(347);
      if (lookahead == '^') ADVANCE(597);
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == 'e') ADVANCE(352);
      if (lookahead == 'i') ADVANCE(351);
      if (lookahead == '|') ADVANCE(594);
      if (lookahead == '}') ADVANCE(818);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(272)
      END_STATE();
    case 273:
      if (lookahead == '!') ADVANCE(344);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(862);
      if (lookahead == '$') ADVANCE(336);
      if (lookahead == '%') ADVANCE(650);
      if (lookahead == '&') ADVANCE(600);
      if (lookahead == '(') ADVANCE(332);
      if (lookahead == ')') ADVANCE(655);
      if (lookahead == '*') ADVANCE(639);
      if (lookahead == '+') ADVANCE(631);
      if (lookahead == ',') ADVANCE(556);
      if (lookahead == '-') ADVANCE(634);
      if (lookahead == '/') ADVANCE(645);
      if (lookahead == '<') ADVANCE(615);
      if (lookahead == '=') ADVANCE(562);
      if (lookahead == '>') ADVANCE(620);
      if (lookahead == '?') ADVANCE(760);
      if (lookahead == '[' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') SKIP(448)
      if (lookahead == ']') ADVANCE(675);
      if (lookahead == '^') ADVANCE(597);
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == '|') ADVANCE(594);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(273)
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'z')) ADVANCE(879);
      END_STATE();
    case 274:
      if (lookahead == '!') ADVANCE(344);
      if (lookahead == '#') ADVANCE(862);
      if (lookahead == '%') ADVANCE(650);
      if (lookahead == '&') ADVANCE(600);
      if (lookahead == ')') ADVANCE(655);
      if (lookahead == '*') ADVANCE(639);
      if (lookahead == '+') ADVANCE(631);
      if (lookahead == ',') ADVANCE(556);
      if (lookahead == '-') ADVANCE(635);
      if (lookahead == '/') ADVANCE(645);
      if (lookahead == '<') ADVANCE(615);
      if (lookahead == '=') ADVANCE(561);
      if (lookahead == '>') ADVANCE(620);
      if (lookahead == '\\') SKIP(502)
      if (lookahead == '^') ADVANCE(597);
      if (lookahead == '|') ADVANCE(594);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(274)
      END_STATE();
    case 275:
      if (lookahead == '!') ADVANCE(344);
      if (lookahead == '#') ADVANCE(862);
      if (lookahead == '%') ADVANCE(650);
      if (lookahead == '&') ADVANCE(600);
      if (lookahead == ')') ADVANCE(655);
      if (lookahead == '*') ADVANCE(639);
      if (lookahead == '+') ADVANCE(631);
      if (lookahead == '-') ADVANCE(634);
      if (lookahead == '/') ADVANCE(645);
      if (lookahead == '<') ADVANCE(615);
      if (lookahead == '=') ADVANCE(562);
      if (lookahead == '>') ADVANCE(620);
      if (lookahead == '?') ADVANCE(760);
      if (lookahead == '[') ADVANCE(673);
      if (lookahead == '\\') SKIP(457)
      if (lookahead == ']') ADVANCE(675);
      if (lookahead == '^') ADVANCE(597);
      if (lookahead == '`') ADVANCE(348);
      if (lookahead == '|') ADVANCE(594);
      if (lookahead == '}') ADVANCE(818);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(275)
      END_STATE();
    case 276:
      if (lookahead == '!') ADVANCE(344);
      if (lookahead == '#') ADVANCE(862);
      if (lookahead == '%') ADVANCE(650);
      if (lookahead == '&') ADVANCE(600);
      if (lookahead == ')') ADVANCE(655);
      if (lookahead == '*') ADVANCE(639);
      if (lookahead == '+') ADVANCE(631);
      if (lookahead == '-') ADVANCE(634);
      if (lookahead == '/') ADVANCE(645);
      if (lookahead == '<') ADVANCE(615);
      if (lookahead == '=') ADVANCE(562);
      if (lookahead == '>') ADVANCE(620);
      if (lookahead == '?') ADVANCE(760);
      if (lookahead == '[' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') SKIP(452)
      if (lookahead == ']') ADVANCE(675);
      if (lookahead == '^') ADVANCE(597);
      if (lookahead == '`') ADVANCE(348);
      if (lookahead == '|') ADVANCE(594);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(276)
      END_STATE();
    case 277:
      if (lookahead == '!') ADVANCE(344);
      if (lookahead == '#') ADVANCE(862);
      if (lookahead == '%') ADVANCE(650);
      if (lookahead == '&') ADVANCE(600);
      if (lookahead == ')') ADVANCE(655);
      if (lookahead == '*') ADVANCE(639);
      if (lookahead == '+') ADVANCE(631);
      if (lookahead == '-') ADVANCE(634);
      if (lookahead == '/') ADVANCE(645);
      if (lookahead == '<') ADVANCE(615);
      if (lookahead == '=') ADVANCE(562);
      if (lookahead == '>') ADVANCE(620);
      if (lookahead == '?') ADVANCE(760);
      if (lookahead == '[' ||
          lookahead == ']' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') SKIP(454)
      if (lookahead == '^') ADVANCE(597);
      if (lookahead == '`') ADVANCE(348);
      if (lookahead == '|') ADVANCE(594);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(277)
      END_STATE();
    case 278:
      if (lookahead == '!') ADVANCE(344);
      if (lookahead == '#') ADVANCE(862);
      if (lookahead == '%') ADVANCE(650);
      if (lookahead == '&') ADVANCE(600);
      if (lookahead == ')') ADVANCE(337);
      if (lookahead == '*') ADVANCE(639);
      if (lookahead == '+') ADVANCE(631);
      if (lookahead == ',') ADVANCE(556);
      if (lookahead == '-') ADVANCE(634);
      if (lookahead == '/') ADVANCE(645);
      if (lookahead == ':') ADVANCE(762);
      if (lookahead == '<') ADVANCE(615);
      if (lookahead == '=') ADVANCE(562);
      if (lookahead == '>') ADVANCE(620);
      if (lookahead == '?') ADVANCE(760);
      if (lookahead == '[' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') SKIP(450)
      if (lookahead == ']') ADVANCE(787);
      if (lookahead == '^') ADVANCE(597);
      if (lookahead == '`') ADVANCE(348);
      if (lookahead == '|') ADVANCE(594);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(278)
      END_STATE();
    case 279:
      if (lookahead == '!') ADVANCE(344);
      if (lookahead == '#') ADVANCE(862);
      if (lookahead == '%') ADVANCE(650);
      if (lookahead == '&') ADVANCE(600);
      if (lookahead == ')') ADVANCE(337);
      if (lookahead == '*') ADVANCE(639);
      if (lookahead == '+') ADVANCE(631);
      if (lookahead == ',') ADVANCE(556);
      if (lookahead == '-') ADVANCE(635);
      if (lookahead == '/') ADVANCE(645);
      if (lookahead == '<') ADVANCE(615);
      if (lookahead == '=') ADVANCE(561);
      if (lookahead == '>') ADVANCE(620);
      if (lookahead == '\\') SKIP(458)
      if (lookahead == '^') ADVANCE(597);
      if (lookahead == '|') ADVANCE(594);
      if (lookahead == '}') ADVANCE(778);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(280)
      END_STATE();
    case 280:
      if (lookahead == '!') ADVANCE(344);
      if (lookahead == '#') ADVANCE(862);
      if (lookahead == '%') ADVANCE(650);
      if (lookahead == '&') ADVANCE(600);
      if (lookahead == ')') ADVANCE(337);
      if (lookahead == '*') ADVANCE(639);
      if (lookahead == '+') ADVANCE(631);
      if (lookahead == ',') ADVANCE(556);
      if (lookahead == '-') ADVANCE(635);
      if (lookahead == '/') ADVANCE(645);
      if (lookahead == '<') ADVANCE(615);
      if (lookahead == '=') ADVANCE(561);
      if (lookahead == '>') ADVANCE(620);
      if (lookahead == '\\') SKIP(458)
      if (lookahead == '^') ADVANCE(597);
      if (lookahead == '|') ADVANCE(594);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(280)
      END_STATE();
    case 281:
      if (lookahead == '!') ADVANCE(344);
      if (lookahead == '#') ADVANCE(862);
      if (lookahead == '%') ADVANCE(650);
      if (lookahead == '&') ADVANCE(600);
      if (lookahead == ')') ADVANCE(337);
      if (lookahead == '*') ADVANCE(639);
      if (lookahead == '+') ADVANCE(631);
      if (lookahead == '-') ADVANCE(634);
      if (lookahead == '/') ADVANCE(645);
      if (lookahead == ':') ADVANCE(762);
      if (lookahead == '<') ADVANCE(615);
      if (lookahead == '=') ADVANCE(562);
      if (lookahead == '>') ADVANCE(620);
      if (lookahead == '?') ADVANCE(760);
      if (lookahead == '[' ||
          lookahead == ']' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') SKIP(453)
      if (lookahead == '^') ADVANCE(597);
      if (lookahead == '`') ADVANCE(348);
      if (lookahead == '|') ADVANCE(594);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(281)
      END_STATE();
    case 282:
      if (lookahead == '!') ADVANCE(669);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(780);
      if (lookahead == '&') ADVANCE(327);
      if (lookahead == '*') ADVANCE(637);
      if (lookahead == '-') ADVANCE(632);
      if (lookahead == '0') ADVANCE(945);
      if (lookahead == '<') ADVANCE(612);
      if (lookahead == '>') ADVANCE(618);
      if (lookahead == '?') ADVANCE(760);
      if (lookahead == '@') ADVANCE(941);
      if (lookahead == '\\') ADVANCE(379);
      if (lookahead == ']') ADVANCE(675);
      if (lookahead == '_') ADVANCE(948);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(282)
      if (('1' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(879);
      END_STATE();
    case 283:
      if (lookahead == '!') ADVANCE(669);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(780);
      if (lookahead == '&') ADVANCE(327);
      if (lookahead == '*') ADVANCE(637);
      if (lookahead == '-') ADVANCE(632);
      if (lookahead == '0') ADVANCE(945);
      if (lookahead == '<') ADVANCE(612);
      if (lookahead == '>') ADVANCE(618);
      if (lookahead == '?') ADVANCE(760);
      if (lookahead == '@') ADVANCE(941);
      if (lookahead == '\\') ADVANCE(382);
      if (lookahead == '_') ADVANCE(948);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(283)
      if (('1' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(879);
      END_STATE();
    case 284:
      if (lookahead == '!') ADVANCE(669);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(780);
      if (lookahead == '&') ADVANCE(327);
      if (lookahead == '*') ADVANCE(637);
      if (lookahead == '-') ADVANCE(632);
      if (lookahead == '0') ADVANCE(945);
      if (lookahead == '<') ADVANCE(613);
      if (lookahead == '>') ADVANCE(618);
      if (lookahead == '?') ADVANCE(760);
      if (lookahead == '@') ADVANCE(941);
      if (lookahead == '\\') ADVANCE(402);
      if (lookahead == ']') ADVANCE(675);
      if (lookahead == '_') ADVANCE(948);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(284)
      if (('1' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(879);
      END_STATE();
    case 285:
      if (lookahead == '!') ADVANCE(669);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(780);
      if (lookahead == '&') ADVANCE(327);
      if (lookahead == '*') ADVANCE(637);
      if (lookahead == '-') ADVANCE(632);
      if (lookahead == '0') ADVANCE(945);
      if (lookahead == '<') ADVANCE(613);
      if (lookahead == '>') ADVANCE(618);
      if (lookahead == '?') ADVANCE(760);
      if (lookahead == '@') ADVANCE(941);
      if (lookahead == '\\') ADVANCE(403);
      if (lookahead == '_') ADVANCE(948);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(285)
      if (('1' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(879);
      END_STATE();
    case 286:
      if (lookahead == '!') ADVANCE(669);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(780);
      if (lookahead == ')') ADVANCE(655);
      if (lookahead == '*') ADVANCE(637);
      if (lookahead == '-') ADVANCE(632);
      if (lookahead == '0') ADVANCE(945);
      if (lookahead == '?') ADVANCE(760);
      if (lookahead == '@') ADVANCE(941);
      if (lookahead == '\\') ADVANCE(384);
      if (lookahead == '_') ADVANCE(948);
      if (lookahead == '|') ADVANCE(590);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(286)
      if (('1' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(879);
      END_STATE();
    case 287:
      if (lookahead == '!') ADVANCE(669);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(780);
      if (lookahead == '*') ADVANCE(637);
      if (lookahead == '-') ADVANCE(632);
      if (lookahead == '0') ADVANCE(945);
      if (lookahead == '?') ADVANCE(760);
      if (lookahead == '@') ADVANCE(941);
      if (lookahead == '\\') ADVANCE(386);
      if (lookahead == '_') ADVANCE(948);
      if (lookahead == '\n' ||
          lookahead == '\r') SKIP(287)
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') ADVANCE(793);
      if (('1' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(879);
      if (lookahead != 0 &&
          lookahead != '`') ADVANCE(798);
      END_STATE();
    case 288:
      if (lookahead == '!') ADVANCE(669);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(780);
      if (lookahead == '*') ADVANCE(637);
      if (lookahead == '-') ADVANCE(632);
      if (lookahead == '0') ADVANCE(945);
      if (lookahead == '?') ADVANCE(760);
      if (lookahead == '@') ADVANCE(941);
      if (lookahead == '\\') ADVANCE(385);
      if (lookahead == '_') ADVANCE(948);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(288)
      if (('1' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(879);
      END_STATE();
    case 289:
      if (lookahead == '!') ADVANCE(669);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(862);
      if (lookahead == '$') ADVANCE(785);
      if (lookahead == '&') ADVANCE(327);
      if (lookahead == '(') ADVANCE(653);
      if (lookahead == '+') ADVANCE(771);
      if (lookahead == '-') ADVANCE(769);
      if (lookahead == '0') ADVANCE(808);
      if (lookahead == '<') ADVANCE(612);
      if (lookahead == '>') ADVANCE(618);
      if (lookahead == '[' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') SKIP(464)
      if (lookahead == ']') ADVANCE(675);
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == '|') ADVANCE(593);
      if (lookahead == '~') ADVANCE(772);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(289)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(810);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(879);
      END_STATE();
    case 290:
      if (lookahead == '!') ADVANCE(669);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(780);
      if (lookahead == '*') ADVANCE(637);
      if (lookahead == '-') ADVANCE(632);
      if (lookahead == '0') ADVANCE(945);
      if (lookahead == '?') ADVANCE(760);
      if (lookahead == '@') ADVANCE(941);
      if (lookahead == '\\') ADVANCE(387);
      if (lookahead == '_') ADVANCE(948);
      if (lookahead == '\n' ||
          lookahead == '\r') SKIP(290)
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') ADVANCE(794);
      if (('1' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(879);
      if (lookahead != 0 &&
          lookahead != '"' &&
          lookahead != '`') ADVANCE(798);
      END_STATE();
    case 291:
      if (lookahead == '!') ADVANCE(669);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(780);
      if (lookahead == '*') ADVANCE(637);
      if (lookahead == '-') ADVANCE(632);
      if (lookahead == '0') ADVANCE(945);
      if (lookahead == '?') ADVANCE(760);
      if (lookahead == '@') ADVANCE(941);
      if (lookahead == '\\') ADVANCE(388);
      if (lookahead == '_') ADVANCE(948);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(291)
      if (('1' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(879);
      END_STATE();
    case 292:
      if (lookahead == '!') ADVANCE(669);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(784);
      if (lookahead == '*') ADVANCE(637);
      if (lookahead == '-') ADVANCE(632);
      if (lookahead == '0') ADVANCE(945);
      if (lookahead == '?') ADVANCE(760);
      if (lookahead == '@') ADVANCE(941);
      if (lookahead == '\\') SKIP(467)
      if (lookahead == '_') ADVANCE(948);
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == '}') ADVANCE(818);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(292)
      if (('1' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(879);
      END_STATE();
    case 293:
      if (lookahead == '!') ADVANCE(669);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(784);
      if (lookahead == '*') ADVANCE(637);
      if (lookahead == '-') ADVANCE(632);
      if (lookahead == '0') ADVANCE(945);
      if (lookahead == '?') ADVANCE(760);
      if (lookahead == '@') ADVANCE(941);
      if (lookahead == '\\') SKIP(476)
      if (lookahead == '_') ADVANCE(948);
      if (lookahead == '`') ADVANCE(855);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(293)
      if (('1' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(879);
      END_STATE();
    case 294:
      if (lookahead == '!') ADVANCE(819);
      if (lookahead == '#') ADVANCE(825);
      if (lookahead == '$') ADVANCE(784);
      if (lookahead == '*') ADVANCE(637);
      if (lookahead == '-') ADVANCE(632);
      if (lookahead == '0') ADVANCE(945);
      if (lookahead == '=') ADVANCE(826);
      if (lookahead == '?') ADVANCE(760);
      if (lookahead == '@') ADVANCE(941);
      if (lookahead == '\\') SKIP(467)
      if (lookahead == '_') ADVANCE(948);
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == '}') ADVANCE(818);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(292)
      if (('1' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(879);
      END_STATE();
    case 295:
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(862);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(346);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(654);
      if (lookahead == '-') ADVANCE(950);
      if (lookahead == '0') ADVANCE(803);
      if (lookahead == '<') ADVANCE(608);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '[') ADVANCE(674);
      if (lookahead == '\\') ADVANCE(439);
      if (lookahead == ']' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '`') ADVANCE(855);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(295)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(805);
      if (lookahead != 0 &&
          lookahead != ')' &&
          lookahead != ';' &&
          lookahead != '|') ADVANCE(958);
      END_STATE();
    case 296:
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(862);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(346);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(332);
      if (lookahead == ')') ADVANCE(655);
      if (lookahead == '-') ADVANCE(950);
      if (lookahead == '0') ADVANCE(803);
      if (lookahead == '<') ADVANCE(608);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '[' ||
          lookahead == ']' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') ADVANCE(500);
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == '|') ADVANCE(590);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(296)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(805);
      if (lookahead != 0 &&
          lookahead != ';') ADVANCE(958);
      END_STATE();
    case 297:
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(862);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(346);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(332);
      if (lookahead == ')') ADVANCE(655);
      if (lookahead == '-') ADVANCE(950);
      if (lookahead == '0') ADVANCE(803);
      if (lookahead == '<') ADVANCE(608);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '[' ||
          lookahead == ']' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') ADVANCE(511);
      if (lookahead == '`') ADVANCE(856);
      if (lookahead == '|') ADVANCE(590);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(297)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(805);
      if (lookahead != 0 &&
          lookahead != ';') ADVANCE(958);
      END_STATE();
    case 298:
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(862);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(327);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(654);
      if (lookahead == ')') ADVANCE(655);
      if (lookahead == '-') ADVANCE(950);
      if (lookahead == '0') ADVANCE(803);
      if (lookahead == '<') ADVANCE(609);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '[' ||
          lookahead == ']' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') ADVANCE(445);
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(298)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(805);
      if (lookahead != 0 &&
          lookahead != ';') ADVANCE(958);
      END_STATE();
    case 299:
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(862);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(327);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(654);
      if (lookahead == '-') ADVANCE(950);
      if (lookahead == '0') ADVANCE(803);
      if (lookahead == '<') ADVANCE(609);
      if (lookahead == '=') ADVANCE(952);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '[' ||
          lookahead == ']' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') ADVANCE(438);
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(299)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(805);
      if (lookahead != 0 &&
          lookahead != ')' &&
          lookahead != ';') ADVANCE(958);
      END_STATE();
    case 300:
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(862);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(327);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(654);
      if (lookahead == '-') ADVANCE(950);
      if (lookahead == '0') ADVANCE(803);
      if (lookahead == '<') ADVANCE(609);
      if (lookahead == '=') ADVANCE(952);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '[' ||
          lookahead == ']' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') ADVANCE(495);
      if (lookahead == '`') ADVANCE(856);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(300)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(805);
      if (lookahead != 0 &&
          lookahead != ')' &&
          lookahead != ';') ADVANCE(958);
      END_STATE();
    case 301:
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(862);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(327);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(332);
      if (lookahead == ')') ADVANCE(655);
      if (lookahead == '-') ADVANCE(950);
      if (lookahead == '0') ADVANCE(803);
      if (lookahead == '<') ADVANCE(609);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '[' ||
          lookahead == ']' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') ADVANCE(510);
      if (lookahead == '`') ADVANCE(856);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(301)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(805);
      if (lookahead != 0 &&
          lookahead != ';') ADVANCE(958);
      END_STATE();
    case 302:
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(862);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(327);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(332);
      if (lookahead == '-') ADVANCE(950);
      if (lookahead == '0') ADVANCE(803);
      if (lookahead == '<') ADVANCE(609);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '[' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') ADVANCE(444);
      if (lookahead == ']') ADVANCE(675);
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(302)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(805);
      if (lookahead != 0 &&
          lookahead != ')' &&
          lookahead != ';') ADVANCE(958);
      END_STATE();
    case 303:
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(862);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(327);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(332);
      if (lookahead == '-') ADVANCE(950);
      if (lookahead == '0') ADVANCE(803);
      if (lookahead == '<') ADVANCE(609);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '[' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') ADVANCE(499);
      if (lookahead == ']') ADVANCE(675);
      if (lookahead == '`') ADVANCE(856);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(303)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(805);
      if (lookahead != 0 &&
          lookahead != ')' &&
          lookahead != ';') ADVANCE(958);
      END_STATE();
    case 304:
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(862);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(327);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(332);
      if (lookahead == '-') ADVANCE(950);
      if (lookahead == '0') ADVANCE(803);
      if (lookahead == '<') ADVANCE(610);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '[' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') ADVANCE(509);
      if (lookahead == ']') ADVANCE(675);
      if (lookahead == '`') ADVANCE(856);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(304)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(805);
      if (lookahead != 0 &&
          lookahead != ')' &&
          lookahead != ';') ADVANCE(958);
      END_STATE();
    case 305:
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(862);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(327);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(332);
      if (lookahead == '-') ADVANCE(950);
      if (lookahead == '0') ADVANCE(803);
      if (lookahead == '<') ADVANCE(610);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '[' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') ADVANCE(515);
      if (lookahead == ']') ADVANCE(675);
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(305)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(805);
      if (lookahead != 0 &&
          lookahead != ')' &&
          lookahead != ';') ADVANCE(958);
      END_STATE();
    case 306:
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(862);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(327);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(332);
      if (lookahead == '-') ADVANCE(950);
      if (lookahead == '0') ADVANCE(803);
      if (lookahead == '<') ADVANCE(610);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '[' ||
          lookahead == ']' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') ADVANCE(446);
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(306)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(805);
      if (lookahead != 0 &&
          lookahead != ')' &&
          lookahead != ';') ADVANCE(958);
      END_STATE();
    case 307:
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(862);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(327);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(332);
      if (lookahead == '-') ADVANCE(950);
      if (lookahead == '0') ADVANCE(803);
      if (lookahead == '<') ADVANCE(610);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '[' ||
          lookahead == ']' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') ADVANCE(501);
      if (lookahead == '`') ADVANCE(856);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(307)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(805);
      if (lookahead != 0 &&
          lookahead != ')' &&
          lookahead != ';') ADVANCE(958);
      END_STATE();
    case 308:
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(862);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(327);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(332);
      if (lookahead == '-') ADVANCE(950);
      if (lookahead == '0') ADVANCE(804);
      if (lookahead == '<') ADVANCE(610);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '[' ||
          lookahead == ']' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') ADVANCE(443);
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(308)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(806);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(868);
      if (lookahead != 0 &&
          lookahead != ')' &&
          lookahead != ';') ADVANCE(958);
      END_STATE();
    case 309:
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(862);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(327);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(332);
      if (lookahead == '-') ADVANCE(950);
      if (lookahead == '0') ADVANCE(804);
      if (lookahead == '<') ADVANCE(610);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '[' ||
          lookahead == ']' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') ADVANCE(498);
      if (lookahead == '`') ADVANCE(856);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(309)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(806);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(868);
      if (lookahead != 0 &&
          lookahead != ')' &&
          lookahead != ';') ADVANCE(958);
      END_STATE();
    case 310:
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(862);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(654);
      if (lookahead == '-') ADVANCE(950);
      if (lookahead == '0') ADVANCE(803);
      if (lookahead == '<') ADVANCE(333);
      if (lookahead == '>') ADVANCE(334);
      if (lookahead == '[' ||
          lookahead == ']' ||
          lookahead == '{') ADVANCE(786);
      if (lookahead == '\\') ADVANCE(460);
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == '}') ADVANCE(818);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(310)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(805);
      if (lookahead != 0 &&
          (lookahead < '&' || ')' < lookahead) &&
          lookahead != ';' &&
          lookahead != '|') ADVANCE(958);
      END_STATE();
    case 311:
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(862);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(654);
      if (lookahead == '-') ADVANCE(950);
      if (lookahead == '0') ADVANCE(803);
      if (lookahead == '<') ADVANCE(333);
      if (lookahead == '>') ADVANCE(334);
      if (lookahead == '[' ||
          lookahead == ']' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') ADVANCE(456);
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == 'e') ADVANCE(956);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(311)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(805);
      if (lookahead != 0 &&
          (lookahead < '&' || ')' < lookahead) &&
          lookahead != ';' &&
          lookahead != '|') ADVANCE(958);
      END_STATE();
    case 312:
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(862);
      if (lookahead == '$') ADVANCE(785);
      if (lookahead == '&') ADVANCE(327);
      if (lookahead == '(') ADVANCE(653);
      if (lookahead == ')') ADVANCE(337);
      if (lookahead == '+') ADVANCE(339);
      if (lookahead == '-') ADVANCE(340);
      if (lookahead == '0') ADVANCE(809);
      if (lookahead == '<') ADVANCE(613);
      if (lookahead == '=') ADVANCE(559);
      if (lookahead == '>') ADVANCE(618);
      if (lookahead == '[' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') SKIP(468)
      if (lookahead == ']') ADVANCE(675);
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(312)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(811);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(656);
      END_STATE();
    case 313:
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(862);
      if (lookahead == '$') ADVANCE(782);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(653);
      if (lookahead == ')') ADVANCE(655);
      if (lookahead == ';') ADVANCE(553);
      if (lookahead == '<') ADVANCE(333);
      if (lookahead == '>') ADVANCE(334);
      if (lookahead == '\\') ADVANCE(474);
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == '{') ADVANCE(666);
      if (lookahead == '|') ADVANCE(590);
      if (lookahead == '}') ADVANCE(818);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(313)
      if (lookahead != 0 &&
          lookahead != '&' &&
          (lookahead < '[' || ']' < lookahead)) ADVANCE(958);
      END_STATE();
    case 314:
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(862);
      if (lookahead == '$') ADVANCE(335);
      if (lookahead == '%') ADVANCE(646);
      if (lookahead == '*') ADVANCE(637);
      if (lookahead == '+') ADVANCE(629);
      if (lookahead == '-') ADVANCE(632);
      if (lookahead == '/') ADVANCE(641);
      if (lookahead == ':') ADVANCE(762);
      if (lookahead == '[' ||
          lookahead == ']' ||
          lookahead == '{') ADVANCE(786);
      if (lookahead == '\\') SKIP(479)
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == '}') ADVANCE(818);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(314)
      END_STATE();
    case 315:
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(862);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == ')') ADVANCE(655);
      if (lookahead == '\\') SKIP(481)
      if (lookahead == '}') ADVANCE(818);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(842);
      END_STATE();
    case 316:
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(863);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(654);
      if (lookahead == '-') ADVANCE(950);
      if (lookahead == '0') ADVANCE(803);
      if (lookahead == '<') ADVANCE(333);
      if (lookahead == '>') ADVANCE(334);
      if (lookahead == '[' ||
          lookahead == ']' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') ADVANCE(461);
      if (lookahead == '`') ADVANCE(855);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(316)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(805);
      if (lookahead != 0 &&
          (lookahead < '&' || ')' < lookahead) &&
          lookahead != ';' &&
          lookahead != '|') ADVANCE(958);
      END_STATE();
    case 317:
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(863);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(332);
      if (lookahead == '-') ADVANCE(950);
      if (lookahead == '0') ADVANCE(803);
      if (lookahead == '<') ADVANCE(333);
      if (lookahead == '>') ADVANCE(334);
      if (lookahead == '[' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') ADVANCE(463);
      if (lookahead == ']') ADVANCE(675);
      if (lookahead == '`') ADVANCE(855);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(317)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(805);
      if (lookahead != 0 &&
          (lookahead < '&' || ')' < lookahead) &&
          lookahead != ';' &&
          lookahead != '|') ADVANCE(958);
      END_STATE();
    case 318:
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(797);
      if (lookahead == '$') ADVANCE(783);
      if (lookahead == '(') ADVANCE(796);
      if (lookahead == '\\') ADVANCE(477);
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == '\n' ||
          lookahead == '\r') SKIP(318)
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') ADVANCE(795);
      if (lookahead != 0) ADVANCE(798);
      END_STATE();
    case 319:
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '%') ADVANCE(648);
      if (lookahead == '*') ADVANCE(823);
      if (lookahead == '+') ADVANCE(833);
      if (lookahead == ',') ADVANCE(558);
      if (lookahead == '-') ADVANCE(830);
      if (lookahead == '/') ADVANCE(643);
      if (lookahead == ':') ADVANCE(764);
      if (lookahead == '=') ADVANCE(826);
      if (lookahead == '?') ADVANCE(836);
      if (lookahead == '@') ADVANCE(821);
      if (lookahead == '[') ADVANCE(673);
      if (lookahead == '\\') SKIP(462)
      if (lookahead == '^') ADVANCE(598);
      if (lookahead == '}') ADVANCE(818);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(320)
      END_STATE();
    case 320:
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '%') ADVANCE(648);
      if (lookahead == ',') ADVANCE(558);
      if (lookahead == '/') ADVANCE(643);
      if (lookahead == ':') ADVANCE(762);
      if (lookahead == '[') ADVANCE(673);
      if (lookahead == '\\') SKIP(462)
      if (lookahead == '^') ADVANCE(598);
      if (lookahead == '}') ADVANCE(818);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(320)
      END_STATE();
    case 321:
      if (lookahead == '#') ADVANCE(862);
      if (lookahead == '$') ADVANCE(353);
      if (lookahead == '&') ADVANCE(327);
      if (lookahead == '-') ADVANCE(342);
      if (lookahead == '0') ADVANCE(808);
      if (lookahead == '<') ADVANCE(613);
      if (lookahead == '>') ADVANCE(618);
      if (lookahead == '[' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') SKIP(470)
      if (lookahead == ']') ADVANCE(675);
      if (lookahead == '`') ADVANCE(348);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(321)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(810);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(879);
      END_STATE();
    case 322:
      if (lookahead == '#') ADVANCE(862);
      if (lookahead == '%') ADVANCE(646);
      if (lookahead == '&') ADVANCE(327);
      if (lookahead == '*') ADVANCE(637);
      if (lookahead == '+') ADVANCE(629);
      if (lookahead == '-') ADVANCE(632);
      if (lookahead == '/') ADVANCE(641);
      if (lookahead == '<') ADVANCE(613);
      if (lookahead == '>') ADVANCE(618);
      if (lookahead == '[' ||
          lookahead == ']' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') SKIP(472)
      if (lookahead == '`') ADVANCE(348);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(322)
      END_STATE();
    case 323:
      if (lookahead == '#') ADVANCE(862);
      if (lookahead == '&') ADVANCE(346);
      if (lookahead == '(') ADVANCE(654);
      if (lookahead == '<') ADVANCE(611);
      if (lookahead == '>') ADVANCE(618);
      if (lookahead == '[') ADVANCE(674);
      if (lookahead == '\\') ADVANCE(478);
      if (lookahead == '{') ADVANCE(666);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(323)
      if (lookahead != 0 &&
          (lookahead < '"' || '$' < lookahead) &&
          (lookahead < '\'' || ')' < lookahead) &&
          lookahead != ';' &&
          lookahead != ']' &&
          lookahead != '`' &&
          lookahead != '|' &&
          lookahead != '}') ADVANCE(958);
      END_STATE();
    case 324:
      if (lookahead == '#') ADVANCE(862);
      if (lookahead == '&') ADVANCE(327);
      if (lookahead == '<') ADVANCE(612);
      if (lookahead == '>') ADVANCE(618);
      if (lookahead == '[' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') ADVANCE(469);
      if (lookahead == ']') ADVANCE(675);
      if (lookahead == '`') ADVANCE(348);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(324)
      if (lookahead != 0 &&
          (lookahead < '"' || '$' < lookahead) &&
          (lookahead < '\'' || ')' < lookahead) &&
          lookahead != ';') ADVANCE(958);
      END_STATE();
    case 325:
      if (lookahead == '#') ADVANCE(862);
      if (lookahead == '&') ADVANCE(327);
      if (lookahead == '<') ADVANCE(612);
      if (lookahead == '>') ADVANCE(618);
      if (lookahead == '[' ||
          lookahead == ']' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') SKIP(473)
      if (lookahead == '`') ADVANCE(348);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(325)
      END_STATE();
    case 326:
      if (lookahead == '#') ADVANCE(862);
      if (lookahead == '+') ADVANCE(345);
      if (lookahead == '/') ADVANCE(641);
      if (lookahead == '=') ADVANCE(559);
      if (lookahead == '[') ADVANCE(673);
      if (lookahead == '\\') SKIP(482)
      if (lookahead == '`') ADVANCE(348);
      if (lookahead == '}') ADVANCE(818);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(326)
      END_STATE();
    case 327:
      if (lookahead == '&') ADVANCE(588);
      if (lookahead == '>') ADVANCE(680);
      END_STATE();
    case 328:
      if (lookahead == '&') ADVANCE(664);
      if (lookahead == ';') ADVANCE(663);
      END_STATE();
    case 329:
      if (lookahead == '\'') ADVANCE(800);
      if (lookahead != 0) ADVANCE(329);
      END_STATE();
    case 330:
      if (lookahead == '\'') ADVANCE(801);
      if (lookahead == '\\') ADVANCE(331);
      if (lookahead != 0) ADVANCE(330);
      END_STATE();
    case 331:
      if (lookahead == '\'') ADVANCE(802);
      if (lookahead == '\\') ADVANCE(331);
      if (lookahead != 0) ADVANCE(330);
      END_STATE();
    case 332:
      if (lookahead == '(') ADVANCE(551);
      END_STATE();
    case 333:
      if (lookahead == '(') ADVANCE(858);
      END_STATE();
    case 334:
      if (lookahead == '(') ADVANCE(859);
      END_STATE();
    case 335:
      if (lookahead == '(') ADVANCE(853);
      if (lookahead == '`') ADVANCE(857);
      END_STATE();
    case 336:
      if (lookahead == '(') ADVANCE(853);
      if (lookahead == '`') ADVANCE(857);
      if (lookahead == '{') ADVANCE(817);
      END_STATE();
    case 337:
      if (lookahead == ')') ADVANCE(552);
      END_STATE();
    case 338:
      if (lookahead == '+') ADVANCE(564);
      END_STATE();
    case 339:
      if (lookahead == '+') ADVANCE(564);
      if (lookahead == '=') ADVANCE(568);
      END_STATE();
    case 340:
      if (lookahead == '-') ADVANCE(566);
      if (lookahead == '0') ADVANCE(809);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(811);
      END_STATE();
    case 341:
      if (lookahead == '.') ADVANCE(776);
      END_STATE();
    case 342:
      if (lookahead == '0') ADVANCE(809);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(811);
      END_STATE();
    case 343:
      if (lookahead == '<') ADVANCE(759);
      END_STATE();
    case 344:
      if (lookahead == '=') ADVANCE(605);
      END_STATE();
    case 345:
      if (lookahead == '=') ADVANCE(568);
      END_STATE();
    case 346:
      if (lookahead == '>') ADVANCE(680);
      END_STATE();
    case 347:
      if (lookahead == ']') ADVANCE(677);
      END_STATE();
    case 348:
      if (lookahead == '`') ADVANCE(779);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(348);
      END_STATE();
    case 349:
      if (lookahead == 'a') ADVANCE(350);
      END_STATE();
    case 350:
      if (lookahead == 'c') ADVANCE(657);
      END_STATE();
    case 351:
      if (lookahead == 'n') ADVANCE(548);
      END_STATE();
    case 352:
      if (lookahead == 's') ADVANCE(349);
      END_STATE();
    case 353:
      if (lookahead == '{') ADVANCE(817);
      END_STATE();
    case 354:
      if (lookahead == '|') ADVANCE(586);
      END_STATE();
    case 355:
      if (lookahead == '\t' ||
          lookahead == 11 ||
          lookahead == '\f' ||
          lookahead == ' ') SKIP(12)
      if (lookahead == '\n') ADVANCE(880);
      if (lookahead == '\r') ADVANCE(13);
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 356:
      if (lookahead == '\t' ||
          lookahead == 11 ||
          lookahead == '\f' ||
          lookahead == ' ') SKIP(240)
      if (lookahead == '\n') ADVANCE(918);
      if (lookahead == '\r') ADVANCE(17);
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 357:
      if (lookahead == '\t' ||
          lookahead == 11 ||
          lookahead == '\f' ||
          lookahead == ' ') SKIP(163)
      if (lookahead == '\n') ADVANCE(882);
      if (lookahead == '\r') ADVANCE(24);
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 358:
      if (lookahead == '\t' ||
          lookahead == 11 ||
          lookahead == '\f' ||
          lookahead == ' ') SKIP(176)
      if (lookahead == '\n') ADVANCE(884);
      if (lookahead == '\r') ADVANCE(26);
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 359:
      if (lookahead == '\t' ||
          lookahead == 11 ||
          lookahead == '\f' ||
          lookahead == ' ') SKIP(182)
      if (lookahead == '\n') ADVANCE(888);
      if (lookahead == '\r') ADVANCE(27);
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 360:
      if (lookahead == '\t' ||
          lookahead == 11 ||
          lookahead == '\f' ||
          lookahead == ' ') SKIP(197)
      if (lookahead == '\n') ADVANCE(898);
      if (lookahead == '\r') ADVANCE(31);
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 361:
      if (lookahead == '\t' ||
          lookahead == 11 ||
          lookahead == '\f' ||
          lookahead == ' ') SKIP(203)
      if (lookahead == '\n') ADVANCE(902);
      if (lookahead == '\r') ADVANCE(36);
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 362:
      if (lookahead == '\t' ||
          lookahead == 11 ||
          lookahead == '\f' ||
          lookahead == ' ') SKIP(206)
      if (lookahead == '\n') ADVANCE(904);
      if (lookahead == '\r') ADVANCE(38);
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 363:
      if (lookahead == '\t' ||
          lookahead == 11 ||
          lookahead == '\f' ||
          lookahead == ' ') SKIP(214)
      if (lookahead == '\n') ADVANCE(909);
      if (lookahead == '\r') ADVANCE(39);
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 364:
      if (lookahead == '\t' ||
          lookahead == 11 ||
          lookahead == '\f' ||
          lookahead == ' ') SKIP(248)
      if (lookahead == '\n') ADVANCE(920);
      if (lookahead == '\r') ADVANCE(41);
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 365:
      if (lookahead == '\t' ||
          lookahead == 11 ||
          lookahead == '\f' ||
          lookahead == ' ') SKIP(250)
      if (lookahead == '\n') ADVANCE(922);
      if (lookahead == '\r') ADVANCE(44);
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 366:
      if (lookahead == '\t' ||
          lookahead == 11 ||
          lookahead == '\f' ||
          lookahead == ' ') SKIP(222)
      if (lookahead == '\n') ADVANCE(913);
      if (lookahead == '\r') ADVANCE(46);
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 367:
      if (lookahead == '\t' ||
          lookahead == 11 ||
          lookahead == '\f' ||
          lookahead == ' ') SKIP(268)
      if (lookahead == '\n') ADVANCE(930);
      if (lookahead == '\r') ADVANCE(47);
      END_STATE();
    case 368:
      if (lookahead == '\t' ||
          lookahead == 11 ||
          lookahead == '\f' ||
          lookahead == ' ') SKIP(251)
      if (lookahead == '\n') ADVANCE(923);
      if (lookahead == '\r') ADVANCE(48);
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 369:
      if (lookahead == '\t' ||
          lookahead == 11 ||
          lookahead == '\f' ||
          lookahead == ' ') SKIP(267)
      if (lookahead == '\n') ADVANCE(929);
      if (lookahead == '\r') ADVANCE(49);
      END_STATE();
    case 370:
      if (lookahead == '\t' ||
          lookahead == 11 ||
          lookahead == '\f' ||
          lookahead == ' ') SKIP(265)
      if (lookahead == '\n') ADVANCE(927);
      if (lookahead == '\r') ADVANCE(51);
      END_STATE();
    case 371:
      if (lookahead == '\t' ||
          lookahead == 11 ||
          lookahead == '\f' ||
          lookahead == ' ') SKIP(270)
      if (lookahead == '\n') ADVANCE(932);
      if (lookahead == '\r') ADVANCE(52);
      END_STATE();
    case 372:
      if (lookahead == '\t' ||
          lookahead == 11 ||
          lookahead == '\f' ||
          lookahead == ' ') SKIP(247)
      if (lookahead == '\n') ADVANCE(919);
      if (lookahead == '\r') ADVANCE(56);
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 373:
      if (lookahead == '\t' ||
          lookahead == 11 ||
          lookahead == '\f' ||
          lookahead == ' ') SKIP(150)
      if (lookahead == '\n') ADVANCE(881);
      if (lookahead == '\r') ADVANCE(62);
      END_STATE();
    case 374:
      if (lookahead == '\t' ||
          lookahead == 11 ||
          lookahead == '\f' ||
          lookahead == ' ') SKIP(175)
      if (lookahead == '\n') ADVANCE(883);
      if (lookahead == '\r') ADVANCE(71);
      END_STATE();
    case 375:
      if (lookahead == '\t' ||
          lookahead == 11 ||
          lookahead == '\f' ||
          lookahead == ' ') SKIP(184)
      if (lookahead == '\n') ADVANCE(889);
      if (lookahead == '\r') ADVANCE(75);
      END_STATE();
    case 376:
      if (lookahead == '\t' ||
          lookahead == 11 ||
          lookahead == '\f' ||
          lookahead == ' ') SKIP(187)
      if (lookahead == '\n') ADVANCE(891);
      if (lookahead == '\r') ADVANCE(76);
      END_STATE();
    case 377:
      if (lookahead == '\t' ||
          lookahead == 11 ||
          lookahead == '\f' ||
          lookahead == ' ') SKIP(228)
      if (lookahead == '\n') ADVANCE(916);
      if (lookahead == '\r') ADVANCE(77);
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 378:
      if (lookahead == '\t' ||
          lookahead == 11 ||
          lookahead == '\f' ||
          lookahead == ' ') SKIP(202)
      if (lookahead == '\n') ADVANCE(901);
      if (lookahead == '\r') ADVANCE(82);
      END_STATE();
    case 379:
      if (lookahead == '\t' ||
          lookahead == 11 ||
          lookahead == '\f' ||
          lookahead == ' ') SKIP(282)
      if (lookahead == '\n') ADVANCE(933);
      if (lookahead == '\r') ADVANCE(83);
      END_STATE();
    case 380:
      if (lookahead == '\t' ||
          lookahead == 11 ||
          lookahead == '\f' ||
          lookahead == ' ') SKIP(254)
      if (lookahead == '\n') ADVANCE(926);
      if (lookahead == '\r') ADVANCE(84);
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 381:
      if (lookahead == '\t' ||
          lookahead == 11 ||
          lookahead == '\f' ||
          lookahead == ' ') SKIP(230)
      if (lookahead == '\n') ADVANCE(917);
      if (lookahead == '\r') ADVANCE(85);
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 382:
      if (lookahead == '\t' ||
          lookahead == 11 ||
          lookahead == '\f' ||
          lookahead == ' ') SKIP(283)
      if (lookahead == '\n') ADVANCE(934);
      if (lookahead == '\r') ADVANCE(86);
      END_STATE();
    case 383:
      if (lookahead == '\t' ||
          lookahead == 11 ||
          lookahead == '\f' ||
          lookahead == ' ') SKIP(211)
      if (lookahead == '\n') ADVANCE(907);
      if (lookahead == '\r') ADVANCE(99);
      END_STATE();
    case 384:
      if (lookahead == '\t' ||
          lookahead == 11 ||
          lookahead == '\f' ||
          lookahead == ' ') SKIP(286)
      if (lookahead == '\n') ADVANCE(937);
      if (lookahead == '\r') ADVANCE(102);
      END_STATE();
    case 385:
      if (lookahead == '\t' ||
          lookahead == 11 ||
          lookahead == '\f' ||
          lookahead == ' ') SKIP(288)
      if (lookahead == '\n') ADVANCE(938);
      if (lookahead == '\r') ADVANCE(103);
      END_STATE();
    case 386:
      if (lookahead == '\t' ||
          lookahead == 11 ||
          lookahead == '\f' ||
          lookahead == ' ') ADVANCE(793);
      if (lookahead == '\n') ADVANCE(793);
      if (lookahead == '\r') ADVANCE(790);
      if (lookahead != 0) ADVANCE(798);
      END_STATE();
    case 387:
      if (lookahead == '\t' ||
          lookahead == 11 ||
          lookahead == '\f' ||
          lookahead == ' ') ADVANCE(794);
      if (lookahead == '\n') ADVANCE(794);
      if (lookahead == '\r') ADVANCE(791);
      if (lookahead != 0) ADVANCE(798);
      END_STATE();
    case 388:
      if (lookahead == '\t' ||
          lookahead == 11 ||
          lookahead == '\f' ||
          lookahead == ' ') SKIP(291)
      if (lookahead == '\n') ADVANCE(939);
      if (lookahead == '\r') ADVANCE(104);
      END_STATE();
    case 389:
      if (lookahead == '\t' ||
          lookahead == 11 ||
          lookahead == '\f' ||
          lookahead == ' ') SKIP(179)
      if (lookahead == '\n') ADVANCE(886);
      if (lookahead == '\r') ADVANCE(118);
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 390:
      if (lookahead == '\t' ||
          lookahead == 11 ||
          lookahead == '\f' ||
          lookahead == ' ') SKIP(185)
      if (lookahead == '\n') ADVANCE(890);
      if (lookahead == '\r') ADVANCE(120);
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 391:
      if (lookahead == '\t' ||
          lookahead == 11 ||
          lookahead == '\f' ||
          lookahead == ' ') SKIP(191)
      if (lookahead == '\n') ADVANCE(894);
      if (lookahead == '\r') ADVANCE(121);
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 392:
      if (lookahead == '\t' ||
          lookahead == 11 ||
          lookahead == '\f' ||
          lookahead == ' ') SKIP(212)
      if (lookahead == '\n') ADVANCE(908);
      if (lookahead == '\r') ADVANCE(128);
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 393:
      if (lookahead == '\t' ||
          lookahead == 11 ||
          lookahead == '\f' ||
          lookahead == ' ') SKIP(220)
      if (lookahead == '\n') ADVANCE(912);
      if (lookahead == '\r') ADVANCE(129);
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 394:
      if (lookahead == '\t' ||
          lookahead == 11 ||
          lookahead == '\f' ||
          lookahead == ' ') SKIP(249)
      if (lookahead == '\n') ADVANCE(921);
      if (lookahead == '\r') ADVANCE(131);
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 395:
      if (lookahead == '\t' ||
          lookahead == 11 ||
          lookahead == '\f' ||
          lookahead == ' ') SKIP(226)
      if (lookahead == '\n') ADVANCE(915);
      if (lookahead == '\r') ADVANCE(135);
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 396:
      if (lookahead == '\t' ||
          lookahead == 11 ||
          lookahead == '\f' ||
          lookahead == ' ') SKIP(269)
      if (lookahead == '\n') ADVANCE(931);
      if (lookahead == '\r') ADVANCE(136);
      END_STATE();
    case 397:
      if (lookahead == '\t' ||
          lookahead == 11 ||
          lookahead == '\f' ||
          lookahead == ' ') SKIP(266)
      if (lookahead == '\n') ADVANCE(928);
      if (lookahead == '\r') ADVANCE(138);
      END_STATE();
    case 398:
      if (lookahead == '\t' ||
          lookahead == 11 ||
          lookahead == '\f' ||
          lookahead == ' ') SKIP(181)
      if (lookahead == '\n') ADVANCE(887);
      if (lookahead == '\r') ADVANCE(140);
      END_STATE();
    case 399:
      if (lookahead == '\t' ||
          lookahead == 11 ||
          lookahead == '\f' ||
          lookahead == ' ') SKIP(193)
      if (lookahead == '\n') ADVANCE(895);
      if (lookahead == '\r') ADVANCE(142);
      END_STATE();
    case 400:
      if (lookahead == '\t' ||
          lookahead == 11 ||
          lookahead == '\f' ||
          lookahead == ' ') SKIP(199)
      if (lookahead == '\n') ADVANCE(899);
      if (lookahead == '\r') ADVANCE(143);
      END_STATE();
    case 401:
      if (lookahead == '\t' ||
          lookahead == 11 ||
          lookahead == '\f' ||
          lookahead == ' ') SKIP(205)
      if (lookahead == '\n') ADVANCE(903);
      if (lookahead == '\r') ADVANCE(145);
      END_STATE();
    case 402:
      if (lookahead == '\t' ||
          lookahead == 11 ||
          lookahead == '\f' ||
          lookahead == ' ') SKIP(284)
      if (lookahead == '\n') ADVANCE(935);
      if (lookahead == '\r') ADVANCE(146);
      END_STATE();
    case 403:
      if (lookahead == '\t' ||
          lookahead == 11 ||
          lookahead == '\f' ||
          lookahead == ' ') SKIP(285)
      if (lookahead == '\n') ADVANCE(936);
      if (lookahead == '\r') ADVANCE(147);
      END_STATE();
    case 404:
      if (lookahead == '\t' ||
          lookahead == 11 ||
          lookahead == '\f' ||
          lookahead == ' ') SKIP(188)
      if (lookahead == '\n') ADVANCE(892);
      if (lookahead == '\r') ADVANCE(151);
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 405:
      if (lookahead == '\t' ||
          lookahead == 11 ||
          lookahead == '\f' ||
          lookahead == ' ') SKIP(218)
      if (lookahead == '\n') ADVANCE(911);
      if (lookahead == '\r') ADVANCE(155);
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 406:
      if (lookahead == '\t' ||
          lookahead == 11 ||
          lookahead == '\f' ||
          lookahead == ' ') SKIP(190)
      if (lookahead == '\n') ADVANCE(893);
      if (lookahead == '\r') ADVANCE(158);
      END_STATE();
    case 407:
      if (lookahead == '\t' ||
          lookahead == 11 ||
          lookahead == '\f' ||
          lookahead == ' ') SKIP(208)
      if (lookahead == '\n') ADVANCE(905);
      if (lookahead == '\r') ADVANCE(159);
      END_STATE();
    case 408:
      if (lookahead == '\t' ||
          lookahead == 11 ||
          lookahead == '\f' ||
          lookahead == ' ') SKIP(194)
      if (lookahead == '\n') ADVANCE(896);
      if (lookahead == '\r') ADVANCE(160);
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 409:
      if (lookahead == '\t' ||
          lookahead == 11 ||
          lookahead == '\f' ||
          lookahead == ' ') SKIP(196)
      if (lookahead == '\n') ADVANCE(897);
      if (lookahead == '\r') ADVANCE(162);
      END_STATE();
    case 410:
      if (lookahead == '\t' ||
          lookahead == 11 ||
          lookahead == '\f' ||
          lookahead == ' ') SKIP(200)
      if (lookahead == '\n') ADVANCE(900);
      if (lookahead == '\r') ADVANCE(165);
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 411:
      if (lookahead == '\t' ||
          lookahead == 11 ||
          lookahead == '\f' ||
          lookahead == ' ') SKIP(209)
      if (lookahead == '\n') ADVANCE(906);
      if (lookahead == '\r') ADVANCE(168);
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 412:
      if (lookahead == '\t' ||
          lookahead == 11 ||
          lookahead == '\f' ||
          lookahead == ' ') SKIP(216)
      if (lookahead == '\n') ADVANCE(910);
      if (lookahead == '\r') ADVANCE(169);
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 413:
      if (lookahead == '\t' ||
          lookahead == 11 ||
          lookahead == '\f' ||
          lookahead == ' ') SKIP(252)
      if (lookahead == '\n') ADVANCE(924);
      if (lookahead == '\r') ADVANCE(171);
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 414:
      if (lookahead == '\t' ||
          lookahead == 11 ||
          lookahead == '\f' ||
          lookahead == ' ') SKIP(224)
      if (lookahead == '\n') ADVANCE(914);
      if (lookahead == '\r') ADVANCE(172);
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 415:
      if (lookahead == '\t' ||
          lookahead == 11 ||
          lookahead == '\f' ||
          lookahead == ' ') SKIP(253)
      if (lookahead == '\n') ADVANCE(925);
      if (lookahead == '\r') ADVANCE(173);
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 416:
      if (lookahead == '\t' ||
          lookahead == 11 ||
          lookahead == '\f' ||
          lookahead == ' ') SKIP(178)
      if (lookahead == '\n') ADVANCE(885);
      if (lookahead == '\r') ADVANCE(139);
      END_STATE();
    case 417:
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') SKIP(239)
      if (lookahead == '\r') SKIP(1)
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 418:
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') SKIP(255)
      if (lookahead == '\r') SKIP(2)
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 419:
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') SKIP(256)
      if (lookahead == '\r') SKIP(3)
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 420:
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') SKIP(257)
      if (lookahead == '\r') SKIP(4)
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 421:
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') SKIP(6)
      if (lookahead == '\r') SKIP(5)
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 422:
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') SKIP(241)
      if (lookahead == '\r') SKIP(7)
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 423:
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') SKIP(259)
      if (lookahead == '\r') SKIP(8)
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 424:
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') SKIP(260)
      if (lookahead == '\r') SKIP(9)
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 425:
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') SKIP(242)
      if (lookahead == '\r') SKIP(10)
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 426:
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') SKIP(244)
      if (lookahead == '\r') SKIP(11)
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 427:
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') SKIP(112)
      if (lookahead == '\r') SKIP(18)
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 428:
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') SKIP(263)
      if (lookahead == '\r') SKIP(19)
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 429:
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') SKIP(258)
      if (lookahead == '\r') SKIP(20)
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 430:
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') SKIP(177)
      if (lookahead == '\r') SKIP(21)
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 431:
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') SKIP(180)
      if (lookahead == '\r') SKIP(22)
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 432:
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') SKIP(183)
      if (lookahead == '\r') SKIP(23)
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 433:
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') SKIP(186)
      if (lookahead == '\r') SKIP(25)
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 434:
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') SKIP(189)
      if (lookahead == '\r') SKIP(28)
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 435:
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') SKIP(192)
      if (lookahead == '\r') SKIP(29)
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 436:
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') SKIP(195)
      if (lookahead == '\r') SKIP(30)
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 437:
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') SKIP(198)
      if (lookahead == '\r') SKIP(32)
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 438:
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') SKIP(299)
      if (lookahead == '\r') SKIP(33)
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 439:
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') SKIP(295)
      if (lookahead == '\r') SKIP(34)
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 440:
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') SKIP(204)
      if (lookahead == '\r') SKIP(35)
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 441:
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') SKIP(262)
      if (lookahead == '\r') SKIP(37)
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 442:
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') SKIP(213)
      if (lookahead == '\r') SKIP(40)
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 443:
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') SKIP(308)
      if (lookahead == '\r') SKIP(42)
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 444:
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') SKIP(302)
      if (lookahead == '\r') SKIP(43)
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 445:
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') SKIP(298)
      if (lookahead == '\r') SKIP(45)
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 446:
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') SKIP(306)
      if (lookahead == '\r') SKIP(50)
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 447:
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') SKIP(272)
      if (lookahead == '\r') SKIP(53)
      END_STATE();
    case 448:
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') SKIP(273)
      if (lookahead == '\r') SKIP(54)
      END_STATE();
    case 449:
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') SKIP(261)
      if (lookahead == '\r') SKIP(55)
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 450:
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') SKIP(278)
      if (lookahead == '\r') SKIP(57)
      END_STATE();
    case 451:
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') SKIP(235)
      if (lookahead == '\r') SKIP(58)
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 452:
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') SKIP(276)
      if (lookahead == '\r') SKIP(59)
      END_STATE();
    case 453:
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') SKIP(281)
      if (lookahead == '\r') SKIP(60)
      END_STATE();
    case 454:
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') SKIP(277)
      if (lookahead == '\r') SKIP(61)
      END_STATE();
    case 455:
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') SKIP(64)
      if (lookahead == '\r') SKIP(63)
      END_STATE();
    case 456:
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') SKIP(311)
      if (lookahead == '\r') SKIP(72)
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 457:
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') SKIP(275)
      if (lookahead == '\r') SKIP(73)
      END_STATE();
    case 458:
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') SKIP(280)
      if (lookahead == '\r') SKIP(74)
      END_STATE();
    case 459:
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') SKIP(237)
      if (lookahead == '\r') SKIP(78)
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 460:
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') SKIP(310)
      if (lookahead == '\r') SKIP(79)
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 461:
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') SKIP(316)
      if (lookahead == '\r') SKIP(80)
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 462:
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') SKIP(320)
      if (lookahead == '\r') SKIP(81)
      END_STATE();
    case 463:
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') SKIP(317)
      if (lookahead == '\r') SKIP(87)
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 464:
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') SKIP(289)
      if (lookahead == '\r') SKIP(88)
      END_STATE();
    case 465:
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') SKIP(65)
      if (lookahead == '\r') SKIP(89)
      END_STATE();
    case 466:
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') SKIP(67)
      if (lookahead == '\r') SKIP(90)
      END_STATE();
    case 467:
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') SKIP(292)
      if (lookahead == '\r') SKIP(91)
      END_STATE();
    case 468:
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') SKIP(312)
      if (lookahead == '\r') SKIP(92)
      END_STATE();
    case 469:
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') SKIP(324)
      if (lookahead == '\r') SKIP(93)
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 470:
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') SKIP(321)
      if (lookahead == '\r') SKIP(94)
      END_STATE();
    case 471:
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') SKIP(66)
      if (lookahead == '\r') SKIP(95)
      END_STATE();
    case 472:
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') SKIP(322)
      if (lookahead == '\r') SKIP(96)
      END_STATE();
    case 473:
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') SKIP(325)
      if (lookahead == '\r') SKIP(97)
      END_STATE();
    case 474:
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') SKIP(313)
      if (lookahead == '\r') SKIP(98)
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 475:
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') SKIP(69)
      if (lookahead == '\r') SKIP(100)
      END_STATE();
    case 476:
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') SKIP(293)
      if (lookahead == '\r') SKIP(101)
      END_STATE();
    case 477:
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') ADVANCE(795);
      if (lookahead == '\r') ADVANCE(789);
      if (lookahead != 0) ADVANCE(798);
      END_STATE();
    case 478:
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') SKIP(323)
      if (lookahead == '\r') SKIP(105)
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 479:
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') SKIP(314)
      if (lookahead == '\r') SKIP(106)
      END_STATE();
    case 480:
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') SKIP(70)
      if (lookahead == '\r') SKIP(107)
      END_STATE();
    case 481:
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') SKIP(315)
      if (lookahead == '\r') SKIP(108)
      END_STATE();
    case 482:
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') SKIP(326)
      if (lookahead == '\r') SKIP(109)
      END_STATE();
    case 483:
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') SKIP(245)
      if (lookahead == '\r') SKIP(110)
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 484:
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') SKIP(243)
      if (lookahead == '\r') SKIP(111)
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 485:
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') SKIP(174)
      if (lookahead == '\r') SKIP(113)
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 486:
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') SKIP(264)
      if (lookahead == '\r') SKIP(114)
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 487:
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') SKIP(215)
      if (lookahead == '\r') SKIP(115)
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 488:
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') SKIP(217)
      if (lookahead == '\r') SKIP(116)
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 489:
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') SKIP(219)
      if (lookahead == '\r') SKIP(117)
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 490:
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') SKIP(221)
      if (lookahead == '\r') SKIP(119)
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 491:
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') SKIP(225)
      if (lookahead == '\r') SKIP(122)
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 492:
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') SKIP(231)
      if (lookahead == '\r') SKIP(123)
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 493:
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') SKIP(223)
      if (lookahead == '\r') SKIP(124)
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 494:
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') SKIP(227)
      if (lookahead == '\r') SKIP(125)
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 495:
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') SKIP(300)
      if (lookahead == '\r') SKIP(126)
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 496:
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') SKIP(232)
      if (lookahead == '\r') SKIP(127)
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 497:
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') SKIP(236)
      if (lookahead == '\r') SKIP(130)
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 498:
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') SKIP(309)
      if (lookahead == '\r') SKIP(132)
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 499:
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') SKIP(303)
      if (lookahead == '\r') SKIP(133)
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 500:
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') SKIP(296)
      if (lookahead == '\r') SKIP(134)
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 501:
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') SKIP(307)
      if (lookahead == '\r') SKIP(137)
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 502:
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') SKIP(274)
      if (lookahead == '\r') SKIP(141)
      END_STATE();
    case 503:
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') SKIP(238)
      if (lookahead == '\r') SKIP(144)
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 504:
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') SKIP(68)
      if (lookahead == '\r') SKIP(148)
      END_STATE();
    case 505:
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') SKIP(246)
      if (lookahead == '\r') SKIP(149)
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 506:
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') SKIP(229)
      if (lookahead == '\r') SKIP(152)
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 507:
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') SKIP(233)
      if (lookahead == '\r') SKIP(153)
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 508:
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') SKIP(234)
      if (lookahead == '\r') SKIP(154)
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 509:
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') SKIP(304)
      if (lookahead == '\r') SKIP(156)
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 510:
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') SKIP(301)
      if (lookahead == '\r') SKIP(157)
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 511:
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') SKIP(297)
      if (lookahead == '\r') SKIP(161)
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 512:
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') SKIP(201)
      if (lookahead == '\r') SKIP(164)
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 513:
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') SKIP(207)
      if (lookahead == '\r') SKIP(166)
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 514:
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') SKIP(210)
      if (lookahead == '\r') SKIP(167)
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 515:
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') SKIP(305)
      if (lookahead == '\r') SKIP(170)
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 516:
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(811);
      END_STATE();
    case 517:
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead)) ADVANCE(958);
      END_STATE();
    case 518:
      if (lookahead != 0 &&
          lookahead != '\r') ADVANCE(798);
      if (lookahead == '\r') ADVANCE(799);
      END_STATE();
    case 519:
      if (eof) ADVANCE(547);
      if (lookahead == '\n') ADVANCE(690);
      if (lookahead == '#') ADVANCE(862);
      if (lookahead == '&') ADVANCE(602);
      if (lookahead == ')') ADVANCE(655);
      if (lookahead == ';') ADVANCE(555);
      if (lookahead == '<') ADVANCE(612);
      if (lookahead == '>') ADVANCE(618);
      if (lookahead == '[' ||
          lookahead == ']' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') SKIP(544)
      if (lookahead == '`') ADVANCE(856);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(519)
      END_STATE();
    case 520:
      if (eof) ADVANCE(547);
      if (lookahead == '\n') ADVANCE(690);
      if (lookahead == '#') ADVANCE(862);
      if (lookahead == '&') ADVANCE(602);
      if (lookahead == ')') ADVANCE(655);
      if (lookahead == ';') ADVANCE(555);
      if (lookahead == '<') ADVANCE(612);
      if (lookahead == '>') ADVANCE(618);
      if (lookahead == '[' ||
          lookahead == ']' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') SKIP(546)
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(520)
      END_STATE();
    case 521:
      if (eof) ADVANCE(547);
      if (lookahead == '\n') ADVANCE(690);
      if (lookahead == '#') ADVANCE(862);
      if (lookahead == '&') ADVANCE(602);
      if (lookahead == ')') ADVANCE(655);
      if (lookahead == ';') ADVANCE(555);
      if (lookahead == '<') ADVANCE(613);
      if (lookahead == '>') ADVANCE(618);
      if (lookahead == '[' ||
          lookahead == ']' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') SKIP(543)
      if (lookahead == '`') ADVANCE(856);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(521)
      END_STATE();
    case 522:
      if (eof) ADVANCE(547);
      if (lookahead == '\n') ADVANCE(690);
      if (lookahead == '#') ADVANCE(862);
      if (lookahead == '&') ADVANCE(602);
      if (lookahead == ')') ADVANCE(655);
      if (lookahead == ';') ADVANCE(555);
      if (lookahead == '<') ADVANCE(613);
      if (lookahead == '>') ADVANCE(618);
      if (lookahead == '[' ||
          lookahead == ']' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') SKIP(545)
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(522)
      END_STATE();
    case 523:
      if (eof) ADVANCE(547);
      if (lookahead == '\n') SKIP(521)
      END_STATE();
    case 524:
      if (eof) ADVANCE(547);
      if (lookahead == '\n') SKIP(519)
      END_STATE();
    case 525:
      if (eof) ADVANCE(547);
      if (lookahead == '\n') SKIP(522)
      END_STATE();
    case 526:
      if (eof) ADVANCE(547);
      if (lookahead == '\n') SKIP(520)
      END_STATE();
    case 527:
      if (eof) ADVANCE(547);
      if (lookahead == '\n') ADVANCE(703);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(862);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(602);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(654);
      if (lookahead == ')') ADVANCE(655);
      if (lookahead == '-') ADVANCE(950);
      if (lookahead == '0') ADVANCE(803);
      if (lookahead == ';') ADVANCE(555);
      if (lookahead == '<') ADVANCE(609);
      if (lookahead == '=') ADVANCE(952);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '[' ||
          lookahead == ']' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') ADVANCE(432);
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(527)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(805);
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 528:
      if (eof) ADVANCE(547);
      if (lookahead == '\n') ADVANCE(705);
      if (lookahead == '!') ADVANCE(672);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(602);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(654);
      if (lookahead == '*') ADVANCE(640);
      if (lookahead == '-') ADVANCE(636);
      if (lookahead == '0') ADVANCE(943);
      if (lookahead == ';') ADVANCE(555);
      if (lookahead == '<') ADVANCE(609);
      if (lookahead == '=') ADVANCE(952);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '?') ADVANCE(761);
      if (lookahead == '@') ADVANCE(942);
      if (lookahead == '[' ||
          lookahead == ']' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') ADVANCE(390);
      if (lookahead == '_') ADVANCE(947);
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(528)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(807);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(873);
      if (lookahead != 0 &&
          lookahead != ')') ADVANCE(958);
      END_STATE();
    case 529:
      if (eof) ADVANCE(547);
      if (lookahead == '\n') ADVANCE(710);
      if (lookahead == '!') ADVANCE(669);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(780);
      if (lookahead == '&') ADVANCE(602);
      if (lookahead == '*') ADVANCE(637);
      if (lookahead == '-') ADVANCE(632);
      if (lookahead == '0') ADVANCE(945);
      if (lookahead == ';') ADVANCE(555);
      if (lookahead == '<') ADVANCE(612);
      if (lookahead == '>') ADVANCE(618);
      if (lookahead == '?') ADVANCE(760);
      if (lookahead == '@') ADVANCE(941);
      if (lookahead == '\\') ADVANCE(406);
      if (lookahead == '_') ADVANCE(948);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(529)
      if (('1' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(879);
      END_STATE();
    case 530:
      if (eof) ADVANCE(547);
      if (lookahead == '\n') ADVANCE(712);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(862);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(602);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(332);
      if (lookahead == ')') ADVANCE(655);
      if (lookahead == '-') ADVANCE(950);
      if (lookahead == '0') ADVANCE(804);
      if (lookahead == ';') ADVANCE(555);
      if (lookahead == '<') ADVANCE(610);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '[' ||
          lookahead == ']' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') ADVANCE(435);
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(530)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(806);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(868);
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 531:
      if (eof) ADVANCE(547);
      if (lookahead == '\n') ADVANCE(714);
      if (lookahead == '!') ADVANCE(672);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(602);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(332);
      if (lookahead == '*') ADVANCE(640);
      if (lookahead == '-') ADVANCE(636);
      if (lookahead == '0') ADVANCE(943);
      if (lookahead == ';') ADVANCE(555);
      if (lookahead == '<') ADVANCE(609);
      if (lookahead == '=') ADVANCE(952);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '?') ADVANCE(761);
      if (lookahead == '@') ADVANCE(942);
      if (lookahead == '[' ||
          lookahead == ']' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') ADVANCE(408);
      if (lookahead == '_') ADVANCE(947);
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(531)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(807);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(873);
      if (lookahead != 0 &&
          lookahead != ')') ADVANCE(958);
      END_STATE();
    case 532:
      if (eof) ADVANCE(547);
      if (lookahead == '\n') ADVANCE(716);
      if (lookahead == '!') ADVANCE(669);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(780);
      if (lookahead == '&') ADVANCE(602);
      if (lookahead == '*') ADVANCE(637);
      if (lookahead == '-') ADVANCE(632);
      if (lookahead == '0') ADVANCE(945);
      if (lookahead == ';') ADVANCE(555);
      if (lookahead == '<') ADVANCE(613);
      if (lookahead == '>') ADVANCE(618);
      if (lookahead == '?') ADVANCE(760);
      if (lookahead == '@') ADVANCE(941);
      if (lookahead == '\\') ADVANCE(409);
      if (lookahead == '_') ADVANCE(948);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(532)
      if (('1' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(879);
      END_STATE();
    case 533:
      if (eof) ADVANCE(547);
      if (lookahead == '\n') ADVANCE(724);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(862);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(602);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(332);
      if (lookahead == ')') ADVANCE(655);
      if (lookahead == '-') ADVANCE(950);
      if (lookahead == '0') ADVANCE(803);
      if (lookahead == ';') ADVANCE(555);
      if (lookahead == '<') ADVANCE(609);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '[' ||
          lookahead == ']' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') ADVANCE(440);
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(533)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(805);
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 534:
      if (eof) ADVANCE(547);
      if (lookahead == '\n') ADVANCE(730);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(862);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(602);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(332);
      if (lookahead == ')') ADVANCE(655);
      if (lookahead == '-') ADVANCE(950);
      if (lookahead == '0') ADVANCE(803);
      if (lookahead == ';') ADVANCE(555);
      if (lookahead == '<') ADVANCE(610);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '[' ||
          lookahead == ']' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') ADVANCE(514);
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(534)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(805);
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 535:
      if (eof) ADVANCE(547);
      if (lookahead == '\n') ADVANCE(732);
      if (lookahead == '!') ADVANCE(672);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(602);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(332);
      if (lookahead == '*') ADVANCE(640);
      if (lookahead == '-') ADVANCE(636);
      if (lookahead == '0') ADVANCE(943);
      if (lookahead == ';') ADVANCE(555);
      if (lookahead == '<') ADVANCE(609);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '?') ADVANCE(761);
      if (lookahead == '@') ADVANCE(942);
      if (lookahead == '[' ||
          lookahead == ']' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') ADVANCE(392);
      if (lookahead == '_') ADVANCE(947);
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(535)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(807);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(873);
      if (lookahead != 0 &&
          lookahead != ')') ADVANCE(958);
      END_STATE();
    case 536:
      if (eof) ADVANCE(547);
      if (lookahead == '\n') ADVANCE(738);
      if (lookahead == '!') ADVANCE(672);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(602);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(332);
      if (lookahead == '*') ADVANCE(640);
      if (lookahead == '-') ADVANCE(636);
      if (lookahead == '0') ADVANCE(943);
      if (lookahead == ';') ADVANCE(555);
      if (lookahead == '<') ADVANCE(610);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '?') ADVANCE(761);
      if (lookahead == '@') ADVANCE(942);
      if (lookahead == '[' ||
          lookahead == ']' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') ADVANCE(405);
      if (lookahead == '_') ADVANCE(947);
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(536)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(807);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(873);
      if (lookahead != 0 &&
          lookahead != ')') ADVANCE(958);
      END_STATE();
    case 537:
      if (eof) ADVANCE(547);
      if (lookahead == '\n') ADVANCE(739);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(862);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(602);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(654);
      if (lookahead == ')') ADVANCE(655);
      if (lookahead == '-') ADVANCE(950);
      if (lookahead == '0') ADVANCE(803);
      if (lookahead == ';') ADVANCE(555);
      if (lookahead == '<') ADVANCE(609);
      if (lookahead == '=') ADVANCE(952);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '[' ||
          lookahead == ']' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') ADVANCE(489);
      if (lookahead == '`') ADVANCE(856);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(537)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(805);
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 538:
      if (eof) ADVANCE(547);
      if (lookahead == '\n') ADVANCE(751);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(862);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(602);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(332);
      if (lookahead == ')') ADVANCE(655);
      if (lookahead == '-') ADVANCE(950);
      if (lookahead == '0') ADVANCE(804);
      if (lookahead == ';') ADVANCE(555);
      if (lookahead == '<') ADVANCE(610);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '[' ||
          lookahead == ']' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') ADVANCE(492);
      if (lookahead == '`') ADVANCE(856);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(538)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(806);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(868);
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 539:
      if (eof) ADVANCE(547);
      if (lookahead == '\n') ADVANCE(752);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(862);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(602);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(332);
      if (lookahead == ')') ADVANCE(655);
      if (lookahead == '-') ADVANCE(950);
      if (lookahead == '0') ADVANCE(803);
      if (lookahead == ';') ADVANCE(555);
      if (lookahead == '<') ADVANCE(609);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '[' ||
          lookahead == ']' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') ADVANCE(496);
      if (lookahead == '`') ADVANCE(856);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(539)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(805);
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 540:
      if (eof) ADVANCE(547);
      if (lookahead == '\n') ADVANCE(754);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(862);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(602);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(332);
      if (lookahead == ')') ADVANCE(655);
      if (lookahead == '-') ADVANCE(950);
      if (lookahead == '0') ADVANCE(803);
      if (lookahead == ';') ADVANCE(555);
      if (lookahead == '<') ADVANCE(610);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '[' ||
          lookahead == ']' ||
          lookahead == '{' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '\\') ADVANCE(508);
      if (lookahead == '`') ADVANCE(856);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(540)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(805);
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 541:
      if (eof) ADVANCE(547);
      if (lookahead == '!') ADVANCE(670);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '%') ADVANCE(647);
      if (lookahead == '&') ADVANCE(601);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(654);
      if (lookahead == ')') ADVANCE(655);
      if (lookahead == '*') ADVANCE(638);
      if (lookahead == '+') ADVANCE(770);
      if (lookahead == ',') ADVANCE(557);
      if (lookahead == '-') ADVANCE(768);
      if (lookahead == '/') ADVANCE(642);
      if (lookahead == '0') ADVANCE(944);
      if (lookahead == ':') ADVANCE(765);
      if (lookahead == ';') ADVANCE(554);
      if (lookahead == '<') ADVANCE(607);
      if (lookahead == '=') ADVANCE(563);
      if (lookahead == '>') ADVANCE(616);
      if (lookahead == '?') ADVANCE(761);
      if (lookahead == '@') ADVANCE(942);
      if (lookahead == '[') ADVANCE(674);
      if (lookahead == '\\') ADVANCE(417);
      if (lookahead == ']') ADVANCE(675);
      if (lookahead == '^') ADVANCE(595);
      if (lookahead == '_') ADVANCE(946);
      if (lookahead == '`') ADVANCE(856);
      if (lookahead == 'e') ADVANCE(956);
      if (lookahead == 'i') ADVANCE(955);
      if (lookahead == '{') ADVANCE(666);
      if (lookahead == '|') ADVANCE(592);
      if (lookahead == '}') ADVANCE(818);
      if (lookahead == '~') ADVANCE(772);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(541)
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 542:
      if (eof) ADVANCE(547);
      if (lookahead == '!') ADVANCE(672);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(862);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(346);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(654);
      if (lookahead == ')') ADVANCE(655);
      if (lookahead == '-') ADVANCE(950);
      if (lookahead == '0') ADVANCE(803);
      if (lookahead == ';') ADVANCE(328);
      if (lookahead == '<') ADVANCE(608);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '[') ADVANCE(674);
      if (lookahead == '\\') ADVANCE(418);
      if (lookahead == ']' ||
          lookahead == '}') ADVANCE(786);
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == '{') ADVANCE(666);
      if (lookahead == '|') ADVANCE(590);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(542)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(805);
      if (lookahead != 0) ADVANCE(958);
      END_STATE();
    case 543:
      if (eof) ADVANCE(547);
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') SKIP(521)
      if (lookahead == '\r') SKIP(523)
      END_STATE();
    case 544:
      if (eof) ADVANCE(547);
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') SKIP(519)
      if (lookahead == '\r') SKIP(524)
      END_STATE();
    case 545:
      if (eof) ADVANCE(547);
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') SKIP(522)
      if (lookahead == '\r') SKIP(525)
      END_STATE();
    case 546:
      if (eof) ADVANCE(547);
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') SKIP(520)
      if (lookahead == '\r') SKIP(526)
      END_STATE();
    case 547:
      ACCEPT_TOKEN(ts_builtin_sym_end);
      END_STATE();
    case 548:
      ACCEPT_TOKEN(anon_sym_in);
      END_STATE();
    case 549:
      ACCEPT_TOKEN(anon_sym_in);
      if (lookahead == '\\') ADVANCE(517);
      if (!sym_word_character_set_1(lookahead)) ADVANCE(958);
      END_STATE();
    case 550:
      ACCEPT_TOKEN(anon_sym_in);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(879);
      END_STATE();
    case 551:
      ACCEPT_TOKEN(anon_sym_LPAREN_LPAREN);
      END_STATE();
    case 552:
      ACCEPT_TOKEN(anon_sym_RPAREN_RPAREN);
      END_STATE();
    case 553:
      ACCEPT_TOKEN(anon_sym_SEMI);
      END_STATE();
    case 554:
      ACCEPT_TOKEN(anon_sym_SEMI);
      if (lookahead == '&') ADVANCE(664);
      if (lookahead == ';') ADVANCE(663);
      END_STATE();
    case 555:
      ACCEPT_TOKEN(anon_sym_SEMI);
      if (lookahead == ';') ADVANCE(662);
      END_STATE();
    case 556:
      ACCEPT_TOKEN(anon_sym_COMMA);
      END_STATE();
    case 557:
      ACCEPT_TOKEN(anon_sym_COMMA);
      if (lookahead == ',') ADVANCE(850);
      if (lookahead == '\\') ADVANCE(517);
      if (!sym_word_character_set_1(lookahead)) ADVANCE(958);
      END_STATE();
    case 558:
      ACCEPT_TOKEN(anon_sym_COMMA);
      if (lookahead == ',') ADVANCE(849);
      END_STATE();
    case 559:
      ACCEPT_TOKEN(anon_sym_EQ);
      END_STATE();
    case 560:
      ACCEPT_TOKEN(anon_sym_EQ);
      if (lookahead == '=') ADVANCE(604);
      if (lookahead == '\\') ADVANCE(517);
      if (lookahead == '~') ADVANCE(679);
      if (!sym_word_character_set_2(lookahead)) ADVANCE(958);
      END_STATE();
    case 561:
      ACCEPT_TOKEN(anon_sym_EQ);
      if (lookahead == '=') ADVANCE(603);
      END_STATE();
    case 562:
      ACCEPT_TOKEN(anon_sym_EQ);
      if (lookahead == '=') ADVANCE(603);
      if (lookahead == '~') ADVANCE(678);
      END_STATE();
    case 563:
      ACCEPT_TOKEN(anon_sym_EQ);
      if (lookahead == '\\') ADVANCE(517);
      if (!sym_word_character_set_1(lookahead)) ADVANCE(958);
      END_STATE();
    case 564:
      ACCEPT_TOKEN(anon_sym_PLUS_PLUS);
      END_STATE();
    case 565:
      ACCEPT_TOKEN(anon_sym_PLUS_PLUS);
      if (lookahead == '\\') ADVANCE(517);
      if (!sym_word_character_set_1(lookahead)) ADVANCE(958);
      END_STATE();
    case 566:
      ACCEPT_TOKEN(anon_sym_DASH_DASH);
      END_STATE();
    case 567:
      ACCEPT_TOKEN(anon_sym_DASH_DASH);
      if (lookahead == '\\') ADVANCE(517);
      if (!sym_word_character_set_1(lookahead)) ADVANCE(958);
      END_STATE();
    case 568:
      ACCEPT_TOKEN(anon_sym_PLUS_EQ);
      END_STATE();
    case 569:
      ACCEPT_TOKEN(anon_sym_PLUS_EQ);
      if (lookahead == '\\') ADVANCE(517);
      if (!sym_word_character_set_1(lookahead)) ADVANCE(958);
      END_STATE();
    case 570:
      ACCEPT_TOKEN(anon_sym_DASH_EQ);
      END_STATE();
    case 571:
      ACCEPT_TOKEN(anon_sym_DASH_EQ);
      if (lookahead == '\\') ADVANCE(517);
      if (!sym_word_character_set_1(lookahead)) ADVANCE(958);
      END_STATE();
    case 572:
      ACCEPT_TOKEN(anon_sym_STAR_EQ);
      END_STATE();
    case 573:
      ACCEPT_TOKEN(anon_sym_STAR_EQ);
      if (lookahead == '\\') ADVANCE(517);
      if (!sym_word_character_set_1(lookahead)) ADVANCE(958);
      END_STATE();
    case 574:
      ACCEPT_TOKEN(anon_sym_SLASH_EQ);
      END_STATE();
    case 575:
      ACCEPT_TOKEN(anon_sym_SLASH_EQ);
      if (lookahead == '\\') ADVANCE(517);
      if (!sym_word_character_set_1(lookahead)) ADVANCE(958);
      END_STATE();
    case 576:
      ACCEPT_TOKEN(anon_sym_PERCENT_EQ);
      END_STATE();
    case 577:
      ACCEPT_TOKEN(anon_sym_PERCENT_EQ);
      if (lookahead == '\\') ADVANCE(517);
      if (!sym_word_character_set_1(lookahead)) ADVANCE(958);
      END_STATE();
    case 578:
      ACCEPT_TOKEN(anon_sym_STAR_STAR_EQ);
      END_STATE();
    case 579:
      ACCEPT_TOKEN(anon_sym_STAR_STAR_EQ);
      if (lookahead == '\\') ADVANCE(517);
      if (!sym_word_character_set_1(lookahead)) ADVANCE(958);
      END_STATE();
    case 580:
      ACCEPT_TOKEN(anon_sym_LT_LT_EQ);
      END_STATE();
    case 581:
      ACCEPT_TOKEN(anon_sym_GT_GT_EQ);
      END_STATE();
    case 582:
      ACCEPT_TOKEN(anon_sym_AMP_EQ);
      END_STATE();
    case 583:
      ACCEPT_TOKEN(anon_sym_CARET_EQ);
      END_STATE();
    case 584:
      ACCEPT_TOKEN(anon_sym_CARET_EQ);
      if (lookahead == '\\') ADVANCE(517);
      if (!sym_word_character_set_1(lookahead)) ADVANCE(958);
      END_STATE();
    case 585:
      ACCEPT_TOKEN(anon_sym_PIPE_EQ);
      END_STATE();
    case 586:
      ACCEPT_TOKEN(anon_sym_PIPE_PIPE);
      END_STATE();
    case 587:
      ACCEPT_TOKEN(anon_sym_DASHo);
      END_STATE();
    case 588:
      ACCEPT_TOKEN(anon_sym_AMP_AMP);
      END_STATE();
    case 589:
      ACCEPT_TOKEN(anon_sym_DASHa);
      END_STATE();
    case 590:
      ACCEPT_TOKEN(anon_sym_PIPE);
      END_STATE();
    case 591:
      ACCEPT_TOKEN(anon_sym_PIPE);
      if (lookahead == '&') ADVANCE(668);
      END_STATE();
    case 592:
      ACCEPT_TOKEN(anon_sym_PIPE);
      if (lookahead == '&') ADVANCE(668);
      if (lookahead == '=') ADVANCE(585);
      if (lookahead == '|') ADVANCE(586);
      END_STATE();
    case 593:
      ACCEPT_TOKEN(anon_sym_PIPE);
      if (lookahead == '&') ADVANCE(668);
      if (lookahead == '|') ADVANCE(586);
      END_STATE();
    case 594:
      ACCEPT_TOKEN(anon_sym_PIPE);
      if (lookahead == '=') ADVANCE(585);
      if (lookahead == '|') ADVANCE(586);
      END_STATE();
    case 595:
      ACCEPT_TOKEN(anon_sym_CARET);
      if (lookahead == '=') ADVANCE(584);
      if (lookahead == '\\') ADVANCE(517);
      if (lookahead == '^') ADVANCE(852);
      if (!sym_word_character_set_2(lookahead)) ADVANCE(958);
      END_STATE();
    case 596:
      ACCEPT_TOKEN(anon_sym_CARET);
      if (lookahead == '=') ADVANCE(584);
      if (lookahead == '\\') ADVANCE(517);
      if (!sym_word_character_set_2(lookahead)) ADVANCE(958);
      END_STATE();
    case 597:
      ACCEPT_TOKEN(anon_sym_CARET);
      if (lookahead == '=') ADVANCE(583);
      END_STATE();
    case 598:
      ACCEPT_TOKEN(anon_sym_CARET);
      if (lookahead == '^') ADVANCE(851);
      END_STATE();
    case 599:
      ACCEPT_TOKEN(anon_sym_AMP);
      END_STATE();
    case 600:
      ACCEPT_TOKEN(anon_sym_AMP);
      if (lookahead == '&') ADVANCE(588);
      if (lookahead == '=') ADVANCE(582);
      END_STATE();
    case 601:
      ACCEPT_TOKEN(anon_sym_AMP);
      if (lookahead == '&') ADVANCE(588);
      if (lookahead == '=') ADVANCE(582);
      if (lookahead == '>') ADVANCE(680);
      END_STATE();
    case 602:
      ACCEPT_TOKEN(anon_sym_AMP);
      if (lookahead == '&') ADVANCE(588);
      if (lookahead == '>') ADVANCE(680);
      END_STATE();
    case 603:
      ACCEPT_TOKEN(anon_sym_EQ_EQ);
      END_STATE();
    case 604:
      ACCEPT_TOKEN(anon_sym_EQ_EQ);
      if (lookahead == '\\') ADVANCE(517);
      if (!sym_word_character_set_1(lookahead)) ADVANCE(958);
      END_STATE();
    case 605:
      ACCEPT_TOKEN(anon_sym_BANG_EQ);
      END_STATE();
    case 606:
      ACCEPT_TOKEN(anon_sym_BANG_EQ);
      if (lookahead == '\\') ADVANCE(517);
      if (!sym_word_character_set_1(lookahead)) ADVANCE(958);
      END_STATE();
    case 607:
      ACCEPT_TOKEN(anon_sym_LT);
      if (lookahead == '&') ADVANCE(682);
      if (lookahead == '(') ADVANCE(858);
      if (lookahead == '<') ADVANCE(625);
      if (lookahead == '=') ADVANCE(621);
      END_STATE();
    case 608:
      ACCEPT_TOKEN(anon_sym_LT);
      if (lookahead == '&') ADVANCE(682);
      if (lookahead == '(') ADVANCE(858);
      if (lookahead == '<') ADVANCE(343);
      END_STATE();
    case 609:
      ACCEPT_TOKEN(anon_sym_LT);
      if (lookahead == '&') ADVANCE(682);
      if (lookahead == '(') ADVANCE(858);
      if (lookahead == '<') ADVANCE(624);
      END_STATE();
    case 610:
      ACCEPT_TOKEN(anon_sym_LT);
      if (lookahead == '&') ADVANCE(682);
      if (lookahead == '(') ADVANCE(858);
      if (lookahead == '<') ADVANCE(623);
      END_STATE();
    case 611:
      ACCEPT_TOKEN(anon_sym_LT);
      if (lookahead == '&') ADVANCE(682);
      if (lookahead == '<') ADVANCE(343);
      END_STATE();
    case 612:
      ACCEPT_TOKEN(anon_sym_LT);
      if (lookahead == '&') ADVANCE(682);
      if (lookahead == '<') ADVANCE(624);
      END_STATE();
    case 613:
      ACCEPT_TOKEN(anon_sym_LT);
      if (lookahead == '&') ADVANCE(682);
      if (lookahead == '<') ADVANCE(623);
      END_STATE();
    case 614:
      ACCEPT_TOKEN(anon_sym_LT);
      if (lookahead == '(') ADVANCE(858);
      if (lookahead == '<') ADVANCE(626);
      if (lookahead == '=') ADVANCE(621);
      END_STATE();
    case 615:
      ACCEPT_TOKEN(anon_sym_LT);
      if (lookahead == '<') ADVANCE(626);
      if (lookahead == '=') ADVANCE(621);
      END_STATE();
    case 616:
      ACCEPT_TOKEN(anon_sym_GT);
      if (lookahead == '&') ADVANCE(683);
      if (lookahead == '(') ADVANCE(859);
      if (lookahead == '=') ADVANCE(622);
      if (lookahead == '>') ADVANCE(628);
      if (lookahead == '|') ADVANCE(684);
      END_STATE();
    case 617:
      ACCEPT_TOKEN(anon_sym_GT);
      if (lookahead == '&') ADVANCE(683);
      if (lookahead == '(') ADVANCE(859);
      if (lookahead == '>') ADVANCE(627);
      if (lookahead == '|') ADVANCE(684);
      END_STATE();
    case 618:
      ACCEPT_TOKEN(anon_sym_GT);
      if (lookahead == '&') ADVANCE(683);
      if (lookahead == '>') ADVANCE(627);
      if (lookahead == '|') ADVANCE(684);
      END_STATE();
    case 619:
      ACCEPT_TOKEN(anon_sym_GT);
      if (lookahead == '(') ADVANCE(859);
      if (lookahead == '=') ADVANCE(622);
      if (lookahead == '>') ADVANCE(628);
      END_STATE();
    case 620:
      ACCEPT_TOKEN(anon_sym_GT);
      if (lookahead == '=') ADVANCE(622);
      if (lookahead == '>') ADVANCE(628);
      END_STATE();
    case 621:
      ACCEPT_TOKEN(anon_sym_LT_EQ);
      END_STATE();
    case 622:
      ACCEPT_TOKEN(anon_sym_GT_EQ);
      END_STATE();
    case 623:
      ACCEPT_TOKEN(anon_sym_LT_LT);
      if (lookahead == '-') ADVANCE(687);
      END_STATE();
    case 624:
      ACCEPT_TOKEN(anon_sym_LT_LT);
      if (lookahead == '-') ADVANCE(687);
      if (lookahead == '<') ADVANCE(759);
      END_STATE();
    case 625:
      ACCEPT_TOKEN(anon_sym_LT_LT);
      if (lookahead == '-') ADVANCE(687);
      if (lookahead == '<') ADVANCE(759);
      if (lookahead == '=') ADVANCE(580);
      END_STATE();
    case 626:
      ACCEPT_TOKEN(anon_sym_LT_LT);
      if (lookahead == '=') ADVANCE(580);
      END_STATE();
    case 627:
      ACCEPT_TOKEN(anon_sym_GT_GT);
      END_STATE();
    case 628:
      ACCEPT_TOKEN(anon_sym_GT_GT);
      if (lookahead == '=') ADVANCE(581);
      END_STATE();
    case 629:
      ACCEPT_TOKEN(anon_sym_PLUS);
      END_STATE();
    case 630:
      ACCEPT_TOKEN(anon_sym_PLUS);
      if (lookahead == '+') ADVANCE(565);
      if (lookahead == '=') ADVANCE(569);
      if (lookahead == '\\') ADVANCE(517);
      if (!sym_word_character_set_2(lookahead)) ADVANCE(958);
      END_STATE();
    case 631:
      ACCEPT_TOKEN(anon_sym_PLUS);
      if (lookahead == '+') ADVANCE(564);
      if (lookahead == '=') ADVANCE(568);
      END_STATE();
    case 632:
      ACCEPT_TOKEN(anon_sym_DASH);
      END_STATE();
    case 633:
      ACCEPT_TOKEN(anon_sym_DASH);
      if (lookahead == '-') ADVANCE(567);
      if (lookahead == '0') ADVANCE(803);
      if (lookahead == '=') ADVANCE(571);
      if (lookahead == '\\') ADVANCE(517);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(805);
      if (!sym_word_character_set_2(lookahead)) ADVANCE(958);
      END_STATE();
    case 634:
      ACCEPT_TOKEN(anon_sym_DASH);
      if (lookahead == '-') ADVANCE(566);
      if (lookahead == '=') ADVANCE(570);
      END_STATE();
    case 635:
      ACCEPT_TOKEN(anon_sym_DASH);
      if (lookahead == '-') ADVANCE(566);
      if (lookahead == '=') ADVANCE(570);
      if (lookahead == 'a') ADVANCE(589);
      if (lookahead == 'o') ADVANCE(587);
      END_STATE();
    case 636:
      ACCEPT_TOKEN(anon_sym_DASH);
      if (lookahead == '0') ADVANCE(803);
      if (lookahead == '\\') ADVANCE(517);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(805);
      if (!sym_word_character_set_1(lookahead)) ADVANCE(958);
      END_STATE();
    case 637:
      ACCEPT_TOKEN(anon_sym_STAR);
      END_STATE();
    case 638:
      ACCEPT_TOKEN(anon_sym_STAR);
      if (lookahead == '*') ADVANCE(651);
      if (lookahead == '=') ADVANCE(573);
      if (lookahead == '\\') ADVANCE(517);
      if (!sym_word_character_set_2(lookahead)) ADVANCE(958);
      END_STATE();
    case 639:
      ACCEPT_TOKEN(anon_sym_STAR);
      if (lookahead == '*') ADVANCE(652);
      if (lookahead == '=') ADVANCE(572);
      END_STATE();
    case 640:
      ACCEPT_TOKEN(anon_sym_STAR);
      if (lookahead == '\\') ADVANCE(517);
      if (!sym_word_character_set_1(lookahead)) ADVANCE(958);
      END_STATE();
    case 641:
      ACCEPT_TOKEN(anon_sym_SLASH);
      END_STATE();
    case 642:
      ACCEPT_TOKEN(anon_sym_SLASH);
      if (lookahead == '#') ADVANCE(846);
      if (lookahead == '%') ADVANCE(848);
      if (lookahead == '/') ADVANCE(844);
      if (lookahead == '=') ADVANCE(575);
      if (lookahead == '\\') ADVANCE(517);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          (lookahead < '"' || ')' < lookahead) &&
          (lookahead < ';' || '>' < lookahead) &&
          (lookahead < '[' || ']' < lookahead) &&
          lookahead != '`' &&
          (lookahead < '{' || '}' < lookahead)) ADVANCE(958);
      END_STATE();
    case 643:
      ACCEPT_TOKEN(anon_sym_SLASH);
      if (lookahead == '#') ADVANCE(845);
      if (lookahead == '%') ADVANCE(847);
      if (lookahead == '/') ADVANCE(843);
      END_STATE();
    case 644:
      ACCEPT_TOKEN(anon_sym_SLASH);
      if (lookahead == '=') ADVANCE(575);
      if (lookahead == '\\') ADVANCE(517);
      if (!sym_word_character_set_2(lookahead)) ADVANCE(958);
      END_STATE();
    case 645:
      ACCEPT_TOKEN(anon_sym_SLASH);
      if (lookahead == '=') ADVANCE(574);
      END_STATE();
    case 646:
      ACCEPT_TOKEN(anon_sym_PERCENT);
      END_STATE();
    case 647:
      ACCEPT_TOKEN(anon_sym_PERCENT);
      if (lookahead == '%') ADVANCE(841);
      if (lookahead == '=') ADVANCE(577);
      if (lookahead == '\\') ADVANCE(517);
      if (!sym_word_character_set_3(lookahead)) ADVANCE(958);
      END_STATE();
    case 648:
      ACCEPT_TOKEN(anon_sym_PERCENT);
      if (lookahead == '%') ADVANCE(840);
      END_STATE();
    case 649:
      ACCEPT_TOKEN(anon_sym_PERCENT);
      if (lookahead == '=') ADVANCE(577);
      if (lookahead == '\\') ADVANCE(517);
      if (!sym_word_character_set_2(lookahead)) ADVANCE(958);
      END_STATE();
    case 650:
      ACCEPT_TOKEN(anon_sym_PERCENT);
      if (lookahead == '=') ADVANCE(576);
      END_STATE();
    case 651:
      ACCEPT_TOKEN(anon_sym_STAR_STAR);
      if (lookahead == '=') ADVANCE(579);
      if (lookahead == '\\') ADVANCE(517);
      if (!sym_word_character_set_2(lookahead)) ADVANCE(958);
      END_STATE();
    case 652:
      ACCEPT_TOKEN(anon_sym_STAR_STAR);
      if (lookahead == '=') ADVANCE(578);
      END_STATE();
    case 653:
      ACCEPT_TOKEN(anon_sym_LPAREN);
      END_STATE();
    case 654:
      ACCEPT_TOKEN(anon_sym_LPAREN);
      if (lookahead == '(') ADVANCE(551);
      END_STATE();
    case 655:
      ACCEPT_TOKEN(anon_sym_RPAREN);
      END_STATE();
    case 656:
      ACCEPT_TOKEN(aux_sym__c_word_token1);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(656);
      END_STATE();
    case 657:
      ACCEPT_TOKEN(anon_sym_esac);
      END_STATE();
    case 658:
      ACCEPT_TOKEN(anon_sym_esac);
      if (lookahead == '\\') ADVANCE(517);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(868);
      if (!sym_word_character_set_4(lookahead)) ADVANCE(958);
      END_STATE();
    case 659:
      ACCEPT_TOKEN(anon_sym_esac);
      if (lookahead == '\\') ADVANCE(517);
      if (!sym_word_character_set_1(lookahead)) ADVANCE(958);
      END_STATE();
    case 660:
      ACCEPT_TOKEN(anon_sym_esac);
      if (lookahead == '\\') ADVANCE(16);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(873);
      if (!sym_word_character_set_4(lookahead)) ADVANCE(958);
      END_STATE();
    case 661:
      ACCEPT_TOKEN(anon_sym_esac);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(879);
      END_STATE();
    case 662:
      ACCEPT_TOKEN(anon_sym_SEMI_SEMI);
      END_STATE();
    case 663:
      ACCEPT_TOKEN(anon_sym_SEMI_SEMI);
      if (lookahead == '&') ADVANCE(665);
      END_STATE();
    case 664:
      ACCEPT_TOKEN(anon_sym_SEMI_AMP);
      END_STATE();
    case 665:
      ACCEPT_TOKEN(anon_sym_SEMI_SEMI_AMP);
      END_STATE();
    case 666:
      ACCEPT_TOKEN(anon_sym_LBRACE);
      END_STATE();
    case 667:
      ACCEPT_TOKEN(anon_sym_RBRACE);
      END_STATE();
    case 668:
      ACCEPT_TOKEN(anon_sym_PIPE_AMP);
      END_STATE();
    case 669:
      ACCEPT_TOKEN(anon_sym_BANG);
      END_STATE();
    case 670:
      ACCEPT_TOKEN(anon_sym_BANG);
      if (lookahead == '=') ADVANCE(606);
      if (lookahead == '\\') ADVANCE(517);
      if (!sym_word_character_set_2(lookahead)) ADVANCE(958);
      END_STATE();
    case 671:
      ACCEPT_TOKEN(anon_sym_BANG);
      if (lookahead == '=') ADVANCE(605);
      END_STATE();
    case 672:
      ACCEPT_TOKEN(anon_sym_BANG);
      if (lookahead == '\\') ADVANCE(517);
      if (!sym_word_character_set_1(lookahead)) ADVANCE(958);
      END_STATE();
    case 673:
      ACCEPT_TOKEN(anon_sym_LBRACK);
      END_STATE();
    case 674:
      ACCEPT_TOKEN(anon_sym_LBRACK);
      if (lookahead == '[') ADVANCE(676);
      END_STATE();
    case 675:
      ACCEPT_TOKEN(anon_sym_RBRACK);
      END_STATE();
    case 676:
      ACCEPT_TOKEN(anon_sym_LBRACK_LBRACK);
      END_STATE();
    case 677:
      ACCEPT_TOKEN(anon_sym_RBRACK_RBRACK);
      END_STATE();
    case 678:
      ACCEPT_TOKEN(anon_sym_EQ_TILDE);
      END_STATE();
    case 679:
      ACCEPT_TOKEN(anon_sym_EQ_TILDE);
      if (lookahead == '\\') ADVANCE(517);
      if (!sym_word_character_set_1(lookahead)) ADVANCE(958);
      END_STATE();
    case 680:
      ACCEPT_TOKEN(anon_sym_AMP_GT);
      if (lookahead == '>') ADVANCE(681);
      END_STATE();
    case 681:
      ACCEPT_TOKEN(anon_sym_AMP_GT_GT);
      END_STATE();
    case 682:
      ACCEPT_TOKEN(anon_sym_LT_AMP);
      if (lookahead == '-') ADVANCE(685);
      END_STATE();
    case 683:
      ACCEPT_TOKEN(anon_sym_GT_AMP);
      if (lookahead == '-') ADVANCE(686);
      END_STATE();
    case 684:
      ACCEPT_TOKEN(anon_sym_GT_PIPE);
      END_STATE();
    case 685:
      ACCEPT_TOKEN(anon_sym_LT_AMP_DASH);
      END_STATE();
    case 686:
      ACCEPT_TOKEN(anon_sym_GT_AMP_DASH);
      END_STATE();
    case 687:
      ACCEPT_TOKEN(anon_sym_LT_LT_DASH);
      END_STATE();
    case 688:
      ACCEPT_TOKEN(aux_sym_heredoc_redirect_token1);
      if (lookahead == '\n') ADVANCE(688);
      if (lookahead == '+') ADVANCE(771);
      if (lookahead == '-') ADVANCE(769);
      if (lookahead == '\\') ADVANCE(421);
      if (lookahead == '~') ADVANCE(772);
      END_STATE();
    case 689:
      ACCEPT_TOKEN(aux_sym_heredoc_redirect_token1);
      if (lookahead == '\n') ADVANCE(689);
      if (lookahead == '\\') ADVANCE(355);
      END_STATE();
    case 690:
      ACCEPT_TOKEN(aux_sym_heredoc_redirect_token1);
      if (lookahead == '\n') ADVANCE(690);
      END_STATE();
    case 691:
      ACCEPT_TOKEN(aux_sym_heredoc_redirect_token1);
      if (lookahead == '\n') ADVANCE(691);
      if (lookahead == '\\') ADVANCE(427);
      END_STATE();
    case 692:
      ACCEPT_TOKEN(aux_sym_heredoc_redirect_token1);
      if (lookahead == '\n') ADVANCE(692);
      if (lookahead == '\\') ADVANCE(373);
      END_STATE();
    case 693:
      ACCEPT_TOKEN(aux_sym_heredoc_redirect_token1);
      if (lookahead == '\n') ADVANCE(693);
      if (lookahead == '\\') ADVANCE(357);
      END_STATE();
    case 694:
      ACCEPT_TOKEN(aux_sym_heredoc_redirect_token1);
      if (lookahead == '\n') ADVANCE(694);
      if (lookahead == '\\') ADVANCE(485);
      END_STATE();
    case 695:
      ACCEPT_TOKEN(aux_sym_heredoc_redirect_token1);
      if (lookahead == '\n') ADVANCE(695);
      if (lookahead == '\\') ADVANCE(374);
      END_STATE();
    case 696:
      ACCEPT_TOKEN(aux_sym_heredoc_redirect_token1);
      if (lookahead == '\n') ADVANCE(696);
      if (lookahead == '\\') ADVANCE(358);
      END_STATE();
    case 697:
      ACCEPT_TOKEN(aux_sym_heredoc_redirect_token1);
      if (lookahead == '\n') ADVANCE(697);
      if (lookahead == '\\') ADVANCE(430);
      END_STATE();
    case 698:
      ACCEPT_TOKEN(aux_sym_heredoc_redirect_token1);
      if (lookahead == '\n') ADVANCE(698);
      if (lookahead == '\\') ADVANCE(416);
      END_STATE();
    case 699:
      ACCEPT_TOKEN(aux_sym_heredoc_redirect_token1);
      if (lookahead == '\n') ADVANCE(699);
      if (lookahead == '\\') ADVANCE(389);
      END_STATE();
    case 700:
      ACCEPT_TOKEN(aux_sym_heredoc_redirect_token1);
      if (lookahead == '\n') ADVANCE(700);
      if (lookahead == '\\') ADVANCE(431);
      END_STATE();
    case 701:
      ACCEPT_TOKEN(aux_sym_heredoc_redirect_token1);
      if (lookahead == '\n') ADVANCE(701);
      if (lookahead == '\\') ADVANCE(398);
      END_STATE();
    case 702:
      ACCEPT_TOKEN(aux_sym_heredoc_redirect_token1);
      if (lookahead == '\n') ADVANCE(702);
      if (lookahead == '\\') ADVANCE(359);
      END_STATE();
    case 703:
      ACCEPT_TOKEN(aux_sym_heredoc_redirect_token1);
      if (lookahead == '\n') ADVANCE(703);
      if (lookahead == '\\') ADVANCE(432);
      END_STATE();
    case 704:
      ACCEPT_TOKEN(aux_sym_heredoc_redirect_token1);
      if (lookahead == '\n') ADVANCE(704);
      if (lookahead == '\\') ADVANCE(375);
      END_STATE();
    case 705:
      ACCEPT_TOKEN(aux_sym_heredoc_redirect_token1);
      if (lookahead == '\n') ADVANCE(705);
      if (lookahead == '\\') ADVANCE(390);
      END_STATE();
    case 706:
      ACCEPT_TOKEN(aux_sym_heredoc_redirect_token1);
      if (lookahead == '\n') ADVANCE(706);
      if (lookahead == '\\') ADVANCE(433);
      END_STATE();
    case 707:
      ACCEPT_TOKEN(aux_sym_heredoc_redirect_token1);
      if (lookahead == '\n') ADVANCE(707);
      if (lookahead == '\\') ADVANCE(376);
      END_STATE();
    case 708:
      ACCEPT_TOKEN(aux_sym_heredoc_redirect_token1);
      if (lookahead == '\n') ADVANCE(708);
      if (lookahead == '\\') ADVANCE(404);
      END_STATE();
    case 709:
      ACCEPT_TOKEN(aux_sym_heredoc_redirect_token1);
      if (lookahead == '\n') ADVANCE(709);
      if (lookahead == '\\') ADVANCE(434);
      END_STATE();
    case 710:
      ACCEPT_TOKEN(aux_sym_heredoc_redirect_token1);
      if (lookahead == '\n') ADVANCE(710);
      if (lookahead == '\\') ADVANCE(406);
      END_STATE();
    case 711:
      ACCEPT_TOKEN(aux_sym_heredoc_redirect_token1);
      if (lookahead == '\n') ADVANCE(711);
      if (lookahead == '\\') ADVANCE(391);
      END_STATE();
    case 712:
      ACCEPT_TOKEN(aux_sym_heredoc_redirect_token1);
      if (lookahead == '\n') ADVANCE(712);
      if (lookahead == '\\') ADVANCE(435);
      END_STATE();
    case 713:
      ACCEPT_TOKEN(aux_sym_heredoc_redirect_token1);
      if (lookahead == '\n') ADVANCE(713);
      if (lookahead == '\\') ADVANCE(399);
      END_STATE();
    case 714:
      ACCEPT_TOKEN(aux_sym_heredoc_redirect_token1);
      if (lookahead == '\n') ADVANCE(714);
      if (lookahead == '\\') ADVANCE(408);
      END_STATE();
    case 715:
      ACCEPT_TOKEN(aux_sym_heredoc_redirect_token1);
      if (lookahead == '\n') ADVANCE(715);
      if (lookahead == '\\') ADVANCE(436);
      END_STATE();
    case 716:
      ACCEPT_TOKEN(aux_sym_heredoc_redirect_token1);
      if (lookahead == '\n') ADVANCE(716);
      if (lookahead == '\\') ADVANCE(409);
      END_STATE();
    case 717:
      ACCEPT_TOKEN(aux_sym_heredoc_redirect_token1);
      if (lookahead == '\n') ADVANCE(717);
      if (lookahead == '\\') ADVANCE(360);
      END_STATE();
    case 718:
      ACCEPT_TOKEN(aux_sym_heredoc_redirect_token1);
      if (lookahead == '\n') ADVANCE(718);
      if (lookahead == '\\') ADVANCE(437);
      END_STATE();
    case 719:
      ACCEPT_TOKEN(aux_sym_heredoc_redirect_token1);
      if (lookahead == '\n') ADVANCE(719);
      if (lookahead == '\\') ADVANCE(400);
      END_STATE();
    case 720:
      ACCEPT_TOKEN(aux_sym_heredoc_redirect_token1);
      if (lookahead == '\n') ADVANCE(720);
      if (lookahead == '\\') ADVANCE(410);
      END_STATE();
    case 721:
      ACCEPT_TOKEN(aux_sym_heredoc_redirect_token1);
      if (lookahead == '\n') ADVANCE(721);
      if (lookahead == '\\') ADVANCE(512);
      END_STATE();
    case 722:
      ACCEPT_TOKEN(aux_sym_heredoc_redirect_token1);
      if (lookahead == '\n') ADVANCE(722);
      if (lookahead == '\\') ADVANCE(378);
      END_STATE();
    case 723:
      ACCEPT_TOKEN(aux_sym_heredoc_redirect_token1);
      if (lookahead == '\n') ADVANCE(723);
      if (lookahead == '\\') ADVANCE(361);
      END_STATE();
    case 724:
      ACCEPT_TOKEN(aux_sym_heredoc_redirect_token1);
      if (lookahead == '\n') ADVANCE(724);
      if (lookahead == '\\') ADVANCE(440);
      END_STATE();
    case 725:
      ACCEPT_TOKEN(aux_sym_heredoc_redirect_token1);
      if (lookahead == '\n') ADVANCE(725);
      if (lookahead == '\\') ADVANCE(401);
      END_STATE();
    case 726:
      ACCEPT_TOKEN(aux_sym_heredoc_redirect_token1);
      if (lookahead == '\n') ADVANCE(726);
      if (lookahead == '\\') ADVANCE(362);
      END_STATE();
    case 727:
      ACCEPT_TOKEN(aux_sym_heredoc_redirect_token1);
      if (lookahead == '\n') ADVANCE(727);
      if (lookahead == '\\') ADVANCE(513);
      END_STATE();
    case 728:
      ACCEPT_TOKEN(aux_sym_heredoc_redirect_token1);
      if (lookahead == '\n') ADVANCE(728);
      if (lookahead == '\\') ADVANCE(407);
      END_STATE();
    case 729:
      ACCEPT_TOKEN(aux_sym_heredoc_redirect_token1);
      if (lookahead == '\n') ADVANCE(729);
      if (lookahead == '\\') ADVANCE(411);
      END_STATE();
    case 730:
      ACCEPT_TOKEN(aux_sym_heredoc_redirect_token1);
      if (lookahead == '\n') ADVANCE(730);
      if (lookahead == '\\') ADVANCE(514);
      END_STATE();
    case 731:
      ACCEPT_TOKEN(aux_sym_heredoc_redirect_token1);
      if (lookahead == '\n') ADVANCE(731);
      if (lookahead == '\\') ADVANCE(383);
      END_STATE();
    case 732:
      ACCEPT_TOKEN(aux_sym_heredoc_redirect_token1);
      if (lookahead == '\n') ADVANCE(732);
      if (lookahead == '\\') ADVANCE(392);
      END_STATE();
    case 733:
      ACCEPT_TOKEN(aux_sym_heredoc_redirect_token1);
      if (lookahead == '\n') ADVANCE(733);
      if (lookahead == '\\') ADVANCE(442);
      END_STATE();
    case 734:
      ACCEPT_TOKEN(aux_sym_heredoc_redirect_token1);
      if (lookahead == '\n') ADVANCE(734);
      if (lookahead == '\\') ADVANCE(363);
      END_STATE();
    case 735:
      ACCEPT_TOKEN(aux_sym_heredoc_redirect_token1);
      if (lookahead == '\n') ADVANCE(735);
      if (lookahead == '\\') ADVANCE(487);
      END_STATE();
    case 736:
      ACCEPT_TOKEN(aux_sym_heredoc_redirect_token1);
      if (lookahead == '\n') ADVANCE(736);
      if (lookahead == '\\') ADVANCE(412);
      END_STATE();
    case 737:
      ACCEPT_TOKEN(aux_sym_heredoc_redirect_token1);
      if (lookahead == '\n') ADVANCE(737);
      if (lookahead == '\\') ADVANCE(488);
      END_STATE();
    case 738:
      ACCEPT_TOKEN(aux_sym_heredoc_redirect_token1);
      if (lookahead == '\n') ADVANCE(738);
      if (lookahead == '\\') ADVANCE(405);
      END_STATE();
    case 739:
      ACCEPT_TOKEN(aux_sym_heredoc_redirect_token1);
      if (lookahead == '\n') ADVANCE(739);
      if (lookahead == '\\') ADVANCE(489);
      END_STATE();
    case 740:
      ACCEPT_TOKEN(aux_sym_heredoc_redirect_token1);
      if (lookahead == '\n') ADVANCE(740);
      if (lookahead == '\\') ADVANCE(393);
      END_STATE();
    case 741:
      ACCEPT_TOKEN(aux_sym_heredoc_redirect_token1);
      if (lookahead == '\n') ADVANCE(741);
      if (lookahead == '\\') ADVANCE(490);
      END_STATE();
    case 742:
      ACCEPT_TOKEN(aux_sym_heredoc_redirect_token1);
      if (lookahead == '\n') ADVANCE(742);
      if (lookahead == '\\') ADVANCE(366);
      END_STATE();
    case 743:
      ACCEPT_TOKEN(aux_sym_heredoc_redirect_token1);
      if (lookahead == '\n') ADVANCE(743);
      if (lookahead == '\\') ADVANCE(493);
      END_STATE();
    case 744:
      ACCEPT_TOKEN(aux_sym_heredoc_redirect_token1);
      if (lookahead == '\n') ADVANCE(744);
      if (lookahead == '\\') ADVANCE(414);
      END_STATE();
    case 745:
      ACCEPT_TOKEN(aux_sym_heredoc_redirect_token1);
      if (lookahead == '\n') ADVANCE(745);
      if (lookahead == '\\') ADVANCE(491);
      END_STATE();
    case 746:
      ACCEPT_TOKEN(aux_sym_heredoc_redirect_token1);
      if (lookahead == '\n') ADVANCE(746);
      if (lookahead == '\\') ADVANCE(395);
      END_STATE();
    case 747:
      ACCEPT_TOKEN(aux_sym_heredoc_redirect_token1);
      if (lookahead == '\n') ADVANCE(747);
      if (lookahead == '\\') ADVANCE(494);
      END_STATE();
    case 748:
      ACCEPT_TOKEN(aux_sym_heredoc_redirect_token1);
      if (lookahead == '\n') ADVANCE(748);
      if (lookahead == '\\') ADVANCE(377);
      END_STATE();
    case 749:
      ACCEPT_TOKEN(aux_sym_heredoc_redirect_token1);
      if (lookahead == '\n') ADVANCE(749);
      if (lookahead == '\\') ADVANCE(506);
      END_STATE();
    case 750:
      ACCEPT_TOKEN(aux_sym_heredoc_redirect_token1);
      if (lookahead == '\n') ADVANCE(750);
      if (lookahead == '\\') ADVANCE(381);
      END_STATE();
    case 751:
      ACCEPT_TOKEN(aux_sym_heredoc_redirect_token1);
      if (lookahead == '\n') ADVANCE(751);
      if (lookahead == '\\') ADVANCE(492);
      END_STATE();
    case 752:
      ACCEPT_TOKEN(aux_sym_heredoc_redirect_token1);
      if (lookahead == '\n') ADVANCE(752);
      if (lookahead == '\\') ADVANCE(496);
      END_STATE();
    case 753:
      ACCEPT_TOKEN(aux_sym_heredoc_redirect_token1);
      if (lookahead == '\n') ADVANCE(753);
      if (lookahead == '\\') ADVANCE(507);
      END_STATE();
    case 754:
      ACCEPT_TOKEN(aux_sym_heredoc_redirect_token1);
      if (lookahead == '\n') ADVANCE(754);
      if (lookahead == '\\') ADVANCE(508);
      END_STATE();
    case 755:
      ACCEPT_TOKEN(aux_sym_heredoc_redirect_token1);
      if (lookahead == '\n') ADVANCE(755);
      if (lookahead == '\\') ADVANCE(451);
      END_STATE();
    case 756:
      ACCEPT_TOKEN(aux_sym_heredoc_redirect_token1);
      if (lookahead == '\n') ADVANCE(756);
      if (lookahead == '\\') ADVANCE(497);
      END_STATE();
    case 757:
      ACCEPT_TOKEN(aux_sym_heredoc_redirect_token1);
      if (lookahead == '\n') ADVANCE(757);
      if (lookahead == '\\') ADVANCE(459);
      END_STATE();
    case 758:
      ACCEPT_TOKEN(aux_sym_heredoc_redirect_token1);
      if (lookahead == '\n') ADVANCE(758);
      if (lookahead == '\\') ADVANCE(503);
      END_STATE();
    case 759:
      ACCEPT_TOKEN(anon_sym_LT_LT_LT);
      END_STATE();
    case 760:
      ACCEPT_TOKEN(anon_sym_QMARK);
      END_STATE();
    case 761:
      ACCEPT_TOKEN(anon_sym_QMARK);
      if (lookahead == '\\') ADVANCE(517);
      if (!sym_word_character_set_1(lookahead)) ADVANCE(958);
      END_STATE();
    case 762:
      ACCEPT_TOKEN(anon_sym_COLON);
      END_STATE();
    case 763:
      ACCEPT_TOKEN(anon_sym_COLON);
      if (lookahead == '+') ADVANCE(835);
      if (lookahead == '-') ADVANCE(832);
      if (lookahead == '=') ADVANCE(829);
      if (lookahead == '?') ADVANCE(839);
      if (lookahead == '\\') ADVANCE(517);
      if (!sym_word_character_set_2(lookahead)) ADVANCE(958);
      END_STATE();
    case 764:
      ACCEPT_TOKEN(anon_sym_COLON);
      if (lookahead == '+') ADVANCE(834);
      if (lookahead == '-') ADVANCE(831);
      if (lookahead == '=') ADVANCE(828);
      if (lookahead == '?') ADVANCE(838);
      END_STATE();
    case 765:
      ACCEPT_TOKEN(anon_sym_COLON);
      if (lookahead == '\\') ADVANCE(517);
      if (!sym_word_character_set_1(lookahead)) ADVANCE(958);
      END_STATE();
    case 766:
      ACCEPT_TOKEN(anon_sym_PLUS_PLUS2);
      END_STATE();
    case 767:
      ACCEPT_TOKEN(anon_sym_DASH_DASH2);
      END_STATE();
    case 768:
      ACCEPT_TOKEN(anon_sym_DASH2);
      END_STATE();
    case 769:
      ACCEPT_TOKEN(anon_sym_DASH2);
      if (lookahead == '-') ADVANCE(767);
      END_STATE();
    case 770:
      ACCEPT_TOKEN(anon_sym_PLUS2);
      END_STATE();
    case 771:
      ACCEPT_TOKEN(anon_sym_PLUS2);
      if (lookahead == '+') ADVANCE(766);
      END_STATE();
    case 772:
      ACCEPT_TOKEN(anon_sym_TILDE);
      END_STATE();
    case 773:
      ACCEPT_TOKEN(anon_sym_DOLLAR_LPAREN_LPAREN);
      END_STATE();
    case 774:
      ACCEPT_TOKEN(anon_sym_DOLLAR_LBRACK);
      END_STATE();
    case 775:
      ACCEPT_TOKEN(aux_sym_brace_expression_token1);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(775);
      END_STATE();
    case 776:
      ACCEPT_TOKEN(anon_sym_DOT_DOT);
      END_STATE();
    case 777:
      ACCEPT_TOKEN(anon_sym_DOT_DOT);
      if (lookahead == '\\') ADVANCE(517);
      if (!sym_word_character_set_1(lookahead)) ADVANCE(958);
      END_STATE();
    case 778:
      ACCEPT_TOKEN(anon_sym_RBRACE2);
      END_STATE();
    case 779:
      ACCEPT_TOKEN(aux_sym_concatenation_token1);
      END_STATE();
    case 780:
      ACCEPT_TOKEN(anon_sym_DOLLAR);
      END_STATE();
    case 781:
      ACCEPT_TOKEN(anon_sym_DOLLAR);
      if (lookahead == '\'') ADVANCE(330);
      if (lookahead == '(') ADVANCE(854);
      if (lookahead == '[') ADVANCE(774);
      if (lookahead == '`') ADVANCE(857);
      if (lookahead == '{') ADVANCE(817);
      END_STATE();
    case 782:
      ACCEPT_TOKEN(anon_sym_DOLLAR);
      if (lookahead == '\'') ADVANCE(330);
      if (lookahead == '(') ADVANCE(853);
      if (lookahead == '`') ADVANCE(857);
      if (lookahead == '{') ADVANCE(817);
      END_STATE();
    case 783:
      ACCEPT_TOKEN(anon_sym_DOLLAR);
      if (lookahead == '(') ADVANCE(854);
      if (lookahead == '[') ADVANCE(774);
      if (lookahead == '`') ADVANCE(857);
      if (lookahead == '{') ADVANCE(817);
      END_STATE();
    case 784:
      ACCEPT_TOKEN(anon_sym_DOLLAR);
      if (lookahead == '(') ADVANCE(853);
      if (lookahead == '`') ADVANCE(857);
      END_STATE();
    case 785:
      ACCEPT_TOKEN(anon_sym_DOLLAR);
      if (lookahead == '(') ADVANCE(853);
      if (lookahead == '`') ADVANCE(857);
      if (lookahead == '{') ADVANCE(817);
      END_STATE();
    case 786:
      ACCEPT_TOKEN(sym__special_character);
      END_STATE();
    case 787:
      ACCEPT_TOKEN(sym__special_character);
      if (lookahead == ']') ADVANCE(677);
      END_STATE();
    case 788:
      ACCEPT_TOKEN(anon_sym_DQUOTE);
      END_STATE();
    case 789:
      ACCEPT_TOKEN(sym_string_content);
      if (lookahead == '\n') ADVANCE(795);
      if (lookahead == '\\') ADVANCE(518);
      if (lookahead != 0 &&
          lookahead != '\r' &&
          lookahead != '"' &&
          lookahead != '$' &&
          lookahead != '`') ADVANCE(798);
      END_STATE();
    case 790:
      ACCEPT_TOKEN(sym_string_content);
      if (lookahead == '\n') ADVANCE(793);
      if (lookahead == '\\') ADVANCE(518);
      if (lookahead != 0 &&
          lookahead != '\r' &&
          lookahead != '"' &&
          lookahead != '$' &&
          lookahead != '`') ADVANCE(798);
      END_STATE();
    case 791:
      ACCEPT_TOKEN(sym_string_content);
      if (lookahead == '\n') ADVANCE(794);
      if (lookahead == '\\') ADVANCE(518);
      if (lookahead != 0 &&
          lookahead != '\r' &&
          lookahead != '"' &&
          lookahead != '$' &&
          lookahead != '`') ADVANCE(798);
      END_STATE();
    case 792:
      ACCEPT_TOKEN(sym_string_content);
      if (lookahead == '\n') ADVANCE(798);
      if (lookahead == '\\') ADVANCE(860);
      if (lookahead != 0 &&
          lookahead != '\r' &&
          lookahead != '"' &&
          lookahead != '$' &&
          lookahead != '`') ADVANCE(797);
      END_STATE();
    case 793:
      ACCEPT_TOKEN(sym_string_content);
      if (lookahead == '!') ADVANCE(669);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(780);
      if (lookahead == '*') ADVANCE(637);
      if (lookahead == '-') ADVANCE(632);
      if (lookahead == '0') ADVANCE(945);
      if (lookahead == '?') ADVANCE(760);
      if (lookahead == '@') ADVANCE(941);
      if (lookahead == '\\') ADVANCE(386);
      if (lookahead == '_') ADVANCE(948);
      if (lookahead == '\n' ||
          lookahead == '\r') SKIP(287)
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') ADVANCE(793);
      if (('1' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(879);
      if (lookahead != 0 &&
          lookahead != '`') ADVANCE(798);
      END_STATE();
    case 794:
      ACCEPT_TOKEN(sym_string_content);
      if (lookahead == '!') ADVANCE(669);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(780);
      if (lookahead == '*') ADVANCE(637);
      if (lookahead == '-') ADVANCE(632);
      if (lookahead == '0') ADVANCE(945);
      if (lookahead == '?') ADVANCE(760);
      if (lookahead == '@') ADVANCE(941);
      if (lookahead == '\\') ADVANCE(387);
      if (lookahead == '_') ADVANCE(948);
      if (lookahead == '\n' ||
          lookahead == '\r') SKIP(290)
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') ADVANCE(794);
      if (('1' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(879);
      if (lookahead != 0 &&
          lookahead != '"' &&
          lookahead != '`') ADVANCE(798);
      END_STATE();
    case 795:
      ACCEPT_TOKEN(sym_string_content);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(797);
      if (lookahead == '$') ADVANCE(783);
      if (lookahead == '(') ADVANCE(796);
      if (lookahead == '\\') ADVANCE(477);
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == '\n' ||
          lookahead == '\r') SKIP(318)
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') ADVANCE(795);
      if (lookahead != 0) ADVANCE(798);
      END_STATE();
    case 796:
      ACCEPT_TOKEN(sym_string_content);
      if (lookahead == '(') ADVANCE(551);
      if (lookahead == '\\') ADVANCE(518);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '\r' &&
          lookahead != '"' &&
          lookahead != '$' &&
          lookahead != '`') ADVANCE(798);
      END_STATE();
    case 797:
      ACCEPT_TOKEN(sym_string_content);
      if (lookahead == '\\') ADVANCE(860);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '\r' &&
          lookahead != '"' &&
          lookahead != '$' &&
          lookahead != '`') ADVANCE(797);
      END_STATE();
    case 798:
      ACCEPT_TOKEN(sym_string_content);
      if (lookahead == '\\') ADVANCE(518);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '\r' &&
          lookahead != '"' &&
          lookahead != '$' &&
          lookahead != '`') ADVANCE(798);
      END_STATE();
    case 799:
      ACCEPT_TOKEN(sym_string_content);
      if (lookahead != 0 &&
          lookahead != '\r' &&
          lookahead != '"' &&
          lookahead != '$' &&
          lookahead != '\\' &&
          lookahead != '`') ADVANCE(798);
      if (lookahead == '\\') ADVANCE(518);
      END_STATE();
    case 800:
      ACCEPT_TOKEN(sym_raw_string);
      END_STATE();
    case 801:
      ACCEPT_TOKEN(sym_ansi_c_string);
      END_STATE();
    case 802:
      ACCEPT_TOKEN(sym_ansi_c_string);
      if (lookahead == '\'') ADVANCE(801);
      if (lookahead == '\\') ADVANCE(331);
      if (lookahead != 0) ADVANCE(330);
      END_STATE();
    case 803:
      ACCEPT_TOKEN(aux_sym_number_token1);
      if (lookahead == '#') ADVANCE(814);
      if (lookahead == '\\') ADVANCE(517);
      if (lookahead == 'x') ADVANCE(957);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(805);
      if (!sym_word_character_set_5(lookahead)) ADVANCE(958);
      END_STATE();
    case 804:
      ACCEPT_TOKEN(aux_sym_number_token1);
      if (lookahead == '#') ADVANCE(814);
      if (lookahead == '\\') ADVANCE(517);
      if (lookahead == 'x') ADVANCE(867);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(806);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(868);
      if (!sym_word_character_set_6(lookahead)) ADVANCE(958);
      END_STATE();
    case 805:
      ACCEPT_TOKEN(aux_sym_number_token1);
      if (lookahead == '#') ADVANCE(814);
      if (lookahead == '\\') ADVANCE(517);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(805);
      if (!sym_word_character_set_5(lookahead)) ADVANCE(958);
      END_STATE();
    case 806:
      ACCEPT_TOKEN(aux_sym_number_token1);
      if (lookahead == '#') ADVANCE(814);
      if (lookahead == '\\') ADVANCE(517);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(806);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(868);
      if (!sym_word_character_set_6(lookahead)) ADVANCE(958);
      END_STATE();
    case 807:
      ACCEPT_TOKEN(aux_sym_number_token1);
      if (lookahead == '#') ADVANCE(814);
      if (lookahead == '\\') ADVANCE(16);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(807);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(873);
      if (!sym_word_character_set_6(lookahead)) ADVANCE(958);
      END_STATE();
    case 808:
      ACCEPT_TOKEN(aux_sym_number_token1);
      if (lookahead == '#') ADVANCE(815);
      if (lookahead == 'x') ADVANCE(878);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(810);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(879);
      END_STATE();
    case 809:
      ACCEPT_TOKEN(aux_sym_number_token1);
      if (lookahead == '#') ADVANCE(815);
      if (lookahead == 'x') ADVANCE(516);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(811);
      END_STATE();
    case 810:
      ACCEPT_TOKEN(aux_sym_number_token1);
      if (lookahead == '#') ADVANCE(815);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(810);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(879);
      END_STATE();
    case 811:
      ACCEPT_TOKEN(aux_sym_number_token1);
      if (lookahead == '#') ADVANCE(815);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(811);
      END_STATE();
    case 812:
      ACCEPT_TOKEN(aux_sym_number_token1);
      if (lookahead == '\\') ADVANCE(517);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('@' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(812);
      if (!sym_word_character_set_4(lookahead)) ADVANCE(958);
      END_STATE();
    case 813:
      ACCEPT_TOKEN(aux_sym_number_token1);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('@' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(813);
      END_STATE();
    case 814:
      ACCEPT_TOKEN(aux_sym_number_token2);
      if (lookahead == '\\') ADVANCE(517);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('@' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(812);
      if (!sym_word_character_set_4(lookahead)) ADVANCE(958);
      END_STATE();
    case 815:
      ACCEPT_TOKEN(aux_sym_number_token2);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('@' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(813);
      END_STATE();
    case 816:
      ACCEPT_TOKEN(anon_sym_POUND);
      END_STATE();
    case 817:
      ACCEPT_TOKEN(anon_sym_DOLLAR_LBRACE);
      END_STATE();
    case 818:
      ACCEPT_TOKEN(anon_sym_RBRACE3);
      END_STATE();
    case 819:
      ACCEPT_TOKEN(anon_sym_BANG2);
      END_STATE();
    case 820:
      ACCEPT_TOKEN(anon_sym_BANG2);
      if (lookahead == '=') ADVANCE(606);
      if (lookahead == '\\') ADVANCE(517);
      if (!sym_word_character_set_2(lookahead)) ADVANCE(958);
      END_STATE();
    case 821:
      ACCEPT_TOKEN(anon_sym_AT);
      END_STATE();
    case 822:
      ACCEPT_TOKEN(anon_sym_AT);
      if (lookahead == '\\') ADVANCE(517);
      if (!sym_word_character_set_1(lookahead)) ADVANCE(958);
      END_STATE();
    case 823:
      ACCEPT_TOKEN(anon_sym_STAR2);
      END_STATE();
    case 824:
      ACCEPT_TOKEN(anon_sym_STAR2);
      if (lookahead == '*') ADVANCE(651);
      if (lookahead == '=') ADVANCE(573);
      if (lookahead == '\\') ADVANCE(517);
      if (!sym_word_character_set_2(lookahead)) ADVANCE(958);
      END_STATE();
    case 825:
      ACCEPT_TOKEN(anon_sym_POUND2);
      END_STATE();
    case 826:
      ACCEPT_TOKEN(anon_sym_EQ2);
      END_STATE();
    case 827:
      ACCEPT_TOKEN(anon_sym_EQ2);
      if (lookahead == '\\') ADVANCE(517);
      if (!sym_word_character_set_1(lookahead)) ADVANCE(958);
      END_STATE();
    case 828:
      ACCEPT_TOKEN(anon_sym_COLON_EQ);
      END_STATE();
    case 829:
      ACCEPT_TOKEN(anon_sym_COLON_EQ);
      if (lookahead == '\\') ADVANCE(517);
      if (!sym_word_character_set_1(lookahead)) ADVANCE(958);
      END_STATE();
    case 830:
      ACCEPT_TOKEN(anon_sym_DASH3);
      END_STATE();
    case 831:
      ACCEPT_TOKEN(anon_sym_COLON_DASH);
      END_STATE();
    case 832:
      ACCEPT_TOKEN(anon_sym_COLON_DASH);
      if (lookahead == '\\') ADVANCE(517);
      if (!sym_word_character_set_1(lookahead)) ADVANCE(958);
      END_STATE();
    case 833:
      ACCEPT_TOKEN(anon_sym_PLUS3);
      END_STATE();
    case 834:
      ACCEPT_TOKEN(anon_sym_COLON_PLUS);
      END_STATE();
    case 835:
      ACCEPT_TOKEN(anon_sym_COLON_PLUS);
      if (lookahead == '\\') ADVANCE(517);
      if (!sym_word_character_set_1(lookahead)) ADVANCE(958);
      END_STATE();
    case 836:
      ACCEPT_TOKEN(anon_sym_QMARK2);
      END_STATE();
    case 837:
      ACCEPT_TOKEN(anon_sym_QMARK2);
      if (lookahead == '\\') ADVANCE(517);
      if (!sym_word_character_set_1(lookahead)) ADVANCE(958);
      END_STATE();
    case 838:
      ACCEPT_TOKEN(anon_sym_COLON_QMARK);
      END_STATE();
    case 839:
      ACCEPT_TOKEN(anon_sym_COLON_QMARK);
      if (lookahead == '\\') ADVANCE(517);
      if (!sym_word_character_set_1(lookahead)) ADVANCE(958);
      END_STATE();
    case 840:
      ACCEPT_TOKEN(anon_sym_PERCENT_PERCENT);
      END_STATE();
    case 841:
      ACCEPT_TOKEN(anon_sym_PERCENT_PERCENT);
      if (lookahead == '\\') ADVANCE(517);
      if (!sym_word_character_set_1(lookahead)) ADVANCE(958);
      END_STATE();
    case 842:
      ACCEPT_TOKEN(aux_sym__expansion_regex_token1);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(842);
      END_STATE();
    case 843:
      ACCEPT_TOKEN(anon_sym_SLASH_SLASH);
      END_STATE();
    case 844:
      ACCEPT_TOKEN(anon_sym_SLASH_SLASH);
      if (lookahead == '\\') ADVANCE(517);
      if (!sym_word_character_set_1(lookahead)) ADVANCE(958);
      END_STATE();
    case 845:
      ACCEPT_TOKEN(anon_sym_SLASH_POUND);
      END_STATE();
    case 846:
      ACCEPT_TOKEN(anon_sym_SLASH_POUND);
      if (lookahead == '\\') ADVANCE(517);
      if (!sym_word_character_set_1(lookahead)) ADVANCE(958);
      END_STATE();
    case 847:
      ACCEPT_TOKEN(anon_sym_SLASH_PERCENT);
      END_STATE();
    case 848:
      ACCEPT_TOKEN(anon_sym_SLASH_PERCENT);
      if (lookahead == '\\') ADVANCE(517);
      if (!sym_word_character_set_1(lookahead)) ADVANCE(958);
      END_STATE();
    case 849:
      ACCEPT_TOKEN(anon_sym_COMMA_COMMA);
      END_STATE();
    case 850:
      ACCEPT_TOKEN(anon_sym_COMMA_COMMA);
      if (lookahead == '\\') ADVANCE(517);
      if (!sym_word_character_set_1(lookahead)) ADVANCE(958);
      END_STATE();
    case 851:
      ACCEPT_TOKEN(anon_sym_CARET_CARET);
      END_STATE();
    case 852:
      ACCEPT_TOKEN(anon_sym_CARET_CARET);
      if (lookahead == '\\') ADVANCE(517);
      if (!sym_word_character_set_1(lookahead)) ADVANCE(958);
      END_STATE();
    case 853:
      ACCEPT_TOKEN(anon_sym_DOLLAR_LPAREN);
      END_STATE();
    case 854:
      ACCEPT_TOKEN(anon_sym_DOLLAR_LPAREN);
      if (lookahead == '(') ADVANCE(773);
      END_STATE();
    case 855:
      ACCEPT_TOKEN(anon_sym_BQUOTE);
      END_STATE();
    case 856:
      ACCEPT_TOKEN(anon_sym_BQUOTE);
      if (lookahead == '`') ADVANCE(779);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(348);
      END_STATE();
    case 857:
      ACCEPT_TOKEN(anon_sym_DOLLAR_BQUOTE);
      END_STATE();
    case 858:
      ACCEPT_TOKEN(anon_sym_LT_LPAREN);
      END_STATE();
    case 859:
      ACCEPT_TOKEN(anon_sym_GT_LPAREN);
      END_STATE();
    case 860:
      ACCEPT_TOKEN(sym_comment);
      if (lookahead == '\n') ADVANCE(798);
      if (lookahead == '\r') ADVANCE(792);
      if (lookahead != 0) ADVANCE(797);
      END_STATE();
    case 861:
      ACCEPT_TOKEN(sym_comment);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead)) ADVANCE(863);
      if (lookahead == '\t' ||
          (11 <= lookahead && lookahead <= '\r')) ADVANCE(862);
      END_STATE();
    case 862:
      ACCEPT_TOKEN(sym_comment);
      if (lookahead != 0 &&
          lookahead != '\n') ADVANCE(862);
      END_STATE();
    case 863:
      ACCEPT_TOKEN(sym__comment_word);
      if (lookahead == '\\') ADVANCE(861);
      if (!sym_word_character_set_1(lookahead)) ADVANCE(863);
      END_STATE();
    case 864:
      ACCEPT_TOKEN(aux_sym__simple_variable_name_token1);
      if (lookahead == '\\') ADVANCE(517);
      if (lookahead == 'a') ADVANCE(865);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(868);
      if (!sym_word_character_set_4(lookahead)) ADVANCE(958);
      END_STATE();
    case 865:
      ACCEPT_TOKEN(aux_sym__simple_variable_name_token1);
      if (lookahead == '\\') ADVANCE(517);
      if (lookahead == 'c') ADVANCE(658);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(868);
      if (!sym_word_character_set_4(lookahead)) ADVANCE(958);
      END_STATE();
    case 866:
      ACCEPT_TOKEN(aux_sym__simple_variable_name_token1);
      if (lookahead == '\\') ADVANCE(517);
      if (lookahead == 's') ADVANCE(864);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(868);
      if (!sym_word_character_set_4(lookahead)) ADVANCE(958);
      END_STATE();
    case 867:
      ACCEPT_TOKEN(aux_sym__simple_variable_name_token1);
      if (lookahead == '\\') ADVANCE(517);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(806);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(868);
      if (!sym_word_character_set_4(lookahead)) ADVANCE(958);
      END_STATE();
    case 868:
      ACCEPT_TOKEN(aux_sym__simple_variable_name_token1);
      if (lookahead == '\\') ADVANCE(517);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(868);
      if (!sym_word_character_set_4(lookahead)) ADVANCE(958);
      END_STATE();
    case 869:
      ACCEPT_TOKEN(aux_sym__simple_variable_name_token1);
      if (lookahead == '\\') ADVANCE(16);
      if (lookahead == 'a') ADVANCE(870);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(873);
      if (!sym_word_character_set_4(lookahead)) ADVANCE(958);
      END_STATE();
    case 870:
      ACCEPT_TOKEN(aux_sym__simple_variable_name_token1);
      if (lookahead == '\\') ADVANCE(16);
      if (lookahead == 'c') ADVANCE(660);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(873);
      if (!sym_word_character_set_4(lookahead)) ADVANCE(958);
      END_STATE();
    case 871:
      ACCEPT_TOKEN(aux_sym__simple_variable_name_token1);
      if (lookahead == '\\') ADVANCE(16);
      if (lookahead == 's') ADVANCE(869);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(873);
      if (!sym_word_character_set_4(lookahead)) ADVANCE(958);
      END_STATE();
    case 872:
      ACCEPT_TOKEN(aux_sym__simple_variable_name_token1);
      if (lookahead == '\\') ADVANCE(16);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(807);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(873);
      if (!sym_word_character_set_4(lookahead)) ADVANCE(958);
      END_STATE();
    case 873:
      ACCEPT_TOKEN(aux_sym__simple_variable_name_token1);
      if (lookahead == '\\') ADVANCE(16);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(873);
      if (!sym_word_character_set_4(lookahead)) ADVANCE(958);
      END_STATE();
    case 874:
      ACCEPT_TOKEN(aux_sym__simple_variable_name_token1);
      if (lookahead == 'a') ADVANCE(875);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(879);
      END_STATE();
    case 875:
      ACCEPT_TOKEN(aux_sym__simple_variable_name_token1);
      if (lookahead == 'c') ADVANCE(661);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(879);
      END_STATE();
    case 876:
      ACCEPT_TOKEN(aux_sym__simple_variable_name_token1);
      if (lookahead == 'n') ADVANCE(550);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(879);
      END_STATE();
    case 877:
      ACCEPT_TOKEN(aux_sym__simple_variable_name_token1);
      if (lookahead == 's') ADVANCE(874);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(879);
      END_STATE();
    case 878:
      ACCEPT_TOKEN(aux_sym__simple_variable_name_token1);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(810);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(879);
      END_STATE();
    case 879:
      ACCEPT_TOKEN(aux_sym__simple_variable_name_token1);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(879);
      END_STATE();
    case 880:
      ACCEPT_TOKEN(aux_sym__multiline_variable_name_token1);
      if (lookahead == '\n') ADVANCE(689);
      if (lookahead == '!') ADVANCE(670);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '%') ADVANCE(649);
      if (lookahead == '&') ADVANCE(601);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(654);
      if (lookahead == ')') ADVANCE(655);
      if (lookahead == '*') ADVANCE(638);
      if (lookahead == '+') ADVANCE(630);
      if (lookahead == '-') ADVANCE(633);
      if (lookahead == '/') ADVANCE(644);
      if (lookahead == '0') ADVANCE(943);
      if (lookahead == ';') ADVANCE(555);
      if (lookahead == '<') ADVANCE(607);
      if (lookahead == '=') ADVANCE(560);
      if (lookahead == '>') ADVANCE(616);
      if (lookahead == '?') ADVANCE(761);
      if (lookahead == '@') ADVANCE(942);
      if (lookahead == '\\') ADVANCE(355);
      if (lookahead == '^') ADVANCE(596);
      if (lookahead == '_') ADVANCE(947);
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == '|') ADVANCE(592);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(12)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(807);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(873);
      if (lookahead != 0 &&
          (lookahead < '[' || '}' < lookahead)) ADVANCE(958);
      END_STATE();
    case 881:
      ACCEPT_TOKEN(aux_sym__multiline_variable_name_token1);
      if (lookahead == '\n') ADVANCE(692);
      if (lookahead == '!') ADVANCE(669);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(780);
      if (lookahead == '&') ADVANCE(602);
      if (lookahead == '*') ADVANCE(637);
      if (lookahead == '-') ADVANCE(632);
      if (lookahead == '0') ADVANCE(945);
      if (lookahead == ';') ADVANCE(554);
      if (lookahead == '<') ADVANCE(612);
      if (lookahead == '>') ADVANCE(618);
      if (lookahead == '?') ADVANCE(760);
      if (lookahead == '@') ADVANCE(941);
      if (lookahead == '\\') ADVANCE(373);
      if (lookahead == '_') ADVANCE(948);
      if (lookahead == 'e') ADVANCE(877);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(150)
      if (('1' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(879);
      END_STATE();
    case 882:
      ACCEPT_TOKEN(aux_sym__multiline_variable_name_token1);
      if (lookahead == '\n') ADVANCE(693);
      if (lookahead == '!') ADVANCE(672);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(602);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(654);
      if (lookahead == '*') ADVANCE(640);
      if (lookahead == '-') ADVANCE(636);
      if (lookahead == '0') ADVANCE(943);
      if (lookahead == ';') ADVANCE(554);
      if (lookahead == '<') ADVANCE(609);
      if (lookahead == '=') ADVANCE(952);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '?') ADVANCE(761);
      if (lookahead == '@') ADVANCE(942);
      if (lookahead == '\\') ADVANCE(357);
      if (lookahead == '_') ADVANCE(947);
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == 'e') ADVANCE(871);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(163)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(807);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(873);
      if (lookahead != 0 &&
          lookahead != ')' &&
          (lookahead < '[' || ']' < lookahead) &&
          (lookahead < '{' || '}' < lookahead)) ADVANCE(958);
      END_STATE();
    case 883:
      ACCEPT_TOKEN(aux_sym__multiline_variable_name_token1);
      if (lookahead == '\n') ADVANCE(695);
      if (lookahead == '!') ADVANCE(669);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(780);
      if (lookahead == '&') ADVANCE(602);
      if (lookahead == '*') ADVANCE(637);
      if (lookahead == '-') ADVANCE(632);
      if (lookahead == '0') ADVANCE(945);
      if (lookahead == ';') ADVANCE(554);
      if (lookahead == '<') ADVANCE(612);
      if (lookahead == '>') ADVANCE(618);
      if (lookahead == '?') ADVANCE(760);
      if (lookahead == '@') ADVANCE(941);
      if (lookahead == '\\') ADVANCE(374);
      if (lookahead == '_') ADVANCE(948);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(175)
      if (('1' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(879);
      END_STATE();
    case 884:
      ACCEPT_TOKEN(aux_sym__multiline_variable_name_token1);
      if (lookahead == '\n') ADVANCE(696);
      if (lookahead == '!') ADVANCE(672);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(602);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(654);
      if (lookahead == '*') ADVANCE(640);
      if (lookahead == '-') ADVANCE(636);
      if (lookahead == '0') ADVANCE(943);
      if (lookahead == ';') ADVANCE(554);
      if (lookahead == '<') ADVANCE(609);
      if (lookahead == '=') ADVANCE(952);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '?') ADVANCE(761);
      if (lookahead == '@') ADVANCE(942);
      if (lookahead == '\\') ADVANCE(358);
      if (lookahead == '_') ADVANCE(947);
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(176)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(807);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(873);
      if (lookahead != 0 &&
          lookahead != ')' &&
          (lookahead < '[' || ']' < lookahead) &&
          (lookahead < '{' || '}' < lookahead)) ADVANCE(958);
      END_STATE();
    case 885:
      ACCEPT_TOKEN(aux_sym__multiline_variable_name_token1);
      if (lookahead == '\n') ADVANCE(698);
      if (lookahead == '!') ADVANCE(669);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(780);
      if (lookahead == '&') ADVANCE(602);
      if (lookahead == '*') ADVANCE(637);
      if (lookahead == '-') ADVANCE(632);
      if (lookahead == '0') ADVANCE(945);
      if (lookahead == ';') ADVANCE(554);
      if (lookahead == '<') ADVANCE(613);
      if (lookahead == '>') ADVANCE(618);
      if (lookahead == '?') ADVANCE(760);
      if (lookahead == '@') ADVANCE(941);
      if (lookahead == '\\') ADVANCE(416);
      if (lookahead == '_') ADVANCE(948);
      if (lookahead == 'e') ADVANCE(877);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(178)
      if (('1' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(879);
      END_STATE();
    case 886:
      ACCEPT_TOKEN(aux_sym__multiline_variable_name_token1);
      if (lookahead == '\n') ADVANCE(699);
      if (lookahead == '!') ADVANCE(672);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(602);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(332);
      if (lookahead == '*') ADVANCE(640);
      if (lookahead == '-') ADVANCE(636);
      if (lookahead == '0') ADVANCE(943);
      if (lookahead == ';') ADVANCE(554);
      if (lookahead == '<') ADVANCE(609);
      if (lookahead == '=') ADVANCE(952);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '?') ADVANCE(761);
      if (lookahead == '@') ADVANCE(942);
      if (lookahead == '\\') ADVANCE(389);
      if (lookahead == '_') ADVANCE(947);
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == 'e') ADVANCE(871);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(179)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(807);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(873);
      if (lookahead != 0 &&
          lookahead != ')' &&
          (lookahead < '[' || ']' < lookahead) &&
          (lookahead < '{' || '}' < lookahead)) ADVANCE(958);
      END_STATE();
    case 887:
      ACCEPT_TOKEN(aux_sym__multiline_variable_name_token1);
      if (lookahead == '\n') ADVANCE(701);
      if (lookahead == '!') ADVANCE(669);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(780);
      if (lookahead == '&') ADVANCE(602);
      if (lookahead == '*') ADVANCE(637);
      if (lookahead == '-') ADVANCE(632);
      if (lookahead == '0') ADVANCE(945);
      if (lookahead == ';') ADVANCE(554);
      if (lookahead == '<') ADVANCE(613);
      if (lookahead == '>') ADVANCE(618);
      if (lookahead == '?') ADVANCE(760);
      if (lookahead == '@') ADVANCE(941);
      if (lookahead == '\\') ADVANCE(398);
      if (lookahead == '_') ADVANCE(948);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(181)
      if (('1' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(879);
      END_STATE();
    case 888:
      ACCEPT_TOKEN(aux_sym__multiline_variable_name_token1);
      if (lookahead == '\n') ADVANCE(702);
      if (lookahead == '!') ADVANCE(672);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(602);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(654);
      if (lookahead == ')') ADVANCE(655);
      if (lookahead == '*') ADVANCE(640);
      if (lookahead == '-') ADVANCE(636);
      if (lookahead == '0') ADVANCE(943);
      if (lookahead == ';') ADVANCE(555);
      if (lookahead == '<') ADVANCE(609);
      if (lookahead == '=') ADVANCE(952);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '?') ADVANCE(761);
      if (lookahead == '@') ADVANCE(942);
      if (lookahead == '\\') ADVANCE(359);
      if (lookahead == '_') ADVANCE(947);
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(182)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(807);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(873);
      if (lookahead != 0 &&
          (lookahead < '[' || ']' < lookahead) &&
          (lookahead < '{' || '}' < lookahead)) ADVANCE(958);
      END_STATE();
    case 889:
      ACCEPT_TOKEN(aux_sym__multiline_variable_name_token1);
      if (lookahead == '\n') ADVANCE(704);
      if (lookahead == '!') ADVANCE(669);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(780);
      if (lookahead == '&') ADVANCE(602);
      if (lookahead == '*') ADVANCE(637);
      if (lookahead == '-') ADVANCE(632);
      if (lookahead == '0') ADVANCE(945);
      if (lookahead == ';') ADVANCE(555);
      if (lookahead == '<') ADVANCE(612);
      if (lookahead == '>') ADVANCE(618);
      if (lookahead == '?') ADVANCE(760);
      if (lookahead == '@') ADVANCE(941);
      if (lookahead == '\\') ADVANCE(375);
      if (lookahead == '_') ADVANCE(948);
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(184)
      if (('1' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(879);
      END_STATE();
    case 890:
      ACCEPT_TOKEN(aux_sym__multiline_variable_name_token1);
      if (lookahead == '\n') ADVANCE(705);
      if (lookahead == '!') ADVANCE(672);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(602);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(654);
      if (lookahead == '*') ADVANCE(640);
      if (lookahead == '-') ADVANCE(636);
      if (lookahead == '0') ADVANCE(943);
      if (lookahead == ';') ADVANCE(555);
      if (lookahead == '<') ADVANCE(609);
      if (lookahead == '=') ADVANCE(952);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '?') ADVANCE(761);
      if (lookahead == '@') ADVANCE(942);
      if (lookahead == '\\') ADVANCE(390);
      if (lookahead == '_') ADVANCE(947);
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(185)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(807);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(873);
      if (lookahead != 0 &&
          lookahead != ')' &&
          (lookahead < '[' || ']' < lookahead) &&
          (lookahead < '{' || '}' < lookahead)) ADVANCE(958);
      END_STATE();
    case 891:
      ACCEPT_TOKEN(aux_sym__multiline_variable_name_token1);
      if (lookahead == '\n') ADVANCE(707);
      if (lookahead == '!') ADVANCE(669);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(780);
      if (lookahead == '&') ADVANCE(602);
      if (lookahead == ')') ADVANCE(655);
      if (lookahead == '*') ADVANCE(637);
      if (lookahead == '-') ADVANCE(632);
      if (lookahead == '0') ADVANCE(945);
      if (lookahead == ';') ADVANCE(555);
      if (lookahead == '<') ADVANCE(612);
      if (lookahead == '>') ADVANCE(618);
      if (lookahead == '?') ADVANCE(760);
      if (lookahead == '@') ADVANCE(941);
      if (lookahead == '\\') ADVANCE(376);
      if (lookahead == '_') ADVANCE(948);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(187)
      if (('1' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(879);
      END_STATE();
    case 892:
      ACCEPT_TOKEN(aux_sym__multiline_variable_name_token1);
      if (lookahead == '\n') ADVANCE(708);
      if (lookahead == '!') ADVANCE(672);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(602);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(332);
      if (lookahead == '*') ADVANCE(640);
      if (lookahead == '-') ADVANCE(636);
      if (lookahead == '0') ADVANCE(943);
      if (lookahead == ';') ADVANCE(554);
      if (lookahead == '<') ADVANCE(609);
      if (lookahead == '=') ADVANCE(952);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '?') ADVANCE(761);
      if (lookahead == '@') ADVANCE(942);
      if (lookahead == '\\') ADVANCE(404);
      if (lookahead == '_') ADVANCE(947);
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(188)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(807);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(873);
      if (lookahead != 0 &&
          lookahead != ')' &&
          (lookahead < '[' || ']' < lookahead) &&
          (lookahead < '{' || '}' < lookahead)) ADVANCE(958);
      END_STATE();
    case 893:
      ACCEPT_TOKEN(aux_sym__multiline_variable_name_token1);
      if (lookahead == '\n') ADVANCE(710);
      if (lookahead == '!') ADVANCE(669);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(780);
      if (lookahead == '&') ADVANCE(602);
      if (lookahead == '*') ADVANCE(637);
      if (lookahead == '-') ADVANCE(632);
      if (lookahead == '0') ADVANCE(945);
      if (lookahead == ';') ADVANCE(555);
      if (lookahead == '<') ADVANCE(612);
      if (lookahead == '>') ADVANCE(618);
      if (lookahead == '?') ADVANCE(760);
      if (lookahead == '@') ADVANCE(941);
      if (lookahead == '\\') ADVANCE(406);
      if (lookahead == '_') ADVANCE(948);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(190)
      if (('1' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(879);
      END_STATE();
    case 894:
      ACCEPT_TOKEN(aux_sym__multiline_variable_name_token1);
      if (lookahead == '\n') ADVANCE(711);
      if (lookahead == '!') ADVANCE(672);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(602);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(332);
      if (lookahead == ')') ADVANCE(655);
      if (lookahead == '*') ADVANCE(640);
      if (lookahead == '-') ADVANCE(636);
      if (lookahead == '0') ADVANCE(943);
      if (lookahead == ';') ADVANCE(555);
      if (lookahead == '<') ADVANCE(609);
      if (lookahead == '=') ADVANCE(952);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '?') ADVANCE(761);
      if (lookahead == '@') ADVANCE(942);
      if (lookahead == '\\') ADVANCE(391);
      if (lookahead == '_') ADVANCE(947);
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(191)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(807);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(873);
      if (lookahead != 0 &&
          (lookahead < '[' || ']' < lookahead) &&
          (lookahead < '{' || '}' < lookahead)) ADVANCE(958);
      END_STATE();
    case 895:
      ACCEPT_TOKEN(aux_sym__multiline_variable_name_token1);
      if (lookahead == '\n') ADVANCE(713);
      if (lookahead == '!') ADVANCE(669);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(780);
      if (lookahead == '&') ADVANCE(602);
      if (lookahead == '*') ADVANCE(637);
      if (lookahead == '-') ADVANCE(632);
      if (lookahead == '0') ADVANCE(945);
      if (lookahead == ';') ADVANCE(555);
      if (lookahead == '<') ADVANCE(613);
      if (lookahead == '>') ADVANCE(618);
      if (lookahead == '?') ADVANCE(760);
      if (lookahead == '@') ADVANCE(941);
      if (lookahead == '\\') ADVANCE(399);
      if (lookahead == '_') ADVANCE(948);
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(193)
      if (('1' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(879);
      END_STATE();
    case 896:
      ACCEPT_TOKEN(aux_sym__multiline_variable_name_token1);
      if (lookahead == '\n') ADVANCE(714);
      if (lookahead == '!') ADVANCE(672);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(602);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(332);
      if (lookahead == '*') ADVANCE(640);
      if (lookahead == '-') ADVANCE(636);
      if (lookahead == '0') ADVANCE(943);
      if (lookahead == ';') ADVANCE(555);
      if (lookahead == '<') ADVANCE(609);
      if (lookahead == '=') ADVANCE(952);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '?') ADVANCE(761);
      if (lookahead == '@') ADVANCE(942);
      if (lookahead == '\\') ADVANCE(408);
      if (lookahead == '_') ADVANCE(947);
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(194)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(807);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(873);
      if (lookahead != 0 &&
          lookahead != ')' &&
          (lookahead < '[' || ']' < lookahead) &&
          (lookahead < '{' || '}' < lookahead)) ADVANCE(958);
      END_STATE();
    case 897:
      ACCEPT_TOKEN(aux_sym__multiline_variable_name_token1);
      if (lookahead == '\n') ADVANCE(716);
      if (lookahead == '!') ADVANCE(669);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(780);
      if (lookahead == '&') ADVANCE(602);
      if (lookahead == '*') ADVANCE(637);
      if (lookahead == '-') ADVANCE(632);
      if (lookahead == '0') ADVANCE(945);
      if (lookahead == ';') ADVANCE(555);
      if (lookahead == '<') ADVANCE(613);
      if (lookahead == '>') ADVANCE(618);
      if (lookahead == '?') ADVANCE(760);
      if (lookahead == '@') ADVANCE(941);
      if (lookahead == '\\') ADVANCE(409);
      if (lookahead == '_') ADVANCE(948);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(196)
      if (('1' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(879);
      END_STATE();
    case 898:
      ACCEPT_TOKEN(aux_sym__multiline_variable_name_token1);
      if (lookahead == '\n') ADVANCE(717);
      if (lookahead == '!') ADVANCE(672);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(602);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(332);
      if (lookahead == '*') ADVANCE(640);
      if (lookahead == '-') ADVANCE(636);
      if (lookahead == '0') ADVANCE(943);
      if (lookahead == ';') ADVANCE(554);
      if (lookahead == '<') ADVANCE(609);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '?') ADVANCE(761);
      if (lookahead == '@') ADVANCE(942);
      if (lookahead == '\\') ADVANCE(360);
      if (lookahead == '_') ADVANCE(947);
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == 'e') ADVANCE(871);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(197)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(807);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(873);
      if (lookahead != 0 &&
          lookahead != ')' &&
          (lookahead < '[' || ']' < lookahead) &&
          (lookahead < '{' || '}' < lookahead)) ADVANCE(958);
      END_STATE();
    case 899:
      ACCEPT_TOKEN(aux_sym__multiline_variable_name_token1);
      if (lookahead == '\n') ADVANCE(719);
      if (lookahead == '!') ADVANCE(669);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(780);
      if (lookahead == '&') ADVANCE(602);
      if (lookahead == ')') ADVANCE(655);
      if (lookahead == '*') ADVANCE(637);
      if (lookahead == '-') ADVANCE(632);
      if (lookahead == '0') ADVANCE(945);
      if (lookahead == ';') ADVANCE(555);
      if (lookahead == '<') ADVANCE(613);
      if (lookahead == '>') ADVANCE(618);
      if (lookahead == '?') ADVANCE(760);
      if (lookahead == '@') ADVANCE(941);
      if (lookahead == '\\') ADVANCE(400);
      if (lookahead == '_') ADVANCE(948);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(199)
      if (('1' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(879);
      END_STATE();
    case 900:
      ACCEPT_TOKEN(aux_sym__multiline_variable_name_token1);
      if (lookahead == '\n') ADVANCE(720);
      if (lookahead == '!') ADVANCE(672);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(602);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(332);
      if (lookahead == '*') ADVANCE(640);
      if (lookahead == '-') ADVANCE(636);
      if (lookahead == '0') ADVANCE(943);
      if (lookahead == ';') ADVANCE(554);
      if (lookahead == '<') ADVANCE(610);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '?') ADVANCE(761);
      if (lookahead == '@') ADVANCE(942);
      if (lookahead == '\\') ADVANCE(410);
      if (lookahead == '_') ADVANCE(947);
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == 'e') ADVANCE(871);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(200)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(807);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(873);
      if (lookahead != 0 &&
          lookahead != ')' &&
          (lookahead < '[' || ']' < lookahead) &&
          (lookahead < '{' || '}' < lookahead)) ADVANCE(958);
      END_STATE();
    case 901:
      ACCEPT_TOKEN(aux_sym__multiline_variable_name_token1);
      if (lookahead == '\n') ADVANCE(722);
      if (lookahead == '!') ADVANCE(669);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(780);
      if (lookahead == '&') ADVANCE(327);
      if (lookahead == '*') ADVANCE(637);
      if (lookahead == '-') ADVANCE(632);
      if (lookahead == '0') ADVANCE(945);
      if (lookahead == '<') ADVANCE(612);
      if (lookahead == '>') ADVANCE(618);
      if (lookahead == '?') ADVANCE(760);
      if (lookahead == '@') ADVANCE(941);
      if (lookahead == '\\') ADVANCE(378);
      if (lookahead == '_') ADVANCE(948);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(202)
      if (('1' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(879);
      END_STATE();
    case 902:
      ACCEPT_TOKEN(aux_sym__multiline_variable_name_token1);
      if (lookahead == '\n') ADVANCE(723);
      if (lookahead == '!') ADVANCE(672);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(602);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(332);
      if (lookahead == '*') ADVANCE(640);
      if (lookahead == '-') ADVANCE(636);
      if (lookahead == '0') ADVANCE(943);
      if (lookahead == ';') ADVANCE(554);
      if (lookahead == '<') ADVANCE(609);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '?') ADVANCE(761);
      if (lookahead == '@') ADVANCE(942);
      if (lookahead == '\\') ADVANCE(361);
      if (lookahead == '_') ADVANCE(947);
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(203)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(807);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(873);
      if (lookahead != 0 &&
          lookahead != ')' &&
          (lookahead < '[' || ']' < lookahead) &&
          (lookahead < '{' || '}' < lookahead)) ADVANCE(958);
      END_STATE();
    case 903:
      ACCEPT_TOKEN(aux_sym__multiline_variable_name_token1);
      if (lookahead == '\n') ADVANCE(725);
      if (lookahead == '!') ADVANCE(669);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(780);
      if (lookahead == '&') ADVANCE(327);
      if (lookahead == '*') ADVANCE(637);
      if (lookahead == '-') ADVANCE(632);
      if (lookahead == '0') ADVANCE(945);
      if (lookahead == '<') ADVANCE(613);
      if (lookahead == '>') ADVANCE(618);
      if (lookahead == '?') ADVANCE(760);
      if (lookahead == '@') ADVANCE(941);
      if (lookahead == '\\') ADVANCE(401);
      if (lookahead == '_') ADVANCE(948);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(205)
      if (('1' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(879);
      END_STATE();
    case 904:
      ACCEPT_TOKEN(aux_sym__multiline_variable_name_token1);
      if (lookahead == '\n') ADVANCE(726);
      if (lookahead == '!') ADVANCE(672);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(602);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(332);
      if (lookahead == ')') ADVANCE(655);
      if (lookahead == '*') ADVANCE(640);
      if (lookahead == '-') ADVANCE(636);
      if (lookahead == '0') ADVANCE(943);
      if (lookahead == ';') ADVANCE(555);
      if (lookahead == '<') ADVANCE(609);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '?') ADVANCE(761);
      if (lookahead == '@') ADVANCE(942);
      if (lookahead == '\\') ADVANCE(362);
      if (lookahead == '_') ADVANCE(947);
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(206)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(807);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(873);
      if (lookahead != 0 &&
          (lookahead < '[' || ']' < lookahead) &&
          (lookahead < '{' || '}' < lookahead)) ADVANCE(958);
      END_STATE();
    case 905:
      ACCEPT_TOKEN(aux_sym__multiline_variable_name_token1);
      if (lookahead == '\n') ADVANCE(728);
      if (lookahead == '!') ADVANCE(669);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(780);
      if (lookahead == '&') ADVANCE(327);
      if (lookahead == '*') ADVANCE(637);
      if (lookahead == '-') ADVANCE(632);
      if (lookahead == '0') ADVANCE(945);
      if (lookahead == '<') ADVANCE(611);
      if (lookahead == '>') ADVANCE(618);
      if (lookahead == '?') ADVANCE(760);
      if (lookahead == '@') ADVANCE(941);
      if (lookahead == '\\') ADVANCE(407);
      if (lookahead == '_') ADVANCE(948);
      if (lookahead == '|') ADVANCE(354);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(208)
      if (('1' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(879);
      END_STATE();
    case 906:
      ACCEPT_TOKEN(aux_sym__multiline_variable_name_token1);
      if (lookahead == '\n') ADVANCE(729);
      if (lookahead == '!') ADVANCE(672);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(602);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(332);
      if (lookahead == '*') ADVANCE(640);
      if (lookahead == '-') ADVANCE(636);
      if (lookahead == '0') ADVANCE(943);
      if (lookahead == ';') ADVANCE(554);
      if (lookahead == '<') ADVANCE(610);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '?') ADVANCE(761);
      if (lookahead == '@') ADVANCE(942);
      if (lookahead == '\\') ADVANCE(411);
      if (lookahead == '_') ADVANCE(947);
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(209)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(807);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(873);
      if (lookahead != 0 &&
          lookahead != ')' &&
          (lookahead < '[' || ']' < lookahead) &&
          (lookahead < '{' || '}' < lookahead)) ADVANCE(958);
      END_STATE();
    case 907:
      ACCEPT_TOKEN(aux_sym__multiline_variable_name_token1);
      if (lookahead == '\n') ADVANCE(731);
      if (lookahead == '!') ADVANCE(669);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(780);
      if (lookahead == '&') ADVANCE(599);
      if (lookahead == '*') ADVANCE(637);
      if (lookahead == '-') ADVANCE(632);
      if (lookahead == '0') ADVANCE(945);
      if (lookahead == ';') ADVANCE(555);
      if (lookahead == '?') ADVANCE(760);
      if (lookahead == '@') ADVANCE(941);
      if (lookahead == '\\') ADVANCE(383);
      if (lookahead == '_') ADVANCE(948);
      if (lookahead == 'i') ADVANCE(876);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(211)
      if (('1' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(879);
      END_STATE();
    case 908:
      ACCEPT_TOKEN(aux_sym__multiline_variable_name_token1);
      if (lookahead == '\n') ADVANCE(732);
      if (lookahead == '!') ADVANCE(672);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(602);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(332);
      if (lookahead == '*') ADVANCE(640);
      if (lookahead == '-') ADVANCE(636);
      if (lookahead == '0') ADVANCE(943);
      if (lookahead == ';') ADVANCE(555);
      if (lookahead == '<') ADVANCE(609);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '?') ADVANCE(761);
      if (lookahead == '@') ADVANCE(942);
      if (lookahead == '\\') ADVANCE(392);
      if (lookahead == '_') ADVANCE(947);
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(212)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(807);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(873);
      if (lookahead != 0 &&
          lookahead != ')' &&
          (lookahead < '[' || ']' < lookahead) &&
          (lookahead < '{' || '}' < lookahead)) ADVANCE(958);
      END_STATE();
    case 909:
      ACCEPT_TOKEN(aux_sym__multiline_variable_name_token1);
      if (lookahead == '\n') ADVANCE(734);
      if (lookahead == '!') ADVANCE(672);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(327);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(654);
      if (lookahead == '*') ADVANCE(640);
      if (lookahead == '-') ADVANCE(636);
      if (lookahead == '0') ADVANCE(943);
      if (lookahead == '<') ADVANCE(609);
      if (lookahead == '=') ADVANCE(952);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '?') ADVANCE(761);
      if (lookahead == '@') ADVANCE(942);
      if (lookahead == '\\') ADVANCE(363);
      if (lookahead == '_') ADVANCE(947);
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(214)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(807);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(873);
      if (lookahead != 0 &&
          lookahead != ')' &&
          (lookahead < ';' || ']' < lookahead) &&
          (lookahead < '{' || '}' < lookahead)) ADVANCE(958);
      END_STATE();
    case 910:
      ACCEPT_TOKEN(aux_sym__multiline_variable_name_token1);
      if (lookahead == '\n') ADVANCE(736);
      if (lookahead == '!') ADVANCE(672);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(602);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(332);
      if (lookahead == ')') ADVANCE(655);
      if (lookahead == '*') ADVANCE(640);
      if (lookahead == '-') ADVANCE(636);
      if (lookahead == '0') ADVANCE(943);
      if (lookahead == ';') ADVANCE(555);
      if (lookahead == '<') ADVANCE(610);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '?') ADVANCE(761);
      if (lookahead == '@') ADVANCE(942);
      if (lookahead == '\\') ADVANCE(412);
      if (lookahead == '_') ADVANCE(947);
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(216)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(807);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(873);
      if (lookahead != 0 &&
          (lookahead < '[' || ']' < lookahead) &&
          (lookahead < '{' || '}' < lookahead)) ADVANCE(958);
      END_STATE();
    case 911:
      ACCEPT_TOKEN(aux_sym__multiline_variable_name_token1);
      if (lookahead == '\n') ADVANCE(738);
      if (lookahead == '!') ADVANCE(672);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(602);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(332);
      if (lookahead == '*') ADVANCE(640);
      if (lookahead == '-') ADVANCE(636);
      if (lookahead == '0') ADVANCE(943);
      if (lookahead == ';') ADVANCE(555);
      if (lookahead == '<') ADVANCE(610);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '?') ADVANCE(761);
      if (lookahead == '@') ADVANCE(942);
      if (lookahead == '\\') ADVANCE(405);
      if (lookahead == '_') ADVANCE(947);
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(218)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(807);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(873);
      if (lookahead != 0 &&
          lookahead != ')' &&
          (lookahead < '[' || ']' < lookahead) &&
          (lookahead < '{' || '}' < lookahead)) ADVANCE(958);
      END_STATE();
    case 912:
      ACCEPT_TOKEN(aux_sym__multiline_variable_name_token1);
      if (lookahead == '\n') ADVANCE(740);
      if (lookahead == '!') ADVANCE(672);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(327);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(332);
      if (lookahead == '*') ADVANCE(640);
      if (lookahead == '-') ADVANCE(636);
      if (lookahead == '0') ADVANCE(943);
      if (lookahead == '<') ADVANCE(609);
      if (lookahead == '=') ADVANCE(952);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '?') ADVANCE(761);
      if (lookahead == '@') ADVANCE(942);
      if (lookahead == '\\') ADVANCE(393);
      if (lookahead == '_') ADVANCE(947);
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(220)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(807);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(873);
      if (lookahead != 0 &&
          lookahead != ')' &&
          (lookahead < ';' || ']' < lookahead) &&
          (lookahead < '{' || '}' < lookahead)) ADVANCE(958);
      END_STATE();
    case 913:
      ACCEPT_TOKEN(aux_sym__multiline_variable_name_token1);
      if (lookahead == '\n') ADVANCE(742);
      if (lookahead == '!') ADVANCE(672);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(327);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(332);
      if (lookahead == '*') ADVANCE(640);
      if (lookahead == '-') ADVANCE(636);
      if (lookahead == '0') ADVANCE(943);
      if (lookahead == '<') ADVANCE(609);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '?') ADVANCE(761);
      if (lookahead == '@') ADVANCE(942);
      if (lookahead == '\\') ADVANCE(366);
      if (lookahead == '_') ADVANCE(947);
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(222)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(807);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(873);
      if (lookahead != 0 &&
          lookahead != ')' &&
          lookahead != ';' &&
          (lookahead < '[' || ']' < lookahead) &&
          (lookahead < '{' || '}' < lookahead)) ADVANCE(958);
      END_STATE();
    case 914:
      ACCEPT_TOKEN(aux_sym__multiline_variable_name_token1);
      if (lookahead == '\n') ADVANCE(744);
      if (lookahead == '!') ADVANCE(672);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(327);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(332);
      if (lookahead == '*') ADVANCE(640);
      if (lookahead == '-') ADVANCE(636);
      if (lookahead == '0') ADVANCE(943);
      if (lookahead == '<') ADVANCE(610);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '?') ADVANCE(761);
      if (lookahead == '@') ADVANCE(942);
      if (lookahead == '\\') ADVANCE(414);
      if (lookahead == '_') ADVANCE(947);
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(224)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(807);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(873);
      if (lookahead != 0 &&
          lookahead != ')' &&
          lookahead != ';' &&
          (lookahead < '[' || ']' < lookahead) &&
          (lookahead < '{' || '}' < lookahead)) ADVANCE(958);
      END_STATE();
    case 915:
      ACCEPT_TOKEN(aux_sym__multiline_variable_name_token1);
      if (lookahead == '\n') ADVANCE(746);
      if (lookahead == '!') ADVANCE(672);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(327);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(332);
      if (lookahead == '*') ADVANCE(640);
      if (lookahead == '-') ADVANCE(636);
      if (lookahead == '0') ADVANCE(943);
      if (lookahead == '<') ADVANCE(608);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '?') ADVANCE(761);
      if (lookahead == '@') ADVANCE(942);
      if (lookahead == '\\') ADVANCE(395);
      if (lookahead == '_') ADVANCE(947);
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == '|') ADVANCE(354);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(226)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(807);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(873);
      if (lookahead != 0 &&
          lookahead != ')' &&
          lookahead != ';' &&
          (lookahead < '[' || ']' < lookahead) &&
          (lookahead < '{' || '}' < lookahead)) ADVANCE(958);
      END_STATE();
    case 916:
      ACCEPT_TOKEN(aux_sym__multiline_variable_name_token1);
      if (lookahead == '\n') ADVANCE(748);
      if (lookahead == '!') ADVANCE(672);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(599);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(332);
      if (lookahead == '*') ADVANCE(640);
      if (lookahead == '-') ADVANCE(636);
      if (lookahead == '0') ADVANCE(943);
      if (lookahead == ';') ADVANCE(555);
      if (lookahead == '<') ADVANCE(333);
      if (lookahead == '>') ADVANCE(334);
      if (lookahead == '?') ADVANCE(761);
      if (lookahead == '@') ADVANCE(942);
      if (lookahead == '\\') ADVANCE(377);
      if (lookahead == '_') ADVANCE(947);
      if (lookahead == '`') ADVANCE(855);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(228)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(807);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(873);
      if (lookahead != 0 &&
          lookahead != ')' &&
          (lookahead < '[' || ']' < lookahead) &&
          (lookahead < '{' || '}' < lookahead)) ADVANCE(958);
      END_STATE();
    case 917:
      ACCEPT_TOKEN(aux_sym__multiline_variable_name_token1);
      if (lookahead == '\n') ADVANCE(750);
      if (lookahead == '!') ADVANCE(672);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(332);
      if (lookahead == '*') ADVANCE(640);
      if (lookahead == '-') ADVANCE(636);
      if (lookahead == '0') ADVANCE(943);
      if (lookahead == '<') ADVANCE(333);
      if (lookahead == '>') ADVANCE(334);
      if (lookahead == '?') ADVANCE(761);
      if (lookahead == '@') ADVANCE(942);
      if (lookahead == '\\') ADVANCE(381);
      if (lookahead == '_') ADVANCE(947);
      if (lookahead == '`') ADVANCE(855);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(230)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(807);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(873);
      if (lookahead != 0 &&
          (lookahead < '&' || ')' < lookahead) &&
          lookahead != ';' &&
          (lookahead < '[' || ']' < lookahead) &&
          (lookahead < '{' || '}' < lookahead)) ADVANCE(958);
      END_STATE();
    case 918:
      ACCEPT_TOKEN(aux_sym__multiline_variable_name_token1);
      if (lookahead == '!') ADVANCE(670);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '%') ADVANCE(649);
      if (lookahead == '&') ADVANCE(601);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(654);
      if (lookahead == '*') ADVANCE(638);
      if (lookahead == '+') ADVANCE(630);
      if (lookahead == '-') ADVANCE(633);
      if (lookahead == '/') ADVANCE(644);
      if (lookahead == '0') ADVANCE(943);
      if (lookahead == '<') ADVANCE(607);
      if (lookahead == '=') ADVANCE(560);
      if (lookahead == '>') ADVANCE(616);
      if (lookahead == '?') ADVANCE(761);
      if (lookahead == '@') ADVANCE(942);
      if (lookahead == '\\') ADVANCE(356);
      if (lookahead == ']') ADVANCE(675);
      if (lookahead == '^') ADVANCE(596);
      if (lookahead == '_') ADVANCE(947);
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == '|') ADVANCE(592);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(240)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(807);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(873);
      if (lookahead != 0 &&
          lookahead != ')' &&
          (lookahead < ';' || '}' < lookahead)) ADVANCE(958);
      END_STATE();
    case 919:
      ACCEPT_TOKEN(aux_sym__multiline_variable_name_token1);
      if (lookahead == '!') ADVANCE(672);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(346);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(332);
      if (lookahead == '*') ADVANCE(640);
      if (lookahead == '-') ADVANCE(636);
      if (lookahead == '0') ADVANCE(943);
      if (lookahead == '<') ADVANCE(608);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '?') ADVANCE(761);
      if (lookahead == '@') ADVANCE(942);
      if (lookahead == '\\') ADVANCE(372);
      if (lookahead == '_') ADVANCE(947);
      if (lookahead == '`') ADVANCE(855);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(247)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(807);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(873);
      if (lookahead != 0 &&
          lookahead != ')' &&
          lookahead != ';' &&
          (lookahead < '[' || ']' < lookahead) &&
          (lookahead < '{' || '}' < lookahead)) ADVANCE(958);
      END_STATE();
    case 920:
      ACCEPT_TOKEN(aux_sym__multiline_variable_name_token1);
      if (lookahead == '!') ADVANCE(672);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(327);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(654);
      if (lookahead == '*') ADVANCE(640);
      if (lookahead == '-') ADVANCE(636);
      if (lookahead == '0') ADVANCE(943);
      if (lookahead == '<') ADVANCE(609);
      if (lookahead == '=') ADVANCE(952);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '?') ADVANCE(761);
      if (lookahead == '@') ADVANCE(942);
      if (lookahead == '\\') ADVANCE(364);
      if (lookahead == '_') ADVANCE(947);
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(248)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(807);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(873);
      if (lookahead != 0 &&
          lookahead != ')' &&
          (lookahead < ';' || ']' < lookahead) &&
          (lookahead < '{' || '}' < lookahead)) ADVANCE(958);
      END_STATE();
    case 921:
      ACCEPT_TOKEN(aux_sym__multiline_variable_name_token1);
      if (lookahead == '!') ADVANCE(672);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(327);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(332);
      if (lookahead == '*') ADVANCE(640);
      if (lookahead == '-') ADVANCE(636);
      if (lookahead == '0') ADVANCE(943);
      if (lookahead == '<') ADVANCE(609);
      if (lookahead == '=') ADVANCE(952);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '?') ADVANCE(761);
      if (lookahead == '@') ADVANCE(942);
      if (lookahead == '\\') ADVANCE(394);
      if (lookahead == '_') ADVANCE(947);
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(249)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(807);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(873);
      if (lookahead != 0 &&
          lookahead != ')' &&
          (lookahead < ';' || ']' < lookahead) &&
          (lookahead < '{' || '}' < lookahead)) ADVANCE(958);
      END_STATE();
    case 922:
      ACCEPT_TOKEN(aux_sym__multiline_variable_name_token1);
      if (lookahead == '!') ADVANCE(672);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(327);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(332);
      if (lookahead == '*') ADVANCE(640);
      if (lookahead == '-') ADVANCE(636);
      if (lookahead == '0') ADVANCE(943);
      if (lookahead == '<') ADVANCE(609);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '?') ADVANCE(761);
      if (lookahead == '@') ADVANCE(942);
      if (lookahead == '\\') ADVANCE(365);
      if (lookahead == ']') ADVANCE(675);
      if (lookahead == '_') ADVANCE(947);
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(250)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(807);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(873);
      if (lookahead != 0 &&
          lookahead != ')' &&
          lookahead != ';' &&
          lookahead != '[' &&
          (lookahead < '{' || '}' < lookahead)) ADVANCE(958);
      END_STATE();
    case 923:
      ACCEPT_TOKEN(aux_sym__multiline_variable_name_token1);
      if (lookahead == '!') ADVANCE(672);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(327);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(332);
      if (lookahead == '*') ADVANCE(640);
      if (lookahead == '-') ADVANCE(636);
      if (lookahead == '0') ADVANCE(943);
      if (lookahead == '<') ADVANCE(609);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '?') ADVANCE(761);
      if (lookahead == '@') ADVANCE(942);
      if (lookahead == '\\') ADVANCE(368);
      if (lookahead == '_') ADVANCE(947);
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(251)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(807);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(873);
      if (lookahead != 0 &&
          lookahead != ')' &&
          lookahead != ';' &&
          (lookahead < '[' || ']' < lookahead) &&
          (lookahead < '{' || '}' < lookahead)) ADVANCE(958);
      END_STATE();
    case 924:
      ACCEPT_TOKEN(aux_sym__multiline_variable_name_token1);
      if (lookahead == '!') ADVANCE(672);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(327);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(332);
      if (lookahead == '*') ADVANCE(640);
      if (lookahead == '-') ADVANCE(636);
      if (lookahead == '0') ADVANCE(943);
      if (lookahead == '<') ADVANCE(610);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '?') ADVANCE(761);
      if (lookahead == '@') ADVANCE(942);
      if (lookahead == '\\') ADVANCE(413);
      if (lookahead == ']') ADVANCE(675);
      if (lookahead == '_') ADVANCE(947);
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(252)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(807);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(873);
      if (lookahead != 0 &&
          lookahead != ')' &&
          lookahead != ';' &&
          lookahead != '[' &&
          (lookahead < '{' || '}' < lookahead)) ADVANCE(958);
      END_STATE();
    case 925:
      ACCEPT_TOKEN(aux_sym__multiline_variable_name_token1);
      if (lookahead == '!') ADVANCE(672);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '&') ADVANCE(327);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(332);
      if (lookahead == '*') ADVANCE(640);
      if (lookahead == '-') ADVANCE(636);
      if (lookahead == '0') ADVANCE(943);
      if (lookahead == '<') ADVANCE(610);
      if (lookahead == '>') ADVANCE(617);
      if (lookahead == '?') ADVANCE(761);
      if (lookahead == '@') ADVANCE(942);
      if (lookahead == '\\') ADVANCE(415);
      if (lookahead == '_') ADVANCE(947);
      if (lookahead == '`') ADVANCE(855);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(253)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(807);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(873);
      if (lookahead != 0 &&
          lookahead != ')' &&
          lookahead != ';' &&
          (lookahead < '[' || ']' < lookahead) &&
          (lookahead < '{' || '}' < lookahead)) ADVANCE(958);
      END_STATE();
    case 926:
      ACCEPT_TOKEN(aux_sym__multiline_variable_name_token1);
      if (lookahead == '!') ADVANCE(672);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(781);
      if (lookahead == '\'') ADVANCE(329);
      if (lookahead == '(') ADVANCE(332);
      if (lookahead == ')') ADVANCE(655);
      if (lookahead == '*') ADVANCE(640);
      if (lookahead == '-') ADVANCE(636);
      if (lookahead == '0') ADVANCE(943);
      if (lookahead == '<') ADVANCE(333);
      if (lookahead == '>') ADVANCE(334);
      if (lookahead == '?') ADVANCE(761);
      if (lookahead == '@') ADVANCE(942);
      if (lookahead == '\\') ADVANCE(380);
      if (lookahead == '_') ADVANCE(947);
      if (lookahead == '`') ADVANCE(855);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(254)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(807);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(873);
      if (lookahead != 0 &&
          lookahead != '&' &&
          lookahead != ';' &&
          (lookahead < '[' || ']' < lookahead) &&
          (lookahead < '{' || '}' < lookahead)) ADVANCE(958);
      END_STATE();
    case 927:
      ACCEPT_TOKEN(aux_sym__multiline_variable_name_token1);
      if (lookahead == '!') ADVANCE(671);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(780);
      if (lookahead == '%') ADVANCE(650);
      if (lookahead == '&') ADVANCE(600);
      if (lookahead == ')') ADVANCE(655);
      if (lookahead == '*') ADVANCE(639);
      if (lookahead == '+') ADVANCE(631);
      if (lookahead == '-') ADVANCE(634);
      if (lookahead == '/') ADVANCE(645);
      if (lookahead == '0') ADVANCE(945);
      if (lookahead == '<') ADVANCE(615);
      if (lookahead == '=') ADVANCE(562);
      if (lookahead == '>') ADVANCE(620);
      if (lookahead == '?') ADVANCE(760);
      if (lookahead == '@') ADVANCE(941);
      if (lookahead == '\\') ADVANCE(370);
      if (lookahead == '^') ADVANCE(597);
      if (lookahead == '_') ADVANCE(948);
      if (lookahead == '|') ADVANCE(594);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(265)
      if (('1' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(879);
      END_STATE();
    case 928:
      ACCEPT_TOKEN(aux_sym__multiline_variable_name_token1);
      if (lookahead == '!') ADVANCE(671);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(780);
      if (lookahead == '%') ADVANCE(650);
      if (lookahead == '&') ADVANCE(600);
      if (lookahead == ')') ADVANCE(337);
      if (lookahead == '*') ADVANCE(639);
      if (lookahead == '+') ADVANCE(631);
      if (lookahead == '-') ADVANCE(634);
      if (lookahead == '/') ADVANCE(645);
      if (lookahead == '0') ADVANCE(945);
      if (lookahead == '<') ADVANCE(615);
      if (lookahead == '=') ADVANCE(562);
      if (lookahead == '>') ADVANCE(620);
      if (lookahead == '?') ADVANCE(760);
      if (lookahead == '@') ADVANCE(941);
      if (lookahead == '\\') ADVANCE(397);
      if (lookahead == '^') ADVANCE(597);
      if (lookahead == '_') ADVANCE(948);
      if (lookahead == '|') ADVANCE(594);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(266)
      if (('1' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(879);
      END_STATE();
    case 929:
      ACCEPT_TOKEN(aux_sym__multiline_variable_name_token1);
      if (lookahead == '!') ADVANCE(671);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(780);
      if (lookahead == '%') ADVANCE(650);
      if (lookahead == '&') ADVANCE(600);
      if (lookahead == '*') ADVANCE(639);
      if (lookahead == '+') ADVANCE(631);
      if (lookahead == '-') ADVANCE(634);
      if (lookahead == '/') ADVANCE(645);
      if (lookahead == '0') ADVANCE(945);
      if (lookahead == ':') ADVANCE(762);
      if (lookahead == '<') ADVANCE(615);
      if (lookahead == '=') ADVANCE(562);
      if (lookahead == '>') ADVANCE(620);
      if (lookahead == '?') ADVANCE(760);
      if (lookahead == '@') ADVANCE(941);
      if (lookahead == '\\') ADVANCE(369);
      if (lookahead == '^') ADVANCE(597);
      if (lookahead == '_') ADVANCE(948);
      if (lookahead == '|') ADVANCE(594);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(267)
      if (('1' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(879);
      END_STATE();
    case 930:
      ACCEPT_TOKEN(aux_sym__multiline_variable_name_token1);
      if (lookahead == '!') ADVANCE(671);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(780);
      if (lookahead == '%') ADVANCE(650);
      if (lookahead == '&') ADVANCE(600);
      if (lookahead == '*') ADVANCE(639);
      if (lookahead == '+') ADVANCE(631);
      if (lookahead == '-') ADVANCE(634);
      if (lookahead == '/') ADVANCE(645);
      if (lookahead == '0') ADVANCE(945);
      if (lookahead == '<') ADVANCE(615);
      if (lookahead == '=') ADVANCE(562);
      if (lookahead == '>') ADVANCE(620);
      if (lookahead == '?') ADVANCE(760);
      if (lookahead == '@') ADVANCE(941);
      if (lookahead == '\\') ADVANCE(367);
      if (lookahead == ']') ADVANCE(675);
      if (lookahead == '^') ADVANCE(597);
      if (lookahead == '_') ADVANCE(948);
      if (lookahead == '|') ADVANCE(594);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(268)
      if (('1' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(879);
      END_STATE();
    case 931:
      ACCEPT_TOKEN(aux_sym__multiline_variable_name_token1);
      if (lookahead == '!') ADVANCE(671);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(780);
      if (lookahead == '%') ADVANCE(650);
      if (lookahead == '&') ADVANCE(600);
      if (lookahead == '*') ADVANCE(639);
      if (lookahead == '+') ADVANCE(631);
      if (lookahead == '-') ADVANCE(634);
      if (lookahead == '/') ADVANCE(645);
      if (lookahead == '0') ADVANCE(945);
      if (lookahead == '<') ADVANCE(615);
      if (lookahead == '=') ADVANCE(562);
      if (lookahead == '>') ADVANCE(620);
      if (lookahead == '?') ADVANCE(760);
      if (lookahead == '@') ADVANCE(941);
      if (lookahead == '\\') ADVANCE(396);
      if (lookahead == ']') ADVANCE(347);
      if (lookahead == '^') ADVANCE(597);
      if (lookahead == '_') ADVANCE(948);
      if (lookahead == '|') ADVANCE(594);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(269)
      if (('1' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(879);
      END_STATE();
    case 932:
      ACCEPT_TOKEN(aux_sym__multiline_variable_name_token1);
      if (lookahead == '!') ADVANCE(671);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(780);
      if (lookahead == '%') ADVANCE(650);
      if (lookahead == '&') ADVANCE(600);
      if (lookahead == '*') ADVANCE(639);
      if (lookahead == '+') ADVANCE(631);
      if (lookahead == '-') ADVANCE(634);
      if (lookahead == '/') ADVANCE(645);
      if (lookahead == '0') ADVANCE(945);
      if (lookahead == '<') ADVANCE(615);
      if (lookahead == '=') ADVANCE(562);
      if (lookahead == '>') ADVANCE(620);
      if (lookahead == '?') ADVANCE(760);
      if (lookahead == '@') ADVANCE(941);
      if (lookahead == '\\') ADVANCE(371);
      if (lookahead == '^') ADVANCE(597);
      if (lookahead == '_') ADVANCE(948);
      if (lookahead == '|') ADVANCE(594);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(270)
      if (('1' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(879);
      END_STATE();
    case 933:
      ACCEPT_TOKEN(aux_sym__multiline_variable_name_token1);
      if (lookahead == '!') ADVANCE(669);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(780);
      if (lookahead == '&') ADVANCE(327);
      if (lookahead == '*') ADVANCE(637);
      if (lookahead == '-') ADVANCE(632);
      if (lookahead == '0') ADVANCE(945);
      if (lookahead == '<') ADVANCE(612);
      if (lookahead == '>') ADVANCE(618);
      if (lookahead == '?') ADVANCE(760);
      if (lookahead == '@') ADVANCE(941);
      if (lookahead == '\\') ADVANCE(379);
      if (lookahead == ']') ADVANCE(675);
      if (lookahead == '_') ADVANCE(948);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(282)
      if (('1' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(879);
      END_STATE();
    case 934:
      ACCEPT_TOKEN(aux_sym__multiline_variable_name_token1);
      if (lookahead == '!') ADVANCE(669);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(780);
      if (lookahead == '&') ADVANCE(327);
      if (lookahead == '*') ADVANCE(637);
      if (lookahead == '-') ADVANCE(632);
      if (lookahead == '0') ADVANCE(945);
      if (lookahead == '<') ADVANCE(612);
      if (lookahead == '>') ADVANCE(618);
      if (lookahead == '?') ADVANCE(760);
      if (lookahead == '@') ADVANCE(941);
      if (lookahead == '\\') ADVANCE(382);
      if (lookahead == '_') ADVANCE(948);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(283)
      if (('1' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(879);
      END_STATE();
    case 935:
      ACCEPT_TOKEN(aux_sym__multiline_variable_name_token1);
      if (lookahead == '!') ADVANCE(669);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(780);
      if (lookahead == '&') ADVANCE(327);
      if (lookahead == '*') ADVANCE(637);
      if (lookahead == '-') ADVANCE(632);
      if (lookahead == '0') ADVANCE(945);
      if (lookahead == '<') ADVANCE(613);
      if (lookahead == '>') ADVANCE(618);
      if (lookahead == '?') ADVANCE(760);
      if (lookahead == '@') ADVANCE(941);
      if (lookahead == '\\') ADVANCE(402);
      if (lookahead == ']') ADVANCE(675);
      if (lookahead == '_') ADVANCE(948);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(284)
      if (('1' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(879);
      END_STATE();
    case 936:
      ACCEPT_TOKEN(aux_sym__multiline_variable_name_token1);
      if (lookahead == '!') ADVANCE(669);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(780);
      if (lookahead == '&') ADVANCE(327);
      if (lookahead == '*') ADVANCE(637);
      if (lookahead == '-') ADVANCE(632);
      if (lookahead == '0') ADVANCE(945);
      if (lookahead == '<') ADVANCE(613);
      if (lookahead == '>') ADVANCE(618);
      if (lookahead == '?') ADVANCE(760);
      if (lookahead == '@') ADVANCE(941);
      if (lookahead == '\\') ADVANCE(403);
      if (lookahead == '_') ADVANCE(948);
      if (lookahead == '|') ADVANCE(593);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(285)
      if (('1' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(879);
      END_STATE();
    case 937:
      ACCEPT_TOKEN(aux_sym__multiline_variable_name_token1);
      if (lookahead == '!') ADVANCE(669);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(780);
      if (lookahead == ')') ADVANCE(655);
      if (lookahead == '*') ADVANCE(637);
      if (lookahead == '-') ADVANCE(632);
      if (lookahead == '0') ADVANCE(945);
      if (lookahead == '?') ADVANCE(760);
      if (lookahead == '@') ADVANCE(941);
      if (lookahead == '\\') ADVANCE(384);
      if (lookahead == '_') ADVANCE(948);
      if (lookahead == '|') ADVANCE(590);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(286)
      if (('1' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(879);
      END_STATE();
    case 938:
      ACCEPT_TOKEN(aux_sym__multiline_variable_name_token1);
      if (lookahead == '!') ADVANCE(669);
      if (lookahead == '"') ADVANCE(788);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(780);
      if (lookahead == '*') ADVANCE(637);
      if (lookahead == '-') ADVANCE(632);
      if (lookahead == '0') ADVANCE(945);
      if (lookahead == '?') ADVANCE(760);
      if (lookahead == '@') ADVANCE(941);
      if (lookahead == '\\') ADVANCE(385);
      if (lookahead == '_') ADVANCE(948);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(288)
      if (('1' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(879);
      END_STATE();
    case 939:
      ACCEPT_TOKEN(aux_sym__multiline_variable_name_token1);
      if (lookahead == '!') ADVANCE(669);
      if (lookahead == '#') ADVANCE(816);
      if (lookahead == '$') ADVANCE(780);
      if (lookahead == '*') ADVANCE(637);
      if (lookahead == '-') ADVANCE(632);
      if (lookahead == '0') ADVANCE(945);
      if (lookahead == '?') ADVANCE(760);
      if (lookahead == '@') ADVANCE(941);
      if (lookahead == '\\') ADVANCE(388);
      if (lookahead == '_') ADVANCE(948);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(291)
      if (('1' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(879);
      END_STATE();
    case 940:
      ACCEPT_TOKEN(aux_sym__multiline_variable_name_token1);
      if (lookahead == '\\') ADVANCE(15);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(940);
      END_STATE();
    case 941:
      ACCEPT_TOKEN(anon_sym_AT2);
      END_STATE();
    case 942:
      ACCEPT_TOKEN(anon_sym_AT2);
      if (lookahead == '\\') ADVANCE(517);
      if (!sym_word_character_set_1(lookahead)) ADVANCE(958);
      END_STATE();
    case 943:
      ACCEPT_TOKEN(anon_sym_0);
      if (lookahead == '#') ADVANCE(814);
      if (lookahead == '\\') ADVANCE(16);
      if (lookahead == 'x') ADVANCE(872);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(807);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(873);
      if (!sym_word_character_set_6(lookahead)) ADVANCE(958);
      END_STATE();
    case 944:
      ACCEPT_TOKEN(anon_sym_0);
      if (lookahead == '\\') ADVANCE(517);
      if (!sym_word_character_set_1(lookahead)) ADVANCE(958);
      END_STATE();
    case 945:
      ACCEPT_TOKEN(anon_sym_0);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(879);
      END_STATE();
    case 946:
      ACCEPT_TOKEN(anon_sym__);
      if (lookahead == '\\') ADVANCE(517);
      if (!sym_word_character_set_1(lookahead)) ADVANCE(958);
      END_STATE();
    case 947:
      ACCEPT_TOKEN(anon_sym__);
      if (lookahead == '\\') ADVANCE(16);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(873);
      if (!sym_word_character_set_4(lookahead)) ADVANCE(958);
      END_STATE();
    case 948:
      ACCEPT_TOKEN(anon_sym__);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(879);
      END_STATE();
    case 949:
      ACCEPT_TOKEN(sym_word);
      if (lookahead == '.') ADVANCE(777);
      if (lookahead == '\\') ADVANCE(517);
      if (!sym_word_character_set_1(lookahead)) ADVANCE(958);
      END_STATE();
    case 950:
      ACCEPT_TOKEN(sym_word);
      if (lookahead == '0') ADVANCE(803);
      if (lookahead == '\\') ADVANCE(517);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(805);
      if (!sym_word_character_set_1(lookahead)) ADVANCE(958);
      END_STATE();
    case 951:
      ACCEPT_TOKEN(sym_word);
      if (lookahead == '=') ADVANCE(606);
      if (lookahead == '\\') ADVANCE(517);
      if (!sym_word_character_set_2(lookahead)) ADVANCE(958);
      END_STATE();
    case 952:
      ACCEPT_TOKEN(sym_word);
      if (lookahead == '=') ADVANCE(604);
      if (lookahead == '\\') ADVANCE(517);
      if (lookahead == '~') ADVANCE(679);
      if (!sym_word_character_set_2(lookahead)) ADVANCE(958);
      END_STATE();
    case 953:
      ACCEPT_TOKEN(sym_word);
      if (lookahead == '\\') ADVANCE(517);
      if (lookahead == 'a') ADVANCE(954);
      if (!sym_word_character_set_1(lookahead)) ADVANCE(958);
      END_STATE();
    case 954:
      ACCEPT_TOKEN(sym_word);
      if (lookahead == '\\') ADVANCE(517);
      if (lookahead == 'c') ADVANCE(659);
      if (!sym_word_character_set_1(lookahead)) ADVANCE(958);
      END_STATE();
    case 955:
      ACCEPT_TOKEN(sym_word);
      if (lookahead == '\\') ADVANCE(517);
      if (lookahead == 'n') ADVANCE(549);
      if (!sym_word_character_set_1(lookahead)) ADVANCE(958);
      END_STATE();
    case 956:
      ACCEPT_TOKEN(sym_word);
      if (lookahead == '\\') ADVANCE(517);
      if (lookahead == 's') ADVANCE(953);
      if (!sym_word_character_set_1(lookahead)) ADVANCE(958);
      END_STATE();
    case 957:
      ACCEPT_TOKEN(sym_word);
      if (lookahead == '\\') ADVANCE(517);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(805);
      if (!sym_word_character_set_1(lookahead)) ADVANCE(958);
      END_STATE();
    case 958:
      ACCEPT_TOKEN(sym_word);
      if (lookahead == '\\') ADVANCE(517);
      if (!sym_word_character_set_1(lookahead)) ADVANCE(958);
      END_STATE();
    default:
      return false;
  }
}

static bool ts_lex_keywords(TSLexer *lexer, TSStateId state) {
  START_LEXER();
  eof = lexer->eof(lexer);
  switch (state) {
    case 0:
      if (lookahead == 'A') ADVANCE(1);
      if (lookahead == 'E') ADVANCE(2);
      if (lookahead == 'K') ADVANCE(3);
      if (lookahead == 'L') ADVANCE(4);
      if (lookahead == 'P') ADVANCE(5);
      if (lookahead == 'Q') ADVANCE(6);
      if (lookahead == 'U') ADVANCE(7);
      if (lookahead == '\\') SKIP(8)
      if (lookahead == 'a') ADVANCE(9);
      if (lookahead == 'c') ADVANCE(10);
      if (lookahead == 'd') ADVANCE(11);
      if (lookahead == 'e') ADVANCE(12);
      if (lookahead == 'f') ADVANCE(13);
      if (lookahead == 'i') ADVANCE(14);
      if (lookahead == 'k') ADVANCE(15);
      if (lookahead == 'l') ADVANCE(16);
      if (lookahead == 'r') ADVANCE(17);
      if (lookahead == 's') ADVANCE(18);
      if (lookahead == 't') ADVANCE(19);
      if (lookahead == 'u') ADVANCE(20);
      if (lookahead == 'w') ADVANCE(21);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(22)
      END_STATE();
    case 1:
      ACCEPT_TOKEN(anon_sym_A);
      END_STATE();
    case 2:
      ACCEPT_TOKEN(anon_sym_E);
      END_STATE();
    case 3:
      ACCEPT_TOKEN(anon_sym_K);
      END_STATE();
    case 4:
      ACCEPT_TOKEN(anon_sym_L);
      END_STATE();
    case 5:
      ACCEPT_TOKEN(anon_sym_P);
      END_STATE();
    case 6:
      ACCEPT_TOKEN(anon_sym_Q);
      END_STATE();
    case 7:
      ACCEPT_TOKEN(anon_sym_U);
      END_STATE();
    case 8:
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') SKIP(22)
      if (lookahead == '\r') SKIP(23)
      END_STATE();
    case 9:
      ACCEPT_TOKEN(anon_sym_a);
      END_STATE();
    case 10:
      if (lookahead == 'a') ADVANCE(24);
      END_STATE();
    case 11:
      if (lookahead == 'e') ADVANCE(25);
      if (lookahead == 'o') ADVANCE(26);
      END_STATE();
    case 12:
      if (lookahead == 'l') ADVANCE(27);
      if (lookahead == 'x') ADVANCE(28);
      END_STATE();
    case 13:
      if (lookahead == 'i') ADVANCE(29);
      if (lookahead == 'o') ADVANCE(30);
      if (lookahead == 'u') ADVANCE(31);
      END_STATE();
    case 14:
      if (lookahead == 'f') ADVANCE(32);
      END_STATE();
    case 15:
      ACCEPT_TOKEN(anon_sym_k);
      END_STATE();
    case 16:
      if (lookahead == 'o') ADVANCE(33);
      END_STATE();
    case 17:
      if (lookahead == 'e') ADVANCE(34);
      END_STATE();
    case 18:
      if (lookahead == 'e') ADVANCE(35);
      END_STATE();
    case 19:
      if (lookahead == 'h') ADVANCE(36);
      if (lookahead == 'y') ADVANCE(37);
      END_STATE();
    case 20:
      ACCEPT_TOKEN(anon_sym_u);
      if (lookahead == 'n') ADVANCE(38);
      END_STATE();
    case 21:
      if (lookahead == 'h') ADVANCE(39);
      END_STATE();
    case 22:
      if (lookahead == '\\') SKIP(8)
      if (lookahead == 'c') ADVANCE(10);
      if (lookahead == 'd') ADVANCE(11);
      if (lookahead == 'e') ADVANCE(12);
      if (lookahead == 'f') ADVANCE(13);
      if (lookahead == 'i') ADVANCE(14);
      if (lookahead == 'l') ADVANCE(16);
      if (lookahead == 'r') ADVANCE(17);
      if (lookahead == 's') ADVANCE(18);
      if (lookahead == 't') ADVANCE(19);
      if (lookahead == 'u') ADVANCE(40);
      if (lookahead == 'w') ADVANCE(21);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(22)
      END_STATE();
    case 23:
      if (lookahead == '\n') SKIP(22)
      END_STATE();
    case 24:
      if (lookahead == 's') ADVANCE(41);
      END_STATE();
    case 25:
      if (lookahead == 'c') ADVANCE(42);
      END_STATE();
    case 26:
      ACCEPT_TOKEN(anon_sym_do);
      if (lookahead == 'n') ADVANCE(43);
      END_STATE();
    case 27:
      if (lookahead == 'i') ADVANCE(44);
      if (lookahead == 's') ADVANCE(45);
      END_STATE();
    case 28:
      if (lookahead == 'p') ADVANCE(46);
      END_STATE();
    case 29:
      ACCEPT_TOKEN(anon_sym_fi);
      END_STATE();
    case 30:
      if (lookahead == 'r') ADVANCE(47);
      END_STATE();
    case 31:
      if (lookahead == 'n') ADVANCE(48);
      END_STATE();
    case 32:
      ACCEPT_TOKEN(anon_sym_if);
      END_STATE();
    case 33:
      if (lookahead == 'c') ADVANCE(49);
      END_STATE();
    case 34:
      if (lookahead == 'a') ADVANCE(50);
      END_STATE();
    case 35:
      if (lookahead == 'l') ADVANCE(51);
      END_STATE();
    case 36:
      if (lookahead == 'e') ADVANCE(52);
      END_STATE();
    case 37:
      if (lookahead == 'p') ADVANCE(53);
      END_STATE();
    case 38:
      if (lookahead == 's') ADVANCE(54);
      if (lookahead == 't') ADVANCE(55);
      END_STATE();
    case 39:
      if (lookahead == 'i') ADVANCE(56);
      END_STATE();
    case 40:
      if (lookahead == 'n') ADVANCE(38);
      END_STATE();
    case 41:
      if (lookahead == 'e') ADVANCE(57);
      END_STATE();
    case 42:
      if (lookahead == 'l') ADVANCE(58);
      END_STATE();
    case 43:
      if (lookahead == 'e') ADVANCE(59);
      END_STATE();
    case 44:
      if (lookahead == 'f') ADVANCE(60);
      END_STATE();
    case 45:
      if (lookahead == 'e') ADVANCE(61);
      END_STATE();
    case 46:
      if (lookahead == 'o') ADVANCE(62);
      END_STATE();
    case 47:
      ACCEPT_TOKEN(anon_sym_for);
      END_STATE();
    case 48:
      if (lookahead == 'c') ADVANCE(63);
      END_STATE();
    case 49:
      if (lookahead == 'a') ADVANCE(64);
      END_STATE();
    case 50:
      if (lookahead == 'd') ADVANCE(65);
      END_STATE();
    case 51:
      if (lookahead == 'e') ADVANCE(66);
      END_STATE();
    case 52:
      if (lookahead == 'n') ADVANCE(67);
      END_STATE();
    case 53:
      if (lookahead == 'e') ADVANCE(68);
      END_STATE();
    case 54:
      if (lookahead == 'e') ADVANCE(69);
      END_STATE();
    case 55:
      if (lookahead == 'i') ADVANCE(70);
      END_STATE();
    case 56:
      if (lookahead == 'l') ADVANCE(71);
      END_STATE();
    case 57:
      ACCEPT_TOKEN(anon_sym_case);
      END_STATE();
    case 58:
      if (lookahead == 'a') ADVANCE(72);
      END_STATE();
    case 59:
      ACCEPT_TOKEN(anon_sym_done);
      END_STATE();
    case 60:
      ACCEPT_TOKEN(anon_sym_elif);
      END_STATE();
    case 61:
      ACCEPT_TOKEN(anon_sym_else);
      END_STATE();
    case 62:
      if (lookahead == 'r') ADVANCE(73);
      END_STATE();
    case 63:
      if (lookahead == 't') ADVANCE(74);
      END_STATE();
    case 64:
      if (lookahead == 'l') ADVANCE(75);
      END_STATE();
    case 65:
      if (lookahead == 'o') ADVANCE(76);
      END_STATE();
    case 66:
      if (lookahead == 'c') ADVANCE(77);
      END_STATE();
    case 67:
      ACCEPT_TOKEN(anon_sym_then);
      END_STATE();
    case 68:
      if (lookahead == 's') ADVANCE(78);
      END_STATE();
    case 69:
      if (lookahead == 't') ADVANCE(79);
      END_STATE();
    case 70:
      if (lookahead == 'l') ADVANCE(80);
      END_STATE();
    case 71:
      if (lookahead == 'e') ADVANCE(81);
      END_STATE();
    case 72:
      if (lookahead == 'r') ADVANCE(82);
      END_STATE();
    case 73:
      if (lookahead == 't') ADVANCE(83);
      END_STATE();
    case 74:
      if (lookahead == 'i') ADVANCE(84);
      END_STATE();
    case 75:
      ACCEPT_TOKEN(anon_sym_local);
      END_STATE();
    case 76:
      if (lookahead == 'n') ADVANCE(85);
      END_STATE();
    case 77:
      if (lookahead == 't') ADVANCE(86);
      END_STATE();
    case 78:
      if (lookahead == 'e') ADVANCE(87);
      END_STATE();
    case 79:
      ACCEPT_TOKEN(anon_sym_unset);
      if (lookahead == 'e') ADVANCE(88);
      END_STATE();
    case 80:
      ACCEPT_TOKEN(anon_sym_until);
      END_STATE();
    case 81:
      ACCEPT_TOKEN(anon_sym_while);
      END_STATE();
    case 82:
      if (lookahead == 'e') ADVANCE(89);
      END_STATE();
    case 83:
      ACCEPT_TOKEN(anon_sym_export);
      END_STATE();
    case 84:
      if (lookahead == 'o') ADVANCE(90);
      END_STATE();
    case 85:
      if (lookahead == 'l') ADVANCE(91);
      END_STATE();
    case 86:
      ACCEPT_TOKEN(anon_sym_select);
      END_STATE();
    case 87:
      if (lookahead == 't') ADVANCE(92);
      END_STATE();
    case 88:
      if (lookahead == 'n') ADVANCE(93);
      END_STATE();
    case 89:
      ACCEPT_TOKEN(anon_sym_declare);
      END_STATE();
    case 90:
      if (lookahead == 'n') ADVANCE(94);
      END_STATE();
    case 91:
      if (lookahead == 'y') ADVANCE(95);
      END_STATE();
    case 92:
      ACCEPT_TOKEN(anon_sym_typeset);
      END_STATE();
    case 93:
      if (lookahead == 'v') ADVANCE(96);
      END_STATE();
    case 94:
      ACCEPT_TOKEN(anon_sym_function);
      END_STATE();
    case 95:
      ACCEPT_TOKEN(anon_sym_readonly);
      END_STATE();
    case 96:
      ACCEPT_TOKEN(anon_sym_unsetenv);
      END_STATE();
    default:
      return false;
  }
}

static const TSLexMode ts_lex_modes[STATE_COUNT] = {
  [0] = {.lex_state = 0, .external_lex_state = 1},
  [1] = {.lex_state = 542, .external_lex_state = 2},
  [2] = {.lex_state = 256, .external_lex_state = 2},
  [3] = {.lex_state = 256, .external_lex_state = 2},
  [4] = {.lex_state = 257, .external_lex_state = 3},
  [5] = {.lex_state = 257, .external_lex_state = 3},
  [6] = {.lex_state = 257, .external_lex_state = 3},
  [7] = {.lex_state = 257, .external_lex_state = 3},
  [8] = {.lex_state = 257, .external_lex_state = 3},
  [9] = {.lex_state = 257, .external_lex_state = 3},
  [10] = {.lex_state = 257, .external_lex_state = 3},
  [11] = {.lex_state = 257, .external_lex_state = 3},
  [12] = {.lex_state = 6, .external_lex_state = 4},
  [13] = {.lex_state = 542, .external_lex_state = 2},
  [14] = {.lex_state = 542, .external_lex_state = 2},
  [15] = {.lex_state = 542, .external_lex_state = 2},
  [16] = {.lex_state = 241, .external_lex_state = 5},
  [17] = {.lex_state = 542, .external_lex_state = 2},
  [18] = {.lex_state = 542, .external_lex_state = 2},
  [19] = {.lex_state = 542, .external_lex_state = 2},
  [20] = {.lex_state = 542, .external_lex_state = 2},
  [21] = {.lex_state = 542, .external_lex_state = 2},
  [22] = {.lex_state = 259, .external_lex_state = 6},
  [23] = {.lex_state = 259, .external_lex_state = 6},
  [24] = {.lex_state = 259, .external_lex_state = 6},
  [25] = {.lex_state = 259, .external_lex_state = 6},
  [26] = {.lex_state = 259, .external_lex_state = 6},
  [27] = {.lex_state = 259, .external_lex_state = 6},
  [28] = {.lex_state = 259, .external_lex_state = 6},
  [29] = {.lex_state = 259, .external_lex_state = 6},
  [30] = {.lex_state = 259, .external_lex_state = 6},
  [31] = {.lex_state = 259, .external_lex_state = 6},
  [32] = {.lex_state = 259, .external_lex_state = 6},
  [33] = {.lex_state = 259, .external_lex_state = 6},
  [34] = {.lex_state = 542, .external_lex_state = 2},
  [35] = {.lex_state = 542, .external_lex_state = 2},
  [36] = {.lex_state = 542, .external_lex_state = 2},
  [37] = {.lex_state = 542, .external_lex_state = 2},
  [38] = {.lex_state = 542, .external_lex_state = 2},
  [39] = {.lex_state = 542, .external_lex_state = 2},
  [40] = {.lex_state = 542, .external_lex_state = 2},
  [41] = {.lex_state = 542, .external_lex_state = 2},
  [42] = {.lex_state = 542, .external_lex_state = 2},
  [43] = {.lex_state = 542, .external_lex_state = 2},
  [44] = {.lex_state = 542, .external_lex_state = 2},
  [45] = {.lex_state = 542, .external_lex_state = 2},
  [46] = {.lex_state = 542, .external_lex_state = 2},
  [47] = {.lex_state = 542, .external_lex_state = 2},
  [48] = {.lex_state = 542, .external_lex_state = 2},
  [49] = {.lex_state = 542, .external_lex_state = 2},
  [50] = {.lex_state = 542, .external_lex_state = 2},
  [51] = {.lex_state = 542, .external_lex_state = 2},
  [52] = {.lex_state = 542, .external_lex_state = 2},
  [53] = {.lex_state = 542, .external_lex_state = 2},
  [54] = {.lex_state = 542, .external_lex_state = 2},
  [55] = {.lex_state = 542, .external_lex_state = 2},
  [56] = {.lex_state = 542, .external_lex_state = 2},
  [57] = {.lex_state = 542, .external_lex_state = 2},
  [58] = {.lex_state = 542, .external_lex_state = 2},
  [59] = {.lex_state = 542, .external_lex_state = 2},
  [60] = {.lex_state = 542, .external_lex_state = 2},
  [61] = {.lex_state = 542, .external_lex_state = 2},
  [62] = {.lex_state = 260, .external_lex_state = 2},
  [63] = {.lex_state = 260, .external_lex_state = 2},
  [64] = {.lex_state = 542, .external_lex_state = 2},
  [65] = {.lex_state = 542, .external_lex_state = 2},
  [66] = {.lex_state = 542, .external_lex_state = 2},
  [67] = {.lex_state = 542, .external_lex_state = 2},
  [68] = {.lex_state = 260, .external_lex_state = 2},
  [69] = {.lex_state = 542, .external_lex_state = 2},
  [70] = {.lex_state = 542, .external_lex_state = 2},
  [71] = {.lex_state = 542, .external_lex_state = 2},
  [72] = {.lex_state = 260, .external_lex_state = 2},
  [73] = {.lex_state = 542, .external_lex_state = 2},
  [74] = {.lex_state = 260, .external_lex_state = 2},
  [75] = {.lex_state = 542, .external_lex_state = 2},
  [76] = {.lex_state = 542, .external_lex_state = 2},
  [77] = {.lex_state = 542, .external_lex_state = 2},
  [78] = {.lex_state = 542, .external_lex_state = 2},
  [79] = {.lex_state = 542, .external_lex_state = 2},
  [80] = {.lex_state = 542, .external_lex_state = 2},
  [81] = {.lex_state = 260, .external_lex_state = 2},
  [82] = {.lex_state = 260, .external_lex_state = 2},
  [83] = {.lex_state = 542, .external_lex_state = 2},
  [84] = {.lex_state = 542, .external_lex_state = 2},
  [85] = {.lex_state = 542, .external_lex_state = 2},
  [86] = {.lex_state = 260, .external_lex_state = 2},
  [87] = {.lex_state = 260, .external_lex_state = 2},
  [88] = {.lex_state = 542, .external_lex_state = 2},
  [89] = {.lex_state = 542, .external_lex_state = 2},
  [90] = {.lex_state = 542, .external_lex_state = 2},
  [91] = {.lex_state = 542, .external_lex_state = 2},
  [92] = {.lex_state = 260, .external_lex_state = 2},
  [93] = {.lex_state = 260, .external_lex_state = 2},
  [94] = {.lex_state = 542, .external_lex_state = 2},
  [95] = {.lex_state = 542, .external_lex_state = 2},
  [96] = {.lex_state = 260, .external_lex_state = 2},
  [97] = {.lex_state = 542, .external_lex_state = 2},
  [98] = {.lex_state = 542, .external_lex_state = 2},
  [99] = {.lex_state = 260, .external_lex_state = 2},
  [100] = {.lex_state = 542, .external_lex_state = 2},
  [101] = {.lex_state = 542, .external_lex_state = 2},
  [102] = {.lex_state = 542, .external_lex_state = 2},
  [103] = {.lex_state = 542, .external_lex_state = 2},
  [104] = {.lex_state = 542, .external_lex_state = 2},
  [105] = {.lex_state = 542, .external_lex_state = 2},
  [106] = {.lex_state = 542, .external_lex_state = 2},
  [107] = {.lex_state = 542, .external_lex_state = 2},
  [108] = {.lex_state = 542, .external_lex_state = 2},
  [109] = {.lex_state = 542, .external_lex_state = 2},
  [110] = {.lex_state = 542, .external_lex_state = 2},
  [111] = {.lex_state = 542, .external_lex_state = 2},
  [112] = {.lex_state = 542, .external_lex_state = 2},
  [113] = {.lex_state = 542, .external_lex_state = 2},
  [114] = {.lex_state = 542, .external_lex_state = 2},
  [115] = {.lex_state = 542, .external_lex_state = 2},
  [116] = {.lex_state = 542, .external_lex_state = 2},
  [117] = {.lex_state = 542, .external_lex_state = 2},
  [118] = {.lex_state = 542, .external_lex_state = 2},
  [119] = {.lex_state = 542, .external_lex_state = 2},
  [120] = {.lex_state = 542, .external_lex_state = 2},
  [121] = {.lex_state = 542, .external_lex_state = 2},
  [122] = {.lex_state = 542, .external_lex_state = 2},
  [123] = {.lex_state = 542, .external_lex_state = 2},
  [124] = {.lex_state = 542, .external_lex_state = 2},
  [125] = {.lex_state = 542, .external_lex_state = 2},
  [126] = {.lex_state = 542, .external_lex_state = 2},
  [127] = {.lex_state = 542, .external_lex_state = 2},
  [128] = {.lex_state = 542, .external_lex_state = 2},
  [129] = {.lex_state = 542, .external_lex_state = 2},
  [130] = {.lex_state = 542, .external_lex_state = 2},
  [131] = {.lex_state = 542, .external_lex_state = 2},
  [132] = {.lex_state = 542, .external_lex_state = 2},
  [133] = {.lex_state = 542, .external_lex_state = 2},
  [134] = {.lex_state = 542, .external_lex_state = 2},
  [135] = {.lex_state = 542, .external_lex_state = 2},
  [136] = {.lex_state = 542, .external_lex_state = 2},
  [137] = {.lex_state = 542, .external_lex_state = 2},
  [138] = {.lex_state = 542, .external_lex_state = 2},
  [139] = {.lex_state = 542, .external_lex_state = 2},
  [140] = {.lex_state = 542, .external_lex_state = 2},
  [141] = {.lex_state = 542, .external_lex_state = 2},
  [142] = {.lex_state = 542, .external_lex_state = 2},
  [143] = {.lex_state = 542, .external_lex_state = 2},
  [144] = {.lex_state = 542, .external_lex_state = 2},
  [145] = {.lex_state = 542, .external_lex_state = 2},
  [146] = {.lex_state = 542, .external_lex_state = 2},
  [147] = {.lex_state = 542, .external_lex_state = 2},
  [148] = {.lex_state = 542, .external_lex_state = 2},
  [149] = {.lex_state = 542, .external_lex_state = 2},
  [150] = {.lex_state = 542, .external_lex_state = 2},
  [151] = {.lex_state = 542, .external_lex_state = 2},
  [152] = {.lex_state = 542, .external_lex_state = 2},
  [153] = {.lex_state = 542, .external_lex_state = 2},
  [154] = {.lex_state = 260, .external_lex_state = 2},
  [155] = {.lex_state = 542, .external_lex_state = 2},
  [156] = {.lex_state = 542, .external_lex_state = 2},
  [157] = {.lex_state = 542, .external_lex_state = 2},
  [158] = {.lex_state = 542, .external_lex_state = 2},
  [159] = {.lex_state = 542, .external_lex_state = 2},
  [160] = {.lex_state = 542, .external_lex_state = 2},
  [161] = {.lex_state = 542, .external_lex_state = 2},
  [162] = {.lex_state = 542, .external_lex_state = 2},
  [163] = {.lex_state = 542, .external_lex_state = 2},
  [164] = {.lex_state = 542, .external_lex_state = 2},
  [165] = {.lex_state = 542, .external_lex_state = 2},
  [166] = {.lex_state = 542, .external_lex_state = 2},
  [167] = {.lex_state = 542, .external_lex_state = 2},
  [168] = {.lex_state = 542, .external_lex_state = 2},
  [169] = {.lex_state = 542, .external_lex_state = 2},
  [170] = {.lex_state = 542, .external_lex_state = 2},
  [171] = {.lex_state = 542, .external_lex_state = 2},
  [172] = {.lex_state = 542, .external_lex_state = 2},
  [173] = {.lex_state = 542, .external_lex_state = 2},
  [174] = {.lex_state = 542, .external_lex_state = 2},
  [175] = {.lex_state = 542, .external_lex_state = 2},
  [176] = {.lex_state = 542, .external_lex_state = 2},
  [177] = {.lex_state = 542, .external_lex_state = 2},
  [178] = {.lex_state = 542, .external_lex_state = 2},
  [179] = {.lex_state = 542, .external_lex_state = 2},
  [180] = {.lex_state = 542, .external_lex_state = 2},
  [181] = {.lex_state = 542, .external_lex_state = 2},
  [182] = {.lex_state = 542, .external_lex_state = 2},
  [183] = {.lex_state = 542, .external_lex_state = 2},
  [184] = {.lex_state = 542, .external_lex_state = 2},
  [185] = {.lex_state = 542, .external_lex_state = 2},
  [186] = {.lex_state = 542, .external_lex_state = 2},
  [187] = {.lex_state = 542, .external_lex_state = 2},
  [188] = {.lex_state = 542, .external_lex_state = 2},
  [189] = {.lex_state = 542, .external_lex_state = 2},
  [190] = {.lex_state = 542, .external_lex_state = 2},
  [191] = {.lex_state = 542, .external_lex_state = 2},
  [192] = {.lex_state = 542, .external_lex_state = 2},
  [193] = {.lex_state = 542, .external_lex_state = 2},
  [194] = {.lex_state = 542, .external_lex_state = 2},
  [195] = {.lex_state = 542, .external_lex_state = 2},
  [196] = {.lex_state = 542, .external_lex_state = 2},
  [197] = {.lex_state = 542, .external_lex_state = 2},
  [198] = {.lex_state = 542, .external_lex_state = 2},
  [199] = {.lex_state = 542, .external_lex_state = 2},
  [200] = {.lex_state = 542, .external_lex_state = 2},
  [201] = {.lex_state = 542, .external_lex_state = 2},
  [202] = {.lex_state = 542, .external_lex_state = 2},
  [203] = {.lex_state = 542, .external_lex_state = 2},
  [204] = {.lex_state = 542, .external_lex_state = 2},
  [205] = {.lex_state = 542, .external_lex_state = 2},
  [206] = {.lex_state = 542, .external_lex_state = 2},
  [207] = {.lex_state = 542, .external_lex_state = 2},
  [208] = {.lex_state = 542, .external_lex_state = 2},
  [209] = {.lex_state = 542, .external_lex_state = 2},
  [210] = {.lex_state = 542, .external_lex_state = 2},
  [211] = {.lex_state = 542, .external_lex_state = 2},
  [212] = {.lex_state = 542, .external_lex_state = 2},
  [213] = {.lex_state = 542, .external_lex_state = 2},
  [214] = {.lex_state = 542, .external_lex_state = 2},
  [215] = {.lex_state = 542, .external_lex_state = 2},
  [216] = {.lex_state = 542, .external_lex_state = 2},
  [217] = {.lex_state = 542, .external_lex_state = 2},
  [218] = {.lex_state = 542, .external_lex_state = 2},
  [219] = {.lex_state = 542, .external_lex_state = 2},
  [220] = {.lex_state = 542, .external_lex_state = 2},
  [221] = {.lex_state = 542, .external_lex_state = 2},
  [222] = {.lex_state = 542, .external_lex_state = 2},
  [223] = {.lex_state = 542, .external_lex_state = 2},
  [224] = {.lex_state = 542, .external_lex_state = 2},
  [225] = {.lex_state = 542, .external_lex_state = 2},
  [226] = {.lex_state = 542, .external_lex_state = 2},
  [227] = {.lex_state = 542, .external_lex_state = 2},
  [228] = {.lex_state = 542, .external_lex_state = 2},
  [229] = {.lex_state = 542, .external_lex_state = 2},
  [230] = {.lex_state = 542, .external_lex_state = 2},
  [231] = {.lex_state = 542, .external_lex_state = 2},
  [232] = {.lex_state = 542, .external_lex_state = 2},
  [233] = {.lex_state = 542, .external_lex_state = 2},
  [234] = {.lex_state = 542, .external_lex_state = 2},
  [235] = {.lex_state = 542, .external_lex_state = 2},
  [236] = {.lex_state = 542, .external_lex_state = 2},
  [237] = {.lex_state = 542, .external_lex_state = 2},
  [238] = {.lex_state = 542, .external_lex_state = 2},
  [239] = {.lex_state = 542, .external_lex_state = 2},
  [240] = {.lex_state = 542, .external_lex_state = 2},
  [241] = {.lex_state = 542, .external_lex_state = 2},
  [242] = {.lex_state = 542, .external_lex_state = 2},
  [243] = {.lex_state = 542, .external_lex_state = 2},
  [244] = {.lex_state = 542, .external_lex_state = 2},
  [245] = {.lex_state = 542, .external_lex_state = 2},
  [246] = {.lex_state = 542, .external_lex_state = 2},
  [247] = {.lex_state = 542, .external_lex_state = 2},
  [248] = {.lex_state = 542, .external_lex_state = 2},
  [249] = {.lex_state = 542, .external_lex_state = 2},
  [250] = {.lex_state = 542, .external_lex_state = 2},
  [251] = {.lex_state = 542, .external_lex_state = 2},
  [252] = {.lex_state = 542, .external_lex_state = 2},
  [253] = {.lex_state = 542, .external_lex_state = 2},
  [254] = {.lex_state = 542, .external_lex_state = 2},
  [255] = {.lex_state = 542, .external_lex_state = 2},
  [256] = {.lex_state = 542, .external_lex_state = 2},
  [257] = {.lex_state = 542, .external_lex_state = 2},
  [258] = {.lex_state = 542, .external_lex_state = 2},
  [259] = {.lex_state = 542, .external_lex_state = 2},
  [260] = {.lex_state = 542, .external_lex_state = 2},
  [261] = {.lex_state = 542, .external_lex_state = 2},
  [262] = {.lex_state = 542, .external_lex_state = 2},
  [263] = {.lex_state = 542, .external_lex_state = 2},
  [264] = {.lex_state = 542, .external_lex_state = 2},
  [265] = {.lex_state = 542, .external_lex_state = 2},
  [266] = {.lex_state = 542, .external_lex_state = 2},
  [267] = {.lex_state = 542, .external_lex_state = 2},
  [268] = {.lex_state = 542, .external_lex_state = 2},
  [269] = {.lex_state = 542, .external_lex_state = 2},
  [270] = {.lex_state = 542, .external_lex_state = 2},
  [271] = {.lex_state = 542, .external_lex_state = 2},
  [272] = {.lex_state = 542, .external_lex_state = 2},
  [273] = {.lex_state = 542, .external_lex_state = 2},
  [274] = {.lex_state = 542, .external_lex_state = 2},
  [275] = {.lex_state = 542, .external_lex_state = 2},
  [276] = {.lex_state = 542, .external_lex_state = 2},
  [277] = {.lex_state = 542, .external_lex_state = 2},
  [278] = {.lex_state = 542, .external_lex_state = 2},
  [279] = {.lex_state = 542, .external_lex_state = 2},
  [280] = {.lex_state = 542, .external_lex_state = 2},
  [281] = {.lex_state = 542, .external_lex_state = 2},
  [282] = {.lex_state = 260, .external_lex_state = 2},
  [283] = {.lex_state = 542, .external_lex_state = 2},
  [284] = {.lex_state = 260, .external_lex_state = 2},
  [285] = {.lex_state = 542, .external_lex_state = 2},
  [286] = {.lex_state = 542, .external_lex_state = 2},
  [287] = {.lex_state = 542, .external_lex_state = 2},
  [288] = {.lex_state = 542, .external_lex_state = 2},
  [289] = {.lex_state = 542, .external_lex_state = 2},
  [290] = {.lex_state = 260, .external_lex_state = 2},
  [291] = {.lex_state = 542, .external_lex_state = 2},
  [292] = {.lex_state = 542, .external_lex_state = 2},
  [293] = {.lex_state = 542, .external_lex_state = 2},
  [294] = {.lex_state = 542, .external_lex_state = 2},
  [295] = {.lex_state = 542, .external_lex_state = 2},
  [296] = {.lex_state = 542, .external_lex_state = 2},
  [297] = {.lex_state = 542, .external_lex_state = 2},
  [298] = {.lex_state = 542, .external_lex_state = 2},
  [299] = {.lex_state = 542, .external_lex_state = 2},
  [300] = {.lex_state = 542, .external_lex_state = 2},
  [301] = {.lex_state = 542, .external_lex_state = 2},
  [302] = {.lex_state = 542, .external_lex_state = 2},
  [303] = {.lex_state = 542, .external_lex_state = 2},
  [304] = {.lex_state = 542, .external_lex_state = 2},
  [305] = {.lex_state = 542, .external_lex_state = 2},
  [306] = {.lex_state = 542, .external_lex_state = 2},
  [307] = {.lex_state = 542, .external_lex_state = 2},
  [308] = {.lex_state = 542, .external_lex_state = 2},
  [309] = {.lex_state = 542, .external_lex_state = 2},
  [310] = {.lex_state = 542, .external_lex_state = 2},
  [311] = {.lex_state = 542, .external_lex_state = 2},
  [312] = {.lex_state = 542, .external_lex_state = 2},
  [313] = {.lex_state = 542, .external_lex_state = 2},
  [314] = {.lex_state = 542, .external_lex_state = 2},
  [315] = {.lex_state = 542, .external_lex_state = 2},
  [316] = {.lex_state = 542, .external_lex_state = 2},
  [317] = {.lex_state = 542, .external_lex_state = 2},
  [318] = {.lex_state = 542, .external_lex_state = 2},
  [319] = {.lex_state = 542, .external_lex_state = 2},
  [320] = {.lex_state = 542, .external_lex_state = 2},
  [321] = {.lex_state = 542, .external_lex_state = 2},
  [322] = {.lex_state = 542, .external_lex_state = 2},
  [323] = {.lex_state = 542, .external_lex_state = 2},
  [324] = {.lex_state = 542, .external_lex_state = 2},
  [325] = {.lex_state = 542, .external_lex_state = 2},
  [326] = {.lex_state = 542, .external_lex_state = 2},
  [327] = {.lex_state = 542, .external_lex_state = 2},
  [328] = {.lex_state = 542, .external_lex_state = 2},
  [329] = {.lex_state = 542, .external_lex_state = 2},
  [330] = {.lex_state = 542, .external_lex_state = 2},
  [331] = {.lex_state = 542, .external_lex_state = 2},
  [332] = {.lex_state = 542, .external_lex_state = 2},
  [333] = {.lex_state = 542, .external_lex_state = 2},
  [334] = {.lex_state = 542, .external_lex_state = 2},
  [335] = {.lex_state = 542, .external_lex_state = 2},
  [336] = {.lex_state = 542, .external_lex_state = 2},
  [337] = {.lex_state = 542, .external_lex_state = 2},
  [338] = {.lex_state = 542, .external_lex_state = 2},
  [339] = {.lex_state = 542, .external_lex_state = 2},
  [340] = {.lex_state = 542, .external_lex_state = 2},
  [341] = {.lex_state = 542, .external_lex_state = 2},
  [342] = {.lex_state = 542, .external_lex_state = 2},
  [343] = {.lex_state = 542, .external_lex_state = 2},
  [344] = {.lex_state = 542, .external_lex_state = 2},
  [345] = {.lex_state = 542, .external_lex_state = 2},
  [346] = {.lex_state = 542, .external_lex_state = 2},
  [347] = {.lex_state = 542, .external_lex_state = 2},
  [348] = {.lex_state = 542, .external_lex_state = 2},
  [349] = {.lex_state = 542, .external_lex_state = 2},
  [350] = {.lex_state = 542, .external_lex_state = 2},
  [351] = {.lex_state = 542, .external_lex_state = 2},
  [352] = {.lex_state = 542, .external_lex_state = 2},
  [353] = {.lex_state = 542, .external_lex_state = 2},
  [354] = {.lex_state = 542, .external_lex_state = 2},
  [355] = {.lex_state = 542, .external_lex_state = 2},
  [356] = {.lex_state = 542, .external_lex_state = 2},
  [357] = {.lex_state = 542, .external_lex_state = 2},
  [358] = {.lex_state = 542, .external_lex_state = 2},
  [359] = {.lex_state = 542, .external_lex_state = 2},
  [360] = {.lex_state = 542, .external_lex_state = 2},
  [361] = {.lex_state = 542, .external_lex_state = 2},
  [362] = {.lex_state = 542, .external_lex_state = 2},
  [363] = {.lex_state = 542, .external_lex_state = 2},
  [364] = {.lex_state = 542, .external_lex_state = 2},
  [365] = {.lex_state = 542, .external_lex_state = 2},
  [366] = {.lex_state = 542, .external_lex_state = 2},
  [367] = {.lex_state = 542, .external_lex_state = 2},
  [368] = {.lex_state = 542, .external_lex_state = 2},
  [369] = {.lex_state = 542, .external_lex_state = 2},
  [370] = {.lex_state = 542, .external_lex_state = 2},
  [371] = {.lex_state = 542, .external_lex_state = 2},
  [372] = {.lex_state = 542, .external_lex_state = 2},
  [373] = {.lex_state = 542, .external_lex_state = 2},
  [374] = {.lex_state = 542, .external_lex_state = 2},
  [375] = {.lex_state = 542, .external_lex_state = 2},
  [376] = {.lex_state = 542, .external_lex_state = 2},
  [377] = {.lex_state = 542, .external_lex_state = 2},
  [378] = {.lex_state = 542, .external_lex_state = 2},
  [379] = {.lex_state = 542, .external_lex_state = 2},
  [380] = {.lex_state = 542, .external_lex_state = 2},
  [381] = {.lex_state = 542, .external_lex_state = 2},
  [382] = {.lex_state = 542, .external_lex_state = 2},
  [383] = {.lex_state = 542, .external_lex_state = 2},
  [384] = {.lex_state = 542, .external_lex_state = 2},
  [385] = {.lex_state = 542, .external_lex_state = 2},
  [386] = {.lex_state = 542, .external_lex_state = 2},
  [387] = {.lex_state = 542, .external_lex_state = 2},
  [388] = {.lex_state = 542, .external_lex_state = 2},
  [389] = {.lex_state = 542, .external_lex_state = 2},
  [390] = {.lex_state = 542, .external_lex_state = 2},
  [391] = {.lex_state = 542, .external_lex_state = 2},
  [392] = {.lex_state = 542, .external_lex_state = 2},
  [393] = {.lex_state = 542, .external_lex_state = 2},
  [394] = {.lex_state = 542, .external_lex_state = 2},
  [395] = {.lex_state = 542, .external_lex_state = 2},
  [396] = {.lex_state = 542, .external_lex_state = 2},
  [397] = {.lex_state = 542, .external_lex_state = 2},
  [398] = {.lex_state = 542, .external_lex_state = 2},
  [399] = {.lex_state = 542, .external_lex_state = 2},
  [400] = {.lex_state = 542, .external_lex_state = 2},
  [401] = {.lex_state = 542, .external_lex_state = 2},
  [402] = {.lex_state = 542, .external_lex_state = 2},
  [403] = {.lex_state = 542, .external_lex_state = 2},
  [404] = {.lex_state = 542, .external_lex_state = 2},
  [405] = {.lex_state = 542, .external_lex_state = 2},
  [406] = {.lex_state = 542, .external_lex_state = 2},
  [407] = {.lex_state = 542, .external_lex_state = 2},
  [408] = {.lex_state = 542, .external_lex_state = 2},
  [409] = {.lex_state = 542, .external_lex_state = 2},
  [410] = {.lex_state = 542, .external_lex_state = 2},
  [411] = {.lex_state = 542, .external_lex_state = 2},
  [412] = {.lex_state = 542, .external_lex_state = 2},
  [413] = {.lex_state = 542, .external_lex_state = 2},
  [414] = {.lex_state = 542, .external_lex_state = 2},
  [415] = {.lex_state = 542, .external_lex_state = 2},
  [416] = {.lex_state = 542, .external_lex_state = 2},
  [417] = {.lex_state = 542, .external_lex_state = 2},
  [418] = {.lex_state = 542, .external_lex_state = 2},
  [419] = {.lex_state = 542, .external_lex_state = 2},
  [420] = {.lex_state = 542, .external_lex_state = 2},
  [421] = {.lex_state = 542, .external_lex_state = 2},
  [422] = {.lex_state = 542, .external_lex_state = 2},
  [423] = {.lex_state = 542, .external_lex_state = 2},
  [424] = {.lex_state = 542, .external_lex_state = 2},
  [425] = {.lex_state = 542, .external_lex_state = 2},
  [426] = {.lex_state = 542, .external_lex_state = 2},
  [427] = {.lex_state = 542, .external_lex_state = 2},
  [428] = {.lex_state = 542, .external_lex_state = 2},
  [429] = {.lex_state = 542, .external_lex_state = 2},
  [430] = {.lex_state = 542, .external_lex_state = 2},
  [431] = {.lex_state = 542, .external_lex_state = 2},
  [432] = {.lex_state = 542, .external_lex_state = 2},
  [433] = {.lex_state = 542, .external_lex_state = 2},
  [434] = {.lex_state = 542, .external_lex_state = 2},
  [435] = {.lex_state = 542, .external_lex_state = 2},
  [436] = {.lex_state = 542, .external_lex_state = 2},
  [437] = {.lex_state = 542, .external_lex_state = 2},
  [438] = {.lex_state = 542, .external_lex_state = 2},
  [439] = {.lex_state = 542, .external_lex_state = 2},
  [440] = {.lex_state = 542, .external_lex_state = 2},
  [441] = {.lex_state = 542, .external_lex_state = 2},
  [442] = {.lex_state = 542, .external_lex_state = 2},
  [443] = {.lex_state = 542, .external_lex_state = 2},
  [444] = {.lex_state = 542, .external_lex_state = 2},
  [445] = {.lex_state = 542, .external_lex_state = 2},
  [446] = {.lex_state = 542, .external_lex_state = 2},
  [447] = {.lex_state = 542, .external_lex_state = 2},
  [448] = {.lex_state = 542, .external_lex_state = 2},
  [449] = {.lex_state = 542, .external_lex_state = 2},
  [450] = {.lex_state = 542, .external_lex_state = 2},
  [451] = {.lex_state = 242, .external_lex_state = 7},
  [452] = {.lex_state = 245, .external_lex_state = 8},
  [453] = {.lex_state = 245, .external_lex_state = 8},
  [454] = {.lex_state = 245, .external_lex_state = 8},
  [455] = {.lex_state = 245, .external_lex_state = 8},
  [456] = {.lex_state = 245, .external_lex_state = 8},
  [457] = {.lex_state = 245, .external_lex_state = 8},
  [458] = {.lex_state = 244, .external_lex_state = 7},
  [459] = {.lex_state = 243, .external_lex_state = 7},
  [460] = {.lex_state = 246, .external_lex_state = 7},
  [461] = {.lex_state = 245, .external_lex_state = 8},
  [462] = {.lex_state = 245, .external_lex_state = 8},
  [463] = {.lex_state = 242, .external_lex_state = 7},
  [464] = {.lex_state = 12, .external_lex_state = 9},
  [465] = {.lex_state = 12, .external_lex_state = 9},
  [466] = {.lex_state = 240, .external_lex_state = 10},
  [467] = {.lex_state = 240, .external_lex_state = 10},
  [468] = {.lex_state = 112, .external_lex_state = 4},
  [469] = {.lex_state = 112, .external_lex_state = 4},
  [470] = {.lex_state = 112, .external_lex_state = 4},
  [471] = {.lex_state = 112, .external_lex_state = 4},
  [472] = {.lex_state = 112, .external_lex_state = 4},
  [473] = {.lex_state = 112, .external_lex_state = 4},
  [474] = {.lex_state = 112, .external_lex_state = 4},
  [475] = {.lex_state = 112, .external_lex_state = 4},
  [476] = {.lex_state = 112, .external_lex_state = 4},
  [477] = {.lex_state = 112, .external_lex_state = 4},
  [478] = {.lex_state = 112, .external_lex_state = 4},
  [479] = {.lex_state = 112, .external_lex_state = 4},
  [480] = {.lex_state = 112, .external_lex_state = 4},
  [481] = {.lex_state = 112, .external_lex_state = 4},
  [482] = {.lex_state = 112, .external_lex_state = 4},
  [483] = {.lex_state = 112, .external_lex_state = 4},
  [484] = {.lex_state = 112, .external_lex_state = 4},
  [485] = {.lex_state = 112, .external_lex_state = 4},
  [486] = {.lex_state = 112, .external_lex_state = 4},
  [487] = {.lex_state = 112, .external_lex_state = 4},
  [488] = {.lex_state = 112, .external_lex_state = 4},
  [489] = {.lex_state = 112, .external_lex_state = 4},
  [490] = {.lex_state = 112, .external_lex_state = 4},
  [491] = {.lex_state = 112, .external_lex_state = 4},
  [492] = {.lex_state = 112, .external_lex_state = 4},
  [493] = {.lex_state = 174, .external_lex_state = 11},
  [494] = {.lex_state = 174, .external_lex_state = 11},
  [495] = {.lex_state = 174, .external_lex_state = 11},
  [496] = {.lex_state = 263, .external_lex_state = 5},
  [497] = {.lex_state = 263, .external_lex_state = 5},
  [498] = {.lex_state = 263, .external_lex_state = 5},
  [499] = {.lex_state = 263, .external_lex_state = 5},
  [500] = {.lex_state = 263, .external_lex_state = 5},
  [501] = {.lex_state = 263, .external_lex_state = 5},
  [502] = {.lex_state = 174, .external_lex_state = 11},
  [503] = {.lex_state = 263, .external_lex_state = 5},
  [504] = {.lex_state = 263, .external_lex_state = 5},
  [505] = {.lex_state = 263, .external_lex_state = 5},
  [506] = {.lex_state = 263, .external_lex_state = 5},
  [507] = {.lex_state = 263, .external_lex_state = 5},
  [508] = {.lex_state = 263, .external_lex_state = 5},
  [509] = {.lex_state = 263, .external_lex_state = 5},
  [510] = {.lex_state = 263, .external_lex_state = 5},
  [511] = {.lex_state = 263, .external_lex_state = 5},
  [512] = {.lex_state = 263, .external_lex_state = 5},
  [513] = {.lex_state = 263, .external_lex_state = 5},
  [514] = {.lex_state = 263, .external_lex_state = 5},
  [515] = {.lex_state = 263, .external_lex_state = 5},
  [516] = {.lex_state = 263, .external_lex_state = 5},
  [517] = {.lex_state = 263, .external_lex_state = 5},
  [518] = {.lex_state = 263, .external_lex_state = 5},
  [519] = {.lex_state = 263, .external_lex_state = 5},
  [520] = {.lex_state = 263, .external_lex_state = 5},
  [521] = {.lex_state = 263, .external_lex_state = 5},
  [522] = {.lex_state = 264, .external_lex_state = 12},
  [523] = {.lex_state = 264, .external_lex_state = 12},
  [524] = {.lex_state = 264, .external_lex_state = 12},
  [525] = {.lex_state = 264, .external_lex_state = 12},
  [526] = {.lex_state = 258, .external_lex_state = 2},
  [527] = {.lex_state = 258, .external_lex_state = 2},
  [528] = {.lex_state = 177, .external_lex_state = 13},
  [529] = {.lex_state = 177, .external_lex_state = 13},
  [530] = {.lex_state = 180, .external_lex_state = 11},
  [531] = {.lex_state = 180, .external_lex_state = 11},
  [532] = {.lex_state = 177, .external_lex_state = 14},
  [533] = {.lex_state = 527, .external_lex_state = 11},
  [534] = {.lex_state = 527, .external_lex_state = 11},
  [535] = {.lex_state = 177, .external_lex_state = 14},
  [536] = {.lex_state = 527, .external_lex_state = 11},
  [537] = {.lex_state = 527, .external_lex_state = 11},
  [538] = {.lex_state = 163, .external_lex_state = 15},
  [539] = {.lex_state = 163, .external_lex_state = 15},
  [540] = {.lex_state = 177, .external_lex_state = 14},
  [541] = {.lex_state = 180, .external_lex_state = 16},
  [542] = {.lex_state = 180, .external_lex_state = 16},
  [543] = {.lex_state = 527, .external_lex_state = 11},
  [544] = {.lex_state = 527, .external_lex_state = 11},
  [545] = {.lex_state = 527, .external_lex_state = 11},
  [546] = {.lex_state = 186, .external_lex_state = 17},
  [547] = {.lex_state = 180, .external_lex_state = 16},
  [548] = {.lex_state = 186, .external_lex_state = 17},
  [549] = {.lex_state = 176, .external_lex_state = 9},
  [550] = {.lex_state = 176, .external_lex_state = 9},
  [551] = {.lex_state = 186, .external_lex_state = 17},
  [552] = {.lex_state = 179, .external_lex_state = 18},
  [553] = {.lex_state = 179, .external_lex_state = 18},
  [554] = {.lex_state = 527, .external_lex_state = 11},
  [555] = {.lex_state = 182, .external_lex_state = 9},
  [556] = {.lex_state = 528, .external_lex_state = 9},
  [557] = {.lex_state = 527, .external_lex_state = 16},
  [558] = {.lex_state = 528, .external_lex_state = 9},
  [559] = {.lex_state = 189, .external_lex_state = 19},
  [560] = {.lex_state = 527, .external_lex_state = 16},
  [561] = {.lex_state = 527, .external_lex_state = 16},
  [562] = {.lex_state = 527, .external_lex_state = 16},
  [563] = {.lex_state = 189, .external_lex_state = 19},
  [564] = {.lex_state = 182, .external_lex_state = 9},
  [565] = {.lex_state = 189, .external_lex_state = 19},
  [566] = {.lex_state = 527, .external_lex_state = 16},
  [567] = {.lex_state = 188, .external_lex_state = 20},
  [568] = {.lex_state = 188, .external_lex_state = 20},
  [569] = {.lex_state = 527, .external_lex_state = 16},
  [570] = {.lex_state = 530, .external_lex_state = 19},
  [571] = {.lex_state = 527, .external_lex_state = 16},
  [572] = {.lex_state = 191, .external_lex_state = 20},
  [573] = {.lex_state = 195, .external_lex_state = 17},
  [574] = {.lex_state = 527, .external_lex_state = 16},
  [575] = {.lex_state = 195, .external_lex_state = 17},
  [576] = {.lex_state = 530, .external_lex_state = 19},
  [577] = {.lex_state = 530, .external_lex_state = 19},
  [578] = {.lex_state = 531, .external_lex_state = 20},
  [579] = {.lex_state = 528, .external_lex_state = 9},
  [580] = {.lex_state = 528, .external_lex_state = 9},
  [581] = {.lex_state = 530, .external_lex_state = 19},
  [582] = {.lex_state = 527, .external_lex_state = 16},
  [583] = {.lex_state = 195, .external_lex_state = 17},
  [584] = {.lex_state = 527, .external_lex_state = 16},
  [585] = {.lex_state = 530, .external_lex_state = 19},
  [586] = {.lex_state = 530, .external_lex_state = 19},
  [587] = {.lex_state = 527, .external_lex_state = 16},
  [588] = {.lex_state = 531, .external_lex_state = 20},
  [589] = {.lex_state = 191, .external_lex_state = 20},
  [590] = {.lex_state = 186, .external_lex_state = 21},
  [591] = {.lex_state = 195, .external_lex_state = 21},
  [592] = {.lex_state = 530, .external_lex_state = 19},
  [593] = {.lex_state = 195, .external_lex_state = 21},
  [594] = {.lex_state = 530, .external_lex_state = 19},
  [595] = {.lex_state = 195, .external_lex_state = 21},
  [596] = {.lex_state = 180, .external_lex_state = 11},
  [597] = {.lex_state = 197, .external_lex_state = 17},
  [598] = {.lex_state = 186, .external_lex_state = 21},
  [599] = {.lex_state = 542, .external_lex_state = 2},
  [600] = {.lex_state = 180, .external_lex_state = 11},
  [601] = {.lex_state = 198, .external_lex_state = 19},
  [602] = {.lex_state = 530, .external_lex_state = 19},
  [603] = {.lex_state = 197, .external_lex_state = 17},
  [604] = {.lex_state = 195, .external_lex_state = 17},
  [605] = {.lex_state = 197, .external_lex_state = 17},
  [606] = {.lex_state = 531, .external_lex_state = 20},
  [607] = {.lex_state = 195, .external_lex_state = 17},
  [608] = {.lex_state = 531, .external_lex_state = 20},
  [609] = {.lex_state = 198, .external_lex_state = 19},
  [610] = {.lex_state = 530, .external_lex_state = 19},
  [611] = {.lex_state = 197, .external_lex_state = 17},
  [612] = {.lex_state = 530, .external_lex_state = 19},
  [613] = {.lex_state = 186, .external_lex_state = 21},
  [614] = {.lex_state = 198, .external_lex_state = 19},
  [615] = {.lex_state = 198, .external_lex_state = 22},
  [616] = {.lex_state = 201, .external_lex_state = 21},
  [617] = {.lex_state = 299, .external_lex_state = 23},
  [618] = {.lex_state = 295, .external_lex_state = 2},
  [619] = {.lex_state = 295, .external_lex_state = 2},
  [620] = {.lex_state = 299, .external_lex_state = 23},
  [621] = {.lex_state = 198, .external_lex_state = 22},
  [622] = {.lex_state = 259, .external_lex_state = 6},
  [623] = {.lex_state = 295, .external_lex_state = 2},
  [624] = {.lex_state = 542, .external_lex_state = 2},
  [625] = {.lex_state = 542, .external_lex_state = 2},
  [626] = {.lex_state = 295, .external_lex_state = 2},
  [627] = {.lex_state = 533, .external_lex_state = 19},
  [628] = {.lex_state = 200, .external_lex_state = 17},
  [629] = {.lex_state = 189, .external_lex_state = 22},
  [630] = {.lex_state = 201, .external_lex_state = 21},
  [631] = {.lex_state = 195, .external_lex_state = 21},
  [632] = {.lex_state = 189, .external_lex_state = 22},
  [633] = {.lex_state = 189, .external_lex_state = 22},
  [634] = {.lex_state = 200, .external_lex_state = 17},
  [635] = {.lex_state = 195, .external_lex_state = 21},
  [636] = {.lex_state = 295, .external_lex_state = 2},
  [637] = {.lex_state = 200, .external_lex_state = 17},
  [638] = {.lex_state = 200, .external_lex_state = 17},
  [639] = {.lex_state = 203, .external_lex_state = 19},
  [640] = {.lex_state = 203, .external_lex_state = 19},
  [641] = {.lex_state = 200, .external_lex_state = 17},
  [642] = {.lex_state = 533, .external_lex_state = 19},
  [643] = {.lex_state = 200, .external_lex_state = 17},
  [644] = {.lex_state = 295, .external_lex_state = 2},
  [645] = {.lex_state = 198, .external_lex_state = 19},
  [646] = {.lex_state = 533, .external_lex_state = 19},
  [647] = {.lex_state = 203, .external_lex_state = 19},
  [648] = {.lex_state = 198, .external_lex_state = 19},
  [649] = {.lex_state = 533, .external_lex_state = 19},
  [650] = {.lex_state = 295, .external_lex_state = 2},
  [651] = {.lex_state = 533, .external_lex_state = 19},
  [652] = {.lex_state = 203, .external_lex_state = 19},
  [653] = {.lex_state = 198, .external_lex_state = 22},
  [654] = {.lex_state = 533, .external_lex_state = 19},
  [655] = {.lex_state = 259, .external_lex_state = 6},
  [656] = {.lex_state = 295, .external_lex_state = 2},
  [657] = {.lex_state = 201, .external_lex_state = 21},
  [658] = {.lex_state = 198, .external_lex_state = 22},
  [659] = {.lex_state = 262, .external_lex_state = 24},
  [660] = {.lex_state = 207, .external_lex_state = 22},
  [661] = {.lex_state = 262, .external_lex_state = 24},
  [662] = {.lex_state = 206, .external_lex_state = 19},
  [663] = {.lex_state = 209, .external_lex_state = 19},
  [664] = {.lex_state = 530, .external_lex_state = 22},
  [665] = {.lex_state = 198, .external_lex_state = 22},
  [666] = {.lex_state = 209, .external_lex_state = 19},
  [667] = {.lex_state = 530, .external_lex_state = 22},
  [668] = {.lex_state = 206, .external_lex_state = 19},
  [669] = {.lex_state = 535, .external_lex_state = 19},
  [670] = {.lex_state = 209, .external_lex_state = 19},
  [671] = {.lex_state = 262, .external_lex_state = 24},
  [672] = {.lex_state = 533, .external_lex_state = 22},
  [673] = {.lex_state = 533, .external_lex_state = 19},
  [674] = {.lex_state = 209, .external_lex_state = 19},
  [675] = {.lex_state = 180, .external_lex_state = 16},
  [676] = {.lex_state = 209, .external_lex_state = 19},
  [677] = {.lex_state = 206, .external_lex_state = 19},
  [678] = {.lex_state = 533, .external_lex_state = 22},
  [679] = {.lex_state = 209, .external_lex_state = 19},
  [680] = {.lex_state = 207, .external_lex_state = 22},
  [681] = {.lex_state = 201, .external_lex_state = 21},
  [682] = {.lex_state = 533, .external_lex_state = 19},
  [683] = {.lex_state = 535, .external_lex_state = 19},
  [684] = {.lex_state = 533, .external_lex_state = 22},
  [685] = {.lex_state = 262, .external_lex_state = 24},
  [686] = {.lex_state = 533, .external_lex_state = 19},
  [687] = {.lex_state = 535, .external_lex_state = 19},
  [688] = {.lex_state = 180, .external_lex_state = 16},
  [689] = {.lex_state = 533, .external_lex_state = 19},
  [690] = {.lex_state = 530, .external_lex_state = 22},
  [691] = {.lex_state = 533, .external_lex_state = 22},
  [692] = {.lex_state = 206, .external_lex_state = 19},
  [693] = {.lex_state = 530, .external_lex_state = 22},
  [694] = {.lex_state = 533, .external_lex_state = 19},
  [695] = {.lex_state = 530, .external_lex_state = 22},
  [696] = {.lex_state = 214, .external_lex_state = 9},
  [697] = {.lex_state = 262, .external_lex_state = 24},
  [698] = {.lex_state = 262, .external_lex_state = 24},
  [699] = {.lex_state = 214, .external_lex_state = 9},
  [700] = {.lex_state = 535, .external_lex_state = 19},
  [701] = {.lex_state = 262, .external_lex_state = 24},
  [702] = {.lex_state = 533, .external_lex_state = 22},
  [703] = {.lex_state = 180, .external_lex_state = 16},
  [704] = {.lex_state = 533, .external_lex_state = 19},
  [705] = {.lex_state = 533, .external_lex_state = 22},
  [706] = {.lex_state = 201, .external_lex_state = 21},
  [707] = {.lex_state = 207, .external_lex_state = 22},
  [708] = {.lex_state = 533, .external_lex_state = 19},
  [709] = {.lex_state = 530, .external_lex_state = 22},
  [710] = {.lex_state = 216, .external_lex_state = 19},
  [711] = {.lex_state = 262, .external_lex_state = 24},
  [712] = {.lex_state = 262, .external_lex_state = 24},
  [713] = {.lex_state = 262, .external_lex_state = 24},
  [714] = {.lex_state = 534, .external_lex_state = 22},
  [715] = {.lex_state = 533, .external_lex_state = 22},
  [716] = {.lex_state = 536, .external_lex_state = 19},
  [717] = {.lex_state = 536, .external_lex_state = 19},
  [718] = {.lex_state = 530, .external_lex_state = 22},
  [719] = {.lex_state = 533, .external_lex_state = 22},
  [720] = {.lex_state = 533, .external_lex_state = 22},
  [721] = {.lex_state = 536, .external_lex_state = 19},
  [722] = {.lex_state = 207, .external_lex_state = 22},
  [723] = {.lex_state = 536, .external_lex_state = 19},
  [724] = {.lex_state = 207, .external_lex_state = 22},
  [725] = {.lex_state = 536, .external_lex_state = 19},
  [726] = {.lex_state = 530, .external_lex_state = 22},
  [727] = {.lex_state = 299, .external_lex_state = 25},
  [728] = {.lex_state = 213, .external_lex_state = 26},
  [729] = {.lex_state = 213, .external_lex_state = 26},
  [730] = {.lex_state = 262, .external_lex_state = 24},
  [731] = {.lex_state = 530, .external_lex_state = 22},
  [732] = {.lex_state = 248, .external_lex_state = 27},
  [733] = {.lex_state = 533, .external_lex_state = 22},
  [734] = {.lex_state = 262, .external_lex_state = 24},
  [735] = {.lex_state = 536, .external_lex_state = 19},
  [736] = {.lex_state = 213, .external_lex_state = 26},
  [737] = {.lex_state = 248, .external_lex_state = 27},
  [738] = {.lex_state = 533, .external_lex_state = 22},
  [739] = {.lex_state = 533, .external_lex_state = 22},
  [740] = {.lex_state = 216, .external_lex_state = 19},
  [741] = {.lex_state = 534, .external_lex_state = 22},
  [742] = {.lex_state = 534, .external_lex_state = 22},
  [743] = {.lex_state = 530, .external_lex_state = 22},
  [744] = {.lex_state = 213, .external_lex_state = 26},
  [745] = {.lex_state = 213, .external_lex_state = 26},
  [746] = {.lex_state = 189, .external_lex_state = 19},
  [747] = {.lex_state = 534, .external_lex_state = 22},
  [748] = {.lex_state = 189, .external_lex_state = 19},
  [749] = {.lex_state = 535, .external_lex_state = 19},
  [750] = {.lex_state = 535, .external_lex_state = 19},
  [751] = {.lex_state = 299, .external_lex_state = 25},
  [752] = {.lex_state = 299, .external_lex_state = 25},
  [753] = {.lex_state = 213, .external_lex_state = 26},
  [754] = {.lex_state = 533, .external_lex_state = 22},
  [755] = {.lex_state = 533, .external_lex_state = 22},
  [756] = {.lex_state = 216, .external_lex_state = 19},
  [757] = {.lex_state = 534, .external_lex_state = 22},
  [758] = {.lex_state = 213, .external_lex_state = 26},
  [759] = {.lex_state = 533, .external_lex_state = 19},
  [760] = {.lex_state = 262, .external_lex_state = 24},
  [761] = {.lex_state = 530, .external_lex_state = 22},
  [762] = {.lex_state = 262, .external_lex_state = 24},
  [763] = {.lex_state = 262, .external_lex_state = 24},
  [764] = {.lex_state = 533, .external_lex_state = 22},
  [765] = {.lex_state = 216, .external_lex_state = 19},
  [766] = {.lex_state = 216, .external_lex_state = 19},
  [767] = {.lex_state = 220, .external_lex_state = 20},
  [768] = {.lex_state = 220, .external_lex_state = 20},
  [769] = {.lex_state = 213, .external_lex_state = 26},
  [770] = {.lex_state = 213, .external_lex_state = 26},
  [771] = {.lex_state = 213, .external_lex_state = 26},
  [772] = {.lex_state = 533, .external_lex_state = 19},
  [773] = {.lex_state = 534, .external_lex_state = 22},
  [774] = {.lex_state = 216, .external_lex_state = 19},
  [775] = {.lex_state = 535, .external_lex_state = 19},
  [776] = {.lex_state = 535, .external_lex_state = 19},
  [777] = {.lex_state = 189, .external_lex_state = 19},
  [778] = {.lex_state = 533, .external_lex_state = 22},
  [779] = {.lex_state = 534, .external_lex_state = 22},
  [780] = {.lex_state = 260, .external_lex_state = 2},
  [781] = {.lex_state = 534, .external_lex_state = 22},
  [782] = {.lex_state = 249, .external_lex_state = 28},
  [783] = {.lex_state = 534, .external_lex_state = 22},
  [784] = {.lex_state = 536, .external_lex_state = 19},
  [785] = {.lex_state = 533, .external_lex_state = 22},
  [786] = {.lex_state = 534, .external_lex_state = 22},
  [787] = {.lex_state = 536, .external_lex_state = 19},
  [788] = {.lex_state = 534, .external_lex_state = 22},
  [789] = {.lex_state = 534, .external_lex_state = 22},
  [790] = {.lex_state = 533, .external_lex_state = 22},
  [791] = {.lex_state = 534, .external_lex_state = 22},
  [792] = {.lex_state = 533, .external_lex_state = 22},
  [793] = {.lex_state = 536, .external_lex_state = 19},
  [794] = {.lex_state = 542, .external_lex_state = 2},
  [795] = {.lex_state = 536, .external_lex_state = 19},
  [796] = {.lex_state = 536, .external_lex_state = 19},
  [797] = {.lex_state = 308, .external_lex_state = 29},
  [798] = {.lex_state = 534, .external_lex_state = 22},
  [799] = {.lex_state = 308, .external_lex_state = 29},
  [800] = {.lex_state = 542, .external_lex_state = 2},
  [801] = {.lex_state = 536, .external_lex_state = 19},
  [802] = {.lex_state = 542, .external_lex_state = 2},
  [803] = {.lex_state = 308, .external_lex_state = 29},
  [804] = {.lex_state = 534, .external_lex_state = 22},
  [805] = {.lex_state = 249, .external_lex_state = 28},
  [806] = {.lex_state = 198, .external_lex_state = 19},
  [807] = {.lex_state = 195, .external_lex_state = 17},
  [808] = {.lex_state = 215, .external_lex_state = 30},
  [809] = {.lex_state = 195, .external_lex_state = 17},
  [810] = {.lex_state = 542, .external_lex_state = 2},
  [811] = {.lex_state = 198, .external_lex_state = 19},
  [812] = {.lex_state = 542, .external_lex_state = 2},
  [813] = {.lex_state = 215, .external_lex_state = 30},
  [814] = {.lex_state = 215, .external_lex_state = 30},
  [815] = {.lex_state = 215, .external_lex_state = 30},
  [816] = {.lex_state = 198, .external_lex_state = 19},
  [817] = {.lex_state = 302, .external_lex_state = 31},
  [818] = {.lex_state = 195, .external_lex_state = 17},
  [819] = {.lex_state = 215, .external_lex_state = 30},
  [820] = {.lex_state = 215, .external_lex_state = 30},
  [821] = {.lex_state = 302, .external_lex_state = 31},
  [822] = {.lex_state = 195, .external_lex_state = 17},
  [823] = {.lex_state = 302, .external_lex_state = 31},
  [824] = {.lex_state = 215, .external_lex_state = 30},
  [825] = {.lex_state = 534, .external_lex_state = 22},
  [826] = {.lex_state = 534, .external_lex_state = 22},
  [827] = {.lex_state = 542, .external_lex_state = 2},
  [828] = {.lex_state = 534, .external_lex_state = 22},
  [829] = {.lex_state = 534, .external_lex_state = 22},
  [830] = {.lex_state = 302, .external_lex_state = 31},
  [831] = {.lex_state = 215, .external_lex_state = 30},
  [832] = {.lex_state = 215, .external_lex_state = 32},
  [833] = {.lex_state = 198, .external_lex_state = 19},
  [834] = {.lex_state = 215, .external_lex_state = 32},
  [835] = {.lex_state = 198, .external_lex_state = 19},
  [836] = {.lex_state = 198, .external_lex_state = 19},
  [837] = {.lex_state = 217, .external_lex_state = 4},
  [838] = {.lex_state = 217, .external_lex_state = 4},
  [839] = {.lex_state = 198, .external_lex_state = 19},
  [840] = {.lex_state = 250, .external_lex_state = 31},
  [841] = {.lex_state = 217, .external_lex_state = 4},
  [842] = {.lex_state = 250, .external_lex_state = 31},
  [843] = {.lex_state = 215, .external_lex_state = 30},
  [844] = {.lex_state = 215, .external_lex_state = 30},
  [845] = {.lex_state = 215, .external_lex_state = 30},
  [846] = {.lex_state = 215, .external_lex_state = 30},
  [847] = {.lex_state = 198, .external_lex_state = 19},
  [848] = {.lex_state = 215, .external_lex_state = 30},
  [849] = {.lex_state = 198, .external_lex_state = 19},
  [850] = {.lex_state = 217, .external_lex_state = 4},
  [851] = {.lex_state = 298, .external_lex_state = 29},
  [852] = {.lex_state = 217, .external_lex_state = 4},
  [853] = {.lex_state = 217, .external_lex_state = 4},
  [854] = {.lex_state = 215, .external_lex_state = 32},
  [855] = {.lex_state = 215, .external_lex_state = 30},
  [856] = {.lex_state = 215, .external_lex_state = 30},
  [857] = {.lex_state = 215, .external_lex_state = 30},
  [858] = {.lex_state = 298, .external_lex_state = 29},
  [859] = {.lex_state = 215, .external_lex_state = 32},
  [860] = {.lex_state = 198, .external_lex_state = 22},
  [861] = {.lex_state = 215, .external_lex_state = 32},
  [862] = {.lex_state = 198, .external_lex_state = 22},
  [863] = {.lex_state = 198, .external_lex_state = 22},
  [864] = {.lex_state = 215, .external_lex_state = 30},
  [865] = {.lex_state = 189, .external_lex_state = 22},
  [866] = {.lex_state = 215, .external_lex_state = 30},
  [867] = {.lex_state = 302, .external_lex_state = 33},
  [868] = {.lex_state = 215, .external_lex_state = 30},
  [869] = {.lex_state = 215, .external_lex_state = 30},
  [870] = {.lex_state = 298, .external_lex_state = 29},
  [871] = {.lex_state = 215, .external_lex_state = 30},
  [872] = {.lex_state = 215, .external_lex_state = 30},
  [873] = {.lex_state = 222, .external_lex_state = 19},
  [874] = {.lex_state = 215, .external_lex_state = 32},
  [875] = {.lex_state = 189, .external_lex_state = 22},
  [876] = {.lex_state = 215, .external_lex_state = 32},
  [877] = {.lex_state = 222, .external_lex_state = 19},
  [878] = {.lex_state = 189, .external_lex_state = 22},
  [879] = {.lex_state = 302, .external_lex_state = 33},
  [880] = {.lex_state = 215, .external_lex_state = 32},
  [881] = {.lex_state = 250, .external_lex_state = 31},
  [882] = {.lex_state = 302, .external_lex_state = 31},
  [883] = {.lex_state = 250, .external_lex_state = 31},
  [884] = {.lex_state = 215, .external_lex_state = 30},
  [885] = {.lex_state = 215, .external_lex_state = 32},
  [886] = {.lex_state = 217, .external_lex_state = 4},
  [887] = {.lex_state = 215, .external_lex_state = 30},
  [888] = {.lex_state = 215, .external_lex_state = 30},
  [889] = {.lex_state = 215, .external_lex_state = 30},
  [890] = {.lex_state = 222, .external_lex_state = 19},
  [891] = {.lex_state = 215, .external_lex_state = 32},
  [892] = {.lex_state = 222, .external_lex_state = 19},
  [893] = {.lex_state = 215, .external_lex_state = 32},
  [894] = {.lex_state = 215, .external_lex_state = 32},
  [895] = {.lex_state = 302, .external_lex_state = 33},
  [896] = {.lex_state = 224, .external_lex_state = 19},
  [897] = {.lex_state = 298, .external_lex_state = 34},
  [898] = {.lex_state = 215, .external_lex_state = 32},
  [899] = {.lex_state = 224, .external_lex_state = 19},
  [900] = {.lex_state = 215, .external_lex_state = 32},
  [901] = {.lex_state = 215, .external_lex_state = 32},
  [902] = {.lex_state = 268, .external_lex_state = 35},
  [903] = {.lex_state = 298, .external_lex_state = 34},
  [904] = {.lex_state = 268, .external_lex_state = 35},
  [905] = {.lex_state = 298, .external_lex_state = 29},
  [906] = {.lex_state = 224, .external_lex_state = 19},
  [907] = {.lex_state = 537, .external_lex_state = 4},
  [908] = {.lex_state = 198, .external_lex_state = 22},
  [909] = {.lex_state = 198, .external_lex_state = 22},
  [910] = {.lex_state = 215, .external_lex_state = 32},
  [911] = {.lex_state = 224, .external_lex_state = 19},
  [912] = {.lex_state = 215, .external_lex_state = 32},
  [913] = {.lex_state = 533, .external_lex_state = 19},
  [914] = {.lex_state = 217, .external_lex_state = 4},
  [915] = {.lex_state = 217, .external_lex_state = 4},
  [916] = {.lex_state = 217, .external_lex_state = 4},
  [917] = {.lex_state = 215, .external_lex_state = 32},
  [918] = {.lex_state = 215, .external_lex_state = 32},
  [919] = {.lex_state = 215, .external_lex_state = 32},
  [920] = {.lex_state = 537, .external_lex_state = 4},
  [921] = {.lex_state = 217, .external_lex_state = 4},
  [922] = {.lex_state = 215, .external_lex_state = 32},
  [923] = {.lex_state = 217, .external_lex_state = 4},
  [924] = {.lex_state = 215, .external_lex_state = 32},
  [925] = {.lex_state = 296, .external_lex_state = 36},
  [926] = {.lex_state = 177, .external_lex_state = 13},
  [927] = {.lex_state = 217, .external_lex_state = 4},
  [928] = {.lex_state = 217, .external_lex_state = 4},
  [929] = {.lex_state = 217, .external_lex_state = 4},
  [930] = {.lex_state = 215, .external_lex_state = 32},
  [931] = {.lex_state = 217, .external_lex_state = 4},
  [932] = {.lex_state = 537, .external_lex_state = 4},
  [933] = {.lex_state = 296, .external_lex_state = 36},
  [934] = {.lex_state = 217, .external_lex_state = 4},
  [935] = {.lex_state = 537, .external_lex_state = 4},
  [936] = {.lex_state = 298, .external_lex_state = 34},
  [937] = {.lex_state = 296, .external_lex_state = 36},
  [938] = {.lex_state = 537, .external_lex_state = 4},
  [939] = {.lex_state = 302, .external_lex_state = 33},
  [940] = {.lex_state = 537, .external_lex_state = 4},
  [941] = {.lex_state = 224, .external_lex_state = 19},
  [942] = {.lex_state = 224, .external_lex_state = 19},
  [943] = {.lex_state = 252, .external_lex_state = 31},
  [944] = {.lex_state = 302, .external_lex_state = 33},
  [945] = {.lex_state = 533, .external_lex_state = 19},
  [946] = {.lex_state = 252, .external_lex_state = 31},
  [947] = {.lex_state = 537, .external_lex_state = 4},
  [948] = {.lex_state = 533, .external_lex_state = 19},
  [949] = {.lex_state = 298, .external_lex_state = 29},
  [950] = {.lex_state = 308, .external_lex_state = 34},
  [951] = {.lex_state = 296, .external_lex_state = 36},
  [952] = {.lex_state = 177, .external_lex_state = 13},
  [953] = {.lex_state = 251, .external_lex_state = 29},
  [954] = {.lex_state = 251, .external_lex_state = 29},
  [955] = {.lex_state = 296, .external_lex_state = 36},
  [956] = {.lex_state = 251, .external_lex_state = 29},
  [957] = {.lex_state = 537, .external_lex_state = 4},
  [958] = {.lex_state = 296, .external_lex_state = 36},
  [959] = {.lex_state = 537, .external_lex_state = 4},
  [960] = {.lex_state = 251, .external_lex_state = 29},
  [961] = {.lex_state = 533, .external_lex_state = 19},
  [962] = {.lex_state = 308, .external_lex_state = 34},
  [963] = {.lex_state = 215, .external_lex_state = 32},
  [964] = {.lex_state = 217, .external_lex_state = 4},
  [965] = {.lex_state = 537, .external_lex_state = 4},
  [966] = {.lex_state = 537, .external_lex_state = 4},
  [967] = {.lex_state = 305, .external_lex_state = 33},
  [968] = {.lex_state = 217, .external_lex_state = 4},
  [969] = {.lex_state = 217, .external_lex_state = 37},
  [970] = {.lex_state = 217, .external_lex_state = 37},
  [971] = {.lex_state = 305, .external_lex_state = 33},
  [972] = {.lex_state = 207, .external_lex_state = 22},
  [973] = {.lex_state = 207, .external_lex_state = 22},
  [974] = {.lex_state = 217, .external_lex_state = 37},
  [975] = {.lex_state = 215, .external_lex_state = 32},
  [976] = {.lex_state = 207, .external_lex_state = 22},
  [977] = {.lex_state = 215, .external_lex_state = 32},
  [978] = {.lex_state = 217, .external_lex_state = 37},
  [979] = {.lex_state = 308, .external_lex_state = 34},
  [980] = {.lex_state = 217, .external_lex_state = 4},
  [981] = {.lex_state = 305, .external_lex_state = 33},
  [982] = {.lex_state = 217, .external_lex_state = 4},
  [983] = {.lex_state = 537, .external_lex_state = 4},
  [984] = {.lex_state = 215, .external_lex_state = 32},
  [985] = {.lex_state = 217, .external_lex_state = 37},
  [986] = {.lex_state = 533, .external_lex_state = 19},
  [987] = {.lex_state = 215, .external_lex_state = 32},
  [988] = {.lex_state = 217, .external_lex_state = 4},
  [989] = {.lex_state = 296, .external_lex_state = 36},
  [990] = {.lex_state = 217, .external_lex_state = 4},
  [991] = {.lex_state = 533, .external_lex_state = 19},
  [992] = {.lex_state = 217, .external_lex_state = 37},
  [993] = {.lex_state = 217, .external_lex_state = 37},
  [994] = {.lex_state = 217, .external_lex_state = 37},
  [995] = {.lex_state = 217, .external_lex_state = 37},
  [996] = {.lex_state = 537, .external_lex_state = 4},
  [997] = {.lex_state = 296, .external_lex_state = 36},
  [998] = {.lex_state = 217, .external_lex_state = 4},
  [999] = {.lex_state = 217, .external_lex_state = 4},
  [1000] = {.lex_state = 217, .external_lex_state = 37},
  [1001] = {.lex_state = 217, .external_lex_state = 37},
  [1002] = {.lex_state = 533, .external_lex_state = 19},
  [1003] = {.lex_state = 217, .external_lex_state = 37},
  [1004] = {.lex_state = 217, .external_lex_state = 4},
  [1005] = {.lex_state = 215, .external_lex_state = 32},
  [1006] = {.lex_state = 533, .external_lex_state = 19},
  [1007] = {.lex_state = 537, .external_lex_state = 4},
  [1008] = {.lex_state = 215, .external_lex_state = 32},
  [1009] = {.lex_state = 215, .external_lex_state = 32},
  [1010] = {.lex_state = 253, .external_lex_state = 29},
  [1011] = {.lex_state = 221, .external_lex_state = 38},
  [1012] = {.lex_state = 537, .external_lex_state = 4},
  [1013] = {.lex_state = 537, .external_lex_state = 4},
  [1014] = {.lex_state = 177, .external_lex_state = 14},
  [1015] = {.lex_state = 537, .external_lex_state = 4},
  [1016] = {.lex_state = 537, .external_lex_state = 4},
  [1017] = {.lex_state = 537, .external_lex_state = 4},
  [1018] = {.lex_state = 537, .external_lex_state = 4},
  [1019] = {.lex_state = 537, .external_lex_state = 4},
  [1020] = {.lex_state = 537, .external_lex_state = 4},
  [1021] = {.lex_state = 537, .external_lex_state = 37},
  [1022] = {.lex_state = 537, .external_lex_state = 4},
  [1023] = {.lex_state = 537, .external_lex_state = 37},
  [1024] = {.lex_state = 267, .external_lex_state = 39},
  [1025] = {.lex_state = 267, .external_lex_state = 39},
  [1026] = {.lex_state = 180, .external_lex_state = 11},
  [1027] = {.lex_state = 537, .external_lex_state = 4},
  [1028] = {.lex_state = 533, .external_lex_state = 19},
  [1029] = {.lex_state = 537, .external_lex_state = 4},
  [1030] = {.lex_state = 533, .external_lex_state = 19},
  [1031] = {.lex_state = 537, .external_lex_state = 4},
  [1032] = {.lex_state = 537, .external_lex_state = 37},
  [1033] = {.lex_state = 537, .external_lex_state = 4},
  [1034] = {.lex_state = 537, .external_lex_state = 4},
  [1035] = {.lex_state = 306, .external_lex_state = 34},
  [1036] = {.lex_state = 533, .external_lex_state = 19},
  [1037] = {.lex_state = 537, .external_lex_state = 4},
  [1038] = {.lex_state = 537, .external_lex_state = 4},
  [1039] = {.lex_state = 537, .external_lex_state = 37},
  [1040] = {.lex_state = 537, .external_lex_state = 37},
  [1041] = {.lex_state = 253, .external_lex_state = 29},
  [1042] = {.lex_state = 537, .external_lex_state = 37},
  [1043] = {.lex_state = 537, .external_lex_state = 4},
  [1044] = {.lex_state = 537, .external_lex_state = 4},
  [1045] = {.lex_state = 177, .external_lex_state = 14},
  [1046] = {.lex_state = 537, .external_lex_state = 4},
  [1047] = {.lex_state = 537, .external_lex_state = 4},
  [1048] = {.lex_state = 253, .external_lex_state = 29},
  [1049] = {.lex_state = 537, .external_lex_state = 4},
  [1050] = {.lex_state = 253, .external_lex_state = 29},
  [1051] = {.lex_state = 265, .external_lex_state = 39},
  [1052] = {.lex_state = 177, .external_lex_state = 13},
  [1053] = {.lex_state = 265, .external_lex_state = 39},
  [1054] = {.lex_state = 537, .external_lex_state = 4},
  [1055] = {.lex_state = 537, .external_lex_state = 4},
  [1056] = {.lex_state = 306, .external_lex_state = 34},
  [1057] = {.lex_state = 537, .external_lex_state = 4},
  [1058] = {.lex_state = 537, .external_lex_state = 4},
  [1059] = {.lex_state = 533, .external_lex_state = 19},
  [1060] = {.lex_state = 537, .external_lex_state = 4},
  [1061] = {.lex_state = 537, .external_lex_state = 4},
  [1062] = {.lex_state = 537, .external_lex_state = 4},
  [1063] = {.lex_state = 537, .external_lex_state = 4},
  [1064] = {.lex_state = 537, .external_lex_state = 4},
  [1065] = {.lex_state = 537, .external_lex_state = 37},
  [1066] = {.lex_state = 537, .external_lex_state = 37},
  [1067] = {.lex_state = 221, .external_lex_state = 38},
  [1068] = {.lex_state = 221, .external_lex_state = 38},
  [1069] = {.lex_state = 537, .external_lex_state = 37},
  [1070] = {.lex_state = 268, .external_lex_state = 40},
  [1071] = {.lex_state = 268, .external_lex_state = 40},
  [1072] = {.lex_state = 537, .external_lex_state = 37},
  [1073] = {.lex_state = 223, .external_lex_state = 38},
  [1074] = {.lex_state = 223, .external_lex_state = 38},
  [1075] = {.lex_state = 306, .external_lex_state = 34},
  [1076] = {.lex_state = 537, .external_lex_state = 37},
  [1077] = {.lex_state = 537, .external_lex_state = 4},
  [1078] = {.lex_state = 537, .external_lex_state = 4},
  [1079] = {.lex_state = 223, .external_lex_state = 38},
  [1080] = {.lex_state = 537, .external_lex_state = 4},
  [1081] = {.lex_state = 223, .external_lex_state = 38},
  [1082] = {.lex_state = 177, .external_lex_state = 13},
  [1083] = {.lex_state = 537, .external_lex_state = 4},
  [1084] = {.lex_state = 537, .external_lex_state = 4},
  [1085] = {.lex_state = 537, .external_lex_state = 4},
  [1086] = {.lex_state = 221, .external_lex_state = 38},
  [1087] = {.lex_state = 221, .external_lex_state = 38},
  [1088] = {.lex_state = 221, .external_lex_state = 38},
  [1089] = {.lex_state = 223, .external_lex_state = 38},
  [1090] = {.lex_state = 298, .external_lex_state = 34},
  [1091] = {.lex_state = 537, .external_lex_state = 37},
  [1092] = {.lex_state = 221, .external_lex_state = 38},
  [1093] = {.lex_state = 537, .external_lex_state = 37},
  [1094] = {.lex_state = 537, .external_lex_state = 37},
  [1095] = {.lex_state = 177, .external_lex_state = 14},
  [1096] = {.lex_state = 533, .external_lex_state = 19},
  [1097] = {.lex_state = 180, .external_lex_state = 11},
  [1098] = {.lex_state = 533, .external_lex_state = 19},
  [1099] = {.lex_state = 217, .external_lex_state = 37},
  [1100] = {.lex_state = 223, .external_lex_state = 38},
  [1101] = {.lex_state = 217, .external_lex_state = 37},
  [1102] = {.lex_state = 298, .external_lex_state = 34},
  [1103] = {.lex_state = 217, .external_lex_state = 37},
  [1104] = {.lex_state = 537, .external_lex_state = 4},
  [1105] = {.lex_state = 537, .external_lex_state = 4},
  [1106] = {.lex_state = 537, .external_lex_state = 37},
  [1107] = {.lex_state = 537, .external_lex_state = 37},
  [1108] = {.lex_state = 537, .external_lex_state = 37},
  [1109] = {.lex_state = 537, .external_lex_state = 37},
  [1110] = {.lex_state = 207, .external_lex_state = 22},
  [1111] = {.lex_state = 207, .external_lex_state = 22},
  [1112] = {.lex_state = 537, .external_lex_state = 4},
  [1113] = {.lex_state = 537, .external_lex_state = 4},
  [1114] = {.lex_state = 269, .external_lex_state = 39},
  [1115] = {.lex_state = 269, .external_lex_state = 39},
  [1116] = {.lex_state = 537, .external_lex_state = 4},
  [1117] = {.lex_state = 305, .external_lex_state = 33},
  [1118] = {.lex_state = 305, .external_lex_state = 33},
  [1119] = {.lex_state = 217, .external_lex_state = 37},
  [1120] = {.lex_state = 217, .external_lex_state = 37},
  [1121] = {.lex_state = 223, .external_lex_state = 38},
  [1122] = {.lex_state = 537, .external_lex_state = 37},
  [1123] = {.lex_state = 537, .external_lex_state = 37},
  [1124] = {.lex_state = 537, .external_lex_state = 37},
  [1125] = {.lex_state = 537, .external_lex_state = 4},
  [1126] = {.lex_state = 537, .external_lex_state = 4},
  [1127] = {.lex_state = 223, .external_lex_state = 38},
  [1128] = {.lex_state = 537, .external_lex_state = 4},
  [1129] = {.lex_state = 217, .external_lex_state = 37},
  [1130] = {.lex_state = 217, .external_lex_state = 37},
  [1131] = {.lex_state = 537, .external_lex_state = 37},
  [1132] = {.lex_state = 533, .external_lex_state = 19},
  [1133] = {.lex_state = 217, .external_lex_state = 37},
  [1134] = {.lex_state = 266, .external_lex_state = 39},
  [1135] = {.lex_state = 253, .external_lex_state = 29},
  [1136] = {.lex_state = 217, .external_lex_state = 37},
  [1137] = {.lex_state = 217, .external_lex_state = 37},
  [1138] = {.lex_state = 217, .external_lex_state = 37},
  [1139] = {.lex_state = 217, .external_lex_state = 37},
  [1140] = {.lex_state = 537, .external_lex_state = 37},
  [1141] = {.lex_state = 253, .external_lex_state = 29},
  [1142] = {.lex_state = 217, .external_lex_state = 37},
  [1143] = {.lex_state = 217, .external_lex_state = 37},
  [1144] = {.lex_state = 537, .external_lex_state = 37},
  [1145] = {.lex_state = 217, .external_lex_state = 37},
  [1146] = {.lex_state = 177, .external_lex_state = 14},
  [1147] = {.lex_state = 223, .external_lex_state = 38},
  [1148] = {.lex_state = 223, .external_lex_state = 38},
  [1149] = {.lex_state = 177, .external_lex_state = 14},
  [1150] = {.lex_state = 533, .external_lex_state = 19},
  [1151] = {.lex_state = 533, .external_lex_state = 19},
  [1152] = {.lex_state = 537, .external_lex_state = 4},
  [1153] = {.lex_state = 533, .external_lex_state = 19},
  [1154] = {.lex_state = 266, .external_lex_state = 39},
  [1155] = {.lex_state = 217, .external_lex_state = 37},
  [1156] = {.lex_state = 221, .external_lex_state = 38},
  [1157] = {.lex_state = 217, .external_lex_state = 37},
  [1158] = {.lex_state = 223, .external_lex_state = 38},
  [1159] = {.lex_state = 217, .external_lex_state = 37},
  [1160] = {.lex_state = 223, .external_lex_state = 38},
  [1161] = {.lex_state = 217, .external_lex_state = 37},
  [1162] = {.lex_state = 223, .external_lex_state = 38},
  [1163] = {.lex_state = 223, .external_lex_state = 38},
  [1164] = {.lex_state = 270, .external_lex_state = 39},
  [1165] = {.lex_state = 223, .external_lex_state = 38},
  [1166] = {.lex_state = 225, .external_lex_state = 41},
  [1167] = {.lex_state = 225, .external_lex_state = 41},
  [1168] = {.lex_state = 177, .external_lex_state = 14},
  [1169] = {.lex_state = 180, .external_lex_state = 11},
  [1170] = {.lex_state = 225, .external_lex_state = 41},
  [1171] = {.lex_state = 225, .external_lex_state = 41},
  [1172] = {.lex_state = 306, .external_lex_state = 34},
  [1173] = {.lex_state = 271, .external_lex_state = 42},
  [1174] = {.lex_state = 537, .external_lex_state = 37},
  [1175] = {.lex_state = 180, .external_lex_state = 16},
  [1176] = {.lex_state = 537, .external_lex_state = 37},
  [1177] = {.lex_state = 537, .external_lex_state = 37},
  [1178] = {.lex_state = 306, .external_lex_state = 34},
  [1179] = {.lex_state = 527, .external_lex_state = 11},
  [1180] = {.lex_state = 180, .external_lex_state = 16},
  [1181] = {.lex_state = 177, .external_lex_state = 14},
  [1182] = {.lex_state = 177, .external_lex_state = 14},
  [1183] = {.lex_state = 537, .external_lex_state = 37},
  [1184] = {.lex_state = 227, .external_lex_state = 41},
  [1185] = {.lex_state = 537, .external_lex_state = 37},
  [1186] = {.lex_state = 537, .external_lex_state = 37},
  [1187] = {.lex_state = 537, .external_lex_state = 37},
  [1188] = {.lex_state = 537, .external_lex_state = 37},
  [1189] = {.lex_state = 537, .external_lex_state = 37},
  [1190] = {.lex_state = 221, .external_lex_state = 43},
  [1191] = {.lex_state = 221, .external_lex_state = 43},
  [1192] = {.lex_state = 227, .external_lex_state = 41},
  [1193] = {.lex_state = 227, .external_lex_state = 41},
  [1194] = {.lex_state = 537, .external_lex_state = 37},
  [1195] = {.lex_state = 537, .external_lex_state = 37},
  [1196] = {.lex_state = 537, .external_lex_state = 37},
  [1197] = {.lex_state = 270, .external_lex_state = 39},
  [1198] = {.lex_state = 223, .external_lex_state = 38},
  [1199] = {.lex_state = 223, .external_lex_state = 38},
  [1200] = {.lex_state = 537, .external_lex_state = 37},
  [1201] = {.lex_state = 180, .external_lex_state = 16},
  [1202] = {.lex_state = 180, .external_lex_state = 16},
  [1203] = {.lex_state = 537, .external_lex_state = 37},
  [1204] = {.lex_state = 223, .external_lex_state = 38},
  [1205] = {.lex_state = 537, .external_lex_state = 37},
  [1206] = {.lex_state = 221, .external_lex_state = 38},
  [1207] = {.lex_state = 221, .external_lex_state = 38},
  [1208] = {.lex_state = 223, .external_lex_state = 38},
  [1209] = {.lex_state = 527, .external_lex_state = 11},
  [1210] = {.lex_state = 223, .external_lex_state = 38},
  [1211] = {.lex_state = 221, .external_lex_state = 43},
  [1212] = {.lex_state = 527, .external_lex_state = 11},
  [1213] = {.lex_state = 221, .external_lex_state = 38},
  [1214] = {.lex_state = 221, .external_lex_state = 43},
  [1215] = {.lex_state = 537, .external_lex_state = 37},
  [1216] = {.lex_state = 537, .external_lex_state = 37},
  [1217] = {.lex_state = 221, .external_lex_state = 43},
  [1218] = {.lex_state = 223, .external_lex_state = 38},
  [1219] = {.lex_state = 537, .external_lex_state = 37},
  [1220] = {.lex_state = 221, .external_lex_state = 38},
  [1221] = {.lex_state = 223, .external_lex_state = 38},
  [1222] = {.lex_state = 221, .external_lex_state = 38},
  [1223] = {.lex_state = 537, .external_lex_state = 37},
  [1224] = {.lex_state = 221, .external_lex_state = 38},
  [1225] = {.lex_state = 221, .external_lex_state = 38},
  [1226] = {.lex_state = 221, .external_lex_state = 43},
  [1227] = {.lex_state = 223, .external_lex_state = 43},
  [1228] = {.lex_state = 223, .external_lex_state = 43},
  [1229] = {.lex_state = 177, .external_lex_state = 14},
  [1230] = {.lex_state = 223, .external_lex_state = 43},
  [1231] = {.lex_state = 537, .external_lex_state = 37},
  [1232] = {.lex_state = 223, .external_lex_state = 38},
  [1233] = {.lex_state = 223, .external_lex_state = 38},
  [1234] = {.lex_state = 223, .external_lex_state = 43},
  [1235] = {.lex_state = 221, .external_lex_state = 38},
  [1236] = {.lex_state = 221, .external_lex_state = 38},
  [1237] = {.lex_state = 537, .external_lex_state = 37},
  [1238] = {.lex_state = 537, .external_lex_state = 37},
  [1239] = {.lex_state = 537, .external_lex_state = 37},
  [1240] = {.lex_state = 221, .external_lex_state = 38},
  [1241] = {.lex_state = 221, .external_lex_state = 38},
  [1242] = {.lex_state = 221, .external_lex_state = 38},
  [1243] = {.lex_state = 537, .external_lex_state = 37},
  [1244] = {.lex_state = 537, .external_lex_state = 37},
  [1245] = {.lex_state = 227, .external_lex_state = 41},
  [1246] = {.lex_state = 223, .external_lex_state = 38},
  [1247] = {.lex_state = 223, .external_lex_state = 38},
  [1248] = {.lex_state = 537, .external_lex_state = 37},
  [1249] = {.lex_state = 227, .external_lex_state = 41},
  [1250] = {.lex_state = 227, .external_lex_state = 41},
  [1251] = {.lex_state = 225, .external_lex_state = 41},
  [1252] = {.lex_state = 221, .external_lex_state = 38},
  [1253] = {.lex_state = 221, .external_lex_state = 38},
  [1254] = {.lex_state = 223, .external_lex_state = 38},
  [1255] = {.lex_state = 223, .external_lex_state = 38},
  [1256] = {.lex_state = 223, .external_lex_state = 38},
  [1257] = {.lex_state = 225, .external_lex_state = 41},
  [1258] = {.lex_state = 227, .external_lex_state = 41},
  [1259] = {.lex_state = 537, .external_lex_state = 37},
  [1260] = {.lex_state = 180, .external_lex_state = 11},
  [1261] = {.lex_state = 227, .external_lex_state = 41},
  [1262] = {.lex_state = 225, .external_lex_state = 41},
  [1263] = {.lex_state = 227, .external_lex_state = 41},
  [1264] = {.lex_state = 227, .external_lex_state = 41},
  [1265] = {.lex_state = 227, .external_lex_state = 41},
  [1266] = {.lex_state = 225, .external_lex_state = 41},
  [1267] = {.lex_state = 527, .external_lex_state = 11},
  [1268] = {.lex_state = 537, .external_lex_state = 37},
  [1269] = {.lex_state = 227, .external_lex_state = 41},
  [1270] = {.lex_state = 227, .external_lex_state = 41},
  [1271] = {.lex_state = 537, .external_lex_state = 37},
  [1272] = {.lex_state = 537, .external_lex_state = 37},
  [1273] = {.lex_state = 537, .external_lex_state = 37},
  [1274] = {.lex_state = 537, .external_lex_state = 37},
  [1275] = {.lex_state = 227, .external_lex_state = 41},
  [1276] = {.lex_state = 537, .external_lex_state = 37},
  [1277] = {.lex_state = 537, .external_lex_state = 37},
  [1278] = {.lex_state = 221, .external_lex_state = 38},
  [1279] = {.lex_state = 537, .external_lex_state = 37},
  [1280] = {.lex_state = 537, .external_lex_state = 37},
  [1281] = {.lex_state = 177, .external_lex_state = 14},
  [1282] = {.lex_state = 537, .external_lex_state = 37},
  [1283] = {.lex_state = 537, .external_lex_state = 37},
  [1284] = {.lex_state = 537, .external_lex_state = 37},
  [1285] = {.lex_state = 537, .external_lex_state = 37},
  [1286] = {.lex_state = 221, .external_lex_state = 38},
  [1287] = {.lex_state = 537, .external_lex_state = 37},
  [1288] = {.lex_state = 221, .external_lex_state = 38},
  [1289] = {.lex_state = 537, .external_lex_state = 37},
  [1290] = {.lex_state = 223, .external_lex_state = 43},
  [1291] = {.lex_state = 223, .external_lex_state = 43},
  [1292] = {.lex_state = 537, .external_lex_state = 37},
  [1293] = {.lex_state = 537, .external_lex_state = 37},
  [1294] = {.lex_state = 537, .external_lex_state = 37},
  [1295] = {.lex_state = 180, .external_lex_state = 16},
  [1296] = {.lex_state = 537, .external_lex_state = 37},
  [1297] = {.lex_state = 221, .external_lex_state = 38},
  [1298] = {.lex_state = 223, .external_lex_state = 38},
  [1299] = {.lex_state = 221, .external_lex_state = 38},
  [1300] = {.lex_state = 537, .external_lex_state = 37},
  [1301] = {.lex_state = 177, .external_lex_state = 14},
  [1302] = {.lex_state = 223, .external_lex_state = 38},
  [1303] = {.lex_state = 537, .external_lex_state = 37},
  [1304] = {.lex_state = 537, .external_lex_state = 37},
  [1305] = {.lex_state = 223, .external_lex_state = 38},
  [1306] = {.lex_state = 223, .external_lex_state = 38},
  [1307] = {.lex_state = 227, .external_lex_state = 44},
  [1308] = {.lex_state = 229, .external_lex_state = 43},
  [1309] = {.lex_state = 223, .external_lex_state = 43},
  [1310] = {.lex_state = 223, .external_lex_state = 43},
  [1311] = {.lex_state = 227, .external_lex_state = 44},
  [1312] = {.lex_state = 273, .external_lex_state = 45},
  [1313] = {.lex_state = 527, .external_lex_state = 16},
  [1314] = {.lex_state = 213, .external_lex_state = 26},
  [1315] = {.lex_state = 225, .external_lex_state = 41},
  [1316] = {.lex_state = 225, .external_lex_state = 41},
  [1317] = {.lex_state = 213, .external_lex_state = 26},
  [1318] = {.lex_state = 538, .external_lex_state = 41},
  [1319] = {.lex_state = 225, .external_lex_state = 41},
  [1320] = {.lex_state = 539, .external_lex_state = 41},
  [1321] = {.lex_state = 527, .external_lex_state = 16},
  [1322] = {.lex_state = 538, .external_lex_state = 41},
  [1323] = {.lex_state = 227, .external_lex_state = 41},
  [1324] = {.lex_state = 227, .external_lex_state = 41},
  [1325] = {.lex_state = 225, .external_lex_state = 44},
  [1326] = {.lex_state = 225, .external_lex_state = 44},
  [1327] = {.lex_state = 227, .external_lex_state = 41},
  [1328] = {.lex_state = 180, .external_lex_state = 16},
  [1329] = {.lex_state = 180, .external_lex_state = 16},
  [1330] = {.lex_state = 539, .external_lex_state = 41},
  [1331] = {.lex_state = 539, .external_lex_state = 41},
  [1332] = {.lex_state = 213, .external_lex_state = 26},
  [1333] = {.lex_state = 539, .external_lex_state = 41},
  [1334] = {.lex_state = 227, .external_lex_state = 41},
  [1335] = {.lex_state = 527, .external_lex_state = 16},
  [1336] = {.lex_state = 539, .external_lex_state = 41},
  [1337] = {.lex_state = 538, .external_lex_state = 41},
  [1338] = {.lex_state = 180, .external_lex_state = 16},
  [1339] = {.lex_state = 539, .external_lex_state = 41},
  [1340] = {.lex_state = 227, .external_lex_state = 41},
  [1341] = {.lex_state = 538, .external_lex_state = 41},
  [1342] = {.lex_state = 539, .external_lex_state = 41},
  [1343] = {.lex_state = 539, .external_lex_state = 41},
  [1344] = {.lex_state = 539, .external_lex_state = 41},
  [1345] = {.lex_state = 227, .external_lex_state = 41},
  [1346] = {.lex_state = 180, .external_lex_state = 16},
  [1347] = {.lex_state = 227, .external_lex_state = 44},
  [1348] = {.lex_state = 227, .external_lex_state = 41},
  [1349] = {.lex_state = 195, .external_lex_state = 17},
  [1350] = {.lex_state = 225, .external_lex_state = 41},
  [1351] = {.lex_state = 527, .external_lex_state = 11},
  [1352] = {.lex_state = 227, .external_lex_state = 41},
  [1353] = {.lex_state = 225, .external_lex_state = 41},
  [1354] = {.lex_state = 539, .external_lex_state = 41},
  [1355] = {.lex_state = 539, .external_lex_state = 41},
  [1356] = {.lex_state = 538, .external_lex_state = 41},
  [1357] = {.lex_state = 539, .external_lex_state = 41},
  [1358] = {.lex_state = 527, .external_lex_state = 16},
  [1359] = {.lex_state = 225, .external_lex_state = 41},
  [1360] = {.lex_state = 538, .external_lex_state = 41},
  [1361] = {.lex_state = 195, .external_lex_state = 17},
  [1362] = {.lex_state = 538, .external_lex_state = 41},
  [1363] = {.lex_state = 539, .external_lex_state = 41},
  [1364] = {.lex_state = 223, .external_lex_state = 43},
  [1365] = {.lex_state = 223, .external_lex_state = 43},
  [1366] = {.lex_state = 223, .external_lex_state = 43},
  [1367] = {.lex_state = 539, .external_lex_state = 41},
  [1368] = {.lex_state = 186, .external_lex_state = 17},
  [1369] = {.lex_state = 225, .external_lex_state = 44},
  [1370] = {.lex_state = 539, .external_lex_state = 41},
  [1371] = {.lex_state = 538, .external_lex_state = 41},
  [1372] = {.lex_state = 539, .external_lex_state = 41},
  [1373] = {.lex_state = 539, .external_lex_state = 41},
  [1374] = {.lex_state = 527, .external_lex_state = 16},
  [1375] = {.lex_state = 225, .external_lex_state = 41},
  [1376] = {.lex_state = 527, .external_lex_state = 11},
  [1377] = {.lex_state = 527, .external_lex_state = 16},
  [1378] = {.lex_state = 225, .external_lex_state = 41},
  [1379] = {.lex_state = 527, .external_lex_state = 16},
  [1380] = {.lex_state = 195, .external_lex_state = 17},
  [1381] = {.lex_state = 229, .external_lex_state = 43},
  [1382] = {.lex_state = 195, .external_lex_state = 17},
  [1383] = {.lex_state = 538, .external_lex_state = 41},
  [1384] = {.lex_state = 223, .external_lex_state = 43},
  [1385] = {.lex_state = 225, .external_lex_state = 41},
  [1386] = {.lex_state = 527, .external_lex_state = 11},
  [1387] = {.lex_state = 180, .external_lex_state = 16},
  [1388] = {.lex_state = 229, .external_lex_state = 43},
  [1389] = {.lex_state = 225, .external_lex_state = 41},
  [1390] = {.lex_state = 180, .external_lex_state = 16},
  [1391] = {.lex_state = 539, .external_lex_state = 41},
  [1392] = {.lex_state = 539, .external_lex_state = 41},
  [1393] = {.lex_state = 227, .external_lex_state = 41},
  [1394] = {.lex_state = 227, .external_lex_state = 44},
  [1395] = {.lex_state = 539, .external_lex_state = 41},
  [1396] = {.lex_state = 539, .external_lex_state = 41},
  [1397] = {.lex_state = 539, .external_lex_state = 41},
  [1398] = {.lex_state = 538, .external_lex_state = 41},
  [1399] = {.lex_state = 539, .external_lex_state = 41},
  [1400] = {.lex_state = 223, .external_lex_state = 43},
  [1401] = {.lex_state = 527, .external_lex_state = 16},
  [1402] = {.lex_state = 223, .external_lex_state = 43},
  [1403] = {.lex_state = 223, .external_lex_state = 43},
  [1404] = {.lex_state = 527, .external_lex_state = 11},
  [1405] = {.lex_state = 539, .external_lex_state = 41},
  [1406] = {.lex_state = 223, .external_lex_state = 43},
  [1407] = {.lex_state = 539, .external_lex_state = 41},
  [1408] = {.lex_state = 225, .external_lex_state = 41},
  [1409] = {.lex_state = 227, .external_lex_state = 41},
  [1410] = {.lex_state = 225, .external_lex_state = 41},
  [1411] = {.lex_state = 227, .external_lex_state = 41},
  [1412] = {.lex_state = 225, .external_lex_state = 44},
  [1413] = {.lex_state = 223, .external_lex_state = 43},
  [1414] = {.lex_state = 225, .external_lex_state = 41},
  [1415] = {.lex_state = 225, .external_lex_state = 44},
  [1416] = {.lex_state = 225, .external_lex_state = 41},
  [1417] = {.lex_state = 221, .external_lex_state = 43},
  [1418] = {.lex_state = 221, .external_lex_state = 43},
  [1419] = {.lex_state = 225, .external_lex_state = 41},
  [1420] = {.lex_state = 225, .external_lex_state = 44},
  [1421] = {.lex_state = 539, .external_lex_state = 41},
  [1422] = {.lex_state = 221, .external_lex_state = 43},
  [1423] = {.lex_state = 221, .external_lex_state = 43},
  [1424] = {.lex_state = 227, .external_lex_state = 41},
  [1425] = {.lex_state = 226, .external_lex_state = 46},
  [1426] = {.lex_state = 226, .external_lex_state = 46},
  [1427] = {.lex_state = 227, .external_lex_state = 41},
  [1428] = {.lex_state = 227, .external_lex_state = 41},
  [1429] = {.lex_state = 195, .external_lex_state = 17},
  [1430] = {.lex_state = 227, .external_lex_state = 41},
  [1431] = {.lex_state = 527, .external_lex_state = 16},
  [1432] = {.lex_state = 221, .external_lex_state = 43},
  [1433] = {.lex_state = 195, .external_lex_state = 17},
  [1434] = {.lex_state = 221, .external_lex_state = 43},
  [1435] = {.lex_state = 221, .external_lex_state = 43},
  [1436] = {.lex_state = 527, .external_lex_state = 16},
  [1437] = {.lex_state = 221, .external_lex_state = 43},
  [1438] = {.lex_state = 221, .external_lex_state = 43},
  [1439] = {.lex_state = 221, .external_lex_state = 43},
  [1440] = {.lex_state = 225, .external_lex_state = 41},
  [1441] = {.lex_state = 229, .external_lex_state = 43},
  [1442] = {.lex_state = 225, .external_lex_state = 41},
  [1443] = {.lex_state = 538, .external_lex_state = 41},
  [1444] = {.lex_state = 229, .external_lex_state = 43},
  [1445] = {.lex_state = 539, .external_lex_state = 41},
  [1446] = {.lex_state = 539, .external_lex_state = 41},
  [1447] = {.lex_state = 538, .external_lex_state = 41},
  [1448] = {.lex_state = 186, .external_lex_state = 17},
  [1449] = {.lex_state = 227, .external_lex_state = 41},
  [1450] = {.lex_state = 221, .external_lex_state = 43},
  [1451] = {.lex_state = 229, .external_lex_state = 43},
  [1452] = {.lex_state = 221, .external_lex_state = 43},
  [1453] = {.lex_state = 221, .external_lex_state = 43},
  [1454] = {.lex_state = 227, .external_lex_state = 41},
  [1455] = {.lex_state = 221, .external_lex_state = 43},
  [1456] = {.lex_state = 221, .external_lex_state = 43},
  [1457] = {.lex_state = 227, .external_lex_state = 44},
  [1458] = {.lex_state = 221, .external_lex_state = 43},
  [1459] = {.lex_state = 221, .external_lex_state = 43},
  [1460] = {.lex_state = 538, .external_lex_state = 41},
  [1461] = {.lex_state = 223, .external_lex_state = 43},
  [1462] = {.lex_state = 223, .external_lex_state = 43},
  [1463] = {.lex_state = 223, .external_lex_state = 43},
  [1464] = {.lex_state = 223, .external_lex_state = 43},
  [1465] = {.lex_state = 227, .external_lex_state = 41},
  [1466] = {.lex_state = 227, .external_lex_state = 41},
  [1467] = {.lex_state = 538, .external_lex_state = 41},
  [1468] = {.lex_state = 227, .external_lex_state = 44},
  [1469] = {.lex_state = 225, .external_lex_state = 41},
  [1470] = {.lex_state = 538, .external_lex_state = 41},
  [1471] = {.lex_state = 223, .external_lex_state = 43},
  [1472] = {.lex_state = 225, .external_lex_state = 41},
  [1473] = {.lex_state = 223, .external_lex_state = 43},
  [1474] = {.lex_state = 195, .external_lex_state = 17},
  [1475] = {.lex_state = 538, .external_lex_state = 41},
  [1476] = {.lex_state = 223, .external_lex_state = 43},
  [1477] = {.lex_state = 223, .external_lex_state = 43},
  [1478] = {.lex_state = 221, .external_lex_state = 43},
  [1479] = {.lex_state = 221, .external_lex_state = 43},
  [1480] = {.lex_state = 186, .external_lex_state = 17},
  [1481] = {.lex_state = 538, .external_lex_state = 41},
  [1482] = {.lex_state = 539, .external_lex_state = 41},
  [1483] = {.lex_state = 538, .external_lex_state = 41},
  [1484] = {.lex_state = 538, .external_lex_state = 41},
  [1485] = {.lex_state = 227, .external_lex_state = 44},
  [1486] = {.lex_state = 538, .external_lex_state = 41},
  [1487] = {.lex_state = 538, .external_lex_state = 41},
  [1488] = {.lex_state = 538, .external_lex_state = 41},
  [1489] = {.lex_state = 539, .external_lex_state = 41},
  [1490] = {.lex_state = 538, .external_lex_state = 41},
  [1491] = {.lex_state = 538, .external_lex_state = 41},
  [1492] = {.lex_state = 538, .external_lex_state = 41},
  [1493] = {.lex_state = 538, .external_lex_state = 41},
  [1494] = {.lex_state = 538, .external_lex_state = 41},
  [1495] = {.lex_state = 538, .external_lex_state = 41},
  [1496] = {.lex_state = 225, .external_lex_state = 44},
  [1497] = {.lex_state = 227, .external_lex_state = 44},
  [1498] = {.lex_state = 186, .external_lex_state = 17},
  [1499] = {.lex_state = 538, .external_lex_state = 41},
  [1500] = {.lex_state = 538, .external_lex_state = 41},
  [1501] = {.lex_state = 229, .external_lex_state = 43},
  [1502] = {.lex_state = 538, .external_lex_state = 41},
  [1503] = {.lex_state = 225, .external_lex_state = 44},
  [1504] = {.lex_state = 229, .external_lex_state = 43},
  [1505] = {.lex_state = 539, .external_lex_state = 41},
  [1506] = {.lex_state = 229, .external_lex_state = 43},
  [1507] = {.lex_state = 229, .external_lex_state = 43},
  [1508] = {.lex_state = 538, .external_lex_state = 41},
  [1509] = {.lex_state = 198, .external_lex_state = 19},
  [1510] = {.lex_state = 227, .external_lex_state = 44},
  [1511] = {.lex_state = 539, .external_lex_state = 41},
  [1512] = {.lex_state = 539, .external_lex_state = 41},
  [1513] = {.lex_state = 229, .external_lex_state = 43},
  [1514] = {.lex_state = 229, .external_lex_state = 43},
  [1515] = {.lex_state = 229, .external_lex_state = 43},
  [1516] = {.lex_state = 227, .external_lex_state = 44},
  [1517] = {.lex_state = 229, .external_lex_state = 43},
  [1518] = {.lex_state = 227, .external_lex_state = 44},
  [1519] = {.lex_state = 539, .external_lex_state = 41},
  [1520] = {.lex_state = 229, .external_lex_state = 43},
  [1521] = {.lex_state = 527, .external_lex_state = 16},
  [1522] = {.lex_state = 186, .external_lex_state = 21},
  [1523] = {.lex_state = 539, .external_lex_state = 41},
  [1524] = {.lex_state = 233, .external_lex_state = 44},
  [1525] = {.lex_state = 229, .external_lex_state = 43},
  [1526] = {.lex_state = 233, .external_lex_state = 44},
  [1527] = {.lex_state = 227, .external_lex_state = 44},
  [1528] = {.lex_state = 538, .external_lex_state = 44},
  [1529] = {.lex_state = 539, .external_lex_state = 41},
  [1530] = {.lex_state = 229, .external_lex_state = 43},
  [1531] = {.lex_state = 538, .external_lex_state = 41},
  [1532] = {.lex_state = 229, .external_lex_state = 43},
  [1533] = {.lex_state = 229, .external_lex_state = 43},
  [1534] = {.lex_state = 195, .external_lex_state = 17},
  [1535] = {.lex_state = 186, .external_lex_state = 17},
  [1536] = {.lex_state = 538, .external_lex_state = 41},
  [1537] = {.lex_state = 538, .external_lex_state = 41},
  [1538] = {.lex_state = 538, .external_lex_state = 41},
  [1539] = {.lex_state = 198, .external_lex_state = 19},
  [1540] = {.lex_state = 538, .external_lex_state = 41},
  [1541] = {.lex_state = 539, .external_lex_state = 41},
  [1542] = {.lex_state = 186, .external_lex_state = 17},
  [1543] = {.lex_state = 527, .external_lex_state = 16},
  [1544] = {.lex_state = 539, .external_lex_state = 41},
  [1545] = {.lex_state = 539, .external_lex_state = 41},
  [1546] = {.lex_state = 229, .external_lex_state = 43},
  [1547] = {.lex_state = 538, .external_lex_state = 41},
  [1548] = {.lex_state = 538, .external_lex_state = 41},
  [1549] = {.lex_state = 538, .external_lex_state = 41},
  [1550] = {.lex_state = 539, .external_lex_state = 41},
  [1551] = {.lex_state = 195, .external_lex_state = 17},
  [1552] = {.lex_state = 527, .external_lex_state = 16},
  [1553] = {.lex_state = 539, .external_lex_state = 41},
  [1554] = {.lex_state = 539, .external_lex_state = 41},
  [1555] = {.lex_state = 539, .external_lex_state = 41},
  [1556] = {.lex_state = 195, .external_lex_state = 17},
  [1557] = {.lex_state = 538, .external_lex_state = 44},
  [1558] = {.lex_state = 538, .external_lex_state = 41},
  [1559] = {.lex_state = 539, .external_lex_state = 41},
  [1560] = {.lex_state = 539, .external_lex_state = 41},
  [1561] = {.lex_state = 195, .external_lex_state = 17},
  [1562] = {.lex_state = 198, .external_lex_state = 19},
  [1563] = {.lex_state = 538, .external_lex_state = 41},
  [1564] = {.lex_state = 539, .external_lex_state = 41},
  [1565] = {.lex_state = 539, .external_lex_state = 41},
  [1566] = {.lex_state = 539, .external_lex_state = 41},
  [1567] = {.lex_state = 198, .external_lex_state = 19},
  [1568] = {.lex_state = 225, .external_lex_state = 44},
  [1569] = {.lex_state = 225, .external_lex_state = 44},
  [1570] = {.lex_state = 539, .external_lex_state = 41},
  [1571] = {.lex_state = 538, .external_lex_state = 41},
  [1572] = {.lex_state = 217, .external_lex_state = 4},
  [1573] = {.lex_state = 527, .external_lex_state = 16},
  [1574] = {.lex_state = 186, .external_lex_state = 17},
  [1575] = {.lex_state = 539, .external_lex_state = 41},
  [1576] = {.lex_state = 261, .external_lex_state = 47},
  [1577] = {.lex_state = 227, .external_lex_state = 44},
  [1578] = {.lex_state = 229, .external_lex_state = 43},
  [1579] = {.lex_state = 229, .external_lex_state = 43},
  [1580] = {.lex_state = 538, .external_lex_state = 41},
  [1581] = {.lex_state = 229, .external_lex_state = 43},
  [1582] = {.lex_state = 538, .external_lex_state = 44},
  [1583] = {.lex_state = 539, .external_lex_state = 41},
  [1584] = {.lex_state = 227, .external_lex_state = 44},
  [1585] = {.lex_state = 227, .external_lex_state = 44},
  [1586] = {.lex_state = 539, .external_lex_state = 41},
  [1587] = {.lex_state = 539, .external_lex_state = 41},
  [1588] = {.lex_state = 225, .external_lex_state = 44},
  [1589] = {.lex_state = 538, .external_lex_state = 41},
  [1590] = {.lex_state = 539, .external_lex_state = 41},
  [1591] = {.lex_state = 225, .external_lex_state = 44},
  [1592] = {.lex_state = 227, .external_lex_state = 44},
  [1593] = {.lex_state = 225, .external_lex_state = 44},
  [1594] = {.lex_state = 227, .external_lex_state = 44},
  [1595] = {.lex_state = 538, .external_lex_state = 41},
  [1596] = {.lex_state = 539, .external_lex_state = 41},
  [1597] = {.lex_state = 538, .external_lex_state = 44},
  [1598] = {.lex_state = 538, .external_lex_state = 41},
  [1599] = {.lex_state = 539, .external_lex_state = 41},
  [1600] = {.lex_state = 225, .external_lex_state = 44},
  [1601] = {.lex_state = 539, .external_lex_state = 41},
  [1602] = {.lex_state = 538, .external_lex_state = 41},
  [1603] = {.lex_state = 273, .external_lex_state = 48},
  [1604] = {.lex_state = 195, .external_lex_state = 21},
  [1605] = {.lex_state = 539, .external_lex_state = 41},
  [1606] = {.lex_state = 198, .external_lex_state = 19},
  [1607] = {.lex_state = 195, .external_lex_state = 17},
  [1608] = {.lex_state = 198, .external_lex_state = 19},
  [1609] = {.lex_state = 225, .external_lex_state = 44},
  [1610] = {.lex_state = 225, .external_lex_state = 44},
  [1611] = {.lex_state = 527, .external_lex_state = 16},
  [1612] = {.lex_state = 225, .external_lex_state = 44},
  [1613] = {.lex_state = 527, .external_lex_state = 16},
  [1614] = {.lex_state = 539, .external_lex_state = 41},
  [1615] = {.lex_state = 261, .external_lex_state = 47},
  [1616] = {.lex_state = 227, .external_lex_state = 44},
  [1617] = {.lex_state = 227, .external_lex_state = 44},
  [1618] = {.lex_state = 198, .external_lex_state = 19},
  [1619] = {.lex_state = 538, .external_lex_state = 41},
  [1620] = {.lex_state = 261, .external_lex_state = 47},
  [1621] = {.lex_state = 539, .external_lex_state = 41},
  [1622] = {.lex_state = 527, .external_lex_state = 16},
  [1623] = {.lex_state = 538, .external_lex_state = 41},
  [1624] = {.lex_state = 225, .external_lex_state = 44},
  [1625] = {.lex_state = 539, .external_lex_state = 41},
  [1626] = {.lex_state = 539, .external_lex_state = 41},
  [1627] = {.lex_state = 198, .external_lex_state = 19},
  [1628] = {.lex_state = 539, .external_lex_state = 41},
  [1629] = {.lex_state = 195, .external_lex_state = 17},
  [1630] = {.lex_state = 539, .external_lex_state = 44},
  [1631] = {.lex_state = 186, .external_lex_state = 21},
  [1632] = {.lex_state = 539, .external_lex_state = 44},
  [1633] = {.lex_state = 539, .external_lex_state = 41},
  [1634] = {.lex_state = 195, .external_lex_state = 17},
  [1635] = {.lex_state = 195, .external_lex_state = 17},
  [1636] = {.lex_state = 261, .external_lex_state = 47},
  [1637] = {.lex_state = 539, .external_lex_state = 44},
  [1638] = {.lex_state = 539, .external_lex_state = 44},
  [1639] = {.lex_state = 195, .external_lex_state = 17},
  [1640] = {.lex_state = 233, .external_lex_state = 44},
  [1641] = {.lex_state = 229, .external_lex_state = 43},
  [1642] = {.lex_state = 229, .external_lex_state = 43},
  [1643] = {.lex_state = 198, .external_lex_state = 19},
  [1644] = {.lex_state = 538, .external_lex_state = 41},
  [1645] = {.lex_state = 233, .external_lex_state = 44},
  [1646] = {.lex_state = 189, .external_lex_state = 19},
  [1647] = {.lex_state = 538, .external_lex_state = 44},
  [1648] = {.lex_state = 273, .external_lex_state = 42},
  [1649] = {.lex_state = 539, .external_lex_state = 44},
  [1650] = {.lex_state = 538, .external_lex_state = 44},
  [1651] = {.lex_state = 539, .external_lex_state = 41},
  [1652] = {.lex_state = 538, .external_lex_state = 41},
  [1653] = {.lex_state = 217, .external_lex_state = 4},
  [1654] = {.lex_state = 213, .external_lex_state = 26},
  [1655] = {.lex_state = 225, .external_lex_state = 44},
  [1656] = {.lex_state = 527, .external_lex_state = 16},
  [1657] = {.lex_state = 225, .external_lex_state = 44},
  [1658] = {.lex_state = 195, .external_lex_state = 21},
  [1659] = {.lex_state = 233, .external_lex_state = 44},
  [1660] = {.lex_state = 539, .external_lex_state = 41},
  [1661] = {.lex_state = 539, .external_lex_state = 41},
  [1662] = {.lex_state = 539, .external_lex_state = 41},
  [1663] = {.lex_state = 538, .external_lex_state = 41},
  [1664] = {.lex_state = 538, .external_lex_state = 41},
  [1665] = {.lex_state = 538, .external_lex_state = 41},
  [1666] = {.lex_state = 225, .external_lex_state = 44},
  [1667] = {.lex_state = 195, .external_lex_state = 17},
  [1668] = {.lex_state = 539, .external_lex_state = 41},
  [1669] = {.lex_state = 227, .external_lex_state = 44},
  [1670] = {.lex_state = 213, .external_lex_state = 26},
  [1671] = {.lex_state = 225, .external_lex_state = 44},
  [1672] = {.lex_state = 538, .external_lex_state = 44},
  [1673] = {.lex_state = 527, .external_lex_state = 16},
  [1674] = {.lex_state = 227, .external_lex_state = 44},
  [1675] = {.lex_state = 539, .external_lex_state = 44},
  [1676] = {.lex_state = 227, .external_lex_state = 44},
  [1677] = {.lex_state = 225, .external_lex_state = 44},
  [1678] = {.lex_state = 227, .external_lex_state = 44},
  [1679] = {.lex_state = 261, .external_lex_state = 47},
  [1680] = {.lex_state = 539, .external_lex_state = 41},
  [1681] = {.lex_state = 538, .external_lex_state = 44},
  [1682] = {.lex_state = 539, .external_lex_state = 41},
  [1683] = {.lex_state = 539, .external_lex_state = 41},
  [1684] = {.lex_state = 539, .external_lex_state = 41},
  [1685] = {.lex_state = 261, .external_lex_state = 47},
  [1686] = {.lex_state = 539, .external_lex_state = 41},
  [1687] = {.lex_state = 539, .external_lex_state = 41},
  [1688] = {.lex_state = 539, .external_lex_state = 41},
  [1689] = {.lex_state = 538, .external_lex_state = 41},
  [1690] = {.lex_state = 539, .external_lex_state = 41},
  [1691] = {.lex_state = 538, .external_lex_state = 41},
  [1692] = {.lex_state = 538, .external_lex_state = 41},
  [1693] = {.lex_state = 538, .external_lex_state = 41},
  [1694] = {.lex_state = 539, .external_lex_state = 44},
  [1695] = {.lex_state = 539, .external_lex_state = 44},
  [1696] = {.lex_state = 538, .external_lex_state = 44},
  [1697] = {.lex_state = 538, .external_lex_state = 41},
  [1698] = {.lex_state = 261, .external_lex_state = 47},
  [1699] = {.lex_state = 189, .external_lex_state = 19},
  [1700] = {.lex_state = 539, .external_lex_state = 41},
  [1701] = {.lex_state = 227, .external_lex_state = 44},
  [1702] = {.lex_state = 539, .external_lex_state = 41},
  [1703] = {.lex_state = 195, .external_lex_state = 17},
  [1704] = {.lex_state = 539, .external_lex_state = 44},
  [1705] = {.lex_state = 538, .external_lex_state = 44},
  [1706] = {.lex_state = 198, .external_lex_state = 19},
  [1707] = {.lex_state = 198, .external_lex_state = 19},
  [1708] = {.lex_state = 539, .external_lex_state = 44},
  [1709] = {.lex_state = 227, .external_lex_state = 44},
  [1710] = {.lex_state = 539, .external_lex_state = 44},
  [1711] = {.lex_state = 539, .external_lex_state = 44},
  [1712] = {.lex_state = 527, .external_lex_state = 16},
  [1713] = {.lex_state = 225, .external_lex_state = 44},
  [1714] = {.lex_state = 538, .external_lex_state = 44},
  [1715] = {.lex_state = 217, .external_lex_state = 4},
  [1716] = {.lex_state = 527, .external_lex_state = 16},
  [1717] = {.lex_state = 217, .external_lex_state = 4},
  [1718] = {.lex_state = 225, .external_lex_state = 44},
  [1719] = {.lex_state = 217, .external_lex_state = 4},
  [1720] = {.lex_state = 189, .external_lex_state = 19},
  [1721] = {.lex_state = 217, .external_lex_state = 4},
  [1722] = {.lex_state = 538, .external_lex_state = 44},
  [1723] = {.lex_state = 538, .external_lex_state = 41},
  [1724] = {.lex_state = 527, .external_lex_state = 16},
  [1725] = {.lex_state = 233, .external_lex_state = 44},
  [1726] = {.lex_state = 538, .external_lex_state = 44},
  [1727] = {.lex_state = 538, .external_lex_state = 44},
  [1728] = {.lex_state = 538, .external_lex_state = 44},
  [1729] = {.lex_state = 198, .external_lex_state = 19},
  [1730] = {.lex_state = 261, .external_lex_state = 49},
  [1731] = {.lex_state = 198, .external_lex_state = 19},
  [1732] = {.lex_state = 538, .external_lex_state = 44},
  [1733] = {.lex_state = 538, .external_lex_state = 44},
  [1734] = {.lex_state = 538, .external_lex_state = 44},
  [1735] = {.lex_state = 189, .external_lex_state = 19},
  [1736] = {.lex_state = 530, .external_lex_state = 19},
  [1737] = {.lex_state = 538, .external_lex_state = 44},
  [1738] = {.lex_state = 533, .external_lex_state = 19},
  [1739] = {.lex_state = 533, .external_lex_state = 19},
  [1740] = {.lex_state = 533, .external_lex_state = 19},
  [1741] = {.lex_state = 533, .external_lex_state = 19},
  [1742] = {.lex_state = 533, .external_lex_state = 19},
  [1743] = {.lex_state = 530, .external_lex_state = 19},
  [1744] = {.lex_state = 533, .external_lex_state = 19},
  [1745] = {.lex_state = 217, .external_lex_state = 37},
  [1746] = {.lex_state = 538, .external_lex_state = 44},
  [1747] = {.lex_state = 198, .external_lex_state = 19},
  [1748] = {.lex_state = 189, .external_lex_state = 19},
  [1749] = {.lex_state = 261, .external_lex_state = 50},
  [1750] = {.lex_state = 217, .external_lex_state = 37},
  [1751] = {.lex_state = 533, .external_lex_state = 19},
  [1752] = {.lex_state = 217, .external_lex_state = 37},
  [1753] = {.lex_state = 538, .external_lex_state = 44},
  [1754] = {.lex_state = 539, .external_lex_state = 44},
  [1755] = {.lex_state = 539, .external_lex_state = 44},
  [1756] = {.lex_state = 539, .external_lex_state = 44},
  [1757] = {.lex_state = 261, .external_lex_state = 49},
  [1758] = {.lex_state = 217, .external_lex_state = 37},
  [1759] = {.lex_state = 189, .external_lex_state = 19},
  [1760] = {.lex_state = 233, .external_lex_state = 44},
  [1761] = {.lex_state = 233, .external_lex_state = 44},
  [1762] = {.lex_state = 233, .external_lex_state = 44},
  [1763] = {.lex_state = 233, .external_lex_state = 44},
  [1764] = {.lex_state = 233, .external_lex_state = 44},
  [1765] = {.lex_state = 233, .external_lex_state = 44},
  [1766] = {.lex_state = 539, .external_lex_state = 44},
  [1767] = {.lex_state = 198, .external_lex_state = 19},
  [1768] = {.lex_state = 539, .external_lex_state = 44},
  [1769] = {.lex_state = 539, .external_lex_state = 44},
  [1770] = {.lex_state = 233, .external_lex_state = 44},
  [1771] = {.lex_state = 539, .external_lex_state = 44},
  [1772] = {.lex_state = 233, .external_lex_state = 44},
  [1773] = {.lex_state = 233, .external_lex_state = 44},
  [1774] = {.lex_state = 539, .external_lex_state = 44},
  [1775] = {.lex_state = 233, .external_lex_state = 44},
  [1776] = {.lex_state = 198, .external_lex_state = 19},
  [1777] = {.lex_state = 539, .external_lex_state = 44},
  [1778] = {.lex_state = 261, .external_lex_state = 50},
  [1779] = {.lex_state = 539, .external_lex_state = 44},
  [1780] = {.lex_state = 539, .external_lex_state = 44},
  [1781] = {.lex_state = 233, .external_lex_state = 44},
  [1782] = {.lex_state = 233, .external_lex_state = 44},
  [1783] = {.lex_state = 539, .external_lex_state = 44},
  [1784] = {.lex_state = 217, .external_lex_state = 37},
  [1785] = {.lex_state = 539, .external_lex_state = 44},
  [1786] = {.lex_state = 539, .external_lex_state = 44},
  [1787] = {.lex_state = 217, .external_lex_state = 37},
  [1788] = {.lex_state = 538, .external_lex_state = 44},
  [1789] = {.lex_state = 198, .external_lex_state = 22},
  [1790] = {.lex_state = 539, .external_lex_state = 44},
  [1791] = {.lex_state = 261, .external_lex_state = 49},
  [1792] = {.lex_state = 538, .external_lex_state = 44},
  [1793] = {.lex_state = 538, .external_lex_state = 44},
  [1794] = {.lex_state = 217, .external_lex_state = 37},
  [1795] = {.lex_state = 217, .external_lex_state = 37},
  [1796] = {.lex_state = 538, .external_lex_state = 44},
  [1797] = {.lex_state = 300, .external_lex_state = 51},
  [1798] = {.lex_state = 538, .external_lex_state = 44},
  [1799] = {.lex_state = 300, .external_lex_state = 51},
  [1800] = {.lex_state = 300, .external_lex_state = 51},
  [1801] = {.lex_state = 533, .external_lex_state = 19},
  [1802] = {.lex_state = 233, .external_lex_state = 44},
  [1803] = {.lex_state = 201, .external_lex_state = 21},
  [1804] = {.lex_state = 538, .external_lex_state = 44},
  [1805] = {.lex_state = 538, .external_lex_state = 44},
  [1806] = {.lex_state = 261, .external_lex_state = 49},
  [1807] = {.lex_state = 233, .external_lex_state = 44},
  [1808] = {.lex_state = 296, .external_lex_state = 36},
  [1809] = {.lex_state = 233, .external_lex_state = 44},
  [1810] = {.lex_state = 538, .external_lex_state = 44},
  [1811] = {.lex_state = 298, .external_lex_state = 29},
  [1812] = {.lex_state = 233, .external_lex_state = 44},
  [1813] = {.lex_state = 538, .external_lex_state = 44},
  [1814] = {.lex_state = 189, .external_lex_state = 22},
  [1815] = {.lex_state = 539, .external_lex_state = 44},
  [1816] = {.lex_state = 530, .external_lex_state = 19},
  [1817] = {.lex_state = 539, .external_lex_state = 44},
  [1818] = {.lex_state = 539, .external_lex_state = 44},
  [1819] = {.lex_state = 539, .external_lex_state = 44},
  [1820] = {.lex_state = 539, .external_lex_state = 44},
  [1821] = {.lex_state = 217, .external_lex_state = 37},
  [1822] = {.lex_state = 217, .external_lex_state = 37},
  [1823] = {.lex_state = 217, .external_lex_state = 37},
  [1824] = {.lex_state = 261, .external_lex_state = 49},
  [1825] = {.lex_state = 233, .external_lex_state = 44},
  [1826] = {.lex_state = 533, .external_lex_state = 19},
  [1827] = {.lex_state = 233, .external_lex_state = 44},
  [1828] = {.lex_state = 533, .external_lex_state = 19},
  [1829] = {.lex_state = 533, .external_lex_state = 19},
  [1830] = {.lex_state = 261, .external_lex_state = 49},
  [1831] = {.lex_state = 296, .external_lex_state = 36},
  [1832] = {.lex_state = 540, .external_lex_state = 44},
  [1833] = {.lex_state = 539, .external_lex_state = 44},
  [1834] = {.lex_state = 233, .external_lex_state = 44},
  [1835] = {.lex_state = 189, .external_lex_state = 19},
  [1836] = {.lex_state = 538, .external_lex_state = 44},
  [1837] = {.lex_state = 538, .external_lex_state = 44},
  [1838] = {.lex_state = 538, .external_lex_state = 44},
  [1839] = {.lex_state = 540, .external_lex_state = 44},
  [1840] = {.lex_state = 539, .external_lex_state = 44},
  [1841] = {.lex_state = 539, .external_lex_state = 44},
  [1842] = {.lex_state = 261, .external_lex_state = 49},
  [1843] = {.lex_state = 540, .external_lex_state = 44},
  [1844] = {.lex_state = 198, .external_lex_state = 19},
  [1845] = {.lex_state = 540, .external_lex_state = 44},
  [1846] = {.lex_state = 540, .external_lex_state = 44},
  [1847] = {.lex_state = 538, .external_lex_state = 44},
  [1848] = {.lex_state = 538, .external_lex_state = 44},
  [1849] = {.lex_state = 538, .external_lex_state = 44},
  [1850] = {.lex_state = 261, .external_lex_state = 50},
  [1851] = {.lex_state = 539, .external_lex_state = 44},
  [1852] = {.lex_state = 538, .external_lex_state = 44},
  [1853] = {.lex_state = 538, .external_lex_state = 44},
  [1854] = {.lex_state = 538, .external_lex_state = 44},
  [1855] = {.lex_state = 538, .external_lex_state = 44},
  [1856] = {.lex_state = 538, .external_lex_state = 44},
  [1857] = {.lex_state = 538, .external_lex_state = 44},
  [1858] = {.lex_state = 538, .external_lex_state = 44},
  [1859] = {.lex_state = 538, .external_lex_state = 44},
  [1860] = {.lex_state = 189, .external_lex_state = 22},
  [1861] = {.lex_state = 539, .external_lex_state = 44},
  [1862] = {.lex_state = 539, .external_lex_state = 44},
  [1863] = {.lex_state = 538, .external_lex_state = 44},
  [1864] = {.lex_state = 539, .external_lex_state = 44},
  [1865] = {.lex_state = 539, .external_lex_state = 44},
  [1866] = {.lex_state = 539, .external_lex_state = 44},
  [1867] = {.lex_state = 261, .external_lex_state = 50},
  [1868] = {.lex_state = 296, .external_lex_state = 36},
  [1869] = {.lex_state = 261, .external_lex_state = 49},
  [1870] = {.lex_state = 539, .external_lex_state = 44},
  [1871] = {.lex_state = 261, .external_lex_state = 49},
  [1872] = {.lex_state = 198, .external_lex_state = 19},
  [1873] = {.lex_state = 530, .external_lex_state = 19},
  [1874] = {.lex_state = 538, .external_lex_state = 44},
  [1875] = {.lex_state = 538, .external_lex_state = 44},
  [1876] = {.lex_state = 538, .external_lex_state = 44},
  [1877] = {.lex_state = 539, .external_lex_state = 44},
  [1878] = {.lex_state = 538, .external_lex_state = 44},
  [1879] = {.lex_state = 540, .external_lex_state = 44},
  [1880] = {.lex_state = 538, .external_lex_state = 44},
  [1881] = {.lex_state = 538, .external_lex_state = 44},
  [1882] = {.lex_state = 298, .external_lex_state = 29},
  [1883] = {.lex_state = 198, .external_lex_state = 19},
  [1884] = {.lex_state = 261, .external_lex_state = 49},
  [1885] = {.lex_state = 195, .external_lex_state = 21},
  [1886] = {.lex_state = 201, .external_lex_state = 21},
  [1887] = {.lex_state = 530, .external_lex_state = 19},
  [1888] = {.lex_state = 539, .external_lex_state = 44},
  [1889] = {.lex_state = 540, .external_lex_state = 44},
  [1890] = {.lex_state = 186, .external_lex_state = 21},
  [1891] = {.lex_state = 539, .external_lex_state = 44},
  [1892] = {.lex_state = 198, .external_lex_state = 19},
  [1893] = {.lex_state = 198, .external_lex_state = 19},
  [1894] = {.lex_state = 540, .external_lex_state = 44},
  [1895] = {.lex_state = 533, .external_lex_state = 19},
  [1896] = {.lex_state = 539, .external_lex_state = 44},
  [1897] = {.lex_state = 300, .external_lex_state = 51},
  [1898] = {.lex_state = 533, .external_lex_state = 19},
  [1899] = {.lex_state = 198, .external_lex_state = 22},
  [1900] = {.lex_state = 533, .external_lex_state = 19},
  [1901] = {.lex_state = 300, .external_lex_state = 51},
  [1902] = {.lex_state = 538, .external_lex_state = 44},
  [1903] = {.lex_state = 538, .external_lex_state = 44},
  [1904] = {.lex_state = 300, .external_lex_state = 51},
  [1905] = {.lex_state = 261, .external_lex_state = 49},
  [1906] = {.lex_state = 261, .external_lex_state = 50},
  [1907] = {.lex_state = 298, .external_lex_state = 29},
  [1908] = {.lex_state = 540, .external_lex_state = 44},
  [1909] = {.lex_state = 198, .external_lex_state = 19},
  [1910] = {.lex_state = 540, .external_lex_state = 44},
  [1911] = {.lex_state = 540, .external_lex_state = 44},
  [1912] = {.lex_state = 300, .external_lex_state = 51},
  [1913] = {.lex_state = 261, .external_lex_state = 49},
  [1914] = {.lex_state = 538, .external_lex_state = 44},
  [1915] = {.lex_state = 538, .external_lex_state = 44},
  [1916] = {.lex_state = 540, .external_lex_state = 44},
  [1917] = {.lex_state = 539, .external_lex_state = 44},
  [1918] = {.lex_state = 539, .external_lex_state = 44},
  [1919] = {.lex_state = 539, .external_lex_state = 44},
  [1920] = {.lex_state = 539, .external_lex_state = 44},
  [1921] = {.lex_state = 539, .external_lex_state = 44},
  [1922] = {.lex_state = 539, .external_lex_state = 44},
  [1923] = {.lex_state = 539, .external_lex_state = 44},
  [1924] = {.lex_state = 539, .external_lex_state = 44},
  [1925] = {.lex_state = 261, .external_lex_state = 50},
  [1926] = {.lex_state = 261, .external_lex_state = 50},
  [1927] = {.lex_state = 261, .external_lex_state = 50},
  [1928] = {.lex_state = 539, .external_lex_state = 44},
  [1929] = {.lex_state = 530, .external_lex_state = 19},
  [1930] = {.lex_state = 533, .external_lex_state = 19},
  [1931] = {.lex_state = 533, .external_lex_state = 19},
  [1932] = {.lex_state = 261, .external_lex_state = 49},
  [1933] = {.lex_state = 261, .external_lex_state = 49},
  [1934] = {.lex_state = 261, .external_lex_state = 49},
  [1935] = {.lex_state = 261, .external_lex_state = 49},
  [1936] = {.lex_state = 261, .external_lex_state = 49},
  [1937] = {.lex_state = 261, .external_lex_state = 49},
  [1938] = {.lex_state = 261, .external_lex_state = 49},
  [1939] = {.lex_state = 533, .external_lex_state = 19},
  [1940] = {.lex_state = 261, .external_lex_state = 49},
  [1941] = {.lex_state = 261, .external_lex_state = 49},
  [1942] = {.lex_state = 261, .external_lex_state = 49},
  [1943] = {.lex_state = 261, .external_lex_state = 49},
  [1944] = {.lex_state = 261, .external_lex_state = 49},
  [1945] = {.lex_state = 540, .external_lex_state = 44},
  [1946] = {.lex_state = 540, .external_lex_state = 44},
  [1947] = {.lex_state = 530, .external_lex_state = 19},
  [1948] = {.lex_state = 261, .external_lex_state = 49},
  [1949] = {.lex_state = 261, .external_lex_state = 49},
  [1950] = {.lex_state = 261, .external_lex_state = 49},
  [1951] = {.lex_state = 530, .external_lex_state = 19},
  [1952] = {.lex_state = 533, .external_lex_state = 19},
  [1953] = {.lex_state = 533, .external_lex_state = 19},
  [1954] = {.lex_state = 300, .external_lex_state = 51},
  [1955] = {.lex_state = 300, .external_lex_state = 51},
  [1956] = {.lex_state = 300, .external_lex_state = 51},
  [1957] = {.lex_state = 533, .external_lex_state = 19},
  [1958] = {.lex_state = 533, .external_lex_state = 19},
  [1959] = {.lex_state = 533, .external_lex_state = 19},
  [1960] = {.lex_state = 300, .external_lex_state = 51},
  [1961] = {.lex_state = 530, .external_lex_state = 19},
  [1962] = {.lex_state = 300, .external_lex_state = 51},
  [1963] = {.lex_state = 261, .external_lex_state = 49},
  [1964] = {.lex_state = 530, .external_lex_state = 22},
  [1965] = {.lex_state = 300, .external_lex_state = 51},
  [1966] = {.lex_state = 530, .external_lex_state = 22},
  [1967] = {.lex_state = 261, .external_lex_state = 49},
  [1968] = {.lex_state = 540, .external_lex_state = 44},
  [1969] = {.lex_state = 261, .external_lex_state = 49},
  [1970] = {.lex_state = 261, .external_lex_state = 49},
  [1971] = {.lex_state = 300, .external_lex_state = 51},
  [1972] = {.lex_state = 540, .external_lex_state = 44},
  [1973] = {.lex_state = 300, .external_lex_state = 51},
  [1974] = {.lex_state = 533, .external_lex_state = 19},
  [1975] = {.lex_state = 261, .external_lex_state = 49},
  [1976] = {.lex_state = 261, .external_lex_state = 49},
  [1977] = {.lex_state = 300, .external_lex_state = 51},
  [1978] = {.lex_state = 261, .external_lex_state = 49},
  [1979] = {.lex_state = 540, .external_lex_state = 44},
  [1980] = {.lex_state = 261, .external_lex_state = 49},
  [1981] = {.lex_state = 261, .external_lex_state = 49},
  [1982] = {.lex_state = 261, .external_lex_state = 49},
  [1983] = {.lex_state = 261, .external_lex_state = 49},
  [1984] = {.lex_state = 261, .external_lex_state = 49},
  [1985] = {.lex_state = 261, .external_lex_state = 49},
  [1986] = {.lex_state = 540, .external_lex_state = 44},
  [1987] = {.lex_state = 261, .external_lex_state = 49},
  [1988] = {.lex_state = 261, .external_lex_state = 49},
  [1989] = {.lex_state = 261, .external_lex_state = 49},
  [1990] = {.lex_state = 261, .external_lex_state = 49},
  [1991] = {.lex_state = 261, .external_lex_state = 49},
  [1992] = {.lex_state = 300, .external_lex_state = 51},
  [1993] = {.lex_state = 540, .external_lex_state = 44},
  [1994] = {.lex_state = 533, .external_lex_state = 19},
  [1995] = {.lex_state = 300, .external_lex_state = 51},
  [1996] = {.lex_state = 540, .external_lex_state = 44},
  [1997] = {.lex_state = 540, .external_lex_state = 44},
  [1998] = {.lex_state = 261, .external_lex_state = 49},
  [1999] = {.lex_state = 261, .external_lex_state = 49},
  [2000] = {.lex_state = 261, .external_lex_state = 49},
  [2001] = {.lex_state = 261, .external_lex_state = 49},
  [2002] = {.lex_state = 300, .external_lex_state = 51},
  [2003] = {.lex_state = 261, .external_lex_state = 49},
  [2004] = {.lex_state = 540, .external_lex_state = 44},
  [2005] = {.lex_state = 540, .external_lex_state = 44},
  [2006] = {.lex_state = 540, .external_lex_state = 44},
  [2007] = {.lex_state = 261, .external_lex_state = 49},
  [2008] = {.lex_state = 261, .external_lex_state = 49},
  [2009] = {.lex_state = 261, .external_lex_state = 49},
  [2010] = {.lex_state = 261, .external_lex_state = 49},
  [2011] = {.lex_state = 261, .external_lex_state = 49},
  [2012] = {.lex_state = 540, .external_lex_state = 44},
  [2013] = {.lex_state = 261, .external_lex_state = 49},
  [2014] = {.lex_state = 261, .external_lex_state = 49},
  [2015] = {.lex_state = 530, .external_lex_state = 19},
  [2016] = {.lex_state = 261, .external_lex_state = 49},
  [2017] = {.lex_state = 533, .external_lex_state = 19},
  [2018] = {.lex_state = 261, .external_lex_state = 49},
  [2019] = {.lex_state = 533, .external_lex_state = 19},
  [2020] = {.lex_state = 533, .external_lex_state = 19},
  [2021] = {.lex_state = 261, .external_lex_state = 49},
  [2022] = {.lex_state = 261, .external_lex_state = 49},
  [2023] = {.lex_state = 261, .external_lex_state = 49},
  [2024] = {.lex_state = 533, .external_lex_state = 22},
  [2025] = {.lex_state = 540, .external_lex_state = 44},
  [2026] = {.lex_state = 533, .external_lex_state = 22},
  [2027] = {.lex_state = 261, .external_lex_state = 49},
  [2028] = {.lex_state = 261, .external_lex_state = 49},
  [2029] = {.lex_state = 261, .external_lex_state = 49},
  [2030] = {.lex_state = 261, .external_lex_state = 49},
  [2031] = {.lex_state = 261, .external_lex_state = 49},
  [2032] = {.lex_state = 540, .external_lex_state = 44},
  [2033] = {.lex_state = 300, .external_lex_state = 52},
  [2034] = {.lex_state = 261, .external_lex_state = 49},
  [2035] = {.lex_state = 261, .external_lex_state = 49},
  [2036] = {.lex_state = 261, .external_lex_state = 49},
  [2037] = {.lex_state = 300, .external_lex_state = 51},
  [2038] = {.lex_state = 261, .external_lex_state = 49},
  [2039] = {.lex_state = 261, .external_lex_state = 49},
  [2040] = {.lex_state = 261, .external_lex_state = 49},
  [2041] = {.lex_state = 261, .external_lex_state = 49},
  [2042] = {.lex_state = 261, .external_lex_state = 49},
  [2043] = {.lex_state = 261, .external_lex_state = 49},
  [2044] = {.lex_state = 261, .external_lex_state = 49},
  [2045] = {.lex_state = 261, .external_lex_state = 49},
  [2046] = {.lex_state = 300, .external_lex_state = 51},
  [2047] = {.lex_state = 261, .external_lex_state = 49},
  [2048] = {.lex_state = 261, .external_lex_state = 49},
  [2049] = {.lex_state = 261, .external_lex_state = 49},
  [2050] = {.lex_state = 261, .external_lex_state = 49},
  [2051] = {.lex_state = 261, .external_lex_state = 49},
  [2052] = {.lex_state = 261, .external_lex_state = 49},
  [2053] = {.lex_state = 261, .external_lex_state = 49},
  [2054] = {.lex_state = 261, .external_lex_state = 49},
  [2055] = {.lex_state = 533, .external_lex_state = 19},
  [2056] = {.lex_state = 261, .external_lex_state = 49},
  [2057] = {.lex_state = 300, .external_lex_state = 51},
  [2058] = {.lex_state = 540, .external_lex_state = 44},
  [2059] = {.lex_state = 261, .external_lex_state = 49},
  [2060] = {.lex_state = 533, .external_lex_state = 19},
  [2061] = {.lex_state = 261, .external_lex_state = 49},
  [2062] = {.lex_state = 261, .external_lex_state = 49},
  [2063] = {.lex_state = 261, .external_lex_state = 49},
  [2064] = {.lex_state = 261, .external_lex_state = 49},
  [2065] = {.lex_state = 300, .external_lex_state = 51},
  [2066] = {.lex_state = 261, .external_lex_state = 49},
  [2067] = {.lex_state = 540, .external_lex_state = 44},
  [2068] = {.lex_state = 540, .external_lex_state = 44},
  [2069] = {.lex_state = 261, .external_lex_state = 49},
  [2070] = {.lex_state = 533, .external_lex_state = 19},
  [2071] = {.lex_state = 533, .external_lex_state = 19},
  [2072] = {.lex_state = 261, .external_lex_state = 49},
  [2073] = {.lex_state = 540, .external_lex_state = 44},
  [2074] = {.lex_state = 540, .external_lex_state = 44},
  [2075] = {.lex_state = 300, .external_lex_state = 51},
  [2076] = {.lex_state = 533, .external_lex_state = 19},
  [2077] = {.lex_state = 300, .external_lex_state = 51},
  [2078] = {.lex_state = 533, .external_lex_state = 19},
  [2079] = {.lex_state = 533, .external_lex_state = 19},
  [2080] = {.lex_state = 540, .external_lex_state = 44},
  [2081] = {.lex_state = 540, .external_lex_state = 44},
  [2082] = {.lex_state = 533, .external_lex_state = 19},
  [2083] = {.lex_state = 533, .external_lex_state = 19},
  [2084] = {.lex_state = 533, .external_lex_state = 19},
  [2085] = {.lex_state = 533, .external_lex_state = 19},
  [2086] = {.lex_state = 540, .external_lex_state = 44},
  [2087] = {.lex_state = 261, .external_lex_state = 49},
  [2088] = {.lex_state = 533, .external_lex_state = 19},
  [2089] = {.lex_state = 261, .external_lex_state = 49},
  [2090] = {.lex_state = 261, .external_lex_state = 49},
  [2091] = {.lex_state = 540, .external_lex_state = 44},
  [2092] = {.lex_state = 261, .external_lex_state = 49},
  [2093] = {.lex_state = 300, .external_lex_state = 51},
  [2094] = {.lex_state = 198, .external_lex_state = 22},
  [2095] = {.lex_state = 540, .external_lex_state = 44},
  [2096] = {.lex_state = 533, .external_lex_state = 19},
  [2097] = {.lex_state = 533, .external_lex_state = 19},
  [2098] = {.lex_state = 533, .external_lex_state = 19},
  [2099] = {.lex_state = 533, .external_lex_state = 19},
  [2100] = {.lex_state = 540, .external_lex_state = 44},
  [2101] = {.lex_state = 533, .external_lex_state = 19},
  [2102] = {.lex_state = 533, .external_lex_state = 19},
  [2103] = {.lex_state = 533, .external_lex_state = 19},
  [2104] = {.lex_state = 533, .external_lex_state = 19},
  [2105] = {.lex_state = 540, .external_lex_state = 44},
  [2106] = {.lex_state = 533, .external_lex_state = 19},
  [2107] = {.lex_state = 533, .external_lex_state = 19},
  [2108] = {.lex_state = 540, .external_lex_state = 44},
  [2109] = {.lex_state = 540, .external_lex_state = 44},
  [2110] = {.lex_state = 533, .external_lex_state = 19},
  [2111] = {.lex_state = 540, .external_lex_state = 44},
  [2112] = {.lex_state = 533, .external_lex_state = 19},
  [2113] = {.lex_state = 247, .external_lex_state = 36},
  [2114] = {.lex_state = 540, .external_lex_state = 44},
  [2115] = {.lex_state = 247, .external_lex_state = 36},
  [2116] = {.lex_state = 540, .external_lex_state = 44},
  [2117] = {.lex_state = 533, .external_lex_state = 19},
  [2118] = {.lex_state = 207, .external_lex_state = 22},
  [2119] = {.lex_state = 261, .external_lex_state = 49},
  [2120] = {.lex_state = 533, .external_lex_state = 19},
  [2121] = {.lex_state = 540, .external_lex_state = 44},
  [2122] = {.lex_state = 533, .external_lex_state = 19},
  [2123] = {.lex_state = 296, .external_lex_state = 36},
  [2124] = {.lex_state = 533, .external_lex_state = 22},
  [2125] = {.lex_state = 300, .external_lex_state = 52},
  [2126] = {.lex_state = 300, .external_lex_state = 52},
  [2127] = {.lex_state = 300, .external_lex_state = 52},
  [2128] = {.lex_state = 533, .external_lex_state = 19},
  [2129] = {.lex_state = 261, .external_lex_state = 49},
  [2130] = {.lex_state = 261, .external_lex_state = 49},
  [2131] = {.lex_state = 261, .external_lex_state = 49},
  [2132] = {.lex_state = 261, .external_lex_state = 49},
  [2133] = {.lex_state = 533, .external_lex_state = 19},
  [2134] = {.lex_state = 261, .external_lex_state = 49},
  [2135] = {.lex_state = 300, .external_lex_state = 52},
  [2136] = {.lex_state = 533, .external_lex_state = 19},
  [2137] = {.lex_state = 261, .external_lex_state = 49},
  [2138] = {.lex_state = 296, .external_lex_state = 36},
  [2139] = {.lex_state = 261, .external_lex_state = 49},
  [2140] = {.lex_state = 533, .external_lex_state = 19},
  [2141] = {.lex_state = 533, .external_lex_state = 19},
  [2142] = {.lex_state = 533, .external_lex_state = 19},
  [2143] = {.lex_state = 533, .external_lex_state = 19},
  [2144] = {.lex_state = 533, .external_lex_state = 19},
  [2145] = {.lex_state = 261, .external_lex_state = 49},
  [2146] = {.lex_state = 533, .external_lex_state = 19},
  [2147] = {.lex_state = 207, .external_lex_state = 22},
  [2148] = {.lex_state = 533, .external_lex_state = 19},
  [2149] = {.lex_state = 533, .external_lex_state = 19},
  [2150] = {.lex_state = 533, .external_lex_state = 19},
  [2151] = {.lex_state = 533, .external_lex_state = 19},
  [2152] = {.lex_state = 300, .external_lex_state = 52},
  [2153] = {.lex_state = 533, .external_lex_state = 19},
  [2154] = {.lex_state = 300, .external_lex_state = 52},
  [2155] = {.lex_state = 300, .external_lex_state = 52},
  [2156] = {.lex_state = 533, .external_lex_state = 19},
  [2157] = {.lex_state = 533, .external_lex_state = 19},
  [2158] = {.lex_state = 530, .external_lex_state = 19},
  [2159] = {.lex_state = 533, .external_lex_state = 19},
  [2160] = {.lex_state = 201, .external_lex_state = 21},
  [2161] = {.lex_state = 533, .external_lex_state = 22},
  [2162] = {.lex_state = 540, .external_lex_state = 44},
  [2163] = {.lex_state = 533, .external_lex_state = 19},
  [2164] = {.lex_state = 533, .external_lex_state = 19},
  [2165] = {.lex_state = 533, .external_lex_state = 19},
  [2166] = {.lex_state = 261, .external_lex_state = 49},
  [2167] = {.lex_state = 533, .external_lex_state = 19},
  [2168] = {.lex_state = 533, .external_lex_state = 19},
  [2169] = {.lex_state = 533, .external_lex_state = 19},
  [2170] = {.lex_state = 533, .external_lex_state = 19},
  [2171] = {.lex_state = 533, .external_lex_state = 19},
  [2172] = {.lex_state = 533, .external_lex_state = 19},
  [2173] = {.lex_state = 533, .external_lex_state = 19},
  [2174] = {.lex_state = 261, .external_lex_state = 49},
  [2175] = {.lex_state = 533, .external_lex_state = 19},
  [2176] = {.lex_state = 533, .external_lex_state = 19},
  [2177] = {.lex_state = 533, .external_lex_state = 19},
  [2178] = {.lex_state = 533, .external_lex_state = 19},
  [2179] = {.lex_state = 533, .external_lex_state = 19},
  [2180] = {.lex_state = 533, .external_lex_state = 19},
  [2181] = {.lex_state = 533, .external_lex_state = 19},
  [2182] = {.lex_state = 189, .external_lex_state = 22},
  [2183] = {.lex_state = 533, .external_lex_state = 19},
  [2184] = {.lex_state = 261, .external_lex_state = 49},
  [2185] = {.lex_state = 533, .external_lex_state = 19},
  [2186] = {.lex_state = 533, .external_lex_state = 19},
  [2187] = {.lex_state = 533, .external_lex_state = 19},
  [2188] = {.lex_state = 261, .external_lex_state = 49},
  [2189] = {.lex_state = 533, .external_lex_state = 19},
  [2190] = {.lex_state = 530, .external_lex_state = 22},
  [2191] = {.lex_state = 533, .external_lex_state = 19},
  [2192] = {.lex_state = 261, .external_lex_state = 49},
  [2193] = {.lex_state = 533, .external_lex_state = 19},
  [2194] = {.lex_state = 533, .external_lex_state = 19},
  [2195] = {.lex_state = 533, .external_lex_state = 19},
  [2196] = {.lex_state = 533, .external_lex_state = 19},
  [2197] = {.lex_state = 261, .external_lex_state = 49},
  [2198] = {.lex_state = 261, .external_lex_state = 49},
  [2199] = {.lex_state = 300, .external_lex_state = 52},
  [2200] = {.lex_state = 533, .external_lex_state = 19},
  [2201] = {.lex_state = 533, .external_lex_state = 19},
  [2202] = {.lex_state = 261, .external_lex_state = 49},
  [2203] = {.lex_state = 300, .external_lex_state = 52},
  [2204] = {.lex_state = 300, .external_lex_state = 52},
  [2205] = {.lex_state = 261, .external_lex_state = 49},
  [2206] = {.lex_state = 533, .external_lex_state = 19},
  [2207] = {.lex_state = 533, .external_lex_state = 19},
  [2208] = {.lex_state = 261, .external_lex_state = 49},
  [2209] = {.lex_state = 261, .external_lex_state = 49},
  [2210] = {.lex_state = 540, .external_lex_state = 44},
  [2211] = {.lex_state = 300, .external_lex_state = 52},
  [2212] = {.lex_state = 533, .external_lex_state = 19},
  [2213] = {.lex_state = 261, .external_lex_state = 49},
  [2214] = {.lex_state = 540, .external_lex_state = 44},
  [2215] = {.lex_state = 530, .external_lex_state = 19},
  [2216] = {.lex_state = 540, .external_lex_state = 44},
  [2217] = {.lex_state = 533, .external_lex_state = 19},
  [2218] = {.lex_state = 540, .external_lex_state = 44},
  [2219] = {.lex_state = 533, .external_lex_state = 19},
  [2220] = {.lex_state = 533, .external_lex_state = 19},
  [2221] = {.lex_state = 533, .external_lex_state = 19},
  [2222] = {.lex_state = 540, .external_lex_state = 44},
  [2223] = {.lex_state = 533, .external_lex_state = 19},
  [2224] = {.lex_state = 540, .external_lex_state = 44},
  [2225] = {.lex_state = 261, .external_lex_state = 49},
  [2226] = {.lex_state = 261, .external_lex_state = 49},
  [2227] = {.lex_state = 533, .external_lex_state = 19},
  [2228] = {.lex_state = 533, .external_lex_state = 19},
  [2229] = {.lex_state = 261, .external_lex_state = 49},
  [2230] = {.lex_state = 261, .external_lex_state = 49},
  [2231] = {.lex_state = 261, .external_lex_state = 49},
  [2232] = {.lex_state = 261, .external_lex_state = 49},
  [2233] = {.lex_state = 261, .external_lex_state = 49},
  [2234] = {.lex_state = 261, .external_lex_state = 49},
  [2235] = {.lex_state = 261, .external_lex_state = 49},
  [2236] = {.lex_state = 533, .external_lex_state = 19},
  [2237] = {.lex_state = 533, .external_lex_state = 19},
  [2238] = {.lex_state = 261, .external_lex_state = 49},
  [2239] = {.lex_state = 261, .external_lex_state = 49},
  [2240] = {.lex_state = 261, .external_lex_state = 49},
  [2241] = {.lex_state = 261, .external_lex_state = 49},
  [2242] = {.lex_state = 261, .external_lex_state = 49},
  [2243] = {.lex_state = 261, .external_lex_state = 49},
  [2244] = {.lex_state = 261, .external_lex_state = 49},
  [2245] = {.lex_state = 261, .external_lex_state = 49},
  [2246] = {.lex_state = 261, .external_lex_state = 49},
  [2247] = {.lex_state = 530, .external_lex_state = 19},
  [2248] = {.lex_state = 261, .external_lex_state = 49},
  [2249] = {.lex_state = 261, .external_lex_state = 49},
  [2250] = {.lex_state = 261, .external_lex_state = 49},
  [2251] = {.lex_state = 261, .external_lex_state = 49},
  [2252] = {.lex_state = 261, .external_lex_state = 49},
  [2253] = {.lex_state = 530, .external_lex_state = 19},
  [2254] = {.lex_state = 530, .external_lex_state = 22},
  [2255] = {.lex_state = 533, .external_lex_state = 19},
  [2256] = {.lex_state = 261, .external_lex_state = 49},
  [2257] = {.lex_state = 540, .external_lex_state = 44},
  [2258] = {.lex_state = 261, .external_lex_state = 49},
  [2259] = {.lex_state = 533, .external_lex_state = 19},
  [2260] = {.lex_state = 540, .external_lex_state = 44},
  [2261] = {.lex_state = 261, .external_lex_state = 49},
  [2262] = {.lex_state = 533, .external_lex_state = 19},
  [2263] = {.lex_state = 540, .external_lex_state = 44},
  [2264] = {.lex_state = 533, .external_lex_state = 19},
  [2265] = {.lex_state = 533, .external_lex_state = 19},
  [2266] = {.lex_state = 261, .external_lex_state = 49},
  [2267] = {.lex_state = 261, .external_lex_state = 49},
  [2268] = {.lex_state = 300, .external_lex_state = 52},
  [2269] = {.lex_state = 300, .external_lex_state = 52},
  [2270] = {.lex_state = 534, .external_lex_state = 22},
  [2271] = {.lex_state = 303, .external_lex_state = 53},
  [2272] = {.lex_state = 533, .external_lex_state = 19},
  [2273] = {.lex_state = 533, .external_lex_state = 19},
  [2274] = {.lex_state = 278, .external_lex_state = 54},
  [2275] = {.lex_state = 278, .external_lex_state = 54},
  [2276] = {.lex_state = 278, .external_lex_state = 54},
  [2277] = {.lex_state = 278, .external_lex_state = 54},
  [2278] = {.lex_state = 303, .external_lex_state = 53},
  [2279] = {.lex_state = 227, .external_lex_state = 41},
  [2280] = {.lex_state = 530, .external_lex_state = 22},
  [2281] = {.lex_state = 534, .external_lex_state = 22},
  [2282] = {.lex_state = 227, .external_lex_state = 41},
  [2283] = {.lex_state = 227, .external_lex_state = 41},
  [2284] = {.lex_state = 303, .external_lex_state = 53},
  [2285] = {.lex_state = 303, .external_lex_state = 53},
  [2286] = {.lex_state = 533, .external_lex_state = 19},
  [2287] = {.lex_state = 300, .external_lex_state = 52},
  [2288] = {.lex_state = 227, .external_lex_state = 41},
  [2289] = {.lex_state = 300, .external_lex_state = 52},
  [2290] = {.lex_state = 227, .external_lex_state = 41},
  [2291] = {.lex_state = 227, .external_lex_state = 41},
  [2292] = {.lex_state = 278, .external_lex_state = 54},
  [2293] = {.lex_state = 303, .external_lex_state = 53},
  [2294] = {.lex_state = 300, .external_lex_state = 52},
  [2295] = {.lex_state = 300, .external_lex_state = 52},
  [2296] = {.lex_state = 303, .external_lex_state = 53},
  [2297] = {.lex_state = 227, .external_lex_state = 41},
  [2298] = {.lex_state = 303, .external_lex_state = 53},
  [2299] = {.lex_state = 300, .external_lex_state = 52},
  [2300] = {.lex_state = 303, .external_lex_state = 53},
  [2301] = {.lex_state = 300, .external_lex_state = 52},
  [2302] = {.lex_state = 227, .external_lex_state = 41},
  [2303] = {.lex_state = 303, .external_lex_state = 53},
  [2304] = {.lex_state = 278, .external_lex_state = 54},
  [2305] = {.lex_state = 207, .external_lex_state = 22},
  [2306] = {.lex_state = 533, .external_lex_state = 22},
  [2307] = {.lex_state = 278, .external_lex_state = 54},
  [2308] = {.lex_state = 303, .external_lex_state = 53},
  [2309] = {.lex_state = 303, .external_lex_state = 53},
  [2310] = {.lex_state = 530, .external_lex_state = 22},
  [2311] = {.lex_state = 534, .external_lex_state = 22},
  [2312] = {.lex_state = 278, .external_lex_state = 54},
  [2313] = {.lex_state = 300, .external_lex_state = 52},
  [2314] = {.lex_state = 303, .external_lex_state = 53},
  [2315] = {.lex_state = 278, .external_lex_state = 54},
  [2316] = {.lex_state = 534, .external_lex_state = 22},
  [2317] = {.lex_state = 278, .external_lex_state = 54},
  [2318] = {.lex_state = 300, .external_lex_state = 52},
  [2319] = {.lex_state = 278, .external_lex_state = 54},
  [2320] = {.lex_state = 300, .external_lex_state = 52},
  [2321] = {.lex_state = 278, .external_lex_state = 54},
  [2322] = {.lex_state = 278, .external_lex_state = 54},
  [2323] = {.lex_state = 225, .external_lex_state = 41},
  [2324] = {.lex_state = 225, .external_lex_state = 41},
  [2325] = {.lex_state = 227, .external_lex_state = 41},
  [2326] = {.lex_state = 300, .external_lex_state = 52},
  [2327] = {.lex_state = 227, .external_lex_state = 41},
  [2328] = {.lex_state = 227, .external_lex_state = 41},
  [2329] = {.lex_state = 225, .external_lex_state = 41},
  [2330] = {.lex_state = 225, .external_lex_state = 41},
  [2331] = {.lex_state = 225, .external_lex_state = 41},
  [2332] = {.lex_state = 299, .external_lex_state = 23},
  [2333] = {.lex_state = 300, .external_lex_state = 52},
  [2334] = {.lex_state = 300, .external_lex_state = 52},
  [2335] = {.lex_state = 300, .external_lex_state = 52},
  [2336] = {.lex_state = 225, .external_lex_state = 41},
  [2337] = {.lex_state = 227, .external_lex_state = 41},
  [2338] = {.lex_state = 533, .external_lex_state = 22},
  [2339] = {.lex_state = 300, .external_lex_state = 52},
  [2340] = {.lex_state = 227, .external_lex_state = 41},
  [2341] = {.lex_state = 300, .external_lex_state = 52},
  [2342] = {.lex_state = 225, .external_lex_state = 41},
  [2343] = {.lex_state = 300, .external_lex_state = 52},
  [2344] = {.lex_state = 300, .external_lex_state = 52},
  [2345] = {.lex_state = 299, .external_lex_state = 23},
  [2346] = {.lex_state = 303, .external_lex_state = 53},
  [2347] = {.lex_state = 278, .external_lex_state = 54},
  [2348] = {.lex_state = 303, .external_lex_state = 53},
  [2349] = {.lex_state = 235, .external_lex_state = 55},
  [2350] = {.lex_state = 301, .external_lex_state = 56},
  [2351] = {.lex_state = 299, .external_lex_state = 23},
  [2352] = {.lex_state = 299, .external_lex_state = 25},
  [2353] = {.lex_state = 235, .external_lex_state = 55},
  [2354] = {.lex_state = 278, .external_lex_state = 54},
  [2355] = {.lex_state = 278, .external_lex_state = 54},
  [2356] = {.lex_state = 303, .external_lex_state = 57},
  [2357] = {.lex_state = 301, .external_lex_state = 56},
  [2358] = {.lex_state = 235, .external_lex_state = 55},
  [2359] = {.lex_state = 301, .external_lex_state = 56},
  [2360] = {.lex_state = 235, .external_lex_state = 55},
  [2361] = {.lex_state = 235, .external_lex_state = 55},
  [2362] = {.lex_state = 303, .external_lex_state = 53},
  [2363] = {.lex_state = 303, .external_lex_state = 53},
  [2364] = {.lex_state = 235, .external_lex_state = 55},
  [2365] = {.lex_state = 225, .external_lex_state = 44},
  [2366] = {.lex_state = 299, .external_lex_state = 25},
  [2367] = {.lex_state = 225, .external_lex_state = 44},
  [2368] = {.lex_state = 299, .external_lex_state = 23},
  [2369] = {.lex_state = 301, .external_lex_state = 56},
  [2370] = {.lex_state = 299, .external_lex_state = 25},
  [2371] = {.lex_state = 301, .external_lex_state = 56},
  [2372] = {.lex_state = 235, .external_lex_state = 55},
  [2373] = {.lex_state = 309, .external_lex_state = 56},
  [2374] = {.lex_state = 276, .external_lex_state = 45},
  [2375] = {.lex_state = 303, .external_lex_state = 53},
  [2376] = {.lex_state = 235, .external_lex_state = 55},
  [2377] = {.lex_state = 235, .external_lex_state = 55},
  [2378] = {.lex_state = 235, .external_lex_state = 55},
  [2379] = {.lex_state = 235, .external_lex_state = 55},
  [2380] = {.lex_state = 303, .external_lex_state = 53},
  [2381] = {.lex_state = 534, .external_lex_state = 22},
  [2382] = {.lex_state = 276, .external_lex_state = 45},
  [2383] = {.lex_state = 278, .external_lex_state = 54},
  [2384] = {.lex_state = 303, .external_lex_state = 53},
  [2385] = {.lex_state = 225, .external_lex_state = 44},
  [2386] = {.lex_state = 225, .external_lex_state = 44},
  [2387] = {.lex_state = 278, .external_lex_state = 54},
  [2388] = {.lex_state = 303, .external_lex_state = 53},
  [2389] = {.lex_state = 227, .external_lex_state = 44},
  [2390] = {.lex_state = 303, .external_lex_state = 53},
  [2391] = {.lex_state = 227, .external_lex_state = 44},
  [2392] = {.lex_state = 278, .external_lex_state = 54},
  [2393] = {.lex_state = 278, .external_lex_state = 54},
  [2394] = {.lex_state = 299, .external_lex_state = 25},
  [2395] = {.lex_state = 309, .external_lex_state = 56},
  [2396] = {.lex_state = 301, .external_lex_state = 56},
  [2397] = {.lex_state = 278, .external_lex_state = 54},
  [2398] = {.lex_state = 303, .external_lex_state = 53},
  [2399] = {.lex_state = 303, .external_lex_state = 53},
  [2400] = {.lex_state = 303, .external_lex_state = 53},
  [2401] = {.lex_state = 303, .external_lex_state = 53},
  [2402] = {.lex_state = 303, .external_lex_state = 53},
  [2403] = {.lex_state = 225, .external_lex_state = 44},
  [2404] = {.lex_state = 227, .external_lex_state = 44},
  [2405] = {.lex_state = 309, .external_lex_state = 56},
  [2406] = {.lex_state = 309, .external_lex_state = 56},
  [2407] = {.lex_state = 309, .external_lex_state = 56},
  [2408] = {.lex_state = 281, .external_lex_state = 54},
  [2409] = {.lex_state = 281, .external_lex_state = 54},
  [2410] = {.lex_state = 235, .external_lex_state = 55},
  [2411] = {.lex_state = 235, .external_lex_state = 55},
  [2412] = {.lex_state = 303, .external_lex_state = 53},
  [2413] = {.lex_state = 301, .external_lex_state = 56},
  [2414] = {.lex_state = 235, .external_lex_state = 55},
  [2415] = {.lex_state = 301, .external_lex_state = 56},
  [2416] = {.lex_state = 235, .external_lex_state = 55},
  [2417] = {.lex_state = 299, .external_lex_state = 25},
  [2418] = {.lex_state = 534, .external_lex_state = 22},
  [2419] = {.lex_state = 309, .external_lex_state = 56},
  [2420] = {.lex_state = 303, .external_lex_state = 53},
  [2421] = {.lex_state = 303, .external_lex_state = 53},
  [2422] = {.lex_state = 227, .external_lex_state = 44},
  [2423] = {.lex_state = 303, .external_lex_state = 53},
  [2424] = {.lex_state = 303, .external_lex_state = 57},
  [2425] = {.lex_state = 303, .external_lex_state = 57},
  [2426] = {.lex_state = 309, .external_lex_state = 56},
  [2427] = {.lex_state = 301, .external_lex_state = 56},
  [2428] = {.lex_state = 235, .external_lex_state = 55},
  [2429] = {.lex_state = 301, .external_lex_state = 56},
  [2430] = {.lex_state = 303, .external_lex_state = 53},
  [2431] = {.lex_state = 227, .external_lex_state = 44},
  [2432] = {.lex_state = 301, .external_lex_state = 56},
  [2433] = {.lex_state = 277, .external_lex_state = 54},
  [2434] = {.lex_state = 301, .external_lex_state = 56},
  [2435] = {.lex_state = 309, .external_lex_state = 56},
  [2436] = {.lex_state = 301, .external_lex_state = 56},
  [2437] = {.lex_state = 303, .external_lex_state = 57},
  [2438] = {.lex_state = 303, .external_lex_state = 57},
  [2439] = {.lex_state = 278, .external_lex_state = 54},
  [2440] = {.lex_state = 301, .external_lex_state = 56},
  [2441] = {.lex_state = 303, .external_lex_state = 57},
  [2442] = {.lex_state = 278, .external_lex_state = 54},
  [2443] = {.lex_state = 271, .external_lex_state = 58},
  [2444] = {.lex_state = 309, .external_lex_state = 56},
  [2445] = {.lex_state = 273, .external_lex_state = 45},
  [2446] = {.lex_state = 303, .external_lex_state = 57},
  [2447] = {.lex_state = 303, .external_lex_state = 57},
  [2448] = {.lex_state = 303, .external_lex_state = 57},
  [2449] = {.lex_state = 303, .external_lex_state = 57},
  [2450] = {.lex_state = 303, .external_lex_state = 57},
  [2451] = {.lex_state = 276, .external_lex_state = 45},
  [2452] = {.lex_state = 309, .external_lex_state = 56},
  [2453] = {.lex_state = 309, .external_lex_state = 59},
  [2454] = {.lex_state = 309, .external_lex_state = 56},
  [2455] = {.lex_state = 276, .external_lex_state = 45},
  [2456] = {.lex_state = 278, .external_lex_state = 54},
  [2457] = {.lex_state = 302, .external_lex_state = 31},
  [2458] = {.lex_state = 301, .external_lex_state = 59},
  [2459] = {.lex_state = 278, .external_lex_state = 54},
  [2460] = {.lex_state = 304, .external_lex_state = 57},
  [2461] = {.lex_state = 276, .external_lex_state = 45},
  [2462] = {.lex_state = 303, .external_lex_state = 57},
  [2463] = {.lex_state = 273, .external_lex_state = 45},
  [2464] = {.lex_state = 276, .external_lex_state = 54},
  [2465] = {.lex_state = 309, .external_lex_state = 56},
  [2466] = {.lex_state = 278, .external_lex_state = 54},
  [2467] = {.lex_state = 309, .external_lex_state = 56},
  [2468] = {.lex_state = 309, .external_lex_state = 56},
  [2469] = {.lex_state = 309, .external_lex_state = 56},
  [2470] = {.lex_state = 233, .external_lex_state = 44},
  [2471] = {.lex_state = 233, .external_lex_state = 44},
  [2472] = {.lex_state = 273, .external_lex_state = 45},
  [2473] = {.lex_state = 278, .external_lex_state = 54},
  [2474] = {.lex_state = 273, .external_lex_state = 45},
  [2475] = {.lex_state = 276, .external_lex_state = 54},
  [2476] = {.lex_state = 302, .external_lex_state = 31},
  [2477] = {.lex_state = 309, .external_lex_state = 56},
  [2478] = {.lex_state = 303, .external_lex_state = 57},
  [2479] = {.lex_state = 301, .external_lex_state = 56},
  [2480] = {.lex_state = 233, .external_lex_state = 44},
  [2481] = {.lex_state = 301, .external_lex_state = 56},
  [2482] = {.lex_state = 303, .external_lex_state = 57},
  [2483] = {.lex_state = 301, .external_lex_state = 59},
  [2484] = {.lex_state = 281, .external_lex_state = 42},
  [2485] = {.lex_state = 271, .external_lex_state = 58},
  [2486] = {.lex_state = 304, .external_lex_state = 57},
  [2487] = {.lex_state = 301, .external_lex_state = 59},
  [2488] = {.lex_state = 304, .external_lex_state = 57},
  [2489] = {.lex_state = 304, .external_lex_state = 57},
  [2490] = {.lex_state = 304, .external_lex_state = 57},
  [2491] = {.lex_state = 276, .external_lex_state = 54},
  [2492] = {.lex_state = 299, .external_lex_state = 25},
  [2493] = {.lex_state = 301, .external_lex_state = 59},
  [2494] = {.lex_state = 304, .external_lex_state = 57},
  [2495] = {.lex_state = 278, .external_lex_state = 54},
  [2496] = {.lex_state = 303, .external_lex_state = 57},
  [2497] = {.lex_state = 276, .external_lex_state = 45},
  [2498] = {.lex_state = 301, .external_lex_state = 56},
  [2499] = {.lex_state = 309, .external_lex_state = 56},
  [2500] = {.lex_state = 276, .external_lex_state = 54},
  [2501] = {.lex_state = 273, .external_lex_state = 45},
  [2502] = {.lex_state = 301, .external_lex_state = 56},
  [2503] = {.lex_state = 278, .external_lex_state = 54},
  [2504] = {.lex_state = 303, .external_lex_state = 57},
  [2505] = {.lex_state = 271, .external_lex_state = 58},
  [2506] = {.lex_state = 276, .external_lex_state = 54},
  [2507] = {.lex_state = 301, .external_lex_state = 56},
  [2508] = {.lex_state = 301, .external_lex_state = 56},
  [2509] = {.lex_state = 276, .external_lex_state = 54},
  [2510] = {.lex_state = 301, .external_lex_state = 56},
  [2511] = {.lex_state = 276, .external_lex_state = 45},
  [2512] = {.lex_state = 301, .external_lex_state = 56},
  [2513] = {.lex_state = 301, .external_lex_state = 56},
  [2514] = {.lex_state = 301, .external_lex_state = 56},
  [2515] = {.lex_state = 301, .external_lex_state = 56},
  [2516] = {.lex_state = 301, .external_lex_state = 56},
  [2517] = {.lex_state = 301, .external_lex_state = 56},
  [2518] = {.lex_state = 301, .external_lex_state = 56},
  [2519] = {.lex_state = 271, .external_lex_state = 58},
  [2520] = {.lex_state = 301, .external_lex_state = 56},
  [2521] = {.lex_state = 276, .external_lex_state = 45},
  [2522] = {.lex_state = 299, .external_lex_state = 25},
  [2523] = {.lex_state = 303, .external_lex_state = 57},
  [2524] = {.lex_state = 271, .external_lex_state = 58},
  [2525] = {.lex_state = 301, .external_lex_state = 56},
  [2526] = {.lex_state = 309, .external_lex_state = 56},
  [2527] = {.lex_state = 309, .external_lex_state = 56},
  [2528] = {.lex_state = 309, .external_lex_state = 56},
  [2529] = {.lex_state = 271, .external_lex_state = 58},
  [2530] = {.lex_state = 271, .external_lex_state = 58},
  [2531] = {.lex_state = 271, .external_lex_state = 58},
  [2532] = {.lex_state = 302, .external_lex_state = 31},
  [2533] = {.lex_state = 276, .external_lex_state = 45},
  [2534] = {.lex_state = 299, .external_lex_state = 25},
  [2535] = {.lex_state = 301, .external_lex_state = 56},
  [2536] = {.lex_state = 303, .external_lex_state = 57},
  [2537] = {.lex_state = 281, .external_lex_state = 54},
  [2538] = {.lex_state = 309, .external_lex_state = 59},
  [2539] = {.lex_state = 309, .external_lex_state = 59},
  [2540] = {.lex_state = 276, .external_lex_state = 45},
  [2541] = {.lex_state = 302, .external_lex_state = 31},
  [2542] = {.lex_state = 301, .external_lex_state = 56},
  [2543] = {.lex_state = 301, .external_lex_state = 59},
  [2544] = {.lex_state = 309, .external_lex_state = 56},
  [2545] = {.lex_state = 309, .external_lex_state = 56},
  [2546] = {.lex_state = 276, .external_lex_state = 45},
  [2547] = {.lex_state = 273, .external_lex_state = 45},
  [2548] = {.lex_state = 309, .external_lex_state = 59},
  [2549] = {.lex_state = 276, .external_lex_state = 45},
  [2550] = {.lex_state = 301, .external_lex_state = 59},
  [2551] = {.lex_state = 276, .external_lex_state = 45},
  [2552] = {.lex_state = 273, .external_lex_state = 45},
  [2553] = {.lex_state = 309, .external_lex_state = 59},
  [2554] = {.lex_state = 301, .external_lex_state = 56},
  [2555] = {.lex_state = 276, .external_lex_state = 45},
  [2556] = {.lex_state = 299, .external_lex_state = 25},
  [2557] = {.lex_state = 281, .external_lex_state = 42},
  [2558] = {.lex_state = 299, .external_lex_state = 25},
  [2559] = {.lex_state = 271, .external_lex_state = 58},
  [2560] = {.lex_state = 303, .external_lex_state = 57},
  [2561] = {.lex_state = 273, .external_lex_state = 45},
  [2562] = {.lex_state = 299, .external_lex_state = 25},
  [2563] = {.lex_state = 276, .external_lex_state = 45},
  [2564] = {.lex_state = 309, .external_lex_state = 59},
  [2565] = {.lex_state = 309, .external_lex_state = 56},
  [2566] = {.lex_state = 309, .external_lex_state = 56},
  [2567] = {.lex_state = 303, .external_lex_state = 57},
  [2568] = {.lex_state = 302, .external_lex_state = 31},
  [2569] = {.lex_state = 302, .external_lex_state = 31},
  [2570] = {.lex_state = 303, .external_lex_state = 57},
  [2571] = {.lex_state = 303, .external_lex_state = 57},
  [2572] = {.lex_state = 233, .external_lex_state = 44},
  [2573] = {.lex_state = 278, .external_lex_state = 54},
  [2574] = {.lex_state = 309, .external_lex_state = 56},
  [2575] = {.lex_state = 278, .external_lex_state = 54},
  [2576] = {.lex_state = 303, .external_lex_state = 57},
  [2577] = {.lex_state = 303, .external_lex_state = 57},
  [2578] = {.lex_state = 233, .external_lex_state = 44},
  [2579] = {.lex_state = 309, .external_lex_state = 56},
  [2580] = {.lex_state = 309, .external_lex_state = 56},
  [2581] = {.lex_state = 303, .external_lex_state = 57},
  [2582] = {.lex_state = 278, .external_lex_state = 54},
  [2583] = {.lex_state = 301, .external_lex_state = 59},
  [2584] = {.lex_state = 273, .external_lex_state = 45},
  [2585] = {.lex_state = 304, .external_lex_state = 57},
  [2586] = {.lex_state = 276, .external_lex_state = 54},
  [2587] = {.lex_state = 304, .external_lex_state = 57},
  [2588] = {.lex_state = 278, .external_lex_state = 54},
  [2589] = {.lex_state = 271, .external_lex_state = 60},
  [2590] = {.lex_state = 304, .external_lex_state = 57},
  [2591] = {.lex_state = 304, .external_lex_state = 57},
  [2592] = {.lex_state = 304, .external_lex_state = 57},
  [2593] = {.lex_state = 276, .external_lex_state = 54},
  [2594] = {.lex_state = 271, .external_lex_state = 42},
  [2595] = {.lex_state = 298, .external_lex_state = 29},
  [2596] = {.lex_state = 276, .external_lex_state = 45},
  [2597] = {.lex_state = 276, .external_lex_state = 45},
  [2598] = {.lex_state = 276, .external_lex_state = 45},
  [2599] = {.lex_state = 304, .external_lex_state = 57},
  [2600] = {.lex_state = 304, .external_lex_state = 57},
  [2601] = {.lex_state = 273, .external_lex_state = 61},
  [2602] = {.lex_state = 302, .external_lex_state = 33},
  [2603] = {.lex_state = 276, .external_lex_state = 54},
  [2604] = {.lex_state = 304, .external_lex_state = 57},
  [2605] = {.lex_state = 304, .external_lex_state = 57},
  [2606] = {.lex_state = 304, .external_lex_state = 57},
  [2607] = {.lex_state = 304, .external_lex_state = 57},
  [2608] = {.lex_state = 304, .external_lex_state = 57},
  [2609] = {.lex_state = 304, .external_lex_state = 57},
  [2610] = {.lex_state = 304, .external_lex_state = 57},
  [2611] = {.lex_state = 304, .external_lex_state = 57},
  [2612] = {.lex_state = 307, .external_lex_state = 59},
  [2613] = {.lex_state = 307, .external_lex_state = 59},
  [2614] = {.lex_state = 304, .external_lex_state = 57},
  [2615] = {.lex_state = 304, .external_lex_state = 57},
  [2616] = {.lex_state = 304, .external_lex_state = 57},
  [2617] = {.lex_state = 304, .external_lex_state = 57},
  [2618] = {.lex_state = 307, .external_lex_state = 59},
  [2619] = {.lex_state = 307, .external_lex_state = 59},
  [2620] = {.lex_state = 307, .external_lex_state = 59},
  [2621] = {.lex_state = 271, .external_lex_state = 42},
  [2622] = {.lex_state = 302, .external_lex_state = 31},
  [2623] = {.lex_state = 302, .external_lex_state = 31},
  [2624] = {.lex_state = 276, .external_lex_state = 54},
  [2625] = {.lex_state = 276, .external_lex_state = 54},
  [2626] = {.lex_state = 309, .external_lex_state = 59},
  [2627] = {.lex_state = 271, .external_lex_state = 42},
  [2628] = {.lex_state = 309, .external_lex_state = 59},
  [2629] = {.lex_state = 309, .external_lex_state = 59},
  [2630] = {.lex_state = 276, .external_lex_state = 54},
  [2631] = {.lex_state = 276, .external_lex_state = 54},
  [2632] = {.lex_state = 307, .external_lex_state = 59},
  [2633] = {.lex_state = 301, .external_lex_state = 59},
  [2634] = {.lex_state = 301, .external_lex_state = 59},
  [2635] = {.lex_state = 302, .external_lex_state = 31},
  [2636] = {.lex_state = 271, .external_lex_state = 42},
  [2637] = {.lex_state = 302, .external_lex_state = 31},
  [2638] = {.lex_state = 301, .external_lex_state = 59},
  [2639] = {.lex_state = 301, .external_lex_state = 59},
  [2640] = {.lex_state = 301, .external_lex_state = 59},
  [2641] = {.lex_state = 273, .external_lex_state = 48},
  [2642] = {.lex_state = 309, .external_lex_state = 59},
  [2643] = {.lex_state = 301, .external_lex_state = 59},
  [2644] = {.lex_state = 301, .external_lex_state = 59},
  [2645] = {.lex_state = 309, .external_lex_state = 59},
  [2646] = {.lex_state = 309, .external_lex_state = 59},
  [2647] = {.lex_state = 276, .external_lex_state = 45},
  [2648] = {.lex_state = 271, .external_lex_state = 42},
  [2649] = {.lex_state = 309, .external_lex_state = 59},
  [2650] = {.lex_state = 309, .external_lex_state = 59},
  [2651] = {.lex_state = 301, .external_lex_state = 59},
  [2652] = {.lex_state = 309, .external_lex_state = 59},
  [2653] = {.lex_state = 301, .external_lex_state = 59},
  [2654] = {.lex_state = 276, .external_lex_state = 54},
  [2655] = {.lex_state = 301, .external_lex_state = 59},
  [2656] = {.lex_state = 273, .external_lex_state = 61},
  [2657] = {.lex_state = 273, .external_lex_state = 61},
  [2658] = {.lex_state = 273, .external_lex_state = 61},
  [2659] = {.lex_state = 271, .external_lex_state = 42},
  [2660] = {.lex_state = 273, .external_lex_state = 61},
  [2661] = {.lex_state = 276, .external_lex_state = 45},
  [2662] = {.lex_state = 273, .external_lex_state = 61},
  [2663] = {.lex_state = 309, .external_lex_state = 59},
  [2664] = {.lex_state = 276, .external_lex_state = 45},
  [2665] = {.lex_state = 276, .external_lex_state = 45},
  [2666] = {.lex_state = 309, .external_lex_state = 59},
  [2667] = {.lex_state = 309, .external_lex_state = 59},
  [2668] = {.lex_state = 273, .external_lex_state = 61},
  [2669] = {.lex_state = 273, .external_lex_state = 61},
  [2670] = {.lex_state = 273, .external_lex_state = 61},
  [2671] = {.lex_state = 298, .external_lex_state = 29},
  [2672] = {.lex_state = 301, .external_lex_state = 59},
  [2673] = {.lex_state = 276, .external_lex_state = 45},
  [2674] = {.lex_state = 301, .external_lex_state = 59},
  [2675] = {.lex_state = 276, .external_lex_state = 54},
  [2676] = {.lex_state = 301, .external_lex_state = 59},
  [2677] = {.lex_state = 301, .external_lex_state = 59},
  [2678] = {.lex_state = 298, .external_lex_state = 29},
  [2679] = {.lex_state = 276, .external_lex_state = 45},
  [2680] = {.lex_state = 276, .external_lex_state = 45},
  [2681] = {.lex_state = 276, .external_lex_state = 54},
  [2682] = {.lex_state = 301, .external_lex_state = 59},
  [2683] = {.lex_state = 309, .external_lex_state = 59},
  [2684] = {.lex_state = 301, .external_lex_state = 59},
  [2685] = {.lex_state = 309, .external_lex_state = 59},
  [2686] = {.lex_state = 301, .external_lex_state = 59},
  [2687] = {.lex_state = 301, .external_lex_state = 59},
  [2688] = {.lex_state = 308, .external_lex_state = 29},
  [2689] = {.lex_state = 298, .external_lex_state = 29},
  [2690] = {.lex_state = 277, .external_lex_state = 42},
  [2691] = {.lex_state = 302, .external_lex_state = 31},
  [2692] = {.lex_state = 308, .external_lex_state = 29},
  [2693] = {.lex_state = 302, .external_lex_state = 31},
  [2694] = {.lex_state = 271, .external_lex_state = 62},
  [2695] = {.lex_state = 277, .external_lex_state = 42},
  [2696] = {.lex_state = 271, .external_lex_state = 62},
  [2697] = {.lex_state = 276, .external_lex_state = 45},
  [2698] = {.lex_state = 276, .external_lex_state = 45},
  [2699] = {.lex_state = 278, .external_lex_state = 42},
  [2700] = {.lex_state = 309, .external_lex_state = 59},
  [2701] = {.lex_state = 309, .external_lex_state = 59},
  [2702] = {.lex_state = 271, .external_lex_state = 62},
  [2703] = {.lex_state = 298, .external_lex_state = 29},
  [2704] = {.lex_state = 271, .external_lex_state = 62},
  [2705] = {.lex_state = 150, .external_lex_state = 63},
  [2706] = {.lex_state = 150, .external_lex_state = 63},
  [2707] = {.lex_state = 309, .external_lex_state = 59},
  [2708] = {.lex_state = 276, .external_lex_state = 54},
  [2709] = {.lex_state = 276, .external_lex_state = 54},
  [2710] = {.lex_state = 276, .external_lex_state = 54},
  [2711] = {.lex_state = 309, .external_lex_state = 59},
  [2712] = {.lex_state = 309, .external_lex_state = 59},
  [2713] = {.lex_state = 281, .external_lex_state = 42},
  [2714] = {.lex_state = 308, .external_lex_state = 29},
  [2715] = {.lex_state = 302, .external_lex_state = 33},
  [2716] = {.lex_state = 276, .external_lex_state = 45},
  [2717] = {.lex_state = 278, .external_lex_state = 42},
  [2718] = {.lex_state = 298, .external_lex_state = 29},
  [2719] = {.lex_state = 276, .external_lex_state = 54},
  [2720] = {.lex_state = 278, .external_lex_state = 54},
  [2721] = {.lex_state = 276, .external_lex_state = 45},
  [2722] = {.lex_state = 276, .external_lex_state = 54},
  [2723] = {.lex_state = 278, .external_lex_state = 54},
  [2724] = {.lex_state = 271, .external_lex_state = 42},
  [2725] = {.lex_state = 302, .external_lex_state = 31},
  [2726] = {.lex_state = 298, .external_lex_state = 29},
  [2727] = {.lex_state = 273, .external_lex_state = 48},
  [2728] = {.lex_state = 276, .external_lex_state = 54},
  [2729] = {.lex_state = 276, .external_lex_state = 45},
  [2730] = {.lex_state = 276, .external_lex_state = 45},
  [2731] = {.lex_state = 276, .external_lex_state = 45},
  [2732] = {.lex_state = 276, .external_lex_state = 54},
  [2733] = {.lex_state = 276, .external_lex_state = 54},
  [2734] = {.lex_state = 276, .external_lex_state = 54},
  [2735] = {.lex_state = 276, .external_lex_state = 45},
  [2736] = {.lex_state = 276, .external_lex_state = 45},
  [2737] = {.lex_state = 273, .external_lex_state = 58},
  [2738] = {.lex_state = 273, .external_lex_state = 45},
  [2739] = {.lex_state = 271, .external_lex_state = 60},
  [2740] = {.lex_state = 271, .external_lex_state = 60},
  [2741] = {.lex_state = 271, .external_lex_state = 60},
  [2742] = {.lex_state = 271, .external_lex_state = 60},
  [2743] = {.lex_state = 271, .external_lex_state = 60},
  [2744] = {.lex_state = 271, .external_lex_state = 60},
  [2745] = {.lex_state = 271, .external_lex_state = 60},
  [2746] = {.lex_state = 271, .external_lex_state = 60},
  [2747] = {.lex_state = 271, .external_lex_state = 60},
  [2748] = {.lex_state = 271, .external_lex_state = 60},
  [2749] = {.lex_state = 271, .external_lex_state = 60},
  [2750] = {.lex_state = 271, .external_lex_state = 60},
  [2751] = {.lex_state = 271, .external_lex_state = 60},
  [2752] = {.lex_state = 271, .external_lex_state = 60},
  [2753] = {.lex_state = 271, .external_lex_state = 60},
  [2754] = {.lex_state = 271, .external_lex_state = 60},
  [2755] = {.lex_state = 271, .external_lex_state = 60},
  [2756] = {.lex_state = 271, .external_lex_state = 60},
  [2757] = {.lex_state = 271, .external_lex_state = 60},
  [2758] = {.lex_state = 271, .external_lex_state = 60},
  [2759] = {.lex_state = 271, .external_lex_state = 60},
  [2760] = {.lex_state = 271, .external_lex_state = 60},
  [2761] = {.lex_state = 271, .external_lex_state = 60},
  [2762] = {.lex_state = 271, .external_lex_state = 60},
  [2763] = {.lex_state = 271, .external_lex_state = 60},
  [2764] = {.lex_state = 271, .external_lex_state = 60},
  [2765] = {.lex_state = 271, .external_lex_state = 60},
  [2766] = {.lex_state = 271, .external_lex_state = 60},
  [2767] = {.lex_state = 271, .external_lex_state = 60},
  [2768] = {.lex_state = 271, .external_lex_state = 60},
  [2769] = {.lex_state = 271, .external_lex_state = 60},
  [2770] = {.lex_state = 271, .external_lex_state = 60},
  [2771] = {.lex_state = 271, .external_lex_state = 60},
  [2772] = {.lex_state = 271, .external_lex_state = 60},
  [2773] = {.lex_state = 271, .external_lex_state = 60},
  [2774] = {.lex_state = 271, .external_lex_state = 60},
  [2775] = {.lex_state = 271, .external_lex_state = 60},
  [2776] = {.lex_state = 305, .external_lex_state = 33},
  [2777] = {.lex_state = 271, .external_lex_state = 60},
  [2778] = {.lex_state = 271, .external_lex_state = 60},
  [2779] = {.lex_state = 271, .external_lex_state = 60},
  [2780] = {.lex_state = 298, .external_lex_state = 29},
  [2781] = {.lex_state = 271, .external_lex_state = 60},
  [2782] = {.lex_state = 271, .external_lex_state = 60},
  [2783] = {.lex_state = 271, .external_lex_state = 60},
  [2784] = {.lex_state = 271, .external_lex_state = 60},
  [2785] = {.lex_state = 271, .external_lex_state = 60},
  [2786] = {.lex_state = 271, .external_lex_state = 60},
  [2787] = {.lex_state = 271, .external_lex_state = 60},
  [2788] = {.lex_state = 271, .external_lex_state = 60},
  [2789] = {.lex_state = 271, .external_lex_state = 60},
  [2790] = {.lex_state = 271, .external_lex_state = 60},
  [2791] = {.lex_state = 271, .external_lex_state = 60},
  [2792] = {.lex_state = 271, .external_lex_state = 60},
  [2793] = {.lex_state = 271, .external_lex_state = 60},
  [2794] = {.lex_state = 271, .external_lex_state = 60},
  [2795] = {.lex_state = 271, .external_lex_state = 60},
  [2796] = {.lex_state = 273, .external_lex_state = 48},
  [2797] = {.lex_state = 271, .external_lex_state = 60},
  [2798] = {.lex_state = 298, .external_lex_state = 29},
  [2799] = {.lex_state = 271, .external_lex_state = 60},
  [2800] = {.lex_state = 271, .external_lex_state = 60},
  [2801] = {.lex_state = 271, .external_lex_state = 60},
  [2802] = {.lex_state = 273, .external_lex_state = 45},
  [2803] = {.lex_state = 271, .external_lex_state = 60},
  [2804] = {.lex_state = 271, .external_lex_state = 60},
  [2805] = {.lex_state = 273, .external_lex_state = 45},
  [2806] = {.lex_state = 305, .external_lex_state = 33},
  [2807] = {.lex_state = 308, .external_lex_state = 29},
  [2808] = {.lex_state = 273, .external_lex_state = 45},
  [2809] = {.lex_state = 271, .external_lex_state = 60},
  [2810] = {.lex_state = 273, .external_lex_state = 45},
  [2811] = {.lex_state = 271, .external_lex_state = 60},
  [2812] = {.lex_state = 273, .external_lex_state = 45},
  [2813] = {.lex_state = 273, .external_lex_state = 45},
  [2814] = {.lex_state = 273, .external_lex_state = 45},
  [2815] = {.lex_state = 273, .external_lex_state = 45},
  [2816] = {.lex_state = 273, .external_lex_state = 45},
  [2817] = {.lex_state = 308, .external_lex_state = 29},
  [2818] = {.lex_state = 271, .external_lex_state = 60},
  [2819] = {.lex_state = 302, .external_lex_state = 33},
  [2820] = {.lex_state = 271, .external_lex_state = 60},
  [2821] = {.lex_state = 271, .external_lex_state = 60},
  [2822] = {.lex_state = 271, .external_lex_state = 60},
  [2823] = {.lex_state = 271, .external_lex_state = 60},
  [2824] = {.lex_state = 271, .external_lex_state = 60},
  [2825] = {.lex_state = 271, .external_lex_state = 60},
  [2826] = {.lex_state = 271, .external_lex_state = 60},
  [2827] = {.lex_state = 273, .external_lex_state = 45},
  [2828] = {.lex_state = 273, .external_lex_state = 45},
  [2829] = {.lex_state = 273, .external_lex_state = 45},
  [2830] = {.lex_state = 273, .external_lex_state = 45},
  [2831] = {.lex_state = 273, .external_lex_state = 45},
  [2832] = {.lex_state = 273, .external_lex_state = 45},
  [2833] = {.lex_state = 273, .external_lex_state = 45},
  [2834] = {.lex_state = 273, .external_lex_state = 45},
  [2835] = {.lex_state = 273, .external_lex_state = 45},
  [2836] = {.lex_state = 271, .external_lex_state = 60},
  [2837] = {.lex_state = 273, .external_lex_state = 45},
  [2838] = {.lex_state = 64, .external_lex_state = 64},
  [2839] = {.lex_state = 273, .external_lex_state = 45},
  [2840] = {.lex_state = 271, .external_lex_state = 60},
  [2841] = {.lex_state = 307, .external_lex_state = 59},
  [2842] = {.lex_state = 307, .external_lex_state = 59},
  [2843] = {.lex_state = 271, .external_lex_state = 60},
  [2844] = {.lex_state = 271, .external_lex_state = 60},
  [2845] = {.lex_state = 271, .external_lex_state = 60},
  [2846] = {.lex_state = 298, .external_lex_state = 29},
  [2847] = {.lex_state = 271, .external_lex_state = 60},
  [2848] = {.lex_state = 271, .external_lex_state = 60},
  [2849] = {.lex_state = 271, .external_lex_state = 60},
  [2850] = {.lex_state = 307, .external_lex_state = 59},
  [2851] = {.lex_state = 273, .external_lex_state = 45},
  [2852] = {.lex_state = 271, .external_lex_state = 60},
  [2853] = {.lex_state = 271, .external_lex_state = 60},
  [2854] = {.lex_state = 307, .external_lex_state = 59},
  [2855] = {.lex_state = 271, .external_lex_state = 60},
  [2856] = {.lex_state = 271, .external_lex_state = 60},
  [2857] = {.lex_state = 307, .external_lex_state = 59},
  [2858] = {.lex_state = 271, .external_lex_state = 60},
  [2859] = {.lex_state = 307, .external_lex_state = 59},
  [2860] = {.lex_state = 271, .external_lex_state = 60},
  [2861] = {.lex_state = 307, .external_lex_state = 59},
  [2862] = {.lex_state = 273, .external_lex_state = 58},
  [2863] = {.lex_state = 273, .external_lex_state = 45},
  [2864] = {.lex_state = 298, .external_lex_state = 29},
  [2865] = {.lex_state = 273, .external_lex_state = 65},
  [2866] = {.lex_state = 271, .external_lex_state = 60},
  [2867] = {.lex_state = 298, .external_lex_state = 34},
  [2868] = {.lex_state = 273, .external_lex_state = 45},
  [2869] = {.lex_state = 307, .external_lex_state = 59},
  [2870] = {.lex_state = 307, .external_lex_state = 59},
  [2871] = {.lex_state = 273, .external_lex_state = 65},
  [2872] = {.lex_state = 273, .external_lex_state = 65},
  [2873] = {.lex_state = 273, .external_lex_state = 65},
  [2874] = {.lex_state = 273, .external_lex_state = 45},
  [2875] = {.lex_state = 273, .external_lex_state = 65},
  [2876] = {.lex_state = 273, .external_lex_state = 65},
  [2877] = {.lex_state = 307, .external_lex_state = 59},
  [2878] = {.lex_state = 308, .external_lex_state = 29},
  [2879] = {.lex_state = 307, .external_lex_state = 59},
  [2880] = {.lex_state = 271, .external_lex_state = 60},
  [2881] = {.lex_state = 273, .external_lex_state = 65},
  [2882] = {.lex_state = 273, .external_lex_state = 65},
  [2883] = {.lex_state = 273, .external_lex_state = 65},
  [2884] = {.lex_state = 298, .external_lex_state = 29},
  [2885] = {.lex_state = 273, .external_lex_state = 58},
  [2886] = {.lex_state = 298, .external_lex_state = 29},
  [2887] = {.lex_state = 175, .external_lex_state = 66},
  [2888] = {.lex_state = 175, .external_lex_state = 66},
  [2889] = {.lex_state = 273, .external_lex_state = 45},
  [2890] = {.lex_state = 271, .external_lex_state = 60},
  [2891] = {.lex_state = 273, .external_lex_state = 58},
  [2892] = {.lex_state = 273, .external_lex_state = 58},
  [2893] = {.lex_state = 273, .external_lex_state = 58},
  [2894] = {.lex_state = 271, .external_lex_state = 60},
  [2895] = {.lex_state = 178, .external_lex_state = 63},
  [2896] = {.lex_state = 178, .external_lex_state = 63},
  [2897] = {.lex_state = 273, .external_lex_state = 45},
  [2898] = {.lex_state = 271, .external_lex_state = 60},
  [2899] = {.lex_state = 271, .external_lex_state = 60},
  [2900] = {.lex_state = 273, .external_lex_state = 45},
  [2901] = {.lex_state = 307, .external_lex_state = 59},
  [2902] = {.lex_state = 178, .external_lex_state = 63},
  [2903] = {.lex_state = 178, .external_lex_state = 63},
  [2904] = {.lex_state = 307, .external_lex_state = 59},
  [2905] = {.lex_state = 298, .external_lex_state = 29},
  [2906] = {.lex_state = 307, .external_lex_state = 59},
  [2907] = {.lex_state = 307, .external_lex_state = 59},
  [2908] = {.lex_state = 273, .external_lex_state = 58},
  [2909] = {.lex_state = 273, .external_lex_state = 58},
  [2910] = {.lex_state = 273, .external_lex_state = 58},
  [2911] = {.lex_state = 307, .external_lex_state = 59},
  [2912] = {.lex_state = 308, .external_lex_state = 34},
  [2913] = {.lex_state = 298, .external_lex_state = 29},
  [2914] = {.lex_state = 298, .external_lex_state = 34},
  [2915] = {.lex_state = 271, .external_lex_state = 60},
  [2916] = {.lex_state = 271, .external_lex_state = 60},
  [2917] = {.lex_state = 271, .external_lex_state = 60},
  [2918] = {.lex_state = 307, .external_lex_state = 59},
  [2919] = {.lex_state = 307, .external_lex_state = 59},
  [2920] = {.lex_state = 307, .external_lex_state = 59},
  [2921] = {.lex_state = 271, .external_lex_state = 60},
  [2922] = {.lex_state = 271, .external_lex_state = 60},
  [2923] = {.lex_state = 271, .external_lex_state = 60},
  [2924] = {.lex_state = 273, .external_lex_state = 45},
  [2925] = {.lex_state = 273, .external_lex_state = 45},
  [2926] = {.lex_state = 271, .external_lex_state = 60},
  [2927] = {.lex_state = 277, .external_lex_state = 42},
  [2928] = {.lex_state = 278, .external_lex_state = 42},
  [2929] = {.lex_state = 271, .external_lex_state = 60},
  [2930] = {.lex_state = 271, .external_lex_state = 60},
  [2931] = {.lex_state = 271, .external_lex_state = 60},
  [2932] = {.lex_state = 271, .external_lex_state = 60},
  [2933] = {.lex_state = 273, .external_lex_state = 45},
  [2934] = {.lex_state = 298, .external_lex_state = 29},
  [2935] = {.lex_state = 308, .external_lex_state = 29},
  [2936] = {.lex_state = 271, .external_lex_state = 60},
  [2937] = {.lex_state = 271, .external_lex_state = 60},
  [2938] = {.lex_state = 271, .external_lex_state = 60},
  [2939] = {.lex_state = 308, .external_lex_state = 34},
  [2940] = {.lex_state = 298, .external_lex_state = 29},
  [2941] = {.lex_state = 298, .external_lex_state = 29},
  [2942] = {.lex_state = 271, .external_lex_state = 42},
  [2943] = {.lex_state = 271, .external_lex_state = 42},
  [2944] = {.lex_state = 311, .external_lex_state = 67},
  [2945] = {.lex_state = 275, .external_lex_state = 60},
  [2946] = {.lex_state = 64, .external_lex_state = 64},
  [2947] = {.lex_state = 64, .external_lex_state = 64},
  [2948] = {.lex_state = 64, .external_lex_state = 64},
  [2949] = {.lex_state = 279, .external_lex_state = 60},
  [2950] = {.lex_state = 273, .external_lex_state = 48},
  [2951] = {.lex_state = 273, .external_lex_state = 48},
  [2952] = {.lex_state = 64, .external_lex_state = 64},
  [2953] = {.lex_state = 273, .external_lex_state = 48},
  [2954] = {.lex_state = 273, .external_lex_state = 42},
  [2955] = {.lex_state = 271, .external_lex_state = 42},
  [2956] = {.lex_state = 64, .external_lex_state = 64},
  [2957] = {.lex_state = 273, .external_lex_state = 48},
  [2958] = {.lex_state = 273, .external_lex_state = 48},
  [2959] = {.lex_state = 273, .external_lex_state = 48},
  [2960] = {.lex_state = 273, .external_lex_state = 48},
  [2961] = {.lex_state = 273, .external_lex_state = 48},
  [2962] = {.lex_state = 273, .external_lex_state = 48},
  [2963] = {.lex_state = 273, .external_lex_state = 48},
  [2964] = {.lex_state = 273, .external_lex_state = 48},
  [2965] = {.lex_state = 64, .external_lex_state = 64},
  [2966] = {.lex_state = 273, .external_lex_state = 48},
  [2967] = {.lex_state = 271, .external_lex_state = 42},
  [2968] = {.lex_state = 271, .external_lex_state = 42},
  [2969] = {.lex_state = 271, .external_lex_state = 42},
  [2970] = {.lex_state = 271, .external_lex_state = 42},
  [2971] = {.lex_state = 271, .external_lex_state = 42},
  [2972] = {.lex_state = 271, .external_lex_state = 42},
  [2973] = {.lex_state = 271, .external_lex_state = 42},
  [2974] = {.lex_state = 271, .external_lex_state = 42},
  [2975] = {.lex_state = 271, .external_lex_state = 42},
  [2976] = {.lex_state = 271, .external_lex_state = 42},
  [2977] = {.lex_state = 271, .external_lex_state = 42},
  [2978] = {.lex_state = 271, .external_lex_state = 42},
  [2979] = {.lex_state = 271, .external_lex_state = 42},
  [2980] = {.lex_state = 308, .external_lex_state = 34},
  [2981] = {.lex_state = 273, .external_lex_state = 48},
  [2982] = {.lex_state = 311, .external_lex_state = 67},
  [2983] = {.lex_state = 273, .external_lex_state = 68},
  [2984] = {.lex_state = 273, .external_lex_state = 68},
  [2985] = {.lex_state = 273, .external_lex_state = 48},
  [2986] = {.lex_state = 273, .external_lex_state = 48},
  [2987] = {.lex_state = 271, .external_lex_state = 42},
  [2988] = {.lex_state = 273, .external_lex_state = 48},
  [2989] = {.lex_state = 273, .external_lex_state = 48},
  [2990] = {.lex_state = 273, .external_lex_state = 48},
  [2991] = {.lex_state = 64, .external_lex_state = 64},
  [2992] = {.lex_state = 64, .external_lex_state = 64},
  [2993] = {.lex_state = 64, .external_lex_state = 64},
  [2994] = {.lex_state = 271, .external_lex_state = 42},
  [2995] = {.lex_state = 271, .external_lex_state = 42},
  [2996] = {.lex_state = 271, .external_lex_state = 42},
  [2997] = {.lex_state = 64, .external_lex_state = 64},
  [2998] = {.lex_state = 273, .external_lex_state = 48},
  [2999] = {.lex_state = 64, .external_lex_state = 64},
  [3000] = {.lex_state = 271, .external_lex_state = 42},
  [3001] = {.lex_state = 271, .external_lex_state = 42},
  [3002] = {.lex_state = 271, .external_lex_state = 42},
  [3003] = {.lex_state = 271, .external_lex_state = 42},
  [3004] = {.lex_state = 271, .external_lex_state = 42},
  [3005] = {.lex_state = 64, .external_lex_state = 64},
  [3006] = {.lex_state = 273, .external_lex_state = 48},
  [3007] = {.lex_state = 271, .external_lex_state = 42},
  [3008] = {.lex_state = 271, .external_lex_state = 42},
  [3009] = {.lex_state = 273, .external_lex_state = 48},
  [3010] = {.lex_state = 271, .external_lex_state = 42},
  [3011] = {.lex_state = 273, .external_lex_state = 48},
  [3012] = {.lex_state = 273, .external_lex_state = 48},
  [3013] = {.lex_state = 271, .external_lex_state = 42},
  [3014] = {.lex_state = 271, .external_lex_state = 42},
  [3015] = {.lex_state = 181, .external_lex_state = 66},
  [3016] = {.lex_state = 181, .external_lex_state = 66},
  [3017] = {.lex_state = 271, .external_lex_state = 60},
  [3018] = {.lex_state = 181, .external_lex_state = 66},
  [3019] = {.lex_state = 64, .external_lex_state = 64},
  [3020] = {.lex_state = 64, .external_lex_state = 64},
  [3021] = {.lex_state = 64, .external_lex_state = 64},
  [3022] = {.lex_state = 271, .external_lex_state = 42},
  [3023] = {.lex_state = 271, .external_lex_state = 42},
  [3024] = {.lex_state = 64, .external_lex_state = 64},
  [3025] = {.lex_state = 64, .external_lex_state = 64},
  [3026] = {.lex_state = 64, .external_lex_state = 64},
  [3027] = {.lex_state = 181, .external_lex_state = 66},
  [3028] = {.lex_state = 273, .external_lex_state = 68},
  [3029] = {.lex_state = 273, .external_lex_state = 68},
  [3030] = {.lex_state = 271, .external_lex_state = 42},
  [3031] = {.lex_state = 274, .external_lex_state = 60},
  [3032] = {.lex_state = 311, .external_lex_state = 67},
  [3033] = {.lex_state = 271, .external_lex_state = 42},
  [3034] = {.lex_state = 271, .external_lex_state = 42},
  [3035] = {.lex_state = 275, .external_lex_state = 69},
  [3036] = {.lex_state = 64, .external_lex_state = 64},
  [3037] = {.lex_state = 271, .external_lex_state = 42},
  [3038] = {.lex_state = 311, .external_lex_state = 67},
  [3039] = {.lex_state = 271, .external_lex_state = 42},
  [3040] = {.lex_state = 271, .external_lex_state = 42},
  [3041] = {.lex_state = 271, .external_lex_state = 42},
  [3042] = {.lex_state = 273, .external_lex_state = 42},
  [3043] = {.lex_state = 271, .external_lex_state = 42},
  [3044] = {.lex_state = 184, .external_lex_state = 66},
  [3045] = {.lex_state = 64, .external_lex_state = 64},
  [3046] = {.lex_state = 271, .external_lex_state = 42},
  [3047] = {.lex_state = 184, .external_lex_state = 66},
  [3048] = {.lex_state = 271, .external_lex_state = 42},
  [3049] = {.lex_state = 273, .external_lex_state = 48},
  [3050] = {.lex_state = 271, .external_lex_state = 42},
  [3051] = {.lex_state = 271, .external_lex_state = 42},
  [3052] = {.lex_state = 273, .external_lex_state = 48},
  [3053] = {.lex_state = 273, .external_lex_state = 42},
  [3054] = {.lex_state = 271, .external_lex_state = 42},
  [3055] = {.lex_state = 298, .external_lex_state = 34},
  [3056] = {.lex_state = 64, .external_lex_state = 64},
  [3057] = {.lex_state = 311, .external_lex_state = 67},
  [3058] = {.lex_state = 273, .external_lex_state = 42},
  [3059] = {.lex_state = 271, .external_lex_state = 42},
  [3060] = {.lex_state = 271, .external_lex_state = 42},
  [3061] = {.lex_state = 271, .external_lex_state = 42},
  [3062] = {.lex_state = 271, .external_lex_state = 42},
  [3063] = {.lex_state = 271, .external_lex_state = 42},
  [3064] = {.lex_state = 271, .external_lex_state = 42},
  [3065] = {.lex_state = 306, .external_lex_state = 34},
  [3066] = {.lex_state = 311, .external_lex_state = 67},
  [3067] = {.lex_state = 271, .external_lex_state = 42},
  [3068] = {.lex_state = 271, .external_lex_state = 42},
  [3069] = {.lex_state = 273, .external_lex_state = 62},
  [3070] = {.lex_state = 311, .external_lex_state = 67},
  [3071] = {.lex_state = 273, .external_lex_state = 48},
  [3072] = {.lex_state = 273, .external_lex_state = 62},
  [3073] = {.lex_state = 271, .external_lex_state = 42},
  [3074] = {.lex_state = 64, .external_lex_state = 64},
  [3075] = {.lex_state = 64, .external_lex_state = 64},
  [3076] = {.lex_state = 64, .external_lex_state = 64},
  [3077] = {.lex_state = 273, .external_lex_state = 42},
  [3078] = {.lex_state = 273, .external_lex_state = 42},
  [3079] = {.lex_state = 273, .external_lex_state = 42},
  [3080] = {.lex_state = 273, .external_lex_state = 42},
  [3081] = {.lex_state = 273, .external_lex_state = 42},
  [3082] = {.lex_state = 273, .external_lex_state = 42},
  [3083] = {.lex_state = 273, .external_lex_state = 42},
  [3084] = {.lex_state = 273, .external_lex_state = 42},
  [3085] = {.lex_state = 273, .external_lex_state = 48},
  [3086] = {.lex_state = 273, .external_lex_state = 42},
  [3087] = {.lex_state = 271, .external_lex_state = 42},
  [3088] = {.lex_state = 273, .external_lex_state = 42},
  [3089] = {.lex_state = 271, .external_lex_state = 42},
  [3090] = {.lex_state = 273, .external_lex_state = 42},
  [3091] = {.lex_state = 273, .external_lex_state = 42},
  [3092] = {.lex_state = 236, .external_lex_state = 70},
  [3093] = {.lex_state = 236, .external_lex_state = 70},
  [3094] = {.lex_state = 273, .external_lex_state = 42},
  [3095] = {.lex_state = 273, .external_lex_state = 42},
  [3096] = {.lex_state = 271, .external_lex_state = 42},
  [3097] = {.lex_state = 273, .external_lex_state = 42},
  [3098] = {.lex_state = 273, .external_lex_state = 42},
  [3099] = {.lex_state = 273, .external_lex_state = 62},
  [3100] = {.lex_state = 273, .external_lex_state = 42},
  [3101] = {.lex_state = 271, .external_lex_state = 42},
  [3102] = {.lex_state = 271, .external_lex_state = 42},
  [3103] = {.lex_state = 273, .external_lex_state = 62},
  [3104] = {.lex_state = 271, .external_lex_state = 42},
  [3105] = {.lex_state = 311, .external_lex_state = 67},
  [3106] = {.lex_state = 311, .external_lex_state = 67},
  [3107] = {.lex_state = 273, .external_lex_state = 42},
  [3108] = {.lex_state = 271, .external_lex_state = 42},
  [3109] = {.lex_state = 311, .external_lex_state = 67},
  [3110] = {.lex_state = 236, .external_lex_state = 70},
  [3111] = {.lex_state = 273, .external_lex_state = 42},
  [3112] = {.lex_state = 271, .external_lex_state = 42},
  [3113] = {.lex_state = 64, .external_lex_state = 64},
  [3114] = {.lex_state = 306, .external_lex_state = 34},
  [3115] = {.lex_state = 236, .external_lex_state = 70},
  [3116] = {.lex_state = 64, .external_lex_state = 64},
  [3117] = {.lex_state = 236, .external_lex_state = 70},
  [3118] = {.lex_state = 64, .external_lex_state = 64},
  [3119] = {.lex_state = 64, .external_lex_state = 64},
  [3120] = {.lex_state = 236, .external_lex_state = 70},
  [3121] = {.lex_state = 273, .external_lex_state = 42},
  [3122] = {.lex_state = 311, .external_lex_state = 67},
  [3123] = {.lex_state = 271, .external_lex_state = 42},
  [3124] = {.lex_state = 273, .external_lex_state = 48},
  [3125] = {.lex_state = 271, .external_lex_state = 42},
  [3126] = {.lex_state = 273, .external_lex_state = 42},
  [3127] = {.lex_state = 296, .external_lex_state = 36},
  [3128] = {.lex_state = 271, .external_lex_state = 42},
  [3129] = {.lex_state = 271, .external_lex_state = 42},
  [3130] = {.lex_state = 273, .external_lex_state = 42},
  [3131] = {.lex_state = 273, .external_lex_state = 48},
  [3132] = {.lex_state = 187, .external_lex_state = 66},
  [3133] = {.lex_state = 311, .external_lex_state = 67},
  [3134] = {.lex_state = 187, .external_lex_state = 66},
  [3135] = {.lex_state = 311, .external_lex_state = 67},
  [3136] = {.lex_state = 311, .external_lex_state = 67},
  [3137] = {.lex_state = 311, .external_lex_state = 67},
  [3138] = {.lex_state = 273, .external_lex_state = 42},
  [3139] = {.lex_state = 271, .external_lex_state = 42},
  [3140] = {.lex_state = 273, .external_lex_state = 48},
  [3141] = {.lex_state = 271, .external_lex_state = 42},
  [3142] = {.lex_state = 273, .external_lex_state = 42},
  [3143] = {.lex_state = 271, .external_lex_state = 42},
  [3144] = {.lex_state = 271, .external_lex_state = 42},
  [3145] = {.lex_state = 273, .external_lex_state = 48},
  [3146] = {.lex_state = 273, .external_lex_state = 42},
  [3147] = {.lex_state = 271, .external_lex_state = 42},
  [3148] = {.lex_state = 271, .external_lex_state = 42},
  [3149] = {.lex_state = 305, .external_lex_state = 33},
  [3150] = {.lex_state = 271, .external_lex_state = 42},
  [3151] = {.lex_state = 273, .external_lex_state = 42},
  [3152] = {.lex_state = 273, .external_lex_state = 48},
  [3153] = {.lex_state = 271, .external_lex_state = 42},
  [3154] = {.lex_state = 271, .external_lex_state = 42},
  [3155] = {.lex_state = 273, .external_lex_state = 48},
  [3156] = {.lex_state = 271, .external_lex_state = 42},
  [3157] = {.lex_state = 64, .external_lex_state = 64},
  [3158] = {.lex_state = 271, .external_lex_state = 42},
  [3159] = {.lex_state = 64, .external_lex_state = 64},
  [3160] = {.lex_state = 64, .external_lex_state = 64},
  [3161] = {.lex_state = 64, .external_lex_state = 64},
  [3162] = {.lex_state = 529, .external_lex_state = 66},
  [3163] = {.lex_state = 529, .external_lex_state = 66},
  [3164] = {.lex_state = 311, .external_lex_state = 67},
  [3165] = {.lex_state = 273, .external_lex_state = 69},
  [3166] = {.lex_state = 273, .external_lex_state = 69},
  [3167] = {.lex_state = 279, .external_lex_state = 60},
  [3168] = {.lex_state = 273, .external_lex_state = 60},
  [3169] = {.lex_state = 273, .external_lex_state = 60},
  [3170] = {.lex_state = 279, .external_lex_state = 60},
  [3171] = {.lex_state = 279, .external_lex_state = 60},
  [3172] = {.lex_state = 279, .external_lex_state = 60},
  [3173] = {.lex_state = 279, .external_lex_state = 60},
  [3174] = {.lex_state = 279, .external_lex_state = 60},
  [3175] = {.lex_state = 236, .external_lex_state = 70},
  [3176] = {.lex_state = 279, .external_lex_state = 60},
  [3177] = {.lex_state = 236, .external_lex_state = 70},
  [3178] = {.lex_state = 273, .external_lex_state = 60},
  [3179] = {.lex_state = 273, .external_lex_state = 60},
  [3180] = {.lex_state = 273, .external_lex_state = 60},
  [3181] = {.lex_state = 236, .external_lex_state = 70},
  [3182] = {.lex_state = 273, .external_lex_state = 69},
  [3183] = {.lex_state = 271, .external_lex_state = 42},
  [3184] = {.lex_state = 236, .external_lex_state = 70},
  [3185] = {.lex_state = 273, .external_lex_state = 69},
  [3186] = {.lex_state = 193, .external_lex_state = 66},
  [3187] = {.lex_state = 193, .external_lex_state = 66},
  [3188] = {.lex_state = 273, .external_lex_state = 60},
  [3189] = {.lex_state = 279, .external_lex_state = 60},
  [3190] = {.lex_state = 273, .external_lex_state = 69},
  [3191] = {.lex_state = 273, .external_lex_state = 69},
  [3192] = {.lex_state = 273, .external_lex_state = 69},
  [3193] = {.lex_state = 271, .external_lex_state = 42},
  [3194] = {.lex_state = 273, .external_lex_state = 69},
  [3195] = {.lex_state = 271, .external_lex_state = 60},
  [3196] = {.lex_state = 279, .external_lex_state = 60},
  [3197] = {.lex_state = 271, .external_lex_state = 42},
  [3198] = {.lex_state = 279, .external_lex_state = 60},
  [3199] = {.lex_state = 279, .external_lex_state = 60},
  [3200] = {.lex_state = 279, .external_lex_state = 60},
  [3201] = {.lex_state = 279, .external_lex_state = 60},
  [3202] = {.lex_state = 279, .external_lex_state = 60},
  [3203] = {.lex_state = 273, .external_lex_state = 69},
  [3204] = {.lex_state = 273, .external_lex_state = 69},
  [3205] = {.lex_state = 279, .external_lex_state = 60},
  [3206] = {.lex_state = 306, .external_lex_state = 34},
  [3207] = {.lex_state = 279, .external_lex_state = 60},
  [3208] = {.lex_state = 273, .external_lex_state = 60},
  [3209] = {.lex_state = 273, .external_lex_state = 69},
  [3210] = {.lex_state = 279, .external_lex_state = 60},
  [3211] = {.lex_state = 273, .external_lex_state = 69},
  [3212] = {.lex_state = 273, .external_lex_state = 69},
  [3213] = {.lex_state = 279, .external_lex_state = 60},
  [3214] = {.lex_state = 279, .external_lex_state = 60},
  [3215] = {.lex_state = 273, .external_lex_state = 69},
  [3216] = {.lex_state = 279, .external_lex_state = 60},
  [3217] = {.lex_state = 279, .external_lex_state = 60},
  [3218] = {.lex_state = 279, .external_lex_state = 60},
  [3219] = {.lex_state = 273, .external_lex_state = 60},
  [3220] = {.lex_state = 236, .external_lex_state = 70},
  [3221] = {.lex_state = 236, .external_lex_state = 70},
  [3222] = {.lex_state = 273, .external_lex_state = 69},
  [3223] = {.lex_state = 273, .external_lex_state = 69},
  [3224] = {.lex_state = 273, .external_lex_state = 69},
  [3225] = {.lex_state = 236, .external_lex_state = 70},
  [3226] = {.lex_state = 236, .external_lex_state = 70},
  [3227] = {.lex_state = 532, .external_lex_state = 66},
  [3228] = {.lex_state = 279, .external_lex_state = 60},
  [3229] = {.lex_state = 236, .external_lex_state = 70},
  [3230] = {.lex_state = 532, .external_lex_state = 66},
  [3231] = {.lex_state = 273, .external_lex_state = 69},
  [3232] = {.lex_state = 273, .external_lex_state = 69},
  [3233] = {.lex_state = 298, .external_lex_state = 47},
  [3234] = {.lex_state = 273, .external_lex_state = 69},
  [3235] = {.lex_state = 273, .external_lex_state = 69},
  [3236] = {.lex_state = 298, .external_lex_state = 47},
  [3237] = {.lex_state = 199, .external_lex_state = 66},
  [3238] = {.lex_state = 199, .external_lex_state = 66},
  [3239] = {.lex_state = 273, .external_lex_state = 69},
  [3240] = {.lex_state = 532, .external_lex_state = 66},
  [3241] = {.lex_state = 273, .external_lex_state = 69},
  [3242] = {.lex_state = 273, .external_lex_state = 60},
  [3243] = {.lex_state = 532, .external_lex_state = 66},
  [3244] = {.lex_state = 273, .external_lex_state = 69},
  [3245] = {.lex_state = 279, .external_lex_state = 60},
  [3246] = {.lex_state = 273, .external_lex_state = 60},
  [3247] = {.lex_state = 273, .external_lex_state = 60},
  [3248] = {.lex_state = 273, .external_lex_state = 60},
  [3249] = {.lex_state = 273, .external_lex_state = 60},
  [3250] = {.lex_state = 236, .external_lex_state = 70},
  [3251] = {.lex_state = 273, .external_lex_state = 69},
  [3252] = {.lex_state = 273, .external_lex_state = 69},
  [3253] = {.lex_state = 236, .external_lex_state = 70},
  [3254] = {.lex_state = 279, .external_lex_state = 60},
  [3255] = {.lex_state = 273, .external_lex_state = 60},
  [3256] = {.lex_state = 236, .external_lex_state = 70},
  [3257] = {.lex_state = 273, .external_lex_state = 69},
  [3258] = {.lex_state = 273, .external_lex_state = 60},
  [3259] = {.lex_state = 273, .external_lex_state = 60},
  [3260] = {.lex_state = 273, .external_lex_state = 69},
  [3261] = {.lex_state = 273, .external_lex_state = 60},
  [3262] = {.lex_state = 273, .external_lex_state = 60},
  [3263] = {.lex_state = 273, .external_lex_state = 60},
  [3264] = {.lex_state = 273, .external_lex_state = 60},
  [3265] = {.lex_state = 273, .external_lex_state = 60},
  [3266] = {.lex_state = 273, .external_lex_state = 60},
  [3267] = {.lex_state = 273, .external_lex_state = 69},
  [3268] = {.lex_state = 273, .external_lex_state = 69},
  [3269] = {.lex_state = 193, .external_lex_state = 66},
  [3270] = {.lex_state = 193, .external_lex_state = 66},
  [3271] = {.lex_state = 273, .external_lex_state = 60},
  [3272] = {.lex_state = 279, .external_lex_state = 60},
  [3273] = {.lex_state = 273, .external_lex_state = 69},
  [3274] = {.lex_state = 273, .external_lex_state = 69},
  [3275] = {.lex_state = 274, .external_lex_state = 60},
  [3276] = {.lex_state = 279, .external_lex_state = 60},
  [3277] = {.lex_state = 273, .external_lex_state = 60},
  [3278] = {.lex_state = 273, .external_lex_state = 60},
  [3279] = {.lex_state = 273, .external_lex_state = 69},
  [3280] = {.lex_state = 274, .external_lex_state = 60},
  [3281] = {.lex_state = 273, .external_lex_state = 60},
  [3282] = {.lex_state = 274, .external_lex_state = 60},
  [3283] = {.lex_state = 273, .external_lex_state = 60},
  [3284] = {.lex_state = 273, .external_lex_state = 69},
  [3285] = {.lex_state = 279, .external_lex_state = 60},
  [3286] = {.lex_state = 279, .external_lex_state = 60},
  [3287] = {.lex_state = 273, .external_lex_state = 60},
  [3288] = {.lex_state = 273, .external_lex_state = 69},
  [3289] = {.lex_state = 273, .external_lex_state = 69},
  [3290] = {.lex_state = 279, .external_lex_state = 60},
  [3291] = {.lex_state = 274, .external_lex_state = 60},
  [3292] = {.lex_state = 279, .external_lex_state = 60},
  [3293] = {.lex_state = 273, .external_lex_state = 69},
  [3294] = {.lex_state = 273, .external_lex_state = 60},
  [3295] = {.lex_state = 273, .external_lex_state = 60},
  [3296] = {.lex_state = 273, .external_lex_state = 69},
  [3297] = {.lex_state = 279, .external_lex_state = 60},
  [3298] = {.lex_state = 529, .external_lex_state = 66},
  [3299] = {.lex_state = 273, .external_lex_state = 69},
  [3300] = {.lex_state = 274, .external_lex_state = 60},
  [3301] = {.lex_state = 273, .external_lex_state = 69},
  [3302] = {.lex_state = 273, .external_lex_state = 69},
  [3303] = {.lex_state = 273, .external_lex_state = 69},
  [3304] = {.lex_state = 273, .external_lex_state = 69},
  [3305] = {.lex_state = 271, .external_lex_state = 42},
  [3306] = {.lex_state = 274, .external_lex_state = 60},
  [3307] = {.lex_state = 274, .external_lex_state = 60},
  [3308] = {.lex_state = 273, .external_lex_state = 69},
  [3309] = {.lex_state = 273, .external_lex_state = 69},
  [3310] = {.lex_state = 298, .external_lex_state = 47},
  [3311] = {.lex_state = 236, .external_lex_state = 70},
  [3312] = {.lex_state = 274, .external_lex_state = 60},
  [3313] = {.lex_state = 529, .external_lex_state = 66},
  [3314] = {.lex_state = 271, .external_lex_state = 42},
  [3315] = {.lex_state = 273, .external_lex_state = 69},
  [3316] = {.lex_state = 273, .external_lex_state = 69},
  [3317] = {.lex_state = 273, .external_lex_state = 69},
  [3318] = {.lex_state = 271, .external_lex_state = 42},
  [3319] = {.lex_state = 271, .external_lex_state = 42},
  [3320] = {.lex_state = 273, .external_lex_state = 69},
  [3321] = {.lex_state = 273, .external_lex_state = 69},
  [3322] = {.lex_state = 274, .external_lex_state = 60},
  [3323] = {.lex_state = 271, .external_lex_state = 42},
  [3324] = {.lex_state = 273, .external_lex_state = 69},
  [3325] = {.lex_state = 273, .external_lex_state = 69},
  [3326] = {.lex_state = 273, .external_lex_state = 69},
  [3327] = {.lex_state = 271, .external_lex_state = 42},
  [3328] = {.lex_state = 273, .external_lex_state = 69},
  [3329] = {.lex_state = 273, .external_lex_state = 69},
  [3330] = {.lex_state = 271, .external_lex_state = 42},
  [3331] = {.lex_state = 236, .external_lex_state = 70},
  [3332] = {.lex_state = 271, .external_lex_state = 42},
  [3333] = {.lex_state = 271, .external_lex_state = 42},
  [3334] = {.lex_state = 273, .external_lex_state = 69},
  [3335] = {.lex_state = 271, .external_lex_state = 42},
  [3336] = {.lex_state = 273, .external_lex_state = 69},
  [3337] = {.lex_state = 273, .external_lex_state = 69},
  [3338] = {.lex_state = 273, .external_lex_state = 69},
  [3339] = {.lex_state = 273, .external_lex_state = 69},
  [3340] = {.lex_state = 273, .external_lex_state = 69},
  [3341] = {.lex_state = 273, .external_lex_state = 69},
  [3342] = {.lex_state = 273, .external_lex_state = 69},
  [3343] = {.lex_state = 273, .external_lex_state = 69},
  [3344] = {.lex_state = 273, .external_lex_state = 69},
  [3345] = {.lex_state = 271, .external_lex_state = 42},
  [3346] = {.lex_state = 273, .external_lex_state = 69},
  [3347] = {.lex_state = 271, .external_lex_state = 42},
  [3348] = {.lex_state = 273, .external_lex_state = 69},
  [3349] = {.lex_state = 273, .external_lex_state = 69},
  [3350] = {.lex_state = 271, .external_lex_state = 60},
  [3351] = {.lex_state = 271, .external_lex_state = 42},
  [3352] = {.lex_state = 274, .external_lex_state = 60},
  [3353] = {.lex_state = 274, .external_lex_state = 60},
  [3354] = {.lex_state = 274, .external_lex_state = 60},
  [3355] = {.lex_state = 273, .external_lex_state = 69},
  [3356] = {.lex_state = 271, .external_lex_state = 42},
  [3357] = {.lex_state = 273, .external_lex_state = 69},
  [3358] = {.lex_state = 271, .external_lex_state = 42},
  [3359] = {.lex_state = 273, .external_lex_state = 60},
  [3360] = {.lex_state = 236, .external_lex_state = 70},
  [3361] = {.lex_state = 273, .external_lex_state = 60},
  [3362] = {.lex_state = 273, .external_lex_state = 69},
  [3363] = {.lex_state = 273, .external_lex_state = 69},
  [3364] = {.lex_state = 273, .external_lex_state = 69},
  [3365] = {.lex_state = 273, .external_lex_state = 60},
  [3366] = {.lex_state = 273, .external_lex_state = 69},
  [3367] = {.lex_state = 273, .external_lex_state = 60},
  [3368] = {.lex_state = 273, .external_lex_state = 60},
  [3369] = {.lex_state = 273, .external_lex_state = 69},
  [3370] = {.lex_state = 273, .external_lex_state = 69},
  [3371] = {.lex_state = 273, .external_lex_state = 69},
  [3372] = {.lex_state = 273, .external_lex_state = 69},
  [3373] = {.lex_state = 273, .external_lex_state = 69},
  [3374] = {.lex_state = 273, .external_lex_state = 69},
  [3375] = {.lex_state = 273, .external_lex_state = 60},
  [3376] = {.lex_state = 236, .external_lex_state = 70},
  [3377] = {.lex_state = 298, .external_lex_state = 47},
  [3378] = {.lex_state = 298, .external_lex_state = 47},
  [3379] = {.lex_state = 273, .external_lex_state = 69},
  [3380] = {.lex_state = 274, .external_lex_state = 60},
  [3381] = {.lex_state = 274, .external_lex_state = 60},
  [3382] = {.lex_state = 298, .external_lex_state = 47},
  [3383] = {.lex_state = 298, .external_lex_state = 47},
  [3384] = {.lex_state = 274, .external_lex_state = 60},
  [3385] = {.lex_state = 298, .external_lex_state = 47},
  [3386] = {.lex_state = 298, .external_lex_state = 47},
  [3387] = {.lex_state = 274, .external_lex_state = 60},
  [3388] = {.lex_state = 298, .external_lex_state = 47},
  [3389] = {.lex_state = 298, .external_lex_state = 47},
  [3390] = {.lex_state = 274, .external_lex_state = 60},
  [3391] = {.lex_state = 273, .external_lex_state = 69},
  [3392] = {.lex_state = 274, .external_lex_state = 60},
  [3393] = {.lex_state = 236, .external_lex_state = 70},
  [3394] = {.lex_state = 274, .external_lex_state = 60},
  [3395] = {.lex_state = 274, .external_lex_state = 60},
  [3396] = {.lex_state = 273, .external_lex_state = 69},
  [3397] = {.lex_state = 271, .external_lex_state = 42},
  [3398] = {.lex_state = 273, .external_lex_state = 69},
  [3399] = {.lex_state = 274, .external_lex_state = 60},
  [3400] = {.lex_state = 273, .external_lex_state = 69},
  [3401] = {.lex_state = 271, .external_lex_state = 60},
  [3402] = {.lex_state = 274, .external_lex_state = 60},
  [3403] = {.lex_state = 274, .external_lex_state = 60},
  [3404] = {.lex_state = 274, .external_lex_state = 60},
  [3405] = {.lex_state = 274, .external_lex_state = 60},
  [3406] = {.lex_state = 274, .external_lex_state = 60},
  [3407] = {.lex_state = 298, .external_lex_state = 47},
  [3408] = {.lex_state = 273, .external_lex_state = 60},
  [3409] = {.lex_state = 273, .external_lex_state = 69},
  [3410] = {.lex_state = 273, .external_lex_state = 69},
  [3411] = {.lex_state = 298, .external_lex_state = 47},
  [3412] = {.lex_state = 273, .external_lex_state = 69},
  [3413] = {.lex_state = 273, .external_lex_state = 69},
  [3414] = {.lex_state = 199, .external_lex_state = 66},
  [3415] = {.lex_state = 236, .external_lex_state = 70},
  [3416] = {.lex_state = 298, .external_lex_state = 47},
  [3417] = {.lex_state = 199, .external_lex_state = 66},
  [3418] = {.lex_state = 236, .external_lex_state = 70},
  [3419] = {.lex_state = 298, .external_lex_state = 47},
  [3420] = {.lex_state = 298, .external_lex_state = 47},
  [3421] = {.lex_state = 273, .external_lex_state = 69},
  [3422] = {.lex_state = 298, .external_lex_state = 47},
  [3423] = {.lex_state = 298, .external_lex_state = 47},
  [3424] = {.lex_state = 298, .external_lex_state = 47},
  [3425] = {.lex_state = 273, .external_lex_state = 69},
  [3426] = {.lex_state = 273, .external_lex_state = 69},
  [3427] = {.lex_state = 273, .external_lex_state = 60},
  [3428] = {.lex_state = 273, .external_lex_state = 69},
  [3429] = {.lex_state = 273, .external_lex_state = 69},
  [3430] = {.lex_state = 273, .external_lex_state = 69},
  [3431] = {.lex_state = 273, .external_lex_state = 69},
  [3432] = {.lex_state = 273, .external_lex_state = 69},
  [3433] = {.lex_state = 273, .external_lex_state = 69},
  [3434] = {.lex_state = 273, .external_lex_state = 69},
  [3435] = {.lex_state = 273, .external_lex_state = 69},
  [3436] = {.lex_state = 273, .external_lex_state = 69},
  [3437] = {.lex_state = 298, .external_lex_state = 47},
  [3438] = {.lex_state = 298, .external_lex_state = 47},
  [3439] = {.lex_state = 273, .external_lex_state = 60},
  [3440] = {.lex_state = 298, .external_lex_state = 47},
  [3441] = {.lex_state = 274, .external_lex_state = 60},
  [3442] = {.lex_state = 274, .external_lex_state = 60},
  [3443] = {.lex_state = 274, .external_lex_state = 60},
  [3444] = {.lex_state = 274, .external_lex_state = 60},
  [3445] = {.lex_state = 273, .external_lex_state = 60},
  [3446] = {.lex_state = 273, .external_lex_state = 69},
  [3447] = {.lex_state = 273, .external_lex_state = 69},
  [3448] = {.lex_state = 273, .external_lex_state = 60},
  [3449] = {.lex_state = 273, .external_lex_state = 69},
  [3450] = {.lex_state = 273, .external_lex_state = 69},
  [3451] = {.lex_state = 298, .external_lex_state = 47},
  [3452] = {.lex_state = 298, .external_lex_state = 47},
  [3453] = {.lex_state = 273, .external_lex_state = 60},
  [3454] = {.lex_state = 274, .external_lex_state = 60},
  [3455] = {.lex_state = 274, .external_lex_state = 60},
  [3456] = {.lex_state = 297, .external_lex_state = 71},
  [3457] = {.lex_state = 228, .external_lex_state = 72},
  [3458] = {.lex_state = 237, .external_lex_state = 73},
  [3459] = {.lex_state = 297, .external_lex_state = 71},
  [3460] = {.lex_state = 237, .external_lex_state = 73},
  [3461] = {.lex_state = 297, .external_lex_state = 71},
  [3462] = {.lex_state = 310, .external_lex_state = 74},
  [3463] = {.lex_state = 532, .external_lex_state = 66},
  [3464] = {.lex_state = 532, .external_lex_state = 66},
  [3465] = {.lex_state = 297, .external_lex_state = 71},
  [3466] = {.lex_state = 297, .external_lex_state = 71},
  [3467] = {.lex_state = 297, .external_lex_state = 71},
  [3468] = {.lex_state = 297, .external_lex_state = 71},
  [3469] = {.lex_state = 213, .external_lex_state = 26},
  [3470] = {.lex_state = 310, .external_lex_state = 74},
  [3471] = {.lex_state = 297, .external_lex_state = 71},
  [3472] = {.lex_state = 297, .external_lex_state = 71},
  [3473] = {.lex_state = 213, .external_lex_state = 26},
  [3474] = {.lex_state = 228, .external_lex_state = 72},
  [3475] = {.lex_state = 237, .external_lex_state = 73},
  [3476] = {.lex_state = 297, .external_lex_state = 71},
  [3477] = {.lex_state = 298, .external_lex_state = 47},
  [3478] = {.lex_state = 237, .external_lex_state = 73},
  [3479] = {.lex_state = 297, .external_lex_state = 71},
  [3480] = {.lex_state = 297, .external_lex_state = 71},
  [3481] = {.lex_state = 297, .external_lex_state = 71},
  [3482] = {.lex_state = 532, .external_lex_state = 66},
  [3483] = {.lex_state = 297, .external_lex_state = 71},
  [3484] = {.lex_state = 237, .external_lex_state = 73},
  [3485] = {.lex_state = 532, .external_lex_state = 66},
  [3486] = {.lex_state = 310, .external_lex_state = 74},
  [3487] = {.lex_state = 316, .external_lex_state = 75},
  [3488] = {.lex_state = 297, .external_lex_state = 71},
  [3489] = {.lex_state = 316, .external_lex_state = 75},
  [3490] = {.lex_state = 297, .external_lex_state = 71},
  [3491] = {.lex_state = 319, .external_lex_state = 76},
  [3492] = {.lex_state = 297, .external_lex_state = 71},
  [3493] = {.lex_state = 316, .external_lex_state = 75},
  [3494] = {.lex_state = 297, .external_lex_state = 71},
  [3495] = {.lex_state = 316, .external_lex_state = 75},
  [3496] = {.lex_state = 316, .external_lex_state = 75},
  [3497] = {.lex_state = 316, .external_lex_state = 75},
  [3498] = {.lex_state = 297, .external_lex_state = 71},
  [3499] = {.lex_state = 316, .external_lex_state = 75},
  [3500] = {.lex_state = 316, .external_lex_state = 75},
  [3501] = {.lex_state = 297, .external_lex_state = 71},
  [3502] = {.lex_state = 316, .external_lex_state = 75},
  [3503] = {.lex_state = 316, .external_lex_state = 75},
  [3504] = {.lex_state = 297, .external_lex_state = 71},
  [3505] = {.lex_state = 297, .external_lex_state = 71},
  [3506] = {.lex_state = 297, .external_lex_state = 71},
  [3507] = {.lex_state = 316, .external_lex_state = 75},
  [3508] = {.lex_state = 297, .external_lex_state = 71},
  [3509] = {.lex_state = 316, .external_lex_state = 75},
  [3510] = {.lex_state = 297, .external_lex_state = 71},
  [3511] = {.lex_state = 213, .external_lex_state = 26},
  [3512] = {.lex_state = 297, .external_lex_state = 71},
  [3513] = {.lex_state = 297, .external_lex_state = 71},
  [3514] = {.lex_state = 316, .external_lex_state = 75},
  [3515] = {.lex_state = 297, .external_lex_state = 71},
  [3516] = {.lex_state = 297, .external_lex_state = 71},
  [3517] = {.lex_state = 316, .external_lex_state = 75},
  [3518] = {.lex_state = 316, .external_lex_state = 75},
  [3519] = {.lex_state = 316, .external_lex_state = 75},
  [3520] = {.lex_state = 316, .external_lex_state = 75},
  [3521] = {.lex_state = 316, .external_lex_state = 75},
  [3522] = {.lex_state = 297, .external_lex_state = 71},
  [3523] = {.lex_state = 316, .external_lex_state = 75},
  [3524] = {.lex_state = 316, .external_lex_state = 75},
  [3525] = {.lex_state = 297, .external_lex_state = 71},
  [3526] = {.lex_state = 316, .external_lex_state = 75},
  [3527] = {.lex_state = 316, .external_lex_state = 75},
  [3528] = {.lex_state = 297, .external_lex_state = 71},
  [3529] = {.lex_state = 297, .external_lex_state = 71},
  [3530] = {.lex_state = 316, .external_lex_state = 75},
  [3531] = {.lex_state = 319, .external_lex_state = 76},
  [3532] = {.lex_state = 319, .external_lex_state = 76},
  [3533] = {.lex_state = 319, .external_lex_state = 76},
  [3534] = {.lex_state = 319, .external_lex_state = 76},
  [3535] = {.lex_state = 202, .external_lex_state = 66},
  [3536] = {.lex_state = 296, .external_lex_state = 36},
  [3537] = {.lex_state = 282, .external_lex_state = 77},
  [3538] = {.lex_state = 282, .external_lex_state = 77},
  [3539] = {.lex_state = 296, .external_lex_state = 36},
  [3540] = {.lex_state = 202, .external_lex_state = 66},
  [3541] = {.lex_state = 296, .external_lex_state = 36},
  [3542] = {.lex_state = 296, .external_lex_state = 36},
  [3543] = {.lex_state = 296, .external_lex_state = 36},
  [3544] = {.lex_state = 296, .external_lex_state = 36},
  [3545] = {.lex_state = 296, .external_lex_state = 36},
  [3546] = {.lex_state = 298, .external_lex_state = 78},
  [3547] = {.lex_state = 296, .external_lex_state = 36},
  [3548] = {.lex_state = 298, .external_lex_state = 78},
  [3549] = {.lex_state = 298, .external_lex_state = 78},
  [3550] = {.lex_state = 298, .external_lex_state = 78},
  [3551] = {.lex_state = 298, .external_lex_state = 78},
  [3552] = {.lex_state = 296, .external_lex_state = 36},
  [3553] = {.lex_state = 298, .external_lex_state = 78},
  [3554] = {.lex_state = 298, .external_lex_state = 78},
  [3555] = {.lex_state = 298, .external_lex_state = 79},
  [3556] = {.lex_state = 296, .external_lex_state = 36},
  [3557] = {.lex_state = 298, .external_lex_state = 78},
  [3558] = {.lex_state = 298, .external_lex_state = 78},
  [3559] = {.lex_state = 254, .external_lex_state = 80},
  [3560] = {.lex_state = 254, .external_lex_state = 80},
  [3561] = {.lex_state = 298, .external_lex_state = 78},
  [3562] = {.lex_state = 230, .external_lex_state = 72},
  [3563] = {.lex_state = 298, .external_lex_state = 78},
  [3564] = {.lex_state = 298, .external_lex_state = 78},
  [3565] = {.lex_state = 198, .external_lex_state = 73},
  [3566] = {.lex_state = 298, .external_lex_state = 78},
  [3567] = {.lex_state = 198, .external_lex_state = 73},
  [3568] = {.lex_state = 298, .external_lex_state = 78},
  [3569] = {.lex_state = 296, .external_lex_state = 36},
  [3570] = {.lex_state = 198, .external_lex_state = 73},
  [3571] = {.lex_state = 198, .external_lex_state = 73},
  [3572] = {.lex_state = 284, .external_lex_state = 77},
  [3573] = {.lex_state = 296, .external_lex_state = 36},
  [3574] = {.lex_state = 298, .external_lex_state = 79},
  [3575] = {.lex_state = 298, .external_lex_state = 78},
  [3576] = {.lex_state = 298, .external_lex_state = 78},
  [3577] = {.lex_state = 298, .external_lex_state = 78},
  [3578] = {.lex_state = 296, .external_lex_state = 36},
  [3579] = {.lex_state = 205, .external_lex_state = 66},
  [3580] = {.lex_state = 283, .external_lex_state = 81},
  [3581] = {.lex_state = 296, .external_lex_state = 36},
  [3582] = {.lex_state = 283, .external_lex_state = 81},
  [3583] = {.lex_state = 198, .external_lex_state = 73},
  [3584] = {.lex_state = 230, .external_lex_state = 72},
  [3585] = {.lex_state = 198, .external_lex_state = 73},
  [3586] = {.lex_state = 298, .external_lex_state = 78},
  [3587] = {.lex_state = 298, .external_lex_state = 78},
  [3588] = {.lex_state = 298, .external_lex_state = 78},
  [3589] = {.lex_state = 298, .external_lex_state = 78},
  [3590] = {.lex_state = 298, .external_lex_state = 78},
  [3591] = {.lex_state = 198, .external_lex_state = 73},
  [3592] = {.lex_state = 298, .external_lex_state = 78},
  [3593] = {.lex_state = 298, .external_lex_state = 78},
  [3594] = {.lex_state = 298, .external_lex_state = 78},
  [3595] = {.lex_state = 298, .external_lex_state = 78},
  [3596] = {.lex_state = 198, .external_lex_state = 73},
  [3597] = {.lex_state = 296, .external_lex_state = 36},
  [3598] = {.lex_state = 284, .external_lex_state = 77},
  [3599] = {.lex_state = 298, .external_lex_state = 78},
  [3600] = {.lex_state = 205, .external_lex_state = 66},
  [3601] = {.lex_state = 296, .external_lex_state = 36},
  [3602] = {.lex_state = 298, .external_lex_state = 78},
  [3603] = {.lex_state = 198, .external_lex_state = 73},
  [3604] = {.lex_state = 298, .external_lex_state = 78},
  [3605] = {.lex_state = 198, .external_lex_state = 73},
  [3606] = {.lex_state = 298, .external_lex_state = 78},
  [3607] = {.lex_state = 205, .external_lex_state = 66},
  [3608] = {.lex_state = 198, .external_lex_state = 73},
  [3609] = {.lex_state = 298, .external_lex_state = 79},
  [3610] = {.lex_state = 298, .external_lex_state = 78},
  [3611] = {.lex_state = 298, .external_lex_state = 78},
  [3612] = {.lex_state = 205, .external_lex_state = 66},
  [3613] = {.lex_state = 298, .external_lex_state = 78},
  [3614] = {.lex_state = 298, .external_lex_state = 78},
  [3615] = {.lex_state = 317, .external_lex_state = 82},
  [3616] = {.lex_state = 317, .external_lex_state = 82},
  [3617] = {.lex_state = 298, .external_lex_state = 78},
  [3618] = {.lex_state = 298, .external_lex_state = 78},
  [3619] = {.lex_state = 285, .external_lex_state = 81},
  [3620] = {.lex_state = 298, .external_lex_state = 83},
  [3621] = {.lex_state = 298, .external_lex_state = 78},
  [3622] = {.lex_state = 298, .external_lex_state = 78},
  [3623] = {.lex_state = 298, .external_lex_state = 78},
  [3624] = {.lex_state = 298, .external_lex_state = 78},
  [3625] = {.lex_state = 298, .external_lex_state = 78},
  [3626] = {.lex_state = 298, .external_lex_state = 78},
  [3627] = {.lex_state = 298, .external_lex_state = 78},
  [3628] = {.lex_state = 319, .external_lex_state = 84},
  [3629] = {.lex_state = 298, .external_lex_state = 78},
  [3630] = {.lex_state = 298, .external_lex_state = 78},
  [3631] = {.lex_state = 298, .external_lex_state = 78},
  [3632] = {.lex_state = 298, .external_lex_state = 78},
  [3633] = {.lex_state = 298, .external_lex_state = 78},
  [3634] = {.lex_state = 298, .external_lex_state = 78},
  [3635] = {.lex_state = 298, .external_lex_state = 78},
  [3636] = {.lex_state = 285, .external_lex_state = 81},
  [3637] = {.lex_state = 298, .external_lex_state = 78},
  [3638] = {.lex_state = 298, .external_lex_state = 78},
  [3639] = {.lex_state = 298, .external_lex_state = 78},
  [3640] = {.lex_state = 298, .external_lex_state = 78},
  [3641] = {.lex_state = 298, .external_lex_state = 78},
  [3642] = {.lex_state = 298, .external_lex_state = 78},
  [3643] = {.lex_state = 317, .external_lex_state = 82},
  [3644] = {.lex_state = 317, .external_lex_state = 82},
  [3645] = {.lex_state = 298, .external_lex_state = 78},
  [3646] = {.lex_state = 298, .external_lex_state = 78},
  [3647] = {.lex_state = 298, .external_lex_state = 78},
  [3648] = {.lex_state = 298, .external_lex_state = 78},
  [3649] = {.lex_state = 298, .external_lex_state = 78},
  [3650] = {.lex_state = 298, .external_lex_state = 78},
  [3651] = {.lex_state = 298, .external_lex_state = 78},
  [3652] = {.lex_state = 298, .external_lex_state = 78},
  [3653] = {.lex_state = 298, .external_lex_state = 78},
  [3654] = {.lex_state = 298, .external_lex_state = 78},
  [3655] = {.lex_state = 298, .external_lex_state = 78},
  [3656] = {.lex_state = 317, .external_lex_state = 82},
  [3657] = {.lex_state = 317, .external_lex_state = 82},
  [3658] = {.lex_state = 317, .external_lex_state = 82},
  [3659] = {.lex_state = 298, .external_lex_state = 78},
  [3660] = {.lex_state = 298, .external_lex_state = 83},
  [3661] = {.lex_state = 298, .external_lex_state = 78},
  [3662] = {.lex_state = 298, .external_lex_state = 78},
  [3663] = {.lex_state = 317, .external_lex_state = 82},
  [3664] = {.lex_state = 285, .external_lex_state = 81},
  [3665] = {.lex_state = 298, .external_lex_state = 78},
  [3666] = {.lex_state = 285, .external_lex_state = 81},
  [3667] = {.lex_state = 298, .external_lex_state = 78},
  [3668] = {.lex_state = 317, .external_lex_state = 82},
  [3669] = {.lex_state = 298, .external_lex_state = 78},
  [3670] = {.lex_state = 298, .external_lex_state = 78},
  [3671] = {.lex_state = 317, .external_lex_state = 82},
  [3672] = {.lex_state = 298, .external_lex_state = 83},
  [3673] = {.lex_state = 298, .external_lex_state = 78},
  [3674] = {.lex_state = 298, .external_lex_state = 78},
  [3675] = {.lex_state = 298, .external_lex_state = 78},
  [3676] = {.lex_state = 298, .external_lex_state = 78},
  [3677] = {.lex_state = 298, .external_lex_state = 78},
  [3678] = {.lex_state = 298, .external_lex_state = 78},
  [3679] = {.lex_state = 298, .external_lex_state = 78},
  [3680] = {.lex_state = 298, .external_lex_state = 78},
  [3681] = {.lex_state = 298, .external_lex_state = 78},
  [3682] = {.lex_state = 298, .external_lex_state = 83},
  [3683] = {.lex_state = 317, .external_lex_state = 82},
  [3684] = {.lex_state = 298, .external_lex_state = 78},
  [3685] = {.lex_state = 317, .external_lex_state = 82},
  [3686] = {.lex_state = 298, .external_lex_state = 83},
  [3687] = {.lex_state = 298, .external_lex_state = 78},
  [3688] = {.lex_state = 298, .external_lex_state = 78},
  [3689] = {.lex_state = 298, .external_lex_state = 78},
  [3690] = {.lex_state = 298, .external_lex_state = 78},
  [3691] = {.lex_state = 298, .external_lex_state = 78},
  [3692] = {.lex_state = 298, .external_lex_state = 78},
  [3693] = {.lex_state = 298, .external_lex_state = 78},
  [3694] = {.lex_state = 298, .external_lex_state = 78},
  [3695] = {.lex_state = 298, .external_lex_state = 83},
  [3696] = {.lex_state = 298, .external_lex_state = 78},
  [3697] = {.lex_state = 298, .external_lex_state = 83},
  [3698] = {.lex_state = 316, .external_lex_state = 85},
  [3699] = {.lex_state = 289, .external_lex_state = 86},
  [3700] = {.lex_state = 316, .external_lex_state = 85},
  [3701] = {.lex_state = 316, .external_lex_state = 85},
  [3702] = {.lex_state = 316, .external_lex_state = 85},
  [3703] = {.lex_state = 298, .external_lex_state = 78},
  [3704] = {.lex_state = 316, .external_lex_state = 85},
  [3705] = {.lex_state = 289, .external_lex_state = 86},
  [3706] = {.lex_state = 298, .external_lex_state = 78},
  [3707] = {.lex_state = 316, .external_lex_state = 85},
  [3708] = {.lex_state = 289, .external_lex_state = 86},
  [3709] = {.lex_state = 289, .external_lex_state = 86},
  [3710] = {.lex_state = 289, .external_lex_state = 86},
  [3711] = {.lex_state = 316, .external_lex_state = 85},
  [3712] = {.lex_state = 316, .external_lex_state = 85},
  [3713] = {.lex_state = 289, .external_lex_state = 86},
  [3714] = {.lex_state = 289, .external_lex_state = 86},
  [3715] = {.lex_state = 298, .external_lex_state = 78},
  [3716] = {.lex_state = 298, .external_lex_state = 78},
  [3717] = {.lex_state = 316, .external_lex_state = 85},
  [3718] = {.lex_state = 298, .external_lex_state = 78},
  [3719] = {.lex_state = 316, .external_lex_state = 85},
  [3720] = {.lex_state = 298, .external_lex_state = 78},
  [3721] = {.lex_state = 298, .external_lex_state = 78},
  [3722] = {.lex_state = 289, .external_lex_state = 86},
  [3723] = {.lex_state = 316, .external_lex_state = 85},
  [3724] = {.lex_state = 289, .external_lex_state = 86},
  [3725] = {.lex_state = 298, .external_lex_state = 78},
  [3726] = {.lex_state = 289, .external_lex_state = 86},
  [3727] = {.lex_state = 316, .external_lex_state = 85},
  [3728] = {.lex_state = 289, .external_lex_state = 86},
  [3729] = {.lex_state = 289, .external_lex_state = 86},
  [3730] = {.lex_state = 289, .external_lex_state = 86},
  [3731] = {.lex_state = 289, .external_lex_state = 86},
  [3732] = {.lex_state = 316, .external_lex_state = 85},
  [3733] = {.lex_state = 316, .external_lex_state = 85},
  [3734] = {.lex_state = 298, .external_lex_state = 78},
  [3735] = {.lex_state = 316, .external_lex_state = 85},
  [3736] = {.lex_state = 298, .external_lex_state = 78},
  [3737] = {.lex_state = 316, .external_lex_state = 85},
  [3738] = {.lex_state = 316, .external_lex_state = 85},
  [3739] = {.lex_state = 298, .external_lex_state = 78},
  [3740] = {.lex_state = 298, .external_lex_state = 78},
  [3741] = {.lex_state = 289, .external_lex_state = 86},
  [3742] = {.lex_state = 289, .external_lex_state = 86},
  [3743] = {.lex_state = 289, .external_lex_state = 86},
  [3744] = {.lex_state = 289, .external_lex_state = 86},
  [3745] = {.lex_state = 316, .external_lex_state = 85},
  [3746] = {.lex_state = 316, .external_lex_state = 85},
  [3747] = {.lex_state = 316, .external_lex_state = 85},
  [3748] = {.lex_state = 289, .external_lex_state = 86},
  [3749] = {.lex_state = 289, .external_lex_state = 86},
  [3750] = {.lex_state = 316, .external_lex_state = 85},
  [3751] = {.lex_state = 289, .external_lex_state = 86},
  [3752] = {.lex_state = 316, .external_lex_state = 85},
  [3753] = {.lex_state = 289, .external_lex_state = 86},
  [3754] = {.lex_state = 289, .external_lex_state = 86},
  [3755] = {.lex_state = 289, .external_lex_state = 86},
  [3756] = {.lex_state = 316, .external_lex_state = 85},
  [3757] = {.lex_state = 316, .external_lex_state = 85},
  [3758] = {.lex_state = 316, .external_lex_state = 85},
  [3759] = {.lex_state = 289, .external_lex_state = 86},
  [3760] = {.lex_state = 316, .external_lex_state = 85},
  [3761] = {.lex_state = 298, .external_lex_state = 78},
  [3762] = {.lex_state = 298, .external_lex_state = 78},
  [3763] = {.lex_state = 298, .external_lex_state = 78},
  [3764] = {.lex_state = 289, .external_lex_state = 86},
  [3765] = {.lex_state = 298, .external_lex_state = 78},
  [3766] = {.lex_state = 316, .external_lex_state = 85},
  [3767] = {.lex_state = 316, .external_lex_state = 85},
  [3768] = {.lex_state = 316, .external_lex_state = 85},
  [3769] = {.lex_state = 298, .external_lex_state = 78},
  [3770] = {.lex_state = 316, .external_lex_state = 85},
  [3771] = {.lex_state = 316, .external_lex_state = 85},
  [3772] = {.lex_state = 298, .external_lex_state = 78},
  [3773] = {.lex_state = 316, .external_lex_state = 85},
  [3774] = {.lex_state = 316, .external_lex_state = 85},
  [3775] = {.lex_state = 298, .external_lex_state = 78},
  [3776] = {.lex_state = 316, .external_lex_state = 85},
  [3777] = {.lex_state = 298, .external_lex_state = 78},
  [3778] = {.lex_state = 316, .external_lex_state = 85},
  [3779] = {.lex_state = 298, .external_lex_state = 78},
  [3780] = {.lex_state = 298, .external_lex_state = 78},
  [3781] = {.lex_state = 316, .external_lex_state = 85},
  [3782] = {.lex_state = 298, .external_lex_state = 78},
  [3783] = {.lex_state = 289, .external_lex_state = 86},
  [3784] = {.lex_state = 289, .external_lex_state = 86},
  [3785] = {.lex_state = 316, .external_lex_state = 85},
  [3786] = {.lex_state = 289, .external_lex_state = 86},
  [3787] = {.lex_state = 298, .external_lex_state = 78},
  [3788] = {.lex_state = 316, .external_lex_state = 85},
  [3789] = {.lex_state = 289, .external_lex_state = 86},
  [3790] = {.lex_state = 316, .external_lex_state = 85},
  [3791] = {.lex_state = 316, .external_lex_state = 85},
  [3792] = {.lex_state = 316, .external_lex_state = 85},
  [3793] = {.lex_state = 289, .external_lex_state = 86},
  [3794] = {.lex_state = 316, .external_lex_state = 85},
  [3795] = {.lex_state = 316, .external_lex_state = 85},
  [3796] = {.lex_state = 316, .external_lex_state = 85},
  [3797] = {.lex_state = 289, .external_lex_state = 86},
  [3798] = {.lex_state = 316, .external_lex_state = 85},
  [3799] = {.lex_state = 298, .external_lex_state = 78},
  [3800] = {.lex_state = 289, .external_lex_state = 86},
  [3801] = {.lex_state = 298, .external_lex_state = 78},
  [3802] = {.lex_state = 316, .external_lex_state = 85},
  [3803] = {.lex_state = 316, .external_lex_state = 85},
  [3804] = {.lex_state = 289, .external_lex_state = 86},
  [3805] = {.lex_state = 298, .external_lex_state = 78},
  [3806] = {.lex_state = 316, .external_lex_state = 85},
  [3807] = {.lex_state = 298, .external_lex_state = 78},
  [3808] = {.lex_state = 298, .external_lex_state = 78},
  [3809] = {.lex_state = 289, .external_lex_state = 86},
  [3810] = {.lex_state = 316, .external_lex_state = 85},
  [3811] = {.lex_state = 316, .external_lex_state = 85},
  [3812] = {.lex_state = 289, .external_lex_state = 86},
  [3813] = {.lex_state = 316, .external_lex_state = 85},
  [3814] = {.lex_state = 316, .external_lex_state = 85},
  [3815] = {.lex_state = 316, .external_lex_state = 85},
  [3816] = {.lex_state = 289, .external_lex_state = 86},
  [3817] = {.lex_state = 289, .external_lex_state = 86},
  [3818] = {.lex_state = 289, .external_lex_state = 86},
  [3819] = {.lex_state = 316, .external_lex_state = 85},
  [3820] = {.lex_state = 289, .external_lex_state = 86},
  [3821] = {.lex_state = 298, .external_lex_state = 78},
  [3822] = {.lex_state = 298, .external_lex_state = 78},
  [3823] = {.lex_state = 316, .external_lex_state = 85},
  [3824] = {.lex_state = 289, .external_lex_state = 86},
  [3825] = {.lex_state = 289, .external_lex_state = 86},
  [3826] = {.lex_state = 316, .external_lex_state = 85},
  [3827] = {.lex_state = 298, .external_lex_state = 78},
  [3828] = {.lex_state = 298, .external_lex_state = 78},
  [3829] = {.lex_state = 316, .external_lex_state = 85},
  [3830] = {.lex_state = 289, .external_lex_state = 86},
  [3831] = {.lex_state = 289, .external_lex_state = 86},
  [3832] = {.lex_state = 298, .external_lex_state = 78},
  [3833] = {.lex_state = 298, .external_lex_state = 78},
  [3834] = {.lex_state = 316, .external_lex_state = 85},
  [3835] = {.lex_state = 289, .external_lex_state = 86},
  [3836] = {.lex_state = 298, .external_lex_state = 78},
  [3837] = {.lex_state = 316, .external_lex_state = 85},
  [3838] = {.lex_state = 298, .external_lex_state = 78},
  [3839] = {.lex_state = 316, .external_lex_state = 85},
  [3840] = {.lex_state = 298, .external_lex_state = 78},
  [3841] = {.lex_state = 316, .external_lex_state = 85},
  [3842] = {.lex_state = 289, .external_lex_state = 86},
  [3843] = {.lex_state = 316, .external_lex_state = 85},
  [3844] = {.lex_state = 316, .external_lex_state = 85},
  [3845] = {.lex_state = 316, .external_lex_state = 85},
  [3846] = {.lex_state = 316, .external_lex_state = 85},
  [3847] = {.lex_state = 316, .external_lex_state = 85},
  [3848] = {.lex_state = 316, .external_lex_state = 85},
  [3849] = {.lex_state = 289, .external_lex_state = 86},
  [3850] = {.lex_state = 289, .external_lex_state = 86},
  [3851] = {.lex_state = 316, .external_lex_state = 85},
  [3852] = {.lex_state = 316, .external_lex_state = 85},
  [3853] = {.lex_state = 316, .external_lex_state = 85},
  [3854] = {.lex_state = 316, .external_lex_state = 85},
  [3855] = {.lex_state = 289, .external_lex_state = 86},
  [3856] = {.lex_state = 316, .external_lex_state = 85},
  [3857] = {.lex_state = 289, .external_lex_state = 86},
  [3858] = {.lex_state = 316, .external_lex_state = 85},
  [3859] = {.lex_state = 298, .external_lex_state = 78},
  [3860] = {.lex_state = 316, .external_lex_state = 85},
  [3861] = {.lex_state = 316, .external_lex_state = 85},
  [3862] = {.lex_state = 289, .external_lex_state = 86},
  [3863] = {.lex_state = 289, .external_lex_state = 86},
  [3864] = {.lex_state = 316, .external_lex_state = 85},
  [3865] = {.lex_state = 316, .external_lex_state = 85},
  [3866] = {.lex_state = 316, .external_lex_state = 85},
  [3867] = {.lex_state = 316, .external_lex_state = 85},
  [3868] = {.lex_state = 316, .external_lex_state = 85},
  [3869] = {.lex_state = 316, .external_lex_state = 85},
  [3870] = {.lex_state = 316, .external_lex_state = 85},
  [3871] = {.lex_state = 316, .external_lex_state = 85},
  [3872] = {.lex_state = 316, .external_lex_state = 85},
  [3873] = {.lex_state = 316, .external_lex_state = 85},
  [3874] = {.lex_state = 289, .external_lex_state = 86},
  [3875] = {.lex_state = 289, .external_lex_state = 86},
  [3876] = {.lex_state = 289, .external_lex_state = 86},
  [3877] = {.lex_state = 316, .external_lex_state = 85},
  [3878] = {.lex_state = 316, .external_lex_state = 85},
  [3879] = {.lex_state = 289, .external_lex_state = 86},
  [3880] = {.lex_state = 316, .external_lex_state = 85},
  [3881] = {.lex_state = 316, .external_lex_state = 85},
  [3882] = {.lex_state = 316, .external_lex_state = 85},
  [3883] = {.lex_state = 316, .external_lex_state = 85},
  [3884] = {.lex_state = 316, .external_lex_state = 85},
  [3885] = {.lex_state = 316, .external_lex_state = 85},
  [3886] = {.lex_state = 316, .external_lex_state = 85},
  [3887] = {.lex_state = 316, .external_lex_state = 85},
  [3888] = {.lex_state = 289, .external_lex_state = 86},
  [3889] = {.lex_state = 316, .external_lex_state = 85},
  [3890] = {.lex_state = 316, .external_lex_state = 85},
  [3891] = {.lex_state = 316, .external_lex_state = 85},
  [3892] = {.lex_state = 289, .external_lex_state = 86},
  [3893] = {.lex_state = 289, .external_lex_state = 86},
  [3894] = {.lex_state = 316, .external_lex_state = 85},
  [3895] = {.lex_state = 316, .external_lex_state = 85},
  [3896] = {.lex_state = 316, .external_lex_state = 85},
  [3897] = {.lex_state = 316, .external_lex_state = 85},
  [3898] = {.lex_state = 298, .external_lex_state = 78},
  [3899] = {.lex_state = 316, .external_lex_state = 85},
  [3900] = {.lex_state = 316, .external_lex_state = 85},
  [3901] = {.lex_state = 316, .external_lex_state = 85},
  [3902] = {.lex_state = 298, .external_lex_state = 78},
  [3903] = {.lex_state = 316, .external_lex_state = 85},
  [3904] = {.lex_state = 316, .external_lex_state = 85},
  [3905] = {.lex_state = 316, .external_lex_state = 85},
  [3906] = {.lex_state = 316, .external_lex_state = 85},
  [3907] = {.lex_state = 298, .external_lex_state = 78},
  [3908] = {.lex_state = 316, .external_lex_state = 85},
  [3909] = {.lex_state = 316, .external_lex_state = 85},
  [3910] = {.lex_state = 316, .external_lex_state = 85},
  [3911] = {.lex_state = 316, .external_lex_state = 85},
  [3912] = {.lex_state = 316, .external_lex_state = 85},
  [3913] = {.lex_state = 316, .external_lex_state = 85},
  [3914] = {.lex_state = 316, .external_lex_state = 85},
  [3915] = {.lex_state = 289, .external_lex_state = 86},
  [3916] = {.lex_state = 289, .external_lex_state = 86},
  [3917] = {.lex_state = 316, .external_lex_state = 85},
  [3918] = {.lex_state = 316, .external_lex_state = 85},
  [3919] = {.lex_state = 289, .external_lex_state = 86},
  [3920] = {.lex_state = 289, .external_lex_state = 86},
  [3921] = {.lex_state = 316, .external_lex_state = 85},
  [3922] = {.lex_state = 316, .external_lex_state = 85},
  [3923] = {.lex_state = 316, .external_lex_state = 85},
  [3924] = {.lex_state = 316, .external_lex_state = 85},
  [3925] = {.lex_state = 316, .external_lex_state = 85},
  [3926] = {.lex_state = 316, .external_lex_state = 85},
  [3927] = {.lex_state = 316, .external_lex_state = 85},
  [3928] = {.lex_state = 316, .external_lex_state = 85},
  [3929] = {.lex_state = 316, .external_lex_state = 85},
  [3930] = {.lex_state = 316, .external_lex_state = 85},
  [3931] = {.lex_state = 316, .external_lex_state = 85},
  [3932] = {.lex_state = 289, .external_lex_state = 86},
  [3933] = {.lex_state = 289, .external_lex_state = 86},
  [3934] = {.lex_state = 289, .external_lex_state = 86},
  [3935] = {.lex_state = 316, .external_lex_state = 85},
  [3936] = {.lex_state = 316, .external_lex_state = 85},
  [3937] = {.lex_state = 316, .external_lex_state = 85},
  [3938] = {.lex_state = 316, .external_lex_state = 85},
  [3939] = {.lex_state = 316, .external_lex_state = 85},
  [3940] = {.lex_state = 316, .external_lex_state = 85},
  [3941] = {.lex_state = 316, .external_lex_state = 85},
  [3942] = {.lex_state = 316, .external_lex_state = 85},
  [3943] = {.lex_state = 316, .external_lex_state = 85},
  [3944] = {.lex_state = 289, .external_lex_state = 86},
  [3945] = {.lex_state = 289, .external_lex_state = 86},
  [3946] = {.lex_state = 316, .external_lex_state = 85},
  [3947] = {.lex_state = 316, .external_lex_state = 85},
  [3948] = {.lex_state = 316, .external_lex_state = 85},
  [3949] = {.lex_state = 316, .external_lex_state = 85},
  [3950] = {.lex_state = 316, .external_lex_state = 85},
  [3951] = {.lex_state = 316, .external_lex_state = 85},
  [3952] = {.lex_state = 316, .external_lex_state = 85},
  [3953] = {.lex_state = 316, .external_lex_state = 85},
  [3954] = {.lex_state = 316, .external_lex_state = 85},
  [3955] = {.lex_state = 316, .external_lex_state = 85},
  [3956] = {.lex_state = 316, .external_lex_state = 85},
  [3957] = {.lex_state = 316, .external_lex_state = 85},
  [3958] = {.lex_state = 316, .external_lex_state = 85},
  [3959] = {.lex_state = 316, .external_lex_state = 85},
  [3960] = {.lex_state = 289, .external_lex_state = 86},
  [3961] = {.lex_state = 316, .external_lex_state = 85},
  [3962] = {.lex_state = 316, .external_lex_state = 85},
  [3963] = {.lex_state = 298, .external_lex_state = 78},
  [3964] = {.lex_state = 316, .external_lex_state = 85},
  [3965] = {.lex_state = 298, .external_lex_state = 78},
  [3966] = {.lex_state = 289, .external_lex_state = 86},
  [3967] = {.lex_state = 289, .external_lex_state = 86},
  [3968] = {.lex_state = 316, .external_lex_state = 85},
  [3969] = {.lex_state = 316, .external_lex_state = 85},
  [3970] = {.lex_state = 289, .external_lex_state = 86},
  [3971] = {.lex_state = 316, .external_lex_state = 85},
  [3972] = {.lex_state = 316, .external_lex_state = 85},
  [3973] = {.lex_state = 289, .external_lex_state = 86},
  [3974] = {.lex_state = 289, .external_lex_state = 86},
  [3975] = {.lex_state = 316, .external_lex_state = 85},
  [3976] = {.lex_state = 316, .external_lex_state = 85},
  [3977] = {.lex_state = 289, .external_lex_state = 86},
  [3978] = {.lex_state = 289, .external_lex_state = 86},
  [3979] = {.lex_state = 316, .external_lex_state = 85},
  [3980] = {.lex_state = 316, .external_lex_state = 85},
  [3981] = {.lex_state = 289, .external_lex_state = 86},
  [3982] = {.lex_state = 289, .external_lex_state = 86},
  [3983] = {.lex_state = 289, .external_lex_state = 86},
  [3984] = {.lex_state = 316, .external_lex_state = 85},
  [3985] = {.lex_state = 289, .external_lex_state = 86},
  [3986] = {.lex_state = 289, .external_lex_state = 86},
  [3987] = {.lex_state = 289, .external_lex_state = 86},
  [3988] = {.lex_state = 316, .external_lex_state = 85},
  [3989] = {.lex_state = 316, .external_lex_state = 85},
  [3990] = {.lex_state = 316, .external_lex_state = 85},
  [3991] = {.lex_state = 316, .external_lex_state = 85},
  [3992] = {.lex_state = 316, .external_lex_state = 85},
  [3993] = {.lex_state = 316, .external_lex_state = 85},
  [3994] = {.lex_state = 289, .external_lex_state = 86},
  [3995] = {.lex_state = 289, .external_lex_state = 86},
  [3996] = {.lex_state = 316, .external_lex_state = 85},
  [3997] = {.lex_state = 289, .external_lex_state = 86},
  [3998] = {.lex_state = 316, .external_lex_state = 85},
  [3999] = {.lex_state = 316, .external_lex_state = 85},
  [4000] = {.lex_state = 289, .external_lex_state = 86},
  [4001] = {.lex_state = 289, .external_lex_state = 86},
  [4002] = {.lex_state = 316, .external_lex_state = 85},
  [4003] = {.lex_state = 316, .external_lex_state = 85},
  [4004] = {.lex_state = 316, .external_lex_state = 85},
  [4005] = {.lex_state = 289, .external_lex_state = 86},
  [4006] = {.lex_state = 289, .external_lex_state = 86},
  [4007] = {.lex_state = 316, .external_lex_state = 85},
  [4008] = {.lex_state = 289, .external_lex_state = 86},
  [4009] = {.lex_state = 316, .external_lex_state = 85},
  [4010] = {.lex_state = 316, .external_lex_state = 85},
  [4011] = {.lex_state = 316, .external_lex_state = 85},
  [4012] = {.lex_state = 289, .external_lex_state = 86},
  [4013] = {.lex_state = 316, .external_lex_state = 85},
  [4014] = {.lex_state = 316, .external_lex_state = 85},
  [4015] = {.lex_state = 316, .external_lex_state = 85},
  [4016] = {.lex_state = 316, .external_lex_state = 85},
  [4017] = {.lex_state = 316, .external_lex_state = 85},
  [4018] = {.lex_state = 298, .external_lex_state = 78},
  [4019] = {.lex_state = 316, .external_lex_state = 85},
  [4020] = {.lex_state = 316, .external_lex_state = 85},
  [4021] = {.lex_state = 289, .external_lex_state = 86},
  [4022] = {.lex_state = 316, .external_lex_state = 85},
  [4023] = {.lex_state = 316, .external_lex_state = 85},
  [4024] = {.lex_state = 298, .external_lex_state = 78},
  [4025] = {.lex_state = 289, .external_lex_state = 86},
  [4026] = {.lex_state = 316, .external_lex_state = 85},
  [4027] = {.lex_state = 316, .external_lex_state = 85},
  [4028] = {.lex_state = 316, .external_lex_state = 85},
  [4029] = {.lex_state = 316, .external_lex_state = 85},
  [4030] = {.lex_state = 298, .external_lex_state = 78},
  [4031] = {.lex_state = 298, .external_lex_state = 78},
  [4032] = {.lex_state = 289, .external_lex_state = 86},
  [4033] = {.lex_state = 289, .external_lex_state = 86},
  [4034] = {.lex_state = 316, .external_lex_state = 85},
  [4035] = {.lex_state = 289, .external_lex_state = 86},
  [4036] = {.lex_state = 298, .external_lex_state = 78},
  [4037] = {.lex_state = 289, .external_lex_state = 86},
  [4038] = {.lex_state = 316, .external_lex_state = 85},
  [4039] = {.lex_state = 298, .external_lex_state = 78},
  [4040] = {.lex_state = 289, .external_lex_state = 86},
  [4041] = {.lex_state = 289, .external_lex_state = 86},
  [4042] = {.lex_state = 316, .external_lex_state = 85},
  [4043] = {.lex_state = 289, .external_lex_state = 86},
  [4044] = {.lex_state = 289, .external_lex_state = 86},
  [4045] = {.lex_state = 298, .external_lex_state = 78},
  [4046] = {.lex_state = 316, .external_lex_state = 85},
  [4047] = {.lex_state = 289, .external_lex_state = 86},
  [4048] = {.lex_state = 289, .external_lex_state = 86},
  [4049] = {.lex_state = 289, .external_lex_state = 86},
  [4050] = {.lex_state = 298, .external_lex_state = 78},
  [4051] = {.lex_state = 289, .external_lex_state = 86},
  [4052] = {.lex_state = 298, .external_lex_state = 78},
  [4053] = {.lex_state = 298, .external_lex_state = 78},
  [4054] = {.lex_state = 289, .external_lex_state = 86},
  [4055] = {.lex_state = 316, .external_lex_state = 85},
  [4056] = {.lex_state = 289, .external_lex_state = 86},
  [4057] = {.lex_state = 289, .external_lex_state = 86},
  [4058] = {.lex_state = 289, .external_lex_state = 86},
  [4059] = {.lex_state = 316, .external_lex_state = 85},
  [4060] = {.lex_state = 289, .external_lex_state = 86},
  [4061] = {.lex_state = 289, .external_lex_state = 86},
  [4062] = {.lex_state = 289, .external_lex_state = 86},
  [4063] = {.lex_state = 289, .external_lex_state = 86},
  [4064] = {.lex_state = 289, .external_lex_state = 86},
  [4065] = {.lex_state = 316, .external_lex_state = 85},
  [4066] = {.lex_state = 289, .external_lex_state = 86},
  [4067] = {.lex_state = 289, .external_lex_state = 86},
  [4068] = {.lex_state = 289, .external_lex_state = 86},
  [4069] = {.lex_state = 289, .external_lex_state = 86},
  [4070] = {.lex_state = 289, .external_lex_state = 86},
  [4071] = {.lex_state = 316, .external_lex_state = 85},
  [4072] = {.lex_state = 289, .external_lex_state = 86},
  [4073] = {.lex_state = 289, .external_lex_state = 86},
  [4074] = {.lex_state = 289, .external_lex_state = 86},
  [4075] = {.lex_state = 316, .external_lex_state = 85},
  [4076] = {.lex_state = 289, .external_lex_state = 86},
  [4077] = {.lex_state = 316, .external_lex_state = 85},
  [4078] = {.lex_state = 298, .external_lex_state = 78},
  [4079] = {.lex_state = 289, .external_lex_state = 86},
  [4080] = {.lex_state = 316, .external_lex_state = 85},
  [4081] = {.lex_state = 316, .external_lex_state = 85},
  [4082] = {.lex_state = 289, .external_lex_state = 86},
  [4083] = {.lex_state = 289, .external_lex_state = 86},
  [4084] = {.lex_state = 289, .external_lex_state = 86},
  [4085] = {.lex_state = 289, .external_lex_state = 86},
  [4086] = {.lex_state = 316, .external_lex_state = 85},
  [4087] = {.lex_state = 316, .external_lex_state = 85},
  [4088] = {.lex_state = 316, .external_lex_state = 85},
  [4089] = {.lex_state = 316, .external_lex_state = 85},
  [4090] = {.lex_state = 289, .external_lex_state = 86},
  [4091] = {.lex_state = 289, .external_lex_state = 86},
  [4092] = {.lex_state = 316, .external_lex_state = 85},
  [4093] = {.lex_state = 289, .external_lex_state = 86},
  [4094] = {.lex_state = 316, .external_lex_state = 85},
  [4095] = {.lex_state = 316, .external_lex_state = 85},
  [4096] = {.lex_state = 289, .external_lex_state = 86},
  [4097] = {.lex_state = 316, .external_lex_state = 85},
  [4098] = {.lex_state = 289, .external_lex_state = 86},
  [4099] = {.lex_state = 316, .external_lex_state = 85},
  [4100] = {.lex_state = 316, .external_lex_state = 85},
  [4101] = {.lex_state = 289, .external_lex_state = 86},
  [4102] = {.lex_state = 289, .external_lex_state = 86},
  [4103] = {.lex_state = 289, .external_lex_state = 86},
  [4104] = {.lex_state = 316, .external_lex_state = 85},
  [4105] = {.lex_state = 316, .external_lex_state = 85},
  [4106] = {.lex_state = 316, .external_lex_state = 85},
  [4107] = {.lex_state = 298, .external_lex_state = 78},
  [4108] = {.lex_state = 289, .external_lex_state = 86},
  [4109] = {.lex_state = 298, .external_lex_state = 78},
  [4110] = {.lex_state = 289, .external_lex_state = 86},
  [4111] = {.lex_state = 298, .external_lex_state = 78},
  [4112] = {.lex_state = 289, .external_lex_state = 86},
  [4113] = {.lex_state = 316, .external_lex_state = 85},
  [4114] = {.lex_state = 289, .external_lex_state = 86},
  [4115] = {.lex_state = 298, .external_lex_state = 78},
  [4116] = {.lex_state = 316, .external_lex_state = 85},
  [4117] = {.lex_state = 316, .external_lex_state = 85},
  [4118] = {.lex_state = 298, .external_lex_state = 78},
  [4119] = {.lex_state = 298, .external_lex_state = 78},
  [4120] = {.lex_state = 289, .external_lex_state = 86},
  [4121] = {.lex_state = 289, .external_lex_state = 86},
  [4122] = {.lex_state = 298, .external_lex_state = 78},
  [4123] = {.lex_state = 289, .external_lex_state = 86},
  [4124] = {.lex_state = 289, .external_lex_state = 86},
  [4125] = {.lex_state = 298, .external_lex_state = 78},
  [4126] = {.lex_state = 298, .external_lex_state = 78},
  [4127] = {.lex_state = 316, .external_lex_state = 85},
  [4128] = {.lex_state = 316, .external_lex_state = 85},
  [4129] = {.lex_state = 316, .external_lex_state = 85},
  [4130] = {.lex_state = 289, .external_lex_state = 86},
  [4131] = {.lex_state = 289, .external_lex_state = 86},
  [4132] = {.lex_state = 289, .external_lex_state = 86},
  [4133] = {.lex_state = 316, .external_lex_state = 85},
  [4134] = {.lex_state = 289, .external_lex_state = 86},
  [4135] = {.lex_state = 289, .external_lex_state = 86},
  [4136] = {.lex_state = 289, .external_lex_state = 86},
  [4137] = {.lex_state = 289, .external_lex_state = 86},
  [4138] = {.lex_state = 298, .external_lex_state = 78},
  [4139] = {.lex_state = 316, .external_lex_state = 85},
  [4140] = {.lex_state = 316, .external_lex_state = 85},
  [4141] = {.lex_state = 316, .external_lex_state = 85},
  [4142] = {.lex_state = 289, .external_lex_state = 86},
  [4143] = {.lex_state = 289, .external_lex_state = 86},
  [4144] = {.lex_state = 289, .external_lex_state = 86},
  [4145] = {.lex_state = 289, .external_lex_state = 86},
  [4146] = {.lex_state = 289, .external_lex_state = 86},
  [4147] = {.lex_state = 289, .external_lex_state = 86},
  [4148] = {.lex_state = 289, .external_lex_state = 86},
  [4149] = {.lex_state = 289, .external_lex_state = 86},
  [4150] = {.lex_state = 289, .external_lex_state = 86},
  [4151] = {.lex_state = 289, .external_lex_state = 86},
  [4152] = {.lex_state = 298, .external_lex_state = 78},
  [4153] = {.lex_state = 316, .external_lex_state = 85},
  [4154] = {.lex_state = 289, .external_lex_state = 86},
  [4155] = {.lex_state = 289, .external_lex_state = 86},
  [4156] = {.lex_state = 316, .external_lex_state = 85},
  [4157] = {.lex_state = 316, .external_lex_state = 85},
  [4158] = {.lex_state = 316, .external_lex_state = 85},
  [4159] = {.lex_state = 316, .external_lex_state = 85},
  [4160] = {.lex_state = 316, .external_lex_state = 85},
  [4161] = {.lex_state = 316, .external_lex_state = 85},
  [4162] = {.lex_state = 316, .external_lex_state = 85},
  [4163] = {.lex_state = 316, .external_lex_state = 85},
  [4164] = {.lex_state = 289, .external_lex_state = 86},
  [4165] = {.lex_state = 298, .external_lex_state = 78},
  [4166] = {.lex_state = 289, .external_lex_state = 86},
  [4167] = {.lex_state = 298, .external_lex_state = 78},
  [4168] = {.lex_state = 316, .external_lex_state = 85},
  [4169] = {.lex_state = 289, .external_lex_state = 86},
  [4170] = {.lex_state = 298, .external_lex_state = 78},
  [4171] = {.lex_state = 316, .external_lex_state = 85},
  [4172] = {.lex_state = 316, .external_lex_state = 85},
  [4173] = {.lex_state = 316, .external_lex_state = 85},
  [4174] = {.lex_state = 298, .external_lex_state = 78},
  [4175] = {.lex_state = 289, .external_lex_state = 86},
  [4176] = {.lex_state = 289, .external_lex_state = 86},
  [4177] = {.lex_state = 316, .external_lex_state = 85},
  [4178] = {.lex_state = 316, .external_lex_state = 85},
  [4179] = {.lex_state = 316, .external_lex_state = 85},
  [4180] = {.lex_state = 316, .external_lex_state = 85},
  [4181] = {.lex_state = 316, .external_lex_state = 85},
  [4182] = {.lex_state = 298, .external_lex_state = 78},
  [4183] = {.lex_state = 289, .external_lex_state = 86},
  [4184] = {.lex_state = 289, .external_lex_state = 86},
  [4185] = {.lex_state = 289, .external_lex_state = 86},
  [4186] = {.lex_state = 298, .external_lex_state = 78},
  [4187] = {.lex_state = 316, .external_lex_state = 85},
  [4188] = {.lex_state = 298, .external_lex_state = 78},
  [4189] = {.lex_state = 298, .external_lex_state = 78},
  [4190] = {.lex_state = 289, .external_lex_state = 86},
  [4191] = {.lex_state = 316, .external_lex_state = 85},
  [4192] = {.lex_state = 289, .external_lex_state = 86},
  [4193] = {.lex_state = 289, .external_lex_state = 86},
  [4194] = {.lex_state = 289, .external_lex_state = 86},
  [4195] = {.lex_state = 289, .external_lex_state = 86},
  [4196] = {.lex_state = 289, .external_lex_state = 86},
  [4197] = {.lex_state = 316, .external_lex_state = 85},
  [4198] = {.lex_state = 316, .external_lex_state = 85},
  [4199] = {.lex_state = 316, .external_lex_state = 85},
  [4200] = {.lex_state = 289, .external_lex_state = 86},
  [4201] = {.lex_state = 289, .external_lex_state = 86},
  [4202] = {.lex_state = 289, .external_lex_state = 86},
  [4203] = {.lex_state = 316, .external_lex_state = 85},
  [4204] = {.lex_state = 289, .external_lex_state = 86},
  [4205] = {.lex_state = 289, .external_lex_state = 86},
  [4206] = {.lex_state = 289, .external_lex_state = 86},
  [4207] = {.lex_state = 289, .external_lex_state = 86},
  [4208] = {.lex_state = 289, .external_lex_state = 86},
  [4209] = {.lex_state = 289, .external_lex_state = 86},
  [4210] = {.lex_state = 289, .external_lex_state = 86},
  [4211] = {.lex_state = 289, .external_lex_state = 86},
  [4212] = {.lex_state = 289, .external_lex_state = 86},
  [4213] = {.lex_state = 316, .external_lex_state = 85},
  [4214] = {.lex_state = 316, .external_lex_state = 85},
  [4215] = {.lex_state = 289, .external_lex_state = 86},
  [4216] = {.lex_state = 208, .external_lex_state = 87},
  [4217] = {.lex_state = 65, .external_lex_state = 88},
  [4218] = {.lex_state = 67, .external_lex_state = 89},
  [4219] = {.lex_state = 65, .external_lex_state = 88},
  [4220] = {.lex_state = 65, .external_lex_state = 88},
  [4221] = {.lex_state = 319, .external_lex_state = 90},
  [4222] = {.lex_state = 68, .external_lex_state = 91},
  [4223] = {.lex_state = 67, .external_lex_state = 92},
  [4224] = {.lex_state = 67, .external_lex_state = 92},
  [4225] = {.lex_state = 67, .external_lex_state = 92},
  [4226] = {.lex_state = 65, .external_lex_state = 88},
  [4227] = {.lex_state = 67, .external_lex_state = 92},
  [4228] = {.lex_state = 208, .external_lex_state = 87},
  [4229] = {.lex_state = 68, .external_lex_state = 91},
  [4230] = {.lex_state = 68, .external_lex_state = 63},
  [4231] = {.lex_state = 67, .external_lex_state = 89},
  [4232] = {.lex_state = 67, .external_lex_state = 89},
  [4233] = {.lex_state = 67, .external_lex_state = 93},
  [4234] = {.lex_state = 67, .external_lex_state = 94},
  [4235] = {.lex_state = 67, .external_lex_state = 92},
  [4236] = {.lex_state = 65, .external_lex_state = 88},
  [4237] = {.lex_state = 68, .external_lex_state = 63},
  [4238] = {.lex_state = 68, .external_lex_state = 91},
  [4239] = {.lex_state = 67, .external_lex_state = 94},
  [4240] = {.lex_state = 67, .external_lex_state = 89},
  [4241] = {.lex_state = 65, .external_lex_state = 88},
  [4242] = {.lex_state = 67, .external_lex_state = 89},
  [4243] = {.lex_state = 65, .external_lex_state = 88},
  [4244] = {.lex_state = 521, .external_lex_state = 95},
  [4245] = {.lex_state = 67, .external_lex_state = 89},
  [4246] = {.lex_state = 67, .external_lex_state = 89},
  [4247] = {.lex_state = 67, .external_lex_state = 89},
  [4248] = {.lex_state = 67, .external_lex_state = 89},
  [4249] = {.lex_state = 67, .external_lex_state = 94},
  [4250] = {.lex_state = 67, .external_lex_state = 92},
  [4251] = {.lex_state = 68, .external_lex_state = 91},
  [4252] = {.lex_state = 67, .external_lex_state = 89},
  [4253] = {.lex_state = 519, .external_lex_state = 93},
  [4254] = {.lex_state = 67, .external_lex_state = 94},
  [4255] = {.lex_state = 68, .external_lex_state = 95},
  [4256] = {.lex_state = 68, .external_lex_state = 91},
  [4257] = {.lex_state = 67, .external_lex_state = 89},
  [4258] = {.lex_state = 68, .external_lex_state = 91},
  [4259] = {.lex_state = 68, .external_lex_state = 89},
  [4260] = {.lex_state = 67, .external_lex_state = 89},
  [4261] = {.lex_state = 522, .external_lex_state = 66},
  [4262] = {.lex_state = 67, .external_lex_state = 92},
  [4263] = {.lex_state = 67, .external_lex_state = 89},
  [4264] = {.lex_state = 519, .external_lex_state = 93},
  [4265] = {.lex_state = 67, .external_lex_state = 89},
  [4266] = {.lex_state = 67, .external_lex_state = 93},
  [4267] = {.lex_state = 67, .external_lex_state = 89},
  [4268] = {.lex_state = 521, .external_lex_state = 93},
  [4269] = {.lex_state = 520, .external_lex_state = 94},
  [4270] = {.lex_state = 68, .external_lex_state = 89},
  [4271] = {.lex_state = 68, .external_lex_state = 89},
  [4272] = {.lex_state = 68, .external_lex_state = 89},
  [4273] = {.lex_state = 67, .external_lex_state = 92},
  [4274] = {.lex_state = 67, .external_lex_state = 93},
  [4275] = {.lex_state = 519, .external_lex_state = 93},
  [4276] = {.lex_state = 67, .external_lex_state = 93},
  [4277] = {.lex_state = 520, .external_lex_state = 94},
  [4278] = {.lex_state = 68, .external_lex_state = 92},
  [4279] = {.lex_state = 67, .external_lex_state = 94},
  [4280] = {.lex_state = 519, .external_lex_state = 93},
  [4281] = {.lex_state = 519, .external_lex_state = 94},
  [4282] = {.lex_state = 68, .external_lex_state = 66},
  [4283] = {.lex_state = 68, .external_lex_state = 92},
  [4284] = {.lex_state = 67, .external_lex_state = 93},
  [4285] = {.lex_state = 68, .external_lex_state = 91},
  [4286] = {.lex_state = 68, .external_lex_state = 89},
  [4287] = {.lex_state = 522, .external_lex_state = 66},
  [4288] = {.lex_state = 521, .external_lex_state = 95},
  [4289] = {.lex_state = 520, .external_lex_state = 94},
  [4290] = {.lex_state = 67, .external_lex_state = 92},
  [4291] = {.lex_state = 319, .external_lex_state = 96},
  [4292] = {.lex_state = 68, .external_lex_state = 89},
  [4293] = {.lex_state = 319, .external_lex_state = 96},
  [4294] = {.lex_state = 520, .external_lex_state = 94},
  [4295] = {.lex_state = 67, .external_lex_state = 89},
  [4296] = {.lex_state = 68, .external_lex_state = 89},
  [4297] = {.lex_state = 68, .external_lex_state = 89},
  [4298] = {.lex_state = 67, .external_lex_state = 89},
  [4299] = {.lex_state = 67, .external_lex_state = 93},
  [4300] = {.lex_state = 67, .external_lex_state = 89},
  [4301] = {.lex_state = 67, .external_lex_state = 89},
  [4302] = {.lex_state = 67, .external_lex_state = 89},
  [4303] = {.lex_state = 68, .external_lex_state = 91},
  [4304] = {.lex_state = 68, .external_lex_state = 89},
  [4305] = {.lex_state = 520, .external_lex_state = 94},
  [4306] = {.lex_state = 67, .external_lex_state = 89},
  [4307] = {.lex_state = 67, .external_lex_state = 89},
  [4308] = {.lex_state = 67, .external_lex_state = 89},
  [4309] = {.lex_state = 67, .external_lex_state = 93},
  [4310] = {.lex_state = 67, .external_lex_state = 93},
  [4311] = {.lex_state = 519, .external_lex_state = 93},
  [4312] = {.lex_state = 519, .external_lex_state = 94},
  [4313] = {.lex_state = 521, .external_lex_state = 95},
  [4314] = {.lex_state = 520, .external_lex_state = 94},
  [4315] = {.lex_state = 68, .external_lex_state = 93},
  [4316] = {.lex_state = 521, .external_lex_state = 95},
  [4317] = {.lex_state = 67, .external_lex_state = 89},
  [4318] = {.lex_state = 521, .external_lex_state = 95},
  [4319] = {.lex_state = 67, .external_lex_state = 93},
  [4320] = {.lex_state = 521, .external_lex_state = 95},
  [4321] = {.lex_state = 519, .external_lex_state = 94},
  [4322] = {.lex_state = 68, .external_lex_state = 91},
  [4323] = {.lex_state = 68, .external_lex_state = 91},
  [4324] = {.lex_state = 67, .external_lex_state = 92},
  [4325] = {.lex_state = 68, .external_lex_state = 91},
  [4326] = {.lex_state = 68, .external_lex_state = 91},
  [4327] = {.lex_state = 68, .external_lex_state = 91},
  [4328] = {.lex_state = 67, .external_lex_state = 89},
  [4329] = {.lex_state = 67, .external_lex_state = 89},
  [4330] = {.lex_state = 67, .external_lex_state = 89},
  [4331] = {.lex_state = 67, .external_lex_state = 93},
  [4332] = {.lex_state = 68, .external_lex_state = 91},
  [4333] = {.lex_state = 521, .external_lex_state = 95},
  [4334] = {.lex_state = 521, .external_lex_state = 95},
  [4335] = {.lex_state = 68, .external_lex_state = 91},
  [4336] = {.lex_state = 68, .external_lex_state = 91},
  [4337] = {.lex_state = 519, .external_lex_state = 93},
  [4338] = {.lex_state = 519, .external_lex_state = 93},
  [4339] = {.lex_state = 68, .external_lex_state = 92},
  [4340] = {.lex_state = 68, .external_lex_state = 91},
  [4341] = {.lex_state = 68, .external_lex_state = 92},
  [4342] = {.lex_state = 68, .external_lex_state = 92},
  [4343] = {.lex_state = 68, .external_lex_state = 91},
  [4344] = {.lex_state = 319, .external_lex_state = 96},
  [4345] = {.lex_state = 521, .external_lex_state = 95},
  [4346] = {.lex_state = 67, .external_lex_state = 92},
  [4347] = {.lex_state = 520, .external_lex_state = 94},
  [4348] = {.lex_state = 519, .external_lex_state = 93},
  [4349] = {.lex_state = 68, .external_lex_state = 92},
  [4350] = {.lex_state = 521, .external_lex_state = 95},
  [4351] = {.lex_state = 67, .external_lex_state = 92},
  [4352] = {.lex_state = 67, .external_lex_state = 93},
  [4353] = {.lex_state = 519, .external_lex_state = 94},
  [4354] = {.lex_state = 68, .external_lex_state = 91},
  [4355] = {.lex_state = 68, .external_lex_state = 91},
  [4356] = {.lex_state = 68, .external_lex_state = 91},
  [4357] = {.lex_state = 521, .external_lex_state = 95},
  [4358] = {.lex_state = 68, .external_lex_state = 66},
  [4359] = {.lex_state = 519, .external_lex_state = 93},
  [4360] = {.lex_state = 521, .external_lex_state = 95},
  [4361] = {.lex_state = 521, .external_lex_state = 95},
  [4362] = {.lex_state = 519, .external_lex_state = 93},
  [4363] = {.lex_state = 319, .external_lex_state = 96},
  [4364] = {.lex_state = 68, .external_lex_state = 89},
  [4365] = {.lex_state = 521, .external_lex_state = 95},
  [4366] = {.lex_state = 519, .external_lex_state = 93},
  [4367] = {.lex_state = 519, .external_lex_state = 93},
  [4368] = {.lex_state = 68, .external_lex_state = 63},
  [4369] = {.lex_state = 519, .external_lex_state = 93},
  [4370] = {.lex_state = 68, .external_lex_state = 89},
  [4371] = {.lex_state = 67, .external_lex_state = 94},
  [4372] = {.lex_state = 519, .external_lex_state = 93},
  [4373] = {.lex_state = 521, .external_lex_state = 95},
  [4374] = {.lex_state = 521, .external_lex_state = 95},
  [4375] = {.lex_state = 67, .external_lex_state = 93},
  [4376] = {.lex_state = 519, .external_lex_state = 94},
  [4377] = {.lex_state = 521, .external_lex_state = 95},
  [4378] = {.lex_state = 521, .external_lex_state = 95},
  [4379] = {.lex_state = 519, .external_lex_state = 93},
  [4380] = {.lex_state = 68, .external_lex_state = 95},
  [4381] = {.lex_state = 519, .external_lex_state = 94},
  [4382] = {.lex_state = 67, .external_lex_state = 92},
  [4383] = {.lex_state = 68, .external_lex_state = 63},
  [4384] = {.lex_state = 68, .external_lex_state = 91},
  [4385] = {.lex_state = 68, .external_lex_state = 91},
  [4386] = {.lex_state = 519, .external_lex_state = 93},
  [4387] = {.lex_state = 67, .external_lex_state = 92},
  [4388] = {.lex_state = 67, .external_lex_state = 89},
  [4389] = {.lex_state = 519, .external_lex_state = 93},
  [4390] = {.lex_state = 68, .external_lex_state = 95},
  [4391] = {.lex_state = 519, .external_lex_state = 93},
  [4392] = {.lex_state = 519, .external_lex_state = 93},
  [4393] = {.lex_state = 521, .external_lex_state = 95},
  [4394] = {.lex_state = 520, .external_lex_state = 94},
  [4395] = {.lex_state = 519, .external_lex_state = 93},
  [4396] = {.lex_state = 519, .external_lex_state = 94},
  [4397] = {.lex_state = 520, .external_lex_state = 94},
  [4398] = {.lex_state = 521, .external_lex_state = 95},
  [4399] = {.lex_state = 67, .external_lex_state = 89},
  [4400] = {.lex_state = 519, .external_lex_state = 94},
  [4401] = {.lex_state = 519, .external_lex_state = 93},
  [4402] = {.lex_state = 521, .external_lex_state = 95},
  [4403] = {.lex_state = 519, .external_lex_state = 93},
  [4404] = {.lex_state = 67, .external_lex_state = 89},
  [4405] = {.lex_state = 68, .external_lex_state = 91},
  [4406] = {.lex_state = 68, .external_lex_state = 89},
  [4407] = {.lex_state = 521, .external_lex_state = 95},
  [4408] = {.lex_state = 521, .external_lex_state = 95},
  [4409] = {.lex_state = 67, .external_lex_state = 89},
  [4410] = {.lex_state = 68, .external_lex_state = 95},
  [4411] = {.lex_state = 519, .external_lex_state = 93},
  [4412] = {.lex_state = 68, .external_lex_state = 91},
  [4413] = {.lex_state = 68, .external_lex_state = 95},
  [4414] = {.lex_state = 68, .external_lex_state = 95},
  [4415] = {.lex_state = 519, .external_lex_state = 94},
  [4416] = {.lex_state = 68, .external_lex_state = 94},
  [4417] = {.lex_state = 68, .external_lex_state = 89},
  [4418] = {.lex_state = 519, .external_lex_state = 93},
  [4419] = {.lex_state = 68, .external_lex_state = 95},
  [4420] = {.lex_state = 67, .external_lex_state = 94},
  [4421] = {.lex_state = 522, .external_lex_state = 94},
  [4422] = {.lex_state = 238, .external_lex_state = 97},
  [4423] = {.lex_state = 520, .external_lex_state = 94},
  [4424] = {.lex_state = 67, .external_lex_state = 93},
  [4425] = {.lex_state = 67, .external_lex_state = 93},
  [4426] = {.lex_state = 67, .external_lex_state = 93},
  [4427] = {.lex_state = 67, .external_lex_state = 93},
  [4428] = {.lex_state = 68, .external_lex_state = 94},
  [4429] = {.lex_state = 521, .external_lex_state = 93},
  [4430] = {.lex_state = 68, .external_lex_state = 93},
  [4431] = {.lex_state = 68, .external_lex_state = 66},
  [4432] = {.lex_state = 67, .external_lex_state = 93},
  [4433] = {.lex_state = 68, .external_lex_state = 94},
  [4434] = {.lex_state = 68, .external_lex_state = 95},
  [4435] = {.lex_state = 68, .external_lex_state = 95},
  [4436] = {.lex_state = 68, .external_lex_state = 95},
  [4437] = {.lex_state = 68, .external_lex_state = 95},
  [4438] = {.lex_state = 67, .external_lex_state = 93},
  [4439] = {.lex_state = 521, .external_lex_state = 66},
  [4440] = {.lex_state = 519, .external_lex_state = 93},
  [4441] = {.lex_state = 68, .external_lex_state = 95},
  [4442] = {.lex_state = 68, .external_lex_state = 89},
  [4443] = {.lex_state = 67, .external_lex_state = 93},
  [4444] = {.lex_state = 67, .external_lex_state = 93},
  [4445] = {.lex_state = 67, .external_lex_state = 93},
  [4446] = {.lex_state = 67, .external_lex_state = 93},
  [4447] = {.lex_state = 68, .external_lex_state = 95},
  [4448] = {.lex_state = 68, .external_lex_state = 95},
  [4449] = {.lex_state = 68, .external_lex_state = 95},
  [4450] = {.lex_state = 519, .external_lex_state = 93},
  [4451] = {.lex_state = 521, .external_lex_state = 93},
  [4452] = {.lex_state = 519, .external_lex_state = 93},
  [4453] = {.lex_state = 68, .external_lex_state = 93},
  [4454] = {.lex_state = 67, .external_lex_state = 93},
  [4455] = {.lex_state = 68, .external_lex_state = 94},
  [4456] = {.lex_state = 521, .external_lex_state = 93},
  [4457] = {.lex_state = 68, .external_lex_state = 95},
  [4458] = {.lex_state = 238, .external_lex_state = 97},
  [4459] = {.lex_state = 519, .external_lex_state = 93},
  [4460] = {.lex_state = 238, .external_lex_state = 97},
  [4461] = {.lex_state = 68, .external_lex_state = 89},
  [4462] = {.lex_state = 68, .external_lex_state = 92},
  [4463] = {.lex_state = 519, .external_lex_state = 93},
  [4464] = {.lex_state = 67, .external_lex_state = 93},
  [4465] = {.lex_state = 68, .external_lex_state = 94},
  [4466] = {.lex_state = 519, .external_lex_state = 93},
  [4467] = {.lex_state = 519, .external_lex_state = 93},
  [4468] = {.lex_state = 68, .external_lex_state = 93},
  [4469] = {.lex_state = 68, .external_lex_state = 89},
  [4470] = {.lex_state = 68, .external_lex_state = 89},
  [4471] = {.lex_state = 68, .external_lex_state = 93},
  [4472] = {.lex_state = 519, .external_lex_state = 93},
  [4473] = {.lex_state = 521, .external_lex_state = 93},
  [4474] = {.lex_state = 68, .external_lex_state = 89},
  [4475] = {.lex_state = 521, .external_lex_state = 93},
  [4476] = {.lex_state = 519, .external_lex_state = 93},
  [4477] = {.lex_state = 519, .external_lex_state = 93},
  [4478] = {.lex_state = 68, .external_lex_state = 95},
  [4479] = {.lex_state = 519, .external_lex_state = 93},
  [4480] = {.lex_state = 519, .external_lex_state = 94},
  [4481] = {.lex_state = 519, .external_lex_state = 93},
  [4482] = {.lex_state = 67, .external_lex_state = 93},
  [4483] = {.lex_state = 68, .external_lex_state = 92},
  [4484] = {.lex_state = 519, .external_lex_state = 93},
  [4485] = {.lex_state = 522, .external_lex_state = 94},
  [4486] = {.lex_state = 519, .external_lex_state = 93},
  [4487] = {.lex_state = 521, .external_lex_state = 95},
  [4488] = {.lex_state = 519, .external_lex_state = 93},
  [4489] = {.lex_state = 519, .external_lex_state = 94},
  [4490] = {.lex_state = 68, .external_lex_state = 95},
  [4491] = {.lex_state = 238, .external_lex_state = 97},
  [4492] = {.lex_state = 67, .external_lex_state = 94},
  [4493] = {.lex_state = 67, .external_lex_state = 93},
  [4494] = {.lex_state = 67, .external_lex_state = 93},
  [4495] = {.lex_state = 519, .external_lex_state = 93},
  [4496] = {.lex_state = 68, .external_lex_state = 89},
  [4497] = {.lex_state = 68, .external_lex_state = 89},
  [4498] = {.lex_state = 67, .external_lex_state = 94},
  [4499] = {.lex_state = 68, .external_lex_state = 95},
  [4500] = {.lex_state = 522, .external_lex_state = 94},
  [4501] = {.lex_state = 68, .external_lex_state = 89},
  [4502] = {.lex_state = 68, .external_lex_state = 89},
  [4503] = {.lex_state = 522, .external_lex_state = 66},
  [4504] = {.lex_state = 68, .external_lex_state = 93},
  [4505] = {.lex_state = 521, .external_lex_state = 95},
  [4506] = {.lex_state = 68, .external_lex_state = 93},
  [4507] = {.lex_state = 68, .external_lex_state = 92},
  [4508] = {.lex_state = 68, .external_lex_state = 95},
  [4509] = {.lex_state = 319, .external_lex_state = 76},
  [4510] = {.lex_state = 519, .external_lex_state = 93},
  [4511] = {.lex_state = 521, .external_lex_state = 95},
  [4512] = {.lex_state = 68, .external_lex_state = 89},
  [4513] = {.lex_state = 519, .external_lex_state = 93},
  [4514] = {.lex_state = 521, .external_lex_state = 95},
  [4515] = {.lex_state = 521, .external_lex_state = 93},
  [4516] = {.lex_state = 68, .external_lex_state = 89},
  [4517] = {.lex_state = 519, .external_lex_state = 94},
  [4518] = {.lex_state = 519, .external_lex_state = 93},
  [4519] = {.lex_state = 521, .external_lex_state = 66},
  [4520] = {.lex_state = 238, .external_lex_state = 97},
  [4521] = {.lex_state = 522, .external_lex_state = 94},
  [4522] = {.lex_state = 519, .external_lex_state = 93},
  [4523] = {.lex_state = 519, .external_lex_state = 93},
  [4524] = {.lex_state = 521, .external_lex_state = 93},
  [4525] = {.lex_state = 521, .external_lex_state = 95},
  [4526] = {.lex_state = 68, .external_lex_state = 95},
  [4527] = {.lex_state = 67, .external_lex_state = 94},
  [4528] = {.lex_state = 68, .external_lex_state = 94},
  [4529] = {.lex_state = 519, .external_lex_state = 93},
  [4530] = {.lex_state = 519, .external_lex_state = 94},
  [4531] = {.lex_state = 521, .external_lex_state = 95},
  [4532] = {.lex_state = 68, .external_lex_state = 89},
  [4533] = {.lex_state = 68, .external_lex_state = 66},
  [4534] = {.lex_state = 521, .external_lex_state = 66},
  [4535] = {.lex_state = 521, .external_lex_state = 95},
  [4536] = {.lex_state = 521, .external_lex_state = 93},
  [4537] = {.lex_state = 521, .external_lex_state = 93},
  [4538] = {.lex_state = 68, .external_lex_state = 92},
  [4539] = {.lex_state = 68, .external_lex_state = 89},
  [4540] = {.lex_state = 521, .external_lex_state = 93},
  [4541] = {.lex_state = 319, .external_lex_state = 76},
  [4542] = {.lex_state = 319, .external_lex_state = 76},
  [4543] = {.lex_state = 521, .external_lex_state = 93},
  [4544] = {.lex_state = 521, .external_lex_state = 93},
  [4545] = {.lex_state = 521, .external_lex_state = 93},
  [4546] = {.lex_state = 522, .external_lex_state = 94},
  [4547] = {.lex_state = 520, .external_lex_state = 94},
  [4548] = {.lex_state = 521, .external_lex_state = 93},
  [4549] = {.lex_state = 67, .external_lex_state = 94},
  [4550] = {.lex_state = 521, .external_lex_state = 95},
  [4551] = {.lex_state = 519, .external_lex_state = 94},
  [4552] = {.lex_state = 68, .external_lex_state = 89},
  [4553] = {.lex_state = 68, .external_lex_state = 89},
  [4554] = {.lex_state = 68, .external_lex_state = 92},
  [4555] = {.lex_state = 68, .external_lex_state = 93},
  [4556] = {.lex_state = 68, .external_lex_state = 95},
  [4557] = {.lex_state = 68, .external_lex_state = 95},
  [4558] = {.lex_state = 238, .external_lex_state = 97},
  [4559] = {.lex_state = 521, .external_lex_state = 95},
  [4560] = {.lex_state = 68, .external_lex_state = 89},
  [4561] = {.lex_state = 521, .external_lex_state = 93},
  [4562] = {.lex_state = 68, .external_lex_state = 93},
  [4563] = {.lex_state = 68, .external_lex_state = 63},
  [4564] = {.lex_state = 521, .external_lex_state = 93},
  [4565] = {.lex_state = 519, .external_lex_state = 94},
  [4566] = {.lex_state = 521, .external_lex_state = 93},
  [4567] = {.lex_state = 519, .external_lex_state = 94},
  [4568] = {.lex_state = 522, .external_lex_state = 66},
  [4569] = {.lex_state = 521, .external_lex_state = 93},
  [4570] = {.lex_state = 519, .external_lex_state = 94},
  [4571] = {.lex_state = 519, .external_lex_state = 94},
  [4572] = {.lex_state = 68, .external_lex_state = 89},
  [4573] = {.lex_state = 520, .external_lex_state = 94},
  [4574] = {.lex_state = 521, .external_lex_state = 66},
  [4575] = {.lex_state = 67, .external_lex_state = 93},
  [4576] = {.lex_state = 67, .external_lex_state = 93},
  [4577] = {.lex_state = 67, .external_lex_state = 92},
  [4578] = {.lex_state = 521, .external_lex_state = 93},
  [4579] = {.lex_state = 68, .external_lex_state = 93},
  [4580] = {.lex_state = 67, .external_lex_state = 94},
  [4581] = {.lex_state = 522, .external_lex_state = 66},
  [4582] = {.lex_state = 519, .external_lex_state = 93},
  [4583] = {.lex_state = 519, .external_lex_state = 93},
  [4584] = {.lex_state = 519, .external_lex_state = 93},
  [4585] = {.lex_state = 521, .external_lex_state = 95},
  [4586] = {.lex_state = 521, .external_lex_state = 95},
  [4587] = {.lex_state = 68, .external_lex_state = 95},
  [4588] = {.lex_state = 67, .external_lex_state = 93},
  [4589] = {.lex_state = 68, .external_lex_state = 93},
  [4590] = {.lex_state = 67, .external_lex_state = 93},
  [4591] = {.lex_state = 519, .external_lex_state = 93},
  [4592] = {.lex_state = 519, .external_lex_state = 93},
  [4593] = {.lex_state = 521, .external_lex_state = 93},
  [4594] = {.lex_state = 521, .external_lex_state = 95},
  [4595] = {.lex_state = 68, .external_lex_state = 95},
  [4596] = {.lex_state = 67, .external_lex_state = 94},
  [4597] = {.lex_state = 67, .external_lex_state = 94},
  [4598] = {.lex_state = 521, .external_lex_state = 93},
  [4599] = {.lex_state = 519, .external_lex_state = 93},
  [4600] = {.lex_state = 68, .external_lex_state = 89},
  [4601] = {.lex_state = 68, .external_lex_state = 93},
  [4602] = {.lex_state = 519, .external_lex_state = 93},
  [4603] = {.lex_state = 521, .external_lex_state = 95},
  [4604] = {.lex_state = 519, .external_lex_state = 93},
  [4605] = {.lex_state = 319, .external_lex_state = 76},
  [4606] = {.lex_state = 521, .external_lex_state = 93},
  [4607] = {.lex_state = 67, .external_lex_state = 92},
  [4608] = {.lex_state = 521, .external_lex_state = 94},
  [4609] = {.lex_state = 520, .external_lex_state = 94},
  [4610] = {.lex_state = 520, .external_lex_state = 94},
  [4611] = {.lex_state = 521, .external_lex_state = 93},
  [4612] = {.lex_state = 520, .external_lex_state = 94},
  [4613] = {.lex_state = 67, .external_lex_state = 92},
  [4614] = {.lex_state = 521, .external_lex_state = 93},
  [4615] = {.lex_state = 520, .external_lex_state = 94},
  [4616] = {.lex_state = 68, .external_lex_state = 92},
  [4617] = {.lex_state = 520, .external_lex_state = 94},
  [4618] = {.lex_state = 68, .external_lex_state = 93},
  [4619] = {.lex_state = 521, .external_lex_state = 93},
  [4620] = {.lex_state = 68, .external_lex_state = 92},
  [4621] = {.lex_state = 521, .external_lex_state = 93},
  [4622] = {.lex_state = 521, .external_lex_state = 93},
  [4623] = {.lex_state = 520, .external_lex_state = 94},
  [4624] = {.lex_state = 294, .external_lex_state = 98},
  [4625] = {.lex_state = 519, .external_lex_state = 94},
  [4626] = {.lex_state = 521, .external_lex_state = 93},
  [4627] = {.lex_state = 521, .external_lex_state = 93},
  [4628] = {.lex_state = 238, .external_lex_state = 97},
  [4629] = {.lex_state = 294, .external_lex_state = 98},
  [4630] = {.lex_state = 521, .external_lex_state = 93},
  [4631] = {.lex_state = 238, .external_lex_state = 97},
  [4632] = {.lex_state = 67, .external_lex_state = 92},
  [4633] = {.lex_state = 521, .external_lex_state = 93},
  [4634] = {.lex_state = 521, .external_lex_state = 93},
  [4635] = {.lex_state = 521, .external_lex_state = 93},
  [4636] = {.lex_state = 522, .external_lex_state = 66},
  [4637] = {.lex_state = 68, .external_lex_state = 93},
  [4638] = {.lex_state = 68, .external_lex_state = 93},
  [4639] = {.lex_state = 68, .external_lex_state = 63},
  [4640] = {.lex_state = 521, .external_lex_state = 93},
  [4641] = {.lex_state = 294, .external_lex_state = 98},
  [4642] = {.lex_state = 520, .external_lex_state = 94},
  [4643] = {.lex_state = 67, .external_lex_state = 92},
  [4644] = {.lex_state = 519, .external_lex_state = 94},
  [4645] = {.lex_state = 68, .external_lex_state = 94},
  [4646] = {.lex_state = 312, .external_lex_state = 99},
  [4647] = {.lex_state = 67, .external_lex_state = 92},
  [4648] = {.lex_state = 67, .external_lex_state = 92},
  [4649] = {.lex_state = 294, .external_lex_state = 98},
  [4650] = {.lex_state = 521, .external_lex_state = 94},
  [4651] = {.lex_state = 312, .external_lex_state = 99},
  [4652] = {.lex_state = 294, .external_lex_state = 98},
  [4653] = {.lex_state = 238, .external_lex_state = 97},
  [4654] = {.lex_state = 294, .external_lex_state = 98},
  [4655] = {.lex_state = 520, .external_lex_state = 94},
  [4656] = {.lex_state = 68, .external_lex_state = 93},
  [4657] = {.lex_state = 294, .external_lex_state = 98},
  [4658] = {.lex_state = 294, .external_lex_state = 98},
  [4659] = {.lex_state = 519, .external_lex_state = 94},
  [4660] = {.lex_state = 294, .external_lex_state = 98},
  [4661] = {.lex_state = 294, .external_lex_state = 98},
  [4662] = {.lex_state = 522, .external_lex_state = 94},
  [4663] = {.lex_state = 521, .external_lex_state = 94},
  [4664] = {.lex_state = 68, .external_lex_state = 92},
  [4665] = {.lex_state = 519, .external_lex_state = 94},
  [4666] = {.lex_state = 520, .external_lex_state = 94},
  [4667] = {.lex_state = 519, .external_lex_state = 94},
  [4668] = {.lex_state = 294, .external_lex_state = 98},
  [4669] = {.lex_state = 521, .external_lex_state = 93},
  [4670] = {.lex_state = 238, .external_lex_state = 97},
  [4671] = {.lex_state = 294, .external_lex_state = 98},
  [4672] = {.lex_state = 294, .external_lex_state = 98},
  [4673] = {.lex_state = 521, .external_lex_state = 93},
  [4674] = {.lex_state = 521, .external_lex_state = 93},
  [4675] = {.lex_state = 294, .external_lex_state = 98},
  [4676] = {.lex_state = 294, .external_lex_state = 98},
  [4677] = {.lex_state = 294, .external_lex_state = 98},
  [4678] = {.lex_state = 294, .external_lex_state = 98},
  [4679] = {.lex_state = 294, .external_lex_state = 98},
  [4680] = {.lex_state = 294, .external_lex_state = 98},
  [4681] = {.lex_state = 294, .external_lex_state = 98},
  [4682] = {.lex_state = 294, .external_lex_state = 98},
  [4683] = {.lex_state = 521, .external_lex_state = 93},
  [4684] = {.lex_state = 521, .external_lex_state = 93},
  [4685] = {.lex_state = 519, .external_lex_state = 94},
  [4686] = {.lex_state = 294, .external_lex_state = 98},
  [4687] = {.lex_state = 521, .external_lex_state = 93},
  [4688] = {.lex_state = 522, .external_lex_state = 66},
  [4689] = {.lex_state = 521, .external_lex_state = 93},
  [4690] = {.lex_state = 294, .external_lex_state = 98},
  [4691] = {.lex_state = 294, .external_lex_state = 98},
  [4692] = {.lex_state = 521, .external_lex_state = 93},
  [4693] = {.lex_state = 294, .external_lex_state = 98},
  [4694] = {.lex_state = 294, .external_lex_state = 98},
  [4695] = {.lex_state = 68, .external_lex_state = 94},
  [4696] = {.lex_state = 294, .external_lex_state = 98},
  [4697] = {.lex_state = 522, .external_lex_state = 94},
  [4698] = {.lex_state = 521, .external_lex_state = 93},
  [4699] = {.lex_state = 521, .external_lex_state = 93},
  [4700] = {.lex_state = 312, .external_lex_state = 99},
  [4701] = {.lex_state = 68, .external_lex_state = 93},
  [4702] = {.lex_state = 520, .external_lex_state = 94},
  [4703] = {.lex_state = 294, .external_lex_state = 98},
  [4704] = {.lex_state = 68, .external_lex_state = 94},
  [4705] = {.lex_state = 521, .external_lex_state = 93},
  [4706] = {.lex_state = 68, .external_lex_state = 93},
  [4707] = {.lex_state = 294, .external_lex_state = 98},
  [4708] = {.lex_state = 238, .external_lex_state = 97},
  [4709] = {.lex_state = 67, .external_lex_state = 92},
  [4710] = {.lex_state = 521, .external_lex_state = 94},
  [4711] = {.lex_state = 521, .external_lex_state = 93},
  [4712] = {.lex_state = 67, .external_lex_state = 92},
  [4713] = {.lex_state = 294, .external_lex_state = 98},
  [4714] = {.lex_state = 294, .external_lex_state = 98},
  [4715] = {.lex_state = 294, .external_lex_state = 98},
  [4716] = {.lex_state = 521, .external_lex_state = 94},
  [4717] = {.lex_state = 521, .external_lex_state = 94},
  [4718] = {.lex_state = 294, .external_lex_state = 98},
  [4719] = {.lex_state = 520, .external_lex_state = 94},
  [4720] = {.lex_state = 294, .external_lex_state = 98},
  [4721] = {.lex_state = 312, .external_lex_state = 99},
  [4722] = {.lex_state = 522, .external_lex_state = 94},
  [4723] = {.lex_state = 519, .external_lex_state = 94},
  [4724] = {.lex_state = 522, .external_lex_state = 94},
  [4725] = {.lex_state = 522, .external_lex_state = 94},
  [4726] = {.lex_state = 68, .external_lex_state = 66},
  [4727] = {.lex_state = 520, .external_lex_state = 94},
  [4728] = {.lex_state = 294, .external_lex_state = 98},
  [4729] = {.lex_state = 294, .external_lex_state = 98},
  [4730] = {.lex_state = 294, .external_lex_state = 98},
  [4731] = {.lex_state = 294, .external_lex_state = 98},
  [4732] = {.lex_state = 294, .external_lex_state = 98},
  [4733] = {.lex_state = 521, .external_lex_state = 93},
  [4734] = {.lex_state = 521, .external_lex_state = 94},
  [4735] = {.lex_state = 238, .external_lex_state = 97},
  [4736] = {.lex_state = 520, .external_lex_state = 94},
  [4737] = {.lex_state = 294, .external_lex_state = 98},
  [4738] = {.lex_state = 294, .external_lex_state = 98},
  [4739] = {.lex_state = 68, .external_lex_state = 93},
  [4740] = {.lex_state = 294, .external_lex_state = 98},
  [4741] = {.lex_state = 312, .external_lex_state = 99},
  [4742] = {.lex_state = 312, .external_lex_state = 99},
  [4743] = {.lex_state = 312, .external_lex_state = 99},
  [4744] = {.lex_state = 294, .external_lex_state = 98},
  [4745] = {.lex_state = 521, .external_lex_state = 94},
  [4746] = {.lex_state = 521, .external_lex_state = 94},
  [4747] = {.lex_state = 520, .external_lex_state = 94},
  [4748] = {.lex_state = 521, .external_lex_state = 93},
  [4749] = {.lex_state = 294, .external_lex_state = 98},
  [4750] = {.lex_state = 519, .external_lex_state = 93},
  [4751] = {.lex_state = 294, .external_lex_state = 98},
  [4752] = {.lex_state = 294, .external_lex_state = 98},
  [4753] = {.lex_state = 68, .external_lex_state = 93},
  [4754] = {.lex_state = 519, .external_lex_state = 93},
  [4755] = {.lex_state = 67, .external_lex_state = 92},
  [4756] = {.lex_state = 294, .external_lex_state = 98},
  [4757] = {.lex_state = 312, .external_lex_state = 99},
  [4758] = {.lex_state = 520, .external_lex_state = 94},
  [4759] = {.lex_state = 522, .external_lex_state = 66},
  [4760] = {.lex_state = 521, .external_lex_state = 93},
  [4761] = {.lex_state = 294, .external_lex_state = 98},
  [4762] = {.lex_state = 68, .external_lex_state = 93},
  [4763] = {.lex_state = 294, .external_lex_state = 98},
  [4764] = {.lex_state = 294, .external_lex_state = 98},
  [4765] = {.lex_state = 294, .external_lex_state = 98},
  [4766] = {.lex_state = 238, .external_lex_state = 97},
  [4767] = {.lex_state = 68, .external_lex_state = 94},
  [4768] = {.lex_state = 294, .external_lex_state = 98},
  [4769] = {.lex_state = 520, .external_lex_state = 94},
  [4770] = {.lex_state = 68, .external_lex_state = 93},
  [4771] = {.lex_state = 521, .external_lex_state = 93},
  [4772] = {.lex_state = 67, .external_lex_state = 94},
  [4773] = {.lex_state = 294, .external_lex_state = 98},
  [4774] = {.lex_state = 521, .external_lex_state = 93},
  [4775] = {.lex_state = 521, .external_lex_state = 95},
  [4776] = {.lex_state = 519, .external_lex_state = 93},
  [4777] = {.lex_state = 519, .external_lex_state = 93},
  [4778] = {.lex_state = 521, .external_lex_state = 94},
  [4779] = {.lex_state = 294, .external_lex_state = 98},
  [4780] = {.lex_state = 68, .external_lex_state = 93},
  [4781] = {.lex_state = 521, .external_lex_state = 95},
  [4782] = {.lex_state = 520, .external_lex_state = 94},
  [4783] = {.lex_state = 521, .external_lex_state = 95},
  [4784] = {.lex_state = 521, .external_lex_state = 66},
  [4785] = {.lex_state = 67, .external_lex_state = 92},
  [4786] = {.lex_state = 67, .external_lex_state = 92},
  [4787] = {.lex_state = 521, .external_lex_state = 66},
  [4788] = {.lex_state = 68, .external_lex_state = 92},
  [4789] = {.lex_state = 67, .external_lex_state = 92},
  [4790] = {.lex_state = 68, .external_lex_state = 93},
  [4791] = {.lex_state = 294, .external_lex_state = 98},
  [4792] = {.lex_state = 68, .external_lex_state = 92},
  [4793] = {.lex_state = 522, .external_lex_state = 94},
  [4794] = {.lex_state = 522, .external_lex_state = 94},
  [4795] = {.lex_state = 521, .external_lex_state = 93},
  [4796] = {.lex_state = 68, .external_lex_state = 93},
  [4797] = {.lex_state = 521, .external_lex_state = 94},
  [4798] = {.lex_state = 294, .external_lex_state = 98},
  [4799] = {.lex_state = 294, .external_lex_state = 98},
  [4800] = {.lex_state = 294, .external_lex_state = 98},
  [4801] = {.lex_state = 238, .external_lex_state = 97},
  [4802] = {.lex_state = 521, .external_lex_state = 95},
  [4803] = {.lex_state = 312, .external_lex_state = 99},
  [4804] = {.lex_state = 68, .external_lex_state = 63},
  [4805] = {.lex_state = 68, .external_lex_state = 93},
  [4806] = {.lex_state = 519, .external_lex_state = 93},
  [4807] = {.lex_state = 238, .external_lex_state = 97},
  [4808] = {.lex_state = 238, .external_lex_state = 97},
  [4809] = {.lex_state = 294, .external_lex_state = 98},
  [4810] = {.lex_state = 238, .external_lex_state = 97},
  [4811] = {.lex_state = 238, .external_lex_state = 97},
  [4812] = {.lex_state = 238, .external_lex_state = 97},
  [4813] = {.lex_state = 294, .external_lex_state = 98},
  [4814] = {.lex_state = 294, .external_lex_state = 98},
  [4815] = {.lex_state = 294, .external_lex_state = 98},
  [4816] = {.lex_state = 519, .external_lex_state = 93},
  [4817] = {.lex_state = 294, .external_lex_state = 98},
  [4818] = {.lex_state = 67, .external_lex_state = 92},
  [4819] = {.lex_state = 294, .external_lex_state = 98},
  [4820] = {.lex_state = 67, .external_lex_state = 92},
  [4821] = {.lex_state = 67, .external_lex_state = 92},
  [4822] = {.lex_state = 519, .external_lex_state = 93},
  [4823] = {.lex_state = 294, .external_lex_state = 98},
  [4824] = {.lex_state = 68, .external_lex_state = 92},
  [4825] = {.lex_state = 238, .external_lex_state = 97},
  [4826] = {.lex_state = 294, .external_lex_state = 98},
  [4827] = {.lex_state = 67, .external_lex_state = 92},
  [4828] = {.lex_state = 521, .external_lex_state = 93},
  [4829] = {.lex_state = 519, .external_lex_state = 94},
  [4830] = {.lex_state = 294, .external_lex_state = 98},
  [4831] = {.lex_state = 68, .external_lex_state = 93},
  [4832] = {.lex_state = 294, .external_lex_state = 98},
  [4833] = {.lex_state = 238, .external_lex_state = 97},
  [4834] = {.lex_state = 519, .external_lex_state = 93},
  [4835] = {.lex_state = 519, .external_lex_state = 93},
  [4836] = {.lex_state = 294, .external_lex_state = 98},
  [4837] = {.lex_state = 238, .external_lex_state = 97},
  [4838] = {.lex_state = 519, .external_lex_state = 93},
  [4839] = {.lex_state = 294, .external_lex_state = 98},
  [4840] = {.lex_state = 68, .external_lex_state = 63},
  [4841] = {.lex_state = 294, .external_lex_state = 98},
  [4842] = {.lex_state = 238, .external_lex_state = 97},
  [4843] = {.lex_state = 294, .external_lex_state = 98},
  [4844] = {.lex_state = 67, .external_lex_state = 92},
  [4845] = {.lex_state = 294, .external_lex_state = 98},
  [4846] = {.lex_state = 522, .external_lex_state = 94},
  [4847] = {.lex_state = 238, .external_lex_state = 97},
  [4848] = {.lex_state = 294, .external_lex_state = 98},
  [4849] = {.lex_state = 68, .external_lex_state = 93},
  [4850] = {.lex_state = 294, .external_lex_state = 98},
  [4851] = {.lex_state = 294, .external_lex_state = 98},
  [4852] = {.lex_state = 68, .external_lex_state = 93},
  [4853] = {.lex_state = 68, .external_lex_state = 93},
  [4854] = {.lex_state = 294, .external_lex_state = 98},
  [4855] = {.lex_state = 68, .external_lex_state = 94},
  [4856] = {.lex_state = 520, .external_lex_state = 94},
  [4857] = {.lex_state = 294, .external_lex_state = 98},
  [4858] = {.lex_state = 520, .external_lex_state = 94},
  [4859] = {.lex_state = 68, .external_lex_state = 93},
  [4860] = {.lex_state = 521, .external_lex_state = 94},
  [4861] = {.lex_state = 520, .external_lex_state = 94},
  [4862] = {.lex_state = 238, .external_lex_state = 97},
  [4863] = {.lex_state = 294, .external_lex_state = 98},
  [4864] = {.lex_state = 67, .external_lex_state = 94},
  [4865] = {.lex_state = 522, .external_lex_state = 94},
  [4866] = {.lex_state = 68, .external_lex_state = 92},
  [4867] = {.lex_state = 68, .external_lex_state = 92},
  [4868] = {.lex_state = 522, .external_lex_state = 94},
  [4869] = {.lex_state = 522, .external_lex_state = 94},
  [4870] = {.lex_state = 68, .external_lex_state = 66},
  [4871] = {.lex_state = 522, .external_lex_state = 94},
  [4872] = {.lex_state = 68, .external_lex_state = 92},
  [4873] = {.lex_state = 522, .external_lex_state = 94},
  [4874] = {.lex_state = 522, .external_lex_state = 94},
  [4875] = {.lex_state = 522, .external_lex_state = 94},
  [4876] = {.lex_state = 522, .external_lex_state = 94},
  [4877] = {.lex_state = 522, .external_lex_state = 94},
  [4878] = {.lex_state = 68, .external_lex_state = 92},
  [4879] = {.lex_state = 68, .external_lex_state = 92},
  [4880] = {.lex_state = 68, .external_lex_state = 92},
  [4881] = {.lex_state = 68, .external_lex_state = 92},
  [4882] = {.lex_state = 68, .external_lex_state = 92},
  [4883] = {.lex_state = 68, .external_lex_state = 92},
  [4884] = {.lex_state = 68, .external_lex_state = 92},
  [4885] = {.lex_state = 67, .external_lex_state = 94},
  [4886] = {.lex_state = 522, .external_lex_state = 94},
  [4887] = {.lex_state = 522, .external_lex_state = 94},
  [4888] = {.lex_state = 68, .external_lex_state = 92},
  [4889] = {.lex_state = 68, .external_lex_state = 92},
  [4890] = {.lex_state = 68, .external_lex_state = 92},
  [4891] = {.lex_state = 312, .external_lex_state = 99},
  [4892] = {.lex_state = 522, .external_lex_state = 94},
  [4893] = {.lex_state = 67, .external_lex_state = 94},
  [4894] = {.lex_state = 68, .external_lex_state = 92},
  [4895] = {.lex_state = 522, .external_lex_state = 94},
  [4896] = {.lex_state = 521, .external_lex_state = 93},
  [4897] = {.lex_state = 521, .external_lex_state = 93},
  [4898] = {.lex_state = 522, .external_lex_state = 94},
  [4899] = {.lex_state = 68, .external_lex_state = 92},
  [4900] = {.lex_state = 522, .external_lex_state = 94},
  [4901] = {.lex_state = 521, .external_lex_state = 93},
  [4902] = {.lex_state = 67, .external_lex_state = 94},
  [4903] = {.lex_state = 522, .external_lex_state = 94},
  [4904] = {.lex_state = 68, .external_lex_state = 94},
  [4905] = {.lex_state = 522, .external_lex_state = 94},
  [4906] = {.lex_state = 521, .external_lex_state = 93},
  [4907] = {.lex_state = 522, .external_lex_state = 94},
  [4908] = {.lex_state = 68, .external_lex_state = 92},
  [4909] = {.lex_state = 522, .external_lex_state = 94},
  [4910] = {.lex_state = 68, .external_lex_state = 94},
  [4911] = {.lex_state = 522, .external_lex_state = 94},
  [4912] = {.lex_state = 522, .external_lex_state = 94},
  [4913] = {.lex_state = 68, .external_lex_state = 94},
  [4914] = {.lex_state = 67, .external_lex_state = 94},
  [4915] = {.lex_state = 522, .external_lex_state = 94},
  [4916] = {.lex_state = 68, .external_lex_state = 92},
  [4917] = {.lex_state = 67, .external_lex_state = 94},
  [4918] = {.lex_state = 312, .external_lex_state = 99},
  [4919] = {.lex_state = 521, .external_lex_state = 93},
  [4920] = {.lex_state = 522, .external_lex_state = 94},
  [4921] = {.lex_state = 521, .external_lex_state = 93},
  [4922] = {.lex_state = 68, .external_lex_state = 92},
  [4923] = {.lex_state = 68, .external_lex_state = 92},
  [4924] = {.lex_state = 312, .external_lex_state = 99},
  [4925] = {.lex_state = 68, .external_lex_state = 66},
  [4926] = {.lex_state = 68, .external_lex_state = 92},
  [4927] = {.lex_state = 68, .external_lex_state = 92},
  [4928] = {.lex_state = 521, .external_lex_state = 94},
  [4929] = {.lex_state = 68, .external_lex_state = 92},
  [4930] = {.lex_state = 68, .external_lex_state = 92},
  [4931] = {.lex_state = 522, .external_lex_state = 94},
  [4932] = {.lex_state = 68, .external_lex_state = 92},
  [4933] = {.lex_state = 68, .external_lex_state = 92},
  [4934] = {.lex_state = 312, .external_lex_state = 99},
  [4935] = {.lex_state = 522, .external_lex_state = 94},
  [4936] = {.lex_state = 68, .external_lex_state = 92},
  [4937] = {.lex_state = 68, .external_lex_state = 92},
  [4938] = {.lex_state = 522, .external_lex_state = 94},
  [4939] = {.lex_state = 68, .external_lex_state = 94},
  [4940] = {.lex_state = 522, .external_lex_state = 94},
  [4941] = {.lex_state = 521, .external_lex_state = 93},
  [4942] = {.lex_state = 68, .external_lex_state = 94},
  [4943] = {.lex_state = 68, .external_lex_state = 92},
  [4944] = {.lex_state = 68, .external_lex_state = 92},
  [4945] = {.lex_state = 68, .external_lex_state = 92},
  [4946] = {.lex_state = 67, .external_lex_state = 94},
  [4947] = {.lex_state = 68, .external_lex_state = 92},
  [4948] = {.lex_state = 522, .external_lex_state = 94},
  [4949] = {.lex_state = 68, .external_lex_state = 92},
  [4950] = {.lex_state = 67, .external_lex_state = 94},
  [4951] = {.lex_state = 521, .external_lex_state = 94},
  [4952] = {.lex_state = 522, .external_lex_state = 94},
  [4953] = {.lex_state = 67, .external_lex_state = 94},
  [4954] = {.lex_state = 522, .external_lex_state = 94},
  [4955] = {.lex_state = 68, .external_lex_state = 92},
  [4956] = {.lex_state = 68, .external_lex_state = 92},
  [4957] = {.lex_state = 68, .external_lex_state = 92},
  [4958] = {.lex_state = 522, .external_lex_state = 94},
  [4959] = {.lex_state = 68, .external_lex_state = 92},
  [4960] = {.lex_state = 522, .external_lex_state = 94},
  [4961] = {.lex_state = 522, .external_lex_state = 94},
  [4962] = {.lex_state = 67, .external_lex_state = 94},
  [4963] = {.lex_state = 68, .external_lex_state = 92},
  [4964] = {.lex_state = 67, .external_lex_state = 94},
  [4965] = {.lex_state = 522, .external_lex_state = 94},
  [4966] = {.lex_state = 68, .external_lex_state = 92},
  [4967] = {.lex_state = 312, .external_lex_state = 99},
  [4968] = {.lex_state = 522, .external_lex_state = 94},
  [4969] = {.lex_state = 68, .external_lex_state = 92},
  [4970] = {.lex_state = 522, .external_lex_state = 94},
  [4971] = {.lex_state = 68, .external_lex_state = 94},
  [4972] = {.lex_state = 312, .external_lex_state = 99},
  [4973] = {.lex_state = 522, .external_lex_state = 94},
  [4974] = {.lex_state = 312, .external_lex_state = 99},
  [4975] = {.lex_state = 237, .external_lex_state = 73},
  [4976] = {.lex_state = 68, .external_lex_state = 92},
  [4977] = {.lex_state = 68, .external_lex_state = 92},
  [4978] = {.lex_state = 68, .external_lex_state = 92},
  [4979] = {.lex_state = 522, .external_lex_state = 94},
  [4980] = {.lex_state = 68, .external_lex_state = 92},
  [4981] = {.lex_state = 68, .external_lex_state = 92},
  [4982] = {.lex_state = 67, .external_lex_state = 94},
  [4983] = {.lex_state = 522, .external_lex_state = 94},
  [4984] = {.lex_state = 68, .external_lex_state = 92},
  [4985] = {.lex_state = 68, .external_lex_state = 92},
  [4986] = {.lex_state = 521, .external_lex_state = 94},
  [4987] = {.lex_state = 522, .external_lex_state = 94},
  [4988] = {.lex_state = 522, .external_lex_state = 94},
  [4989] = {.lex_state = 522, .external_lex_state = 94},
  [4990] = {.lex_state = 68, .external_lex_state = 66},
  [4991] = {.lex_state = 68, .external_lex_state = 92},
  [4992] = {.lex_state = 521, .external_lex_state = 93},
  [4993] = {.lex_state = 521, .external_lex_state = 94},
  [4994] = {.lex_state = 522, .external_lex_state = 94},
  [4995] = {.lex_state = 519, .external_lex_state = 94},
  [4996] = {.lex_state = 521, .external_lex_state = 94},
  [4997] = {.lex_state = 521, .external_lex_state = 94},
  [4998] = {.lex_state = 521, .external_lex_state = 94},
  [4999] = {.lex_state = 312, .external_lex_state = 99},
  [5000] = {.lex_state = 519, .external_lex_state = 94},
  [5001] = {.lex_state = 519, .external_lex_state = 94},
  [5002] = {.lex_state = 522, .external_lex_state = 94},
  [5003] = {.lex_state = 68, .external_lex_state = 92},
  [5004] = {.lex_state = 68, .external_lex_state = 92},
  [5005] = {.lex_state = 522, .external_lex_state = 94},
  [5006] = {.lex_state = 68, .external_lex_state = 92},
  [5007] = {.lex_state = 521, .external_lex_state = 94},
  [5008] = {.lex_state = 68, .external_lex_state = 92},
  [5009] = {.lex_state = 522, .external_lex_state = 94},
  [5010] = {.lex_state = 522, .external_lex_state = 94},
  [5011] = {.lex_state = 522, .external_lex_state = 94},
  [5012] = {.lex_state = 68, .external_lex_state = 92},
  [5013] = {.lex_state = 68, .external_lex_state = 92},
  [5014] = {.lex_state = 522, .external_lex_state = 94},
  [5015] = {.lex_state = 312, .external_lex_state = 99},
  [5016] = {.lex_state = 68, .external_lex_state = 92},
  [5017] = {.lex_state = 522, .external_lex_state = 94},
  [5018] = {.lex_state = 67, .external_lex_state = 94},
  [5019] = {.lex_state = 67, .external_lex_state = 94},
  [5020] = {.lex_state = 67, .external_lex_state = 94},
  [5021] = {.lex_state = 237, .external_lex_state = 73},
  [5022] = {.lex_state = 522, .external_lex_state = 94},
  [5023] = {.lex_state = 522, .external_lex_state = 94},
  [5024] = {.lex_state = 67, .external_lex_state = 94},
  [5025] = {.lex_state = 67, .external_lex_state = 94},
  [5026] = {.lex_state = 522, .external_lex_state = 94},
  [5027] = {.lex_state = 522, .external_lex_state = 94},
  [5028] = {.lex_state = 68, .external_lex_state = 92},
  [5029] = {.lex_state = 522, .external_lex_state = 94},
  [5030] = {.lex_state = 522, .external_lex_state = 94},
  [5031] = {.lex_state = 521, .external_lex_state = 93},
  [5032] = {.lex_state = 68, .external_lex_state = 92},
  [5033] = {.lex_state = 521, .external_lex_state = 93},
  [5034] = {.lex_state = 227, .external_lex_state = 97},
  [5035] = {.lex_state = 68, .external_lex_state = 94},
  [5036] = {.lex_state = 521, .external_lex_state = 94},
  [5037] = {.lex_state = 68, .external_lex_state = 94},
  [5038] = {.lex_state = 301, .external_lex_state = 100},
  [5039] = {.lex_state = 68, .external_lex_state = 94},
  [5040] = {.lex_state = 522, .external_lex_state = 94},
  [5041] = {.lex_state = 301, .external_lex_state = 100},
  [5042] = {.lex_state = 68, .external_lex_state = 94},
  [5043] = {.lex_state = 68, .external_lex_state = 94},
  [5044] = {.lex_state = 237, .external_lex_state = 73},
  [5045] = {.lex_state = 301, .external_lex_state = 100},
  [5046] = {.lex_state = 68, .external_lex_state = 94},
  [5047] = {.lex_state = 68, .external_lex_state = 94},
  [5048] = {.lex_state = 227, .external_lex_state = 97},
  [5049] = {.lex_state = 68, .external_lex_state = 94},
  [5050] = {.lex_state = 68, .external_lex_state = 94},
  [5051] = {.lex_state = 521, .external_lex_state = 94},
  [5052] = {.lex_state = 68, .external_lex_state = 94},
  [5053] = {.lex_state = 68, .external_lex_state = 94},
  [5054] = {.lex_state = 521, .external_lex_state = 94},
  [5055] = {.lex_state = 227, .external_lex_state = 97},
  [5056] = {.lex_state = 67, .external_lex_state = 94},
  [5057] = {.lex_state = 227, .external_lex_state = 97},
  [5058] = {.lex_state = 67, .external_lex_state = 94},
  [5059] = {.lex_state = 301, .external_lex_state = 100},
  [5060] = {.lex_state = 68, .external_lex_state = 94},
  [5061] = {.lex_state = 68, .external_lex_state = 94},
  [5062] = {.lex_state = 227, .external_lex_state = 97},
  [5063] = {.lex_state = 68, .external_lex_state = 94},
  [5064] = {.lex_state = 324, .external_lex_state = 101},
  [5065] = {.lex_state = 301, .external_lex_state = 100},
  [5066] = {.lex_state = 68, .external_lex_state = 94},
  [5067] = {.lex_state = 68, .external_lex_state = 94},
  [5068] = {.lex_state = 68, .external_lex_state = 94},
  [5069] = {.lex_state = 68, .external_lex_state = 94},
  [5070] = {.lex_state = 522, .external_lex_state = 94},
  [5071] = {.lex_state = 68, .external_lex_state = 94},
  [5072] = {.lex_state = 68, .external_lex_state = 94},
  [5073] = {.lex_state = 68, .external_lex_state = 94},
  [5074] = {.lex_state = 68, .external_lex_state = 94},
  [5075] = {.lex_state = 68, .external_lex_state = 94},
  [5076] = {.lex_state = 68, .external_lex_state = 94},
  [5077] = {.lex_state = 68, .external_lex_state = 94},
  [5078] = {.lex_state = 68, .external_lex_state = 94},
  [5079] = {.lex_state = 521, .external_lex_state = 94},
  [5080] = {.lex_state = 68, .external_lex_state = 94},
  [5081] = {.lex_state = 522, .external_lex_state = 94},
  [5082] = {.lex_state = 521, .external_lex_state = 94},
  [5083] = {.lex_state = 68, .external_lex_state = 94},
  [5084] = {.lex_state = 68, .external_lex_state = 94},
  [5085] = {.lex_state = 68, .external_lex_state = 94},
  [5086] = {.lex_state = 68, .external_lex_state = 94},
  [5087] = {.lex_state = 68, .external_lex_state = 94},
  [5088] = {.lex_state = 68, .external_lex_state = 94},
  [5089] = {.lex_state = 68, .external_lex_state = 94},
  [5090] = {.lex_state = 68, .external_lex_state = 94},
  [5091] = {.lex_state = 68, .external_lex_state = 94},
  [5092] = {.lex_state = 68, .external_lex_state = 94},
  [5093] = {.lex_state = 68, .external_lex_state = 94},
  [5094] = {.lex_state = 68, .external_lex_state = 94},
  [5095] = {.lex_state = 521, .external_lex_state = 94},
  [5096] = {.lex_state = 68, .external_lex_state = 94},
  [5097] = {.lex_state = 301, .external_lex_state = 100},
  [5098] = {.lex_state = 68, .external_lex_state = 94},
  [5099] = {.lex_state = 68, .external_lex_state = 94},
  [5100] = {.lex_state = 68, .external_lex_state = 94},
  [5101] = {.lex_state = 68, .external_lex_state = 94},
  [5102] = {.lex_state = 522, .external_lex_state = 94},
  [5103] = {.lex_state = 68, .external_lex_state = 94},
  [5104] = {.lex_state = 68, .external_lex_state = 94},
  [5105] = {.lex_state = 68, .external_lex_state = 94},
  [5106] = {.lex_state = 68, .external_lex_state = 94},
  [5107] = {.lex_state = 68, .external_lex_state = 94},
  [5108] = {.lex_state = 521, .external_lex_state = 94},
  [5109] = {.lex_state = 67, .external_lex_state = 93},
  [5110] = {.lex_state = 68, .external_lex_state = 95},
  [5111] = {.lex_state = 68, .external_lex_state = 94},
  [5112] = {.lex_state = 68, .external_lex_state = 94},
  [5113] = {.lex_state = 68, .external_lex_state = 94},
  [5114] = {.lex_state = 68, .external_lex_state = 94},
  [5115] = {.lex_state = 521, .external_lex_state = 94},
  [5116] = {.lex_state = 68, .external_lex_state = 94},
  [5117] = {.lex_state = 67, .external_lex_state = 94},
  [5118] = {.lex_state = 67, .external_lex_state = 94},
  [5119] = {.lex_state = 68, .external_lex_state = 94},
  [5120] = {.lex_state = 68, .external_lex_state = 94},
  [5121] = {.lex_state = 311, .external_lex_state = 67},
  [5122] = {.lex_state = 521, .external_lex_state = 94},
  [5123] = {.lex_state = 301, .external_lex_state = 100},
  [5124] = {.lex_state = 324, .external_lex_state = 101},
  [5125] = {.lex_state = 311, .external_lex_state = 67},
  [5126] = {.lex_state = 311, .external_lex_state = 67},
  [5127] = {.lex_state = 324, .external_lex_state = 101},
  [5128] = {.lex_state = 67, .external_lex_state = 93},
  [5129] = {.lex_state = 67, .external_lex_state = 93},
  [5130] = {.lex_state = 67, .external_lex_state = 93},
  [5131] = {.lex_state = 67, .external_lex_state = 93},
  [5132] = {.lex_state = 311, .external_lex_state = 67},
  [5133] = {.lex_state = 67, .external_lex_state = 93},
  [5134] = {.lex_state = 67, .external_lex_state = 93},
  [5135] = {.lex_state = 67, .external_lex_state = 93},
  [5136] = {.lex_state = 67, .external_lex_state = 93},
  [5137] = {.lex_state = 68, .external_lex_state = 95},
  [5138] = {.lex_state = 68, .external_lex_state = 95},
  [5139] = {.lex_state = 311, .external_lex_state = 67},
  [5140] = {.lex_state = 324, .external_lex_state = 101},
  [5141] = {.lex_state = 311, .external_lex_state = 67},
  [5142] = {.lex_state = 312, .external_lex_state = 99},
  [5143] = {.lex_state = 311, .external_lex_state = 67},
  [5144] = {.lex_state = 311, .external_lex_state = 67},
  [5145] = {.lex_state = 311, .external_lex_state = 67},
  [5146] = {.lex_state = 311, .external_lex_state = 67},
  [5147] = {.lex_state = 68, .external_lex_state = 66},
  [5148] = {.lex_state = 68, .external_lex_state = 66},
  [5149] = {.lex_state = 301, .external_lex_state = 100},
  [5150] = {.lex_state = 301, .external_lex_state = 100},
  [5151] = {.lex_state = 301, .external_lex_state = 100},
  [5152] = {.lex_state = 301, .external_lex_state = 100},
  [5153] = {.lex_state = 301, .external_lex_state = 100},
  [5154] = {.lex_state = 301, .external_lex_state = 100},
  [5155] = {.lex_state = 301, .external_lex_state = 100},
  [5156] = {.lex_state = 301, .external_lex_state = 100},
  [5157] = {.lex_state = 301, .external_lex_state = 100},
  [5158] = {.lex_state = 301, .external_lex_state = 100},
  [5159] = {.lex_state = 301, .external_lex_state = 100},
  [5160] = {.lex_state = 301, .external_lex_state = 100},
  [5161] = {.lex_state = 301, .external_lex_state = 100},
  [5162] = {.lex_state = 311, .external_lex_state = 67},
  [5163] = {.lex_state = 521, .external_lex_state = 94},
  [5164] = {.lex_state = 311, .external_lex_state = 67},
  [5165] = {.lex_state = 301, .external_lex_state = 100},
  [5166] = {.lex_state = 311, .external_lex_state = 67},
  [5167] = {.lex_state = 289, .external_lex_state = 102},
  [5168] = {.lex_state = 312, .external_lex_state = 99},
  [5169] = {.lex_state = 311, .external_lex_state = 67},
  [5170] = {.lex_state = 324, .external_lex_state = 101},
  [5171] = {.lex_state = 301, .external_lex_state = 100},
  [5172] = {.lex_state = 311, .external_lex_state = 67},
  [5173] = {.lex_state = 324, .external_lex_state = 101},
  [5174] = {.lex_state = 321, .external_lex_state = 101},
  [5175] = {.lex_state = 324, .external_lex_state = 101},
  [5176] = {.lex_state = 67, .external_lex_state = 94},
  [5177] = {.lex_state = 311, .external_lex_state = 67},
  [5178] = {.lex_state = 324, .external_lex_state = 101},
  [5179] = {.lex_state = 324, .external_lex_state = 101},
  [5180] = {.lex_state = 324, .external_lex_state = 101},
  [5181] = {.lex_state = 67, .external_lex_state = 94},
  [5182] = {.lex_state = 324, .external_lex_state = 101},
  [5183] = {.lex_state = 311, .external_lex_state = 67},
  [5184] = {.lex_state = 289, .external_lex_state = 103},
  [5185] = {.lex_state = 312, .external_lex_state = 99},
  [5186] = {.lex_state = 312, .external_lex_state = 99},
  [5187] = {.lex_state = 312, .external_lex_state = 99},
  [5188] = {.lex_state = 324, .external_lex_state = 101},
  [5189] = {.lex_state = 521, .external_lex_state = 94},
  [5190] = {.lex_state = 312, .external_lex_state = 99},
  [5191] = {.lex_state = 312, .external_lex_state = 99},
  [5192] = {.lex_state = 312, .external_lex_state = 99},
  [5193] = {.lex_state = 311, .external_lex_state = 67},
  [5194] = {.lex_state = 312, .external_lex_state = 99},
  [5195] = {.lex_state = 289, .external_lex_state = 103},
  [5196] = {.lex_state = 312, .external_lex_state = 99},
  [5197] = {.lex_state = 312, .external_lex_state = 99},
  [5198] = {.lex_state = 312, .external_lex_state = 99},
  [5199] = {.lex_state = 301, .external_lex_state = 100},
  [5200] = {.lex_state = 301, .external_lex_state = 100},
  [5201] = {.lex_state = 311, .external_lex_state = 67},
  [5202] = {.lex_state = 312, .external_lex_state = 99},
  [5203] = {.lex_state = 312, .external_lex_state = 99},
  [5204] = {.lex_state = 301, .external_lex_state = 100},
  [5205] = {.lex_state = 289, .external_lex_state = 102},
  [5206] = {.lex_state = 68, .external_lex_state = 93},
  [5207] = {.lex_state = 67, .external_lex_state = 93},
  [5208] = {.lex_state = 67, .external_lex_state = 93},
  [5209] = {.lex_state = 68, .external_lex_state = 95},
  [5210] = {.lex_state = 68, .external_lex_state = 95},
  [5211] = {.lex_state = 66, .external_lex_state = 104},
  [5212] = {.lex_state = 66, .external_lex_state = 104},
  [5213] = {.lex_state = 289, .external_lex_state = 103},
  [5214] = {.lex_state = 312, .external_lex_state = 99},
  [5215] = {.lex_state = 311, .external_lex_state = 67},
  [5216] = {.lex_state = 521, .external_lex_state = 94},
  [5217] = {.lex_state = 312, .external_lex_state = 99},
  [5218] = {.lex_state = 312, .external_lex_state = 99},
  [5219] = {.lex_state = 312, .external_lex_state = 99},
  [5220] = {.lex_state = 312, .external_lex_state = 99},
  [5221] = {.lex_state = 322, .external_lex_state = 105},
  [5222] = {.lex_state = 312, .external_lex_state = 99},
  [5223] = {.lex_state = 312, .external_lex_state = 99},
  [5224] = {.lex_state = 521, .external_lex_state = 94},
  [5225] = {.lex_state = 312, .external_lex_state = 99},
  [5226] = {.lex_state = 312, .external_lex_state = 99},
  [5227] = {.lex_state = 311, .external_lex_state = 67},
  [5228] = {.lex_state = 312, .external_lex_state = 99},
  [5229] = {.lex_state = 312, .external_lex_state = 99},
  [5230] = {.lex_state = 312, .external_lex_state = 99},
  [5231] = {.lex_state = 312, .external_lex_state = 99},
  [5232] = {.lex_state = 66, .external_lex_state = 104},
  [5233] = {.lex_state = 312, .external_lex_state = 99},
  [5234] = {.lex_state = 312, .external_lex_state = 99},
  [5235] = {.lex_state = 312, .external_lex_state = 99},
  [5236] = {.lex_state = 312, .external_lex_state = 99},
  [5237] = {.lex_state = 312, .external_lex_state = 99},
  [5238] = {.lex_state = 312, .external_lex_state = 99},
  [5239] = {.lex_state = 312, .external_lex_state = 99},
  [5240] = {.lex_state = 325, .external_lex_state = 106},
  [5241] = {.lex_state = 312, .external_lex_state = 99},
  [5242] = {.lex_state = 312, .external_lex_state = 99},
  [5243] = {.lex_state = 312, .external_lex_state = 99},
  [5244] = {.lex_state = 312, .external_lex_state = 99},
  [5245] = {.lex_state = 311, .external_lex_state = 67},
  [5246] = {.lex_state = 312, .external_lex_state = 99},
  [5247] = {.lex_state = 311, .external_lex_state = 67},
  [5248] = {.lex_state = 521, .external_lex_state = 94},
  [5249] = {.lex_state = 313, .external_lex_state = 107},
  [5250] = {.lex_state = 311, .external_lex_state = 67},
  [5251] = {.lex_state = 298, .external_lex_state = 47},
  [5252] = {.lex_state = 298, .external_lex_state = 47},
  [5253] = {.lex_state = 321, .external_lex_state = 105},
  [5254] = {.lex_state = 298, .external_lex_state = 47},
  [5255] = {.lex_state = 324, .external_lex_state = 101},
  [5256] = {.lex_state = 321, .external_lex_state = 105},
  [5257] = {.lex_state = 298, .external_lex_state = 47},
  [5258] = {.lex_state = 289, .external_lex_state = 102},
  [5259] = {.lex_state = 298, .external_lex_state = 47},
  [5260] = {.lex_state = 66, .external_lex_state = 104},
  [5261] = {.lex_state = 298, .external_lex_state = 47},
  [5262] = {.lex_state = 324, .external_lex_state = 101},
  [5263] = {.lex_state = 324, .external_lex_state = 101},
  [5264] = {.lex_state = 298, .external_lex_state = 47},
  [5265] = {.lex_state = 298, .external_lex_state = 47},
  [5266] = {.lex_state = 312, .external_lex_state = 102},
  [5267] = {.lex_state = 322, .external_lex_state = 106},
  [5268] = {.lex_state = 298, .external_lex_state = 47},
  [5269] = {.lex_state = 324, .external_lex_state = 101},
  [5270] = {.lex_state = 324, .external_lex_state = 106},
  [5271] = {.lex_state = 324, .external_lex_state = 106},
  [5272] = {.lex_state = 298, .external_lex_state = 47},
  [5273] = {.lex_state = 298, .external_lex_state = 47},
  [5274] = {.lex_state = 298, .external_lex_state = 78},
  [5275] = {.lex_state = 298, .external_lex_state = 47},
  [5276] = {.lex_state = 324, .external_lex_state = 106},
  [5277] = {.lex_state = 298, .external_lex_state = 47},
  [5278] = {.lex_state = 321, .external_lex_state = 101},
  [5279] = {.lex_state = 324, .external_lex_state = 101},
  [5280] = {.lex_state = 298, .external_lex_state = 47},
  [5281] = {.lex_state = 298, .external_lex_state = 47},
  [5282] = {.lex_state = 298, .external_lex_state = 47},
  [5283] = {.lex_state = 298, .external_lex_state = 47},
  [5284] = {.lex_state = 324, .external_lex_state = 101},
  [5285] = {.lex_state = 298, .external_lex_state = 47},
  [5286] = {.lex_state = 68, .external_lex_state = 93},
  [5287] = {.lex_state = 68, .external_lex_state = 93},
  [5288] = {.lex_state = 324, .external_lex_state = 101},
  [5289] = {.lex_state = 312, .external_lex_state = 102},
  [5290] = {.lex_state = 321, .external_lex_state = 101},
  [5291] = {.lex_state = 298, .external_lex_state = 47},
  [5292] = {.lex_state = 289, .external_lex_state = 102},
  [5293] = {.lex_state = 68, .external_lex_state = 93},
  [5294] = {.lex_state = 324, .external_lex_state = 101},
  [5295] = {.lex_state = 298, .external_lex_state = 47},
  [5296] = {.lex_state = 298, .external_lex_state = 47},
  [5297] = {.lex_state = 321, .external_lex_state = 105},
  [5298] = {.lex_state = 324, .external_lex_state = 101},
  [5299] = {.lex_state = 298, .external_lex_state = 47},
  [5300] = {.lex_state = 324, .external_lex_state = 106},
  [5301] = {.lex_state = 324, .external_lex_state = 106},
  [5302] = {.lex_state = 298, .external_lex_state = 47},
  [5303] = {.lex_state = 324, .external_lex_state = 106},
  [5304] = {.lex_state = 298, .external_lex_state = 47},
  [5305] = {.lex_state = 68, .external_lex_state = 93},
  [5306] = {.lex_state = 68, .external_lex_state = 93},
  [5307] = {.lex_state = 298, .external_lex_state = 47},
  [5308] = {.lex_state = 298, .external_lex_state = 47},
  [5309] = {.lex_state = 298, .external_lex_state = 47},
  [5310] = {.lex_state = 298, .external_lex_state = 47},
  [5311] = {.lex_state = 298, .external_lex_state = 47},
  [5312] = {.lex_state = 68, .external_lex_state = 93},
  [5313] = {.lex_state = 298, .external_lex_state = 47},
  [5314] = {.lex_state = 298, .external_lex_state = 47},
  [5315] = {.lex_state = 312, .external_lex_state = 102},
  [5316] = {.lex_state = 298, .external_lex_state = 47},
  [5317] = {.lex_state = 312, .external_lex_state = 102},
  [5318] = {.lex_state = 68, .external_lex_state = 93},
  [5319] = {.lex_state = 298, .external_lex_state = 47},
  [5320] = {.lex_state = 298, .external_lex_state = 47},
  [5321] = {.lex_state = 298, .external_lex_state = 47},
  [5322] = {.lex_state = 321, .external_lex_state = 101},
  [5323] = {.lex_state = 324, .external_lex_state = 106},
  [5324] = {.lex_state = 298, .external_lex_state = 47},
  [5325] = {.lex_state = 298, .external_lex_state = 47},
  [5326] = {.lex_state = 298, .external_lex_state = 47},
  [5327] = {.lex_state = 324, .external_lex_state = 101},
  [5328] = {.lex_state = 321, .external_lex_state = 101},
  [5329] = {.lex_state = 298, .external_lex_state = 47},
  [5330] = {.lex_state = 298, .external_lex_state = 47},
  [5331] = {.lex_state = 321, .external_lex_state = 101},
  [5332] = {.lex_state = 298, .external_lex_state = 47},
  [5333] = {.lex_state = 298, .external_lex_state = 47},
  [5334] = {.lex_state = 298, .external_lex_state = 47},
  [5335] = {.lex_state = 298, .external_lex_state = 47},
  [5336] = {.lex_state = 298, .external_lex_state = 47},
  [5337] = {.lex_state = 324, .external_lex_state = 101},
  [5338] = {.lex_state = 298, .external_lex_state = 47},
  [5339] = {.lex_state = 298, .external_lex_state = 47},
  [5340] = {.lex_state = 298, .external_lex_state = 47},
  [5341] = {.lex_state = 298, .external_lex_state = 47},
  [5342] = {.lex_state = 298, .external_lex_state = 47},
  [5343] = {.lex_state = 298, .external_lex_state = 47},
  [5344] = {.lex_state = 298, .external_lex_state = 47},
  [5345] = {.lex_state = 298, .external_lex_state = 47},
  [5346] = {.lex_state = 298, .external_lex_state = 47},
  [5347] = {.lex_state = 321, .external_lex_state = 101},
  [5348] = {.lex_state = 321, .external_lex_state = 101},
  [5349] = {.lex_state = 321, .external_lex_state = 101},
  [5350] = {.lex_state = 312, .external_lex_state = 102},
  [5351] = {.lex_state = 198, .external_lex_state = 73},
  [5352] = {.lex_state = 324, .external_lex_state = 106},
  [5353] = {.lex_state = 298, .external_lex_state = 47},
  [5354] = {.lex_state = 298, .external_lex_state = 47},
  [5355] = {.lex_state = 298, .external_lex_state = 47},
  [5356] = {.lex_state = 312, .external_lex_state = 102},
  [5357] = {.lex_state = 289, .external_lex_state = 102},
  [5358] = {.lex_state = 312, .external_lex_state = 102},
  [5359] = {.lex_state = 324, .external_lex_state = 106},
  [5360] = {.lex_state = 312, .external_lex_state = 81},
  [5361] = {.lex_state = 312, .external_lex_state = 102},
  [5362] = {.lex_state = 289, .external_lex_state = 103},
  [5363] = {.lex_state = 289, .external_lex_state = 103},
  [5364] = {.lex_state = 324, .external_lex_state = 101},
  [5365] = {.lex_state = 298, .external_lex_state = 78},
  [5366] = {.lex_state = 68, .external_lex_state = 94},
  [5367] = {.lex_state = 67, .external_lex_state = 94},
  [5368] = {.lex_state = 68, .external_lex_state = 94},
  [5369] = {.lex_state = 68, .external_lex_state = 94},
  [5370] = {.lex_state = 68, .external_lex_state = 94},
  [5371] = {.lex_state = 321, .external_lex_state = 101},
  [5372] = {.lex_state = 67, .external_lex_state = 94},
  [5373] = {.lex_state = 324, .external_lex_state = 106},
  [5374] = {.lex_state = 321, .external_lex_state = 105},
  [5375] = {.lex_state = 67, .external_lex_state = 94},
  [5376] = {.lex_state = 289, .external_lex_state = 102},
  [5377] = {.lex_state = 321, .external_lex_state = 101},
  [5378] = {.lex_state = 324, .external_lex_state = 106},
  [5379] = {.lex_state = 321, .external_lex_state = 101},
  [5380] = {.lex_state = 324, .external_lex_state = 101},
  [5381] = {.lex_state = 68, .external_lex_state = 94},
  [5382] = {.lex_state = 312, .external_lex_state = 102},
  [5383] = {.lex_state = 68, .external_lex_state = 94},
  [5384] = {.lex_state = 298, .external_lex_state = 47},
  [5385] = {.lex_state = 312, .external_lex_state = 81},
  [5386] = {.lex_state = 298, .external_lex_state = 47},
  [5387] = {.lex_state = 324, .external_lex_state = 101},
  [5388] = {.lex_state = 289, .external_lex_state = 102},
  [5389] = {.lex_state = 324, .external_lex_state = 101},
  [5390] = {.lex_state = 321, .external_lex_state = 105},
  [5391] = {.lex_state = 298, .external_lex_state = 47},
  [5392] = {.lex_state = 68, .external_lex_state = 93},
  [5393] = {.lex_state = 68, .external_lex_state = 93},
  [5394] = {.lex_state = 312, .external_lex_state = 102},
  [5395] = {.lex_state = 298, .external_lex_state = 47},
  [5396] = {.lex_state = 298, .external_lex_state = 47},
  [5397] = {.lex_state = 298, .external_lex_state = 47},
  [5398] = {.lex_state = 68, .external_lex_state = 93},
  [5399] = {.lex_state = 324, .external_lex_state = 101},
  [5400] = {.lex_state = 324, .external_lex_state = 101},
  [5401] = {.lex_state = 298, .external_lex_state = 47},
  [5402] = {.lex_state = 324, .external_lex_state = 101},
  [5403] = {.lex_state = 324, .external_lex_state = 101},
  [5404] = {.lex_state = 324, .external_lex_state = 106},
  [5405] = {.lex_state = 289, .external_lex_state = 103},
  [5406] = {.lex_state = 312, .external_lex_state = 103},
  [5407] = {.lex_state = 321, .external_lex_state = 106},
  [5408] = {.lex_state = 322, .external_lex_state = 81},
  [5409] = {.lex_state = 321, .external_lex_state = 105},
  [5410] = {.lex_state = 289, .external_lex_state = 102},
  [5411] = {.lex_state = 321, .external_lex_state = 106},
  [5412] = {.lex_state = 321, .external_lex_state = 106},
  [5413] = {.lex_state = 321, .external_lex_state = 105},
  [5414] = {.lex_state = 321, .external_lex_state = 105},
  [5415] = {.lex_state = 321, .external_lex_state = 106},
  [5416] = {.lex_state = 298, .external_lex_state = 78},
  [5417] = {.lex_state = 321, .external_lex_state = 106},
  [5418] = {.lex_state = 321, .external_lex_state = 106},
  [5419] = {.lex_state = 312, .external_lex_state = 103},
  [5420] = {.lex_state = 321, .external_lex_state = 106},
  [5421] = {.lex_state = 312, .external_lex_state = 103},
  [5422] = {.lex_state = 321, .external_lex_state = 105},
  [5423] = {.lex_state = 289, .external_lex_state = 103},
  [5424] = {.lex_state = 312, .external_lex_state = 102},
  [5425] = {.lex_state = 313, .external_lex_state = 108},
  [5426] = {.lex_state = 312, .external_lex_state = 102},
  [5427] = {.lex_state = 321, .external_lex_state = 106},
  [5428] = {.lex_state = 321, .external_lex_state = 105},
  [5429] = {.lex_state = 66, .external_lex_state = 104},
  [5430] = {.lex_state = 324, .external_lex_state = 106},
  [5431] = {.lex_state = 321, .external_lex_state = 105},
  [5432] = {.lex_state = 324, .external_lex_state = 106},
  [5433] = {.lex_state = 324, .external_lex_state = 106},
  [5434] = {.lex_state = 324, .external_lex_state = 106},
  [5435] = {.lex_state = 321, .external_lex_state = 105},
  [5436] = {.lex_state = 312, .external_lex_state = 103},
  [5437] = {.lex_state = 325, .external_lex_state = 103},
  [5438] = {.lex_state = 324, .external_lex_state = 106},
  [5439] = {.lex_state = 324, .external_lex_state = 106},
  [5440] = {.lex_state = 324, .external_lex_state = 106},
  [5441] = {.lex_state = 321, .external_lex_state = 101},
  [5442] = {.lex_state = 321, .external_lex_state = 105},
  [5443] = {.lex_state = 321, .external_lex_state = 105},
  [5444] = {.lex_state = 324, .external_lex_state = 106},
  [5445] = {.lex_state = 324, .external_lex_state = 106},
  [5446] = {.lex_state = 321, .external_lex_state = 105},
  [5447] = {.lex_state = 312, .external_lex_state = 102},
  [5448] = {.lex_state = 321, .external_lex_state = 101},
  [5449] = {.lex_state = 324, .external_lex_state = 106},
  [5450] = {.lex_state = 321, .external_lex_state = 101},
  [5451] = {.lex_state = 321, .external_lex_state = 101},
  [5452] = {.lex_state = 324, .external_lex_state = 106},
  [5453] = {.lex_state = 324, .external_lex_state = 106},
  [5454] = {.lex_state = 321, .external_lex_state = 101},
  [5455] = {.lex_state = 321, .external_lex_state = 101},
  [5456] = {.lex_state = 198, .external_lex_state = 73},
  [5457] = {.lex_state = 324, .external_lex_state = 106},
  [5458] = {.lex_state = 66, .external_lex_state = 104},
  [5459] = {.lex_state = 321, .external_lex_state = 101},
  [5460] = {.lex_state = 66, .external_lex_state = 104},
  [5461] = {.lex_state = 321, .external_lex_state = 101},
  [5462] = {.lex_state = 325, .external_lex_state = 103},
  [5463] = {.lex_state = 321, .external_lex_state = 101},
  [5464] = {.lex_state = 321, .external_lex_state = 101},
  [5465] = {.lex_state = 312, .external_lex_state = 103},
  [5466] = {.lex_state = 325, .external_lex_state = 103},
  [5467] = {.lex_state = 324, .external_lex_state = 106},
  [5468] = {.lex_state = 321, .external_lex_state = 101},
  [5469] = {.lex_state = 325, .external_lex_state = 103},
  [5470] = {.lex_state = 321, .external_lex_state = 101},
  [5471] = {.lex_state = 321, .external_lex_state = 106},
  [5472] = {.lex_state = 321, .external_lex_state = 105},
  [5473] = {.lex_state = 321, .external_lex_state = 101},
  [5474] = {.lex_state = 321, .external_lex_state = 106},
  [5475] = {.lex_state = 289, .external_lex_state = 103},
  [5476] = {.lex_state = 321, .external_lex_state = 106},
  [5477] = {.lex_state = 321, .external_lex_state = 101},
  [5478] = {.lex_state = 321, .external_lex_state = 101},
  [5479] = {.lex_state = 322, .external_lex_state = 81},
  [5480] = {.lex_state = 321, .external_lex_state = 101},
  [5481] = {.lex_state = 321, .external_lex_state = 105},
  [5482] = {.lex_state = 321, .external_lex_state = 101},
  [5483] = {.lex_state = 321, .external_lex_state = 101},
  [5484] = {.lex_state = 321, .external_lex_state = 105},
  [5485] = {.lex_state = 324, .external_lex_state = 106},
  [5486] = {.lex_state = 324, .external_lex_state = 106},
  [5487] = {.lex_state = 321, .external_lex_state = 101},
  [5488] = {.lex_state = 321, .external_lex_state = 105},
  [5489] = {.lex_state = 324, .external_lex_state = 106},
  [5490] = {.lex_state = 321, .external_lex_state = 105},
  [5491] = {.lex_state = 321, .external_lex_state = 105},
  [5492] = {.lex_state = 324, .external_lex_state = 106},
  [5493] = {.lex_state = 321, .external_lex_state = 105},
  [5494] = {.lex_state = 325, .external_lex_state = 103},
  [5495] = {.lex_state = 321, .external_lex_state = 105},
  [5496] = {.lex_state = 312, .external_lex_state = 102},
  [5497] = {.lex_state = 321, .external_lex_state = 105},
  [5498] = {.lex_state = 312, .external_lex_state = 102},
  [5499] = {.lex_state = 68, .external_lex_state = 94},
  [5500] = {.lex_state = 322, .external_lex_state = 103},
  [5501] = {.lex_state = 321, .external_lex_state = 106},
  [5502] = {.lex_state = 321, .external_lex_state = 106},
  [5503] = {.lex_state = 211, .external_lex_state = 109},
  [5504] = {.lex_state = 66, .external_lex_state = 110},
  [5505] = {.lex_state = 66, .external_lex_state = 110},
  [5506] = {.lex_state = 325, .external_lex_state = 103},
  [5507] = {.lex_state = 321, .external_lex_state = 106},
  [5508] = {.lex_state = 321, .external_lex_state = 106},
  [5509] = {.lex_state = 321, .external_lex_state = 106},
  [5510] = {.lex_state = 321, .external_lex_state = 106},
  [5511] = {.lex_state = 321, .external_lex_state = 106},
  [5512] = {.lex_state = 322, .external_lex_state = 81},
  [5513] = {.lex_state = 321, .external_lex_state = 106},
  [5514] = {.lex_state = 211, .external_lex_state = 109},
  [5515] = {.lex_state = 66, .external_lex_state = 110},
  [5516] = {.lex_state = 66, .external_lex_state = 110},
  [5517] = {.lex_state = 66, .external_lex_state = 110},
  [5518] = {.lex_state = 69, .external_lex_state = 111},
  [5519] = {.lex_state = 322, .external_lex_state = 103},
  [5520] = {.lex_state = 322, .external_lex_state = 103},
  [5521] = {.lex_state = 321, .external_lex_state = 106},
  [5522] = {.lex_state = 321, .external_lex_state = 106},
  [5523] = {.lex_state = 321, .external_lex_state = 106},
  [5524] = {.lex_state = 321, .external_lex_state = 106},
  [5525] = {.lex_state = 321, .external_lex_state = 106},
  [5526] = {.lex_state = 289, .external_lex_state = 102},
  [5527] = {.lex_state = 321, .external_lex_state = 106},
  [5528] = {.lex_state = 68, .external_lex_state = 94},
  [5529] = {.lex_state = 312, .external_lex_state = 102},
  [5530] = {.lex_state = 289, .external_lex_state = 102},
  [5531] = {.lex_state = 66, .external_lex_state = 110},
  [5532] = {.lex_state = 322, .external_lex_state = 103},
  [5533] = {.lex_state = 68, .external_lex_state = 94},
  [5534] = {.lex_state = 321, .external_lex_state = 106},
  [5535] = {.lex_state = 68, .external_lex_state = 94},
  [5536] = {.lex_state = 321, .external_lex_state = 106},
  [5537] = {.lex_state = 321, .external_lex_state = 106},
  [5538] = {.lex_state = 66, .external_lex_state = 110},
  [5539] = {.lex_state = 289, .external_lex_state = 102},
  [5540] = {.lex_state = 66, .external_lex_state = 110},
  [5541] = {.lex_state = 321, .external_lex_state = 106},
  [5542] = {.lex_state = 321, .external_lex_state = 106},
  [5543] = {.lex_state = 289, .external_lex_state = 102},
  [5544] = {.lex_state = 322, .external_lex_state = 103},
  [5545] = {.lex_state = 66, .external_lex_state = 110},
  [5546] = {.lex_state = 66, .external_lex_state = 110},
  [5547] = {.lex_state = 289, .external_lex_state = 103},
  [5548] = {.lex_state = 69, .external_lex_state = 111},
  [5549] = {.lex_state = 312, .external_lex_state = 102},
  [5550] = {.lex_state = 312, .external_lex_state = 103},
  [5551] = {.lex_state = 312, .external_lex_state = 103},
  [5552] = {.lex_state = 312, .external_lex_state = 81},
  [5553] = {.lex_state = 69, .external_lex_state = 111},
  [5554] = {.lex_state = 312, .external_lex_state = 103},
  [5555] = {.lex_state = 312, .external_lex_state = 102},
  [5556] = {.lex_state = 312, .external_lex_state = 102},
  [5557] = {.lex_state = 312, .external_lex_state = 103},
  [5558] = {.lex_state = 312, .external_lex_state = 103},
  [5559] = {.lex_state = 312, .external_lex_state = 103},
  [5560] = {.lex_state = 289, .external_lex_state = 103},
  [5561] = {.lex_state = 312, .external_lex_state = 103},
  [5562] = {.lex_state = 69, .external_lex_state = 111},
  [5563] = {.lex_state = 312, .external_lex_state = 102},
  [5564] = {.lex_state = 289, .external_lex_state = 103},
  [5565] = {.lex_state = 312, .external_lex_state = 102},
  [5566] = {.lex_state = 312, .external_lex_state = 81},
  [5567] = {.lex_state = 289, .external_lex_state = 103},
  [5568] = {.lex_state = 312, .external_lex_state = 102},
  [5569] = {.lex_state = 312, .external_lex_state = 103},
  [5570] = {.lex_state = 289, .external_lex_state = 103},
  [5571] = {.lex_state = 312, .external_lex_state = 102},
  [5572] = {.lex_state = 312, .external_lex_state = 102},
  [5573] = {.lex_state = 312, .external_lex_state = 102},
  [5574] = {.lex_state = 312, .external_lex_state = 102},
  [5575] = {.lex_state = 312, .external_lex_state = 102},
  [5576] = {.lex_state = 312, .external_lex_state = 102},
  [5577] = {.lex_state = 69, .external_lex_state = 111},
  [5578] = {.lex_state = 312, .external_lex_state = 102},
  [5579] = {.lex_state = 312, .external_lex_state = 102},
  [5580] = {.lex_state = 69, .external_lex_state = 111},
  [5581] = {.lex_state = 312, .external_lex_state = 102},
  [5582] = {.lex_state = 289, .external_lex_state = 103},
  [5583] = {.lex_state = 69, .external_lex_state = 111},
  [5584] = {.lex_state = 69, .external_lex_state = 111},
  [5585] = {.lex_state = 312, .external_lex_state = 102},
  [5586] = {.lex_state = 312, .external_lex_state = 102},
  [5587] = {.lex_state = 289, .external_lex_state = 103},
  [5588] = {.lex_state = 289, .external_lex_state = 103},
  [5589] = {.lex_state = 312, .external_lex_state = 102},
  [5590] = {.lex_state = 312, .external_lex_state = 102},
  [5591] = {.lex_state = 289, .external_lex_state = 103},
  [5592] = {.lex_state = 289, .external_lex_state = 103},
  [5593] = {.lex_state = 312, .external_lex_state = 103},
  [5594] = {.lex_state = 289, .external_lex_state = 103},
  [5595] = {.lex_state = 312, .external_lex_state = 102},
  [5596] = {.lex_state = 312, .external_lex_state = 102},
  [5597] = {.lex_state = 312, .external_lex_state = 102},
  [5598] = {.lex_state = 312, .external_lex_state = 102},
  [5599] = {.lex_state = 312, .external_lex_state = 102},
  [5600] = {.lex_state = 312, .external_lex_state = 102},
  [5601] = {.lex_state = 289, .external_lex_state = 103},
  [5602] = {.lex_state = 312, .external_lex_state = 103},
  [5603] = {.lex_state = 289, .external_lex_state = 103},
  [5604] = {.lex_state = 289, .external_lex_state = 103},
  [5605] = {.lex_state = 69, .external_lex_state = 111},
  [5606] = {.lex_state = 69, .external_lex_state = 111},
  [5607] = {.lex_state = 322, .external_lex_state = 103},
  [5608] = {.lex_state = 69, .external_lex_state = 111},
  [5609] = {.lex_state = 66, .external_lex_state = 110},
  [5610] = {.lex_state = 312, .external_lex_state = 81},
  [5611] = {.lex_state = 312, .external_lex_state = 103},
  [5612] = {.lex_state = 289, .external_lex_state = 103},
  [5613] = {.lex_state = 289, .external_lex_state = 103},
  [5614] = {.lex_state = 312, .external_lex_state = 103},
  [5615] = {.lex_state = 289, .external_lex_state = 103},
  [5616] = {.lex_state = 69, .external_lex_state = 111},
  [5617] = {.lex_state = 312, .external_lex_state = 103},
  [5618] = {.lex_state = 69, .external_lex_state = 111},
  [5619] = {.lex_state = 312, .external_lex_state = 103},
  [5620] = {.lex_state = 312, .external_lex_state = 103},
  [5621] = {.lex_state = 312, .external_lex_state = 103},
  [5622] = {.lex_state = 69, .external_lex_state = 111},
  [5623] = {.lex_state = 312, .external_lex_state = 103},
  [5624] = {.lex_state = 312, .external_lex_state = 103},
  [5625] = {.lex_state = 69, .external_lex_state = 111},
  [5626] = {.lex_state = 312, .external_lex_state = 103},
  [5627] = {.lex_state = 312, .external_lex_state = 103},
  [5628] = {.lex_state = 69, .external_lex_state = 111},
  [5629] = {.lex_state = 312, .external_lex_state = 103},
  [5630] = {.lex_state = 312, .external_lex_state = 103},
  [5631] = {.lex_state = 312, .external_lex_state = 103},
  [5632] = {.lex_state = 69, .external_lex_state = 110},
  [5633] = {.lex_state = 312, .external_lex_state = 103},
  [5634] = {.lex_state = 312, .external_lex_state = 103},
  [5635] = {.lex_state = 312, .external_lex_state = 103},
  [5636] = {.lex_state = 312, .external_lex_state = 103},
  [5637] = {.lex_state = 312, .external_lex_state = 103},
  [5638] = {.lex_state = 312, .external_lex_state = 103},
  [5639] = {.lex_state = 312, .external_lex_state = 103},
  [5640] = {.lex_state = 312, .external_lex_state = 103},
  [5641] = {.lex_state = 312, .external_lex_state = 103},
  [5642] = {.lex_state = 312, .external_lex_state = 103},
  [5643] = {.lex_state = 69, .external_lex_state = 110},
  [5644] = {.lex_state = 312, .external_lex_state = 103},
  [5645] = {.lex_state = 312, .external_lex_state = 103},
  [5646] = {.lex_state = 312, .external_lex_state = 103},
  [5647] = {.lex_state = 312, .external_lex_state = 103},
  [5648] = {.lex_state = 312, .external_lex_state = 103},
  [5649] = {.lex_state = 312, .external_lex_state = 103},
  [5650] = {.lex_state = 312, .external_lex_state = 103},
  [5651] = {.lex_state = 69, .external_lex_state = 110},
  [5652] = {.lex_state = 312, .external_lex_state = 103},
  [5653] = {.lex_state = 312, .external_lex_state = 103},
  [5654] = {.lex_state = 69, .external_lex_state = 111},
  [5655] = {.lex_state = 69, .external_lex_state = 111},
  [5656] = {.lex_state = 312, .external_lex_state = 103},
  [5657] = {.lex_state = 69, .external_lex_state = 110},
  [5658] = {.lex_state = 312, .external_lex_state = 103},
  [5659] = {.lex_state = 312, .external_lex_state = 103},
  [5660] = {.lex_state = 69, .external_lex_state = 111},
  [5661] = {.lex_state = 312, .external_lex_state = 103},
  [5662] = {.lex_state = 69, .external_lex_state = 111},
  [5663] = {.lex_state = 69, .external_lex_state = 111},
  [5664] = {.lex_state = 69, .external_lex_state = 111},
  [5665] = {.lex_state = 312, .external_lex_state = 103},
  [5666] = {.lex_state = 69, .external_lex_state = 111},
  [5667] = {.lex_state = 312, .external_lex_state = 103},
  [5668] = {.lex_state = 69, .external_lex_state = 111},
  [5669] = {.lex_state = 69, .external_lex_state = 111},
  [5670] = {.lex_state = 69, .external_lex_state = 111},
  [5671] = {.lex_state = 312, .external_lex_state = 103},
  [5672] = {.lex_state = 312, .external_lex_state = 103},
  [5673] = {.lex_state = 312, .external_lex_state = 103},
  [5674] = {.lex_state = 69, .external_lex_state = 110},
  [5675] = {.lex_state = 312, .external_lex_state = 103},
  [5676] = {.lex_state = 312, .external_lex_state = 103},
  [5677] = {.lex_state = 312, .external_lex_state = 103},
  [5678] = {.lex_state = 312, .external_lex_state = 103},
  [5679] = {.lex_state = 312, .external_lex_state = 103},
  [5680] = {.lex_state = 69, .external_lex_state = 111},
  [5681] = {.lex_state = 312, .external_lex_state = 103},
  [5682] = {.lex_state = 312, .external_lex_state = 103},
  [5683] = {.lex_state = 69, .external_lex_state = 111},
  [5684] = {.lex_state = 312, .external_lex_state = 103},
  [5685] = {.lex_state = 312, .external_lex_state = 103},
  [5686] = {.lex_state = 312, .external_lex_state = 103},
  [5687] = {.lex_state = 312, .external_lex_state = 103},
  [5688] = {.lex_state = 69, .external_lex_state = 111},
  [5689] = {.lex_state = 312, .external_lex_state = 103},
  [5690] = {.lex_state = 312, .external_lex_state = 103},
  [5691] = {.lex_state = 312, .external_lex_state = 103},
  [5692] = {.lex_state = 312, .external_lex_state = 103},
  [5693] = {.lex_state = 69, .external_lex_state = 111},
  [5694] = {.lex_state = 312, .external_lex_state = 103},
  [5695] = {.lex_state = 312, .external_lex_state = 103},
  [5696] = {.lex_state = 69, .external_lex_state = 111},
  [5697] = {.lex_state = 293, .external_lex_state = 112},
  [5698] = {.lex_state = 286, .external_lex_state = 112},
  [5699] = {.lex_state = 286, .external_lex_state = 112},
  [5700] = {.lex_state = 293, .external_lex_state = 112},
  [5701] = {.lex_state = 69, .external_lex_state = 110},
  [5702] = {.lex_state = 318},
  [5703] = {.lex_state = 318},
  [5704] = {.lex_state = 318},
  [5705] = {.lex_state = 318},
  [5706] = {.lex_state = 318},
  [5707] = {.lex_state = 318},
  [5708] = {.lex_state = 318},
  [5709] = {.lex_state = 318},
  [5710] = {.lex_state = 318},
  [5711] = {.lex_state = 318},
  [5712] = {.lex_state = 318},
  [5713] = {.lex_state = 318},
  [5714] = {.lex_state = 318},
  [5715] = {.lex_state = 318},
  [5716] = {.lex_state = 318},
  [5717] = {.lex_state = 318},
  [5718] = {.lex_state = 318},
  [5719] = {.lex_state = 318},
  [5720] = {.lex_state = 318},
  [5721] = {.lex_state = 318},
  [5722] = {.lex_state = 318},
  [5723] = {.lex_state = 318},
  [5724] = {.lex_state = 318},
  [5725] = {.lex_state = 318},
  [5726] = {.lex_state = 318},
  [5727] = {.lex_state = 318},
  [5728] = {.lex_state = 318},
  [5729] = {.lex_state = 318},
  [5730] = {.lex_state = 318},
  [5731] = {.lex_state = 318},
  [5732] = {.lex_state = 318},
  [5733] = {.lex_state = 318},
  [5734] = {.lex_state = 318},
  [5735] = {.lex_state = 318},
  [5736] = {.lex_state = 318},
  [5737] = {.lex_state = 318},
  [5738] = {.lex_state = 66, .external_lex_state = 110},
  [5739] = {.lex_state = 318},
  [5740] = {.lex_state = 318},
  [5741] = {.lex_state = 318},
  [5742] = {.lex_state = 318},
  [5743] = {.lex_state = 318},
  [5744] = {.lex_state = 66, .external_lex_state = 110},
  [5745] = {.lex_state = 318},
  [5746] = {.lex_state = 318},
  [5747] = {.lex_state = 66, .external_lex_state = 110},
  [5748] = {.lex_state = 318},
  [5749] = {.lex_state = 318},
  [5750] = {.lex_state = 318},
  [5751] = {.lex_state = 318},
  [5752] = {.lex_state = 318},
  [5753] = {.lex_state = 318},
  [5754] = {.lex_state = 318},
  [5755] = {.lex_state = 318},
  [5756] = {.lex_state = 318},
  [5757] = {.lex_state = 318},
  [5758] = {.lex_state = 318},
  [5759] = {.lex_state = 318},
  [5760] = {.lex_state = 318},
  [5761] = {.lex_state = 318},
  [5762] = {.lex_state = 318},
  [5763] = {.lex_state = 318},
  [5764] = {.lex_state = 318},
  [5765] = {.lex_state = 318},
  [5766] = {.lex_state = 318},
  [5767] = {.lex_state = 318},
  [5768] = {.lex_state = 318},
  [5769] = {.lex_state = 318},
  [5770] = {.lex_state = 318},
  [5771] = {.lex_state = 318},
  [5772] = {.lex_state = 318},
  [5773] = {.lex_state = 318},
  [5774] = {.lex_state = 318},
  [5775] = {.lex_state = 318},
  [5776] = {.lex_state = 318},
  [5777] = {.lex_state = 318},
  [5778] = {.lex_state = 318},
  [5779] = {.lex_state = 318},
  [5780] = {.lex_state = 318},
  [5781] = {.lex_state = 318},
  [5782] = {.lex_state = 318},
  [5783] = {.lex_state = 318},
  [5784] = {.lex_state = 318},
  [5785] = {.lex_state = 318},
  [5786] = {.lex_state = 318},
  [5787] = {.lex_state = 318},
  [5788] = {.lex_state = 318},
  [5789] = {.lex_state = 66, .external_lex_state = 110},
  [5790] = {.lex_state = 318},
  [5791] = {.lex_state = 318},
  [5792] = {.lex_state = 318},
  [5793] = {.lex_state = 318},
  [5794] = {.lex_state = 318},
  [5795] = {.lex_state = 318},
  [5796] = {.lex_state = 318},
  [5797] = {.lex_state = 318},
  [5798] = {.lex_state = 318},
  [5799] = {.lex_state = 318},
  [5800] = {.lex_state = 318},
  [5801] = {.lex_state = 318},
  [5802] = {.lex_state = 318},
  [5803] = {.lex_state = 318},
  [5804] = {.lex_state = 318},
  [5805] = {.lex_state = 318},
  [5806] = {.lex_state = 318},
  [5807] = {.lex_state = 318},
  [5808] = {.lex_state = 318},
  [5809] = {.lex_state = 318},
  [5810] = {.lex_state = 318},
  [5811] = {.lex_state = 318},
  [5812] = {.lex_state = 318},
  [5813] = {.lex_state = 318},
  [5814] = {.lex_state = 318},
  [5815] = {.lex_state = 318},
  [5816] = {.lex_state = 318},
  [5817] = {.lex_state = 318},
  [5818] = {.lex_state = 318},
  [5819] = {.lex_state = 318},
  [5820] = {.lex_state = 318},
  [5821] = {.lex_state = 318},
  [5822] = {.lex_state = 318},
  [5823] = {.lex_state = 318},
  [5824] = {.lex_state = 318},
  [5825] = {.lex_state = 318},
  [5826] = {.lex_state = 318},
  [5827] = {.lex_state = 318},
  [5828] = {.lex_state = 318},
  [5829] = {.lex_state = 318},
  [5830] = {.lex_state = 318},
  [5831] = {.lex_state = 318},
  [5832] = {.lex_state = 318},
  [5833] = {.lex_state = 318},
  [5834] = {.lex_state = 318},
  [5835] = {.lex_state = 318},
  [5836] = {.lex_state = 318},
  [5837] = {.lex_state = 318},
  [5838] = {.lex_state = 318},
  [5839] = {.lex_state = 318},
  [5840] = {.lex_state = 318},
  [5841] = {.lex_state = 318},
  [5842] = {.lex_state = 318},
  [5843] = {.lex_state = 318},
  [5844] = {.lex_state = 318},
  [5845] = {.lex_state = 318},
  [5846] = {.lex_state = 318},
  [5847] = {.lex_state = 318},
  [5848] = {.lex_state = 318},
  [5849] = {.lex_state = 318},
  [5850] = {.lex_state = 318},
  [5851] = {.lex_state = 318},
  [5852] = {.lex_state = 318},
  [5853] = {.lex_state = 318},
  [5854] = {.lex_state = 318},
  [5855] = {.lex_state = 318},
  [5856] = {.lex_state = 318},
  [5857] = {.lex_state = 318},
  [5858] = {.lex_state = 318},
  [5859] = {.lex_state = 318},
  [5860] = {.lex_state = 318},
  [5861] = {.lex_state = 318},
  [5862] = {.lex_state = 318},
  [5863] = {.lex_state = 318},
  [5864] = {.lex_state = 318},
  [5865] = {.lex_state = 288, .external_lex_state = 112},
  [5866] = {.lex_state = 287, .external_lex_state = 112},
  [5867] = {.lex_state = 287, .external_lex_state = 112},
  [5868] = {.lex_state = 287, .external_lex_state = 112},
  [5869] = {.lex_state = 287, .external_lex_state = 112},
  [5870] = {.lex_state = 288, .external_lex_state = 112},
  [5871] = {.lex_state = 287, .external_lex_state = 112},
  [5872] = {.lex_state = 287, .external_lex_state = 112},
  [5873] = {.lex_state = 287, .external_lex_state = 112},
  [5874] = {.lex_state = 287, .external_lex_state = 112},
  [5875] = {.lex_state = 288, .external_lex_state = 112},
  [5876] = {.lex_state = 288, .external_lex_state = 112},
  [5877] = {.lex_state = 287, .external_lex_state = 112},
  [5878] = {.lex_state = 287, .external_lex_state = 112},
  [5879] = {.lex_state = 288, .external_lex_state = 112},
  [5880] = {.lex_state = 287, .external_lex_state = 112},
  [5881] = {.lex_state = 287, .external_lex_state = 112},
  [5882] = {.lex_state = 287, .external_lex_state = 112},
  [5883] = {.lex_state = 288, .external_lex_state = 112},
  [5884] = {.lex_state = 288, .external_lex_state = 112},
  [5885] = {.lex_state = 287, .external_lex_state = 112},
  [5886] = {.lex_state = 287, .external_lex_state = 112},
  [5887] = {.lex_state = 287, .external_lex_state = 112},
  [5888] = {.lex_state = 287, .external_lex_state = 112},
  [5889] = {.lex_state = 287, .external_lex_state = 112},
  [5890] = {.lex_state = 287, .external_lex_state = 112},
  [5891] = {.lex_state = 287, .external_lex_state = 112},
  [5892] = {.lex_state = 288, .external_lex_state = 112},
  [5893] = {.lex_state = 287, .external_lex_state = 112},
  [5894] = {.lex_state = 288, .external_lex_state = 112},
  [5895] = {.lex_state = 288, .external_lex_state = 112},
  [5896] = {.lex_state = 287, .external_lex_state = 112},
  [5897] = {.lex_state = 287, .external_lex_state = 112},
  [5898] = {.lex_state = 287, .external_lex_state = 112},
  [5899] = {.lex_state = 287, .external_lex_state = 112},
  [5900] = {.lex_state = 288, .external_lex_state = 112},
  [5901] = {.lex_state = 287, .external_lex_state = 112},
  [5902] = {.lex_state = 287, .external_lex_state = 112},
  [5903] = {.lex_state = 288, .external_lex_state = 112},
  [5904] = {.lex_state = 287, .external_lex_state = 112},
  [5905] = {.lex_state = 287, .external_lex_state = 112},
  [5906] = {.lex_state = 287, .external_lex_state = 112},
  [5907] = {.lex_state = 288, .external_lex_state = 112},
  [5908] = {.lex_state = 288, .external_lex_state = 112},
  [5909] = {.lex_state = 287, .external_lex_state = 112},
  [5910] = {.lex_state = 287, .external_lex_state = 112},
  [5911] = {.lex_state = 287, .external_lex_state = 112},
  [5912] = {.lex_state = 287, .external_lex_state = 112},
  [5913] = {.lex_state = 287, .external_lex_state = 112},
  [5914] = {.lex_state = 287, .external_lex_state = 112},
  [5915] = {.lex_state = 288, .external_lex_state = 112},
  [5916] = {.lex_state = 287, .external_lex_state = 112},
  [5917] = {.lex_state = 287, .external_lex_state = 112},
  [5918] = {.lex_state = 287, .external_lex_state = 112},
  [5919] = {.lex_state = 288, .external_lex_state = 112},
  [5920] = {.lex_state = 288, .external_lex_state = 112},
  [5921] = {.lex_state = 287, .external_lex_state = 112},
  [5922] = {.lex_state = 287, .external_lex_state = 112},
  [5923] = {.lex_state = 288, .external_lex_state = 112},
  [5924] = {.lex_state = 288, .external_lex_state = 112},
  [5925] = {.lex_state = 288, .external_lex_state = 112},
  [5926] = {.lex_state = 288, .external_lex_state = 112},
  [5927] = {.lex_state = 287, .external_lex_state = 112},
  [5928] = {.lex_state = 287, .external_lex_state = 112},
  [5929] = {.lex_state = 287, .external_lex_state = 112},
  [5930] = {.lex_state = 288, .external_lex_state = 112},
  [5931] = {.lex_state = 287, .external_lex_state = 112},
  [5932] = {.lex_state = 287, .external_lex_state = 112},
  [5933] = {.lex_state = 287, .external_lex_state = 112},
  [5934] = {.lex_state = 287, .external_lex_state = 112},
  [5935] = {.lex_state = 288, .external_lex_state = 112},
  [5936] = {.lex_state = 288, .external_lex_state = 112},
  [5937] = {.lex_state = 287, .external_lex_state = 112},
  [5938] = {.lex_state = 287, .external_lex_state = 112},
  [5939] = {.lex_state = 287, .external_lex_state = 112},
  [5940] = {.lex_state = 288, .external_lex_state = 112},
  [5941] = {.lex_state = 287, .external_lex_state = 112},
  [5942] = {.lex_state = 287, .external_lex_state = 112},
  [5943] = {.lex_state = 287, .external_lex_state = 112},
  [5944] = {.lex_state = 287, .external_lex_state = 112},
  [5945] = {.lex_state = 287, .external_lex_state = 112},
  [5946] = {.lex_state = 287, .external_lex_state = 112},
  [5947] = {.lex_state = 287, .external_lex_state = 112},
  [5948] = {.lex_state = 287, .external_lex_state = 112},
  [5949] = {.lex_state = 288, .external_lex_state = 112},
  [5950] = {.lex_state = 288, .external_lex_state = 112},
  [5951] = {.lex_state = 287, .external_lex_state = 112},
  [5952] = {.lex_state = 288, .external_lex_state = 112},
  [5953] = {.lex_state = 287, .external_lex_state = 112},
  [5954] = {.lex_state = 287, .external_lex_state = 112},
  [5955] = {.lex_state = 287, .external_lex_state = 112},
  [5956] = {.lex_state = 288, .external_lex_state = 112},
  [5957] = {.lex_state = 287, .external_lex_state = 112},
  [5958] = {.lex_state = 288, .external_lex_state = 112},
  [5959] = {.lex_state = 287, .external_lex_state = 112},
  [5960] = {.lex_state = 287, .external_lex_state = 112},
  [5961] = {.lex_state = 287, .external_lex_state = 112},
  [5962] = {.lex_state = 287, .external_lex_state = 112},
  [5963] = {.lex_state = 287, .external_lex_state = 112},
  [5964] = {.lex_state = 287, .external_lex_state = 112},
  [5965] = {.lex_state = 287, .external_lex_state = 112},
  [5966] = {.lex_state = 287, .external_lex_state = 112},
  [5967] = {.lex_state = 287, .external_lex_state = 112},
  [5968] = {.lex_state = 288, .external_lex_state = 112},
  [5969] = {.lex_state = 288, .external_lex_state = 112},
  [5970] = {.lex_state = 287, .external_lex_state = 112},
  [5971] = {.lex_state = 288, .external_lex_state = 112},
  [5972] = {.lex_state = 287, .external_lex_state = 112},
  [5973] = {.lex_state = 287, .external_lex_state = 112},
  [5974] = {.lex_state = 288, .external_lex_state = 112},
  [5975] = {.lex_state = 287, .external_lex_state = 112},
  [5976] = {.lex_state = 288, .external_lex_state = 112},
  [5977] = {.lex_state = 287, .external_lex_state = 112},
  [5978] = {.lex_state = 287, .external_lex_state = 112},
  [5979] = {.lex_state = 288, .external_lex_state = 112},
  [5980] = {.lex_state = 287, .external_lex_state = 112},
  [5981] = {.lex_state = 288, .external_lex_state = 112},
  [5982] = {.lex_state = 287, .external_lex_state = 112},
  [5983] = {.lex_state = 287, .external_lex_state = 112},
  [5984] = {.lex_state = 287, .external_lex_state = 112},
  [5985] = {.lex_state = 287, .external_lex_state = 112},
  [5986] = {.lex_state = 287, .external_lex_state = 112},
  [5987] = {.lex_state = 287, .external_lex_state = 112},
  [5988] = {.lex_state = 287, .external_lex_state = 112},
  [5989] = {.lex_state = 287, .external_lex_state = 112},
  [5990] = {.lex_state = 287, .external_lex_state = 112},
  [5991] = {.lex_state = 287, .external_lex_state = 112},
  [5992] = {.lex_state = 288, .external_lex_state = 112},
  [5993] = {.lex_state = 287, .external_lex_state = 112},
  [5994] = {.lex_state = 288, .external_lex_state = 112},
  [5995] = {.lex_state = 287, .external_lex_state = 112},
  [5996] = {.lex_state = 287, .external_lex_state = 112},
  [5997] = {.lex_state = 288, .external_lex_state = 112},
  [5998] = {.lex_state = 287, .external_lex_state = 112},
  [5999] = {.lex_state = 287, .external_lex_state = 112},
  [6000] = {.lex_state = 287, .external_lex_state = 112},
  [6001] = {.lex_state = 287, .external_lex_state = 112},
  [6002] = {.lex_state = 287, .external_lex_state = 112},
  [6003] = {.lex_state = 287, .external_lex_state = 112},
  [6004] = {.lex_state = 287, .external_lex_state = 112},
  [6005] = {.lex_state = 288, .external_lex_state = 112},
  [6006] = {.lex_state = 287, .external_lex_state = 112},
  [6007] = {.lex_state = 288, .external_lex_state = 112},
  [6008] = {.lex_state = 287, .external_lex_state = 112},
  [6009] = {.lex_state = 287, .external_lex_state = 112},
  [6010] = {.lex_state = 287, .external_lex_state = 112},
  [6011] = {.lex_state = 287, .external_lex_state = 112},
  [6012] = {.lex_state = 288, .external_lex_state = 112},
  [6013] = {.lex_state = 287, .external_lex_state = 112},
  [6014] = {.lex_state = 288, .external_lex_state = 112},
  [6015] = {.lex_state = 287, .external_lex_state = 112},
  [6016] = {.lex_state = 287, .external_lex_state = 112},
  [6017] = {.lex_state = 287, .external_lex_state = 112},
  [6018] = {.lex_state = 287, .external_lex_state = 112},
  [6019] = {.lex_state = 287, .external_lex_state = 112},
  [6020] = {.lex_state = 287, .external_lex_state = 112},
  [6021] = {.lex_state = 287, .external_lex_state = 112},
  [6022] = {.lex_state = 287, .external_lex_state = 112},
  [6023] = {.lex_state = 287, .external_lex_state = 112},
  [6024] = {.lex_state = 288, .external_lex_state = 112},
  [6025] = {.lex_state = 287, .external_lex_state = 112},
  [6026] = {.lex_state = 287, .external_lex_state = 112},
  [6027] = {.lex_state = 287, .external_lex_state = 112},
  [6028] = {.lex_state = 288, .external_lex_state = 112},
  [6029] = {.lex_state = 287, .external_lex_state = 112},
  [6030] = {.lex_state = 288, .external_lex_state = 112},
  [6031] = {.lex_state = 287, .external_lex_state = 112},
  [6032] = {.lex_state = 287, .external_lex_state = 112},
  [6033] = {.lex_state = 287, .external_lex_state = 112},
  [6034] = {.lex_state = 288, .external_lex_state = 112},
  [6035] = {.lex_state = 288, .external_lex_state = 112},
  [6036] = {.lex_state = 287, .external_lex_state = 112},
  [6037] = {.lex_state = 287, .external_lex_state = 112},
  [6038] = {.lex_state = 288, .external_lex_state = 112},
  [6039] = {.lex_state = 288, .external_lex_state = 112},
  [6040] = {.lex_state = 287, .external_lex_state = 112},
  [6041] = {.lex_state = 287, .external_lex_state = 112},
  [6042] = {.lex_state = 287, .external_lex_state = 112},
  [6043] = {.lex_state = 287, .external_lex_state = 112},
  [6044] = {.lex_state = 288, .external_lex_state = 112},
  [6045] = {.lex_state = 287, .external_lex_state = 112},
  [6046] = {.lex_state = 287, .external_lex_state = 112},
  [6047] = {.lex_state = 288, .external_lex_state = 112},
  [6048] = {.lex_state = 288, .external_lex_state = 112},
  [6049] = {.lex_state = 287, .external_lex_state = 112},
  [6050] = {.lex_state = 287, .external_lex_state = 112},
  [6051] = {.lex_state = 288, .external_lex_state = 112},
  [6052] = {.lex_state = 287, .external_lex_state = 112},
  [6053] = {.lex_state = 288, .external_lex_state = 112},
  [6054] = {.lex_state = 287, .external_lex_state = 112},
  [6055] = {.lex_state = 287, .external_lex_state = 112},
  [6056] = {.lex_state = 287, .external_lex_state = 112},
  [6057] = {.lex_state = 287, .external_lex_state = 112},
  [6058] = {.lex_state = 287, .external_lex_state = 112},
  [6059] = {.lex_state = 287, .external_lex_state = 112},
  [6060] = {.lex_state = 288, .external_lex_state = 112},
  [6061] = {.lex_state = 288, .external_lex_state = 112},
  [6062] = {.lex_state = 287, .external_lex_state = 112},
  [6063] = {.lex_state = 287, .external_lex_state = 112},
  [6064] = {.lex_state = 287, .external_lex_state = 112},
  [6065] = {.lex_state = 287, .external_lex_state = 112},
  [6066] = {.lex_state = 288, .external_lex_state = 112},
  [6067] = {.lex_state = 287, .external_lex_state = 112},
  [6068] = {.lex_state = 287, .external_lex_state = 112},
  [6069] = {.lex_state = 287, .external_lex_state = 112},
  [6070] = {.lex_state = 288, .external_lex_state = 112},
  [6071] = {.lex_state = 288, .external_lex_state = 112},
  [6072] = {.lex_state = 287, .external_lex_state = 112},
  [6073] = {.lex_state = 287, .external_lex_state = 112},
  [6074] = {.lex_state = 287, .external_lex_state = 112},
  [6075] = {.lex_state = 287, .external_lex_state = 112},
  [6076] = {.lex_state = 287, .external_lex_state = 112},
  [6077] = {.lex_state = 287, .external_lex_state = 112},
  [6078] = {.lex_state = 288, .external_lex_state = 112},
  [6079] = {.lex_state = 287, .external_lex_state = 112},
  [6080] = {.lex_state = 287, .external_lex_state = 112},
  [6081] = {.lex_state = 287, .external_lex_state = 112},
  [6082] = {.lex_state = 288, .external_lex_state = 112},
  [6083] = {.lex_state = 287, .external_lex_state = 112},
  [6084] = {.lex_state = 288, .external_lex_state = 112},
  [6085] = {.lex_state = 288, .external_lex_state = 112},
  [6086] = {.lex_state = 287, .external_lex_state = 112},
  [6087] = {.lex_state = 287, .external_lex_state = 112},
  [6088] = {.lex_state = 289, .external_lex_state = 113},
  [6089] = {.lex_state = 289, .external_lex_state = 113},
  [6090] = {.lex_state = 289, .external_lex_state = 113},
  [6091] = {.lex_state = 289, .external_lex_state = 113},
  [6092] = {.lex_state = 289, .external_lex_state = 113},
  [6093] = {.lex_state = 289, .external_lex_state = 113},
  [6094] = {.lex_state = 289, .external_lex_state = 113},
  [6095] = {.lex_state = 66, .external_lex_state = 104},
  [6096] = {.lex_state = 66, .external_lex_state = 104},
  [6097] = {.lex_state = 289, .external_lex_state = 113},
  [6098] = {.lex_state = 290, .external_lex_state = 112},
  [6099] = {.lex_state = 291, .external_lex_state = 112},
  [6100] = {.lex_state = 312, .external_lex_state = 113},
  [6101] = {.lex_state = 291, .external_lex_state = 112},
  [6102] = {.lex_state = 291, .external_lex_state = 112},
  [6103] = {.lex_state = 291, .external_lex_state = 112},
  [6104] = {.lex_state = 312, .external_lex_state = 113},
  [6105] = {.lex_state = 312, .external_lex_state = 113},
  [6106] = {.lex_state = 312, .external_lex_state = 113},
  [6107] = {.lex_state = 312, .external_lex_state = 113},
  [6108] = {.lex_state = 291, .external_lex_state = 112},
  [6109] = {.lex_state = 312, .external_lex_state = 113},
  [6110] = {.lex_state = 291, .external_lex_state = 112},
  [6111] = {.lex_state = 291, .external_lex_state = 112},
  [6112] = {.lex_state = 291, .external_lex_state = 112},
  [6113] = {.lex_state = 312, .external_lex_state = 113},
  [6114] = {.lex_state = 291, .external_lex_state = 112},
  [6115] = {.lex_state = 312, .external_lex_state = 113},
  [6116] = {.lex_state = 312, .external_lex_state = 113},
  [6117] = {.lex_state = 323},
  [6118] = {.lex_state = 323},
  [6119] = {.lex_state = 323},
  [6120] = {.lex_state = 323},
  [6121] = {.lex_state = 313, .external_lex_state = 114},
  [6122] = {.lex_state = 318, .external_lex_state = 115},
  [6123] = {.lex_state = 318, .external_lex_state = 115},
  [6124] = {.lex_state = 318, .external_lex_state = 115},
  [6125] = {.lex_state = 318, .external_lex_state = 115},
  [6126] = {.lex_state = 323},
  [6127] = {.lex_state = 318, .external_lex_state = 115},
  [6128] = {.lex_state = 271, .external_lex_state = 116},
  [6129] = {.lex_state = 323},
  [6130] = {.lex_state = 323},
  [6131] = {.lex_state = 271, .external_lex_state = 116},
  [6132] = {.lex_state = 323},
  [6133] = {.lex_state = 323},
  [6134] = {.lex_state = 323},
  [6135] = {.lex_state = 323},
  [6136] = {.lex_state = 323},
  [6137] = {.lex_state = 271, .external_lex_state = 116},
  [6138] = {.lex_state = 318, .external_lex_state = 115},
  [6139] = {.lex_state = 323},
  [6140] = {.lex_state = 318, .external_lex_state = 115},
  [6141] = {.lex_state = 318, .external_lex_state = 115},
  [6142] = {.lex_state = 323},
  [6143] = {.lex_state = 318, .external_lex_state = 115},
  [6144] = {.lex_state = 323},
  [6145] = {.lex_state = 323},
  [6146] = {.lex_state = 323},
  [6147] = {.lex_state = 318, .external_lex_state = 115},
  [6148] = {.lex_state = 323},
  [6149] = {.lex_state = 323},
  [6150] = {.lex_state = 323},
  [6151] = {.lex_state = 323},
  [6152] = {.lex_state = 323},
  [6153] = {.lex_state = 323},
  [6154] = {.lex_state = 318, .external_lex_state = 115},
  [6155] = {.lex_state = 323},
  [6156] = {.lex_state = 323},
  [6157] = {.lex_state = 323},
  [6158] = {.lex_state = 323},
  [6159] = {.lex_state = 323, .external_lex_state = 99},
  [6160] = {.lex_state = 323, .external_lex_state = 99},
  [6161] = {.lex_state = 323, .external_lex_state = 99},
  [6162] = {.lex_state = 323, .external_lex_state = 99},
  [6163] = {.lex_state = 323, .external_lex_state = 99},
  [6164] = {.lex_state = 313},
  [6165] = {.lex_state = 323, .external_lex_state = 99},
  [6166] = {.lex_state = 323, .external_lex_state = 99},
  [6167] = {.lex_state = 323, .external_lex_state = 99},
  [6168] = {.lex_state = 323, .external_lex_state = 99},
  [6169] = {.lex_state = 323, .external_lex_state = 99},
  [6170] = {.lex_state = 323, .external_lex_state = 99},
  [6171] = {.lex_state = 323, .external_lex_state = 99},
  [6172] = {.lex_state = 323, .external_lex_state = 99},
  [6173] = {.lex_state = 323, .external_lex_state = 99},
  [6174] = {.lex_state = 323, .external_lex_state = 99},
  [6175] = {.lex_state = 323, .external_lex_state = 99},
  [6176] = {.lex_state = 324},
  [6177] = {.lex_state = 323, .external_lex_state = 99},
  [6178] = {.lex_state = 323, .external_lex_state = 99},
  [6179] = {.lex_state = 318},
  [6180] = {.lex_state = 323, .external_lex_state = 99},
  [6181] = {.lex_state = 323, .external_lex_state = 99},
  [6182] = {.lex_state = 323, .external_lex_state = 99},
  [6183] = {.lex_state = 318},
  [6184] = {.lex_state = 323, .external_lex_state = 99},
  [6185] = {.lex_state = 323, .external_lex_state = 99},
  [6186] = {.lex_state = 323, .external_lex_state = 99},
  [6187] = {.lex_state = 314, .external_lex_state = 117},
  [6188] = {.lex_state = 70, .external_lex_state = 118},
  [6189] = {.lex_state = 70, .external_lex_state = 118},
  [6190] = {.lex_state = 315, .external_lex_state = 119},
  [6191] = {.lex_state = 70, .external_lex_state = 118},
  [6192] = {.lex_state = 70, .external_lex_state = 118},
  [6193] = {.lex_state = 70, .external_lex_state = 118},
  [6194] = {.lex_state = 315, .external_lex_state = 119},
  [6195] = {.lex_state = 321},
  [6196] = {.lex_state = 321},
  [6197] = {.lex_state = 315, .external_lex_state = 119},
  [6198] = {.lex_state = 70, .external_lex_state = 118},
  [6199] = {.lex_state = 70, .external_lex_state = 118},
  [6200] = {.lex_state = 70, .external_lex_state = 118},
  [6201] = {.lex_state = 70, .external_lex_state = 118},
  [6202] = {.lex_state = 70, .external_lex_state = 118},
  [6203] = {.lex_state = 70, .external_lex_state = 118},
  [6204] = {.lex_state = 70, .external_lex_state = 118},
  [6205] = {.lex_state = 70, .external_lex_state = 118},
  [6206] = {.lex_state = 70, .external_lex_state = 118},
  [6207] = {.lex_state = 70, .external_lex_state = 118},
  [6208] = {.lex_state = 314, .external_lex_state = 114},
  [6209] = {.lex_state = 70, .external_lex_state = 118},
  [6210] = {.lex_state = 70, .external_lex_state = 118},
  [6211] = {.lex_state = 271, .external_lex_state = 116},
  [6212] = {.lex_state = 314, .external_lex_state = 114},
  [6213] = {.lex_state = 70, .external_lex_state = 120},
  [6214] = {.lex_state = 70, .external_lex_state = 118},
  [6215] = {.lex_state = 314, .external_lex_state = 114},
  [6216] = {.lex_state = 314, .external_lex_state = 114},
  [6217] = {.lex_state = 70, .external_lex_state = 118},
  [6218] = {.lex_state = 314, .external_lex_state = 114},
  [6219] = {.lex_state = 70, .external_lex_state = 120},
  [6220] = {.lex_state = 70, .external_lex_state = 120},
  [6221] = {.lex_state = 70, .external_lex_state = 118},
  [6222] = {.lex_state = 70, .external_lex_state = 118},
  [6223] = {.lex_state = 70, .external_lex_state = 120},
  [6224] = {.lex_state = 271, .external_lex_state = 116},
  [6225] = {.lex_state = 70, .external_lex_state = 118},
  [6226] = {.lex_state = 271, .external_lex_state = 116},
  [6227] = {.lex_state = 314, .external_lex_state = 114},
  [6228] = {.lex_state = 314, .external_lex_state = 114},
  [6229] = {.lex_state = 314, .external_lex_state = 114},
  [6230] = {.lex_state = 70, .external_lex_state = 118},
  [6231] = {.lex_state = 70, .external_lex_state = 118},
  [6232] = {.lex_state = 70, .external_lex_state = 118},
  [6233] = {.lex_state = 271, .external_lex_state = 116},
  [6234] = {.lex_state = 314, .external_lex_state = 114},
  [6235] = {.lex_state = 271, .external_lex_state = 116},
  [6236] = {.lex_state = 70, .external_lex_state = 118},
  [6237] = {.lex_state = 271, .external_lex_state = 116},
  [6238] = {.lex_state = 70, .external_lex_state = 118},
  [6239] = {.lex_state = 271, .external_lex_state = 116},
  [6240] = {.lex_state = 70, .external_lex_state = 120},
  [6241] = {.lex_state = 70, .external_lex_state = 118},
  [6242] = {.lex_state = 70, .external_lex_state = 118},
  [6243] = {.lex_state = 70, .external_lex_state = 118},
  [6244] = {.lex_state = 70, .external_lex_state = 118},
  [6245] = {.lex_state = 314, .external_lex_state = 114},
  [6246] = {.lex_state = 314, .external_lex_state = 114},
  [6247] = {.lex_state = 314, .external_lex_state = 114},
  [6248] = {.lex_state = 271},
  [6249] = {.lex_state = 271},
  [6250] = {.lex_state = 271},
  [6251] = {.lex_state = 271},
  [6252] = {.lex_state = 271},
  [6253] = {.lex_state = 271},
  [6254] = {.lex_state = 271},
  [6255] = {.lex_state = 271},
  [6256] = {.lex_state = 314, .external_lex_state = 114},
  [6257] = {.lex_state = 271},
  [6258] = {.lex_state = 314, .external_lex_state = 114},
  [6259] = {.lex_state = 314, .external_lex_state = 114},
  [6260] = {.lex_state = 314, .external_lex_state = 114},
  [6261] = {.lex_state = 314, .external_lex_state = 114},
  [6262] = {.lex_state = 314, .external_lex_state = 114},
  [6263] = {.lex_state = 271},
  [6264] = {.lex_state = 314, .external_lex_state = 114},
  [6265] = {.lex_state = 324},
  [6266] = {.lex_state = 314, .external_lex_state = 114},
  [6267] = {.lex_state = 314, .external_lex_state = 114},
  [6268] = {.lex_state = 314, .external_lex_state = 114},
  [6269] = {.lex_state = 271},
  [6270] = {.lex_state = 65, .external_lex_state = 120},
  [6271] = {.lex_state = 271},
  [6272] = {.lex_state = 271},
  [6273] = {.lex_state = 271},
  [6274] = {.lex_state = 271},
  [6275] = {.lex_state = 314, .external_lex_state = 114},
  [6276] = {.lex_state = 271},
  [6277] = {.lex_state = 271},
  [6278] = {.lex_state = 271},
  [6279] = {.lex_state = 314, .external_lex_state = 114},
  [6280] = {.lex_state = 65, .external_lex_state = 120},
  [6281] = {.lex_state = 271},
  [6282] = {.lex_state = 271},
  [6283] = {.lex_state = 271},
  [6284] = {.lex_state = 271},
  [6285] = {.lex_state = 271},
  [6286] = {.lex_state = 65, .external_lex_state = 120},
  [6287] = {.lex_state = 314, .external_lex_state = 114},
  [6288] = {.lex_state = 271},
  [6289] = {.lex_state = 324},
  [6290] = {.lex_state = 271},
  [6291] = {.lex_state = 271},
  [6292] = {.lex_state = 271},
  [6293] = {.lex_state = 315, .external_lex_state = 119},
  [6294] = {.lex_state = 271},
  [6295] = {.lex_state = 271},
  [6296] = {.lex_state = 271},
  [6297] = {.lex_state = 271},
  [6298] = {.lex_state = 271},
  [6299] = {.lex_state = 314, .external_lex_state = 114},
  [6300] = {.lex_state = 65, .external_lex_state = 120},
  [6301] = {.lex_state = 314, .external_lex_state = 114},
  [6302] = {.lex_state = 297, .external_lex_state = 115},
  [6303] = {.lex_state = 297, .external_lex_state = 115},
  [6304] = {.lex_state = 70, .external_lex_state = 120},
  [6305] = {.lex_state = 271},
  [6306] = {.lex_state = 315, .external_lex_state = 119},
  [6307] = {.lex_state = 271},
  [6308] = {.lex_state = 271},
  [6309] = {.lex_state = 271},
  [6310] = {.lex_state = 297, .external_lex_state = 115},
  [6311] = {.lex_state = 271},
  [6312] = {.lex_state = 297, .external_lex_state = 115},
  [6313] = {.lex_state = 271},
  [6314] = {.lex_state = 326, .external_lex_state = 121},
  [6315] = {.lex_state = 271},
  [6316] = {.lex_state = 271},
  [6317] = {.lex_state = 315, .external_lex_state = 119},
  [6318] = {.lex_state = 297, .external_lex_state = 115},
  [6319] = {.lex_state = 65, .external_lex_state = 120},
  [6320] = {.lex_state = 271},
  [6321] = {.lex_state = 271},
  [6322] = {.lex_state = 271},
  [6323] = {.lex_state = 271},
  [6324] = {.lex_state = 271},
  [6325] = {.lex_state = 271},
  [6326] = {.lex_state = 297, .external_lex_state = 115},
  [6327] = {.lex_state = 324},
  [6328] = {.lex_state = 297, .external_lex_state = 115},
  [6329] = {.lex_state = 271},
  [6330] = {.lex_state = 65, .external_lex_state = 120},
  [6331] = {.lex_state = 324},
  [6332] = {.lex_state = 271},
  [6333] = {.lex_state = 271},
  [6334] = {.lex_state = 65, .external_lex_state = 120},
  [6335] = {.lex_state = 271},
  [6336] = {.lex_state = 271},
  [6337] = {.lex_state = 324},
  [6338] = {.lex_state = 297, .external_lex_state = 115},
  [6339] = {.lex_state = 324},
  [6340] = {.lex_state = 324},
  [6341] = {.lex_state = 271},
  [6342] = {.lex_state = 271},
  [6343] = {.lex_state = 271},
  [6344] = {.lex_state = 314, .external_lex_state = 114},
  [6345] = {.lex_state = 271},
  [6346] = {.lex_state = 326, .external_lex_state = 121},
  [6347] = {.lex_state = 271},
  [6348] = {.lex_state = 271},
  [6349] = {.lex_state = 271},
  [6350] = {.lex_state = 271},
  [6351] = {.lex_state = 271},
  [6352] = {.lex_state = 271},
  [6353] = {.lex_state = 326, .external_lex_state = 121},
  [6354] = {.lex_state = 271},
  [6355] = {.lex_state = 271},
  [6356] = {.lex_state = 271},
  [6357] = {.lex_state = 271},
  [6358] = {.lex_state = 271},
  [6359] = {.lex_state = 271},
  [6360] = {.lex_state = 271},
  [6361] = {.lex_state = 315, .external_lex_state = 119},
  [6362] = {.lex_state = 271},
  [6363] = {.lex_state = 324},
  [6364] = {.lex_state = 297, .external_lex_state = 115},
  [6365] = {.lex_state = 65, .external_lex_state = 120},
  [6366] = {.lex_state = 271},
  [6367] = {.lex_state = 271},
  [6368] = {.lex_state = 542, .external_lex_state = 122},
  [6369] = {.lex_state = 542, .external_lex_state = 122},
  [6370] = {.lex_state = 271, .external_lex_state = 123},
  [6371] = {.lex_state = 542, .external_lex_state = 122},
  [6372] = {.lex_state = 326, .external_lex_state = 124},
  [6373] = {.lex_state = 297, .external_lex_state = 115},
  [6374] = {.lex_state = 70, .external_lex_state = 120},
  [6375] = {.lex_state = 542, .external_lex_state = 122},
  [6376] = {.lex_state = 70, .external_lex_state = 120},
  [6377] = {.lex_state = 542, .external_lex_state = 122},
  [6378] = {.lex_state = 542, .external_lex_state = 122},
  [6379] = {.lex_state = 326, .external_lex_state = 124},
  [6380] = {.lex_state = 542, .external_lex_state = 122},
  [6381] = {.lex_state = 70, .external_lex_state = 120},
  [6382] = {.lex_state = 542, .external_lex_state = 122},
  [6383] = {.lex_state = 542, .external_lex_state = 122},
  [6384] = {.lex_state = 542, .external_lex_state = 122},
  [6385] = {.lex_state = 542, .external_lex_state = 122},
  [6386] = {.lex_state = 313},
  [6387] = {.lex_state = 542, .external_lex_state = 122},
  [6388] = {.lex_state = 271, .external_lex_state = 123},
  [6389] = {.lex_state = 271, .external_lex_state = 125},
  [6390] = {.lex_state = 271, .external_lex_state = 125},
  [6391] = {.lex_state = 542, .external_lex_state = 122},
  [6392] = {.lex_state = 542, .external_lex_state = 122},
  [6393] = {.lex_state = 297, .external_lex_state = 115},
  [6394] = {.lex_state = 271, .external_lex_state = 125},
  [6395] = {.lex_state = 542, .external_lex_state = 122},
  [6396] = {.lex_state = 313},
  [6397] = {.lex_state = 542, .external_lex_state = 122},
  [6398] = {.lex_state = 542, .external_lex_state = 122},
  [6399] = {.lex_state = 542, .external_lex_state = 122},
  [6400] = {.lex_state = 324},
  [6401] = {.lex_state = 542, .external_lex_state = 122},
  [6402] = {.lex_state = 271, .external_lex_state = 125},
  [6403] = {.lex_state = 271, .external_lex_state = 123},
  [6404] = {.lex_state = 542, .external_lex_state = 122},
  [6405] = {.lex_state = 296},
  [6406] = {.lex_state = 542, .external_lex_state = 122},
  [6407] = {.lex_state = 271, .external_lex_state = 123},
  [6408] = {.lex_state = 70, .external_lex_state = 120},
  [6409] = {.lex_state = 296},
  [6410] = {.lex_state = 326, .external_lex_state = 124},
  [6411] = {.lex_state = 542, .external_lex_state = 122},
  [6412] = {.lex_state = 542, .external_lex_state = 122},
  [6413] = {.lex_state = 70, .external_lex_state = 120},
  [6414] = {.lex_state = 542, .external_lex_state = 122},
  [6415] = {.lex_state = 296},
  [6416] = {.lex_state = 542, .external_lex_state = 122},
  [6417] = {.lex_state = 313},
  [6418] = {.lex_state = 542, .external_lex_state = 122},
  [6419] = {.lex_state = 297, .external_lex_state = 115},
  [6420] = {.lex_state = 297, .external_lex_state = 115},
  [6421] = {.lex_state = 542, .external_lex_state = 122},
  [6422] = {.lex_state = 542, .external_lex_state = 122},
  [6423] = {.lex_state = 542, .external_lex_state = 122},
  [6424] = {.lex_state = 271, .external_lex_state = 125},
  [6425] = {.lex_state = 326, .external_lex_state = 124},
  [6426] = {.lex_state = 296},
  [6427] = {.lex_state = 326, .external_lex_state = 124},
  [6428] = {.lex_state = 271, .external_lex_state = 125},
  [6429] = {.lex_state = 542, .external_lex_state = 122},
  [6430] = {.lex_state = 542, .external_lex_state = 122},
  [6431] = {.lex_state = 542, .external_lex_state = 122},
  [6432] = {.lex_state = 542, .external_lex_state = 122},
  [6433] = {.lex_state = 542, .external_lex_state = 122},
  [6434] = {.lex_state = 542, .external_lex_state = 122},
  [6435] = {.lex_state = 326, .external_lex_state = 121},
  [6436] = {.lex_state = 326, .external_lex_state = 121},
  [6437] = {.lex_state = 542, .external_lex_state = 122},
  [6438] = {.lex_state = 326, .external_lex_state = 121},
  [6439] = {.lex_state = 297, .external_lex_state = 115},
  [6440] = {.lex_state = 542, .external_lex_state = 122},
  [6441] = {.lex_state = 313},
  [6442] = {.lex_state = 542, .external_lex_state = 122},
  [6443] = {.lex_state = 70, .external_lex_state = 120},
  [6444] = {.lex_state = 542, .external_lex_state = 122},
  [6445] = {.lex_state = 542, .external_lex_state = 122},
  [6446] = {.lex_state = 542, .external_lex_state = 122},
  [6447] = {.lex_state = 542, .external_lex_state = 122},
  [6448] = {.lex_state = 271, .external_lex_state = 125},
  [6449] = {.lex_state = 326, .external_lex_state = 124},
  [6450] = {.lex_state = 542, .external_lex_state = 122},
  [6451] = {.lex_state = 542, .external_lex_state = 122},
  [6452] = {.lex_state = 542, .external_lex_state = 122},
  [6453] = {.lex_state = 542, .external_lex_state = 122},
  [6454] = {.lex_state = 542, .external_lex_state = 122},
  [6455] = {.lex_state = 542, .external_lex_state = 122},
  [6456] = {.lex_state = 542, .external_lex_state = 122},
  [6457] = {.lex_state = 542, .external_lex_state = 122},
  [6458] = {.lex_state = 542, .external_lex_state = 122},
  [6459] = {.lex_state = 542, .external_lex_state = 122},
  [6460] = {.lex_state = 326, .external_lex_state = 124},
  [6461] = {.lex_state = 542, .external_lex_state = 122},
  [6462] = {.lex_state = 542, .external_lex_state = 122},
  [6463] = {.lex_state = 65, .external_lex_state = 120},
  [6464] = {.lex_state = 542, .external_lex_state = 122},
  [6465] = {.lex_state = 271, .external_lex_state = 125},
  [6466] = {.lex_state = 542, .external_lex_state = 122},
  [6467] = {.lex_state = 542, .external_lex_state = 122},
  [6468] = {.lex_state = 542, .external_lex_state = 122},
  [6469] = {.lex_state = 70, .external_lex_state = 120},
  [6470] = {.lex_state = 322},
  [6471] = {.lex_state = 70, .external_lex_state = 120},
  [6472] = {.lex_state = 271, .external_lex_state = 125},
  [6473] = {.lex_state = 542, .external_lex_state = 122},
  [6474] = {.lex_state = 271, .external_lex_state = 125},
  [6475] = {.lex_state = 326, .external_lex_state = 124},
  [6476] = {.lex_state = 271, .external_lex_state = 125},
  [6477] = {.lex_state = 271, .external_lex_state = 125},
  [6478] = {.lex_state = 271, .external_lex_state = 125},
  [6479] = {.lex_state = 542, .external_lex_state = 122},
  [6480] = {.lex_state = 326, .external_lex_state = 124},
  [6481] = {.lex_state = 65, .external_lex_state = 120},
  [6482] = {.lex_state = 314, .external_lex_state = 114},
  [6483] = {.lex_state = 297, .external_lex_state = 115},
  [6484] = {.lex_state = 297, .external_lex_state = 115},
  [6485] = {.lex_state = 297, .external_lex_state = 115},
  [6486] = {.lex_state = 297, .external_lex_state = 115},
  [6487] = {.lex_state = 326, .external_lex_state = 124},
  [6488] = {.lex_state = 314, .external_lex_state = 114},
  [6489] = {.lex_state = 271, .external_lex_state = 125},
  [6490] = {.lex_state = 326, .external_lex_state = 124},
  [6491] = {.lex_state = 271, .external_lex_state = 126},
  [6492] = {.lex_state = 271, .external_lex_state = 126},
  [6493] = {.lex_state = 65, .external_lex_state = 120},
  [6494] = {.lex_state = 297, .external_lex_state = 115},
  [6495] = {.lex_state = 313},
  [6496] = {.lex_state = 271, .external_lex_state = 126},
  [6497] = {.lex_state = 297, .external_lex_state = 115},
  [6498] = {.lex_state = 275, .external_lex_state = 124},
  [6499] = {.lex_state = 271, .external_lex_state = 125},
  [6500] = {.lex_state = 314, .external_lex_state = 114},
  [6501] = {.lex_state = 271, .external_lex_state = 126},
  [6502] = {.lex_state = 297, .external_lex_state = 115},
  [6503] = {.lex_state = 297, .external_lex_state = 115},
  [6504] = {.lex_state = 297, .external_lex_state = 115},
  [6505] = {.lex_state = 326, .external_lex_state = 124},
  [6506] = {.lex_state = 326, .external_lex_state = 124},
  [6507] = {.lex_state = 326, .external_lex_state = 124},
  [6508] = {.lex_state = 313},
  [6509] = {.lex_state = 297, .external_lex_state = 115},
  [6510] = {.lex_state = 297, .external_lex_state = 115},
  [6511] = {.lex_state = 271, .external_lex_state = 126},
  [6512] = {.lex_state = 313},
  [6513] = {.lex_state = 297, .external_lex_state = 115},
  [6514] = {.lex_state = 326, .external_lex_state = 124},
  [6515] = {.lex_state = 326, .external_lex_state = 124},
  [6516] = {.lex_state = 326, .external_lex_state = 124},
  [6517] = {.lex_state = 326, .external_lex_state = 124},
  [6518] = {.lex_state = 326, .external_lex_state = 124},
  [6519] = {.lex_state = 297, .external_lex_state = 115},
  [6520] = {.lex_state = 297, .external_lex_state = 115},
  [6521] = {.lex_state = 271, .external_lex_state = 126},
  [6522] = {.lex_state = 297, .external_lex_state = 115},
  [6523] = {.lex_state = 297, .external_lex_state = 115},
  [6524] = {.lex_state = 271, .external_lex_state = 126},
  [6525] = {.lex_state = 326, .external_lex_state = 124},
  [6526] = {.lex_state = 297, .external_lex_state = 115},
  [6527] = {.lex_state = 271, .external_lex_state = 125},
  [6528] = {.lex_state = 271, .external_lex_state = 125},
  [6529] = {.lex_state = 271, .external_lex_state = 125},
  [6530] = {.lex_state = 314, .external_lex_state = 114},
  [6531] = {.lex_state = 65, .external_lex_state = 120},
  [6532] = {.lex_state = 326, .external_lex_state = 124},
  [6533] = {.lex_state = 271, .external_lex_state = 126},
  [6534] = {.lex_state = 297, .external_lex_state = 115},
  [6535] = {.lex_state = 297, .external_lex_state = 115},
  [6536] = {.lex_state = 271, .external_lex_state = 126},
  [6537] = {.lex_state = 326, .external_lex_state = 124},
  [6538] = {.lex_state = 271, .external_lex_state = 125},
  [6539] = {.lex_state = 271, .external_lex_state = 126},
  [6540] = {.lex_state = 271, .external_lex_state = 125},
  [6541] = {.lex_state = 271, .external_lex_state = 126},
  [6542] = {.lex_state = 271, .external_lex_state = 126},
  [6543] = {.lex_state = 296},
  [6544] = {.lex_state = 313},
  [6545] = {.lex_state = 296},
  [6546] = {.lex_state = 271, .external_lex_state = 125},
  [6547] = {.lex_state = 326},
  [6548] = {.lex_state = 271},
  [6549] = {.lex_state = 326},
  [6550] = {.lex_state = 271},
  [6551] = {.lex_state = 273},
  [6552] = {.lex_state = 271},
  [6553] = {.lex_state = 271},
  [6554] = {.lex_state = 542, .external_lex_state = 127},
  [6555] = {.lex_state = 542, .external_lex_state = 127},
  [6556] = {.lex_state = 542, .external_lex_state = 127},
  [6557] = {.lex_state = 542, .external_lex_state = 127},
  [6558] = {.lex_state = 256},
  [6559] = {.lex_state = 271},
  [6560] = {.lex_state = 326},
  [6561] = {.lex_state = 314, .external_lex_state = 114},
  [6562] = {.lex_state = 271},
  [6563] = {.lex_state = 542, .external_lex_state = 127},
  [6564] = {.lex_state = 542, .external_lex_state = 127},
  [6565] = {.lex_state = 271},
  [6566] = {.lex_state = 256},
  [6567] = {.lex_state = 542},
  [6568] = {.lex_state = 256},
  [6569] = {.lex_state = 542},
  [6570] = {.lex_state = 542},
  [6571] = {.lex_state = 542, .external_lex_state = 127},
  [6572] = {.lex_state = 256},
  [6573] = {.lex_state = 542, .external_lex_state = 127},
  [6574] = {.lex_state = 271},
  [6575] = {.lex_state = 326},
  [6576] = {.lex_state = 271},
  [6577] = {.lex_state = 271},
  [6578] = {.lex_state = 326},
  [6579] = {.lex_state = 271},
  [6580] = {.lex_state = 271},
  [6581] = {.lex_state = 296},
  [6582] = {.lex_state = 314, .external_lex_state = 117},
  [6583] = {.lex_state = 312, .external_lex_state = 115},
  [6584] = {.lex_state = 256},
  [6585] = {.lex_state = 312, .external_lex_state = 115},
  [6586] = {.lex_state = 271},
  [6587] = {.lex_state = 271},
  [6588] = {.lex_state = 271},
  [6589] = {.lex_state = 273},
  [6590] = {.lex_state = 542},
  [6591] = {.lex_state = 542},
  [6592] = {.lex_state = 271},
  [6593] = {.lex_state = 271},
  [6594] = {.lex_state = 271},
  [6595] = {.lex_state = 542},
  [6596] = {.lex_state = 542},
  [6597] = {.lex_state = 542},
  [6598] = {.lex_state = 542},
  [6599] = {.lex_state = 542},
  [6600] = {.lex_state = 271},
  [6601] = {.lex_state = 271},
  [6602] = {.lex_state = 271},
  [6603] = {.lex_state = 326},
  [6604] = {.lex_state = 326},
  [6605] = {.lex_state = 273},
  [6606] = {.lex_state = 271},
  [6607] = {.lex_state = 271},
  [6608] = {.lex_state = 271},
  [6609] = {.lex_state = 312, .external_lex_state = 115},
  [6610] = {.lex_state = 312, .external_lex_state = 115},
  [6611] = {.lex_state = 542},
  [6612] = {.lex_state = 271},
  [6613] = {.lex_state = 271},
  [6614] = {.lex_state = 256},
  [6615] = {.lex_state = 314, .external_lex_state = 114},
  [6616] = {.lex_state = 326},
  [6617] = {.lex_state = 271},
  [6618] = {.lex_state = 271},
  [6619] = {.lex_state = 326},
  [6620] = {.lex_state = 542},
  [6621] = {.lex_state = 271},
  [6622] = {.lex_state = 271},
  [6623] = {.lex_state = 271},
  [6624] = {.lex_state = 271},
  [6625] = {.lex_state = 271},
  [6626] = {.lex_state = 271},
  [6627] = {.lex_state = 542},
  [6628] = {.lex_state = 326},
  [6629] = {.lex_state = 271},
  [6630] = {.lex_state = 271},
  [6631] = {.lex_state = 241, .external_lex_state = 115},
  [6632] = {.lex_state = 271},
  [6633] = {.lex_state = 271},
  [6634] = {.lex_state = 326},
  [6635] = {.lex_state = 314, .external_lex_state = 117},
  [6636] = {.lex_state = 326},
  [6637] = {.lex_state = 542},
  [6638] = {.lex_state = 326},
  [6639] = {.lex_state = 271},
  [6640] = {.lex_state = 273},
  [6641] = {.lex_state = 326},
  [6642] = {.lex_state = 271},
  [6643] = {.lex_state = 271},
  [6644] = {.lex_state = 271},
  [6645] = {.lex_state = 314, .external_lex_state = 117},
  [6646] = {.lex_state = 271},
  [6647] = {.lex_state = 271},
  [6648] = {.lex_state = 271},
  [6649] = {.lex_state = 542},
  [6650] = {.lex_state = 271},
  [6651] = {.lex_state = 314, .external_lex_state = 114},
  [6652] = {.lex_state = 542},
  [6653] = {.lex_state = 271},
  [6654] = {.lex_state = 271},
  [6655] = {.lex_state = 271},
  [6656] = {.lex_state = 271},
  [6657] = {.lex_state = 271},
  [6658] = {.lex_state = 271},
  [6659] = {.lex_state = 273},
  [6660] = {.lex_state = 271},
  [6661] = {.lex_state = 271},
  [6662] = {.lex_state = 326},
  [6663] = {.lex_state = 314, .external_lex_state = 117},
  [6664] = {.lex_state = 326},
  [6665] = {.lex_state = 271},
  [6666] = {.lex_state = 542},
  [6667] = {.lex_state = 542},
  [6668] = {.lex_state = 326},
  [6669] = {.lex_state = 271},
  [6670] = {.lex_state = 271},
  [6671] = {.lex_state = 273},
  [6672] = {.lex_state = 542},
  [6673] = {.lex_state = 326},
  [6674] = {.lex_state = 256},
  [6675] = {.lex_state = 271},
  [6676] = {.lex_state = 542},
  [6677] = {.lex_state = 542, .external_lex_state = 127},
  [6678] = {.lex_state = 271},
  [6679] = {.lex_state = 273},
  [6680] = {.lex_state = 271},
  [6681] = {.lex_state = 271},
  [6682] = {.lex_state = 271},
  [6683] = {.lex_state = 271},
  [6684] = {.lex_state = 271},
  [6685] = {.lex_state = 271},
  [6686] = {.lex_state = 326},
  [6687] = {.lex_state = 271},
  [6688] = {.lex_state = 271},
  [6689] = {.lex_state = 271},
  [6690] = {.lex_state = 542},
  [6691] = {.lex_state = 271},
  [6692] = {.lex_state = 271},
  [6693] = {.lex_state = 271},
  [6694] = {.lex_state = 542},
  [6695] = {.lex_state = 542},
  [6696] = {.lex_state = 326},
  [6697] = {.lex_state = 271},
  [6698] = {.lex_state = 271},
  [6699] = {.lex_state = 271},
  [6700] = {.lex_state = 542},
  [6701] = {.lex_state = 271},
  [6702] = {.lex_state = 271},
  [6703] = {.lex_state = 542},
  [6704] = {.lex_state = 326},
  [6705] = {.lex_state = 326},
  [6706] = {.lex_state = 271},
  [6707] = {.lex_state = 542},
  [6708] = {.lex_state = 271},
  [6709] = {.lex_state = 542},
  [6710] = {.lex_state = 542},
  [6711] = {.lex_state = 542},
  [6712] = {.lex_state = 271},
  [6713] = {.lex_state = 271},
  [6714] = {.lex_state = 271},
  [6715] = {.lex_state = 271},
  [6716] = {.lex_state = 542},
  [6717] = {.lex_state = 326},
  [6718] = {.lex_state = 256},
  [6719] = {.lex_state = 542},
  [6720] = {.lex_state = 271},
  [6721] = {.lex_state = 271},
  [6722] = {.lex_state = 542},
  [6723] = {.lex_state = 271},
  [6724] = {.lex_state = 271},
  [6725] = {.lex_state = 542},
  [6726] = {.lex_state = 312},
  [6727] = {.lex_state = 312},
  [6728] = {.lex_state = 273},
  [6729] = {.lex_state = 273},
  [6730] = {.lex_state = 324},
  [6731] = {.lex_state = 312},
  [6732] = {.lex_state = 324},
  [6733] = {.lex_state = 312},
  [6734] = {.lex_state = 312},
  [6735] = {.lex_state = 271, .external_lex_state = 114},
  [6736] = {.lex_state = 312},
  [6737] = {.lex_state = 312},
  [6738] = {.lex_state = 271},
  [6739] = {.lex_state = 312},
  [6740] = {.lex_state = 314, .external_lex_state = 114},
  [6741] = {.lex_state = 271},
  [6742] = {.lex_state = 273},
  [6743] = {.lex_state = 324},
  [6744] = {.lex_state = 324},
  [6745] = {.lex_state = 324},
  [6746] = {.lex_state = 312},
  [6747] = {.lex_state = 314, .external_lex_state = 114},
  [6748] = {.lex_state = 314, .external_lex_state = 114},
  [6749] = {.lex_state = 314, .external_lex_state = 114},
  [6750] = {.lex_state = 312},
  [6751] = {.lex_state = 312},
  [6752] = {.lex_state = 273},
  [6753] = {.lex_state = 314, .external_lex_state = 114},
  [6754] = {.lex_state = 312},
  [6755] = {.lex_state = 271, .external_lex_state = 119},
  [6756] = {.lex_state = 312},
  [6757] = {.lex_state = 312},
  [6758] = {.lex_state = 312},
  [6759] = {.lex_state = 312},
  [6760] = {.lex_state = 312},
  [6761] = {.lex_state = 314, .external_lex_state = 114},
  [6762] = {.lex_state = 273},
  [6763] = {.lex_state = 324},
  [6764] = {.lex_state = 314, .external_lex_state = 114},
  [6765] = {.lex_state = 314, .external_lex_state = 114},
  [6766] = {.lex_state = 542},
  [6767] = {.lex_state = 314, .external_lex_state = 114},
  [6768] = {.lex_state = 271},
  [6769] = {.lex_state = 312},
  [6770] = {.lex_state = 324},
  [6771] = {.lex_state = 312},
  [6772] = {.lex_state = 312},
  [6773] = {.lex_state = 312},
  [6774] = {.lex_state = 312},
  [6775] = {.lex_state = 542},
  [6776] = {.lex_state = 312},
  [6777] = {.lex_state = 273},
  [6778] = {.lex_state = 312},
  [6779] = {.lex_state = 314, .external_lex_state = 114},
  [6780] = {.lex_state = 271, .external_lex_state = 114},
  [6781] = {.lex_state = 312},
  [6782] = {.lex_state = 314, .external_lex_state = 114},
  [6783] = {.lex_state = 324},
  [6784] = {.lex_state = 314, .external_lex_state = 114},
  [6785] = {.lex_state = 273},
  [6786] = {.lex_state = 312},
  [6787] = {.lex_state = 314, .external_lex_state = 114},
  [6788] = {.lex_state = 314, .external_lex_state = 114},
  [6789] = {.lex_state = 314, .external_lex_state = 114},
  [6790] = {.lex_state = 312},
  [6791] = {.lex_state = 314, .external_lex_state = 114},
  [6792] = {.lex_state = 112, .external_lex_state = 120},
  [6793] = {.lex_state = 271},
  [6794] = {.lex_state = 542},
  [6795] = {.lex_state = 542},
  [6796] = {.lex_state = 542, .external_lex_state = 128},
  [6797] = {.lex_state = 271},
  [6798] = {.lex_state = 271, .external_lex_state = 114},
  [6799] = {.lex_state = 542},
  [6800] = {.lex_state = 271, .external_lex_state = 114},
  [6801] = {.lex_state = 542},
  [6802] = {.lex_state = 271},
  [6803] = {.lex_state = 271},
  [6804] = {.lex_state = 542},
  [6805] = {.lex_state = 542},
  [6806] = {.lex_state = 112, .external_lex_state = 120},
  [6807] = {.lex_state = 271, .external_lex_state = 114},
  [6808] = {.lex_state = 542},
  [6809] = {.lex_state = 542},
  [6810] = {.lex_state = 271, .external_lex_state = 114},
  [6811] = {.lex_state = 271},
  [6812] = {.lex_state = 271},
  [6813] = {.lex_state = 542},
  [6814] = {.lex_state = 279},
  [6815] = {.lex_state = 542},
  [6816] = {.lex_state = 542},
  [6817] = {.lex_state = 279},
  [6818] = {.lex_state = 542},
  [6819] = {.lex_state = 542},
  [6820] = {.lex_state = 271},
  [6821] = {.lex_state = 542},
  [6822] = {.lex_state = 542},
  [6823] = {.lex_state = 542},
  [6824] = {.lex_state = 271, .external_lex_state = 126},
  [6825] = {.lex_state = 271, .external_lex_state = 114},
  [6826] = {.lex_state = 542, .external_lex_state = 128},
  [6827] = {.lex_state = 542},
  [6828] = {.lex_state = 279},
  [6829] = {.lex_state = 271},
  [6830] = {.lex_state = 279},
  [6831] = {.lex_state = 542},
  [6832] = {.lex_state = 271, .external_lex_state = 114},
  [6833] = {.lex_state = 257, .external_lex_state = 129},
  [6834] = {.lex_state = 257, .external_lex_state = 129},
  [6835] = {.lex_state = 112, .external_lex_state = 120},
  [6836] = {.lex_state = 112, .external_lex_state = 120},
  [6837] = {.lex_state = 112, .external_lex_state = 120},
  [6838] = {.lex_state = 271},
  [6839] = {.lex_state = 112, .external_lex_state = 120},
  [6840] = {.lex_state = 542},
  [6841] = {.lex_state = 271, .external_lex_state = 114},
  [6842] = {.lex_state = 271, .external_lex_state = 114},
  [6843] = {.lex_state = 112, .external_lex_state = 120},
  [6844] = {.lex_state = 112, .external_lex_state = 120},
  [6845] = {.lex_state = 542, .external_lex_state = 128},
  [6846] = {.lex_state = 271},
  [6847] = {.lex_state = 271},
  [6848] = {.lex_state = 542},
  [6849] = {.lex_state = 542},
  [6850] = {.lex_state = 542},
  [6851] = {.lex_state = 271, .external_lex_state = 126},
  [6852] = {.lex_state = 279},
  [6853] = {.lex_state = 112, .external_lex_state = 120},
  [6854] = {.lex_state = 542},
  [6855] = {.lex_state = 542},
  [6856] = {.lex_state = 271},
  [6857] = {.lex_state = 542},
  [6858] = {.lex_state = 279},
  [6859] = {.lex_state = 542},
  [6860] = {.lex_state = 271, .external_lex_state = 114},
  [6861] = {.lex_state = 542},
  [6862] = {.lex_state = 324},
  [6863] = {.lex_state = 542},
  [6864] = {.lex_state = 542},
  [6865] = {.lex_state = 271},
  [6866] = {.lex_state = 112, .external_lex_state = 120},
  [6867] = {.lex_state = 542},
  [6868] = {.lex_state = 112, .external_lex_state = 120},
  [6869] = {.lex_state = 279},
  [6870] = {.lex_state = 112, .external_lex_state = 120},
  [6871] = {.lex_state = 112, .external_lex_state = 120},
  [6872] = {.lex_state = 542},
  [6873] = {.lex_state = 112, .external_lex_state = 120},
  [6874] = {.lex_state = 271},
  [6875] = {.lex_state = 112, .external_lex_state = 120},
  [6876] = {.lex_state = 271, .external_lex_state = 114},
  [6877] = {.lex_state = 542},
  [6878] = {.lex_state = 542},
  [6879] = {.lex_state = 542},
  [6880] = {.lex_state = 542},
  [6881] = {.lex_state = 542},
  [6882] = {.lex_state = 271},
  [6883] = {.lex_state = 271},
  [6884] = {.lex_state = 271},
  [6885] = {.lex_state = 279},
  [6886] = {.lex_state = 256},
  [6887] = {.lex_state = 271},
  [6888] = {.lex_state = 542},
  [6889] = {.lex_state = 542},
  [6890] = {.lex_state = 542},
  [6891] = {.lex_state = 271, .external_lex_state = 114},
  [6892] = {.lex_state = 271},
  [6893] = {.lex_state = 542},
  [6894] = {.lex_state = 324},
  [6895] = {.lex_state = 271, .external_lex_state = 114},
  [6896] = {.lex_state = 542},
  [6897] = {.lex_state = 542},
  [6898] = {.lex_state = 271, .external_lex_state = 126},
  [6899] = {.lex_state = 542},
  [6900] = {.lex_state = 271, .external_lex_state = 114},
  [6901] = {.lex_state = 271},
  [6902] = {.lex_state = 542},
  [6903] = {.lex_state = 542},
  [6904] = {.lex_state = 271, .external_lex_state = 114},
  [6905] = {.lex_state = 542},
  [6906] = {.lex_state = 542},
  [6907] = {.lex_state = 542},
  [6908] = {.lex_state = 271, .external_lex_state = 114},
  [6909] = {.lex_state = 542},
  [6910] = {.lex_state = 271},
  [6911] = {.lex_state = 542},
  [6912] = {.lex_state = 279},
  [6913] = {.lex_state = 246},
  [6914] = {.lex_state = 542},
  [6915] = {.lex_state = 542},
  [6916] = {.lex_state = 279},
  [6917] = {.lex_state = 542},
  [6918] = {.lex_state = 542},
  [6919] = {.lex_state = 271},
  [6920] = {.lex_state = 246},
  [6921] = {.lex_state = 324},
  [6922] = {.lex_state = 271, .external_lex_state = 114},
  [6923] = {.lex_state = 271, .external_lex_state = 126},
  [6924] = {.lex_state = 542},
  [6925] = {.lex_state = 271, .external_lex_state = 114},
  [6926] = {.lex_state = 271},
  [6927] = {.lex_state = 542},
  [6928] = {.lex_state = 112, .external_lex_state = 120},
  [6929] = {.lex_state = 542},
  [6930] = {.lex_state = 542},
  [6931] = {.lex_state = 112, .external_lex_state = 120},
  [6932] = {.lex_state = 542},
  [6933] = {.lex_state = 271},
  [6934] = {.lex_state = 542},
  [6935] = {.lex_state = 112, .external_lex_state = 120},
  [6936] = {.lex_state = 112, .external_lex_state = 120},
  [6937] = {.lex_state = 542},
  [6938] = {.lex_state = 542},
  [6939] = {.lex_state = 112, .external_lex_state = 120},
  [6940] = {.lex_state = 271},
  [6941] = {.lex_state = 112, .external_lex_state = 120},
  [6942] = {.lex_state = 271, .external_lex_state = 126},
  [6943] = {.lex_state = 542},
  [6944] = {.lex_state = 112, .external_lex_state = 120},
  [6945] = {.lex_state = 112, .external_lex_state = 120},
  [6946] = {.lex_state = 271, .external_lex_state = 114},
  [6947] = {.lex_state = 271},
  [6948] = {.lex_state = 271},
  [6949] = {.lex_state = 279},
  [6950] = {.lex_state = 271},
  [6951] = {.lex_state = 542},
  [6952] = {.lex_state = 256},
  [6953] = {.lex_state = 271},
  [6954] = {.lex_state = 271},
  [6955] = {.lex_state = 542},
  [6956] = {.lex_state = 542},
  [6957] = {.lex_state = 271, .external_lex_state = 114},
  [6958] = {.lex_state = 271, .external_lex_state = 114},
  [6959] = {.lex_state = 271, .external_lex_state = 114},
  [6960] = {.lex_state = 271, .external_lex_state = 114},
  [6961] = {.lex_state = 271},
  [6962] = {.lex_state = 257, .external_lex_state = 129},
  [6963] = {.lex_state = 542},
  [6964] = {.lex_state = 271, .external_lex_state = 114},
  [6965] = {.lex_state = 271, .external_lex_state = 114},
  [6966] = {.lex_state = 271, .external_lex_state = 114},
  [6967] = {.lex_state = 271, .external_lex_state = 114},
  [6968] = {.lex_state = 271},
  [6969] = {.lex_state = 257, .external_lex_state = 129},
  [6970] = {.lex_state = 324},
  [6971] = {.lex_state = 279},
  [6972] = {.lex_state = 271, .external_lex_state = 114},
  [6973] = {.lex_state = 271, .external_lex_state = 114},
  [6974] = {.lex_state = 271, .external_lex_state = 114},
  [6975] = {.lex_state = 271},
  [6976] = {.lex_state = 542},
  [6977] = {.lex_state = 271, .external_lex_state = 114},
  [6978] = {.lex_state = 271, .external_lex_state = 114},
  [6979] = {.lex_state = 271, .external_lex_state = 114},
  [6980] = {.lex_state = 271, .external_lex_state = 114},
  [6981] = {.lex_state = 324},
  [6982] = {.lex_state = 271},
  [6983] = {.lex_state = 542},
  [6984] = {.lex_state = 271, .external_lex_state = 114},
  [6985] = {.lex_state = 271, .external_lex_state = 114},
  [6986] = {.lex_state = 271, .external_lex_state = 114},
  [6987] = {.lex_state = 271, .external_lex_state = 114},
  [6988] = {.lex_state = 271, .external_lex_state = 114},
  [6989] = {.lex_state = 271},
  [6990] = {.lex_state = 257, .external_lex_state = 129},
  [6991] = {.lex_state = 271, .external_lex_state = 126},
  [6992] = {.lex_state = 271, .external_lex_state = 114},
  [6993] = {.lex_state = 271, .external_lex_state = 114},
  [6994] = {.lex_state = 271, .external_lex_state = 114},
  [6995] = {.lex_state = 271, .external_lex_state = 114},
  [6996] = {.lex_state = 271},
  [6997] = {.lex_state = 271, .external_lex_state = 126},
  [6998] = {.lex_state = 112, .external_lex_state = 120},
  [6999] = {.lex_state = 271, .external_lex_state = 114},
  [7000] = {.lex_state = 271, .external_lex_state = 114},
  [7001] = {.lex_state = 271, .external_lex_state = 114},
  [7002] = {.lex_state = 271, .external_lex_state = 114},
  [7003] = {.lex_state = 271},
  [7004] = {.lex_state = 324},
  [7005] = {.lex_state = 271, .external_lex_state = 126},
  [7006] = {.lex_state = 271, .external_lex_state = 114},
  [7007] = {.lex_state = 271, .external_lex_state = 126},
  [7008] = {.lex_state = 271, .external_lex_state = 114},
  [7009] = {.lex_state = 542, .external_lex_state = 128},
  [7010] = {.lex_state = 271},
  [7011] = {.lex_state = 542, .external_lex_state = 128},
  [7012] = {.lex_state = 324},
  [7013] = {.lex_state = 271, .external_lex_state = 126},
  [7014] = {.lex_state = 271, .external_lex_state = 114},
  [7015] = {.lex_state = 279},
  [7016] = {.lex_state = 271, .external_lex_state = 126},
  [7017] = {.lex_state = 271},
  [7018] = {.lex_state = 257, .external_lex_state = 129},
  [7019] = {.lex_state = 279},
  [7020] = {.lex_state = 271, .external_lex_state = 126},
  [7021] = {.lex_state = 271, .external_lex_state = 126},
  [7022] = {.lex_state = 271, .external_lex_state = 114},
  [7023] = {.lex_state = 271, .external_lex_state = 114},
  [7024] = {.lex_state = 271},
  [7025] = {.lex_state = 271, .external_lex_state = 114},
  [7026] = {.lex_state = 271, .external_lex_state = 114},
  [7027] = {.lex_state = 271, .external_lex_state = 114},
  [7028] = {.lex_state = 112, .external_lex_state = 120},
  [7029] = {.lex_state = 271, .external_lex_state = 114},
  [7030] = {.lex_state = 112, .external_lex_state = 120},
  [7031] = {.lex_state = 271},
  [7032] = {.lex_state = 542},
  [7033] = {.lex_state = 542},
  [7034] = {.lex_state = 542},
  [7035] = {.lex_state = 271, .external_lex_state = 126},
  [7036] = {.lex_state = 112, .external_lex_state = 120},
  [7037] = {.lex_state = 246},
  [7038] = {.lex_state = 271},
  [7039] = {.lex_state = 112, .external_lex_state = 120},
  [7040] = {.lex_state = 542},
  [7041] = {.lex_state = 112, .external_lex_state = 120},
  [7042] = {.lex_state = 112, .external_lex_state = 120},
  [7043] = {.lex_state = 542},
  [7044] = {.lex_state = 271, .external_lex_state = 126},
  [7045] = {.lex_state = 271},
  [7046] = {.lex_state = 271, .external_lex_state = 114},
  [7047] = {.lex_state = 542},
  [7048] = {.lex_state = 542},
  [7049] = {.lex_state = 271, .external_lex_state = 126},
  [7050] = {.lex_state = 542},
  [7051] = {.lex_state = 271, .external_lex_state = 114},
  [7052] = {.lex_state = 271},
  [7053] = {.lex_state = 542},
  [7054] = {.lex_state = 279},
  [7055] = {.lex_state = 271},
  [7056] = {.lex_state = 271, .external_lex_state = 126},
  [7057] = {.lex_state = 271},
  [7058] = {.lex_state = 542},
  [7059] = {.lex_state = 271},
  [7060] = {.lex_state = 542},
  [7061] = {.lex_state = 256},
  [7062] = {.lex_state = 271},
  [7063] = {.lex_state = 542},
  [7064] = {.lex_state = 542},
  [7065] = {.lex_state = 271, .external_lex_state = 114},
  [7066] = {.lex_state = 271},
  [7067] = {.lex_state = 542},
  [7068] = {.lex_state = 324},
  [7069] = {.lex_state = 324},
  [7070] = {.lex_state = 542},
  [7071] = {.lex_state = 279},
  [7072] = {.lex_state = 542},
  [7073] = {.lex_state = 271},
  [7074] = {.lex_state = 279},
  [7075] = {.lex_state = 542},
  [7076] = {.lex_state = 279},
  [7077] = {.lex_state = 542},
  [7078] = {.lex_state = 542},
  [7079] = {.lex_state = 542},
  [7080] = {.lex_state = 271},
  [7081] = {.lex_state = 542},
  [7082] = {.lex_state = 271, .external_lex_state = 114},
  [7083] = {.lex_state = 542},
  [7084] = {.lex_state = 542},
  [7085] = {.lex_state = 542},
  [7086] = {.lex_state = 542},
  [7087] = {.lex_state = 271},
  [7088] = {.lex_state = 542},
  [7089] = {.lex_state = 271, .external_lex_state = 114},
  [7090] = {.lex_state = 279},
  [7091] = {.lex_state = 542},
  [7092] = {.lex_state = 542},
  [7093] = {.lex_state = 542},
  [7094] = {.lex_state = 271},
  [7095] = {.lex_state = 542},
  [7096] = {.lex_state = 542},
  [7097] = {.lex_state = 542},
  [7098] = {.lex_state = 542},
  [7099] = {.lex_state = 542},
  [7100] = {.lex_state = 542},
  [7101] = {.lex_state = 271},
  [7102] = {.lex_state = 271, .external_lex_state = 114},
  [7103] = {.lex_state = 542},
  [7104] = {.lex_state = 271, .external_lex_state = 114},
  [7105] = {.lex_state = 271, .external_lex_state = 114},
  [7106] = {.lex_state = 279},
  [7107] = {.lex_state = 542},
  [7108] = {.lex_state = 271},
  [7109] = {.lex_state = 542},
  [7110] = {.lex_state = 542},
  [7111] = {.lex_state = 542},
  [7112] = {.lex_state = 271, .external_lex_state = 114},
  [7113] = {.lex_state = 271, .external_lex_state = 114},
  [7114] = {.lex_state = 324},
  [7115] = {.lex_state = 271},
  [7116] = {.lex_state = 279},
  [7117] = {.lex_state = 542},
  [7118] = {.lex_state = 542},
  [7119] = {.lex_state = 542},
  [7120] = {.lex_state = 542},
  [7121] = {.lex_state = 246},
  [7122] = {.lex_state = 271},
  [7123] = {.lex_state = 542},
  [7124] = {.lex_state = 542},
  [7125] = {.lex_state = 542},
  [7126] = {.lex_state = 271, .external_lex_state = 114},
  [7127] = {.lex_state = 542},
  [7128] = {.lex_state = 271, .external_lex_state = 114},
  [7129] = {.lex_state = 271},
  [7130] = {.lex_state = 542},
  [7131] = {.lex_state = 271, .external_lex_state = 114},
  [7132] = {.lex_state = 542},
  [7133] = {.lex_state = 542},
  [7134] = {.lex_state = 271, .external_lex_state = 126},
  [7135] = {.lex_state = 271, .external_lex_state = 114},
  [7136] = {.lex_state = 271},
  [7137] = {.lex_state = 271, .external_lex_state = 114},
  [7138] = {.lex_state = 542},
  [7139] = {.lex_state = 542},
  [7140] = {.lex_state = 542},
  [7141] = {.lex_state = 271, .external_lex_state = 114},
  [7142] = {.lex_state = 542, .external_lex_state = 128},
  [7143] = {.lex_state = 271},
  [7144] = {.lex_state = 542},
  [7145] = {.lex_state = 279},
  [7146] = {.lex_state = 271, .external_lex_state = 114},
  [7147] = {.lex_state = 542},
  [7148] = {.lex_state = 542},
  [7149] = {.lex_state = 279},
  [7150] = {.lex_state = 271},
  [7151] = {.lex_state = 542},
  [7152] = {.lex_state = 542},
  [7153] = {.lex_state = 542},
  [7154] = {.lex_state = 542},
  [7155] = {.lex_state = 542},
  [7156] = {.lex_state = 542},
  [7157] = {.lex_state = 271},
  [7158] = {.lex_state = 542},
  [7159] = {.lex_state = 246},
  [7160] = {.lex_state = 542},
  [7161] = {.lex_state = 542},
  [7162] = {.lex_state = 271, .external_lex_state = 114},
  [7163] = {.lex_state = 271, .external_lex_state = 114},
  [7164] = {.lex_state = 271},
  [7165] = {.lex_state = 542},
  [7166] = {.lex_state = 542},
  [7167] = {.lex_state = 542},
  [7168] = {.lex_state = 542},
  [7169] = {.lex_state = 542},
  [7170] = {.lex_state = 271, .external_lex_state = 114},
  [7171] = {.lex_state = 271},
  [7172] = {.lex_state = 279},
  [7173] = {.lex_state = 542},
  [7174] = {.lex_state = 542},
  [7175] = {.lex_state = 542},
  [7176] = {.lex_state = 279},
  [7177] = {.lex_state = 257, .external_lex_state = 129},
  [7178] = {.lex_state = 271},
  [7179] = {.lex_state = 271, .external_lex_state = 114},
  [7180] = {.lex_state = 542},
  [7181] = {.lex_state = 257, .external_lex_state = 129},
  [7182] = {.lex_state = 542},
  [7183] = {.lex_state = 542},
  [7184] = {.lex_state = 271, .external_lex_state = 114},
  [7185] = {.lex_state = 271},
  [7186] = {.lex_state = 542},
  [7187] = {.lex_state = 271, .external_lex_state = 114},
  [7188] = {.lex_state = 246},
  [7189] = {.lex_state = 542},
  [7190] = {.lex_state = 542},
  [7191] = {.lex_state = 279},
  [7192] = {.lex_state = 271},
  [7193] = {.lex_state = 257, .external_lex_state = 129},
  [7194] = {.lex_state = 542},
  [7195] = {.lex_state = 542},
  [7196] = {.lex_state = 271, .external_lex_state = 114},
  [7197] = {.lex_state = 271, .external_lex_state = 114},
  [7198] = {.lex_state = 257, .external_lex_state = 129},
  [7199] = {.lex_state = 271},
  [7200] = {.lex_state = 542},
  [7201] = {.lex_state = 542},
  [7202] = {.lex_state = 542},
  [7203] = {.lex_state = 542},
  [7204] = {.lex_state = 542},
  [7205] = {.lex_state = 542, .external_lex_state = 130},
  [7206] = {.lex_state = 271},
  [7207] = {.lex_state = 271, .external_lex_state = 114},
  [7208] = {.lex_state = 542},
  [7209] = {.lex_state = 542},
  [7210] = {.lex_state = 542},
  [7211] = {.lex_state = 542},
  [7212] = {.lex_state = 542},
  [7213] = {.lex_state = 271},
  [7214] = {.lex_state = 542},
  [7215] = {.lex_state = 542},
  [7216] = {.lex_state = 542},
  [7217] = {.lex_state = 542},
  [7218] = {.lex_state = 271, .external_lex_state = 114},
  [7219] = {.lex_state = 271, .external_lex_state = 114},
  [7220] = {.lex_state = 271, .external_lex_state = 114},
  [7221] = {.lex_state = 271, .external_lex_state = 114},
  [7222] = {.lex_state = 271, .external_lex_state = 114},
  [7223] = {.lex_state = 271, .external_lex_state = 114},
  [7224] = {.lex_state = 542},
  [7225] = {.lex_state = 271},
  [7226] = {.lex_state = 279},
  [7227] = {.lex_state = 542},
  [7228] = {.lex_state = 542},
  [7229] = {.lex_state = 542},
  [7230] = {.lex_state = 542},
  [7231] = {.lex_state = 542},
  [7232] = {.lex_state = 271, .external_lex_state = 114},
  [7233] = {.lex_state = 271, .external_lex_state = 114},
  [7234] = {.lex_state = 271, .external_lex_state = 114},
  [7235] = {.lex_state = 271, .external_lex_state = 114},
  [7236] = {.lex_state = 271, .external_lex_state = 114},
  [7237] = {.lex_state = 271, .external_lex_state = 114},
  [7238] = {.lex_state = 271, .external_lex_state = 114},
  [7239] = {.lex_state = 271, .external_lex_state = 114},
  [7240] = {.lex_state = 271, .external_lex_state = 114},
  [7241] = {.lex_state = 271, .external_lex_state = 114},
  [7242] = {.lex_state = 271, .external_lex_state = 114},
  [7243] = {.lex_state = 271, .external_lex_state = 114},
  [7244] = {.lex_state = 279},
  [7245] = {.lex_state = 271, .external_lex_state = 114},
  [7246] = {.lex_state = 542},
  [7247] = {.lex_state = 279},
  [7248] = {.lex_state = 279},
  [7249] = {.lex_state = 542},
  [7250] = {.lex_state = 542},
  [7251] = {.lex_state = 271, .external_lex_state = 114},
  [7252] = {.lex_state = 542},
  [7253] = {.lex_state = 542},
  [7254] = {.lex_state = 542},
  [7255] = {.lex_state = 271, .external_lex_state = 114},
  [7256] = {.lex_state = 542},
  [7257] = {.lex_state = 271, .external_lex_state = 114},
  [7258] = {.lex_state = 271},
  [7259] = {.lex_state = 271},
  [7260] = {.lex_state = 271, .external_lex_state = 114},
  [7261] = {.lex_state = 271, .external_lex_state = 114},
  [7262] = {.lex_state = 542},
  [7263] = {.lex_state = 279},
  [7264] = {.lex_state = 542},
  [7265] = {.lex_state = 273},
  [7266] = {.lex_state = 542},
  [7267] = {.lex_state = 542},
  [7268] = {.lex_state = 324},
  [7269] = {.lex_state = 271},
  [7270] = {.lex_state = 542},
  [7271] = {.lex_state = 271, .external_lex_state = 114},
  [7272] = {.lex_state = 271, .external_lex_state = 114},
  [7273] = {.lex_state = 542},
  [7274] = {.lex_state = 542, .external_lex_state = 130},
  [7275] = {.lex_state = 271, .external_lex_state = 126},
  [7276] = {.lex_state = 279},
  [7277] = {.lex_state = 542},
  [7278] = {.lex_state = 279},
  [7279] = {.lex_state = 542},
  [7280] = {.lex_state = 542},
  [7281] = {.lex_state = 279},
  [7282] = {.lex_state = 542},
  [7283] = {.lex_state = 542},
  [7284] = {.lex_state = 271, .external_lex_state = 114},
  [7285] = {.lex_state = 271, .external_lex_state = 114},
  [7286] = {.lex_state = 542},
  [7287] = {.lex_state = 279},
  [7288] = {.lex_state = 542},
  [7289] = {.lex_state = 542},
  [7290] = {.lex_state = 542},
  [7291] = {.lex_state = 271, .external_lex_state = 126},
  [7292] = {.lex_state = 542, .external_lex_state = 130},
  [7293] = {.lex_state = 542},
  [7294] = {.lex_state = 542},
  [7295] = {.lex_state = 542},
  [7296] = {.lex_state = 271, .external_lex_state = 114},
  [7297] = {.lex_state = 279},
  [7298] = {.lex_state = 271, .external_lex_state = 114},
  [7299] = {.lex_state = 542},
  [7300] = {.lex_state = 542},
  [7301] = {.lex_state = 542},
  [7302] = {.lex_state = 271, .external_lex_state = 114},
  [7303] = {.lex_state = 271, .external_lex_state = 114},
  [7304] = {.lex_state = 542},
  [7305] = {.lex_state = 542},
  [7306] = {.lex_state = 542},
  [7307] = {.lex_state = 542},
  [7308] = {.lex_state = 271, .external_lex_state = 114},
  [7309] = {.lex_state = 279},
  [7310] = {.lex_state = 542},
  [7311] = {.lex_state = 542},
  [7312] = {.lex_state = 542},
  [7313] = {.lex_state = 271, .external_lex_state = 114},
  [7314] = {.lex_state = 542},
  [7315] = {.lex_state = 271, .external_lex_state = 126},
  [7316] = {.lex_state = 542},
  [7317] = {.lex_state = 271, .external_lex_state = 114},
  [7318] = {.lex_state = 246},
  [7319] = {.lex_state = 542},
  [7320] = {.lex_state = 324},
  [7321] = {.lex_state = 542},
  [7322] = {.lex_state = 279},
  [7323] = {.lex_state = 542},
  [7324] = {.lex_state = 279},
  [7325] = {.lex_state = 542},
  [7326] = {.lex_state = 256},
  [7327] = {.lex_state = 542},
  [7328] = {.lex_state = 542},
  [7329] = {.lex_state = 542},
  [7330] = {.lex_state = 271},
  [7331] = {.lex_state = 271, .external_lex_state = 114},
  [7332] = {.lex_state = 271, .external_lex_state = 114},
  [7333] = {.lex_state = 542},
  [7334] = {.lex_state = 542},
  [7335] = {.lex_state = 542},
  [7336] = {.lex_state = 542},
  [7337] = {.lex_state = 279},
  [7338] = {.lex_state = 542},
  [7339] = {.lex_state = 542},
  [7340] = {.lex_state = 542},
  [7341] = {.lex_state = 542},
  [7342] = {.lex_state = 542},
  [7343] = {.lex_state = 279},
  [7344] = {.lex_state = 542},
  [7345] = {.lex_state = 271, .external_lex_state = 114},
  [7346] = {.lex_state = 279},
  [7347] = {.lex_state = 542},
  [7348] = {.lex_state = 271, .external_lex_state = 114},
  [7349] = {.lex_state = 542, .external_lex_state = 130},
  [7350] = {.lex_state = 542},
  [7351] = {.lex_state = 279},
  [7352] = {.lex_state = 542},
  [7353] = {.lex_state = 542},
  [7354] = {.lex_state = 542},
  [7355] = {.lex_state = 542},
  [7356] = {.lex_state = 542},
  [7357] = {.lex_state = 542},
  [7358] = {.lex_state = 542},
  [7359] = {.lex_state = 542},
  [7360] = {.lex_state = 542},
  [7361] = {.lex_state = 271, .external_lex_state = 114},
  [7362] = {.lex_state = 257, .external_lex_state = 129},
  [7363] = {.lex_state = 257, .external_lex_state = 129},
  [7364] = {.lex_state = 542},
  [7365] = {.lex_state = 279},
  [7366] = {.lex_state = 542},
  [7367] = {.lex_state = 542},
  [7368] = {.lex_state = 542},
  [7369] = {.lex_state = 542},
  [7370] = {.lex_state = 542},
  [7371] = {.lex_state = 542},
  [7372] = {.lex_state = 542},
  [7373] = {.lex_state = 271, .external_lex_state = 114},
  [7374] = {.lex_state = 271, .external_lex_state = 114},
  [7375] = {.lex_state = 542},
  [7376] = {.lex_state = 279},
  [7377] = {.lex_state = 271, .external_lex_state = 114},
  [7378] = {.lex_state = 542},
  [7379] = {.lex_state = 542},
  [7380] = {.lex_state = 271, .external_lex_state = 114},
  [7381] = {.lex_state = 279},
  [7382] = {.lex_state = 542},
  [7383] = {.lex_state = 542},
  [7384] = {.lex_state = 542},
  [7385] = {.lex_state = 542},
  [7386] = {.lex_state = 271, .external_lex_state = 114},
  [7387] = {.lex_state = 279},
  [7388] = {.lex_state = 542},
  [7389] = {.lex_state = 273},
  [7390] = {.lex_state = 542},
  [7391] = {.lex_state = 542},
  [7392] = {.lex_state = 324},
  [7393] = {.lex_state = 271},
  [7394] = {.lex_state = 542},
  [7395] = {.lex_state = 271, .external_lex_state = 114},
  [7396] = {.lex_state = 542},
  [7397] = {.lex_state = 542, .external_lex_state = 130},
  [7398] = {.lex_state = 271, .external_lex_state = 114},
  [7399] = {.lex_state = 542},
  [7400] = {.lex_state = 542},
  [7401] = {.lex_state = 542, .external_lex_state = 130},
  [7402] = {.lex_state = 271, .external_lex_state = 126},
  [7403] = {.lex_state = 542},
  [7404] = {.lex_state = 279},
  [7405] = {.lex_state = 542},
  [7406] = {.lex_state = 279},
  [7407] = {.lex_state = 542},
  [7408] = {.lex_state = 246},
  [7409] = {.lex_state = 542},
  [7410] = {.lex_state = 542},
  [7411] = {.lex_state = 542},
  [7412] = {.lex_state = 271, .external_lex_state = 114},
  [7413] = {.lex_state = 271, .external_lex_state = 126},
  [7414] = {.lex_state = 542},
  [7415] = {.lex_state = 279},
  [7416] = {.lex_state = 324},
  [7417] = {.lex_state = 324},
  [7418] = {.lex_state = 542},
  [7419] = {.lex_state = 542},
  [7420] = {.lex_state = 542},
  [7421] = {.lex_state = 542},
  [7422] = {.lex_state = 271, .external_lex_state = 114},
  [7423] = {.lex_state = 542},
  [7424] = {.lex_state = 542},
  [7425] = {.lex_state = 279},
  [7426] = {.lex_state = 542},
  [7427] = {.lex_state = 542},
  [7428] = {.lex_state = 542},
  [7429] = {.lex_state = 542},
  [7430] = {.lex_state = 542},
  [7431] = {.lex_state = 542},
  [7432] = {.lex_state = 542},
  [7433] = {.lex_state = 271, .external_lex_state = 114},
  [7434] = {.lex_state = 542},
  [7435] = {.lex_state = 279},
  [7436] = {.lex_state = 542},
  [7437] = {.lex_state = 271, .external_lex_state = 114},
  [7438] = {.lex_state = 542},
  [7439] = {.lex_state = 271},
  [7440] = {.lex_state = 271, .external_lex_state = 126},
  [7441] = {.lex_state = 542},
  [7442] = {.lex_state = 279},
  [7443] = {.lex_state = 279},
  [7444] = {.lex_state = 246},
  [7445] = {.lex_state = 542},
  [7446] = {.lex_state = 542},
  [7447] = {.lex_state = 542},
  [7448] = {.lex_state = 542},
  [7449] = {.lex_state = 271, .external_lex_state = 114},
  [7450] = {.lex_state = 279},
  [7451] = {.lex_state = 271, .external_lex_state = 126},
  [7452] = {.lex_state = 279},
  [7453] = {.lex_state = 279},
  [7454] = {.lex_state = 542},
  [7455] = {.lex_state = 542},
  [7456] = {.lex_state = 542},
  [7457] = {.lex_state = 542},
  [7458] = {.lex_state = 542},
  [7459] = {.lex_state = 542},
  [7460] = {.lex_state = 542},
  [7461] = {.lex_state = 271, .external_lex_state = 114},
  [7462] = {.lex_state = 542},
  [7463] = {.lex_state = 542},
  [7464] = {.lex_state = 271, .external_lex_state = 114},
  [7465] = {.lex_state = 542, .external_lex_state = 128},
  [7466] = {.lex_state = 542, .external_lex_state = 128},
  [7467] = {.lex_state = 542},
  [7468] = {.lex_state = 279},
  [7469] = {.lex_state = 542},
  [7470] = {.lex_state = 542},
  [7471] = {.lex_state = 542},
  [7472] = {.lex_state = 542},
  [7473] = {.lex_state = 542},
  [7474] = {.lex_state = 542},
  [7475] = {.lex_state = 542},
  [7476] = {.lex_state = 271, .external_lex_state = 114},
  [7477] = {.lex_state = 271, .external_lex_state = 114},
  [7478] = {.lex_state = 324},
  [7479] = {.lex_state = 273},
  [7480] = {.lex_state = 324},
  [7481] = {.lex_state = 279},
  [7482] = {.lex_state = 324},
  [7483] = {.lex_state = 271},
  [7484] = {.lex_state = 542},
  [7485] = {.lex_state = 542},
  [7486] = {.lex_state = 542},
  [7487] = {.lex_state = 542, .external_lex_state = 130},
  [7488] = {.lex_state = 542},
  [7489] = {.lex_state = 271, .external_lex_state = 126},
  [7490] = {.lex_state = 279},
  [7491] = {.lex_state = 542, .external_lex_state = 130},
  [7492] = {.lex_state = 542},
  [7493] = {.lex_state = 271, .external_lex_state = 114},
  [7494] = {.lex_state = 271, .external_lex_state = 126},
  [7495] = {.lex_state = 271, .external_lex_state = 126},
  [7496] = {.lex_state = 279},
  [7497] = {.lex_state = 542},
  [7498] = {.lex_state = 542},
  [7499] = {.lex_state = 542},
  [7500] = {.lex_state = 542},
  [7501] = {.lex_state = 271, .external_lex_state = 114},
  [7502] = {.lex_state = 542},
  [7503] = {.lex_state = 271, .external_lex_state = 126},
  [7504] = {.lex_state = 271, .external_lex_state = 126},
  [7505] = {.lex_state = 279},
  [7506] = {.lex_state = 542},
  [7507] = {.lex_state = 542},
  [7508] = {.lex_state = 542},
  [7509] = {.lex_state = 542},
  [7510] = {.lex_state = 271, .external_lex_state = 114},
  [7511] = {.lex_state = 324},
  [7512] = {.lex_state = 271},
  [7513] = {.lex_state = 542},
  [7514] = {.lex_state = 542, .external_lex_state = 130},
  [7515] = {.lex_state = 542},
  [7516] = {.lex_state = 324},
  [7517] = {.lex_state = 542, .external_lex_state = 130},
  [7518] = {.lex_state = 542},
  [7519] = {.lex_state = 542, .external_lex_state = 128},
  [7520] = {.lex_state = 542},
  [7521] = {.lex_state = 271, .external_lex_state = 114},
  [7522] = {.lex_state = 542},
  [7523] = {.lex_state = 542, .external_lex_state = 128},
  [7524] = {.lex_state = 271, .external_lex_state = 114},
  [7525] = {.lex_state = 279},
  [7526] = {.lex_state = 542},
  [7527] = {.lex_state = 542},
  [7528] = {.lex_state = 271, .external_lex_state = 126},
  [7529] = {.lex_state = 542},
  [7530] = {.lex_state = 271, .external_lex_state = 114},
  [7531] = {.lex_state = 542},
  [7532] = {.lex_state = 279},
  [7533] = {.lex_state = 271, .external_lex_state = 114},
  [7534] = {.lex_state = 279},
  [7535] = {.lex_state = 542},
  [7536] = {.lex_state = 542},
  [7537] = {.lex_state = 271},
  [7538] = {.lex_state = 271},
  [7539] = {.lex_state = 271},
  [7540] = {.lex_state = 271},
  [7541] = {.lex_state = 271},
  [7542] = {.lex_state = 271},
  [7543] = {.lex_state = 271},
  [7544] = {.lex_state = 271},
  [7545] = {.lex_state = 271},
  [7546] = {.lex_state = 271},
  [7547] = {.lex_state = 271},
  [7548] = {.lex_state = 271},
  [7549] = {.lex_state = 271},
  [7550] = {.lex_state = 271},
  [7551] = {.lex_state = 271},
  [7552] = {.lex_state = 271},
  [7553] = {.lex_state = 271},
  [7554] = {.lex_state = 271},
  [7555] = {.lex_state = 271},
  [7556] = {.lex_state = 271},
  [7557] = {.lex_state = 271},
  [7558] = {.lex_state = 271},
  [7559] = {.lex_state = 271},
  [7560] = {.lex_state = 271},
  [7561] = {.lex_state = 271},
  [7562] = {.lex_state = 271},
  [7563] = {.lex_state = 271},
  [7564] = {.lex_state = 271},
  [7565] = {.lex_state = 271},
  [7566] = {.lex_state = 271},
  [7567] = {.lex_state = 271},
  [7568] = {.lex_state = 271},
  [7569] = {.lex_state = 271},
  [7570] = {.lex_state = 271},
  [7571] = {.lex_state = 271},
  [7572] = {.lex_state = 271},
  [7573] = {.lex_state = 271},
  [7574] = {.lex_state = 271},
  [7575] = {.lex_state = 271},
  [7576] = {.lex_state = 271},
  [7577] = {.lex_state = 271},
  [7578] = {.lex_state = 271},
  [7579] = {.lex_state = 271},
  [7580] = {.lex_state = 271},
  [7581] = {.lex_state = 271},
  [7582] = {.lex_state = 271},
  [7583] = {.lex_state = 271},
  [7584] = {.lex_state = 271},
  [7585] = {.lex_state = 271},
  [7586] = {.lex_state = 271},
  [7587] = {.lex_state = 271},
  [7588] = {.lex_state = 271},
  [7589] = {.lex_state = 271},
  [7590] = {.lex_state = 271},
  [7591] = {.lex_state = 271},
  [7592] = {.lex_state = 271},
  [7593] = {.lex_state = 271},
  [7594] = {.lex_state = 271},
  [7595] = {.lex_state = 271},
  [7596] = {.lex_state = 271},
  [7597] = {.lex_state = 271},
  [7598] = {.lex_state = 542},
  [7599] = {.lex_state = 542},
  [7600] = {.lex_state = 542},
  [7601] = {.lex_state = 271, .external_lex_state = 114},
  [7602] = {.lex_state = 542},
  [7603] = {.lex_state = 271, .external_lex_state = 126},
  [7604] = {.lex_state = 273},
  [7605] = {.lex_state = 271},
  [7606] = {.lex_state = 542},
  [7607] = {.lex_state = 271},
  [7608] = {.lex_state = 271},
  [7609] = {.lex_state = 271},
  [7610] = {.lex_state = 271},
  [7611] = {.lex_state = 271},
  [7612] = {.lex_state = 271},
  [7613] = {.lex_state = 271},
  [7614] = {.lex_state = 271},
  [7615] = {.lex_state = 271},
  [7616] = {.lex_state = 271},
  [7617] = {.lex_state = 271},
  [7618] = {.lex_state = 271},
  [7619] = {.lex_state = 271},
  [7620] = {.lex_state = 271},
  [7621] = {.lex_state = 271},
  [7622] = {.lex_state = 271},
  [7623] = {.lex_state = 271},
  [7624] = {.lex_state = 271},
  [7625] = {.lex_state = 271},
  [7626] = {.lex_state = 271},
  [7627] = {.lex_state = 271},
  [7628] = {.lex_state = 271},
  [7629] = {.lex_state = 271},
  [7630] = {.lex_state = 271},
  [7631] = {.lex_state = 271},
  [7632] = {.lex_state = 271},
  [7633] = {.lex_state = 271},
  [7634] = {.lex_state = 271},
  [7635] = {.lex_state = 271},
  [7636] = {.lex_state = 271},
  [7637] = {.lex_state = 271},
  [7638] = {.lex_state = 271},
  [7639] = {.lex_state = 271},
  [7640] = {.lex_state = 271},
  [7641] = {.lex_state = 271},
  [7642] = {.lex_state = 271},
  [7643] = {.lex_state = 271},
  [7644] = {.lex_state = 271},
  [7645] = {.lex_state = 271},
  [7646] = {.lex_state = 271},
  [7647] = {.lex_state = 271},
  [7648] = {.lex_state = 271},
  [7649] = {.lex_state = 271},
  [7650] = {.lex_state = 271},
  [7651] = {.lex_state = 271},
  [7652] = {.lex_state = 271},
  [7653] = {.lex_state = 271},
  [7654] = {.lex_state = 271},
  [7655] = {.lex_state = 271},
  [7656] = {.lex_state = 271},
  [7657] = {.lex_state = 271},
  [7658] = {.lex_state = 271},
  [7659] = {.lex_state = 271},
  [7660] = {.lex_state = 271},
  [7661] = {.lex_state = 271},
  [7662] = {.lex_state = 271},
  [7663] = {.lex_state = 271},
  [7664] = {.lex_state = 271},
  [7665] = {.lex_state = 271},
  [7666] = {.lex_state = 271},
  [7667] = {.lex_state = 271},
  [7668] = {.lex_state = 271},
  [7669] = {.lex_state = 271},
  [7670] = {.lex_state = 271, .external_lex_state = 114},
};

static const uint16_t ts_parse_table[LARGE_STATE_COUNT][SYMBOL_COUNT] = {
  [0] = {
    [ts_builtin_sym_end] = ACTIONS(1),
    [sym_word] = ACTIONS(1),
    [anon_sym_for] = ACTIONS(1),
    [anon_sym_select] = ACTIONS(1),
    [anon_sym_in] = ACTIONS(1),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(1),
    [anon_sym_SEMI] = ACTIONS(1),
    [anon_sym_COMMA] = ACTIONS(1),
    [anon_sym_EQ] = ACTIONS(1),
    [anon_sym_STAR_EQ] = ACTIONS(1),
    [anon_sym_SLASH_EQ] = ACTIONS(1),
    [anon_sym_PERCENT_EQ] = ACTIONS(1),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(1),
    [anon_sym_LT_LT_EQ] = ACTIONS(1),
    [anon_sym_GT_GT_EQ] = ACTIONS(1),
    [anon_sym_AMP_EQ] = ACTIONS(1),
    [anon_sym_CARET_EQ] = ACTIONS(1),
    [anon_sym_PIPE_EQ] = ACTIONS(1),
    [anon_sym_PIPE_PIPE] = ACTIONS(1),
    [anon_sym_AMP_AMP] = ACTIONS(1),
    [anon_sym_PIPE] = ACTIONS(1),
    [anon_sym_CARET] = ACTIONS(1),
    [anon_sym_AMP] = ACTIONS(1),
    [anon_sym_BANG_EQ] = ACTIONS(1),
    [anon_sym_LT] = ACTIONS(1),
    [anon_sym_GT] = ACTIONS(1),
    [anon_sym_LT_EQ] = ACTIONS(1),
    [anon_sym_GT_EQ] = ACTIONS(1),
    [anon_sym_LT_LT] = ACTIONS(1),
    [anon_sym_GT_GT] = ACTIONS(1),
    [anon_sym_PLUS] = ACTIONS(1),
    [anon_sym_DASH] = ACTIONS(1),
    [anon_sym_STAR] = ACTIONS(1),
    [anon_sym_SLASH] = ACTIONS(1),
    [anon_sym_PERCENT] = ACTIONS(1),
    [anon_sym_STAR_STAR] = ACTIONS(1),
    [anon_sym_LPAREN] = ACTIONS(1),
    [anon_sym_RPAREN] = ACTIONS(1),
    [anon_sym_while] = ACTIONS(1),
    [anon_sym_until] = ACTIONS(1),
    [anon_sym_do] = ACTIONS(1),
    [anon_sym_done] = ACTIONS(1),
    [anon_sym_if] = ACTIONS(1),
    [anon_sym_then] = ACTIONS(1),
    [anon_sym_fi] = ACTIONS(1),
    [anon_sym_elif] = ACTIONS(1),
    [anon_sym_else] = ACTIONS(1),
    [anon_sym_case] = ACTIONS(1),
    [anon_sym_esac] = ACTIONS(1),
    [anon_sym_SEMI_SEMI] = ACTIONS(1),
    [anon_sym_SEMI_AMP] = ACTIONS(1),
    [anon_sym_SEMI_SEMI_AMP] = ACTIONS(1),
    [anon_sym_function] = ACTIONS(1),
    [anon_sym_LBRACE] = ACTIONS(1),
    [anon_sym_RBRACE] = ACTIONS(1),
    [anon_sym_PIPE_AMP] = ACTIONS(1),
    [anon_sym_BANG] = ACTIONS(1),
    [anon_sym_LBRACK] = ACTIONS(1),
    [anon_sym_RBRACK] = ACTIONS(1),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1),
    [anon_sym_declare] = ACTIONS(1),
    [anon_sym_typeset] = ACTIONS(1),
    [anon_sym_export] = ACTIONS(1),
    [anon_sym_readonly] = ACTIONS(1),
    [anon_sym_local] = ACTIONS(1),
    [anon_sym_unset] = ACTIONS(1),
    [anon_sym_unsetenv] = ACTIONS(1),
    [anon_sym_AMP_GT] = ACTIONS(1),
    [anon_sym_AMP_GT_GT] = ACTIONS(1),
    [anon_sym_LT_AMP] = ACTIONS(1),
    [anon_sym_GT_AMP] = ACTIONS(1),
    [anon_sym_GT_PIPE] = ACTIONS(1),
    [anon_sym_LT_AMP_DASH] = ACTIONS(1),
    [anon_sym_GT_AMP_DASH] = ACTIONS(1),
    [anon_sym_LT_LT_DASH] = ACTIONS(1),
    [anon_sym_LT_LT_LT] = ACTIONS(1),
    [anon_sym_QMARK] = ACTIONS(1),
    [anon_sym_COLON] = ACTIONS(1),
    [anon_sym_DASH2] = ACTIONS(1),
    [anon_sym_PLUS2] = ACTIONS(1),
    [anon_sym_TILDE] = ACTIONS(1),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(1),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(1),
    [anon_sym_DOT_DOT] = ACTIONS(1),
    [anon_sym_RBRACE2] = ACTIONS(1),
    [aux_sym_concatenation_token1] = ACTIONS(1),
    [anon_sym_DOLLAR] = ACTIONS(1),
    [sym__special_character] = ACTIONS(1),
    [anon_sym_DQUOTE] = ACTIONS(1),
    [sym_raw_string] = ACTIONS(1),
    [sym_ansi_c_string] = ACTIONS(1),
    [anon_sym_POUND] = ACTIONS(1),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(1),
    [anon_sym_RBRACE3] = ACTIONS(1),
    [anon_sym_BANG2] = ACTIONS(1),
    [anon_sym_AT] = ACTIONS(1),
    [anon_sym_STAR2] = ACTIONS(1),
    [anon_sym_POUND2] = ACTIONS(1),
    [anon_sym_EQ2] = ACTIONS(1),
    [anon_sym_COLON_EQ] = ACTIONS(1),
    [anon_sym_DASH3] = ACTIONS(1),
    [anon_sym_COLON_DASH] = ACTIONS(1),
    [anon_sym_PLUS3] = ACTIONS(1),
    [anon_sym_COLON_PLUS] = ACTIONS(1),
    [anon_sym_QMARK2] = ACTIONS(1),
    [anon_sym_COLON_QMARK] = ACTIONS(1),
    [anon_sym_PERCENT_PERCENT] = ACTIONS(1),
    [anon_sym_SLASH_SLASH] = ACTIONS(1),
    [anon_sym_SLASH_POUND] = ACTIONS(1),
    [anon_sym_SLASH_PERCENT] = ACTIONS(1),
    [anon_sym_COMMA_COMMA] = ACTIONS(1),
    [anon_sym_CARET_CARET] = ACTIONS(1),
    [anon_sym_U] = ACTIONS(1),
    [anon_sym_u] = ACTIONS(1),
    [anon_sym_L] = ACTIONS(1),
    [anon_sym_Q] = ACTIONS(1),
    [anon_sym_E] = ACTIONS(1),
    [anon_sym_P] = ACTIONS(1),
    [anon_sym_A] = ACTIONS(1),
    [anon_sym_K] = ACTIONS(1),
    [anon_sym_a] = ACTIONS(1),
    [anon_sym_k] = ACTIONS(1),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1),
    [anon_sym_BQUOTE] = ACTIONS(1),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(1),
    [anon_sym_LT_LPAREN] = ACTIONS(1),
    [anon_sym_GT_LPAREN] = ACTIONS(1),
    [sym_comment] = ACTIONS(3),
    [sym__comment_word] = ACTIONS(1),
    [anon_sym_AT2] = ACTIONS(1),
    [anon_sym_0] = ACTIONS(1),
    [anon_sym__] = ACTIONS(1),
    [sym_heredoc_start] = ACTIONS(1),
    [sym_simple_heredoc_body] = ACTIONS(1),
    [sym__heredoc_body_beginning] = ACTIONS(1),
    [sym_heredoc_content] = ACTIONS(1),
    [sym_heredoc_end] = ACTIONS(1),
    [sym_file_descriptor] = ACTIONS(1),
    [sym__empty_value] = ACTIONS(1),
    [sym__concat] = ACTIONS(1),
    [sym_variable_name] = ACTIONS(1),
    [sym_test_operator] = ACTIONS(1),
    [sym_regex] = ACTIONS(1),
    [sym__regex_no_slash] = ACTIONS(1),
    [sym__regex_no_space] = ACTIONS(1),
    [sym__expansion_word] = ACTIONS(1),
    [sym_extglob_pattern] = ACTIONS(1),
    [sym__bare_dollar] = ACTIONS(1),
    [sym__brace_start] = ACTIONS(1),
    [sym__immediate_double_hash] = ACTIONS(1),
    [sym__external_expansion_sym_hash] = ACTIONS(1),
    [sym__external_expansion_sym_bang] = ACTIONS(1),
    [sym__external_expansion_sym_equal] = ACTIONS(1),
    [sym___error_recovery] = ACTIONS(1),
  },
  [1] = {
    [sym_program] = STATE(7470),
    [sym__statements] = STATE(7469),
    [sym__statement_not_pipeline] = STATE(6572),
    [sym_redirected_statement] = STATE(4860),
    [sym_for_statement] = STATE(4860),
    [sym_c_style_for_statement] = STATE(4860),
    [sym_while_statement] = STATE(4321),
    [sym_if_statement] = STATE(4321),
    [sym_case_statement] = STATE(4860),
    [sym_function_definition] = STATE(4860),
    [sym_compound_statement] = STATE(4860),
    [sym_subshell] = STATE(4860),
    [sym_pipeline] = STATE(5079),
    [sym_list] = STATE(4860),
    [sym_negated_command] = STATE(4860),
    [sym_test_command] = STATE(4860),
    [sym_declaration_command] = STATE(4860),
    [sym_unset_command] = STATE(4860),
    [sym_command] = STATE(4860),
    [sym_command_name] = STATE(534),
    [sym_variable_assignment] = STATE(913),
    [sym_variable_assignments] = STATE(4860),
    [sym_subscript] = STATE(6774),
    [sym_file_redirect] = STATE(2055),
    [sym_herestring_redirect] = STATE(2156),
    [sym_arithmetic_expansion] = STATE(957),
    [sym_brace_expression] = STATE(957),
    [sym_concatenation] = STATE(1386),
    [sym_string] = STATE(957),
    [sym_translated_string] = STATE(957),
    [sym_number] = STATE(957),
    [sym_simple_expansion] = STATE(957),
    [sym_expansion] = STATE(957),
    [sym_command_substitution] = STATE(957),
    [sym_process_substitution] = STATE(957),
    [aux_sym__statements_repeat1] = STATE(427),
    [aux_sym_redirected_statement_repeat2] = STATE(4571),
    [aux_sym_command_repeat1] = STATE(937),
    [aux_sym__literal_repeat1] = STATE(1212),
    [ts_builtin_sym_end] = ACTIONS(5),
    [sym_word] = ACTIONS(7),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(13),
    [anon_sym_LT] = ACTIONS(15),
    [anon_sym_GT] = ACTIONS(15),
    [anon_sym_GT_GT] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(31),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(37),
    [anon_sym_typeset] = ACTIONS(37),
    [anon_sym_export] = ACTIONS(37),
    [anon_sym_readonly] = ACTIONS(37),
    [anon_sym_local] = ACTIONS(37),
    [anon_sym_unset] = ACTIONS(39),
    [anon_sym_unsetenv] = ACTIONS(39),
    [anon_sym_AMP_GT] = ACTIONS(15),
    [anon_sym_AMP_GT_GT] = ACTIONS(17),
    [anon_sym_LT_AMP] = ACTIONS(15),
    [anon_sym_GT_AMP] = ACTIONS(15),
    [anon_sym_GT_PIPE] = ACTIONS(17),
    [anon_sym_LT_AMP_DASH] = ACTIONS(41),
    [anon_sym_GT_AMP_DASH] = ACTIONS(41),
    [anon_sym_LT_LT_LT] = ACTIONS(43),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(45),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(47),
    [anon_sym_DOLLAR] = ACTIONS(49),
    [sym__special_character] = ACTIONS(51),
    [anon_sym_DQUOTE] = ACTIONS(53),
    [sym_raw_string] = ACTIONS(55),
    [sym_ansi_c_string] = ACTIONS(55),
    [aux_sym_number_token1] = ACTIONS(57),
    [aux_sym_number_token2] = ACTIONS(59),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(61),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(63),
    [anon_sym_BQUOTE] = ACTIONS(65),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(67),
    [anon_sym_LT_LPAREN] = ACTIONS(69),
    [anon_sym_GT_LPAREN] = ACTIONS(69),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(73),
    [sym_variable_name] = ACTIONS(75),
    [sym_test_operator] = ACTIONS(77),
    [sym__brace_start] = ACTIONS(79),
  },
  [2] = {
    [sym__statements] = STATE(7427),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym__expression] = STATE(3126),
    [sym_binary_expression] = STATE(3053),
    [sym_ternary_expression] = STATE(3053),
    [sym_unary_expression] = STATE(3053),
    [sym_postfix_expression] = STATE(3053),
    [sym_parenthesized_expression] = STATE(3053),
    [sym_arithmetic_expansion] = STATE(472),
    [sym_brace_expression] = STATE(472),
    [sym_concatenation] = STATE(502),
    [sym_string] = STATE(472),
    [sym_translated_string] = STATE(472),
    [sym_number] = STATE(472),
    [sym_simple_expansion] = STATE(472),
    [sym_expansion] = STATE(472),
    [sym_command_substitution] = STATE(472),
    [sym_process_substitution] = STATE(472),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(493),
    [sym_word] = ACTIONS(81),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(89),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_PLUS_PLUS2] = ACTIONS(103),
    [anon_sym_DASH_DASH2] = ACTIONS(103),
    [anon_sym_DASH2] = ACTIONS(105),
    [anon_sym_PLUS2] = ACTIONS(105),
    [anon_sym_TILDE] = ACTIONS(107),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(109),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(111),
    [anon_sym_DOLLAR] = ACTIONS(113),
    [sym__special_character] = ACTIONS(115),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [sym_raw_string] = ACTIONS(119),
    [sym_ansi_c_string] = ACTIONS(119),
    [aux_sym_number_token1] = ACTIONS(121),
    [aux_sym_number_token2] = ACTIONS(123),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(125),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(127),
    [anon_sym_BQUOTE] = ACTIONS(129),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(131),
    [anon_sym_LT_LPAREN] = ACTIONS(133),
    [anon_sym_GT_LPAREN] = ACTIONS(133),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(139),
    [sym__brace_start] = ACTIONS(141),
  },
  [3] = {
    [sym__statements] = STATE(6914),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym__expression] = STATE(3042),
    [sym_binary_expression] = STATE(3053),
    [sym_ternary_expression] = STATE(3053),
    [sym_unary_expression] = STATE(3053),
    [sym_postfix_expression] = STATE(3053),
    [sym_parenthesized_expression] = STATE(3053),
    [sym_arithmetic_expansion] = STATE(472),
    [sym_brace_expression] = STATE(472),
    [sym_concatenation] = STATE(502),
    [sym_string] = STATE(472),
    [sym_translated_string] = STATE(472),
    [sym_number] = STATE(472),
    [sym_simple_expansion] = STATE(472),
    [sym_expansion] = STATE(472),
    [sym_command_substitution] = STATE(472),
    [sym_process_substitution] = STATE(472),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(493),
    [sym_word] = ACTIONS(81),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(89),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_PLUS_PLUS2] = ACTIONS(103),
    [anon_sym_DASH_DASH2] = ACTIONS(103),
    [anon_sym_DASH2] = ACTIONS(105),
    [anon_sym_PLUS2] = ACTIONS(105),
    [anon_sym_TILDE] = ACTIONS(107),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(109),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(111),
    [anon_sym_DOLLAR] = ACTIONS(113),
    [sym__special_character] = ACTIONS(115),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [sym_raw_string] = ACTIONS(119),
    [sym_ansi_c_string] = ACTIONS(119),
    [aux_sym_number_token1] = ACTIONS(121),
    [aux_sym_number_token2] = ACTIONS(123),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(125),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(127),
    [anon_sym_BQUOTE] = ACTIONS(129),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(131),
    [anon_sym_LT_LPAREN] = ACTIONS(133),
    [anon_sym_GT_LPAREN] = ACTIONS(133),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(139),
    [sym__brace_start] = ACTIONS(141),
  },
  [4] = {
    [sym__statement_not_pipeline] = STATE(6568),
    [sym_redirected_statement] = STATE(5394),
    [sym_for_statement] = STATE(5406),
    [sym_c_style_for_statement] = STATE(5406),
    [sym_while_statement] = STATE(5195),
    [sym_if_statement] = STATE(5195),
    [sym_case_statement] = STATE(5406),
    [sym_function_definition] = STATE(5406),
    [sym_compound_statement] = STATE(5406),
    [sym_subshell] = STATE(5406),
    [sym_pipeline] = STATE(5569),
    [sym_list] = STATE(5406),
    [sym_negated_command] = STATE(5406),
    [sym_test_command] = STATE(5406),
    [sym_declaration_command] = STATE(5406),
    [sym_unset_command] = STATE(5406),
    [sym_command] = STATE(5406),
    [sym_command_name] = STATE(620),
    [sym_variable_assignment] = STATE(1811),
    [sym_variable_assignments] = STATE(5406),
    [sym_subscript] = STATE(6756),
    [sym_file_redirect] = STATE(2693),
    [sym_herestring_redirect] = STATE(2691),
    [sym__expression] = STATE(3131),
    [sym_binary_expression] = STATE(3071),
    [sym_ternary_expression] = STATE(3071),
    [sym_unary_expression] = STATE(3071),
    [sym_postfix_expression] = STATE(3071),
    [sym_parenthesized_expression] = STATE(3071),
    [sym_arithmetic_expansion] = STATE(496),
    [sym_brace_expression] = STATE(496),
    [sym_concatenation] = STATE(524),
    [sym_string] = STATE(496),
    [sym_translated_string] = STATE(496),
    [sym_number] = STATE(496),
    [sym_simple_expansion] = STATE(496),
    [sym_expansion] = STATE(496),
    [sym_command_substitution] = STATE(496),
    [sym_process_substitution] = STATE(496),
    [aux_sym_redirected_statement_repeat2] = STATE(5167),
    [aux_sym_command_repeat1] = STATE(955),
    [aux_sym__literal_repeat1] = STATE(522),
    [sym_word] = ACTIONS(143),
    [anon_sym_for] = ACTIONS(145),
    [anon_sym_select] = ACTIONS(147),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(149),
    [anon_sym_LT] = ACTIONS(151),
    [anon_sym_GT] = ACTIONS(151),
    [anon_sym_GT_GT] = ACTIONS(153),
    [anon_sym_LPAREN] = ACTIONS(155),
    [anon_sym_while] = ACTIONS(157),
    [anon_sym_until] = ACTIONS(157),
    [anon_sym_if] = ACTIONS(159),
    [anon_sym_case] = ACTIONS(161),
    [anon_sym_function] = ACTIONS(163),
    [anon_sym_LBRACE] = ACTIONS(165),
    [anon_sym_BANG] = ACTIONS(167),
    [anon_sym_LBRACK] = ACTIONS(169),
    [anon_sym_RBRACK] = ACTIONS(171),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(173),
    [anon_sym_declare] = ACTIONS(175),
    [anon_sym_typeset] = ACTIONS(175),
    [anon_sym_export] = ACTIONS(175),
    [anon_sym_readonly] = ACTIONS(175),
    [anon_sym_local] = ACTIONS(175),
    [anon_sym_unset] = ACTIONS(177),
    [anon_sym_unsetenv] = ACTIONS(177),
    [anon_sym_AMP_GT] = ACTIONS(151),
    [anon_sym_AMP_GT_GT] = ACTIONS(153),
    [anon_sym_LT_AMP] = ACTIONS(151),
    [anon_sym_GT_AMP] = ACTIONS(151),
    [anon_sym_GT_PIPE] = ACTIONS(153),
    [anon_sym_LT_AMP_DASH] = ACTIONS(179),
    [anon_sym_GT_AMP_DASH] = ACTIONS(179),
    [anon_sym_LT_LT_LT] = ACTIONS(181),
    [anon_sym_PLUS_PLUS2] = ACTIONS(183),
    [anon_sym_DASH_DASH2] = ACTIONS(183),
    [anon_sym_DASH2] = ACTIONS(185),
    [anon_sym_PLUS2] = ACTIONS(185),
    [anon_sym_TILDE] = ACTIONS(187),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(189),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(191),
    [anon_sym_DOLLAR] = ACTIONS(193),
    [sym__special_character] = ACTIONS(195),
    [anon_sym_DQUOTE] = ACTIONS(197),
    [sym_raw_string] = ACTIONS(199),
    [sym_ansi_c_string] = ACTIONS(199),
    [aux_sym_number_token1] = ACTIONS(201),
    [aux_sym_number_token2] = ACTIONS(203),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(205),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(207),
    [anon_sym_BQUOTE] = ACTIONS(209),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(211),
    [anon_sym_LT_LPAREN] = ACTIONS(213),
    [anon_sym_GT_LPAREN] = ACTIONS(213),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(215),
    [sym_variable_name] = ACTIONS(217),
    [sym_test_operator] = ACTIONS(219),
    [sym__brace_start] = ACTIONS(221),
  },
  [5] = {
    [sym__statement_not_pipeline] = STATE(6568),
    [sym_redirected_statement] = STATE(5266),
    [sym_for_statement] = STATE(5406),
    [sym_c_style_for_statement] = STATE(5406),
    [sym_while_statement] = STATE(5195),
    [sym_if_statement] = STATE(5195),
    [sym_case_statement] = STATE(5406),
    [sym_function_definition] = STATE(5406),
    [sym_compound_statement] = STATE(5406),
    [sym_subshell] = STATE(5406),
    [sym_pipeline] = STATE(5569),
    [sym_list] = STATE(5406),
    [sym_negated_command] = STATE(5406),
    [sym_test_command] = STATE(5406),
    [sym_declaration_command] = STATE(5406),
    [sym_unset_command] = STATE(5406),
    [sym_command] = STATE(5406),
    [sym_command_name] = STATE(620),
    [sym_variable_assignment] = STATE(1811),
    [sym_variable_assignments] = STATE(5406),
    [sym_subscript] = STATE(6756),
    [sym_file_redirect] = STATE(2693),
    [sym_herestring_redirect] = STATE(2691),
    [sym__expression] = STATE(3124),
    [sym_binary_expression] = STATE(3071),
    [sym_ternary_expression] = STATE(3071),
    [sym_unary_expression] = STATE(3071),
    [sym_postfix_expression] = STATE(3071),
    [sym_parenthesized_expression] = STATE(3071),
    [sym_arithmetic_expansion] = STATE(496),
    [sym_brace_expression] = STATE(496),
    [sym_concatenation] = STATE(524),
    [sym_string] = STATE(496),
    [sym_translated_string] = STATE(496),
    [sym_number] = STATE(496),
    [sym_simple_expansion] = STATE(496),
    [sym_expansion] = STATE(496),
    [sym_command_substitution] = STATE(496),
    [sym_process_substitution] = STATE(496),
    [aux_sym_redirected_statement_repeat2] = STATE(5167),
    [aux_sym_command_repeat1] = STATE(955),
    [aux_sym__literal_repeat1] = STATE(522),
    [sym_word] = ACTIONS(143),
    [anon_sym_for] = ACTIONS(145),
    [anon_sym_select] = ACTIONS(147),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(149),
    [anon_sym_LT] = ACTIONS(151),
    [anon_sym_GT] = ACTIONS(151),
    [anon_sym_GT_GT] = ACTIONS(153),
    [anon_sym_LPAREN] = ACTIONS(155),
    [anon_sym_while] = ACTIONS(157),
    [anon_sym_until] = ACTIONS(157),
    [anon_sym_if] = ACTIONS(159),
    [anon_sym_case] = ACTIONS(161),
    [anon_sym_function] = ACTIONS(163),
    [anon_sym_LBRACE] = ACTIONS(165),
    [anon_sym_BANG] = ACTIONS(167),
    [anon_sym_LBRACK] = ACTIONS(169),
    [anon_sym_RBRACK] = ACTIONS(223),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(173),
    [anon_sym_declare] = ACTIONS(175),
    [anon_sym_typeset] = ACTIONS(175),
    [anon_sym_export] = ACTIONS(175),
    [anon_sym_readonly] = ACTIONS(175),
    [anon_sym_local] = ACTIONS(175),
    [anon_sym_unset] = ACTIONS(177),
    [anon_sym_unsetenv] = ACTIONS(177),
    [anon_sym_AMP_GT] = ACTIONS(151),
    [anon_sym_AMP_GT_GT] = ACTIONS(153),
    [anon_sym_LT_AMP] = ACTIONS(151),
    [anon_sym_GT_AMP] = ACTIONS(151),
    [anon_sym_GT_PIPE] = ACTIONS(153),
    [anon_sym_LT_AMP_DASH] = ACTIONS(179),
    [anon_sym_GT_AMP_DASH] = ACTIONS(179),
    [anon_sym_LT_LT_LT] = ACTIONS(181),
    [anon_sym_PLUS_PLUS2] = ACTIONS(183),
    [anon_sym_DASH_DASH2] = ACTIONS(183),
    [anon_sym_DASH2] = ACTIONS(185),
    [anon_sym_PLUS2] = ACTIONS(185),
    [anon_sym_TILDE] = ACTIONS(187),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(189),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(191),
    [anon_sym_DOLLAR] = ACTIONS(193),
    [sym__special_character] = ACTIONS(195),
    [anon_sym_DQUOTE] = ACTIONS(197),
    [sym_raw_string] = ACTIONS(199),
    [sym_ansi_c_string] = ACTIONS(199),
    [aux_sym_number_token1] = ACTIONS(201),
    [aux_sym_number_token2] = ACTIONS(203),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(205),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(207),
    [anon_sym_BQUOTE] = ACTIONS(209),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(211),
    [anon_sym_LT_LPAREN] = ACTIONS(213),
    [anon_sym_GT_LPAREN] = ACTIONS(213),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(215),
    [sym_variable_name] = ACTIONS(217),
    [sym_test_operator] = ACTIONS(219),
    [sym__brace_start] = ACTIONS(221),
  },
  [6] = {
    [sym__statement_not_pipeline] = STATE(6568),
    [sym_redirected_statement] = STATE(5317),
    [sym_for_statement] = STATE(5406),
    [sym_c_style_for_statement] = STATE(5406),
    [sym_while_statement] = STATE(5195),
    [sym_if_statement] = STATE(5195),
    [sym_case_statement] = STATE(5406),
    [sym_function_definition] = STATE(5406),
    [sym_compound_statement] = STATE(5406),
    [sym_subshell] = STATE(5406),
    [sym_pipeline] = STATE(5569),
    [sym_list] = STATE(5406),
    [sym_negated_command] = STATE(5406),
    [sym_test_command] = STATE(5406),
    [sym_declaration_command] = STATE(5406),
    [sym_unset_command] = STATE(5406),
    [sym_command] = STATE(5406),
    [sym_command_name] = STATE(620),
    [sym_variable_assignment] = STATE(1811),
    [sym_variable_assignments] = STATE(5406),
    [sym_subscript] = STATE(6756),
    [sym_file_redirect] = STATE(2693),
    [sym_herestring_redirect] = STATE(2691),
    [sym__expression] = STATE(3145),
    [sym_binary_expression] = STATE(3071),
    [sym_ternary_expression] = STATE(3071),
    [sym_unary_expression] = STATE(3071),
    [sym_postfix_expression] = STATE(3071),
    [sym_parenthesized_expression] = STATE(3071),
    [sym_arithmetic_expansion] = STATE(496),
    [sym_brace_expression] = STATE(496),
    [sym_concatenation] = STATE(524),
    [sym_string] = STATE(496),
    [sym_translated_string] = STATE(496),
    [sym_number] = STATE(496),
    [sym_simple_expansion] = STATE(496),
    [sym_expansion] = STATE(496),
    [sym_command_substitution] = STATE(496),
    [sym_process_substitution] = STATE(496),
    [aux_sym_redirected_statement_repeat2] = STATE(5167),
    [aux_sym_command_repeat1] = STATE(955),
    [aux_sym__literal_repeat1] = STATE(522),
    [sym_word] = ACTIONS(143),
    [anon_sym_for] = ACTIONS(145),
    [anon_sym_select] = ACTIONS(147),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(149),
    [anon_sym_LT] = ACTIONS(151),
    [anon_sym_GT] = ACTIONS(151),
    [anon_sym_GT_GT] = ACTIONS(153),
    [anon_sym_LPAREN] = ACTIONS(155),
    [anon_sym_while] = ACTIONS(157),
    [anon_sym_until] = ACTIONS(157),
    [anon_sym_if] = ACTIONS(159),
    [anon_sym_case] = ACTIONS(161),
    [anon_sym_function] = ACTIONS(163),
    [anon_sym_LBRACE] = ACTIONS(165),
    [anon_sym_BANG] = ACTIONS(167),
    [anon_sym_LBRACK] = ACTIONS(169),
    [anon_sym_RBRACK] = ACTIONS(225),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(173),
    [anon_sym_declare] = ACTIONS(175),
    [anon_sym_typeset] = ACTIONS(175),
    [anon_sym_export] = ACTIONS(175),
    [anon_sym_readonly] = ACTIONS(175),
    [anon_sym_local] = ACTIONS(175),
    [anon_sym_unset] = ACTIONS(177),
    [anon_sym_unsetenv] = ACTIONS(177),
    [anon_sym_AMP_GT] = ACTIONS(151),
    [anon_sym_AMP_GT_GT] = ACTIONS(153),
    [anon_sym_LT_AMP] = ACTIONS(151),
    [anon_sym_GT_AMP] = ACTIONS(151),
    [anon_sym_GT_PIPE] = ACTIONS(153),
    [anon_sym_LT_AMP_DASH] = ACTIONS(179),
    [anon_sym_GT_AMP_DASH] = ACTIONS(179),
    [anon_sym_LT_LT_LT] = ACTIONS(181),
    [anon_sym_PLUS_PLUS2] = ACTIONS(183),
    [anon_sym_DASH_DASH2] = ACTIONS(183),
    [anon_sym_DASH2] = ACTIONS(185),
    [anon_sym_PLUS2] = ACTIONS(185),
    [anon_sym_TILDE] = ACTIONS(187),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(189),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(191),
    [anon_sym_DOLLAR] = ACTIONS(193),
    [sym__special_character] = ACTIONS(195),
    [anon_sym_DQUOTE] = ACTIONS(197),
    [sym_raw_string] = ACTIONS(199),
    [sym_ansi_c_string] = ACTIONS(199),
    [aux_sym_number_token1] = ACTIONS(201),
    [aux_sym_number_token2] = ACTIONS(203),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(205),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(207),
    [anon_sym_BQUOTE] = ACTIONS(209),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(211),
    [anon_sym_LT_LPAREN] = ACTIONS(213),
    [anon_sym_GT_LPAREN] = ACTIONS(213),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(215),
    [sym_variable_name] = ACTIONS(217),
    [sym_test_operator] = ACTIONS(219),
    [sym__brace_start] = ACTIONS(221),
  },
  [7] = {
    [sym__statement_not_pipeline] = STATE(6568),
    [sym_redirected_statement] = STATE(5361),
    [sym_for_statement] = STATE(5406),
    [sym_c_style_for_statement] = STATE(5406),
    [sym_while_statement] = STATE(5195),
    [sym_if_statement] = STATE(5195),
    [sym_case_statement] = STATE(5406),
    [sym_function_definition] = STATE(5406),
    [sym_compound_statement] = STATE(5406),
    [sym_subshell] = STATE(5406),
    [sym_pipeline] = STATE(5569),
    [sym_list] = STATE(5406),
    [sym_negated_command] = STATE(5406),
    [sym_test_command] = STATE(5406),
    [sym_declaration_command] = STATE(5406),
    [sym_unset_command] = STATE(5406),
    [sym_command] = STATE(5406),
    [sym_command_name] = STATE(620),
    [sym_variable_assignment] = STATE(1811),
    [sym_variable_assignments] = STATE(5406),
    [sym_subscript] = STATE(6756),
    [sym_file_redirect] = STATE(2693),
    [sym_herestring_redirect] = STATE(2691),
    [sym__expression] = STATE(3155),
    [sym_binary_expression] = STATE(3071),
    [sym_ternary_expression] = STATE(3071),
    [sym_unary_expression] = STATE(3071),
    [sym_postfix_expression] = STATE(3071),
    [sym_parenthesized_expression] = STATE(3071),
    [sym_arithmetic_expansion] = STATE(496),
    [sym_brace_expression] = STATE(496),
    [sym_concatenation] = STATE(524),
    [sym_string] = STATE(496),
    [sym_translated_string] = STATE(496),
    [sym_number] = STATE(496),
    [sym_simple_expansion] = STATE(496),
    [sym_expansion] = STATE(496),
    [sym_command_substitution] = STATE(496),
    [sym_process_substitution] = STATE(496),
    [aux_sym_redirected_statement_repeat2] = STATE(5167),
    [aux_sym_command_repeat1] = STATE(955),
    [aux_sym__literal_repeat1] = STATE(522),
    [sym_word] = ACTIONS(143),
    [anon_sym_for] = ACTIONS(145),
    [anon_sym_select] = ACTIONS(147),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(149),
    [anon_sym_LT] = ACTIONS(151),
    [anon_sym_GT] = ACTIONS(151),
    [anon_sym_GT_GT] = ACTIONS(153),
    [anon_sym_LPAREN] = ACTIONS(155),
    [anon_sym_while] = ACTIONS(157),
    [anon_sym_until] = ACTIONS(157),
    [anon_sym_if] = ACTIONS(159),
    [anon_sym_case] = ACTIONS(161),
    [anon_sym_function] = ACTIONS(163),
    [anon_sym_LBRACE] = ACTIONS(165),
    [anon_sym_BANG] = ACTIONS(167),
    [anon_sym_LBRACK] = ACTIONS(169),
    [anon_sym_RBRACK] = ACTIONS(227),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(173),
    [anon_sym_declare] = ACTIONS(175),
    [anon_sym_typeset] = ACTIONS(175),
    [anon_sym_export] = ACTIONS(175),
    [anon_sym_readonly] = ACTIONS(175),
    [anon_sym_local] = ACTIONS(175),
    [anon_sym_unset] = ACTIONS(177),
    [anon_sym_unsetenv] = ACTIONS(177),
    [anon_sym_AMP_GT] = ACTIONS(151),
    [anon_sym_AMP_GT_GT] = ACTIONS(153),
    [anon_sym_LT_AMP] = ACTIONS(151),
    [anon_sym_GT_AMP] = ACTIONS(151),
    [anon_sym_GT_PIPE] = ACTIONS(153),
    [anon_sym_LT_AMP_DASH] = ACTIONS(179),
    [anon_sym_GT_AMP_DASH] = ACTIONS(179),
    [anon_sym_LT_LT_LT] = ACTIONS(181),
    [anon_sym_PLUS_PLUS2] = ACTIONS(183),
    [anon_sym_DASH_DASH2] = ACTIONS(183),
    [anon_sym_DASH2] = ACTIONS(185),
    [anon_sym_PLUS2] = ACTIONS(185),
    [anon_sym_TILDE] = ACTIONS(187),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(189),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(191),
    [anon_sym_DOLLAR] = ACTIONS(193),
    [sym__special_character] = ACTIONS(195),
    [anon_sym_DQUOTE] = ACTIONS(197),
    [sym_raw_string] = ACTIONS(199),
    [sym_ansi_c_string] = ACTIONS(199),
    [aux_sym_number_token1] = ACTIONS(201),
    [aux_sym_number_token2] = ACTIONS(203),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(205),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(207),
    [anon_sym_BQUOTE] = ACTIONS(209),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(211),
    [anon_sym_LT_LPAREN] = ACTIONS(213),
    [anon_sym_GT_LPAREN] = ACTIONS(213),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(215),
    [sym_variable_name] = ACTIONS(217),
    [sym_test_operator] = ACTIONS(219),
    [sym__brace_start] = ACTIONS(221),
  },
  [8] = {
    [sym__statement_not_pipeline] = STATE(6568),
    [sym_redirected_statement] = STATE(5358),
    [sym_for_statement] = STATE(5406),
    [sym_c_style_for_statement] = STATE(5406),
    [sym_while_statement] = STATE(5195),
    [sym_if_statement] = STATE(5195),
    [sym_case_statement] = STATE(5406),
    [sym_function_definition] = STATE(5406),
    [sym_compound_statement] = STATE(5406),
    [sym_subshell] = STATE(5406),
    [sym_pipeline] = STATE(5569),
    [sym_list] = STATE(5406),
    [sym_negated_command] = STATE(5406),
    [sym_test_command] = STATE(5406),
    [sym_declaration_command] = STATE(5406),
    [sym_unset_command] = STATE(5406),
    [sym_command] = STATE(5406),
    [sym_command_name] = STATE(620),
    [sym_variable_assignment] = STATE(1811),
    [sym_variable_assignments] = STATE(5406),
    [sym_subscript] = STATE(6756),
    [sym_file_redirect] = STATE(2693),
    [sym_herestring_redirect] = STATE(2691),
    [sym__expression] = STATE(3052),
    [sym_binary_expression] = STATE(3071),
    [sym_ternary_expression] = STATE(3071),
    [sym_unary_expression] = STATE(3071),
    [sym_postfix_expression] = STATE(3071),
    [sym_parenthesized_expression] = STATE(3071),
    [sym_arithmetic_expansion] = STATE(496),
    [sym_brace_expression] = STATE(496),
    [sym_concatenation] = STATE(524),
    [sym_string] = STATE(496),
    [sym_translated_string] = STATE(496),
    [sym_number] = STATE(496),
    [sym_simple_expansion] = STATE(496),
    [sym_expansion] = STATE(496),
    [sym_command_substitution] = STATE(496),
    [sym_process_substitution] = STATE(496),
    [aux_sym_redirected_statement_repeat2] = STATE(5167),
    [aux_sym_command_repeat1] = STATE(955),
    [aux_sym__literal_repeat1] = STATE(522),
    [sym_word] = ACTIONS(143),
    [anon_sym_for] = ACTIONS(145),
    [anon_sym_select] = ACTIONS(147),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(149),
    [anon_sym_LT] = ACTIONS(151),
    [anon_sym_GT] = ACTIONS(151),
    [anon_sym_GT_GT] = ACTIONS(153),
    [anon_sym_LPAREN] = ACTIONS(155),
    [anon_sym_while] = ACTIONS(157),
    [anon_sym_until] = ACTIONS(157),
    [anon_sym_if] = ACTIONS(159),
    [anon_sym_case] = ACTIONS(161),
    [anon_sym_function] = ACTIONS(163),
    [anon_sym_LBRACE] = ACTIONS(165),
    [anon_sym_BANG] = ACTIONS(167),
    [anon_sym_LBRACK] = ACTIONS(169),
    [anon_sym_RBRACK] = ACTIONS(229),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(173),
    [anon_sym_declare] = ACTIONS(175),
    [anon_sym_typeset] = ACTIONS(175),
    [anon_sym_export] = ACTIONS(175),
    [anon_sym_readonly] = ACTIONS(175),
    [anon_sym_local] = ACTIONS(175),
    [anon_sym_unset] = ACTIONS(177),
    [anon_sym_unsetenv] = ACTIONS(177),
    [anon_sym_AMP_GT] = ACTIONS(151),
    [anon_sym_AMP_GT_GT] = ACTIONS(153),
    [anon_sym_LT_AMP] = ACTIONS(151),
    [anon_sym_GT_AMP] = ACTIONS(151),
    [anon_sym_GT_PIPE] = ACTIONS(153),
    [anon_sym_LT_AMP_DASH] = ACTIONS(179),
    [anon_sym_GT_AMP_DASH] = ACTIONS(179),
    [anon_sym_LT_LT_LT] = ACTIONS(181),
    [anon_sym_PLUS_PLUS2] = ACTIONS(183),
    [anon_sym_DASH_DASH2] = ACTIONS(183),
    [anon_sym_DASH2] = ACTIONS(185),
    [anon_sym_PLUS2] = ACTIONS(185),
    [anon_sym_TILDE] = ACTIONS(187),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(189),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(191),
    [anon_sym_DOLLAR] = ACTIONS(193),
    [sym__special_character] = ACTIONS(195),
    [anon_sym_DQUOTE] = ACTIONS(197),
    [sym_raw_string] = ACTIONS(199),
    [sym_ansi_c_string] = ACTIONS(199),
    [aux_sym_number_token1] = ACTIONS(201),
    [aux_sym_number_token2] = ACTIONS(203),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(205),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(207),
    [anon_sym_BQUOTE] = ACTIONS(209),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(211),
    [anon_sym_LT_LPAREN] = ACTIONS(213),
    [anon_sym_GT_LPAREN] = ACTIONS(213),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(215),
    [sym_variable_name] = ACTIONS(217),
    [sym_test_operator] = ACTIONS(219),
    [sym__brace_start] = ACTIONS(221),
  },
  [9] = {
    [sym__statement_not_pipeline] = STATE(6568),
    [sym_redirected_statement] = STATE(5356),
    [sym_for_statement] = STATE(5406),
    [sym_c_style_for_statement] = STATE(5406),
    [sym_while_statement] = STATE(5195),
    [sym_if_statement] = STATE(5195),
    [sym_case_statement] = STATE(5406),
    [sym_function_definition] = STATE(5406),
    [sym_compound_statement] = STATE(5406),
    [sym_subshell] = STATE(5406),
    [sym_pipeline] = STATE(5569),
    [sym_list] = STATE(5406),
    [sym_negated_command] = STATE(5406),
    [sym_test_command] = STATE(5406),
    [sym_declaration_command] = STATE(5406),
    [sym_unset_command] = STATE(5406),
    [sym_command] = STATE(5406),
    [sym_command_name] = STATE(620),
    [sym_variable_assignment] = STATE(1811),
    [sym_variable_assignments] = STATE(5406),
    [sym_subscript] = STATE(6756),
    [sym_file_redirect] = STATE(2693),
    [sym_herestring_redirect] = STATE(2691),
    [sym__expression] = STATE(3140),
    [sym_binary_expression] = STATE(3071),
    [sym_ternary_expression] = STATE(3071),
    [sym_unary_expression] = STATE(3071),
    [sym_postfix_expression] = STATE(3071),
    [sym_parenthesized_expression] = STATE(3071),
    [sym_arithmetic_expansion] = STATE(496),
    [sym_brace_expression] = STATE(496),
    [sym_concatenation] = STATE(524),
    [sym_string] = STATE(496),
    [sym_translated_string] = STATE(496),
    [sym_number] = STATE(496),
    [sym_simple_expansion] = STATE(496),
    [sym_expansion] = STATE(496),
    [sym_command_substitution] = STATE(496),
    [sym_process_substitution] = STATE(496),
    [aux_sym_redirected_statement_repeat2] = STATE(5167),
    [aux_sym_command_repeat1] = STATE(955),
    [aux_sym__literal_repeat1] = STATE(522),
    [sym_word] = ACTIONS(143),
    [anon_sym_for] = ACTIONS(145),
    [anon_sym_select] = ACTIONS(147),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(149),
    [anon_sym_LT] = ACTIONS(151),
    [anon_sym_GT] = ACTIONS(151),
    [anon_sym_GT_GT] = ACTIONS(153),
    [anon_sym_LPAREN] = ACTIONS(155),
    [anon_sym_while] = ACTIONS(157),
    [anon_sym_until] = ACTIONS(157),
    [anon_sym_if] = ACTIONS(159),
    [anon_sym_case] = ACTIONS(161),
    [anon_sym_function] = ACTIONS(163),
    [anon_sym_LBRACE] = ACTIONS(165),
    [anon_sym_BANG] = ACTIONS(167),
    [anon_sym_LBRACK] = ACTIONS(169),
    [anon_sym_RBRACK] = ACTIONS(231),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(173),
    [anon_sym_declare] = ACTIONS(175),
    [anon_sym_typeset] = ACTIONS(175),
    [anon_sym_export] = ACTIONS(175),
    [anon_sym_readonly] = ACTIONS(175),
    [anon_sym_local] = ACTIONS(175),
    [anon_sym_unset] = ACTIONS(177),
    [anon_sym_unsetenv] = ACTIONS(177),
    [anon_sym_AMP_GT] = ACTIONS(151),
    [anon_sym_AMP_GT_GT] = ACTIONS(153),
    [anon_sym_LT_AMP] = ACTIONS(151),
    [anon_sym_GT_AMP] = ACTIONS(151),
    [anon_sym_GT_PIPE] = ACTIONS(153),
    [anon_sym_LT_AMP_DASH] = ACTIONS(179),
    [anon_sym_GT_AMP_DASH] = ACTIONS(179),
    [anon_sym_LT_LT_LT] = ACTIONS(181),
    [anon_sym_PLUS_PLUS2] = ACTIONS(183),
    [anon_sym_DASH_DASH2] = ACTIONS(183),
    [anon_sym_DASH2] = ACTIONS(185),
    [anon_sym_PLUS2] = ACTIONS(185),
    [anon_sym_TILDE] = ACTIONS(187),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(189),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(191),
    [anon_sym_DOLLAR] = ACTIONS(193),
    [sym__special_character] = ACTIONS(195),
    [anon_sym_DQUOTE] = ACTIONS(197),
    [sym_raw_string] = ACTIONS(199),
    [sym_ansi_c_string] = ACTIONS(199),
    [aux_sym_number_token1] = ACTIONS(201),
    [aux_sym_number_token2] = ACTIONS(203),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(205),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(207),
    [anon_sym_BQUOTE] = ACTIONS(209),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(211),
    [anon_sym_LT_LPAREN] = ACTIONS(213),
    [anon_sym_GT_LPAREN] = ACTIONS(213),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(215),
    [sym_variable_name] = ACTIONS(217),
    [sym_test_operator] = ACTIONS(219),
    [sym__brace_start] = ACTIONS(221),
  },
  [10] = {
    [sym__statement_not_pipeline] = STATE(6568),
    [sym_redirected_statement] = STATE(5315),
    [sym_for_statement] = STATE(5406),
    [sym_c_style_for_statement] = STATE(5406),
    [sym_while_statement] = STATE(5195),
    [sym_if_statement] = STATE(5195),
    [sym_case_statement] = STATE(5406),
    [sym_function_definition] = STATE(5406),
    [sym_compound_statement] = STATE(5406),
    [sym_subshell] = STATE(5406),
    [sym_pipeline] = STATE(5569),
    [sym_list] = STATE(5406),
    [sym_negated_command] = STATE(5406),
    [sym_test_command] = STATE(5406),
    [sym_declaration_command] = STATE(5406),
    [sym_unset_command] = STATE(5406),
    [sym_command] = STATE(5406),
    [sym_command_name] = STATE(620),
    [sym_variable_assignment] = STATE(1811),
    [sym_variable_assignments] = STATE(5406),
    [sym_subscript] = STATE(6756),
    [sym_file_redirect] = STATE(2693),
    [sym_herestring_redirect] = STATE(2691),
    [sym__expression] = STATE(3085),
    [sym_binary_expression] = STATE(3071),
    [sym_ternary_expression] = STATE(3071),
    [sym_unary_expression] = STATE(3071),
    [sym_postfix_expression] = STATE(3071),
    [sym_parenthesized_expression] = STATE(3071),
    [sym_arithmetic_expansion] = STATE(496),
    [sym_brace_expression] = STATE(496),
    [sym_concatenation] = STATE(524),
    [sym_string] = STATE(496),
    [sym_translated_string] = STATE(496),
    [sym_number] = STATE(496),
    [sym_simple_expansion] = STATE(496),
    [sym_expansion] = STATE(496),
    [sym_command_substitution] = STATE(496),
    [sym_process_substitution] = STATE(496),
    [aux_sym_redirected_statement_repeat2] = STATE(5167),
    [aux_sym_command_repeat1] = STATE(955),
    [aux_sym__literal_repeat1] = STATE(522),
    [sym_word] = ACTIONS(143),
    [anon_sym_for] = ACTIONS(145),
    [anon_sym_select] = ACTIONS(147),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(149),
    [anon_sym_LT] = ACTIONS(151),
    [anon_sym_GT] = ACTIONS(151),
    [anon_sym_GT_GT] = ACTIONS(153),
    [anon_sym_LPAREN] = ACTIONS(155),
    [anon_sym_while] = ACTIONS(157),
    [anon_sym_until] = ACTIONS(157),
    [anon_sym_if] = ACTIONS(159),
    [anon_sym_case] = ACTIONS(161),
    [anon_sym_function] = ACTIONS(163),
    [anon_sym_LBRACE] = ACTIONS(165),
    [anon_sym_BANG] = ACTIONS(167),
    [anon_sym_LBRACK] = ACTIONS(169),
    [anon_sym_RBRACK] = ACTIONS(233),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(173),
    [anon_sym_declare] = ACTIONS(175),
    [anon_sym_typeset] = ACTIONS(175),
    [anon_sym_export] = ACTIONS(175),
    [anon_sym_readonly] = ACTIONS(175),
    [anon_sym_local] = ACTIONS(175),
    [anon_sym_unset] = ACTIONS(177),
    [anon_sym_unsetenv] = ACTIONS(177),
    [anon_sym_AMP_GT] = ACTIONS(151),
    [anon_sym_AMP_GT_GT] = ACTIONS(153),
    [anon_sym_LT_AMP] = ACTIONS(151),
    [anon_sym_GT_AMP] = ACTIONS(151),
    [anon_sym_GT_PIPE] = ACTIONS(153),
    [anon_sym_LT_AMP_DASH] = ACTIONS(179),
    [anon_sym_GT_AMP_DASH] = ACTIONS(179),
    [anon_sym_LT_LT_LT] = ACTIONS(181),
    [anon_sym_PLUS_PLUS2] = ACTIONS(183),
    [anon_sym_DASH_DASH2] = ACTIONS(183),
    [anon_sym_DASH2] = ACTIONS(185),
    [anon_sym_PLUS2] = ACTIONS(185),
    [anon_sym_TILDE] = ACTIONS(187),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(189),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(191),
    [anon_sym_DOLLAR] = ACTIONS(193),
    [sym__special_character] = ACTIONS(195),
    [anon_sym_DQUOTE] = ACTIONS(197),
    [sym_raw_string] = ACTIONS(199),
    [sym_ansi_c_string] = ACTIONS(199),
    [aux_sym_number_token1] = ACTIONS(201),
    [aux_sym_number_token2] = ACTIONS(203),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(205),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(207),
    [anon_sym_BQUOTE] = ACTIONS(209),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(211),
    [anon_sym_LT_LPAREN] = ACTIONS(213),
    [anon_sym_GT_LPAREN] = ACTIONS(213),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(215),
    [sym_variable_name] = ACTIONS(217),
    [sym_test_operator] = ACTIONS(219),
    [sym__brace_start] = ACTIONS(221),
  },
  [11] = {
    [sym__statement_not_pipeline] = STATE(6568),
    [sym_redirected_statement] = STATE(5289),
    [sym_for_statement] = STATE(5406),
    [sym_c_style_for_statement] = STATE(5406),
    [sym_while_statement] = STATE(5195),
    [sym_if_statement] = STATE(5195),
    [sym_case_statement] = STATE(5406),
    [sym_function_definition] = STATE(5406),
    [sym_compound_statement] = STATE(5406),
    [sym_subshell] = STATE(5406),
    [sym_pipeline] = STATE(5569),
    [sym_list] = STATE(5406),
    [sym_negated_command] = STATE(5406),
    [sym_test_command] = STATE(5406),
    [sym_declaration_command] = STATE(5406),
    [sym_unset_command] = STATE(5406),
    [sym_command] = STATE(5406),
    [sym_command_name] = STATE(620),
    [sym_variable_assignment] = STATE(1811),
    [sym_variable_assignments] = STATE(5406),
    [sym_subscript] = STATE(6756),
    [sym_file_redirect] = STATE(2693),
    [sym_herestring_redirect] = STATE(2691),
    [sym__expression] = STATE(3152),
    [sym_binary_expression] = STATE(3071),
    [sym_ternary_expression] = STATE(3071),
    [sym_unary_expression] = STATE(3071),
    [sym_postfix_expression] = STATE(3071),
    [sym_parenthesized_expression] = STATE(3071),
    [sym_arithmetic_expansion] = STATE(496),
    [sym_brace_expression] = STATE(496),
    [sym_concatenation] = STATE(524),
    [sym_string] = STATE(496),
    [sym_translated_string] = STATE(496),
    [sym_number] = STATE(496),
    [sym_simple_expansion] = STATE(496),
    [sym_expansion] = STATE(496),
    [sym_command_substitution] = STATE(496),
    [sym_process_substitution] = STATE(496),
    [aux_sym_redirected_statement_repeat2] = STATE(5167),
    [aux_sym_command_repeat1] = STATE(955),
    [aux_sym__literal_repeat1] = STATE(522),
    [sym_word] = ACTIONS(143),
    [anon_sym_for] = ACTIONS(145),
    [anon_sym_select] = ACTIONS(147),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(149),
    [anon_sym_LT] = ACTIONS(151),
    [anon_sym_GT] = ACTIONS(151),
    [anon_sym_GT_GT] = ACTIONS(153),
    [anon_sym_LPAREN] = ACTIONS(155),
    [anon_sym_while] = ACTIONS(157),
    [anon_sym_until] = ACTIONS(157),
    [anon_sym_if] = ACTIONS(159),
    [anon_sym_case] = ACTIONS(161),
    [anon_sym_function] = ACTIONS(163),
    [anon_sym_LBRACE] = ACTIONS(165),
    [anon_sym_BANG] = ACTIONS(167),
    [anon_sym_LBRACK] = ACTIONS(169),
    [anon_sym_RBRACK] = ACTIONS(235),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(173),
    [anon_sym_declare] = ACTIONS(175),
    [anon_sym_typeset] = ACTIONS(175),
    [anon_sym_export] = ACTIONS(175),
    [anon_sym_readonly] = ACTIONS(175),
    [anon_sym_local] = ACTIONS(175),
    [anon_sym_unset] = ACTIONS(177),
    [anon_sym_unsetenv] = ACTIONS(177),
    [anon_sym_AMP_GT] = ACTIONS(151),
    [anon_sym_AMP_GT_GT] = ACTIONS(153),
    [anon_sym_LT_AMP] = ACTIONS(151),
    [anon_sym_GT_AMP] = ACTIONS(151),
    [anon_sym_GT_PIPE] = ACTIONS(153),
    [anon_sym_LT_AMP_DASH] = ACTIONS(179),
    [anon_sym_GT_AMP_DASH] = ACTIONS(179),
    [anon_sym_LT_LT_LT] = ACTIONS(181),
    [anon_sym_PLUS_PLUS2] = ACTIONS(183),
    [anon_sym_DASH_DASH2] = ACTIONS(183),
    [anon_sym_DASH2] = ACTIONS(185),
    [anon_sym_PLUS2] = ACTIONS(185),
    [anon_sym_TILDE] = ACTIONS(187),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(189),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(191),
    [anon_sym_DOLLAR] = ACTIONS(193),
    [sym__special_character] = ACTIONS(195),
    [anon_sym_DQUOTE] = ACTIONS(197),
    [sym_raw_string] = ACTIONS(199),
    [sym_ansi_c_string] = ACTIONS(199),
    [aux_sym_number_token1] = ACTIONS(201),
    [aux_sym_number_token2] = ACTIONS(203),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(205),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(207),
    [anon_sym_BQUOTE] = ACTIONS(209),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(211),
    [anon_sym_LT_LPAREN] = ACTIONS(213),
    [anon_sym_GT_LPAREN] = ACTIONS(213),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(215),
    [sym_variable_name] = ACTIONS(217),
    [sym_test_operator] = ACTIONS(219),
    [sym__brace_start] = ACTIONS(221),
  },
  [12] = {
    [sym__expression] = STATE(3098),
    [sym_binary_expression] = STATE(3053),
    [sym_ternary_expression] = STATE(3053),
    [sym_unary_expression] = STATE(3053),
    [sym_postfix_expression] = STATE(3053),
    [sym_parenthesized_expression] = STATE(3053),
    [sym_arithmetic_expansion] = STATE(2500),
    [sym_brace_expression] = STATE(2500),
    [sym_concatenation] = STATE(3053),
    [sym_string] = STATE(2500),
    [sym_translated_string] = STATE(2500),
    [sym_number] = STATE(2500),
    [sym_simple_expansion] = STATE(2500),
    [sym_expansion] = STATE(2500),
    [sym_command_substitution] = STATE(2500),
    [sym_process_substitution] = STATE(2500),
    [aux_sym__literal_repeat1] = STATE(2690),
    [aux_sym_concatenation_repeat1] = STATE(470),
    [sym_word] = ACTIONS(237),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(239),
    [anon_sym_SEMI] = ACTIONS(241),
    [anon_sym_EQ] = ACTIONS(243),
    [anon_sym_PLUS_PLUS] = ACTIONS(243),
    [anon_sym_DASH_DASH] = ACTIONS(243),
    [anon_sym_PLUS_EQ] = ACTIONS(243),
    [anon_sym_DASH_EQ] = ACTIONS(243),
    [anon_sym_STAR_EQ] = ACTIONS(243),
    [anon_sym_SLASH_EQ] = ACTIONS(243),
    [anon_sym_PERCENT_EQ] = ACTIONS(243),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(243),
    [anon_sym_LT_LT_EQ] = ACTIONS(243),
    [anon_sym_GT_GT_EQ] = ACTIONS(243),
    [anon_sym_AMP_EQ] = ACTIONS(243),
    [anon_sym_CARET_EQ] = ACTIONS(243),
    [anon_sym_PIPE_EQ] = ACTIONS(243),
    [anon_sym_PIPE_PIPE] = ACTIONS(245),
    [anon_sym_AMP_AMP] = ACTIONS(245),
    [anon_sym_PIPE] = ACTIONS(245),
    [anon_sym_CARET] = ACTIONS(243),
    [anon_sym_AMP] = ACTIONS(245),
    [anon_sym_EQ_EQ] = ACTIONS(245),
    [anon_sym_BANG_EQ] = ACTIONS(243),
    [anon_sym_LT] = ACTIONS(245),
    [anon_sym_GT] = ACTIONS(245),
    [anon_sym_LT_EQ] = ACTIONS(243),
    [anon_sym_GT_EQ] = ACTIONS(243),
    [anon_sym_LT_LT] = ACTIONS(245),
    [anon_sym_GT_GT] = ACTIONS(245),
    [anon_sym_PLUS] = ACTIONS(243),
    [anon_sym_DASH] = ACTIONS(243),
    [anon_sym_STAR] = ACTIONS(243),
    [anon_sym_SLASH] = ACTIONS(243),
    [anon_sym_PERCENT] = ACTIONS(243),
    [anon_sym_STAR_STAR] = ACTIONS(243),
    [anon_sym_LPAREN] = ACTIONS(248),
    [anon_sym_RPAREN] = ACTIONS(245),
    [anon_sym_SEMI_SEMI] = ACTIONS(241),
    [anon_sym_PIPE_AMP] = ACTIONS(241),
    [anon_sym_BANG] = ACTIONS(250),
    [anon_sym_EQ_TILDE] = ACTIONS(245),
    [anon_sym_AMP_GT] = ACTIONS(241),
    [anon_sym_AMP_GT_GT] = ACTIONS(241),
    [anon_sym_LT_AMP] = ACTIONS(241),
    [anon_sym_GT_AMP] = ACTIONS(241),
    [anon_sym_GT_PIPE] = ACTIONS(241),
    [anon_sym_LT_AMP_DASH] = ACTIONS(241),
    [anon_sym_GT_AMP_DASH] = ACTIONS(241),
    [anon_sym_LT_LT_DASH] = ACTIONS(241),
    [aux_sym_heredoc_redirect_token1] = ACTIONS(241),
    [anon_sym_LT_LT_LT] = ACTIONS(241),
    [anon_sym_QMARK] = ACTIONS(243),
    [anon_sym_PLUS_PLUS2] = ACTIONS(252),
    [anon_sym_DASH_DASH2] = ACTIONS(252),
    [anon_sym_DASH2] = ACTIONS(105),
    [anon_sym_PLUS2] = ACTIONS(105),
    [anon_sym_TILDE] = ACTIONS(105),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(239),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(254),
    [aux_sym_concatenation_token1] = ACTIONS(256),
    [anon_sym_DOLLAR] = ACTIONS(258),
    [sym__special_character] = ACTIONS(260),
    [anon_sym_DQUOTE] = ACTIONS(262),
    [sym_raw_string] = ACTIONS(237),
    [sym_ansi_c_string] = ACTIONS(237),
    [aux_sym_number_token1] = ACTIONS(264),
    [aux_sym_number_token2] = ACTIONS(266),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(268),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(270),
    [anon_sym_BQUOTE] = ACTIONS(272),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(274),
    [anon_sym_LT_LPAREN] = ACTIONS(276),
    [anon_sym_GT_LPAREN] = ACTIONS(276),
    [sym_comment] = ACTIONS(3),
    [sym_file_descriptor] = ACTIONS(278),
    [sym__concat] = ACTIONS(280),
    [sym_test_operator] = ACTIONS(282),
    [sym__bare_dollar] = ACTIONS(278),
    [sym__brace_start] = ACTIONS(284),
  },
  [13] = {
    [aux_sym__terminated_statement] = STATE(34),
    [sym__statement_not_pipeline] = STATE(6718),
    [sym_redirected_statement] = STATE(5007),
    [sym_for_statement] = STATE(5007),
    [sym_c_style_for_statement] = STATE(5007),
    [sym_while_statement] = STATE(4551),
    [sym_if_statement] = STATE(4551),
    [sym_elif_clause] = STATE(6289),
    [sym_else_clause] = STATE(7417),
    [sym_case_statement] = STATE(5007),
    [sym_function_definition] = STATE(5007),
    [sym_compound_statement] = STATE(5007),
    [sym_subshell] = STATE(5007),
    [sym_pipeline] = STATE(5189),
    [sym_list] = STATE(5007),
    [sym_negated_command] = STATE(5007),
    [sym_test_command] = STATE(5007),
    [sym_declaration_command] = STATE(5007),
    [sym_unset_command] = STATE(5007),
    [sym_command] = STATE(5007),
    [sym_command_name] = STATE(544),
    [sym_variable_assignment] = STATE(1151),
    [sym_variable_assignments] = STATE(5007),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym_redirected_statement_repeat2] = STATE(4644),
    [aux_sym_if_statement_repeat1] = STATE(6289),
    [aux_sym_command_repeat1] = STATE(997),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(286),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_fi] = ACTIONS(294),
    [anon_sym_elif] = ACTIONS(296),
    [anon_sym_else] = ACTIONS(298),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(300),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(302),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(304),
    [anon_sym_typeset] = ACTIONS(304),
    [anon_sym_export] = ACTIONS(304),
    [anon_sym_readonly] = ACTIONS(304),
    [anon_sym_local] = ACTIONS(304),
    [anon_sym_unset] = ACTIONS(306),
    [anon_sym_unsetenv] = ACTIONS(306),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [14] = {
    [aux_sym__terminated_statement] = STATE(34),
    [sym__statement_not_pipeline] = STATE(6718),
    [sym_redirected_statement] = STATE(5007),
    [sym_for_statement] = STATE(5007),
    [sym_c_style_for_statement] = STATE(5007),
    [sym_while_statement] = STATE(4551),
    [sym_if_statement] = STATE(4551),
    [sym_elif_clause] = STATE(6327),
    [sym_else_clause] = STATE(7004),
    [sym_case_statement] = STATE(5007),
    [sym_function_definition] = STATE(5007),
    [sym_compound_statement] = STATE(5007),
    [sym_subshell] = STATE(5007),
    [sym_pipeline] = STATE(5189),
    [sym_list] = STATE(5007),
    [sym_negated_command] = STATE(5007),
    [sym_test_command] = STATE(5007),
    [sym_declaration_command] = STATE(5007),
    [sym_unset_command] = STATE(5007),
    [sym_command] = STATE(5007),
    [sym_command_name] = STATE(544),
    [sym_variable_assignment] = STATE(1151),
    [sym_variable_assignments] = STATE(5007),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym_redirected_statement_repeat2] = STATE(4644),
    [aux_sym_if_statement_repeat1] = STATE(6327),
    [aux_sym_command_repeat1] = STATE(997),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(286),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_fi] = ACTIONS(346),
    [anon_sym_elif] = ACTIONS(296),
    [anon_sym_else] = ACTIONS(298),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(300),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(302),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(304),
    [anon_sym_typeset] = ACTIONS(304),
    [anon_sym_export] = ACTIONS(304),
    [anon_sym_readonly] = ACTIONS(304),
    [anon_sym_local] = ACTIONS(304),
    [anon_sym_unset] = ACTIONS(306),
    [anon_sym_unsetenv] = ACTIONS(306),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [15] = {
    [aux_sym__terminated_statement] = STATE(34),
    [sym__statement_not_pipeline] = STATE(6718),
    [sym_redirected_statement] = STATE(5007),
    [sym_for_statement] = STATE(5007),
    [sym_c_style_for_statement] = STATE(5007),
    [sym_while_statement] = STATE(4551),
    [sym_if_statement] = STATE(4551),
    [sym_elif_clause] = STATE(6265),
    [sym_else_clause] = STATE(7416),
    [sym_case_statement] = STATE(5007),
    [sym_function_definition] = STATE(5007),
    [sym_compound_statement] = STATE(5007),
    [sym_subshell] = STATE(5007),
    [sym_pipeline] = STATE(5189),
    [sym_list] = STATE(5007),
    [sym_negated_command] = STATE(5007),
    [sym_test_command] = STATE(5007),
    [sym_declaration_command] = STATE(5007),
    [sym_unset_command] = STATE(5007),
    [sym_command] = STATE(5007),
    [sym_command_name] = STATE(544),
    [sym_variable_assignment] = STATE(1151),
    [sym_variable_assignments] = STATE(5007),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym_redirected_statement_repeat2] = STATE(4644),
    [aux_sym_if_statement_repeat1] = STATE(6265),
    [aux_sym_command_repeat1] = STATE(997),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(286),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_fi] = ACTIONS(348),
    [anon_sym_elif] = ACTIONS(296),
    [anon_sym_else] = ACTIONS(298),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(300),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(302),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(304),
    [anon_sym_typeset] = ACTIONS(304),
    [anon_sym_export] = ACTIONS(304),
    [anon_sym_readonly] = ACTIONS(304),
    [anon_sym_local] = ACTIONS(304),
    [anon_sym_unset] = ACTIONS(306),
    [anon_sym_unsetenv] = ACTIONS(306),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [16] = {
    [sym__expression] = STATE(3012),
    [sym_binary_expression] = STATE(3071),
    [sym_ternary_expression] = STATE(3071),
    [sym_unary_expression] = STATE(3071),
    [sym_postfix_expression] = STATE(3071),
    [sym_parenthesized_expression] = STATE(3071),
    [sym_arithmetic_expansion] = STATE(2521),
    [sym_brace_expression] = STATE(2521),
    [sym_concatenation] = STATE(3071),
    [sym_string] = STATE(2521),
    [sym_translated_string] = STATE(2521),
    [sym_number] = STATE(2521),
    [sym_simple_expansion] = STATE(2521),
    [sym_expansion] = STATE(2521),
    [sym_command_substitution] = STATE(2521),
    [sym_process_substitution] = STATE(2521),
    [aux_sym__literal_repeat1] = STATE(2641),
    [aux_sym_concatenation_repeat1] = STATE(499),
    [sym_word] = ACTIONS(350),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(352),
    [anon_sym_EQ] = ACTIONS(243),
    [anon_sym_PLUS_PLUS] = ACTIONS(243),
    [anon_sym_DASH_DASH] = ACTIONS(243),
    [anon_sym_PLUS_EQ] = ACTIONS(243),
    [anon_sym_DASH_EQ] = ACTIONS(243),
    [anon_sym_STAR_EQ] = ACTIONS(243),
    [anon_sym_SLASH_EQ] = ACTIONS(243),
    [anon_sym_PERCENT_EQ] = ACTIONS(243),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(243),
    [anon_sym_LT_LT_EQ] = ACTIONS(354),
    [anon_sym_GT_GT_EQ] = ACTIONS(354),
    [anon_sym_AMP_EQ] = ACTIONS(354),
    [anon_sym_CARET_EQ] = ACTIONS(243),
    [anon_sym_PIPE_EQ] = ACTIONS(354),
    [anon_sym_PIPE_PIPE] = ACTIONS(356),
    [anon_sym_AMP_AMP] = ACTIONS(356),
    [anon_sym_PIPE] = ACTIONS(245),
    [anon_sym_CARET] = ACTIONS(243),
    [anon_sym_AMP] = ACTIONS(243),
    [anon_sym_EQ_EQ] = ACTIONS(245),
    [anon_sym_BANG_EQ] = ACTIONS(243),
    [anon_sym_LT] = ACTIONS(245),
    [anon_sym_GT] = ACTIONS(245),
    [anon_sym_LT_EQ] = ACTIONS(354),
    [anon_sym_GT_EQ] = ACTIONS(354),
    [anon_sym_LT_LT] = ACTIONS(245),
    [anon_sym_GT_GT] = ACTIONS(245),
    [anon_sym_PLUS] = ACTIONS(243),
    [anon_sym_DASH] = ACTIONS(243),
    [anon_sym_STAR] = ACTIONS(243),
    [anon_sym_SLASH] = ACTIONS(243),
    [anon_sym_PERCENT] = ACTIONS(243),
    [anon_sym_STAR_STAR] = ACTIONS(243),
    [anon_sym_LPAREN] = ACTIONS(359),
    [anon_sym_PIPE_AMP] = ACTIONS(278),
    [anon_sym_BANG] = ACTIONS(361),
    [anon_sym_RBRACK] = ACTIONS(354),
    [anon_sym_EQ_TILDE] = ACTIONS(245),
    [anon_sym_AMP_GT] = ACTIONS(241),
    [anon_sym_AMP_GT_GT] = ACTIONS(278),
    [anon_sym_LT_AMP] = ACTIONS(241),
    [anon_sym_GT_AMP] = ACTIONS(241),
    [anon_sym_GT_PIPE] = ACTIONS(278),
    [anon_sym_LT_AMP_DASH] = ACTIONS(278),
    [anon_sym_GT_AMP_DASH] = ACTIONS(278),
    [anon_sym_LT_LT_DASH] = ACTIONS(278),
    [anon_sym_LT_LT_LT] = ACTIONS(278),
    [anon_sym_QMARK] = ACTIONS(243),
    [anon_sym_PLUS_PLUS2] = ACTIONS(183),
    [anon_sym_DASH_DASH2] = ACTIONS(183),
    [anon_sym_DASH2] = ACTIONS(185),
    [anon_sym_PLUS2] = ACTIONS(185),
    [anon_sym_TILDE] = ACTIONS(187),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(352),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(363),
    [aux_sym_concatenation_token1] = ACTIONS(365),
    [anon_sym_DOLLAR] = ACTIONS(367),
    [sym__special_character] = ACTIONS(369),
    [anon_sym_DQUOTE] = ACTIONS(371),
    [sym_raw_string] = ACTIONS(373),
    [sym_ansi_c_string] = ACTIONS(373),
    [aux_sym_number_token1] = ACTIONS(375),
    [aux_sym_number_token2] = ACTIONS(377),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(379),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(381),
    [anon_sym_BQUOTE] = ACTIONS(383),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(385),
    [anon_sym_LT_LPAREN] = ACTIONS(387),
    [anon_sym_GT_LPAREN] = ACTIONS(387),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(278),
    [sym__concat] = ACTIONS(365),
    [sym_test_operator] = ACTIONS(389),
    [sym__bare_dollar] = ACTIONS(278),
    [sym__brace_start] = ACTIONS(391),
  },
  [17] = {
    [aux_sym__terminated_statement] = STATE(15),
    [sym__statement_not_pipeline] = STATE(6718),
    [sym_redirected_statement] = STATE(5007),
    [sym_for_statement] = STATE(5007),
    [sym_c_style_for_statement] = STATE(5007),
    [sym_while_statement] = STATE(4551),
    [sym_if_statement] = STATE(4551),
    [sym_elif_clause] = STATE(6340),
    [sym_else_clause] = STATE(7114),
    [sym_case_statement] = STATE(5007),
    [sym_function_definition] = STATE(5007),
    [sym_compound_statement] = STATE(5007),
    [sym_subshell] = STATE(5007),
    [sym_pipeline] = STATE(5189),
    [sym_list] = STATE(5007),
    [sym_negated_command] = STATE(5007),
    [sym_test_command] = STATE(5007),
    [sym_declaration_command] = STATE(5007),
    [sym_unset_command] = STATE(5007),
    [sym_command] = STATE(5007),
    [sym_command_name] = STATE(544),
    [sym_variable_assignment] = STATE(1151),
    [sym_variable_assignments] = STATE(5007),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym_redirected_statement_repeat2] = STATE(4644),
    [aux_sym_if_statement_repeat1] = STATE(6340),
    [aux_sym_command_repeat1] = STATE(997),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(286),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_fi] = ACTIONS(393),
    [anon_sym_elif] = ACTIONS(296),
    [anon_sym_else] = ACTIONS(298),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(300),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(302),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(304),
    [anon_sym_typeset] = ACTIONS(304),
    [anon_sym_export] = ACTIONS(304),
    [anon_sym_readonly] = ACTIONS(304),
    [anon_sym_local] = ACTIONS(304),
    [anon_sym_unset] = ACTIONS(306),
    [anon_sym_unsetenv] = ACTIONS(306),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [18] = {
    [aux_sym__terminated_statement] = STATE(13),
    [sym__statement_not_pipeline] = STATE(6718),
    [sym_redirected_statement] = STATE(5007),
    [sym_for_statement] = STATE(5007),
    [sym_c_style_for_statement] = STATE(5007),
    [sym_while_statement] = STATE(4551),
    [sym_if_statement] = STATE(4551),
    [sym_elif_clause] = STATE(6363),
    [sym_else_clause] = STATE(7478),
    [sym_case_statement] = STATE(5007),
    [sym_function_definition] = STATE(5007),
    [sym_compound_statement] = STATE(5007),
    [sym_subshell] = STATE(5007),
    [sym_pipeline] = STATE(5189),
    [sym_list] = STATE(5007),
    [sym_negated_command] = STATE(5007),
    [sym_test_command] = STATE(5007),
    [sym_declaration_command] = STATE(5007),
    [sym_unset_command] = STATE(5007),
    [sym_command] = STATE(5007),
    [sym_command_name] = STATE(544),
    [sym_variable_assignment] = STATE(1151),
    [sym_variable_assignments] = STATE(5007),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym_redirected_statement_repeat2] = STATE(4644),
    [aux_sym_if_statement_repeat1] = STATE(6363),
    [aux_sym_command_repeat1] = STATE(997),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(286),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_fi] = ACTIONS(395),
    [anon_sym_elif] = ACTIONS(296),
    [anon_sym_else] = ACTIONS(298),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(300),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(302),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(304),
    [anon_sym_typeset] = ACTIONS(304),
    [anon_sym_export] = ACTIONS(304),
    [anon_sym_readonly] = ACTIONS(304),
    [anon_sym_local] = ACTIONS(304),
    [anon_sym_unset] = ACTIONS(306),
    [anon_sym_unsetenv] = ACTIONS(306),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [19] = {
    [aux_sym__terminated_statement] = STATE(20),
    [sym__statement_not_pipeline] = STATE(6718),
    [sym_redirected_statement] = STATE(5007),
    [sym_for_statement] = STATE(5007),
    [sym_c_style_for_statement] = STATE(5007),
    [sym_while_statement] = STATE(4551),
    [sym_if_statement] = STATE(4551),
    [sym_elif_clause] = STATE(6339),
    [sym_else_clause] = STATE(7069),
    [sym_case_statement] = STATE(5007),
    [sym_function_definition] = STATE(5007),
    [sym_compound_statement] = STATE(5007),
    [sym_subshell] = STATE(5007),
    [sym_pipeline] = STATE(5189),
    [sym_list] = STATE(5007),
    [sym_negated_command] = STATE(5007),
    [sym_test_command] = STATE(5007),
    [sym_declaration_command] = STATE(5007),
    [sym_unset_command] = STATE(5007),
    [sym_command] = STATE(5007),
    [sym_command_name] = STATE(544),
    [sym_variable_assignment] = STATE(1151),
    [sym_variable_assignments] = STATE(5007),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym_redirected_statement_repeat2] = STATE(4644),
    [aux_sym_if_statement_repeat1] = STATE(6339),
    [aux_sym_command_repeat1] = STATE(997),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(286),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_fi] = ACTIONS(397),
    [anon_sym_elif] = ACTIONS(296),
    [anon_sym_else] = ACTIONS(298),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(300),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(302),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(304),
    [anon_sym_typeset] = ACTIONS(304),
    [anon_sym_export] = ACTIONS(304),
    [anon_sym_readonly] = ACTIONS(304),
    [anon_sym_local] = ACTIONS(304),
    [anon_sym_unset] = ACTIONS(306),
    [anon_sym_unsetenv] = ACTIONS(306),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [20] = {
    [aux_sym__terminated_statement] = STATE(34),
    [sym__statement_not_pipeline] = STATE(6718),
    [sym_redirected_statement] = STATE(5007),
    [sym_for_statement] = STATE(5007),
    [sym_c_style_for_statement] = STATE(5007),
    [sym_while_statement] = STATE(4551),
    [sym_if_statement] = STATE(4551),
    [sym_elif_clause] = STATE(6331),
    [sym_else_clause] = STATE(6862),
    [sym_case_statement] = STATE(5007),
    [sym_function_definition] = STATE(5007),
    [sym_compound_statement] = STATE(5007),
    [sym_subshell] = STATE(5007),
    [sym_pipeline] = STATE(5189),
    [sym_list] = STATE(5007),
    [sym_negated_command] = STATE(5007),
    [sym_test_command] = STATE(5007),
    [sym_declaration_command] = STATE(5007),
    [sym_unset_command] = STATE(5007),
    [sym_command] = STATE(5007),
    [sym_command_name] = STATE(544),
    [sym_variable_assignment] = STATE(1151),
    [sym_variable_assignments] = STATE(5007),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym_redirected_statement_repeat2] = STATE(4644),
    [aux_sym_if_statement_repeat1] = STATE(6331),
    [aux_sym_command_repeat1] = STATE(997),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(286),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_fi] = ACTIONS(399),
    [anon_sym_elif] = ACTIONS(296),
    [anon_sym_else] = ACTIONS(298),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(300),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(302),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(304),
    [anon_sym_typeset] = ACTIONS(304),
    [anon_sym_export] = ACTIONS(304),
    [anon_sym_readonly] = ACTIONS(304),
    [anon_sym_local] = ACTIONS(304),
    [anon_sym_unset] = ACTIONS(306),
    [anon_sym_unsetenv] = ACTIONS(306),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [21] = {
    [aux_sym__terminated_statement] = STATE(14),
    [sym__statement_not_pipeline] = STATE(6718),
    [sym_redirected_statement] = STATE(5007),
    [sym_for_statement] = STATE(5007),
    [sym_c_style_for_statement] = STATE(5007),
    [sym_while_statement] = STATE(4551),
    [sym_if_statement] = STATE(4551),
    [sym_elif_clause] = STATE(6337),
    [sym_else_clause] = STATE(6981),
    [sym_case_statement] = STATE(5007),
    [sym_function_definition] = STATE(5007),
    [sym_compound_statement] = STATE(5007),
    [sym_subshell] = STATE(5007),
    [sym_pipeline] = STATE(5189),
    [sym_list] = STATE(5007),
    [sym_negated_command] = STATE(5007),
    [sym_test_command] = STATE(5007),
    [sym_declaration_command] = STATE(5007),
    [sym_unset_command] = STATE(5007),
    [sym_command] = STATE(5007),
    [sym_command_name] = STATE(544),
    [sym_variable_assignment] = STATE(1151),
    [sym_variable_assignments] = STATE(5007),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym_redirected_statement_repeat2] = STATE(4644),
    [aux_sym_if_statement_repeat1] = STATE(6337),
    [aux_sym_command_repeat1] = STATE(997),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(286),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_fi] = ACTIONS(401),
    [anon_sym_elif] = ACTIONS(296),
    [anon_sym_else] = ACTIONS(298),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(300),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(302),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(304),
    [anon_sym_typeset] = ACTIONS(304),
    [anon_sym_export] = ACTIONS(304),
    [anon_sym_readonly] = ACTIONS(304),
    [anon_sym_local] = ACTIONS(304),
    [anon_sym_unset] = ACTIONS(306),
    [anon_sym_unsetenv] = ACTIONS(306),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [22] = {
    [sym__statements] = STATE(6501),
    [sym__statement_not_pipeline] = STATE(6584),
    [sym_redirected_statement] = STATE(4349),
    [sym_for_statement] = STATE(4349),
    [sym_c_style_for_statement] = STATE(4349),
    [sym_while_statement] = STATE(4225),
    [sym_if_statement] = STATE(4225),
    [sym_case_statement] = STATE(4349),
    [sym_function_definition] = STATE(4349),
    [sym_compound_statement] = STATE(4349),
    [sym_subshell] = STATE(4349),
    [sym_pipeline] = STATE(4620),
    [sym_list] = STATE(4349),
    [sym_negated_command] = STATE(4349),
    [sym_test_command] = STATE(4349),
    [sym_declaration_command] = STATE(4349),
    [sym_unset_command] = STATE(4349),
    [sym_command] = STATE(4349),
    [sym_command_name] = STATE(528),
    [sym_variable_assignment] = STATE(807),
    [sym_variable_assignments] = STATE(4349),
    [sym_subscript] = STATE(6736),
    [sym_file_redirect] = STATE(1635),
    [sym_herestring_redirect] = STATE(1634),
    [sym_arithmetic_expansion] = STATE(815),
    [sym_brace_expression] = STATE(815),
    [sym_concatenation] = STATE(1082),
    [sym_string] = STATE(815),
    [sym_translated_string] = STATE(815),
    [sym_number] = STATE(815),
    [sym_simple_expansion] = STATE(815),
    [sym_expansion] = STATE(815),
    [sym_command_substitution] = STATE(815),
    [sym_process_substitution] = STATE(815),
    [aux_sym__statements_repeat1] = STATE(431),
    [aux_sym_redirected_statement_repeat2] = STATE(4250),
    [aux_sym_command_repeat1] = STATE(925),
    [aux_sym__literal_repeat1] = STATE(926),
    [sym_word] = ACTIONS(403),
    [anon_sym_for] = ACTIONS(405),
    [anon_sym_select] = ACTIONS(407),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(409),
    [anon_sym_LT] = ACTIONS(411),
    [anon_sym_GT] = ACTIONS(411),
    [anon_sym_GT_GT] = ACTIONS(413),
    [anon_sym_LPAREN] = ACTIONS(415),
    [anon_sym_while] = ACTIONS(417),
    [anon_sym_until] = ACTIONS(417),
    [anon_sym_if] = ACTIONS(419),
    [anon_sym_case] = ACTIONS(421),
    [anon_sym_esac] = ACTIONS(423),
    [anon_sym_SEMI_SEMI] = ACTIONS(425),
    [anon_sym_SEMI_AMP] = ACTIONS(427),
    [anon_sym_SEMI_SEMI_AMP] = ACTIONS(429),
    [anon_sym_function] = ACTIONS(431),
    [anon_sym_LBRACE] = ACTIONS(433),
    [anon_sym_BANG] = ACTIONS(435),
    [anon_sym_LBRACK] = ACTIONS(437),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(439),
    [anon_sym_declare] = ACTIONS(441),
    [anon_sym_typeset] = ACTIONS(441),
    [anon_sym_export] = ACTIONS(441),
    [anon_sym_readonly] = ACTIONS(441),
    [anon_sym_local] = ACTIONS(441),
    [anon_sym_unset] = ACTIONS(443),
    [anon_sym_unsetenv] = ACTIONS(443),
    [anon_sym_AMP_GT] = ACTIONS(411),
    [anon_sym_AMP_GT_GT] = ACTIONS(413),
    [anon_sym_LT_AMP] = ACTIONS(411),
    [anon_sym_GT_AMP] = ACTIONS(411),
    [anon_sym_GT_PIPE] = ACTIONS(413),
    [anon_sym_LT_AMP_DASH] = ACTIONS(445),
    [anon_sym_GT_AMP_DASH] = ACTIONS(445),
    [anon_sym_LT_LT_LT] = ACTIONS(447),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(449),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(451),
    [anon_sym_DOLLAR] = ACTIONS(453),
    [sym__special_character] = ACTIONS(455),
    [anon_sym_DQUOTE] = ACTIONS(457),
    [sym_raw_string] = ACTIONS(459),
    [sym_ansi_c_string] = ACTIONS(459),
    [aux_sym_number_token1] = ACTIONS(461),
    [aux_sym_number_token2] = ACTIONS(463),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(465),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(467),
    [anon_sym_BQUOTE] = ACTIONS(469),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(471),
    [anon_sym_LT_LPAREN] = ACTIONS(473),
    [anon_sym_GT_LPAREN] = ACTIONS(473),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(475),
    [sym_variable_name] = ACTIONS(477),
    [sym_test_operator] = ACTIONS(479),
    [sym__brace_start] = ACTIONS(481),
  },
  [23] = {
    [sym__statements] = STATE(6539),
    [sym__statement_not_pipeline] = STATE(6584),
    [sym_redirected_statement] = STATE(4349),
    [sym_for_statement] = STATE(4349),
    [sym_c_style_for_statement] = STATE(4349),
    [sym_while_statement] = STATE(4225),
    [sym_if_statement] = STATE(4225),
    [sym_case_statement] = STATE(4349),
    [sym_function_definition] = STATE(4349),
    [sym_compound_statement] = STATE(4349),
    [sym_subshell] = STATE(4349),
    [sym_pipeline] = STATE(4620),
    [sym_list] = STATE(4349),
    [sym_negated_command] = STATE(4349),
    [sym_test_command] = STATE(4349),
    [sym_declaration_command] = STATE(4349),
    [sym_unset_command] = STATE(4349),
    [sym_command] = STATE(4349),
    [sym_command_name] = STATE(528),
    [sym_variable_assignment] = STATE(807),
    [sym_variable_assignments] = STATE(4349),
    [sym_subscript] = STATE(6736),
    [sym_file_redirect] = STATE(1635),
    [sym_herestring_redirect] = STATE(1634),
    [sym_arithmetic_expansion] = STATE(815),
    [sym_brace_expression] = STATE(815),
    [sym_concatenation] = STATE(1082),
    [sym_string] = STATE(815),
    [sym_translated_string] = STATE(815),
    [sym_number] = STATE(815),
    [sym_simple_expansion] = STATE(815),
    [sym_expansion] = STATE(815),
    [sym_command_substitution] = STATE(815),
    [sym_process_substitution] = STATE(815),
    [aux_sym__statements_repeat1] = STATE(431),
    [aux_sym_redirected_statement_repeat2] = STATE(4250),
    [aux_sym_command_repeat1] = STATE(925),
    [aux_sym__literal_repeat1] = STATE(926),
    [sym_word] = ACTIONS(403),
    [anon_sym_for] = ACTIONS(405),
    [anon_sym_select] = ACTIONS(407),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(409),
    [anon_sym_LT] = ACTIONS(411),
    [anon_sym_GT] = ACTIONS(411),
    [anon_sym_GT_GT] = ACTIONS(413),
    [anon_sym_LPAREN] = ACTIONS(415),
    [anon_sym_while] = ACTIONS(417),
    [anon_sym_until] = ACTIONS(417),
    [anon_sym_if] = ACTIONS(419),
    [anon_sym_case] = ACTIONS(421),
    [anon_sym_esac] = ACTIONS(483),
    [anon_sym_SEMI_SEMI] = ACTIONS(485),
    [anon_sym_SEMI_AMP] = ACTIONS(487),
    [anon_sym_SEMI_SEMI_AMP] = ACTIONS(487),
    [anon_sym_function] = ACTIONS(431),
    [anon_sym_LBRACE] = ACTIONS(433),
    [anon_sym_BANG] = ACTIONS(435),
    [anon_sym_LBRACK] = ACTIONS(437),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(439),
    [anon_sym_declare] = ACTIONS(441),
    [anon_sym_typeset] = ACTIONS(441),
    [anon_sym_export] = ACTIONS(441),
    [anon_sym_readonly] = ACTIONS(441),
    [anon_sym_local] = ACTIONS(441),
    [anon_sym_unset] = ACTIONS(443),
    [anon_sym_unsetenv] = ACTIONS(443),
    [anon_sym_AMP_GT] = ACTIONS(411),
    [anon_sym_AMP_GT_GT] = ACTIONS(413),
    [anon_sym_LT_AMP] = ACTIONS(411),
    [anon_sym_GT_AMP] = ACTIONS(411),
    [anon_sym_GT_PIPE] = ACTIONS(413),
    [anon_sym_LT_AMP_DASH] = ACTIONS(445),
    [anon_sym_GT_AMP_DASH] = ACTIONS(445),
    [anon_sym_LT_LT_LT] = ACTIONS(447),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(449),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(451),
    [anon_sym_DOLLAR] = ACTIONS(453),
    [sym__special_character] = ACTIONS(455),
    [anon_sym_DQUOTE] = ACTIONS(457),
    [sym_raw_string] = ACTIONS(459),
    [sym_ansi_c_string] = ACTIONS(459),
    [aux_sym_number_token1] = ACTIONS(461),
    [aux_sym_number_token2] = ACTIONS(463),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(465),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(467),
    [anon_sym_BQUOTE] = ACTIONS(469),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(471),
    [anon_sym_LT_LPAREN] = ACTIONS(473),
    [anon_sym_GT_LPAREN] = ACTIONS(473),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(475),
    [sym_variable_name] = ACTIONS(477),
    [sym_test_operator] = ACTIONS(479),
    [sym__brace_start] = ACTIONS(481),
  },
  [24] = {
    [sym__statements] = STATE(6511),
    [sym__statement_not_pipeline] = STATE(6584),
    [sym_redirected_statement] = STATE(4349),
    [sym_for_statement] = STATE(4349),
    [sym_c_style_for_statement] = STATE(4349),
    [sym_while_statement] = STATE(4225),
    [sym_if_statement] = STATE(4225),
    [sym_case_statement] = STATE(4349),
    [sym_function_definition] = STATE(4349),
    [sym_compound_statement] = STATE(4349),
    [sym_subshell] = STATE(4349),
    [sym_pipeline] = STATE(4620),
    [sym_list] = STATE(4349),
    [sym_negated_command] = STATE(4349),
    [sym_test_command] = STATE(4349),
    [sym_declaration_command] = STATE(4349),
    [sym_unset_command] = STATE(4349),
    [sym_command] = STATE(4349),
    [sym_command_name] = STATE(528),
    [sym_variable_assignment] = STATE(807),
    [sym_variable_assignments] = STATE(4349),
    [sym_subscript] = STATE(6736),
    [sym_file_redirect] = STATE(1635),
    [sym_herestring_redirect] = STATE(1634),
    [sym_arithmetic_expansion] = STATE(815),
    [sym_brace_expression] = STATE(815),
    [sym_concatenation] = STATE(1082),
    [sym_string] = STATE(815),
    [sym_translated_string] = STATE(815),
    [sym_number] = STATE(815),
    [sym_simple_expansion] = STATE(815),
    [sym_expansion] = STATE(815),
    [sym_command_substitution] = STATE(815),
    [sym_process_substitution] = STATE(815),
    [aux_sym__statements_repeat1] = STATE(431),
    [aux_sym_redirected_statement_repeat2] = STATE(4250),
    [aux_sym_command_repeat1] = STATE(925),
    [aux_sym__literal_repeat1] = STATE(926),
    [sym_word] = ACTIONS(403),
    [anon_sym_for] = ACTIONS(405),
    [anon_sym_select] = ACTIONS(407),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(409),
    [anon_sym_LT] = ACTIONS(411),
    [anon_sym_GT] = ACTIONS(411),
    [anon_sym_GT_GT] = ACTIONS(413),
    [anon_sym_LPAREN] = ACTIONS(415),
    [anon_sym_while] = ACTIONS(417),
    [anon_sym_until] = ACTIONS(417),
    [anon_sym_if] = ACTIONS(419),
    [anon_sym_case] = ACTIONS(421),
    [anon_sym_esac] = ACTIONS(489),
    [anon_sym_SEMI_SEMI] = ACTIONS(491),
    [anon_sym_SEMI_AMP] = ACTIONS(493),
    [anon_sym_SEMI_SEMI_AMP] = ACTIONS(495),
    [anon_sym_function] = ACTIONS(431),
    [anon_sym_LBRACE] = ACTIONS(433),
    [anon_sym_BANG] = ACTIONS(435),
    [anon_sym_LBRACK] = ACTIONS(437),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(439),
    [anon_sym_declare] = ACTIONS(441),
    [anon_sym_typeset] = ACTIONS(441),
    [anon_sym_export] = ACTIONS(441),
    [anon_sym_readonly] = ACTIONS(441),
    [anon_sym_local] = ACTIONS(441),
    [anon_sym_unset] = ACTIONS(443),
    [anon_sym_unsetenv] = ACTIONS(443),
    [anon_sym_AMP_GT] = ACTIONS(411),
    [anon_sym_AMP_GT_GT] = ACTIONS(413),
    [anon_sym_LT_AMP] = ACTIONS(411),
    [anon_sym_GT_AMP] = ACTIONS(411),
    [anon_sym_GT_PIPE] = ACTIONS(413),
    [anon_sym_LT_AMP_DASH] = ACTIONS(445),
    [anon_sym_GT_AMP_DASH] = ACTIONS(445),
    [anon_sym_LT_LT_LT] = ACTIONS(447),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(449),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(451),
    [anon_sym_DOLLAR] = ACTIONS(453),
    [sym__special_character] = ACTIONS(455),
    [anon_sym_DQUOTE] = ACTIONS(457),
    [sym_raw_string] = ACTIONS(459),
    [sym_ansi_c_string] = ACTIONS(459),
    [aux_sym_number_token1] = ACTIONS(461),
    [aux_sym_number_token2] = ACTIONS(463),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(465),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(467),
    [anon_sym_BQUOTE] = ACTIONS(469),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(471),
    [anon_sym_LT_LPAREN] = ACTIONS(473),
    [anon_sym_GT_LPAREN] = ACTIONS(473),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(475),
    [sym_variable_name] = ACTIONS(477),
    [sym_test_operator] = ACTIONS(479),
    [sym__brace_start] = ACTIONS(481),
  },
  [25] = {
    [sym__statements] = STATE(6492),
    [sym__statement_not_pipeline] = STATE(6584),
    [sym_redirected_statement] = STATE(4349),
    [sym_for_statement] = STATE(4349),
    [sym_c_style_for_statement] = STATE(4349),
    [sym_while_statement] = STATE(4225),
    [sym_if_statement] = STATE(4225),
    [sym_case_statement] = STATE(4349),
    [sym_function_definition] = STATE(4349),
    [sym_compound_statement] = STATE(4349),
    [sym_subshell] = STATE(4349),
    [sym_pipeline] = STATE(4620),
    [sym_list] = STATE(4349),
    [sym_negated_command] = STATE(4349),
    [sym_test_command] = STATE(4349),
    [sym_declaration_command] = STATE(4349),
    [sym_unset_command] = STATE(4349),
    [sym_command] = STATE(4349),
    [sym_command_name] = STATE(528),
    [sym_variable_assignment] = STATE(807),
    [sym_variable_assignments] = STATE(4349),
    [sym_subscript] = STATE(6736),
    [sym_file_redirect] = STATE(1635),
    [sym_herestring_redirect] = STATE(1634),
    [sym_arithmetic_expansion] = STATE(815),
    [sym_brace_expression] = STATE(815),
    [sym_concatenation] = STATE(1082),
    [sym_string] = STATE(815),
    [sym_translated_string] = STATE(815),
    [sym_number] = STATE(815),
    [sym_simple_expansion] = STATE(815),
    [sym_expansion] = STATE(815),
    [sym_command_substitution] = STATE(815),
    [sym_process_substitution] = STATE(815),
    [aux_sym__statements_repeat1] = STATE(431),
    [aux_sym_redirected_statement_repeat2] = STATE(4250),
    [aux_sym_command_repeat1] = STATE(925),
    [aux_sym__literal_repeat1] = STATE(926),
    [sym_word] = ACTIONS(403),
    [anon_sym_for] = ACTIONS(405),
    [anon_sym_select] = ACTIONS(407),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(409),
    [anon_sym_LT] = ACTIONS(411),
    [anon_sym_GT] = ACTIONS(411),
    [anon_sym_GT_GT] = ACTIONS(413),
    [anon_sym_LPAREN] = ACTIONS(415),
    [anon_sym_while] = ACTIONS(417),
    [anon_sym_until] = ACTIONS(417),
    [anon_sym_if] = ACTIONS(419),
    [anon_sym_case] = ACTIONS(421),
    [anon_sym_esac] = ACTIONS(489),
    [anon_sym_SEMI_SEMI] = ACTIONS(497),
    [anon_sym_SEMI_AMP] = ACTIONS(499),
    [anon_sym_SEMI_SEMI_AMP] = ACTIONS(501),
    [anon_sym_function] = ACTIONS(431),
    [anon_sym_LBRACE] = ACTIONS(433),
    [anon_sym_BANG] = ACTIONS(435),
    [anon_sym_LBRACK] = ACTIONS(437),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(439),
    [anon_sym_declare] = ACTIONS(441),
    [anon_sym_typeset] = ACTIONS(441),
    [anon_sym_export] = ACTIONS(441),
    [anon_sym_readonly] = ACTIONS(441),
    [anon_sym_local] = ACTIONS(441),
    [anon_sym_unset] = ACTIONS(443),
    [anon_sym_unsetenv] = ACTIONS(443),
    [anon_sym_AMP_GT] = ACTIONS(411),
    [anon_sym_AMP_GT_GT] = ACTIONS(413),
    [anon_sym_LT_AMP] = ACTIONS(411),
    [anon_sym_GT_AMP] = ACTIONS(411),
    [anon_sym_GT_PIPE] = ACTIONS(413),
    [anon_sym_LT_AMP_DASH] = ACTIONS(445),
    [anon_sym_GT_AMP_DASH] = ACTIONS(445),
    [anon_sym_LT_LT_LT] = ACTIONS(447),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(449),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(451),
    [anon_sym_DOLLAR] = ACTIONS(453),
    [sym__special_character] = ACTIONS(455),
    [anon_sym_DQUOTE] = ACTIONS(457),
    [sym_raw_string] = ACTIONS(459),
    [sym_ansi_c_string] = ACTIONS(459),
    [aux_sym_number_token1] = ACTIONS(461),
    [aux_sym_number_token2] = ACTIONS(463),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(465),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(467),
    [anon_sym_BQUOTE] = ACTIONS(469),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(471),
    [anon_sym_LT_LPAREN] = ACTIONS(473),
    [anon_sym_GT_LPAREN] = ACTIONS(473),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(475),
    [sym_variable_name] = ACTIONS(477),
    [sym_test_operator] = ACTIONS(479),
    [sym__brace_start] = ACTIONS(481),
  },
  [26] = {
    [sym__statements] = STATE(6524),
    [sym__statement_not_pipeline] = STATE(6584),
    [sym_redirected_statement] = STATE(4349),
    [sym_for_statement] = STATE(4349),
    [sym_c_style_for_statement] = STATE(4349),
    [sym_while_statement] = STATE(4225),
    [sym_if_statement] = STATE(4225),
    [sym_case_statement] = STATE(4349),
    [sym_function_definition] = STATE(4349),
    [sym_compound_statement] = STATE(4349),
    [sym_subshell] = STATE(4349),
    [sym_pipeline] = STATE(4620),
    [sym_list] = STATE(4349),
    [sym_negated_command] = STATE(4349),
    [sym_test_command] = STATE(4349),
    [sym_declaration_command] = STATE(4349),
    [sym_unset_command] = STATE(4349),
    [sym_command] = STATE(4349),
    [sym_command_name] = STATE(528),
    [sym_variable_assignment] = STATE(807),
    [sym_variable_assignments] = STATE(4349),
    [sym_subscript] = STATE(6736),
    [sym_file_redirect] = STATE(1635),
    [sym_herestring_redirect] = STATE(1634),
    [sym_arithmetic_expansion] = STATE(815),
    [sym_brace_expression] = STATE(815),
    [sym_concatenation] = STATE(1082),
    [sym_string] = STATE(815),
    [sym_translated_string] = STATE(815),
    [sym_number] = STATE(815),
    [sym_simple_expansion] = STATE(815),
    [sym_expansion] = STATE(815),
    [sym_command_substitution] = STATE(815),
    [sym_process_substitution] = STATE(815),
    [aux_sym__statements_repeat1] = STATE(431),
    [aux_sym_redirected_statement_repeat2] = STATE(4250),
    [aux_sym_command_repeat1] = STATE(925),
    [aux_sym__literal_repeat1] = STATE(926),
    [sym_word] = ACTIONS(403),
    [anon_sym_for] = ACTIONS(405),
    [anon_sym_select] = ACTIONS(407),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(409),
    [anon_sym_LT] = ACTIONS(411),
    [anon_sym_GT] = ACTIONS(411),
    [anon_sym_GT_GT] = ACTIONS(413),
    [anon_sym_LPAREN] = ACTIONS(415),
    [anon_sym_while] = ACTIONS(417),
    [anon_sym_until] = ACTIONS(417),
    [anon_sym_if] = ACTIONS(419),
    [anon_sym_case] = ACTIONS(421),
    [anon_sym_esac] = ACTIONS(503),
    [anon_sym_SEMI_SEMI] = ACTIONS(505),
    [anon_sym_SEMI_AMP] = ACTIONS(507),
    [anon_sym_SEMI_SEMI_AMP] = ACTIONS(509),
    [anon_sym_function] = ACTIONS(431),
    [anon_sym_LBRACE] = ACTIONS(433),
    [anon_sym_BANG] = ACTIONS(435),
    [anon_sym_LBRACK] = ACTIONS(437),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(439),
    [anon_sym_declare] = ACTIONS(441),
    [anon_sym_typeset] = ACTIONS(441),
    [anon_sym_export] = ACTIONS(441),
    [anon_sym_readonly] = ACTIONS(441),
    [anon_sym_local] = ACTIONS(441),
    [anon_sym_unset] = ACTIONS(443),
    [anon_sym_unsetenv] = ACTIONS(443),
    [anon_sym_AMP_GT] = ACTIONS(411),
    [anon_sym_AMP_GT_GT] = ACTIONS(413),
    [anon_sym_LT_AMP] = ACTIONS(411),
    [anon_sym_GT_AMP] = ACTIONS(411),
    [anon_sym_GT_PIPE] = ACTIONS(413),
    [anon_sym_LT_AMP_DASH] = ACTIONS(445),
    [anon_sym_GT_AMP_DASH] = ACTIONS(445),
    [anon_sym_LT_LT_LT] = ACTIONS(447),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(449),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(451),
    [anon_sym_DOLLAR] = ACTIONS(453),
    [sym__special_character] = ACTIONS(455),
    [anon_sym_DQUOTE] = ACTIONS(457),
    [sym_raw_string] = ACTIONS(459),
    [sym_ansi_c_string] = ACTIONS(459),
    [aux_sym_number_token1] = ACTIONS(461),
    [aux_sym_number_token2] = ACTIONS(463),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(465),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(467),
    [anon_sym_BQUOTE] = ACTIONS(469),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(471),
    [anon_sym_LT_LPAREN] = ACTIONS(473),
    [anon_sym_GT_LPAREN] = ACTIONS(473),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(475),
    [sym_variable_name] = ACTIONS(477),
    [sym_test_operator] = ACTIONS(479),
    [sym__brace_start] = ACTIONS(481),
  },
  [27] = {
    [sym__statements] = STATE(6496),
    [sym__statement_not_pipeline] = STATE(6584),
    [sym_redirected_statement] = STATE(4349),
    [sym_for_statement] = STATE(4349),
    [sym_c_style_for_statement] = STATE(4349),
    [sym_while_statement] = STATE(4225),
    [sym_if_statement] = STATE(4225),
    [sym_case_statement] = STATE(4349),
    [sym_function_definition] = STATE(4349),
    [sym_compound_statement] = STATE(4349),
    [sym_subshell] = STATE(4349),
    [sym_pipeline] = STATE(4620),
    [sym_list] = STATE(4349),
    [sym_negated_command] = STATE(4349),
    [sym_test_command] = STATE(4349),
    [sym_declaration_command] = STATE(4349),
    [sym_unset_command] = STATE(4349),
    [sym_command] = STATE(4349),
    [sym_command_name] = STATE(528),
    [sym_variable_assignment] = STATE(807),
    [sym_variable_assignments] = STATE(4349),
    [sym_subscript] = STATE(6736),
    [sym_file_redirect] = STATE(1635),
    [sym_herestring_redirect] = STATE(1634),
    [sym_arithmetic_expansion] = STATE(815),
    [sym_brace_expression] = STATE(815),
    [sym_concatenation] = STATE(1082),
    [sym_string] = STATE(815),
    [sym_translated_string] = STATE(815),
    [sym_number] = STATE(815),
    [sym_simple_expansion] = STATE(815),
    [sym_expansion] = STATE(815),
    [sym_command_substitution] = STATE(815),
    [sym_process_substitution] = STATE(815),
    [aux_sym__statements_repeat1] = STATE(431),
    [aux_sym_redirected_statement_repeat2] = STATE(4250),
    [aux_sym_command_repeat1] = STATE(925),
    [aux_sym__literal_repeat1] = STATE(926),
    [sym_word] = ACTIONS(403),
    [anon_sym_for] = ACTIONS(405),
    [anon_sym_select] = ACTIONS(407),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(409),
    [anon_sym_LT] = ACTIONS(411),
    [anon_sym_GT] = ACTIONS(411),
    [anon_sym_GT_GT] = ACTIONS(413),
    [anon_sym_LPAREN] = ACTIONS(415),
    [anon_sym_while] = ACTIONS(417),
    [anon_sym_until] = ACTIONS(417),
    [anon_sym_if] = ACTIONS(419),
    [anon_sym_case] = ACTIONS(421),
    [anon_sym_esac] = ACTIONS(511),
    [anon_sym_SEMI_SEMI] = ACTIONS(513),
    [anon_sym_SEMI_AMP] = ACTIONS(515),
    [anon_sym_SEMI_SEMI_AMP] = ACTIONS(517),
    [anon_sym_function] = ACTIONS(431),
    [anon_sym_LBRACE] = ACTIONS(433),
    [anon_sym_BANG] = ACTIONS(435),
    [anon_sym_LBRACK] = ACTIONS(437),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(439),
    [anon_sym_declare] = ACTIONS(441),
    [anon_sym_typeset] = ACTIONS(441),
    [anon_sym_export] = ACTIONS(441),
    [anon_sym_readonly] = ACTIONS(441),
    [anon_sym_local] = ACTIONS(441),
    [anon_sym_unset] = ACTIONS(443),
    [anon_sym_unsetenv] = ACTIONS(443),
    [anon_sym_AMP_GT] = ACTIONS(411),
    [anon_sym_AMP_GT_GT] = ACTIONS(413),
    [anon_sym_LT_AMP] = ACTIONS(411),
    [anon_sym_GT_AMP] = ACTIONS(411),
    [anon_sym_GT_PIPE] = ACTIONS(413),
    [anon_sym_LT_AMP_DASH] = ACTIONS(445),
    [anon_sym_GT_AMP_DASH] = ACTIONS(445),
    [anon_sym_LT_LT_LT] = ACTIONS(447),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(449),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(451),
    [anon_sym_DOLLAR] = ACTIONS(453),
    [sym__special_character] = ACTIONS(455),
    [anon_sym_DQUOTE] = ACTIONS(457),
    [sym_raw_string] = ACTIONS(459),
    [sym_ansi_c_string] = ACTIONS(459),
    [aux_sym_number_token1] = ACTIONS(461),
    [aux_sym_number_token2] = ACTIONS(463),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(465),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(467),
    [anon_sym_BQUOTE] = ACTIONS(469),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(471),
    [anon_sym_LT_LPAREN] = ACTIONS(473),
    [anon_sym_GT_LPAREN] = ACTIONS(473),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(475),
    [sym_variable_name] = ACTIONS(477),
    [sym_test_operator] = ACTIONS(479),
    [sym__brace_start] = ACTIONS(481),
  },
  [28] = {
    [sym__statements] = STATE(6491),
    [sym__statement_not_pipeline] = STATE(6584),
    [sym_redirected_statement] = STATE(4349),
    [sym_for_statement] = STATE(4349),
    [sym_c_style_for_statement] = STATE(4349),
    [sym_while_statement] = STATE(4225),
    [sym_if_statement] = STATE(4225),
    [sym_case_statement] = STATE(4349),
    [sym_function_definition] = STATE(4349),
    [sym_compound_statement] = STATE(4349),
    [sym_subshell] = STATE(4349),
    [sym_pipeline] = STATE(4620),
    [sym_list] = STATE(4349),
    [sym_negated_command] = STATE(4349),
    [sym_test_command] = STATE(4349),
    [sym_declaration_command] = STATE(4349),
    [sym_unset_command] = STATE(4349),
    [sym_command] = STATE(4349),
    [sym_command_name] = STATE(528),
    [sym_variable_assignment] = STATE(807),
    [sym_variable_assignments] = STATE(4349),
    [sym_subscript] = STATE(6736),
    [sym_file_redirect] = STATE(1635),
    [sym_herestring_redirect] = STATE(1634),
    [sym_arithmetic_expansion] = STATE(815),
    [sym_brace_expression] = STATE(815),
    [sym_concatenation] = STATE(1082),
    [sym_string] = STATE(815),
    [sym_translated_string] = STATE(815),
    [sym_number] = STATE(815),
    [sym_simple_expansion] = STATE(815),
    [sym_expansion] = STATE(815),
    [sym_command_substitution] = STATE(815),
    [sym_process_substitution] = STATE(815),
    [aux_sym__statements_repeat1] = STATE(431),
    [aux_sym_redirected_statement_repeat2] = STATE(4250),
    [aux_sym_command_repeat1] = STATE(925),
    [aux_sym__literal_repeat1] = STATE(926),
    [sym_word] = ACTIONS(403),
    [anon_sym_for] = ACTIONS(405),
    [anon_sym_select] = ACTIONS(407),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(409),
    [anon_sym_LT] = ACTIONS(411),
    [anon_sym_GT] = ACTIONS(411),
    [anon_sym_GT_GT] = ACTIONS(413),
    [anon_sym_LPAREN] = ACTIONS(415),
    [anon_sym_while] = ACTIONS(417),
    [anon_sym_until] = ACTIONS(417),
    [anon_sym_if] = ACTIONS(419),
    [anon_sym_case] = ACTIONS(421),
    [anon_sym_esac] = ACTIONS(511),
    [anon_sym_SEMI_SEMI] = ACTIONS(519),
    [anon_sym_SEMI_AMP] = ACTIONS(521),
    [anon_sym_SEMI_SEMI_AMP] = ACTIONS(523),
    [anon_sym_function] = ACTIONS(431),
    [anon_sym_LBRACE] = ACTIONS(433),
    [anon_sym_BANG] = ACTIONS(435),
    [anon_sym_LBRACK] = ACTIONS(437),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(439),
    [anon_sym_declare] = ACTIONS(441),
    [anon_sym_typeset] = ACTIONS(441),
    [anon_sym_export] = ACTIONS(441),
    [anon_sym_readonly] = ACTIONS(441),
    [anon_sym_local] = ACTIONS(441),
    [anon_sym_unset] = ACTIONS(443),
    [anon_sym_unsetenv] = ACTIONS(443),
    [anon_sym_AMP_GT] = ACTIONS(411),
    [anon_sym_AMP_GT_GT] = ACTIONS(413),
    [anon_sym_LT_AMP] = ACTIONS(411),
    [anon_sym_GT_AMP] = ACTIONS(411),
    [anon_sym_GT_PIPE] = ACTIONS(413),
    [anon_sym_LT_AMP_DASH] = ACTIONS(445),
    [anon_sym_GT_AMP_DASH] = ACTIONS(445),
    [anon_sym_LT_LT_LT] = ACTIONS(447),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(449),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(451),
    [anon_sym_DOLLAR] = ACTIONS(453),
    [sym__special_character] = ACTIONS(455),
    [anon_sym_DQUOTE] = ACTIONS(457),
    [sym_raw_string] = ACTIONS(459),
    [sym_ansi_c_string] = ACTIONS(459),
    [aux_sym_number_token1] = ACTIONS(461),
    [aux_sym_number_token2] = ACTIONS(463),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(465),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(467),
    [anon_sym_BQUOTE] = ACTIONS(469),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(471),
    [anon_sym_LT_LPAREN] = ACTIONS(473),
    [anon_sym_GT_LPAREN] = ACTIONS(473),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(475),
    [sym_variable_name] = ACTIONS(477),
    [sym_test_operator] = ACTIONS(479),
    [sym__brace_start] = ACTIONS(481),
  },
  [29] = {
    [sym__statements] = STATE(6542),
    [sym__statement_not_pipeline] = STATE(6584),
    [sym_redirected_statement] = STATE(4349),
    [sym_for_statement] = STATE(4349),
    [sym_c_style_for_statement] = STATE(4349),
    [sym_while_statement] = STATE(4225),
    [sym_if_statement] = STATE(4225),
    [sym_case_statement] = STATE(4349),
    [sym_function_definition] = STATE(4349),
    [sym_compound_statement] = STATE(4349),
    [sym_subshell] = STATE(4349),
    [sym_pipeline] = STATE(4620),
    [sym_list] = STATE(4349),
    [sym_negated_command] = STATE(4349),
    [sym_test_command] = STATE(4349),
    [sym_declaration_command] = STATE(4349),
    [sym_unset_command] = STATE(4349),
    [sym_command] = STATE(4349),
    [sym_command_name] = STATE(528),
    [sym_variable_assignment] = STATE(807),
    [sym_variable_assignments] = STATE(4349),
    [sym_subscript] = STATE(6736),
    [sym_file_redirect] = STATE(1635),
    [sym_herestring_redirect] = STATE(1634),
    [sym_arithmetic_expansion] = STATE(815),
    [sym_brace_expression] = STATE(815),
    [sym_concatenation] = STATE(1082),
    [sym_string] = STATE(815),
    [sym_translated_string] = STATE(815),
    [sym_number] = STATE(815),
    [sym_simple_expansion] = STATE(815),
    [sym_expansion] = STATE(815),
    [sym_command_substitution] = STATE(815),
    [sym_process_substitution] = STATE(815),
    [aux_sym__statements_repeat1] = STATE(431),
    [aux_sym_redirected_statement_repeat2] = STATE(4250),
    [aux_sym_command_repeat1] = STATE(925),
    [aux_sym__literal_repeat1] = STATE(926),
    [sym_word] = ACTIONS(403),
    [anon_sym_for] = ACTIONS(405),
    [anon_sym_select] = ACTIONS(407),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(409),
    [anon_sym_LT] = ACTIONS(411),
    [anon_sym_GT] = ACTIONS(411),
    [anon_sym_GT_GT] = ACTIONS(413),
    [anon_sym_LPAREN] = ACTIONS(415),
    [anon_sym_while] = ACTIONS(417),
    [anon_sym_until] = ACTIONS(417),
    [anon_sym_if] = ACTIONS(419),
    [anon_sym_case] = ACTIONS(421),
    [anon_sym_esac] = ACTIONS(503),
    [anon_sym_SEMI_SEMI] = ACTIONS(525),
    [anon_sym_SEMI_AMP] = ACTIONS(527),
    [anon_sym_SEMI_SEMI_AMP] = ACTIONS(529),
    [anon_sym_function] = ACTIONS(431),
    [anon_sym_LBRACE] = ACTIONS(433),
    [anon_sym_BANG] = ACTIONS(435),
    [anon_sym_LBRACK] = ACTIONS(437),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(439),
    [anon_sym_declare] = ACTIONS(441),
    [anon_sym_typeset] = ACTIONS(441),
    [anon_sym_export] = ACTIONS(441),
    [anon_sym_readonly] = ACTIONS(441),
    [anon_sym_local] = ACTIONS(441),
    [anon_sym_unset] = ACTIONS(443),
    [anon_sym_unsetenv] = ACTIONS(443),
    [anon_sym_AMP_GT] = ACTIONS(411),
    [anon_sym_AMP_GT_GT] = ACTIONS(413),
    [anon_sym_LT_AMP] = ACTIONS(411),
    [anon_sym_GT_AMP] = ACTIONS(411),
    [anon_sym_GT_PIPE] = ACTIONS(413),
    [anon_sym_LT_AMP_DASH] = ACTIONS(445),
    [anon_sym_GT_AMP_DASH] = ACTIONS(445),
    [anon_sym_LT_LT_LT] = ACTIONS(447),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(449),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(451),
    [anon_sym_DOLLAR] = ACTIONS(453),
    [sym__special_character] = ACTIONS(455),
    [anon_sym_DQUOTE] = ACTIONS(457),
    [sym_raw_string] = ACTIONS(459),
    [sym_ansi_c_string] = ACTIONS(459),
    [aux_sym_number_token1] = ACTIONS(461),
    [aux_sym_number_token2] = ACTIONS(463),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(465),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(467),
    [anon_sym_BQUOTE] = ACTIONS(469),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(471),
    [anon_sym_LT_LPAREN] = ACTIONS(473),
    [anon_sym_GT_LPAREN] = ACTIONS(473),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(475),
    [sym_variable_name] = ACTIONS(477),
    [sym_test_operator] = ACTIONS(479),
    [sym__brace_start] = ACTIONS(481),
  },
  [30] = {
    [sym__statements] = STATE(6541),
    [sym__statement_not_pipeline] = STATE(6584),
    [sym_redirected_statement] = STATE(4349),
    [sym_for_statement] = STATE(4349),
    [sym_c_style_for_statement] = STATE(4349),
    [sym_while_statement] = STATE(4225),
    [sym_if_statement] = STATE(4225),
    [sym_case_statement] = STATE(4349),
    [sym_function_definition] = STATE(4349),
    [sym_compound_statement] = STATE(4349),
    [sym_subshell] = STATE(4349),
    [sym_pipeline] = STATE(4620),
    [sym_list] = STATE(4349),
    [sym_negated_command] = STATE(4349),
    [sym_test_command] = STATE(4349),
    [sym_declaration_command] = STATE(4349),
    [sym_unset_command] = STATE(4349),
    [sym_command] = STATE(4349),
    [sym_command_name] = STATE(528),
    [sym_variable_assignment] = STATE(807),
    [sym_variable_assignments] = STATE(4349),
    [sym_subscript] = STATE(6736),
    [sym_file_redirect] = STATE(1635),
    [sym_herestring_redirect] = STATE(1634),
    [sym_arithmetic_expansion] = STATE(815),
    [sym_brace_expression] = STATE(815),
    [sym_concatenation] = STATE(1082),
    [sym_string] = STATE(815),
    [sym_translated_string] = STATE(815),
    [sym_number] = STATE(815),
    [sym_simple_expansion] = STATE(815),
    [sym_expansion] = STATE(815),
    [sym_command_substitution] = STATE(815),
    [sym_process_substitution] = STATE(815),
    [aux_sym__statements_repeat1] = STATE(431),
    [aux_sym_redirected_statement_repeat2] = STATE(4250),
    [aux_sym_command_repeat1] = STATE(925),
    [aux_sym__literal_repeat1] = STATE(926),
    [sym_word] = ACTIONS(403),
    [anon_sym_for] = ACTIONS(405),
    [anon_sym_select] = ACTIONS(407),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(409),
    [anon_sym_LT] = ACTIONS(411),
    [anon_sym_GT] = ACTIONS(411),
    [anon_sym_GT_GT] = ACTIONS(413),
    [anon_sym_LPAREN] = ACTIONS(415),
    [anon_sym_while] = ACTIONS(417),
    [anon_sym_until] = ACTIONS(417),
    [anon_sym_if] = ACTIONS(419),
    [anon_sym_case] = ACTIONS(421),
    [anon_sym_esac] = ACTIONS(531),
    [anon_sym_SEMI_SEMI] = ACTIONS(533),
    [anon_sym_SEMI_AMP] = ACTIONS(535),
    [anon_sym_SEMI_SEMI_AMP] = ACTIONS(535),
    [anon_sym_function] = ACTIONS(431),
    [anon_sym_LBRACE] = ACTIONS(433),
    [anon_sym_BANG] = ACTIONS(435),
    [anon_sym_LBRACK] = ACTIONS(437),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(439),
    [anon_sym_declare] = ACTIONS(441),
    [anon_sym_typeset] = ACTIONS(441),
    [anon_sym_export] = ACTIONS(441),
    [anon_sym_readonly] = ACTIONS(441),
    [anon_sym_local] = ACTIONS(441),
    [anon_sym_unset] = ACTIONS(443),
    [anon_sym_unsetenv] = ACTIONS(443),
    [anon_sym_AMP_GT] = ACTIONS(411),
    [anon_sym_AMP_GT_GT] = ACTIONS(413),
    [anon_sym_LT_AMP] = ACTIONS(411),
    [anon_sym_GT_AMP] = ACTIONS(411),
    [anon_sym_GT_PIPE] = ACTIONS(413),
    [anon_sym_LT_AMP_DASH] = ACTIONS(445),
    [anon_sym_GT_AMP_DASH] = ACTIONS(445),
    [anon_sym_LT_LT_LT] = ACTIONS(447),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(449),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(451),
    [anon_sym_DOLLAR] = ACTIONS(453),
    [sym__special_character] = ACTIONS(455),
    [anon_sym_DQUOTE] = ACTIONS(457),
    [sym_raw_string] = ACTIONS(459),
    [sym_ansi_c_string] = ACTIONS(459),
    [aux_sym_number_token1] = ACTIONS(461),
    [aux_sym_number_token2] = ACTIONS(463),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(465),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(467),
    [anon_sym_BQUOTE] = ACTIONS(469),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(471),
    [anon_sym_LT_LPAREN] = ACTIONS(473),
    [anon_sym_GT_LPAREN] = ACTIONS(473),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(475),
    [sym_variable_name] = ACTIONS(477),
    [sym_test_operator] = ACTIONS(479),
    [sym__brace_start] = ACTIONS(481),
  },
  [31] = {
    [sym__statements] = STATE(6536),
    [sym__statement_not_pipeline] = STATE(6584),
    [sym_redirected_statement] = STATE(4349),
    [sym_for_statement] = STATE(4349),
    [sym_c_style_for_statement] = STATE(4349),
    [sym_while_statement] = STATE(4225),
    [sym_if_statement] = STATE(4225),
    [sym_case_statement] = STATE(4349),
    [sym_function_definition] = STATE(4349),
    [sym_compound_statement] = STATE(4349),
    [sym_subshell] = STATE(4349),
    [sym_pipeline] = STATE(4620),
    [sym_list] = STATE(4349),
    [sym_negated_command] = STATE(4349),
    [sym_test_command] = STATE(4349),
    [sym_declaration_command] = STATE(4349),
    [sym_unset_command] = STATE(4349),
    [sym_command] = STATE(4349),
    [sym_command_name] = STATE(528),
    [sym_variable_assignment] = STATE(807),
    [sym_variable_assignments] = STATE(4349),
    [sym_subscript] = STATE(6736),
    [sym_file_redirect] = STATE(1635),
    [sym_herestring_redirect] = STATE(1634),
    [sym_arithmetic_expansion] = STATE(815),
    [sym_brace_expression] = STATE(815),
    [sym_concatenation] = STATE(1082),
    [sym_string] = STATE(815),
    [sym_translated_string] = STATE(815),
    [sym_number] = STATE(815),
    [sym_simple_expansion] = STATE(815),
    [sym_expansion] = STATE(815),
    [sym_command_substitution] = STATE(815),
    [sym_process_substitution] = STATE(815),
    [aux_sym__statements_repeat1] = STATE(431),
    [aux_sym_redirected_statement_repeat2] = STATE(4250),
    [aux_sym_command_repeat1] = STATE(925),
    [aux_sym__literal_repeat1] = STATE(926),
    [sym_word] = ACTIONS(403),
    [anon_sym_for] = ACTIONS(405),
    [anon_sym_select] = ACTIONS(407),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(409),
    [anon_sym_LT] = ACTIONS(411),
    [anon_sym_GT] = ACTIONS(411),
    [anon_sym_GT_GT] = ACTIONS(413),
    [anon_sym_LPAREN] = ACTIONS(415),
    [anon_sym_while] = ACTIONS(417),
    [anon_sym_until] = ACTIONS(417),
    [anon_sym_if] = ACTIONS(419),
    [anon_sym_case] = ACTIONS(421),
    [anon_sym_esac] = ACTIONS(537),
    [anon_sym_SEMI_SEMI] = ACTIONS(539),
    [anon_sym_SEMI_AMP] = ACTIONS(541),
    [anon_sym_SEMI_SEMI_AMP] = ACTIONS(541),
    [anon_sym_function] = ACTIONS(431),
    [anon_sym_LBRACE] = ACTIONS(433),
    [anon_sym_BANG] = ACTIONS(435),
    [anon_sym_LBRACK] = ACTIONS(437),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(439),
    [anon_sym_declare] = ACTIONS(441),
    [anon_sym_typeset] = ACTIONS(441),
    [anon_sym_export] = ACTIONS(441),
    [anon_sym_readonly] = ACTIONS(441),
    [anon_sym_local] = ACTIONS(441),
    [anon_sym_unset] = ACTIONS(443),
    [anon_sym_unsetenv] = ACTIONS(443),
    [anon_sym_AMP_GT] = ACTIONS(411),
    [anon_sym_AMP_GT_GT] = ACTIONS(413),
    [anon_sym_LT_AMP] = ACTIONS(411),
    [anon_sym_GT_AMP] = ACTIONS(411),
    [anon_sym_GT_PIPE] = ACTIONS(413),
    [anon_sym_LT_AMP_DASH] = ACTIONS(445),
    [anon_sym_GT_AMP_DASH] = ACTIONS(445),
    [anon_sym_LT_LT_LT] = ACTIONS(447),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(449),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(451),
    [anon_sym_DOLLAR] = ACTIONS(453),
    [sym__special_character] = ACTIONS(455),
    [anon_sym_DQUOTE] = ACTIONS(457),
    [sym_raw_string] = ACTIONS(459),
    [sym_ansi_c_string] = ACTIONS(459),
    [aux_sym_number_token1] = ACTIONS(461),
    [aux_sym_number_token2] = ACTIONS(463),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(465),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(467),
    [anon_sym_BQUOTE] = ACTIONS(469),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(471),
    [anon_sym_LT_LPAREN] = ACTIONS(473),
    [anon_sym_GT_LPAREN] = ACTIONS(473),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(475),
    [sym_variable_name] = ACTIONS(477),
    [sym_test_operator] = ACTIONS(479),
    [sym__brace_start] = ACTIONS(481),
  },
  [32] = {
    [sym__statements] = STATE(6533),
    [sym__statement_not_pipeline] = STATE(6584),
    [sym_redirected_statement] = STATE(4349),
    [sym_for_statement] = STATE(4349),
    [sym_c_style_for_statement] = STATE(4349),
    [sym_while_statement] = STATE(4225),
    [sym_if_statement] = STATE(4225),
    [sym_case_statement] = STATE(4349),
    [sym_function_definition] = STATE(4349),
    [sym_compound_statement] = STATE(4349),
    [sym_subshell] = STATE(4349),
    [sym_pipeline] = STATE(4620),
    [sym_list] = STATE(4349),
    [sym_negated_command] = STATE(4349),
    [sym_test_command] = STATE(4349),
    [sym_declaration_command] = STATE(4349),
    [sym_unset_command] = STATE(4349),
    [sym_command] = STATE(4349),
    [sym_command_name] = STATE(528),
    [sym_variable_assignment] = STATE(807),
    [sym_variable_assignments] = STATE(4349),
    [sym_subscript] = STATE(6736),
    [sym_file_redirect] = STATE(1635),
    [sym_herestring_redirect] = STATE(1634),
    [sym_arithmetic_expansion] = STATE(815),
    [sym_brace_expression] = STATE(815),
    [sym_concatenation] = STATE(1082),
    [sym_string] = STATE(815),
    [sym_translated_string] = STATE(815),
    [sym_number] = STATE(815),
    [sym_simple_expansion] = STATE(815),
    [sym_expansion] = STATE(815),
    [sym_command_substitution] = STATE(815),
    [sym_process_substitution] = STATE(815),
    [aux_sym__statements_repeat1] = STATE(431),
    [aux_sym_redirected_statement_repeat2] = STATE(4250),
    [aux_sym_command_repeat1] = STATE(925),
    [aux_sym__literal_repeat1] = STATE(926),
    [sym_word] = ACTIONS(403),
    [anon_sym_for] = ACTIONS(405),
    [anon_sym_select] = ACTIONS(407),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(409),
    [anon_sym_LT] = ACTIONS(411),
    [anon_sym_GT] = ACTIONS(411),
    [anon_sym_GT_GT] = ACTIONS(413),
    [anon_sym_LPAREN] = ACTIONS(415),
    [anon_sym_while] = ACTIONS(417),
    [anon_sym_until] = ACTIONS(417),
    [anon_sym_if] = ACTIONS(419),
    [anon_sym_case] = ACTIONS(421),
    [anon_sym_esac] = ACTIONS(543),
    [anon_sym_SEMI_SEMI] = ACTIONS(545),
    [anon_sym_SEMI_AMP] = ACTIONS(547),
    [anon_sym_SEMI_SEMI_AMP] = ACTIONS(547),
    [anon_sym_function] = ACTIONS(431),
    [anon_sym_LBRACE] = ACTIONS(433),
    [anon_sym_BANG] = ACTIONS(435),
    [anon_sym_LBRACK] = ACTIONS(437),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(439),
    [anon_sym_declare] = ACTIONS(441),
    [anon_sym_typeset] = ACTIONS(441),
    [anon_sym_export] = ACTIONS(441),
    [anon_sym_readonly] = ACTIONS(441),
    [anon_sym_local] = ACTIONS(441),
    [anon_sym_unset] = ACTIONS(443),
    [anon_sym_unsetenv] = ACTIONS(443),
    [anon_sym_AMP_GT] = ACTIONS(411),
    [anon_sym_AMP_GT_GT] = ACTIONS(413),
    [anon_sym_LT_AMP] = ACTIONS(411),
    [anon_sym_GT_AMP] = ACTIONS(411),
    [anon_sym_GT_PIPE] = ACTIONS(413),
    [anon_sym_LT_AMP_DASH] = ACTIONS(445),
    [anon_sym_GT_AMP_DASH] = ACTIONS(445),
    [anon_sym_LT_LT_LT] = ACTIONS(447),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(449),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(451),
    [anon_sym_DOLLAR] = ACTIONS(453),
    [sym__special_character] = ACTIONS(455),
    [anon_sym_DQUOTE] = ACTIONS(457),
    [sym_raw_string] = ACTIONS(459),
    [sym_ansi_c_string] = ACTIONS(459),
    [aux_sym_number_token1] = ACTIONS(461),
    [aux_sym_number_token2] = ACTIONS(463),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(465),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(467),
    [anon_sym_BQUOTE] = ACTIONS(469),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(471),
    [anon_sym_LT_LPAREN] = ACTIONS(473),
    [anon_sym_GT_LPAREN] = ACTIONS(473),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(475),
    [sym_variable_name] = ACTIONS(477),
    [sym_test_operator] = ACTIONS(479),
    [sym__brace_start] = ACTIONS(481),
  },
  [33] = {
    [sym__statements] = STATE(6521),
    [sym__statement_not_pipeline] = STATE(6584),
    [sym_redirected_statement] = STATE(4349),
    [sym_for_statement] = STATE(4349),
    [sym_c_style_for_statement] = STATE(4349),
    [sym_while_statement] = STATE(4225),
    [sym_if_statement] = STATE(4225),
    [sym_case_statement] = STATE(4349),
    [sym_function_definition] = STATE(4349),
    [sym_compound_statement] = STATE(4349),
    [sym_subshell] = STATE(4349),
    [sym_pipeline] = STATE(4620),
    [sym_list] = STATE(4349),
    [sym_negated_command] = STATE(4349),
    [sym_test_command] = STATE(4349),
    [sym_declaration_command] = STATE(4349),
    [sym_unset_command] = STATE(4349),
    [sym_command] = STATE(4349),
    [sym_command_name] = STATE(528),
    [sym_variable_assignment] = STATE(807),
    [sym_variable_assignments] = STATE(4349),
    [sym_subscript] = STATE(6736),
    [sym_file_redirect] = STATE(1635),
    [sym_herestring_redirect] = STATE(1634),
    [sym_arithmetic_expansion] = STATE(815),
    [sym_brace_expression] = STATE(815),
    [sym_concatenation] = STATE(1082),
    [sym_string] = STATE(815),
    [sym_translated_string] = STATE(815),
    [sym_number] = STATE(815),
    [sym_simple_expansion] = STATE(815),
    [sym_expansion] = STATE(815),
    [sym_command_substitution] = STATE(815),
    [sym_process_substitution] = STATE(815),
    [aux_sym__statements_repeat1] = STATE(431),
    [aux_sym_redirected_statement_repeat2] = STATE(4250),
    [aux_sym_command_repeat1] = STATE(925),
    [aux_sym__literal_repeat1] = STATE(926),
    [sym_word] = ACTIONS(403),
    [anon_sym_for] = ACTIONS(405),
    [anon_sym_select] = ACTIONS(407),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(409),
    [anon_sym_LT] = ACTIONS(411),
    [anon_sym_GT] = ACTIONS(411),
    [anon_sym_GT_GT] = ACTIONS(413),
    [anon_sym_LPAREN] = ACTIONS(415),
    [anon_sym_while] = ACTIONS(417),
    [anon_sym_until] = ACTIONS(417),
    [anon_sym_if] = ACTIONS(419),
    [anon_sym_case] = ACTIONS(421),
    [anon_sym_esac] = ACTIONS(423),
    [anon_sym_SEMI_SEMI] = ACTIONS(549),
    [anon_sym_SEMI_AMP] = ACTIONS(551),
    [anon_sym_SEMI_SEMI_AMP] = ACTIONS(553),
    [anon_sym_function] = ACTIONS(431),
    [anon_sym_LBRACE] = ACTIONS(433),
    [anon_sym_BANG] = ACTIONS(435),
    [anon_sym_LBRACK] = ACTIONS(437),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(439),
    [anon_sym_declare] = ACTIONS(441),
    [anon_sym_typeset] = ACTIONS(441),
    [anon_sym_export] = ACTIONS(441),
    [anon_sym_readonly] = ACTIONS(441),
    [anon_sym_local] = ACTIONS(441),
    [anon_sym_unset] = ACTIONS(443),
    [anon_sym_unsetenv] = ACTIONS(443),
    [anon_sym_AMP_GT] = ACTIONS(411),
    [anon_sym_AMP_GT_GT] = ACTIONS(413),
    [anon_sym_LT_AMP] = ACTIONS(411),
    [anon_sym_GT_AMP] = ACTIONS(411),
    [anon_sym_GT_PIPE] = ACTIONS(413),
    [anon_sym_LT_AMP_DASH] = ACTIONS(445),
    [anon_sym_GT_AMP_DASH] = ACTIONS(445),
    [anon_sym_LT_LT_LT] = ACTIONS(447),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(449),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(451),
    [anon_sym_DOLLAR] = ACTIONS(453),
    [sym__special_character] = ACTIONS(455),
    [anon_sym_DQUOTE] = ACTIONS(457),
    [sym_raw_string] = ACTIONS(459),
    [sym_ansi_c_string] = ACTIONS(459),
    [aux_sym_number_token1] = ACTIONS(461),
    [aux_sym_number_token2] = ACTIONS(463),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(465),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(467),
    [anon_sym_BQUOTE] = ACTIONS(469),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(471),
    [anon_sym_LT_LPAREN] = ACTIONS(473),
    [anon_sym_GT_LPAREN] = ACTIONS(473),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(475),
    [sym_variable_name] = ACTIONS(477),
    [sym_test_operator] = ACTIONS(479),
    [sym__brace_start] = ACTIONS(481),
  },
  [34] = {
    [aux_sym__terminated_statement] = STATE(34),
    [sym__statement_not_pipeline] = STATE(6718),
    [sym_redirected_statement] = STATE(5007),
    [sym_for_statement] = STATE(5007),
    [sym_c_style_for_statement] = STATE(5007),
    [sym_while_statement] = STATE(4551),
    [sym_if_statement] = STATE(4551),
    [sym_case_statement] = STATE(5007),
    [sym_function_definition] = STATE(5007),
    [sym_compound_statement] = STATE(5007),
    [sym_subshell] = STATE(5007),
    [sym_pipeline] = STATE(5189),
    [sym_list] = STATE(5007),
    [sym_negated_command] = STATE(5007),
    [sym_test_command] = STATE(5007),
    [sym_declaration_command] = STATE(5007),
    [sym_unset_command] = STATE(5007),
    [sym_command] = STATE(5007),
    [sym_command_name] = STATE(544),
    [sym_variable_assignment] = STATE(1151),
    [sym_variable_assignments] = STATE(5007),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym_redirected_statement_repeat2] = STATE(4644),
    [aux_sym_command_repeat1] = STATE(997),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(555),
    [anon_sym_for] = ACTIONS(558),
    [anon_sym_select] = ACTIONS(561),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(564),
    [anon_sym_LT] = ACTIONS(567),
    [anon_sym_GT] = ACTIONS(567),
    [anon_sym_GT_GT] = ACTIONS(570),
    [anon_sym_LPAREN] = ACTIONS(573),
    [anon_sym_while] = ACTIONS(576),
    [anon_sym_until] = ACTIONS(576),
    [anon_sym_do] = ACTIONS(579),
    [anon_sym_if] = ACTIONS(581),
    [anon_sym_then] = ACTIONS(579),
    [anon_sym_fi] = ACTIONS(579),
    [anon_sym_elif] = ACTIONS(579),
    [anon_sym_else] = ACTIONS(579),
    [anon_sym_case] = ACTIONS(584),
    [anon_sym_function] = ACTIONS(587),
    [anon_sym_LBRACE] = ACTIONS(590),
    [anon_sym_BANG] = ACTIONS(593),
    [anon_sym_LBRACK] = ACTIONS(596),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(599),
    [anon_sym_declare] = ACTIONS(602),
    [anon_sym_typeset] = ACTIONS(602),
    [anon_sym_export] = ACTIONS(602),
    [anon_sym_readonly] = ACTIONS(602),
    [anon_sym_local] = ACTIONS(602),
    [anon_sym_unset] = ACTIONS(605),
    [anon_sym_unsetenv] = ACTIONS(605),
    [anon_sym_AMP_GT] = ACTIONS(567),
    [anon_sym_AMP_GT_GT] = ACTIONS(570),
    [anon_sym_LT_AMP] = ACTIONS(567),
    [anon_sym_GT_AMP] = ACTIONS(567),
    [anon_sym_GT_PIPE] = ACTIONS(570),
    [anon_sym_LT_AMP_DASH] = ACTIONS(608),
    [anon_sym_GT_AMP_DASH] = ACTIONS(608),
    [anon_sym_LT_LT_LT] = ACTIONS(611),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(614),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(617),
    [anon_sym_DOLLAR] = ACTIONS(620),
    [sym__special_character] = ACTIONS(623),
    [anon_sym_DQUOTE] = ACTIONS(626),
    [sym_raw_string] = ACTIONS(629),
    [sym_ansi_c_string] = ACTIONS(629),
    [aux_sym_number_token1] = ACTIONS(632),
    [aux_sym_number_token2] = ACTIONS(635),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(638),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(644),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(647),
    [anon_sym_LT_LPAREN] = ACTIONS(650),
    [anon_sym_GT_LPAREN] = ACTIONS(650),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(653),
    [sym_variable_name] = ACTIONS(656),
    [sym_test_operator] = ACTIONS(659),
    [sym__brace_start] = ACTIONS(662),
  },
  [35] = {
    [sym__statements] = STATE(6649),
    [sym__statement_not_pipeline] = STATE(6674),
    [sym_redirected_statement] = STATE(4416),
    [sym_for_statement] = STATE(4416),
    [sym_c_style_for_statement] = STATE(4416),
    [sym_while_statement] = STATE(4239),
    [sym_if_statement] = STATE(4239),
    [sym_case_statement] = STATE(4416),
    [sym_function_definition] = STATE(4416),
    [sym_compound_statement] = STATE(4416),
    [sym_subshell] = STATE(4416),
    [sym_pipeline] = STATE(4910),
    [sym_list] = STATE(4416),
    [sym_negated_command] = STATE(4416),
    [sym_test_command] = STATE(4416),
    [sym_declaration_command] = STATE(4416),
    [sym_unset_command] = STATE(4416),
    [sym_command] = STATE(4416),
    [sym_command_name] = STATE(530),
    [sym_variable_assignment] = STATE(849),
    [sym_variable_assignments] = STATE(4416),
    [sym_subscript] = STATE(6760),
    [sym_file_redirect] = STATE(1893),
    [sym_herestring_redirect] = STATE(1892),
    [sym_arithmetic_expansion] = STATE(852),
    [sym_brace_expression] = STATE(852),
    [sym_concatenation] = STATE(1169),
    [sym_string] = STATE(852),
    [sym_translated_string] = STATE(852),
    [sym_number] = STATE(852),
    [sym_simple_expansion] = STATE(852),
    [sym_expansion] = STATE(852),
    [sym_command_substitution] = STATE(852),
    [sym_process_substitution] = STATE(852),
    [aux_sym__statements_repeat1] = STATE(426),
    [aux_sym_redirected_statement_repeat2] = STATE(4371),
    [aux_sym_command_repeat1] = STATE(989),
    [aux_sym__literal_repeat1] = STATE(1026),
    [sym_word] = ACTIONS(665),
    [anon_sym_for] = ACTIONS(667),
    [anon_sym_select] = ACTIONS(669),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(671),
    [anon_sym_LT] = ACTIONS(673),
    [anon_sym_GT] = ACTIONS(673),
    [anon_sym_GT_GT] = ACTIONS(675),
    [anon_sym_LPAREN] = ACTIONS(677),
    [anon_sym_while] = ACTIONS(679),
    [anon_sym_until] = ACTIONS(679),
    [anon_sym_if] = ACTIONS(681),
    [anon_sym_case] = ACTIONS(683),
    [anon_sym_SEMI_SEMI] = ACTIONS(685),
    [anon_sym_SEMI_AMP] = ACTIONS(535),
    [anon_sym_SEMI_SEMI_AMP] = ACTIONS(535),
    [anon_sym_function] = ACTIONS(687),
    [anon_sym_LBRACE] = ACTIONS(689),
    [anon_sym_BANG] = ACTIONS(691),
    [anon_sym_LBRACK] = ACTIONS(693),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(695),
    [anon_sym_declare] = ACTIONS(697),
    [anon_sym_typeset] = ACTIONS(697),
    [anon_sym_export] = ACTIONS(697),
    [anon_sym_readonly] = ACTIONS(697),
    [anon_sym_local] = ACTIONS(697),
    [anon_sym_unset] = ACTIONS(699),
    [anon_sym_unsetenv] = ACTIONS(699),
    [anon_sym_AMP_GT] = ACTIONS(673),
    [anon_sym_AMP_GT_GT] = ACTIONS(675),
    [anon_sym_LT_AMP] = ACTIONS(673),
    [anon_sym_GT_AMP] = ACTIONS(673),
    [anon_sym_GT_PIPE] = ACTIONS(675),
    [anon_sym_LT_AMP_DASH] = ACTIONS(701),
    [anon_sym_GT_AMP_DASH] = ACTIONS(701),
    [anon_sym_LT_LT_LT] = ACTIONS(703),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(705),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(707),
    [anon_sym_DOLLAR] = ACTIONS(709),
    [sym__special_character] = ACTIONS(711),
    [anon_sym_DQUOTE] = ACTIONS(713),
    [sym_raw_string] = ACTIONS(715),
    [sym_ansi_c_string] = ACTIONS(715),
    [aux_sym_number_token1] = ACTIONS(717),
    [aux_sym_number_token2] = ACTIONS(719),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(721),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(723),
    [anon_sym_BQUOTE] = ACTIONS(725),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(727),
    [anon_sym_LT_LPAREN] = ACTIONS(729),
    [anon_sym_GT_LPAREN] = ACTIONS(729),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(731),
    [sym_variable_name] = ACTIONS(733),
    [sym_test_operator] = ACTIONS(735),
    [sym__brace_start] = ACTIONS(737),
  },
  [36] = {
    [sym__statements] = STATE(6620),
    [sym__statement_not_pipeline] = STATE(6674),
    [sym_redirected_statement] = STATE(4416),
    [sym_for_statement] = STATE(4416),
    [sym_c_style_for_statement] = STATE(4416),
    [sym_while_statement] = STATE(4239),
    [sym_if_statement] = STATE(4239),
    [sym_case_statement] = STATE(4416),
    [sym_function_definition] = STATE(4416),
    [sym_compound_statement] = STATE(4416),
    [sym_subshell] = STATE(4416),
    [sym_pipeline] = STATE(4910),
    [sym_list] = STATE(4416),
    [sym_negated_command] = STATE(4416),
    [sym_test_command] = STATE(4416),
    [sym_declaration_command] = STATE(4416),
    [sym_unset_command] = STATE(4416),
    [sym_command] = STATE(4416),
    [sym_command_name] = STATE(530),
    [sym_variable_assignment] = STATE(849),
    [sym_variable_assignments] = STATE(4416),
    [sym_subscript] = STATE(6760),
    [sym_file_redirect] = STATE(1893),
    [sym_herestring_redirect] = STATE(1892),
    [sym_arithmetic_expansion] = STATE(852),
    [sym_brace_expression] = STATE(852),
    [sym_concatenation] = STATE(1169),
    [sym_string] = STATE(852),
    [sym_translated_string] = STATE(852),
    [sym_number] = STATE(852),
    [sym_simple_expansion] = STATE(852),
    [sym_expansion] = STATE(852),
    [sym_command_substitution] = STATE(852),
    [sym_process_substitution] = STATE(852),
    [aux_sym__statements_repeat1] = STATE(426),
    [aux_sym_redirected_statement_repeat2] = STATE(4371),
    [aux_sym_command_repeat1] = STATE(989),
    [aux_sym__literal_repeat1] = STATE(1026),
    [sym_word] = ACTIONS(665),
    [anon_sym_for] = ACTIONS(667),
    [anon_sym_select] = ACTIONS(669),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(671),
    [anon_sym_LT] = ACTIONS(673),
    [anon_sym_GT] = ACTIONS(673),
    [anon_sym_GT_GT] = ACTIONS(675),
    [anon_sym_LPAREN] = ACTIONS(677),
    [anon_sym_while] = ACTIONS(679),
    [anon_sym_until] = ACTIONS(679),
    [anon_sym_if] = ACTIONS(681),
    [anon_sym_case] = ACTIONS(683),
    [anon_sym_SEMI_SEMI] = ACTIONS(739),
    [anon_sym_SEMI_AMP] = ACTIONS(499),
    [anon_sym_SEMI_SEMI_AMP] = ACTIONS(501),
    [anon_sym_function] = ACTIONS(687),
    [anon_sym_LBRACE] = ACTIONS(689),
    [anon_sym_BANG] = ACTIONS(691),
    [anon_sym_LBRACK] = ACTIONS(693),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(695),
    [anon_sym_declare] = ACTIONS(697),
    [anon_sym_typeset] = ACTIONS(697),
    [anon_sym_export] = ACTIONS(697),
    [anon_sym_readonly] = ACTIONS(697),
    [anon_sym_local] = ACTIONS(697),
    [anon_sym_unset] = ACTIONS(699),
    [anon_sym_unsetenv] = ACTIONS(699),
    [anon_sym_AMP_GT] = ACTIONS(673),
    [anon_sym_AMP_GT_GT] = ACTIONS(675),
    [anon_sym_LT_AMP] = ACTIONS(673),
    [anon_sym_GT_AMP] = ACTIONS(673),
    [anon_sym_GT_PIPE] = ACTIONS(675),
    [anon_sym_LT_AMP_DASH] = ACTIONS(701),
    [anon_sym_GT_AMP_DASH] = ACTIONS(701),
    [anon_sym_LT_LT_LT] = ACTIONS(703),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(705),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(707),
    [anon_sym_DOLLAR] = ACTIONS(709),
    [sym__special_character] = ACTIONS(711),
    [anon_sym_DQUOTE] = ACTIONS(713),
    [sym_raw_string] = ACTIONS(715),
    [sym_ansi_c_string] = ACTIONS(715),
    [aux_sym_number_token1] = ACTIONS(717),
    [aux_sym_number_token2] = ACTIONS(719),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(721),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(723),
    [anon_sym_BQUOTE] = ACTIONS(725),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(727),
    [anon_sym_LT_LPAREN] = ACTIONS(729),
    [anon_sym_GT_LPAREN] = ACTIONS(729),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(731),
    [sym_variable_name] = ACTIONS(733),
    [sym_test_operator] = ACTIONS(735),
    [sym__brace_start] = ACTIONS(737),
  },
  [37] = {
    [sym__statements] = STATE(6722),
    [sym__statement_not_pipeline] = STATE(6674),
    [sym_redirected_statement] = STATE(4416),
    [sym_for_statement] = STATE(4416),
    [sym_c_style_for_statement] = STATE(4416),
    [sym_while_statement] = STATE(4239),
    [sym_if_statement] = STATE(4239),
    [sym_case_statement] = STATE(4416),
    [sym_function_definition] = STATE(4416),
    [sym_compound_statement] = STATE(4416),
    [sym_subshell] = STATE(4416),
    [sym_pipeline] = STATE(4910),
    [sym_list] = STATE(4416),
    [sym_negated_command] = STATE(4416),
    [sym_test_command] = STATE(4416),
    [sym_declaration_command] = STATE(4416),
    [sym_unset_command] = STATE(4416),
    [sym_command] = STATE(4416),
    [sym_command_name] = STATE(530),
    [sym_variable_assignment] = STATE(849),
    [sym_variable_assignments] = STATE(4416),
    [sym_subscript] = STATE(6760),
    [sym_file_redirect] = STATE(1893),
    [sym_herestring_redirect] = STATE(1892),
    [sym_arithmetic_expansion] = STATE(852),
    [sym_brace_expression] = STATE(852),
    [sym_concatenation] = STATE(1169),
    [sym_string] = STATE(852),
    [sym_translated_string] = STATE(852),
    [sym_number] = STATE(852),
    [sym_simple_expansion] = STATE(852),
    [sym_expansion] = STATE(852),
    [sym_command_substitution] = STATE(852),
    [sym_process_substitution] = STATE(852),
    [aux_sym__statements_repeat1] = STATE(426),
    [aux_sym_redirected_statement_repeat2] = STATE(4371),
    [aux_sym_command_repeat1] = STATE(989),
    [aux_sym__literal_repeat1] = STATE(1026),
    [sym_word] = ACTIONS(665),
    [anon_sym_for] = ACTIONS(667),
    [anon_sym_select] = ACTIONS(669),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(671),
    [anon_sym_LT] = ACTIONS(673),
    [anon_sym_GT] = ACTIONS(673),
    [anon_sym_GT_GT] = ACTIONS(675),
    [anon_sym_LPAREN] = ACTIONS(677),
    [anon_sym_while] = ACTIONS(679),
    [anon_sym_until] = ACTIONS(679),
    [anon_sym_if] = ACTIONS(681),
    [anon_sym_case] = ACTIONS(683),
    [anon_sym_SEMI_SEMI] = ACTIONS(741),
    [anon_sym_SEMI_AMP] = ACTIONS(515),
    [anon_sym_SEMI_SEMI_AMP] = ACTIONS(517),
    [anon_sym_function] = ACTIONS(687),
    [anon_sym_LBRACE] = ACTIONS(689),
    [anon_sym_BANG] = ACTIONS(691),
    [anon_sym_LBRACK] = ACTIONS(693),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(695),
    [anon_sym_declare] = ACTIONS(697),
    [anon_sym_typeset] = ACTIONS(697),
    [anon_sym_export] = ACTIONS(697),
    [anon_sym_readonly] = ACTIONS(697),
    [anon_sym_local] = ACTIONS(697),
    [anon_sym_unset] = ACTIONS(699),
    [anon_sym_unsetenv] = ACTIONS(699),
    [anon_sym_AMP_GT] = ACTIONS(673),
    [anon_sym_AMP_GT_GT] = ACTIONS(675),
    [anon_sym_LT_AMP] = ACTIONS(673),
    [anon_sym_GT_AMP] = ACTIONS(673),
    [anon_sym_GT_PIPE] = ACTIONS(675),
    [anon_sym_LT_AMP_DASH] = ACTIONS(701),
    [anon_sym_GT_AMP_DASH] = ACTIONS(701),
    [anon_sym_LT_LT_LT] = ACTIONS(703),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(705),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(707),
    [anon_sym_DOLLAR] = ACTIONS(709),
    [sym__special_character] = ACTIONS(711),
    [anon_sym_DQUOTE] = ACTIONS(713),
    [sym_raw_string] = ACTIONS(715),
    [sym_ansi_c_string] = ACTIONS(715),
    [aux_sym_number_token1] = ACTIONS(717),
    [aux_sym_number_token2] = ACTIONS(719),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(721),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(723),
    [anon_sym_BQUOTE] = ACTIONS(725),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(727),
    [anon_sym_LT_LPAREN] = ACTIONS(729),
    [anon_sym_GT_LPAREN] = ACTIONS(729),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(731),
    [sym_variable_name] = ACTIONS(733),
    [sym_test_operator] = ACTIONS(735),
    [sym__brace_start] = ACTIONS(737),
  },
  [38] = {
    [sym__statements] = STATE(6719),
    [sym__statement_not_pipeline] = STATE(6674),
    [sym_redirected_statement] = STATE(4416),
    [sym_for_statement] = STATE(4416),
    [sym_c_style_for_statement] = STATE(4416),
    [sym_while_statement] = STATE(4239),
    [sym_if_statement] = STATE(4239),
    [sym_case_statement] = STATE(4416),
    [sym_function_definition] = STATE(4416),
    [sym_compound_statement] = STATE(4416),
    [sym_subshell] = STATE(4416),
    [sym_pipeline] = STATE(4910),
    [sym_list] = STATE(4416),
    [sym_negated_command] = STATE(4416),
    [sym_test_command] = STATE(4416),
    [sym_declaration_command] = STATE(4416),
    [sym_unset_command] = STATE(4416),
    [sym_command] = STATE(4416),
    [sym_command_name] = STATE(530),
    [sym_variable_assignment] = STATE(849),
    [sym_variable_assignments] = STATE(4416),
    [sym_subscript] = STATE(6760),
    [sym_file_redirect] = STATE(1893),
    [sym_herestring_redirect] = STATE(1892),
    [sym_arithmetic_expansion] = STATE(852),
    [sym_brace_expression] = STATE(852),
    [sym_concatenation] = STATE(1169),
    [sym_string] = STATE(852),
    [sym_translated_string] = STATE(852),
    [sym_number] = STATE(852),
    [sym_simple_expansion] = STATE(852),
    [sym_expansion] = STATE(852),
    [sym_command_substitution] = STATE(852),
    [sym_process_substitution] = STATE(852),
    [aux_sym__statements_repeat1] = STATE(426),
    [aux_sym_redirected_statement_repeat2] = STATE(4371),
    [aux_sym_command_repeat1] = STATE(989),
    [aux_sym__literal_repeat1] = STATE(1026),
    [sym_word] = ACTIONS(665),
    [anon_sym_for] = ACTIONS(667),
    [anon_sym_select] = ACTIONS(669),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(671),
    [anon_sym_LT] = ACTIONS(673),
    [anon_sym_GT] = ACTIONS(673),
    [anon_sym_GT_GT] = ACTIONS(675),
    [anon_sym_LPAREN] = ACTIONS(677),
    [anon_sym_while] = ACTIONS(679),
    [anon_sym_until] = ACTIONS(679),
    [anon_sym_if] = ACTIONS(681),
    [anon_sym_case] = ACTIONS(683),
    [anon_sym_SEMI_SEMI] = ACTIONS(743),
    [anon_sym_SEMI_AMP] = ACTIONS(427),
    [anon_sym_SEMI_SEMI_AMP] = ACTIONS(429),
    [anon_sym_function] = ACTIONS(687),
    [anon_sym_LBRACE] = ACTIONS(689),
    [anon_sym_BANG] = ACTIONS(691),
    [anon_sym_LBRACK] = ACTIONS(693),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(695),
    [anon_sym_declare] = ACTIONS(697),
    [anon_sym_typeset] = ACTIONS(697),
    [anon_sym_export] = ACTIONS(697),
    [anon_sym_readonly] = ACTIONS(697),
    [anon_sym_local] = ACTIONS(697),
    [anon_sym_unset] = ACTIONS(699),
    [anon_sym_unsetenv] = ACTIONS(699),
    [anon_sym_AMP_GT] = ACTIONS(673),
    [anon_sym_AMP_GT_GT] = ACTIONS(675),
    [anon_sym_LT_AMP] = ACTIONS(673),
    [anon_sym_GT_AMP] = ACTIONS(673),
    [anon_sym_GT_PIPE] = ACTIONS(675),
    [anon_sym_LT_AMP_DASH] = ACTIONS(701),
    [anon_sym_GT_AMP_DASH] = ACTIONS(701),
    [anon_sym_LT_LT_LT] = ACTIONS(703),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(705),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(707),
    [anon_sym_DOLLAR] = ACTIONS(709),
    [sym__special_character] = ACTIONS(711),
    [anon_sym_DQUOTE] = ACTIONS(713),
    [sym_raw_string] = ACTIONS(715),
    [sym_ansi_c_string] = ACTIONS(715),
    [aux_sym_number_token1] = ACTIONS(717),
    [aux_sym_number_token2] = ACTIONS(719),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(721),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(723),
    [anon_sym_BQUOTE] = ACTIONS(725),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(727),
    [anon_sym_LT_LPAREN] = ACTIONS(729),
    [anon_sym_GT_LPAREN] = ACTIONS(729),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(731),
    [sym_variable_name] = ACTIONS(733),
    [sym_test_operator] = ACTIONS(735),
    [sym__brace_start] = ACTIONS(737),
  },
  [39] = {
    [sym__statements] = STATE(6716),
    [sym__statement_not_pipeline] = STATE(6674),
    [sym_redirected_statement] = STATE(4416),
    [sym_for_statement] = STATE(4416),
    [sym_c_style_for_statement] = STATE(4416),
    [sym_while_statement] = STATE(4239),
    [sym_if_statement] = STATE(4239),
    [sym_case_statement] = STATE(4416),
    [sym_function_definition] = STATE(4416),
    [sym_compound_statement] = STATE(4416),
    [sym_subshell] = STATE(4416),
    [sym_pipeline] = STATE(4910),
    [sym_list] = STATE(4416),
    [sym_negated_command] = STATE(4416),
    [sym_test_command] = STATE(4416),
    [sym_declaration_command] = STATE(4416),
    [sym_unset_command] = STATE(4416),
    [sym_command] = STATE(4416),
    [sym_command_name] = STATE(530),
    [sym_variable_assignment] = STATE(849),
    [sym_variable_assignments] = STATE(4416),
    [sym_subscript] = STATE(6760),
    [sym_file_redirect] = STATE(1893),
    [sym_herestring_redirect] = STATE(1892),
    [sym_arithmetic_expansion] = STATE(852),
    [sym_brace_expression] = STATE(852),
    [sym_concatenation] = STATE(1169),
    [sym_string] = STATE(852),
    [sym_translated_string] = STATE(852),
    [sym_number] = STATE(852),
    [sym_simple_expansion] = STATE(852),
    [sym_expansion] = STATE(852),
    [sym_command_substitution] = STATE(852),
    [sym_process_substitution] = STATE(852),
    [aux_sym__statements_repeat1] = STATE(426),
    [aux_sym_redirected_statement_repeat2] = STATE(4371),
    [aux_sym_command_repeat1] = STATE(989),
    [aux_sym__literal_repeat1] = STATE(1026),
    [sym_word] = ACTIONS(665),
    [anon_sym_for] = ACTIONS(667),
    [anon_sym_select] = ACTIONS(669),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(671),
    [anon_sym_LT] = ACTIONS(673),
    [anon_sym_GT] = ACTIONS(673),
    [anon_sym_GT_GT] = ACTIONS(675),
    [anon_sym_LPAREN] = ACTIONS(677),
    [anon_sym_while] = ACTIONS(679),
    [anon_sym_until] = ACTIONS(679),
    [anon_sym_if] = ACTIONS(681),
    [anon_sym_case] = ACTIONS(683),
    [anon_sym_SEMI_SEMI] = ACTIONS(745),
    [anon_sym_SEMI_AMP] = ACTIONS(551),
    [anon_sym_SEMI_SEMI_AMP] = ACTIONS(553),
    [anon_sym_function] = ACTIONS(687),
    [anon_sym_LBRACE] = ACTIONS(689),
    [anon_sym_BANG] = ACTIONS(691),
    [anon_sym_LBRACK] = ACTIONS(693),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(695),
    [anon_sym_declare] = ACTIONS(697),
    [anon_sym_typeset] = ACTIONS(697),
    [anon_sym_export] = ACTIONS(697),
    [anon_sym_readonly] = ACTIONS(697),
    [anon_sym_local] = ACTIONS(697),
    [anon_sym_unset] = ACTIONS(699),
    [anon_sym_unsetenv] = ACTIONS(699),
    [anon_sym_AMP_GT] = ACTIONS(673),
    [anon_sym_AMP_GT_GT] = ACTIONS(675),
    [anon_sym_LT_AMP] = ACTIONS(673),
    [anon_sym_GT_AMP] = ACTIONS(673),
    [anon_sym_GT_PIPE] = ACTIONS(675),
    [anon_sym_LT_AMP_DASH] = ACTIONS(701),
    [anon_sym_GT_AMP_DASH] = ACTIONS(701),
    [anon_sym_LT_LT_LT] = ACTIONS(703),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(705),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(707),
    [anon_sym_DOLLAR] = ACTIONS(709),
    [sym__special_character] = ACTIONS(711),
    [anon_sym_DQUOTE] = ACTIONS(713),
    [sym_raw_string] = ACTIONS(715),
    [sym_ansi_c_string] = ACTIONS(715),
    [aux_sym_number_token1] = ACTIONS(717),
    [aux_sym_number_token2] = ACTIONS(719),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(721),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(723),
    [anon_sym_BQUOTE] = ACTIONS(725),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(727),
    [anon_sym_LT_LPAREN] = ACTIONS(729),
    [anon_sym_GT_LPAREN] = ACTIONS(729),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(731),
    [sym_variable_name] = ACTIONS(733),
    [sym_test_operator] = ACTIONS(735),
    [sym__brace_start] = ACTIONS(737),
  },
  [40] = {
    [sym__statements] = STATE(6700),
    [sym__statement_not_pipeline] = STATE(6674),
    [sym_redirected_statement] = STATE(4416),
    [sym_for_statement] = STATE(4416),
    [sym_c_style_for_statement] = STATE(4416),
    [sym_while_statement] = STATE(4239),
    [sym_if_statement] = STATE(4239),
    [sym_case_statement] = STATE(4416),
    [sym_function_definition] = STATE(4416),
    [sym_compound_statement] = STATE(4416),
    [sym_subshell] = STATE(4416),
    [sym_pipeline] = STATE(4910),
    [sym_list] = STATE(4416),
    [sym_negated_command] = STATE(4416),
    [sym_test_command] = STATE(4416),
    [sym_declaration_command] = STATE(4416),
    [sym_unset_command] = STATE(4416),
    [sym_command] = STATE(4416),
    [sym_command_name] = STATE(530),
    [sym_variable_assignment] = STATE(849),
    [sym_variable_assignments] = STATE(4416),
    [sym_subscript] = STATE(6760),
    [sym_file_redirect] = STATE(1893),
    [sym_herestring_redirect] = STATE(1892),
    [sym_arithmetic_expansion] = STATE(852),
    [sym_brace_expression] = STATE(852),
    [sym_concatenation] = STATE(1169),
    [sym_string] = STATE(852),
    [sym_translated_string] = STATE(852),
    [sym_number] = STATE(852),
    [sym_simple_expansion] = STATE(852),
    [sym_expansion] = STATE(852),
    [sym_command_substitution] = STATE(852),
    [sym_process_substitution] = STATE(852),
    [aux_sym__statements_repeat1] = STATE(426),
    [aux_sym_redirected_statement_repeat2] = STATE(4371),
    [aux_sym_command_repeat1] = STATE(989),
    [aux_sym__literal_repeat1] = STATE(1026),
    [sym_word] = ACTIONS(665),
    [anon_sym_for] = ACTIONS(667),
    [anon_sym_select] = ACTIONS(669),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(671),
    [anon_sym_LT] = ACTIONS(673),
    [anon_sym_GT] = ACTIONS(673),
    [anon_sym_GT_GT] = ACTIONS(675),
    [anon_sym_LPAREN] = ACTIONS(677),
    [anon_sym_while] = ACTIONS(679),
    [anon_sym_until] = ACTIONS(679),
    [anon_sym_if] = ACTIONS(681),
    [anon_sym_case] = ACTIONS(683),
    [anon_sym_SEMI_SEMI] = ACTIONS(747),
    [anon_sym_SEMI_AMP] = ACTIONS(521),
    [anon_sym_SEMI_SEMI_AMP] = ACTIONS(523),
    [anon_sym_function] = ACTIONS(687),
    [anon_sym_LBRACE] = ACTIONS(689),
    [anon_sym_BANG] = ACTIONS(691),
    [anon_sym_LBRACK] = ACTIONS(693),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(695),
    [anon_sym_declare] = ACTIONS(697),
    [anon_sym_typeset] = ACTIONS(697),
    [anon_sym_export] = ACTIONS(697),
    [anon_sym_readonly] = ACTIONS(697),
    [anon_sym_local] = ACTIONS(697),
    [anon_sym_unset] = ACTIONS(699),
    [anon_sym_unsetenv] = ACTIONS(699),
    [anon_sym_AMP_GT] = ACTIONS(673),
    [anon_sym_AMP_GT_GT] = ACTIONS(675),
    [anon_sym_LT_AMP] = ACTIONS(673),
    [anon_sym_GT_AMP] = ACTIONS(673),
    [anon_sym_GT_PIPE] = ACTIONS(675),
    [anon_sym_LT_AMP_DASH] = ACTIONS(701),
    [anon_sym_GT_AMP_DASH] = ACTIONS(701),
    [anon_sym_LT_LT_LT] = ACTIONS(703),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(705),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(707),
    [anon_sym_DOLLAR] = ACTIONS(709),
    [sym__special_character] = ACTIONS(711),
    [anon_sym_DQUOTE] = ACTIONS(713),
    [sym_raw_string] = ACTIONS(715),
    [sym_ansi_c_string] = ACTIONS(715),
    [aux_sym_number_token1] = ACTIONS(717),
    [aux_sym_number_token2] = ACTIONS(719),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(721),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(723),
    [anon_sym_BQUOTE] = ACTIONS(725),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(727),
    [anon_sym_LT_LPAREN] = ACTIONS(729),
    [anon_sym_GT_LPAREN] = ACTIONS(729),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(731),
    [sym_variable_name] = ACTIONS(733),
    [sym_test_operator] = ACTIONS(735),
    [sym__brace_start] = ACTIONS(737),
  },
  [41] = {
    [sym__statements] = STATE(6611),
    [sym__statement_not_pipeline] = STATE(6674),
    [sym_redirected_statement] = STATE(4416),
    [sym_for_statement] = STATE(4416),
    [sym_c_style_for_statement] = STATE(4416),
    [sym_while_statement] = STATE(4239),
    [sym_if_statement] = STATE(4239),
    [sym_case_statement] = STATE(4416),
    [sym_function_definition] = STATE(4416),
    [sym_compound_statement] = STATE(4416),
    [sym_subshell] = STATE(4416),
    [sym_pipeline] = STATE(4910),
    [sym_list] = STATE(4416),
    [sym_negated_command] = STATE(4416),
    [sym_test_command] = STATE(4416),
    [sym_declaration_command] = STATE(4416),
    [sym_unset_command] = STATE(4416),
    [sym_command] = STATE(4416),
    [sym_command_name] = STATE(530),
    [sym_variable_assignment] = STATE(849),
    [sym_variable_assignments] = STATE(4416),
    [sym_subscript] = STATE(6760),
    [sym_file_redirect] = STATE(1893),
    [sym_herestring_redirect] = STATE(1892),
    [sym_arithmetic_expansion] = STATE(852),
    [sym_brace_expression] = STATE(852),
    [sym_concatenation] = STATE(1169),
    [sym_string] = STATE(852),
    [sym_translated_string] = STATE(852),
    [sym_number] = STATE(852),
    [sym_simple_expansion] = STATE(852),
    [sym_expansion] = STATE(852),
    [sym_command_substitution] = STATE(852),
    [sym_process_substitution] = STATE(852),
    [aux_sym__statements_repeat1] = STATE(426),
    [aux_sym_redirected_statement_repeat2] = STATE(4371),
    [aux_sym_command_repeat1] = STATE(989),
    [aux_sym__literal_repeat1] = STATE(1026),
    [sym_word] = ACTIONS(665),
    [anon_sym_for] = ACTIONS(667),
    [anon_sym_select] = ACTIONS(669),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(671),
    [anon_sym_LT] = ACTIONS(673),
    [anon_sym_GT] = ACTIONS(673),
    [anon_sym_GT_GT] = ACTIONS(675),
    [anon_sym_LPAREN] = ACTIONS(677),
    [anon_sym_while] = ACTIONS(679),
    [anon_sym_until] = ACTIONS(679),
    [anon_sym_if] = ACTIONS(681),
    [anon_sym_case] = ACTIONS(683),
    [anon_sym_SEMI_SEMI] = ACTIONS(749),
    [anon_sym_SEMI_AMP] = ACTIONS(487),
    [anon_sym_SEMI_SEMI_AMP] = ACTIONS(487),
    [anon_sym_function] = ACTIONS(687),
    [anon_sym_LBRACE] = ACTIONS(689),
    [anon_sym_BANG] = ACTIONS(691),
    [anon_sym_LBRACK] = ACTIONS(693),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(695),
    [anon_sym_declare] = ACTIONS(697),
    [anon_sym_typeset] = ACTIONS(697),
    [anon_sym_export] = ACTIONS(697),
    [anon_sym_readonly] = ACTIONS(697),
    [anon_sym_local] = ACTIONS(697),
    [anon_sym_unset] = ACTIONS(699),
    [anon_sym_unsetenv] = ACTIONS(699),
    [anon_sym_AMP_GT] = ACTIONS(673),
    [anon_sym_AMP_GT_GT] = ACTIONS(675),
    [anon_sym_LT_AMP] = ACTIONS(673),
    [anon_sym_GT_AMP] = ACTIONS(673),
    [anon_sym_GT_PIPE] = ACTIONS(675),
    [anon_sym_LT_AMP_DASH] = ACTIONS(701),
    [anon_sym_GT_AMP_DASH] = ACTIONS(701),
    [anon_sym_LT_LT_LT] = ACTIONS(703),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(705),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(707),
    [anon_sym_DOLLAR] = ACTIONS(709),
    [sym__special_character] = ACTIONS(711),
    [anon_sym_DQUOTE] = ACTIONS(713),
    [sym_raw_string] = ACTIONS(715),
    [sym_ansi_c_string] = ACTIONS(715),
    [aux_sym_number_token1] = ACTIONS(717),
    [aux_sym_number_token2] = ACTIONS(719),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(721),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(723),
    [anon_sym_BQUOTE] = ACTIONS(725),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(727),
    [anon_sym_LT_LPAREN] = ACTIONS(729),
    [anon_sym_GT_LPAREN] = ACTIONS(729),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(731),
    [sym_variable_name] = ACTIONS(733),
    [sym_test_operator] = ACTIONS(735),
    [sym__brace_start] = ACTIONS(737),
  },
  [42] = {
    [sym__statements] = STATE(6672),
    [sym__statement_not_pipeline] = STATE(6674),
    [sym_redirected_statement] = STATE(4416),
    [sym_for_statement] = STATE(4416),
    [sym_c_style_for_statement] = STATE(4416),
    [sym_while_statement] = STATE(4239),
    [sym_if_statement] = STATE(4239),
    [sym_case_statement] = STATE(4416),
    [sym_function_definition] = STATE(4416),
    [sym_compound_statement] = STATE(4416),
    [sym_subshell] = STATE(4416),
    [sym_pipeline] = STATE(4910),
    [sym_list] = STATE(4416),
    [sym_negated_command] = STATE(4416),
    [sym_test_command] = STATE(4416),
    [sym_declaration_command] = STATE(4416),
    [sym_unset_command] = STATE(4416),
    [sym_command] = STATE(4416),
    [sym_command_name] = STATE(530),
    [sym_variable_assignment] = STATE(849),
    [sym_variable_assignments] = STATE(4416),
    [sym_subscript] = STATE(6760),
    [sym_file_redirect] = STATE(1893),
    [sym_herestring_redirect] = STATE(1892),
    [sym_arithmetic_expansion] = STATE(852),
    [sym_brace_expression] = STATE(852),
    [sym_concatenation] = STATE(1169),
    [sym_string] = STATE(852),
    [sym_translated_string] = STATE(852),
    [sym_number] = STATE(852),
    [sym_simple_expansion] = STATE(852),
    [sym_expansion] = STATE(852),
    [sym_command_substitution] = STATE(852),
    [sym_process_substitution] = STATE(852),
    [aux_sym__statements_repeat1] = STATE(426),
    [aux_sym_redirected_statement_repeat2] = STATE(4371),
    [aux_sym_command_repeat1] = STATE(989),
    [aux_sym__literal_repeat1] = STATE(1026),
    [sym_word] = ACTIONS(665),
    [anon_sym_for] = ACTIONS(667),
    [anon_sym_select] = ACTIONS(669),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(671),
    [anon_sym_LT] = ACTIONS(673),
    [anon_sym_GT] = ACTIONS(673),
    [anon_sym_GT_GT] = ACTIONS(675),
    [anon_sym_LPAREN] = ACTIONS(677),
    [anon_sym_while] = ACTIONS(679),
    [anon_sym_until] = ACTIONS(679),
    [anon_sym_if] = ACTIONS(681),
    [anon_sym_case] = ACTIONS(683),
    [anon_sym_SEMI_SEMI] = ACTIONS(751),
    [anon_sym_SEMI_AMP] = ACTIONS(547),
    [anon_sym_SEMI_SEMI_AMP] = ACTIONS(547),
    [anon_sym_function] = ACTIONS(687),
    [anon_sym_LBRACE] = ACTIONS(689),
    [anon_sym_BANG] = ACTIONS(691),
    [anon_sym_LBRACK] = ACTIONS(693),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(695),
    [anon_sym_declare] = ACTIONS(697),
    [anon_sym_typeset] = ACTIONS(697),
    [anon_sym_export] = ACTIONS(697),
    [anon_sym_readonly] = ACTIONS(697),
    [anon_sym_local] = ACTIONS(697),
    [anon_sym_unset] = ACTIONS(699),
    [anon_sym_unsetenv] = ACTIONS(699),
    [anon_sym_AMP_GT] = ACTIONS(673),
    [anon_sym_AMP_GT_GT] = ACTIONS(675),
    [anon_sym_LT_AMP] = ACTIONS(673),
    [anon_sym_GT_AMP] = ACTIONS(673),
    [anon_sym_GT_PIPE] = ACTIONS(675),
    [anon_sym_LT_AMP_DASH] = ACTIONS(701),
    [anon_sym_GT_AMP_DASH] = ACTIONS(701),
    [anon_sym_LT_LT_LT] = ACTIONS(703),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(705),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(707),
    [anon_sym_DOLLAR] = ACTIONS(709),
    [sym__special_character] = ACTIONS(711),
    [anon_sym_DQUOTE] = ACTIONS(713),
    [sym_raw_string] = ACTIONS(715),
    [sym_ansi_c_string] = ACTIONS(715),
    [aux_sym_number_token1] = ACTIONS(717),
    [aux_sym_number_token2] = ACTIONS(719),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(721),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(723),
    [anon_sym_BQUOTE] = ACTIONS(725),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(727),
    [anon_sym_LT_LPAREN] = ACTIONS(729),
    [anon_sym_GT_LPAREN] = ACTIONS(729),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(731),
    [sym_variable_name] = ACTIONS(733),
    [sym_test_operator] = ACTIONS(735),
    [sym__brace_start] = ACTIONS(737),
  },
  [43] = {
    [sym__statements] = STATE(6667),
    [sym__statement_not_pipeline] = STATE(6674),
    [sym_redirected_statement] = STATE(4416),
    [sym_for_statement] = STATE(4416),
    [sym_c_style_for_statement] = STATE(4416),
    [sym_while_statement] = STATE(4239),
    [sym_if_statement] = STATE(4239),
    [sym_case_statement] = STATE(4416),
    [sym_function_definition] = STATE(4416),
    [sym_compound_statement] = STATE(4416),
    [sym_subshell] = STATE(4416),
    [sym_pipeline] = STATE(4910),
    [sym_list] = STATE(4416),
    [sym_negated_command] = STATE(4416),
    [sym_test_command] = STATE(4416),
    [sym_declaration_command] = STATE(4416),
    [sym_unset_command] = STATE(4416),
    [sym_command] = STATE(4416),
    [sym_command_name] = STATE(530),
    [sym_variable_assignment] = STATE(849),
    [sym_variable_assignments] = STATE(4416),
    [sym_subscript] = STATE(6760),
    [sym_file_redirect] = STATE(1893),
    [sym_herestring_redirect] = STATE(1892),
    [sym_arithmetic_expansion] = STATE(852),
    [sym_brace_expression] = STATE(852),
    [sym_concatenation] = STATE(1169),
    [sym_string] = STATE(852),
    [sym_translated_string] = STATE(852),
    [sym_number] = STATE(852),
    [sym_simple_expansion] = STATE(852),
    [sym_expansion] = STATE(852),
    [sym_command_substitution] = STATE(852),
    [sym_process_substitution] = STATE(852),
    [aux_sym__statements_repeat1] = STATE(426),
    [aux_sym_redirected_statement_repeat2] = STATE(4371),
    [aux_sym_command_repeat1] = STATE(989),
    [aux_sym__literal_repeat1] = STATE(1026),
    [sym_word] = ACTIONS(665),
    [anon_sym_for] = ACTIONS(667),
    [anon_sym_select] = ACTIONS(669),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(671),
    [anon_sym_LT] = ACTIONS(673),
    [anon_sym_GT] = ACTIONS(673),
    [anon_sym_GT_GT] = ACTIONS(675),
    [anon_sym_LPAREN] = ACTIONS(677),
    [anon_sym_while] = ACTIONS(679),
    [anon_sym_until] = ACTIONS(679),
    [anon_sym_if] = ACTIONS(681),
    [anon_sym_case] = ACTIONS(683),
    [anon_sym_SEMI_SEMI] = ACTIONS(753),
    [anon_sym_SEMI_AMP] = ACTIONS(493),
    [anon_sym_SEMI_SEMI_AMP] = ACTIONS(495),
    [anon_sym_function] = ACTIONS(687),
    [anon_sym_LBRACE] = ACTIONS(689),
    [anon_sym_BANG] = ACTIONS(691),
    [anon_sym_LBRACK] = ACTIONS(693),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(695),
    [anon_sym_declare] = ACTIONS(697),
    [anon_sym_typeset] = ACTIONS(697),
    [anon_sym_export] = ACTIONS(697),
    [anon_sym_readonly] = ACTIONS(697),
    [anon_sym_local] = ACTIONS(697),
    [anon_sym_unset] = ACTIONS(699),
    [anon_sym_unsetenv] = ACTIONS(699),
    [anon_sym_AMP_GT] = ACTIONS(673),
    [anon_sym_AMP_GT_GT] = ACTIONS(675),
    [anon_sym_LT_AMP] = ACTIONS(673),
    [anon_sym_GT_AMP] = ACTIONS(673),
    [anon_sym_GT_PIPE] = ACTIONS(675),
    [anon_sym_LT_AMP_DASH] = ACTIONS(701),
    [anon_sym_GT_AMP_DASH] = ACTIONS(701),
    [anon_sym_LT_LT_LT] = ACTIONS(703),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(705),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(707),
    [anon_sym_DOLLAR] = ACTIONS(709),
    [sym__special_character] = ACTIONS(711),
    [anon_sym_DQUOTE] = ACTIONS(713),
    [sym_raw_string] = ACTIONS(715),
    [sym_ansi_c_string] = ACTIONS(715),
    [aux_sym_number_token1] = ACTIONS(717),
    [aux_sym_number_token2] = ACTIONS(719),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(721),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(723),
    [anon_sym_BQUOTE] = ACTIONS(725),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(727),
    [anon_sym_LT_LPAREN] = ACTIONS(729),
    [anon_sym_GT_LPAREN] = ACTIONS(729),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(731),
    [sym_variable_name] = ACTIONS(733),
    [sym_test_operator] = ACTIONS(735),
    [sym__brace_start] = ACTIONS(737),
  },
  [44] = {
    [sym__statements] = STATE(6637),
    [sym__statement_not_pipeline] = STATE(6674),
    [sym_redirected_statement] = STATE(4416),
    [sym_for_statement] = STATE(4416),
    [sym_c_style_for_statement] = STATE(4416),
    [sym_while_statement] = STATE(4239),
    [sym_if_statement] = STATE(4239),
    [sym_case_statement] = STATE(4416),
    [sym_function_definition] = STATE(4416),
    [sym_compound_statement] = STATE(4416),
    [sym_subshell] = STATE(4416),
    [sym_pipeline] = STATE(4910),
    [sym_list] = STATE(4416),
    [sym_negated_command] = STATE(4416),
    [sym_test_command] = STATE(4416),
    [sym_declaration_command] = STATE(4416),
    [sym_unset_command] = STATE(4416),
    [sym_command] = STATE(4416),
    [sym_command_name] = STATE(530),
    [sym_variable_assignment] = STATE(849),
    [sym_variable_assignments] = STATE(4416),
    [sym_subscript] = STATE(6760),
    [sym_file_redirect] = STATE(1893),
    [sym_herestring_redirect] = STATE(1892),
    [sym_arithmetic_expansion] = STATE(852),
    [sym_brace_expression] = STATE(852),
    [sym_concatenation] = STATE(1169),
    [sym_string] = STATE(852),
    [sym_translated_string] = STATE(852),
    [sym_number] = STATE(852),
    [sym_simple_expansion] = STATE(852),
    [sym_expansion] = STATE(852),
    [sym_command_substitution] = STATE(852),
    [sym_process_substitution] = STATE(852),
    [aux_sym__statements_repeat1] = STATE(426),
    [aux_sym_redirected_statement_repeat2] = STATE(4371),
    [aux_sym_command_repeat1] = STATE(989),
    [aux_sym__literal_repeat1] = STATE(1026),
    [sym_word] = ACTIONS(665),
    [anon_sym_for] = ACTIONS(667),
    [anon_sym_select] = ACTIONS(669),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(671),
    [anon_sym_LT] = ACTIONS(673),
    [anon_sym_GT] = ACTIONS(673),
    [anon_sym_GT_GT] = ACTIONS(675),
    [anon_sym_LPAREN] = ACTIONS(677),
    [anon_sym_while] = ACTIONS(679),
    [anon_sym_until] = ACTIONS(679),
    [anon_sym_if] = ACTIONS(681),
    [anon_sym_case] = ACTIONS(683),
    [anon_sym_SEMI_SEMI] = ACTIONS(755),
    [anon_sym_SEMI_AMP] = ACTIONS(507),
    [anon_sym_SEMI_SEMI_AMP] = ACTIONS(509),
    [anon_sym_function] = ACTIONS(687),
    [anon_sym_LBRACE] = ACTIONS(689),
    [anon_sym_BANG] = ACTIONS(691),
    [anon_sym_LBRACK] = ACTIONS(693),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(695),
    [anon_sym_declare] = ACTIONS(697),
    [anon_sym_typeset] = ACTIONS(697),
    [anon_sym_export] = ACTIONS(697),
    [anon_sym_readonly] = ACTIONS(697),
    [anon_sym_local] = ACTIONS(697),
    [anon_sym_unset] = ACTIONS(699),
    [anon_sym_unsetenv] = ACTIONS(699),
    [anon_sym_AMP_GT] = ACTIONS(673),
    [anon_sym_AMP_GT_GT] = ACTIONS(675),
    [anon_sym_LT_AMP] = ACTIONS(673),
    [anon_sym_GT_AMP] = ACTIONS(673),
    [anon_sym_GT_PIPE] = ACTIONS(675),
    [anon_sym_LT_AMP_DASH] = ACTIONS(701),
    [anon_sym_GT_AMP_DASH] = ACTIONS(701),
    [anon_sym_LT_LT_LT] = ACTIONS(703),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(705),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(707),
    [anon_sym_DOLLAR] = ACTIONS(709),
    [sym__special_character] = ACTIONS(711),
    [anon_sym_DQUOTE] = ACTIONS(713),
    [sym_raw_string] = ACTIONS(715),
    [sym_ansi_c_string] = ACTIONS(715),
    [aux_sym_number_token1] = ACTIONS(717),
    [aux_sym_number_token2] = ACTIONS(719),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(721),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(723),
    [anon_sym_BQUOTE] = ACTIONS(725),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(727),
    [anon_sym_LT_LPAREN] = ACTIONS(729),
    [anon_sym_GT_LPAREN] = ACTIONS(729),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(731),
    [sym_variable_name] = ACTIONS(733),
    [sym_test_operator] = ACTIONS(735),
    [sym__brace_start] = ACTIONS(737),
  },
  [45] = {
    [sym__statements] = STATE(6666),
    [sym__statement_not_pipeline] = STATE(6674),
    [sym_redirected_statement] = STATE(4416),
    [sym_for_statement] = STATE(4416),
    [sym_c_style_for_statement] = STATE(4416),
    [sym_while_statement] = STATE(4239),
    [sym_if_statement] = STATE(4239),
    [sym_case_statement] = STATE(4416),
    [sym_function_definition] = STATE(4416),
    [sym_compound_statement] = STATE(4416),
    [sym_subshell] = STATE(4416),
    [sym_pipeline] = STATE(4910),
    [sym_list] = STATE(4416),
    [sym_negated_command] = STATE(4416),
    [sym_test_command] = STATE(4416),
    [sym_declaration_command] = STATE(4416),
    [sym_unset_command] = STATE(4416),
    [sym_command] = STATE(4416),
    [sym_command_name] = STATE(530),
    [sym_variable_assignment] = STATE(849),
    [sym_variable_assignments] = STATE(4416),
    [sym_subscript] = STATE(6760),
    [sym_file_redirect] = STATE(1893),
    [sym_herestring_redirect] = STATE(1892),
    [sym_arithmetic_expansion] = STATE(852),
    [sym_brace_expression] = STATE(852),
    [sym_concatenation] = STATE(1169),
    [sym_string] = STATE(852),
    [sym_translated_string] = STATE(852),
    [sym_number] = STATE(852),
    [sym_simple_expansion] = STATE(852),
    [sym_expansion] = STATE(852),
    [sym_command_substitution] = STATE(852),
    [sym_process_substitution] = STATE(852),
    [aux_sym__statements_repeat1] = STATE(426),
    [aux_sym_redirected_statement_repeat2] = STATE(4371),
    [aux_sym_command_repeat1] = STATE(989),
    [aux_sym__literal_repeat1] = STATE(1026),
    [sym_word] = ACTIONS(665),
    [anon_sym_for] = ACTIONS(667),
    [anon_sym_select] = ACTIONS(669),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(671),
    [anon_sym_LT] = ACTIONS(673),
    [anon_sym_GT] = ACTIONS(673),
    [anon_sym_GT_GT] = ACTIONS(675),
    [anon_sym_LPAREN] = ACTIONS(677),
    [anon_sym_while] = ACTIONS(679),
    [anon_sym_until] = ACTIONS(679),
    [anon_sym_if] = ACTIONS(681),
    [anon_sym_case] = ACTIONS(683),
    [anon_sym_SEMI_SEMI] = ACTIONS(757),
    [anon_sym_SEMI_AMP] = ACTIONS(541),
    [anon_sym_SEMI_SEMI_AMP] = ACTIONS(541),
    [anon_sym_function] = ACTIONS(687),
    [anon_sym_LBRACE] = ACTIONS(689),
    [anon_sym_BANG] = ACTIONS(691),
    [anon_sym_LBRACK] = ACTIONS(693),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(695),
    [anon_sym_declare] = ACTIONS(697),
    [anon_sym_typeset] = ACTIONS(697),
    [anon_sym_export] = ACTIONS(697),
    [anon_sym_readonly] = ACTIONS(697),
    [anon_sym_local] = ACTIONS(697),
    [anon_sym_unset] = ACTIONS(699),
    [anon_sym_unsetenv] = ACTIONS(699),
    [anon_sym_AMP_GT] = ACTIONS(673),
    [anon_sym_AMP_GT_GT] = ACTIONS(675),
    [anon_sym_LT_AMP] = ACTIONS(673),
    [anon_sym_GT_AMP] = ACTIONS(673),
    [anon_sym_GT_PIPE] = ACTIONS(675),
    [anon_sym_LT_AMP_DASH] = ACTIONS(701),
    [anon_sym_GT_AMP_DASH] = ACTIONS(701),
    [anon_sym_LT_LT_LT] = ACTIONS(703),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(705),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(707),
    [anon_sym_DOLLAR] = ACTIONS(709),
    [sym__special_character] = ACTIONS(711),
    [anon_sym_DQUOTE] = ACTIONS(713),
    [sym_raw_string] = ACTIONS(715),
    [sym_ansi_c_string] = ACTIONS(715),
    [aux_sym_number_token1] = ACTIONS(717),
    [aux_sym_number_token2] = ACTIONS(719),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(721),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(723),
    [anon_sym_BQUOTE] = ACTIONS(725),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(727),
    [anon_sym_LT_LPAREN] = ACTIONS(729),
    [anon_sym_GT_LPAREN] = ACTIONS(729),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(731),
    [sym_variable_name] = ACTIONS(733),
    [sym_test_operator] = ACTIONS(735),
    [sym__brace_start] = ACTIONS(737),
  },
  [46] = {
    [sym__statements] = STATE(6690),
    [sym__statement_not_pipeline] = STATE(6674),
    [sym_redirected_statement] = STATE(4416),
    [sym_for_statement] = STATE(4416),
    [sym_c_style_for_statement] = STATE(4416),
    [sym_while_statement] = STATE(4239),
    [sym_if_statement] = STATE(4239),
    [sym_case_statement] = STATE(4416),
    [sym_function_definition] = STATE(4416),
    [sym_compound_statement] = STATE(4416),
    [sym_subshell] = STATE(4416),
    [sym_pipeline] = STATE(4910),
    [sym_list] = STATE(4416),
    [sym_negated_command] = STATE(4416),
    [sym_test_command] = STATE(4416),
    [sym_declaration_command] = STATE(4416),
    [sym_unset_command] = STATE(4416),
    [sym_command] = STATE(4416),
    [sym_command_name] = STATE(530),
    [sym_variable_assignment] = STATE(849),
    [sym_variable_assignments] = STATE(4416),
    [sym_subscript] = STATE(6760),
    [sym_file_redirect] = STATE(1893),
    [sym_herestring_redirect] = STATE(1892),
    [sym_arithmetic_expansion] = STATE(852),
    [sym_brace_expression] = STATE(852),
    [sym_concatenation] = STATE(1169),
    [sym_string] = STATE(852),
    [sym_translated_string] = STATE(852),
    [sym_number] = STATE(852),
    [sym_simple_expansion] = STATE(852),
    [sym_expansion] = STATE(852),
    [sym_command_substitution] = STATE(852),
    [sym_process_substitution] = STATE(852),
    [aux_sym__statements_repeat1] = STATE(426),
    [aux_sym_redirected_statement_repeat2] = STATE(4371),
    [aux_sym_command_repeat1] = STATE(989),
    [aux_sym__literal_repeat1] = STATE(1026),
    [sym_word] = ACTIONS(665),
    [anon_sym_for] = ACTIONS(667),
    [anon_sym_select] = ACTIONS(669),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(671),
    [anon_sym_LT] = ACTIONS(673),
    [anon_sym_GT] = ACTIONS(673),
    [anon_sym_GT_GT] = ACTIONS(675),
    [anon_sym_LPAREN] = ACTIONS(677),
    [anon_sym_while] = ACTIONS(679),
    [anon_sym_until] = ACTIONS(679),
    [anon_sym_if] = ACTIONS(681),
    [anon_sym_case] = ACTIONS(683),
    [anon_sym_SEMI_SEMI] = ACTIONS(759),
    [anon_sym_SEMI_AMP] = ACTIONS(527),
    [anon_sym_SEMI_SEMI_AMP] = ACTIONS(529),
    [anon_sym_function] = ACTIONS(687),
    [anon_sym_LBRACE] = ACTIONS(689),
    [anon_sym_BANG] = ACTIONS(691),
    [anon_sym_LBRACK] = ACTIONS(693),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(695),
    [anon_sym_declare] = ACTIONS(697),
    [anon_sym_typeset] = ACTIONS(697),
    [anon_sym_export] = ACTIONS(697),
    [anon_sym_readonly] = ACTIONS(697),
    [anon_sym_local] = ACTIONS(697),
    [anon_sym_unset] = ACTIONS(699),
    [anon_sym_unsetenv] = ACTIONS(699),
    [anon_sym_AMP_GT] = ACTIONS(673),
    [anon_sym_AMP_GT_GT] = ACTIONS(675),
    [anon_sym_LT_AMP] = ACTIONS(673),
    [anon_sym_GT_AMP] = ACTIONS(673),
    [anon_sym_GT_PIPE] = ACTIONS(675),
    [anon_sym_LT_AMP_DASH] = ACTIONS(701),
    [anon_sym_GT_AMP_DASH] = ACTIONS(701),
    [anon_sym_LT_LT_LT] = ACTIONS(703),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(705),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(707),
    [anon_sym_DOLLAR] = ACTIONS(709),
    [sym__special_character] = ACTIONS(711),
    [anon_sym_DQUOTE] = ACTIONS(713),
    [sym_raw_string] = ACTIONS(715),
    [sym_ansi_c_string] = ACTIONS(715),
    [aux_sym_number_token1] = ACTIONS(717),
    [aux_sym_number_token2] = ACTIONS(719),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(721),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(723),
    [anon_sym_BQUOTE] = ACTIONS(725),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(727),
    [anon_sym_LT_LPAREN] = ACTIONS(729),
    [anon_sym_GT_LPAREN] = ACTIONS(729),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(731),
    [sym_variable_name] = ACTIONS(733),
    [sym_test_operator] = ACTIONS(735),
    [sym__brace_start] = ACTIONS(737),
  },
  [47] = {
    [aux_sym__terminated_statement] = STATE(34),
    [sym__statement_not_pipeline] = STATE(6718),
    [sym_redirected_statement] = STATE(5007),
    [sym_for_statement] = STATE(5007),
    [sym_c_style_for_statement] = STATE(5007),
    [sym_while_statement] = STATE(4551),
    [sym_if_statement] = STATE(4551),
    [sym_case_statement] = STATE(5007),
    [sym_function_definition] = STATE(5007),
    [sym_compound_statement] = STATE(5007),
    [sym_subshell] = STATE(5007),
    [sym_pipeline] = STATE(5189),
    [sym_list] = STATE(5007),
    [sym_negated_command] = STATE(5007),
    [sym_test_command] = STATE(5007),
    [sym_declaration_command] = STATE(5007),
    [sym_unset_command] = STATE(5007),
    [sym_command] = STATE(5007),
    [sym_command_name] = STATE(544),
    [sym_variable_assignment] = STATE(1151),
    [sym_variable_assignments] = STATE(5007),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym_redirected_statement_repeat2] = STATE(4644),
    [aux_sym_command_repeat1] = STATE(997),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(286),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_fi] = ACTIONS(761),
    [anon_sym_elif] = ACTIONS(761),
    [anon_sym_else] = ACTIONS(761),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(300),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(302),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(304),
    [anon_sym_typeset] = ACTIONS(304),
    [anon_sym_export] = ACTIONS(304),
    [anon_sym_readonly] = ACTIONS(304),
    [anon_sym_local] = ACTIONS(304),
    [anon_sym_unset] = ACTIONS(306),
    [anon_sym_unsetenv] = ACTIONS(306),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [48] = {
    [aux_sym__terminated_statement] = STATE(47),
    [sym__statement_not_pipeline] = STATE(6718),
    [sym_redirected_statement] = STATE(5007),
    [sym_for_statement] = STATE(5007),
    [sym_c_style_for_statement] = STATE(5007),
    [sym_while_statement] = STATE(4551),
    [sym_if_statement] = STATE(4551),
    [sym_case_statement] = STATE(5007),
    [sym_function_definition] = STATE(5007),
    [sym_compound_statement] = STATE(5007),
    [sym_subshell] = STATE(5007),
    [sym_pipeline] = STATE(5189),
    [sym_list] = STATE(5007),
    [sym_negated_command] = STATE(5007),
    [sym_test_command] = STATE(5007),
    [sym_declaration_command] = STATE(5007),
    [sym_unset_command] = STATE(5007),
    [sym_command] = STATE(5007),
    [sym_command_name] = STATE(544),
    [sym_variable_assignment] = STATE(1151),
    [sym_variable_assignments] = STATE(5007),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym_redirected_statement_repeat2] = STATE(4644),
    [aux_sym_command_repeat1] = STATE(997),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(286),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_fi] = ACTIONS(763),
    [anon_sym_elif] = ACTIONS(763),
    [anon_sym_else] = ACTIONS(763),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(300),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(302),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(304),
    [anon_sym_typeset] = ACTIONS(304),
    [anon_sym_export] = ACTIONS(304),
    [anon_sym_readonly] = ACTIONS(304),
    [anon_sym_local] = ACTIONS(304),
    [anon_sym_unset] = ACTIONS(306),
    [anon_sym_unsetenv] = ACTIONS(306),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [49] = {
    [sym__statements] = STATE(7117),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_RPAREN] = ACTIONS(767),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [50] = {
    [aux_sym__terminated_statement] = STATE(34),
    [sym__statement_not_pipeline] = STATE(6718),
    [sym_redirected_statement] = STATE(5007),
    [sym_for_statement] = STATE(5007),
    [sym_c_style_for_statement] = STATE(5007),
    [sym_while_statement] = STATE(4551),
    [sym_do_group] = STATE(4946),
    [sym_if_statement] = STATE(4551),
    [sym_case_statement] = STATE(5007),
    [sym_function_definition] = STATE(5007),
    [sym_compound_statement] = STATE(5007),
    [sym_subshell] = STATE(5007),
    [sym_pipeline] = STATE(5189),
    [sym_list] = STATE(5007),
    [sym_negated_command] = STATE(5007),
    [sym_test_command] = STATE(5007),
    [sym_declaration_command] = STATE(5007),
    [sym_unset_command] = STATE(5007),
    [sym_command] = STATE(5007),
    [sym_command_name] = STATE(544),
    [sym_variable_assignment] = STATE(1151),
    [sym_variable_assignments] = STATE(5007),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym_redirected_statement_repeat2] = STATE(4644),
    [aux_sym_command_repeat1] = STATE(997),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(286),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_do] = ACTIONS(777),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(300),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(302),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(304),
    [anon_sym_typeset] = ACTIONS(304),
    [anon_sym_export] = ACTIONS(304),
    [anon_sym_readonly] = ACTIONS(304),
    [anon_sym_local] = ACTIONS(304),
    [anon_sym_unset] = ACTIONS(306),
    [anon_sym_unsetenv] = ACTIONS(306),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [51] = {
    [aux_sym__terminated_statement] = STATE(34),
    [sym__statement_not_pipeline] = STATE(6718),
    [sym_redirected_statement] = STATE(5007),
    [sym_for_statement] = STATE(5007),
    [sym_c_style_for_statement] = STATE(5007),
    [sym_while_statement] = STATE(4551),
    [sym_do_group] = STATE(4632),
    [sym_if_statement] = STATE(4551),
    [sym_case_statement] = STATE(5007),
    [sym_function_definition] = STATE(5007),
    [sym_compound_statement] = STATE(5007),
    [sym_subshell] = STATE(5007),
    [sym_pipeline] = STATE(5189),
    [sym_list] = STATE(5007),
    [sym_negated_command] = STATE(5007),
    [sym_test_command] = STATE(5007),
    [sym_declaration_command] = STATE(5007),
    [sym_unset_command] = STATE(5007),
    [sym_command] = STATE(5007),
    [sym_command_name] = STATE(544),
    [sym_variable_assignment] = STATE(1151),
    [sym_variable_assignments] = STATE(5007),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym_redirected_statement_repeat2] = STATE(4644),
    [aux_sym_command_repeat1] = STATE(997),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(286),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_do] = ACTIONS(779),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(300),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(302),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(304),
    [anon_sym_typeset] = ACTIONS(304),
    [anon_sym_export] = ACTIONS(304),
    [anon_sym_readonly] = ACTIONS(304),
    [anon_sym_local] = ACTIONS(304),
    [anon_sym_unset] = ACTIONS(306),
    [anon_sym_unsetenv] = ACTIONS(306),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [52] = {
    [sym__statements] = STATE(7117),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_RPAREN] = ACTIONS(781),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [53] = {
    [sym__statements] = STATE(7053),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_RPAREN] = ACTIONS(783),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [54] = {
    [aux_sym__terminated_statement] = STATE(34),
    [sym__statement_not_pipeline] = STATE(6718),
    [sym_redirected_statement] = STATE(5007),
    [sym_for_statement] = STATE(5007),
    [sym_c_style_for_statement] = STATE(5007),
    [sym_while_statement] = STATE(4551),
    [sym_do_group] = STATE(4666),
    [sym_if_statement] = STATE(4551),
    [sym_case_statement] = STATE(5007),
    [sym_function_definition] = STATE(5007),
    [sym_compound_statement] = STATE(5007),
    [sym_subshell] = STATE(5007),
    [sym_pipeline] = STATE(5189),
    [sym_list] = STATE(5007),
    [sym_negated_command] = STATE(5007),
    [sym_test_command] = STATE(5007),
    [sym_declaration_command] = STATE(5007),
    [sym_unset_command] = STATE(5007),
    [sym_command] = STATE(5007),
    [sym_command_name] = STATE(544),
    [sym_variable_assignment] = STATE(1151),
    [sym_variable_assignments] = STATE(5007),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym_redirected_statement_repeat2] = STATE(4644),
    [aux_sym_command_repeat1] = STATE(997),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(286),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_do] = ACTIONS(785),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(300),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(302),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(304),
    [anon_sym_typeset] = ACTIONS(304),
    [anon_sym_export] = ACTIONS(304),
    [anon_sym_readonly] = ACTIONS(304),
    [anon_sym_local] = ACTIONS(304),
    [anon_sym_unset] = ACTIONS(306),
    [anon_sym_unsetenv] = ACTIONS(306),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [55] = {
    [sym__statements] = STATE(7441),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_RPAREN] = ACTIONS(787),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [56] = {
    [sym__statements] = STATE(7117),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_RPAREN] = ACTIONS(789),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [57] = {
    [sym__statements] = STATE(7117),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_RPAREN] = ACTIONS(791),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [58] = {
    [sym__statements] = STATE(7350),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_RPAREN] = ACTIONS(793),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [59] = {
    [aux_sym__terminated_statement] = STATE(34),
    [sym__statement_not_pipeline] = STATE(6718),
    [sym_redirected_statement] = STATE(5007),
    [sym_for_statement] = STATE(5007),
    [sym_c_style_for_statement] = STATE(5007),
    [sym_while_statement] = STATE(4551),
    [sym_do_group] = STATE(5560),
    [sym_if_statement] = STATE(4551),
    [sym_case_statement] = STATE(5007),
    [sym_function_definition] = STATE(5007),
    [sym_compound_statement] = STATE(5007),
    [sym_subshell] = STATE(5007),
    [sym_pipeline] = STATE(5189),
    [sym_list] = STATE(5007),
    [sym_negated_command] = STATE(5007),
    [sym_test_command] = STATE(5007),
    [sym_declaration_command] = STATE(5007),
    [sym_unset_command] = STATE(5007),
    [sym_command] = STATE(5007),
    [sym_command_name] = STATE(544),
    [sym_variable_assignment] = STATE(1151),
    [sym_variable_assignments] = STATE(5007),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym_redirected_statement_repeat2] = STATE(4644),
    [aux_sym_command_repeat1] = STATE(997),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(286),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_do] = ACTIONS(795),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(300),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(302),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(304),
    [anon_sym_typeset] = ACTIONS(304),
    [anon_sym_export] = ACTIONS(304),
    [anon_sym_readonly] = ACTIONS(304),
    [anon_sym_local] = ACTIONS(304),
    [anon_sym_unset] = ACTIONS(306),
    [anon_sym_unsetenv] = ACTIONS(306),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [60] = {
    [sym__statements] = STATE(7117),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_RPAREN] = ACTIONS(797),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [61] = {
    [sym__statements] = STATE(7328),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [62] = {
    [aux_sym__terminated_statement] = STATE(62),
    [sym__statement_not_pipeline] = STATE(6718),
    [sym_redirected_statement] = STATE(4993),
    [sym_for_statement] = STATE(4993),
    [sym_c_style_for_statement] = STATE(4993),
    [sym_while_statement] = STATE(4489),
    [sym_if_statement] = STATE(4489),
    [sym_case_statement] = STATE(4993),
    [sym_function_definition] = STATE(4993),
    [sym_compound_statement] = STATE(4993),
    [sym_subshell] = STATE(4993),
    [sym_pipeline] = STATE(5122),
    [sym_list] = STATE(4993),
    [sym_negated_command] = STATE(4993),
    [sym_test_command] = STATE(4993),
    [sym_declaration_command] = STATE(4993),
    [sym_unset_command] = STATE(4993),
    [sym_command] = STATE(4993),
    [sym_command_name] = STATE(544),
    [sym_variable_assignment] = STATE(1030),
    [sym_variable_assignments] = STATE(4993),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym_redirected_statement_repeat2] = STATE(4644),
    [aux_sym_command_repeat1] = STATE(997),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(555),
    [anon_sym_for] = ACTIONS(558),
    [anon_sym_select] = ACTIONS(561),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(564),
    [anon_sym_LT] = ACTIONS(567),
    [anon_sym_GT] = ACTIONS(567),
    [anon_sym_GT_GT] = ACTIONS(570),
    [anon_sym_LPAREN] = ACTIONS(573),
    [anon_sym_while] = ACTIONS(576),
    [anon_sym_until] = ACTIONS(576),
    [anon_sym_if] = ACTIONS(581),
    [anon_sym_case] = ACTIONS(584),
    [anon_sym_function] = ACTIONS(587),
    [anon_sym_LBRACE] = ACTIONS(590),
    [anon_sym_RBRACE] = ACTIONS(809),
    [anon_sym_BANG] = ACTIONS(593),
    [anon_sym_LBRACK] = ACTIONS(596),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(599),
    [anon_sym_declare] = ACTIONS(602),
    [anon_sym_typeset] = ACTIONS(602),
    [anon_sym_export] = ACTIONS(602),
    [anon_sym_readonly] = ACTIONS(602),
    [anon_sym_local] = ACTIONS(602),
    [anon_sym_unset] = ACTIONS(605),
    [anon_sym_unsetenv] = ACTIONS(605),
    [anon_sym_AMP_GT] = ACTIONS(567),
    [anon_sym_AMP_GT_GT] = ACTIONS(570),
    [anon_sym_LT_AMP] = ACTIONS(567),
    [anon_sym_GT_AMP] = ACTIONS(567),
    [anon_sym_GT_PIPE] = ACTIONS(570),
    [anon_sym_LT_AMP_DASH] = ACTIONS(608),
    [anon_sym_GT_AMP_DASH] = ACTIONS(608),
    [anon_sym_LT_LT_LT] = ACTIONS(611),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(614),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(617),
    [anon_sym_DOLLAR] = ACTIONS(620),
    [sym__special_character] = ACTIONS(623),
    [anon_sym_DQUOTE] = ACTIONS(626),
    [sym_raw_string] = ACTIONS(629),
    [sym_ansi_c_string] = ACTIONS(629),
    [aux_sym_number_token1] = ACTIONS(632),
    [aux_sym_number_token2] = ACTIONS(635),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(638),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(644),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(647),
    [anon_sym_LT_LPAREN] = ACTIONS(650),
    [anon_sym_GT_LPAREN] = ACTIONS(650),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(653),
    [sym_variable_name] = ACTIONS(656),
    [sym_test_operator] = ACTIONS(659),
    [sym__brace_start] = ACTIONS(662),
  },
  [63] = {
    [aux_sym__terminated_statement] = STATE(68),
    [sym__statement_not_pipeline] = STATE(6718),
    [sym_redirected_statement] = STATE(4993),
    [sym_for_statement] = STATE(4993),
    [sym_c_style_for_statement] = STATE(4993),
    [sym_while_statement] = STATE(4489),
    [sym_if_statement] = STATE(4489),
    [sym_case_statement] = STATE(4993),
    [sym_function_definition] = STATE(4993),
    [sym_compound_statement] = STATE(4993),
    [sym_subshell] = STATE(4993),
    [sym_pipeline] = STATE(5122),
    [sym_list] = STATE(4993),
    [sym_negated_command] = STATE(4993),
    [sym_test_command] = STATE(4993),
    [sym_declaration_command] = STATE(4993),
    [sym_unset_command] = STATE(4993),
    [sym_command] = STATE(4993),
    [sym_command_name] = STATE(544),
    [sym_variable_assignment] = STATE(1030),
    [sym_variable_assignments] = STATE(4993),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym_redirected_statement_repeat2] = STATE(4644),
    [aux_sym_command_repeat1] = STATE(997),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(286),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(300),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_RBRACE] = ACTIONS(811),
    [anon_sym_BANG] = ACTIONS(302),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(304),
    [anon_sym_typeset] = ACTIONS(304),
    [anon_sym_export] = ACTIONS(304),
    [anon_sym_readonly] = ACTIONS(304),
    [anon_sym_local] = ACTIONS(304),
    [anon_sym_unset] = ACTIONS(306),
    [anon_sym_unsetenv] = ACTIONS(306),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [64] = {
    [sym__statements] = STATE(7356),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2191),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [65] = {
    [sym__statements] = STATE(7355),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [66] = {
    [sym__statements] = STATE(7354),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [67] = {
    [sym__statements] = STATE(7353),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [68] = {
    [aux_sym__terminated_statement] = STATE(62),
    [sym__statement_not_pipeline] = STATE(6718),
    [sym_redirected_statement] = STATE(4993),
    [sym_for_statement] = STATE(4993),
    [sym_c_style_for_statement] = STATE(4993),
    [sym_while_statement] = STATE(4489),
    [sym_if_statement] = STATE(4489),
    [sym_case_statement] = STATE(4993),
    [sym_function_definition] = STATE(4993),
    [sym_compound_statement] = STATE(4993),
    [sym_subshell] = STATE(4993),
    [sym_pipeline] = STATE(5122),
    [sym_list] = STATE(4993),
    [sym_negated_command] = STATE(4993),
    [sym_test_command] = STATE(4993),
    [sym_declaration_command] = STATE(4993),
    [sym_unset_command] = STATE(4993),
    [sym_command] = STATE(4993),
    [sym_command_name] = STATE(544),
    [sym_variable_assignment] = STATE(1030),
    [sym_variable_assignments] = STATE(4993),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym_redirected_statement_repeat2] = STATE(4644),
    [aux_sym_command_repeat1] = STATE(997),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(286),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(300),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_RBRACE] = ACTIONS(813),
    [anon_sym_BANG] = ACTIONS(302),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(304),
    [anon_sym_typeset] = ACTIONS(304),
    [anon_sym_export] = ACTIONS(304),
    [anon_sym_readonly] = ACTIONS(304),
    [anon_sym_local] = ACTIONS(304),
    [anon_sym_unset] = ACTIONS(306),
    [anon_sym_unsetenv] = ACTIONS(306),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [69] = {
    [sym__statements] = STATE(7127),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2186),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [70] = {
    [aux_sym__terminated_statement] = STATE(71),
    [sym__statement_not_pipeline] = STATE(6718),
    [sym_redirected_statement] = STATE(4951),
    [sym_for_statement] = STATE(4951),
    [sym_c_style_for_statement] = STATE(4951),
    [sym_while_statement] = STATE(4567),
    [sym_if_statement] = STATE(4567),
    [sym_case_statement] = STATE(4951),
    [sym_function_definition] = STATE(4951),
    [sym_compound_statement] = STATE(4951),
    [sym_subshell] = STATE(4951),
    [sym_pipeline] = STATE(5163),
    [sym_list] = STATE(4951),
    [sym_negated_command] = STATE(4951),
    [sym_test_command] = STATE(4951),
    [sym_declaration_command] = STATE(4951),
    [sym_unset_command] = STATE(4951),
    [sym_command] = STATE(4951),
    [sym_command_name] = STATE(544),
    [sym_variable_assignment] = STATE(1153),
    [sym_variable_assignments] = STATE(4951),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym_redirected_statement_repeat2] = STATE(4644),
    [aux_sym_command_repeat1] = STATE(997),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(286),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_done] = ACTIONS(815),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(300),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(302),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(304),
    [anon_sym_typeset] = ACTIONS(304),
    [anon_sym_export] = ACTIONS(304),
    [anon_sym_readonly] = ACTIONS(304),
    [anon_sym_local] = ACTIONS(304),
    [anon_sym_unset] = ACTIONS(306),
    [anon_sym_unsetenv] = ACTIONS(306),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [71] = {
    [aux_sym__terminated_statement] = STATE(105),
    [sym__statement_not_pipeline] = STATE(6718),
    [sym_redirected_statement] = STATE(4951),
    [sym_for_statement] = STATE(4951),
    [sym_c_style_for_statement] = STATE(4951),
    [sym_while_statement] = STATE(4567),
    [sym_if_statement] = STATE(4567),
    [sym_case_statement] = STATE(4951),
    [sym_function_definition] = STATE(4951),
    [sym_compound_statement] = STATE(4951),
    [sym_subshell] = STATE(4951),
    [sym_pipeline] = STATE(5163),
    [sym_list] = STATE(4951),
    [sym_negated_command] = STATE(4951),
    [sym_test_command] = STATE(4951),
    [sym_declaration_command] = STATE(4951),
    [sym_unset_command] = STATE(4951),
    [sym_command] = STATE(4951),
    [sym_command_name] = STATE(544),
    [sym_variable_assignment] = STATE(1153),
    [sym_variable_assignments] = STATE(4951),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym_redirected_statement_repeat2] = STATE(4644),
    [aux_sym_command_repeat1] = STATE(997),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(286),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_done] = ACTIONS(817),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(300),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(302),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(304),
    [anon_sym_typeset] = ACTIONS(304),
    [anon_sym_export] = ACTIONS(304),
    [anon_sym_readonly] = ACTIONS(304),
    [anon_sym_local] = ACTIONS(304),
    [anon_sym_unset] = ACTIONS(306),
    [anon_sym_unsetenv] = ACTIONS(306),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [72] = {
    [aux_sym__terminated_statement] = STATE(74),
    [sym__statement_not_pipeline] = STATE(6718),
    [sym_redirected_statement] = STATE(4993),
    [sym_for_statement] = STATE(4993),
    [sym_c_style_for_statement] = STATE(4993),
    [sym_while_statement] = STATE(4489),
    [sym_if_statement] = STATE(4489),
    [sym_case_statement] = STATE(4993),
    [sym_function_definition] = STATE(4993),
    [sym_compound_statement] = STATE(4993),
    [sym_subshell] = STATE(4993),
    [sym_pipeline] = STATE(5122),
    [sym_list] = STATE(4993),
    [sym_negated_command] = STATE(4993),
    [sym_test_command] = STATE(4993),
    [sym_declaration_command] = STATE(4993),
    [sym_unset_command] = STATE(4993),
    [sym_command] = STATE(4993),
    [sym_command_name] = STATE(544),
    [sym_variable_assignment] = STATE(1030),
    [sym_variable_assignments] = STATE(4993),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym_redirected_statement_repeat2] = STATE(4644),
    [aux_sym_command_repeat1] = STATE(997),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(286),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(300),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_RBRACE] = ACTIONS(819),
    [anon_sym_BANG] = ACTIONS(302),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(304),
    [anon_sym_typeset] = ACTIONS(304),
    [anon_sym_export] = ACTIONS(304),
    [anon_sym_readonly] = ACTIONS(304),
    [anon_sym_local] = ACTIONS(304),
    [anon_sym_unset] = ACTIONS(306),
    [anon_sym_unsetenv] = ACTIONS(306),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [73] = {
    [sym__statements] = STATE(7124),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [74] = {
    [aux_sym__terminated_statement] = STATE(62),
    [sym__statement_not_pipeline] = STATE(6718),
    [sym_redirected_statement] = STATE(4993),
    [sym_for_statement] = STATE(4993),
    [sym_c_style_for_statement] = STATE(4993),
    [sym_while_statement] = STATE(4489),
    [sym_if_statement] = STATE(4489),
    [sym_case_statement] = STATE(4993),
    [sym_function_definition] = STATE(4993),
    [sym_compound_statement] = STATE(4993),
    [sym_subshell] = STATE(4993),
    [sym_pipeline] = STATE(5122),
    [sym_list] = STATE(4993),
    [sym_negated_command] = STATE(4993),
    [sym_test_command] = STATE(4993),
    [sym_declaration_command] = STATE(4993),
    [sym_unset_command] = STATE(4993),
    [sym_command] = STATE(4993),
    [sym_command_name] = STATE(544),
    [sym_variable_assignment] = STATE(1030),
    [sym_variable_assignments] = STATE(4993),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym_redirected_statement_repeat2] = STATE(4644),
    [aux_sym_command_repeat1] = STATE(997),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(286),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(300),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_RBRACE] = ACTIONS(821),
    [anon_sym_BANG] = ACTIONS(302),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(304),
    [anon_sym_typeset] = ACTIONS(304),
    [anon_sym_export] = ACTIONS(304),
    [anon_sym_readonly] = ACTIONS(304),
    [anon_sym_local] = ACTIONS(304),
    [anon_sym_unset] = ACTIONS(306),
    [anon_sym_unsetenv] = ACTIONS(306),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [75] = {
    [sym__statements] = STATE(7123),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [76] = {
    [aux_sym__terminated_statement] = STATE(78),
    [sym__statement_not_pipeline] = STATE(6718),
    [sym_redirected_statement] = STATE(4951),
    [sym_for_statement] = STATE(4951),
    [sym_c_style_for_statement] = STATE(4951),
    [sym_while_statement] = STATE(4567),
    [sym_if_statement] = STATE(4567),
    [sym_case_statement] = STATE(4951),
    [sym_function_definition] = STATE(4951),
    [sym_compound_statement] = STATE(4951),
    [sym_subshell] = STATE(4951),
    [sym_pipeline] = STATE(5163),
    [sym_list] = STATE(4951),
    [sym_negated_command] = STATE(4951),
    [sym_test_command] = STATE(4951),
    [sym_declaration_command] = STATE(4951),
    [sym_unset_command] = STATE(4951),
    [sym_command] = STATE(4951),
    [sym_command_name] = STATE(544),
    [sym_variable_assignment] = STATE(1153),
    [sym_variable_assignments] = STATE(4951),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym_redirected_statement_repeat2] = STATE(4644),
    [aux_sym_command_repeat1] = STATE(997),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(286),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_done] = ACTIONS(823),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(300),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(302),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(304),
    [anon_sym_typeset] = ACTIONS(304),
    [anon_sym_export] = ACTIONS(304),
    [anon_sym_readonly] = ACTIONS(304),
    [anon_sym_local] = ACTIONS(304),
    [anon_sym_unset] = ACTIONS(306),
    [anon_sym_unsetenv] = ACTIONS(306),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [77] = {
    [sym__statements] = STATE(7120),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [78] = {
    [aux_sym__terminated_statement] = STATE(105),
    [sym__statement_not_pipeline] = STATE(6718),
    [sym_redirected_statement] = STATE(4951),
    [sym_for_statement] = STATE(4951),
    [sym_c_style_for_statement] = STATE(4951),
    [sym_while_statement] = STATE(4567),
    [sym_if_statement] = STATE(4567),
    [sym_case_statement] = STATE(4951),
    [sym_function_definition] = STATE(4951),
    [sym_compound_statement] = STATE(4951),
    [sym_subshell] = STATE(4951),
    [sym_pipeline] = STATE(5163),
    [sym_list] = STATE(4951),
    [sym_negated_command] = STATE(4951),
    [sym_test_command] = STATE(4951),
    [sym_declaration_command] = STATE(4951),
    [sym_unset_command] = STATE(4951),
    [sym_command] = STATE(4951),
    [sym_command_name] = STATE(544),
    [sym_variable_assignment] = STATE(1153),
    [sym_variable_assignments] = STATE(4951),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym_redirected_statement_repeat2] = STATE(4644),
    [aux_sym_command_repeat1] = STATE(997),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(286),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_done] = ACTIONS(825),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(300),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(302),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(304),
    [anon_sym_typeset] = ACTIONS(304),
    [anon_sym_export] = ACTIONS(304),
    [anon_sym_readonly] = ACTIONS(304),
    [anon_sym_local] = ACTIONS(304),
    [anon_sym_unset] = ACTIONS(306),
    [anon_sym_unsetenv] = ACTIONS(306),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [79] = {
    [sym__statements] = STATE(7103),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2183),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [80] = {
    [sym__statements] = STATE(7100),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [81] = {
    [aux_sym__terminated_statement] = STATE(82),
    [sym__statement_not_pipeline] = STATE(6718),
    [sym_redirected_statement] = STATE(4993),
    [sym_for_statement] = STATE(4993),
    [sym_c_style_for_statement] = STATE(4993),
    [sym_while_statement] = STATE(4489),
    [sym_if_statement] = STATE(4489),
    [sym_case_statement] = STATE(4993),
    [sym_function_definition] = STATE(4993),
    [sym_compound_statement] = STATE(4993),
    [sym_subshell] = STATE(4993),
    [sym_pipeline] = STATE(5122),
    [sym_list] = STATE(4993),
    [sym_negated_command] = STATE(4993),
    [sym_test_command] = STATE(4993),
    [sym_declaration_command] = STATE(4993),
    [sym_unset_command] = STATE(4993),
    [sym_command] = STATE(4993),
    [sym_command_name] = STATE(544),
    [sym_variable_assignment] = STATE(1030),
    [sym_variable_assignments] = STATE(4993),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym_redirected_statement_repeat2] = STATE(4644),
    [aux_sym_command_repeat1] = STATE(997),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(286),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(300),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_RBRACE] = ACTIONS(827),
    [anon_sym_BANG] = ACTIONS(302),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(304),
    [anon_sym_typeset] = ACTIONS(304),
    [anon_sym_export] = ACTIONS(304),
    [anon_sym_readonly] = ACTIONS(304),
    [anon_sym_local] = ACTIONS(304),
    [anon_sym_unset] = ACTIONS(306),
    [anon_sym_unsetenv] = ACTIONS(306),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [82] = {
    [aux_sym__terminated_statement] = STATE(62),
    [sym__statement_not_pipeline] = STATE(6718),
    [sym_redirected_statement] = STATE(4993),
    [sym_for_statement] = STATE(4993),
    [sym_c_style_for_statement] = STATE(4993),
    [sym_while_statement] = STATE(4489),
    [sym_if_statement] = STATE(4489),
    [sym_case_statement] = STATE(4993),
    [sym_function_definition] = STATE(4993),
    [sym_compound_statement] = STATE(4993),
    [sym_subshell] = STATE(4993),
    [sym_pipeline] = STATE(5122),
    [sym_list] = STATE(4993),
    [sym_negated_command] = STATE(4993),
    [sym_test_command] = STATE(4993),
    [sym_declaration_command] = STATE(4993),
    [sym_unset_command] = STATE(4993),
    [sym_command] = STATE(4993),
    [sym_command_name] = STATE(544),
    [sym_variable_assignment] = STATE(1030),
    [sym_variable_assignments] = STATE(4993),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym_redirected_statement_repeat2] = STATE(4644),
    [aux_sym_command_repeat1] = STATE(997),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(286),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(300),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_RBRACE] = ACTIONS(829),
    [anon_sym_BANG] = ACTIONS(302),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(304),
    [anon_sym_typeset] = ACTIONS(304),
    [anon_sym_export] = ACTIONS(304),
    [anon_sym_readonly] = ACTIONS(304),
    [anon_sym_local] = ACTIONS(304),
    [anon_sym_unset] = ACTIONS(306),
    [anon_sym_unsetenv] = ACTIONS(306),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [83] = {
    [sym__statements] = STATE(7099),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [84] = {
    [aux_sym__terminated_statement] = STATE(85),
    [sym__statement_not_pipeline] = STATE(6718),
    [sym_redirected_statement] = STATE(4951),
    [sym_for_statement] = STATE(4951),
    [sym_c_style_for_statement] = STATE(4951),
    [sym_while_statement] = STATE(4567),
    [sym_if_statement] = STATE(4567),
    [sym_case_statement] = STATE(4951),
    [sym_function_definition] = STATE(4951),
    [sym_compound_statement] = STATE(4951),
    [sym_subshell] = STATE(4951),
    [sym_pipeline] = STATE(5163),
    [sym_list] = STATE(4951),
    [sym_negated_command] = STATE(4951),
    [sym_test_command] = STATE(4951),
    [sym_declaration_command] = STATE(4951),
    [sym_unset_command] = STATE(4951),
    [sym_command] = STATE(4951),
    [sym_command_name] = STATE(544),
    [sym_variable_assignment] = STATE(1153),
    [sym_variable_assignments] = STATE(4951),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym_redirected_statement_repeat2] = STATE(4644),
    [aux_sym_command_repeat1] = STATE(997),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(286),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_done] = ACTIONS(831),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(300),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(302),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(304),
    [anon_sym_typeset] = ACTIONS(304),
    [anon_sym_export] = ACTIONS(304),
    [anon_sym_readonly] = ACTIONS(304),
    [anon_sym_local] = ACTIONS(304),
    [anon_sym_unset] = ACTIONS(306),
    [anon_sym_unsetenv] = ACTIONS(306),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [85] = {
    [aux_sym__terminated_statement] = STATE(105),
    [sym__statement_not_pipeline] = STATE(6718),
    [sym_redirected_statement] = STATE(4951),
    [sym_for_statement] = STATE(4951),
    [sym_c_style_for_statement] = STATE(4951),
    [sym_while_statement] = STATE(4567),
    [sym_if_statement] = STATE(4567),
    [sym_case_statement] = STATE(4951),
    [sym_function_definition] = STATE(4951),
    [sym_compound_statement] = STATE(4951),
    [sym_subshell] = STATE(4951),
    [sym_pipeline] = STATE(5163),
    [sym_list] = STATE(4951),
    [sym_negated_command] = STATE(4951),
    [sym_test_command] = STATE(4951),
    [sym_declaration_command] = STATE(4951),
    [sym_unset_command] = STATE(4951),
    [sym_command] = STATE(4951),
    [sym_command_name] = STATE(544),
    [sym_variable_assignment] = STATE(1153),
    [sym_variable_assignments] = STATE(4951),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym_redirected_statement_repeat2] = STATE(4644),
    [aux_sym_command_repeat1] = STATE(997),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(286),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_done] = ACTIONS(833),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(300),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(302),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(304),
    [anon_sym_typeset] = ACTIONS(304),
    [anon_sym_export] = ACTIONS(304),
    [anon_sym_readonly] = ACTIONS(304),
    [anon_sym_local] = ACTIONS(304),
    [anon_sym_unset] = ACTIONS(306),
    [anon_sym_unsetenv] = ACTIONS(306),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [86] = {
    [aux_sym__terminated_statement] = STATE(87),
    [sym__statement_not_pipeline] = STATE(6718),
    [sym_redirected_statement] = STATE(4993),
    [sym_for_statement] = STATE(4993),
    [sym_c_style_for_statement] = STATE(4993),
    [sym_while_statement] = STATE(4489),
    [sym_if_statement] = STATE(4489),
    [sym_case_statement] = STATE(4993),
    [sym_function_definition] = STATE(4993),
    [sym_compound_statement] = STATE(4993),
    [sym_subshell] = STATE(4993),
    [sym_pipeline] = STATE(5122),
    [sym_list] = STATE(4993),
    [sym_negated_command] = STATE(4993),
    [sym_test_command] = STATE(4993),
    [sym_declaration_command] = STATE(4993),
    [sym_unset_command] = STATE(4993),
    [sym_command] = STATE(4993),
    [sym_command_name] = STATE(544),
    [sym_variable_assignment] = STATE(1030),
    [sym_variable_assignments] = STATE(4993),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym_redirected_statement_repeat2] = STATE(4644),
    [aux_sym_command_repeat1] = STATE(997),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(286),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(300),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_RBRACE] = ACTIONS(835),
    [anon_sym_BANG] = ACTIONS(302),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(304),
    [anon_sym_typeset] = ACTIONS(304),
    [anon_sym_export] = ACTIONS(304),
    [anon_sym_readonly] = ACTIONS(304),
    [anon_sym_local] = ACTIONS(304),
    [anon_sym_unset] = ACTIONS(306),
    [anon_sym_unsetenv] = ACTIONS(306),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [87] = {
    [aux_sym__terminated_statement] = STATE(62),
    [sym__statement_not_pipeline] = STATE(6718),
    [sym_redirected_statement] = STATE(4993),
    [sym_for_statement] = STATE(4993),
    [sym_c_style_for_statement] = STATE(4993),
    [sym_while_statement] = STATE(4489),
    [sym_if_statement] = STATE(4489),
    [sym_case_statement] = STATE(4993),
    [sym_function_definition] = STATE(4993),
    [sym_compound_statement] = STATE(4993),
    [sym_subshell] = STATE(4993),
    [sym_pipeline] = STATE(5122),
    [sym_list] = STATE(4993),
    [sym_negated_command] = STATE(4993),
    [sym_test_command] = STATE(4993),
    [sym_declaration_command] = STATE(4993),
    [sym_unset_command] = STATE(4993),
    [sym_command] = STATE(4993),
    [sym_command_name] = STATE(544),
    [sym_variable_assignment] = STATE(1030),
    [sym_variable_assignments] = STATE(4993),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym_redirected_statement_repeat2] = STATE(4644),
    [aux_sym_command_repeat1] = STATE(997),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(286),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(300),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_RBRACE] = ACTIONS(837),
    [anon_sym_BANG] = ACTIONS(302),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(304),
    [anon_sym_typeset] = ACTIONS(304),
    [anon_sym_export] = ACTIONS(304),
    [anon_sym_readonly] = ACTIONS(304),
    [anon_sym_local] = ACTIONS(304),
    [anon_sym_unset] = ACTIONS(306),
    [anon_sym_unsetenv] = ACTIONS(306),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [88] = {
    [sym__statements] = STATE(7098),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [89] = {
    [sym__statements] = STATE(7047),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2181),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [90] = {
    [aux_sym__terminated_statement] = STATE(91),
    [sym__statement_not_pipeline] = STATE(6718),
    [sym_redirected_statement] = STATE(4951),
    [sym_for_statement] = STATE(4951),
    [sym_c_style_for_statement] = STATE(4951),
    [sym_while_statement] = STATE(4567),
    [sym_if_statement] = STATE(4567),
    [sym_case_statement] = STATE(4951),
    [sym_function_definition] = STATE(4951),
    [sym_compound_statement] = STATE(4951),
    [sym_subshell] = STATE(4951),
    [sym_pipeline] = STATE(5163),
    [sym_list] = STATE(4951),
    [sym_negated_command] = STATE(4951),
    [sym_test_command] = STATE(4951),
    [sym_declaration_command] = STATE(4951),
    [sym_unset_command] = STATE(4951),
    [sym_command] = STATE(4951),
    [sym_command_name] = STATE(544),
    [sym_variable_assignment] = STATE(1153),
    [sym_variable_assignments] = STATE(4951),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym_redirected_statement_repeat2] = STATE(4644),
    [aux_sym_command_repeat1] = STATE(997),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(286),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_done] = ACTIONS(839),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(300),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(302),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(304),
    [anon_sym_typeset] = ACTIONS(304),
    [anon_sym_export] = ACTIONS(304),
    [anon_sym_readonly] = ACTIONS(304),
    [anon_sym_local] = ACTIONS(304),
    [anon_sym_unset] = ACTIONS(306),
    [anon_sym_unsetenv] = ACTIONS(306),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [91] = {
    [aux_sym__terminated_statement] = STATE(105),
    [sym__statement_not_pipeline] = STATE(6718),
    [sym_redirected_statement] = STATE(4951),
    [sym_for_statement] = STATE(4951),
    [sym_c_style_for_statement] = STATE(4951),
    [sym_while_statement] = STATE(4567),
    [sym_if_statement] = STATE(4567),
    [sym_case_statement] = STATE(4951),
    [sym_function_definition] = STATE(4951),
    [sym_compound_statement] = STATE(4951),
    [sym_subshell] = STATE(4951),
    [sym_pipeline] = STATE(5163),
    [sym_list] = STATE(4951),
    [sym_negated_command] = STATE(4951),
    [sym_test_command] = STATE(4951),
    [sym_declaration_command] = STATE(4951),
    [sym_unset_command] = STATE(4951),
    [sym_command] = STATE(4951),
    [sym_command_name] = STATE(544),
    [sym_variable_assignment] = STATE(1153),
    [sym_variable_assignments] = STATE(4951),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym_redirected_statement_repeat2] = STATE(4644),
    [aux_sym_command_repeat1] = STATE(997),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(286),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_done] = ACTIONS(841),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(300),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(302),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(304),
    [anon_sym_typeset] = ACTIONS(304),
    [anon_sym_export] = ACTIONS(304),
    [anon_sym_readonly] = ACTIONS(304),
    [anon_sym_local] = ACTIONS(304),
    [anon_sym_unset] = ACTIONS(306),
    [anon_sym_unsetenv] = ACTIONS(306),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [92] = {
    [aux_sym__terminated_statement] = STATE(93),
    [sym__statement_not_pipeline] = STATE(6718),
    [sym_redirected_statement] = STATE(4993),
    [sym_for_statement] = STATE(4993),
    [sym_c_style_for_statement] = STATE(4993),
    [sym_while_statement] = STATE(4489),
    [sym_if_statement] = STATE(4489),
    [sym_case_statement] = STATE(4993),
    [sym_function_definition] = STATE(4993),
    [sym_compound_statement] = STATE(4993),
    [sym_subshell] = STATE(4993),
    [sym_pipeline] = STATE(5122),
    [sym_list] = STATE(4993),
    [sym_negated_command] = STATE(4993),
    [sym_test_command] = STATE(4993),
    [sym_declaration_command] = STATE(4993),
    [sym_unset_command] = STATE(4993),
    [sym_command] = STATE(4993),
    [sym_command_name] = STATE(544),
    [sym_variable_assignment] = STATE(1030),
    [sym_variable_assignments] = STATE(4993),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym_redirected_statement_repeat2] = STATE(4644),
    [aux_sym_command_repeat1] = STATE(997),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(286),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(300),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_RBRACE] = ACTIONS(843),
    [anon_sym_BANG] = ACTIONS(302),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(304),
    [anon_sym_typeset] = ACTIONS(304),
    [anon_sym_export] = ACTIONS(304),
    [anon_sym_readonly] = ACTIONS(304),
    [anon_sym_local] = ACTIONS(304),
    [anon_sym_unset] = ACTIONS(306),
    [anon_sym_unsetenv] = ACTIONS(306),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [93] = {
    [aux_sym__terminated_statement] = STATE(62),
    [sym__statement_not_pipeline] = STATE(6718),
    [sym_redirected_statement] = STATE(4993),
    [sym_for_statement] = STATE(4993),
    [sym_c_style_for_statement] = STATE(4993),
    [sym_while_statement] = STATE(4489),
    [sym_if_statement] = STATE(4489),
    [sym_case_statement] = STATE(4993),
    [sym_function_definition] = STATE(4993),
    [sym_compound_statement] = STATE(4993),
    [sym_subshell] = STATE(4993),
    [sym_pipeline] = STATE(5122),
    [sym_list] = STATE(4993),
    [sym_negated_command] = STATE(4993),
    [sym_test_command] = STATE(4993),
    [sym_declaration_command] = STATE(4993),
    [sym_unset_command] = STATE(4993),
    [sym_command] = STATE(4993),
    [sym_command_name] = STATE(544),
    [sym_variable_assignment] = STATE(1030),
    [sym_variable_assignments] = STATE(4993),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym_redirected_statement_repeat2] = STATE(4644),
    [aux_sym_command_repeat1] = STATE(997),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(286),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(300),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_RBRACE] = ACTIONS(845),
    [anon_sym_BANG] = ACTIONS(302),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(304),
    [anon_sym_typeset] = ACTIONS(304),
    [anon_sym_export] = ACTIONS(304),
    [anon_sym_readonly] = ACTIONS(304),
    [anon_sym_local] = ACTIONS(304),
    [anon_sym_unset] = ACTIONS(306),
    [anon_sym_unsetenv] = ACTIONS(306),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [94] = {
    [aux_sym__terminated_statement] = STATE(95),
    [sym__statement_not_pipeline] = STATE(6718),
    [sym_redirected_statement] = STATE(4951),
    [sym_for_statement] = STATE(4951),
    [sym_c_style_for_statement] = STATE(4951),
    [sym_while_statement] = STATE(4567),
    [sym_if_statement] = STATE(4567),
    [sym_case_statement] = STATE(4951),
    [sym_function_definition] = STATE(4951),
    [sym_compound_statement] = STATE(4951),
    [sym_subshell] = STATE(4951),
    [sym_pipeline] = STATE(5163),
    [sym_list] = STATE(4951),
    [sym_negated_command] = STATE(4951),
    [sym_test_command] = STATE(4951),
    [sym_declaration_command] = STATE(4951),
    [sym_unset_command] = STATE(4951),
    [sym_command] = STATE(4951),
    [sym_command_name] = STATE(544),
    [sym_variable_assignment] = STATE(1153),
    [sym_variable_assignments] = STATE(4951),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym_redirected_statement_repeat2] = STATE(4644),
    [aux_sym_command_repeat1] = STATE(997),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(286),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_done] = ACTIONS(847),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(300),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(302),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(304),
    [anon_sym_typeset] = ACTIONS(304),
    [anon_sym_export] = ACTIONS(304),
    [anon_sym_readonly] = ACTIONS(304),
    [anon_sym_local] = ACTIONS(304),
    [anon_sym_unset] = ACTIONS(306),
    [anon_sym_unsetenv] = ACTIONS(306),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [95] = {
    [aux_sym__terminated_statement] = STATE(105),
    [sym__statement_not_pipeline] = STATE(6718),
    [sym_redirected_statement] = STATE(4951),
    [sym_for_statement] = STATE(4951),
    [sym_c_style_for_statement] = STATE(4951),
    [sym_while_statement] = STATE(4567),
    [sym_if_statement] = STATE(4567),
    [sym_case_statement] = STATE(4951),
    [sym_function_definition] = STATE(4951),
    [sym_compound_statement] = STATE(4951),
    [sym_subshell] = STATE(4951),
    [sym_pipeline] = STATE(5163),
    [sym_list] = STATE(4951),
    [sym_negated_command] = STATE(4951),
    [sym_test_command] = STATE(4951),
    [sym_declaration_command] = STATE(4951),
    [sym_unset_command] = STATE(4951),
    [sym_command] = STATE(4951),
    [sym_command_name] = STATE(544),
    [sym_variable_assignment] = STATE(1153),
    [sym_variable_assignments] = STATE(4951),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym_redirected_statement_repeat2] = STATE(4644),
    [aux_sym_command_repeat1] = STATE(997),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(286),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_done] = ACTIONS(849),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(300),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(302),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(304),
    [anon_sym_typeset] = ACTIONS(304),
    [anon_sym_export] = ACTIONS(304),
    [anon_sym_readonly] = ACTIONS(304),
    [anon_sym_local] = ACTIONS(304),
    [anon_sym_unset] = ACTIONS(306),
    [anon_sym_unsetenv] = ACTIONS(306),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [96] = {
    [aux_sym__terminated_statement] = STATE(99),
    [sym__statement_not_pipeline] = STATE(6718),
    [sym_redirected_statement] = STATE(4993),
    [sym_for_statement] = STATE(4993),
    [sym_c_style_for_statement] = STATE(4993),
    [sym_while_statement] = STATE(4489),
    [sym_if_statement] = STATE(4489),
    [sym_case_statement] = STATE(4993),
    [sym_function_definition] = STATE(4993),
    [sym_compound_statement] = STATE(4993),
    [sym_subshell] = STATE(4993),
    [sym_pipeline] = STATE(5122),
    [sym_list] = STATE(4993),
    [sym_negated_command] = STATE(4993),
    [sym_test_command] = STATE(4993),
    [sym_declaration_command] = STATE(4993),
    [sym_unset_command] = STATE(4993),
    [sym_command] = STATE(4993),
    [sym_command_name] = STATE(544),
    [sym_variable_assignment] = STATE(1030),
    [sym_variable_assignments] = STATE(4993),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym_redirected_statement_repeat2] = STATE(4644),
    [aux_sym_command_repeat1] = STATE(997),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(286),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(300),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_RBRACE] = ACTIONS(851),
    [anon_sym_BANG] = ACTIONS(302),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(304),
    [anon_sym_typeset] = ACTIONS(304),
    [anon_sym_export] = ACTIONS(304),
    [anon_sym_readonly] = ACTIONS(304),
    [anon_sym_local] = ACTIONS(304),
    [anon_sym_unset] = ACTIONS(306),
    [anon_sym_unsetenv] = ACTIONS(306),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [97] = {
    [sym__statements] = STATE(7043),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [98] = {
    [sym__statements] = STATE(7034),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [99] = {
    [aux_sym__terminated_statement] = STATE(62),
    [sym__statement_not_pipeline] = STATE(6718),
    [sym_redirected_statement] = STATE(4993),
    [sym_for_statement] = STATE(4993),
    [sym_c_style_for_statement] = STATE(4993),
    [sym_while_statement] = STATE(4489),
    [sym_if_statement] = STATE(4489),
    [sym_case_statement] = STATE(4993),
    [sym_function_definition] = STATE(4993),
    [sym_compound_statement] = STATE(4993),
    [sym_subshell] = STATE(4993),
    [sym_pipeline] = STATE(5122),
    [sym_list] = STATE(4993),
    [sym_negated_command] = STATE(4993),
    [sym_test_command] = STATE(4993),
    [sym_declaration_command] = STATE(4993),
    [sym_unset_command] = STATE(4993),
    [sym_command] = STATE(4993),
    [sym_command_name] = STATE(544),
    [sym_variable_assignment] = STATE(1030),
    [sym_variable_assignments] = STATE(4993),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym_redirected_statement_repeat2] = STATE(4644),
    [aux_sym_command_repeat1] = STATE(997),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(286),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(300),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_RBRACE] = ACTIONS(853),
    [anon_sym_BANG] = ACTIONS(302),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(304),
    [anon_sym_typeset] = ACTIONS(304),
    [anon_sym_export] = ACTIONS(304),
    [anon_sym_readonly] = ACTIONS(304),
    [anon_sym_local] = ACTIONS(304),
    [anon_sym_unset] = ACTIONS(306),
    [anon_sym_unsetenv] = ACTIONS(306),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [100] = {
    [sym__statements] = STATE(7032),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [101] = {
    [sym__statements] = STATE(6888),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2180),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [102] = {
    [aux_sym__terminated_statement] = STATE(34),
    [sym__statement_not_pipeline] = STATE(6718),
    [sym_redirected_statement] = STATE(5007),
    [sym_for_statement] = STATE(5007),
    [sym_c_style_for_statement] = STATE(5007),
    [sym_while_statement] = STATE(4551),
    [sym_if_statement] = STATE(4551),
    [sym_case_statement] = STATE(5007),
    [sym_function_definition] = STATE(5007),
    [sym_compound_statement] = STATE(5007),
    [sym_subshell] = STATE(5007),
    [sym_pipeline] = STATE(5189),
    [sym_list] = STATE(5007),
    [sym_negated_command] = STATE(5007),
    [sym_test_command] = STATE(5007),
    [sym_declaration_command] = STATE(5007),
    [sym_unset_command] = STATE(5007),
    [sym_command] = STATE(5007),
    [sym_command_name] = STATE(544),
    [sym_variable_assignment] = STATE(1151),
    [sym_variable_assignments] = STATE(5007),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym_redirected_statement_repeat2] = STATE(4644),
    [aux_sym_command_repeat1] = STATE(997),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(286),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_then] = ACTIONS(855),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(300),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(302),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(304),
    [anon_sym_typeset] = ACTIONS(304),
    [anon_sym_export] = ACTIONS(304),
    [anon_sym_readonly] = ACTIONS(304),
    [anon_sym_local] = ACTIONS(304),
    [anon_sym_unset] = ACTIONS(306),
    [anon_sym_unsetenv] = ACTIONS(306),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [103] = {
    [aux_sym__terminated_statement] = STATE(104),
    [sym__statement_not_pipeline] = STATE(6718),
    [sym_redirected_statement] = STATE(4951),
    [sym_for_statement] = STATE(4951),
    [sym_c_style_for_statement] = STATE(4951),
    [sym_while_statement] = STATE(4567),
    [sym_if_statement] = STATE(4567),
    [sym_case_statement] = STATE(4951),
    [sym_function_definition] = STATE(4951),
    [sym_compound_statement] = STATE(4951),
    [sym_subshell] = STATE(4951),
    [sym_pipeline] = STATE(5163),
    [sym_list] = STATE(4951),
    [sym_negated_command] = STATE(4951),
    [sym_test_command] = STATE(4951),
    [sym_declaration_command] = STATE(4951),
    [sym_unset_command] = STATE(4951),
    [sym_command] = STATE(4951),
    [sym_command_name] = STATE(544),
    [sym_variable_assignment] = STATE(1153),
    [sym_variable_assignments] = STATE(4951),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym_redirected_statement_repeat2] = STATE(4644),
    [aux_sym_command_repeat1] = STATE(997),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(286),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_done] = ACTIONS(857),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(300),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(302),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(304),
    [anon_sym_typeset] = ACTIONS(304),
    [anon_sym_export] = ACTIONS(304),
    [anon_sym_readonly] = ACTIONS(304),
    [anon_sym_local] = ACTIONS(304),
    [anon_sym_unset] = ACTIONS(306),
    [anon_sym_unsetenv] = ACTIONS(306),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [104] = {
    [aux_sym__terminated_statement] = STATE(105),
    [sym__statement_not_pipeline] = STATE(6718),
    [sym_redirected_statement] = STATE(4951),
    [sym_for_statement] = STATE(4951),
    [sym_c_style_for_statement] = STATE(4951),
    [sym_while_statement] = STATE(4567),
    [sym_if_statement] = STATE(4567),
    [sym_case_statement] = STATE(4951),
    [sym_function_definition] = STATE(4951),
    [sym_compound_statement] = STATE(4951),
    [sym_subshell] = STATE(4951),
    [sym_pipeline] = STATE(5163),
    [sym_list] = STATE(4951),
    [sym_negated_command] = STATE(4951),
    [sym_test_command] = STATE(4951),
    [sym_declaration_command] = STATE(4951),
    [sym_unset_command] = STATE(4951),
    [sym_command] = STATE(4951),
    [sym_command_name] = STATE(544),
    [sym_variable_assignment] = STATE(1153),
    [sym_variable_assignments] = STATE(4951),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym_redirected_statement_repeat2] = STATE(4644),
    [aux_sym_command_repeat1] = STATE(997),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(286),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_done] = ACTIONS(859),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(300),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(302),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(304),
    [anon_sym_typeset] = ACTIONS(304),
    [anon_sym_export] = ACTIONS(304),
    [anon_sym_readonly] = ACTIONS(304),
    [anon_sym_local] = ACTIONS(304),
    [anon_sym_unset] = ACTIONS(306),
    [anon_sym_unsetenv] = ACTIONS(306),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [105] = {
    [aux_sym__terminated_statement] = STATE(105),
    [sym__statement_not_pipeline] = STATE(6718),
    [sym_redirected_statement] = STATE(4951),
    [sym_for_statement] = STATE(4951),
    [sym_c_style_for_statement] = STATE(4951),
    [sym_while_statement] = STATE(4567),
    [sym_if_statement] = STATE(4567),
    [sym_case_statement] = STATE(4951),
    [sym_function_definition] = STATE(4951),
    [sym_compound_statement] = STATE(4951),
    [sym_subshell] = STATE(4951),
    [sym_pipeline] = STATE(5163),
    [sym_list] = STATE(4951),
    [sym_negated_command] = STATE(4951),
    [sym_test_command] = STATE(4951),
    [sym_declaration_command] = STATE(4951),
    [sym_unset_command] = STATE(4951),
    [sym_command] = STATE(4951),
    [sym_command_name] = STATE(544),
    [sym_variable_assignment] = STATE(1153),
    [sym_variable_assignments] = STATE(4951),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym_redirected_statement_repeat2] = STATE(4644),
    [aux_sym_command_repeat1] = STATE(997),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(555),
    [anon_sym_for] = ACTIONS(558),
    [anon_sym_select] = ACTIONS(561),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(564),
    [anon_sym_LT] = ACTIONS(567),
    [anon_sym_GT] = ACTIONS(567),
    [anon_sym_GT_GT] = ACTIONS(570),
    [anon_sym_LPAREN] = ACTIONS(573),
    [anon_sym_while] = ACTIONS(576),
    [anon_sym_until] = ACTIONS(576),
    [anon_sym_done] = ACTIONS(579),
    [anon_sym_if] = ACTIONS(581),
    [anon_sym_case] = ACTIONS(584),
    [anon_sym_function] = ACTIONS(587),
    [anon_sym_LBRACE] = ACTIONS(590),
    [anon_sym_BANG] = ACTIONS(593),
    [anon_sym_LBRACK] = ACTIONS(596),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(599),
    [anon_sym_declare] = ACTIONS(602),
    [anon_sym_typeset] = ACTIONS(602),
    [anon_sym_export] = ACTIONS(602),
    [anon_sym_readonly] = ACTIONS(602),
    [anon_sym_local] = ACTIONS(602),
    [anon_sym_unset] = ACTIONS(605),
    [anon_sym_unsetenv] = ACTIONS(605),
    [anon_sym_AMP_GT] = ACTIONS(567),
    [anon_sym_AMP_GT_GT] = ACTIONS(570),
    [anon_sym_LT_AMP] = ACTIONS(567),
    [anon_sym_GT_AMP] = ACTIONS(567),
    [anon_sym_GT_PIPE] = ACTIONS(570),
    [anon_sym_LT_AMP_DASH] = ACTIONS(608),
    [anon_sym_GT_AMP_DASH] = ACTIONS(608),
    [anon_sym_LT_LT_LT] = ACTIONS(611),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(614),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(617),
    [anon_sym_DOLLAR] = ACTIONS(620),
    [sym__special_character] = ACTIONS(623),
    [anon_sym_DQUOTE] = ACTIONS(626),
    [sym_raw_string] = ACTIONS(629),
    [sym_ansi_c_string] = ACTIONS(629),
    [aux_sym_number_token1] = ACTIONS(632),
    [aux_sym_number_token2] = ACTIONS(635),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(638),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(641),
    [anon_sym_BQUOTE] = ACTIONS(644),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(647),
    [anon_sym_LT_LPAREN] = ACTIONS(650),
    [anon_sym_GT_LPAREN] = ACTIONS(650),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(653),
    [sym_variable_name] = ACTIONS(656),
    [sym_test_operator] = ACTIONS(659),
    [sym__brace_start] = ACTIONS(662),
  },
  [106] = {
    [sym__statements] = STATE(7083),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [107] = {
    [sym__statements] = STATE(6881),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [108] = {
    [sym__statements] = STATE(6880),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [109] = {
    [sym__statements] = STATE(6859),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [110] = {
    [sym__statements] = STATE(6831),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2179),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [111] = {
    [sym__statements] = STATE(6823),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [112] = {
    [sym__statements] = STATE(6914),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [113] = {
    [sym__statements] = STATE(6822),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [114] = {
    [sym__statements] = STATE(6927),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(1939),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [115] = {
    [sym__statements] = STATE(6818),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [116] = {
    [sym__statements] = STATE(6799),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2178),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [117] = {
    [sym__statements] = STATE(6930),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [118] = {
    [sym__statements] = STATE(6934),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [119] = {
    [sym__statements] = STATE(6794),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [120] = {
    [sym__statements] = STATE(6937),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [121] = {
    [sym__statements] = STATE(6808),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [122] = {
    [aux_sym__terminated_statement] = STATE(34),
    [sym__statement_not_pipeline] = STATE(6718),
    [sym_redirected_statement] = STATE(5007),
    [sym_for_statement] = STATE(5007),
    [sym_c_style_for_statement] = STATE(5007),
    [sym_while_statement] = STATE(4551),
    [sym_if_statement] = STATE(4551),
    [sym_case_statement] = STATE(5007),
    [sym_function_definition] = STATE(5007),
    [sym_compound_statement] = STATE(5007),
    [sym_subshell] = STATE(5007),
    [sym_pipeline] = STATE(5189),
    [sym_list] = STATE(5007),
    [sym_negated_command] = STATE(5007),
    [sym_test_command] = STATE(5007),
    [sym_declaration_command] = STATE(5007),
    [sym_unset_command] = STATE(5007),
    [sym_command] = STATE(5007),
    [sym_command_name] = STATE(544),
    [sym_variable_assignment] = STATE(1151),
    [sym_variable_assignments] = STATE(5007),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym_redirected_statement_repeat2] = STATE(4644),
    [aux_sym_command_repeat1] = STATE(997),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(286),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_then] = ACTIONS(861),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(300),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(302),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(304),
    [anon_sym_typeset] = ACTIONS(304),
    [anon_sym_export] = ACTIONS(304),
    [anon_sym_readonly] = ACTIONS(304),
    [anon_sym_local] = ACTIONS(304),
    [anon_sym_unset] = ACTIONS(306),
    [anon_sym_unsetenv] = ACTIONS(306),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [123] = {
    [sym__statements] = STATE(6801),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [124] = {
    [sym__statements] = STATE(6849),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2177),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [125] = {
    [sym__statements] = STATE(7117),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [126] = {
    [sym__statements] = STATE(6854),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [127] = {
    [sym__statements] = STATE(6855),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [128] = {
    [sym__statements] = STATE(7262),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2060),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [129] = {
    [sym__statements] = STATE(7277),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [130] = {
    [sym__statements] = STATE(7321),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [131] = {
    [sym__statements] = STATE(6857),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [132] = {
    [sym__statements] = STATE(7286),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [133] = {
    [aux_sym__terminated_statement] = STATE(34),
    [sym__statement_not_pipeline] = STATE(6718),
    [sym_redirected_statement] = STATE(5007),
    [sym_for_statement] = STATE(5007),
    [sym_c_style_for_statement] = STATE(5007),
    [sym_while_statement] = STATE(4551),
    [sym_if_statement] = STATE(4551),
    [sym_case_statement] = STATE(5007),
    [sym_function_definition] = STATE(5007),
    [sym_compound_statement] = STATE(5007),
    [sym_subshell] = STATE(5007),
    [sym_pipeline] = STATE(5189),
    [sym_list] = STATE(5007),
    [sym_negated_command] = STATE(5007),
    [sym_test_command] = STATE(5007),
    [sym_declaration_command] = STATE(5007),
    [sym_unset_command] = STATE(5007),
    [sym_command] = STATE(5007),
    [sym_command_name] = STATE(544),
    [sym_variable_assignment] = STATE(1151),
    [sym_variable_assignments] = STATE(5007),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym_redirected_statement_repeat2] = STATE(4644),
    [aux_sym_command_repeat1] = STATE(997),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(286),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_then] = ACTIONS(863),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(300),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(302),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(304),
    [anon_sym_typeset] = ACTIONS(304),
    [anon_sym_export] = ACTIONS(304),
    [anon_sym_readonly] = ACTIONS(304),
    [anon_sym_local] = ACTIONS(304),
    [anon_sym_unset] = ACTIONS(306),
    [anon_sym_unsetenv] = ACTIONS(306),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [134] = {
    [sym__statements] = STATE(6809),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2176),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [135] = {
    [sym__statements] = STATE(6897),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [136] = {
    [sym__statements] = STATE(7203),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [137] = {
    [aux_sym__terminated_statement] = STATE(105),
    [sym__statement_not_pipeline] = STATE(6718),
    [sym_redirected_statement] = STATE(4951),
    [sym_for_statement] = STATE(4951),
    [sym_c_style_for_statement] = STATE(4951),
    [sym_while_statement] = STATE(4567),
    [sym_if_statement] = STATE(4567),
    [sym_case_statement] = STATE(4951),
    [sym_function_definition] = STATE(4951),
    [sym_compound_statement] = STATE(4951),
    [sym_subshell] = STATE(4951),
    [sym_pipeline] = STATE(5163),
    [sym_list] = STATE(4951),
    [sym_negated_command] = STATE(4951),
    [sym_test_command] = STATE(4951),
    [sym_declaration_command] = STATE(4951),
    [sym_unset_command] = STATE(4951),
    [sym_command] = STATE(4951),
    [sym_command_name] = STATE(544),
    [sym_variable_assignment] = STATE(1153),
    [sym_variable_assignments] = STATE(4951),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym_redirected_statement_repeat2] = STATE(4644),
    [aux_sym_command_repeat1] = STATE(997),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(286),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_done] = ACTIONS(865),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(300),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(302),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(304),
    [anon_sym_typeset] = ACTIONS(304),
    [anon_sym_export] = ACTIONS(304),
    [anon_sym_readonly] = ACTIONS(304),
    [anon_sym_local] = ACTIONS(304),
    [anon_sym_unset] = ACTIONS(306),
    [anon_sym_unsetenv] = ACTIONS(306),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [138] = {
    [sym__statements] = STATE(6911),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [139] = {
    [sym__statements] = STATE(6915),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [140] = {
    [sym__statements] = STATE(7147),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2175),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [141] = {
    [sym__statements] = STATE(7040),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [142] = {
    [sym__statements] = STATE(7033),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [143] = {
    [aux_sym__terminated_statement] = STATE(137),
    [sym__statement_not_pipeline] = STATE(6718),
    [sym_redirected_statement] = STATE(4951),
    [sym_for_statement] = STATE(4951),
    [sym_c_style_for_statement] = STATE(4951),
    [sym_while_statement] = STATE(4567),
    [sym_if_statement] = STATE(4567),
    [sym_case_statement] = STATE(4951),
    [sym_function_definition] = STATE(4951),
    [sym_compound_statement] = STATE(4951),
    [sym_subshell] = STATE(4951),
    [sym_pipeline] = STATE(5163),
    [sym_list] = STATE(4951),
    [sym_negated_command] = STATE(4951),
    [sym_test_command] = STATE(4951),
    [sym_declaration_command] = STATE(4951),
    [sym_unset_command] = STATE(4951),
    [sym_command] = STATE(4951),
    [sym_command_name] = STATE(544),
    [sym_variable_assignment] = STATE(1153),
    [sym_variable_assignments] = STATE(4951),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym_redirected_statement_repeat2] = STATE(4644),
    [aux_sym_command_repeat1] = STATE(997),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(286),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_done] = ACTIONS(867),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(300),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(302),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(304),
    [anon_sym_typeset] = ACTIONS(304),
    [anon_sym_export] = ACTIONS(304),
    [anon_sym_readonly] = ACTIONS(304),
    [anon_sym_local] = ACTIONS(304),
    [anon_sym_unset] = ACTIONS(306),
    [anon_sym_unsetenv] = ACTIONS(306),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [144] = {
    [sym__statements] = STATE(7092),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [145] = {
    [sym__statements] = STATE(7350),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [146] = {
    [sym__statements] = STATE(7093),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [147] = {
    [sym__statements] = STATE(7144),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2172),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [148] = {
    [sym__statements] = STATE(6951),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2120),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [149] = {
    [sym__statements] = STATE(7250),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [150] = {
    [sym__statements] = STATE(6943),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [151] = {
    [sym__statements] = STATE(6850),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [152] = {
    [sym__statements] = STATE(6893),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [153] = {
    [aux_sym__terminated_statement] = STATE(34),
    [sym__statement_not_pipeline] = STATE(6718),
    [sym_redirected_statement] = STATE(5007),
    [sym_for_statement] = STATE(5007),
    [sym_c_style_for_statement] = STATE(5007),
    [sym_while_statement] = STATE(4551),
    [sym_if_statement] = STATE(4551),
    [sym_case_statement] = STATE(5007),
    [sym_function_definition] = STATE(5007),
    [sym_compound_statement] = STATE(5007),
    [sym_subshell] = STATE(5007),
    [sym_pipeline] = STATE(5189),
    [sym_list] = STATE(5007),
    [sym_negated_command] = STATE(5007),
    [sym_test_command] = STATE(5007),
    [sym_declaration_command] = STATE(5007),
    [sym_unset_command] = STATE(5007),
    [sym_command] = STATE(5007),
    [sym_command_name] = STATE(544),
    [sym_variable_assignment] = STATE(1151),
    [sym_variable_assignments] = STATE(5007),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym_redirected_statement_repeat2] = STATE(4644),
    [aux_sym_command_repeat1] = STATE(997),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(286),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_then] = ACTIONS(869),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(300),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(302),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(304),
    [anon_sym_typeset] = ACTIONS(304),
    [anon_sym_export] = ACTIONS(304),
    [anon_sym_readonly] = ACTIONS(304),
    [anon_sym_local] = ACTIONS(304),
    [anon_sym_unset] = ACTIONS(306),
    [anon_sym_unsetenv] = ACTIONS(306),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [154] = {
    [aux_sym__terminated_statement] = STATE(282),
    [sym__statement_not_pipeline] = STATE(6718),
    [sym_redirected_statement] = STATE(4993),
    [sym_for_statement] = STATE(4993),
    [sym_c_style_for_statement] = STATE(4993),
    [sym_while_statement] = STATE(4489),
    [sym_if_statement] = STATE(4489),
    [sym_case_statement] = STATE(4993),
    [sym_function_definition] = STATE(4993),
    [sym_compound_statement] = STATE(4993),
    [sym_subshell] = STATE(4993),
    [sym_pipeline] = STATE(5122),
    [sym_list] = STATE(4993),
    [sym_negated_command] = STATE(4993),
    [sym_test_command] = STATE(4993),
    [sym_declaration_command] = STATE(4993),
    [sym_unset_command] = STATE(4993),
    [sym_command] = STATE(4993),
    [sym_command_name] = STATE(544),
    [sym_variable_assignment] = STATE(1030),
    [sym_variable_assignments] = STATE(4993),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym_redirected_statement_repeat2] = STATE(4644),
    [aux_sym_command_repeat1] = STATE(997),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(286),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(300),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_RBRACE] = ACTIONS(871),
    [anon_sym_BANG] = ACTIONS(302),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(304),
    [anon_sym_typeset] = ACTIONS(304),
    [anon_sym_export] = ACTIONS(304),
    [anon_sym_readonly] = ACTIONS(304),
    [anon_sym_local] = ACTIONS(304),
    [anon_sym_unset] = ACTIONS(306),
    [anon_sym_unsetenv] = ACTIONS(306),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [155] = {
    [sym__statements] = STATE(6917),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [156] = {
    [sym__statements] = STATE(7208),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [157] = {
    [sym__statements] = STATE(7091),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [158] = {
    [sym__statements] = STATE(6879),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2164),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [159] = {
    [sym__statements] = STATE(6878),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [160] = {
    [sym__statements] = STATE(6877),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [161] = {
    [sym__statements] = STATE(6872),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [162] = {
    [sym__statements] = STATE(7165),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [163] = {
    [sym__statements] = STATE(7293),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [164] = {
    [sym__statements] = STATE(7200),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2171),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [165] = {
    [sym__statements] = STATE(7256),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2187),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [166] = {
    [sym__statements] = STATE(7148),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2185),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [167] = {
    [sym__statements] = STATE(7153),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [168] = {
    [sym__statements] = STATE(7209),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [169] = {
    [sym__statements] = STATE(7211),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [170] = {
    [sym__statements] = STATE(7158),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [171] = {
    [sym__statements] = STATE(7167),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [172] = {
    [sym__statements] = STATE(7224),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [173] = {
    [sym__statements] = STATE(7053),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [174] = {
    [sym__statements] = STATE(7336),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2200),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [175] = {
    [sym__statements] = STATE(7342),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [176] = {
    [sym__statements] = STATE(7347),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [177] = {
    [sym__statements] = STATE(7364),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [178] = {
    [sym__statements] = STATE(7299),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2170),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [179] = {
    [sym__statements] = STATE(7441),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [180] = {
    [sym__statements] = STATE(7599),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2228),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [181] = {
    [sym__statements] = STATE(7598),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [182] = {
    [sym__statements] = STATE(7294),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [183] = {
    [sym__statements] = STATE(7319),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [184] = {
    [sym__statements] = STATE(7536),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [185] = {
    [sym__statements] = STATE(7535),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [186] = {
    [sym__statements] = STATE(7520),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2236),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [187] = {
    [sym__statements] = STATE(7518),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [188] = {
    [sym__statements] = STATE(7515),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [189] = {
    [sym__statements] = STATE(7502),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [190] = {
    [sym__statements] = STATE(7460),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2262),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [191] = {
    [sym__statements] = STATE(7323),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [192] = {
    [sym__statements] = STATE(7458),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [193] = {
    [sym__statements] = STATE(7457),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [194] = {
    [sym__statements] = STATE(7456),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [195] = {
    [sym__statements] = STATE(7334),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [196] = {
    [sym__statements] = STATE(7432),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2265),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [197] = {
    [sym__statements] = STATE(7430),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [198] = {
    [sym__statements] = STATE(7429),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [199] = {
    [sym__statements] = STATE(7403),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2169),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [200] = {
    [sym__statements] = STATE(7426),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [201] = {
    [sym__statements] = STATE(7375),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2264),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [202] = {
    [sym__statements] = STATE(7371),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [203] = {
    [sym__statements] = STATE(7370),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [204] = {
    [sym__statements] = STATE(7369),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [205] = {
    [sym__statements] = STATE(7312),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2259),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [206] = {
    [sym__statements] = STATE(7307),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [207] = {
    [sym__statements] = STATE(7305),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [208] = {
    [sym__statements] = STATE(7414),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [209] = {
    [sym__statements] = STATE(7423),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [210] = {
    [sym__statements] = STATE(7424),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [211] = {
    [sym__statements] = STATE(7600),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2168),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [212] = {
    [sym__statements] = STATE(7606),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [213] = {
    [sym__statements] = STATE(7602),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [214] = {
    [sym__statements] = STATE(7304),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [215] = {
    [sym__statements] = STATE(7246),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(1931),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [216] = {
    [sym__statements] = STATE(7110),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [217] = {
    [sym__statements] = STATE(7438),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [218] = {
    [sym__statements] = STATE(7215),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [219] = {
    [sym__statements] = STATE(7210),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [220] = {
    [sym__statements] = STATE(7161),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2255),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [221] = {
    [sym__statements] = STATE(7156),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [222] = {
    [sym__statements] = STATE(7155),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [223] = {
    [sym__statements] = STATE(7154),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [224] = {
    [sym__statements] = STATE(7081),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2223),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [225] = {
    [sym__statements] = STATE(7075),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [226] = {
    [aux_sym__terminated_statement] = STATE(232),
    [sym__statement_not_pipeline] = STATE(6718),
    [sym_redirected_statement] = STATE(5007),
    [sym_for_statement] = STATE(5007),
    [sym_c_style_for_statement] = STATE(5007),
    [sym_while_statement] = STATE(4551),
    [sym_if_statement] = STATE(4551),
    [sym_case_statement] = STATE(5007),
    [sym_function_definition] = STATE(5007),
    [sym_compound_statement] = STATE(5007),
    [sym_subshell] = STATE(5007),
    [sym_pipeline] = STATE(5189),
    [sym_list] = STATE(5007),
    [sym_negated_command] = STATE(5007),
    [sym_test_command] = STATE(5007),
    [sym_declaration_command] = STATE(5007),
    [sym_unset_command] = STATE(5007),
    [sym_command] = STATE(5007),
    [sym_command_name] = STATE(544),
    [sym_variable_assignment] = STATE(1151),
    [sym_variable_assignments] = STATE(5007),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym_redirected_statement_repeat2] = STATE(4644),
    [aux_sym_command_repeat1] = STATE(997),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(286),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_fi] = ACTIONS(873),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(300),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(302),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(304),
    [anon_sym_typeset] = ACTIONS(304),
    [anon_sym_export] = ACTIONS(304),
    [anon_sym_readonly] = ACTIONS(304),
    [anon_sym_local] = ACTIONS(304),
    [anon_sym_unset] = ACTIONS(306),
    [anon_sym_unsetenv] = ACTIONS(306),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [227] = {
    [sym__statements] = STATE(7070),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [228] = {
    [sym__statements] = STATE(7067),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [229] = {
    [sym__statements] = STATE(6983),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2221),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [230] = {
    [sym__statements] = STATE(6976),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [231] = {
    [sym__statements] = STATE(6840),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [232] = {
    [aux_sym__terminated_statement] = STATE(34),
    [sym__statement_not_pipeline] = STATE(6718),
    [sym_redirected_statement] = STATE(5007),
    [sym_for_statement] = STATE(5007),
    [sym_c_style_for_statement] = STATE(5007),
    [sym_while_statement] = STATE(4551),
    [sym_if_statement] = STATE(4551),
    [sym_case_statement] = STATE(5007),
    [sym_function_definition] = STATE(5007),
    [sym_compound_statement] = STATE(5007),
    [sym_subshell] = STATE(5007),
    [sym_pipeline] = STATE(5189),
    [sym_list] = STATE(5007),
    [sym_negated_command] = STATE(5007),
    [sym_test_command] = STATE(5007),
    [sym_declaration_command] = STATE(5007),
    [sym_unset_command] = STATE(5007),
    [sym_command] = STATE(5007),
    [sym_command_name] = STATE(544),
    [sym_variable_assignment] = STATE(1151),
    [sym_variable_assignments] = STATE(5007),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym_redirected_statement_repeat2] = STATE(4644),
    [aux_sym_command_repeat1] = STATE(997),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(286),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_fi] = ACTIONS(875),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(300),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(302),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(304),
    [anon_sym_typeset] = ACTIONS(304),
    [anon_sym_export] = ACTIONS(304),
    [anon_sym_readonly] = ACTIONS(304),
    [anon_sym_local] = ACTIONS(304),
    [anon_sym_unset] = ACTIONS(306),
    [anon_sym_unsetenv] = ACTIONS(306),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [233] = {
    [sym__statements] = STATE(7531),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2167),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [234] = {
    [aux_sym__terminated_statement] = STATE(34),
    [sym__statement_not_pipeline] = STATE(6718),
    [sym_redirected_statement] = STATE(5007),
    [sym_for_statement] = STATE(5007),
    [sym_c_style_for_statement] = STATE(5007),
    [sym_while_statement] = STATE(4551),
    [sym_if_statement] = STATE(4551),
    [sym_case_statement] = STATE(5007),
    [sym_function_definition] = STATE(5007),
    [sym_compound_statement] = STATE(5007),
    [sym_subshell] = STATE(5007),
    [sym_pipeline] = STATE(5189),
    [sym_list] = STATE(5007),
    [sym_negated_command] = STATE(5007),
    [sym_test_command] = STATE(5007),
    [sym_declaration_command] = STATE(5007),
    [sym_unset_command] = STATE(5007),
    [sym_command] = STATE(5007),
    [sym_command_name] = STATE(544),
    [sym_variable_assignment] = STATE(1151),
    [sym_variable_assignments] = STATE(5007),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym_redirected_statement_repeat2] = STATE(4644),
    [aux_sym_command_repeat1] = STATE(997),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(286),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_then] = ACTIONS(877),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(300),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(302),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(304),
    [anon_sym_typeset] = ACTIONS(304),
    [anon_sym_export] = ACTIONS(304),
    [anon_sym_readonly] = ACTIONS(304),
    [anon_sym_local] = ACTIONS(304),
    [anon_sym_unset] = ACTIONS(306),
    [anon_sym_unsetenv] = ACTIONS(306),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [235] = {
    [sym__statements] = STATE(6963),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [236] = {
    [sym__statements] = STATE(6906),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2220),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [237] = {
    [aux_sym__terminated_statement] = STATE(105),
    [sym__statement_not_pipeline] = STATE(6718),
    [sym_redirected_statement] = STATE(4951),
    [sym_for_statement] = STATE(4951),
    [sym_c_style_for_statement] = STATE(4951),
    [sym_while_statement] = STATE(4567),
    [sym_if_statement] = STATE(4567),
    [sym_case_statement] = STATE(4951),
    [sym_function_definition] = STATE(4951),
    [sym_compound_statement] = STATE(4951),
    [sym_subshell] = STATE(4951),
    [sym_pipeline] = STATE(5163),
    [sym_list] = STATE(4951),
    [sym_negated_command] = STATE(4951),
    [sym_test_command] = STATE(4951),
    [sym_declaration_command] = STATE(4951),
    [sym_unset_command] = STATE(4951),
    [sym_command] = STATE(4951),
    [sym_command_name] = STATE(544),
    [sym_variable_assignment] = STATE(1153),
    [sym_variable_assignments] = STATE(4951),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym_redirected_statement_repeat2] = STATE(4644),
    [aux_sym_command_repeat1] = STATE(997),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(286),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_done] = ACTIONS(879),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(300),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(302),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(304),
    [anon_sym_typeset] = ACTIONS(304),
    [anon_sym_export] = ACTIONS(304),
    [anon_sym_readonly] = ACTIONS(304),
    [anon_sym_local] = ACTIONS(304),
    [anon_sym_unset] = ACTIONS(306),
    [anon_sym_unsetenv] = ACTIONS(306),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [238] = {
    [sym__statements] = STATE(6903),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [239] = {
    [sym__statements] = STATE(7529),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [240] = {
    [sym__statements] = STATE(7427),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [241] = {
    [sym__statements] = STATE(6899),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [242] = {
    [sym__statements] = STATE(6896),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [243] = {
    [sym__statements] = STATE(6815),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2219),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [244] = {
    [sym__statements] = STATE(6816),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [245] = {
    [sym__statements] = STATE(6819),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [246] = {
    [sym__statements] = STATE(7527),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [247] = {
    [sym__statements] = STATE(7300),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [248] = {
    [sym__statements] = STATE(6821),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [249] = {
    [sym__statements] = STATE(6861),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2217),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [250] = {
    [sym__statements] = STATE(7526),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [251] = {
    [sym__statements] = STATE(6863),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [252] = {
    [sym__statements] = STATE(7509),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2165),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [253] = {
    [sym__statements] = STATE(6864),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [254] = {
    [sym__statements] = STATE(6867),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [255] = {
    [sym__statements] = STATE(6902),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2133),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [256] = {
    [sym__statements] = STATE(6905),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [257] = {
    [sym__statements] = STATE(6907),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [258] = {
    [sym__statements] = STATE(6909),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [259] = {
    [sym__statements] = STATE(7508),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [260] = {
    [sym__statements] = STATE(7507),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [261] = {
    [sym__statements] = STATE(6924),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2207),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [262] = {
    [sym__statements] = STATE(6929),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [263] = {
    [sym__statements] = STATE(6932),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [264] = {
    [sym__statements] = STATE(7506),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [265] = {
    [sym__statements] = STATE(6938),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [266] = {
    [sym__statements] = STATE(7048),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2206),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [267] = {
    [sym__statements] = STATE(7050),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [268] = {
    [sym__statements] = STATE(7500),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2163),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [269] = {
    [sym__statements] = STATE(7499),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [270] = {
    [sym__statements] = STATE(7498),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [271] = {
    [sym__statements] = STATE(7058),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [272] = {
    [sym__statements] = STATE(7060),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [273] = {
    [sym__statements] = STATE(7497),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [274] = {
    [sym__statements] = STATE(7492),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2159),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [275] = {
    [sym__statements] = STATE(7072),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [276] = {
    [sym__statements] = STATE(7077),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [277] = {
    [sym__statements] = STATE(7078),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2076),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [278] = {
    [sym__statements] = STATE(7194),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2153),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [279] = {
    [sym__statements] = STATE(7079),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [280] = {
    [aux_sym__terminated_statement] = STATE(237),
    [sym__statement_not_pipeline] = STATE(6718),
    [sym_redirected_statement] = STATE(4951),
    [sym_for_statement] = STATE(4951),
    [sym_c_style_for_statement] = STATE(4951),
    [sym_while_statement] = STATE(4567),
    [sym_if_statement] = STATE(4567),
    [sym_case_statement] = STATE(4951),
    [sym_function_definition] = STATE(4951),
    [sym_compound_statement] = STATE(4951),
    [sym_subshell] = STATE(4951),
    [sym_pipeline] = STATE(5163),
    [sym_list] = STATE(4951),
    [sym_negated_command] = STATE(4951),
    [sym_test_command] = STATE(4951),
    [sym_declaration_command] = STATE(4951),
    [sym_unset_command] = STATE(4951),
    [sym_command] = STATE(4951),
    [sym_command_name] = STATE(544),
    [sym_variable_assignment] = STATE(1153),
    [sym_variable_assignments] = STATE(4951),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym_redirected_statement_repeat2] = STATE(4644),
    [aux_sym_command_repeat1] = STATE(997),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(286),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_done] = ACTIONS(881),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(300),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(302),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(304),
    [anon_sym_typeset] = ACTIONS(304),
    [anon_sym_export] = ACTIONS(304),
    [anon_sym_readonly] = ACTIONS(304),
    [anon_sym_local] = ACTIONS(304),
    [anon_sym_unset] = ACTIONS(306),
    [anon_sym_unsetenv] = ACTIONS(306),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [281] = {
    [sym__statements] = STATE(7488),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [282] = {
    [aux_sym__terminated_statement] = STATE(62),
    [sym__statement_not_pipeline] = STATE(6718),
    [sym_redirected_statement] = STATE(4993),
    [sym_for_statement] = STATE(4993),
    [sym_c_style_for_statement] = STATE(4993),
    [sym_while_statement] = STATE(4489),
    [sym_if_statement] = STATE(4489),
    [sym_case_statement] = STATE(4993),
    [sym_function_definition] = STATE(4993),
    [sym_compound_statement] = STATE(4993),
    [sym_subshell] = STATE(4993),
    [sym_pipeline] = STATE(5122),
    [sym_list] = STATE(4993),
    [sym_negated_command] = STATE(4993),
    [sym_test_command] = STATE(4993),
    [sym_declaration_command] = STATE(4993),
    [sym_unset_command] = STATE(4993),
    [sym_command] = STATE(4993),
    [sym_command_name] = STATE(544),
    [sym_variable_assignment] = STATE(1030),
    [sym_variable_assignments] = STATE(4993),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym_redirected_statement_repeat2] = STATE(4644),
    [aux_sym_command_repeat1] = STATE(997),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(286),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(300),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_RBRACE] = ACTIONS(883),
    [anon_sym_BANG] = ACTIONS(302),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(304),
    [anon_sym_typeset] = ACTIONS(304),
    [anon_sym_export] = ACTIONS(304),
    [anon_sym_readonly] = ACTIONS(304),
    [anon_sym_local] = ACTIONS(304),
    [anon_sym_unset] = ACTIONS(306),
    [anon_sym_unsetenv] = ACTIONS(306),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [283] = {
    [sym__statements] = STATE(7485),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [284] = {
    [aux_sym__terminated_statement] = STATE(290),
    [sym__statement_not_pipeline] = STATE(6718),
    [sym_redirected_statement] = STATE(4993),
    [sym_for_statement] = STATE(4993),
    [sym_c_style_for_statement] = STATE(4993),
    [sym_while_statement] = STATE(4489),
    [sym_if_statement] = STATE(4489),
    [sym_case_statement] = STATE(4993),
    [sym_function_definition] = STATE(4993),
    [sym_compound_statement] = STATE(4993),
    [sym_subshell] = STATE(4993),
    [sym_pipeline] = STATE(5122),
    [sym_list] = STATE(4993),
    [sym_negated_command] = STATE(4993),
    [sym_test_command] = STATE(4993),
    [sym_declaration_command] = STATE(4993),
    [sym_unset_command] = STATE(4993),
    [sym_command] = STATE(4993),
    [sym_command_name] = STATE(544),
    [sym_variable_assignment] = STATE(1030),
    [sym_variable_assignments] = STATE(4993),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym_redirected_statement_repeat2] = STATE(4644),
    [aux_sym_command_repeat1] = STATE(997),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(286),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(300),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_RBRACE] = ACTIONS(885),
    [anon_sym_BANG] = ACTIONS(302),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(304),
    [anon_sym_typeset] = ACTIONS(304),
    [anon_sym_export] = ACTIONS(304),
    [anon_sym_readonly] = ACTIONS(304),
    [anon_sym_local] = ACTIONS(304),
    [anon_sym_unset] = ACTIONS(306),
    [anon_sym_unsetenv] = ACTIONS(306),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [285] = {
    [sym__statements] = STATE(7484),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [286] = {
    [sym__statements] = STATE(7084),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2078),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [287] = {
    [sym__statements] = STATE(7475),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2157),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [288] = {
    [sym__statements] = STATE(7085),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [289] = {
    [sym__statements] = STATE(7216),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [290] = {
    [aux_sym__terminated_statement] = STATE(62),
    [sym__statement_not_pipeline] = STATE(6718),
    [sym_redirected_statement] = STATE(4993),
    [sym_for_statement] = STATE(4993),
    [sym_c_style_for_statement] = STATE(4993),
    [sym_while_statement] = STATE(4489),
    [sym_if_statement] = STATE(4489),
    [sym_case_statement] = STATE(4993),
    [sym_function_definition] = STATE(4993),
    [sym_compound_statement] = STATE(4993),
    [sym_subshell] = STATE(4993),
    [sym_pipeline] = STATE(5122),
    [sym_list] = STATE(4993),
    [sym_negated_command] = STATE(4993),
    [sym_test_command] = STATE(4993),
    [sym_declaration_command] = STATE(4993),
    [sym_unset_command] = STATE(4993),
    [sym_command] = STATE(4993),
    [sym_command_name] = STATE(544),
    [sym_variable_assignment] = STATE(1030),
    [sym_variable_assignments] = STATE(4993),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym_redirected_statement_repeat2] = STATE(4644),
    [aux_sym_command_repeat1] = STATE(997),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(286),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(300),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_RBRACE] = ACTIONS(887),
    [anon_sym_BANG] = ACTIONS(302),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(304),
    [anon_sym_typeset] = ACTIONS(304),
    [anon_sym_export] = ACTIONS(304),
    [anon_sym_readonly] = ACTIONS(304),
    [anon_sym_local] = ACTIONS(304),
    [anon_sym_unset] = ACTIONS(306),
    [anon_sym_unsetenv] = ACTIONS(306),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [291] = {
    [sym__statements] = STATE(7086),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [292] = {
    [sym__statements] = STATE(7088),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2079),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [293] = {
    [sym__statements] = STATE(7095),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [294] = {
    [sym__statements] = STATE(7230),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [295] = {
    [sym__statements] = STATE(7474),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [296] = {
    [sym__statements] = STATE(7378),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2196),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [297] = {
    [sym__statements] = STATE(7399),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [298] = {
    [sym__statements] = STATE(7096),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [299] = {
    [sym__statements] = STATE(7097),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2082),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [300] = {
    [sym__statements] = STATE(7107),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [301] = {
    [sym__statements] = STATE(7109),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [302] = {
    [sym__statements] = STATE(7111),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2083),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [303] = {
    [sym__statements] = STATE(7118),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [304] = {
    [sym__statements] = STATE(7119),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [305] = {
    [sym__statements] = STATE(7125),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2084),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [306] = {
    [sym__statements] = STATE(7130),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [307] = {
    [sym__statements] = STATE(7132),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [308] = {
    [sym__statements] = STATE(7133),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2085),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [309] = {
    [sym__statements] = STATE(7138),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [310] = {
    [sym__statements] = STATE(7139),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [311] = {
    [sym__statements] = STATE(7140),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2088),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [312] = {
    [sym__statements] = STATE(7151),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [313] = {
    [sym__statements] = STATE(7152),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [314] = {
    [sym__statements] = STATE(7160),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2096),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [315] = {
    [sym__statements] = STATE(7166),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [316] = {
    [sym__statements] = STATE(7168),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [317] = {
    [sym__statements] = STATE(7169),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2097),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [318] = {
    [sym__statements] = STATE(7173),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [319] = {
    [sym__statements] = STATE(7174),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [320] = {
    [sym__statements] = STATE(7175),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2098),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [321] = {
    [sym__statements] = STATE(7180),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [322] = {
    [sym__statements] = STATE(7182),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [323] = {
    [sym__statements] = STATE(7186),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2099),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [324] = {
    [sym__statements] = STATE(7189),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [325] = {
    [sym__statements] = STATE(7190),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [326] = {
    [sym__statements] = STATE(7195),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2101),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [327] = {
    [sym__statements] = STATE(7201),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [328] = {
    [sym__statements] = STATE(7202),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [329] = {
    [sym__statements] = STATE(7204),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2102),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [330] = {
    [sym__statements] = STATE(7212),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [331] = {
    [sym__statements] = STATE(7214),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [332] = {
    [sym__statements] = STATE(7217),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2104),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [333] = {
    [sym__statements] = STATE(7228),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [334] = {
    [sym__statements] = STATE(7229),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [335] = {
    [sym__statements] = STATE(7231),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2110),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [336] = {
    [sym__statements] = STATE(7400),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [337] = {
    [sym__statements] = STATE(7249),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [338] = {
    [sym__statements] = STATE(7252),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [339] = {
    [sym__statements] = STATE(7405),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [340] = {
    [sym__statements] = STATE(6918),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2195),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [341] = {
    [sym__statements] = STATE(6827),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [342] = {
    [sym__statements] = STATE(6795),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [343] = {
    [sym__statements] = STATE(7253),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [344] = {
    [sym__statements] = STATE(7254),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2112),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [345] = {
    [sym__statements] = STATE(7264),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [346] = {
    [sym__statements] = STATE(7266),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [347] = {
    [sym__statements] = STATE(7267),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [348] = {
    [sym__statements] = STATE(7270),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2128),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [349] = {
    [sym__statements] = STATE(7279),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [350] = {
    [sym__statements] = STATE(6804),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [351] = {
    [sym__statements] = STATE(7522),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2194),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [352] = {
    [sym__statements] = STATE(7280),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [353] = {
    [sym__statements] = STATE(7473),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [354] = {
    [sym__statements] = STATE(7467),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [355] = {
    [sym__statements] = STATE(7462),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [356] = {
    [sym__statements] = STATE(7472),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [357] = {
    [sym__statements] = STATE(7471),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [358] = {
    [sym__statements] = STATE(7282),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [359] = {
    [sym__statements] = STATE(7379),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2193),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [360] = {
    [sym__statements] = STATE(7283),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2141),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [361] = {
    [sym__statements] = STATE(7288),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [362] = {
    [sym__statements] = STATE(7289),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [363] = {
    [sym__statements] = STATE(7227),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [364] = {
    [sym__statements] = STATE(7290),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [365] = {
    [sym__statements] = STATE(7335),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [366] = {
    [sym__statements] = STATE(7333),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [367] = {
    [sym__statements] = STATE(7295),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2136),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [368] = {
    [sym__statements] = STATE(7310),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [369] = {
    [sym__statements] = STATE(7311),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [370] = {
    [sym__statements] = STATE(7314),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [371] = {
    [sym__statements] = STATE(7316),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2140),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [372] = {
    [sym__statements] = STATE(7325),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [373] = {
    [sym__statements] = STATE(7327),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [374] = {
    [sym__statements] = STATE(7306),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2189),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [375] = {
    [sym__statements] = STATE(7183),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [376] = {
    [sym__statements] = STATE(7329),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2212),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [377] = {
    [sym__statements] = STATE(7338),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [378] = {
    [sym__statements] = STATE(7339),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [379] = {
    [sym__statements] = STATE(7340),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [380] = {
    [sym__statements] = STATE(7341),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2142),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [381] = {
    [sym__statements] = STATE(7357),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [382] = {
    [sym__statements] = STATE(7358),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [383] = {
    [sym__statements] = STATE(7359),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [384] = {
    [sym__statements] = STATE(7360),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2143),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [385] = {
    [sym__statements] = STATE(7366),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [386] = {
    [sym__statements] = STATE(7367),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [387] = {
    [sym__statements] = STATE(7368),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [388] = {
    [sym__statements] = STATE(7372),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2144),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [389] = {
    [sym__statements] = STATE(7382),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [390] = {
    [sym__statements] = STATE(7383),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [391] = {
    [sym__statements] = STATE(7384),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [392] = {
    [sym__statements] = STATE(7385),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2146),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [393] = {
    [sym__statements] = STATE(7388),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [394] = {
    [sym__statements] = STATE(7390),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [395] = {
    [sym__statements] = STATE(7391),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [396] = {
    [sym__statements] = STATE(7394),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2148),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [397] = {
    [sym__statements] = STATE(7407),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [398] = {
    [sym__statements] = STATE(7409),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [399] = {
    [sym__statements] = STATE(7410),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [400] = {
    [sym__statements] = STATE(7411),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2149),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [401] = {
    [sym__statements] = STATE(7418),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [402] = {
    [sym__statements] = STATE(7419),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [403] = {
    [sym__statements] = STATE(7420),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [404] = {
    [sym__statements] = STATE(7421),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2150),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [405] = {
    [sym__statements] = STATE(7428),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [406] = {
    [sym__statements] = STATE(7431),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [407] = {
    [sym__statements] = STATE(7434),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [408] = {
    [sym__statements] = STATE(7436),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2151),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [409] = {
    [sym__statements] = STATE(7445),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [410] = {
    [sym__statements] = STATE(7301),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [411] = {
    [sym__statements] = STATE(7446),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [412] = {
    [sym__statements] = STATE(7447),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [413] = {
    [sym__statements] = STATE(7448),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2070),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [414] = {
    [sym__statements] = STATE(7454),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [415] = {
    [sym__statements] = STATE(7455),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [416] = {
    [sym__statements] = STATE(7459),
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4662),
    [sym_for_statement] = STATE(4662),
    [sym_c_style_for_statement] = STATE(4662),
    [sym_while_statement] = STATE(4314),
    [sym_if_statement] = STATE(4314),
    [sym_case_statement] = STATE(4662),
    [sym_function_definition] = STATE(4662),
    [sym_compound_statement] = STATE(4662),
    [sym_subshell] = STATE(4662),
    [sym_pipeline] = STATE(5070),
    [sym_list] = STATE(4662),
    [sym_negated_command] = STATE(4662),
    [sym_test_command] = STATE(4662),
    [sym_declaration_command] = STATE(4662),
    [sym_unset_command] = STATE(4662),
    [sym_command] = STATE(4662),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1150),
    [sym_variable_assignments] = STATE(4662),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(429),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [417] = {
    [sym__statements] = STATE(7463),
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4797),
    [sym_for_statement] = STATE(4797),
    [sym_c_style_for_statement] = STATE(4797),
    [sym_while_statement] = STATE(4400),
    [sym_if_statement] = STATE(4400),
    [sym_case_statement] = STATE(4797),
    [sym_function_definition] = STATE(4797),
    [sym_compound_statement] = STATE(4797),
    [sym_subshell] = STATE(4797),
    [sym_pipeline] = STATE(5054),
    [sym_list] = STATE(4797),
    [sym_negated_command] = STATE(4797),
    [sym_test_command] = STATE(4797),
    [sym_declaration_command] = STATE(4797),
    [sym_unset_command] = STATE(4797),
    [sym_command] = STATE(4797),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(945),
    [sym_variable_assignments] = STATE(4797),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(1930),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(423),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [418] = {
    [aux_sym__terminated_statement] = STATE(234),
    [sym__statement_not_pipeline] = STATE(6718),
    [sym_redirected_statement] = STATE(5007),
    [sym_for_statement] = STATE(5007),
    [sym_c_style_for_statement] = STATE(5007),
    [sym_while_statement] = STATE(4551),
    [sym_if_statement] = STATE(4551),
    [sym_case_statement] = STATE(5007),
    [sym_function_definition] = STATE(5007),
    [sym_compound_statement] = STATE(5007),
    [sym_subshell] = STATE(5007),
    [sym_pipeline] = STATE(5189),
    [sym_list] = STATE(5007),
    [sym_negated_command] = STATE(5007),
    [sym_test_command] = STATE(5007),
    [sym_declaration_command] = STATE(5007),
    [sym_unset_command] = STATE(5007),
    [sym_command] = STATE(5007),
    [sym_command_name] = STATE(544),
    [sym_variable_assignment] = STATE(1151),
    [sym_variable_assignments] = STATE(5007),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym_redirected_statement_repeat2] = STATE(4644),
    [aux_sym_command_repeat1] = STATE(997),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(286),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(300),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(302),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(304),
    [anon_sym_typeset] = ACTIONS(304),
    [anon_sym_export] = ACTIONS(304),
    [anon_sym_readonly] = ACTIONS(304),
    [anon_sym_local] = ACTIONS(304),
    [anon_sym_unset] = ACTIONS(306),
    [anon_sym_unsetenv] = ACTIONS(306),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [419] = {
    [sym__statement_not_pipeline] = STATE(6718),
    [sym_redirected_statement] = STATE(4928),
    [sym_for_statement] = STATE(4928),
    [sym_c_style_for_statement] = STATE(4928),
    [sym_while_statement] = STATE(4415),
    [sym_if_statement] = STATE(4415),
    [sym_case_statement] = STATE(4928),
    [sym_function_definition] = STATE(4928),
    [sym_compound_statement] = STATE(4928),
    [sym_subshell] = STATE(4928),
    [sym_pipeline] = STATE(5248),
    [sym_list] = STATE(4928),
    [sym_negated_command] = STATE(4928),
    [sym_test_command] = STATE(4928),
    [sym_declaration_command] = STATE(4928),
    [sym_unset_command] = STATE(4928),
    [sym_command] = STATE(4928),
    [sym_command_name] = STATE(544),
    [sym_variable_assignment] = STATE(1036),
    [sym_variable_assignments] = STATE(4928),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(419),
    [aux_sym_redirected_statement_repeat2] = STATE(4644),
    [aux_sym_command_repeat1] = STATE(997),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(889),
    [anon_sym_for] = ACTIONS(892),
    [anon_sym_select] = ACTIONS(895),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(898),
    [anon_sym_LT] = ACTIONS(901),
    [anon_sym_GT] = ACTIONS(901),
    [anon_sym_GT_GT] = ACTIONS(904),
    [anon_sym_LPAREN] = ACTIONS(907),
    [anon_sym_while] = ACTIONS(910),
    [anon_sym_until] = ACTIONS(910),
    [anon_sym_if] = ACTIONS(913),
    [anon_sym_case] = ACTIONS(916),
    [anon_sym_function] = ACTIONS(919),
    [anon_sym_LBRACE] = ACTIONS(922),
    [anon_sym_BANG] = ACTIONS(925),
    [anon_sym_LBRACK] = ACTIONS(928),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(931),
    [anon_sym_declare] = ACTIONS(934),
    [anon_sym_typeset] = ACTIONS(934),
    [anon_sym_export] = ACTIONS(934),
    [anon_sym_readonly] = ACTIONS(934),
    [anon_sym_local] = ACTIONS(934),
    [anon_sym_unset] = ACTIONS(937),
    [anon_sym_unsetenv] = ACTIONS(937),
    [anon_sym_AMP_GT] = ACTIONS(901),
    [anon_sym_AMP_GT_GT] = ACTIONS(904),
    [anon_sym_LT_AMP] = ACTIONS(901),
    [anon_sym_GT_AMP] = ACTIONS(901),
    [anon_sym_GT_PIPE] = ACTIONS(904),
    [anon_sym_LT_AMP_DASH] = ACTIONS(940),
    [anon_sym_GT_AMP_DASH] = ACTIONS(940),
    [anon_sym_LT_LT_LT] = ACTIONS(943),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(946),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(949),
    [anon_sym_DOLLAR] = ACTIONS(952),
    [sym__special_character] = ACTIONS(955),
    [anon_sym_DQUOTE] = ACTIONS(958),
    [sym_raw_string] = ACTIONS(961),
    [sym_ansi_c_string] = ACTIONS(961),
    [aux_sym_number_token1] = ACTIONS(964),
    [aux_sym_number_token2] = ACTIONS(967),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(970),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(973),
    [anon_sym_BQUOTE] = ACTIONS(976),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(979),
    [anon_sym_LT_LPAREN] = ACTIONS(982),
    [anon_sym_GT_LPAREN] = ACTIONS(982),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(985),
    [sym_variable_name] = ACTIONS(988),
    [sym_test_operator] = ACTIONS(991),
    [sym__brace_start] = ACTIONS(994),
  },
  [420] = {
    [aux_sym__terminated_statement] = STATE(59),
    [sym__statement_not_pipeline] = STATE(6718),
    [sym_redirected_statement] = STATE(5007),
    [sym_for_statement] = STATE(5007),
    [sym_c_style_for_statement] = STATE(5007),
    [sym_while_statement] = STATE(4551),
    [sym_if_statement] = STATE(4551),
    [sym_case_statement] = STATE(5007),
    [sym_function_definition] = STATE(5007),
    [sym_compound_statement] = STATE(5007),
    [sym_subshell] = STATE(5007),
    [sym_pipeline] = STATE(5189),
    [sym_list] = STATE(5007),
    [sym_negated_command] = STATE(5007),
    [sym_test_command] = STATE(5007),
    [sym_declaration_command] = STATE(5007),
    [sym_unset_command] = STATE(5007),
    [sym_command] = STATE(5007),
    [sym_command_name] = STATE(544),
    [sym_variable_assignment] = STATE(1151),
    [sym_variable_assignments] = STATE(5007),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym_redirected_statement_repeat2] = STATE(4644),
    [aux_sym_command_repeat1] = STATE(997),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(286),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(300),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(302),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(304),
    [anon_sym_typeset] = ACTIONS(304),
    [anon_sym_export] = ACTIONS(304),
    [anon_sym_readonly] = ACTIONS(304),
    [anon_sym_local] = ACTIONS(304),
    [anon_sym_unset] = ACTIONS(306),
    [anon_sym_unsetenv] = ACTIONS(306),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [421] = {
    [aux_sym__terminated_statement] = STATE(153),
    [sym__statement_not_pipeline] = STATE(6718),
    [sym_redirected_statement] = STATE(5007),
    [sym_for_statement] = STATE(5007),
    [sym_c_style_for_statement] = STATE(5007),
    [sym_while_statement] = STATE(4551),
    [sym_if_statement] = STATE(4551),
    [sym_case_statement] = STATE(5007),
    [sym_function_definition] = STATE(5007),
    [sym_compound_statement] = STATE(5007),
    [sym_subshell] = STATE(5007),
    [sym_pipeline] = STATE(5189),
    [sym_list] = STATE(5007),
    [sym_negated_command] = STATE(5007),
    [sym_test_command] = STATE(5007),
    [sym_declaration_command] = STATE(5007),
    [sym_unset_command] = STATE(5007),
    [sym_command] = STATE(5007),
    [sym_command_name] = STATE(544),
    [sym_variable_assignment] = STATE(1151),
    [sym_variable_assignments] = STATE(5007),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym_redirected_statement_repeat2] = STATE(4644),
    [aux_sym_command_repeat1] = STATE(997),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(286),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(300),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(302),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(304),
    [anon_sym_typeset] = ACTIONS(304),
    [anon_sym_export] = ACTIONS(304),
    [anon_sym_readonly] = ACTIONS(304),
    [anon_sym_local] = ACTIONS(304),
    [anon_sym_unset] = ACTIONS(306),
    [anon_sym_unsetenv] = ACTIONS(306),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [422] = {
    [aux_sym__terminated_statement] = STATE(133),
    [sym__statement_not_pipeline] = STATE(6718),
    [sym_redirected_statement] = STATE(5007),
    [sym_for_statement] = STATE(5007),
    [sym_c_style_for_statement] = STATE(5007),
    [sym_while_statement] = STATE(4551),
    [sym_if_statement] = STATE(4551),
    [sym_case_statement] = STATE(5007),
    [sym_function_definition] = STATE(5007),
    [sym_compound_statement] = STATE(5007),
    [sym_subshell] = STATE(5007),
    [sym_pipeline] = STATE(5189),
    [sym_list] = STATE(5007),
    [sym_negated_command] = STATE(5007),
    [sym_test_command] = STATE(5007),
    [sym_declaration_command] = STATE(5007),
    [sym_unset_command] = STATE(5007),
    [sym_command] = STATE(5007),
    [sym_command_name] = STATE(544),
    [sym_variable_assignment] = STATE(1151),
    [sym_variable_assignments] = STATE(5007),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym_redirected_statement_repeat2] = STATE(4644),
    [aux_sym_command_repeat1] = STATE(997),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(286),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(300),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(302),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(304),
    [anon_sym_typeset] = ACTIONS(304),
    [anon_sym_export] = ACTIONS(304),
    [anon_sym_readonly] = ACTIONS(304),
    [anon_sym_local] = ACTIONS(304),
    [anon_sym_unset] = ACTIONS(306),
    [anon_sym_unsetenv] = ACTIONS(306),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [423] = {
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4778),
    [sym_for_statement] = STATE(4778),
    [sym_c_style_for_statement] = STATE(4778),
    [sym_while_statement] = STATE(4396),
    [sym_if_statement] = STATE(4396),
    [sym_case_statement] = STATE(4778),
    [sym_function_definition] = STATE(4778),
    [sym_compound_statement] = STATE(4778),
    [sym_subshell] = STATE(4778),
    [sym_pipeline] = STATE(5051),
    [sym_list] = STATE(4778),
    [sym_negated_command] = STATE(4778),
    [sym_test_command] = STATE(4778),
    [sym_declaration_command] = STATE(4778),
    [sym_unset_command] = STATE(4778),
    [sym_command] = STATE(4778),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(1002),
    [sym_variable_assignments] = STATE(4778),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym__statements_repeat1] = STATE(419),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [424] = {
    [aux_sym__terminated_statement] = STATE(50),
    [sym__statement_not_pipeline] = STATE(6718),
    [sym_redirected_statement] = STATE(5007),
    [sym_for_statement] = STATE(5007),
    [sym_c_style_for_statement] = STATE(5007),
    [sym_while_statement] = STATE(4551),
    [sym_if_statement] = STATE(4551),
    [sym_case_statement] = STATE(5007),
    [sym_function_definition] = STATE(5007),
    [sym_compound_statement] = STATE(5007),
    [sym_subshell] = STATE(5007),
    [sym_pipeline] = STATE(5189),
    [sym_list] = STATE(5007),
    [sym_negated_command] = STATE(5007),
    [sym_test_command] = STATE(5007),
    [sym_declaration_command] = STATE(5007),
    [sym_unset_command] = STATE(5007),
    [sym_command] = STATE(5007),
    [sym_command_name] = STATE(544),
    [sym_variable_assignment] = STATE(1151),
    [sym_variable_assignments] = STATE(5007),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym_redirected_statement_repeat2] = STATE(4644),
    [aux_sym_command_repeat1] = STATE(997),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(286),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(300),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(302),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(304),
    [anon_sym_typeset] = ACTIONS(304),
    [anon_sym_export] = ACTIONS(304),
    [anon_sym_readonly] = ACTIONS(304),
    [anon_sym_local] = ACTIONS(304),
    [anon_sym_unset] = ACTIONS(306),
    [anon_sym_unsetenv] = ACTIONS(306),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [425] = {
    [aux_sym__terminated_statement] = STATE(122),
    [sym__statement_not_pipeline] = STATE(6718),
    [sym_redirected_statement] = STATE(5007),
    [sym_for_statement] = STATE(5007),
    [sym_c_style_for_statement] = STATE(5007),
    [sym_while_statement] = STATE(4551),
    [sym_if_statement] = STATE(4551),
    [sym_case_statement] = STATE(5007),
    [sym_function_definition] = STATE(5007),
    [sym_compound_statement] = STATE(5007),
    [sym_subshell] = STATE(5007),
    [sym_pipeline] = STATE(5189),
    [sym_list] = STATE(5007),
    [sym_negated_command] = STATE(5007),
    [sym_test_command] = STATE(5007),
    [sym_declaration_command] = STATE(5007),
    [sym_unset_command] = STATE(5007),
    [sym_command] = STATE(5007),
    [sym_command_name] = STATE(544),
    [sym_variable_assignment] = STATE(1151),
    [sym_variable_assignments] = STATE(5007),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym_redirected_statement_repeat2] = STATE(4644),
    [aux_sym_command_repeat1] = STATE(997),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(286),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(300),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(302),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(304),
    [anon_sym_typeset] = ACTIONS(304),
    [anon_sym_export] = ACTIONS(304),
    [anon_sym_readonly] = ACTIONS(304),
    [anon_sym_local] = ACTIONS(304),
    [anon_sym_unset] = ACTIONS(306),
    [anon_sym_unsetenv] = ACTIONS(306),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [426] = {
    [sym__statement_not_pipeline] = STATE(6674),
    [sym_redirected_statement] = STATE(4433),
    [sym_for_statement] = STATE(4433),
    [sym_c_style_for_statement] = STATE(4433),
    [sym_while_statement] = STATE(4249),
    [sym_if_statement] = STATE(4249),
    [sym_case_statement] = STATE(4433),
    [sym_function_definition] = STATE(4433),
    [sym_compound_statement] = STATE(4433),
    [sym_subshell] = STATE(4433),
    [sym_pipeline] = STATE(4939),
    [sym_list] = STATE(4433),
    [sym_negated_command] = STATE(4433),
    [sym_test_command] = STATE(4433),
    [sym_declaration_command] = STATE(4433),
    [sym_unset_command] = STATE(4433),
    [sym_command] = STATE(4433),
    [sym_command_name] = STATE(530),
    [sym_variable_assignment] = STATE(835),
    [sym_variable_assignments] = STATE(4433),
    [sym_subscript] = STATE(6760),
    [sym_file_redirect] = STATE(1893),
    [sym_herestring_redirect] = STATE(1892),
    [sym_arithmetic_expansion] = STATE(852),
    [sym_brace_expression] = STATE(852),
    [sym_concatenation] = STATE(1169),
    [sym_string] = STATE(852),
    [sym_translated_string] = STATE(852),
    [sym_number] = STATE(852),
    [sym_simple_expansion] = STATE(852),
    [sym_expansion] = STATE(852),
    [sym_command_substitution] = STATE(852),
    [sym_process_substitution] = STATE(852),
    [aux_sym__statements_repeat1] = STATE(419),
    [aux_sym_redirected_statement_repeat2] = STATE(4371),
    [aux_sym_command_repeat1] = STATE(989),
    [aux_sym__literal_repeat1] = STATE(1026),
    [sym_word] = ACTIONS(665),
    [anon_sym_for] = ACTIONS(667),
    [anon_sym_select] = ACTIONS(669),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(671),
    [anon_sym_LT] = ACTIONS(673),
    [anon_sym_GT] = ACTIONS(673),
    [anon_sym_GT_GT] = ACTIONS(675),
    [anon_sym_LPAREN] = ACTIONS(677),
    [anon_sym_while] = ACTIONS(679),
    [anon_sym_until] = ACTIONS(679),
    [anon_sym_if] = ACTIONS(681),
    [anon_sym_case] = ACTIONS(683),
    [anon_sym_function] = ACTIONS(687),
    [anon_sym_LBRACE] = ACTIONS(689),
    [anon_sym_BANG] = ACTIONS(691),
    [anon_sym_LBRACK] = ACTIONS(693),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(695),
    [anon_sym_declare] = ACTIONS(697),
    [anon_sym_typeset] = ACTIONS(697),
    [anon_sym_export] = ACTIONS(697),
    [anon_sym_readonly] = ACTIONS(697),
    [anon_sym_local] = ACTIONS(697),
    [anon_sym_unset] = ACTIONS(699),
    [anon_sym_unsetenv] = ACTIONS(699),
    [anon_sym_AMP_GT] = ACTIONS(673),
    [anon_sym_AMP_GT_GT] = ACTIONS(675),
    [anon_sym_LT_AMP] = ACTIONS(673),
    [anon_sym_GT_AMP] = ACTIONS(673),
    [anon_sym_GT_PIPE] = ACTIONS(675),
    [anon_sym_LT_AMP_DASH] = ACTIONS(701),
    [anon_sym_GT_AMP_DASH] = ACTIONS(701),
    [anon_sym_LT_LT_LT] = ACTIONS(703),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(705),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(707),
    [anon_sym_DOLLAR] = ACTIONS(709),
    [sym__special_character] = ACTIONS(711),
    [anon_sym_DQUOTE] = ACTIONS(713),
    [sym_raw_string] = ACTIONS(715),
    [sym_ansi_c_string] = ACTIONS(715),
    [aux_sym_number_token1] = ACTIONS(717),
    [aux_sym_number_token2] = ACTIONS(719),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(721),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(723),
    [anon_sym_BQUOTE] = ACTIONS(725),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(727),
    [anon_sym_LT_LPAREN] = ACTIONS(729),
    [anon_sym_GT_LPAREN] = ACTIONS(729),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(731),
    [sym_variable_name] = ACTIONS(733),
    [sym_test_operator] = ACTIONS(735),
    [sym__brace_start] = ACTIONS(737),
  },
  [427] = {
    [sym__statement_not_pipeline] = STATE(6572),
    [sym_redirected_statement] = STATE(4663),
    [sym_for_statement] = STATE(4663),
    [sym_c_style_for_statement] = STATE(4663),
    [sym_while_statement] = STATE(4312),
    [sym_if_statement] = STATE(4312),
    [sym_case_statement] = STATE(4663),
    [sym_function_definition] = STATE(4663),
    [sym_compound_statement] = STATE(4663),
    [sym_subshell] = STATE(4663),
    [sym_pipeline] = STATE(5108),
    [sym_list] = STATE(4663),
    [sym_negated_command] = STATE(4663),
    [sym_test_command] = STATE(4663),
    [sym_declaration_command] = STATE(4663),
    [sym_unset_command] = STATE(4663),
    [sym_command] = STATE(4663),
    [sym_command_name] = STATE(534),
    [sym_variable_assignment] = STATE(991),
    [sym_variable_assignments] = STATE(4663),
    [sym_subscript] = STATE(6774),
    [sym_file_redirect] = STATE(2055),
    [sym_herestring_redirect] = STATE(2156),
    [sym_arithmetic_expansion] = STATE(957),
    [sym_brace_expression] = STATE(957),
    [sym_concatenation] = STATE(1386),
    [sym_string] = STATE(957),
    [sym_translated_string] = STATE(957),
    [sym_number] = STATE(957),
    [sym_simple_expansion] = STATE(957),
    [sym_expansion] = STATE(957),
    [sym_command_substitution] = STATE(957),
    [sym_process_substitution] = STATE(957),
    [aux_sym__statements_repeat1] = STATE(419),
    [aux_sym_redirected_statement_repeat2] = STATE(4571),
    [aux_sym_command_repeat1] = STATE(937),
    [aux_sym__literal_repeat1] = STATE(1212),
    [sym_word] = ACTIONS(7),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(13),
    [anon_sym_LT] = ACTIONS(15),
    [anon_sym_GT] = ACTIONS(15),
    [anon_sym_GT_GT] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(31),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(37),
    [anon_sym_typeset] = ACTIONS(37),
    [anon_sym_export] = ACTIONS(37),
    [anon_sym_readonly] = ACTIONS(37),
    [anon_sym_local] = ACTIONS(37),
    [anon_sym_unset] = ACTIONS(39),
    [anon_sym_unsetenv] = ACTIONS(39),
    [anon_sym_AMP_GT] = ACTIONS(15),
    [anon_sym_AMP_GT_GT] = ACTIONS(17),
    [anon_sym_LT_AMP] = ACTIONS(15),
    [anon_sym_GT_AMP] = ACTIONS(15),
    [anon_sym_GT_PIPE] = ACTIONS(17),
    [anon_sym_LT_AMP_DASH] = ACTIONS(41),
    [anon_sym_GT_AMP_DASH] = ACTIONS(41),
    [anon_sym_LT_LT_LT] = ACTIONS(43),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(45),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(47),
    [anon_sym_DOLLAR] = ACTIONS(49),
    [sym__special_character] = ACTIONS(51),
    [anon_sym_DQUOTE] = ACTIONS(53),
    [sym_raw_string] = ACTIONS(55),
    [sym_ansi_c_string] = ACTIONS(55),
    [aux_sym_number_token1] = ACTIONS(57),
    [aux_sym_number_token2] = ACTIONS(59),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(61),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(63),
    [anon_sym_BQUOTE] = ACTIONS(65),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(67),
    [anon_sym_LT_LPAREN] = ACTIONS(69),
    [anon_sym_GT_LPAREN] = ACTIONS(69),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(73),
    [sym_variable_name] = ACTIONS(75),
    [sym_test_operator] = ACTIONS(77),
    [sym__brace_start] = ACTIONS(79),
  },
  [428] = {
    [aux_sym__terminated_statement] = STATE(102),
    [sym__statement_not_pipeline] = STATE(6718),
    [sym_redirected_statement] = STATE(5007),
    [sym_for_statement] = STATE(5007),
    [sym_c_style_for_statement] = STATE(5007),
    [sym_while_statement] = STATE(4551),
    [sym_if_statement] = STATE(4551),
    [sym_case_statement] = STATE(5007),
    [sym_function_definition] = STATE(5007),
    [sym_compound_statement] = STATE(5007),
    [sym_subshell] = STATE(5007),
    [sym_pipeline] = STATE(5189),
    [sym_list] = STATE(5007),
    [sym_negated_command] = STATE(5007),
    [sym_test_command] = STATE(5007),
    [sym_declaration_command] = STATE(5007),
    [sym_unset_command] = STATE(5007),
    [sym_command] = STATE(5007),
    [sym_command_name] = STATE(544),
    [sym_variable_assignment] = STATE(1151),
    [sym_variable_assignments] = STATE(5007),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym_redirected_statement_repeat2] = STATE(4644),
    [aux_sym_command_repeat1] = STATE(997),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(286),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(300),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(302),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(304),
    [anon_sym_typeset] = ACTIONS(304),
    [anon_sym_export] = ACTIONS(304),
    [anon_sym_readonly] = ACTIONS(304),
    [anon_sym_local] = ACTIONS(304),
    [anon_sym_unset] = ACTIONS(306),
    [anon_sym_unsetenv] = ACTIONS(306),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [429] = {
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4793),
    [sym_for_statement] = STATE(4793),
    [sym_c_style_for_statement] = STATE(4793),
    [sym_while_statement] = STATE(4397),
    [sym_if_statement] = STATE(4397),
    [sym_case_statement] = STATE(4793),
    [sym_function_definition] = STATE(4793),
    [sym_compound_statement] = STATE(4793),
    [sym_subshell] = STATE(4793),
    [sym_pipeline] = STATE(5040),
    [sym_list] = STATE(4793),
    [sym_negated_command] = STATE(4793),
    [sym_test_command] = STATE(4793),
    [sym_declaration_command] = STATE(4793),
    [sym_unset_command] = STATE(4793),
    [sym_command] = STATE(4793),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1132),
    [sym_variable_assignments] = STATE(4793),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym__statements_repeat1] = STATE(419),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [430] = {
    [aux_sym__terminated_statement] = STATE(51),
    [sym__statement_not_pipeline] = STATE(6718),
    [sym_redirected_statement] = STATE(5007),
    [sym_for_statement] = STATE(5007),
    [sym_c_style_for_statement] = STATE(5007),
    [sym_while_statement] = STATE(4551),
    [sym_if_statement] = STATE(4551),
    [sym_case_statement] = STATE(5007),
    [sym_function_definition] = STATE(5007),
    [sym_compound_statement] = STATE(5007),
    [sym_subshell] = STATE(5007),
    [sym_pipeline] = STATE(5189),
    [sym_list] = STATE(5007),
    [sym_negated_command] = STATE(5007),
    [sym_test_command] = STATE(5007),
    [sym_declaration_command] = STATE(5007),
    [sym_unset_command] = STATE(5007),
    [sym_command] = STATE(5007),
    [sym_command_name] = STATE(544),
    [sym_variable_assignment] = STATE(1151),
    [sym_variable_assignments] = STATE(5007),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym_redirected_statement_repeat2] = STATE(4644),
    [aux_sym_command_repeat1] = STATE(997),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(286),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(300),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(302),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(304),
    [anon_sym_typeset] = ACTIONS(304),
    [anon_sym_export] = ACTIONS(304),
    [anon_sym_readonly] = ACTIONS(304),
    [anon_sym_local] = ACTIONS(304),
    [anon_sym_unset] = ACTIONS(306),
    [anon_sym_unsetenv] = ACTIONS(306),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [431] = {
    [sym__statement_not_pipeline] = STATE(6584),
    [sym_redirected_statement] = STATE(4341),
    [sym_for_statement] = STATE(4341),
    [sym_c_style_for_statement] = STATE(4341),
    [sym_while_statement] = STATE(4223),
    [sym_if_statement] = STATE(4223),
    [sym_case_statement] = STATE(4341),
    [sym_function_definition] = STATE(4341),
    [sym_compound_statement] = STATE(4341),
    [sym_subshell] = STATE(4341),
    [sym_pipeline] = STATE(4792),
    [sym_list] = STATE(4341),
    [sym_negated_command] = STATE(4341),
    [sym_test_command] = STATE(4341),
    [sym_declaration_command] = STATE(4341),
    [sym_unset_command] = STATE(4341),
    [sym_command] = STATE(4341),
    [sym_command_name] = STATE(528),
    [sym_variable_assignment] = STATE(818),
    [sym_variable_assignments] = STATE(4341),
    [sym_subscript] = STATE(6736),
    [sym_file_redirect] = STATE(1635),
    [sym_herestring_redirect] = STATE(1634),
    [sym_arithmetic_expansion] = STATE(815),
    [sym_brace_expression] = STATE(815),
    [sym_concatenation] = STATE(1082),
    [sym_string] = STATE(815),
    [sym_translated_string] = STATE(815),
    [sym_number] = STATE(815),
    [sym_simple_expansion] = STATE(815),
    [sym_expansion] = STATE(815),
    [sym_command_substitution] = STATE(815),
    [sym_process_substitution] = STATE(815),
    [aux_sym__statements_repeat1] = STATE(419),
    [aux_sym_redirected_statement_repeat2] = STATE(4250),
    [aux_sym_command_repeat1] = STATE(925),
    [aux_sym__literal_repeat1] = STATE(926),
    [sym_word] = ACTIONS(403),
    [anon_sym_for] = ACTIONS(405),
    [anon_sym_select] = ACTIONS(407),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(409),
    [anon_sym_LT] = ACTIONS(411),
    [anon_sym_GT] = ACTIONS(411),
    [anon_sym_GT_GT] = ACTIONS(413),
    [anon_sym_LPAREN] = ACTIONS(415),
    [anon_sym_while] = ACTIONS(417),
    [anon_sym_until] = ACTIONS(417),
    [anon_sym_if] = ACTIONS(419),
    [anon_sym_case] = ACTIONS(421),
    [anon_sym_function] = ACTIONS(431),
    [anon_sym_LBRACE] = ACTIONS(433),
    [anon_sym_BANG] = ACTIONS(435),
    [anon_sym_LBRACK] = ACTIONS(437),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(439),
    [anon_sym_declare] = ACTIONS(441),
    [anon_sym_typeset] = ACTIONS(441),
    [anon_sym_export] = ACTIONS(441),
    [anon_sym_readonly] = ACTIONS(441),
    [anon_sym_local] = ACTIONS(441),
    [anon_sym_unset] = ACTIONS(443),
    [anon_sym_unsetenv] = ACTIONS(443),
    [anon_sym_AMP_GT] = ACTIONS(411),
    [anon_sym_AMP_GT_GT] = ACTIONS(413),
    [anon_sym_LT_AMP] = ACTIONS(411),
    [anon_sym_GT_AMP] = ACTIONS(411),
    [anon_sym_GT_PIPE] = ACTIONS(413),
    [anon_sym_LT_AMP_DASH] = ACTIONS(445),
    [anon_sym_GT_AMP_DASH] = ACTIONS(445),
    [anon_sym_LT_LT_LT] = ACTIONS(447),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(449),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(451),
    [anon_sym_DOLLAR] = ACTIONS(453),
    [sym__special_character] = ACTIONS(455),
    [anon_sym_DQUOTE] = ACTIONS(457),
    [sym_raw_string] = ACTIONS(459),
    [sym_ansi_c_string] = ACTIONS(459),
    [aux_sym_number_token1] = ACTIONS(461),
    [aux_sym_number_token2] = ACTIONS(463),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(465),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(467),
    [anon_sym_BQUOTE] = ACTIONS(469),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(471),
    [anon_sym_LT_LPAREN] = ACTIONS(473),
    [anon_sym_GT_LPAREN] = ACTIONS(473),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(475),
    [sym_variable_name] = ACTIONS(477),
    [sym_test_operator] = ACTIONS(479),
    [sym__brace_start] = ACTIONS(481),
  },
  [432] = {
    [aux_sym__terminated_statement] = STATE(54),
    [sym__statement_not_pipeline] = STATE(6718),
    [sym_redirected_statement] = STATE(5007),
    [sym_for_statement] = STATE(5007),
    [sym_c_style_for_statement] = STATE(5007),
    [sym_while_statement] = STATE(4551),
    [sym_if_statement] = STATE(4551),
    [sym_case_statement] = STATE(5007),
    [sym_function_definition] = STATE(5007),
    [sym_compound_statement] = STATE(5007),
    [sym_subshell] = STATE(5007),
    [sym_pipeline] = STATE(5189),
    [sym_list] = STATE(5007),
    [sym_negated_command] = STATE(5007),
    [sym_test_command] = STATE(5007),
    [sym_declaration_command] = STATE(5007),
    [sym_unset_command] = STATE(5007),
    [sym_command] = STATE(5007),
    [sym_command_name] = STATE(544),
    [sym_variable_assignment] = STATE(1151),
    [sym_variable_assignments] = STATE(5007),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym_redirected_statement_repeat2] = STATE(4644),
    [aux_sym_command_repeat1] = STATE(997),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(286),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(300),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(302),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(304),
    [anon_sym_typeset] = ACTIONS(304),
    [anon_sym_export] = ACTIONS(304),
    [anon_sym_readonly] = ACTIONS(304),
    [anon_sym_local] = ACTIONS(304),
    [anon_sym_unset] = ACTIONS(306),
    [anon_sym_unsetenv] = ACTIONS(306),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [433] = {
    [sym__statement_not_pipeline] = STATE(6572),
    [sym_redirected_statement] = STATE(4745),
    [sym_for_statement] = STATE(4745),
    [sym_c_style_for_statement] = STATE(4745),
    [sym_while_statement] = STATE(4281),
    [sym_if_statement] = STATE(4281),
    [sym_case_statement] = STATE(4745),
    [sym_function_definition] = STATE(4745),
    [sym_compound_statement] = STATE(4745),
    [sym_subshell] = STATE(4745),
    [sym_pipeline] = STATE(4608),
    [sym_list] = STATE(4745),
    [sym_negated_command] = STATE(4745),
    [sym_test_command] = STATE(4745),
    [sym_declaration_command] = STATE(4745),
    [sym_unset_command] = STATE(4745),
    [sym_command] = STATE(4745),
    [sym_command_name] = STATE(534),
    [sym_variable_assignment] = STATE(1006),
    [sym_variable_assignments] = STATE(4745),
    [sym_subscript] = STATE(6774),
    [sym_file_redirect] = STATE(2055),
    [sym_herestring_redirect] = STATE(2156),
    [sym_arithmetic_expansion] = STATE(957),
    [sym_brace_expression] = STATE(957),
    [sym_concatenation] = STATE(1386),
    [sym_string] = STATE(957),
    [sym_translated_string] = STATE(957),
    [sym_number] = STATE(957),
    [sym_simple_expansion] = STATE(957),
    [sym_expansion] = STATE(957),
    [sym_command_substitution] = STATE(957),
    [sym_process_substitution] = STATE(957),
    [aux_sym_redirected_statement_repeat2] = STATE(4571),
    [aux_sym_command_repeat1] = STATE(937),
    [aux_sym__literal_repeat1] = STATE(1212),
    [sym_word] = ACTIONS(7),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(13),
    [anon_sym_LT] = ACTIONS(15),
    [anon_sym_GT] = ACTIONS(15),
    [anon_sym_GT_GT] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(31),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(37),
    [anon_sym_typeset] = ACTIONS(37),
    [anon_sym_export] = ACTIONS(37),
    [anon_sym_readonly] = ACTIONS(37),
    [anon_sym_local] = ACTIONS(37),
    [anon_sym_unset] = ACTIONS(39),
    [anon_sym_unsetenv] = ACTIONS(39),
    [anon_sym_AMP_GT] = ACTIONS(15),
    [anon_sym_AMP_GT_GT] = ACTIONS(17),
    [anon_sym_LT_AMP] = ACTIONS(15),
    [anon_sym_GT_AMP] = ACTIONS(15),
    [anon_sym_GT_PIPE] = ACTIONS(17),
    [anon_sym_LT_AMP_DASH] = ACTIONS(41),
    [anon_sym_GT_AMP_DASH] = ACTIONS(41),
    [anon_sym_LT_LT_LT] = ACTIONS(43),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(45),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(47),
    [anon_sym_DOLLAR] = ACTIONS(49),
    [sym__special_character] = ACTIONS(51),
    [anon_sym_DQUOTE] = ACTIONS(53),
    [sym_raw_string] = ACTIONS(55),
    [sym_ansi_c_string] = ACTIONS(55),
    [aux_sym_number_token1] = ACTIONS(57),
    [aux_sym_number_token2] = ACTIONS(59),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(61),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(63),
    [anon_sym_BQUOTE] = ACTIONS(65),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(67),
    [anon_sym_LT_LPAREN] = ACTIONS(69),
    [anon_sym_GT_LPAREN] = ACTIONS(69),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(73),
    [sym_variable_name] = ACTIONS(75),
    [sym_test_operator] = ACTIONS(77),
    [sym__brace_start] = ACTIONS(79),
  },
  [434] = {
    [sym__statement_not_pipeline] = STATE(6566),
    [sym_redirected_statement] = STATE(5381),
    [sym_for_statement] = STATE(5381),
    [sym_c_style_for_statement] = STATE(5381),
    [sym_while_statement] = STATE(5118),
    [sym_if_statement] = STATE(5118),
    [sym_case_statement] = STATE(5381),
    [sym_function_definition] = STATE(5381),
    [sym_compound_statement] = STATE(5381),
    [sym_subshell] = STATE(5381),
    [sym_pipeline] = STATE(5533),
    [sym_list] = STATE(5381),
    [sym_negated_command] = STATE(5381),
    [sym_test_command] = STATE(5381),
    [sym_declaration_command] = STATE(5381),
    [sym_unset_command] = STATE(5381),
    [sym_command] = STATE(5381),
    [sym_command_name] = STATE(596),
    [sym_variable_assignment] = STATE(1706),
    [sym_variable_assignments] = STATE(5381),
    [sym_subscript] = STATE(6759),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(1717),
    [sym_brace_expression] = STATE(1717),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1717),
    [sym_translated_string] = STATE(1717),
    [sym_number] = STATE(1717),
    [sym_simple_expansion] = STATE(1717),
    [sym_expansion] = STATE(1717),
    [sym_command_substitution] = STATE(1717),
    [sym_process_substitution] = STATE(1717),
    [aux_sym_redirected_statement_repeat2] = STATE(5176),
    [aux_sym_command_repeat1] = STATE(933),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(997),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(999),
    [anon_sym_GT] = ACTIONS(999),
    [anon_sym_GT_GT] = ACTIONS(1001),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(1003),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(1005),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(1007),
    [anon_sym_typeset] = ACTIONS(1007),
    [anon_sym_export] = ACTIONS(1007),
    [anon_sym_readonly] = ACTIONS(1007),
    [anon_sym_local] = ACTIONS(1007),
    [anon_sym_unset] = ACTIONS(1009),
    [anon_sym_unsetenv] = ACTIONS(1009),
    [anon_sym_AMP_GT] = ACTIONS(999),
    [anon_sym_AMP_GT_GT] = ACTIONS(1001),
    [anon_sym_LT_AMP] = ACTIONS(999),
    [anon_sym_GT_AMP] = ACTIONS(999),
    [anon_sym_GT_PIPE] = ACTIONS(1001),
    [anon_sym_LT_AMP_DASH] = ACTIONS(1011),
    [anon_sym_GT_AMP_DASH] = ACTIONS(1011),
    [anon_sym_LT_LT_LT] = ACTIONS(1013),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(1015),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(1017),
    [sym_ansi_c_string] = ACTIONS(1017),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(1019),
    [sym_variable_name] = ACTIONS(1021),
    [sym_test_operator] = ACTIONS(1023),
    [sym__brace_start] = ACTIONS(344),
  },
  [435] = {
    [sym__statement_not_pipeline] = STATE(4942),
    [sym_redirected_statement] = STATE(4942),
    [sym_for_statement] = STATE(4942),
    [sym_c_style_for_statement] = STATE(4942),
    [sym_while_statement] = STATE(4234),
    [sym_if_statement] = STATE(4234),
    [sym_case_statement] = STATE(4942),
    [sym_function_definition] = STATE(4942),
    [sym_compound_statement] = STATE(4942),
    [sym_subshell] = STATE(4942),
    [sym_pipeline] = STATE(5593),
    [sym_list] = STATE(4942),
    [sym_negated_command] = STATE(4942),
    [sym_test_command] = STATE(4942),
    [sym_declaration_command] = STATE(4942),
    [sym_unset_command] = STATE(4942),
    [sym_command] = STATE(4942),
    [sym_command_name] = STATE(530),
    [sym_variable_assignment] = STATE(847),
    [sym_variable_assignments] = STATE(4942),
    [sym_subscript] = STATE(6760),
    [sym_file_redirect] = STATE(1893),
    [sym_herestring_redirect] = STATE(1892),
    [sym_arithmetic_expansion] = STATE(852),
    [sym_brace_expression] = STATE(852),
    [sym_concatenation] = STATE(1169),
    [sym_string] = STATE(852),
    [sym_translated_string] = STATE(852),
    [sym_number] = STATE(852),
    [sym_simple_expansion] = STATE(852),
    [sym_expansion] = STATE(852),
    [sym_command_substitution] = STATE(852),
    [sym_process_substitution] = STATE(852),
    [aux_sym_redirected_statement_repeat2] = STATE(4371),
    [aux_sym_command_repeat1] = STATE(989),
    [aux_sym__literal_repeat1] = STATE(1026),
    [sym_word] = ACTIONS(665),
    [anon_sym_for] = ACTIONS(667),
    [anon_sym_select] = ACTIONS(669),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(671),
    [anon_sym_LT] = ACTIONS(673),
    [anon_sym_GT] = ACTIONS(673),
    [anon_sym_GT_GT] = ACTIONS(675),
    [anon_sym_LPAREN] = ACTIONS(677),
    [anon_sym_while] = ACTIONS(679),
    [anon_sym_until] = ACTIONS(679),
    [anon_sym_if] = ACTIONS(681),
    [anon_sym_case] = ACTIONS(683),
    [anon_sym_function] = ACTIONS(687),
    [anon_sym_LBRACE] = ACTIONS(689),
    [anon_sym_BANG] = ACTIONS(691),
    [anon_sym_LBRACK] = ACTIONS(693),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(695),
    [anon_sym_declare] = ACTIONS(697),
    [anon_sym_typeset] = ACTIONS(697),
    [anon_sym_export] = ACTIONS(697),
    [anon_sym_readonly] = ACTIONS(697),
    [anon_sym_local] = ACTIONS(697),
    [anon_sym_unset] = ACTIONS(699),
    [anon_sym_unsetenv] = ACTIONS(699),
    [anon_sym_AMP_GT] = ACTIONS(673),
    [anon_sym_AMP_GT_GT] = ACTIONS(675),
    [anon_sym_LT_AMP] = ACTIONS(673),
    [anon_sym_GT_AMP] = ACTIONS(673),
    [anon_sym_GT_PIPE] = ACTIONS(675),
    [anon_sym_LT_AMP_DASH] = ACTIONS(701),
    [anon_sym_GT_AMP_DASH] = ACTIONS(701),
    [anon_sym_LT_LT_LT] = ACTIONS(703),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(705),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(707),
    [anon_sym_DOLLAR] = ACTIONS(709),
    [sym__special_character] = ACTIONS(711),
    [anon_sym_DQUOTE] = ACTIONS(713),
    [sym_raw_string] = ACTIONS(715),
    [sym_ansi_c_string] = ACTIONS(715),
    [aux_sym_number_token1] = ACTIONS(717),
    [aux_sym_number_token2] = ACTIONS(719),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(721),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(723),
    [anon_sym_BQUOTE] = ACTIONS(725),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(727),
    [anon_sym_LT_LPAREN] = ACTIONS(729),
    [anon_sym_GT_LPAREN] = ACTIONS(729),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(731),
    [sym_variable_name] = ACTIONS(733),
    [sym_test_operator] = ACTIONS(735),
    [sym__brace_start] = ACTIONS(737),
  },
  [436] = {
    [sym__statement_not_pipeline] = STATE(6674),
    [sym_redirected_statement] = STATE(4528),
    [sym_for_statement] = STATE(4528),
    [sym_c_style_for_statement] = STATE(4528),
    [sym_while_statement] = STATE(4254),
    [sym_if_statement] = STATE(4254),
    [sym_case_statement] = STATE(4528),
    [sym_function_definition] = STATE(4528),
    [sym_compound_statement] = STATE(4528),
    [sym_subshell] = STATE(4528),
    [sym_pipeline] = STATE(4428),
    [sym_list] = STATE(4528),
    [sym_negated_command] = STATE(4528),
    [sym_test_command] = STATE(4528),
    [sym_declaration_command] = STATE(4528),
    [sym_unset_command] = STATE(4528),
    [sym_command] = STATE(4528),
    [sym_command_name] = STATE(530),
    [sym_variable_assignment] = STATE(839),
    [sym_variable_assignments] = STATE(4528),
    [sym_subscript] = STATE(6760),
    [sym_file_redirect] = STATE(1893),
    [sym_herestring_redirect] = STATE(1892),
    [sym_arithmetic_expansion] = STATE(852),
    [sym_brace_expression] = STATE(852),
    [sym_concatenation] = STATE(1169),
    [sym_string] = STATE(852),
    [sym_translated_string] = STATE(852),
    [sym_number] = STATE(852),
    [sym_simple_expansion] = STATE(852),
    [sym_expansion] = STATE(852),
    [sym_command_substitution] = STATE(852),
    [sym_process_substitution] = STATE(852),
    [aux_sym_redirected_statement_repeat2] = STATE(4371),
    [aux_sym_command_repeat1] = STATE(989),
    [aux_sym__literal_repeat1] = STATE(1026),
    [sym_word] = ACTIONS(665),
    [anon_sym_for] = ACTIONS(667),
    [anon_sym_select] = ACTIONS(669),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(671),
    [anon_sym_LT] = ACTIONS(673),
    [anon_sym_GT] = ACTIONS(673),
    [anon_sym_GT_GT] = ACTIONS(675),
    [anon_sym_LPAREN] = ACTIONS(677),
    [anon_sym_while] = ACTIONS(679),
    [anon_sym_until] = ACTIONS(679),
    [anon_sym_if] = ACTIONS(681),
    [anon_sym_case] = ACTIONS(683),
    [anon_sym_function] = ACTIONS(687),
    [anon_sym_LBRACE] = ACTIONS(689),
    [anon_sym_BANG] = ACTIONS(691),
    [anon_sym_LBRACK] = ACTIONS(693),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(695),
    [anon_sym_declare] = ACTIONS(697),
    [anon_sym_typeset] = ACTIONS(697),
    [anon_sym_export] = ACTIONS(697),
    [anon_sym_readonly] = ACTIONS(697),
    [anon_sym_local] = ACTIONS(697),
    [anon_sym_unset] = ACTIONS(699),
    [anon_sym_unsetenv] = ACTIONS(699),
    [anon_sym_AMP_GT] = ACTIONS(673),
    [anon_sym_AMP_GT_GT] = ACTIONS(675),
    [anon_sym_LT_AMP] = ACTIONS(673),
    [anon_sym_GT_AMP] = ACTIONS(673),
    [anon_sym_GT_PIPE] = ACTIONS(675),
    [anon_sym_LT_AMP_DASH] = ACTIONS(701),
    [anon_sym_GT_AMP_DASH] = ACTIONS(701),
    [anon_sym_LT_LT_LT] = ACTIONS(703),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(705),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(707),
    [anon_sym_DOLLAR] = ACTIONS(709),
    [sym__special_character] = ACTIONS(711),
    [anon_sym_DQUOTE] = ACTIONS(713),
    [sym_raw_string] = ACTIONS(715),
    [sym_ansi_c_string] = ACTIONS(715),
    [aux_sym_number_token1] = ACTIONS(717),
    [aux_sym_number_token2] = ACTIONS(719),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(721),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(723),
    [anon_sym_BQUOTE] = ACTIONS(725),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(727),
    [anon_sym_LT_LPAREN] = ACTIONS(729),
    [anon_sym_GT_LPAREN] = ACTIONS(729),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(731),
    [sym_variable_name] = ACTIONS(733),
    [sym_test_operator] = ACTIONS(735),
    [sym__brace_start] = ACTIONS(737),
  },
  [437] = {
    [sym__statement_not_pipeline] = STATE(4794),
    [sym_redirected_statement] = STATE(4794),
    [sym_for_statement] = STATE(4794),
    [sym_c_style_for_statement] = STATE(4794),
    [sym_while_statement] = STATE(4269),
    [sym_if_statement] = STATE(4269),
    [sym_case_statement] = STATE(4794),
    [sym_function_definition] = STATE(4794),
    [sym_compound_statement] = STATE(4794),
    [sym_subshell] = STATE(4794),
    [sym_pipeline] = STATE(5558),
    [sym_list] = STATE(4794),
    [sym_negated_command] = STATE(4794),
    [sym_test_command] = STATE(4794),
    [sym_declaration_command] = STATE(4794),
    [sym_unset_command] = STATE(4794),
    [sym_command] = STATE(4794),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1028),
    [sym_variable_assignments] = STATE(4794),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [438] = {
    [sym__statement_not_pipeline] = STATE(4616),
    [sym_redirected_statement] = STATE(4616),
    [sym_for_statement] = STATE(4616),
    [sym_c_style_for_statement] = STATE(4616),
    [sym_while_statement] = STATE(4227),
    [sym_if_statement] = STATE(4227),
    [sym_case_statement] = STATE(4616),
    [sym_function_definition] = STATE(4616),
    [sym_compound_statement] = STATE(4616),
    [sym_subshell] = STATE(4616),
    [sym_pipeline] = STATE(5614),
    [sym_list] = STATE(4616),
    [sym_negated_command] = STATE(4616),
    [sym_test_command] = STATE(4616),
    [sym_declaration_command] = STATE(4616),
    [sym_unset_command] = STATE(4616),
    [sym_command] = STATE(4616),
    [sym_command_name] = STATE(528),
    [sym_variable_assignment] = STATE(822),
    [sym_variable_assignments] = STATE(4616),
    [sym_subscript] = STATE(6736),
    [sym_file_redirect] = STATE(1635),
    [sym_herestring_redirect] = STATE(1634),
    [sym_arithmetic_expansion] = STATE(815),
    [sym_brace_expression] = STATE(815),
    [sym_concatenation] = STATE(1082),
    [sym_string] = STATE(815),
    [sym_translated_string] = STATE(815),
    [sym_number] = STATE(815),
    [sym_simple_expansion] = STATE(815),
    [sym_expansion] = STATE(815),
    [sym_command_substitution] = STATE(815),
    [sym_process_substitution] = STATE(815),
    [aux_sym_redirected_statement_repeat2] = STATE(4250),
    [aux_sym_command_repeat1] = STATE(925),
    [aux_sym__literal_repeat1] = STATE(926),
    [sym_word] = ACTIONS(403),
    [anon_sym_for] = ACTIONS(405),
    [anon_sym_select] = ACTIONS(407),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(409),
    [anon_sym_LT] = ACTIONS(411),
    [anon_sym_GT] = ACTIONS(411),
    [anon_sym_GT_GT] = ACTIONS(413),
    [anon_sym_LPAREN] = ACTIONS(415),
    [anon_sym_while] = ACTIONS(417),
    [anon_sym_until] = ACTIONS(417),
    [anon_sym_if] = ACTIONS(419),
    [anon_sym_case] = ACTIONS(421),
    [anon_sym_function] = ACTIONS(431),
    [anon_sym_LBRACE] = ACTIONS(433),
    [anon_sym_BANG] = ACTIONS(435),
    [anon_sym_LBRACK] = ACTIONS(437),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(439),
    [anon_sym_declare] = ACTIONS(441),
    [anon_sym_typeset] = ACTIONS(441),
    [anon_sym_export] = ACTIONS(441),
    [anon_sym_readonly] = ACTIONS(441),
    [anon_sym_local] = ACTIONS(441),
    [anon_sym_unset] = ACTIONS(443),
    [anon_sym_unsetenv] = ACTIONS(443),
    [anon_sym_AMP_GT] = ACTIONS(411),
    [anon_sym_AMP_GT_GT] = ACTIONS(413),
    [anon_sym_LT_AMP] = ACTIONS(411),
    [anon_sym_GT_AMP] = ACTIONS(411),
    [anon_sym_GT_PIPE] = ACTIONS(413),
    [anon_sym_LT_AMP_DASH] = ACTIONS(445),
    [anon_sym_GT_AMP_DASH] = ACTIONS(445),
    [anon_sym_LT_LT_LT] = ACTIONS(447),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(449),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(451),
    [anon_sym_DOLLAR] = ACTIONS(453),
    [sym__special_character] = ACTIONS(455),
    [anon_sym_DQUOTE] = ACTIONS(457),
    [sym_raw_string] = ACTIONS(459),
    [sym_ansi_c_string] = ACTIONS(459),
    [aux_sym_number_token1] = ACTIONS(461),
    [aux_sym_number_token2] = ACTIONS(463),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(465),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(467),
    [anon_sym_BQUOTE] = ACTIONS(469),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(471),
    [anon_sym_LT_LPAREN] = ACTIONS(473),
    [anon_sym_GT_LPAREN] = ACTIONS(473),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(475),
    [sym_variable_name] = ACTIONS(477),
    [sym_test_operator] = ACTIONS(479),
    [sym__brace_start] = ACTIONS(481),
  },
  [439] = {
    [sym__statement_not_pipeline] = STATE(5561),
    [sym_redirected_statement] = STATE(5561),
    [sym_for_statement] = STATE(5561),
    [sym_c_style_for_statement] = STATE(5561),
    [sym_while_statement] = STATE(5213),
    [sym_if_statement] = STATE(5213),
    [sym_case_statement] = STATE(5561),
    [sym_function_definition] = STATE(5561),
    [sym_compound_statement] = STATE(5561),
    [sym_subshell] = STATE(5561),
    [sym_pipeline] = STATE(5551),
    [sym_list] = STATE(5561),
    [sym_negated_command] = STATE(5561),
    [sym_test_command] = STATE(5561),
    [sym_declaration_command] = STATE(5561),
    [sym_unset_command] = STATE(5561),
    [sym_command] = STATE(5561),
    [sym_command_name] = STATE(620),
    [sym_variable_assignment] = STATE(1907),
    [sym_variable_assignments] = STATE(5561),
    [sym_subscript] = STATE(6756),
    [sym_file_redirect] = STATE(2780),
    [sym_herestring_redirect] = STATE(2886),
    [sym_arithmetic_expansion] = STATE(1799),
    [sym_brace_expression] = STATE(1799),
    [sym_concatenation] = STATE(2368),
    [sym_string] = STATE(1799),
    [sym_translated_string] = STATE(1799),
    [sym_number] = STATE(1799),
    [sym_simple_expansion] = STATE(1799),
    [sym_expansion] = STATE(1799),
    [sym_command_substitution] = STATE(1799),
    [sym_process_substitution] = STATE(1799),
    [aux_sym_redirected_statement_repeat2] = STATE(5362),
    [aux_sym_command_repeat1] = STATE(955),
    [aux_sym__literal_repeat1] = STATE(2332),
    [sym_word] = ACTIONS(1025),
    [anon_sym_for] = ACTIONS(145),
    [anon_sym_select] = ACTIONS(147),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(1027),
    [anon_sym_LT] = ACTIONS(1029),
    [anon_sym_GT] = ACTIONS(1029),
    [anon_sym_GT_GT] = ACTIONS(1031),
    [anon_sym_LPAREN] = ACTIONS(1033),
    [anon_sym_while] = ACTIONS(157),
    [anon_sym_until] = ACTIONS(157),
    [anon_sym_if] = ACTIONS(159),
    [anon_sym_case] = ACTIONS(161),
    [anon_sym_function] = ACTIONS(163),
    [anon_sym_LBRACE] = ACTIONS(165),
    [anon_sym_BANG] = ACTIONS(1035),
    [anon_sym_LBRACK] = ACTIONS(169),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(173),
    [anon_sym_declare] = ACTIONS(175),
    [anon_sym_typeset] = ACTIONS(175),
    [anon_sym_export] = ACTIONS(175),
    [anon_sym_readonly] = ACTIONS(175),
    [anon_sym_local] = ACTIONS(175),
    [anon_sym_unset] = ACTIONS(177),
    [anon_sym_unsetenv] = ACTIONS(177),
    [anon_sym_AMP_GT] = ACTIONS(1029),
    [anon_sym_AMP_GT_GT] = ACTIONS(1031),
    [anon_sym_LT_AMP] = ACTIONS(1029),
    [anon_sym_GT_AMP] = ACTIONS(1029),
    [anon_sym_GT_PIPE] = ACTIONS(1031),
    [anon_sym_LT_AMP_DASH] = ACTIONS(1037),
    [anon_sym_GT_AMP_DASH] = ACTIONS(1037),
    [anon_sym_LT_LT_LT] = ACTIONS(1039),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(1041),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(1043),
    [anon_sym_DOLLAR] = ACTIONS(1045),
    [sym__special_character] = ACTIONS(1047),
    [anon_sym_DQUOTE] = ACTIONS(1049),
    [sym_raw_string] = ACTIONS(1051),
    [sym_ansi_c_string] = ACTIONS(1051),
    [aux_sym_number_token1] = ACTIONS(1053),
    [aux_sym_number_token2] = ACTIONS(1055),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(1057),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1059),
    [anon_sym_BQUOTE] = ACTIONS(1061),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(1063),
    [anon_sym_LT_LPAREN] = ACTIONS(1065),
    [anon_sym_GT_LPAREN] = ACTIONS(1065),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(1067),
    [sym_variable_name] = ACTIONS(217),
    [sym_test_operator] = ACTIONS(1069),
    [sym__brace_start] = ACTIONS(1071),
  },
  [440] = {
    [sym__statement_not_pipeline] = STATE(6566),
    [sym_redirected_statement] = STATE(5383),
    [sym_for_statement] = STATE(5383),
    [sym_c_style_for_statement] = STATE(5383),
    [sym_while_statement] = STATE(5117),
    [sym_if_statement] = STATE(5117),
    [sym_case_statement] = STATE(5383),
    [sym_function_definition] = STATE(5383),
    [sym_compound_statement] = STATE(5383),
    [sym_subshell] = STATE(5383),
    [sym_pipeline] = STATE(5535),
    [sym_list] = STATE(5383),
    [sym_negated_command] = STATE(5383),
    [sym_test_command] = STATE(5383),
    [sym_declaration_command] = STATE(5383),
    [sym_unset_command] = STATE(5383),
    [sym_command] = STATE(5383),
    [sym_command_name] = STATE(596),
    [sym_variable_assignment] = STATE(1707),
    [sym_variable_assignments] = STATE(5383),
    [sym_subscript] = STATE(6759),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(1717),
    [sym_brace_expression] = STATE(1717),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1717),
    [sym_translated_string] = STATE(1717),
    [sym_number] = STATE(1717),
    [sym_simple_expansion] = STATE(1717),
    [sym_expansion] = STATE(1717),
    [sym_command_substitution] = STATE(1717),
    [sym_process_substitution] = STATE(1717),
    [aux_sym_redirected_statement_repeat2] = STATE(5176),
    [aux_sym_command_repeat1] = STATE(933),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(997),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(999),
    [anon_sym_GT] = ACTIONS(999),
    [anon_sym_GT_GT] = ACTIONS(1001),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(1003),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(1005),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(1007),
    [anon_sym_typeset] = ACTIONS(1007),
    [anon_sym_export] = ACTIONS(1007),
    [anon_sym_readonly] = ACTIONS(1007),
    [anon_sym_local] = ACTIONS(1007),
    [anon_sym_unset] = ACTIONS(1009),
    [anon_sym_unsetenv] = ACTIONS(1009),
    [anon_sym_AMP_GT] = ACTIONS(999),
    [anon_sym_AMP_GT_GT] = ACTIONS(1001),
    [anon_sym_LT_AMP] = ACTIONS(999),
    [anon_sym_GT_AMP] = ACTIONS(999),
    [anon_sym_GT_PIPE] = ACTIONS(1001),
    [anon_sym_LT_AMP_DASH] = ACTIONS(1011),
    [anon_sym_GT_AMP_DASH] = ACTIONS(1011),
    [anon_sym_LT_LT_LT] = ACTIONS(1013),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(1015),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(1017),
    [sym_ansi_c_string] = ACTIONS(1017),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(1019),
    [sym_variable_name] = ACTIONS(1021),
    [sym_test_operator] = ACTIONS(1023),
    [sym__brace_start] = ACTIONS(344),
  },
  [441] = {
    [sym__statement_not_pipeline] = STATE(6614),
    [sym_redirected_statement] = STATE(4717),
    [sym_for_statement] = STATE(4717),
    [sym_c_style_for_statement] = STATE(4717),
    [sym_while_statement] = STATE(4376),
    [sym_if_statement] = STATE(4376),
    [sym_case_statement] = STATE(4717),
    [sym_function_definition] = STATE(4717),
    [sym_compound_statement] = STATE(4717),
    [sym_subshell] = STATE(4717),
    [sym_pipeline] = STATE(4716),
    [sym_list] = STATE(4717),
    [sym_negated_command] = STATE(4717),
    [sym_test_command] = STATE(4717),
    [sym_declaration_command] = STATE(4717),
    [sym_unset_command] = STATE(4717),
    [sym_command] = STATE(4717),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(948),
    [sym_variable_assignments] = STATE(4717),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [442] = {
    [sym__statement_not_pipeline] = STATE(4794),
    [sym_redirected_statement] = STATE(4794),
    [sym_for_statement] = STATE(4794),
    [sym_c_style_for_statement] = STATE(4794),
    [sym_while_statement] = STATE(4353),
    [sym_if_statement] = STATE(4353),
    [sym_case_statement] = STATE(4794),
    [sym_function_definition] = STATE(4794),
    [sym_compound_statement] = STATE(4794),
    [sym_subshell] = STATE(4794),
    [sym_pipeline] = STATE(5550),
    [sym_list] = STATE(4794),
    [sym_negated_command] = STATE(4794),
    [sym_test_command] = STATE(4794),
    [sym_declaration_command] = STATE(4794),
    [sym_unset_command] = STATE(4794),
    [sym_command] = STATE(4794),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(961),
    [sym_variable_assignments] = STATE(4794),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(938),
    [sym_brace_expression] = STATE(938),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(938),
    [sym_translated_string] = STATE(938),
    [sym_number] = STATE(938),
    [sym_simple_expansion] = STATE(938),
    [sym_expansion] = STATE(938),
    [sym_command_substitution] = STATE(938),
    [sym_process_substitution] = STATE(938),
    [aux_sym_redirected_statement_repeat2] = STATE(4480),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(765),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(85),
    [anon_sym_GT] = ACTIONS(85),
    [anon_sym_GT_GT] = ACTIONS(87),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(91),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(769),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(95),
    [anon_sym_typeset] = ACTIONS(95),
    [anon_sym_export] = ACTIONS(95),
    [anon_sym_readonly] = ACTIONS(95),
    [anon_sym_local] = ACTIONS(95),
    [anon_sym_unset] = ACTIONS(97),
    [anon_sym_unsetenv] = ACTIONS(97),
    [anon_sym_AMP_GT] = ACTIONS(85),
    [anon_sym_AMP_GT_GT] = ACTIONS(87),
    [anon_sym_LT_AMP] = ACTIONS(85),
    [anon_sym_GT_AMP] = ACTIONS(85),
    [anon_sym_GT_PIPE] = ACTIONS(87),
    [anon_sym_LT_AMP_DASH] = ACTIONS(99),
    [anon_sym_GT_AMP_DASH] = ACTIONS(99),
    [anon_sym_LT_LT_LT] = ACTIONS(101),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(771),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(773),
    [sym_ansi_c_string] = ACTIONS(773),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(135),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(775),
    [sym__brace_start] = ACTIONS(344),
  },
  [443] = {
    [sym__statement_not_pipeline] = STATE(4794),
    [sym_redirected_statement] = STATE(4794),
    [sym_for_statement] = STATE(4794),
    [sym_c_style_for_statement] = STATE(4794),
    [sym_while_statement] = STATE(4565),
    [sym_if_statement] = STATE(4565),
    [sym_case_statement] = STATE(4794),
    [sym_function_definition] = STATE(4794),
    [sym_compound_statement] = STATE(4794),
    [sym_subshell] = STATE(4794),
    [sym_pipeline] = STATE(5554),
    [sym_list] = STATE(4794),
    [sym_negated_command] = STATE(4794),
    [sym_test_command] = STATE(4794),
    [sym_declaration_command] = STATE(4794),
    [sym_unset_command] = STATE(4794),
    [sym_command] = STATE(4794),
    [sym_command_name] = STATE(544),
    [sym_variable_assignment] = STATE(1098),
    [sym_variable_assignments] = STATE(4794),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym_redirected_statement_repeat2] = STATE(4644),
    [aux_sym_command_repeat1] = STATE(997),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(286),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(300),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(302),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(304),
    [anon_sym_typeset] = ACTIONS(304),
    [anon_sym_export] = ACTIONS(304),
    [anon_sym_readonly] = ACTIONS(304),
    [anon_sym_local] = ACTIONS(304),
    [anon_sym_unset] = ACTIONS(306),
    [anon_sym_unsetenv] = ACTIONS(306),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [444] = {
    [sym__statement_not_pipeline] = STATE(6718),
    [sym_redirected_statement] = STATE(4996),
    [sym_for_statement] = STATE(4996),
    [sym_c_style_for_statement] = STATE(4996),
    [sym_while_statement] = STATE(4570),
    [sym_if_statement] = STATE(4570),
    [sym_case_statement] = STATE(4996),
    [sym_function_definition] = STATE(4996),
    [sym_compound_statement] = STATE(4996),
    [sym_subshell] = STATE(4996),
    [sym_pipeline] = STATE(4997),
    [sym_list] = STATE(4996),
    [sym_negated_command] = STATE(4996),
    [sym_test_command] = STATE(4996),
    [sym_declaration_command] = STATE(4996),
    [sym_unset_command] = STATE(4996),
    [sym_command] = STATE(4996),
    [sym_command_name] = STATE(544),
    [sym_variable_assignment] = STATE(1096),
    [sym_variable_assignments] = STATE(4996),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym_redirected_statement_repeat2] = STATE(4644),
    [aux_sym_command_repeat1] = STATE(997),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(286),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(300),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(302),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(304),
    [anon_sym_typeset] = ACTIONS(304),
    [anon_sym_export] = ACTIONS(304),
    [anon_sym_readonly] = ACTIONS(304),
    [anon_sym_local] = ACTIONS(304),
    [anon_sym_unset] = ACTIONS(306),
    [anon_sym_unsetenv] = ACTIONS(306),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [445] = {
    [sym__statement_not_pipeline] = STATE(6566),
    [sym_redirected_statement] = STATE(5368),
    [sym_for_statement] = STATE(5368),
    [sym_c_style_for_statement] = STATE(5368),
    [sym_while_statement] = STATE(5056),
    [sym_if_statement] = STATE(5056),
    [sym_case_statement] = STATE(5368),
    [sym_function_definition] = STATE(5368),
    [sym_compound_statement] = STATE(5368),
    [sym_subshell] = STATE(5368),
    [sym_pipeline] = STATE(5369),
    [sym_list] = STATE(5368),
    [sym_negated_command] = STATE(5368),
    [sym_test_command] = STATE(5368),
    [sym_declaration_command] = STATE(5368),
    [sym_unset_command] = STATE(5368),
    [sym_command] = STATE(5368),
    [sym_command_name] = STATE(596),
    [sym_variable_assignment] = STATE(1562),
    [sym_variable_assignments] = STATE(5368),
    [sym_subscript] = STATE(6759),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(1717),
    [sym_brace_expression] = STATE(1717),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1717),
    [sym_translated_string] = STATE(1717),
    [sym_number] = STATE(1717),
    [sym_simple_expansion] = STATE(1717),
    [sym_expansion] = STATE(1717),
    [sym_command_substitution] = STATE(1717),
    [sym_process_substitution] = STATE(1717),
    [aux_sym_redirected_statement_repeat2] = STATE(5176),
    [aux_sym_command_repeat1] = STATE(933),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(997),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(999),
    [anon_sym_GT] = ACTIONS(999),
    [anon_sym_GT_GT] = ACTIONS(1001),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(1003),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(1005),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(1007),
    [anon_sym_typeset] = ACTIONS(1007),
    [anon_sym_export] = ACTIONS(1007),
    [anon_sym_readonly] = ACTIONS(1007),
    [anon_sym_local] = ACTIONS(1007),
    [anon_sym_unset] = ACTIONS(1009),
    [anon_sym_unsetenv] = ACTIONS(1009),
    [anon_sym_AMP_GT] = ACTIONS(999),
    [anon_sym_AMP_GT_GT] = ACTIONS(1001),
    [anon_sym_LT_AMP] = ACTIONS(999),
    [anon_sym_GT_AMP] = ACTIONS(999),
    [anon_sym_GT_PIPE] = ACTIONS(1001),
    [anon_sym_LT_AMP_DASH] = ACTIONS(1011),
    [anon_sym_GT_AMP_DASH] = ACTIONS(1011),
    [anon_sym_LT_LT_LT] = ACTIONS(1013),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(1015),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(1017),
    [sym_ansi_c_string] = ACTIONS(1017),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(1019),
    [sym_variable_name] = ACTIONS(1021),
    [sym_test_operator] = ACTIONS(1023),
    [sym__brace_start] = ACTIONS(344),
  },
  [446] = {
    [sym__statement_not_pipeline] = STATE(4794),
    [sym_redirected_statement] = STATE(4794),
    [sym_for_statement] = STATE(4794),
    [sym_c_style_for_statement] = STATE(4794),
    [sym_while_statement] = STATE(5058),
    [sym_if_statement] = STATE(5058),
    [sym_case_statement] = STATE(4794),
    [sym_function_definition] = STATE(4794),
    [sym_compound_statement] = STATE(4794),
    [sym_subshell] = STATE(4794),
    [sym_pipeline] = STATE(5559),
    [sym_list] = STATE(4794),
    [sym_negated_command] = STATE(4794),
    [sym_test_command] = STATE(4794),
    [sym_declaration_command] = STATE(4794),
    [sym_unset_command] = STATE(4794),
    [sym_command] = STATE(4794),
    [sym_command_name] = STATE(596),
    [sym_variable_assignment] = STATE(1627),
    [sym_variable_assignments] = STATE(4794),
    [sym_subscript] = STATE(6759),
    [sym_file_redirect] = STATE(2019),
    [sym_herestring_redirect] = STATE(2017),
    [sym_arithmetic_expansion] = STATE(1717),
    [sym_brace_expression] = STATE(1717),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1717),
    [sym_translated_string] = STATE(1717),
    [sym_number] = STATE(1717),
    [sym_simple_expansion] = STATE(1717),
    [sym_expansion] = STATE(1717),
    [sym_command_substitution] = STATE(1717),
    [sym_process_substitution] = STATE(1717),
    [aux_sym_redirected_statement_repeat2] = STATE(5176),
    [aux_sym_command_repeat1] = STATE(933),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(997),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(999),
    [anon_sym_GT] = ACTIONS(999),
    [anon_sym_GT_GT] = ACTIONS(1001),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(1003),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(1005),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(1007),
    [anon_sym_typeset] = ACTIONS(1007),
    [anon_sym_export] = ACTIONS(1007),
    [anon_sym_readonly] = ACTIONS(1007),
    [anon_sym_local] = ACTIONS(1007),
    [anon_sym_unset] = ACTIONS(1009),
    [anon_sym_unsetenv] = ACTIONS(1009),
    [anon_sym_AMP_GT] = ACTIONS(999),
    [anon_sym_AMP_GT_GT] = ACTIONS(1001),
    [anon_sym_LT_AMP] = ACTIONS(999),
    [anon_sym_GT_AMP] = ACTIONS(999),
    [anon_sym_GT_PIPE] = ACTIONS(1001),
    [anon_sym_LT_AMP_DASH] = ACTIONS(1011),
    [anon_sym_GT_AMP_DASH] = ACTIONS(1011),
    [anon_sym_LT_LT_LT] = ACTIONS(1013),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(1015),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(1017),
    [sym_ansi_c_string] = ACTIONS(1017),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(1019),
    [sym_variable_name] = ACTIONS(1021),
    [sym_test_operator] = ACTIONS(1023),
    [sym__brace_start] = ACTIONS(344),
  },
  [447] = {
    [sym__statement_not_pipeline] = STATE(6568),
    [sym_redirected_statement] = STATE(5421),
    [sym_for_statement] = STATE(5421),
    [sym_c_style_for_statement] = STATE(5421),
    [sym_while_statement] = STATE(5184),
    [sym_if_statement] = STATE(5184),
    [sym_case_statement] = STATE(5421),
    [sym_function_definition] = STATE(5421),
    [sym_compound_statement] = STATE(5421),
    [sym_subshell] = STATE(5421),
    [sym_pipeline] = STATE(5419),
    [sym_list] = STATE(5421),
    [sym_negated_command] = STATE(5421),
    [sym_test_command] = STATE(5421),
    [sym_declaration_command] = STATE(5421),
    [sym_unset_command] = STATE(5421),
    [sym_command] = STATE(5421),
    [sym_command_name] = STATE(620),
    [sym_variable_assignment] = STATE(1882),
    [sym_variable_assignments] = STATE(5421),
    [sym_subscript] = STATE(6756),
    [sym_file_redirect] = STATE(2780),
    [sym_herestring_redirect] = STATE(2886),
    [sym_arithmetic_expansion] = STATE(1799),
    [sym_brace_expression] = STATE(1799),
    [sym_concatenation] = STATE(2368),
    [sym_string] = STATE(1799),
    [sym_translated_string] = STATE(1799),
    [sym_number] = STATE(1799),
    [sym_simple_expansion] = STATE(1799),
    [sym_expansion] = STATE(1799),
    [sym_command_substitution] = STATE(1799),
    [sym_process_substitution] = STATE(1799),
    [aux_sym_redirected_statement_repeat2] = STATE(5362),
    [aux_sym_command_repeat1] = STATE(955),
    [aux_sym__literal_repeat1] = STATE(2332),
    [sym_word] = ACTIONS(1025),
    [anon_sym_for] = ACTIONS(145),
    [anon_sym_select] = ACTIONS(147),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(1027),
    [anon_sym_LT] = ACTIONS(1029),
    [anon_sym_GT] = ACTIONS(1029),
    [anon_sym_GT_GT] = ACTIONS(1031),
    [anon_sym_LPAREN] = ACTIONS(1033),
    [anon_sym_while] = ACTIONS(157),
    [anon_sym_until] = ACTIONS(157),
    [anon_sym_if] = ACTIONS(159),
    [anon_sym_case] = ACTIONS(161),
    [anon_sym_function] = ACTIONS(163),
    [anon_sym_LBRACE] = ACTIONS(165),
    [anon_sym_BANG] = ACTIONS(1035),
    [anon_sym_LBRACK] = ACTIONS(169),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(173),
    [anon_sym_declare] = ACTIONS(175),
    [anon_sym_typeset] = ACTIONS(175),
    [anon_sym_export] = ACTIONS(175),
    [anon_sym_readonly] = ACTIONS(175),
    [anon_sym_local] = ACTIONS(175),
    [anon_sym_unset] = ACTIONS(177),
    [anon_sym_unsetenv] = ACTIONS(177),
    [anon_sym_AMP_GT] = ACTIONS(1029),
    [anon_sym_AMP_GT_GT] = ACTIONS(1031),
    [anon_sym_LT_AMP] = ACTIONS(1029),
    [anon_sym_GT_AMP] = ACTIONS(1029),
    [anon_sym_GT_PIPE] = ACTIONS(1031),
    [anon_sym_LT_AMP_DASH] = ACTIONS(1037),
    [anon_sym_GT_AMP_DASH] = ACTIONS(1037),
    [anon_sym_LT_LT_LT] = ACTIONS(1039),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(1041),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(1043),
    [anon_sym_DOLLAR] = ACTIONS(1045),
    [sym__special_character] = ACTIONS(1047),
    [anon_sym_DQUOTE] = ACTIONS(1049),
    [sym_raw_string] = ACTIONS(1051),
    [sym_ansi_c_string] = ACTIONS(1051),
    [aux_sym_number_token1] = ACTIONS(1053),
    [aux_sym_number_token2] = ACTIONS(1055),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(1057),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1059),
    [anon_sym_BQUOTE] = ACTIONS(1061),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(1063),
    [anon_sym_LT_LPAREN] = ACTIONS(1065),
    [anon_sym_GT_LPAREN] = ACTIONS(1065),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(1067),
    [sym_variable_name] = ACTIONS(217),
    [sym_test_operator] = ACTIONS(1069),
    [sym__brace_start] = ACTIONS(1071),
  },
  [448] = {
    [sym__statement_not_pipeline] = STATE(4794),
    [sym_redirected_statement] = STATE(4794),
    [sym_for_statement] = STATE(4794),
    [sym_c_style_for_statement] = STATE(4794),
    [sym_while_statement] = STATE(4381),
    [sym_if_statement] = STATE(4381),
    [sym_case_statement] = STATE(4794),
    [sym_function_definition] = STATE(4794),
    [sym_compound_statement] = STATE(4794),
    [sym_subshell] = STATE(4794),
    [sym_pipeline] = STATE(5557),
    [sym_list] = STATE(4794),
    [sym_negated_command] = STATE(4794),
    [sym_test_command] = STATE(4794),
    [sym_declaration_command] = STATE(4794),
    [sym_unset_command] = STATE(4794),
    [sym_command] = STATE(4794),
    [sym_command_name] = STATE(534),
    [sym_variable_assignment] = STATE(986),
    [sym_variable_assignments] = STATE(4794),
    [sym_subscript] = STATE(6774),
    [sym_file_redirect] = STATE(2055),
    [sym_herestring_redirect] = STATE(2156),
    [sym_arithmetic_expansion] = STATE(957),
    [sym_brace_expression] = STATE(957),
    [sym_concatenation] = STATE(1386),
    [sym_string] = STATE(957),
    [sym_translated_string] = STATE(957),
    [sym_number] = STATE(957),
    [sym_simple_expansion] = STATE(957),
    [sym_expansion] = STATE(957),
    [sym_command_substitution] = STATE(957),
    [sym_process_substitution] = STATE(957),
    [aux_sym_redirected_statement_repeat2] = STATE(4571),
    [aux_sym_command_repeat1] = STATE(937),
    [aux_sym__literal_repeat1] = STATE(1212),
    [sym_word] = ACTIONS(7),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(13),
    [anon_sym_LT] = ACTIONS(15),
    [anon_sym_GT] = ACTIONS(15),
    [anon_sym_GT_GT] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(31),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(37),
    [anon_sym_typeset] = ACTIONS(37),
    [anon_sym_export] = ACTIONS(37),
    [anon_sym_readonly] = ACTIONS(37),
    [anon_sym_local] = ACTIONS(37),
    [anon_sym_unset] = ACTIONS(39),
    [anon_sym_unsetenv] = ACTIONS(39),
    [anon_sym_AMP_GT] = ACTIONS(15),
    [anon_sym_AMP_GT_GT] = ACTIONS(17),
    [anon_sym_LT_AMP] = ACTIONS(15),
    [anon_sym_GT_AMP] = ACTIONS(15),
    [anon_sym_GT_PIPE] = ACTIONS(17),
    [anon_sym_LT_AMP_DASH] = ACTIONS(41),
    [anon_sym_GT_AMP_DASH] = ACTIONS(41),
    [anon_sym_LT_LT_LT] = ACTIONS(43),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(45),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(47),
    [anon_sym_DOLLAR] = ACTIONS(49),
    [sym__special_character] = ACTIONS(51),
    [anon_sym_DQUOTE] = ACTIONS(53),
    [sym_raw_string] = ACTIONS(55),
    [sym_ansi_c_string] = ACTIONS(55),
    [aux_sym_number_token1] = ACTIONS(57),
    [aux_sym_number_token2] = ACTIONS(59),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(61),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(63),
    [anon_sym_BQUOTE] = ACTIONS(65),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(67),
    [anon_sym_LT_LPAREN] = ACTIONS(69),
    [anon_sym_GT_LPAREN] = ACTIONS(69),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(73),
    [sym_variable_name] = ACTIONS(75),
    [sym_test_operator] = ACTIONS(77),
    [sym__brace_start] = ACTIONS(79),
  },
  [449] = {
    [sym__statement_not_pipeline] = STATE(6558),
    [sym_redirected_statement] = STATE(4722),
    [sym_for_statement] = STATE(4722),
    [sym_c_style_for_statement] = STATE(4722),
    [sym_while_statement] = STATE(4289),
    [sym_if_statement] = STATE(4289),
    [sym_case_statement] = STATE(4722),
    [sym_function_definition] = STATE(4722),
    [sym_compound_statement] = STATE(4722),
    [sym_subshell] = STATE(4722),
    [sym_pipeline] = STATE(4724),
    [sym_list] = STATE(4722),
    [sym_negated_command] = STATE(4722),
    [sym_test_command] = STATE(4722),
    [sym_declaration_command] = STATE(4722),
    [sym_unset_command] = STATE(4722),
    [sym_command] = STATE(4722),
    [sym_command_name] = STATE(554),
    [sym_variable_assignment] = STATE(1059),
    [sym_variable_assignments] = STATE(4722),
    [sym_subscript] = STATE(6790),
    [sym_file_redirect] = STATE(2273),
    [sym_herestring_redirect] = STATE(2272),
    [sym_arithmetic_expansion] = STATE(1078),
    [sym_brace_expression] = STATE(1078),
    [sym_concatenation] = STATE(1376),
    [sym_string] = STATE(1078),
    [sym_translated_string] = STATE(1078),
    [sym_number] = STATE(1078),
    [sym_simple_expansion] = STATE(1078),
    [sym_expansion] = STATE(1078),
    [sym_command_substitution] = STATE(1078),
    [sym_process_substitution] = STATE(1078),
    [aux_sym_redirected_statement_repeat2] = STATE(4573),
    [aux_sym_command_repeat1] = STATE(958),
    [aux_sym__literal_repeat1] = STATE(1267),
    [sym_word] = ACTIONS(799),
    [anon_sym_for] = ACTIONS(9),
    [anon_sym_select] = ACTIONS(11),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(290),
    [anon_sym_GT_GT] = ACTIONS(292),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_while] = ACTIONS(21),
    [anon_sym_until] = ACTIONS(21),
    [anon_sym_if] = ACTIONS(23),
    [anon_sym_case] = ACTIONS(25),
    [anon_sym_function] = ACTIONS(801),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_declare] = ACTIONS(805),
    [anon_sym_typeset] = ACTIONS(805),
    [anon_sym_export] = ACTIONS(805),
    [anon_sym_readonly] = ACTIONS(805),
    [anon_sym_local] = ACTIONS(805),
    [anon_sym_unset] = ACTIONS(807),
    [anon_sym_unsetenv] = ACTIONS(807),
    [anon_sym_AMP_GT] = ACTIONS(290),
    [anon_sym_AMP_GT_GT] = ACTIONS(292),
    [anon_sym_LT_AMP] = ACTIONS(290),
    [anon_sym_GT_AMP] = ACTIONS(290),
    [anon_sym_GT_PIPE] = ACTIONS(292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(308),
    [anon_sym_LT_LT_LT] = ACTIONS(310),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(314),
    [anon_sym_DOLLAR] = ACTIONS(316),
    [sym__special_character] = ACTIONS(318),
    [anon_sym_DQUOTE] = ACTIONS(320),
    [sym_raw_string] = ACTIONS(322),
    [sym_ansi_c_string] = ACTIONS(322),
    [aux_sym_number_token1] = ACTIONS(324),
    [aux_sym_number_token2] = ACTIONS(326),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(330),
    [anon_sym_BQUOTE] = ACTIONS(332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(334),
    [anon_sym_LT_LPAREN] = ACTIONS(336),
    [anon_sym_GT_LPAREN] = ACTIONS(336),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(338),
    [sym_variable_name] = ACTIONS(340),
    [sym_test_operator] = ACTIONS(342),
    [sym__brace_start] = ACTIONS(344),
  },
  [450] = {
    [sym__statement_not_pipeline] = STATE(6584),
    [sym_redirected_statement] = STATE(4342),
    [sym_for_statement] = STATE(4342),
    [sym_c_style_for_statement] = STATE(4342),
    [sym_while_statement] = STATE(4224),
    [sym_if_statement] = STATE(4224),
    [sym_case_statement] = STATE(4342),
    [sym_function_definition] = STATE(4342),
    [sym_compound_statement] = STATE(4342),
    [sym_subshell] = STATE(4342),
    [sym_pipeline] = STATE(4339),
    [sym_list] = STATE(4342),
    [sym_negated_command] = STATE(4342),
    [sym_test_command] = STATE(4342),
    [sym_declaration_command] = STATE(4342),
    [sym_unset_command] = STATE(4342),
    [sym_command] = STATE(4342),
    [sym_command_name] = STATE(528),
    [sym_variable_assignment] = STATE(809),
    [sym_variable_assignments] = STATE(4342),
    [sym_subscript] = STATE(6736),
    [sym_file_redirect] = STATE(1635),
    [sym_herestring_redirect] = STATE(1634),
    [sym_arithmetic_expansion] = STATE(815),
    [sym_brace_expression] = STATE(815),
    [sym_concatenation] = STATE(1082),
    [sym_string] = STATE(815),
    [sym_translated_string] = STATE(815),
    [sym_number] = STATE(815),
    [sym_simple_expansion] = STATE(815),
    [sym_expansion] = STATE(815),
    [sym_command_substitution] = STATE(815),
    [sym_process_substitution] = STATE(815),
    [aux_sym_redirected_statement_repeat2] = STATE(4250),
    [aux_sym_command_repeat1] = STATE(925),
    [aux_sym__literal_repeat1] = STATE(926),
    [sym_word] = ACTIONS(403),
    [anon_sym_for] = ACTIONS(405),
    [anon_sym_select] = ACTIONS(407),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(409),
    [anon_sym_LT] = ACTIONS(411),
    [anon_sym_GT] = ACTIONS(411),
    [anon_sym_GT_GT] = ACTIONS(413),
    [anon_sym_LPAREN] = ACTIONS(415),
    [anon_sym_while] = ACTIONS(417),
    [anon_sym_until] = ACTIONS(417),
    [anon_sym_if] = ACTIONS(419),
    [anon_sym_case] = ACTIONS(421),
    [anon_sym_function] = ACTIONS(431),
    [anon_sym_LBRACE] = ACTIONS(433),
    [anon_sym_BANG] = ACTIONS(435),
    [anon_sym_LBRACK] = ACTIONS(437),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(439),
    [anon_sym_declare] = ACTIONS(441),
    [anon_sym_typeset] = ACTIONS(441),
    [anon_sym_export] = ACTIONS(441),
    [anon_sym_readonly] = ACTIONS(441),
    [anon_sym_local] = ACTIONS(441),
    [anon_sym_unset] = ACTIONS(443),
    [anon_sym_unsetenv] = ACTIONS(443),
    [anon_sym_AMP_GT] = ACTIONS(411),
    [anon_sym_AMP_GT_GT] = ACTIONS(413),
    [anon_sym_LT_AMP] = ACTIONS(411),
    [anon_sym_GT_AMP] = ACTIONS(411),
    [anon_sym_GT_PIPE] = ACTIONS(413),
    [anon_sym_LT_AMP_DASH] = ACTIONS(445),
    [anon_sym_GT_AMP_DASH] = ACTIONS(445),
    [anon_sym_LT_LT_LT] = ACTIONS(447),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(449),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(451),
    [anon_sym_DOLLAR] = ACTIONS(453),
    [sym__special_character] = ACTIONS(455),
    [anon_sym_DQUOTE] = ACTIONS(457),
    [sym_raw_string] = ACTIONS(459),
    [sym_ansi_c_string] = ACTIONS(459),
    [aux_sym_number_token1] = ACTIONS(461),
    [aux_sym_number_token2] = ACTIONS(463),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(465),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(467),
    [anon_sym_BQUOTE] = ACTIONS(469),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(471),
    [anon_sym_LT_LPAREN] = ACTIONS(473),
    [anon_sym_GT_LPAREN] = ACTIONS(473),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(475),
    [sym_variable_name] = ACTIONS(477),
    [sym_test_operator] = ACTIONS(479),
    [sym__brace_start] = ACTIONS(481),
  },
  [451] = {
    [sym__expression] = STATE(3098),
    [sym_binary_expression] = STATE(3053),
    [sym_ternary_expression] = STATE(3053),
    [sym_unary_expression] = STATE(3053),
    [sym_postfix_expression] = STATE(3053),
    [sym_parenthesized_expression] = STATE(3053),
    [sym_arithmetic_expansion] = STATE(2500),
    [sym_brace_expression] = STATE(2500),
    [sym_concatenation] = STATE(3053),
    [sym_string] = STATE(2500),
    [sym_translated_string] = STATE(2500),
    [sym_number] = STATE(2500),
    [sym_simple_expansion] = STATE(2500),
    [sym_expansion] = STATE(2500),
    [sym_command_substitution] = STATE(2500),
    [sym_process_substitution] = STATE(2500),
    [aux_sym__literal_repeat1] = STATE(2690),
    [aux_sym_concatenation_repeat1] = STATE(2491),
    [sym_word] = ACTIONS(237),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(1073),
    [anon_sym_EQ] = ACTIONS(243),
    [anon_sym_PLUS_PLUS] = ACTIONS(243),
    [anon_sym_DASH_DASH] = ACTIONS(243),
    [anon_sym_PLUS_EQ] = ACTIONS(243),
    [anon_sym_DASH_EQ] = ACTIONS(243),
    [anon_sym_STAR_EQ] = ACTIONS(243),
    [anon_sym_SLASH_EQ] = ACTIONS(243),
    [anon_sym_PERCENT_EQ] = ACTIONS(243),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(243),
    [anon_sym_LT_LT_EQ] = ACTIONS(354),
    [anon_sym_GT_GT_EQ] = ACTIONS(354),
    [anon_sym_AMP_EQ] = ACTIONS(354),
    [anon_sym_CARET_EQ] = ACTIONS(243),
    [anon_sym_PIPE_EQ] = ACTIONS(354),
    [anon_sym_PIPE_PIPE] = ACTIONS(354),
    [anon_sym_AMP_AMP] = ACTIONS(354),
    [anon_sym_PIPE] = ACTIONS(243),
    [anon_sym_CARET] = ACTIONS(243),
    [anon_sym_AMP] = ACTIONS(243),
    [anon_sym_EQ_EQ] = ACTIONS(243),
    [anon_sym_BANG_EQ] = ACTIONS(243),
    [anon_sym_LT] = ACTIONS(243),
    [anon_sym_GT] = ACTIONS(243),
    [anon_sym_LT_EQ] = ACTIONS(354),
    [anon_sym_GT_EQ] = ACTIONS(354),
    [anon_sym_LT_LT] = ACTIONS(243),
    [anon_sym_GT_GT] = ACTIONS(243),
    [anon_sym_PLUS] = ACTIONS(243),
    [anon_sym_DASH] = ACTIONS(243),
    [anon_sym_STAR] = ACTIONS(243),
    [anon_sym_SLASH] = ACTIONS(243),
    [anon_sym_PERCENT] = ACTIONS(243),
    [anon_sym_STAR_STAR] = ACTIONS(243),
    [anon_sym_LPAREN] = ACTIONS(248),
    [anon_sym_RPAREN] = ACTIONS(354),
    [anon_sym_BANG] = ACTIONS(250),
    [anon_sym_EQ_TILDE] = ACTIONS(243),
    [anon_sym_QMARK] = ACTIONS(243),
    [anon_sym_PLUS_PLUS2] = ACTIONS(103),
    [anon_sym_DASH_DASH2] = ACTIONS(103),
    [anon_sym_DASH2] = ACTIONS(105),
    [anon_sym_PLUS2] = ACTIONS(105),
    [anon_sym_TILDE] = ACTIONS(107),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(1073),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(1075),
    [aux_sym_concatenation_token1] = ACTIONS(1077),
    [anon_sym_DOLLAR] = ACTIONS(258),
    [sym__special_character] = ACTIONS(1079),
    [anon_sym_DQUOTE] = ACTIONS(1081),
    [sym_raw_string] = ACTIONS(1083),
    [sym_ansi_c_string] = ACTIONS(1083),
    [aux_sym_number_token1] = ACTIONS(264),
    [aux_sym_number_token2] = ACTIONS(266),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(1085),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(270),
    [anon_sym_BQUOTE] = ACTIONS(272),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(1087),
    [anon_sym_LT_LPAREN] = ACTIONS(1089),
    [anon_sym_GT_LPAREN] = ACTIONS(1089),
    [sym_comment] = ACTIONS(71),
    [sym__concat] = ACTIONS(1077),
    [sym_test_operator] = ACTIONS(282),
    [sym__brace_start] = ACTIONS(284),
  },
  [452] = {
    [sym__expression] = STATE(2933),
    [sym_binary_expression] = STATE(2863),
    [sym_ternary_expression] = STATE(2863),
    [sym_unary_expression] = STATE(2863),
    [sym_postfix_expression] = STATE(2863),
    [sym_parenthesized_expression] = STATE(2863),
    [sym_arithmetic_expansion] = STATE(2451),
    [sym_brace_expression] = STATE(2451),
    [sym_concatenation] = STATE(2863),
    [sym_string] = STATE(2451),
    [sym_translated_string] = STATE(2451),
    [sym_number] = STATE(2451),
    [sym_simple_expansion] = STATE(2451),
    [sym_expansion] = STATE(2451),
    [sym_command_substitution] = STATE(2451),
    [sym_process_substitution] = STATE(2451),
    [aux_sym__literal_repeat1] = STATE(2472),
    [aux_sym_concatenation_repeat1] = STATE(2540),
    [sym_word] = ACTIONS(1091),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(352),
    [anon_sym_EQ] = ACTIONS(243),
    [anon_sym_PLUS_PLUS] = ACTIONS(243),
    [anon_sym_DASH_DASH] = ACTIONS(243),
    [anon_sym_PLUS_EQ] = ACTIONS(243),
    [anon_sym_DASH_EQ] = ACTIONS(243),
    [anon_sym_STAR_EQ] = ACTIONS(243),
    [anon_sym_SLASH_EQ] = ACTIONS(243),
    [anon_sym_PERCENT_EQ] = ACTIONS(243),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(243),
    [anon_sym_LT_LT_EQ] = ACTIONS(354),
    [anon_sym_GT_GT_EQ] = ACTIONS(354),
    [anon_sym_AMP_EQ] = ACTIONS(354),
    [anon_sym_CARET_EQ] = ACTIONS(243),
    [anon_sym_PIPE_EQ] = ACTIONS(354),
    [anon_sym_PIPE_PIPE] = ACTIONS(354),
    [anon_sym_AMP_AMP] = ACTIONS(354),
    [anon_sym_PIPE] = ACTIONS(243),
    [anon_sym_CARET] = ACTIONS(243),
    [anon_sym_AMP] = ACTIONS(243),
    [anon_sym_EQ_EQ] = ACTIONS(243),
    [anon_sym_BANG_EQ] = ACTIONS(243),
    [anon_sym_LT] = ACTIONS(243),
    [anon_sym_GT] = ACTIONS(243),
    [anon_sym_LT_EQ] = ACTIONS(354),
    [anon_sym_GT_EQ] = ACTIONS(354),
    [anon_sym_LT_LT] = ACTIONS(243),
    [anon_sym_GT_GT] = ACTIONS(243),
    [anon_sym_PLUS] = ACTIONS(243),
    [anon_sym_DASH] = ACTIONS(243),
    [anon_sym_STAR] = ACTIONS(243),
    [anon_sym_SLASH] = ACTIONS(243),
    [anon_sym_PERCENT] = ACTIONS(243),
    [anon_sym_STAR_STAR] = ACTIONS(243),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [anon_sym_BANG] = ACTIONS(1095),
    [anon_sym_RBRACK] = ACTIONS(1097),
    [anon_sym_EQ_TILDE] = ACTIONS(243),
    [anon_sym_QMARK] = ACTIONS(243),
    [anon_sym_PLUS_PLUS2] = ACTIONS(1099),
    [anon_sym_DASH_DASH2] = ACTIONS(1099),
    [anon_sym_DASH2] = ACTIONS(1101),
    [anon_sym_PLUS2] = ACTIONS(1101),
    [anon_sym_TILDE] = ACTIONS(1103),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(352),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(363),
    [aux_sym_concatenation_token1] = ACTIONS(1105),
    [anon_sym_DOLLAR] = ACTIONS(367),
    [sym__special_character] = ACTIONS(1107),
    [anon_sym_DQUOTE] = ACTIONS(371),
    [sym_raw_string] = ACTIONS(1109),
    [sym_ansi_c_string] = ACTIONS(1109),
    [aux_sym_number_token1] = ACTIONS(375),
    [aux_sym_number_token2] = ACTIONS(377),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(379),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(381),
    [anon_sym_BQUOTE] = ACTIONS(383),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(385),
    [anon_sym_LT_LPAREN] = ACTIONS(387),
    [anon_sym_GT_LPAREN] = ACTIONS(387),
    [sym_comment] = ACTIONS(71),
    [sym__concat] = ACTIONS(1111),
    [sym_test_operator] = ACTIONS(1113),
    [sym__brace_start] = ACTIONS(391),
  },
  [453] = {
    [sym__expression] = STATE(2933),
    [sym_binary_expression] = STATE(2863),
    [sym_ternary_expression] = STATE(2863),
    [sym_unary_expression] = STATE(2863),
    [sym_postfix_expression] = STATE(2863),
    [sym_parenthesized_expression] = STATE(2863),
    [sym_arithmetic_expansion] = STATE(2451),
    [sym_brace_expression] = STATE(2451),
    [sym_concatenation] = STATE(2863),
    [sym_string] = STATE(2451),
    [sym_translated_string] = STATE(2451),
    [sym_number] = STATE(2451),
    [sym_simple_expansion] = STATE(2451),
    [sym_expansion] = STATE(2451),
    [sym_command_substitution] = STATE(2451),
    [sym_process_substitution] = STATE(2451),
    [aux_sym__literal_repeat1] = STATE(2472),
    [aux_sym_concatenation_repeat1] = STATE(2540),
    [sym_word] = ACTIONS(1091),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(352),
    [anon_sym_EQ] = ACTIONS(243),
    [anon_sym_PLUS_PLUS] = ACTIONS(243),
    [anon_sym_DASH_DASH] = ACTIONS(243),
    [anon_sym_PLUS_EQ] = ACTIONS(243),
    [anon_sym_DASH_EQ] = ACTIONS(243),
    [anon_sym_STAR_EQ] = ACTIONS(243),
    [anon_sym_SLASH_EQ] = ACTIONS(243),
    [anon_sym_PERCENT_EQ] = ACTIONS(243),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(243),
    [anon_sym_LT_LT_EQ] = ACTIONS(354),
    [anon_sym_GT_GT_EQ] = ACTIONS(354),
    [anon_sym_AMP_EQ] = ACTIONS(354),
    [anon_sym_CARET_EQ] = ACTIONS(243),
    [anon_sym_PIPE_EQ] = ACTIONS(354),
    [anon_sym_PIPE_PIPE] = ACTIONS(354),
    [anon_sym_AMP_AMP] = ACTIONS(354),
    [anon_sym_PIPE] = ACTIONS(243),
    [anon_sym_CARET] = ACTIONS(243),
    [anon_sym_AMP] = ACTIONS(243),
    [anon_sym_EQ_EQ] = ACTIONS(243),
    [anon_sym_BANG_EQ] = ACTIONS(243),
    [anon_sym_LT] = ACTIONS(243),
    [anon_sym_GT] = ACTIONS(243),
    [anon_sym_LT_EQ] = ACTIONS(354),
    [anon_sym_GT_EQ] = ACTIONS(354),
    [anon_sym_LT_LT] = ACTIONS(243),
    [anon_sym_GT_GT] = ACTIONS(243),
    [anon_sym_PLUS] = ACTIONS(243),
    [anon_sym_DASH] = ACTIONS(243),
    [anon_sym_STAR] = ACTIONS(243),
    [anon_sym_SLASH] = ACTIONS(243),
    [anon_sym_PERCENT] = ACTIONS(243),
    [anon_sym_STAR_STAR] = ACTIONS(243),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [anon_sym_BANG] = ACTIONS(1095),
    [anon_sym_RBRACK] = ACTIONS(1115),
    [anon_sym_EQ_TILDE] = ACTIONS(243),
    [anon_sym_QMARK] = ACTIONS(243),
    [anon_sym_PLUS_PLUS2] = ACTIONS(1099),
    [anon_sym_DASH_DASH2] = ACTIONS(1099),
    [anon_sym_DASH2] = ACTIONS(1101),
    [anon_sym_PLUS2] = ACTIONS(1101),
    [anon_sym_TILDE] = ACTIONS(1103),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(352),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(363),
    [aux_sym_concatenation_token1] = ACTIONS(1105),
    [anon_sym_DOLLAR] = ACTIONS(367),
    [sym__special_character] = ACTIONS(1107),
    [anon_sym_DQUOTE] = ACTIONS(371),
    [sym_raw_string] = ACTIONS(1109),
    [sym_ansi_c_string] = ACTIONS(1109),
    [aux_sym_number_token1] = ACTIONS(375),
    [aux_sym_number_token2] = ACTIONS(377),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(379),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(381),
    [anon_sym_BQUOTE] = ACTIONS(383),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(385),
    [anon_sym_LT_LPAREN] = ACTIONS(387),
    [anon_sym_GT_LPAREN] = ACTIONS(387),
    [sym_comment] = ACTIONS(71),
    [sym__concat] = ACTIONS(1117),
    [sym_test_operator] = ACTIONS(1113),
    [sym__brace_start] = ACTIONS(391),
  },
  [454] = {
    [sym__expression] = STATE(3012),
    [sym_binary_expression] = STATE(3071),
    [sym_ternary_expression] = STATE(3071),
    [sym_unary_expression] = STATE(3071),
    [sym_postfix_expression] = STATE(3071),
    [sym_parenthesized_expression] = STATE(3071),
    [sym_arithmetic_expansion] = STATE(2521),
    [sym_brace_expression] = STATE(2521),
    [sym_concatenation] = STATE(3071),
    [sym_string] = STATE(2521),
    [sym_translated_string] = STATE(2521),
    [sym_number] = STATE(2521),
    [sym_simple_expansion] = STATE(2521),
    [sym_expansion] = STATE(2521),
    [sym_command_substitution] = STATE(2521),
    [sym_process_substitution] = STATE(2521),
    [aux_sym__literal_repeat1] = STATE(2641),
    [aux_sym_concatenation_repeat1] = STATE(2555),
    [sym_word] = ACTIONS(350),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(352),
    [anon_sym_EQ] = ACTIONS(243),
    [anon_sym_PLUS_PLUS] = ACTIONS(243),
    [anon_sym_DASH_DASH] = ACTIONS(243),
    [anon_sym_PLUS_EQ] = ACTIONS(243),
    [anon_sym_DASH_EQ] = ACTIONS(243),
    [anon_sym_STAR_EQ] = ACTIONS(243),
    [anon_sym_SLASH_EQ] = ACTIONS(243),
    [anon_sym_PERCENT_EQ] = ACTIONS(243),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(243),
    [anon_sym_LT_LT_EQ] = ACTIONS(354),
    [anon_sym_GT_GT_EQ] = ACTIONS(354),
    [anon_sym_AMP_EQ] = ACTIONS(354),
    [anon_sym_CARET_EQ] = ACTIONS(243),
    [anon_sym_PIPE_EQ] = ACTIONS(354),
    [anon_sym_PIPE_PIPE] = ACTIONS(354),
    [anon_sym_AMP_AMP] = ACTIONS(354),
    [anon_sym_PIPE] = ACTIONS(243),
    [anon_sym_CARET] = ACTIONS(243),
    [anon_sym_AMP] = ACTIONS(243),
    [anon_sym_EQ_EQ] = ACTIONS(243),
    [anon_sym_BANG_EQ] = ACTIONS(243),
    [anon_sym_LT] = ACTIONS(243),
    [anon_sym_GT] = ACTIONS(243),
    [anon_sym_LT_EQ] = ACTIONS(354),
    [anon_sym_GT_EQ] = ACTIONS(354),
    [anon_sym_LT_LT] = ACTIONS(243),
    [anon_sym_GT_GT] = ACTIONS(243),
    [anon_sym_PLUS] = ACTIONS(243),
    [anon_sym_DASH] = ACTIONS(243),
    [anon_sym_STAR] = ACTIONS(243),
    [anon_sym_SLASH] = ACTIONS(243),
    [anon_sym_PERCENT] = ACTIONS(243),
    [anon_sym_STAR_STAR] = ACTIONS(243),
    [anon_sym_LPAREN] = ACTIONS(359),
    [anon_sym_BANG] = ACTIONS(361),
    [anon_sym_RBRACK] = ACTIONS(354),
    [anon_sym_EQ_TILDE] = ACTIONS(243),
    [anon_sym_QMARK] = ACTIONS(243),
    [anon_sym_PLUS_PLUS2] = ACTIONS(183),
    [anon_sym_DASH_DASH2] = ACTIONS(183),
    [anon_sym_DASH2] = ACTIONS(185),
    [anon_sym_PLUS2] = ACTIONS(185),
    [anon_sym_TILDE] = ACTIONS(187),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(352),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(363),
    [aux_sym_concatenation_token1] = ACTIONS(1105),
    [anon_sym_DOLLAR] = ACTIONS(367),
    [sym__special_character] = ACTIONS(369),
    [anon_sym_DQUOTE] = ACTIONS(371),
    [sym_raw_string] = ACTIONS(373),
    [sym_ansi_c_string] = ACTIONS(373),
    [aux_sym_number_token1] = ACTIONS(375),
    [aux_sym_number_token2] = ACTIONS(377),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(379),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(381),
    [anon_sym_BQUOTE] = ACTIONS(383),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(385),
    [anon_sym_LT_LPAREN] = ACTIONS(387),
    [anon_sym_GT_LPAREN] = ACTIONS(387),
    [sym_comment] = ACTIONS(71),
    [sym__concat] = ACTIONS(1105),
    [sym_test_operator] = ACTIONS(389),
    [sym__brace_start] = ACTIONS(391),
  },
  [455] = {
    [sym__expression] = STATE(2933),
    [sym_binary_expression] = STATE(2863),
    [sym_ternary_expression] = STATE(2863),
    [sym_unary_expression] = STATE(2863),
    [sym_postfix_expression] = STATE(2863),
    [sym_parenthesized_expression] = STATE(2863),
    [sym_arithmetic_expansion] = STATE(2451),
    [sym_brace_expression] = STATE(2451),
    [sym_concatenation] = STATE(2863),
    [sym_string] = STATE(2451),
    [sym_translated_string] = STATE(2451),
    [sym_number] = STATE(2451),
    [sym_simple_expansion] = STATE(2451),
    [sym_expansion] = STATE(2451),
    [sym_command_substitution] = STATE(2451),
    [sym_process_substitution] = STATE(2451),
    [aux_sym__literal_repeat1] = STATE(2472),
    [aux_sym_concatenation_repeat1] = STATE(2540),
    [sym_word] = ACTIONS(1091),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(352),
    [anon_sym_EQ] = ACTIONS(243),
    [anon_sym_PLUS_PLUS] = ACTIONS(243),
    [anon_sym_DASH_DASH] = ACTIONS(243),
    [anon_sym_PLUS_EQ] = ACTIONS(243),
    [anon_sym_DASH_EQ] = ACTIONS(243),
    [anon_sym_STAR_EQ] = ACTIONS(243),
    [anon_sym_SLASH_EQ] = ACTIONS(243),
    [anon_sym_PERCENT_EQ] = ACTIONS(243),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(243),
    [anon_sym_LT_LT_EQ] = ACTIONS(354),
    [anon_sym_GT_GT_EQ] = ACTIONS(354),
    [anon_sym_AMP_EQ] = ACTIONS(354),
    [anon_sym_CARET_EQ] = ACTIONS(243),
    [anon_sym_PIPE_EQ] = ACTIONS(354),
    [anon_sym_PIPE_PIPE] = ACTIONS(354),
    [anon_sym_AMP_AMP] = ACTIONS(354),
    [anon_sym_PIPE] = ACTIONS(243),
    [anon_sym_CARET] = ACTIONS(243),
    [anon_sym_AMP] = ACTIONS(243),
    [anon_sym_EQ_EQ] = ACTIONS(243),
    [anon_sym_BANG_EQ] = ACTIONS(243),
    [anon_sym_LT] = ACTIONS(243),
    [anon_sym_GT] = ACTIONS(243),
    [anon_sym_LT_EQ] = ACTIONS(354),
    [anon_sym_GT_EQ] = ACTIONS(354),
    [anon_sym_LT_LT] = ACTIONS(243),
    [anon_sym_GT_GT] = ACTIONS(243),
    [anon_sym_PLUS] = ACTIONS(243),
    [anon_sym_DASH] = ACTIONS(243),
    [anon_sym_STAR] = ACTIONS(243),
    [anon_sym_SLASH] = ACTIONS(243),
    [anon_sym_PERCENT] = ACTIONS(243),
    [anon_sym_STAR_STAR] = ACTIONS(243),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [anon_sym_BANG] = ACTIONS(1095),
    [anon_sym_RBRACK] = ACTIONS(1119),
    [anon_sym_EQ_TILDE] = ACTIONS(243),
    [anon_sym_QMARK] = ACTIONS(243),
    [anon_sym_PLUS_PLUS2] = ACTIONS(1099),
    [anon_sym_DASH_DASH2] = ACTIONS(1099),
    [anon_sym_DASH2] = ACTIONS(1101),
    [anon_sym_PLUS2] = ACTIONS(1101),
    [anon_sym_TILDE] = ACTIONS(1103),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(352),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(363),
    [aux_sym_concatenation_token1] = ACTIONS(1105),
    [anon_sym_DOLLAR] = ACTIONS(367),
    [sym__special_character] = ACTIONS(1107),
    [anon_sym_DQUOTE] = ACTIONS(371),
    [sym_raw_string] = ACTIONS(1109),
    [sym_ansi_c_string] = ACTIONS(1109),
    [aux_sym_number_token1] = ACTIONS(375),
    [aux_sym_number_token2] = ACTIONS(377),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(379),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(381),
    [anon_sym_BQUOTE] = ACTIONS(383),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(385),
    [anon_sym_LT_LPAREN] = ACTIONS(387),
    [anon_sym_GT_LPAREN] = ACTIONS(387),
    [sym_comment] = ACTIONS(71),
    [sym__concat] = ACTIONS(1121),
    [sym_test_operator] = ACTIONS(1113),
    [sym__brace_start] = ACTIONS(391),
  },
  [456] = {
    [sym__expression] = STATE(2933),
    [sym_binary_expression] = STATE(2863),
    [sym_ternary_expression] = STATE(2863),
    [sym_unary_expression] = STATE(2863),
    [sym_postfix_expression] = STATE(2863),
    [sym_parenthesized_expression] = STATE(2863),
    [sym_arithmetic_expansion] = STATE(2451),
    [sym_brace_expression] = STATE(2451),
    [sym_concatenation] = STATE(2863),
    [sym_string] = STATE(2451),
    [sym_translated_string] = STATE(2451),
    [sym_number] = STATE(2451),
    [sym_simple_expansion] = STATE(2451),
    [sym_expansion] = STATE(2451),
    [sym_command_substitution] = STATE(2451),
    [sym_process_substitution] = STATE(2451),
    [aux_sym__literal_repeat1] = STATE(2472),
    [aux_sym_concatenation_repeat1] = STATE(2540),
    [sym_word] = ACTIONS(1091),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(352),
    [anon_sym_EQ] = ACTIONS(243),
    [anon_sym_PLUS_PLUS] = ACTIONS(243),
    [anon_sym_DASH_DASH] = ACTIONS(243),
    [anon_sym_PLUS_EQ] = ACTIONS(243),
    [anon_sym_DASH_EQ] = ACTIONS(243),
    [anon_sym_STAR_EQ] = ACTIONS(243),
    [anon_sym_SLASH_EQ] = ACTIONS(243),
    [anon_sym_PERCENT_EQ] = ACTIONS(243),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(243),
    [anon_sym_LT_LT_EQ] = ACTIONS(354),
    [anon_sym_GT_GT_EQ] = ACTIONS(354),
    [anon_sym_AMP_EQ] = ACTIONS(354),
    [anon_sym_CARET_EQ] = ACTIONS(243),
    [anon_sym_PIPE_EQ] = ACTIONS(354),
    [anon_sym_PIPE_PIPE] = ACTIONS(354),
    [anon_sym_AMP_AMP] = ACTIONS(354),
    [anon_sym_PIPE] = ACTIONS(243),
    [anon_sym_CARET] = ACTIONS(243),
    [anon_sym_AMP] = ACTIONS(243),
    [anon_sym_EQ_EQ] = ACTIONS(243),
    [anon_sym_BANG_EQ] = ACTIONS(243),
    [anon_sym_LT] = ACTIONS(243),
    [anon_sym_GT] = ACTIONS(243),
    [anon_sym_LT_EQ] = ACTIONS(354),
    [anon_sym_GT_EQ] = ACTIONS(354),
    [anon_sym_LT_LT] = ACTIONS(243),
    [anon_sym_GT_GT] = ACTIONS(243),
    [anon_sym_PLUS] = ACTIONS(243),
    [anon_sym_DASH] = ACTIONS(243),
    [anon_sym_STAR] = ACTIONS(243),
    [anon_sym_SLASH] = ACTIONS(243),
    [anon_sym_PERCENT] = ACTIONS(243),
    [anon_sym_STAR_STAR] = ACTIONS(243),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [anon_sym_BANG] = ACTIONS(1095),
    [anon_sym_RBRACK] = ACTIONS(354),
    [anon_sym_EQ_TILDE] = ACTIONS(243),
    [anon_sym_QMARK] = ACTIONS(243),
    [anon_sym_PLUS_PLUS2] = ACTIONS(1099),
    [anon_sym_DASH_DASH2] = ACTIONS(1099),
    [anon_sym_DASH2] = ACTIONS(1101),
    [anon_sym_PLUS2] = ACTIONS(1101),
    [anon_sym_TILDE] = ACTIONS(1103),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(352),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(363),
    [aux_sym_concatenation_token1] = ACTIONS(1105),
    [anon_sym_DOLLAR] = ACTIONS(367),
    [sym__special_character] = ACTIONS(1107),
    [anon_sym_DQUOTE] = ACTIONS(371),
    [sym_raw_string] = ACTIONS(1109),
    [sym_ansi_c_string] = ACTIONS(1109),
    [aux_sym_number_token1] = ACTIONS(375),
    [aux_sym_number_token2] = ACTIONS(377),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(379),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(381),
    [anon_sym_BQUOTE] = ACTIONS(383),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(385),
    [anon_sym_LT_LPAREN] = ACTIONS(387),
    [anon_sym_GT_LPAREN] = ACTIONS(387),
    [sym_comment] = ACTIONS(71),
    [sym__concat] = ACTIONS(354),
    [sym_test_operator] = ACTIONS(1113),
    [sym__brace_start] = ACTIONS(391),
  },
  [457] = {
    [sym__expression] = STATE(2933),
    [sym_binary_expression] = STATE(2863),
    [sym_ternary_expression] = STATE(2863),
    [sym_unary_expression] = STATE(2863),
    [sym_postfix_expression] = STATE(2863),
    [sym_parenthesized_expression] = STATE(2863),
    [sym_arithmetic_expansion] = STATE(2451),
    [sym_brace_expression] = STATE(2451),
    [sym_concatenation] = STATE(2863),
    [sym_string] = STATE(2451),
    [sym_translated_string] = STATE(2451),
    [sym_number] = STATE(2451),
    [sym_simple_expansion] = STATE(2451),
    [sym_expansion] = STATE(2451),
    [sym_command_substitution] = STATE(2451),
    [sym_process_substitution] = STATE(2451),
    [aux_sym__literal_repeat1] = STATE(2472),
    [aux_sym_concatenation_repeat1] = STATE(2540),
    [sym_word] = ACTIONS(1091),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(352),
    [anon_sym_EQ] = ACTIONS(243),
    [anon_sym_PLUS_PLUS] = ACTIONS(243),
    [anon_sym_DASH_DASH] = ACTIONS(243),
    [anon_sym_PLUS_EQ] = ACTIONS(243),
    [anon_sym_DASH_EQ] = ACTIONS(243),
    [anon_sym_STAR_EQ] = ACTIONS(243),
    [anon_sym_SLASH_EQ] = ACTIONS(243),
    [anon_sym_PERCENT_EQ] = ACTIONS(243),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(243),
    [anon_sym_LT_LT_EQ] = ACTIONS(354),
    [anon_sym_GT_GT_EQ] = ACTIONS(354),
    [anon_sym_AMP_EQ] = ACTIONS(354),
    [anon_sym_CARET_EQ] = ACTIONS(243),
    [anon_sym_PIPE_EQ] = ACTIONS(354),
    [anon_sym_PIPE_PIPE] = ACTIONS(354),
    [anon_sym_AMP_AMP] = ACTIONS(354),
    [anon_sym_PIPE] = ACTIONS(243),
    [anon_sym_CARET] = ACTIONS(243),
    [anon_sym_AMP] = ACTIONS(243),
    [anon_sym_EQ_EQ] = ACTIONS(243),
    [anon_sym_BANG_EQ] = ACTIONS(243),
    [anon_sym_LT] = ACTIONS(243),
    [anon_sym_GT] = ACTIONS(243),
    [anon_sym_LT_EQ] = ACTIONS(354),
    [anon_sym_GT_EQ] = ACTIONS(354),
    [anon_sym_LT_LT] = ACTIONS(243),
    [anon_sym_GT_GT] = ACTIONS(243),
    [anon_sym_PLUS] = ACTIONS(243),
    [anon_sym_DASH] = ACTIONS(243),
    [anon_sym_STAR] = ACTIONS(243),
    [anon_sym_SLASH] = ACTIONS(243),
    [anon_sym_PERCENT] = ACTIONS(243),
    [anon_sym_STAR_STAR] = ACTIONS(243),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [anon_sym_BANG] = ACTIONS(1095),
    [anon_sym_RBRACK] = ACTIONS(1123),
    [anon_sym_EQ_TILDE] = ACTIONS(243),
    [anon_sym_QMARK] = ACTIONS(243),
    [anon_sym_PLUS_PLUS2] = ACTIONS(1099),
    [anon_sym_DASH_DASH2] = ACTIONS(1099),
    [anon_sym_DASH2] = ACTIONS(1101),
    [anon_sym_PLUS2] = ACTIONS(1101),
    [anon_sym_TILDE] = ACTIONS(1103),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(352),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(363),
    [aux_sym_concatenation_token1] = ACTIONS(1105),
    [anon_sym_DOLLAR] = ACTIONS(367),
    [sym__special_character] = ACTIONS(1107),
    [anon_sym_DQUOTE] = ACTIONS(371),
    [sym_raw_string] = ACTIONS(1109),
    [sym_ansi_c_string] = ACTIONS(1109),
    [aux_sym_number_token1] = ACTIONS(375),
    [aux_sym_number_token2] = ACTIONS(377),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(379),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(381),
    [anon_sym_BQUOTE] = ACTIONS(383),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(385),
    [anon_sym_LT_LPAREN] = ACTIONS(387),
    [anon_sym_GT_LPAREN] = ACTIONS(387),
    [sym_comment] = ACTIONS(71),
    [sym__concat] = ACTIONS(1125),
    [sym_test_operator] = ACTIONS(1113),
    [sym__brace_start] = ACTIONS(391),
  },
  [458] = {
    [sym__expression] = STATE(3089),
    [sym_binary_expression] = STATE(2594),
    [sym_ternary_expression] = STATE(2594),
    [sym_unary_expression] = STATE(2594),
    [sym_postfix_expression] = STATE(2594),
    [sym_parenthesized_expression] = STATE(2594),
    [sym_arithmetic_expansion] = STATE(2573),
    [sym_brace_expression] = STATE(2573),
    [sym_concatenation] = STATE(2594),
    [sym_string] = STATE(2573),
    [sym_translated_string] = STATE(2573),
    [sym_number] = STATE(2573),
    [sym_simple_expansion] = STATE(2573),
    [sym_expansion] = STATE(2573),
    [sym_command_substitution] = STATE(2573),
    [sym_process_substitution] = STATE(2573),
    [aux_sym__literal_repeat1] = STATE(2484),
    [aux_sym_concatenation_repeat1] = STATE(2459),
    [sym_word] = ACTIONS(1127),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(1129),
    [anon_sym_EQ] = ACTIONS(243),
    [anon_sym_PLUS_PLUS] = ACTIONS(243),
    [anon_sym_DASH_DASH] = ACTIONS(243),
    [anon_sym_PLUS_EQ] = ACTIONS(243),
    [anon_sym_DASH_EQ] = ACTIONS(243),
    [anon_sym_STAR_EQ] = ACTIONS(243),
    [anon_sym_SLASH_EQ] = ACTIONS(243),
    [anon_sym_PERCENT_EQ] = ACTIONS(243),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(243),
    [anon_sym_LT_LT_EQ] = ACTIONS(354),
    [anon_sym_GT_GT_EQ] = ACTIONS(354),
    [anon_sym_AMP_EQ] = ACTIONS(354),
    [anon_sym_CARET_EQ] = ACTIONS(243),
    [anon_sym_PIPE_EQ] = ACTIONS(354),
    [anon_sym_PIPE_PIPE] = ACTIONS(354),
    [anon_sym_AMP_AMP] = ACTIONS(354),
    [anon_sym_PIPE] = ACTIONS(243),
    [anon_sym_CARET] = ACTIONS(243),
    [anon_sym_AMP] = ACTIONS(243),
    [anon_sym_EQ_EQ] = ACTIONS(243),
    [anon_sym_BANG_EQ] = ACTIONS(243),
    [anon_sym_LT] = ACTIONS(243),
    [anon_sym_GT] = ACTIONS(243),
    [anon_sym_LT_EQ] = ACTIONS(354),
    [anon_sym_GT_EQ] = ACTIONS(354),
    [anon_sym_LT_LT] = ACTIONS(243),
    [anon_sym_GT_GT] = ACTIONS(243),
    [anon_sym_PLUS] = ACTIONS(243),
    [anon_sym_DASH] = ACTIONS(243),
    [anon_sym_STAR] = ACTIONS(243),
    [anon_sym_SLASH] = ACTIONS(243),
    [anon_sym_PERCENT] = ACTIONS(243),
    [anon_sym_STAR_STAR] = ACTIONS(243),
    [anon_sym_LPAREN] = ACTIONS(1131),
    [anon_sym_BANG] = ACTIONS(1133),
    [anon_sym_EQ_TILDE] = ACTIONS(243),
    [anon_sym_QMARK] = ACTIONS(243),
    [anon_sym_COLON] = ACTIONS(243),
    [anon_sym_PLUS_PLUS2] = ACTIONS(1135),
    [anon_sym_DASH_DASH2] = ACTIONS(1135),
    [anon_sym_DASH2] = ACTIONS(1137),
    [anon_sym_PLUS2] = ACTIONS(1137),
    [anon_sym_TILDE] = ACTIONS(1139),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(1129),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(1141),
    [aux_sym_concatenation_token1] = ACTIONS(1143),
    [anon_sym_DOLLAR] = ACTIONS(1145),
    [sym__special_character] = ACTIONS(1147),
    [anon_sym_DQUOTE] = ACTIONS(1149),
    [sym_raw_string] = ACTIONS(1151),
    [sym_ansi_c_string] = ACTIONS(1151),
    [aux_sym_number_token1] = ACTIONS(1153),
    [aux_sym_number_token2] = ACTIONS(1155),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(1157),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1159),
    [anon_sym_BQUOTE] = ACTIONS(1161),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(1163),
    [anon_sym_LT_LPAREN] = ACTIONS(1165),
    [anon_sym_GT_LPAREN] = ACTIONS(1165),
    [sym_comment] = ACTIONS(71),
    [sym__concat] = ACTIONS(1143),
    [sym_test_operator] = ACTIONS(1167),
    [sym__brace_start] = ACTIONS(1169),
  },
  [459] = {
    [sym__expression] = STATE(3112),
    [sym_binary_expression] = STATE(2594),
    [sym_ternary_expression] = STATE(2594),
    [sym_unary_expression] = STATE(2594),
    [sym_postfix_expression] = STATE(2594),
    [sym_parenthesized_expression] = STATE(2594),
    [sym_arithmetic_expansion] = STATE(2456),
    [sym_brace_expression] = STATE(2456),
    [sym_concatenation] = STATE(2594),
    [sym_string] = STATE(2456),
    [sym_translated_string] = STATE(2456),
    [sym_number] = STATE(2456),
    [sym_simple_expansion] = STATE(2456),
    [sym_expansion] = STATE(2456),
    [sym_command_substitution] = STATE(2456),
    [sym_process_substitution] = STATE(2456),
    [aux_sym__literal_repeat1] = STATE(2484),
    [aux_sym_concatenation_repeat1] = STATE(2495),
    [sym_word] = ACTIONS(1171),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(1129),
    [anon_sym_RPAREN_RPAREN] = ACTIONS(354),
    [anon_sym_EQ] = ACTIONS(243),
    [anon_sym_PLUS_PLUS] = ACTIONS(243),
    [anon_sym_DASH_DASH] = ACTIONS(243),
    [anon_sym_PLUS_EQ] = ACTIONS(243),
    [anon_sym_DASH_EQ] = ACTIONS(243),
    [anon_sym_STAR_EQ] = ACTIONS(243),
    [anon_sym_SLASH_EQ] = ACTIONS(243),
    [anon_sym_PERCENT_EQ] = ACTIONS(243),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(243),
    [anon_sym_LT_LT_EQ] = ACTIONS(354),
    [anon_sym_GT_GT_EQ] = ACTIONS(354),
    [anon_sym_AMP_EQ] = ACTIONS(354),
    [anon_sym_CARET_EQ] = ACTIONS(243),
    [anon_sym_PIPE_EQ] = ACTIONS(354),
    [anon_sym_PIPE_PIPE] = ACTIONS(354),
    [anon_sym_AMP_AMP] = ACTIONS(354),
    [anon_sym_PIPE] = ACTIONS(243),
    [anon_sym_CARET] = ACTIONS(243),
    [anon_sym_AMP] = ACTIONS(243),
    [anon_sym_EQ_EQ] = ACTIONS(243),
    [anon_sym_BANG_EQ] = ACTIONS(243),
    [anon_sym_LT] = ACTIONS(243),
    [anon_sym_GT] = ACTIONS(243),
    [anon_sym_LT_EQ] = ACTIONS(354),
    [anon_sym_GT_EQ] = ACTIONS(354),
    [anon_sym_LT_LT] = ACTIONS(243),
    [anon_sym_GT_GT] = ACTIONS(243),
    [anon_sym_PLUS] = ACTIONS(243),
    [anon_sym_DASH] = ACTIONS(243),
    [anon_sym_STAR] = ACTIONS(243),
    [anon_sym_SLASH] = ACTIONS(243),
    [anon_sym_PERCENT] = ACTIONS(243),
    [anon_sym_STAR_STAR] = ACTIONS(243),
    [anon_sym_LPAREN] = ACTIONS(1131),
    [anon_sym_BANG] = ACTIONS(1173),
    [anon_sym_EQ_TILDE] = ACTIONS(243),
    [anon_sym_QMARK] = ACTIONS(243),
    [anon_sym_PLUS_PLUS2] = ACTIONS(1175),
    [anon_sym_DASH_DASH2] = ACTIONS(1175),
    [anon_sym_DASH2] = ACTIONS(1177),
    [anon_sym_PLUS2] = ACTIONS(1177),
    [anon_sym_TILDE] = ACTIONS(1179),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(1129),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(1141),
    [aux_sym_concatenation_token1] = ACTIONS(1143),
    [anon_sym_DOLLAR] = ACTIONS(1145),
    [sym__special_character] = ACTIONS(1181),
    [anon_sym_DQUOTE] = ACTIONS(1149),
    [sym_raw_string] = ACTIONS(1183),
    [sym_ansi_c_string] = ACTIONS(1183),
    [aux_sym_number_token1] = ACTIONS(1153),
    [aux_sym_number_token2] = ACTIONS(1155),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(1157),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1159),
    [anon_sym_BQUOTE] = ACTIONS(1161),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(1163),
    [anon_sym_LT_LPAREN] = ACTIONS(1165),
    [anon_sym_GT_LPAREN] = ACTIONS(1165),
    [sym_comment] = ACTIONS(71),
    [sym__concat] = ACTIONS(1143),
    [sym_test_operator] = ACTIONS(1185),
    [sym__brace_start] = ACTIONS(1169),
  },
  [460] = {
    [sym__expression] = STATE(3041),
    [sym_binary_expression] = STATE(2594),
    [sym_ternary_expression] = STATE(2594),
    [sym_unary_expression] = STATE(2594),
    [sym_postfix_expression] = STATE(2594),
    [sym_parenthesized_expression] = STATE(2594),
    [sym_arithmetic_expansion] = STATE(2503),
    [sym_brace_expression] = STATE(2503),
    [sym_concatenation] = STATE(2594),
    [sym_string] = STATE(2503),
    [sym_translated_string] = STATE(2503),
    [sym_number] = STATE(2503),
    [sym_simple_expansion] = STATE(2503),
    [sym_expansion] = STATE(2503),
    [sym_command_substitution] = STATE(2503),
    [sym_process_substitution] = STATE(2503),
    [aux_sym__literal_repeat1] = STATE(2717),
    [aux_sym_concatenation_repeat1] = STATE(2466),
    [sym_word] = ACTIONS(1187),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(1129),
    [anon_sym_EQ] = ACTIONS(243),
    [anon_sym_PLUS_PLUS] = ACTIONS(243),
    [anon_sym_DASH_DASH] = ACTIONS(243),
    [anon_sym_PLUS_EQ] = ACTIONS(243),
    [anon_sym_DASH_EQ] = ACTIONS(243),
    [anon_sym_STAR_EQ] = ACTIONS(243),
    [anon_sym_SLASH_EQ] = ACTIONS(243),
    [anon_sym_PERCENT_EQ] = ACTIONS(243),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(243),
    [anon_sym_LT_LT_EQ] = ACTIONS(354),
    [anon_sym_GT_GT_EQ] = ACTIONS(354),
    [anon_sym_AMP_EQ] = ACTIONS(354),
    [anon_sym_CARET_EQ] = ACTIONS(243),
    [anon_sym_PIPE_EQ] = ACTIONS(354),
    [anon_sym_PIPE_PIPE] = ACTIONS(354),
    [anon_sym_AMP_AMP] = ACTIONS(354),
    [anon_sym_PIPE] = ACTIONS(243),
    [anon_sym_CARET] = ACTIONS(243),
    [anon_sym_AMP] = ACTIONS(243),
    [anon_sym_EQ_EQ] = ACTIONS(243),
    [anon_sym_BANG_EQ] = ACTIONS(243),
    [anon_sym_LT] = ACTIONS(243),
    [anon_sym_GT] = ACTIONS(243),
    [anon_sym_LT_EQ] = ACTIONS(354),
    [anon_sym_GT_EQ] = ACTIONS(354),
    [anon_sym_LT_LT] = ACTIONS(243),
    [anon_sym_GT_GT] = ACTIONS(243),
    [anon_sym_PLUS] = ACTIONS(243),
    [anon_sym_DASH] = ACTIONS(243),
    [anon_sym_STAR] = ACTIONS(243),
    [anon_sym_SLASH] = ACTIONS(243),
    [anon_sym_PERCENT] = ACTIONS(243),
    [anon_sym_STAR_STAR] = ACTIONS(243),
    [anon_sym_LPAREN] = ACTIONS(1131),
    [anon_sym_BANG] = ACTIONS(1189),
    [anon_sym_RBRACK_RBRACK] = ACTIONS(354),
    [anon_sym_EQ_TILDE] = ACTIONS(243),
    [anon_sym_QMARK] = ACTIONS(243),
    [anon_sym_PLUS_PLUS2] = ACTIONS(1191),
    [anon_sym_DASH_DASH2] = ACTIONS(1191),
    [anon_sym_DASH2] = ACTIONS(1193),
    [anon_sym_PLUS2] = ACTIONS(1193),
    [anon_sym_TILDE] = ACTIONS(1195),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(1129),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(1141),
    [aux_sym_concatenation_token1] = ACTIONS(1143),
    [anon_sym_DOLLAR] = ACTIONS(1145),
    [sym__special_character] = ACTIONS(1197),
    [anon_sym_DQUOTE] = ACTIONS(1149),
    [sym_raw_string] = ACTIONS(1199),
    [sym_ansi_c_string] = ACTIONS(1199),
    [aux_sym_number_token1] = ACTIONS(1153),
    [aux_sym_number_token2] = ACTIONS(1155),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(1157),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1159),
    [anon_sym_BQUOTE] = ACTIONS(1161),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(1163),
    [anon_sym_LT_LPAREN] = ACTIONS(1165),
    [anon_sym_GT_LPAREN] = ACTIONS(1165),
    [sym_comment] = ACTIONS(71),
    [sym__concat] = ACTIONS(1143),
    [sym_test_operator] = ACTIONS(1201),
    [sym__brace_start] = ACTIONS(1169),
  },
  [461] = {
    [sym__expression] = STATE(2933),
    [sym_binary_expression] = STATE(2863),
    [sym_ternary_expression] = STATE(2863),
    [sym_unary_expression] = STATE(2863),
    [sym_postfix_expression] = STATE(2863),
    [sym_parenthesized_expression] = STATE(2863),
    [sym_arithmetic_expansion] = STATE(2451),
    [sym_brace_expression] = STATE(2451),
    [sym_concatenation] = STATE(2863),
    [sym_string] = STATE(2451),
    [sym_translated_string] = STATE(2451),
    [sym_number] = STATE(2451),
    [sym_simple_expansion] = STATE(2451),
    [sym_expansion] = STATE(2451),
    [sym_command_substitution] = STATE(2451),
    [sym_process_substitution] = STATE(2451),
    [aux_sym__literal_repeat1] = STATE(2472),
    [aux_sym_concatenation_repeat1] = STATE(2540),
    [sym_word] = ACTIONS(1091),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(352),
    [anon_sym_EQ] = ACTIONS(243),
    [anon_sym_PLUS_PLUS] = ACTIONS(243),
    [anon_sym_DASH_DASH] = ACTIONS(243),
    [anon_sym_PLUS_EQ] = ACTIONS(243),
    [anon_sym_DASH_EQ] = ACTIONS(243),
    [anon_sym_STAR_EQ] = ACTIONS(243),
    [anon_sym_SLASH_EQ] = ACTIONS(243),
    [anon_sym_PERCENT_EQ] = ACTIONS(243),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(243),
    [anon_sym_LT_LT_EQ] = ACTIONS(354),
    [anon_sym_GT_GT_EQ] = ACTIONS(354),
    [anon_sym_AMP_EQ] = ACTIONS(354),
    [anon_sym_CARET_EQ] = ACTIONS(243),
    [anon_sym_PIPE_EQ] = ACTIONS(354),
    [anon_sym_PIPE_PIPE] = ACTIONS(354),
    [anon_sym_AMP_AMP] = ACTIONS(354),
    [anon_sym_PIPE] = ACTIONS(243),
    [anon_sym_CARET] = ACTIONS(243),
    [anon_sym_AMP] = ACTIONS(243),
    [anon_sym_EQ_EQ] = ACTIONS(243),
    [anon_sym_BANG_EQ] = ACTIONS(243),
    [anon_sym_LT] = ACTIONS(243),
    [anon_sym_GT] = ACTIONS(243),
    [anon_sym_LT_EQ] = ACTIONS(354),
    [anon_sym_GT_EQ] = ACTIONS(354),
    [anon_sym_LT_LT] = ACTIONS(243),
    [anon_sym_GT_GT] = ACTIONS(243),
    [anon_sym_PLUS] = ACTIONS(243),
    [anon_sym_DASH] = ACTIONS(243),
    [anon_sym_STAR] = ACTIONS(243),
    [anon_sym_SLASH] = ACTIONS(243),
    [anon_sym_PERCENT] = ACTIONS(243),
    [anon_sym_STAR_STAR] = ACTIONS(243),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [anon_sym_BANG] = ACTIONS(1095),
    [anon_sym_RBRACK] = ACTIONS(1203),
    [anon_sym_EQ_TILDE] = ACTIONS(243),
    [anon_sym_QMARK] = ACTIONS(243),
    [anon_sym_PLUS_PLUS2] = ACTIONS(1099),
    [anon_sym_DASH_DASH2] = ACTIONS(1099),
    [anon_sym_DASH2] = ACTIONS(1101),
    [anon_sym_PLUS2] = ACTIONS(1101),
    [anon_sym_TILDE] = ACTIONS(1103),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(352),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(363),
    [aux_sym_concatenation_token1] = ACTIONS(1105),
    [anon_sym_DOLLAR] = ACTIONS(367),
    [sym__special_character] = ACTIONS(1107),
    [anon_sym_DQUOTE] = ACTIONS(371),
    [sym_raw_string] = ACTIONS(1109),
    [sym_ansi_c_string] = ACTIONS(1109),
    [aux_sym_number_token1] = ACTIONS(375),
    [aux_sym_number_token2] = ACTIONS(377),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(379),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(381),
    [anon_sym_BQUOTE] = ACTIONS(383),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(385),
    [anon_sym_LT_LPAREN] = ACTIONS(387),
    [anon_sym_GT_LPAREN] = ACTIONS(387),
    [sym_comment] = ACTIONS(71),
    [sym__concat] = ACTIONS(1205),
    [sym_test_operator] = ACTIONS(1113),
    [sym__brace_start] = ACTIONS(391),
  },
  [462] = {
    [sym__expression] = STATE(2933),
    [sym_binary_expression] = STATE(2863),
    [sym_ternary_expression] = STATE(2863),
    [sym_unary_expression] = STATE(2863),
    [sym_postfix_expression] = STATE(2863),
    [sym_parenthesized_expression] = STATE(2863),
    [sym_arithmetic_expansion] = STATE(2451),
    [sym_brace_expression] = STATE(2451),
    [sym_concatenation] = STATE(2863),
    [sym_string] = STATE(2451),
    [sym_translated_string] = STATE(2451),
    [sym_number] = STATE(2451),
    [sym_simple_expansion] = STATE(2451),
    [sym_expansion] = STATE(2451),
    [sym_command_substitution] = STATE(2451),
    [sym_process_substitution] = STATE(2451),
    [aux_sym__literal_repeat1] = STATE(2472),
    [aux_sym_concatenation_repeat1] = STATE(2540),
    [sym_word] = ACTIONS(1091),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(352),
    [anon_sym_EQ] = ACTIONS(243),
    [anon_sym_PLUS_PLUS] = ACTIONS(243),
    [anon_sym_DASH_DASH] = ACTIONS(243),
    [anon_sym_PLUS_EQ] = ACTIONS(243),
    [anon_sym_DASH_EQ] = ACTIONS(243),
    [anon_sym_STAR_EQ] = ACTIONS(243),
    [anon_sym_SLASH_EQ] = ACTIONS(243),
    [anon_sym_PERCENT_EQ] = ACTIONS(243),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(243),
    [anon_sym_LT_LT_EQ] = ACTIONS(354),
    [anon_sym_GT_GT_EQ] = ACTIONS(354),
    [anon_sym_AMP_EQ] = ACTIONS(354),
    [anon_sym_CARET_EQ] = ACTIONS(243),
    [anon_sym_PIPE_EQ] = ACTIONS(354),
    [anon_sym_PIPE_PIPE] = ACTIONS(354),
    [anon_sym_AMP_AMP] = ACTIONS(354),
    [anon_sym_PIPE] = ACTIONS(243),
    [anon_sym_CARET] = ACTIONS(243),
    [anon_sym_AMP] = ACTIONS(243),
    [anon_sym_EQ_EQ] = ACTIONS(243),
    [anon_sym_BANG_EQ] = ACTIONS(243),
    [anon_sym_LT] = ACTIONS(243),
    [anon_sym_GT] = ACTIONS(243),
    [anon_sym_LT_EQ] = ACTIONS(354),
    [anon_sym_GT_EQ] = ACTIONS(354),
    [anon_sym_LT_LT] = ACTIONS(243),
    [anon_sym_GT_GT] = ACTIONS(243),
    [anon_sym_PLUS] = ACTIONS(243),
    [anon_sym_DASH] = ACTIONS(243),
    [anon_sym_STAR] = ACTIONS(243),
    [anon_sym_SLASH] = ACTIONS(243),
    [anon_sym_PERCENT] = ACTIONS(243),
    [anon_sym_STAR_STAR] = ACTIONS(243),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [anon_sym_BANG] = ACTIONS(1095),
    [anon_sym_RBRACK] = ACTIONS(1207),
    [anon_sym_EQ_TILDE] = ACTIONS(243),
    [anon_sym_QMARK] = ACTIONS(243),
    [anon_sym_PLUS_PLUS2] = ACTIONS(1099),
    [anon_sym_DASH_DASH2] = ACTIONS(1099),
    [anon_sym_DASH2] = ACTIONS(1101),
    [anon_sym_PLUS2] = ACTIONS(1101),
    [anon_sym_TILDE] = ACTIONS(1103),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(352),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(363),
    [aux_sym_concatenation_token1] = ACTIONS(1105),
    [anon_sym_DOLLAR] = ACTIONS(367),
    [sym__special_character] = ACTIONS(1107),
    [anon_sym_DQUOTE] = ACTIONS(371),
    [sym_raw_string] = ACTIONS(1109),
    [sym_ansi_c_string] = ACTIONS(1109),
    [aux_sym_number_token1] = ACTIONS(375),
    [aux_sym_number_token2] = ACTIONS(377),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(379),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(381),
    [anon_sym_BQUOTE] = ACTIONS(383),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(385),
    [anon_sym_LT_LPAREN] = ACTIONS(387),
    [anon_sym_GT_LPAREN] = ACTIONS(387),
    [sym_comment] = ACTIONS(71),
    [sym__concat] = ACTIONS(1209),
    [sym_test_operator] = ACTIONS(1113),
    [sym__brace_start] = ACTIONS(391),
  },
  [463] = {
    [sym__expression] = STATE(3197),
    [sym_binary_expression] = STATE(2594),
    [sym_ternary_expression] = STATE(2594),
    [sym_unary_expression] = STATE(2594),
    [sym_postfix_expression] = STATE(2594),
    [sym_parenthesized_expression] = STATE(2594),
    [sym_arithmetic_expansion] = STATE(2588),
    [sym_brace_expression] = STATE(2588),
    [sym_concatenation] = STATE(2594),
    [sym_string] = STATE(2588),
    [sym_translated_string] = STATE(2588),
    [sym_number] = STATE(2588),
    [sym_simple_expansion] = STATE(2588),
    [sym_expansion] = STATE(2588),
    [sym_command_substitution] = STATE(2588),
    [sym_process_substitution] = STATE(2588),
    [aux_sym__literal_repeat1] = STATE(2484),
    [aux_sym_concatenation_repeat1] = STATE(2723),
    [sym_word] = ACTIONS(1211),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(1129),
    [anon_sym_EQ] = ACTIONS(243),
    [anon_sym_PLUS_PLUS] = ACTIONS(243),
    [anon_sym_DASH_DASH] = ACTIONS(243),
    [anon_sym_PLUS_EQ] = ACTIONS(243),
    [anon_sym_DASH_EQ] = ACTIONS(243),
    [anon_sym_STAR_EQ] = ACTIONS(243),
    [anon_sym_SLASH_EQ] = ACTIONS(243),
    [anon_sym_PERCENT_EQ] = ACTIONS(243),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(243),
    [anon_sym_LT_LT_EQ] = ACTIONS(354),
    [anon_sym_GT_GT_EQ] = ACTIONS(354),
    [anon_sym_AMP_EQ] = ACTIONS(354),
    [anon_sym_CARET_EQ] = ACTIONS(243),
    [anon_sym_PIPE_EQ] = ACTIONS(354),
    [anon_sym_PIPE_PIPE] = ACTIONS(354),
    [anon_sym_AMP_AMP] = ACTIONS(354),
    [anon_sym_PIPE] = ACTIONS(243),
    [anon_sym_CARET] = ACTIONS(243),
    [anon_sym_AMP] = ACTIONS(243),
    [anon_sym_EQ_EQ] = ACTIONS(243),
    [anon_sym_BANG_EQ] = ACTIONS(243),
    [anon_sym_LT] = ACTIONS(243),
    [anon_sym_GT] = ACTIONS(243),
    [anon_sym_LT_EQ] = ACTIONS(354),
    [anon_sym_GT_EQ] = ACTIONS(354),
    [anon_sym_LT_LT] = ACTIONS(243),
    [anon_sym_GT_GT] = ACTIONS(243),
    [anon_sym_PLUS] = ACTIONS(243),
    [anon_sym_DASH] = ACTIONS(243),
    [anon_sym_STAR] = ACTIONS(243),
    [anon_sym_SLASH] = ACTIONS(243),
    [anon_sym_PERCENT] = ACTIONS(243),
    [anon_sym_STAR_STAR] = ACTIONS(243),
    [anon_sym_LPAREN] = ACTIONS(1131),
    [anon_sym_BANG] = ACTIONS(1213),
    [anon_sym_EQ_TILDE] = ACTIONS(243),
    [anon_sym_QMARK] = ACTIONS(243),
    [anon_sym_PLUS_PLUS2] = ACTIONS(1215),
    [anon_sym_DASH_DASH2] = ACTIONS(1215),
    [anon_sym_DASH2] = ACTIONS(1217),
    [anon_sym_PLUS2] = ACTIONS(1217),
    [anon_sym_TILDE] = ACTIONS(1219),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(1129),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(1141),
    [aux_sym_concatenation_token1] = ACTIONS(1143),
    [anon_sym_DOLLAR] = ACTIONS(1145),
    [sym__special_character] = ACTIONS(1221),
    [anon_sym_DQUOTE] = ACTIONS(1149),
    [sym_raw_string] = ACTIONS(1223),
    [sym_ansi_c_string] = ACTIONS(1223),
    [aux_sym_number_token1] = ACTIONS(1153),
    [aux_sym_number_token2] = ACTIONS(1155),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(1157),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1159),
    [anon_sym_BQUOTE] = ACTIONS(1161),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(1163),
    [anon_sym_LT_LPAREN] = ACTIONS(1165),
    [anon_sym_GT_LPAREN] = ACTIONS(1165),
    [sym_comment] = ACTIONS(71),
    [sym__concat] = ACTIONS(1143),
    [sym_test_operator] = ACTIONS(1225),
    [sym__brace_start] = ACTIONS(1169),
  },
  [464] = {
    [sym_string] = STATE(489),
    [sym_word] = ACTIONS(1227),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(1227),
    [anon_sym_SEMI] = ACTIONS(1227),
    [anon_sym_EQ] = ACTIONS(1227),
    [anon_sym_PLUS_PLUS] = ACTIONS(1227),
    [anon_sym_DASH_DASH] = ACTIONS(1227),
    [anon_sym_PLUS_EQ] = ACTIONS(1227),
    [anon_sym_DASH_EQ] = ACTIONS(1227),
    [anon_sym_STAR_EQ] = ACTIONS(1227),
    [anon_sym_SLASH_EQ] = ACTIONS(1227),
    [anon_sym_PERCENT_EQ] = ACTIONS(1227),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(1227),
    [anon_sym_LT_LT_EQ] = ACTIONS(1227),
    [anon_sym_GT_GT_EQ] = ACTIONS(1227),
    [anon_sym_AMP_EQ] = ACTIONS(1227),
    [anon_sym_CARET_EQ] = ACTIONS(1227),
    [anon_sym_PIPE_EQ] = ACTIONS(1227),
    [anon_sym_PIPE_PIPE] = ACTIONS(1227),
    [anon_sym_AMP_AMP] = ACTIONS(1227),
    [anon_sym_PIPE] = ACTIONS(1227),
    [anon_sym_CARET] = ACTIONS(1227),
    [anon_sym_AMP] = ACTIONS(1227),
    [anon_sym_EQ_EQ] = ACTIONS(1227),
    [anon_sym_BANG_EQ] = ACTIONS(1227),
    [anon_sym_LT] = ACTIONS(1227),
    [anon_sym_GT] = ACTIONS(1227),
    [anon_sym_LT_EQ] = ACTIONS(1227),
    [anon_sym_GT_EQ] = ACTIONS(1227),
    [anon_sym_LT_LT] = ACTIONS(1227),
    [anon_sym_GT_GT] = ACTIONS(1227),
    [anon_sym_PLUS] = ACTIONS(1227),
    [anon_sym_DASH] = ACTIONS(1229),
    [anon_sym_STAR] = ACTIONS(1229),
    [anon_sym_SLASH] = ACTIONS(1227),
    [anon_sym_PERCENT] = ACTIONS(1227),
    [anon_sym_STAR_STAR] = ACTIONS(1227),
    [anon_sym_LPAREN] = ACTIONS(1227),
    [anon_sym_RPAREN] = ACTIONS(1227),
    [anon_sym_SEMI_SEMI] = ACTIONS(1227),
    [anon_sym_PIPE_AMP] = ACTIONS(1227),
    [anon_sym_BANG] = ACTIONS(1229),
    [anon_sym_EQ_TILDE] = ACTIONS(1227),
    [anon_sym_AMP_GT] = ACTIONS(1227),
    [anon_sym_AMP_GT_GT] = ACTIONS(1227),
    [anon_sym_LT_AMP] = ACTIONS(1227),
    [anon_sym_GT_AMP] = ACTIONS(1227),
    [anon_sym_GT_PIPE] = ACTIONS(1227),
    [anon_sym_LT_AMP_DASH] = ACTIONS(1227),
    [anon_sym_GT_AMP_DASH] = ACTIONS(1227),
    [anon_sym_LT_LT_DASH] = ACTIONS(1227),
    [aux_sym_heredoc_redirect_token1] = ACTIONS(1227),
    [anon_sym_LT_LT_LT] = ACTIONS(1227),
    [anon_sym_QMARK] = ACTIONS(1229),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(1227),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(1227),
    [anon_sym_DOLLAR] = ACTIONS(1229),
    [sym__special_character] = ACTIONS(1227),
    [anon_sym_DQUOTE] = ACTIONS(1231),
    [sym_raw_string] = ACTIONS(1227),
    [sym_ansi_c_string] = ACTIONS(1227),
    [aux_sym_number_token1] = ACTIONS(1227),
    [aux_sym_number_token2] = ACTIONS(1227),
    [anon_sym_POUND] = ACTIONS(1229),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(1227),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1227),
    [anon_sym_BQUOTE] = ACTIONS(1227),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(1227),
    [anon_sym_LT_LPAREN] = ACTIONS(1227),
    [anon_sym_GT_LPAREN] = ACTIONS(1227),
    [sym_comment] = ACTIONS(3),
    [aux_sym__simple_variable_name_token1] = ACTIONS(1233),
    [aux_sym__multiline_variable_name_token1] = ACTIONS(1233),
    [anon_sym_AT2] = ACTIONS(1229),
    [anon_sym_0] = ACTIONS(1229),
    [anon_sym__] = ACTIONS(1229),
    [sym_file_descriptor] = ACTIONS(1235),
    [sym_variable_name] = ACTIONS(1237),
    [sym_test_operator] = ACTIONS(1235),
    [sym__bare_dollar] = ACTIONS(1235),
    [sym__brace_start] = ACTIONS(1235),
  },
  [465] = {
    [sym_string] = STATE(489),
    [sym_word] = ACTIONS(1239),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(1239),
    [anon_sym_SEMI] = ACTIONS(1239),
    [anon_sym_EQ] = ACTIONS(1239),
    [anon_sym_PLUS_PLUS] = ACTIONS(1239),
    [anon_sym_DASH_DASH] = ACTIONS(1239),
    [anon_sym_PLUS_EQ] = ACTIONS(1239),
    [anon_sym_DASH_EQ] = ACTIONS(1239),
    [anon_sym_STAR_EQ] = ACTIONS(1239),
    [anon_sym_SLASH_EQ] = ACTIONS(1239),
    [anon_sym_PERCENT_EQ] = ACTIONS(1239),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(1239),
    [anon_sym_LT_LT_EQ] = ACTIONS(1239),
    [anon_sym_GT_GT_EQ] = ACTIONS(1239),
    [anon_sym_AMP_EQ] = ACTIONS(1239),
    [anon_sym_CARET_EQ] = ACTIONS(1239),
    [anon_sym_PIPE_EQ] = ACTIONS(1239),
    [anon_sym_PIPE_PIPE] = ACTIONS(1239),
    [anon_sym_AMP_AMP] = ACTIONS(1239),
    [anon_sym_PIPE] = ACTIONS(1239),
    [anon_sym_CARET] = ACTIONS(1239),
    [anon_sym_AMP] = ACTIONS(1239),
    [anon_sym_EQ_EQ] = ACTIONS(1239),
    [anon_sym_BANG_EQ] = ACTIONS(1239),
    [anon_sym_LT] = ACTIONS(1239),
    [anon_sym_GT] = ACTIONS(1239),
    [anon_sym_LT_EQ] = ACTIONS(1239),
    [anon_sym_GT_EQ] = ACTIONS(1239),
    [anon_sym_LT_LT] = ACTIONS(1239),
    [anon_sym_GT_GT] = ACTIONS(1239),
    [anon_sym_PLUS] = ACTIONS(1239),
    [anon_sym_DASH] = ACTIONS(1229),
    [anon_sym_STAR] = ACTIONS(1229),
    [anon_sym_SLASH] = ACTIONS(1239),
    [anon_sym_PERCENT] = ACTIONS(1239),
    [anon_sym_STAR_STAR] = ACTIONS(1239),
    [anon_sym_LPAREN] = ACTIONS(1239),
    [anon_sym_RPAREN] = ACTIONS(1239),
    [anon_sym_SEMI_SEMI] = ACTIONS(1239),
    [anon_sym_PIPE_AMP] = ACTIONS(1239),
    [anon_sym_BANG] = ACTIONS(1229),
    [anon_sym_EQ_TILDE] = ACTIONS(1239),
    [anon_sym_AMP_GT] = ACTIONS(1239),
    [anon_sym_AMP_GT_GT] = ACTIONS(1239),
    [anon_sym_LT_AMP] = ACTIONS(1239),
    [anon_sym_GT_AMP] = ACTIONS(1239),
    [anon_sym_GT_PIPE] = ACTIONS(1239),
    [anon_sym_LT_AMP_DASH] = ACTIONS(1239),
    [anon_sym_GT_AMP_DASH] = ACTIONS(1239),
    [anon_sym_LT_LT_DASH] = ACTIONS(1239),
    [aux_sym_heredoc_redirect_token1] = ACTIONS(1239),
    [anon_sym_LT_LT_LT] = ACTIONS(1239),
    [anon_sym_QMARK] = ACTIONS(1229),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(1239),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(1239),
    [anon_sym_DOLLAR] = ACTIONS(1229),
    [sym__special_character] = ACTIONS(1239),
    [anon_sym_DQUOTE] = ACTIONS(1231),
    [sym_raw_string] = ACTIONS(1239),
    [sym_ansi_c_string] = ACTIONS(1239),
    [aux_sym_number_token1] = ACTIONS(1239),
    [aux_sym_number_token2] = ACTIONS(1239),
    [anon_sym_POUND] = ACTIONS(1229),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(1239),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1239),
    [anon_sym_BQUOTE] = ACTIONS(1239),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(1239),
    [anon_sym_LT_LPAREN] = ACTIONS(1239),
    [anon_sym_GT_LPAREN] = ACTIONS(1239),
    [sym_comment] = ACTIONS(3),
    [aux_sym__simple_variable_name_token1] = ACTIONS(1233),
    [aux_sym__multiline_variable_name_token1] = ACTIONS(1233),
    [anon_sym_AT2] = ACTIONS(1229),
    [anon_sym_0] = ACTIONS(1229),
    [anon_sym__] = ACTIONS(1229),
    [sym_file_descriptor] = ACTIONS(1241),
    [sym_variable_name] = ACTIONS(1237),
    [sym_test_operator] = ACTIONS(1241),
    [sym__bare_dollar] = ACTIONS(1241),
    [sym__brace_start] = ACTIONS(1241),
  },
  [466] = {
    [sym_string] = STATE(503),
    [sym_word] = ACTIONS(1227),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(1227),
    [anon_sym_EQ] = ACTIONS(1227),
    [anon_sym_PLUS_PLUS] = ACTIONS(1227),
    [anon_sym_DASH_DASH] = ACTIONS(1227),
    [anon_sym_PLUS_EQ] = ACTIONS(1227),
    [anon_sym_DASH_EQ] = ACTIONS(1227),
    [anon_sym_STAR_EQ] = ACTIONS(1227),
    [anon_sym_SLASH_EQ] = ACTIONS(1227),
    [anon_sym_PERCENT_EQ] = ACTIONS(1227),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(1227),
    [anon_sym_LT_LT_EQ] = ACTIONS(1227),
    [anon_sym_GT_GT_EQ] = ACTIONS(1227),
    [anon_sym_AMP_EQ] = ACTIONS(1227),
    [anon_sym_CARET_EQ] = ACTIONS(1227),
    [anon_sym_PIPE_EQ] = ACTIONS(1227),
    [anon_sym_PIPE_PIPE] = ACTIONS(1227),
    [anon_sym_AMP_AMP] = ACTIONS(1227),
    [anon_sym_PIPE] = ACTIONS(1227),
    [anon_sym_CARET] = ACTIONS(1227),
    [anon_sym_AMP] = ACTIONS(1227),
    [anon_sym_EQ_EQ] = ACTIONS(1227),
    [anon_sym_BANG_EQ] = ACTIONS(1227),
    [anon_sym_LT] = ACTIONS(1227),
    [anon_sym_GT] = ACTIONS(1227),
    [anon_sym_LT_EQ] = ACTIONS(1227),
    [anon_sym_GT_EQ] = ACTIONS(1227),
    [anon_sym_LT_LT] = ACTIONS(1227),
    [anon_sym_GT_GT] = ACTIONS(1227),
    [anon_sym_PLUS] = ACTIONS(1227),
    [anon_sym_DASH] = ACTIONS(1243),
    [anon_sym_STAR] = ACTIONS(1243),
    [anon_sym_SLASH] = ACTIONS(1227),
    [anon_sym_PERCENT] = ACTIONS(1227),
    [anon_sym_STAR_STAR] = ACTIONS(1227),
    [anon_sym_LPAREN] = ACTIONS(1227),
    [anon_sym_PIPE_AMP] = ACTIONS(1227),
    [anon_sym_BANG] = ACTIONS(1243),
    [anon_sym_RBRACK] = ACTIONS(1227),
    [anon_sym_EQ_TILDE] = ACTIONS(1227),
    [anon_sym_AMP_GT] = ACTIONS(1227),
    [anon_sym_AMP_GT_GT] = ACTIONS(1227),
    [anon_sym_LT_AMP] = ACTIONS(1227),
    [anon_sym_GT_AMP] = ACTIONS(1227),
    [anon_sym_GT_PIPE] = ACTIONS(1227),
    [anon_sym_LT_AMP_DASH] = ACTIONS(1227),
    [anon_sym_GT_AMP_DASH] = ACTIONS(1227),
    [anon_sym_LT_LT_DASH] = ACTIONS(1227),
    [anon_sym_LT_LT_LT] = ACTIONS(1227),
    [anon_sym_QMARK] = ACTIONS(1243),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(1227),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(1227),
    [anon_sym_DOLLAR] = ACTIONS(1243),
    [sym__special_character] = ACTIONS(1227),
    [anon_sym_DQUOTE] = ACTIONS(1245),
    [sym_raw_string] = ACTIONS(1227),
    [sym_ansi_c_string] = ACTIONS(1227),
    [aux_sym_number_token1] = ACTIONS(1227),
    [aux_sym_number_token2] = ACTIONS(1227),
    [anon_sym_POUND] = ACTIONS(1243),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(1227),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1227),
    [anon_sym_BQUOTE] = ACTIONS(1227),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(1227),
    [anon_sym_LT_LPAREN] = ACTIONS(1227),
    [anon_sym_GT_LPAREN] = ACTIONS(1227),
    [sym_comment] = ACTIONS(3),
    [aux_sym__simple_variable_name_token1] = ACTIONS(1247),
    [aux_sym__multiline_variable_name_token1] = ACTIONS(1247),
    [anon_sym_AT2] = ACTIONS(1243),
    [anon_sym_0] = ACTIONS(1243),
    [anon_sym__] = ACTIONS(1243),
    [sym_file_descriptor] = ACTIONS(1235),
    [sym_variable_name] = ACTIONS(1249),
    [sym_test_operator] = ACTIONS(1235),
    [sym__bare_dollar] = ACTIONS(1235),
    [sym__brace_start] = ACTIONS(1235),
  },
  [467] = {
    [sym_string] = STATE(503),
    [sym_word] = ACTIONS(1239),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(1239),
    [anon_sym_EQ] = ACTIONS(1239),
    [anon_sym_PLUS_PLUS] = ACTIONS(1239),
    [anon_sym_DASH_DASH] = ACTIONS(1239),
    [anon_sym_PLUS_EQ] = ACTIONS(1239),
    [anon_sym_DASH_EQ] = ACTIONS(1239),
    [anon_sym_STAR_EQ] = ACTIONS(1239),
    [anon_sym_SLASH_EQ] = ACTIONS(1239),
    [anon_sym_PERCENT_EQ] = ACTIONS(1239),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(1239),
    [anon_sym_LT_LT_EQ] = ACTIONS(1239),
    [anon_sym_GT_GT_EQ] = ACTIONS(1239),
    [anon_sym_AMP_EQ] = ACTIONS(1239),
    [anon_sym_CARET_EQ] = ACTIONS(1239),
    [anon_sym_PIPE_EQ] = ACTIONS(1239),
    [anon_sym_PIPE_PIPE] = ACTIONS(1239),
    [anon_sym_AMP_AMP] = ACTIONS(1239),
    [anon_sym_PIPE] = ACTIONS(1239),
    [anon_sym_CARET] = ACTIONS(1239),
    [anon_sym_AMP] = ACTIONS(1239),
    [anon_sym_EQ_EQ] = ACTIONS(1239),
    [anon_sym_BANG_EQ] = ACTIONS(1239),
    [anon_sym_LT] = ACTIONS(1239),
    [anon_sym_GT] = ACTIONS(1239),
    [anon_sym_LT_EQ] = ACTIONS(1239),
    [anon_sym_GT_EQ] = ACTIONS(1239),
    [anon_sym_LT_LT] = ACTIONS(1239),
    [anon_sym_GT_GT] = ACTIONS(1239),
    [anon_sym_PLUS] = ACTIONS(1239),
    [anon_sym_DASH] = ACTIONS(1243),
    [anon_sym_STAR] = ACTIONS(1243),
    [anon_sym_SLASH] = ACTIONS(1239),
    [anon_sym_PERCENT] = ACTIONS(1239),
    [anon_sym_STAR_STAR] = ACTIONS(1239),
    [anon_sym_LPAREN] = ACTIONS(1239),
    [anon_sym_PIPE_AMP] = ACTIONS(1239),
    [anon_sym_BANG] = ACTIONS(1243),
    [anon_sym_RBRACK] = ACTIONS(1239),
    [anon_sym_EQ_TILDE] = ACTIONS(1239),
    [anon_sym_AMP_GT] = ACTIONS(1239),
    [anon_sym_AMP_GT_GT] = ACTIONS(1239),
    [anon_sym_LT_AMP] = ACTIONS(1239),
    [anon_sym_GT_AMP] = ACTIONS(1239),
    [anon_sym_GT_PIPE] = ACTIONS(1239),
    [anon_sym_LT_AMP_DASH] = ACTIONS(1239),
    [anon_sym_GT_AMP_DASH] = ACTIONS(1239),
    [anon_sym_LT_LT_DASH] = ACTIONS(1239),
    [anon_sym_LT_LT_LT] = ACTIONS(1239),
    [anon_sym_QMARK] = ACTIONS(1243),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(1239),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(1239),
    [anon_sym_DOLLAR] = ACTIONS(1243),
    [sym__special_character] = ACTIONS(1239),
    [anon_sym_DQUOTE] = ACTIONS(1245),
    [sym_raw_string] = ACTIONS(1239),
    [sym_ansi_c_string] = ACTIONS(1239),
    [aux_sym_number_token1] = ACTIONS(1239),
    [aux_sym_number_token2] = ACTIONS(1239),
    [anon_sym_POUND] = ACTIONS(1243),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(1239),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1239),
    [anon_sym_BQUOTE] = ACTIONS(1239),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(1239),
    [anon_sym_LT_LPAREN] = ACTIONS(1239),
    [anon_sym_GT_LPAREN] = ACTIONS(1239),
    [sym_comment] = ACTIONS(3),
    [aux_sym__simple_variable_name_token1] = ACTIONS(1247),
    [aux_sym__multiline_variable_name_token1] = ACTIONS(1247),
    [anon_sym_AT2] = ACTIONS(1243),
    [anon_sym_0] = ACTIONS(1243),
    [anon_sym__] = ACTIONS(1243),
    [sym_file_descriptor] = ACTIONS(1241),
    [sym_variable_name] = ACTIONS(1249),
    [sym_test_operator] = ACTIONS(1241),
    [sym__bare_dollar] = ACTIONS(1241),
    [sym__brace_start] = ACTIONS(1241),
  },
  [468] = {
    [aux_sym_concatenation_repeat1] = STATE(468),
    [sym_word] = ACTIONS(1251),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(1251),
    [anon_sym_SEMI] = ACTIONS(1251),
    [anon_sym_EQ] = ACTIONS(1251),
    [anon_sym_PLUS_PLUS] = ACTIONS(1251),
    [anon_sym_DASH_DASH] = ACTIONS(1251),
    [anon_sym_PLUS_EQ] = ACTIONS(1251),
    [anon_sym_DASH_EQ] = ACTIONS(1251),
    [anon_sym_STAR_EQ] = ACTIONS(1251),
    [anon_sym_SLASH_EQ] = ACTIONS(1251),
    [anon_sym_PERCENT_EQ] = ACTIONS(1251),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(1251),
    [anon_sym_LT_LT_EQ] = ACTIONS(1251),
    [anon_sym_GT_GT_EQ] = ACTIONS(1251),
    [anon_sym_AMP_EQ] = ACTIONS(1251),
    [anon_sym_CARET_EQ] = ACTIONS(1251),
    [anon_sym_PIPE_EQ] = ACTIONS(1251),
    [anon_sym_PIPE_PIPE] = ACTIONS(1251),
    [anon_sym_AMP_AMP] = ACTIONS(1251),
    [anon_sym_PIPE] = ACTIONS(1251),
    [anon_sym_CARET] = ACTIONS(1251),
    [anon_sym_AMP] = ACTIONS(1251),
    [anon_sym_EQ_EQ] = ACTIONS(1251),
    [anon_sym_BANG_EQ] = ACTIONS(1251),
    [anon_sym_LT] = ACTIONS(1251),
    [anon_sym_GT] = ACTIONS(1251),
    [anon_sym_LT_EQ] = ACTIONS(1251),
    [anon_sym_GT_EQ] = ACTIONS(1251),
    [anon_sym_LT_LT] = ACTIONS(1251),
    [anon_sym_GT_GT] = ACTIONS(1251),
    [anon_sym_PLUS] = ACTIONS(1251),
    [anon_sym_DASH] = ACTIONS(1251),
    [anon_sym_STAR] = ACTIONS(1251),
    [anon_sym_SLASH] = ACTIONS(1251),
    [anon_sym_PERCENT] = ACTIONS(1251),
    [anon_sym_STAR_STAR] = ACTIONS(1251),
    [anon_sym_LPAREN] = ACTIONS(1251),
    [anon_sym_RPAREN] = ACTIONS(1251),
    [anon_sym_SEMI_SEMI] = ACTIONS(1251),
    [anon_sym_PIPE_AMP] = ACTIONS(1251),
    [anon_sym_EQ_TILDE] = ACTIONS(1251),
    [anon_sym_AMP_GT] = ACTIONS(1251),
    [anon_sym_AMP_GT_GT] = ACTIONS(1251),
    [anon_sym_LT_AMP] = ACTIONS(1251),
    [anon_sym_GT_AMP] = ACTIONS(1251),
    [anon_sym_GT_PIPE] = ACTIONS(1251),
    [anon_sym_LT_AMP_DASH] = ACTIONS(1251),
    [anon_sym_GT_AMP_DASH] = ACTIONS(1251),
    [anon_sym_LT_LT_DASH] = ACTIONS(1251),
    [aux_sym_heredoc_redirect_token1] = ACTIONS(1253),
    [anon_sym_LT_LT_LT] = ACTIONS(1251),
    [anon_sym_QMARK] = ACTIONS(1251),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(1251),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(1251),
    [aux_sym_concatenation_token1] = ACTIONS(1255),
    [anon_sym_DOLLAR] = ACTIONS(1251),
    [sym__special_character] = ACTIONS(1251),
    [anon_sym_DQUOTE] = ACTIONS(1251),
    [sym_raw_string] = ACTIONS(1251),
    [sym_ansi_c_string] = ACTIONS(1251),
    [aux_sym_number_token1] = ACTIONS(1251),
    [aux_sym_number_token2] = ACTIONS(1251),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(1251),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1251),
    [anon_sym_BQUOTE] = ACTIONS(1251),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(1251),
    [anon_sym_LT_LPAREN] = ACTIONS(1251),
    [anon_sym_GT_LPAREN] = ACTIONS(1251),
    [sym_comment] = ACTIONS(3),
    [sym_file_descriptor] = ACTIONS(1253),
    [sym__concat] = ACTIONS(1258),
    [sym_test_operator] = ACTIONS(1253),
    [sym__bare_dollar] = ACTIONS(1253),
    [sym__brace_start] = ACTIONS(1253),
  },
  [469] = {
    [aux_sym_concatenation_repeat1] = STATE(470),
    [sym_word] = ACTIONS(1261),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(1261),
    [anon_sym_SEMI] = ACTIONS(1261),
    [anon_sym_EQ] = ACTIONS(1261),
    [anon_sym_PLUS_PLUS] = ACTIONS(1261),
    [anon_sym_DASH_DASH] = ACTIONS(1261),
    [anon_sym_PLUS_EQ] = ACTIONS(1261),
    [anon_sym_DASH_EQ] = ACTIONS(1261),
    [anon_sym_STAR_EQ] = ACTIONS(1261),
    [anon_sym_SLASH_EQ] = ACTIONS(1261),
    [anon_sym_PERCENT_EQ] = ACTIONS(1261),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(1261),
    [anon_sym_LT_LT_EQ] = ACTIONS(1261),
    [anon_sym_GT_GT_EQ] = ACTIONS(1261),
    [anon_sym_AMP_EQ] = ACTIONS(1261),
    [anon_sym_CARET_EQ] = ACTIONS(1261),
    [anon_sym_PIPE_EQ] = ACTIONS(1261),
    [anon_sym_PIPE_PIPE] = ACTIONS(1261),
    [anon_sym_AMP_AMP] = ACTIONS(1261),
    [anon_sym_PIPE] = ACTIONS(1261),
    [anon_sym_CARET] = ACTIONS(1261),
    [anon_sym_AMP] = ACTIONS(1261),
    [anon_sym_EQ_EQ] = ACTIONS(1261),
    [anon_sym_BANG_EQ] = ACTIONS(1261),
    [anon_sym_LT] = ACTIONS(1261),
    [anon_sym_GT] = ACTIONS(1261),
    [anon_sym_LT_EQ] = ACTIONS(1261),
    [anon_sym_GT_EQ] = ACTIONS(1261),
    [anon_sym_LT_LT] = ACTIONS(1261),
    [anon_sym_GT_GT] = ACTIONS(1261),
    [anon_sym_PLUS] = ACTIONS(1261),
    [anon_sym_DASH] = ACTIONS(1261),
    [anon_sym_STAR] = ACTIONS(1261),
    [anon_sym_SLASH] = ACTIONS(1261),
    [anon_sym_PERCENT] = ACTIONS(1261),
    [anon_sym_STAR_STAR] = ACTIONS(1261),
    [anon_sym_LPAREN] = ACTIONS(1261),
    [anon_sym_RPAREN] = ACTIONS(1261),
    [anon_sym_SEMI_SEMI] = ACTIONS(1261),
    [anon_sym_PIPE_AMP] = ACTIONS(1261),
    [anon_sym_EQ_TILDE] = ACTIONS(1261),
    [anon_sym_AMP_GT] = ACTIONS(1261),
    [anon_sym_AMP_GT_GT] = ACTIONS(1261),
    [anon_sym_LT_AMP] = ACTIONS(1261),
    [anon_sym_GT_AMP] = ACTIONS(1261),
    [anon_sym_GT_PIPE] = ACTIONS(1261),
    [anon_sym_LT_AMP_DASH] = ACTIONS(1261),
    [anon_sym_GT_AMP_DASH] = ACTIONS(1261),
    [anon_sym_LT_LT_DASH] = ACTIONS(1261),
    [aux_sym_heredoc_redirect_token1] = ACTIONS(1263),
    [anon_sym_LT_LT_LT] = ACTIONS(1261),
    [anon_sym_QMARK] = ACTIONS(1261),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(1261),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(1261),
    [aux_sym_concatenation_token1] = ACTIONS(256),
    [anon_sym_DOLLAR] = ACTIONS(1261),
    [sym__special_character] = ACTIONS(1261),
    [anon_sym_DQUOTE] = ACTIONS(1261),
    [sym_raw_string] = ACTIONS(1261),
    [sym_ansi_c_string] = ACTIONS(1261),
    [aux_sym_number_token1] = ACTIONS(1261),
    [aux_sym_number_token2] = ACTIONS(1261),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(1261),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1261),
    [anon_sym_BQUOTE] = ACTIONS(1261),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(1261),
    [anon_sym_LT_LPAREN] = ACTIONS(1261),
    [anon_sym_GT_LPAREN] = ACTIONS(1261),
    [sym_comment] = ACTIONS(3),
    [sym_file_descriptor] = ACTIONS(1263),
    [sym__concat] = ACTIONS(280),
    [sym_test_operator] = ACTIONS(1263),
    [sym__bare_dollar] = ACTIONS(1263),
    [sym__brace_start] = ACTIONS(1263),
  },
  [470] = {
    [aux_sym_concatenation_repeat1] = STATE(468),
    [sym_word] = ACTIONS(1265),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(1265),
    [anon_sym_SEMI] = ACTIONS(1265),
    [anon_sym_EQ] = ACTIONS(1265),
    [anon_sym_PLUS_PLUS] = ACTIONS(1265),
    [anon_sym_DASH_DASH] = ACTIONS(1265),
    [anon_sym_PLUS_EQ] = ACTIONS(1265),
    [anon_sym_DASH_EQ] = ACTIONS(1265),
    [anon_sym_STAR_EQ] = ACTIONS(1265),
    [anon_sym_SLASH_EQ] = ACTIONS(1265),
    [anon_sym_PERCENT_EQ] = ACTIONS(1265),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(1265),
    [anon_sym_LT_LT_EQ] = ACTIONS(1265),
    [anon_sym_GT_GT_EQ] = ACTIONS(1265),
    [anon_sym_AMP_EQ] = ACTIONS(1265),
    [anon_sym_CARET_EQ] = ACTIONS(1265),
    [anon_sym_PIPE_EQ] = ACTIONS(1265),
    [anon_sym_PIPE_PIPE] = ACTIONS(1265),
    [anon_sym_AMP_AMP] = ACTIONS(1265),
    [anon_sym_PIPE] = ACTIONS(1265),
    [anon_sym_CARET] = ACTIONS(1265),
    [anon_sym_AMP] = ACTIONS(1265),
    [anon_sym_EQ_EQ] = ACTIONS(1265),
    [anon_sym_BANG_EQ] = ACTIONS(1265),
    [anon_sym_LT] = ACTIONS(1265),
    [anon_sym_GT] = ACTIONS(1265),
    [anon_sym_LT_EQ] = ACTIONS(1265),
    [anon_sym_GT_EQ] = ACTIONS(1265),
    [anon_sym_LT_LT] = ACTIONS(1265),
    [anon_sym_GT_GT] = ACTIONS(1265),
    [anon_sym_PLUS] = ACTIONS(1265),
    [anon_sym_DASH] = ACTIONS(1265),
    [anon_sym_STAR] = ACTIONS(1265),
    [anon_sym_SLASH] = ACTIONS(1265),
    [anon_sym_PERCENT] = ACTIONS(1265),
    [anon_sym_STAR_STAR] = ACTIONS(1265),
    [anon_sym_LPAREN] = ACTIONS(1265),
    [anon_sym_RPAREN] = ACTIONS(1265),
    [anon_sym_SEMI_SEMI] = ACTIONS(1265),
    [anon_sym_PIPE_AMP] = ACTIONS(1265),
    [anon_sym_EQ_TILDE] = ACTIONS(1265),
    [anon_sym_AMP_GT] = ACTIONS(1265),
    [anon_sym_AMP_GT_GT] = ACTIONS(1265),
    [anon_sym_LT_AMP] = ACTIONS(1265),
    [anon_sym_GT_AMP] = ACTIONS(1265),
    [anon_sym_GT_PIPE] = ACTIONS(1265),
    [anon_sym_LT_AMP_DASH] = ACTIONS(1265),
    [anon_sym_GT_AMP_DASH] = ACTIONS(1265),
    [anon_sym_LT_LT_DASH] = ACTIONS(1265),
    [aux_sym_heredoc_redirect_token1] = ACTIONS(1267),
    [anon_sym_LT_LT_LT] = ACTIONS(1265),
    [anon_sym_QMARK] = ACTIONS(1265),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(1265),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(1265),
    [aux_sym_concatenation_token1] = ACTIONS(256),
    [anon_sym_DOLLAR] = ACTIONS(1265),
    [sym__special_character] = ACTIONS(1265),
    [anon_sym_DQUOTE] = ACTIONS(1265),
    [sym_raw_string] = ACTIONS(1265),
    [sym_ansi_c_string] = ACTIONS(1265),
    [aux_sym_number_token1] = ACTIONS(1265),
    [aux_sym_number_token2] = ACTIONS(1265),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(1265),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1265),
    [anon_sym_BQUOTE] = ACTIONS(1265),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(1265),
    [anon_sym_LT_LPAREN] = ACTIONS(1265),
    [anon_sym_GT_LPAREN] = ACTIONS(1265),
    [sym_comment] = ACTIONS(3),
    [sym_file_descriptor] = ACTIONS(1267),
    [sym__concat] = ACTIONS(1269),
    [sym_test_operator] = ACTIONS(1267),
    [sym__bare_dollar] = ACTIONS(1267),
    [sym__brace_start] = ACTIONS(1267),
  },
  [471] = {
    [aux_sym_concatenation_repeat1] = STATE(473),
    [sym_word] = ACTIONS(1271),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(1271),
    [anon_sym_SEMI] = ACTIONS(1271),
    [anon_sym_EQ] = ACTIONS(1273),
    [anon_sym_PLUS_PLUS] = ACTIONS(1273),
    [anon_sym_DASH_DASH] = ACTIONS(1273),
    [anon_sym_PLUS_EQ] = ACTIONS(1273),
    [anon_sym_DASH_EQ] = ACTIONS(1273),
    [anon_sym_STAR_EQ] = ACTIONS(1273),
    [anon_sym_SLASH_EQ] = ACTIONS(1273),
    [anon_sym_PERCENT_EQ] = ACTIONS(1273),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(1273),
    [anon_sym_LT_LT_EQ] = ACTIONS(1273),
    [anon_sym_GT_GT_EQ] = ACTIONS(1273),
    [anon_sym_AMP_EQ] = ACTIONS(1273),
    [anon_sym_CARET_EQ] = ACTIONS(1273),
    [anon_sym_PIPE_EQ] = ACTIONS(1273),
    [anon_sym_PIPE_PIPE] = ACTIONS(1275),
    [anon_sym_AMP_AMP] = ACTIONS(1275),
    [anon_sym_PIPE] = ACTIONS(1275),
    [anon_sym_CARET] = ACTIONS(1273),
    [anon_sym_AMP] = ACTIONS(1275),
    [anon_sym_EQ_EQ] = ACTIONS(1275),
    [anon_sym_BANG_EQ] = ACTIONS(1273),
    [anon_sym_LT] = ACTIONS(1275),
    [anon_sym_GT] = ACTIONS(1275),
    [anon_sym_LT_EQ] = ACTIONS(1273),
    [anon_sym_GT_EQ] = ACTIONS(1273),
    [anon_sym_LT_LT] = ACTIONS(1275),
    [anon_sym_GT_GT] = ACTIONS(1275),
    [anon_sym_PLUS] = ACTIONS(1273),
    [anon_sym_DASH] = ACTIONS(1273),
    [anon_sym_STAR] = ACTIONS(1273),
    [anon_sym_SLASH] = ACTIONS(1273),
    [anon_sym_PERCENT] = ACTIONS(1273),
    [anon_sym_STAR_STAR] = ACTIONS(1273),
    [anon_sym_LPAREN] = ACTIONS(1278),
    [anon_sym_RPAREN] = ACTIONS(1275),
    [anon_sym_SEMI_SEMI] = ACTIONS(1271),
    [anon_sym_PIPE_AMP] = ACTIONS(1271),
    [anon_sym_EQ_TILDE] = ACTIONS(1275),
    [anon_sym_AMP_GT] = ACTIONS(1271),
    [anon_sym_AMP_GT_GT] = ACTIONS(1271),
    [anon_sym_LT_AMP] = ACTIONS(1271),
    [anon_sym_GT_AMP] = ACTIONS(1271),
    [anon_sym_GT_PIPE] = ACTIONS(1271),
    [anon_sym_LT_AMP_DASH] = ACTIONS(1271),
    [anon_sym_GT_AMP_DASH] = ACTIONS(1271),
    [anon_sym_LT_LT_DASH] = ACTIONS(1271),
    [aux_sym_heredoc_redirect_token1] = ACTIONS(1281),
    [anon_sym_LT_LT_LT] = ACTIONS(1271),
    [anon_sym_QMARK] = ACTIONS(1273),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(1271),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(1271),
    [aux_sym_concatenation_token1] = ACTIONS(256),
    [anon_sym_DOLLAR] = ACTIONS(1271),
    [sym__special_character] = ACTIONS(1271),
    [anon_sym_DQUOTE] = ACTIONS(1271),
    [sym_raw_string] = ACTIONS(1271),
    [sym_ansi_c_string] = ACTIONS(1271),
    [aux_sym_number_token1] = ACTIONS(1271),
    [aux_sym_number_token2] = ACTIONS(1271),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(1271),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1271),
    [anon_sym_BQUOTE] = ACTIONS(1271),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(1271),
    [anon_sym_LT_LPAREN] = ACTIONS(1271),
    [anon_sym_GT_LPAREN] = ACTIONS(1271),
    [sym_comment] = ACTIONS(3),
    [sym_file_descriptor] = ACTIONS(1281),
    [sym__concat] = ACTIONS(280),
    [sym_test_operator] = ACTIONS(1283),
    [sym__bare_dollar] = ACTIONS(1281),
    [sym__brace_start] = ACTIONS(1281),
  },
  [472] = {
    [aux_sym_concatenation_repeat1] = STATE(473),
    [sym_word] = ACTIONS(1271),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(1271),
    [anon_sym_SEMI] = ACTIONS(1271),
    [anon_sym_EQ] = ACTIONS(1273),
    [anon_sym_PLUS_PLUS] = ACTIONS(1273),
    [anon_sym_DASH_DASH] = ACTIONS(1273),
    [anon_sym_PLUS_EQ] = ACTIONS(1273),
    [anon_sym_DASH_EQ] = ACTIONS(1273),
    [anon_sym_STAR_EQ] = ACTIONS(1273),
    [anon_sym_SLASH_EQ] = ACTIONS(1273),
    [anon_sym_PERCENT_EQ] = ACTIONS(1273),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(1273),
    [anon_sym_LT_LT_EQ] = ACTIONS(1273),
    [anon_sym_GT_GT_EQ] = ACTIONS(1273),
    [anon_sym_AMP_EQ] = ACTIONS(1273),
    [anon_sym_CARET_EQ] = ACTIONS(1273),
    [anon_sym_PIPE_EQ] = ACTIONS(1273),
    [anon_sym_PIPE_PIPE] = ACTIONS(1275),
    [anon_sym_AMP_AMP] = ACTIONS(1275),
    [anon_sym_PIPE] = ACTIONS(1275),
    [anon_sym_CARET] = ACTIONS(1273),
    [anon_sym_AMP] = ACTIONS(1275),
    [anon_sym_EQ_EQ] = ACTIONS(1275),
    [anon_sym_BANG_EQ] = ACTIONS(1273),
    [anon_sym_LT] = ACTIONS(1275),
    [anon_sym_GT] = ACTIONS(1275),
    [anon_sym_LT_EQ] = ACTIONS(1273),
    [anon_sym_GT_EQ] = ACTIONS(1273),
    [anon_sym_LT_LT] = ACTIONS(1275),
    [anon_sym_GT_GT] = ACTIONS(1275),
    [anon_sym_PLUS] = ACTIONS(1273),
    [anon_sym_DASH] = ACTIONS(1273),
    [anon_sym_STAR] = ACTIONS(1273),
    [anon_sym_SLASH] = ACTIONS(1273),
    [anon_sym_PERCENT] = ACTIONS(1273),
    [anon_sym_STAR_STAR] = ACTIONS(1273),
    [anon_sym_LPAREN] = ACTIONS(1271),
    [anon_sym_RPAREN] = ACTIONS(1275),
    [anon_sym_SEMI_SEMI] = ACTIONS(1271),
    [anon_sym_PIPE_AMP] = ACTIONS(1271),
    [anon_sym_EQ_TILDE] = ACTIONS(1275),
    [anon_sym_AMP_GT] = ACTIONS(1271),
    [anon_sym_AMP_GT_GT] = ACTIONS(1271),
    [anon_sym_LT_AMP] = ACTIONS(1271),
    [anon_sym_GT_AMP] = ACTIONS(1271),
    [anon_sym_GT_PIPE] = ACTIONS(1271),
    [anon_sym_LT_AMP_DASH] = ACTIONS(1271),
    [anon_sym_GT_AMP_DASH] = ACTIONS(1271),
    [anon_sym_LT_LT_DASH] = ACTIONS(1271),
    [aux_sym_heredoc_redirect_token1] = ACTIONS(1281),
    [anon_sym_LT_LT_LT] = ACTIONS(1271),
    [anon_sym_QMARK] = ACTIONS(1273),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(1271),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(1271),
    [aux_sym_concatenation_token1] = ACTIONS(256),
    [anon_sym_DOLLAR] = ACTIONS(1271),
    [sym__special_character] = ACTIONS(1271),
    [anon_sym_DQUOTE] = ACTIONS(1271),
    [sym_raw_string] = ACTIONS(1271),
    [sym_ansi_c_string] = ACTIONS(1271),
    [aux_sym_number_token1] = ACTIONS(1271),
    [aux_sym_number_token2] = ACTIONS(1271),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(1271),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1271),
    [anon_sym_BQUOTE] = ACTIONS(1271),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(1271),
    [anon_sym_LT_LPAREN] = ACTIONS(1271),
    [anon_sym_GT_LPAREN] = ACTIONS(1271),
    [sym_comment] = ACTIONS(3),
    [sym_file_descriptor] = ACTIONS(1281),
    [sym__concat] = ACTIONS(280),
    [sym_test_operator] = ACTIONS(1283),
    [sym__bare_dollar] = ACTIONS(1281),
    [sym__brace_start] = ACTIONS(1281),
  },
  [473] = {
    [aux_sym_concatenation_repeat1] = STATE(468),
    [sym_word] = ACTIONS(1286),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(1286),
    [anon_sym_SEMI] = ACTIONS(1286),
    [anon_sym_EQ] = ACTIONS(1286),
    [anon_sym_PLUS_PLUS] = ACTIONS(1286),
    [anon_sym_DASH_DASH] = ACTIONS(1286),
    [anon_sym_PLUS_EQ] = ACTIONS(1286),
    [anon_sym_DASH_EQ] = ACTIONS(1286),
    [anon_sym_STAR_EQ] = ACTIONS(1286),
    [anon_sym_SLASH_EQ] = ACTIONS(1286),
    [anon_sym_PERCENT_EQ] = ACTIONS(1286),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(1286),
    [anon_sym_LT_LT_EQ] = ACTIONS(1286),
    [anon_sym_GT_GT_EQ] = ACTIONS(1286),
    [anon_sym_AMP_EQ] = ACTIONS(1286),
    [anon_sym_CARET_EQ] = ACTIONS(1286),
    [anon_sym_PIPE_EQ] = ACTIONS(1286),
    [anon_sym_PIPE_PIPE] = ACTIONS(1286),
    [anon_sym_AMP_AMP] = ACTIONS(1286),
    [anon_sym_PIPE] = ACTIONS(1286),
    [anon_sym_CARET] = ACTIONS(1286),
    [anon_sym_AMP] = ACTIONS(1286),
    [anon_sym_EQ_EQ] = ACTIONS(1286),
    [anon_sym_BANG_EQ] = ACTIONS(1286),
    [anon_sym_LT] = ACTIONS(1286),
    [anon_sym_GT] = ACTIONS(1286),
    [anon_sym_LT_EQ] = ACTIONS(1286),
    [anon_sym_GT_EQ] = ACTIONS(1286),
    [anon_sym_LT_LT] = ACTIONS(1286),
    [anon_sym_GT_GT] = ACTIONS(1286),
    [anon_sym_PLUS] = ACTIONS(1286),
    [anon_sym_DASH] = ACTIONS(1286),
    [anon_sym_STAR] = ACTIONS(1286),
    [anon_sym_SLASH] = ACTIONS(1286),
    [anon_sym_PERCENT] = ACTIONS(1286),
    [anon_sym_STAR_STAR] = ACTIONS(1286),
    [anon_sym_LPAREN] = ACTIONS(1286),
    [anon_sym_RPAREN] = ACTIONS(1286),
    [anon_sym_SEMI_SEMI] = ACTIONS(1286),
    [anon_sym_PIPE_AMP] = ACTIONS(1286),
    [anon_sym_EQ_TILDE] = ACTIONS(1286),
    [anon_sym_AMP_GT] = ACTIONS(1286),
    [anon_sym_AMP_GT_GT] = ACTIONS(1286),
    [anon_sym_LT_AMP] = ACTIONS(1286),
    [anon_sym_GT_AMP] = ACTIONS(1286),
    [anon_sym_GT_PIPE] = ACTIONS(1286),
    [anon_sym_LT_AMP_DASH] = ACTIONS(1286),
    [anon_sym_GT_AMP_DASH] = ACTIONS(1286),
    [anon_sym_LT_LT_DASH] = ACTIONS(1286),
    [aux_sym_heredoc_redirect_token1] = ACTIONS(1288),
    [anon_sym_LT_LT_LT] = ACTIONS(1286),
    [anon_sym_QMARK] = ACTIONS(1286),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(1286),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(1286),
    [aux_sym_concatenation_token1] = ACTIONS(256),
    [anon_sym_DOLLAR] = ACTIONS(1286),
    [sym__special_character] = ACTIONS(1286),
    [anon_sym_DQUOTE] = ACTIONS(1286),
    [sym_raw_string] = ACTIONS(1286),
    [sym_ansi_c_string] = ACTIONS(1286),
    [aux_sym_number_token1] = ACTIONS(1286),
    [aux_sym_number_token2] = ACTIONS(1286),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(1286),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1286),
    [anon_sym_BQUOTE] = ACTIONS(1286),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(1286),
    [anon_sym_LT_LPAREN] = ACTIONS(1286),
    [anon_sym_GT_LPAREN] = ACTIONS(1286),
    [sym_comment] = ACTIONS(3),
    [sym_file_descriptor] = ACTIONS(1288),
    [sym__concat] = ACTIONS(1290),
    [sym_test_operator] = ACTIONS(1288),
    [sym__bare_dollar] = ACTIONS(1288),
    [sym__brace_start] = ACTIONS(1288),
  },
  [474] = {
    [sym_word] = ACTIONS(1292),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(1292),
    [anon_sym_SEMI] = ACTIONS(1292),
    [anon_sym_EQ] = ACTIONS(1292),
    [anon_sym_PLUS_PLUS] = ACTIONS(1292),
    [anon_sym_DASH_DASH] = ACTIONS(1292),
    [anon_sym_PLUS_EQ] = ACTIONS(1292),
    [anon_sym_DASH_EQ] = ACTIONS(1292),
    [anon_sym_STAR_EQ] = ACTIONS(1292),
    [anon_sym_SLASH_EQ] = ACTIONS(1292),
    [anon_sym_PERCENT_EQ] = ACTIONS(1292),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(1292),
    [anon_sym_LT_LT_EQ] = ACTIONS(1292),
    [anon_sym_GT_GT_EQ] = ACTIONS(1292),
    [anon_sym_AMP_EQ] = ACTIONS(1292),
    [anon_sym_CARET_EQ] = ACTIONS(1292),
    [anon_sym_PIPE_EQ] = ACTIONS(1292),
    [anon_sym_PIPE_PIPE] = ACTIONS(1292),
    [anon_sym_AMP_AMP] = ACTIONS(1292),
    [anon_sym_PIPE] = ACTIONS(1292),
    [anon_sym_CARET] = ACTIONS(1292),
    [anon_sym_AMP] = ACTIONS(1292),
    [anon_sym_EQ_EQ] = ACTIONS(1292),
    [anon_sym_BANG_EQ] = ACTIONS(1292),
    [anon_sym_LT] = ACTIONS(1292),
    [anon_sym_GT] = ACTIONS(1292),
    [anon_sym_LT_EQ] = ACTIONS(1292),
    [anon_sym_GT_EQ] = ACTIONS(1292),
    [anon_sym_LT_LT] = ACTIONS(1292),
    [anon_sym_GT_GT] = ACTIONS(1292),
    [anon_sym_PLUS] = ACTIONS(1292),
    [anon_sym_DASH] = ACTIONS(1292),
    [anon_sym_STAR] = ACTIONS(1292),
    [anon_sym_SLASH] = ACTIONS(1292),
    [anon_sym_PERCENT] = ACTIONS(1292),
    [anon_sym_STAR_STAR] = ACTIONS(1292),
    [anon_sym_LPAREN] = ACTIONS(1292),
    [anon_sym_RPAREN] = ACTIONS(1292),
    [anon_sym_SEMI_SEMI] = ACTIONS(1292),
    [anon_sym_PIPE_AMP] = ACTIONS(1292),
    [anon_sym_EQ_TILDE] = ACTIONS(1292),
    [anon_sym_AMP_GT] = ACTIONS(1292),
    [anon_sym_AMP_GT_GT] = ACTIONS(1292),
    [anon_sym_LT_AMP] = ACTIONS(1292),
    [anon_sym_GT_AMP] = ACTIONS(1292),
    [anon_sym_GT_PIPE] = ACTIONS(1292),
    [anon_sym_LT_AMP_DASH] = ACTIONS(1292),
    [anon_sym_GT_AMP_DASH] = ACTIONS(1292),
    [anon_sym_LT_LT_DASH] = ACTIONS(1292),
    [aux_sym_heredoc_redirect_token1] = ACTIONS(1294),
    [anon_sym_LT_LT_LT] = ACTIONS(1292),
    [anon_sym_QMARK] = ACTIONS(1292),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(1292),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(1292),
    [aux_sym_concatenation_token1] = ACTIONS(1292),
    [anon_sym_DOLLAR] = ACTIONS(1292),
    [sym__special_character] = ACTIONS(1292),
    [anon_sym_DQUOTE] = ACTIONS(1292),
    [sym_raw_string] = ACTIONS(1292),
    [sym_ansi_c_string] = ACTIONS(1292),
    [aux_sym_number_token1] = ACTIONS(1292),
    [aux_sym_number_token2] = ACTIONS(1292),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(1292),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1292),
    [anon_sym_BQUOTE] = ACTIONS(1292),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(1292),
    [anon_sym_LT_LPAREN] = ACTIONS(1292),
    [anon_sym_GT_LPAREN] = ACTIONS(1292),
    [sym_comment] = ACTIONS(3),
    [sym_file_descriptor] = ACTIONS(1294),
    [sym__concat] = ACTIONS(1294),
    [sym_test_operator] = ACTIONS(1294),
    [sym__bare_dollar] = ACTIONS(1294),
    [sym__brace_start] = ACTIONS(1294),
  },
  [475] = {
    [sym_word] = ACTIONS(1296),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(1296),
    [anon_sym_SEMI] = ACTIONS(1296),
    [anon_sym_EQ] = ACTIONS(1296),
    [anon_sym_PLUS_PLUS] = ACTIONS(1296),
    [anon_sym_DASH_DASH] = ACTIONS(1296),
    [anon_sym_PLUS_EQ] = ACTIONS(1296),
    [anon_sym_DASH_EQ] = ACTIONS(1296),
    [anon_sym_STAR_EQ] = ACTIONS(1296),
    [anon_sym_SLASH_EQ] = ACTIONS(1296),
    [anon_sym_PERCENT_EQ] = ACTIONS(1296),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(1296),
    [anon_sym_LT_LT_EQ] = ACTIONS(1296),
    [anon_sym_GT_GT_EQ] = ACTIONS(1296),
    [anon_sym_AMP_EQ] = ACTIONS(1296),
    [anon_sym_CARET_EQ] = ACTIONS(1296),
    [anon_sym_PIPE_EQ] = ACTIONS(1296),
    [anon_sym_PIPE_PIPE] = ACTIONS(1296),
    [anon_sym_AMP_AMP] = ACTIONS(1296),
    [anon_sym_PIPE] = ACTIONS(1296),
    [anon_sym_CARET] = ACTIONS(1296),
    [anon_sym_AMP] = ACTIONS(1296),
    [anon_sym_EQ_EQ] = ACTIONS(1296),
    [anon_sym_BANG_EQ] = ACTIONS(1296),
    [anon_sym_LT] = ACTIONS(1296),
    [anon_sym_GT] = ACTIONS(1296),
    [anon_sym_LT_EQ] = ACTIONS(1296),
    [anon_sym_GT_EQ] = ACTIONS(1296),
    [anon_sym_LT_LT] = ACTIONS(1296),
    [anon_sym_GT_GT] = ACTIONS(1296),
    [anon_sym_PLUS] = ACTIONS(1296),
    [anon_sym_DASH] = ACTIONS(1296),
    [anon_sym_STAR] = ACTIONS(1296),
    [anon_sym_SLASH] = ACTIONS(1296),
    [anon_sym_PERCENT] = ACTIONS(1296),
    [anon_sym_STAR_STAR] = ACTIONS(1296),
    [anon_sym_LPAREN] = ACTIONS(1296),
    [anon_sym_RPAREN] = ACTIONS(1296),
    [anon_sym_SEMI_SEMI] = ACTIONS(1296),
    [anon_sym_PIPE_AMP] = ACTIONS(1296),
    [anon_sym_EQ_TILDE] = ACTIONS(1296),
    [anon_sym_AMP_GT] = ACTIONS(1296),
    [anon_sym_AMP_GT_GT] = ACTIONS(1296),
    [anon_sym_LT_AMP] = ACTIONS(1296),
    [anon_sym_GT_AMP] = ACTIONS(1296),
    [anon_sym_GT_PIPE] = ACTIONS(1296),
    [anon_sym_LT_AMP_DASH] = ACTIONS(1296),
    [anon_sym_GT_AMP_DASH] = ACTIONS(1296),
    [anon_sym_LT_LT_DASH] = ACTIONS(1296),
    [aux_sym_heredoc_redirect_token1] = ACTIONS(1298),
    [anon_sym_LT_LT_LT] = ACTIONS(1296),
    [anon_sym_QMARK] = ACTIONS(1296),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(1296),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(1296),
    [aux_sym_concatenation_token1] = ACTIONS(1296),
    [anon_sym_DOLLAR] = ACTIONS(1296),
    [sym__special_character] = ACTIONS(1296),
    [anon_sym_DQUOTE] = ACTIONS(1296),
    [sym_raw_string] = ACTIONS(1296),
    [sym_ansi_c_string] = ACTIONS(1296),
    [aux_sym_number_token1] = ACTIONS(1296),
    [aux_sym_number_token2] = ACTIONS(1296),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(1296),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1296),
    [anon_sym_BQUOTE] = ACTIONS(1296),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(1296),
    [anon_sym_LT_LPAREN] = ACTIONS(1296),
    [anon_sym_GT_LPAREN] = ACTIONS(1296),
    [sym_comment] = ACTIONS(3),
    [sym_file_descriptor] = ACTIONS(1298),
    [sym__concat] = ACTIONS(1298),
    [sym_test_operator] = ACTIONS(1298),
    [sym__bare_dollar] = ACTIONS(1298),
    [sym__brace_start] = ACTIONS(1298),
  },
  [476] = {
    [sym_word] = ACTIONS(1300),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(1300),
    [anon_sym_SEMI] = ACTIONS(1300),
    [anon_sym_EQ] = ACTIONS(1300),
    [anon_sym_PLUS_PLUS] = ACTIONS(1300),
    [anon_sym_DASH_DASH] = ACTIONS(1300),
    [anon_sym_PLUS_EQ] = ACTIONS(1300),
    [anon_sym_DASH_EQ] = ACTIONS(1300),
    [anon_sym_STAR_EQ] = ACTIONS(1300),
    [anon_sym_SLASH_EQ] = ACTIONS(1300),
    [anon_sym_PERCENT_EQ] = ACTIONS(1300),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(1300),
    [anon_sym_LT_LT_EQ] = ACTIONS(1300),
    [anon_sym_GT_GT_EQ] = ACTIONS(1300),
    [anon_sym_AMP_EQ] = ACTIONS(1300),
    [anon_sym_CARET_EQ] = ACTIONS(1300),
    [anon_sym_PIPE_EQ] = ACTIONS(1300),
    [anon_sym_PIPE_PIPE] = ACTIONS(1300),
    [anon_sym_AMP_AMP] = ACTIONS(1300),
    [anon_sym_PIPE] = ACTIONS(1300),
    [anon_sym_CARET] = ACTIONS(1300),
    [anon_sym_AMP] = ACTIONS(1300),
    [anon_sym_EQ_EQ] = ACTIONS(1300),
    [anon_sym_BANG_EQ] = ACTIONS(1300),
    [anon_sym_LT] = ACTIONS(1300),
    [anon_sym_GT] = ACTIONS(1300),
    [anon_sym_LT_EQ] = ACTIONS(1300),
    [anon_sym_GT_EQ] = ACTIONS(1300),
    [anon_sym_LT_LT] = ACTIONS(1300),
    [anon_sym_GT_GT] = ACTIONS(1300),
    [anon_sym_PLUS] = ACTIONS(1300),
    [anon_sym_DASH] = ACTIONS(1300),
    [anon_sym_STAR] = ACTIONS(1300),
    [anon_sym_SLASH] = ACTIONS(1300),
    [anon_sym_PERCENT] = ACTIONS(1300),
    [anon_sym_STAR_STAR] = ACTIONS(1300),
    [anon_sym_LPAREN] = ACTIONS(1300),
    [anon_sym_RPAREN] = ACTIONS(1300),
    [anon_sym_SEMI_SEMI] = ACTIONS(1300),
    [anon_sym_PIPE_AMP] = ACTIONS(1300),
    [anon_sym_EQ_TILDE] = ACTIONS(1300),
    [anon_sym_AMP_GT] = ACTIONS(1300),
    [anon_sym_AMP_GT_GT] = ACTIONS(1300),
    [anon_sym_LT_AMP] = ACTIONS(1300),
    [anon_sym_GT_AMP] = ACTIONS(1300),
    [anon_sym_GT_PIPE] = ACTIONS(1300),
    [anon_sym_LT_AMP_DASH] = ACTIONS(1300),
    [anon_sym_GT_AMP_DASH] = ACTIONS(1300),
    [anon_sym_LT_LT_DASH] = ACTIONS(1300),
    [aux_sym_heredoc_redirect_token1] = ACTIONS(1302),
    [anon_sym_LT_LT_LT] = ACTIONS(1300),
    [anon_sym_QMARK] = ACTIONS(1300),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(1300),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(1300),
    [aux_sym_concatenation_token1] = ACTIONS(1300),
    [anon_sym_DOLLAR] = ACTIONS(1300),
    [sym__special_character] = ACTIONS(1300),
    [anon_sym_DQUOTE] = ACTIONS(1300),
    [sym_raw_string] = ACTIONS(1300),
    [sym_ansi_c_string] = ACTIONS(1300),
    [aux_sym_number_token1] = ACTIONS(1300),
    [aux_sym_number_token2] = ACTIONS(1300),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(1300),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1300),
    [anon_sym_BQUOTE] = ACTIONS(1300),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(1300),
    [anon_sym_LT_LPAREN] = ACTIONS(1300),
    [anon_sym_GT_LPAREN] = ACTIONS(1300),
    [sym_comment] = ACTIONS(3),
    [sym_file_descriptor] = ACTIONS(1302),
    [sym__concat] = ACTIONS(1302),
    [sym_test_operator] = ACTIONS(1302),
    [sym__bare_dollar] = ACTIONS(1302),
    [sym__brace_start] = ACTIONS(1302),
  },
  [477] = {
    [sym_word] = ACTIONS(1304),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(1304),
    [anon_sym_SEMI] = ACTIONS(1304),
    [anon_sym_EQ] = ACTIONS(1304),
    [anon_sym_PLUS_PLUS] = ACTIONS(1304),
    [anon_sym_DASH_DASH] = ACTIONS(1304),
    [anon_sym_PLUS_EQ] = ACTIONS(1304),
    [anon_sym_DASH_EQ] = ACTIONS(1304),
    [anon_sym_STAR_EQ] = ACTIONS(1304),
    [anon_sym_SLASH_EQ] = ACTIONS(1304),
    [anon_sym_PERCENT_EQ] = ACTIONS(1304),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(1304),
    [anon_sym_LT_LT_EQ] = ACTIONS(1304),
    [anon_sym_GT_GT_EQ] = ACTIONS(1304),
    [anon_sym_AMP_EQ] = ACTIONS(1304),
    [anon_sym_CARET_EQ] = ACTIONS(1304),
    [anon_sym_PIPE_EQ] = ACTIONS(1304),
    [anon_sym_PIPE_PIPE] = ACTIONS(1304),
    [anon_sym_AMP_AMP] = ACTIONS(1304),
    [anon_sym_PIPE] = ACTIONS(1304),
    [anon_sym_CARET] = ACTIONS(1304),
    [anon_sym_AMP] = ACTIONS(1304),
    [anon_sym_EQ_EQ] = ACTIONS(1304),
    [anon_sym_BANG_EQ] = ACTIONS(1304),
    [anon_sym_LT] = ACTIONS(1304),
    [anon_sym_GT] = ACTIONS(1304),
    [anon_sym_LT_EQ] = ACTIONS(1304),
    [anon_sym_GT_EQ] = ACTIONS(1304),
    [anon_sym_LT_LT] = ACTIONS(1304),
    [anon_sym_GT_GT] = ACTIONS(1304),
    [anon_sym_PLUS] = ACTIONS(1304),
    [anon_sym_DASH] = ACTIONS(1304),
    [anon_sym_STAR] = ACTIONS(1304),
    [anon_sym_SLASH] = ACTIONS(1304),
    [anon_sym_PERCENT] = ACTIONS(1304),
    [anon_sym_STAR_STAR] = ACTIONS(1304),
    [anon_sym_LPAREN] = ACTIONS(1304),
    [anon_sym_RPAREN] = ACTIONS(1304),
    [anon_sym_SEMI_SEMI] = ACTIONS(1304),
    [anon_sym_PIPE_AMP] = ACTIONS(1304),
    [anon_sym_EQ_TILDE] = ACTIONS(1304),
    [anon_sym_AMP_GT] = ACTIONS(1304),
    [anon_sym_AMP_GT_GT] = ACTIONS(1304),
    [anon_sym_LT_AMP] = ACTIONS(1304),
    [anon_sym_GT_AMP] = ACTIONS(1304),
    [anon_sym_GT_PIPE] = ACTIONS(1304),
    [anon_sym_LT_AMP_DASH] = ACTIONS(1304),
    [anon_sym_GT_AMP_DASH] = ACTIONS(1304),
    [anon_sym_LT_LT_DASH] = ACTIONS(1304),
    [aux_sym_heredoc_redirect_token1] = ACTIONS(1306),
    [anon_sym_LT_LT_LT] = ACTIONS(1304),
    [anon_sym_QMARK] = ACTIONS(1304),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(1304),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(1304),
    [aux_sym_concatenation_token1] = ACTIONS(1304),
    [anon_sym_DOLLAR] = ACTIONS(1304),
    [sym__special_character] = ACTIONS(1304),
    [anon_sym_DQUOTE] = ACTIONS(1304),
    [sym_raw_string] = ACTIONS(1304),
    [sym_ansi_c_string] = ACTIONS(1304),
    [aux_sym_number_token1] = ACTIONS(1304),
    [aux_sym_number_token2] = ACTIONS(1304),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(1304),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1304),
    [anon_sym_BQUOTE] = ACTIONS(1304),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(1304),
    [anon_sym_LT_LPAREN] = ACTIONS(1304),
    [anon_sym_GT_LPAREN] = ACTIONS(1304),
    [sym_comment] = ACTIONS(3),
    [sym_file_descriptor] = ACTIONS(1306),
    [sym__concat] = ACTIONS(1306),
    [sym_test_operator] = ACTIONS(1306),
    [sym__bare_dollar] = ACTIONS(1306),
    [sym__brace_start] = ACTIONS(1306),
  },
  [478] = {
    [sym_word] = ACTIONS(1308),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(1308),
    [anon_sym_SEMI] = ACTIONS(1308),
    [anon_sym_EQ] = ACTIONS(1308),
    [anon_sym_PLUS_PLUS] = ACTIONS(1308),
    [anon_sym_DASH_DASH] = ACTIONS(1308),
    [anon_sym_PLUS_EQ] = ACTIONS(1308),
    [anon_sym_DASH_EQ] = ACTIONS(1308),
    [anon_sym_STAR_EQ] = ACTIONS(1308),
    [anon_sym_SLASH_EQ] = ACTIONS(1308),
    [anon_sym_PERCENT_EQ] = ACTIONS(1308),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(1308),
    [anon_sym_LT_LT_EQ] = ACTIONS(1308),
    [anon_sym_GT_GT_EQ] = ACTIONS(1308),
    [anon_sym_AMP_EQ] = ACTIONS(1308),
    [anon_sym_CARET_EQ] = ACTIONS(1308),
    [anon_sym_PIPE_EQ] = ACTIONS(1308),
    [anon_sym_PIPE_PIPE] = ACTIONS(1308),
    [anon_sym_AMP_AMP] = ACTIONS(1308),
    [anon_sym_PIPE] = ACTIONS(1308),
    [anon_sym_CARET] = ACTIONS(1308),
    [anon_sym_AMP] = ACTIONS(1308),
    [anon_sym_EQ_EQ] = ACTIONS(1308),
    [anon_sym_BANG_EQ] = ACTIONS(1308),
    [anon_sym_LT] = ACTIONS(1308),
    [anon_sym_GT] = ACTIONS(1308),
    [anon_sym_LT_EQ] = ACTIONS(1308),
    [anon_sym_GT_EQ] = ACTIONS(1308),
    [anon_sym_LT_LT] = ACTIONS(1308),
    [anon_sym_GT_GT] = ACTIONS(1308),
    [anon_sym_PLUS] = ACTIONS(1308),
    [anon_sym_DASH] = ACTIONS(1308),
    [anon_sym_STAR] = ACTIONS(1308),
    [anon_sym_SLASH] = ACTIONS(1308),
    [anon_sym_PERCENT] = ACTIONS(1308),
    [anon_sym_STAR_STAR] = ACTIONS(1308),
    [anon_sym_LPAREN] = ACTIONS(1308),
    [anon_sym_RPAREN] = ACTIONS(1308),
    [anon_sym_SEMI_SEMI] = ACTIONS(1308),
    [anon_sym_PIPE_AMP] = ACTIONS(1308),
    [anon_sym_EQ_TILDE] = ACTIONS(1308),
    [anon_sym_AMP_GT] = ACTIONS(1308),
    [anon_sym_AMP_GT_GT] = ACTIONS(1308),
    [anon_sym_LT_AMP] = ACTIONS(1308),
    [anon_sym_GT_AMP] = ACTIONS(1308),
    [anon_sym_GT_PIPE] = ACTIONS(1308),
    [anon_sym_LT_AMP_DASH] = ACTIONS(1308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(1308),
    [anon_sym_LT_LT_DASH] = ACTIONS(1308),
    [aux_sym_heredoc_redirect_token1] = ACTIONS(1310),
    [anon_sym_LT_LT_LT] = ACTIONS(1308),
    [anon_sym_QMARK] = ACTIONS(1308),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(1308),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(1308),
    [aux_sym_concatenation_token1] = ACTIONS(1308),
    [anon_sym_DOLLAR] = ACTIONS(1308),
    [sym__special_character] = ACTIONS(1308),
    [anon_sym_DQUOTE] = ACTIONS(1308),
    [sym_raw_string] = ACTIONS(1308),
    [sym_ansi_c_string] = ACTIONS(1308),
    [aux_sym_number_token1] = ACTIONS(1308),
    [aux_sym_number_token2] = ACTIONS(1308),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(1308),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1308),
    [anon_sym_BQUOTE] = ACTIONS(1308),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(1308),
    [anon_sym_LT_LPAREN] = ACTIONS(1308),
    [anon_sym_GT_LPAREN] = ACTIONS(1308),
    [sym_comment] = ACTIONS(3),
    [sym_file_descriptor] = ACTIONS(1310),
    [sym__concat] = ACTIONS(1310),
    [sym_test_operator] = ACTIONS(1310),
    [sym__bare_dollar] = ACTIONS(1310),
    [sym__brace_start] = ACTIONS(1310),
  },
  [479] = {
    [sym_word] = ACTIONS(1312),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(1312),
    [anon_sym_SEMI] = ACTIONS(1312),
    [anon_sym_EQ] = ACTIONS(1312),
    [anon_sym_PLUS_PLUS] = ACTIONS(1312),
    [anon_sym_DASH_DASH] = ACTIONS(1312),
    [anon_sym_PLUS_EQ] = ACTIONS(1312),
    [anon_sym_DASH_EQ] = ACTIONS(1312),
    [anon_sym_STAR_EQ] = ACTIONS(1312),
    [anon_sym_SLASH_EQ] = ACTIONS(1312),
    [anon_sym_PERCENT_EQ] = ACTIONS(1312),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(1312),
    [anon_sym_LT_LT_EQ] = ACTIONS(1312),
    [anon_sym_GT_GT_EQ] = ACTIONS(1312),
    [anon_sym_AMP_EQ] = ACTIONS(1312),
    [anon_sym_CARET_EQ] = ACTIONS(1312),
    [anon_sym_PIPE_EQ] = ACTIONS(1312),
    [anon_sym_PIPE_PIPE] = ACTIONS(1312),
    [anon_sym_AMP_AMP] = ACTIONS(1312),
    [anon_sym_PIPE] = ACTIONS(1312),
    [anon_sym_CARET] = ACTIONS(1312),
    [anon_sym_AMP] = ACTIONS(1312),
    [anon_sym_EQ_EQ] = ACTIONS(1312),
    [anon_sym_BANG_EQ] = ACTIONS(1312),
    [anon_sym_LT] = ACTIONS(1312),
    [anon_sym_GT] = ACTIONS(1312),
    [anon_sym_LT_EQ] = ACTIONS(1312),
    [anon_sym_GT_EQ] = ACTIONS(1312),
    [anon_sym_LT_LT] = ACTIONS(1312),
    [anon_sym_GT_GT] = ACTIONS(1312),
    [anon_sym_PLUS] = ACTIONS(1312),
    [anon_sym_DASH] = ACTIONS(1312),
    [anon_sym_STAR] = ACTIONS(1312),
    [anon_sym_SLASH] = ACTIONS(1312),
    [anon_sym_PERCENT] = ACTIONS(1312),
    [anon_sym_STAR_STAR] = ACTIONS(1312),
    [anon_sym_LPAREN] = ACTIONS(1312),
    [anon_sym_RPAREN] = ACTIONS(1312),
    [anon_sym_SEMI_SEMI] = ACTIONS(1312),
    [anon_sym_PIPE_AMP] = ACTIONS(1312),
    [anon_sym_EQ_TILDE] = ACTIONS(1312),
    [anon_sym_AMP_GT] = ACTIONS(1312),
    [anon_sym_AMP_GT_GT] = ACTIONS(1312),
    [anon_sym_LT_AMP] = ACTIONS(1312),
    [anon_sym_GT_AMP] = ACTIONS(1312),
    [anon_sym_GT_PIPE] = ACTIONS(1312),
    [anon_sym_LT_AMP_DASH] = ACTIONS(1312),
    [anon_sym_GT_AMP_DASH] = ACTIONS(1312),
    [anon_sym_LT_LT_DASH] = ACTIONS(1312),
    [aux_sym_heredoc_redirect_token1] = ACTIONS(1314),
    [anon_sym_LT_LT_LT] = ACTIONS(1312),
    [anon_sym_QMARK] = ACTIONS(1312),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(1312),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(1312),
    [aux_sym_concatenation_token1] = ACTIONS(1312),
    [anon_sym_DOLLAR] = ACTIONS(1312),
    [sym__special_character] = ACTIONS(1312),
    [anon_sym_DQUOTE] = ACTIONS(1312),
    [sym_raw_string] = ACTIONS(1312),
    [sym_ansi_c_string] = ACTIONS(1312),
    [aux_sym_number_token1] = ACTIONS(1312),
    [aux_sym_number_token2] = ACTIONS(1312),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(1312),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1312),
    [anon_sym_BQUOTE] = ACTIONS(1312),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(1312),
    [anon_sym_LT_LPAREN] = ACTIONS(1312),
    [anon_sym_GT_LPAREN] = ACTIONS(1312),
    [sym_comment] = ACTIONS(3),
    [sym_file_descriptor] = ACTIONS(1314),
    [sym__concat] = ACTIONS(1314),
    [sym_test_operator] = ACTIONS(1314),
    [sym__bare_dollar] = ACTIONS(1314),
    [sym__brace_start] = ACTIONS(1314),
  },
  [480] = {
    [sym_word] = ACTIONS(1316),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(1316),
    [anon_sym_SEMI] = ACTIONS(1316),
    [anon_sym_EQ] = ACTIONS(1316),
    [anon_sym_PLUS_PLUS] = ACTIONS(1316),
    [anon_sym_DASH_DASH] = ACTIONS(1316),
    [anon_sym_PLUS_EQ] = ACTIONS(1316),
    [anon_sym_DASH_EQ] = ACTIONS(1316),
    [anon_sym_STAR_EQ] = ACTIONS(1316),
    [anon_sym_SLASH_EQ] = ACTIONS(1316),
    [anon_sym_PERCENT_EQ] = ACTIONS(1316),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(1316),
    [anon_sym_LT_LT_EQ] = ACTIONS(1316),
    [anon_sym_GT_GT_EQ] = ACTIONS(1316),
    [anon_sym_AMP_EQ] = ACTIONS(1316),
    [anon_sym_CARET_EQ] = ACTIONS(1316),
    [anon_sym_PIPE_EQ] = ACTIONS(1316),
    [anon_sym_PIPE_PIPE] = ACTIONS(1316),
    [anon_sym_AMP_AMP] = ACTIONS(1316),
    [anon_sym_PIPE] = ACTIONS(1316),
    [anon_sym_CARET] = ACTIONS(1316),
    [anon_sym_AMP] = ACTIONS(1316),
    [anon_sym_EQ_EQ] = ACTIONS(1316),
    [anon_sym_BANG_EQ] = ACTIONS(1316),
    [anon_sym_LT] = ACTIONS(1316),
    [anon_sym_GT] = ACTIONS(1316),
    [anon_sym_LT_EQ] = ACTIONS(1316),
    [anon_sym_GT_EQ] = ACTIONS(1316),
    [anon_sym_LT_LT] = ACTIONS(1316),
    [anon_sym_GT_GT] = ACTIONS(1316),
    [anon_sym_PLUS] = ACTIONS(1316),
    [anon_sym_DASH] = ACTIONS(1316),
    [anon_sym_STAR] = ACTIONS(1316),
    [anon_sym_SLASH] = ACTIONS(1316),
    [anon_sym_PERCENT] = ACTIONS(1316),
    [anon_sym_STAR_STAR] = ACTIONS(1316),
    [anon_sym_LPAREN] = ACTIONS(1316),
    [anon_sym_RPAREN] = ACTIONS(1316),
    [anon_sym_SEMI_SEMI] = ACTIONS(1316),
    [anon_sym_PIPE_AMP] = ACTIONS(1316),
    [anon_sym_EQ_TILDE] = ACTIONS(1316),
    [anon_sym_AMP_GT] = ACTIONS(1316),
    [anon_sym_AMP_GT_GT] = ACTIONS(1316),
    [anon_sym_LT_AMP] = ACTIONS(1316),
    [anon_sym_GT_AMP] = ACTIONS(1316),
    [anon_sym_GT_PIPE] = ACTIONS(1316),
    [anon_sym_LT_AMP_DASH] = ACTIONS(1316),
    [anon_sym_GT_AMP_DASH] = ACTIONS(1316),
    [anon_sym_LT_LT_DASH] = ACTIONS(1316),
    [aux_sym_heredoc_redirect_token1] = ACTIONS(1318),
    [anon_sym_LT_LT_LT] = ACTIONS(1316),
    [anon_sym_QMARK] = ACTIONS(1316),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(1316),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(1316),
    [aux_sym_concatenation_token1] = ACTIONS(1316),
    [anon_sym_DOLLAR] = ACTIONS(1316),
    [sym__special_character] = ACTIONS(1316),
    [anon_sym_DQUOTE] = ACTIONS(1316),
    [sym_raw_string] = ACTIONS(1316),
    [sym_ansi_c_string] = ACTIONS(1316),
    [aux_sym_number_token1] = ACTIONS(1316),
    [aux_sym_number_token2] = ACTIONS(1316),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(1316),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1316),
    [anon_sym_BQUOTE] = ACTIONS(1316),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(1316),
    [anon_sym_LT_LPAREN] = ACTIONS(1316),
    [anon_sym_GT_LPAREN] = ACTIONS(1316),
    [sym_comment] = ACTIONS(3),
    [sym_file_descriptor] = ACTIONS(1318),
    [sym__concat] = ACTIONS(1318),
    [sym_test_operator] = ACTIONS(1318),
    [sym__bare_dollar] = ACTIONS(1318),
    [sym__brace_start] = ACTIONS(1318),
  },
  [481] = {
    [sym_word] = ACTIONS(1320),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(1320),
    [anon_sym_SEMI] = ACTIONS(1320),
    [anon_sym_EQ] = ACTIONS(1320),
    [anon_sym_PLUS_PLUS] = ACTIONS(1320),
    [anon_sym_DASH_DASH] = ACTIONS(1320),
    [anon_sym_PLUS_EQ] = ACTIONS(1320),
    [anon_sym_DASH_EQ] = ACTIONS(1320),
    [anon_sym_STAR_EQ] = ACTIONS(1320),
    [anon_sym_SLASH_EQ] = ACTIONS(1320),
    [anon_sym_PERCENT_EQ] = ACTIONS(1320),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(1320),
    [anon_sym_LT_LT_EQ] = ACTIONS(1320),
    [anon_sym_GT_GT_EQ] = ACTIONS(1320),
    [anon_sym_AMP_EQ] = ACTIONS(1320),
    [anon_sym_CARET_EQ] = ACTIONS(1320),
    [anon_sym_PIPE_EQ] = ACTIONS(1320),
    [anon_sym_PIPE_PIPE] = ACTIONS(1320),
    [anon_sym_AMP_AMP] = ACTIONS(1320),
    [anon_sym_PIPE] = ACTIONS(1320),
    [anon_sym_CARET] = ACTIONS(1320),
    [anon_sym_AMP] = ACTIONS(1320),
    [anon_sym_EQ_EQ] = ACTIONS(1320),
    [anon_sym_BANG_EQ] = ACTIONS(1320),
    [anon_sym_LT] = ACTIONS(1320),
    [anon_sym_GT] = ACTIONS(1320),
    [anon_sym_LT_EQ] = ACTIONS(1320),
    [anon_sym_GT_EQ] = ACTIONS(1320),
    [anon_sym_LT_LT] = ACTIONS(1320),
    [anon_sym_GT_GT] = ACTIONS(1320),
    [anon_sym_PLUS] = ACTIONS(1320),
    [anon_sym_DASH] = ACTIONS(1320),
    [anon_sym_STAR] = ACTIONS(1320),
    [anon_sym_SLASH] = ACTIONS(1320),
    [anon_sym_PERCENT] = ACTIONS(1320),
    [anon_sym_STAR_STAR] = ACTIONS(1320),
    [anon_sym_LPAREN] = ACTIONS(1320),
    [anon_sym_RPAREN] = ACTIONS(1320),
    [anon_sym_SEMI_SEMI] = ACTIONS(1320),
    [anon_sym_PIPE_AMP] = ACTIONS(1320),
    [anon_sym_EQ_TILDE] = ACTIONS(1320),
    [anon_sym_AMP_GT] = ACTIONS(1320),
    [anon_sym_AMP_GT_GT] = ACTIONS(1320),
    [anon_sym_LT_AMP] = ACTIONS(1320),
    [anon_sym_GT_AMP] = ACTIONS(1320),
    [anon_sym_GT_PIPE] = ACTIONS(1320),
    [anon_sym_LT_AMP_DASH] = ACTIONS(1320),
    [anon_sym_GT_AMP_DASH] = ACTIONS(1320),
    [anon_sym_LT_LT_DASH] = ACTIONS(1320),
    [aux_sym_heredoc_redirect_token1] = ACTIONS(1322),
    [anon_sym_LT_LT_LT] = ACTIONS(1320),
    [anon_sym_QMARK] = ACTIONS(1320),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(1320),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(1320),
    [aux_sym_concatenation_token1] = ACTIONS(1320),
    [anon_sym_DOLLAR] = ACTIONS(1320),
    [sym__special_character] = ACTIONS(1320),
    [anon_sym_DQUOTE] = ACTIONS(1320),
    [sym_raw_string] = ACTIONS(1320),
    [sym_ansi_c_string] = ACTIONS(1320),
    [aux_sym_number_token1] = ACTIONS(1320),
    [aux_sym_number_token2] = ACTIONS(1320),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(1320),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1320),
    [anon_sym_BQUOTE] = ACTIONS(1320),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(1320),
    [anon_sym_LT_LPAREN] = ACTIONS(1320),
    [anon_sym_GT_LPAREN] = ACTIONS(1320),
    [sym_comment] = ACTIONS(3),
    [sym_file_descriptor] = ACTIONS(1322),
    [sym__concat] = ACTIONS(1322),
    [sym_test_operator] = ACTIONS(1322),
    [sym__bare_dollar] = ACTIONS(1322),
    [sym__brace_start] = ACTIONS(1322),
  },
  [482] = {
    [sym_word] = ACTIONS(1251),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(1251),
    [anon_sym_SEMI] = ACTIONS(1251),
    [anon_sym_EQ] = ACTIONS(1251),
    [anon_sym_PLUS_PLUS] = ACTIONS(1251),
    [anon_sym_DASH_DASH] = ACTIONS(1251),
    [anon_sym_PLUS_EQ] = ACTIONS(1251),
    [anon_sym_DASH_EQ] = ACTIONS(1251),
    [anon_sym_STAR_EQ] = ACTIONS(1251),
    [anon_sym_SLASH_EQ] = ACTIONS(1251),
    [anon_sym_PERCENT_EQ] = ACTIONS(1251),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(1251),
    [anon_sym_LT_LT_EQ] = ACTIONS(1251),
    [anon_sym_GT_GT_EQ] = ACTIONS(1251),
    [anon_sym_AMP_EQ] = ACTIONS(1251),
    [anon_sym_CARET_EQ] = ACTIONS(1251),
    [anon_sym_PIPE_EQ] = ACTIONS(1251),
    [anon_sym_PIPE_PIPE] = ACTIONS(1251),
    [anon_sym_AMP_AMP] = ACTIONS(1251),
    [anon_sym_PIPE] = ACTIONS(1251),
    [anon_sym_CARET] = ACTIONS(1251),
    [anon_sym_AMP] = ACTIONS(1251),
    [anon_sym_EQ_EQ] = ACTIONS(1251),
    [anon_sym_BANG_EQ] = ACTIONS(1251),
    [anon_sym_LT] = ACTIONS(1251),
    [anon_sym_GT] = ACTIONS(1251),
    [anon_sym_LT_EQ] = ACTIONS(1251),
    [anon_sym_GT_EQ] = ACTIONS(1251),
    [anon_sym_LT_LT] = ACTIONS(1251),
    [anon_sym_GT_GT] = ACTIONS(1251),
    [anon_sym_PLUS] = ACTIONS(1251),
    [anon_sym_DASH] = ACTIONS(1251),
    [anon_sym_STAR] = ACTIONS(1251),
    [anon_sym_SLASH] = ACTIONS(1251),
    [anon_sym_PERCENT] = ACTIONS(1251),
    [anon_sym_STAR_STAR] = ACTIONS(1251),
    [anon_sym_LPAREN] = ACTIONS(1251),
    [anon_sym_RPAREN] = ACTIONS(1251),
    [anon_sym_SEMI_SEMI] = ACTIONS(1251),
    [anon_sym_PIPE_AMP] = ACTIONS(1251),
    [anon_sym_EQ_TILDE] = ACTIONS(1251),
    [anon_sym_AMP_GT] = ACTIONS(1251),
    [anon_sym_AMP_GT_GT] = ACTIONS(1251),
    [anon_sym_LT_AMP] = ACTIONS(1251),
    [anon_sym_GT_AMP] = ACTIONS(1251),
    [anon_sym_GT_PIPE] = ACTIONS(1251),
    [anon_sym_LT_AMP_DASH] = ACTIONS(1251),
    [anon_sym_GT_AMP_DASH] = ACTIONS(1251),
    [anon_sym_LT_LT_DASH] = ACTIONS(1251),
    [aux_sym_heredoc_redirect_token1] = ACTIONS(1253),
    [anon_sym_LT_LT_LT] = ACTIONS(1251),
    [anon_sym_QMARK] = ACTIONS(1251),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(1251),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(1251),
    [aux_sym_concatenation_token1] = ACTIONS(1251),
    [anon_sym_DOLLAR] = ACTIONS(1251),
    [sym__special_character] = ACTIONS(1251),
    [anon_sym_DQUOTE] = ACTIONS(1251),
    [sym_raw_string] = ACTIONS(1251),
    [sym_ansi_c_string] = ACTIONS(1251),
    [aux_sym_number_token1] = ACTIONS(1251),
    [aux_sym_number_token2] = ACTIONS(1251),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(1251),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1251),
    [anon_sym_BQUOTE] = ACTIONS(1251),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(1251),
    [anon_sym_LT_LPAREN] = ACTIONS(1251),
    [anon_sym_GT_LPAREN] = ACTIONS(1251),
    [sym_comment] = ACTIONS(3),
    [sym_file_descriptor] = ACTIONS(1253),
    [sym__concat] = ACTIONS(1253),
    [sym_test_operator] = ACTIONS(1253),
    [sym__bare_dollar] = ACTIONS(1253),
    [sym__brace_start] = ACTIONS(1253),
  },
  [483] = {
    [sym_word] = ACTIONS(1324),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(1324),
    [anon_sym_SEMI] = ACTIONS(1324),
    [anon_sym_EQ] = ACTIONS(1324),
    [anon_sym_PLUS_PLUS] = ACTIONS(1324),
    [anon_sym_DASH_DASH] = ACTIONS(1324),
    [anon_sym_PLUS_EQ] = ACTIONS(1324),
    [anon_sym_DASH_EQ] = ACTIONS(1324),
    [anon_sym_STAR_EQ] = ACTIONS(1324),
    [anon_sym_SLASH_EQ] = ACTIONS(1324),
    [anon_sym_PERCENT_EQ] = ACTIONS(1324),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(1324),
    [anon_sym_LT_LT_EQ] = ACTIONS(1324),
    [anon_sym_GT_GT_EQ] = ACTIONS(1324),
    [anon_sym_AMP_EQ] = ACTIONS(1324),
    [anon_sym_CARET_EQ] = ACTIONS(1324),
    [anon_sym_PIPE_EQ] = ACTIONS(1324),
    [anon_sym_PIPE_PIPE] = ACTIONS(1324),
    [anon_sym_AMP_AMP] = ACTIONS(1324),
    [anon_sym_PIPE] = ACTIONS(1324),
    [anon_sym_CARET] = ACTIONS(1324),
    [anon_sym_AMP] = ACTIONS(1324),
    [anon_sym_EQ_EQ] = ACTIONS(1324),
    [anon_sym_BANG_EQ] = ACTIONS(1324),
    [anon_sym_LT] = ACTIONS(1324),
    [anon_sym_GT] = ACTIONS(1324),
    [anon_sym_LT_EQ] = ACTIONS(1324),
    [anon_sym_GT_EQ] = ACTIONS(1324),
    [anon_sym_LT_LT] = ACTIONS(1324),
    [anon_sym_GT_GT] = ACTIONS(1324),
    [anon_sym_PLUS] = ACTIONS(1324),
    [anon_sym_DASH] = ACTIONS(1324),
    [anon_sym_STAR] = ACTIONS(1324),
    [anon_sym_SLASH] = ACTIONS(1324),
    [anon_sym_PERCENT] = ACTIONS(1324),
    [anon_sym_STAR_STAR] = ACTIONS(1324),
    [anon_sym_LPAREN] = ACTIONS(1324),
    [anon_sym_RPAREN] = ACTIONS(1324),
    [anon_sym_SEMI_SEMI] = ACTIONS(1324),
    [anon_sym_PIPE_AMP] = ACTIONS(1324),
    [anon_sym_EQ_TILDE] = ACTIONS(1324),
    [anon_sym_AMP_GT] = ACTIONS(1324),
    [anon_sym_AMP_GT_GT] = ACTIONS(1324),
    [anon_sym_LT_AMP] = ACTIONS(1324),
    [anon_sym_GT_AMP] = ACTIONS(1324),
    [anon_sym_GT_PIPE] = ACTIONS(1324),
    [anon_sym_LT_AMP_DASH] = ACTIONS(1324),
    [anon_sym_GT_AMP_DASH] = ACTIONS(1324),
    [anon_sym_LT_LT_DASH] = ACTIONS(1324),
    [aux_sym_heredoc_redirect_token1] = ACTIONS(1326),
    [anon_sym_LT_LT_LT] = ACTIONS(1324),
    [anon_sym_QMARK] = ACTIONS(1324),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(1324),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(1324),
    [aux_sym_concatenation_token1] = ACTIONS(1324),
    [anon_sym_DOLLAR] = ACTIONS(1324),
    [sym__special_character] = ACTIONS(1324),
    [anon_sym_DQUOTE] = ACTIONS(1324),
    [sym_raw_string] = ACTIONS(1324),
    [sym_ansi_c_string] = ACTIONS(1324),
    [aux_sym_number_token1] = ACTIONS(1324),
    [aux_sym_number_token2] = ACTIONS(1324),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(1324),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1324),
    [anon_sym_BQUOTE] = ACTIONS(1324),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(1324),
    [anon_sym_LT_LPAREN] = ACTIONS(1324),
    [anon_sym_GT_LPAREN] = ACTIONS(1324),
    [sym_comment] = ACTIONS(3),
    [sym_file_descriptor] = ACTIONS(1326),
    [sym__concat] = ACTIONS(1326),
    [sym_test_operator] = ACTIONS(1326),
    [sym__bare_dollar] = ACTIONS(1326),
    [sym__brace_start] = ACTIONS(1326),
  },
  [484] = {
    [sym_word] = ACTIONS(1328),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(1328),
    [anon_sym_SEMI] = ACTIONS(1328),
    [anon_sym_EQ] = ACTIONS(1328),
    [anon_sym_PLUS_PLUS] = ACTIONS(1328),
    [anon_sym_DASH_DASH] = ACTIONS(1328),
    [anon_sym_PLUS_EQ] = ACTIONS(1328),
    [anon_sym_DASH_EQ] = ACTIONS(1328),
    [anon_sym_STAR_EQ] = ACTIONS(1328),
    [anon_sym_SLASH_EQ] = ACTIONS(1328),
    [anon_sym_PERCENT_EQ] = ACTIONS(1328),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(1328),
    [anon_sym_LT_LT_EQ] = ACTIONS(1328),
    [anon_sym_GT_GT_EQ] = ACTIONS(1328),
    [anon_sym_AMP_EQ] = ACTIONS(1328),
    [anon_sym_CARET_EQ] = ACTIONS(1328),
    [anon_sym_PIPE_EQ] = ACTIONS(1328),
    [anon_sym_PIPE_PIPE] = ACTIONS(1328),
    [anon_sym_AMP_AMP] = ACTIONS(1328),
    [anon_sym_PIPE] = ACTIONS(1328),
    [anon_sym_CARET] = ACTIONS(1328),
    [anon_sym_AMP] = ACTIONS(1328),
    [anon_sym_EQ_EQ] = ACTIONS(1328),
    [anon_sym_BANG_EQ] = ACTIONS(1328),
    [anon_sym_LT] = ACTIONS(1328),
    [anon_sym_GT] = ACTIONS(1328),
    [anon_sym_LT_EQ] = ACTIONS(1328),
    [anon_sym_GT_EQ] = ACTIONS(1328),
    [anon_sym_LT_LT] = ACTIONS(1328),
    [anon_sym_GT_GT] = ACTIONS(1328),
    [anon_sym_PLUS] = ACTIONS(1328),
    [anon_sym_DASH] = ACTIONS(1328),
    [anon_sym_STAR] = ACTIONS(1328),
    [anon_sym_SLASH] = ACTIONS(1328),
    [anon_sym_PERCENT] = ACTIONS(1328),
    [anon_sym_STAR_STAR] = ACTIONS(1328),
    [anon_sym_LPAREN] = ACTIONS(1328),
    [anon_sym_RPAREN] = ACTIONS(1328),
    [anon_sym_SEMI_SEMI] = ACTIONS(1328),
    [anon_sym_PIPE_AMP] = ACTIONS(1328),
    [anon_sym_EQ_TILDE] = ACTIONS(1328),
    [anon_sym_AMP_GT] = ACTIONS(1328),
    [anon_sym_AMP_GT_GT] = ACTIONS(1328),
    [anon_sym_LT_AMP] = ACTIONS(1328),
    [anon_sym_GT_AMP] = ACTIONS(1328),
    [anon_sym_GT_PIPE] = ACTIONS(1328),
    [anon_sym_LT_AMP_DASH] = ACTIONS(1328),
    [anon_sym_GT_AMP_DASH] = ACTIONS(1328),
    [anon_sym_LT_LT_DASH] = ACTIONS(1328),
    [aux_sym_heredoc_redirect_token1] = ACTIONS(1330),
    [anon_sym_LT_LT_LT] = ACTIONS(1328),
    [anon_sym_QMARK] = ACTIONS(1328),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(1328),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(1328),
    [aux_sym_concatenation_token1] = ACTIONS(1328),
    [anon_sym_DOLLAR] = ACTIONS(1328),
    [sym__special_character] = ACTIONS(1328),
    [anon_sym_DQUOTE] = ACTIONS(1328),
    [sym_raw_string] = ACTIONS(1328),
    [sym_ansi_c_string] = ACTIONS(1328),
    [aux_sym_number_token1] = ACTIONS(1328),
    [aux_sym_number_token2] = ACTIONS(1328),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(1328),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1328),
    [anon_sym_BQUOTE] = ACTIONS(1328),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(1328),
    [anon_sym_LT_LPAREN] = ACTIONS(1328),
    [anon_sym_GT_LPAREN] = ACTIONS(1328),
    [sym_comment] = ACTIONS(3),
    [sym_file_descriptor] = ACTIONS(1330),
    [sym__concat] = ACTIONS(1330),
    [sym_test_operator] = ACTIONS(1330),
    [sym__bare_dollar] = ACTIONS(1330),
    [sym__brace_start] = ACTIONS(1330),
  },
  [485] = {
    [sym_word] = ACTIONS(1332),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(1332),
    [anon_sym_SEMI] = ACTIONS(1332),
    [anon_sym_EQ] = ACTIONS(1332),
    [anon_sym_PLUS_PLUS] = ACTIONS(1332),
    [anon_sym_DASH_DASH] = ACTIONS(1332),
    [anon_sym_PLUS_EQ] = ACTIONS(1332),
    [anon_sym_DASH_EQ] = ACTIONS(1332),
    [anon_sym_STAR_EQ] = ACTIONS(1332),
    [anon_sym_SLASH_EQ] = ACTIONS(1332),
    [anon_sym_PERCENT_EQ] = ACTIONS(1332),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(1332),
    [anon_sym_LT_LT_EQ] = ACTIONS(1332),
    [anon_sym_GT_GT_EQ] = ACTIONS(1332),
    [anon_sym_AMP_EQ] = ACTIONS(1332),
    [anon_sym_CARET_EQ] = ACTIONS(1332),
    [anon_sym_PIPE_EQ] = ACTIONS(1332),
    [anon_sym_PIPE_PIPE] = ACTIONS(1332),
    [anon_sym_AMP_AMP] = ACTIONS(1332),
    [anon_sym_PIPE] = ACTIONS(1332),
    [anon_sym_CARET] = ACTIONS(1332),
    [anon_sym_AMP] = ACTIONS(1332),
    [anon_sym_EQ_EQ] = ACTIONS(1332),
    [anon_sym_BANG_EQ] = ACTIONS(1332),
    [anon_sym_LT] = ACTIONS(1332),
    [anon_sym_GT] = ACTIONS(1332),
    [anon_sym_LT_EQ] = ACTIONS(1332),
    [anon_sym_GT_EQ] = ACTIONS(1332),
    [anon_sym_LT_LT] = ACTIONS(1332),
    [anon_sym_GT_GT] = ACTIONS(1332),
    [anon_sym_PLUS] = ACTIONS(1332),
    [anon_sym_DASH] = ACTIONS(1332),
    [anon_sym_STAR] = ACTIONS(1332),
    [anon_sym_SLASH] = ACTIONS(1332),
    [anon_sym_PERCENT] = ACTIONS(1332),
    [anon_sym_STAR_STAR] = ACTIONS(1332),
    [anon_sym_LPAREN] = ACTIONS(1332),
    [anon_sym_RPAREN] = ACTIONS(1332),
    [anon_sym_SEMI_SEMI] = ACTIONS(1332),
    [anon_sym_PIPE_AMP] = ACTIONS(1332),
    [anon_sym_EQ_TILDE] = ACTIONS(1332),
    [anon_sym_AMP_GT] = ACTIONS(1332),
    [anon_sym_AMP_GT_GT] = ACTIONS(1332),
    [anon_sym_LT_AMP] = ACTIONS(1332),
    [anon_sym_GT_AMP] = ACTIONS(1332),
    [anon_sym_GT_PIPE] = ACTIONS(1332),
    [anon_sym_LT_AMP_DASH] = ACTIONS(1332),
    [anon_sym_GT_AMP_DASH] = ACTIONS(1332),
    [anon_sym_LT_LT_DASH] = ACTIONS(1332),
    [aux_sym_heredoc_redirect_token1] = ACTIONS(1334),
    [anon_sym_LT_LT_LT] = ACTIONS(1332),
    [anon_sym_QMARK] = ACTIONS(1332),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(1332),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(1332),
    [aux_sym_concatenation_token1] = ACTIONS(1332),
    [anon_sym_DOLLAR] = ACTIONS(1332),
    [sym__special_character] = ACTIONS(1332),
    [anon_sym_DQUOTE] = ACTIONS(1332),
    [sym_raw_string] = ACTIONS(1332),
    [sym_ansi_c_string] = ACTIONS(1332),
    [aux_sym_number_token1] = ACTIONS(1332),
    [aux_sym_number_token2] = ACTIONS(1332),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(1332),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1332),
    [anon_sym_BQUOTE] = ACTIONS(1332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(1332),
    [anon_sym_LT_LPAREN] = ACTIONS(1332),
    [anon_sym_GT_LPAREN] = ACTIONS(1332),
    [sym_comment] = ACTIONS(3),
    [sym_file_descriptor] = ACTIONS(1334),
    [sym__concat] = ACTIONS(1334),
    [sym_test_operator] = ACTIONS(1334),
    [sym__bare_dollar] = ACTIONS(1334),
    [sym__brace_start] = ACTIONS(1334),
  },
  [486] = {
    [sym_word] = ACTIONS(1336),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(1336),
    [anon_sym_SEMI] = ACTIONS(1336),
    [anon_sym_EQ] = ACTIONS(1336),
    [anon_sym_PLUS_PLUS] = ACTIONS(1336),
    [anon_sym_DASH_DASH] = ACTIONS(1336),
    [anon_sym_PLUS_EQ] = ACTIONS(1336),
    [anon_sym_DASH_EQ] = ACTIONS(1336),
    [anon_sym_STAR_EQ] = ACTIONS(1336),
    [anon_sym_SLASH_EQ] = ACTIONS(1336),
    [anon_sym_PERCENT_EQ] = ACTIONS(1336),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(1336),
    [anon_sym_LT_LT_EQ] = ACTIONS(1336),
    [anon_sym_GT_GT_EQ] = ACTIONS(1336),
    [anon_sym_AMP_EQ] = ACTIONS(1336),
    [anon_sym_CARET_EQ] = ACTIONS(1336),
    [anon_sym_PIPE_EQ] = ACTIONS(1336),
    [anon_sym_PIPE_PIPE] = ACTIONS(1336),
    [anon_sym_AMP_AMP] = ACTIONS(1336),
    [anon_sym_PIPE] = ACTIONS(1336),
    [anon_sym_CARET] = ACTIONS(1336),
    [anon_sym_AMP] = ACTIONS(1336),
    [anon_sym_EQ_EQ] = ACTIONS(1336),
    [anon_sym_BANG_EQ] = ACTIONS(1336),
    [anon_sym_LT] = ACTIONS(1336),
    [anon_sym_GT] = ACTIONS(1336),
    [anon_sym_LT_EQ] = ACTIONS(1336),
    [anon_sym_GT_EQ] = ACTIONS(1336),
    [anon_sym_LT_LT] = ACTIONS(1336),
    [anon_sym_GT_GT] = ACTIONS(1336),
    [anon_sym_PLUS] = ACTIONS(1336),
    [anon_sym_DASH] = ACTIONS(1336),
    [anon_sym_STAR] = ACTIONS(1336),
    [anon_sym_SLASH] = ACTIONS(1336),
    [anon_sym_PERCENT] = ACTIONS(1336),
    [anon_sym_STAR_STAR] = ACTIONS(1336),
    [anon_sym_LPAREN] = ACTIONS(1336),
    [anon_sym_RPAREN] = ACTIONS(1336),
    [anon_sym_SEMI_SEMI] = ACTIONS(1336),
    [anon_sym_PIPE_AMP] = ACTIONS(1336),
    [anon_sym_EQ_TILDE] = ACTIONS(1336),
    [anon_sym_AMP_GT] = ACTIONS(1336),
    [anon_sym_AMP_GT_GT] = ACTIONS(1336),
    [anon_sym_LT_AMP] = ACTIONS(1336),
    [anon_sym_GT_AMP] = ACTIONS(1336),
    [anon_sym_GT_PIPE] = ACTIONS(1336),
    [anon_sym_LT_AMP_DASH] = ACTIONS(1336),
    [anon_sym_GT_AMP_DASH] = ACTIONS(1336),
    [anon_sym_LT_LT_DASH] = ACTIONS(1336),
    [aux_sym_heredoc_redirect_token1] = ACTIONS(1338),
    [anon_sym_LT_LT_LT] = ACTIONS(1336),
    [anon_sym_QMARK] = ACTIONS(1336),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(1336),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(1336),
    [aux_sym_concatenation_token1] = ACTIONS(1336),
    [anon_sym_DOLLAR] = ACTIONS(1336),
    [sym__special_character] = ACTIONS(1336),
    [anon_sym_DQUOTE] = ACTIONS(1336),
    [sym_raw_string] = ACTIONS(1336),
    [sym_ansi_c_string] = ACTIONS(1336),
    [aux_sym_number_token1] = ACTIONS(1336),
    [aux_sym_number_token2] = ACTIONS(1336),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(1336),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1336),
    [anon_sym_BQUOTE] = ACTIONS(1336),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(1336),
    [anon_sym_LT_LPAREN] = ACTIONS(1336),
    [anon_sym_GT_LPAREN] = ACTIONS(1336),
    [sym_comment] = ACTIONS(3),
    [sym_file_descriptor] = ACTIONS(1338),
    [sym__concat] = ACTIONS(1338),
    [sym_test_operator] = ACTIONS(1338),
    [sym__bare_dollar] = ACTIONS(1338),
    [sym__brace_start] = ACTIONS(1338),
  },
  [487] = {
    [sym_word] = ACTIONS(1340),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(1340),
    [anon_sym_SEMI] = ACTIONS(1340),
    [anon_sym_EQ] = ACTIONS(1340),
    [anon_sym_PLUS_PLUS] = ACTIONS(1340),
    [anon_sym_DASH_DASH] = ACTIONS(1340),
    [anon_sym_PLUS_EQ] = ACTIONS(1340),
    [anon_sym_DASH_EQ] = ACTIONS(1340),
    [anon_sym_STAR_EQ] = ACTIONS(1340),
    [anon_sym_SLASH_EQ] = ACTIONS(1340),
    [anon_sym_PERCENT_EQ] = ACTIONS(1340),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(1340),
    [anon_sym_LT_LT_EQ] = ACTIONS(1340),
    [anon_sym_GT_GT_EQ] = ACTIONS(1340),
    [anon_sym_AMP_EQ] = ACTIONS(1340),
    [anon_sym_CARET_EQ] = ACTIONS(1340),
    [anon_sym_PIPE_EQ] = ACTIONS(1340),
    [anon_sym_PIPE_PIPE] = ACTIONS(1340),
    [anon_sym_AMP_AMP] = ACTIONS(1340),
    [anon_sym_PIPE] = ACTIONS(1340),
    [anon_sym_CARET] = ACTIONS(1340),
    [anon_sym_AMP] = ACTIONS(1340),
    [anon_sym_EQ_EQ] = ACTIONS(1340),
    [anon_sym_BANG_EQ] = ACTIONS(1340),
    [anon_sym_LT] = ACTIONS(1340),
    [anon_sym_GT] = ACTIONS(1340),
    [anon_sym_LT_EQ] = ACTIONS(1340),
    [anon_sym_GT_EQ] = ACTIONS(1340),
    [anon_sym_LT_LT] = ACTIONS(1340),
    [anon_sym_GT_GT] = ACTIONS(1340),
    [anon_sym_PLUS] = ACTIONS(1340),
    [anon_sym_DASH] = ACTIONS(1340),
    [anon_sym_STAR] = ACTIONS(1340),
    [anon_sym_SLASH] = ACTIONS(1340),
    [anon_sym_PERCENT] = ACTIONS(1340),
    [anon_sym_STAR_STAR] = ACTIONS(1340),
    [anon_sym_LPAREN] = ACTIONS(1340),
    [anon_sym_RPAREN] = ACTIONS(1340),
    [anon_sym_SEMI_SEMI] = ACTIONS(1340),
    [anon_sym_PIPE_AMP] = ACTIONS(1340),
    [anon_sym_EQ_TILDE] = ACTIONS(1340),
    [anon_sym_AMP_GT] = ACTIONS(1340),
    [anon_sym_AMP_GT_GT] = ACTIONS(1340),
    [anon_sym_LT_AMP] = ACTIONS(1340),
    [anon_sym_GT_AMP] = ACTIONS(1340),
    [anon_sym_GT_PIPE] = ACTIONS(1340),
    [anon_sym_LT_AMP_DASH] = ACTIONS(1340),
    [anon_sym_GT_AMP_DASH] = ACTIONS(1340),
    [anon_sym_LT_LT_DASH] = ACTIONS(1340),
    [aux_sym_heredoc_redirect_token1] = ACTIONS(1342),
    [anon_sym_LT_LT_LT] = ACTIONS(1340),
    [anon_sym_QMARK] = ACTIONS(1340),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(1340),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(1340),
    [aux_sym_concatenation_token1] = ACTIONS(1340),
    [anon_sym_DOLLAR] = ACTIONS(1340),
    [sym__special_character] = ACTIONS(1340),
    [anon_sym_DQUOTE] = ACTIONS(1340),
    [sym_raw_string] = ACTIONS(1340),
    [sym_ansi_c_string] = ACTIONS(1340),
    [aux_sym_number_token1] = ACTIONS(1340),
    [aux_sym_number_token2] = ACTIONS(1340),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(1340),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1340),
    [anon_sym_BQUOTE] = ACTIONS(1340),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(1340),
    [anon_sym_LT_LPAREN] = ACTIONS(1340),
    [anon_sym_GT_LPAREN] = ACTIONS(1340),
    [sym_comment] = ACTIONS(3),
    [sym_file_descriptor] = ACTIONS(1342),
    [sym__concat] = ACTIONS(1342),
    [sym_test_operator] = ACTIONS(1342),
    [sym__bare_dollar] = ACTIONS(1342),
    [sym__brace_start] = ACTIONS(1342),
  },
  [488] = {
    [sym_word] = ACTIONS(1344),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(1344),
    [anon_sym_SEMI] = ACTIONS(1344),
    [anon_sym_EQ] = ACTIONS(1344),
    [anon_sym_PLUS_PLUS] = ACTIONS(1344),
    [anon_sym_DASH_DASH] = ACTIONS(1344),
    [anon_sym_PLUS_EQ] = ACTIONS(1344),
    [anon_sym_DASH_EQ] = ACTIONS(1344),
    [anon_sym_STAR_EQ] = ACTIONS(1344),
    [anon_sym_SLASH_EQ] = ACTIONS(1344),
    [anon_sym_PERCENT_EQ] = ACTIONS(1344),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(1344),
    [anon_sym_LT_LT_EQ] = ACTIONS(1344),
    [anon_sym_GT_GT_EQ] = ACTIONS(1344),
    [anon_sym_AMP_EQ] = ACTIONS(1344),
    [anon_sym_CARET_EQ] = ACTIONS(1344),
    [anon_sym_PIPE_EQ] = ACTIONS(1344),
    [anon_sym_PIPE_PIPE] = ACTIONS(1344),
    [anon_sym_AMP_AMP] = ACTIONS(1344),
    [anon_sym_PIPE] = ACTIONS(1344),
    [anon_sym_CARET] = ACTIONS(1344),
    [anon_sym_AMP] = ACTIONS(1344),
    [anon_sym_EQ_EQ] = ACTIONS(1344),
    [anon_sym_BANG_EQ] = ACTIONS(1344),
    [anon_sym_LT] = ACTIONS(1344),
    [anon_sym_GT] = ACTIONS(1344),
    [anon_sym_LT_EQ] = ACTIONS(1344),
    [anon_sym_GT_EQ] = ACTIONS(1344),
    [anon_sym_LT_LT] = ACTIONS(1344),
    [anon_sym_GT_GT] = ACTIONS(1344),
    [anon_sym_PLUS] = ACTIONS(1344),
    [anon_sym_DASH] = ACTIONS(1344),
    [anon_sym_STAR] = ACTIONS(1344),
    [anon_sym_SLASH] = ACTIONS(1344),
    [anon_sym_PERCENT] = ACTIONS(1344),
    [anon_sym_STAR_STAR] = ACTIONS(1344),
    [anon_sym_LPAREN] = ACTIONS(1344),
    [anon_sym_RPAREN] = ACTIONS(1344),
    [anon_sym_SEMI_SEMI] = ACTIONS(1344),
    [anon_sym_PIPE_AMP] = ACTIONS(1344),
    [anon_sym_EQ_TILDE] = ACTIONS(1344),
    [anon_sym_AMP_GT] = ACTIONS(1344),
    [anon_sym_AMP_GT_GT] = ACTIONS(1344),
    [anon_sym_LT_AMP] = ACTIONS(1344),
    [anon_sym_GT_AMP] = ACTIONS(1344),
    [anon_sym_GT_PIPE] = ACTIONS(1344),
    [anon_sym_LT_AMP_DASH] = ACTIONS(1344),
    [anon_sym_GT_AMP_DASH] = ACTIONS(1344),
    [anon_sym_LT_LT_DASH] = ACTIONS(1344),
    [aux_sym_heredoc_redirect_token1] = ACTIONS(1346),
    [anon_sym_LT_LT_LT] = ACTIONS(1344),
    [anon_sym_QMARK] = ACTIONS(1344),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(1344),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(1344),
    [aux_sym_concatenation_token1] = ACTIONS(1344),
    [anon_sym_DOLLAR] = ACTIONS(1344),
    [sym__special_character] = ACTIONS(1344),
    [anon_sym_DQUOTE] = ACTIONS(1344),
    [sym_raw_string] = ACTIONS(1344),
    [sym_ansi_c_string] = ACTIONS(1344),
    [aux_sym_number_token1] = ACTIONS(1344),
    [aux_sym_number_token2] = ACTIONS(1344),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(1344),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1344),
    [anon_sym_BQUOTE] = ACTIONS(1344),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(1344),
    [anon_sym_LT_LPAREN] = ACTIONS(1344),
    [anon_sym_GT_LPAREN] = ACTIONS(1344),
    [sym_comment] = ACTIONS(3),
    [sym_file_descriptor] = ACTIONS(1346),
    [sym__concat] = ACTIONS(1346),
    [sym_test_operator] = ACTIONS(1346),
    [sym__bare_dollar] = ACTIONS(1346),
    [sym__brace_start] = ACTIONS(1346),
  },
  [489] = {
    [sym_word] = ACTIONS(1348),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(1348),
    [anon_sym_SEMI] = ACTIONS(1348),
    [anon_sym_EQ] = ACTIONS(1348),
    [anon_sym_PLUS_PLUS] = ACTIONS(1348),
    [anon_sym_DASH_DASH] = ACTIONS(1348),
    [anon_sym_PLUS_EQ] = ACTIONS(1348),
    [anon_sym_DASH_EQ] = ACTIONS(1348),
    [anon_sym_STAR_EQ] = ACTIONS(1348),
    [anon_sym_SLASH_EQ] = ACTIONS(1348),
    [anon_sym_PERCENT_EQ] = ACTIONS(1348),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(1348),
    [anon_sym_LT_LT_EQ] = ACTIONS(1348),
    [anon_sym_GT_GT_EQ] = ACTIONS(1348),
    [anon_sym_AMP_EQ] = ACTIONS(1348),
    [anon_sym_CARET_EQ] = ACTIONS(1348),
    [anon_sym_PIPE_EQ] = ACTIONS(1348),
    [anon_sym_PIPE_PIPE] = ACTIONS(1348),
    [anon_sym_AMP_AMP] = ACTIONS(1348),
    [anon_sym_PIPE] = ACTIONS(1348),
    [anon_sym_CARET] = ACTIONS(1348),
    [anon_sym_AMP] = ACTIONS(1348),
    [anon_sym_EQ_EQ] = ACTIONS(1348),
    [anon_sym_BANG_EQ] = ACTIONS(1348),
    [anon_sym_LT] = ACTIONS(1348),
    [anon_sym_GT] = ACTIONS(1348),
    [anon_sym_LT_EQ] = ACTIONS(1348),
    [anon_sym_GT_EQ] = ACTIONS(1348),
    [anon_sym_LT_LT] = ACTIONS(1348),
    [anon_sym_GT_GT] = ACTIONS(1348),
    [anon_sym_PLUS] = ACTIONS(1348),
    [anon_sym_DASH] = ACTIONS(1348),
    [anon_sym_STAR] = ACTIONS(1348),
    [anon_sym_SLASH] = ACTIONS(1348),
    [anon_sym_PERCENT] = ACTIONS(1348),
    [anon_sym_STAR_STAR] = ACTIONS(1348),
    [anon_sym_LPAREN] = ACTIONS(1348),
    [anon_sym_RPAREN] = ACTIONS(1348),
    [anon_sym_SEMI_SEMI] = ACTIONS(1348),
    [anon_sym_PIPE_AMP] = ACTIONS(1348),
    [anon_sym_EQ_TILDE] = ACTIONS(1348),
    [anon_sym_AMP_GT] = ACTIONS(1348),
    [anon_sym_AMP_GT_GT] = ACTIONS(1348),
    [anon_sym_LT_AMP] = ACTIONS(1348),
    [anon_sym_GT_AMP] = ACTIONS(1348),
    [anon_sym_GT_PIPE] = ACTIONS(1348),
    [anon_sym_LT_AMP_DASH] = ACTIONS(1348),
    [anon_sym_GT_AMP_DASH] = ACTIONS(1348),
    [anon_sym_LT_LT_DASH] = ACTIONS(1348),
    [aux_sym_heredoc_redirect_token1] = ACTIONS(1350),
    [anon_sym_LT_LT_LT] = ACTIONS(1348),
    [anon_sym_QMARK] = ACTIONS(1348),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(1348),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(1348),
    [aux_sym_concatenation_token1] = ACTIONS(1348),
    [anon_sym_DOLLAR] = ACTIONS(1348),
    [sym__special_character] = ACTIONS(1348),
    [anon_sym_DQUOTE] = ACTIONS(1348),
    [sym_raw_string] = ACTIONS(1348),
    [sym_ansi_c_string] = ACTIONS(1348),
    [aux_sym_number_token1] = ACTIONS(1348),
    [aux_sym_number_token2] = ACTIONS(1348),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(1348),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1348),
    [anon_sym_BQUOTE] = ACTIONS(1348),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(1348),
    [anon_sym_LT_LPAREN] = ACTIONS(1348),
    [anon_sym_GT_LPAREN] = ACTIONS(1348),
    [sym_comment] = ACTIONS(3),
    [sym_file_descriptor] = ACTIONS(1350),
    [sym__concat] = ACTIONS(1350),
    [sym_test_operator] = ACTIONS(1350),
    [sym__bare_dollar] = ACTIONS(1350),
    [sym__brace_start] = ACTIONS(1350),
  },
  [490] = {
    [sym_word] = ACTIONS(1352),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(1352),
    [anon_sym_SEMI] = ACTIONS(1352),
    [anon_sym_EQ] = ACTIONS(1352),
    [anon_sym_PLUS_PLUS] = ACTIONS(1352),
    [anon_sym_DASH_DASH] = ACTIONS(1352),
    [anon_sym_PLUS_EQ] = ACTIONS(1352),
    [anon_sym_DASH_EQ] = ACTIONS(1352),
    [anon_sym_STAR_EQ] = ACTIONS(1352),
    [anon_sym_SLASH_EQ] = ACTIONS(1352),
    [anon_sym_PERCENT_EQ] = ACTIONS(1352),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(1352),
    [anon_sym_LT_LT_EQ] = ACTIONS(1352),
    [anon_sym_GT_GT_EQ] = ACTIONS(1352),
    [anon_sym_AMP_EQ] = ACTIONS(1352),
    [anon_sym_CARET_EQ] = ACTIONS(1352),
    [anon_sym_PIPE_EQ] = ACTIONS(1352),
    [anon_sym_PIPE_PIPE] = ACTIONS(1352),
    [anon_sym_AMP_AMP] = ACTIONS(1352),
    [anon_sym_PIPE] = ACTIONS(1352),
    [anon_sym_CARET] = ACTIONS(1352),
    [anon_sym_AMP] = ACTIONS(1352),
    [anon_sym_EQ_EQ] = ACTIONS(1352),
    [anon_sym_BANG_EQ] = ACTIONS(1352),
    [anon_sym_LT] = ACTIONS(1352),
    [anon_sym_GT] = ACTIONS(1352),
    [anon_sym_LT_EQ] = ACTIONS(1352),
    [anon_sym_GT_EQ] = ACTIONS(1352),
    [anon_sym_LT_LT] = ACTIONS(1352),
    [anon_sym_GT_GT] = ACTIONS(1352),
    [anon_sym_PLUS] = ACTIONS(1352),
    [anon_sym_DASH] = ACTIONS(1352),
    [anon_sym_STAR] = ACTIONS(1352),
    [anon_sym_SLASH] = ACTIONS(1352),
    [anon_sym_PERCENT] = ACTIONS(1352),
    [anon_sym_STAR_STAR] = ACTIONS(1352),
    [anon_sym_LPAREN] = ACTIONS(1352),
    [anon_sym_RPAREN] = ACTIONS(1352),
    [anon_sym_SEMI_SEMI] = ACTIONS(1352),
    [anon_sym_PIPE_AMP] = ACTIONS(1352),
    [anon_sym_EQ_TILDE] = ACTIONS(1352),
    [anon_sym_AMP_GT] = ACTIONS(1352),
    [anon_sym_AMP_GT_GT] = ACTIONS(1352),
    [anon_sym_LT_AMP] = ACTIONS(1352),
    [anon_sym_GT_AMP] = ACTIONS(1352),
    [anon_sym_GT_PIPE] = ACTIONS(1352),
    [anon_sym_LT_AMP_DASH] = ACTIONS(1352),
    [anon_sym_GT_AMP_DASH] = ACTIONS(1352),
    [anon_sym_LT_LT_DASH] = ACTIONS(1352),
    [aux_sym_heredoc_redirect_token1] = ACTIONS(1354),
    [anon_sym_LT_LT_LT] = ACTIONS(1352),
    [anon_sym_QMARK] = ACTIONS(1352),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(1352),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(1352),
    [aux_sym_concatenation_token1] = ACTIONS(1352),
    [anon_sym_DOLLAR] = ACTIONS(1352),
    [sym__special_character] = ACTIONS(1352),
    [anon_sym_DQUOTE] = ACTIONS(1352),
    [sym_raw_string] = ACTIONS(1352),
    [sym_ansi_c_string] = ACTIONS(1352),
    [aux_sym_number_token1] = ACTIONS(1352),
    [aux_sym_number_token2] = ACTIONS(1352),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(1352),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1352),
    [anon_sym_BQUOTE] = ACTIONS(1352),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(1352),
    [anon_sym_LT_LPAREN] = ACTIONS(1352),
    [anon_sym_GT_LPAREN] = ACTIONS(1352),
    [sym_comment] = ACTIONS(3),
    [sym_file_descriptor] = ACTIONS(1354),
    [sym__concat] = ACTIONS(1354),
    [sym_test_operator] = ACTIONS(1354),
    [sym__bare_dollar] = ACTIONS(1354),
    [sym__brace_start] = ACTIONS(1354),
  },
  [491] = {
    [sym_word] = ACTIONS(1308),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(1308),
    [anon_sym_SEMI] = ACTIONS(1308),
    [anon_sym_EQ] = ACTIONS(1308),
    [anon_sym_PLUS_PLUS] = ACTIONS(1308),
    [anon_sym_DASH_DASH] = ACTIONS(1308),
    [anon_sym_PLUS_EQ] = ACTIONS(1308),
    [anon_sym_DASH_EQ] = ACTIONS(1308),
    [anon_sym_STAR_EQ] = ACTIONS(1308),
    [anon_sym_SLASH_EQ] = ACTIONS(1308),
    [anon_sym_PERCENT_EQ] = ACTIONS(1308),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(1308),
    [anon_sym_LT_LT_EQ] = ACTIONS(1308),
    [anon_sym_GT_GT_EQ] = ACTIONS(1308),
    [anon_sym_AMP_EQ] = ACTIONS(1308),
    [anon_sym_CARET_EQ] = ACTIONS(1308),
    [anon_sym_PIPE_EQ] = ACTIONS(1308),
    [anon_sym_PIPE_PIPE] = ACTIONS(1308),
    [anon_sym_AMP_AMP] = ACTIONS(1308),
    [anon_sym_PIPE] = ACTIONS(1308),
    [anon_sym_CARET] = ACTIONS(1308),
    [anon_sym_AMP] = ACTIONS(1308),
    [anon_sym_EQ_EQ] = ACTIONS(1308),
    [anon_sym_BANG_EQ] = ACTIONS(1308),
    [anon_sym_LT] = ACTIONS(1308),
    [anon_sym_GT] = ACTIONS(1308),
    [anon_sym_LT_EQ] = ACTIONS(1308),
    [anon_sym_GT_EQ] = ACTIONS(1308),
    [anon_sym_LT_LT] = ACTIONS(1308),
    [anon_sym_GT_GT] = ACTIONS(1308),
    [anon_sym_PLUS] = ACTIONS(1308),
    [anon_sym_DASH] = ACTIONS(1308),
    [anon_sym_STAR] = ACTIONS(1308),
    [anon_sym_SLASH] = ACTIONS(1308),
    [anon_sym_PERCENT] = ACTIONS(1308),
    [anon_sym_STAR_STAR] = ACTIONS(1308),
    [anon_sym_LPAREN] = ACTIONS(1308),
    [anon_sym_RPAREN] = ACTIONS(1308),
    [anon_sym_SEMI_SEMI] = ACTIONS(1308),
    [anon_sym_PIPE_AMP] = ACTIONS(1308),
    [anon_sym_EQ_TILDE] = ACTIONS(1308),
    [anon_sym_AMP_GT] = ACTIONS(1308),
    [anon_sym_AMP_GT_GT] = ACTIONS(1308),
    [anon_sym_LT_AMP] = ACTIONS(1308),
    [anon_sym_GT_AMP] = ACTIONS(1308),
    [anon_sym_GT_PIPE] = ACTIONS(1308),
    [anon_sym_LT_AMP_DASH] = ACTIONS(1308),
    [anon_sym_GT_AMP_DASH] = ACTIONS(1308),
    [anon_sym_LT_LT_DASH] = ACTIONS(1308),
    [aux_sym_heredoc_redirect_token1] = ACTIONS(1310),
    [anon_sym_LT_LT_LT] = ACTIONS(1308),
    [anon_sym_QMARK] = ACTIONS(1308),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(1308),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(1308),
    [aux_sym_concatenation_token1] = ACTIONS(1308),
    [anon_sym_DOLLAR] = ACTIONS(1308),
    [sym__special_character] = ACTIONS(1308),
    [anon_sym_DQUOTE] = ACTIONS(1308),
    [sym_raw_string] = ACTIONS(1308),
    [sym_ansi_c_string] = ACTIONS(1308),
    [aux_sym_number_token1] = ACTIONS(1308),
    [aux_sym_number_token2] = ACTIONS(1308),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(1308),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1308),
    [anon_sym_BQUOTE] = ACTIONS(1308),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(1308),
    [anon_sym_LT_LPAREN] = ACTIONS(1308),
    [anon_sym_GT_LPAREN] = ACTIONS(1308),
    [sym_comment] = ACTIONS(3),
    [sym_file_descriptor] = ACTIONS(1310),
    [sym__concat] = ACTIONS(1310),
    [sym_test_operator] = ACTIONS(1310),
    [sym__bare_dollar] = ACTIONS(1310),
    [sym__brace_start] = ACTIONS(1310),
  },
  [492] = {
    [sym_word] = ACTIONS(1356),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(1356),
    [anon_sym_SEMI] = ACTIONS(1356),
    [anon_sym_EQ] = ACTIONS(1356),
    [anon_sym_PLUS_PLUS] = ACTIONS(1356),
    [anon_sym_DASH_DASH] = ACTIONS(1356),
    [anon_sym_PLUS_EQ] = ACTIONS(1356),
    [anon_sym_DASH_EQ] = ACTIONS(1356),
    [anon_sym_STAR_EQ] = ACTIONS(1356),
    [anon_sym_SLASH_EQ] = ACTIONS(1356),
    [anon_sym_PERCENT_EQ] = ACTIONS(1356),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(1356),
    [anon_sym_LT_LT_EQ] = ACTIONS(1356),
    [anon_sym_GT_GT_EQ] = ACTIONS(1356),
    [anon_sym_AMP_EQ] = ACTIONS(1356),
    [anon_sym_CARET_EQ] = ACTIONS(1356),
    [anon_sym_PIPE_EQ] = ACTIONS(1356),
    [anon_sym_PIPE_PIPE] = ACTIONS(1356),
    [anon_sym_AMP_AMP] = ACTIONS(1356),
    [anon_sym_PIPE] = ACTIONS(1356),
    [anon_sym_CARET] = ACTIONS(1356),
    [anon_sym_AMP] = ACTIONS(1356),
    [anon_sym_EQ_EQ] = ACTIONS(1356),
    [anon_sym_BANG_EQ] = ACTIONS(1356),
    [anon_sym_LT] = ACTIONS(1356),
    [anon_sym_GT] = ACTIONS(1356),
    [anon_sym_LT_EQ] = ACTIONS(1356),
    [anon_sym_GT_EQ] = ACTIONS(1356),
    [anon_sym_LT_LT] = ACTIONS(1356),
    [anon_sym_GT_GT] = ACTIONS(1356),
    [anon_sym_PLUS] = ACTIONS(1356),
    [anon_sym_DASH] = ACTIONS(1356),
    [anon_sym_STAR] = ACTIONS(1356),
    [anon_sym_SLASH] = ACTIONS(1356),
    [anon_sym_PERCENT] = ACTIONS(1356),
    [anon_sym_STAR_STAR] = ACTIONS(1356),
    [anon_sym_LPAREN] = ACTIONS(1356),
    [anon_sym_RPAREN] = ACTIONS(1356),
    [anon_sym_SEMI_SEMI] = ACTIONS(1356),
    [anon_sym_PIPE_AMP] = ACTIONS(1356),
    [anon_sym_EQ_TILDE] = ACTIONS(1356),
    [anon_sym_AMP_GT] = ACTIONS(1356),
    [anon_sym_AMP_GT_GT] = ACTIONS(1356),
    [anon_sym_LT_AMP] = ACTIONS(1356),
    [anon_sym_GT_AMP] = ACTIONS(1356),
    [anon_sym_GT_PIPE] = ACTIONS(1356),
    [anon_sym_LT_AMP_DASH] = ACTIONS(1356),
    [anon_sym_GT_AMP_DASH] = ACTIONS(1356),
    [anon_sym_LT_LT_DASH] = ACTIONS(1356),
    [aux_sym_heredoc_redirect_token1] = ACTIONS(1358),
    [anon_sym_LT_LT_LT] = ACTIONS(1356),
    [anon_sym_QMARK] = ACTIONS(1356),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(1356),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(1356),
    [aux_sym_concatenation_token1] = ACTIONS(1356),
    [anon_sym_DOLLAR] = ACTIONS(1356),
    [sym__special_character] = ACTIONS(1356),
    [anon_sym_DQUOTE] = ACTIONS(1356),
    [sym_raw_string] = ACTIONS(1356),
    [sym_ansi_c_string] = ACTIONS(1356),
    [aux_sym_number_token1] = ACTIONS(1356),
    [aux_sym_number_token2] = ACTIONS(1356),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(1356),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1356),
    [anon_sym_BQUOTE] = ACTIONS(1356),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(1356),
    [anon_sym_LT_LPAREN] = ACTIONS(1356),
    [anon_sym_GT_LPAREN] = ACTIONS(1356),
    [sym_comment] = ACTIONS(3),
    [sym_file_descriptor] = ACTIONS(1358),
    [sym__concat] = ACTIONS(1358),
    [sym_test_operator] = ACTIONS(1358),
    [sym__bare_dollar] = ACTIONS(1358),
    [sym__brace_start] = ACTIONS(1358),
  },
  [493] = {
    [aux_sym__literal_repeat1] = STATE(494),
    [sym_word] = ACTIONS(241),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(241),
    [anon_sym_SEMI] = ACTIONS(241),
    [anon_sym_EQ] = ACTIONS(243),
    [anon_sym_PLUS_PLUS] = ACTIONS(243),
    [anon_sym_DASH_DASH] = ACTIONS(243),
    [anon_sym_PLUS_EQ] = ACTIONS(243),
    [anon_sym_DASH_EQ] = ACTIONS(243),
    [anon_sym_STAR_EQ] = ACTIONS(243),
    [anon_sym_SLASH_EQ] = ACTIONS(243),
    [anon_sym_PERCENT_EQ] = ACTIONS(243),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(243),
    [anon_sym_LT_LT_EQ] = ACTIONS(243),
    [anon_sym_GT_GT_EQ] = ACTIONS(243),
    [anon_sym_AMP_EQ] = ACTIONS(243),
    [anon_sym_CARET_EQ] = ACTIONS(243),
    [anon_sym_PIPE_EQ] = ACTIONS(243),
    [anon_sym_PIPE_PIPE] = ACTIONS(245),
    [anon_sym_AMP_AMP] = ACTIONS(245),
    [anon_sym_PIPE] = ACTIONS(245),
    [anon_sym_CARET] = ACTIONS(243),
    [anon_sym_AMP] = ACTIONS(245),
    [anon_sym_EQ_EQ] = ACTIONS(245),
    [anon_sym_BANG_EQ] = ACTIONS(243),
    [anon_sym_LT] = ACTIONS(245),
    [anon_sym_GT] = ACTIONS(245),
    [anon_sym_LT_EQ] = ACTIONS(243),
    [anon_sym_GT_EQ] = ACTIONS(243),
    [anon_sym_LT_LT] = ACTIONS(245),
    [anon_sym_GT_GT] = ACTIONS(245),
    [anon_sym_PLUS] = ACTIONS(243),
    [anon_sym_DASH] = ACTIONS(243),
    [anon_sym_STAR] = ACTIONS(243),
    [anon_sym_SLASH] = ACTIONS(243),
    [anon_sym_PERCENT] = ACTIONS(243),
    [anon_sym_STAR_STAR] = ACTIONS(243),
    [anon_sym_LPAREN] = ACTIONS(241),
    [anon_sym_RPAREN] = ACTIONS(245),
    [anon_sym_SEMI_SEMI] = ACTIONS(241),
    [anon_sym_PIPE_AMP] = ACTIONS(241),
    [anon_sym_EQ_TILDE] = ACTIONS(245),
    [anon_sym_AMP_GT] = ACTIONS(241),
    [anon_sym_AMP_GT_GT] = ACTIONS(241),
    [anon_sym_LT_AMP] = ACTIONS(241),
    [anon_sym_GT_AMP] = ACTIONS(241),
    [anon_sym_GT_PIPE] = ACTIONS(241),
    [anon_sym_LT_AMP_DASH] = ACTIONS(241),
    [anon_sym_GT_AMP_DASH] = ACTIONS(241),
    [anon_sym_LT_LT_DASH] = ACTIONS(241),
    [aux_sym_heredoc_redirect_token1] = ACTIONS(278),
    [anon_sym_LT_LT_LT] = ACTIONS(241),
    [anon_sym_QMARK] = ACTIONS(243),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(241),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(241),
    [anon_sym_DOLLAR] = ACTIONS(241),
    [sym__special_character] = ACTIONS(1360),
    [anon_sym_DQUOTE] = ACTIONS(241),
    [sym_raw_string] = ACTIONS(241),
    [sym_ansi_c_string] = ACTIONS(241),
    [aux_sym_number_token1] = ACTIONS(241),
    [aux_sym_number_token2] = ACTIONS(241),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(241),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(241),
    [anon_sym_BQUOTE] = ACTIONS(241),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(241),
    [anon_sym_LT_LPAREN] = ACTIONS(241),
    [anon_sym_GT_LPAREN] = ACTIONS(241),
    [sym_comment] = ACTIONS(3),
    [sym_file_descriptor] = ACTIONS(278),
    [sym_test_operator] = ACTIONS(356),
    [sym__bare_dollar] = ACTIONS(278),
    [sym__brace_start] = ACTIONS(278),
  },
  [494] = {
    [aux_sym__literal_repeat1] = STATE(494),
    [sym_word] = ACTIONS(1362),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(1362),
    [anon_sym_SEMI] = ACTIONS(1362),
    [anon_sym_EQ] = ACTIONS(1362),
    [anon_sym_PLUS_PLUS] = ACTIONS(1362),
    [anon_sym_DASH_DASH] = ACTIONS(1362),
    [anon_sym_PLUS_EQ] = ACTIONS(1362),
    [anon_sym_DASH_EQ] = ACTIONS(1362),
    [anon_sym_STAR_EQ] = ACTIONS(1362),
    [anon_sym_SLASH_EQ] = ACTIONS(1362),
    [anon_sym_PERCENT_EQ] = ACTIONS(1362),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(1362),
    [anon_sym_LT_LT_EQ] = ACTIONS(1362),
    [anon_sym_GT_GT_EQ] = ACTIONS(1362),
    [anon_sym_AMP_EQ] = ACTIONS(1362),
    [anon_sym_CARET_EQ] = ACTIONS(1362),
    [anon_sym_PIPE_EQ] = ACTIONS(1362),
    [anon_sym_PIPE_PIPE] = ACTIONS(1362),
    [anon_sym_AMP_AMP] = ACTIONS(1362),
    [anon_sym_PIPE] = ACTIONS(1362),
    [anon_sym_CARET] = ACTIONS(1362),
    [anon_sym_AMP] = ACTIONS(1362),
    [anon_sym_EQ_EQ] = ACTIONS(1362),
    [anon_sym_BANG_EQ] = ACTIONS(1362),
    [anon_sym_LT] = ACTIONS(1362),
    [anon_sym_GT] = ACTIONS(1362),
    [anon_sym_LT_EQ] = ACTIONS(1362),
    [anon_sym_GT_EQ] = ACTIONS(1362),
    [anon_sym_LT_LT] = ACTIONS(1362),
    [anon_sym_GT_GT] = ACTIONS(1362),
    [anon_sym_PLUS] = ACTIONS(1362),
    [anon_sym_DASH] = ACTIONS(1362),
    [anon_sym_STAR] = ACTIONS(1362),
    [anon_sym_SLASH] = ACTIONS(1362),
    [anon_sym_PERCENT] = ACTIONS(1362),
    [anon_sym_STAR_STAR] = ACTIONS(1362),
    [anon_sym_LPAREN] = ACTIONS(1362),
    [anon_sym_RPAREN] = ACTIONS(1362),
    [anon_sym_SEMI_SEMI] = ACTIONS(1362),
    [anon_sym_PIPE_AMP] = ACTIONS(1362),
    [anon_sym_EQ_TILDE] = ACTIONS(1362),
    [anon_sym_AMP_GT] = ACTIONS(1362),
    [anon_sym_AMP_GT_GT] = ACTIONS(1362),
    [anon_sym_LT_AMP] = ACTIONS(1362),
    [anon_sym_GT_AMP] = ACTIONS(1362),
    [anon_sym_GT_PIPE] = ACTIONS(1362),
    [anon_sym_LT_AMP_DASH] = ACTIONS(1362),
    [anon_sym_GT_AMP_DASH] = ACTIONS(1362),
    [anon_sym_LT_LT_DASH] = ACTIONS(1362),
    [aux_sym_heredoc_redirect_token1] = ACTIONS(1364),
    [anon_sym_LT_LT_LT] = ACTIONS(1362),
    [anon_sym_QMARK] = ACTIONS(1362),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(1362),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(1362),
    [anon_sym_DOLLAR] = ACTIONS(1362),
    [sym__special_character] = ACTIONS(1366),
    [anon_sym_DQUOTE] = ACTIONS(1362),
    [sym_raw_string] = ACTIONS(1362),
    [sym_ansi_c_string] = ACTIONS(1362),
    [aux_sym_number_token1] = ACTIONS(1362),
    [aux_sym_number_token2] = ACTIONS(1362),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(1362),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1362),
    [anon_sym_BQUOTE] = ACTIONS(1362),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(1362),
    [anon_sym_LT_LPAREN] = ACTIONS(1362),
    [anon_sym_GT_LPAREN] = ACTIONS(1362),
    [sym_comment] = ACTIONS(3),
    [sym_file_descriptor] = ACTIONS(1364),
    [sym_test_operator] = ACTIONS(1364),
    [sym__bare_dollar] = ACTIONS(1364),
    [sym__brace_start] = ACTIONS(1364),
  },
  [495] = {
    [sym_word] = ACTIONS(1261),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(1261),
    [anon_sym_SEMI] = ACTIONS(1261),
    [anon_sym_EQ] = ACTIONS(1261),
    [anon_sym_PLUS_PLUS] = ACTIONS(1261),
    [anon_sym_DASH_DASH] = ACTIONS(1261),
    [anon_sym_PLUS_EQ] = ACTIONS(1261),
    [anon_sym_DASH_EQ] = ACTIONS(1261),
    [anon_sym_STAR_EQ] = ACTIONS(1261),
    [anon_sym_SLASH_EQ] = ACTIONS(1261),
    [anon_sym_PERCENT_EQ] = ACTIONS(1261),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(1261),
    [anon_sym_LT_LT_EQ] = ACTIONS(1261),
    [anon_sym_GT_GT_EQ] = ACTIONS(1261),
    [anon_sym_AMP_EQ] = ACTIONS(1261),
    [anon_sym_CARET_EQ] = ACTIONS(1261),
    [anon_sym_PIPE_EQ] = ACTIONS(1261),
    [anon_sym_PIPE_PIPE] = ACTIONS(1261),
    [anon_sym_AMP_AMP] = ACTIONS(1261),
    [anon_sym_PIPE] = ACTIONS(1261),
    [anon_sym_CARET] = ACTIONS(1261),
    [anon_sym_AMP] = ACTIONS(1261),
    [anon_sym_EQ_EQ] = ACTIONS(1261),
    [anon_sym_BANG_EQ] = ACTIONS(1261),
    [anon_sym_LT] = ACTIONS(1261),
    [anon_sym_GT] = ACTIONS(1261),
    [anon_sym_LT_EQ] = ACTIONS(1261),
    [anon_sym_GT_EQ] = ACTIONS(1261),
    [anon_sym_LT_LT] = ACTIONS(1261),
    [anon_sym_GT_GT] = ACTIONS(1261),
    [anon_sym_PLUS] = ACTIONS(1261),
    [anon_sym_DASH] = ACTIONS(1261),
    [anon_sym_STAR] = ACTIONS(1261),
    [anon_sym_SLASH] = ACTIONS(1261),
    [anon_sym_PERCENT] = ACTIONS(1261),
    [anon_sym_STAR_STAR] = ACTIONS(1261),
    [anon_sym_LPAREN] = ACTIONS(1261),
    [anon_sym_RPAREN] = ACTIONS(1261),
    [anon_sym_SEMI_SEMI] = ACTIONS(1261),
    [anon_sym_PIPE_AMP] = ACTIONS(1261),
    [anon_sym_EQ_TILDE] = ACTIONS(1261),
    [anon_sym_AMP_GT] = ACTIONS(1261),
    [anon_sym_AMP_GT_GT] = ACTIONS(1261),
    [anon_sym_LT_AMP] = ACTIONS(1261),
    [anon_sym_GT_AMP] = ACTIONS(1261),
    [anon_sym_GT_PIPE] = ACTIONS(1261),
    [anon_sym_LT_AMP_DASH] = ACTIONS(1261),
    [anon_sym_GT_AMP_DASH] = ACTIONS(1261),
    [anon_sym_LT_LT_DASH] = ACTIONS(1261),
    [aux_sym_heredoc_redirect_token1] = ACTIONS(1263),
    [anon_sym_LT_LT_LT] = ACTIONS(1261),
    [anon_sym_QMARK] = ACTIONS(1261),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(1261),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(1261),
    [anon_sym_DOLLAR] = ACTIONS(1261),
    [sym__special_character] = ACTIONS(1261),
    [anon_sym_DQUOTE] = ACTIONS(1261),
    [sym_raw_string] = ACTIONS(1261),
    [sym_ansi_c_string] = ACTIONS(1261),
    [aux_sym_number_token1] = ACTIONS(1261),
    [aux_sym_number_token2] = ACTIONS(1261),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(1261),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1261),
    [anon_sym_BQUOTE] = ACTIONS(1261),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(1261),
    [anon_sym_LT_LPAREN] = ACTIONS(1261),
    [anon_sym_GT_LPAREN] = ACTIONS(1261),
    [sym_comment] = ACTIONS(3),
    [sym_file_descriptor] = ACTIONS(1263),
    [sym_test_operator] = ACTIONS(1263),
    [sym__bare_dollar] = ACTIONS(1263),
    [sym__brace_start] = ACTIONS(1263),
  },
  [496] = {
    [aux_sym_concatenation_repeat1] = STATE(498),
    [sym_word] = ACTIONS(1271),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(1281),
    [anon_sym_EQ] = ACTIONS(1273),
    [anon_sym_PLUS_PLUS] = ACTIONS(1273),
    [anon_sym_DASH_DASH] = ACTIONS(1273),
    [anon_sym_PLUS_EQ] = ACTIONS(1273),
    [anon_sym_DASH_EQ] = ACTIONS(1273),
    [anon_sym_STAR_EQ] = ACTIONS(1273),
    [anon_sym_SLASH_EQ] = ACTIONS(1273),
    [anon_sym_PERCENT_EQ] = ACTIONS(1273),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(1273),
    [anon_sym_LT_LT_EQ] = ACTIONS(1369),
    [anon_sym_GT_GT_EQ] = ACTIONS(1369),
    [anon_sym_AMP_EQ] = ACTIONS(1369),
    [anon_sym_CARET_EQ] = ACTIONS(1273),
    [anon_sym_PIPE_EQ] = ACTIONS(1369),
    [anon_sym_PIPE_PIPE] = ACTIONS(1283),
    [anon_sym_AMP_AMP] = ACTIONS(1283),
    [anon_sym_PIPE] = ACTIONS(1275),
    [anon_sym_CARET] = ACTIONS(1273),
    [anon_sym_AMP] = ACTIONS(1273),
    [anon_sym_EQ_EQ] = ACTIONS(1275),
    [anon_sym_BANG_EQ] = ACTIONS(1273),
    [anon_sym_LT] = ACTIONS(1275),
    [anon_sym_GT] = ACTIONS(1275),
    [anon_sym_LT_EQ] = ACTIONS(1369),
    [anon_sym_GT_EQ] = ACTIONS(1369),
    [anon_sym_LT_LT] = ACTIONS(1275),
    [anon_sym_GT_GT] = ACTIONS(1275),
    [anon_sym_PLUS] = ACTIONS(1273),
    [anon_sym_DASH] = ACTIONS(1273),
    [anon_sym_STAR] = ACTIONS(1273),
    [anon_sym_SLASH] = ACTIONS(1273),
    [anon_sym_PERCENT] = ACTIONS(1273),
    [anon_sym_STAR_STAR] = ACTIONS(1273),
    [anon_sym_LPAREN] = ACTIONS(1271),
    [anon_sym_PIPE_AMP] = ACTIONS(1281),
    [anon_sym_RBRACK] = ACTIONS(1369),
    [anon_sym_EQ_TILDE] = ACTIONS(1275),
    [anon_sym_AMP_GT] = ACTIONS(1271),
    [anon_sym_AMP_GT_GT] = ACTIONS(1281),
    [anon_sym_LT_AMP] = ACTIONS(1271),
    [anon_sym_GT_AMP] = ACTIONS(1271),
    [anon_sym_GT_PIPE] = ACTIONS(1281),
    [anon_sym_LT_AMP_DASH] = ACTIONS(1281),
    [anon_sym_GT_AMP_DASH] = ACTIONS(1281),
    [anon_sym_LT_LT_DASH] = ACTIONS(1281),
    [anon_sym_LT_LT_LT] = ACTIONS(1281),
    [anon_sym_QMARK] = ACTIONS(1273),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(1281),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(1281),
    [aux_sym_concatenation_token1] = ACTIONS(365),
    [anon_sym_DOLLAR] = ACTIONS(1271),
    [sym__special_character] = ACTIONS(1271),
    [anon_sym_DQUOTE] = ACTIONS(1281),
    [sym_raw_string] = ACTIONS(1281),
    [sym_ansi_c_string] = ACTIONS(1281),
    [aux_sym_number_token1] = ACTIONS(1271),
    [aux_sym_number_token2] = ACTIONS(1271),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(1281),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1271),
    [anon_sym_BQUOTE] = ACTIONS(1271),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(1281),
    [anon_sym_LT_LPAREN] = ACTIONS(1281),
    [anon_sym_GT_LPAREN] = ACTIONS(1281),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(1281),
    [sym__concat] = ACTIONS(365),
    [sym_test_operator] = ACTIONS(1283),
    [sym__bare_dollar] = ACTIONS(1281),
    [sym__brace_start] = ACTIONS(1281),
  },
  [497] = {
    [aux_sym_concatenation_repeat1] = STATE(498),
    [sym_word] = ACTIONS(1271),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(1281),
    [anon_sym_EQ] = ACTIONS(1273),
    [anon_sym_PLUS_PLUS] = ACTIONS(1273),
    [anon_sym_DASH_DASH] = ACTIONS(1273),
    [anon_sym_PLUS_EQ] = ACTIONS(1273),
    [anon_sym_DASH_EQ] = ACTIONS(1273),
    [anon_sym_STAR_EQ] = ACTIONS(1273),
    [anon_sym_SLASH_EQ] = ACTIONS(1273),
    [anon_sym_PERCENT_EQ] = ACTIONS(1273),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(1273),
    [anon_sym_LT_LT_EQ] = ACTIONS(1369),
    [anon_sym_GT_GT_EQ] = ACTIONS(1369),
    [anon_sym_AMP_EQ] = ACTIONS(1369),
    [anon_sym_CARET_EQ] = ACTIONS(1273),
    [anon_sym_PIPE_EQ] = ACTIONS(1369),
    [anon_sym_PIPE_PIPE] = ACTIONS(1283),
    [anon_sym_AMP_AMP] = ACTIONS(1283),
    [anon_sym_PIPE] = ACTIONS(1275),
    [anon_sym_CARET] = ACTIONS(1273),
    [anon_sym_AMP] = ACTIONS(1273),
    [anon_sym_EQ_EQ] = ACTIONS(1275),
    [anon_sym_BANG_EQ] = ACTIONS(1273),
    [anon_sym_LT] = ACTIONS(1275),
    [anon_sym_GT] = ACTIONS(1275),
    [anon_sym_LT_EQ] = ACTIONS(1369),
    [anon_sym_GT_EQ] = ACTIONS(1369),
    [anon_sym_LT_LT] = ACTIONS(1275),
    [anon_sym_GT_GT] = ACTIONS(1275),
    [anon_sym_PLUS] = ACTIONS(1273),
    [anon_sym_DASH] = ACTIONS(1273),
    [anon_sym_STAR] = ACTIONS(1273),
    [anon_sym_SLASH] = ACTIONS(1273),
    [anon_sym_PERCENT] = ACTIONS(1273),
    [anon_sym_STAR_STAR] = ACTIONS(1273),
    [anon_sym_LPAREN] = ACTIONS(1371),
    [anon_sym_PIPE_AMP] = ACTIONS(1281),
    [anon_sym_RBRACK] = ACTIONS(1369),
    [anon_sym_EQ_TILDE] = ACTIONS(1275),
    [anon_sym_AMP_GT] = ACTIONS(1271),
    [anon_sym_AMP_GT_GT] = ACTIONS(1281),
    [anon_sym_LT_AMP] = ACTIONS(1271),
    [anon_sym_GT_AMP] = ACTIONS(1271),
    [anon_sym_GT_PIPE] = ACTIONS(1281),
    [anon_sym_LT_AMP_DASH] = ACTIONS(1281),
    [anon_sym_GT_AMP_DASH] = ACTIONS(1281),
    [anon_sym_LT_LT_DASH] = ACTIONS(1281),
    [anon_sym_LT_LT_LT] = ACTIONS(1281),
    [anon_sym_QMARK] = ACTIONS(1273),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(1281),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(1281),
    [aux_sym_concatenation_token1] = ACTIONS(365),
    [anon_sym_DOLLAR] = ACTIONS(1271),
    [sym__special_character] = ACTIONS(1271),
    [anon_sym_DQUOTE] = ACTIONS(1281),
    [sym_raw_string] = ACTIONS(1281),
    [sym_ansi_c_string] = ACTIONS(1281),
    [aux_sym_number_token1] = ACTIONS(1271),
    [aux_sym_number_token2] = ACTIONS(1271),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(1281),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1271),
    [anon_sym_BQUOTE] = ACTIONS(1271),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(1281),
    [anon_sym_LT_LPAREN] = ACTIONS(1281),
    [anon_sym_GT_LPAREN] = ACTIONS(1281),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(1281),
    [sym__concat] = ACTIONS(365),
    [sym_test_operator] = ACTIONS(1283),
    [sym__bare_dollar] = ACTIONS(1281),
    [sym__brace_start] = ACTIONS(1281),
  },
  [498] = {
    [aux_sym_concatenation_repeat1] = STATE(501),
    [sym_word] = ACTIONS(1286),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(1288),
    [anon_sym_EQ] = ACTIONS(1286),
    [anon_sym_PLUS_PLUS] = ACTIONS(1286),
    [anon_sym_DASH_DASH] = ACTIONS(1286),
    [anon_sym_PLUS_EQ] = ACTIONS(1286),
    [anon_sym_DASH_EQ] = ACTIONS(1286),
    [anon_sym_STAR_EQ] = ACTIONS(1286),
    [anon_sym_SLASH_EQ] = ACTIONS(1286),
    [anon_sym_PERCENT_EQ] = ACTIONS(1286),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(1286),
    [anon_sym_LT_LT_EQ] = ACTIONS(1288),
    [anon_sym_GT_GT_EQ] = ACTIONS(1288),
    [anon_sym_AMP_EQ] = ACTIONS(1288),
    [anon_sym_CARET_EQ] = ACTIONS(1286),
    [anon_sym_PIPE_EQ] = ACTIONS(1288),
    [anon_sym_PIPE_PIPE] = ACTIONS(1288),
    [anon_sym_AMP_AMP] = ACTIONS(1288),
    [anon_sym_PIPE] = ACTIONS(1286),
    [anon_sym_CARET] = ACTIONS(1286),
    [anon_sym_AMP] = ACTIONS(1286),
    [anon_sym_EQ_EQ] = ACTIONS(1286),
    [anon_sym_BANG_EQ] = ACTIONS(1286),
    [anon_sym_LT] = ACTIONS(1286),
    [anon_sym_GT] = ACTIONS(1286),
    [anon_sym_LT_EQ] = ACTIONS(1288),
    [anon_sym_GT_EQ] = ACTIONS(1288),
    [anon_sym_LT_LT] = ACTIONS(1286),
    [anon_sym_GT_GT] = ACTIONS(1286),
    [anon_sym_PLUS] = ACTIONS(1286),
    [anon_sym_DASH] = ACTIONS(1286),
    [anon_sym_STAR] = ACTIONS(1286),
    [anon_sym_SLASH] = ACTIONS(1286),
    [anon_sym_PERCENT] = ACTIONS(1286),
    [anon_sym_STAR_STAR] = ACTIONS(1286),
    [anon_sym_LPAREN] = ACTIONS(1286),
    [anon_sym_PIPE_AMP] = ACTIONS(1288),
    [anon_sym_RBRACK] = ACTIONS(1288),
    [anon_sym_EQ_TILDE] = ACTIONS(1286),
    [anon_sym_AMP_GT] = ACTIONS(1286),
    [anon_sym_AMP_GT_GT] = ACTIONS(1288),
    [anon_sym_LT_AMP] = ACTIONS(1286),
    [anon_sym_GT_AMP] = ACTIONS(1286),
    [anon_sym_GT_PIPE] = ACTIONS(1288),
    [anon_sym_LT_AMP_DASH] = ACTIONS(1288),
    [anon_sym_GT_AMP_DASH] = ACTIONS(1288),
    [anon_sym_LT_LT_DASH] = ACTIONS(1288),
    [anon_sym_LT_LT_LT] = ACTIONS(1288),
    [anon_sym_QMARK] = ACTIONS(1286),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(1288),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(1288),
    [aux_sym_concatenation_token1] = ACTIONS(365),
    [anon_sym_DOLLAR] = ACTIONS(1286),
    [sym__special_character] = ACTIONS(1286),
    [anon_sym_DQUOTE] = ACTIONS(1288),
    [sym_raw_string] = ACTIONS(1288),
    [sym_ansi_c_string] = ACTIONS(1288),
    [aux_sym_number_token1] = ACTIONS(1286),
    [aux_sym_number_token2] = ACTIONS(1286),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(1288),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1286),
    [anon_sym_BQUOTE] = ACTIONS(1286),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(1288),
    [anon_sym_LT_LPAREN] = ACTIONS(1288),
    [anon_sym_GT_LPAREN] = ACTIONS(1288),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(1288),
    [sym__concat] = ACTIONS(1374),
    [sym_test_operator] = ACTIONS(1288),
    [sym__bare_dollar] = ACTIONS(1288),
    [sym__brace_start] = ACTIONS(1288),
  },
  [499] = {
    [aux_sym_concatenation_repeat1] = STATE(501),
    [sym_word] = ACTIONS(1265),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(1267),
    [anon_sym_EQ] = ACTIONS(1265),
    [anon_sym_PLUS_PLUS] = ACTIONS(1265),
    [anon_sym_DASH_DASH] = ACTIONS(1265),
    [anon_sym_PLUS_EQ] = ACTIONS(1265),
    [anon_sym_DASH_EQ] = ACTIONS(1265),
    [anon_sym_STAR_EQ] = ACTIONS(1265),
    [anon_sym_SLASH_EQ] = ACTIONS(1265),
    [anon_sym_PERCENT_EQ] = ACTIONS(1265),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(1265),
    [anon_sym_LT_LT_EQ] = ACTIONS(1267),
    [anon_sym_GT_GT_EQ] = ACTIONS(1267),
    [anon_sym_AMP_EQ] = ACTIONS(1267),
    [anon_sym_CARET_EQ] = ACTIONS(1265),
    [anon_sym_PIPE_EQ] = ACTIONS(1267),
    [anon_sym_PIPE_PIPE] = ACTIONS(1267),
    [anon_sym_AMP_AMP] = ACTIONS(1267),
    [anon_sym_PIPE] = ACTIONS(1265),
    [anon_sym_CARET] = ACTIONS(1265),
    [anon_sym_AMP] = ACTIONS(1265),
    [anon_sym_EQ_EQ] = ACTIONS(1265),
    [anon_sym_BANG_EQ] = ACTIONS(1265),
    [anon_sym_LT] = ACTIONS(1265),
    [anon_sym_GT] = ACTIONS(1265),
    [anon_sym_LT_EQ] = ACTIONS(1267),
    [anon_sym_GT_EQ] = ACTIONS(1267),
    [anon_sym_LT_LT] = ACTIONS(1265),
    [anon_sym_GT_GT] = ACTIONS(1265),
    [anon_sym_PLUS] = ACTIONS(1265),
    [anon_sym_DASH] = ACTIONS(1265),
    [anon_sym_STAR] = ACTIONS(1265),
    [anon_sym_SLASH] = ACTIONS(1265),
    [anon_sym_PERCENT] = ACTIONS(1265),
    [anon_sym_STAR_STAR] = ACTIONS(1265),
    [anon_sym_LPAREN] = ACTIONS(1265),
    [anon_sym_PIPE_AMP] = ACTIONS(1267),
    [anon_sym_RBRACK] = ACTIONS(1267),
    [anon_sym_EQ_TILDE] = ACTIONS(1265),
    [anon_sym_AMP_GT] = ACTIONS(1265),
    [anon_sym_AMP_GT_GT] = ACTIONS(1267),
    [anon_sym_LT_AMP] = ACTIONS(1265),
    [anon_sym_GT_AMP] = ACTIONS(1265),
    [anon_sym_GT_PIPE] = ACTIONS(1267),
    [anon_sym_LT_AMP_DASH] = ACTIONS(1267),
    [anon_sym_GT_AMP_DASH] = ACTIONS(1267),
    [anon_sym_LT_LT_DASH] = ACTIONS(1267),
    [anon_sym_LT_LT_LT] = ACTIONS(1267),
    [anon_sym_QMARK] = ACTIONS(1265),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(1267),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(1267),
    [aux_sym_concatenation_token1] = ACTIONS(365),
    [anon_sym_DOLLAR] = ACTIONS(1265),
    [sym__special_character] = ACTIONS(1265),
    [anon_sym_DQUOTE] = ACTIONS(1267),
    [sym_raw_string] = ACTIONS(1267),
    [sym_ansi_c_string] = ACTIONS(1267),
    [aux_sym_number_token1] = ACTIONS(1265),
    [aux_sym_number_token2] = ACTIONS(1265),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(1267),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1265),
    [anon_sym_BQUOTE] = ACTIONS(1265),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(1267),
    [anon_sym_LT_LPAREN] = ACTIONS(1267),
    [anon_sym_GT_LPAREN] = ACTIONS(1267),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(1267),
    [sym__concat] = ACTIONS(1376),
    [sym_test_operator] = ACTIONS(1267),
    [sym__bare_dollar] = ACTIONS(1267),
    [sym__brace_start] = ACTIONS(1267),
  },
  [500] = {
    [aux_sym_concatenation_repeat1] = STATE(499),
    [sym_word] = ACTIONS(1261),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(1263),
    [anon_sym_EQ] = ACTIONS(1261),
    [anon_sym_PLUS_PLUS] = ACTIONS(1261),
    [anon_sym_DASH_DASH] = ACTIONS(1261),
    [anon_sym_PLUS_EQ] = ACTIONS(1261),
    [anon_sym_DASH_EQ] = ACTIONS(1261),
    [anon_sym_STAR_EQ] = ACTIONS(1261),
    [anon_sym_SLASH_EQ] = ACTIONS(1261),
    [anon_sym_PERCENT_EQ] = ACTIONS(1261),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(1261),
    [anon_sym_LT_LT_EQ] = ACTIONS(1263),
    [anon_sym_GT_GT_EQ] = ACTIONS(1263),
    [anon_sym_AMP_EQ] = ACTIONS(1263),
    [anon_sym_CARET_EQ] = ACTIONS(1261),
    [anon_sym_PIPE_EQ] = ACTIONS(1263),
    [anon_sym_PIPE_PIPE] = ACTIONS(1263),
    [anon_sym_AMP_AMP] = ACTIONS(1263),
    [anon_sym_PIPE] = ACTIONS(1261),
    [anon_sym_CARET] = ACTIONS(1261),
    [anon_sym_AMP] = ACTIONS(1261),
    [anon_sym_EQ_EQ] = ACTIONS(1261),
    [anon_sym_BANG_EQ] = ACTIONS(1261),
    [anon_sym_LT] = ACTIONS(1261),
    [anon_sym_GT] = ACTIONS(1261),
    [anon_sym_LT_EQ] = ACTIONS(1263),
    [anon_sym_GT_EQ] = ACTIONS(1263),
    [anon_sym_LT_LT] = ACTIONS(1261),
    [anon_sym_GT_GT] = ACTIONS(1261),
    [anon_sym_PLUS] = ACTIONS(1261),
    [anon_sym_DASH] = ACTIONS(1261),
    [anon_sym_STAR] = ACTIONS(1261),
    [anon_sym_SLASH] = ACTIONS(1261),
    [anon_sym_PERCENT] = ACTIONS(1261),
    [anon_sym_STAR_STAR] = ACTIONS(1261),
    [anon_sym_LPAREN] = ACTIONS(1261),
    [anon_sym_PIPE_AMP] = ACTIONS(1263),
    [anon_sym_RBRACK] = ACTIONS(1263),
    [anon_sym_EQ_TILDE] = ACTIONS(1261),
    [anon_sym_AMP_GT] = ACTIONS(1261),
    [anon_sym_AMP_GT_GT] = ACTIONS(1263),
    [anon_sym_LT_AMP] = ACTIONS(1261),
    [anon_sym_GT_AMP] = ACTIONS(1261),
    [anon_sym_GT_PIPE] = ACTIONS(1263),
    [anon_sym_LT_AMP_DASH] = ACTIONS(1263),
    [anon_sym_GT_AMP_DASH] = ACTIONS(1263),
    [anon_sym_LT_LT_DASH] = ACTIONS(1263),
    [anon_sym_LT_LT_LT] = ACTIONS(1263),
    [anon_sym_QMARK] = ACTIONS(1261),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(1263),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(1263),
    [aux_sym_concatenation_token1] = ACTIONS(365),
    [anon_sym_DOLLAR] = ACTIONS(1261),
    [sym__special_character] = ACTIONS(1261),
    [anon_sym_DQUOTE] = ACTIONS(1263),
    [sym_raw_string] = ACTIONS(1263),
    [sym_ansi_c_string] = ACTIONS(1263),
    [aux_sym_number_token1] = ACTIONS(1261),
    [aux_sym_number_token2] = ACTIONS(1261),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(1263),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1261),
    [anon_sym_BQUOTE] = ACTIONS(1261),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(1263),
    [anon_sym_LT_LPAREN] = ACTIONS(1263),
    [anon_sym_GT_LPAREN] = ACTIONS(1263),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(1263),
    [sym__concat] = ACTIONS(365),
    [sym_test_operator] = ACTIONS(1263),
    [sym__bare_dollar] = ACTIONS(1263),
    [sym__brace_start] = ACTIONS(1263),
  },
  [501] = {
    [aux_sym_concatenation_repeat1] = STATE(501),
    [sym_word] = ACTIONS(1251),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(1253),
    [anon_sym_EQ] = ACTIONS(1251),
    [anon_sym_PLUS_PLUS] = ACTIONS(1251),
    [anon_sym_DASH_DASH] = ACTIONS(1251),
    [anon_sym_PLUS_EQ] = ACTIONS(1251),
    [anon_sym_DASH_EQ] = ACTIONS(1251),
    [anon_sym_STAR_EQ] = ACTIONS(1251),
    [anon_sym_SLASH_EQ] = ACTIONS(1251),
    [anon_sym_PERCENT_EQ] = ACTIONS(1251),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(1251),
    [anon_sym_LT_LT_EQ] = ACTIONS(1253),
    [anon_sym_GT_GT_EQ] = ACTIONS(1253),
    [anon_sym_AMP_EQ] = ACTIONS(1253),
    [anon_sym_CARET_EQ] = ACTIONS(1251),
    [anon_sym_PIPE_EQ] = ACTIONS(1253),
    [anon_sym_PIPE_PIPE] = ACTIONS(1253),
    [anon_sym_AMP_AMP] = ACTIONS(1253),
    [anon_sym_PIPE] = ACTIONS(1251),
    [anon_sym_CARET] = ACTIONS(1251),
    [anon_sym_AMP] = ACTIONS(1251),
    [anon_sym_EQ_EQ] = ACTIONS(1251),
    [anon_sym_BANG_EQ] = ACTIONS(1251),
    [anon_sym_LT] = ACTIONS(1251),
    [anon_sym_GT] = ACTIONS(1251),
    [anon_sym_LT_EQ] = ACTIONS(1253),
    [anon_sym_GT_EQ] = ACTIONS(1253),
    [anon_sym_LT_LT] = ACTIONS(1251),
    [anon_sym_GT_GT] = ACTIONS(1251),
    [anon_sym_PLUS] = ACTIONS(1251),
    [anon_sym_DASH] = ACTIONS(1251),
    [anon_sym_STAR] = ACTIONS(1251),
    [anon_sym_SLASH] = ACTIONS(1251),
    [anon_sym_PERCENT] = ACTIONS(1251),
    [anon_sym_STAR_STAR] = ACTIONS(1251),
    [anon_sym_LPAREN] = ACTIONS(1251),
    [anon_sym_PIPE_AMP] = ACTIONS(1253),
    [anon_sym_RBRACK] = ACTIONS(1253),
    [anon_sym_EQ_TILDE] = ACTIONS(1251),
    [anon_sym_AMP_GT] = ACTIONS(1251),
    [anon_sym_AMP_GT_GT] = ACTIONS(1253),
    [anon_sym_LT_AMP] = ACTIONS(1251),
    [anon_sym_GT_AMP] = ACTIONS(1251),
    [anon_sym_GT_PIPE] = ACTIONS(1253),
    [anon_sym_LT_AMP_DASH] = ACTIONS(1253),
    [anon_sym_GT_AMP_DASH] = ACTIONS(1253),
    [anon_sym_LT_LT_DASH] = ACTIONS(1253),
    [anon_sym_LT_LT_LT] = ACTIONS(1253),
    [anon_sym_QMARK] = ACTIONS(1251),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(1253),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(1253),
    [aux_sym_concatenation_token1] = ACTIONS(1378),
    [anon_sym_DOLLAR] = ACTIONS(1251),
    [sym__special_character] = ACTIONS(1251),
    [anon_sym_DQUOTE] = ACTIONS(1253),
    [sym_raw_string] = ACTIONS(1253),
    [sym_ansi_c_string] = ACTIONS(1253),
    [aux_sym_number_token1] = ACTIONS(1251),
    [aux_sym_number_token2] = ACTIONS(1251),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(1253),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1251),
    [anon_sym_BQUOTE] = ACTIONS(1251),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(1253),
    [anon_sym_LT_LPAREN] = ACTIONS(1253),
    [anon_sym_GT_LPAREN] = ACTIONS(1253),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(1253),
    [sym__concat] = ACTIONS(1378),
    [sym_test_operator] = ACTIONS(1253),
    [sym__bare_dollar] = ACTIONS(1253),
    [sym__brace_start] = ACTIONS(1253),
  },
  [502] = {
    [sym_word] = ACTIONS(1271),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(1271),
    [anon_sym_SEMI] = ACTIONS(1271),
    [anon_sym_EQ] = ACTIONS(1273),
    [anon_sym_PLUS_PLUS] = ACTIONS(1273),
    [anon_sym_DASH_DASH] = ACTIONS(1273),
    [anon_sym_PLUS_EQ] = ACTIONS(1273),
    [anon_sym_DASH_EQ] = ACTIONS(1273),
    [anon_sym_STAR_EQ] = ACTIONS(1273),
    [anon_sym_SLASH_EQ] = ACTIONS(1273),
    [anon_sym_PERCENT_EQ] = ACTIONS(1273),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(1273),
    [anon_sym_LT_LT_EQ] = ACTIONS(1273),
    [anon_sym_GT_GT_EQ] = ACTIONS(1273),
    [anon_sym_AMP_EQ] = ACTIONS(1273),
    [anon_sym_CARET_EQ] = ACTIONS(1273),
    [anon_sym_PIPE_EQ] = ACTIONS(1273),
    [anon_sym_PIPE_PIPE] = ACTIONS(1275),
    [anon_sym_AMP_AMP] = ACTIONS(1275),
    [anon_sym_PIPE] = ACTIONS(1275),
    [anon_sym_CARET] = ACTIONS(1273),
    [anon_sym_AMP] = ACTIONS(1275),
    [anon_sym_EQ_EQ] = ACTIONS(1275),
    [anon_sym_BANG_EQ] = ACTIONS(1273),
    [anon_sym_LT] = ACTIONS(1275),
    [anon_sym_GT] = ACTIONS(1275),
    [anon_sym_LT_EQ] = ACTIONS(1273),
    [anon_sym_GT_EQ] = ACTIONS(1273),
    [anon_sym_LT_LT] = ACTIONS(1275),
    [anon_sym_GT_GT] = ACTIONS(1275),
    [anon_sym_PLUS] = ACTIONS(1273),
    [anon_sym_DASH] = ACTIONS(1273),
    [anon_sym_STAR] = ACTIONS(1273),
    [anon_sym_SLASH] = ACTIONS(1273),
    [anon_sym_PERCENT] = ACTIONS(1273),
    [anon_sym_STAR_STAR] = ACTIONS(1273),
    [anon_sym_LPAREN] = ACTIONS(1271),
    [anon_sym_RPAREN] = ACTIONS(1275),
    [anon_sym_SEMI_SEMI] = ACTIONS(1271),
    [anon_sym_PIPE_AMP] = ACTIONS(1271),
    [anon_sym_EQ_TILDE] = ACTIONS(1275),
    [anon_sym_AMP_GT] = ACTIONS(1271),
    [anon_sym_AMP_GT_GT] = ACTIONS(1271),
    [anon_sym_LT_AMP] = ACTIONS(1271),
    [anon_sym_GT_AMP] = ACTIONS(1271),
    [anon_sym_GT_PIPE] = ACTIONS(1271),
    [anon_sym_LT_AMP_DASH] = ACTIONS(1271),
    [anon_sym_GT_AMP_DASH] = ACTIONS(1271),
    [anon_sym_LT_LT_DASH] = ACTIONS(1271),
    [aux_sym_heredoc_redirect_token1] = ACTIONS(1281),
    [anon_sym_LT_LT_LT] = ACTIONS(1271),
    [anon_sym_QMARK] = ACTIONS(1273),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(1271),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(1271),
    [anon_sym_DOLLAR] = ACTIONS(1271),
    [sym__special_character] = ACTIONS(1271),
    [anon_sym_DQUOTE] = ACTIONS(1271),
    [sym_raw_string] = ACTIONS(1271),
    [sym_ansi_c_string] = ACTIONS(1271),
    [aux_sym_number_token1] = ACTIONS(1271),
    [aux_sym_number_token2] = ACTIONS(1271),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(1271),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1271),
    [anon_sym_BQUOTE] = ACTIONS(1271),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(1271),
    [anon_sym_LT_LPAREN] = ACTIONS(1271),
    [anon_sym_GT_LPAREN] = ACTIONS(1271),
    [sym_comment] = ACTIONS(3),
    [sym_file_descriptor] = ACTIONS(1281),
    [sym_test_operator] = ACTIONS(1283),
    [sym__bare_dollar] = ACTIONS(1281),
    [sym__brace_start] = ACTIONS(1281),
  },
  [503] = {
    [sym_word] = ACTIONS(1348),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(1350),
    [anon_sym_EQ] = ACTIONS(1348),
    [anon_sym_PLUS_PLUS] = ACTIONS(1348),
    [anon_sym_DASH_DASH] = ACTIONS(1348),
    [anon_sym_PLUS_EQ] = ACTIONS(1348),
    [anon_sym_DASH_EQ] = ACTIONS(1348),
    [anon_sym_STAR_EQ] = ACTIONS(1348),
    [anon_sym_SLASH_EQ] = ACTIONS(1348),
    [anon_sym_PERCENT_EQ] = ACTIONS(1348),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(1348),
    [anon_sym_LT_LT_EQ] = ACTIONS(1350),
    [anon_sym_GT_GT_EQ] = ACTIONS(1350),
    [anon_sym_AMP_EQ] = ACTIONS(1350),
    [anon_sym_CARET_EQ] = ACTIONS(1348),
    [anon_sym_PIPE_EQ] = ACTIONS(1350),
    [anon_sym_PIPE_PIPE] = ACTIONS(1350),
    [anon_sym_AMP_AMP] = ACTIONS(1350),
    [anon_sym_PIPE] = ACTIONS(1348),
    [anon_sym_CARET] = ACTIONS(1348),
    [anon_sym_AMP] = ACTIONS(1348),
    [anon_sym_EQ_EQ] = ACTIONS(1348),
    [anon_sym_BANG_EQ] = ACTIONS(1348),
    [anon_sym_LT] = ACTIONS(1348),
    [anon_sym_GT] = ACTIONS(1348),
    [anon_sym_LT_EQ] = ACTIONS(1350),
    [anon_sym_GT_EQ] = ACTIONS(1350),
    [anon_sym_LT_LT] = ACTIONS(1348),
    [anon_sym_GT_GT] = ACTIONS(1348),
    [anon_sym_PLUS] = ACTIONS(1348),
    [anon_sym_DASH] = ACTIONS(1348),
    [anon_sym_STAR] = ACTIONS(1348),
    [anon_sym_SLASH] = ACTIONS(1348),
    [anon_sym_PERCENT] = ACTIONS(1348),
    [anon_sym_STAR_STAR] = ACTIONS(1348),
    [anon_sym_LPAREN] = ACTIONS(1348),
    [anon_sym_PIPE_AMP] = ACTIONS(1350),
    [anon_sym_RBRACK] = ACTIONS(1350),
    [anon_sym_EQ_TILDE] = ACTIONS(1348),
    [anon_sym_AMP_GT] = ACTIONS(1348),
    [anon_sym_AMP_GT_GT] = ACTIONS(1350),
    [anon_sym_LT_AMP] = ACTIONS(1348),
    [anon_sym_GT_AMP] = ACTIONS(1348),
    [anon_sym_GT_PIPE] = ACTIONS(1350),
    [anon_sym_LT_AMP_DASH] = ACTIONS(1350),
    [anon_sym_GT_AMP_DASH] = ACTIONS(1350),
    [anon_sym_LT_LT_DASH] = ACTIONS(1350),
    [anon_sym_LT_LT_LT] = ACTIONS(1350),
    [anon_sym_QMARK] = ACTIONS(1348),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(1350),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(1350),
    [aux_sym_concatenation_token1] = ACTIONS(1350),
    [anon_sym_DOLLAR] = ACTIONS(1348),
    [sym__special_character] = ACTIONS(1348),
    [anon_sym_DQUOTE] = ACTIONS(1350),
    [sym_raw_string] = ACTIONS(1350),
    [sym_ansi_c_string] = ACTIONS(1350),
    [aux_sym_number_token1] = ACTIONS(1348),
    [aux_sym_number_token2] = ACTIONS(1348),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(1350),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1348),
    [anon_sym_BQUOTE] = ACTIONS(1348),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(1350),
    [anon_sym_LT_LPAREN] = ACTIONS(1350),
    [anon_sym_GT_LPAREN] = ACTIONS(1350),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(1350),
    [sym__concat] = ACTIONS(1350),
    [sym_test_operator] = ACTIONS(1350),
    [sym__bare_dollar] = ACTIONS(1350),
    [sym__brace_start] = ACTIONS(1350),
  },
  [504] = {
    [sym_word] = ACTIONS(1352),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(1354),
    [anon_sym_EQ] = ACTIONS(1352),
    [anon_sym_PLUS_PLUS] = ACTIONS(1352),
    [anon_sym_DASH_DASH] = ACTIONS(1352),
    [anon_sym_PLUS_EQ] = ACTIONS(1352),
    [anon_sym_DASH_EQ] = ACTIONS(1352),
    [anon_sym_STAR_EQ] = ACTIONS(1352),
    [anon_sym_SLASH_EQ] = ACTIONS(1352),
    [anon_sym_PERCENT_EQ] = ACTIONS(1352),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(1352),
    [anon_sym_LT_LT_EQ] = ACTIONS(1354),
    [anon_sym_GT_GT_EQ] = ACTIONS(1354),
    [anon_sym_AMP_EQ] = ACTIONS(1354),
    [anon_sym_CARET_EQ] = ACTIONS(1352),
    [anon_sym_PIPE_EQ] = ACTIONS(1354),
    [anon_sym_PIPE_PIPE] = ACTIONS(1354),
    [anon_sym_AMP_AMP] = ACTIONS(1354),
    [anon_sym_PIPE] = ACTIONS(1352),
    [anon_sym_CARET] = ACTIONS(1352),
    [anon_sym_AMP] = ACTIONS(1352),
    [anon_sym_EQ_EQ] = ACTIONS(1352),
    [anon_sym_BANG_EQ] = ACTIONS(1352),
    [anon_sym_LT] = ACTIONS(1352),
    [anon_sym_GT] = ACTIONS(1352),
    [anon_sym_LT_EQ] = ACTIONS(1354),
    [anon_sym_GT_EQ] = ACTIONS(1354),
    [anon_sym_LT_LT] = ACTIONS(1352),
    [anon_sym_GT_GT] = ACTIONS(1352),
    [anon_sym_PLUS] = ACTIONS(1352),
    [anon_sym_DASH] = ACTIONS(1352),
    [anon_sym_STAR] = ACTIONS(1352),
    [anon_sym_SLASH] = ACTIONS(1352),
    [anon_sym_PERCENT] = ACTIONS(1352),
    [anon_sym_STAR_STAR] = ACTIONS(1352),
    [anon_sym_LPAREN] = ACTIONS(1352),
    [anon_sym_PIPE_AMP] = ACTIONS(1354),
    [anon_sym_RBRACK] = ACTIONS(1354),
    [anon_sym_EQ_TILDE] = ACTIONS(1352),
    [anon_sym_AMP_GT] = ACTIONS(1352),
    [anon_sym_AMP_GT_GT] = ACTIONS(1354),
    [anon_sym_LT_AMP] = ACTIONS(1352),
    [anon_sym_GT_AMP] = ACTIONS(1352),
    [anon_sym_GT_PIPE] = ACTIONS(1354),
    [anon_sym_LT_AMP_DASH] = ACTIONS(1354),
    [anon_sym_GT_AMP_DASH] = ACTIONS(1354),
    [anon_sym_LT_LT_DASH] = ACTIONS(1354),
    [anon_sym_LT_LT_LT] = ACTIONS(1354),
    [anon_sym_QMARK] = ACTIONS(1352),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(1354),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(1354),
    [aux_sym_concatenation_token1] = ACTIONS(1354),
    [anon_sym_DOLLAR] = ACTIONS(1352),
    [sym__special_character] = ACTIONS(1352),
    [anon_sym_DQUOTE] = ACTIONS(1354),
    [sym_raw_string] = ACTIONS(1354),
    [sym_ansi_c_string] = ACTIONS(1354),
    [aux_sym_number_token1] = ACTIONS(1352),
    [aux_sym_number_token2] = ACTIONS(1352),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(1354),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1352),
    [anon_sym_BQUOTE] = ACTIONS(1352),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(1354),
    [anon_sym_LT_LPAREN] = ACTIONS(1354),
    [anon_sym_GT_LPAREN] = ACTIONS(1354),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(1354),
    [sym__concat] = ACTIONS(1354),
    [sym_test_operator] = ACTIONS(1354),
    [sym__bare_dollar] = ACTIONS(1354),
    [sym__brace_start] = ACTIONS(1354),
  },
  [505] = {
    [sym_word] = ACTIONS(1328),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(1330),
    [anon_sym_EQ] = ACTIONS(1328),
    [anon_sym_PLUS_PLUS] = ACTIONS(1328),
    [anon_sym_DASH_DASH] = ACTIONS(1328),
    [anon_sym_PLUS_EQ] = ACTIONS(1328),
    [anon_sym_DASH_EQ] = ACTIONS(1328),
    [anon_sym_STAR_EQ] = ACTIONS(1328),
    [anon_sym_SLASH_EQ] = ACTIONS(1328),
    [anon_sym_PERCENT_EQ] = ACTIONS(1328),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(1328),
    [anon_sym_LT_LT_EQ] = ACTIONS(1330),
    [anon_sym_GT_GT_EQ] = ACTIONS(1330),
    [anon_sym_AMP_EQ] = ACTIONS(1330),
    [anon_sym_CARET_EQ] = ACTIONS(1328),
    [anon_sym_PIPE_EQ] = ACTIONS(1330),
    [anon_sym_PIPE_PIPE] = ACTIONS(1330),
    [anon_sym_AMP_AMP] = ACTIONS(1330),
    [anon_sym_PIPE] = ACTIONS(1328),
    [anon_sym_CARET] = ACTIONS(1328),
    [anon_sym_AMP] = ACTIONS(1328),
    [anon_sym_EQ_EQ] = ACTIONS(1328),
    [anon_sym_BANG_EQ] = ACTIONS(1328),
    [anon_sym_LT] = ACTIONS(1328),
    [anon_sym_GT] = ACTIONS(1328),
    [anon_sym_LT_EQ] = ACTIONS(1330),
    [anon_sym_GT_EQ] = ACTIONS(1330),
    [anon_sym_LT_LT] = ACTIONS(1328),
    [anon_sym_GT_GT] = ACTIONS(1328),
    [anon_sym_PLUS] = ACTIONS(1328),
    [anon_sym_DASH] = ACTIONS(1328),
    [anon_sym_STAR] = ACTIONS(1328),
    [anon_sym_SLASH] = ACTIONS(1328),
    [anon_sym_PERCENT] = ACTIONS(1328),
    [anon_sym_STAR_STAR] = ACTIONS(1328),
    [anon_sym_LPAREN] = ACTIONS(1328),
    [anon_sym_PIPE_AMP] = ACTIONS(1330),
    [anon_sym_RBRACK] = ACTIONS(1330),
    [anon_sym_EQ_TILDE] = ACTIONS(1328),
    [anon_sym_AMP_GT] = ACTIONS(1328),
    [anon_sym_AMP_GT_GT] = ACTIONS(1330),
    [anon_sym_LT_AMP] = ACTIONS(1328),
    [anon_sym_GT_AMP] = ACTIONS(1328),
    [anon_sym_GT_PIPE] = ACTIONS(1330),
    [anon_sym_LT_AMP_DASH] = ACTIONS(1330),
    [anon_sym_GT_AMP_DASH] = ACTIONS(1330),
    [anon_sym_LT_LT_DASH] = ACTIONS(1330),
    [anon_sym_LT_LT_LT] = ACTIONS(1330),
    [anon_sym_QMARK] = ACTIONS(1328),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(1330),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(1330),
    [aux_sym_concatenation_token1] = ACTIONS(1330),
    [anon_sym_DOLLAR] = ACTIONS(1328),
    [sym__special_character] = ACTIONS(1328),
    [anon_sym_DQUOTE] = ACTIONS(1330),
    [sym_raw_string] = ACTIONS(1330),
    [sym_ansi_c_string] = ACTIONS(1330),
    [aux_sym_number_token1] = ACTIONS(1328),
    [aux_sym_number_token2] = ACTIONS(1328),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(1330),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1328),
    [anon_sym_BQUOTE] = ACTIONS(1328),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(1330),
    [anon_sym_LT_LPAREN] = ACTIONS(1330),
    [anon_sym_GT_LPAREN] = ACTIONS(1330),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(1330),
    [sym__concat] = ACTIONS(1330),
    [sym_test_operator] = ACTIONS(1330),
    [sym__bare_dollar] = ACTIONS(1330),
    [sym__brace_start] = ACTIONS(1330),
  },
  [506] = {
    [sym_word] = ACTIONS(1344),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(1346),
    [anon_sym_EQ] = ACTIONS(1344),
    [anon_sym_PLUS_PLUS] = ACTIONS(1344),
    [anon_sym_DASH_DASH] = ACTIONS(1344),
    [anon_sym_PLUS_EQ] = ACTIONS(1344),
    [anon_sym_DASH_EQ] = ACTIONS(1344),
    [anon_sym_STAR_EQ] = ACTIONS(1344),
    [anon_sym_SLASH_EQ] = ACTIONS(1344),
    [anon_sym_PERCENT_EQ] = ACTIONS(1344),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(1344),
    [anon_sym_LT_LT_EQ] = ACTIONS(1346),
    [anon_sym_GT_GT_EQ] = ACTIONS(1346),
    [anon_sym_AMP_EQ] = ACTIONS(1346),
    [anon_sym_CARET_EQ] = ACTIONS(1344),
    [anon_sym_PIPE_EQ] = ACTIONS(1346),
    [anon_sym_PIPE_PIPE] = ACTIONS(1346),
    [anon_sym_AMP_AMP] = ACTIONS(1346),
    [anon_sym_PIPE] = ACTIONS(1344),
    [anon_sym_CARET] = ACTIONS(1344),
    [anon_sym_AMP] = ACTIONS(1344),
    [anon_sym_EQ_EQ] = ACTIONS(1344),
    [anon_sym_BANG_EQ] = ACTIONS(1344),
    [anon_sym_LT] = ACTIONS(1344),
    [anon_sym_GT] = ACTIONS(1344),
    [anon_sym_LT_EQ] = ACTIONS(1346),
    [anon_sym_GT_EQ] = ACTIONS(1346),
    [anon_sym_LT_LT] = ACTIONS(1344),
    [anon_sym_GT_GT] = ACTIONS(1344),
    [anon_sym_PLUS] = ACTIONS(1344),
    [anon_sym_DASH] = ACTIONS(1344),
    [anon_sym_STAR] = ACTIONS(1344),
    [anon_sym_SLASH] = ACTIONS(1344),
    [anon_sym_PERCENT] = ACTIONS(1344),
    [anon_sym_STAR_STAR] = ACTIONS(1344),
    [anon_sym_LPAREN] = ACTIONS(1344),
    [anon_sym_PIPE_AMP] = ACTIONS(1346),
    [anon_sym_RBRACK] = ACTIONS(1346),
    [anon_sym_EQ_TILDE] = ACTIONS(1344),
    [anon_sym_AMP_GT] = ACTIONS(1344),
    [anon_sym_AMP_GT_GT] = ACTIONS(1346),
    [anon_sym_LT_AMP] = ACTIONS(1344),
    [anon_sym_GT_AMP] = ACTIONS(1344),
    [anon_sym_GT_PIPE] = ACTIONS(1346),
    [anon_sym_LT_AMP_DASH] = ACTIONS(1346),
    [anon_sym_GT_AMP_DASH] = ACTIONS(1346),
    [anon_sym_LT_LT_DASH] = ACTIONS(1346),
    [anon_sym_LT_LT_LT] = ACTIONS(1346),
    [anon_sym_QMARK] = ACTIONS(1344),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(1346),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(1346),
    [aux_sym_concatenation_token1] = ACTIONS(1346),
    [anon_sym_DOLLAR] = ACTIONS(1344),
    [sym__special_character] = ACTIONS(1344),
    [anon_sym_DQUOTE] = ACTIONS(1346),
    [sym_raw_string] = ACTIONS(1346),
    [sym_ansi_c_string] = ACTIONS(1346),
    [aux_sym_number_token1] = ACTIONS(1344),
    [aux_sym_number_token2] = ACTIONS(1344),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(1346),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1344),
    [anon_sym_BQUOTE] = ACTIONS(1344),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(1346),
    [anon_sym_LT_LPAREN] = ACTIONS(1346),
    [anon_sym_GT_LPAREN] = ACTIONS(1346),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(1346),
    [sym__concat] = ACTIONS(1346),
    [sym_test_operator] = ACTIONS(1346),
    [sym__bare_dollar] = ACTIONS(1346),
    [sym__brace_start] = ACTIONS(1346),
  },
  [507] = {
    [sym_word] = ACTIONS(1356),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(1358),
    [anon_sym_EQ] = ACTIONS(1356),
    [anon_sym_PLUS_PLUS] = ACTIONS(1356),
    [anon_sym_DASH_DASH] = ACTIONS(1356),
    [anon_sym_PLUS_EQ] = ACTIONS(1356),
    [anon_sym_DASH_EQ] = ACTIONS(1356),
    [anon_sym_STAR_EQ] = ACTIONS(1356),
    [anon_sym_SLASH_EQ] = ACTIONS(1356),
    [anon_sym_PERCENT_EQ] = ACTIONS(1356),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(1356),
    [anon_sym_LT_LT_EQ] = ACTIONS(1358),
    [anon_sym_GT_GT_EQ] = ACTIONS(1358),
    [anon_sym_AMP_EQ] = ACTIONS(1358),
    [anon_sym_CARET_EQ] = ACTIONS(1356),
    [anon_sym_PIPE_EQ] = ACTIONS(1358),
    [anon_sym_PIPE_PIPE] = ACTIONS(1358),
    [anon_sym_AMP_AMP] = ACTIONS(1358),
    [anon_sym_PIPE] = ACTIONS(1356),
    [anon_sym_CARET] = ACTIONS(1356),
    [anon_sym_AMP] = ACTIONS(1356),
    [anon_sym_EQ_EQ] = ACTIONS(1356),
    [anon_sym_BANG_EQ] = ACTIONS(1356),
    [anon_sym_LT] = ACTIONS(1356),
    [anon_sym_GT] = ACTIONS(1356),
    [anon_sym_LT_EQ] = ACTIONS(1358),
    [anon_sym_GT_EQ] = ACTIONS(1358),
    [anon_sym_LT_LT] = ACTIONS(1356),
    [anon_sym_GT_GT] = ACTIONS(1356),
    [anon_sym_PLUS] = ACTIONS(1356),
    [anon_sym_DASH] = ACTIONS(1356),
    [anon_sym_STAR] = ACTIONS(1356),
    [anon_sym_SLASH] = ACTIONS(1356),
    [anon_sym_PERCENT] = ACTIONS(1356),
    [anon_sym_STAR_STAR] = ACTIONS(1356),
    [anon_sym_LPAREN] = ACTIONS(1356),
    [anon_sym_PIPE_AMP] = ACTIONS(1358),
    [anon_sym_RBRACK] = ACTIONS(1358),
    [anon_sym_EQ_TILDE] = ACTIONS(1356),
    [anon_sym_AMP_GT] = ACTIONS(1356),
    [anon_sym_AMP_GT_GT] = ACTIONS(1358),
    [anon_sym_LT_AMP] = ACTIONS(1356),
    [anon_sym_GT_AMP] = ACTIONS(1356),
    [anon_sym_GT_PIPE] = ACTIONS(1358),
    [anon_sym_LT_AMP_DASH] = ACTIONS(1358),
    [anon_sym_GT_AMP_DASH] = ACTIONS(1358),
    [anon_sym_LT_LT_DASH] = ACTIONS(1358),
    [anon_sym_LT_LT_LT] = ACTIONS(1358),
    [anon_sym_QMARK] = ACTIONS(1356),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(1358),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(1358),
    [aux_sym_concatenation_token1] = ACTIONS(1358),
    [anon_sym_DOLLAR] = ACTIONS(1356),
    [sym__special_character] = ACTIONS(1356),
    [anon_sym_DQUOTE] = ACTIONS(1358),
    [sym_raw_string] = ACTIONS(1358),
    [sym_ansi_c_string] = ACTIONS(1358),
    [aux_sym_number_token1] = ACTIONS(1356),
    [aux_sym_number_token2] = ACTIONS(1356),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(1358),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1356),
    [anon_sym_BQUOTE] = ACTIONS(1356),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(1358),
    [anon_sym_LT_LPAREN] = ACTIONS(1358),
    [anon_sym_GT_LPAREN] = ACTIONS(1358),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(1358),
    [sym__concat] = ACTIONS(1358),
    [sym_test_operator] = ACTIONS(1358),
    [sym__bare_dollar] = ACTIONS(1358),
    [sym__brace_start] = ACTIONS(1358),
  },
  [508] = {
    [sym_word] = ACTIONS(1300),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(1302),
    [anon_sym_EQ] = ACTIONS(1300),
    [anon_sym_PLUS_PLUS] = ACTIONS(1300),
    [anon_sym_DASH_DASH] = ACTIONS(1300),
    [anon_sym_PLUS_EQ] = ACTIONS(1300),
    [anon_sym_DASH_EQ] = ACTIONS(1300),
    [anon_sym_STAR_EQ] = ACTIONS(1300),
    [anon_sym_SLASH_EQ] = ACTIONS(1300),
    [anon_sym_PERCENT_EQ] = ACTIONS(1300),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(1300),
    [anon_sym_LT_LT_EQ] = ACTIONS(1302),
    [anon_sym_GT_GT_EQ] = ACTIONS(1302),
    [anon_sym_AMP_EQ] = ACTIONS(1302),
    [anon_sym_CARET_EQ] = ACTIONS(1300),
    [anon_sym_PIPE_EQ] = ACTIONS(1302),
    [anon_sym_PIPE_PIPE] = ACTIONS(1302),
    [anon_sym_AMP_AMP] = ACTIONS(1302),
    [anon_sym_PIPE] = ACTIONS(1300),
    [anon_sym_CARET] = ACTIONS(1300),
    [anon_sym_AMP] = ACTIONS(1300),
    [anon_sym_EQ_EQ] = ACTIONS(1300),
    [anon_sym_BANG_EQ] = ACTIONS(1300),
    [anon_sym_LT] = ACTIONS(1300),
    [anon_sym_GT] = ACTIONS(1300),
    [anon_sym_LT_EQ] = ACTIONS(1302),
    [anon_sym_GT_EQ] = ACTIONS(1302),
    [anon_sym_LT_LT] = ACTIONS(1300),
    [anon_sym_GT_GT] = ACTIONS(1300),
    [anon_sym_PLUS] = ACTIONS(1300),
    [anon_sym_DASH] = ACTIONS(1300),
    [anon_sym_STAR] = ACTIONS(1300),
    [anon_sym_SLASH] = ACTIONS(1300),
    [anon_sym_PERCENT] = ACTIONS(1300),
    [anon_sym_STAR_STAR] = ACTIONS(1300),
    [anon_sym_LPAREN] = ACTIONS(1300),
    [anon_sym_PIPE_AMP] = ACTIONS(1302),
    [anon_sym_RBRACK] = ACTIONS(1302),
    [anon_sym_EQ_TILDE] = ACTIONS(1300),
    [anon_sym_AMP_GT] = ACTIONS(1300),
    [anon_sym_AMP_GT_GT] = ACTIONS(1302),
    [anon_sym_LT_AMP] = ACTIONS(1300),
    [anon_sym_GT_AMP] = ACTIONS(1300),
    [anon_sym_GT_PIPE] = ACTIONS(1302),
    [anon_sym_LT_AMP_DASH] = ACTIONS(1302),
    [anon_sym_GT_AMP_DASH] = ACTIONS(1302),
    [anon_sym_LT_LT_DASH] = ACTIONS(1302),
    [anon_sym_LT_LT_LT] = ACTIONS(1302),
    [anon_sym_QMARK] = ACTIONS(1300),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(1302),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(1302),
    [aux_sym_concatenation_token1] = ACTIONS(1302),
    [anon_sym_DOLLAR] = ACTIONS(1300),
    [sym__special_character] = ACTIONS(1300),
    [anon_sym_DQUOTE] = ACTIONS(1302),
    [sym_raw_string] = ACTIONS(1302),
    [sym_ansi_c_string] = ACTIONS(1302),
    [aux_sym_number_token1] = ACTIONS(1300),
    [aux_sym_number_token2] = ACTIONS(1300),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(1302),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1300),
    [anon_sym_BQUOTE] = ACTIONS(1300),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(1302),
    [anon_sym_LT_LPAREN] = ACTIONS(1302),
    [anon_sym_GT_LPAREN] = ACTIONS(1302),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(1302),
    [sym__concat] = ACTIONS(1302),
    [sym_test_operator] = ACTIONS(1302),
    [sym__bare_dollar] = ACTIONS(1302),
    [sym__brace_start] = ACTIONS(1302),
  },
  [509] = {
    [sym_word] = ACTIONS(1320),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(1322),
    [anon_sym_EQ] = ACTIONS(1320),
    [anon_sym_PLUS_PLUS] = ACTIONS(1320),
    [anon_sym_DASH_DASH] = ACTIONS(1320),
    [anon_sym_PLUS_EQ] = ACTIONS(1320),
    [anon_sym_DASH_EQ] = ACTIONS(1320),
    [anon_sym_STAR_EQ] = ACTIONS(1320),
    [anon_sym_SLASH_EQ] = ACTIONS(1320),
    [anon_sym_PERCENT_EQ] = ACTIONS(1320),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(1320),
    [anon_sym_LT_LT_EQ] = ACTIONS(1322),
    [anon_sym_GT_GT_EQ] = ACTIONS(1322),
    [anon_sym_AMP_EQ] = ACTIONS(1322),
    [anon_sym_CARET_EQ] = ACTIONS(1320),
    [anon_sym_PIPE_EQ] = ACTIONS(1322),
    [anon_sym_PIPE_PIPE] = ACTIONS(1322),
    [anon_sym_AMP_AMP] = ACTIONS(1322),
    [anon_sym_PIPE] = ACTIONS(1320),
    [anon_sym_CARET] = ACTIONS(1320),
    [anon_sym_AMP] = ACTIONS(1320),
    [anon_sym_EQ_EQ] = ACTIONS(1320),
    [anon_sym_BANG_EQ] = ACTIONS(1320),
    [anon_sym_LT] = ACTIONS(1320),
    [anon_sym_GT] = ACTIONS(1320),
    [anon_sym_LT_EQ] = ACTIONS(1322),
    [anon_sym_GT_EQ] = ACTIONS(1322),
    [anon_sym_LT_LT] = ACTIONS(1320),
    [anon_sym_GT_GT] = ACTIONS(1320),
    [anon_sym_PLUS] = ACTIONS(1320),
    [anon_sym_DASH] = ACTIONS(1320),
    [anon_sym_STAR] = ACTIONS(1320),
    [anon_sym_SLASH] = ACTIONS(1320),
    [anon_sym_PERCENT] = ACTIONS(1320),
    [anon_sym_STAR_STAR] = ACTIONS(1320),
    [anon_sym_LPAREN] = ACTIONS(1320),
    [anon_sym_PIPE_AMP] = ACTIONS(1322),
    [anon_sym_RBRACK] = ACTIONS(1322),
    [anon_sym_EQ_TILDE] = ACTIONS(1320),
    [anon_sym_AMP_GT] = ACTIONS(1320),
    [anon_sym_AMP_GT_GT] = ACTIONS(1322),
    [anon_sym_LT_AMP] = ACTIONS(1320),
    [anon_sym_GT_AMP] = ACTIONS(1320),
    [anon_sym_GT_PIPE] = ACTIONS(1322),
    [anon_sym_LT_AMP_DASH] = ACTIONS(1322),
    [anon_sym_GT_AMP_DASH] = ACTIONS(1322),
    [anon_sym_LT_LT_DASH] = ACTIONS(1322),
    [anon_sym_LT_LT_LT] = ACTIONS(1322),
    [anon_sym_QMARK] = ACTIONS(1320),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(1322),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(1322),
    [aux_sym_concatenation_token1] = ACTIONS(1322),
    [anon_sym_DOLLAR] = ACTIONS(1320),
    [sym__special_character] = ACTIONS(1320),
    [anon_sym_DQUOTE] = ACTIONS(1322),
    [sym_raw_string] = ACTIONS(1322),
    [sym_ansi_c_string] = ACTIONS(1322),
    [aux_sym_number_token1] = ACTIONS(1320),
    [aux_sym_number_token2] = ACTIONS(1320),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(1322),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1320),
    [anon_sym_BQUOTE] = ACTIONS(1320),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(1322),
    [anon_sym_LT_LPAREN] = ACTIONS(1322),
    [anon_sym_GT_LPAREN] = ACTIONS(1322),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(1322),
    [sym__concat] = ACTIONS(1322),
    [sym_test_operator] = ACTIONS(1322),
    [sym__bare_dollar] = ACTIONS(1322),
    [sym__brace_start] = ACTIONS(1322),
  },
  [510] = {
    [sym_word] = ACTIONS(1316),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(1318),
    [anon_sym_EQ] = ACTIONS(1316),
    [anon_sym_PLUS_PLUS] = ACTIONS(1316),
    [anon_sym_DASH_DASH] = ACTIONS(1316),
    [anon_sym_PLUS_EQ] = ACTIONS(1316),
    [anon_sym_DASH_EQ] = ACTIONS(1316),
    [anon_sym_STAR_EQ] = ACTIONS(1316),
    [anon_sym_SLASH_EQ] = ACTIONS(1316),
    [anon_sym_PERCENT_EQ] = ACTIONS(1316),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(1316),
    [anon_sym_LT_LT_EQ] = ACTIONS(1318),
    [anon_sym_GT_GT_EQ] = ACTIONS(1318),
    [anon_sym_AMP_EQ] = ACTIONS(1318),
    [anon_sym_CARET_EQ] = ACTIONS(1316),
    [anon_sym_PIPE_EQ] = ACTIONS(1318),
    [anon_sym_PIPE_PIPE] = ACTIONS(1318),
    [anon_sym_AMP_AMP] = ACTIONS(1318),
    [anon_sym_PIPE] = ACTIONS(1316),
    [anon_sym_CARET] = ACTIONS(1316),
    [anon_sym_AMP] = ACTIONS(1316),
    [anon_sym_EQ_EQ] = ACTIONS(1316),
    [anon_sym_BANG_EQ] = ACTIONS(1316),
    [anon_sym_LT] = ACTIONS(1316),
    [anon_sym_GT] = ACTIONS(1316),
    [anon_sym_LT_EQ] = ACTIONS(1318),
    [anon_sym_GT_EQ] = ACTIONS(1318),
    [anon_sym_LT_LT] = ACTIONS(1316),
    [anon_sym_GT_GT] = ACTIONS(1316),
    [anon_sym_PLUS] = ACTIONS(1316),
    [anon_sym_DASH] = ACTIONS(1316),
    [anon_sym_STAR] = ACTIONS(1316),
    [anon_sym_SLASH] = ACTIONS(1316),
    [anon_sym_PERCENT] = ACTIONS(1316),
    [anon_sym_STAR_STAR] = ACTIONS(1316),
    [anon_sym_LPAREN] = ACTIONS(1316),
    [anon_sym_PIPE_AMP] = ACTIONS(1318),
    [anon_sym_RBRACK] = ACTIONS(1318),
    [anon_sym_EQ_TILDE] = ACTIONS(1316),
    [anon_sym_AMP_GT] = ACTIONS(1316),
    [anon_sym_AMP_GT_GT] = ACTIONS(1318),
    [anon_sym_LT_AMP] = ACTIONS(1316),
    [anon_sym_GT_AMP] = ACTIONS(1316),
    [anon_sym_GT_PIPE] = ACTIONS(1318),
    [anon_sym_LT_AMP_DASH] = ACTIONS(1318),
    [anon_sym_GT_AMP_DASH] = ACTIONS(1318),
    [anon_sym_LT_LT_DASH] = ACTIONS(1318),
    [anon_sym_LT_LT_LT] = ACTIONS(1318),
    [anon_sym_QMARK] = ACTIONS(1316),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(1318),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(1318),
    [aux_sym_concatenation_token1] = ACTIONS(1318),
    [anon_sym_DOLLAR] = ACTIONS(1316),
    [sym__special_character] = ACTIONS(1316),
    [anon_sym_DQUOTE] = ACTIONS(1318),
    [sym_raw_string] = ACTIONS(1318),
    [sym_ansi_c_string] = ACTIONS(1318),
    [aux_sym_number_token1] = ACTIONS(1316),
    [aux_sym_number_token2] = ACTIONS(1316),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(1318),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1316),
    [anon_sym_BQUOTE] = ACTIONS(1316),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(1318),
    [anon_sym_LT_LPAREN] = ACTIONS(1318),
    [anon_sym_GT_LPAREN] = ACTIONS(1318),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(1318),
    [sym__concat] = ACTIONS(1318),
    [sym_test_operator] = ACTIONS(1318),
    [sym__bare_dollar] = ACTIONS(1318),
    [sym__brace_start] = ACTIONS(1318),
  },
  [511] = {
    [sym_word] = ACTIONS(1336),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(1338),
    [anon_sym_EQ] = ACTIONS(1336),
    [anon_sym_PLUS_PLUS] = ACTIONS(1336),
    [anon_sym_DASH_DASH] = ACTIONS(1336),
    [anon_sym_PLUS_EQ] = ACTIONS(1336),
    [anon_sym_DASH_EQ] = ACTIONS(1336),
    [anon_sym_STAR_EQ] = ACTIONS(1336),
    [anon_sym_SLASH_EQ] = ACTIONS(1336),
    [anon_sym_PERCENT_EQ] = ACTIONS(1336),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(1336),
    [anon_sym_LT_LT_EQ] = ACTIONS(1338),
    [anon_sym_GT_GT_EQ] = ACTIONS(1338),
    [anon_sym_AMP_EQ] = ACTIONS(1338),
    [anon_sym_CARET_EQ] = ACTIONS(1336),
    [anon_sym_PIPE_EQ] = ACTIONS(1338),
    [anon_sym_PIPE_PIPE] = ACTIONS(1338),
    [anon_sym_AMP_AMP] = ACTIONS(1338),
    [anon_sym_PIPE] = ACTIONS(1336),
    [anon_sym_CARET] = ACTIONS(1336),
    [anon_sym_AMP] = ACTIONS(1336),
    [anon_sym_EQ_EQ] = ACTIONS(1336),
    [anon_sym_BANG_EQ] = ACTIONS(1336),
    [anon_sym_LT] = ACTIONS(1336),
    [anon_sym_GT] = ACTIONS(1336),
    [anon_sym_LT_EQ] = ACTIONS(1338),
    [anon_sym_GT_EQ] = ACTIONS(1338),
    [anon_sym_LT_LT] = ACTIONS(1336),
    [anon_sym_GT_GT] = ACTIONS(1336),
    [anon_sym_PLUS] = ACTIONS(1336),
    [anon_sym_DASH] = ACTIONS(1336),
    [anon_sym_STAR] = ACTIONS(1336),
    [anon_sym_SLASH] = ACTIONS(1336),
    [anon_sym_PERCENT] = ACTIONS(1336),
    [anon_sym_STAR_STAR] = ACTIONS(1336),
    [anon_sym_LPAREN] = ACTIONS(1336),
    [anon_sym_PIPE_AMP] = ACTIONS(1338),
    [anon_sym_RBRACK] = ACTIONS(1338),
    [anon_sym_EQ_TILDE] = ACTIONS(1336),
    [anon_sym_AMP_GT] = ACTIONS(1336),
    [anon_sym_AMP_GT_GT] = ACTIONS(1338),
    [anon_sym_LT_AMP] = ACTIONS(1336),
    [anon_sym_GT_AMP] = ACTIONS(1336),
    [anon_sym_GT_PIPE] = ACTIONS(1338),
    [anon_sym_LT_AMP_DASH] = ACTIONS(1338),
    [anon_sym_GT_AMP_DASH] = ACTIONS(1338),
    [anon_sym_LT_LT_DASH] = ACTIONS(1338),
    [anon_sym_LT_LT_LT] = ACTIONS(1338),
    [anon_sym_QMARK] = ACTIONS(1336),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(1338),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(1338),
    [aux_sym_concatenation_token1] = ACTIONS(1338),
    [anon_sym_DOLLAR] = ACTIONS(1336),
    [sym__special_character] = ACTIONS(1336),
    [anon_sym_DQUOTE] = ACTIONS(1338),
    [sym_raw_string] = ACTIONS(1338),
    [sym_ansi_c_string] = ACTIONS(1338),
    [aux_sym_number_token1] = ACTIONS(1336),
    [aux_sym_number_token2] = ACTIONS(1336),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(1338),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1336),
    [anon_sym_BQUOTE] = ACTIONS(1336),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(1338),
    [anon_sym_LT_LPAREN] = ACTIONS(1338),
    [anon_sym_GT_LPAREN] = ACTIONS(1338),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(1338),
    [sym__concat] = ACTIONS(1338),
    [sym_test_operator] = ACTIONS(1338),
    [sym__bare_dollar] = ACTIONS(1338),
    [sym__brace_start] = ACTIONS(1338),
  },
  [512] = {
    [sym_word] = ACTIONS(1324),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(1326),
    [anon_sym_EQ] = ACTIONS(1324),
    [anon_sym_PLUS_PLUS] = ACTIONS(1324),
    [anon_sym_DASH_DASH] = ACTIONS(1324),
    [anon_sym_PLUS_EQ] = ACTIONS(1324),
    [anon_sym_DASH_EQ] = ACTIONS(1324),
    [anon_sym_STAR_EQ] = ACTIONS(1324),
    [anon_sym_SLASH_EQ] = ACTIONS(1324),
    [anon_sym_PERCENT_EQ] = ACTIONS(1324),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(1324),
    [anon_sym_LT_LT_EQ] = ACTIONS(1326),
    [anon_sym_GT_GT_EQ] = ACTIONS(1326),
    [anon_sym_AMP_EQ] = ACTIONS(1326),
    [anon_sym_CARET_EQ] = ACTIONS(1324),
    [anon_sym_PIPE_EQ] = ACTIONS(1326),
    [anon_sym_PIPE_PIPE] = ACTIONS(1326),
    [anon_sym_AMP_AMP] = ACTIONS(1326),
    [anon_sym_PIPE] = ACTIONS(1324),
    [anon_sym_CARET] = ACTIONS(1324),
    [anon_sym_AMP] = ACTIONS(1324),
    [anon_sym_EQ_EQ] = ACTIONS(1324),
    [anon_sym_BANG_EQ] = ACTIONS(1324),
    [anon_sym_LT] = ACTIONS(1324),
    [anon_sym_GT] = ACTIONS(1324),
    [anon_sym_LT_EQ] = ACTIONS(1326),
    [anon_sym_GT_EQ] = ACTIONS(1326),
    [anon_sym_LT_LT] = ACTIONS(1324),
    [anon_sym_GT_GT] = ACTIONS(1324),
    [anon_sym_PLUS] = ACTIONS(1324),
    [anon_sym_DASH] = ACTIONS(1324),
    [anon_sym_STAR] = ACTIONS(1324),
    [anon_sym_SLASH] = ACTIONS(1324),
    [anon_sym_PERCENT] = ACTIONS(1324),
    [anon_sym_STAR_STAR] = ACTIONS(1324),
    [anon_sym_LPAREN] = ACTIONS(1324),
    [anon_sym_PIPE_AMP] = ACTIONS(1326),
    [anon_sym_RBRACK] = ACTIONS(1326),
    [anon_sym_EQ_TILDE] = ACTIONS(1324),
    [anon_sym_AMP_GT] = ACTIONS(1324),
    [anon_sym_AMP_GT_GT] = ACTIONS(1326),
    [anon_sym_LT_AMP] = ACTIONS(1324),
    [anon_sym_GT_AMP] = ACTIONS(1324),
    [anon_sym_GT_PIPE] = ACTIONS(1326),
    [anon_sym_LT_AMP_DASH] = ACTIONS(1326),
    [anon_sym_GT_AMP_DASH] = ACTIONS(1326),
    [anon_sym_LT_LT_DASH] = ACTIONS(1326),
    [anon_sym_LT_LT_LT] = ACTIONS(1326),
    [anon_sym_QMARK] = ACTIONS(1324),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(1326),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(1326),
    [aux_sym_concatenation_token1] = ACTIONS(1326),
    [anon_sym_DOLLAR] = ACTIONS(1324),
    [sym__special_character] = ACTIONS(1324),
    [anon_sym_DQUOTE] = ACTIONS(1326),
    [sym_raw_string] = ACTIONS(1326),
    [sym_ansi_c_string] = ACTIONS(1326),
    [aux_sym_number_token1] = ACTIONS(1324),
    [aux_sym_number_token2] = ACTIONS(1324),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(1326),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1324),
    [anon_sym_BQUOTE] = ACTIONS(1324),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(1326),
    [anon_sym_LT_LPAREN] = ACTIONS(1326),
    [anon_sym_GT_LPAREN] = ACTIONS(1326),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(1326),
    [sym__concat] = ACTIONS(1326),
    [sym_test_operator] = ACTIONS(1326),
    [sym__bare_dollar] = ACTIONS(1326),
    [sym__brace_start] = ACTIONS(1326),
  },
  [513] = {
    [sym_word] = ACTIONS(1251),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(1253),
    [anon_sym_EQ] = ACTIONS(1251),
    [anon_sym_PLUS_PLUS] = ACTIONS(1251),
    [anon_sym_DASH_DASH] = ACTIONS(1251),
    [anon_sym_PLUS_EQ] = ACTIONS(1251),
    [anon_sym_DASH_EQ] = ACTIONS(1251),
    [anon_sym_STAR_EQ] = ACTIONS(1251),
    [anon_sym_SLASH_EQ] = ACTIONS(1251),
    [anon_sym_PERCENT_EQ] = ACTIONS(1251),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(1251),
    [anon_sym_LT_LT_EQ] = ACTIONS(1253),
    [anon_sym_GT_GT_EQ] = ACTIONS(1253),
    [anon_sym_AMP_EQ] = ACTIONS(1253),
    [anon_sym_CARET_EQ] = ACTIONS(1251),
    [anon_sym_PIPE_EQ] = ACTIONS(1253),
    [anon_sym_PIPE_PIPE] = ACTIONS(1253),
    [anon_sym_AMP_AMP] = ACTIONS(1253),
    [anon_sym_PIPE] = ACTIONS(1251),
    [anon_sym_CARET] = ACTIONS(1251),
    [anon_sym_AMP] = ACTIONS(1251),
    [anon_sym_EQ_EQ] = ACTIONS(1251),
    [anon_sym_BANG_EQ] = ACTIONS(1251),
    [anon_sym_LT] = ACTIONS(1251),
    [anon_sym_GT] = ACTIONS(1251),
    [anon_sym_LT_EQ] = ACTIONS(1253),
    [anon_sym_GT_EQ] = ACTIONS(1253),
    [anon_sym_LT_LT] = ACTIONS(1251),
    [anon_sym_GT_GT] = ACTIONS(1251),
    [anon_sym_PLUS] = ACTIONS(1251),
    [anon_sym_DASH] = ACTIONS(1251),
    [anon_sym_STAR] = ACTIONS(1251),
    [anon_sym_SLASH] = ACTIONS(1251),
    [anon_sym_PERCENT] = ACTIONS(1251),
    [anon_sym_STAR_STAR] = ACTIONS(1251),
    [anon_sym_LPAREN] = ACTIONS(1251),
    [anon_sym_PIPE_AMP] = ACTIONS(1253),
    [anon_sym_RBRACK] = ACTIONS(1253),
    [anon_sym_EQ_TILDE] = ACTIONS(1251),
    [anon_sym_AMP_GT] = ACTIONS(1251),
    [anon_sym_AMP_GT_GT] = ACTIONS(1253),
    [anon_sym_LT_AMP] = ACTIONS(1251),
    [anon_sym_GT_AMP] = ACTIONS(1251),
    [anon_sym_GT_PIPE] = ACTIONS(1253),
    [anon_sym_LT_AMP_DASH] = ACTIONS(1253),
    [anon_sym_GT_AMP_DASH] = ACTIONS(1253),
    [anon_sym_LT_LT_DASH] = ACTIONS(1253),
    [anon_sym_LT_LT_LT] = ACTIONS(1253),
    [anon_sym_QMARK] = ACTIONS(1251),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(1253),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(1253),
    [aux_sym_concatenation_token1] = ACTIONS(1253),
    [anon_sym_DOLLAR] = ACTIONS(1251),
    [sym__special_character] = ACTIONS(1251),
    [anon_sym_DQUOTE] = ACTIONS(1253),
    [sym_raw_string] = ACTIONS(1253),
    [sym_ansi_c_string] = ACTIONS(1253),
    [aux_sym_number_token1] = ACTIONS(1251),
    [aux_sym_number_token2] = ACTIONS(1251),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(1253),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1251),
    [anon_sym_BQUOTE] = ACTIONS(1251),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(1253),
    [anon_sym_LT_LPAREN] = ACTIONS(1253),
    [anon_sym_GT_LPAREN] = ACTIONS(1253),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(1253),
    [sym__concat] = ACTIONS(1253),
    [sym_test_operator] = ACTIONS(1253),
    [sym__bare_dollar] = ACTIONS(1253),
    [sym__brace_start] = ACTIONS(1253),
  },
  [514] = {
    [sym_word] = ACTIONS(1332),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(1334),
    [anon_sym_EQ] = ACTIONS(1332),
    [anon_sym_PLUS_PLUS] = ACTIONS(1332),
    [anon_sym_DASH_DASH] = ACTIONS(1332),
    [anon_sym_PLUS_EQ] = ACTIONS(1332),
    [anon_sym_DASH_EQ] = ACTIONS(1332),
    [anon_sym_STAR_EQ] = ACTIONS(1332),
    [anon_sym_SLASH_EQ] = ACTIONS(1332),
    [anon_sym_PERCENT_EQ] = ACTIONS(1332),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(1332),
    [anon_sym_LT_LT_EQ] = ACTIONS(1334),
    [anon_sym_GT_GT_EQ] = ACTIONS(1334),
    [anon_sym_AMP_EQ] = ACTIONS(1334),
    [anon_sym_CARET_EQ] = ACTIONS(1332),
    [anon_sym_PIPE_EQ] = ACTIONS(1334),
    [anon_sym_PIPE_PIPE] = ACTIONS(1334),
    [anon_sym_AMP_AMP] = ACTIONS(1334),
    [anon_sym_PIPE] = ACTIONS(1332),
    [anon_sym_CARET] = ACTIONS(1332),
    [anon_sym_AMP] = ACTIONS(1332),
    [anon_sym_EQ_EQ] = ACTIONS(1332),
    [anon_sym_BANG_EQ] = ACTIONS(1332),
    [anon_sym_LT] = ACTIONS(1332),
    [anon_sym_GT] = ACTIONS(1332),
    [anon_sym_LT_EQ] = ACTIONS(1334),
    [anon_sym_GT_EQ] = ACTIONS(1334),
    [anon_sym_LT_LT] = ACTIONS(1332),
    [anon_sym_GT_GT] = ACTIONS(1332),
    [anon_sym_PLUS] = ACTIONS(1332),
    [anon_sym_DASH] = ACTIONS(1332),
    [anon_sym_STAR] = ACTIONS(1332),
    [anon_sym_SLASH] = ACTIONS(1332),
    [anon_sym_PERCENT] = ACTIONS(1332),
    [anon_sym_STAR_STAR] = ACTIONS(1332),
    [anon_sym_LPAREN] = ACTIONS(1332),
    [anon_sym_PIPE_AMP] = ACTIONS(1334),
    [anon_sym_RBRACK] = ACTIONS(1334),
    [anon_sym_EQ_TILDE] = ACTIONS(1332),
    [anon_sym_AMP_GT] = ACTIONS(1332),
    [anon_sym_AMP_GT_GT] = ACTIONS(1334),
    [anon_sym_LT_AMP] = ACTIONS(1332),
    [anon_sym_GT_AMP] = ACTIONS(1332),
    [anon_sym_GT_PIPE] = ACTIONS(1334),
    [anon_sym_LT_AMP_DASH] = ACTIONS(1334),
    [anon_sym_GT_AMP_DASH] = ACTIONS(1334),
    [anon_sym_LT_LT_DASH] = ACTIONS(1334),
    [anon_sym_LT_LT_LT] = ACTIONS(1334),
    [anon_sym_QMARK] = ACTIONS(1332),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(1334),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(1334),
    [aux_sym_concatenation_token1] = ACTIONS(1334),
    [anon_sym_DOLLAR] = ACTIONS(1332),
    [sym__special_character] = ACTIONS(1332),
    [anon_sym_DQUOTE] = ACTIONS(1334),
    [sym_raw_string] = ACTIONS(1334),
    [sym_ansi_c_string] = ACTIONS(1334),
    [aux_sym_number_token1] = ACTIONS(1332),
    [aux_sym_number_token2] = ACTIONS(1332),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(1334),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1332),
    [anon_sym_BQUOTE] = ACTIONS(1332),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(1334),
    [anon_sym_LT_LPAREN] = ACTIONS(1334),
    [anon_sym_GT_LPAREN] = ACTIONS(1334),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(1334),
    [sym__concat] = ACTIONS(1334),
    [sym_test_operator] = ACTIONS(1334),
    [sym__bare_dollar] = ACTIONS(1334),
    [sym__brace_start] = ACTIONS(1334),
  },
  [515] = {
    [sym_word] = ACTIONS(1292),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(1294),
    [anon_sym_EQ] = ACTIONS(1292),
    [anon_sym_PLUS_PLUS] = ACTIONS(1292),
    [anon_sym_DASH_DASH] = ACTIONS(1292),
    [anon_sym_PLUS_EQ] = ACTIONS(1292),
    [anon_sym_DASH_EQ] = ACTIONS(1292),
    [anon_sym_STAR_EQ] = ACTIONS(1292),
    [anon_sym_SLASH_EQ] = ACTIONS(1292),
    [anon_sym_PERCENT_EQ] = ACTIONS(1292),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(1292),
    [anon_sym_LT_LT_EQ] = ACTIONS(1294),
    [anon_sym_GT_GT_EQ] = ACTIONS(1294),
    [anon_sym_AMP_EQ] = ACTIONS(1294),
    [anon_sym_CARET_EQ] = ACTIONS(1292),
    [anon_sym_PIPE_EQ] = ACTIONS(1294),
    [anon_sym_PIPE_PIPE] = ACTIONS(1294),
    [anon_sym_AMP_AMP] = ACTIONS(1294),
    [anon_sym_PIPE] = ACTIONS(1292),
    [anon_sym_CARET] = ACTIONS(1292),
    [anon_sym_AMP] = ACTIONS(1292),
    [anon_sym_EQ_EQ] = ACTIONS(1292),
    [anon_sym_BANG_EQ] = ACTIONS(1292),
    [anon_sym_LT] = ACTIONS(1292),
    [anon_sym_GT] = ACTIONS(1292),
    [anon_sym_LT_EQ] = ACTIONS(1294),
    [anon_sym_GT_EQ] = ACTIONS(1294),
    [anon_sym_LT_LT] = ACTIONS(1292),
    [anon_sym_GT_GT] = ACTIONS(1292),
    [anon_sym_PLUS] = ACTIONS(1292),
    [anon_sym_DASH] = ACTIONS(1292),
    [anon_sym_STAR] = ACTIONS(1292),
    [anon_sym_SLASH] = ACTIONS(1292),
    [anon_sym_PERCENT] = ACTIONS(1292),
    [anon_sym_STAR_STAR] = ACTIONS(1292),
    [anon_sym_LPAREN] = ACTIONS(1292),
    [anon_sym_PIPE_AMP] = ACTIONS(1294),
    [anon_sym_RBRACK] = ACTIONS(1294),
    [anon_sym_EQ_TILDE] = ACTIONS(1292),
    [anon_sym_AMP_GT] = ACTIONS(1292),
    [anon_sym_AMP_GT_GT] = ACTIONS(1294),
    [anon_sym_LT_AMP] = ACTIONS(1292),
    [anon_sym_GT_AMP] = ACTIONS(1292),
    [anon_sym_GT_PIPE] = ACTIONS(1294),
    [anon_sym_LT_AMP_DASH] = ACTIONS(1294),
    [anon_sym_GT_AMP_DASH] = ACTIONS(1294),
    [anon_sym_LT_LT_DASH] = ACTIONS(1294),
    [anon_sym_LT_LT_LT] = ACTIONS(1294),
    [anon_sym_QMARK] = ACTIONS(1292),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(1294),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(1294),
    [aux_sym_concatenation_token1] = ACTIONS(1294),
    [anon_sym_DOLLAR] = ACTIONS(1292),
    [sym__special_character] = ACTIONS(1292),
    [anon_sym_DQUOTE] = ACTIONS(1294),
    [sym_raw_string] = ACTIONS(1294),
    [sym_ansi_c_string] = ACTIONS(1294),
    [aux_sym_number_token1] = ACTIONS(1292),
    [aux_sym_number_token2] = ACTIONS(1292),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(1294),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1292),
    [anon_sym_BQUOTE] = ACTIONS(1292),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(1294),
    [anon_sym_LT_LPAREN] = ACTIONS(1294),
    [anon_sym_GT_LPAREN] = ACTIONS(1294),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(1294),
    [sym__concat] = ACTIONS(1294),
    [sym_test_operator] = ACTIONS(1294),
    [sym__bare_dollar] = ACTIONS(1294),
    [sym__brace_start] = ACTIONS(1294),
  },
  [516] = {
    [sym_word] = ACTIONS(1296),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(1298),
    [anon_sym_EQ] = ACTIONS(1296),
    [anon_sym_PLUS_PLUS] = ACTIONS(1296),
    [anon_sym_DASH_DASH] = ACTIONS(1296),
    [anon_sym_PLUS_EQ] = ACTIONS(1296),
    [anon_sym_DASH_EQ] = ACTIONS(1296),
    [anon_sym_STAR_EQ] = ACTIONS(1296),
    [anon_sym_SLASH_EQ] = ACTIONS(1296),
    [anon_sym_PERCENT_EQ] = ACTIONS(1296),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(1296),
    [anon_sym_LT_LT_EQ] = ACTIONS(1298),
    [anon_sym_GT_GT_EQ] = ACTIONS(1298),
    [anon_sym_AMP_EQ] = ACTIONS(1298),
    [anon_sym_CARET_EQ] = ACTIONS(1296),
    [anon_sym_PIPE_EQ] = ACTIONS(1298),
    [anon_sym_PIPE_PIPE] = ACTIONS(1298),
    [anon_sym_AMP_AMP] = ACTIONS(1298),
    [anon_sym_PIPE] = ACTIONS(1296),
    [anon_sym_CARET] = ACTIONS(1296),
    [anon_sym_AMP] = ACTIONS(1296),
    [anon_sym_EQ_EQ] = ACTIONS(1296),
    [anon_sym_BANG_EQ] = ACTIONS(1296),
    [anon_sym_LT] = ACTIONS(1296),
    [anon_sym_GT] = ACTIONS(1296),
    [anon_sym_LT_EQ] = ACTIONS(1298),
    [anon_sym_GT_EQ] = ACTIONS(1298),
    [anon_sym_LT_LT] = ACTIONS(1296),
    [anon_sym_GT_GT] = ACTIONS(1296),
    [anon_sym_PLUS] = ACTIONS(1296),
    [anon_sym_DASH] = ACTIONS(1296),
    [anon_sym_STAR] = ACTIONS(1296),
    [anon_sym_SLASH] = ACTIONS(1296),
    [anon_sym_PERCENT] = ACTIONS(1296),
    [anon_sym_STAR_STAR] = ACTIONS(1296),
    [anon_sym_LPAREN] = ACTIONS(1296),
    [anon_sym_PIPE_AMP] = ACTIONS(1298),
    [anon_sym_RBRACK] = ACTIONS(1298),
    [anon_sym_EQ_TILDE] = ACTIONS(1296),
    [anon_sym_AMP_GT] = ACTIONS(1296),
    [anon_sym_AMP_GT_GT] = ACTIONS(1298),
    [anon_sym_LT_AMP] = ACTIONS(1296),
    [anon_sym_GT_AMP] = ACTIONS(1296),
    [anon_sym_GT_PIPE] = ACTIONS(1298),
    [anon_sym_LT_AMP_DASH] = ACTIONS(1298),
    [anon_sym_GT_AMP_DASH] = ACTIONS(1298),
    [anon_sym_LT_LT_DASH] = ACTIONS(1298),
    [anon_sym_LT_LT_LT] = ACTIONS(1298),
    [anon_sym_QMARK] = ACTIONS(1296),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(1298),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(1298),
    [aux_sym_concatenation_token1] = ACTIONS(1298),
    [anon_sym_DOLLAR] = ACTIONS(1296),
    [sym__special_character] = ACTIONS(1296),
    [anon_sym_DQUOTE] = ACTIONS(1298),
    [sym_raw_string] = ACTIONS(1298),
    [sym_ansi_c_string] = ACTIONS(1298),
    [aux_sym_number_token1] = ACTIONS(1296),
    [aux_sym_number_token2] = ACTIONS(1296),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(1298),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1296),
    [anon_sym_BQUOTE] = ACTIONS(1296),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(1298),
    [anon_sym_LT_LPAREN] = ACTIONS(1298),
    [anon_sym_GT_LPAREN] = ACTIONS(1298),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(1298),
    [sym__concat] = ACTIONS(1298),
    [sym_test_operator] = ACTIONS(1298),
    [sym__bare_dollar] = ACTIONS(1298),
    [sym__brace_start] = ACTIONS(1298),
  },
  [517] = {
    [sym_word] = ACTIONS(1308),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(1310),
    [anon_sym_EQ] = ACTIONS(1308),
    [anon_sym_PLUS_PLUS] = ACTIONS(1308),
    [anon_sym_DASH_DASH] = ACTIONS(1308),
    [anon_sym_PLUS_EQ] = ACTIONS(1308),
    [anon_sym_DASH_EQ] = ACTIONS(1308),
    [anon_sym_STAR_EQ] = ACTIONS(1308),
    [anon_sym_SLASH_EQ] = ACTIONS(1308),
    [anon_sym_PERCENT_EQ] = ACTIONS(1308),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(1308),
    [anon_sym_LT_LT_EQ] = ACTIONS(1310),
    [anon_sym_GT_GT_EQ] = ACTIONS(1310),
    [anon_sym_AMP_EQ] = ACTIONS(1310),
    [anon_sym_CARET_EQ] = ACTIONS(1308),
    [anon_sym_PIPE_EQ] = ACTIONS(1310),
    [anon_sym_PIPE_PIPE] = ACTIONS(1310),
    [anon_sym_AMP_AMP] = ACTIONS(1310),
    [anon_sym_PIPE] = ACTIONS(1308),
    [anon_sym_CARET] = ACTIONS(1308),
    [anon_sym_AMP] = ACTIONS(1308),
    [anon_sym_EQ_EQ] = ACTIONS(1308),
    [anon_sym_BANG_EQ] = ACTIONS(1308),
    [anon_sym_LT] = ACTIONS(1308),
    [anon_sym_GT] = ACTIONS(1308),
    [anon_sym_LT_EQ] = ACTIONS(1310),
    [anon_sym_GT_EQ] = ACTIONS(1310),
    [anon_sym_LT_LT] = ACTIONS(1308),
    [anon_sym_GT_GT] = ACTIONS(1308),
    [anon_sym_PLUS] = ACTIONS(1308),
    [anon_sym_DASH] = ACTIONS(1308),
    [anon_sym_STAR] = ACTIONS(1308),
    [anon_sym_SLASH] = ACTIONS(1308),
    [anon_sym_PERCENT] = ACTIONS(1308),
    [anon_sym_STAR_STAR] = ACTIONS(1308),
    [anon_sym_LPAREN] = ACTIONS(1308),
    [anon_sym_PIPE_AMP] = ACTIONS(1310),
    [anon_sym_RBRACK] = ACTIONS(1310),
    [anon_sym_EQ_TILDE] = ACTIONS(1308),
    [anon_sym_AMP_GT] = ACTIONS(1308),
    [anon_sym_AMP_GT_GT] = ACTIONS(1310),
    [anon_sym_LT_AMP] = ACTIONS(1308),
    [anon_sym_GT_AMP] = ACTIONS(1308),
    [anon_sym_GT_PIPE] = ACTIONS(1310),
    [anon_sym_LT_AMP_DASH] = ACTIONS(1310),
    [anon_sym_GT_AMP_DASH] = ACTIONS(1310),
    [anon_sym_LT_LT_DASH] = ACTIONS(1310),
    [anon_sym_LT_LT_LT] = ACTIONS(1310),
    [anon_sym_QMARK] = ACTIONS(1308),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(1310),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(1310),
    [aux_sym_concatenation_token1] = ACTIONS(1310),
    [anon_sym_DOLLAR] = ACTIONS(1308),
    [sym__special_character] = ACTIONS(1308),
    [anon_sym_DQUOTE] = ACTIONS(1310),
    [sym_raw_string] = ACTIONS(1310),
    [sym_ansi_c_string] = ACTIONS(1310),
    [aux_sym_number_token1] = ACTIONS(1308),
    [aux_sym_number_token2] = ACTIONS(1308),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(1310),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1308),
    [anon_sym_BQUOTE] = ACTIONS(1308),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(1310),
    [anon_sym_LT_LPAREN] = ACTIONS(1310),
    [anon_sym_GT_LPAREN] = ACTIONS(1310),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(1310),
    [sym__concat] = ACTIONS(1310),
    [sym_test_operator] = ACTIONS(1310),
    [sym__bare_dollar] = ACTIONS(1310),
    [sym__brace_start] = ACTIONS(1310),
  },
  [518] = {
    [sym_word] = ACTIONS(1304),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(1306),
    [anon_sym_EQ] = ACTIONS(1304),
    [anon_sym_PLUS_PLUS] = ACTIONS(1304),
    [anon_sym_DASH_DASH] = ACTIONS(1304),
    [anon_sym_PLUS_EQ] = ACTIONS(1304),
    [anon_sym_DASH_EQ] = ACTIONS(1304),
    [anon_sym_STAR_EQ] = ACTIONS(1304),
    [anon_sym_SLASH_EQ] = ACTIONS(1304),
    [anon_sym_PERCENT_EQ] = ACTIONS(1304),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(1304),
    [anon_sym_LT_LT_EQ] = ACTIONS(1306),
    [anon_sym_GT_GT_EQ] = ACTIONS(1306),
    [anon_sym_AMP_EQ] = ACTIONS(1306),
    [anon_sym_CARET_EQ] = ACTIONS(1304),
    [anon_sym_PIPE_EQ] = ACTIONS(1306),
    [anon_sym_PIPE_PIPE] = ACTIONS(1306),
    [anon_sym_AMP_AMP] = ACTIONS(1306),
    [anon_sym_PIPE] = ACTIONS(1304),
    [anon_sym_CARET] = ACTIONS(1304),
    [anon_sym_AMP] = ACTIONS(1304),
    [anon_sym_EQ_EQ] = ACTIONS(1304),
    [anon_sym_BANG_EQ] = ACTIONS(1304),
    [anon_sym_LT] = ACTIONS(1304),
    [anon_sym_GT] = ACTIONS(1304),
    [anon_sym_LT_EQ] = ACTIONS(1306),
    [anon_sym_GT_EQ] = ACTIONS(1306),
    [anon_sym_LT_LT] = ACTIONS(1304),
    [anon_sym_GT_GT] = ACTIONS(1304),
    [anon_sym_PLUS] = ACTIONS(1304),
    [anon_sym_DASH] = ACTIONS(1304),
    [anon_sym_STAR] = ACTIONS(1304),
    [anon_sym_SLASH] = ACTIONS(1304),
    [anon_sym_PERCENT] = ACTIONS(1304),
    [anon_sym_STAR_STAR] = ACTIONS(1304),
    [anon_sym_LPAREN] = ACTIONS(1304),
    [anon_sym_PIPE_AMP] = ACTIONS(1306),
    [anon_sym_RBRACK] = ACTIONS(1306),
    [anon_sym_EQ_TILDE] = ACTIONS(1304),
    [anon_sym_AMP_GT] = ACTIONS(1304),
    [anon_sym_AMP_GT_GT] = ACTIONS(1306),
    [anon_sym_LT_AMP] = ACTIONS(1304),
    [anon_sym_GT_AMP] = ACTIONS(1304),
    [anon_sym_GT_PIPE] = ACTIONS(1306),
    [anon_sym_LT_AMP_DASH] = ACTIONS(1306),
    [anon_sym_GT_AMP_DASH] = ACTIONS(1306),
    [anon_sym_LT_LT_DASH] = ACTIONS(1306),
    [anon_sym_LT_LT_LT] = ACTIONS(1306),
    [anon_sym_QMARK] = ACTIONS(1304),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(1306),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(1306),
    [aux_sym_concatenation_token1] = ACTIONS(1306),
    [anon_sym_DOLLAR] = ACTIONS(1304),
    [sym__special_character] = ACTIONS(1304),
    [anon_sym_DQUOTE] = ACTIONS(1306),
    [sym_raw_string] = ACTIONS(1306),
    [sym_ansi_c_string] = ACTIONS(1306),
    [aux_sym_number_token1] = ACTIONS(1304),
    [aux_sym_number_token2] = ACTIONS(1304),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(1306),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1304),
    [anon_sym_BQUOTE] = ACTIONS(1304),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(1306),
    [anon_sym_LT_LPAREN] = ACTIONS(1306),
    [anon_sym_GT_LPAREN] = ACTIONS(1306),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(1306),
    [sym__concat] = ACTIONS(1306),
    [sym_test_operator] = ACTIONS(1306),
    [sym__bare_dollar] = ACTIONS(1306),
    [sym__brace_start] = ACTIONS(1306),
  },
  [519] = {
    [sym_word] = ACTIONS(1308),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(1310),
    [anon_sym_EQ] = ACTIONS(1308),
    [anon_sym_PLUS_PLUS] = ACTIONS(1308),
    [anon_sym_DASH_DASH] = ACTIONS(1308),
    [anon_sym_PLUS_EQ] = ACTIONS(1308),
    [anon_sym_DASH_EQ] = ACTIONS(1308),
    [anon_sym_STAR_EQ] = ACTIONS(1308),
    [anon_sym_SLASH_EQ] = ACTIONS(1308),
    [anon_sym_PERCENT_EQ] = ACTIONS(1308),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(1308),
    [anon_sym_LT_LT_EQ] = ACTIONS(1310),
    [anon_sym_GT_GT_EQ] = ACTIONS(1310),
    [anon_sym_AMP_EQ] = ACTIONS(1310),
    [anon_sym_CARET_EQ] = ACTIONS(1308),
    [anon_sym_PIPE_EQ] = ACTIONS(1310),
    [anon_sym_PIPE_PIPE] = ACTIONS(1310),
    [anon_sym_AMP_AMP] = ACTIONS(1310),
    [anon_sym_PIPE] = ACTIONS(1308),
    [anon_sym_CARET] = ACTIONS(1308),
    [anon_sym_AMP] = ACTIONS(1308),
    [anon_sym_EQ_EQ] = ACTIONS(1308),
    [anon_sym_BANG_EQ] = ACTIONS(1308),
    [anon_sym_LT] = ACTIONS(1308),
    [anon_sym_GT] = ACTIONS(1308),
    [anon_sym_LT_EQ] = ACTIONS(1310),
    [anon_sym_GT_EQ] = ACTIONS(1310),
    [anon_sym_LT_LT] = ACTIONS(1308),
    [anon_sym_GT_GT] = ACTIONS(1308),
    [anon_sym_PLUS] = ACTIONS(1308),
    [anon_sym_DASH] = ACTIONS(1308),
    [anon_sym_STAR] = ACTIONS(1308),
    [anon_sym_SLASH] = ACTIONS(1308),
    [anon_sym_PERCENT] = ACTIONS(1308),
    [anon_sym_STAR_STAR] = ACTIONS(1308),
    [anon_sym_LPAREN] = ACTIONS(1308),
    [anon_sym_PIPE_AMP] = ACTIONS(1310),
    [anon_sym_RBRACK] = ACTIONS(1310),
    [anon_sym_EQ_TILDE] = ACTIONS(1308),
    [anon_sym_AMP_GT] = ACTIONS(1308),
    [anon_sym_AMP_GT_GT] = ACTIONS(1310),
    [anon_sym_LT_AMP] = ACTIONS(1308),
    [anon_sym_GT_AMP] = ACTIONS(1308),
    [anon_sym_GT_PIPE] = ACTIONS(1310),
    [anon_sym_LT_AMP_DASH] = ACTIONS(1310),
    [anon_sym_GT_AMP_DASH] = ACTIONS(1310),
    [anon_sym_LT_LT_DASH] = ACTIONS(1310),
    [anon_sym_LT_LT_LT] = ACTIONS(1310),
    [anon_sym_QMARK] = ACTIONS(1308),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(1310),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(1310),
    [aux_sym_concatenation_token1] = ACTIONS(1310),
    [anon_sym_DOLLAR] = ACTIONS(1308),
    [sym__special_character] = ACTIONS(1308),
    [anon_sym_DQUOTE] = ACTIONS(1310),
    [sym_raw_string] = ACTIONS(1310),
    [sym_ansi_c_string] = ACTIONS(1310),
    [aux_sym_number_token1] = ACTIONS(1308),
    [aux_sym_number_token2] = ACTIONS(1308),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(1310),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1308),
    [anon_sym_BQUOTE] = ACTIONS(1308),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(1310),
    [anon_sym_LT_LPAREN] = ACTIONS(1310),
    [anon_sym_GT_LPAREN] = ACTIONS(1310),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(1310),
    [sym__concat] = ACTIONS(1310),
    [sym_test_operator] = ACTIONS(1310),
    [sym__bare_dollar] = ACTIONS(1310),
    [sym__brace_start] = ACTIONS(1310),
  },
  [520] = {
    [sym_word] = ACTIONS(1312),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(1314),
    [anon_sym_EQ] = ACTIONS(1312),
    [anon_sym_PLUS_PLUS] = ACTIONS(1312),
    [anon_sym_DASH_DASH] = ACTIONS(1312),
    [anon_sym_PLUS_EQ] = ACTIONS(1312),
    [anon_sym_DASH_EQ] = ACTIONS(1312),
    [anon_sym_STAR_EQ] = ACTIONS(1312),
    [anon_sym_SLASH_EQ] = ACTIONS(1312),
    [anon_sym_PERCENT_EQ] = ACTIONS(1312),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(1312),
    [anon_sym_LT_LT_EQ] = ACTIONS(1314),
    [anon_sym_GT_GT_EQ] = ACTIONS(1314),
    [anon_sym_AMP_EQ] = ACTIONS(1314),
    [anon_sym_CARET_EQ] = ACTIONS(1312),
    [anon_sym_PIPE_EQ] = ACTIONS(1314),
    [anon_sym_PIPE_PIPE] = ACTIONS(1314),
    [anon_sym_AMP_AMP] = ACTIONS(1314),
    [anon_sym_PIPE] = ACTIONS(1312),
    [anon_sym_CARET] = ACTIONS(1312),
    [anon_sym_AMP] = ACTIONS(1312),
    [anon_sym_EQ_EQ] = ACTIONS(1312),
    [anon_sym_BANG_EQ] = ACTIONS(1312),
    [anon_sym_LT] = ACTIONS(1312),
    [anon_sym_GT] = ACTIONS(1312),
    [anon_sym_LT_EQ] = ACTIONS(1314),
    [anon_sym_GT_EQ] = ACTIONS(1314),
    [anon_sym_LT_LT] = ACTIONS(1312),
    [anon_sym_GT_GT] = ACTIONS(1312),
    [anon_sym_PLUS] = ACTIONS(1312),
    [anon_sym_DASH] = ACTIONS(1312),
    [anon_sym_STAR] = ACTIONS(1312),
    [anon_sym_SLASH] = ACTIONS(1312),
    [anon_sym_PERCENT] = ACTIONS(1312),
    [anon_sym_STAR_STAR] = ACTIONS(1312),
    [anon_sym_LPAREN] = ACTIONS(1312),
    [anon_sym_PIPE_AMP] = ACTIONS(1314),
    [anon_sym_RBRACK] = ACTIONS(1314),
    [anon_sym_EQ_TILDE] = ACTIONS(1312),
    [anon_sym_AMP_GT] = ACTIONS(1312),
    [anon_sym_AMP_GT_GT] = ACTIONS(1314),
    [anon_sym_LT_AMP] = ACTIONS(1312),
    [anon_sym_GT_AMP] = ACTIONS(1312),
    [anon_sym_GT_PIPE] = ACTIONS(1314),
    [anon_sym_LT_AMP_DASH] = ACTIONS(1314),
    [anon_sym_GT_AMP_DASH] = ACTIONS(1314),
    [anon_sym_LT_LT_DASH] = ACTIONS(1314),
    [anon_sym_LT_LT_LT] = ACTIONS(1314),
    [anon_sym_QMARK] = ACTIONS(1312),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(1314),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(1314),
    [aux_sym_concatenation_token1] = ACTIONS(1314),
    [anon_sym_DOLLAR] = ACTIONS(1312),
    [sym__special_character] = ACTIONS(1312),
    [anon_sym_DQUOTE] = ACTIONS(1314),
    [sym_raw_string] = ACTIONS(1314),
    [sym_ansi_c_string] = ACTIONS(1314),
    [aux_sym_number_token1] = ACTIONS(1312),
    [aux_sym_number_token2] = ACTIONS(1312),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(1314),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1312),
    [anon_sym_BQUOTE] = ACTIONS(1312),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(1314),
    [anon_sym_LT_LPAREN] = ACTIONS(1314),
    [anon_sym_GT_LPAREN] = ACTIONS(1314),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(1314),
    [sym__concat] = ACTIONS(1314),
    [sym_test_operator] = ACTIONS(1314),
    [sym__bare_dollar] = ACTIONS(1314),
    [sym__brace_start] = ACTIONS(1314),
  },
  [521] = {
    [sym_word] = ACTIONS(1340),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(1342),
    [anon_sym_EQ] = ACTIONS(1340),
    [anon_sym_PLUS_PLUS] = ACTIONS(1340),
    [anon_sym_DASH_DASH] = ACTIONS(1340),
    [anon_sym_PLUS_EQ] = ACTIONS(1340),
    [anon_sym_DASH_EQ] = ACTIONS(1340),
    [anon_sym_STAR_EQ] = ACTIONS(1340),
    [anon_sym_SLASH_EQ] = ACTIONS(1340),
    [anon_sym_PERCENT_EQ] = ACTIONS(1340),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(1340),
    [anon_sym_LT_LT_EQ] = ACTIONS(1342),
    [anon_sym_GT_GT_EQ] = ACTIONS(1342),
    [anon_sym_AMP_EQ] = ACTIONS(1342),
    [anon_sym_CARET_EQ] = ACTIONS(1340),
    [anon_sym_PIPE_EQ] = ACTIONS(1342),
    [anon_sym_PIPE_PIPE] = ACTIONS(1342),
    [anon_sym_AMP_AMP] = ACTIONS(1342),
    [anon_sym_PIPE] = ACTIONS(1340),
    [anon_sym_CARET] = ACTIONS(1340),
    [anon_sym_AMP] = ACTIONS(1340),
    [anon_sym_EQ_EQ] = ACTIONS(1340),
    [anon_sym_BANG_EQ] = ACTIONS(1340),
    [anon_sym_LT] = ACTIONS(1340),
    [anon_sym_GT] = ACTIONS(1340),
    [anon_sym_LT_EQ] = ACTIONS(1342),
    [anon_sym_GT_EQ] = ACTIONS(1342),
    [anon_sym_LT_LT] = ACTIONS(1340),
    [anon_sym_GT_GT] = ACTIONS(1340),
    [anon_sym_PLUS] = ACTIONS(1340),
    [anon_sym_DASH] = ACTIONS(1340),
    [anon_sym_STAR] = ACTIONS(1340),
    [anon_sym_SLASH] = ACTIONS(1340),
    [anon_sym_PERCENT] = ACTIONS(1340),
    [anon_sym_STAR_STAR] = ACTIONS(1340),
    [anon_sym_LPAREN] = ACTIONS(1340),
    [anon_sym_PIPE_AMP] = ACTIONS(1342),
    [anon_sym_RBRACK] = ACTIONS(1342),
    [anon_sym_EQ_TILDE] = ACTIONS(1340),
    [anon_sym_AMP_GT] = ACTIONS(1340),
    [anon_sym_AMP_GT_GT] = ACTIONS(1342),
    [anon_sym_LT_AMP] = ACTIONS(1340),
    [anon_sym_GT_AMP] = ACTIONS(1340),
    [anon_sym_GT_PIPE] = ACTIONS(1342),
    [anon_sym_LT_AMP_DASH] = ACTIONS(1342),
    [anon_sym_GT_AMP_DASH] = ACTIONS(1342),
    [anon_sym_LT_LT_DASH] = ACTIONS(1342),
    [anon_sym_LT_LT_LT] = ACTIONS(1342),
    [anon_sym_QMARK] = ACTIONS(1340),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(1342),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(1342),
    [aux_sym_concatenation_token1] = ACTIONS(1342),
    [anon_sym_DOLLAR] = ACTIONS(1340),
    [sym__special_character] = ACTIONS(1340),
    [anon_sym_DQUOTE] = ACTIONS(1342),
    [sym_raw_string] = ACTIONS(1342),
    [sym_ansi_c_string] = ACTIONS(1342),
    [aux_sym_number_token1] = ACTIONS(1340),
    [aux_sym_number_token2] = ACTIONS(1340),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(1342),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1340),
    [anon_sym_BQUOTE] = ACTIONS(1340),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(1342),
    [anon_sym_LT_LPAREN] = ACTIONS(1342),
    [anon_sym_GT_LPAREN] = ACTIONS(1342),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(1342),
    [sym__concat] = ACTIONS(1342),
    [sym_test_operator] = ACTIONS(1342),
    [sym__bare_dollar] = ACTIONS(1342),
    [sym__brace_start] = ACTIONS(1342),
  },
  [522] = {
    [aux_sym__literal_repeat1] = STATE(523),
    [sym_word] = ACTIONS(241),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(278),
    [anon_sym_EQ] = ACTIONS(243),
    [anon_sym_PLUS_PLUS] = ACTIONS(243),
    [anon_sym_DASH_DASH] = ACTIONS(243),
    [anon_sym_PLUS_EQ] = ACTIONS(243),
    [anon_sym_DASH_EQ] = ACTIONS(243),
    [anon_sym_STAR_EQ] = ACTIONS(243),
    [anon_sym_SLASH_EQ] = ACTIONS(243),
    [anon_sym_PERCENT_EQ] = ACTIONS(243),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(243),
    [anon_sym_LT_LT_EQ] = ACTIONS(354),
    [anon_sym_GT_GT_EQ] = ACTIONS(354),
    [anon_sym_AMP_EQ] = ACTIONS(354),
    [anon_sym_CARET_EQ] = ACTIONS(243),
    [anon_sym_PIPE_EQ] = ACTIONS(354),
    [anon_sym_PIPE_PIPE] = ACTIONS(356),
    [anon_sym_AMP_AMP] = ACTIONS(356),
    [anon_sym_PIPE] = ACTIONS(245),
    [anon_sym_CARET] = ACTIONS(243),
    [anon_sym_AMP] = ACTIONS(243),
    [anon_sym_EQ_EQ] = ACTIONS(245),
    [anon_sym_BANG_EQ] = ACTIONS(243),
    [anon_sym_LT] = ACTIONS(245),
    [anon_sym_GT] = ACTIONS(245),
    [anon_sym_LT_EQ] = ACTIONS(354),
    [anon_sym_GT_EQ] = ACTIONS(354),
    [anon_sym_LT_LT] = ACTIONS(245),
    [anon_sym_GT_GT] = ACTIONS(245),
    [anon_sym_PLUS] = ACTIONS(243),
    [anon_sym_DASH] = ACTIONS(243),
    [anon_sym_STAR] = ACTIONS(243),
    [anon_sym_SLASH] = ACTIONS(243),
    [anon_sym_PERCENT] = ACTIONS(243),
    [anon_sym_STAR_STAR] = ACTIONS(243),
    [anon_sym_LPAREN] = ACTIONS(241),
    [anon_sym_PIPE_AMP] = ACTIONS(278),
    [anon_sym_RBRACK] = ACTIONS(354),
    [anon_sym_EQ_TILDE] = ACTIONS(245),
    [anon_sym_AMP_GT] = ACTIONS(241),
    [anon_sym_AMP_GT_GT] = ACTIONS(278),
    [anon_sym_LT_AMP] = ACTIONS(241),
    [anon_sym_GT_AMP] = ACTIONS(241),
    [anon_sym_GT_PIPE] = ACTIONS(278),
    [anon_sym_LT_AMP_DASH] = ACTIONS(278),
    [anon_sym_GT_AMP_DASH] = ACTIONS(278),
    [anon_sym_LT_LT_DASH] = ACTIONS(278),
    [anon_sym_LT_LT_LT] = ACTIONS(278),
    [anon_sym_QMARK] = ACTIONS(243),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(278),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(278),
    [anon_sym_DOLLAR] = ACTIONS(241),
    [sym__special_character] = ACTIONS(1381),
    [anon_sym_DQUOTE] = ACTIONS(278),
    [sym_raw_string] = ACTIONS(278),
    [sym_ansi_c_string] = ACTIONS(278),
    [aux_sym_number_token1] = ACTIONS(241),
    [aux_sym_number_token2] = ACTIONS(241),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(278),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(241),
    [anon_sym_BQUOTE] = ACTIONS(278),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(278),
    [anon_sym_LT_LPAREN] = ACTIONS(278),
    [anon_sym_GT_LPAREN] = ACTIONS(278),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(278),
    [sym_test_operator] = ACTIONS(356),
    [sym__bare_dollar] = ACTIONS(278),
    [sym__brace_start] = ACTIONS(278),
  },
  [523] = {
    [aux_sym__literal_repeat1] = STATE(523),
    [sym_word] = ACTIONS(1362),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(1364),
    [anon_sym_EQ] = ACTIONS(1362),
    [anon_sym_PLUS_PLUS] = ACTIONS(1362),
    [anon_sym_DASH_DASH] = ACTIONS(1362),
    [anon_sym_PLUS_EQ] = ACTIONS(1362),
    [anon_sym_DASH_EQ] = ACTIONS(1362),
    [anon_sym_STAR_EQ] = ACTIONS(1362),
    [anon_sym_SLASH_EQ] = ACTIONS(1362),
    [anon_sym_PERCENT_EQ] = ACTIONS(1362),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(1362),
    [anon_sym_LT_LT_EQ] = ACTIONS(1364),
    [anon_sym_GT_GT_EQ] = ACTIONS(1364),
    [anon_sym_AMP_EQ] = ACTIONS(1364),
    [anon_sym_CARET_EQ] = ACTIONS(1362),
    [anon_sym_PIPE_EQ] = ACTIONS(1364),
    [anon_sym_PIPE_PIPE] = ACTIONS(1364),
    [anon_sym_AMP_AMP] = ACTIONS(1364),
    [anon_sym_PIPE] = ACTIONS(1362),
    [anon_sym_CARET] = ACTIONS(1362),
    [anon_sym_AMP] = ACTIONS(1362),
    [anon_sym_EQ_EQ] = ACTIONS(1362),
    [anon_sym_BANG_EQ] = ACTIONS(1362),
    [anon_sym_LT] = ACTIONS(1362),
    [anon_sym_GT] = ACTIONS(1362),
    [anon_sym_LT_EQ] = ACTIONS(1364),
    [anon_sym_GT_EQ] = ACTIONS(1364),
    [anon_sym_LT_LT] = ACTIONS(1362),
    [anon_sym_GT_GT] = ACTIONS(1362),
    [anon_sym_PLUS] = ACTIONS(1362),
    [anon_sym_DASH] = ACTIONS(1362),
    [anon_sym_STAR] = ACTIONS(1362),
    [anon_sym_SLASH] = ACTIONS(1362),
    [anon_sym_PERCENT] = ACTIONS(1362),
    [anon_sym_STAR_STAR] = ACTIONS(1362),
    [anon_sym_LPAREN] = ACTIONS(1362),
    [anon_sym_PIPE_AMP] = ACTIONS(1364),
    [anon_sym_RBRACK] = ACTIONS(1364),
    [anon_sym_EQ_TILDE] = ACTIONS(1362),
    [anon_sym_AMP_GT] = ACTIONS(1362),
    [anon_sym_AMP_GT_GT] = ACTIONS(1364),
    [anon_sym_LT_AMP] = ACTIONS(1362),
    [anon_sym_GT_AMP] = ACTIONS(1362),
    [anon_sym_GT_PIPE] = ACTIONS(1364),
    [anon_sym_LT_AMP_DASH] = ACTIONS(1364),
    [anon_sym_GT_AMP_DASH] = ACTIONS(1364),
    [anon_sym_LT_LT_DASH] = ACTIONS(1364),
    [anon_sym_LT_LT_LT] = ACTIONS(1364),
    [anon_sym_QMARK] = ACTIONS(1362),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(1364),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(1364),
    [anon_sym_DOLLAR] = ACTIONS(1362),
    [sym__special_character] = ACTIONS(1383),
    [anon_sym_DQUOTE] = ACTIONS(1364),
    [sym_raw_string] = ACTIONS(1364),
    [sym_ansi_c_string] = ACTIONS(1364),
    [aux_sym_number_token1] = ACTIONS(1362),
    [aux_sym_number_token2] = ACTIONS(1362),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(1364),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1362),
    [anon_sym_BQUOTE] = ACTIONS(1364),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(1364),
    [anon_sym_LT_LPAREN] = ACTIONS(1364),
    [anon_sym_GT_LPAREN] = ACTIONS(1364),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(1364),
    [sym_test_operator] = ACTIONS(1364),
    [sym__bare_dollar] = ACTIONS(1364),
    [sym__brace_start] = ACTIONS(1364),
  },
  [524] = {
    [sym_word] = ACTIONS(1271),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(1281),
    [anon_sym_EQ] = ACTIONS(1273),
    [anon_sym_PLUS_PLUS] = ACTIONS(1273),
    [anon_sym_DASH_DASH] = ACTIONS(1273),
    [anon_sym_PLUS_EQ] = ACTIONS(1273),
    [anon_sym_DASH_EQ] = ACTIONS(1273),
    [anon_sym_STAR_EQ] = ACTIONS(1273),
    [anon_sym_SLASH_EQ] = ACTIONS(1273),
    [anon_sym_PERCENT_EQ] = ACTIONS(1273),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(1273),
    [anon_sym_LT_LT_EQ] = ACTIONS(1369),
    [anon_sym_GT_GT_EQ] = ACTIONS(1369),
    [anon_sym_AMP_EQ] = ACTIONS(1369),
    [anon_sym_CARET_EQ] = ACTIONS(1273),
    [anon_sym_PIPE_EQ] = ACTIONS(1369),
    [anon_sym_PIPE_PIPE] = ACTIONS(1283),
    [anon_sym_AMP_AMP] = ACTIONS(1283),
    [anon_sym_PIPE] = ACTIONS(1275),
    [anon_sym_CARET] = ACTIONS(1273),
    [anon_sym_AMP] = ACTIONS(1273),
    [anon_sym_EQ_EQ] = ACTIONS(1275),
    [anon_sym_BANG_EQ] = ACTIONS(1273),
    [anon_sym_LT] = ACTIONS(1275),
    [anon_sym_GT] = ACTIONS(1275),
    [anon_sym_LT_EQ] = ACTIONS(1369),
    [anon_sym_GT_EQ] = ACTIONS(1369),
    [anon_sym_LT_LT] = ACTIONS(1275),
    [anon_sym_GT_GT] = ACTIONS(1275),
    [anon_sym_PLUS] = ACTIONS(1273),
    [anon_sym_DASH] = ACTIONS(1273),
    [anon_sym_STAR] = ACTIONS(1273),
    [anon_sym_SLASH] = ACTIONS(1273),
    [anon_sym_PERCENT] = ACTIONS(1273),
    [anon_sym_STAR_STAR] = ACTIONS(1273),
    [anon_sym_LPAREN] = ACTIONS(1271),
    [anon_sym_PIPE_AMP] = ACTIONS(1281),
    [anon_sym_RBRACK] = ACTIONS(1369),
    [anon_sym_EQ_TILDE] = ACTIONS(1275),
    [anon_sym_AMP_GT] = ACTIONS(1271),
    [anon_sym_AMP_GT_GT] = ACTIONS(1281),
    [anon_sym_LT_AMP] = ACTIONS(1271),
    [anon_sym_GT_AMP] = ACTIONS(1271),
    [anon_sym_GT_PIPE] = ACTIONS(1281),
    [anon_sym_LT_AMP_DASH] = ACTIONS(1281),
    [anon_sym_GT_AMP_DASH] = ACTIONS(1281),
    [anon_sym_LT_LT_DASH] = ACTIONS(1281),
    [anon_sym_LT_LT_LT] = ACTIONS(1281),
    [anon_sym_QMARK] = ACTIONS(1273),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(1281),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(1281),
    [anon_sym_DOLLAR] = ACTIONS(1271),
    [sym__special_character] = ACTIONS(1271),
    [anon_sym_DQUOTE] = ACTIONS(1281),
    [sym_raw_string] = ACTIONS(1281),
    [sym_ansi_c_string] = ACTIONS(1281),
    [aux_sym_number_token1] = ACTIONS(1271),
    [aux_sym_number_token2] = ACTIONS(1271),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(1281),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1271),
    [anon_sym_BQUOTE] = ACTIONS(1281),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(1281),
    [anon_sym_LT_LPAREN] = ACTIONS(1281),
    [anon_sym_GT_LPAREN] = ACTIONS(1281),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(1281),
    [sym_test_operator] = ACTIONS(1283),
    [sym__bare_dollar] = ACTIONS(1281),
    [sym__brace_start] = ACTIONS(1281),
  },
  [525] = {
    [sym_word] = ACTIONS(1261),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(1263),
    [anon_sym_EQ] = ACTIONS(1261),
    [anon_sym_PLUS_PLUS] = ACTIONS(1261),
    [anon_sym_DASH_DASH] = ACTIONS(1261),
    [anon_sym_PLUS_EQ] = ACTIONS(1261),
    [anon_sym_DASH_EQ] = ACTIONS(1261),
    [anon_sym_STAR_EQ] = ACTIONS(1261),
    [anon_sym_SLASH_EQ] = ACTIONS(1261),
    [anon_sym_PERCENT_EQ] = ACTIONS(1261),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(1261),
    [anon_sym_LT_LT_EQ] = ACTIONS(1263),
    [anon_sym_GT_GT_EQ] = ACTIONS(1263),
    [anon_sym_AMP_EQ] = ACTIONS(1263),
    [anon_sym_CARET_EQ] = ACTIONS(1261),
    [anon_sym_PIPE_EQ] = ACTIONS(1263),
    [anon_sym_PIPE_PIPE] = ACTIONS(1263),
    [anon_sym_AMP_AMP] = ACTIONS(1263),
    [anon_sym_PIPE] = ACTIONS(1261),
    [anon_sym_CARET] = ACTIONS(1261),
    [anon_sym_AMP] = ACTIONS(1261),
    [anon_sym_EQ_EQ] = ACTIONS(1261),
    [anon_sym_BANG_EQ] = ACTIONS(1261),
    [anon_sym_LT] = ACTIONS(1261),
    [anon_sym_GT] = ACTIONS(1261),
    [anon_sym_LT_EQ] = ACTIONS(1263),
    [anon_sym_GT_EQ] = ACTIONS(1263),
    [anon_sym_LT_LT] = ACTIONS(1261),
    [anon_sym_GT_GT] = ACTIONS(1261),
    [anon_sym_PLUS] = ACTIONS(1261),
    [anon_sym_DASH] = ACTIONS(1261),
    [anon_sym_STAR] = ACTIONS(1261),
    [anon_sym_SLASH] = ACTIONS(1261),
    [anon_sym_PERCENT] = ACTIONS(1261),
    [anon_sym_STAR_STAR] = ACTIONS(1261),
    [anon_sym_LPAREN] = ACTIONS(1261),
    [anon_sym_PIPE_AMP] = ACTIONS(1263),
    [anon_sym_RBRACK] = ACTIONS(1263),
    [anon_sym_EQ_TILDE] = ACTIONS(1261),
    [anon_sym_AMP_GT] = ACTIONS(1261),
    [anon_sym_AMP_GT_GT] = ACTIONS(1263),
    [anon_sym_LT_AMP] = ACTIONS(1261),
    [anon_sym_GT_AMP] = ACTIONS(1261),
    [anon_sym_GT_PIPE] = ACTIONS(1263),
    [anon_sym_LT_AMP_DASH] = ACTIONS(1263),
    [anon_sym_GT_AMP_DASH] = ACTIONS(1263),
    [anon_sym_LT_LT_DASH] = ACTIONS(1263),
    [anon_sym_LT_LT_LT] = ACTIONS(1263),
    [anon_sym_QMARK] = ACTIONS(1261),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(1263),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(1263),
    [anon_sym_DOLLAR] = ACTIONS(1261),
    [sym__special_character] = ACTIONS(1261),
    [anon_sym_DQUOTE] = ACTIONS(1263),
    [sym_raw_string] = ACTIONS(1263),
    [sym_ansi_c_string] = ACTIONS(1263),
    [aux_sym_number_token1] = ACTIONS(1261),
    [aux_sym_number_token2] = ACTIONS(1261),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(1263),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(1261),
    [anon_sym_BQUOTE] = ACTIONS(1263),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(1263),
    [anon_sym_LT_LPAREN] = ACTIONS(1263),
    [anon_sym_GT_LPAREN] = ACTIONS(1263),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(1263),
    [sym_test_operator] = ACTIONS(1263),
    [sym__bare_dollar] = ACTIONS(1263),
    [sym__brace_start] = ACTIONS(1263),
  },
  [526] = {
    [sym_subshell] = STATE(5014),
    [sym_test_command] = STATE(5014),
    [sym_command] = STATE(5026),
    [sym_command_name] = STATE(533),
    [sym_variable_assignment] = STATE(1974),
    [sym_subscript] = STATE(6754),
    [sym_file_redirect] = STATE(3581),
    [sym_herestring_redirect] = STATE(3581),
    [sym__expression] = STATE(3100),
    [sym_binary_expression] = STATE(3053),
    [sym_ternary_expression] = STATE(3053),
    [sym_unary_expression] = STATE(3053),
    [sym_postfix_expression] = STATE(3053),
    [sym_parenthesized_expression] = STATE(3053),
    [sym_arithmetic_expansion] = STATE(472),
    [sym_brace_expression] = STATE(472),
    [sym_concatenation] = STATE(502),
    [sym_string] = STATE(472),
    [sym_translated_string] = STATE(472),
    [sym_number] = STATE(472),
    [sym_simple_expansion] = STATE(472),
    [sym_expansion] = STATE(472),
    [sym_command_substitution] = STATE(472),
    [sym_process_substitution] = STATE(472),
    [aux_sym_command_repeat1] = STATE(951),
    [aux_sym__literal_repeat1] = STATE(493),
    [sym_word] = ACTIONS(1386),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(1388),
    [anon_sym_GT] = ACTIONS(1388),
    [anon_sym_GT_GT] = ACTIONS(1390),
    [anon_sym_LPAREN] = ACTIONS(89),
    [anon_sym_BANG] = ACTIONS(250),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(35),
    [anon_sym_AMP_GT] = ACTIONS(1388),
    [anon_sym_AMP_GT_GT] = ACTIONS(1390),
    [anon_sym_LT_AMP] = ACTIONS(1388),
    [anon_sym_GT_AMP] = ACTIONS(1388),
    [anon_sym_GT_PIPE] = ACTIONS(1390),
    [anon_sym_LT_AMP_DASH] = ACTIONS(1392),
    [anon_sym_GT_AMP_DASH] = ACTIONS(1392),
    [anon_sym_LT_LT_LT] = ACTIONS(1394),
    [anon_sym_PLUS_PLUS2] = ACTIONS(103),
    [anon_sym_DASH_DASH2] = ACTIONS(103),
    [anon_sym_DASH2] = ACTIONS(105),
    [anon_sym_PLUS2] = ACTIONS(105),
    [anon_sym_TILDE] = ACTIONS(107),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(109),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(111),
    [anon_sym_DOLLAR] = ACTIONS(113),
    [sym__special_character] = ACTIONS(115),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [sym_raw_string] = ACTIONS(119),
    [sym_ansi_c_string] = ACTIONS(119),
    [aux_sym_number_token1] = ACTIONS(121),
    [aux_sym_number_token2] = ACTIONS(123),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(125),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(127),
    [anon_sym_BQUOTE] = ACTIONS(129),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(131),
    [anon_sym_LT_LPAREN] = ACTIONS(133),
    [anon_sym_GT_LPAREN] = ACTIONS(133),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(1396),
    [sym_variable_name] = ACTIONS(137),
    [sym_test_operator] = ACTIONS(139),
    [sym__brace_start] = ACTIONS(141),
  },
  [527] = {
    [sym_subshell] = STATE(5623),
    [sym_test_command] = STATE(5623),
    [sym_command] = STATE(5624),
    [sym_command_name] = STATE(620),
    [sym_variable_assignment] = STATE(2941),
    [sym_subscript] = STATE(6756),
    [sym_file_redirect] = STATE(3581),
    [sym_herestring_redirect] = STATE(3581),
    [sym__expression] = STATE(3011),
    [sym_binary_expression] = STATE(3071),
    [sym_ternary_expression] = STATE(3071),
    [sym_unary_expression] = STATE(3071),
    [sym_postfix_expression] = STATE(3071),
    [sym_parenthesized_expression] = STATE(3071),
    [sym_arithmetic_expansion] = STATE(496),
    [sym_brace_expression] = STATE(496),
    [sym_concatenation] = STATE(524),
    [sym_string] = STATE(496),
    [sym_translated_string] = STATE(496),
    [sym_number] = STATE(496),
    [sym_simple_expansion] = STATE(496),
    [sym_expansion] = STATE(496),
    [sym_command_substitution] = STATE(496),
    [sym_process_substitution] = STATE(496),
    [aux_sym_command_repeat1] = STATE(955),
    [aux_sym__literal_repeat1] = STATE(522),
    [sym_word] = ACTIONS(1398),
    [anon_sym_LPAREN_LPAREN] = ACTIONS(149),
    [anon_sym_LT] = ACTIONS(1388),
    [anon_sym_GT] = ACTIONS(1388),
    [anon_sym_GT_GT] = ACTIONS(1390),
    [anon_sym_LPAREN] = ACTIONS(155),
    [anon_sym_BANG] = ACTIONS(361),
    [anon_sym_LBRACK] = ACTIONS(169),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(173),
    [anon_sym_AMP_GT] = ACTIONS(1388),
    [anon_sym_AMP_GT_GT] = ACTIONS(1390),
    [anon_sym_LT_AMP] = ACTIONS(1388),
    [anon_sym_GT_AMP] = ACTIONS(1388),
    [anon_sym_GT_PIPE] = ACTIONS(1390),
    [anon_sym_LT_AMP_DASH] = ACTIONS(1392),
    [anon_sym_GT_AMP_DASH] = ACTIONS(1392),
    [anon_sym_LT_LT_LT] = ACTIONS(1394),
    [anon_sym_PLUS_PLUS2] = ACTIONS(183),
    [anon_sym_DASH_DASH2] = ACTIONS(183),
    [anon_sym_DASH2] = ACTIONS(185),
    [anon_sym_PLUS2] = ACTIONS(185),
    [anon_sym_TILDE] = ACTIONS(187),
    [anon_sym_DOLLAR_LPAREN_LPAREN] = ACTIONS(189),
    [anon_sym_DOLLAR_LBRACK] = ACTIONS(191),
    [anon_sym_DOLLAR] = ACTIONS(193),
    [sym__special_character] = ACTIONS(195),
    [anon_sym_DQUOTE] = ACTIONS(197),
    [sym_raw_string] = ACTIONS(199),
    [sym_ansi_c_string] = ACTIONS(199),
    [aux_sym_number_token1] = ACTIONS(201),
    [aux_sym_number_token2] = ACTIONS(203),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(205),
    [anon_sym_DOLLAR_LPAREN] = ACTIONS(207),
    [anon_sym_BQUOTE] = ACTIONS(209),
    [anon_sym_DOLLAR_BQUOTE] = ACTIONS(211),
    [anon_sym_LT_LPAREN] = ACTIONS(213),
    [anon_sym_GT_LPAREN] = ACTIONS(213),
    [sym_comment] = ACTIONS(71),
    [sym_file_descriptor] = ACTIONS(1396),
    [sym_variable_name] = ACTIONS(217),
    [sym_test_operator] = ACTIONS(219),
    [sym__brace_start] = ACTIONS(221),
  },
};

static const uint16_t ts_small_parse_table[] = {
  [0] = 28,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LPAREN,
    ACTIONS(1410), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(1412), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1414), 1,
      anon_sym_DOLLAR,
    ACTIONS(1416), 1,
      sym__special_character,
    ACTIONS(1418), 1,
      anon_sym_DQUOTE,
    ACTIONS(1420), 1,
      aux_sym_number_token1,
    ACTIONS(1422), 1,
      aux_sym_number_token2,
    ACTIONS(1424), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1426), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1428), 1,
      anon_sym_BQUOTE,
    ACTIONS(1430), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1434), 1,
      sym_test_operator,
    ACTIONS(1436), 1,
      sym__bare_dollar,
    ACTIONS(1438), 1,
      sym__brace_start,
    STATE(540), 1,
      aux_sym_command_repeat2,
    STATE(1045), 1,
      aux_sym__literal_repeat1,
    STATE(1181), 1,
      sym_herestring_redirect,
    STATE(1182), 1,
      sym_concatenation,
    STATE(4991), 1,
      sym_subshell,
    ACTIONS(1402), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1406), 2,
      anon_sym_EQ_EQ,
      anon_sym_EQ_TILDE,
    ACTIONS(1408), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1432), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1400), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(874), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(1404), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [120] = 28,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LPAREN,
    ACTIONS(1410), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(1412), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1414), 1,
      anon_sym_DOLLAR,
    ACTIONS(1416), 1,
      sym__special_character,
    ACTIONS(1418), 1,
      anon_sym_DQUOTE,
    ACTIONS(1420), 1,
      aux_sym_number_token1,
    ACTIONS(1422), 1,
      aux_sym_number_token2,
    ACTIONS(1424), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1426), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1428), 1,
      anon_sym_BQUOTE,
    ACTIONS(1430), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1434), 1,
      sym_test_operator,
    ACTIONS(1436), 1,
      sym__bare_dollar,
    ACTIONS(1438), 1,
      sym__brace_start,
    STATE(535), 1,
      aux_sym_command_repeat2,
    STATE(1045), 1,
      aux_sym__literal_repeat1,
    STATE(1181), 1,
      sym_herestring_redirect,
    STATE(1182), 1,
      sym_concatenation,
    STATE(4977), 1,
      sym_subshell,
    ACTIONS(1402), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1406), 2,
      anon_sym_EQ_EQ,
      anon_sym_EQ_TILDE,
    ACTIONS(1432), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1442), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1400), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(874), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(1440), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [240] = 28,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(677), 1,
      anon_sym_LPAREN,
    ACTIONS(1450), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(1452), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1454), 1,
      anon_sym_DOLLAR,
    ACTIONS(1456), 1,
      sym__special_character,
    ACTIONS(1458), 1,
      anon_sym_DQUOTE,
    ACTIONS(1460), 1,
      aux_sym_number_token1,
    ACTIONS(1462), 1,
      aux_sym_number_token2,
    ACTIONS(1464), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1466), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1468), 1,
      anon_sym_BQUOTE,
    ACTIONS(1470), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1474), 1,
      sym_test_operator,
    ACTIONS(1476), 1,
      sym__bare_dollar,
    ACTIONS(1478), 1,
      sym__brace_start,
    STATE(542), 1,
      aux_sym_command_repeat2,
    STATE(1175), 1,
      aux_sym__literal_repeat1,
    STATE(1328), 1,
      sym_concatenation,
    STATE(1329), 1,
      sym_herestring_redirect,
    STATE(5116), 1,
      sym_subshell,
    ACTIONS(1408), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1446), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1448), 2,
      anon_sym_EQ_EQ,
      anon_sym_EQ_TILDE,
    ACTIONS(1472), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1444), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(985), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(1404), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [359] = 28,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(677), 1,
      anon_sym_LPAREN,
    ACTIONS(1450), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(1452), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1454), 1,
      anon_sym_DOLLAR,
    ACTIONS(1456), 1,
      sym__special_character,
    ACTIONS(1458), 1,
      anon_sym_DQUOTE,
    ACTIONS(1460), 1,
      aux_sym_number_token1,
    ACTIONS(1462), 1,
      aux_sym_number_token2,
    ACTIONS(1464), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1466), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1468), 1,
      anon_sym_BQUOTE,
    ACTIONS(1470), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1474), 1,
      sym_test_operator,
    ACTIONS(1476), 1,
      sym__bare_dollar,
    ACTIONS(1478), 1,
      sym__brace_start,
    STATE(547), 1,
      aux_sym_command_repeat2,
    STATE(1175), 1,
      aux_sym__literal_repeat1,
    STATE(1328), 1,
      sym_concatenation,
    STATE(1329), 1,
      sym_herestring_redirect,
    STATE(5111), 1,
      sym_subshell,
    ACTIONS(1442), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1446), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1448), 2,
      anon_sym_EQ_EQ,
      anon_sym_EQ_TILDE,
    ACTIONS(1472), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1444), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(985), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(1440), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [478] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1491), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1493), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(1496), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1499), 1,
      anon_sym_DOLLAR,
    ACTIONS(1502), 1,
      sym__special_character,
    ACTIONS(1505), 1,
      anon_sym_DQUOTE,
    ACTIONS(1508), 1,
      aux_sym_number_token1,
    ACTIONS(1511), 1,
      aux_sym_number_token2,
    ACTIONS(1514), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1517), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1520), 1,
      anon_sym_BQUOTE,
    ACTIONS(1523), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1529), 1,
      sym_file_descriptor,
    ACTIONS(1532), 1,
      sym_test_operator,
    ACTIONS(1535), 1,
      sym__bare_dollar,
    ACTIONS(1538), 1,
      sym__brace_start,
    STATE(532), 1,
      aux_sym_command_repeat2,
    STATE(1045), 1,
      aux_sym__literal_repeat1,
    STATE(1181), 1,
      sym_herestring_redirect,
    STATE(1182), 1,
      sym_concatenation,
    ACTIONS(1483), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1488), 2,
      anon_sym_EQ_EQ,
      anon_sym_EQ_TILDE,
    ACTIONS(1526), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1480), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(874), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(1486), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [594] = 28,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(19), 1,
      anon_sym_LPAREN,
    ACTIONS(1547), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(1549), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1551), 1,
      anon_sym_DOLLAR,
    ACTIONS(1553), 1,
      sym__special_character,
    ACTIONS(1555), 1,
      anon_sym_DQUOTE,
    ACTIONS(1557), 1,
      aux_sym_number_token1,
    ACTIONS(1559), 1,
      aux_sym_number_token2,
    ACTIONS(1561), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1563), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1565), 1,
      anon_sym_BQUOTE,
    ACTIONS(1567), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1571), 1,
      sym_test_operator,
    ACTIONS(1573), 1,
      sym__bare_dollar,
    ACTIONS(1575), 1,
      sym__brace_start,
    STATE(566), 1,
      aux_sym_command_repeat2,
    STATE(1321), 1,
      aux_sym__literal_repeat1,
    STATE(1712), 1,
      sym_herestring_redirect,
    STATE(1716), 1,
      sym_concatenation,
    STATE(4911), 1,
      sym_subshell,
    ACTIONS(1408), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1543), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1545), 2,
      anon_sym_EQ_EQ,
      anon_sym_EQ_TILDE,
    ACTIONS(1569), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1541), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(1023), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(1404), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [712] = 28,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(19), 1,
      anon_sym_LPAREN,
    ACTIONS(1583), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(1585), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1587), 1,
      anon_sym_DOLLAR,
    ACTIONS(1589), 1,
      sym__special_character,
    ACTIONS(1591), 1,
      anon_sym_DQUOTE,
    ACTIONS(1593), 1,
      aux_sym_number_token1,
    ACTIONS(1595), 1,
      aux_sym_number_token2,
    ACTIONS(1597), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1599), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1601), 1,
      anon_sym_BQUOTE,
    ACTIONS(1603), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1607), 1,
      sym_test_operator,
    ACTIONS(1609), 1,
      sym__bare_dollar,
    ACTIONS(1611), 1,
      sym__brace_start,
    STATE(557), 1,
      aux_sym_command_repeat2,
    STATE(1431), 1,
      aux_sym__literal_repeat1,
    STATE(1521), 1,
      sym_concatenation,
    STATE(1552), 1,
      sym_herestring_redirect,
    STATE(4911), 1,
      sym_subshell,
    ACTIONS(1579), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1581), 2,
      anon_sym_EQ_EQ,
      anon_sym_EQ_TILDE,
    ACTIONS(1605), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1408), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1577), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(1093), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(1404), 19,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [830] = 26,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1410), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(1412), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1414), 1,
      anon_sym_DOLLAR,
    ACTIONS(1416), 1,
      sym__special_character,
    ACTIONS(1418), 1,
      anon_sym_DQUOTE,
    ACTIONS(1420), 1,
      aux_sym_number_token1,
    ACTIONS(1422), 1,
      aux_sym_number_token2,
    ACTIONS(1424), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1426), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1428), 1,
      anon_sym_BQUOTE,
    ACTIONS(1430), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1434), 1,
      sym_test_operator,
    ACTIONS(1436), 1,
      sym__bare_dollar,
    ACTIONS(1438), 1,
      sym__brace_start,
    STATE(532), 1,
      aux_sym_command_repeat2,
    STATE(1045), 1,
      aux_sym__literal_repeat1,
    STATE(1181), 1,
      sym_herestring_redirect,
    STATE(1182), 1,
      sym_concatenation,
    ACTIONS(1402), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1406), 2,
      anon_sym_EQ_EQ,
      anon_sym_EQ_TILDE,
    ACTIONS(1432), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1615), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1400), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(874), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(1613), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [944] = 28,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(19), 1,
      anon_sym_LPAREN,
    ACTIONS(1547), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(1549), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1551), 1,
      anon_sym_DOLLAR,
    ACTIONS(1553), 1,
      sym__special_character,
    ACTIONS(1555), 1,
      anon_sym_DQUOTE,
    ACTIONS(1557), 1,
      aux_sym_number_token1,
    ACTIONS(1559), 1,
      aux_sym_number_token2,
    ACTIONS(1561), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1563), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1565), 1,
      anon_sym_BQUOTE,
    ACTIONS(1567), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1571), 1,
      sym_test_operator,
    ACTIONS(1573), 1,
      sym__bare_dollar,
    ACTIONS(1575), 1,
      sym__brace_start,
    STATE(561), 1,
      aux_sym_command_repeat2,
    STATE(1321), 1,
      aux_sym__literal_repeat1,
    STATE(1712), 1,
      sym_herestring_redirect,
    STATE(1716), 1,
      sym_concatenation,
    STATE(4952), 1,
      sym_subshell,
    ACTIONS(1442), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1543), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1545), 2,
      anon_sym_EQ_EQ,
      anon_sym_EQ_TILDE,
    ACTIONS(1569), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1541), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(1023), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(1440), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [1062] = 28,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(19), 1,
      anon_sym_LPAREN,
    ACTIONS(1583), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(1585), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1587), 1,
      anon_sym_DOLLAR,
    ACTIONS(1589), 1,
      sym__special_character,
    ACTIONS(1591), 1,
      anon_sym_DQUOTE,
    ACTIONS(1593), 1,
      aux_sym_number_token1,
    ACTIONS(1595), 1,
      aux_sym_number_token2,
    ACTIONS(1597), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1599), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1601), 1,
      anon_sym_BQUOTE,
    ACTIONS(1603), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1607), 1,
      sym_test_operator,
    ACTIONS(1609), 1,
      sym__bare_dollar,
    ACTIONS(1611), 1,
      sym__brace_start,
    STATE(560), 1,
      aux_sym_command_repeat2,
    STATE(1431), 1,
      aux_sym__literal_repeat1,
    STATE(1521), 1,
      sym_concatenation,
    STATE(1552), 1,
      sym_herestring_redirect,
    STATE(4952), 1,
      sym_subshell,
    ACTIONS(1579), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1581), 2,
      anon_sym_EQ_EQ,
      anon_sym_EQ_TILDE,
    ACTIONS(1605), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1442), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1577), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(1093), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(1440), 19,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [1180] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1619), 1,
      anon_sym_DQUOTE,
    ACTIONS(1623), 1,
      sym_variable_name,
    STATE(868), 1,
      sym_string,
    ACTIONS(1621), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1241), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
    ACTIONS(1617), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1239), 42,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [1258] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1619), 1,
      anon_sym_DQUOTE,
    ACTIONS(1623), 1,
      sym_variable_name,
    STATE(868), 1,
      sym_string,
    ACTIONS(1621), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1235), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
    ACTIONS(1617), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1227), 42,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [1336] = 26,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1410), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(1412), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1414), 1,
      anon_sym_DOLLAR,
    ACTIONS(1416), 1,
      sym__special_character,
    ACTIONS(1418), 1,
      anon_sym_DQUOTE,
    ACTIONS(1420), 1,
      aux_sym_number_token1,
    ACTIONS(1422), 1,
      aux_sym_number_token2,
    ACTIONS(1424), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1426), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1428), 1,
      anon_sym_BQUOTE,
    ACTIONS(1430), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1434), 1,
      sym_test_operator,
    ACTIONS(1436), 1,
      sym__bare_dollar,
    ACTIONS(1438), 1,
      sym__brace_start,
    STATE(532), 1,
      aux_sym_command_repeat2,
    STATE(1045), 1,
      aux_sym__literal_repeat1,
    STATE(1181), 1,
      sym_herestring_redirect,
    STATE(1182), 1,
      sym_concatenation,
    ACTIONS(1402), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1406), 2,
      anon_sym_EQ_EQ,
      anon_sym_EQ_TILDE,
    ACTIONS(1432), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1627), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1400), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(874), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(1625), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [1450] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1491), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1638), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(1641), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1644), 1,
      anon_sym_DOLLAR,
    ACTIONS(1647), 1,
      sym__special_character,
    ACTIONS(1650), 1,
      anon_sym_DQUOTE,
    ACTIONS(1653), 1,
      aux_sym_number_token1,
    ACTIONS(1656), 1,
      aux_sym_number_token2,
    ACTIONS(1659), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1662), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1665), 1,
      anon_sym_BQUOTE,
    ACTIONS(1668), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1674), 1,
      sym_file_descriptor,
    ACTIONS(1677), 1,
      sym_test_operator,
    ACTIONS(1680), 1,
      sym__bare_dollar,
    ACTIONS(1683), 1,
      sym__brace_start,
    STATE(541), 1,
      aux_sym_command_repeat2,
    STATE(1175), 1,
      aux_sym__literal_repeat1,
    STATE(1328), 1,
      sym_concatenation,
    STATE(1329), 1,
      sym_herestring_redirect,
    ACTIONS(1632), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1635), 2,
      anon_sym_EQ_EQ,
      anon_sym_EQ_TILDE,
    ACTIONS(1671), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1629), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(985), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(1486), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [1565] = 26,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1450), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(1452), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1454), 1,
      anon_sym_DOLLAR,
    ACTIONS(1456), 1,
      sym__special_character,
    ACTIONS(1458), 1,
      anon_sym_DQUOTE,
    ACTIONS(1460), 1,
      aux_sym_number_token1,
    ACTIONS(1462), 1,
      aux_sym_number_token2,
    ACTIONS(1464), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1466), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1468), 1,
      anon_sym_BQUOTE,
    ACTIONS(1470), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1474), 1,
      sym_test_operator,
    ACTIONS(1476), 1,
      sym__bare_dollar,
    ACTIONS(1478), 1,
      sym__brace_start,
    STATE(541), 1,
      aux_sym_command_repeat2,
    STATE(1175), 1,
      aux_sym__literal_repeat1,
    STATE(1328), 1,
      sym_concatenation,
    STATE(1329), 1,
      sym_herestring_redirect,
    ACTIONS(1446), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1448), 2,
      anon_sym_EQ_EQ,
      anon_sym_EQ_TILDE,
    ACTIONS(1472), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1627), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1444), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(985), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(1625), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [1678] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(19), 1,
      anon_sym_LPAREN,
    ACTIONS(1549), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1551), 1,
      anon_sym_DOLLAR,
    ACTIONS(1555), 1,
      anon_sym_DQUOTE,
    ACTIONS(1557), 1,
      aux_sym_number_token1,
    ACTIONS(1559), 1,
      aux_sym_number_token2,
    ACTIONS(1561), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1563), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1567), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1573), 1,
      sym__bare_dollar,
    ACTIONS(1575), 1,
      sym__brace_start,
    ACTIONS(1690), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(1692), 1,
      sym__special_character,
    ACTIONS(1694), 1,
      sym_test_operator,
    STATE(587), 1,
      aux_sym_command_repeat2,
    STATE(1321), 1,
      aux_sym__literal_repeat1,
    STATE(1712), 1,
      sym_herestring_redirect,
    STATE(1716), 1,
      sym_concatenation,
    STATE(4952), 1,
      sym_subshell,
    ACTIONS(1442), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1543), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1569), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1688), 2,
      anon_sym_EQ_EQ,
      anon_sym_EQ_TILDE,
    ACTIONS(1686), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(1219), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(1440), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_BQUOTE,
  [1793] = 28,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(19), 1,
      anon_sym_LPAREN,
    ACTIONS(1549), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1551), 1,
      anon_sym_DOLLAR,
    ACTIONS(1555), 1,
      anon_sym_DQUOTE,
    ACTIONS(1557), 1,
      aux_sym_number_token1,
    ACTIONS(1559), 1,
      aux_sym_number_token2,
    ACTIONS(1561), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1563), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1565), 1,
      anon_sym_BQUOTE,
    ACTIONS(1567), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1573), 1,
      sym__bare_dollar,
    ACTIONS(1575), 1,
      sym__brace_start,
    ACTIONS(1690), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(1692), 1,
      sym__special_character,
    ACTIONS(1694), 1,
      sym_test_operator,
    STATE(582), 1,
      aux_sym_command_repeat2,
    STATE(1321), 1,
      aux_sym__literal_repeat1,
    STATE(1712), 1,
      sym_herestring_redirect,
    STATE(1716), 1,
      sym_concatenation,
    STATE(4911), 1,
      sym_subshell,
    ACTIONS(1408), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1543), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1569), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1688), 2,
      anon_sym_EQ_EQ,
      anon_sym_EQ_TILDE,
    ACTIONS(1686), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(1219), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(1404), 19,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [1910] = 28,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(19), 1,
      anon_sym_LPAREN,
    ACTIONS(1549), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1551), 1,
      anon_sym_DOLLAR,
    ACTIONS(1555), 1,
      anon_sym_DQUOTE,
    ACTIONS(1557), 1,
      aux_sym_number_token1,
    ACTIONS(1559), 1,
      aux_sym_number_token2,
    ACTIONS(1561), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1563), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1565), 1,
      anon_sym_BQUOTE,
    ACTIONS(1567), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1573), 1,
      sym__bare_dollar,
    ACTIONS(1575), 1,
      sym__brace_start,
    ACTIONS(1690), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(1692), 1,
      sym__special_character,
    ACTIONS(1694), 1,
      sym_test_operator,
    STATE(571), 1,
      aux_sym_command_repeat2,
    STATE(1321), 1,
      aux_sym__literal_repeat1,
    STATE(1712), 1,
      sym_herestring_redirect,
    STATE(1716), 1,
      sym_concatenation,
    STATE(4952), 1,
      sym_subshell,
    ACTIONS(1442), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1543), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1569), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1688), 2,
      anon_sym_EQ_EQ,
      anon_sym_EQ_TILDE,
    ACTIONS(1686), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(1219), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(1440), 19,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [2027] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1706), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1709), 1,
      anon_sym_DOLLAR,
    ACTIONS(1712), 1,
      sym__special_character,
    ACTIONS(1715), 1,
      anon_sym_DQUOTE,
    ACTIONS(1718), 1,
      aux_sym_number_token1,
    ACTIONS(1721), 1,
      aux_sym_number_token2,
    ACTIONS(1724), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1727), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1730), 1,
      anon_sym_BQUOTE,
    ACTIONS(1733), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1739), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(1742), 1,
      sym_variable_name,
    ACTIONS(1745), 1,
      sym_test_operator,
    ACTIONS(1748), 1,
      sym__brace_start,
    STATE(1368), 1,
      aux_sym__literal_repeat1,
    STATE(6750), 1,
      sym_subscript,
    ACTIONS(1699), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1704), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1736), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1696), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(546), 3,
      sym_variable_assignment,
      sym_concatenation,
      aux_sym_declaration_command_repeat1,
    STATE(1067), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(1702), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [2136] = 26,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1450), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(1452), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1454), 1,
      anon_sym_DOLLAR,
    ACTIONS(1456), 1,
      sym__special_character,
    ACTIONS(1458), 1,
      anon_sym_DQUOTE,
    ACTIONS(1460), 1,
      aux_sym_number_token1,
    ACTIONS(1462), 1,
      aux_sym_number_token2,
    ACTIONS(1464), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1466), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1468), 1,
      anon_sym_BQUOTE,
    ACTIONS(1470), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1474), 1,
      sym_test_operator,
    ACTIONS(1476), 1,
      sym__bare_dollar,
    ACTIONS(1478), 1,
      sym__brace_start,
    STATE(541), 1,
      aux_sym_command_repeat2,
    STATE(1175), 1,
      aux_sym__literal_repeat1,
    STATE(1328), 1,
      sym_concatenation,
    STATE(1329), 1,
      sym_herestring_redirect,
    ACTIONS(1446), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1448), 2,
      anon_sym_EQ_EQ,
      anon_sym_EQ_TILDE,
    ACTIONS(1472), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1615), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1444), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(985), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(1613), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [2249] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1759), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1761), 1,
      anon_sym_DOLLAR,
    ACTIONS(1763), 1,
      sym__special_character,
    ACTIONS(1765), 1,
      anon_sym_DQUOTE,
    ACTIONS(1767), 1,
      aux_sym_number_token1,
    ACTIONS(1769), 1,
      aux_sym_number_token2,
    ACTIONS(1771), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1773), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1775), 1,
      anon_sym_BQUOTE,
    ACTIONS(1777), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1781), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(1783), 1,
      sym_variable_name,
    ACTIONS(1785), 1,
      sym_test_operator,
    ACTIONS(1787), 1,
      sym__brace_start,
    STATE(1368), 1,
      aux_sym__literal_repeat1,
    STATE(6750), 1,
      sym_subscript,
    ACTIONS(1753), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1757), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1779), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1751), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(551), 3,
      sym_variable_assignment,
      sym_concatenation,
      aux_sym_declaration_command_repeat1,
    STATE(1067), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(1755), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [2358] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1791), 1,
      anon_sym_DQUOTE,
    ACTIONS(1795), 1,
      sym_variable_name,
    STATE(998), 1,
      sym_string,
    ACTIONS(1793), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1241), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
    ACTIONS(1789), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1239), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [2435] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1791), 1,
      anon_sym_DQUOTE,
    ACTIONS(1795), 1,
      sym_variable_name,
    STATE(998), 1,
      sym_string,
    ACTIONS(1793), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1235), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
    ACTIONS(1789), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1227), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [2512] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1759), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1761), 1,
      anon_sym_DOLLAR,
    ACTIONS(1763), 1,
      sym__special_character,
    ACTIONS(1765), 1,
      anon_sym_DQUOTE,
    ACTIONS(1767), 1,
      aux_sym_number_token1,
    ACTIONS(1769), 1,
      aux_sym_number_token2,
    ACTIONS(1771), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1773), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1775), 1,
      anon_sym_BQUOTE,
    ACTIONS(1777), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1783), 1,
      sym_variable_name,
    ACTIONS(1785), 1,
      sym_test_operator,
    ACTIONS(1787), 1,
      sym__brace_start,
    ACTIONS(1801), 1,
      aux_sym__simple_variable_name_token1,
    STATE(1368), 1,
      aux_sym__literal_repeat1,
    STATE(6750), 1,
      sym_subscript,
    ACTIONS(1753), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1779), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1799), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1751), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(546), 3,
      sym_variable_assignment,
      sym_concatenation,
      aux_sym_declaration_command_repeat1,
    STATE(1067), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(1797), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [2621] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1418), 1,
      anon_sym_DQUOTE,
    ACTIONS(1807), 1,
      sym_variable_name,
    STATE(922), 1,
      sym_string,
    ACTIONS(1805), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1241), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
    ACTIONS(1803), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1239), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [2698] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1418), 1,
      anon_sym_DQUOTE,
    ACTIONS(1807), 1,
      sym_variable_name,
    STATE(922), 1,
      sym_string,
    ACTIONS(1805), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1235), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
    ACTIONS(1803), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1227), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [2775] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(19), 1,
      anon_sym_LPAREN,
    ACTIONS(1549), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1551), 1,
      anon_sym_DOLLAR,
    ACTIONS(1555), 1,
      anon_sym_DQUOTE,
    ACTIONS(1557), 1,
      aux_sym_number_token1,
    ACTIONS(1559), 1,
      aux_sym_number_token2,
    ACTIONS(1561), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1563), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1567), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1573), 1,
      sym__bare_dollar,
    ACTIONS(1575), 1,
      sym__brace_start,
    ACTIONS(1690), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(1692), 1,
      sym__special_character,
    ACTIONS(1694), 1,
      sym_test_operator,
    STATE(584), 1,
      aux_sym_command_repeat2,
    STATE(1321), 1,
      aux_sym__literal_repeat1,
    STATE(1712), 1,
      sym_herestring_redirect,
    STATE(1716), 1,
      sym_concatenation,
    STATE(4911), 1,
      sym_subshell,
    ACTIONS(1408), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1543), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1569), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1688), 2,
      anon_sym_EQ_EQ,
      anon_sym_EQ_TILDE,
    ACTIONS(1686), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(1219), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(1404), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_BQUOTE,
  [2890] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1811), 1,
      anon_sym_DQUOTE,
    ACTIONS(1815), 1,
      sym_variable_name,
    STATE(1061), 1,
      sym_string,
    ACTIONS(1813), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1235), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
    ACTIONS(1809), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1227), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [2966] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1819), 1,
      anon_sym_DQUOTE,
    ACTIONS(1823), 1,
      sym_variable_name,
    STATE(1125), 1,
      sym_string,
    ACTIONS(1821), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1241), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      ts_builtin_sym_end,
    ACTIONS(1817), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1239), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [3042] = 26,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1583), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(1585), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1587), 1,
      anon_sym_DOLLAR,
    ACTIONS(1589), 1,
      sym__special_character,
    ACTIONS(1591), 1,
      anon_sym_DQUOTE,
    ACTIONS(1593), 1,
      aux_sym_number_token1,
    ACTIONS(1595), 1,
      aux_sym_number_token2,
    ACTIONS(1597), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1599), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1601), 1,
      anon_sym_BQUOTE,
    ACTIONS(1603), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1607), 1,
      sym_test_operator,
    ACTIONS(1609), 1,
      sym__bare_dollar,
    ACTIONS(1611), 1,
      sym__brace_start,
    STATE(569), 1,
      aux_sym_command_repeat2,
    STATE(1431), 1,
      aux_sym__literal_repeat1,
    STATE(1521), 1,
      sym_concatenation,
    STATE(1552), 1,
      sym_herestring_redirect,
    ACTIONS(1579), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1581), 2,
      anon_sym_EQ_EQ,
      anon_sym_EQ_TILDE,
    ACTIONS(1605), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1577), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    ACTIONS(1627), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    STATE(1093), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(1625), 19,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [3154] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1819), 1,
      anon_sym_DQUOTE,
    ACTIONS(1823), 1,
      sym_variable_name,
    STATE(1125), 1,
      sym_string,
    ACTIONS(1821), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1235), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      ts_builtin_sym_end,
    ACTIONS(1817), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1227), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [3230] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1829), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1831), 1,
      anon_sym_DOLLAR,
    ACTIONS(1833), 1,
      sym__special_character,
    ACTIONS(1835), 1,
      anon_sym_DQUOTE,
    ACTIONS(1837), 1,
      aux_sym_number_token1,
    ACTIONS(1839), 1,
      aux_sym_number_token2,
    ACTIONS(1841), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1843), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1845), 1,
      anon_sym_BQUOTE,
    ACTIONS(1847), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1851), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(1853), 1,
      sym_variable_name,
    ACTIONS(1855), 1,
      sym_test_operator,
    ACTIONS(1857), 1,
      sym__brace_start,
    STATE(1699), 1,
      aux_sym__literal_repeat1,
    STATE(6769), 1,
      sym_subscript,
    ACTIONS(1799), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1827), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1849), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1825), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(565), 3,
      sym_variable_assignment,
      sym_concatenation,
      aux_sym_declaration_command_repeat1,
    STATE(1171), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(1797), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [3338] = 26,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1583), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(1585), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1587), 1,
      anon_sym_DOLLAR,
    ACTIONS(1589), 1,
      sym__special_character,
    ACTIONS(1591), 1,
      anon_sym_DQUOTE,
    ACTIONS(1593), 1,
      aux_sym_number_token1,
    ACTIONS(1595), 1,
      aux_sym_number_token2,
    ACTIONS(1597), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1599), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1601), 1,
      anon_sym_BQUOTE,
    ACTIONS(1603), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1607), 1,
      sym_test_operator,
    ACTIONS(1609), 1,
      sym__bare_dollar,
    ACTIONS(1611), 1,
      sym__brace_start,
    STATE(569), 1,
      aux_sym_command_repeat2,
    STATE(1431), 1,
      aux_sym__literal_repeat1,
    STATE(1521), 1,
      sym_concatenation,
    STATE(1552), 1,
      sym_herestring_redirect,
    ACTIONS(1579), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1581), 2,
      anon_sym_EQ_EQ,
      anon_sym_EQ_TILDE,
    ACTIONS(1605), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1577), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    ACTIONS(1615), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    STATE(1093), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(1613), 19,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [3450] = 26,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1547), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(1549), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1551), 1,
      anon_sym_DOLLAR,
    ACTIONS(1553), 1,
      sym__special_character,
    ACTIONS(1555), 1,
      anon_sym_DQUOTE,
    ACTIONS(1557), 1,
      aux_sym_number_token1,
    ACTIONS(1559), 1,
      aux_sym_number_token2,
    ACTIONS(1561), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1563), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1565), 1,
      anon_sym_BQUOTE,
    ACTIONS(1567), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1571), 1,
      sym_test_operator,
    ACTIONS(1573), 1,
      sym__bare_dollar,
    ACTIONS(1575), 1,
      sym__brace_start,
    STATE(562), 1,
      aux_sym_command_repeat2,
    STATE(1321), 1,
      aux_sym__literal_repeat1,
    STATE(1712), 1,
      sym_herestring_redirect,
    STATE(1716), 1,
      sym_concatenation,
    ACTIONS(1543), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1545), 2,
      anon_sym_EQ_EQ,
      anon_sym_EQ_TILDE,
    ACTIONS(1569), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1615), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1541), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(1023), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(1613), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [3562] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1491), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1868), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(1871), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1874), 1,
      anon_sym_DOLLAR,
    ACTIONS(1877), 1,
      sym__special_character,
    ACTIONS(1880), 1,
      anon_sym_DQUOTE,
    ACTIONS(1883), 1,
      aux_sym_number_token1,
    ACTIONS(1886), 1,
      aux_sym_number_token2,
    ACTIONS(1889), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1892), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1895), 1,
      anon_sym_BQUOTE,
    ACTIONS(1898), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1904), 1,
      sym_file_descriptor,
    ACTIONS(1907), 1,
      sym_test_operator,
    ACTIONS(1910), 1,
      sym__bare_dollar,
    ACTIONS(1913), 1,
      sym__brace_start,
    STATE(562), 1,
      aux_sym_command_repeat2,
    STATE(1321), 1,
      aux_sym__literal_repeat1,
    STATE(1712), 1,
      sym_herestring_redirect,
    STATE(1716), 1,
      sym_concatenation,
    ACTIONS(1862), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1865), 2,
      anon_sym_EQ_EQ,
      anon_sym_EQ_TILDE,
    ACTIONS(1901), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1859), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(1023), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(1486), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [3676] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1829), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1831), 1,
      anon_sym_DOLLAR,
    ACTIONS(1833), 1,
      sym__special_character,
    ACTIONS(1835), 1,
      anon_sym_DQUOTE,
    ACTIONS(1837), 1,
      aux_sym_number_token1,
    ACTIONS(1839), 1,
      aux_sym_number_token2,
    ACTIONS(1841), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1843), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1845), 1,
      anon_sym_BQUOTE,
    ACTIONS(1847), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1853), 1,
      sym_variable_name,
    ACTIONS(1855), 1,
      sym_test_operator,
    ACTIONS(1857), 1,
      sym__brace_start,
    ACTIONS(1916), 1,
      aux_sym__simple_variable_name_token1,
    STATE(1699), 1,
      aux_sym__literal_repeat1,
    STATE(6769), 1,
      sym_subscript,
    ACTIONS(1757), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1827), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1849), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1825), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(559), 3,
      sym_variable_assignment,
      sym_concatenation,
      aux_sym_declaration_command_repeat1,
    STATE(1171), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(1755), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [3784] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1811), 1,
      anon_sym_DQUOTE,
    ACTIONS(1815), 1,
      sym_variable_name,
    STATE(1061), 1,
      sym_string,
    ACTIONS(1813), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1241), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
    ACTIONS(1809), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1239), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [3860] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1924), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1927), 1,
      anon_sym_DOLLAR,
    ACTIONS(1930), 1,
      sym__special_character,
    ACTIONS(1933), 1,
      anon_sym_DQUOTE,
    ACTIONS(1936), 1,
      aux_sym_number_token1,
    ACTIONS(1939), 1,
      aux_sym_number_token2,
    ACTIONS(1942), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1945), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1948), 1,
      anon_sym_BQUOTE,
    ACTIONS(1951), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1957), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(1960), 1,
      sym_variable_name,
    ACTIONS(1963), 1,
      sym_test_operator,
    ACTIONS(1966), 1,
      sym__brace_start,
    STATE(1699), 1,
      aux_sym__literal_repeat1,
    STATE(6769), 1,
      sym_subscript,
    ACTIONS(1704), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1921), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1954), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1918), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(565), 3,
      sym_variable_assignment,
      sym_concatenation,
      aux_sym_declaration_command_repeat1,
    STATE(1171), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(1702), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [3968] = 26,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1547), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(1549), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1551), 1,
      anon_sym_DOLLAR,
    ACTIONS(1553), 1,
      sym__special_character,
    ACTIONS(1555), 1,
      anon_sym_DQUOTE,
    ACTIONS(1557), 1,
      aux_sym_number_token1,
    ACTIONS(1559), 1,
      aux_sym_number_token2,
    ACTIONS(1561), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1563), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1565), 1,
      anon_sym_BQUOTE,
    ACTIONS(1567), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1571), 1,
      sym_test_operator,
    ACTIONS(1573), 1,
      sym__bare_dollar,
    ACTIONS(1575), 1,
      sym__brace_start,
    STATE(562), 1,
      aux_sym_command_repeat2,
    STATE(1321), 1,
      aux_sym__literal_repeat1,
    STATE(1712), 1,
      sym_herestring_redirect,
    STATE(1716), 1,
      sym_concatenation,
    ACTIONS(1543), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1545), 2,
      anon_sym_EQ_EQ,
      anon_sym_EQ_TILDE,
    ACTIONS(1569), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1627), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1541), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(1023), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(1625), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [4080] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1458), 1,
      anon_sym_DQUOTE,
    ACTIONS(1973), 1,
      sym_variable_name,
    STATE(1130), 1,
      sym_string,
    ACTIONS(1971), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1235), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
    ACTIONS(1969), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1227), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [4156] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1458), 1,
      anon_sym_DQUOTE,
    ACTIONS(1973), 1,
      sym_variable_name,
    STATE(1130), 1,
      sym_string,
    ACTIONS(1971), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1241), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
    ACTIONS(1969), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1239), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [4232] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1984), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(1987), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1990), 1,
      anon_sym_DOLLAR,
    ACTIONS(1993), 1,
      sym__special_character,
    ACTIONS(1996), 1,
      anon_sym_DQUOTE,
    ACTIONS(1999), 1,
      aux_sym_number_token1,
    ACTIONS(2002), 1,
      aux_sym_number_token2,
    ACTIONS(2005), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(2008), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2011), 1,
      anon_sym_BQUOTE,
    ACTIONS(2014), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(2020), 1,
      sym_file_descriptor,
    ACTIONS(2023), 1,
      sym_test_operator,
    ACTIONS(2026), 1,
      sym__bare_dollar,
    ACTIONS(2029), 1,
      sym__brace_start,
    STATE(569), 1,
      aux_sym_command_repeat2,
    STATE(1431), 1,
      aux_sym__literal_repeat1,
    STATE(1521), 1,
      sym_concatenation,
    STATE(1552), 1,
      sym_herestring_redirect,
    ACTIONS(1491), 2,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1978), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1981), 2,
      anon_sym_EQ_EQ,
      anon_sym_EQ_TILDE,
    ACTIONS(2017), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1975), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(1093), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(1486), 19,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [4346] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2036), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(2038), 1,
      anon_sym_DOLLAR,
    ACTIONS(2040), 1,
      sym__special_character,
    ACTIONS(2042), 1,
      anon_sym_DQUOTE,
    ACTIONS(2044), 1,
      aux_sym_number_token1,
    ACTIONS(2046), 1,
      aux_sym_number_token2,
    ACTIONS(2048), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(2050), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2052), 1,
      anon_sym_BQUOTE,
    ACTIONS(2054), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(2058), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(2060), 1,
      sym_variable_name,
    ACTIONS(2062), 1,
      sym_test_operator,
    ACTIONS(2064), 1,
      sym__brace_start,
    STATE(1816), 1,
      aux_sym__literal_repeat1,
    STATE(6731), 1,
      sym_subscript,
    ACTIONS(2034), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(2056), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1799), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(2032), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(577), 3,
      sym_variable_assignment,
      sym_concatenation,
      aux_sym_declaration_command_repeat1,
    STATE(1322), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(1797), 19,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [4453] = 26,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1549), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1551), 1,
      anon_sym_DOLLAR,
    ACTIONS(1555), 1,
      anon_sym_DQUOTE,
    ACTIONS(1557), 1,
      aux_sym_number_token1,
    ACTIONS(1559), 1,
      aux_sym_number_token2,
    ACTIONS(1561), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1563), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1565), 1,
      anon_sym_BQUOTE,
    ACTIONS(1567), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1573), 1,
      sym__bare_dollar,
    ACTIONS(1575), 1,
      sym__brace_start,
    ACTIONS(1690), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(1692), 1,
      sym__special_character,
    ACTIONS(1694), 1,
      sym_test_operator,
    STATE(574), 1,
      aux_sym_command_repeat2,
    STATE(1321), 1,
      aux_sym__literal_repeat1,
    STATE(1712), 1,
      sym_herestring_redirect,
    STATE(1716), 1,
      sym_concatenation,
    ACTIONS(1543), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1569), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1615), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1688), 2,
      anon_sym_EQ_EQ,
      anon_sym_EQ_TILDE,
    ACTIONS(1686), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(1219), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(1613), 19,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [4564] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1555), 1,
      anon_sym_DQUOTE,
    ACTIONS(2070), 1,
      sym_variable_name,
    STATE(1176), 1,
      sym_string,
    ACTIONS(2068), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1235), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
    ACTIONS(2066), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1227), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [4639] = 6,
    ACTIONS(3), 1,
      sym_comment,
    STATE(1474), 1,
      aux_sym__literal_repeat1,
    STATE(583), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    ACTIONS(2074), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    STATE(1158), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2072), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [4710] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1491), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1871), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1874), 1,
      anon_sym_DOLLAR,
    ACTIONS(1880), 1,
      anon_sym_DQUOTE,
    ACTIONS(1883), 1,
      aux_sym_number_token1,
    ACTIONS(1886), 1,
      aux_sym_number_token2,
    ACTIONS(1889), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1892), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1895), 1,
      anon_sym_BQUOTE,
    ACTIONS(1898), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1910), 1,
      sym__bare_dollar,
    ACTIONS(1913), 1,
      sym__brace_start,
    ACTIONS(2082), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(2085), 1,
      sym__special_character,
    ACTIONS(2088), 1,
      sym_file_descriptor,
    ACTIONS(2091), 1,
      sym_test_operator,
    STATE(574), 1,
      aux_sym_command_repeat2,
    STATE(1321), 1,
      aux_sym__literal_repeat1,
    STATE(1712), 1,
      sym_herestring_redirect,
    STATE(1716), 1,
      sym_concatenation,
    ACTIONS(1862), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1901), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(2079), 2,
      anon_sym_EQ_EQ,
      anon_sym_EQ_TILDE,
    ACTIONS(2076), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(1219), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(1486), 19,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [4823] = 6,
    ACTIONS(3), 1,
      sym_comment,
    STATE(1474), 1,
      aux_sym__literal_repeat1,
    STATE(583), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    ACTIONS(2096), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    STATE(1158), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2094), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [4894] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2036), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(2038), 1,
      anon_sym_DOLLAR,
    ACTIONS(2040), 1,
      sym__special_character,
    ACTIONS(2042), 1,
      anon_sym_DQUOTE,
    ACTIONS(2044), 1,
      aux_sym_number_token1,
    ACTIONS(2046), 1,
      aux_sym_number_token2,
    ACTIONS(2048), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(2050), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2052), 1,
      anon_sym_BQUOTE,
    ACTIONS(2054), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(2060), 1,
      sym_variable_name,
    ACTIONS(2062), 1,
      sym_test_operator,
    ACTIONS(2064), 1,
      sym__brace_start,
    ACTIONS(2098), 1,
      aux_sym__simple_variable_name_token1,
    STATE(1816), 1,
      aux_sym__literal_repeat1,
    STATE(6731), 1,
      sym_subscript,
    ACTIONS(2034), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(2056), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1757), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(2032), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(570), 3,
      sym_variable_assignment,
      sym_concatenation,
      aux_sym_declaration_command_repeat1,
    STATE(1322), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(1755), 19,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [5001] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2106), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(2109), 1,
      anon_sym_DOLLAR,
    ACTIONS(2112), 1,
      sym__special_character,
    ACTIONS(2115), 1,
      anon_sym_DQUOTE,
    ACTIONS(2118), 1,
      aux_sym_number_token1,
    ACTIONS(2121), 1,
      aux_sym_number_token2,
    ACTIONS(2124), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(2127), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2130), 1,
      anon_sym_BQUOTE,
    ACTIONS(2133), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(2139), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(2142), 1,
      sym_variable_name,
    ACTIONS(2145), 1,
      sym_test_operator,
    ACTIONS(2148), 1,
      sym__brace_start,
    STATE(1816), 1,
      aux_sym__literal_repeat1,
    STATE(6731), 1,
      sym_subscript,
    ACTIONS(2103), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(2136), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1704), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(2100), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(577), 3,
      sym_variable_assignment,
      sym_concatenation,
      aux_sym_declaration_command_repeat1,
    STATE(1322), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(1702), 19,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [5108] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1591), 1,
      anon_sym_DQUOTE,
    ACTIONS(2155), 1,
      sym_variable_name,
    STATE(1282), 1,
      sym_string,
    ACTIONS(2153), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1241), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      ts_builtin_sym_end,
    ACTIONS(2151), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1239), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [5183] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1811), 1,
      anon_sym_DQUOTE,
    ACTIONS(1815), 1,
      sym_variable_name,
    STATE(1061), 1,
      sym_string,
    ACTIONS(1813), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1241), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
    ACTIONS(1809), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1239), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [5258] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1811), 1,
      anon_sym_DQUOTE,
    ACTIONS(1815), 1,
      sym_variable_name,
    STATE(1061), 1,
      sym_string,
    ACTIONS(1813), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1235), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
    ACTIONS(1809), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1227), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [5333] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2163), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(2166), 1,
      anon_sym_DOLLAR,
    ACTIONS(2169), 1,
      sym__special_character,
    ACTIONS(2172), 1,
      anon_sym_DQUOTE,
    ACTIONS(2175), 1,
      aux_sym_number_token1,
    ACTIONS(2178), 1,
      aux_sym_number_token2,
    ACTIONS(2181), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(2184), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2187), 1,
      anon_sym_BQUOTE,
    ACTIONS(2190), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(2196), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(2199), 1,
      sym_variable_name,
    ACTIONS(2202), 1,
      sym_test_operator,
    ACTIONS(2205), 1,
      sym__brace_start,
    STATE(1743), 1,
      aux_sym__literal_repeat1,
    STATE(6781), 1,
      sym_subscript,
    ACTIONS(1704), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(2160), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(2193), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(2157), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(581), 3,
      sym_variable_assignment,
      sym_concatenation,
      aux_sym_declaration_command_repeat1,
    STATE(1360), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(1702), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [5440] = 26,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1549), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1551), 1,
      anon_sym_DOLLAR,
    ACTIONS(1555), 1,
      anon_sym_DQUOTE,
    ACTIONS(1557), 1,
      aux_sym_number_token1,
    ACTIONS(1559), 1,
      aux_sym_number_token2,
    ACTIONS(1561), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1563), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1565), 1,
      anon_sym_BQUOTE,
    ACTIONS(1567), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1573), 1,
      sym__bare_dollar,
    ACTIONS(1575), 1,
      sym__brace_start,
    ACTIONS(1690), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(1692), 1,
      sym__special_character,
    ACTIONS(1694), 1,
      sym_test_operator,
    STATE(574), 1,
      aux_sym_command_repeat2,
    STATE(1321), 1,
      aux_sym__literal_repeat1,
    STATE(1712), 1,
      sym_herestring_redirect,
    STATE(1716), 1,
      sym_concatenation,
    ACTIONS(1543), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1569), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1627), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1688), 2,
      anon_sym_EQ_EQ,
      anon_sym_EQ_TILDE,
    ACTIONS(1686), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(1219), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(1625), 19,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [5551] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2218), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(2221), 1,
      anon_sym_DOLLAR,
    ACTIONS(2224), 1,
      sym__special_character,
    ACTIONS(2227), 1,
      anon_sym_DQUOTE,
    ACTIONS(2230), 1,
      aux_sym_number_token1,
    ACTIONS(2233), 1,
      aux_sym_number_token2,
    ACTIONS(2236), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(2239), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2242), 1,
      anon_sym_BQUOTE,
    ACTIONS(2245), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(2251), 1,
      sym_test_operator,
    ACTIONS(2254), 1,
      sym__brace_start,
    STATE(1474), 1,
      aux_sym__literal_repeat1,
    ACTIONS(2211), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(2248), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(583), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    ACTIONS(2208), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    ACTIONS(2216), 3,
      sym_file_descriptor,
      sym_variable_name,
      aux_sym_heredoc_redirect_token1,
    STATE(1158), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2214), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [5652] = 25,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1549), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1551), 1,
      anon_sym_DOLLAR,
    ACTIONS(1555), 1,
      anon_sym_DQUOTE,
    ACTIONS(1557), 1,
      aux_sym_number_token1,
    ACTIONS(1559), 1,
      aux_sym_number_token2,
    ACTIONS(1561), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1563), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1567), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1573), 1,
      sym__bare_dollar,
    ACTIONS(1575), 1,
      sym__brace_start,
    ACTIONS(1690), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(1692), 1,
      sym__special_character,
    ACTIONS(1694), 1,
      sym_test_operator,
    STATE(574), 1,
      aux_sym_command_repeat2,
    STATE(1321), 1,
      aux_sym__literal_repeat1,
    STATE(1712), 1,
      sym_herestring_redirect,
    STATE(1716), 1,
      sym_concatenation,
    ACTIONS(1543), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1569), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1627), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1688), 2,
      anon_sym_EQ_EQ,
      anon_sym_EQ_TILDE,
    ACTIONS(1686), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(1219), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(1625), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_BQUOTE,
  [5761] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2261), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(2263), 1,
      anon_sym_DOLLAR,
    ACTIONS(2265), 1,
      sym__special_character,
    ACTIONS(2267), 1,
      anon_sym_DQUOTE,
    ACTIONS(2269), 1,
      aux_sym_number_token1,
    ACTIONS(2271), 1,
      aux_sym_number_token2,
    ACTIONS(2273), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(2275), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2277), 1,
      anon_sym_BQUOTE,
    ACTIONS(2279), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(2283), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(2285), 1,
      sym_variable_name,
    ACTIONS(2287), 1,
      sym_test_operator,
    ACTIONS(2289), 1,
      sym__brace_start,
    STATE(1743), 1,
      aux_sym__literal_repeat1,
    STATE(6781), 1,
      sym_subscript,
    ACTIONS(1799), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(2259), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(2281), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(2257), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(581), 3,
      sym_variable_assignment,
      sym_concatenation,
      aux_sym_declaration_command_repeat1,
    STATE(1360), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(1797), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [5868] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2261), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(2263), 1,
      anon_sym_DOLLAR,
    ACTIONS(2265), 1,
      sym__special_character,
    ACTIONS(2267), 1,
      anon_sym_DQUOTE,
    ACTIONS(2269), 1,
      aux_sym_number_token1,
    ACTIONS(2271), 1,
      aux_sym_number_token2,
    ACTIONS(2273), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(2275), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2277), 1,
      anon_sym_BQUOTE,
    ACTIONS(2279), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(2285), 1,
      sym_variable_name,
    ACTIONS(2287), 1,
      sym_test_operator,
    ACTIONS(2289), 1,
      sym__brace_start,
    ACTIONS(2291), 1,
      aux_sym__simple_variable_name_token1,
    STATE(1743), 1,
      aux_sym__literal_repeat1,
    STATE(6781), 1,
      sym_subscript,
    ACTIONS(1757), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(2259), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(2281), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(2257), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(585), 3,
      sym_variable_assignment,
      sym_concatenation,
      aux_sym_declaration_command_repeat1,
    STATE(1360), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(1755), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [5975] = 25,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1549), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1551), 1,
      anon_sym_DOLLAR,
    ACTIONS(1555), 1,
      anon_sym_DQUOTE,
    ACTIONS(1557), 1,
      aux_sym_number_token1,
    ACTIONS(1559), 1,
      aux_sym_number_token2,
    ACTIONS(1561), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1563), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1567), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1573), 1,
      sym__bare_dollar,
    ACTIONS(1575), 1,
      sym__brace_start,
    ACTIONS(1690), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(1692), 1,
      sym__special_character,
    ACTIONS(1694), 1,
      sym_test_operator,
    STATE(574), 1,
      aux_sym_command_repeat2,
    STATE(1321), 1,
      aux_sym__literal_repeat1,
    STATE(1712), 1,
      sym_herestring_redirect,
    STATE(1716), 1,
      sym_concatenation,
    ACTIONS(1543), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1569), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1615), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1688), 2,
      anon_sym_EQ_EQ,
      anon_sym_EQ_TILDE,
    ACTIONS(1686), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(1219), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(1613), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_BQUOTE,
  [6084] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1591), 1,
      anon_sym_DQUOTE,
    ACTIONS(2155), 1,
      sym_variable_name,
    STATE(1282), 1,
      sym_string,
    ACTIONS(2153), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1235), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      ts_builtin_sym_end,
    ACTIONS(2151), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1227), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [6159] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1555), 1,
      anon_sym_DQUOTE,
    ACTIONS(2070), 1,
      sym_variable_name,
    STATE(1176), 1,
      sym_string,
    ACTIONS(2068), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1241), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
    ACTIONS(2066), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1239), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [6234] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2301), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(2303), 1,
      anon_sym_DOLLAR,
    ACTIONS(2305), 1,
      sym__special_character,
    ACTIONS(2307), 1,
      anon_sym_DQUOTE,
    ACTIONS(2309), 1,
      aux_sym_number_token1,
    ACTIONS(2311), 1,
      aux_sym_number_token2,
    ACTIONS(2313), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(2315), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2317), 1,
      anon_sym_BQUOTE,
    ACTIONS(2319), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(2323), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(2325), 1,
      sym_test_operator,
    ACTIONS(2327), 1,
      sym__brace_start,
    STATE(1631), 1,
      aux_sym__literal_repeat1,
    ACTIONS(2295), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(2299), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(2321), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(613), 2,
      sym_concatenation,
      aux_sym_unset_command_repeat1,
    ACTIONS(2293), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(1190), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2297), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [6336] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2333), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(2335), 1,
      anon_sym_DOLLAR,
    ACTIONS(2337), 1,
      sym__special_character,
    ACTIONS(2339), 1,
      anon_sym_DQUOTE,
    ACTIONS(2341), 1,
      aux_sym_number_token1,
    ACTIONS(2343), 1,
      aux_sym_number_token2,
    ACTIONS(2345), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(2347), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2349), 1,
      anon_sym_BQUOTE,
    ACTIONS(2351), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(2355), 1,
      sym_test_operator,
    ACTIONS(2357), 1,
      sym__brace_start,
    STATE(1658), 1,
      aux_sym__literal_repeat1,
    ACTIONS(2096), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(2331), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(2353), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(593), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    ACTIONS(2329), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(1290), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2094), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [6436] = 23,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2261), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(2263), 1,
      anon_sym_DOLLAR,
    ACTIONS(2267), 1,
      anon_sym_DQUOTE,
    ACTIONS(2269), 1,
      aux_sym_number_token1,
    ACTIONS(2271), 1,
      aux_sym_number_token2,
    ACTIONS(2273), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(2275), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2279), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(2289), 1,
      sym__brace_start,
    ACTIONS(2361), 1,
      sym__special_character,
    ACTIONS(2363), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(2365), 1,
      sym_variable_name,
    ACTIONS(2367), 1,
      sym_test_operator,
    STATE(1743), 1,
      aux_sym__literal_repeat1,
    STATE(6773), 1,
      sym_subscript,
    ACTIONS(1757), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(2259), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(2281), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(2359), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(594), 3,
      sym_variable_assignment,
      sym_concatenation,
      aux_sym_declaration_command_repeat1,
    STATE(1602), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(1755), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_BQUOTE,
  [6540] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2375), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(2378), 1,
      anon_sym_DOLLAR,
    ACTIONS(2381), 1,
      sym__special_character,
    ACTIONS(2384), 1,
      anon_sym_DQUOTE,
    ACTIONS(2387), 1,
      aux_sym_number_token1,
    ACTIONS(2390), 1,
      aux_sym_number_token2,
    ACTIONS(2393), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(2396), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2399), 1,
      anon_sym_BQUOTE,
    ACTIONS(2402), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(2408), 1,
      sym_test_operator,
    ACTIONS(2411), 1,
      sym__brace_start,
    STATE(1658), 1,
      aux_sym__literal_repeat1,
    ACTIONS(2216), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(2372), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(2405), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(593), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    ACTIONS(2369), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(1290), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2214), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [6640] = 23,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2261), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(2263), 1,
      anon_sym_DOLLAR,
    ACTIONS(2267), 1,
      anon_sym_DQUOTE,
    ACTIONS(2269), 1,
      aux_sym_number_token1,
    ACTIONS(2271), 1,
      aux_sym_number_token2,
    ACTIONS(2273), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(2275), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2279), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(2289), 1,
      sym__brace_start,
    ACTIONS(2361), 1,
      sym__special_character,
    ACTIONS(2365), 1,
      sym_variable_name,
    ACTIONS(2367), 1,
      sym_test_operator,
    ACTIONS(2414), 1,
      aux_sym__simple_variable_name_token1,
    STATE(1743), 1,
      aux_sym__literal_repeat1,
    STATE(6773), 1,
      sym_subscript,
    ACTIONS(1799), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(2259), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(2281), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(2359), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(612), 3,
      sym_variable_assignment,
      sym_concatenation,
      aux_sym_declaration_command_repeat1,
    STATE(1602), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(1797), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_BQUOTE,
  [6744] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2333), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(2335), 1,
      anon_sym_DOLLAR,
    ACTIONS(2337), 1,
      sym__special_character,
    ACTIONS(2339), 1,
      anon_sym_DQUOTE,
    ACTIONS(2341), 1,
      aux_sym_number_token1,
    ACTIONS(2343), 1,
      aux_sym_number_token2,
    ACTIONS(2345), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(2347), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2349), 1,
      anon_sym_BQUOTE,
    ACTIONS(2351), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(2355), 1,
      sym_test_operator,
    ACTIONS(2357), 1,
      sym__brace_start,
    STATE(1658), 1,
      aux_sym__literal_repeat1,
    ACTIONS(2074), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(2331), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(2353), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(593), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    ACTIONS(2329), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(1290), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2072), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [6844] = 28,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(19), 1,
      anon_sym_LPAREN,
    ACTIONS(1549), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1551), 1,
      anon_sym_DOLLAR,
    ACTIONS(1555), 1,
      anon_sym_DQUOTE,
    ACTIONS(1557), 1,
      aux_sym_number_token1,
    ACTIONS(1559), 1,
      aux_sym_number_token2,
    ACTIONS(1561), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1563), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1565), 1,
      anon_sym_BQUOTE,
    ACTIONS(1567), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1573), 1,
      sym__bare_dollar,
    ACTIONS(1575), 1,
      sym__brace_start,
    ACTIONS(2420), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(2422), 1,
      sym__special_character,
    ACTIONS(2424), 1,
      sym_test_operator,
    STATE(688), 1,
      aux_sym_command_repeat2,
    STATE(1321), 1,
      aux_sym__literal_repeat1,
    STATE(1712), 1,
      sym_herestring_redirect,
    STATE(1716), 1,
      sym_concatenation,
    STATE(4911), 1,
      sym_subshell,
    ACTIONS(1408), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1543), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1569), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(2418), 2,
      anon_sym_EQ_EQ,
      anon_sym_EQ_TILDE,
    ACTIONS(2416), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(1794), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(1404), 16,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [6958] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2428), 1,
      anon_sym_DQUOTE,
    ACTIONS(2432), 1,
      sym_variable_name,
    STATE(1204), 1,
      sym_string,
    ACTIONS(2430), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1235), 3,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(2426), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1227), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [7032] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2444), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(2447), 1,
      anon_sym_DOLLAR,
    ACTIONS(2450), 1,
      sym__special_character,
    ACTIONS(2453), 1,
      anon_sym_DQUOTE,
    ACTIONS(2456), 1,
      aux_sym_number_token1,
    ACTIONS(2459), 1,
      aux_sym_number_token2,
    ACTIONS(2462), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(2465), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2468), 1,
      anon_sym_BQUOTE,
    ACTIONS(2471), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(2477), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(2480), 1,
      sym_test_operator,
    ACTIONS(2483), 1,
      sym__brace_start,
    STATE(1631), 1,
      aux_sym__literal_repeat1,
    ACTIONS(2437), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(2442), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(2474), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(598), 2,
      sym_concatenation,
      aux_sym_unset_command_repeat1,
    ACTIONS(2434), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(1190), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2440), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [7134] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(809), 23,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_GT_GT,
      anon_sym_LBRACE,
      anon_sym_LBRACK_LBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(579), 33,
      anon_sym_for,
      anon_sym_select,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LPAREN,
      anon_sym_while,
      anon_sym_until,
      anon_sym_do,
      anon_sym_if,
      anon_sym_then,
      anon_sym_fi,
      anon_sym_elif,
      anon_sym_else,
      anon_sym_case,
      anon_sym_function,
      anon_sym_BANG,
      anon_sym_LBRACK,
      anon_sym_declare,
      anon_sym_typeset,
      anon_sym_export,
      anon_sym_readonly,
      anon_sym_local,
      anon_sym_unset,
      anon_sym_unsetenv,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      sym__special_character,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
  [7198] = 28,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(19), 1,
      anon_sym_LPAREN,
    ACTIONS(1549), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1551), 1,
      anon_sym_DOLLAR,
    ACTIONS(1555), 1,
      anon_sym_DQUOTE,
    ACTIONS(1557), 1,
      aux_sym_number_token1,
    ACTIONS(1559), 1,
      aux_sym_number_token2,
    ACTIONS(1561), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1563), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1565), 1,
      anon_sym_BQUOTE,
    ACTIONS(1567), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1573), 1,
      sym__bare_dollar,
    ACTIONS(1575), 1,
      sym__brace_start,
    ACTIONS(2420), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(2422), 1,
      sym__special_character,
    ACTIONS(2424), 1,
      sym_test_operator,
    STATE(675), 1,
      aux_sym_command_repeat2,
    STATE(1321), 1,
      aux_sym__literal_repeat1,
    STATE(1712), 1,
      sym_herestring_redirect,
    STATE(1716), 1,
      sym_concatenation,
    STATE(4952), 1,
      sym_subshell,
    ACTIONS(1442), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1543), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1569), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(2418), 2,
      anon_sym_EQ_EQ,
      anon_sym_EQ_TILDE,
    ACTIONS(2416), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(1794), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(1440), 16,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [7312] = 6,
    ACTIONS(3), 1,
      sym_comment,
    STATE(1567), 1,
      aux_sym__literal_repeat1,
    STATE(609), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    ACTIONS(2096), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    STATE(1269), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2094), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [7382] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2261), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(2263), 1,
      anon_sym_DOLLAR,
    ACTIONS(2267), 1,
      anon_sym_DQUOTE,
    ACTIONS(2269), 1,
      aux_sym_number_token1,
    ACTIONS(2271), 1,
      aux_sym_number_token2,
    ACTIONS(2273), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(2275), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2277), 1,
      anon_sym_BQUOTE,
    ACTIONS(2279), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(2289), 1,
      sym__brace_start,
    ACTIONS(2361), 1,
      sym__special_character,
    ACTIONS(2365), 1,
      sym_variable_name,
    ACTIONS(2367), 1,
      sym_test_operator,
    ACTIONS(2486), 1,
      aux_sym__simple_variable_name_token1,
    STATE(1743), 1,
      aux_sym__literal_repeat1,
    STATE(6773), 1,
      sym_subscript,
    ACTIONS(1757), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(2259), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(2281), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(2359), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(610), 3,
      sym_variable_assignment,
      sym_concatenation,
      aux_sym_declaration_command_repeat1,
    STATE(1602), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(1755), 19,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [7488] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2339), 1,
      anon_sym_DQUOTE,
    ACTIONS(2492), 1,
      sym_variable_name,
    STATE(1463), 1,
      sym_string,
    ACTIONS(2490), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1241), 3,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(2488), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1239), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [7562] = 6,
    ACTIONS(3), 1,
      sym_comment,
    STATE(1429), 1,
      aux_sym__literal_repeat1,
    STATE(1534), 1,
      sym_concatenation,
    ACTIONS(2496), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    STATE(1100), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2494), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [7632] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2339), 1,
      anon_sym_DQUOTE,
    ACTIONS(2492), 1,
      sym_variable_name,
    STATE(1463), 1,
      sym_string,
    ACTIONS(2490), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1235), 3,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(2488), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1227), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [7706] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1555), 1,
      anon_sym_DQUOTE,
    ACTIONS(2070), 1,
      sym_variable_name,
    STATE(1176), 1,
      sym_string,
    ACTIONS(2068), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1235), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
    ACTIONS(2066), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1227), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [7780] = 6,
    ACTIONS(3), 1,
      sym_comment,
    STATE(1382), 1,
      aux_sym__literal_repeat1,
    STATE(1607), 1,
      sym_concatenation,
    ACTIONS(2500), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    STATE(1081), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2498), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [7850] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1555), 1,
      anon_sym_DQUOTE,
    ACTIONS(2070), 1,
      sym_variable_name,
    STATE(1176), 1,
      sym_string,
    ACTIONS(2068), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1241), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
    ACTIONS(2066), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1239), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [7924] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2508), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(2511), 1,
      anon_sym_DOLLAR,
    ACTIONS(2514), 1,
      sym__special_character,
    ACTIONS(2517), 1,
      anon_sym_DQUOTE,
    ACTIONS(2520), 1,
      aux_sym_number_token1,
    ACTIONS(2523), 1,
      aux_sym_number_token2,
    ACTIONS(2526), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(2529), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2532), 1,
      anon_sym_BQUOTE,
    ACTIONS(2535), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(2541), 1,
      sym_test_operator,
    ACTIONS(2544), 1,
      sym__brace_start,
    STATE(1567), 1,
      aux_sym__literal_repeat1,
    ACTIONS(2505), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(2538), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(609), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    ACTIONS(2216), 3,
      sym_file_descriptor,
      sym_variable_name,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(2502), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(1269), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2214), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [8024] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2261), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(2263), 1,
      anon_sym_DOLLAR,
    ACTIONS(2267), 1,
      anon_sym_DQUOTE,
    ACTIONS(2269), 1,
      aux_sym_number_token1,
    ACTIONS(2271), 1,
      aux_sym_number_token2,
    ACTIONS(2273), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(2275), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2277), 1,
      anon_sym_BQUOTE,
    ACTIONS(2279), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(2289), 1,
      sym__brace_start,
    ACTIONS(2361), 1,
      sym__special_character,
    ACTIONS(2365), 1,
      sym_variable_name,
    ACTIONS(2367), 1,
      sym_test_operator,
    ACTIONS(2414), 1,
      aux_sym__simple_variable_name_token1,
    STATE(1743), 1,
      aux_sym__literal_repeat1,
    STATE(6773), 1,
      sym_subscript,
    ACTIONS(1799), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(2259), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(2281), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(2359), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(612), 3,
      sym_variable_assignment,
      sym_concatenation,
      aux_sym_declaration_command_repeat1,
    STATE(1602), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(1797), 19,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [8130] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2428), 1,
      anon_sym_DQUOTE,
    ACTIONS(2432), 1,
      sym_variable_name,
    STATE(1204), 1,
      sym_string,
    ACTIONS(2430), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1241), 3,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(2426), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1239), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [8204] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2163), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(2166), 1,
      anon_sym_DOLLAR,
    ACTIONS(2172), 1,
      anon_sym_DQUOTE,
    ACTIONS(2175), 1,
      aux_sym_number_token1,
    ACTIONS(2178), 1,
      aux_sym_number_token2,
    ACTIONS(2181), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(2184), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2187), 1,
      anon_sym_BQUOTE,
    ACTIONS(2190), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(2205), 1,
      sym__brace_start,
    ACTIONS(2550), 1,
      sym__special_character,
    ACTIONS(2553), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(2556), 1,
      sym_variable_name,
    ACTIONS(2559), 1,
      sym_test_operator,
    STATE(1743), 1,
      aux_sym__literal_repeat1,
    STATE(6773), 1,
      sym_subscript,
    ACTIONS(1704), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(2160), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(2193), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(2547), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(612), 3,
      sym_variable_assignment,
      sym_concatenation,
      aux_sym_declaration_command_repeat1,
    STATE(1602), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(1702), 19,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [8310] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2301), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(2303), 1,
      anon_sym_DOLLAR,
    ACTIONS(2305), 1,
      sym__special_character,
    ACTIONS(2307), 1,
      anon_sym_DQUOTE,
    ACTIONS(2309), 1,
      aux_sym_number_token1,
    ACTIONS(2311), 1,
      aux_sym_number_token2,
    ACTIONS(2313), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(2315), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2317), 1,
      anon_sym_BQUOTE,
    ACTIONS(2319), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(2325), 1,
      sym_test_operator,
    ACTIONS(2327), 1,
      sym__brace_start,
    ACTIONS(2566), 1,
      aux_sym__simple_variable_name_token1,
    STATE(1631), 1,
      aux_sym__literal_repeat1,
    ACTIONS(2295), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(2321), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(2564), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    STATE(598), 2,
      sym_concatenation,
      aux_sym_unset_command_repeat1,
    ACTIONS(2293), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(1190), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2562), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [8412] = 6,
    ACTIONS(3), 1,
      sym_comment,
    STATE(1567), 1,
      aux_sym__literal_repeat1,
    STATE(609), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    ACTIONS(2074), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    STATE(1269), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2072), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [8482] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2572), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(2574), 1,
      anon_sym_DOLLAR,
    ACTIONS(2576), 1,
      sym__special_character,
    ACTIONS(2578), 1,
      anon_sym_DQUOTE,
    ACTIONS(2580), 1,
      aux_sym_number_token1,
    ACTIONS(2582), 1,
      aux_sym_number_token2,
    ACTIONS(2584), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(2586), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2588), 1,
      anon_sym_BQUOTE,
    ACTIONS(2590), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(2594), 1,
      sym_test_operator,
    ACTIONS(2596), 1,
      sym__brace_start,
    STATE(1899), 1,
      aux_sym__literal_repeat1,
    ACTIONS(2074), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(2570), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(2592), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(621), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    ACTIONS(2568), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(1394), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2072), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [8581] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2602), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(2604), 1,
      anon_sym_DOLLAR,
    ACTIONS(2606), 1,
      sym__special_character,
    ACTIONS(2608), 1,
      anon_sym_DQUOTE,
    ACTIONS(2610), 1,
      aux_sym_number_token1,
    ACTIONS(2612), 1,
      aux_sym_number_token2,
    ACTIONS(2614), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(2616), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2618), 1,
      anon_sym_BQUOTE,
    ACTIONS(2620), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(2624), 1,
      sym_test_operator,
    ACTIONS(2626), 1,
      sym__brace_start,
    STATE(1803), 1,
      aux_sym__literal_repeat1,
    ACTIONS(2074), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(2600), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(2622), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(630), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    ACTIONS(2598), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(1381), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2072), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [8680] = 29,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1033), 1,
      anon_sym_LPAREN,
    ACTIONS(2628), 1,
      sym_word,
    ACTIONS(2634), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(2636), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(2638), 1,
      anon_sym_DOLLAR,
    ACTIONS(2640), 1,
      sym__special_character,
    ACTIONS(2642), 1,
      anon_sym_DQUOTE,
    ACTIONS(2646), 1,
      aux_sym_number_token1,
    ACTIONS(2648), 1,
      aux_sym_number_token2,
    ACTIONS(2650), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(2652), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2654), 1,
      anon_sym_BQUOTE,
    ACTIONS(2656), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(2660), 1,
      sym_test_operator,
    ACTIONS(2662), 1,
      sym__bare_dollar,
    ACTIONS(2664), 1,
      sym__brace_start,
    STATE(752), 1,
      aux_sym_command_repeat2,
    STATE(2417), 1,
      aux_sym__literal_repeat1,
    STATE(2558), 1,
      sym_concatenation,
    STATE(2562), 1,
      sym_herestring_redirect,
    STATE(5656), 1,
      sym_subshell,
    ACTIONS(2630), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(2632), 2,
      anon_sym_EQ_EQ,
      anon_sym_EQ_TILDE,
    ACTIONS(2644), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(2658), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1440), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    STATE(2211), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(1442), 10,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [8795] = 37,
    ACTIONS(13), 1,
      anon_sym_LPAREN_LPAREN,
    ACTIONS(19), 1,
      anon_sym_LPAREN,
    ACTIONS(33), 1,
      anon_sym_LBRACK,
    ACTIONS(35), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(45), 1,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(47), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(49), 1,
      anon_sym_DOLLAR,
    ACTIONS(51), 1,
      sym__special_character,
    ACTIONS(53), 1,
      anon_sym_DQUOTE,
    ACTIONS(57), 1,
      aux_sym_number_token1,
    ACTIONS(59), 1,
      aux_sym_number_token2,
    ACTIONS(61), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(63), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(65), 1,
      anon_sym_BQUOTE,
    ACTIONS(67), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(75), 1,
      sym_variable_name,
    ACTIONS(77), 1,
      sym_test_operator,
    ACTIONS(79), 1,
      sym__brace_start,
    ACTIONS(1394), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(1396), 1,
      sym_file_descriptor,
    ACTIONS(2666), 1,
      sym_word,
    STATE(534), 1,
      sym_command_name,
    STATE(937), 1,
      aux_sym_command_repeat1,
    STATE(1212), 1,
      aux_sym__literal_repeat1,
    STATE(1386), 1,
      sym_concatenation,
    STATE(2122), 1,
      sym_variable_assignment,
    STATE(5026), 1,
      sym_command,
    STATE(6774), 1,
      sym_subscript,
    ACTIONS(55), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(69), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1392), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    STATE(3581), 2,
      sym_file_redirect,
      sym_herestring_redirect,
    STATE(5014), 2,
      sym_subshell,
      sym_test_command,
    ACTIONS(1390), 3,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
    ACTIONS(1388), 5,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    STATE(957), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [8926] = 37,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(169), 1,
      anon_sym_LBRACK,
    ACTIONS(173), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(217), 1,
      sym_variable_name,
    ACTIONS(1027), 1,
      anon_sym_LPAREN_LPAREN,
    ACTIONS(1033), 1,
      anon_sym_LPAREN,
    ACTIONS(1041), 1,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1043), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1045), 1,
      anon_sym_DOLLAR,
    ACTIONS(1047), 1,
      sym__special_character,
    ACTIONS(1049), 1,
      anon_sym_DQUOTE,
    ACTIONS(1053), 1,
      aux_sym_number_token1,
    ACTIONS(1055), 1,
      aux_sym_number_token2,
    ACTIONS(1057), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1059), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1061), 1,
      anon_sym_BQUOTE,
    ACTIONS(1063), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1069), 1,
      sym_test_operator,
    ACTIONS(1071), 1,
      sym__brace_start,
    ACTIONS(1394), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(1396), 1,
      sym_file_descriptor,
    ACTIONS(2668), 1,
      sym_word,
    STATE(620), 1,
      sym_command_name,
    STATE(955), 1,
      aux_sym_command_repeat1,
    STATE(2332), 1,
      aux_sym__literal_repeat1,
    STATE(2368), 1,
      sym_concatenation,
    STATE(2941), 1,
      sym_variable_assignment,
    STATE(5624), 1,
      sym_command,
    STATE(6756), 1,
      sym_subscript,
    ACTIONS(1051), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(1065), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1392), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    STATE(3581), 2,
      sym_file_redirect,
      sym_herestring_redirect,
    STATE(5623), 2,
      sym_subshell,
      sym_test_command,
    ACTIONS(1390), 3,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
    ACTIONS(1388), 5,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    STATE(1799), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [9057] = 29,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1033), 1,
      anon_sym_LPAREN,
    ACTIONS(2628), 1,
      sym_word,
    ACTIONS(2634), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(2636), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(2638), 1,
      anon_sym_DOLLAR,
    ACTIONS(2640), 1,
      sym__special_character,
    ACTIONS(2642), 1,
      anon_sym_DQUOTE,
    ACTIONS(2646), 1,
      aux_sym_number_token1,
    ACTIONS(2648), 1,
      aux_sym_number_token2,
    ACTIONS(2650), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(2652), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2654), 1,
      anon_sym_BQUOTE,
    ACTIONS(2656), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(2660), 1,
      sym_test_operator,
    ACTIONS(2662), 1,
      sym__bare_dollar,
    ACTIONS(2664), 1,
      sym__brace_start,
    STATE(727), 1,
      aux_sym_command_repeat2,
    STATE(2417), 1,
      aux_sym__literal_repeat1,
    STATE(2558), 1,
      sym_concatenation,
    STATE(2562), 1,
      sym_herestring_redirect,
    STATE(5634), 1,
      sym_subshell,
    ACTIONS(2630), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(2632), 2,
      anon_sym_EQ_EQ,
      anon_sym_EQ_TILDE,
    ACTIONS(2644), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(2658), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1404), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    STATE(2211), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(1408), 10,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [9172] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2676), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(2679), 1,
      anon_sym_DOLLAR,
    ACTIONS(2682), 1,
      sym__special_character,
    ACTIONS(2685), 1,
      anon_sym_DQUOTE,
    ACTIONS(2688), 1,
      aux_sym_number_token1,
    ACTIONS(2691), 1,
      aux_sym_number_token2,
    ACTIONS(2694), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(2697), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2700), 1,
      anon_sym_BQUOTE,
    ACTIONS(2703), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(2709), 1,
      sym_test_operator,
    ACTIONS(2712), 1,
      sym__brace_start,
    STATE(1899), 1,
      aux_sym__literal_repeat1,
    ACTIONS(2216), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(2673), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(2706), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(621), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    ACTIONS(2670), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(1394), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2214), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [9271] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(2719), 2,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
    ACTIONS(2721), 2,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
    ACTIONS(2717), 23,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_GT_GT,
      anon_sym_LBRACE,
      anon_sym_LBRACK_LBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(2715), 28,
      anon_sym_for,
      anon_sym_select,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LPAREN,
      anon_sym_while,
      anon_sym_until,
      anon_sym_if,
      anon_sym_case,
      anon_sym_function,
      anon_sym_BANG,
      anon_sym_LBRACK,
      anon_sym_declare,
      anon_sym_typeset,
      anon_sym_export,
      anon_sym_readonly,
      anon_sym_local,
      anon_sym_unset,
      anon_sym_unsetenv,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      sym__special_character,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
  [9338] = 37,
    ACTIONS(19), 1,
      anon_sym_LPAREN,
    ACTIONS(33), 1,
      anon_sym_LBRACK,
    ACTIONS(35), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(288), 1,
      anon_sym_LPAREN_LPAREN,
    ACTIONS(312), 1,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(314), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(316), 1,
      anon_sym_DOLLAR,
    ACTIONS(318), 1,
      sym__special_character,
    ACTIONS(320), 1,
      anon_sym_DQUOTE,
    ACTIONS(324), 1,
      aux_sym_number_token1,
    ACTIONS(326), 1,
      aux_sym_number_token2,
    ACTIONS(328), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(330), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(332), 1,
      anon_sym_BQUOTE,
    ACTIONS(334), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(340), 1,
      sym_variable_name,
    ACTIONS(342), 1,
      sym_test_operator,
    ACTIONS(344), 1,
      sym__brace_start,
    ACTIONS(1394), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(1396), 1,
      sym_file_descriptor,
    ACTIONS(2723), 1,
      sym_word,
    STATE(544), 1,
      sym_command_name,
    STATE(997), 1,
      aux_sym_command_repeat1,
    STATE(1267), 1,
      aux_sym__literal_repeat1,
    STATE(1376), 1,
      sym_concatenation,
    STATE(1974), 1,
      sym_variable_assignment,
    STATE(5026), 1,
      sym_command,
    STATE(6790), 1,
      sym_subscript,
    ACTIONS(322), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(336), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1392), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    STATE(3581), 2,
      sym_file_redirect,
      sym_herestring_redirect,
    STATE(5014), 2,
      sym_subshell,
      sym_test_command,
    ACTIONS(1390), 3,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
    ACTIONS(1388), 5,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    STATE(1078), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [9469] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(2719), 1,
      anon_sym_SEMI_SEMI,
    ACTIONS(2721), 3,
      anon_sym_RPAREN,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
    ACTIONS(2717), 23,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_GT_GT,
      anon_sym_LBRACE,
      anon_sym_LBRACK_LBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(2715), 28,
      anon_sym_for,
      anon_sym_select,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LPAREN,
      anon_sym_while,
      anon_sym_until,
      anon_sym_if,
      anon_sym_case,
      anon_sym_function,
      anon_sym_BANG,
      anon_sym_LBRACK,
      anon_sym_declare,
      anon_sym_typeset,
      anon_sym_export,
      anon_sym_readonly,
      anon_sym_local,
      anon_sym_unset,
      anon_sym_unsetenv,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      sym__special_character,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
  [9536] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(2727), 1,
      anon_sym_SEMI_SEMI,
    ACTIONS(2725), 3,
      anon_sym_RPAREN,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
    ACTIONS(2717), 23,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_GT_GT,
      anon_sym_LBRACE,
      anon_sym_LBRACK_LBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(2715), 28,
      anon_sym_for,
      anon_sym_select,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LPAREN,
      anon_sym_while,
      anon_sym_until,
      anon_sym_if,
      anon_sym_case,
      anon_sym_function,
      anon_sym_BANG,
      anon_sym_LBRACK,
      anon_sym_declare,
      anon_sym_typeset,
      anon_sym_export,
      anon_sym_readonly,
      anon_sym_local,
      anon_sym_unset,
      anon_sym_unsetenv,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      sym__special_character,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
  [9603] = 37,
    ACTIONS(19), 1,
      anon_sym_LPAREN,
    ACTIONS(33), 1,
      anon_sym_LBRACK,
    ACTIONS(35), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(288), 1,
      anon_sym_LPAREN_LPAREN,
    ACTIONS(312), 1,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(314), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(316), 1,
      anon_sym_DOLLAR,
    ACTIONS(318), 1,
      sym__special_character,
    ACTIONS(320), 1,
      anon_sym_DQUOTE,
    ACTIONS(324), 1,
      aux_sym_number_token1,
    ACTIONS(326), 1,
      aux_sym_number_token2,
    ACTIONS(328), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(330), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(332), 1,
      anon_sym_BQUOTE,
    ACTIONS(334), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(340), 1,
      sym_variable_name,
    ACTIONS(342), 1,
      sym_test_operator,
    ACTIONS(344), 1,
      sym__brace_start,
    ACTIONS(1394), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(1396), 1,
      sym_file_descriptor,
    ACTIONS(2723), 1,
      sym_word,
    STATE(554), 1,
      sym_command_name,
    STATE(958), 1,
      aux_sym_command_repeat1,
    STATE(1267), 1,
      aux_sym__literal_repeat1,
    STATE(1376), 1,
      sym_concatenation,
    STATE(2286), 1,
      sym_variable_assignment,
    STATE(5026), 1,
      sym_command,
    STATE(6790), 1,
      sym_subscript,
    ACTIONS(322), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(336), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1392), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    STATE(3581), 2,
      sym_file_redirect,
      sym_herestring_redirect,
    STATE(5014), 2,
      sym_subshell,
      sym_test_command,
    ACTIONS(1390), 3,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
    ACTIONS(1388), 5,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    STATE(1078), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [9734] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2735), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(2738), 1,
      anon_sym_DOLLAR,
    ACTIONS(2741), 1,
      sym__special_character,
    ACTIONS(2744), 1,
      anon_sym_DQUOTE,
    ACTIONS(2747), 1,
      aux_sym_number_token1,
    ACTIONS(2750), 1,
      aux_sym_number_token2,
    ACTIONS(2753), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(2756), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2759), 1,
      anon_sym_BQUOTE,
    ACTIONS(2762), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(2768), 1,
      sym_test_operator,
    ACTIONS(2771), 1,
      sym__brace_start,
    STATE(1744), 1,
      aux_sym__literal_repeat1,
    ACTIONS(2732), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(2765), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(627), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    ACTIONS(2216), 3,
      sym_file_descriptor,
      sym_variable_name,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(2729), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(1370), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2214), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [9833] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2608), 1,
      anon_sym_DQUOTE,
    ACTIONS(2778), 1,
      sym_variable_name,
    STATE(1525), 1,
      sym_string,
    ACTIONS(2776), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1241), 3,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(2774), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1239), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [9906] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2784), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(2786), 1,
      anon_sym_DOLLAR,
    ACTIONS(2788), 1,
      sym__special_character,
    ACTIONS(2790), 1,
      anon_sym_DQUOTE,
    ACTIONS(2792), 1,
      aux_sym_number_token1,
    ACTIONS(2794), 1,
      aux_sym_number_token2,
    ACTIONS(2796), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(2798), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2800), 1,
      anon_sym_BQUOTE,
    ACTIONS(2802), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(2806), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(2808), 1,
      sym_test_operator,
    ACTIONS(2810), 1,
      sym__brace_start,
    STATE(1860), 1,
      aux_sym__literal_repeat1,
    ACTIONS(2564), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(2782), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(2804), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(633), 2,
      sym_concatenation,
      aux_sym_unset_command_repeat1,
    ACTIONS(2780), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(1326), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2562), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [10007] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2818), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(2821), 1,
      anon_sym_DOLLAR,
    ACTIONS(2824), 1,
      sym__special_character,
    ACTIONS(2827), 1,
      anon_sym_DQUOTE,
    ACTIONS(2830), 1,
      aux_sym_number_token1,
    ACTIONS(2833), 1,
      aux_sym_number_token2,
    ACTIONS(2836), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(2839), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2842), 1,
      anon_sym_BQUOTE,
    ACTIONS(2845), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(2851), 1,
      sym_test_operator,
    ACTIONS(2854), 1,
      sym__brace_start,
    STATE(1803), 1,
      aux_sym__literal_repeat1,
    ACTIONS(2216), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(2815), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(2848), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(630), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    ACTIONS(2812), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(1381), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2214), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [10106] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2861), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(2863), 1,
      anon_sym_DOLLAR,
    ACTIONS(2865), 1,
      sym__special_character,
    ACTIONS(2867), 1,
      anon_sym_DQUOTE,
    ACTIONS(2869), 1,
      aux_sym_number_token1,
    ACTIONS(2871), 1,
      aux_sym_number_token2,
    ACTIONS(2873), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(2875), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2877), 1,
      anon_sym_BQUOTE,
    ACTIONS(2879), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(2883), 1,
      sym_test_operator,
    ACTIONS(2885), 1,
      sym__brace_start,
    STATE(4262), 1,
      aux_sym__literal_repeat1,
    STATE(4820), 1,
      sym_concatenation,
    ACTIONS(2496), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(2859), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(2881), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(2857), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(4240), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2494), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [10205] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2784), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(2786), 1,
      anon_sym_DOLLAR,
    ACTIONS(2788), 1,
      sym__special_character,
    ACTIONS(2790), 1,
      anon_sym_DQUOTE,
    ACTIONS(2792), 1,
      aux_sym_number_token1,
    ACTIONS(2794), 1,
      aux_sym_number_token2,
    ACTIONS(2796), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(2798), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2800), 1,
      anon_sym_BQUOTE,
    ACTIONS(2802), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(2808), 1,
      sym_test_operator,
    ACTIONS(2810), 1,
      sym__brace_start,
    ACTIONS(2887), 1,
      aux_sym__simple_variable_name_token1,
    STATE(1860), 1,
      aux_sym__literal_repeat1,
    ACTIONS(2299), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(2782), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(2804), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(629), 2,
      sym_concatenation,
      aux_sym_unset_command_repeat1,
    ACTIONS(2780), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(1326), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2297), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [10306] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2895), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(2898), 1,
      anon_sym_DOLLAR,
    ACTIONS(2901), 1,
      sym__special_character,
    ACTIONS(2904), 1,
      anon_sym_DQUOTE,
    ACTIONS(2907), 1,
      aux_sym_number_token1,
    ACTIONS(2910), 1,
      aux_sym_number_token2,
    ACTIONS(2913), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(2916), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2919), 1,
      anon_sym_BQUOTE,
    ACTIONS(2922), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(2928), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(2931), 1,
      sym_test_operator,
    ACTIONS(2934), 1,
      sym__brace_start,
    STATE(1860), 1,
      aux_sym__literal_repeat1,
    ACTIONS(2442), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(2892), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(2925), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(633), 2,
      sym_concatenation,
      aux_sym_unset_command_repeat1,
    ACTIONS(2889), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(1326), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2440), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [10407] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2608), 1,
      anon_sym_DQUOTE,
    ACTIONS(2778), 1,
      sym_variable_name,
    STATE(1525), 1,
      sym_string,
    ACTIONS(2776), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1235), 3,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(2774), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1227), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [10480] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2861), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(2863), 1,
      anon_sym_DOLLAR,
    ACTIONS(2865), 1,
      sym__special_character,
    ACTIONS(2867), 1,
      anon_sym_DQUOTE,
    ACTIONS(2869), 1,
      aux_sym_number_token1,
    ACTIONS(2871), 1,
      aux_sym_number_token2,
    ACTIONS(2873), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(2875), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2877), 1,
      anon_sym_BQUOTE,
    ACTIONS(2879), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(2885), 1,
      sym__brace_start,
    ACTIONS(2939), 1,
      sym_test_operator,
    STATE(4351), 1,
      aux_sym__literal_repeat1,
    STATE(4789), 1,
      sym_concatenation,
    ACTIONS(2500), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(2859), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(2881), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(2937), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(4247), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2498), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [10579] = 37,
    ACTIONS(19), 1,
      anon_sym_LPAREN,
    ACTIONS(33), 1,
      anon_sym_LBRACK,
    ACTIONS(35), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(137), 1,
      sym_variable_name,
    ACTIONS(288), 1,
      anon_sym_LPAREN_LPAREN,
    ACTIONS(312), 1,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(314), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(316), 1,
      anon_sym_DOLLAR,
    ACTIONS(320), 1,
      anon_sym_DQUOTE,
    ACTIONS(324), 1,
      aux_sym_number_token1,
    ACTIONS(326), 1,
      aux_sym_number_token2,
    ACTIONS(328), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(330), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(332), 1,
      anon_sym_BQUOTE,
    ACTIONS(334), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(344), 1,
      sym__brace_start,
    ACTIONS(771), 1,
      sym__special_character,
    ACTIONS(775), 1,
      sym_test_operator,
    ACTIONS(1394), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(1396), 1,
      sym_file_descriptor,
    ACTIONS(2941), 1,
      sym_word,
    STATE(533), 1,
      sym_command_name,
    STATE(951), 1,
      aux_sym_command_repeat1,
    STATE(1267), 1,
      aux_sym__literal_repeat1,
    STATE(1376), 1,
      sym_concatenation,
    STATE(1974), 1,
      sym_variable_assignment,
    STATE(5026), 1,
      sym_command,
    STATE(6754), 1,
      sym_subscript,
    ACTIONS(336), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(773), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(1392), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    STATE(3581), 2,
      sym_file_redirect,
      sym_herestring_redirect,
    STATE(5014), 2,
      sym_subshell,
      sym_test_command,
    ACTIONS(1390), 3,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
    ACTIONS(1388), 5,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    STATE(938), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [10710] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2307), 1,
      anon_sym_DQUOTE,
    ACTIONS(2947), 1,
      sym_variable_name,
    STATE(1432), 1,
      sym_string,
    ACTIONS(2945), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1235), 3,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(2943), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1227), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [10783] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2307), 1,
      anon_sym_DQUOTE,
    ACTIONS(2947), 1,
      sym_variable_name,
    STATE(1432), 1,
      sym_string,
    ACTIONS(2945), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1241), 3,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(2943), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1239), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [10856] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2578), 1,
      anon_sym_DQUOTE,
    ACTIONS(2953), 1,
      sym_variable_name,
    STATE(1585), 1,
      sym_string,
    ACTIONS(2951), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1235), 3,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(2949), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1227), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [10929] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2578), 1,
      anon_sym_DQUOTE,
    ACTIONS(2953), 1,
      sym_variable_name,
    STATE(1585), 1,
      sym_string,
    ACTIONS(2951), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1241), 3,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(2949), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1239), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [11002] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1765), 1,
      anon_sym_DQUOTE,
    ACTIONS(2959), 1,
      sym_variable_name,
    STATE(1213), 1,
      sym_string,
    ACTIONS(2957), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1235), 3,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(2955), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1227), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [11075] = 6,
    ACTIONS(3), 1,
      sym_comment,
    STATE(1801), 1,
      aux_sym__literal_repeat1,
    STATE(651), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    ACTIONS(2096), 6,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    STATE(1399), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2094), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [11144] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1765), 1,
      anon_sym_DQUOTE,
    ACTIONS(2959), 1,
      sym_variable_name,
    STATE(1213), 1,
      sym_string,
    ACTIONS(2957), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1241), 3,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(2955), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1239), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [11217] = 37,
    ACTIONS(19), 1,
      anon_sym_LPAREN,
    ACTIONS(33), 1,
      anon_sym_LBRACK,
    ACTIONS(35), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(288), 1,
      anon_sym_LPAREN_LPAREN,
    ACTIONS(312), 1,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(314), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(316), 1,
      anon_sym_DOLLAR,
    ACTIONS(320), 1,
      anon_sym_DQUOTE,
    ACTIONS(324), 1,
      aux_sym_number_token1,
    ACTIONS(326), 1,
      aux_sym_number_token2,
    ACTIONS(328), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(330), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(332), 1,
      anon_sym_BQUOTE,
    ACTIONS(334), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(344), 1,
      sym__brace_start,
    ACTIONS(1015), 1,
      sym__special_character,
    ACTIONS(1021), 1,
      sym_variable_name,
    ACTIONS(1023), 1,
      sym_test_operator,
    ACTIONS(1394), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(1396), 1,
      sym_file_descriptor,
    ACTIONS(2961), 1,
      sym_word,
    STATE(596), 1,
      sym_command_name,
    STATE(933), 1,
      aux_sym_command_repeat1,
    STATE(1267), 1,
      aux_sym__literal_repeat1,
    STATE(1376), 1,
      sym_concatenation,
    STATE(1974), 1,
      sym_variable_assignment,
    STATE(5026), 1,
      sym_command,
    STATE(6759), 1,
      sym_subscript,
    ACTIONS(336), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1017), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(1392), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    STATE(3581), 2,
      sym_file_redirect,
      sym_herestring_redirect,
    STATE(5014), 2,
      sym_subshell,
      sym_test_command,
    ACTIONS(1390), 3,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
    ACTIONS(1388), 5,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    STATE(1717), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [11348] = 6,
    ACTIONS(3), 1,
      sym_comment,
    STATE(1643), 1,
      aux_sym__literal_repeat1,
    STATE(1776), 1,
      sym_concatenation,
    ACTIONS(2500), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    STATE(1184), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2498), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [11417] = 6,
    ACTIONS(3), 1,
      sym_comment,
    STATE(1744), 1,
      aux_sym__literal_repeat1,
    STATE(627), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    ACTIONS(2096), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    STATE(1370), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2094), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [11486] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2965), 1,
      anon_sym_DQUOTE,
    ACTIONS(2969), 1,
      sym_variable_name,
    STATE(1454), 1,
      sym_string,
    ACTIONS(2967), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1241), 3,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(2963), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1239), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [11559] = 6,
    ACTIONS(3), 1,
      sym_comment,
    STATE(1606), 1,
      aux_sym__literal_repeat1,
    STATE(1767), 1,
      sym_concatenation,
    ACTIONS(2496), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    STATE(1275), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2494), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [11628] = 6,
    ACTIONS(3), 1,
      sym_comment,
    STATE(1744), 1,
      aux_sym__literal_repeat1,
    STATE(627), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    ACTIONS(2074), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    STATE(1370), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2072), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [11697] = 37,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(671), 1,
      anon_sym_LPAREN_LPAREN,
    ACTIONS(677), 1,
      anon_sym_LPAREN,
    ACTIONS(693), 1,
      anon_sym_LBRACK,
    ACTIONS(695), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(705), 1,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(707), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(709), 1,
      anon_sym_DOLLAR,
    ACTIONS(711), 1,
      sym__special_character,
    ACTIONS(713), 1,
      anon_sym_DQUOTE,
    ACTIONS(717), 1,
      aux_sym_number_token1,
    ACTIONS(719), 1,
      aux_sym_number_token2,
    ACTIONS(721), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(723), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(725), 1,
      anon_sym_BQUOTE,
    ACTIONS(727), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(733), 1,
      sym_variable_name,
    ACTIONS(735), 1,
      sym_test_operator,
    ACTIONS(737), 1,
      sym__brace_start,
    ACTIONS(1394), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(1396), 1,
      sym_file_descriptor,
    ACTIONS(2971), 1,
      sym_word,
    STATE(530), 1,
      sym_command_name,
    STATE(989), 1,
      aux_sym_command_repeat1,
    STATE(1026), 1,
      aux_sym__literal_repeat1,
    STATE(1169), 1,
      sym_concatenation,
    STATE(1909), 1,
      sym_variable_assignment,
    STATE(5046), 1,
      sym_command,
    STATE(6760), 1,
      sym_subscript,
    ACTIONS(715), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(729), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1392), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    STATE(3581), 2,
      sym_file_redirect,
      sym_herestring_redirect,
    STATE(5071), 2,
      sym_subshell,
      sym_test_command,
    ACTIONS(1390), 3,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
    ACTIONS(1388), 5,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    STATE(852), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [11828] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2979), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(2982), 1,
      anon_sym_DOLLAR,
    ACTIONS(2985), 1,
      sym__special_character,
    ACTIONS(2988), 1,
      anon_sym_DQUOTE,
    ACTIONS(2991), 1,
      aux_sym_number_token1,
    ACTIONS(2994), 1,
      aux_sym_number_token2,
    ACTIONS(2997), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(3000), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3003), 1,
      anon_sym_BQUOTE,
    ACTIONS(3006), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3012), 1,
      sym_test_operator,
    ACTIONS(3015), 1,
      sym__brace_start,
    STATE(1801), 1,
      aux_sym__literal_repeat1,
    ACTIONS(2976), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(3009), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(651), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    ACTIONS(2973), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    ACTIONS(2216), 4,
      sym_file_descriptor,
      sym_variable_name,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    STATE(1399), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2214), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [11927] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2965), 1,
      anon_sym_DQUOTE,
    ACTIONS(2969), 1,
      sym_variable_name,
    STATE(1454), 1,
      sym_string,
    ACTIONS(2967), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1235), 3,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(2963), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1227), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [12000] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2572), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(2574), 1,
      anon_sym_DOLLAR,
    ACTIONS(2576), 1,
      sym__special_character,
    ACTIONS(2578), 1,
      anon_sym_DQUOTE,
    ACTIONS(2580), 1,
      aux_sym_number_token1,
    ACTIONS(2582), 1,
      aux_sym_number_token2,
    ACTIONS(2584), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(2586), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2588), 1,
      anon_sym_BQUOTE,
    ACTIONS(2590), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(2594), 1,
      sym_test_operator,
    ACTIONS(2596), 1,
      sym__brace_start,
    STATE(1899), 1,
      aux_sym__literal_repeat1,
    ACTIONS(2096), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(2570), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(2592), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(621), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    ACTIONS(2568), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(1394), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2094), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [12099] = 6,
    ACTIONS(3), 1,
      sym_comment,
    STATE(1801), 1,
      aux_sym__literal_repeat1,
    STATE(651), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    ACTIONS(2074), 6,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    STATE(1399), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2072), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [12168] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(2725), 2,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
    ACTIONS(2727), 2,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
    ACTIONS(2717), 23,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_GT_GT,
      anon_sym_LBRACE,
      anon_sym_LBRACK_LBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(2715), 28,
      anon_sym_for,
      anon_sym_select,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LPAREN,
      anon_sym_while,
      anon_sym_until,
      anon_sym_if,
      anon_sym_case,
      anon_sym_function,
      anon_sym_BANG,
      anon_sym_LBRACK,
      anon_sym_declare,
      anon_sym_typeset,
      anon_sym_export,
      anon_sym_readonly,
      anon_sym_local,
      anon_sym_unset,
      anon_sym_unsetenv,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      sym__special_character,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
  [12235] = 37,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(409), 1,
      anon_sym_LPAREN_LPAREN,
    ACTIONS(415), 1,
      anon_sym_LPAREN,
    ACTIONS(437), 1,
      anon_sym_LBRACK,
    ACTIONS(439), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(449), 1,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(451), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(453), 1,
      anon_sym_DOLLAR,
    ACTIONS(455), 1,
      sym__special_character,
    ACTIONS(457), 1,
      anon_sym_DQUOTE,
    ACTIONS(461), 1,
      aux_sym_number_token1,
    ACTIONS(463), 1,
      aux_sym_number_token2,
    ACTIONS(465), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(467), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(469), 1,
      anon_sym_BQUOTE,
    ACTIONS(471), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(477), 1,
      sym_variable_name,
    ACTIONS(479), 1,
      sym_test_operator,
    ACTIONS(481), 1,
      sym__brace_start,
    ACTIONS(1394), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(1396), 1,
      sym_file_descriptor,
    ACTIONS(3018), 1,
      sym_word,
    STATE(528), 1,
      sym_command_name,
    STATE(925), 1,
      aux_sym_command_repeat1,
    STATE(926), 1,
      aux_sym__literal_repeat1,
    STATE(1082), 1,
      sym_concatenation,
    STATE(1703), 1,
      sym_variable_assignment,
    STATE(4945), 1,
      sym_command,
    STATE(6736), 1,
      sym_subscript,
    ACTIONS(459), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(473), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1392), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    STATE(3581), 2,
      sym_file_redirect,
      sym_herestring_redirect,
    STATE(4943), 2,
      sym_subshell,
      sym_test_command,
    ACTIONS(1390), 3,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
    ACTIONS(1388), 5,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    STATE(815), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [12366] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2602), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(2604), 1,
      anon_sym_DOLLAR,
    ACTIONS(2606), 1,
      sym__special_character,
    ACTIONS(2608), 1,
      anon_sym_DQUOTE,
    ACTIONS(2610), 1,
      aux_sym_number_token1,
    ACTIONS(2612), 1,
      aux_sym_number_token2,
    ACTIONS(2614), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(2616), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2618), 1,
      anon_sym_BQUOTE,
    ACTIONS(2620), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(2624), 1,
      sym_test_operator,
    ACTIONS(2626), 1,
      sym__brace_start,
    STATE(1803), 1,
      aux_sym__literal_repeat1,
    ACTIONS(2096), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(2600), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(2622), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(630), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    ACTIONS(2598), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(1381), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2094), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [12465] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3024), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(3026), 1,
      anon_sym_DOLLAR,
    ACTIONS(3028), 1,
      sym__special_character,
    ACTIONS(3030), 1,
      anon_sym_DQUOTE,
    ACTIONS(3032), 1,
      aux_sym_number_token1,
    ACTIONS(3034), 1,
      aux_sym_number_token2,
    ACTIONS(3036), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(3038), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3040), 1,
      anon_sym_BQUOTE,
    ACTIONS(3042), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3046), 1,
      sym_test_operator,
    ACTIONS(3048), 1,
      sym__brace_start,
    STATE(4597), 1,
      aux_sym__literal_repeat1,
    STATE(4964), 1,
      sym_concatenation,
    ACTIONS(2496), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(3022), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(3044), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(3020), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(4310), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2494), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [12563] = 36,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(171), 1,
      anon_sym_RPAREN_RPAREN,
    ACTIONS(1141), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1145), 1,
      anon_sym_DOLLAR,
    ACTIONS(1149), 1,
      anon_sym_DQUOTE,
    ACTIONS(1153), 1,
      aux_sym_number_token1,
    ACTIONS(1155), 1,
      aux_sym_number_token2,
    ACTIONS(1157), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1159), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1163), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1169), 1,
      sym__brace_start,
    ACTIONS(1171), 1,
      sym_word,
    ACTIONS(1181), 1,
      sym__special_character,
    ACTIONS(1185), 1,
      sym_test_operator,
    ACTIONS(3050), 1,
      anon_sym_LPAREN,
    ACTIONS(3052), 1,
      anon_sym_BANG,
    ACTIONS(3058), 1,
      anon_sym_TILDE,
    ACTIONS(3060), 1,
      anon_sym_BQUOTE,
    ACTIONS(3062), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(3064), 1,
      sym_variable_name,
    STATE(2387), 1,
      sym_command_substitution,
    STATE(2484), 1,
      aux_sym__literal_repeat1,
    STATE(2801), 1,
      sym__arithmetic_binary_expression,
    STATE(2803), 1,
      sym__arithmetic_ternary_expression,
    STATE(2809), 1,
      sym__arithmetic_unary_expression,
    STATE(2825), 1,
      sym__arithmetic_postfix_expression,
    STATE(3158), 1,
      sym__expression,
    ACTIONS(1129), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1165), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1183), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(3054), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(3056), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(2397), 4,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
    STATE(2456), 4,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_translated_string,
      sym_process_substitution,
    STATE(2836), 4,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
    STATE(2594), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
  [12691] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3072), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(3075), 1,
      anon_sym_DOLLAR,
    ACTIONS(3078), 1,
      sym__special_character,
    ACTIONS(3081), 1,
      anon_sym_DQUOTE,
    ACTIONS(3084), 1,
      aux_sym_number_token1,
    ACTIONS(3087), 1,
      aux_sym_number_token2,
    ACTIONS(3090), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(3093), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3096), 1,
      anon_sym_BQUOTE,
    ACTIONS(3099), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3105), 1,
      sym_test_operator,
    ACTIONS(3108), 1,
      sym__brace_start,
    STATE(2147), 1,
      aux_sym__literal_repeat1,
    ACTIONS(2216), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(3069), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(3102), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(660), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    ACTIONS(3066), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(1526), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2214), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [12789] = 36,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(171), 1,
      anon_sym_RPAREN_RPAREN,
    ACTIONS(1141), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1145), 1,
      anon_sym_DOLLAR,
    ACTIONS(1149), 1,
      anon_sym_DQUOTE,
    ACTIONS(1153), 1,
      aux_sym_number_token1,
    ACTIONS(1155), 1,
      aux_sym_number_token2,
    ACTIONS(1157), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1159), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1163), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1169), 1,
      sym__brace_start,
    ACTIONS(1171), 1,
      sym_word,
    ACTIONS(1181), 1,
      sym__special_character,
    ACTIONS(1185), 1,
      sym_test_operator,
    ACTIONS(3050), 1,
      anon_sym_LPAREN,
    ACTIONS(3052), 1,
      anon_sym_BANG,
    ACTIONS(3058), 1,
      anon_sym_TILDE,
    ACTIONS(3060), 1,
      anon_sym_BQUOTE,
    ACTIONS(3064), 1,
      sym_variable_name,
    ACTIONS(3111), 1,
      aux_sym__simple_variable_name_token1,
    STATE(2387), 1,
      sym_command_substitution,
    STATE(2484), 1,
      aux_sym__literal_repeat1,
    STATE(2801), 1,
      sym__arithmetic_binary_expression,
    STATE(2803), 1,
      sym__arithmetic_ternary_expression,
    STATE(2809), 1,
      sym__arithmetic_unary_expression,
    STATE(2825), 1,
      sym__arithmetic_postfix_expression,
    STATE(3158), 1,
      sym__expression,
    ACTIONS(1129), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1165), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1183), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(3054), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(3056), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(2397), 4,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
    STATE(2456), 4,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_translated_string,
      sym_process_substitution,
    STATE(2799), 4,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
    STATE(2594), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
  [12917] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3115), 1,
      anon_sym_DQUOTE,
    ACTIONS(3119), 1,
      sym_variable_name,
    STATE(1864), 1,
      sym_string,
    ACTIONS(3117), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1235), 3,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(3113), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1227), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [12989] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1835), 1,
      anon_sym_DQUOTE,
    ACTIONS(3125), 1,
      sym_variable_name,
    STATE(1416), 1,
      sym_string,
    ACTIONS(3123), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1235), 3,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(3121), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1227), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [13061] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3133), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(3136), 1,
      anon_sym_DOLLAR,
    ACTIONS(3139), 1,
      sym__special_character,
    ACTIONS(3142), 1,
      anon_sym_DQUOTE,
    ACTIONS(3145), 1,
      aux_sym_number_token1,
    ACTIONS(3148), 1,
      aux_sym_number_token2,
    ACTIONS(3151), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(3154), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3157), 1,
      anon_sym_BQUOTE,
    ACTIONS(3160), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3166), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(3169), 1,
      sym_test_operator,
    ACTIONS(3172), 1,
      sym__brace_start,
    STATE(2190), 1,
      aux_sym__literal_repeat1,
    ACTIONS(3130), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(3163), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(664), 2,
      sym_concatenation,
      aux_sym_unset_command_repeat1,
    ACTIONS(2442), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(3127), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(1681), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2440), 19,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [13161] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3024), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(3026), 1,
      anon_sym_DOLLAR,
    ACTIONS(3028), 1,
      sym__special_character,
    ACTIONS(3030), 1,
      anon_sym_DQUOTE,
    ACTIONS(3032), 1,
      aux_sym_number_token1,
    ACTIONS(3034), 1,
      aux_sym_number_token2,
    ACTIONS(3036), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(3038), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3040), 1,
      anon_sym_BQUOTE,
    ACTIONS(3042), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3048), 1,
      sym__brace_start,
    ACTIONS(3177), 1,
      sym_test_operator,
    STATE(4596), 1,
      aux_sym__literal_repeat1,
    STATE(4914), 1,
      sym_concatenation,
    ACTIONS(2500), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(3022), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(3044), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(3175), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(4331), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2498), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [13259] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1835), 1,
      anon_sym_DQUOTE,
    ACTIONS(3125), 1,
      sym_variable_name,
    STATE(1416), 1,
      sym_string,
    ACTIONS(3123), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1241), 3,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(3121), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1239), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [13331] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3183), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(3185), 1,
      anon_sym_DOLLAR,
    ACTIONS(3187), 1,
      sym__special_character,
    ACTIONS(3189), 1,
      anon_sym_DQUOTE,
    ACTIONS(3191), 1,
      aux_sym_number_token1,
    ACTIONS(3193), 1,
      aux_sym_number_token2,
    ACTIONS(3195), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(3197), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3199), 1,
      anon_sym_BQUOTE,
    ACTIONS(3201), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3205), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(3207), 1,
      sym_test_operator,
    ACTIONS(3209), 1,
      sym__brace_start,
    STATE(1966), 1,
      aux_sym__literal_repeat1,
    ACTIONS(2564), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(3181), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(3203), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(709), 2,
      sym_concatenation,
      aux_sym_unset_command_repeat1,
    ACTIONS(3179), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(1650), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2562), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [13431] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3115), 1,
      anon_sym_DQUOTE,
    ACTIONS(3119), 1,
      sym_variable_name,
    STATE(1864), 1,
      sym_string,
    ACTIONS(3117), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1241), 3,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(3113), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1239), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [13503] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3213), 1,
      anon_sym_DQUOTE,
    ACTIONS(3217), 1,
      sym_variable_name,
    STATE(1919), 1,
      sym_string,
    ACTIONS(3215), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1241), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
    ACTIONS(3211), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1239), 36,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [13575] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2790), 1,
      anon_sym_DQUOTE,
    ACTIONS(3223), 1,
      sym_variable_name,
    STATE(1600), 1,
      sym_string,
    ACTIONS(3221), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1241), 3,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(3219), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1239), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [13647] = 36,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(231), 1,
      anon_sym_RPAREN_RPAREN,
    ACTIONS(1141), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1145), 1,
      anon_sym_DOLLAR,
    ACTIONS(1149), 1,
      anon_sym_DQUOTE,
    ACTIONS(1153), 1,
      aux_sym_number_token1,
    ACTIONS(1155), 1,
      aux_sym_number_token2,
    ACTIONS(1157), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1159), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1163), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1169), 1,
      sym__brace_start,
    ACTIONS(1171), 1,
      sym_word,
    ACTIONS(1181), 1,
      sym__special_character,
    ACTIONS(1185), 1,
      sym_test_operator,
    ACTIONS(3050), 1,
      anon_sym_LPAREN,
    ACTIONS(3052), 1,
      anon_sym_BANG,
    ACTIONS(3058), 1,
      anon_sym_TILDE,
    ACTIONS(3060), 1,
      anon_sym_BQUOTE,
    ACTIONS(3064), 1,
      sym_variable_name,
    ACTIONS(3225), 1,
      aux_sym__simple_variable_name_token1,
    STATE(2387), 1,
      sym_command_substitution,
    STATE(2484), 1,
      aux_sym__literal_repeat1,
    STATE(2801), 1,
      sym__arithmetic_binary_expression,
    STATE(2803), 1,
      sym__arithmetic_ternary_expression,
    STATE(2809), 1,
      sym__arithmetic_unary_expression,
    STATE(2825), 1,
      sym__arithmetic_postfix_expression,
    STATE(3139), 1,
      sym__expression,
    ACTIONS(1129), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1165), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1183), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(3054), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(3056), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(2397), 4,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
    STATE(2456), 4,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_translated_string,
      sym_process_substitution,
    STATE(2756), 4,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
    STATE(2594), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
  [13775] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3233), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(3236), 1,
      anon_sym_DOLLAR,
    ACTIONS(3239), 1,
      sym__special_character,
    ACTIONS(3242), 1,
      anon_sym_DQUOTE,
    ACTIONS(3245), 1,
      aux_sym_number_token1,
    ACTIONS(3248), 1,
      aux_sym_number_token2,
    ACTIONS(3251), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(3254), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3257), 1,
      anon_sym_BQUOTE,
    ACTIONS(3260), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3266), 1,
      sym_test_operator,
    ACTIONS(3269), 1,
      sym__brace_start,
    STATE(2161), 1,
      aux_sym__literal_repeat1,
    ACTIONS(2216), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(3230), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(3263), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(672), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    ACTIONS(3227), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(1711), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2214), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [13873] = 6,
    ACTIONS(3), 1,
      sym_comment,
    STATE(1744), 1,
      aux_sym__literal_repeat1,
    STATE(704), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    ACTIONS(2074), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    STATE(1688), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2072), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [13941] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2790), 1,
      anon_sym_DQUOTE,
    ACTIONS(3223), 1,
      sym_variable_name,
    STATE(1600), 1,
      sym_string,
    ACTIONS(3221), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1235), 3,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(3219), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1227), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [14013] = 26,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1549), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1551), 1,
      anon_sym_DOLLAR,
    ACTIONS(1555), 1,
      anon_sym_DQUOTE,
    ACTIONS(1557), 1,
      aux_sym_number_token1,
    ACTIONS(1559), 1,
      aux_sym_number_token2,
    ACTIONS(1561), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1563), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1565), 1,
      anon_sym_BQUOTE,
    ACTIONS(1567), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1573), 1,
      sym__bare_dollar,
    ACTIONS(1575), 1,
      sym__brace_start,
    ACTIONS(2420), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(2422), 1,
      sym__special_character,
    ACTIONS(2424), 1,
      sym_test_operator,
    STATE(703), 1,
      aux_sym_command_repeat2,
    STATE(1321), 1,
      aux_sym__literal_repeat1,
    STATE(1712), 1,
      sym_herestring_redirect,
    STATE(1716), 1,
      sym_concatenation,
    ACTIONS(1543), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1569), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1615), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(2418), 2,
      anon_sym_EQ_EQ,
      anon_sym_EQ_TILDE,
    ACTIONS(2416), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(1794), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(1613), 16,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [14121] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3274), 1,
      anon_sym_DQUOTE,
    ACTIONS(3278), 1,
      sym_variable_name,
    STATE(1825), 1,
      sym_string,
    ACTIONS(3276), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1241), 3,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(3272), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1239), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [14193] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3282), 1,
      anon_sym_DQUOTE,
    ACTIONS(3286), 1,
      sym_variable_name,
    STATE(1628), 1,
      sym_string,
    ACTIONS(3284), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1235), 3,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(3280), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1227), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [14265] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3294), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(3297), 1,
      anon_sym_DOLLAR,
    ACTIONS(3300), 1,
      sym__special_character,
    ACTIONS(3303), 1,
      anon_sym_DQUOTE,
    ACTIONS(3306), 1,
      aux_sym_number_token1,
    ACTIONS(3309), 1,
      aux_sym_number_token2,
    ACTIONS(3312), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(3315), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3318), 1,
      anon_sym_BQUOTE,
    ACTIONS(3321), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3327), 1,
      sym_test_operator,
    ACTIONS(3330), 1,
      sym__brace_start,
    STATE(2024), 1,
      aux_sym__literal_repeat1,
    ACTIONS(3291), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(3324), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(678), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    ACTIONS(2216), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(3288), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(1630), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2214), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [14363] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3274), 1,
      anon_sym_DQUOTE,
    ACTIONS(3278), 1,
      sym_variable_name,
    STATE(1825), 1,
      sym_string,
    ACTIONS(3276), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1235), 3,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(3272), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1227), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [14435] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3274), 1,
      anon_sym_DQUOTE,
    ACTIONS(3337), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(3339), 1,
      anon_sym_DOLLAR,
    ACTIONS(3341), 1,
      sym__special_character,
    ACTIONS(3343), 1,
      aux_sym_number_token1,
    ACTIONS(3345), 1,
      aux_sym_number_token2,
    ACTIONS(3347), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(3349), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3351), 1,
      anon_sym_BQUOTE,
    ACTIONS(3353), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3357), 1,
      sym_test_operator,
    ACTIONS(3359), 1,
      sym__brace_start,
    STATE(2147), 1,
      aux_sym__literal_repeat1,
    ACTIONS(2074), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(3335), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(3355), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(660), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    ACTIONS(3333), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(1526), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2072), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [14533] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3365), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(3367), 1,
      anon_sym_DOLLAR,
    ACTIONS(3369), 1,
      sym__special_character,
    ACTIONS(3371), 1,
      anon_sym_DQUOTE,
    ACTIONS(3373), 1,
      aux_sym_number_token1,
    ACTIONS(3375), 1,
      aux_sym_number_token2,
    ACTIONS(3377), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(3379), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3381), 1,
      anon_sym_BQUOTE,
    ACTIONS(3383), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3387), 1,
      sym_test_operator,
    ACTIONS(3389), 1,
      sym__brace_start,
    STATE(4507), 1,
      aux_sym__literal_repeat1,
    STATE(4879), 1,
      sym_concatenation,
    ACTIONS(2496), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(3363), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(3385), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(3361), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(4271), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2494), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [14631] = 6,
    ACTIONS(3), 1,
      sym_comment,
    STATE(1826), 1,
      aux_sym__literal_repeat1,
    STATE(2201), 1,
      sym_concatenation,
    ACTIONS(2496), 6,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    STATE(1446), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2494), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [14699] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3213), 1,
      anon_sym_DQUOTE,
    ACTIONS(3217), 1,
      sym_variable_name,
    STATE(1919), 1,
      sym_string,
    ACTIONS(3215), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1235), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
    ACTIONS(3211), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1227), 36,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [14771] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3115), 1,
      anon_sym_DQUOTE,
    ACTIONS(3395), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(3397), 1,
      anon_sym_DOLLAR,
    ACTIONS(3399), 1,
      sym__special_character,
    ACTIONS(3401), 1,
      aux_sym_number_token1,
    ACTIONS(3403), 1,
      aux_sym_number_token2,
    ACTIONS(3405), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(3407), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3409), 1,
      anon_sym_BQUOTE,
    ACTIONS(3411), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3415), 1,
      sym_test_operator,
    ACTIONS(3417), 1,
      sym__brace_start,
    STATE(2161), 1,
      aux_sym__literal_repeat1,
    ACTIONS(2074), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(3393), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(3413), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(672), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    ACTIONS(3391), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(1711), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2072), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [14869] = 36,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(229), 1,
      anon_sym_RPAREN_RPAREN,
    ACTIONS(1141), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1145), 1,
      anon_sym_DOLLAR,
    ACTIONS(1149), 1,
      anon_sym_DQUOTE,
    ACTIONS(1153), 1,
      aux_sym_number_token1,
    ACTIONS(1155), 1,
      aux_sym_number_token2,
    ACTIONS(1157), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1159), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1163), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1169), 1,
      sym__brace_start,
    ACTIONS(1171), 1,
      sym_word,
    ACTIONS(1181), 1,
      sym__special_character,
    ACTIONS(1185), 1,
      sym_test_operator,
    ACTIONS(3050), 1,
      anon_sym_LPAREN,
    ACTIONS(3052), 1,
      anon_sym_BANG,
    ACTIONS(3058), 1,
      anon_sym_TILDE,
    ACTIONS(3060), 1,
      anon_sym_BQUOTE,
    ACTIONS(3064), 1,
      sym_variable_name,
    ACTIONS(3419), 1,
      aux_sym__simple_variable_name_token1,
    STATE(2387), 1,
      sym_command_substitution,
    STATE(2484), 1,
      aux_sym__literal_repeat1,
    STATE(2801), 1,
      sym__arithmetic_binary_expression,
    STATE(2803), 1,
      sym__arithmetic_ternary_expression,
    STATE(2809), 1,
      sym__arithmetic_unary_expression,
    STATE(2825), 1,
      sym__arithmetic_postfix_expression,
    STATE(3033), 1,
      sym__expression,
    ACTIONS(1129), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1165), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1183), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(3054), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(3056), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(2397), 4,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
    STATE(2456), 4,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_translated_string,
      sym_process_substitution,
    STATE(2811), 4,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
    STATE(2594), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
  [14997] = 6,
    ACTIONS(3), 1,
      sym_comment,
    STATE(1744), 1,
      aux_sym__literal_repeat1,
    STATE(704), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    ACTIONS(2096), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    STATE(1688), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2094), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [15065] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3423), 1,
      anon_sym_DQUOTE,
    ACTIONS(3427), 1,
      sym_variable_name,
    STATE(1686), 1,
      sym_string,
    ACTIONS(3425), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1235), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
    ACTIONS(3421), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1227), 36,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [15137] = 26,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1549), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1551), 1,
      anon_sym_DOLLAR,
    ACTIONS(1555), 1,
      anon_sym_DQUOTE,
    ACTIONS(1557), 1,
      aux_sym_number_token1,
    ACTIONS(1559), 1,
      aux_sym_number_token2,
    ACTIONS(1561), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1563), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1565), 1,
      anon_sym_BQUOTE,
    ACTIONS(1567), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1573), 1,
      sym__bare_dollar,
    ACTIONS(1575), 1,
      sym__brace_start,
    ACTIONS(2420), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(2422), 1,
      sym__special_character,
    ACTIONS(2424), 1,
      sym_test_operator,
    STATE(703), 1,
      aux_sym_command_repeat2,
    STATE(1321), 1,
      aux_sym__literal_repeat1,
    STATE(1712), 1,
      sym_herestring_redirect,
    STATE(1716), 1,
      sym_concatenation,
    ACTIONS(1543), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1569), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1627), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(2418), 2,
      anon_sym_EQ_EQ,
      anon_sym_EQ_TILDE,
    ACTIONS(2416), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(1794), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(1625), 16,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [15245] = 6,
    ACTIONS(3), 1,
      sym_comment,
    STATE(1895), 1,
      aux_sym__literal_repeat1,
    STATE(2117), 1,
      sym_concatenation,
    ACTIONS(2500), 6,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    STATE(1397), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2498), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [15313] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3433), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(3435), 1,
      anon_sym_DOLLAR,
    ACTIONS(3437), 1,
      sym__special_character,
    ACTIONS(3439), 1,
      anon_sym_DQUOTE,
    ACTIONS(3441), 1,
      aux_sym_number_token1,
    ACTIONS(3443), 1,
      aux_sym_number_token2,
    ACTIONS(3445), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(3447), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3449), 1,
      anon_sym_BQUOTE,
    ACTIONS(3451), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3455), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(3457), 1,
      sym_test_operator,
    ACTIONS(3459), 1,
      sym__brace_start,
    STATE(2190), 1,
      aux_sym__literal_repeat1,
    ACTIONS(3431), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(3453), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(695), 2,
      sym_concatenation,
      aux_sym_unset_command_repeat1,
    ACTIONS(2299), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(3429), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(1681), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2297), 19,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [15413] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3115), 1,
      anon_sym_DQUOTE,
    ACTIONS(3395), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(3397), 1,
      anon_sym_DOLLAR,
    ACTIONS(3399), 1,
      sym__special_character,
    ACTIONS(3401), 1,
      aux_sym_number_token1,
    ACTIONS(3403), 1,
      aux_sym_number_token2,
    ACTIONS(3405), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(3407), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3409), 1,
      anon_sym_BQUOTE,
    ACTIONS(3411), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3415), 1,
      sym_test_operator,
    ACTIONS(3417), 1,
      sym__brace_start,
    STATE(2161), 1,
      aux_sym__literal_repeat1,
    ACTIONS(2096), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(3393), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(3413), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(672), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    ACTIONS(3391), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(1711), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2094), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [15511] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3282), 1,
      anon_sym_DQUOTE,
    ACTIONS(3286), 1,
      sym_variable_name,
    STATE(1628), 1,
      sym_string,
    ACTIONS(3284), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1241), 3,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(3280), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1239), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [15583] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3183), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(3185), 1,
      anon_sym_DOLLAR,
    ACTIONS(3187), 1,
      sym__special_character,
    ACTIONS(3189), 1,
      anon_sym_DQUOTE,
    ACTIONS(3191), 1,
      aux_sym_number_token1,
    ACTIONS(3193), 1,
      aux_sym_number_token2,
    ACTIONS(3195), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(3197), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3199), 1,
      anon_sym_BQUOTE,
    ACTIONS(3201), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3207), 1,
      sym_test_operator,
    ACTIONS(3209), 1,
      sym__brace_start,
    ACTIONS(3461), 1,
      aux_sym__simple_variable_name_token1,
    STATE(1966), 1,
      aux_sym__literal_repeat1,
    ACTIONS(2299), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(3181), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(3203), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(667), 2,
      sym_concatenation,
      aux_sym_unset_command_repeat1,
    ACTIONS(3179), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(1650), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2297), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [15683] = 6,
    ACTIONS(3), 1,
      sym_comment,
    STATE(1739), 1,
      aux_sym__literal_repeat1,
    STATE(1953), 1,
      sym_concatenation,
    ACTIONS(2496), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    STATE(1320), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2494), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [15751] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3433), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(3435), 1,
      anon_sym_DOLLAR,
    ACTIONS(3437), 1,
      sym__special_character,
    ACTIONS(3439), 1,
      anon_sym_DQUOTE,
    ACTIONS(3441), 1,
      aux_sym_number_token1,
    ACTIONS(3443), 1,
      aux_sym_number_token2,
    ACTIONS(3445), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(3447), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3449), 1,
      anon_sym_BQUOTE,
    ACTIONS(3451), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3457), 1,
      sym_test_operator,
    ACTIONS(3459), 1,
      sym__brace_start,
    ACTIONS(3463), 1,
      aux_sym__simple_variable_name_token1,
    STATE(2190), 1,
      aux_sym__literal_repeat1,
    ACTIONS(3431), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(3453), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(664), 2,
      sym_concatenation,
      aux_sym_unset_command_repeat1,
    ACTIONS(2564), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(3429), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(1681), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2562), 19,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [15851] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1811), 1,
      anon_sym_DQUOTE,
    ACTIONS(1815), 1,
      sym_variable_name,
    STATE(1061), 1,
      sym_string,
    ACTIONS(1813), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1241), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
    ACTIONS(1809), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1239), 36,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [15923] = 36,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(171), 1,
      anon_sym_RPAREN_RPAREN,
    ACTIONS(1141), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1145), 1,
      anon_sym_DOLLAR,
    ACTIONS(1149), 1,
      anon_sym_DQUOTE,
    ACTIONS(1153), 1,
      aux_sym_number_token1,
    ACTIONS(1155), 1,
      aux_sym_number_token2,
    ACTIONS(1157), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1159), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1163), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1169), 1,
      sym__brace_start,
    ACTIONS(1171), 1,
      sym_word,
    ACTIONS(1181), 1,
      sym__special_character,
    ACTIONS(1185), 1,
      sym_test_operator,
    ACTIONS(3050), 1,
      anon_sym_LPAREN,
    ACTIONS(3052), 1,
      anon_sym_BANG,
    ACTIONS(3058), 1,
      anon_sym_TILDE,
    ACTIONS(3060), 1,
      anon_sym_BQUOTE,
    ACTIONS(3064), 1,
      sym_variable_name,
    ACTIONS(3465), 1,
      aux_sym__simple_variable_name_token1,
    STATE(2387), 1,
      sym_command_substitution,
    STATE(2484), 1,
      aux_sym__literal_repeat1,
    STATE(2801), 1,
      sym__arithmetic_binary_expression,
    STATE(2803), 1,
      sym__arithmetic_ternary_expression,
    STATE(2809), 1,
      sym__arithmetic_unary_expression,
    STATE(2825), 1,
      sym__arithmetic_postfix_expression,
    STATE(3158), 1,
      sym__expression,
    ACTIONS(1129), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1165), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1183), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(3054), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(3056), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(2397), 4,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
    STATE(2456), 4,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_translated_string,
      sym_process_substitution,
    STATE(2782), 4,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
    STATE(2594), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
  [16051] = 36,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(229), 1,
      anon_sym_RPAREN_RPAREN,
    ACTIONS(1141), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1145), 1,
      anon_sym_DOLLAR,
    ACTIONS(1149), 1,
      anon_sym_DQUOTE,
    ACTIONS(1153), 1,
      aux_sym_number_token1,
    ACTIONS(1155), 1,
      aux_sym_number_token2,
    ACTIONS(1157), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1159), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1163), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1169), 1,
      sym__brace_start,
    ACTIONS(1171), 1,
      sym_word,
    ACTIONS(1181), 1,
      sym__special_character,
    ACTIONS(1185), 1,
      sym_test_operator,
    ACTIONS(3050), 1,
      anon_sym_LPAREN,
    ACTIONS(3052), 1,
      anon_sym_BANG,
    ACTIONS(3058), 1,
      anon_sym_TILDE,
    ACTIONS(3060), 1,
      anon_sym_BQUOTE,
    ACTIONS(3064), 1,
      sym_variable_name,
    ACTIONS(3467), 1,
      aux_sym__simple_variable_name_token1,
    STATE(2387), 1,
      sym_command_substitution,
    STATE(2484), 1,
      aux_sym__literal_repeat1,
    STATE(2801), 1,
      sym__arithmetic_binary_expression,
    STATE(2803), 1,
      sym__arithmetic_ternary_expression,
    STATE(2809), 1,
      sym__arithmetic_unary_expression,
    STATE(2825), 1,
      sym__arithmetic_postfix_expression,
    STATE(3033), 1,
      sym__expression,
    ACTIONS(1129), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1165), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1183), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(3054), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(3056), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(2397), 4,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
    STATE(2456), 4,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_translated_string,
      sym_process_substitution,
    STATE(2777), 4,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
    STATE(2594), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
  [16179] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1811), 1,
      anon_sym_DQUOTE,
    ACTIONS(1815), 1,
      sym_variable_name,
    STATE(1061), 1,
      sym_string,
    ACTIONS(1813), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1235), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
    ACTIONS(1809), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1227), 36,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [16251] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3423), 1,
      anon_sym_DQUOTE,
    ACTIONS(3427), 1,
      sym_variable_name,
    STATE(1686), 1,
      sym_string,
    ACTIONS(3425), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1241), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
    ACTIONS(3421), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1239), 36,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [16323] = 36,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(235), 1,
      anon_sym_RPAREN_RPAREN,
    ACTIONS(1141), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1145), 1,
      anon_sym_DOLLAR,
    ACTIONS(1149), 1,
      anon_sym_DQUOTE,
    ACTIONS(1153), 1,
      aux_sym_number_token1,
    ACTIONS(1155), 1,
      aux_sym_number_token2,
    ACTIONS(1157), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1159), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1163), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1169), 1,
      sym__brace_start,
    ACTIONS(1171), 1,
      sym_word,
    ACTIONS(1181), 1,
      sym__special_character,
    ACTIONS(1185), 1,
      sym_test_operator,
    ACTIONS(3050), 1,
      anon_sym_LPAREN,
    ACTIONS(3052), 1,
      anon_sym_BANG,
    ACTIONS(3058), 1,
      anon_sym_TILDE,
    ACTIONS(3060), 1,
      anon_sym_BQUOTE,
    ACTIONS(3064), 1,
      sym_variable_name,
    ACTIONS(3469), 1,
      aux_sym__simple_variable_name_token1,
    STATE(2387), 1,
      sym_command_substitution,
    STATE(2484), 1,
      aux_sym__literal_repeat1,
    STATE(2801), 1,
      sym__arithmetic_binary_expression,
    STATE(2803), 1,
      sym__arithmetic_ternary_expression,
    STATE(2809), 1,
      sym__arithmetic_unary_expression,
    STATE(2825), 1,
      sym__arithmetic_postfix_expression,
    STATE(3150), 1,
      sym__expression,
    ACTIONS(1129), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1165), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1183), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(3054), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(3056), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(2397), 4,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
    STATE(2456), 4,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_translated_string,
      sym_process_substitution,
    STATE(2743), 4,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
    STATE(2594), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
  [16451] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3213), 1,
      anon_sym_DQUOTE,
    ACTIONS(3475), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(3477), 1,
      anon_sym_DOLLAR,
    ACTIONS(3479), 1,
      sym__special_character,
    ACTIONS(3481), 1,
      aux_sym_number_token1,
    ACTIONS(3483), 1,
      aux_sym_number_token2,
    ACTIONS(3485), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(3487), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3489), 1,
      anon_sym_BQUOTE,
    ACTIONS(3491), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3495), 1,
      sym_test_operator,
    ACTIONS(3497), 1,
      sym__brace_start,
    STATE(2024), 1,
      aux_sym__literal_repeat1,
    ACTIONS(3473), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(3493), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(678), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    ACTIONS(2074), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(3471), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(1630), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2072), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [16549] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1491), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1871), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1874), 1,
      anon_sym_DOLLAR,
    ACTIONS(1880), 1,
      anon_sym_DQUOTE,
    ACTIONS(1883), 1,
      aux_sym_number_token1,
    ACTIONS(1886), 1,
      aux_sym_number_token2,
    ACTIONS(1889), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1892), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1895), 1,
      anon_sym_BQUOTE,
    ACTIONS(1898), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1910), 1,
      sym__bare_dollar,
    ACTIONS(1913), 1,
      sym__brace_start,
    ACTIONS(3505), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(3508), 1,
      sym__special_character,
    ACTIONS(3511), 1,
      sym_file_descriptor,
    ACTIONS(3514), 1,
      sym_test_operator,
    STATE(703), 1,
      aux_sym_command_repeat2,
    STATE(1321), 1,
      aux_sym__literal_repeat1,
    STATE(1712), 1,
      sym_herestring_redirect,
    STATE(1716), 1,
      sym_concatenation,
    ACTIONS(1862), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1901), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(3502), 2,
      anon_sym_EQ_EQ,
      anon_sym_EQ_TILDE,
    ACTIONS(3499), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(1794), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(1486), 16,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [16659] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2735), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(2738), 1,
      anon_sym_DOLLAR,
    ACTIONS(2744), 1,
      anon_sym_DQUOTE,
    ACTIONS(2747), 1,
      aux_sym_number_token1,
    ACTIONS(2750), 1,
      aux_sym_number_token2,
    ACTIONS(2753), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(2756), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2759), 1,
      anon_sym_BQUOTE,
    ACTIONS(2762), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(2771), 1,
      sym__brace_start,
    ACTIONS(3520), 1,
      sym__special_character,
    ACTIONS(3523), 1,
      sym_test_operator,
    STATE(1744), 1,
      aux_sym__literal_repeat1,
    ACTIONS(2732), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(2765), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(704), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    ACTIONS(2216), 3,
      sym_file_descriptor,
      sym_variable_name,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(3517), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(1688), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2214), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [16757] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3213), 1,
      anon_sym_DQUOTE,
    ACTIONS(3475), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(3477), 1,
      anon_sym_DOLLAR,
    ACTIONS(3479), 1,
      sym__special_character,
    ACTIONS(3481), 1,
      aux_sym_number_token1,
    ACTIONS(3483), 1,
      aux_sym_number_token2,
    ACTIONS(3485), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(3487), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3489), 1,
      anon_sym_BQUOTE,
    ACTIONS(3491), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3495), 1,
      sym_test_operator,
    ACTIONS(3497), 1,
      sym__brace_start,
    STATE(2024), 1,
      aux_sym__literal_repeat1,
    ACTIONS(3473), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(3493), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(678), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    ACTIONS(2096), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(3471), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(1630), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2094), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [16855] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3365), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(3367), 1,
      anon_sym_DOLLAR,
    ACTIONS(3369), 1,
      sym__special_character,
    ACTIONS(3371), 1,
      anon_sym_DQUOTE,
    ACTIONS(3373), 1,
      aux_sym_number_token1,
    ACTIONS(3375), 1,
      aux_sym_number_token2,
    ACTIONS(3377), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(3379), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3381), 1,
      anon_sym_BQUOTE,
    ACTIONS(3383), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3389), 1,
      sym__brace_start,
    ACTIONS(3528), 1,
      sym_test_operator,
    STATE(4554), 1,
      aux_sym__literal_repeat1,
    STATE(5004), 1,
      sym_concatenation,
    ACTIONS(2500), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(3363), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(3385), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(3526), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(4370), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2498), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [16953] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3274), 1,
      anon_sym_DQUOTE,
    ACTIONS(3337), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(3339), 1,
      anon_sym_DOLLAR,
    ACTIONS(3341), 1,
      sym__special_character,
    ACTIONS(3343), 1,
      aux_sym_number_token1,
    ACTIONS(3345), 1,
      aux_sym_number_token2,
    ACTIONS(3347), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(3349), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3351), 1,
      anon_sym_BQUOTE,
    ACTIONS(3353), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3357), 1,
      sym_test_operator,
    ACTIONS(3359), 1,
      sym__brace_start,
    STATE(2147), 1,
      aux_sym__literal_repeat1,
    ACTIONS(2096), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(3335), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(3355), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(660), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    ACTIONS(3333), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(1526), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2094), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [17051] = 6,
    ACTIONS(3), 1,
      sym_comment,
    STATE(1740), 1,
      aux_sym__literal_repeat1,
    STATE(1957), 1,
      sym_concatenation,
    ACTIONS(2500), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    STATE(1336), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2498), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [17119] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3536), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(3539), 1,
      anon_sym_DOLLAR,
    ACTIONS(3542), 1,
      sym__special_character,
    ACTIONS(3545), 1,
      anon_sym_DQUOTE,
    ACTIONS(3548), 1,
      aux_sym_number_token1,
    ACTIONS(3551), 1,
      aux_sym_number_token2,
    ACTIONS(3554), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(3557), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3560), 1,
      anon_sym_BQUOTE,
    ACTIONS(3563), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3569), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(3572), 1,
      sym_test_operator,
    ACTIONS(3575), 1,
      sym__brace_start,
    STATE(1966), 1,
      aux_sym__literal_repeat1,
    ACTIONS(2442), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(3533), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(3566), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(709), 2,
      sym_concatenation,
      aux_sym_unset_command_repeat1,
    ACTIONS(3530), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(1650), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2440), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [17219] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3580), 1,
      anon_sym_DQUOTE,
    ACTIONS(3584), 1,
      sym_variable_name,
    STATE(2121), 1,
      sym_string,
    ACTIONS(3582), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1235), 3,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(3578), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1227), 36,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [17290] = 35,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1141), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1145), 1,
      anon_sym_DOLLAR,
    ACTIONS(1149), 1,
      anon_sym_DQUOTE,
    ACTIONS(1153), 1,
      aux_sym_number_token1,
    ACTIONS(1155), 1,
      aux_sym_number_token2,
    ACTIONS(1157), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1159), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1163), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1169), 1,
      sym__brace_start,
    ACTIONS(1171), 1,
      sym_word,
    ACTIONS(1181), 1,
      sym__special_character,
    ACTIONS(1185), 1,
      sym_test_operator,
    ACTIONS(3050), 1,
      anon_sym_LPAREN,
    ACTIONS(3052), 1,
      anon_sym_BANG,
    ACTIONS(3058), 1,
      anon_sym_TILDE,
    ACTIONS(3060), 1,
      anon_sym_BQUOTE,
    ACTIONS(3064), 1,
      sym_variable_name,
    ACTIONS(3586), 1,
      aux_sym__simple_variable_name_token1,
    STATE(2387), 1,
      sym_command_substitution,
    STATE(2484), 1,
      aux_sym__literal_repeat1,
    STATE(2801), 1,
      sym__arithmetic_binary_expression,
    STATE(2803), 1,
      sym__arithmetic_ternary_expression,
    STATE(2809), 1,
      sym__arithmetic_unary_expression,
    STATE(2825), 1,
      sym__arithmetic_postfix_expression,
    STATE(3108), 1,
      sym__expression,
    ACTIONS(1129), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1165), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1183), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(3054), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(3056), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(2397), 4,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
    STATE(2456), 4,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_translated_string,
      sym_process_substitution,
    STATE(2840), 4,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
    STATE(2594), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
  [17415] = 35,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(237), 1,
      sym_word,
    ACTIONS(258), 1,
      anon_sym_DOLLAR,
    ACTIONS(264), 1,
      aux_sym_number_token1,
    ACTIONS(266), 1,
      aux_sym_number_token2,
    ACTIONS(270), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(282), 1,
      sym_test_operator,
    ACTIONS(284), 1,
      sym__brace_start,
    ACTIONS(1075), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1079), 1,
      sym__special_character,
    ACTIONS(1081), 1,
      anon_sym_DQUOTE,
    ACTIONS(1085), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1087), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3588), 1,
      anon_sym_LPAREN,
    ACTIONS(3590), 1,
      anon_sym_BANG,
    ACTIONS(3596), 1,
      anon_sym_TILDE,
    ACTIONS(3598), 1,
      anon_sym_BQUOTE,
    ACTIONS(3600), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(3602), 1,
      sym_variable_name,
    STATE(2509), 1,
      sym_command_substitution,
    STATE(2690), 1,
      aux_sym__literal_repeat1,
    STATE(3130), 1,
      sym__expression,
    STATE(3359), 1,
      sym__arithmetic_postfix_expression,
    STATE(3361), 1,
      sym__arithmetic_unary_expression,
    STATE(3367), 1,
      sym__arithmetic_ternary_expression,
    STATE(3375), 1,
      sym__arithmetic_binary_expression,
    ACTIONS(1073), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1083), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(1089), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(3592), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(3594), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(2500), 4,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_translated_string,
      sym_process_substitution,
    STATE(2506), 4,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
    STATE(3281), 4,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
    STATE(3053), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
  [17540] = 35,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(237), 1,
      sym_word,
    ACTIONS(258), 1,
      anon_sym_DOLLAR,
    ACTIONS(264), 1,
      aux_sym_number_token1,
    ACTIONS(266), 1,
      aux_sym_number_token2,
    ACTIONS(270), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(282), 1,
      sym_test_operator,
    ACTIONS(284), 1,
      sym__brace_start,
    ACTIONS(1075), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1079), 1,
      sym__special_character,
    ACTIONS(1081), 1,
      anon_sym_DQUOTE,
    ACTIONS(1085), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1087), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3588), 1,
      anon_sym_LPAREN,
    ACTIONS(3590), 1,
      anon_sym_BANG,
    ACTIONS(3596), 1,
      anon_sym_TILDE,
    ACTIONS(3598), 1,
      anon_sym_BQUOTE,
    ACTIONS(3602), 1,
      sym_variable_name,
    ACTIONS(3604), 1,
      aux_sym__simple_variable_name_token1,
    STATE(2509), 1,
      sym_command_substitution,
    STATE(2690), 1,
      aux_sym__literal_repeat1,
    STATE(3100), 1,
      sym__expression,
    STATE(3359), 1,
      sym__arithmetic_postfix_expression,
    STATE(3361), 1,
      sym__arithmetic_unary_expression,
    STATE(3367), 1,
      sym__arithmetic_ternary_expression,
    STATE(3375), 1,
      sym__arithmetic_binary_expression,
    ACTIONS(1073), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1083), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(1089), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(3592), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(3594), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(2500), 4,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_translated_string,
      sym_process_substitution,
    STATE(2506), 4,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
    STATE(3283), 4,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
    STATE(3053), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
  [17665] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3580), 1,
      anon_sym_DQUOTE,
    ACTIONS(3610), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(3612), 1,
      anon_sym_DOLLAR,
    ACTIONS(3614), 1,
      sym__special_character,
    ACTIONS(3616), 1,
      aux_sym_number_token1,
    ACTIONS(3618), 1,
      aux_sym_number_token2,
    ACTIONS(3620), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(3622), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3624), 1,
      anon_sym_BQUOTE,
    ACTIONS(3626), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3630), 1,
      sym_test_operator,
    ACTIONS(3632), 1,
      sym__brace_start,
    STATE(2270), 1,
      aux_sym__literal_repeat1,
    ACTIONS(2074), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(3608), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(3628), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(741), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    ACTIONS(3606), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(1839), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2072), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [17762] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3233), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(3236), 1,
      anon_sym_DOLLAR,
    ACTIONS(3242), 1,
      anon_sym_DQUOTE,
    ACTIONS(3245), 1,
      aux_sym_number_token1,
    ACTIONS(3248), 1,
      aux_sym_number_token2,
    ACTIONS(3251), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(3254), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3257), 1,
      anon_sym_BQUOTE,
    ACTIONS(3260), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3269), 1,
      sym__brace_start,
    ACTIONS(3637), 1,
      sym__special_character,
    ACTIONS(3640), 1,
      sym_test_operator,
    STATE(2161), 1,
      aux_sym__literal_repeat1,
    ACTIONS(2216), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(3230), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(3263), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(715), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    ACTIONS(3634), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(1918), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2214), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [17859] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2042), 1,
      anon_sym_DQUOTE,
    ACTIONS(3647), 1,
      sym_variable_name,
    STATE(1547), 1,
      sym_string,
    ACTIONS(3645), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1235), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
    ACTIONS(3643), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1227), 35,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [17930] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3439), 1,
      anon_sym_DQUOTE,
    ACTIONS(3653), 1,
      sym_variable_name,
    STATE(1878), 1,
      sym_string,
    ACTIONS(3651), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1235), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
    ACTIONS(3649), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1227), 35,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [18001] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3183), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(3185), 1,
      anon_sym_DOLLAR,
    ACTIONS(3189), 1,
      anon_sym_DQUOTE,
    ACTIONS(3191), 1,
      aux_sym_number_token1,
    ACTIONS(3193), 1,
      aux_sym_number_token2,
    ACTIONS(3195), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(3197), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3199), 1,
      anon_sym_BQUOTE,
    ACTIONS(3201), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3209), 1,
      sym__brace_start,
    ACTIONS(3657), 1,
      sym__special_character,
    ACTIONS(3659), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(3661), 1,
      sym_test_operator,
    STATE(1966), 1,
      aux_sym__literal_repeat1,
    ACTIONS(2299), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(3181), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(3203), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(743), 2,
      sym_concatenation,
      aux_sym_unset_command_repeat1,
    ACTIONS(3655), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(1902), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2297), 19,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [18100] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3115), 1,
      anon_sym_DQUOTE,
    ACTIONS(3395), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(3397), 1,
      anon_sym_DOLLAR,
    ACTIONS(3401), 1,
      aux_sym_number_token1,
    ACTIONS(3403), 1,
      aux_sym_number_token2,
    ACTIONS(3405), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(3407), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3409), 1,
      anon_sym_BQUOTE,
    ACTIONS(3411), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3417), 1,
      sym__brace_start,
    ACTIONS(3665), 1,
      sym__special_character,
    ACTIONS(3667), 1,
      sym_test_operator,
    STATE(2161), 1,
      aux_sym__literal_repeat1,
    ACTIONS(2074), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(3393), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(3413), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(715), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    ACTIONS(3663), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(1918), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2072), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [18197] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3115), 1,
      anon_sym_DQUOTE,
    ACTIONS(3395), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(3397), 1,
      anon_sym_DOLLAR,
    ACTIONS(3401), 1,
      aux_sym_number_token1,
    ACTIONS(3403), 1,
      aux_sym_number_token2,
    ACTIONS(3405), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(3407), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3409), 1,
      anon_sym_BQUOTE,
    ACTIONS(3411), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3417), 1,
      sym__brace_start,
    ACTIONS(3665), 1,
      sym__special_character,
    ACTIONS(3667), 1,
      sym_test_operator,
    STATE(2161), 1,
      aux_sym__literal_repeat1,
    ACTIONS(2096), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(3393), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(3413), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(715), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    ACTIONS(3663), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(1918), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2094), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [18294] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3439), 1,
      anon_sym_DQUOTE,
    ACTIONS(3653), 1,
      sym_variable_name,
    STATE(1878), 1,
      sym_string,
    ACTIONS(3651), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1241), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
    ACTIONS(3649), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1239), 35,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [18365] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3673), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(3675), 1,
      anon_sym_DOLLAR,
    ACTIONS(3677), 1,
      sym__special_character,
    ACTIONS(3679), 1,
      anon_sym_DQUOTE,
    ACTIONS(3681), 1,
      aux_sym_number_token1,
    ACTIONS(3683), 1,
      aux_sym_number_token2,
    ACTIONS(3685), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(3687), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3689), 1,
      anon_sym_BQUOTE,
    ACTIONS(3691), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3695), 1,
      sym_test_operator,
    ACTIONS(3697), 1,
      sym__brace_start,
    STATE(4767), 1,
      aux_sym__literal_repeat1,
    STATE(5114), 1,
      sym_concatenation,
    ACTIONS(2500), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(3671), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(3693), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(3669), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(4504), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2498), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [18462] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3701), 1,
      anon_sym_DQUOTE,
    ACTIONS(3705), 1,
      sym_variable_name,
    STATE(2025), 1,
      sym_string,
    ACTIONS(3703), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1235), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
    ACTIONS(3699), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1227), 35,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [18533] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3673), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(3675), 1,
      anon_sym_DOLLAR,
    ACTIONS(3677), 1,
      sym__special_character,
    ACTIONS(3679), 1,
      anon_sym_DQUOTE,
    ACTIONS(3681), 1,
      aux_sym_number_token1,
    ACTIONS(3683), 1,
      aux_sym_number_token2,
    ACTIONS(3685), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(3687), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3689), 1,
      anon_sym_BQUOTE,
    ACTIONS(3691), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3697), 1,
      sym__brace_start,
    ACTIONS(3709), 1,
      sym_test_operator,
    STATE(4695), 1,
      aux_sym__literal_repeat1,
    STATE(5096), 1,
      sym_concatenation,
    ACTIONS(2496), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(3671), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(3693), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(3707), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(4555), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2494), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [18630] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3701), 1,
      anon_sym_DQUOTE,
    ACTIONS(3705), 1,
      sym_variable_name,
    STATE(2025), 1,
      sym_string,
    ACTIONS(3703), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1241), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
    ACTIONS(3699), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1239), 35,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [18701] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3183), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(3185), 1,
      anon_sym_DOLLAR,
    ACTIONS(3189), 1,
      anon_sym_DQUOTE,
    ACTIONS(3191), 1,
      aux_sym_number_token1,
    ACTIONS(3193), 1,
      aux_sym_number_token2,
    ACTIONS(3195), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(3197), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3201), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3209), 1,
      sym__brace_start,
    ACTIONS(3657), 1,
      sym__special_character,
    ACTIONS(3661), 1,
      sym_test_operator,
    ACTIONS(3711), 1,
      aux_sym__simple_variable_name_token1,
    STATE(1966), 1,
      aux_sym__literal_repeat1,
    ACTIONS(2564), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(3181), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(3203), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(731), 2,
      sym_concatenation,
      aux_sym_unset_command_repeat1,
    ACTIONS(3655), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(1902), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2562), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_BQUOTE,
  [18798] = 27,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(2628), 1,
      sym_word,
    ACTIONS(2634), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(2636), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(2638), 1,
      anon_sym_DOLLAR,
    ACTIONS(2640), 1,
      sym__special_character,
    ACTIONS(2642), 1,
      anon_sym_DQUOTE,
    ACTIONS(2646), 1,
      aux_sym_number_token1,
    ACTIONS(2648), 1,
      aux_sym_number_token2,
    ACTIONS(2650), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(2652), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2654), 1,
      anon_sym_BQUOTE,
    ACTIONS(2656), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(2660), 1,
      sym_test_operator,
    ACTIONS(2662), 1,
      sym__bare_dollar,
    ACTIONS(2664), 1,
      sym__brace_start,
    STATE(751), 1,
      aux_sym_command_repeat2,
    STATE(2417), 1,
      aux_sym__literal_repeat1,
    STATE(2558), 1,
      sym_concatenation,
    STATE(2562), 1,
      sym_herestring_redirect,
    ACTIONS(2630), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(2632), 2,
      anon_sym_EQ_EQ,
      anon_sym_EQ_TILDE,
    ACTIONS(2644), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(2658), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1625), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    STATE(2211), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(1627), 10,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [18907] = 30,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3725), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(3727), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(3729), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(3731), 1,
      anon_sym_DOLLAR,
    ACTIONS(3733), 1,
      sym__special_character,
    ACTIONS(3735), 1,
      anon_sym_DQUOTE,
    ACTIONS(3737), 1,
      aux_sym_number_token1,
    ACTIONS(3739), 1,
      aux_sym_number_token2,
    ACTIONS(3741), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(3743), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3745), 1,
      anon_sym_BQUOTE,
    ACTIONS(3747), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3751), 1,
      sym_file_descriptor,
    ACTIONS(3753), 1,
      sym_test_operator,
    ACTIONS(3755), 1,
      sym__brace_start,
    STATE(3570), 1,
      aux_sym__heredoc_command,
    STATE(5351), 1,
      aux_sym__literal_repeat1,
    STATE(5456), 1,
      sym_concatenation,
    STATE(7036), 1,
      sym__heredoc_expression,
    STATE(7039), 1,
      sym__heredoc_pipeline,
    ACTIONS(3715), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(3717), 2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
    ACTIONS(3719), 2,
      anon_sym_PIPE,
      anon_sym_PIPE_AMP,
    ACTIONS(3723), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(3749), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(3713), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(5505), 3,
      sym_file_redirect,
      sym_herestring_redirect,
      aux_sym_redirected_statement_repeat2,
    ACTIONS(3721), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
    STATE(5062), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [19022] = 30,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3727), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(3729), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(3731), 1,
      anon_sym_DOLLAR,
    ACTIONS(3733), 1,
      sym__special_character,
    ACTIONS(3735), 1,
      anon_sym_DQUOTE,
    ACTIONS(3737), 1,
      aux_sym_number_token1,
    ACTIONS(3739), 1,
      aux_sym_number_token2,
    ACTIONS(3741), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(3743), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3745), 1,
      anon_sym_BQUOTE,
    ACTIONS(3747), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3751), 1,
      sym_file_descriptor,
    ACTIONS(3753), 1,
      sym_test_operator,
    ACTIONS(3755), 1,
      sym__brace_start,
    ACTIONS(3757), 1,
      aux_sym_heredoc_redirect_token1,
    STATE(3605), 1,
      aux_sym__heredoc_command,
    STATE(5351), 1,
      aux_sym__literal_repeat1,
    STATE(5456), 1,
      sym_concatenation,
    STATE(7041), 1,
      sym__heredoc_expression,
    STATE(7042), 1,
      sym__heredoc_pipeline,
    ACTIONS(3715), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(3717), 2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
    ACTIONS(3719), 2,
      anon_sym_PIPE,
      anon_sym_PIPE_AMP,
    ACTIONS(3723), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(3749), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(3713), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(5504), 3,
      sym_file_redirect,
      sym_herestring_redirect,
      aux_sym_redirected_statement_repeat2,
    ACTIONS(3721), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
    STATE(5062), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [19137] = 35,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(237), 1,
      sym_word,
    ACTIONS(258), 1,
      anon_sym_DOLLAR,
    ACTIONS(264), 1,
      aux_sym_number_token1,
    ACTIONS(266), 1,
      aux_sym_number_token2,
    ACTIONS(270), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(282), 1,
      sym_test_operator,
    ACTIONS(284), 1,
      sym__brace_start,
    ACTIONS(1075), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1079), 1,
      sym__special_character,
    ACTIONS(1081), 1,
      anon_sym_DQUOTE,
    ACTIONS(1085), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1087), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3588), 1,
      anon_sym_LPAREN,
    ACTIONS(3590), 1,
      anon_sym_BANG,
    ACTIONS(3596), 1,
      anon_sym_TILDE,
    ACTIONS(3598), 1,
      anon_sym_BQUOTE,
    ACTIONS(3602), 1,
      sym_variable_name,
    ACTIONS(3759), 1,
      aux_sym__simple_variable_name_token1,
    STATE(2509), 1,
      sym_command_substitution,
    STATE(2690), 1,
      aux_sym__literal_repeat1,
    STATE(3107), 1,
      sym__expression,
    STATE(3359), 1,
      sym__arithmetic_postfix_expression,
    STATE(3361), 1,
      sym__arithmetic_unary_expression,
    STATE(3367), 1,
      sym__arithmetic_ternary_expression,
    STATE(3375), 1,
      sym__arithmetic_binary_expression,
    ACTIONS(1073), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1083), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(1089), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(3592), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(3594), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(2500), 4,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_translated_string,
      sym_process_substitution,
    STATE(2506), 4,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
    STATE(3278), 4,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
    STATE(3053), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
  [19262] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3536), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(3539), 1,
      anon_sym_DOLLAR,
    ACTIONS(3545), 1,
      anon_sym_DQUOTE,
    ACTIONS(3548), 1,
      aux_sym_number_token1,
    ACTIONS(3551), 1,
      aux_sym_number_token2,
    ACTIONS(3554), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(3557), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3560), 1,
      anon_sym_BQUOTE,
    ACTIONS(3563), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3575), 1,
      sym__brace_start,
    ACTIONS(3764), 1,
      sym__special_character,
    ACTIONS(3767), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(3770), 1,
      sym_test_operator,
    STATE(1966), 1,
      aux_sym__literal_repeat1,
    ACTIONS(2442), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(3533), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(3566), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(731), 2,
      sym_concatenation,
      aux_sym_unset_command_repeat1,
    ACTIONS(3761), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(1902), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2440), 19,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [19361] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3775), 1,
      anon_sym_DQUOTE,
    ACTIONS(3779), 1,
      sym_variable_name,
    STATE(2093), 1,
      sym_string,
    ACTIONS(3777), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1235), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
    ACTIONS(3773), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1227), 35,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [19432] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3785), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(3787), 1,
      anon_sym_DOLLAR,
    ACTIONS(3789), 1,
      sym__special_character,
    ACTIONS(3791), 1,
      anon_sym_DQUOTE,
    ACTIONS(3793), 1,
      aux_sym_number_token1,
    ACTIONS(3795), 1,
      aux_sym_number_token2,
    ACTIONS(3797), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(3799), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3801), 1,
      anon_sym_BQUOTE,
    ACTIONS(3803), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3807), 1,
      sym_test_operator,
    ACTIONS(3809), 1,
      sym__brace_start,
    STATE(4277), 1,
      aux_sym__literal_repeat1,
    STATE(4782), 1,
      sym_concatenation,
    ACTIONS(2496), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(3783), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(3805), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(3781), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(4522), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2494), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [19529] = 35,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(237), 1,
      sym_word,
    ACTIONS(258), 1,
      anon_sym_DOLLAR,
    ACTIONS(264), 1,
      aux_sym_number_token1,
    ACTIONS(266), 1,
      aux_sym_number_token2,
    ACTIONS(270), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(282), 1,
      sym_test_operator,
    ACTIONS(284), 1,
      sym__brace_start,
    ACTIONS(1075), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1079), 1,
      sym__special_character,
    ACTIONS(1081), 1,
      anon_sym_DQUOTE,
    ACTIONS(1085), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1087), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3588), 1,
      anon_sym_LPAREN,
    ACTIONS(3590), 1,
      anon_sym_BANG,
    ACTIONS(3596), 1,
      anon_sym_TILDE,
    ACTIONS(3598), 1,
      anon_sym_BQUOTE,
    ACTIONS(3602), 1,
      sym_variable_name,
    ACTIONS(3811), 1,
      aux_sym__simple_variable_name_token1,
    STATE(2509), 1,
      sym_command_substitution,
    STATE(2690), 1,
      aux_sym__literal_repeat1,
    STATE(3126), 1,
      sym__expression,
    STATE(3359), 1,
      sym__arithmetic_postfix_expression,
    STATE(3361), 1,
      sym__arithmetic_unary_expression,
    STATE(3367), 1,
      sym__arithmetic_ternary_expression,
    STATE(3375), 1,
      sym__arithmetic_binary_expression,
    ACTIONS(1073), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1083), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(1089), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(3592), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(3594), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(2500), 4,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_translated_string,
      sym_process_substitution,
    STATE(2506), 4,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
    STATE(3365), 4,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
    STATE(3053), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
  [19654] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2042), 1,
      anon_sym_DQUOTE,
    ACTIONS(3647), 1,
      sym_variable_name,
    STATE(1547), 1,
      sym_string,
    ACTIONS(3645), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1241), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
    ACTIONS(3643), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1239), 35,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [19725] = 30,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3727), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(3729), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(3731), 1,
      anon_sym_DOLLAR,
    ACTIONS(3733), 1,
      sym__special_character,
    ACTIONS(3735), 1,
      anon_sym_DQUOTE,
    ACTIONS(3737), 1,
      aux_sym_number_token1,
    ACTIONS(3739), 1,
      aux_sym_number_token2,
    ACTIONS(3741), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(3743), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3745), 1,
      anon_sym_BQUOTE,
    ACTIONS(3747), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3751), 1,
      sym_file_descriptor,
    ACTIONS(3753), 1,
      sym_test_operator,
    ACTIONS(3755), 1,
      sym__brace_start,
    ACTIONS(3813), 1,
      aux_sym_heredoc_redirect_token1,
    STATE(3603), 1,
      aux_sym__heredoc_command,
    STATE(5351), 1,
      aux_sym__literal_repeat1,
    STATE(5456), 1,
      sym_concatenation,
    STATE(6792), 1,
      sym__heredoc_expression,
    STATE(6806), 1,
      sym__heredoc_pipeline,
    ACTIONS(3715), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(3717), 2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
    ACTIONS(3719), 2,
      anon_sym_PIPE,
      anon_sym_PIPE_AMP,
    ACTIONS(3723), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(3749), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(3713), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(5531), 3,
      sym_file_redirect,
      sym_herestring_redirect,
      aux_sym_redirected_statement_repeat2,
    ACTIONS(3721), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
    STATE(5062), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [19840] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3775), 1,
      anon_sym_DQUOTE,
    ACTIONS(3779), 1,
      sym_variable_name,
    STATE(2093), 1,
      sym_string,
    ACTIONS(3777), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1241), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
    ACTIONS(3773), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1239), 35,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [19911] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3785), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(3787), 1,
      anon_sym_DOLLAR,
    ACTIONS(3791), 1,
      anon_sym_DQUOTE,
    ACTIONS(3793), 1,
      aux_sym_number_token1,
    ACTIONS(3795), 1,
      aux_sym_number_token2,
    ACTIONS(3797), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(3799), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3801), 1,
      anon_sym_BQUOTE,
    ACTIONS(3803), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3809), 1,
      sym__brace_start,
    ACTIONS(3817), 1,
      sym__special_character,
    ACTIONS(3819), 1,
      sym_test_operator,
    STATE(4305), 1,
      aux_sym__literal_repeat1,
    STATE(4736), 1,
      sym_concatenation,
    ACTIONS(3783), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(3805), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(2500), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(3815), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(4518), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2498), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [20008] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3785), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(3787), 1,
      anon_sym_DOLLAR,
    ACTIONS(3791), 1,
      anon_sym_DQUOTE,
    ACTIONS(3793), 1,
      aux_sym_number_token1,
    ACTIONS(3795), 1,
      aux_sym_number_token2,
    ACTIONS(3797), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(3799), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3801), 1,
      anon_sym_BQUOTE,
    ACTIONS(3803), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3809), 1,
      sym__brace_start,
    ACTIONS(3817), 1,
      sym__special_character,
    ACTIONS(3823), 1,
      sym_test_operator,
    STATE(4277), 1,
      aux_sym__literal_repeat1,
    STATE(4782), 1,
      sym_concatenation,
    ACTIONS(3783), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(3805), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(2496), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(3821), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(4486), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2494), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [20105] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2267), 1,
      anon_sym_DQUOTE,
    ACTIONS(3829), 1,
      sym_variable_name,
    STATE(1595), 1,
      sym_string,
    ACTIONS(3827), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1235), 3,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(3825), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1227), 36,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [20176] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3837), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(3840), 1,
      anon_sym_DOLLAR,
    ACTIONS(3843), 1,
      sym__special_character,
    ACTIONS(3846), 1,
      anon_sym_DQUOTE,
    ACTIONS(3849), 1,
      aux_sym_number_token1,
    ACTIONS(3852), 1,
      aux_sym_number_token2,
    ACTIONS(3855), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(3858), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3861), 1,
      anon_sym_BQUOTE,
    ACTIONS(3864), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3870), 1,
      sym_test_operator,
    ACTIONS(3873), 1,
      sym__brace_start,
    STATE(2270), 1,
      aux_sym__literal_repeat1,
    ACTIONS(2216), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(3834), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(3867), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(741), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    ACTIONS(3831), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(1839), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2214), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [20273] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3701), 1,
      anon_sym_DQUOTE,
    ACTIONS(3880), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(3882), 1,
      anon_sym_DOLLAR,
    ACTIONS(3884), 1,
      sym__special_character,
    ACTIONS(3886), 1,
      aux_sym_number_token1,
    ACTIONS(3888), 1,
      aux_sym_number_token2,
    ACTIONS(3890), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(3892), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3894), 1,
      anon_sym_BQUOTE,
    ACTIONS(3896), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3900), 1,
      sym_test_operator,
    ACTIONS(3902), 1,
      sym__brace_start,
    STATE(2281), 1,
      aux_sym__literal_repeat1,
    ACTIONS(3878), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(3898), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(747), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    ACTIONS(2096), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(3876), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(1846), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2094), 19,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [20370] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3183), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(3185), 1,
      anon_sym_DOLLAR,
    ACTIONS(3189), 1,
      anon_sym_DQUOTE,
    ACTIONS(3191), 1,
      aux_sym_number_token1,
    ACTIONS(3193), 1,
      aux_sym_number_token2,
    ACTIONS(3195), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(3197), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3199), 1,
      anon_sym_BQUOTE,
    ACTIONS(3201), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3209), 1,
      sym__brace_start,
    ACTIONS(3657), 1,
      sym__special_character,
    ACTIONS(3661), 1,
      sym_test_operator,
    ACTIONS(3711), 1,
      aux_sym__simple_variable_name_token1,
    STATE(1966), 1,
      aux_sym__literal_repeat1,
    ACTIONS(2564), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(3181), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(3203), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(731), 2,
      sym_concatenation,
      aux_sym_unset_command_repeat1,
    ACTIONS(3655), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(1902), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2562), 19,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [20469] = 30,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3727), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(3729), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(3731), 1,
      anon_sym_DOLLAR,
    ACTIONS(3733), 1,
      sym__special_character,
    ACTIONS(3735), 1,
      anon_sym_DQUOTE,
    ACTIONS(3737), 1,
      aux_sym_number_token1,
    ACTIONS(3739), 1,
      aux_sym_number_token2,
    ACTIONS(3741), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(3743), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3745), 1,
      anon_sym_BQUOTE,
    ACTIONS(3747), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3751), 1,
      sym_file_descriptor,
    ACTIONS(3753), 1,
      sym_test_operator,
    ACTIONS(3755), 1,
      sym__brace_start,
    ACTIONS(3904), 1,
      aux_sym_heredoc_redirect_token1,
    STATE(3583), 1,
      aux_sym__heredoc_command,
    STATE(5351), 1,
      aux_sym__literal_repeat1,
    STATE(5456), 1,
      sym_concatenation,
    STATE(6939), 1,
      sym__heredoc_expression,
    STATE(6941), 1,
      sym__heredoc_pipeline,
    ACTIONS(3715), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(3717), 2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
    ACTIONS(3719), 2,
      anon_sym_PIPE,
      anon_sym_PIPE_AMP,
    ACTIONS(3723), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(3749), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(3713), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(5515), 3,
      sym_file_redirect,
      sym_herestring_redirect,
      aux_sym_redirected_statement_repeat2,
    ACTIONS(3721), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
    STATE(5062), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [20584] = 30,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3727), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(3729), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(3731), 1,
      anon_sym_DOLLAR,
    ACTIONS(3733), 1,
      sym__special_character,
    ACTIONS(3735), 1,
      anon_sym_DQUOTE,
    ACTIONS(3737), 1,
      aux_sym_number_token1,
    ACTIONS(3739), 1,
      aux_sym_number_token2,
    ACTIONS(3741), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(3743), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3745), 1,
      anon_sym_BQUOTE,
    ACTIONS(3747), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3751), 1,
      sym_file_descriptor,
    ACTIONS(3753), 1,
      sym_test_operator,
    ACTIONS(3755), 1,
      sym__brace_start,
    ACTIONS(3906), 1,
      aux_sym_heredoc_redirect_token1,
    STATE(3585), 1,
      aux_sym__heredoc_command,
    STATE(5351), 1,
      aux_sym__literal_repeat1,
    STATE(5456), 1,
      sym_concatenation,
    STATE(6935), 1,
      sym__heredoc_expression,
    STATE(6936), 1,
      sym__heredoc_pipeline,
    ACTIONS(3715), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(3717), 2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
    ACTIONS(3719), 2,
      anon_sym_PIPE,
      anon_sym_PIPE_AMP,
    ACTIONS(3723), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(3749), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(3713), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(5516), 3,
      sym_file_redirect,
      sym_herestring_redirect,
      aux_sym_redirected_statement_repeat2,
    ACTIONS(3721), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
    STATE(5062), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [20699] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2163), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(2166), 1,
      anon_sym_DOLLAR,
    ACTIONS(2172), 1,
      anon_sym_DQUOTE,
    ACTIONS(2175), 1,
      aux_sym_number_token1,
    ACTIONS(2178), 1,
      aux_sym_number_token2,
    ACTIONS(2181), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(2184), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2187), 1,
      anon_sym_BQUOTE,
    ACTIONS(2190), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(2205), 1,
      sym__brace_start,
    ACTIONS(3911), 1,
      sym__special_character,
    ACTIONS(3914), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(3917), 1,
      sym_variable_name,
    ACTIONS(3920), 1,
      sym_test_operator,
    STATE(1743), 1,
      aux_sym__literal_repeat1,
    STATE(6739), 1,
      sym_subscript,
    ACTIONS(1704), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(2160), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(2193), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(3908), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(746), 3,
      sym_variable_assignment,
      sym_concatenation,
      aux_sym_declaration_command_repeat1,
    STATE(2330), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(1702), 16,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [20802] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3929), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(3932), 1,
      anon_sym_DOLLAR,
    ACTIONS(3935), 1,
      sym__special_character,
    ACTIONS(3938), 1,
      anon_sym_DQUOTE,
    ACTIONS(3941), 1,
      aux_sym_number_token1,
    ACTIONS(3944), 1,
      aux_sym_number_token2,
    ACTIONS(3947), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(3950), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3953), 1,
      anon_sym_BQUOTE,
    ACTIONS(3956), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3962), 1,
      sym_test_operator,
    ACTIONS(3965), 1,
      sym__brace_start,
    STATE(2281), 1,
      aux_sym__literal_repeat1,
    ACTIONS(3926), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(3959), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(747), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    ACTIONS(2216), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(3923), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(1846), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2214), 19,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [20899] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2261), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(2263), 1,
      anon_sym_DOLLAR,
    ACTIONS(2267), 1,
      anon_sym_DQUOTE,
    ACTIONS(2269), 1,
      aux_sym_number_token1,
    ACTIONS(2271), 1,
      aux_sym_number_token2,
    ACTIONS(2273), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(2275), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2277), 1,
      anon_sym_BQUOTE,
    ACTIONS(2279), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(2289), 1,
      sym__brace_start,
    ACTIONS(3970), 1,
      sym__special_character,
    ACTIONS(3972), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(3974), 1,
      sym_variable_name,
    ACTIONS(3976), 1,
      sym_test_operator,
    STATE(1743), 1,
      aux_sym__literal_repeat1,
    STATE(6739), 1,
      sym_subscript,
    ACTIONS(1757), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(2259), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(2281), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(3968), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(777), 3,
      sym_variable_assignment,
      sym_concatenation,
      aux_sym_declaration_command_repeat1,
    STATE(2330), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(1755), 16,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [21002] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3115), 1,
      anon_sym_DQUOTE,
    ACTIONS(3119), 1,
      sym_variable_name,
    STATE(1864), 1,
      sym_string,
    ACTIONS(3117), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1235), 3,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(3113), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1227), 36,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [21073] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3115), 1,
      anon_sym_DQUOTE,
    ACTIONS(3119), 1,
      sym_variable_name,
    STATE(1864), 1,
      sym_string,
    ACTIONS(3117), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1241), 3,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(3113), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1239), 36,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [21144] = 28,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3978), 1,
      sym_word,
    ACTIONS(3987), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(3990), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(3993), 1,
      anon_sym_DOLLAR,
    ACTIONS(3996), 1,
      sym__special_character,
    ACTIONS(3999), 1,
      anon_sym_DQUOTE,
    ACTIONS(4005), 1,
      aux_sym_number_token1,
    ACTIONS(4008), 1,
      aux_sym_number_token2,
    ACTIONS(4011), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(4014), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(4017), 1,
      anon_sym_BQUOTE,
    ACTIONS(4020), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(4026), 1,
      sym_file_descriptor,
    ACTIONS(4029), 1,
      sym_test_operator,
    ACTIONS(4032), 1,
      sym__bare_dollar,
    ACTIONS(4035), 1,
      sym__brace_start,
    STATE(751), 1,
      aux_sym_command_repeat2,
    STATE(2417), 1,
      aux_sym__literal_repeat1,
    STATE(2558), 1,
      sym_concatenation,
    STATE(2562), 1,
      sym_herestring_redirect,
    ACTIONS(3981), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(3984), 2,
      anon_sym_EQ_EQ,
      anon_sym_EQ_TILDE,
    ACTIONS(4002), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(4023), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1486), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1491), 9,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
    STATE(2211), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [21255] = 27,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(2628), 1,
      sym_word,
    ACTIONS(2634), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(2636), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(2638), 1,
      anon_sym_DOLLAR,
    ACTIONS(2640), 1,
      sym__special_character,
    ACTIONS(2642), 1,
      anon_sym_DQUOTE,
    ACTIONS(2646), 1,
      aux_sym_number_token1,
    ACTIONS(2648), 1,
      aux_sym_number_token2,
    ACTIONS(2650), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(2652), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2654), 1,
      anon_sym_BQUOTE,
    ACTIONS(2656), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(2660), 1,
      sym_test_operator,
    ACTIONS(2662), 1,
      sym__bare_dollar,
    ACTIONS(2664), 1,
      sym__brace_start,
    STATE(751), 1,
      aux_sym_command_repeat2,
    STATE(2417), 1,
      aux_sym__literal_repeat1,
    STATE(2558), 1,
      sym_concatenation,
    STATE(2562), 1,
      sym_herestring_redirect,
    ACTIONS(2630), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(2632), 2,
      anon_sym_EQ_EQ,
      anon_sym_EQ_TILDE,
    ACTIONS(2644), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(2658), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1613), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    STATE(2211), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(1615), 10,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [21364] = 30,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3727), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(3729), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(3731), 1,
      anon_sym_DOLLAR,
    ACTIONS(3733), 1,
      sym__special_character,
    ACTIONS(3735), 1,
      anon_sym_DQUOTE,
    ACTIONS(3737), 1,
      aux_sym_number_token1,
    ACTIONS(3739), 1,
      aux_sym_number_token2,
    ACTIONS(3741), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(3743), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3745), 1,
      anon_sym_BQUOTE,
    ACTIONS(3747), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3751), 1,
      sym_file_descriptor,
    ACTIONS(3753), 1,
      sym_test_operator,
    ACTIONS(3755), 1,
      sym__brace_start,
    ACTIONS(4038), 1,
      aux_sym_heredoc_redirect_token1,
    STATE(3565), 1,
      aux_sym__heredoc_command,
    STATE(5351), 1,
      aux_sym__literal_repeat1,
    STATE(5456), 1,
      sym_concatenation,
    STATE(6837), 1,
      sym__heredoc_expression,
    STATE(6839), 1,
      sym__heredoc_pipeline,
    ACTIONS(3715), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(3717), 2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
    ACTIONS(3719), 2,
      anon_sym_PIPE,
      anon_sym_PIPE_AMP,
    ACTIONS(3723), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(3749), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(3713), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(5545), 3,
      sym_file_redirect,
      sym_herestring_redirect,
      aux_sym_redirected_statement_repeat2,
    ACTIONS(3721), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
    STATE(5062), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [21479] = 20,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3115), 1,
      anon_sym_DQUOTE,
    ACTIONS(3395), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(3397), 1,
      anon_sym_DOLLAR,
    ACTIONS(3401), 1,
      aux_sym_number_token1,
    ACTIONS(3403), 1,
      aux_sym_number_token2,
    ACTIONS(3405), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(3407), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3411), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3417), 1,
      sym__brace_start,
    ACTIONS(3665), 1,
      sym__special_character,
    ACTIONS(3667), 1,
      sym_test_operator,
    STATE(2161), 1,
      aux_sym__literal_repeat1,
    ACTIONS(2074), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(3393), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(3413), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(715), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    ACTIONS(3663), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(1918), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2072), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_BQUOTE,
  [21574] = 20,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3115), 1,
      anon_sym_DQUOTE,
    ACTIONS(3395), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(3397), 1,
      anon_sym_DOLLAR,
    ACTIONS(3401), 1,
      aux_sym_number_token1,
    ACTIONS(3403), 1,
      aux_sym_number_token2,
    ACTIONS(3405), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(3407), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3411), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3417), 1,
      sym__brace_start,
    ACTIONS(3665), 1,
      sym__special_character,
    ACTIONS(3667), 1,
      sym_test_operator,
    STATE(2161), 1,
      aux_sym__literal_repeat1,
    ACTIONS(2096), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(3393), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(3413), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(715), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    ACTIONS(3663), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(1918), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2094), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_BQUOTE,
  [21669] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2267), 1,
      anon_sym_DQUOTE,
    ACTIONS(3829), 1,
      sym_variable_name,
    STATE(1595), 1,
      sym_string,
    ACTIONS(3827), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1241), 3,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(3825), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1239), 36,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [21740] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3701), 1,
      anon_sym_DQUOTE,
    ACTIONS(3880), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(3882), 1,
      anon_sym_DOLLAR,
    ACTIONS(3884), 1,
      sym__special_character,
    ACTIONS(3886), 1,
      aux_sym_number_token1,
    ACTIONS(3888), 1,
      aux_sym_number_token2,
    ACTIONS(3890), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(3892), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3894), 1,
      anon_sym_BQUOTE,
    ACTIONS(3896), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3900), 1,
      sym_test_operator,
    ACTIONS(3902), 1,
      sym__brace_start,
    STATE(2281), 1,
      aux_sym__literal_repeat1,
    ACTIONS(3878), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(3898), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(747), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    ACTIONS(2074), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(3876), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(1846), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2072), 19,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [21837] = 30,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3727), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(3729), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(3731), 1,
      anon_sym_DOLLAR,
    ACTIONS(3733), 1,
      sym__special_character,
    ACTIONS(3735), 1,
      anon_sym_DQUOTE,
    ACTIONS(3737), 1,
      aux_sym_number_token1,
    ACTIONS(3739), 1,
      aux_sym_number_token2,
    ACTIONS(3741), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(3743), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3745), 1,
      anon_sym_BQUOTE,
    ACTIONS(3747), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3751), 1,
      sym_file_descriptor,
    ACTIONS(3753), 1,
      sym_test_operator,
    ACTIONS(3755), 1,
      sym__brace_start,
    ACTIONS(4040), 1,
      aux_sym_heredoc_redirect_token1,
    STATE(3567), 1,
      aux_sym__heredoc_command,
    STATE(5351), 1,
      aux_sym__literal_repeat1,
    STATE(5456), 1,
      sym_concatenation,
    STATE(6843), 1,
      sym__heredoc_expression,
    STATE(6844), 1,
      sym__heredoc_pipeline,
    ACTIONS(3715), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(3717), 2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
    ACTIONS(3719), 2,
      anon_sym_PIPE,
      anon_sym_PIPE_AMP,
    ACTIONS(3723), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(3749), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(3713), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(5546), 3,
      sym_file_redirect,
      sym_herestring_redirect,
      aux_sym_redirected_statement_repeat2,
    ACTIONS(3721), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
    STATE(5062), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [21952] = 6,
    ACTIONS(3), 1,
      sym_comment,
    STATE(1739), 1,
      aux_sym__literal_repeat1,
    STATE(1953), 1,
      sym_concatenation,
    ACTIONS(2496), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    STATE(1601), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2494), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [22019] = 35,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(237), 1,
      sym_word,
    ACTIONS(258), 1,
      anon_sym_DOLLAR,
    ACTIONS(264), 1,
      aux_sym_number_token1,
    ACTIONS(266), 1,
      aux_sym_number_token2,
    ACTIONS(270), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(282), 1,
      sym_test_operator,
    ACTIONS(284), 1,
      sym__brace_start,
    ACTIONS(1075), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1079), 1,
      sym__special_character,
    ACTIONS(1081), 1,
      anon_sym_DQUOTE,
    ACTIONS(1085), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1087), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3588), 1,
      anon_sym_LPAREN,
    ACTIONS(3590), 1,
      anon_sym_BANG,
    ACTIONS(3596), 1,
      anon_sym_TILDE,
    ACTIONS(3598), 1,
      anon_sym_BQUOTE,
    ACTIONS(3602), 1,
      sym_variable_name,
    ACTIONS(4042), 1,
      aux_sym__simple_variable_name_token1,
    STATE(2509), 1,
      sym_command_substitution,
    STATE(2690), 1,
      aux_sym__literal_repeat1,
    STATE(3111), 1,
      sym__expression,
    STATE(3359), 1,
      sym__arithmetic_postfix_expression,
    STATE(3361), 1,
      sym__arithmetic_unary_expression,
    STATE(3367), 1,
      sym__arithmetic_ternary_expression,
    STATE(3375), 1,
      sym__arithmetic_binary_expression,
    ACTIONS(1073), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1083), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(1089), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(3592), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(3594), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(2500), 4,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_translated_string,
      sym_process_substitution,
    STATE(2506), 4,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
    STATE(3368), 4,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
    STATE(3053), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
  [22144] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3183), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(3185), 1,
      anon_sym_DOLLAR,
    ACTIONS(3189), 1,
      anon_sym_DQUOTE,
    ACTIONS(3191), 1,
      aux_sym_number_token1,
    ACTIONS(3193), 1,
      aux_sym_number_token2,
    ACTIONS(3195), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(3197), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3201), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3209), 1,
      sym__brace_start,
    ACTIONS(3657), 1,
      sym__special_character,
    ACTIONS(3661), 1,
      sym_test_operator,
    ACTIONS(4044), 1,
      aux_sym__simple_variable_name_token1,
    STATE(1966), 1,
      aux_sym__literal_repeat1,
    ACTIONS(2299), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(3181), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(3203), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(726), 2,
      sym_concatenation,
      aux_sym_unset_command_repeat1,
    ACTIONS(3655), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(1902), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2297), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_BQUOTE,
  [22241] = 35,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1141), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1145), 1,
      anon_sym_DOLLAR,
    ACTIONS(1149), 1,
      anon_sym_DQUOTE,
    ACTIONS(1153), 1,
      aux_sym_number_token1,
    ACTIONS(1155), 1,
      aux_sym_number_token2,
    ACTIONS(1157), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1159), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1163), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1169), 1,
      sym__brace_start,
    ACTIONS(1171), 1,
      sym_word,
    ACTIONS(1181), 1,
      sym__special_character,
    ACTIONS(1185), 1,
      sym_test_operator,
    ACTIONS(3050), 1,
      anon_sym_LPAREN,
    ACTIONS(3052), 1,
      anon_sym_BANG,
    ACTIONS(3058), 1,
      anon_sym_TILDE,
    ACTIONS(3060), 1,
      anon_sym_BQUOTE,
    ACTIONS(3064), 1,
      sym_variable_name,
    ACTIONS(4046), 1,
      aux_sym__simple_variable_name_token1,
    STATE(2387), 1,
      sym_command_substitution,
    STATE(2484), 1,
      aux_sym__literal_repeat1,
    STATE(2621), 1,
      sym__expression,
    STATE(2801), 1,
      sym__arithmetic_binary_expression,
    STATE(2803), 1,
      sym__arithmetic_ternary_expression,
    STATE(2809), 1,
      sym__arithmetic_unary_expression,
    STATE(2825), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(1129), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1165), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1183), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(3054), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(3056), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(2397), 4,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
    STATE(2456), 4,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_translated_string,
      sym_process_substitution,
    STATE(2843), 4,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
    STATE(2594), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
  [22366] = 35,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1141), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1145), 1,
      anon_sym_DOLLAR,
    ACTIONS(1149), 1,
      anon_sym_DQUOTE,
    ACTIONS(1153), 1,
      aux_sym_number_token1,
    ACTIONS(1155), 1,
      aux_sym_number_token2,
    ACTIONS(1157), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1159), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1163), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1169), 1,
      sym__brace_start,
    ACTIONS(1171), 1,
      sym_word,
    ACTIONS(1181), 1,
      sym__special_character,
    ACTIONS(1185), 1,
      sym_test_operator,
    ACTIONS(3050), 1,
      anon_sym_LPAREN,
    ACTIONS(3052), 1,
      anon_sym_BANG,
    ACTIONS(3058), 1,
      anon_sym_TILDE,
    ACTIONS(3060), 1,
      anon_sym_BQUOTE,
    ACTIONS(3064), 1,
      sym_variable_name,
    ACTIONS(4048), 1,
      aux_sym__simple_variable_name_token1,
    STATE(2387), 1,
      sym_command_substitution,
    STATE(2484), 1,
      aux_sym__literal_repeat1,
    STATE(2801), 1,
      sym__arithmetic_binary_expression,
    STATE(2803), 1,
      sym__arithmetic_ternary_expression,
    STATE(2809), 1,
      sym__arithmetic_unary_expression,
    STATE(2825), 1,
      sym__arithmetic_postfix_expression,
    STATE(3101), 1,
      sym__expression,
    ACTIONS(1129), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1165), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1183), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(3054), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(3056), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(2397), 4,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
    STATE(2456), 4,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_translated_string,
      sym_process_substitution,
    STATE(2848), 4,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
    STATE(2594), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
  [22491] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3785), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(3787), 1,
      anon_sym_DOLLAR,
    ACTIONS(3789), 1,
      sym__special_character,
    ACTIONS(3791), 1,
      anon_sym_DQUOTE,
    ACTIONS(3793), 1,
      aux_sym_number_token1,
    ACTIONS(3795), 1,
      aux_sym_number_token2,
    ACTIONS(3797), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(3799), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3801), 1,
      anon_sym_BQUOTE,
    ACTIONS(3803), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3809), 1,
      sym__brace_start,
    ACTIONS(4052), 1,
      sym_test_operator,
    STATE(4305), 1,
      aux_sym__literal_repeat1,
    STATE(4736), 1,
      sym_concatenation,
    ACTIONS(2500), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(3783), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(3805), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(4050), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(4582), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2498), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [22588] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3189), 1,
      anon_sym_DQUOTE,
    ACTIONS(4058), 1,
      sym_variable_name,
    STATE(1737), 1,
      sym_string,
    ACTIONS(4056), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1241), 3,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(4054), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1239), 36,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [22659] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3189), 1,
      anon_sym_DQUOTE,
    ACTIONS(4058), 1,
      sym_variable_name,
    STATE(1737), 1,
      sym_string,
    ACTIONS(4056), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1235), 3,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(4054), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1227), 36,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [22730] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1555), 1,
      anon_sym_DQUOTE,
    ACTIONS(2070), 1,
      sym_variable_name,
    STATE(1176), 1,
      sym_string,
    ACTIONS(2068), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1235), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
    ACTIONS(2066), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1227), 35,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [22801] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1555), 1,
      anon_sym_DQUOTE,
    ACTIONS(2070), 1,
      sym_variable_name,
    STATE(1176), 1,
      sym_string,
    ACTIONS(2068), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1241), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
    ACTIONS(2066), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1239), 35,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [22872] = 30,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3727), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(3729), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(3731), 1,
      anon_sym_DOLLAR,
    ACTIONS(3733), 1,
      sym__special_character,
    ACTIONS(3735), 1,
      anon_sym_DQUOTE,
    ACTIONS(3737), 1,
      aux_sym_number_token1,
    ACTIONS(3739), 1,
      aux_sym_number_token2,
    ACTIONS(3741), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(3743), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3745), 1,
      anon_sym_BQUOTE,
    ACTIONS(3747), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3751), 1,
      sym_file_descriptor,
    ACTIONS(3753), 1,
      sym_test_operator,
    ACTIONS(3755), 1,
      sym__brace_start,
    ACTIONS(4060), 1,
      aux_sym_heredoc_redirect_token1,
    STATE(3596), 1,
      aux_sym__heredoc_command,
    STATE(5351), 1,
      aux_sym__literal_repeat1,
    STATE(5456), 1,
      sym_concatenation,
    STATE(6944), 1,
      sym__heredoc_expression,
    STATE(6945), 1,
      sym__heredoc_pipeline,
    ACTIONS(3715), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(3717), 2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
    ACTIONS(3719), 2,
      anon_sym_PIPE,
      anon_sym_PIPE_AMP,
    ACTIONS(3723), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(3749), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(3713), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(5517), 3,
      sym_file_redirect,
      sym_herestring_redirect,
      aux_sym_redirected_statement_repeat2,
    ACTIONS(3721), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
    STATE(5062), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [22987] = 30,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3727), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(3729), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(3731), 1,
      anon_sym_DOLLAR,
    ACTIONS(3733), 1,
      sym__special_character,
    ACTIONS(3735), 1,
      anon_sym_DQUOTE,
    ACTIONS(3737), 1,
      aux_sym_number_token1,
    ACTIONS(3739), 1,
      aux_sym_number_token2,
    ACTIONS(3741), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(3743), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3745), 1,
      anon_sym_BQUOTE,
    ACTIONS(3747), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3751), 1,
      sym_file_descriptor,
    ACTIONS(3753), 1,
      sym_test_operator,
    ACTIONS(3755), 1,
      sym__brace_start,
    ACTIONS(4062), 1,
      aux_sym_heredoc_redirect_token1,
    STATE(3608), 1,
      aux_sym__heredoc_command,
    STATE(5351), 1,
      aux_sym__literal_repeat1,
    STATE(5456), 1,
      sym_concatenation,
    STATE(6870), 1,
      sym__heredoc_expression,
    STATE(6871), 1,
      sym__heredoc_pipeline,
    ACTIONS(3715), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(3717), 2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
    ACTIONS(3719), 2,
      anon_sym_PIPE,
      anon_sym_PIPE_AMP,
    ACTIONS(3723), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(3749), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(3713), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(5540), 3,
      sym_file_redirect,
      sym_herestring_redirect,
      aux_sym_redirected_statement_repeat2,
    ACTIONS(3721), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
    STATE(5062), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [23102] = 30,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3727), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(3729), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(3731), 1,
      anon_sym_DOLLAR,
    ACTIONS(3733), 1,
      sym__special_character,
    ACTIONS(3735), 1,
      anon_sym_DQUOTE,
    ACTIONS(3737), 1,
      aux_sym_number_token1,
    ACTIONS(3739), 1,
      aux_sym_number_token2,
    ACTIONS(3741), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(3743), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3745), 1,
      anon_sym_BQUOTE,
    ACTIONS(3747), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3751), 1,
      sym_file_descriptor,
    ACTIONS(3753), 1,
      sym_test_operator,
    ACTIONS(3755), 1,
      sym__brace_start,
    ACTIONS(4064), 1,
      aux_sym_heredoc_redirect_token1,
    STATE(3591), 1,
      aux_sym__heredoc_command,
    STATE(5351), 1,
      aux_sym__literal_repeat1,
    STATE(5456), 1,
      sym_concatenation,
    STATE(6873), 1,
      sym__heredoc_expression,
    STATE(6875), 1,
      sym__heredoc_pipeline,
    ACTIONS(3715), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(3717), 2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
    ACTIONS(3719), 2,
      anon_sym_PIPE,
      anon_sym_PIPE_AMP,
    ACTIONS(3723), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(3749), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(3713), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(5538), 3,
      sym_file_redirect,
      sym_herestring_redirect,
      aux_sym_redirected_statement_repeat2,
    ACTIONS(3721), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
    STATE(5062), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [23217] = 6,
    ACTIONS(3), 1,
      sym_comment,
    STATE(1740), 1,
      aux_sym__literal_repeat1,
    STATE(1957), 1,
      sym_concatenation,
    ACTIONS(2500), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    STATE(1560), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2498), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [23284] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3580), 1,
      anon_sym_DQUOTE,
    ACTIONS(3610), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(3612), 1,
      anon_sym_DOLLAR,
    ACTIONS(3614), 1,
      sym__special_character,
    ACTIONS(3616), 1,
      aux_sym_number_token1,
    ACTIONS(3618), 1,
      aux_sym_number_token2,
    ACTIONS(3620), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(3622), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3624), 1,
      anon_sym_BQUOTE,
    ACTIONS(3626), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3630), 1,
      sym_test_operator,
    ACTIONS(3632), 1,
      sym__brace_start,
    STATE(2270), 1,
      aux_sym__literal_repeat1,
    ACTIONS(2096), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(3608), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(3628), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(741), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    ACTIONS(3606), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(1839), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2094), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [23381] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3580), 1,
      anon_sym_DQUOTE,
    ACTIONS(3584), 1,
      sym_variable_name,
    STATE(2121), 1,
      sym_string,
    ACTIONS(3582), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1241), 3,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(3578), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1239), 36,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [23452] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3282), 1,
      anon_sym_DQUOTE,
    ACTIONS(3286), 1,
      sym_variable_name,
    STATE(1628), 1,
      sym_string,
    ACTIONS(3284), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1241), 3,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(3280), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1239), 36,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [23523] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3282), 1,
      anon_sym_DQUOTE,
    ACTIONS(3286), 1,
      sym_variable_name,
    STATE(1628), 1,
      sym_string,
    ACTIONS(3284), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1235), 3,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(3280), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1227), 36,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [23594] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2261), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(2263), 1,
      anon_sym_DOLLAR,
    ACTIONS(2267), 1,
      anon_sym_DQUOTE,
    ACTIONS(2269), 1,
      aux_sym_number_token1,
    ACTIONS(2271), 1,
      aux_sym_number_token2,
    ACTIONS(2273), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(2275), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2277), 1,
      anon_sym_BQUOTE,
    ACTIONS(2279), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(2289), 1,
      sym__brace_start,
    ACTIONS(3970), 1,
      sym__special_character,
    ACTIONS(3974), 1,
      sym_variable_name,
    ACTIONS(3976), 1,
      sym_test_operator,
    ACTIONS(4066), 1,
      aux_sym__simple_variable_name_token1,
    STATE(1743), 1,
      aux_sym__literal_repeat1,
    STATE(6739), 1,
      sym_subscript,
    ACTIONS(1799), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(2259), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(2281), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(3968), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(746), 3,
      sym_variable_assignment,
      sym_concatenation,
      aux_sym_declaration_command_repeat1,
    STATE(2330), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(1797), 16,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [23697] = 20,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3785), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(3787), 1,
      anon_sym_DOLLAR,
    ACTIONS(3791), 1,
      anon_sym_DQUOTE,
    ACTIONS(3793), 1,
      aux_sym_number_token1,
    ACTIONS(3795), 1,
      aux_sym_number_token2,
    ACTIONS(3797), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(3799), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3803), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3809), 1,
      sym__brace_start,
    ACTIONS(4070), 1,
      sym__special_character,
    ACTIONS(4072), 1,
      sym_test_operator,
    STATE(4305), 1,
      aux_sym__literal_repeat1,
    STATE(4736), 1,
      sym_concatenation,
    ACTIONS(2500), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(3783), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(3805), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(4068), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(4472), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2498), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_BQUOTE,
  [23791] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4078), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(4080), 1,
      anon_sym_DOLLAR,
    ACTIONS(4082), 1,
      sym__special_character,
    ACTIONS(4084), 1,
      anon_sym_DQUOTE,
    ACTIONS(4086), 1,
      aux_sym_number_token1,
    ACTIONS(4088), 1,
      aux_sym_number_token2,
    ACTIONS(4090), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(4092), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(4094), 1,
      anon_sym_BQUOTE,
    ACTIONS(4096), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(4100), 1,
      sym_test_operator,
    ACTIONS(4102), 1,
      sym__brace_start,
    STATE(4521), 1,
      aux_sym__literal_repeat1,
    STATE(5023), 1,
      sym_concatenation,
    ACTIONS(2496), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4076), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(4098), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(4074), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(4828), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2494), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [23887] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(809), 24,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_GT_GT,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LBRACK_LBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(579), 28,
      anon_sym_for,
      anon_sym_select,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LPAREN,
      anon_sym_while,
      anon_sym_until,
      anon_sym_if,
      anon_sym_case,
      anon_sym_function,
      anon_sym_BANG,
      anon_sym_LBRACK,
      anon_sym_declare,
      anon_sym_typeset,
      anon_sym_export,
      anon_sym_readonly,
      anon_sym_local,
      anon_sym_unset,
      anon_sym_unsetenv,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      sym__special_character,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
  [23947] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3837), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(3840), 1,
      anon_sym_DOLLAR,
    ACTIONS(3846), 1,
      anon_sym_DQUOTE,
    ACTIONS(3849), 1,
      aux_sym_number_token1,
    ACTIONS(3852), 1,
      aux_sym_number_token2,
    ACTIONS(3855), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(3858), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3861), 1,
      anon_sym_BQUOTE,
    ACTIONS(3864), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3873), 1,
      sym__brace_start,
    ACTIONS(4107), 1,
      sym__special_character,
    ACTIONS(4110), 1,
      sym_test_operator,
    STATE(2270), 1,
      aux_sym__literal_repeat1,
    ACTIONS(2216), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(3834), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(3867), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(781), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    ACTIONS(4104), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(1968), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2214), 19,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [24043] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4115), 1,
      anon_sym_DQUOTE,
    ACTIONS(4119), 1,
      sym_variable_name,
    STATE(2334), 1,
      sym_string,
    ACTIONS(4117), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1235), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
    ACTIONS(4113), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1227), 34,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [24113] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4078), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(4080), 1,
      anon_sym_DOLLAR,
    ACTIONS(4084), 1,
      anon_sym_DQUOTE,
    ACTIONS(4086), 1,
      aux_sym_number_token1,
    ACTIONS(4088), 1,
      aux_sym_number_token2,
    ACTIONS(4090), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(4092), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(4094), 1,
      anon_sym_BQUOTE,
    ACTIONS(4096), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(4102), 1,
      sym__brace_start,
    ACTIONS(4123), 1,
      sym__special_character,
    ACTIONS(4125), 1,
      sym_test_operator,
    STATE(4521), 1,
      aux_sym__literal_repeat1,
    STATE(5023), 1,
      sym_concatenation,
    ACTIONS(4076), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(4098), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(2496), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4121), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(4705), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2494), 19,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [24209] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3189), 1,
      anon_sym_DQUOTE,
    ACTIONS(4058), 1,
      sym_variable_name,
    STATE(1737), 1,
      sym_string,
    ACTIONS(4056), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1235), 3,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(4054), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1227), 35,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [24279] = 20,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3785), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(3787), 1,
      anon_sym_DOLLAR,
    ACTIONS(3791), 1,
      anon_sym_DQUOTE,
    ACTIONS(3793), 1,
      aux_sym_number_token1,
    ACTIONS(3795), 1,
      aux_sym_number_token2,
    ACTIONS(3797), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(3799), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3803), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3809), 1,
      sym__brace_start,
    ACTIONS(4070), 1,
      sym__special_character,
    ACTIONS(4129), 1,
      sym_test_operator,
    STATE(4277), 1,
      aux_sym__literal_repeat1,
    STATE(4782), 1,
      sym_concatenation,
    ACTIONS(2496), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(3783), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(3805), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(4127), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(4479), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2494), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_BQUOTE,
  [24373] = 20,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3580), 1,
      anon_sym_DQUOTE,
    ACTIONS(3610), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(3612), 1,
      anon_sym_DOLLAR,
    ACTIONS(3616), 1,
      aux_sym_number_token1,
    ACTIONS(3618), 1,
      aux_sym_number_token2,
    ACTIONS(3620), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(3622), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3626), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3632), 1,
      sym__brace_start,
    ACTIONS(4133), 1,
      sym__special_character,
    ACTIONS(4135), 1,
      sym_test_operator,
    STATE(2270), 1,
      aux_sym__literal_repeat1,
    ACTIONS(2074), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(3608), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(3628), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(781), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    ACTIONS(4131), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(1968), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2072), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_BQUOTE,
  [24467] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3189), 1,
      anon_sym_DQUOTE,
    ACTIONS(4058), 1,
      sym_variable_name,
    STATE(1737), 1,
      sym_string,
    ACTIONS(4056), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1241), 3,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(4054), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1239), 35,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [24537] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3580), 1,
      anon_sym_DQUOTE,
    ACTIONS(3610), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(3612), 1,
      anon_sym_DOLLAR,
    ACTIONS(3616), 1,
      aux_sym_number_token1,
    ACTIONS(3618), 1,
      aux_sym_number_token2,
    ACTIONS(3620), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(3622), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3624), 1,
      anon_sym_BQUOTE,
    ACTIONS(3626), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3632), 1,
      sym__brace_start,
    ACTIONS(4133), 1,
      sym__special_character,
    ACTIONS(4135), 1,
      sym_test_operator,
    STATE(2270), 1,
      aux_sym__literal_repeat1,
    ACTIONS(2074), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(3608), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(3628), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(781), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    ACTIONS(4131), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(1968), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2072), 19,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [24633] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3580), 1,
      anon_sym_DQUOTE,
    ACTIONS(3610), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(3612), 1,
      anon_sym_DOLLAR,
    ACTIONS(3616), 1,
      aux_sym_number_token1,
    ACTIONS(3618), 1,
      aux_sym_number_token2,
    ACTIONS(3620), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(3622), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3624), 1,
      anon_sym_BQUOTE,
    ACTIONS(3626), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3632), 1,
      sym__brace_start,
    ACTIONS(4133), 1,
      sym__special_character,
    ACTIONS(4135), 1,
      sym_test_operator,
    STATE(2270), 1,
      aux_sym__literal_repeat1,
    ACTIONS(2096), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(3608), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(3628), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(781), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    ACTIONS(4131), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(1968), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2094), 19,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [24729] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3785), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(3787), 1,
      anon_sym_DOLLAR,
    ACTIONS(3791), 1,
      anon_sym_DQUOTE,
    ACTIONS(3793), 1,
      aux_sym_number_token1,
    ACTIONS(3795), 1,
      aux_sym_number_token2,
    ACTIONS(3797), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(3799), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3801), 1,
      anon_sym_BQUOTE,
    ACTIONS(3803), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3809), 1,
      sym__brace_start,
    ACTIONS(4139), 1,
      sym__special_character,
    ACTIONS(4141), 1,
      sym_test_operator,
    STATE(4277), 1,
      aux_sym__literal_repeat1,
    STATE(4782), 1,
      sym_concatenation,
    ACTIONS(2496), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(3783), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(3805), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(4137), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(4834), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2494), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [24825] = 20,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3580), 1,
      anon_sym_DQUOTE,
    ACTIONS(3610), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(3612), 1,
      anon_sym_DOLLAR,
    ACTIONS(3616), 1,
      aux_sym_number_token1,
    ACTIONS(3618), 1,
      aux_sym_number_token2,
    ACTIONS(3620), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(3622), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3626), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3632), 1,
      sym__brace_start,
    ACTIONS(4133), 1,
      sym__special_character,
    ACTIONS(4135), 1,
      sym_test_operator,
    STATE(2270), 1,
      aux_sym__literal_repeat1,
    ACTIONS(2096), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(3608), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(3628), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(781), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    ACTIONS(4131), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(1968), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2094), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_BQUOTE,
  [24919] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3785), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(3787), 1,
      anon_sym_DOLLAR,
    ACTIONS(3791), 1,
      anon_sym_DQUOTE,
    ACTIONS(3793), 1,
      aux_sym_number_token1,
    ACTIONS(3795), 1,
      aux_sym_number_token2,
    ACTIONS(3797), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(3799), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3801), 1,
      anon_sym_BQUOTE,
    ACTIONS(3803), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3809), 1,
      sym__brace_start,
    ACTIONS(4139), 1,
      sym__special_character,
    ACTIONS(4145), 1,
      sym_test_operator,
    STATE(4305), 1,
      aux_sym__literal_repeat1,
    STATE(4736), 1,
      sym_concatenation,
    ACTIONS(2500), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(3783), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(3805), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(4143), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(4838), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2498), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [25015] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3580), 1,
      anon_sym_DQUOTE,
    ACTIONS(3584), 1,
      sym_variable_name,
    STATE(2121), 1,
      sym_string,
    ACTIONS(3582), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1241), 3,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(3578), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1239), 35,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [25085] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(2725), 1,
      ts_builtin_sym_end,
    ACTIONS(2717), 23,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_GT_GT,
      anon_sym_LBRACE,
      anon_sym_LBRACK_LBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(2715), 28,
      anon_sym_for,
      anon_sym_select,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LPAREN,
      anon_sym_while,
      anon_sym_until,
      anon_sym_if,
      anon_sym_case,
      anon_sym_function,
      anon_sym_BANG,
      anon_sym_LBRACK,
      anon_sym_declare,
      anon_sym_typeset,
      anon_sym_export,
      anon_sym_readonly,
      anon_sym_local,
      anon_sym_unset,
      anon_sym_unsetenv,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      sym__special_character,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
  [25147] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2267), 1,
      anon_sym_DQUOTE,
    ACTIONS(3829), 1,
      sym_variable_name,
    STATE(1595), 1,
      sym_string,
    ACTIONS(3827), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1235), 3,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(3825), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1227), 35,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [25217] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3580), 1,
      anon_sym_DQUOTE,
    ACTIONS(3584), 1,
      sym_variable_name,
    STATE(2121), 1,
      sym_string,
    ACTIONS(3582), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1235), 3,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(3578), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1227), 35,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [25287] = 25,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(4147), 1,
      sym_word,
    ACTIONS(4151), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(4153), 1,
      anon_sym_DOLLAR,
    ACTIONS(4155), 1,
      sym__special_character,
    ACTIONS(4157), 1,
      anon_sym_DQUOTE,
    ACTIONS(4161), 1,
      aux_sym_number_token1,
    ACTIONS(4163), 1,
      aux_sym_number_token2,
    ACTIONS(4165), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(4167), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(4169), 1,
      anon_sym_BQUOTE,
    ACTIONS(4171), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(4175), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(4177), 1,
      sym_variable_name,
    ACTIONS(4179), 1,
      sym_test_operator,
    ACTIONS(4181), 1,
      sym__brace_start,
    STATE(2688), 1,
      aux_sym__literal_repeat1,
    STATE(6751), 1,
      sym_subscript,
    ACTIONS(4149), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(4159), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(4173), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(803), 3,
      sym_variable_assignment,
      sym_concatenation,
      aux_sym_declaration_command_repeat1,
    ACTIONS(1797), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    STATE(2419), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(1799), 10,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [25391] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4078), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(4080), 1,
      anon_sym_DOLLAR,
    ACTIONS(4082), 1,
      sym__special_character,
    ACTIONS(4084), 1,
      anon_sym_DQUOTE,
    ACTIONS(4086), 1,
      aux_sym_number_token1,
    ACTIONS(4088), 1,
      aux_sym_number_token2,
    ACTIONS(4090), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(4092), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(4094), 1,
      anon_sym_BQUOTE,
    ACTIONS(4096), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(4102), 1,
      sym__brace_start,
    ACTIONS(4185), 1,
      sym_test_operator,
    STATE(4546), 1,
      aux_sym__literal_repeat1,
    STATE(4907), 1,
      sym_concatenation,
    ACTIONS(2500), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4076), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(4098), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(4183), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(4692), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2498), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [25487] = 25,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(4147), 1,
      sym_word,
    ACTIONS(4151), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(4153), 1,
      anon_sym_DOLLAR,
    ACTIONS(4155), 1,
      sym__special_character,
    ACTIONS(4157), 1,
      anon_sym_DQUOTE,
    ACTIONS(4161), 1,
      aux_sym_number_token1,
    ACTIONS(4163), 1,
      aux_sym_number_token2,
    ACTIONS(4165), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(4167), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(4169), 1,
      anon_sym_BQUOTE,
    ACTIONS(4171), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(4177), 1,
      sym_variable_name,
    ACTIONS(4179), 1,
      sym_test_operator,
    ACTIONS(4181), 1,
      sym__brace_start,
    ACTIONS(4187), 1,
      aux_sym__simple_variable_name_token1,
    STATE(2688), 1,
      aux_sym__literal_repeat1,
    STATE(6751), 1,
      sym_subscript,
    ACTIONS(4149), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(4159), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(4173), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(797), 3,
      sym_variable_assignment,
      sym_concatenation,
      aux_sym_declaration_command_repeat1,
    ACTIONS(1755), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    STATE(2419), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(1757), 10,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [25591] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(2721), 1,
      ts_builtin_sym_end,
    ACTIONS(2717), 23,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_GT_GT,
      anon_sym_LBRACE,
      anon_sym_LBRACK_LBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(2715), 28,
      anon_sym_for,
      anon_sym_select,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LPAREN,
      anon_sym_while,
      anon_sym_until,
      anon_sym_if,
      anon_sym_case,
      anon_sym_function,
      anon_sym_BANG,
      anon_sym_LBRACK,
      anon_sym_declare,
      anon_sym_typeset,
      anon_sym_export,
      anon_sym_readonly,
      anon_sym_local,
      anon_sym_unset,
      anon_sym_unsetenv,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      sym__special_character,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
  [25653] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2267), 1,
      anon_sym_DQUOTE,
    ACTIONS(3829), 1,
      sym_variable_name,
    STATE(1595), 1,
      sym_string,
    ACTIONS(3827), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1241), 3,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(3825), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1239), 35,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [25723] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(809), 23,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_GT_GT,
      anon_sym_LBRACE,
      anon_sym_LBRACK_LBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(579), 29,
      anon_sym_for,
      anon_sym_select,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LPAREN,
      anon_sym_while,
      anon_sym_until,
      anon_sym_done,
      anon_sym_if,
      anon_sym_case,
      anon_sym_function,
      anon_sym_BANG,
      anon_sym_LBRACK,
      anon_sym_declare,
      anon_sym_typeset,
      anon_sym_export,
      anon_sym_readonly,
      anon_sym_local,
      anon_sym_unset,
      anon_sym_unsetenv,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      sym__special_character,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
  [25783] = 25,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(4189), 1,
      sym_word,
    ACTIONS(4195), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(4198), 1,
      anon_sym_DOLLAR,
    ACTIONS(4201), 1,
      sym__special_character,
    ACTIONS(4204), 1,
      anon_sym_DQUOTE,
    ACTIONS(4210), 1,
      aux_sym_number_token1,
    ACTIONS(4213), 1,
      aux_sym_number_token2,
    ACTIONS(4216), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(4219), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(4222), 1,
      anon_sym_BQUOTE,
    ACTIONS(4225), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(4231), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(4234), 1,
      sym_variable_name,
    ACTIONS(4237), 1,
      sym_test_operator,
    ACTIONS(4240), 1,
      sym__brace_start,
    STATE(2688), 1,
      aux_sym__literal_repeat1,
    STATE(6751), 1,
      sym_subscript,
    ACTIONS(4192), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(4207), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(4228), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(803), 3,
      sym_variable_assignment,
      sym_concatenation,
      aux_sym_declaration_command_repeat1,
    ACTIONS(1702), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    STATE(2419), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(1704), 10,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [25887] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4078), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(4080), 1,
      anon_sym_DOLLAR,
    ACTIONS(4084), 1,
      anon_sym_DQUOTE,
    ACTIONS(4086), 1,
      aux_sym_number_token1,
    ACTIONS(4088), 1,
      aux_sym_number_token2,
    ACTIONS(4090), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(4092), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(4094), 1,
      anon_sym_BQUOTE,
    ACTIONS(4096), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(4102), 1,
      sym__brace_start,
    ACTIONS(4123), 1,
      sym__special_character,
    ACTIONS(4245), 1,
      sym_test_operator,
    STATE(4546), 1,
      aux_sym__literal_repeat1,
    STATE(4907), 1,
      sym_concatenation,
    ACTIONS(4076), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(4098), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(2500), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4243), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(4614), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2498), 19,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [25983] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4115), 1,
      anon_sym_DQUOTE,
    ACTIONS(4119), 1,
      sym_variable_name,
    STATE(2334), 1,
      sym_string,
    ACTIONS(4117), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1241), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
    ACTIONS(4113), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1239), 34,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [26053] = 6,
    ACTIONS(3), 1,
      sym_comment,
    STATE(1744), 1,
      aux_sym__literal_repeat1,
    STATE(811), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    ACTIONS(2074), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    STATE(2337), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2072), 34,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [26118] = 13,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4259), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4263), 1,
      sym_variable_name,
    STATE(6726), 1,
      sym_subscript,
    ACTIONS(4249), 2,
      anon_sym_SEMI,
      anon_sym_AMP,
    ACTIONS(4251), 2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
    ACTIONS(4253), 2,
      anon_sym_PIPE,
      anon_sym_PIPE_AMP,
    ACTIONS(4255), 2,
      anon_sym_LT_LT,
      anon_sym_LT_LT_DASH,
    STATE(4237), 2,
      sym_variable_assignment,
      aux_sym_variable_assignments_repeat1,
    ACTIONS(4261), 3,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
    STATE(4283), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(4257), 4,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
    ACTIONS(4247), 28,
      anon_sym_LPAREN_LPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [26197] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4266), 1,
      aux_sym_concatenation_token1,
    ACTIONS(4268), 1,
      sym__concat,
    STATE(820), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1267), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1265), 43,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [26262] = 10,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4263), 1,
      sym_variable_name,
    ACTIONS(4272), 1,
      aux_sym_heredoc_redirect_token1,
    STATE(6726), 1,
      sym_subscript,
    ACTIONS(4253), 2,
      anon_sym_PIPE,
      anon_sym_PIPE_AMP,
    STATE(4237), 2,
      sym_variable_assignment,
      aux_sym_variable_assignments_repeat1,
    ACTIONS(4261), 3,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
    STATE(4283), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(4270), 10,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_LT_LT_DASH,
    ACTIONS(4247), 28,
      anon_sym_LPAREN_LPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [26335] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(2725), 1,
      anon_sym_BQUOTE,
    ACTIONS(2717), 22,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_GT_GT,
      anon_sym_LBRACE,
      anon_sym_LBRACK_LBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(2715), 28,
      anon_sym_for,
      anon_sym_select,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LPAREN,
      anon_sym_while,
      anon_sym_until,
      anon_sym_if,
      anon_sym_case,
      anon_sym_function,
      anon_sym_BANG,
      anon_sym_LBRACK,
      anon_sym_declare,
      anon_sym_typeset,
      anon_sym_export,
      anon_sym_readonly,
      anon_sym_local,
      anon_sym_unset,
      anon_sym_unsetenv,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      sym__special_character,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
  [26396] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2735), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(2738), 1,
      anon_sym_DOLLAR,
    ACTIONS(2744), 1,
      anon_sym_DQUOTE,
    ACTIONS(2747), 1,
      aux_sym_number_token1,
    ACTIONS(2750), 1,
      aux_sym_number_token2,
    ACTIONS(2753), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(2756), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2759), 1,
      anon_sym_BQUOTE,
    ACTIONS(2762), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(2771), 1,
      sym__brace_start,
    ACTIONS(4277), 1,
      sym__special_character,
    ACTIONS(4280), 1,
      sym_test_operator,
    STATE(1744), 1,
      aux_sym__literal_repeat1,
    ACTIONS(2732), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(2765), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(811), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    ACTIONS(2216), 3,
      sym_file_descriptor,
      sym_variable_name,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4274), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(2337), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2214), 17,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [26491] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(2721), 1,
      anon_sym_BQUOTE,
    ACTIONS(2717), 22,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_GT_GT,
      anon_sym_LBRACE,
      anon_sym_LBRACK_LBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(2715), 28,
      anon_sym_for,
      anon_sym_select,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LPAREN,
      anon_sym_while,
      anon_sym_until,
      anon_sym_if,
      anon_sym_case,
      anon_sym_function,
      anon_sym_BANG,
      anon_sym_LBRACK,
      anon_sym_declare,
      anon_sym_typeset,
      anon_sym_export,
      anon_sym_readonly,
      anon_sym_local,
      anon_sym_unset,
      anon_sym_unsetenv,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      sym__special_character,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
  [26552] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4266), 1,
      aux_sym_concatenation_token1,
    ACTIONS(4283), 1,
      anon_sym_LPAREN,
    ACTIONS(4286), 1,
      sym__concat,
    STATE(814), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1281), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1271), 42,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [26619] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4266), 1,
      aux_sym_concatenation_token1,
    ACTIONS(4288), 1,
      sym__concat,
    STATE(820), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1288), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1286), 43,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [26684] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4266), 1,
      aux_sym_concatenation_token1,
    ACTIONS(4286), 1,
      sym__concat,
    STATE(814), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1281), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1271), 43,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [26749] = 6,
    ACTIONS(3), 1,
      sym_comment,
    STATE(1744), 1,
      aux_sym__literal_repeat1,
    STATE(811), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    ACTIONS(2096), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    STATE(2337), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2094), 34,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [26814] = 22,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(4290), 1,
      sym_word,
    ACTIONS(4296), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(4299), 1,
      anon_sym_DOLLAR,
    ACTIONS(4302), 1,
      sym__special_character,
    ACTIONS(4305), 1,
      anon_sym_DQUOTE,
    ACTIONS(4311), 1,
      aux_sym_number_token1,
    ACTIONS(4314), 1,
      aux_sym_number_token2,
    ACTIONS(4317), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(4320), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(4323), 1,
      anon_sym_BQUOTE,
    ACTIONS(4326), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(4332), 1,
      sym_test_operator,
    ACTIONS(4335), 1,
      sym__brace_start,
    STATE(2457), 1,
      aux_sym__literal_repeat1,
    ACTIONS(4293), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(4308), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(4329), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(817), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    ACTIONS(2214), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    STATE(2271), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2216), 13,
      sym_file_descriptor,
      sym_variable_name,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [26911] = 13,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4263), 1,
      sym_variable_name,
    ACTIONS(4340), 1,
      aux_sym_heredoc_redirect_token1,
    STATE(6726), 1,
      sym_subscript,
    ACTIONS(4251), 2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
    ACTIONS(4253), 2,
      anon_sym_PIPE,
      anon_sym_PIPE_AMP,
    ACTIONS(4255), 2,
      anon_sym_LT_LT,
      anon_sym_LT_LT_DASH,
    ACTIONS(4338), 2,
      anon_sym_SEMI,
      anon_sym_AMP,
    STATE(4237), 2,
      sym_variable_assignment,
      aux_sym_variable_assignments_repeat1,
    ACTIONS(4261), 3,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
    STATE(4283), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(2719), 4,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
    ACTIONS(4247), 28,
      anon_sym_LPAREN_LPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [26990] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4266), 1,
      aux_sym_concatenation_token1,
    ACTIONS(4286), 1,
      sym__concat,
    STATE(808), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(278), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(241), 43,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [27055] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4342), 1,
      aux_sym_concatenation_token1,
    ACTIONS(4345), 1,
      sym__concat,
    STATE(820), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1253), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1251), 43,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [27120] = 6,
    ACTIONS(71), 1,
      sym_comment,
    STATE(2457), 1,
      aux_sym__literal_repeat1,
    STATE(817), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    STATE(2271), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2094), 13,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      sym__special_character,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(2096), 26,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [27185] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4263), 1,
      sym_variable_name,
    STATE(6726), 1,
      sym_subscript,
    ACTIONS(4261), 2,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(4348), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    STATE(4237), 2,
      sym_variable_assignment,
      aux_sym_variable_assignments_repeat1,
    STATE(4283), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(4247), 18,
      anon_sym_LPAREN_LPAREN,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
    ACTIONS(4253), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [27256] = 6,
    ACTIONS(71), 1,
      sym_comment,
    STATE(2457), 1,
      aux_sym__literal_repeat1,
    STATE(817), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    STATE(2271), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2072), 13,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      sym__special_character,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(2074), 26,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [27321] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4266), 1,
      aux_sym_concatenation_token1,
    ACTIONS(4286), 1,
      sym__concat,
    STATE(808), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1263), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1261), 43,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [27386] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4078), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(4080), 1,
      anon_sym_DOLLAR,
    ACTIONS(4084), 1,
      anon_sym_DQUOTE,
    ACTIONS(4086), 1,
      aux_sym_number_token1,
    ACTIONS(4088), 1,
      aux_sym_number_token2,
    ACTIONS(4090), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(4092), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(4094), 1,
      anon_sym_BQUOTE,
    ACTIONS(4096), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(4102), 1,
      sym__brace_start,
    ACTIONS(4352), 1,
      sym__special_character,
    ACTIONS(4354), 1,
      sym_test_operator,
    STATE(4521), 1,
      aux_sym__literal_repeat1,
    STATE(5023), 1,
      sym_concatenation,
    ACTIONS(2496), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4076), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(4098), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(4350), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(4921), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2494), 19,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [27481] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4078), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(4080), 1,
      anon_sym_DOLLAR,
    ACTIONS(4084), 1,
      anon_sym_DQUOTE,
    ACTIONS(4086), 1,
      aux_sym_number_token1,
    ACTIONS(4088), 1,
      aux_sym_number_token2,
    ACTIONS(4090), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(4092), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(4094), 1,
      anon_sym_BQUOTE,
    ACTIONS(4096), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(4102), 1,
      sym__brace_start,
    ACTIONS(4352), 1,
      sym__special_character,
    ACTIONS(4358), 1,
      sym_test_operator,
    STATE(4546), 1,
      aux_sym__literal_repeat1,
    STATE(4907), 1,
      sym_concatenation,
    ACTIONS(2500), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4076), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(4098), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(4356), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(4901), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2498), 19,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [27576] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(2717), 23,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_GT_GT,
      anon_sym_LBRACE,
      anon_sym_LBRACK_LBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(2715), 28,
      anon_sym_for,
      anon_sym_select,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LPAREN,
      anon_sym_while,
      anon_sym_until,
      anon_sym_if,
      anon_sym_case,
      anon_sym_function,
      anon_sym_BANG,
      anon_sym_LBRACK,
      anon_sym_declare,
      anon_sym_typeset,
      anon_sym_export,
      anon_sym_readonly,
      anon_sym_local,
      anon_sym_unset,
      anon_sym_unsetenv,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      sym__special_character,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
  [27635] = 20,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4078), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(4080), 1,
      anon_sym_DOLLAR,
    ACTIONS(4084), 1,
      anon_sym_DQUOTE,
    ACTIONS(4086), 1,
      aux_sym_number_token1,
    ACTIONS(4088), 1,
      aux_sym_number_token2,
    ACTIONS(4090), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(4092), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(4096), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(4102), 1,
      sym__brace_start,
    ACTIONS(4362), 1,
      sym__special_character,
    ACTIONS(4364), 1,
      sym_test_operator,
    STATE(4546), 1,
      aux_sym__literal_repeat1,
    STATE(4907), 1,
      sym_concatenation,
    ACTIONS(2500), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4076), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(4098), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(4360), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(4640), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2498), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_BQUOTE,
  [27728] = 20,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4078), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(4080), 1,
      anon_sym_DOLLAR,
    ACTIONS(4084), 1,
      anon_sym_DQUOTE,
    ACTIONS(4086), 1,
      aux_sym_number_token1,
    ACTIONS(4088), 1,
      aux_sym_number_token2,
    ACTIONS(4090), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(4092), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(4096), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(4102), 1,
      sym__brace_start,
    ACTIONS(4362), 1,
      sym__special_character,
    ACTIONS(4368), 1,
      sym_test_operator,
    STATE(4521), 1,
      aux_sym__literal_repeat1,
    STATE(5023), 1,
      sym_concatenation,
    ACTIONS(2496), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4076), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(4098), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(4366), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(4627), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2494), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_BQUOTE,
  [27821] = 6,
    ACTIONS(71), 1,
      sym_comment,
    STATE(2476), 1,
      aux_sym__literal_repeat1,
    STATE(2637), 1,
      sym_concatenation,
    STATE(2303), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2494), 13,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      sym__special_character,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(2496), 26,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [27885] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1302), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1300), 44,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [27943] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4374), 1,
      aux_sym_concatenation_token1,
    ACTIONS(4376), 1,
      sym__concat,
    STATE(893), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4372), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4370), 42,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [28007] = 6,
    ACTIONS(3), 1,
      sym_comment,
    STATE(1740), 1,
      aux_sym__literal_repeat1,
    STATE(1957), 1,
      sym_concatenation,
    ACTIONS(2500), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    STATE(2291), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2498), 34,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [28071] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4374), 1,
      aux_sym_concatenation_token1,
    ACTIONS(4376), 1,
      sym__concat,
    STATE(891), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1263), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1261), 42,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [28135] = 13,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4384), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4386), 1,
      sym_variable_name,
    STATE(6772), 1,
      sym_subscript,
    ACTIONS(4253), 2,
      anon_sym_PIPE,
      anon_sym_PIPE_AMP,
    ACTIONS(4378), 2,
      anon_sym_SEMI,
      anon_sym_AMP,
    ACTIONS(4380), 2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
    ACTIONS(4382), 2,
      anon_sym_LT_LT,
      anon_sym_LT_LT_DASH,
    STATE(4358), 2,
      sym_variable_assignment,
      aux_sym_variable_assignments_repeat1,
    ACTIONS(2719), 3,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
    ACTIONS(4261), 3,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
    STATE(4455), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(4247), 28,
      anon_sym_LPAREN_LPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [28213] = 6,
    ACTIONS(3), 1,
      sym_comment,
    STATE(1739), 1,
      aux_sym__literal_repeat1,
    STATE(1953), 1,
      sym_concatenation,
    ACTIONS(2496), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    STATE(2328), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2494), 34,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [28277] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4389), 1,
      aux_sym_concatenation_token1,
    ACTIONS(4391), 1,
      sym__concat,
    STATE(841), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1267), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1265), 42,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [28341] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4389), 1,
      aux_sym_concatenation_token1,
    ACTIONS(4393), 1,
      sym__concat,
    STATE(841), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1288), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1286), 42,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [28405] = 10,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4272), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4386), 1,
      sym_variable_name,
    STATE(6772), 1,
      sym_subscript,
    ACTIONS(4253), 2,
      anon_sym_PIPE,
      anon_sym_PIPE_AMP,
    STATE(4358), 2,
      sym_variable_assignment,
      aux_sym_variable_assignments_repeat1,
    ACTIONS(4261), 3,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
    STATE(4455), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(4270), 9,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_LT_LT_DASH,
    ACTIONS(4247), 28,
      anon_sym_LPAREN_LPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [28477] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4397), 1,
      anon_sym_DQUOTE,
    ACTIONS(4401), 1,
      sym_variable_name,
    STATE(2412), 1,
      sym_string,
    ACTIONS(4399), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1235), 3,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(4395), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1227), 33,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [28545] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4403), 1,
      aux_sym_concatenation_token1,
    ACTIONS(4406), 1,
      sym__concat,
    STATE(841), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1253), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1251), 42,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [28609] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4397), 1,
      anon_sym_DQUOTE,
    ACTIONS(4401), 1,
      sym_variable_name,
    STATE(2412), 1,
      sym_string,
    ACTIONS(4399), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1241), 3,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(4395), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1239), 33,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [28677] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1322), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1320), 44,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [28735] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1338), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1336), 44,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [28793] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1342), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1340), 44,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [28851] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1314), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1312), 44,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [28909] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4386), 1,
      sym_variable_name,
    STATE(6772), 1,
      sym_subscript,
    ACTIONS(4261), 2,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(4348), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    STATE(4358), 2,
      sym_variable_assignment,
      aux_sym_variable_assignments_repeat1,
    STATE(4455), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(4247), 18,
      anon_sym_LPAREN_LPAREN,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
    ACTIONS(4253), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [28979] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1354), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1352), 44,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [29037] = 13,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4386), 1,
      sym_variable_name,
    ACTIONS(4411), 1,
      aux_sym_heredoc_redirect_token1,
    STATE(6772), 1,
      sym_subscript,
    ACTIONS(4253), 2,
      anon_sym_PIPE,
      anon_sym_PIPE_AMP,
    ACTIONS(4380), 2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
    ACTIONS(4382), 2,
      anon_sym_LT_LT,
      anon_sym_LT_LT_DASH,
    ACTIONS(4409), 2,
      anon_sym_SEMI,
      anon_sym_AMP,
    STATE(4358), 2,
      sym_variable_assignment,
      aux_sym_variable_assignments_repeat1,
    ACTIONS(4257), 3,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
    ACTIONS(4261), 3,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
    STATE(4455), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(4247), 28,
      anon_sym_LPAREN_LPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [29115] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4389), 1,
      aux_sym_concatenation_token1,
    ACTIONS(4413), 1,
      sym__concat,
    STATE(837), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1263), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1261), 42,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [29179] = 22,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(4415), 1,
      sym_word,
    ACTIONS(4421), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(4424), 1,
      anon_sym_DOLLAR,
    ACTIONS(4427), 1,
      sym__special_character,
    ACTIONS(4430), 1,
      anon_sym_DQUOTE,
    ACTIONS(4436), 1,
      aux_sym_number_token1,
    ACTIONS(4439), 1,
      aux_sym_number_token2,
    ACTIONS(4442), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(4445), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(4448), 1,
      anon_sym_BQUOTE,
    ACTIONS(4451), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(4457), 1,
      sym_test_operator,
    ACTIONS(4460), 1,
      sym__brace_start,
    STATE(2671), 1,
      aux_sym__literal_repeat1,
    ACTIONS(4418), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(4433), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(4454), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(851), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    ACTIONS(2214), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    STATE(2369), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2216), 12,
      sym_file_descriptor,
      sym_variable_name,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [29275] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4389), 1,
      aux_sym_concatenation_token1,
    ACTIONS(4413), 1,
      sym__concat,
    STATE(838), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1281), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1271), 42,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [29339] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4389), 1,
      aux_sym_concatenation_token1,
    ACTIONS(4413), 1,
      sym__concat,
    STATE(837), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(278), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(241), 42,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [29403] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4374), 1,
      aux_sym_concatenation_token1,
    ACTIONS(4376), 1,
      sym__concat,
    STATE(893), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4465), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4463), 42,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [29467] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1310), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1308), 44,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [29525] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1306), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1304), 44,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [29583] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1310), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1308), 44,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [29641] = 6,
    ACTIONS(71), 1,
      sym_comment,
    STATE(2671), 1,
      aux_sym__literal_repeat1,
    STATE(851), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    STATE(2369), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2072), 12,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(2074), 26,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [29705] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4374), 1,
      aux_sym_concatenation_token1,
    ACTIONS(4376), 1,
      sym__concat,
    STATE(891), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4469), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4467), 42,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [29769] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3233), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(3236), 1,
      anon_sym_DOLLAR,
    ACTIONS(3242), 1,
      anon_sym_DQUOTE,
    ACTIONS(3245), 1,
      aux_sym_number_token1,
    ACTIONS(3248), 1,
      aux_sym_number_token2,
    ACTIONS(3251), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(3254), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3257), 1,
      anon_sym_BQUOTE,
    ACTIONS(3260), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3269), 1,
      sym__brace_start,
    ACTIONS(4474), 1,
      sym__special_character,
    ACTIONS(4477), 1,
      sym_test_operator,
    STATE(2161), 1,
      aux_sym__literal_repeat1,
    ACTIONS(2216), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(3230), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(3263), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(860), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    ACTIONS(4471), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(2422), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2214), 17,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [29863] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4374), 1,
      aux_sym_concatenation_token1,
    ACTIONS(4376), 1,
      sym__concat,
    STATE(891), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4482), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4480), 42,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [29927] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3115), 1,
      anon_sym_DQUOTE,
    ACTIONS(3395), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(3397), 1,
      anon_sym_DOLLAR,
    ACTIONS(3401), 1,
      aux_sym_number_token1,
    ACTIONS(3403), 1,
      aux_sym_number_token2,
    ACTIONS(3405), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(3407), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3409), 1,
      anon_sym_BQUOTE,
    ACTIONS(3411), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3417), 1,
      sym__brace_start,
    ACTIONS(4486), 1,
      sym__special_character,
    ACTIONS(4488), 1,
      sym_test_operator,
    STATE(2161), 1,
      aux_sym__literal_repeat1,
    ACTIONS(2074), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(3393), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(3413), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(860), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    ACTIONS(4484), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(2422), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2072), 17,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [30021] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3115), 1,
      anon_sym_DQUOTE,
    ACTIONS(3395), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(3397), 1,
      anon_sym_DOLLAR,
    ACTIONS(3401), 1,
      aux_sym_number_token1,
    ACTIONS(3403), 1,
      aux_sym_number_token2,
    ACTIONS(3405), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(3407), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3409), 1,
      anon_sym_BQUOTE,
    ACTIONS(3411), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3417), 1,
      sym__brace_start,
    ACTIONS(4486), 1,
      sym__special_character,
    ACTIONS(4488), 1,
      sym_test_operator,
    STATE(2161), 1,
      aux_sym__literal_repeat1,
    ACTIONS(2096), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(3393), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(3413), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(860), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    ACTIONS(4484), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(2422), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2094), 17,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [30115] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1326), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1324), 44,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [30173] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3183), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(3185), 1,
      anon_sym_DOLLAR,
    ACTIONS(3189), 1,
      anon_sym_DQUOTE,
    ACTIONS(3191), 1,
      aux_sym_number_token1,
    ACTIONS(3193), 1,
      aux_sym_number_token2,
    ACTIONS(3195), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(3197), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3199), 1,
      anon_sym_BQUOTE,
    ACTIONS(3201), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3209), 1,
      sym__brace_start,
    ACTIONS(4492), 1,
      sym__special_character,
    ACTIONS(4494), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(4496), 1,
      sym_test_operator,
    STATE(1966), 1,
      aux_sym__literal_repeat1,
    ACTIONS(2564), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(3181), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(3203), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(875), 2,
      sym_concatenation,
      aux_sym_unset_command_repeat1,
    ACTIONS(4490), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(2367), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2562), 16,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [30269] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1253), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1251), 44,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [30327] = 22,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(4498), 1,
      sym_word,
    ACTIONS(4502), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(4504), 1,
      anon_sym_DOLLAR,
    ACTIONS(4506), 1,
      sym__special_character,
    ACTIONS(4508), 1,
      anon_sym_DQUOTE,
    ACTIONS(4512), 1,
      aux_sym_number_token1,
    ACTIONS(4514), 1,
      aux_sym_number_token2,
    ACTIONS(4516), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(4518), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(4520), 1,
      anon_sym_BQUOTE,
    ACTIONS(4522), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(4526), 1,
      sym_test_operator,
    ACTIONS(4528), 1,
      sym__brace_start,
    STATE(2715), 1,
      aux_sym__literal_repeat1,
    ACTIONS(4500), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(4510), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(4524), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(895), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    ACTIONS(2072), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    STATE(2438), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2074), 12,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [30423] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1350), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1348), 44,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [30481] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1298), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1296), 44,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [30539] = 6,
    ACTIONS(71), 1,
      sym_comment,
    STATE(2671), 1,
      aux_sym__literal_repeat1,
    STATE(851), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    STATE(2369), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2094), 12,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(2096), 26,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [30603] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1334), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1332), 44,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [30661] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1294), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1292), 44,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [30719] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3115), 1,
      anon_sym_DQUOTE,
    ACTIONS(3119), 1,
      sym_variable_name,
    STATE(1864), 1,
      sym_string,
    ACTIONS(3117), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1235), 3,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(3113), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1227), 33,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [30787] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4374), 1,
      aux_sym_concatenation_token1,
    ACTIONS(4376), 1,
      sym__concat,
    STATE(893), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4532), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4530), 42,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [30851] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3536), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(3539), 1,
      anon_sym_DOLLAR,
    ACTIONS(3545), 1,
      anon_sym_DQUOTE,
    ACTIONS(3548), 1,
      aux_sym_number_token1,
    ACTIONS(3551), 1,
      aux_sym_number_token2,
    ACTIONS(3554), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(3557), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3560), 1,
      anon_sym_BQUOTE,
    ACTIONS(3563), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3575), 1,
      sym__brace_start,
    ACTIONS(4537), 1,
      sym__special_character,
    ACTIONS(4540), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(4543), 1,
      sym_test_operator,
    STATE(1966), 1,
      aux_sym__literal_repeat1,
    ACTIONS(2442), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(3533), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(3566), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(875), 2,
      sym_concatenation,
      aux_sym_unset_command_repeat1,
    ACTIONS(4534), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(2367), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2440), 16,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [30947] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4374), 1,
      aux_sym_concatenation_token1,
    ACTIONS(4376), 1,
      sym__concat,
    STATE(891), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4548), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4546), 42,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [31011] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3115), 1,
      anon_sym_DQUOTE,
    ACTIONS(3119), 1,
      sym_variable_name,
    STATE(1864), 1,
      sym_string,
    ACTIONS(3117), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1241), 3,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(3113), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1239), 33,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [31079] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3183), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(3185), 1,
      anon_sym_DOLLAR,
    ACTIONS(3189), 1,
      anon_sym_DQUOTE,
    ACTIONS(3191), 1,
      aux_sym_number_token1,
    ACTIONS(3193), 1,
      aux_sym_number_token2,
    ACTIONS(3195), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(3197), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3199), 1,
      anon_sym_BQUOTE,
    ACTIONS(3201), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3209), 1,
      sym__brace_start,
    ACTIONS(4492), 1,
      sym__special_character,
    ACTIONS(4496), 1,
      sym_test_operator,
    ACTIONS(4550), 1,
      aux_sym__simple_variable_name_token1,
    STATE(1966), 1,
      aux_sym__literal_repeat1,
    ACTIONS(2299), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(3181), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(3203), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(865), 2,
      sym_concatenation,
      aux_sym_unset_command_repeat1,
    ACTIONS(4490), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(2367), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2297), 16,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [31175] = 22,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(4498), 1,
      sym_word,
    ACTIONS(4502), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(4504), 1,
      anon_sym_DOLLAR,
    ACTIONS(4506), 1,
      sym__special_character,
    ACTIONS(4508), 1,
      anon_sym_DQUOTE,
    ACTIONS(4512), 1,
      aux_sym_number_token1,
    ACTIONS(4514), 1,
      aux_sym_number_token2,
    ACTIONS(4516), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(4518), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(4520), 1,
      anon_sym_BQUOTE,
    ACTIONS(4522), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(4526), 1,
      sym_test_operator,
    ACTIONS(4528), 1,
      sym__brace_start,
    STATE(2715), 1,
      aux_sym__literal_repeat1,
    ACTIONS(4500), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(4510), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(4524), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(895), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    ACTIONS(2094), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    STATE(2438), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2096), 12,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [31271] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4374), 1,
      aux_sym_concatenation_token1,
    ACTIONS(4376), 1,
      sym__concat,
    STATE(893), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4554), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4552), 42,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [31335] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4558), 1,
      anon_sym_DQUOTE,
    ACTIONS(4562), 1,
      sym_variable_name,
    STATE(2447), 1,
      sym_string,
    ACTIONS(4560), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1241), 3,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(4556), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1239), 33,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [31403] = 6,
    ACTIONS(71), 1,
      sym_comment,
    STATE(2541), 1,
      aux_sym__literal_repeat1,
    STATE(2622), 1,
      sym_concatenation,
    STATE(2300), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2498), 13,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      sym__special_character,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(2500), 26,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [31467] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4558), 1,
      anon_sym_DQUOTE,
    ACTIONS(4562), 1,
      sym_variable_name,
    STATE(2447), 1,
      sym_string,
    ACTIONS(4560), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1235), 3,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(4556), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1227), 33,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [31535] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1318), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1316), 44,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [31593] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4374), 1,
      aux_sym_concatenation_token1,
    ACTIONS(4376), 1,
      sym__concat,
    STATE(891), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4566), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4564), 42,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [31657] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4389), 1,
      aux_sym_concatenation_token1,
    ACTIONS(4413), 1,
      sym__concat,
    ACTIONS(4568), 1,
      anon_sym_LPAREN,
    STATE(838), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1281), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1271), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [31723] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1330), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1328), 44,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [31781] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1346), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1344), 44,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [31839] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1358), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1356), 44,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [31897] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3282), 1,
      anon_sym_DQUOTE,
    ACTIONS(3286), 1,
      sym_variable_name,
    STATE(1628), 1,
      sym_string,
    ACTIONS(3284), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1241), 3,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(3280), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1239), 33,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [31965] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4374), 1,
      aux_sym_concatenation_token1,
    ACTIONS(4571), 1,
      sym__concat,
    STATE(894), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1267), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1265), 42,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [32029] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3282), 1,
      anon_sym_DQUOTE,
    ACTIONS(3286), 1,
      sym_variable_name,
    STATE(1628), 1,
      sym_string,
    ACTIONS(3284), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1235), 3,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(3280), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1227), 33,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [32097] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4374), 1,
      aux_sym_concatenation_token1,
    ACTIONS(4573), 1,
      sym__concat,
    STATE(894), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1288), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1286), 42,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [32161] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4575), 1,
      aux_sym_concatenation_token1,
    ACTIONS(4578), 1,
      sym__concat,
    STATE(894), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1253), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1251), 42,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [32225] = 22,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(4581), 1,
      sym_word,
    ACTIONS(4587), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(4590), 1,
      anon_sym_DOLLAR,
    ACTIONS(4593), 1,
      sym__special_character,
    ACTIONS(4596), 1,
      anon_sym_DQUOTE,
    ACTIONS(4602), 1,
      aux_sym_number_token1,
    ACTIONS(4605), 1,
      aux_sym_number_token2,
    ACTIONS(4608), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(4611), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(4614), 1,
      anon_sym_BQUOTE,
    ACTIONS(4617), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(4623), 1,
      sym_test_operator,
    ACTIONS(4626), 1,
      sym__brace_start,
    STATE(2715), 1,
      aux_sym__literal_repeat1,
    ACTIONS(4584), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(4599), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(4620), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(895), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    ACTIONS(2214), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    STATE(2438), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2216), 12,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [32321] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2267), 1,
      anon_sym_DQUOTE,
    ACTIONS(3829), 1,
      sym_variable_name,
    STATE(1595), 1,
      sym_string,
    ACTIONS(3827), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1235), 3,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(3825), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1227), 32,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [32388] = 22,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(4629), 1,
      sym_word,
    ACTIONS(4633), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(4635), 1,
      anon_sym_DOLLAR,
    ACTIONS(4637), 1,
      sym__special_character,
    ACTIONS(4639), 1,
      anon_sym_DQUOTE,
    ACTIONS(4643), 1,
      aux_sym_number_token1,
    ACTIONS(4645), 1,
      aux_sym_number_token2,
    ACTIONS(4647), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(4649), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(4651), 1,
      anon_sym_BQUOTE,
    ACTIONS(4653), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(4657), 1,
      sym_test_operator,
    ACTIONS(4659), 1,
      sym__brace_start,
    STATE(2914), 1,
      aux_sym__literal_repeat1,
    ACTIONS(4631), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(4641), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(4655), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(936), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    ACTIONS(2072), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    STATE(2487), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2074), 11,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [32483] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1310), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1308), 43,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [32540] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2267), 1,
      anon_sym_DQUOTE,
    ACTIONS(3829), 1,
      sym_variable_name,
    STATE(1595), 1,
      sym_string,
    ACTIONS(3827), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1241), 3,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(3825), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1239), 32,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [32607] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1306), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1304), 43,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [32664] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1310), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1308), 43,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [32721] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4663), 1,
      anon_sym_DQUOTE,
    ACTIONS(4667), 1,
      sym_variable_name,
    STATE(2716), 1,
      sym_string,
    ACTIONS(1241), 2,
      sym__concat,
      sym_test_operator,
    ACTIONS(4665), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(4661), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1239), 33,
      anon_sym_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
  [32788] = 22,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(4629), 1,
      sym_word,
    ACTIONS(4633), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(4635), 1,
      anon_sym_DOLLAR,
    ACTIONS(4637), 1,
      sym__special_character,
    ACTIONS(4639), 1,
      anon_sym_DQUOTE,
    ACTIONS(4643), 1,
      aux_sym_number_token1,
    ACTIONS(4645), 1,
      aux_sym_number_token2,
    ACTIONS(4647), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(4649), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(4651), 1,
      anon_sym_BQUOTE,
    ACTIONS(4653), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(4657), 1,
      sym_test_operator,
    ACTIONS(4659), 1,
      sym__brace_start,
    STATE(2914), 1,
      aux_sym__literal_repeat1,
    ACTIONS(4631), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(4641), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(4655), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(936), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    ACTIONS(2094), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    STATE(2487), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2096), 11,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [32883] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4663), 1,
      anon_sym_DQUOTE,
    ACTIONS(4667), 1,
      sym_variable_name,
    STATE(2716), 1,
      sym_string,
    ACTIONS(1235), 2,
      sym__concat,
      sym_test_operator,
    ACTIONS(4665), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(4661), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1227), 33,
      anon_sym_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
  [32950] = 6,
    ACTIONS(71), 1,
      sym_comment,
    STATE(2718), 1,
      aux_sym__literal_repeat1,
    STATE(2864), 1,
      sym_concatenation,
    STATE(2357), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2494), 12,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(2496), 26,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [33013] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3189), 1,
      anon_sym_DQUOTE,
    ACTIONS(4058), 1,
      sym_variable_name,
    STATE(1737), 1,
      sym_string,
    ACTIONS(4056), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1241), 3,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(4054), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1239), 32,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [33080] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4669), 1,
      anon_sym_LPAREN,
    ACTIONS(4672), 1,
      aux_sym_concatenation_token1,
    ACTIONS(4674), 1,
      sym__concat,
    STATE(983), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1281), 6,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1271), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [33145] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3785), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(3787), 1,
      anon_sym_DOLLAR,
    ACTIONS(3791), 1,
      anon_sym_DQUOTE,
    ACTIONS(3793), 1,
      aux_sym_number_token1,
    ACTIONS(3795), 1,
      aux_sym_number_token2,
    ACTIONS(3797), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(3799), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3801), 1,
      anon_sym_BQUOTE,
    ACTIONS(3803), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3809), 1,
      sym__brace_start,
    ACTIONS(4678), 1,
      sym__special_character,
    ACTIONS(4680), 1,
      sym_test_operator,
    STATE(4277), 1,
      aux_sym__literal_repeat1,
    STATE(4782), 1,
      sym_concatenation,
    ACTIONS(2496), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(3783), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(3805), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(4676), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(5135), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2494), 17,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [33238] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3785), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(3787), 1,
      anon_sym_DOLLAR,
    ACTIONS(3791), 1,
      anon_sym_DQUOTE,
    ACTIONS(3793), 1,
      aux_sym_number_token1,
    ACTIONS(3795), 1,
      aux_sym_number_token2,
    ACTIONS(3797), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(3799), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3801), 1,
      anon_sym_BQUOTE,
    ACTIONS(3803), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3809), 1,
      sym__brace_start,
    ACTIONS(4678), 1,
      sym__special_character,
    ACTIONS(4684), 1,
      sym_test_operator,
    STATE(4305), 1,
      aux_sym__literal_repeat1,
    STATE(4736), 1,
      sym_concatenation,
    ACTIONS(2500), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(3783), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(3805), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(4682), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(5133), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2498), 17,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [33331] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1253), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1251), 43,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [33388] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3189), 1,
      anon_sym_DQUOTE,
    ACTIONS(4058), 1,
      sym_variable_name,
    STATE(1737), 1,
      sym_string,
    ACTIONS(4056), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1235), 3,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(4054), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1227), 32,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [33455] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1326), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1324), 43,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [33512] = 13,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4686), 1,
      ts_builtin_sym_end,
    ACTIONS(4694), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4696), 1,
      sym_variable_name,
    STATE(6757), 1,
      sym_subscript,
    ACTIONS(4253), 2,
      anon_sym_PIPE,
      anon_sym_PIPE_AMP,
    ACTIONS(4690), 2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
    ACTIONS(4692), 2,
      anon_sym_LT_LT,
      anon_sym_LT_LT_DASH,
    STATE(4574), 2,
      sym_variable_assignment,
      aux_sym_variable_assignments_repeat1,
    ACTIONS(4261), 3,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(4688), 3,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
    STATE(4650), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(4247), 28,
      anon_sym_LPAREN_LPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [33589] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1322), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1320), 43,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [33646] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1338), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1336), 43,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [33703] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1342), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1340), 43,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [33760] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1354), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1352), 43,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [33817] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1314), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1312), 43,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [33874] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1342), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1340), 43,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [33931] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4699), 1,
      aux_sym_concatenation_token1,
    ACTIONS(4701), 1,
      sym__concat,
    STATE(935), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1263), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1261), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [33994] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1314), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1312), 43,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [34051] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1350), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1348), 43,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [34108] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1354), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1352), 43,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [34165] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1338), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1336), 43,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [34222] = 30,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(451), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(453), 1,
      anon_sym_DOLLAR,
    ACTIONS(457), 1,
      anon_sym_DQUOTE,
    ACTIONS(461), 1,
      aux_sym_number_token1,
    ACTIONS(463), 1,
      aux_sym_number_token2,
    ACTIONS(465), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(467), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(469), 1,
      anon_sym_BQUOTE,
    ACTIONS(471), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(479), 1,
      sym_test_operator,
    ACTIONS(481), 1,
      sym__brace_start,
    ACTIONS(1394), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(1396), 1,
      sym_file_descriptor,
    ACTIONS(3018), 1,
      sym_word,
    ACTIONS(4703), 1,
      sym__special_character,
    ACTIONS(4705), 1,
      sym_variable_name,
    STATE(529), 1,
      sym_command_name,
    STATE(926), 1,
      aux_sym__literal_repeat1,
    STATE(1082), 1,
      sym_concatenation,
    STATE(6771), 1,
      sym_subscript,
    ACTIONS(449), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(459), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(473), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1392), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    STATE(3127), 2,
      sym_variable_assignment,
      aux_sym_command_repeat1,
    STATE(3581), 2,
      sym_file_redirect,
      sym_herestring_redirect,
    ACTIONS(1390), 3,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
    ACTIONS(1388), 5,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    STATE(815), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [34333] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4707), 1,
      sym__special_character,
    STATE(952), 1,
      aux_sym__literal_repeat1,
    ACTIONS(278), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(241), 42,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [34394] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1310), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1308), 43,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [34451] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1306), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1304), 43,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [34508] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1310), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1308), 43,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [34565] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1322), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1320), 43,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [34622] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1346), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1344), 43,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [34679] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4699), 1,
      aux_sym_concatenation_token1,
    ACTIONS(4701), 1,
      sym__concat,
    STATE(935), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(278), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(241), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [34742] = 30,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(314), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(316), 1,
      anon_sym_DOLLAR,
    ACTIONS(320), 1,
      anon_sym_DQUOTE,
    ACTIONS(324), 1,
      aux_sym_number_token1,
    ACTIONS(326), 1,
      aux_sym_number_token2,
    ACTIONS(328), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(330), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(332), 1,
      anon_sym_BQUOTE,
    ACTIONS(334), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(344), 1,
      sym__brace_start,
    ACTIONS(1023), 1,
      sym_test_operator,
    ACTIONS(1394), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(1396), 1,
      sym_file_descriptor,
    ACTIONS(2961), 1,
      sym_word,
    ACTIONS(4705), 1,
      sym_variable_name,
    ACTIONS(4709), 1,
      sym__special_character,
    STATE(600), 1,
      sym_command_name,
    STATE(1267), 1,
      aux_sym__literal_repeat1,
    STATE(1376), 1,
      sym_concatenation,
    STATE(6771), 1,
      sym_subscript,
    ACTIONS(312), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(336), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1017), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(1392), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    STATE(3127), 2,
      sym_variable_assignment,
      aux_sym_command_repeat1,
    STATE(3581), 2,
      sym_file_redirect,
      sym_herestring_redirect,
    ACTIONS(1390), 3,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
    ACTIONS(1388), 5,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    STATE(1717), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [34853] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1358), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1356), 43,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [34910] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4699), 1,
      aux_sym_concatenation_token1,
    ACTIONS(4711), 1,
      sym__concat,
    STATE(965), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1267), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1265), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [34973] = 22,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(4713), 1,
      sym_word,
    ACTIONS(4719), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(4722), 1,
      anon_sym_DOLLAR,
    ACTIONS(4725), 1,
      sym__special_character,
    ACTIONS(4728), 1,
      anon_sym_DQUOTE,
    ACTIONS(4734), 1,
      aux_sym_number_token1,
    ACTIONS(4737), 1,
      aux_sym_number_token2,
    ACTIONS(4740), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(4743), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(4746), 1,
      anon_sym_BQUOTE,
    ACTIONS(4749), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(4755), 1,
      sym_test_operator,
    ACTIONS(4758), 1,
      sym__brace_start,
    STATE(2914), 1,
      aux_sym__literal_repeat1,
    ACTIONS(4716), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(4731), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(4752), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(936), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    ACTIONS(2214), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    STATE(2487), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2216), 11,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [35068] = 30,
    ACTIONS(47), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(49), 1,
      anon_sym_DOLLAR,
    ACTIONS(53), 1,
      anon_sym_DQUOTE,
    ACTIONS(57), 1,
      aux_sym_number_token1,
    ACTIONS(59), 1,
      aux_sym_number_token2,
    ACTIONS(61), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(63), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(65), 1,
      anon_sym_BQUOTE,
    ACTIONS(67), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(77), 1,
      sym_test_operator,
    ACTIONS(79), 1,
      sym__brace_start,
    ACTIONS(1394), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(1396), 1,
      sym_file_descriptor,
    ACTIONS(2666), 1,
      sym_word,
    ACTIONS(4705), 1,
      sym_variable_name,
    ACTIONS(4761), 1,
      sym__special_character,
    STATE(537), 1,
      sym_command_name,
    STATE(1212), 1,
      aux_sym__literal_repeat1,
    STATE(1386), 1,
      sym_concatenation,
    STATE(6771), 1,
      sym_subscript,
    ACTIONS(45), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(55), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(69), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1392), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    STATE(3127), 2,
      sym_variable_assignment,
      aux_sym_command_repeat1,
    STATE(3581), 2,
      sym_file_redirect,
      sym_herestring_redirect,
    ACTIONS(1390), 3,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
    ACTIONS(1388), 5,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    STATE(957), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [35179] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4699), 1,
      aux_sym_concatenation_token1,
    ACTIONS(4701), 1,
      sym__concat,
    STATE(947), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1281), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1271), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [35242] = 22,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(4763), 1,
      sym_word,
    ACTIONS(4767), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(4769), 1,
      anon_sym_DOLLAR,
    ACTIONS(4771), 1,
      sym__special_character,
    ACTIONS(4773), 1,
      anon_sym_DQUOTE,
    ACTIONS(4777), 1,
      aux_sym_number_token1,
    ACTIONS(4779), 1,
      aux_sym_number_token2,
    ACTIONS(4781), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(4783), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(4785), 1,
      anon_sym_BQUOTE,
    ACTIONS(4787), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(4791), 1,
      sym_test_operator,
    ACTIONS(4793), 1,
      sym__brace_start,
    STATE(5357), 1,
      aux_sym__literal_repeat1,
    STATE(5539), 1,
      sym_concatenation,
    ACTIONS(4765), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(4775), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(4789), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(2498), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    STATE(5178), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2500), 12,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [35337] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1278), 1,
      anon_sym_LPAREN,
    ACTIONS(4699), 1,
      aux_sym_concatenation_token1,
    ACTIONS(4701), 1,
      sym__concat,
    STATE(947), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1281), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1271), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [35402] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3580), 1,
      anon_sym_DQUOTE,
    ACTIONS(3584), 1,
      sym_variable_name,
    STATE(2121), 1,
      sym_string,
    ACTIONS(3582), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1241), 3,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(3578), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1239), 32,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [35469] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3580), 1,
      anon_sym_DQUOTE,
    ACTIONS(3584), 1,
      sym_variable_name,
    STATE(2121), 1,
      sym_string,
    ACTIONS(3582), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1235), 3,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(3578), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1227), 32,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [35536] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4797), 1,
      anon_sym_DQUOTE,
    ACTIONS(4801), 1,
      sym_variable_name,
    STATE(2585), 1,
      sym_string,
    ACTIONS(4799), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1235), 3,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(4795), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1227), 32,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [35603] = 22,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(4767), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(4769), 1,
      anon_sym_DOLLAR,
    ACTIONS(4771), 1,
      sym__special_character,
    ACTIONS(4773), 1,
      anon_sym_DQUOTE,
    ACTIONS(4777), 1,
      aux_sym_number_token1,
    ACTIONS(4779), 1,
      aux_sym_number_token2,
    ACTIONS(4781), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(4783), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(4785), 1,
      anon_sym_BQUOTE,
    ACTIONS(4787), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(4793), 1,
      sym__brace_start,
    ACTIONS(4803), 1,
      sym_word,
    ACTIONS(4807), 1,
      sym_test_operator,
    STATE(5376), 1,
      aux_sym__literal_repeat1,
    STATE(5530), 1,
      sym_concatenation,
    ACTIONS(4765), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(4789), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(4805), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(2494), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    STATE(5182), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2496), 12,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [35698] = 13,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4257), 1,
      anon_sym_RPAREN,
    ACTIONS(4411), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4811), 1,
      sym_variable_name,
    STATE(6786), 1,
      sym_subscript,
    ACTIONS(4253), 2,
      anon_sym_PIPE,
      anon_sym_PIPE_AMP,
    ACTIONS(4692), 2,
      anon_sym_LT_LT,
      anon_sym_LT_LT_DASH,
    ACTIONS(4809), 2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
    STATE(4534), 2,
      sym_variable_assignment,
      aux_sym_variable_assignments_repeat1,
    ACTIONS(4261), 3,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(4409), 3,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
    STATE(4746), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(4247), 28,
      anon_sym_LPAREN_LPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [35775] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4797), 1,
      anon_sym_DQUOTE,
    ACTIONS(4801), 1,
      sym_variable_name,
    STATE(2585), 1,
      sym_string,
    ACTIONS(4799), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1241), 3,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(4795), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1239), 32,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [35842] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4699), 1,
      aux_sym_concatenation_token1,
    ACTIONS(4814), 1,
      sym__concat,
    STATE(965), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1288), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1286), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [35905] = 10,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4272), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4811), 1,
      sym_variable_name,
    STATE(6786), 1,
      sym_subscript,
    ACTIONS(4253), 2,
      anon_sym_PIPE,
      anon_sym_PIPE_AMP,
    STATE(4534), 2,
      sym_variable_assignment,
      aux_sym_variable_assignments_repeat1,
    ACTIONS(4261), 3,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
    STATE(4746), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(4270), 8,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_LT_LT_DASH,
    ACTIONS(4247), 28,
      anon_sym_LPAREN_LPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [35976] = 6,
    ACTIONS(71), 1,
      sym_comment,
    STATE(2726), 1,
      aux_sym__literal_repeat1,
    STATE(2846), 1,
      sym_concatenation,
    STATE(2427), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2498), 12,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(2500), 26,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [36039] = 23,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(4816), 1,
      sym_word,
    ACTIONS(4820), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(4822), 1,
      anon_sym_DOLLAR,
    ACTIONS(4824), 1,
      sym__special_character,
    ACTIONS(4826), 1,
      anon_sym_DQUOTE,
    ACTIONS(4830), 1,
      aux_sym_number_token1,
    ACTIONS(4832), 1,
      aux_sym_number_token2,
    ACTIONS(4834), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(4836), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(4838), 1,
      anon_sym_BQUOTE,
    ACTIONS(4840), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(4844), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(4846), 1,
      sym_test_operator,
    ACTIONS(4848), 1,
      sym__brace_start,
    STATE(2939), 1,
      aux_sym__literal_repeat1,
    ACTIONS(4818), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(4828), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(4842), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(962), 2,
      sym_concatenation,
      aux_sym_unset_command_repeat1,
    ACTIONS(2297), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    STATE(2539), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2299), 10,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [36136] = 30,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(314), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(316), 1,
      anon_sym_DOLLAR,
    ACTIONS(320), 1,
      anon_sym_DQUOTE,
    ACTIONS(324), 1,
      aux_sym_number_token1,
    ACTIONS(326), 1,
      aux_sym_number_token2,
    ACTIONS(328), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(330), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(332), 1,
      anon_sym_BQUOTE,
    ACTIONS(334), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(344), 1,
      sym__brace_start,
    ACTIONS(775), 1,
      sym_test_operator,
    ACTIONS(1394), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(1396), 1,
      sym_file_descriptor,
    ACTIONS(2941), 1,
      sym_word,
    ACTIONS(4705), 1,
      sym_variable_name,
    ACTIONS(4850), 1,
      sym__special_character,
    STATE(536), 1,
      sym_command_name,
    STATE(1267), 1,
      aux_sym__literal_repeat1,
    STATE(1376), 1,
      sym_concatenation,
    STATE(6771), 1,
      sym_subscript,
    ACTIONS(312), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(336), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(773), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(1392), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    STATE(3127), 2,
      sym_variable_assignment,
      aux_sym_command_repeat1,
    STATE(3581), 2,
      sym_file_redirect,
      sym_herestring_redirect,
    ACTIONS(1390), 3,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
    ACTIONS(1388), 5,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    STATE(938), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [36247] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4852), 1,
      sym__special_character,
    STATE(952), 1,
      aux_sym__literal_repeat1,
    ACTIONS(1364), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1362), 42,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [36308] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4857), 1,
      anon_sym_DQUOTE,
    ACTIONS(4861), 1,
      sym_variable_name,
    STATE(2638), 1,
      sym_string,
    ACTIONS(4859), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1235), 3,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(4855), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1227), 32,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [36375] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4857), 1,
      anon_sym_DQUOTE,
    ACTIONS(4861), 1,
      sym_variable_name,
    STATE(2638), 1,
      sym_string,
    ACTIONS(4859), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1241), 3,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(4855), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1239), 32,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [36442] = 30,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1043), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1045), 1,
      anon_sym_DOLLAR,
    ACTIONS(1049), 1,
      anon_sym_DQUOTE,
    ACTIONS(1053), 1,
      aux_sym_number_token1,
    ACTIONS(1055), 1,
      aux_sym_number_token2,
    ACTIONS(1057), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1059), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1061), 1,
      anon_sym_BQUOTE,
    ACTIONS(1063), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1069), 1,
      sym_test_operator,
    ACTIONS(1071), 1,
      sym__brace_start,
    ACTIONS(1394), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(1396), 1,
      sym_file_descriptor,
    ACTIONS(2668), 1,
      sym_word,
    ACTIONS(4705), 1,
      sym_variable_name,
    ACTIONS(4863), 1,
      sym__special_character,
    STATE(617), 1,
      sym_command_name,
    STATE(2332), 1,
      aux_sym__literal_repeat1,
    STATE(2368), 1,
      sym_concatenation,
    STATE(6771), 1,
      sym_subscript,
    ACTIONS(1041), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1051), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(1065), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1392), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    STATE(3127), 2,
      sym_variable_assignment,
      aux_sym_command_repeat1,
    STATE(3581), 2,
      sym_file_redirect,
      sym_herestring_redirect,
    ACTIONS(1390), 3,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
    ACTIONS(1388), 5,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    STATE(1799), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [36553] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4867), 1,
      anon_sym_DQUOTE,
    ACTIONS(4871), 1,
      sym_variable_name,
    STATE(2525), 1,
      sym_string,
    ACTIONS(4869), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1235), 3,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(4865), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1227), 32,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [36620] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4672), 1,
      aux_sym_concatenation_token1,
    ACTIONS(4674), 1,
      sym__concat,
    STATE(983), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1281), 6,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1271), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [36683] = 30,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(314), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(316), 1,
      anon_sym_DOLLAR,
    ACTIONS(320), 1,
      anon_sym_DQUOTE,
    ACTIONS(324), 1,
      aux_sym_number_token1,
    ACTIONS(326), 1,
      aux_sym_number_token2,
    ACTIONS(328), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(330), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(332), 1,
      anon_sym_BQUOTE,
    ACTIONS(334), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(342), 1,
      sym_test_operator,
    ACTIONS(344), 1,
      sym__brace_start,
    ACTIONS(1394), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(1396), 1,
      sym_file_descriptor,
    ACTIONS(2723), 1,
      sym_word,
    ACTIONS(4705), 1,
      sym_variable_name,
    ACTIONS(4873), 1,
      sym__special_character,
    STATE(543), 1,
      sym_command_name,
    STATE(1267), 1,
      aux_sym__literal_repeat1,
    STATE(1376), 1,
      sym_concatenation,
    STATE(6771), 1,
      sym_subscript,
    ACTIONS(312), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(322), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(336), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1392), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    STATE(3127), 2,
      sym_variable_assignment,
      aux_sym_command_repeat1,
    STATE(3581), 2,
      sym_file_redirect,
      sym_herestring_redirect,
    ACTIONS(1390), 3,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
    ACTIONS(1388), 5,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    STATE(1078), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [36794] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4672), 1,
      aux_sym_concatenation_token1,
    ACTIONS(4875), 1,
      sym__concat,
    STATE(996), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1267), 6,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1265), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [36857] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4867), 1,
      anon_sym_DQUOTE,
    ACTIONS(4871), 1,
      sym_variable_name,
    STATE(2525), 1,
      sym_string,
    ACTIONS(4869), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1241), 3,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(4865), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1239), 32,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [36924] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4811), 1,
      sym_variable_name,
    STATE(6786), 1,
      sym_subscript,
    ACTIONS(4261), 2,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(4348), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    STATE(4534), 2,
      sym_variable_assignment,
      aux_sym_variable_assignments_repeat1,
    STATE(4746), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(4247), 18,
      anon_sym_LPAREN_LPAREN,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
    ACTIONS(4253), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [36993] = 23,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(4816), 1,
      sym_word,
    ACTIONS(4820), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(4822), 1,
      anon_sym_DOLLAR,
    ACTIONS(4824), 1,
      sym__special_character,
    ACTIONS(4826), 1,
      anon_sym_DQUOTE,
    ACTIONS(4830), 1,
      aux_sym_number_token1,
    ACTIONS(4832), 1,
      aux_sym_number_token2,
    ACTIONS(4834), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(4836), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(4838), 1,
      anon_sym_BQUOTE,
    ACTIONS(4840), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(4846), 1,
      sym_test_operator,
    ACTIONS(4848), 1,
      sym__brace_start,
    ACTIONS(4877), 1,
      aux_sym__simple_variable_name_token1,
    STATE(2939), 1,
      aux_sym__literal_repeat1,
    ACTIONS(4818), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(4828), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(4842), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(979), 2,
      sym_concatenation,
      aux_sym_unset_command_repeat1,
    ACTIONS(2562), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    STATE(2539), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2564), 10,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [37090] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1298), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1296), 43,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [37147] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1330), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1328), 43,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [37204] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4879), 1,
      aux_sym_concatenation_token1,
    ACTIONS(4882), 1,
      sym__concat,
    STATE(965), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1253), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1251), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [37267] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4672), 1,
      aux_sym_concatenation_token1,
    ACTIONS(4674), 1,
      sym__concat,
    STATE(959), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1263), 6,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1261), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [37330] = 22,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(4885), 1,
      sym_word,
    ACTIONS(4891), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(4894), 1,
      anon_sym_DOLLAR,
    ACTIONS(4897), 1,
      sym__special_character,
    ACTIONS(4900), 1,
      anon_sym_DQUOTE,
    ACTIONS(4906), 1,
      aux_sym_number_token1,
    ACTIONS(4909), 1,
      aux_sym_number_token2,
    ACTIONS(4912), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(4915), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(4918), 1,
      anon_sym_BQUOTE,
    ACTIONS(4921), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(4927), 1,
      sym_test_operator,
    ACTIONS(4930), 1,
      sym__brace_start,
    STATE(2776), 1,
      aux_sym__literal_repeat1,
    ACTIONS(4888), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(4903), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(4924), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(967), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    ACTIONS(2214), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    STATE(2489), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2216), 11,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [37425] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1318), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1316), 43,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [37482] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4933), 1,
      aux_sym_concatenation_token1,
    ACTIONS(4935), 1,
      sym__concat,
    STATE(978), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1263), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1261), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [37545] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4937), 1,
      aux_sym_concatenation_token1,
    ACTIONS(4940), 1,
      sym__concat,
    STATE(970), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1253), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1251), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [37608] = 22,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(4943), 1,
      sym_word,
    ACTIONS(4947), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(4949), 1,
      anon_sym_DOLLAR,
    ACTIONS(4951), 1,
      sym__special_character,
    ACTIONS(4953), 1,
      anon_sym_DQUOTE,
    ACTIONS(4957), 1,
      aux_sym_number_token1,
    ACTIONS(4959), 1,
      aux_sym_number_token2,
    ACTIONS(4961), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(4963), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(4965), 1,
      anon_sym_BQUOTE,
    ACTIONS(4967), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(4971), 1,
      sym_test_operator,
    ACTIONS(4973), 1,
      sym__brace_start,
    STATE(2776), 1,
      aux_sym__literal_repeat1,
    ACTIONS(4945), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(4955), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(4969), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(967), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    ACTIONS(2072), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    STATE(2489), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2074), 11,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [37703] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3580), 1,
      anon_sym_DQUOTE,
    ACTIONS(3610), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(3612), 1,
      anon_sym_DOLLAR,
    ACTIONS(3616), 1,
      aux_sym_number_token1,
    ACTIONS(3618), 1,
      aux_sym_number_token2,
    ACTIONS(3620), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(3622), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3624), 1,
      anon_sym_BQUOTE,
    ACTIONS(3626), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3632), 1,
      sym__brace_start,
    ACTIONS(4977), 1,
      sym__special_character,
    ACTIONS(4979), 1,
      sym_test_operator,
    STATE(2270), 1,
      aux_sym__literal_repeat1,
    ACTIONS(2096), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(3608), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(3628), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(976), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    ACTIONS(4975), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(2572), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2094), 16,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [37796] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3580), 1,
      anon_sym_DQUOTE,
    ACTIONS(3610), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(3612), 1,
      anon_sym_DOLLAR,
    ACTIONS(3616), 1,
      aux_sym_number_token1,
    ACTIONS(3618), 1,
      aux_sym_number_token2,
    ACTIONS(3620), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(3622), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3624), 1,
      anon_sym_BQUOTE,
    ACTIONS(3626), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3632), 1,
      sym__brace_start,
    ACTIONS(4977), 1,
      sym__special_character,
    ACTIONS(4979), 1,
      sym_test_operator,
    STATE(2270), 1,
      aux_sym__literal_repeat1,
    ACTIONS(2074), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(3608), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(3628), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(976), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    ACTIONS(4975), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(2572), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2072), 16,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [37889] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4933), 1,
      aux_sym_concatenation_token1,
    ACTIONS(4981), 1,
      sym__concat,
    STATE(970), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1288), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1286), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [37952] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1346), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1344), 43,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [38009] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3837), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(3840), 1,
      anon_sym_DOLLAR,
    ACTIONS(3846), 1,
      anon_sym_DQUOTE,
    ACTIONS(3849), 1,
      aux_sym_number_token1,
    ACTIONS(3852), 1,
      aux_sym_number_token2,
    ACTIONS(3855), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(3858), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3861), 1,
      anon_sym_BQUOTE,
    ACTIONS(3864), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3873), 1,
      sym__brace_start,
    ACTIONS(4986), 1,
      sym__special_character,
    ACTIONS(4989), 1,
      sym_test_operator,
    STATE(2270), 1,
      aux_sym__literal_repeat1,
    ACTIONS(2216), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(3834), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(3867), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(976), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    ACTIONS(4983), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(2572), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2214), 16,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [38102] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1358), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1356), 43,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [38159] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4933), 1,
      aux_sym_concatenation_token1,
    ACTIONS(4992), 1,
      sym__concat,
    STATE(970), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1267), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1265), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [38222] = 23,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(4994), 1,
      sym_word,
    ACTIONS(5000), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(5003), 1,
      anon_sym_DOLLAR,
    ACTIONS(5006), 1,
      sym__special_character,
    ACTIONS(5009), 1,
      anon_sym_DQUOTE,
    ACTIONS(5015), 1,
      aux_sym_number_token1,
    ACTIONS(5018), 1,
      aux_sym_number_token2,
    ACTIONS(5021), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(5024), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(5027), 1,
      anon_sym_BQUOTE,
    ACTIONS(5030), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(5036), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(5039), 1,
      sym_test_operator,
    ACTIONS(5042), 1,
      sym__brace_start,
    STATE(2939), 1,
      aux_sym__literal_repeat1,
    ACTIONS(4997), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(5012), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(5033), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(979), 2,
      sym_concatenation,
      aux_sym_unset_command_repeat1,
    ACTIONS(2440), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    STATE(2539), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2442), 10,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [38319] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1294), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1292), 43,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [38376] = 22,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(4943), 1,
      sym_word,
    ACTIONS(4947), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(4949), 1,
      anon_sym_DOLLAR,
    ACTIONS(4951), 1,
      sym__special_character,
    ACTIONS(4953), 1,
      anon_sym_DQUOTE,
    ACTIONS(4957), 1,
      aux_sym_number_token1,
    ACTIONS(4959), 1,
      aux_sym_number_token2,
    ACTIONS(4961), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(4963), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(4965), 1,
      anon_sym_BQUOTE,
    ACTIONS(4967), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(4971), 1,
      sym_test_operator,
    ACTIONS(4973), 1,
      sym__brace_start,
    STATE(2776), 1,
      aux_sym__literal_repeat1,
    ACTIONS(4945), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(4955), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(4969), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(967), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    ACTIONS(2094), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    STATE(2489), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2096), 11,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [38471] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1334), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1332), 43,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [38528] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4672), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5045), 1,
      sym__concat,
    STATE(996), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1288), 6,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1286), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [38591] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1330), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1328), 43,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [38648] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4933), 1,
      aux_sym_concatenation_token1,
    ACTIONS(4935), 1,
      sym__concat,
    STATE(974), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4532), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4530), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [38711] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4696), 1,
      sym_variable_name,
    STATE(6757), 1,
      sym_subscript,
    ACTIONS(4261), 2,
      sym_test_operator,
      sym__brace_start,
    STATE(4574), 2,
      sym_variable_assignment,
      aux_sym_variable_assignments_repeat1,
    ACTIONS(4348), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    STATE(4650), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(4247), 18,
      anon_sym_LPAREN_LPAREN,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
    ACTIONS(4253), 19,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [38780] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1318), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1316), 43,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [38837] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1302), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1300), 43,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [38894] = 30,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(707), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(709), 1,
      anon_sym_DOLLAR,
    ACTIONS(713), 1,
      anon_sym_DQUOTE,
    ACTIONS(717), 1,
      aux_sym_number_token1,
    ACTIONS(719), 1,
      aux_sym_number_token2,
    ACTIONS(721), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(723), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(725), 1,
      anon_sym_BQUOTE,
    ACTIONS(727), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(735), 1,
      sym_test_operator,
    ACTIONS(737), 1,
      sym__brace_start,
    ACTIONS(1394), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(1396), 1,
      sym_file_descriptor,
    ACTIONS(2971), 1,
      sym_word,
    ACTIONS(4705), 1,
      sym_variable_name,
    ACTIONS(5047), 1,
      sym__special_character,
    STATE(531), 1,
      sym_command_name,
    STATE(1026), 1,
      aux_sym__literal_repeat1,
    STATE(1169), 1,
      sym_concatenation,
    STATE(6771), 1,
      sym_subscript,
    ACTIONS(705), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(715), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(729), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1392), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    STATE(3127), 2,
      sym_variable_assignment,
      aux_sym_command_repeat1,
    STATE(3581), 2,
      sym_file_redirect,
      sym_herestring_redirect,
    ACTIONS(1390), 3,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
    ACTIONS(1388), 5,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    STATE(852), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [39005] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1298), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1296), 43,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [39062] = 13,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2721), 1,
      ts_builtin_sym_end,
    ACTIONS(4696), 1,
      sym_variable_name,
    ACTIONS(5051), 1,
      aux_sym_heredoc_redirect_token1,
    STATE(6757), 1,
      sym_subscript,
    ACTIONS(4253), 2,
      anon_sym_PIPE,
      anon_sym_PIPE_AMP,
    ACTIONS(4690), 2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
    ACTIONS(4692), 2,
      anon_sym_LT_LT,
      anon_sym_LT_LT_DASH,
    STATE(4574), 2,
      sym_variable_assignment,
      aux_sym_variable_assignments_repeat1,
    ACTIONS(4261), 3,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(5049), 3,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
    STATE(4650), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(4247), 28,
      anon_sym_LPAREN_LPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [39139] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4933), 1,
      aux_sym_concatenation_token1,
    ACTIONS(4935), 1,
      sym__concat,
    STATE(978), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4469), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4467), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [39202] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4933), 1,
      aux_sym_concatenation_token1,
    ACTIONS(4935), 1,
      sym__concat,
    STATE(974), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4372), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4370), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [39265] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4933), 1,
      aux_sym_concatenation_token1,
    ACTIONS(4935), 1,
      sym__concat,
    STATE(978), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4566), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4564), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [39328] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4933), 1,
      aux_sym_concatenation_token1,
    ACTIONS(4935), 1,
      sym__concat,
    STATE(974), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4554), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4552), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [39391] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5053), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5056), 1,
      sym__concat,
    STATE(996), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1253), 6,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1251), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [39454] = 30,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(314), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(316), 1,
      anon_sym_DOLLAR,
    ACTIONS(320), 1,
      anon_sym_DQUOTE,
    ACTIONS(324), 1,
      aux_sym_number_token1,
    ACTIONS(326), 1,
      aux_sym_number_token2,
    ACTIONS(328), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(330), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(332), 1,
      anon_sym_BQUOTE,
    ACTIONS(334), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(342), 1,
      sym_test_operator,
    ACTIONS(344), 1,
      sym__brace_start,
    ACTIONS(1394), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(1396), 1,
      sym_file_descriptor,
    ACTIONS(2723), 1,
      sym_word,
    ACTIONS(4705), 1,
      sym_variable_name,
    ACTIONS(4873), 1,
      sym__special_character,
    STATE(545), 1,
      sym_command_name,
    STATE(1267), 1,
      aux_sym__literal_repeat1,
    STATE(1376), 1,
      sym_concatenation,
    STATE(6771), 1,
      sym_subscript,
    ACTIONS(312), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(322), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(336), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1392), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    STATE(3127), 2,
      sym_variable_assignment,
      aux_sym_command_repeat1,
    STATE(3581), 2,
      sym_file_redirect,
      sym_herestring_redirect,
    ACTIONS(1390), 3,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
    ACTIONS(1388), 5,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    STATE(1078), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [39565] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1350), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1348), 43,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [39622] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1253), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1251), 43,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [39679] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4933), 1,
      aux_sym_concatenation_token1,
    ACTIONS(4935), 1,
      sym__concat,
    STATE(974), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4465), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4463), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [39742] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4933), 1,
      aux_sym_concatenation_token1,
    ACTIONS(4935), 1,
      sym__concat,
    STATE(978), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4482), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4480), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [39805] = 13,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2719), 1,
      anon_sym_RPAREN,
    ACTIONS(4384), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4811), 1,
      sym_variable_name,
    STATE(6786), 1,
      sym_subscript,
    ACTIONS(4253), 2,
      anon_sym_PIPE,
      anon_sym_PIPE_AMP,
    ACTIONS(4692), 2,
      anon_sym_LT_LT,
      anon_sym_LT_LT_DASH,
    ACTIONS(4809), 2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
    STATE(4534), 2,
      sym_variable_assignment,
      aux_sym_variable_assignments_repeat1,
    ACTIONS(4261), 3,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(4378), 3,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
    STATE(4746), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(4247), 28,
      anon_sym_LPAREN_LPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [39882] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4933), 1,
      aux_sym_concatenation_token1,
    ACTIONS(4935), 1,
      sym__concat,
    STATE(978), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4548), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4546), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [39945] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1326), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1324), 43,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [40002] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1294), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1292), 43,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [40059] = 10,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4696), 1,
      sym_variable_name,
    STATE(6757), 1,
      sym_subscript,
    ACTIONS(4253), 2,
      anon_sym_PIPE,
      anon_sym_PIPE_AMP,
    ACTIONS(4272), 2,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    STATE(4574), 2,
      sym_variable_assignment,
      aux_sym_variable_assignments_repeat1,
    ACTIONS(4261), 3,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
    STATE(4650), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(4270), 7,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_SEMI_SEMI,
      anon_sym_LT_LT_DASH,
    ACTIONS(4247), 28,
      anon_sym_LPAREN_LPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [40130] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4672), 1,
      aux_sym_concatenation_token1,
    ACTIONS(4674), 1,
      sym__concat,
    STATE(959), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(278), 6,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(241), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [40193] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1334), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1332), 43,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [40250] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1302), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1300), 43,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [40307] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5061), 1,
      anon_sym_DQUOTE,
    ACTIONS(5065), 1,
      sym_variable_name,
    STATE(2712), 1,
      sym_string,
    ACTIONS(5063), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1241), 3,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(5059), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1239), 31,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [40373] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5071), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5073), 1,
      sym__concat,
    STATE(1086), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5069), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5067), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [40435] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4699), 1,
      aux_sym_concatenation_token1,
    ACTIONS(4701), 1,
      sym__concat,
    ACTIONS(5075), 1,
      anon_sym_LPAREN,
    STATE(1054), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1281), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1271), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [40499] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1302), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1300), 42,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [40555] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5078), 1,
      sym__special_character,
    STATE(1095), 1,
      aux_sym__literal_repeat1,
    ACTIONS(4482), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4480), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [40615] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1294), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1292), 42,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [40671] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1298), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1296), 42,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [40727] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1310), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1308), 42,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [40783] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1306), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1304), 42,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [40839] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1310), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1308), 42,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [40895] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1314), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1312), 42,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [40951] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5080), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5082), 1,
      sym__concat,
    STATE(1032), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4548), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4546), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [41013] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1342), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1340), 42,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [41069] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5080), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5082), 1,
      sym__concat,
    STATE(1040), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4532), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4530), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [41131] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1235), 1,
      sym_test_operator,
    ACTIONS(5086), 1,
      anon_sym_DQUOTE,
    ACTIONS(5090), 1,
      sym_variable_name,
    STATE(2442), 1,
      sym_string,
    ACTIONS(5088), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(5084), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1227), 33,
      anon_sym_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_EQ_TILDE,
      anon_sym_COLON,
  [41197] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1241), 1,
      sym_test_operator,
    ACTIONS(5086), 1,
      anon_sym_DQUOTE,
    ACTIONS(5090), 1,
      sym_variable_name,
    STATE(2442), 1,
      sym_string,
    ACTIONS(5088), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(5084), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1239), 33,
      anon_sym_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_EQ_TILDE,
      anon_sym_COLON,
  [41263] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5092), 1,
      sym__special_character,
    STATE(1097), 1,
      aux_sym__literal_repeat1,
    ACTIONS(278), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(241), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [41323] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1334), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1332), 42,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [41379] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5094), 1,
      sym_variable_name,
    STATE(6758), 1,
      sym_subscript,
    ACTIONS(4261), 2,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(4348), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    STATE(4568), 2,
      sym_variable_assignment,
      aux_sym_variable_assignments_repeat1,
    STATE(4697), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(4247), 17,
      anon_sym_LPAREN_LPAREN,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
    ACTIONS(4253), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_BQUOTE,
  [41447] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1302), 7,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1300), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [41503] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5101), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5103), 1,
      sym_variable_name,
    STATE(6727), 1,
      sym_subscript,
    ACTIONS(4253), 2,
      anon_sym_PIPE,
      anon_sym_PIPE_AMP,
    ACTIONS(4692), 2,
      anon_sym_LT_LT,
      anon_sym_LT_LT_DASH,
    ACTIONS(5099), 2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
    STATE(4787), 2,
      sym_variable_assignment,
      aux_sym_variable_assignments_repeat1,
    ACTIONS(4261), 3,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(5097), 3,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
    STATE(4986), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(4247), 28,
      anon_sym_LPAREN_LPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [41577] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1354), 7,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1352), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [41633] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5080), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5106), 1,
      sym__concat,
    STATE(1065), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1267), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1265), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [41695] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1294), 7,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1292), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [41751] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1298), 7,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1296), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [41807] = 22,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(5108), 1,
      sym_word,
    ACTIONS(5112), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(5114), 1,
      anon_sym_DOLLAR,
    ACTIONS(5116), 1,
      sym__special_character,
    ACTIONS(5118), 1,
      anon_sym_DQUOTE,
    ACTIONS(5122), 1,
      aux_sym_number_token1,
    ACTIONS(5124), 1,
      aux_sym_number_token2,
    ACTIONS(5126), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(5128), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(5130), 1,
      anon_sym_BQUOTE,
    ACTIONS(5132), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(5136), 1,
      sym_test_operator,
    ACTIONS(5138), 1,
      sym__brace_start,
    STATE(3114), 1,
      aux_sym__literal_repeat1,
    ACTIONS(5110), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(5120), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(5134), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(1075), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    ACTIONS(2094), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    STATE(2613), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2096), 10,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [41901] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5103), 1,
      sym_variable_name,
    ACTIONS(5142), 1,
      aux_sym_heredoc_redirect_token1,
    STATE(6727), 1,
      sym_subscript,
    ACTIONS(4253), 2,
      anon_sym_PIPE,
      anon_sym_PIPE_AMP,
    ACTIONS(4692), 2,
      anon_sym_LT_LT,
      anon_sym_LT_LT_DASH,
    ACTIONS(5099), 2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
    STATE(4787), 2,
      sym_variable_assignment,
      aux_sym_variable_assignments_repeat1,
    ACTIONS(4261), 3,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(5140), 3,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
    STATE(4986), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(4247), 28,
      anon_sym_LPAREN_LPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [41975] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1253), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1251), 42,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [42031] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1326), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1324), 42,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [42087] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5144), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5146), 1,
      sym__concat,
    STATE(1131), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4465), 6,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4463), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [42149] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5080), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5148), 1,
      sym__concat,
    STATE(1065), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1288), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1286), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [42211] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5061), 1,
      anon_sym_DQUOTE,
    ACTIONS(5065), 1,
      sym_variable_name,
    STATE(2712), 1,
      sym_string,
    ACTIONS(5063), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1235), 3,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(5059), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1227), 31,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [42277] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5144), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5146), 1,
      sym__concat,
    STATE(1123), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4482), 6,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4480), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [42339] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1310), 7,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1308), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [42395] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1306), 7,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1304), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [42451] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5078), 1,
      sym__special_character,
    STATE(1095), 1,
      aux_sym__literal_repeat1,
    ACTIONS(4548), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4546), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [42511] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1310), 7,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1308), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [42567] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1314), 7,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1312), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [42623] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5152), 1,
      anon_sym_DQUOTE,
    ACTIONS(5156), 1,
      sym_variable_name,
    STATE(2566), 1,
      sym_string,
    ACTIONS(5154), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1235), 3,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(5150), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1227), 31,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [42689] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1342), 7,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1340), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [42745] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5152), 1,
      anon_sym_DQUOTE,
    ACTIONS(5156), 1,
      sym_variable_name,
    STATE(2566), 1,
      sym_string,
    ACTIONS(5154), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1241), 3,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(5150), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1239), 31,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [42811] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(262), 1,
      anon_sym_DQUOTE,
    ACTIONS(1235), 1,
      sym_test_operator,
    ACTIONS(5162), 1,
      sym_variable_name,
    STATE(2734), 1,
      sym_string,
    ACTIONS(5160), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(5158), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1227), 33,
      anon_sym_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_RPAREN,
      anon_sym_EQ_TILDE,
  [42877] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1263), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1261), 43,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [42933] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(262), 1,
      anon_sym_DQUOTE,
    ACTIONS(1241), 1,
      sym_test_operator,
    ACTIONS(5162), 1,
      sym_variable_name,
    STATE(2734), 1,
      sym_string,
    ACTIONS(5160), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(5158), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1239), 33,
      anon_sym_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_RPAREN,
      anon_sym_EQ_TILDE,
  [42999] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4699), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5164), 1,
      sym__concat,
    STATE(965), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1288), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1286), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [43061] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4699), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5166), 1,
      sym__concat,
    STATE(965), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1267), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1265), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [43123] = 22,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(5108), 1,
      sym_word,
    ACTIONS(5112), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(5114), 1,
      anon_sym_DOLLAR,
    ACTIONS(5116), 1,
      sym__special_character,
    ACTIONS(5118), 1,
      anon_sym_DQUOTE,
    ACTIONS(5122), 1,
      aux_sym_number_token1,
    ACTIONS(5124), 1,
      aux_sym_number_token2,
    ACTIONS(5126), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(5128), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(5130), 1,
      anon_sym_BQUOTE,
    ACTIONS(5132), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(5136), 1,
      sym_test_operator,
    ACTIONS(5138), 1,
      sym__brace_start,
    STATE(3114), 1,
      aux_sym__literal_repeat1,
    ACTIONS(5110), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(5120), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(5134), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(1075), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    ACTIONS(2072), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    STATE(2613), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2074), 10,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [43217] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1338), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1336), 42,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [43273] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1330), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1328), 42,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [43329] = 10,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4272), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5094), 1,
      sym_variable_name,
    STATE(6758), 1,
      sym_subscript,
    ACTIONS(4253), 2,
      anon_sym_PIPE,
      anon_sym_PIPE_AMP,
    STATE(4568), 2,
      sym_variable_assignment,
      aux_sym_variable_assignments_repeat1,
    ACTIONS(4261), 3,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
    STATE(4697), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(4270), 7,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_SEMI_SEMI,
      anon_sym_LT_LT_DASH,
    ACTIONS(4247), 28,
      anon_sym_LPAREN_LPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [43399] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1322), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1320), 42,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [43455] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1350), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1348), 42,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [43511] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1346), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1344), 42,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [43567] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1358), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1356), 42,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [43623] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1334), 7,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1332), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [43679] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5168), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5171), 1,
      sym__concat,
    STATE(1065), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1253), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1251), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [43741] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5144), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5146), 1,
      sym__concat,
    STATE(1123), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4469), 6,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4467), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [43803] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5071), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5073), 1,
      sym__concat,
    STATE(1087), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5176), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5174), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [43865] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5071), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5073), 1,
      sym__concat,
    STATE(1086), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5180), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5178), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [43927] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5144), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5146), 1,
      sym__concat,
    STATE(1131), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4372), 6,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4370), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [43989] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1235), 1,
      sym_test_operator,
    ACTIONS(4663), 1,
      anon_sym_DQUOTE,
    ACTIONS(4667), 1,
      sym_variable_name,
    STATE(2716), 1,
      sym_string,
    ACTIONS(4665), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(4661), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1227), 33,
      anon_sym_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
  [44055] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1241), 1,
      sym_test_operator,
    ACTIONS(4663), 1,
      anon_sym_DQUOTE,
    ACTIONS(4667), 1,
      sym_variable_name,
    STATE(2716), 1,
      sym_string,
    ACTIONS(4665), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(4661), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1239), 33,
      anon_sym_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
  [44121] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5144), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5146), 1,
      sym__concat,
    STATE(1123), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4566), 6,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4564), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [44183] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5182), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5184), 1,
      sym__concat,
    STATE(1079), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1267), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1265), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [44245] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5182), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5186), 1,
      sym__concat,
    STATE(1079), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1288), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1286), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [44307] = 22,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(5188), 1,
      sym_word,
    ACTIONS(5194), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(5197), 1,
      anon_sym_DOLLAR,
    ACTIONS(5200), 1,
      sym__special_character,
    ACTIONS(5203), 1,
      anon_sym_DQUOTE,
    ACTIONS(5209), 1,
      aux_sym_number_token1,
    ACTIONS(5212), 1,
      aux_sym_number_token2,
    ACTIONS(5215), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(5218), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(5221), 1,
      anon_sym_BQUOTE,
    ACTIONS(5224), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(5230), 1,
      sym_test_operator,
    ACTIONS(5233), 1,
      sym__brace_start,
    STATE(3114), 1,
      aux_sym__literal_repeat1,
    ACTIONS(5191), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(5206), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(5227), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(1075), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    ACTIONS(2214), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    STATE(2613), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2216), 10,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [44401] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5144), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5146), 1,
      sym__concat,
    STATE(1131), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4554), 6,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4552), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [44463] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4699), 1,
      aux_sym_concatenation_token1,
    ACTIONS(4701), 1,
      sym__concat,
    STATE(1055), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(278), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(241), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [44525] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4699), 1,
      aux_sym_concatenation_token1,
    ACTIONS(4701), 1,
      sym__concat,
    STATE(1054), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1281), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1271), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [44587] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5236), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5239), 1,
      sym__concat,
    STATE(1079), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1253), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1251), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [44649] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4699), 1,
      aux_sym_concatenation_token1,
    ACTIONS(4701), 1,
      sym__concat,
    ACTIONS(5242), 1,
      anon_sym_LPAREN,
    STATE(1054), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1281), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1271), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [44713] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5182), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5245), 1,
      sym__concat,
    STATE(1074), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(2074), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(2072), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [44775] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1281), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1271), 43,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [44831] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1253), 7,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1251), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [44887] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1326), 7,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1324), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [44943] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1318), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1316), 42,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [44999] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5071), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5247), 1,
      sym__concat,
    STATE(1088), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1267), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1265), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [45061] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5071), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5249), 1,
      sym__concat,
    STATE(1088), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1288), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1286), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [45123] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5251), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5254), 1,
      sym__concat,
    STATE(1088), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1253), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1251), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [45185] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5182), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5245), 1,
      sym__concat,
    STATE(1073), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1263), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1261), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [45247] = 22,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(5257), 1,
      sym_word,
    ACTIONS(5261), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(5263), 1,
      anon_sym_DOLLAR,
    ACTIONS(5265), 1,
      sym__special_character,
    ACTIONS(5267), 1,
      anon_sym_DQUOTE,
    ACTIONS(5271), 1,
      aux_sym_number_token1,
    ACTIONS(5273), 1,
      aux_sym_number_token2,
    ACTIONS(5275), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(5277), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(5279), 1,
      anon_sym_BQUOTE,
    ACTIONS(5281), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(5285), 1,
      sym_test_operator,
    ACTIONS(5287), 1,
      sym__brace_start,
    STATE(5437), 1,
      aux_sym__literal_repeat1,
    STATE(5604), 1,
      sym_concatenation,
    ACTIONS(5259), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(5269), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(5283), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(2494), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    STATE(5378), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2496), 11,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [45341] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5080), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5082), 1,
      sym__concat,
    STATE(1032), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1263), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1261), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [45403] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5071), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5073), 1,
      sym__concat,
    STATE(1086), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1263), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1261), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [45465] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5144), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5146), 1,
      sym__concat,
    STATE(1131), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4532), 6,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4530), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [45527] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5144), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5146), 1,
      sym__concat,
    STATE(1123), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4548), 6,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4546), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [45589] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5289), 1,
      sym__special_character,
    STATE(1095), 1,
      aux_sym__literal_repeat1,
    ACTIONS(1364), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1362), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [45649] = 10,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4272), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5103), 1,
      sym_variable_name,
    STATE(6727), 1,
      sym_subscript,
    ACTIONS(4253), 2,
      anon_sym_PIPE,
      anon_sym_PIPE_AMP,
    STATE(4787), 2,
      sym_variable_assignment,
      aux_sym_variable_assignments_repeat1,
    ACTIONS(4261), 3,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
    STATE(4986), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(4270), 7,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_SEMI_SEMI,
      anon_sym_LT_LT_DASH,
    ACTIONS(4247), 28,
      anon_sym_LPAREN_LPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [45719] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5292), 1,
      sym__special_character,
    STATE(1097), 1,
      aux_sym__literal_repeat1,
    ACTIONS(1364), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1362), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [45779] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5103), 1,
      sym_variable_name,
    STATE(6727), 1,
      sym_subscript,
    ACTIONS(4261), 2,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(4348), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    STATE(4787), 2,
      sym_variable_assignment,
      aux_sym_variable_assignments_repeat1,
    STATE(4986), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(4247), 18,
      anon_sym_LPAREN_LPAREN,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
    ACTIONS(4253), 19,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [45847] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1310), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1308), 42,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [45903] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5182), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5245), 1,
      sym__concat,
    STATE(1074), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(2096), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(2094), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [45965] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1306), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1304), 42,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [46021] = 22,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(5261), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(5263), 1,
      anon_sym_DOLLAR,
    ACTIONS(5265), 1,
      sym__special_character,
    ACTIONS(5267), 1,
      anon_sym_DQUOTE,
    ACTIONS(5271), 1,
      aux_sym_number_token1,
    ACTIONS(5273), 1,
      aux_sym_number_token2,
    ACTIONS(5275), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(5277), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(5279), 1,
      anon_sym_BQUOTE,
    ACTIONS(5281), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(5287), 1,
      sym__brace_start,
    ACTIONS(5295), 1,
      sym_word,
    ACTIONS(5299), 1,
      sym_test_operator,
    STATE(5494), 1,
      aux_sym__literal_repeat1,
    STATE(5601), 1,
      sym_concatenation,
    ACTIONS(5259), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(5283), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(5297), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(2498), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    STATE(5352), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2500), 11,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [46115] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1310), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1308), 42,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [46171] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1318), 7,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1316), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [46227] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4699), 1,
      aux_sym_concatenation_token1,
    ACTIONS(4701), 1,
      sym__concat,
    STATE(1055), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1263), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1261), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [46289] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5080), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5082), 1,
      sym__concat,
    STATE(1032), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4469), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4467), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [46351] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5080), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5082), 1,
      sym__concat,
    STATE(1040), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4372), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4370), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [46413] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5080), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5082), 1,
      sym__concat,
    STATE(1032), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4566), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4564), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [46475] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5080), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5082), 1,
      sym__concat,
    STATE(1040), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4554), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4552), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [46537] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4078), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(4080), 1,
      anon_sym_DOLLAR,
    ACTIONS(4084), 1,
      anon_sym_DQUOTE,
    ACTIONS(4086), 1,
      aux_sym_number_token1,
    ACTIONS(4088), 1,
      aux_sym_number_token2,
    ACTIONS(4090), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(4092), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(4094), 1,
      anon_sym_BQUOTE,
    ACTIONS(4096), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(4102), 1,
      sym__brace_start,
    ACTIONS(5303), 1,
      sym__special_character,
    ACTIONS(5305), 1,
      sym_test_operator,
    STATE(4546), 1,
      aux_sym__literal_repeat1,
    STATE(4907), 1,
      sym_concatenation,
    ACTIONS(2500), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4076), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(4098), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(5301), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(5293), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2498), 16,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [46629] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4078), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(4080), 1,
      anon_sym_DOLLAR,
    ACTIONS(4084), 1,
      anon_sym_DQUOTE,
    ACTIONS(4086), 1,
      aux_sym_number_token1,
    ACTIONS(4088), 1,
      aux_sym_number_token2,
    ACTIONS(4090), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(4092), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(4094), 1,
      anon_sym_BQUOTE,
    ACTIONS(4096), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(4102), 1,
      sym__brace_start,
    ACTIONS(5303), 1,
      sym__special_character,
    ACTIONS(5309), 1,
      sym_test_operator,
    STATE(4521), 1,
      aux_sym__literal_repeat1,
    STATE(5023), 1,
      sym_concatenation,
    ACTIONS(2496), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4076), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(4098), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(5307), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(5398), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2494), 16,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [46721] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1338), 7,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1336), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [46777] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1330), 7,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1328), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [46833] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1241), 1,
      sym_test_operator,
    ACTIONS(5086), 1,
      anon_sym_DQUOTE,
    ACTIONS(5090), 1,
      sym_variable_name,
    STATE(2442), 1,
      sym_string,
    ACTIONS(5088), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(5084), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1239), 33,
      anon_sym_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_RBRACK_RBRACK,
      anon_sym_EQ_TILDE,
  [46899] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1235), 1,
      sym_test_operator,
    ACTIONS(5086), 1,
      anon_sym_DQUOTE,
    ACTIONS(5090), 1,
      sym_variable_name,
    STATE(2442), 1,
      sym_string,
    ACTIONS(5088), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(5084), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1227), 33,
      anon_sym_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_RBRACK_RBRACK,
      anon_sym_EQ_TILDE,
  [46965] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1322), 7,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1320), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [47021] = 22,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(5311), 1,
      sym_word,
    ACTIONS(5315), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(5317), 1,
      anon_sym_DOLLAR,
    ACTIONS(5319), 1,
      sym__special_character,
    ACTIONS(5321), 1,
      anon_sym_DQUOTE,
    ACTIONS(5325), 1,
      aux_sym_number_token1,
    ACTIONS(5327), 1,
      aux_sym_number_token2,
    ACTIONS(5329), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(5331), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(5333), 1,
      anon_sym_BQUOTE,
    ACTIONS(5335), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(5339), 1,
      sym_test_operator,
    ACTIONS(5341), 1,
      sym__brace_start,
    STATE(5424), 1,
      aux_sym__literal_repeat1,
    STATE(5575), 1,
      sym_concatenation,
    ACTIONS(5313), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(5323), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(5337), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(2494), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    STATE(5379), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2496), 11,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [47115] = 22,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(5315), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(5317), 1,
      anon_sym_DOLLAR,
    ACTIONS(5319), 1,
      sym__special_character,
    ACTIONS(5321), 1,
      anon_sym_DQUOTE,
    ACTIONS(5325), 1,
      aux_sym_number_token1,
    ACTIONS(5327), 1,
      aux_sym_number_token2,
    ACTIONS(5329), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(5331), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(5333), 1,
      anon_sym_BQUOTE,
    ACTIONS(5335), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(5341), 1,
      sym__brace_start,
    ACTIONS(5343), 1,
      sym_word,
    ACTIONS(5347), 1,
      sym_test_operator,
    STATE(5496), 1,
      aux_sym__literal_repeat1,
    STATE(5563), 1,
      sym_concatenation,
    ACTIONS(5313), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(5337), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(5345), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(2498), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    STATE(5331), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2500), 11,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [47209] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1326), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1324), 42,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [47265] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1253), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1251), 42,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [47321] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5182), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5245), 1,
      sym__concat,
    STATE(1073), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4469), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4467), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [47383] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5080), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5082), 1,
      sym__concat,
    STATE(1032), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4482), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4480), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [47445] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5144), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5349), 1,
      sym__concat,
    STATE(1140), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1267), 6,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1265), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [47507] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5080), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5082), 1,
      sym__concat,
    STATE(1040), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4465), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4463), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [47569] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1350), 7,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1348), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [47625] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1346), 7,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1344), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [47681] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5182), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5245), 1,
      sym__concat,
    STATE(1074), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4372), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4370), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [47743] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1358), 7,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1356), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [47799] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1354), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1352), 42,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [47855] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1350), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1348), 42,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [47911] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5144), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5351), 1,
      sym__concat,
    STATE(1140), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1288), 6,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1286), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [47973] = 13,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2719), 1,
      anon_sym_BQUOTE,
    ACTIONS(5094), 1,
      sym_variable_name,
    ACTIONS(5357), 1,
      aux_sym_heredoc_redirect_token1,
    STATE(6758), 1,
      sym_subscript,
    ACTIONS(4253), 2,
      anon_sym_PIPE,
      anon_sym_PIPE_AMP,
    ACTIONS(4692), 2,
      anon_sym_LT_LT,
      anon_sym_LT_LT_DASH,
    ACTIONS(5355), 2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
    STATE(4568), 2,
      sym_variable_assignment,
      aux_sym_variable_assignments_repeat1,
    ACTIONS(4261), 3,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(5353), 3,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
    STATE(4697), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(4247), 27,
      anon_sym_LPAREN_LPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [48049] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1298), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1296), 42,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [48105] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1241), 1,
      sym_test_operator,
    ACTIONS(5086), 1,
      anon_sym_DQUOTE,
    ACTIONS(5090), 1,
      sym_variable_name,
    STATE(2442), 1,
      sym_string,
    ACTIONS(5088), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(5084), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1239), 33,
      anon_sym_RPAREN_RPAREN,
      anon_sym_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_EQ_TILDE,
  [48171] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5361), 1,
      anon_sym_DQUOTE,
    ACTIONS(5365), 1,
      sym_variable_name,
    STATE(2850), 1,
      sym_string,
    ACTIONS(5363), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1241), 3,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(5359), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1239), 31,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [48237] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1314), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1312), 42,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [48293] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1342), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1340), 42,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [48349] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1338), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1336), 42,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [48405] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1322), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1320), 42,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [48461] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5367), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5370), 1,
      sym__concat,
    STATE(1140), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1253), 6,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1251), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [48523] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5361), 1,
      anon_sym_DQUOTE,
    ACTIONS(5365), 1,
      sym_variable_name,
    STATE(2850), 1,
      sym_string,
    ACTIONS(5363), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1235), 3,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(5359), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1227), 31,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [48589] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1302), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1300), 42,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [48645] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1334), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1332), 42,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [48701] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5144), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5146), 1,
      sym__concat,
    STATE(1123), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1263), 6,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1261), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [48763] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1294), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1292), 42,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [48819] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5078), 1,
      sym__special_character,
    STATE(1095), 1,
      aux_sym__literal_repeat1,
    ACTIONS(4566), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4564), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [48879] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5182), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5245), 1,
      sym__concat,
    STATE(1073), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5069), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5067), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [48941] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5182), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5245), 1,
      sym__concat,
    STATE(1074), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5375), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5373), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [49003] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5078), 1,
      sym__special_character,
    STATE(1095), 1,
      aux_sym__literal_repeat1,
    ACTIONS(4469), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4467), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [49063] = 13,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4257), 1,
      anon_sym_BQUOTE,
    ACTIONS(5094), 1,
      sym_variable_name,
    ACTIONS(5379), 1,
      aux_sym_heredoc_redirect_token1,
    STATE(6758), 1,
      sym_subscript,
    ACTIONS(4253), 2,
      anon_sym_PIPE,
      anon_sym_PIPE_AMP,
    ACTIONS(4692), 2,
      anon_sym_LT_LT,
      anon_sym_LT_LT_DASH,
    ACTIONS(5355), 2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
    STATE(4568), 2,
      sym_variable_assignment,
      aux_sym_variable_assignments_repeat1,
    ACTIONS(4261), 3,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(5377), 3,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
    STATE(4697), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(4247), 27,
      anon_sym_LPAREN_LPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [49139] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5103), 1,
      sym_variable_name,
    ACTIONS(5383), 1,
      aux_sym_heredoc_redirect_token1,
    STATE(6727), 1,
      sym_subscript,
    ACTIONS(4253), 2,
      anon_sym_PIPE,
      anon_sym_PIPE_AMP,
    ACTIONS(4692), 2,
      anon_sym_LT_LT,
      anon_sym_LT_LT_DASH,
    ACTIONS(5099), 2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
    STATE(4787), 2,
      sym_variable_assignment,
      aux_sym_variable_assignments_repeat1,
    ACTIONS(4261), 3,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(5381), 3,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
    STATE(4986), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(4247), 28,
      anon_sym_LPAREN_LPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [49213] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1354), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1352), 42,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [49269] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5103), 1,
      sym_variable_name,
    ACTIONS(5387), 1,
      aux_sym_heredoc_redirect_token1,
    STATE(6727), 1,
      sym_subscript,
    ACTIONS(4253), 2,
      anon_sym_PIPE,
      anon_sym_PIPE_AMP,
    ACTIONS(4692), 2,
      anon_sym_LT_LT,
      anon_sym_LT_LT_DASH,
    ACTIONS(5099), 2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
    STATE(4787), 2,
      sym_variable_assignment,
      aux_sym_variable_assignments_repeat1,
    ACTIONS(4261), 3,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(5385), 3,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
    STATE(4986), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(4247), 28,
      anon_sym_LPAREN_LPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [49343] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1235), 1,
      sym_test_operator,
    ACTIONS(5086), 1,
      anon_sym_DQUOTE,
    ACTIONS(5090), 1,
      sym_variable_name,
    STATE(2442), 1,
      sym_string,
    ACTIONS(5088), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(5084), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1227), 33,
      anon_sym_RPAREN_RPAREN,
      anon_sym_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_EQ_TILDE,
  [49409] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1318), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1316), 42,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [49465] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5071), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5073), 1,
      sym__concat,
    STATE(1087), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5375), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5373), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [49527] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1330), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1328), 42,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [49583] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5182), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5245), 1,
      sym__concat,
    STATE(1074), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5391), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5389), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [49645] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1346), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1344), 42,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [49701] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5182), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5245), 1,
      sym__concat,
    STATE(1073), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4566), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4564), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [49763] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1358), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1356), 42,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [49819] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5182), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5245), 1,
      sym__concat,
    STATE(1074), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4554), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4552), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [49881] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5182), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5245), 1,
      sym__concat,
    STATE(1073), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5395), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5393), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [49943] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1241), 1,
      sym_test_operator,
    ACTIONS(5086), 1,
      anon_sym_DQUOTE,
    ACTIONS(5090), 1,
      sym_variable_name,
    STATE(2442), 1,
      sym_string,
    ACTIONS(5088), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(5084), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1239), 32,
      anon_sym_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_EQ_TILDE,
  [50008] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1342), 6,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1340), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [50063] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5397), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5399), 1,
      sym__concat,
    STATE(1257), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5375), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5373), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [50124] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5397), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5399), 1,
      sym__concat,
    STATE(1251), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5069), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5067), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [50185] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4465), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4463), 42,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [50240] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1281), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1271), 42,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [50295] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5397), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5399), 1,
      sym__concat,
    STATE(1251), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5180), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5178), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [50356] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5397), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5399), 1,
      sym__concat,
    STATE(1257), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5176), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5174), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [50417] = 22,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(5401), 1,
      sym_word,
    ACTIONS(5405), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(5407), 1,
      anon_sym_DOLLAR,
    ACTIONS(5409), 1,
      sym__special_character,
    ACTIONS(5411), 1,
      anon_sym_DQUOTE,
    ACTIONS(5415), 1,
      aux_sym_number_token1,
    ACTIONS(5417), 1,
      aux_sym_number_token2,
    ACTIONS(5419), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(5421), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(5423), 1,
      anon_sym_BQUOTE,
    ACTIONS(5425), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(5429), 1,
      sym_test_operator,
    ACTIONS(5431), 1,
      sym__brace_start,
    STATE(5532), 1,
      aux_sym__literal_repeat1,
    STATE(5692), 1,
      sym_concatenation,
    ACTIONS(5403), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(5413), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(5427), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(2498), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    STATE(5418), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2500), 10,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [50510] = 9,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(5437), 1,
      anon_sym_DQUOTE,
    ACTIONS(5439), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(5441), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(5443), 1,
      anon_sym_BQUOTE,
    ACTIONS(5445), 1,
      anon_sym_DOLLAR_BQUOTE,
    STATE(2559), 3,
      sym_string,
      sym_expansion,
      sym_command_substitution,
    ACTIONS(5435), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(5433), 25,
      sym_test_operator,
      anon_sym_RPAREN_RPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      anon_sym_COLON,
  [50577] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5080), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5082), 1,
      sym__concat,
    STATE(1177), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1263), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1261), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [50638] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5447), 1,
      sym__special_character,
    STATE(1295), 1,
      aux_sym__literal_repeat1,
    ACTIONS(4548), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4546), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [50697] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1350), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1348), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [50752] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5080), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5449), 1,
      sym__concat,
    STATE(1065), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1267), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1265), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [50813] = 22,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(5405), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(5407), 1,
      anon_sym_DOLLAR,
    ACTIONS(5409), 1,
      sym__special_character,
    ACTIONS(5411), 1,
      anon_sym_DQUOTE,
    ACTIONS(5415), 1,
      aux_sym_number_token1,
    ACTIONS(5417), 1,
      aux_sym_number_token2,
    ACTIONS(5419), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(5421), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(5423), 1,
      anon_sym_BQUOTE,
    ACTIONS(5425), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(5431), 1,
      sym__brace_start,
    ACTIONS(5451), 1,
      sym_word,
    ACTIONS(5455), 1,
      sym_test_operator,
    STATE(5519), 1,
      aux_sym__literal_repeat1,
    STATE(5619), 1,
      sym_concatenation,
    ACTIONS(5403), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(5427), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(5453), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(2494), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    STATE(5427), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2496), 10,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [50906] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5457), 1,
      sym__special_character,
    STATE(1179), 1,
      aux_sym__literal_repeat1,
    ACTIONS(1364), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1362), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [50965] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5447), 1,
      sym__special_character,
    STATE(1295), 1,
      aux_sym__literal_repeat1,
    ACTIONS(4482), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4480), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [51024] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5462), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5460), 42,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [51079] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4532), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4530), 42,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [51134] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1310), 7,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1308), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [51189] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5464), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5466), 1,
      sym__concat,
    STATE(1249), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(2074), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(2072), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [51250] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1306), 7,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1304), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [51305] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1338), 7,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1336), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [51360] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1310), 7,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1308), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [51415] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1326), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1324), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [51470] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1253), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1251), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [51525] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5472), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5474), 1,
      sym__concat,
    STATE(1214), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5470), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5468), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [51586] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5472), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5474), 1,
      sym__concat,
    STATE(1211), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5478), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5476), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [51647] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5464), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5466), 1,
      sym__concat,
    STATE(1249), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5375), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5373), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [51708] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5464), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5466), 1,
      sym__concat,
    STATE(1245), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5069), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5067), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [51769] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1346), 7,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1344), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [51824] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1358), 7,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1356), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [51879] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1298), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1296), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [51934] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1235), 1,
      sym_test_operator,
    ACTIONS(5086), 1,
      anon_sym_DQUOTE,
    ACTIONS(5090), 1,
      sym_variable_name,
    STATE(2442), 1,
      sym_string,
    ACTIONS(5088), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(5084), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1227), 32,
      anon_sym_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_EQ_TILDE,
  [51999] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1326), 6,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1324), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [52054] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1253), 6,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1251), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [52109] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1334), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1332), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [52164] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5447), 1,
      sym__special_character,
    STATE(1295), 1,
      aux_sym__literal_repeat1,
    ACTIONS(4566), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4564), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [52223] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5447), 1,
      sym__special_character,
    STATE(1295), 1,
      aux_sym__literal_repeat1,
    ACTIONS(4469), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4467), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [52282] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5080), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5082), 1,
      sym__concat,
    STATE(1177), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4548), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4546), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [52343] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1350), 6,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1348), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [52398] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1302), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1300), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [52453] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1326), 6,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1324), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [52508] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1253), 6,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1251), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [52563] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1298), 6,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1296), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [52618] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5480), 1,
      sym__special_character,
    STATE(1209), 1,
      aux_sym__literal_repeat1,
    ACTIONS(1364), 6,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1362), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [52677] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1302), 6,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1300), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [52732] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5472), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5483), 1,
      sym__concat,
    STATE(1217), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1267), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1265), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [52793] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5485), 1,
      sym__special_character,
    STATE(1209), 1,
      aux_sym__literal_repeat1,
    ACTIONS(278), 6,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(241), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [52852] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1350), 6,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1348), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [52907] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5472), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5487), 1,
      sym__concat,
    STATE(1217), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1288), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1286), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [52968] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5080), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5489), 1,
      sym__concat,
    STATE(1065), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1288), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1286), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [53029] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1294), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1292), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [53084] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5491), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5494), 1,
      sym__concat,
    STATE(1217), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1253), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1251), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [53145] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1334), 6,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1332), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [53200] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5080), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5082), 1,
      sym__concat,
    STATE(1215), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4532), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4530), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [53261] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1298), 6,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1296), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [53316] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1294), 6,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1292), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [53371] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1302), 6,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1300), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [53426] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1354), 7,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1352), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [53481] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1334), 6,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1332), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [53536] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1294), 6,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1292), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [53591] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5472), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5474), 1,
      sym__concat,
    STATE(1211), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1263), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1261), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [53652] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5497), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5499), 1,
      sym__concat,
    STATE(1230), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1267), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1265), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [53713] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5497), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5501), 1,
      sym__concat,
    STATE(1230), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1288), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1286), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [53774] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1263), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1261), 42,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [53829] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5503), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5506), 1,
      sym__concat,
    STATE(1230), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1253), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1251), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [53890] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1342), 7,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1340), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [53945] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1318), 6,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1316), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [54000] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1330), 6,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1328), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [54055] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5497), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5509), 1,
      sym__concat,
    STATE(1227), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1263), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1261), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [54116] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1318), 6,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1316), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [54171] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1330), 6,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1328), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [54226] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1310), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1308), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [54281] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1306), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1304), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [54336] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1310), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1308), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [54391] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1310), 6,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1308), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [54446] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1306), 6,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1304), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [54501] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1310), 6,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1308), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [54556] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5080), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5082), 1,
      sym__concat,
    STATE(1177), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4482), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4480), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [54617] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5080), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5082), 1,
      sym__concat,
    STATE(1215), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4465), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4463), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [54678] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5464), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5511), 1,
      sym__concat,
    STATE(1250), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1267), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1265), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [54739] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1358), 6,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1356), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [54794] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1346), 6,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1344), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [54849] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1330), 7,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1328), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [54904] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5464), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5513), 1,
      sym__concat,
    STATE(1250), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1288), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1286), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [54965] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5515), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5518), 1,
      sym__concat,
    STATE(1250), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1253), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1251), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [55026] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5397), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5521), 1,
      sym__concat,
    STATE(1262), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1267), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1265), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [55087] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1358), 6,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1356), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [55142] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1346), 6,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1344), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [55197] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1310), 6,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1308), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [55252] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1306), 6,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1304), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [55307] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1310), 6,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1308), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [55362] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5397), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5523), 1,
      sym__concat,
    STATE(1262), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1288), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1286), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [55423] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5464), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5466), 1,
      sym__concat,
    STATE(1249), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4554), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4552), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [55484] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1322), 7,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1320), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [55539] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1263), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1261), 42,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [55594] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5464), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5466), 1,
      sym__concat,
    STATE(1245), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4566), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4564), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [55655] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5525), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5528), 1,
      sym__concat,
    STATE(1262), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1253), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1251), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [55716] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5464), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5466), 1,
      sym__concat,
    STATE(1249), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4372), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4370), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [55777] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5464), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5466), 1,
      sym__concat,
    STATE(1245), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1263), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1261), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [55838] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5464), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5466), 1,
      sym__concat,
    STATE(1245), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4469), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4467), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [55899] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5397), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5399), 1,
      sym__concat,
    STATE(1251), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1263), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1261), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [55960] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5531), 1,
      sym__special_character,
    STATE(1179), 1,
      aux_sym__literal_repeat1,
    ACTIONS(278), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(241), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [56019] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1318), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1316), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [56074] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5464), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5466), 1,
      sym__concat,
    STATE(1249), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5391), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5389), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [56135] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5464), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5466), 1,
      sym__concat,
    STATE(1245), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5395), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5393), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [56196] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5080), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5082), 1,
      sym__concat,
    STATE(1177), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4469), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4467), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [56257] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1318), 7,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1316), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [56312] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5080), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5082), 1,
      sym__concat,
    STATE(1215), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4372), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4370), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [56373] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5080), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5082), 1,
      sym__concat,
    STATE(1177), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4566), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4564), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [56434] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5464), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5466), 1,
      sym__concat,
    STATE(1249), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(2096), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(2094), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [56495] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5080), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5082), 1,
      sym__concat,
    STATE(1215), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4554), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4552), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [56556] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1314), 7,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1312), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [56611] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1354), 6,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1352), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [56666] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1294), 7,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1292), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [56721] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1314), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1312), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [56776] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4372), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4370), 42,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [56831] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1350), 7,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1348), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [56886] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1334), 7,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1332), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [56941] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1322), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1320), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [56996] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1338), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1336), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [57051] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1314), 6,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1312), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [57106] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1346), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1344), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [57161] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1342), 6,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1340), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [57216] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1358), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1356), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [57271] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5497), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5509), 1,
      sym__concat,
    STATE(1228), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5391), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5389), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [57332] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5497), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5509), 1,
      sym__concat,
    STATE(1227), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5395), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5393), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [57393] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1326), 7,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1324), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [57448] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1253), 7,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1251), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [57503] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1354), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1352), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [57558] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5533), 1,
      sym__special_character,
    STATE(1295), 1,
      aux_sym__literal_repeat1,
    ACTIONS(1364), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1362), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [57617] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1298), 7,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1296), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [57672] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1338), 6,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1336), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [57727] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1354), 6,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1352), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [57782] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1322), 6,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1320), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [57837] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1330), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1328), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [57892] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4554), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4552), 42,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [57947] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1338), 6,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1336), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [58002] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1302), 7,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1300), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [58057] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1342), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1340), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [58112] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1322), 6,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1320), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [58167] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1314), 6,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1312), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [58222] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5536), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5538), 1,
      sym__concat,
    STATE(1311), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5395), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5393), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [58282] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5540), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5542), 1,
      sym__concat,
    STATE(1441), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1263), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1261), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [58342] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1346), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1344), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [58396] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1358), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1356), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [58450] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5536), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5544), 1,
      sym__concat,
    STATE(1347), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1267), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1265), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [58510] = 9,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(5546), 1,
      anon_sym_DQUOTE,
    ACTIONS(5548), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(5550), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(5552), 1,
      anon_sym_BQUOTE,
    ACTIONS(5554), 1,
      anon_sym_DOLLAR_BQUOTE,
    STATE(2601), 3,
      sym_string,
      sym_expansion,
      sym_command_substitution,
    ACTIONS(5435), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(5433), 24,
      sym__concat,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [58576] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5556), 1,
      sym__special_character,
    STATE(1313), 1,
      aux_sym__literal_repeat1,
    ACTIONS(1364), 6,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1362), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [58634] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5565), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(5568), 1,
      anon_sym_DOLLAR,
    ACTIONS(5571), 1,
      sym__special_character,
    ACTIONS(5574), 1,
      anon_sym_DQUOTE,
    ACTIONS(5577), 1,
      aux_sym_number_token1,
    ACTIONS(5580), 1,
      aux_sym_number_token2,
    ACTIONS(5583), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(5586), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(5589), 1,
      anon_sym_BQUOTE,
    ACTIONS(5592), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(5598), 1,
      sym_test_operator,
    ACTIONS(5601), 1,
      sym__brace_start,
    STATE(3469), 1,
      aux_sym__literal_repeat1,
    ACTIONS(2216), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5562), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(5595), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(1314), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    ACTIONS(5559), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(3093), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2214), 13,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
  [58724] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1342), 6,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1340), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [58778] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1338), 6,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1336), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [58832] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5608), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(5610), 1,
      anon_sym_DOLLAR,
    ACTIONS(5612), 1,
      sym__special_character,
    ACTIONS(5614), 1,
      anon_sym_DQUOTE,
    ACTIONS(5616), 1,
      aux_sym_number_token1,
    ACTIONS(5618), 1,
      aux_sym_number_token2,
    ACTIONS(5620), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(5622), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(5624), 1,
      anon_sym_BQUOTE,
    ACTIONS(5626), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(5630), 1,
      sym_test_operator,
    ACTIONS(5632), 1,
      sym__brace_start,
    STATE(3469), 1,
      aux_sym__literal_repeat1,
    ACTIONS(2074), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5606), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(5628), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(1314), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    ACTIONS(5604), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(3093), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2072), 13,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
  [58922] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5634), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5636), 1,
      sym__concat,
    STATE(1398), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5180), 6,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5178), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [58982] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1322), 6,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1320), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [59036] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5638), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5640), 1,
      sym__concat,
    STATE(1367), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(2096), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(2094), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [59096] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5642), 1,
      sym__special_character,
    STATE(1358), 1,
      aux_sym__literal_repeat1,
    ACTIONS(4548), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4546), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [59154] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5634), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5636), 1,
      sym__concat,
    STATE(1443), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5176), 6,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5174), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [59214] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1310), 6,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1308), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [59268] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1306), 6,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1304), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [59322] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5644), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5646), 1,
      sym__concat,
    STATE(1420), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5478), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5476), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [59382] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5644), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5646), 1,
      sym__concat,
    STATE(1415), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5470), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5468), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [59442] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1354), 6,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1352), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [59496] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4532), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4530), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [59550] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5462), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5460), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [59604] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5638), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5640), 1,
      sym__concat,
    STATE(1367), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4372), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4370), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [59664] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5638), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5640), 1,
      sym__concat,
    STATE(1421), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4469), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4467), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [59724] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5608), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(5610), 1,
      anon_sym_DOLLAR,
    ACTIONS(5612), 1,
      sym__special_character,
    ACTIONS(5614), 1,
      anon_sym_DQUOTE,
    ACTIONS(5616), 1,
      aux_sym_number_token1,
    ACTIONS(5618), 1,
      aux_sym_number_token2,
    ACTIONS(5620), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(5622), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(5624), 1,
      anon_sym_BQUOTE,
    ACTIONS(5626), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(5630), 1,
      sym_test_operator,
    ACTIONS(5632), 1,
      sym__brace_start,
    STATE(3469), 1,
      aux_sym__literal_repeat1,
    ACTIONS(2096), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5606), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(5628), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(1314), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    ACTIONS(5604), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(3093), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2094), 13,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
  [59814] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5648), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5650), 1,
      sym__concat,
    STATE(1373), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1288), 6,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1286), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [59874] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1314), 6,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1312), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [59928] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5652), 1,
      sym__special_character,
    STATE(1313), 1,
      aux_sym__literal_repeat1,
    ACTIONS(4482), 6,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4480), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [59986] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5638), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5640), 1,
      sym__concat,
    STATE(1367), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(2074), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(2072), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [60046] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5634), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5636), 1,
      sym__concat,
    STATE(1443), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5375), 6,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5373), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [60106] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1263), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1261), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [60160] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5648), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5654), 1,
      sym__concat,
    STATE(1357), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4469), 6,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4467), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [60220] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1310), 6,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1308), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [60274] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5634), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5636), 1,
      sym__concat,
    STATE(1398), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5069), 6,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5067), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [60334] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5648), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5654), 1,
      sym__concat,
    STATE(1333), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4372), 6,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4370), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [60394] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5638), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5640), 1,
      sym__concat,
    STATE(1421), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4566), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4564), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [60454] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5638), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5640), 1,
      sym__concat,
    STATE(1367), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4554), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4552), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [60514] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1342), 6,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1340), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [60568] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4465), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4463), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [60622] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5656), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5659), 1,
      sym__concat,
    STATE(1347), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1253), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1251), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [60682] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1338), 6,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1336), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [60736] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5662), 1,
      sym__special_character,
    STATE(1433), 1,
      aux_sym__literal_repeat1,
    ACTIONS(4469), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4467), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [60794] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1310), 6,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1308), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [60848] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1263), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1261), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [60902] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1322), 6,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1320), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [60956] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1306), 6,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1304), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [61010] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5638), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5640), 1,
      sym__concat,
    STATE(1421), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5069), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5067), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [61070] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5638), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5640), 1,
      sym__concat,
    STATE(1367), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5375), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5373), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [61130] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5664), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5666), 1,
      sym__concat,
    STATE(1447), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1263), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1261), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [61190] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5648), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5668), 1,
      sym__concat,
    STATE(1373), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1267), 6,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1265), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [61250] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5670), 1,
      sym__special_character,
    STATE(1358), 1,
      aux_sym__literal_repeat1,
    ACTIONS(1364), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1362), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [61308] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1310), 6,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1308), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [61362] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5664), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5666), 1,
      sym__concat,
    STATE(1383), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5176), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5174), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [61422] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5662), 1,
      sym__special_character,
    STATE(1433), 1,
      aux_sym__literal_repeat1,
    ACTIONS(5069), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5067), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [61480] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5664), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5666), 1,
      sym__concat,
    STATE(1447), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5180), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5178), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [61540] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5673), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5676), 1,
      sym__concat,
    STATE(1363), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1253), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1251), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [61600] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1310), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1308), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [61654] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1306), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1304), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [61708] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1310), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1308), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [61762] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5638), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5679), 1,
      sym__concat,
    STATE(1363), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1288), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1286), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [61822] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5681), 1,
      sym__special_character,
    STATE(1448), 1,
      aux_sym__literal_repeat1,
    ACTIONS(5180), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5178), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [61880] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5644), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5646), 1,
      sym__concat,
    STATE(1420), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1263), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1261), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [61940] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5638), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5640), 1,
      sym__concat,
    STATE(1367), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5391), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5389), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [62000] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5683), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5686), 1,
      sym__concat,
    STATE(1371), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1253), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1251), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [62060] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5638), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5640), 1,
      sym__concat,
    STATE(1421), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5395), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5393), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [62120] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5689), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5692), 1,
      sym__concat,
    STATE(1373), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1253), 6,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1251), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [62180] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5652), 1,
      sym__special_character,
    STATE(1313), 1,
      aux_sym__literal_repeat1,
    ACTIONS(4469), 6,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4467), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [62238] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1294), 6,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1292), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [62292] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1281), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1271), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [62346] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5652), 1,
      sym__special_character,
    STATE(1313), 1,
      aux_sym__literal_repeat1,
    ACTIONS(4566), 6,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4564), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [62404] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1334), 6,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1332), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [62458] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5642), 1,
      sym__special_character,
    STATE(1358), 1,
      aux_sym__literal_repeat1,
    ACTIONS(4482), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4480), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [62516] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5662), 1,
      sym__special_character,
    STATE(1433), 1,
      aux_sym__literal_repeat1,
    ACTIONS(4566), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4564), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [62574] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5540), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5542), 1,
      sym__concat,
    STATE(1444), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5391), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5389), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [62634] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5662), 1,
      sym__special_character,
    STATE(1433), 1,
      aux_sym__literal_repeat1,
    ACTIONS(5697), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5695), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [62692] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5664), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5699), 1,
      sym__concat,
    STATE(1371), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1288), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1286), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [62752] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1330), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1328), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [62806] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1354), 6,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1352), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [62860] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1281), 6,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1271), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [62914] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4554), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4552), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [62968] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5540), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5542), 1,
      sym__concat,
    STATE(1441), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5395), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5393), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [63028] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1314), 6,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1312), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [63082] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4372), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4370), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [63136] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5648), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5654), 1,
      sym__concat,
    STATE(1333), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5375), 6,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5373), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [63196] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5648), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5654), 1,
      sym__concat,
    STATE(1357), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5069), 6,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5067), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [63256] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1318), 6,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1316), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [63310] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5536), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5538), 1,
      sym__concat,
    STATE(1457), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5391), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5389), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [63370] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5648), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5654), 1,
      sym__concat,
    STATE(1333), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4554), 6,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4552), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [63430] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5648), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5654), 1,
      sym__concat,
    STATE(1357), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4566), 6,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4564), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [63490] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5648), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5654), 1,
      sym__concat,
    STATE(1333), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(2074), 6,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(2072), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [63550] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5634), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5701), 1,
      sym__concat,
    STATE(1475), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1267), 6,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1265), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [63610] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5648), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5654), 1,
      sym__concat,
    STATE(1333), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5391), 6,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5389), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [63670] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1314), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1312), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [63724] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5642), 1,
      sym__special_character,
    STATE(1358), 1,
      aux_sym__literal_repeat1,
    ACTIONS(4566), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4564), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [63782] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1342), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1340), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [63836] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1338), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1336), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [63890] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1263), 6,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1261), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [63944] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5648), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5654), 1,
      sym__concat,
    STATE(1357), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1263), 6,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1261), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [64004] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1318), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1316), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [64058] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5648), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5654), 1,
      sym__concat,
    STATE(1357), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5395), 6,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5393), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [64118] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1302), 6,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1300), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [64172] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1294), 6,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1292), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [64226] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1298), 6,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1296), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [64280] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1334), 6,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1332), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [64334] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5703), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5706), 1,
      sym__concat,
    STATE(1412), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1253), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1251), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [64394] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1322), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1320), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [64448] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1346), 6,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1344), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [64502] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5644), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5709), 1,
      sym__concat,
    STATE(1412), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1288), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1286), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [64562] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1350), 6,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1348), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [64616] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1346), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1344), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [64670] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1358), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1356), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [64724] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1358), 6,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1356), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [64778] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5644), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5711), 1,
      sym__concat,
    STATE(1412), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1267), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1265), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [64838] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5638), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5713), 1,
      sym__concat,
    STATE(1363), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1267), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1265), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [64898] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1326), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1324), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [64952] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1253), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1251), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [65006] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1302), 6,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1300), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [65060] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5614), 1,
      anon_sym_DQUOTE,
    ACTIONS(5719), 1,
      sym_variable_name,
    STATE(3393), 1,
      sym_string,
    ACTIONS(5717), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1235), 3,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(5715), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1227), 29,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      aux_sym_heredoc_redirect_token1,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [65124] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5614), 1,
      anon_sym_DQUOTE,
    ACTIONS(5719), 1,
      sym_variable_name,
    STATE(3393), 1,
      sym_string,
    ACTIONS(5717), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1241), 3,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(5715), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1239), 29,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      aux_sym_heredoc_redirect_token1,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [65188] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1346), 6,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1344), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [65242] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1358), 6,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1356), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [65296] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5662), 1,
      sym__special_character,
    STATE(1433), 1,
      aux_sym__literal_repeat1,
    ACTIONS(5723), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5721), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [65354] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1298), 6,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1296), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [65408] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5652), 1,
      sym__special_character,
    STATE(1313), 1,
      aux_sym__literal_repeat1,
    ACTIONS(4548), 6,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4546), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [65466] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1350), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1348), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [65520] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5725), 1,
      sym__special_character,
    STATE(1433), 1,
      aux_sym__literal_repeat1,
    ACTIONS(1364), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1362), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [65578] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1298), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1296), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [65632] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1310), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1308), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [65686] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5642), 1,
      sym__special_character,
    STATE(1358), 1,
      aux_sym__literal_repeat1,
    ACTIONS(4469), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4467), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [65744] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1302), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1300), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [65798] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1306), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1304), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [65852] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1322), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1320), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [65906] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1253), 6,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1251), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [65960] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5540), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5728), 1,
      sym__concat,
    STATE(1451), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1267), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1265), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [66020] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1326), 6,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1324), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [66074] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5634), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5730), 1,
      sym__concat,
    STATE(1475), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1288), 6,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1286), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [66134] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5540), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5732), 1,
      sym__concat,
    STATE(1451), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1288), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1286), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [66194] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5638), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5640), 1,
      sym__concat,
    STATE(1421), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1263), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1261), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [66254] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5648), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5654), 1,
      sym__concat,
    STATE(1333), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(2096), 6,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(2094), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [66314] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5664), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5734), 1,
      sym__concat,
    STATE(1371), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1267), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1265), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [66374] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5736), 1,
      sym__special_character,
    STATE(1448), 1,
      aux_sym__literal_repeat1,
    ACTIONS(1364), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1362), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [66432] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1330), 6,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1328), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [66486] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1338), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1336), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [66540] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5739), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5742), 1,
      sym__concat,
    STATE(1451), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1253), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1251), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [66600] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1334), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1332), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [66654] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1342), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1340), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [66708] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1350), 6,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1348), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [66762] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1294), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1292), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [66816] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1314), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1312), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [66870] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5536), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5745), 1,
      sym__concat,
    STATE(1347), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1288), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1286), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [66930] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1354), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1352), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [66984] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1310), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1308), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [67038] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5634), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5636), 1,
      sym__concat,
    STATE(1398), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1263), 6,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1261), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [67098] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1326), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1324), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [67152] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1253), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1251), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [67206] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1350), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1348), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [67260] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1354), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1352), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [67314] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1253), 6,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1251), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [67368] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1326), 6,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1324), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [67422] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5664), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5666), 1,
      sym__concat,
    STATE(1447), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5069), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5067), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [67482] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5536), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5538), 1,
      sym__concat,
    STATE(1311), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1263), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1261), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [67542] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1318), 6,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1316), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [67596] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5664), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5666), 1,
      sym__concat,
    STATE(1383), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5375), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5373), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [67656] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1298), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1296), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [67710] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1330), 6,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1328), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [67764] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1302), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1300), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [67818] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5662), 1,
      sym__special_character,
    STATE(1433), 1,
      aux_sym__literal_repeat1,
    ACTIONS(5395), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5393), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [67876] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5747), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5750), 1,
      sym__concat,
    STATE(1475), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1253), 6,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1251), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [67936] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1334), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1332), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [67990] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1294), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1292), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [68044] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1318), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1316), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [68098] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1330), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1328), 41,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [68152] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5681), 1,
      sym__special_character,
    STATE(1448), 1,
      aux_sym__literal_repeat1,
    ACTIONS(5069), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5067), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [68210] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5664), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5666), 1,
      sym__concat,
    STATE(1663), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5180), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5178), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [68269] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1346), 7,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1344), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [68322] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1342), 7,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1340), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [68375] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1346), 6,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1344), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [68428] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1294), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1292), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [68481] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1314), 7,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1312), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [68534] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1310), 7,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1308), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [68587] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1306), 7,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1304), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [68640] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1354), 6,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1352), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [68693] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1334), 7,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1332), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [68746] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1310), 7,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1308), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [68799] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1358), 6,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1356), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [68852] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1253), 7,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1251), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [68905] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1326), 7,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1324), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [68958] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1298), 7,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1296), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [69011] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1318), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1316), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [69064] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1334), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1332), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [69117] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5375), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5373), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [69170] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1338), 6,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1336), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [69223] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1294), 7,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1292), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [69276] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1322), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1320), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [69329] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1354), 7,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1352), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [69382] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1330), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1328), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [69435] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1294), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1292), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [69488] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1294), 7,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1292), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [69541] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1334), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1332), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [69594] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1338), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1336), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [69647] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1302), 7,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1300), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [69700] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5753), 1,
      sym__special_character,
    STATE(1608), 1,
      aux_sym__literal_repeat1,
    ACTIONS(4566), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4564), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [69757] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1338), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1336), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [69810] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1354), 7,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1352), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [69863] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5638), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5640), 1,
      sym__concat,
    STATE(1661), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5375), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5373), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [69922] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1342), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1340), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [69975] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1314), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1312), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [70028] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1318), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1316), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [70081] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1298), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1296), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [70134] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1302), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1300), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [70187] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1342), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1340), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [70240] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1302), 7,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1300), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [70293] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1298), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1296), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [70346] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4532), 6,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4530), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [70399] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5755), 1,
      sym__special_character,
    STATE(1522), 1,
      aux_sym__literal_repeat1,
    ACTIONS(1364), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1362), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [70456] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5638), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5640), 1,
      sym__concat,
    STATE(1660), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5069), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5067), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [70515] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5758), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5760), 1,
      sym__concat,
    STATE(1640), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5395), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5393), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [70574] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1350), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1348), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [70627] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5758), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5760), 1,
      sym__concat,
    STATE(1645), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5391), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5389), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [70686] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1314), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1312), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [70739] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5762), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5764), 1,
      sym__concat,
    STATE(1582), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1267), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1265), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [70798] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5638), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5640), 1,
      sym__concat,
    STATE(1661), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4554), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4552), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [70857] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1330), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1328), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [70910] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1314), 6,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1312), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [70963] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1253), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1251), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [71016] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1326), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1324), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [71069] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2096), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(2094), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [71122] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1263), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1261), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [71175] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5664), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5666), 1,
      sym__concat,
    STATE(1663), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5069), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5067), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [71234] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1338), 7,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1336), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [71287] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1330), 7,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1328), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [71340] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5753), 1,
      sym__special_character,
    STATE(1608), 1,
      aux_sym__literal_repeat1,
    ACTIONS(4469), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4467), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [71397] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1322), 7,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1320), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [71450] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1302), 6,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1300), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [71503] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5768), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5766), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [71556] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4372), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4370), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [71609] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1298), 7,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1296), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [71662] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1310), 7,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1308), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [71715] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1354), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1352), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [71768] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1350), 7,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1348), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [71821] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1346), 7,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1344), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [71874] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1358), 7,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1356), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [71927] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1306), 7,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1304), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [71980] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1263), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1261), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [72033] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5462), 6,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5460), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [72086] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1310), 7,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1308), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [72139] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1314), 7,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1312), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [72192] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1342), 7,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1340), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [72245] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5772), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5770), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [72298] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5762), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5774), 1,
      sym__concat,
    STATE(1582), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1288), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1286), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [72357] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5664), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5666), 1,
      sym__concat,
    STATE(1664), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5375), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5373), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [72416] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5638), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5640), 1,
      sym__concat,
    STATE(1660), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4566), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4564), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [72475] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5638), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5640), 1,
      sym__concat,
    STATE(1661), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(2074), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(2072), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [72534] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5768), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5766), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [72587] = 10,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4272), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5776), 1,
      sym_variable_name,
    STATE(6746), 1,
      sym_subscript,
    ACTIONS(4253), 2,
      anon_sym_PIPE,
      anon_sym_PIPE_AMP,
    STATE(5148), 2,
      sym_variable_assignment,
      aux_sym_variable_assignments_repeat1,
    ACTIONS(4261), 3,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
    STATE(5366), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(4270), 4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_LT,
      anon_sym_LT_LT_DASH,
    ACTIONS(4247), 28,
      anon_sym_LPAREN_LPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [72654] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1354), 6,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1352), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [72707] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1310), 6,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1308), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [72760] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1306), 6,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1304), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [72813] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1310), 6,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1308), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [72866] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5753), 1,
      sym__special_character,
    STATE(1608), 1,
      aux_sym__literal_repeat1,
    ACTIONS(5395), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5393), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [72923] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1326), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1324), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [72976] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1253), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1251), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [73029] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1334), 7,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1332), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [73082] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1318), 6,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1316), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [73135] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4699), 1,
      aux_sym_concatenation_token1,
    ACTIONS(4701), 1,
      sym__concat,
    STATE(1715), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1263), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1261), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [73194] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4554), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4552), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [73247] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5772), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5770), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [73300] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5638), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5640), 1,
      sym__concat,
    STATE(1660), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5395), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5393), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [73359] = 28,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1131), 1,
      anon_sym_LPAREN,
    ACTIONS(1141), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1145), 1,
      anon_sym_DOLLAR,
    ACTIONS(1149), 1,
      anon_sym_DQUOTE,
    ACTIONS(1153), 1,
      aux_sym_number_token1,
    ACTIONS(1155), 1,
      aux_sym_number_token2,
    ACTIONS(1157), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1159), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1163), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1169), 1,
      sym__brace_start,
    ACTIONS(1187), 1,
      sym_word,
    ACTIONS(1189), 1,
      anon_sym_BANG,
    ACTIONS(1195), 1,
      anon_sym_TILDE,
    ACTIONS(1201), 1,
      sym_test_operator,
    ACTIONS(3060), 1,
      anon_sym_BQUOTE,
    ACTIONS(5779), 1,
      sym__special_character,
    ACTIONS(5781), 1,
      sym_extglob_pattern,
    STATE(2648), 1,
      sym__extglob_blob,
    STATE(2717), 1,
      aux_sym__literal_repeat1,
    STATE(3004), 1,
      sym__expression,
    ACTIONS(1129), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1165), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1191), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1193), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1199), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2594), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2503), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [73462] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1354), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1352), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [73515] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1310), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1308), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [73568] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1306), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1304), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [73621] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1318), 7,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1316), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [73674] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1310), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1308), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [73727] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5783), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5786), 1,
      sym__concat,
    STATE(1582), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1253), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1251), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [73786] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1314), 6,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1312), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [73839] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1310), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1308), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [73892] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1350), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1348), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [73945] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1358), 6,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1356), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [73998] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1298), 6,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1296), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [74051] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1310), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1308), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [74104] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1342), 6,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1340), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [74157] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1253), 6,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1251), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [74210] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1306), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1304), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [74263] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1306), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1304), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [74316] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1310), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1308), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [74369] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1310), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1308), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [74422] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1350), 6,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1348), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [74475] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5638), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5640), 1,
      sym__concat,
    STATE(1661), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4372), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4370), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [74534] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5762), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5789), 1,
      sym__concat,
    STATE(1528), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1263), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1261), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [74593] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1294), 6,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1292), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [74646] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5638), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5640), 1,
      sym__concat,
    STATE(1660), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4469), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4467), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [74705] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1350), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1348), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [74758] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5638), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5640), 1,
      sym__concat,
    STATE(1661), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(2096), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(2094), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [74817] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5664), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5666), 1,
      sym__concat,
    STATE(1664), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5176), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5174), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [74876] = 9,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(5791), 1,
      anon_sym_DQUOTE,
    ACTIONS(5793), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(5795), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(5797), 1,
      anon_sym_BQUOTE,
    ACTIONS(5799), 1,
      anon_sym_DOLLAR_BQUOTE,
    STATE(2865), 3,
      sym_string,
      sym_expansion,
      sym_command_substitution,
    ACTIONS(5435), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(5433), 23,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [74941] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5801), 1,
      sym__special_character,
    STATE(1604), 1,
      aux_sym__literal_repeat1,
    ACTIONS(1364), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1362), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [74998] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1318), 7,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1316), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [75051] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5753), 1,
      sym__special_character,
    STATE(1608), 1,
      aux_sym__literal_repeat1,
    ACTIONS(5723), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5721), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [75108] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2074), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(2072), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [75161] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5804), 1,
      sym__special_character,
    STATE(1608), 1,
      aux_sym__literal_repeat1,
    ACTIONS(1364), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1362), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [75218] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1358), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1356), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [75271] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1346), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1344), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [75324] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4465), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4463), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [75377] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1298), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1296), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [75430] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4554), 6,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4552), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [75483] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1342), 6,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1340), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [75536] = 28,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(187), 1,
      anon_sym_TILDE,
    ACTIONS(350), 1,
      sym_word,
    ACTIONS(359), 1,
      anon_sym_LPAREN,
    ACTIONS(361), 1,
      anon_sym_BANG,
    ACTIONS(363), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(367), 1,
      anon_sym_DOLLAR,
    ACTIONS(371), 1,
      anon_sym_DQUOTE,
    ACTIONS(375), 1,
      aux_sym_number_token1,
    ACTIONS(377), 1,
      aux_sym_number_token2,
    ACTIONS(379), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(381), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(385), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(389), 1,
      sym_test_operator,
    ACTIONS(391), 1,
      sym__brace_start,
    ACTIONS(5807), 1,
      sym__special_character,
    ACTIONS(5809), 1,
      anon_sym_BQUOTE,
    ACTIONS(5811), 1,
      sym_extglob_pattern,
    STATE(2641), 1,
      aux_sym__literal_repeat1,
    STATE(2950), 1,
      sym__extglob_blob,
    STATE(2963), 1,
      sym__expression,
    ACTIONS(183), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(185), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(352), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(373), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(387), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3071), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2521), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [75639] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1318), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1316), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [75692] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1330), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1328), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [75745] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5753), 1,
      sym__special_character,
    STATE(1608), 1,
      aux_sym__literal_repeat1,
    ACTIONS(5069), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5067), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [75802] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1326), 6,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1324), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [75855] = 28,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1131), 1,
      anon_sym_LPAREN,
    ACTIONS(1141), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1145), 1,
      anon_sym_DOLLAR,
    ACTIONS(1149), 1,
      anon_sym_DQUOTE,
    ACTIONS(1153), 1,
      aux_sym_number_token1,
    ACTIONS(1155), 1,
      aux_sym_number_token2,
    ACTIONS(1157), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1159), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1163), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1169), 1,
      sym__brace_start,
    ACTIONS(1171), 1,
      sym_word,
    ACTIONS(1173), 1,
      anon_sym_BANG,
    ACTIONS(1179), 1,
      anon_sym_TILDE,
    ACTIONS(1181), 1,
      sym__special_character,
    ACTIONS(1185), 1,
      sym_test_operator,
    ACTIONS(3060), 1,
      anon_sym_BQUOTE,
    ACTIONS(5781), 1,
      sym_extglob_pattern,
    STATE(2484), 1,
      aux_sym__literal_repeat1,
    STATE(2648), 1,
      sym__extglob_blob,
    STATE(3051), 1,
      sym__expression,
    ACTIONS(1129), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1165), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1175), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1177), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1183), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2594), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2456), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [75958] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1326), 6,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1324), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [76011] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4372), 6,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4370), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [76064] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1253), 6,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1251), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [76117] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1302), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1300), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [76170] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1253), 7,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1251), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [76223] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1326), 7,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1324), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [76276] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5776), 1,
      sym_variable_name,
    STATE(6746), 1,
      sym_subscript,
    ACTIONS(4261), 2,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(4348), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    STATE(5148), 2,
      sym_variable_assignment,
      aux_sym_variable_assignments_repeat1,
    STATE(5366), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(4253), 16,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
    ACTIONS(4247), 18,
      anon_sym_LPAREN_LPAREN,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [76341] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1350), 6,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1348), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [76394] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4554), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4552), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [76447] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5813), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5815), 1,
      sym__concat,
    STATE(1638), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5391), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5389), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [76506] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5817), 1,
      sym__special_character,
    STATE(1522), 1,
      aux_sym__literal_repeat1,
    ACTIONS(5478), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5476), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [76563] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5813), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5815), 1,
      sym__concat,
    STATE(1637), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5395), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5393), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [76622] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1334), 6,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1332), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [76675] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5826), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5828), 1,
      sym_file_descriptor,
    ACTIONS(5831), 3,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(5823), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
    ACTIONS(5821), 12,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
    ACTIONS(5819), 17,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [76736] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5826), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5828), 1,
      sym_file_descriptor,
    ACTIONS(5831), 3,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(5823), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
    ACTIONS(5821), 12,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
    ACTIONS(5819), 17,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [76797] = 28,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1131), 1,
      anon_sym_LPAREN,
    ACTIONS(1141), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1145), 1,
      anon_sym_DOLLAR,
    ACTIONS(1149), 1,
      anon_sym_DQUOTE,
    ACTIONS(1153), 1,
      aux_sym_number_token1,
    ACTIONS(1155), 1,
      aux_sym_number_token2,
    ACTIONS(1157), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1159), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1163), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1169), 1,
      sym__brace_start,
    ACTIONS(1211), 1,
      sym_word,
    ACTIONS(1213), 1,
      anon_sym_BANG,
    ACTIONS(1219), 1,
      anon_sym_TILDE,
    ACTIONS(1221), 1,
      sym__special_character,
    ACTIONS(1225), 1,
      sym_test_operator,
    ACTIONS(3060), 1,
      anon_sym_BQUOTE,
    ACTIONS(5781), 1,
      sym_extglob_pattern,
    STATE(2484), 1,
      aux_sym__literal_repeat1,
    STATE(2648), 1,
      sym__extglob_blob,
    STATE(3333), 1,
      sym__expression,
    ACTIONS(1129), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1165), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1215), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1217), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1223), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2594), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2588), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [76900] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5813), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5833), 1,
      sym__concat,
    STATE(1649), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1267), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1265), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [76959] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5813), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5835), 1,
      sym__concat,
    STATE(1649), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1288), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1286), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [77018] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4372), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4370), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [77071] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5758), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5837), 1,
      sym__concat,
    STATE(1659), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1267), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1265), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [77130] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1358), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1356), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [77183] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1346), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1344), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [77236] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5753), 1,
      sym__special_character,
    STATE(1608), 1,
      aux_sym__literal_repeat1,
    ACTIONS(5697), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5695), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [77293] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1298), 6,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1296), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [77346] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5758), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5839), 1,
      sym__concat,
    STATE(1659), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1288), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1286), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [77405] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5841), 1,
      sym__special_character,
    STATE(1646), 1,
      aux_sym__literal_repeat1,
    ACTIONS(1364), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1362), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [77462] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5844), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5846), 1,
      sym__concat,
    STATE(1672), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5478), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5476), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [77521] = 9,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(5848), 1,
      anon_sym_DQUOTE,
    ACTIONS(5850), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(5852), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(5854), 1,
      anon_sym_BQUOTE,
    ACTIONS(5856), 1,
      anon_sym_DOLLAR_BQUOTE,
    STATE(2862), 3,
      sym_string,
      sym_expansion,
      sym_command_substitution,
    ACTIONS(5435), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(5433), 23,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [77586] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5858), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5861), 1,
      sym__concat,
    STATE(1649), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1253), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1251), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [77645] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5844), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5846), 1,
      sym__concat,
    STATE(1714), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5470), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5468), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [77704] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1294), 6,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1292), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [77757] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1302), 6,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1300), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [77810] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4699), 1,
      aux_sym_concatenation_token1,
    ACTIONS(4701), 1,
      sym__concat,
    STATE(1715), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(278), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(241), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [77869] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5868), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(5870), 1,
      anon_sym_DOLLAR,
    ACTIONS(5872), 1,
      sym__special_character,
    ACTIONS(5874), 1,
      anon_sym_DQUOTE,
    ACTIONS(5876), 1,
      aux_sym_number_token1,
    ACTIONS(5878), 1,
      aux_sym_number_token2,
    ACTIONS(5880), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(5882), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(5884), 1,
      anon_sym_BQUOTE,
    ACTIONS(5886), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(5890), 1,
      sym_test_operator,
    ACTIONS(5892), 1,
      sym__brace_start,
    STATE(5674), 1,
      aux_sym__literal_repeat1,
    STATE(5738), 1,
      sym_concatenation,
    ACTIONS(2496), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5866), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(5888), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(5864), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(5608), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2494), 13,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
  [77958] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1322), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1320), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [78011] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1263), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1261), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [78064] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1338), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1336), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [78117] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5894), 1,
      sym__special_character,
    STATE(1604), 1,
      aux_sym__literal_repeat1,
    ACTIONS(5395), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5393), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [78174] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5896), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5899), 1,
      sym__concat,
    STATE(1659), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1253), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1251), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [78233] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5638), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5902), 1,
      sym__concat,
    STATE(1363), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1267), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1265), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [78292] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5638), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5904), 1,
      sym__concat,
    STATE(1363), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1288), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1286), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [78351] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5638), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5640), 1,
      sym__concat,
    STATE(1660), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1263), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1261), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [78410] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5664), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5906), 1,
      sym__concat,
    STATE(1371), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1267), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1265), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [78469] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5664), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5908), 1,
      sym__concat,
    STATE(1371), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1288), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1286), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [78528] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5664), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5666), 1,
      sym__concat,
    STATE(1663), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1263), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1261), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [78587] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1334), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1332), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [78640] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5375), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5373), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [78693] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1330), 6,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1328), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [78746] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1253), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1251), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [78799] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5868), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(5870), 1,
      anon_sym_DOLLAR,
    ACTIONS(5872), 1,
      sym__special_character,
    ACTIONS(5874), 1,
      anon_sym_DQUOTE,
    ACTIONS(5876), 1,
      aux_sym_number_token1,
    ACTIONS(5878), 1,
      aux_sym_number_token2,
    ACTIONS(5880), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(5882), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(5884), 1,
      anon_sym_BQUOTE,
    ACTIONS(5886), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(5892), 1,
      sym__brace_start,
    ACTIONS(5912), 1,
      sym_test_operator,
    STATE(5632), 1,
      aux_sym__literal_repeat1,
    STATE(5789), 1,
      sym_concatenation,
    ACTIONS(2500), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5866), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(5888), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(5910), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(5553), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2498), 13,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
  [78888] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1294), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1292), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [78941] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5844), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5914), 1,
      sym__concat,
    STATE(1722), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1267), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1265), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [79000] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4465), 6,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4463), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [79053] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1302), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1300), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [79106] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5813), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5815), 1,
      sym__concat,
    STATE(1637), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1263), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1261), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [79165] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1322), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1320), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [79218] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1354), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1352), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [79271] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1326), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1324), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [79324] = 28,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1127), 1,
      sym_word,
    ACTIONS(1131), 1,
      anon_sym_LPAREN,
    ACTIONS(1133), 1,
      anon_sym_BANG,
    ACTIONS(1139), 1,
      anon_sym_TILDE,
    ACTIONS(1141), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1145), 1,
      anon_sym_DOLLAR,
    ACTIONS(1147), 1,
      sym__special_character,
    ACTIONS(1149), 1,
      anon_sym_DQUOTE,
    ACTIONS(1153), 1,
      aux_sym_number_token1,
    ACTIONS(1155), 1,
      aux_sym_number_token2,
    ACTIONS(1157), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1159), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1163), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1167), 1,
      sym_test_operator,
    ACTIONS(1169), 1,
      sym__brace_start,
    ACTIONS(3060), 1,
      anon_sym_BQUOTE,
    ACTIONS(5781), 1,
      sym_extglob_pattern,
    STATE(2484), 1,
      aux_sym__literal_repeat1,
    STATE(2648), 1,
      sym__extglob_blob,
    STATE(2972), 1,
      sym__expression,
    ACTIONS(1129), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1135), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1137), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1151), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(1165), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(2594), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2573), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [79427] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1322), 6,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1320), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [79480] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5762), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5789), 1,
      sym__concat,
    STATE(1557), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5470), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5468), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [79539] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1338), 7,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1336), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [79592] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1330), 7,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1328), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [79645] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1322), 7,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1320), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [79698] = 28,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(363), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(367), 1,
      anon_sym_DOLLAR,
    ACTIONS(371), 1,
      anon_sym_DQUOTE,
    ACTIONS(375), 1,
      aux_sym_number_token1,
    ACTIONS(377), 1,
      aux_sym_number_token2,
    ACTIONS(379), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(381), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(385), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(391), 1,
      sym__brace_start,
    ACTIONS(1091), 1,
      sym_word,
    ACTIONS(1093), 1,
      anon_sym_LPAREN,
    ACTIONS(1095), 1,
      anon_sym_BANG,
    ACTIONS(1103), 1,
      anon_sym_TILDE,
    ACTIONS(1113), 1,
      sym_test_operator,
    ACTIONS(5809), 1,
      anon_sym_BQUOTE,
    ACTIONS(5916), 1,
      sym__special_character,
    ACTIONS(5918), 1,
      sym_extglob_pattern,
    STATE(2472), 1,
      aux_sym__literal_repeat1,
    STATE(2827), 1,
      sym__expression,
    STATE(2897), 1,
      sym__extglob_blob,
    ACTIONS(352), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(387), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1099), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1101), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1109), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2863), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2451), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [79801] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1350), 7,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1348), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [79854] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1338), 6,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1336), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [79907] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5638), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5640), 1,
      sym__concat,
    STATE(1661), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5391), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5389), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [79966] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1330), 6,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1328), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [80019] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1358), 7,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1356), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [80072] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1322), 6,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1320), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [80125] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1310), 6,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1308), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [80178] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1306), 6,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1304), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [80231] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5920), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5922), 1,
      sym__concat,
    STATE(1704), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1267), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1265), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [80290] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5920), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5924), 1,
      sym__concat,
    STATE(1704), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1288), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1286), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [80349] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5762), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5789), 1,
      sym__concat,
    STATE(1528), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5478), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5476), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [80408] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1310), 6,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1308), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [80461] = 28,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(107), 1,
      anon_sym_TILDE,
    ACTIONS(237), 1,
      sym_word,
    ACTIONS(248), 1,
      anon_sym_LPAREN,
    ACTIONS(250), 1,
      anon_sym_BANG,
    ACTIONS(258), 1,
      anon_sym_DOLLAR,
    ACTIONS(264), 1,
      aux_sym_number_token1,
    ACTIONS(266), 1,
      aux_sym_number_token2,
    ACTIONS(270), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(282), 1,
      sym_test_operator,
    ACTIONS(284), 1,
      sym__brace_start,
    ACTIONS(1075), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1079), 1,
      sym__special_character,
    ACTIONS(1081), 1,
      anon_sym_DQUOTE,
    ACTIONS(1085), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1087), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3598), 1,
      anon_sym_BQUOTE,
    ACTIONS(5926), 1,
      sym_extglob_pattern,
    STATE(2690), 1,
      aux_sym__literal_repeat1,
    STATE(3082), 1,
      sym__expression,
    STATE(3138), 1,
      sym__extglob_blob,
    ACTIONS(103), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(105), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1073), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1083), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(1089), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3053), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2500), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [80564] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5928), 1,
      sym__special_character,
    STATE(1646), 1,
      aux_sym__literal_repeat1,
    ACTIONS(5180), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5178), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [80621] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1346), 6,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1344), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [80674] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1358), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1356), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [80727] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1318), 6,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1316), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [80780] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5932), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4261), 3,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(4247), 18,
      anon_sym_LPAREN_LPAREN,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
    ACTIONS(5930), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [80837] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5934), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5937), 1,
      sym__concat,
    STATE(1704), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1253), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1251), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [80896] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5844), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5846), 1,
      sym__concat,
    STATE(1672), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1263), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1261), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [80955] = 11,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5776), 1,
      sym_variable_name,
    ACTIONS(5942), 1,
      aux_sym_heredoc_redirect_token1,
    STATE(6746), 1,
      sym_subscript,
    ACTIONS(4253), 2,
      anon_sym_PIPE,
      anon_sym_PIPE_AMP,
    ACTIONS(4692), 2,
      anon_sym_LT_LT,
      anon_sym_LT_LT_DASH,
    ACTIONS(5940), 2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
    STATE(5148), 2,
      sym_variable_assignment,
      aux_sym_variable_assignments_repeat1,
    ACTIONS(4261), 3,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
    STATE(5366), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(4247), 28,
      anon_sym_LPAREN_LPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [81024] = 11,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5776), 1,
      sym_variable_name,
    ACTIONS(5944), 1,
      aux_sym_heredoc_redirect_token1,
    STATE(6746), 1,
      sym_subscript,
    ACTIONS(4253), 2,
      anon_sym_PIPE,
      anon_sym_PIPE_AMP,
    ACTIONS(4692), 2,
      anon_sym_LT_LT,
      anon_sym_LT_LT_DASH,
    ACTIONS(5940), 2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
    STATE(5148), 2,
      sym_variable_assignment,
      aux_sym_variable_assignments_repeat1,
    ACTIONS(4261), 3,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
    STATE(5366), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(4247), 28,
      anon_sym_LPAREN_LPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [81093] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5920), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5946), 1,
      sym__concat,
    STATE(1694), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5395), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5393), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [81152] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1346), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1344), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [81205] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5920), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5946), 1,
      sym__concat,
    STATE(1694), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1263), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1261), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [81264] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5920), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5946), 1,
      sym__concat,
    STATE(1695), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5391), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5389), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [81323] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5462), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5460), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [81376] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1314), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1312), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [81429] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5844), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5948), 1,
      sym__concat,
    STATE(1722), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1288), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1286), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [81488] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4699), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5950), 1,
      sym__concat,
    STATE(965), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1267), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1265), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [81547] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4532), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4530), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [81600] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4699), 1,
      aux_sym_concatenation_token1,
    ACTIONS(4701), 1,
      sym__concat,
    STATE(1721), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1281), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1271), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [81659] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1342), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1340), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [81712] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4699), 1,
      aux_sym_concatenation_token1,
    ACTIONS(4701), 1,
      sym__concat,
    ACTIONS(5952), 1,
      anon_sym_LPAREN,
    STATE(1721), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1281), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1271), 36,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [81773] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5928), 1,
      sym__special_character,
    STATE(1646), 1,
      aux_sym__literal_repeat1,
    ACTIONS(5069), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5067), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [81830] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4699), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5955), 1,
      sym__concat,
    STATE(965), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1288), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1286), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LPAREN,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [81889] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5957), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5960), 1,
      sym__concat,
    STATE(1722), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1253), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1251), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [81948] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1334), 6,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1332), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [82001] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1263), 6,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1261), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [82054] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5758), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5760), 1,
      sym__concat,
    STATE(1640), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1263), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1261), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [82113] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1318), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1316), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [82165] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5844), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5963), 1,
      sym__concat,
    STATE(1722), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1288), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1286), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [82223] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1302), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1300), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [82275] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5768), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5766), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [82327] = 27,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1131), 1,
      anon_sym_LPAREN,
    ACTIONS(1141), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1145), 1,
      anon_sym_DOLLAR,
    ACTIONS(1149), 1,
      anon_sym_DQUOTE,
    ACTIONS(1153), 1,
      aux_sym_number_token1,
    ACTIONS(1155), 1,
      aux_sym_number_token2,
    ACTIONS(1157), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1159), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1163), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1169), 1,
      sym__brace_start,
    ACTIONS(1187), 1,
      sym_word,
    ACTIONS(1189), 1,
      anon_sym_BANG,
    ACTIONS(1195), 1,
      anon_sym_TILDE,
    ACTIONS(1201), 1,
      sym_test_operator,
    ACTIONS(3060), 1,
      anon_sym_BQUOTE,
    ACTIONS(5779), 1,
      sym__special_character,
    STATE(2717), 1,
      aux_sym__literal_repeat1,
    STATE(3128), 1,
      sym__expression,
    STATE(7408), 1,
      sym__test_command_binary_expression,
    ACTIONS(1129), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1165), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1191), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1193), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1199), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2594), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2503), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [82427] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5772), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5770), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [82479] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1298), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1296), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [82531] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1253), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1251), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [82583] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1326), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1324), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [82635] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5375), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5373), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [82687] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5965), 1,
      sym__special_character,
    STATE(1736), 1,
      aux_sym__literal_repeat1,
    ACTIONS(1364), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1362), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [82743] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1350), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1348), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [82795] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5968), 1,
      sym__special_character,
    STATE(1751), 1,
      aux_sym__literal_repeat1,
    ACTIONS(4469), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4467), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [82851] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5968), 1,
      sym__special_character,
    STATE(1751), 1,
      aux_sym__literal_repeat1,
    ACTIONS(5723), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5721), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [82907] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5968), 1,
      sym__special_character,
    STATE(1751), 1,
      aux_sym__literal_repeat1,
    ACTIONS(5697), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5695), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [82963] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5968), 1,
      sym__special_character,
    STATE(1751), 1,
      aux_sym__literal_repeat1,
    ACTIONS(4566), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4564), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [83019] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5968), 1,
      sym__special_character,
    STATE(1751), 1,
      aux_sym__literal_repeat1,
    ACTIONS(5069), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5067), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [83075] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5970), 1,
      sym__special_character,
    STATE(1736), 1,
      aux_sym__literal_repeat1,
    ACTIONS(5180), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5178), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [83131] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5968), 1,
      sym__special_character,
    STATE(1751), 1,
      aux_sym__literal_repeat1,
    ACTIONS(5395), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5393), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [83187] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5080), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5082), 1,
      sym__concat,
    STATE(1822), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4554), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4552), 36,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [83245] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1334), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1332), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [83297] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5375), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5373), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [83349] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5772), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5770), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [83401] = 27,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(187), 1,
      anon_sym_TILDE,
    ACTIONS(350), 1,
      sym_word,
    ACTIONS(359), 1,
      anon_sym_LPAREN,
    ACTIONS(361), 1,
      anon_sym_BANG,
    ACTIONS(363), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(367), 1,
      anon_sym_DOLLAR,
    ACTIONS(371), 1,
      anon_sym_DQUOTE,
    ACTIONS(375), 1,
      aux_sym_number_token1,
    ACTIONS(377), 1,
      aux_sym_number_token2,
    ACTIONS(379), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(381), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(385), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(389), 1,
      sym_test_operator,
    ACTIONS(391), 1,
      sym__brace_start,
    ACTIONS(5807), 1,
      sym__special_character,
    ACTIONS(5809), 1,
      anon_sym_BQUOTE,
    ACTIONS(5972), 1,
      sym__regex_no_space,
    STATE(2641), 1,
      aux_sym__literal_repeat1,
    STATE(2989), 1,
      sym__expression,
    ACTIONS(183), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(185), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(352), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(373), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(387), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3071), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2521), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [83501] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5080), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5082), 1,
      sym__concat,
    STATE(1821), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4566), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4564), 36,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [83559] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5974), 1,
      sym__special_character,
    STATE(1751), 1,
      aux_sym__literal_repeat1,
    ACTIONS(1364), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1362), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [83615] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5080), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5082), 1,
      sym__concat,
    STATE(1822), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4372), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4370), 36,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [83673] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1294), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1292), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [83725] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1310), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1308), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [83777] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1306), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1304), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [83829] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1310), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1308), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [83881] = 27,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(223), 1,
      anon_sym_RPAREN_RPAREN,
    ACTIONS(1131), 1,
      anon_sym_LPAREN,
    ACTIONS(1141), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1145), 1,
      anon_sym_DOLLAR,
    ACTIONS(1149), 1,
      anon_sym_DQUOTE,
    ACTIONS(1153), 1,
      aux_sym_number_token1,
    ACTIONS(1155), 1,
      aux_sym_number_token2,
    ACTIONS(1157), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1159), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1163), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1169), 1,
      sym__brace_start,
    ACTIONS(1171), 1,
      sym_word,
    ACTIONS(1173), 1,
      anon_sym_BANG,
    ACTIONS(1179), 1,
      anon_sym_TILDE,
    ACTIONS(1181), 1,
      sym__special_character,
    ACTIONS(1185), 1,
      sym_test_operator,
    ACTIONS(3060), 1,
      anon_sym_BQUOTE,
    STATE(2484), 1,
      aux_sym__literal_repeat1,
    STATE(3123), 1,
      sym__expression,
    ACTIONS(1129), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1165), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1175), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1177), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1183), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2594), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2456), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [83981] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5080), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5082), 1,
      sym__concat,
    STATE(1821), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4469), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4467), 36,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [84039] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5768), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5766), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [84091] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1310), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1308), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [84143] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1306), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1304), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [84195] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1310), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1308), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [84247] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1354), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1352), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [84299] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1314), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1312), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [84351] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1342), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1340), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [84403] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1314), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1312), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [84455] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2096), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(2094), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [84507] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1338), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1336), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [84559] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1346), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1344), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [84611] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1338), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1336), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [84663] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1358), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1356), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [84715] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1322), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1320), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [84767] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1346), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1344), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [84819] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1354), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1352), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [84871] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1358), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1356), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [84923] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2074), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(2072), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [84975] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1342), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1340), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [85027] = 27,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(363), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(367), 1,
      anon_sym_DOLLAR,
    ACTIONS(371), 1,
      anon_sym_DQUOTE,
    ACTIONS(375), 1,
      aux_sym_number_token1,
    ACTIONS(377), 1,
      aux_sym_number_token2,
    ACTIONS(379), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(381), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(385), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(391), 1,
      sym__brace_start,
    ACTIONS(1091), 1,
      sym_word,
    ACTIONS(1093), 1,
      anon_sym_LPAREN,
    ACTIONS(1095), 1,
      anon_sym_BANG,
    ACTIONS(1103), 1,
      anon_sym_TILDE,
    ACTIONS(1113), 1,
      sym_test_operator,
    ACTIONS(5809), 1,
      anon_sym_BQUOTE,
    ACTIONS(5916), 1,
      sym__special_character,
    ACTIONS(5977), 1,
      sym__regex_no_space,
    STATE(2472), 1,
      aux_sym__literal_repeat1,
    STATE(2835), 1,
      sym__expression,
    ACTIONS(352), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(387), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1099), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1101), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1109), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2863), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2451), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [85127] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5920), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5979), 1,
      sym__concat,
    STATE(1704), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1267), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1265), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [85185] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5920), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5981), 1,
      sym__concat,
    STATE(1704), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1288), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1286), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [85243] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1330), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1328), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [85295] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1318), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1316), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [85347] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1330), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1328), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [85399] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5080), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5082), 1,
      sym__concat,
    STATE(1822), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4465), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4463), 36,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [85457] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1322), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1320), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [85509] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5920), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5946), 1,
      sym__concat,
    STATE(1779), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1263), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1261), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [85567] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5080), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5082), 1,
      sym__concat,
    STATE(1821), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4482), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4480), 36,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [85625] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1318), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1316), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [85677] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5983), 1,
      sym__special_character,
    STATE(1789), 1,
      aux_sym__literal_repeat1,
    ACTIONS(1364), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1362), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [85733] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1318), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1316), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [85785] = 27,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(225), 1,
      anon_sym_RPAREN_RPAREN,
    ACTIONS(1131), 1,
      anon_sym_LPAREN,
    ACTIONS(1141), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1145), 1,
      anon_sym_DOLLAR,
    ACTIONS(1149), 1,
      anon_sym_DQUOTE,
    ACTIONS(1153), 1,
      aux_sym_number_token1,
    ACTIONS(1155), 1,
      aux_sym_number_token2,
    ACTIONS(1157), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1159), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1163), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1169), 1,
      sym__brace_start,
    ACTIONS(1171), 1,
      sym_word,
    ACTIONS(1173), 1,
      anon_sym_BANG,
    ACTIONS(1179), 1,
      anon_sym_TILDE,
    ACTIONS(1181), 1,
      sym__special_character,
    ACTIONS(1185), 1,
      sym_test_operator,
    ACTIONS(3060), 1,
      anon_sym_BQUOTE,
    STATE(2484), 1,
      aux_sym__literal_repeat1,
    STATE(3144), 1,
      sym__expression,
    ACTIONS(1129), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1165), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1175), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1177), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1183), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2594), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2456), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [85885] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1322), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1320), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [85937] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1330), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1328), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [85989] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5080), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5082), 1,
      sym__concat,
    STATE(1822), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4532), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4530), 36,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [86047] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5080), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5082), 1,
      sym__concat,
    STATE(1821), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4548), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4546), 36,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [86105] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1342), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1340), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [86157] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(1897), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5986), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(241), 16,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_LPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(278), 25,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [86213] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1354), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1352), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [86265] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(1901), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5986), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(1271), 16,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_LPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1281), 25,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [86321] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1371), 1,
      anon_sym_LPAREN,
    STATE(1901), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5986), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(1271), 15,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1281), 25,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [86379] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5988), 1,
      sym__special_character,
    STATE(1829), 1,
      aux_sym__literal_repeat1,
    ACTIONS(5395), 6,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5393), 36,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [86435] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1294), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1292), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [86487] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5990), 1,
      sym__special_character,
    STATE(1886), 1,
      aux_sym__literal_repeat1,
    ACTIONS(5395), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5393), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [86543] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1358), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1356), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [86595] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1346), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1344), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [86647] = 27,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(227), 1,
      anon_sym_RPAREN_RPAREN,
    ACTIONS(1131), 1,
      anon_sym_LPAREN,
    ACTIONS(1141), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1145), 1,
      anon_sym_DOLLAR,
    ACTIONS(1149), 1,
      anon_sym_DQUOTE,
    ACTIONS(1153), 1,
      aux_sym_number_token1,
    ACTIONS(1155), 1,
      aux_sym_number_token2,
    ACTIONS(1157), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1159), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1163), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1169), 1,
      sym__brace_start,
    ACTIONS(1171), 1,
      sym_word,
    ACTIONS(1173), 1,
      anon_sym_BANG,
    ACTIONS(1179), 1,
      anon_sym_TILDE,
    ACTIONS(1181), 1,
      sym__special_character,
    ACTIONS(1185), 1,
      sym_test_operator,
    ACTIONS(3060), 1,
      anon_sym_BQUOTE,
    STATE(2484), 1,
      aux_sym__literal_repeat1,
    STATE(3154), 1,
      sym__expression,
    ACTIONS(1129), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1165), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1175), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1177), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1183), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2594), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2456), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [86747] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1334), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1332), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [86799] = 6,
    ACTIONS(71), 1,
      sym_comment,
    STATE(3543), 1,
      aux_sym__literal_repeat1,
    STATE(1868), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    STATE(3480), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2094), 10,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(2096), 22,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [86857] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1302), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1300), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [86909] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1338), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1336), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [86961] = 12,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(4253), 1,
      anon_sym_PIPE,
    ACTIONS(4348), 1,
      anon_sym_PIPE_AMP,
    ACTIONS(5994), 1,
      anon_sym_LT_LT,
    ACTIONS(5996), 1,
      anon_sym_LT_LT_DASH,
    ACTIONS(5998), 1,
      sym_variable_name,
    STATE(6737), 1,
      sym_subscript,
    ACTIONS(5992), 2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
    STATE(5360), 2,
      sym_variable_assignment,
      aux_sym_variable_assignments_repeat1,
    STATE(5350), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(4247), 10,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(4261), 21,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [87031] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1298), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1296), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [87083] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1314), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1312), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [87135] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(6001), 1,
      sym__special_character,
    STATE(1814), 1,
      aux_sym__literal_repeat1,
    ACTIONS(1364), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1362), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [87191] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1294), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1292), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [87243] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(6004), 1,
      sym__special_character,
    STATE(1873), 1,
      aux_sym__literal_repeat1,
    ACTIONS(5180), 6,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5178), 36,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [87299] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1310), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1308), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [87351] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1306), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1304), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [87403] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1310), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1308), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [87455] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1334), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1332), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [87507] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5080), 1,
      aux_sym_concatenation_token1,
    ACTIONS(6006), 1,
      sym__concat,
    STATE(1065), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1267), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1265), 36,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [87565] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5080), 1,
      aux_sym_concatenation_token1,
    ACTIONS(6008), 1,
      sym__concat,
    STATE(1065), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1288), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1286), 36,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [87623] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5080), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5082), 1,
      sym__concat,
    STATE(1821), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1263), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1261), 36,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [87681] = 27,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1131), 1,
      anon_sym_LPAREN,
    ACTIONS(1141), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1145), 1,
      anon_sym_DOLLAR,
    ACTIONS(1149), 1,
      anon_sym_DQUOTE,
    ACTIONS(1153), 1,
      aux_sym_number_token1,
    ACTIONS(1155), 1,
      aux_sym_number_token2,
    ACTIONS(1157), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1159), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1163), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1169), 1,
      sym__brace_start,
    ACTIONS(1187), 1,
      sym_word,
    ACTIONS(1189), 1,
      anon_sym_BANG,
    ACTIONS(1195), 1,
      anon_sym_TILDE,
    ACTIONS(1201), 1,
      sym_test_operator,
    ACTIONS(3060), 1,
      anon_sym_BQUOTE,
    ACTIONS(5779), 1,
      sym__special_character,
    STATE(2717), 1,
      aux_sym__literal_repeat1,
    STATE(3125), 1,
      sym__expression,
    STATE(7188), 1,
      sym__test_command_binary_expression,
    ACTIONS(1129), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1165), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1191), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1193), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1199), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2594), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2503), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [87781] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1350), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1348), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [87833] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5988), 1,
      sym__special_character,
    STATE(1829), 1,
      aux_sym__literal_repeat1,
    ACTIONS(5723), 6,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5721), 36,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [87889] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1253), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1251), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [87941] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5988), 1,
      sym__special_character,
    STATE(1829), 1,
      aux_sym__literal_repeat1,
    ACTIONS(4469), 6,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4467), 36,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [87997] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(6010), 1,
      sym__special_character,
    STATE(1829), 1,
      aux_sym__literal_repeat1,
    ACTIONS(1364), 6,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1362), 36,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [88053] = 27,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1131), 1,
      anon_sym_LPAREN,
    ACTIONS(1141), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1145), 1,
      anon_sym_DOLLAR,
    ACTIONS(1149), 1,
      anon_sym_DQUOTE,
    ACTIONS(1153), 1,
      aux_sym_number_token1,
    ACTIONS(1155), 1,
      aux_sym_number_token2,
    ACTIONS(1157), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1159), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1163), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1169), 1,
      sym__brace_start,
    ACTIONS(1187), 1,
      sym_word,
    ACTIONS(1189), 1,
      anon_sym_BANG,
    ACTIONS(1195), 1,
      anon_sym_TILDE,
    ACTIONS(1201), 1,
      sym_test_operator,
    ACTIONS(3060), 1,
      anon_sym_BQUOTE,
    ACTIONS(5779), 1,
      sym__special_character,
    STATE(2717), 1,
      aux_sym__literal_repeat1,
    STATE(3141), 1,
      sym__expression,
    STATE(6913), 1,
      sym__test_command_binary_expression,
    ACTIONS(1129), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1165), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1191), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1193), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1199), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2594), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2503), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [88153] = 6,
    ACTIONS(71), 1,
      sym_comment,
    STATE(3543), 1,
      aux_sym__literal_repeat1,
    STATE(1868), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    STATE(3480), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2072), 10,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(2074), 22,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [88211] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(6013), 1,
      aux_sym_concatenation_token1,
    ACTIONS(6015), 1,
      sym__concat,
    STATE(1894), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1263), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1261), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [88269] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1302), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1300), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [88321] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1326), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1324), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [88373] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1263), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1261), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [88425] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1310), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1308), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [88477] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1306), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1304), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [88529] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1310), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1308), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [88581] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(6013), 1,
      aux_sym_concatenation_token1,
    ACTIONS(6015), 1,
      sym__concat,
    STATE(1889), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5391), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5389), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [88639] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1298), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1296), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [88691] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1314), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1312), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [88743] = 27,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1131), 1,
      anon_sym_LPAREN,
    ACTIONS(1141), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1145), 1,
      anon_sym_DOLLAR,
    ACTIONS(1149), 1,
      anon_sym_DQUOTE,
    ACTIONS(1153), 1,
      aux_sym_number_token1,
    ACTIONS(1155), 1,
      aux_sym_number_token2,
    ACTIONS(1157), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1159), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1163), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1169), 1,
      sym__brace_start,
    ACTIONS(1187), 1,
      sym_word,
    ACTIONS(1189), 1,
      anon_sym_BANG,
    ACTIONS(1195), 1,
      anon_sym_TILDE,
    ACTIONS(1201), 1,
      sym_test_operator,
    ACTIONS(3060), 1,
      anon_sym_BQUOTE,
    ACTIONS(5779), 1,
      sym__special_character,
    STATE(2717), 1,
      aux_sym__literal_repeat1,
    STATE(3147), 1,
      sym__expression,
    STATE(7121), 1,
      sym__test_command_binary_expression,
    ACTIONS(1129), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1165), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1191), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1193), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1199), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2594), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2503), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [88843] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(6013), 1,
      aux_sym_concatenation_token1,
    ACTIONS(6015), 1,
      sym__concat,
    STATE(1894), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5395), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5393), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [88901] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1263), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1261), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [88953] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(6017), 1,
      aux_sym_concatenation_token1,
    ACTIONS(6019), 1,
      sym__concat,
    STATE(1908), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5395), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5393), 36,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [89011] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(6017), 1,
      aux_sym_concatenation_token1,
    ACTIONS(6019), 1,
      sym__concat,
    STATE(1910), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5391), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5389), 36,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [89069] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1302), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1300), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [89121] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1354), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1352), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [89173] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1294), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1292), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [89225] = 27,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(107), 1,
      anon_sym_TILDE,
    ACTIONS(237), 1,
      sym_word,
    ACTIONS(248), 1,
      anon_sym_LPAREN,
    ACTIONS(250), 1,
      anon_sym_BANG,
    ACTIONS(258), 1,
      anon_sym_DOLLAR,
    ACTIONS(264), 1,
      aux_sym_number_token1,
    ACTIONS(266), 1,
      aux_sym_number_token2,
    ACTIONS(270), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(282), 1,
      sym_test_operator,
    ACTIONS(284), 1,
      sym__brace_start,
    ACTIONS(1075), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1079), 1,
      sym__special_character,
    ACTIONS(1081), 1,
      anon_sym_DQUOTE,
    ACTIONS(1085), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1087), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3598), 1,
      anon_sym_BQUOTE,
    ACTIONS(6021), 1,
      sym__regex_no_space,
    STATE(2690), 1,
      aux_sym__literal_repeat1,
    STATE(3095), 1,
      sym__expression,
    ACTIONS(103), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(105), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1073), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1083), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(1089), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3053), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2500), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [89325] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1253), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1251), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [89377] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1298), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1296), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [89429] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1310), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1308), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [89481] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1306), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1304), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [89533] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1310), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1308), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [89585] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1314), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1312), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [89637] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1342), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1340), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [89689] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1334), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1332), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [89741] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1253), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1251), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [89793] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(6023), 1,
      sym__special_character,
    STATE(1814), 1,
      aux_sym__literal_repeat1,
    ACTIONS(5478), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5476), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [89849] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1326), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1324), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [89901] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1338), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1336), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [89953] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1326), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1324), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [90005] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1350), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1348), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [90057] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1346), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1344), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [90109] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1358), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1356), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [90161] = 27,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1131), 1,
      anon_sym_LPAREN,
    ACTIONS(1141), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1145), 1,
      anon_sym_DOLLAR,
    ACTIONS(1149), 1,
      anon_sym_DQUOTE,
    ACTIONS(1153), 1,
      aux_sym_number_token1,
    ACTIONS(1155), 1,
      aux_sym_number_token2,
    ACTIONS(1157), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1159), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1163), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1169), 1,
      sym__brace_start,
    ACTIONS(1187), 1,
      sym_word,
    ACTIONS(1189), 1,
      anon_sym_BANG,
    ACTIONS(1195), 1,
      anon_sym_TILDE,
    ACTIONS(1201), 1,
      sym_test_operator,
    ACTIONS(3060), 1,
      anon_sym_BQUOTE,
    ACTIONS(5779), 1,
      sym__special_character,
    ACTIONS(6025), 1,
      sym__regex_no_space,
    STATE(2717), 1,
      aux_sym__literal_repeat1,
    STATE(2995), 1,
      sym__expression,
    ACTIONS(1129), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1165), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1191), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1193), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1199), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2594), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2503), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [90261] = 22,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6027), 1,
      sym_word,
    ACTIONS(6033), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(6036), 1,
      anon_sym_DOLLAR,
    ACTIONS(6039), 1,
      sym__special_character,
    ACTIONS(6042), 1,
      anon_sym_DQUOTE,
    ACTIONS(6048), 1,
      aux_sym_number_token1,
    ACTIONS(6051), 1,
      aux_sym_number_token2,
    ACTIONS(6054), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(6057), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(6060), 1,
      anon_sym_BQUOTE,
    ACTIONS(6063), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(6069), 1,
      sym_test_operator,
    ACTIONS(6072), 1,
      sym__brace_start,
    STATE(3543), 1,
      aux_sym__literal_repeat1,
    ACTIONS(6030), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(6045), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(6066), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(1868), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    ACTIONS(2214), 5,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(2216), 8,
      sym_file_descriptor,
      sym_variable_name,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
    STATE(3480), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [90351] = 27,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1131), 1,
      anon_sym_LPAREN,
    ACTIONS(1141), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1145), 1,
      anon_sym_DOLLAR,
    ACTIONS(1149), 1,
      anon_sym_DQUOTE,
    ACTIONS(1153), 1,
      aux_sym_number_token1,
    ACTIONS(1155), 1,
      aux_sym_number_token2,
    ACTIONS(1157), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1159), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1163), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1169), 1,
      sym__brace_start,
    ACTIONS(1187), 1,
      sym_word,
    ACTIONS(1189), 1,
      anon_sym_BANG,
    ACTIONS(1195), 1,
      anon_sym_TILDE,
    ACTIONS(1201), 1,
      sym_test_operator,
    ACTIONS(3060), 1,
      anon_sym_BQUOTE,
    ACTIONS(5779), 1,
      sym__special_character,
    STATE(2717), 1,
      aux_sym__literal_repeat1,
    STATE(3153), 1,
      sym__expression,
    STATE(7318), 1,
      sym__test_command_binary_expression,
    ACTIONS(1129), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1165), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1191), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1193), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1199), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2594), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2503), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [90451] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1354), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1352), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [90503] = 27,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1131), 1,
      anon_sym_LPAREN,
    ACTIONS(1141), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1145), 1,
      anon_sym_DOLLAR,
    ACTIONS(1149), 1,
      anon_sym_DQUOTE,
    ACTIONS(1153), 1,
      aux_sym_number_token1,
    ACTIONS(1155), 1,
      aux_sym_number_token2,
    ACTIONS(1157), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1159), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1163), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1169), 1,
      sym__brace_start,
    ACTIONS(1187), 1,
      sym_word,
    ACTIONS(1189), 1,
      anon_sym_BANG,
    ACTIONS(1195), 1,
      anon_sym_TILDE,
    ACTIONS(1201), 1,
      sym_test_operator,
    ACTIONS(3060), 1,
      anon_sym_BQUOTE,
    ACTIONS(5779), 1,
      sym__special_character,
    STATE(2717), 1,
      aux_sym__literal_repeat1,
    STATE(3156), 1,
      sym__expression,
    STATE(7444), 1,
      sym__test_command_binary_expression,
    ACTIONS(1129), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1165), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1191), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1193), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1199), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2594), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2503), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [90603] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4372), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4370), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [90655] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(6075), 1,
      sym__special_character,
    STATE(1873), 1,
      aux_sym__literal_repeat1,
    ACTIONS(1364), 6,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1362), 36,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [90711] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1338), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1336), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [90763] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1330), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1328), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [90815] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1322), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1320), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [90867] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1342), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1340), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [90919] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1350), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1348), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [90971] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(6078), 1,
      aux_sym_concatenation_token1,
    ACTIONS(6081), 1,
      sym__concat,
    STATE(1879), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1253), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1251), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [91029] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1346), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1344), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [91081] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1358), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1356), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [91133] = 11,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(4253), 1,
      anon_sym_PIPE,
    ACTIONS(4270), 1,
      anon_sym_LT_LT,
    ACTIONS(4348), 1,
      anon_sym_PIPE_AMP,
    ACTIONS(5998), 1,
      sym_variable_name,
    STATE(6737), 1,
      sym_subscript,
    STATE(5360), 2,
      sym_variable_assignment,
      aux_sym_variable_assignments_repeat1,
    ACTIONS(4272), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_LT_DASH,
    STATE(5436), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(4247), 10,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(4261), 21,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [91201] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4554), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4552), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [91253] = 27,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1131), 1,
      anon_sym_LPAREN,
    ACTIONS(1141), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1145), 1,
      anon_sym_DOLLAR,
    ACTIONS(1149), 1,
      anon_sym_DQUOTE,
    ACTIONS(1153), 1,
      aux_sym_number_token1,
    ACTIONS(1155), 1,
      aux_sym_number_token2,
    ACTIONS(1157), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1159), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1163), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1169), 1,
      sym__brace_start,
    ACTIONS(1187), 1,
      sym_word,
    ACTIONS(1189), 1,
      anon_sym_BANG,
    ACTIONS(1195), 1,
      anon_sym_TILDE,
    ACTIONS(1201), 1,
      sym_test_operator,
    ACTIONS(3060), 1,
      anon_sym_BQUOTE,
    ACTIONS(5779), 1,
      sym__special_character,
    STATE(2717), 1,
      aux_sym__literal_repeat1,
    STATE(3087), 1,
      sym__expression,
    STATE(7159), 1,
      sym__test_command_binary_expression,
    ACTIONS(1129), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1165), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1191), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1193), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1199), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2594), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2503), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [91353] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1263), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1261), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [91405] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(6084), 1,
      sym__special_character,
    STATE(1886), 1,
      aux_sym__literal_repeat1,
    ACTIONS(1364), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1362), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [91461] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5970), 1,
      sym__special_character,
    STATE(1736), 1,
      aux_sym__literal_repeat1,
    ACTIONS(5069), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5067), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [91517] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1330), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1328), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [91569] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(6013), 1,
      aux_sym_concatenation_token1,
    ACTIONS(6087), 1,
      sym__concat,
    STATE(1879), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1288), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1286), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [91627] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1263), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1261), 40,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [91679] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1322), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1320), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [91731] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5826), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5828), 1,
      sym_file_descriptor,
    ACTIONS(5831), 3,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(5821), 11,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
    ACTIONS(5823), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
    ACTIONS(5819), 17,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [91791] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5826), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5828), 1,
      sym_file_descriptor,
    ACTIONS(5831), 3,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(5821), 11,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
    ACTIONS(5823), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
    ACTIONS(5819), 17,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [91851] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(6013), 1,
      aux_sym_concatenation_token1,
    ACTIONS(6089), 1,
      sym__concat,
    STATE(1879), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1267), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1265), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [91909] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5988), 1,
      sym__special_character,
    STATE(1829), 1,
      aux_sym__literal_repeat1,
    ACTIONS(5697), 6,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5695), 36,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [91965] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1318), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1316), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [92017] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(5986), 1,
      aux_sym_concatenation_token1,
    ACTIONS(6091), 1,
      sym__concat,
    STATE(1904), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1265), 16,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_LPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1267), 25,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [92075] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5988), 1,
      sym__special_character,
    STATE(1829), 1,
      aux_sym__literal_repeat1,
    ACTIONS(4566), 6,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4564), 36,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [92131] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(6093), 1,
      sym__special_character,
    STATE(1789), 1,
      aux_sym__literal_repeat1,
    ACTIONS(5395), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5393), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [92187] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5988), 1,
      sym__special_character,
    STATE(1829), 1,
      aux_sym__literal_repeat1,
    ACTIONS(5069), 6,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5067), 36,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [92243] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(5986), 1,
      aux_sym_concatenation_token1,
    ACTIONS(6095), 1,
      sym__concat,
    STATE(1904), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1286), 16,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_LPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1288), 25,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [92301] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5844), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5846), 1,
      sym__concat,
    STATE(1727), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5470), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5468), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [92359] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5844), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5846), 1,
      sym__concat,
    STATE(1914), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5478), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5476), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [92417] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(1904), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(6097), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(1251), 16,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_LPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1253), 25,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [92473] = 27,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1131), 1,
      anon_sym_LPAREN,
    ACTIONS(1141), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1145), 1,
      anon_sym_DOLLAR,
    ACTIONS(1149), 1,
      anon_sym_DQUOTE,
    ACTIONS(1153), 1,
      aux_sym_number_token1,
    ACTIONS(1155), 1,
      aux_sym_number_token2,
    ACTIONS(1157), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1159), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1163), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1169), 1,
      sym__brace_start,
    ACTIONS(1187), 1,
      sym_word,
    ACTIONS(1189), 1,
      anon_sym_BANG,
    ACTIONS(1195), 1,
      anon_sym_TILDE,
    ACTIONS(1201), 1,
      sym_test_operator,
    ACTIONS(3060), 1,
      anon_sym_BQUOTE,
    ACTIONS(5779), 1,
      sym__special_character,
    STATE(2717), 1,
      aux_sym__literal_repeat1,
    STATE(3040), 1,
      sym__expression,
    STATE(6920), 1,
      sym__test_command_binary_expression,
    ACTIONS(1129), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1165), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1191), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1193), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1199), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2594), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2503), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [92573] = 27,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1131), 1,
      anon_sym_LPAREN,
    ACTIONS(1141), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1145), 1,
      anon_sym_DOLLAR,
    ACTIONS(1149), 1,
      anon_sym_DQUOTE,
    ACTIONS(1153), 1,
      aux_sym_number_token1,
    ACTIONS(1155), 1,
      aux_sym_number_token2,
    ACTIONS(1157), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1159), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1163), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1169), 1,
      sym__brace_start,
    ACTIONS(1211), 1,
      sym_word,
    ACTIONS(1213), 1,
      anon_sym_BANG,
    ACTIONS(1219), 1,
      anon_sym_TILDE,
    ACTIONS(1221), 1,
      sym__special_character,
    ACTIONS(1225), 1,
      sym_test_operator,
    ACTIONS(3060), 1,
      anon_sym_BQUOTE,
    ACTIONS(6025), 1,
      sym__regex_no_space,
    STATE(2484), 1,
      aux_sym__literal_repeat1,
    STATE(3314), 1,
      sym__expression,
    ACTIONS(1129), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1165), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1215), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1217), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1223), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2594), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2588), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [92673] = 9,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(5998), 1,
      sym_variable_name,
    STATE(6737), 1,
      sym_subscript,
    STATE(5360), 2,
      sym_variable_assignment,
      aux_sym_variable_assignments_repeat1,
    STATE(5436), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(4247), 5,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(4253), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(4348), 10,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
    ACTIONS(4261), 15,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [92737] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(6017), 1,
      aux_sym_concatenation_token1,
    ACTIONS(6100), 1,
      sym__concat,
    STATE(1911), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1267), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1265), 36,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [92795] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5932), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4261), 3,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(4247), 18,
      anon_sym_LPAREN_LPAREN,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
    ACTIONS(5930), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [92851] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(6017), 1,
      aux_sym_concatenation_token1,
    ACTIONS(6102), 1,
      sym__concat,
    STATE(1911), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1288), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1286), 36,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [92909] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(6104), 1,
      aux_sym_concatenation_token1,
    ACTIONS(6107), 1,
      sym__concat,
    STATE(1911), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1253), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1251), 36,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [92967] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(1897), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5986), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(1261), 16,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_LPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1263), 25,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [93023] = 27,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(233), 1,
      anon_sym_RPAREN_RPAREN,
    ACTIONS(1131), 1,
      anon_sym_LPAREN,
    ACTIONS(1141), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1145), 1,
      anon_sym_DOLLAR,
    ACTIONS(1149), 1,
      anon_sym_DQUOTE,
    ACTIONS(1153), 1,
      aux_sym_number_token1,
    ACTIONS(1155), 1,
      aux_sym_number_token2,
    ACTIONS(1157), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1159), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1163), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1169), 1,
      sym__brace_start,
    ACTIONS(1171), 1,
      sym_word,
    ACTIONS(1173), 1,
      anon_sym_BANG,
    ACTIONS(1179), 1,
      anon_sym_TILDE,
    ACTIONS(1181), 1,
      sym__special_character,
    ACTIONS(1185), 1,
      sym_test_operator,
    ACTIONS(3060), 1,
      anon_sym_BQUOTE,
    STATE(2484), 1,
      aux_sym__literal_repeat1,
    STATE(3073), 1,
      sym__expression,
    ACTIONS(1129), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1165), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1175), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1177), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1183), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2594), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2456), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [93123] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5844), 1,
      aux_sym_concatenation_token1,
    ACTIONS(6110), 1,
      sym__concat,
    STATE(1722), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1267), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1265), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [93181] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5844), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5846), 1,
      sym__concat,
    STATE(1914), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1263), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1261), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [93239] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(6017), 1,
      aux_sym_concatenation_token1,
    ACTIONS(6019), 1,
      sym__concat,
    STATE(1908), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1263), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1261), 36,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [93297] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1294), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1292), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [93349] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5920), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5946), 1,
      sym__concat,
    STATE(1780), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5391), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5389), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [93407] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1350), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1348), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [93459] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1302), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1300), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [93511] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1253), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1251), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [93563] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5920), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5946), 1,
      sym__concat,
    STATE(1779), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5395), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5393), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [93621] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1334), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1332), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [93673] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1326), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1324), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [93725] = 27,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1131), 1,
      anon_sym_LPAREN,
    ACTIONS(1141), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1145), 1,
      anon_sym_DOLLAR,
    ACTIONS(1149), 1,
      anon_sym_DQUOTE,
    ACTIONS(1153), 1,
      aux_sym_number_token1,
    ACTIONS(1155), 1,
      aux_sym_number_token2,
    ACTIONS(1157), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1159), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1163), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1169), 1,
      sym__brace_start,
    ACTIONS(1171), 1,
      sym_word,
    ACTIONS(1173), 1,
      anon_sym_BANG,
    ACTIONS(1179), 1,
      anon_sym_TILDE,
    ACTIONS(1181), 1,
      sym__special_character,
    ACTIONS(1185), 1,
      sym_test_operator,
    ACTIONS(3060), 1,
      anon_sym_BQUOTE,
    ACTIONS(6025), 1,
      sym__regex_no_space,
    STATE(2484), 1,
      aux_sym__literal_repeat1,
    STATE(3067), 1,
      sym__expression,
    ACTIONS(1129), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1165), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1175), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1177), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1183), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2594), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2456), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [93825] = 27,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1131), 1,
      anon_sym_LPAREN,
    ACTIONS(1141), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1145), 1,
      anon_sym_DOLLAR,
    ACTIONS(1149), 1,
      anon_sym_DQUOTE,
    ACTIONS(1153), 1,
      aux_sym_number_token1,
    ACTIONS(1155), 1,
      aux_sym_number_token2,
    ACTIONS(1157), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1159), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1163), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1169), 1,
      sym__brace_start,
    ACTIONS(1187), 1,
      sym_word,
    ACTIONS(1189), 1,
      anon_sym_BANG,
    ACTIONS(1195), 1,
      anon_sym_TILDE,
    ACTIONS(1201), 1,
      sym_test_operator,
    ACTIONS(3060), 1,
      anon_sym_BQUOTE,
    ACTIONS(5779), 1,
      sym__special_character,
    ACTIONS(6112), 1,
      sym__regex_no_space,
    STATE(2717), 1,
      aux_sym__literal_repeat1,
    STATE(2995), 1,
      sym__expression,
    ACTIONS(1129), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1165), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1191), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1193), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1199), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2594), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2503), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [93925] = 27,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1127), 1,
      sym_word,
    ACTIONS(1131), 1,
      anon_sym_LPAREN,
    ACTIONS(1133), 1,
      anon_sym_BANG,
    ACTIONS(1139), 1,
      anon_sym_TILDE,
    ACTIONS(1141), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1145), 1,
      anon_sym_DOLLAR,
    ACTIONS(1147), 1,
      sym__special_character,
    ACTIONS(1149), 1,
      anon_sym_DQUOTE,
    ACTIONS(1153), 1,
      aux_sym_number_token1,
    ACTIONS(1155), 1,
      aux_sym_number_token2,
    ACTIONS(1157), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1159), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1163), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1167), 1,
      sym_test_operator,
    ACTIONS(1169), 1,
      sym__brace_start,
    ACTIONS(3060), 1,
      anon_sym_BQUOTE,
    ACTIONS(6025), 1,
      sym__regex_no_space,
    STATE(2484), 1,
      aux_sym__literal_repeat1,
    STATE(2979), 1,
      sym__expression,
    ACTIONS(1129), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1135), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1137), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1151), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(1165), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(2594), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2573), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [94025] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1298), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1296), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [94077] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(6004), 1,
      sym__special_character,
    STATE(1873), 1,
      aux_sym__literal_repeat1,
    ACTIONS(5069), 6,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5067), 36,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [94133] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5826), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5828), 1,
      sym_file_descriptor,
    ACTIONS(6114), 1,
      anon_sym_RPAREN,
    ACTIONS(5831), 3,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(5821), 9,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
    ACTIONS(5823), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
    ACTIONS(5819), 17,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [94194] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5826), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5828), 1,
      sym_file_descriptor,
    ACTIONS(6117), 1,
      anon_sym_RPAREN,
    ACTIONS(5831), 3,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(5821), 9,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
    ACTIONS(5823), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
    ACTIONS(5819), 17,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [94255] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1127), 1,
      sym_word,
    ACTIONS(1131), 1,
      anon_sym_LPAREN,
    ACTIONS(1133), 1,
      anon_sym_BANG,
    ACTIONS(1139), 1,
      anon_sym_TILDE,
    ACTIONS(1141), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1145), 1,
      anon_sym_DOLLAR,
    ACTIONS(1147), 1,
      sym__special_character,
    ACTIONS(1149), 1,
      anon_sym_DQUOTE,
    ACTIONS(1153), 1,
      aux_sym_number_token1,
    ACTIONS(1155), 1,
      aux_sym_number_token2,
    ACTIONS(1157), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1159), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1163), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1167), 1,
      sym_test_operator,
    ACTIONS(1169), 1,
      sym__brace_start,
    ACTIONS(3060), 1,
      anon_sym_BQUOTE,
    STATE(2484), 1,
      aux_sym__literal_repeat1,
    STATE(2979), 1,
      sym__expression,
    ACTIONS(1129), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1135), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1137), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1151), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(1165), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(2594), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2573), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [94352] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1127), 1,
      sym_word,
    ACTIONS(1131), 1,
      anon_sym_LPAREN,
    ACTIONS(1133), 1,
      anon_sym_BANG,
    ACTIONS(1139), 1,
      anon_sym_TILDE,
    ACTIONS(1141), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1145), 1,
      anon_sym_DOLLAR,
    ACTIONS(1147), 1,
      sym__special_character,
    ACTIONS(1149), 1,
      anon_sym_DQUOTE,
    ACTIONS(1153), 1,
      aux_sym_number_token1,
    ACTIONS(1155), 1,
      aux_sym_number_token2,
    ACTIONS(1157), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1159), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1163), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1167), 1,
      sym_test_operator,
    ACTIONS(1169), 1,
      sym__brace_start,
    ACTIONS(3060), 1,
      anon_sym_BQUOTE,
    STATE(2484), 1,
      aux_sym__literal_repeat1,
    STATE(2978), 1,
      sym__expression,
    ACTIONS(1129), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1135), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1137), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1151), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(1165), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(2594), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2573), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [94449] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1127), 1,
      sym_word,
    ACTIONS(1131), 1,
      anon_sym_LPAREN,
    ACTIONS(1133), 1,
      anon_sym_BANG,
    ACTIONS(1139), 1,
      anon_sym_TILDE,
    ACTIONS(1141), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1145), 1,
      anon_sym_DOLLAR,
    ACTIONS(1147), 1,
      sym__special_character,
    ACTIONS(1149), 1,
      anon_sym_DQUOTE,
    ACTIONS(1153), 1,
      aux_sym_number_token1,
    ACTIONS(1155), 1,
      aux_sym_number_token2,
    ACTIONS(1157), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1159), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1163), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1167), 1,
      sym_test_operator,
    ACTIONS(1169), 1,
      sym__brace_start,
    ACTIONS(3060), 1,
      anon_sym_BQUOTE,
    STATE(2484), 1,
      aux_sym__literal_repeat1,
    STATE(2977), 1,
      sym__expression,
    ACTIONS(1129), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1135), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1137), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1151), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(1165), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(2594), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2573), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [94546] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1127), 1,
      sym_word,
    ACTIONS(1131), 1,
      anon_sym_LPAREN,
    ACTIONS(1133), 1,
      anon_sym_BANG,
    ACTIONS(1139), 1,
      anon_sym_TILDE,
    ACTIONS(1141), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1145), 1,
      anon_sym_DOLLAR,
    ACTIONS(1147), 1,
      sym__special_character,
    ACTIONS(1149), 1,
      anon_sym_DQUOTE,
    ACTIONS(1153), 1,
      aux_sym_number_token1,
    ACTIONS(1155), 1,
      aux_sym_number_token2,
    ACTIONS(1157), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1159), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1163), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1167), 1,
      sym_test_operator,
    ACTIONS(1169), 1,
      sym__brace_start,
    ACTIONS(3060), 1,
      anon_sym_BQUOTE,
    STATE(2484), 1,
      aux_sym__literal_repeat1,
    STATE(2976), 1,
      sym__expression,
    ACTIONS(1129), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1135), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1137), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1151), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(1165), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(2594), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2573), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [94643] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1127), 1,
      sym_word,
    ACTIONS(1131), 1,
      anon_sym_LPAREN,
    ACTIONS(1133), 1,
      anon_sym_BANG,
    ACTIONS(1139), 1,
      anon_sym_TILDE,
    ACTIONS(1141), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1145), 1,
      anon_sym_DOLLAR,
    ACTIONS(1147), 1,
      sym__special_character,
    ACTIONS(1149), 1,
      anon_sym_DQUOTE,
    ACTIONS(1153), 1,
      aux_sym_number_token1,
    ACTIONS(1155), 1,
      aux_sym_number_token2,
    ACTIONS(1157), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1159), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1163), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1167), 1,
      sym_test_operator,
    ACTIONS(1169), 1,
      sym__brace_start,
    ACTIONS(3060), 1,
      anon_sym_BQUOTE,
    STATE(2484), 1,
      aux_sym__literal_repeat1,
    STATE(2975), 1,
      sym__expression,
    ACTIONS(1129), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1135), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1137), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1151), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(1165), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(2594), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2573), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [94740] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1127), 1,
      sym_word,
    ACTIONS(1131), 1,
      anon_sym_LPAREN,
    ACTIONS(1133), 1,
      anon_sym_BANG,
    ACTIONS(1139), 1,
      anon_sym_TILDE,
    ACTIONS(1141), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1145), 1,
      anon_sym_DOLLAR,
    ACTIONS(1147), 1,
      sym__special_character,
    ACTIONS(1149), 1,
      anon_sym_DQUOTE,
    ACTIONS(1153), 1,
      aux_sym_number_token1,
    ACTIONS(1155), 1,
      aux_sym_number_token2,
    ACTIONS(1157), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1159), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1163), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1167), 1,
      sym_test_operator,
    ACTIONS(1169), 1,
      sym__brace_start,
    ACTIONS(3060), 1,
      anon_sym_BQUOTE,
    STATE(2484), 1,
      aux_sym__literal_repeat1,
    STATE(2974), 1,
      sym__expression,
    ACTIONS(1129), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1135), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1137), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1151), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(1165), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(2594), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2573), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [94837] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1127), 1,
      sym_word,
    ACTIONS(1131), 1,
      anon_sym_LPAREN,
    ACTIONS(1133), 1,
      anon_sym_BANG,
    ACTIONS(1139), 1,
      anon_sym_TILDE,
    ACTIONS(1141), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1145), 1,
      anon_sym_DOLLAR,
    ACTIONS(1147), 1,
      sym__special_character,
    ACTIONS(1149), 1,
      anon_sym_DQUOTE,
    ACTIONS(1153), 1,
      aux_sym_number_token1,
    ACTIONS(1155), 1,
      aux_sym_number_token2,
    ACTIONS(1157), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1159), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1163), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1167), 1,
      sym_test_operator,
    ACTIONS(1169), 1,
      sym__brace_start,
    ACTIONS(3060), 1,
      anon_sym_BQUOTE,
    STATE(2484), 1,
      aux_sym__literal_repeat1,
    STATE(2973), 1,
      sym__expression,
    ACTIONS(1129), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1135), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1137), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1151), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(1165), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(2594), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2573), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [94934] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5826), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5828), 1,
      sym_file_descriptor,
    ACTIONS(6120), 1,
      anon_sym_RPAREN,
    ACTIONS(5831), 3,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(5821), 9,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
    ACTIONS(5823), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
    ACTIONS(5819), 17,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [94995] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1127), 1,
      sym_word,
    ACTIONS(1131), 1,
      anon_sym_LPAREN,
    ACTIONS(1133), 1,
      anon_sym_BANG,
    ACTIONS(1139), 1,
      anon_sym_TILDE,
    ACTIONS(1141), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1145), 1,
      anon_sym_DOLLAR,
    ACTIONS(1147), 1,
      sym__special_character,
    ACTIONS(1149), 1,
      anon_sym_DQUOTE,
    ACTIONS(1153), 1,
      aux_sym_number_token1,
    ACTIONS(1155), 1,
      aux_sym_number_token2,
    ACTIONS(1157), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1159), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1163), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1167), 1,
      sym_test_operator,
    ACTIONS(1169), 1,
      sym__brace_start,
    ACTIONS(3060), 1,
      anon_sym_BQUOTE,
    STATE(2484), 1,
      aux_sym__literal_repeat1,
    STATE(2971), 1,
      sym__expression,
    ACTIONS(1129), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1135), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1137), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1151), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(1165), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(2594), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2573), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [95092] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1127), 1,
      sym_word,
    ACTIONS(1131), 1,
      anon_sym_LPAREN,
    ACTIONS(1133), 1,
      anon_sym_BANG,
    ACTIONS(1139), 1,
      anon_sym_TILDE,
    ACTIONS(1141), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1145), 1,
      anon_sym_DOLLAR,
    ACTIONS(1147), 1,
      sym__special_character,
    ACTIONS(1149), 1,
      anon_sym_DQUOTE,
    ACTIONS(1153), 1,
      aux_sym_number_token1,
    ACTIONS(1155), 1,
      aux_sym_number_token2,
    ACTIONS(1157), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1159), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1163), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1167), 1,
      sym_test_operator,
    ACTIONS(1169), 1,
      sym__brace_start,
    ACTIONS(3060), 1,
      anon_sym_BQUOTE,
    STATE(2484), 1,
      aux_sym__literal_repeat1,
    STATE(2970), 1,
      sym__expression,
    ACTIONS(1129), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1135), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1137), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1151), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(1165), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(2594), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2573), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [95189] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1127), 1,
      sym_word,
    ACTIONS(1131), 1,
      anon_sym_LPAREN,
    ACTIONS(1133), 1,
      anon_sym_BANG,
    ACTIONS(1139), 1,
      anon_sym_TILDE,
    ACTIONS(1141), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1145), 1,
      anon_sym_DOLLAR,
    ACTIONS(1147), 1,
      sym__special_character,
    ACTIONS(1149), 1,
      anon_sym_DQUOTE,
    ACTIONS(1153), 1,
      aux_sym_number_token1,
    ACTIONS(1155), 1,
      aux_sym_number_token2,
    ACTIONS(1157), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1159), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1163), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1167), 1,
      sym_test_operator,
    ACTIONS(1169), 1,
      sym__brace_start,
    ACTIONS(3060), 1,
      anon_sym_BQUOTE,
    STATE(2484), 1,
      aux_sym__literal_repeat1,
    STATE(2969), 1,
      sym__expression,
    ACTIONS(1129), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1135), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1137), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1151), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(1165), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(2594), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2573), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [95286] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1127), 1,
      sym_word,
    ACTIONS(1131), 1,
      anon_sym_LPAREN,
    ACTIONS(1133), 1,
      anon_sym_BANG,
    ACTIONS(1139), 1,
      anon_sym_TILDE,
    ACTIONS(1141), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1145), 1,
      anon_sym_DOLLAR,
    ACTIONS(1147), 1,
      sym__special_character,
    ACTIONS(1149), 1,
      anon_sym_DQUOTE,
    ACTIONS(1153), 1,
      aux_sym_number_token1,
    ACTIONS(1155), 1,
      aux_sym_number_token2,
    ACTIONS(1157), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1159), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1163), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1167), 1,
      sym_test_operator,
    ACTIONS(1169), 1,
      sym__brace_start,
    ACTIONS(3060), 1,
      anon_sym_BQUOTE,
    STATE(2484), 1,
      aux_sym__literal_repeat1,
    STATE(2968), 1,
      sym__expression,
    ACTIONS(1129), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1135), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1137), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1151), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(1165), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(2594), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2573), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [95383] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1127), 1,
      sym_word,
    ACTIONS(1131), 1,
      anon_sym_LPAREN,
    ACTIONS(1133), 1,
      anon_sym_BANG,
    ACTIONS(1139), 1,
      anon_sym_TILDE,
    ACTIONS(1141), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1145), 1,
      anon_sym_DOLLAR,
    ACTIONS(1147), 1,
      sym__special_character,
    ACTIONS(1149), 1,
      anon_sym_DQUOTE,
    ACTIONS(1153), 1,
      aux_sym_number_token1,
    ACTIONS(1155), 1,
      aux_sym_number_token2,
    ACTIONS(1157), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1159), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1163), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1167), 1,
      sym_test_operator,
    ACTIONS(1169), 1,
      sym__brace_start,
    ACTIONS(3060), 1,
      anon_sym_BQUOTE,
    STATE(2484), 1,
      aux_sym__literal_repeat1,
    STATE(2967), 1,
      sym__expression,
    ACTIONS(1129), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1135), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1137), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1151), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(1165), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(2594), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2573), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [95480] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1322), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1320), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [95531] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1318), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1316), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [95582] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5375), 6,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5373), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [95633] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1127), 1,
      sym_word,
    ACTIONS(1131), 1,
      anon_sym_LPAREN,
    ACTIONS(1133), 1,
      anon_sym_BANG,
    ACTIONS(1139), 1,
      anon_sym_TILDE,
    ACTIONS(1141), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1145), 1,
      anon_sym_DOLLAR,
    ACTIONS(1147), 1,
      sym__special_character,
    ACTIONS(1149), 1,
      anon_sym_DQUOTE,
    ACTIONS(1153), 1,
      aux_sym_number_token1,
    ACTIONS(1155), 1,
      aux_sym_number_token2,
    ACTIONS(1157), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1159), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1163), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1167), 1,
      sym_test_operator,
    ACTIONS(1169), 1,
      sym__brace_start,
    ACTIONS(3060), 1,
      anon_sym_BQUOTE,
    STATE(2484), 1,
      aux_sym__literal_repeat1,
    STATE(3096), 1,
      sym__expression,
    ACTIONS(1129), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1135), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1137), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1151), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(1165), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(2594), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2573), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [95730] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(363), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(367), 1,
      anon_sym_DOLLAR,
    ACTIONS(371), 1,
      anon_sym_DQUOTE,
    ACTIONS(375), 1,
      aux_sym_number_token1,
    ACTIONS(377), 1,
      aux_sym_number_token2,
    ACTIONS(379), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(381), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(385), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(391), 1,
      sym__brace_start,
    ACTIONS(1091), 1,
      sym_word,
    ACTIONS(1093), 1,
      anon_sym_LPAREN,
    ACTIONS(1095), 1,
      anon_sym_BANG,
    ACTIONS(1103), 1,
      anon_sym_TILDE,
    ACTIONS(1113), 1,
      sym_test_operator,
    ACTIONS(5809), 1,
      anon_sym_BQUOTE,
    ACTIONS(5916), 1,
      sym__special_character,
    STATE(2472), 1,
      aux_sym__literal_repeat1,
    STATE(2924), 1,
      sym__expression,
    ACTIONS(352), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(387), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1099), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1101), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1109), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2863), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2451), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [95827] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(363), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(367), 1,
      anon_sym_DOLLAR,
    ACTIONS(371), 1,
      anon_sym_DQUOTE,
    ACTIONS(375), 1,
      aux_sym_number_token1,
    ACTIONS(377), 1,
      aux_sym_number_token2,
    ACTIONS(379), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(381), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(385), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(391), 1,
      sym__brace_start,
    ACTIONS(1091), 1,
      sym_word,
    ACTIONS(1093), 1,
      anon_sym_LPAREN,
    ACTIONS(1095), 1,
      anon_sym_BANG,
    ACTIONS(1103), 1,
      anon_sym_TILDE,
    ACTIONS(1113), 1,
      sym_test_operator,
    ACTIONS(5809), 1,
      anon_sym_BQUOTE,
    ACTIONS(5916), 1,
      sym__special_character,
    STATE(2472), 1,
      aux_sym__literal_repeat1,
    STATE(2925), 1,
      sym__expression,
    ACTIONS(352), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(387), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1099), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1101), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1109), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2863), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2451), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [95924] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1263), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1261), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [95975] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4372), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4370), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [96026] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2096), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(2094), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [96077] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1312), 16,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_LPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1314), 27,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [96128] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1336), 16,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_LPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1338), 27,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [96179] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1344), 16,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_LPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1346), 27,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [96230] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2074), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(2072), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [96281] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4554), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4552), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [96332] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5375), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5373), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [96383] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1356), 16,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_LPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1358), 27,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [96434] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5772), 6,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5770), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [96485] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1352), 16,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_LPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1354), 27,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [96536] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(363), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(367), 1,
      anon_sym_DOLLAR,
    ACTIONS(371), 1,
      anon_sym_DQUOTE,
    ACTIONS(375), 1,
      aux_sym_number_token1,
    ACTIONS(377), 1,
      aux_sym_number_token2,
    ACTIONS(379), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(381), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(385), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(391), 1,
      sym__brace_start,
    ACTIONS(1091), 1,
      sym_word,
    ACTIONS(1093), 1,
      anon_sym_LPAREN,
    ACTIONS(1095), 1,
      anon_sym_BANG,
    ACTIONS(1103), 1,
      anon_sym_TILDE,
    ACTIONS(1113), 1,
      sym_test_operator,
    ACTIONS(5809), 1,
      anon_sym_BQUOTE,
    ACTIONS(5916), 1,
      sym__special_character,
    STATE(2472), 1,
      aux_sym__literal_repeat1,
    STATE(2813), 1,
      sym__expression,
    ACTIONS(352), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(387), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1099), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1101), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1109), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2863), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2451), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [96633] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(6123), 1,
      sym__special_character,
    STATE(1964), 1,
      aux_sym__literal_repeat1,
    ACTIONS(1364), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1362), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [96688] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1340), 16,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_LPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1342), 27,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [96739] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(6126), 1,
      sym__special_character,
    STATE(1964), 1,
      aux_sym__literal_repeat1,
    ACTIONS(5478), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5476), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [96794] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(363), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(367), 1,
      anon_sym_DOLLAR,
    ACTIONS(371), 1,
      anon_sym_DQUOTE,
    ACTIONS(375), 1,
      aux_sym_number_token1,
    ACTIONS(377), 1,
      aux_sym_number_token2,
    ACTIONS(379), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(381), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(385), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(391), 1,
      sym__brace_start,
    ACTIONS(1091), 1,
      sym_word,
    ACTIONS(1093), 1,
      anon_sym_LPAREN,
    ACTIONS(1095), 1,
      anon_sym_BANG,
    ACTIONS(1103), 1,
      anon_sym_TILDE,
    ACTIONS(1113), 1,
      sym_test_operator,
    ACTIONS(5809), 1,
      anon_sym_BQUOTE,
    ACTIONS(5916), 1,
      sym__special_character,
    STATE(2472), 1,
      aux_sym__literal_repeat1,
    STATE(2738), 1,
      sym__expression,
    ACTIONS(352), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(387), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1099), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1101), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1109), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2863), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2451), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [96891] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(6013), 1,
      aux_sym_concatenation_token1,
    ACTIONS(6015), 1,
      sym__concat,
    STATE(2068), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5391), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5389), 36,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [96948] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1131), 1,
      anon_sym_LPAREN,
    ACTIONS(1141), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1145), 1,
      anon_sym_DOLLAR,
    ACTIONS(1149), 1,
      anon_sym_DQUOTE,
    ACTIONS(1153), 1,
      aux_sym_number_token1,
    ACTIONS(1155), 1,
      aux_sym_number_token2,
    ACTIONS(1157), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1159), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1163), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1169), 1,
      sym__brace_start,
    ACTIONS(1187), 1,
      sym_word,
    ACTIONS(1189), 1,
      anon_sym_BANG,
    ACTIONS(1195), 1,
      anon_sym_TILDE,
    ACTIONS(1201), 1,
      sym_test_operator,
    ACTIONS(3060), 1,
      anon_sym_BQUOTE,
    ACTIONS(5779), 1,
      sym__special_character,
    STATE(2717), 1,
      aux_sym__literal_repeat1,
    STATE(3048), 1,
      sym__expression,
    ACTIONS(1129), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1165), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1191), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1193), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1199), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2594), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2503), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [97045] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1131), 1,
      anon_sym_LPAREN,
    ACTIONS(1141), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1145), 1,
      anon_sym_DOLLAR,
    ACTIONS(1149), 1,
      anon_sym_DQUOTE,
    ACTIONS(1153), 1,
      aux_sym_number_token1,
    ACTIONS(1155), 1,
      aux_sym_number_token2,
    ACTIONS(1157), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1159), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1163), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1169), 1,
      sym__brace_start,
    ACTIONS(1187), 1,
      sym_word,
    ACTIONS(1189), 1,
      anon_sym_BANG,
    ACTIONS(1195), 1,
      anon_sym_TILDE,
    ACTIONS(1201), 1,
      sym_test_operator,
    ACTIONS(3060), 1,
      anon_sym_BQUOTE,
    ACTIONS(5779), 1,
      sym__special_character,
    STATE(2621), 1,
      sym__expression,
    STATE(2717), 1,
      aux_sym__literal_repeat1,
    ACTIONS(1129), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1165), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1191), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1193), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1199), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2594), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2503), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [97142] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1328), 16,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_LPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1330), 27,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [97193] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(6013), 1,
      aux_sym_concatenation_token1,
    ACTIONS(6015), 1,
      sym__concat,
    STATE(2067), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5395), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5393), 36,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [97250] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1320), 16,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_LPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1322), 27,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [97301] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5932), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4261), 3,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(4247), 18,
      anon_sym_LPAREN_LPAREN,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
    ACTIONS(5930), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [97356] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(187), 1,
      anon_sym_TILDE,
    ACTIONS(350), 1,
      sym_word,
    ACTIONS(359), 1,
      anon_sym_LPAREN,
    ACTIONS(361), 1,
      anon_sym_BANG,
    ACTIONS(363), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(367), 1,
      anon_sym_DOLLAR,
    ACTIONS(371), 1,
      anon_sym_DQUOTE,
    ACTIONS(375), 1,
      aux_sym_number_token1,
    ACTIONS(377), 1,
      aux_sym_number_token2,
    ACTIONS(379), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(381), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(385), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(389), 1,
      sym_test_operator,
    ACTIONS(391), 1,
      sym__brace_start,
    ACTIONS(5807), 1,
      sym__special_character,
    ACTIONS(5809), 1,
      anon_sym_BQUOTE,
    STATE(2641), 1,
      aux_sym__literal_repeat1,
    STATE(2962), 1,
      sym__expression,
    ACTIONS(183), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(185), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(352), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(373), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(387), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3071), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2521), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [97453] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(187), 1,
      anon_sym_TILDE,
    ACTIONS(350), 1,
      sym_word,
    ACTIONS(359), 1,
      anon_sym_LPAREN,
    ACTIONS(361), 1,
      anon_sym_BANG,
    ACTIONS(363), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(367), 1,
      anon_sym_DOLLAR,
    ACTIONS(371), 1,
      anon_sym_DQUOTE,
    ACTIONS(375), 1,
      aux_sym_number_token1,
    ACTIONS(377), 1,
      aux_sym_number_token2,
    ACTIONS(379), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(381), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(385), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(389), 1,
      sym_test_operator,
    ACTIONS(391), 1,
      sym__brace_start,
    ACTIONS(5807), 1,
      sym__special_character,
    ACTIONS(5809), 1,
      anon_sym_BQUOTE,
    STATE(2641), 1,
      aux_sym__literal_repeat1,
    STATE(2989), 1,
      sym__expression,
    ACTIONS(183), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(185), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(352), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(373), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(387), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3071), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2521), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [97550] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1316), 16,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_LPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1318), 27,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [97601] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(187), 1,
      anon_sym_TILDE,
    ACTIONS(350), 1,
      sym_word,
    ACTIONS(359), 1,
      anon_sym_LPAREN,
    ACTIONS(361), 1,
      anon_sym_BANG,
    ACTIONS(363), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(367), 1,
      anon_sym_DOLLAR,
    ACTIONS(371), 1,
      anon_sym_DQUOTE,
    ACTIONS(375), 1,
      aux_sym_number_token1,
    ACTIONS(377), 1,
      aux_sym_number_token2,
    ACTIONS(379), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(381), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(385), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(389), 1,
      sym_test_operator,
    ACTIONS(391), 1,
      sym__brace_start,
    ACTIONS(5807), 1,
      sym__special_character,
    ACTIONS(5809), 1,
      anon_sym_BQUOTE,
    STATE(2641), 1,
      aux_sym__literal_repeat1,
    STATE(2988), 1,
      sym__expression,
    ACTIONS(183), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(185), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(352), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(373), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(387), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3071), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2521), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [97698] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1294), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1292), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [97749] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(187), 1,
      anon_sym_TILDE,
    ACTIONS(350), 1,
      sym_word,
    ACTIONS(359), 1,
      anon_sym_LPAREN,
    ACTIONS(361), 1,
      anon_sym_BANG,
    ACTIONS(363), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(367), 1,
      anon_sym_DOLLAR,
    ACTIONS(371), 1,
      anon_sym_DQUOTE,
    ACTIONS(375), 1,
      aux_sym_number_token1,
    ACTIONS(377), 1,
      aux_sym_number_token2,
    ACTIONS(379), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(381), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(385), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(389), 1,
      sym_test_operator,
    ACTIONS(391), 1,
      sym__brace_start,
    ACTIONS(5807), 1,
      sym__special_character,
    ACTIONS(5809), 1,
      anon_sym_BQUOTE,
    STATE(2641), 1,
      aux_sym__literal_repeat1,
    STATE(2986), 1,
      sym__expression,
    ACTIONS(183), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(185), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(352), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(373), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(387), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3071), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2521), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [97846] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(187), 1,
      anon_sym_TILDE,
    ACTIONS(350), 1,
      sym_word,
    ACTIONS(359), 1,
      anon_sym_LPAREN,
    ACTIONS(361), 1,
      anon_sym_BANG,
    ACTIONS(363), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(367), 1,
      anon_sym_DOLLAR,
    ACTIONS(371), 1,
      anon_sym_DQUOTE,
    ACTIONS(375), 1,
      aux_sym_number_token1,
    ACTIONS(377), 1,
      aux_sym_number_token2,
    ACTIONS(379), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(381), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(385), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(389), 1,
      sym_test_operator,
    ACTIONS(391), 1,
      sym__brace_start,
    ACTIONS(5807), 1,
      sym__special_character,
    ACTIONS(5809), 1,
      anon_sym_BQUOTE,
    STATE(2641), 1,
      aux_sym__literal_repeat1,
    STATE(2985), 1,
      sym__expression,
    ACTIONS(183), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(185), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(352), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(373), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(387), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3071), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2521), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [97943] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1127), 1,
      sym_word,
    ACTIONS(1131), 1,
      anon_sym_LPAREN,
    ACTIONS(1133), 1,
      anon_sym_BANG,
    ACTIONS(1139), 1,
      anon_sym_TILDE,
    ACTIONS(1141), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1145), 1,
      anon_sym_DOLLAR,
    ACTIONS(1147), 1,
      sym__special_character,
    ACTIONS(1149), 1,
      anon_sym_DQUOTE,
    ACTIONS(1153), 1,
      aux_sym_number_token1,
    ACTIONS(1155), 1,
      aux_sym_number_token2,
    ACTIONS(1157), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1159), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1163), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1167), 1,
      sym_test_operator,
    ACTIONS(1169), 1,
      sym__brace_start,
    ACTIONS(3060), 1,
      anon_sym_BQUOTE,
    STATE(2484), 1,
      aux_sym__literal_repeat1,
    STATE(2955), 1,
      sym__expression,
    ACTIONS(1129), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1135), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1137), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1151), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(1165), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(2594), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2573), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [98040] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(187), 1,
      anon_sym_TILDE,
    ACTIONS(350), 1,
      sym_word,
    ACTIONS(359), 1,
      anon_sym_LPAREN,
    ACTIONS(361), 1,
      anon_sym_BANG,
    ACTIONS(363), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(367), 1,
      anon_sym_DOLLAR,
    ACTIONS(371), 1,
      anon_sym_DQUOTE,
    ACTIONS(375), 1,
      aux_sym_number_token1,
    ACTIONS(377), 1,
      aux_sym_number_token2,
    ACTIONS(379), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(381), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(385), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(389), 1,
      sym_test_operator,
    ACTIONS(391), 1,
      sym__brace_start,
    ACTIONS(5807), 1,
      sym__special_character,
    ACTIONS(5809), 1,
      anon_sym_BQUOTE,
    STATE(2641), 1,
      aux_sym__literal_repeat1,
    STATE(2981), 1,
      sym__expression,
    ACTIONS(183), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(185), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(352), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(373), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(387), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3071), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2521), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [98137] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(187), 1,
      anon_sym_TILDE,
    ACTIONS(350), 1,
      sym_word,
    ACTIONS(359), 1,
      anon_sym_LPAREN,
    ACTIONS(361), 1,
      anon_sym_BANG,
    ACTIONS(363), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(367), 1,
      anon_sym_DOLLAR,
    ACTIONS(371), 1,
      anon_sym_DQUOTE,
    ACTIONS(375), 1,
      aux_sym_number_token1,
    ACTIONS(377), 1,
      aux_sym_number_token2,
    ACTIONS(379), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(381), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(385), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(389), 1,
      sym_test_operator,
    ACTIONS(391), 1,
      sym__brace_start,
    ACTIONS(5807), 1,
      sym__special_character,
    ACTIONS(5809), 1,
      anon_sym_BQUOTE,
    STATE(2641), 1,
      aux_sym__literal_repeat1,
    STATE(2966), 1,
      sym__expression,
    ACTIONS(183), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(185), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(352), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(373), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(387), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3071), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2521), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [98234] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(187), 1,
      anon_sym_TILDE,
    ACTIONS(350), 1,
      sym_word,
    ACTIONS(359), 1,
      anon_sym_LPAREN,
    ACTIONS(361), 1,
      anon_sym_BANG,
    ACTIONS(363), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(367), 1,
      anon_sym_DOLLAR,
    ACTIONS(371), 1,
      anon_sym_DQUOTE,
    ACTIONS(375), 1,
      aux_sym_number_token1,
    ACTIONS(377), 1,
      aux_sym_number_token2,
    ACTIONS(379), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(381), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(385), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(389), 1,
      sym_test_operator,
    ACTIONS(391), 1,
      sym__brace_start,
    ACTIONS(5807), 1,
      sym__special_character,
    ACTIONS(5809), 1,
      anon_sym_BQUOTE,
    STATE(2641), 1,
      aux_sym__literal_repeat1,
    STATE(2964), 1,
      sym__expression,
    ACTIONS(183), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(185), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(352), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(373), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(387), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3071), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2521), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [98331] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1334), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1332), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [98382] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(187), 1,
      anon_sym_TILDE,
    ACTIONS(350), 1,
      sym_word,
    ACTIONS(359), 1,
      anon_sym_LPAREN,
    ACTIONS(361), 1,
      anon_sym_BANG,
    ACTIONS(363), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(367), 1,
      anon_sym_DOLLAR,
    ACTIONS(371), 1,
      anon_sym_DQUOTE,
    ACTIONS(375), 1,
      aux_sym_number_token1,
    ACTIONS(377), 1,
      aux_sym_number_token2,
    ACTIONS(379), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(381), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(385), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(389), 1,
      sym_test_operator,
    ACTIONS(391), 1,
      sym__brace_start,
    ACTIONS(5807), 1,
      sym__special_character,
    ACTIONS(5809), 1,
      anon_sym_BQUOTE,
    STATE(2641), 1,
      aux_sym__literal_repeat1,
    STATE(2961), 1,
      sym__expression,
    ACTIONS(183), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(185), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(352), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(373), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(387), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3071), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2521), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [98479] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(187), 1,
      anon_sym_TILDE,
    ACTIONS(350), 1,
      sym_word,
    ACTIONS(359), 1,
      anon_sym_LPAREN,
    ACTIONS(361), 1,
      anon_sym_BANG,
    ACTIONS(363), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(367), 1,
      anon_sym_DOLLAR,
    ACTIONS(371), 1,
      anon_sym_DQUOTE,
    ACTIONS(375), 1,
      aux_sym_number_token1,
    ACTIONS(377), 1,
      aux_sym_number_token2,
    ACTIONS(379), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(381), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(385), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(389), 1,
      sym_test_operator,
    ACTIONS(391), 1,
      sym__brace_start,
    ACTIONS(5807), 1,
      sym__special_character,
    ACTIONS(5809), 1,
      anon_sym_BQUOTE,
    STATE(2641), 1,
      aux_sym__literal_repeat1,
    STATE(2960), 1,
      sym__expression,
    ACTIONS(183), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(185), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(352), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(373), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(387), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3071), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2521), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [98576] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(187), 1,
      anon_sym_TILDE,
    ACTIONS(350), 1,
      sym_word,
    ACTIONS(359), 1,
      anon_sym_LPAREN,
    ACTIONS(361), 1,
      anon_sym_BANG,
    ACTIONS(363), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(367), 1,
      anon_sym_DOLLAR,
    ACTIONS(371), 1,
      anon_sym_DQUOTE,
    ACTIONS(375), 1,
      aux_sym_number_token1,
    ACTIONS(377), 1,
      aux_sym_number_token2,
    ACTIONS(379), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(381), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(385), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(389), 1,
      sym_test_operator,
    ACTIONS(391), 1,
      sym__brace_start,
    ACTIONS(5807), 1,
      sym__special_character,
    ACTIONS(5809), 1,
      anon_sym_BQUOTE,
    STATE(2641), 1,
      aux_sym__literal_repeat1,
    STATE(2959), 1,
      sym__expression,
    ACTIONS(183), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(185), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(352), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(373), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(387), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3071), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2521), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [98673] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(187), 1,
      anon_sym_TILDE,
    ACTIONS(350), 1,
      sym_word,
    ACTIONS(359), 1,
      anon_sym_LPAREN,
    ACTIONS(361), 1,
      anon_sym_BANG,
    ACTIONS(363), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(367), 1,
      anon_sym_DOLLAR,
    ACTIONS(371), 1,
      anon_sym_DQUOTE,
    ACTIONS(375), 1,
      aux_sym_number_token1,
    ACTIONS(377), 1,
      aux_sym_number_token2,
    ACTIONS(379), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(381), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(385), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(389), 1,
      sym_test_operator,
    ACTIONS(391), 1,
      sym__brace_start,
    ACTIONS(5807), 1,
      sym__special_character,
    ACTIONS(5809), 1,
      anon_sym_BQUOTE,
    STATE(2641), 1,
      aux_sym__literal_repeat1,
    STATE(2958), 1,
      sym__expression,
    ACTIONS(183), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(185), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(352), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(373), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(387), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3071), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2521), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [98770] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(187), 1,
      anon_sym_TILDE,
    ACTIONS(350), 1,
      sym_word,
    ACTIONS(359), 1,
      anon_sym_LPAREN,
    ACTIONS(361), 1,
      anon_sym_BANG,
    ACTIONS(363), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(367), 1,
      anon_sym_DOLLAR,
    ACTIONS(371), 1,
      anon_sym_DQUOTE,
    ACTIONS(375), 1,
      aux_sym_number_token1,
    ACTIONS(377), 1,
      aux_sym_number_token2,
    ACTIONS(379), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(381), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(385), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(389), 1,
      sym_test_operator,
    ACTIONS(391), 1,
      sym__brace_start,
    ACTIONS(5807), 1,
      sym__special_character,
    ACTIONS(5809), 1,
      anon_sym_BQUOTE,
    STATE(2641), 1,
      aux_sym__literal_repeat1,
    STATE(2957), 1,
      sym__expression,
    ACTIONS(183), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(185), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(352), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(373), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(387), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3071), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2521), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [98867] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1308), 16,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_LPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1310), 27,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [98918] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1302), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1300), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [98969] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1263), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1261), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [99020] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1292), 16,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_LPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1294), 27,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [99071] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1310), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1308), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [99122] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1298), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1296), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [99173] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1131), 1,
      anon_sym_LPAREN,
    ACTIONS(1141), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1145), 1,
      anon_sym_DOLLAR,
    ACTIONS(1149), 1,
      anon_sym_DQUOTE,
    ACTIONS(1153), 1,
      aux_sym_number_token1,
    ACTIONS(1155), 1,
      aux_sym_number_token2,
    ACTIONS(1157), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1159), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1163), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1169), 1,
      sym__brace_start,
    ACTIONS(1211), 1,
      sym_word,
    ACTIONS(1213), 1,
      anon_sym_BANG,
    ACTIONS(1219), 1,
      anon_sym_TILDE,
    ACTIONS(1221), 1,
      sym__special_character,
    ACTIONS(1225), 1,
      sym_test_operator,
    ACTIONS(3060), 1,
      anon_sym_BQUOTE,
    STATE(2484), 1,
      aux_sym__literal_repeat1,
    STATE(3305), 1,
      sym__expression,
    ACTIONS(1129), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1165), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1215), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1217), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1223), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2594), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2588), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [99270] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1131), 1,
      anon_sym_LPAREN,
    ACTIONS(1141), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1145), 1,
      anon_sym_DOLLAR,
    ACTIONS(1149), 1,
      anon_sym_DQUOTE,
    ACTIONS(1153), 1,
      aux_sym_number_token1,
    ACTIONS(1155), 1,
      aux_sym_number_token2,
    ACTIONS(1157), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1159), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1163), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1169), 1,
      sym__brace_start,
    ACTIONS(1211), 1,
      sym_word,
    ACTIONS(1213), 1,
      anon_sym_BANG,
    ACTIONS(1219), 1,
      anon_sym_TILDE,
    ACTIONS(1221), 1,
      sym__special_character,
    ACTIONS(1225), 1,
      sym_test_operator,
    ACTIONS(3060), 1,
      anon_sym_BQUOTE,
    STATE(2484), 1,
      aux_sym__literal_repeat1,
    STATE(3314), 1,
      sym__expression,
    ACTIONS(1129), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1165), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1215), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1217), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1223), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2594), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2588), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [99367] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1131), 1,
      anon_sym_LPAREN,
    ACTIONS(1141), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1145), 1,
      anon_sym_DOLLAR,
    ACTIONS(1149), 1,
      anon_sym_DQUOTE,
    ACTIONS(1153), 1,
      aux_sym_number_token1,
    ACTIONS(1155), 1,
      aux_sym_number_token2,
    ACTIONS(1157), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1159), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1163), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1169), 1,
      sym__brace_start,
    ACTIONS(1211), 1,
      sym_word,
    ACTIONS(1213), 1,
      anon_sym_BANG,
    ACTIONS(1219), 1,
      anon_sym_TILDE,
    ACTIONS(1221), 1,
      sym__special_character,
    ACTIONS(1225), 1,
      sym_test_operator,
    ACTIONS(3060), 1,
      anon_sym_BQUOTE,
    STATE(2484), 1,
      aux_sym__literal_repeat1,
    STATE(3319), 1,
      sym__expression,
    ACTIONS(1129), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1165), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1215), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1217), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1223), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2594), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2588), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [99464] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(187), 1,
      anon_sym_TILDE,
    ACTIONS(350), 1,
      sym_word,
    ACTIONS(359), 1,
      anon_sym_LPAREN,
    ACTIONS(361), 1,
      anon_sym_BANG,
    ACTIONS(363), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(367), 1,
      anon_sym_DOLLAR,
    ACTIONS(371), 1,
      anon_sym_DQUOTE,
    ACTIONS(375), 1,
      aux_sym_number_token1,
    ACTIONS(377), 1,
      aux_sym_number_token2,
    ACTIONS(379), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(381), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(385), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(389), 1,
      sym_test_operator,
    ACTIONS(391), 1,
      sym__brace_start,
    ACTIONS(5807), 1,
      sym__special_character,
    ACTIONS(5809), 1,
      anon_sym_BQUOTE,
    STATE(2641), 1,
      aux_sym__literal_repeat1,
    STATE(3011), 1,
      sym__expression,
    ACTIONS(183), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(185), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(352), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(373), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(387), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3071), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2521), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [99561] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1332), 16,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_LPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1334), 27,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [99612] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1127), 1,
      sym_word,
    ACTIONS(1131), 1,
      anon_sym_LPAREN,
    ACTIONS(1133), 1,
      anon_sym_BANG,
    ACTIONS(1139), 1,
      anon_sym_TILDE,
    ACTIONS(1141), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1145), 1,
      anon_sym_DOLLAR,
    ACTIONS(1147), 1,
      sym__special_character,
    ACTIONS(1149), 1,
      anon_sym_DQUOTE,
    ACTIONS(1153), 1,
      aux_sym_number_token1,
    ACTIONS(1155), 1,
      aux_sym_number_token2,
    ACTIONS(1157), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1159), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1163), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1167), 1,
      sym_test_operator,
    ACTIONS(1169), 1,
      sym__brace_start,
    ACTIONS(3060), 1,
      anon_sym_BQUOTE,
    STATE(2484), 1,
      aux_sym__literal_repeat1,
    STATE(3054), 1,
      sym__expression,
    ACTIONS(1129), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1135), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1137), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1151), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(1165), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(2594), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2573), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [99709] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1253), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1251), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [99760] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1326), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1324), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [99811] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1306), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1304), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [99862] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(363), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(367), 1,
      anon_sym_DOLLAR,
    ACTIONS(371), 1,
      anon_sym_DQUOTE,
    ACTIONS(375), 1,
      aux_sym_number_token1,
    ACTIONS(377), 1,
      aux_sym_number_token2,
    ACTIONS(379), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(381), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(385), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(391), 1,
      sym__brace_start,
    ACTIONS(1091), 1,
      sym_word,
    ACTIONS(1093), 1,
      anon_sym_LPAREN,
    ACTIONS(1095), 1,
      anon_sym_BANG,
    ACTIONS(1103), 1,
      anon_sym_TILDE,
    ACTIONS(1113), 1,
      sym_test_operator,
    ACTIONS(5809), 1,
      anon_sym_BQUOTE,
    ACTIONS(5916), 1,
      sym__special_character,
    STATE(2472), 1,
      aux_sym__literal_repeat1,
    STATE(2837), 1,
      sym__expression,
    ACTIONS(352), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(387), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1099), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1101), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1109), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2863), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2451), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [99959] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1131), 1,
      anon_sym_LPAREN,
    ACTIONS(1141), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1145), 1,
      anon_sym_DOLLAR,
    ACTIONS(1149), 1,
      anon_sym_DQUOTE,
    ACTIONS(1153), 1,
      aux_sym_number_token1,
    ACTIONS(1155), 1,
      aux_sym_number_token2,
    ACTIONS(1157), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1159), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1163), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1169), 1,
      sym__brace_start,
    ACTIONS(1211), 1,
      sym_word,
    ACTIONS(1213), 1,
      anon_sym_BANG,
    ACTIONS(1219), 1,
      anon_sym_TILDE,
    ACTIONS(1221), 1,
      sym__special_character,
    ACTIONS(1225), 1,
      sym_test_operator,
    ACTIONS(3060), 1,
      anon_sym_BQUOTE,
    STATE(2484), 1,
      aux_sym__literal_repeat1,
    STATE(3323), 1,
      sym__expression,
    ACTIONS(1129), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1165), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1215), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1217), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1223), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2594), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2588), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [100056] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1131), 1,
      anon_sym_LPAREN,
    ACTIONS(1141), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1145), 1,
      anon_sym_DOLLAR,
    ACTIONS(1149), 1,
      anon_sym_DQUOTE,
    ACTIONS(1153), 1,
      aux_sym_number_token1,
    ACTIONS(1155), 1,
      aux_sym_number_token2,
    ACTIONS(1157), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1159), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1163), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1169), 1,
      sym__brace_start,
    ACTIONS(1211), 1,
      sym_word,
    ACTIONS(1213), 1,
      anon_sym_BANG,
    ACTIONS(1219), 1,
      anon_sym_TILDE,
    ACTIONS(1221), 1,
      sym__special_character,
    ACTIONS(1225), 1,
      sym_test_operator,
    ACTIONS(3060), 1,
      anon_sym_BQUOTE,
    STATE(2484), 1,
      aux_sym__literal_repeat1,
    STATE(3327), 1,
      sym__expression,
    ACTIONS(1129), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1165), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1215), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1217), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1223), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2594), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2588), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [100153] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1131), 1,
      anon_sym_LPAREN,
    ACTIONS(1141), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1145), 1,
      anon_sym_DOLLAR,
    ACTIONS(1149), 1,
      anon_sym_DQUOTE,
    ACTIONS(1153), 1,
      aux_sym_number_token1,
    ACTIONS(1155), 1,
      aux_sym_number_token2,
    ACTIONS(1157), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1159), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1163), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1169), 1,
      sym__brace_start,
    ACTIONS(1211), 1,
      sym_word,
    ACTIONS(1213), 1,
      anon_sym_BANG,
    ACTIONS(1219), 1,
      anon_sym_TILDE,
    ACTIONS(1221), 1,
      sym__special_character,
    ACTIONS(1225), 1,
      sym_test_operator,
    ACTIONS(3060), 1,
      anon_sym_BQUOTE,
    STATE(2484), 1,
      aux_sym__literal_repeat1,
    STATE(3330), 1,
      sym__expression,
    ACTIONS(1129), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1165), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1215), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1217), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1223), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2594), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2588), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [100250] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1131), 1,
      anon_sym_LPAREN,
    ACTIONS(1141), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1145), 1,
      anon_sym_DOLLAR,
    ACTIONS(1149), 1,
      anon_sym_DQUOTE,
    ACTIONS(1153), 1,
      aux_sym_number_token1,
    ACTIONS(1155), 1,
      aux_sym_number_token2,
    ACTIONS(1157), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1159), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1163), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1169), 1,
      sym__brace_start,
    ACTIONS(1211), 1,
      sym_word,
    ACTIONS(1213), 1,
      anon_sym_BANG,
    ACTIONS(1219), 1,
      anon_sym_TILDE,
    ACTIONS(1221), 1,
      sym__special_character,
    ACTIONS(1225), 1,
      sym_test_operator,
    ACTIONS(3060), 1,
      anon_sym_BQUOTE,
    STATE(2484), 1,
      aux_sym__literal_repeat1,
    STATE(3332), 1,
      sym__expression,
    ACTIONS(1129), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1165), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1215), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1217), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1223), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2594), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2588), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [100347] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1310), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1308), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [100398] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1131), 1,
      anon_sym_LPAREN,
    ACTIONS(1141), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1145), 1,
      anon_sym_DOLLAR,
    ACTIONS(1149), 1,
      anon_sym_DQUOTE,
    ACTIONS(1153), 1,
      aux_sym_number_token1,
    ACTIONS(1155), 1,
      aux_sym_number_token2,
    ACTIONS(1157), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1159), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1163), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1169), 1,
      sym__brace_start,
    ACTIONS(1211), 1,
      sym_word,
    ACTIONS(1213), 1,
      anon_sym_BANG,
    ACTIONS(1219), 1,
      anon_sym_TILDE,
    ACTIONS(1221), 1,
      sym__special_character,
    ACTIONS(1225), 1,
      sym_test_operator,
    ACTIONS(3060), 1,
      anon_sym_BQUOTE,
    STATE(2484), 1,
      aux_sym__literal_repeat1,
    STATE(3335), 1,
      sym__expression,
    ACTIONS(1129), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1165), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1215), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1217), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1223), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2594), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2588), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [100495] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(363), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(367), 1,
      anon_sym_DOLLAR,
    ACTIONS(371), 1,
      anon_sym_DQUOTE,
    ACTIONS(375), 1,
      aux_sym_number_token1,
    ACTIONS(377), 1,
      aux_sym_number_token2,
    ACTIONS(379), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(381), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(385), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(391), 1,
      sym__brace_start,
    ACTIONS(1091), 1,
      sym_word,
    ACTIONS(1093), 1,
      anon_sym_LPAREN,
    ACTIONS(1095), 1,
      anon_sym_BANG,
    ACTIONS(1103), 1,
      anon_sym_TILDE,
    ACTIONS(1113), 1,
      sym_test_operator,
    ACTIONS(5809), 1,
      anon_sym_BQUOTE,
    ACTIONS(5916), 1,
      sym__special_character,
    STATE(2472), 1,
      aux_sym__literal_repeat1,
    STATE(2835), 1,
      sym__expression,
    ACTIONS(352), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(387), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1099), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1101), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1109), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2863), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2451), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [100592] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5768), 6,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5766), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [100643] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1131), 1,
      anon_sym_LPAREN,
    ACTIONS(1141), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1145), 1,
      anon_sym_DOLLAR,
    ACTIONS(1149), 1,
      anon_sym_DQUOTE,
    ACTIONS(1153), 1,
      aux_sym_number_token1,
    ACTIONS(1155), 1,
      aux_sym_number_token2,
    ACTIONS(1157), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1159), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1163), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1169), 1,
      sym__brace_start,
    ACTIONS(1211), 1,
      sym_word,
    ACTIONS(1213), 1,
      anon_sym_BANG,
    ACTIONS(1219), 1,
      anon_sym_TILDE,
    ACTIONS(1221), 1,
      sym__special_character,
    ACTIONS(1225), 1,
      sym_test_operator,
    ACTIONS(3060), 1,
      anon_sym_BQUOTE,
    STATE(2484), 1,
      aux_sym__literal_repeat1,
    STATE(3345), 1,
      sym__expression,
    ACTIONS(1129), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1165), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1215), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1217), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1223), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2594), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2588), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [100740] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5826), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5828), 1,
      sym_file_descriptor,
    ACTIONS(5831), 3,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(5821), 10,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
    ACTIONS(5823), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
    ACTIONS(5819), 17,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [100799] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1131), 1,
      anon_sym_LPAREN,
    ACTIONS(1141), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1145), 1,
      anon_sym_DOLLAR,
    ACTIONS(1149), 1,
      anon_sym_DQUOTE,
    ACTIONS(1153), 1,
      aux_sym_number_token1,
    ACTIONS(1155), 1,
      aux_sym_number_token2,
    ACTIONS(1157), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1159), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1163), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1169), 1,
      sym__brace_start,
    ACTIONS(1211), 1,
      sym_word,
    ACTIONS(1213), 1,
      anon_sym_BANG,
    ACTIONS(1219), 1,
      anon_sym_TILDE,
    ACTIONS(1221), 1,
      sym__special_character,
    ACTIONS(1225), 1,
      sym_test_operator,
    ACTIONS(3060), 1,
      anon_sym_BQUOTE,
    STATE(2484), 1,
      aux_sym__literal_repeat1,
    STATE(3347), 1,
      sym__expression,
    ACTIONS(1129), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1165), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1215), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1217), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1223), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2594), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2588), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [100896] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5826), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5828), 1,
      sym_file_descriptor,
    ACTIONS(5831), 3,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(5821), 10,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
    ACTIONS(5823), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
    ACTIONS(5819), 17,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [100955] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5772), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5770), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [101006] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(363), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(367), 1,
      anon_sym_DOLLAR,
    ACTIONS(371), 1,
      anon_sym_DQUOTE,
    ACTIONS(375), 1,
      aux_sym_number_token1,
    ACTIONS(377), 1,
      aux_sym_number_token2,
    ACTIONS(379), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(381), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(385), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(391), 1,
      sym__brace_start,
    ACTIONS(1091), 1,
      sym_word,
    ACTIONS(1093), 1,
      anon_sym_LPAREN,
    ACTIONS(1095), 1,
      anon_sym_BANG,
    ACTIONS(1103), 1,
      anon_sym_TILDE,
    ACTIONS(1113), 1,
      sym_test_operator,
    ACTIONS(5809), 1,
      anon_sym_BQUOTE,
    ACTIONS(5916), 1,
      sym__special_character,
    STATE(2472), 1,
      aux_sym__literal_repeat1,
    STATE(2831), 1,
      sym__expression,
    ACTIONS(352), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(387), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1099), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1101), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1109), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2863), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2451), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [101103] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(363), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(367), 1,
      anon_sym_DOLLAR,
    ACTIONS(371), 1,
      anon_sym_DQUOTE,
    ACTIONS(375), 1,
      aux_sym_number_token1,
    ACTIONS(377), 1,
      aux_sym_number_token2,
    ACTIONS(379), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(381), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(385), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(391), 1,
      sym__brace_start,
    ACTIONS(1091), 1,
      sym_word,
    ACTIONS(1093), 1,
      anon_sym_LPAREN,
    ACTIONS(1095), 1,
      anon_sym_BANG,
    ACTIONS(1103), 1,
      anon_sym_TILDE,
    ACTIONS(1113), 1,
      sym_test_operator,
    ACTIONS(5809), 1,
      anon_sym_BQUOTE,
    ACTIONS(5916), 1,
      sym__special_character,
    STATE(2472), 1,
      aux_sym__literal_repeat1,
    STATE(2802), 1,
      sym__expression,
    ACTIONS(352), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(387), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1099), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1101), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1109), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2863), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2451), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [101200] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1127), 1,
      sym_word,
    ACTIONS(1131), 1,
      anon_sym_LPAREN,
    ACTIONS(1133), 1,
      anon_sym_BANG,
    ACTIONS(1139), 1,
      anon_sym_TILDE,
    ACTIONS(1141), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1145), 1,
      anon_sym_DOLLAR,
    ACTIONS(1147), 1,
      sym__special_character,
    ACTIONS(1149), 1,
      anon_sym_DQUOTE,
    ACTIONS(1153), 1,
      aux_sym_number_token1,
    ACTIONS(1155), 1,
      aux_sym_number_token2,
    ACTIONS(1157), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1159), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1163), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1167), 1,
      sym_test_operator,
    ACTIONS(1169), 1,
      sym__brace_start,
    ACTIONS(3060), 1,
      anon_sym_BQUOTE,
    STATE(2484), 1,
      aux_sym__literal_repeat1,
    STATE(3022), 1,
      sym__expression,
    ACTIONS(1129), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1135), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1137), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1151), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(1165), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(2594), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2573), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [101297] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(6128), 1,
      sym__special_character,
    STATE(2026), 1,
      aux_sym__literal_repeat1,
    ACTIONS(5395), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5393), 36,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [101352] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1350), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1348), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [101403] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(6130), 1,
      sym__special_character,
    STATE(2026), 1,
      aux_sym__literal_repeat1,
    ACTIONS(1364), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1362), 36,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [101458] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(363), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(367), 1,
      anon_sym_DOLLAR,
    ACTIONS(371), 1,
      anon_sym_DQUOTE,
    ACTIONS(375), 1,
      aux_sym_number_token1,
    ACTIONS(377), 1,
      aux_sym_number_token2,
    ACTIONS(379), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(381), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(385), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(391), 1,
      sym__brace_start,
    ACTIONS(1091), 1,
      sym_word,
    ACTIONS(1093), 1,
      anon_sym_LPAREN,
    ACTIONS(1095), 1,
      anon_sym_BANG,
    ACTIONS(1103), 1,
      anon_sym_TILDE,
    ACTIONS(1113), 1,
      sym_test_operator,
    ACTIONS(5809), 1,
      anon_sym_BQUOTE,
    ACTIONS(5916), 1,
      sym__special_character,
    STATE(2472), 1,
      aux_sym__literal_repeat1,
    STATE(2830), 1,
      sym__expression,
    ACTIONS(352), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(387), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1099), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1101), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1109), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2863), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2451), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [101555] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1131), 1,
      anon_sym_LPAREN,
    ACTIONS(1141), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1145), 1,
      anon_sym_DOLLAR,
    ACTIONS(1149), 1,
      anon_sym_DQUOTE,
    ACTIONS(1153), 1,
      aux_sym_number_token1,
    ACTIONS(1155), 1,
      aux_sym_number_token2,
    ACTIONS(1157), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1159), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1163), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1169), 1,
      sym__brace_start,
    ACTIONS(1171), 1,
      sym_word,
    ACTIONS(1173), 1,
      anon_sym_BANG,
    ACTIONS(1179), 1,
      anon_sym_TILDE,
    ACTIONS(1181), 1,
      sym__special_character,
    ACTIONS(1185), 1,
      sym_test_operator,
    ACTIONS(3060), 1,
      anon_sym_BQUOTE,
    STATE(2484), 1,
      aux_sym__literal_repeat1,
    STATE(3023), 1,
      sym__expression,
    ACTIONS(1129), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1165), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1175), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1177), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1183), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2594), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2456), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [101652] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1131), 1,
      anon_sym_LPAREN,
    ACTIONS(1141), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1145), 1,
      anon_sym_DOLLAR,
    ACTIONS(1149), 1,
      anon_sym_DQUOTE,
    ACTIONS(1153), 1,
      aux_sym_number_token1,
    ACTIONS(1155), 1,
      aux_sym_number_token2,
    ACTIONS(1157), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1159), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1163), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1169), 1,
      sym__brace_start,
    ACTIONS(1171), 1,
      sym_word,
    ACTIONS(1173), 1,
      anon_sym_BANG,
    ACTIONS(1179), 1,
      anon_sym_TILDE,
    ACTIONS(1181), 1,
      sym__special_character,
    ACTIONS(1185), 1,
      sym_test_operator,
    ACTIONS(3060), 1,
      anon_sym_BQUOTE,
    STATE(2484), 1,
      aux_sym__literal_repeat1,
    STATE(3030), 1,
      sym__expression,
    ACTIONS(1129), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1165), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1175), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1177), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1183), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2594), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2456), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [101749] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(363), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(367), 1,
      anon_sym_DOLLAR,
    ACTIONS(371), 1,
      anon_sym_DQUOTE,
    ACTIONS(375), 1,
      aux_sym_number_token1,
    ACTIONS(377), 1,
      aux_sym_number_token2,
    ACTIONS(379), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(381), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(385), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(391), 1,
      sym__brace_start,
    ACTIONS(1091), 1,
      sym_word,
    ACTIONS(1093), 1,
      anon_sym_LPAREN,
    ACTIONS(1095), 1,
      anon_sym_BANG,
    ACTIONS(1103), 1,
      anon_sym_TILDE,
    ACTIONS(1113), 1,
      sym_test_operator,
    ACTIONS(5809), 1,
      anon_sym_BQUOTE,
    ACTIONS(5916), 1,
      sym__special_character,
    STATE(2472), 1,
      aux_sym__literal_repeat1,
    STATE(2829), 1,
      sym__expression,
    ACTIONS(352), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(387), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1099), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1101), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1109), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2863), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2451), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [101846] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(363), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(367), 1,
      anon_sym_DOLLAR,
    ACTIONS(371), 1,
      anon_sym_DQUOTE,
    ACTIONS(375), 1,
      aux_sym_number_token1,
    ACTIONS(377), 1,
      aux_sym_number_token2,
    ACTIONS(379), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(381), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(385), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(391), 1,
      sym__brace_start,
    ACTIONS(1091), 1,
      sym_word,
    ACTIONS(1093), 1,
      anon_sym_LPAREN,
    ACTIONS(1095), 1,
      anon_sym_BANG,
    ACTIONS(1103), 1,
      anon_sym_TILDE,
    ACTIONS(1113), 1,
      sym_test_operator,
    ACTIONS(5809), 1,
      anon_sym_BQUOTE,
    ACTIONS(5916), 1,
      sym__special_character,
    STATE(2472), 1,
      aux_sym__literal_repeat1,
    STATE(2828), 1,
      sym__expression,
    ACTIONS(352), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(387), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1099), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1101), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1109), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2863), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2451), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [101943] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1342), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1340), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [101994] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(2152), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(6133), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(1261), 15,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1263), 25,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [102049] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(363), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(367), 1,
      anon_sym_DOLLAR,
    ACTIONS(371), 1,
      anon_sym_DQUOTE,
    ACTIONS(375), 1,
      aux_sym_number_token1,
    ACTIONS(377), 1,
      aux_sym_number_token2,
    ACTIONS(379), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(381), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(385), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(391), 1,
      sym__brace_start,
    ACTIONS(1091), 1,
      sym_word,
    ACTIONS(1093), 1,
      anon_sym_LPAREN,
    ACTIONS(1095), 1,
      anon_sym_BANG,
    ACTIONS(1103), 1,
      anon_sym_TILDE,
    ACTIONS(1113), 1,
      sym_test_operator,
    ACTIONS(5809), 1,
      anon_sym_BQUOTE,
    ACTIONS(5916), 1,
      sym__special_character,
    STATE(2472), 1,
      aux_sym__literal_repeat1,
    STATE(2816), 1,
      sym__expression,
    ACTIONS(352), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(387), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1099), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1101), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1109), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2863), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2451), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [102146] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1127), 1,
      sym_word,
    ACTIONS(1131), 1,
      anon_sym_LPAREN,
    ACTIONS(1133), 1,
      anon_sym_BANG,
    ACTIONS(1139), 1,
      anon_sym_TILDE,
    ACTIONS(1141), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1145), 1,
      anon_sym_DOLLAR,
    ACTIONS(1147), 1,
      sym__special_character,
    ACTIONS(1149), 1,
      anon_sym_DQUOTE,
    ACTIONS(1153), 1,
      aux_sym_number_token1,
    ACTIONS(1155), 1,
      aux_sym_number_token2,
    ACTIONS(1157), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1159), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1163), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1167), 1,
      sym_test_operator,
    ACTIONS(1169), 1,
      sym__brace_start,
    ACTIONS(3060), 1,
      anon_sym_BQUOTE,
    STATE(2484), 1,
      aux_sym__literal_repeat1,
    STATE(2942), 1,
      sym__expression,
    ACTIONS(1129), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1135), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1137), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1151), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(1165), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(2594), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2573), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [102243] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(107), 1,
      anon_sym_TILDE,
    ACTIONS(237), 1,
      sym_word,
    ACTIONS(248), 1,
      anon_sym_LPAREN,
    ACTIONS(250), 1,
      anon_sym_BANG,
    ACTIONS(258), 1,
      anon_sym_DOLLAR,
    ACTIONS(264), 1,
      aux_sym_number_token1,
    ACTIONS(266), 1,
      aux_sym_number_token2,
    ACTIONS(270), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(282), 1,
      sym_test_operator,
    ACTIONS(284), 1,
      sym__brace_start,
    ACTIONS(1075), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1079), 1,
      sym__special_character,
    ACTIONS(1081), 1,
      anon_sym_DQUOTE,
    ACTIONS(1085), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1087), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3598), 1,
      anon_sym_BQUOTE,
    STATE(2690), 1,
      aux_sym__literal_repeat1,
    STATE(3100), 1,
      sym__expression,
    ACTIONS(103), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(105), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1073), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1083), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(1089), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3053), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2500), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [102340] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1308), 16,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_LPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1310), 27,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [102391] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1131), 1,
      anon_sym_LPAREN,
    ACTIONS(1141), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1145), 1,
      anon_sym_DOLLAR,
    ACTIONS(1149), 1,
      anon_sym_DQUOTE,
    ACTIONS(1153), 1,
      aux_sym_number_token1,
    ACTIONS(1155), 1,
      aux_sym_number_token2,
    ACTIONS(1157), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1159), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1163), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1169), 1,
      sym__brace_start,
    ACTIONS(1171), 1,
      sym_word,
    ACTIONS(1173), 1,
      anon_sym_BANG,
    ACTIONS(1179), 1,
      anon_sym_TILDE,
    ACTIONS(1181), 1,
      sym__special_character,
    ACTIONS(1185), 1,
      sym_test_operator,
    ACTIONS(3060), 1,
      anon_sym_BQUOTE,
    STATE(2484), 1,
      aux_sym__literal_repeat1,
    STATE(2987), 1,
      sym__expression,
    ACTIONS(1129), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1165), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1175), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1177), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1183), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2594), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2456), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [102488] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1131), 1,
      anon_sym_LPAREN,
    ACTIONS(1141), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1145), 1,
      anon_sym_DOLLAR,
    ACTIONS(1149), 1,
      anon_sym_DQUOTE,
    ACTIONS(1153), 1,
      aux_sym_number_token1,
    ACTIONS(1155), 1,
      aux_sym_number_token2,
    ACTIONS(1157), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1159), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1163), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1169), 1,
      sym__brace_start,
    ACTIONS(1171), 1,
      sym_word,
    ACTIONS(1173), 1,
      anon_sym_BANG,
    ACTIONS(1179), 1,
      anon_sym_TILDE,
    ACTIONS(1181), 1,
      sym__special_character,
    ACTIONS(1185), 1,
      sym_test_operator,
    ACTIONS(3060), 1,
      anon_sym_BQUOTE,
    STATE(2484), 1,
      aux_sym__literal_repeat1,
    STATE(3039), 1,
      sym__expression,
    ACTIONS(1129), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1165), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1175), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1177), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1183), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2594), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2456), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [102585] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1131), 1,
      anon_sym_LPAREN,
    ACTIONS(1141), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1145), 1,
      anon_sym_DOLLAR,
    ACTIONS(1149), 1,
      anon_sym_DQUOTE,
    ACTIONS(1153), 1,
      aux_sym_number_token1,
    ACTIONS(1155), 1,
      aux_sym_number_token2,
    ACTIONS(1157), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1159), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1163), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1169), 1,
      sym__brace_start,
    ACTIONS(1171), 1,
      sym_word,
    ACTIONS(1173), 1,
      anon_sym_BANG,
    ACTIONS(1179), 1,
      anon_sym_TILDE,
    ACTIONS(1181), 1,
      sym__special_character,
    ACTIONS(1185), 1,
      sym_test_operator,
    ACTIONS(3060), 1,
      anon_sym_BQUOTE,
    STATE(2484), 1,
      aux_sym__literal_repeat1,
    STATE(3046), 1,
      sym__expression,
    ACTIONS(1129), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1165), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1175), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1177), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1183), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2594), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2456), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [102682] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(107), 1,
      anon_sym_TILDE,
    ACTIONS(237), 1,
      sym_word,
    ACTIONS(248), 1,
      anon_sym_LPAREN,
    ACTIONS(250), 1,
      anon_sym_BANG,
    ACTIONS(258), 1,
      anon_sym_DOLLAR,
    ACTIONS(264), 1,
      aux_sym_number_token1,
    ACTIONS(266), 1,
      aux_sym_number_token2,
    ACTIONS(270), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(282), 1,
      sym_test_operator,
    ACTIONS(284), 1,
      sym__brace_start,
    ACTIONS(1075), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1079), 1,
      sym__special_character,
    ACTIONS(1081), 1,
      anon_sym_DQUOTE,
    ACTIONS(1085), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1087), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3598), 1,
      anon_sym_BQUOTE,
    STATE(2690), 1,
      aux_sym__literal_repeat1,
    STATE(3091), 1,
      sym__expression,
    ACTIONS(103), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(105), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1073), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1083), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(1089), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3053), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2500), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [102779] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1131), 1,
      anon_sym_LPAREN,
    ACTIONS(1141), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1145), 1,
      anon_sym_DOLLAR,
    ACTIONS(1149), 1,
      anon_sym_DQUOTE,
    ACTIONS(1153), 1,
      aux_sym_number_token1,
    ACTIONS(1155), 1,
      aux_sym_number_token2,
    ACTIONS(1157), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1159), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1163), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1169), 1,
      sym__brace_start,
    ACTIONS(1171), 1,
      sym_word,
    ACTIONS(1173), 1,
      anon_sym_BANG,
    ACTIONS(1179), 1,
      anon_sym_TILDE,
    ACTIONS(1181), 1,
      sym__special_character,
    ACTIONS(1185), 1,
      sym_test_operator,
    ACTIONS(3060), 1,
      anon_sym_BQUOTE,
    STATE(2484), 1,
      aux_sym__literal_repeat1,
    STATE(2943), 1,
      sym__expression,
    ACTIONS(1129), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1165), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1175), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1177), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1183), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2594), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2456), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [102876] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1131), 1,
      anon_sym_LPAREN,
    ACTIONS(1141), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1145), 1,
      anon_sym_DOLLAR,
    ACTIONS(1149), 1,
      anon_sym_DQUOTE,
    ACTIONS(1153), 1,
      aux_sym_number_token1,
    ACTIONS(1155), 1,
      aux_sym_number_token2,
    ACTIONS(1157), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1159), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1163), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1169), 1,
      sym__brace_start,
    ACTIONS(1171), 1,
      sym_word,
    ACTIONS(1173), 1,
      anon_sym_BANG,
    ACTIONS(1179), 1,
      anon_sym_TILDE,
    ACTIONS(1181), 1,
      sym__special_character,
    ACTIONS(1185), 1,
      sym_test_operator,
    ACTIONS(3060), 1,
      anon_sym_BQUOTE,
    STATE(2484), 1,
      aux_sym__literal_repeat1,
    STATE(3059), 1,
      sym__expression,
    ACTIONS(1129), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1165), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1175), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1177), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1183), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2594), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2456), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [102973] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1131), 1,
      anon_sym_LPAREN,
    ACTIONS(1141), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1145), 1,
      anon_sym_DOLLAR,
    ACTIONS(1149), 1,
      anon_sym_DQUOTE,
    ACTIONS(1153), 1,
      aux_sym_number_token1,
    ACTIONS(1155), 1,
      aux_sym_number_token2,
    ACTIONS(1157), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1159), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1163), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1169), 1,
      sym__brace_start,
    ACTIONS(1171), 1,
      sym_word,
    ACTIONS(1173), 1,
      anon_sym_BANG,
    ACTIONS(1179), 1,
      anon_sym_TILDE,
    ACTIONS(1181), 1,
      sym__special_character,
    ACTIONS(1185), 1,
      sym_test_operator,
    ACTIONS(3060), 1,
      anon_sym_BQUOTE,
    STATE(2484), 1,
      aux_sym__literal_repeat1,
    STATE(3061), 1,
      sym__expression,
    ACTIONS(1129), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1165), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1175), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1177), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1183), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2594), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2456), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [103070] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1131), 1,
      anon_sym_LPAREN,
    ACTIONS(1141), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1145), 1,
      anon_sym_DOLLAR,
    ACTIONS(1149), 1,
      anon_sym_DQUOTE,
    ACTIONS(1153), 1,
      aux_sym_number_token1,
    ACTIONS(1155), 1,
      aux_sym_number_token2,
    ACTIONS(1157), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1159), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1163), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1169), 1,
      sym__brace_start,
    ACTIONS(1171), 1,
      sym_word,
    ACTIONS(1173), 1,
      anon_sym_BANG,
    ACTIONS(1179), 1,
      anon_sym_TILDE,
    ACTIONS(1181), 1,
      sym__special_character,
    ACTIONS(1185), 1,
      sym_test_operator,
    ACTIONS(3060), 1,
      anon_sym_BQUOTE,
    STATE(2484), 1,
      aux_sym__literal_repeat1,
    STATE(3062), 1,
      sym__expression,
    ACTIONS(1129), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1165), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1175), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1177), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1183), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2594), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2456), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [103167] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1300), 16,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_LPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1302), 27,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [103218] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1131), 1,
      anon_sym_LPAREN,
    ACTIONS(1141), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1145), 1,
      anon_sym_DOLLAR,
    ACTIONS(1149), 1,
      anon_sym_DQUOTE,
    ACTIONS(1153), 1,
      aux_sym_number_token1,
    ACTIONS(1155), 1,
      aux_sym_number_token2,
    ACTIONS(1157), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1159), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1163), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1169), 1,
      sym__brace_start,
    ACTIONS(1171), 1,
      sym_word,
    ACTIONS(1173), 1,
      anon_sym_BANG,
    ACTIONS(1179), 1,
      anon_sym_TILDE,
    ACTIONS(1181), 1,
      sym__special_character,
    ACTIONS(1185), 1,
      sym_test_operator,
    ACTIONS(3060), 1,
      anon_sym_BQUOTE,
    STATE(2484), 1,
      aux_sym__literal_repeat1,
    STATE(3063), 1,
      sym__expression,
    ACTIONS(1129), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1165), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1175), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1177), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1183), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2594), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2456), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [103315] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1131), 1,
      anon_sym_LPAREN,
    ACTIONS(1141), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1145), 1,
      anon_sym_DOLLAR,
    ACTIONS(1149), 1,
      anon_sym_DQUOTE,
    ACTIONS(1153), 1,
      aux_sym_number_token1,
    ACTIONS(1155), 1,
      aux_sym_number_token2,
    ACTIONS(1157), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1159), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1163), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1169), 1,
      sym__brace_start,
    ACTIONS(1171), 1,
      sym_word,
    ACTIONS(1173), 1,
      anon_sym_BANG,
    ACTIONS(1179), 1,
      anon_sym_TILDE,
    ACTIONS(1181), 1,
      sym__special_character,
    ACTIONS(1185), 1,
      sym_test_operator,
    ACTIONS(3060), 1,
      anon_sym_BQUOTE,
    STATE(2484), 1,
      aux_sym__literal_repeat1,
    STATE(3064), 1,
      sym__expression,
    ACTIONS(1129), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1165), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1175), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1177), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1183), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2594), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2456), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [103412] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1127), 1,
      sym_word,
    ACTIONS(1131), 1,
      anon_sym_LPAREN,
    ACTIONS(1133), 1,
      anon_sym_BANG,
    ACTIONS(1139), 1,
      anon_sym_TILDE,
    ACTIONS(1141), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1145), 1,
      anon_sym_DOLLAR,
    ACTIONS(1147), 1,
      sym__special_character,
    ACTIONS(1149), 1,
      anon_sym_DQUOTE,
    ACTIONS(1153), 1,
      aux_sym_number_token1,
    ACTIONS(1155), 1,
      aux_sym_number_token2,
    ACTIONS(1157), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1159), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1163), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1167), 1,
      sym_test_operator,
    ACTIONS(1169), 1,
      sym__brace_start,
    ACTIONS(3060), 1,
      anon_sym_BQUOTE,
    STATE(2484), 1,
      aux_sym__literal_repeat1,
    STATE(3102), 1,
      sym__expression,
    ACTIONS(1129), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1135), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1137), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1151), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(1165), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(2594), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2573), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [103509] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1131), 1,
      anon_sym_LPAREN,
    ACTIONS(1141), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1145), 1,
      anon_sym_DOLLAR,
    ACTIONS(1149), 1,
      anon_sym_DQUOTE,
    ACTIONS(1153), 1,
      aux_sym_number_token1,
    ACTIONS(1155), 1,
      aux_sym_number_token2,
    ACTIONS(1157), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1159), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1163), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1169), 1,
      sym__brace_start,
    ACTIONS(1211), 1,
      sym_word,
    ACTIONS(1213), 1,
      anon_sym_BANG,
    ACTIONS(1219), 1,
      anon_sym_TILDE,
    ACTIONS(1221), 1,
      sym__special_character,
    ACTIONS(1225), 1,
      sym_test_operator,
    ACTIONS(3060), 1,
      anon_sym_BQUOTE,
    STATE(2484), 1,
      aux_sym__literal_repeat1,
    STATE(3351), 1,
      sym__expression,
    ACTIONS(1129), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1165), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1215), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1217), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1223), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2594), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2588), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [103606] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1131), 1,
      anon_sym_LPAREN,
    ACTIONS(1141), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1145), 1,
      anon_sym_DOLLAR,
    ACTIONS(1149), 1,
      anon_sym_DQUOTE,
    ACTIONS(1153), 1,
      aux_sym_number_token1,
    ACTIONS(1155), 1,
      aux_sym_number_token2,
    ACTIONS(1157), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1159), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1163), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1169), 1,
      sym__brace_start,
    ACTIONS(1171), 1,
      sym_word,
    ACTIONS(1173), 1,
      anon_sym_BANG,
    ACTIONS(1179), 1,
      anon_sym_TILDE,
    ACTIONS(1181), 1,
      sym__special_character,
    ACTIONS(1185), 1,
      sym_test_operator,
    ACTIONS(3060), 1,
      anon_sym_BQUOTE,
    STATE(2484), 1,
      aux_sym__literal_repeat1,
    STATE(3067), 1,
      sym__expression,
    ACTIONS(1129), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1165), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1175), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1177), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1183), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2594), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2456), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [103703] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1131), 1,
      anon_sym_LPAREN,
    ACTIONS(1141), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1145), 1,
      anon_sym_DOLLAR,
    ACTIONS(1149), 1,
      anon_sym_DQUOTE,
    ACTIONS(1153), 1,
      aux_sym_number_token1,
    ACTIONS(1155), 1,
      aux_sym_number_token2,
    ACTIONS(1157), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1159), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1163), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1169), 1,
      sym__brace_start,
    ACTIONS(1211), 1,
      sym_word,
    ACTIONS(1213), 1,
      anon_sym_BANG,
    ACTIONS(1219), 1,
      anon_sym_TILDE,
    ACTIONS(1221), 1,
      sym__special_character,
    ACTIONS(1225), 1,
      sym_test_operator,
    ACTIONS(3060), 1,
      anon_sym_BQUOTE,
    STATE(2484), 1,
      aux_sym__literal_repeat1,
    STATE(3356), 1,
      sym__expression,
    ACTIONS(1129), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1165), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1215), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1217), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1223), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2594), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2588), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [103800] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1131), 1,
      anon_sym_LPAREN,
    ACTIONS(1141), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1145), 1,
      anon_sym_DOLLAR,
    ACTIONS(1149), 1,
      anon_sym_DQUOTE,
    ACTIONS(1153), 1,
      aux_sym_number_token1,
    ACTIONS(1155), 1,
      aux_sym_number_token2,
    ACTIONS(1157), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1159), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1163), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1169), 1,
      sym__brace_start,
    ACTIONS(1171), 1,
      sym_word,
    ACTIONS(1173), 1,
      anon_sym_BANG,
    ACTIONS(1179), 1,
      anon_sym_TILDE,
    ACTIONS(1181), 1,
      sym__special_character,
    ACTIONS(1185), 1,
      sym_test_operator,
    ACTIONS(3060), 1,
      anon_sym_BQUOTE,
    STATE(2484), 1,
      aux_sym__literal_repeat1,
    STATE(3068), 1,
      sym__expression,
    ACTIONS(1129), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1165), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1175), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1177), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1183), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2594), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2456), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [103897] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(363), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(367), 1,
      anon_sym_DOLLAR,
    ACTIONS(371), 1,
      anon_sym_DQUOTE,
    ACTIONS(375), 1,
      aux_sym_number_token1,
    ACTIONS(377), 1,
      aux_sym_number_token2,
    ACTIONS(379), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(381), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(385), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(391), 1,
      sym__brace_start,
    ACTIONS(1091), 1,
      sym_word,
    ACTIONS(1093), 1,
      anon_sym_LPAREN,
    ACTIONS(1095), 1,
      anon_sym_BANG,
    ACTIONS(1103), 1,
      anon_sym_TILDE,
    ACTIONS(1113), 1,
      sym_test_operator,
    ACTIONS(5809), 1,
      anon_sym_BQUOTE,
    ACTIONS(5916), 1,
      sym__special_character,
    STATE(2472), 1,
      aux_sym__literal_repeat1,
    STATE(2814), 1,
      sym__expression,
    ACTIONS(352), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(387), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1099), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1101), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1109), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2863), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2451), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [103994] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5828), 1,
      sym_file_descriptor,
    ACTIONS(5826), 2,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5831), 3,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(5821), 9,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
    ACTIONS(5823), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
    ACTIONS(5819), 17,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [104053] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1127), 1,
      sym_word,
    ACTIONS(1131), 1,
      anon_sym_LPAREN,
    ACTIONS(1133), 1,
      anon_sym_BANG,
    ACTIONS(1139), 1,
      anon_sym_TILDE,
    ACTIONS(1141), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1145), 1,
      anon_sym_DOLLAR,
    ACTIONS(1147), 1,
      sym__special_character,
    ACTIONS(1149), 1,
      anon_sym_DQUOTE,
    ACTIONS(1153), 1,
      aux_sym_number_token1,
    ACTIONS(1155), 1,
      aux_sym_number_token2,
    ACTIONS(1157), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1159), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1163), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1167), 1,
      sym_test_operator,
    ACTIONS(1169), 1,
      sym__brace_start,
    ACTIONS(3060), 1,
      anon_sym_BQUOTE,
    STATE(2484), 1,
      aux_sym__literal_repeat1,
    STATE(2621), 1,
      sym__expression,
    ACTIONS(1129), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1135), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1137), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1151), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(1165), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(2594), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2573), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [104150] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1296), 16,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_LPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1298), 27,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [104201] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1314), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1312), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [104252] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1127), 1,
      sym_word,
    ACTIONS(1131), 1,
      anon_sym_LPAREN,
    ACTIONS(1133), 1,
      anon_sym_BANG,
    ACTIONS(1139), 1,
      anon_sym_TILDE,
    ACTIONS(1141), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1145), 1,
      anon_sym_DOLLAR,
    ACTIONS(1147), 1,
      sym__special_character,
    ACTIONS(1149), 1,
      anon_sym_DQUOTE,
    ACTIONS(1153), 1,
      aux_sym_number_token1,
    ACTIONS(1155), 1,
      aux_sym_number_token2,
    ACTIONS(1157), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1159), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1163), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1167), 1,
      sym_test_operator,
    ACTIONS(1169), 1,
      sym__brace_start,
    ACTIONS(3060), 1,
      anon_sym_BQUOTE,
    STATE(2484), 1,
      aux_sym__literal_repeat1,
    STATE(3037), 1,
      sym__expression,
    ACTIONS(1129), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1135), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1137), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1151), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(1165), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(2594), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2573), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [104349] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5826), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5828), 1,
      sym_file_descriptor,
    ACTIONS(6135), 1,
      anon_sym_RPAREN,
    ACTIONS(5831), 3,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(5821), 9,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
    ACTIONS(5823), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
    ACTIONS(5819), 17,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [104410] = 27,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(363), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(367), 1,
      anon_sym_DOLLAR,
    ACTIONS(371), 1,
      anon_sym_DQUOTE,
    ACTIONS(375), 1,
      aux_sym_number_token1,
    ACTIONS(377), 1,
      aux_sym_number_token2,
    ACTIONS(379), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(381), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(385), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(391), 1,
      sym__brace_start,
    ACTIONS(1093), 1,
      anon_sym_LPAREN,
    ACTIONS(1095), 1,
      anon_sym_BANG,
    ACTIONS(1103), 1,
      anon_sym_TILDE,
    ACTIONS(5809), 1,
      anon_sym_BQUOTE,
    ACTIONS(5916), 1,
      sym__special_character,
    ACTIONS(6138), 1,
      sym_word,
    ACTIONS(6142), 1,
      sym_test_operator,
    STATE(2561), 1,
      aux_sym__literal_repeat1,
    STATE(3397), 1,
      sym__expression,
    ACTIONS(352), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(387), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1099), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1101), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(6140), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2594), 2,
      sym_ternary_expression,
      sym_postfix_expression,
    STATE(2874), 4,
      sym_binary_expression,
      sym_unary_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2563), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [104509] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(363), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(367), 1,
      anon_sym_DOLLAR,
    ACTIONS(371), 1,
      anon_sym_DQUOTE,
    ACTIONS(375), 1,
      aux_sym_number_token1,
    ACTIONS(377), 1,
      aux_sym_number_token2,
    ACTIONS(379), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(381), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(385), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(391), 1,
      sym__brace_start,
    ACTIONS(1091), 1,
      sym_word,
    ACTIONS(1093), 1,
      anon_sym_LPAREN,
    ACTIONS(1095), 1,
      anon_sym_BANG,
    ACTIONS(1103), 1,
      anon_sym_TILDE,
    ACTIONS(1113), 1,
      sym_test_operator,
    ACTIONS(5809), 1,
      anon_sym_BQUOTE,
    ACTIONS(5916), 1,
      sym__special_character,
    STATE(2472), 1,
      aux_sym__literal_repeat1,
    STATE(2812), 1,
      sym__expression,
    ACTIONS(352), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(387), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1099), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1101), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1109), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2863), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2451), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [104606] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(363), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(367), 1,
      anon_sym_DOLLAR,
    ACTIONS(371), 1,
      anon_sym_DQUOTE,
    ACTIONS(375), 1,
      aux_sym_number_token1,
    ACTIONS(377), 1,
      aux_sym_number_token2,
    ACTIONS(379), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(381), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(385), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(391), 1,
      sym__brace_start,
    ACTIONS(1091), 1,
      sym_word,
    ACTIONS(1093), 1,
      anon_sym_LPAREN,
    ACTIONS(1095), 1,
      anon_sym_BANG,
    ACTIONS(1103), 1,
      anon_sym_TILDE,
    ACTIONS(1113), 1,
      sym_test_operator,
    ACTIONS(5809), 1,
      anon_sym_BQUOTE,
    ACTIONS(5916), 1,
      sym__special_character,
    STATE(2472), 1,
      aux_sym__literal_repeat1,
    STATE(2810), 1,
      sym__expression,
    ACTIONS(352), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(387), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1099), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1101), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1109), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2863), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2451), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [104703] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(363), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(367), 1,
      anon_sym_DOLLAR,
    ACTIONS(371), 1,
      anon_sym_DQUOTE,
    ACTIONS(375), 1,
      aux_sym_number_token1,
    ACTIONS(377), 1,
      aux_sym_number_token2,
    ACTIONS(379), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(381), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(385), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(391), 1,
      sym__brace_start,
    ACTIONS(1091), 1,
      sym_word,
    ACTIONS(1093), 1,
      anon_sym_LPAREN,
    ACTIONS(1095), 1,
      anon_sym_BANG,
    ACTIONS(1103), 1,
      anon_sym_TILDE,
    ACTIONS(1113), 1,
      sym_test_operator,
    ACTIONS(5809), 1,
      anon_sym_BQUOTE,
    ACTIONS(5916), 1,
      sym__special_character,
    STATE(2472), 1,
      aux_sym__literal_repeat1,
    STATE(2832), 1,
      sym__expression,
    ACTIONS(352), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(387), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1099), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1101), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1109), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2863), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2451), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [104800] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1304), 16,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_LPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1306), 27,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [104851] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1131), 1,
      anon_sym_LPAREN,
    ACTIONS(1141), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1145), 1,
      anon_sym_DOLLAR,
    ACTIONS(1149), 1,
      anon_sym_DQUOTE,
    ACTIONS(1153), 1,
      aux_sym_number_token1,
    ACTIONS(1155), 1,
      aux_sym_number_token2,
    ACTIONS(1157), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1159), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1163), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1169), 1,
      sym__brace_start,
    ACTIONS(1211), 1,
      sym_word,
    ACTIONS(1213), 1,
      anon_sym_BANG,
    ACTIONS(1219), 1,
      anon_sym_TILDE,
    ACTIONS(1221), 1,
      sym__special_character,
    ACTIONS(1225), 1,
      sym_test_operator,
    ACTIONS(3060), 1,
      anon_sym_BQUOTE,
    STATE(2484), 1,
      aux_sym__literal_repeat1,
    STATE(3358), 1,
      sym__expression,
    ACTIONS(1129), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1165), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1215), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1217), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1223), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2594), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2588), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [104948] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(6013), 1,
      aux_sym_concatenation_token1,
    ACTIONS(6144), 1,
      sym__concat,
    STATE(1879), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1267), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1265), 36,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [105005] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(6013), 1,
      aux_sym_concatenation_token1,
    ACTIONS(6146), 1,
      sym__concat,
    STATE(1879), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1288), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1286), 36,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [105062] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1131), 1,
      anon_sym_LPAREN,
    ACTIONS(1141), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1145), 1,
      anon_sym_DOLLAR,
    ACTIONS(1149), 1,
      anon_sym_DQUOTE,
    ACTIONS(1153), 1,
      aux_sym_number_token1,
    ACTIONS(1155), 1,
      aux_sym_number_token2,
    ACTIONS(1157), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1159), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1163), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1169), 1,
      sym__brace_start,
    ACTIONS(1211), 1,
      sym_word,
    ACTIONS(1213), 1,
      anon_sym_BANG,
    ACTIONS(1219), 1,
      anon_sym_TILDE,
    ACTIONS(1221), 1,
      sym__special_character,
    ACTIONS(1225), 1,
      sym_test_operator,
    ACTIONS(3060), 1,
      anon_sym_BQUOTE,
    STATE(2484), 1,
      aux_sym__literal_repeat1,
    STATE(3193), 1,
      sym__expression,
    ACTIONS(1129), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1165), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1215), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1217), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1223), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2594), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2588), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [105159] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5826), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5828), 1,
      sym_file_descriptor,
    ACTIONS(6148), 1,
      anon_sym_RPAREN,
    ACTIONS(5831), 3,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(5821), 9,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
    ACTIONS(5823), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
    ACTIONS(5819), 17,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [105220] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5768), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5766), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [105271] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(107), 1,
      anon_sym_TILDE,
    ACTIONS(237), 1,
      sym_word,
    ACTIONS(248), 1,
      anon_sym_LPAREN,
    ACTIONS(250), 1,
      anon_sym_BANG,
    ACTIONS(258), 1,
      anon_sym_DOLLAR,
    ACTIONS(264), 1,
      aux_sym_number_token1,
    ACTIONS(266), 1,
      aux_sym_number_token2,
    ACTIONS(270), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(282), 1,
      sym_test_operator,
    ACTIONS(284), 1,
      sym__brace_start,
    ACTIONS(1075), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1079), 1,
      sym__special_character,
    ACTIONS(1081), 1,
      anon_sym_DQUOTE,
    ACTIONS(1085), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1087), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3598), 1,
      anon_sym_BQUOTE,
    STATE(2690), 1,
      aux_sym__literal_repeat1,
    STATE(3094), 1,
      sym__expression,
    ACTIONS(103), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(105), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1073), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1083), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(1089), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3053), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2500), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [105368] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1338), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1336), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [105419] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(6013), 1,
      aux_sym_concatenation_token1,
    ACTIONS(6015), 1,
      sym__concat,
    STATE(2067), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1263), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1261), 36,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [105476] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1251), 16,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_LPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1253), 27,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [105527] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5826), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5828), 1,
      sym_file_descriptor,
    ACTIONS(6151), 1,
      anon_sym_RPAREN,
    ACTIONS(5831), 3,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(5821), 9,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
    ACTIONS(5823), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
    ACTIONS(5819), 17,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [105588] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1324), 16,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_LPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1326), 27,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [105639] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5826), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5828), 1,
      sym_file_descriptor,
    ACTIONS(6154), 1,
      anon_sym_RPAREN,
    ACTIONS(5831), 3,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(5821), 9,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
    ACTIONS(5823), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
    ACTIONS(5819), 17,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [105700] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5826), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5828), 1,
      sym_file_descriptor,
    ACTIONS(6157), 1,
      anon_sym_RPAREN,
    ACTIONS(5831), 3,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(5821), 9,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
    ACTIONS(5823), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
    ACTIONS(5819), 17,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [105761] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1354), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1352), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [105812] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1346), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1344), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [105863] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5826), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5828), 1,
      sym_file_descriptor,
    ACTIONS(6160), 1,
      anon_sym_RPAREN,
    ACTIONS(5831), 3,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(5821), 9,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
    ACTIONS(5823), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
    ACTIONS(5819), 17,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [105924] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5826), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5828), 1,
      sym_file_descriptor,
    ACTIONS(6163), 1,
      anon_sym_RPAREN,
    ACTIONS(5831), 3,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(5821), 9,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
    ACTIONS(5823), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
    ACTIONS(5819), 17,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [105985] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5826), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5828), 1,
      sym_file_descriptor,
    ACTIONS(6166), 1,
      anon_sym_RPAREN,
    ACTIONS(5831), 3,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(5821), 9,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
    ACTIONS(5823), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
    ACTIONS(5819), 17,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [106046] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5826), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5828), 1,
      sym_file_descriptor,
    ACTIONS(6169), 1,
      anon_sym_RPAREN,
    ACTIONS(5831), 3,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(5821), 9,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
    ACTIONS(5823), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
    ACTIONS(5819), 17,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [106107] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1358), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1356), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [106158] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(107), 1,
      anon_sym_TILDE,
    ACTIONS(237), 1,
      sym_word,
    ACTIONS(248), 1,
      anon_sym_LPAREN,
    ACTIONS(250), 1,
      anon_sym_BANG,
    ACTIONS(258), 1,
      anon_sym_DOLLAR,
    ACTIONS(264), 1,
      aux_sym_number_token1,
    ACTIONS(266), 1,
      aux_sym_number_token2,
    ACTIONS(270), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(282), 1,
      sym_test_operator,
    ACTIONS(284), 1,
      sym__brace_start,
    ACTIONS(1075), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1079), 1,
      sym__special_character,
    ACTIONS(1081), 1,
      anon_sym_DQUOTE,
    ACTIONS(1085), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1087), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3598), 1,
      anon_sym_BQUOTE,
    STATE(2690), 1,
      aux_sym__literal_repeat1,
    STATE(3130), 1,
      sym__expression,
    ACTIONS(103), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(105), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1073), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1083), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(1089), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3053), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2500), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [106255] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5826), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5828), 1,
      sym_file_descriptor,
    ACTIONS(6172), 1,
      anon_sym_RPAREN,
    ACTIONS(5831), 3,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(5821), 9,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
    ACTIONS(5823), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
    ACTIONS(5819), 17,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [106316] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(107), 1,
      anon_sym_TILDE,
    ACTIONS(237), 1,
      sym_word,
    ACTIONS(248), 1,
      anon_sym_LPAREN,
    ACTIONS(250), 1,
      anon_sym_BANG,
    ACTIONS(258), 1,
      anon_sym_DOLLAR,
    ACTIONS(264), 1,
      aux_sym_number_token1,
    ACTIONS(266), 1,
      aux_sym_number_token2,
    ACTIONS(270), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(282), 1,
      sym_test_operator,
    ACTIONS(284), 1,
      sym__brace_start,
    ACTIONS(1075), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1079), 1,
      sym__special_character,
    ACTIONS(1081), 1,
      anon_sym_DQUOTE,
    ACTIONS(1085), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1087), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3598), 1,
      anon_sym_BQUOTE,
    STATE(2690), 1,
      aux_sym__literal_repeat1,
    STATE(3107), 1,
      sym__expression,
    ACTIONS(103), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(105), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1073), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1083), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(1089), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3053), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2500), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [106413] = 27,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(363), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(367), 1,
      anon_sym_DOLLAR,
    ACTIONS(371), 1,
      anon_sym_DQUOTE,
    ACTIONS(375), 1,
      aux_sym_number_token1,
    ACTIONS(377), 1,
      aux_sym_number_token2,
    ACTIONS(379), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(381), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(385), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(391), 1,
      sym__brace_start,
    ACTIONS(1093), 1,
      anon_sym_LPAREN,
    ACTIONS(1095), 1,
      anon_sym_BANG,
    ACTIONS(1103), 1,
      anon_sym_TILDE,
    ACTIONS(5809), 1,
      anon_sym_BQUOTE,
    ACTIONS(5916), 1,
      sym__special_character,
    ACTIONS(6175), 1,
      sym_word,
    ACTIONS(6179), 1,
      sym_test_operator,
    STATE(2445), 1,
      aux_sym__literal_repeat1,
    STATE(3397), 1,
      sym__expression,
    ACTIONS(352), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(387), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1099), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1101), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(6177), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2594), 2,
      sym_ternary_expression,
      sym_postfix_expression,
    STATE(2834), 4,
      sym_binary_expression,
      sym_unary_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2461), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [106512] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1358), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1356), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [106563] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(187), 1,
      anon_sym_TILDE,
    ACTIONS(350), 1,
      sym_word,
    ACTIONS(359), 1,
      anon_sym_LPAREN,
    ACTIONS(361), 1,
      anon_sym_BANG,
    ACTIONS(363), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(367), 1,
      anon_sym_DOLLAR,
    ACTIONS(371), 1,
      anon_sym_DQUOTE,
    ACTIONS(375), 1,
      aux_sym_number_token1,
    ACTIONS(377), 1,
      aux_sym_number_token2,
    ACTIONS(379), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(381), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(385), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(389), 1,
      sym_test_operator,
    ACTIONS(391), 1,
      sym__brace_start,
    ACTIONS(5807), 1,
      sym__special_character,
    ACTIONS(5809), 1,
      anon_sym_BQUOTE,
    STATE(2641), 1,
      aux_sym__literal_repeat1,
    STATE(2951), 1,
      sym__expression,
    ACTIONS(183), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(185), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(352), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(373), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(387), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3071), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2521), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [106660] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1348), 16,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_LPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1350), 27,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [106711] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1263), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1261), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [106762] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1346), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1344), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [106813] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5826), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5828), 1,
      sym_file_descriptor,
    ACTIONS(6181), 1,
      anon_sym_RPAREN,
    ACTIONS(5831), 3,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(5821), 9,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
    ACTIONS(5823), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
    ACTIONS(5819), 17,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [106874] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5826), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5828), 1,
      sym_file_descriptor,
    ACTIONS(6184), 1,
      anon_sym_RPAREN,
    ACTIONS(5831), 3,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(5821), 9,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
    ACTIONS(5823), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
    ACTIONS(5819), 17,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [106935] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5826), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5828), 1,
      sym_file_descriptor,
    ACTIONS(6187), 1,
      anon_sym_RPAREN,
    ACTIONS(5831), 3,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(5821), 9,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
    ACTIONS(5823), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
    ACTIONS(5819), 17,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [106996] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5826), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5828), 1,
      sym_file_descriptor,
    ACTIONS(6190), 1,
      anon_sym_RPAREN,
    ACTIONS(5831), 3,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(5821), 9,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
    ACTIONS(5823), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
    ACTIONS(5819), 17,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [107057] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1338), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1336), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [107108] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5826), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5828), 1,
      sym_file_descriptor,
    ACTIONS(6193), 1,
      anon_sym_RPAREN,
    ACTIONS(5831), 3,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(5821), 9,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
    ACTIONS(5823), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
    ACTIONS(5819), 17,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [107169] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5826), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5828), 1,
      sym_file_descriptor,
    ACTIONS(6196), 1,
      anon_sym_RPAREN,
    ACTIONS(5831), 3,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(5821), 9,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
    ACTIONS(5823), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
    ACTIONS(5819), 17,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [107230] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5375), 6,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5373), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [107281] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5826), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5828), 1,
      sym_file_descriptor,
    ACTIONS(6199), 1,
      anon_sym_RPAREN,
    ACTIONS(5831), 3,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(5821), 9,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
    ACTIONS(5823), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
    ACTIONS(5819), 17,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [107342] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1314), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1312), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [107393] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5772), 6,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5770), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [107444] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4554), 6,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4552), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [107495] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1354), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1352), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [107546] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1342), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1340), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [107597] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5826), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5828), 1,
      sym_file_descriptor,
    ACTIONS(6202), 1,
      anon_sym_RPAREN,
    ACTIONS(5831), 3,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(5821), 9,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
    ACTIONS(5823), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
    ACTIONS(5819), 17,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [107658] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1330), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1328), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [107709] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5826), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5828), 1,
      sym_file_descriptor,
    ACTIONS(6205), 1,
      anon_sym_RPAREN,
    ACTIONS(5831), 3,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(5821), 9,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
    ACTIONS(5823), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
    ACTIONS(5819), 17,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [107770] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(6210), 1,
      anon_sym_DQUOTE,
    ACTIONS(6214), 1,
      sym_variable_name,
    STATE(3525), 1,
      sym_string,
    ACTIONS(6212), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1235), 3,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(6208), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1227), 26,
      anon_sym_LPAREN_LPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [107831] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1322), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1320), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [107882] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(6210), 1,
      anon_sym_DQUOTE,
    ACTIONS(6214), 1,
      sym_variable_name,
    STATE(3525), 1,
      sym_string,
    ACTIONS(6212), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1241), 3,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(6208), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1239), 26,
      anon_sym_LPAREN_LPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [107943] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1318), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1316), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [107994] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2074), 6,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(2072), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [108045] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(6216), 1,
      sym__special_character,
    STATE(2118), 1,
      aux_sym__literal_repeat1,
    ACTIONS(1364), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1362), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [108100] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(363), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(367), 1,
      anon_sym_DOLLAR,
    ACTIONS(371), 1,
      anon_sym_DQUOTE,
    ACTIONS(375), 1,
      aux_sym_number_token1,
    ACTIONS(377), 1,
      aux_sym_number_token2,
    ACTIONS(379), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(381), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(385), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(391), 1,
      sym__brace_start,
    ACTIONS(1091), 1,
      sym_word,
    ACTIONS(1093), 1,
      anon_sym_LPAREN,
    ACTIONS(1095), 1,
      anon_sym_BANG,
    ACTIONS(1103), 1,
      anon_sym_TILDE,
    ACTIONS(1113), 1,
      sym_test_operator,
    ACTIONS(5809), 1,
      anon_sym_BQUOTE,
    ACTIONS(5916), 1,
      sym__special_character,
    STATE(2472), 1,
      aux_sym__literal_repeat1,
    STATE(2868), 1,
      sym__expression,
    ACTIONS(352), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(387), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1099), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1101), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1109), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2863), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2451), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [108197] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5826), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5828), 1,
      sym_file_descriptor,
    ACTIONS(6219), 1,
      anon_sym_RPAREN,
    ACTIONS(5831), 3,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(5821), 9,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
    ACTIONS(5823), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
    ACTIONS(5819), 17,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [108258] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1350), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1348), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [108309] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4261), 3,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(5932), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4247), 18,
      anon_sym_LPAREN_LPAREN,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
    ACTIONS(5930), 19,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [108364] = 6,
    ACTIONS(71), 1,
      sym_comment,
    STATE(3541), 1,
      aux_sym__literal_repeat1,
    STATE(3573), 1,
      sym_concatenation,
    STATE(3481), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2498), 10,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(2500), 22,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [108421] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(6222), 1,
      sym__special_character,
    STATE(2124), 1,
      aux_sym__literal_repeat1,
    ACTIONS(1364), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1362), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [108476] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(2135), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(6133), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(4552), 15,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(4554), 25,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [108531] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(2152), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(6133), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(4564), 15,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(4566), 25,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [108586] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(2127), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(6225), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(1251), 15,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1253), 25,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [108641] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5826), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5828), 1,
      sym_file_descriptor,
    ACTIONS(6228), 1,
      anon_sym_RPAREN,
    ACTIONS(5831), 3,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(5821), 9,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
    ACTIONS(5823), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
    ACTIONS(5819), 17,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [108702] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(187), 1,
      anon_sym_TILDE,
    ACTIONS(350), 1,
      sym_word,
    ACTIONS(359), 1,
      anon_sym_LPAREN,
    ACTIONS(361), 1,
      anon_sym_BANG,
    ACTIONS(363), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(367), 1,
      anon_sym_DOLLAR,
    ACTIONS(371), 1,
      anon_sym_DQUOTE,
    ACTIONS(375), 1,
      aux_sym_number_token1,
    ACTIONS(377), 1,
      aux_sym_number_token2,
    ACTIONS(379), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(381), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(385), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(389), 1,
      sym_test_operator,
    ACTIONS(391), 1,
      sym__brace_start,
    ACTIONS(5807), 1,
      sym__special_character,
    ACTIONS(5809), 1,
      anon_sym_BQUOTE,
    STATE(2641), 1,
      aux_sym__literal_repeat1,
    STATE(3009), 1,
      sym__expression,
    ACTIONS(183), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(185), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(352), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(373), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(387), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3071), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2521), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [108799] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(187), 1,
      anon_sym_TILDE,
    ACTIONS(350), 1,
      sym_word,
    ACTIONS(359), 1,
      anon_sym_LPAREN,
    ACTIONS(361), 1,
      anon_sym_BANG,
    ACTIONS(363), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(367), 1,
      anon_sym_DOLLAR,
    ACTIONS(371), 1,
      anon_sym_DQUOTE,
    ACTIONS(375), 1,
      aux_sym_number_token1,
    ACTIONS(377), 1,
      aux_sym_number_token2,
    ACTIONS(379), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(381), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(385), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(389), 1,
      sym_test_operator,
    ACTIONS(391), 1,
      sym__brace_start,
    ACTIONS(5807), 1,
      sym__special_character,
    ACTIONS(5809), 1,
      anon_sym_BQUOTE,
    STATE(2641), 1,
      aux_sym__literal_repeat1,
    STATE(3006), 1,
      sym__expression,
    ACTIONS(183), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(185), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(352), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(373), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(387), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3071), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2521), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [108896] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1131), 1,
      anon_sym_LPAREN,
    ACTIONS(1141), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1145), 1,
      anon_sym_DOLLAR,
    ACTIONS(1149), 1,
      anon_sym_DQUOTE,
    ACTIONS(1153), 1,
      aux_sym_number_token1,
    ACTIONS(1155), 1,
      aux_sym_number_token2,
    ACTIONS(1157), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1159), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1163), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1169), 1,
      sym__brace_start,
    ACTIONS(1171), 1,
      sym_word,
    ACTIONS(1173), 1,
      anon_sym_BANG,
    ACTIONS(1179), 1,
      anon_sym_TILDE,
    ACTIONS(1181), 1,
      sym__special_character,
    ACTIONS(1185), 1,
      sym_test_operator,
    ACTIONS(3060), 1,
      anon_sym_BQUOTE,
    STATE(2484), 1,
      aux_sym__literal_repeat1,
    STATE(2621), 1,
      sym__expression,
    ACTIONS(1129), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1165), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1175), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1177), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1183), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2594), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2456), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [108993] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1131), 1,
      anon_sym_LPAREN,
    ACTIONS(1141), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1145), 1,
      anon_sym_DOLLAR,
    ACTIONS(1149), 1,
      anon_sym_DQUOTE,
    ACTIONS(1153), 1,
      aux_sym_number_token1,
    ACTIONS(1155), 1,
      aux_sym_number_token2,
    ACTIONS(1157), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1159), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1163), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1169), 1,
      sym__brace_start,
    ACTIONS(1171), 1,
      sym_word,
    ACTIONS(1173), 1,
      anon_sym_BANG,
    ACTIONS(1179), 1,
      anon_sym_TILDE,
    ACTIONS(1181), 1,
      sym__special_character,
    ACTIONS(1185), 1,
      sym_test_operator,
    ACTIONS(3060), 1,
      anon_sym_BQUOTE,
    STATE(2484), 1,
      aux_sym__literal_repeat1,
    STATE(3101), 1,
      sym__expression,
    ACTIONS(1129), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1165), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1175), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1177), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1183), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2594), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2456), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [109090] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5826), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5828), 1,
      sym_file_descriptor,
    ACTIONS(6231), 1,
      anon_sym_RPAREN,
    ACTIONS(5831), 3,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(5821), 9,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
    ACTIONS(5823), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
    ACTIONS(5819), 17,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [109151] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(107), 1,
      anon_sym_TILDE,
    ACTIONS(237), 1,
      sym_word,
    ACTIONS(248), 1,
      anon_sym_LPAREN,
    ACTIONS(250), 1,
      anon_sym_BANG,
    ACTIONS(258), 1,
      anon_sym_DOLLAR,
    ACTIONS(264), 1,
      aux_sym_number_token1,
    ACTIONS(266), 1,
      aux_sym_number_token2,
    ACTIONS(270), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(282), 1,
      sym_test_operator,
    ACTIONS(284), 1,
      sym__brace_start,
    ACTIONS(1075), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1079), 1,
      sym__special_character,
    ACTIONS(1081), 1,
      anon_sym_DQUOTE,
    ACTIONS(1085), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1087), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3598), 1,
      anon_sym_BQUOTE,
    STATE(2690), 1,
      aux_sym__literal_repeat1,
    STATE(3151), 1,
      sym__expression,
    ACTIONS(103), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(105), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1073), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1083), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(1089), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3053), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2500), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [109248] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6133), 1,
      aux_sym_concatenation_token1,
    ACTIONS(6234), 1,
      sym__concat,
    STATE(2127), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1286), 15,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1288), 25,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [109305] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5826), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5828), 1,
      sym_file_descriptor,
    ACTIONS(6236), 1,
      anon_sym_RPAREN,
    ACTIONS(5831), 3,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(5821), 9,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
    ACTIONS(5823), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
    ACTIONS(5819), 17,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [109366] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1131), 1,
      anon_sym_LPAREN,
    ACTIONS(1141), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1145), 1,
      anon_sym_DOLLAR,
    ACTIONS(1149), 1,
      anon_sym_DQUOTE,
    ACTIONS(1153), 1,
      aux_sym_number_token1,
    ACTIONS(1155), 1,
      aux_sym_number_token2,
    ACTIONS(1157), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1159), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1163), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1169), 1,
      sym__brace_start,
    ACTIONS(1171), 1,
      sym_word,
    ACTIONS(1173), 1,
      anon_sym_BANG,
    ACTIONS(1179), 1,
      anon_sym_TILDE,
    ACTIONS(1181), 1,
      sym__special_character,
    ACTIONS(1185), 1,
      sym_test_operator,
    ACTIONS(3060), 1,
      anon_sym_BQUOTE,
    STATE(2484), 1,
      aux_sym__literal_repeat1,
    STATE(3108), 1,
      sym__expression,
    ACTIONS(1129), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1165), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1175), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1177), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1183), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2594), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2456), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [109463] = 6,
    ACTIONS(71), 1,
      sym_comment,
    STATE(3545), 1,
      aux_sym__literal_repeat1,
    STATE(3556), 1,
      sym_concatenation,
    STATE(3471), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
    ACTIONS(2494), 10,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(2496), 22,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [109520] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(107), 1,
      anon_sym_TILDE,
    ACTIONS(237), 1,
      sym_word,
    ACTIONS(248), 1,
      anon_sym_LPAREN,
    ACTIONS(250), 1,
      anon_sym_BANG,
    ACTIONS(258), 1,
      anon_sym_DOLLAR,
    ACTIONS(264), 1,
      aux_sym_number_token1,
    ACTIONS(266), 1,
      aux_sym_number_token2,
    ACTIONS(270), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(282), 1,
      sym_test_operator,
    ACTIONS(284), 1,
      sym__brace_start,
    ACTIONS(1075), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1079), 1,
      sym__special_character,
    ACTIONS(1081), 1,
      anon_sym_DQUOTE,
    ACTIONS(1085), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1087), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3598), 1,
      anon_sym_BQUOTE,
    STATE(2690), 1,
      aux_sym__literal_repeat1,
    STATE(3111), 1,
      sym__expression,
    ACTIONS(103), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(105), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1073), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1083), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(1089), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3053), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2500), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [109617] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5826), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5828), 1,
      sym_file_descriptor,
    ACTIONS(6239), 1,
      anon_sym_RPAREN,
    ACTIONS(5831), 3,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(5821), 9,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
    ACTIONS(5823), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
    ACTIONS(5819), 17,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [109678] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5826), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5828), 1,
      sym_file_descriptor,
    ACTIONS(6242), 1,
      anon_sym_RPAREN,
    ACTIONS(5831), 3,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(5821), 9,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
    ACTIONS(5823), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
    ACTIONS(5819), 17,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [109739] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5826), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5828), 1,
      sym_file_descriptor,
    ACTIONS(6245), 1,
      anon_sym_RPAREN,
    ACTIONS(5831), 3,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(5821), 9,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
    ACTIONS(5823), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
    ACTIONS(5819), 17,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [109800] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5826), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5828), 1,
      sym_file_descriptor,
    ACTIONS(6248), 1,
      anon_sym_RPAREN,
    ACTIONS(5831), 3,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(5821), 9,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
    ACTIONS(5823), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
    ACTIONS(5819), 17,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [109861] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5826), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5828), 1,
      sym_file_descriptor,
    ACTIONS(6251), 1,
      anon_sym_RPAREN,
    ACTIONS(5831), 3,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(5821), 9,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
    ACTIONS(5823), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
    ACTIONS(5819), 17,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [109922] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1131), 1,
      anon_sym_LPAREN,
    ACTIONS(1141), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1145), 1,
      anon_sym_DOLLAR,
    ACTIONS(1149), 1,
      anon_sym_DQUOTE,
    ACTIONS(1153), 1,
      aux_sym_number_token1,
    ACTIONS(1155), 1,
      aux_sym_number_token2,
    ACTIONS(1157), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1159), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1163), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1169), 1,
      sym__brace_start,
    ACTIONS(1187), 1,
      sym_word,
    ACTIONS(1189), 1,
      anon_sym_BANG,
    ACTIONS(1195), 1,
      anon_sym_TILDE,
    ACTIONS(1201), 1,
      sym_test_operator,
    ACTIONS(3060), 1,
      anon_sym_BQUOTE,
    ACTIONS(5779), 1,
      sym__special_character,
    STATE(2717), 1,
      aux_sym__literal_repeat1,
    STATE(3050), 1,
      sym__expression,
    ACTIONS(1129), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1165), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1191), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1193), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1199), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2594), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2503), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [110019] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5826), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5828), 1,
      sym_file_descriptor,
    ACTIONS(6254), 1,
      anon_sym_RPAREN,
    ACTIONS(5831), 3,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(5821), 9,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
    ACTIONS(5823), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
    ACTIONS(5819), 17,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [110080] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(6257), 1,
      sym__special_character,
    STATE(2118), 1,
      aux_sym__literal_repeat1,
    ACTIONS(5395), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5393), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [110135] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5826), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5828), 1,
      sym_file_descriptor,
    ACTIONS(6259), 1,
      anon_sym_RPAREN,
    ACTIONS(5831), 3,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(5821), 9,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
    ACTIONS(5823), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
    ACTIONS(5819), 17,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [110196] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5826), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5828), 1,
      sym_file_descriptor,
    ACTIONS(6262), 1,
      anon_sym_RPAREN,
    ACTIONS(5831), 3,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(5821), 9,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
    ACTIONS(5823), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
    ACTIONS(5819), 17,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [110257] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5826), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5828), 1,
      sym_file_descriptor,
    ACTIONS(6265), 1,
      anon_sym_RPAREN,
    ACTIONS(5831), 3,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(5821), 9,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
    ACTIONS(5823), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
    ACTIONS(5819), 17,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [110318] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5826), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5828), 1,
      sym_file_descriptor,
    ACTIONS(6268), 1,
      anon_sym_RPAREN,
    ACTIONS(5831), 3,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(5821), 9,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
    ACTIONS(5823), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
    ACTIONS(5819), 17,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [110379] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6133), 1,
      aux_sym_concatenation_token1,
    ACTIONS(6271), 1,
      sym__concat,
    STATE(2127), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1265), 15,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1267), 25,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [110436] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5826), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5828), 1,
      sym_file_descriptor,
    ACTIONS(6273), 1,
      anon_sym_RPAREN,
    ACTIONS(5831), 3,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(5821), 9,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
    ACTIONS(5823), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
    ACTIONS(5819), 17,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [110497] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(2135), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(6133), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(4370), 15,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(4372), 25,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [110552] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(2152), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(6133), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(4467), 15,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(4469), 25,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [110607] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5828), 1,
      sym_file_descriptor,
    ACTIONS(5826), 2,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5831), 3,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(5821), 9,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
    ACTIONS(5823), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
    ACTIONS(5819), 17,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [110666] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5826), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5828), 1,
      sym_file_descriptor,
    ACTIONS(6276), 1,
      anon_sym_RPAREN,
    ACTIONS(5831), 3,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(5821), 9,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
    ACTIONS(5823), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
    ACTIONS(5819), 17,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [110727] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5375), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5373), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [110778] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5826), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5828), 1,
      sym_file_descriptor,
    ACTIONS(6279), 1,
      anon_sym_RPAREN,
    ACTIONS(5831), 3,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(5821), 9,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
    ACTIONS(5823), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
    ACTIONS(5819), 17,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [110839] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1263), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1261), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [110890] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(6282), 1,
      sym__special_character,
    STATE(2124), 1,
      aux_sym__literal_repeat1,
    ACTIONS(5395), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5393), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [110945] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1326), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1324), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [110996] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5826), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5828), 1,
      sym_file_descriptor,
    ACTIONS(6284), 1,
      anon_sym_RPAREN,
    ACTIONS(5831), 3,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(5821), 9,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
    ACTIONS(5823), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
    ACTIONS(5819), 17,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [111057] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5826), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5828), 1,
      sym_file_descriptor,
    ACTIONS(6287), 1,
      anon_sym_RPAREN,
    ACTIONS(5831), 3,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(5821), 9,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
    ACTIONS(5823), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
    ACTIONS(5819), 17,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [111118] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5826), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5828), 1,
      sym_file_descriptor,
    ACTIONS(6290), 1,
      anon_sym_RPAREN,
    ACTIONS(5831), 3,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(5821), 9,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
    ACTIONS(5823), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
    ACTIONS(5819), 17,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [111179] = 27,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(363), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(367), 1,
      anon_sym_DOLLAR,
    ACTIONS(371), 1,
      anon_sym_DQUOTE,
    ACTIONS(375), 1,
      aux_sym_number_token1,
    ACTIONS(377), 1,
      aux_sym_number_token2,
    ACTIONS(379), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(381), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(385), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(391), 1,
      sym__brace_start,
    ACTIONS(1093), 1,
      anon_sym_LPAREN,
    ACTIONS(1095), 1,
      anon_sym_BANG,
    ACTIONS(1103), 1,
      anon_sym_TILDE,
    ACTIONS(5809), 1,
      anon_sym_BQUOTE,
    ACTIONS(5916), 1,
      sym__special_character,
    ACTIONS(6293), 1,
      sym_word,
    ACTIONS(6297), 1,
      sym_test_operator,
    STATE(2501), 1,
      aux_sym__literal_repeat1,
    STATE(3397), 1,
      sym__expression,
    ACTIONS(352), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(387), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1099), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1101), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(6295), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2594), 2,
      sym_ternary_expression,
      sym_postfix_expression,
    STATE(2833), 4,
      sym_binary_expression,
      sym_unary_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2497), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [111278] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5826), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5828), 1,
      sym_file_descriptor,
    ACTIONS(6299), 1,
      anon_sym_RPAREN,
    ACTIONS(5831), 3,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(5821), 9,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
    ACTIONS(5823), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
    ACTIONS(5819), 17,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [111339] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5826), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5828), 1,
      sym_file_descriptor,
    ACTIONS(6302), 1,
      anon_sym_RPAREN,
    ACTIONS(5831), 3,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(5821), 9,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
    ACTIONS(5823), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
    ACTIONS(5819), 17,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [111400] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5826), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5828), 1,
      sym_file_descriptor,
    ACTIONS(6305), 1,
      anon_sym_RPAREN,
    ACTIONS(5831), 3,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(5821), 9,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
    ACTIONS(5823), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
    ACTIONS(5819), 17,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [111461] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5826), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5828), 1,
      sym_file_descriptor,
    ACTIONS(6308), 1,
      anon_sym_RPAREN,
    ACTIONS(5831), 3,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(5821), 9,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
    ACTIONS(5823), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
    ACTIONS(5819), 17,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [111522] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5826), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5828), 1,
      sym_file_descriptor,
    ACTIONS(6311), 1,
      anon_sym_RPAREN,
    ACTIONS(5831), 3,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(5821), 9,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
    ACTIONS(5823), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
    ACTIONS(5819), 17,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [111583] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5826), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5828), 1,
      sym_file_descriptor,
    ACTIONS(6314), 1,
      anon_sym_RPAREN,
    ACTIONS(5831), 3,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(5821), 9,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
    ACTIONS(5823), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
    ACTIONS(5819), 17,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [111644] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5768), 6,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5766), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [111695] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(107), 1,
      anon_sym_TILDE,
    ACTIONS(237), 1,
      sym_word,
    ACTIONS(248), 1,
      anon_sym_LPAREN,
    ACTIONS(250), 1,
      anon_sym_BANG,
    ACTIONS(258), 1,
      anon_sym_DOLLAR,
    ACTIONS(264), 1,
      aux_sym_number_token1,
    ACTIONS(266), 1,
      aux_sym_number_token2,
    ACTIONS(270), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(282), 1,
      sym_test_operator,
    ACTIONS(284), 1,
      sym__brace_start,
    ACTIONS(1075), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1079), 1,
      sym__special_character,
    ACTIONS(1081), 1,
      anon_sym_DQUOTE,
    ACTIONS(1085), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1087), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3598), 1,
      anon_sym_BQUOTE,
    STATE(2690), 1,
      aux_sym__literal_repeat1,
    STATE(3097), 1,
      sym__expression,
    ACTIONS(103), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(105), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1073), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1083), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(1089), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3053), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2500), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [111792] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5826), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5828), 1,
      sym_file_descriptor,
    ACTIONS(6317), 1,
      anon_sym_RPAREN,
    ACTIONS(5831), 3,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(5821), 9,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
    ACTIONS(5823), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
    ACTIONS(5819), 17,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [111853] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5826), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5828), 1,
      sym_file_descriptor,
    ACTIONS(6320), 1,
      anon_sym_RPAREN,
    ACTIONS(5831), 3,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(5821), 9,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
    ACTIONS(5823), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
    ACTIONS(5819), 17,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [111914] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5826), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5828), 1,
      sym_file_descriptor,
    ACTIONS(6323), 1,
      anon_sym_RPAREN,
    ACTIONS(5831), 3,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(5821), 9,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
    ACTIONS(5823), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
    ACTIONS(5819), 17,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [111975] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5826), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5828), 1,
      sym_file_descriptor,
    ACTIONS(6326), 1,
      anon_sym_RPAREN,
    ACTIONS(5831), 3,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(5821), 9,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
    ACTIONS(5823), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
    ACTIONS(5819), 17,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [112036] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5826), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5828), 1,
      sym_file_descriptor,
    ACTIONS(6329), 1,
      anon_sym_RPAREN,
    ACTIONS(5831), 3,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(5821), 9,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
    ACTIONS(5823), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
    ACTIONS(5819), 17,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [112097] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5826), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5828), 1,
      sym_file_descriptor,
    ACTIONS(6332), 1,
      anon_sym_RPAREN,
    ACTIONS(5831), 3,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(5821), 9,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
    ACTIONS(5823), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
    ACTIONS(5819), 17,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [112158] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5826), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5828), 1,
      sym_file_descriptor,
    ACTIONS(6335), 1,
      anon_sym_RPAREN,
    ACTIONS(5831), 3,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(5821), 9,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
    ACTIONS(5823), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
    ACTIONS(5819), 17,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [112219] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1263), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1261), 39,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [112270] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5826), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5828), 1,
      sym_file_descriptor,
    ACTIONS(6338), 1,
      anon_sym_RPAREN,
    ACTIONS(5831), 3,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(5821), 9,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
    ACTIONS(5823), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
    ACTIONS(5819), 17,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [112331] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1131), 1,
      anon_sym_LPAREN,
    ACTIONS(1141), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1145), 1,
      anon_sym_DOLLAR,
    ACTIONS(1149), 1,
      anon_sym_DQUOTE,
    ACTIONS(1153), 1,
      aux_sym_number_token1,
    ACTIONS(1155), 1,
      aux_sym_number_token2,
    ACTIONS(1157), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1159), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1163), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1169), 1,
      sym__brace_start,
    ACTIONS(1211), 1,
      sym_word,
    ACTIONS(1213), 1,
      anon_sym_BANG,
    ACTIONS(1219), 1,
      anon_sym_TILDE,
    ACTIONS(1221), 1,
      sym__special_character,
    ACTIONS(1225), 1,
      sym_test_operator,
    ACTIONS(3060), 1,
      anon_sym_BQUOTE,
    STATE(2484), 1,
      aux_sym__literal_repeat1,
    STATE(2621), 1,
      sym__expression,
    ACTIONS(1129), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1165), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1215), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1217), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1223), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2594), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2588), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [112428] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5826), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5828), 1,
      sym_file_descriptor,
    ACTIONS(6341), 1,
      anon_sym_RPAREN,
    ACTIONS(5831), 3,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(5821), 9,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
    ACTIONS(5823), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
    ACTIONS(5819), 17,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [112489] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5826), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5828), 1,
      sym_file_descriptor,
    ACTIONS(6344), 1,
      anon_sym_RPAREN,
    ACTIONS(5831), 3,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(5821), 9,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
    ACTIONS(5823), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
    ACTIONS(5819), 17,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [112550] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5826), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5828), 1,
      sym_file_descriptor,
    ACTIONS(6347), 1,
      anon_sym_RPAREN,
    ACTIONS(5831), 3,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(5821), 9,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
    ACTIONS(5823), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
    ACTIONS(5819), 17,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [112611] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1131), 1,
      anon_sym_LPAREN,
    ACTIONS(1141), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1145), 1,
      anon_sym_DOLLAR,
    ACTIONS(1149), 1,
      anon_sym_DQUOTE,
    ACTIONS(1153), 1,
      aux_sym_number_token1,
    ACTIONS(1155), 1,
      aux_sym_number_token2,
    ACTIONS(1157), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1159), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1163), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1169), 1,
      sym__brace_start,
    ACTIONS(1211), 1,
      sym_word,
    ACTIONS(1213), 1,
      anon_sym_BANG,
    ACTIONS(1219), 1,
      anon_sym_TILDE,
    ACTIONS(1221), 1,
      sym__special_character,
    ACTIONS(1225), 1,
      sym_test_operator,
    ACTIONS(3060), 1,
      anon_sym_BQUOTE,
    STATE(2484), 1,
      aux_sym__literal_repeat1,
    STATE(3183), 1,
      sym__expression,
    ACTIONS(1129), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1165), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1215), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1217), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1223), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2594), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2588), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [112708] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5826), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5828), 1,
      sym_file_descriptor,
    ACTIONS(6350), 1,
      anon_sym_RPAREN,
    ACTIONS(5831), 3,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(5821), 9,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
    ACTIONS(5823), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
    ACTIONS(5819), 17,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [112769] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(6353), 1,
      sym__special_character,
    STATE(2254), 1,
      aux_sym__literal_repeat1,
    ACTIONS(5478), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5476), 36,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [112824] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5826), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5828), 1,
      sym_file_descriptor,
    ACTIONS(6355), 1,
      anon_sym_RPAREN,
    ACTIONS(5831), 3,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(5821), 9,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
    ACTIONS(5823), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
    ACTIONS(5819), 17,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [112885] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1131), 1,
      anon_sym_LPAREN,
    ACTIONS(1141), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1145), 1,
      anon_sym_DOLLAR,
    ACTIONS(1149), 1,
      anon_sym_DQUOTE,
    ACTIONS(1153), 1,
      aux_sym_number_token1,
    ACTIONS(1155), 1,
      aux_sym_number_token2,
    ACTIONS(1157), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1159), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1163), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1169), 1,
      sym__brace_start,
    ACTIONS(1187), 1,
      sym_word,
    ACTIONS(1189), 1,
      anon_sym_BANG,
    ACTIONS(1195), 1,
      anon_sym_TILDE,
    ACTIONS(1201), 1,
      sym_test_operator,
    ACTIONS(3060), 1,
      anon_sym_BQUOTE,
    ACTIONS(5779), 1,
      sym__special_character,
    STATE(2717), 1,
      aux_sym__literal_repeat1,
    STATE(2995), 1,
      sym__expression,
    ACTIONS(1129), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1165), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1191), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1193), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1199), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2594), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2503), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [112982] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5826), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5828), 1,
      sym_file_descriptor,
    ACTIONS(6358), 1,
      anon_sym_RPAREN,
    ACTIONS(5831), 3,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(5821), 9,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
    ACTIONS(5823), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
    ACTIONS(5819), 17,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [113043] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5826), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5828), 1,
      sym_file_descriptor,
    ACTIONS(6361), 1,
      anon_sym_RPAREN,
    ACTIONS(5831), 3,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(5821), 9,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
    ACTIONS(5823), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
    ACTIONS(5819), 17,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [113104] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5826), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5828), 1,
      sym_file_descriptor,
    ACTIONS(6364), 1,
      anon_sym_RPAREN,
    ACTIONS(5831), 3,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(5821), 9,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
    ACTIONS(5823), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
    ACTIONS(5819), 17,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [113165] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5826), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5828), 1,
      sym_file_descriptor,
    ACTIONS(6367), 1,
      anon_sym_RPAREN,
    ACTIONS(5831), 3,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(5821), 9,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
    ACTIONS(5823), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
    ACTIONS(5819), 17,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [113226] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1131), 1,
      anon_sym_LPAREN,
    ACTIONS(1141), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1145), 1,
      anon_sym_DOLLAR,
    ACTIONS(1149), 1,
      anon_sym_DQUOTE,
    ACTIONS(1153), 1,
      aux_sym_number_token1,
    ACTIONS(1155), 1,
      aux_sym_number_token2,
    ACTIONS(1157), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1159), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1163), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1169), 1,
      sym__brace_start,
    ACTIONS(1187), 1,
      sym_word,
    ACTIONS(1189), 1,
      anon_sym_BANG,
    ACTIONS(1195), 1,
      anon_sym_TILDE,
    ACTIONS(1201), 1,
      sym_test_operator,
    ACTIONS(3060), 1,
      anon_sym_BQUOTE,
    ACTIONS(5779), 1,
      sym__special_character,
    STATE(2717), 1,
      aux_sym__literal_repeat1,
    STATE(3003), 1,
      sym__expression,
    ACTIONS(1129), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1165), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1191), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1193), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1199), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2594), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2503), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [113323] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1131), 1,
      anon_sym_LPAREN,
    ACTIONS(1141), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1145), 1,
      anon_sym_DOLLAR,
    ACTIONS(1149), 1,
      anon_sym_DQUOTE,
    ACTIONS(1153), 1,
      aux_sym_number_token1,
    ACTIONS(1155), 1,
      aux_sym_number_token2,
    ACTIONS(1157), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1159), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1163), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1169), 1,
      sym__brace_start,
    ACTIONS(1187), 1,
      sym_word,
    ACTIONS(1189), 1,
      anon_sym_BANG,
    ACTIONS(1195), 1,
      anon_sym_TILDE,
    ACTIONS(1201), 1,
      sym_test_operator,
    ACTIONS(3060), 1,
      anon_sym_BQUOTE,
    ACTIONS(5779), 1,
      sym__special_character,
    STATE(2717), 1,
      aux_sym__literal_repeat1,
    STATE(3043), 1,
      sym__expression,
    ACTIONS(1129), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1165), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1191), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1193), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1199), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2594), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2503), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [113420] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(2135), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(6133), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(4463), 15,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(4465), 25,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [113475] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5826), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5828), 1,
      sym_file_descriptor,
    ACTIONS(6370), 1,
      anon_sym_RPAREN,
    ACTIONS(5831), 3,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(5821), 9,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
    ACTIONS(5823), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
    ACTIONS(5819), 17,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [113536] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2096), 6,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(2094), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [113587] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(107), 1,
      anon_sym_TILDE,
    ACTIONS(237), 1,
      sym_word,
    ACTIONS(248), 1,
      anon_sym_LPAREN,
    ACTIONS(250), 1,
      anon_sym_BANG,
    ACTIONS(258), 1,
      anon_sym_DOLLAR,
    ACTIONS(264), 1,
      aux_sym_number_token1,
    ACTIONS(266), 1,
      aux_sym_number_token2,
    ACTIONS(270), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(282), 1,
      sym_test_operator,
    ACTIONS(284), 1,
      sym__brace_start,
    ACTIONS(1075), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1079), 1,
      sym__special_character,
    ACTIONS(1081), 1,
      anon_sym_DQUOTE,
    ACTIONS(1085), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1087), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3598), 1,
      anon_sym_BQUOTE,
    STATE(2690), 1,
      aux_sym__literal_repeat1,
    STATE(3126), 1,
      sym__expression,
    ACTIONS(103), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(105), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1073), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1083), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(1089), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3053), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2500), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [113684] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(2152), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(6133), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(4546), 15,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(4548), 25,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [113739] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(2152), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(6133), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(4480), 15,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(4482), 25,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [113794] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(107), 1,
      anon_sym_TILDE,
    ACTIONS(237), 1,
      sym_word,
    ACTIONS(248), 1,
      anon_sym_LPAREN,
    ACTIONS(250), 1,
      anon_sym_BANG,
    ACTIONS(258), 1,
      anon_sym_DOLLAR,
    ACTIONS(264), 1,
      aux_sym_number_token1,
    ACTIONS(266), 1,
      aux_sym_number_token2,
    ACTIONS(270), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(282), 1,
      sym_test_operator,
    ACTIONS(284), 1,
      sym__brace_start,
    ACTIONS(1075), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1079), 1,
      sym__special_character,
    ACTIONS(1081), 1,
      anon_sym_DQUOTE,
    ACTIONS(1085), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1087), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3598), 1,
      anon_sym_BQUOTE,
    STATE(2690), 1,
      aux_sym__literal_repeat1,
    STATE(3080), 1,
      sym__expression,
    ACTIONS(103), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(105), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1073), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1083), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(1089), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3053), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2500), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [113891] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5826), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5828), 1,
      sym_file_descriptor,
    ACTIONS(6373), 1,
      anon_sym_RPAREN,
    ACTIONS(5831), 3,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(5821), 9,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
    ACTIONS(5823), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
    ACTIONS(5819), 17,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [113952] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5826), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5828), 1,
      sym_file_descriptor,
    ACTIONS(6376), 1,
      anon_sym_RPAREN,
    ACTIONS(5831), 3,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(5821), 9,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
    ACTIONS(5823), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
    ACTIONS(5819), 17,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [114013] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(107), 1,
      anon_sym_TILDE,
    ACTIONS(237), 1,
      sym_word,
    ACTIONS(248), 1,
      anon_sym_LPAREN,
    ACTIONS(250), 1,
      anon_sym_BANG,
    ACTIONS(258), 1,
      anon_sym_DOLLAR,
    ACTIONS(264), 1,
      aux_sym_number_token1,
    ACTIONS(266), 1,
      aux_sym_number_token2,
    ACTIONS(270), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(282), 1,
      sym_test_operator,
    ACTIONS(284), 1,
      sym__brace_start,
    ACTIONS(1075), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1079), 1,
      sym__special_character,
    ACTIONS(1081), 1,
      anon_sym_DQUOTE,
    ACTIONS(1085), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1087), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3598), 1,
      anon_sym_BQUOTE,
    STATE(2690), 1,
      aux_sym__literal_repeat1,
    STATE(3090), 1,
      sym__expression,
    ACTIONS(103), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(105), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1073), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1083), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(1089), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3053), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2500), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [114110] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(107), 1,
      anon_sym_TILDE,
    ACTIONS(237), 1,
      sym_word,
    ACTIONS(248), 1,
      anon_sym_LPAREN,
    ACTIONS(250), 1,
      anon_sym_BANG,
    ACTIONS(258), 1,
      anon_sym_DOLLAR,
    ACTIONS(264), 1,
      aux_sym_number_token1,
    ACTIONS(266), 1,
      aux_sym_number_token2,
    ACTIONS(270), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(282), 1,
      sym_test_operator,
    ACTIONS(284), 1,
      sym__brace_start,
    ACTIONS(1075), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1079), 1,
      sym__special_character,
    ACTIONS(1081), 1,
      anon_sym_DQUOTE,
    ACTIONS(1085), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1087), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3598), 1,
      anon_sym_BQUOTE,
    STATE(2690), 1,
      aux_sym__literal_repeat1,
    STATE(3088), 1,
      sym__expression,
    ACTIONS(103), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(105), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1073), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1083), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(1089), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3053), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2500), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [114207] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1253), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1251), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [114258] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(2135), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(6133), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(4530), 15,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(4532), 25,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [114313] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5826), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5828), 1,
      sym_file_descriptor,
    ACTIONS(6379), 1,
      anon_sym_RPAREN,
    ACTIONS(5831), 3,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(5821), 9,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
    ACTIONS(5823), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
    ACTIONS(5819), 17,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [114374] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1131), 1,
      anon_sym_LPAREN,
    ACTIONS(1141), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1145), 1,
      anon_sym_DOLLAR,
    ACTIONS(1149), 1,
      anon_sym_DQUOTE,
    ACTIONS(1153), 1,
      aux_sym_number_token1,
    ACTIONS(1155), 1,
      aux_sym_number_token2,
    ACTIONS(1157), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1159), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1163), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1169), 1,
      sym__brace_start,
    ACTIONS(1171), 1,
      sym_word,
    ACTIONS(1173), 1,
      anon_sym_BANG,
    ACTIONS(1179), 1,
      anon_sym_TILDE,
    ACTIONS(1181), 1,
      sym__special_character,
    ACTIONS(1185), 1,
      sym_test_operator,
    ACTIONS(3060), 1,
      anon_sym_BQUOTE,
    STATE(2484), 1,
      aux_sym__literal_repeat1,
    STATE(3060), 1,
      sym__expression,
    ACTIONS(1129), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1165), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1175), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1177), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1183), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2594), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2456), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [114471] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1310), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1308), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [114522] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1263), 6,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1261), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [114573] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1306), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1304), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [114624] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5826), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5828), 1,
      sym_file_descriptor,
    ACTIONS(6382), 1,
      anon_sym_RPAREN,
    ACTIONS(5831), 3,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(5821), 9,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
    ACTIONS(5823), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
    ACTIONS(5819), 17,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [114685] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1310), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1308), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [114736] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5826), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5828), 1,
      sym_file_descriptor,
    ACTIONS(6385), 1,
      anon_sym_RPAREN,
    ACTIONS(5831), 3,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(5821), 9,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
    ACTIONS(5823), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
    ACTIONS(5819), 17,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [114797] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5826), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5828), 1,
      sym_file_descriptor,
    ACTIONS(6388), 1,
      anon_sym_RPAREN,
    ACTIONS(5831), 3,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(5821), 9,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
    ACTIONS(5823), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
    ACTIONS(5819), 17,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [114858] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5826), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5828), 1,
      sym_file_descriptor,
    ACTIONS(6391), 1,
      anon_sym_RPAREN,
    ACTIONS(5831), 3,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(5821), 9,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
    ACTIONS(5823), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
    ACTIONS(5819), 17,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [114919] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1298), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1296), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [114970] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5826), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5828), 1,
      sym_file_descriptor,
    ACTIONS(6394), 1,
      anon_sym_RPAREN,
    ACTIONS(5831), 3,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(5821), 9,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
    ACTIONS(5823), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
    ACTIONS(5819), 17,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [115031] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1302), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1300), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [115082] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1131), 1,
      anon_sym_LPAREN,
    ACTIONS(1141), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1145), 1,
      anon_sym_DOLLAR,
    ACTIONS(1149), 1,
      anon_sym_DQUOTE,
    ACTIONS(1153), 1,
      aux_sym_number_token1,
    ACTIONS(1155), 1,
      aux_sym_number_token2,
    ACTIONS(1157), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1159), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1163), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1169), 1,
      sym__brace_start,
    ACTIONS(1187), 1,
      sym_word,
    ACTIONS(1189), 1,
      anon_sym_BANG,
    ACTIONS(1195), 1,
      anon_sym_TILDE,
    ACTIONS(1201), 1,
      sym_test_operator,
    ACTIONS(3060), 1,
      anon_sym_BQUOTE,
    ACTIONS(5779), 1,
      sym__special_character,
    STATE(2717), 1,
      aux_sym__literal_repeat1,
    STATE(3014), 1,
      sym__expression,
    ACTIONS(1129), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1165), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1191), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1193), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1199), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2594), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2503), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [115179] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1131), 1,
      anon_sym_LPAREN,
    ACTIONS(1141), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1145), 1,
      anon_sym_DOLLAR,
    ACTIONS(1149), 1,
      anon_sym_DQUOTE,
    ACTIONS(1153), 1,
      aux_sym_number_token1,
    ACTIONS(1155), 1,
      aux_sym_number_token2,
    ACTIONS(1157), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1159), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1163), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1169), 1,
      sym__brace_start,
    ACTIONS(1187), 1,
      sym_word,
    ACTIONS(1189), 1,
      anon_sym_BANG,
    ACTIONS(1195), 1,
      anon_sym_TILDE,
    ACTIONS(1201), 1,
      sym_test_operator,
    ACTIONS(3060), 1,
      anon_sym_BQUOTE,
    ACTIONS(5779), 1,
      sym__special_character,
    STATE(2717), 1,
      aux_sym__literal_repeat1,
    STATE(3013), 1,
      sym__expression,
    ACTIONS(1129), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1165), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1191), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1193), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1199), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2594), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2503), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [115276] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4372), 6,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4370), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [115327] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5826), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5828), 1,
      sym_file_descriptor,
    ACTIONS(6397), 1,
      anon_sym_RPAREN,
    ACTIONS(5831), 3,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(5821), 9,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
    ACTIONS(5823), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
    ACTIONS(5819), 17,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [115388] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1131), 1,
      anon_sym_LPAREN,
    ACTIONS(1141), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1145), 1,
      anon_sym_DOLLAR,
    ACTIONS(1149), 1,
      anon_sym_DQUOTE,
    ACTIONS(1153), 1,
      aux_sym_number_token1,
    ACTIONS(1155), 1,
      aux_sym_number_token2,
    ACTIONS(1157), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1159), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1163), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1169), 1,
      sym__brace_start,
    ACTIONS(1187), 1,
      sym_word,
    ACTIONS(1189), 1,
      anon_sym_BANG,
    ACTIONS(1195), 1,
      anon_sym_TILDE,
    ACTIONS(1201), 1,
      sym_test_operator,
    ACTIONS(3060), 1,
      anon_sym_BQUOTE,
    ACTIONS(5779), 1,
      sym__special_character,
    STATE(2717), 1,
      aux_sym__literal_repeat1,
    STATE(3010), 1,
      sym__expression,
    ACTIONS(1129), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1165), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1191), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1193), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1199), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2594), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2503), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [115485] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(107), 1,
      anon_sym_TILDE,
    ACTIONS(237), 1,
      sym_word,
    ACTIONS(248), 1,
      anon_sym_LPAREN,
    ACTIONS(250), 1,
      anon_sym_BANG,
    ACTIONS(258), 1,
      anon_sym_DOLLAR,
    ACTIONS(264), 1,
      aux_sym_number_token1,
    ACTIONS(266), 1,
      aux_sym_number_token2,
    ACTIONS(270), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(282), 1,
      sym_test_operator,
    ACTIONS(284), 1,
      sym__brace_start,
    ACTIONS(1075), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1079), 1,
      sym__special_character,
    ACTIONS(1081), 1,
      anon_sym_DQUOTE,
    ACTIONS(1085), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1087), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3598), 1,
      anon_sym_BQUOTE,
    STATE(2690), 1,
      aux_sym__literal_repeat1,
    STATE(3086), 1,
      sym__expression,
    ACTIONS(103), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(105), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1073), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1083), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(1089), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3053), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2500), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [115582] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(107), 1,
      anon_sym_TILDE,
    ACTIONS(237), 1,
      sym_word,
    ACTIONS(248), 1,
      anon_sym_LPAREN,
    ACTIONS(250), 1,
      anon_sym_BANG,
    ACTIONS(258), 1,
      anon_sym_DOLLAR,
    ACTIONS(264), 1,
      aux_sym_number_token1,
    ACTIONS(266), 1,
      aux_sym_number_token2,
    ACTIONS(270), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(282), 1,
      sym_test_operator,
    ACTIONS(284), 1,
      sym__brace_start,
    ACTIONS(1075), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1079), 1,
      sym__special_character,
    ACTIONS(1081), 1,
      anon_sym_DQUOTE,
    ACTIONS(1085), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1087), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3598), 1,
      anon_sym_BQUOTE,
    STATE(2690), 1,
      aux_sym__literal_repeat1,
    STATE(3084), 1,
      sym__expression,
    ACTIONS(103), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(105), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1073), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1083), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(1089), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3053), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2500), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [115679] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1131), 1,
      anon_sym_LPAREN,
    ACTIONS(1141), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1145), 1,
      anon_sym_DOLLAR,
    ACTIONS(1149), 1,
      anon_sym_DQUOTE,
    ACTIONS(1153), 1,
      aux_sym_number_token1,
    ACTIONS(1155), 1,
      aux_sym_number_token2,
    ACTIONS(1157), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1159), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1163), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1169), 1,
      sym__brace_start,
    ACTIONS(1187), 1,
      sym_word,
    ACTIONS(1189), 1,
      anon_sym_BANG,
    ACTIONS(1195), 1,
      anon_sym_TILDE,
    ACTIONS(1201), 1,
      sym_test_operator,
    ACTIONS(3060), 1,
      anon_sym_BQUOTE,
    ACTIONS(5779), 1,
      sym__special_character,
    STATE(2717), 1,
      aux_sym__literal_repeat1,
    STATE(3008), 1,
      sym__expression,
    ACTIONS(1129), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1165), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1191), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1193), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1199), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2594), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2503), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [115776] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1131), 1,
      anon_sym_LPAREN,
    ACTIONS(1141), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1145), 1,
      anon_sym_DOLLAR,
    ACTIONS(1149), 1,
      anon_sym_DQUOTE,
    ACTIONS(1153), 1,
      aux_sym_number_token1,
    ACTIONS(1155), 1,
      aux_sym_number_token2,
    ACTIONS(1157), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1159), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1163), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1169), 1,
      sym__brace_start,
    ACTIONS(1187), 1,
      sym_word,
    ACTIONS(1189), 1,
      anon_sym_BANG,
    ACTIONS(1195), 1,
      anon_sym_TILDE,
    ACTIONS(1201), 1,
      sym_test_operator,
    ACTIONS(3060), 1,
      anon_sym_BQUOTE,
    ACTIONS(5779), 1,
      sym__special_character,
    STATE(2717), 1,
      aux_sym__literal_repeat1,
    STATE(3007), 1,
      sym__expression,
    ACTIONS(1129), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1165), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1191), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1193), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1199), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2594), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2503), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [115873] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1127), 1,
      sym_word,
    ACTIONS(1131), 1,
      anon_sym_LPAREN,
    ACTIONS(1133), 1,
      anon_sym_BANG,
    ACTIONS(1139), 1,
      anon_sym_TILDE,
    ACTIONS(1141), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1145), 1,
      anon_sym_DOLLAR,
    ACTIONS(1147), 1,
      sym__special_character,
    ACTIONS(1149), 1,
      anon_sym_DQUOTE,
    ACTIONS(1153), 1,
      aux_sym_number_token1,
    ACTIONS(1155), 1,
      aux_sym_number_token2,
    ACTIONS(1157), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1159), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1163), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1167), 1,
      sym_test_operator,
    ACTIONS(1169), 1,
      sym__brace_start,
    ACTIONS(3060), 1,
      anon_sym_BQUOTE,
    STATE(2484), 1,
      aux_sym__literal_repeat1,
    STATE(3129), 1,
      sym__expression,
    ACTIONS(1129), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1135), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1137), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1151), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(1165), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(2594), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2573), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [115970] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1127), 1,
      sym_word,
    ACTIONS(1131), 1,
      anon_sym_LPAREN,
    ACTIONS(1133), 1,
      anon_sym_BANG,
    ACTIONS(1139), 1,
      anon_sym_TILDE,
    ACTIONS(1141), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1145), 1,
      anon_sym_DOLLAR,
    ACTIONS(1147), 1,
      sym__special_character,
    ACTIONS(1149), 1,
      anon_sym_DQUOTE,
    ACTIONS(1153), 1,
      aux_sym_number_token1,
    ACTIONS(1155), 1,
      aux_sym_number_token2,
    ACTIONS(1157), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1159), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1163), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1167), 1,
      sym_test_operator,
    ACTIONS(1169), 1,
      sym__brace_start,
    ACTIONS(3060), 1,
      anon_sym_BQUOTE,
    STATE(2484), 1,
      aux_sym__literal_repeat1,
    STATE(3148), 1,
      sym__expression,
    ACTIONS(1129), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1135), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1137), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1151), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(1165), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(2594), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2573), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [116067] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5826), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5828), 1,
      sym_file_descriptor,
    ACTIONS(6400), 1,
      anon_sym_RPAREN,
    ACTIONS(5831), 3,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(5821), 9,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
    ACTIONS(5823), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
    ACTIONS(5819), 17,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [116128] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1263), 6,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1261), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [116179] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1131), 1,
      anon_sym_LPAREN,
    ACTIONS(1141), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1145), 1,
      anon_sym_DOLLAR,
    ACTIONS(1149), 1,
      anon_sym_DQUOTE,
    ACTIONS(1153), 1,
      aux_sym_number_token1,
    ACTIONS(1155), 1,
      aux_sym_number_token2,
    ACTIONS(1157), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1159), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1163), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1169), 1,
      sym__brace_start,
    ACTIONS(1187), 1,
      sym_word,
    ACTIONS(1189), 1,
      anon_sym_BANG,
    ACTIONS(1195), 1,
      anon_sym_TILDE,
    ACTIONS(1201), 1,
      sym_test_operator,
    ACTIONS(3060), 1,
      anon_sym_BQUOTE,
    ACTIONS(5779), 1,
      sym__special_character,
    STATE(2717), 1,
      aux_sym__literal_repeat1,
    STATE(3002), 1,
      sym__expression,
    ACTIONS(1129), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1165), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1191), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1193), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1199), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2594), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2503), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [116276] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1131), 1,
      anon_sym_LPAREN,
    ACTIONS(1141), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1145), 1,
      anon_sym_DOLLAR,
    ACTIONS(1149), 1,
      anon_sym_DQUOTE,
    ACTIONS(1153), 1,
      aux_sym_number_token1,
    ACTIONS(1155), 1,
      aux_sym_number_token2,
    ACTIONS(1157), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1159), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1163), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1169), 1,
      sym__brace_start,
    ACTIONS(1187), 1,
      sym_word,
    ACTIONS(1189), 1,
      anon_sym_BANG,
    ACTIONS(1195), 1,
      anon_sym_TILDE,
    ACTIONS(1201), 1,
      sym_test_operator,
    ACTIONS(3060), 1,
      anon_sym_BQUOTE,
    ACTIONS(5779), 1,
      sym__special_character,
    STATE(2717), 1,
      aux_sym__literal_repeat1,
    STATE(3001), 1,
      sym__expression,
    ACTIONS(1129), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1165), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1191), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1193), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1199), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2594), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2503), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [116373] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(107), 1,
      anon_sym_TILDE,
    ACTIONS(237), 1,
      sym_word,
    ACTIONS(248), 1,
      anon_sym_LPAREN,
    ACTIONS(250), 1,
      anon_sym_BANG,
    ACTIONS(258), 1,
      anon_sym_DOLLAR,
    ACTIONS(264), 1,
      aux_sym_number_token1,
    ACTIONS(266), 1,
      aux_sym_number_token2,
    ACTIONS(270), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(282), 1,
      sym_test_operator,
    ACTIONS(284), 1,
      sym__brace_start,
    ACTIONS(1075), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1079), 1,
      sym__special_character,
    ACTIONS(1081), 1,
      anon_sym_DQUOTE,
    ACTIONS(1085), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1087), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3598), 1,
      anon_sym_BQUOTE,
    STATE(2690), 1,
      aux_sym__literal_repeat1,
    STATE(3083), 1,
      sym__expression,
    ACTIONS(103), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(105), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1073), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1083), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(1089), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3053), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2500), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [116470] = 27,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(363), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(367), 1,
      anon_sym_DOLLAR,
    ACTIONS(371), 1,
      anon_sym_DQUOTE,
    ACTIONS(375), 1,
      aux_sym_number_token1,
    ACTIONS(377), 1,
      aux_sym_number_token2,
    ACTIONS(379), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(381), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(385), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(391), 1,
      sym__brace_start,
    ACTIONS(1093), 1,
      anon_sym_LPAREN,
    ACTIONS(1095), 1,
      anon_sym_BANG,
    ACTIONS(1103), 1,
      anon_sym_TILDE,
    ACTIONS(5809), 1,
      anon_sym_BQUOTE,
    ACTIONS(5916), 1,
      sym__special_character,
    ACTIONS(6403), 1,
      sym_word,
    ACTIONS(6407), 1,
      sym_test_operator,
    STATE(2547), 1,
      aux_sym__literal_repeat1,
    STATE(3397), 1,
      sym__expression,
    ACTIONS(352), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(387), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1099), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1101), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(6405), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2594), 2,
      sym_ternary_expression,
      sym_postfix_expression,
    STATE(2815), 4,
      sym_binary_expression,
      sym_unary_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2546), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [116569] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(107), 1,
      anon_sym_TILDE,
    ACTIONS(237), 1,
      sym_word,
    ACTIONS(248), 1,
      anon_sym_LPAREN,
    ACTIONS(250), 1,
      anon_sym_BANG,
    ACTIONS(258), 1,
      anon_sym_DOLLAR,
    ACTIONS(264), 1,
      aux_sym_number_token1,
    ACTIONS(266), 1,
      aux_sym_number_token2,
    ACTIONS(270), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(282), 1,
      sym_test_operator,
    ACTIONS(284), 1,
      sym__brace_start,
    ACTIONS(1075), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1079), 1,
      sym__special_character,
    ACTIONS(1081), 1,
      anon_sym_DQUOTE,
    ACTIONS(1085), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1087), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3598), 1,
      anon_sym_BQUOTE,
    STATE(2690), 1,
      aux_sym__literal_repeat1,
    STATE(3081), 1,
      sym__expression,
    ACTIONS(103), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(105), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1073), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1083), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(1089), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3053), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2500), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [116666] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(107), 1,
      anon_sym_TILDE,
    ACTIONS(237), 1,
      sym_word,
    ACTIONS(248), 1,
      anon_sym_LPAREN,
    ACTIONS(250), 1,
      anon_sym_BANG,
    ACTIONS(258), 1,
      anon_sym_DOLLAR,
    ACTIONS(264), 1,
      aux_sym_number_token1,
    ACTIONS(266), 1,
      aux_sym_number_token2,
    ACTIONS(270), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(282), 1,
      sym_test_operator,
    ACTIONS(284), 1,
      sym__brace_start,
    ACTIONS(1075), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1079), 1,
      sym__special_character,
    ACTIONS(1081), 1,
      anon_sym_DQUOTE,
    ACTIONS(1085), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1087), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3598), 1,
      anon_sym_BQUOTE,
    STATE(2690), 1,
      aux_sym__literal_repeat1,
    STATE(3142), 1,
      sym__expression,
    ACTIONS(103), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(105), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1073), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1083), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(1089), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3053), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2500), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [116763] = 27,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(363), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(367), 1,
      anon_sym_DOLLAR,
    ACTIONS(371), 1,
      anon_sym_DQUOTE,
    ACTIONS(375), 1,
      aux_sym_number_token1,
    ACTIONS(377), 1,
      aux_sym_number_token2,
    ACTIONS(379), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(381), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(385), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(391), 1,
      sym__brace_start,
    ACTIONS(1093), 1,
      anon_sym_LPAREN,
    ACTIONS(1095), 1,
      anon_sym_BANG,
    ACTIONS(1103), 1,
      anon_sym_TILDE,
    ACTIONS(5809), 1,
      anon_sym_BQUOTE,
    ACTIONS(5916), 1,
      sym__special_character,
    ACTIONS(6409), 1,
      sym_word,
    ACTIONS(6413), 1,
      sym_test_operator,
    STATE(2552), 1,
      aux_sym__literal_repeat1,
    STATE(3397), 1,
      sym__expression,
    ACTIONS(352), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(387), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1099), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1101), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(6411), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2594), 2,
      sym_ternary_expression,
      sym_postfix_expression,
    STATE(2808), 4,
      sym_binary_expression,
      sym_unary_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2549), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [116862] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(107), 1,
      anon_sym_TILDE,
    ACTIONS(237), 1,
      sym_word,
    ACTIONS(248), 1,
      anon_sym_LPAREN,
    ACTIONS(250), 1,
      anon_sym_BANG,
    ACTIONS(258), 1,
      anon_sym_DOLLAR,
    ACTIONS(264), 1,
      aux_sym_number_token1,
    ACTIONS(266), 1,
      aux_sym_number_token2,
    ACTIONS(270), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(282), 1,
      sym_test_operator,
    ACTIONS(284), 1,
      sym__brace_start,
    ACTIONS(1075), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1079), 1,
      sym__special_character,
    ACTIONS(1081), 1,
      anon_sym_DQUOTE,
    ACTIONS(1085), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1087), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3598), 1,
      anon_sym_BQUOTE,
    STATE(2690), 1,
      aux_sym__literal_repeat1,
    STATE(3079), 1,
      sym__expression,
    ACTIONS(103), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(105), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1073), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1083), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(1089), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3053), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2500), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [116959] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(107), 1,
      anon_sym_TILDE,
    ACTIONS(237), 1,
      sym_word,
    ACTIONS(248), 1,
      anon_sym_LPAREN,
    ACTIONS(250), 1,
      anon_sym_BANG,
    ACTIONS(258), 1,
      anon_sym_DOLLAR,
    ACTIONS(264), 1,
      aux_sym_number_token1,
    ACTIONS(266), 1,
      aux_sym_number_token2,
    ACTIONS(270), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(282), 1,
      sym_test_operator,
    ACTIONS(284), 1,
      sym__brace_start,
    ACTIONS(1075), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1079), 1,
      sym__special_character,
    ACTIONS(1081), 1,
      anon_sym_DQUOTE,
    ACTIONS(1085), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1087), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3598), 1,
      anon_sym_BQUOTE,
    STATE(2690), 1,
      aux_sym__literal_repeat1,
    STATE(3078), 1,
      sym__expression,
    ACTIONS(103), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(105), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1073), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1083), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(1089), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3053), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2500), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [117056] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5772), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5770), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [117107] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1131), 1,
      anon_sym_LPAREN,
    ACTIONS(1141), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1145), 1,
      anon_sym_DOLLAR,
    ACTIONS(1149), 1,
      anon_sym_DQUOTE,
    ACTIONS(1153), 1,
      aux_sym_number_token1,
    ACTIONS(1155), 1,
      aux_sym_number_token2,
    ACTIONS(1157), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1159), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1163), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1169), 1,
      sym__brace_start,
    ACTIONS(1187), 1,
      sym_word,
    ACTIONS(1189), 1,
      anon_sym_BANG,
    ACTIONS(1195), 1,
      anon_sym_TILDE,
    ACTIONS(1201), 1,
      sym_test_operator,
    ACTIONS(3060), 1,
      anon_sym_BQUOTE,
    ACTIONS(5779), 1,
      sym__special_character,
    STATE(2717), 1,
      aux_sym__literal_repeat1,
    STATE(3000), 1,
      sym__expression,
    ACTIONS(1129), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1165), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1191), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1193), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1199), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2594), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2503), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [117204] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(107), 1,
      anon_sym_TILDE,
    ACTIONS(237), 1,
      sym_word,
    ACTIONS(248), 1,
      anon_sym_LPAREN,
    ACTIONS(250), 1,
      anon_sym_BANG,
    ACTIONS(258), 1,
      anon_sym_DOLLAR,
    ACTIONS(264), 1,
      aux_sym_number_token1,
    ACTIONS(266), 1,
      aux_sym_number_token2,
    ACTIONS(270), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(282), 1,
      sym_test_operator,
    ACTIONS(284), 1,
      sym__brace_start,
    ACTIONS(1075), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1079), 1,
      sym__special_character,
    ACTIONS(1081), 1,
      anon_sym_DQUOTE,
    ACTIONS(1085), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1087), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3598), 1,
      anon_sym_BQUOTE,
    STATE(2690), 1,
      aux_sym__literal_repeat1,
    STATE(3077), 1,
      sym__expression,
    ACTIONS(103), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(105), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1073), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1083), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(1089), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3053), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2500), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [117301] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1131), 1,
      anon_sym_LPAREN,
    ACTIONS(1141), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1145), 1,
      anon_sym_DOLLAR,
    ACTIONS(1149), 1,
      anon_sym_DQUOTE,
    ACTIONS(1153), 1,
      aux_sym_number_token1,
    ACTIONS(1155), 1,
      aux_sym_number_token2,
    ACTIONS(1157), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1159), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1163), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1169), 1,
      sym__brace_start,
    ACTIONS(1187), 1,
      sym_word,
    ACTIONS(1189), 1,
      anon_sym_BANG,
    ACTIONS(1195), 1,
      anon_sym_TILDE,
    ACTIONS(1201), 1,
      sym_test_operator,
    ACTIONS(3060), 1,
      anon_sym_BQUOTE,
    ACTIONS(5779), 1,
      sym__special_character,
    STATE(2717), 1,
      aux_sym__literal_repeat1,
    STATE(3034), 1,
      sym__expression,
    ACTIONS(1129), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1165), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1191), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1193), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1199), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2594), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2503), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [117398] = 27,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(363), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(367), 1,
      anon_sym_DOLLAR,
    ACTIONS(371), 1,
      anon_sym_DQUOTE,
    ACTIONS(375), 1,
      aux_sym_number_token1,
    ACTIONS(377), 1,
      aux_sym_number_token2,
    ACTIONS(379), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(381), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(385), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(391), 1,
      sym__brace_start,
    ACTIONS(1093), 1,
      anon_sym_LPAREN,
    ACTIONS(1095), 1,
      anon_sym_BANG,
    ACTIONS(1103), 1,
      anon_sym_TILDE,
    ACTIONS(5809), 1,
      anon_sym_BQUOTE,
    ACTIONS(5916), 1,
      sym__special_character,
    ACTIONS(6415), 1,
      sym_word,
    ACTIONS(6419), 1,
      sym_test_operator,
    STATE(2463), 1,
      aux_sym__literal_repeat1,
    STATE(3397), 1,
      sym__expression,
    ACTIONS(352), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(387), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1099), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1101), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(6417), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2594), 2,
      sym_ternary_expression,
      sym_postfix_expression,
    STATE(2851), 4,
      sym_binary_expression,
      sym_unary_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2455), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [117497] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1131), 1,
      anon_sym_LPAREN,
    ACTIONS(1141), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1145), 1,
      anon_sym_DOLLAR,
    ACTIONS(1149), 1,
      anon_sym_DQUOTE,
    ACTIONS(1153), 1,
      aux_sym_number_token1,
    ACTIONS(1155), 1,
      aux_sym_number_token2,
    ACTIONS(1157), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1159), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1163), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1169), 1,
      sym__brace_start,
    ACTIONS(1187), 1,
      sym_word,
    ACTIONS(1189), 1,
      anon_sym_BANG,
    ACTIONS(1195), 1,
      anon_sym_TILDE,
    ACTIONS(1201), 1,
      sym_test_operator,
    ACTIONS(3060), 1,
      anon_sym_BQUOTE,
    ACTIONS(5779), 1,
      sym__special_character,
    STATE(2717), 1,
      aux_sym__literal_repeat1,
    STATE(2996), 1,
      sym__expression,
    ACTIONS(1129), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1165), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1191), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1193), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1199), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2594), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2503), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [117594] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5768), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5766), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [117645] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(6421), 1,
      sym__special_character,
    STATE(2254), 1,
      aux_sym__literal_repeat1,
    ACTIONS(1364), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1362), 36,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [117700] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5826), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5828), 1,
      sym_file_descriptor,
    ACTIONS(6424), 1,
      anon_sym_RPAREN,
    ACTIONS(5831), 3,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(5821), 9,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
    ACTIONS(5823), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
    ACTIONS(5819), 17,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [117761] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(107), 1,
      anon_sym_TILDE,
    ACTIONS(237), 1,
      sym_word,
    ACTIONS(248), 1,
      anon_sym_LPAREN,
    ACTIONS(250), 1,
      anon_sym_BANG,
    ACTIONS(258), 1,
      anon_sym_DOLLAR,
    ACTIONS(264), 1,
      aux_sym_number_token1,
    ACTIONS(266), 1,
      aux_sym_number_token2,
    ACTIONS(270), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(282), 1,
      sym_test_operator,
    ACTIONS(284), 1,
      sym__brace_start,
    ACTIONS(1075), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1079), 1,
      sym__special_character,
    ACTIONS(1081), 1,
      anon_sym_DQUOTE,
    ACTIONS(1085), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1087), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3598), 1,
      anon_sym_BQUOTE,
    STATE(2690), 1,
      aux_sym__literal_repeat1,
    STATE(3095), 1,
      sym__expression,
    ACTIONS(103), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(105), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1073), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1083), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(1089), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3053), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2500), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [117858] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1330), 6,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1328), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [117909] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1131), 1,
      anon_sym_LPAREN,
    ACTIONS(1141), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1145), 1,
      anon_sym_DOLLAR,
    ACTIONS(1149), 1,
      anon_sym_DQUOTE,
    ACTIONS(1153), 1,
      aux_sym_number_token1,
    ACTIONS(1155), 1,
      aux_sym_number_token2,
    ACTIONS(1157), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1159), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1163), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1169), 1,
      sym__brace_start,
    ACTIONS(1187), 1,
      sym_word,
    ACTIONS(1189), 1,
      anon_sym_BANG,
    ACTIONS(1195), 1,
      anon_sym_TILDE,
    ACTIONS(1201), 1,
      sym_test_operator,
    ACTIONS(3060), 1,
      anon_sym_BQUOTE,
    ACTIONS(5779), 1,
      sym__special_character,
    STATE(2717), 1,
      aux_sym__literal_repeat1,
    STATE(2994), 1,
      sym__expression,
    ACTIONS(1129), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1165), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(1191), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1193), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1199), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2594), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2503), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [118006] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5826), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5828), 1,
      sym_file_descriptor,
    ACTIONS(6427), 1,
      anon_sym_RPAREN,
    ACTIONS(5831), 3,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(5821), 9,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
    ACTIONS(5823), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
    ACTIONS(5819), 17,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [118067] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1294), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1292), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [118118] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(107), 1,
      anon_sym_TILDE,
    ACTIONS(237), 1,
      sym_word,
    ACTIONS(248), 1,
      anon_sym_LPAREN,
    ACTIONS(250), 1,
      anon_sym_BANG,
    ACTIONS(258), 1,
      anon_sym_DOLLAR,
    ACTIONS(264), 1,
      aux_sym_number_token1,
    ACTIONS(266), 1,
      aux_sym_number_token2,
    ACTIONS(270), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(282), 1,
      sym_test_operator,
    ACTIONS(284), 1,
      sym__brace_start,
    ACTIONS(1075), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1079), 1,
      sym__special_character,
    ACTIONS(1081), 1,
      anon_sym_DQUOTE,
    ACTIONS(1085), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1087), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3598), 1,
      anon_sym_BQUOTE,
    STATE(2690), 1,
      aux_sym__literal_repeat1,
    STATE(3042), 1,
      sym__expression,
    ACTIONS(103), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(105), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1073), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1083), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(1089), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3053), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2500), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [118215] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5826), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5828), 1,
      sym_file_descriptor,
    ACTIONS(6430), 1,
      anon_sym_RPAREN,
    ACTIONS(5831), 3,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(5821), 9,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
    ACTIONS(5823), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
    ACTIONS(5819), 17,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [118276] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1334), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1332), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [118327] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5826), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5828), 1,
      sym_file_descriptor,
    ACTIONS(6433), 1,
      anon_sym_RPAREN,
    ACTIONS(5831), 3,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(5821), 9,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
    ACTIONS(5823), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
    ACTIONS(5819), 17,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [118388] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5826), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5828), 1,
      sym_file_descriptor,
    ACTIONS(6436), 1,
      anon_sym_RPAREN,
    ACTIONS(5831), 3,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(5821), 9,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
    ACTIONS(5823), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
    ACTIONS(5819), 17,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [118449] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1127), 1,
      sym_word,
    ACTIONS(1131), 1,
      anon_sym_LPAREN,
    ACTIONS(1133), 1,
      anon_sym_BANG,
    ACTIONS(1139), 1,
      anon_sym_TILDE,
    ACTIONS(1141), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1145), 1,
      anon_sym_DOLLAR,
    ACTIONS(1147), 1,
      sym__special_character,
    ACTIONS(1149), 1,
      anon_sym_DQUOTE,
    ACTIONS(1153), 1,
      aux_sym_number_token1,
    ACTIONS(1155), 1,
      aux_sym_number_token2,
    ACTIONS(1157), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1159), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1163), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1167), 1,
      sym_test_operator,
    ACTIONS(1169), 1,
      sym__brace_start,
    ACTIONS(3060), 1,
      anon_sym_BQUOTE,
    STATE(2484), 1,
      aux_sym__literal_repeat1,
    STATE(3143), 1,
      sym__expression,
    ACTIONS(1129), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1135), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1137), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1151), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(1165), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(2594), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2573), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [118546] = 26,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1127), 1,
      sym_word,
    ACTIONS(1131), 1,
      anon_sym_LPAREN,
    ACTIONS(1133), 1,
      anon_sym_BANG,
    ACTIONS(1139), 1,
      anon_sym_TILDE,
    ACTIONS(1141), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1145), 1,
      anon_sym_DOLLAR,
    ACTIONS(1147), 1,
      sym__special_character,
    ACTIONS(1149), 1,
      anon_sym_DQUOTE,
    ACTIONS(1153), 1,
      aux_sym_number_token1,
    ACTIONS(1155), 1,
      aux_sym_number_token2,
    ACTIONS(1157), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1159), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1163), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1167), 1,
      sym_test_operator,
    ACTIONS(1169), 1,
      sym__brace_start,
    ACTIONS(3060), 1,
      anon_sym_BQUOTE,
    STATE(2484), 1,
      aux_sym__literal_repeat1,
    STATE(3104), 1,
      sym__expression,
    ACTIONS(1129), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1135), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(1137), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    ACTIONS(1151), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(1165), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(2594), 6,
      sym_binary_expression,
      sym_ternary_expression,
      sym_unary_expression,
      sym_postfix_expression,
      sym_parenthesized_expression,
      sym_concatenation,
    STATE(2573), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [118643] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1316), 15,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1318), 27,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [118693] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1332), 15,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1334), 27,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [118743] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(6439), 1,
      sym__special_character,
    STATE(2311), 1,
      aux_sym__literal_repeat1,
    ACTIONS(5395), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5393), 36,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [118797] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(2293), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(6441), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(5389), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      sym__special_character,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(5391), 25,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [118851] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5826), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5828), 1,
      sym_file_descriptor,
    ACTIONS(5831), 3,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(5821), 9,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
    ACTIONS(5823), 12,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_BQUOTE,
    ACTIONS(5819), 16,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [118909] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5826), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5828), 1,
      sym_file_descriptor,
    ACTIONS(5831), 3,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(5821), 9,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
    ACTIONS(5823), 12,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_BQUOTE,
    ACTIONS(5819), 16,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [118967] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1320), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1322), 28,
      sym__concat,
      sym_test_operator,
      anon_sym_RPAREN_RPAREN,
      anon_sym_COMMA,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      anon_sym_COLON,
      aux_sym_concatenation_token1,
  [119017] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1352), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1354), 28,
      sym__concat,
      sym_test_operator,
      anon_sym_RPAREN_RPAREN,
      anon_sym_COMMA,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      anon_sym_COLON,
      aux_sym_concatenation_token1,
  [119067] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1328), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1330), 28,
      sym__concat,
      sym_test_operator,
      anon_sym_RPAREN_RPAREN,
      anon_sym_COMMA,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      anon_sym_COLON,
      aux_sym_concatenation_token1,
  [119117] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1336), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1338), 28,
      sym__concat,
      sym_test_operator,
      anon_sym_RPAREN_RPAREN,
      anon_sym_COMMA,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      anon_sym_COLON,
      aux_sym_concatenation_token1,
  [119167] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(2293), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(6441), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(4552), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      sym__special_character,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(4554), 25,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [119221] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5638), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5640), 1,
      sym__concat,
    STATE(2297), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5375), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5373), 34,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [119277] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1263), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1261), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [119327] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(6443), 1,
      sym__special_character,
    STATE(2316), 1,
      aux_sym__literal_repeat1,
    ACTIONS(5395), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5393), 35,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [119381] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5638), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5640), 1,
      sym__concat,
    STATE(2290), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5069), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5067), 34,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [119437] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5638), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5640), 1,
      sym__concat,
    STATE(2297), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4554), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4552), 34,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [119493] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(2285), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(6441), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(4564), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      sym__special_character,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(4566), 25,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [119547] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6441), 1,
      aux_sym_concatenation_token1,
    ACTIONS(6445), 1,
      sym__concat,
    STATE(2296), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1265), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      sym__special_character,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1267), 25,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [119603] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5932), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4261), 3,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(4247), 17,
      anon_sym_LPAREN_LPAREN,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
    ACTIONS(5930), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_BQUOTE,
  [119657] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1308), 15,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1310), 27,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [119707] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5638), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5640), 1,
      sym__concat,
    STATE(2290), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4566), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4564), 34,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [119763] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1292), 15,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1294), 27,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [119813] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5638), 1,
      aux_sym_concatenation_token1,
    ACTIONS(6447), 1,
      sym__concat,
    STATE(1363), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1267), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1265), 34,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [119869] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5638), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5640), 1,
      sym__concat,
    STATE(2297), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(2074), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(2072), 34,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [119925] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1316), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1318), 28,
      sym__concat,
      sym_test_operator,
      anon_sym_RPAREN_RPAREN,
      anon_sym_COMMA,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      anon_sym_COLON,
      aux_sym_concatenation_token1,
  [119975] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6441), 1,
      aux_sym_concatenation_token1,
    ACTIONS(6449), 1,
      sym__concat,
    STATE(2296), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1286), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      sym__special_character,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1288), 25,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [120031] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1300), 15,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1302), 27,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [120081] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1304), 15,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1306), 27,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [120131] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(2296), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(6451), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(1251), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      sym__special_character,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1253), 25,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [120185] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5638), 1,
      aux_sym_concatenation_token1,
    ACTIONS(6454), 1,
      sym__concat,
    STATE(1363), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1288), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1286), 34,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [120241] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(2285), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(6441), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(1261), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      sym__special_character,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1263), 25,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [120295] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1308), 15,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1310), 27,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [120345] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(2293), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(6441), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(2072), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      sym__special_character,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(2074), 25,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [120399] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1296), 15,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1298), 27,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [120449] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5638), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5640), 1,
      sym__concat,
    STATE(2290), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1263), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1261), 34,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [120505] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(2293), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(6441), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(2094), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      sym__special_character,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(2096), 25,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [120559] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1344), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1346), 28,
      sym__concat,
      sym_test_operator,
      anon_sym_RPAREN_RPAREN,
      anon_sym_COMMA,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      anon_sym_COLON,
      aux_sym_concatenation_token1,
  [120609] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1263), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1261), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [120659] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1263), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1261), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [120709] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1356), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1358), 28,
      sym__concat,
      sym_test_operator,
      anon_sym_RPAREN_RPAREN,
      anon_sym_COMMA,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      anon_sym_COLON,
      aux_sym_concatenation_token1,
  [120759] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(2285), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(6441), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(4467), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      sym__special_character,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(4469), 25,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [120813] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(2285), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(6441), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(5393), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      sym__special_character,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(5395), 25,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [120867] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1263), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1261), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [120917] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(6456), 1,
      sym__special_character,
    STATE(2311), 1,
      aux_sym__literal_repeat1,
    ACTIONS(1364), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1362), 36,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [120971] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(2312), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(6459), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(1251), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1253), 25,
      sym_test_operator,
      anon_sym_RPAREN_RPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      anon_sym_COLON,
  [121025] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1312), 15,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1314), 27,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [121075] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(2293), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(6441), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(4370), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      sym__special_character,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(4372), 25,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [121129] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1308), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1310), 28,
      sym__concat,
      sym_test_operator,
      anon_sym_RPAREN_RPAREN,
      anon_sym_COMMA,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      anon_sym_COLON,
      aux_sym_concatenation_token1,
  [121179] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(6462), 1,
      sym__special_character,
    STATE(2316), 1,
      aux_sym__literal_repeat1,
    ACTIONS(1364), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1362), 35,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [121233] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1304), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1306), 28,
      sym__concat,
      sym_test_operator,
      anon_sym_RPAREN_RPAREN,
      anon_sym_COMMA,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      anon_sym_COLON,
      aux_sym_concatenation_token1,
  [121283] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1336), 15,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1338), 27,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [121333] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1308), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1310), 28,
      sym__concat,
      sym_test_operator,
      anon_sym_RPAREN_RPAREN,
      anon_sym_COMMA,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      anon_sym_COLON,
      aux_sym_concatenation_token1,
  [121383] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1344), 15,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1346), 27,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [121433] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1312), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1314), 28,
      sym__concat,
      sym_test_operator,
      anon_sym_RPAREN_RPAREN,
      anon_sym_COMMA,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      anon_sym_COLON,
      aux_sym_concatenation_token1,
  [121483] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1340), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1342), 28,
      sym__concat,
      sym_test_operator,
      anon_sym_RPAREN_RPAREN,
      anon_sym_COMMA,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      anon_sym_COLON,
      aux_sym_concatenation_token1,
  [121533] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5664), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5666), 1,
      sym__concat,
    STATE(2329), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5069), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5067), 34,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [121589] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5664), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5666), 1,
      sym__concat,
    STATE(2336), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5375), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5373), 34,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [121645] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5638), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5640), 1,
      sym__concat,
    STATE(2297), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4372), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4370), 34,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [121701] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1356), 15,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1358), 27,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [121751] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5638), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5640), 1,
      sym__concat,
    STATE(2290), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4469), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4467), 34,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [121807] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5638), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5640), 1,
      sym__concat,
    STATE(2297), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(2096), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(2094), 34,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [121863] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5664), 1,
      aux_sym_concatenation_token1,
    ACTIONS(6465), 1,
      sym__concat,
    STATE(1371), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1267), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1265), 34,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [121919] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5664), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5666), 1,
      sym__concat,
    STATE(2336), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5176), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5174), 34,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [121975] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5664), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5666), 1,
      sym__concat,
    STATE(2329), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5180), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5178), 34,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [122031] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6467), 1,
      sym__special_character,
    STATE(2345), 1,
      aux_sym__literal_repeat1,
    ACTIONS(241), 15,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_LPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(278), 25,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [122085] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1328), 15,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1330), 27,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [122135] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1348), 15,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1350), 27,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [122185] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1352), 15,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1354), 27,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [122235] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5664), 1,
      aux_sym_concatenation_token1,
    ACTIONS(6469), 1,
      sym__concat,
    STATE(1371), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1288), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1286), 34,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [122291] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5638), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5640), 1,
      sym__concat,
    STATE(2297), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5391), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5389), 34,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [122347] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1263), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1261), 38,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [122397] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1340), 15,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1342), 27,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [122447] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5638), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5640), 1,
      sym__concat,
    STATE(2290), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5395), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5393), 34,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [122503] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1324), 15,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1326), 27,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [122553] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5664), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5666), 1,
      sym__concat,
    STATE(2329), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1263), 5,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1261), 34,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [122609] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1320), 15,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1322), 27,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [122659] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1251), 15,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1253), 27,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [122709] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6471), 1,
      sym__special_character,
    STATE(2345), 1,
      aux_sym__literal_repeat1,
    ACTIONS(1362), 15,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_LPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(1364), 25,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [122763] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1312), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      sym__special_character,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1314), 27,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [122812] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(2466), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1143), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(1261), 15,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      sym__special_character,
    ACTIONS(1263), 23,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [122865] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1336), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      sym__special_character,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1338), 27,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [122914] = 26,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(6480), 1,
      anon_sym_LPAREN,
    ACTIONS(6482), 1,
      anon_sym_esac,
    ACTIONS(6484), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(6486), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(6488), 1,
      anon_sym_DOLLAR,
    ACTIONS(6490), 1,
      sym__special_character,
    ACTIONS(6492), 1,
      anon_sym_DQUOTE,
    ACTIONS(6494), 1,
      aux_sym_number_token1,
    ACTIONS(6496), 1,
      aux_sym_number_token2,
    ACTIONS(6498), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(6500), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(6502), 1,
      anon_sym_BQUOTE,
    ACTIONS(6504), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(6508), 1,
      sym_test_operator,
    ACTIONS(6510), 1,
      sym_extglob_pattern,
    ACTIONS(6512), 1,
      sym__brace_start,
    STATE(6426), 1,
      aux_sym__literal_repeat1,
    STATE(7044), 1,
      sym_last_case_item,
    ACTIONS(6476), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(6506), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3236), 2,
      sym_case_item,
      aux_sym_case_statement_repeat1,
    STATE(6695), 2,
      sym_concatenation,
      sym__extglob_blob,
    ACTIONS(6474), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    ACTIONS(6478), 3,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
    STATE(6303), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [123009] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6514), 1,
      aux_sym_concatenation_token1,
    ACTIONS(6516), 1,
      sym__concat,
    STATE(2359), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1286), 13,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1288), 25,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [123064] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1261), 15,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_LPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(1263), 26,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [123113] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6518), 1,
      sym__special_character,
    STATE(2366), 1,
      aux_sym__literal_repeat1,
    ACTIONS(4564), 14,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(4566), 25,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [123166] = 26,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(6480), 1,
      anon_sym_LPAREN,
    ACTIONS(6486), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(6488), 1,
      anon_sym_DOLLAR,
    ACTIONS(6490), 1,
      sym__special_character,
    ACTIONS(6492), 1,
      anon_sym_DQUOTE,
    ACTIONS(6494), 1,
      aux_sym_number_token1,
    ACTIONS(6496), 1,
      aux_sym_number_token2,
    ACTIONS(6498), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(6500), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(6502), 1,
      anon_sym_BQUOTE,
    ACTIONS(6504), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(6508), 1,
      sym_test_operator,
    ACTIONS(6510), 1,
      sym_extglob_pattern,
    ACTIONS(6512), 1,
      sym__brace_start,
    ACTIONS(6522), 1,
      anon_sym_esac,
    ACTIONS(6524), 1,
      aux_sym_heredoc_redirect_token1,
    STATE(6426), 1,
      aux_sym__literal_repeat1,
    STATE(7056), 1,
      sym_last_case_item,
    ACTIONS(6476), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(6506), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3233), 2,
      sym_case_item,
      aux_sym_case_statement_repeat1,
    STATE(6695), 2,
      sym_concatenation,
      sym__extglob_blob,
    ACTIONS(6474), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    ACTIONS(6520), 3,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
    STATE(6303), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [123261] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1300), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1302), 27,
      sym__concat,
      sym_test_operator,
      anon_sym_RPAREN_RPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      anon_sym_COLON,
      aux_sym_concatenation_token1,
  [123310] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1332), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1334), 27,
      sym__concat,
      sym_test_operator,
      anon_sym_RPAREN_RPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      anon_sym_COLON,
      aux_sym_concatenation_token1,
  [123359] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(2424), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(6526), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(1261), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      sym__special_character,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1263), 24,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [123412] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(2350), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(6514), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(2094), 13,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(2096), 25,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [123465] = 26,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(6480), 1,
      anon_sym_LPAREN,
    ACTIONS(6486), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(6488), 1,
      anon_sym_DOLLAR,
    ACTIONS(6490), 1,
      sym__special_character,
    ACTIONS(6492), 1,
      anon_sym_DQUOTE,
    ACTIONS(6494), 1,
      aux_sym_number_token1,
    ACTIONS(6496), 1,
      aux_sym_number_token2,
    ACTIONS(6498), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(6500), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(6502), 1,
      anon_sym_BQUOTE,
    ACTIONS(6504), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(6508), 1,
      sym_test_operator,
    ACTIONS(6510), 1,
      sym_extglob_pattern,
    ACTIONS(6512), 1,
      sym__brace_start,
    ACTIONS(6530), 1,
      anon_sym_esac,
    ACTIONS(6532), 1,
      aux_sym_heredoc_redirect_token1,
    STATE(6426), 1,
      aux_sym__literal_repeat1,
    STATE(7005), 1,
      sym_last_case_item,
    ACTIONS(6476), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(6506), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3385), 2,
      sym_case_item,
      aux_sym_case_statement_repeat1,
    STATE(6695), 2,
      sym_concatenation,
      sym__extglob_blob,
    ACTIONS(6474), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    ACTIONS(6528), 3,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
    STATE(6303), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [123560] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(2359), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(6534), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(1251), 13,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1253), 25,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [123613] = 26,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(6480), 1,
      anon_sym_LPAREN,
    ACTIONS(6486), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(6488), 1,
      anon_sym_DOLLAR,
    ACTIONS(6490), 1,
      sym__special_character,
    ACTIONS(6492), 1,
      anon_sym_DQUOTE,
    ACTIONS(6494), 1,
      aux_sym_number_token1,
    ACTIONS(6496), 1,
      aux_sym_number_token2,
    ACTIONS(6498), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(6500), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(6502), 1,
      anon_sym_BQUOTE,
    ACTIONS(6504), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(6508), 1,
      sym_test_operator,
    ACTIONS(6510), 1,
      sym_extglob_pattern,
    ACTIONS(6512), 1,
      sym__brace_start,
    ACTIONS(6539), 1,
      anon_sym_esac,
    ACTIONS(6541), 1,
      aux_sym_heredoc_redirect_token1,
    STATE(6426), 1,
      aux_sym__literal_repeat1,
    STATE(7275), 1,
      sym_last_case_item,
    ACTIONS(6476), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(6506), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3452), 2,
      sym_case_item,
      aux_sym_case_statement_repeat1,
    STATE(6695), 2,
      sym_concatenation,
      sym__extglob_blob,
    ACTIONS(6474), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    ACTIONS(6537), 3,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
    STATE(6303), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [123708] = 26,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(6480), 1,
      anon_sym_LPAREN,
    ACTIONS(6486), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(6488), 1,
      anon_sym_DOLLAR,
    ACTIONS(6490), 1,
      sym__special_character,
    ACTIONS(6492), 1,
      anon_sym_DQUOTE,
    ACTIONS(6494), 1,
      aux_sym_number_token1,
    ACTIONS(6496), 1,
      aux_sym_number_token2,
    ACTIONS(6498), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(6500), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(6502), 1,
      anon_sym_BQUOTE,
    ACTIONS(6504), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(6508), 1,
      sym_test_operator,
    ACTIONS(6510), 1,
      sym_extglob_pattern,
    ACTIONS(6512), 1,
      sym__brace_start,
    ACTIONS(6545), 1,
      anon_sym_esac,
    ACTIONS(6547), 1,
      aux_sym_heredoc_redirect_token1,
    STATE(6426), 1,
      aux_sym__literal_repeat1,
    STATE(7134), 1,
      sym_last_case_item,
    ACTIONS(6476), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(6506), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3451), 2,
      sym_case_item,
      aux_sym_case_statement_repeat1,
    STATE(6695), 2,
      sym_concatenation,
      sym__extglob_blob,
    ACTIONS(6474), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    ACTIONS(6543), 3,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
    STATE(6303), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [123803] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1344), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      sym__special_character,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1346), 27,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [123852] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1356), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      sym__special_character,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1358), 27,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [123901] = 26,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(6480), 1,
      anon_sym_LPAREN,
    ACTIONS(6486), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(6488), 1,
      anon_sym_DOLLAR,
    ACTIONS(6490), 1,
      sym__special_character,
    ACTIONS(6492), 1,
      anon_sym_DQUOTE,
    ACTIONS(6494), 1,
      aux_sym_number_token1,
    ACTIONS(6496), 1,
      aux_sym_number_token2,
    ACTIONS(6498), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(6500), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(6502), 1,
      anon_sym_BQUOTE,
    ACTIONS(6504), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(6508), 1,
      sym_test_operator,
    ACTIONS(6510), 1,
      sym_extglob_pattern,
    ACTIONS(6512), 1,
      sym__brace_start,
    ACTIONS(6551), 1,
      anon_sym_esac,
    ACTIONS(6553), 1,
      aux_sym_heredoc_redirect_token1,
    STATE(6426), 1,
      aux_sym__literal_repeat1,
    STATE(7007), 1,
      sym_last_case_item,
    ACTIONS(6476), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(6506), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3386), 2,
      sym_case_item,
      aux_sym_case_statement_repeat1,
    STATE(6695), 2,
      sym_concatenation,
      sym__extglob_blob,
    ACTIONS(6474), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    ACTIONS(6549), 3,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
    STATE(6303), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [123996] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5844), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5846), 1,
      sym__concat,
    STATE(2385), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5478), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5476), 34,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [124051] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6555), 1,
      sym__special_character,
    STATE(2366), 1,
      aux_sym__literal_repeat1,
    ACTIONS(1362), 14,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(1364), 25,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [124104] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5844), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5846), 1,
      sym__concat,
    STATE(2403), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5470), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5468), 34,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [124159] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1271), 15,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_LPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(1281), 26,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [124208] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(2350), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(6514), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(5389), 13,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(5391), 25,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [124261] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6518), 1,
      sym__special_character,
    STATE(2366), 1,
      aux_sym__literal_repeat1,
    ACTIONS(4467), 14,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(4469), 25,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [124314] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(2429), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(6514), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(5393), 13,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(5395), 25,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [124367] = 26,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(6480), 1,
      anon_sym_LPAREN,
    ACTIONS(6486), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(6488), 1,
      anon_sym_DOLLAR,
    ACTIONS(6490), 1,
      sym__special_character,
    ACTIONS(6492), 1,
      anon_sym_DQUOTE,
    ACTIONS(6494), 1,
      aux_sym_number_token1,
    ACTIONS(6496), 1,
      aux_sym_number_token2,
    ACTIONS(6498), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(6500), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(6502), 1,
      anon_sym_BQUOTE,
    ACTIONS(6504), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(6508), 1,
      sym_test_operator,
    ACTIONS(6510), 1,
      sym_extglob_pattern,
    ACTIONS(6512), 1,
      sym__brace_start,
    ACTIONS(6560), 1,
      anon_sym_esac,
    ACTIONS(6562), 1,
      aux_sym_heredoc_redirect_token1,
    STATE(6426), 1,
      aux_sym__literal_repeat1,
    STATE(6851), 1,
      sym_last_case_item,
    ACTIONS(6476), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(6506), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3377), 2,
      sym_case_item,
      aux_sym_case_statement_repeat1,
    STATE(6695), 2,
      sym_concatenation,
      sym__extglob_blob,
    ACTIONS(6474), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    ACTIONS(6558), 3,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
    STATE(6303), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [124462] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(2407), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(6564), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(1261), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      aux_sym__simple_variable_name_token1,
      sym_word,
    ACTIONS(1263), 24,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [124515] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1105), 1,
      aux_sym_concatenation_token1,
    STATE(2540), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1261), 15,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      sym__special_character,
    ACTIONS(1263), 24,
      sym__concat,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [124568] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1332), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      sym__special_character,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1334), 27,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [124617] = 26,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(6480), 1,
      anon_sym_LPAREN,
    ACTIONS(6486), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(6488), 1,
      anon_sym_DOLLAR,
    ACTIONS(6490), 1,
      sym__special_character,
    ACTIONS(6492), 1,
      anon_sym_DQUOTE,
    ACTIONS(6494), 1,
      aux_sym_number_token1,
    ACTIONS(6496), 1,
      aux_sym_number_token2,
    ACTIONS(6498), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(6500), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(6502), 1,
      anon_sym_BQUOTE,
    ACTIONS(6504), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(6508), 1,
      sym_test_operator,
    ACTIONS(6510), 1,
      sym_extglob_pattern,
    ACTIONS(6512), 1,
      sym__brace_start,
    ACTIONS(6568), 1,
      anon_sym_esac,
    ACTIONS(6570), 1,
      aux_sym_heredoc_redirect_token1,
    STATE(6426), 1,
      aux_sym__literal_repeat1,
    STATE(6997), 1,
      sym_last_case_item,
    ACTIONS(6476), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(6506), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3383), 2,
      sym_case_item,
      aux_sym_case_statement_repeat1,
    STATE(6695), 2,
      sym_concatenation,
      sym__extglob_blob,
    ACTIONS(6474), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    ACTIONS(6566), 3,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
    STATE(6303), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [124712] = 26,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(6480), 1,
      anon_sym_LPAREN,
    ACTIONS(6486), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(6488), 1,
      anon_sym_DOLLAR,
    ACTIONS(6490), 1,
      sym__special_character,
    ACTIONS(6492), 1,
      anon_sym_DQUOTE,
    ACTIONS(6494), 1,
      aux_sym_number_token1,
    ACTIONS(6496), 1,
      aux_sym_number_token2,
    ACTIONS(6498), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(6500), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(6502), 1,
      anon_sym_BQUOTE,
    ACTIONS(6504), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(6508), 1,
      sym_test_operator,
    ACTIONS(6510), 1,
      sym_extglob_pattern,
    ACTIONS(6512), 1,
      sym__brace_start,
    ACTIONS(6574), 1,
      anon_sym_esac,
    ACTIONS(6576), 1,
      aux_sym_heredoc_redirect_token1,
    STATE(6426), 1,
      aux_sym__literal_repeat1,
    STATE(7603), 1,
      sym_last_case_item,
    ACTIONS(6476), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(6506), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3440), 2,
      sym_case_item,
      aux_sym_case_statement_repeat1,
    STATE(6695), 2,
      sym_concatenation,
      sym__extglob_blob,
    ACTIONS(6474), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    ACTIONS(6572), 3,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
    STATE(6303), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [124807] = 26,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(6480), 1,
      anon_sym_LPAREN,
    ACTIONS(6486), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(6488), 1,
      anon_sym_DOLLAR,
    ACTIONS(6490), 1,
      sym__special_character,
    ACTIONS(6492), 1,
      anon_sym_DQUOTE,
    ACTIONS(6494), 1,
      aux_sym_number_token1,
    ACTIONS(6496), 1,
      aux_sym_number_token2,
    ACTIONS(6498), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(6500), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(6502), 1,
      anon_sym_BQUOTE,
    ACTIONS(6504), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(6508), 1,
      sym_test_operator,
    ACTIONS(6510), 1,
      sym_extglob_pattern,
    ACTIONS(6512), 1,
      sym__brace_start,
    ACTIONS(6580), 1,
      anon_sym_esac,
    ACTIONS(6582), 1,
      aux_sym_heredoc_redirect_token1,
    STATE(6426), 1,
      aux_sym__literal_repeat1,
    STATE(6824), 1,
      sym_last_case_item,
    ACTIONS(6476), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(6506), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3378), 2,
      sym_case_item,
      aux_sym_case_statement_repeat1,
    STATE(6695), 2,
      sym_concatenation,
      sym__extglob_blob,
    ACTIONS(6474), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    ACTIONS(6578), 3,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
    STATE(6303), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [124902] = 26,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(6480), 1,
      anon_sym_LPAREN,
    ACTIONS(6486), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(6488), 1,
      anon_sym_DOLLAR,
    ACTIONS(6490), 1,
      sym__special_character,
    ACTIONS(6492), 1,
      anon_sym_DQUOTE,
    ACTIONS(6494), 1,
      aux_sym_number_token1,
    ACTIONS(6496), 1,
      aux_sym_number_token2,
    ACTIONS(6498), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(6500), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(6502), 1,
      anon_sym_BQUOTE,
    ACTIONS(6504), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(6508), 1,
      sym_test_operator,
    ACTIONS(6510), 1,
      sym_extglob_pattern,
    ACTIONS(6512), 1,
      sym__brace_start,
    ACTIONS(6586), 1,
      anon_sym_esac,
    ACTIONS(6588), 1,
      aux_sym_heredoc_redirect_token1,
    STATE(6426), 1,
      aux_sym__literal_repeat1,
    STATE(6991), 1,
      sym_last_case_item,
    ACTIONS(6476), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(6506), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3382), 2,
      sym_case_item,
      aux_sym_case_statement_repeat1,
    STATE(6695), 2,
      sym_concatenation,
      sym__extglob_blob,
    ACTIONS(6474), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    ACTIONS(6584), 3,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
    STATE(6303), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [124997] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1352), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      sym__special_character,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1354), 27,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [125046] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1263), 5,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1261), 36,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [125095] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(2555), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1105), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(1261), 15,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      sym__special_character,
    ACTIONS(1263), 23,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [125148] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1296), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1298), 27,
      sym__concat,
      sym_test_operator,
      anon_sym_RPAREN_RPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      anon_sym_COLON,
      aux_sym_concatenation_token1,
  [125197] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1300), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      sym__special_character,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1302), 27,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [125246] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5844), 1,
      aux_sym_concatenation_token1,
    ACTIONS(6590), 1,
      sym__concat,
    STATE(1722), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1267), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1265), 34,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [125301] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5844), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5846), 1,
      sym__concat,
    STATE(2385), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1263), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1261), 34,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [125356] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1369), 1,
      sym_test_operator,
    STATE(2575), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1143), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(6594), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(6592), 23,
      anon_sym_RPAREN_RPAREN,
      anon_sym_COMMA,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [125411] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1296), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      sym__special_character,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1298), 27,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [125460] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5920), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5946), 1,
      sym__concat,
    STATE(2404), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1263), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1261), 34,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [125515] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1340), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      sym__special_character,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1342), 27,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [125564] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5920), 1,
      aux_sym_concatenation_token1,
    ACTIONS(6596), 1,
      sym__concat,
    STATE(1704), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1288), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1286), 34,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [125619] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1251), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1253), 27,
      sym__concat,
      sym_test_operator,
      anon_sym_RPAREN_RPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      anon_sym_COLON,
      aux_sym_concatenation_token1,
  [125668] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1324), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1326), 27,
      sym__concat,
      sym_test_operator,
      anon_sym_RPAREN_RPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      anon_sym_COLON,
      aux_sym_concatenation_token1,
  [125717] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6518), 1,
      sym__special_character,
    STATE(2366), 1,
      aux_sym__literal_repeat1,
    ACTIONS(4480), 14,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(4482), 25,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [125770] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(2407), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(6564), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(5178), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      aux_sym__simple_variable_name_token1,
      sym_word,
    ACTIONS(5180), 24,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [125823] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(2429), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(6514), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(1261), 13,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1263), 25,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [125876] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1369), 1,
      sym_test_operator,
    STATE(2575), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1143), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(6600), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(6598), 23,
      anon_sym_RPAREN_RPAREN,
      anon_sym_COMMA,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [125931] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1292), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      sym__special_character,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1294), 27,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [125980] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1328), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      sym__special_character,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1330), 27,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [126029] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1320), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      sym__special_character,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1322), 27,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [126078] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1251), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      sym__special_character,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1253), 27,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [126127] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1324), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      sym__special_character,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1326), 27,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [126176] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5844), 1,
      aux_sym_concatenation_token1,
    ACTIONS(6602), 1,
      sym__concat,
    STATE(1722), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1288), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1286), 34,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
  [126231] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5920), 1,
      aux_sym_concatenation_token1,
    ACTIONS(6604), 1,
      sym__concat,
    STATE(1704), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1267), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1265), 34,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [126286] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(2407), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(6564), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(5067), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      aux_sym__simple_variable_name_token1,
      sym_word,
    ACTIONS(5069), 24,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [126339] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(2426), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(6564), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(5373), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      aux_sym__simple_variable_name_token1,
      sym_word,
    ACTIONS(5375), 24,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [126392] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6564), 1,
      aux_sym_concatenation_token1,
    ACTIONS(6606), 1,
      sym__concat,
    STATE(2435), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1265), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      aux_sym__simple_variable_name_token1,
      sym_word,
    ACTIONS(1267), 24,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [126447] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(2495), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1143), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(1261), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1263), 24,
      sym_test_operator,
      anon_sym_RPAREN_RPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      sym__special_character,
  [126500] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(2459), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1143), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(1261), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1263), 24,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      anon_sym_COLON,
      sym__special_character,
  [126553] = 26,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(6480), 1,
      anon_sym_LPAREN,
    ACTIONS(6486), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(6488), 1,
      anon_sym_DOLLAR,
    ACTIONS(6490), 1,
      sym__special_character,
    ACTIONS(6492), 1,
      anon_sym_DQUOTE,
    ACTIONS(6494), 1,
      aux_sym_number_token1,
    ACTIONS(6496), 1,
      aux_sym_number_token2,
    ACTIONS(6498), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(6500), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(6502), 1,
      anon_sym_BQUOTE,
    ACTIONS(6504), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(6508), 1,
      sym_test_operator,
    ACTIONS(6510), 1,
      sym_extglob_pattern,
    ACTIONS(6512), 1,
      sym__brace_start,
    ACTIONS(6610), 1,
      anon_sym_esac,
    ACTIONS(6612), 1,
      aux_sym_heredoc_redirect_token1,
    STATE(6426), 1,
      aux_sym__literal_repeat1,
    STATE(7451), 1,
      sym_last_case_item,
    ACTIONS(6476), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(6506), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3420), 2,
      sym_case_item,
      aux_sym_case_statement_repeat1,
    STATE(6695), 2,
      sym_concatenation,
      sym__extglob_blob,
    ACTIONS(6474), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    ACTIONS(6608), 3,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
    STATE(6303), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [126648] = 26,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(6480), 1,
      anon_sym_LPAREN,
    ACTIONS(6486), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(6488), 1,
      anon_sym_DOLLAR,
    ACTIONS(6490), 1,
      sym__special_character,
    ACTIONS(6492), 1,
      anon_sym_DQUOTE,
    ACTIONS(6494), 1,
      aux_sym_number_token1,
    ACTIONS(6496), 1,
      aux_sym_number_token2,
    ACTIONS(6498), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(6500), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(6502), 1,
      anon_sym_BQUOTE,
    ACTIONS(6504), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(6508), 1,
      sym_test_operator,
    ACTIONS(6510), 1,
      sym_extglob_pattern,
    ACTIONS(6512), 1,
      sym__brace_start,
    ACTIONS(6616), 1,
      anon_sym_esac,
    ACTIONS(6618), 1,
      aux_sym_heredoc_redirect_token1,
    STATE(6426), 1,
      aux_sym__literal_repeat1,
    STATE(7440), 1,
      sym_last_case_item,
    ACTIONS(6476), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(6506), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3419), 2,
      sym_case_item,
      aux_sym_case_statement_repeat1,
    STATE(6695), 2,
      sym_concatenation,
      sym__extglob_blob,
    ACTIONS(6474), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    ACTIONS(6614), 3,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
    STATE(6303), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [126743] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1348), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      sym__special_character,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1350), 27,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [126792] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(2350), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(6514), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(4552), 13,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(4554), 25,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [126845] = 26,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(6480), 1,
      anon_sym_LPAREN,
    ACTIONS(6486), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(6488), 1,
      anon_sym_DOLLAR,
    ACTIONS(6490), 1,
      sym__special_character,
    ACTIONS(6492), 1,
      anon_sym_DQUOTE,
    ACTIONS(6494), 1,
      aux_sym_number_token1,
    ACTIONS(6496), 1,
      aux_sym_number_token2,
    ACTIONS(6498), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(6500), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(6502), 1,
      anon_sym_BQUOTE,
    ACTIONS(6504), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(6508), 1,
      sym_test_operator,
    ACTIONS(6510), 1,
      sym_extglob_pattern,
    ACTIONS(6512), 1,
      sym__brace_start,
    ACTIONS(6622), 1,
      anon_sym_esac,
    ACTIONS(6624), 1,
      aux_sym_heredoc_redirect_token1,
    STATE(6426), 1,
      aux_sym__literal_repeat1,
    STATE(7413), 1,
      sym_last_case_item,
    ACTIONS(6476), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(6506), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3416), 2,
      sym_case_item,
      aux_sym_case_statement_repeat1,
    STATE(6695), 2,
      sym_concatenation,
      sym__extglob_blob,
    ACTIONS(6474), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    ACTIONS(6620), 3,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
    STATE(6303), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [126940] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(2429), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(6514), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(4564), 13,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(4566), 25,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [126993] = 26,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(6480), 1,
      anon_sym_LPAREN,
    ACTIONS(6486), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(6488), 1,
      anon_sym_DOLLAR,
    ACTIONS(6490), 1,
      sym__special_character,
    ACTIONS(6492), 1,
      anon_sym_DQUOTE,
    ACTIONS(6494), 1,
      aux_sym_number_token1,
    ACTIONS(6496), 1,
      aux_sym_number_token2,
    ACTIONS(6498), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(6500), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(6502), 1,
      anon_sym_BQUOTE,
    ACTIONS(6504), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(6508), 1,
      sym_test_operator,
    ACTIONS(6510), 1,
      sym_extglob_pattern,
    ACTIONS(6512), 1,
      sym__brace_start,
    ACTIONS(6628), 1,
      anon_sym_esac,
    ACTIONS(6630), 1,
      aux_sym_heredoc_redirect_token1,
    STATE(6426), 1,
      aux_sym__literal_repeat1,
    STATE(7402), 1,
      sym_last_case_item,
    ACTIONS(6476), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(6506), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3411), 2,
      sym_case_item,
      aux_sym_case_statement_repeat1,
    STATE(6695), 2,
      sym_concatenation,
      sym__extglob_blob,
    ACTIONS(6474), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    ACTIONS(6626), 3,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
    STATE(6303), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [127088] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6518), 1,
      sym__special_character,
    STATE(2366), 1,
      aux_sym__literal_repeat1,
    ACTIONS(4546), 14,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(4548), 25,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [127141] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1263), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1261), 37,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [127190] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(2426), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(6564), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(5174), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      aux_sym__simple_variable_name_token1,
      sym_word,
    ACTIONS(5176), 24,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [127243] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1316), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      sym__special_character,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1318), 27,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [127292] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1308), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      sym__special_character,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1310), 27,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [127341] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5920), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5946), 1,
      sym__concat,
    STATE(2391), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5391), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5389), 34,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [127396] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1304), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      sym__special_character,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1306), 27,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [127445] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6526), 1,
      aux_sym_concatenation_token1,
    ACTIONS(6632), 1,
      sym__concat,
    STATE(2425), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1265), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      sym__special_character,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1267), 24,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [127500] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(2425), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(6634), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(1251), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      sym__special_character,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1253), 24,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [127553] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6564), 1,
      aux_sym_concatenation_token1,
    ACTIONS(6637), 1,
      sym__concat,
    STATE(2435), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1286), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      aux_sym__simple_variable_name_token1,
      sym_word,
    ACTIONS(1288), 24,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [127608] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(2350), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(6514), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(2072), 13,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(2074), 25,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [127661] = 26,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(6480), 1,
      anon_sym_LPAREN,
    ACTIONS(6486), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(6488), 1,
      anon_sym_DOLLAR,
    ACTIONS(6490), 1,
      sym__special_character,
    ACTIONS(6492), 1,
      anon_sym_DQUOTE,
    ACTIONS(6494), 1,
      aux_sym_number_token1,
    ACTIONS(6496), 1,
      aux_sym_number_token2,
    ACTIONS(6498), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(6500), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(6502), 1,
      anon_sym_BQUOTE,
    ACTIONS(6504), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(6508), 1,
      sym_test_operator,
    ACTIONS(6510), 1,
      sym_extglob_pattern,
    ACTIONS(6512), 1,
      sym__brace_start,
    ACTIONS(6641), 1,
      anon_sym_esac,
    ACTIONS(6643), 1,
      aux_sym_heredoc_redirect_token1,
    STATE(6426), 1,
      aux_sym__literal_repeat1,
    STATE(7528), 1,
      sym_last_case_item,
    ACTIONS(6476), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(6506), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3310), 2,
      sym_case_item,
      aux_sym_case_statement_repeat1,
    STATE(6695), 2,
      sym_concatenation,
      sym__extglob_blob,
    ACTIONS(6474), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    ACTIONS(6639), 3,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
    STATE(6303), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [127756] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6514), 1,
      aux_sym_concatenation_token1,
    ACTIONS(6645), 1,
      sym__concat,
    STATE(2359), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1265), 13,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1267), 25,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [127811] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1308), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      sym__special_character,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1310), 27,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [127860] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5920), 1,
      aux_sym_concatenation_token1,
    ACTIONS(5946), 1,
      sym__concat,
    STATE(2404), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5395), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5393), 34,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [127915] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(2429), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(6514), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(4467), 13,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(4469), 25,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [127968] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(2491), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1077), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(1261), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1263), 24,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      sym__special_character,
  [128021] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(2429), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(6514), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(5067), 13,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(5069), 25,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [128074] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(2435), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(6647), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(1251), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      aux_sym__simple_variable_name_token1,
      sym_word,
    ACTIONS(1253), 24,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [128127] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(2350), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(6514), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(4370), 13,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(4372), 25,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [128180] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(2424), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(6526), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(5393), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      sym__special_character,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(5395), 24,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [128233] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(2441), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(6526), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(5389), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      sym__special_character,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(5391), 24,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [128286] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1292), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1294), 27,
      sym__concat,
      sym_test_operator,
      anon_sym_RPAREN_RPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      anon_sym_COLON,
      aux_sym_concatenation_token1,
  [128335] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(2350), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(6514), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(5373), 13,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(5375), 25,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [128388] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6526), 1,
      aux_sym_concatenation_token1,
    ACTIONS(6650), 1,
      sym__concat,
    STATE(2425), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1286), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      sym__special_character,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1288), 24,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [128443] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1348), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1350), 27,
      sym__concat,
      sym_test_operator,
      anon_sym_RPAREN_RPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      anon_sym_COLON,
      aux_sym_concatenation_token1,
  [128492] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1312), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1314), 26,
      sym_test_operator,
      sym_extglob_pattern,
      anon_sym_RPAREN_RPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      anon_sym_COLON,
  [128540] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1328), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      aux_sym__simple_variable_name_token1,
      sym_word,
    ACTIONS(1330), 26,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [128588] = 7,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1115), 1,
      anon_sym_RBRACK,
    ACTIONS(6652), 1,
      sym__special_character,
    ACTIONS(6654), 1,
      sym__concat,
    STATE(2474), 1,
      aux_sym__literal_repeat1,
    ACTIONS(243), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(354), 22,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [128644] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1308), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      sym__special_character,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1310), 26,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [128692] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1348), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      sym__special_character,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1350), 26,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [128740] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1304), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      sym__special_character,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1306), 26,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [128788] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1324), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      sym__special_character,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1326), 26,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [128836] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1251), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      sym__special_character,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1253), 26,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [128884] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1105), 1,
      aux_sym_concatenation_token1,
    STATE(2533), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1273), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1369), 24,
      sym__concat,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [128936] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1340), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      aux_sym__simple_variable_name_token1,
      sym_word,
    ACTIONS(1342), 26,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [128984] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(2564), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(6656), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(1261), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      aux_sym__simple_variable_name_token1,
      sym_word,
    ACTIONS(1263), 23,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [129036] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1320), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      aux_sym__simple_variable_name_token1,
      sym_word,
    ACTIONS(1322), 26,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [129084] = 7,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1105), 1,
      aux_sym_concatenation_token1,
    ACTIONS(6658), 1,
      anon_sym_RBRACK,
    ACTIONS(6660), 1,
      sym__concat,
    STATE(2533), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1273), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1369), 22,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [129140] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(2575), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1143), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(1273), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1369), 23,
      sym_test_operator,
      anon_sym_RPAREN_RPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [129192] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6662), 1,
      sym__special_character,
    STATE(2568), 1,
      aux_sym__literal_repeat1,
    ACTIONS(5393), 12,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(5395), 26,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [129244] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(2550), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(6664), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(1261), 13,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1263), 24,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [129296] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1143), 1,
      aux_sym_concatenation_token1,
    ACTIONS(6666), 1,
      sym__concat,
    STATE(2312), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1265), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1267), 23,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      anon_sym_COLON,
  [129350] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(2494), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(6668), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(1261), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      sym__special_character,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1263), 23,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [129402] = 7,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1105), 1,
      aux_sym_concatenation_token1,
    ACTIONS(6670), 1,
      anon_sym_RBRACK,
    ACTIONS(6672), 1,
      sym__concat,
    STATE(2533), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1273), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1369), 22,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [129458] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1296), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      sym__special_character,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1298), 26,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [129506] = 7,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1207), 1,
      anon_sym_RBRACK,
    ACTIONS(6652), 1,
      sym__special_character,
    ACTIONS(6674), 1,
      sym__concat,
    STATE(2474), 1,
      aux_sym__literal_repeat1,
    ACTIONS(243), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(354), 22,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [129562] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(2464), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(6676), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(1251), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1253), 23,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [129614] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1352), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      aux_sym__simple_variable_name_token1,
      sym_word,
    ACTIONS(1354), 26,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [129662] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1143), 1,
      aux_sym_concatenation_token1,
    ACTIONS(6679), 1,
      sym__concat,
    STATE(2312), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1265), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1267), 23,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [129716] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1356), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      aux_sym__simple_variable_name_token1,
      sym_word,
    ACTIONS(1358), 26,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [129764] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1344), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      aux_sym__simple_variable_name_token1,
      sym_word,
    ACTIONS(1346), 26,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [129812] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1316), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      aux_sym__simple_variable_name_token1,
      sym_word,
    ACTIONS(1318), 26,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [129860] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(6013), 1,
      aux_sym_concatenation_token1,
    ACTIONS(6681), 1,
      sym__concat,
    STATE(1879), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1267), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1265), 33,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [129914] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(6013), 1,
      aux_sym_concatenation_token1,
    ACTIONS(6683), 1,
      sym__concat,
    STATE(1879), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1288), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1286), 33,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [129968] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6652), 1,
      sym__special_character,
    STATE(2474), 1,
      aux_sym__literal_repeat1,
    ACTIONS(243), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(354), 24,
      sym__concat,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [130020] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1143), 1,
      aux_sym_concatenation_token1,
    ACTIONS(6685), 1,
      sym__concat,
    STATE(2312), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1286), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1288), 23,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [130074] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6687), 1,
      sym__special_character,
    STATE(2474), 1,
      aux_sym__literal_repeat1,
    ACTIONS(1362), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1364), 24,
      sym__concat,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [130126] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1077), 1,
      aux_sym_concatenation_token1,
    ACTIONS(6690), 1,
      sym__concat,
    STATE(2464), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1286), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1288), 23,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [130180] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6662), 1,
      sym__special_character,
    STATE(2568), 1,
      aux_sym__literal_repeat1,
    ACTIONS(5721), 12,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(5723), 26,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [130232] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1336), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      aux_sym__simple_variable_name_token1,
      sym_word,
    ACTIONS(1338), 26,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [130280] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1300), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      sym__special_character,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1302), 26,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [130328] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1292), 13,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1294), 27,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [130376] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(6013), 1,
      aux_sym_concatenation_token1,
    ACTIONS(6015), 1,
      sym__concat,
    STATE(2470), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1263), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1261), 33,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [130430] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1332), 13,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1334), 27,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [130478] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1332), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      sym__special_character,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1334), 26,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [130526] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(2483), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(6692), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(1251), 13,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1253), 24,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [130578] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6695), 1,
      sym__special_character,
    STATE(2557), 1,
      aux_sym__literal_repeat1,
    ACTIONS(243), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(354), 24,
      sym_test_operator,
      anon_sym_RPAREN_RPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      anon_sym_COLON,
  [130630] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1320), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1322), 26,
      sym_test_operator,
      sym_extglob_pattern,
      anon_sym_RPAREN_RPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      anon_sym_COLON,
  [130678] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(2486), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(6697), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(1251), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      sym__special_character,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1253), 23,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [130730] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(2543), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(6664), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(5389), 13,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(5391), 24,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [130782] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6668), 1,
      aux_sym_concatenation_token1,
    ACTIONS(6700), 1,
      sym__concat,
    STATE(2486), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1286), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      sym__special_character,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1288), 23,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [130836] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(2488), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(6668), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(5389), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      sym__special_character,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(5391), 23,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [130888] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(2494), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(6668), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(5393), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      sym__special_character,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(5395), 23,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [130940] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1077), 1,
      aux_sym_concatenation_token1,
    ACTIONS(6702), 1,
      sym__concat,
    STATE(2464), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1265), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1267), 23,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [130994] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1261), 14,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(1263), 26,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [131042] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(2550), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(6664), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(5393), 13,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(5395), 24,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [131094] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6668), 1,
      aux_sym_concatenation_token1,
    ACTIONS(6704), 1,
      sym__concat,
    STATE(2486), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1265), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      sym__special_character,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1267), 23,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [131148] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1143), 1,
      aux_sym_concatenation_token1,
    ACTIONS(6706), 1,
      sym__concat,
    STATE(2312), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1265), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1267), 23,
      sym_test_operator,
      anon_sym_RPAREN_RPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [131202] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1312), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      sym__special_character,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1314), 26,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [131250] = 7,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1105), 1,
      aux_sym_concatenation_token1,
    ACTIONS(6708), 1,
      anon_sym_RBRACK,
    ACTIONS(6710), 1,
      sym__concat,
    STATE(2533), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1273), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1369), 22,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [131306] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1300), 13,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1302), 27,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [131354] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1312), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      aux_sym__simple_variable_name_token1,
      sym_word,
    ACTIONS(1314), 26,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [131402] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(2475), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1077), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(1273), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1369), 23,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [131454] = 7,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1203), 1,
      anon_sym_RBRACK,
    ACTIONS(6652), 1,
      sym__special_character,
    ACTIONS(6712), 1,
      sym__concat,
    STATE(2474), 1,
      aux_sym__literal_repeat1,
    ACTIONS(243), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(354), 22,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [131510] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1296), 13,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1298), 27,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [131558] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(2473), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1143), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(1273), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1369), 23,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [131610] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1292), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      sym__special_character,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1294), 26,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [131658] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1336), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1338), 26,
      sym_test_operator,
      sym_extglob_pattern,
      anon_sym_RPAREN_RPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      anon_sym_COLON,
  [131706] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1369), 1,
      sym_test_operator,
    STATE(2475), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1077), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(6600), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(6598), 22,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [131760] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1316), 13,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1318), 27,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [131808] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1320), 13,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1322), 27,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [131856] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1369), 1,
      sym_test_operator,
    STATE(2475), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1077), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(6594), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(6592), 22,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [131910] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1328), 13,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1330), 27,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [131958] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(2511), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(6714), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(1251), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1253), 23,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [132010] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1340), 13,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1342), 27,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [132058] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1352), 13,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1354), 27,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [132106] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1251), 13,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1253), 27,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [132154] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1356), 13,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1358), 27,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [132202] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1344), 13,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1346), 27,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [132250] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1336), 13,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1338), 27,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [132298] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1324), 13,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1326), 27,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [132346] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1340), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1342), 26,
      sym_test_operator,
      sym_extglob_pattern,
      anon_sym_RPAREN_RPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      anon_sym_COLON,
  [132394] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1312), 13,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1314), 27,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [132442] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(2551), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1105), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(1273), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1369), 23,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [132494] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(4552), 14,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(4554), 26,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [132542] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1336), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      sym__special_character,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1338), 26,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [132590] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1352), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1354), 26,
      sym_test_operator,
      sym_extglob_pattern,
      anon_sym_RPAREN_RPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      anon_sym_COLON,
  [132638] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1348), 13,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1350), 27,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [132686] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1308), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      aux_sym__simple_variable_name_token1,
      sym_word,
    ACTIONS(1310), 26,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [132734] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1304), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      aux_sym__simple_variable_name_token1,
      sym_word,
    ACTIONS(1306), 26,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [132782] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1308), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      aux_sym__simple_variable_name_token1,
      sym_word,
    ACTIONS(1310), 26,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [132830] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1308), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1310), 26,
      sym_test_operator,
      sym_extglob_pattern,
      anon_sym_RPAREN_RPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      anon_sym_COLON,
  [132878] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1304), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1306), 26,
      sym_test_operator,
      sym_extglob_pattern,
      anon_sym_RPAREN_RPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      anon_sym_COLON,
  [132926] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1308), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1310), 26,
      sym_test_operator,
      sym_extglob_pattern,
      anon_sym_RPAREN_RPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      anon_sym_COLON,
  [132974] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6662), 1,
      sym__special_character,
    STATE(2568), 1,
      aux_sym__literal_repeat1,
    ACTIONS(4467), 12,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(4469), 26,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [133026] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1105), 1,
      aux_sym_concatenation_token1,
    ACTIONS(6717), 1,
      sym__concat,
    STATE(2511), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1286), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1288), 23,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [133080] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(4370), 14,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(4372), 26,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [133128] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1308), 13,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1310), 27,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [133176] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1356), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      sym__special_character,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1358), 26,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [133224] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(2723), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1143), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(1261), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1263), 23,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      sym__special_character,
  [133276] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(2564), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(6656), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(5476), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      aux_sym__simple_variable_name_token1,
      sym_word,
    ACTIONS(5478), 23,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [133328] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(2553), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(6656), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(5468), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      aux_sym__simple_variable_name_token1,
      sym_word,
    ACTIONS(5470), 23,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [133380] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1105), 1,
      aux_sym_concatenation_token1,
    ACTIONS(6719), 1,
      sym__concat,
    STATE(2511), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1265), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1267), 23,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [133434] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6662), 1,
      sym__special_character,
    STATE(2568), 1,
      aux_sym__literal_repeat1,
    ACTIONS(5695), 12,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(5697), 26,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [133486] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1304), 13,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1306), 27,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [133534] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6664), 1,
      aux_sym_concatenation_token1,
    ACTIONS(6721), 1,
      sym__concat,
    STATE(2483), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1286), 13,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1288), 24,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [133588] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1292), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      aux_sym__simple_variable_name_token1,
      sym_word,
    ACTIONS(1294), 26,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [133636] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1332), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      aux_sym__simple_variable_name_token1,
      sym_word,
    ACTIONS(1334), 26,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [133684] = 7,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1105), 1,
      aux_sym_concatenation_token1,
    ACTIONS(6723), 1,
      anon_sym_RBRACK,
    ACTIONS(6725), 1,
      sym__concat,
    STATE(2533), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1273), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1369), 22,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [133740] = 7,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1123), 1,
      anon_sym_RBRACK,
    ACTIONS(6652), 1,
      sym__special_character,
    ACTIONS(6727), 1,
      sym__concat,
    STATE(2474), 1,
      aux_sym__literal_repeat1,
    ACTIONS(243), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(354), 22,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [133796] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(2548), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(6729), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(1251), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      aux_sym__simple_variable_name_token1,
      sym_word,
    ACTIONS(1253), 23,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [133848] = 7,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1105), 1,
      aux_sym_concatenation_token1,
    ACTIONS(6732), 1,
      anon_sym_RBRACK,
    ACTIONS(6734), 1,
      sym__concat,
    STATE(2533), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1273), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1369), 22,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [133904] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6664), 1,
      aux_sym_concatenation_token1,
    ACTIONS(6736), 1,
      sym__concat,
    STATE(2483), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1265), 13,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1267), 24,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [133958] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1105), 1,
      aux_sym_concatenation_token1,
    ACTIONS(6738), 1,
      sym__concat,
    STATE(2511), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1286), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1288), 23,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [134012] = 7,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1119), 1,
      anon_sym_RBRACK,
    ACTIONS(6652), 1,
      sym__special_character,
    ACTIONS(6740), 1,
      sym__concat,
    STATE(2474), 1,
      aux_sym__literal_repeat1,
    ACTIONS(243), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(354), 22,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [134068] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6656), 1,
      aux_sym_concatenation_token1,
    ACTIONS(6742), 1,
      sym__concat,
    STATE(2548), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1286), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      aux_sym__simple_variable_name_token1,
      sym_word,
    ACTIONS(1288), 23,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [134122] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1308), 13,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1310), 27,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [134170] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1105), 1,
      aux_sym_concatenation_token1,
    ACTIONS(6744), 1,
      sym__concat,
    STATE(2511), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1265), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1267), 23,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [134224] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(4463), 14,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(4465), 26,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [134272] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6746), 1,
      sym__special_character,
    STATE(2557), 1,
      aux_sym__literal_repeat1,
    ACTIONS(1362), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1364), 24,
      sym_test_operator,
      anon_sym_RPAREN_RPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      anon_sym_COLON,
  [134324] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(4530), 14,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(4532), 26,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [134372] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6753), 1,
      sym_extglob_pattern,
    ACTIONS(6751), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(6749), 25,
      sym_test_operator,
      anon_sym_RPAREN_RPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      anon_sym_COLON,
  [134422] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1316), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      sym__special_character,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1318), 26,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [134470] = 7,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1097), 1,
      anon_sym_RBRACK,
    ACTIONS(6652), 1,
      sym__special_character,
    ACTIONS(6755), 1,
      sym__concat,
    STATE(2474), 1,
      aux_sym__literal_repeat1,
    ACTIONS(243), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(354), 22,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [134526] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(5460), 14,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_EQ_TILDE,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(5462), 26,
      sym_file_descriptor,
      sym_test_operator,
      sym__bare_dollar,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [134574] = 7,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1105), 1,
      aux_sym_concatenation_token1,
    ACTIONS(6757), 1,
      anon_sym_RBRACK,
    ACTIONS(6759), 1,
      sym__concat,
    STATE(2533), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1273), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1369), 22,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [134630] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6656), 1,
      aux_sym_concatenation_token1,
    ACTIONS(6761), 1,
      sym__concat,
    STATE(2548), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1265), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      aux_sym__simple_variable_name_token1,
      sym_word,
    ACTIONS(1267), 23,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [134684] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1300), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      aux_sym__simple_variable_name_token1,
      sym_word,
    ACTIONS(1302), 26,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [134732] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1348), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      aux_sym__simple_variable_name_token1,
      sym_word,
    ACTIONS(1350), 26,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [134780] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1344), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      sym__special_character,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1346), 26,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [134828] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6763), 1,
      sym__special_character,
    STATE(2568), 1,
      aux_sym__literal_repeat1,
    ACTIONS(1362), 12,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(1364), 26,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [134880] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6662), 1,
      sym__special_character,
    STATE(2568), 1,
      aux_sym__literal_repeat1,
    ACTIONS(4564), 12,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(4566), 26,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [134932] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1308), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      sym__special_character,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1310), 26,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [134980] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1320), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      sym__special_character,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1322), 26,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [135028] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(6013), 1,
      aux_sym_concatenation_token1,
    ACTIONS(6015), 1,
      sym__concat,
    STATE(2471), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5391), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5389), 33,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [135082] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(2582), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1143), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(1273), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1369), 23,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      anon_sym_COLON,
  [135134] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1296), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      aux_sym__simple_variable_name_token1,
      sym_word,
    ACTIONS(1298), 26,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [135182] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1143), 1,
      aux_sym_concatenation_token1,
    ACTIONS(6766), 1,
      sym__concat,
    STATE(2312), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1286), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1288), 23,
      sym_test_operator,
      anon_sym_RPAREN_RPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [135236] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1328), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      sym__special_character,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1330), 26,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [135284] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1340), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      sym__special_character,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1342), 26,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [135332] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(6013), 1,
      aux_sym_concatenation_token1,
    ACTIONS(6015), 1,
      sym__concat,
    STATE(2470), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5395), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5393), 33,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [135386] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1324), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      aux_sym__simple_variable_name_token1,
      sym_word,
    ACTIONS(1326), 26,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [135434] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1251), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      aux_sym__simple_variable_name_token1,
      sym_word,
    ACTIONS(1253), 26,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [135482] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1352), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      sym__special_character,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1354), 26,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [135530] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1143), 1,
      aux_sym_concatenation_token1,
    ACTIONS(6768), 1,
      sym__concat,
    STATE(2312), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1286), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1288), 23,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      anon_sym_COLON,
  [135584] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1328), 13,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1330), 26,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [135631] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1261), 15,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      sym__special_character,
    ACTIONS(1263), 24,
      sym__concat,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [135678] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1348), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      sym__special_character,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1350), 25,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [135725] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1328), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1330), 25,
      sym__concat,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      aux_sym_concatenation_token1,
  [135772] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1324), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      sym__special_character,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1326), 25,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [135819] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(2720), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1143), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(1273), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1369), 22,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [135870] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6770), 1,
      anon_sym_LBRACK,
    ACTIONS(6600), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(6598), 24,
      anon_sym_RPAREN_RPAREN,
      anon_sym_COMMA,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      anon_sym_COLON,
  [135919] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1251), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      sym__special_character,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1253), 25,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [135966] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1296), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      sym__special_character,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1298), 25,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [136013] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1300), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      sym__special_character,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1302), 25,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [136060] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1340), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1342), 25,
      sym__concat,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      aux_sym_concatenation_token1,
  [136107] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1273), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1369), 25,
      sym_test_operator,
      anon_sym_RPAREN_RPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      anon_sym_COLON,
  [136154] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6772), 1,
      sym__special_character,
    STATE(2595), 1,
      aux_sym__literal_repeat1,
    ACTIONS(1362), 12,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(1364), 25,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [136205] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1308), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1310), 25,
      sym__concat,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      aux_sym_concatenation_token1,
  [136252] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1304), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1306), 25,
      sym__concat,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      aux_sym_concatenation_token1,
  [136299] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1308), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1310), 25,
      sym__concat,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      aux_sym_concatenation_token1,
  [136346] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1332), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      sym__special_character,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1334), 25,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [136393] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1292), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      sym__special_character,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1294), 25,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [136440] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6775), 1,
      sym_extglob_pattern,
    ACTIONS(6751), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(6749), 24,
      sym__concat,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [136489] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6777), 1,
      sym__special_character,
    STATE(2602), 1,
      aux_sym__literal_repeat1,
    ACTIONS(1362), 12,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(1364), 25,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [136540] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1352), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1354), 25,
      sym__concat,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      aux_sym_concatenation_token1,
  [136587] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1316), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      sym__special_character,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1318), 25,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [136634] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1320), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      sym__special_character,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1322), 25,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [136681] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1328), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      sym__special_character,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1330), 25,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [136728] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1340), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      sym__special_character,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1342), 25,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [136775] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1352), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      sym__special_character,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1354), 25,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [136822] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1356), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      sym__special_character,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1358), 25,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [136869] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1344), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      sym__special_character,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1346), 25,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [136916] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1336), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      sym__special_character,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1338), 25,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [136963] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(2618), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(6780), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(5393), 13,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(5395), 23,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [137014] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(2619), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(6780), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(5389), 13,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(5391), 23,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [137065] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1312), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      sym__special_character,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1314), 25,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [137112] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1308), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      sym__special_character,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1310), 25,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [137159] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1304), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      sym__special_character,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1306), 25,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [137206] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1308), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      sym__special_character,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1310), 25,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [137253] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6780), 1,
      aux_sym_concatenation_token1,
    ACTIONS(6782), 1,
      sym__concat,
    STATE(2620), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1265), 13,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1267), 23,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [137306] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6780), 1,
      aux_sym_concatenation_token1,
    ACTIONS(6784), 1,
      sym__concat,
    STATE(2620), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1286), 13,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1288), 23,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [137359] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(2620), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(6786), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(1251), 13,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1253), 23,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [137410] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6791), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(6789), 23,
      sym_test_operator,
      anon_sym_RPAREN_RPAREN,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      anon_sym_COLON,
  [137459] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(2072), 13,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      sym__special_character,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(2074), 26,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [137506] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(4552), 13,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      sym__special_character,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(4554), 26,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [137553] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1356), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1358), 25,
      sym__concat,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      aux_sym_concatenation_token1,
  [137600] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1344), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1346), 25,
      sym__concat,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      aux_sym_concatenation_token1,
  [137647] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1308), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      aux_sym__simple_variable_name_token1,
      sym_word,
    ACTIONS(1310), 25,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [137694] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6797), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(6795), 25,
      sym_test_operator,
      anon_sym_RPAREN_RPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      anon_sym_COLON,
  [137741] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1304), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      aux_sym__simple_variable_name_token1,
      sym_word,
    ACTIONS(1306), 25,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [137788] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1308), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      aux_sym__simple_variable_name_token1,
      sym_word,
    ACTIONS(1310), 25,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [137835] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1336), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1338), 25,
      sym__concat,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      aux_sym_concatenation_token1,
  [137882] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1316), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1318), 25,
      sym__concat,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      aux_sym_concatenation_token1,
  [137929] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(2618), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(6780), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(1261), 13,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1263), 23,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [137980] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1324), 13,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1326), 26,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [138027] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1251), 13,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1253), 26,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [138074] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(4370), 13,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      sym__special_character,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(4372), 26,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [138121] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6801), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(6799), 25,
      sym_test_operator,
      anon_sym_RPAREN_RPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      anon_sym_COLON,
  [138168] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(2094), 13,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      sym__special_character,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(2096), 26,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [138215] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1348), 13,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1350), 26,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [138262] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1296), 13,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1298), 26,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [138309] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1300), 13,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1302), 26,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [138356] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6803), 1,
      sym__special_character,
    STATE(2727), 1,
      aux_sym__literal_repeat1,
    ACTIONS(243), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(354), 23,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [138407] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1312), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      aux_sym__simple_variable_name_token1,
      sym_word,
    ACTIONS(1314), 25,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [138454] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1332), 13,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1334), 26,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [138501] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1292), 13,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1294), 26,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [138548] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1336), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      aux_sym__simple_variable_name_token1,
      sym_word,
    ACTIONS(1338), 25,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [138595] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1344), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      aux_sym__simple_variable_name_token1,
      sym_word,
    ACTIONS(1346), 25,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [138642] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1312), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1314), 25,
      sym__concat,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      aux_sym_concatenation_token1,
  [138689] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6807), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(6805), 25,
      sym_test_operator,
      anon_sym_RPAREN_RPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      anon_sym_COLON,
  [138736] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1356), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      aux_sym__simple_variable_name_token1,
      sym_word,
    ACTIONS(1358), 25,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [138783] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1352), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      aux_sym__simple_variable_name_token1,
      sym_word,
    ACTIONS(1354), 25,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [138830] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1316), 13,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1318), 26,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [138877] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1340), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      aux_sym__simple_variable_name_token1,
      sym_word,
    ACTIONS(1342), 25,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [138924] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1356), 13,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1358), 26,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [138971] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1312), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1314), 25,
      sym__concat,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      aux_sym_concatenation_token1,
  [139018] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1344), 13,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1346), 26,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [139065] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1320), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1322), 25,
      sym__concat,
      sym_test_operator,
      sym_extglob_pattern,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [139112] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1336), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1338), 25,
      sym__concat,
      sym_test_operator,
      sym_extglob_pattern,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [139159] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1340), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1342), 25,
      sym__concat,
      sym_test_operator,
      sym_extglob_pattern,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [139206] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6807), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(6805), 25,
      sym_test_operator,
      anon_sym_RPAREN_RPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      anon_sym_COLON,
  [139253] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1312), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1314), 25,
      sym__concat,
      sym_test_operator,
      sym_extglob_pattern,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [139300] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1336), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1338), 25,
      sym__concat,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      aux_sym_concatenation_token1,
  [139347] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1352), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1354), 25,
      sym__concat,
      sym_test_operator,
      sym_extglob_pattern,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [139394] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1328), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      aux_sym__simple_variable_name_token1,
      sym_word,
    ACTIONS(1330), 25,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [139441] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1344), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1346), 25,
      sym__concat,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      aux_sym_concatenation_token1,
  [139488] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1356), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1358), 25,
      sym__concat,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      aux_sym_concatenation_token1,
  [139535] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1320), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      aux_sym__simple_variable_name_token1,
      sym_word,
    ACTIONS(1322), 25,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [139582] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1316), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      aux_sym__simple_variable_name_token1,
      sym_word,
    ACTIONS(1318), 25,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [139629] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1308), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1310), 25,
      sym__concat,
      sym_test_operator,
      sym_extglob_pattern,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [139676] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1304), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1306), 25,
      sym__concat,
      sym_test_operator,
      sym_extglob_pattern,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [139723] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1308), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1310), 25,
      sym__concat,
      sym_test_operator,
      sym_extglob_pattern,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [139770] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6809), 1,
      sym__special_character,
    STATE(2595), 1,
      aux_sym__literal_repeat1,
    ACTIONS(5393), 12,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(5395), 25,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [139821] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1320), 13,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1322), 26,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [139868] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1352), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1354), 25,
      sym__concat,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      aux_sym_concatenation_token1,
  [139915] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1336), 13,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1338), 26,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [139962] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1292), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1294), 25,
      sym__concat,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      aux_sym_concatenation_token1,
  [140009] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1340), 13,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1342), 26,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [140056] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1312), 13,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1314), 26,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [140103] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6809), 1,
      sym__special_character,
    STATE(2595), 1,
      aux_sym__literal_repeat1,
    ACTIONS(5067), 12,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(5069), 25,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [140154] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1316), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1318), 25,
      sym__concat,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      aux_sym_concatenation_token1,
  [140201] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1340), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1342), 25,
      sym__concat,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      aux_sym_concatenation_token1,
  [140248] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1332), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1334), 25,
      sym__concat,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      aux_sym_concatenation_token1,
  [140295] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1352), 13,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1354), 26,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [140342] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1292), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      aux_sym__simple_variable_name_token1,
      sym_word,
    ACTIONS(1294), 25,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [140389] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1308), 13,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1310), 26,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [140436] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1332), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      aux_sym__simple_variable_name_token1,
      sym_word,
    ACTIONS(1334), 25,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [140483] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1304), 13,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1306), 26,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [140530] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1308), 13,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1310), 26,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [140577] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6811), 1,
      sym__special_character,
    STATE(2692), 1,
      aux_sym__literal_repeat1,
    ACTIONS(5178), 13,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
    ACTIONS(5180), 24,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [140628] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6809), 1,
      sym__special_character,
    STATE(2595), 1,
      aux_sym__literal_repeat1,
    ACTIONS(4564), 12,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(4566), 25,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [140679] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6813), 1,
      sym__special_character,
    STATE(2695), 1,
      aux_sym__literal_repeat1,
    ACTIONS(243), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(354), 23,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [140730] = 7,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(5821), 2,
      anon_sym_PIPE,
      anon_sym_LT_LT,
    ACTIONS(5823), 5,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(5826), 5,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_LT_LT_DASH,
    ACTIONS(5819), 6,
      anon_sym_DOLLAR,
      sym__special_character,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(5828), 7,
      sym_file_descriptor,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
    ACTIONS(5831), 14,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [140785] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6815), 1,
      sym__special_character,
    STATE(2692), 1,
      aux_sym__literal_repeat1,
    ACTIONS(1362), 13,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
    ACTIONS(1364), 24,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [140836] = 7,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(5821), 2,
      anon_sym_PIPE,
      anon_sym_LT_LT,
    ACTIONS(5823), 5,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(5826), 5,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_LT_LT_DASH,
    ACTIONS(5819), 6,
      anon_sym_DOLLAR,
      sym__special_character,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(5828), 7,
      sym_file_descriptor,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
    ACTIONS(5831), 14,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [140891] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6822), 1,
      sym__concat,
    ACTIONS(6820), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(6818), 24,
      anon_sym_RPAREN_RPAREN,
      anon_sym_COMMA,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      anon_sym_COLON,
  [140940] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6824), 1,
      sym__special_character,
    STATE(2695), 1,
      aux_sym__literal_repeat1,
    ACTIONS(1362), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1364), 23,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [140991] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6831), 1,
      sym__concat,
    ACTIONS(6829), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(6827), 24,
      anon_sym_RPAREN_RPAREN,
      anon_sym_COMMA,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      anon_sym_COLON,
  [141040] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1328), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1330), 25,
      sym__concat,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      aux_sym_concatenation_token1,
  [141087] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1320), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1322), 25,
      sym__concat,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      aux_sym_concatenation_token1,
  [141134] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6833), 1,
      sym__special_character,
    STATE(2699), 1,
      aux_sym__literal_repeat1,
    ACTIONS(1362), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1364), 23,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [141185] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1300), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      aux_sym__simple_variable_name_token1,
      sym_word,
    ACTIONS(1302), 25,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [141232] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1296), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      aux_sym__simple_variable_name_token1,
      sym_word,
    ACTIONS(1298), 25,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [141279] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6840), 1,
      sym__concat,
    ACTIONS(6838), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(6836), 24,
      anon_sym_RPAREN_RPAREN,
      anon_sym_COMMA,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      anon_sym_COLON,
  [141328] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6809), 1,
      sym__special_character,
    STATE(2595), 1,
      aux_sym__literal_repeat1,
    ACTIONS(4467), 12,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(4469), 25,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [141379] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6846), 1,
      sym__concat,
    ACTIONS(6844), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(6842), 24,
      anon_sym_RPAREN_RPAREN,
      anon_sym_COMMA,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      anon_sym_COLON,
  [141428] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1241), 1,
      sym_file_descriptor,
    ACTIONS(2867), 1,
      anon_sym_DQUOTE,
    ACTIONS(6852), 1,
      sym_variable_name,
    STATE(4263), 1,
      sym_string,
    ACTIONS(6850), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(6848), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1239), 24,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
      anon_sym_LT_LT_LT,
  [141485] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1235), 1,
      sym_file_descriptor,
    ACTIONS(2867), 1,
      anon_sym_DQUOTE,
    ACTIONS(6852), 1,
      sym_variable_name,
    STATE(4263), 1,
      sym_string,
    ACTIONS(6850), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(6848), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1227), 24,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
      anon_sym_LT_LT_LT,
  [141542] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1251), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      aux_sym__simple_variable_name_token1,
      sym_word,
    ACTIONS(1253), 25,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [141589] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1308), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1310), 25,
      sym__concat,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      aux_sym_concatenation_token1,
  [141636] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1304), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1306), 25,
      sym__concat,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      aux_sym_concatenation_token1,
  [141683] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1308), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1310), 25,
      sym__concat,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      aux_sym_concatenation_token1,
  [141730] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1324), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      aux_sym__simple_variable_name_token1,
      sym_word,
    ACTIONS(1326), 25,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [141777] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1348), 14,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      aux_sym__simple_variable_name_token1,
      sym_word,
    ACTIONS(1350), 25,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [141824] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1261), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1263), 25,
      sym_test_operator,
      anon_sym_RPAREN_RPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      anon_sym_COLON,
      sym__special_character,
  [141871] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6811), 1,
      sym__special_character,
    STATE(2692), 1,
      aux_sym__literal_repeat1,
    ACTIONS(5067), 13,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
    ACTIONS(5069), 24,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [141922] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6854), 1,
      sym__special_character,
    STATE(2602), 1,
      aux_sym__literal_repeat1,
    ACTIONS(5393), 12,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(5395), 25,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [141973] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1348), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1350), 25,
      sym__concat,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      aux_sym_concatenation_token1,
  [142020] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6856), 1,
      sym__special_character,
    STATE(2699), 1,
      aux_sym__literal_repeat1,
    ACTIONS(243), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(354), 23,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [142071] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6809), 1,
      sym__special_character,
    STATE(2595), 1,
      aux_sym__literal_repeat1,
    ACTIONS(5721), 12,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(5723), 25,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [142122] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1300), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1302), 25,
      sym__concat,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      aux_sym_concatenation_token1,
  [142169] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1143), 1,
      aux_sym_concatenation_token1,
    ACTIONS(6858), 1,
      sym__concat,
    STATE(2312), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1286), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1288), 22,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [142222] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1292), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1294), 25,
      sym__concat,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      aux_sym_concatenation_token1,
  [142269] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1296), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1298), 25,
      sym__concat,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      aux_sym_concatenation_token1,
  [142316] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1143), 1,
      aux_sym_concatenation_token1,
    ACTIONS(6860), 1,
      sym__concat,
    STATE(2312), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1265), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1267), 22,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [142369] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6864), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(6862), 25,
      sym_test_operator,
      anon_sym_RPAREN_RPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      anon_sym_COLON,
  [142416] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1261), 13,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      sym__special_character,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(1263), 26,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [142463] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6809), 1,
      sym__special_character,
    STATE(2595), 1,
      aux_sym__literal_repeat1,
    ACTIONS(5695), 12,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(5697), 25,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [142514] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6866), 1,
      sym__special_character,
    STATE(2727), 1,
      aux_sym__literal_repeat1,
    ACTIONS(1362), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1364), 23,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [142565] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1320), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1322), 25,
      sym__concat,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      aux_sym_concatenation_token1,
  [142612] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1324), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1326), 25,
      sym__concat,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      aux_sym_concatenation_token1,
  [142659] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1251), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1253), 25,
      sym__concat,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      aux_sym_concatenation_token1,
  [142706] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1296), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1298), 25,
      sym__concat,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      aux_sym_concatenation_token1,
  [142753] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1251), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1253), 25,
      sym__concat,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      aux_sym_concatenation_token1,
  [142800] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1324), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1326), 25,
      sym__concat,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      aux_sym_concatenation_token1,
  [142847] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1348), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1350), 25,
      sym__concat,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      aux_sym_concatenation_token1,
  [142894] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1300), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1302), 25,
      sym__concat,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      aux_sym_concatenation_token1,
  [142941] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1332), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1334), 25,
      sym__concat,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      aux_sym_concatenation_token1,
  [142988] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1336), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1338), 24,
      sym_test_operator,
      sym_extglob_pattern,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [143034] = 8,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6877), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6869), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6871), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(6873), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(6875), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6791), 6,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(6789), 22,
      sym__concat,
      sym_test_operator,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [143090] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6879), 1,
      anon_sym_RPAREN_RPAREN,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(6883), 1,
      anon_sym_EQ,
    ACTIONS(6889), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(6891), 1,
      anon_sym_AMP_AMP,
    ACTIONS(6893), 1,
      anon_sym_PIPE,
    ACTIONS(6895), 1,
      anon_sym_CARET,
    ACTIONS(6897), 1,
      anon_sym_AMP,
    ACTIONS(6911), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6913), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(6915), 1,
      anon_sym_QMARK,
    STATE(6721), 1,
      aux_sym_arithmetic_expansion_repeat1,
    ACTIONS(6885), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6899), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(6901), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(6903), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(6905), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(6907), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(6909), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6887), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [143172] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(6883), 1,
      anon_sym_EQ,
    ACTIONS(6889), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(6891), 1,
      anon_sym_AMP_AMP,
    ACTIONS(6893), 1,
      anon_sym_PIPE,
    ACTIONS(6895), 1,
      anon_sym_CARET,
    ACTIONS(6897), 1,
      anon_sym_AMP,
    ACTIONS(6911), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6913), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(6915), 1,
      anon_sym_QMARK,
    ACTIONS(6917), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6687), 1,
      aux_sym_arithmetic_expansion_repeat1,
    ACTIONS(6885), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6899), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(6901), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(6903), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(6905), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(6907), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(6909), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6887), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [143254] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(6883), 1,
      anon_sym_EQ,
    ACTIONS(6889), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(6891), 1,
      anon_sym_AMP_AMP,
    ACTIONS(6893), 1,
      anon_sym_PIPE,
    ACTIONS(6895), 1,
      anon_sym_CARET,
    ACTIONS(6897), 1,
      anon_sym_AMP,
    ACTIONS(6911), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6913), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(6915), 1,
      anon_sym_QMARK,
    ACTIONS(6919), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6625), 1,
      aux_sym_arithmetic_expansion_repeat1,
    ACTIONS(6885), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6899), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(6901), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(6903), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(6905), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(6907), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(6909), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6887), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [143336] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(6883), 1,
      anon_sym_EQ,
    ACTIONS(6889), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(6891), 1,
      anon_sym_AMP_AMP,
    ACTIONS(6893), 1,
      anon_sym_PIPE,
    ACTIONS(6895), 1,
      anon_sym_CARET,
    ACTIONS(6897), 1,
      anon_sym_AMP,
    ACTIONS(6911), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6913), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(6915), 1,
      anon_sym_QMARK,
    ACTIONS(6921), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6602), 1,
      aux_sym_arithmetic_expansion_repeat1,
    ACTIONS(6885), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6899), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(6901), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(6903), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(6905), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(6907), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(6909), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6887), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [143418] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(6883), 1,
      anon_sym_EQ,
    ACTIONS(6889), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(6891), 1,
      anon_sym_AMP_AMP,
    ACTIONS(6893), 1,
      anon_sym_PIPE,
    ACTIONS(6895), 1,
      anon_sym_CARET,
    ACTIONS(6897), 1,
      anon_sym_AMP,
    ACTIONS(6911), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6913), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(6915), 1,
      anon_sym_QMARK,
    ACTIONS(6923), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6630), 1,
      aux_sym_arithmetic_expansion_repeat1,
    ACTIONS(6885), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6899), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(6901), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(6903), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(6905), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(6907), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(6909), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6887), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [143500] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(6883), 1,
      anon_sym_EQ,
    ACTIONS(6889), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(6891), 1,
      anon_sym_AMP_AMP,
    ACTIONS(6893), 1,
      anon_sym_PIPE,
    ACTIONS(6895), 1,
      anon_sym_CARET,
    ACTIONS(6897), 1,
      anon_sym_AMP,
    ACTIONS(6911), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6913), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(6915), 1,
      anon_sym_QMARK,
    ACTIONS(6925), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6720), 1,
      aux_sym_arithmetic_expansion_repeat1,
    ACTIONS(6885), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6899), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(6901), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(6903), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(6905), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(6907), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(6909), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6887), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [143582] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(6883), 1,
      anon_sym_EQ,
    ACTIONS(6889), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(6891), 1,
      anon_sym_AMP_AMP,
    ACTIONS(6893), 1,
      anon_sym_PIPE,
    ACTIONS(6895), 1,
      anon_sym_CARET,
    ACTIONS(6897), 1,
      anon_sym_AMP,
    ACTIONS(6911), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6913), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(6915), 1,
      anon_sym_QMARK,
    ACTIONS(6927), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6682), 1,
      aux_sym_arithmetic_expansion_repeat1,
    ACTIONS(6885), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6899), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(6901), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(6903), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(6905), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(6907), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(6909), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6887), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [143664] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(6883), 1,
      anon_sym_EQ,
    ACTIONS(6889), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(6891), 1,
      anon_sym_AMP_AMP,
    ACTIONS(6893), 1,
      anon_sym_PIPE,
    ACTIONS(6895), 1,
      anon_sym_CARET,
    ACTIONS(6897), 1,
      anon_sym_AMP,
    ACTIONS(6911), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6913), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(6915), 1,
      anon_sym_QMARK,
    ACTIONS(6929), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6655), 1,
      aux_sym_arithmetic_expansion_repeat1,
    ACTIONS(6885), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6899), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(6901), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(6903), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(6905), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(6907), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(6909), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6887), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [143746] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(6883), 1,
      anon_sym_EQ,
    ACTIONS(6889), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(6891), 1,
      anon_sym_AMP_AMP,
    ACTIONS(6893), 1,
      anon_sym_PIPE,
    ACTIONS(6895), 1,
      anon_sym_CARET,
    ACTIONS(6897), 1,
      anon_sym_AMP,
    ACTIONS(6911), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6913), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(6915), 1,
      anon_sym_QMARK,
    ACTIONS(6931), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6623), 1,
      aux_sym_arithmetic_expansion_repeat1,
    ACTIONS(6885), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6899), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(6901), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(6903), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(6905), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(6907), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(6909), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6887), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [143828] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(6883), 1,
      anon_sym_EQ,
    ACTIONS(6889), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(6891), 1,
      anon_sym_AMP_AMP,
    ACTIONS(6893), 1,
      anon_sym_PIPE,
    ACTIONS(6895), 1,
      anon_sym_CARET,
    ACTIONS(6897), 1,
      anon_sym_AMP,
    ACTIONS(6911), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6913), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(6915), 1,
      anon_sym_QMARK,
    ACTIONS(6933), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6608), 1,
      aux_sym_arithmetic_expansion_repeat1,
    ACTIONS(6885), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6899), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(6901), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(6903), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(6905), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(6907), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(6909), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6887), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [143910] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(6883), 1,
      anon_sym_EQ,
    ACTIONS(6889), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(6891), 1,
      anon_sym_AMP_AMP,
    ACTIONS(6893), 1,
      anon_sym_PIPE,
    ACTIONS(6895), 1,
      anon_sym_CARET,
    ACTIONS(6897), 1,
      anon_sym_AMP,
    ACTIONS(6911), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6913), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(6915), 1,
      anon_sym_QMARK,
    ACTIONS(6935), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6592), 1,
      aux_sym_arithmetic_expansion_repeat1,
    ACTIONS(6885), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6899), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(6901), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(6903), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(6905), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(6907), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(6909), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6887), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [143992] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(6883), 1,
      anon_sym_EQ,
    ACTIONS(6889), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(6891), 1,
      anon_sym_AMP_AMP,
    ACTIONS(6893), 1,
      anon_sym_PIPE,
    ACTIONS(6895), 1,
      anon_sym_CARET,
    ACTIONS(6897), 1,
      anon_sym_AMP,
    ACTIONS(6911), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6913), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(6915), 1,
      anon_sym_QMARK,
    ACTIONS(6937), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6577), 1,
      aux_sym_arithmetic_expansion_repeat1,
    ACTIONS(6885), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6899), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(6901), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(6903), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(6905), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(6907), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(6909), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6887), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [144074] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(6883), 1,
      anon_sym_EQ,
    ACTIONS(6889), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(6891), 1,
      anon_sym_AMP_AMP,
    ACTIONS(6893), 1,
      anon_sym_PIPE,
    ACTIONS(6895), 1,
      anon_sym_CARET,
    ACTIONS(6897), 1,
      anon_sym_AMP,
    ACTIONS(6911), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6913), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(6915), 1,
      anon_sym_QMARK,
    ACTIONS(6939), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6559), 1,
      aux_sym_arithmetic_expansion_repeat1,
    ACTIONS(6885), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6899), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(6901), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(6903), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(6905), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(6907), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(6909), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6887), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [144156] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(6883), 1,
      anon_sym_EQ,
    ACTIONS(6889), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(6891), 1,
      anon_sym_AMP_AMP,
    ACTIONS(6893), 1,
      anon_sym_PIPE,
    ACTIONS(6895), 1,
      anon_sym_CARET,
    ACTIONS(6897), 1,
      anon_sym_AMP,
    ACTIONS(6911), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6913), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(6915), 1,
      anon_sym_QMARK,
    ACTIONS(6941), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6712), 1,
      aux_sym_arithmetic_expansion_repeat1,
    ACTIONS(6885), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6899), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(6901), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(6903), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(6905), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(6907), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(6909), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6887), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [144238] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(6883), 1,
      anon_sym_EQ,
    ACTIONS(6889), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(6891), 1,
      anon_sym_AMP_AMP,
    ACTIONS(6893), 1,
      anon_sym_PIPE,
    ACTIONS(6895), 1,
      anon_sym_CARET,
    ACTIONS(6897), 1,
      anon_sym_AMP,
    ACTIONS(6911), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6913), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(6915), 1,
      anon_sym_QMARK,
    ACTIONS(6943), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6606), 1,
      aux_sym_arithmetic_expansion_repeat1,
    ACTIONS(6885), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6899), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(6901), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(6903), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(6905), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(6907), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(6909), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6887), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [144320] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(6883), 1,
      anon_sym_EQ,
    ACTIONS(6889), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(6891), 1,
      anon_sym_AMP_AMP,
    ACTIONS(6893), 1,
      anon_sym_PIPE,
    ACTIONS(6895), 1,
      anon_sym_CARET,
    ACTIONS(6897), 1,
      anon_sym_AMP,
    ACTIONS(6911), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6913), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(6915), 1,
      anon_sym_QMARK,
    ACTIONS(6945), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6697), 1,
      aux_sym_arithmetic_expansion_repeat1,
    ACTIONS(6885), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6899), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(6901), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(6903), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(6905), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(6907), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(6909), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6887), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [144402] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(6883), 1,
      anon_sym_EQ,
    ACTIONS(6889), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(6891), 1,
      anon_sym_AMP_AMP,
    ACTIONS(6893), 1,
      anon_sym_PIPE,
    ACTIONS(6895), 1,
      anon_sym_CARET,
    ACTIONS(6897), 1,
      anon_sym_AMP,
    ACTIONS(6911), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6913), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(6915), 1,
      anon_sym_QMARK,
    ACTIONS(6947), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6653), 1,
      aux_sym_arithmetic_expansion_repeat1,
    ACTIONS(6885), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6899), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(6901), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(6903), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(6905), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(6907), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(6909), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6887), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [144484] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(6883), 1,
      anon_sym_EQ,
    ACTIONS(6889), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(6891), 1,
      anon_sym_AMP_AMP,
    ACTIONS(6893), 1,
      anon_sym_PIPE,
    ACTIONS(6895), 1,
      anon_sym_CARET,
    ACTIONS(6897), 1,
      anon_sym_AMP,
    ACTIONS(6911), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6913), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(6915), 1,
      anon_sym_QMARK,
    ACTIONS(6949), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6612), 1,
      aux_sym_arithmetic_expansion_repeat1,
    ACTIONS(6885), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6899), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(6901), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(6903), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(6905), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(6907), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(6909), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6887), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [144566] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(6883), 1,
      anon_sym_EQ,
    ACTIONS(6889), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(6891), 1,
      anon_sym_AMP_AMP,
    ACTIONS(6893), 1,
      anon_sym_PIPE,
    ACTIONS(6895), 1,
      anon_sym_CARET,
    ACTIONS(6897), 1,
      anon_sym_AMP,
    ACTIONS(6911), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6913), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(6915), 1,
      anon_sym_QMARK,
    ACTIONS(6951), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6715), 1,
      aux_sym_arithmetic_expansion_repeat1,
    ACTIONS(6885), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6899), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(6901), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(6903), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(6905), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(6907), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(6909), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6887), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [144648] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(6883), 1,
      anon_sym_EQ,
    ACTIONS(6889), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(6891), 1,
      anon_sym_AMP_AMP,
    ACTIONS(6893), 1,
      anon_sym_PIPE,
    ACTIONS(6895), 1,
      anon_sym_CARET,
    ACTIONS(6897), 1,
      anon_sym_AMP,
    ACTIONS(6911), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6913), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(6915), 1,
      anon_sym_QMARK,
    ACTIONS(6953), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6639), 1,
      aux_sym_arithmetic_expansion_repeat1,
    ACTIONS(6885), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6899), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(6901), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(6903), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(6905), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(6907), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(6909), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6887), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [144730] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(6883), 1,
      anon_sym_EQ,
    ACTIONS(6889), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(6891), 1,
      anon_sym_AMP_AMP,
    ACTIONS(6893), 1,
      anon_sym_PIPE,
    ACTIONS(6895), 1,
      anon_sym_CARET,
    ACTIONS(6897), 1,
      anon_sym_AMP,
    ACTIONS(6911), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6913), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(6915), 1,
      anon_sym_QMARK,
    ACTIONS(6955), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6600), 1,
      aux_sym_arithmetic_expansion_repeat1,
    ACTIONS(6885), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6899), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(6901), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(6903), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(6905), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(6907), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(6909), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6887), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [144812] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(6883), 1,
      anon_sym_EQ,
    ACTIONS(6889), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(6891), 1,
      anon_sym_AMP_AMP,
    ACTIONS(6893), 1,
      anon_sym_PIPE,
    ACTIONS(6895), 1,
      anon_sym_CARET,
    ACTIONS(6897), 1,
      anon_sym_AMP,
    ACTIONS(6911), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6913), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(6915), 1,
      anon_sym_QMARK,
    ACTIONS(6957), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6562), 1,
      aux_sym_arithmetic_expansion_repeat1,
    ACTIONS(6885), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6899), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(6901), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(6903), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(6905), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(6907), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(6909), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6887), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [144894] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(6883), 1,
      anon_sym_EQ,
    ACTIONS(6889), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(6891), 1,
      anon_sym_AMP_AMP,
    ACTIONS(6893), 1,
      anon_sym_PIPE,
    ACTIONS(6895), 1,
      anon_sym_CARET,
    ACTIONS(6897), 1,
      anon_sym_AMP,
    ACTIONS(6911), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6913), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(6915), 1,
      anon_sym_QMARK,
    ACTIONS(6959), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6607), 1,
      aux_sym_arithmetic_expansion_repeat1,
    ACTIONS(6885), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6899), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(6901), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(6903), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(6905), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(6907), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(6909), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6887), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [144976] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(6883), 1,
      anon_sym_EQ,
    ACTIONS(6889), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(6891), 1,
      anon_sym_AMP_AMP,
    ACTIONS(6893), 1,
      anon_sym_PIPE,
    ACTIONS(6895), 1,
      anon_sym_CARET,
    ACTIONS(6897), 1,
      anon_sym_AMP,
    ACTIONS(6911), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6913), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(6915), 1,
      anon_sym_QMARK,
    ACTIONS(6961), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6701), 1,
      aux_sym_arithmetic_expansion_repeat1,
    ACTIONS(6885), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6899), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(6901), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(6903), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(6905), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(6907), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(6909), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6887), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [145058] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(6883), 1,
      anon_sym_EQ,
    ACTIONS(6889), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(6891), 1,
      anon_sym_AMP_AMP,
    ACTIONS(6893), 1,
      anon_sym_PIPE,
    ACTIONS(6895), 1,
      anon_sym_CARET,
    ACTIONS(6897), 1,
      anon_sym_AMP,
    ACTIONS(6911), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6913), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(6915), 1,
      anon_sym_QMARK,
    ACTIONS(6963), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6633), 1,
      aux_sym_arithmetic_expansion_repeat1,
    ACTIONS(6885), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6899), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(6901), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(6903), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(6905), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(6907), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(6909), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6887), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [145140] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(6883), 1,
      anon_sym_EQ,
    ACTIONS(6889), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(6891), 1,
      anon_sym_AMP_AMP,
    ACTIONS(6893), 1,
      anon_sym_PIPE,
    ACTIONS(6895), 1,
      anon_sym_CARET,
    ACTIONS(6897), 1,
      anon_sym_AMP,
    ACTIONS(6911), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6913), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(6915), 1,
      anon_sym_QMARK,
    ACTIONS(6965), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6660), 1,
      aux_sym_arithmetic_expansion_repeat1,
    ACTIONS(6885), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6899), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(6901), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(6903), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(6905), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(6907), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(6909), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6887), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [145222] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(6883), 1,
      anon_sym_EQ,
    ACTIONS(6889), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(6891), 1,
      anon_sym_AMP_AMP,
    ACTIONS(6893), 1,
      anon_sym_PIPE,
    ACTIONS(6895), 1,
      anon_sym_CARET,
    ACTIONS(6897), 1,
      anon_sym_AMP,
    ACTIONS(6911), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6913), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(6915), 1,
      anon_sym_QMARK,
    ACTIONS(6967), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6693), 1,
      aux_sym_arithmetic_expansion_repeat1,
    ACTIONS(6885), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6899), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(6901), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(6903), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(6905), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(6907), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(6909), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6887), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [145304] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(6883), 1,
      anon_sym_EQ,
    ACTIONS(6889), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(6891), 1,
      anon_sym_AMP_AMP,
    ACTIONS(6893), 1,
      anon_sym_PIPE,
    ACTIONS(6895), 1,
      anon_sym_CARET,
    ACTIONS(6897), 1,
      anon_sym_AMP,
    ACTIONS(6911), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6913), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(6915), 1,
      anon_sym_QMARK,
    ACTIONS(6969), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6586), 1,
      aux_sym_arithmetic_expansion_repeat1,
    ACTIONS(6885), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6899), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(6901), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(6903), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(6905), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(6907), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(6909), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6887), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [145386] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(6883), 1,
      anon_sym_EQ,
    ACTIONS(6889), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(6891), 1,
      anon_sym_AMP_AMP,
    ACTIONS(6893), 1,
      anon_sym_PIPE,
    ACTIONS(6895), 1,
      anon_sym_CARET,
    ACTIONS(6897), 1,
      anon_sym_AMP,
    ACTIONS(6911), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6913), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(6915), 1,
      anon_sym_QMARK,
    ACTIONS(6971), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6565), 1,
      aux_sym_arithmetic_expansion_repeat1,
    ACTIONS(6885), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6899), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(6901), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(6903), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(6905), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(6907), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(6909), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6887), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [145468] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(6883), 1,
      anon_sym_EQ,
    ACTIONS(6889), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(6891), 1,
      anon_sym_AMP_AMP,
    ACTIONS(6893), 1,
      anon_sym_PIPE,
    ACTIONS(6895), 1,
      anon_sym_CARET,
    ACTIONS(6897), 1,
      anon_sym_AMP,
    ACTIONS(6911), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6913), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(6915), 1,
      anon_sym_QMARK,
    ACTIONS(6973), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6579), 1,
      aux_sym_arithmetic_expansion_repeat1,
    ACTIONS(6885), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6899), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(6901), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(6903), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(6905), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(6907), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(6909), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6887), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [145550] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(6883), 1,
      anon_sym_EQ,
    ACTIONS(6889), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(6891), 1,
      anon_sym_AMP_AMP,
    ACTIONS(6893), 1,
      anon_sym_PIPE,
    ACTIONS(6895), 1,
      anon_sym_CARET,
    ACTIONS(6897), 1,
      anon_sym_AMP,
    ACTIONS(6911), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6913), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(6915), 1,
      anon_sym_QMARK,
    ACTIONS(6975), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6601), 1,
      aux_sym_arithmetic_expansion_repeat1,
    ACTIONS(6885), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6899), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(6901), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(6903), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(6905), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(6907), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(6909), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6887), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [145632] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(6883), 1,
      anon_sym_EQ,
    ACTIONS(6889), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(6891), 1,
      anon_sym_AMP_AMP,
    ACTIONS(6893), 1,
      anon_sym_PIPE,
    ACTIONS(6895), 1,
      anon_sym_CARET,
    ACTIONS(6897), 1,
      anon_sym_AMP,
    ACTIONS(6911), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6913), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(6915), 1,
      anon_sym_QMARK,
    ACTIONS(6977), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6683), 1,
      aux_sym_arithmetic_expansion_repeat1,
    ACTIONS(6885), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6899), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(6901), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(6903), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(6905), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(6907), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(6909), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6887), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [145714] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(6883), 1,
      anon_sym_EQ,
    ACTIONS(6889), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(6891), 1,
      anon_sym_AMP_AMP,
    ACTIONS(6893), 1,
      anon_sym_PIPE,
    ACTIONS(6895), 1,
      anon_sym_CARET,
    ACTIONS(6897), 1,
      anon_sym_AMP,
    ACTIONS(6911), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6913), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(6915), 1,
      anon_sym_QMARK,
    ACTIONS(6979), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6723), 1,
      aux_sym_arithmetic_expansion_repeat1,
    ACTIONS(6885), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6899), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(6901), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(6903), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(6905), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(6907), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(6909), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6887), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [145796] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(6883), 1,
      anon_sym_EQ,
    ACTIONS(6889), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(6891), 1,
      anon_sym_AMP_AMP,
    ACTIONS(6893), 1,
      anon_sym_PIPE,
    ACTIONS(6895), 1,
      anon_sym_CARET,
    ACTIONS(6897), 1,
      anon_sym_AMP,
    ACTIONS(6911), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6913), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(6915), 1,
      anon_sym_QMARK,
    ACTIONS(6981), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6594), 1,
      aux_sym_arithmetic_expansion_repeat1,
    ACTIONS(6885), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6899), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(6901), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(6903), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(6905), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(6907), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(6909), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6887), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [145878] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(6883), 1,
      anon_sym_EQ,
    ACTIONS(6889), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(6891), 1,
      anon_sym_AMP_AMP,
    ACTIONS(6893), 1,
      anon_sym_PIPE,
    ACTIONS(6895), 1,
      anon_sym_CARET,
    ACTIONS(6897), 1,
      anon_sym_AMP,
    ACTIONS(6911), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6913), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(6915), 1,
      anon_sym_QMARK,
    ACTIONS(6983), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6708), 1,
      aux_sym_arithmetic_expansion_repeat1,
    ACTIONS(6885), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6899), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(6901), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(6903), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(6905), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(6907), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(6909), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6887), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [145960] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(6883), 1,
      anon_sym_EQ,
    ACTIONS(6889), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(6891), 1,
      anon_sym_AMP_AMP,
    ACTIONS(6893), 1,
      anon_sym_PIPE,
    ACTIONS(6895), 1,
      anon_sym_CARET,
    ACTIONS(6897), 1,
      anon_sym_AMP,
    ACTIONS(6911), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6913), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(6915), 1,
      anon_sym_QMARK,
    ACTIONS(6985), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6669), 1,
      aux_sym_arithmetic_expansion_repeat1,
    ACTIONS(6885), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6899), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(6901), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(6903), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(6905), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(6907), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(6909), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6887), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [146042] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(6883), 1,
      anon_sym_EQ,
    ACTIONS(6889), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(6891), 1,
      anon_sym_AMP_AMP,
    ACTIONS(6893), 1,
      anon_sym_PIPE,
    ACTIONS(6895), 1,
      anon_sym_CARET,
    ACTIONS(6897), 1,
      anon_sym_AMP,
    ACTIONS(6911), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6913), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(6915), 1,
      anon_sym_QMARK,
    ACTIONS(6987), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6647), 1,
      aux_sym_arithmetic_expansion_repeat1,
    ACTIONS(6885), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6899), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(6901), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(6903), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(6905), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(6907), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(6909), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6887), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [146124] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6989), 1,
      sym__special_character,
    STATE(2806), 1,
      aux_sym__literal_repeat1,
    ACTIONS(5393), 12,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(5395), 24,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [146174] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(6883), 1,
      anon_sym_EQ,
    ACTIONS(6889), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(6891), 1,
      anon_sym_AMP_AMP,
    ACTIONS(6893), 1,
      anon_sym_PIPE,
    ACTIONS(6895), 1,
      anon_sym_CARET,
    ACTIONS(6897), 1,
      anon_sym_AMP,
    ACTIONS(6911), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6913), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(6915), 1,
      anon_sym_QMARK,
    ACTIONS(6991), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6624), 1,
      aux_sym_arithmetic_expansion_repeat1,
    ACTIONS(6885), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6899), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(6901), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(6903), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(6905), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(6907), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(6909), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6887), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [146256] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(6883), 1,
      anon_sym_EQ,
    ACTIONS(6889), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(6891), 1,
      anon_sym_AMP_AMP,
    ACTIONS(6893), 1,
      anon_sym_PIPE,
    ACTIONS(6895), 1,
      anon_sym_CARET,
    ACTIONS(6897), 1,
      anon_sym_AMP,
    ACTIONS(6911), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6913), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(6915), 1,
      anon_sym_QMARK,
    ACTIONS(6993), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6576), 1,
      aux_sym_arithmetic_expansion_repeat1,
    ACTIONS(6885), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6899), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(6901), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(6903), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(6905), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(6907), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(6909), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6887), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [146338] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(6883), 1,
      anon_sym_EQ,
    ACTIONS(6889), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(6891), 1,
      anon_sym_AMP_AMP,
    ACTIONS(6893), 1,
      anon_sym_PIPE,
    ACTIONS(6895), 1,
      anon_sym_CARET,
    ACTIONS(6897), 1,
      anon_sym_AMP,
    ACTIONS(6911), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6913), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(6915), 1,
      anon_sym_QMARK,
    ACTIONS(6995), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6621), 1,
      aux_sym_arithmetic_expansion_repeat1,
    ACTIONS(6885), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6899), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(6901), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(6903), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(6905), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(6907), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(6909), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6887), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [146420] = 7,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(5821), 2,
      anon_sym_PIPE,
      anon_sym_LT_LT,
    ACTIONS(5826), 4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
    ACTIONS(5819), 5,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(5823), 5,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(5828), 7,
      sym_file_descriptor,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
    ACTIONS(5831), 15,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [146474] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(6883), 1,
      anon_sym_EQ,
    ACTIONS(6889), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(6891), 1,
      anon_sym_AMP_AMP,
    ACTIONS(6893), 1,
      anon_sym_PIPE,
    ACTIONS(6895), 1,
      anon_sym_CARET,
    ACTIONS(6897), 1,
      anon_sym_AMP,
    ACTIONS(6911), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6913), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(6915), 1,
      anon_sym_QMARK,
    ACTIONS(6997), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6685), 1,
      aux_sym_arithmetic_expansion_repeat1,
    ACTIONS(6885), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6899), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(6901), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(6903), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(6905), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(6907), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(6909), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6887), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [146556] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(6883), 1,
      anon_sym_EQ,
    ACTIONS(6889), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(6891), 1,
      anon_sym_AMP_AMP,
    ACTIONS(6893), 1,
      anon_sym_PIPE,
    ACTIONS(6895), 1,
      anon_sym_CARET,
    ACTIONS(6897), 1,
      anon_sym_AMP,
    ACTIONS(6911), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6913), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(6915), 1,
      anon_sym_QMARK,
    ACTIONS(6999), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6684), 1,
      aux_sym_arithmetic_expansion_repeat1,
    ACTIONS(6885), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6899), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(6901), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(6903), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(6905), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(6907), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(6909), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6887), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [146638] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(6883), 1,
      anon_sym_EQ,
    ACTIONS(6889), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(6891), 1,
      anon_sym_AMP_AMP,
    ACTIONS(6893), 1,
      anon_sym_PIPE,
    ACTIONS(6895), 1,
      anon_sym_CARET,
    ACTIONS(6897), 1,
      anon_sym_AMP,
    ACTIONS(6911), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6913), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(6915), 1,
      anon_sym_QMARK,
    ACTIONS(7001), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6644), 1,
      aux_sym_arithmetic_expansion_repeat1,
    ACTIONS(6885), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6899), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(6901), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(6903), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(6905), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(6907), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(6909), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6887), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [146720] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(6883), 1,
      anon_sym_EQ,
    ACTIONS(6889), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(6891), 1,
      anon_sym_AMP_AMP,
    ACTIONS(6893), 1,
      anon_sym_PIPE,
    ACTIONS(6895), 1,
      anon_sym_CARET,
    ACTIONS(6897), 1,
      anon_sym_AMP,
    ACTIONS(6911), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6913), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(6915), 1,
      anon_sym_QMARK,
    ACTIONS(7003), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6552), 1,
      aux_sym_arithmetic_expansion_repeat1,
    ACTIONS(6885), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6899), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(6901), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(6903), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(6905), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(6907), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(6909), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6887), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [146802] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(6883), 1,
      anon_sym_EQ,
    ACTIONS(6889), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(6891), 1,
      anon_sym_AMP_AMP,
    ACTIONS(6893), 1,
      anon_sym_PIPE,
    ACTIONS(6895), 1,
      anon_sym_CARET,
    ACTIONS(6897), 1,
      anon_sym_AMP,
    ACTIONS(6911), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6913), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(6915), 1,
      anon_sym_QMARK,
    ACTIONS(7005), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6680), 1,
      aux_sym_arithmetic_expansion_repeat1,
    ACTIONS(6885), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6899), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(6901), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(6903), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(6905), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(6907), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(6909), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6887), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [146884] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(6883), 1,
      anon_sym_EQ,
    ACTIONS(6889), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(6891), 1,
      anon_sym_AMP_AMP,
    ACTIONS(6893), 1,
      anon_sym_PIPE,
    ACTIONS(6895), 1,
      anon_sym_CARET,
    ACTIONS(6897), 1,
      anon_sym_AMP,
    ACTIONS(6911), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6913), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(6915), 1,
      anon_sym_QMARK,
    ACTIONS(7007), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6593), 1,
      aux_sym_arithmetic_expansion_repeat1,
    ACTIONS(6885), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6899), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(6901), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(6903), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(6905), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(6907), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(6909), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6887), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [146966] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(6883), 1,
      anon_sym_EQ,
    ACTIONS(6889), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(6891), 1,
      anon_sym_AMP_AMP,
    ACTIONS(6893), 1,
      anon_sym_PIPE,
    ACTIONS(6895), 1,
      anon_sym_CARET,
    ACTIONS(6897), 1,
      anon_sym_AMP,
    ACTIONS(6911), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6913), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(6915), 1,
      anon_sym_QMARK,
    ACTIONS(7009), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6617), 1,
      aux_sym_arithmetic_expansion_repeat1,
    ACTIONS(6885), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6899), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(6901), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(6903), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(6905), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(6907), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(6909), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6887), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [147048] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(6883), 1,
      anon_sym_EQ,
    ACTIONS(6889), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(6891), 1,
      anon_sym_AMP_AMP,
    ACTIONS(6893), 1,
      anon_sym_PIPE,
    ACTIONS(6895), 1,
      anon_sym_CARET,
    ACTIONS(6897), 1,
      anon_sym_AMP,
    ACTIONS(6911), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6913), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(6915), 1,
      anon_sym_QMARK,
    ACTIONS(7011), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6691), 1,
      aux_sym_arithmetic_expansion_repeat1,
    ACTIONS(6885), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6899), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(6901), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(6903), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(6905), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(6907), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(6909), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6887), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [147130] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(6883), 1,
      anon_sym_EQ,
    ACTIONS(6889), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(6891), 1,
      anon_sym_AMP_AMP,
    ACTIONS(6893), 1,
      anon_sym_PIPE,
    ACTIONS(6895), 1,
      anon_sym_CARET,
    ACTIONS(6897), 1,
      anon_sym_AMP,
    ACTIONS(6911), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6913), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(6915), 1,
      anon_sym_QMARK,
    ACTIONS(7013), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6622), 1,
      aux_sym_arithmetic_expansion_repeat1,
    ACTIONS(6885), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6899), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(6901), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(6903), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(6905), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(6907), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(6909), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6887), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [147212] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(6883), 1,
      anon_sym_EQ,
    ACTIONS(6889), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(6891), 1,
      anon_sym_AMP_AMP,
    ACTIONS(6893), 1,
      anon_sym_PIPE,
    ACTIONS(6895), 1,
      anon_sym_CARET,
    ACTIONS(6897), 1,
      anon_sym_AMP,
    ACTIONS(6911), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6913), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(6915), 1,
      anon_sym_QMARK,
    ACTIONS(7015), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6629), 1,
      aux_sym_arithmetic_expansion_repeat1,
    ACTIONS(6885), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6899), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(6901), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(6903), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(6905), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(6907), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(6909), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6887), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [147294] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(6883), 1,
      anon_sym_EQ,
    ACTIONS(6889), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(6891), 1,
      anon_sym_AMP_AMP,
    ACTIONS(6893), 1,
      anon_sym_PIPE,
    ACTIONS(6895), 1,
      anon_sym_CARET,
    ACTIONS(6897), 1,
      anon_sym_AMP,
    ACTIONS(6911), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6913), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(6915), 1,
      anon_sym_QMARK,
    ACTIONS(7017), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6588), 1,
      aux_sym_arithmetic_expansion_repeat1,
    ACTIONS(6885), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6899), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(6901), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(6903), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(6905), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(6907), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(6909), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6887), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [147376] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(6883), 1,
      anon_sym_EQ,
    ACTIONS(6889), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(6891), 1,
      anon_sym_AMP_AMP,
    ACTIONS(6893), 1,
      anon_sym_PIPE,
    ACTIONS(6895), 1,
      anon_sym_CARET,
    ACTIONS(6897), 1,
      anon_sym_AMP,
    ACTIONS(6911), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6913), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(6915), 1,
      anon_sym_QMARK,
    ACTIONS(7019), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6688), 1,
      aux_sym_arithmetic_expansion_repeat1,
    ACTIONS(6885), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6899), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(6901), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(6903), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(6905), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(6907), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(6909), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6887), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [147458] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(6883), 1,
      anon_sym_EQ,
    ACTIONS(6889), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(6891), 1,
      anon_sym_AMP_AMP,
    ACTIONS(6893), 1,
      anon_sym_PIPE,
    ACTIONS(6895), 1,
      anon_sym_CARET,
    ACTIONS(6897), 1,
      anon_sym_AMP,
    ACTIONS(6911), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6913), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(6915), 1,
      anon_sym_QMARK,
    ACTIONS(7021), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6548), 1,
      aux_sym_arithmetic_expansion_repeat1,
    ACTIONS(6885), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6899), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(6901), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(6903), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(6905), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(6907), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(6909), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6887), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [147540] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(6883), 1,
      anon_sym_EQ,
    ACTIONS(6889), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(6891), 1,
      anon_sym_AMP_AMP,
    ACTIONS(6893), 1,
      anon_sym_PIPE,
    ACTIONS(6895), 1,
      anon_sym_CARET,
    ACTIONS(6897), 1,
      anon_sym_AMP,
    ACTIONS(6911), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6913), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(6915), 1,
      anon_sym_QMARK,
    ACTIONS(7023), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6580), 1,
      aux_sym_arithmetic_expansion_repeat1,
    ACTIONS(6885), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6899), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(6901), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(6903), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(6905), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(6907), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(6909), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6887), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [147622] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(6883), 1,
      anon_sym_EQ,
    ACTIONS(6889), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(6891), 1,
      anon_sym_AMP_AMP,
    ACTIONS(6893), 1,
      anon_sym_PIPE,
    ACTIONS(6895), 1,
      anon_sym_CARET,
    ACTIONS(6897), 1,
      anon_sym_AMP,
    ACTIONS(6911), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6913), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(6915), 1,
      anon_sym_QMARK,
    ACTIONS(7025), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6656), 1,
      aux_sym_arithmetic_expansion_repeat1,
    ACTIONS(6885), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6899), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(6901), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(6903), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(6905), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(6907), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(6909), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6887), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [147704] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1261), 15,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      sym__special_character,
    ACTIONS(1263), 23,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [147750] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(6883), 1,
      anon_sym_EQ,
    ACTIONS(6889), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(6891), 1,
      anon_sym_AMP_AMP,
    ACTIONS(6893), 1,
      anon_sym_PIPE,
    ACTIONS(6895), 1,
      anon_sym_CARET,
    ACTIONS(6897), 1,
      anon_sym_AMP,
    ACTIONS(6911), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6913), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(6915), 1,
      anon_sym_QMARK,
    ACTIONS(7027), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6650), 1,
      aux_sym_arithmetic_expansion_repeat1,
    ACTIONS(6885), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6899), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(6901), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(6903), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(6905), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(6907), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(6909), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6887), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [147832] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1261), 12,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(1263), 26,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [147878] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(6883), 1,
      anon_sym_EQ,
    ACTIONS(6889), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(6891), 1,
      anon_sym_AMP_AMP,
    ACTIONS(6893), 1,
      anon_sym_PIPE,
    ACTIONS(6895), 1,
      anon_sym_CARET,
    ACTIONS(6897), 1,
      anon_sym_AMP,
    ACTIONS(6911), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6913), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(6915), 1,
      anon_sym_QMARK,
    ACTIONS(7029), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6657), 1,
      aux_sym_arithmetic_expansion_repeat1,
    ACTIONS(6885), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6899), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(6901), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(6903), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(6905), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(6907), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(6909), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6887), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [147960] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(6883), 1,
      anon_sym_EQ,
    ACTIONS(6889), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(6891), 1,
      anon_sym_AMP_AMP,
    ACTIONS(6893), 1,
      anon_sym_PIPE,
    ACTIONS(6895), 1,
      anon_sym_CARET,
    ACTIONS(6897), 1,
      anon_sym_AMP,
    ACTIONS(6911), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6913), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(6915), 1,
      anon_sym_QMARK,
    ACTIONS(7031), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6643), 1,
      aux_sym_arithmetic_expansion_repeat1,
    ACTIONS(6885), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6899), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(6901), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(6903), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(6905), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(6907), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(6909), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6887), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [148042] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7035), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(7033), 24,
      anon_sym_RPAREN_RPAREN,
      anon_sym_COMMA,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      anon_sym_COLON,
  [148088] = 15,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6807), 1,
      anon_sym_EQ,
    ACTIONS(6877), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7037), 1,
      anon_sym_PIPE,
    ACTIONS(7039), 1,
      anon_sym_CARET,
    ACTIONS(7041), 1,
      anon_sym_AMP,
    ACTIONS(7049), 1,
      sym_test_operator,
    ACTIONS(6869), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6871), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(6873), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7043), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7045), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7047), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(6875), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6805), 17,
      sym__concat,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [148158] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7053), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(7051), 24,
      anon_sym_RPAREN_RPAREN,
      anon_sym_COMMA,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      anon_sym_COLON,
  [148204] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(6883), 1,
      anon_sym_EQ,
    ACTIONS(6889), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(6891), 1,
      anon_sym_AMP_AMP,
    ACTIONS(6893), 1,
      anon_sym_PIPE,
    ACTIONS(6895), 1,
      anon_sym_CARET,
    ACTIONS(6897), 1,
      anon_sym_AMP,
    ACTIONS(6911), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6913), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(6915), 1,
      anon_sym_QMARK,
    ACTIONS(7055), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6692), 1,
      aux_sym_arithmetic_expansion_repeat1,
    ACTIONS(6885), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6899), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(6901), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(6903), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(6905), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(6907), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(6909), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6887), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [148286] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6864), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(6862), 24,
      sym__concat,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [148332] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7057), 1,
      sym__special_character,
    STATE(2806), 1,
      aux_sym__literal_repeat1,
    ACTIONS(1362), 12,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(1364), 24,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [148382] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(5770), 13,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
    ACTIONS(5772), 25,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [148428] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6732), 1,
      anon_sym_RBRACK,
    ACTIONS(7060), 1,
      sym__concat,
    ACTIONS(1273), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1369), 22,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [148478] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7064), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(7062), 24,
      anon_sym_RPAREN_RPAREN,
      anon_sym_COMMA,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      anon_sym_COLON,
  [148524] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6877), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6869), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6807), 13,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6805), 22,
      sym__concat,
      sym_test_operator,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [148574] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(6883), 1,
      anon_sym_EQ,
    ACTIONS(6889), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(6891), 1,
      anon_sym_AMP_AMP,
    ACTIONS(6893), 1,
      anon_sym_PIPE,
    ACTIONS(6895), 1,
      anon_sym_CARET,
    ACTIONS(6897), 1,
      anon_sym_AMP,
    ACTIONS(6911), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6913), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(6915), 1,
      anon_sym_QMARK,
    ACTIONS(7066), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6613), 1,
      aux_sym_arithmetic_expansion_repeat1,
    ACTIONS(6885), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6899), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(6901), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(6903), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(6905), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(6907), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(6909), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6887), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [148656] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6877), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6869), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6807), 13,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6805), 22,
      sym__concat,
      sym_test_operator,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [148706] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6877), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6869), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6875), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6807), 10,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(6805), 22,
      sym__concat,
      sym_test_operator,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [148758] = 7,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6877), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6869), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6873), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(6875), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6807), 8,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(6805), 22,
      sym__concat,
      sym_test_operator,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [148812] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6723), 1,
      anon_sym_RBRACK,
    ACTIONS(7068), 1,
      sym__concat,
    ACTIONS(1273), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1369), 22,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [148862] = 9,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6877), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7049), 1,
      sym_test_operator,
    ACTIONS(6869), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6871), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(6873), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(6875), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6807), 6,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(6805), 21,
      sym__concat,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [148920] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(5766), 13,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
    ACTIONS(5768), 25,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [148966] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1352), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1354), 24,
      anon_sym_RPAREN_RPAREN,
      anon_sym_COMMA,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      anon_sym_COLON,
  [149012] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1261), 13,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      sym__special_character,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(1263), 25,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [149058] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(6883), 1,
      anon_sym_EQ,
    ACTIONS(6889), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(6891), 1,
      anon_sym_AMP_AMP,
    ACTIONS(6893), 1,
      anon_sym_PIPE,
    ACTIONS(6895), 1,
      anon_sym_CARET,
    ACTIONS(6897), 1,
      anon_sym_AMP,
    ACTIONS(6911), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6913), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(6915), 1,
      anon_sym_QMARK,
    ACTIONS(7070), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6698), 1,
      aux_sym_arithmetic_expansion_repeat1,
    ACTIONS(6885), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6899), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(6901), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(6903), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(6905), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(6907), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(6909), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6887), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [149140] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1308), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1310), 24,
      anon_sym_RPAREN_RPAREN,
      anon_sym_COMMA,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      anon_sym_COLON,
  [149186] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1304), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1306), 24,
      anon_sym_RPAREN_RPAREN,
      anon_sym_COMMA,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      anon_sym_COLON,
  [149232] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1308), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1310), 24,
      anon_sym_RPAREN_RPAREN,
      anon_sym_COMMA,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      anon_sym_COLON,
  [149278] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1312), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1314), 24,
      anon_sym_RPAREN_RPAREN,
      anon_sym_COMMA,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      anon_sym_COLON,
  [149324] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7064), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(7062), 24,
      anon_sym_RPAREN_RPAREN,
      anon_sym_COMMA,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      anon_sym_COLON,
  [149370] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1340), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1342), 24,
      anon_sym_RPAREN_RPAREN,
      anon_sym_COMMA,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      anon_sym_COLON,
  [149416] = 11,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6877), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7049), 1,
      sym_test_operator,
    ACTIONS(6869), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6871), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(6873), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7045), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7047), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(6875), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6807), 4,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(6805), 19,
      sym__concat,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [149478] = 12,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6877), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7049), 1,
      sym_test_operator,
    ACTIONS(6869), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6871), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(6873), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7043), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7045), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7047), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(6875), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6807), 4,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(6805), 17,
      sym__concat,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [149542] = 13,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6877), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7041), 1,
      anon_sym_AMP,
    ACTIONS(7049), 1,
      sym_test_operator,
    ACTIONS(6869), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6871), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(6873), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7043), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7045), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7047), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(6807), 3,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
    ACTIONS(6875), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6805), 17,
      sym__concat,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [149608] = 14,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6877), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7039), 1,
      anon_sym_CARET,
    ACTIONS(7041), 1,
      anon_sym_AMP,
    ACTIONS(7049), 1,
      sym_test_operator,
    ACTIONS(6807), 2,
      anon_sym_EQ,
      anon_sym_PIPE,
    ACTIONS(6869), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6871), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(6873), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7043), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7045), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7047), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(6875), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6805), 17,
      sym__concat,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [149676] = 16,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6807), 1,
      anon_sym_EQ,
    ACTIONS(6877), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7037), 1,
      anon_sym_PIPE,
    ACTIONS(7039), 1,
      anon_sym_CARET,
    ACTIONS(7041), 1,
      anon_sym_AMP,
    ACTIONS(7049), 1,
      sym_test_operator,
    ACTIONS(7072), 1,
      anon_sym_AMP_AMP,
    ACTIONS(6869), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6871), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(6873), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7043), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7045), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7047), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(6875), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6805), 16,
      sym__concat,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [149748] = 19,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6877), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7037), 1,
      anon_sym_PIPE,
    ACTIONS(7039), 1,
      anon_sym_CARET,
    ACTIONS(7041), 1,
      anon_sym_AMP,
    ACTIONS(7049), 1,
      sym_test_operator,
    ACTIONS(7072), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7074), 1,
      anon_sym_EQ,
    ACTIONS(7076), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7078), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(7080), 1,
      anon_sym_QMARK,
    ACTIONS(6869), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6871), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(6873), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7043), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7045), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7047), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(6875), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6805), 13,
      sym__concat,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_RBRACK,
  [149826] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6708), 1,
      anon_sym_RBRACK,
    ACTIONS(7082), 1,
      sym__concat,
    ACTIONS(1273), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1369), 22,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [149876] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6670), 1,
      anon_sym_RBRACK,
    ACTIONS(7084), 1,
      sym__concat,
    ACTIONS(1273), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1369), 22,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [149926] = 18,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6807), 1,
      anon_sym_EQ,
    ACTIONS(6877), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7037), 1,
      anon_sym_PIPE,
    ACTIONS(7039), 1,
      anon_sym_CARET,
    ACTIONS(7041), 1,
      anon_sym_AMP,
    ACTIONS(7049), 1,
      sym_test_operator,
    ACTIONS(7072), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7076), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7080), 1,
      anon_sym_QMARK,
    ACTIONS(6869), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6871), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(6873), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7043), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7045), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7047), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(6875), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6805), 14,
      sym__concat,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
  [150002] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(6883), 1,
      anon_sym_EQ,
    ACTIONS(6889), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(6891), 1,
      anon_sym_AMP_AMP,
    ACTIONS(6893), 1,
      anon_sym_PIPE,
    ACTIONS(6895), 1,
      anon_sym_CARET,
    ACTIONS(6897), 1,
      anon_sym_AMP,
    ACTIONS(6911), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6913), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(6915), 1,
      anon_sym_QMARK,
    ACTIONS(7086), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6661), 1,
      aux_sym_arithmetic_expansion_repeat1,
    ACTIONS(6885), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6899), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(6901), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(6903), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(6905), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(6907), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(6909), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6887), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [150084] = 8,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6877), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6869), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6871), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(6873), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(6875), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6807), 6,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(6805), 22,
      sym__concat,
      sym_test_operator,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [150140] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(7090), 1,
      anon_sym_EQ,
    ACTIONS(7092), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(7088), 36,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_DASHo,
      anon_sym_AMP_AMP,
      anon_sym_DASHa,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
  [150188] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6797), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(6795), 24,
      sym__concat,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [150234] = 8,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6911), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6885), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6905), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(6907), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(6909), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7096), 6,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7094), 22,
      anon_sym_RPAREN_RPAREN,
      anon_sym_COMMA,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      anon_sym_COLON,
  [150290] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1324), 13,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1326), 25,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [150336] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1251), 13,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1253), 25,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [150382] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6885), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7096), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(7094), 22,
      anon_sym_RPAREN_RPAREN,
      anon_sym_COMMA,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      anon_sym_COLON,
  [150430] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(6883), 1,
      anon_sym_EQ,
    ACTIONS(6889), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(6891), 1,
      anon_sym_AMP_AMP,
    ACTIONS(6893), 1,
      anon_sym_PIPE,
    ACTIONS(6895), 1,
      anon_sym_CARET,
    ACTIONS(6897), 1,
      anon_sym_AMP,
    ACTIONS(6911), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6913), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(6915), 1,
      anon_sym_QMARK,
    ACTIONS(7098), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6642), 1,
      aux_sym_arithmetic_expansion_repeat1,
    ACTIONS(6885), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6899), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(6901), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(6903), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(6905), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(6907), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(6909), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6887), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [150512] = 16,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6889), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(6891), 1,
      anon_sym_AMP_AMP,
    ACTIONS(6893), 1,
      anon_sym_PIPE,
    ACTIONS(6895), 1,
      anon_sym_CARET,
    ACTIONS(6897), 1,
      anon_sym_AMP,
    ACTIONS(6911), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7102), 1,
      anon_sym_EQ,
    ACTIONS(6885), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6899), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(6901), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(6903), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(6905), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(6907), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(6909), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7100), 16,
      anon_sym_RPAREN_RPAREN,
      anon_sym_COMMA,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      anon_sym_COLON,
  [150584] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(2072), 12,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(2074), 26,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [150630] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1336), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1338), 24,
      anon_sym_RPAREN_RPAREN,
      anon_sym_COMMA,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      anon_sym_COLON,
  [150676] = 8,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6911), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6885), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6905), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(6907), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(6909), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7096), 6,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7094), 22,
      anon_sym_RPAREN_RPAREN,
      anon_sym_COMMA,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      anon_sym_COLON,
  [150732] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7106), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(7104), 24,
      anon_sym_RPAREN_RPAREN,
      anon_sym_COMMA,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      anon_sym_COLON,
  [150778] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1348), 13,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1350), 25,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [150824] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6658), 1,
      anon_sym_RBRACK,
    ACTIONS(7108), 1,
      sym__concat,
    ACTIONS(1273), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1369), 22,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [150874] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1328), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1330), 24,
      anon_sym_RPAREN_RPAREN,
      anon_sym_COMMA,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      anon_sym_COLON,
  [150920] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1320), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1322), 24,
      anon_sym_RPAREN_RPAREN,
      anon_sym_COMMA,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      anon_sym_COLON,
  [150966] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1296), 13,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1298), 25,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [151012] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1344), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1346), 24,
      anon_sym_RPAREN_RPAREN,
      anon_sym_COMMA,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      anon_sym_COLON,
  [151058] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1356), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1358), 24,
      anon_sym_RPAREN_RPAREN,
      anon_sym_COMMA,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      anon_sym_COLON,
  [151104] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1300), 13,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1302), 25,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [151150] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(6883), 1,
      anon_sym_EQ,
    ACTIONS(6889), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(6891), 1,
      anon_sym_AMP_AMP,
    ACTIONS(6893), 1,
      anon_sym_PIPE,
    ACTIONS(6895), 1,
      anon_sym_CARET,
    ACTIONS(6897), 1,
      anon_sym_AMP,
    ACTIONS(6911), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6913), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(6915), 1,
      anon_sym_QMARK,
    ACTIONS(7110), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6724), 1,
      aux_sym_arithmetic_expansion_repeat1,
    ACTIONS(6885), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6899), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(6901), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(6903), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(6905), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(6907), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(6909), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6887), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [151232] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1332), 13,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1334), 25,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [151278] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7114), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(7112), 24,
      anon_sym_RPAREN_RPAREN,
      anon_sym_COMMA,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      anon_sym_COLON,
  [151324] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1292), 13,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1294), 25,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [151370] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7116), 1,
      sym_extglob_pattern,
    ACTIONS(6751), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(6749), 23,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [151418] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1273), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1369), 24,
      sym__concat,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [151464] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(2094), 12,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(2096), 26,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [151510] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7118), 1,
      sym_extglob_pattern,
    ACTIONS(6751), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(6749), 23,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [151558] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1316), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1318), 24,
      anon_sym_RPAREN_RPAREN,
      anon_sym_COMMA,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      anon_sym_COLON,
  [151604] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7120), 1,
      sym__special_character,
    STATE(2867), 1,
      aux_sym__literal_repeat1,
    ACTIONS(1362), 12,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(1364), 24,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [151654] = 17,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6877), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7037), 1,
      anon_sym_PIPE,
    ACTIONS(7039), 1,
      anon_sym_CARET,
    ACTIONS(7041), 1,
      anon_sym_AMP,
    ACTIONS(7049), 1,
      sym_test_operator,
    ACTIONS(7072), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7076), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7123), 1,
      anon_sym_EQ,
    ACTIONS(6869), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6871), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(6873), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7043), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7045), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7047), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(6875), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7125), 15,
      sym__concat,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [151728] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1316), 13,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1318), 25,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [151774] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1328), 13,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1330), 25,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [151820] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1320), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1322), 24,
      sym_test_operator,
      sym_extglob_pattern,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [151866] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1336), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1338), 24,
      sym_test_operator,
      sym_extglob_pattern,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [151912] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1340), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1342), 24,
      sym_test_operator,
      sym_extglob_pattern,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [151958] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6757), 1,
      anon_sym_RBRACK,
    ACTIONS(7127), 1,
      sym__concat,
    ACTIONS(1273), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1369), 22,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [152008] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1312), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1314), 24,
      sym_test_operator,
      sym_extglob_pattern,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [152054] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1352), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1354), 24,
      sym_test_operator,
      sym_extglob_pattern,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [152100] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1356), 13,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1358), 25,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [152146] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(5373), 13,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
    ACTIONS(5375), 25,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [152192] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1344), 13,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1346), 25,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [152238] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(6883), 1,
      anon_sym_EQ,
    ACTIONS(6889), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(6891), 1,
      anon_sym_AMP_AMP,
    ACTIONS(6893), 1,
      anon_sym_PIPE,
    ACTIONS(6895), 1,
      anon_sym_CARET,
    ACTIONS(6897), 1,
      anon_sym_AMP,
    ACTIONS(6911), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6913), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(6915), 1,
      anon_sym_QMARK,
    ACTIONS(7129), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6574), 1,
      aux_sym_arithmetic_expansion_repeat1,
    ACTIONS(6885), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6899), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(6901), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(6903), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(6905), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(6907), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(6909), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6887), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [152320] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1308), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1310), 24,
      sym_test_operator,
      sym_extglob_pattern,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [152366] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1304), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1306), 24,
      sym_test_operator,
      sym_extglob_pattern,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [152412] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1308), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1310), 24,
      sym_test_operator,
      sym_extglob_pattern,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [152458] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(5766), 12,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(5768), 26,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [152504] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1320), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1322), 24,
      sym_test_operator,
      sym_extglob_pattern,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [152550] = 7,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(5821), 2,
      anon_sym_PIPE,
      anon_sym_LT_LT,
    ACTIONS(5826), 4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
    ACTIONS(5819), 5,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(5823), 5,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(5828), 7,
      sym_file_descriptor,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
    ACTIONS(5831), 15,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [152604] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1235), 1,
      sym_file_descriptor,
    ACTIONS(3030), 1,
      anon_sym_DQUOTE,
    ACTIONS(7135), 1,
      sym_variable_name,
    STATE(4446), 1,
      sym_string,
    ACTIONS(7133), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(7131), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1227), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
      anon_sym_LT_LT_LT,
  [152660] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1241), 1,
      sym_file_descriptor,
    ACTIONS(3030), 1,
      anon_sym_DQUOTE,
    ACTIONS(7135), 1,
      sym_variable_name,
    STATE(4446), 1,
      sym_string,
    ACTIONS(7133), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(7131), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1239), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
      anon_sym_LT_LT_LT,
  [152716] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6807), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(6805), 24,
      sym__concat,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [152762] = 17,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6889), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(6891), 1,
      anon_sym_AMP_AMP,
    ACTIONS(6893), 1,
      anon_sym_PIPE,
    ACTIONS(6895), 1,
      anon_sym_CARET,
    ACTIONS(6897), 1,
      anon_sym_AMP,
    ACTIONS(6911), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6915), 1,
      anon_sym_QMARK,
    ACTIONS(7139), 1,
      anon_sym_EQ,
    ACTIONS(6885), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6899), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(6901), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(6903), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(6905), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(6907), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(6909), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7137), 15,
      anon_sym_RPAREN_RPAREN,
      anon_sym_COMMA,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_COLON,
  [152836] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1340), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1342), 24,
      sym_test_operator,
      sym_extglob_pattern,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [152882] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1312), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1314), 24,
      sym_test_operator,
      sym_extglob_pattern,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [152928] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1352), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1354), 24,
      sym_test_operator,
      sym_extglob_pattern,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [152974] = 18,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6883), 1,
      anon_sym_EQ,
    ACTIONS(6889), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(6891), 1,
      anon_sym_AMP_AMP,
    ACTIONS(6893), 1,
      anon_sym_PIPE,
    ACTIONS(6895), 1,
      anon_sym_CARET,
    ACTIONS(6897), 1,
      anon_sym_AMP,
    ACTIONS(6911), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6913), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(6915), 1,
      anon_sym_QMARK,
    ACTIONS(6885), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6899), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(6901), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(6903), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(6905), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(6907), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(6909), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7137), 14,
      anon_sym_RPAREN_RPAREN,
      anon_sym_COMMA,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_COLON,
  [153050] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1235), 1,
      sym_file_descriptor,
    ACTIONS(7143), 1,
      anon_sym_DQUOTE,
    ACTIONS(7147), 1,
      sym_variable_name,
    STATE(4322), 1,
      sym_string,
    ACTIONS(7145), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(7141), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1227), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
  [153106] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1241), 1,
      sym_file_descriptor,
    ACTIONS(7143), 1,
      anon_sym_DQUOTE,
    ACTIONS(7147), 1,
      sym_variable_name,
    STATE(4322), 1,
      sym_string,
    ACTIONS(7145), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(7141), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1239), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
  [153162] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6807), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(6805), 24,
      sym__concat,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [153208] = 15,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6891), 1,
      anon_sym_AMP_AMP,
    ACTIONS(6893), 1,
      anon_sym_PIPE,
    ACTIONS(6895), 1,
      anon_sym_CARET,
    ACTIONS(6897), 1,
      anon_sym_AMP,
    ACTIONS(6911), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7139), 1,
      anon_sym_EQ,
    ACTIONS(6885), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6899), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(6901), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(6903), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(6905), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(6907), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(6909), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7137), 17,
      anon_sym_RPAREN_RPAREN,
      anon_sym_COMMA,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      anon_sym_COLON,
  [153278] = 14,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6893), 1,
      anon_sym_PIPE,
    ACTIONS(6895), 1,
      anon_sym_CARET,
    ACTIONS(6897), 1,
      anon_sym_AMP,
    ACTIONS(6911), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7139), 1,
      anon_sym_EQ,
    ACTIONS(6885), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6899), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(6901), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(6903), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(6905), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(6907), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(6909), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7137), 18,
      anon_sym_RPAREN_RPAREN,
      anon_sym_COMMA,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      anon_sym_COLON,
  [153346] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6801), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(6799), 24,
      sym__concat,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [153392] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1320), 13,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1322), 25,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [153438] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1235), 1,
      sym_file_descriptor,
    ACTIONS(3371), 1,
      anon_sym_DQUOTE,
    ACTIONS(7153), 1,
      sym_variable_name,
    STATE(4512), 1,
      sym_string,
    ACTIONS(7151), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(7149), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1227), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
  [153494] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1241), 1,
      sym_file_descriptor,
    ACTIONS(3371), 1,
      anon_sym_DQUOTE,
    ACTIONS(7153), 1,
      sym_variable_name,
    STATE(4512), 1,
      sym_string,
    ACTIONS(7151), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(7149), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1239), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
  [153550] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1336), 13,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1338), 25,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [153596] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(4370), 12,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(4372), 26,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [153642] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1340), 13,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1342), 25,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [153688] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1312), 13,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1314), 25,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [153734] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1308), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1310), 24,
      sym_test_operator,
      sym_extglob_pattern,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [153780] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1304), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1306), 24,
      sym_test_operator,
      sym_extglob_pattern,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [153826] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1308), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1310), 24,
      sym_test_operator,
      sym_extglob_pattern,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [153872] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1352), 13,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1354), 25,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [153918] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7155), 1,
      sym__special_character,
    STATE(2912), 1,
      aux_sym__literal_repeat1,
    ACTIONS(1362), 13,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
    ACTIONS(1364), 23,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [153968] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(5770), 12,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(5772), 26,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [154014] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7158), 1,
      sym__special_character,
    STATE(2867), 1,
      aux_sym__literal_repeat1,
    ACTIONS(5393), 12,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(5395), 24,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [154064] = 13,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6895), 1,
      anon_sym_CARET,
    ACTIONS(6897), 1,
      anon_sym_AMP,
    ACTIONS(6911), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6885), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6899), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(6901), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(6903), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(6905), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(6907), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7139), 2,
      anon_sym_EQ,
      anon_sym_PIPE,
    ACTIONS(6909), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7137), 18,
      anon_sym_RPAREN_RPAREN,
      anon_sym_COMMA,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      anon_sym_COLON,
  [154130] = 12,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6897), 1,
      anon_sym_AMP,
    ACTIONS(6911), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6885), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6899), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(6901), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(6903), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(6905), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(6907), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(6909), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7139), 3,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
    ACTIONS(7137), 18,
      anon_sym_RPAREN_RPAREN,
      anon_sym_COMMA,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      anon_sym_COLON,
  [154194] = 11,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6911), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6885), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6899), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(6901), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(6903), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(6905), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(6907), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(6909), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7139), 4,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(7137), 18,
      anon_sym_RPAREN_RPAREN,
      anon_sym_COMMA,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      anon_sym_COLON,
  [154256] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1308), 13,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1310), 25,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [154302] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1304), 13,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1306), 25,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [154348] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1308), 13,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1310), 25,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [154394] = 10,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6911), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6885), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6901), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(6903), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(6905), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(6907), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(6909), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7139), 4,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(7137), 20,
      anon_sym_RPAREN_RPAREN,
      anon_sym_COMMA,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      anon_sym_COLON,
  [154454] = 8,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6911), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6885), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6905), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(6907), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(6909), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7139), 6,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7137), 22,
      anon_sym_RPAREN_RPAREN,
      anon_sym_COMMA,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      anon_sym_COLON,
  [154510] = 7,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6911), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6885), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6907), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(6909), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7139), 8,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7137), 22,
      anon_sym_RPAREN_RPAREN,
      anon_sym_COMMA,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      anon_sym_COLON,
  [154564] = 8,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6877), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6869), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6871), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(6873), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(6875), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6791), 6,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(6789), 22,
      sym__concat,
      sym_test_operator,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [154620] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6869), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6791), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(6789), 22,
      sym__concat,
      sym_test_operator,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [154668] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6829), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(6827), 24,
      anon_sym_RPAREN_RPAREN,
      anon_sym_COMMA,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      anon_sym_COLON,
  [154714] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1261), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1263), 24,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      sym__special_character,
  [154760] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1261), 15,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      sym__special_character,
    ACTIONS(1263), 23,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [154806] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6820), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(6818), 24,
      anon_sym_RPAREN_RPAREN,
      anon_sym_COMMA,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      anon_sym_COLON,
  [154852] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7162), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(7160), 24,
      anon_sym_RPAREN_RPAREN,
      anon_sym_COMMA,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      anon_sym_COLON,
  [154898] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7166), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(7164), 24,
      anon_sym_RPAREN_RPAREN,
      anon_sym_COMMA,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      anon_sym_COLON,
  [154944] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(6883), 1,
      anon_sym_EQ,
    ACTIONS(6889), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(6891), 1,
      anon_sym_AMP_AMP,
    ACTIONS(6893), 1,
      anon_sym_PIPE,
    ACTIONS(6895), 1,
      anon_sym_CARET,
    ACTIONS(6897), 1,
      anon_sym_AMP,
    ACTIONS(6911), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6913), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(6915), 1,
      anon_sym_QMARK,
    ACTIONS(7168), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6699), 1,
      aux_sym_arithmetic_expansion_repeat1,
    ACTIONS(6885), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6899), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(6901), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(6903), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(6905), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(6907), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(6909), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6887), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [155026] = 9,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6877), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7049), 1,
      sym_test_operator,
    ACTIONS(6869), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6871), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(6873), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(6875), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6791), 6,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(6789), 21,
      sym__concat,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [155084] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(5373), 12,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(5375), 26,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [155130] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1261), 13,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
    ACTIONS(1263), 25,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [155176] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6911), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6885), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6909), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7139), 10,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7137), 22,
      anon_sym_RPAREN_RPAREN,
      anon_sym_COMMA,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      anon_sym_COLON,
  [155228] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6911), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6885), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7139), 13,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7137), 22,
      anon_sym_RPAREN_RPAREN,
      anon_sym_COMMA,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      anon_sym_COLON,
  [155278] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6885), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7139), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(7137), 22,
      anon_sym_RPAREN_RPAREN,
      anon_sym_COMMA,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      anon_sym_COLON,
  [155326] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7170), 1,
      sym__special_character,
    STATE(2912), 1,
      aux_sym__literal_repeat1,
    ACTIONS(5476), 13,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
    ACTIONS(5478), 23,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [155376] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(4552), 12,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(4554), 26,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [155422] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(4247), 5,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(5930), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(5932), 10,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
    ACTIONS(4261), 16,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [155472] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7172), 1,
      anon_sym_EQ,
    ACTIONS(7176), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7178), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7180), 1,
      anon_sym_PIPE,
    ACTIONS(7182), 1,
      anon_sym_CARET,
    ACTIONS(7184), 1,
      anon_sym_AMP,
    ACTIONS(7198), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7200), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(7202), 1,
      anon_sym_QMARK,
    ACTIONS(7204), 1,
      anon_sym_COLON,
    ACTIONS(7206), 1,
      sym_test_operator,
    ACTIONS(6793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7186), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7188), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7190), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7192), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7194), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7196), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7174), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [155551] = 12,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7220), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7222), 1,
      sym_test_operator,
    ACTIONS(6793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7208), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7210), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7212), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7214), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7216), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7218), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6807), 4,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(6805), 16,
      anon_sym_RPAREN_RPAREN,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [155614] = 25,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6474), 1,
      sym_word,
    ACTIONS(6480), 1,
      anon_sym_LPAREN,
    ACTIONS(6488), 1,
      anon_sym_DOLLAR,
    ACTIONS(6494), 1,
      aux_sym_number_token1,
    ACTIONS(6496), 1,
      aux_sym_number_token2,
    ACTIONS(6500), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(6508), 1,
      sym_test_operator,
    ACTIONS(6510), 1,
      sym_extglob_pattern,
    ACTIONS(6512), 1,
      sym__brace_start,
    ACTIONS(6560), 1,
      anon_sym_esac,
    ACTIONS(7226), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(7228), 1,
      sym__special_character,
    ACTIONS(7230), 1,
      anon_sym_DQUOTE,
    ACTIONS(7234), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(7236), 1,
      anon_sym_BQUOTE,
    ACTIONS(7238), 1,
      anon_sym_DOLLAR_BQUOTE,
    STATE(6426), 1,
      aux_sym__literal_repeat1,
    STATE(6851), 1,
      sym_last_case_item,
    ACTIONS(7224), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(7232), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(7240), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3377), 2,
      sym_case_item,
      aux_sym_case_statement_repeat1,
    STATE(6695), 2,
      sym_concatenation,
      sym__extglob_blob,
    STATE(6303), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [155703] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7242), 1,
      anon_sym_LBRACK,
    ACTIONS(6600), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(6598), 22,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [155750] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(7246), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(7244), 36,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_DASHo,
      anon_sym_AMP_AMP,
      anon_sym_DASHa,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
  [155795] = 15,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(7256), 1,
      anon_sym_PIPE,
    ACTIONS(7258), 1,
      anon_sym_CARET,
    ACTIONS(7260), 1,
      anon_sym_AMP,
    ACTIONS(7272), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7274), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(7250), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7252), 2,
      anon_sym_PIPE_PIPE,
      anon_sym_DASHo,
    ACTIONS(7254), 2,
      anon_sym_AMP_AMP,
      anon_sym_DASHa,
    ACTIONS(7262), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7266), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7268), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7270), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7264), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7248), 13,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [155864] = 13,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(7256), 1,
      anon_sym_PIPE,
    ACTIONS(7258), 1,
      anon_sym_CARET,
    ACTIONS(7260), 1,
      anon_sym_AMP,
    ACTIONS(7272), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7274), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(7250), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7262), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7266), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7268), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7270), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7264), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7248), 17,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_DASHo,
      anon_sym_AMP_AMP,
      anon_sym_DASHa,
  [155929] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7276), 1,
      anon_sym_EQ,
    ACTIONS(7088), 13,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(7092), 23,
      anon_sym_RPAREN_RPAREN,
      anon_sym_COMMA,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_DASHo,
      anon_sym_AMP_AMP,
      anon_sym_DASHa,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
  [155976] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6807), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(6805), 23,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [156021] = 17,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7123), 1,
      anon_sym_EQ,
    ACTIONS(7280), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7282), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7284), 1,
      anon_sym_PIPE,
    ACTIONS(7286), 1,
      anon_sym_CARET,
    ACTIONS(7288), 1,
      anon_sym_AMP,
    ACTIONS(7302), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7304), 1,
      sym_test_operator,
    ACTIONS(7278), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7290), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7292), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7294), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7296), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7298), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7300), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7125), 14,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [156094] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(7258), 1,
      anon_sym_CARET,
    ACTIONS(7260), 1,
      anon_sym_AMP,
    ACTIONS(7272), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7274), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(7250), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7262), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7266), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7268), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7270), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7264), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7248), 18,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_DASHo,
      anon_sym_AMP_AMP,
      anon_sym_DASHa,
      anon_sym_PIPE,
  [156157] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6807), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(6805), 23,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [156202] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6797), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(6795), 23,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [156247] = 17,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7123), 1,
      anon_sym_EQ,
    ACTIONS(7176), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7178), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7180), 1,
      anon_sym_PIPE,
    ACTIONS(7182), 1,
      anon_sym_CARET,
    ACTIONS(7184), 1,
      anon_sym_AMP,
    ACTIONS(7198), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7206), 1,
      sym_test_operator,
    ACTIONS(6793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7186), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7188), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7190), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7192), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7194), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7196), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7125), 14,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      anon_sym_COLON,
  [156320] = 11,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(7260), 1,
      anon_sym_AMP,
    ACTIONS(7272), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7274), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(7250), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7262), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7266), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7268), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7270), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7264), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7248), 19,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_DASHo,
      anon_sym_AMP_AMP,
      anon_sym_DASHa,
      anon_sym_PIPE,
      anon_sym_CARET,
  [156381] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7302), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7278), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6807), 13,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6805), 21,
      sym_test_operator,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [156430] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7302), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7278), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6807), 13,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6805), 21,
      sym_test_operator,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [156479] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7302), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7278), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7300), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6807), 10,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(6805), 21,
      sym_test_operator,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [156530] = 7,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7302), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7278), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7298), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7300), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6807), 8,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(6805), 21,
      sym_test_operator,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [156583] = 9,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7302), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7304), 1,
      sym_test_operator,
    ACTIONS(7278), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7296), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7298), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7300), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6807), 6,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(6805), 20,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [156640] = 8,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7302), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7278), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7296), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7298), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7300), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6807), 6,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(6805), 21,
      sym_test_operator,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [156695] = 11,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7302), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7304), 1,
      sym_test_operator,
    ACTIONS(7278), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7292), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7294), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7296), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7298), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7300), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6807), 4,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(6805), 18,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [156756] = 12,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7302), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7304), 1,
      sym_test_operator,
    ACTIONS(7278), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7290), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7292), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7294), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7296), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7298), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7300), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6807), 4,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(6805), 16,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [156819] = 10,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(7272), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7274), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(7250), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7262), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7266), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7268), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7270), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7264), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7248), 20,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_DASHo,
      anon_sym_AMP_AMP,
      anon_sym_DASHa,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
  [156878] = 13,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7288), 1,
      anon_sym_AMP,
    ACTIONS(7302), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7304), 1,
      sym_test_operator,
    ACTIONS(7278), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7290), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7292), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7294), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7296), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7298), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(6807), 3,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
    ACTIONS(7300), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6805), 16,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [156943] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7198), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6807), 13,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6805), 21,
      sym_test_operator,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      anon_sym_COLON,
  [156992] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7198), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6807), 13,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6805), 21,
      sym_test_operator,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      anon_sym_COLON,
  [157041] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7198), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7196), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6807), 10,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(6805), 21,
      sym_test_operator,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      anon_sym_COLON,
  [157092] = 7,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7198), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7194), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7196), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6807), 8,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(6805), 21,
      sym_test_operator,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      anon_sym_COLON,
  [157145] = 9,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7198), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7206), 1,
      sym_test_operator,
    ACTIONS(6793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7192), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7194), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7196), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6807), 6,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(6805), 20,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      anon_sym_COLON,
  [157202] = 11,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7198), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7206), 1,
      sym_test_operator,
    ACTIONS(6793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7188), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7190), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7192), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7194), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7196), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6807), 4,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(6805), 18,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      anon_sym_COLON,
  [157263] = 12,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7198), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7206), 1,
      sym_test_operator,
    ACTIONS(6793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7186), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7188), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7190), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7192), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7194), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7196), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6807), 4,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(6805), 16,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      anon_sym_COLON,
  [157326] = 13,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7184), 1,
      anon_sym_AMP,
    ACTIONS(7198), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7206), 1,
      sym_test_operator,
    ACTIONS(6793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7186), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7188), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7190), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7192), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7194), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(6807), 3,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
    ACTIONS(7196), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6805), 16,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      anon_sym_COLON,
  [157391] = 14,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7182), 1,
      anon_sym_CARET,
    ACTIONS(7184), 1,
      anon_sym_AMP,
    ACTIONS(7198), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7206), 1,
      sym_test_operator,
    ACTIONS(6793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6807), 2,
      anon_sym_EQ,
      anon_sym_PIPE,
    ACTIONS(7186), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7188), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7190), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7192), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7194), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7196), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6805), 16,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      anon_sym_COLON,
  [157458] = 15,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6807), 1,
      anon_sym_EQ,
    ACTIONS(7180), 1,
      anon_sym_PIPE,
    ACTIONS(7182), 1,
      anon_sym_CARET,
    ACTIONS(7184), 1,
      anon_sym_AMP,
    ACTIONS(7198), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7206), 1,
      sym_test_operator,
    ACTIONS(6793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7186), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7188), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7190), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7192), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7194), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7196), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6805), 16,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      anon_sym_COLON,
  [157527] = 16,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6807), 1,
      anon_sym_EQ,
    ACTIONS(7178), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7180), 1,
      anon_sym_PIPE,
    ACTIONS(7182), 1,
      anon_sym_CARET,
    ACTIONS(7184), 1,
      anon_sym_AMP,
    ACTIONS(7198), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7206), 1,
      sym_test_operator,
    ACTIONS(6793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7186), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7188), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7190), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7192), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7194), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7196), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6805), 15,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      anon_sym_COLON,
  [157598] = 19,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7172), 1,
      anon_sym_EQ,
    ACTIONS(7176), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7178), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7180), 1,
      anon_sym_PIPE,
    ACTIONS(7182), 1,
      anon_sym_CARET,
    ACTIONS(7184), 1,
      anon_sym_AMP,
    ACTIONS(7198), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7200), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(7202), 1,
      anon_sym_QMARK,
    ACTIONS(7206), 1,
      sym_test_operator,
    ACTIONS(6793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7186), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7188), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7190), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7192), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7194), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7196), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6805), 12,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_COLON,
  [157675] = 18,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6807), 1,
      anon_sym_EQ,
    ACTIONS(7176), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7178), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7180), 1,
      anon_sym_PIPE,
    ACTIONS(7182), 1,
      anon_sym_CARET,
    ACTIONS(7184), 1,
      anon_sym_AMP,
    ACTIONS(7198), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7202), 1,
      anon_sym_QMARK,
    ACTIONS(7206), 1,
      sym_test_operator,
    ACTIONS(6793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7186), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7188), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7190), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7192), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7194), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7196), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6805), 13,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_COLON,
  [157750] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1261), 13,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      aux_sym__simple_variable_name_token1,
      sym_word,
    ACTIONS(1263), 24,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [157795] = 14,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7286), 1,
      anon_sym_CARET,
    ACTIONS(7288), 1,
      anon_sym_AMP,
    ACTIONS(7302), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7304), 1,
      sym_test_operator,
    ACTIONS(6807), 2,
      anon_sym_EQ,
      anon_sym_PIPE,
    ACTIONS(7278), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7290), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7292), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7294), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7296), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7298), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7300), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6805), 16,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [157862] = 25,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6474), 1,
      sym_word,
    ACTIONS(6480), 1,
      anon_sym_LPAREN,
    ACTIONS(6488), 1,
      anon_sym_DOLLAR,
    ACTIONS(6494), 1,
      aux_sym_number_token1,
    ACTIONS(6496), 1,
      aux_sym_number_token2,
    ACTIONS(6500), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(6508), 1,
      sym_test_operator,
    ACTIONS(6510), 1,
      sym_extglob_pattern,
    ACTIONS(6512), 1,
      sym__brace_start,
    ACTIONS(6530), 1,
      anon_sym_esac,
    ACTIONS(7226), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(7228), 1,
      sym__special_character,
    ACTIONS(7230), 1,
      anon_sym_DQUOTE,
    ACTIONS(7234), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(7236), 1,
      anon_sym_BQUOTE,
    ACTIONS(7238), 1,
      anon_sym_DOLLAR_BQUOTE,
    STATE(6426), 1,
      aux_sym__literal_repeat1,
    STATE(7005), 1,
      sym_last_case_item,
    ACTIONS(7224), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(7232), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(7240), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3385), 2,
      sym_case_item,
      aux_sym_case_statement_repeat1,
    STATE(6695), 2,
      sym_concatenation,
      sym__extglob_blob,
    STATE(6303), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [157951] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7306), 1,
      sym__concat,
    ACTIONS(6844), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(6842), 22,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [157998] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7308), 1,
      sym__concat,
    ACTIONS(6838), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(6836), 22,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [158045] = 15,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6807), 1,
      anon_sym_EQ,
    ACTIONS(7284), 1,
      anon_sym_PIPE,
    ACTIONS(7286), 1,
      anon_sym_CARET,
    ACTIONS(7288), 1,
      anon_sym_AMP,
    ACTIONS(7302), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7304), 1,
      sym_test_operator,
    ACTIONS(7278), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7290), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7292), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7294), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7296), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7298), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7300), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6805), 16,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [158114] = 16,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6807), 1,
      anon_sym_EQ,
    ACTIONS(7282), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7284), 1,
      anon_sym_PIPE,
    ACTIONS(7286), 1,
      anon_sym_CARET,
    ACTIONS(7288), 1,
      anon_sym_AMP,
    ACTIONS(7302), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7304), 1,
      sym_test_operator,
    ACTIONS(7278), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7290), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7292), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7294), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7296), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7298), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7300), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6805), 15,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [158185] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7220), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7218), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6807), 10,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(6805), 21,
      sym_test_operator,
      anon_sym_RPAREN_RPAREN,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [158236] = 19,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7280), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7282), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7284), 1,
      anon_sym_PIPE,
    ACTIONS(7286), 1,
      anon_sym_CARET,
    ACTIONS(7288), 1,
      anon_sym_AMP,
    ACTIONS(7302), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7304), 1,
      sym_test_operator,
    ACTIONS(7310), 1,
      anon_sym_EQ,
    ACTIONS(7312), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(7314), 1,
      anon_sym_QMARK,
    ACTIONS(7278), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7290), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7292), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7294), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7296), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7298), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7300), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6805), 12,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_RBRACK,
  [158313] = 18,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6807), 1,
      anon_sym_EQ,
    ACTIONS(7280), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7282), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7284), 1,
      anon_sym_PIPE,
    ACTIONS(7286), 1,
      anon_sym_CARET,
    ACTIONS(7288), 1,
      anon_sym_AMP,
    ACTIONS(7302), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7304), 1,
      sym_test_operator,
    ACTIONS(7314), 1,
      anon_sym_QMARK,
    ACTIONS(7278), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7290), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7292), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7294), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7296), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7298), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7300), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6805), 13,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
  [158388] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6801), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(6799), 23,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [158433] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(7272), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7274), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(7250), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7266), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7268), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7270), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7248), 26,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_DASHo,
      anon_sym_AMP_AMP,
      anon_sym_DASHa,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
  [158488] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(7272), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7274), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(7250), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7268), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7270), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7248), 28,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_DASHo,
      anon_sym_AMP_AMP,
      anon_sym_DASHa,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [158541] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(7272), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7274), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(7250), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7270), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7248), 30,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_DASHo,
      anon_sym_AMP_AMP,
      anon_sym_DASHa,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
  [158592] = 8,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7322), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7316), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7318), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7320), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6807), 6,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(6805), 21,
      sym_test_operator,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [158647] = 18,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6807), 1,
      anon_sym_EQ,
    ACTIONS(7322), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7324), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7326), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7328), 1,
      anon_sym_PIPE,
    ACTIONS(7330), 1,
      anon_sym_CARET,
    ACTIONS(7332), 1,
      anon_sym_AMP,
    ACTIONS(7340), 1,
      anon_sym_QMARK,
    ACTIONS(7342), 1,
      sym_test_operator,
    ACTIONS(6793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7316), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7318), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7334), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7336), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7338), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7320), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6805), 13,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_RBRACK_RBRACK,
      anon_sym_EQ_TILDE,
  [158722] = 19,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7322), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7324), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7326), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7328), 1,
      anon_sym_PIPE,
    ACTIONS(7330), 1,
      anon_sym_CARET,
    ACTIONS(7332), 1,
      anon_sym_AMP,
    ACTIONS(7340), 1,
      anon_sym_QMARK,
    ACTIONS(7342), 1,
      sym_test_operator,
    ACTIONS(7344), 1,
      anon_sym_EQ,
    ACTIONS(7346), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(6793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7316), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7318), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7334), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7336), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7338), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7320), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6805), 12,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_RBRACK_RBRACK,
  [158799] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1318), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1316), 36,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_DASHo,
      anon_sym_AMP_AMP,
      anon_sym_DASHa,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
  [158844] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6797), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(6795), 23,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [158889] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(7272), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7274), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(7250), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7248), 33,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_DASHo,
      anon_sym_AMP_AMP,
      anon_sym_DASHa,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
  [158938] = 15,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6807), 1,
      anon_sym_EQ,
    ACTIONS(7322), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7328), 1,
      anon_sym_PIPE,
    ACTIONS(7330), 1,
      anon_sym_CARET,
    ACTIONS(7332), 1,
      anon_sym_AMP,
    ACTIONS(7342), 1,
      sym_test_operator,
    ACTIONS(6793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7316), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7318), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7334), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7336), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7338), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7320), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6805), 16,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_RBRACK_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [159007] = 14,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7322), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7330), 1,
      anon_sym_CARET,
    ACTIONS(7332), 1,
      anon_sym_AMP,
    ACTIONS(7342), 1,
      sym_test_operator,
    ACTIONS(6793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6807), 2,
      anon_sym_EQ,
      anon_sym_PIPE,
    ACTIONS(7316), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7318), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7334), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7336), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7338), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7320), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6805), 16,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_RBRACK_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [159074] = 13,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7322), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7332), 1,
      anon_sym_AMP,
    ACTIONS(7342), 1,
      sym_test_operator,
    ACTIONS(6793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7316), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7318), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7334), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7336), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7338), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(6807), 3,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
    ACTIONS(7320), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6805), 16,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_RBRACK_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [159139] = 12,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7322), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7342), 1,
      sym_test_operator,
    ACTIONS(6793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7316), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7318), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7334), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7336), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7338), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7320), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6807), 4,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(6805), 16,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_RBRACK_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [159202] = 11,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7322), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7342), 1,
      sym_test_operator,
    ACTIONS(6793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7316), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7318), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7336), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7338), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7320), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6807), 4,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(6805), 18,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_RBRACK_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [159263] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(7272), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7274), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(7250), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7248), 33,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_DASHo,
      anon_sym_AMP_AMP,
      anon_sym_DASHa,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
  [159312] = 8,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7302), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7278), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7296), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7298), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7300), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6791), 6,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(6789), 21,
      sym_test_operator,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [159367] = 9,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7322), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7342), 1,
      sym_test_operator,
    ACTIONS(6793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7316), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7318), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7320), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6807), 6,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(6805), 20,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [159424] = 7,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7322), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7318), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7320), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6807), 8,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(6805), 21,
      sym_test_operator,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [159477] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7278), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6791), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(6789), 21,
      sym_test_operator,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [159524] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7322), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7320), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6807), 10,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(6805), 21,
      sym_test_operator,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [159575] = 8,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7302), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7278), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7296), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7298), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7300), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6791), 6,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(6789), 21,
      sym_test_operator,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [159630] = 9,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7302), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7304), 1,
      sym_test_operator,
    ACTIONS(7278), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7296), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7298), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7300), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6791), 6,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(6789), 20,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [159687] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7322), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6807), 13,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6805), 21,
      sym_test_operator,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [159736] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7322), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6807), 13,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6805), 21,
      sym_test_operator,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [159785] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1241), 1,
      sym_file_descriptor,
    ACTIONS(3679), 1,
      anon_sym_DQUOTE,
    ACTIONS(7352), 1,
      sym_variable_name,
    STATE(4859), 1,
      sym_string,
    ACTIONS(7350), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(7348), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1239), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
  [159840] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1235), 1,
      sym_file_descriptor,
    ACTIONS(3679), 1,
      anon_sym_DQUOTE,
    ACTIONS(7352), 1,
      sym_variable_name,
    STATE(4859), 1,
      sym_string,
    ACTIONS(7350), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(7348), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1227), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
  [159895] = 19,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6883), 1,
      anon_sym_EQ,
    ACTIONS(6889), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(6891), 1,
      anon_sym_AMP_AMP,
    ACTIONS(6893), 1,
      anon_sym_PIPE,
    ACTIONS(6895), 1,
      anon_sym_CARET,
    ACTIONS(6897), 1,
      anon_sym_AMP,
    ACTIONS(6911), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6913), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(6915), 1,
      anon_sym_QMARK,
    ACTIONS(6885), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6899), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(6901), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(6903), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(6905), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(6907), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7354), 2,
      anon_sym_RPAREN_RPAREN,
      anon_sym_COMMA,
    ACTIONS(6909), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6887), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [159972] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1241), 1,
      sym_file_descriptor,
    ACTIONS(7358), 1,
      anon_sym_DQUOTE,
    ACTIONS(7362), 1,
      sym_variable_name,
    STATE(4457), 1,
      sym_string,
    ACTIONS(7360), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(7356), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1239), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
  [160027] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(7272), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7274), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(7250), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7266), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7268), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7270), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7264), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7248), 22,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_DASHo,
      anon_sym_AMP_AMP,
      anon_sym_DASHa,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
  [160084] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1322), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1320), 36,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_DASHo,
      anon_sym_AMP_AMP,
      anon_sym_DASHa,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
  [160129] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1330), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1328), 36,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_DASHo,
      anon_sym_AMP_AMP,
      anon_sym_DASHa,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
  [160174] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7172), 1,
      anon_sym_EQ,
    ACTIONS(7176), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7178), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7180), 1,
      anon_sym_PIPE,
    ACTIONS(7182), 1,
      anon_sym_CARET,
    ACTIONS(7184), 1,
      anon_sym_AMP,
    ACTIONS(7198), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7200), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(7202), 1,
      anon_sym_QMARK,
    ACTIONS(7206), 1,
      sym_test_operator,
    ACTIONS(7364), 1,
      anon_sym_COLON,
    ACTIONS(6793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7186), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7188), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7190), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7192), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7194), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7196), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7174), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [160253] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7220), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6807), 13,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6805), 21,
      sym_test_operator,
      anon_sym_RPAREN_RPAREN,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [160302] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1310), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1308), 36,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_DASHo,
      anon_sym_AMP_AMP,
      anon_sym_DASHa,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
  [160347] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1306), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1304), 36,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_DASHo,
      anon_sym_AMP_AMP,
      anon_sym_DASHa,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
  [160392] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1310), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1308), 36,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_DASHo,
      anon_sym_AMP_AMP,
      anon_sym_DASHa,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
  [160437] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1235), 1,
      sym_file_descriptor,
    ACTIONS(7358), 1,
      anon_sym_DQUOTE,
    ACTIONS(7362), 1,
      sym_variable_name,
    STATE(4457), 1,
      sym_string,
    ACTIONS(7360), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(7356), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1227), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
  [160492] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7366), 1,
      sym__concat,
    ACTIONS(6829), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(6827), 22,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [160539] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7368), 1,
      sym__concat,
    ACTIONS(6820), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(6818), 22,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [160586] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7220), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6807), 13,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6805), 21,
      sym_test_operator,
      anon_sym_RPAREN_RPAREN,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [160635] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7370), 1,
      anon_sym_EQ,
    ACTIONS(7088), 13,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(7092), 23,
      anon_sym_COMMA,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_DASHo,
      anon_sym_AMP_AMP,
      anon_sym_DASHa,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RPAREN,
  [160682] = 25,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6474), 1,
      sym_word,
    ACTIONS(6480), 1,
      anon_sym_LPAREN,
    ACTIONS(6488), 1,
      anon_sym_DOLLAR,
    ACTIONS(6494), 1,
      aux_sym_number_token1,
    ACTIONS(6496), 1,
      aux_sym_number_token2,
    ACTIONS(6500), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(6508), 1,
      sym_test_operator,
    ACTIONS(6510), 1,
      sym_extglob_pattern,
    ACTIONS(6512), 1,
      sym__brace_start,
    ACTIONS(7226), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(7228), 1,
      sym__special_character,
    ACTIONS(7230), 1,
      anon_sym_DQUOTE,
    ACTIONS(7234), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(7236), 1,
      anon_sym_BQUOTE,
    ACTIONS(7238), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(7372), 1,
      anon_sym_esac,
    STATE(6426), 1,
      aux_sym__literal_repeat1,
    STATE(6942), 1,
      sym_last_case_item,
    ACTIONS(7224), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(7232), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(7240), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3407), 2,
      sym_case_item,
      aux_sym_case_statement_repeat1,
    STATE(6695), 2,
      sym_concatenation,
      sym__extglob_blob,
    STATE(6303), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [160771] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7220), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7222), 1,
      sym_test_operator,
    ACTIONS(7374), 1,
      anon_sym_RPAREN_RPAREN,
    ACTIONS(7376), 1,
      anon_sym_EQ,
    ACTIONS(7380), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7382), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7384), 1,
      anon_sym_PIPE,
    ACTIONS(7386), 1,
      anon_sym_CARET,
    ACTIONS(7388), 1,
      anon_sym_AMP,
    ACTIONS(7390), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(7392), 1,
      anon_sym_QMARK,
    ACTIONS(6793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7208), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7210), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7212), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7214), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7216), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7218), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7378), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [160850] = 16,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6807), 1,
      anon_sym_EQ,
    ACTIONS(7322), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7326), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7328), 1,
      anon_sym_PIPE,
    ACTIONS(7330), 1,
      anon_sym_CARET,
    ACTIONS(7332), 1,
      anon_sym_AMP,
    ACTIONS(7342), 1,
      sym_test_operator,
    ACTIONS(6793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7316), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7318), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7334), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7336), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7338), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7320), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6805), 15,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_RBRACK_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [160921] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7394), 1,
      anon_sym_LBRACK,
    ACTIONS(6600), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(6598), 22,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [160968] = 14,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(7256), 1,
      anon_sym_PIPE,
    ACTIONS(7258), 1,
      anon_sym_CARET,
    ACTIONS(7260), 1,
      anon_sym_AMP,
    ACTIONS(7272), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7274), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(7250), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7254), 2,
      anon_sym_AMP_AMP,
      anon_sym_DASHa,
    ACTIONS(7262), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7266), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7268), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7270), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7264), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7248), 15,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_DASHo,
  [161035] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7172), 1,
      anon_sym_EQ,
    ACTIONS(7176), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7178), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7180), 1,
      anon_sym_PIPE,
    ACTIONS(7182), 1,
      anon_sym_CARET,
    ACTIONS(7184), 1,
      anon_sym_AMP,
    ACTIONS(7198), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7200), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(7202), 1,
      anon_sym_QMARK,
    ACTIONS(7206), 1,
      sym_test_operator,
    ACTIONS(7396), 1,
      anon_sym_COLON,
    ACTIONS(6793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7186), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7188), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7190), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7192), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7194), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7196), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7174), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [161114] = 25,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6474), 1,
      sym_word,
    ACTIONS(6480), 1,
      anon_sym_LPAREN,
    ACTIONS(6488), 1,
      anon_sym_DOLLAR,
    ACTIONS(6494), 1,
      aux_sym_number_token1,
    ACTIONS(6496), 1,
      aux_sym_number_token2,
    ACTIONS(6500), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(6508), 1,
      sym_test_operator,
    ACTIONS(6510), 1,
      sym_extglob_pattern,
    ACTIONS(6512), 1,
      sym__brace_start,
    ACTIONS(6580), 1,
      anon_sym_esac,
    ACTIONS(7226), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(7228), 1,
      sym__special_character,
    ACTIONS(7230), 1,
      anon_sym_DQUOTE,
    ACTIONS(7234), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(7236), 1,
      anon_sym_BQUOTE,
    ACTIONS(7238), 1,
      anon_sym_DOLLAR_BQUOTE,
    STATE(6426), 1,
      aux_sym__literal_repeat1,
    STATE(6824), 1,
      sym_last_case_item,
    ACTIONS(7224), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(7232), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(7240), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3378), 2,
      sym_case_item,
      aux_sym_case_statement_repeat1,
    STATE(6695), 2,
      sym_concatenation,
      sym__extglob_blob,
    STATE(6303), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [161203] = 7,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7220), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7216), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7218), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6807), 8,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(6805), 21,
      sym_test_operator,
      anon_sym_RPAREN_RPAREN,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [161256] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7322), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7324), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7326), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7328), 1,
      anon_sym_PIPE,
    ACTIONS(7330), 1,
      anon_sym_CARET,
    ACTIONS(7332), 1,
      anon_sym_AMP,
    ACTIONS(7340), 1,
      anon_sym_QMARK,
    ACTIONS(7342), 1,
      sym_test_operator,
    ACTIONS(7346), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(7374), 1,
      anon_sym_RBRACK_RBRACK,
    ACTIONS(7398), 1,
      anon_sym_EQ,
    ACTIONS(6793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7316), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7318), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7334), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7336), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7338), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7320), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7400), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [161335] = 9,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7322), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7342), 1,
      sym_test_operator,
    ACTIONS(6793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7316), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7318), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7320), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6791), 6,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(6789), 20,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [161392] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7402), 1,
      anon_sym_EQ,
    ACTIONS(7408), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7410), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7412), 1,
      anon_sym_PIPE,
    ACTIONS(7414), 1,
      anon_sym_CARET,
    ACTIONS(7416), 1,
      anon_sym_AMP,
    ACTIONS(7430), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7432), 1,
      anon_sym_RPAREN,
    ACTIONS(7434), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(7436), 1,
      anon_sym_QMARK,
    ACTIONS(7438), 1,
      sym_test_operator,
    ACTIONS(7404), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7418), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7420), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7422), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7424), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7426), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7428), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7406), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [161471] = 8,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7322), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7316), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7318), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7320), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6791), 6,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(6789), 21,
      sym_test_operator,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [161526] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1235), 1,
      sym_file_descriptor,
    ACTIONS(3791), 1,
      anon_sym_DQUOTE,
    ACTIONS(7444), 1,
      sym_variable_name,
    STATE(4389), 1,
      sym_string,
    ACTIONS(7442), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(7440), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1227), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
      anon_sym_LT_LT_LT,
      anon_sym_BQUOTE,
  [161581] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1342), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1340), 36,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_DASHo,
      anon_sym_AMP_AMP,
      anon_sym_DASHa,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
  [161626] = 9,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7220), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7222), 1,
      sym_test_operator,
    ACTIONS(6793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7214), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7216), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7218), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6807), 6,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(6805), 20,
      anon_sym_RPAREN_RPAREN,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [161683] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1241), 1,
      sym_file_descriptor,
    ACTIONS(3791), 1,
      anon_sym_DQUOTE,
    ACTIONS(7444), 1,
      sym_variable_name,
    STATE(4389), 1,
      sym_string,
    ACTIONS(7442), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(7440), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1239), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
      anon_sym_LT_LT_LT,
      anon_sym_BQUOTE,
  [161738] = 8,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7322), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7316), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7318), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7320), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6791), 6,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(6789), 21,
      sym_test_operator,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [161793] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6864), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(6862), 23,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [161838] = 17,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7123), 1,
      anon_sym_EQ,
    ACTIONS(7322), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7324), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7326), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7328), 1,
      anon_sym_PIPE,
    ACTIONS(7330), 1,
      anon_sym_CARET,
    ACTIONS(7332), 1,
      anon_sym_AMP,
    ACTIONS(7342), 1,
      sym_test_operator,
    ACTIONS(6793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7316), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7318), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7334), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7336), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7338), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7320), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7125), 14,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_RBRACK_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [161911] = 11,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7220), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7222), 1,
      sym_test_operator,
    ACTIONS(6793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7210), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7212), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7214), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7216), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7218), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6807), 4,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(6805), 18,
      anon_sym_RPAREN_RPAREN,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [161972] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7280), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7282), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7284), 1,
      anon_sym_PIPE,
    ACTIONS(7286), 1,
      anon_sym_CARET,
    ACTIONS(7288), 1,
      anon_sym_AMP,
    ACTIONS(7302), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7304), 1,
      sym_test_operator,
    ACTIONS(7310), 1,
      anon_sym_EQ,
    ACTIONS(7312), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(7314), 1,
      anon_sym_QMARK,
    ACTIONS(7374), 1,
      anon_sym_RBRACK,
    ACTIONS(7278), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7290), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7292), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7294), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7296), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7298), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7300), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7446), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [162051] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1273), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1369), 23,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [162096] = 8,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7198), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7192), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7194), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7196), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6807), 6,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(6805), 21,
      sym_test_operator,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      anon_sym_COLON,
  [162151] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1261), 12,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(1263), 25,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [162196] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(7450), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(7448), 36,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_DASHo,
      anon_sym_AMP_AMP,
      anon_sym_DASHa,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
  [162241] = 25,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6474), 1,
      sym_word,
    ACTIONS(6480), 1,
      anon_sym_LPAREN,
    ACTIONS(6488), 1,
      anon_sym_DOLLAR,
    ACTIONS(6494), 1,
      aux_sym_number_token1,
    ACTIONS(6496), 1,
      aux_sym_number_token2,
    ACTIONS(6500), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(6508), 1,
      sym_test_operator,
    ACTIONS(6510), 1,
      sym_extglob_pattern,
    ACTIONS(6512), 1,
      sym__brace_start,
    ACTIONS(6551), 1,
      anon_sym_esac,
    ACTIONS(7226), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(7228), 1,
      sym__special_character,
    ACTIONS(7230), 1,
      anon_sym_DQUOTE,
    ACTIONS(7234), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(7236), 1,
      anon_sym_BQUOTE,
    ACTIONS(7238), 1,
      anon_sym_DOLLAR_BQUOTE,
    STATE(6426), 1,
      aux_sym__literal_repeat1,
    STATE(7007), 1,
      sym_last_case_item,
    ACTIONS(7224), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(7232), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(7240), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3386), 2,
      sym_case_item,
      aux_sym_case_statement_repeat1,
    STATE(6695), 2,
      sym_concatenation,
      sym__extglob_blob,
    STATE(6303), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [162330] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6864), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(6862), 23,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [162375] = 13,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7220), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7222), 1,
      sym_test_operator,
    ACTIONS(7388), 1,
      anon_sym_AMP,
    ACTIONS(6793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7208), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7210), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7212), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7214), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7216), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(6807), 3,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
    ACTIONS(7218), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6805), 16,
      anon_sym_RPAREN_RPAREN,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [162440] = 17,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7123), 1,
      anon_sym_EQ,
    ACTIONS(7220), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7222), 1,
      sym_test_operator,
    ACTIONS(7380), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7382), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7384), 1,
      anon_sym_PIPE,
    ACTIONS(7386), 1,
      anon_sym_CARET,
    ACTIONS(7388), 1,
      anon_sym_AMP,
    ACTIONS(6793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7208), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7210), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7212), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7214), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7216), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7218), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7125), 14,
      anon_sym_RPAREN_RPAREN,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [162513] = 14,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7220), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7222), 1,
      sym_test_operator,
    ACTIONS(7386), 1,
      anon_sym_CARET,
    ACTIONS(7388), 1,
      anon_sym_AMP,
    ACTIONS(6793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6807), 2,
      anon_sym_EQ,
      anon_sym_PIPE,
    ACTIONS(7208), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7210), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7212), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7214), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7216), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7218), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6805), 16,
      anon_sym_RPAREN_RPAREN,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [162580] = 15,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6807), 1,
      anon_sym_EQ,
    ACTIONS(7220), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7222), 1,
      sym_test_operator,
    ACTIONS(7384), 1,
      anon_sym_PIPE,
    ACTIONS(7386), 1,
      anon_sym_CARET,
    ACTIONS(7388), 1,
      anon_sym_AMP,
    ACTIONS(6793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7208), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7210), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7212), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7214), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7216), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7218), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6805), 16,
      anon_sym_RPAREN_RPAREN,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [162649] = 16,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6807), 1,
      anon_sym_EQ,
    ACTIONS(7220), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7222), 1,
      sym_test_operator,
    ACTIONS(7382), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7384), 1,
      anon_sym_PIPE,
    ACTIONS(7386), 1,
      anon_sym_CARET,
    ACTIONS(7388), 1,
      anon_sym_AMP,
    ACTIONS(6793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7208), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7210), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7212), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7214), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7216), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7218), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6805), 15,
      anon_sym_RPAREN_RPAREN,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [162720] = 19,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7220), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7222), 1,
      sym_test_operator,
    ACTIONS(7376), 1,
      anon_sym_EQ,
    ACTIONS(7380), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7382), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7384), 1,
      anon_sym_PIPE,
    ACTIONS(7386), 1,
      anon_sym_CARET,
    ACTIONS(7388), 1,
      anon_sym_AMP,
    ACTIONS(7390), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(7392), 1,
      anon_sym_QMARK,
    ACTIONS(6793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7208), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7210), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7212), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7214), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7216), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7218), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6805), 12,
      anon_sym_RPAREN_RPAREN,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [162797] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7452), 1,
      sym__special_character,
    STATE(3065), 1,
      aux_sym__literal_repeat1,
    ACTIONS(1362), 12,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(1364), 23,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [162846] = 25,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6474), 1,
      sym_word,
    ACTIONS(6480), 1,
      anon_sym_LPAREN,
    ACTIONS(6488), 1,
      anon_sym_DOLLAR,
    ACTIONS(6494), 1,
      aux_sym_number_token1,
    ACTIONS(6496), 1,
      aux_sym_number_token2,
    ACTIONS(6500), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(6508), 1,
      sym_test_operator,
    ACTIONS(6510), 1,
      sym_extglob_pattern,
    ACTIONS(6512), 1,
      sym__brace_start,
    ACTIONS(7226), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(7228), 1,
      sym__special_character,
    ACTIONS(7230), 1,
      anon_sym_DQUOTE,
    ACTIONS(7234), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(7236), 1,
      anon_sym_BQUOTE,
    ACTIONS(7238), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(7455), 1,
      anon_sym_esac,
    STATE(6426), 1,
      aux_sym__literal_repeat1,
    STATE(7013), 1,
      sym_last_case_item,
    ACTIONS(7224), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(7232), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(7240), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3388), 2,
      sym_case_item,
      aux_sym_case_statement_repeat1,
    STATE(6695), 2,
      sym_concatenation,
      sym__extglob_blob,
    STATE(6303), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [162935] = 18,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6807), 1,
      anon_sym_EQ,
    ACTIONS(7220), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7222), 1,
      sym_test_operator,
    ACTIONS(7380), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7382), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7384), 1,
      anon_sym_PIPE,
    ACTIONS(7386), 1,
      anon_sym_CARET,
    ACTIONS(7388), 1,
      anon_sym_AMP,
    ACTIONS(7392), 1,
      anon_sym_QMARK,
    ACTIONS(6793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7208), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7210), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7212), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7214), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7216), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7218), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6805), 13,
      anon_sym_RPAREN_RPAREN,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_EQ_TILDE,
  [163010] = 8,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7220), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7214), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7216), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7218), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6807), 6,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(6805), 21,
      sym_test_operator,
      anon_sym_RPAREN_RPAREN,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [163065] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7457), 1,
      sym__concat,
    ACTIONS(6820), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(6818), 22,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [163112] = 25,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6474), 1,
      sym_word,
    ACTIONS(6480), 1,
      anon_sym_LPAREN,
    ACTIONS(6488), 1,
      anon_sym_DOLLAR,
    ACTIONS(6494), 1,
      aux_sym_number_token1,
    ACTIONS(6496), 1,
      aux_sym_number_token2,
    ACTIONS(6500), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(6508), 1,
      sym_test_operator,
    ACTIONS(6510), 1,
      sym_extglob_pattern,
    ACTIONS(6512), 1,
      sym__brace_start,
    ACTIONS(7226), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(7228), 1,
      sym__special_character,
    ACTIONS(7230), 1,
      anon_sym_DQUOTE,
    ACTIONS(7234), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(7236), 1,
      anon_sym_BQUOTE,
    ACTIONS(7238), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(7459), 1,
      anon_sym_esac,
    STATE(6426), 1,
      aux_sym__literal_repeat1,
    STATE(7016), 1,
      sym_last_case_item,
    ACTIONS(7224), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(7232), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(7240), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3389), 2,
      sym_case_item,
      aux_sym_case_statement_repeat1,
    STATE(6695), 2,
      sym_concatenation,
      sym__extglob_blob,
    STATE(6303), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [163201] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1273), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1369), 23,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [163246] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7461), 1,
      sym__concat,
    ACTIONS(6829), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(6827), 22,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [163293] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7220), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7222), 1,
      sym_test_operator,
    ACTIONS(7376), 1,
      anon_sym_EQ,
    ACTIONS(7380), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7382), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7384), 1,
      anon_sym_PIPE,
    ACTIONS(7386), 1,
      anon_sym_CARET,
    ACTIONS(7388), 1,
      anon_sym_AMP,
    ACTIONS(7390), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(7392), 1,
      anon_sym_QMARK,
    ACTIONS(7463), 1,
      anon_sym_RPAREN_RPAREN,
    ACTIONS(6793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7208), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7210), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7212), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7214), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7216), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7218), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7378), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [163372] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(7467), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(7465), 36,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_DASHo,
      anon_sym_AMP_AMP,
      anon_sym_DASHa,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
  [163417] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(7471), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(7250), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7469), 34,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_DASHo,
      anon_sym_AMP_AMP,
      anon_sym_DASHa,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
  [163464] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1354), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1352), 36,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_DASHo,
      anon_sym_AMP_AMP,
      anon_sym_DASHa,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
  [163509] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7430), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7404), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6807), 13,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6805), 21,
      sym_test_operator,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [163558] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7430), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7404), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6807), 13,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6805), 21,
      sym_test_operator,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [163607] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7430), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7404), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7428), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6807), 10,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(6805), 21,
      sym_test_operator,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [163658] = 7,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7430), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7404), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7426), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7428), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6807), 8,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(6805), 21,
      sym_test_operator,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [163711] = 9,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7430), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7438), 1,
      sym_test_operator,
    ACTIONS(7404), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7424), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7426), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7428), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6807), 6,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(6805), 20,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [163768] = 11,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7430), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7438), 1,
      sym_test_operator,
    ACTIONS(7404), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7420), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7422), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7424), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7426), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7428), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6807), 4,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(6805), 18,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_RPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [163829] = 12,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7430), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7438), 1,
      sym_test_operator,
    ACTIONS(7404), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7418), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7420), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7422), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7424), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7426), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7428), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6807), 4,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(6805), 16,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_RPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [163892] = 13,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7416), 1,
      anon_sym_AMP,
    ACTIONS(7430), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7438), 1,
      sym_test_operator,
    ACTIONS(7404), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7418), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7420), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7422), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7424), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7426), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(6807), 3,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
    ACTIONS(7428), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6805), 16,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_RPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [163957] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7280), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7282), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7284), 1,
      anon_sym_PIPE,
    ACTIONS(7286), 1,
      anon_sym_CARET,
    ACTIONS(7288), 1,
      anon_sym_AMP,
    ACTIONS(7302), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7304), 1,
      sym_test_operator,
    ACTIONS(7310), 1,
      anon_sym_EQ,
    ACTIONS(7312), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(7314), 1,
      anon_sym_QMARK,
    ACTIONS(7463), 1,
      anon_sym_RBRACK,
    ACTIONS(7278), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7290), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7292), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7294), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7296), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7298), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7300), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7446), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [164036] = 14,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7414), 1,
      anon_sym_CARET,
    ACTIONS(7416), 1,
      anon_sym_AMP,
    ACTIONS(7430), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7438), 1,
      sym_test_operator,
    ACTIONS(6807), 2,
      anon_sym_EQ,
      anon_sym_PIPE,
    ACTIONS(7404), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7418), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7420), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7422), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7424), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7426), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7428), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6805), 16,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_RPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [164103] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7322), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7324), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7326), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7328), 1,
      anon_sym_PIPE,
    ACTIONS(7330), 1,
      anon_sym_CARET,
    ACTIONS(7332), 1,
      anon_sym_AMP,
    ACTIONS(7340), 1,
      anon_sym_QMARK,
    ACTIONS(7342), 1,
      sym_test_operator,
    ACTIONS(7346), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(7398), 1,
      anon_sym_EQ,
    ACTIONS(7463), 1,
      anon_sym_RBRACK_RBRACK,
    ACTIONS(6793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7316), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7318), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7334), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7336), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7338), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7320), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7400), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [164182] = 15,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6807), 1,
      anon_sym_EQ,
    ACTIONS(7412), 1,
      anon_sym_PIPE,
    ACTIONS(7414), 1,
      anon_sym_CARET,
    ACTIONS(7416), 1,
      anon_sym_AMP,
    ACTIONS(7430), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7438), 1,
      sym_test_operator,
    ACTIONS(7404), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7418), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7420), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7422), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7424), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7426), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7428), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6805), 16,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_RPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [164251] = 9,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7198), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7206), 1,
      sym_test_operator,
    ACTIONS(6793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7192), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7194), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7196), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6791), 6,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(6789), 20,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      anon_sym_COLON,
  [164308] = 16,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6807), 1,
      anon_sym_EQ,
    ACTIONS(7410), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7412), 1,
      anon_sym_PIPE,
    ACTIONS(7414), 1,
      anon_sym_CARET,
    ACTIONS(7416), 1,
      anon_sym_AMP,
    ACTIONS(7430), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7438), 1,
      sym_test_operator,
    ACTIONS(7404), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7418), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7420), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7422), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7424), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7426), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7428), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6805), 15,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_RPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [164379] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7402), 1,
      anon_sym_EQ,
    ACTIONS(7408), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7410), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7412), 1,
      anon_sym_PIPE,
    ACTIONS(7414), 1,
      anon_sym_CARET,
    ACTIONS(7416), 1,
      anon_sym_AMP,
    ACTIONS(7430), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7434), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(7436), 1,
      anon_sym_QMARK,
    ACTIONS(7438), 1,
      sym_test_operator,
    ACTIONS(7473), 1,
      anon_sym_RPAREN,
    ACTIONS(7404), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7418), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7420), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7422), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7424), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7426), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7428), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7406), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [164458] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(7475), 1,
      aux_sym_concatenation_token1,
    ACTIONS(7477), 1,
      sym__concat,
    STATE(3120), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5395), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5393), 30,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [164509] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(7475), 1,
      aux_sym_concatenation_token1,
    ACTIONS(7477), 1,
      sym__concat,
    STATE(3117), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5391), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5389), 30,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [164560] = 19,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7402), 1,
      anon_sym_EQ,
    ACTIONS(7408), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7410), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7412), 1,
      anon_sym_PIPE,
    ACTIONS(7414), 1,
      anon_sym_CARET,
    ACTIONS(7416), 1,
      anon_sym_AMP,
    ACTIONS(7430), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7434), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(7436), 1,
      anon_sym_QMARK,
    ACTIONS(7438), 1,
      sym_test_operator,
    ACTIONS(7404), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7418), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7420), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7422), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7424), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7426), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7428), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6805), 12,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_RPAREN,
  [164637] = 18,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6807), 1,
      anon_sym_EQ,
    ACTIONS(7408), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7410), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7412), 1,
      anon_sym_PIPE,
    ACTIONS(7414), 1,
      anon_sym_CARET,
    ACTIONS(7416), 1,
      anon_sym_AMP,
    ACTIONS(7430), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7436), 1,
      anon_sym_QMARK,
    ACTIONS(7438), 1,
      sym_test_operator,
    ACTIONS(7404), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7418), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7420), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7422), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7424), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7426), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7428), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6805), 13,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_RPAREN,
      anon_sym_EQ_TILDE,
  [164712] = 8,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7198), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7192), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7194), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7196), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6791), 6,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(6789), 21,
      sym_test_operator,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      anon_sym_COLON,
  [164767] = 8,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7430), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7404), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7424), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7426), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7428), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6807), 6,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(6805), 21,
      sym_test_operator,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [164822] = 9,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7430), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7438), 1,
      sym_test_operator,
    ACTIONS(7404), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7424), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7426), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7428), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6791), 6,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(6789), 20,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [164879] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7479), 1,
      sym__concat,
    ACTIONS(6838), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(6836), 22,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [164926] = 8,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7430), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7404), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7424), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7426), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7428), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6791), 6,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(6789), 21,
      sym_test_operator,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [164981] = 8,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7220), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7214), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7216), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7218), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6791), 6,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(6789), 21,
      sym_test_operator,
      anon_sym_RPAREN_RPAREN,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [165036] = 8,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7198), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7192), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7194), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7196), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6791), 6,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(6789), 21,
      sym_test_operator,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
      anon_sym_COLON,
  [165091] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7481), 1,
      sym__concat,
    ACTIONS(6844), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(6842), 22,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [165138] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7172), 1,
      anon_sym_EQ,
    ACTIONS(7176), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7178), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7180), 1,
      anon_sym_PIPE,
    ACTIONS(7182), 1,
      anon_sym_CARET,
    ACTIONS(7184), 1,
      anon_sym_AMP,
    ACTIONS(7198), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7200), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(7202), 1,
      anon_sym_QMARK,
    ACTIONS(7206), 1,
      sym_test_operator,
    ACTIONS(7483), 1,
      anon_sym_COLON,
    ACTIONS(6793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7186), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7188), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7190), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7192), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7194), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7196), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7174), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [165217] = 25,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6474), 1,
      sym_word,
    ACTIONS(6480), 1,
      anon_sym_LPAREN,
    ACTIONS(6488), 1,
      anon_sym_DOLLAR,
    ACTIONS(6494), 1,
      aux_sym_number_token1,
    ACTIONS(6496), 1,
      aux_sym_number_token2,
    ACTIONS(6500), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(6508), 1,
      sym_test_operator,
    ACTIONS(6510), 1,
      sym_extglob_pattern,
    ACTIONS(6512), 1,
      sym__brace_start,
    ACTIONS(6616), 1,
      anon_sym_esac,
    ACTIONS(7226), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(7228), 1,
      sym__special_character,
    ACTIONS(7230), 1,
      anon_sym_DQUOTE,
    ACTIONS(7234), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(7236), 1,
      anon_sym_BQUOTE,
    ACTIONS(7238), 1,
      anon_sym_DOLLAR_BQUOTE,
    STATE(6426), 1,
      aux_sym__literal_repeat1,
    STATE(7440), 1,
      sym_last_case_item,
    ACTIONS(7224), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(7232), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(7240), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3419), 2,
      sym_case_item,
      aux_sym_case_statement_repeat1,
    STATE(6695), 2,
      sym_concatenation,
      sym__extglob_blob,
    STATE(6303), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [165306] = 25,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6474), 1,
      sym_word,
    ACTIONS(6480), 1,
      anon_sym_LPAREN,
    ACTIONS(6488), 1,
      anon_sym_DOLLAR,
    ACTIONS(6494), 1,
      aux_sym_number_token1,
    ACTIONS(6496), 1,
      aux_sym_number_token2,
    ACTIONS(6500), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(6508), 1,
      sym_test_operator,
    ACTIONS(6510), 1,
      sym_extglob_pattern,
    ACTIONS(6512), 1,
      sym__brace_start,
    ACTIONS(6610), 1,
      anon_sym_esac,
    ACTIONS(7226), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(7228), 1,
      sym__special_character,
    ACTIONS(7230), 1,
      anon_sym_DQUOTE,
    ACTIONS(7234), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(7236), 1,
      anon_sym_BQUOTE,
    ACTIONS(7238), 1,
      anon_sym_DOLLAR_BQUOTE,
    STATE(6426), 1,
      aux_sym__literal_repeat1,
    STATE(7451), 1,
      sym_last_case_item,
    ACTIONS(7224), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(7232), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(7240), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3420), 2,
      sym_case_item,
      aux_sym_case_statement_repeat1,
    STATE(6695), 2,
      sym_concatenation,
      sym__extglob_blob,
    STATE(6303), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [165395] = 8,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7430), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7404), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7424), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7426), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7428), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6791), 6,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(6789), 21,
      sym_test_operator,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [165450] = 8,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7220), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7214), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7216), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7218), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6791), 6,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(6789), 21,
      sym_test_operator,
      anon_sym_RPAREN_RPAREN,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [165505] = 25,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6474), 1,
      sym_word,
    ACTIONS(6480), 1,
      anon_sym_LPAREN,
    ACTIONS(6488), 1,
      anon_sym_DOLLAR,
    ACTIONS(6494), 1,
      aux_sym_number_token1,
    ACTIONS(6496), 1,
      aux_sym_number_token2,
    ACTIONS(6500), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(6508), 1,
      sym_test_operator,
    ACTIONS(6510), 1,
      sym_extglob_pattern,
    ACTIONS(6512), 1,
      sym__brace_start,
    ACTIONS(7226), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(7228), 1,
      sym__special_character,
    ACTIONS(7230), 1,
      anon_sym_DQUOTE,
    ACTIONS(7234), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(7236), 1,
      anon_sym_BQUOTE,
    ACTIONS(7238), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(7485), 1,
      anon_sym_esac,
    STATE(6426), 1,
      aux_sym__literal_repeat1,
    STATE(7494), 1,
      sym_last_case_item,
    ACTIONS(7224), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(7232), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(7240), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3422), 2,
      sym_case_item,
      aux_sym_case_statement_repeat1,
    STATE(6695), 2,
      sym_concatenation,
      sym__extglob_blob,
    STATE(6303), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [165594] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(7475), 1,
      aux_sym_concatenation_token1,
    ACTIONS(7477), 1,
      sym__concat,
    STATE(3120), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1263), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1261), 30,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [165645] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7402), 1,
      anon_sym_EQ,
    ACTIONS(7408), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7410), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7412), 1,
      anon_sym_PIPE,
    ACTIONS(7414), 1,
      anon_sym_CARET,
    ACTIONS(7416), 1,
      anon_sym_AMP,
    ACTIONS(7430), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7434), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(7436), 1,
      anon_sym_QMARK,
    ACTIONS(7438), 1,
      sym_test_operator,
    ACTIONS(7487), 1,
      anon_sym_RPAREN,
    ACTIONS(7404), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7418), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7420), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7422), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7424), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7426), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7428), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7406), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [165724] = 9,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7220), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7222), 1,
      sym_test_operator,
    ACTIONS(6793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7214), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7216), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7218), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6791), 6,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(6789), 20,
      anon_sym_RPAREN_RPAREN,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [165781] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(7491), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(7489), 36,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_DASHo,
      anon_sym_AMP_AMP,
      anon_sym_DASHa,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
  [165826] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7493), 1,
      sym__special_character,
    STATE(3065), 1,
      aux_sym__literal_repeat1,
    ACTIONS(5393), 12,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(5395), 23,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [165875] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(7495), 1,
      aux_sym_concatenation_token1,
    ACTIONS(7498), 1,
      sym__concat,
    STATE(3115), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1253), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1251), 30,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [165926] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(7503), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(7501), 36,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_DASHo,
      anon_sym_AMP_AMP,
      anon_sym_DASHa,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
  [165971] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(7475), 1,
      aux_sym_concatenation_token1,
    ACTIONS(7505), 1,
      sym__concat,
    STATE(3115), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1288), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1286), 30,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [166022] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(7491), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(7489), 36,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_DASHo,
      anon_sym_AMP_AMP,
      anon_sym_DASHa,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
  [166067] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(7256), 1,
      anon_sym_PIPE,
    ACTIONS(7258), 1,
      anon_sym_CARET,
    ACTIONS(7260), 1,
      anon_sym_AMP,
    ACTIONS(7272), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7511), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(7250), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7252), 2,
      anon_sym_PIPE_PIPE,
      anon_sym_DASHo,
    ACTIONS(7254), 2,
      anon_sym_AMP_AMP,
      anon_sym_DASHa,
    ACTIONS(7262), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7266), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7268), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7507), 2,
      anon_sym_SEMI,
      anon_sym_COMMA,
    ACTIONS(7270), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7264), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7509), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [166138] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(7475), 1,
      aux_sym_concatenation_token1,
    ACTIONS(7513), 1,
      sym__concat,
    STATE(3115), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1267), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1265), 30,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [166189] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6801), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(6799), 23,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [166234] = 25,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6474), 1,
      sym_word,
    ACTIONS(6480), 1,
      anon_sym_LPAREN,
    ACTIONS(6488), 1,
      anon_sym_DOLLAR,
    ACTIONS(6494), 1,
      aux_sym_number_token1,
    ACTIONS(6496), 1,
      aux_sym_number_token2,
    ACTIONS(6500), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(6508), 1,
      sym_test_operator,
    ACTIONS(6510), 1,
      sym_extglob_pattern,
    ACTIONS(6512), 1,
      sym__brace_start,
    ACTIONS(7226), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(7228), 1,
      sym__special_character,
    ACTIONS(7230), 1,
      anon_sym_DQUOTE,
    ACTIONS(7234), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(7236), 1,
      anon_sym_BQUOTE,
    ACTIONS(7238), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(7515), 1,
      anon_sym_esac,
    STATE(6426), 1,
      aux_sym__literal_repeat1,
    STATE(7495), 1,
      sym_last_case_item,
    ACTIONS(7224), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(7232), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(7240), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3423), 2,
      sym_case_item,
      aux_sym_case_statement_repeat1,
    STATE(6695), 2,
      sym_concatenation,
      sym__extglob_blob,
    STATE(6303), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [166323] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7220), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7222), 1,
      sym_test_operator,
    ACTIONS(7376), 1,
      anon_sym_EQ,
    ACTIONS(7380), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7382), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7384), 1,
      anon_sym_PIPE,
    ACTIONS(7386), 1,
      anon_sym_CARET,
    ACTIONS(7388), 1,
      anon_sym_AMP,
    ACTIONS(7390), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(7392), 1,
      anon_sym_QMARK,
    ACTIONS(7517), 1,
      anon_sym_RPAREN_RPAREN,
    ACTIONS(6793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7208), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7210), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7212), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7214), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7216), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7218), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7378), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [166402] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7280), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7282), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7284), 1,
      anon_sym_PIPE,
    ACTIONS(7286), 1,
      anon_sym_CARET,
    ACTIONS(7288), 1,
      anon_sym_AMP,
    ACTIONS(7302), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7304), 1,
      sym_test_operator,
    ACTIONS(7310), 1,
      anon_sym_EQ,
    ACTIONS(7312), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(7314), 1,
      anon_sym_QMARK,
    ACTIONS(7517), 1,
      anon_sym_RBRACK,
    ACTIONS(7278), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7290), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7292), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7294), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7296), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7298), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7300), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7446), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [166481] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7322), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7324), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7326), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7328), 1,
      anon_sym_PIPE,
    ACTIONS(7330), 1,
      anon_sym_CARET,
    ACTIONS(7332), 1,
      anon_sym_AMP,
    ACTIONS(7340), 1,
      anon_sym_QMARK,
    ACTIONS(7342), 1,
      sym_test_operator,
    ACTIONS(7346), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(7398), 1,
      anon_sym_EQ,
    ACTIONS(7517), 1,
      anon_sym_RBRACK_RBRACK,
    ACTIONS(6793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7316), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7318), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7334), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7336), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7338), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7320), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7400), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [166560] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7402), 1,
      anon_sym_EQ,
    ACTIONS(7408), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7410), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7412), 1,
      anon_sym_PIPE,
    ACTIONS(7414), 1,
      anon_sym_CARET,
    ACTIONS(7416), 1,
      anon_sym_AMP,
    ACTIONS(7430), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7434), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(7436), 1,
      anon_sym_QMARK,
    ACTIONS(7438), 1,
      sym_test_operator,
    ACTIONS(7519), 1,
      anon_sym_RPAREN,
    ACTIONS(7404), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7418), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7420), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7422), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7424), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7426), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7428), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7406), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [166639] = 12,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7534), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(7537), 1,
      sym_file_descriptor,
    ACTIONS(7540), 1,
      sym_variable_name,
    STATE(6771), 1,
      sym_subscript,
    ACTIONS(7531), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    STATE(3127), 2,
      sym_variable_assignment,
      aux_sym_command_repeat1,
    STATE(3581), 2,
      sym_file_redirect,
      sym_herestring_redirect,
    ACTIONS(7528), 3,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
    ACTIONS(7521), 5,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(7525), 5,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(7523), 14,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [166702] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7322), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7324), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7326), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7328), 1,
      anon_sym_PIPE,
    ACTIONS(7330), 1,
      anon_sym_CARET,
    ACTIONS(7332), 1,
      anon_sym_AMP,
    ACTIONS(7340), 1,
      anon_sym_QMARK,
    ACTIONS(7342), 1,
      sym_test_operator,
    ACTIONS(7346), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(7398), 1,
      anon_sym_EQ,
    ACTIONS(7543), 1,
      anon_sym_RBRACK_RBRACK,
    ACTIONS(6793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7316), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7318), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7334), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7336), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7338), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7320), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7400), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [166781] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7172), 1,
      anon_sym_EQ,
    ACTIONS(7176), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7178), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7180), 1,
      anon_sym_PIPE,
    ACTIONS(7182), 1,
      anon_sym_CARET,
    ACTIONS(7184), 1,
      anon_sym_AMP,
    ACTIONS(7198), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7200), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(7202), 1,
      anon_sym_QMARK,
    ACTIONS(7206), 1,
      sym_test_operator,
    ACTIONS(7545), 1,
      anon_sym_COLON,
    ACTIONS(6793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7186), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7188), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7190), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7192), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7194), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7196), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7174), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [166860] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7404), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6791), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(6789), 21,
      sym_test_operator,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [166907] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7280), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7282), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7284), 1,
      anon_sym_PIPE,
    ACTIONS(7286), 1,
      anon_sym_CARET,
    ACTIONS(7288), 1,
      anon_sym_AMP,
    ACTIONS(7302), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7304), 1,
      sym_test_operator,
    ACTIONS(7310), 1,
      anon_sym_EQ,
    ACTIONS(7312), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(7314), 1,
      anon_sym_QMARK,
    ACTIONS(7543), 1,
      anon_sym_RBRACK,
    ACTIONS(7278), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7290), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7292), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7294), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7296), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7298), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7300), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7446), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [166986] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1235), 1,
      sym_file_descriptor,
    ACTIONS(3791), 1,
      anon_sym_DQUOTE,
    ACTIONS(7444), 1,
      sym_variable_name,
    STATE(4389), 1,
      sym_string,
    ACTIONS(7442), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(7440), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1227), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
      anon_sym_LT_LT_LT,
  [167041] = 25,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6474), 1,
      sym_word,
    ACTIONS(6480), 1,
      anon_sym_LPAREN,
    ACTIONS(6488), 1,
      anon_sym_DOLLAR,
    ACTIONS(6494), 1,
      aux_sym_number_token1,
    ACTIONS(6496), 1,
      aux_sym_number_token2,
    ACTIONS(6500), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(6508), 1,
      sym_test_operator,
    ACTIONS(6510), 1,
      sym_extglob_pattern,
    ACTIONS(6512), 1,
      sym__brace_start,
    ACTIONS(6641), 1,
      anon_sym_esac,
    ACTIONS(7226), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(7228), 1,
      sym__special_character,
    ACTIONS(7230), 1,
      anon_sym_DQUOTE,
    ACTIONS(7234), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(7236), 1,
      anon_sym_BQUOTE,
    ACTIONS(7238), 1,
      anon_sym_DOLLAR_BQUOTE,
    STATE(6426), 1,
      aux_sym__literal_repeat1,
    STATE(7528), 1,
      sym_last_case_item,
    ACTIONS(7224), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(7232), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(7240), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3310), 2,
      sym_case_item,
      aux_sym_case_statement_repeat1,
    STATE(6695), 2,
      sym_concatenation,
      sym__extglob_blob,
    STATE(6303), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [167130] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1241), 1,
      sym_file_descriptor,
    ACTIONS(3791), 1,
      anon_sym_DQUOTE,
    ACTIONS(7444), 1,
      sym_variable_name,
    STATE(4389), 1,
      sym_string,
    ACTIONS(7442), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(7440), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1239), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
      anon_sym_LT_LT_LT,
  [167185] = 25,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6474), 1,
      sym_word,
    ACTIONS(6480), 1,
      anon_sym_LPAREN,
    ACTIONS(6488), 1,
      anon_sym_DOLLAR,
    ACTIONS(6494), 1,
      aux_sym_number_token1,
    ACTIONS(6496), 1,
      aux_sym_number_token2,
    ACTIONS(6500), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(6508), 1,
      sym_test_operator,
    ACTIONS(6510), 1,
      sym_extglob_pattern,
    ACTIONS(6512), 1,
      sym__brace_start,
    ACTIONS(6574), 1,
      anon_sym_esac,
    ACTIONS(7226), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(7228), 1,
      sym__special_character,
    ACTIONS(7230), 1,
      anon_sym_DQUOTE,
    ACTIONS(7234), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(7236), 1,
      anon_sym_BQUOTE,
    ACTIONS(7238), 1,
      anon_sym_DOLLAR_BQUOTE,
    STATE(6426), 1,
      aux_sym__literal_repeat1,
    STATE(7603), 1,
      sym_last_case_item,
    ACTIONS(7224), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(7232), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(7240), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3440), 2,
      sym_case_item,
      aux_sym_case_statement_repeat1,
    STATE(6695), 2,
      sym_concatenation,
      sym__extglob_blob,
    STATE(6303), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [167274] = 25,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6474), 1,
      sym_word,
    ACTIONS(6480), 1,
      anon_sym_LPAREN,
    ACTIONS(6488), 1,
      anon_sym_DOLLAR,
    ACTIONS(6494), 1,
      aux_sym_number_token1,
    ACTIONS(6496), 1,
      aux_sym_number_token2,
    ACTIONS(6500), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(6508), 1,
      sym_test_operator,
    ACTIONS(6510), 1,
      sym_extglob_pattern,
    ACTIONS(6512), 1,
      sym__brace_start,
    ACTIONS(7226), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(7228), 1,
      sym__special_character,
    ACTIONS(7230), 1,
      anon_sym_DQUOTE,
    ACTIONS(7234), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(7236), 1,
      anon_sym_BQUOTE,
    ACTIONS(7238), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(7547), 1,
      anon_sym_esac,
    STATE(6426), 1,
      aux_sym__literal_repeat1,
    STATE(7291), 1,
      sym_last_case_item,
    ACTIONS(7224), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(7232), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(7240), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3438), 2,
      sym_case_item,
      aux_sym_case_statement_repeat1,
    STATE(6695), 2,
      sym_concatenation,
      sym__extglob_blob,
    STATE(6303), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [167363] = 25,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6474), 1,
      sym_word,
    ACTIONS(6480), 1,
      anon_sym_LPAREN,
    ACTIONS(6488), 1,
      anon_sym_DOLLAR,
    ACTIONS(6494), 1,
      aux_sym_number_token1,
    ACTIONS(6496), 1,
      aux_sym_number_token2,
    ACTIONS(6500), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(6508), 1,
      sym_test_operator,
    ACTIONS(6510), 1,
      sym_extglob_pattern,
    ACTIONS(6512), 1,
      sym__brace_start,
    ACTIONS(7226), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(7228), 1,
      sym__special_character,
    ACTIONS(7230), 1,
      anon_sym_DQUOTE,
    ACTIONS(7234), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(7236), 1,
      anon_sym_BQUOTE,
    ACTIONS(7238), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(7549), 1,
      anon_sym_esac,
    STATE(6426), 1,
      aux_sym__literal_repeat1,
    STATE(7315), 1,
      sym_last_case_item,
    ACTIONS(7224), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(7232), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(7240), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3437), 2,
      sym_case_item,
      aux_sym_case_statement_repeat1,
    STATE(6695), 2,
      sym_concatenation,
      sym__extglob_blob,
    STATE(6303), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [167452] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6807), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(6805), 23,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [167497] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7220), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7222), 1,
      sym_test_operator,
    ACTIONS(7376), 1,
      anon_sym_EQ,
    ACTIONS(7380), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7382), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7384), 1,
      anon_sym_PIPE,
    ACTIONS(7386), 1,
      anon_sym_CARET,
    ACTIONS(7388), 1,
      anon_sym_AMP,
    ACTIONS(7390), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(7392), 1,
      anon_sym_QMARK,
    ACTIONS(7551), 1,
      anon_sym_RPAREN_RPAREN,
    ACTIONS(6793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7208), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7210), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7212), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7214), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7216), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7218), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7378), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [167576] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7280), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7282), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7284), 1,
      anon_sym_PIPE,
    ACTIONS(7286), 1,
      anon_sym_CARET,
    ACTIONS(7288), 1,
      anon_sym_AMP,
    ACTIONS(7302), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7304), 1,
      sym_test_operator,
    ACTIONS(7310), 1,
      anon_sym_EQ,
    ACTIONS(7312), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(7314), 1,
      anon_sym_QMARK,
    ACTIONS(7551), 1,
      anon_sym_RBRACK,
    ACTIONS(7278), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7290), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7292), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7294), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7296), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7298), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7300), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7446), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [167655] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7322), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7324), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7326), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7328), 1,
      anon_sym_PIPE,
    ACTIONS(7330), 1,
      anon_sym_CARET,
    ACTIONS(7332), 1,
      anon_sym_AMP,
    ACTIONS(7340), 1,
      anon_sym_QMARK,
    ACTIONS(7342), 1,
      sym_test_operator,
    ACTIONS(7346), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(7398), 1,
      anon_sym_EQ,
    ACTIONS(7551), 1,
      anon_sym_RBRACK_RBRACK,
    ACTIONS(6793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7316), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7318), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7334), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7336), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7338), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7320), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7400), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [167734] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7402), 1,
      anon_sym_EQ,
    ACTIONS(7408), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7410), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7412), 1,
      anon_sym_PIPE,
    ACTIONS(7414), 1,
      anon_sym_CARET,
    ACTIONS(7416), 1,
      anon_sym_AMP,
    ACTIONS(7430), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7434), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(7436), 1,
      anon_sym_QMARK,
    ACTIONS(7438), 1,
      sym_test_operator,
    ACTIONS(7553), 1,
      anon_sym_RPAREN,
    ACTIONS(7404), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7418), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7420), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7422), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7424), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7426), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7428), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7406), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [167813] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7172), 1,
      anon_sym_EQ,
    ACTIONS(7176), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7178), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7180), 1,
      anon_sym_PIPE,
    ACTIONS(7182), 1,
      anon_sym_CARET,
    ACTIONS(7184), 1,
      anon_sym_AMP,
    ACTIONS(7198), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7200), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(7202), 1,
      anon_sym_QMARK,
    ACTIONS(7206), 1,
      sym_test_operator,
    ACTIONS(7555), 1,
      anon_sym_COLON,
    ACTIONS(6793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7186), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7188), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7190), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7192), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7194), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7196), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7174), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [167892] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7220), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7222), 1,
      sym_test_operator,
    ACTIONS(7376), 1,
      anon_sym_EQ,
    ACTIONS(7380), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7382), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7384), 1,
      anon_sym_PIPE,
    ACTIONS(7386), 1,
      anon_sym_CARET,
    ACTIONS(7388), 1,
      anon_sym_AMP,
    ACTIONS(7390), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(7392), 1,
      anon_sym_QMARK,
    ACTIONS(7557), 1,
      anon_sym_RPAREN_RPAREN,
    ACTIONS(6793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7208), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7210), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7212), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7214), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7216), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7218), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7378), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [167971] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7280), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7282), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7284), 1,
      anon_sym_PIPE,
    ACTIONS(7286), 1,
      anon_sym_CARET,
    ACTIONS(7288), 1,
      anon_sym_AMP,
    ACTIONS(7302), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7304), 1,
      sym_test_operator,
    ACTIONS(7310), 1,
      anon_sym_EQ,
    ACTIONS(7312), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(7314), 1,
      anon_sym_QMARK,
    ACTIONS(7557), 1,
      anon_sym_RBRACK,
    ACTIONS(7278), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7290), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7292), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7294), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7296), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7298), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7300), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7446), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [168050] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6807), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(6805), 23,
      sym_test_operator,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [168095] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7322), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7324), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7326), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7328), 1,
      anon_sym_PIPE,
    ACTIONS(7330), 1,
      anon_sym_CARET,
    ACTIONS(7332), 1,
      anon_sym_AMP,
    ACTIONS(7340), 1,
      anon_sym_QMARK,
    ACTIONS(7342), 1,
      sym_test_operator,
    ACTIONS(7346), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(7398), 1,
      anon_sym_EQ,
    ACTIONS(7557), 1,
      anon_sym_RBRACK_RBRACK,
    ACTIONS(6793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7316), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7318), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7334), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7336), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7338), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7320), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7400), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [168174] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7172), 1,
      anon_sym_EQ,
    ACTIONS(7176), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7178), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7180), 1,
      anon_sym_PIPE,
    ACTIONS(7182), 1,
      anon_sym_CARET,
    ACTIONS(7184), 1,
      anon_sym_AMP,
    ACTIONS(7198), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7200), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(7202), 1,
      anon_sym_QMARK,
    ACTIONS(7206), 1,
      sym_test_operator,
    ACTIONS(7559), 1,
      anon_sym_COLON,
    ACTIONS(6793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7186), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7188), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7190), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7192), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7194), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7196), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7174), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [168253] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1261), 13,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      sym__special_character,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(1263), 24,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [168298] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7220), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7222), 1,
      sym_test_operator,
    ACTIONS(7376), 1,
      anon_sym_EQ,
    ACTIONS(7380), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7382), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7384), 1,
      anon_sym_PIPE,
    ACTIONS(7386), 1,
      anon_sym_CARET,
    ACTIONS(7388), 1,
      anon_sym_AMP,
    ACTIONS(7390), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(7392), 1,
      anon_sym_QMARK,
    ACTIONS(7561), 1,
      anon_sym_RPAREN_RPAREN,
    ACTIONS(6793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7208), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7210), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7212), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7214), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7216), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7218), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7378), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [168377] = 17,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7123), 1,
      anon_sym_EQ,
    ACTIONS(7408), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7410), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7412), 1,
      anon_sym_PIPE,
    ACTIONS(7414), 1,
      anon_sym_CARET,
    ACTIONS(7416), 1,
      anon_sym_AMP,
    ACTIONS(7430), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7438), 1,
      sym_test_operator,
    ACTIONS(7404), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7418), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7420), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7422), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7424), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7426), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7428), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7125), 14,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_RPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [168450] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7280), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7282), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7284), 1,
      anon_sym_PIPE,
    ACTIONS(7286), 1,
      anon_sym_CARET,
    ACTIONS(7288), 1,
      anon_sym_AMP,
    ACTIONS(7302), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7304), 1,
      sym_test_operator,
    ACTIONS(7310), 1,
      anon_sym_EQ,
    ACTIONS(7312), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(7314), 1,
      anon_sym_QMARK,
    ACTIONS(7561), 1,
      anon_sym_RBRACK,
    ACTIONS(7278), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7290), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7292), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7294), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7296), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7298), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7300), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7446), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [168529] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7322), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7324), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7326), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7328), 1,
      anon_sym_PIPE,
    ACTIONS(7330), 1,
      anon_sym_CARET,
    ACTIONS(7332), 1,
      anon_sym_AMP,
    ACTIONS(7340), 1,
      anon_sym_QMARK,
    ACTIONS(7342), 1,
      sym_test_operator,
    ACTIONS(7346), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(7398), 1,
      anon_sym_EQ,
    ACTIONS(7561), 1,
      anon_sym_RBRACK_RBRACK,
    ACTIONS(6793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7316), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7318), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7334), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7336), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7338), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7320), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7400), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [168608] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7220), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7222), 1,
      sym_test_operator,
    ACTIONS(7376), 1,
      anon_sym_EQ,
    ACTIONS(7380), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7382), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7384), 1,
      anon_sym_PIPE,
    ACTIONS(7386), 1,
      anon_sym_CARET,
    ACTIONS(7388), 1,
      anon_sym_AMP,
    ACTIONS(7390), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(7392), 1,
      anon_sym_QMARK,
    ACTIONS(7563), 1,
      anon_sym_RPAREN_RPAREN,
    ACTIONS(6793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7208), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7210), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7212), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7214), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7216), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7218), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7378), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [168687] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7280), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7282), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7284), 1,
      anon_sym_PIPE,
    ACTIONS(7286), 1,
      anon_sym_CARET,
    ACTIONS(7288), 1,
      anon_sym_AMP,
    ACTIONS(7302), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7304), 1,
      sym_test_operator,
    ACTIONS(7310), 1,
      anon_sym_EQ,
    ACTIONS(7312), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(7314), 1,
      anon_sym_QMARK,
    ACTIONS(7563), 1,
      anon_sym_RBRACK,
    ACTIONS(7278), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7290), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7292), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7294), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7296), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7298), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7300), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7446), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [168766] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7322), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7324), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7326), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7328), 1,
      anon_sym_PIPE,
    ACTIONS(7330), 1,
      anon_sym_CARET,
    ACTIONS(7332), 1,
      anon_sym_AMP,
    ACTIONS(7340), 1,
      anon_sym_QMARK,
    ACTIONS(7342), 1,
      sym_test_operator,
    ACTIONS(7346), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(7398), 1,
      anon_sym_EQ,
    ACTIONS(7563), 1,
      anon_sym_RBRACK_RBRACK,
    ACTIONS(6793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7316), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7318), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7334), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7336), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7338), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7320), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7400), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [168845] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1314), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1312), 36,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_DASHo,
      anon_sym_AMP_AMP,
      anon_sym_DASHa,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
  [168890] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7220), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7222), 1,
      sym_test_operator,
    ACTIONS(7376), 1,
      anon_sym_EQ,
    ACTIONS(7380), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7382), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7384), 1,
      anon_sym_PIPE,
    ACTIONS(7386), 1,
      anon_sym_CARET,
    ACTIONS(7388), 1,
      anon_sym_AMP,
    ACTIONS(7390), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(7392), 1,
      anon_sym_QMARK,
    ACTIONS(7543), 1,
      anon_sym_RPAREN_RPAREN,
    ACTIONS(6793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7208), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7210), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7212), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7214), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7216), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7218), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7378), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [168969] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1346), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1344), 36,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_DASHo,
      anon_sym_AMP_AMP,
      anon_sym_DASHa,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
  [169014] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1358), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1356), 36,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_DASHo,
      anon_sym_AMP_AMP,
      anon_sym_DASHa,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
  [169059] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1338), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1336), 36,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_DASHo,
      anon_sym_AMP_AMP,
      anon_sym_DASHa,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
  [169104] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3791), 1,
      anon_sym_DQUOTE,
    ACTIONS(7444), 1,
      sym_variable_name,
    STATE(4389), 1,
      sym_string,
    ACTIONS(1235), 2,
      sym_file_descriptor,
      ts_builtin_sym_end,
    ACTIONS(7442), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(7440), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1227), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
      anon_sym_LT_LT_LT,
  [169159] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3791), 1,
      anon_sym_DQUOTE,
    ACTIONS(7444), 1,
      sym_variable_name,
    STATE(4389), 1,
      sym_string,
    ACTIONS(1241), 2,
      sym_file_descriptor,
      ts_builtin_sym_end,
    ACTIONS(7442), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(7440), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1239), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
      anon_sym_LT_LT_LT,
  [169214] = 25,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6474), 1,
      sym_word,
    ACTIONS(6480), 1,
      anon_sym_LPAREN,
    ACTIONS(6488), 1,
      anon_sym_DOLLAR,
    ACTIONS(6494), 1,
      aux_sym_number_token1,
    ACTIONS(6496), 1,
      aux_sym_number_token2,
    ACTIONS(6500), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(6508), 1,
      sym_test_operator,
    ACTIONS(6510), 1,
      sym_extglob_pattern,
    ACTIONS(6512), 1,
      sym__brace_start,
    ACTIONS(7226), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(7228), 1,
      sym__special_character,
    ACTIONS(7230), 1,
      anon_sym_DQUOTE,
    ACTIONS(7234), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(7236), 1,
      anon_sym_BQUOTE,
    ACTIONS(7238), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(7565), 1,
      anon_sym_esac,
    STATE(6426), 1,
      aux_sym__literal_repeat1,
    STATE(6923), 1,
      sym_last_case_item,
    ACTIONS(7224), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(7232), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(7240), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3424), 2,
      sym_case_item,
      aux_sym_case_statement_repeat1,
    STATE(6695), 2,
      sym_concatenation,
      sym__extglob_blob,
    STATE(6303), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [169303] = 19,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7025), 1,
      anon_sym_RBRACK,
    ACTIONS(7567), 1,
      anon_sym_EQ,
    ACTIONS(7573), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7575), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7577), 1,
      anon_sym_PIPE,
    ACTIONS(7579), 1,
      anon_sym_CARET,
    ACTIONS(7581), 1,
      anon_sym_AMP,
    ACTIONS(7595), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7597), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(7599), 1,
      anon_sym_QMARK,
    ACTIONS(7569), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7583), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7585), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7587), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7589), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7591), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7593), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7571), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [169379] = 19,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6969), 1,
      anon_sym_RBRACK,
    ACTIONS(7567), 1,
      anon_sym_EQ,
    ACTIONS(7573), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7575), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7577), 1,
      anon_sym_PIPE,
    ACTIONS(7579), 1,
      anon_sym_CARET,
    ACTIONS(7581), 1,
      anon_sym_AMP,
    ACTIONS(7595), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7597), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(7599), 1,
      anon_sym_QMARK,
    ACTIONS(7569), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7583), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7585), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7587), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7589), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7591), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7593), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7571), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [169455] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1340), 13,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1342), 23,
      anon_sym_RPAREN_RPAREN,
      anon_sym_COMMA,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_DASHo,
      anon_sym_AMP_AMP,
      anon_sym_DASHa,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
  [169499] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7166), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(7164), 22,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [169543] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7162), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(7160), 22,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [169587] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1312), 13,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1314), 23,
      anon_sym_RPAREN_RPAREN,
      anon_sym_COMMA,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_DASHo,
      anon_sym_AMP_AMP,
      anon_sym_DASHa,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
  [169631] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1308), 13,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1310), 23,
      anon_sym_RPAREN_RPAREN,
      anon_sym_COMMA,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_DASHo,
      anon_sym_AMP_AMP,
      anon_sym_DASHa,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
  [169675] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1304), 13,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1306), 23,
      anon_sym_RPAREN_RPAREN,
      anon_sym_COMMA,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_DASHo,
      anon_sym_AMP_AMP,
      anon_sym_DASHa,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
  [169719] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1308), 13,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1310), 23,
      anon_sym_RPAREN_RPAREN,
      anon_sym_COMMA,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_DASHo,
      anon_sym_AMP_AMP,
      anon_sym_DASHa,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
  [169763] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1336), 13,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1338), 23,
      anon_sym_RPAREN_RPAREN,
      anon_sym_COMMA,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_DASHo,
      anon_sym_AMP_AMP,
      anon_sym_DASHa,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
  [169807] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1330), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1328), 31,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [169851] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1320), 13,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1322), 23,
      anon_sym_RPAREN_RPAREN,
      anon_sym_COMMA,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_DASHo,
      anon_sym_AMP_AMP,
      anon_sym_DASHa,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
  [169895] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1358), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1356), 31,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [169939] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1308), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1310), 22,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [169983] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1304), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1306), 22,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [170027] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1308), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1310), 22,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [170071] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1346), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1344), 31,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [170115] = 19,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7017), 1,
      anon_sym_RBRACK,
    ACTIONS(7567), 1,
      anon_sym_EQ,
    ACTIONS(7573), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7575), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7577), 1,
      anon_sym_PIPE,
    ACTIONS(7579), 1,
      anon_sym_CARET,
    ACTIONS(7581), 1,
      anon_sym_AMP,
    ACTIONS(7595), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7597), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(7599), 1,
      anon_sym_QMARK,
    ACTIONS(7569), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7583), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7585), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7587), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7589), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7591), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7593), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7571), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [170191] = 8,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7607), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7601), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7603), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7605), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6791), 6,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(6789), 20,
      sym_test_operator,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [170245] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1318), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1316), 31,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [170289] = 19,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7031), 1,
      anon_sym_RBRACK,
    ACTIONS(7567), 1,
      anon_sym_EQ,
    ACTIONS(7573), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7575), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7577), 1,
      anon_sym_PIPE,
    ACTIONS(7579), 1,
      anon_sym_CARET,
    ACTIONS(7581), 1,
      anon_sym_AMP,
    ACTIONS(7595), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7597), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(7599), 1,
      anon_sym_QMARK,
    ACTIONS(7569), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7583), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7585), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7587), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7589), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7591), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7593), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7571), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [170365] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1241), 1,
      sym_file_descriptor,
    ACTIONS(4084), 1,
      anon_sym_DQUOTE,
    ACTIONS(7613), 1,
      sym_variable_name,
    STATE(4606), 1,
      sym_string,
    ACTIONS(7611), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(7609), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1239), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
      anon_sym_BQUOTE,
  [170419] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1235), 1,
      sym_file_descriptor,
    ACTIONS(4084), 1,
      anon_sym_DQUOTE,
    ACTIONS(7613), 1,
      sym_variable_name,
    STATE(4606), 1,
      sym_string,
    ACTIONS(7611), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(7609), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1227), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
      anon_sym_BQUOTE,
  [170473] = 16,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7102), 1,
      anon_sym_EQ,
    ACTIONS(7617), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7619), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7621), 1,
      anon_sym_PIPE,
    ACTIONS(7623), 1,
      anon_sym_CARET,
    ACTIONS(7625), 1,
      anon_sym_AMP,
    ACTIONS(7639), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7615), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7627), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7629), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7631), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7633), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7635), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7637), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7100), 14,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_RPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [170543] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7448), 13,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(7450), 23,
      anon_sym_RPAREN_RPAREN,
      anon_sym_COMMA,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_DASHo,
      anon_sym_AMP_AMP,
      anon_sym_DASHa,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
  [170587] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7166), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(7164), 22,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [170631] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1304), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1306), 22,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [170675] = 19,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7015), 1,
      anon_sym_RBRACK,
    ACTIONS(7567), 1,
      anon_sym_EQ,
    ACTIONS(7573), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7575), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7577), 1,
      anon_sym_PIPE,
    ACTIONS(7579), 1,
      anon_sym_CARET,
    ACTIONS(7581), 1,
      anon_sym_AMP,
    ACTIONS(7595), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7597), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(7599), 1,
      anon_sym_QMARK,
    ACTIONS(7569), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7583), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7585), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7587), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7589), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7591), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7593), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7571), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [170751] = 8,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7607), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7601), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7603), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7605), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6791), 6,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(6789), 20,
      sym_test_operator,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [170805] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7162), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(7160), 22,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [170849] = 19,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6883), 1,
      anon_sym_EQ,
    ACTIONS(6889), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(6891), 1,
      anon_sym_AMP_AMP,
    ACTIONS(6893), 1,
      anon_sym_PIPE,
    ACTIONS(6895), 1,
      anon_sym_CARET,
    ACTIONS(6897), 1,
      anon_sym_AMP,
    ACTIONS(6911), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6913), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(6915), 1,
      anon_sym_QMARK,
    ACTIONS(7641), 1,
      anon_sym_COLON,
    ACTIONS(6885), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6899), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(6901), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(6903), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(6905), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(6907), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(6909), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6887), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [170925] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7645), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7643), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7248), 12,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7274), 21,
      anon_sym_RPAREN_RPAREN,
      anon_sym_COMMA,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_DASHo,
      anon_sym_AMP_AMP,
      anon_sym_DASHa,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
  [170973] = 9,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7607), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7647), 1,
      sym_test_operator,
    ACTIONS(6793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7601), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7603), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7605), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6791), 6,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(6789), 19,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [171029] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7645), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7643), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7248), 12,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7274), 21,
      anon_sym_RPAREN_RPAREN,
      anon_sym_COMMA,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_DASHo,
      anon_sym_AMP_AMP,
      anon_sym_DASHa,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
  [171077] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7645), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7643), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7649), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7248), 9,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7274), 21,
      anon_sym_RPAREN_RPAREN,
      anon_sym_COMMA,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_DASHo,
      anon_sym_AMP_AMP,
      anon_sym_DASHa,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
  [171127] = 7,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7645), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7643), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7651), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7649), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7248), 7,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7274), 21,
      anon_sym_RPAREN_RPAREN,
      anon_sym_COMMA,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_DASHo,
      anon_sym_AMP_AMP,
      anon_sym_DASHa,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
  [171179] = 8,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7645), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7643), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7651), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7653), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7649), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7248), 5,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7274), 21,
      anon_sym_RPAREN_RPAREN,
      anon_sym_COMMA,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_DASHo,
      anon_sym_AMP_AMP,
      anon_sym_DASHa,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
  [171233] = 10,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7645), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7643), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7651), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7653), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7655), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7657), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7248), 3,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(7649), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7274), 19,
      anon_sym_RPAREN_RPAREN,
      anon_sym_COMMA,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_DASHo,
      anon_sym_AMP_AMP,
      anon_sym_DASHa,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
  [171291] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1340), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1342), 22,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [171335] = 19,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7013), 1,
      anon_sym_RBRACK,
    ACTIONS(7567), 1,
      anon_sym_EQ,
    ACTIONS(7573), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7575), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7577), 1,
      anon_sym_PIPE,
    ACTIONS(7579), 1,
      anon_sym_CARET,
    ACTIONS(7581), 1,
      anon_sym_AMP,
    ACTIONS(7595), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7597), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(7599), 1,
      anon_sym_QMARK,
    ACTIONS(7569), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7583), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7585), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7587), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7589), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7591), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7593), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7571), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [171411] = 11,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7645), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7643), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7651), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7653), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7655), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7657), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7659), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7248), 3,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(7649), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7274), 17,
      anon_sym_RPAREN_RPAREN,
      anon_sym_COMMA,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_DASHo,
      anon_sym_AMP_AMP,
      anon_sym_DASHa,
  [171471] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1261), 12,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(1263), 24,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [171515] = 12,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7645), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7661), 1,
      anon_sym_AMP,
    ACTIONS(7248), 2,
      anon_sym_PIPE,
      anon_sym_CARET,
    ACTIONS(7643), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7651), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7653), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7655), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7657), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7659), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7649), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7274), 17,
      anon_sym_RPAREN_RPAREN,
      anon_sym_COMMA,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_DASHo,
      anon_sym_AMP_AMP,
      anon_sym_DASHa,
  [171577] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6820), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(6818), 22,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [171621] = 19,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7055), 1,
      anon_sym_RBRACK,
    ACTIONS(7567), 1,
      anon_sym_EQ,
    ACTIONS(7573), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7575), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7577), 1,
      anon_sym_PIPE,
    ACTIONS(7579), 1,
      anon_sym_CARET,
    ACTIONS(7581), 1,
      anon_sym_AMP,
    ACTIONS(7595), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7597), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(7599), 1,
      anon_sym_QMARK,
    ACTIONS(7569), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7583), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7585), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7587), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7589), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7591), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7593), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7571), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [171697] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1344), 13,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1346), 23,
      anon_sym_RPAREN_RPAREN,
      anon_sym_COMMA,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_DASHo,
      anon_sym_AMP_AMP,
      anon_sym_DASHa,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
  [171741] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1336), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1338), 22,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [171785] = 19,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7011), 1,
      anon_sym_RBRACK,
    ACTIONS(7567), 1,
      anon_sym_EQ,
    ACTIONS(7573), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7575), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7577), 1,
      anon_sym_PIPE,
    ACTIONS(7579), 1,
      anon_sym_CARET,
    ACTIONS(7581), 1,
      anon_sym_AMP,
    ACTIONS(7595), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7597), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(7599), 1,
      anon_sym_QMARK,
    ACTIONS(7569), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7583), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7585), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7587), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7589), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7591), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7593), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7571), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [171861] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1356), 13,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1358), 23,
      anon_sym_RPAREN_RPAREN,
      anon_sym_COMMA,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_DASHo,
      anon_sym_AMP_AMP,
      anon_sym_DASHa,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
  [171905] = 13,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7248), 1,
      anon_sym_PIPE,
    ACTIONS(7645), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7661), 1,
      anon_sym_AMP,
    ACTIONS(7663), 1,
      anon_sym_CARET,
    ACTIONS(7643), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7651), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7653), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7655), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7657), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7659), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7649), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7274), 17,
      anon_sym_RPAREN_RPAREN,
      anon_sym_COMMA,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_DASHo,
      anon_sym_AMP_AMP,
      anon_sym_DASHa,
  [171969] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1328), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1330), 22,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [172013] = 13,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7645), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7661), 1,
      anon_sym_AMP,
    ACTIONS(7663), 1,
      anon_sym_CARET,
    ACTIONS(7665), 1,
      anon_sym_PIPE,
    ACTIONS(7643), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7651), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7653), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7655), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7657), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7659), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7649), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7274), 17,
      anon_sym_RPAREN_RPAREN,
      anon_sym_COMMA,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_DASHo,
      anon_sym_AMP_AMP,
      anon_sym_DASHa,
  [172077] = 14,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7645), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7661), 1,
      anon_sym_AMP,
    ACTIONS(7663), 1,
      anon_sym_CARET,
    ACTIONS(7665), 1,
      anon_sym_PIPE,
    ACTIONS(7643), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7651), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7653), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7655), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7657), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7659), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7667), 2,
      anon_sym_AMP_AMP,
      anon_sym_DASHa,
    ACTIONS(7649), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7274), 15,
      anon_sym_RPAREN_RPAREN,
      anon_sym_COMMA,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_DASHo,
  [172143] = 15,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7645), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7661), 1,
      anon_sym_AMP,
    ACTIONS(7663), 1,
      anon_sym_CARET,
    ACTIONS(7665), 1,
      anon_sym_PIPE,
    ACTIONS(7643), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7651), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7653), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7655), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7657), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7659), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7667), 2,
      anon_sym_AMP_AMP,
      anon_sym_DASHa,
    ACTIONS(7669), 2,
      anon_sym_PIPE_PIPE,
      anon_sym_DASHo,
    ACTIONS(7649), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7274), 13,
      anon_sym_RPAREN_RPAREN,
      anon_sym_COMMA,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [172211] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6829), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(6827), 22,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [172255] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1322), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1320), 31,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [172299] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1338), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1336), 31,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [172343] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1312), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1314), 22,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [172387] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1344), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1346), 22,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [172431] = 19,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7009), 1,
      anon_sym_RBRACK,
    ACTIONS(7567), 1,
      anon_sym_EQ,
    ACTIONS(7573), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7575), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7577), 1,
      anon_sym_PIPE,
    ACTIONS(7579), 1,
      anon_sym_CARET,
    ACTIONS(7581), 1,
      anon_sym_AMP,
    ACTIONS(7595), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7597), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(7599), 1,
      anon_sym_QMARK,
    ACTIONS(7569), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7583), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7585), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7587), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7589), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7591), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7593), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7571), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [172507] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1342), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1340), 31,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [172551] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1314), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1312), 31,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [172595] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(7673), 1,
      anon_sym_DQUOTE,
    ACTIONS(7677), 1,
      sym_variable_name,
    STATE(4393), 1,
      sym_string,
    ACTIONS(1241), 2,
      sym_file_descriptor,
      ts_builtin_sym_end,
    ACTIONS(7675), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(7671), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1239), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
  [172649] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7244), 13,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(7246), 23,
      anon_sym_RPAREN_RPAREN,
      anon_sym_COMMA,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_DASHo,
      anon_sym_AMP_AMP,
      anon_sym_DASHa,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
  [172693] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1354), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1352), 31,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [172737] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(7673), 1,
      anon_sym_DQUOTE,
    ACTIONS(7677), 1,
      sym_variable_name,
    STATE(4393), 1,
      sym_string,
    ACTIONS(1235), 2,
      sym_file_descriptor,
      ts_builtin_sym_end,
    ACTIONS(7675), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(7671), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1227), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
  [172791] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6820), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(6818), 22,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [172835] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6829), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(6827), 22,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [172879] = 24,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6474), 1,
      sym_word,
    ACTIONS(6480), 1,
      anon_sym_LPAREN,
    ACTIONS(6488), 1,
      anon_sym_DOLLAR,
    ACTIONS(6494), 1,
      aux_sym_number_token1,
    ACTIONS(6496), 1,
      aux_sym_number_token2,
    ACTIONS(6500), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(6508), 1,
      sym_test_operator,
    ACTIONS(6510), 1,
      sym_extglob_pattern,
    ACTIONS(6512), 1,
      sym__brace_start,
    ACTIONS(7226), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(7228), 1,
      sym__special_character,
    ACTIONS(7230), 1,
      anon_sym_DQUOTE,
    ACTIONS(7234), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(7236), 1,
      anon_sym_BQUOTE,
    ACTIONS(7238), 1,
      anon_sym_DOLLAR_BQUOTE,
    STATE(6426), 1,
      aux_sym__literal_repeat1,
    STATE(6851), 1,
      sym_last_case_item,
    ACTIONS(7224), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(7232), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(7240), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3477), 2,
      sym_case_item,
      aux_sym_case_statement_repeat1,
    STATE(6695), 2,
      sym_concatenation,
      sym__extglob_blob,
    STATE(6303), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [172965] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1308), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1310), 22,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [173009] = 19,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7007), 1,
      anon_sym_RBRACK,
    ACTIONS(7567), 1,
      anon_sym_EQ,
    ACTIONS(7573), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7575), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7577), 1,
      anon_sym_PIPE,
    ACTIONS(7579), 1,
      anon_sym_CARET,
    ACTIONS(7581), 1,
      anon_sym_AMP,
    ACTIONS(7595), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7597), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(7599), 1,
      anon_sym_QMARK,
    ACTIONS(7569), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7583), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7585), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7587), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7589), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7591), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7593), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7571), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [173085] = 24,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6474), 1,
      sym_word,
    ACTIONS(6480), 1,
      anon_sym_LPAREN,
    ACTIONS(6488), 1,
      anon_sym_DOLLAR,
    ACTIONS(6494), 1,
      aux_sym_number_token1,
    ACTIONS(6496), 1,
      aux_sym_number_token2,
    ACTIONS(6500), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(6508), 1,
      sym_test_operator,
    ACTIONS(6510), 1,
      sym_extglob_pattern,
    ACTIONS(6512), 1,
      sym__brace_start,
    ACTIONS(7226), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(7228), 1,
      sym__special_character,
    ACTIONS(7230), 1,
      anon_sym_DQUOTE,
    ACTIONS(7234), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(7236), 1,
      anon_sym_BQUOTE,
    ACTIONS(7238), 1,
      anon_sym_DOLLAR_BQUOTE,
    STATE(6426), 1,
      aux_sym__literal_repeat1,
    STATE(6824), 1,
      sym_last_case_item,
    ACTIONS(7224), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(7232), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(7240), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3477), 2,
      sym_case_item,
      aux_sym_case_statement_repeat1,
    STATE(6695), 2,
      sym_concatenation,
      sym__extglob_blob,
    STATE(6303), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [173171] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1241), 1,
      sym_file_descriptor,
    ACTIONS(7673), 1,
      anon_sym_DQUOTE,
    ACTIONS(7677), 1,
      sym_variable_name,
    STATE(4393), 1,
      sym_string,
    ACTIONS(7675), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(7671), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1239), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
  [173225] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1235), 1,
      sym_file_descriptor,
    ACTIONS(7673), 1,
      anon_sym_DQUOTE,
    ACTIONS(7677), 1,
      sym_variable_name,
    STATE(4393), 1,
      sym_string,
    ACTIONS(7675), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(7671), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1227), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
  [173279] = 19,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7066), 1,
      anon_sym_RBRACK,
    ACTIONS(7567), 1,
      anon_sym_EQ,
    ACTIONS(7573), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7575), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7577), 1,
      anon_sym_PIPE,
    ACTIONS(7579), 1,
      anon_sym_CARET,
    ACTIONS(7581), 1,
      anon_sym_AMP,
    ACTIONS(7595), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7597), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(7599), 1,
      anon_sym_QMARK,
    ACTIONS(7569), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7583), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7585), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7587), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7589), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7591), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7593), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7571), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [173355] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4084), 1,
      anon_sym_DQUOTE,
    ACTIONS(7613), 1,
      sym_variable_name,
    STATE(4606), 1,
      sym_string,
    ACTIONS(1241), 2,
      sym_file_descriptor,
      ts_builtin_sym_end,
    ACTIONS(7611), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(7609), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1239), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
  [173409] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1356), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1358), 22,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [173453] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7615), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7139), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(7137), 20,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [173499] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4084), 1,
      anon_sym_DQUOTE,
    ACTIONS(7613), 1,
      sym_variable_name,
    STATE(4606), 1,
      sym_string,
    ACTIONS(1235), 2,
      sym_file_descriptor,
      ts_builtin_sym_end,
    ACTIONS(7611), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(7609), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1227), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
  [173553] = 19,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7005), 1,
      anon_sym_RBRACK,
    ACTIONS(7567), 1,
      anon_sym_EQ,
    ACTIONS(7573), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7575), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7577), 1,
      anon_sym_PIPE,
    ACTIONS(7579), 1,
      anon_sym_CARET,
    ACTIONS(7581), 1,
      anon_sym_AMP,
    ACTIONS(7595), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7597), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(7599), 1,
      anon_sym_QMARK,
    ACTIONS(7569), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7583), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7585), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7587), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7589), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7591), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7593), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7571), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [173629] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1328), 13,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1330), 23,
      anon_sym_RPAREN_RPAREN,
      anon_sym_COMMA,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_DASHo,
      anon_sym_AMP_AMP,
      anon_sym_DASHa,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
  [173673] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7639), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7615), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7139), 13,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7137), 20,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [173721] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7639), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7615), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7637), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7139), 10,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7137), 20,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [173771] = 7,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7639), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7615), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7635), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7637), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7139), 8,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7137), 20,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [173823] = 8,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7639), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7615), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7633), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7635), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7637), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7139), 6,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7137), 20,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [173877] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1310), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1308), 31,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [173921] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1308), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1310), 22,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [173965] = 19,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7003), 1,
      anon_sym_RBRACK,
    ACTIONS(7567), 1,
      anon_sym_EQ,
    ACTIONS(7573), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7575), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7577), 1,
      anon_sym_PIPE,
    ACTIONS(7579), 1,
      anon_sym_CARET,
    ACTIONS(7581), 1,
      anon_sym_AMP,
    ACTIONS(7595), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7597), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(7599), 1,
      anon_sym_QMARK,
    ACTIONS(7569), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7583), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7585), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7587), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7589), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7591), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7593), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7571), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [174041] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1306), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1304), 31,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [174085] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1316), 13,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1318), 23,
      anon_sym_RPAREN_RPAREN,
      anon_sym_COMMA,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_DASHo,
      anon_sym_AMP_AMP,
      anon_sym_DASHa,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
  [174129] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1312), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1314), 22,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [174173] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1310), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1308), 31,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [174217] = 19,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7001), 1,
      anon_sym_RBRACK,
    ACTIONS(7567), 1,
      anon_sym_EQ,
    ACTIONS(7573), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7575), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7577), 1,
      anon_sym_PIPE,
    ACTIONS(7579), 1,
      anon_sym_CARET,
    ACTIONS(7581), 1,
      anon_sym_AMP,
    ACTIONS(7595), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7597), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(7599), 1,
      anon_sym_QMARK,
    ACTIONS(7569), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7583), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7585), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7587), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7589), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7591), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7593), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7571), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [174293] = 10,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7639), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7615), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7629), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7631), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7633), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7635), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7637), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7139), 4,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(7137), 18,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_RPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [174351] = 11,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7639), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7615), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7627), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7629), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7631), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7633), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7635), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7637), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7139), 4,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(7137), 16,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_RPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [174411] = 19,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6999), 1,
      anon_sym_RBRACK,
    ACTIONS(7567), 1,
      anon_sym_EQ,
    ACTIONS(7573), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7575), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7577), 1,
      anon_sym_PIPE,
    ACTIONS(7579), 1,
      anon_sym_CARET,
    ACTIONS(7581), 1,
      anon_sym_AMP,
    ACTIONS(7595), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7597), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(7599), 1,
      anon_sym_QMARK,
    ACTIONS(7569), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7583), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7585), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7587), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7589), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7591), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7593), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7571), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [174487] = 12,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7625), 1,
      anon_sym_AMP,
    ACTIONS(7639), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7615), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7627), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7629), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7631), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7633), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7635), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7139), 3,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
    ACTIONS(7637), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7137), 16,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_RPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [174549] = 13,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7623), 1,
      anon_sym_CARET,
    ACTIONS(7625), 1,
      anon_sym_AMP,
    ACTIONS(7639), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7139), 2,
      anon_sym_EQ,
      anon_sym_PIPE,
    ACTIONS(7615), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7627), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7629), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7631), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7633), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7635), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7637), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7137), 16,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_RPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [174613] = 14,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7139), 1,
      anon_sym_EQ,
    ACTIONS(7621), 1,
      anon_sym_PIPE,
    ACTIONS(7623), 1,
      anon_sym_CARET,
    ACTIONS(7625), 1,
      anon_sym_AMP,
    ACTIONS(7639), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7615), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7627), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7629), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7631), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7633), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7635), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7637), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7137), 16,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_RPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [174679] = 15,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7139), 1,
      anon_sym_EQ,
    ACTIONS(7619), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7621), 1,
      anon_sym_PIPE,
    ACTIONS(7623), 1,
      anon_sym_CARET,
    ACTIONS(7625), 1,
      anon_sym_AMP,
    ACTIONS(7639), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7615), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7627), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7629), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7631), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7633), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7635), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7637), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7137), 15,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_RPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [174747] = 18,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7617), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7619), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7621), 1,
      anon_sym_PIPE,
    ACTIONS(7623), 1,
      anon_sym_CARET,
    ACTIONS(7625), 1,
      anon_sym_AMP,
    ACTIONS(7639), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7679), 1,
      anon_sym_EQ,
    ACTIONS(7681), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(7683), 1,
      anon_sym_QMARK,
    ACTIONS(7615), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7627), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7629), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7631), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7633), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7635), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7637), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7137), 12,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_RPAREN,
  [174821] = 17,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7139), 1,
      anon_sym_EQ,
    ACTIONS(7617), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7619), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7621), 1,
      anon_sym_PIPE,
    ACTIONS(7623), 1,
      anon_sym_CARET,
    ACTIONS(7625), 1,
      anon_sym_AMP,
    ACTIONS(7639), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7683), 1,
      anon_sym_QMARK,
    ACTIONS(7615), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7627), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7629), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7631), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7633), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7635), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7637), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7137), 13,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_RPAREN,
      anon_sym_EQ_TILDE,
  [174893] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1316), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1318), 22,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [174937] = 19,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6997), 1,
      anon_sym_RBRACK,
    ACTIONS(7567), 1,
      anon_sym_EQ,
    ACTIONS(7573), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7575), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7577), 1,
      anon_sym_PIPE,
    ACTIONS(7579), 1,
      anon_sym_CARET,
    ACTIONS(7581), 1,
      anon_sym_AMP,
    ACTIONS(7595), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7597), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(7599), 1,
      anon_sym_QMARK,
    ACTIONS(7569), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7583), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7585), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7587), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7589), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7591), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7593), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7571), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [175013] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1241), 1,
      sym_file_descriptor,
    ACTIONS(7673), 1,
      anon_sym_DQUOTE,
    ACTIONS(7677), 1,
      sym_variable_name,
    STATE(4393), 1,
      sym_string,
    ACTIONS(7675), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(7671), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1239), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
      anon_sym_BQUOTE,
  [175067] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1235), 1,
      sym_file_descriptor,
    ACTIONS(7673), 1,
      anon_sym_DQUOTE,
    ACTIONS(7677), 1,
      sym_variable_name,
    STATE(4393), 1,
      sym_string,
    ACTIONS(7675), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(7671), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1227), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
      anon_sym_BQUOTE,
  [175121] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7114), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(7112), 22,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [175165] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7465), 13,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(7467), 23,
      anon_sym_RPAREN_RPAREN,
      anon_sym_COMMA,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_DASHo,
      anon_sym_AMP_AMP,
      anon_sym_DASHa,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
  [175209] = 19,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6995), 1,
      anon_sym_RBRACK,
    ACTIONS(7567), 1,
      anon_sym_EQ,
    ACTIONS(7573), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7575), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7577), 1,
      anon_sym_PIPE,
    ACTIONS(7579), 1,
      anon_sym_CARET,
    ACTIONS(7581), 1,
      anon_sym_AMP,
    ACTIONS(7595), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7597), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(7599), 1,
      anon_sym_QMARK,
    ACTIONS(7569), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7583), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7585), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7587), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7589), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7591), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7593), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7571), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [175285] = 19,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7070), 1,
      anon_sym_RBRACK,
    ACTIONS(7567), 1,
      anon_sym_EQ,
    ACTIONS(7573), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7575), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7577), 1,
      anon_sym_PIPE,
    ACTIONS(7579), 1,
      anon_sym_CARET,
    ACTIONS(7581), 1,
      anon_sym_AMP,
    ACTIONS(7595), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7597), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(7599), 1,
      anon_sym_QMARK,
    ACTIONS(7569), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7583), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7585), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7587), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7589), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7591), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7593), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7571), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [175361] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1316), 13,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1318), 23,
      anon_sym_COMMA,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_DASHo,
      anon_sym_AMP_AMP,
      anon_sym_DASHa,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RPAREN,
  [175405] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7643), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7469), 13,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(7471), 21,
      anon_sym_RPAREN_RPAREN,
      anon_sym_COMMA,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_DASHo,
      anon_sym_AMP_AMP,
      anon_sym_DASHa,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
  [175451] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7106), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(7104), 22,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [175495] = 8,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7639), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7615), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7633), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7635), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7637), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7096), 6,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7094), 20,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [175549] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1320), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1322), 22,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [175593] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1320), 13,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1322), 23,
      anon_sym_COMMA,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_DASHo,
      anon_sym_AMP_AMP,
      anon_sym_DASHa,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RPAREN,
  [175637] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7615), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7096), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(7094), 20,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [175683] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1328), 13,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1330), 23,
      anon_sym_COMMA,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_DASHo,
      anon_sym_AMP_AMP,
      anon_sym_DASHa,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RPAREN,
  [175727] = 8,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7639), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7615), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7633), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7635), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7637), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7096), 6,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7094), 20,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [175781] = 19,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6993), 1,
      anon_sym_RBRACK,
    ACTIONS(7567), 1,
      anon_sym_EQ,
    ACTIONS(7573), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7575), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7577), 1,
      anon_sym_PIPE,
    ACTIONS(7579), 1,
      anon_sym_CARET,
    ACTIONS(7581), 1,
      anon_sym_AMP,
    ACTIONS(7595), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7597), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(7599), 1,
      anon_sym_QMARK,
    ACTIONS(7569), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7583), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7585), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7587), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7589), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7591), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7593), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7571), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [175857] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7489), 13,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(7491), 23,
      anon_sym_RPAREN_RPAREN,
      anon_sym_COMMA,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_DASHo,
      anon_sym_AMP_AMP,
      anon_sym_DASHa,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
  [175901] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7501), 13,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(7503), 23,
      anon_sym_RPAREN_RPAREN,
      anon_sym_COMMA,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_DASHo,
      anon_sym_AMP_AMP,
      anon_sym_DASHa,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
  [175945] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1316), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1318), 22,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [175989] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1352), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1354), 22,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [176033] = 19,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6991), 1,
      anon_sym_RBRACK,
    ACTIONS(7567), 1,
      anon_sym_EQ,
    ACTIONS(7573), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7575), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7577), 1,
      anon_sym_PIPE,
    ACTIONS(7579), 1,
      anon_sym_CARET,
    ACTIONS(7581), 1,
      anon_sym_AMP,
    ACTIONS(7595), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7597), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(7599), 1,
      anon_sym_QMARK,
    ACTIONS(7569), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7583), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7585), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7587), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7589), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7591), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7593), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7571), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [176109] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7489), 13,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(7491), 23,
      anon_sym_RPAREN_RPAREN,
      anon_sym_COMMA,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_DASHo,
      anon_sym_AMP_AMP,
      anon_sym_DASHa,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
  [176153] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1340), 13,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1342), 23,
      anon_sym_COMMA,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_DASHo,
      anon_sym_AMP_AMP,
      anon_sym_DASHa,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RPAREN,
  [176197] = 16,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7645), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7661), 1,
      anon_sym_AMP,
    ACTIONS(7663), 1,
      anon_sym_CARET,
    ACTIONS(7665), 1,
      anon_sym_PIPE,
    ACTIONS(7511), 2,
      anon_sym_RPAREN_RPAREN,
      anon_sym_COMMA,
    ACTIONS(7643), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7651), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7653), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7655), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7657), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7659), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7667), 2,
      anon_sym_AMP_AMP,
      anon_sym_DASHa,
    ACTIONS(7669), 2,
      anon_sym_PIPE_PIPE,
      anon_sym_DASHo,
    ACTIONS(7649), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7685), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [176267] = 19,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6987), 1,
      anon_sym_RBRACK,
    ACTIONS(7567), 1,
      anon_sym_EQ,
    ACTIONS(7573), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7575), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7577), 1,
      anon_sym_PIPE,
    ACTIONS(7579), 1,
      anon_sym_CARET,
    ACTIONS(7581), 1,
      anon_sym_AMP,
    ACTIONS(7595), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7597), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(7599), 1,
      anon_sym_QMARK,
    ACTIONS(7569), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7583), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7585), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7587), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7589), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7591), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7593), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7571), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [176343] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1320), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1322), 22,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [176387] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1328), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1330), 22,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [176431] = 19,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6985), 1,
      anon_sym_RBRACK,
    ACTIONS(7567), 1,
      anon_sym_EQ,
    ACTIONS(7573), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7575), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7577), 1,
      anon_sym_PIPE,
    ACTIONS(7579), 1,
      anon_sym_CARET,
    ACTIONS(7581), 1,
      anon_sym_AMP,
    ACTIONS(7595), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7597), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(7599), 1,
      anon_sym_QMARK,
    ACTIONS(7569), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7583), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7585), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7587), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7589), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7591), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7593), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7571), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [176507] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1352), 13,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1354), 23,
      anon_sym_RPAREN_RPAREN,
      anon_sym_COMMA,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_DASHo,
      anon_sym_AMP_AMP,
      anon_sym_DASHa,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
  [176551] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1235), 1,
      sym_file_descriptor,
    ACTIONS(3791), 1,
      anon_sym_DQUOTE,
    ACTIONS(7444), 1,
      sym_variable_name,
    STATE(4389), 1,
      sym_string,
    ACTIONS(7442), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(7440), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1227), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
      anon_sym_LT_LT_LT,
  [176605] = 19,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6983), 1,
      anon_sym_RBRACK,
    ACTIONS(7567), 1,
      anon_sym_EQ,
    ACTIONS(7573), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7575), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7577), 1,
      anon_sym_PIPE,
    ACTIONS(7579), 1,
      anon_sym_CARET,
    ACTIONS(7581), 1,
      anon_sym_AMP,
    ACTIONS(7595), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7597), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(7599), 1,
      anon_sym_QMARK,
    ACTIONS(7569), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7583), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7585), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7587), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7589), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7591), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7593), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7571), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [176681] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1352), 13,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1354), 23,
      anon_sym_COMMA,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_DASHo,
      anon_sym_AMP_AMP,
      anon_sym_DASHa,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RPAREN,
  [176725] = 19,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6981), 1,
      anon_sym_RBRACK,
    ACTIONS(7567), 1,
      anon_sym_EQ,
    ACTIONS(7573), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7575), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7577), 1,
      anon_sym_PIPE,
    ACTIONS(7579), 1,
      anon_sym_CARET,
    ACTIONS(7581), 1,
      anon_sym_AMP,
    ACTIONS(7595), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7597), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(7599), 1,
      anon_sym_QMARK,
    ACTIONS(7569), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7583), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7585), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7587), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7589), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7591), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7593), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7571), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [176801] = 19,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6979), 1,
      anon_sym_RBRACK,
    ACTIONS(7567), 1,
      anon_sym_EQ,
    ACTIONS(7573), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7575), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7577), 1,
      anon_sym_PIPE,
    ACTIONS(7579), 1,
      anon_sym_CARET,
    ACTIONS(7581), 1,
      anon_sym_AMP,
    ACTIONS(7595), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7597), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(7599), 1,
      anon_sym_QMARK,
    ACTIONS(7569), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7583), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7585), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7587), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7589), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7591), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7593), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7571), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [176877] = 19,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6977), 1,
      anon_sym_RBRACK,
    ACTIONS(7567), 1,
      anon_sym_EQ,
    ACTIONS(7573), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7575), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7577), 1,
      anon_sym_PIPE,
    ACTIONS(7579), 1,
      anon_sym_CARET,
    ACTIONS(7581), 1,
      anon_sym_AMP,
    ACTIONS(7595), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7597), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(7599), 1,
      anon_sym_QMARK,
    ACTIONS(7569), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7583), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7585), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7587), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7589), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7591), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7593), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7571), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [176953] = 19,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6975), 1,
      anon_sym_RBRACK,
    ACTIONS(7567), 1,
      anon_sym_EQ,
    ACTIONS(7573), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7575), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7577), 1,
      anon_sym_PIPE,
    ACTIONS(7579), 1,
      anon_sym_CARET,
    ACTIONS(7581), 1,
      anon_sym_AMP,
    ACTIONS(7595), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7597), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(7599), 1,
      anon_sym_QMARK,
    ACTIONS(7569), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7583), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7585), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7587), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7589), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7591), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7593), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7571), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [177029] = 8,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7607), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7601), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7603), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7605), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6807), 6,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(6805), 20,
      sym_test_operator,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [177083] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1356), 13,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1358), 23,
      anon_sym_COMMA,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_DASHo,
      anon_sym_AMP_AMP,
      anon_sym_DASHa,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RPAREN,
  [177127] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1344), 13,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1346), 23,
      anon_sym_COMMA,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_DASHo,
      anon_sym_AMP_AMP,
      anon_sym_DASHa,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RPAREN,
  [177171] = 19,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6973), 1,
      anon_sym_RBRACK,
    ACTIONS(7567), 1,
      anon_sym_EQ,
    ACTIONS(7573), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7575), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7577), 1,
      anon_sym_PIPE,
    ACTIONS(7579), 1,
      anon_sym_CARET,
    ACTIONS(7581), 1,
      anon_sym_AMP,
    ACTIONS(7595), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7597), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(7599), 1,
      anon_sym_QMARK,
    ACTIONS(7569), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7583), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7585), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7587), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7589), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7591), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7593), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7571), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [177247] = 19,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6971), 1,
      anon_sym_RBRACK,
    ACTIONS(7567), 1,
      anon_sym_EQ,
    ACTIONS(7573), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7575), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7577), 1,
      anon_sym_PIPE,
    ACTIONS(7579), 1,
      anon_sym_CARET,
    ACTIONS(7581), 1,
      anon_sym_AMP,
    ACTIONS(7595), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7597), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(7599), 1,
      anon_sym_QMARK,
    ACTIONS(7569), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7583), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7585), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7587), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7589), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7591), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7593), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7571), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [177323] = 24,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6474), 1,
      sym_word,
    ACTIONS(6480), 1,
      anon_sym_LPAREN,
    ACTIONS(6488), 1,
      anon_sym_DOLLAR,
    ACTIONS(6494), 1,
      aux_sym_number_token1,
    ACTIONS(6496), 1,
      aux_sym_number_token2,
    ACTIONS(6500), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(6508), 1,
      sym_test_operator,
    ACTIONS(6510), 1,
      sym_extglob_pattern,
    ACTIONS(6512), 1,
      sym__brace_start,
    ACTIONS(7226), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(7228), 1,
      sym__special_character,
    ACTIONS(7230), 1,
      anon_sym_DQUOTE,
    ACTIONS(7234), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(7236), 1,
      anon_sym_BQUOTE,
    ACTIONS(7238), 1,
      anon_sym_DOLLAR_BQUOTE,
    STATE(6426), 1,
      aux_sym__literal_repeat1,
    STATE(7291), 1,
      sym_last_case_item,
    ACTIONS(7224), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(7232), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(7240), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3477), 2,
      sym_case_item,
      aux_sym_case_statement_repeat1,
    STATE(6695), 2,
      sym_concatenation,
      sym__extglob_blob,
    STATE(6303), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [177409] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1294), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1292), 31,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [177453] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1336), 13,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1338), 23,
      anon_sym_COMMA,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_DASHo,
      anon_sym_AMP_AMP,
      anon_sym_DASHa,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RPAREN,
  [177497] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1241), 1,
      sym_file_descriptor,
    ACTIONS(3791), 1,
      anon_sym_DQUOTE,
    ACTIONS(7444), 1,
      sym_variable_name,
    STATE(4389), 1,
      sym_string,
    ACTIONS(7442), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(7440), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1239), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
      anon_sym_LT_LT_LT,
  [177551] = 18,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6807), 1,
      anon_sym_EQ,
    ACTIONS(7607), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7647), 1,
      sym_test_operator,
    ACTIONS(7687), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7689), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7691), 1,
      anon_sym_PIPE,
    ACTIONS(7693), 1,
      anon_sym_CARET,
    ACTIONS(7695), 1,
      anon_sym_AMP,
    ACTIONS(7703), 1,
      anon_sym_QMARK,
    ACTIONS(6793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7601), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7603), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7697), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7699), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7701), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7605), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6805), 12,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_EQ_TILDE,
  [177625] = 19,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6967), 1,
      anon_sym_RBRACK,
    ACTIONS(7567), 1,
      anon_sym_EQ,
    ACTIONS(7573), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7575), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7577), 1,
      anon_sym_PIPE,
    ACTIONS(7579), 1,
      anon_sym_CARET,
    ACTIONS(7581), 1,
      anon_sym_AMP,
    ACTIONS(7595), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7597), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(7599), 1,
      anon_sym_QMARK,
    ACTIONS(7569), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7583), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7585), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7587), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7589), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7591), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7593), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7571), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [177701] = 19,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6965), 1,
      anon_sym_RBRACK,
    ACTIONS(7567), 1,
      anon_sym_EQ,
    ACTIONS(7573), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7575), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7577), 1,
      anon_sym_PIPE,
    ACTIONS(7579), 1,
      anon_sym_CARET,
    ACTIONS(7581), 1,
      anon_sym_AMP,
    ACTIONS(7595), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7597), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(7599), 1,
      anon_sym_QMARK,
    ACTIONS(7569), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7583), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7585), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7587), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7589), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7591), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7593), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7571), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [177777] = 19,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6963), 1,
      anon_sym_RBRACK,
    ACTIONS(7567), 1,
      anon_sym_EQ,
    ACTIONS(7573), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7575), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7577), 1,
      anon_sym_PIPE,
    ACTIONS(7579), 1,
      anon_sym_CARET,
    ACTIONS(7581), 1,
      anon_sym_AMP,
    ACTIONS(7595), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7597), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(7599), 1,
      anon_sym_QMARK,
    ACTIONS(7569), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7583), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7585), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7587), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7589), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7591), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7593), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7571), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [177853] = 19,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7607), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7647), 1,
      sym_test_operator,
    ACTIONS(7687), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7689), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7691), 1,
      anon_sym_PIPE,
    ACTIONS(7693), 1,
      anon_sym_CARET,
    ACTIONS(7695), 1,
      anon_sym_AMP,
    ACTIONS(7703), 1,
      anon_sym_QMARK,
    ACTIONS(7705), 1,
      anon_sym_EQ,
    ACTIONS(7707), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(6793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7601), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7603), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7697), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7699), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7701), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7605), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6805), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [177929] = 16,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6807), 1,
      anon_sym_EQ,
    ACTIONS(7607), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7647), 1,
      sym_test_operator,
    ACTIONS(7689), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7691), 1,
      anon_sym_PIPE,
    ACTIONS(7693), 1,
      anon_sym_CARET,
    ACTIONS(7695), 1,
      anon_sym_AMP,
    ACTIONS(6793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7601), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7603), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7697), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7699), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7701), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7605), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6805), 14,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [177999] = 19,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6961), 1,
      anon_sym_RBRACK,
    ACTIONS(7567), 1,
      anon_sym_EQ,
    ACTIONS(7573), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7575), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7577), 1,
      anon_sym_PIPE,
    ACTIONS(7579), 1,
      anon_sym_CARET,
    ACTIONS(7581), 1,
      anon_sym_AMP,
    ACTIONS(7595), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7597), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(7599), 1,
      anon_sym_QMARK,
    ACTIONS(7569), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7583), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7585), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7587), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7589), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7591), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7593), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7571), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [178075] = 19,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6959), 1,
      anon_sym_RBRACK,
    ACTIONS(7567), 1,
      anon_sym_EQ,
    ACTIONS(7573), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7575), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7577), 1,
      anon_sym_PIPE,
    ACTIONS(7579), 1,
      anon_sym_CARET,
    ACTIONS(7581), 1,
      anon_sym_AMP,
    ACTIONS(7595), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7597), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(7599), 1,
      anon_sym_QMARK,
    ACTIONS(7569), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7583), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7585), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7587), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7589), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7591), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7593), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7571), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [178151] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1312), 13,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1314), 23,
      anon_sym_COMMA,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_DASHo,
      anon_sym_AMP_AMP,
      anon_sym_DASHa,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RPAREN,
  [178195] = 15,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6807), 1,
      anon_sym_EQ,
    ACTIONS(7607), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7647), 1,
      sym_test_operator,
    ACTIONS(7691), 1,
      anon_sym_PIPE,
    ACTIONS(7693), 1,
      anon_sym_CARET,
    ACTIONS(7695), 1,
      anon_sym_AMP,
    ACTIONS(6793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7601), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7603), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7697), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7699), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7701), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7605), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6805), 15,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [178263] = 19,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6957), 1,
      anon_sym_RBRACK,
    ACTIONS(7567), 1,
      anon_sym_EQ,
    ACTIONS(7573), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7575), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7577), 1,
      anon_sym_PIPE,
    ACTIONS(7579), 1,
      anon_sym_CARET,
    ACTIONS(7581), 1,
      anon_sym_AMP,
    ACTIONS(7595), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7597), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(7599), 1,
      anon_sym_QMARK,
    ACTIONS(7569), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7583), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7585), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7587), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7589), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7591), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7593), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7571), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [178339] = 19,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7086), 1,
      anon_sym_RBRACK,
    ACTIONS(7567), 1,
      anon_sym_EQ,
    ACTIONS(7573), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7575), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7577), 1,
      anon_sym_PIPE,
    ACTIONS(7579), 1,
      anon_sym_CARET,
    ACTIONS(7581), 1,
      anon_sym_AMP,
    ACTIONS(7595), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7597), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(7599), 1,
      anon_sym_QMARK,
    ACTIONS(7569), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7583), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7585), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7587), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7589), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7591), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7593), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7571), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [178415] = 19,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6955), 1,
      anon_sym_RBRACK,
    ACTIONS(7567), 1,
      anon_sym_EQ,
    ACTIONS(7573), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7575), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7577), 1,
      anon_sym_PIPE,
    ACTIONS(7579), 1,
      anon_sym_CARET,
    ACTIONS(7581), 1,
      anon_sym_AMP,
    ACTIONS(7595), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7597), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(7599), 1,
      anon_sym_QMARK,
    ACTIONS(7569), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7583), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7585), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7587), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7589), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7591), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7593), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7571), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [178491] = 14,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7607), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7647), 1,
      sym_test_operator,
    ACTIONS(7693), 1,
      anon_sym_CARET,
    ACTIONS(7695), 1,
      anon_sym_AMP,
    ACTIONS(6793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6807), 2,
      anon_sym_EQ,
      anon_sym_PIPE,
    ACTIONS(7601), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7603), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7697), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7699), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7701), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7605), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6805), 15,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [178557] = 19,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6953), 1,
      anon_sym_RBRACK,
    ACTIONS(7567), 1,
      anon_sym_EQ,
    ACTIONS(7573), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7575), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7577), 1,
      anon_sym_PIPE,
    ACTIONS(7579), 1,
      anon_sym_CARET,
    ACTIONS(7581), 1,
      anon_sym_AMP,
    ACTIONS(7595), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7597), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(7599), 1,
      anon_sym_QMARK,
    ACTIONS(7569), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7583), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7585), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7587), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7589), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7591), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7593), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7571), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [178633] = 19,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6951), 1,
      anon_sym_RBRACK,
    ACTIONS(7567), 1,
      anon_sym_EQ,
    ACTIONS(7573), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7575), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7577), 1,
      anon_sym_PIPE,
    ACTIONS(7579), 1,
      anon_sym_CARET,
    ACTIONS(7581), 1,
      anon_sym_AMP,
    ACTIONS(7595), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7597), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(7599), 1,
      anon_sym_QMARK,
    ACTIONS(7569), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7583), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7585), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7587), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7589), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7591), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7593), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7571), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [178709] = 13,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7607), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7647), 1,
      sym_test_operator,
    ACTIONS(7695), 1,
      anon_sym_AMP,
    ACTIONS(6793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7601), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7603), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7697), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7699), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7701), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(6807), 3,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
    ACTIONS(7605), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6805), 15,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [178773] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1334), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1332), 31,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [178817] = 12,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7607), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7647), 1,
      sym_test_operator,
    ACTIONS(6793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7601), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7603), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7697), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7699), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7701), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7605), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6807), 4,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(6805), 15,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [178879] = 11,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7607), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7647), 1,
      sym_test_operator,
    ACTIONS(6793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7601), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7603), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7699), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7701), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7605), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6807), 4,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(6805), 17,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [178939] = 19,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6949), 1,
      anon_sym_RBRACK,
    ACTIONS(7567), 1,
      anon_sym_EQ,
    ACTIONS(7573), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7575), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7577), 1,
      anon_sym_PIPE,
    ACTIONS(7579), 1,
      anon_sym_CARET,
    ACTIONS(7581), 1,
      anon_sym_AMP,
    ACTIONS(7595), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7597), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(7599), 1,
      anon_sym_QMARK,
    ACTIONS(7569), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7583), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7585), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7587), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7589), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7591), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7593), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7571), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [179015] = 9,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7607), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7647), 1,
      sym_test_operator,
    ACTIONS(6793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7601), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7603), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7605), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6807), 6,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(6805), 19,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [179071] = 19,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6947), 1,
      anon_sym_RBRACK,
    ACTIONS(7567), 1,
      anon_sym_EQ,
    ACTIONS(7573), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7575), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7577), 1,
      anon_sym_PIPE,
    ACTIONS(7579), 1,
      anon_sym_CARET,
    ACTIONS(7581), 1,
      anon_sym_AMP,
    ACTIONS(7595), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7597), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(7599), 1,
      anon_sym_QMARK,
    ACTIONS(7569), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7583), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7585), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7587), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7589), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7591), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7593), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7571), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [179147] = 19,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6945), 1,
      anon_sym_RBRACK,
    ACTIONS(7567), 1,
      anon_sym_EQ,
    ACTIONS(7573), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7575), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7577), 1,
      anon_sym_PIPE,
    ACTIONS(7579), 1,
      anon_sym_CARET,
    ACTIONS(7581), 1,
      anon_sym_AMP,
    ACTIONS(7595), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7597), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(7599), 1,
      anon_sym_QMARK,
    ACTIONS(7569), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7583), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7585), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7587), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7589), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7591), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7593), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7571), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [179223] = 19,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6943), 1,
      anon_sym_RBRACK,
    ACTIONS(7567), 1,
      anon_sym_EQ,
    ACTIONS(7573), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7575), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7577), 1,
      anon_sym_PIPE,
    ACTIONS(7579), 1,
      anon_sym_CARET,
    ACTIONS(7581), 1,
      anon_sym_AMP,
    ACTIONS(7595), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7597), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(7599), 1,
      anon_sym_QMARK,
    ACTIONS(7569), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7583), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7585), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7587), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7589), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7591), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7593), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7571), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [179299] = 19,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6941), 1,
      anon_sym_RBRACK,
    ACTIONS(7567), 1,
      anon_sym_EQ,
    ACTIONS(7573), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7575), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7577), 1,
      anon_sym_PIPE,
    ACTIONS(7579), 1,
      anon_sym_CARET,
    ACTIONS(7581), 1,
      anon_sym_AMP,
    ACTIONS(7595), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7597), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(7599), 1,
      anon_sym_QMARK,
    ACTIONS(7569), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7583), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7585), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7587), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7589), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7591), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7593), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7571), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [179375] = 19,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6939), 1,
      anon_sym_RBRACK,
    ACTIONS(7567), 1,
      anon_sym_EQ,
    ACTIONS(7573), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7575), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7577), 1,
      anon_sym_PIPE,
    ACTIONS(7579), 1,
      anon_sym_CARET,
    ACTIONS(7581), 1,
      anon_sym_AMP,
    ACTIONS(7595), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7597), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(7599), 1,
      anon_sym_QMARK,
    ACTIONS(7569), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7583), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7585), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7587), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7589), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7591), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7593), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7571), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [179451] = 19,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6937), 1,
      anon_sym_RBRACK,
    ACTIONS(7567), 1,
      anon_sym_EQ,
    ACTIONS(7573), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7575), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7577), 1,
      anon_sym_PIPE,
    ACTIONS(7579), 1,
      anon_sym_CARET,
    ACTIONS(7581), 1,
      anon_sym_AMP,
    ACTIONS(7595), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7597), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(7599), 1,
      anon_sym_QMARK,
    ACTIONS(7569), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7583), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7585), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7587), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7589), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7591), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7593), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7571), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [179527] = 19,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6935), 1,
      anon_sym_RBRACK,
    ACTIONS(7567), 1,
      anon_sym_EQ,
    ACTIONS(7573), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7575), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7577), 1,
      anon_sym_PIPE,
    ACTIONS(7579), 1,
      anon_sym_CARET,
    ACTIONS(7581), 1,
      anon_sym_AMP,
    ACTIONS(7595), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7597), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(7599), 1,
      anon_sym_QMARK,
    ACTIONS(7569), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7583), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7585), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7587), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7589), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7591), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7593), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7571), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [179603] = 19,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6933), 1,
      anon_sym_RBRACK,
    ACTIONS(7567), 1,
      anon_sym_EQ,
    ACTIONS(7573), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7575), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7577), 1,
      anon_sym_PIPE,
    ACTIONS(7579), 1,
      anon_sym_CARET,
    ACTIONS(7581), 1,
      anon_sym_AMP,
    ACTIONS(7595), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7597), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(7599), 1,
      anon_sym_QMARK,
    ACTIONS(7569), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7583), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7585), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7587), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7589), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7591), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7593), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7571), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [179679] = 19,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6931), 1,
      anon_sym_RBRACK,
    ACTIONS(7567), 1,
      anon_sym_EQ,
    ACTIONS(7573), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7575), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7577), 1,
      anon_sym_PIPE,
    ACTIONS(7579), 1,
      anon_sym_CARET,
    ACTIONS(7581), 1,
      anon_sym_AMP,
    ACTIONS(7595), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7597), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(7599), 1,
      anon_sym_QMARK,
    ACTIONS(7569), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7583), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7585), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7587), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7589), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7591), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7593), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7571), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [179755] = 7,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7607), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7603), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7605), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6807), 8,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(6805), 20,
      sym_test_operator,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [179807] = 19,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6929), 1,
      anon_sym_RBRACK,
    ACTIONS(7567), 1,
      anon_sym_EQ,
    ACTIONS(7573), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7575), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7577), 1,
      anon_sym_PIPE,
    ACTIONS(7579), 1,
      anon_sym_CARET,
    ACTIONS(7581), 1,
      anon_sym_AMP,
    ACTIONS(7595), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7597), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(7599), 1,
      anon_sym_QMARK,
    ACTIONS(7569), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7583), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7585), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7587), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7589), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7591), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7593), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7571), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [179883] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7607), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7605), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6807), 10,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(6805), 20,
      sym_test_operator,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [179933] = 19,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6927), 1,
      anon_sym_RBRACK,
    ACTIONS(7567), 1,
      anon_sym_EQ,
    ACTIONS(7573), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7575), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7577), 1,
      anon_sym_PIPE,
    ACTIONS(7579), 1,
      anon_sym_CARET,
    ACTIONS(7581), 1,
      anon_sym_AMP,
    ACTIONS(7595), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7597), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(7599), 1,
      anon_sym_QMARK,
    ACTIONS(7569), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7583), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7585), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7587), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7589), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7591), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7593), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7571), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [180009] = 19,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6925), 1,
      anon_sym_RBRACK,
    ACTIONS(7567), 1,
      anon_sym_EQ,
    ACTIONS(7573), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7575), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7577), 1,
      anon_sym_PIPE,
    ACTIONS(7579), 1,
      anon_sym_CARET,
    ACTIONS(7581), 1,
      anon_sym_AMP,
    ACTIONS(7595), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7597), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(7599), 1,
      anon_sym_QMARK,
    ACTIONS(7569), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7583), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7585), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7587), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7589), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7591), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7593), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7571), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [180085] = 19,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6883), 1,
      anon_sym_EQ,
    ACTIONS(6889), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(6891), 1,
      anon_sym_AMP_AMP,
    ACTIONS(6893), 1,
      anon_sym_PIPE,
    ACTIONS(6895), 1,
      anon_sym_CARET,
    ACTIONS(6897), 1,
      anon_sym_AMP,
    ACTIONS(6911), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6913), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(6915), 1,
      anon_sym_QMARK,
    ACTIONS(7709), 1,
      anon_sym_COLON,
    ACTIONS(6885), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6899), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(6901), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(6903), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(6905), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(6907), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(6909), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6887), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [180161] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7607), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6807), 13,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6805), 20,
      sym_test_operator,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [180209] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1308), 13,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1310), 23,
      anon_sym_COMMA,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_DASHo,
      anon_sym_AMP_AMP,
      anon_sym_DASHa,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RPAREN,
  [180253] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1304), 13,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1306), 23,
      anon_sym_COMMA,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_DASHo,
      anon_sym_AMP_AMP,
      anon_sym_DASHa,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RPAREN,
  [180297] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1308), 13,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1310), 23,
      anon_sym_COMMA,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_DASHo,
      anon_sym_AMP_AMP,
      anon_sym_DASHa,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RPAREN,
  [180341] = 19,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6923), 1,
      anon_sym_RBRACK,
    ACTIONS(7567), 1,
      anon_sym_EQ,
    ACTIONS(7573), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7575), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7577), 1,
      anon_sym_PIPE,
    ACTIONS(7579), 1,
      anon_sym_CARET,
    ACTIONS(7581), 1,
      anon_sym_AMP,
    ACTIONS(7595), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7597), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(7599), 1,
      anon_sym_QMARK,
    ACTIONS(7569), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7583), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7585), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7587), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7589), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7591), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7593), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7571), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [180417] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7607), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6807), 13,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6805), 20,
      sym_test_operator,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [180465] = 19,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6921), 1,
      anon_sym_RBRACK,
    ACTIONS(7567), 1,
      anon_sym_EQ,
    ACTIONS(7573), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7575), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7577), 1,
      anon_sym_PIPE,
    ACTIONS(7579), 1,
      anon_sym_CARET,
    ACTIONS(7581), 1,
      anon_sym_AMP,
    ACTIONS(7595), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7597), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(7599), 1,
      anon_sym_QMARK,
    ACTIONS(7569), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7583), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7585), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7587), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7589), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7591), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7593), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7571), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [180541] = 17,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7123), 1,
      anon_sym_EQ,
    ACTIONS(7607), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7647), 1,
      sym_test_operator,
    ACTIONS(7687), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7689), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7691), 1,
      anon_sym_PIPE,
    ACTIONS(7693), 1,
      anon_sym_CARET,
    ACTIONS(7695), 1,
      anon_sym_AMP,
    ACTIONS(6793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7601), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7603), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7697), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7699), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7701), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7605), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7125), 13,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [180613] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7064), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(7062), 22,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [180657] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1302), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1300), 31,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [180701] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7064), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(7062), 22,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [180745] = 19,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6919), 1,
      anon_sym_RBRACK,
    ACTIONS(7567), 1,
      anon_sym_EQ,
    ACTIONS(7573), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7575), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7577), 1,
      anon_sym_PIPE,
    ACTIONS(7579), 1,
      anon_sym_CARET,
    ACTIONS(7581), 1,
      anon_sym_AMP,
    ACTIONS(7595), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7597), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(7599), 1,
      anon_sym_QMARK,
    ACTIONS(7569), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7583), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7585), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7587), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7589), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7591), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7593), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7571), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [180821] = 19,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6917), 1,
      anon_sym_RBRACK,
    ACTIONS(7567), 1,
      anon_sym_EQ,
    ACTIONS(7573), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7575), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7577), 1,
      anon_sym_PIPE,
    ACTIONS(7579), 1,
      anon_sym_CARET,
    ACTIONS(7581), 1,
      anon_sym_AMP,
    ACTIONS(7595), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7597), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(7599), 1,
      anon_sym_QMARK,
    ACTIONS(7569), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7583), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7585), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7587), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7589), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7591), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7593), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7571), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [180897] = 19,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6879), 1,
      anon_sym_RBRACK,
    ACTIONS(7567), 1,
      anon_sym_EQ,
    ACTIONS(7573), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7575), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7577), 1,
      anon_sym_PIPE,
    ACTIONS(7579), 1,
      anon_sym_CARET,
    ACTIONS(7581), 1,
      anon_sym_AMP,
    ACTIONS(7595), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7597), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(7599), 1,
      anon_sym_QMARK,
    ACTIONS(7569), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7583), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7585), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7587), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7589), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7591), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7593), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7571), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [180973] = 19,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7617), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7619), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7621), 1,
      anon_sym_PIPE,
    ACTIONS(7623), 1,
      anon_sym_CARET,
    ACTIONS(7625), 1,
      anon_sym_AMP,
    ACTIONS(7639), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7679), 1,
      anon_sym_EQ,
    ACTIONS(7681), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(7683), 1,
      anon_sym_QMARK,
    ACTIONS(7713), 1,
      anon_sym_RPAREN,
    ACTIONS(7615), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7627), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7629), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7631), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7633), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7635), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7637), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7711), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [181049] = 19,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7168), 1,
      anon_sym_RBRACK,
    ACTIONS(7567), 1,
      anon_sym_EQ,
    ACTIONS(7573), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7575), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7577), 1,
      anon_sym_PIPE,
    ACTIONS(7579), 1,
      anon_sym_CARET,
    ACTIONS(7581), 1,
      anon_sym_AMP,
    ACTIONS(7595), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7597), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(7599), 1,
      anon_sym_QMARK,
    ACTIONS(7569), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7583), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7585), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7587), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7589), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7591), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7593), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7571), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [181125] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7053), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(7051), 22,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [181169] = 19,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7617), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7619), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7621), 1,
      anon_sym_PIPE,
    ACTIONS(7623), 1,
      anon_sym_CARET,
    ACTIONS(7625), 1,
      anon_sym_AMP,
    ACTIONS(7639), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7679), 1,
      anon_sym_EQ,
    ACTIONS(7681), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(7683), 1,
      anon_sym_QMARK,
    ACTIONS(7715), 1,
      anon_sym_RPAREN,
    ACTIONS(7615), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7627), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7629), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7631), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7633), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7635), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7637), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7711), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [181245] = 19,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7019), 1,
      anon_sym_RBRACK,
    ACTIONS(7567), 1,
      anon_sym_EQ,
    ACTIONS(7573), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7575), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7577), 1,
      anon_sym_PIPE,
    ACTIONS(7579), 1,
      anon_sym_CARET,
    ACTIONS(7581), 1,
      anon_sym_AMP,
    ACTIONS(7595), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7597), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(7599), 1,
      anon_sym_QMARK,
    ACTIONS(7569), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7583), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7585), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7587), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7589), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7591), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7593), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7571), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [181321] = 19,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7021), 1,
      anon_sym_RBRACK,
    ACTIONS(7567), 1,
      anon_sym_EQ,
    ACTIONS(7573), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7575), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7577), 1,
      anon_sym_PIPE,
    ACTIONS(7579), 1,
      anon_sym_CARET,
    ACTIONS(7581), 1,
      anon_sym_AMP,
    ACTIONS(7595), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7597), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(7599), 1,
      anon_sym_QMARK,
    ACTIONS(7569), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7583), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7585), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7587), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7589), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7591), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7593), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7571), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [181397] = 19,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7023), 1,
      anon_sym_RBRACK,
    ACTIONS(7567), 1,
      anon_sym_EQ,
    ACTIONS(7573), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7575), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7577), 1,
      anon_sym_PIPE,
    ACTIONS(7579), 1,
      anon_sym_CARET,
    ACTIONS(7581), 1,
      anon_sym_AMP,
    ACTIONS(7595), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7597), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(7599), 1,
      anon_sym_QMARK,
    ACTIONS(7569), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7583), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7585), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7587), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7589), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7591), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7593), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7571), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [181473] = 19,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7027), 1,
      anon_sym_RBRACK,
    ACTIONS(7567), 1,
      anon_sym_EQ,
    ACTIONS(7573), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7575), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7577), 1,
      anon_sym_PIPE,
    ACTIONS(7579), 1,
      anon_sym_CARET,
    ACTIONS(7581), 1,
      anon_sym_AMP,
    ACTIONS(7595), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7597), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(7599), 1,
      anon_sym_QMARK,
    ACTIONS(7569), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7583), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7585), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7587), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7589), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7591), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7593), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7571), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [181549] = 19,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7110), 1,
      anon_sym_RBRACK,
    ACTIONS(7567), 1,
      anon_sym_EQ,
    ACTIONS(7573), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7575), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7577), 1,
      anon_sym_PIPE,
    ACTIONS(7579), 1,
      anon_sym_CARET,
    ACTIONS(7581), 1,
      anon_sym_AMP,
    ACTIONS(7595), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7597), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(7599), 1,
      anon_sym_QMARK,
    ACTIONS(7569), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7583), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7585), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7587), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7589), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7591), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7593), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7571), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [181625] = 19,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7098), 1,
      anon_sym_RBRACK,
    ACTIONS(7567), 1,
      anon_sym_EQ,
    ACTIONS(7573), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7575), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7577), 1,
      anon_sym_PIPE,
    ACTIONS(7579), 1,
      anon_sym_CARET,
    ACTIONS(7581), 1,
      anon_sym_AMP,
    ACTIONS(7595), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7597), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(7599), 1,
      anon_sym_QMARK,
    ACTIONS(7569), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7583), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7585), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7587), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7589), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7591), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7593), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7571), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [181701] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7035), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(7033), 22,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [181745] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1298), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1296), 31,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [181789] = 24,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6474), 1,
      sym_word,
    ACTIONS(6480), 1,
      anon_sym_LPAREN,
    ACTIONS(6488), 1,
      anon_sym_DOLLAR,
    ACTIONS(6494), 1,
      aux_sym_number_token1,
    ACTIONS(6496), 1,
      aux_sym_number_token2,
    ACTIONS(6500), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(6508), 1,
      sym_test_operator,
    ACTIONS(6510), 1,
      sym_extglob_pattern,
    ACTIONS(6512), 1,
      sym__brace_start,
    ACTIONS(7226), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(7228), 1,
      sym__special_character,
    ACTIONS(7230), 1,
      anon_sym_DQUOTE,
    ACTIONS(7234), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(7236), 1,
      anon_sym_BQUOTE,
    ACTIONS(7238), 1,
      anon_sym_DOLLAR_BQUOTE,
    STATE(6426), 1,
      aux_sym__literal_repeat1,
    STATE(6923), 1,
      sym_last_case_item,
    ACTIONS(7224), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(7232), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(7240), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3477), 2,
      sym_case_item,
      aux_sym_case_statement_repeat1,
    STATE(6695), 2,
      sym_concatenation,
      sym__extglob_blob,
    STATE(6303), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [181875] = 24,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6474), 1,
      sym_word,
    ACTIONS(6480), 1,
      anon_sym_LPAREN,
    ACTIONS(6488), 1,
      anon_sym_DOLLAR,
    ACTIONS(6494), 1,
      aux_sym_number_token1,
    ACTIONS(6496), 1,
      aux_sym_number_token2,
    ACTIONS(6500), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(6508), 1,
      sym_test_operator,
    ACTIONS(6510), 1,
      sym_extglob_pattern,
    ACTIONS(6512), 1,
      sym__brace_start,
    ACTIONS(7226), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(7228), 1,
      sym__special_character,
    ACTIONS(7230), 1,
      anon_sym_DQUOTE,
    ACTIONS(7234), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(7236), 1,
      anon_sym_BQUOTE,
    ACTIONS(7238), 1,
      anon_sym_DOLLAR_BQUOTE,
    STATE(6426), 1,
      aux_sym__literal_repeat1,
    STATE(6942), 1,
      sym_last_case_item,
    ACTIONS(7224), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(7232), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(7240), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3477), 2,
      sym_case_item,
      aux_sym_case_statement_repeat1,
    STATE(6695), 2,
      sym_concatenation,
      sym__extglob_blob,
    STATE(6303), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [181961] = 16,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7102), 1,
      anon_sym_EQ,
    ACTIONS(7573), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7575), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7577), 1,
      anon_sym_PIPE,
    ACTIONS(7579), 1,
      anon_sym_CARET,
    ACTIONS(7581), 1,
      anon_sym_AMP,
    ACTIONS(7595), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7569), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7583), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7585), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7587), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7589), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7591), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7593), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7100), 14,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [182031] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7448), 13,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(7450), 23,
      anon_sym_COMMA,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_DASHo,
      anon_sym_AMP_AMP,
      anon_sym_DASHa,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RPAREN,
  [182075] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7719), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7717), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7248), 12,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7274), 21,
      anon_sym_COMMA,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_DASHo,
      anon_sym_AMP_AMP,
      anon_sym_DASHa,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RPAREN,
  [182123] = 24,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6474), 1,
      sym_word,
    ACTIONS(6480), 1,
      anon_sym_LPAREN,
    ACTIONS(6488), 1,
      anon_sym_DOLLAR,
    ACTIONS(6494), 1,
      aux_sym_number_token1,
    ACTIONS(6496), 1,
      aux_sym_number_token2,
    ACTIONS(6500), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(6508), 1,
      sym_test_operator,
    ACTIONS(6510), 1,
      sym_extglob_pattern,
    ACTIONS(6512), 1,
      sym__brace_start,
    ACTIONS(7226), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(7228), 1,
      sym__special_character,
    ACTIONS(7230), 1,
      anon_sym_DQUOTE,
    ACTIONS(7234), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(7236), 1,
      anon_sym_BQUOTE,
    ACTIONS(7238), 1,
      anon_sym_DOLLAR_BQUOTE,
    STATE(6426), 1,
      aux_sym__literal_repeat1,
    STATE(7005), 1,
      sym_last_case_item,
    ACTIONS(7224), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(7232), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(7240), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3477), 2,
      sym_case_item,
      aux_sym_case_statement_repeat1,
    STATE(6695), 2,
      sym_concatenation,
      sym__extglob_blob,
    STATE(6303), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [182209] = 24,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6474), 1,
      sym_word,
    ACTIONS(6480), 1,
      anon_sym_LPAREN,
    ACTIONS(6488), 1,
      anon_sym_DOLLAR,
    ACTIONS(6494), 1,
      aux_sym_number_token1,
    ACTIONS(6496), 1,
      aux_sym_number_token2,
    ACTIONS(6500), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(6508), 1,
      sym_test_operator,
    ACTIONS(6510), 1,
      sym_extglob_pattern,
    ACTIONS(6512), 1,
      sym__brace_start,
    ACTIONS(7226), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(7228), 1,
      sym__special_character,
    ACTIONS(7230), 1,
      anon_sym_DQUOTE,
    ACTIONS(7234), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(7236), 1,
      anon_sym_BQUOTE,
    ACTIONS(7238), 1,
      anon_sym_DOLLAR_BQUOTE,
    STATE(6426), 1,
      aux_sym__literal_repeat1,
    STATE(7007), 1,
      sym_last_case_item,
    ACTIONS(7224), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(7232), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(7240), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3477), 2,
      sym_case_item,
      aux_sym_case_statement_repeat1,
    STATE(6695), 2,
      sym_concatenation,
      sym__extglob_blob,
    STATE(6303), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [182295] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7719), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7717), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7248), 12,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7274), 21,
      anon_sym_COMMA,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_DASHo,
      anon_sym_AMP_AMP,
      anon_sym_DASHa,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RPAREN,
  [182343] = 24,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6474), 1,
      sym_word,
    ACTIONS(6480), 1,
      anon_sym_LPAREN,
    ACTIONS(6488), 1,
      anon_sym_DOLLAR,
    ACTIONS(6494), 1,
      aux_sym_number_token1,
    ACTIONS(6496), 1,
      aux_sym_number_token2,
    ACTIONS(6500), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(6508), 1,
      sym_test_operator,
    ACTIONS(6510), 1,
      sym_extglob_pattern,
    ACTIONS(6512), 1,
      sym__brace_start,
    ACTIONS(7226), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(7228), 1,
      sym__special_character,
    ACTIONS(7230), 1,
      anon_sym_DQUOTE,
    ACTIONS(7234), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(7236), 1,
      anon_sym_BQUOTE,
    ACTIONS(7238), 1,
      anon_sym_DOLLAR_BQUOTE,
    STATE(6426), 1,
      aux_sym__literal_repeat1,
    STATE(7013), 1,
      sym_last_case_item,
    ACTIONS(7224), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(7232), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(7240), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3477), 2,
      sym_case_item,
      aux_sym_case_statement_repeat1,
    STATE(6695), 2,
      sym_concatenation,
      sym__extglob_blob,
    STATE(6303), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [182429] = 24,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6474), 1,
      sym_word,
    ACTIONS(6480), 1,
      anon_sym_LPAREN,
    ACTIONS(6488), 1,
      anon_sym_DOLLAR,
    ACTIONS(6494), 1,
      aux_sym_number_token1,
    ACTIONS(6496), 1,
      aux_sym_number_token2,
    ACTIONS(6500), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(6508), 1,
      sym_test_operator,
    ACTIONS(6510), 1,
      sym_extglob_pattern,
    ACTIONS(6512), 1,
      sym__brace_start,
    ACTIONS(7226), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(7228), 1,
      sym__special_character,
    ACTIONS(7230), 1,
      anon_sym_DQUOTE,
    ACTIONS(7234), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(7236), 1,
      anon_sym_BQUOTE,
    ACTIONS(7238), 1,
      anon_sym_DOLLAR_BQUOTE,
    STATE(6426), 1,
      aux_sym__literal_repeat1,
    STATE(7016), 1,
      sym_last_case_item,
    ACTIONS(7224), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(7232), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(7240), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3477), 2,
      sym_case_item,
      aux_sym_case_statement_repeat1,
    STATE(6695), 2,
      sym_concatenation,
      sym__extglob_blob,
    STATE(6303), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [182515] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7719), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7717), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7721), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7248), 9,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7274), 21,
      anon_sym_COMMA,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_DASHo,
      anon_sym_AMP_AMP,
      anon_sym_DASHa,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RPAREN,
  [182565] = 24,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6474), 1,
      sym_word,
    ACTIONS(6480), 1,
      anon_sym_LPAREN,
    ACTIONS(6488), 1,
      anon_sym_DOLLAR,
    ACTIONS(6494), 1,
      aux_sym_number_token1,
    ACTIONS(6496), 1,
      aux_sym_number_token2,
    ACTIONS(6500), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(6508), 1,
      sym_test_operator,
    ACTIONS(6510), 1,
      sym_extglob_pattern,
    ACTIONS(6512), 1,
      sym__brace_start,
    ACTIONS(7226), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(7228), 1,
      sym__special_character,
    ACTIONS(7230), 1,
      anon_sym_DQUOTE,
    ACTIONS(7234), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(7236), 1,
      anon_sym_BQUOTE,
    ACTIONS(7238), 1,
      anon_sym_DOLLAR_BQUOTE,
    STATE(6426), 1,
      aux_sym__literal_repeat1,
    STATE(7020), 1,
      sym_last_case_item,
    ACTIONS(7224), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(7232), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(7240), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3477), 2,
      sym_case_item,
      aux_sym_case_statement_repeat1,
    STATE(6695), 2,
      sym_concatenation,
      sym__extglob_blob,
    STATE(6303), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [182651] = 24,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6474), 1,
      sym_word,
    ACTIONS(6480), 1,
      anon_sym_LPAREN,
    ACTIONS(6488), 1,
      anon_sym_DOLLAR,
    ACTIONS(6494), 1,
      aux_sym_number_token1,
    ACTIONS(6496), 1,
      aux_sym_number_token2,
    ACTIONS(6500), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(6508), 1,
      sym_test_operator,
    ACTIONS(6510), 1,
      sym_extglob_pattern,
    ACTIONS(6512), 1,
      sym__brace_start,
    ACTIONS(7226), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(7228), 1,
      sym__special_character,
    ACTIONS(7230), 1,
      anon_sym_DQUOTE,
    ACTIONS(7234), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(7236), 1,
      anon_sym_BQUOTE,
    ACTIONS(7238), 1,
      anon_sym_DOLLAR_BQUOTE,
    STATE(6426), 1,
      aux_sym__literal_repeat1,
    STATE(7021), 1,
      sym_last_case_item,
    ACTIONS(7224), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(7232), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(7240), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3477), 2,
      sym_case_item,
      aux_sym_case_statement_repeat1,
    STATE(6695), 2,
      sym_concatenation,
      sym__extglob_blob,
    STATE(6303), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [182737] = 7,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7719), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7717), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7723), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7721), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7248), 7,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7274), 21,
      anon_sym_COMMA,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_DASHo,
      anon_sym_AMP_AMP,
      anon_sym_DASHa,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RPAREN,
  [182789] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7035), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(7033), 22,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [182833] = 8,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7719), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7717), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7723), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7725), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7721), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7248), 5,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7274), 21,
      anon_sym_COMMA,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_DASHo,
      anon_sym_AMP_AMP,
      anon_sym_DASHa,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RPAREN,
  [182887] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1350), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1348), 31,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [182931] = 10,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7719), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7717), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7723), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7725), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7727), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7729), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7248), 3,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(7721), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7274), 19,
      anon_sym_COMMA,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_DASHo,
      anon_sym_AMP_AMP,
      anon_sym_DASHa,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_RPAREN,
  [182989] = 11,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7719), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7717), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7723), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7725), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7727), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7729), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7731), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7248), 3,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(7721), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7274), 17,
      anon_sym_COMMA,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_DASHo,
      anon_sym_AMP_AMP,
      anon_sym_DASHa,
      anon_sym_RPAREN,
  [183049] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7053), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(7051), 22,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [183093] = 19,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6877), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7037), 1,
      anon_sym_PIPE,
    ACTIONS(7039), 1,
      anon_sym_CARET,
    ACTIONS(7041), 1,
      anon_sym_AMP,
    ACTIONS(7049), 1,
      sym_test_operator,
    ACTIONS(7072), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7074), 1,
      anon_sym_EQ,
    ACTIONS(7076), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7078), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(7703), 1,
      anon_sym_QMARK,
    ACTIONS(6793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6871), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(6873), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7043), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7045), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7047), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(6875), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7733), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [183169] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7064), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(7062), 22,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [183213] = 12,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7719), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7735), 1,
      anon_sym_AMP,
    ACTIONS(7248), 2,
      anon_sym_PIPE,
      anon_sym_CARET,
    ACTIONS(7717), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7723), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7725), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7727), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7729), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7731), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7721), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7274), 17,
      anon_sym_COMMA,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_DASHo,
      anon_sym_AMP_AMP,
      anon_sym_DASHa,
      anon_sym_RPAREN,
  [183275] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7064), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(7062), 22,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [183319] = 19,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6883), 1,
      anon_sym_EQ,
    ACTIONS(6889), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(6891), 1,
      anon_sym_AMP_AMP,
    ACTIONS(6893), 1,
      anon_sym_PIPE,
    ACTIONS(6895), 1,
      anon_sym_CARET,
    ACTIONS(6897), 1,
      anon_sym_AMP,
    ACTIONS(6911), 1,
      anon_sym_STAR_STAR,
    ACTIONS(6913), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(6915), 1,
      anon_sym_QMARK,
    ACTIONS(7737), 1,
      anon_sym_COLON,
    ACTIONS(6885), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(6899), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(6901), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(6903), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(6905), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(6907), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(6909), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(6887), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [183395] = 13,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7248), 1,
      anon_sym_PIPE,
    ACTIONS(7719), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7735), 1,
      anon_sym_AMP,
    ACTIONS(7739), 1,
      anon_sym_CARET,
    ACTIONS(7717), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7723), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7725), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7727), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7729), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7731), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7721), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7274), 17,
      anon_sym_COMMA,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_DASHo,
      anon_sym_AMP_AMP,
      anon_sym_DASHa,
      anon_sym_RPAREN,
  [183459] = 13,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7719), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7735), 1,
      anon_sym_AMP,
    ACTIONS(7739), 1,
      anon_sym_CARET,
    ACTIONS(7741), 1,
      anon_sym_PIPE,
    ACTIONS(7717), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7723), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7725), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7727), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7729), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7731), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7721), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7274), 17,
      anon_sym_COMMA,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_DASHo,
      anon_sym_AMP_AMP,
      anon_sym_DASHa,
      anon_sym_RPAREN,
  [183523] = 14,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7719), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7735), 1,
      anon_sym_AMP,
    ACTIONS(7739), 1,
      anon_sym_CARET,
    ACTIONS(7741), 1,
      anon_sym_PIPE,
    ACTIONS(7717), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7723), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7725), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7727), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7729), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7731), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7743), 2,
      anon_sym_AMP_AMP,
      anon_sym_DASHa,
    ACTIONS(7721), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7274), 15,
      anon_sym_COMMA,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_DASHo,
      anon_sym_RPAREN,
  [183589] = 15,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7719), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7735), 1,
      anon_sym_AMP,
    ACTIONS(7739), 1,
      anon_sym_CARET,
    ACTIONS(7741), 1,
      anon_sym_PIPE,
    ACTIONS(7717), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7723), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7725), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7727), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7729), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7731), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7743), 2,
      anon_sym_AMP_AMP,
      anon_sym_DASHa,
    ACTIONS(7745), 2,
      anon_sym_PIPE_PIPE,
      anon_sym_DASHo,
    ACTIONS(7721), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7274), 13,
      anon_sym_COMMA,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_RPAREN,
  [183657] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7244), 13,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(7246), 23,
      anon_sym_COMMA,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_DASHo,
      anon_sym_AMP_AMP,
      anon_sym_DASHa,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RPAREN,
  [183701] = 24,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6474), 1,
      sym_word,
    ACTIONS(6480), 1,
      anon_sym_LPAREN,
    ACTIONS(6488), 1,
      anon_sym_DOLLAR,
    ACTIONS(6494), 1,
      aux_sym_number_token1,
    ACTIONS(6496), 1,
      aux_sym_number_token2,
    ACTIONS(6500), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(6508), 1,
      sym_test_operator,
    ACTIONS(6510), 1,
      sym_extglob_pattern,
    ACTIONS(6512), 1,
      sym__brace_start,
    ACTIONS(7226), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(7228), 1,
      sym__special_character,
    ACTIONS(7230), 1,
      anon_sym_DQUOTE,
    ACTIONS(7234), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(7236), 1,
      anon_sym_BQUOTE,
    ACTIONS(7238), 1,
      anon_sym_DOLLAR_BQUOTE,
    STATE(6426), 1,
      aux_sym__literal_repeat1,
    STATE(6898), 1,
      sym_last_case_item,
    ACTIONS(7224), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(7232), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(7240), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3477), 2,
      sym_case_item,
      aux_sym_case_statement_repeat1,
    STATE(6695), 2,
      sym_concatenation,
      sym__extglob_blob,
    STATE(6303), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [183787] = 19,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7617), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7619), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7621), 1,
      anon_sym_PIPE,
    ACTIONS(7623), 1,
      anon_sym_CARET,
    ACTIONS(7625), 1,
      anon_sym_AMP,
    ACTIONS(7639), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7679), 1,
      anon_sym_EQ,
    ACTIONS(7681), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(7683), 1,
      anon_sym_QMARK,
    ACTIONS(7747), 1,
      anon_sym_RPAREN,
    ACTIONS(7615), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7627), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7629), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7631), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7633), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7635), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7637), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7711), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [183863] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7569), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7139), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(7137), 20,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [183909] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7595), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7569), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7139), 13,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7137), 20,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [183957] = 24,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6474), 1,
      sym_word,
    ACTIONS(6480), 1,
      anon_sym_LPAREN,
    ACTIONS(6488), 1,
      anon_sym_DOLLAR,
    ACTIONS(6494), 1,
      aux_sym_number_token1,
    ACTIONS(6496), 1,
      aux_sym_number_token2,
    ACTIONS(6500), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(6508), 1,
      sym_test_operator,
    ACTIONS(6510), 1,
      sym_extglob_pattern,
    ACTIONS(6512), 1,
      sym__brace_start,
    ACTIONS(7226), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(7228), 1,
      sym__special_character,
    ACTIONS(7230), 1,
      anon_sym_DQUOTE,
    ACTIONS(7234), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(7236), 1,
      anon_sym_BQUOTE,
    ACTIONS(7238), 1,
      anon_sym_DOLLAR_BQUOTE,
    STATE(6426), 1,
      aux_sym__literal_repeat1,
    STATE(7440), 1,
      sym_last_case_item,
    ACTIONS(7224), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(7232), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(7240), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3477), 2,
      sym_case_item,
      aux_sym_case_statement_repeat1,
    STATE(6695), 2,
      sym_concatenation,
      sym__extglob_blob,
    STATE(6303), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [184043] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7595), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7569), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7593), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7139), 10,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7137), 20,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [184093] = 7,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7595), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7569), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7591), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7593), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7139), 8,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7137), 20,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [184145] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1241), 1,
      sym_file_descriptor,
    ACTIONS(4084), 1,
      anon_sym_DQUOTE,
    ACTIONS(7613), 1,
      sym_variable_name,
    STATE(4606), 1,
      sym_string,
    ACTIONS(7611), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(7609), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1239), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
  [184199] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1253), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1251), 31,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [184243] = 24,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6474), 1,
      sym_word,
    ACTIONS(6480), 1,
      anon_sym_LPAREN,
    ACTIONS(6488), 1,
      anon_sym_DOLLAR,
    ACTIONS(6494), 1,
      aux_sym_number_token1,
    ACTIONS(6496), 1,
      aux_sym_number_token2,
    ACTIONS(6500), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(6508), 1,
      sym_test_operator,
    ACTIONS(6510), 1,
      sym_extglob_pattern,
    ACTIONS(6512), 1,
      sym__brace_start,
    ACTIONS(7226), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(7228), 1,
      sym__special_character,
    ACTIONS(7230), 1,
      anon_sym_DQUOTE,
    ACTIONS(7234), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(7236), 1,
      anon_sym_BQUOTE,
    ACTIONS(7238), 1,
      anon_sym_DOLLAR_BQUOTE,
    STATE(6426), 1,
      aux_sym__literal_repeat1,
    STATE(7451), 1,
      sym_last_case_item,
    ACTIONS(7224), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(7232), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(7240), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3477), 2,
      sym_case_item,
      aux_sym_case_statement_repeat1,
    STATE(6695), 2,
      sym_concatenation,
      sym__extglob_blob,
    STATE(6303), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [184329] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1235), 1,
      sym_file_descriptor,
    ACTIONS(4084), 1,
      anon_sym_DQUOTE,
    ACTIONS(7613), 1,
      sym_variable_name,
    STATE(4606), 1,
      sym_string,
    ACTIONS(7611), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(7609), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1227), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
  [184383] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1326), 5,
      sym_file_descriptor,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1324), 31,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [184427] = 24,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6474), 1,
      sym_word,
    ACTIONS(6480), 1,
      anon_sym_LPAREN,
    ACTIONS(6488), 1,
      anon_sym_DOLLAR,
    ACTIONS(6494), 1,
      aux_sym_number_token1,
    ACTIONS(6496), 1,
      aux_sym_number_token2,
    ACTIONS(6500), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(6508), 1,
      sym_test_operator,
    ACTIONS(6510), 1,
      sym_extglob_pattern,
    ACTIONS(6512), 1,
      sym__brace_start,
    ACTIONS(7226), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(7228), 1,
      sym__special_character,
    ACTIONS(7230), 1,
      anon_sym_DQUOTE,
    ACTIONS(7234), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(7236), 1,
      anon_sym_BQUOTE,
    ACTIONS(7238), 1,
      anon_sym_DOLLAR_BQUOTE,
    STATE(6426), 1,
      aux_sym__literal_repeat1,
    STATE(7494), 1,
      sym_last_case_item,
    ACTIONS(7224), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(7232), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(7240), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3477), 2,
      sym_case_item,
      aux_sym_case_statement_repeat1,
    STATE(6695), 2,
      sym_concatenation,
      sym__extglob_blob,
    STATE(6303), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [184513] = 24,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6474), 1,
      sym_word,
    ACTIONS(6480), 1,
      anon_sym_LPAREN,
    ACTIONS(6488), 1,
      anon_sym_DOLLAR,
    ACTIONS(6494), 1,
      aux_sym_number_token1,
    ACTIONS(6496), 1,
      aux_sym_number_token2,
    ACTIONS(6500), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(6508), 1,
      sym_test_operator,
    ACTIONS(6510), 1,
      sym_extglob_pattern,
    ACTIONS(6512), 1,
      sym__brace_start,
    ACTIONS(7226), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(7228), 1,
      sym__special_character,
    ACTIONS(7230), 1,
      anon_sym_DQUOTE,
    ACTIONS(7234), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(7236), 1,
      anon_sym_BQUOTE,
    ACTIONS(7238), 1,
      anon_sym_DOLLAR_BQUOTE,
    STATE(6426), 1,
      aux_sym__literal_repeat1,
    STATE(7495), 1,
      sym_last_case_item,
    ACTIONS(7224), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(7232), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(7240), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3477), 2,
      sym_case_item,
      aux_sym_case_statement_repeat1,
    STATE(6695), 2,
      sym_concatenation,
      sym__extglob_blob,
    STATE(6303), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [184599] = 8,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7595), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7569), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7589), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7591), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7593), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7139), 6,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7137), 20,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [184653] = 24,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6474), 1,
      sym_word,
    ACTIONS(6480), 1,
      anon_sym_LPAREN,
    ACTIONS(6488), 1,
      anon_sym_DOLLAR,
    ACTIONS(6494), 1,
      aux_sym_number_token1,
    ACTIONS(6496), 1,
      aux_sym_number_token2,
    ACTIONS(6500), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(6508), 1,
      sym_test_operator,
    ACTIONS(6510), 1,
      sym_extglob_pattern,
    ACTIONS(6512), 1,
      sym__brace_start,
    ACTIONS(7226), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(7228), 1,
      sym__special_character,
    ACTIONS(7230), 1,
      anon_sym_DQUOTE,
    ACTIONS(7234), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(7236), 1,
      anon_sym_BQUOTE,
    ACTIONS(7238), 1,
      anon_sym_DOLLAR_BQUOTE,
    STATE(6426), 1,
      aux_sym__literal_repeat1,
    STATE(7503), 1,
      sym_last_case_item,
    ACTIONS(7224), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(7232), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(7240), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3477), 2,
      sym_case_item,
      aux_sym_case_statement_repeat1,
    STATE(6695), 2,
      sym_concatenation,
      sym__extglob_blob,
    STATE(6303), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [184739] = 24,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6474), 1,
      sym_word,
    ACTIONS(6480), 1,
      anon_sym_LPAREN,
    ACTIONS(6488), 1,
      anon_sym_DOLLAR,
    ACTIONS(6494), 1,
      aux_sym_number_token1,
    ACTIONS(6496), 1,
      aux_sym_number_token2,
    ACTIONS(6500), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(6508), 1,
      sym_test_operator,
    ACTIONS(6510), 1,
      sym_extglob_pattern,
    ACTIONS(6512), 1,
      sym__brace_start,
    ACTIONS(7226), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(7228), 1,
      sym__special_character,
    ACTIONS(7230), 1,
      anon_sym_DQUOTE,
    ACTIONS(7234), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(7236), 1,
      anon_sym_BQUOTE,
    ACTIONS(7238), 1,
      anon_sym_DOLLAR_BQUOTE,
    STATE(6426), 1,
      aux_sym__literal_repeat1,
    STATE(7504), 1,
      sym_last_case_item,
    ACTIONS(7224), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(7232), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(7240), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3477), 2,
      sym_case_item,
      aux_sym_case_statement_repeat1,
    STATE(6695), 2,
      sym_concatenation,
      sym__extglob_blob,
    STATE(6303), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [184825] = 24,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6474), 1,
      sym_word,
    ACTIONS(6480), 1,
      anon_sym_LPAREN,
    ACTIONS(6488), 1,
      anon_sym_DOLLAR,
    ACTIONS(6494), 1,
      aux_sym_number_token1,
    ACTIONS(6496), 1,
      aux_sym_number_token2,
    ACTIONS(6500), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(6508), 1,
      sym_test_operator,
    ACTIONS(6510), 1,
      sym_extglob_pattern,
    ACTIONS(6512), 1,
      sym__brace_start,
    ACTIONS(7226), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(7228), 1,
      sym__special_character,
    ACTIONS(7230), 1,
      anon_sym_DQUOTE,
    ACTIONS(7234), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(7236), 1,
      anon_sym_BQUOTE,
    ACTIONS(7238), 1,
      anon_sym_DOLLAR_BQUOTE,
    STATE(6426), 1,
      aux_sym__literal_repeat1,
    STATE(7489), 1,
      sym_last_case_item,
    ACTIONS(7224), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(7232), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(7240), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3477), 2,
      sym_case_item,
      aux_sym_case_statement_repeat1,
    STATE(6695), 2,
      sym_concatenation,
      sym__extglob_blob,
    STATE(6303), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [184911] = 10,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7595), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7569), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7585), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7587), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7589), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7591), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7593), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7139), 4,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(7137), 18,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [184969] = 19,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7129), 1,
      anon_sym_RBRACK,
    ACTIONS(7567), 1,
      anon_sym_EQ,
    ACTIONS(7573), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7575), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7577), 1,
      anon_sym_PIPE,
    ACTIONS(7579), 1,
      anon_sym_CARET,
    ACTIONS(7581), 1,
      anon_sym_AMP,
    ACTIONS(7595), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7597), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(7599), 1,
      anon_sym_QMARK,
    ACTIONS(7569), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7583), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7585), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7587), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7589), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7591), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7593), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7571), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [185045] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1340), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1342), 22,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [185089] = 11,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7595), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7569), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7583), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7585), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7587), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7589), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7591), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7593), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7139), 4,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(7137), 16,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [185149] = 12,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7581), 1,
      anon_sym_AMP,
    ACTIONS(7595), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7569), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7583), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7585), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7587), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7589), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7591), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7139), 3,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
    ACTIONS(7593), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7137), 16,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [185211] = 13,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7579), 1,
      anon_sym_CARET,
    ACTIONS(7581), 1,
      anon_sym_AMP,
    ACTIONS(7595), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7139), 2,
      anon_sym_EQ,
      anon_sym_PIPE,
    ACTIONS(7569), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7583), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7585), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7587), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7589), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7591), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7593), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7137), 16,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [185275] = 14,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7139), 1,
      anon_sym_EQ,
    ACTIONS(7577), 1,
      anon_sym_PIPE,
    ACTIONS(7579), 1,
      anon_sym_CARET,
    ACTIONS(7581), 1,
      anon_sym_AMP,
    ACTIONS(7595), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7569), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7583), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7585), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7587), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7589), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7591), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7593), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7137), 16,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [185341] = 15,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7139), 1,
      anon_sym_EQ,
    ACTIONS(7575), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7577), 1,
      anon_sym_PIPE,
    ACTIONS(7579), 1,
      anon_sym_CARET,
    ACTIONS(7581), 1,
      anon_sym_AMP,
    ACTIONS(7595), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7569), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7583), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7585), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7587), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7589), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7591), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7593), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7137), 15,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [185409] = 18,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7567), 1,
      anon_sym_EQ,
    ACTIONS(7573), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7575), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7577), 1,
      anon_sym_PIPE,
    ACTIONS(7579), 1,
      anon_sym_CARET,
    ACTIONS(7581), 1,
      anon_sym_AMP,
    ACTIONS(7595), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7597), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(7599), 1,
      anon_sym_QMARK,
    ACTIONS(7569), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7583), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7585), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7587), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7589), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7591), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7593), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7137), 12,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_RBRACK,
  [185483] = 17,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7139), 1,
      anon_sym_EQ,
    ACTIONS(7573), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7575), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7577), 1,
      anon_sym_PIPE,
    ACTIONS(7579), 1,
      anon_sym_CARET,
    ACTIONS(7581), 1,
      anon_sym_AMP,
    ACTIONS(7595), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7599), 1,
      anon_sym_QMARK,
    ACTIONS(7569), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7583), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7585), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7587), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7589), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7591), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7593), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7137), 13,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
  [185555] = 19,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7029), 1,
      anon_sym_RBRACK,
    ACTIONS(7567), 1,
      anon_sym_EQ,
    ACTIONS(7573), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(7575), 1,
      anon_sym_AMP_AMP,
    ACTIONS(7577), 1,
      anon_sym_PIPE,
    ACTIONS(7579), 1,
      anon_sym_CARET,
    ACTIONS(7581), 1,
      anon_sym_AMP,
    ACTIONS(7595), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7597), 1,
      anon_sym_EQ_TILDE,
    ACTIONS(7599), 1,
      anon_sym_QMARK,
    ACTIONS(7569), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7583), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7585), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7587), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7589), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7591), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7593), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7571), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [185631] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7114), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(7112), 22,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [185675] = 24,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6474), 1,
      sym_word,
    ACTIONS(6480), 1,
      anon_sym_LPAREN,
    ACTIONS(6488), 1,
      anon_sym_DOLLAR,
    ACTIONS(6494), 1,
      aux_sym_number_token1,
    ACTIONS(6496), 1,
      aux_sym_number_token2,
    ACTIONS(6500), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(6508), 1,
      sym_test_operator,
    ACTIONS(6510), 1,
      sym_extglob_pattern,
    ACTIONS(6512), 1,
      sym__brace_start,
    ACTIONS(7226), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(7228), 1,
      sym__special_character,
    ACTIONS(7230), 1,
      anon_sym_DQUOTE,
    ACTIONS(7234), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(7236), 1,
      anon_sym_BQUOTE,
    ACTIONS(7238), 1,
      anon_sym_DOLLAR_BQUOTE,
    STATE(6426), 1,
      aux_sym__literal_repeat1,
    STATE(7035), 1,
      sym_last_case_item,
    ACTIONS(7224), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(7232), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(7240), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3477), 2,
      sym_case_item,
      aux_sym_case_statement_repeat1,
    STATE(6695), 2,
      sym_concatenation,
      sym__extglob_blob,
    STATE(6303), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [185761] = 24,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6474), 1,
      sym_word,
    ACTIONS(6480), 1,
      anon_sym_LPAREN,
    ACTIONS(6488), 1,
      anon_sym_DOLLAR,
    ACTIONS(6494), 1,
      aux_sym_number_token1,
    ACTIONS(6496), 1,
      aux_sym_number_token2,
    ACTIONS(6500), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(6508), 1,
      sym_test_operator,
    ACTIONS(6510), 1,
      sym_extglob_pattern,
    ACTIONS(6512), 1,
      sym__brace_start,
    ACTIONS(7226), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(7228), 1,
      sym__special_character,
    ACTIONS(7230), 1,
      anon_sym_DQUOTE,
    ACTIONS(7234), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(7236), 1,
      anon_sym_BQUOTE,
    ACTIONS(7238), 1,
      anon_sym_DOLLAR_BQUOTE,
    STATE(6426), 1,
      aux_sym__literal_repeat1,
    STATE(7049), 1,
      sym_last_case_item,
    ACTIONS(7224), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(7232), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(7240), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3477), 2,
      sym_case_item,
      aux_sym_case_statement_repeat1,
    STATE(6695), 2,
      sym_concatenation,
      sym__extglob_blob,
    STATE(6303), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [185847] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1336), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1338), 22,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [185891] = 24,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6474), 1,
      sym_word,
    ACTIONS(6480), 1,
      anon_sym_LPAREN,
    ACTIONS(6488), 1,
      anon_sym_DOLLAR,
    ACTIONS(6494), 1,
      aux_sym_number_token1,
    ACTIONS(6496), 1,
      aux_sym_number_token2,
    ACTIONS(6500), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(6508), 1,
      sym_test_operator,
    ACTIONS(6510), 1,
      sym_extglob_pattern,
    ACTIONS(6512), 1,
      sym__brace_start,
    ACTIONS(7226), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(7228), 1,
      sym__special_character,
    ACTIONS(7230), 1,
      anon_sym_DQUOTE,
    ACTIONS(7234), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(7236), 1,
      anon_sym_BQUOTE,
    ACTIONS(7238), 1,
      anon_sym_DOLLAR_BQUOTE,
    STATE(6426), 1,
      aux_sym__literal_repeat1,
    STATE(7315), 1,
      sym_last_case_item,
    ACTIONS(7224), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(7232), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(7240), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3477), 2,
      sym_case_item,
      aux_sym_case_statement_repeat1,
    STATE(6695), 2,
      sym_concatenation,
      sym__extglob_blob,
    STATE(6303), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [185977] = 16,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7719), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7735), 1,
      anon_sym_AMP,
    ACTIONS(7739), 1,
      anon_sym_CARET,
    ACTIONS(7741), 1,
      anon_sym_PIPE,
    ACTIONS(7511), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
    ACTIONS(7717), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7723), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7725), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7727), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7729), 2,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(7731), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(7743), 2,
      anon_sym_AMP_AMP,
      anon_sym_DASHa,
    ACTIONS(7745), 2,
      anon_sym_PIPE_PIPE,
      anon_sym_DASHo,
    ACTIONS(7721), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7749), 11,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [186047] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7489), 13,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(7491), 23,
      anon_sym_COMMA,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_DASHo,
      anon_sym_AMP_AMP,
      anon_sym_DASHa,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RPAREN,
  [186091] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7501), 13,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(7503), 23,
      anon_sym_COMMA,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_DASHo,
      anon_sym_AMP_AMP,
      anon_sym_DASHa,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RPAREN,
  [186135] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7489), 13,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(7491), 23,
      anon_sym_COMMA,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_DASHo,
      anon_sym_AMP_AMP,
      anon_sym_DASHa,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RPAREN,
  [186179] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1344), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1346), 22,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [186223] = 8,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7595), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7569), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7589), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7591), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7593), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7096), 6,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7094), 20,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [186277] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7569), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7096), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(7094), 20,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [186323] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1356), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1358), 22,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [186367] = 8,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7595), 1,
      anon_sym_STAR_STAR,
    ACTIONS(7569), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7589), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(7591), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(7593), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(7096), 6,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(7094), 20,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [186421] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7106), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(7104), 22,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RBRACK,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [186465] = 24,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6474), 1,
      sym_word,
    ACTIONS(6480), 1,
      anon_sym_LPAREN,
    ACTIONS(6488), 1,
      anon_sym_DOLLAR,
    ACTIONS(6494), 1,
      aux_sym_number_token1,
    ACTIONS(6496), 1,
      aux_sym_number_token2,
    ACTIONS(6500), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(6508), 1,
      sym_test_operator,
    ACTIONS(6510), 1,
      sym_extglob_pattern,
    ACTIONS(6512), 1,
      sym__brace_start,
    ACTIONS(7226), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(7228), 1,
      sym__special_character,
    ACTIONS(7230), 1,
      anon_sym_DQUOTE,
    ACTIONS(7234), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(7236), 1,
      anon_sym_BQUOTE,
    ACTIONS(7238), 1,
      anon_sym_DOLLAR_BQUOTE,
    STATE(6426), 1,
      aux_sym__literal_repeat1,
    STATE(7528), 1,
      sym_last_case_item,
    ACTIONS(7224), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(7232), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(7240), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3477), 2,
      sym_case_item,
      aux_sym_case_statement_repeat1,
    STATE(6695), 2,
      sym_concatenation,
      sym__extglob_blob,
    STATE(6303), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [186551] = 24,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6474), 1,
      sym_word,
    ACTIONS(6480), 1,
      anon_sym_LPAREN,
    ACTIONS(6488), 1,
      anon_sym_DOLLAR,
    ACTIONS(6494), 1,
      aux_sym_number_token1,
    ACTIONS(6496), 1,
      aux_sym_number_token2,
    ACTIONS(6500), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(6508), 1,
      sym_test_operator,
    ACTIONS(6510), 1,
      sym_extglob_pattern,
    ACTIONS(6512), 1,
      sym__brace_start,
    ACTIONS(7226), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(7228), 1,
      sym__special_character,
    ACTIONS(7230), 1,
      anon_sym_DQUOTE,
    ACTIONS(7234), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(7236), 1,
      anon_sym_BQUOTE,
    ACTIONS(7238), 1,
      anon_sym_DOLLAR_BQUOTE,
    STATE(6426), 1,
      aux_sym__literal_repeat1,
    STATE(7603), 1,
      sym_last_case_item,
    ACTIONS(7224), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(7232), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(7240), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3477), 2,
      sym_case_item,
      aux_sym_case_statement_repeat1,
    STATE(6695), 2,
      sym_concatenation,
      sym__extglob_blob,
    STATE(6303), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [186637] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1352), 14,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(1354), 22,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RPAREN,
      anon_sym_EQ_TILDE,
      anon_sym_QMARK,
  [186681] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7717), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(7469), 13,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(7471), 21,
      anon_sym_COMMA,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_DASHo,
      anon_sym_AMP_AMP,
      anon_sym_DASHa,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RPAREN,
  [186727] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7465), 13,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
    ACTIONS(7467), 23,
      anon_sym_COMMA,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_DASHo,
      anon_sym_AMP_AMP,
      anon_sym_DASHa,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_RPAREN,
  [186771] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(3461), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(7751), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(4370), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(4372), 21,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [186818] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3735), 1,
      anon_sym_DQUOTE,
    ACTIONS(7757), 1,
      sym_variable_name,
    STATE(4808), 1,
      sym_string,
    ACTIONS(1241), 2,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(7755), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(7753), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1239), 19,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
      aux_sym_heredoc_redirect_token1,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [186871] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3729), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(3731), 1,
      anon_sym_DOLLAR,
    ACTIONS(3735), 1,
      anon_sym_DQUOTE,
    ACTIONS(3737), 1,
      aux_sym_number_token1,
    ACTIONS(3739), 1,
      aux_sym_number_token2,
    ACTIONS(3741), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(3743), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3745), 1,
      anon_sym_BQUOTE,
    ACTIONS(3747), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3755), 1,
      sym__brace_start,
    ACTIONS(7763), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(7765), 1,
      sym__special_character,
    ACTIONS(7767), 1,
      sym_test_operator,
    STATE(4975), 1,
      aux_sym__literal_repeat1,
    ACTIONS(3715), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(3749), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3478), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    ACTIONS(7759), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    ACTIONS(7761), 3,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
    STATE(4460), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [186950] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7751), 1,
      aux_sym_concatenation_token1,
    ACTIONS(7769), 1,
      sym__concat,
    STATE(3468), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1265), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1267), 21,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [186999] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3729), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(3731), 1,
      anon_sym_DOLLAR,
    ACTIONS(3735), 1,
      anon_sym_DQUOTE,
    ACTIONS(3737), 1,
      aux_sym_number_token1,
    ACTIONS(3739), 1,
      aux_sym_number_token2,
    ACTIONS(3741), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(3743), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3745), 1,
      anon_sym_BQUOTE,
    ACTIONS(3747), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3755), 1,
      sym__brace_start,
    ACTIONS(7765), 1,
      sym__special_character,
    ACTIONS(7767), 1,
      sym_test_operator,
    ACTIONS(7773), 1,
      aux_sym_heredoc_redirect_token1,
    STATE(4975), 1,
      aux_sym__literal_repeat1,
    ACTIONS(3715), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(3749), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3478), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    ACTIONS(7759), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    ACTIONS(7771), 3,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
    STATE(4460), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [187078] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7751), 1,
      aux_sym_concatenation_token1,
    ACTIONS(7775), 1,
      sym__concat,
    STATE(3468), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1286), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1288), 21,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [187127] = 25,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7777), 1,
      sym_word,
    ACTIONS(7781), 1,
      anon_sym_LPAREN,
    ACTIONS(7783), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(7785), 1,
      anon_sym_DOLLAR,
    ACTIONS(7787), 1,
      sym__special_character,
    ACTIONS(7789), 1,
      anon_sym_DQUOTE,
    ACTIONS(7793), 1,
      aux_sym_number_token1,
    ACTIONS(7795), 1,
      aux_sym_number_token2,
    ACTIONS(7797), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(7799), 1,
      anon_sym_RBRACE3,
    ACTIONS(7801), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(7803), 1,
      anon_sym_BQUOTE,
    ACTIONS(7805), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(7809), 1,
      sym_variable_name,
    ACTIONS(7811), 1,
      sym_test_operator,
    ACTIONS(7813), 1,
      sym__expansion_word,
    ACTIONS(7815), 1,
      sym__brace_start,
    STATE(6346), 1,
      sym_command_substitution,
    STATE(6482), 1,
      aux_sym__literal_repeat1,
    ACTIONS(7779), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(7791), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(7807), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(6427), 5,
      sym_string,
      sym_array,
      sym_simple_expansion,
      sym_expansion,
      sym_process_substitution,
    STATE(6782), 5,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_translated_string,
      sym_number,
      sym__concatenation_in_expansion,
  [187214] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1241), 1,
      sym_file_descriptor,
    ACTIONS(7673), 1,
      anon_sym_DQUOTE,
    ACTIONS(7677), 1,
      sym_variable_name,
    STATE(4393), 1,
      sym_string,
    ACTIONS(7675), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(7671), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1239), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
  [187267] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1235), 1,
      sym_file_descriptor,
    ACTIONS(7673), 1,
      anon_sym_DQUOTE,
    ACTIONS(7677), 1,
      sym_variable_name,
    STATE(4393), 1,
      sym_string,
    ACTIONS(7675), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(7671), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1227), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
  [187320] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(3459), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(7751), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(5067), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(5069), 21,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [187367] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(3461), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(7751), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(5373), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(5375), 21,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [187414] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(3461), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(7751), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(4552), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(4554), 21,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [187461] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(3468), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(7817), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(1251), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1253), 21,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [187508] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(7820), 1,
      sym__special_character,
    STATE(3473), 1,
      aux_sym__literal_repeat1,
    ACTIONS(5395), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5393), 29,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [187555] = 25,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7781), 1,
      anon_sym_LPAREN,
    ACTIONS(7783), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(7785), 1,
      anon_sym_DOLLAR,
    ACTIONS(7787), 1,
      sym__special_character,
    ACTIONS(7789), 1,
      anon_sym_DQUOTE,
    ACTIONS(7793), 1,
      aux_sym_number_token1,
    ACTIONS(7795), 1,
      aux_sym_number_token2,
    ACTIONS(7797), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(7801), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(7803), 1,
      anon_sym_BQUOTE,
    ACTIONS(7805), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(7809), 1,
      sym_variable_name,
    ACTIONS(7815), 1,
      sym__brace_start,
    ACTIONS(7822), 1,
      sym_word,
    ACTIONS(7826), 1,
      anon_sym_RBRACE3,
    ACTIONS(7828), 1,
      sym_test_operator,
    ACTIONS(7830), 1,
      sym__expansion_word,
    STATE(6353), 1,
      sym_command_substitution,
    STATE(6530), 1,
      aux_sym__literal_repeat1,
    ACTIONS(7779), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(7807), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(7824), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(6379), 5,
      sym_string,
      sym_array,
      sym_simple_expansion,
      sym_expansion,
      sym_process_substitution,
    STATE(6788), 5,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_translated_string,
      sym_number,
      sym__concatenation_in_expansion,
  [187642] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(3461), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(7751), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(2094), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(2096), 21,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [187689] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(3459), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(7751), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(4564), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(4566), 21,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [187736] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(7832), 1,
      sym__special_character,
    STATE(3473), 1,
      aux_sym__literal_repeat1,
    ACTIONS(1364), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1362), 29,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [187783] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3735), 1,
      anon_sym_DQUOTE,
    ACTIONS(7757), 1,
      sym_variable_name,
    STATE(4808), 1,
      sym_string,
    ACTIONS(1235), 2,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(7755), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(7753), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1227), 19,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
      aux_sym_heredoc_redirect_token1,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [187836] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3729), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(3731), 1,
      anon_sym_DOLLAR,
    ACTIONS(3735), 1,
      anon_sym_DQUOTE,
    ACTIONS(3737), 1,
      aux_sym_number_token1,
    ACTIONS(3739), 1,
      aux_sym_number_token2,
    ACTIONS(3741), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(3743), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3745), 1,
      anon_sym_BQUOTE,
    ACTIONS(3747), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3755), 1,
      sym__brace_start,
    ACTIONS(7765), 1,
      sym__special_character,
    ACTIONS(7767), 1,
      sym_test_operator,
    ACTIONS(7837), 1,
      aux_sym_heredoc_redirect_token1,
    STATE(4975), 1,
      aux_sym__literal_repeat1,
    ACTIONS(3715), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(3749), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3478), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    ACTIONS(7759), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    ACTIONS(7835), 3,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
    STATE(4460), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [187915] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(3459), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(7751), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(1261), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1263), 21,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [187962] = 23,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7839), 1,
      sym_word,
    ACTIONS(7845), 1,
      anon_sym_LPAREN,
    ACTIONS(7848), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(7851), 1,
      anon_sym_DOLLAR,
    ACTIONS(7854), 1,
      sym__special_character,
    ACTIONS(7857), 1,
      anon_sym_DQUOTE,
    ACTIONS(7863), 1,
      aux_sym_number_token1,
    ACTIONS(7866), 1,
      aux_sym_number_token2,
    ACTIONS(7869), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(7872), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(7875), 1,
      anon_sym_BQUOTE,
    ACTIONS(7878), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(7884), 1,
      sym_test_operator,
    ACTIONS(7887), 1,
      sym_extglob_pattern,
    ACTIONS(7890), 1,
      sym__brace_start,
    STATE(6409), 1,
      aux_sym__literal_repeat1,
    ACTIONS(7842), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(7860), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(7881), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3477), 2,
      sym_case_item,
      aux_sym_case_statement_repeat1,
    STATE(6591), 2,
      sym_concatenation,
      sym__extglob_blob,
    STATE(6328), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [188045] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2216), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(7899), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(7902), 1,
      anon_sym_DOLLAR,
    ACTIONS(7905), 1,
      sym__special_character,
    ACTIONS(7908), 1,
      anon_sym_DQUOTE,
    ACTIONS(7911), 1,
      aux_sym_number_token1,
    ACTIONS(7914), 1,
      aux_sym_number_token2,
    ACTIONS(7917), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(7920), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(7923), 1,
      anon_sym_BQUOTE,
    ACTIONS(7926), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(7932), 1,
      sym_test_operator,
    ACTIONS(7935), 1,
      sym__brace_start,
    STATE(4975), 1,
      aux_sym__literal_repeat1,
    ACTIONS(7896), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(7929), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3478), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    ACTIONS(2214), 3,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
    ACTIONS(7893), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(4460), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [188124] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(3459), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(7751), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(5393), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(5395), 21,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [188171] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(3461), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(7751), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(5389), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(5391), 21,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [188218] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(3461), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(7751), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(2072), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(2074), 21,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [188265] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1241), 1,
      sym_file_descriptor,
    ACTIONS(4084), 1,
      anon_sym_DQUOTE,
    ACTIONS(7613), 1,
      sym_variable_name,
    STATE(4606), 1,
      sym_string,
    ACTIONS(7611), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(7609), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1239), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
  [188318] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(3459), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(7751), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(4467), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(4469), 21,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [188365] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3729), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(3731), 1,
      anon_sym_DOLLAR,
    ACTIONS(3735), 1,
      anon_sym_DQUOTE,
    ACTIONS(3737), 1,
      aux_sym_number_token1,
    ACTIONS(3739), 1,
      aux_sym_number_token2,
    ACTIONS(3741), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(3743), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3745), 1,
      anon_sym_BQUOTE,
    ACTIONS(3747), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3755), 1,
      sym__brace_start,
    ACTIONS(7765), 1,
      sym__special_character,
    ACTIONS(7767), 1,
      sym_test_operator,
    ACTIONS(7940), 1,
      aux_sym_heredoc_redirect_token1,
    STATE(4975), 1,
      aux_sym__literal_repeat1,
    ACTIONS(3715), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(3749), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3478), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    ACTIONS(7759), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    ACTIONS(7938), 3,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
    STATE(4460), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [188444] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1235), 1,
      sym_file_descriptor,
    ACTIONS(4084), 1,
      anon_sym_DQUOTE,
    ACTIONS(7613), 1,
      sym_variable_name,
    STATE(4606), 1,
      sym_string,
    ACTIONS(7611), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(7609), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1227), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
  [188497] = 25,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7781), 1,
      anon_sym_LPAREN,
    ACTIONS(7783), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(7785), 1,
      anon_sym_DOLLAR,
    ACTIONS(7787), 1,
      sym__special_character,
    ACTIONS(7789), 1,
      anon_sym_DQUOTE,
    ACTIONS(7793), 1,
      aux_sym_number_token1,
    ACTIONS(7795), 1,
      aux_sym_number_token2,
    ACTIONS(7797), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(7801), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(7803), 1,
      anon_sym_BQUOTE,
    ACTIONS(7805), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(7809), 1,
      sym_variable_name,
    ACTIONS(7815), 1,
      sym__brace_start,
    ACTIONS(7942), 1,
      sym_word,
    ACTIONS(7946), 1,
      anon_sym_RBRACE3,
    ACTIONS(7948), 1,
      sym_test_operator,
    ACTIONS(7950), 1,
      sym__expansion_word,
    STATE(6314), 1,
      sym_command_substitution,
    STATE(6488), 1,
      aux_sym__literal_repeat1,
    ACTIONS(7779), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(7807), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(7944), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(6449), 5,
      sym_string,
      sym_array,
      sym_simple_expansion,
      sym_expansion,
      sym_process_substitution,
    STATE(6787), 5,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_translated_string,
      sym_number,
      sym__concatenation_in_expansion,
  [188584] = 23,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(7952), 1,
      sym_word,
    ACTIONS(7956), 1,
      anon_sym_LPAREN,
    ACTIONS(7958), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(7960), 1,
      anon_sym_DOLLAR,
    ACTIONS(7962), 1,
      sym__special_character,
    ACTIONS(7964), 1,
      anon_sym_DQUOTE,
    ACTIONS(7968), 1,
      aux_sym_number_token1,
    ACTIONS(7970), 1,
      aux_sym_number_token2,
    ACTIONS(7972), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(7974), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(7976), 1,
      anon_sym_BQUOTE,
    ACTIONS(7978), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(7982), 1,
      sym__comment_word,
    ACTIONS(7984), 1,
      sym__empty_value,
    ACTIONS(7986), 1,
      sym_test_operator,
    ACTIONS(7988), 1,
      sym__brace_start,
    STATE(4287), 1,
      aux_sym__literal_repeat1,
    ACTIONS(7954), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(7966), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(7980), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(4759), 2,
      sym_concatenation,
      sym_array,
    STATE(5137), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [188666] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1320), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1322), 23,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [188708] = 23,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(7990), 1,
      sym_word,
    ACTIONS(7994), 1,
      anon_sym_LPAREN,
    ACTIONS(7996), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(7998), 1,
      anon_sym_DOLLAR,
    ACTIONS(8000), 1,
      sym__special_character,
    ACTIONS(8002), 1,
      anon_sym_DQUOTE,
    ACTIONS(8006), 1,
      aux_sym_number_token1,
    ACTIONS(8008), 1,
      aux_sym_number_token2,
    ACTIONS(8010), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8012), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(8014), 1,
      anon_sym_BQUOTE,
    ACTIONS(8016), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8020), 1,
      sym__comment_word,
    ACTIONS(8022), 1,
      sym__empty_value,
    ACTIONS(8024), 1,
      sym_test_operator,
    ACTIONS(8026), 1,
      sym__brace_start,
    STATE(3536), 1,
      aux_sym__literal_repeat1,
    ACTIONS(7992), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8004), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(8018), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3601), 2,
      sym_concatenation,
      sym_array,
    STATE(3466), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [188790] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1312), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1314), 23,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [188832] = 20,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8030), 1,
      anon_sym_SLASH,
    ACTIONS(8032), 1,
      anon_sym_PERCENT,
    ACTIONS(8034), 1,
      anon_sym_COLON,
    ACTIONS(8038), 1,
      anon_sym_RBRACE3,
    ACTIONS(8040), 1,
      anon_sym_AT,
    ACTIONS(8042), 1,
      anon_sym_STAR2,
    STATE(6394), 1,
      aux_sym__expansion_body_repeat1,
    STATE(6972), 1,
      sym__expansion_operator,
    STATE(6973), 1,
      sym__expansion_max_length,
    STATE(6974), 1,
      sym__expansion_regex_removal,
    STATE(6978), 1,
      sym__expansion_regex_replacement,
    STATE(6979), 1,
      sym__expansion_regex,
    STATE(6980), 1,
      sym__expansion_expression,
    ACTIONS(8028), 2,
      anon_sym_COMMA,
      anon_sym_CARET,
    ACTIONS(8048), 2,
      anon_sym_COMMA_COMMA,
      anon_sym_CARET_CARET,
    ACTIONS(8036), 3,
      sym__immediate_double_hash,
      anon_sym_POUND,
      anon_sym_PERCENT_PERCENT,
    ACTIONS(8046), 3,
      anon_sym_SLASH_SLASH,
      anon_sym_SLASH_POUND,
      anon_sym_SLASH_PERCENT,
    ACTIONS(8050), 3,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
    ACTIONS(8044), 8,
      anon_sym_EQ2,
      anon_sym_COLON_EQ,
      anon_sym_DASH3,
      anon_sym_COLON_DASH,
      anon_sym_PLUS3,
      anon_sym_COLON_PLUS,
      anon_sym_QMARK2,
      anon_sym_COLON_QMARK,
  [188908] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1336), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1338), 23,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [188950] = 23,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(7956), 1,
      anon_sym_LPAREN,
    ACTIONS(7958), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(7960), 1,
      anon_sym_DOLLAR,
    ACTIONS(7964), 1,
      anon_sym_DQUOTE,
    ACTIONS(7968), 1,
      aux_sym_number_token1,
    ACTIONS(7970), 1,
      aux_sym_number_token2,
    ACTIONS(7972), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(7974), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(7976), 1,
      anon_sym_BQUOTE,
    ACTIONS(7978), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(7982), 1,
      sym__comment_word,
    ACTIONS(7984), 1,
      sym__empty_value,
    ACTIONS(7988), 1,
      sym__brace_start,
    ACTIONS(8052), 1,
      sym_word,
    ACTIONS(8054), 1,
      sym__special_character,
    ACTIONS(8058), 1,
      sym_test_operator,
    STATE(4287), 1,
      aux_sym__literal_repeat1,
    ACTIONS(7954), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(7980), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(8056), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(4759), 2,
      sym_concatenation,
      sym_array,
    STATE(4802), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [189032] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1344), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1346), 23,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [189074] = 23,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2263), 1,
      anon_sym_DOLLAR,
    ACTIONS(2269), 1,
      aux_sym_number_token1,
    ACTIONS(2271), 1,
      aux_sym_number_token2,
    ACTIONS(2275), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2289), 1,
      sym__brace_start,
    ACTIONS(8060), 1,
      sym_word,
    ACTIONS(8064), 1,
      anon_sym_LPAREN,
    ACTIONS(8066), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8068), 1,
      sym__special_character,
    ACTIONS(8070), 1,
      anon_sym_DQUOTE,
    ACTIONS(8074), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8076), 1,
      anon_sym_BQUOTE,
    ACTIONS(8078), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8082), 1,
      sym__comment_word,
    ACTIONS(8084), 1,
      sym__empty_value,
    ACTIONS(8086), 1,
      sym_test_operator,
    STATE(1887), 1,
      aux_sym__literal_repeat1,
    ACTIONS(8062), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8072), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(8080), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(2158), 2,
      sym_concatenation,
      sym_array,
    STATE(2324), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [189156] = 23,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(7956), 1,
      anon_sym_LPAREN,
    ACTIONS(7958), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(7960), 1,
      anon_sym_DOLLAR,
    ACTIONS(7964), 1,
      anon_sym_DQUOTE,
    ACTIONS(7968), 1,
      aux_sym_number_token1,
    ACTIONS(7970), 1,
      aux_sym_number_token2,
    ACTIONS(7972), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(7974), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(7976), 1,
      anon_sym_BQUOTE,
    ACTIONS(7978), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(7982), 1,
      sym__comment_word,
    ACTIONS(7984), 1,
      sym__empty_value,
    ACTIONS(7988), 1,
      sym__brace_start,
    ACTIONS(8088), 1,
      sym_word,
    ACTIONS(8090), 1,
      sym__special_character,
    ACTIONS(8094), 1,
      sym_test_operator,
    STATE(4287), 1,
      aux_sym__literal_repeat1,
    ACTIONS(7954), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(7980), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(8092), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(4759), 2,
      sym_concatenation,
      sym_array,
    STATE(4550), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [189238] = 23,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8096), 1,
      sym_word,
    ACTIONS(8100), 1,
      anon_sym_LPAREN,
    ACTIONS(8102), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8104), 1,
      anon_sym_DOLLAR,
    ACTIONS(8106), 1,
      sym__special_character,
    ACTIONS(8108), 1,
      anon_sym_DQUOTE,
    ACTIONS(8112), 1,
      aux_sym_number_token1,
    ACTIONS(8114), 1,
      aux_sym_number_token2,
    ACTIONS(8116), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8118), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(8120), 1,
      anon_sym_BQUOTE,
    ACTIONS(8122), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8126), 1,
      sym__comment_word,
    ACTIONS(8128), 1,
      sym__empty_value,
    ACTIONS(8130), 1,
      sym_test_operator,
    ACTIONS(8132), 1,
      sym__brace_start,
    STATE(1361), 1,
      aux_sym__literal_repeat1,
    ACTIONS(8098), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8110), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(8124), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(1667), 2,
      sym_concatenation,
      sym_array,
    STATE(1148), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [189320] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1356), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1358), 23,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [189362] = 23,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8134), 1,
      sym_word,
    ACTIONS(8138), 1,
      anon_sym_LPAREN,
    ACTIONS(8140), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8142), 1,
      anon_sym_DOLLAR,
    ACTIONS(8144), 1,
      sym__special_character,
    ACTIONS(8146), 1,
      anon_sym_DQUOTE,
    ACTIONS(8150), 1,
      aux_sym_number_token1,
    ACTIONS(8152), 1,
      aux_sym_number_token2,
    ACTIONS(8154), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8156), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(8158), 1,
      anon_sym_BQUOTE,
    ACTIONS(8160), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8164), 1,
      sym__comment_word,
    ACTIONS(8166), 1,
      sym__empty_value,
    ACTIONS(8168), 1,
      sym_test_operator,
    ACTIONS(8170), 1,
      sym__brace_start,
    STATE(4431), 1,
      aux_sym__literal_repeat1,
    ACTIONS(8136), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8148), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(8162), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(4925), 2,
      sym_concatenation,
      sym_array,
    STATE(4410), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [189444] = 23,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8172), 1,
      sym_word,
    ACTIONS(8176), 1,
      anon_sym_LPAREN,
    ACTIONS(8178), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8180), 1,
      anon_sym_DOLLAR,
    ACTIONS(8182), 1,
      sym__special_character,
    ACTIONS(8184), 1,
      anon_sym_DQUOTE,
    ACTIONS(8188), 1,
      aux_sym_number_token1,
    ACTIONS(8190), 1,
      aux_sym_number_token2,
    ACTIONS(8192), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8194), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(8196), 1,
      anon_sym_BQUOTE,
    ACTIONS(8198), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8202), 1,
      sym__comment_word,
    ACTIONS(8204), 1,
      sym__empty_value,
    ACTIONS(8206), 1,
      sym_test_operator,
    ACTIONS(8208), 1,
      sym__brace_start,
    STATE(1742), 1,
      aux_sym__literal_repeat1,
    ACTIONS(8174), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8186), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(8200), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(1959), 2,
      sym_concatenation,
      sym_array,
    STATE(2279), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [189526] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1300), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1302), 23,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [189568] = 23,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(7956), 1,
      anon_sym_LPAREN,
    ACTIONS(7958), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(7960), 1,
      anon_sym_DOLLAR,
    ACTIONS(7964), 1,
      anon_sym_DQUOTE,
    ACTIONS(7968), 1,
      aux_sym_number_token1,
    ACTIONS(7970), 1,
      aux_sym_number_token2,
    ACTIONS(7972), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(7974), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(7976), 1,
      anon_sym_BQUOTE,
    ACTIONS(7978), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(7982), 1,
      sym__comment_word,
    ACTIONS(7984), 1,
      sym__empty_value,
    ACTIONS(7988), 1,
      sym__brace_start,
    ACTIONS(8210), 1,
      sym_word,
    ACTIONS(8212), 1,
      sym__special_character,
    ACTIONS(8216), 1,
      sym_test_operator,
    STATE(4287), 1,
      aux_sym__literal_repeat1,
    ACTIONS(7954), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(7980), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(8214), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(4759), 2,
      sym_concatenation,
      sym_array,
    STATE(4525), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [189650] = 23,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8176), 1,
      anon_sym_LPAREN,
    ACTIONS(8178), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8180), 1,
      anon_sym_DOLLAR,
    ACTIONS(8184), 1,
      anon_sym_DQUOTE,
    ACTIONS(8188), 1,
      aux_sym_number_token1,
    ACTIONS(8190), 1,
      aux_sym_number_token2,
    ACTIONS(8192), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8194), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(8196), 1,
      anon_sym_BQUOTE,
    ACTIONS(8198), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8202), 1,
      sym__comment_word,
    ACTIONS(8204), 1,
      sym__empty_value,
    ACTIONS(8208), 1,
      sym__brace_start,
    ACTIONS(8218), 1,
      sym_word,
    ACTIONS(8220), 1,
      sym__special_character,
    ACTIONS(8224), 1,
      sym_test_operator,
    STATE(1742), 1,
      aux_sym__literal_repeat1,
    ACTIONS(8174), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8200), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(8222), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1959), 2,
      sym_concatenation,
      sym_array,
    STATE(1355), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [189732] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1308), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1310), 23,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [189774] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1304), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1306), 23,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [189816] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1308), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1310), 23,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [189858] = 23,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1831), 1,
      anon_sym_DOLLAR,
    ACTIONS(1837), 1,
      aux_sym_number_token1,
    ACTIONS(1839), 1,
      aux_sym_number_token2,
    ACTIONS(1843), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1857), 1,
      sym__brace_start,
    ACTIONS(8226), 1,
      sym_word,
    ACTIONS(8230), 1,
      anon_sym_LPAREN,
    ACTIONS(8232), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8234), 1,
      sym__special_character,
    ACTIONS(8236), 1,
      anon_sym_DQUOTE,
    ACTIONS(8240), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8242), 1,
      anon_sym_BQUOTE,
    ACTIONS(8244), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8248), 1,
      sym__comment_word,
    ACTIONS(8250), 1,
      sym__empty_value,
    ACTIONS(8252), 1,
      sym_test_operator,
    STATE(1720), 1,
      aux_sym__literal_repeat1,
    ACTIONS(8228), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8238), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(8246), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(1735), 2,
      sym_concatenation,
      sym_array,
    STATE(1166), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [189940] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1324), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1326), 23,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [189982] = 23,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2263), 1,
      anon_sym_DOLLAR,
    ACTIONS(2269), 1,
      aux_sym_number_token1,
    ACTIONS(2271), 1,
      aux_sym_number_token2,
    ACTIONS(2275), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2289), 1,
      sym__brace_start,
    ACTIONS(8064), 1,
      anon_sym_LPAREN,
    ACTIONS(8066), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8070), 1,
      anon_sym_DQUOTE,
    ACTIONS(8074), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8076), 1,
      anon_sym_BQUOTE,
    ACTIONS(8078), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8082), 1,
      sym__comment_word,
    ACTIONS(8084), 1,
      sym__empty_value,
    ACTIONS(8254), 1,
      sym_word,
    ACTIONS(8256), 1,
      sym__special_character,
    ACTIONS(8260), 1,
      sym_test_operator,
    STATE(1887), 1,
      aux_sym__literal_repeat1,
    ACTIONS(8062), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8080), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(8258), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2158), 2,
      sym_concatenation,
      sym_array,
    STATE(1470), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [190064] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1251), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1253), 23,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [190106] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1263), 4,
      sym_file_descriptor,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1261), 30,
      anon_sym_LPAREN_LPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [190148] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1296), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1298), 23,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [190190] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1292), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1294), 23,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [190232] = 23,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8262), 1,
      sym_word,
    ACTIONS(8266), 1,
      anon_sym_LPAREN,
    ACTIONS(8268), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8270), 1,
      anon_sym_DOLLAR,
    ACTIONS(8272), 1,
      sym__special_character,
    ACTIONS(8274), 1,
      anon_sym_DQUOTE,
    ACTIONS(8278), 1,
      aux_sym_number_token1,
    ACTIONS(8280), 1,
      aux_sym_number_token2,
    ACTIONS(8282), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8284), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(8286), 1,
      anon_sym_BQUOTE,
    ACTIONS(8288), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8292), 1,
      sym__comment_word,
    ACTIONS(8294), 1,
      sym__empty_value,
    ACTIONS(8296), 1,
      sym_test_operator,
    ACTIONS(8298), 1,
      sym__brace_start,
    STATE(1618), 1,
      aux_sym__literal_repeat1,
    ACTIONS(8264), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8276), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(8290), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(1747), 2,
      sym_concatenation,
      sym_array,
    STATE(1192), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [190314] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1332), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1334), 23,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [190356] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1316), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1318), 23,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [190398] = 23,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8300), 1,
      sym_word,
    ACTIONS(8304), 1,
      anon_sym_LPAREN,
    ACTIONS(8306), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8308), 1,
      anon_sym_DOLLAR,
    ACTIONS(8310), 1,
      sym__special_character,
    ACTIONS(8312), 1,
      anon_sym_DQUOTE,
    ACTIONS(8316), 1,
      aux_sym_number_token1,
    ACTIONS(8318), 1,
      aux_sym_number_token2,
    ACTIONS(8320), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8322), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(8324), 1,
      anon_sym_BQUOTE,
    ACTIONS(8326), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8330), 1,
      sym__comment_word,
    ACTIONS(8332), 1,
      sym__empty_value,
    ACTIONS(8334), 1,
      sym_test_operator,
    ACTIONS(8336), 1,
      sym__brace_start,
    STATE(5408), 1,
      aux_sym__literal_repeat1,
    ACTIONS(8302), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8314), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(8328), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(5566), 2,
      sym_concatenation,
      sym_array,
    STATE(5253), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [190480] = 23,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2038), 1,
      anon_sym_DOLLAR,
    ACTIONS(2044), 1,
      aux_sym_number_token1,
    ACTIONS(2046), 1,
      aux_sym_number_token2,
    ACTIONS(2050), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2064), 1,
      sym__brace_start,
    ACTIONS(8338), 1,
      sym_word,
    ACTIONS(8342), 1,
      anon_sym_LPAREN,
    ACTIONS(8344), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8346), 1,
      sym__special_character,
    ACTIONS(8348), 1,
      anon_sym_DQUOTE,
    ACTIONS(8352), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8354), 1,
      anon_sym_BQUOTE,
    ACTIONS(8356), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8360), 1,
      sym__comment_word,
    ACTIONS(8362), 1,
      sym__empty_value,
    ACTIONS(8364), 1,
      sym_test_operator,
    STATE(1929), 1,
      aux_sym__literal_repeat1,
    ACTIONS(8340), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8350), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(8358), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(1947), 2,
      sym_concatenation,
      sym_array,
    STATE(1337), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [190562] = 23,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8366), 1,
      sym_word,
    ACTIONS(8370), 1,
      anon_sym_LPAREN,
    ACTIONS(8372), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8374), 1,
      anon_sym_DOLLAR,
    ACTIONS(8376), 1,
      sym__special_character,
    ACTIONS(8378), 1,
      anon_sym_DQUOTE,
    ACTIONS(8382), 1,
      aux_sym_number_token1,
    ACTIONS(8384), 1,
      aux_sym_number_token2,
    ACTIONS(8386), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8388), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(8390), 1,
      anon_sym_BQUOTE,
    ACTIONS(8392), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8396), 1,
      sym__comment_word,
    ACTIONS(8398), 1,
      sym__empty_value,
    ACTIONS(8400), 1,
      sym_test_operator,
    ACTIONS(8402), 1,
      sym__brace_start,
    STATE(4383), 1,
      aux_sym__literal_repeat1,
    ACTIONS(8368), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8380), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(8394), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(4639), 2,
      sym_concatenation,
      sym_array,
    STATE(4238), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [190644] = 23,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2263), 1,
      anon_sym_DOLLAR,
    ACTIONS(2269), 1,
      aux_sym_number_token1,
    ACTIONS(2271), 1,
      aux_sym_number_token2,
    ACTIONS(2275), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2289), 1,
      sym__brace_start,
    ACTIONS(8064), 1,
      anon_sym_LPAREN,
    ACTIONS(8066), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8070), 1,
      anon_sym_DQUOTE,
    ACTIONS(8074), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8076), 1,
      anon_sym_BQUOTE,
    ACTIONS(8078), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8082), 1,
      sym__comment_word,
    ACTIONS(8084), 1,
      sym__empty_value,
    ACTIONS(8404), 1,
      sym_word,
    ACTIONS(8406), 1,
      sym__special_character,
    ACTIONS(8410), 1,
      sym_test_operator,
    STATE(1887), 1,
      aux_sym__literal_repeat1,
    ACTIONS(8062), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8080), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(8408), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2158), 2,
      sym_concatenation,
      sym_array,
    STATE(1558), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [190726] = 23,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(7956), 1,
      anon_sym_LPAREN,
    ACTIONS(7958), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(7960), 1,
      anon_sym_DOLLAR,
    ACTIONS(7964), 1,
      anon_sym_DQUOTE,
    ACTIONS(7968), 1,
      aux_sym_number_token1,
    ACTIONS(7970), 1,
      aux_sym_number_token2,
    ACTIONS(7972), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(7974), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(7976), 1,
      anon_sym_BQUOTE,
    ACTIONS(7978), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(7982), 1,
      sym__comment_word,
    ACTIONS(7984), 1,
      sym__empty_value,
    ACTIONS(7988), 1,
      sym__brace_start,
    ACTIONS(8412), 1,
      sym_word,
    ACTIONS(8414), 1,
      sym__special_character,
    ACTIONS(8418), 1,
      sym_test_operator,
    STATE(4287), 1,
      aux_sym__literal_repeat1,
    ACTIONS(7954), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(7980), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(8416), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(4759), 2,
      sym_concatenation,
      sym_array,
    STATE(4531), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [190808] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1352), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1354), 23,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [190850] = 23,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1761), 1,
      anon_sym_DOLLAR,
    ACTIONS(1767), 1,
      aux_sym_number_token1,
    ACTIONS(1769), 1,
      aux_sym_number_token2,
    ACTIONS(1773), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1787), 1,
      sym__brace_start,
    ACTIONS(8420), 1,
      sym_word,
    ACTIONS(8424), 1,
      anon_sym_LPAREN,
    ACTIONS(8426), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8428), 1,
      sym__special_character,
    ACTIONS(8430), 1,
      anon_sym_DQUOTE,
    ACTIONS(8434), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8436), 1,
      anon_sym_BQUOTE,
    ACTIONS(8438), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8442), 1,
      sym__comment_word,
    ACTIONS(8444), 1,
      sym__empty_value,
    ACTIONS(8446), 1,
      sym_test_operator,
    STATE(1480), 1,
      aux_sym__literal_repeat1,
    ACTIONS(8422), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8432), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(8440), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(1498), 2,
      sym_concatenation,
      sym_array,
    STATE(1156), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [190932] = 23,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8176), 1,
      anon_sym_LPAREN,
    ACTIONS(8178), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8180), 1,
      anon_sym_DOLLAR,
    ACTIONS(8184), 1,
      anon_sym_DQUOTE,
    ACTIONS(8188), 1,
      aux_sym_number_token1,
    ACTIONS(8190), 1,
      aux_sym_number_token2,
    ACTIONS(8192), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8194), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(8196), 1,
      anon_sym_BQUOTE,
    ACTIONS(8198), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8202), 1,
      sym__comment_word,
    ACTIONS(8204), 1,
      sym__empty_value,
    ACTIONS(8208), 1,
      sym__brace_start,
    ACTIONS(8448), 1,
      sym_word,
    ACTIONS(8450), 1,
      sym__special_character,
    ACTIONS(8454), 1,
      sym_test_operator,
    STATE(1742), 1,
      aux_sym__literal_repeat1,
    ACTIONS(8174), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8200), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(8452), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1959), 2,
      sym_concatenation,
      sym_array,
    STATE(1512), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [191014] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1348), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1350), 23,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [191056] = 23,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4151), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(4153), 1,
      anon_sym_DOLLAR,
    ACTIONS(4155), 1,
      sym__special_character,
    ACTIONS(4157), 1,
      anon_sym_DQUOTE,
    ACTIONS(4161), 1,
      aux_sym_number_token1,
    ACTIONS(4163), 1,
      aux_sym_number_token2,
    ACTIONS(4165), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(4167), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(4169), 1,
      anon_sym_BQUOTE,
    ACTIONS(4171), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(4181), 1,
      sym__brace_start,
    ACTIONS(8456), 1,
      sym_word,
    ACTIONS(8458), 1,
      anon_sym_LPAREN,
    ACTIONS(8462), 1,
      sym__comment_word,
    ACTIONS(8464), 1,
      sym__empty_value,
    ACTIONS(8466), 1,
      sym_test_operator,
    STATE(2714), 1,
      aux_sym__literal_repeat1,
    ACTIONS(4149), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(4173), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(8460), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2878), 2,
      sym_concatenation,
      sym_array,
    STATE(2406), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [191138] = 23,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8468), 1,
      sym_word,
    ACTIONS(8472), 1,
      anon_sym_LPAREN,
    ACTIONS(8474), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8476), 1,
      anon_sym_DOLLAR,
    ACTIONS(8478), 1,
      sym__special_character,
    ACTIONS(8480), 1,
      anon_sym_DQUOTE,
    ACTIONS(8484), 1,
      aux_sym_number_token1,
    ACTIONS(8486), 1,
      aux_sym_number_token2,
    ACTIONS(8488), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8490), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(8492), 1,
      anon_sym_BQUOTE,
    ACTIONS(8494), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8498), 1,
      sym__comment_word,
    ACTIONS(8500), 1,
      sym__empty_value,
    ACTIONS(8502), 1,
      sym_test_operator,
    ACTIONS(8504), 1,
      sym__brace_start,
    STATE(1900), 1,
      aux_sym__literal_repeat1,
    ACTIONS(8470), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8482), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(8496), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(2103), 2,
      sym_concatenation,
      sym_array,
    STATE(1391), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [191220] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1340), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1342), 23,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [191262] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1328), 11,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1330), 23,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [191304] = 23,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8506), 1,
      sym_word,
    ACTIONS(8510), 1,
      anon_sym_LPAREN,
    ACTIONS(8512), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8514), 1,
      anon_sym_DOLLAR,
    ACTIONS(8516), 1,
      sym__special_character,
    ACTIONS(8518), 1,
      anon_sym_DQUOTE,
    ACTIONS(8522), 1,
      aux_sym_number_token1,
    ACTIONS(8524), 1,
      aux_sym_number_token2,
    ACTIONS(8526), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8528), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(8530), 1,
      anon_sym_BQUOTE,
    ACTIONS(8532), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8536), 1,
      sym__comment_word,
    ACTIONS(8538), 1,
      sym__empty_value,
    ACTIONS(8540), 1,
      sym_test_operator,
    ACTIONS(8542), 1,
      sym__brace_start,
    STATE(2678), 1,
      aux_sym__literal_repeat1,
    ACTIONS(8508), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8520), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(8534), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(2934), 2,
      sym_concatenation,
      sym_array,
    STATE(2440), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [191386] = 19,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8030), 1,
      anon_sym_SLASH,
    ACTIONS(8032), 1,
      anon_sym_PERCENT,
    ACTIONS(8034), 1,
      anon_sym_COLON,
    ACTIONS(8544), 1,
      anon_sym_RBRACE3,
    ACTIONS(8546), 1,
      anon_sym_AT,
    STATE(6465), 1,
      aux_sym__expansion_body_repeat1,
    STATE(7232), 1,
      sym__expansion_operator,
    STATE(7233), 1,
      sym__expansion_max_length,
    STATE(7234), 1,
      sym__expansion_regex_removal,
    STATE(7235), 1,
      sym__expansion_regex_replacement,
    STATE(7236), 1,
      sym__expansion_regex,
    STATE(7237), 1,
      sym__expansion_expression,
    ACTIONS(8028), 2,
      anon_sym_COMMA,
      anon_sym_CARET,
    ACTIONS(8048), 2,
      anon_sym_COMMA_COMMA,
      anon_sym_CARET_CARET,
    ACTIONS(8036), 3,
      sym__immediate_double_hash,
      anon_sym_POUND,
      anon_sym_PERCENT_PERCENT,
    ACTIONS(8046), 3,
      anon_sym_SLASH_SLASH,
      anon_sym_SLASH_POUND,
      anon_sym_SLASH_PERCENT,
    ACTIONS(8050), 3,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
    ACTIONS(8044), 8,
      anon_sym_EQ2,
      anon_sym_COLON_EQ,
      anon_sym_DASH3,
      anon_sym_COLON_DASH,
      anon_sym_PLUS3,
      anon_sym_COLON_PLUS,
      anon_sym_QMARK2,
      anon_sym_COLON_QMARK,
  [191459] = 19,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8030), 1,
      anon_sym_SLASH,
    ACTIONS(8032), 1,
      anon_sym_PERCENT,
    ACTIONS(8034), 1,
      anon_sym_COLON,
    ACTIONS(8546), 1,
      anon_sym_AT,
    ACTIONS(8548), 1,
      anon_sym_RBRACE3,
    STATE(6477), 1,
      aux_sym__expansion_body_repeat1,
    STATE(6876), 1,
      sym__expansion_max_length,
    STATE(7218), 1,
      sym__expansion_operator,
    STATE(7220), 1,
      sym__expansion_regex_removal,
    STATE(7221), 1,
      sym__expansion_regex_replacement,
    STATE(7222), 1,
      sym__expansion_regex,
    STATE(7223), 1,
      sym__expansion_expression,
    ACTIONS(8028), 2,
      anon_sym_COMMA,
      anon_sym_CARET,
    ACTIONS(8048), 2,
      anon_sym_COMMA_COMMA,
      anon_sym_CARET_CARET,
    ACTIONS(8036), 3,
      sym__immediate_double_hash,
      anon_sym_POUND,
      anon_sym_PERCENT_PERCENT,
    ACTIONS(8046), 3,
      anon_sym_SLASH_SLASH,
      anon_sym_SLASH_POUND,
      anon_sym_SLASH_PERCENT,
    ACTIONS(8050), 3,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
    ACTIONS(8044), 8,
      anon_sym_EQ2,
      anon_sym_COLON_EQ,
      anon_sym_DASH3,
      anon_sym_COLON_DASH,
      anon_sym_PLUS3,
      anon_sym_COLON_PLUS,
      anon_sym_QMARK2,
      anon_sym_COLON_QMARK,
  [191532] = 19,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8030), 1,
      anon_sym_SLASH,
    ACTIONS(8032), 1,
      anon_sym_PERCENT,
    ACTIONS(8034), 1,
      anon_sym_COLON,
    ACTIONS(8038), 1,
      anon_sym_RBRACE3,
    ACTIONS(8546), 1,
      anon_sym_AT,
    STATE(6390), 1,
      aux_sym__expansion_body_repeat1,
    STATE(6959), 1,
      sym__expansion_operator,
    STATE(6960), 1,
      sym__expansion_max_length,
    STATE(6964), 1,
      sym__expansion_regex_removal,
    STATE(6965), 1,
      sym__expansion_regex_replacement,
    STATE(6966), 1,
      sym__expansion_regex,
    STATE(6967), 1,
      sym__expansion_expression,
    ACTIONS(8028), 2,
      anon_sym_COMMA,
      anon_sym_CARET,
    ACTIONS(8048), 2,
      anon_sym_COMMA_COMMA,
      anon_sym_CARET_CARET,
    ACTIONS(8036), 3,
      sym__immediate_double_hash,
      anon_sym_POUND,
      anon_sym_PERCENT_PERCENT,
    ACTIONS(8046), 3,
      anon_sym_SLASH_SLASH,
      anon_sym_SLASH_POUND,
      anon_sym_SLASH_PERCENT,
    ACTIONS(8050), 3,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
    ACTIONS(8044), 8,
      anon_sym_EQ2,
      anon_sym_COLON_EQ,
      anon_sym_DASH3,
      anon_sym_COLON_DASH,
      anon_sym_PLUS3,
      anon_sym_COLON_PLUS,
      anon_sym_QMARK2,
      anon_sym_COLON_QMARK,
  [191605] = 19,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8030), 1,
      anon_sym_SLASH,
    ACTIONS(8032), 1,
      anon_sym_PERCENT,
    ACTIONS(8034), 1,
      anon_sym_COLON,
    ACTIONS(8546), 1,
      anon_sym_AT,
    ACTIONS(8550), 1,
      anon_sym_RBRACE3,
    STATE(6402), 1,
      aux_sym__expansion_body_repeat1,
    STATE(6985), 1,
      sym__expansion_operator,
    STATE(6986), 1,
      sym__expansion_max_length,
    STATE(6987), 1,
      sym__expansion_regex_removal,
    STATE(6988), 1,
      sym__expansion_regex_replacement,
    STATE(6992), 1,
      sym__expansion_regex,
    STATE(6993), 1,
      sym__expansion_expression,
    ACTIONS(8028), 2,
      anon_sym_COMMA,
      anon_sym_CARET,
    ACTIONS(8048), 2,
      anon_sym_COMMA_COMMA,
      anon_sym_CARET_CARET,
    ACTIONS(8036), 3,
      sym__immediate_double_hash,
      anon_sym_POUND,
      anon_sym_PERCENT_PERCENT,
    ACTIONS(8046), 3,
      anon_sym_SLASH_SLASH,
      anon_sym_SLASH_POUND,
      anon_sym_SLASH_PERCENT,
    ACTIONS(8050), 3,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
    ACTIONS(8044), 8,
      anon_sym_EQ2,
      anon_sym_COLON_EQ,
      anon_sym_DASH3,
      anon_sym_COLON_DASH,
      anon_sym_PLUS3,
      anon_sym_COLON_PLUS,
      anon_sym_QMARK2,
      anon_sym_COLON_QMARK,
  [191678] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1235), 1,
      sym_file_descriptor,
    ACTIONS(3791), 1,
      anon_sym_DQUOTE,
    ACTIONS(7444), 1,
      sym_variable_name,
    STATE(4389), 1,
      sym_string,
    ACTIONS(7442), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(7440), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1227), 18,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
      anon_sym_LT_LT_LT,
  [191729] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(8552), 1,
      sym__special_character,
    STATE(3544), 1,
      aux_sym__literal_repeat1,
    ACTIONS(5067), 10,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(5069), 21,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [191774] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1241), 1,
      sym_file_descriptor,
    ACTIONS(8556), 1,
      anon_sym_DQUOTE,
    ACTIONS(8560), 1,
      sym_variable_name,
    STATE(5364), 1,
      sym_string,
    ACTIONS(8558), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(8554), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1239), 18,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [191825] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1235), 1,
      sym_file_descriptor,
    ACTIONS(8556), 1,
      anon_sym_DQUOTE,
    ACTIONS(8560), 1,
      sym_variable_name,
    STATE(5364), 1,
      sym_string,
    ACTIONS(8558), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(8554), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1227), 18,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [191876] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(8552), 1,
      sym__special_character,
    STATE(3544), 1,
      aux_sym__literal_repeat1,
    ACTIONS(4564), 10,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(4566), 21,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [191921] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1241), 1,
      sym_file_descriptor,
    ACTIONS(3791), 1,
      anon_sym_DQUOTE,
    ACTIONS(7444), 1,
      sym_variable_name,
    STATE(4389), 1,
      sym_string,
    ACTIONS(7442), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(7440), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1239), 18,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
      anon_sym_LT_LT_LT,
  [191972] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(8552), 1,
      sym__special_character,
    STATE(3544), 1,
      aux_sym__literal_repeat1,
    ACTIONS(5695), 10,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(5697), 21,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [192017] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(8552), 1,
      sym__special_character,
    STATE(3544), 1,
      aux_sym__literal_repeat1,
    ACTIONS(4467), 10,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(4469), 21,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [192062] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(8552), 1,
      sym__special_character,
    STATE(3544), 1,
      aux_sym__literal_repeat1,
    ACTIONS(5393), 10,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(5395), 21,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [192107] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(8562), 1,
      sym__special_character,
    STATE(3544), 1,
      aux_sym__literal_repeat1,
    ACTIONS(1362), 10,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(1364), 21,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [192152] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(8552), 1,
      sym__special_character,
    STATE(3544), 1,
      aux_sym__literal_repeat1,
    ACTIONS(5721), 10,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(5723), 21,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [192197] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(8565), 1,
      sym_word,
    ACTIONS(8569), 1,
      anon_sym_RPAREN,
    ACTIONS(8571), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8573), 1,
      anon_sym_DOLLAR,
    ACTIONS(8575), 1,
      sym__special_character,
    ACTIONS(8577), 1,
      anon_sym_DQUOTE,
    ACTIONS(8581), 1,
      aux_sym_number_token1,
    ACTIONS(8583), 1,
      aux_sym_number_token2,
    ACTIONS(8585), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8587), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(8589), 1,
      anon_sym_BQUOTE,
    ACTIONS(8591), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8595), 1,
      sym_test_operator,
    ACTIONS(8597), 1,
      sym__brace_start,
    STATE(5365), 1,
      aux_sym__literal_repeat1,
    ACTIONS(8567), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8579), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(8593), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3604), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    STATE(5045), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [192273] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(4552), 10,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(4554), 22,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [192313] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(8565), 1,
      sym_word,
    ACTIONS(8571), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8573), 1,
      anon_sym_DOLLAR,
    ACTIONS(8575), 1,
      sym__special_character,
    ACTIONS(8577), 1,
      anon_sym_DQUOTE,
    ACTIONS(8581), 1,
      aux_sym_number_token1,
    ACTIONS(8583), 1,
      aux_sym_number_token2,
    ACTIONS(8585), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8587), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(8589), 1,
      anon_sym_BQUOTE,
    ACTIONS(8591), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8595), 1,
      sym_test_operator,
    ACTIONS(8597), 1,
      sym__brace_start,
    ACTIONS(8599), 1,
      anon_sym_RPAREN,
    STATE(5365), 1,
      aux_sym__literal_repeat1,
    ACTIONS(8567), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8579), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(8593), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3549), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    STATE(5045), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [192389] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(8565), 1,
      sym_word,
    ACTIONS(8571), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8573), 1,
      anon_sym_DOLLAR,
    ACTIONS(8575), 1,
      sym__special_character,
    ACTIONS(8577), 1,
      anon_sym_DQUOTE,
    ACTIONS(8581), 1,
      aux_sym_number_token1,
    ACTIONS(8583), 1,
      aux_sym_number_token2,
    ACTIONS(8585), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8587), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(8589), 1,
      anon_sym_BQUOTE,
    ACTIONS(8591), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8595), 1,
      sym_test_operator,
    ACTIONS(8597), 1,
      sym__brace_start,
    ACTIONS(8601), 1,
      anon_sym_RPAREN,
    STATE(5365), 1,
      aux_sym__literal_repeat1,
    ACTIONS(8567), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8579), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(8593), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3604), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    STATE(5045), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [192465] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(8565), 1,
      sym_word,
    ACTIONS(8571), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8573), 1,
      anon_sym_DOLLAR,
    ACTIONS(8575), 1,
      sym__special_character,
    ACTIONS(8577), 1,
      anon_sym_DQUOTE,
    ACTIONS(8581), 1,
      aux_sym_number_token1,
    ACTIONS(8583), 1,
      aux_sym_number_token2,
    ACTIONS(8585), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8587), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(8589), 1,
      anon_sym_BQUOTE,
    ACTIONS(8591), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8595), 1,
      sym_test_operator,
    ACTIONS(8597), 1,
      sym__brace_start,
    ACTIONS(8603), 1,
      anon_sym_RPAREN,
    STATE(5365), 1,
      aux_sym__literal_repeat1,
    ACTIONS(8567), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8579), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(8593), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3604), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    STATE(5045), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [192541] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(8565), 1,
      sym_word,
    ACTIONS(8571), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8573), 1,
      anon_sym_DOLLAR,
    ACTIONS(8575), 1,
      sym__special_character,
    ACTIONS(8577), 1,
      anon_sym_DQUOTE,
    ACTIONS(8581), 1,
      aux_sym_number_token1,
    ACTIONS(8583), 1,
      aux_sym_number_token2,
    ACTIONS(8585), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8587), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(8589), 1,
      anon_sym_BQUOTE,
    ACTIONS(8591), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8595), 1,
      sym_test_operator,
    ACTIONS(8597), 1,
      sym__brace_start,
    ACTIONS(8605), 1,
      anon_sym_RPAREN,
    STATE(5365), 1,
      aux_sym__literal_repeat1,
    ACTIONS(8567), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8579), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(8593), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3550), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    STATE(5045), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [192617] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1261), 10,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(1263), 22,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [192657] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(8565), 1,
      sym_word,
    ACTIONS(8571), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8573), 1,
      anon_sym_DOLLAR,
    ACTIONS(8575), 1,
      sym__special_character,
    ACTIONS(8577), 1,
      anon_sym_DQUOTE,
    ACTIONS(8581), 1,
      aux_sym_number_token1,
    ACTIONS(8583), 1,
      aux_sym_number_token2,
    ACTIONS(8585), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8587), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(8589), 1,
      anon_sym_BQUOTE,
    ACTIONS(8591), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8595), 1,
      sym_test_operator,
    ACTIONS(8597), 1,
      sym__brace_start,
    ACTIONS(8607), 1,
      anon_sym_RPAREN,
    STATE(5365), 1,
      aux_sym__literal_repeat1,
    ACTIONS(8567), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8579), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(8593), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3554), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    STATE(5045), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [192733] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(8565), 1,
      sym_word,
    ACTIONS(8571), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8573), 1,
      anon_sym_DOLLAR,
    ACTIONS(8575), 1,
      sym__special_character,
    ACTIONS(8577), 1,
      anon_sym_DQUOTE,
    ACTIONS(8581), 1,
      aux_sym_number_token1,
    ACTIONS(8583), 1,
      aux_sym_number_token2,
    ACTIONS(8585), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8587), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(8589), 1,
      anon_sym_BQUOTE,
    ACTIONS(8591), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8595), 1,
      sym_test_operator,
    ACTIONS(8597), 1,
      sym__brace_start,
    ACTIONS(8609), 1,
      anon_sym_RPAREN,
    STATE(5365), 1,
      aux_sym__literal_repeat1,
    ACTIONS(8567), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8579), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(8593), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3604), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    STATE(5045), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [192809] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6488), 1,
      anon_sym_DOLLAR,
    ACTIONS(6494), 1,
      aux_sym_number_token1,
    ACTIONS(6496), 1,
      aux_sym_number_token2,
    ACTIONS(6500), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(6510), 1,
      sym_extglob_pattern,
    ACTIONS(6512), 1,
      sym__brace_start,
    ACTIONS(7226), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(7228), 1,
      sym__special_character,
    ACTIONS(7230), 1,
      anon_sym_DQUOTE,
    ACTIONS(7234), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(7236), 1,
      anon_sym_BQUOTE,
    ACTIONS(7238), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8611), 1,
      sym_word,
    ACTIONS(8615), 1,
      sym_test_operator,
    STATE(6405), 1,
      aux_sym__literal_repeat1,
    ACTIONS(7224), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(7240), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(8613), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(6725), 2,
      sym_concatenation,
      sym__extglob_blob,
    STATE(6338), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [192885] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(2094), 10,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(2096), 22,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [192925] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(8565), 1,
      sym_word,
    ACTIONS(8571), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8573), 1,
      anon_sym_DOLLAR,
    ACTIONS(8575), 1,
      sym__special_character,
    ACTIONS(8577), 1,
      anon_sym_DQUOTE,
    ACTIONS(8581), 1,
      aux_sym_number_token1,
    ACTIONS(8583), 1,
      aux_sym_number_token2,
    ACTIONS(8585), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8587), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(8589), 1,
      anon_sym_BQUOTE,
    ACTIONS(8591), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8595), 1,
      sym_test_operator,
    ACTIONS(8597), 1,
      sym__brace_start,
    ACTIONS(8617), 1,
      anon_sym_RPAREN,
    STATE(5365), 1,
      aux_sym__literal_repeat1,
    ACTIONS(8567), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8579), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(8593), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3546), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    STATE(5045), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [193001] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(8565), 1,
      sym_word,
    ACTIONS(8571), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8573), 1,
      anon_sym_DOLLAR,
    ACTIONS(8575), 1,
      sym__special_character,
    ACTIONS(8577), 1,
      anon_sym_DQUOTE,
    ACTIONS(8581), 1,
      aux_sym_number_token1,
    ACTIONS(8583), 1,
      aux_sym_number_token2,
    ACTIONS(8585), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8587), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(8589), 1,
      anon_sym_BQUOTE,
    ACTIONS(8591), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8595), 1,
      sym_test_operator,
    ACTIONS(8597), 1,
      sym__brace_start,
    ACTIONS(8619), 1,
      anon_sym_RPAREN,
    STATE(5365), 1,
      aux_sym__literal_repeat1,
    ACTIONS(8567), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8579), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(8593), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3613), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    STATE(5045), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [193077] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8623), 1,
      anon_sym_DQUOTE,
    ACTIONS(8627), 1,
      sym_variable_name,
    STATE(5204), 1,
      sym_string,
    ACTIONS(1235), 2,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(8625), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(8621), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1227), 16,
      anon_sym_LPAREN_LPAREN,
      anon_sym_RPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [193127] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8623), 1,
      anon_sym_DQUOTE,
    ACTIONS(8627), 1,
      sym_variable_name,
    STATE(5204), 1,
      sym_string,
    ACTIONS(1241), 2,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(8625), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(8621), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1239), 16,
      anon_sym_LPAREN_LPAREN,
      anon_sym_RPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [193177] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(8565), 1,
      sym_word,
    ACTIONS(8571), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8573), 1,
      anon_sym_DOLLAR,
    ACTIONS(8575), 1,
      sym__special_character,
    ACTIONS(8577), 1,
      anon_sym_DQUOTE,
    ACTIONS(8581), 1,
      aux_sym_number_token1,
    ACTIONS(8583), 1,
      aux_sym_number_token2,
    ACTIONS(8585), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8587), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(8589), 1,
      anon_sym_BQUOTE,
    ACTIONS(8591), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8595), 1,
      sym_test_operator,
    ACTIONS(8597), 1,
      sym__brace_start,
    ACTIONS(8629), 1,
      anon_sym_RPAREN,
    STATE(5365), 1,
      aux_sym__literal_repeat1,
    ACTIONS(8567), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8579), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(8593), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3604), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    STATE(5045), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [193253] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3735), 1,
      anon_sym_DQUOTE,
    ACTIONS(7757), 1,
      sym_variable_name,
    STATE(4808), 1,
      sym_string,
    ACTIONS(1241), 2,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(7755), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(7753), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1239), 16,
      anon_sym_LPAREN_LPAREN,
      aux_sym_heredoc_redirect_token1,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [193303] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(8565), 1,
      sym_word,
    ACTIONS(8571), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8573), 1,
      anon_sym_DOLLAR,
    ACTIONS(8575), 1,
      sym__special_character,
    ACTIONS(8577), 1,
      anon_sym_DQUOTE,
    ACTIONS(8581), 1,
      aux_sym_number_token1,
    ACTIONS(8583), 1,
      aux_sym_number_token2,
    ACTIONS(8585), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8587), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(8589), 1,
      anon_sym_BQUOTE,
    ACTIONS(8591), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8595), 1,
      sym_test_operator,
    ACTIONS(8597), 1,
      sym__brace_start,
    ACTIONS(8631), 1,
      anon_sym_RPAREN,
    STATE(5365), 1,
      aux_sym__literal_repeat1,
    ACTIONS(8567), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8579), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(8593), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3564), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    STATE(5045), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [193379] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(8565), 1,
      sym_word,
    ACTIONS(8571), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8573), 1,
      anon_sym_DOLLAR,
    ACTIONS(8575), 1,
      sym__special_character,
    ACTIONS(8577), 1,
      anon_sym_DQUOTE,
    ACTIONS(8581), 1,
      aux_sym_number_token1,
    ACTIONS(8583), 1,
      aux_sym_number_token2,
    ACTIONS(8585), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8587), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(8589), 1,
      anon_sym_BQUOTE,
    ACTIONS(8591), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8595), 1,
      sym_test_operator,
    ACTIONS(8597), 1,
      sym__brace_start,
    ACTIONS(8633), 1,
      anon_sym_RPAREN,
    STATE(5365), 1,
      aux_sym__literal_repeat1,
    ACTIONS(8567), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8579), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(8593), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3604), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    STATE(5045), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [193455] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3729), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(3731), 1,
      anon_sym_DOLLAR,
    ACTIONS(3733), 1,
      sym__special_character,
    ACTIONS(3735), 1,
      anon_sym_DQUOTE,
    ACTIONS(3737), 1,
      aux_sym_number_token1,
    ACTIONS(3739), 1,
      aux_sym_number_token2,
    ACTIONS(3741), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(3743), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3745), 1,
      anon_sym_BQUOTE,
    ACTIONS(3747), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3753), 1,
      sym_test_operator,
    ACTIONS(3755), 1,
      sym__brace_start,
    ACTIONS(8635), 1,
      aux_sym_heredoc_redirect_token1,
    STATE(3571), 1,
      aux_sym__heredoc_command,
    STATE(5351), 1,
      aux_sym__literal_repeat1,
    STATE(5456), 1,
      sym_concatenation,
    ACTIONS(3715), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(3749), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(3713), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(5062), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [193531] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(8565), 1,
      sym_word,
    ACTIONS(8571), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8573), 1,
      anon_sym_DOLLAR,
    ACTIONS(8575), 1,
      sym__special_character,
    ACTIONS(8577), 1,
      anon_sym_DQUOTE,
    ACTIONS(8581), 1,
      aux_sym_number_token1,
    ACTIONS(8583), 1,
      aux_sym_number_token2,
    ACTIONS(8585), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8587), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(8589), 1,
      anon_sym_BQUOTE,
    ACTIONS(8591), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8595), 1,
      sym_test_operator,
    ACTIONS(8597), 1,
      sym__brace_start,
    ACTIONS(8637), 1,
      anon_sym_RPAREN,
    STATE(5365), 1,
      aux_sym__literal_repeat1,
    ACTIONS(8567), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8579), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(8593), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3568), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    STATE(5045), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [193607] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3729), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(3731), 1,
      anon_sym_DOLLAR,
    ACTIONS(3733), 1,
      sym__special_character,
    ACTIONS(3735), 1,
      anon_sym_DQUOTE,
    ACTIONS(3737), 1,
      aux_sym_number_token1,
    ACTIONS(3739), 1,
      aux_sym_number_token2,
    ACTIONS(3741), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(3743), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3745), 1,
      anon_sym_BQUOTE,
    ACTIONS(3747), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3753), 1,
      sym_test_operator,
    ACTIONS(3755), 1,
      sym__brace_start,
    ACTIONS(8639), 1,
      aux_sym_heredoc_redirect_token1,
    STATE(3571), 1,
      aux_sym__heredoc_command,
    STATE(5351), 1,
      aux_sym__literal_repeat1,
    STATE(5456), 1,
      sym_concatenation,
    ACTIONS(3715), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(3749), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(3713), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(5062), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [193683] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(8565), 1,
      sym_word,
    ACTIONS(8571), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8573), 1,
      anon_sym_DOLLAR,
    ACTIONS(8575), 1,
      sym__special_character,
    ACTIONS(8577), 1,
      anon_sym_DQUOTE,
    ACTIONS(8581), 1,
      aux_sym_number_token1,
    ACTIONS(8583), 1,
      aux_sym_number_token2,
    ACTIONS(8585), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8587), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(8589), 1,
      anon_sym_BQUOTE,
    ACTIONS(8591), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8595), 1,
      sym_test_operator,
    ACTIONS(8597), 1,
      sym__brace_start,
    ACTIONS(8641), 1,
      anon_sym_RPAREN,
    STATE(5365), 1,
      aux_sym__literal_repeat1,
    ACTIONS(8567), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8579), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(8593), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3604), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    STATE(5045), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [193759] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(4370), 10,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(4372), 22,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [193799] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3729), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(3731), 1,
      anon_sym_DOLLAR,
    ACTIONS(3733), 1,
      sym__special_character,
    ACTIONS(3735), 1,
      anon_sym_DQUOTE,
    ACTIONS(3737), 1,
      aux_sym_number_token1,
    ACTIONS(3739), 1,
      aux_sym_number_token2,
    ACTIONS(3741), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(3743), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3745), 1,
      anon_sym_BQUOTE,
    ACTIONS(3747), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3753), 1,
      sym_test_operator,
    ACTIONS(3755), 1,
      sym__brace_start,
    ACTIONS(8643), 1,
      aux_sym_heredoc_redirect_token1,
    STATE(3571), 1,
      aux_sym__heredoc_command,
    STATE(5351), 1,
      aux_sym__literal_repeat1,
    STATE(5456), 1,
      sym_concatenation,
    ACTIONS(3715), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(3749), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(3713), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(5062), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [193875] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8651), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(8653), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8656), 1,
      anon_sym_DOLLAR,
    ACTIONS(8659), 1,
      sym__special_character,
    ACTIONS(8662), 1,
      anon_sym_DQUOTE,
    ACTIONS(8665), 1,
      aux_sym_number_token1,
    ACTIONS(8668), 1,
      aux_sym_number_token2,
    ACTIONS(8671), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8674), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(8677), 1,
      anon_sym_BQUOTE,
    ACTIONS(8680), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8686), 1,
      sym_test_operator,
    ACTIONS(8689), 1,
      sym__brace_start,
    STATE(3571), 1,
      aux_sym__heredoc_command,
    STATE(5351), 1,
      aux_sym__literal_repeat1,
    STATE(5456), 1,
      sym_concatenation,
    ACTIONS(8648), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8683), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(8645), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(5062), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [193951] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1235), 1,
      sym_file_descriptor,
    ACTIONS(8694), 1,
      anon_sym_DQUOTE,
    ACTIONS(8698), 1,
      sym_variable_name,
    STATE(5487), 1,
      sym_string,
    ACTIONS(8696), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(8692), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1227), 17,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [194001] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(2072), 10,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(2074), 22,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [194041] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6488), 1,
      anon_sym_DOLLAR,
    ACTIONS(6494), 1,
      aux_sym_number_token1,
    ACTIONS(6496), 1,
      aux_sym_number_token2,
    ACTIONS(6500), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(6510), 1,
      sym_extglob_pattern,
    ACTIONS(6512), 1,
      sym__brace_start,
    ACTIONS(7226), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(7228), 1,
      sym__special_character,
    ACTIONS(7230), 1,
      anon_sym_DQUOTE,
    ACTIONS(7234), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(7236), 1,
      anon_sym_BQUOTE,
    ACTIONS(7238), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8700), 1,
      sym_word,
    ACTIONS(8704), 1,
      sym_test_operator,
    STATE(6543), 1,
      aux_sym__literal_repeat1,
    ACTIONS(7224), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(7240), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(8702), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(6775), 2,
      sym_concatenation,
      sym__extglob_blob,
    STATE(6420), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [194117] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(8565), 1,
      sym_word,
    ACTIONS(8571), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8573), 1,
      anon_sym_DOLLAR,
    ACTIONS(8575), 1,
      sym__special_character,
    ACTIONS(8577), 1,
      anon_sym_DQUOTE,
    ACTIONS(8581), 1,
      aux_sym_number_token1,
    ACTIONS(8583), 1,
      aux_sym_number_token2,
    ACTIONS(8585), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8587), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(8589), 1,
      anon_sym_BQUOTE,
    ACTIONS(8591), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8595), 1,
      sym_test_operator,
    ACTIONS(8597), 1,
      sym__brace_start,
    ACTIONS(8706), 1,
      anon_sym_RPAREN,
    STATE(5365), 1,
      aux_sym__literal_repeat1,
    ACTIONS(8567), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8579), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(8593), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3561), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    STATE(5045), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [194193] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(8565), 1,
      sym_word,
    ACTIONS(8571), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8573), 1,
      anon_sym_DOLLAR,
    ACTIONS(8575), 1,
      sym__special_character,
    ACTIONS(8577), 1,
      anon_sym_DQUOTE,
    ACTIONS(8581), 1,
      aux_sym_number_token1,
    ACTIONS(8583), 1,
      aux_sym_number_token2,
    ACTIONS(8585), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8587), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(8589), 1,
      anon_sym_BQUOTE,
    ACTIONS(8591), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8595), 1,
      sym_test_operator,
    ACTIONS(8597), 1,
      sym__brace_start,
    ACTIONS(8708), 1,
      anon_sym_RPAREN,
    STATE(5365), 1,
      aux_sym__literal_repeat1,
    ACTIONS(8567), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8579), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(8593), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3604), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    STATE(5045), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [194269] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(8565), 1,
      sym_word,
    ACTIONS(8571), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8573), 1,
      anon_sym_DOLLAR,
    ACTIONS(8575), 1,
      sym__special_character,
    ACTIONS(8577), 1,
      anon_sym_DQUOTE,
    ACTIONS(8581), 1,
      aux_sym_number_token1,
    ACTIONS(8583), 1,
      aux_sym_number_token2,
    ACTIONS(8585), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8587), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(8589), 1,
      anon_sym_BQUOTE,
    ACTIONS(8591), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8595), 1,
      sym_test_operator,
    ACTIONS(8597), 1,
      sym__brace_start,
    ACTIONS(8710), 1,
      anon_sym_RPAREN,
    STATE(5365), 1,
      aux_sym__literal_repeat1,
    ACTIONS(8567), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8579), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(8593), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3604), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    STATE(5045), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [194345] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(5766), 10,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(5768), 22,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [194385] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1235), 1,
      sym_file_descriptor,
    ACTIONS(7673), 1,
      anon_sym_DQUOTE,
    ACTIONS(7677), 1,
      sym_variable_name,
    STATE(4393), 1,
      sym_string,
    ACTIONS(7675), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(7671), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1227), 17,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
  [194435] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1241), 1,
      sym_file_descriptor,
    ACTIONS(8714), 1,
      anon_sym_DQUOTE,
    ACTIONS(8718), 1,
      sym_variable_name,
    STATE(5444), 1,
      sym_string,
    ACTIONS(8716), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(8712), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1239), 17,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [194485] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(5819), 10,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(5831), 22,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [194525] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1235), 1,
      sym_file_descriptor,
    ACTIONS(8714), 1,
      anon_sym_DQUOTE,
    ACTIONS(8718), 1,
      sym_variable_name,
    STATE(5444), 1,
      sym_string,
    ACTIONS(8716), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(8712), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1227), 17,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [194575] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3729), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(3731), 1,
      anon_sym_DOLLAR,
    ACTIONS(3733), 1,
      sym__special_character,
    ACTIONS(3735), 1,
      anon_sym_DQUOTE,
    ACTIONS(3737), 1,
      aux_sym_number_token1,
    ACTIONS(3739), 1,
      aux_sym_number_token2,
    ACTIONS(3741), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(3743), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3745), 1,
      anon_sym_BQUOTE,
    ACTIONS(3747), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3753), 1,
      sym_test_operator,
    ACTIONS(3755), 1,
      sym__brace_start,
    ACTIONS(8720), 1,
      aux_sym_heredoc_redirect_token1,
    STATE(3571), 1,
      aux_sym__heredoc_command,
    STATE(5351), 1,
      aux_sym__literal_repeat1,
    STATE(5456), 1,
      sym_concatenation,
    ACTIONS(3715), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(3749), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(3713), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(5062), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [194651] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3735), 1,
      anon_sym_DQUOTE,
    ACTIONS(7757), 1,
      sym_variable_name,
    STATE(4808), 1,
      sym_string,
    ACTIONS(1235), 2,
      sym_test_operator,
      sym__brace_start,
    ACTIONS(7755), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(7753), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1227), 16,
      anon_sym_LPAREN_LPAREN,
      aux_sym_heredoc_redirect_token1,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [194701] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3729), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(3731), 1,
      anon_sym_DOLLAR,
    ACTIONS(3733), 1,
      sym__special_character,
    ACTIONS(3735), 1,
      anon_sym_DQUOTE,
    ACTIONS(3737), 1,
      aux_sym_number_token1,
    ACTIONS(3739), 1,
      aux_sym_number_token2,
    ACTIONS(3741), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(3743), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3745), 1,
      anon_sym_BQUOTE,
    ACTIONS(3747), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3753), 1,
      sym_test_operator,
    ACTIONS(3755), 1,
      sym__brace_start,
    ACTIONS(8722), 1,
      aux_sym_heredoc_redirect_token1,
    STATE(3571), 1,
      aux_sym__heredoc_command,
    STATE(5351), 1,
      aux_sym__literal_repeat1,
    STATE(5456), 1,
      sym_concatenation,
    ACTIONS(3715), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(3749), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(3713), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(5062), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [194777] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(8565), 1,
      sym_word,
    ACTIONS(8571), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8573), 1,
      anon_sym_DOLLAR,
    ACTIONS(8575), 1,
      sym__special_character,
    ACTIONS(8577), 1,
      anon_sym_DQUOTE,
    ACTIONS(8581), 1,
      aux_sym_number_token1,
    ACTIONS(8583), 1,
      aux_sym_number_token2,
    ACTIONS(8585), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8587), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(8589), 1,
      anon_sym_BQUOTE,
    ACTIONS(8591), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8595), 1,
      sym_test_operator,
    ACTIONS(8597), 1,
      sym__brace_start,
    ACTIONS(8724), 1,
      anon_sym_RPAREN,
    STATE(5365), 1,
      aux_sym__literal_repeat1,
    ACTIONS(8567), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8579), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(8593), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3590), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    STATE(5045), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [194853] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(8565), 1,
      sym_word,
    ACTIONS(8571), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8573), 1,
      anon_sym_DOLLAR,
    ACTIONS(8575), 1,
      sym__special_character,
    ACTIONS(8577), 1,
      anon_sym_DQUOTE,
    ACTIONS(8581), 1,
      aux_sym_number_token1,
    ACTIONS(8583), 1,
      aux_sym_number_token2,
    ACTIONS(8585), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8587), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(8589), 1,
      anon_sym_BQUOTE,
    ACTIONS(8591), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8595), 1,
      sym_test_operator,
    ACTIONS(8597), 1,
      sym__brace_start,
    ACTIONS(8726), 1,
      anon_sym_RPAREN,
    STATE(5365), 1,
      aux_sym__literal_repeat1,
    ACTIONS(8567), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8579), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(8593), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3604), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    STATE(5045), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [194929] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(8565), 1,
      sym_word,
    ACTIONS(8571), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8573), 1,
      anon_sym_DOLLAR,
    ACTIONS(8575), 1,
      sym__special_character,
    ACTIONS(8577), 1,
      anon_sym_DQUOTE,
    ACTIONS(8581), 1,
      aux_sym_number_token1,
    ACTIONS(8583), 1,
      aux_sym_number_token2,
    ACTIONS(8585), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8587), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(8589), 1,
      anon_sym_BQUOTE,
    ACTIONS(8591), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8595), 1,
      sym_test_operator,
    ACTIONS(8597), 1,
      sym__brace_start,
    ACTIONS(8728), 1,
      anon_sym_RPAREN,
    STATE(5365), 1,
      aux_sym__literal_repeat1,
    ACTIONS(8567), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8579), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(8593), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3604), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    STATE(5045), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [195005] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(8565), 1,
      sym_word,
    ACTIONS(8571), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8573), 1,
      anon_sym_DOLLAR,
    ACTIONS(8575), 1,
      sym__special_character,
    ACTIONS(8577), 1,
      anon_sym_DQUOTE,
    ACTIONS(8581), 1,
      aux_sym_number_token1,
    ACTIONS(8583), 1,
      aux_sym_number_token2,
    ACTIONS(8585), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8587), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(8589), 1,
      anon_sym_BQUOTE,
    ACTIONS(8591), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8595), 1,
      sym_test_operator,
    ACTIONS(8597), 1,
      sym__brace_start,
    ACTIONS(8730), 1,
      anon_sym_RPAREN,
    STATE(5365), 1,
      aux_sym__literal_repeat1,
    ACTIONS(8567), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8579), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(8593), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3577), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    STATE(5045), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [195081] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(8565), 1,
      sym_word,
    ACTIONS(8571), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8573), 1,
      anon_sym_DOLLAR,
    ACTIONS(8575), 1,
      sym__special_character,
    ACTIONS(8577), 1,
      anon_sym_DQUOTE,
    ACTIONS(8581), 1,
      aux_sym_number_token1,
    ACTIONS(8583), 1,
      aux_sym_number_token2,
    ACTIONS(8585), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8587), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(8589), 1,
      anon_sym_BQUOTE,
    ACTIONS(8591), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8595), 1,
      sym_test_operator,
    ACTIONS(8597), 1,
      sym__brace_start,
    ACTIONS(8732), 1,
      anon_sym_RPAREN,
    STATE(5365), 1,
      aux_sym__literal_repeat1,
    ACTIONS(8567), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8579), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(8593), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3604), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    STATE(5045), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [195157] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3729), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(3731), 1,
      anon_sym_DOLLAR,
    ACTIONS(3733), 1,
      sym__special_character,
    ACTIONS(3735), 1,
      anon_sym_DQUOTE,
    ACTIONS(3737), 1,
      aux_sym_number_token1,
    ACTIONS(3739), 1,
      aux_sym_number_token2,
    ACTIONS(3741), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(3743), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3745), 1,
      anon_sym_BQUOTE,
    ACTIONS(3747), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3753), 1,
      sym_test_operator,
    ACTIONS(3755), 1,
      sym__brace_start,
    ACTIONS(8734), 1,
      aux_sym_heredoc_redirect_token1,
    STATE(3571), 1,
      aux_sym__heredoc_command,
    STATE(5351), 1,
      aux_sym__literal_repeat1,
    STATE(5456), 1,
      sym_concatenation,
    ACTIONS(3715), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(3749), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(3713), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(5062), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [195233] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(8565), 1,
      sym_word,
    ACTIONS(8571), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8573), 1,
      anon_sym_DOLLAR,
    ACTIONS(8575), 1,
      sym__special_character,
    ACTIONS(8577), 1,
      anon_sym_DQUOTE,
    ACTIONS(8581), 1,
      aux_sym_number_token1,
    ACTIONS(8583), 1,
      aux_sym_number_token2,
    ACTIONS(8585), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8587), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(8589), 1,
      anon_sym_BQUOTE,
    ACTIONS(8591), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8595), 1,
      sym_test_operator,
    ACTIONS(8597), 1,
      sym__brace_start,
    ACTIONS(8736), 1,
      anon_sym_RPAREN,
    STATE(5365), 1,
      aux_sym__literal_repeat1,
    ACTIONS(8567), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8579), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(8593), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3595), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    STATE(5045), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [195309] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(8565), 1,
      sym_word,
    ACTIONS(8571), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8573), 1,
      anon_sym_DOLLAR,
    ACTIONS(8575), 1,
      sym__special_character,
    ACTIONS(8577), 1,
      anon_sym_DQUOTE,
    ACTIONS(8581), 1,
      aux_sym_number_token1,
    ACTIONS(8583), 1,
      aux_sym_number_token2,
    ACTIONS(8585), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8587), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(8589), 1,
      anon_sym_BQUOTE,
    ACTIONS(8591), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8595), 1,
      sym_test_operator,
    ACTIONS(8597), 1,
      sym__brace_start,
    ACTIONS(8738), 1,
      anon_sym_RPAREN,
    STATE(5365), 1,
      aux_sym__literal_repeat1,
    ACTIONS(8567), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8579), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(8593), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3587), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    STATE(5045), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [195385] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(8565), 1,
      sym_word,
    ACTIONS(8571), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8573), 1,
      anon_sym_DOLLAR,
    ACTIONS(8575), 1,
      sym__special_character,
    ACTIONS(8577), 1,
      anon_sym_DQUOTE,
    ACTIONS(8581), 1,
      aux_sym_number_token1,
    ACTIONS(8583), 1,
      aux_sym_number_token2,
    ACTIONS(8585), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8587), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(8589), 1,
      anon_sym_BQUOTE,
    ACTIONS(8591), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8595), 1,
      sym_test_operator,
    ACTIONS(8597), 1,
      sym__brace_start,
    ACTIONS(8740), 1,
      anon_sym_RPAREN,
    STATE(5365), 1,
      aux_sym__literal_repeat1,
    ACTIONS(8567), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8579), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(8593), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3588), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    STATE(5045), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [195461] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(8565), 1,
      sym_word,
    ACTIONS(8571), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8573), 1,
      anon_sym_DOLLAR,
    ACTIONS(8575), 1,
      sym__special_character,
    ACTIONS(8577), 1,
      anon_sym_DQUOTE,
    ACTIONS(8581), 1,
      aux_sym_number_token1,
    ACTIONS(8583), 1,
      aux_sym_number_token2,
    ACTIONS(8585), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8587), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(8589), 1,
      anon_sym_BQUOTE,
    ACTIONS(8591), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8595), 1,
      sym_test_operator,
    ACTIONS(8597), 1,
      sym__brace_start,
    ACTIONS(8742), 1,
      anon_sym_RPAREN,
    STATE(5365), 1,
      aux_sym__literal_repeat1,
    ACTIONS(8567), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8579), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(8593), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3604), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    STATE(5045), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [195537] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3729), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(3731), 1,
      anon_sym_DOLLAR,
    ACTIONS(3733), 1,
      sym__special_character,
    ACTIONS(3735), 1,
      anon_sym_DQUOTE,
    ACTIONS(3737), 1,
      aux_sym_number_token1,
    ACTIONS(3739), 1,
      aux_sym_number_token2,
    ACTIONS(3741), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(3743), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3745), 1,
      anon_sym_BQUOTE,
    ACTIONS(3747), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3753), 1,
      sym_test_operator,
    ACTIONS(3755), 1,
      sym__brace_start,
    ACTIONS(8744), 1,
      aux_sym_heredoc_redirect_token1,
    STATE(3571), 1,
      aux_sym__heredoc_command,
    STATE(5351), 1,
      aux_sym__literal_repeat1,
    STATE(5456), 1,
      sym_concatenation,
    ACTIONS(3715), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(3749), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(3713), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(5062), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [195613] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(5770), 10,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(5772), 22,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [195653] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1241), 1,
      sym_file_descriptor,
    ACTIONS(8694), 1,
      anon_sym_DQUOTE,
    ACTIONS(8698), 1,
      sym_variable_name,
    STATE(5487), 1,
      sym_string,
    ACTIONS(8696), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(8692), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1239), 17,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [195703] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(8565), 1,
      sym_word,
    ACTIONS(8571), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8573), 1,
      anon_sym_DOLLAR,
    ACTIONS(8575), 1,
      sym__special_character,
    ACTIONS(8577), 1,
      anon_sym_DQUOTE,
    ACTIONS(8581), 1,
      aux_sym_number_token1,
    ACTIONS(8583), 1,
      aux_sym_number_token2,
    ACTIONS(8585), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8587), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(8589), 1,
      anon_sym_BQUOTE,
    ACTIONS(8591), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8595), 1,
      sym_test_operator,
    ACTIONS(8597), 1,
      sym__brace_start,
    ACTIONS(8746), 1,
      anon_sym_RPAREN,
    STATE(5365), 1,
      aux_sym__literal_repeat1,
    ACTIONS(8567), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8579), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(8593), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3576), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    STATE(5045), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [195779] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1241), 1,
      sym_file_descriptor,
    ACTIONS(4084), 1,
      anon_sym_DQUOTE,
    ACTIONS(7613), 1,
      sym_variable_name,
    STATE(4606), 1,
      sym_string,
    ACTIONS(7611), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(7609), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1239), 17,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
  [195829] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(5373), 10,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(5375), 22,
      sym_file_descriptor,
      sym_variable_name,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [195869] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(8565), 1,
      sym_word,
    ACTIONS(8571), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8573), 1,
      anon_sym_DOLLAR,
    ACTIONS(8575), 1,
      sym__special_character,
    ACTIONS(8577), 1,
      anon_sym_DQUOTE,
    ACTIONS(8581), 1,
      aux_sym_number_token1,
    ACTIONS(8583), 1,
      aux_sym_number_token2,
    ACTIONS(8585), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8587), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(8589), 1,
      anon_sym_BQUOTE,
    ACTIONS(8591), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8595), 1,
      sym_test_operator,
    ACTIONS(8597), 1,
      sym__brace_start,
    ACTIONS(8748), 1,
      anon_sym_RPAREN,
    STATE(5365), 1,
      aux_sym__literal_repeat1,
    ACTIONS(8567), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8579), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(8593), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3606), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    STATE(5045), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [195945] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3729), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(3731), 1,
      anon_sym_DOLLAR,
    ACTIONS(3733), 1,
      sym__special_character,
    ACTIONS(3735), 1,
      anon_sym_DQUOTE,
    ACTIONS(3737), 1,
      aux_sym_number_token1,
    ACTIONS(3739), 1,
      aux_sym_number_token2,
    ACTIONS(3741), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(3743), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3745), 1,
      anon_sym_BQUOTE,
    ACTIONS(3747), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3753), 1,
      sym_test_operator,
    ACTIONS(3755), 1,
      sym__brace_start,
    ACTIONS(8750), 1,
      aux_sym_heredoc_redirect_token1,
    STATE(3571), 1,
      aux_sym__heredoc_command,
    STATE(5351), 1,
      aux_sym__literal_repeat1,
    STATE(5456), 1,
      sym_concatenation,
    ACTIONS(3715), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(3749), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(3713), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(5062), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [196021] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(2216), 1,
      anon_sym_RPAREN,
    ACTIONS(8752), 1,
      sym_word,
    ACTIONS(8758), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8761), 1,
      anon_sym_DOLLAR,
    ACTIONS(8764), 1,
      sym__special_character,
    ACTIONS(8767), 1,
      anon_sym_DQUOTE,
    ACTIONS(8773), 1,
      aux_sym_number_token1,
    ACTIONS(8776), 1,
      aux_sym_number_token2,
    ACTIONS(8779), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8782), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(8785), 1,
      anon_sym_BQUOTE,
    ACTIONS(8788), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8794), 1,
      sym_test_operator,
    ACTIONS(8797), 1,
      sym__brace_start,
    STATE(5365), 1,
      aux_sym__literal_repeat1,
    ACTIONS(8755), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8770), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(8791), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3604), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    STATE(5045), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [196097] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3729), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(3731), 1,
      anon_sym_DOLLAR,
    ACTIONS(3733), 1,
      sym__special_character,
    ACTIONS(3735), 1,
      anon_sym_DQUOTE,
    ACTIONS(3737), 1,
      aux_sym_number_token1,
    ACTIONS(3739), 1,
      aux_sym_number_token2,
    ACTIONS(3741), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(3743), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3745), 1,
      anon_sym_BQUOTE,
    ACTIONS(3747), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3753), 1,
      sym_test_operator,
    ACTIONS(3755), 1,
      sym__brace_start,
    ACTIONS(8800), 1,
      aux_sym_heredoc_redirect_token1,
    STATE(3571), 1,
      aux_sym__heredoc_command,
    STATE(5351), 1,
      aux_sym__literal_repeat1,
    STATE(5456), 1,
      sym_concatenation,
    ACTIONS(3715), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(3749), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(3713), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(5062), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [196173] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(8565), 1,
      sym_word,
    ACTIONS(8571), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8573), 1,
      anon_sym_DOLLAR,
    ACTIONS(8575), 1,
      sym__special_character,
    ACTIONS(8577), 1,
      anon_sym_DQUOTE,
    ACTIONS(8581), 1,
      aux_sym_number_token1,
    ACTIONS(8583), 1,
      aux_sym_number_token2,
    ACTIONS(8585), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8587), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(8589), 1,
      anon_sym_BQUOTE,
    ACTIONS(8591), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8595), 1,
      sym_test_operator,
    ACTIONS(8597), 1,
      sym__brace_start,
    ACTIONS(8802), 1,
      anon_sym_RPAREN,
    STATE(5365), 1,
      aux_sym__literal_repeat1,
    ACTIONS(8567), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8579), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(8593), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3604), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    STATE(5045), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [196249] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1241), 1,
      sym_file_descriptor,
    ACTIONS(7673), 1,
      anon_sym_DQUOTE,
    ACTIONS(7677), 1,
      sym_variable_name,
    STATE(4393), 1,
      sym_string,
    ACTIONS(7675), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(7671), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1239), 17,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
  [196299] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3729), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(3731), 1,
      anon_sym_DOLLAR,
    ACTIONS(3733), 1,
      sym__special_character,
    ACTIONS(3735), 1,
      anon_sym_DQUOTE,
    ACTIONS(3737), 1,
      aux_sym_number_token1,
    ACTIONS(3739), 1,
      aux_sym_number_token2,
    ACTIONS(3741), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(3743), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3745), 1,
      anon_sym_BQUOTE,
    ACTIONS(3747), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3753), 1,
      sym_test_operator,
    ACTIONS(3755), 1,
      sym__brace_start,
    ACTIONS(8804), 1,
      aux_sym_heredoc_redirect_token1,
    STATE(3571), 1,
      aux_sym__heredoc_command,
    STATE(5351), 1,
      aux_sym__literal_repeat1,
    STATE(5456), 1,
      sym_concatenation,
    ACTIONS(3715), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(3749), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(3713), 3,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(5062), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [196375] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6488), 1,
      anon_sym_DOLLAR,
    ACTIONS(6494), 1,
      aux_sym_number_token1,
    ACTIONS(6496), 1,
      aux_sym_number_token2,
    ACTIONS(6500), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(6510), 1,
      sym_extglob_pattern,
    ACTIONS(6512), 1,
      sym__brace_start,
    ACTIONS(7226), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(7228), 1,
      sym__special_character,
    ACTIONS(7230), 1,
      anon_sym_DQUOTE,
    ACTIONS(7234), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(7236), 1,
      anon_sym_BQUOTE,
    ACTIONS(7238), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8806), 1,
      sym_word,
    ACTIONS(8810), 1,
      sym_test_operator,
    STATE(6415), 1,
      aux_sym__literal_repeat1,
    ACTIONS(7224), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(7240), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(8808), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(6569), 2,
      sym_concatenation,
      sym__extglob_blob,
    STATE(6302), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [196451] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(8565), 1,
      sym_word,
    ACTIONS(8571), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8573), 1,
      anon_sym_DOLLAR,
    ACTIONS(8575), 1,
      sym__special_character,
    ACTIONS(8577), 1,
      anon_sym_DQUOTE,
    ACTIONS(8581), 1,
      aux_sym_number_token1,
    ACTIONS(8583), 1,
      aux_sym_number_token2,
    ACTIONS(8585), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8587), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(8589), 1,
      anon_sym_BQUOTE,
    ACTIONS(8591), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8595), 1,
      sym_test_operator,
    ACTIONS(8597), 1,
      sym__brace_start,
    ACTIONS(8812), 1,
      anon_sym_RPAREN,
    STATE(5365), 1,
      aux_sym__literal_repeat1,
    ACTIONS(8567), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8579), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(8593), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3604), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    STATE(5045), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [196527] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(8565), 1,
      sym_word,
    ACTIONS(8571), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8573), 1,
      anon_sym_DOLLAR,
    ACTIONS(8575), 1,
      sym__special_character,
    ACTIONS(8577), 1,
      anon_sym_DQUOTE,
    ACTIONS(8581), 1,
      aux_sym_number_token1,
    ACTIONS(8583), 1,
      aux_sym_number_token2,
    ACTIONS(8585), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8587), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(8589), 1,
      anon_sym_BQUOTE,
    ACTIONS(8591), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8595), 1,
      sym_test_operator,
    ACTIONS(8597), 1,
      sym__brace_start,
    ACTIONS(8814), 1,
      anon_sym_RPAREN,
    STATE(5365), 1,
      aux_sym__literal_repeat1,
    ACTIONS(8567), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8579), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(8593), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3610), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    STATE(5045), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [196603] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1235), 1,
      sym_file_descriptor,
    ACTIONS(4084), 1,
      anon_sym_DQUOTE,
    ACTIONS(7613), 1,
      sym_variable_name,
    STATE(4606), 1,
      sym_string,
    ACTIONS(7611), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(7609), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1227), 17,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_heredoc_redirect_token1,
  [196653] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(8565), 1,
      sym_word,
    ACTIONS(8571), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8573), 1,
      anon_sym_DOLLAR,
    ACTIONS(8575), 1,
      sym__special_character,
    ACTIONS(8577), 1,
      anon_sym_DQUOTE,
    ACTIONS(8581), 1,
      aux_sym_number_token1,
    ACTIONS(8583), 1,
      aux_sym_number_token2,
    ACTIONS(8585), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8587), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(8589), 1,
      anon_sym_BQUOTE,
    ACTIONS(8591), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8595), 1,
      sym_test_operator,
    ACTIONS(8597), 1,
      sym__brace_start,
    ACTIONS(8816), 1,
      anon_sym_RPAREN,
    STATE(5365), 1,
      aux_sym__literal_repeat1,
    ACTIONS(8567), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8579), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(8593), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3604), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    STATE(5045), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [196729] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3731), 1,
      anon_sym_DOLLAR,
    ACTIONS(3737), 1,
      aux_sym_number_token1,
    ACTIONS(3739), 1,
      aux_sym_number_token2,
    ACTIONS(3743), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3755), 1,
      sym__brace_start,
    ACTIONS(7759), 1,
      sym_word,
    ACTIONS(7767), 1,
      sym_test_operator,
    ACTIONS(8820), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8822), 1,
      sym__special_character,
    ACTIONS(8824), 1,
      anon_sym_DQUOTE,
    ACTIONS(8828), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8830), 1,
      anon_sym_BQUOTE,
    ACTIONS(8832), 1,
      anon_sym_DOLLAR_BQUOTE,
    STATE(4975), 1,
      aux_sym__literal_repeat1,
    ACTIONS(8818), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8826), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(8834), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3484), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    STATE(4460), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [196802] = 19,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(363), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(367), 1,
      anon_sym_DOLLAR,
    ACTIONS(371), 1,
      anon_sym_DQUOTE,
    ACTIONS(375), 1,
      aux_sym_number_token1,
    ACTIONS(377), 1,
      aux_sym_number_token2,
    ACTIONS(379), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(381), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(385), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(391), 1,
      sym__brace_start,
    ACTIONS(5809), 1,
      anon_sym_BQUOTE,
    ACTIONS(8836), 1,
      sym_word,
    ACTIONS(8838), 1,
      anon_sym_RBRACK,
    ACTIONS(8844), 1,
      sym_test_operator,
    ACTIONS(352), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(387), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(8840), 2,
      sym__special_character,
      sym__comment_word,
    ACTIONS(8842), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2730), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [196873] = 19,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(363), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(367), 1,
      anon_sym_DOLLAR,
    ACTIONS(371), 1,
      anon_sym_DQUOTE,
    ACTIONS(375), 1,
      aux_sym_number_token1,
    ACTIONS(377), 1,
      aux_sym_number_token2,
    ACTIONS(379), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(381), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(385), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(391), 1,
      sym__brace_start,
    ACTIONS(5809), 1,
      anon_sym_BQUOTE,
    ACTIONS(8836), 1,
      sym_word,
    ACTIONS(8844), 1,
      sym_test_operator,
    ACTIONS(8846), 1,
      anon_sym_RBRACK,
    ACTIONS(352), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(387), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(8840), 2,
      sym__special_character,
      sym__comment_word,
    ACTIONS(8842), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2730), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [196944] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3731), 1,
      anon_sym_DOLLAR,
    ACTIONS(3737), 1,
      aux_sym_number_token1,
    ACTIONS(3739), 1,
      aux_sym_number_token2,
    ACTIONS(3743), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3755), 1,
      sym__brace_start,
    ACTIONS(7759), 1,
      sym_word,
    ACTIONS(7767), 1,
      sym_test_operator,
    ACTIONS(8820), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8822), 1,
      sym__special_character,
    ACTIONS(8824), 1,
      anon_sym_DQUOTE,
    ACTIONS(8828), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8830), 1,
      anon_sym_BQUOTE,
    ACTIONS(8832), 1,
      anon_sym_DOLLAR_BQUOTE,
    STATE(4975), 1,
      aux_sym__literal_repeat1,
    ACTIONS(8818), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8826), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(8834), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3460), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    STATE(4460), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [197017] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3612), 1,
      anon_sym_DOLLAR,
    ACTIONS(3616), 1,
      aux_sym_number_token1,
    ACTIONS(3618), 1,
      aux_sym_number_token2,
    ACTIONS(3622), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3632), 1,
      sym__brace_start,
    ACTIONS(4975), 1,
      sym_word,
    ACTIONS(4979), 1,
      sym_test_operator,
    ACTIONS(8850), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8852), 1,
      sym__special_character,
    ACTIONS(8854), 1,
      anon_sym_DQUOTE,
    ACTIONS(8858), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8860), 1,
      anon_sym_BQUOTE,
    ACTIONS(8862), 1,
      anon_sym_DOLLAR_BQUOTE,
    STATE(2270), 1,
      aux_sym__literal_repeat1,
    ACTIONS(8848), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8856), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(8864), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(972), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    STATE(2572), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [197090] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1241), 1,
      sym_file_descriptor,
    ACTIONS(8868), 1,
      anon_sym_DQUOTE,
    ACTIONS(8872), 1,
      sym_variable_name,
    STATE(5409), 1,
      sym_string,
    ACTIONS(8870), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(8866), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1239), 16,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [197139] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1454), 1,
      anon_sym_DOLLAR,
    ACTIONS(1460), 1,
      aux_sym_number_token1,
    ACTIONS(1462), 1,
      aux_sym_number_token2,
    ACTIONS(1466), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1478), 1,
      sym__brace_start,
    ACTIONS(8874), 1,
      sym_word,
    ACTIONS(8878), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8880), 1,
      sym__special_character,
    ACTIONS(8882), 1,
      anon_sym_DQUOTE,
    ACTIONS(8886), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8888), 1,
      anon_sym_BQUOTE,
    ACTIONS(8890), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8894), 1,
      sym_test_operator,
    ACTIONS(8896), 1,
      sym_regex,
    STATE(1180), 1,
      aux_sym__literal_repeat1,
    STATE(1346), 1,
      sym_concatenation,
    ACTIONS(8876), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8884), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(8892), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(1000), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [197214] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(8102), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8104), 1,
      anon_sym_DOLLAR,
    ACTIONS(8106), 1,
      sym__special_character,
    ACTIONS(8108), 1,
      anon_sym_DQUOTE,
    ACTIONS(8112), 1,
      aux_sym_number_token1,
    ACTIONS(8114), 1,
      aux_sym_number_token2,
    ACTIONS(8116), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8118), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(8120), 1,
      anon_sym_BQUOTE,
    ACTIONS(8122), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8132), 1,
      sym__brace_start,
    ACTIONS(8898), 1,
      sym_word,
    ACTIONS(8902), 1,
      sym_test_operator,
    STATE(1474), 1,
      aux_sym__literal_repeat1,
    ACTIONS(8098), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8124), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(8900), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(573), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    STATE(1158), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [197287] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(8102), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8104), 1,
      anon_sym_DOLLAR,
    ACTIONS(8106), 1,
      sym__special_character,
    ACTIONS(8108), 1,
      anon_sym_DQUOTE,
    ACTIONS(8112), 1,
      aux_sym_number_token1,
    ACTIONS(8114), 1,
      aux_sym_number_token2,
    ACTIONS(8116), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8118), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(8120), 1,
      anon_sym_BQUOTE,
    ACTIONS(8122), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8132), 1,
      sym__brace_start,
    ACTIONS(8898), 1,
      sym_word,
    ACTIONS(8902), 1,
      sym_test_operator,
    STATE(1474), 1,
      aux_sym__literal_repeat1,
    ACTIONS(8098), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8124), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(8900), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(575), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    STATE(1158), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [197360] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(4629), 1,
      sym_word,
    ACTIONS(4633), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(4635), 1,
      anon_sym_DOLLAR,
    ACTIONS(4637), 1,
      sym__special_character,
    ACTIONS(4639), 1,
      anon_sym_DQUOTE,
    ACTIONS(4643), 1,
      aux_sym_number_token1,
    ACTIONS(4645), 1,
      aux_sym_number_token2,
    ACTIONS(4647), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(4649), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(4651), 1,
      anon_sym_BQUOTE,
    ACTIONS(4653), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(4657), 1,
      sym_test_operator,
    ACTIONS(4659), 1,
      sym__brace_start,
    STATE(2914), 1,
      aux_sym__literal_repeat1,
    ACTIONS(4631), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(4641), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(4655), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(897), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    STATE(2487), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [197433] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(4629), 1,
      sym_word,
    ACTIONS(4633), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(4635), 1,
      anon_sym_DOLLAR,
    ACTIONS(4637), 1,
      sym__special_character,
    ACTIONS(4639), 1,
      anon_sym_DQUOTE,
    ACTIONS(4643), 1,
      aux_sym_number_token1,
    ACTIONS(4645), 1,
      aux_sym_number_token2,
    ACTIONS(4647), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(4649), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(4651), 1,
      anon_sym_BQUOTE,
    ACTIONS(4653), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(4657), 1,
      sym_test_operator,
    ACTIONS(4659), 1,
      sym__brace_start,
    STATE(2914), 1,
      aux_sym__literal_repeat1,
    ACTIONS(4631), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(4641), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(4655), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(903), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    STATE(2487), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [197506] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3471), 1,
      sym_word,
    ACTIONS(3477), 1,
      anon_sym_DOLLAR,
    ACTIONS(3481), 1,
      aux_sym_number_token1,
    ACTIONS(3483), 1,
      aux_sym_number_token2,
    ACTIONS(3487), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3495), 1,
      sym_test_operator,
    ACTIONS(3497), 1,
      sym__brace_start,
    ACTIONS(8906), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8908), 1,
      sym__special_character,
    ACTIONS(8910), 1,
      anon_sym_DQUOTE,
    ACTIONS(8914), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8916), 1,
      anon_sym_BQUOTE,
    ACTIONS(8918), 1,
      anon_sym_DOLLAR_BQUOTE,
    STATE(2024), 1,
      aux_sym__literal_repeat1,
    ACTIONS(8904), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8912), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(8920), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(705), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    STATE(1630), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [197579] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(5108), 1,
      sym_word,
    ACTIONS(5112), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(5114), 1,
      anon_sym_DOLLAR,
    ACTIONS(5116), 1,
      sym__special_character,
    ACTIONS(5118), 1,
      anon_sym_DQUOTE,
    ACTIONS(5122), 1,
      aux_sym_number_token1,
    ACTIONS(5124), 1,
      aux_sym_number_token2,
    ACTIONS(5126), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(5128), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(5130), 1,
      anon_sym_BQUOTE,
    ACTIONS(5132), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(5136), 1,
      sym_test_operator,
    ACTIONS(5138), 1,
      sym__brace_start,
    STATE(3114), 1,
      aux_sym__literal_repeat1,
    ACTIONS(5110), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(5120), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(5134), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(1056), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    STATE(2613), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [197652] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3471), 1,
      sym_word,
    ACTIONS(3477), 1,
      anon_sym_DOLLAR,
    ACTIONS(3481), 1,
      aux_sym_number_token1,
    ACTIONS(3483), 1,
      aux_sym_number_token2,
    ACTIONS(3487), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3495), 1,
      sym_test_operator,
    ACTIONS(3497), 1,
      sym__brace_start,
    ACTIONS(8906), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8908), 1,
      sym__special_character,
    ACTIONS(8910), 1,
      anon_sym_DQUOTE,
    ACTIONS(8914), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8916), 1,
      anon_sym_BQUOTE,
    ACTIONS(8918), 1,
      anon_sym_DOLLAR_BQUOTE,
    STATE(2024), 1,
      aux_sym__literal_repeat1,
    ACTIONS(8904), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8912), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(8920), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(702), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    STATE(1630), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [197725] = 19,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8030), 1,
      anon_sym_SLASH,
    ACTIONS(8032), 1,
      anon_sym_PERCENT,
    ACTIONS(8034), 1,
      anon_sym_COLON,
    ACTIONS(8038), 1,
      anon_sym_RBRACE3,
    ACTIONS(8040), 1,
      anon_sym_AT,
    ACTIONS(8042), 1,
      anon_sym_STAR2,
    ACTIONS(8922), 1,
      anon_sym_LBRACK,
    STATE(6994), 1,
      sym__expansion_operator,
    STATE(6995), 1,
      sym__expansion_max_length,
    STATE(6999), 1,
      sym__expansion_regex_removal,
    STATE(7000), 1,
      sym__expansion_regex_replacement,
    STATE(7001), 1,
      sym__expansion_regex,
    STATE(7002), 1,
      sym__expansion_expression,
    ACTIONS(8028), 2,
      anon_sym_COMMA,
      anon_sym_CARET,
    ACTIONS(8048), 2,
      anon_sym_COMMA_COMMA,
      anon_sym_CARET_CARET,
    ACTIONS(8036), 3,
      sym__immediate_double_hash,
      anon_sym_POUND,
      anon_sym_PERCENT_PERCENT,
    ACTIONS(8046), 3,
      anon_sym_SLASH_SLASH,
      anon_sym_SLASH_POUND,
      anon_sym_SLASH_PERCENT,
    ACTIONS(8044), 8,
      anon_sym_EQ2,
      anon_sym_COLON_EQ,
      anon_sym_DASH3,
      anon_sym_COLON_DASH,
      anon_sym_PLUS3,
      anon_sym_COLON_PLUS,
      anon_sym_QMARK2,
      anon_sym_COLON_QMARK,
  [197796] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(5108), 1,
      sym_word,
    ACTIONS(5112), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(5114), 1,
      anon_sym_DOLLAR,
    ACTIONS(5116), 1,
      sym__special_character,
    ACTIONS(5118), 1,
      anon_sym_DQUOTE,
    ACTIONS(5122), 1,
      aux_sym_number_token1,
    ACTIONS(5124), 1,
      aux_sym_number_token2,
    ACTIONS(5126), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(5128), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(5130), 1,
      anon_sym_BQUOTE,
    ACTIONS(5132), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(5136), 1,
      sym_test_operator,
    ACTIONS(5138), 1,
      sym__brace_start,
    STATE(3114), 1,
      aux_sym__literal_repeat1,
    ACTIONS(5110), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(5120), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(5134), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(1035), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    STATE(2613), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [197869] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3397), 1,
      anon_sym_DOLLAR,
    ACTIONS(3401), 1,
      aux_sym_number_token1,
    ACTIONS(3403), 1,
      aux_sym_number_token2,
    ACTIONS(3407), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3417), 1,
      sym__brace_start,
    ACTIONS(4484), 1,
      sym_word,
    ACTIONS(4488), 1,
      sym_test_operator,
    ACTIONS(8926), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8928), 1,
      sym__special_character,
    ACTIONS(8930), 1,
      anon_sym_DQUOTE,
    ACTIONS(8934), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8936), 1,
      anon_sym_BQUOTE,
    ACTIONS(8938), 1,
      anon_sym_DOLLAR_BQUOTE,
    STATE(2161), 1,
      aux_sym__literal_repeat1,
    ACTIONS(8924), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8932), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(8940), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(862), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    STATE(2422), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [197942] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3612), 1,
      anon_sym_DOLLAR,
    ACTIONS(3616), 1,
      aux_sym_number_token1,
    ACTIONS(3618), 1,
      aux_sym_number_token2,
    ACTIONS(3622), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3632), 1,
      sym__brace_start,
    ACTIONS(4975), 1,
      sym_word,
    ACTIONS(4979), 1,
      sym_test_operator,
    ACTIONS(8850), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8852), 1,
      sym__special_character,
    ACTIONS(8854), 1,
      anon_sym_DQUOTE,
    ACTIONS(8858), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8860), 1,
      anon_sym_BQUOTE,
    ACTIONS(8862), 1,
      anon_sym_DOLLAR_BQUOTE,
    STATE(2270), 1,
      aux_sym__literal_repeat1,
    ACTIONS(8848), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8856), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(8864), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(973), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    STATE(2572), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [198015] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(5604), 1,
      sym_word,
    ACTIONS(5610), 1,
      anon_sym_DOLLAR,
    ACTIONS(5616), 1,
      aux_sym_number_token1,
    ACTIONS(5618), 1,
      aux_sym_number_token2,
    ACTIONS(5622), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(5630), 1,
      sym_test_operator,
    ACTIONS(5632), 1,
      sym__brace_start,
    ACTIONS(8944), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8946), 1,
      sym__special_character,
    ACTIONS(8948), 1,
      anon_sym_DQUOTE,
    ACTIONS(8952), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8954), 1,
      anon_sym_BQUOTE,
    ACTIONS(8956), 1,
      anon_sym_DOLLAR_BQUOTE,
    STATE(3469), 1,
      aux_sym__literal_repeat1,
    ACTIONS(8942), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8950), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(8958), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(1317), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    STATE(3093), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [198088] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3397), 1,
      anon_sym_DOLLAR,
    ACTIONS(3401), 1,
      aux_sym_number_token1,
    ACTIONS(3403), 1,
      aux_sym_number_token2,
    ACTIONS(3407), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3417), 1,
      sym__brace_start,
    ACTIONS(4484), 1,
      sym_word,
    ACTIONS(4488), 1,
      sym_test_operator,
    ACTIONS(8926), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8928), 1,
      sym__special_character,
    ACTIONS(8930), 1,
      anon_sym_DQUOTE,
    ACTIONS(8934), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8936), 1,
      anon_sym_BQUOTE,
    ACTIONS(8938), 1,
      anon_sym_DOLLAR_BQUOTE,
    STATE(2161), 1,
      aux_sym__literal_repeat1,
    ACTIONS(8924), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8932), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(8940), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(863), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    STATE(2422), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [198161] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(5604), 1,
      sym_word,
    ACTIONS(5610), 1,
      anon_sym_DOLLAR,
    ACTIONS(5616), 1,
      aux_sym_number_token1,
    ACTIONS(5618), 1,
      aux_sym_number_token2,
    ACTIONS(5622), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(5630), 1,
      sym_test_operator,
    ACTIONS(5632), 1,
      sym__brace_start,
    ACTIONS(8944), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8946), 1,
      sym__special_character,
    ACTIONS(8948), 1,
      anon_sym_DQUOTE,
    ACTIONS(8952), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8954), 1,
      anon_sym_BQUOTE,
    ACTIONS(8956), 1,
      anon_sym_DOLLAR_BQUOTE,
    STATE(3469), 1,
      aux_sym__literal_repeat1,
    ACTIONS(8942), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8950), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(8958), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(1332), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    STATE(3093), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [198234] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(2598), 1,
      sym_word,
    ACTIONS(2604), 1,
      anon_sym_DOLLAR,
    ACTIONS(2610), 1,
      aux_sym_number_token1,
    ACTIONS(2612), 1,
      aux_sym_number_token2,
    ACTIONS(2616), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2624), 1,
      sym_test_operator,
    ACTIONS(2626), 1,
      sym__brace_start,
    ACTIONS(8962), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8964), 1,
      sym__special_character,
    ACTIONS(8966), 1,
      anon_sym_DQUOTE,
    ACTIONS(8970), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8972), 1,
      anon_sym_BQUOTE,
    ACTIONS(8974), 1,
      anon_sym_DOLLAR_BQUOTE,
    STATE(1803), 1,
      aux_sym__literal_repeat1,
    ACTIONS(8960), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8968), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(8976), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(657), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    STATE(1381), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [198307] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1235), 1,
      sym_file_descriptor,
    ACTIONS(8868), 1,
      anon_sym_DQUOTE,
    ACTIONS(8872), 1,
      sym_variable_name,
    STATE(5409), 1,
      sym_string,
    ACTIONS(8870), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(8866), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1227), 16,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [198356] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(8178), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8180), 1,
      anon_sym_DOLLAR,
    ACTIONS(8184), 1,
      anon_sym_DQUOTE,
    ACTIONS(8188), 1,
      aux_sym_number_token1,
    ACTIONS(8190), 1,
      aux_sym_number_token2,
    ACTIONS(8192), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8194), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(8196), 1,
      anon_sym_BQUOTE,
    ACTIONS(8198), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8208), 1,
      sym__brace_start,
    ACTIONS(8220), 1,
      sym__special_character,
    ACTIONS(8978), 1,
      sym_word,
    ACTIONS(8982), 1,
      sym_test_operator,
    STATE(1744), 1,
      aux_sym__literal_repeat1,
    ACTIONS(8174), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8200), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(8980), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(646), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    STATE(1370), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [198429] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(2598), 1,
      sym_word,
    ACTIONS(2604), 1,
      anon_sym_DOLLAR,
    ACTIONS(2610), 1,
      aux_sym_number_token1,
    ACTIONS(2612), 1,
      aux_sym_number_token2,
    ACTIONS(2616), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2624), 1,
      sym_test_operator,
    ACTIONS(2626), 1,
      sym__brace_start,
    ACTIONS(8962), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8964), 1,
      sym__special_character,
    ACTIONS(8966), 1,
      anon_sym_DQUOTE,
    ACTIONS(8970), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8972), 1,
      anon_sym_BQUOTE,
    ACTIONS(8974), 1,
      anon_sym_DOLLAR_BQUOTE,
    STATE(1803), 1,
      aux_sym__literal_repeat1,
    ACTIONS(8960), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8968), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(8976), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(616), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    STATE(1381), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [198502] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(8512), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8514), 1,
      anon_sym_DOLLAR,
    ACTIONS(8516), 1,
      sym__special_character,
    ACTIONS(8518), 1,
      anon_sym_DQUOTE,
    ACTIONS(8522), 1,
      aux_sym_number_token1,
    ACTIONS(8524), 1,
      aux_sym_number_token2,
    ACTIONS(8526), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8528), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(8530), 1,
      anon_sym_BQUOTE,
    ACTIONS(8532), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8542), 1,
      sym__brace_start,
    ACTIONS(8984), 1,
      sym_word,
    ACTIONS(8988), 1,
      sym_test_operator,
    STATE(2671), 1,
      aux_sym__literal_repeat1,
    ACTIONS(8508), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8534), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(8986), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(858), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    STATE(2369), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [198575] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(8512), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8514), 1,
      anon_sym_DOLLAR,
    ACTIONS(8516), 1,
      sym__special_character,
    ACTIONS(8518), 1,
      anon_sym_DQUOTE,
    ACTIONS(8522), 1,
      aux_sym_number_token1,
    ACTIONS(8524), 1,
      aux_sym_number_token2,
    ACTIONS(8526), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8528), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(8530), 1,
      anon_sym_BQUOTE,
    ACTIONS(8532), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8542), 1,
      sym__brace_start,
    ACTIONS(8984), 1,
      sym_word,
    ACTIONS(8988), 1,
      sym_test_operator,
    STATE(2671), 1,
      aux_sym__literal_repeat1,
    ACTIONS(8508), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8534), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(8986), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(870), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    STATE(2369), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [198648] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3876), 1,
      sym_word,
    ACTIONS(3882), 1,
      anon_sym_DOLLAR,
    ACTIONS(3886), 1,
      aux_sym_number_token1,
    ACTIONS(3888), 1,
      aux_sym_number_token2,
    ACTIONS(3892), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3900), 1,
      sym_test_operator,
    ACTIONS(3902), 1,
      sym__brace_start,
    ACTIONS(8992), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8994), 1,
      sym__special_character,
    ACTIONS(8996), 1,
      anon_sym_DQUOTE,
    ACTIONS(9000), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9002), 1,
      anon_sym_BQUOTE,
    ACTIONS(9004), 1,
      anon_sym_DOLLAR_BQUOTE,
    STATE(2281), 1,
      aux_sym__literal_repeat1,
    ACTIONS(8990), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8998), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(9006), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(757), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    STATE(1846), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [198721] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(2329), 1,
      sym_word,
    ACTIONS(2335), 1,
      anon_sym_DOLLAR,
    ACTIONS(2341), 1,
      aux_sym_number_token1,
    ACTIONS(2343), 1,
      aux_sym_number_token2,
    ACTIONS(2347), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2355), 1,
      sym_test_operator,
    ACTIONS(2357), 1,
      sym__brace_start,
    ACTIONS(9010), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9012), 1,
      sym__special_character,
    ACTIONS(9014), 1,
      anon_sym_DQUOTE,
    ACTIONS(9018), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9020), 1,
      anon_sym_BQUOTE,
    ACTIONS(9022), 1,
      anon_sym_DOLLAR_BQUOTE,
    STATE(1658), 1,
      aux_sym__literal_repeat1,
    ACTIONS(9008), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9016), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(9024), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(591), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    STATE(1290), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [198794] = 19,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(363), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(367), 1,
      anon_sym_DOLLAR,
    ACTIONS(371), 1,
      anon_sym_DQUOTE,
    ACTIONS(375), 1,
      aux_sym_number_token1,
    ACTIONS(377), 1,
      aux_sym_number_token2,
    ACTIONS(379), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(381), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(385), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(391), 1,
      sym__brace_start,
    ACTIONS(5809), 1,
      anon_sym_BQUOTE,
    ACTIONS(8836), 1,
      sym_word,
    ACTIONS(8844), 1,
      sym_test_operator,
    ACTIONS(9026), 1,
      anon_sym_RBRACK,
    ACTIONS(352), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(387), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(8840), 2,
      sym__special_character,
      sym__comment_word,
    ACTIONS(8842), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2730), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [198865] = 19,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(363), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(367), 1,
      anon_sym_DOLLAR,
    ACTIONS(371), 1,
      anon_sym_DQUOTE,
    ACTIONS(375), 1,
      aux_sym_number_token1,
    ACTIONS(377), 1,
      aux_sym_number_token2,
    ACTIONS(379), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(381), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(385), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(391), 1,
      sym__brace_start,
    ACTIONS(5809), 1,
      anon_sym_BQUOTE,
    ACTIONS(8836), 1,
      sym_word,
    ACTIONS(8844), 1,
      sym_test_operator,
    ACTIONS(9028), 1,
      anon_sym_RBRACK,
    ACTIONS(352), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(387), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(8840), 2,
      sym__special_character,
      sym__comment_word,
    ACTIONS(8842), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2730), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [198936] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3876), 1,
      sym_word,
    ACTIONS(3882), 1,
      anon_sym_DOLLAR,
    ACTIONS(3886), 1,
      aux_sym_number_token1,
    ACTIONS(3888), 1,
      aux_sym_number_token2,
    ACTIONS(3892), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3900), 1,
      sym_test_operator,
    ACTIONS(3902), 1,
      sym__brace_start,
    ACTIONS(8992), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8994), 1,
      sym__special_character,
    ACTIONS(8996), 1,
      anon_sym_DQUOTE,
    ACTIONS(9000), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9002), 1,
      anon_sym_BQUOTE,
    ACTIONS(9004), 1,
      anon_sym_DOLLAR_BQUOTE,
    STATE(2281), 1,
      aux_sym__literal_repeat1,
    ACTIONS(8990), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8998), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(9006), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(742), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    STATE(1846), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [199009] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(2329), 1,
      sym_word,
    ACTIONS(2335), 1,
      anon_sym_DOLLAR,
    ACTIONS(2341), 1,
      aux_sym_number_token1,
    ACTIONS(2343), 1,
      aux_sym_number_token2,
    ACTIONS(2347), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2355), 1,
      sym_test_operator,
    ACTIONS(2357), 1,
      sym__brace_start,
    ACTIONS(9010), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9012), 1,
      sym__special_character,
    ACTIONS(9014), 1,
      anon_sym_DQUOTE,
    ACTIONS(9018), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9020), 1,
      anon_sym_BQUOTE,
    ACTIONS(9022), 1,
      anon_sym_DOLLAR_BQUOTE,
    STATE(1658), 1,
      aux_sym__literal_repeat1,
    ACTIONS(9008), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9016), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(9024), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(595), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    STATE(1290), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [199082] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3606), 1,
      sym_word,
    ACTIONS(3612), 1,
      anon_sym_DOLLAR,
    ACTIONS(3616), 1,
      aux_sym_number_token1,
    ACTIONS(3618), 1,
      aux_sym_number_token2,
    ACTIONS(3622), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3630), 1,
      sym_test_operator,
    ACTIONS(3632), 1,
      sym__brace_start,
    ACTIONS(8850), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8854), 1,
      anon_sym_DQUOTE,
    ACTIONS(8858), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8860), 1,
      anon_sym_BQUOTE,
    ACTIONS(8862), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9030), 1,
      sym__special_character,
    STATE(2270), 1,
      aux_sym__literal_repeat1,
    ACTIONS(8848), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8864), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9032), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(773), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    STATE(1839), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [199155] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3397), 1,
      anon_sym_DOLLAR,
    ACTIONS(3401), 1,
      aux_sym_number_token1,
    ACTIONS(3403), 1,
      aux_sym_number_token2,
    ACTIONS(3407), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3417), 1,
      sym__brace_start,
    ACTIONS(3663), 1,
      sym_word,
    ACTIONS(3667), 1,
      sym_test_operator,
    ACTIONS(8926), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8930), 1,
      anon_sym_DQUOTE,
    ACTIONS(8934), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8936), 1,
      anon_sym_BQUOTE,
    ACTIONS(8938), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9034), 1,
      sym__special_character,
    STATE(2161), 1,
      aux_sym__literal_repeat1,
    ACTIONS(8924), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8940), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9036), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(754), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    STATE(1918), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [199228] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3397), 1,
      anon_sym_DOLLAR,
    ACTIONS(3401), 1,
      aux_sym_number_token1,
    ACTIONS(3403), 1,
      aux_sym_number_token2,
    ACTIONS(3407), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3417), 1,
      sym__brace_start,
    ACTIONS(3663), 1,
      sym_word,
    ACTIONS(3667), 1,
      sym_test_operator,
    ACTIONS(8926), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8930), 1,
      anon_sym_DQUOTE,
    ACTIONS(8934), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8936), 1,
      anon_sym_BQUOTE,
    ACTIONS(8938), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9034), 1,
      sym__special_character,
    STATE(2161), 1,
      aux_sym__literal_repeat1,
    ACTIONS(8924), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8940), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9036), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(755), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    STATE(1918), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [199301] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3606), 1,
      sym_word,
    ACTIONS(3612), 1,
      anon_sym_DOLLAR,
    ACTIONS(3616), 1,
      aux_sym_number_token1,
    ACTIONS(3618), 1,
      aux_sym_number_token2,
    ACTIONS(3622), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3630), 1,
      sym_test_operator,
    ACTIONS(3632), 1,
      sym__brace_start,
    ACTIONS(8850), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8854), 1,
      anon_sym_DQUOTE,
    ACTIONS(8858), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8860), 1,
      anon_sym_BQUOTE,
    ACTIONS(8862), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9030), 1,
      sym__special_character,
    STATE(2270), 1,
      aux_sym__literal_repeat1,
    ACTIONS(8848), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8864), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9032), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(714), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    STATE(1839), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [199374] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(8268), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8270), 1,
      anon_sym_DOLLAR,
    ACTIONS(8272), 1,
      sym__special_character,
    ACTIONS(8274), 1,
      anon_sym_DQUOTE,
    ACTIONS(8278), 1,
      aux_sym_number_token1,
    ACTIONS(8280), 1,
      aux_sym_number_token2,
    ACTIONS(8282), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8284), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(8286), 1,
      anon_sym_BQUOTE,
    ACTIONS(8288), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8298), 1,
      sym__brace_start,
    ACTIONS(9038), 1,
      sym_word,
    ACTIONS(9042), 1,
      sym_test_operator,
    STATE(1567), 1,
      aux_sym__literal_repeat1,
    ACTIONS(8264), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8290), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9040), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(601), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    STATE(1269), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [199447] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(8268), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8270), 1,
      anon_sym_DOLLAR,
    ACTIONS(8272), 1,
      sym__special_character,
    ACTIONS(8274), 1,
      anon_sym_DQUOTE,
    ACTIONS(8278), 1,
      aux_sym_number_token1,
    ACTIONS(8280), 1,
      aux_sym_number_token2,
    ACTIONS(8282), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8284), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(8286), 1,
      anon_sym_BQUOTE,
    ACTIONS(8288), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8298), 1,
      sym__brace_start,
    ACTIONS(9038), 1,
      sym_word,
    ACTIONS(9042), 1,
      sym_test_operator,
    STATE(1567), 1,
      aux_sym__literal_repeat1,
    ACTIONS(8264), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8290), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9040), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(614), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    STATE(1269), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [199520] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3333), 1,
      sym_word,
    ACTIONS(3339), 1,
      anon_sym_DOLLAR,
    ACTIONS(3343), 1,
      aux_sym_number_token1,
    ACTIONS(3345), 1,
      aux_sym_number_token2,
    ACTIONS(3349), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3357), 1,
      sym_test_operator,
    ACTIONS(3359), 1,
      sym__brace_start,
    ACTIONS(9046), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9048), 1,
      sym__special_character,
    ACTIONS(9050), 1,
      anon_sym_DQUOTE,
    ACTIONS(9054), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9056), 1,
      anon_sym_BQUOTE,
    ACTIONS(9058), 1,
      anon_sym_DOLLAR_BQUOTE,
    STATE(2147), 1,
      aux_sym__literal_repeat1,
    ACTIONS(9044), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9052), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(9060), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(707), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    STATE(1526), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [199593] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(2568), 1,
      sym_word,
    ACTIONS(2574), 1,
      anon_sym_DOLLAR,
    ACTIONS(2580), 1,
      aux_sym_number_token1,
    ACTIONS(2582), 1,
      aux_sym_number_token2,
    ACTIONS(2586), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2594), 1,
      sym_test_operator,
    ACTIONS(2596), 1,
      sym__brace_start,
    ACTIONS(9064), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9066), 1,
      sym__special_character,
    ACTIONS(9068), 1,
      anon_sym_DQUOTE,
    ACTIONS(9072), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9074), 1,
      anon_sym_BQUOTE,
    ACTIONS(9076), 1,
      anon_sym_DOLLAR_BQUOTE,
    STATE(1899), 1,
      aux_sym__literal_repeat1,
    ACTIONS(9062), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9070), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(9078), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(653), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    STATE(1394), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [199666] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(2568), 1,
      sym_word,
    ACTIONS(2574), 1,
      anon_sym_DOLLAR,
    ACTIONS(2580), 1,
      aux_sym_number_token1,
    ACTIONS(2582), 1,
      aux_sym_number_token2,
    ACTIONS(2586), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2594), 1,
      sym_test_operator,
    ACTIONS(2596), 1,
      sym__brace_start,
    ACTIONS(9064), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9066), 1,
      sym__special_character,
    ACTIONS(9068), 1,
      anon_sym_DQUOTE,
    ACTIONS(9072), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9074), 1,
      anon_sym_BQUOTE,
    ACTIONS(9076), 1,
      anon_sym_DOLLAR_BQUOTE,
    STATE(1899), 1,
      aux_sym__literal_repeat1,
    ACTIONS(9062), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9070), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(9078), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(615), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    STATE(1394), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [199739] = 19,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(363), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(367), 1,
      anon_sym_DOLLAR,
    ACTIONS(371), 1,
      anon_sym_DQUOTE,
    ACTIONS(375), 1,
      aux_sym_number_token1,
    ACTIONS(377), 1,
      aux_sym_number_token2,
    ACTIONS(379), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(381), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(385), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(391), 1,
      sym__brace_start,
    ACTIONS(5809), 1,
      anon_sym_BQUOTE,
    ACTIONS(8836), 1,
      sym_word,
    ACTIONS(8844), 1,
      sym_test_operator,
    ACTIONS(9080), 1,
      anon_sym_RBRACK,
    ACTIONS(352), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(387), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(8840), 2,
      sym__special_character,
      sym__comment_word,
    ACTIONS(8842), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2730), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [199810] = 19,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(363), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(367), 1,
      anon_sym_DOLLAR,
    ACTIONS(371), 1,
      anon_sym_DQUOTE,
    ACTIONS(375), 1,
      aux_sym_number_token1,
    ACTIONS(377), 1,
      aux_sym_number_token2,
    ACTIONS(379), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(381), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(385), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(391), 1,
      sym__brace_start,
    ACTIONS(5809), 1,
      anon_sym_BQUOTE,
    ACTIONS(8836), 1,
      sym_word,
    ACTIONS(8844), 1,
      sym_test_operator,
    ACTIONS(9082), 1,
      anon_sym_RBRACK,
    ACTIONS(352), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(387), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(8840), 2,
      sym__special_character,
      sym__comment_word,
    ACTIONS(8842), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2730), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [199881] = 19,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(363), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(367), 1,
      anon_sym_DOLLAR,
    ACTIONS(371), 1,
      anon_sym_DQUOTE,
    ACTIONS(375), 1,
      aux_sym_number_token1,
    ACTIONS(377), 1,
      aux_sym_number_token2,
    ACTIONS(379), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(381), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(385), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(391), 1,
      sym__brace_start,
    ACTIONS(5809), 1,
      anon_sym_BQUOTE,
    ACTIONS(8836), 1,
      sym_word,
    ACTIONS(8844), 1,
      sym_test_operator,
    ACTIONS(9084), 1,
      anon_sym_RBRACK,
    ACTIONS(352), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(387), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(8840), 2,
      sym__special_character,
      sym__comment_word,
    ACTIONS(8842), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2730), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [199952] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3612), 1,
      anon_sym_DOLLAR,
    ACTIONS(3616), 1,
      aux_sym_number_token1,
    ACTIONS(3618), 1,
      aux_sym_number_token2,
    ACTIONS(3622), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3632), 1,
      sym__brace_start,
    ACTIONS(4131), 1,
      sym_word,
    ACTIONS(4135), 1,
      sym_test_operator,
    ACTIONS(8850), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8854), 1,
      anon_sym_DQUOTE,
    ACTIONS(8858), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8860), 1,
      anon_sym_BQUOTE,
    ACTIONS(8862), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9086), 1,
      sym__special_character,
    STATE(2270), 1,
      aux_sym__literal_repeat1,
    ACTIONS(8848), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8864), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9088), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(786), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    STATE(1968), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [200025] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1414), 1,
      anon_sym_DOLLAR,
    ACTIONS(1420), 1,
      aux_sym_number_token1,
    ACTIONS(1422), 1,
      aux_sym_number_token2,
    ACTIONS(1426), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1438), 1,
      sym__brace_start,
    ACTIONS(9090), 1,
      sym_word,
    ACTIONS(9094), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9096), 1,
      sym__special_character,
    ACTIONS(9098), 1,
      anon_sym_DQUOTE,
    ACTIONS(9102), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9104), 1,
      anon_sym_BQUOTE,
    ACTIONS(9106), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9110), 1,
      sym_test_operator,
    ACTIONS(9112), 1,
      sym_regex,
    STATE(1014), 1,
      aux_sym__literal_repeat1,
    STATE(1168), 1,
      sym_concatenation,
    ACTIONS(9092), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9100), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(9108), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(854), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [200100] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3612), 1,
      anon_sym_DOLLAR,
    ACTIONS(3616), 1,
      aux_sym_number_token1,
    ACTIONS(3618), 1,
      aux_sym_number_token2,
    ACTIONS(3622), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3632), 1,
      sym__brace_start,
    ACTIONS(4131), 1,
      sym_word,
    ACTIONS(4135), 1,
      sym_test_operator,
    ACTIONS(8850), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8854), 1,
      anon_sym_DQUOTE,
    ACTIONS(8858), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8860), 1,
      anon_sym_BQUOTE,
    ACTIONS(8862), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9086), 1,
      sym__special_character,
    STATE(2270), 1,
      aux_sym__literal_repeat1,
    ACTIONS(8848), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8864), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9088), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(791), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    STATE(1968), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [200173] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9114), 1,
      sym_word,
    ACTIONS(9118), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9120), 1,
      anon_sym_DOLLAR,
    ACTIONS(9122), 1,
      sym__special_character,
    ACTIONS(9124), 1,
      anon_sym_DQUOTE,
    ACTIONS(9128), 1,
      aux_sym_number_token1,
    ACTIONS(9130), 1,
      aux_sym_number_token2,
    ACTIONS(9132), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9134), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9136), 1,
      anon_sym_BQUOTE,
    ACTIONS(9138), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9142), 1,
      sym_test_operator,
    ACTIONS(9144), 1,
      sym__brace_start,
    STATE(2457), 1,
      aux_sym__literal_repeat1,
    ACTIONS(9116), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9126), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(9140), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(823), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    STATE(2271), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [200246] = 19,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(363), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(367), 1,
      anon_sym_DOLLAR,
    ACTIONS(371), 1,
      anon_sym_DQUOTE,
    ACTIONS(375), 1,
      aux_sym_number_token1,
    ACTIONS(377), 1,
      aux_sym_number_token2,
    ACTIONS(379), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(381), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(385), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(391), 1,
      sym__brace_start,
    ACTIONS(5809), 1,
      anon_sym_BQUOTE,
    ACTIONS(8836), 1,
      sym_word,
    ACTIONS(8844), 1,
      sym_test_operator,
    ACTIONS(9146), 1,
      anon_sym_RBRACK,
    ACTIONS(352), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(387), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(8840), 2,
      sym__special_character,
      sym__comment_word,
    ACTIONS(8842), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2730), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [200317] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1241), 1,
      sym_file_descriptor,
    ACTIONS(9150), 1,
      anon_sym_DQUOTE,
    ACTIONS(9154), 1,
      sym_variable_name,
    STATE(5537), 1,
      sym_string,
    ACTIONS(9152), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(9148), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1239), 16,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [200366] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9114), 1,
      sym_word,
    ACTIONS(9118), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9120), 1,
      anon_sym_DOLLAR,
    ACTIONS(9122), 1,
      sym__special_character,
    ACTIONS(9124), 1,
      anon_sym_DQUOTE,
    ACTIONS(9128), 1,
      aux_sym_number_token1,
    ACTIONS(9130), 1,
      aux_sym_number_token2,
    ACTIONS(9132), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9134), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9136), 1,
      anon_sym_BQUOTE,
    ACTIONS(9138), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9142), 1,
      sym_test_operator,
    ACTIONS(9144), 1,
      sym__brace_start,
    STATE(2457), 1,
      aux_sym__literal_repeat1,
    ACTIONS(9116), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9126), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(9140), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(821), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    STATE(2271), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [200439] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1235), 1,
      sym_file_descriptor,
    ACTIONS(9150), 1,
      anon_sym_DQUOTE,
    ACTIONS(9154), 1,
      sym_variable_name,
    STATE(5537), 1,
      sym_string,
    ACTIONS(9152), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(9148), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1227), 16,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [200488] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(4498), 1,
      sym_word,
    ACTIONS(4502), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(4504), 1,
      anon_sym_DOLLAR,
    ACTIONS(4508), 1,
      anon_sym_DQUOTE,
    ACTIONS(4512), 1,
      aux_sym_number_token1,
    ACTIONS(4514), 1,
      aux_sym_number_token2,
    ACTIONS(4516), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(4518), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(4520), 1,
      anon_sym_BQUOTE,
    ACTIONS(4522), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(4526), 1,
      sym_test_operator,
    ACTIONS(4528), 1,
      sym__brace_start,
    ACTIONS(9156), 1,
      sym__special_character,
    STATE(2715), 1,
      aux_sym__literal_repeat1,
    ACTIONS(4500), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(4510), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(4524), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(867), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    STATE(2438), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [200561] = 19,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(363), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(367), 1,
      anon_sym_DOLLAR,
    ACTIONS(371), 1,
      anon_sym_DQUOTE,
    ACTIONS(375), 1,
      aux_sym_number_token1,
    ACTIONS(377), 1,
      aux_sym_number_token2,
    ACTIONS(379), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(381), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(385), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(391), 1,
      sym__brace_start,
    ACTIONS(5809), 1,
      anon_sym_BQUOTE,
    ACTIONS(8836), 1,
      sym_word,
    ACTIONS(8844), 1,
      sym_test_operator,
    ACTIONS(9158), 1,
      anon_sym_RBRACK,
    ACTIONS(352), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(387), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(8840), 2,
      sym__special_character,
      sym__comment_word,
    ACTIONS(8842), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2730), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [200632] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(4498), 1,
      sym_word,
    ACTIONS(4502), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(4504), 1,
      anon_sym_DOLLAR,
    ACTIONS(4508), 1,
      anon_sym_DQUOTE,
    ACTIONS(4512), 1,
      aux_sym_number_token1,
    ACTIONS(4514), 1,
      aux_sym_number_token2,
    ACTIONS(4516), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(4518), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(4520), 1,
      anon_sym_BQUOTE,
    ACTIONS(4522), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(4526), 1,
      sym_test_operator,
    ACTIONS(4528), 1,
      sym__brace_start,
    ACTIONS(9156), 1,
      sym__special_character,
    STATE(2715), 1,
      aux_sym__literal_repeat1,
    ACTIONS(4500), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(4510), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(4524), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(879), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    STATE(2438), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [200705] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3391), 1,
      sym_word,
    ACTIONS(3397), 1,
      anon_sym_DOLLAR,
    ACTIONS(3401), 1,
      aux_sym_number_token1,
    ACTIONS(3403), 1,
      aux_sym_number_token2,
    ACTIONS(3407), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3415), 1,
      sym_test_operator,
    ACTIONS(3417), 1,
      sym__brace_start,
    ACTIONS(8926), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8930), 1,
      anon_sym_DQUOTE,
    ACTIONS(8934), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8936), 1,
      anon_sym_BQUOTE,
    ACTIONS(8938), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9160), 1,
      sym__special_character,
    STATE(2161), 1,
      aux_sym__literal_repeat1,
    ACTIONS(8924), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8940), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9162), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(691), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    STATE(1711), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [200778] = 19,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(363), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(367), 1,
      anon_sym_DOLLAR,
    ACTIONS(371), 1,
      anon_sym_DQUOTE,
    ACTIONS(375), 1,
      aux_sym_number_token1,
    ACTIONS(377), 1,
      aux_sym_number_token2,
    ACTIONS(379), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(381), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(385), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(391), 1,
      sym__brace_start,
    ACTIONS(5809), 1,
      anon_sym_BQUOTE,
    ACTIONS(8836), 1,
      sym_word,
    ACTIONS(8844), 1,
      sym_test_operator,
    ACTIONS(9164), 1,
      anon_sym_RBRACK,
    ACTIONS(352), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(387), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(8840), 2,
      sym__special_character,
      sym__comment_word,
    ACTIONS(8842), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2730), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [200849] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(2636), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(2638), 1,
      anon_sym_DOLLAR,
    ACTIONS(2640), 1,
      sym__special_character,
    ACTIONS(2642), 1,
      anon_sym_DQUOTE,
    ACTIONS(2646), 1,
      aux_sym_number_token1,
    ACTIONS(2648), 1,
      aux_sym_number_token2,
    ACTIONS(2650), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(2652), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2654), 1,
      anon_sym_BQUOTE,
    ACTIONS(2656), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(2664), 1,
      sym__brace_start,
    ACTIONS(9166), 1,
      sym_word,
    ACTIONS(9170), 1,
      sym_test_operator,
    ACTIONS(9172), 1,
      sym_regex,
    STATE(2394), 1,
      aux_sym__literal_repeat1,
    STATE(2556), 1,
      sym_concatenation,
    ACTIONS(2630), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(2658), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9168), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2199), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [200924] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(8474), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8476), 1,
      anon_sym_DOLLAR,
    ACTIONS(8478), 1,
      sym__special_character,
    ACTIONS(8480), 1,
      anon_sym_DQUOTE,
    ACTIONS(8484), 1,
      aux_sym_number_token1,
    ACTIONS(8486), 1,
      aux_sym_number_token2,
    ACTIONS(8488), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8490), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(8492), 1,
      anon_sym_BQUOTE,
    ACTIONS(8494), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8504), 1,
      sym__brace_start,
    ACTIONS(9174), 1,
      sym_word,
    ACTIONS(9178), 1,
      sym_test_operator,
    STATE(1801), 1,
      aux_sym__literal_repeat1,
    ACTIONS(8470), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8496), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9176), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(654), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    STATE(1399), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [200997] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3391), 1,
      sym_word,
    ACTIONS(3397), 1,
      anon_sym_DOLLAR,
    ACTIONS(3401), 1,
      aux_sym_number_token1,
    ACTIONS(3403), 1,
      aux_sym_number_token2,
    ACTIONS(3407), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3415), 1,
      sym_test_operator,
    ACTIONS(3417), 1,
      sym__brace_start,
    ACTIONS(8926), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8930), 1,
      anon_sym_DQUOTE,
    ACTIONS(8934), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8936), 1,
      anon_sym_BQUOTE,
    ACTIONS(8938), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9160), 1,
      sym__special_character,
    STATE(2161), 1,
      aux_sym__literal_repeat1,
    ACTIONS(8924), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8940), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9162), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(684), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    STATE(1711), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [201070] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7996), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(7998), 1,
      anon_sym_DOLLAR,
    ACTIONS(8000), 1,
      sym__special_character,
    ACTIONS(8002), 1,
      anon_sym_DQUOTE,
    ACTIONS(8006), 1,
      aux_sym_number_token1,
    ACTIONS(8008), 1,
      aux_sym_number_token2,
    ACTIONS(8010), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8012), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(8014), 1,
      anon_sym_BQUOTE,
    ACTIONS(8016), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8026), 1,
      sym__brace_start,
    ACTIONS(9180), 1,
      sym_word,
    ACTIONS(9184), 1,
      sym_test_operator,
    STATE(3543), 1,
      aux_sym__literal_repeat1,
    ACTIONS(7992), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8018), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9182), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1831), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    STATE(3480), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [201143] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3731), 1,
      anon_sym_DOLLAR,
    ACTIONS(3737), 1,
      aux_sym_number_token1,
    ACTIONS(3739), 1,
      aux_sym_number_token2,
    ACTIONS(3743), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3755), 1,
      sym__brace_start,
    ACTIONS(7759), 1,
      sym_word,
    ACTIONS(7767), 1,
      sym_test_operator,
    ACTIONS(8820), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8822), 1,
      sym__special_character,
    ACTIONS(8824), 1,
      anon_sym_DQUOTE,
    ACTIONS(8828), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8830), 1,
      anon_sym_BQUOTE,
    ACTIONS(8832), 1,
      anon_sym_DOLLAR_BQUOTE,
    STATE(4975), 1,
      aux_sym__literal_repeat1,
    ACTIONS(8818), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8826), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(8834), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3458), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    STATE(4460), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [201216] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7996), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(7998), 1,
      anon_sym_DOLLAR,
    ACTIONS(8000), 1,
      sym__special_character,
    ACTIONS(8002), 1,
      anon_sym_DQUOTE,
    ACTIONS(8006), 1,
      aux_sym_number_token1,
    ACTIONS(8008), 1,
      aux_sym_number_token2,
    ACTIONS(8010), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8012), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(8014), 1,
      anon_sym_BQUOTE,
    ACTIONS(8016), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8026), 1,
      sym__brace_start,
    ACTIONS(9180), 1,
      sym_word,
    ACTIONS(9184), 1,
      sym_test_operator,
    STATE(3543), 1,
      aux_sym__literal_repeat1,
    ACTIONS(7992), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8018), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9182), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1808), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    STATE(3480), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [201289] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(8178), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8180), 1,
      anon_sym_DOLLAR,
    ACTIONS(8184), 1,
      anon_sym_DQUOTE,
    ACTIONS(8188), 1,
      aux_sym_number_token1,
    ACTIONS(8190), 1,
      aux_sym_number_token2,
    ACTIONS(8192), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8194), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(8196), 1,
      anon_sym_BQUOTE,
    ACTIONS(8198), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8208), 1,
      sym__brace_start,
    ACTIONS(8450), 1,
      sym__special_character,
    ACTIONS(9186), 1,
      sym_word,
    ACTIONS(9190), 1,
      sym_test_operator,
    STATE(1744), 1,
      aux_sym__literal_repeat1,
    ACTIONS(8174), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8200), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9188), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(686), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    STATE(1688), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [201362] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(8178), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8180), 1,
      anon_sym_DOLLAR,
    ACTIONS(8184), 1,
      anon_sym_DQUOTE,
    ACTIONS(8188), 1,
      aux_sym_number_token1,
    ACTIONS(8190), 1,
      aux_sym_number_token2,
    ACTIONS(8192), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8194), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(8196), 1,
      anon_sym_BQUOTE,
    ACTIONS(8198), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8208), 1,
      sym__brace_start,
    ACTIONS(8450), 1,
      sym__special_character,
    ACTIONS(9186), 1,
      sym_word,
    ACTIONS(9190), 1,
      sym_test_operator,
    STATE(1744), 1,
      aux_sym__literal_repeat1,
    ACTIONS(8174), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8200), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9188), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(673), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    STATE(1688), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [201435] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(4943), 1,
      sym_word,
    ACTIONS(4947), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(4949), 1,
      anon_sym_DOLLAR,
    ACTIONS(4953), 1,
      anon_sym_DQUOTE,
    ACTIONS(4957), 1,
      aux_sym_number_token1,
    ACTIONS(4959), 1,
      aux_sym_number_token2,
    ACTIONS(4961), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(4963), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(4965), 1,
      anon_sym_BQUOTE,
    ACTIONS(4967), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(4971), 1,
      sym_test_operator,
    ACTIONS(4973), 1,
      sym__brace_start,
    ACTIONS(9192), 1,
      sym__special_character,
    STATE(2776), 1,
      aux_sym__literal_repeat1,
    ACTIONS(4945), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(4955), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(4969), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(971), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    STATE(2489), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [201508] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3612), 1,
      anon_sym_DOLLAR,
    ACTIONS(3616), 1,
      aux_sym_number_token1,
    ACTIONS(3618), 1,
      aux_sym_number_token2,
    ACTIONS(3622), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3632), 1,
      sym__brace_start,
    ACTIONS(4131), 1,
      sym_word,
    ACTIONS(4135), 1,
      sym_test_operator,
    ACTIONS(8850), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8854), 1,
      anon_sym_DQUOTE,
    ACTIONS(8858), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8860), 1,
      anon_sym_BQUOTE,
    ACTIONS(8862), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9086), 1,
      sym__special_character,
    STATE(2270), 1,
      aux_sym__literal_repeat1,
    ACTIONS(8848), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8864), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9088), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(788), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    STATE(1968), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [201581] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1551), 1,
      anon_sym_DOLLAR,
    ACTIONS(1557), 1,
      aux_sym_number_token1,
    ACTIONS(1559), 1,
      aux_sym_number_token2,
    ACTIONS(1563), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1575), 1,
      sym__brace_start,
    ACTIONS(9194), 1,
      sym_word,
    ACTIONS(9198), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9200), 1,
      sym__special_character,
    ACTIONS(9202), 1,
      anon_sym_DQUOTE,
    ACTIONS(9206), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9208), 1,
      anon_sym_BQUOTE,
    ACTIONS(9210), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9214), 1,
      sym_test_operator,
    ACTIONS(9216), 1,
      sym_regex,
    STATE(1379), 1,
      aux_sym__literal_repeat1,
    STATE(1611), 1,
      sym_concatenation,
    ACTIONS(9196), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9204), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(9212), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(1244), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [201656] = 19,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(363), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(367), 1,
      anon_sym_DOLLAR,
    ACTIONS(371), 1,
      anon_sym_DQUOTE,
    ACTIONS(375), 1,
      aux_sym_number_token1,
    ACTIONS(377), 1,
      aux_sym_number_token2,
    ACTIONS(379), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(381), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(385), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(391), 1,
      sym__brace_start,
    ACTIONS(5809), 1,
      anon_sym_BQUOTE,
    ACTIONS(8836), 1,
      sym_word,
    ACTIONS(8844), 1,
      sym_test_operator,
    ACTIONS(9218), 1,
      anon_sym_RBRACK,
    ACTIONS(352), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(387), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(8840), 2,
      sym__special_character,
      sym__comment_word,
    ACTIONS(8842), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2730), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [201727] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3397), 1,
      anon_sym_DOLLAR,
    ACTIONS(3401), 1,
      aux_sym_number_token1,
    ACTIONS(3403), 1,
      aux_sym_number_token2,
    ACTIONS(3407), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3417), 1,
      sym__brace_start,
    ACTIONS(3663), 1,
      sym_word,
    ACTIONS(3667), 1,
      sym_test_operator,
    ACTIONS(8926), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8930), 1,
      anon_sym_DQUOTE,
    ACTIONS(8934), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8936), 1,
      anon_sym_BQUOTE,
    ACTIONS(8938), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9034), 1,
      sym__special_character,
    STATE(2161), 1,
      aux_sym__literal_repeat1,
    ACTIONS(8924), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8940), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9036), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(720), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    STATE(1918), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [201800] = 19,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(363), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(367), 1,
      anon_sym_DOLLAR,
    ACTIONS(371), 1,
      anon_sym_DQUOTE,
    ACTIONS(375), 1,
      aux_sym_number_token1,
    ACTIONS(377), 1,
      aux_sym_number_token2,
    ACTIONS(379), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(381), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(385), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(391), 1,
      sym__brace_start,
    ACTIONS(5809), 1,
      anon_sym_BQUOTE,
    ACTIONS(8836), 1,
      sym_word,
    ACTIONS(8844), 1,
      sym_test_operator,
    ACTIONS(9220), 1,
      anon_sym_RBRACK,
    ACTIONS(352), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(387), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(8840), 2,
      sym__special_character,
      sym__comment_word,
    ACTIONS(8842), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2730), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [201871] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1551), 1,
      anon_sym_DOLLAR,
    ACTIONS(1557), 1,
      aux_sym_number_token1,
    ACTIONS(1559), 1,
      aux_sym_number_token2,
    ACTIONS(1563), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1575), 1,
      sym__brace_start,
    ACTIONS(9198), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9202), 1,
      anon_sym_DQUOTE,
    ACTIONS(9206), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9208), 1,
      anon_sym_BQUOTE,
    ACTIONS(9210), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9216), 1,
      sym_regex,
    ACTIONS(9222), 1,
      sym_word,
    ACTIONS(9224), 1,
      sym__special_character,
    ACTIONS(9228), 1,
      sym_test_operator,
    STATE(1379), 1,
      aux_sym__literal_repeat1,
    STATE(1611), 1,
      sym_concatenation,
    ACTIONS(9196), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9212), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9226), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1124), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [201946] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3333), 1,
      sym_word,
    ACTIONS(3339), 1,
      anon_sym_DOLLAR,
    ACTIONS(3343), 1,
      aux_sym_number_token1,
    ACTIONS(3345), 1,
      aux_sym_number_token2,
    ACTIONS(3349), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3357), 1,
      sym_test_operator,
    ACTIONS(3359), 1,
      sym__brace_start,
    ACTIONS(9046), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9048), 1,
      sym__special_character,
    ACTIONS(9050), 1,
      anon_sym_DQUOTE,
    ACTIONS(9054), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9056), 1,
      anon_sym_BQUOTE,
    ACTIONS(9058), 1,
      anon_sym_DOLLAR_BQUOTE,
    STATE(2147), 1,
      aux_sym__literal_repeat1,
    ACTIONS(9044), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9052), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(9060), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(680), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    STATE(1526), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [202019] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(8178), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8180), 1,
      anon_sym_DOLLAR,
    ACTIONS(8182), 1,
      sym__special_character,
    ACTIONS(8184), 1,
      anon_sym_DQUOTE,
    ACTIONS(8188), 1,
      aux_sym_number_token1,
    ACTIONS(8190), 1,
      aux_sym_number_token2,
    ACTIONS(8192), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8194), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(8196), 1,
      anon_sym_BQUOTE,
    ACTIONS(8198), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8208), 1,
      sym__brace_start,
    ACTIONS(9230), 1,
      sym_word,
    ACTIONS(9234), 1,
      sym_test_operator,
    STATE(1744), 1,
      aux_sym__literal_repeat1,
    ACTIONS(8174), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8200), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9232), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(816), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    STATE(2337), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [202092] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(4943), 1,
      sym_word,
    ACTIONS(4947), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(4949), 1,
      anon_sym_DOLLAR,
    ACTIONS(4953), 1,
      anon_sym_DQUOTE,
    ACTIONS(4957), 1,
      aux_sym_number_token1,
    ACTIONS(4959), 1,
      aux_sym_number_token2,
    ACTIONS(4961), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(4963), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(4965), 1,
      anon_sym_BQUOTE,
    ACTIONS(4967), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(4971), 1,
      sym_test_operator,
    ACTIONS(4973), 1,
      sym__brace_start,
    ACTIONS(9192), 1,
      sym__special_character,
    STATE(2776), 1,
      aux_sym__literal_repeat1,
    ACTIONS(4945), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(4955), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(4969), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(981), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    STATE(2489), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [202165] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(8474), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8476), 1,
      anon_sym_DOLLAR,
    ACTIONS(8478), 1,
      sym__special_character,
    ACTIONS(8480), 1,
      anon_sym_DQUOTE,
    ACTIONS(8484), 1,
      aux_sym_number_token1,
    ACTIONS(8486), 1,
      aux_sym_number_token2,
    ACTIONS(8488), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8490), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(8492), 1,
      anon_sym_BQUOTE,
    ACTIONS(8494), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8504), 1,
      sym__brace_start,
    ACTIONS(9174), 1,
      sym_word,
    ACTIONS(9178), 1,
      sym_test_operator,
    STATE(1801), 1,
      aux_sym__literal_repeat1,
    ACTIONS(8470), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8496), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9176), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(642), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    STATE(1399), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [202238] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(8178), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8180), 1,
      anon_sym_DOLLAR,
    ACTIONS(8182), 1,
      sym__special_character,
    ACTIONS(8184), 1,
      anon_sym_DQUOTE,
    ACTIONS(8188), 1,
      aux_sym_number_token1,
    ACTIONS(8190), 1,
      aux_sym_number_token2,
    ACTIONS(8192), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8194), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(8196), 1,
      anon_sym_BQUOTE,
    ACTIONS(8198), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8208), 1,
      sym__brace_start,
    ACTIONS(9230), 1,
      sym_word,
    ACTIONS(9234), 1,
      sym_test_operator,
    STATE(1744), 1,
      aux_sym__literal_repeat1,
    ACTIONS(8174), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8200), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9232), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(806), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    STATE(2337), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [202311] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3731), 1,
      anon_sym_DOLLAR,
    ACTIONS(3737), 1,
      aux_sym_number_token1,
    ACTIONS(3739), 1,
      aux_sym_number_token2,
    ACTIONS(3743), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3755), 1,
      sym__brace_start,
    ACTIONS(7759), 1,
      sym_word,
    ACTIONS(7767), 1,
      sym_test_operator,
    ACTIONS(8820), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8822), 1,
      sym__special_character,
    ACTIONS(8824), 1,
      anon_sym_DQUOTE,
    ACTIONS(8828), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8830), 1,
      anon_sym_BQUOTE,
    ACTIONS(8832), 1,
      anon_sym_DOLLAR_BQUOTE,
    STATE(4975), 1,
      aux_sym__literal_repeat1,
    ACTIONS(8818), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8826), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(8834), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(3475), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    STATE(4460), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [202384] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(8178), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8180), 1,
      anon_sym_DOLLAR,
    ACTIONS(8184), 1,
      anon_sym_DQUOTE,
    ACTIONS(8188), 1,
      aux_sym_number_token1,
    ACTIONS(8190), 1,
      aux_sym_number_token2,
    ACTIONS(8192), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8194), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(8196), 1,
      anon_sym_BQUOTE,
    ACTIONS(8198), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8208), 1,
      sym__brace_start,
    ACTIONS(8220), 1,
      sym__special_character,
    ACTIONS(8978), 1,
      sym_word,
    ACTIONS(8982), 1,
      sym_test_operator,
    STATE(1744), 1,
      aux_sym__literal_repeat1,
    ACTIONS(8174), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8200), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(8980), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(649), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    STATE(1370), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [202457] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3397), 1,
      anon_sym_DOLLAR,
    ACTIONS(3401), 1,
      aux_sym_number_token1,
    ACTIONS(3403), 1,
      aux_sym_number_token2,
    ACTIONS(3407), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3417), 1,
      sym__brace_start,
    ACTIONS(3663), 1,
      sym_word,
    ACTIONS(3667), 1,
      sym_test_operator,
    ACTIONS(8926), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8930), 1,
      anon_sym_DQUOTE,
    ACTIONS(8934), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8936), 1,
      anon_sym_BQUOTE,
    ACTIONS(8938), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9034), 1,
      sym__special_character,
    STATE(2161), 1,
      aux_sym__literal_repeat1,
    ACTIONS(8924), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8940), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9036), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(719), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    STATE(1918), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [202530] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1587), 1,
      anon_sym_DOLLAR,
    ACTIONS(1593), 1,
      aux_sym_number_token1,
    ACTIONS(1595), 1,
      aux_sym_number_token2,
    ACTIONS(1599), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1611), 1,
      sym__brace_start,
    ACTIONS(9236), 1,
      sym_word,
    ACTIONS(9240), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9242), 1,
      sym__special_character,
    ACTIONS(9244), 1,
      anon_sym_DQUOTE,
    ACTIONS(9248), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9250), 1,
      anon_sym_BQUOTE,
    ACTIONS(9252), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9256), 1,
      sym_test_operator,
    ACTIONS(9258), 1,
      sym_regex,
    STATE(1335), 1,
      aux_sym__literal_repeat1,
    STATE(1673), 1,
      sym_concatenation,
    ACTIONS(9238), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9246), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(9254), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(1039), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [202605] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3612), 1,
      anon_sym_DOLLAR,
    ACTIONS(3616), 1,
      aux_sym_number_token1,
    ACTIONS(3618), 1,
      aux_sym_number_token2,
    ACTIONS(3622), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3632), 1,
      sym__brace_start,
    ACTIONS(4131), 1,
      sym_word,
    ACTIONS(4135), 1,
      sym_test_operator,
    ACTIONS(8850), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8854), 1,
      anon_sym_DQUOTE,
    ACTIONS(8858), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8860), 1,
      anon_sym_BQUOTE,
    ACTIONS(8862), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9086), 1,
      sym__special_character,
    STATE(2270), 1,
      aux_sym__literal_repeat1,
    ACTIONS(8848), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8864), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9088), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(789), 2,
      sym_concatenation,
      aux_sym_for_statement_repeat1,
    STATE(1968), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [202678] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1551), 1,
      anon_sym_DOLLAR,
    ACTIONS(1557), 1,
      aux_sym_number_token1,
    ACTIONS(1559), 1,
      aux_sym_number_token2,
    ACTIONS(1563), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1575), 1,
      sym__brace_start,
    ACTIONS(9198), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9202), 1,
      anon_sym_DQUOTE,
    ACTIONS(9206), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9208), 1,
      anon_sym_BQUOTE,
    ACTIONS(9210), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9216), 1,
      sym_regex,
    ACTIONS(9260), 1,
      sym_word,
    ACTIONS(9262), 1,
      sym__special_character,
    ACTIONS(9266), 1,
      sym_test_operator,
    STATE(1379), 1,
      aux_sym__literal_repeat1,
    STATE(1611), 1,
      sym_concatenation,
    ACTIONS(9196), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9212), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9264), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1784), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [202753] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1557), 1,
      aux_sym_number_token1,
    ACTIONS(1559), 1,
      aux_sym_number_token2,
    ACTIONS(1563), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1575), 1,
      sym__brace_start,
    ACTIONS(9198), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9202), 1,
      anon_sym_DQUOTE,
    ACTIONS(9206), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9208), 1,
      anon_sym_BQUOTE,
    ACTIONS(9210), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9268), 1,
      sym_word,
    ACTIONS(9270), 1,
      anon_sym_DOLLAR,
    ACTIONS(9276), 1,
      sym__comment_word,
    ACTIONS(9196), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9212), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9272), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9274), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1189), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [202821] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3064), 1,
      sym_variable_name,
    ACTIONS(9278), 1,
      anon_sym_LPAREN,
    ACTIONS(9280), 1,
      anon_sym_BANG,
    ACTIONS(9286), 1,
      anon_sym_TILDE,
    ACTIONS(9288), 1,
      anon_sym_DOLLAR,
    ACTIONS(9290), 1,
      anon_sym_DQUOTE,
    ACTIONS(9292), 1,
      aux_sym_number_token1,
    ACTIONS(9294), 1,
      aux_sym_number_token2,
    ACTIONS(9296), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9298), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9300), 1,
      anon_sym_BQUOTE,
    ACTIONS(9302), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9304), 1,
      aux_sym__simple_variable_name_token1,
    STATE(2801), 1,
      sym__arithmetic_binary_expression,
    STATE(2803), 1,
      sym__arithmetic_ternary_expression,
    STATE(2809), 1,
      sym__arithmetic_unary_expression,
    STATE(2825), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9282), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9284), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(2790), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [202895] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4086), 1,
      aux_sym_number_token1,
    ACTIONS(4088), 1,
      aux_sym_number_token2,
    ACTIONS(4092), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(4102), 1,
      sym__brace_start,
    ACTIONS(9306), 1,
      sym_word,
    ACTIONS(9310), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9312), 1,
      anon_sym_DOLLAR,
    ACTIONS(9316), 1,
      anon_sym_DQUOTE,
    ACTIONS(9320), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9322), 1,
      anon_sym_BQUOTE,
    ACTIONS(9324), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9328), 1,
      sym__comment_word,
    ACTIONS(9308), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9314), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9326), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9318), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(4566), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [202963] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4633), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(4639), 1,
      anon_sym_DQUOTE,
    ACTIONS(4643), 1,
      aux_sym_number_token1,
    ACTIONS(4645), 1,
      aux_sym_number_token2,
    ACTIONS(4647), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(4649), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(4651), 1,
      anon_sym_BQUOTE,
    ACTIONS(4653), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(4659), 1,
      sym__brace_start,
    ACTIONS(9330), 1,
      sym_word,
    ACTIONS(9332), 1,
      anon_sym_DOLLAR,
    ACTIONS(9338), 1,
      sym__comment_word,
    ACTIONS(4631), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(4655), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9334), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9336), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2634), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [203031] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(111), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(113), 1,
      anon_sym_DOLLAR,
    ACTIONS(117), 1,
      anon_sym_DQUOTE,
    ACTIONS(121), 1,
      aux_sym_number_token1,
    ACTIONS(123), 1,
      aux_sym_number_token2,
    ACTIONS(125), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(127), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(129), 1,
      anon_sym_BQUOTE,
    ACTIONS(131), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(141), 1,
      sym__brace_start,
    ACTIONS(9340), 1,
      sym_word,
    ACTIONS(9346), 1,
      sym__comment_word,
    ACTIONS(109), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(133), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9342), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9344), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(482), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [203099] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(5315), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(5317), 1,
      anon_sym_DOLLAR,
    ACTIONS(5321), 1,
      anon_sym_DQUOTE,
    ACTIONS(5325), 1,
      aux_sym_number_token1,
    ACTIONS(5327), 1,
      aux_sym_number_token2,
    ACTIONS(5329), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(5331), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(5333), 1,
      anon_sym_BQUOTE,
    ACTIONS(5335), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(5341), 1,
      sym__brace_start,
    ACTIONS(9348), 1,
      sym_word,
    ACTIONS(9350), 1,
      sym__special_character,
    ACTIONS(9354), 1,
      sym_test_operator,
    STATE(5498), 1,
      aux_sym__literal_repeat1,
    STATE(5555), 1,
      sym_concatenation,
    ACTIONS(5313), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(5337), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9352), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(5347), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [203171] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(7996), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(7998), 1,
      anon_sym_DOLLAR,
    ACTIONS(8002), 1,
      anon_sym_DQUOTE,
    ACTIONS(8006), 1,
      aux_sym_number_token1,
    ACTIONS(8008), 1,
      aux_sym_number_token2,
    ACTIONS(8010), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8012), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(8014), 1,
      anon_sym_BQUOTE,
    ACTIONS(8016), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8026), 1,
      sym__brace_start,
    ACTIONS(9356), 1,
      sym_word,
    ACTIONS(9362), 1,
      sym__comment_word,
    ACTIONS(7992), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8018), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9358), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9360), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(3510), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [203239] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9364), 1,
      anon_sym_LPAREN,
    ACTIONS(9366), 1,
      anon_sym_BANG,
    ACTIONS(9372), 1,
      anon_sym_TILDE,
    ACTIONS(9374), 1,
      anon_sym_DOLLAR,
    ACTIONS(9376), 1,
      anon_sym_DQUOTE,
    ACTIONS(9378), 1,
      aux_sym_number_token1,
    ACTIONS(9380), 1,
      aux_sym_number_token2,
    ACTIONS(9382), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9384), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9386), 1,
      anon_sym_BQUOTE,
    ACTIONS(9388), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9390), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(9392), 1,
      sym_variable_name,
    STATE(3391), 1,
      sym__arithmetic_binary_expression,
    STATE(3396), 1,
      sym__arithmetic_ternary_expression,
    STATE(3398), 1,
      sym__arithmetic_unary_expression,
    STATE(3400), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9368), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9370), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(3357), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [203313] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(2636), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(2638), 1,
      anon_sym_DOLLAR,
    ACTIONS(2640), 1,
      sym__special_character,
    ACTIONS(2642), 1,
      anon_sym_DQUOTE,
    ACTIONS(2646), 1,
      aux_sym_number_token1,
    ACTIONS(2648), 1,
      aux_sym_number_token2,
    ACTIONS(2650), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(2652), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2654), 1,
      anon_sym_BQUOTE,
    ACTIONS(2656), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(2664), 1,
      sym__brace_start,
    ACTIONS(9394), 1,
      sym_word,
    ACTIONS(9398), 1,
      sym_test_operator,
    STATE(2352), 1,
      aux_sym__literal_repeat1,
    STATE(2522), 1,
      sym_concatenation,
    ACTIONS(2630), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(2658), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9396), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2125), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [203385] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(314), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(316), 1,
      anon_sym_DOLLAR,
    ACTIONS(320), 1,
      anon_sym_DQUOTE,
    ACTIONS(324), 1,
      aux_sym_number_token1,
    ACTIONS(326), 1,
      aux_sym_number_token2,
    ACTIONS(328), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(330), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(332), 1,
      anon_sym_BQUOTE,
    ACTIONS(334), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(344), 1,
      sym__brace_start,
    ACTIONS(9400), 1,
      sym_word,
    ACTIONS(9406), 1,
      sym__comment_word,
    ACTIONS(312), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(336), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9402), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9404), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1037), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [203453] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3064), 1,
      sym_variable_name,
    ACTIONS(9278), 1,
      anon_sym_LPAREN,
    ACTIONS(9280), 1,
      anon_sym_BANG,
    ACTIONS(9286), 1,
      anon_sym_TILDE,
    ACTIONS(9288), 1,
      anon_sym_DOLLAR,
    ACTIONS(9290), 1,
      anon_sym_DQUOTE,
    ACTIONS(9292), 1,
      aux_sym_number_token1,
    ACTIONS(9294), 1,
      aux_sym_number_token2,
    ACTIONS(9296), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9298), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9300), 1,
      anon_sym_BQUOTE,
    ACTIONS(9302), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9408), 1,
      aux_sym__simple_variable_name_token1,
    STATE(2801), 1,
      sym__arithmetic_binary_expression,
    STATE(2803), 1,
      sym__arithmetic_ternary_expression,
    STATE(2809), 1,
      sym__arithmetic_unary_expression,
    STATE(2825), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9282), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9284), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(2742), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [203527] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9364), 1,
      anon_sym_LPAREN,
    ACTIONS(9366), 1,
      anon_sym_BANG,
    ACTIONS(9372), 1,
      anon_sym_TILDE,
    ACTIONS(9374), 1,
      anon_sym_DOLLAR,
    ACTIONS(9376), 1,
      anon_sym_DQUOTE,
    ACTIONS(9378), 1,
      aux_sym_number_token1,
    ACTIONS(9380), 1,
      aux_sym_number_token2,
    ACTIONS(9382), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9384), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9386), 1,
      anon_sym_BQUOTE,
    ACTIONS(9388), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9392), 1,
      sym_variable_name,
    ACTIONS(9410), 1,
      aux_sym__simple_variable_name_token1,
    STATE(3391), 1,
      sym__arithmetic_binary_expression,
    STATE(3396), 1,
      sym__arithmetic_ternary_expression,
    STATE(3398), 1,
      sym__arithmetic_unary_expression,
    STATE(3400), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9368), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9370), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(3355), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [203601] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3064), 1,
      sym_variable_name,
    ACTIONS(3469), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(9278), 1,
      anon_sym_LPAREN,
    ACTIONS(9280), 1,
      anon_sym_BANG,
    ACTIONS(9286), 1,
      anon_sym_TILDE,
    ACTIONS(9288), 1,
      anon_sym_DOLLAR,
    ACTIONS(9290), 1,
      anon_sym_DQUOTE,
    ACTIONS(9292), 1,
      aux_sym_number_token1,
    ACTIONS(9294), 1,
      aux_sym_number_token2,
    ACTIONS(9296), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9298), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9300), 1,
      anon_sym_BQUOTE,
    ACTIONS(9302), 1,
      anon_sym_DOLLAR_BQUOTE,
    STATE(2801), 1,
      sym__arithmetic_binary_expression,
    STATE(2803), 1,
      sym__arithmetic_ternary_expression,
    STATE(2809), 1,
      sym__arithmetic_unary_expression,
    STATE(2825), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9282), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9284), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(2743), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [203675] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3737), 1,
      aux_sym_number_token1,
    ACTIONS(3739), 1,
      aux_sym_number_token2,
    ACTIONS(3743), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3755), 1,
      sym__brace_start,
    ACTIONS(8820), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8824), 1,
      anon_sym_DQUOTE,
    ACTIONS(8828), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8830), 1,
      anon_sym_BQUOTE,
    ACTIONS(8832), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9412), 1,
      sym_word,
    ACTIONS(9414), 1,
      anon_sym_DOLLAR,
    ACTIONS(9420), 1,
      sym__comment_word,
    ACTIONS(8818), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8834), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9416), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9418), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(4810), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [203743] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3737), 1,
      aux_sym_number_token1,
    ACTIONS(3739), 1,
      aux_sym_number_token2,
    ACTIONS(3743), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3755), 1,
      sym__brace_start,
    ACTIONS(8820), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8824), 1,
      anon_sym_DQUOTE,
    ACTIONS(8828), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8830), 1,
      anon_sym_BQUOTE,
    ACTIONS(8832), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9412), 1,
      sym_word,
    ACTIONS(9420), 1,
      sym__comment_word,
    ACTIONS(9422), 1,
      anon_sym_DOLLAR,
    ACTIONS(8818), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8834), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9416), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9418), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(4810), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [203811] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9364), 1,
      anon_sym_LPAREN,
    ACTIONS(9366), 1,
      anon_sym_BANG,
    ACTIONS(9372), 1,
      anon_sym_TILDE,
    ACTIONS(9374), 1,
      anon_sym_DOLLAR,
    ACTIONS(9376), 1,
      anon_sym_DQUOTE,
    ACTIONS(9378), 1,
      aux_sym_number_token1,
    ACTIONS(9380), 1,
      aux_sym_number_token2,
    ACTIONS(9382), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9384), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9386), 1,
      anon_sym_BQUOTE,
    ACTIONS(9388), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9392), 1,
      sym_variable_name,
    ACTIONS(9424), 1,
      aux_sym__simple_variable_name_token1,
    STATE(3391), 1,
      sym__arithmetic_binary_expression,
    STATE(3396), 1,
      sym__arithmetic_ternary_expression,
    STATE(3398), 1,
      sym__arithmetic_unary_expression,
    STATE(3400), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9368), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9370), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(3349), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [203885] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3064), 1,
      sym_variable_name,
    ACTIONS(9278), 1,
      anon_sym_LPAREN,
    ACTIONS(9280), 1,
      anon_sym_BANG,
    ACTIONS(9286), 1,
      anon_sym_TILDE,
    ACTIONS(9288), 1,
      anon_sym_DOLLAR,
    ACTIONS(9290), 1,
      anon_sym_DQUOTE,
    ACTIONS(9292), 1,
      aux_sym_number_token1,
    ACTIONS(9294), 1,
      aux_sym_number_token2,
    ACTIONS(9296), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9298), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9300), 1,
      anon_sym_BQUOTE,
    ACTIONS(9302), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9426), 1,
      aux_sym__simple_variable_name_token1,
    STATE(2801), 1,
      sym__arithmetic_binary_expression,
    STATE(2803), 1,
      sym__arithmetic_ternary_expression,
    STATE(2809), 1,
      sym__arithmetic_unary_expression,
    STATE(2825), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9282), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9284), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(2744), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [203959] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(2636), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(2638), 1,
      anon_sym_DOLLAR,
    ACTIONS(2640), 1,
      sym__special_character,
    ACTIONS(2642), 1,
      anon_sym_DQUOTE,
    ACTIONS(2646), 1,
      aux_sym_number_token1,
    ACTIONS(2648), 1,
      aux_sym_number_token2,
    ACTIONS(2650), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(2652), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2654), 1,
      anon_sym_BQUOTE,
    ACTIONS(2656), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(2664), 1,
      sym__brace_start,
    ACTIONS(9428), 1,
      sym_word,
    ACTIONS(9432), 1,
      sym_test_operator,
    STATE(2370), 1,
      aux_sym__literal_repeat1,
    STATE(2534), 1,
      sym_concatenation,
    ACTIONS(2630), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(2658), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9430), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2154), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [204031] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(4767), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(4769), 1,
      anon_sym_DOLLAR,
    ACTIONS(4773), 1,
      anon_sym_DQUOTE,
    ACTIONS(4777), 1,
      aux_sym_number_token1,
    ACTIONS(4779), 1,
      aux_sym_number_token2,
    ACTIONS(4781), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(4783), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(4785), 1,
      anon_sym_BQUOTE,
    ACTIONS(4787), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(4793), 1,
      sym__brace_start,
    ACTIONS(9434), 1,
      sym_word,
    ACTIONS(9436), 1,
      sym__special_character,
    ACTIONS(9440), 1,
      sym_test_operator,
    STATE(5258), 1,
      aux_sym__literal_repeat1,
    STATE(5543), 1,
      sym_concatenation,
    ACTIONS(4765), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(4789), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9438), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(5173), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [204103] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9118), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9120), 1,
      anon_sym_DOLLAR,
    ACTIONS(9124), 1,
      anon_sym_DQUOTE,
    ACTIONS(9128), 1,
      aux_sym_number_token1,
    ACTIONS(9130), 1,
      aux_sym_number_token2,
    ACTIONS(9132), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9134), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9136), 1,
      anon_sym_BQUOTE,
    ACTIONS(9138), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9144), 1,
      sym__brace_start,
    ACTIONS(9442), 1,
      sym_word,
    ACTIONS(9448), 1,
      sym__comment_word,
    ACTIONS(9116), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9140), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9444), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9446), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2401), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [204171] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(5315), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(5317), 1,
      anon_sym_DOLLAR,
    ACTIONS(5321), 1,
      anon_sym_DQUOTE,
    ACTIONS(5325), 1,
      aux_sym_number_token1,
    ACTIONS(5327), 1,
      aux_sym_number_token2,
    ACTIONS(5329), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(5331), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(5333), 1,
      anon_sym_BQUOTE,
    ACTIONS(5335), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(5341), 1,
      sym__brace_start,
    ACTIONS(9350), 1,
      sym__special_character,
    ACTIONS(9450), 1,
      sym_word,
    ACTIONS(9454), 1,
      sym_test_operator,
    STATE(5447), 1,
      aux_sym__literal_repeat1,
    STATE(5549), 1,
      sym_concatenation,
    ACTIONS(5313), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(5337), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9452), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(5278), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [204243] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4086), 1,
      aux_sym_number_token1,
    ACTIONS(4088), 1,
      aux_sym_number_token2,
    ACTIONS(4092), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(4102), 1,
      sym__brace_start,
    ACTIONS(9306), 1,
      sym_word,
    ACTIONS(9310), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9316), 1,
      anon_sym_DQUOTE,
    ACTIONS(9320), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9322), 1,
      anon_sym_BQUOTE,
    ACTIONS(9324), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9328), 1,
      sym__comment_word,
    ACTIONS(9456), 1,
      anon_sym_DOLLAR,
    ACTIONS(9308), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9314), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9326), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9318), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(4566), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [204311] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(4767), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(4769), 1,
      anon_sym_DOLLAR,
    ACTIONS(4773), 1,
      anon_sym_DQUOTE,
    ACTIONS(4777), 1,
      aux_sym_number_token1,
    ACTIONS(4779), 1,
      aux_sym_number_token2,
    ACTIONS(4781), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(4783), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(4785), 1,
      anon_sym_BQUOTE,
    ACTIONS(4787), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(4793), 1,
      sym__brace_start,
    ACTIONS(9436), 1,
      sym__special_character,
    ACTIONS(9458), 1,
      sym_word,
    ACTIONS(9462), 1,
      sym_test_operator,
    STATE(5292), 1,
      aux_sym__literal_repeat1,
    STATE(5526), 1,
      sym_concatenation,
    ACTIONS(4765), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(4789), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9460), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(5140), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [204383] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1551), 1,
      anon_sym_DOLLAR,
    ACTIONS(1557), 1,
      aux_sym_number_token1,
    ACTIONS(1559), 1,
      aux_sym_number_token2,
    ACTIONS(1563), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1575), 1,
      sym__brace_start,
    ACTIONS(9198), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9202), 1,
      anon_sym_DQUOTE,
    ACTIONS(9206), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9208), 1,
      anon_sym_BQUOTE,
    ACTIONS(9210), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9224), 1,
      sym__special_character,
    ACTIONS(9464), 1,
      sym_word,
    ACTIONS(9468), 1,
      sym_test_operator,
    STATE(1401), 1,
      aux_sym__literal_repeat1,
    STATE(1573), 1,
      sym_concatenation,
    ACTIONS(9196), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9212), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9466), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1109), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [204455] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9364), 1,
      anon_sym_LPAREN,
    ACTIONS(9366), 1,
      anon_sym_BANG,
    ACTIONS(9372), 1,
      anon_sym_TILDE,
    ACTIONS(9374), 1,
      anon_sym_DOLLAR,
    ACTIONS(9376), 1,
      anon_sym_DQUOTE,
    ACTIONS(9378), 1,
      aux_sym_number_token1,
    ACTIONS(9380), 1,
      aux_sym_number_token2,
    ACTIONS(9382), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9384), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9386), 1,
      anon_sym_BQUOTE,
    ACTIONS(9388), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9392), 1,
      sym_variable_name,
    ACTIONS(9470), 1,
      aux_sym__simple_variable_name_token1,
    STATE(3391), 1,
      sym__arithmetic_binary_expression,
    STATE(3396), 1,
      sym__arithmetic_ternary_expression,
    STATE(3398), 1,
      sym__arithmetic_unary_expression,
    STATE(3400), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9368), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9370), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(3348), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [204529] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5261), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(5267), 1,
      anon_sym_DQUOTE,
    ACTIONS(5271), 1,
      aux_sym_number_token1,
    ACTIONS(5273), 1,
      aux_sym_number_token2,
    ACTIONS(5275), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(5277), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(5279), 1,
      anon_sym_BQUOTE,
    ACTIONS(5281), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(5287), 1,
      sym__brace_start,
    ACTIONS(9472), 1,
      sym_word,
    ACTIONS(9474), 1,
      anon_sym_DOLLAR,
    ACTIONS(9480), 1,
      sym__comment_word,
    ACTIONS(5259), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(5283), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9476), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9478), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(5440), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [204597] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3064), 1,
      sym_variable_name,
    ACTIONS(9278), 1,
      anon_sym_LPAREN,
    ACTIONS(9280), 1,
      anon_sym_BANG,
    ACTIONS(9286), 1,
      anon_sym_TILDE,
    ACTIONS(9288), 1,
      anon_sym_DOLLAR,
    ACTIONS(9290), 1,
      anon_sym_DQUOTE,
    ACTIONS(9292), 1,
      aux_sym_number_token1,
    ACTIONS(9294), 1,
      aux_sym_number_token2,
    ACTIONS(9296), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9298), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9300), 1,
      anon_sym_BQUOTE,
    ACTIONS(9302), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9482), 1,
      aux_sym__simple_variable_name_token1,
    STATE(2801), 1,
      sym__arithmetic_binary_expression,
    STATE(2803), 1,
      sym__arithmetic_ternary_expression,
    STATE(2809), 1,
      sym__arithmetic_unary_expression,
    STATE(2825), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9282), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9284), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(2745), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [204671] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1551), 1,
      anon_sym_DOLLAR,
    ACTIONS(1557), 1,
      aux_sym_number_token1,
    ACTIONS(1559), 1,
      aux_sym_number_token2,
    ACTIONS(1563), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1575), 1,
      sym__brace_start,
    ACTIONS(9198), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9202), 1,
      anon_sym_DQUOTE,
    ACTIONS(9206), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9208), 1,
      anon_sym_BQUOTE,
    ACTIONS(9210), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9224), 1,
      sym__special_character,
    ACTIONS(9484), 1,
      sym_word,
    ACTIONS(9488), 1,
      sym_test_operator,
    STATE(1436), 1,
      aux_sym__literal_repeat1,
    STATE(1543), 1,
      sym_concatenation,
    ACTIONS(9196), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9212), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9486), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1107), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [204743] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9364), 1,
      anon_sym_LPAREN,
    ACTIONS(9366), 1,
      anon_sym_BANG,
    ACTIONS(9372), 1,
      anon_sym_TILDE,
    ACTIONS(9374), 1,
      anon_sym_DOLLAR,
    ACTIONS(9376), 1,
      anon_sym_DQUOTE,
    ACTIONS(9378), 1,
      aux_sym_number_token1,
    ACTIONS(9380), 1,
      aux_sym_number_token2,
    ACTIONS(9382), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9384), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9386), 1,
      anon_sym_BQUOTE,
    ACTIONS(9388), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9392), 1,
      sym_variable_name,
    ACTIONS(9490), 1,
      aux_sym__simple_variable_name_token1,
    STATE(3391), 1,
      sym__arithmetic_binary_expression,
    STATE(3396), 1,
      sym__arithmetic_ternary_expression,
    STATE(3398), 1,
      sym__arithmetic_unary_expression,
    STATE(3400), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9368), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9370), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(3346), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [204817] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5261), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(5267), 1,
      anon_sym_DQUOTE,
    ACTIONS(5271), 1,
      aux_sym_number_token1,
    ACTIONS(5273), 1,
      aux_sym_number_token2,
    ACTIONS(5275), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(5277), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(5279), 1,
      anon_sym_BQUOTE,
    ACTIONS(5281), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(5287), 1,
      sym__brace_start,
    ACTIONS(9472), 1,
      sym_word,
    ACTIONS(9480), 1,
      sym__comment_word,
    ACTIONS(9492), 1,
      anon_sym_DOLLAR,
    ACTIONS(5259), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(5283), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9476), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9478), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(5440), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [204885] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3064), 1,
      sym_variable_name,
    ACTIONS(9278), 1,
      anon_sym_LPAREN,
    ACTIONS(9280), 1,
      anon_sym_BANG,
    ACTIONS(9286), 1,
      anon_sym_TILDE,
    ACTIONS(9288), 1,
      anon_sym_DOLLAR,
    ACTIONS(9290), 1,
      anon_sym_DQUOTE,
    ACTIONS(9292), 1,
      aux_sym_number_token1,
    ACTIONS(9294), 1,
      aux_sym_number_token2,
    ACTIONS(9296), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9298), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9300), 1,
      anon_sym_BQUOTE,
    ACTIONS(9302), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9494), 1,
      aux_sym__simple_variable_name_token1,
    STATE(2801), 1,
      sym__arithmetic_binary_expression,
    STATE(2803), 1,
      sym__arithmetic_ternary_expression,
    STATE(2809), 1,
      sym__arithmetic_unary_expression,
    STATE(2825), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9282), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9284), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(2746), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [204959] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3064), 1,
      sym_variable_name,
    ACTIONS(9278), 1,
      anon_sym_LPAREN,
    ACTIONS(9280), 1,
      anon_sym_BANG,
    ACTIONS(9286), 1,
      anon_sym_TILDE,
    ACTIONS(9288), 1,
      anon_sym_DOLLAR,
    ACTIONS(9290), 1,
      anon_sym_DQUOTE,
    ACTIONS(9292), 1,
      aux_sym_number_token1,
    ACTIONS(9294), 1,
      aux_sym_number_token2,
    ACTIONS(9296), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9298), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9300), 1,
      anon_sym_BQUOTE,
    ACTIONS(9302), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9496), 1,
      aux_sym__simple_variable_name_token1,
    STATE(2801), 1,
      sym__arithmetic_binary_expression,
    STATE(2803), 1,
      sym__arithmetic_ternary_expression,
    STATE(2809), 1,
      sym__arithmetic_unary_expression,
    STATE(2825), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9282), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9284), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(2741), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [205033] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9364), 1,
      anon_sym_LPAREN,
    ACTIONS(9366), 1,
      anon_sym_BANG,
    ACTIONS(9372), 1,
      anon_sym_TILDE,
    ACTIONS(9374), 1,
      anon_sym_DOLLAR,
    ACTIONS(9376), 1,
      anon_sym_DQUOTE,
    ACTIONS(9378), 1,
      aux_sym_number_token1,
    ACTIONS(9380), 1,
      aux_sym_number_token2,
    ACTIONS(9382), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9384), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9386), 1,
      anon_sym_BQUOTE,
    ACTIONS(9388), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9392), 1,
      sym_variable_name,
    ACTIONS(9498), 1,
      aux_sym__simple_variable_name_token1,
    STATE(3391), 1,
      sym__arithmetic_binary_expression,
    STATE(3396), 1,
      sym__arithmetic_ternary_expression,
    STATE(3398), 1,
      sym__arithmetic_unary_expression,
    STATE(3400), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9368), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9370), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(3362), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [205107] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3064), 1,
      sym_variable_name,
    ACTIONS(9278), 1,
      anon_sym_LPAREN,
    ACTIONS(9280), 1,
      anon_sym_BANG,
    ACTIONS(9286), 1,
      anon_sym_TILDE,
    ACTIONS(9288), 1,
      anon_sym_DOLLAR,
    ACTIONS(9290), 1,
      anon_sym_DQUOTE,
    ACTIONS(9292), 1,
      aux_sym_number_token1,
    ACTIONS(9294), 1,
      aux_sym_number_token2,
    ACTIONS(9296), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9298), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9300), 1,
      anon_sym_BQUOTE,
    ACTIONS(9302), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9500), 1,
      aux_sym__simple_variable_name_token1,
    STATE(2801), 1,
      sym__arithmetic_binary_expression,
    STATE(2803), 1,
      sym__arithmetic_ternary_expression,
    STATE(2809), 1,
      sym__arithmetic_unary_expression,
    STATE(2825), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9282), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9284), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(2890), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [205181] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4151), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(4153), 1,
      anon_sym_DOLLAR,
    ACTIONS(4157), 1,
      anon_sym_DQUOTE,
    ACTIONS(4161), 1,
      aux_sym_number_token1,
    ACTIONS(4163), 1,
      aux_sym_number_token2,
    ACTIONS(4165), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(4167), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(4169), 1,
      anon_sym_BQUOTE,
    ACTIONS(4171), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(4181), 1,
      sym__brace_start,
    ACTIONS(9502), 1,
      sym_word,
    ACTIONS(9508), 1,
      sym__comment_word,
    ACTIONS(4149), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(4173), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9504), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9506), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2580), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [205249] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1551), 1,
      anon_sym_DOLLAR,
    ACTIONS(1557), 1,
      aux_sym_number_token1,
    ACTIONS(1559), 1,
      aux_sym_number_token2,
    ACTIONS(1563), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1575), 1,
      sym__brace_start,
    ACTIONS(9198), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9202), 1,
      anon_sym_DQUOTE,
    ACTIONS(9206), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9208), 1,
      anon_sym_BQUOTE,
    ACTIONS(9210), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9268), 1,
      sym_word,
    ACTIONS(9276), 1,
      sym__comment_word,
    ACTIONS(9196), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9212), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9272), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9274), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1189), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [205317] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(8512), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8514), 1,
      anon_sym_DOLLAR,
    ACTIONS(8516), 1,
      sym__special_character,
    ACTIONS(8518), 1,
      anon_sym_DQUOTE,
    ACTIONS(8522), 1,
      aux_sym_number_token1,
    ACTIONS(8524), 1,
      aux_sym_number_token2,
    ACTIONS(8526), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8528), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(8530), 1,
      anon_sym_BQUOTE,
    ACTIONS(8532), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8542), 1,
      sym__brace_start,
    ACTIONS(9510), 1,
      sym_word,
    ACTIONS(9514), 1,
      sym_test_operator,
    STATE(2703), 1,
      aux_sym__literal_repeat1,
    STATE(2905), 1,
      sym_concatenation,
    ACTIONS(8508), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8534), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9512), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2436), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [205389] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3616), 1,
      aux_sym_number_token1,
    ACTIONS(3618), 1,
      aux_sym_number_token2,
    ACTIONS(3622), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3632), 1,
      sym__brace_start,
    ACTIONS(8850), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8854), 1,
      anon_sym_DQUOTE,
    ACTIONS(8858), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8860), 1,
      anon_sym_BQUOTE,
    ACTIONS(8862), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9516), 1,
      sym_word,
    ACTIONS(9518), 1,
      anon_sym_DOLLAR,
    ACTIONS(9524), 1,
      sym__comment_word,
    ACTIONS(8848), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8864), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9520), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9522), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2210), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [205457] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(8512), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8514), 1,
      anon_sym_DOLLAR,
    ACTIONS(8516), 1,
      sym__special_character,
    ACTIONS(8518), 1,
      anon_sym_DQUOTE,
    ACTIONS(8522), 1,
      aux_sym_number_token1,
    ACTIONS(8524), 1,
      aux_sym_number_token2,
    ACTIONS(8526), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8528), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(8530), 1,
      anon_sym_BQUOTE,
    ACTIONS(8532), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8542), 1,
      sym__brace_start,
    ACTIONS(9526), 1,
      sym_word,
    ACTIONS(9530), 1,
      sym_test_operator,
    STATE(2689), 1,
      aux_sym__literal_repeat1,
    STATE(2940), 1,
      sym_concatenation,
    ACTIONS(8508), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8534), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9528), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2413), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [205529] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3616), 1,
      aux_sym_number_token1,
    ACTIONS(3618), 1,
      aux_sym_number_token2,
    ACTIONS(3622), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3632), 1,
      sym__brace_start,
    ACTIONS(8850), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8854), 1,
      anon_sym_DQUOTE,
    ACTIONS(8858), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8860), 1,
      anon_sym_BQUOTE,
    ACTIONS(8862), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9516), 1,
      sym_word,
    ACTIONS(9524), 1,
      sym__comment_word,
    ACTIONS(9532), 1,
      anon_sym_DOLLAR,
    ACTIONS(8848), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8864), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9520), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9522), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2210), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [205597] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4820), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(4822), 1,
      anon_sym_DOLLAR,
    ACTIONS(4826), 1,
      anon_sym_DQUOTE,
    ACTIONS(4830), 1,
      aux_sym_number_token1,
    ACTIONS(4832), 1,
      aux_sym_number_token2,
    ACTIONS(4834), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(4836), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(4838), 1,
      anon_sym_BQUOTE,
    ACTIONS(4840), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(4848), 1,
      sym__brace_start,
    ACTIONS(9534), 1,
      sym_word,
    ACTIONS(9540), 1,
      sym__comment_word,
    ACTIONS(4818), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(4842), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9536), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9538), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2707), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [205665] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(5870), 1,
      anon_sym_DOLLAR,
    ACTIONS(5876), 1,
      aux_sym_number_token1,
    ACTIONS(5878), 1,
      aux_sym_number_token2,
    ACTIONS(5882), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(5892), 1,
      sym__brace_start,
    ACTIONS(9542), 1,
      sym_word,
    ACTIONS(9546), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9548), 1,
      sym__special_character,
    ACTIONS(9550), 1,
      anon_sym_DQUOTE,
    ACTIONS(9554), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9556), 1,
      anon_sym_BQUOTE,
    ACTIONS(9558), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9562), 1,
      sym_test_operator,
    STATE(5651), 1,
      aux_sym__literal_repeat1,
    STATE(5744), 1,
      sym_concatenation,
    ACTIONS(9544), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9552), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(9560), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(5616), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [205737] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(5870), 1,
      anon_sym_DOLLAR,
    ACTIONS(5876), 1,
      aux_sym_number_token1,
    ACTIONS(5878), 1,
      aux_sym_number_token2,
    ACTIONS(5882), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(5892), 1,
      sym__brace_start,
    ACTIONS(9546), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9548), 1,
      sym__special_character,
    ACTIONS(9550), 1,
      anon_sym_DQUOTE,
    ACTIONS(9554), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9556), 1,
      anon_sym_BQUOTE,
    ACTIONS(9558), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9564), 1,
      sym_word,
    ACTIONS(9568), 1,
      sym_test_operator,
    STATE(5643), 1,
      aux_sym__literal_repeat1,
    STATE(5747), 1,
      sym_concatenation,
    ACTIONS(9544), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9560), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9566), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(5605), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [205809] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3064), 1,
      sym_variable_name,
    ACTIONS(9278), 1,
      anon_sym_LPAREN,
    ACTIONS(9280), 1,
      anon_sym_BANG,
    ACTIONS(9286), 1,
      anon_sym_TILDE,
    ACTIONS(9288), 1,
      anon_sym_DOLLAR,
    ACTIONS(9290), 1,
      anon_sym_DQUOTE,
    ACTIONS(9292), 1,
      aux_sym_number_token1,
    ACTIONS(9294), 1,
      aux_sym_number_token2,
    ACTIONS(9296), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9298), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9300), 1,
      anon_sym_BQUOTE,
    ACTIONS(9302), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9570), 1,
      aux_sym__simple_variable_name_token1,
    STATE(2801), 1,
      sym__arithmetic_binary_expression,
    STATE(2803), 1,
      sym__arithmetic_ternary_expression,
    STATE(2809), 1,
      sym__arithmetic_unary_expression,
    STATE(2825), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9282), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9284), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(2740), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [205883] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9364), 1,
      anon_sym_LPAREN,
    ACTIONS(9366), 1,
      anon_sym_BANG,
    ACTIONS(9372), 1,
      anon_sym_TILDE,
    ACTIONS(9374), 1,
      anon_sym_DOLLAR,
    ACTIONS(9376), 1,
      anon_sym_DQUOTE,
    ACTIONS(9378), 1,
      aux_sym_number_token1,
    ACTIONS(9380), 1,
      aux_sym_number_token2,
    ACTIONS(9382), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9384), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9386), 1,
      anon_sym_BQUOTE,
    ACTIONS(9388), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9392), 1,
      sym_variable_name,
    ACTIONS(9572), 1,
      aux_sym__simple_variable_name_token1,
    STATE(3391), 1,
      sym__arithmetic_binary_expression,
    STATE(3396), 1,
      sym__arithmetic_ternary_expression,
    STATE(3398), 1,
      sym__arithmetic_unary_expression,
    STATE(3400), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9368), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9370), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(3363), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [205957] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9364), 1,
      anon_sym_LPAREN,
    ACTIONS(9366), 1,
      anon_sym_BANG,
    ACTIONS(9372), 1,
      anon_sym_TILDE,
    ACTIONS(9374), 1,
      anon_sym_DOLLAR,
    ACTIONS(9376), 1,
      anon_sym_DQUOTE,
    ACTIONS(9378), 1,
      aux_sym_number_token1,
    ACTIONS(9380), 1,
      aux_sym_number_token2,
    ACTIONS(9382), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9384), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9386), 1,
      anon_sym_BQUOTE,
    ACTIONS(9388), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9392), 1,
      sym_variable_name,
    ACTIONS(9574), 1,
      aux_sym__simple_variable_name_token1,
    STATE(3391), 1,
      sym__arithmetic_binary_expression,
    STATE(3396), 1,
      sym__arithmetic_ternary_expression,
    STATE(3398), 1,
      sym__arithmetic_unary_expression,
    STATE(3400), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9368), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9370), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(3344), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [206031] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3064), 1,
      sym_variable_name,
    ACTIONS(9278), 1,
      anon_sym_LPAREN,
    ACTIONS(9280), 1,
      anon_sym_BANG,
    ACTIONS(9286), 1,
      anon_sym_TILDE,
    ACTIONS(9288), 1,
      anon_sym_DOLLAR,
    ACTIONS(9290), 1,
      anon_sym_DQUOTE,
    ACTIONS(9292), 1,
      aux_sym_number_token1,
    ACTIONS(9294), 1,
      aux_sym_number_token2,
    ACTIONS(9296), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9298), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9300), 1,
      anon_sym_BQUOTE,
    ACTIONS(9302), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9576), 1,
      aux_sym__simple_variable_name_token1,
    STATE(2801), 1,
      sym__arithmetic_binary_expression,
    STATE(2803), 1,
      sym__arithmetic_ternary_expression,
    STATE(2809), 1,
      sym__arithmetic_unary_expression,
    STATE(2825), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9282), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9284), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(2747), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [206105] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1557), 1,
      aux_sym_number_token1,
    ACTIONS(1559), 1,
      aux_sym_number_token2,
    ACTIONS(1563), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1575), 1,
      sym__brace_start,
    ACTIONS(9198), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9202), 1,
      anon_sym_DQUOTE,
    ACTIONS(9206), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9208), 1,
      anon_sym_BQUOTE,
    ACTIONS(9210), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9268), 1,
      sym_word,
    ACTIONS(9276), 1,
      sym__comment_word,
    ACTIONS(9578), 1,
      anon_sym_DOLLAR,
    ACTIONS(9196), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9212), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9272), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9274), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1189), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [206173] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2269), 1,
      aux_sym_number_token1,
    ACTIONS(2271), 1,
      aux_sym_number_token2,
    ACTIONS(2275), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2289), 1,
      sym__brace_start,
    ACTIONS(8066), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8070), 1,
      anon_sym_DQUOTE,
    ACTIONS(8074), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8076), 1,
      anon_sym_BQUOTE,
    ACTIONS(8078), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9580), 1,
      sym_word,
    ACTIONS(9582), 1,
      anon_sym_DOLLAR,
    ACTIONS(9588), 1,
      sym__comment_word,
    ACTIONS(8062), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8080), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9584), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9586), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1623), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [206241] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2269), 1,
      aux_sym_number_token1,
    ACTIONS(2271), 1,
      aux_sym_number_token2,
    ACTIONS(2275), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2289), 1,
      sym__brace_start,
    ACTIONS(8066), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8070), 1,
      anon_sym_DQUOTE,
    ACTIONS(8074), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8076), 1,
      anon_sym_BQUOTE,
    ACTIONS(8078), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9580), 1,
      sym_word,
    ACTIONS(9588), 1,
      sym__comment_word,
    ACTIONS(9590), 1,
      anon_sym_DOLLAR,
    ACTIONS(8062), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8080), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9584), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9586), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1623), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [206309] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9364), 1,
      anon_sym_LPAREN,
    ACTIONS(9366), 1,
      anon_sym_BANG,
    ACTIONS(9372), 1,
      anon_sym_TILDE,
    ACTIONS(9374), 1,
      anon_sym_DOLLAR,
    ACTIONS(9376), 1,
      anon_sym_DQUOTE,
    ACTIONS(9378), 1,
      aux_sym_number_token1,
    ACTIONS(9380), 1,
      aux_sym_number_token2,
    ACTIONS(9382), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9384), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9386), 1,
      anon_sym_BQUOTE,
    ACTIONS(9388), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9392), 1,
      sym_variable_name,
    ACTIONS(9592), 1,
      aux_sym__simple_variable_name_token1,
    STATE(3391), 1,
      sym__arithmetic_binary_expression,
    STATE(3396), 1,
      sym__arithmetic_ternary_expression,
    STATE(3398), 1,
      sym__arithmetic_unary_expression,
    STATE(3400), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9368), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9370), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(3343), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [206383] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3064), 1,
      sym_variable_name,
    ACTIONS(9278), 1,
      anon_sym_LPAREN,
    ACTIONS(9280), 1,
      anon_sym_BANG,
    ACTIONS(9286), 1,
      anon_sym_TILDE,
    ACTIONS(9288), 1,
      anon_sym_DOLLAR,
    ACTIONS(9290), 1,
      anon_sym_DQUOTE,
    ACTIONS(9292), 1,
      aux_sym_number_token1,
    ACTIONS(9294), 1,
      aux_sym_number_token2,
    ACTIONS(9296), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9298), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9300), 1,
      anon_sym_BQUOTE,
    ACTIONS(9302), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9594), 1,
      aux_sym__simple_variable_name_token1,
    STATE(2801), 1,
      sym__arithmetic_binary_expression,
    STATE(2803), 1,
      sym__arithmetic_ternary_expression,
    STATE(2809), 1,
      sym__arithmetic_unary_expression,
    STATE(2825), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9282), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9284), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(2748), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [206457] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1557), 1,
      aux_sym_number_token1,
    ACTIONS(1559), 1,
      aux_sym_number_token2,
    ACTIONS(1563), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1575), 1,
      sym__brace_start,
    ACTIONS(9198), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9202), 1,
      anon_sym_DQUOTE,
    ACTIONS(9206), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9208), 1,
      anon_sym_BQUOTE,
    ACTIONS(9210), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9268), 1,
      sym_word,
    ACTIONS(9276), 1,
      sym__comment_word,
    ACTIONS(9596), 1,
      anon_sym_DOLLAR,
    ACTIONS(9196), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9212), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9272), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9274), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1189), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [206525] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9364), 1,
      anon_sym_LPAREN,
    ACTIONS(9366), 1,
      anon_sym_BANG,
    ACTIONS(9372), 1,
      anon_sym_TILDE,
    ACTIONS(9374), 1,
      anon_sym_DOLLAR,
    ACTIONS(9376), 1,
      anon_sym_DQUOTE,
    ACTIONS(9378), 1,
      aux_sym_number_token1,
    ACTIONS(9380), 1,
      aux_sym_number_token2,
    ACTIONS(9382), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9384), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9386), 1,
      anon_sym_BQUOTE,
    ACTIONS(9388), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9392), 1,
      sym_variable_name,
    ACTIONS(9598), 1,
      aux_sym__simple_variable_name_token1,
    STATE(3391), 1,
      sym__arithmetic_binary_expression,
    STATE(3396), 1,
      sym__arithmetic_ternary_expression,
    STATE(3398), 1,
      sym__arithmetic_unary_expression,
    STATE(3400), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9368), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9370), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(3342), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [206599] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8102), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8108), 1,
      anon_sym_DQUOTE,
    ACTIONS(8112), 1,
      aux_sym_number_token1,
    ACTIONS(8114), 1,
      aux_sym_number_token2,
    ACTIONS(8116), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8118), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(8120), 1,
      anon_sym_BQUOTE,
    ACTIONS(8122), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8132), 1,
      sym__brace_start,
    ACTIONS(9600), 1,
      sym_word,
    ACTIONS(9602), 1,
      anon_sym_DOLLAR,
    ACTIONS(9608), 1,
      sym__comment_word,
    ACTIONS(8098), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8124), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9604), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9606), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1199), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [206667] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3064), 1,
      sym_variable_name,
    ACTIONS(9278), 1,
      anon_sym_LPAREN,
    ACTIONS(9280), 1,
      anon_sym_BANG,
    ACTIONS(9286), 1,
      anon_sym_TILDE,
    ACTIONS(9288), 1,
      anon_sym_DOLLAR,
    ACTIONS(9290), 1,
      anon_sym_DQUOTE,
    ACTIONS(9292), 1,
      aux_sym_number_token1,
    ACTIONS(9294), 1,
      aux_sym_number_token2,
    ACTIONS(9296), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9298), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9300), 1,
      anon_sym_BQUOTE,
    ACTIONS(9302), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9610), 1,
      aux_sym__simple_variable_name_token1,
    STATE(2801), 1,
      sym__arithmetic_binary_expression,
    STATE(2803), 1,
      sym__arithmetic_ternary_expression,
    STATE(2809), 1,
      sym__arithmetic_unary_expression,
    STATE(2825), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9282), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9284), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(2749), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [206741] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9364), 1,
      anon_sym_LPAREN,
    ACTIONS(9366), 1,
      anon_sym_BANG,
    ACTIONS(9372), 1,
      anon_sym_TILDE,
    ACTIONS(9374), 1,
      anon_sym_DOLLAR,
    ACTIONS(9376), 1,
      anon_sym_DQUOTE,
    ACTIONS(9378), 1,
      aux_sym_number_token1,
    ACTIONS(9380), 1,
      aux_sym_number_token2,
    ACTIONS(9382), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9384), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9386), 1,
      anon_sym_BQUOTE,
    ACTIONS(9388), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9392), 1,
      sym_variable_name,
    ACTIONS(9612), 1,
      aux_sym__simple_variable_name_token1,
    STATE(3391), 1,
      sym__arithmetic_binary_expression,
    STATE(3396), 1,
      sym__arithmetic_ternary_expression,
    STATE(3398), 1,
      sym__arithmetic_unary_expression,
    STATE(3400), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9368), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9370), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(3341), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [206815] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3064), 1,
      sym_variable_name,
    ACTIONS(9278), 1,
      anon_sym_LPAREN,
    ACTIONS(9280), 1,
      anon_sym_BANG,
    ACTIONS(9286), 1,
      anon_sym_TILDE,
    ACTIONS(9288), 1,
      anon_sym_DOLLAR,
    ACTIONS(9290), 1,
      anon_sym_DQUOTE,
    ACTIONS(9292), 1,
      aux_sym_number_token1,
    ACTIONS(9294), 1,
      aux_sym_number_token2,
    ACTIONS(9296), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9298), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9300), 1,
      anon_sym_BQUOTE,
    ACTIONS(9302), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9614), 1,
      aux_sym__simple_variable_name_token1,
    STATE(2801), 1,
      sym__arithmetic_binary_expression,
    STATE(2803), 1,
      sym__arithmetic_ternary_expression,
    STATE(2809), 1,
      sym__arithmetic_unary_expression,
    STATE(2825), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9282), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9284), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(2750), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [206889] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8102), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8108), 1,
      anon_sym_DQUOTE,
    ACTIONS(8112), 1,
      aux_sym_number_token1,
    ACTIONS(8114), 1,
      aux_sym_number_token2,
    ACTIONS(8116), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8118), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(8120), 1,
      anon_sym_BQUOTE,
    ACTIONS(8122), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8132), 1,
      sym__brace_start,
    ACTIONS(9600), 1,
      sym_word,
    ACTIONS(9608), 1,
      sym__comment_word,
    ACTIONS(9616), 1,
      anon_sym_DOLLAR,
    ACTIONS(8098), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8124), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9604), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9606), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1199), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [206957] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(363), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(371), 1,
      anon_sym_DQUOTE,
    ACTIONS(375), 1,
      aux_sym_number_token1,
    ACTIONS(377), 1,
      aux_sym_number_token2,
    ACTIONS(379), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(381), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(385), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(391), 1,
      sym__brace_start,
    ACTIONS(5809), 1,
      anon_sym_BQUOTE,
    ACTIONS(8836), 1,
      sym_word,
    ACTIONS(8840), 1,
      sym__comment_word,
    ACTIONS(9618), 1,
      anon_sym_DOLLAR,
    ACTIONS(352), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(387), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(8844), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(8842), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2730), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [207025] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2636), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(2638), 1,
      anon_sym_DOLLAR,
    ACTIONS(2642), 1,
      anon_sym_DQUOTE,
    ACTIONS(2646), 1,
      aux_sym_number_token1,
    ACTIONS(2648), 1,
      aux_sym_number_token2,
    ACTIONS(2650), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(2652), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2654), 1,
      anon_sym_BQUOTE,
    ACTIONS(2656), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(2664), 1,
      sym__brace_start,
    ACTIONS(9620), 1,
      sym_word,
    ACTIONS(9626), 1,
      sym__comment_word,
    ACTIONS(2630), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(2658), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9622), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9624), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2344), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [207093] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9364), 1,
      anon_sym_LPAREN,
    ACTIONS(9366), 1,
      anon_sym_BANG,
    ACTIONS(9372), 1,
      anon_sym_TILDE,
    ACTIONS(9374), 1,
      anon_sym_DOLLAR,
    ACTIONS(9376), 1,
      anon_sym_DQUOTE,
    ACTIONS(9378), 1,
      aux_sym_number_token1,
    ACTIONS(9380), 1,
      aux_sym_number_token2,
    ACTIONS(9382), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9384), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9386), 1,
      anon_sym_BQUOTE,
    ACTIONS(9388), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9392), 1,
      sym_variable_name,
    ACTIONS(9628), 1,
      aux_sym__simple_variable_name_token1,
    STATE(3391), 1,
      sym__arithmetic_binary_expression,
    STATE(3396), 1,
      sym__arithmetic_ternary_expression,
    STATE(3398), 1,
      sym__arithmetic_unary_expression,
    STATE(3400), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9368), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9370), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(3340), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [207167] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(363), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(371), 1,
      anon_sym_DQUOTE,
    ACTIONS(375), 1,
      aux_sym_number_token1,
    ACTIONS(377), 1,
      aux_sym_number_token2,
    ACTIONS(379), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(381), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(385), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(391), 1,
      sym__brace_start,
    ACTIONS(5809), 1,
      anon_sym_BQUOTE,
    ACTIONS(8836), 1,
      sym_word,
    ACTIONS(8840), 1,
      sym__comment_word,
    ACTIONS(9630), 1,
      anon_sym_DOLLAR,
    ACTIONS(352), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(387), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(8844), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(8842), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2730), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [207235] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3787), 1,
      anon_sym_DOLLAR,
    ACTIONS(3793), 1,
      aux_sym_number_token1,
    ACTIONS(3795), 1,
      aux_sym_number_token2,
    ACTIONS(3799), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3809), 1,
      sym__brace_start,
    ACTIONS(9632), 1,
      sym_word,
    ACTIONS(9636), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9638), 1,
      sym__special_character,
    ACTIONS(9640), 1,
      anon_sym_DQUOTE,
    ACTIONS(9644), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9646), 1,
      anon_sym_BQUOTE,
    ACTIONS(9648), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9652), 1,
      sym_test_operator,
    STATE(4347), 1,
      aux_sym__literal_repeat1,
    STATE(4617), 1,
      sym_concatenation,
    ACTIONS(9634), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9642), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(9650), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(4604), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [207307] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(5405), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(5407), 1,
      anon_sym_DOLLAR,
    ACTIONS(5409), 1,
      sym__special_character,
    ACTIONS(5411), 1,
      anon_sym_DQUOTE,
    ACTIONS(5415), 1,
      aux_sym_number_token1,
    ACTIONS(5417), 1,
      aux_sym_number_token2,
    ACTIONS(5419), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(5421), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(5423), 1,
      anon_sym_BQUOTE,
    ACTIONS(5425), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(5431), 1,
      sym__brace_start,
    ACTIONS(9654), 1,
      sym_word,
    ACTIONS(9658), 1,
      sym_test_operator,
    STATE(5544), 1,
      aux_sym__literal_repeat1,
    STATE(5629), 1,
      sym_concatenation,
    ACTIONS(5403), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(5427), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9656), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(5412), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [207379] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3787), 1,
      anon_sym_DOLLAR,
    ACTIONS(3793), 1,
      aux_sym_number_token1,
    ACTIONS(3795), 1,
      aux_sym_number_token2,
    ACTIONS(3799), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3809), 1,
      sym__brace_start,
    ACTIONS(9636), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9638), 1,
      sym__special_character,
    ACTIONS(9640), 1,
      anon_sym_DQUOTE,
    ACTIONS(9644), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9646), 1,
      anon_sym_BQUOTE,
    ACTIONS(9648), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9660), 1,
      sym_word,
    ACTIONS(9664), 1,
      sym_test_operator,
    STATE(4294), 1,
      aux_sym__literal_repeat1,
    STATE(4861), 1,
      sym_concatenation,
    ACTIONS(9634), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9650), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9662), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(4477), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [207451] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3064), 1,
      sym_variable_name,
    ACTIONS(9278), 1,
      anon_sym_LPAREN,
    ACTIONS(9280), 1,
      anon_sym_BANG,
    ACTIONS(9286), 1,
      anon_sym_TILDE,
    ACTIONS(9288), 1,
      anon_sym_DOLLAR,
    ACTIONS(9290), 1,
      anon_sym_DQUOTE,
    ACTIONS(9292), 1,
      aux_sym_number_token1,
    ACTIONS(9294), 1,
      aux_sym_number_token2,
    ACTIONS(9296), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9298), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9300), 1,
      anon_sym_BQUOTE,
    ACTIONS(9302), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9666), 1,
      aux_sym__simple_variable_name_token1,
    STATE(2801), 1,
      sym__arithmetic_binary_expression,
    STATE(2803), 1,
      sym__arithmetic_ternary_expression,
    STATE(2809), 1,
      sym__arithmetic_unary_expression,
    STATE(2825), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9282), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9284), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(2751), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [207525] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(5405), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(5407), 1,
      anon_sym_DOLLAR,
    ACTIONS(5409), 1,
      sym__special_character,
    ACTIONS(5411), 1,
      anon_sym_DQUOTE,
    ACTIONS(5415), 1,
      aux_sym_number_token1,
    ACTIONS(5417), 1,
      aux_sym_number_token2,
    ACTIONS(5419), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(5421), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(5423), 1,
      anon_sym_BQUOTE,
    ACTIONS(5425), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(5431), 1,
      sym__brace_start,
    ACTIONS(9668), 1,
      sym_word,
    ACTIONS(9672), 1,
      sym_test_operator,
    STATE(5500), 1,
      aux_sym__literal_repeat1,
    STATE(5671), 1,
      sym_concatenation,
    ACTIONS(5403), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(5427), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9670), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(5415), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [207597] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3185), 1,
      anon_sym_DOLLAR,
    ACTIONS(3191), 1,
      aux_sym_number_token1,
    ACTIONS(3193), 1,
      aux_sym_number_token2,
    ACTIONS(3197), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3209), 1,
      sym__brace_start,
    ACTIONS(9674), 1,
      sym_word,
    ACTIONS(9678), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9682), 1,
      anon_sym_DQUOTE,
    ACTIONS(9686), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9688), 1,
      anon_sym_BQUOTE,
    ACTIONS(9690), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9694), 1,
      sym__comment_word,
    ACTIONS(9676), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9680), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9692), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9684), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1733), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [207665] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3401), 1,
      aux_sym_number_token1,
    ACTIONS(3403), 1,
      aux_sym_number_token2,
    ACTIONS(3407), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3417), 1,
      sym__brace_start,
    ACTIONS(8926), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8930), 1,
      anon_sym_DQUOTE,
    ACTIONS(8934), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8936), 1,
      anon_sym_BQUOTE,
    ACTIONS(8938), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9696), 1,
      sym_word,
    ACTIONS(9698), 1,
      anon_sym_DOLLAR,
    ACTIONS(9704), 1,
      sym__comment_word,
    ACTIONS(8924), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8940), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9700), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9702), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1851), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [207733] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1043), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1045), 1,
      anon_sym_DOLLAR,
    ACTIONS(1049), 1,
      anon_sym_DQUOTE,
    ACTIONS(1053), 1,
      aux_sym_number_token1,
    ACTIONS(1055), 1,
      aux_sym_number_token2,
    ACTIONS(1057), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1059), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1061), 1,
      anon_sym_BQUOTE,
    ACTIONS(1063), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1071), 1,
      sym__brace_start,
    ACTIONS(9706), 1,
      sym_word,
    ACTIONS(9712), 1,
      sym__comment_word,
    ACTIONS(1041), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1065), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9708), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9710), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2075), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [207801] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(5261), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(5263), 1,
      anon_sym_DOLLAR,
    ACTIONS(5265), 1,
      sym__special_character,
    ACTIONS(5267), 1,
      anon_sym_DQUOTE,
    ACTIONS(5271), 1,
      aux_sym_number_token1,
    ACTIONS(5273), 1,
      aux_sym_number_token2,
    ACTIONS(5275), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(5277), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(5279), 1,
      anon_sym_BQUOTE,
    ACTIONS(5281), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(5287), 1,
      sym__brace_start,
    ACTIONS(9714), 1,
      sym_word,
    ACTIONS(9718), 1,
      sym_test_operator,
    STATE(5462), 1,
      aux_sym__literal_repeat1,
    STATE(5587), 1,
      sym_concatenation,
    ACTIONS(5259), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(5283), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9716), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(5300), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [207873] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8178), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8184), 1,
      anon_sym_DQUOTE,
    ACTIONS(8188), 1,
      aux_sym_number_token1,
    ACTIONS(8190), 1,
      aux_sym_number_token2,
    ACTIONS(8192), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8194), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(8196), 1,
      anon_sym_BQUOTE,
    ACTIONS(8198), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8208), 1,
      sym__brace_start,
    ACTIONS(9720), 1,
      sym_word,
    ACTIONS(9722), 1,
      anon_sym_DOLLAR,
    ACTIONS(9728), 1,
      sym__comment_word,
    ACTIONS(8174), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8200), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9724), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9726), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1590), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [207941] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(111), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(117), 1,
      anon_sym_DQUOTE,
    ACTIONS(121), 1,
      aux_sym_number_token1,
    ACTIONS(123), 1,
      aux_sym_number_token2,
    ACTIONS(125), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(127), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(129), 1,
      anon_sym_BQUOTE,
    ACTIONS(131), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(141), 1,
      sym__brace_start,
    ACTIONS(9340), 1,
      sym_word,
    ACTIONS(9346), 1,
      sym__comment_word,
    ACTIONS(9730), 1,
      anon_sym_DOLLAR,
    ACTIONS(109), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(133), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9342), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9344), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(482), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [208009] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(5261), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(5263), 1,
      anon_sym_DOLLAR,
    ACTIONS(5265), 1,
      sym__special_character,
    ACTIONS(5267), 1,
      anon_sym_DQUOTE,
    ACTIONS(5271), 1,
      aux_sym_number_token1,
    ACTIONS(5273), 1,
      aux_sym_number_token2,
    ACTIONS(5275), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(5277), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(5279), 1,
      anon_sym_BQUOTE,
    ACTIONS(5281), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(5287), 1,
      sym__brace_start,
    ACTIONS(9732), 1,
      sym_word,
    ACTIONS(9736), 1,
      sym_test_operator,
    STATE(5466), 1,
      aux_sym__literal_repeat1,
    STATE(5612), 1,
      sym_concatenation,
    ACTIONS(5259), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(5283), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9734), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(5303), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [208081] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3401), 1,
      aux_sym_number_token1,
    ACTIONS(3403), 1,
      aux_sym_number_token2,
    ACTIONS(3407), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3417), 1,
      sym__brace_start,
    ACTIONS(8926), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8930), 1,
      anon_sym_DQUOTE,
    ACTIONS(8934), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8936), 1,
      anon_sym_BQUOTE,
    ACTIONS(8938), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9696), 1,
      sym_word,
    ACTIONS(9704), 1,
      sym__comment_word,
    ACTIONS(9738), 1,
      anon_sym_DOLLAR,
    ACTIONS(8924), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8940), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9700), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9702), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1851), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [208149] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3882), 1,
      anon_sym_DOLLAR,
    ACTIONS(3886), 1,
      aux_sym_number_token1,
    ACTIONS(3888), 1,
      aux_sym_number_token2,
    ACTIONS(3892), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3902), 1,
      sym__brace_start,
    ACTIONS(8992), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8996), 1,
      anon_sym_DQUOTE,
    ACTIONS(9000), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9002), 1,
      anon_sym_BQUOTE,
    ACTIONS(9004), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9740), 1,
      sym_word,
    ACTIONS(9746), 1,
      sym__comment_word,
    ACTIONS(8990), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9006), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9742), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9744), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2004), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [208217] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(8102), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8104), 1,
      anon_sym_DOLLAR,
    ACTIONS(8106), 1,
      sym__special_character,
    ACTIONS(8108), 1,
      anon_sym_DQUOTE,
    ACTIONS(8112), 1,
      aux_sym_number_token1,
    ACTIONS(8114), 1,
      aux_sym_number_token2,
    ACTIONS(8116), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8118), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(8120), 1,
      anon_sym_BQUOTE,
    ACTIONS(8122), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8132), 1,
      sym__brace_start,
    ACTIONS(9748), 1,
      sym_word,
    ACTIONS(9752), 1,
      sym_test_operator,
    STATE(1349), 1,
      aux_sym__literal_repeat1,
    STATE(1639), 1,
      sym_concatenation,
    ACTIONS(8098), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8124), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9750), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1127), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [208289] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9754), 1,
      sym_word,
    ACTIONS(9758), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9760), 1,
      anon_sym_DOLLAR,
    ACTIONS(9764), 1,
      anon_sym_DQUOTE,
    ACTIONS(9768), 1,
      aux_sym_number_token1,
    ACTIONS(9770), 1,
      aux_sym_number_token2,
    ACTIONS(9772), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9774), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9776), 1,
      anon_sym_BQUOTE,
    ACTIONS(9778), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9782), 1,
      sym__comment_word,
    ACTIONS(9784), 1,
      sym__brace_start,
    ACTIONS(9756), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9762), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9780), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9766), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(6225), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [208357] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(8102), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8104), 1,
      anon_sym_DOLLAR,
    ACTIONS(8106), 1,
      sym__special_character,
    ACTIONS(8108), 1,
      anon_sym_DQUOTE,
    ACTIONS(8112), 1,
      aux_sym_number_token1,
    ACTIONS(8114), 1,
      aux_sym_number_token2,
    ACTIONS(8116), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8118), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(8120), 1,
      anon_sym_BQUOTE,
    ACTIONS(8122), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8132), 1,
      sym__brace_start,
    ACTIONS(9786), 1,
      sym_word,
    ACTIONS(9790), 1,
      sym_test_operator,
    STATE(1380), 1,
      aux_sym__literal_repeat1,
    STATE(1629), 1,
      sym_concatenation,
    ACTIONS(8098), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8124), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9788), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1162), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [208429] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4080), 1,
      anon_sym_DOLLAR,
    ACTIONS(4086), 1,
      aux_sym_number_token1,
    ACTIONS(4088), 1,
      aux_sym_number_token2,
    ACTIONS(4092), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(4102), 1,
      sym__brace_start,
    ACTIONS(9306), 1,
      sym_word,
    ACTIONS(9310), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9316), 1,
      anon_sym_DQUOTE,
    ACTIONS(9320), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9322), 1,
      anon_sym_BQUOTE,
    ACTIONS(9324), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9328), 1,
      sym__comment_word,
    ACTIONS(9308), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9314), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9326), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9318), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(4566), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [208497] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1587), 1,
      anon_sym_DOLLAR,
    ACTIONS(1593), 1,
      aux_sym_number_token1,
    ACTIONS(1595), 1,
      aux_sym_number_token2,
    ACTIONS(1599), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1611), 1,
      sym__brace_start,
    ACTIONS(9240), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9242), 1,
      sym__special_character,
    ACTIONS(9244), 1,
      anon_sym_DQUOTE,
    ACTIONS(9248), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9250), 1,
      anon_sym_BQUOTE,
    ACTIONS(9252), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9792), 1,
      sym_word,
    ACTIONS(9796), 1,
      sym_test_operator,
    STATE(1377), 1,
      aux_sym__literal_repeat1,
    STATE(1613), 1,
      sym_concatenation,
    ACTIONS(9238), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9254), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9794), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1076), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [208569] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3367), 1,
      anon_sym_DOLLAR,
    ACTIONS(3373), 1,
      aux_sym_number_token1,
    ACTIONS(3375), 1,
      aux_sym_number_token2,
    ACTIONS(3379), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3389), 1,
      sym__brace_start,
    ACTIONS(9798), 1,
      sym_word,
    ACTIONS(9802), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9804), 1,
      sym__special_character,
    ACTIONS(9806), 1,
      anon_sym_DQUOTE,
    ACTIONS(9810), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9812), 1,
      anon_sym_BQUOTE,
    ACTIONS(9814), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9818), 1,
      sym_test_operator,
    STATE(4538), 1,
      aux_sym__literal_repeat1,
    STATE(4930), 1,
      sym_concatenation,
    ACTIONS(9800), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9808), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(9816), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(4292), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [208641] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1767), 1,
      aux_sym_number_token1,
    ACTIONS(1769), 1,
      aux_sym_number_token2,
    ACTIONS(1773), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1787), 1,
      sym__brace_start,
    ACTIONS(8426), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8430), 1,
      anon_sym_DQUOTE,
    ACTIONS(8434), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8436), 1,
      anon_sym_BQUOTE,
    ACTIONS(8438), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9820), 1,
      sym_word,
    ACTIONS(9822), 1,
      anon_sym_DOLLAR,
    ACTIONS(9828), 1,
      sym__comment_word,
    ACTIONS(8422), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8440), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9824), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9826), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1207), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [208709] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3367), 1,
      anon_sym_DOLLAR,
    ACTIONS(3373), 1,
      aux_sym_number_token1,
    ACTIONS(3375), 1,
      aux_sym_number_token2,
    ACTIONS(3379), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3389), 1,
      sym__brace_start,
    ACTIONS(9802), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9804), 1,
      sym__special_character,
    ACTIONS(9806), 1,
      anon_sym_DQUOTE,
    ACTIONS(9810), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9812), 1,
      anon_sym_BQUOTE,
    ACTIONS(9814), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9830), 1,
      sym_word,
    ACTIONS(9834), 1,
      sym_test_operator,
    STATE(4462), 1,
      aux_sym__literal_repeat1,
    STATE(4888), 1,
      sym_concatenation,
    ACTIONS(9800), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9816), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9832), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(4270), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [208781] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9364), 1,
      anon_sym_LPAREN,
    ACTIONS(9366), 1,
      anon_sym_BANG,
    ACTIONS(9372), 1,
      anon_sym_TILDE,
    ACTIONS(9374), 1,
      anon_sym_DOLLAR,
    ACTIONS(9376), 1,
      anon_sym_DQUOTE,
    ACTIONS(9378), 1,
      aux_sym_number_token1,
    ACTIONS(9380), 1,
      aux_sym_number_token2,
    ACTIONS(9382), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9384), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9386), 1,
      anon_sym_BQUOTE,
    ACTIONS(9388), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9392), 1,
      sym_variable_name,
    ACTIONS(9836), 1,
      aux_sym__simple_variable_name_token1,
    STATE(3391), 1,
      sym__arithmetic_binary_expression,
    STATE(3396), 1,
      sym__arithmetic_ternary_expression,
    STATE(3398), 1,
      sym__arithmetic_unary_expression,
    STATE(3400), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9368), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9370), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(3339), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [208855] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3064), 1,
      sym_variable_name,
    ACTIONS(9278), 1,
      anon_sym_LPAREN,
    ACTIONS(9280), 1,
      anon_sym_BANG,
    ACTIONS(9286), 1,
      anon_sym_TILDE,
    ACTIONS(9288), 1,
      anon_sym_DOLLAR,
    ACTIONS(9290), 1,
      anon_sym_DQUOTE,
    ACTIONS(9292), 1,
      aux_sym_number_token1,
    ACTIONS(9294), 1,
      aux_sym_number_token2,
    ACTIONS(9296), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9298), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9300), 1,
      anon_sym_BQUOTE,
    ACTIONS(9302), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9838), 1,
      aux_sym__simple_variable_name_token1,
    STATE(2801), 1,
      sym__arithmetic_binary_expression,
    STATE(2803), 1,
      sym__arithmetic_ternary_expression,
    STATE(2809), 1,
      sym__arithmetic_unary_expression,
    STATE(2825), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9282), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9284), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(2752), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [208929] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1767), 1,
      aux_sym_number_token1,
    ACTIONS(1769), 1,
      aux_sym_number_token2,
    ACTIONS(1773), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1787), 1,
      sym__brace_start,
    ACTIONS(8426), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8430), 1,
      anon_sym_DQUOTE,
    ACTIONS(8434), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8436), 1,
      anon_sym_BQUOTE,
    ACTIONS(8438), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9820), 1,
      sym_word,
    ACTIONS(9828), 1,
      sym__comment_word,
    ACTIONS(9840), 1,
      anon_sym_DOLLAR,
    ACTIONS(8422), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8440), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9824), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9826), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1207), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [208997] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3064), 1,
      sym_variable_name,
    ACTIONS(9278), 1,
      anon_sym_LPAREN,
    ACTIONS(9280), 1,
      anon_sym_BANG,
    ACTIONS(9286), 1,
      anon_sym_TILDE,
    ACTIONS(9288), 1,
      anon_sym_DOLLAR,
    ACTIONS(9290), 1,
      anon_sym_DQUOTE,
    ACTIONS(9292), 1,
      aux_sym_number_token1,
    ACTIONS(9294), 1,
      aux_sym_number_token2,
    ACTIONS(9296), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9298), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9300), 1,
      anon_sym_BQUOTE,
    ACTIONS(9302), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9842), 1,
      aux_sym__simple_variable_name_token1,
    STATE(2801), 1,
      sym__arithmetic_binary_expression,
    STATE(2803), 1,
      sym__arithmetic_ternary_expression,
    STATE(2809), 1,
      sym__arithmetic_unary_expression,
    STATE(2825), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9282), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9284), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(2739), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [209071] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1587), 1,
      anon_sym_DOLLAR,
    ACTIONS(1593), 1,
      aux_sym_number_token1,
    ACTIONS(1595), 1,
      aux_sym_number_token2,
    ACTIONS(1599), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1611), 1,
      sym__brace_start,
    ACTIONS(9240), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9242), 1,
      sym__special_character,
    ACTIONS(9244), 1,
      anon_sym_DQUOTE,
    ACTIONS(9248), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9250), 1,
      anon_sym_BQUOTE,
    ACTIONS(9252), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9844), 1,
      sym_word,
    ACTIONS(9848), 1,
      sym_test_operator,
    STATE(1374), 1,
      aux_sym__literal_repeat1,
    STATE(1622), 1,
      sym_concatenation,
    ACTIONS(9238), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9254), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9846), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1069), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [209143] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3477), 1,
      anon_sym_DOLLAR,
    ACTIONS(3481), 1,
      aux_sym_number_token1,
    ACTIONS(3483), 1,
      aux_sym_number_token2,
    ACTIONS(3487), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3497), 1,
      sym__brace_start,
    ACTIONS(8906), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8910), 1,
      anon_sym_DQUOTE,
    ACTIONS(8914), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8916), 1,
      anon_sym_BQUOTE,
    ACTIONS(8918), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9850), 1,
      sym_word,
    ACTIONS(9856), 1,
      sym__comment_word,
    ACTIONS(8904), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8920), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9852), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9854), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1921), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [209211] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9364), 1,
      anon_sym_LPAREN,
    ACTIONS(9366), 1,
      anon_sym_BANG,
    ACTIONS(9372), 1,
      anon_sym_TILDE,
    ACTIONS(9374), 1,
      anon_sym_DOLLAR,
    ACTIONS(9376), 1,
      anon_sym_DQUOTE,
    ACTIONS(9378), 1,
      aux_sym_number_token1,
    ACTIONS(9380), 1,
      aux_sym_number_token2,
    ACTIONS(9382), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9384), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9386), 1,
      anon_sym_BQUOTE,
    ACTIONS(9388), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9392), 1,
      sym_variable_name,
    ACTIONS(9858), 1,
      aux_sym__simple_variable_name_token1,
    STATE(3391), 1,
      sym__arithmetic_binary_expression,
    STATE(3396), 1,
      sym__arithmetic_ternary_expression,
    STATE(3398), 1,
      sym__arithmetic_unary_expression,
    STATE(3400), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9368), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9370), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(3364), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [209285] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(7958), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(7964), 1,
      anon_sym_DQUOTE,
    ACTIONS(7968), 1,
      aux_sym_number_token1,
    ACTIONS(7970), 1,
      aux_sym_number_token2,
    ACTIONS(7972), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(7974), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(7976), 1,
      anon_sym_BQUOTE,
    ACTIONS(7978), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(7988), 1,
      sym__brace_start,
    ACTIONS(9860), 1,
      sym_word,
    ACTIONS(9862), 1,
      anon_sym_DOLLAR,
    ACTIONS(9868), 1,
      sym__comment_word,
    ACTIONS(7954), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(7980), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9864), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9866), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(4357), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [209353] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(7958), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(7964), 1,
      anon_sym_DQUOTE,
    ACTIONS(7968), 1,
      aux_sym_number_token1,
    ACTIONS(7970), 1,
      aux_sym_number_token2,
    ACTIONS(7972), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(7974), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(7976), 1,
      anon_sym_BQUOTE,
    ACTIONS(7978), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(7988), 1,
      sym__brace_start,
    ACTIONS(9860), 1,
      sym_word,
    ACTIONS(9868), 1,
      sym__comment_word,
    ACTIONS(9870), 1,
      anon_sym_DOLLAR,
    ACTIONS(7954), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(7980), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9864), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9866), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(4357), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [209421] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3793), 1,
      aux_sym_number_token1,
    ACTIONS(3795), 1,
      aux_sym_number_token2,
    ACTIONS(3799), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3809), 1,
      sym__brace_start,
    ACTIONS(9636), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9640), 1,
      anon_sym_DQUOTE,
    ACTIONS(9644), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9646), 1,
      anon_sym_BQUOTE,
    ACTIONS(9648), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9872), 1,
      sym_word,
    ACTIONS(9874), 1,
      anon_sym_DOLLAR,
    ACTIONS(9880), 1,
      sym__comment_word,
    ACTIONS(9634), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9650), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9876), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9878), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(4337), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [209489] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9364), 1,
      anon_sym_LPAREN,
    ACTIONS(9366), 1,
      anon_sym_BANG,
    ACTIONS(9372), 1,
      anon_sym_TILDE,
    ACTIONS(9374), 1,
      anon_sym_DOLLAR,
    ACTIONS(9376), 1,
      anon_sym_DQUOTE,
    ACTIONS(9378), 1,
      aux_sym_number_token1,
    ACTIONS(9380), 1,
      aux_sym_number_token2,
    ACTIONS(9382), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9384), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9386), 1,
      anon_sym_BQUOTE,
    ACTIONS(9388), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9392), 1,
      sym_variable_name,
    ACTIONS(9882), 1,
      aux_sym__simple_variable_name_token1,
    STATE(3391), 1,
      sym__arithmetic_binary_expression,
    STATE(3396), 1,
      sym__arithmetic_ternary_expression,
    STATE(3398), 1,
      sym__arithmetic_unary_expression,
    STATE(3400), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9368), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9370), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(3338), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [209563] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3793), 1,
      aux_sym_number_token1,
    ACTIONS(3795), 1,
      aux_sym_number_token2,
    ACTIONS(3799), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3809), 1,
      sym__brace_start,
    ACTIONS(9636), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9640), 1,
      anon_sym_DQUOTE,
    ACTIONS(9644), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9646), 1,
      anon_sym_BQUOTE,
    ACTIONS(9648), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9872), 1,
      sym_word,
    ACTIONS(9880), 1,
      sym__comment_word,
    ACTIONS(9884), 1,
      anon_sym_DOLLAR,
    ACTIONS(9634), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9650), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9876), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9878), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(4337), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [209631] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4086), 1,
      aux_sym_number_token1,
    ACTIONS(4088), 1,
      aux_sym_number_token2,
    ACTIONS(4092), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(4102), 1,
      sym__brace_start,
    ACTIONS(9306), 1,
      sym_word,
    ACTIONS(9310), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9316), 1,
      anon_sym_DQUOTE,
    ACTIONS(9320), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9322), 1,
      anon_sym_BQUOTE,
    ACTIONS(9324), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9328), 1,
      sym__comment_word,
    ACTIONS(9886), 1,
      anon_sym_DOLLAR,
    ACTIONS(9308), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9314), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9326), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9318), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(4566), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [209699] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4086), 1,
      aux_sym_number_token1,
    ACTIONS(4088), 1,
      aux_sym_number_token2,
    ACTIONS(4092), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(4102), 1,
      sym__brace_start,
    ACTIONS(9306), 1,
      sym_word,
    ACTIONS(9310), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9316), 1,
      anon_sym_DQUOTE,
    ACTIONS(9320), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9322), 1,
      anon_sym_BQUOTE,
    ACTIONS(9324), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9328), 1,
      sym__comment_word,
    ACTIONS(9888), 1,
      anon_sym_DOLLAR,
    ACTIONS(9308), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9314), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9326), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9318), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(4566), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [209767] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3064), 1,
      sym_variable_name,
    ACTIONS(9278), 1,
      anon_sym_LPAREN,
    ACTIONS(9280), 1,
      anon_sym_BANG,
    ACTIONS(9286), 1,
      anon_sym_TILDE,
    ACTIONS(9288), 1,
      anon_sym_DOLLAR,
    ACTIONS(9290), 1,
      anon_sym_DQUOTE,
    ACTIONS(9292), 1,
      aux_sym_number_token1,
    ACTIONS(9294), 1,
      aux_sym_number_token2,
    ACTIONS(9296), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9298), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9300), 1,
      anon_sym_BQUOTE,
    ACTIONS(9302), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9890), 1,
      aux_sym__simple_variable_name_token1,
    STATE(2801), 1,
      sym__arithmetic_binary_expression,
    STATE(2803), 1,
      sym__arithmetic_ternary_expression,
    STATE(2809), 1,
      sym__arithmetic_unary_expression,
    STATE(2825), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9282), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9284), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(2753), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [209841] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2263), 1,
      anon_sym_DOLLAR,
    ACTIONS(2269), 1,
      aux_sym_number_token1,
    ACTIONS(2271), 1,
      aux_sym_number_token2,
    ACTIONS(2275), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2289), 1,
      sym__brace_start,
    ACTIONS(8066), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8070), 1,
      anon_sym_DQUOTE,
    ACTIONS(8074), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8076), 1,
      anon_sym_BQUOTE,
    ACTIONS(8078), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9580), 1,
      sym_word,
    ACTIONS(9588), 1,
      sym__comment_word,
    ACTIONS(8062), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8080), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9584), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9586), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1623), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [209909] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1414), 1,
      anon_sym_DOLLAR,
    ACTIONS(1420), 1,
      aux_sym_number_token1,
    ACTIONS(1422), 1,
      aux_sym_number_token2,
    ACTIONS(1426), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1438), 1,
      sym__brace_start,
    ACTIONS(9094), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9096), 1,
      sym__special_character,
    ACTIONS(9098), 1,
      anon_sym_DQUOTE,
    ACTIONS(9102), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9104), 1,
      anon_sym_BQUOTE,
    ACTIONS(9106), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9892), 1,
      sym_word,
    ACTIONS(9896), 1,
      sym_test_operator,
    STATE(1149), 1,
      aux_sym__literal_repeat1,
    STATE(1281), 1,
      sym_concatenation,
    ACTIONS(9092), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9108), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9894), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(832), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [209981] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3602), 1,
      sym_variable_name,
    ACTIONS(9898), 1,
      anon_sym_LPAREN,
    ACTIONS(9900), 1,
      anon_sym_BANG,
    ACTIONS(9906), 1,
      anon_sym_TILDE,
    ACTIONS(9908), 1,
      anon_sym_DOLLAR,
    ACTIONS(9910), 1,
      anon_sym_DQUOTE,
    ACTIONS(9912), 1,
      aux_sym_number_token1,
    ACTIONS(9914), 1,
      aux_sym_number_token2,
    ACTIONS(9916), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9918), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9920), 1,
      anon_sym_BQUOTE,
    ACTIONS(9922), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9924), 1,
      aux_sym__simple_variable_name_token1,
    STATE(3359), 1,
      sym__arithmetic_postfix_expression,
    STATE(3361), 1,
      sym__arithmetic_unary_expression,
    STATE(3367), 1,
      sym__arithmetic_ternary_expression,
    STATE(3375), 1,
      sym__arithmetic_binary_expression,
    ACTIONS(9902), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9904), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(3188), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [210055] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1414), 1,
      anon_sym_DOLLAR,
    ACTIONS(1420), 1,
      aux_sym_number_token1,
    ACTIONS(1422), 1,
      aux_sym_number_token2,
    ACTIONS(1426), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1438), 1,
      sym__brace_start,
    ACTIONS(9094), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9096), 1,
      sym__special_character,
    ACTIONS(9098), 1,
      anon_sym_DQUOTE,
    ACTIONS(9102), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9104), 1,
      anon_sym_BQUOTE,
    ACTIONS(9106), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9926), 1,
      sym_word,
    ACTIONS(9930), 1,
      sym_test_operator,
    STATE(1146), 1,
      aux_sym__literal_repeat1,
    STATE(1301), 1,
      sym_concatenation,
    ACTIONS(9092), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9108), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9928), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(880), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [210127] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3787), 1,
      anon_sym_DOLLAR,
    ACTIONS(3793), 1,
      aux_sym_number_token1,
    ACTIONS(3795), 1,
      aux_sym_number_token2,
    ACTIONS(3799), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3809), 1,
      sym__brace_start,
    ACTIONS(9636), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9640), 1,
      anon_sym_DQUOTE,
    ACTIONS(9644), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9646), 1,
      anon_sym_BQUOTE,
    ACTIONS(9648), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9872), 1,
      sym_word,
    ACTIONS(9880), 1,
      sym__comment_word,
    ACTIONS(9634), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9650), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9876), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9878), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(4337), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [210195] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3616), 1,
      aux_sym_number_token1,
    ACTIONS(3618), 1,
      aux_sym_number_token2,
    ACTIONS(3622), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3632), 1,
      sym__brace_start,
    ACTIONS(8850), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8854), 1,
      anon_sym_DQUOTE,
    ACTIONS(8858), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8860), 1,
      anon_sym_BQUOTE,
    ACTIONS(8862), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9516), 1,
      sym_word,
    ACTIONS(9524), 1,
      sym__comment_word,
    ACTIONS(9932), 1,
      anon_sym_DOLLAR,
    ACTIONS(8848), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8864), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9520), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9522), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2210), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [210263] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9364), 1,
      anon_sym_LPAREN,
    ACTIONS(9366), 1,
      anon_sym_BANG,
    ACTIONS(9372), 1,
      anon_sym_TILDE,
    ACTIONS(9374), 1,
      anon_sym_DOLLAR,
    ACTIONS(9376), 1,
      anon_sym_DQUOTE,
    ACTIONS(9378), 1,
      aux_sym_number_token1,
    ACTIONS(9380), 1,
      aux_sym_number_token2,
    ACTIONS(9382), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9384), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9386), 1,
      anon_sym_BQUOTE,
    ACTIONS(9388), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9392), 1,
      sym_variable_name,
    ACTIONS(9934), 1,
      aux_sym__simple_variable_name_token1,
    STATE(3391), 1,
      sym__arithmetic_binary_expression,
    STATE(3396), 1,
      sym__arithmetic_ternary_expression,
    STATE(3398), 1,
      sym__arithmetic_unary_expression,
    STATE(3400), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9368), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9370), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(3337), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [210337] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(2863), 1,
      anon_sym_DOLLAR,
    ACTIONS(2869), 1,
      aux_sym_number_token1,
    ACTIONS(2871), 1,
      aux_sym_number_token2,
    ACTIONS(2875), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2885), 1,
      sym__brace_start,
    ACTIONS(9936), 1,
      sym_word,
    ACTIONS(9940), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9942), 1,
      sym__special_character,
    ACTIONS(9944), 1,
      anon_sym_DQUOTE,
    ACTIONS(9948), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9950), 1,
      anon_sym_BQUOTE,
    ACTIONS(9952), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9956), 1,
      sym_test_operator,
    STATE(4382), 1,
      aux_sym__literal_repeat1,
    STATE(4827), 1,
      sym_concatenation,
    ACTIONS(9938), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9946), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(9954), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(4242), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [210409] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5405), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(5411), 1,
      anon_sym_DQUOTE,
    ACTIONS(5415), 1,
      aux_sym_number_token1,
    ACTIONS(5417), 1,
      aux_sym_number_token2,
    ACTIONS(5419), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(5421), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(5423), 1,
      anon_sym_BQUOTE,
    ACTIONS(5425), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(5431), 1,
      sym__brace_start,
    ACTIONS(9958), 1,
      sym_word,
    ACTIONS(9960), 1,
      anon_sym_DOLLAR,
    ACTIONS(9966), 1,
      sym__comment_word,
    ACTIONS(5403), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(5427), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9962), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9964), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(5541), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [210477] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(2863), 1,
      anon_sym_DOLLAR,
    ACTIONS(2869), 1,
      aux_sym_number_token1,
    ACTIONS(2871), 1,
      aux_sym_number_token2,
    ACTIONS(2875), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2885), 1,
      sym__brace_start,
    ACTIONS(9940), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9942), 1,
      sym__special_character,
    ACTIONS(9944), 1,
      anon_sym_DQUOTE,
    ACTIONS(9948), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9950), 1,
      anon_sym_BQUOTE,
    ACTIONS(9952), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9968), 1,
      sym_word,
    ACTIONS(9972), 1,
      sym_test_operator,
    STATE(4290), 1,
      aux_sym__literal_repeat1,
    STATE(4709), 1,
      sym_concatenation,
    ACTIONS(9938), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9954), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9970), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(4260), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [210549] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(8474), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8476), 1,
      anon_sym_DOLLAR,
    ACTIONS(8478), 1,
      sym__special_character,
    ACTIONS(8480), 1,
      anon_sym_DQUOTE,
    ACTIONS(8484), 1,
      aux_sym_number_token1,
    ACTIONS(8486), 1,
      aux_sym_number_token2,
    ACTIONS(8488), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8490), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(8492), 1,
      anon_sym_BQUOTE,
    ACTIONS(8494), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8504), 1,
      sym__brace_start,
    ACTIONS(9974), 1,
      sym_word,
    ACTIONS(9978), 1,
      sym_test_operator,
    STATE(1828), 1,
      aux_sym__literal_repeat1,
    STATE(2227), 1,
      sym_concatenation,
    ACTIONS(8470), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8496), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9976), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1342), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [210621] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3064), 1,
      sym_variable_name,
    ACTIONS(9278), 1,
      anon_sym_LPAREN,
    ACTIONS(9280), 1,
      anon_sym_BANG,
    ACTIONS(9286), 1,
      anon_sym_TILDE,
    ACTIONS(9288), 1,
      anon_sym_DOLLAR,
    ACTIONS(9290), 1,
      anon_sym_DQUOTE,
    ACTIONS(9292), 1,
      aux_sym_number_token1,
    ACTIONS(9294), 1,
      aux_sym_number_token2,
    ACTIONS(9296), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9298), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9300), 1,
      anon_sym_BQUOTE,
    ACTIONS(9302), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9980), 1,
      aux_sym__simple_variable_name_token1,
    STATE(2801), 1,
      sym__arithmetic_binary_expression,
    STATE(2803), 1,
      sym__arithmetic_ternary_expression,
    STATE(2809), 1,
      sym__arithmetic_unary_expression,
    STATE(2825), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9282), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9284), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(2932), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [210695] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2309), 1,
      aux_sym_number_token1,
    ACTIONS(2311), 1,
      aux_sym_number_token2,
    ACTIONS(2315), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2327), 1,
      sym__brace_start,
    ACTIONS(9982), 1,
      sym_word,
    ACTIONS(9986), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9988), 1,
      anon_sym_DOLLAR,
    ACTIONS(9992), 1,
      anon_sym_DQUOTE,
    ACTIONS(9996), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9998), 1,
      anon_sym_BQUOTE,
    ACTIONS(10000), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10004), 1,
      sym__comment_word,
    ACTIONS(9984), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9990), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10002), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9994), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1423), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [210763] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5405), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(5411), 1,
      anon_sym_DQUOTE,
    ACTIONS(5415), 1,
      aux_sym_number_token1,
    ACTIONS(5417), 1,
      aux_sym_number_token2,
    ACTIONS(5419), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(5421), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(5423), 1,
      anon_sym_BQUOTE,
    ACTIONS(5425), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(5431), 1,
      sym__brace_start,
    ACTIONS(9958), 1,
      sym_word,
    ACTIONS(9966), 1,
      sym__comment_word,
    ACTIONS(10006), 1,
      anon_sym_DOLLAR,
    ACTIONS(5403), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(5427), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9962), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9964), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(5541), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [210831] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9364), 1,
      anon_sym_LPAREN,
    ACTIONS(9366), 1,
      anon_sym_BANG,
    ACTIONS(9372), 1,
      anon_sym_TILDE,
    ACTIONS(9374), 1,
      anon_sym_DOLLAR,
    ACTIONS(9376), 1,
      anon_sym_DQUOTE,
    ACTIONS(9378), 1,
      aux_sym_number_token1,
    ACTIONS(9380), 1,
      aux_sym_number_token2,
    ACTIONS(9382), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9384), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9386), 1,
      anon_sym_BQUOTE,
    ACTIONS(9388), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9392), 1,
      sym_variable_name,
    ACTIONS(10008), 1,
      aux_sym__simple_variable_name_token1,
    STATE(3391), 1,
      sym__arithmetic_binary_expression,
    STATE(3396), 1,
      sym__arithmetic_ternary_expression,
    STATE(3398), 1,
      sym__arithmetic_unary_expression,
    STATE(3400), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9368), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9370), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(3366), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [210905] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(111), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(117), 1,
      anon_sym_DQUOTE,
    ACTIONS(121), 1,
      aux_sym_number_token1,
    ACTIONS(123), 1,
      aux_sym_number_token2,
    ACTIONS(125), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(127), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(129), 1,
      anon_sym_BQUOTE,
    ACTIONS(131), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(141), 1,
      sym__brace_start,
    ACTIONS(9340), 1,
      sym_word,
    ACTIONS(9346), 1,
      sym__comment_word,
    ACTIONS(10010), 1,
      anon_sym_DOLLAR,
    ACTIONS(109), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(133), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9342), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9344), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(482), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [210973] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2309), 1,
      aux_sym_number_token1,
    ACTIONS(2311), 1,
      aux_sym_number_token2,
    ACTIONS(2315), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2327), 1,
      sym__brace_start,
    ACTIONS(9982), 1,
      sym_word,
    ACTIONS(9986), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9992), 1,
      anon_sym_DQUOTE,
    ACTIONS(9996), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9998), 1,
      anon_sym_BQUOTE,
    ACTIONS(10000), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10004), 1,
      sym__comment_word,
    ACTIONS(10012), 1,
      anon_sym_DOLLAR,
    ACTIONS(9984), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9990), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10002), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9994), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1423), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [211041] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8571), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8573), 1,
      anon_sym_DOLLAR,
    ACTIONS(8577), 1,
      anon_sym_DQUOTE,
    ACTIONS(8581), 1,
      aux_sym_number_token1,
    ACTIONS(8583), 1,
      aux_sym_number_token2,
    ACTIONS(8585), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8587), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(8589), 1,
      anon_sym_BQUOTE,
    ACTIONS(8591), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8597), 1,
      sym__brace_start,
    ACTIONS(10014), 1,
      sym_word,
    ACTIONS(10020), 1,
      sym__comment_word,
    ACTIONS(8567), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8593), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10016), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10018), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(5199), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [211109] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3064), 1,
      sym_variable_name,
    ACTIONS(9278), 1,
      anon_sym_LPAREN,
    ACTIONS(9280), 1,
      anon_sym_BANG,
    ACTIONS(9286), 1,
      anon_sym_TILDE,
    ACTIONS(9288), 1,
      anon_sym_DOLLAR,
    ACTIONS(9290), 1,
      anon_sym_DQUOTE,
    ACTIONS(9292), 1,
      aux_sym_number_token1,
    ACTIONS(9294), 1,
      aux_sym_number_token2,
    ACTIONS(9296), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9298), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9300), 1,
      anon_sym_BQUOTE,
    ACTIONS(9302), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10022), 1,
      aux_sym__simple_variable_name_token1,
    STATE(2801), 1,
      sym__arithmetic_binary_expression,
    STATE(2803), 1,
      sym__arithmetic_ternary_expression,
    STATE(2809), 1,
      sym__arithmetic_unary_expression,
    STATE(2825), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9282), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9284), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(2754), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [211183] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3064), 1,
      sym_variable_name,
    ACTIONS(9278), 1,
      anon_sym_LPAREN,
    ACTIONS(9280), 1,
      anon_sym_BANG,
    ACTIONS(9286), 1,
      anon_sym_TILDE,
    ACTIONS(9288), 1,
      anon_sym_DOLLAR,
    ACTIONS(9290), 1,
      anon_sym_DQUOTE,
    ACTIONS(9292), 1,
      aux_sym_number_token1,
    ACTIONS(9294), 1,
      aux_sym_number_token2,
    ACTIONS(9296), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9298), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9300), 1,
      anon_sym_BQUOTE,
    ACTIONS(9302), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10024), 1,
      aux_sym__simple_variable_name_token1,
    STATE(2801), 1,
      sym__arithmetic_binary_expression,
    STATE(2803), 1,
      sym__arithmetic_ternary_expression,
    STATE(2809), 1,
      sym__arithmetic_unary_expression,
    STATE(2825), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9282), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9284), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(2792), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [211257] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9364), 1,
      anon_sym_LPAREN,
    ACTIONS(9366), 1,
      anon_sym_BANG,
    ACTIONS(9372), 1,
      anon_sym_TILDE,
    ACTIONS(9374), 1,
      anon_sym_DOLLAR,
    ACTIONS(9376), 1,
      anon_sym_DQUOTE,
    ACTIONS(9378), 1,
      aux_sym_number_token1,
    ACTIONS(9380), 1,
      aux_sym_number_token2,
    ACTIONS(9382), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9384), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9386), 1,
      anon_sym_BQUOTE,
    ACTIONS(9388), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9392), 1,
      sym_variable_name,
    ACTIONS(10026), 1,
      aux_sym__simple_variable_name_token1,
    STATE(3391), 1,
      sym__arithmetic_binary_expression,
    STATE(3396), 1,
      sym__arithmetic_ternary_expression,
    STATE(3398), 1,
      sym__arithmetic_unary_expression,
    STATE(3400), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9368), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9370), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(3336), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [211331] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3616), 1,
      aux_sym_number_token1,
    ACTIONS(3618), 1,
      aux_sym_number_token2,
    ACTIONS(3622), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3632), 1,
      sym__brace_start,
    ACTIONS(8850), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8854), 1,
      anon_sym_DQUOTE,
    ACTIONS(8858), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8860), 1,
      anon_sym_BQUOTE,
    ACTIONS(8862), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9516), 1,
      sym_word,
    ACTIONS(9524), 1,
      sym__comment_word,
    ACTIONS(10028), 1,
      anon_sym_DOLLAR,
    ACTIONS(8848), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8864), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9520), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9522), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2210), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [211399] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9364), 1,
      anon_sym_LPAREN,
    ACTIONS(9366), 1,
      anon_sym_BANG,
    ACTIONS(9372), 1,
      anon_sym_TILDE,
    ACTIONS(9374), 1,
      anon_sym_DOLLAR,
    ACTIONS(9376), 1,
      anon_sym_DQUOTE,
    ACTIONS(9378), 1,
      aux_sym_number_token1,
    ACTIONS(9380), 1,
      aux_sym_number_token2,
    ACTIONS(9382), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9384), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9386), 1,
      anon_sym_BQUOTE,
    ACTIONS(9388), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9392), 1,
      sym_variable_name,
    ACTIONS(10030), 1,
      aux_sym__simple_variable_name_token1,
    STATE(3391), 1,
      sym__arithmetic_binary_expression,
    STATE(3396), 1,
      sym__arithmetic_ternary_expression,
    STATE(3398), 1,
      sym__arithmetic_unary_expression,
    STATE(3400), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9368), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9370), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(3369), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [211473] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(8268), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8270), 1,
      anon_sym_DOLLAR,
    ACTIONS(8272), 1,
      sym__special_character,
    ACTIONS(8274), 1,
      anon_sym_DQUOTE,
    ACTIONS(8278), 1,
      aux_sym_number_token1,
    ACTIONS(8280), 1,
      aux_sym_number_token2,
    ACTIONS(8282), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8284), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(8286), 1,
      anon_sym_BQUOTE,
    ACTIONS(8288), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8298), 1,
      sym__brace_start,
    ACTIONS(10032), 1,
      sym_word,
    ACTIONS(10036), 1,
      sym_test_operator,
    STATE(1539), 1,
      aux_sym__literal_repeat1,
    STATE(1872), 1,
      sym_concatenation,
    ACTIONS(8264), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8290), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10034), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1263), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [211545] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(8268), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8270), 1,
      anon_sym_DOLLAR,
    ACTIONS(8272), 1,
      sym__special_character,
    ACTIONS(8274), 1,
      anon_sym_DQUOTE,
    ACTIONS(8278), 1,
      aux_sym_number_token1,
    ACTIONS(8280), 1,
      aux_sym_number_token2,
    ACTIONS(8282), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8284), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(8286), 1,
      anon_sym_BQUOTE,
    ACTIONS(8288), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8298), 1,
      sym__brace_start,
    ACTIONS(10038), 1,
      sym_word,
    ACTIONS(10042), 1,
      sym_test_operator,
    STATE(1509), 1,
      aux_sym__literal_repeat1,
    STATE(1883), 1,
      sym_concatenation,
    ACTIONS(8264), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8290), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10040), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1258), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [211617] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3731), 1,
      anon_sym_DOLLAR,
    ACTIONS(3737), 1,
      aux_sym_number_token1,
    ACTIONS(3739), 1,
      aux_sym_number_token2,
    ACTIONS(3743), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3755), 1,
      sym__brace_start,
    ACTIONS(8820), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8824), 1,
      anon_sym_DQUOTE,
    ACTIONS(8828), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8830), 1,
      anon_sym_BQUOTE,
    ACTIONS(8832), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9412), 1,
      sym_word,
    ACTIONS(9420), 1,
      sym__comment_word,
    ACTIONS(8818), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8834), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9416), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9418), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(4810), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [211685] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3064), 1,
      sym_variable_name,
    ACTIONS(9278), 1,
      anon_sym_LPAREN,
    ACTIONS(9280), 1,
      anon_sym_BANG,
    ACTIONS(9286), 1,
      anon_sym_TILDE,
    ACTIONS(9288), 1,
      anon_sym_DOLLAR,
    ACTIONS(9290), 1,
      anon_sym_DQUOTE,
    ACTIONS(9292), 1,
      aux_sym_number_token1,
    ACTIONS(9294), 1,
      aux_sym_number_token2,
    ACTIONS(9296), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9298), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9300), 1,
      anon_sym_BQUOTE,
    ACTIONS(9302), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10044), 1,
      aux_sym__simple_variable_name_token1,
    STATE(2801), 1,
      sym__arithmetic_binary_expression,
    STATE(2803), 1,
      sym__arithmetic_ternary_expression,
    STATE(2809), 1,
      sym__arithmetic_unary_expression,
    STATE(2825), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9282), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9284), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(2793), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [211759] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9364), 1,
      anon_sym_LPAREN,
    ACTIONS(9366), 1,
      anon_sym_BANG,
    ACTIONS(9372), 1,
      anon_sym_TILDE,
    ACTIONS(9374), 1,
      anon_sym_DOLLAR,
    ACTIONS(9376), 1,
      anon_sym_DQUOTE,
    ACTIONS(9378), 1,
      aux_sym_number_token1,
    ACTIONS(9380), 1,
      aux_sym_number_token2,
    ACTIONS(9382), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9384), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9386), 1,
      anon_sym_BQUOTE,
    ACTIONS(9388), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9392), 1,
      sym_variable_name,
    ACTIONS(10046), 1,
      aux_sym__simple_variable_name_token1,
    STATE(3391), 1,
      sym__arithmetic_binary_expression,
    STATE(3396), 1,
      sym__arithmetic_ternary_expression,
    STATE(3398), 1,
      sym__arithmetic_unary_expression,
    STATE(3400), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9368), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9370), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(3370), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [211833] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9118), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9124), 1,
      anon_sym_DQUOTE,
    ACTIONS(9128), 1,
      aux_sym_number_token1,
    ACTIONS(9130), 1,
      aux_sym_number_token2,
    ACTIONS(9132), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9134), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9136), 1,
      anon_sym_BQUOTE,
    ACTIONS(9138), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9144), 1,
      sym__brace_start,
    ACTIONS(9442), 1,
      sym_word,
    ACTIONS(9448), 1,
      sym__comment_word,
    ACTIONS(10048), 1,
      anon_sym_DOLLAR,
    ACTIONS(9116), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9140), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9444), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9446), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2401), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [211901] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9758), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9760), 1,
      anon_sym_DOLLAR,
    ACTIONS(9764), 1,
      anon_sym_DQUOTE,
    ACTIONS(9768), 1,
      aux_sym_number_token1,
    ACTIONS(9770), 1,
      aux_sym_number_token2,
    ACTIONS(9772), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9774), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9776), 1,
      anon_sym_BQUOTE,
    ACTIONS(9778), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9784), 1,
      sym__brace_start,
    ACTIONS(10050), 1,
      sym_word,
    ACTIONS(10052), 1,
      sym__special_character,
    ACTIONS(10056), 1,
      sym_test_operator,
    STATE(6219), 1,
      aux_sym__literal_repeat1,
    STATE(6469), 1,
      sym_concatenation,
    ACTIONS(9756), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9780), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10054), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(6192), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [211973] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3675), 1,
      anon_sym_DOLLAR,
    ACTIONS(3681), 1,
      aux_sym_number_token1,
    ACTIONS(3683), 1,
      aux_sym_number_token2,
    ACTIONS(3687), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3697), 1,
      sym__brace_start,
    ACTIONS(10058), 1,
      sym_word,
    ACTIONS(10062), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(10064), 1,
      sym__special_character,
    ACTIONS(10066), 1,
      anon_sym_DQUOTE,
    ACTIONS(10070), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(10072), 1,
      anon_sym_BQUOTE,
    ACTIONS(10074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10078), 1,
      sym_test_operator,
    STATE(4645), 1,
      aux_sym__literal_repeat1,
    STATE(5086), 1,
      sym_concatenation,
    ACTIONS(10060), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(10068), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(10076), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(4430), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [212045] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3612), 1,
      anon_sym_DOLLAR,
    ACTIONS(3616), 1,
      aux_sym_number_token1,
    ACTIONS(3618), 1,
      aux_sym_number_token2,
    ACTIONS(3622), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3632), 1,
      sym__brace_start,
    ACTIONS(8850), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8854), 1,
      anon_sym_DQUOTE,
    ACTIONS(8858), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8860), 1,
      anon_sym_BQUOTE,
    ACTIONS(8862), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9516), 1,
      sym_word,
    ACTIONS(9524), 1,
      sym__comment_word,
    ACTIONS(8848), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8864), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9520), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9522), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2210), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [212113] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3064), 1,
      sym_variable_name,
    ACTIONS(3111), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(9278), 1,
      anon_sym_LPAREN,
    ACTIONS(9280), 1,
      anon_sym_BANG,
    ACTIONS(9286), 1,
      anon_sym_TILDE,
    ACTIONS(9288), 1,
      anon_sym_DOLLAR,
    ACTIONS(9290), 1,
      anon_sym_DQUOTE,
    ACTIONS(9292), 1,
      aux_sym_number_token1,
    ACTIONS(9294), 1,
      aux_sym_number_token2,
    ACTIONS(9296), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9298), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9300), 1,
      anon_sym_BQUOTE,
    ACTIONS(9302), 1,
      anon_sym_DOLLAR_BQUOTE,
    STATE(2801), 1,
      sym__arithmetic_binary_expression,
    STATE(2803), 1,
      sym__arithmetic_ternary_expression,
    STATE(2809), 1,
      sym__arithmetic_unary_expression,
    STATE(2825), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9282), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9284), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(2799), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [212187] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9364), 1,
      anon_sym_LPAREN,
    ACTIONS(9366), 1,
      anon_sym_BANG,
    ACTIONS(9372), 1,
      anon_sym_TILDE,
    ACTIONS(9374), 1,
      anon_sym_DOLLAR,
    ACTIONS(9376), 1,
      anon_sym_DQUOTE,
    ACTIONS(9378), 1,
      aux_sym_number_token1,
    ACTIONS(9380), 1,
      aux_sym_number_token2,
    ACTIONS(9382), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9384), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9386), 1,
      anon_sym_BQUOTE,
    ACTIONS(9388), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9392), 1,
      sym_variable_name,
    ACTIONS(10080), 1,
      aux_sym__simple_variable_name_token1,
    STATE(3391), 1,
      sym__arithmetic_binary_expression,
    STATE(3396), 1,
      sym__arithmetic_ternary_expression,
    STATE(3398), 1,
      sym__arithmetic_unary_expression,
    STATE(3400), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9368), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9370), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(3435), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [212261] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1454), 1,
      anon_sym_DOLLAR,
    ACTIONS(1460), 1,
      aux_sym_number_token1,
    ACTIONS(1462), 1,
      aux_sym_number_token2,
    ACTIONS(1466), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1478), 1,
      sym__brace_start,
    ACTIONS(8878), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8880), 1,
      sym__special_character,
    ACTIONS(8882), 1,
      anon_sym_DQUOTE,
    ACTIONS(8886), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8888), 1,
      anon_sym_BQUOTE,
    ACTIONS(8890), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10082), 1,
      sym_word,
    ACTIONS(10086), 1,
      sym_test_operator,
    STATE(1202), 1,
      aux_sym__literal_repeat1,
    STATE(1390), 1,
      sym_concatenation,
    ACTIONS(8876), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8892), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10084), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(993), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [212333] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1454), 1,
      anon_sym_DOLLAR,
    ACTIONS(1460), 1,
      aux_sym_number_token1,
    ACTIONS(1462), 1,
      aux_sym_number_token2,
    ACTIONS(1466), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1478), 1,
      sym__brace_start,
    ACTIONS(8878), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8880), 1,
      sym__special_character,
    ACTIONS(8882), 1,
      anon_sym_DQUOTE,
    ACTIONS(8886), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8888), 1,
      anon_sym_BQUOTE,
    ACTIONS(8890), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10088), 1,
      sym_word,
    ACTIONS(10092), 1,
      sym_test_operator,
    STATE(1201), 1,
      aux_sym__literal_repeat1,
    STATE(1387), 1,
      sym_concatenation,
    ACTIONS(8876), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8892), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10090), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(995), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [212405] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1593), 1,
      aux_sym_number_token1,
    ACTIONS(1595), 1,
      aux_sym_number_token2,
    ACTIONS(1599), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1611), 1,
      sym__brace_start,
    ACTIONS(9240), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9244), 1,
      anon_sym_DQUOTE,
    ACTIONS(9248), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9250), 1,
      anon_sym_BQUOTE,
    ACTIONS(9252), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10094), 1,
      sym_word,
    ACTIONS(10096), 1,
      anon_sym_DOLLAR,
    ACTIONS(10102), 1,
      sym__comment_word,
    ACTIONS(9238), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9254), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10098), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10100), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1293), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [212473] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3064), 1,
      sym_variable_name,
    ACTIONS(9278), 1,
      anon_sym_LPAREN,
    ACTIONS(9280), 1,
      anon_sym_BANG,
    ACTIONS(9286), 1,
      anon_sym_TILDE,
    ACTIONS(9288), 1,
      anon_sym_DOLLAR,
    ACTIONS(9290), 1,
      anon_sym_DQUOTE,
    ACTIONS(9292), 1,
      aux_sym_number_token1,
    ACTIONS(9294), 1,
      aux_sym_number_token2,
    ACTIONS(9296), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9298), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9300), 1,
      anon_sym_BQUOTE,
    ACTIONS(9302), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10104), 1,
      aux_sym__simple_variable_name_token1,
    STATE(2801), 1,
      sym__arithmetic_binary_expression,
    STATE(2803), 1,
      sym__arithmetic_ternary_expression,
    STATE(2809), 1,
      sym__arithmetic_unary_expression,
    STATE(2825), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9282), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9284), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(2755), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [212547] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(4080), 1,
      anon_sym_DOLLAR,
    ACTIONS(4086), 1,
      aux_sym_number_token1,
    ACTIONS(4088), 1,
      aux_sym_number_token2,
    ACTIONS(4092), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(4102), 1,
      sym__brace_start,
    ACTIONS(9310), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9316), 1,
      anon_sym_DQUOTE,
    ACTIONS(9320), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9322), 1,
      anon_sym_BQUOTE,
    ACTIONS(9324), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10106), 1,
      sym_word,
    ACTIONS(10108), 1,
      sym__special_character,
    ACTIONS(10112), 1,
      sym_test_operator,
    STATE(4421), 1,
      aux_sym__literal_repeat1,
    STATE(4869), 1,
      sym_concatenation,
    ACTIONS(9308), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9326), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10110), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(4673), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [212619] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3397), 1,
      anon_sym_DOLLAR,
    ACTIONS(3401), 1,
      aux_sym_number_token1,
    ACTIONS(3403), 1,
      aux_sym_number_token2,
    ACTIONS(3407), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3417), 1,
      sym__brace_start,
    ACTIONS(8926), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8930), 1,
      anon_sym_DQUOTE,
    ACTIONS(8934), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8936), 1,
      anon_sym_BQUOTE,
    ACTIONS(8938), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9696), 1,
      sym_word,
    ACTIONS(9704), 1,
      sym__comment_word,
    ACTIONS(8924), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8940), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9700), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9702), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1851), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [212687] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(4080), 1,
      anon_sym_DOLLAR,
    ACTIONS(4086), 1,
      aux_sym_number_token1,
    ACTIONS(4088), 1,
      aux_sym_number_token2,
    ACTIONS(4092), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(4102), 1,
      sym__brace_start,
    ACTIONS(9310), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9316), 1,
      anon_sym_DQUOTE,
    ACTIONS(9320), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9322), 1,
      anon_sym_BQUOTE,
    ACTIONS(9324), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10108), 1,
      sym__special_character,
    ACTIONS(10114), 1,
      sym_word,
    ACTIONS(10118), 1,
      sym_test_operator,
    STATE(4500), 1,
      aux_sym__literal_repeat1,
    STATE(5029), 1,
      sym_concatenation,
    ACTIONS(9308), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9326), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10116), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(4760), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [212759] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1420), 1,
      aux_sym_number_token1,
    ACTIONS(1422), 1,
      aux_sym_number_token2,
    ACTIONS(1426), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1438), 1,
      sym__brace_start,
    ACTIONS(9094), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9098), 1,
      anon_sym_DQUOTE,
    ACTIONS(9102), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9104), 1,
      anon_sym_BQUOTE,
    ACTIONS(9106), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10120), 1,
      sym_word,
    ACTIONS(10122), 1,
      anon_sym_DOLLAR,
    ACTIONS(10128), 1,
      sym__comment_word,
    ACTIONS(9092), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9108), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10124), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10126), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(910), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [212827] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3026), 1,
      anon_sym_DOLLAR,
    ACTIONS(3032), 1,
      aux_sym_number_token1,
    ACTIONS(3034), 1,
      aux_sym_number_token2,
    ACTIONS(3038), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3048), 1,
      sym__brace_start,
    ACTIONS(10130), 1,
      sym_word,
    ACTIONS(10134), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(10136), 1,
      sym__special_character,
    ACTIONS(10138), 1,
      anon_sym_DQUOTE,
    ACTIONS(10142), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(10144), 1,
      anon_sym_BQUOTE,
    ACTIONS(10146), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10150), 1,
      sym_test_operator,
    STATE(4492), 1,
      aux_sym__literal_repeat1,
    STATE(4893), 1,
      sym_concatenation,
    ACTIONS(10132), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(10140), 2,
      sym_raw_string,
      sym_ansi_c_string,
    ACTIONS(10148), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    STATE(4274), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [212899] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3401), 1,
      aux_sym_number_token1,
    ACTIONS(3403), 1,
      aux_sym_number_token2,
    ACTIONS(3407), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3417), 1,
      sym__brace_start,
    ACTIONS(8926), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8930), 1,
      anon_sym_DQUOTE,
    ACTIONS(8934), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8936), 1,
      anon_sym_BQUOTE,
    ACTIONS(8938), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9696), 1,
      sym_word,
    ACTIONS(9704), 1,
      sym__comment_word,
    ACTIONS(10152), 1,
      anon_sym_DOLLAR,
    ACTIONS(8924), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8940), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9700), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9702), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1851), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [212967] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3064), 1,
      sym_variable_name,
    ACTIONS(9278), 1,
      anon_sym_LPAREN,
    ACTIONS(9280), 1,
      anon_sym_BANG,
    ACTIONS(9286), 1,
      anon_sym_TILDE,
    ACTIONS(9288), 1,
      anon_sym_DOLLAR,
    ACTIONS(9290), 1,
      anon_sym_DQUOTE,
    ACTIONS(9292), 1,
      aux_sym_number_token1,
    ACTIONS(9294), 1,
      aux_sym_number_token2,
    ACTIONS(9296), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9298), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9300), 1,
      anon_sym_BQUOTE,
    ACTIONS(9302), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10154), 1,
      aux_sym__simple_variable_name_token1,
    STATE(2801), 1,
      sym__arithmetic_binary_expression,
    STATE(2803), 1,
      sym__arithmetic_ternary_expression,
    STATE(2809), 1,
      sym__arithmetic_unary_expression,
    STATE(2825), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9282), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9284), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(3195), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [213041] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1420), 1,
      aux_sym_number_token1,
    ACTIONS(1422), 1,
      aux_sym_number_token2,
    ACTIONS(1426), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1438), 1,
      sym__brace_start,
    ACTIONS(9094), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9098), 1,
      anon_sym_DQUOTE,
    ACTIONS(9102), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9104), 1,
      anon_sym_BQUOTE,
    ACTIONS(9106), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10120), 1,
      sym_word,
    ACTIONS(10128), 1,
      sym__comment_word,
    ACTIONS(10156), 1,
      anon_sym_DOLLAR,
    ACTIONS(9092), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9108), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10124), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10126), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(910), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [213109] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(7958), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(7964), 1,
      anon_sym_DQUOTE,
    ACTIONS(7968), 1,
      aux_sym_number_token1,
    ACTIONS(7970), 1,
      aux_sym_number_token2,
    ACTIONS(7972), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(7974), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(7976), 1,
      anon_sym_BQUOTE,
    ACTIONS(7978), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(7988), 1,
      sym__brace_start,
    ACTIONS(9860), 1,
      sym_word,
    ACTIONS(9868), 1,
      sym__comment_word,
    ACTIONS(10158), 1,
      anon_sym_DOLLAR,
    ACTIONS(7954), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(7980), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9864), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9866), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(4357), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [213177] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(7958), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(7964), 1,
      anon_sym_DQUOTE,
    ACTIONS(7968), 1,
      aux_sym_number_token1,
    ACTIONS(7970), 1,
      aux_sym_number_token2,
    ACTIONS(7972), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(7974), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(7976), 1,
      anon_sym_BQUOTE,
    ACTIONS(7978), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(7988), 1,
      sym__brace_start,
    ACTIONS(9860), 1,
      sym_word,
    ACTIONS(9868), 1,
      sym__comment_word,
    ACTIONS(10160), 1,
      anon_sym_DOLLAR,
    ACTIONS(7954), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(7980), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9864), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9866), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(4357), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [213245] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3793), 1,
      aux_sym_number_token1,
    ACTIONS(3795), 1,
      aux_sym_number_token2,
    ACTIONS(3799), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3809), 1,
      sym__brace_start,
    ACTIONS(9636), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9640), 1,
      anon_sym_DQUOTE,
    ACTIONS(9644), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9646), 1,
      anon_sym_BQUOTE,
    ACTIONS(9648), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9872), 1,
      sym_word,
    ACTIONS(9880), 1,
      sym__comment_word,
    ACTIONS(10162), 1,
      anon_sym_DOLLAR,
    ACTIONS(9634), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9650), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9876), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9878), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(4337), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [213313] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3793), 1,
      aux_sym_number_token1,
    ACTIONS(3795), 1,
      aux_sym_number_token2,
    ACTIONS(3799), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3809), 1,
      sym__brace_start,
    ACTIONS(9636), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9640), 1,
      anon_sym_DQUOTE,
    ACTIONS(9644), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9646), 1,
      anon_sym_BQUOTE,
    ACTIONS(9648), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9872), 1,
      sym_word,
    ACTIONS(9880), 1,
      sym__comment_word,
    ACTIONS(10164), 1,
      anon_sym_DOLLAR,
    ACTIONS(9634), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9650), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9876), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9878), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(4337), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [213381] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4086), 1,
      aux_sym_number_token1,
    ACTIONS(4088), 1,
      aux_sym_number_token2,
    ACTIONS(4092), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(4102), 1,
      sym__brace_start,
    ACTIONS(9306), 1,
      sym_word,
    ACTIONS(9310), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9316), 1,
      anon_sym_DQUOTE,
    ACTIONS(9320), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9322), 1,
      anon_sym_BQUOTE,
    ACTIONS(9324), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9328), 1,
      sym__comment_word,
    ACTIONS(10166), 1,
      anon_sym_DOLLAR,
    ACTIONS(9308), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9314), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9326), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9318), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(4566), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [213449] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3064), 1,
      sym_variable_name,
    ACTIONS(9278), 1,
      anon_sym_LPAREN,
    ACTIONS(9280), 1,
      anon_sym_BANG,
    ACTIONS(9286), 1,
      anon_sym_TILDE,
    ACTIONS(9288), 1,
      anon_sym_DOLLAR,
    ACTIONS(9290), 1,
      anon_sym_DQUOTE,
    ACTIONS(9292), 1,
      aux_sym_number_token1,
    ACTIONS(9294), 1,
      aux_sym_number_token2,
    ACTIONS(9296), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9298), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9300), 1,
      anon_sym_BQUOTE,
    ACTIONS(9302), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10168), 1,
      aux_sym__simple_variable_name_token1,
    STATE(2801), 1,
      sym__arithmetic_binary_expression,
    STATE(2803), 1,
      sym__arithmetic_ternary_expression,
    STATE(2809), 1,
      sym__arithmetic_unary_expression,
    STATE(2825), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9282), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9284), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(2794), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [213523] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9364), 1,
      anon_sym_LPAREN,
    ACTIONS(9366), 1,
      anon_sym_BANG,
    ACTIONS(9372), 1,
      anon_sym_TILDE,
    ACTIONS(9374), 1,
      anon_sym_DOLLAR,
    ACTIONS(9376), 1,
      anon_sym_DQUOTE,
    ACTIONS(9378), 1,
      aux_sym_number_token1,
    ACTIONS(9380), 1,
      aux_sym_number_token2,
    ACTIONS(9382), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9384), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9386), 1,
      anon_sym_BQUOTE,
    ACTIONS(9388), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9392), 1,
      sym_variable_name,
    ACTIONS(10170), 1,
      aux_sym__simple_variable_name_token1,
    STATE(3391), 1,
      sym__arithmetic_binary_expression,
    STATE(3396), 1,
      sym__arithmetic_ternary_expression,
    STATE(3398), 1,
      sym__arithmetic_unary_expression,
    STATE(3400), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9368), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9370), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(3371), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [213597] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4086), 1,
      aux_sym_number_token1,
    ACTIONS(4088), 1,
      aux_sym_number_token2,
    ACTIONS(4092), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(4102), 1,
      sym__brace_start,
    ACTIONS(9306), 1,
      sym_word,
    ACTIONS(9310), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9316), 1,
      anon_sym_DQUOTE,
    ACTIONS(9320), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9322), 1,
      anon_sym_BQUOTE,
    ACTIONS(9324), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9328), 1,
      sym__comment_word,
    ACTIONS(10172), 1,
      anon_sym_DOLLAR,
    ACTIONS(9308), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9314), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9326), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9318), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(4566), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [213665] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5112), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(5118), 1,
      anon_sym_DQUOTE,
    ACTIONS(5122), 1,
      aux_sym_number_token1,
    ACTIONS(5124), 1,
      aux_sym_number_token2,
    ACTIONS(5126), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(5128), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(5130), 1,
      anon_sym_BQUOTE,
    ACTIONS(5132), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(5138), 1,
      sym__brace_start,
    ACTIONS(10174), 1,
      sym_word,
    ACTIONS(10176), 1,
      anon_sym_DOLLAR,
    ACTIONS(10182), 1,
      sym__comment_word,
    ACTIONS(5110), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(5134), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10178), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10180), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2842), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [213733] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8178), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8180), 1,
      anon_sym_DOLLAR,
    ACTIONS(8184), 1,
      anon_sym_DQUOTE,
    ACTIONS(8188), 1,
      aux_sym_number_token1,
    ACTIONS(8190), 1,
      aux_sym_number_token2,
    ACTIONS(8192), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8194), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(8196), 1,
      anon_sym_BQUOTE,
    ACTIONS(8198), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8208), 1,
      sym__brace_start,
    ACTIONS(9720), 1,
      sym_word,
    ACTIONS(9728), 1,
      sym__comment_word,
    ACTIONS(8174), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8200), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9724), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9726), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1590), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [213801] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3886), 1,
      aux_sym_number_token1,
    ACTIONS(3888), 1,
      aux_sym_number_token2,
    ACTIONS(3892), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3902), 1,
      sym__brace_start,
    ACTIONS(8992), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8996), 1,
      anon_sym_DQUOTE,
    ACTIONS(9000), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9002), 1,
      anon_sym_BQUOTE,
    ACTIONS(9004), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9740), 1,
      sym_word,
    ACTIONS(9746), 1,
      sym__comment_word,
    ACTIONS(10184), 1,
      anon_sym_DOLLAR,
    ACTIONS(8990), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9006), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9742), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9744), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2004), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [213869] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9364), 1,
      anon_sym_LPAREN,
    ACTIONS(9366), 1,
      anon_sym_BANG,
    ACTIONS(9372), 1,
      anon_sym_TILDE,
    ACTIONS(9374), 1,
      anon_sym_DOLLAR,
    ACTIONS(9376), 1,
      anon_sym_DQUOTE,
    ACTIONS(9378), 1,
      aux_sym_number_token1,
    ACTIONS(9380), 1,
      aux_sym_number_token2,
    ACTIONS(9382), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9384), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9386), 1,
      anon_sym_BQUOTE,
    ACTIONS(9388), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9392), 1,
      sym_variable_name,
    ACTIONS(10186), 1,
      aux_sym__simple_variable_name_token1,
    STATE(3391), 1,
      sym__arithmetic_binary_expression,
    STATE(3396), 1,
      sym__arithmetic_ternary_expression,
    STATE(3398), 1,
      sym__arithmetic_unary_expression,
    STATE(3400), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9368), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9370), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(3334), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [213943] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1141), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1149), 1,
      anon_sym_DQUOTE,
    ACTIONS(1153), 1,
      aux_sym_number_token1,
    ACTIONS(1155), 1,
      aux_sym_number_token2,
    ACTIONS(1157), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1159), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1163), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1169), 1,
      sym__brace_start,
    ACTIONS(3060), 1,
      anon_sym_BQUOTE,
    ACTIONS(10188), 1,
      sym_word,
    ACTIONS(10190), 1,
      anon_sym_DOLLAR,
    ACTIONS(10196), 1,
      sym__comment_word,
    ACTIONS(1129), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1165), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10192), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10194), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2392), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [214011] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3064), 1,
      sym_variable_name,
    ACTIONS(3225), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(9278), 1,
      anon_sym_LPAREN,
    ACTIONS(9280), 1,
      anon_sym_BANG,
    ACTIONS(9286), 1,
      anon_sym_TILDE,
    ACTIONS(9288), 1,
      anon_sym_DOLLAR,
    ACTIONS(9290), 1,
      anon_sym_DQUOTE,
    ACTIONS(9292), 1,
      aux_sym_number_token1,
    ACTIONS(9294), 1,
      aux_sym_number_token2,
    ACTIONS(9296), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9298), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9300), 1,
      anon_sym_BQUOTE,
    ACTIONS(9302), 1,
      anon_sym_DOLLAR_BQUOTE,
    STATE(2801), 1,
      sym__arithmetic_binary_expression,
    STATE(2803), 1,
      sym__arithmetic_ternary_expression,
    STATE(2809), 1,
      sym__arithmetic_unary_expression,
    STATE(2825), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9282), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9284), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(2756), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [214085] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1141), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1149), 1,
      anon_sym_DQUOTE,
    ACTIONS(1153), 1,
      aux_sym_number_token1,
    ACTIONS(1155), 1,
      aux_sym_number_token2,
    ACTIONS(1157), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1159), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1163), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1169), 1,
      sym__brace_start,
    ACTIONS(3060), 1,
      anon_sym_BQUOTE,
    ACTIONS(10188), 1,
      sym_word,
    ACTIONS(10196), 1,
      sym__comment_word,
    ACTIONS(10198), 1,
      anon_sym_DOLLAR,
    ACTIONS(1129), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1165), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10192), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10194), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2392), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [214153] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3026), 1,
      anon_sym_DOLLAR,
    ACTIONS(3032), 1,
      aux_sym_number_token1,
    ACTIONS(3034), 1,
      aux_sym_number_token2,
    ACTIONS(3038), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3048), 1,
      sym__brace_start,
    ACTIONS(10134), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(10136), 1,
      sym__special_character,
    ACTIONS(10138), 1,
      anon_sym_DQUOTE,
    ACTIONS(10142), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(10144), 1,
      anon_sym_BQUOTE,
    ACTIONS(10146), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10200), 1,
      sym_word,
    ACTIONS(10204), 1,
      sym_test_operator,
    STATE(4527), 1,
      aux_sym__literal_repeat1,
    STATE(4885), 1,
      sym_concatenation,
    ACTIONS(10132), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(10148), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10202), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(4375), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [214225] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(6488), 1,
      anon_sym_DOLLAR,
    ACTIONS(6494), 1,
      aux_sym_number_token1,
    ACTIONS(6496), 1,
      aux_sym_number_token2,
    ACTIONS(6500), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(6512), 1,
      sym__brace_start,
    ACTIONS(7226), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(7230), 1,
      anon_sym_DQUOTE,
    ACTIONS(7234), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(7236), 1,
      anon_sym_BQUOTE,
    ACTIONS(7238), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10206), 1,
      sym_word,
    ACTIONS(10212), 1,
      sym__comment_word,
    ACTIONS(7224), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(7240), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10208), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10210), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(6494), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [214293] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3401), 1,
      aux_sym_number_token1,
    ACTIONS(3403), 1,
      aux_sym_number_token2,
    ACTIONS(3407), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3417), 1,
      sym__brace_start,
    ACTIONS(8926), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8930), 1,
      anon_sym_DQUOTE,
    ACTIONS(8934), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8936), 1,
      anon_sym_BQUOTE,
    ACTIONS(8938), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9696), 1,
      sym_word,
    ACTIONS(9704), 1,
      sym__comment_word,
    ACTIONS(10214), 1,
      anon_sym_DOLLAR,
    ACTIONS(8924), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8940), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9700), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9702), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1851), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [214361] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9364), 1,
      anon_sym_LPAREN,
    ACTIONS(9366), 1,
      anon_sym_BANG,
    ACTIONS(9372), 1,
      anon_sym_TILDE,
    ACTIONS(9374), 1,
      anon_sym_DOLLAR,
    ACTIONS(9376), 1,
      anon_sym_DQUOTE,
    ACTIONS(9378), 1,
      aux_sym_number_token1,
    ACTIONS(9380), 1,
      aux_sym_number_token2,
    ACTIONS(9382), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9384), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9386), 1,
      anon_sym_BQUOTE,
    ACTIONS(9388), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9392), 1,
      sym_variable_name,
    ACTIONS(10216), 1,
      aux_sym__simple_variable_name_token1,
    STATE(3391), 1,
      sym__arithmetic_binary_expression,
    STATE(3396), 1,
      sym__arithmetic_ternary_expression,
    STATE(3398), 1,
      sym__arithmetic_unary_expression,
    STATE(3400), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9368), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9370), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(3329), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [214435] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3064), 1,
      sym_variable_name,
    ACTIONS(9278), 1,
      anon_sym_LPAREN,
    ACTIONS(9280), 1,
      anon_sym_BANG,
    ACTIONS(9286), 1,
      anon_sym_TILDE,
    ACTIONS(9288), 1,
      anon_sym_DOLLAR,
    ACTIONS(9290), 1,
      anon_sym_DQUOTE,
    ACTIONS(9292), 1,
      aux_sym_number_token1,
    ACTIONS(9294), 1,
      aux_sym_number_token2,
    ACTIONS(9296), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9298), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9300), 1,
      anon_sym_BQUOTE,
    ACTIONS(9302), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10218), 1,
      aux_sym__simple_variable_name_token1,
    STATE(2801), 1,
      sym__arithmetic_binary_expression,
    STATE(2803), 1,
      sym__arithmetic_ternary_expression,
    STATE(2809), 1,
      sym__arithmetic_unary_expression,
    STATE(2825), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9282), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9284), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(2757), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [214509] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8512), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8514), 1,
      anon_sym_DOLLAR,
    ACTIONS(8518), 1,
      anon_sym_DQUOTE,
    ACTIONS(8522), 1,
      aux_sym_number_token1,
    ACTIONS(8524), 1,
      aux_sym_number_token2,
    ACTIONS(8526), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8528), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(8530), 1,
      anon_sym_BQUOTE,
    ACTIONS(8532), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8542), 1,
      sym__brace_start,
    ACTIONS(10220), 1,
      sym_word,
    ACTIONS(10226), 1,
      sym__comment_word,
    ACTIONS(8508), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8534), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10222), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10224), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2514), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [214577] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4947), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(4949), 1,
      anon_sym_DOLLAR,
    ACTIONS(4953), 1,
      anon_sym_DQUOTE,
    ACTIONS(4957), 1,
      aux_sym_number_token1,
    ACTIONS(4959), 1,
      aux_sym_number_token2,
    ACTIONS(4961), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(4963), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(4965), 1,
      anon_sym_BQUOTE,
    ACTIONS(4967), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(4973), 1,
      sym__brace_start,
    ACTIONS(10228), 1,
      sym_word,
    ACTIONS(10234), 1,
      sym__comment_word,
    ACTIONS(4945), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(4969), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10230), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10232), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2590), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [214645] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5315), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(5317), 1,
      anon_sym_DOLLAR,
    ACTIONS(5321), 1,
      anon_sym_DQUOTE,
    ACTIONS(5325), 1,
      aux_sym_number_token1,
    ACTIONS(5327), 1,
      aux_sym_number_token2,
    ACTIONS(5329), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(5331), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(5333), 1,
      anon_sym_BQUOTE,
    ACTIONS(5335), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(5341), 1,
      sym__brace_start,
    ACTIONS(10236), 1,
      sym_word,
    ACTIONS(10242), 1,
      sym__comment_word,
    ACTIONS(5313), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(5337), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10238), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10240), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(5461), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [214713] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1593), 1,
      aux_sym_number_token1,
    ACTIONS(1595), 1,
      aux_sym_number_token2,
    ACTIONS(1599), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1611), 1,
      sym__brace_start,
    ACTIONS(9240), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9244), 1,
      anon_sym_DQUOTE,
    ACTIONS(9248), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9250), 1,
      anon_sym_BQUOTE,
    ACTIONS(9252), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10094), 1,
      sym_word,
    ACTIONS(10102), 1,
      sym__comment_word,
    ACTIONS(10244), 1,
      anon_sym_DOLLAR,
    ACTIONS(9238), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9254), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10098), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10100), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1293), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [214781] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2610), 1,
      aux_sym_number_token1,
    ACTIONS(2612), 1,
      aux_sym_number_token2,
    ACTIONS(2616), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2626), 1,
      sym__brace_start,
    ACTIONS(8962), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8966), 1,
      anon_sym_DQUOTE,
    ACTIONS(8970), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8972), 1,
      anon_sym_BQUOTE,
    ACTIONS(8974), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10246), 1,
      sym_word,
    ACTIONS(10248), 1,
      anon_sym_DOLLAR,
    ACTIONS(10254), 1,
      sym__comment_word,
    ACTIONS(8960), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8976), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10250), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10252), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1532), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [214849] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4502), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(4504), 1,
      anon_sym_DOLLAR,
    ACTIONS(4508), 1,
      anon_sym_DQUOTE,
    ACTIONS(4512), 1,
      aux_sym_number_token1,
    ACTIONS(4514), 1,
      aux_sym_number_token2,
    ACTIONS(4516), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(4518), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(4520), 1,
      anon_sym_BQUOTE,
    ACTIONS(4522), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(4528), 1,
      sym__brace_start,
    ACTIONS(10256), 1,
      sym_word,
    ACTIONS(10262), 1,
      sym__comment_word,
    ACTIONS(4500), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(4524), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10258), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10260), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2450), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [214917] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4767), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(4769), 1,
      anon_sym_DOLLAR,
    ACTIONS(4773), 1,
      anon_sym_DQUOTE,
    ACTIONS(4777), 1,
      aux_sym_number_token1,
    ACTIONS(4779), 1,
      aux_sym_number_token2,
    ACTIONS(4781), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(4783), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(4785), 1,
      anon_sym_BQUOTE,
    ACTIONS(4787), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(4793), 1,
      sym__brace_start,
    ACTIONS(10264), 1,
      sym_word,
    ACTIONS(10270), 1,
      sym__comment_word,
    ACTIONS(4765), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(4789), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10266), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10268), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(5387), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [214985] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5112), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(5118), 1,
      anon_sym_DQUOTE,
    ACTIONS(5122), 1,
      aux_sym_number_token1,
    ACTIONS(5124), 1,
      aux_sym_number_token2,
    ACTIONS(5126), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(5128), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(5130), 1,
      anon_sym_BQUOTE,
    ACTIONS(5132), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(5138), 1,
      sym__brace_start,
    ACTIONS(10174), 1,
      sym_word,
    ACTIONS(10182), 1,
      sym__comment_word,
    ACTIONS(10272), 1,
      anon_sym_DOLLAR,
    ACTIONS(5110), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(5134), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10178), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10180), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2842), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [215053] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2610), 1,
      aux_sym_number_token1,
    ACTIONS(2612), 1,
      aux_sym_number_token2,
    ACTIONS(2616), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2626), 1,
      sym__brace_start,
    ACTIONS(8962), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8966), 1,
      anon_sym_DQUOTE,
    ACTIONS(8970), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8972), 1,
      anon_sym_BQUOTE,
    ACTIONS(8974), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10246), 1,
      sym_word,
    ACTIONS(10254), 1,
      sym__comment_word,
    ACTIONS(10274), 1,
      anon_sym_DOLLAR,
    ACTIONS(8960), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8976), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10250), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10252), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1532), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [215121] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(7958), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(7960), 1,
      anon_sym_DOLLAR,
    ACTIONS(7964), 1,
      anon_sym_DQUOTE,
    ACTIONS(7968), 1,
      aux_sym_number_token1,
    ACTIONS(7970), 1,
      aux_sym_number_token2,
    ACTIONS(7972), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(7974), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(7976), 1,
      anon_sym_BQUOTE,
    ACTIONS(7978), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(7988), 1,
      sym__brace_start,
    ACTIONS(9860), 1,
      sym_word,
    ACTIONS(9868), 1,
      sym__comment_word,
    ACTIONS(7954), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(7980), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9864), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9866), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(4357), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [215189] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9364), 1,
      anon_sym_LPAREN,
    ACTIONS(9366), 1,
      anon_sym_BANG,
    ACTIONS(9372), 1,
      anon_sym_TILDE,
    ACTIONS(9374), 1,
      anon_sym_DOLLAR,
    ACTIONS(9376), 1,
      anon_sym_DQUOTE,
    ACTIONS(9378), 1,
      aux_sym_number_token1,
    ACTIONS(9380), 1,
      aux_sym_number_token2,
    ACTIONS(9382), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9384), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9386), 1,
      anon_sym_BQUOTE,
    ACTIONS(9388), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9392), 1,
      sym_variable_name,
    ACTIONS(10276), 1,
      aux_sym__simple_variable_name_token1,
    STATE(3391), 1,
      sym__arithmetic_binary_expression,
    STATE(3396), 1,
      sym__arithmetic_ternary_expression,
    STATE(3398), 1,
      sym__arithmetic_unary_expression,
    STATE(3400), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9368), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9370), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(3328), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [215263] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3064), 1,
      sym_variable_name,
    ACTIONS(9278), 1,
      anon_sym_LPAREN,
    ACTIONS(9280), 1,
      anon_sym_BANG,
    ACTIONS(9286), 1,
      anon_sym_TILDE,
    ACTIONS(9288), 1,
      anon_sym_DOLLAR,
    ACTIONS(9290), 1,
      anon_sym_DQUOTE,
    ACTIONS(9292), 1,
      aux_sym_number_token1,
    ACTIONS(9294), 1,
      aux_sym_number_token2,
    ACTIONS(9296), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9298), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9300), 1,
      anon_sym_BQUOTE,
    ACTIONS(9302), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10278), 1,
      aux_sym__simple_variable_name_token1,
    STATE(2801), 1,
      sym__arithmetic_binary_expression,
    STATE(2803), 1,
      sym__arithmetic_ternary_expression,
    STATE(2809), 1,
      sym__arithmetic_unary_expression,
    STATE(2825), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9282), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9284), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(2758), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [215337] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9364), 1,
      anon_sym_LPAREN,
    ACTIONS(9366), 1,
      anon_sym_BANG,
    ACTIONS(9372), 1,
      anon_sym_TILDE,
    ACTIONS(9374), 1,
      anon_sym_DOLLAR,
    ACTIONS(9376), 1,
      anon_sym_DQUOTE,
    ACTIONS(9378), 1,
      aux_sym_number_token1,
    ACTIONS(9380), 1,
      aux_sym_number_token2,
    ACTIONS(9382), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9384), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9386), 1,
      anon_sym_BQUOTE,
    ACTIONS(9388), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9392), 1,
      sym_variable_name,
    ACTIONS(10280), 1,
      aux_sym__simple_variable_name_token1,
    STATE(3391), 1,
      sym__arithmetic_binary_expression,
    STATE(3396), 1,
      sym__arithmetic_ternary_expression,
    STATE(3398), 1,
      sym__arithmetic_unary_expression,
    STATE(3400), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9368), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9370), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(3326), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [215411] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8306), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8308), 1,
      anon_sym_DOLLAR,
    ACTIONS(8312), 1,
      anon_sym_DQUOTE,
    ACTIONS(8316), 1,
      aux_sym_number_token1,
    ACTIONS(8318), 1,
      aux_sym_number_token2,
    ACTIONS(8320), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8322), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(8324), 1,
      anon_sym_BQUOTE,
    ACTIONS(8326), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8336), 1,
      sym__brace_start,
    ACTIONS(10282), 1,
      sym_word,
    ACTIONS(10288), 1,
      sym__comment_word,
    ACTIONS(8302), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8328), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10284), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10286), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(5428), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [215479] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5610), 1,
      anon_sym_DOLLAR,
    ACTIONS(5616), 1,
      aux_sym_number_token1,
    ACTIONS(5618), 1,
      aux_sym_number_token2,
    ACTIONS(5622), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(5632), 1,
      sym__brace_start,
    ACTIONS(8944), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8948), 1,
      anon_sym_DQUOTE,
    ACTIONS(8952), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8954), 1,
      anon_sym_BQUOTE,
    ACTIONS(8956), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10290), 1,
      sym_word,
    ACTIONS(10296), 1,
      sym__comment_word,
    ACTIONS(8942), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8958), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10292), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10294), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(3415), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [215547] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3064), 1,
      sym_variable_name,
    ACTIONS(9278), 1,
      anon_sym_LPAREN,
    ACTIONS(9280), 1,
      anon_sym_BANG,
    ACTIONS(9286), 1,
      anon_sym_TILDE,
    ACTIONS(9288), 1,
      anon_sym_DOLLAR,
    ACTIONS(9290), 1,
      anon_sym_DQUOTE,
    ACTIONS(9292), 1,
      aux_sym_number_token1,
    ACTIONS(9294), 1,
      aux_sym_number_token2,
    ACTIONS(9296), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9298), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9300), 1,
      anon_sym_BQUOTE,
    ACTIONS(9302), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10298), 1,
      aux_sym__simple_variable_name_token1,
    STATE(2801), 1,
      sym__arithmetic_binary_expression,
    STATE(2803), 1,
      sym__arithmetic_ternary_expression,
    STATE(2809), 1,
      sym__arithmetic_unary_expression,
    STATE(2825), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9282), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9284), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(2759), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [215621] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5870), 1,
      anon_sym_DOLLAR,
    ACTIONS(5876), 1,
      aux_sym_number_token1,
    ACTIONS(5878), 1,
      aux_sym_number_token2,
    ACTIONS(5882), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(5892), 1,
      sym__brace_start,
    ACTIONS(9546), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9550), 1,
      anon_sym_DQUOTE,
    ACTIONS(9554), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9556), 1,
      anon_sym_BQUOTE,
    ACTIONS(9558), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10300), 1,
      sym_word,
    ACTIONS(10306), 1,
      sym__comment_word,
    ACTIONS(9544), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9560), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10302), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10304), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(5680), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [215689] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(451), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(453), 1,
      anon_sym_DOLLAR,
    ACTIONS(457), 1,
      anon_sym_DQUOTE,
    ACTIONS(461), 1,
      aux_sym_number_token1,
    ACTIONS(463), 1,
      aux_sym_number_token2,
    ACTIONS(465), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(467), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(469), 1,
      anon_sym_BQUOTE,
    ACTIONS(471), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(481), 1,
      sym__brace_start,
    ACTIONS(10308), 1,
      sym_word,
    ACTIONS(10314), 1,
      sym__comment_word,
    ACTIONS(449), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(473), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10310), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10312), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(866), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [215757] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5112), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(5114), 1,
      anon_sym_DOLLAR,
    ACTIONS(5118), 1,
      anon_sym_DQUOTE,
    ACTIONS(5122), 1,
      aux_sym_number_token1,
    ACTIONS(5124), 1,
      aux_sym_number_token2,
    ACTIONS(5126), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(5128), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(5130), 1,
      anon_sym_BQUOTE,
    ACTIONS(5132), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(5138), 1,
      sym__brace_start,
    ACTIONS(10174), 1,
      sym_word,
    ACTIONS(10182), 1,
      sym__comment_word,
    ACTIONS(5110), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(5134), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10178), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10180), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2842), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [215825] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(6494), 1,
      aux_sym_number_token1,
    ACTIONS(6496), 1,
      aux_sym_number_token2,
    ACTIONS(6500), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(6512), 1,
      sym__brace_start,
    ACTIONS(7226), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(7230), 1,
      anon_sym_DQUOTE,
    ACTIONS(7234), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(7236), 1,
      anon_sym_BQUOTE,
    ACTIONS(7238), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10206), 1,
      sym_word,
    ACTIONS(10212), 1,
      sym__comment_word,
    ACTIONS(10316), 1,
      anon_sym_DOLLAR,
    ACTIONS(7224), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(7240), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10208), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10210), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(6494), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [215893] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5405), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(5407), 1,
      anon_sym_DOLLAR,
    ACTIONS(5411), 1,
      anon_sym_DQUOTE,
    ACTIONS(5415), 1,
      aux_sym_number_token1,
    ACTIONS(5417), 1,
      aux_sym_number_token2,
    ACTIONS(5419), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(5421), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(5423), 1,
      anon_sym_BQUOTE,
    ACTIONS(5425), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(5431), 1,
      sym__brace_start,
    ACTIONS(9958), 1,
      sym_word,
    ACTIONS(9966), 1,
      sym__comment_word,
    ACTIONS(5403), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(5427), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9962), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9964), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(5541), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [215961] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4633), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(4635), 1,
      anon_sym_DOLLAR,
    ACTIONS(4639), 1,
      anon_sym_DQUOTE,
    ACTIONS(4643), 1,
      aux_sym_number_token1,
    ACTIONS(4645), 1,
      aux_sym_number_token2,
    ACTIONS(4647), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(4649), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(4651), 1,
      anon_sym_BQUOTE,
    ACTIONS(4653), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(4659), 1,
      sym__brace_start,
    ACTIONS(9330), 1,
      sym_word,
    ACTIONS(9338), 1,
      sym__comment_word,
    ACTIONS(4631), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(4655), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9334), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9336), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2634), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [216029] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3886), 1,
      aux_sym_number_token1,
    ACTIONS(3888), 1,
      aux_sym_number_token2,
    ACTIONS(3892), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3902), 1,
      sym__brace_start,
    ACTIONS(8992), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8996), 1,
      anon_sym_DQUOTE,
    ACTIONS(9000), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9002), 1,
      anon_sym_BQUOTE,
    ACTIONS(9004), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9740), 1,
      sym_word,
    ACTIONS(9746), 1,
      sym__comment_word,
    ACTIONS(10318), 1,
      anon_sym_DOLLAR,
    ACTIONS(8990), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9006), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9742), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9744), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2004), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [216097] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9118), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9124), 1,
      anon_sym_DQUOTE,
    ACTIONS(9128), 1,
      aux_sym_number_token1,
    ACTIONS(9130), 1,
      aux_sym_number_token2,
    ACTIONS(9132), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9134), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9136), 1,
      anon_sym_BQUOTE,
    ACTIONS(9138), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9144), 1,
      sym__brace_start,
    ACTIONS(9442), 1,
      sym_word,
    ACTIONS(9448), 1,
      sym__comment_word,
    ACTIONS(10320), 1,
      anon_sym_DOLLAR,
    ACTIONS(9116), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9140), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9444), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9446), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2401), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [216165] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3064), 1,
      sym_variable_name,
    ACTIONS(9278), 1,
      anon_sym_LPAREN,
    ACTIONS(9280), 1,
      anon_sym_BANG,
    ACTIONS(9286), 1,
      anon_sym_TILDE,
    ACTIONS(9288), 1,
      anon_sym_DOLLAR,
    ACTIONS(9290), 1,
      anon_sym_DQUOTE,
    ACTIONS(9292), 1,
      aux_sym_number_token1,
    ACTIONS(9294), 1,
      aux_sym_number_token2,
    ACTIONS(9296), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9298), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9300), 1,
      anon_sym_BQUOTE,
    ACTIONS(9302), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10322), 1,
      aux_sym__simple_variable_name_token1,
    STATE(2801), 1,
      sym__arithmetic_binary_expression,
    STATE(2803), 1,
      sym__arithmetic_ternary_expression,
    STATE(2809), 1,
      sym__arithmetic_unary_expression,
    STATE(2825), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9282), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9284), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(2797), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [216239] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(6494), 1,
      aux_sym_number_token1,
    ACTIONS(6496), 1,
      aux_sym_number_token2,
    ACTIONS(6500), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(6512), 1,
      sym__brace_start,
    ACTIONS(7226), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(7230), 1,
      anon_sym_DQUOTE,
    ACTIONS(7234), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(7236), 1,
      anon_sym_BQUOTE,
    ACTIONS(7238), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10206), 1,
      sym_word,
    ACTIONS(10212), 1,
      sym__comment_word,
    ACTIONS(10324), 1,
      anon_sym_DOLLAR,
    ACTIONS(7224), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(7240), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10208), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10210), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(6494), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [216307] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5261), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(5263), 1,
      anon_sym_DOLLAR,
    ACTIONS(5267), 1,
      anon_sym_DQUOTE,
    ACTIONS(5271), 1,
      aux_sym_number_token1,
    ACTIONS(5273), 1,
      aux_sym_number_token2,
    ACTIONS(5275), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(5277), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(5279), 1,
      anon_sym_BQUOTE,
    ACTIONS(5281), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(5287), 1,
      sym__brace_start,
    ACTIONS(9472), 1,
      sym_word,
    ACTIONS(9480), 1,
      sym__comment_word,
    ACTIONS(5259), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(5283), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9476), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9478), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(5440), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [216375] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8102), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8104), 1,
      anon_sym_DOLLAR,
    ACTIONS(8108), 1,
      anon_sym_DQUOTE,
    ACTIONS(8112), 1,
      aux_sym_number_token1,
    ACTIONS(8114), 1,
      aux_sym_number_token2,
    ACTIONS(8116), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8118), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(8120), 1,
      anon_sym_BQUOTE,
    ACTIONS(8122), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8132), 1,
      sym__brace_start,
    ACTIONS(9600), 1,
      sym_word,
    ACTIONS(9608), 1,
      sym__comment_word,
    ACTIONS(8098), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8124), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9604), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9606), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1199), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [216443] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9364), 1,
      anon_sym_LPAREN,
    ACTIONS(9366), 1,
      anon_sym_BANG,
    ACTIONS(9372), 1,
      anon_sym_TILDE,
    ACTIONS(9374), 1,
      anon_sym_DOLLAR,
    ACTIONS(9376), 1,
      anon_sym_DQUOTE,
    ACTIONS(9378), 1,
      aux_sym_number_token1,
    ACTIONS(9380), 1,
      aux_sym_number_token2,
    ACTIONS(9382), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9384), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9386), 1,
      anon_sym_BQUOTE,
    ACTIONS(9388), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9392), 1,
      sym_variable_name,
    ACTIONS(10326), 1,
      aux_sym__simple_variable_name_token1,
    STATE(3391), 1,
      sym__arithmetic_binary_expression,
    STATE(3396), 1,
      sym__arithmetic_ternary_expression,
    STATE(3398), 1,
      sym__arithmetic_unary_expression,
    STATE(3400), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9368), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9370), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(3324), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [216517] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9364), 1,
      anon_sym_LPAREN,
    ACTIONS(9366), 1,
      anon_sym_BANG,
    ACTIONS(9372), 1,
      anon_sym_TILDE,
    ACTIONS(9374), 1,
      anon_sym_DOLLAR,
    ACTIONS(9376), 1,
      anon_sym_DQUOTE,
    ACTIONS(9378), 1,
      aux_sym_number_token1,
    ACTIONS(9380), 1,
      aux_sym_number_token2,
    ACTIONS(9382), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9384), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9386), 1,
      anon_sym_BQUOTE,
    ACTIONS(9388), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9392), 1,
      sym_variable_name,
    ACTIONS(10328), 1,
      aux_sym__simple_variable_name_token1,
    STATE(3391), 1,
      sym__arithmetic_binary_expression,
    STATE(3396), 1,
      sym__arithmetic_ternary_expression,
    STATE(3398), 1,
      sym__arithmetic_unary_expression,
    STATE(3400), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9368), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9370), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(3372), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [216591] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1761), 1,
      anon_sym_DOLLAR,
    ACTIONS(1767), 1,
      aux_sym_number_token1,
    ACTIONS(1769), 1,
      aux_sym_number_token2,
    ACTIONS(1773), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1787), 1,
      sym__brace_start,
    ACTIONS(8426), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8430), 1,
      anon_sym_DQUOTE,
    ACTIONS(8434), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8436), 1,
      anon_sym_BQUOTE,
    ACTIONS(8438), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9820), 1,
      sym_word,
    ACTIONS(9828), 1,
      sym__comment_word,
    ACTIONS(8422), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8440), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9824), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9826), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1207), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [216659] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2303), 1,
      anon_sym_DOLLAR,
    ACTIONS(2309), 1,
      aux_sym_number_token1,
    ACTIONS(2311), 1,
      aux_sym_number_token2,
    ACTIONS(2315), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2327), 1,
      sym__brace_start,
    ACTIONS(9982), 1,
      sym_word,
    ACTIONS(9986), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9992), 1,
      anon_sym_DQUOTE,
    ACTIONS(9996), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9998), 1,
      anon_sym_BQUOTE,
    ACTIONS(10000), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10004), 1,
      sym__comment_word,
    ACTIONS(9984), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9990), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10002), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9994), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1423), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [216727] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1414), 1,
      anon_sym_DOLLAR,
    ACTIONS(1420), 1,
      aux_sym_number_token1,
    ACTIONS(1422), 1,
      aux_sym_number_token2,
    ACTIONS(1426), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1438), 1,
      sym__brace_start,
    ACTIONS(9094), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9098), 1,
      anon_sym_DQUOTE,
    ACTIONS(9102), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9104), 1,
      anon_sym_BQUOTE,
    ACTIONS(9106), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10120), 1,
      sym_word,
    ACTIONS(10128), 1,
      sym__comment_word,
    ACTIONS(9092), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9108), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10124), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10126), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(910), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [216795] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3373), 1,
      aux_sym_number_token1,
    ACTIONS(3375), 1,
      aux_sym_number_token2,
    ACTIONS(3379), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3389), 1,
      sym__brace_start,
    ACTIONS(9802), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9806), 1,
      anon_sym_DQUOTE,
    ACTIONS(9810), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9812), 1,
      anon_sym_BQUOTE,
    ACTIONS(9814), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10330), 1,
      sym_word,
    ACTIONS(10332), 1,
      anon_sym_DOLLAR,
    ACTIONS(10338), 1,
      sym__comment_word,
    ACTIONS(9800), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9816), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10334), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10336), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(4532), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [216863] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9118), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9120), 1,
      anon_sym_DOLLAR,
    ACTIONS(9122), 1,
      sym__special_character,
    ACTIONS(9124), 1,
      anon_sym_DQUOTE,
    ACTIONS(9128), 1,
      aux_sym_number_token1,
    ACTIONS(9130), 1,
      aux_sym_number_token2,
    ACTIONS(9132), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9134), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9136), 1,
      anon_sym_BQUOTE,
    ACTIONS(9138), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9144), 1,
      sym__brace_start,
    ACTIONS(10340), 1,
      sym_word,
    ACTIONS(10344), 1,
      sym_test_operator,
    STATE(2569), 1,
      aux_sym__literal_repeat1,
    STATE(2623), 1,
      sym_concatenation,
    ACTIONS(9116), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9140), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10342), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2278), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [216935] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2604), 1,
      anon_sym_DOLLAR,
    ACTIONS(2610), 1,
      aux_sym_number_token1,
    ACTIONS(2612), 1,
      aux_sym_number_token2,
    ACTIONS(2616), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2626), 1,
      sym__brace_start,
    ACTIONS(8962), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8966), 1,
      anon_sym_DQUOTE,
    ACTIONS(8970), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8972), 1,
      anon_sym_BQUOTE,
    ACTIONS(8974), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10246), 1,
      sym_word,
    ACTIONS(10254), 1,
      sym__comment_word,
    ACTIONS(8960), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8976), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10250), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10252), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1532), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [217003] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3367), 1,
      anon_sym_DOLLAR,
    ACTIONS(3373), 1,
      aux_sym_number_token1,
    ACTIONS(3375), 1,
      aux_sym_number_token2,
    ACTIONS(3379), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3389), 1,
      sym__brace_start,
    ACTIONS(9802), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9806), 1,
      anon_sym_DQUOTE,
    ACTIONS(9810), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9812), 1,
      anon_sym_BQUOTE,
    ACTIONS(9814), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10330), 1,
      sym_word,
    ACTIONS(10338), 1,
      sym__comment_word,
    ACTIONS(9800), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9816), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10334), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10336), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(4532), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [217071] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3373), 1,
      aux_sym_number_token1,
    ACTIONS(3375), 1,
      aux_sym_number_token2,
    ACTIONS(3379), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3389), 1,
      sym__brace_start,
    ACTIONS(9802), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9806), 1,
      anon_sym_DQUOTE,
    ACTIONS(9810), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9812), 1,
      anon_sym_BQUOTE,
    ACTIONS(9814), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10330), 1,
      sym_word,
    ACTIONS(10338), 1,
      sym__comment_word,
    ACTIONS(10346), 1,
      anon_sym_DOLLAR,
    ACTIONS(9800), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9816), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10334), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10336), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(4532), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [217139] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9118), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9120), 1,
      anon_sym_DOLLAR,
    ACTIONS(9122), 1,
      sym__special_character,
    ACTIONS(9124), 1,
      anon_sym_DQUOTE,
    ACTIONS(9128), 1,
      aux_sym_number_token1,
    ACTIONS(9130), 1,
      aux_sym_number_token2,
    ACTIONS(9132), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9134), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9136), 1,
      anon_sym_BQUOTE,
    ACTIONS(9138), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9144), 1,
      sym__brace_start,
    ACTIONS(10348), 1,
      sym_word,
    ACTIONS(10352), 1,
      sym_test_operator,
    STATE(2532), 1,
      aux_sym__literal_repeat1,
    STATE(2635), 1,
      sym_concatenation,
    ACTIONS(9116), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9140), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10350), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2314), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [217211] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(7958), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(7964), 1,
      anon_sym_DQUOTE,
    ACTIONS(7968), 1,
      aux_sym_number_token1,
    ACTIONS(7970), 1,
      aux_sym_number_token2,
    ACTIONS(7972), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(7974), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(7976), 1,
      anon_sym_BQUOTE,
    ACTIONS(7978), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(7988), 1,
      sym__brace_start,
    ACTIONS(9860), 1,
      sym_word,
    ACTIONS(9868), 1,
      sym__comment_word,
    ACTIONS(10354), 1,
      anon_sym_DOLLAR,
    ACTIONS(7954), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(7980), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9864), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9866), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(4357), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [217279] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(7958), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(7964), 1,
      anon_sym_DQUOTE,
    ACTIONS(7968), 1,
      aux_sym_number_token1,
    ACTIONS(7970), 1,
      aux_sym_number_token2,
    ACTIONS(7972), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(7974), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(7976), 1,
      anon_sym_BQUOTE,
    ACTIONS(7978), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(7988), 1,
      sym__brace_start,
    ACTIONS(9860), 1,
      sym_word,
    ACTIONS(9868), 1,
      sym__comment_word,
    ACTIONS(10356), 1,
      anon_sym_DOLLAR,
    ACTIONS(7954), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(7980), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9864), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9866), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(4357), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [217347] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2335), 1,
      anon_sym_DOLLAR,
    ACTIONS(2341), 1,
      aux_sym_number_token1,
    ACTIONS(2343), 1,
      aux_sym_number_token2,
    ACTIONS(2347), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2357), 1,
      sym__brace_start,
    ACTIONS(9010), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9014), 1,
      anon_sym_DQUOTE,
    ACTIONS(9018), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9020), 1,
      anon_sym_BQUOTE,
    ACTIONS(9022), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10358), 1,
      sym_word,
    ACTIONS(10364), 1,
      sym__comment_word,
    ACTIONS(9008), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9024), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10360), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10362), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1462), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [217415] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3481), 1,
      aux_sym_number_token1,
    ACTIONS(3483), 1,
      aux_sym_number_token2,
    ACTIONS(3487), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3497), 1,
      sym__brace_start,
    ACTIONS(8906), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8910), 1,
      anon_sym_DQUOTE,
    ACTIONS(8914), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8916), 1,
      anon_sym_BQUOTE,
    ACTIONS(8918), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9850), 1,
      sym_word,
    ACTIONS(9856), 1,
      sym__comment_word,
    ACTIONS(10366), 1,
      anon_sym_DOLLAR,
    ACTIONS(8904), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8920), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9852), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9854), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1921), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [217483] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3675), 1,
      anon_sym_DOLLAR,
    ACTIONS(3681), 1,
      aux_sym_number_token1,
    ACTIONS(3683), 1,
      aux_sym_number_token2,
    ACTIONS(3687), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3697), 1,
      sym__brace_start,
    ACTIONS(10062), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(10064), 1,
      sym__special_character,
    ACTIONS(10066), 1,
      anon_sym_DQUOTE,
    ACTIONS(10070), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(10072), 1,
      anon_sym_BQUOTE,
    ACTIONS(10074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10368), 1,
      sym_word,
    ACTIONS(10372), 1,
      sym_test_operator,
    STATE(4704), 1,
      aux_sym__literal_repeat1,
    STATE(5106), 1,
      sym_concatenation,
    ACTIONS(10060), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(10076), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10370), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(4579), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [217555] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3793), 1,
      aux_sym_number_token1,
    ACTIONS(3795), 1,
      aux_sym_number_token2,
    ACTIONS(3799), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3809), 1,
      sym__brace_start,
    ACTIONS(9636), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9640), 1,
      anon_sym_DQUOTE,
    ACTIONS(9644), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9646), 1,
      anon_sym_BQUOTE,
    ACTIONS(9648), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9872), 1,
      sym_word,
    ACTIONS(9880), 1,
      sym__comment_word,
    ACTIONS(10374), 1,
      anon_sym_DOLLAR,
    ACTIONS(9634), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9650), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9876), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9878), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(4337), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [217623] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3793), 1,
      aux_sym_number_token1,
    ACTIONS(3795), 1,
      aux_sym_number_token2,
    ACTIONS(3799), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3809), 1,
      sym__brace_start,
    ACTIONS(9636), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9640), 1,
      anon_sym_DQUOTE,
    ACTIONS(9644), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9646), 1,
      anon_sym_BQUOTE,
    ACTIONS(9648), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9872), 1,
      sym_word,
    ACTIONS(9880), 1,
      sym__comment_word,
    ACTIONS(10376), 1,
      anon_sym_DOLLAR,
    ACTIONS(9634), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9650), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9876), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9878), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(4337), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [217691] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4086), 1,
      aux_sym_number_token1,
    ACTIONS(4088), 1,
      aux_sym_number_token2,
    ACTIONS(4092), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(4102), 1,
      sym__brace_start,
    ACTIONS(9306), 1,
      sym_word,
    ACTIONS(9310), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9316), 1,
      anon_sym_DQUOTE,
    ACTIONS(9320), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9322), 1,
      anon_sym_BQUOTE,
    ACTIONS(9324), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9328), 1,
      sym__comment_word,
    ACTIONS(10378), 1,
      anon_sym_DOLLAR,
    ACTIONS(9308), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9314), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9326), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9318), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(4566), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [217759] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4086), 1,
      aux_sym_number_token1,
    ACTIONS(4088), 1,
      aux_sym_number_token2,
    ACTIONS(4092), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(4102), 1,
      sym__brace_start,
    ACTIONS(9306), 1,
      sym_word,
    ACTIONS(9310), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9316), 1,
      anon_sym_DQUOTE,
    ACTIONS(9320), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9322), 1,
      anon_sym_BQUOTE,
    ACTIONS(9324), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9328), 1,
      sym__comment_word,
    ACTIONS(10380), 1,
      anon_sym_DOLLAR,
    ACTIONS(9308), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9314), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9326), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9318), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(4566), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [217827] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2863), 1,
      anon_sym_DOLLAR,
    ACTIONS(2869), 1,
      aux_sym_number_token1,
    ACTIONS(2871), 1,
      aux_sym_number_token2,
    ACTIONS(2875), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2885), 1,
      sym__brace_start,
    ACTIONS(9940), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9944), 1,
      anon_sym_DQUOTE,
    ACTIONS(9948), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9950), 1,
      anon_sym_BQUOTE,
    ACTIONS(9952), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10382), 1,
      sym_word,
    ACTIONS(10388), 1,
      sym__comment_word,
    ACTIONS(9938), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9954), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10384), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10386), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(4404), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [217895] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1557), 1,
      aux_sym_number_token1,
    ACTIONS(1559), 1,
      aux_sym_number_token2,
    ACTIONS(1563), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1575), 1,
      sym__brace_start,
    ACTIONS(9198), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9202), 1,
      anon_sym_DQUOTE,
    ACTIONS(9206), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9208), 1,
      anon_sym_BQUOTE,
    ACTIONS(9210), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9268), 1,
      sym_word,
    ACTIONS(9276), 1,
      sym__comment_word,
    ACTIONS(10390), 1,
      anon_sym_DOLLAR,
    ACTIONS(9196), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9212), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9272), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9274), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1189), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [217963] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3481), 1,
      aux_sym_number_token1,
    ACTIONS(3483), 1,
      aux_sym_number_token2,
    ACTIONS(3487), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3497), 1,
      sym__brace_start,
    ACTIONS(8906), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8910), 1,
      anon_sym_DQUOTE,
    ACTIONS(8914), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8916), 1,
      anon_sym_BQUOTE,
    ACTIONS(8918), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9850), 1,
      sym_word,
    ACTIONS(9856), 1,
      sym__comment_word,
    ACTIONS(10392), 1,
      anon_sym_DOLLAR,
    ACTIONS(8904), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8920), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9852), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9854), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1921), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [218031] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3064), 1,
      sym_variable_name,
    ACTIONS(9278), 1,
      anon_sym_LPAREN,
    ACTIONS(9280), 1,
      anon_sym_BANG,
    ACTIONS(9286), 1,
      anon_sym_TILDE,
    ACTIONS(9288), 1,
      anon_sym_DOLLAR,
    ACTIONS(9290), 1,
      anon_sym_DQUOTE,
    ACTIONS(9292), 1,
      aux_sym_number_token1,
    ACTIONS(9294), 1,
      aux_sym_number_token2,
    ACTIONS(9296), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9298), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9300), 1,
      anon_sym_BQUOTE,
    ACTIONS(9302), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10394), 1,
      aux_sym__simple_variable_name_token1,
    STATE(2801), 1,
      sym__arithmetic_binary_expression,
    STATE(2803), 1,
      sym__arithmetic_ternary_expression,
    STATE(2809), 1,
      sym__arithmetic_unary_expression,
    STATE(2825), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9282), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9284), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(2760), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [218105] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9364), 1,
      anon_sym_LPAREN,
    ACTIONS(9366), 1,
      anon_sym_BANG,
    ACTIONS(9372), 1,
      anon_sym_TILDE,
    ACTIONS(9374), 1,
      anon_sym_DOLLAR,
    ACTIONS(9376), 1,
      anon_sym_DQUOTE,
    ACTIONS(9378), 1,
      aux_sym_number_token1,
    ACTIONS(9380), 1,
      aux_sym_number_token2,
    ACTIONS(9382), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9384), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9386), 1,
      anon_sym_BQUOTE,
    ACTIONS(9388), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9392), 1,
      sym_variable_name,
    ACTIONS(10396), 1,
      aux_sym__simple_variable_name_token1,
    STATE(3391), 1,
      sym__arithmetic_binary_expression,
    STATE(3396), 1,
      sym__arithmetic_ternary_expression,
    STATE(3398), 1,
      sym__arithmetic_unary_expression,
    STATE(3400), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9368), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9370), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(3321), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [218179] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1557), 1,
      aux_sym_number_token1,
    ACTIONS(1559), 1,
      aux_sym_number_token2,
    ACTIONS(1563), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1575), 1,
      sym__brace_start,
    ACTIONS(9198), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9202), 1,
      anon_sym_DQUOTE,
    ACTIONS(9206), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9208), 1,
      anon_sym_BQUOTE,
    ACTIONS(9210), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9268), 1,
      sym_word,
    ACTIONS(9276), 1,
      sym__comment_word,
    ACTIONS(10398), 1,
      anon_sym_DOLLAR,
    ACTIONS(9196), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9212), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9272), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9274), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1189), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [218247] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3191), 1,
      aux_sym_number_token1,
    ACTIONS(3193), 1,
      aux_sym_number_token2,
    ACTIONS(3197), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3209), 1,
      sym__brace_start,
    ACTIONS(9674), 1,
      sym_word,
    ACTIONS(9678), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9682), 1,
      anon_sym_DQUOTE,
    ACTIONS(9686), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9688), 1,
      anon_sym_BQUOTE,
    ACTIONS(9690), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9694), 1,
      sym__comment_word,
    ACTIONS(10400), 1,
      anon_sym_DOLLAR,
    ACTIONS(9676), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9680), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9692), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9684), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1733), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [218315] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3064), 1,
      sym_variable_name,
    ACTIONS(9278), 1,
      anon_sym_LPAREN,
    ACTIONS(9280), 1,
      anon_sym_BANG,
    ACTIONS(9286), 1,
      anon_sym_TILDE,
    ACTIONS(9288), 1,
      anon_sym_DOLLAR,
    ACTIONS(9290), 1,
      anon_sym_DQUOTE,
    ACTIONS(9292), 1,
      aux_sym_number_token1,
    ACTIONS(9294), 1,
      aux_sym_number_token2,
    ACTIONS(9296), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9298), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9300), 1,
      anon_sym_BQUOTE,
    ACTIONS(9302), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10402), 1,
      aux_sym__simple_variable_name_token1,
    STATE(2801), 1,
      sym__arithmetic_binary_expression,
    STATE(2803), 1,
      sym__arithmetic_ternary_expression,
    STATE(2809), 1,
      sym__arithmetic_unary_expression,
    STATE(2825), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9282), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9284), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(2844), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [218389] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9364), 1,
      anon_sym_LPAREN,
    ACTIONS(9366), 1,
      anon_sym_BANG,
    ACTIONS(9372), 1,
      anon_sym_TILDE,
    ACTIONS(9374), 1,
      anon_sym_DOLLAR,
    ACTIONS(9376), 1,
      anon_sym_DQUOTE,
    ACTIONS(9378), 1,
      aux_sym_number_token1,
    ACTIONS(9380), 1,
      aux_sym_number_token2,
    ACTIONS(9382), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9384), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9386), 1,
      anon_sym_BQUOTE,
    ACTIONS(9388), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9392), 1,
      sym_variable_name,
    ACTIONS(10404), 1,
      aux_sym__simple_variable_name_token1,
    STATE(3391), 1,
      sym__arithmetic_binary_expression,
    STATE(3396), 1,
      sym__arithmetic_ternary_expression,
    STATE(3398), 1,
      sym__arithmetic_unary_expression,
    STATE(3400), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9368), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9370), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(3374), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [218463] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(191), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(193), 1,
      anon_sym_DOLLAR,
    ACTIONS(197), 1,
      anon_sym_DQUOTE,
    ACTIONS(201), 1,
      aux_sym_number_token1,
    ACTIONS(203), 1,
      aux_sym_number_token2,
    ACTIONS(205), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(207), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(209), 1,
      anon_sym_BQUOTE,
    ACTIONS(211), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(221), 1,
      sym__brace_start,
    ACTIONS(10406), 1,
      sym_word,
    ACTIONS(10412), 1,
      sym__comment_word,
    ACTIONS(189), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(213), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10408), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10410), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(513), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [218531] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3191), 1,
      aux_sym_number_token1,
    ACTIONS(3193), 1,
      aux_sym_number_token2,
    ACTIONS(3197), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3209), 1,
      sym__brace_start,
    ACTIONS(9674), 1,
      sym_word,
    ACTIONS(9678), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9682), 1,
      anon_sym_DQUOTE,
    ACTIONS(9686), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9688), 1,
      anon_sym_BQUOTE,
    ACTIONS(9690), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9694), 1,
      sym__comment_word,
    ACTIONS(10414), 1,
      anon_sym_DOLLAR,
    ACTIONS(9676), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9680), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9692), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9684), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1733), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [218599] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2269), 1,
      aux_sym_number_token1,
    ACTIONS(2271), 1,
      aux_sym_number_token2,
    ACTIONS(2275), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2289), 1,
      sym__brace_start,
    ACTIONS(8066), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8070), 1,
      anon_sym_DQUOTE,
    ACTIONS(8074), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8076), 1,
      anon_sym_BQUOTE,
    ACTIONS(8078), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9580), 1,
      sym_word,
    ACTIONS(9588), 1,
      sym__comment_word,
    ACTIONS(10416), 1,
      anon_sym_DOLLAR,
    ACTIONS(8062), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8080), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9584), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9586), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1623), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [218667] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(451), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(457), 1,
      anon_sym_DQUOTE,
    ACTIONS(461), 1,
      aux_sym_number_token1,
    ACTIONS(463), 1,
      aux_sym_number_token2,
    ACTIONS(465), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(467), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(469), 1,
      anon_sym_BQUOTE,
    ACTIONS(471), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(481), 1,
      sym__brace_start,
    ACTIONS(10308), 1,
      sym_word,
    ACTIONS(10314), 1,
      sym__comment_word,
    ACTIONS(10418), 1,
      anon_sym_DOLLAR,
    ACTIONS(449), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(473), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10310), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10312), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(866), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [218735] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2269), 1,
      aux_sym_number_token1,
    ACTIONS(2271), 1,
      aux_sym_number_token2,
    ACTIONS(2275), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2289), 1,
      sym__brace_start,
    ACTIONS(8066), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8070), 1,
      anon_sym_DQUOTE,
    ACTIONS(8074), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8076), 1,
      anon_sym_BQUOTE,
    ACTIONS(8078), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9580), 1,
      sym_word,
    ACTIONS(9588), 1,
      sym__comment_word,
    ACTIONS(10420), 1,
      anon_sym_DOLLAR,
    ACTIONS(8062), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8080), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9584), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9586), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1623), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [218803] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2341), 1,
      aux_sym_number_token1,
    ACTIONS(2343), 1,
      aux_sym_number_token2,
    ACTIONS(2347), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2357), 1,
      sym__brace_start,
    ACTIONS(9010), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9014), 1,
      anon_sym_DQUOTE,
    ACTIONS(9018), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9020), 1,
      anon_sym_BQUOTE,
    ACTIONS(9022), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10358), 1,
      sym_word,
    ACTIONS(10364), 1,
      sym__comment_word,
    ACTIONS(10422), 1,
      anon_sym_DOLLAR,
    ACTIONS(9008), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9024), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10360), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10362), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1462), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [218871] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8178), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8184), 1,
      anon_sym_DQUOTE,
    ACTIONS(8188), 1,
      aux_sym_number_token1,
    ACTIONS(8190), 1,
      aux_sym_number_token2,
    ACTIONS(8192), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8194), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(8196), 1,
      anon_sym_BQUOTE,
    ACTIONS(8198), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8208), 1,
      sym__brace_start,
    ACTIONS(9720), 1,
      sym_word,
    ACTIONS(9728), 1,
      sym__comment_word,
    ACTIONS(10424), 1,
      anon_sym_DOLLAR,
    ACTIONS(8174), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8200), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9724), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9726), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1590), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [218939] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8178), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8184), 1,
      anon_sym_DQUOTE,
    ACTIONS(8188), 1,
      aux_sym_number_token1,
    ACTIONS(8190), 1,
      aux_sym_number_token2,
    ACTIONS(8192), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8194), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(8196), 1,
      anon_sym_BQUOTE,
    ACTIONS(8198), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8208), 1,
      sym__brace_start,
    ACTIONS(9720), 1,
      sym_word,
    ACTIONS(9728), 1,
      sym__comment_word,
    ACTIONS(10426), 1,
      anon_sym_DOLLAR,
    ACTIONS(8174), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8200), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9724), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9726), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1590), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [219007] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(258), 1,
      anon_sym_DOLLAR,
    ACTIONS(264), 1,
      aux_sym_number_token1,
    ACTIONS(266), 1,
      aux_sym_number_token2,
    ACTIONS(270), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(284), 1,
      sym__brace_start,
    ACTIONS(1075), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1081), 1,
      anon_sym_DQUOTE,
    ACTIONS(1085), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1087), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3598), 1,
      anon_sym_BQUOTE,
    ACTIONS(10428), 1,
      sym_word,
    ACTIONS(10434), 1,
      sym__comment_word,
    ACTIONS(1073), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1089), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10430), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10432), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2732), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [219075] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2341), 1,
      aux_sym_number_token1,
    ACTIONS(2343), 1,
      aux_sym_number_token2,
    ACTIONS(2347), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2357), 1,
      sym__brace_start,
    ACTIONS(9010), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9014), 1,
      anon_sym_DQUOTE,
    ACTIONS(9018), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9020), 1,
      anon_sym_BQUOTE,
    ACTIONS(9022), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10358), 1,
      sym_word,
    ACTIONS(10364), 1,
      sym__comment_word,
    ACTIONS(10436), 1,
      anon_sym_DOLLAR,
    ACTIONS(9008), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9024), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10360), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10362), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1462), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [219143] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(314), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(320), 1,
      anon_sym_DQUOTE,
    ACTIONS(324), 1,
      aux_sym_number_token1,
    ACTIONS(326), 1,
      aux_sym_number_token2,
    ACTIONS(328), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(330), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(332), 1,
      anon_sym_BQUOTE,
    ACTIONS(334), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(344), 1,
      sym__brace_start,
    ACTIONS(9400), 1,
      sym_word,
    ACTIONS(9406), 1,
      sym__comment_word,
    ACTIONS(10438), 1,
      anon_sym_DOLLAR,
    ACTIONS(312), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(336), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9402), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9404), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1037), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [219211] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3064), 1,
      sym_variable_name,
    ACTIONS(9278), 1,
      anon_sym_LPAREN,
    ACTIONS(9280), 1,
      anon_sym_BANG,
    ACTIONS(9286), 1,
      anon_sym_TILDE,
    ACTIONS(9288), 1,
      anon_sym_DOLLAR,
    ACTIONS(9290), 1,
      anon_sym_DQUOTE,
    ACTIONS(9292), 1,
      aux_sym_number_token1,
    ACTIONS(9294), 1,
      aux_sym_number_token2,
    ACTIONS(9296), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9298), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9300), 1,
      anon_sym_BQUOTE,
    ACTIONS(9302), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10440), 1,
      aux_sym__simple_variable_name_token1,
    STATE(2801), 1,
      sym__arithmetic_binary_expression,
    STATE(2803), 1,
      sym__arithmetic_ternary_expression,
    STATE(2809), 1,
      sym__arithmetic_unary_expression,
    STATE(2825), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9282), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9284), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(2761), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [219285] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9364), 1,
      anon_sym_LPAREN,
    ACTIONS(9366), 1,
      anon_sym_BANG,
    ACTIONS(9372), 1,
      anon_sym_TILDE,
    ACTIONS(9374), 1,
      anon_sym_DOLLAR,
    ACTIONS(9376), 1,
      anon_sym_DQUOTE,
    ACTIONS(9378), 1,
      aux_sym_number_token1,
    ACTIONS(9380), 1,
      aux_sym_number_token2,
    ACTIONS(9382), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9384), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9386), 1,
      anon_sym_BQUOTE,
    ACTIONS(9388), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9392), 1,
      sym_variable_name,
    ACTIONS(10442), 1,
      aux_sym__simple_variable_name_token1,
    STATE(3391), 1,
      sym__arithmetic_binary_expression,
    STATE(3396), 1,
      sym__arithmetic_ternary_expression,
    STATE(3398), 1,
      sym__arithmetic_unary_expression,
    STATE(3400), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9368), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9370), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(3320), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [219359] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3064), 1,
      sym_variable_name,
    ACTIONS(9278), 1,
      anon_sym_LPAREN,
    ACTIONS(9280), 1,
      anon_sym_BANG,
    ACTIONS(9286), 1,
      anon_sym_TILDE,
    ACTIONS(9288), 1,
      anon_sym_DOLLAR,
    ACTIONS(9290), 1,
      anon_sym_DQUOTE,
    ACTIONS(9292), 1,
      aux_sym_number_token1,
    ACTIONS(9294), 1,
      aux_sym_number_token2,
    ACTIONS(9296), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9298), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9300), 1,
      anon_sym_BQUOTE,
    ACTIONS(9302), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10444), 1,
      aux_sym__simple_variable_name_token1,
    STATE(2801), 1,
      sym__arithmetic_binary_expression,
    STATE(2803), 1,
      sym__arithmetic_ternary_expression,
    STATE(2809), 1,
      sym__arithmetic_unary_expression,
    STATE(2825), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9282), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9284), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(2762), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [219433] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(314), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(320), 1,
      anon_sym_DQUOTE,
    ACTIONS(324), 1,
      aux_sym_number_token1,
    ACTIONS(326), 1,
      aux_sym_number_token2,
    ACTIONS(328), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(330), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(332), 1,
      anon_sym_BQUOTE,
    ACTIONS(334), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(344), 1,
      sym__brace_start,
    ACTIONS(9400), 1,
      sym_word,
    ACTIONS(9406), 1,
      sym__comment_word,
    ACTIONS(10446), 1,
      anon_sym_DOLLAR,
    ACTIONS(312), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(336), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9402), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9404), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1037), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [219501] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(47), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(53), 1,
      anon_sym_DQUOTE,
    ACTIONS(57), 1,
      aux_sym_number_token1,
    ACTIONS(59), 1,
      aux_sym_number_token2,
    ACTIONS(61), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(63), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(65), 1,
      anon_sym_BQUOTE,
    ACTIONS(67), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(79), 1,
      sym__brace_start,
    ACTIONS(10448), 1,
      sym_word,
    ACTIONS(10450), 1,
      anon_sym_DOLLAR,
    ACTIONS(10456), 1,
      sym__comment_word,
    ACTIONS(45), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(69), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10452), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10454), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1083), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [219569] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1141), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1149), 1,
      anon_sym_DQUOTE,
    ACTIONS(1153), 1,
      aux_sym_number_token1,
    ACTIONS(1155), 1,
      aux_sym_number_token2,
    ACTIONS(1157), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1159), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1163), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1169), 1,
      sym__brace_start,
    ACTIONS(3060), 1,
      anon_sym_BQUOTE,
    ACTIONS(10188), 1,
      sym_word,
    ACTIONS(10196), 1,
      sym__comment_word,
    ACTIONS(10458), 1,
      anon_sym_DOLLAR,
    ACTIONS(1129), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1165), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10192), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10194), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2392), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [219637] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1141), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1149), 1,
      anon_sym_DQUOTE,
    ACTIONS(1153), 1,
      aux_sym_number_token1,
    ACTIONS(1155), 1,
      aux_sym_number_token2,
    ACTIONS(1157), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1159), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1163), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1169), 1,
      sym__brace_start,
    ACTIONS(3060), 1,
      anon_sym_BQUOTE,
    ACTIONS(10188), 1,
      sym_word,
    ACTIONS(10196), 1,
      sym__comment_word,
    ACTIONS(10460), 1,
      anon_sym_DOLLAR,
    ACTIONS(1129), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1165), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10192), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10194), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2392), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [219705] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(451), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(457), 1,
      anon_sym_DQUOTE,
    ACTIONS(461), 1,
      aux_sym_number_token1,
    ACTIONS(463), 1,
      aux_sym_number_token2,
    ACTIONS(465), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(467), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(469), 1,
      anon_sym_BQUOTE,
    ACTIONS(471), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(481), 1,
      sym__brace_start,
    ACTIONS(10308), 1,
      sym_word,
    ACTIONS(10314), 1,
      sym__comment_word,
    ACTIONS(10462), 1,
      anon_sym_DOLLAR,
    ACTIONS(449), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(473), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10310), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10312), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(866), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [219773] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(707), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(709), 1,
      anon_sym_DOLLAR,
    ACTIONS(713), 1,
      anon_sym_DQUOTE,
    ACTIONS(717), 1,
      aux_sym_number_token1,
    ACTIONS(719), 1,
      aux_sym_number_token2,
    ACTIONS(721), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(723), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(725), 1,
      anon_sym_BQUOTE,
    ACTIONS(727), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(737), 1,
      sym__brace_start,
    ACTIONS(10464), 1,
      sym_word,
    ACTIONS(10470), 1,
      sym__comment_word,
    ACTIONS(705), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(729), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10466), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10468), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(999), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [219841] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8372), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8374), 1,
      anon_sym_DOLLAR,
    ACTIONS(8378), 1,
      anon_sym_DQUOTE,
    ACTIONS(8382), 1,
      aux_sym_number_token1,
    ACTIONS(8384), 1,
      aux_sym_number_token2,
    ACTIONS(8386), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8388), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(8390), 1,
      anon_sym_BQUOTE,
    ACTIONS(8392), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8402), 1,
      sym__brace_start,
    ACTIONS(10472), 1,
      sym_word,
    ACTIONS(10478), 1,
      sym__comment_word,
    ACTIONS(8368), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8394), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10474), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10476), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(4285), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [219909] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8268), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8270), 1,
      anon_sym_DOLLAR,
    ACTIONS(8274), 1,
      anon_sym_DQUOTE,
    ACTIONS(8278), 1,
      aux_sym_number_token1,
    ACTIONS(8280), 1,
      aux_sym_number_token2,
    ACTIONS(8282), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8284), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(8286), 1,
      anon_sym_BQUOTE,
    ACTIONS(8288), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8298), 1,
      sym__brace_start,
    ACTIONS(10480), 1,
      sym_word,
    ACTIONS(10486), 1,
      sym__comment_word,
    ACTIONS(8264), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8290), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10482), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10484), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1465), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [219977] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1831), 1,
      anon_sym_DOLLAR,
    ACTIONS(1837), 1,
      aux_sym_number_token1,
    ACTIONS(1839), 1,
      aux_sym_number_token2,
    ACTIONS(1843), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1857), 1,
      sym__brace_start,
    ACTIONS(8232), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8236), 1,
      anon_sym_DQUOTE,
    ACTIONS(8240), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8242), 1,
      anon_sym_BQUOTE,
    ACTIONS(8244), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10488), 1,
      sym_word,
    ACTIONS(10494), 1,
      sym__comment_word,
    ACTIONS(8228), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8246), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10490), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10492), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1440), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [220045] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9364), 1,
      anon_sym_LPAREN,
    ACTIONS(9366), 1,
      anon_sym_BANG,
    ACTIONS(9372), 1,
      anon_sym_TILDE,
    ACTIONS(9374), 1,
      anon_sym_DOLLAR,
    ACTIONS(9376), 1,
      anon_sym_DQUOTE,
    ACTIONS(9378), 1,
      aux_sym_number_token1,
    ACTIONS(9380), 1,
      aux_sym_number_token2,
    ACTIONS(9382), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9384), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9386), 1,
      anon_sym_BQUOTE,
    ACTIONS(9388), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9392), 1,
      sym_variable_name,
    ACTIONS(10496), 1,
      aux_sym__simple_variable_name_token1,
    STATE(3391), 1,
      sym__arithmetic_binary_expression,
    STATE(3396), 1,
      sym__arithmetic_ternary_expression,
    STATE(3398), 1,
      sym__arithmetic_unary_expression,
    STATE(3400), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9368), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9370), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(3317), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [220119] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3064), 1,
      sym_variable_name,
    ACTIONS(9278), 1,
      anon_sym_LPAREN,
    ACTIONS(9280), 1,
      anon_sym_BANG,
    ACTIONS(9286), 1,
      anon_sym_TILDE,
    ACTIONS(9288), 1,
      anon_sym_DOLLAR,
    ACTIONS(9290), 1,
      anon_sym_DQUOTE,
    ACTIONS(9292), 1,
      aux_sym_number_token1,
    ACTIONS(9294), 1,
      aux_sym_number_token2,
    ACTIONS(9296), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9298), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9300), 1,
      anon_sym_BQUOTE,
    ACTIONS(9302), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10498), 1,
      aux_sym__simple_variable_name_token1,
    STATE(2801), 1,
      sym__arithmetic_binary_expression,
    STATE(2803), 1,
      sym__arithmetic_ternary_expression,
    STATE(2809), 1,
      sym__arithmetic_unary_expression,
    STATE(2825), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9282), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9284), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(2763), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [220193] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2786), 1,
      anon_sym_DOLLAR,
    ACTIONS(2792), 1,
      aux_sym_number_token1,
    ACTIONS(2794), 1,
      aux_sym_number_token2,
    ACTIONS(2798), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2810), 1,
      sym__brace_start,
    ACTIONS(10500), 1,
      sym_word,
    ACTIONS(10504), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(10508), 1,
      anon_sym_DQUOTE,
    ACTIONS(10512), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(10514), 1,
      anon_sym_BQUOTE,
    ACTIONS(10516), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10520), 1,
      sym__comment_word,
    ACTIONS(10502), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(10506), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10518), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10510), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1569), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [220261] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1454), 1,
      anon_sym_DOLLAR,
    ACTIONS(1460), 1,
      aux_sym_number_token1,
    ACTIONS(1462), 1,
      aux_sym_number_token2,
    ACTIONS(1466), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1478), 1,
      sym__brace_start,
    ACTIONS(8878), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8882), 1,
      anon_sym_DQUOTE,
    ACTIONS(8886), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8888), 1,
      anon_sym_BQUOTE,
    ACTIONS(8890), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10522), 1,
      sym_word,
    ACTIONS(10528), 1,
      sym__comment_word,
    ACTIONS(8876), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8892), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10524), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10526), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1120), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [220329] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3339), 1,
      anon_sym_DOLLAR,
    ACTIONS(3343), 1,
      aux_sym_number_token1,
    ACTIONS(3345), 1,
      aux_sym_number_token2,
    ACTIONS(3349), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3359), 1,
      sym__brace_start,
    ACTIONS(9046), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9050), 1,
      anon_sym_DQUOTE,
    ACTIONS(9054), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9056), 1,
      anon_sym_BQUOTE,
    ACTIONS(9058), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10530), 1,
      sym_word,
    ACTIONS(10536), 1,
      sym__comment_word,
    ACTIONS(9044), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9060), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10532), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10534), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1827), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [220397] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8512), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8518), 1,
      anon_sym_DQUOTE,
    ACTIONS(8522), 1,
      aux_sym_number_token1,
    ACTIONS(8524), 1,
      aux_sym_number_token2,
    ACTIONS(8526), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8528), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(8530), 1,
      anon_sym_BQUOTE,
    ACTIONS(8532), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8542), 1,
      sym__brace_start,
    ACTIONS(10220), 1,
      sym_word,
    ACTIONS(10226), 1,
      sym__comment_word,
    ACTIONS(10538), 1,
      anon_sym_DOLLAR,
    ACTIONS(8508), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8534), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10222), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10224), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2514), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [220465] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3675), 1,
      anon_sym_DOLLAR,
    ACTIONS(3681), 1,
      aux_sym_number_token1,
    ACTIONS(3683), 1,
      aux_sym_number_token2,
    ACTIONS(3687), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3697), 1,
      sym__brace_start,
    ACTIONS(10062), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(10066), 1,
      anon_sym_DQUOTE,
    ACTIONS(10070), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(10072), 1,
      anon_sym_BQUOTE,
    ACTIONS(10074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10540), 1,
      sym_word,
    ACTIONS(10546), 1,
      sym__comment_word,
    ACTIONS(10060), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(10076), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10542), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10544), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(4831), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [220533] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2574), 1,
      anon_sym_DOLLAR,
    ACTIONS(2580), 1,
      aux_sym_number_token1,
    ACTIONS(2582), 1,
      aux_sym_number_token2,
    ACTIONS(2586), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2596), 1,
      sym__brace_start,
    ACTIONS(9064), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9068), 1,
      anon_sym_DQUOTE,
    ACTIONS(9072), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9074), 1,
      anon_sym_BQUOTE,
    ACTIONS(9076), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10548), 1,
      sym_word,
    ACTIONS(10554), 1,
      sym__comment_word,
    ACTIONS(9062), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9078), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10550), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10552), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1669), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [220601] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3026), 1,
      anon_sym_DOLLAR,
    ACTIONS(3032), 1,
      aux_sym_number_token1,
    ACTIONS(3034), 1,
      aux_sym_number_token2,
    ACTIONS(3038), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3048), 1,
      sym__brace_start,
    ACTIONS(10134), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(10138), 1,
      anon_sym_DQUOTE,
    ACTIONS(10142), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(10144), 1,
      anon_sym_BQUOTE,
    ACTIONS(10146), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10556), 1,
      sym_word,
    ACTIONS(10562), 1,
      sym__comment_word,
    ACTIONS(10132), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(10148), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10558), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10560), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(4493), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [220669] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8140), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8142), 1,
      anon_sym_DOLLAR,
    ACTIONS(8146), 1,
      anon_sym_DQUOTE,
    ACTIONS(8150), 1,
      aux_sym_number_token1,
    ACTIONS(8152), 1,
      aux_sym_number_token2,
    ACTIONS(8154), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8156), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(8158), 1,
      anon_sym_BQUOTE,
    ACTIONS(8160), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8170), 1,
      sym__brace_start,
    ACTIONS(10564), 1,
      sym_word,
    ACTIONS(10570), 1,
      sym__comment_word,
    ACTIONS(8136), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8162), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10566), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10568), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(4526), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [220737] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(264), 1,
      aux_sym_number_token1,
    ACTIONS(266), 1,
      aux_sym_number_token2,
    ACTIONS(270), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(284), 1,
      sym__brace_start,
    ACTIONS(1075), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1081), 1,
      anon_sym_DQUOTE,
    ACTIONS(1085), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1087), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3598), 1,
      anon_sym_BQUOTE,
    ACTIONS(10428), 1,
      sym_word,
    ACTIONS(10434), 1,
      sym__comment_word,
    ACTIONS(10572), 1,
      anon_sym_DOLLAR,
    ACTIONS(1073), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1089), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10430), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10432), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2732), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [220805] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2869), 1,
      aux_sym_number_token1,
    ACTIONS(2871), 1,
      aux_sym_number_token2,
    ACTIONS(2875), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2885), 1,
      sym__brace_start,
    ACTIONS(9940), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9944), 1,
      anon_sym_DQUOTE,
    ACTIONS(9948), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9950), 1,
      anon_sym_BQUOTE,
    ACTIONS(9952), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10382), 1,
      sym_word,
    ACTIONS(10388), 1,
      sym__comment_word,
    ACTIONS(10574), 1,
      anon_sym_DOLLAR,
    ACTIONS(9938), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9954), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10384), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10386), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(4404), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [220873] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4151), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(4157), 1,
      anon_sym_DQUOTE,
    ACTIONS(4161), 1,
      aux_sym_number_token1,
    ACTIONS(4163), 1,
      aux_sym_number_token2,
    ACTIONS(4165), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(4167), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(4169), 1,
      anon_sym_BQUOTE,
    ACTIONS(4171), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(4181), 1,
      sym__brace_start,
    ACTIONS(9502), 1,
      sym_word,
    ACTIONS(9508), 1,
      sym__comment_word,
    ACTIONS(10576), 1,
      anon_sym_DOLLAR,
    ACTIONS(4149), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(4173), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9504), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9506), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2580), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [220941] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(264), 1,
      aux_sym_number_token1,
    ACTIONS(266), 1,
      aux_sym_number_token2,
    ACTIONS(270), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(284), 1,
      sym__brace_start,
    ACTIONS(1075), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1081), 1,
      anon_sym_DQUOTE,
    ACTIONS(1085), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1087), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3598), 1,
      anon_sym_BQUOTE,
    ACTIONS(10428), 1,
      sym_word,
    ACTIONS(10434), 1,
      sym__comment_word,
    ACTIONS(10578), 1,
      anon_sym_DOLLAR,
    ACTIONS(1073), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1089), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10430), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10432), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2732), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [221009] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4947), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(4953), 1,
      anon_sym_DQUOTE,
    ACTIONS(4957), 1,
      aux_sym_number_token1,
    ACTIONS(4959), 1,
      aux_sym_number_token2,
    ACTIONS(4961), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(4963), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(4965), 1,
      anon_sym_BQUOTE,
    ACTIONS(4967), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(4973), 1,
      sym__brace_start,
    ACTIONS(10228), 1,
      sym_word,
    ACTIONS(10234), 1,
      sym__comment_word,
    ACTIONS(10580), 1,
      anon_sym_DOLLAR,
    ACTIONS(4945), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(4969), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10230), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10232), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2590), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [221077] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2869), 1,
      aux_sym_number_token1,
    ACTIONS(2871), 1,
      aux_sym_number_token2,
    ACTIONS(2875), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2885), 1,
      sym__brace_start,
    ACTIONS(9940), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9944), 1,
      anon_sym_DQUOTE,
    ACTIONS(9948), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9950), 1,
      anon_sym_BQUOTE,
    ACTIONS(9952), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10382), 1,
      sym_word,
    ACTIONS(10388), 1,
      sym__comment_word,
    ACTIONS(10582), 1,
      anon_sym_DOLLAR,
    ACTIONS(9938), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9954), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10384), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10386), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(4404), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [221145] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3602), 1,
      sym_variable_name,
    ACTIONS(9898), 1,
      anon_sym_LPAREN,
    ACTIONS(9900), 1,
      anon_sym_BANG,
    ACTIONS(9906), 1,
      anon_sym_TILDE,
    ACTIONS(9908), 1,
      anon_sym_DOLLAR,
    ACTIONS(9910), 1,
      anon_sym_DQUOTE,
    ACTIONS(9912), 1,
      aux_sym_number_token1,
    ACTIONS(9914), 1,
      aux_sym_number_token2,
    ACTIONS(9916), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9918), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9920), 1,
      anon_sym_BQUOTE,
    ACTIONS(9922), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10584), 1,
      aux_sym__simple_variable_name_token1,
    STATE(3359), 1,
      sym__arithmetic_postfix_expression,
    STATE(3361), 1,
      sym__arithmetic_unary_expression,
    STATE(3367), 1,
      sym__arithmetic_ternary_expression,
    STATE(3375), 1,
      sym__arithmetic_binary_expression,
    ACTIONS(9902), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9904), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(3262), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [221219] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4151), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(4157), 1,
      anon_sym_DQUOTE,
    ACTIONS(4161), 1,
      aux_sym_number_token1,
    ACTIONS(4163), 1,
      aux_sym_number_token2,
    ACTIONS(4165), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(4167), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(4169), 1,
      anon_sym_BQUOTE,
    ACTIONS(4171), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(4181), 1,
      sym__brace_start,
    ACTIONS(9502), 1,
      sym_word,
    ACTIONS(9508), 1,
      sym__comment_word,
    ACTIONS(10586), 1,
      anon_sym_DOLLAR,
    ACTIONS(4149), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(4173), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9504), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9506), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2580), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [221287] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4947), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(4953), 1,
      anon_sym_DQUOTE,
    ACTIONS(4957), 1,
      aux_sym_number_token1,
    ACTIONS(4959), 1,
      aux_sym_number_token2,
    ACTIONS(4961), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(4963), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(4965), 1,
      anon_sym_BQUOTE,
    ACTIONS(4967), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(4973), 1,
      sym__brace_start,
    ACTIONS(10228), 1,
      sym_word,
    ACTIONS(10234), 1,
      sym__comment_word,
    ACTIONS(10588), 1,
      anon_sym_DOLLAR,
    ACTIONS(4945), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(4969), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10230), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10232), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2590), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [221355] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7996), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(7998), 1,
      anon_sym_DOLLAR,
    ACTIONS(8000), 1,
      sym__special_character,
    ACTIONS(8002), 1,
      anon_sym_DQUOTE,
    ACTIONS(8006), 1,
      aux_sym_number_token1,
    ACTIONS(8008), 1,
      aux_sym_number_token2,
    ACTIONS(8010), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8012), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(8014), 1,
      anon_sym_BQUOTE,
    ACTIONS(8016), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8026), 1,
      sym__brace_start,
    ACTIONS(10590), 1,
      sym_word,
    ACTIONS(10594), 1,
      sym_test_operator,
    STATE(3539), 1,
      aux_sym__literal_repeat1,
    STATE(3547), 1,
      sym_concatenation,
    ACTIONS(7992), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8018), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10592), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(3467), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [221427] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3793), 1,
      aux_sym_number_token1,
    ACTIONS(3795), 1,
      aux_sym_number_token2,
    ACTIONS(3799), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3809), 1,
      sym__brace_start,
    ACTIONS(9636), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9640), 1,
      anon_sym_DQUOTE,
    ACTIONS(9644), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9646), 1,
      anon_sym_BQUOTE,
    ACTIONS(9648), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9872), 1,
      sym_word,
    ACTIONS(9880), 1,
      sym__comment_word,
    ACTIONS(10596), 1,
      anon_sym_DOLLAR,
    ACTIONS(9634), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9650), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9876), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9878), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(4337), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [221495] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7996), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(7998), 1,
      anon_sym_DOLLAR,
    ACTIONS(8000), 1,
      sym__special_character,
    ACTIONS(8002), 1,
      anon_sym_DQUOTE,
    ACTIONS(8006), 1,
      aux_sym_number_token1,
    ACTIONS(8008), 1,
      aux_sym_number_token2,
    ACTIONS(8010), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8012), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(8014), 1,
      anon_sym_BQUOTE,
    ACTIONS(8016), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8026), 1,
      sym__brace_start,
    ACTIONS(10598), 1,
      sym_word,
    ACTIONS(10602), 1,
      sym_test_operator,
    STATE(3542), 1,
      aux_sym__literal_repeat1,
    STATE(3569), 1,
      sym_concatenation,
    ACTIONS(7992), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8018), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10600), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(3456), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [221567] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3064), 1,
      sym_variable_name,
    ACTIONS(9278), 1,
      anon_sym_LPAREN,
    ACTIONS(9280), 1,
      anon_sym_BANG,
    ACTIONS(9286), 1,
      anon_sym_TILDE,
    ACTIONS(9288), 1,
      anon_sym_DOLLAR,
    ACTIONS(9290), 1,
      anon_sym_DQUOTE,
    ACTIONS(9292), 1,
      aux_sym_number_token1,
    ACTIONS(9294), 1,
      aux_sym_number_token2,
    ACTIONS(9296), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9298), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9300), 1,
      anon_sym_BQUOTE,
    ACTIONS(9302), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10604), 1,
      aux_sym__simple_variable_name_token1,
    STATE(2801), 1,
      sym__arithmetic_binary_expression,
    STATE(2803), 1,
      sym__arithmetic_ternary_expression,
    STATE(2809), 1,
      sym__arithmetic_unary_expression,
    STATE(2825), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9282), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9284), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(2880), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [221641] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9364), 1,
      anon_sym_LPAREN,
    ACTIONS(9366), 1,
      anon_sym_BANG,
    ACTIONS(9372), 1,
      anon_sym_TILDE,
    ACTIONS(9374), 1,
      anon_sym_DOLLAR,
    ACTIONS(9376), 1,
      anon_sym_DQUOTE,
    ACTIONS(9378), 1,
      aux_sym_number_token1,
    ACTIONS(9380), 1,
      aux_sym_number_token2,
    ACTIONS(9382), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9384), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9386), 1,
      anon_sym_BQUOTE,
    ACTIONS(9388), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9392), 1,
      sym_variable_name,
    ACTIONS(10606), 1,
      aux_sym__simple_variable_name_token1,
    STATE(3391), 1,
      sym__arithmetic_binary_expression,
    STATE(3396), 1,
      sym__arithmetic_ternary_expression,
    STATE(3398), 1,
      sym__arithmetic_unary_expression,
    STATE(3400), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9368), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9370), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(3426), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [221715] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(7958), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(7964), 1,
      anon_sym_DQUOTE,
    ACTIONS(7968), 1,
      aux_sym_number_token1,
    ACTIONS(7970), 1,
      aux_sym_number_token2,
    ACTIONS(7972), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(7974), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(7976), 1,
      anon_sym_BQUOTE,
    ACTIONS(7978), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(7988), 1,
      sym__brace_start,
    ACTIONS(9860), 1,
      sym_word,
    ACTIONS(9868), 1,
      sym__comment_word,
    ACTIONS(10608), 1,
      anon_sym_DOLLAR,
    ACTIONS(7954), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(7980), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9864), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9866), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(4357), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [221783] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(191), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(197), 1,
      anon_sym_DQUOTE,
    ACTIONS(201), 1,
      aux_sym_number_token1,
    ACTIONS(203), 1,
      aux_sym_number_token2,
    ACTIONS(205), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(207), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(209), 1,
      anon_sym_BQUOTE,
    ACTIONS(211), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(221), 1,
      sym__brace_start,
    ACTIONS(10406), 1,
      sym_word,
    ACTIONS(10412), 1,
      sym__comment_word,
    ACTIONS(10610), 1,
      anon_sym_DOLLAR,
    ACTIONS(189), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(213), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10408), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10410), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(513), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [221851] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3064), 1,
      sym_variable_name,
    ACTIONS(9278), 1,
      anon_sym_LPAREN,
    ACTIONS(9280), 1,
      anon_sym_BANG,
    ACTIONS(9286), 1,
      anon_sym_TILDE,
    ACTIONS(9288), 1,
      anon_sym_DOLLAR,
    ACTIONS(9290), 1,
      anon_sym_DQUOTE,
    ACTIONS(9292), 1,
      aux_sym_number_token1,
    ACTIONS(9294), 1,
      aux_sym_number_token2,
    ACTIONS(9296), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9298), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9300), 1,
      anon_sym_BQUOTE,
    ACTIONS(9302), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10612), 1,
      aux_sym__simple_variable_name_token1,
    STATE(2801), 1,
      sym__arithmetic_binary_expression,
    STATE(2803), 1,
      sym__arithmetic_ternary_expression,
    STATE(2809), 1,
      sym__arithmetic_unary_expression,
    STATE(2825), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9282), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9284), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(2898), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [221925] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1043), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1049), 1,
      anon_sym_DQUOTE,
    ACTIONS(1053), 1,
      aux_sym_number_token1,
    ACTIONS(1055), 1,
      aux_sym_number_token2,
    ACTIONS(1057), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1059), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1061), 1,
      anon_sym_BQUOTE,
    ACTIONS(1063), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1071), 1,
      sym__brace_start,
    ACTIONS(9706), 1,
      sym_word,
    ACTIONS(9712), 1,
      sym__comment_word,
    ACTIONS(10614), 1,
      anon_sym_DOLLAR,
    ACTIONS(1041), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1065), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9708), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9710), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2075), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [221993] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(7958), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(7964), 1,
      anon_sym_DQUOTE,
    ACTIONS(7968), 1,
      aux_sym_number_token1,
    ACTIONS(7970), 1,
      aux_sym_number_token2,
    ACTIONS(7972), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(7974), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(7976), 1,
      anon_sym_BQUOTE,
    ACTIONS(7978), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(7988), 1,
      sym__brace_start,
    ACTIONS(9860), 1,
      sym_word,
    ACTIONS(9868), 1,
      sym__comment_word,
    ACTIONS(10616), 1,
      anon_sym_DOLLAR,
    ACTIONS(7954), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(7980), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9864), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9866), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(4357), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [222061] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9364), 1,
      anon_sym_LPAREN,
    ACTIONS(9366), 1,
      anon_sym_BANG,
    ACTIONS(9372), 1,
      anon_sym_TILDE,
    ACTIONS(9374), 1,
      anon_sym_DOLLAR,
    ACTIONS(9376), 1,
      anon_sym_DQUOTE,
    ACTIONS(9378), 1,
      aux_sym_number_token1,
    ACTIONS(9380), 1,
      aux_sym_number_token2,
    ACTIONS(9382), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9384), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9386), 1,
      anon_sym_BQUOTE,
    ACTIONS(9388), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9392), 1,
      sym_variable_name,
    ACTIONS(10618), 1,
      aux_sym__simple_variable_name_token1,
    STATE(3391), 1,
      sym__arithmetic_binary_expression,
    STATE(3396), 1,
      sym__arithmetic_ternary_expression,
    STATE(3398), 1,
      sym__arithmetic_unary_expression,
    STATE(3400), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9368), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9370), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(3316), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [222135] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3064), 1,
      sym_variable_name,
    ACTIONS(9278), 1,
      anon_sym_LPAREN,
    ACTIONS(9280), 1,
      anon_sym_BANG,
    ACTIONS(9286), 1,
      anon_sym_TILDE,
    ACTIONS(9288), 1,
      anon_sym_DOLLAR,
    ACTIONS(9290), 1,
      anon_sym_DQUOTE,
    ACTIONS(9292), 1,
      aux_sym_number_token1,
    ACTIONS(9294), 1,
      aux_sym_number_token2,
    ACTIONS(9296), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9298), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9300), 1,
      anon_sym_BQUOTE,
    ACTIONS(9302), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10620), 1,
      aux_sym__simple_variable_name_token1,
    STATE(2801), 1,
      sym__arithmetic_binary_expression,
    STATE(2803), 1,
      sym__arithmetic_ternary_expression,
    STATE(2809), 1,
      sym__arithmetic_unary_expression,
    STATE(2825), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9282), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9284), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(2764), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [222209] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3401), 1,
      aux_sym_number_token1,
    ACTIONS(3403), 1,
      aux_sym_number_token2,
    ACTIONS(3407), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3417), 1,
      sym__brace_start,
    ACTIONS(8926), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8930), 1,
      anon_sym_DQUOTE,
    ACTIONS(8934), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8936), 1,
      anon_sym_BQUOTE,
    ACTIONS(8938), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9696), 1,
      sym_word,
    ACTIONS(9704), 1,
      sym__comment_word,
    ACTIONS(10622), 1,
      anon_sym_DOLLAR,
    ACTIONS(8924), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8940), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9700), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9702), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1851), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [222277] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3401), 1,
      aux_sym_number_token1,
    ACTIONS(3403), 1,
      aux_sym_number_token2,
    ACTIONS(3407), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3417), 1,
      sym__brace_start,
    ACTIONS(8926), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8930), 1,
      anon_sym_DQUOTE,
    ACTIONS(8934), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8936), 1,
      anon_sym_BQUOTE,
    ACTIONS(8938), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9696), 1,
      sym_word,
    ACTIONS(9704), 1,
      sym__comment_word,
    ACTIONS(10624), 1,
      anon_sym_DOLLAR,
    ACTIONS(8924), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8940), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9700), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9702), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1851), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [222345] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9364), 1,
      anon_sym_LPAREN,
    ACTIONS(9366), 1,
      anon_sym_BANG,
    ACTIONS(9372), 1,
      anon_sym_TILDE,
    ACTIONS(9374), 1,
      anon_sym_DOLLAR,
    ACTIONS(9376), 1,
      anon_sym_DQUOTE,
    ACTIONS(9378), 1,
      aux_sym_number_token1,
    ACTIONS(9380), 1,
      aux_sym_number_token2,
    ACTIONS(9382), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9384), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9386), 1,
      anon_sym_BQUOTE,
    ACTIONS(9388), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9392), 1,
      sym_variable_name,
    ACTIONS(10626), 1,
      aux_sym__simple_variable_name_token1,
    STATE(3391), 1,
      sym__arithmetic_binary_expression,
    STATE(3396), 1,
      sym__arithmetic_ternary_expression,
    STATE(3398), 1,
      sym__arithmetic_unary_expression,
    STATE(3400), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9368), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9370), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(3315), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [222419] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3064), 1,
      sym_variable_name,
    ACTIONS(9278), 1,
      anon_sym_LPAREN,
    ACTIONS(9280), 1,
      anon_sym_BANG,
    ACTIONS(9286), 1,
      anon_sym_TILDE,
    ACTIONS(9288), 1,
      anon_sym_DOLLAR,
    ACTIONS(9290), 1,
      anon_sym_DQUOTE,
    ACTIONS(9292), 1,
      aux_sym_number_token1,
    ACTIONS(9294), 1,
      aux_sym_number_token2,
    ACTIONS(9296), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9298), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9300), 1,
      anon_sym_BQUOTE,
    ACTIONS(9302), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10628), 1,
      aux_sym__simple_variable_name_token1,
    STATE(2801), 1,
      sym__arithmetic_binary_expression,
    STATE(2803), 1,
      sym__arithmetic_ternary_expression,
    STATE(2809), 1,
      sym__arithmetic_unary_expression,
    STATE(2825), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9282), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9284), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(2765), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [222493] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3616), 1,
      aux_sym_number_token1,
    ACTIONS(3618), 1,
      aux_sym_number_token2,
    ACTIONS(3622), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3632), 1,
      sym__brace_start,
    ACTIONS(8850), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8854), 1,
      anon_sym_DQUOTE,
    ACTIONS(8858), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8860), 1,
      anon_sym_BQUOTE,
    ACTIONS(8862), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9516), 1,
      sym_word,
    ACTIONS(9524), 1,
      sym__comment_word,
    ACTIONS(10630), 1,
      anon_sym_DOLLAR,
    ACTIONS(8848), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8864), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9520), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9522), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2210), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [222561] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3616), 1,
      aux_sym_number_token1,
    ACTIONS(3618), 1,
      aux_sym_number_token2,
    ACTIONS(3622), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3632), 1,
      sym__brace_start,
    ACTIONS(8850), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8854), 1,
      anon_sym_DQUOTE,
    ACTIONS(8858), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8860), 1,
      anon_sym_BQUOTE,
    ACTIONS(8862), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9516), 1,
      sym_word,
    ACTIONS(9524), 1,
      sym__comment_word,
    ACTIONS(10632), 1,
      anon_sym_DOLLAR,
    ACTIONS(8848), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8864), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9520), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9522), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2210), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [222629] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3602), 1,
      sym_variable_name,
    ACTIONS(9898), 1,
      anon_sym_LPAREN,
    ACTIONS(9900), 1,
      anon_sym_BANG,
    ACTIONS(9906), 1,
      anon_sym_TILDE,
    ACTIONS(9908), 1,
      anon_sym_DOLLAR,
    ACTIONS(9910), 1,
      anon_sym_DQUOTE,
    ACTIONS(9912), 1,
      aux_sym_number_token1,
    ACTIONS(9914), 1,
      aux_sym_number_token2,
    ACTIONS(9916), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9918), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9920), 1,
      anon_sym_BQUOTE,
    ACTIONS(9922), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10634), 1,
      aux_sym__simple_variable_name_token1,
    STATE(3359), 1,
      sym__arithmetic_postfix_expression,
    STATE(3361), 1,
      sym__arithmetic_unary_expression,
    STATE(3367), 1,
      sym__arithmetic_ternary_expression,
    STATE(3375), 1,
      sym__arithmetic_binary_expression,
    ACTIONS(9902), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9904), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(3408), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [222703] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9364), 1,
      anon_sym_LPAREN,
    ACTIONS(9366), 1,
      anon_sym_BANG,
    ACTIONS(9372), 1,
      anon_sym_TILDE,
    ACTIONS(9374), 1,
      anon_sym_DOLLAR,
    ACTIONS(9376), 1,
      anon_sym_DQUOTE,
    ACTIONS(9378), 1,
      aux_sym_number_token1,
    ACTIONS(9380), 1,
      aux_sym_number_token2,
    ACTIONS(9382), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9384), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9386), 1,
      anon_sym_BQUOTE,
    ACTIONS(9388), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9392), 1,
      sym_variable_name,
    ACTIONS(10636), 1,
      aux_sym__simple_variable_name_token1,
    STATE(3391), 1,
      sym__arithmetic_binary_expression,
    STATE(3396), 1,
      sym__arithmetic_ternary_expression,
    STATE(3398), 1,
      sym__arithmetic_unary_expression,
    STATE(3400), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9368), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9370), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(3166), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [222777] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3064), 1,
      sym_variable_name,
    ACTIONS(9278), 1,
      anon_sym_LPAREN,
    ACTIONS(9280), 1,
      anon_sym_BANG,
    ACTIONS(9286), 1,
      anon_sym_TILDE,
    ACTIONS(9288), 1,
      anon_sym_DOLLAR,
    ACTIONS(9290), 1,
      anon_sym_DQUOTE,
    ACTIONS(9292), 1,
      aux_sym_number_token1,
    ACTIONS(9294), 1,
      aux_sym_number_token2,
    ACTIONS(9296), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9298), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9300), 1,
      anon_sym_BQUOTE,
    ACTIONS(9302), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10638), 1,
      aux_sym__simple_variable_name_token1,
    STATE(2801), 1,
      sym__arithmetic_binary_expression,
    STATE(2803), 1,
      sym__arithmetic_ternary_expression,
    STATE(2809), 1,
      sym__arithmetic_unary_expression,
    STATE(2825), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9282), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9284), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(2766), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [222851] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(707), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(713), 1,
      anon_sym_DQUOTE,
    ACTIONS(717), 1,
      aux_sym_number_token1,
    ACTIONS(719), 1,
      aux_sym_number_token2,
    ACTIONS(721), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(723), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(725), 1,
      anon_sym_BQUOTE,
    ACTIONS(727), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(737), 1,
      sym__brace_start,
    ACTIONS(10464), 1,
      sym_word,
    ACTIONS(10470), 1,
      sym__comment_word,
    ACTIONS(10640), 1,
      anon_sym_DOLLAR,
    ACTIONS(705), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(729), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10466), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10468), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(999), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [222919] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9364), 1,
      anon_sym_LPAREN,
    ACTIONS(9366), 1,
      anon_sym_BANG,
    ACTIONS(9372), 1,
      anon_sym_TILDE,
    ACTIONS(9374), 1,
      anon_sym_DOLLAR,
    ACTIONS(9376), 1,
      anon_sym_DQUOTE,
    ACTIONS(9378), 1,
      aux_sym_number_token1,
    ACTIONS(9380), 1,
      aux_sym_number_token2,
    ACTIONS(9382), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9384), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9386), 1,
      anon_sym_BQUOTE,
    ACTIONS(9388), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9392), 1,
      sym_variable_name,
    ACTIONS(10642), 1,
      aux_sym__simple_variable_name_token1,
    STATE(3391), 1,
      sym__arithmetic_binary_expression,
    STATE(3396), 1,
      sym__arithmetic_ternary_expression,
    STATE(3398), 1,
      sym__arithmetic_unary_expression,
    STATE(3400), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9368), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9370), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(3309), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [222993] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3064), 1,
      sym_variable_name,
    ACTIONS(9278), 1,
      anon_sym_LPAREN,
    ACTIONS(9280), 1,
      anon_sym_BANG,
    ACTIONS(9286), 1,
      anon_sym_TILDE,
    ACTIONS(9288), 1,
      anon_sym_DOLLAR,
    ACTIONS(9290), 1,
      anon_sym_DQUOTE,
    ACTIONS(9292), 1,
      aux_sym_number_token1,
    ACTIONS(9294), 1,
      aux_sym_number_token2,
    ACTIONS(9296), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9298), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9300), 1,
      anon_sym_BQUOTE,
    ACTIONS(9302), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10644), 1,
      aux_sym__simple_variable_name_token1,
    STATE(2801), 1,
      sym__arithmetic_binary_expression,
    STATE(2803), 1,
      sym__arithmetic_ternary_expression,
    STATE(2809), 1,
      sym__arithmetic_unary_expression,
    STATE(2825), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9282), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9284), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(2767), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [223067] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3064), 1,
      sym_variable_name,
    ACTIONS(9278), 1,
      anon_sym_LPAREN,
    ACTIONS(9280), 1,
      anon_sym_BANG,
    ACTIONS(9286), 1,
      anon_sym_TILDE,
    ACTIONS(9288), 1,
      anon_sym_DOLLAR,
    ACTIONS(9290), 1,
      anon_sym_DQUOTE,
    ACTIONS(9292), 1,
      aux_sym_number_token1,
    ACTIONS(9294), 1,
      aux_sym_number_token2,
    ACTIONS(9296), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9298), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9300), 1,
      anon_sym_BQUOTE,
    ACTIONS(9302), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10646), 1,
      aux_sym__simple_variable_name_token1,
    STATE(2801), 1,
      sym__arithmetic_binary_expression,
    STATE(2803), 1,
      sym__arithmetic_ternary_expression,
    STATE(2809), 1,
      sym__arithmetic_unary_expression,
    STATE(2825), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9282), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9284), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(3401), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [223141] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(707), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(713), 1,
      anon_sym_DQUOTE,
    ACTIONS(717), 1,
      aux_sym_number_token1,
    ACTIONS(719), 1,
      aux_sym_number_token2,
    ACTIONS(721), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(723), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(725), 1,
      anon_sym_BQUOTE,
    ACTIONS(727), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(737), 1,
      sym__brace_start,
    ACTIONS(10464), 1,
      sym_word,
    ACTIONS(10470), 1,
      sym__comment_word,
    ACTIONS(10648), 1,
      anon_sym_DOLLAR,
    ACTIONS(705), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(729), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10466), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10468), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(999), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [223209] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3737), 1,
      aux_sym_number_token1,
    ACTIONS(3739), 1,
      aux_sym_number_token2,
    ACTIONS(3743), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3755), 1,
      sym__brace_start,
    ACTIONS(8820), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8824), 1,
      anon_sym_DQUOTE,
    ACTIONS(8828), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8830), 1,
      anon_sym_BQUOTE,
    ACTIONS(8832), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9412), 1,
      sym_word,
    ACTIONS(9420), 1,
      sym__comment_word,
    ACTIONS(10650), 1,
      anon_sym_DOLLAR,
    ACTIONS(8818), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8834), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9416), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9418), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(4810), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [223277] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1043), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1049), 1,
      anon_sym_DQUOTE,
    ACTIONS(1053), 1,
      aux_sym_number_token1,
    ACTIONS(1055), 1,
      aux_sym_number_token2,
    ACTIONS(1057), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1059), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1061), 1,
      anon_sym_BQUOTE,
    ACTIONS(1063), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1071), 1,
      sym__brace_start,
    ACTIONS(9706), 1,
      sym_word,
    ACTIONS(9712), 1,
      sym__comment_word,
    ACTIONS(10652), 1,
      anon_sym_DOLLAR,
    ACTIONS(1041), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1065), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9708), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9710), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2075), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [223345] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(191), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(197), 1,
      anon_sym_DQUOTE,
    ACTIONS(201), 1,
      aux_sym_number_token1,
    ACTIONS(203), 1,
      aux_sym_number_token2,
    ACTIONS(205), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(207), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(209), 1,
      anon_sym_BQUOTE,
    ACTIONS(211), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(221), 1,
      sym__brace_start,
    ACTIONS(10406), 1,
      sym_word,
    ACTIONS(10412), 1,
      sym__comment_word,
    ACTIONS(10654), 1,
      anon_sym_DOLLAR,
    ACTIONS(189), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(213), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10408), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10410), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(513), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [223413] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3737), 1,
      aux_sym_number_token1,
    ACTIONS(3739), 1,
      aux_sym_number_token2,
    ACTIONS(3743), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3755), 1,
      sym__brace_start,
    ACTIONS(8820), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8824), 1,
      anon_sym_DQUOTE,
    ACTIONS(8828), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8830), 1,
      anon_sym_BQUOTE,
    ACTIONS(8832), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9412), 1,
      sym_word,
    ACTIONS(9420), 1,
      sym__comment_word,
    ACTIONS(10656), 1,
      anon_sym_DOLLAR,
    ACTIONS(8818), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8834), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9416), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9418), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(4810), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [223481] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3793), 1,
      aux_sym_number_token1,
    ACTIONS(3795), 1,
      aux_sym_number_token2,
    ACTIONS(3799), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3809), 1,
      sym__brace_start,
    ACTIONS(9636), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9640), 1,
      anon_sym_DQUOTE,
    ACTIONS(9644), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9646), 1,
      anon_sym_BQUOTE,
    ACTIONS(9648), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9872), 1,
      sym_word,
    ACTIONS(9880), 1,
      sym__comment_word,
    ACTIONS(10658), 1,
      anon_sym_DOLLAR,
    ACTIONS(9634), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9650), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9876), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9878), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(4337), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [223549] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9364), 1,
      anon_sym_LPAREN,
    ACTIONS(9366), 1,
      anon_sym_BANG,
    ACTIONS(9372), 1,
      anon_sym_TILDE,
    ACTIONS(9374), 1,
      anon_sym_DOLLAR,
    ACTIONS(9376), 1,
      anon_sym_DQUOTE,
    ACTIONS(9378), 1,
      aux_sym_number_token1,
    ACTIONS(9380), 1,
      aux_sym_number_token2,
    ACTIONS(9382), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9384), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9386), 1,
      anon_sym_BQUOTE,
    ACTIONS(9388), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9392), 1,
      sym_variable_name,
    ACTIONS(10660), 1,
      aux_sym__simple_variable_name_token1,
    STATE(3391), 1,
      sym__arithmetic_binary_expression,
    STATE(3396), 1,
      sym__arithmetic_ternary_expression,
    STATE(3398), 1,
      sym__arithmetic_unary_expression,
    STATE(3400), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9368), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9370), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(3308), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [223623] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3064), 1,
      sym_variable_name,
    ACTIONS(9278), 1,
      anon_sym_LPAREN,
    ACTIONS(9280), 1,
      anon_sym_BANG,
    ACTIONS(9286), 1,
      anon_sym_TILDE,
    ACTIONS(9288), 1,
      anon_sym_DOLLAR,
    ACTIONS(9290), 1,
      anon_sym_DQUOTE,
    ACTIONS(9292), 1,
      aux_sym_number_token1,
    ACTIONS(9294), 1,
      aux_sym_number_token2,
    ACTIONS(9296), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9298), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9300), 1,
      anon_sym_BQUOTE,
    ACTIONS(9302), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10662), 1,
      aux_sym__simple_variable_name_token1,
    STATE(2801), 1,
      sym__arithmetic_binary_expression,
    STATE(2803), 1,
      sym__arithmetic_ternary_expression,
    STATE(2809), 1,
      sym__arithmetic_unary_expression,
    STATE(2825), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9282), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9284), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(2768), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [223697] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3793), 1,
      aux_sym_number_token1,
    ACTIONS(3795), 1,
      aux_sym_number_token2,
    ACTIONS(3799), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3809), 1,
      sym__brace_start,
    ACTIONS(9636), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9640), 1,
      anon_sym_DQUOTE,
    ACTIONS(9644), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9646), 1,
      anon_sym_BQUOTE,
    ACTIONS(9648), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9872), 1,
      sym_word,
    ACTIONS(9880), 1,
      sym__comment_word,
    ACTIONS(10664), 1,
      anon_sym_DOLLAR,
    ACTIONS(9634), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9650), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9876), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9878), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(4337), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [223765] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9364), 1,
      anon_sym_LPAREN,
    ACTIONS(9366), 1,
      anon_sym_BANG,
    ACTIONS(9372), 1,
      anon_sym_TILDE,
    ACTIONS(9374), 1,
      anon_sym_DOLLAR,
    ACTIONS(9376), 1,
      anon_sym_DQUOTE,
    ACTIONS(9378), 1,
      aux_sym_number_token1,
    ACTIONS(9380), 1,
      aux_sym_number_token2,
    ACTIONS(9382), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9384), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9386), 1,
      anon_sym_BQUOTE,
    ACTIONS(9388), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9392), 1,
      sym_variable_name,
    ACTIONS(10666), 1,
      aux_sym__simple_variable_name_token1,
    STATE(3391), 1,
      sym__arithmetic_binary_expression,
    STATE(3396), 1,
      sym__arithmetic_ternary_expression,
    STATE(3398), 1,
      sym__arithmetic_unary_expression,
    STATE(3400), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9368), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9370), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(3304), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [223839] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4086), 1,
      aux_sym_number_token1,
    ACTIONS(4088), 1,
      aux_sym_number_token2,
    ACTIONS(4092), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(4102), 1,
      sym__brace_start,
    ACTIONS(9306), 1,
      sym_word,
    ACTIONS(9310), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9316), 1,
      anon_sym_DQUOTE,
    ACTIONS(9320), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9322), 1,
      anon_sym_BQUOTE,
    ACTIONS(9324), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9328), 1,
      sym__comment_word,
    ACTIONS(10668), 1,
      anon_sym_DOLLAR,
    ACTIONS(9308), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9314), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9326), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9318), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(4566), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [223907] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4086), 1,
      aux_sym_number_token1,
    ACTIONS(4088), 1,
      aux_sym_number_token2,
    ACTIONS(4092), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(4102), 1,
      sym__brace_start,
    ACTIONS(9306), 1,
      sym_word,
    ACTIONS(9310), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9316), 1,
      anon_sym_DQUOTE,
    ACTIONS(9320), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9322), 1,
      anon_sym_BQUOTE,
    ACTIONS(9324), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9328), 1,
      sym__comment_word,
    ACTIONS(10670), 1,
      anon_sym_DOLLAR,
    ACTIONS(9308), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9314), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9326), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9318), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(4566), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [223975] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3064), 1,
      sym_variable_name,
    ACTIONS(9278), 1,
      anon_sym_LPAREN,
    ACTIONS(9280), 1,
      anon_sym_BANG,
    ACTIONS(9286), 1,
      anon_sym_TILDE,
    ACTIONS(9288), 1,
      anon_sym_DOLLAR,
    ACTIONS(9290), 1,
      anon_sym_DQUOTE,
    ACTIONS(9292), 1,
      aux_sym_number_token1,
    ACTIONS(9294), 1,
      aux_sym_number_token2,
    ACTIONS(9296), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9298), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9300), 1,
      anon_sym_BQUOTE,
    ACTIONS(9302), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10672), 1,
      aux_sym__simple_variable_name_token1,
    STATE(2801), 1,
      sym__arithmetic_binary_expression,
    STATE(2803), 1,
      sym__arithmetic_ternary_expression,
    STATE(2809), 1,
      sym__arithmetic_unary_expression,
    STATE(2825), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9282), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9284), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(2769), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [224049] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3064), 1,
      sym_variable_name,
    ACTIONS(9278), 1,
      anon_sym_LPAREN,
    ACTIONS(9280), 1,
      anon_sym_BANG,
    ACTIONS(9286), 1,
      anon_sym_TILDE,
    ACTIONS(9288), 1,
      anon_sym_DOLLAR,
    ACTIONS(9290), 1,
      anon_sym_DQUOTE,
    ACTIONS(9292), 1,
      aux_sym_number_token1,
    ACTIONS(9294), 1,
      aux_sym_number_token2,
    ACTIONS(9296), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9298), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9300), 1,
      anon_sym_BQUOTE,
    ACTIONS(9302), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10674), 1,
      aux_sym__simple_variable_name_token1,
    STATE(2801), 1,
      sym__arithmetic_binary_expression,
    STATE(2803), 1,
      sym__arithmetic_ternary_expression,
    STATE(2809), 1,
      sym__arithmetic_unary_expression,
    STATE(2825), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9282), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9284), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(2938), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [224123] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1557), 1,
      aux_sym_number_token1,
    ACTIONS(1559), 1,
      aux_sym_number_token2,
    ACTIONS(1563), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1575), 1,
      sym__brace_start,
    ACTIONS(9198), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9202), 1,
      anon_sym_DQUOTE,
    ACTIONS(9206), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9208), 1,
      anon_sym_BQUOTE,
    ACTIONS(9210), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9268), 1,
      sym_word,
    ACTIONS(9276), 1,
      sym__comment_word,
    ACTIONS(10676), 1,
      anon_sym_DOLLAR,
    ACTIONS(9196), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9212), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9272), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9274), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1189), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [224191] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3191), 1,
      aux_sym_number_token1,
    ACTIONS(3193), 1,
      aux_sym_number_token2,
    ACTIONS(3197), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3209), 1,
      sym__brace_start,
    ACTIONS(9674), 1,
      sym_word,
    ACTIONS(9678), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9682), 1,
      anon_sym_DQUOTE,
    ACTIONS(9686), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9688), 1,
      anon_sym_BQUOTE,
    ACTIONS(9690), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9694), 1,
      sym__comment_word,
    ACTIONS(10678), 1,
      anon_sym_DOLLAR,
    ACTIONS(9676), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9680), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9692), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9684), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1733), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [224259] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3191), 1,
      aux_sym_number_token1,
    ACTIONS(3193), 1,
      aux_sym_number_token2,
    ACTIONS(3197), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3209), 1,
      sym__brace_start,
    ACTIONS(9674), 1,
      sym_word,
    ACTIONS(9678), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9682), 1,
      anon_sym_DQUOTE,
    ACTIONS(9686), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9688), 1,
      anon_sym_BQUOTE,
    ACTIONS(9690), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9694), 1,
      sym__comment_word,
    ACTIONS(10680), 1,
      anon_sym_DOLLAR,
    ACTIONS(9676), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9680), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9692), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9684), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1733), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [224327] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9364), 1,
      anon_sym_LPAREN,
    ACTIONS(9366), 1,
      anon_sym_BANG,
    ACTIONS(9372), 1,
      anon_sym_TILDE,
    ACTIONS(9374), 1,
      anon_sym_DOLLAR,
    ACTIONS(9376), 1,
      anon_sym_DQUOTE,
    ACTIONS(9378), 1,
      aux_sym_number_token1,
    ACTIONS(9380), 1,
      aux_sym_number_token2,
    ACTIONS(9382), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9384), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9386), 1,
      anon_sym_BQUOTE,
    ACTIONS(9388), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9392), 1,
      sym_variable_name,
    ACTIONS(10682), 1,
      aux_sym__simple_variable_name_token1,
    STATE(3391), 1,
      sym__arithmetic_binary_expression,
    STATE(3396), 1,
      sym__arithmetic_ternary_expression,
    STATE(3398), 1,
      sym__arithmetic_unary_expression,
    STATE(3400), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9368), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9370), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(3303), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [224401] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3064), 1,
      sym_variable_name,
    ACTIONS(9278), 1,
      anon_sym_LPAREN,
    ACTIONS(9280), 1,
      anon_sym_BANG,
    ACTIONS(9286), 1,
      anon_sym_TILDE,
    ACTIONS(9288), 1,
      anon_sym_DOLLAR,
    ACTIONS(9290), 1,
      anon_sym_DQUOTE,
    ACTIONS(9292), 1,
      aux_sym_number_token1,
    ACTIONS(9294), 1,
      aux_sym_number_token2,
    ACTIONS(9296), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9298), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9300), 1,
      anon_sym_BQUOTE,
    ACTIONS(9302), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10684), 1,
      aux_sym__simple_variable_name_token1,
    STATE(2801), 1,
      sym__arithmetic_binary_expression,
    STATE(2803), 1,
      sym__arithmetic_ternary_expression,
    STATE(2809), 1,
      sym__arithmetic_unary_expression,
    STATE(2825), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9282), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9284), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(2770), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [224475] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2269), 1,
      aux_sym_number_token1,
    ACTIONS(2271), 1,
      aux_sym_number_token2,
    ACTIONS(2275), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2289), 1,
      sym__brace_start,
    ACTIONS(8066), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8070), 1,
      anon_sym_DQUOTE,
    ACTIONS(8074), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8076), 1,
      anon_sym_BQUOTE,
    ACTIONS(8078), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9580), 1,
      sym_word,
    ACTIONS(9588), 1,
      sym__comment_word,
    ACTIONS(10686), 1,
      anon_sym_DOLLAR,
    ACTIONS(8062), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8080), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9584), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9586), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1623), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [224543] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9364), 1,
      anon_sym_LPAREN,
    ACTIONS(9366), 1,
      anon_sym_BANG,
    ACTIONS(9372), 1,
      anon_sym_TILDE,
    ACTIONS(9374), 1,
      anon_sym_DOLLAR,
    ACTIONS(9376), 1,
      anon_sym_DQUOTE,
    ACTIONS(9378), 1,
      aux_sym_number_token1,
    ACTIONS(9380), 1,
      aux_sym_number_token2,
    ACTIONS(9382), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9384), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9386), 1,
      anon_sym_BQUOTE,
    ACTIONS(9388), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9392), 1,
      sym_variable_name,
    ACTIONS(10688), 1,
      aux_sym__simple_variable_name_token1,
    STATE(3391), 1,
      sym__arithmetic_binary_expression,
    STATE(3396), 1,
      sym__arithmetic_ternary_expression,
    STATE(3398), 1,
      sym__arithmetic_unary_expression,
    STATE(3400), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9368), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9370), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(3302), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [224617] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2269), 1,
      aux_sym_number_token1,
    ACTIONS(2271), 1,
      aux_sym_number_token2,
    ACTIONS(2275), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2289), 1,
      sym__brace_start,
    ACTIONS(8066), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8070), 1,
      anon_sym_DQUOTE,
    ACTIONS(8074), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8076), 1,
      anon_sym_BQUOTE,
    ACTIONS(8078), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9580), 1,
      sym_word,
    ACTIONS(9588), 1,
      sym__comment_word,
    ACTIONS(10690), 1,
      anon_sym_DOLLAR,
    ACTIONS(8062), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8080), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9584), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9586), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1623), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [224685] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8178), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8184), 1,
      anon_sym_DQUOTE,
    ACTIONS(8188), 1,
      aux_sym_number_token1,
    ACTIONS(8190), 1,
      aux_sym_number_token2,
    ACTIONS(8192), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8194), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(8196), 1,
      anon_sym_BQUOTE,
    ACTIONS(8198), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8208), 1,
      sym__brace_start,
    ACTIONS(9720), 1,
      sym_word,
    ACTIONS(9728), 1,
      sym__comment_word,
    ACTIONS(10692), 1,
      anon_sym_DOLLAR,
    ACTIONS(8174), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8200), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9724), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9726), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1590), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [224753] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8178), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8184), 1,
      anon_sym_DQUOTE,
    ACTIONS(8188), 1,
      aux_sym_number_token1,
    ACTIONS(8190), 1,
      aux_sym_number_token2,
    ACTIONS(8192), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8194), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(8196), 1,
      anon_sym_BQUOTE,
    ACTIONS(8198), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8208), 1,
      sym__brace_start,
    ACTIONS(9720), 1,
      sym_word,
    ACTIONS(9728), 1,
      sym__comment_word,
    ACTIONS(10694), 1,
      anon_sym_DOLLAR,
    ACTIONS(8174), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8200), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9724), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9726), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1590), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [224821] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3064), 1,
      sym_variable_name,
    ACTIONS(9278), 1,
      anon_sym_LPAREN,
    ACTIONS(9280), 1,
      anon_sym_BANG,
    ACTIONS(9286), 1,
      anon_sym_TILDE,
    ACTIONS(9288), 1,
      anon_sym_DOLLAR,
    ACTIONS(9290), 1,
      anon_sym_DQUOTE,
    ACTIONS(9292), 1,
      aux_sym_number_token1,
    ACTIONS(9294), 1,
      aux_sym_number_token2,
    ACTIONS(9296), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9298), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9300), 1,
      anon_sym_BQUOTE,
    ACTIONS(9302), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10696), 1,
      aux_sym__simple_variable_name_token1,
    STATE(2801), 1,
      sym__arithmetic_binary_expression,
    STATE(2803), 1,
      sym__arithmetic_ternary_expression,
    STATE(2809), 1,
      sym__arithmetic_unary_expression,
    STATE(2825), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9282), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9284), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(2771), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [224895] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8372), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8378), 1,
      anon_sym_DQUOTE,
    ACTIONS(8382), 1,
      aux_sym_number_token1,
    ACTIONS(8384), 1,
      aux_sym_number_token2,
    ACTIONS(8386), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8388), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(8390), 1,
      anon_sym_BQUOTE,
    ACTIONS(8392), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8402), 1,
      sym__brace_start,
    ACTIONS(10472), 1,
      sym_word,
    ACTIONS(10478), 1,
      sym__comment_word,
    ACTIONS(10698), 1,
      anon_sym_DOLLAR,
    ACTIONS(8368), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8394), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10474), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10476), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(4285), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [224963] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(363), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(371), 1,
      anon_sym_DQUOTE,
    ACTIONS(375), 1,
      aux_sym_number_token1,
    ACTIONS(377), 1,
      aux_sym_number_token2,
    ACTIONS(379), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(381), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(385), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(391), 1,
      sym__brace_start,
    ACTIONS(5809), 1,
      anon_sym_BQUOTE,
    ACTIONS(8836), 1,
      sym_word,
    ACTIONS(8840), 1,
      sym__comment_word,
    ACTIONS(10700), 1,
      anon_sym_DOLLAR,
    ACTIONS(352), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(387), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(8844), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(8842), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2730), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [225031] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(363), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(371), 1,
      anon_sym_DQUOTE,
    ACTIONS(375), 1,
      aux_sym_number_token1,
    ACTIONS(377), 1,
      aux_sym_number_token2,
    ACTIONS(379), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(381), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(385), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(391), 1,
      sym__brace_start,
    ACTIONS(5809), 1,
      anon_sym_BQUOTE,
    ACTIONS(8836), 1,
      sym_word,
    ACTIONS(8840), 1,
      sym__comment_word,
    ACTIONS(10702), 1,
      anon_sym_DOLLAR,
    ACTIONS(352), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(387), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(8844), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(8842), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2730), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [225099] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(363), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(367), 1,
      anon_sym_DOLLAR,
    ACTIONS(371), 1,
      anon_sym_DQUOTE,
    ACTIONS(375), 1,
      aux_sym_number_token1,
    ACTIONS(377), 1,
      aux_sym_number_token2,
    ACTIONS(379), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(381), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(385), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(391), 1,
      sym__brace_start,
    ACTIONS(5809), 1,
      anon_sym_BQUOTE,
    ACTIONS(8836), 1,
      sym_word,
    ACTIONS(8840), 1,
      sym__comment_word,
    ACTIONS(352), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(387), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(8844), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(8842), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2730), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [225167] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8372), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8378), 1,
      anon_sym_DQUOTE,
    ACTIONS(8382), 1,
      aux_sym_number_token1,
    ACTIONS(8384), 1,
      aux_sym_number_token2,
    ACTIONS(8386), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8388), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(8390), 1,
      anon_sym_BQUOTE,
    ACTIONS(8392), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8402), 1,
      sym__brace_start,
    ACTIONS(10472), 1,
      sym_word,
    ACTIONS(10478), 1,
      sym__comment_word,
    ACTIONS(10704), 1,
      anon_sym_DOLLAR,
    ACTIONS(8368), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8394), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10474), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10476), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(4285), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [225235] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9758), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9760), 1,
      anon_sym_DOLLAR,
    ACTIONS(9764), 1,
      anon_sym_DQUOTE,
    ACTIONS(9768), 1,
      aux_sym_number_token1,
    ACTIONS(9770), 1,
      aux_sym_number_token2,
    ACTIONS(9772), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9774), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9776), 1,
      anon_sym_BQUOTE,
    ACTIONS(9778), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9784), 1,
      sym__brace_start,
    ACTIONS(10052), 1,
      sym__special_character,
    ACTIONS(10706), 1,
      sym_word,
    ACTIONS(10710), 1,
      sym_test_operator,
    STATE(6213), 1,
      aux_sym__literal_repeat1,
    STATE(6413), 1,
      sym_concatenation,
    ACTIONS(9756), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9780), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10708), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(6199), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [225307] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(314), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(320), 1,
      anon_sym_DQUOTE,
    ACTIONS(324), 1,
      aux_sym_number_token1,
    ACTIONS(326), 1,
      aux_sym_number_token2,
    ACTIONS(328), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(330), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(332), 1,
      anon_sym_BQUOTE,
    ACTIONS(334), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(344), 1,
      sym__brace_start,
    ACTIONS(9400), 1,
      sym_word,
    ACTIONS(9406), 1,
      sym__comment_word,
    ACTIONS(10712), 1,
      anon_sym_DOLLAR,
    ACTIONS(312), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(336), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9402), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9404), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1037), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [225375] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(314), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(320), 1,
      anon_sym_DQUOTE,
    ACTIONS(324), 1,
      aux_sym_number_token1,
    ACTIONS(326), 1,
      aux_sym_number_token2,
    ACTIONS(328), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(330), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(332), 1,
      anon_sym_BQUOTE,
    ACTIONS(334), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(344), 1,
      sym__brace_start,
    ACTIONS(9400), 1,
      sym_word,
    ACTIONS(9406), 1,
      sym__comment_word,
    ACTIONS(10714), 1,
      anon_sym_DOLLAR,
    ACTIONS(312), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(336), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9402), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9404), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1037), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [225443] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9364), 1,
      anon_sym_LPAREN,
    ACTIONS(9366), 1,
      anon_sym_BANG,
    ACTIONS(9372), 1,
      anon_sym_TILDE,
    ACTIONS(9374), 1,
      anon_sym_DOLLAR,
    ACTIONS(9376), 1,
      anon_sym_DQUOTE,
    ACTIONS(9378), 1,
      aux_sym_number_token1,
    ACTIONS(9380), 1,
      aux_sym_number_token2,
    ACTIONS(9382), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9384), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9386), 1,
      anon_sym_BQUOTE,
    ACTIONS(9388), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9392), 1,
      sym_variable_name,
    ACTIONS(10716), 1,
      aux_sym__simple_variable_name_token1,
    STATE(3391), 1,
      sym__arithmetic_binary_expression,
    STATE(3396), 1,
      sym__arithmetic_ternary_expression,
    STATE(3398), 1,
      sym__arithmetic_unary_expression,
    STATE(3400), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9368), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9370), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(3301), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [225517] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1141), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1149), 1,
      anon_sym_DQUOTE,
    ACTIONS(1153), 1,
      aux_sym_number_token1,
    ACTIONS(1155), 1,
      aux_sym_number_token2,
    ACTIONS(1157), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1159), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1163), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1169), 1,
      sym__brace_start,
    ACTIONS(3060), 1,
      anon_sym_BQUOTE,
    ACTIONS(10188), 1,
      sym_word,
    ACTIONS(10196), 1,
      sym__comment_word,
    ACTIONS(10718), 1,
      anon_sym_DOLLAR,
    ACTIONS(1129), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1165), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10192), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10194), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2392), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [225585] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8474), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8480), 1,
      anon_sym_DQUOTE,
    ACTIONS(8484), 1,
      aux_sym_number_token1,
    ACTIONS(8486), 1,
      aux_sym_number_token2,
    ACTIONS(8488), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8490), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(8492), 1,
      anon_sym_BQUOTE,
    ACTIONS(8494), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8504), 1,
      sym__brace_start,
    ACTIONS(10720), 1,
      sym_word,
    ACTIONS(10722), 1,
      anon_sym_DOLLAR,
    ACTIONS(10728), 1,
      sym__comment_word,
    ACTIONS(8470), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8496), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10724), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10726), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1625), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [225653] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(8474), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8476), 1,
      anon_sym_DOLLAR,
    ACTIONS(8478), 1,
      sym__special_character,
    ACTIONS(8480), 1,
      anon_sym_DQUOTE,
    ACTIONS(8484), 1,
      aux_sym_number_token1,
    ACTIONS(8486), 1,
      aux_sym_number_token2,
    ACTIONS(8488), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8490), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(8492), 1,
      anon_sym_BQUOTE,
    ACTIONS(8494), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8504), 1,
      sym__brace_start,
    ACTIONS(10730), 1,
      sym_word,
    ACTIONS(10734), 1,
      sym_test_operator,
    STATE(1898), 1,
      aux_sym__literal_repeat1,
    STATE(2107), 1,
      sym_concatenation,
    ACTIONS(8470), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8496), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10732), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1395), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [225725] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3064), 1,
      sym_variable_name,
    ACTIONS(9278), 1,
      anon_sym_LPAREN,
    ACTIONS(9280), 1,
      anon_sym_BANG,
    ACTIONS(9286), 1,
      anon_sym_TILDE,
    ACTIONS(9288), 1,
      anon_sym_DOLLAR,
    ACTIONS(9290), 1,
      anon_sym_DQUOTE,
    ACTIONS(9292), 1,
      aux_sym_number_token1,
    ACTIONS(9294), 1,
      aux_sym_number_token2,
    ACTIONS(9296), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9298), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9300), 1,
      anon_sym_BQUOTE,
    ACTIONS(9302), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10736), 1,
      aux_sym__simple_variable_name_token1,
    STATE(2801), 1,
      sym__arithmetic_binary_expression,
    STATE(2803), 1,
      sym__arithmetic_ternary_expression,
    STATE(2809), 1,
      sym__arithmetic_unary_expression,
    STATE(2825), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9282), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9284), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(2772), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [225799] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8512), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8518), 1,
      anon_sym_DQUOTE,
    ACTIONS(8522), 1,
      aux_sym_number_token1,
    ACTIONS(8524), 1,
      aux_sym_number_token2,
    ACTIONS(8526), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8528), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(8530), 1,
      anon_sym_BQUOTE,
    ACTIONS(8532), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8542), 1,
      sym__brace_start,
    ACTIONS(10220), 1,
      sym_word,
    ACTIONS(10226), 1,
      sym__comment_word,
    ACTIONS(10738), 1,
      anon_sym_DOLLAR,
    ACTIONS(8508), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8534), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10222), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10224), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2514), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [225867] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(47), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(49), 1,
      anon_sym_DOLLAR,
    ACTIONS(53), 1,
      anon_sym_DQUOTE,
    ACTIONS(57), 1,
      aux_sym_number_token1,
    ACTIONS(59), 1,
      aux_sym_number_token2,
    ACTIONS(61), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(63), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(65), 1,
      anon_sym_BQUOTE,
    ACTIONS(67), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(79), 1,
      sym__brace_start,
    ACTIONS(10448), 1,
      sym_word,
    ACTIONS(10456), 1,
      sym__comment_word,
    ACTIONS(45), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(69), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10452), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10454), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1083), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [225935] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1141), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1145), 1,
      anon_sym_DOLLAR,
    ACTIONS(1149), 1,
      anon_sym_DQUOTE,
    ACTIONS(1153), 1,
      aux_sym_number_token1,
    ACTIONS(1155), 1,
      aux_sym_number_token2,
    ACTIONS(1157), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1159), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1163), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1169), 1,
      sym__brace_start,
    ACTIONS(3060), 1,
      anon_sym_BQUOTE,
    ACTIONS(10188), 1,
      sym_word,
    ACTIONS(10196), 1,
      sym__comment_word,
    ACTIONS(1129), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1165), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10192), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10194), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2392), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [226003] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1141), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1149), 1,
      anon_sym_DQUOTE,
    ACTIONS(1153), 1,
      aux_sym_number_token1,
    ACTIONS(1155), 1,
      aux_sym_number_token2,
    ACTIONS(1157), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1159), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1163), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1169), 1,
      sym__brace_start,
    ACTIONS(3060), 1,
      anon_sym_BQUOTE,
    ACTIONS(10188), 1,
      sym_word,
    ACTIONS(10196), 1,
      sym__comment_word,
    ACTIONS(10740), 1,
      anon_sym_DOLLAR,
    ACTIONS(1129), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1165), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10192), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10194), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2392), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [226071] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(8178), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8180), 1,
      anon_sym_DOLLAR,
    ACTIONS(8184), 1,
      anon_sym_DQUOTE,
    ACTIONS(8188), 1,
      aux_sym_number_token1,
    ACTIONS(8190), 1,
      aux_sym_number_token2,
    ACTIONS(8192), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8194), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(8196), 1,
      anon_sym_BQUOTE,
    ACTIONS(8198), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8208), 1,
      sym__brace_start,
    ACTIONS(8450), 1,
      sym__special_character,
    ACTIONS(10742), 1,
      sym_word,
    ACTIONS(10746), 1,
      sym_test_operator,
    STATE(1738), 1,
      aux_sym__literal_repeat1,
    STATE(1952), 1,
      sym_concatenation,
    ACTIONS(8174), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8200), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10744), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1596), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [226143] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(8178), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8180), 1,
      anon_sym_DOLLAR,
    ACTIONS(8184), 1,
      anon_sym_DQUOTE,
    ACTIONS(8188), 1,
      aux_sym_number_token1,
    ACTIONS(8190), 1,
      aux_sym_number_token2,
    ACTIONS(8192), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8194), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(8196), 1,
      anon_sym_BQUOTE,
    ACTIONS(8198), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8208), 1,
      sym__brace_start,
    ACTIONS(8450), 1,
      sym__special_character,
    ACTIONS(10748), 1,
      sym_word,
    ACTIONS(10752), 1,
      sym_test_operator,
    STATE(1741), 1,
      aux_sym__literal_repeat1,
    STATE(1958), 1,
      sym_concatenation,
    ACTIONS(8174), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8200), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10750), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1529), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [226215] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3602), 1,
      sym_variable_name,
    ACTIONS(9898), 1,
      anon_sym_LPAREN,
    ACTIONS(9900), 1,
      anon_sym_BANG,
    ACTIONS(9906), 1,
      anon_sym_TILDE,
    ACTIONS(9908), 1,
      anon_sym_DOLLAR,
    ACTIONS(9910), 1,
      anon_sym_DQUOTE,
    ACTIONS(9912), 1,
      aux_sym_number_token1,
    ACTIONS(9914), 1,
      aux_sym_number_token2,
    ACTIONS(9916), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9918), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9920), 1,
      anon_sym_BQUOTE,
    ACTIONS(9922), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10754), 1,
      aux_sym__simple_variable_name_token1,
    STATE(3359), 1,
      sym__arithmetic_postfix_expression,
    STATE(3361), 1,
      sym__arithmetic_unary_expression,
    STATE(3367), 1,
      sym__arithmetic_ternary_expression,
    STATE(3375), 1,
      sym__arithmetic_binary_expression,
    ACTIONS(9902), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9904), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(3242), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [226289] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3602), 1,
      sym_variable_name,
    ACTIONS(9898), 1,
      anon_sym_LPAREN,
    ACTIONS(9900), 1,
      anon_sym_BANG,
    ACTIONS(9906), 1,
      anon_sym_TILDE,
    ACTIONS(9908), 1,
      anon_sym_DOLLAR,
    ACTIONS(9910), 1,
      anon_sym_DQUOTE,
    ACTIONS(9912), 1,
      aux_sym_number_token1,
    ACTIONS(9914), 1,
      aux_sym_number_token2,
    ACTIONS(9916), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9918), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9920), 1,
      anon_sym_BQUOTE,
    ACTIONS(9922), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10756), 1,
      aux_sym__simple_variable_name_token1,
    STATE(3359), 1,
      sym__arithmetic_postfix_expression,
    STATE(3361), 1,
      sym__arithmetic_unary_expression,
    STATE(3367), 1,
      sym__arithmetic_ternary_expression,
    STATE(3375), 1,
      sym__arithmetic_binary_expression,
    ACTIONS(9902), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9904), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(3246), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [226363] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1587), 1,
      anon_sym_DOLLAR,
    ACTIONS(1593), 1,
      aux_sym_number_token1,
    ACTIONS(1595), 1,
      aux_sym_number_token2,
    ACTIONS(1599), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1611), 1,
      sym__brace_start,
    ACTIONS(9240), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9244), 1,
      anon_sym_DQUOTE,
    ACTIONS(9248), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9250), 1,
      anon_sym_BQUOTE,
    ACTIONS(9252), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10094), 1,
      sym_word,
    ACTIONS(10102), 1,
      sym__comment_word,
    ACTIONS(9238), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9254), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10098), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10100), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1293), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [226431] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3602), 1,
      sym_variable_name,
    ACTIONS(9898), 1,
      anon_sym_LPAREN,
    ACTIONS(9900), 1,
      anon_sym_BANG,
    ACTIONS(9906), 1,
      anon_sym_TILDE,
    ACTIONS(9908), 1,
      anon_sym_DOLLAR,
    ACTIONS(9910), 1,
      anon_sym_DQUOTE,
    ACTIONS(9912), 1,
      aux_sym_number_token1,
    ACTIONS(9914), 1,
      aux_sym_number_token2,
    ACTIONS(9916), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9918), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9920), 1,
      anon_sym_BQUOTE,
    ACTIONS(9922), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10758), 1,
      aux_sym__simple_variable_name_token1,
    STATE(3359), 1,
      sym__arithmetic_postfix_expression,
    STATE(3361), 1,
      sym__arithmetic_unary_expression,
    STATE(3367), 1,
      sym__arithmetic_ternary_expression,
    STATE(3375), 1,
      sym__arithmetic_binary_expression,
    ACTIONS(9902), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9904), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(3247), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [226505] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(4080), 1,
      anon_sym_DOLLAR,
    ACTIONS(4086), 1,
      aux_sym_number_token1,
    ACTIONS(4088), 1,
      aux_sym_number_token2,
    ACTIONS(4092), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(4102), 1,
      sym__brace_start,
    ACTIONS(9310), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9316), 1,
      anon_sym_DQUOTE,
    ACTIONS(9320), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9322), 1,
      anon_sym_BQUOTE,
    ACTIONS(9324), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10760), 1,
      sym_word,
    ACTIONS(10762), 1,
      sym__special_character,
    ACTIONS(10766), 1,
      sym_test_operator,
    STATE(4500), 1,
      aux_sym__literal_repeat1,
    STATE(5029), 1,
      sym_concatenation,
    ACTIONS(9308), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9326), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10764), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(4687), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [226577] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9364), 1,
      anon_sym_LPAREN,
    ACTIONS(9366), 1,
      anon_sym_BANG,
    ACTIONS(9372), 1,
      anon_sym_TILDE,
    ACTIONS(9374), 1,
      anon_sym_DOLLAR,
    ACTIONS(9376), 1,
      anon_sym_DQUOTE,
    ACTIONS(9378), 1,
      aux_sym_number_token1,
    ACTIONS(9380), 1,
      aux_sym_number_token2,
    ACTIONS(9382), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9384), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9386), 1,
      anon_sym_BQUOTE,
    ACTIONS(9388), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9392), 1,
      sym_variable_name,
    ACTIONS(10768), 1,
      aux_sym__simple_variable_name_token1,
    STATE(3391), 1,
      sym__arithmetic_binary_expression,
    STATE(3396), 1,
      sym__arithmetic_ternary_expression,
    STATE(3398), 1,
      sym__arithmetic_unary_expression,
    STATE(3400), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9368), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9370), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(3299), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [226651] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3191), 1,
      aux_sym_number_token1,
    ACTIONS(3193), 1,
      aux_sym_number_token2,
    ACTIONS(3197), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3209), 1,
      sym__brace_start,
    ACTIONS(9674), 1,
      sym_word,
    ACTIONS(9678), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9682), 1,
      anon_sym_DQUOTE,
    ACTIONS(9686), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9688), 1,
      anon_sym_BQUOTE,
    ACTIONS(9690), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9694), 1,
      sym__comment_word,
    ACTIONS(10770), 1,
      anon_sym_DOLLAR,
    ACTIONS(9676), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9680), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9692), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9684), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1733), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [226719] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(4080), 1,
      anon_sym_DOLLAR,
    ACTIONS(4086), 1,
      aux_sym_number_token1,
    ACTIONS(4088), 1,
      aux_sym_number_token2,
    ACTIONS(4092), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(4102), 1,
      sym__brace_start,
    ACTIONS(9310), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9316), 1,
      anon_sym_DQUOTE,
    ACTIONS(9320), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9322), 1,
      anon_sym_BQUOTE,
    ACTIONS(9324), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10762), 1,
      sym__special_character,
    ACTIONS(10772), 1,
      sym_word,
    ACTIONS(10776), 1,
      sym_test_operator,
    STATE(4421), 1,
      aux_sym__literal_repeat1,
    STATE(4869), 1,
      sym_concatenation,
    ACTIONS(9308), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9326), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10774), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(4699), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [226791] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3602), 1,
      sym_variable_name,
    ACTIONS(9898), 1,
      anon_sym_LPAREN,
    ACTIONS(9900), 1,
      anon_sym_BANG,
    ACTIONS(9906), 1,
      anon_sym_TILDE,
    ACTIONS(9908), 1,
      anon_sym_DOLLAR,
    ACTIONS(9910), 1,
      anon_sym_DQUOTE,
    ACTIONS(9912), 1,
      aux_sym_number_token1,
    ACTIONS(9914), 1,
      aux_sym_number_token2,
    ACTIONS(9916), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9918), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9920), 1,
      anon_sym_BQUOTE,
    ACTIONS(9922), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10778), 1,
      aux_sym__simple_variable_name_token1,
    STATE(3359), 1,
      sym__arithmetic_postfix_expression,
    STATE(3361), 1,
      sym__arithmetic_unary_expression,
    STATE(3367), 1,
      sym__arithmetic_ternary_expression,
    STATE(3375), 1,
      sym__arithmetic_binary_expression,
    ACTIONS(9902), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9904), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(3248), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [226865] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3602), 1,
      sym_variable_name,
    ACTIONS(9898), 1,
      anon_sym_LPAREN,
    ACTIONS(9900), 1,
      anon_sym_BANG,
    ACTIONS(9906), 1,
      anon_sym_TILDE,
    ACTIONS(9908), 1,
      anon_sym_DOLLAR,
    ACTIONS(9910), 1,
      anon_sym_DQUOTE,
    ACTIONS(9912), 1,
      aux_sym_number_token1,
    ACTIONS(9914), 1,
      aux_sym_number_token2,
    ACTIONS(9916), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9918), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9920), 1,
      anon_sym_BQUOTE,
    ACTIONS(9922), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10780), 1,
      aux_sym__simple_variable_name_token1,
    STATE(3359), 1,
      sym__arithmetic_postfix_expression,
    STATE(3361), 1,
      sym__arithmetic_unary_expression,
    STATE(3367), 1,
      sym__arithmetic_ternary_expression,
    STATE(3375), 1,
      sym__arithmetic_binary_expression,
    ACTIONS(9902), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9904), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(3249), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [226939] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8268), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8274), 1,
      anon_sym_DQUOTE,
    ACTIONS(8278), 1,
      aux_sym_number_token1,
    ACTIONS(8280), 1,
      aux_sym_number_token2,
    ACTIONS(8282), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8284), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(8286), 1,
      anon_sym_BQUOTE,
    ACTIONS(8288), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8298), 1,
      sym__brace_start,
    ACTIONS(10480), 1,
      sym_word,
    ACTIONS(10486), 1,
      sym__comment_word,
    ACTIONS(10782), 1,
      anon_sym_DOLLAR,
    ACTIONS(8264), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8290), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10482), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10484), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1465), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [227007] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3602), 1,
      sym_variable_name,
    ACTIONS(9898), 1,
      anon_sym_LPAREN,
    ACTIONS(9900), 1,
      anon_sym_BANG,
    ACTIONS(9906), 1,
      anon_sym_TILDE,
    ACTIONS(9908), 1,
      anon_sym_DOLLAR,
    ACTIONS(9910), 1,
      anon_sym_DQUOTE,
    ACTIONS(9912), 1,
      aux_sym_number_token1,
    ACTIONS(9914), 1,
      aux_sym_number_token2,
    ACTIONS(9916), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9918), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9920), 1,
      anon_sym_BQUOTE,
    ACTIONS(9922), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10784), 1,
      aux_sym__simple_variable_name_token1,
    STATE(3359), 1,
      sym__arithmetic_postfix_expression,
    STATE(3361), 1,
      sym__arithmetic_unary_expression,
    STATE(3367), 1,
      sym__arithmetic_ternary_expression,
    STATE(3375), 1,
      sym__arithmetic_binary_expression,
    ACTIONS(9902), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9904), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(3258), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [227081] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3602), 1,
      sym_variable_name,
    ACTIONS(9898), 1,
      anon_sym_LPAREN,
    ACTIONS(9900), 1,
      anon_sym_BANG,
    ACTIONS(9906), 1,
      anon_sym_TILDE,
    ACTIONS(9908), 1,
      anon_sym_DOLLAR,
    ACTIONS(9910), 1,
      anon_sym_DQUOTE,
    ACTIONS(9912), 1,
      aux_sym_number_token1,
    ACTIONS(9914), 1,
      aux_sym_number_token2,
    ACTIONS(9916), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9918), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9920), 1,
      anon_sym_BQUOTE,
    ACTIONS(9922), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10786), 1,
      aux_sym__simple_variable_name_token1,
    STATE(3359), 1,
      sym__arithmetic_postfix_expression,
    STATE(3361), 1,
      sym__arithmetic_unary_expression,
    STATE(3367), 1,
      sym__arithmetic_ternary_expression,
    STATE(3375), 1,
      sym__arithmetic_binary_expression,
    ACTIONS(9902), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9904), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(3259), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [227155] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3787), 1,
      anon_sym_DOLLAR,
    ACTIONS(3793), 1,
      aux_sym_number_token1,
    ACTIONS(3795), 1,
      aux_sym_number_token2,
    ACTIONS(3799), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3809), 1,
      sym__brace_start,
    ACTIONS(9636), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9640), 1,
      anon_sym_DQUOTE,
    ACTIONS(9644), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9646), 1,
      anon_sym_BQUOTE,
    ACTIONS(9648), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10788), 1,
      sym_word,
    ACTIONS(10790), 1,
      sym__special_character,
    ACTIONS(10794), 1,
      sym_test_operator,
    STATE(4294), 1,
      aux_sym__literal_repeat1,
    STATE(4861), 1,
      sym_concatenation,
    ACTIONS(9634), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9650), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10792), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(4440), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [227227] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8268), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8274), 1,
      anon_sym_DQUOTE,
    ACTIONS(8278), 1,
      aux_sym_number_token1,
    ACTIONS(8280), 1,
      aux_sym_number_token2,
    ACTIONS(8282), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8284), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(8286), 1,
      anon_sym_BQUOTE,
    ACTIONS(8288), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8298), 1,
      sym__brace_start,
    ACTIONS(10480), 1,
      sym_word,
    ACTIONS(10486), 1,
      sym__comment_word,
    ACTIONS(10796), 1,
      anon_sym_DOLLAR,
    ACTIONS(8264), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8290), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10482), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10484), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1465), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [227295] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3064), 1,
      sym_variable_name,
    ACTIONS(9278), 1,
      anon_sym_LPAREN,
    ACTIONS(9280), 1,
      anon_sym_BANG,
    ACTIONS(9286), 1,
      anon_sym_TILDE,
    ACTIONS(9288), 1,
      anon_sym_DOLLAR,
    ACTIONS(9290), 1,
      anon_sym_DQUOTE,
    ACTIONS(9292), 1,
      aux_sym_number_token1,
    ACTIONS(9294), 1,
      aux_sym_number_token2,
    ACTIONS(9296), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9298), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9300), 1,
      anon_sym_BQUOTE,
    ACTIONS(9302), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10798), 1,
      aux_sym__simple_variable_name_token1,
    STATE(2801), 1,
      sym__arithmetic_binary_expression,
    STATE(2803), 1,
      sym__arithmetic_ternary_expression,
    STATE(2809), 1,
      sym__arithmetic_unary_expression,
    STATE(2825), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9282), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9284), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(2773), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [227369] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9364), 1,
      anon_sym_LPAREN,
    ACTIONS(9366), 1,
      anon_sym_BANG,
    ACTIONS(9372), 1,
      anon_sym_TILDE,
    ACTIONS(9374), 1,
      anon_sym_DOLLAR,
    ACTIONS(9376), 1,
      anon_sym_DQUOTE,
    ACTIONS(9378), 1,
      aux_sym_number_token1,
    ACTIONS(9380), 1,
      aux_sym_number_token2,
    ACTIONS(9382), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9384), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9386), 1,
      anon_sym_BQUOTE,
    ACTIONS(9388), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9392), 1,
      sym_variable_name,
    ACTIONS(10800), 1,
      aux_sym__simple_variable_name_token1,
    STATE(3391), 1,
      sym__arithmetic_binary_expression,
    STATE(3396), 1,
      sym__arithmetic_ternary_expression,
    STATE(3398), 1,
      sym__arithmetic_unary_expression,
    STATE(3400), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9368), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9370), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(3296), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [227443] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3064), 1,
      sym_variable_name,
    ACTIONS(9278), 1,
      anon_sym_LPAREN,
    ACTIONS(9280), 1,
      anon_sym_BANG,
    ACTIONS(9286), 1,
      anon_sym_TILDE,
    ACTIONS(9288), 1,
      anon_sym_DOLLAR,
    ACTIONS(9290), 1,
      anon_sym_DQUOTE,
    ACTIONS(9292), 1,
      aux_sym_number_token1,
    ACTIONS(9294), 1,
      aux_sym_number_token2,
    ACTIONS(9296), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9298), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9300), 1,
      anon_sym_BQUOTE,
    ACTIONS(9302), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10802), 1,
      aux_sym__simple_variable_name_token1,
    STATE(2801), 1,
      sym__arithmetic_binary_expression,
    STATE(2803), 1,
      sym__arithmetic_ternary_expression,
    STATE(2809), 1,
      sym__arithmetic_unary_expression,
    STATE(2825), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9282), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9284), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(2774), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [227517] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3787), 1,
      anon_sym_DOLLAR,
    ACTIONS(3793), 1,
      aux_sym_number_token1,
    ACTIONS(3795), 1,
      aux_sym_number_token2,
    ACTIONS(3799), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3809), 1,
      sym__brace_start,
    ACTIONS(9636), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9640), 1,
      anon_sym_DQUOTE,
    ACTIONS(9644), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9646), 1,
      anon_sym_BQUOTE,
    ACTIONS(9648), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10790), 1,
      sym__special_character,
    ACTIONS(10804), 1,
      sym_word,
    ACTIONS(10808), 1,
      sym_test_operator,
    STATE(4347), 1,
      aux_sym__literal_repeat1,
    STATE(4617), 1,
      sym_concatenation,
    ACTIONS(9634), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9650), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10806), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(4452), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [227589] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3602), 1,
      sym_variable_name,
    ACTIONS(9898), 1,
      anon_sym_LPAREN,
    ACTIONS(9900), 1,
      anon_sym_BANG,
    ACTIONS(9906), 1,
      anon_sym_TILDE,
    ACTIONS(9908), 1,
      anon_sym_DOLLAR,
    ACTIONS(9910), 1,
      anon_sym_DQUOTE,
    ACTIONS(9912), 1,
      aux_sym_number_token1,
    ACTIONS(9914), 1,
      aux_sym_number_token2,
    ACTIONS(9916), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9918), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9920), 1,
      anon_sym_BQUOTE,
    ACTIONS(9922), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10810), 1,
      aux_sym__simple_variable_name_token1,
    STATE(3359), 1,
      sym__arithmetic_postfix_expression,
    STATE(3361), 1,
      sym__arithmetic_unary_expression,
    STATE(3367), 1,
      sym__arithmetic_ternary_expression,
    STATE(3375), 1,
      sym__arithmetic_binary_expression,
    ACTIONS(9902), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9904), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(3261), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [227663] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1551), 1,
      anon_sym_DOLLAR,
    ACTIONS(1557), 1,
      aux_sym_number_token1,
    ACTIONS(1559), 1,
      aux_sym_number_token2,
    ACTIONS(1563), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1575), 1,
      sym__brace_start,
    ACTIONS(9198), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9200), 1,
      sym__special_character,
    ACTIONS(9202), 1,
      anon_sym_DQUOTE,
    ACTIONS(9206), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9208), 1,
      anon_sym_BQUOTE,
    ACTIONS(9210), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10812), 1,
      sym_word,
    ACTIONS(10816), 1,
      sym_test_operator,
    STATE(1436), 1,
      aux_sym__literal_repeat1,
    STATE(1543), 1,
      sym_concatenation,
    ACTIONS(9196), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9212), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10814), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1273), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [227735] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1551), 1,
      anon_sym_DOLLAR,
    ACTIONS(1557), 1,
      aux_sym_number_token1,
    ACTIONS(1559), 1,
      aux_sym_number_token2,
    ACTIONS(1563), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1575), 1,
      sym__brace_start,
    ACTIONS(9198), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9200), 1,
      sym__special_character,
    ACTIONS(9202), 1,
      anon_sym_DQUOTE,
    ACTIONS(9206), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9208), 1,
      anon_sym_BQUOTE,
    ACTIONS(9210), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10818), 1,
      sym_word,
    ACTIONS(10822), 1,
      sym_test_operator,
    STATE(1401), 1,
      aux_sym__literal_repeat1,
    STATE(1573), 1,
      sym_concatenation,
    ACTIONS(9196), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9212), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10820), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1276), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [227807] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9364), 1,
      anon_sym_LPAREN,
    ACTIONS(9366), 1,
      anon_sym_BANG,
    ACTIONS(9372), 1,
      anon_sym_TILDE,
    ACTIONS(9374), 1,
      anon_sym_DOLLAR,
    ACTIONS(9376), 1,
      anon_sym_DQUOTE,
    ACTIONS(9378), 1,
      aux_sym_number_token1,
    ACTIONS(9380), 1,
      aux_sym_number_token2,
    ACTIONS(9382), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9384), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9386), 1,
      anon_sym_BQUOTE,
    ACTIONS(9388), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9392), 1,
      sym_variable_name,
    ACTIONS(10824), 1,
      aux_sym__simple_variable_name_token1,
    STATE(3391), 1,
      sym__arithmetic_binary_expression,
    STATE(3396), 1,
      sym__arithmetic_ternary_expression,
    STATE(3398), 1,
      sym__arithmetic_unary_expression,
    STATE(3400), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9368), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9370), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(3293), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [227881] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4633), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(4639), 1,
      anon_sym_DQUOTE,
    ACTIONS(4643), 1,
      aux_sym_number_token1,
    ACTIONS(4645), 1,
      aux_sym_number_token2,
    ACTIONS(4647), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(4649), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(4651), 1,
      anon_sym_BQUOTE,
    ACTIONS(4653), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(4659), 1,
      sym__brace_start,
    ACTIONS(9330), 1,
      sym_word,
    ACTIONS(9338), 1,
      sym__comment_word,
    ACTIONS(10826), 1,
      anon_sym_DOLLAR,
    ACTIONS(4631), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(4655), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9334), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9336), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2634), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [227949] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3602), 1,
      sym_variable_name,
    ACTIONS(9898), 1,
      anon_sym_LPAREN,
    ACTIONS(9900), 1,
      anon_sym_BANG,
    ACTIONS(9906), 1,
      anon_sym_TILDE,
    ACTIONS(9908), 1,
      anon_sym_DOLLAR,
    ACTIONS(9910), 1,
      anon_sym_DQUOTE,
    ACTIONS(9912), 1,
      aux_sym_number_token1,
    ACTIONS(9914), 1,
      aux_sym_number_token2,
    ACTIONS(9916), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9918), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9920), 1,
      anon_sym_BQUOTE,
    ACTIONS(9922), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10828), 1,
      aux_sym__simple_variable_name_token1,
    STATE(3359), 1,
      sym__arithmetic_postfix_expression,
    STATE(3361), 1,
      sym__arithmetic_unary_expression,
    STATE(3367), 1,
      sym__arithmetic_ternary_expression,
    STATE(3375), 1,
      sym__arithmetic_binary_expression,
    ACTIONS(9902), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9904), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(3263), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [228023] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3602), 1,
      sym_variable_name,
    ACTIONS(9898), 1,
      anon_sym_LPAREN,
    ACTIONS(9900), 1,
      anon_sym_BANG,
    ACTIONS(9906), 1,
      anon_sym_TILDE,
    ACTIONS(9908), 1,
      anon_sym_DOLLAR,
    ACTIONS(9910), 1,
      anon_sym_DQUOTE,
    ACTIONS(9912), 1,
      aux_sym_number_token1,
    ACTIONS(9914), 1,
      aux_sym_number_token2,
    ACTIONS(9916), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9918), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9920), 1,
      anon_sym_BQUOTE,
    ACTIONS(9922), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10830), 1,
      aux_sym__simple_variable_name_token1,
    STATE(3359), 1,
      sym__arithmetic_postfix_expression,
    STATE(3361), 1,
      sym__arithmetic_unary_expression,
    STATE(3367), 1,
      sym__arithmetic_ternary_expression,
    STATE(3375), 1,
      sym__arithmetic_binary_expression,
    ACTIONS(9902), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9904), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(3264), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [228097] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3602), 1,
      sym_variable_name,
    ACTIONS(9898), 1,
      anon_sym_LPAREN,
    ACTIONS(9900), 1,
      anon_sym_BANG,
    ACTIONS(9906), 1,
      anon_sym_TILDE,
    ACTIONS(9908), 1,
      anon_sym_DOLLAR,
    ACTIONS(9910), 1,
      anon_sym_DQUOTE,
    ACTIONS(9912), 1,
      aux_sym_number_token1,
    ACTIONS(9914), 1,
      aux_sym_number_token2,
    ACTIONS(9916), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9918), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9920), 1,
      anon_sym_BQUOTE,
    ACTIONS(9922), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10832), 1,
      aux_sym__simple_variable_name_token1,
    STATE(3359), 1,
      sym__arithmetic_postfix_expression,
    STATE(3361), 1,
      sym__arithmetic_unary_expression,
    STATE(3367), 1,
      sym__arithmetic_ternary_expression,
    STATE(3375), 1,
      sym__arithmetic_binary_expression,
    ACTIONS(9902), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9904), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(3265), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [228171] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5315), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(5321), 1,
      anon_sym_DQUOTE,
    ACTIONS(5325), 1,
      aux_sym_number_token1,
    ACTIONS(5327), 1,
      aux_sym_number_token2,
    ACTIONS(5329), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(5331), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(5333), 1,
      anon_sym_BQUOTE,
    ACTIONS(5335), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(5341), 1,
      sym__brace_start,
    ACTIONS(10236), 1,
      sym_word,
    ACTIONS(10242), 1,
      sym__comment_word,
    ACTIONS(10834), 1,
      anon_sym_DOLLAR,
    ACTIONS(5313), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(5337), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10238), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10240), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(5461), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [228239] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3602), 1,
      sym_variable_name,
    ACTIONS(9898), 1,
      anon_sym_LPAREN,
    ACTIONS(9900), 1,
      anon_sym_BANG,
    ACTIONS(9906), 1,
      anon_sym_TILDE,
    ACTIONS(9908), 1,
      anon_sym_DOLLAR,
    ACTIONS(9910), 1,
      anon_sym_DQUOTE,
    ACTIONS(9912), 1,
      aux_sym_number_token1,
    ACTIONS(9914), 1,
      aux_sym_number_token2,
    ACTIONS(9916), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9918), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9920), 1,
      anon_sym_BQUOTE,
    ACTIONS(9922), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10836), 1,
      aux_sym__simple_variable_name_token1,
    STATE(3359), 1,
      sym__arithmetic_postfix_expression,
    STATE(3361), 1,
      sym__arithmetic_unary_expression,
    STATE(3367), 1,
      sym__arithmetic_ternary_expression,
    STATE(3375), 1,
      sym__arithmetic_binary_expression,
    ACTIONS(9902), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9904), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(3266), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [228313] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3064), 1,
      sym_variable_name,
    ACTIONS(9278), 1,
      anon_sym_LPAREN,
    ACTIONS(9280), 1,
      anon_sym_BANG,
    ACTIONS(9286), 1,
      anon_sym_TILDE,
    ACTIONS(9288), 1,
      anon_sym_DOLLAR,
    ACTIONS(9290), 1,
      anon_sym_DQUOTE,
    ACTIONS(9292), 1,
      aux_sym_number_token1,
    ACTIONS(9294), 1,
      aux_sym_number_token2,
    ACTIONS(9296), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9298), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9300), 1,
      anon_sym_BQUOTE,
    ACTIONS(9302), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10838), 1,
      aux_sym__simple_variable_name_token1,
    STATE(2801), 1,
      sym__arithmetic_binary_expression,
    STATE(2803), 1,
      sym__arithmetic_ternary_expression,
    STATE(2809), 1,
      sym__arithmetic_unary_expression,
    STATE(2825), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9282), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9284), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(2775), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [228387] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3064), 1,
      sym_variable_name,
    ACTIONS(9278), 1,
      anon_sym_LPAREN,
    ACTIONS(9280), 1,
      anon_sym_BANG,
    ACTIONS(9286), 1,
      anon_sym_TILDE,
    ACTIONS(9288), 1,
      anon_sym_DOLLAR,
    ACTIONS(9290), 1,
      anon_sym_DQUOTE,
    ACTIONS(9292), 1,
      aux_sym_number_token1,
    ACTIONS(9294), 1,
      aux_sym_number_token2,
    ACTIONS(9296), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9298), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9300), 1,
      anon_sym_BQUOTE,
    ACTIONS(9302), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10840), 1,
      aux_sym__simple_variable_name_token1,
    STATE(2801), 1,
      sym__arithmetic_binary_expression,
    STATE(2803), 1,
      sym__arithmetic_ternary_expression,
    STATE(2809), 1,
      sym__arithmetic_unary_expression,
    STATE(2825), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9282), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9284), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(2858), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [228461] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9364), 1,
      anon_sym_LPAREN,
    ACTIONS(9366), 1,
      anon_sym_BANG,
    ACTIONS(9372), 1,
      anon_sym_TILDE,
    ACTIONS(9374), 1,
      anon_sym_DOLLAR,
    ACTIONS(9376), 1,
      anon_sym_DQUOTE,
    ACTIONS(9378), 1,
      aux_sym_number_token1,
    ACTIONS(9380), 1,
      aux_sym_number_token2,
    ACTIONS(9382), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9384), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9386), 1,
      anon_sym_BQUOTE,
    ACTIONS(9388), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9392), 1,
      sym_variable_name,
    ACTIONS(10842), 1,
      aux_sym__simple_variable_name_token1,
    STATE(3391), 1,
      sym__arithmetic_binary_expression,
    STATE(3396), 1,
      sym__arithmetic_ternary_expression,
    STATE(3398), 1,
      sym__arithmetic_unary_expression,
    STATE(3400), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9368), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9370), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(3373), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [228535] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3064), 1,
      sym_variable_name,
    ACTIONS(9278), 1,
      anon_sym_LPAREN,
    ACTIONS(9280), 1,
      anon_sym_BANG,
    ACTIONS(9286), 1,
      anon_sym_TILDE,
    ACTIONS(9288), 1,
      anon_sym_DOLLAR,
    ACTIONS(9290), 1,
      anon_sym_DQUOTE,
    ACTIONS(9292), 1,
      aux_sym_number_token1,
    ACTIONS(9294), 1,
      aux_sym_number_token2,
    ACTIONS(9296), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9298), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9300), 1,
      anon_sym_BQUOTE,
    ACTIONS(9302), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10844), 1,
      aux_sym__simple_variable_name_token1,
    STATE(2801), 1,
      sym__arithmetic_binary_expression,
    STATE(2803), 1,
      sym__arithmetic_ternary_expression,
    STATE(2809), 1,
      sym__arithmetic_unary_expression,
    STATE(2825), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9282), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9284), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(2937), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [228609] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2038), 1,
      anon_sym_DOLLAR,
    ACTIONS(2044), 1,
      aux_sym_number_token1,
    ACTIONS(2046), 1,
      aux_sym_number_token2,
    ACTIONS(2050), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2064), 1,
      sym__brace_start,
    ACTIONS(8344), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8348), 1,
      anon_sym_DQUOTE,
    ACTIONS(8352), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8354), 1,
      anon_sym_BQUOTE,
    ACTIONS(8356), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10846), 1,
      sym_word,
    ACTIONS(10852), 1,
      sym__comment_word,
    ACTIONS(8340), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8358), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10848), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10850), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1493), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [228677] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3602), 1,
      sym_variable_name,
    ACTIONS(3811), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(9898), 1,
      anon_sym_LPAREN,
    ACTIONS(9900), 1,
      anon_sym_BANG,
    ACTIONS(9906), 1,
      anon_sym_TILDE,
    ACTIONS(9908), 1,
      anon_sym_DOLLAR,
    ACTIONS(9910), 1,
      anon_sym_DQUOTE,
    ACTIONS(9912), 1,
      aux_sym_number_token1,
    ACTIONS(9914), 1,
      aux_sym_number_token2,
    ACTIONS(9916), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9918), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9920), 1,
      anon_sym_BQUOTE,
    ACTIONS(9922), 1,
      anon_sym_DOLLAR_BQUOTE,
    STATE(3359), 1,
      sym__arithmetic_postfix_expression,
    STATE(3361), 1,
      sym__arithmetic_unary_expression,
    STATE(3367), 1,
      sym__arithmetic_ternary_expression,
    STATE(3375), 1,
      sym__arithmetic_binary_expression,
    ACTIONS(9902), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9904), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(3365), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [228751] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3064), 1,
      sym_variable_name,
    ACTIONS(9278), 1,
      anon_sym_LPAREN,
    ACTIONS(9280), 1,
      anon_sym_BANG,
    ACTIONS(9286), 1,
      anon_sym_TILDE,
    ACTIONS(9288), 1,
      anon_sym_DOLLAR,
    ACTIONS(9290), 1,
      anon_sym_DQUOTE,
    ACTIONS(9292), 1,
      aux_sym_number_token1,
    ACTIONS(9294), 1,
      aux_sym_number_token2,
    ACTIONS(9296), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9298), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9300), 1,
      anon_sym_BQUOTE,
    ACTIONS(9302), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10854), 1,
      aux_sym__simple_variable_name_token1,
    STATE(2801), 1,
      sym__arithmetic_binary_expression,
    STATE(2803), 1,
      sym__arithmetic_ternary_expression,
    STATE(2809), 1,
      sym__arithmetic_unary_expression,
    STATE(2825), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9282), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9284), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(3350), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [228825] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3064), 1,
      sym_variable_name,
    ACTIONS(9278), 1,
      anon_sym_LPAREN,
    ACTIONS(9280), 1,
      anon_sym_BANG,
    ACTIONS(9286), 1,
      anon_sym_TILDE,
    ACTIONS(9288), 1,
      anon_sym_DOLLAR,
    ACTIONS(9290), 1,
      anon_sym_DQUOTE,
    ACTIONS(9292), 1,
      aux_sym_number_token1,
    ACTIONS(9294), 1,
      aux_sym_number_token2,
    ACTIONS(9296), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9298), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9300), 1,
      anon_sym_BQUOTE,
    ACTIONS(9302), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10856), 1,
      aux_sym__simple_variable_name_token1,
    STATE(2801), 1,
      sym__arithmetic_binary_expression,
    STATE(2803), 1,
      sym__arithmetic_ternary_expression,
    STATE(2809), 1,
      sym__arithmetic_unary_expression,
    STATE(2825), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9282), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9284), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(2936), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [228899] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9364), 1,
      anon_sym_LPAREN,
    ACTIONS(9366), 1,
      anon_sym_BANG,
    ACTIONS(9372), 1,
      anon_sym_TILDE,
    ACTIONS(9374), 1,
      anon_sym_DOLLAR,
    ACTIONS(9376), 1,
      anon_sym_DQUOTE,
    ACTIONS(9378), 1,
      aux_sym_number_token1,
    ACTIONS(9380), 1,
      aux_sym_number_token2,
    ACTIONS(9382), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9384), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9386), 1,
      anon_sym_BQUOTE,
    ACTIONS(9388), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9392), 1,
      sym_variable_name,
    ACTIONS(10858), 1,
      aux_sym__simple_variable_name_token1,
    STATE(3391), 1,
      sym__arithmetic_binary_expression,
    STATE(3396), 1,
      sym__arithmetic_ternary_expression,
    STATE(3398), 1,
      sym__arithmetic_unary_expression,
    STATE(3400), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9368), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9370), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(3446), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [228973] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3064), 1,
      sym_variable_name,
    ACTIONS(9278), 1,
      anon_sym_LPAREN,
    ACTIONS(9280), 1,
      anon_sym_BANG,
    ACTIONS(9286), 1,
      anon_sym_TILDE,
    ACTIONS(9288), 1,
      anon_sym_DOLLAR,
    ACTIONS(9290), 1,
      anon_sym_DQUOTE,
    ACTIONS(9292), 1,
      aux_sym_number_token1,
    ACTIONS(9294), 1,
      aux_sym_number_token2,
    ACTIONS(9296), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9298), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9300), 1,
      anon_sym_BQUOTE,
    ACTIONS(9302), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10860), 1,
      aux_sym__simple_variable_name_token1,
    STATE(2801), 1,
      sym__arithmetic_binary_expression,
    STATE(2803), 1,
      sym__arithmetic_ternary_expression,
    STATE(2809), 1,
      sym__arithmetic_unary_expression,
    STATE(2825), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9282), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9284), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(2923), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [229047] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1837), 1,
      aux_sym_number_token1,
    ACTIONS(1839), 1,
      aux_sym_number_token2,
    ACTIONS(1843), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1857), 1,
      sym__brace_start,
    ACTIONS(8232), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8236), 1,
      anon_sym_DQUOTE,
    ACTIONS(8240), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8242), 1,
      anon_sym_BQUOTE,
    ACTIONS(8244), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10488), 1,
      sym_word,
    ACTIONS(10494), 1,
      sym__comment_word,
    ACTIONS(10862), 1,
      anon_sym_DOLLAR,
    ACTIONS(8228), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8246), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10490), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10492), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1440), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [229115] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3064), 1,
      sym_variable_name,
    ACTIONS(9278), 1,
      anon_sym_LPAREN,
    ACTIONS(9280), 1,
      anon_sym_BANG,
    ACTIONS(9286), 1,
      anon_sym_TILDE,
    ACTIONS(9288), 1,
      anon_sym_DOLLAR,
    ACTIONS(9290), 1,
      anon_sym_DQUOTE,
    ACTIONS(9292), 1,
      aux_sym_number_token1,
    ACTIONS(9294), 1,
      aux_sym_number_token2,
    ACTIONS(9296), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9298), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9300), 1,
      anon_sym_BQUOTE,
    ACTIONS(9302), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10864), 1,
      aux_sym__simple_variable_name_token1,
    STATE(2801), 1,
      sym__arithmetic_binary_expression,
    STATE(2803), 1,
      sym__arithmetic_ternary_expression,
    STATE(2809), 1,
      sym__arithmetic_unary_expression,
    STATE(2825), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9282), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9284), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(2922), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [229189] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3064), 1,
      sym_variable_name,
    ACTIONS(9278), 1,
      anon_sym_LPAREN,
    ACTIONS(9280), 1,
      anon_sym_BANG,
    ACTIONS(9286), 1,
      anon_sym_TILDE,
    ACTIONS(9288), 1,
      anon_sym_DOLLAR,
    ACTIONS(9290), 1,
      anon_sym_DQUOTE,
    ACTIONS(9292), 1,
      aux_sym_number_token1,
    ACTIONS(9294), 1,
      aux_sym_number_token2,
    ACTIONS(9296), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9298), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9300), 1,
      anon_sym_BQUOTE,
    ACTIONS(9302), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10866), 1,
      aux_sym__simple_variable_name_token1,
    STATE(2801), 1,
      sym__arithmetic_binary_expression,
    STATE(2803), 1,
      sym__arithmetic_ternary_expression,
    STATE(2809), 1,
      sym__arithmetic_unary_expression,
    STATE(2825), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9282), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9284), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(2921), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [229263] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9364), 1,
      anon_sym_LPAREN,
    ACTIONS(9366), 1,
      anon_sym_BANG,
    ACTIONS(9372), 1,
      anon_sym_TILDE,
    ACTIONS(9374), 1,
      anon_sym_DOLLAR,
    ACTIONS(9376), 1,
      anon_sym_DQUOTE,
    ACTIONS(9378), 1,
      aux_sym_number_token1,
    ACTIONS(9380), 1,
      aux_sym_number_token2,
    ACTIONS(9382), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9384), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9386), 1,
      anon_sym_BQUOTE,
    ACTIONS(9388), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9392), 1,
      sym_variable_name,
    ACTIONS(10868), 1,
      aux_sym__simple_variable_name_token1,
    STATE(3391), 1,
      sym__arithmetic_binary_expression,
    STATE(3396), 1,
      sym__arithmetic_ternary_expression,
    STATE(3398), 1,
      sym__arithmetic_unary_expression,
    STATE(3400), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9368), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9370), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(3289), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [229337] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1837), 1,
      aux_sym_number_token1,
    ACTIONS(1839), 1,
      aux_sym_number_token2,
    ACTIONS(1843), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1857), 1,
      sym__brace_start,
    ACTIONS(8232), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8236), 1,
      anon_sym_DQUOTE,
    ACTIONS(8240), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8242), 1,
      anon_sym_BQUOTE,
    ACTIONS(8244), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10488), 1,
      sym_word,
    ACTIONS(10494), 1,
      sym__comment_word,
    ACTIONS(10870), 1,
      anon_sym_DOLLAR,
    ACTIONS(8228), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8246), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10490), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10492), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1440), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [229405] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3064), 1,
      sym_variable_name,
    ACTIONS(3467), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(9278), 1,
      anon_sym_LPAREN,
    ACTIONS(9280), 1,
      anon_sym_BANG,
    ACTIONS(9286), 1,
      anon_sym_TILDE,
    ACTIONS(9288), 1,
      anon_sym_DOLLAR,
    ACTIONS(9290), 1,
      anon_sym_DQUOTE,
    ACTIONS(9292), 1,
      aux_sym_number_token1,
    ACTIONS(9294), 1,
      aux_sym_number_token2,
    ACTIONS(9296), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9298), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9300), 1,
      anon_sym_BQUOTE,
    ACTIONS(9302), 1,
      anon_sym_DOLLAR_BQUOTE,
    STATE(2801), 1,
      sym__arithmetic_binary_expression,
    STATE(2803), 1,
      sym__arithmetic_ternary_expression,
    STATE(2809), 1,
      sym__arithmetic_unary_expression,
    STATE(2825), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9282), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9284), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(2777), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [229479] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5315), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(5321), 1,
      anon_sym_DQUOTE,
    ACTIONS(5325), 1,
      aux_sym_number_token1,
    ACTIONS(5327), 1,
      aux_sym_number_token2,
    ACTIONS(5329), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(5331), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(5333), 1,
      anon_sym_BQUOTE,
    ACTIONS(5335), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(5341), 1,
      sym__brace_start,
    ACTIONS(10236), 1,
      sym_word,
    ACTIONS(10242), 1,
      sym__comment_word,
    ACTIONS(10872), 1,
      anon_sym_DOLLAR,
    ACTIONS(5313), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(5337), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10238), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10240), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(5461), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [229547] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9758), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9760), 1,
      anon_sym_DOLLAR,
    ACTIONS(9764), 1,
      anon_sym_DQUOTE,
    ACTIONS(9768), 1,
      aux_sym_number_token1,
    ACTIONS(9770), 1,
      aux_sym_number_token2,
    ACTIONS(9772), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9774), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9776), 1,
      anon_sym_BQUOTE,
    ACTIONS(9778), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9784), 1,
      sym__brace_start,
    ACTIONS(10052), 1,
      sym__special_character,
    ACTIONS(10874), 1,
      sym_word,
    ACTIONS(10878), 1,
      sym_test_operator,
    STATE(6223), 1,
      aux_sym__literal_repeat1,
    STATE(6374), 1,
      sym_concatenation,
    ACTIONS(9756), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9780), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10876), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(6202), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [229619] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9364), 1,
      anon_sym_LPAREN,
    ACTIONS(9366), 1,
      anon_sym_BANG,
    ACTIONS(9372), 1,
      anon_sym_TILDE,
    ACTIONS(9374), 1,
      anon_sym_DOLLAR,
    ACTIONS(9376), 1,
      anon_sym_DQUOTE,
    ACTIONS(9378), 1,
      aux_sym_number_token1,
    ACTIONS(9380), 1,
      aux_sym_number_token2,
    ACTIONS(9382), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9384), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9386), 1,
      anon_sym_BQUOTE,
    ACTIONS(9388), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9392), 1,
      sym_variable_name,
    ACTIONS(10880), 1,
      aux_sym__simple_variable_name_token1,
    STATE(3391), 1,
      sym__arithmetic_binary_expression,
    STATE(3396), 1,
      sym__arithmetic_ternary_expression,
    STATE(3398), 1,
      sym__arithmetic_unary_expression,
    STATE(3400), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9368), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9370), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(3284), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [229693] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(314), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(320), 1,
      anon_sym_DQUOTE,
    ACTIONS(324), 1,
      aux_sym_number_token1,
    ACTIONS(326), 1,
      aux_sym_number_token2,
    ACTIONS(328), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(330), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(332), 1,
      anon_sym_BQUOTE,
    ACTIONS(334), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(344), 1,
      sym__brace_start,
    ACTIONS(9400), 1,
      sym_word,
    ACTIONS(9406), 1,
      sym__comment_word,
    ACTIONS(10882), 1,
      anon_sym_DOLLAR,
    ACTIONS(312), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(336), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9402), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9404), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1037), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [229761] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8571), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8577), 1,
      anon_sym_DQUOTE,
    ACTIONS(8581), 1,
      aux_sym_number_token1,
    ACTIONS(8583), 1,
      aux_sym_number_token2,
    ACTIONS(8585), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8587), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(8589), 1,
      anon_sym_BQUOTE,
    ACTIONS(8591), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8597), 1,
      sym__brace_start,
    ACTIONS(10014), 1,
      sym_word,
    ACTIONS(10020), 1,
      sym__comment_word,
    ACTIONS(10884), 1,
      anon_sym_DOLLAR,
    ACTIONS(8567), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8593), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10016), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10018), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(5199), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [229829] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3064), 1,
      sym_variable_name,
    ACTIONS(9278), 1,
      anon_sym_LPAREN,
    ACTIONS(9280), 1,
      anon_sym_BANG,
    ACTIONS(9286), 1,
      anon_sym_TILDE,
    ACTIONS(9288), 1,
      anon_sym_DOLLAR,
    ACTIONS(9290), 1,
      anon_sym_DQUOTE,
    ACTIONS(9292), 1,
      aux_sym_number_token1,
    ACTIONS(9294), 1,
      aux_sym_number_token2,
    ACTIONS(9296), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9298), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9300), 1,
      anon_sym_BQUOTE,
    ACTIONS(9302), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10886), 1,
      aux_sym__simple_variable_name_token1,
    STATE(2801), 1,
      sym__arithmetic_binary_expression,
    STATE(2803), 1,
      sym__arithmetic_ternary_expression,
    STATE(2809), 1,
      sym__arithmetic_unary_expression,
    STATE(2825), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9282), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9284), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(2778), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [229903] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9364), 1,
      anon_sym_LPAREN,
    ACTIONS(9366), 1,
      anon_sym_BANG,
    ACTIONS(9372), 1,
      anon_sym_TILDE,
    ACTIONS(9374), 1,
      anon_sym_DOLLAR,
    ACTIONS(9376), 1,
      anon_sym_DQUOTE,
    ACTIONS(9378), 1,
      aux_sym_number_token1,
    ACTIONS(9380), 1,
      aux_sym_number_token2,
    ACTIONS(9382), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9384), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9386), 1,
      anon_sym_BQUOTE,
    ACTIONS(9388), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9392), 1,
      sym_variable_name,
    ACTIONS(10888), 1,
      aux_sym__simple_variable_name_token1,
    STATE(3391), 1,
      sym__arithmetic_binary_expression,
    STATE(3396), 1,
      sym__arithmetic_ternary_expression,
    STATE(3398), 1,
      sym__arithmetic_unary_expression,
    STATE(3400), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9368), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9370), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(3447), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [229977] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3064), 1,
      sym_variable_name,
    ACTIONS(9278), 1,
      anon_sym_LPAREN,
    ACTIONS(9280), 1,
      anon_sym_BANG,
    ACTIONS(9286), 1,
      anon_sym_TILDE,
    ACTIONS(9288), 1,
      anon_sym_DOLLAR,
    ACTIONS(9290), 1,
      anon_sym_DQUOTE,
    ACTIONS(9292), 1,
      aux_sym_number_token1,
    ACTIONS(9294), 1,
      aux_sym_number_token2,
    ACTIONS(9296), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9298), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9300), 1,
      anon_sym_BQUOTE,
    ACTIONS(9302), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10890), 1,
      aux_sym__simple_variable_name_token1,
    STATE(2801), 1,
      sym__arithmetic_binary_expression,
    STATE(2803), 1,
      sym__arithmetic_ternary_expression,
    STATE(2809), 1,
      sym__arithmetic_unary_expression,
    STATE(2825), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9282), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9284), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(2917), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [230051] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9364), 1,
      anon_sym_LPAREN,
    ACTIONS(9366), 1,
      anon_sym_BANG,
    ACTIONS(9372), 1,
      anon_sym_TILDE,
    ACTIONS(9374), 1,
      anon_sym_DOLLAR,
    ACTIONS(9376), 1,
      anon_sym_DQUOTE,
    ACTIONS(9378), 1,
      aux_sym_number_token1,
    ACTIONS(9380), 1,
      aux_sym_number_token2,
    ACTIONS(9382), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9384), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9386), 1,
      anon_sym_BQUOTE,
    ACTIONS(9388), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9392), 1,
      sym_variable_name,
    ACTIONS(10892), 1,
      aux_sym__simple_variable_name_token1,
    STATE(3391), 1,
      sym__arithmetic_binary_expression,
    STATE(3396), 1,
      sym__arithmetic_ternary_expression,
    STATE(3398), 1,
      sym__arithmetic_unary_expression,
    STATE(3400), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9368), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9370), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(3449), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [230125] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8474), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8480), 1,
      anon_sym_DQUOTE,
    ACTIONS(8484), 1,
      aux_sym_number_token1,
    ACTIONS(8486), 1,
      aux_sym_number_token2,
    ACTIONS(8488), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8490), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(8492), 1,
      anon_sym_BQUOTE,
    ACTIONS(8494), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8504), 1,
      sym__brace_start,
    ACTIONS(10720), 1,
      sym_word,
    ACTIONS(10728), 1,
      sym__comment_word,
    ACTIONS(10894), 1,
      anon_sym_DOLLAR,
    ACTIONS(8470), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8496), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10724), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10726), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1625), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [230193] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9754), 1,
      sym_word,
    ACTIONS(9758), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9764), 1,
      anon_sym_DQUOTE,
    ACTIONS(9768), 1,
      aux_sym_number_token1,
    ACTIONS(9770), 1,
      aux_sym_number_token2,
    ACTIONS(9772), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9774), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9776), 1,
      anon_sym_BQUOTE,
    ACTIONS(9778), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9782), 1,
      sym__comment_word,
    ACTIONS(9784), 1,
      sym__brace_start,
    ACTIONS(10896), 1,
      anon_sym_DOLLAR,
    ACTIONS(9756), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9762), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9780), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9766), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(6225), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [230261] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5876), 1,
      aux_sym_number_token1,
    ACTIONS(5878), 1,
      aux_sym_number_token2,
    ACTIONS(5882), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(5892), 1,
      sym__brace_start,
    ACTIONS(9546), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9550), 1,
      anon_sym_DQUOTE,
    ACTIONS(9554), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9556), 1,
      anon_sym_BQUOTE,
    ACTIONS(9558), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10300), 1,
      sym_word,
    ACTIONS(10306), 1,
      sym__comment_word,
    ACTIONS(10898), 1,
      anon_sym_DOLLAR,
    ACTIONS(9544), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9560), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10302), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10304), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(5680), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [230329] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2636), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(2642), 1,
      anon_sym_DQUOTE,
    ACTIONS(2646), 1,
      aux_sym_number_token1,
    ACTIONS(2648), 1,
      aux_sym_number_token2,
    ACTIONS(2650), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(2652), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2654), 1,
      anon_sym_BQUOTE,
    ACTIONS(2656), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(2664), 1,
      sym__brace_start,
    ACTIONS(9620), 1,
      sym_word,
    ACTIONS(9626), 1,
      sym__comment_word,
    ACTIONS(10900), 1,
      anon_sym_DOLLAR,
    ACTIONS(2630), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(2658), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9622), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9624), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2344), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [230397] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3064), 1,
      sym_variable_name,
    ACTIONS(9278), 1,
      anon_sym_LPAREN,
    ACTIONS(9280), 1,
      anon_sym_BANG,
    ACTIONS(9286), 1,
      anon_sym_TILDE,
    ACTIONS(9288), 1,
      anon_sym_DOLLAR,
    ACTIONS(9290), 1,
      anon_sym_DQUOTE,
    ACTIONS(9292), 1,
      aux_sym_number_token1,
    ACTIONS(9294), 1,
      aux_sym_number_token2,
    ACTIONS(9296), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9298), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9300), 1,
      anon_sym_BQUOTE,
    ACTIONS(9302), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10902), 1,
      aux_sym__simple_variable_name_token1,
    STATE(2801), 1,
      sym__arithmetic_binary_expression,
    STATE(2803), 1,
      sym__arithmetic_ternary_expression,
    STATE(2809), 1,
      sym__arithmetic_unary_expression,
    STATE(2825), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9282), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9284), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(2916), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [230471] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3064), 1,
      sym_variable_name,
    ACTIONS(9278), 1,
      anon_sym_LPAREN,
    ACTIONS(9280), 1,
      anon_sym_BANG,
    ACTIONS(9286), 1,
      anon_sym_TILDE,
    ACTIONS(9288), 1,
      anon_sym_DOLLAR,
    ACTIONS(9290), 1,
      anon_sym_DQUOTE,
    ACTIONS(9292), 1,
      aux_sym_number_token1,
    ACTIONS(9294), 1,
      aux_sym_number_token2,
    ACTIONS(9296), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9298), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9300), 1,
      anon_sym_BQUOTE,
    ACTIONS(9302), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10904), 1,
      aux_sym__simple_variable_name_token1,
    STATE(2801), 1,
      sym__arithmetic_binary_expression,
    STATE(2803), 1,
      sym__arithmetic_ternary_expression,
    STATE(2809), 1,
      sym__arithmetic_unary_expression,
    STATE(2825), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9282), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9284), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(2915), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [230545] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(7996), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8002), 1,
      anon_sym_DQUOTE,
    ACTIONS(8006), 1,
      aux_sym_number_token1,
    ACTIONS(8008), 1,
      aux_sym_number_token2,
    ACTIONS(8010), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8012), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(8014), 1,
      anon_sym_BQUOTE,
    ACTIONS(8016), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8026), 1,
      sym__brace_start,
    ACTIONS(9356), 1,
      sym_word,
    ACTIONS(9362), 1,
      sym__comment_word,
    ACTIONS(10906), 1,
      anon_sym_DOLLAR,
    ACTIONS(7992), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8018), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9358), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9360), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(3510), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [230613] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3064), 1,
      sym_variable_name,
    ACTIONS(9278), 1,
      anon_sym_LPAREN,
    ACTIONS(9280), 1,
      anon_sym_BANG,
    ACTIONS(9286), 1,
      anon_sym_TILDE,
    ACTIONS(9288), 1,
      anon_sym_DOLLAR,
    ACTIONS(9290), 1,
      anon_sym_DQUOTE,
    ACTIONS(9292), 1,
      aux_sym_number_token1,
    ACTIONS(9294), 1,
      aux_sym_number_token2,
    ACTIONS(9296), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9298), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9300), 1,
      anon_sym_BQUOTE,
    ACTIONS(9302), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10908), 1,
      aux_sym__simple_variable_name_token1,
    STATE(2801), 1,
      sym__arithmetic_binary_expression,
    STATE(2803), 1,
      sym__arithmetic_ternary_expression,
    STATE(2809), 1,
      sym__arithmetic_unary_expression,
    STATE(2825), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9282), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9284), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(2899), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [230687] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3191), 1,
      aux_sym_number_token1,
    ACTIONS(3193), 1,
      aux_sym_number_token2,
    ACTIONS(3197), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3209), 1,
      sym__brace_start,
    ACTIONS(9674), 1,
      sym_word,
    ACTIONS(9678), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9682), 1,
      anon_sym_DQUOTE,
    ACTIONS(9686), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9688), 1,
      anon_sym_BQUOTE,
    ACTIONS(9690), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9694), 1,
      sym__comment_word,
    ACTIONS(10910), 1,
      anon_sym_DOLLAR,
    ACTIONS(9676), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9680), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9692), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9684), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1733), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [230755] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5876), 1,
      aux_sym_number_token1,
    ACTIONS(5878), 1,
      aux_sym_number_token2,
    ACTIONS(5882), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(5892), 1,
      sym__brace_start,
    ACTIONS(9546), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9550), 1,
      anon_sym_DQUOTE,
    ACTIONS(9554), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9556), 1,
      anon_sym_BQUOTE,
    ACTIONS(9558), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10300), 1,
      sym_word,
    ACTIONS(10306), 1,
      sym__comment_word,
    ACTIONS(10912), 1,
      anon_sym_DOLLAR,
    ACTIONS(9544), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9560), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10302), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10304), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(5680), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [230823] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3064), 1,
      sym_variable_name,
    ACTIONS(3586), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(9278), 1,
      anon_sym_LPAREN,
    ACTIONS(9280), 1,
      anon_sym_BANG,
    ACTIONS(9286), 1,
      anon_sym_TILDE,
    ACTIONS(9288), 1,
      anon_sym_DOLLAR,
    ACTIONS(9290), 1,
      anon_sym_DQUOTE,
    ACTIONS(9292), 1,
      aux_sym_number_token1,
    ACTIONS(9294), 1,
      aux_sym_number_token2,
    ACTIONS(9296), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9298), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9300), 1,
      anon_sym_BQUOTE,
    ACTIONS(9302), 1,
      anon_sym_DOLLAR_BQUOTE,
    STATE(2801), 1,
      sym__arithmetic_binary_expression,
    STATE(2803), 1,
      sym__arithmetic_ternary_expression,
    STATE(2809), 1,
      sym__arithmetic_unary_expression,
    STATE(2825), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9282), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9284), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(2840), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [230897] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5616), 1,
      aux_sym_number_token1,
    ACTIONS(5618), 1,
      aux_sym_number_token2,
    ACTIONS(5622), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(5632), 1,
      sym__brace_start,
    ACTIONS(8944), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8948), 1,
      anon_sym_DQUOTE,
    ACTIONS(8952), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8954), 1,
      anon_sym_BQUOTE,
    ACTIONS(8956), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10290), 1,
      sym_word,
    ACTIONS(10296), 1,
      sym__comment_word,
    ACTIONS(10914), 1,
      anon_sym_DOLLAR,
    ACTIONS(8942), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8958), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10292), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10294), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(3415), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [230965] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3064), 1,
      sym_variable_name,
    ACTIONS(9278), 1,
      anon_sym_LPAREN,
    ACTIONS(9280), 1,
      anon_sym_BANG,
    ACTIONS(9286), 1,
      anon_sym_TILDE,
    ACTIONS(9288), 1,
      anon_sym_DOLLAR,
    ACTIONS(9290), 1,
      anon_sym_DQUOTE,
    ACTIONS(9292), 1,
      aux_sym_number_token1,
    ACTIONS(9294), 1,
      aux_sym_number_token2,
    ACTIONS(9296), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9298), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9300), 1,
      anon_sym_BQUOTE,
    ACTIONS(9302), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10916), 1,
      aux_sym__simple_variable_name_token1,
    STATE(2801), 1,
      sym__arithmetic_binary_expression,
    STATE(2803), 1,
      sym__arithmetic_ternary_expression,
    STATE(2809), 1,
      sym__arithmetic_unary_expression,
    STATE(2825), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9282), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9284), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(2894), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [231039] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8571), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8577), 1,
      anon_sym_DQUOTE,
    ACTIONS(8581), 1,
      aux_sym_number_token1,
    ACTIONS(8583), 1,
      aux_sym_number_token2,
    ACTIONS(8585), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8587), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(8589), 1,
      anon_sym_BQUOTE,
    ACTIONS(8591), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8597), 1,
      sym__brace_start,
    ACTIONS(10014), 1,
      sym_word,
    ACTIONS(10020), 1,
      sym__comment_word,
    ACTIONS(10918), 1,
      anon_sym_DOLLAR,
    ACTIONS(8567), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8593), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10016), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10018), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(5199), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [231107] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2792), 1,
      aux_sym_number_token1,
    ACTIONS(2794), 1,
      aux_sym_number_token2,
    ACTIONS(2798), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2810), 1,
      sym__brace_start,
    ACTIONS(10500), 1,
      sym_word,
    ACTIONS(10504), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(10508), 1,
      anon_sym_DQUOTE,
    ACTIONS(10512), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(10514), 1,
      anon_sym_BQUOTE,
    ACTIONS(10516), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10520), 1,
      sym__comment_word,
    ACTIONS(10920), 1,
      anon_sym_DOLLAR,
    ACTIONS(10502), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(10506), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10518), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10510), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1569), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [231175] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9364), 1,
      anon_sym_LPAREN,
    ACTIONS(9366), 1,
      anon_sym_BANG,
    ACTIONS(9372), 1,
      anon_sym_TILDE,
    ACTIONS(9374), 1,
      anon_sym_DOLLAR,
    ACTIONS(9376), 1,
      anon_sym_DQUOTE,
    ACTIONS(9378), 1,
      aux_sym_number_token1,
    ACTIONS(9380), 1,
      aux_sym_number_token2,
    ACTIONS(9382), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9384), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9386), 1,
      anon_sym_BQUOTE,
    ACTIONS(9388), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9392), 1,
      sym_variable_name,
    ACTIONS(10922), 1,
      aux_sym__simple_variable_name_token1,
    STATE(3391), 1,
      sym__arithmetic_binary_expression,
    STATE(3396), 1,
      sym__arithmetic_ternary_expression,
    STATE(3398), 1,
      sym__arithmetic_unary_expression,
    STATE(3400), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9368), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9370), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(3273), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [231249] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3064), 1,
      sym_variable_name,
    ACTIONS(4048), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(9278), 1,
      anon_sym_LPAREN,
    ACTIONS(9280), 1,
      anon_sym_BANG,
    ACTIONS(9286), 1,
      anon_sym_TILDE,
    ACTIONS(9288), 1,
      anon_sym_DOLLAR,
    ACTIONS(9290), 1,
      anon_sym_DQUOTE,
    ACTIONS(9292), 1,
      aux_sym_number_token1,
    ACTIONS(9294), 1,
      aux_sym_number_token2,
    ACTIONS(9296), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9298), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9300), 1,
      anon_sym_BQUOTE,
    ACTIONS(9302), 1,
      anon_sym_DOLLAR_BQUOTE,
    STATE(2801), 1,
      sym__arithmetic_binary_expression,
    STATE(2803), 1,
      sym__arithmetic_ternary_expression,
    STATE(2809), 1,
      sym__arithmetic_unary_expression,
    STATE(2825), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9282), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9284), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(2848), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [231323] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3064), 1,
      sym_variable_name,
    ACTIONS(9278), 1,
      anon_sym_LPAREN,
    ACTIONS(9280), 1,
      anon_sym_BANG,
    ACTIONS(9286), 1,
      anon_sym_TILDE,
    ACTIONS(9288), 1,
      anon_sym_DOLLAR,
    ACTIONS(9290), 1,
      anon_sym_DQUOTE,
    ACTIONS(9292), 1,
      aux_sym_number_token1,
    ACTIONS(9294), 1,
      aux_sym_number_token2,
    ACTIONS(9296), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9298), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9300), 1,
      anon_sym_BQUOTE,
    ACTIONS(9302), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10924), 1,
      aux_sym__simple_variable_name_token1,
    STATE(2801), 1,
      sym__arithmetic_binary_expression,
    STATE(2803), 1,
      sym__arithmetic_ternary_expression,
    STATE(2809), 1,
      sym__arithmetic_unary_expression,
    STATE(2825), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9282), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9284), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(2779), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [231397] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2792), 1,
      aux_sym_number_token1,
    ACTIONS(2794), 1,
      aux_sym_number_token2,
    ACTIONS(2798), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2810), 1,
      sym__brace_start,
    ACTIONS(10500), 1,
      sym_word,
    ACTIONS(10504), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(10508), 1,
      anon_sym_DQUOTE,
    ACTIONS(10512), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(10514), 1,
      anon_sym_BQUOTE,
    ACTIONS(10516), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10520), 1,
      sym__comment_word,
    ACTIONS(10926), 1,
      anon_sym_DOLLAR,
    ACTIONS(10502), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(10506), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10518), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10510), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1569), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [231465] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(7996), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8002), 1,
      anon_sym_DQUOTE,
    ACTIONS(8006), 1,
      aux_sym_number_token1,
    ACTIONS(8008), 1,
      aux_sym_number_token2,
    ACTIONS(8010), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8012), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(8014), 1,
      anon_sym_BQUOTE,
    ACTIONS(8016), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8026), 1,
      sym__brace_start,
    ACTIONS(9356), 1,
      sym_word,
    ACTIONS(9362), 1,
      sym__comment_word,
    ACTIONS(10928), 1,
      anon_sym_DOLLAR,
    ACTIONS(7992), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8018), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9358), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9360), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(3510), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [231533] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3441), 1,
      aux_sym_number_token1,
    ACTIONS(3443), 1,
      aux_sym_number_token2,
    ACTIONS(3447), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3459), 1,
      sym__brace_start,
    ACTIONS(10930), 1,
      sym_word,
    ACTIONS(10934), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(10936), 1,
      anon_sym_DOLLAR,
    ACTIONS(10940), 1,
      anon_sym_DQUOTE,
    ACTIONS(10944), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(10946), 1,
      anon_sym_BQUOTE,
    ACTIONS(10948), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10952), 1,
      sym__comment_word,
    ACTIONS(10932), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(10938), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10950), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10942), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1859), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [231601] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(8178), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8180), 1,
      anon_sym_DOLLAR,
    ACTIONS(8182), 1,
      sym__special_character,
    ACTIONS(8184), 1,
      anon_sym_DQUOTE,
    ACTIONS(8188), 1,
      aux_sym_number_token1,
    ACTIONS(8190), 1,
      aux_sym_number_token2,
    ACTIONS(8192), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8194), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(8196), 1,
      anon_sym_BQUOTE,
    ACTIONS(8198), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8208), 1,
      sym__brace_start,
    ACTIONS(10954), 1,
      sym_word,
    ACTIONS(10958), 1,
      sym_test_operator,
    STATE(1738), 1,
      aux_sym__literal_repeat1,
    STATE(1952), 1,
      sym_concatenation,
    ACTIONS(8174), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8200), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10956), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2325), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [231673] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3064), 1,
      sym_variable_name,
    ACTIONS(9278), 1,
      anon_sym_LPAREN,
    ACTIONS(9280), 1,
      anon_sym_BANG,
    ACTIONS(9286), 1,
      anon_sym_TILDE,
    ACTIONS(9288), 1,
      anon_sym_DOLLAR,
    ACTIONS(9290), 1,
      anon_sym_DQUOTE,
    ACTIONS(9292), 1,
      aux_sym_number_token1,
    ACTIONS(9294), 1,
      aux_sym_number_token2,
    ACTIONS(9296), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9298), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9300), 1,
      anon_sym_BQUOTE,
    ACTIONS(9302), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10960), 1,
      aux_sym__simple_variable_name_token1,
    STATE(2801), 1,
      sym__arithmetic_binary_expression,
    STATE(2803), 1,
      sym__arithmetic_ternary_expression,
    STATE(2809), 1,
      sym__arithmetic_unary_expression,
    STATE(2825), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9282), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9284), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(3017), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [231747] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(8178), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8180), 1,
      anon_sym_DOLLAR,
    ACTIONS(8184), 1,
      anon_sym_DQUOTE,
    ACTIONS(8188), 1,
      aux_sym_number_token1,
    ACTIONS(8190), 1,
      aux_sym_number_token2,
    ACTIONS(8192), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8194), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(8196), 1,
      anon_sym_BQUOTE,
    ACTIONS(8198), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8208), 1,
      sym__brace_start,
    ACTIONS(8220), 1,
      sym__special_character,
    ACTIONS(10962), 1,
      sym_word,
    ACTIONS(10966), 1,
      sym_test_operator,
    STATE(1741), 1,
      aux_sym__literal_repeat1,
    STATE(1958), 1,
      sym_concatenation,
    ACTIONS(8174), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8200), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10964), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1344), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [231819] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3064), 1,
      sym_variable_name,
    ACTIONS(4046), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(9278), 1,
      anon_sym_LPAREN,
    ACTIONS(9280), 1,
      anon_sym_BANG,
    ACTIONS(9286), 1,
      anon_sym_TILDE,
    ACTIONS(9288), 1,
      anon_sym_DOLLAR,
    ACTIONS(9290), 1,
      anon_sym_DQUOTE,
    ACTIONS(9292), 1,
      aux_sym_number_token1,
    ACTIONS(9294), 1,
      aux_sym_number_token2,
    ACTIONS(9296), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9298), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9300), 1,
      anon_sym_BQUOTE,
    ACTIONS(9302), 1,
      anon_sym_DOLLAR_BQUOTE,
    STATE(2801), 1,
      sym__arithmetic_binary_expression,
    STATE(2803), 1,
      sym__arithmetic_ternary_expression,
    STATE(2809), 1,
      sym__arithmetic_unary_expression,
    STATE(2825), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9282), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9284), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(2843), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [231893] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(8178), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8180), 1,
      anon_sym_DOLLAR,
    ACTIONS(8182), 1,
      sym__special_character,
    ACTIONS(8184), 1,
      anon_sym_DQUOTE,
    ACTIONS(8188), 1,
      aux_sym_number_token1,
    ACTIONS(8190), 1,
      aux_sym_number_token2,
    ACTIONS(8192), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8194), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(8196), 1,
      anon_sym_BQUOTE,
    ACTIONS(8198), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8208), 1,
      sym__brace_start,
    ACTIONS(10968), 1,
      sym_word,
    ACTIONS(10972), 1,
      sym_test_operator,
    STATE(1741), 1,
      aux_sym__literal_repeat1,
    STATE(1958), 1,
      sym_concatenation,
    ACTIONS(8174), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8200), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10970), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2283), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [231965] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9364), 1,
      anon_sym_LPAREN,
    ACTIONS(9366), 1,
      anon_sym_BANG,
    ACTIONS(9372), 1,
      anon_sym_TILDE,
    ACTIONS(9374), 1,
      anon_sym_DOLLAR,
    ACTIONS(9376), 1,
      anon_sym_DQUOTE,
    ACTIONS(9378), 1,
      aux_sym_number_token1,
    ACTIONS(9380), 1,
      aux_sym_number_token2,
    ACTIONS(9382), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9384), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9386), 1,
      anon_sym_BQUOTE,
    ACTIONS(9388), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9392), 1,
      sym_variable_name,
    ACTIONS(10974), 1,
      aux_sym__simple_variable_name_token1,
    STATE(3391), 1,
      sym__arithmetic_binary_expression,
    STATE(3396), 1,
      sym__arithmetic_ternary_expression,
    STATE(3398), 1,
      sym__arithmetic_unary_expression,
    STATE(3400), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9368), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9370), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(3268), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [232039] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4502), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(4508), 1,
      anon_sym_DQUOTE,
    ACTIONS(4512), 1,
      aux_sym_number_token1,
    ACTIONS(4514), 1,
      aux_sym_number_token2,
    ACTIONS(4516), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(4518), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(4520), 1,
      anon_sym_BQUOTE,
    ACTIONS(4522), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(4528), 1,
      sym__brace_start,
    ACTIONS(10256), 1,
      sym_word,
    ACTIONS(10262), 1,
      sym__comment_word,
    ACTIONS(10976), 1,
      anon_sym_DOLLAR,
    ACTIONS(4500), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(4524), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10258), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10260), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2450), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [232107] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3064), 1,
      sym_variable_name,
    ACTIONS(9278), 1,
      anon_sym_LPAREN,
    ACTIONS(9280), 1,
      anon_sym_BANG,
    ACTIONS(9286), 1,
      anon_sym_TILDE,
    ACTIONS(9288), 1,
      anon_sym_DOLLAR,
    ACTIONS(9290), 1,
      anon_sym_DQUOTE,
    ACTIONS(9292), 1,
      aux_sym_number_token1,
    ACTIONS(9294), 1,
      aux_sym_number_token2,
    ACTIONS(9296), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9298), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9300), 1,
      anon_sym_BQUOTE,
    ACTIONS(9302), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10978), 1,
      aux_sym__simple_variable_name_token1,
    STATE(2801), 1,
      sym__arithmetic_binary_expression,
    STATE(2803), 1,
      sym__arithmetic_ternary_expression,
    STATE(2809), 1,
      sym__arithmetic_unary_expression,
    STATE(2825), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9282), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9284), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(2781), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [232181] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(8178), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8180), 1,
      anon_sym_DOLLAR,
    ACTIONS(8184), 1,
      anon_sym_DQUOTE,
    ACTIONS(8188), 1,
      aux_sym_number_token1,
    ACTIONS(8190), 1,
      aux_sym_number_token2,
    ACTIONS(8192), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8194), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(8196), 1,
      anon_sym_BQUOTE,
    ACTIONS(8198), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8208), 1,
      sym__brace_start,
    ACTIONS(8220), 1,
      sym__special_character,
    ACTIONS(10980), 1,
      sym_word,
    ACTIONS(10984), 1,
      sym_test_operator,
    STATE(1738), 1,
      aux_sym__literal_repeat1,
    STATE(1952), 1,
      sym_concatenation,
    ACTIONS(8174), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8200), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10982), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1330), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [232253] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1141), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1149), 1,
      anon_sym_DQUOTE,
    ACTIONS(1153), 1,
      aux_sym_number_token1,
    ACTIONS(1155), 1,
      aux_sym_number_token2,
    ACTIONS(1157), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1159), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1163), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1169), 1,
      sym__brace_start,
    ACTIONS(3060), 1,
      anon_sym_BQUOTE,
    ACTIONS(10188), 1,
      sym_word,
    ACTIONS(10196), 1,
      sym__comment_word,
    ACTIONS(10986), 1,
      anon_sym_DOLLAR,
    ACTIONS(1129), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1165), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10192), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10194), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2392), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [232321] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4502), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(4508), 1,
      anon_sym_DQUOTE,
    ACTIONS(4512), 1,
      aux_sym_number_token1,
    ACTIONS(4514), 1,
      aux_sym_number_token2,
    ACTIONS(4516), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(4518), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(4520), 1,
      anon_sym_BQUOTE,
    ACTIONS(4522), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(4528), 1,
      sym__brace_start,
    ACTIONS(10256), 1,
      sym_word,
    ACTIONS(10262), 1,
      sym__comment_word,
    ACTIONS(10988), 1,
      anon_sym_DOLLAR,
    ACTIONS(4500), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(4524), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10258), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10260), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2450), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [232389] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(4080), 1,
      anon_sym_DOLLAR,
    ACTIONS(4086), 1,
      aux_sym_number_token1,
    ACTIONS(4088), 1,
      aux_sym_number_token2,
    ACTIONS(4092), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(4102), 1,
      sym__brace_start,
    ACTIONS(9310), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9316), 1,
      anon_sym_DQUOTE,
    ACTIONS(9320), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9322), 1,
      anon_sym_BQUOTE,
    ACTIONS(9324), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10990), 1,
      sym_word,
    ACTIONS(10992), 1,
      sym__special_character,
    ACTIONS(10996), 1,
      sym_test_operator,
    STATE(4500), 1,
      aux_sym__literal_repeat1,
    STATE(5029), 1,
      sym_concatenation,
    ACTIONS(9308), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9326), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10994), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(4992), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [232461] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(4080), 1,
      anon_sym_DOLLAR,
    ACTIONS(4086), 1,
      aux_sym_number_token1,
    ACTIONS(4088), 1,
      aux_sym_number_token2,
    ACTIONS(4092), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(4102), 1,
      sym__brace_start,
    ACTIONS(9310), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9316), 1,
      anon_sym_DQUOTE,
    ACTIONS(9320), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9322), 1,
      anon_sym_BQUOTE,
    ACTIONS(9324), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10992), 1,
      sym__special_character,
    ACTIONS(10998), 1,
      sym_word,
    ACTIONS(11002), 1,
      sym_test_operator,
    STATE(4421), 1,
      aux_sym__literal_repeat1,
    STATE(4869), 1,
      sym_concatenation,
    ACTIONS(9308), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9326), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(11000), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(4896), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [232533] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3602), 1,
      sym_variable_name,
    ACTIONS(3759), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(9898), 1,
      anon_sym_LPAREN,
    ACTIONS(9900), 1,
      anon_sym_BANG,
    ACTIONS(9906), 1,
      anon_sym_TILDE,
    ACTIONS(9908), 1,
      anon_sym_DOLLAR,
    ACTIONS(9910), 1,
      anon_sym_DQUOTE,
    ACTIONS(9912), 1,
      aux_sym_number_token1,
    ACTIONS(9914), 1,
      aux_sym_number_token2,
    ACTIONS(9916), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9918), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9920), 1,
      anon_sym_BQUOTE,
    ACTIONS(9922), 1,
      anon_sym_DOLLAR_BQUOTE,
    STATE(3359), 1,
      sym__arithmetic_postfix_expression,
    STATE(3361), 1,
      sym__arithmetic_unary_expression,
    STATE(3367), 1,
      sym__arithmetic_ternary_expression,
    STATE(3375), 1,
      sym__arithmetic_binary_expression,
    ACTIONS(9902), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9904), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(3278), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [232607] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3600), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(3602), 1,
      sym_variable_name,
    ACTIONS(9898), 1,
      anon_sym_LPAREN,
    ACTIONS(9900), 1,
      anon_sym_BANG,
    ACTIONS(9906), 1,
      anon_sym_TILDE,
    ACTIONS(9908), 1,
      anon_sym_DOLLAR,
    ACTIONS(9910), 1,
      anon_sym_DQUOTE,
    ACTIONS(9912), 1,
      aux_sym_number_token1,
    ACTIONS(9914), 1,
      aux_sym_number_token2,
    ACTIONS(9916), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9918), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9920), 1,
      anon_sym_BQUOTE,
    ACTIONS(9922), 1,
      anon_sym_DOLLAR_BQUOTE,
    STATE(3359), 1,
      sym__arithmetic_postfix_expression,
    STATE(3361), 1,
      sym__arithmetic_unary_expression,
    STATE(3367), 1,
      sym__arithmetic_ternary_expression,
    STATE(3375), 1,
      sym__arithmetic_binary_expression,
    ACTIONS(9902), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9904), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(3281), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [232681] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9758), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9760), 1,
      anon_sym_DOLLAR,
    ACTIONS(9764), 1,
      anon_sym_DQUOTE,
    ACTIONS(9768), 1,
      aux_sym_number_token1,
    ACTIONS(9770), 1,
      aux_sym_number_token2,
    ACTIONS(9772), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9774), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9776), 1,
      anon_sym_BQUOTE,
    ACTIONS(9778), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9784), 1,
      sym__brace_start,
    ACTIONS(10052), 1,
      sym__special_character,
    ACTIONS(11004), 1,
      sym_word,
    ACTIONS(11008), 1,
      sym_test_operator,
    STATE(6220), 1,
      aux_sym__literal_repeat1,
    STATE(6376), 1,
      sym_concatenation,
    ACTIONS(9756), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9780), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(11006), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(6198), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [232753] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3602), 1,
      sym_variable_name,
    ACTIONS(3604), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(9898), 1,
      anon_sym_LPAREN,
    ACTIONS(9900), 1,
      anon_sym_BANG,
    ACTIONS(9906), 1,
      anon_sym_TILDE,
    ACTIONS(9908), 1,
      anon_sym_DOLLAR,
    ACTIONS(9910), 1,
      anon_sym_DQUOTE,
    ACTIONS(9912), 1,
      aux_sym_number_token1,
    ACTIONS(9914), 1,
      aux_sym_number_token2,
    ACTIONS(9916), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9918), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9920), 1,
      anon_sym_BQUOTE,
    ACTIONS(9922), 1,
      anon_sym_DOLLAR_BQUOTE,
    STATE(3359), 1,
      sym__arithmetic_postfix_expression,
    STATE(3361), 1,
      sym__arithmetic_unary_expression,
    STATE(3367), 1,
      sym__arithmetic_ternary_expression,
    STATE(3375), 1,
      sym__arithmetic_binary_expression,
    ACTIONS(9902), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9904), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(3283), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [232827] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3602), 1,
      sym_variable_name,
    ACTIONS(4042), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(9898), 1,
      anon_sym_LPAREN,
    ACTIONS(9900), 1,
      anon_sym_BANG,
    ACTIONS(9906), 1,
      anon_sym_TILDE,
    ACTIONS(9908), 1,
      anon_sym_DOLLAR,
    ACTIONS(9910), 1,
      anon_sym_DQUOTE,
    ACTIONS(9912), 1,
      aux_sym_number_token1,
    ACTIONS(9914), 1,
      aux_sym_number_token2,
    ACTIONS(9916), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9918), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9920), 1,
      anon_sym_BQUOTE,
    ACTIONS(9922), 1,
      anon_sym_DOLLAR_BQUOTE,
    STATE(3359), 1,
      sym__arithmetic_postfix_expression,
    STATE(3361), 1,
      sym__arithmetic_unary_expression,
    STATE(3367), 1,
      sym__arithmetic_ternary_expression,
    STATE(3375), 1,
      sym__arithmetic_binary_expression,
    ACTIONS(9902), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9904), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(3368), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [232901] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3787), 1,
      anon_sym_DOLLAR,
    ACTIONS(3793), 1,
      aux_sym_number_token1,
    ACTIONS(3795), 1,
      aux_sym_number_token2,
    ACTIONS(3799), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3809), 1,
      sym__brace_start,
    ACTIONS(9636), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9640), 1,
      anon_sym_DQUOTE,
    ACTIONS(9644), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9646), 1,
      anon_sym_BQUOTE,
    ACTIONS(9648), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(11010), 1,
      sym_word,
    ACTIONS(11012), 1,
      sym__special_character,
    ACTIONS(11016), 1,
      sym_test_operator,
    STATE(4294), 1,
      aux_sym__literal_repeat1,
    STATE(4861), 1,
      sym_concatenation,
    ACTIONS(9634), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9650), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(11014), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(4806), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [232973] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3787), 1,
      anon_sym_DOLLAR,
    ACTIONS(3793), 1,
      aux_sym_number_token1,
    ACTIONS(3795), 1,
      aux_sym_number_token2,
    ACTIONS(3799), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3809), 1,
      sym__brace_start,
    ACTIONS(9636), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9640), 1,
      anon_sym_DQUOTE,
    ACTIONS(9644), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9646), 1,
      anon_sym_BQUOTE,
    ACTIONS(9648), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(11012), 1,
      sym__special_character,
    ACTIONS(11018), 1,
      sym_word,
    ACTIONS(11022), 1,
      sym_test_operator,
    STATE(4347), 1,
      aux_sym__literal_repeat1,
    STATE(4617), 1,
      sym_concatenation,
    ACTIONS(9634), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9650), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(11020), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(4750), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [233045] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3435), 1,
      anon_sym_DOLLAR,
    ACTIONS(3441), 1,
      aux_sym_number_token1,
    ACTIONS(3443), 1,
      aux_sym_number_token2,
    ACTIONS(3447), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3459), 1,
      sym__brace_start,
    ACTIONS(10930), 1,
      sym_word,
    ACTIONS(10934), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(10940), 1,
      anon_sym_DQUOTE,
    ACTIONS(10944), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(10946), 1,
      anon_sym_BQUOTE,
    ACTIONS(10948), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10952), 1,
      sym__comment_word,
    ACTIONS(10932), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(10938), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10950), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10942), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1859), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [233113] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4820), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(4826), 1,
      anon_sym_DQUOTE,
    ACTIONS(4830), 1,
      aux_sym_number_token1,
    ACTIONS(4832), 1,
      aux_sym_number_token2,
    ACTIONS(4834), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(4836), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(4838), 1,
      anon_sym_BQUOTE,
    ACTIONS(4840), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(4848), 1,
      sym__brace_start,
    ACTIONS(9534), 1,
      sym_word,
    ACTIONS(9540), 1,
      sym__comment_word,
    ACTIONS(11024), 1,
      anon_sym_DOLLAR,
    ACTIONS(4818), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(4842), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9536), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9538), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2707), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [233181] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1460), 1,
      aux_sym_number_token1,
    ACTIONS(1462), 1,
      aux_sym_number_token2,
    ACTIONS(1466), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1478), 1,
      sym__brace_start,
    ACTIONS(8878), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8882), 1,
      anon_sym_DQUOTE,
    ACTIONS(8886), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8888), 1,
      anon_sym_BQUOTE,
    ACTIONS(8890), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10522), 1,
      sym_word,
    ACTIONS(10528), 1,
      sym__comment_word,
    ACTIONS(11026), 1,
      anon_sym_DOLLAR,
    ACTIONS(8876), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8892), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10524), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10526), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1120), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [233249] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9364), 1,
      anon_sym_LPAREN,
    ACTIONS(9366), 1,
      anon_sym_BANG,
    ACTIONS(9372), 1,
      anon_sym_TILDE,
    ACTIONS(9374), 1,
      anon_sym_DOLLAR,
    ACTIONS(9376), 1,
      anon_sym_DQUOTE,
    ACTIONS(9378), 1,
      aux_sym_number_token1,
    ACTIONS(9380), 1,
      aux_sym_number_token2,
    ACTIONS(9382), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9384), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9386), 1,
      anon_sym_BQUOTE,
    ACTIONS(9388), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9392), 1,
      sym_variable_name,
    ACTIONS(11028), 1,
      aux_sym__simple_variable_name_token1,
    STATE(3391), 1,
      sym__arithmetic_binary_expression,
    STATE(3396), 1,
      sym__arithmetic_ternary_expression,
    STATE(3398), 1,
      sym__arithmetic_unary_expression,
    STATE(3400), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9368), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9370), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(3434), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [233323] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9364), 1,
      anon_sym_LPAREN,
    ACTIONS(9366), 1,
      anon_sym_BANG,
    ACTIONS(9372), 1,
      anon_sym_TILDE,
    ACTIONS(9374), 1,
      anon_sym_DOLLAR,
    ACTIONS(9376), 1,
      anon_sym_DQUOTE,
    ACTIONS(9378), 1,
      aux_sym_number_token1,
    ACTIONS(9380), 1,
      aux_sym_number_token2,
    ACTIONS(9382), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9384), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9386), 1,
      anon_sym_BQUOTE,
    ACTIONS(9388), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9392), 1,
      sym_variable_name,
    ACTIONS(11030), 1,
      aux_sym__simple_variable_name_token1,
    STATE(3391), 1,
      sym__arithmetic_binary_expression,
    STATE(3396), 1,
      sym__arithmetic_ternary_expression,
    STATE(3398), 1,
      sym__arithmetic_unary_expression,
    STATE(3400), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9368), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9370), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(3433), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [233397] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9364), 1,
      anon_sym_LPAREN,
    ACTIONS(9366), 1,
      anon_sym_BANG,
    ACTIONS(9372), 1,
      anon_sym_TILDE,
    ACTIONS(9374), 1,
      anon_sym_DOLLAR,
    ACTIONS(9376), 1,
      anon_sym_DQUOTE,
    ACTIONS(9378), 1,
      aux_sym_number_token1,
    ACTIONS(9380), 1,
      aux_sym_number_token2,
    ACTIONS(9382), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9384), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9386), 1,
      anon_sym_BQUOTE,
    ACTIONS(9388), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9392), 1,
      sym_variable_name,
    ACTIONS(11032), 1,
      aux_sym__simple_variable_name_token1,
    STATE(3391), 1,
      sym__arithmetic_binary_expression,
    STATE(3396), 1,
      sym__arithmetic_ternary_expression,
    STATE(3398), 1,
      sym__arithmetic_unary_expression,
    STATE(3400), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9368), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9370), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(3432), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [233471] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1460), 1,
      aux_sym_number_token1,
    ACTIONS(1462), 1,
      aux_sym_number_token2,
    ACTIONS(1466), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1478), 1,
      sym__brace_start,
    ACTIONS(8878), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8882), 1,
      anon_sym_DQUOTE,
    ACTIONS(8886), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8888), 1,
      anon_sym_BQUOTE,
    ACTIONS(8890), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10522), 1,
      sym_word,
    ACTIONS(10528), 1,
      sym__comment_word,
    ACTIONS(11034), 1,
      anon_sym_DOLLAR,
    ACTIONS(8876), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8892), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10524), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10526), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1120), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [233539] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9364), 1,
      anon_sym_LPAREN,
    ACTIONS(9366), 1,
      anon_sym_BANG,
    ACTIONS(9372), 1,
      anon_sym_TILDE,
    ACTIONS(9374), 1,
      anon_sym_DOLLAR,
    ACTIONS(9376), 1,
      anon_sym_DQUOTE,
    ACTIONS(9378), 1,
      aux_sym_number_token1,
    ACTIONS(9380), 1,
      aux_sym_number_token2,
    ACTIONS(9382), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9384), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9386), 1,
      anon_sym_BQUOTE,
    ACTIONS(9388), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9392), 1,
      sym_variable_name,
    ACTIONS(11036), 1,
      aux_sym__simple_variable_name_token1,
    STATE(3391), 1,
      sym__arithmetic_binary_expression,
    STATE(3396), 1,
      sym__arithmetic_ternary_expression,
    STATE(3398), 1,
      sym__arithmetic_unary_expression,
    STATE(3400), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9368), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9370), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(3431), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [233613] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9364), 1,
      anon_sym_LPAREN,
    ACTIONS(9366), 1,
      anon_sym_BANG,
    ACTIONS(9372), 1,
      anon_sym_TILDE,
    ACTIONS(9374), 1,
      anon_sym_DOLLAR,
    ACTIONS(9376), 1,
      anon_sym_DQUOTE,
    ACTIONS(9378), 1,
      aux_sym_number_token1,
    ACTIONS(9380), 1,
      aux_sym_number_token2,
    ACTIONS(9382), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9384), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9386), 1,
      anon_sym_BQUOTE,
    ACTIONS(9388), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9392), 1,
      sym_variable_name,
    ACTIONS(11038), 1,
      aux_sym__simple_variable_name_token1,
    STATE(3391), 1,
      sym__arithmetic_binary_expression,
    STATE(3396), 1,
      sym__arithmetic_ternary_expression,
    STATE(3398), 1,
      sym__arithmetic_unary_expression,
    STATE(3400), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9368), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9370), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(3260), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [233687] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3064), 1,
      sym_variable_name,
    ACTIONS(3465), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(9278), 1,
      anon_sym_LPAREN,
    ACTIONS(9280), 1,
      anon_sym_BANG,
    ACTIONS(9286), 1,
      anon_sym_TILDE,
    ACTIONS(9288), 1,
      anon_sym_DOLLAR,
    ACTIONS(9290), 1,
      anon_sym_DQUOTE,
    ACTIONS(9292), 1,
      aux_sym_number_token1,
    ACTIONS(9294), 1,
      aux_sym_number_token2,
    ACTIONS(9296), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9298), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9300), 1,
      anon_sym_BQUOTE,
    ACTIONS(9302), 1,
      anon_sym_DOLLAR_BQUOTE,
    STATE(2801), 1,
      sym__arithmetic_binary_expression,
    STATE(2803), 1,
      sym__arithmetic_ternary_expression,
    STATE(2809), 1,
      sym__arithmetic_unary_expression,
    STATE(2825), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9282), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9284), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(2782), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [233761] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3064), 1,
      sym_variable_name,
    ACTIONS(9278), 1,
      anon_sym_LPAREN,
    ACTIONS(9280), 1,
      anon_sym_BANG,
    ACTIONS(9286), 1,
      anon_sym_TILDE,
    ACTIONS(9288), 1,
      anon_sym_DOLLAR,
    ACTIONS(9290), 1,
      anon_sym_DQUOTE,
    ACTIONS(9292), 1,
      aux_sym_number_token1,
    ACTIONS(9294), 1,
      aux_sym_number_token2,
    ACTIONS(9296), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9298), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9300), 1,
      anon_sym_BQUOTE,
    ACTIONS(9302), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(11040), 1,
      aux_sym__simple_variable_name_token1,
    STATE(2801), 1,
      sym__arithmetic_binary_expression,
    STATE(2803), 1,
      sym__arithmetic_ternary_expression,
    STATE(2809), 1,
      sym__arithmetic_unary_expression,
    STATE(2825), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9282), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9284), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(2845), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [233835] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1551), 1,
      anon_sym_DOLLAR,
    ACTIONS(1557), 1,
      aux_sym_number_token1,
    ACTIONS(1559), 1,
      aux_sym_number_token2,
    ACTIONS(1563), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1575), 1,
      sym__brace_start,
    ACTIONS(9198), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9202), 1,
      anon_sym_DQUOTE,
    ACTIONS(9206), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9208), 1,
      anon_sym_BQUOTE,
    ACTIONS(9210), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9262), 1,
      sym__special_character,
    ACTIONS(11042), 1,
      sym_word,
    ACTIONS(11046), 1,
      sym_test_operator,
    STATE(1436), 1,
      aux_sym__literal_repeat1,
    STATE(1543), 1,
      sym_concatenation,
    ACTIONS(9196), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9212), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(11044), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1752), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [233907] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5616), 1,
      aux_sym_number_token1,
    ACTIONS(5618), 1,
      aux_sym_number_token2,
    ACTIONS(5622), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(5632), 1,
      sym__brace_start,
    ACTIONS(8944), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8948), 1,
      anon_sym_DQUOTE,
    ACTIONS(8952), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8954), 1,
      anon_sym_BQUOTE,
    ACTIONS(8956), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10290), 1,
      sym_word,
    ACTIONS(10296), 1,
      sym__comment_word,
    ACTIONS(11048), 1,
      anon_sym_DOLLAR,
    ACTIONS(8942), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8958), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10292), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10294), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(3415), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [233975] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2636), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(2642), 1,
      anon_sym_DQUOTE,
    ACTIONS(2646), 1,
      aux_sym_number_token1,
    ACTIONS(2648), 1,
      aux_sym_number_token2,
    ACTIONS(2650), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(2652), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2654), 1,
      anon_sym_BQUOTE,
    ACTIONS(2656), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(2664), 1,
      sym__brace_start,
    ACTIONS(9620), 1,
      sym_word,
    ACTIONS(9626), 1,
      sym__comment_word,
    ACTIONS(11050), 1,
      anon_sym_DOLLAR,
    ACTIONS(2630), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(2658), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9622), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9624), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2344), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [234043] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3441), 1,
      aux_sym_number_token1,
    ACTIONS(3443), 1,
      aux_sym_number_token2,
    ACTIONS(3447), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3459), 1,
      sym__brace_start,
    ACTIONS(10930), 1,
      sym_word,
    ACTIONS(10934), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(10940), 1,
      anon_sym_DQUOTE,
    ACTIONS(10944), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(10946), 1,
      anon_sym_BQUOTE,
    ACTIONS(10948), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10952), 1,
      sym__comment_word,
    ACTIONS(11052), 1,
      anon_sym_DOLLAR,
    ACTIONS(10932), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(10938), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10950), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10942), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1859), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [234111] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9364), 1,
      anon_sym_LPAREN,
    ACTIONS(9366), 1,
      anon_sym_BANG,
    ACTIONS(9372), 1,
      anon_sym_TILDE,
    ACTIONS(9374), 1,
      anon_sym_DOLLAR,
    ACTIONS(9376), 1,
      anon_sym_DQUOTE,
    ACTIONS(9378), 1,
      aux_sym_number_token1,
    ACTIONS(9380), 1,
      aux_sym_number_token2,
    ACTIONS(9382), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9384), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9386), 1,
      anon_sym_BQUOTE,
    ACTIONS(9388), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9392), 1,
      sym_variable_name,
    ACTIONS(11054), 1,
      aux_sym__simple_variable_name_token1,
    STATE(3391), 1,
      sym__arithmetic_binary_expression,
    STATE(3396), 1,
      sym__arithmetic_ternary_expression,
    STATE(3398), 1,
      sym__arithmetic_unary_expression,
    STATE(3400), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9368), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9370), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(3430), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [234185] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9364), 1,
      anon_sym_LPAREN,
    ACTIONS(9366), 1,
      anon_sym_BANG,
    ACTIONS(9372), 1,
      anon_sym_TILDE,
    ACTIONS(9374), 1,
      anon_sym_DOLLAR,
    ACTIONS(9376), 1,
      anon_sym_DQUOTE,
    ACTIONS(9378), 1,
      aux_sym_number_token1,
    ACTIONS(9380), 1,
      aux_sym_number_token2,
    ACTIONS(9382), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9384), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9386), 1,
      anon_sym_BQUOTE,
    ACTIONS(9388), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9392), 1,
      sym_variable_name,
    ACTIONS(11056), 1,
      aux_sym__simple_variable_name_token1,
    STATE(3391), 1,
      sym__arithmetic_binary_expression,
    STATE(3396), 1,
      sym__arithmetic_ternary_expression,
    STATE(3398), 1,
      sym__arithmetic_unary_expression,
    STATE(3400), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9368), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9370), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(3429), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [234259] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9364), 1,
      anon_sym_LPAREN,
    ACTIONS(9366), 1,
      anon_sym_BANG,
    ACTIONS(9372), 1,
      anon_sym_TILDE,
    ACTIONS(9374), 1,
      anon_sym_DOLLAR,
    ACTIONS(9376), 1,
      anon_sym_DQUOTE,
    ACTIONS(9378), 1,
      aux_sym_number_token1,
    ACTIONS(9380), 1,
      aux_sym_number_token2,
    ACTIONS(9382), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9384), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9386), 1,
      anon_sym_BQUOTE,
    ACTIONS(9388), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9392), 1,
      sym_variable_name,
    ACTIONS(11058), 1,
      aux_sym__simple_variable_name_token1,
    STATE(3391), 1,
      sym__arithmetic_binary_expression,
    STATE(3396), 1,
      sym__arithmetic_ternary_expression,
    STATE(3398), 1,
      sym__arithmetic_unary_expression,
    STATE(3400), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9368), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9370), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(3257), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [234333] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9364), 1,
      anon_sym_LPAREN,
    ACTIONS(9366), 1,
      anon_sym_BANG,
    ACTIONS(9372), 1,
      anon_sym_TILDE,
    ACTIONS(9374), 1,
      anon_sym_DOLLAR,
    ACTIONS(9376), 1,
      anon_sym_DQUOTE,
    ACTIONS(9378), 1,
      aux_sym_number_token1,
    ACTIONS(9380), 1,
      aux_sym_number_token2,
    ACTIONS(9382), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9384), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9386), 1,
      anon_sym_BQUOTE,
    ACTIONS(9388), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9392), 1,
      sym_variable_name,
    ACTIONS(11060), 1,
      aux_sym__simple_variable_name_token1,
    STATE(3391), 1,
      sym__arithmetic_binary_expression,
    STATE(3396), 1,
      sym__arithmetic_ternary_expression,
    STATE(3398), 1,
      sym__arithmetic_unary_expression,
    STATE(3400), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9368), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9370), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(3428), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [234407] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9364), 1,
      anon_sym_LPAREN,
    ACTIONS(9366), 1,
      anon_sym_BANG,
    ACTIONS(9372), 1,
      anon_sym_TILDE,
    ACTIONS(9374), 1,
      anon_sym_DOLLAR,
    ACTIONS(9376), 1,
      anon_sym_DQUOTE,
    ACTIONS(9378), 1,
      aux_sym_number_token1,
    ACTIONS(9380), 1,
      aux_sym_number_token2,
    ACTIONS(9382), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9384), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9386), 1,
      anon_sym_BQUOTE,
    ACTIONS(9388), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9392), 1,
      sym_variable_name,
    ACTIONS(11062), 1,
      aux_sym__simple_variable_name_token1,
    STATE(3391), 1,
      sym__arithmetic_binary_expression,
    STATE(3396), 1,
      sym__arithmetic_ternary_expression,
    STATE(3398), 1,
      sym__arithmetic_unary_expression,
    STATE(3400), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9368), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9370), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(3425), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [234481] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9364), 1,
      anon_sym_LPAREN,
    ACTIONS(9366), 1,
      anon_sym_BANG,
    ACTIONS(9372), 1,
      anon_sym_TILDE,
    ACTIONS(9374), 1,
      anon_sym_DOLLAR,
    ACTIONS(9376), 1,
      anon_sym_DQUOTE,
    ACTIONS(9378), 1,
      aux_sym_number_token1,
    ACTIONS(9380), 1,
      aux_sym_number_token2,
    ACTIONS(9382), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9384), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9386), 1,
      anon_sym_BQUOTE,
    ACTIONS(9388), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9392), 1,
      sym_variable_name,
    ACTIONS(11064), 1,
      aux_sym__simple_variable_name_token1,
    STATE(3391), 1,
      sym__arithmetic_binary_expression,
    STATE(3396), 1,
      sym__arithmetic_ternary_expression,
    STATE(3398), 1,
      sym__arithmetic_unary_expression,
    STATE(3400), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9368), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9370), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(3421), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [234555] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9364), 1,
      anon_sym_LPAREN,
    ACTIONS(9366), 1,
      anon_sym_BANG,
    ACTIONS(9372), 1,
      anon_sym_TILDE,
    ACTIONS(9374), 1,
      anon_sym_DOLLAR,
    ACTIONS(9376), 1,
      anon_sym_DQUOTE,
    ACTIONS(9378), 1,
      aux_sym_number_token1,
    ACTIONS(9380), 1,
      aux_sym_number_token2,
    ACTIONS(9382), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9384), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9386), 1,
      anon_sym_BQUOTE,
    ACTIONS(9388), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9392), 1,
      sym_variable_name,
    ACTIONS(11066), 1,
      aux_sym__simple_variable_name_token1,
    STATE(3391), 1,
      sym__arithmetic_binary_expression,
    STATE(3396), 1,
      sym__arithmetic_ternary_expression,
    STATE(3398), 1,
      sym__arithmetic_unary_expression,
    STATE(3400), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9368), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9370), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(3413), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [234629] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9364), 1,
      anon_sym_LPAREN,
    ACTIONS(9366), 1,
      anon_sym_BANG,
    ACTIONS(9372), 1,
      anon_sym_TILDE,
    ACTIONS(9374), 1,
      anon_sym_DOLLAR,
    ACTIONS(9376), 1,
      anon_sym_DQUOTE,
    ACTIONS(9378), 1,
      aux_sym_number_token1,
    ACTIONS(9380), 1,
      aux_sym_number_token2,
    ACTIONS(9382), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9384), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9386), 1,
      anon_sym_BQUOTE,
    ACTIONS(9388), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9392), 1,
      sym_variable_name,
    ACTIONS(11068), 1,
      aux_sym__simple_variable_name_token1,
    STATE(3391), 1,
      sym__arithmetic_binary_expression,
    STATE(3396), 1,
      sym__arithmetic_ternary_expression,
    STATE(3398), 1,
      sym__arithmetic_unary_expression,
    STATE(3400), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9368), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9370), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(3412), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [234703] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9364), 1,
      anon_sym_LPAREN,
    ACTIONS(9366), 1,
      anon_sym_BANG,
    ACTIONS(9372), 1,
      anon_sym_TILDE,
    ACTIONS(9374), 1,
      anon_sym_DOLLAR,
    ACTIONS(9376), 1,
      anon_sym_DQUOTE,
    ACTIONS(9378), 1,
      aux_sym_number_token1,
    ACTIONS(9380), 1,
      aux_sym_number_token2,
    ACTIONS(9382), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9384), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9386), 1,
      anon_sym_BQUOTE,
    ACTIONS(9388), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9392), 1,
      sym_variable_name,
    ACTIONS(11070), 1,
      aux_sym__simple_variable_name_token1,
    STATE(3391), 1,
      sym__arithmetic_binary_expression,
    STATE(3396), 1,
      sym__arithmetic_ternary_expression,
    STATE(3398), 1,
      sym__arithmetic_unary_expression,
    STATE(3400), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9368), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9370), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(3410), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [234777] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9364), 1,
      anon_sym_LPAREN,
    ACTIONS(9366), 1,
      anon_sym_BANG,
    ACTIONS(9372), 1,
      anon_sym_TILDE,
    ACTIONS(9374), 1,
      anon_sym_DOLLAR,
    ACTIONS(9376), 1,
      anon_sym_DQUOTE,
    ACTIONS(9378), 1,
      aux_sym_number_token1,
    ACTIONS(9380), 1,
      aux_sym_number_token2,
    ACTIONS(9382), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9384), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9386), 1,
      anon_sym_BQUOTE,
    ACTIONS(9388), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9392), 1,
      sym_variable_name,
    ACTIONS(11072), 1,
      aux_sym__simple_variable_name_token1,
    STATE(3391), 1,
      sym__arithmetic_binary_expression,
    STATE(3396), 1,
      sym__arithmetic_ternary_expression,
    STATE(3398), 1,
      sym__arithmetic_unary_expression,
    STATE(3400), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9368), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9370), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(3409), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [234851] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1551), 1,
      anon_sym_DOLLAR,
    ACTIONS(1557), 1,
      aux_sym_number_token1,
    ACTIONS(1559), 1,
      aux_sym_number_token2,
    ACTIONS(1563), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1575), 1,
      sym__brace_start,
    ACTIONS(9198), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9202), 1,
      anon_sym_DQUOTE,
    ACTIONS(9206), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9208), 1,
      anon_sym_BQUOTE,
    ACTIONS(9210), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9262), 1,
      sym__special_character,
    ACTIONS(11074), 1,
      sym_word,
    ACTIONS(11078), 1,
      sym_test_operator,
    STATE(1401), 1,
      aux_sym__literal_repeat1,
    STATE(1573), 1,
      sym_concatenation,
    ACTIONS(9196), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9212), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(11076), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1745), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [234923] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(314), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(320), 1,
      anon_sym_DQUOTE,
    ACTIONS(324), 1,
      aux_sym_number_token1,
    ACTIONS(326), 1,
      aux_sym_number_token2,
    ACTIONS(328), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(330), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(332), 1,
      anon_sym_BQUOTE,
    ACTIONS(334), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(344), 1,
      sym__brace_start,
    ACTIONS(9400), 1,
      sym_word,
    ACTIONS(9406), 1,
      sym__comment_word,
    ACTIONS(11080), 1,
      anon_sym_DOLLAR,
    ACTIONS(312), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(336), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9402), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9404), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1037), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [234991] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3062), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(3064), 1,
      sym_variable_name,
    ACTIONS(9278), 1,
      anon_sym_LPAREN,
    ACTIONS(9280), 1,
      anon_sym_BANG,
    ACTIONS(9286), 1,
      anon_sym_TILDE,
    ACTIONS(9288), 1,
      anon_sym_DOLLAR,
    ACTIONS(9290), 1,
      anon_sym_DQUOTE,
    ACTIONS(9292), 1,
      aux_sym_number_token1,
    ACTIONS(9294), 1,
      aux_sym_number_token2,
    ACTIONS(9296), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9298), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9300), 1,
      anon_sym_BQUOTE,
    ACTIONS(9302), 1,
      anon_sym_DOLLAR_BQUOTE,
    STATE(2801), 1,
      sym__arithmetic_binary_expression,
    STATE(2803), 1,
      sym__arithmetic_ternary_expression,
    STATE(2809), 1,
      sym__arithmetic_unary_expression,
    STATE(2825), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9282), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9284), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(2836), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [235065] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9364), 1,
      anon_sym_LPAREN,
    ACTIONS(9366), 1,
      anon_sym_BANG,
    ACTIONS(9372), 1,
      anon_sym_TILDE,
    ACTIONS(9374), 1,
      anon_sym_DOLLAR,
    ACTIONS(9376), 1,
      anon_sym_DQUOTE,
    ACTIONS(9378), 1,
      aux_sym_number_token1,
    ACTIONS(9380), 1,
      aux_sym_number_token2,
    ACTIONS(9382), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9384), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9386), 1,
      anon_sym_BQUOTE,
    ACTIONS(9388), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9392), 1,
      sym_variable_name,
    ACTIONS(11082), 1,
      aux_sym__simple_variable_name_token1,
    STATE(3391), 1,
      sym__arithmetic_binary_expression,
    STATE(3396), 1,
      sym__arithmetic_ternary_expression,
    STATE(3398), 1,
      sym__arithmetic_unary_expression,
    STATE(3400), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9368), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9370), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(3325), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [235139] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1141), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(1149), 1,
      anon_sym_DQUOTE,
    ACTIONS(1153), 1,
      aux_sym_number_token1,
    ACTIONS(1155), 1,
      aux_sym_number_token2,
    ACTIONS(1157), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1159), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(1163), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(1169), 1,
      sym__brace_start,
    ACTIONS(3060), 1,
      anon_sym_BQUOTE,
    ACTIONS(10188), 1,
      sym_word,
    ACTIONS(10196), 1,
      sym__comment_word,
    ACTIONS(11084), 1,
      anon_sym_DOLLAR,
    ACTIONS(1129), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(1165), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10192), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10194), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2392), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [235207] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4767), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(4773), 1,
      anon_sym_DQUOTE,
    ACTIONS(4777), 1,
      aux_sym_number_token1,
    ACTIONS(4779), 1,
      aux_sym_number_token2,
    ACTIONS(4781), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(4783), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(4785), 1,
      anon_sym_BQUOTE,
    ACTIONS(4787), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(4793), 1,
      sym__brace_start,
    ACTIONS(10264), 1,
      sym_word,
    ACTIONS(10270), 1,
      sym__comment_word,
    ACTIONS(11086), 1,
      anon_sym_DOLLAR,
    ACTIONS(4765), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(4789), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10266), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10268), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(5387), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [235275] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3343), 1,
      aux_sym_number_token1,
    ACTIONS(3345), 1,
      aux_sym_number_token2,
    ACTIONS(3349), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3359), 1,
      sym__brace_start,
    ACTIONS(9046), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9050), 1,
      anon_sym_DQUOTE,
    ACTIONS(9054), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9056), 1,
      anon_sym_BQUOTE,
    ACTIONS(9058), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10530), 1,
      sym_word,
    ACTIONS(10536), 1,
      sym__comment_word,
    ACTIONS(11088), 1,
      anon_sym_DOLLAR,
    ACTIONS(9044), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9060), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10532), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10534), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1827), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [235343] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4767), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(4773), 1,
      anon_sym_DQUOTE,
    ACTIONS(4777), 1,
      aux_sym_number_token1,
    ACTIONS(4779), 1,
      aux_sym_number_token2,
    ACTIONS(4781), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(4783), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(4785), 1,
      anon_sym_BQUOTE,
    ACTIONS(4787), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(4793), 1,
      sym__brace_start,
    ACTIONS(10264), 1,
      sym_word,
    ACTIONS(10270), 1,
      sym__comment_word,
    ACTIONS(11090), 1,
      anon_sym_DOLLAR,
    ACTIONS(4765), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(4789), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10266), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10268), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(5387), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [235411] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9754), 1,
      sym_word,
    ACTIONS(9758), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9764), 1,
      anon_sym_DQUOTE,
    ACTIONS(9768), 1,
      aux_sym_number_token1,
    ACTIONS(9770), 1,
      aux_sym_number_token2,
    ACTIONS(9772), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9774), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9776), 1,
      anon_sym_BQUOTE,
    ACTIONS(9778), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9782), 1,
      sym__comment_word,
    ACTIONS(9784), 1,
      sym__brace_start,
    ACTIONS(11092), 1,
      anon_sym_DOLLAR,
    ACTIONS(9756), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9762), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9780), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9766), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(6225), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [235479] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2044), 1,
      aux_sym_number_token1,
    ACTIONS(2046), 1,
      aux_sym_number_token2,
    ACTIONS(2050), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2064), 1,
      sym__brace_start,
    ACTIONS(8344), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8348), 1,
      anon_sym_DQUOTE,
    ACTIONS(8352), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8354), 1,
      anon_sym_BQUOTE,
    ACTIONS(8356), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10846), 1,
      sym_word,
    ACTIONS(10852), 1,
      sym__comment_word,
    ACTIONS(11094), 1,
      anon_sym_DOLLAR,
    ACTIONS(8340), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8358), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10848), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10850), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1493), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [235547] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3343), 1,
      aux_sym_number_token1,
    ACTIONS(3345), 1,
      aux_sym_number_token2,
    ACTIONS(3349), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3359), 1,
      sym__brace_start,
    ACTIONS(9046), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9050), 1,
      anon_sym_DQUOTE,
    ACTIONS(9054), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9056), 1,
      anon_sym_BQUOTE,
    ACTIONS(9058), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10530), 1,
      sym_word,
    ACTIONS(10536), 1,
      sym__comment_word,
    ACTIONS(11096), 1,
      anon_sym_DOLLAR,
    ACTIONS(9044), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9060), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10532), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10534), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1827), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [235615] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2044), 1,
      aux_sym_number_token1,
    ACTIONS(2046), 1,
      aux_sym_number_token2,
    ACTIONS(2050), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2064), 1,
      sym__brace_start,
    ACTIONS(8344), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8348), 1,
      anon_sym_DQUOTE,
    ACTIONS(8352), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8354), 1,
      anon_sym_BQUOTE,
    ACTIONS(8356), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10846), 1,
      sym_word,
    ACTIONS(10852), 1,
      sym__comment_word,
    ACTIONS(11098), 1,
      anon_sym_DOLLAR,
    ACTIONS(8340), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8358), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10848), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10850), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1493), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [235683] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3064), 1,
      sym_variable_name,
    ACTIONS(9278), 1,
      anon_sym_LPAREN,
    ACTIONS(9280), 1,
      anon_sym_BANG,
    ACTIONS(9286), 1,
      anon_sym_TILDE,
    ACTIONS(9288), 1,
      anon_sym_DOLLAR,
    ACTIONS(9290), 1,
      anon_sym_DQUOTE,
    ACTIONS(9292), 1,
      aux_sym_number_token1,
    ACTIONS(9294), 1,
      aux_sym_number_token2,
    ACTIONS(9296), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9298), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9300), 1,
      anon_sym_BQUOTE,
    ACTIONS(9302), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(11100), 1,
      aux_sym__simple_variable_name_token1,
    STATE(2801), 1,
      sym__arithmetic_binary_expression,
    STATE(2803), 1,
      sym__arithmetic_ternary_expression,
    STATE(2809), 1,
      sym__arithmetic_unary_expression,
    STATE(2825), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9282), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9284), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(2783), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [235757] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(4080), 1,
      anon_sym_DOLLAR,
    ACTIONS(4086), 1,
      aux_sym_number_token1,
    ACTIONS(4088), 1,
      aux_sym_number_token2,
    ACTIONS(4092), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(4102), 1,
      sym__brace_start,
    ACTIONS(9310), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9316), 1,
      anon_sym_DQUOTE,
    ACTIONS(9320), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9322), 1,
      anon_sym_BQUOTE,
    ACTIONS(9324), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(11102), 1,
      sym_word,
    ACTIONS(11104), 1,
      sym__special_character,
    ACTIONS(11108), 1,
      sym_test_operator,
    STATE(4500), 1,
      aux_sym__literal_repeat1,
    STATE(5029), 1,
      sym_concatenation,
    ACTIONS(9308), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9326), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(11106), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(4622), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [235829] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9364), 1,
      anon_sym_LPAREN,
    ACTIONS(9366), 1,
      anon_sym_BANG,
    ACTIONS(9372), 1,
      anon_sym_TILDE,
    ACTIONS(9374), 1,
      anon_sym_DOLLAR,
    ACTIONS(9376), 1,
      anon_sym_DQUOTE,
    ACTIONS(9378), 1,
      aux_sym_number_token1,
    ACTIONS(9380), 1,
      aux_sym_number_token2,
    ACTIONS(9382), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9384), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9386), 1,
      anon_sym_BQUOTE,
    ACTIONS(9388), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9392), 1,
      sym_variable_name,
    ACTIONS(11110), 1,
      aux_sym__simple_variable_name_token1,
    STATE(3391), 1,
      sym__arithmetic_binary_expression,
    STATE(3396), 1,
      sym__arithmetic_ternary_expression,
    STATE(3398), 1,
      sym__arithmetic_unary_expression,
    STATE(3400), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9368), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9370), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(3252), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [235903] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(4080), 1,
      anon_sym_DOLLAR,
    ACTIONS(4086), 1,
      aux_sym_number_token1,
    ACTIONS(4088), 1,
      aux_sym_number_token2,
    ACTIONS(4092), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(4102), 1,
      sym__brace_start,
    ACTIONS(9310), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9316), 1,
      anon_sym_DQUOTE,
    ACTIONS(9320), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9322), 1,
      anon_sym_BQUOTE,
    ACTIONS(9324), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(11104), 1,
      sym__special_character,
    ACTIONS(11112), 1,
      sym_word,
    ACTIONS(11116), 1,
      sym_test_operator,
    STATE(4421), 1,
      aux_sym__literal_repeat1,
    STATE(4869), 1,
      sym_concatenation,
    ACTIONS(9308), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9326), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(11114), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(4684), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [235975] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8474), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8476), 1,
      anon_sym_DOLLAR,
    ACTIONS(8480), 1,
      anon_sym_DQUOTE,
    ACTIONS(8484), 1,
      aux_sym_number_token1,
    ACTIONS(8486), 1,
      aux_sym_number_token2,
    ACTIONS(8488), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8490), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(8492), 1,
      anon_sym_BQUOTE,
    ACTIONS(8494), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8504), 1,
      sym__brace_start,
    ACTIONS(10720), 1,
      sym_word,
    ACTIONS(10728), 1,
      sym__comment_word,
    ACTIONS(8470), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8496), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10724), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10726), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1625), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [236043] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3064), 1,
      sym_variable_name,
    ACTIONS(9278), 1,
      anon_sym_LPAREN,
    ACTIONS(9280), 1,
      anon_sym_BANG,
    ACTIONS(9286), 1,
      anon_sym_TILDE,
    ACTIONS(9288), 1,
      anon_sym_DOLLAR,
    ACTIONS(9290), 1,
      anon_sym_DQUOTE,
    ACTIONS(9292), 1,
      aux_sym_number_token1,
    ACTIONS(9294), 1,
      aux_sym_number_token2,
    ACTIONS(9296), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9298), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9300), 1,
      anon_sym_BQUOTE,
    ACTIONS(9302), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(11118), 1,
      aux_sym__simple_variable_name_token1,
    STATE(2801), 1,
      sym__arithmetic_binary_expression,
    STATE(2803), 1,
      sym__arithmetic_ternary_expression,
    STATE(2809), 1,
      sym__arithmetic_unary_expression,
    STATE(2825), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9282), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9284), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(2784), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [236117] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3787), 1,
      anon_sym_DOLLAR,
    ACTIONS(3793), 1,
      aux_sym_number_token1,
    ACTIONS(3795), 1,
      aux_sym_number_token2,
    ACTIONS(3799), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3809), 1,
      sym__brace_start,
    ACTIONS(9636), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9640), 1,
      anon_sym_DQUOTE,
    ACTIONS(9644), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9646), 1,
      anon_sym_BQUOTE,
    ACTIONS(9648), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(11120), 1,
      sym_word,
    ACTIONS(11122), 1,
      sym__special_character,
    ACTIONS(11126), 1,
      sym_test_operator,
    STATE(4294), 1,
      aux_sym__literal_repeat1,
    STATE(4861), 1,
      sym_concatenation,
    ACTIONS(9634), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9650), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(11124), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(4463), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [236189] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3793), 1,
      aux_sym_number_token1,
    ACTIONS(3795), 1,
      aux_sym_number_token2,
    ACTIONS(3799), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3809), 1,
      sym__brace_start,
    ACTIONS(9636), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9640), 1,
      anon_sym_DQUOTE,
    ACTIONS(9644), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9646), 1,
      anon_sym_BQUOTE,
    ACTIONS(9648), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9872), 1,
      sym_word,
    ACTIONS(9880), 1,
      sym__comment_word,
    ACTIONS(11128), 1,
      anon_sym_DOLLAR,
    ACTIONS(9634), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9650), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9876), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9878), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(4337), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [236257] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(7958), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(7964), 1,
      anon_sym_DQUOTE,
    ACTIONS(7968), 1,
      aux_sym_number_token1,
    ACTIONS(7970), 1,
      aux_sym_number_token2,
    ACTIONS(7972), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(7974), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(7976), 1,
      anon_sym_BQUOTE,
    ACTIONS(7978), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(7988), 1,
      sym__brace_start,
    ACTIONS(9860), 1,
      sym_word,
    ACTIONS(9868), 1,
      sym__comment_word,
    ACTIONS(11130), 1,
      anon_sym_DOLLAR,
    ACTIONS(7954), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(7980), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9864), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9866), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(4357), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [236325] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(7958), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(7964), 1,
      anon_sym_DQUOTE,
    ACTIONS(7968), 1,
      aux_sym_number_token1,
    ACTIONS(7970), 1,
      aux_sym_number_token2,
    ACTIONS(7972), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(7974), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(7976), 1,
      anon_sym_BQUOTE,
    ACTIONS(7978), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(7988), 1,
      sym__brace_start,
    ACTIONS(9860), 1,
      sym_word,
    ACTIONS(9868), 1,
      sym__comment_word,
    ACTIONS(11132), 1,
      anon_sym_DOLLAR,
    ACTIONS(7954), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(7980), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9864), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9866), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(4357), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [236393] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3787), 1,
      anon_sym_DOLLAR,
    ACTIONS(3793), 1,
      aux_sym_number_token1,
    ACTIONS(3795), 1,
      aux_sym_number_token2,
    ACTIONS(3799), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3809), 1,
      sym__brace_start,
    ACTIONS(9636), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9640), 1,
      anon_sym_DQUOTE,
    ACTIONS(9644), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9646), 1,
      anon_sym_BQUOTE,
    ACTIONS(9648), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(11122), 1,
      sym__special_character,
    ACTIONS(11134), 1,
      sym_word,
    ACTIONS(11138), 1,
      sym_test_operator,
    STATE(4347), 1,
      aux_sym__literal_repeat1,
    STATE(4617), 1,
      sym_concatenation,
    ACTIONS(9634), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9650), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(11136), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(4513), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [236465] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3064), 1,
      sym_variable_name,
    ACTIONS(9278), 1,
      anon_sym_LPAREN,
    ACTIONS(9280), 1,
      anon_sym_BANG,
    ACTIONS(9286), 1,
      anon_sym_TILDE,
    ACTIONS(9288), 1,
      anon_sym_DOLLAR,
    ACTIONS(9290), 1,
      anon_sym_DQUOTE,
    ACTIONS(9292), 1,
      aux_sym_number_token1,
    ACTIONS(9294), 1,
      aux_sym_number_token2,
    ACTIONS(9296), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9298), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9300), 1,
      anon_sym_BQUOTE,
    ACTIONS(9302), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(11140), 1,
      aux_sym__simple_variable_name_token1,
    STATE(2801), 1,
      sym__arithmetic_binary_expression,
    STATE(2803), 1,
      sym__arithmetic_ternary_expression,
    STATE(2809), 1,
      sym__arithmetic_unary_expression,
    STATE(2825), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9282), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9284), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(2820), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [236539] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9364), 1,
      anon_sym_LPAREN,
    ACTIONS(9366), 1,
      anon_sym_BANG,
    ACTIONS(9372), 1,
      anon_sym_TILDE,
    ACTIONS(9374), 1,
      anon_sym_DOLLAR,
    ACTIONS(9376), 1,
      anon_sym_DQUOTE,
    ACTIONS(9378), 1,
      aux_sym_number_token1,
    ACTIONS(9380), 1,
      aux_sym_number_token2,
    ACTIONS(9382), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9384), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9386), 1,
      anon_sym_BQUOTE,
    ACTIONS(9388), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9392), 1,
      sym_variable_name,
    ACTIONS(11142), 1,
      aux_sym__simple_variable_name_token1,
    STATE(3391), 1,
      sym__arithmetic_binary_expression,
    STATE(3396), 1,
      sym__arithmetic_ternary_expression,
    STATE(3398), 1,
      sym__arithmetic_unary_expression,
    STATE(3400), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9368), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9370), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(3274), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [236613] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4820), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(4826), 1,
      anon_sym_DQUOTE,
    ACTIONS(4830), 1,
      aux_sym_number_token1,
    ACTIONS(4832), 1,
      aux_sym_number_token2,
    ACTIONS(4834), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(4836), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(4838), 1,
      anon_sym_BQUOTE,
    ACTIONS(4840), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(4848), 1,
      sym__brace_start,
    ACTIONS(9534), 1,
      sym_word,
    ACTIONS(9540), 1,
      sym__comment_word,
    ACTIONS(11144), 1,
      anon_sym_DOLLAR,
    ACTIONS(4818), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(4842), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9536), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9538), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(2707), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [236681] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8178), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8184), 1,
      anon_sym_DQUOTE,
    ACTIONS(8188), 1,
      aux_sym_number_token1,
    ACTIONS(8190), 1,
      aux_sym_number_token2,
    ACTIONS(8192), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8194), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(8196), 1,
      anon_sym_BQUOTE,
    ACTIONS(8198), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8208), 1,
      sym__brace_start,
    ACTIONS(9720), 1,
      sym_word,
    ACTIONS(9728), 1,
      sym__comment_word,
    ACTIONS(11146), 1,
      anon_sym_DOLLAR,
    ACTIONS(8174), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8200), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(9724), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(9726), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1590), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [236749] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(47), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(53), 1,
      anon_sym_DQUOTE,
    ACTIONS(57), 1,
      aux_sym_number_token1,
    ACTIONS(59), 1,
      aux_sym_number_token2,
    ACTIONS(61), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(63), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(65), 1,
      anon_sym_BQUOTE,
    ACTIONS(67), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(79), 1,
      sym__brace_start,
    ACTIONS(10448), 1,
      sym_word,
    ACTIONS(10456), 1,
      sym__comment_word,
    ACTIONS(11148), 1,
      anon_sym_DOLLAR,
    ACTIONS(45), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(69), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10452), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10454), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1083), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [236817] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8140), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8146), 1,
      anon_sym_DQUOTE,
    ACTIONS(8150), 1,
      aux_sym_number_token1,
    ACTIONS(8152), 1,
      aux_sym_number_token2,
    ACTIONS(8154), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8156), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(8158), 1,
      anon_sym_BQUOTE,
    ACTIONS(8160), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8170), 1,
      sym__brace_start,
    ACTIONS(10564), 1,
      sym_word,
    ACTIONS(10570), 1,
      sym__comment_word,
    ACTIONS(11150), 1,
      anon_sym_DOLLAR,
    ACTIONS(8136), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8162), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10566), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10568), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(4526), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [236885] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8140), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8146), 1,
      anon_sym_DQUOTE,
    ACTIONS(8150), 1,
      aux_sym_number_token1,
    ACTIONS(8152), 1,
      aux_sym_number_token2,
    ACTIONS(8154), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8156), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(8158), 1,
      anon_sym_BQUOTE,
    ACTIONS(8160), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8170), 1,
      sym__brace_start,
    ACTIONS(10564), 1,
      sym_word,
    ACTIONS(10570), 1,
      sym__comment_word,
    ACTIONS(11152), 1,
      anon_sym_DOLLAR,
    ACTIONS(8136), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8162), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10566), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10568), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(4526), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [236953] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(4080), 1,
      anon_sym_DOLLAR,
    ACTIONS(4086), 1,
      aux_sym_number_token1,
    ACTIONS(4088), 1,
      aux_sym_number_token2,
    ACTIONS(4092), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(4102), 1,
      sym__brace_start,
    ACTIONS(9310), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9316), 1,
      anon_sym_DQUOTE,
    ACTIONS(9320), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9322), 1,
      anon_sym_BQUOTE,
    ACTIONS(9324), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(11154), 1,
      sym_word,
    ACTIONS(11156), 1,
      sym__special_character,
    ACTIONS(11160), 1,
      sym_test_operator,
    STATE(4500), 1,
      aux_sym__literal_repeat1,
    STATE(5029), 1,
      sym_concatenation,
    ACTIONS(9308), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9326), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(11158), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(5306), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [237025] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9364), 1,
      anon_sym_LPAREN,
    ACTIONS(9366), 1,
      anon_sym_BANG,
    ACTIONS(9372), 1,
      anon_sym_TILDE,
    ACTIONS(9374), 1,
      anon_sym_DOLLAR,
    ACTIONS(9376), 1,
      anon_sym_DQUOTE,
    ACTIONS(9378), 1,
      aux_sym_number_token1,
    ACTIONS(9380), 1,
      aux_sym_number_token2,
    ACTIONS(9382), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9384), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9386), 1,
      anon_sym_BQUOTE,
    ACTIONS(9388), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9392), 1,
      sym_variable_name,
    ACTIONS(11162), 1,
      aux_sym__simple_variable_name_token1,
    STATE(3391), 1,
      sym__arithmetic_binary_expression,
    STATE(3396), 1,
      sym__arithmetic_ternary_expression,
    STATE(3398), 1,
      sym__arithmetic_unary_expression,
    STATE(3400), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9368), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9370), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(3244), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [237099] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3064), 1,
      sym_variable_name,
    ACTIONS(9278), 1,
      anon_sym_LPAREN,
    ACTIONS(9280), 1,
      anon_sym_BANG,
    ACTIONS(9286), 1,
      anon_sym_TILDE,
    ACTIONS(9288), 1,
      anon_sym_DOLLAR,
    ACTIONS(9290), 1,
      anon_sym_DQUOTE,
    ACTIONS(9292), 1,
      aux_sym_number_token1,
    ACTIONS(9294), 1,
      aux_sym_number_token2,
    ACTIONS(9296), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9298), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9300), 1,
      anon_sym_BQUOTE,
    ACTIONS(9302), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(11164), 1,
      aux_sym__simple_variable_name_token1,
    STATE(2801), 1,
      sym__arithmetic_binary_expression,
    STATE(2803), 1,
      sym__arithmetic_ternary_expression,
    STATE(2809), 1,
      sym__arithmetic_unary_expression,
    STATE(2825), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9282), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9284), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(2785), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [237173] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9364), 1,
      anon_sym_LPAREN,
    ACTIONS(9366), 1,
      anon_sym_BANG,
    ACTIONS(9372), 1,
      anon_sym_TILDE,
    ACTIONS(9374), 1,
      anon_sym_DOLLAR,
    ACTIONS(9376), 1,
      anon_sym_DQUOTE,
    ACTIONS(9378), 1,
      aux_sym_number_token1,
    ACTIONS(9380), 1,
      aux_sym_number_token2,
    ACTIONS(9382), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9384), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9386), 1,
      anon_sym_BQUOTE,
    ACTIONS(9388), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9392), 1,
      sym_variable_name,
    ACTIONS(11166), 1,
      aux_sym__simple_variable_name_token1,
    STATE(3391), 1,
      sym__arithmetic_binary_expression,
    STATE(3396), 1,
      sym__arithmetic_ternary_expression,
    STATE(3398), 1,
      sym__arithmetic_unary_expression,
    STATE(3400), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9368), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9370), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(3379), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [237247] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(4080), 1,
      anon_sym_DOLLAR,
    ACTIONS(4086), 1,
      aux_sym_number_token1,
    ACTIONS(4088), 1,
      aux_sym_number_token2,
    ACTIONS(4092), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(4102), 1,
      sym__brace_start,
    ACTIONS(9310), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9316), 1,
      anon_sym_DQUOTE,
    ACTIONS(9320), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9322), 1,
      anon_sym_BQUOTE,
    ACTIONS(9324), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(11156), 1,
      sym__special_character,
    ACTIONS(11168), 1,
      sym_word,
    ACTIONS(11172), 1,
      sym_test_operator,
    STATE(4421), 1,
      aux_sym__literal_repeat1,
    STATE(4869), 1,
      sym_concatenation,
    ACTIONS(9308), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9326), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(11170), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(5286), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [237319] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3681), 1,
      aux_sym_number_token1,
    ACTIONS(3683), 1,
      aux_sym_number_token2,
    ACTIONS(3687), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3697), 1,
      sym__brace_start,
    ACTIONS(10062), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(10066), 1,
      anon_sym_DQUOTE,
    ACTIONS(10070), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(10072), 1,
      anon_sym_BQUOTE,
    ACTIONS(10074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10540), 1,
      sym_word,
    ACTIONS(10546), 1,
      sym__comment_word,
    ACTIONS(11174), 1,
      anon_sym_DOLLAR,
    ACTIONS(10060), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(10076), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10542), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10544), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(4831), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [237387] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3787), 1,
      anon_sym_DOLLAR,
    ACTIONS(3793), 1,
      aux_sym_number_token1,
    ACTIONS(3795), 1,
      aux_sym_number_token2,
    ACTIONS(3799), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3809), 1,
      sym__brace_start,
    ACTIONS(9636), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9640), 1,
      anon_sym_DQUOTE,
    ACTIONS(9644), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9646), 1,
      anon_sym_BQUOTE,
    ACTIONS(9648), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(11176), 1,
      sym_word,
    ACTIONS(11178), 1,
      sym__special_character,
    ACTIONS(11182), 1,
      sym_test_operator,
    STATE(4294), 1,
      aux_sym__literal_repeat1,
    STATE(4861), 1,
      sym_concatenation,
    ACTIONS(9634), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9650), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(11180), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(5130), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [237459] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3787), 1,
      anon_sym_DOLLAR,
    ACTIONS(3793), 1,
      aux_sym_number_token1,
    ACTIONS(3795), 1,
      aux_sym_number_token2,
    ACTIONS(3799), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3809), 1,
      sym__brace_start,
    ACTIONS(9636), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9640), 1,
      anon_sym_DQUOTE,
    ACTIONS(9644), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9646), 1,
      anon_sym_BQUOTE,
    ACTIONS(9648), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(11178), 1,
      sym__special_character,
    ACTIONS(11184), 1,
      sym_word,
    ACTIONS(11188), 1,
      sym_test_operator,
    STATE(4347), 1,
      aux_sym__literal_repeat1,
    STATE(4617), 1,
      sym_concatenation,
    ACTIONS(9634), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9650), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(11186), 2,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(5128), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [237531] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3064), 1,
      sym_variable_name,
    ACTIONS(3419), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(9278), 1,
      anon_sym_LPAREN,
    ACTIONS(9280), 1,
      anon_sym_BANG,
    ACTIONS(9286), 1,
      anon_sym_TILDE,
    ACTIONS(9288), 1,
      anon_sym_DOLLAR,
    ACTIONS(9290), 1,
      anon_sym_DQUOTE,
    ACTIONS(9292), 1,
      aux_sym_number_token1,
    ACTIONS(9294), 1,
      aux_sym_number_token2,
    ACTIONS(9296), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9298), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9300), 1,
      anon_sym_BQUOTE,
    ACTIONS(9302), 1,
      anon_sym_DOLLAR_BQUOTE,
    STATE(2801), 1,
      sym__arithmetic_binary_expression,
    STATE(2803), 1,
      sym__arithmetic_ternary_expression,
    STATE(2809), 1,
      sym__arithmetic_unary_expression,
    STATE(2825), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9282), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9284), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(2811), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [237605] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3681), 1,
      aux_sym_number_token1,
    ACTIONS(3683), 1,
      aux_sym_number_token2,
    ACTIONS(3687), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3697), 1,
      sym__brace_start,
    ACTIONS(10062), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(10066), 1,
      anon_sym_DQUOTE,
    ACTIONS(10070), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(10072), 1,
      anon_sym_BQUOTE,
    ACTIONS(10074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10540), 1,
      sym_word,
    ACTIONS(10546), 1,
      sym__comment_word,
    ACTIONS(11190), 1,
      anon_sym_DOLLAR,
    ACTIONS(10060), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(10076), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10542), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10544), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(4831), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [237673] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9364), 1,
      anon_sym_LPAREN,
    ACTIONS(9366), 1,
      anon_sym_BANG,
    ACTIONS(9372), 1,
      anon_sym_TILDE,
    ACTIONS(9374), 1,
      anon_sym_DOLLAR,
    ACTIONS(9376), 1,
      anon_sym_DQUOTE,
    ACTIONS(9378), 1,
      aux_sym_number_token1,
    ACTIONS(9380), 1,
      aux_sym_number_token2,
    ACTIONS(9382), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9384), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9386), 1,
      anon_sym_BQUOTE,
    ACTIONS(9388), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9392), 1,
      sym_variable_name,
    ACTIONS(11192), 1,
      aux_sym__simple_variable_name_token1,
    STATE(3391), 1,
      sym__arithmetic_binary_expression,
    STATE(3396), 1,
      sym__arithmetic_ternary_expression,
    STATE(3398), 1,
      sym__arithmetic_unary_expression,
    STATE(3400), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9368), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9370), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(3235), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [237747] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3064), 1,
      sym_variable_name,
    ACTIONS(9278), 1,
      anon_sym_LPAREN,
    ACTIONS(9280), 1,
      anon_sym_BANG,
    ACTIONS(9286), 1,
      anon_sym_TILDE,
    ACTIONS(9288), 1,
      anon_sym_DOLLAR,
    ACTIONS(9290), 1,
      anon_sym_DQUOTE,
    ACTIONS(9292), 1,
      aux_sym_number_token1,
    ACTIONS(9294), 1,
      aux_sym_number_token2,
    ACTIONS(9296), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9298), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9300), 1,
      anon_sym_BQUOTE,
    ACTIONS(9302), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(11194), 1,
      aux_sym__simple_variable_name_token1,
    STATE(2801), 1,
      sym__arithmetic_binary_expression,
    STATE(2803), 1,
      sym__arithmetic_ternary_expression,
    STATE(2809), 1,
      sym__arithmetic_unary_expression,
    STATE(2825), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9282), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9284), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(2786), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [237821] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9364), 1,
      anon_sym_LPAREN,
    ACTIONS(9366), 1,
      anon_sym_BANG,
    ACTIONS(9372), 1,
      anon_sym_TILDE,
    ACTIONS(9374), 1,
      anon_sym_DOLLAR,
    ACTIONS(9376), 1,
      anon_sym_DQUOTE,
    ACTIONS(9378), 1,
      aux_sym_number_token1,
    ACTIONS(9380), 1,
      aux_sym_number_token2,
    ACTIONS(9382), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9384), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9386), 1,
      anon_sym_BQUOTE,
    ACTIONS(9388), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9392), 1,
      sym_variable_name,
    ACTIONS(11196), 1,
      aux_sym__simple_variable_name_token1,
    STATE(3391), 1,
      sym__arithmetic_binary_expression,
    STATE(3396), 1,
      sym__arithmetic_ternary_expression,
    STATE(3398), 1,
      sym__arithmetic_unary_expression,
    STATE(3400), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9368), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9370), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(3239), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [237895] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9364), 1,
      anon_sym_LPAREN,
    ACTIONS(9366), 1,
      anon_sym_BANG,
    ACTIONS(9372), 1,
      anon_sym_TILDE,
    ACTIONS(9374), 1,
      anon_sym_DOLLAR,
    ACTIONS(9376), 1,
      anon_sym_DQUOTE,
    ACTIONS(9378), 1,
      aux_sym_number_token1,
    ACTIONS(9380), 1,
      aux_sym_number_token2,
    ACTIONS(9382), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9384), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9386), 1,
      anon_sym_BQUOTE,
    ACTIONS(9388), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9392), 1,
      sym_variable_name,
    ACTIONS(11198), 1,
      aux_sym__simple_variable_name_token1,
    STATE(3391), 1,
      sym__arithmetic_binary_expression,
    STATE(3396), 1,
      sym__arithmetic_ternary_expression,
    STATE(3398), 1,
      sym__arithmetic_unary_expression,
    STATE(3400), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9368), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9370), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(3224), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [237969] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3064), 1,
      sym_variable_name,
    ACTIONS(9278), 1,
      anon_sym_LPAREN,
    ACTIONS(9280), 1,
      anon_sym_BANG,
    ACTIONS(9286), 1,
      anon_sym_TILDE,
    ACTIONS(9288), 1,
      anon_sym_DOLLAR,
    ACTIONS(9290), 1,
      anon_sym_DQUOTE,
    ACTIONS(9292), 1,
      aux_sym_number_token1,
    ACTIONS(9294), 1,
      aux_sym_number_token2,
    ACTIONS(9296), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9298), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9300), 1,
      anon_sym_BQUOTE,
    ACTIONS(9302), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(11200), 1,
      aux_sym__simple_variable_name_token1,
    STATE(2801), 1,
      sym__arithmetic_binary_expression,
    STATE(2803), 1,
      sym__arithmetic_ternary_expression,
    STATE(2809), 1,
      sym__arithmetic_unary_expression,
    STATE(2825), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9282), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9284), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(2787), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [238043] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3032), 1,
      aux_sym_number_token1,
    ACTIONS(3034), 1,
      aux_sym_number_token2,
    ACTIONS(3038), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3048), 1,
      sym__brace_start,
    ACTIONS(10134), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(10138), 1,
      anon_sym_DQUOTE,
    ACTIONS(10142), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(10144), 1,
      anon_sym_BQUOTE,
    ACTIONS(10146), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10556), 1,
      sym_word,
    ACTIONS(10562), 1,
      sym__comment_word,
    ACTIONS(11202), 1,
      anon_sym_DOLLAR,
    ACTIONS(10132), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(10148), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10558), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10560), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(4493), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [238111] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3032), 1,
      aux_sym_number_token1,
    ACTIONS(3034), 1,
      aux_sym_number_token2,
    ACTIONS(3038), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(3048), 1,
      sym__brace_start,
    ACTIONS(10134), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(10138), 1,
      anon_sym_DQUOTE,
    ACTIONS(10142), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(10144), 1,
      anon_sym_BQUOTE,
    ACTIONS(10146), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10556), 1,
      sym_word,
    ACTIONS(10562), 1,
      sym__comment_word,
    ACTIONS(11204), 1,
      anon_sym_DOLLAR,
    ACTIONS(10132), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(10148), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10558), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10560), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(4493), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [238179] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8306), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8312), 1,
      anon_sym_DQUOTE,
    ACTIONS(8316), 1,
      aux_sym_number_token1,
    ACTIONS(8318), 1,
      aux_sym_number_token2,
    ACTIONS(8320), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8322), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(8324), 1,
      anon_sym_BQUOTE,
    ACTIONS(8326), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8336), 1,
      sym__brace_start,
    ACTIONS(10282), 1,
      sym_word,
    ACTIONS(10288), 1,
      sym__comment_word,
    ACTIONS(11206), 1,
      anon_sym_DOLLAR,
    ACTIONS(8302), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8328), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10284), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10286), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(5428), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [238247] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3064), 1,
      sym_variable_name,
    ACTIONS(9278), 1,
      anon_sym_LPAREN,
    ACTIONS(9280), 1,
      anon_sym_BANG,
    ACTIONS(9286), 1,
      anon_sym_TILDE,
    ACTIONS(9288), 1,
      anon_sym_DOLLAR,
    ACTIONS(9290), 1,
      anon_sym_DQUOTE,
    ACTIONS(9292), 1,
      aux_sym_number_token1,
    ACTIONS(9294), 1,
      aux_sym_number_token2,
    ACTIONS(9296), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9298), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9300), 1,
      anon_sym_BQUOTE,
    ACTIONS(9302), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(11208), 1,
      aux_sym__simple_variable_name_token1,
    STATE(2801), 1,
      sym__arithmetic_binary_expression,
    STATE(2803), 1,
      sym__arithmetic_ternary_expression,
    STATE(2809), 1,
      sym__arithmetic_unary_expression,
    STATE(2825), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9282), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9284), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(2804), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [238321] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9364), 1,
      anon_sym_LPAREN,
    ACTIONS(9366), 1,
      anon_sym_BANG,
    ACTIONS(9372), 1,
      anon_sym_TILDE,
    ACTIONS(9374), 1,
      anon_sym_DOLLAR,
    ACTIONS(9376), 1,
      anon_sym_DQUOTE,
    ACTIONS(9378), 1,
      aux_sym_number_token1,
    ACTIONS(9380), 1,
      aux_sym_number_token2,
    ACTIONS(9382), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9384), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9386), 1,
      anon_sym_BQUOTE,
    ACTIONS(9388), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9392), 1,
      sym_variable_name,
    ACTIONS(11210), 1,
      aux_sym__simple_variable_name_token1,
    STATE(3391), 1,
      sym__arithmetic_binary_expression,
    STATE(3396), 1,
      sym__arithmetic_ternary_expression,
    STATE(3398), 1,
      sym__arithmetic_unary_expression,
    STATE(3400), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9368), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9370), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(3212), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [238395] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3064), 1,
      sym_variable_name,
    ACTIONS(9278), 1,
      anon_sym_LPAREN,
    ACTIONS(9280), 1,
      anon_sym_BANG,
    ACTIONS(9286), 1,
      anon_sym_TILDE,
    ACTIONS(9288), 1,
      anon_sym_DOLLAR,
    ACTIONS(9290), 1,
      anon_sym_DQUOTE,
    ACTIONS(9292), 1,
      aux_sym_number_token1,
    ACTIONS(9294), 1,
      aux_sym_number_token2,
    ACTIONS(9296), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9298), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9300), 1,
      anon_sym_BQUOTE,
    ACTIONS(9302), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(11212), 1,
      aux_sym__simple_variable_name_token1,
    STATE(2801), 1,
      sym__arithmetic_binary_expression,
    STATE(2803), 1,
      sym__arithmetic_ternary_expression,
    STATE(2809), 1,
      sym__arithmetic_unary_expression,
    STATE(2825), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9282), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9284), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(2788), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [238469] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8306), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(8312), 1,
      anon_sym_DQUOTE,
    ACTIONS(8316), 1,
      aux_sym_number_token1,
    ACTIONS(8318), 1,
      aux_sym_number_token2,
    ACTIONS(8320), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8322), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(8324), 1,
      anon_sym_BQUOTE,
    ACTIONS(8326), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(8336), 1,
      sym__brace_start,
    ACTIONS(10282), 1,
      sym_word,
    ACTIONS(10288), 1,
      sym__comment_word,
    ACTIONS(11214), 1,
      anon_sym_DOLLAR,
    ACTIONS(8302), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(8328), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10284), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10286), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(5428), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [238537] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9364), 1,
      anon_sym_LPAREN,
    ACTIONS(9366), 1,
      anon_sym_BANG,
    ACTIONS(9372), 1,
      anon_sym_TILDE,
    ACTIONS(9374), 1,
      anon_sym_DOLLAR,
    ACTIONS(9376), 1,
      anon_sym_DQUOTE,
    ACTIONS(9378), 1,
      aux_sym_number_token1,
    ACTIONS(9380), 1,
      aux_sym_number_token2,
    ACTIONS(9382), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9384), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9386), 1,
      anon_sym_BQUOTE,
    ACTIONS(9388), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9392), 1,
      sym_variable_name,
    ACTIONS(11216), 1,
      aux_sym__simple_variable_name_token1,
    STATE(3391), 1,
      sym__arithmetic_binary_expression,
    STATE(3396), 1,
      sym__arithmetic_ternary_expression,
    STATE(3398), 1,
      sym__arithmetic_unary_expression,
    STATE(3400), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9368), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9370), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(3209), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [238611] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3064), 1,
      sym_variable_name,
    ACTIONS(9278), 1,
      anon_sym_LPAREN,
    ACTIONS(9280), 1,
      anon_sym_BANG,
    ACTIONS(9286), 1,
      anon_sym_TILDE,
    ACTIONS(9288), 1,
      anon_sym_DOLLAR,
    ACTIONS(9290), 1,
      anon_sym_DQUOTE,
    ACTIONS(9292), 1,
      aux_sym_number_token1,
    ACTIONS(9294), 1,
      aux_sym_number_token2,
    ACTIONS(9296), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9298), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9300), 1,
      anon_sym_BQUOTE,
    ACTIONS(9302), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(11218), 1,
      aux_sym__simple_variable_name_token1,
    STATE(2801), 1,
      sym__arithmetic_binary_expression,
    STATE(2803), 1,
      sym__arithmetic_ternary_expression,
    STATE(2809), 1,
      sym__arithmetic_unary_expression,
    STATE(2825), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9282), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9284), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(2800), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [238685] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9364), 1,
      anon_sym_LPAREN,
    ACTIONS(9366), 1,
      anon_sym_BANG,
    ACTIONS(9372), 1,
      anon_sym_TILDE,
    ACTIONS(9374), 1,
      anon_sym_DOLLAR,
    ACTIONS(9376), 1,
      anon_sym_DQUOTE,
    ACTIONS(9378), 1,
      aux_sym_number_token1,
    ACTIONS(9380), 1,
      aux_sym_number_token2,
    ACTIONS(9382), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9384), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9386), 1,
      anon_sym_BQUOTE,
    ACTIONS(9388), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9392), 1,
      sym_variable_name,
    ACTIONS(11220), 1,
      aux_sym__simple_variable_name_token1,
    STATE(3391), 1,
      sym__arithmetic_binary_expression,
    STATE(3396), 1,
      sym__arithmetic_ternary_expression,
    STATE(3398), 1,
      sym__arithmetic_unary_expression,
    STATE(3400), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9368), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9370), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(3185), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [238759] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3064), 1,
      sym_variable_name,
    ACTIONS(9278), 1,
      anon_sym_LPAREN,
    ACTIONS(9280), 1,
      anon_sym_BANG,
    ACTIONS(9286), 1,
      anon_sym_TILDE,
    ACTIONS(9288), 1,
      anon_sym_DOLLAR,
    ACTIONS(9290), 1,
      anon_sym_DQUOTE,
    ACTIONS(9292), 1,
      aux_sym_number_token1,
    ACTIONS(9294), 1,
      aux_sym_number_token2,
    ACTIONS(9296), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9298), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9300), 1,
      anon_sym_BQUOTE,
    ACTIONS(9302), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(11222), 1,
      aux_sym__simple_variable_name_token1,
    STATE(2801), 1,
      sym__arithmetic_binary_expression,
    STATE(2803), 1,
      sym__arithmetic_ternary_expression,
    STATE(2809), 1,
      sym__arithmetic_unary_expression,
    STATE(2825), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9282), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9284), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(2795), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [238833] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9364), 1,
      anon_sym_LPAREN,
    ACTIONS(9366), 1,
      anon_sym_BANG,
    ACTIONS(9372), 1,
      anon_sym_TILDE,
    ACTIONS(9374), 1,
      anon_sym_DOLLAR,
    ACTIONS(9376), 1,
      anon_sym_DQUOTE,
    ACTIONS(9378), 1,
      aux_sym_number_token1,
    ACTIONS(9380), 1,
      aux_sym_number_token2,
    ACTIONS(9382), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9384), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9386), 1,
      anon_sym_BQUOTE,
    ACTIONS(9388), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9392), 1,
      sym_variable_name,
    ACTIONS(11224), 1,
      aux_sym__simple_variable_name_token1,
    STATE(3391), 1,
      sym__arithmetic_binary_expression,
    STATE(3396), 1,
      sym__arithmetic_ternary_expression,
    STATE(3398), 1,
      sym__arithmetic_unary_expression,
    STATE(3400), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9368), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9370), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(3204), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [238907] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9364), 1,
      anon_sym_LPAREN,
    ACTIONS(9366), 1,
      anon_sym_BANG,
    ACTIONS(9372), 1,
      anon_sym_TILDE,
    ACTIONS(9374), 1,
      anon_sym_DOLLAR,
    ACTIONS(9376), 1,
      anon_sym_DQUOTE,
    ACTIONS(9378), 1,
      aux_sym_number_token1,
    ACTIONS(9380), 1,
      aux_sym_number_token2,
    ACTIONS(9382), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9384), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9386), 1,
      anon_sym_BQUOTE,
    ACTIONS(9388), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9392), 1,
      sym_variable_name,
    ACTIONS(11226), 1,
      aux_sym__simple_variable_name_token1,
    STATE(3391), 1,
      sym__arithmetic_binary_expression,
    STATE(3396), 1,
      sym__arithmetic_ternary_expression,
    STATE(3398), 1,
      sym__arithmetic_unary_expression,
    STATE(3400), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9368), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9370), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(3165), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [238981] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3064), 1,
      sym_variable_name,
    ACTIONS(9278), 1,
      anon_sym_LPAREN,
    ACTIONS(9280), 1,
      anon_sym_BANG,
    ACTIONS(9286), 1,
      anon_sym_TILDE,
    ACTIONS(9288), 1,
      anon_sym_DOLLAR,
    ACTIONS(9290), 1,
      anon_sym_DQUOTE,
    ACTIONS(9292), 1,
      aux_sym_number_token1,
    ACTIONS(9294), 1,
      aux_sym_number_token2,
    ACTIONS(9296), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9298), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9300), 1,
      anon_sym_BQUOTE,
    ACTIONS(9302), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(11228), 1,
      aux_sym__simple_variable_name_token1,
    STATE(2801), 1,
      sym__arithmetic_binary_expression,
    STATE(2803), 1,
      sym__arithmetic_ternary_expression,
    STATE(2809), 1,
      sym__arithmetic_unary_expression,
    STATE(2825), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9282), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9284), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(2789), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [239055] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(3064), 1,
      sym_variable_name,
    ACTIONS(9278), 1,
      anon_sym_LPAREN,
    ACTIONS(9280), 1,
      anon_sym_BANG,
    ACTIONS(9286), 1,
      anon_sym_TILDE,
    ACTIONS(9288), 1,
      anon_sym_DOLLAR,
    ACTIONS(9290), 1,
      anon_sym_DQUOTE,
    ACTIONS(9292), 1,
      aux_sym_number_token1,
    ACTIONS(9294), 1,
      aux_sym_number_token2,
    ACTIONS(9296), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9298), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9300), 1,
      anon_sym_BQUOTE,
    ACTIONS(9302), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(11230), 1,
      aux_sym__simple_variable_name_token1,
    STATE(2801), 1,
      sym__arithmetic_binary_expression,
    STATE(2803), 1,
      sym__arithmetic_ternary_expression,
    STATE(2809), 1,
      sym__arithmetic_unary_expression,
    STATE(2825), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9282), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9284), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(2791), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [239129] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9364), 1,
      anon_sym_LPAREN,
    ACTIONS(9366), 1,
      anon_sym_BANG,
    ACTIONS(9372), 1,
      anon_sym_TILDE,
    ACTIONS(9374), 1,
      anon_sym_DOLLAR,
    ACTIONS(9376), 1,
      anon_sym_DQUOTE,
    ACTIONS(9378), 1,
      aux_sym_number_token1,
    ACTIONS(9380), 1,
      aux_sym_number_token2,
    ACTIONS(9382), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9384), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9386), 1,
      anon_sym_BQUOTE,
    ACTIONS(9388), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9392), 1,
      sym_variable_name,
    ACTIONS(11232), 1,
      aux_sym__simple_variable_name_token1,
    STATE(3391), 1,
      sym__arithmetic_binary_expression,
    STATE(3396), 1,
      sym__arithmetic_ternary_expression,
    STATE(3398), 1,
      sym__arithmetic_unary_expression,
    STATE(3400), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9368), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9370), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(3182), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [239203] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2580), 1,
      aux_sym_number_token1,
    ACTIONS(2582), 1,
      aux_sym_number_token2,
    ACTIONS(2586), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2596), 1,
      sym__brace_start,
    ACTIONS(9064), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9068), 1,
      anon_sym_DQUOTE,
    ACTIONS(9072), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9074), 1,
      anon_sym_BQUOTE,
    ACTIONS(9076), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10548), 1,
      sym_word,
    ACTIONS(10554), 1,
      sym__comment_word,
    ACTIONS(11234), 1,
      anon_sym_DOLLAR,
    ACTIONS(9062), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9078), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10550), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10552), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1669), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [239271] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2580), 1,
      aux_sym_number_token1,
    ACTIONS(2582), 1,
      aux_sym_number_token2,
    ACTIONS(2586), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(2596), 1,
      sym__brace_start,
    ACTIONS(9064), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(9068), 1,
      anon_sym_DQUOTE,
    ACTIONS(9072), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9074), 1,
      anon_sym_BQUOTE,
    ACTIONS(9076), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(10548), 1,
      sym_word,
    ACTIONS(10554), 1,
      sym__comment_word,
    ACTIONS(11236), 1,
      anon_sym_DOLLAR,
    ACTIONS(9062), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    ACTIONS(9078), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(10550), 2,
      sym_test_operator,
      sym__special_character,
    ACTIONS(10552), 3,
      sym__bare_dollar,
      sym_raw_string,
      sym_ansi_c_string,
    STATE(1669), 9,
      sym_arithmetic_expansion,
      sym_brace_expression,
      sym_string,
      sym_translated_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [239339] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9364), 1,
      anon_sym_LPAREN,
    ACTIONS(9366), 1,
      anon_sym_BANG,
    ACTIONS(9372), 1,
      anon_sym_TILDE,
    ACTIONS(9374), 1,
      anon_sym_DOLLAR,
    ACTIONS(9376), 1,
      anon_sym_DQUOTE,
    ACTIONS(9378), 1,
      aux_sym_number_token1,
    ACTIONS(9380), 1,
      aux_sym_number_token2,
    ACTIONS(9382), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9384), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9386), 1,
      anon_sym_BQUOTE,
    ACTIONS(9388), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(9392), 1,
      sym_variable_name,
    ACTIONS(11238), 1,
      aux_sym__simple_variable_name_token1,
    STATE(3391), 1,
      sym__arithmetic_binary_expression,
    STATE(3396), 1,
      sym__arithmetic_ternary_expression,
    STATE(3398), 1,
      sym__arithmetic_unary_expression,
    STATE(3400), 1,
      sym__arithmetic_postfix_expression,
    ACTIONS(9368), 2,
      anon_sym_PLUS_PLUS2,
      anon_sym_DASH_DASH2,
    ACTIONS(9370), 2,
      anon_sym_DASH2,
      anon_sym_PLUS2,
    STATE(3192), 9,
      sym_subscript,
      sym__arithmetic_expression,
      sym__arithmetic_literal,
      sym__arithmetic_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [239413] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1235), 1,
      sym_file_descriptor,
    ACTIONS(5874), 1,
      anon_sym_DQUOTE,
    ACTIONS(11244), 1,
      sym_variable_name,
    STATE(5696), 1,
      sym_string,
    ACTIONS(11242), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(11240), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1227), 14,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      aux_sym_heredoc_redirect_token1,
      anon_sym_LT_LT_LT,
  [239460] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11250), 1,
      anon_sym_LPAREN,
    ACTIONS(11252), 1,
      aux_sym__c_word_token1,
    ACTIONS(11254), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11256), 1,
      anon_sym_DOLLAR,
    ACTIONS(11258), 1,
      anon_sym_DQUOTE,
    ACTIONS(11260), 1,
      aux_sym_number_token1,
    ACTIONS(11262), 1,
      aux_sym_number_token2,
    ACTIONS(11264), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(11266), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(11268), 1,
      anon_sym_BQUOTE,
    ACTIONS(11270), 1,
      anon_sym_DOLLAR_BQUOTE,
    STATE(3113), 1,
      sym__c_postfix_expression,
    STATE(3116), 1,
      sym__c_binary_expression,
    STATE(3118), 1,
      sym__c_unary_expression,
    STATE(4243), 1,
      sym__c_terminator,
    STATE(6365), 1,
      sym__c_expression,
    STATE(6493), 1,
      sym__c_variable_assignment,
    STATE(7061), 1,
      sym__for_body,
    ACTIONS(11246), 2,
      anon_sym_SEMI,
      anon_sym_AMP,
    ACTIONS(11248), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    STATE(3119), 7,
      sym__c_expression_not_assignment,
      sym__c_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [239535] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11272), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11274), 1,
      sym__concat,
    STATE(4248), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1263), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1261), 24,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      sym__special_character,
  [239578] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11250), 1,
      anon_sym_LPAREN,
    ACTIONS(11252), 1,
      aux_sym__c_word_token1,
    ACTIONS(11254), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11256), 1,
      anon_sym_DOLLAR,
    ACTIONS(11258), 1,
      anon_sym_DQUOTE,
    ACTIONS(11260), 1,
      aux_sym_number_token1,
    ACTIONS(11262), 1,
      aux_sym_number_token2,
    ACTIONS(11264), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(11266), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(11268), 1,
      anon_sym_BQUOTE,
    ACTIONS(11270), 1,
      anon_sym_DOLLAR_BQUOTE,
    STATE(3113), 1,
      sym__c_postfix_expression,
    STATE(3116), 1,
      sym__c_binary_expression,
    STATE(3118), 1,
      sym__c_unary_expression,
    STATE(4243), 1,
      sym__c_terminator,
    STATE(6365), 1,
      sym__c_expression,
    STATE(6493), 1,
      sym__c_variable_assignment,
    STATE(6952), 1,
      sym__for_body,
    ACTIONS(11246), 2,
      anon_sym_SEMI,
      anon_sym_AMP,
    ACTIONS(11248), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    STATE(3119), 7,
      sym__c_expression_not_assignment,
      sym__c_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [239653] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11250), 1,
      anon_sym_LPAREN,
    ACTIONS(11252), 1,
      aux_sym__c_word_token1,
    ACTIONS(11254), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11256), 1,
      anon_sym_DOLLAR,
    ACTIONS(11258), 1,
      anon_sym_DQUOTE,
    ACTIONS(11260), 1,
      aux_sym_number_token1,
    ACTIONS(11262), 1,
      aux_sym_number_token2,
    ACTIONS(11264), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(11266), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(11268), 1,
      anon_sym_BQUOTE,
    ACTIONS(11270), 1,
      anon_sym_DOLLAR_BQUOTE,
    STATE(3113), 1,
      sym__c_postfix_expression,
    STATE(3116), 1,
      sym__c_binary_expression,
    STATE(3118), 1,
      sym__c_unary_expression,
    STATE(4243), 1,
      sym__c_terminator,
    STATE(6365), 1,
      sym__c_expression,
    STATE(6493), 1,
      sym__c_variable_assignment,
    STATE(6886), 1,
      sym__for_body,
    ACTIONS(11246), 2,
      anon_sym_SEMI,
      anon_sym_AMP,
    ACTIONS(11248), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    STATE(3119), 7,
      sym__c_expression_not_assignment,
      sym__c_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [239728] = 17,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8030), 1,
      anon_sym_SLASH,
    ACTIONS(8032), 1,
      anon_sym_PERCENT,
    ACTIONS(8034), 1,
      anon_sym_COLON,
    ACTIONS(8546), 1,
      anon_sym_AT,
    ACTIONS(8922), 1,
      anon_sym_LBRACK,
    STATE(7238), 1,
      sym__expansion_operator,
    STATE(7239), 1,
      sym__expansion_max_length,
    STATE(7240), 1,
      sym__expansion_regex_removal,
    STATE(7241), 1,
      sym__expansion_regex_replacement,
    STATE(7242), 1,
      sym__expansion_regex,
    STATE(7243), 1,
      sym__expansion_expression,
    ACTIONS(8028), 2,
      anon_sym_COMMA,
      anon_sym_CARET,
    ACTIONS(8048), 2,
      anon_sym_COMMA_COMMA,
      anon_sym_CARET_CARET,
    ACTIONS(8036), 3,
      sym__immediate_double_hash,
      anon_sym_POUND,
      anon_sym_PERCENT_PERCENT,
    ACTIONS(8046), 3,
      anon_sym_SLASH_SLASH,
      anon_sym_SLASH_POUND,
      anon_sym_SLASH_PERCENT,
    ACTIONS(8044), 8,
      anon_sym_EQ2,
      anon_sym_COLON_EQ,
      anon_sym_DASH3,
      anon_sym_COLON_DASH,
      anon_sym_PLUS3,
      anon_sym_COLON_PLUS,
      anon_sym_QMARK2,
      anon_sym_COLON_QMARK,
  [239793] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11276), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11278), 1,
      sym__concat,
    STATE(4251), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1263), 3,
      sym_file_descriptor,
      sym_variable_name,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1261), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      sym__special_character,
  [239836] = 13,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4340), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11284), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(11286), 1,
      sym_file_descriptor,
    STATE(4866), 1,
      sym_herestring_redirect,
    ACTIONS(4251), 2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
    ACTIONS(4253), 2,
      anon_sym_PIPE,
      anon_sym_PIPE_AMP,
    ACTIONS(4255), 2,
      anon_sym_LT_LT,
      anon_sym_LT_LT_DASH,
    ACTIONS(4338), 2,
      anon_sym_SEMI,
      anon_sym_AMP,
    ACTIONS(11282), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    STATE(4283), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(2719), 4,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
    ACTIONS(11280), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
  [239893] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11284), 1,
      anon_sym_LT_LT_LT,
    STATE(4866), 1,
      sym_herestring_redirect,
    ACTIONS(4253), 2,
      anon_sym_PIPE,
      anon_sym_PIPE_AMP,
    ACTIONS(4272), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    STATE(4283), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(4270), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [239938] = 13,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4259), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11284), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(11286), 1,
      sym_file_descriptor,
    STATE(4866), 1,
      sym_herestring_redirect,
    ACTIONS(4249), 2,
      anon_sym_SEMI,
      anon_sym_AMP,
    ACTIONS(4251), 2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
    ACTIONS(4253), 2,
      anon_sym_PIPE,
      anon_sym_PIPE_AMP,
    ACTIONS(4255), 2,
      anon_sym_LT_LT,
      anon_sym_LT_LT_DASH,
    ACTIONS(11282), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    STATE(4283), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(4257), 4,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
    ACTIONS(11280), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
  [239995] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11250), 1,
      anon_sym_LPAREN,
    ACTIONS(11252), 1,
      aux_sym__c_word_token1,
    ACTIONS(11254), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11256), 1,
      anon_sym_DOLLAR,
    ACTIONS(11258), 1,
      anon_sym_DQUOTE,
    ACTIONS(11260), 1,
      aux_sym_number_token1,
    ACTIONS(11262), 1,
      aux_sym_number_token2,
    ACTIONS(11264), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(11266), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(11268), 1,
      anon_sym_BQUOTE,
    ACTIONS(11270), 1,
      anon_sym_DOLLAR_BQUOTE,
    STATE(3113), 1,
      sym__c_postfix_expression,
    STATE(3116), 1,
      sym__c_binary_expression,
    STATE(3118), 1,
      sym__c_unary_expression,
    STATE(4243), 1,
      sym__c_terminator,
    STATE(6365), 1,
      sym__c_expression,
    STATE(6493), 1,
      sym__c_variable_assignment,
    STATE(7326), 1,
      sym__for_body,
    ACTIONS(11246), 2,
      anon_sym_SEMI,
      anon_sym_AMP,
    ACTIONS(11248), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    STATE(3119), 7,
      sym__c_expression_not_assignment,
      sym__c_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [240070] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11284), 1,
      anon_sym_LT_LT_LT,
    STATE(4866), 1,
      sym_herestring_redirect,
    ACTIONS(4348), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    STATE(4283), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(4253), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [240113] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1241), 1,
      sym_file_descriptor,
    ACTIONS(5874), 1,
      anon_sym_DQUOTE,
    ACTIONS(11244), 1,
      sym_variable_name,
    STATE(5696), 1,
      sym_string,
    ACTIONS(11242), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(11240), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
    ACTIONS(1239), 14,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      aux_sym_heredoc_redirect_token1,
      anon_sym_LT_LT_LT,
  [240160] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11276), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11278), 1,
      sym__concat,
    STATE(4251), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5069), 3,
      sym_file_descriptor,
      sym_variable_name,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5067), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [240202] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11292), 1,
      sym_variable_name,
    STATE(6726), 1,
      sym_subscript,
    ACTIONS(11290), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    STATE(4230), 2,
      sym_variable_assignment,
      aux_sym_variable_assignments_repeat1,
    ACTIONS(11288), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [240244] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11272), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11274), 1,
      sym__concat,
    STATE(4248), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5697), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5695), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [240286] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11295), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11298), 1,
      sym__concat,
    STATE(4232), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1253), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1251), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [240328] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11301), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11303), 1,
      sym__concat,
    STATE(4309), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1263), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1261), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      sym__special_character,
  [240370] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11305), 1,
      anon_sym_LT_LT_LT,
    STATE(5069), 1,
      sym_herestring_redirect,
    ACTIONS(4348), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    STATE(4455), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(4253), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [240412] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11315), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11317), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(11320), 1,
      sym_file_descriptor,
    ACTIONS(11312), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    STATE(4235), 3,
      sym_file_redirect,
      sym_herestring_redirect,
      aux_sym_redirected_statement_repeat2,
    ACTIONS(11309), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
    ACTIONS(11307), 12,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
  [240458] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11250), 1,
      anon_sym_LPAREN,
    ACTIONS(11252), 1,
      aux_sym__c_word_token1,
    ACTIONS(11256), 1,
      anon_sym_DOLLAR,
    ACTIONS(11258), 1,
      anon_sym_DQUOTE,
    ACTIONS(11260), 1,
      aux_sym_number_token1,
    ACTIONS(11262), 1,
      aux_sym_number_token2,
    ACTIONS(11264), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(11266), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(11268), 1,
      anon_sym_BQUOTE,
    ACTIONS(11270), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(11325), 1,
      aux_sym_heredoc_redirect_token1,
    STATE(3113), 1,
      sym__c_postfix_expression,
    STATE(3116), 1,
      sym__c_binary_expression,
    STATE(3118), 1,
      sym__c_unary_expression,
    STATE(4743), 1,
      sym__c_terminator,
    STATE(6270), 1,
      sym__c_expression,
    STATE(6493), 1,
      sym__c_variable_assignment,
    ACTIONS(11248), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(11323), 2,
      anon_sym_SEMI,
      anon_sym_AMP,
    STATE(3119), 7,
      sym__c_expression_not_assignment,
      sym__c_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [240530] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11331), 1,
      sym_variable_name,
    STATE(6726), 1,
      sym_subscript,
    ACTIONS(11329), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    STATE(4230), 2,
      sym_variable_assignment,
      aux_sym_variable_assignments_repeat1,
    ACTIONS(11327), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [240572] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11276), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11278), 1,
      sym__concat,
    STATE(4258), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5375), 3,
      sym_file_descriptor,
      sym_variable_name,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5373), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [240614] = 13,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4411), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11305), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(11337), 1,
      sym_file_descriptor,
    STATE(5069), 1,
      sym_herestring_redirect,
    ACTIONS(4253), 2,
      anon_sym_PIPE,
      anon_sym_PIPE_AMP,
    ACTIONS(4380), 2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
    ACTIONS(4382), 2,
      anon_sym_LT_LT,
      anon_sym_LT_LT_DASH,
    ACTIONS(4409), 2,
      anon_sym_SEMI,
      anon_sym_AMP,
    ACTIONS(11335), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(4257), 3,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
    STATE(4455), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(11333), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
  [240670] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11272), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11274), 1,
      sym__concat,
    STATE(4245), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(2096), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(2094), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [240712] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11250), 1,
      anon_sym_LPAREN,
    ACTIONS(11252), 1,
      aux_sym__c_word_token1,
    ACTIONS(11256), 1,
      anon_sym_DOLLAR,
    ACTIONS(11258), 1,
      anon_sym_DQUOTE,
    ACTIONS(11260), 1,
      aux_sym_number_token1,
    ACTIONS(11262), 1,
      aux_sym_number_token2,
    ACTIONS(11264), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(11266), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(11268), 1,
      anon_sym_BQUOTE,
    ACTIONS(11270), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(11341), 1,
      aux_sym_heredoc_redirect_token1,
    STATE(3113), 1,
      sym__c_postfix_expression,
    STATE(3116), 1,
      sym__c_binary_expression,
    STATE(3118), 1,
      sym__c_unary_expression,
    STATE(4721), 1,
      sym__c_terminator,
    STATE(6300), 1,
      sym__c_expression,
    STATE(6493), 1,
      sym__c_variable_assignment,
    ACTIONS(11248), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(11339), 2,
      anon_sym_SEMI,
      anon_sym_AMP,
    STATE(3119), 7,
      sym__c_expression_not_assignment,
      sym__c_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [240784] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11272), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11274), 1,
      sym__concat,
    STATE(4245), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4372), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4370), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [240826] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11250), 1,
      anon_sym_LPAREN,
    ACTIONS(11252), 1,
      aux_sym__c_word_token1,
    ACTIONS(11256), 1,
      anon_sym_DOLLAR,
    ACTIONS(11258), 1,
      anon_sym_DQUOTE,
    ACTIONS(11260), 1,
      aux_sym_number_token1,
    ACTIONS(11262), 1,
      aux_sym_number_token2,
    ACTIONS(11264), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(11266), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(11268), 1,
      anon_sym_BQUOTE,
    ACTIONS(11270), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(11345), 1,
      aux_sym_heredoc_redirect_token1,
    STATE(3113), 1,
      sym__c_postfix_expression,
    STATE(3116), 1,
      sym__c_binary_expression,
    STATE(3118), 1,
      sym__c_unary_expression,
    STATE(4757), 1,
      sym__c_terminator,
    STATE(6330), 1,
      sym__c_expression,
    STATE(6493), 1,
      sym__c_variable_assignment,
    ACTIONS(11248), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(11343), 2,
      anon_sym_SEMI,
      anon_sym_AMP,
    STATE(3119), 7,
      sym__c_expression_not_assignment,
      sym__c_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [240898] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11347), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11350), 1,
      sym__concat,
    STATE(4244), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1253), 4,
      sym_file_descriptor,
      sym_variable_name,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1251), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_BQUOTE,
  [240940] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11272), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11353), 1,
      sym__concat,
    STATE(4232), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1288), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1286), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [240982] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11272), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11274), 1,
      sym__concat,
    STATE(4248), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4469), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4467), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [241024] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11272), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11274), 1,
      sym__concat,
    STATE(4245), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(2074), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(2072), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [241066] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11272), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11355), 1,
      sym__concat,
    STATE(4232), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1267), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1265), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [241108] = 13,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4384), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11305), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(11337), 1,
      sym_file_descriptor,
    STATE(5069), 1,
      sym_herestring_redirect,
    ACTIONS(4253), 2,
      anon_sym_PIPE,
      anon_sym_PIPE_AMP,
    ACTIONS(4378), 2,
      anon_sym_SEMI,
      anon_sym_AMP,
    ACTIONS(4380), 2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
    ACTIONS(4382), 2,
      anon_sym_LT_LT,
      anon_sym_LT_LT_DASH,
    ACTIONS(11335), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(2719), 3,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
    STATE(4455), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(11333), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
  [241164] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11363), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11365), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(11367), 1,
      sym_file_descriptor,
    ACTIONS(11361), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    STATE(4235), 3,
      sym_file_redirect,
      sym_herestring_redirect,
      aux_sym_redirected_statement_repeat2,
    ACTIONS(11359), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
    ACTIONS(11357), 12,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
  [241210] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11276), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11369), 1,
      sym__concat,
    STATE(4256), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1267), 3,
      sym_file_descriptor,
      sym_variable_name,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1265), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [241252] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11272), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11274), 1,
      sym__concat,
    STATE(4248), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5723), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5721), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [241294] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11371), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11374), 1,
      sym__concat,
    STATE(4253), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1253), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1251), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_BQUOTE,
  [241336] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11305), 1,
      anon_sym_LT_LT_LT,
    STATE(5069), 1,
      sym_herestring_redirect,
    ACTIONS(4253), 2,
      anon_sym_PIPE,
      anon_sym_PIPE_AMP,
    ACTIONS(4272), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    STATE(4455), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(4270), 19,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [241380] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11377), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11379), 1,
      sym__concat,
    STATE(4380), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1263), 3,
      sym_file_descriptor,
      sym_variable_name,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1261), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      sym__special_character,
  [241422] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11381), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11384), 1,
      sym__concat,
    STATE(4256), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1253), 3,
      sym_file_descriptor,
      sym_variable_name,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1251), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [241464] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11272), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11274), 1,
      sym__concat,
    STATE(4248), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4566), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4564), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [241506] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11276), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11387), 1,
      sym__concat,
    STATE(4256), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1288), 3,
      sym_file_descriptor,
      sym_variable_name,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1286), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [241548] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11389), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11391), 1,
      sym__concat,
    STATE(4304), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1263), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1261), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      sym__special_character,
  [241590] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11272), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11274), 1,
      sym__concat,
    STATE(4245), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4554), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4552), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [241632] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11393), 1,
      sym__special_character,
    STATE(4261), 1,
      aux_sym__literal_repeat1,
    ACTIONS(1364), 4,
      sym_file_descriptor,
      sym_variable_name,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1362), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_BQUOTE,
  [241671] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11396), 1,
      sym__special_character,
    STATE(4324), 1,
      aux_sym__literal_repeat1,
    ACTIONS(5723), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5721), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [241710] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1350), 3,
      sym_file_descriptor,
      sym__concat,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1348), 24,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      aux_sym_concatenation_token1,
  [241745] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1310), 4,
      sym_file_descriptor,
      sym__concat,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1308), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      aux_sym_concatenation_token1,
      anon_sym_BQUOTE,
  [241780] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1298), 3,
      sym_file_descriptor,
      sym__concat,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1296), 24,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      aux_sym_concatenation_token1,
  [241815] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11398), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11401), 1,
      sym__concat,
    STATE(4266), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1253), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1251), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [241856] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1302), 3,
      sym_file_descriptor,
      sym__concat,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1300), 24,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      aux_sym_concatenation_token1,
  [241891] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11404), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11407), 1,
      sym__concat,
    STATE(4268), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1253), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1251), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_BQUOTE,
  [241932] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11410), 1,
      anon_sym_LT_LT_LT,
    STATE(4912), 1,
      sym_herestring_redirect,
    ACTIONS(4348), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    STATE(4697), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(4253), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_BQUOTE,
  [241973] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11389), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11391), 1,
      sym__concat,
    STATE(4297), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4554), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4552), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [242014] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11389), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11391), 1,
      sym__concat,
    STATE(4297), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(2096), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(2094), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [242055] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11389), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11391), 1,
      sym__concat,
    STATE(4304), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4566), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4564), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [242096] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11284), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(11414), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11416), 1,
      sym_file_descriptor,
    ACTIONS(11282), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    STATE(4985), 2,
      sym_file_redirect,
      sym_herestring_redirect,
    ACTIONS(11280), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
    ACTIONS(11412), 12,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
  [242141] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11301), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11303), 1,
      sym__concat,
    STATE(4284), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4372), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4370), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [242182] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1310), 4,
      sym_file_descriptor,
      sym__concat,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1308), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      aux_sym_concatenation_token1,
      anon_sym_BQUOTE,
  [242217] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11301), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11303), 1,
      sym__concat,
    STATE(4309), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4469), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4467), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [242258] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11418), 1,
      sym__special_character,
    STATE(4394), 1,
      aux_sym__literal_repeat1,
    ACTIONS(5723), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5721), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_BQUOTE,
  [242297] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11431), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11433), 1,
      sym_file_descriptor,
    ACTIONS(11425), 2,
      anon_sym_LT_LT,
      anon_sym_LT_LT_DASH,
    ACTIONS(11428), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    STATE(4278), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(11422), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
    ACTIONS(11420), 10,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
  [242342] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11315), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11442), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(11445), 1,
      sym_file_descriptor,
    ACTIONS(11439), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    STATE(4279), 3,
      sym_file_redirect,
      sym_herestring_redirect,
      aux_sym_redirected_statement_repeat2,
    ACTIONS(11436), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
    ACTIONS(11307), 11,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
  [242387] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1294), 4,
      sym_file_descriptor,
      sym__concat,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1292), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      aux_sym_concatenation_token1,
      anon_sym_BQUOTE,
  [242422] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11448), 1,
      anon_sym_LT_LT_LT,
    STATE(4912), 1,
      sym_herestring_redirect,
    ACTIONS(4253), 2,
      anon_sym_PIPE,
      anon_sym_PIPE_AMP,
    ACTIONS(4272), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    STATE(4650), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(4270), 17,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [242465] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11450), 1,
      sym_variable_name,
    STATE(6772), 1,
      sym_subscript,
    ACTIONS(11290), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    STATE(4282), 2,
      sym_variable_assignment,
      aux_sym_variable_assignments_repeat1,
    ACTIONS(11288), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [242506] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11455), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11457), 1,
      sym_file_descriptor,
    ACTIONS(4255), 2,
      anon_sym_LT_LT,
      anon_sym_LT_LT_DASH,
    ACTIONS(11282), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    STATE(4278), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(11280), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
    ACTIONS(11453), 10,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
  [242551] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11301), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11459), 1,
      sym__concat,
    STATE(4266), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1288), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1286), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [242592] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1253), 4,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1251), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [242627] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11461), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11464), 1,
      sym__concat,
    STATE(4286), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1253), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1251), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [242668] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11467), 1,
      sym__special_character,
    STATE(4261), 1,
      aux_sym__literal_repeat1,
    ACTIONS(5069), 4,
      sym_file_descriptor,
      sym_variable_name,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5067), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_BQUOTE,
  [242707] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1294), 5,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1292), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
      anon_sym_BQUOTE,
  [242742] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11410), 1,
      anon_sym_LT_LT_LT,
    STATE(4912), 1,
      sym_herestring_redirect,
    ACTIONS(4253), 2,
      anon_sym_PIPE,
      anon_sym_PIPE_AMP,
    ACTIONS(4272), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    STATE(4697), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(4270), 18,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_BQUOTE,
  [242785] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11396), 1,
      sym__special_character,
    STATE(4324), 1,
      aux_sym__literal_repeat1,
    ACTIONS(4566), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4564), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [242824] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11469), 1,
      sym__concat,
    ACTIONS(6844), 5,
      anon_sym_COMMA,
      anon_sym_CARET,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_COLON,
    ACTIONS(6842), 21,
      sym__immediate_double_hash,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
      anon_sym_POUND,
      anon_sym_RBRACE3,
      anon_sym_AT,
      anon_sym_EQ2,
      anon_sym_COLON_EQ,
      anon_sym_DASH3,
      anon_sym_COLON_DASH,
      anon_sym_PLUS3,
      anon_sym_COLON_PLUS,
      anon_sym_QMARK2,
      anon_sym_COLON_QMARK,
      anon_sym_PERCENT_PERCENT,
      anon_sym_SLASH_SLASH,
      anon_sym_SLASH_POUND,
      anon_sym_SLASH_PERCENT,
      anon_sym_COMMA_COMMA,
      anon_sym_CARET_CARET,
  [242861] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11389), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11391), 1,
      sym__concat,
    STATE(4297), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4372), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4370), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [242902] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11471), 1,
      sym__concat,
    ACTIONS(6838), 5,
      anon_sym_COMMA,
      anon_sym_CARET,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_COLON,
    ACTIONS(6836), 21,
      sym__immediate_double_hash,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
      anon_sym_POUND,
      anon_sym_RBRACE3,
      anon_sym_AT,
      anon_sym_EQ2,
      anon_sym_COLON_EQ,
      anon_sym_DASH3,
      anon_sym_COLON_DASH,
      anon_sym_PLUS3,
      anon_sym_COLON_PLUS,
      anon_sym_QMARK2,
      anon_sym_COLON_QMARK,
      anon_sym_PERCENT_PERCENT,
      anon_sym_SLASH_SLASH,
      anon_sym_SLASH_POUND,
      anon_sym_SLASH_PERCENT,
      anon_sym_COMMA_COMMA,
      anon_sym_CARET_CARET,
  [242939] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11418), 1,
      sym__special_character,
    STATE(4394), 1,
      aux_sym__literal_repeat1,
    ACTIONS(4469), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4467), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_BQUOTE,
  [242978] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1334), 3,
      sym_file_descriptor,
      sym__concat,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1332), 24,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      aux_sym_concatenation_token1,
  [243013] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11389), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11391), 1,
      sym__concat,
    STATE(4304), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4469), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4467), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [243054] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11389), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11473), 1,
      sym__concat,
    STATE(4286), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1288), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1286), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [243095] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1322), 3,
      sym_file_descriptor,
      sym__concat,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1320), 24,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      aux_sym_concatenation_token1,
  [243130] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11301), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11303), 1,
      sym__concat,
    STATE(4309), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5723), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5721), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [243171] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1338), 3,
      sym_file_descriptor,
      sym__concat,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1336), 24,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      aux_sym_concatenation_token1,
  [243206] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1342), 3,
      sym_file_descriptor,
      sym__concat,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1340), 24,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      aux_sym_concatenation_token1,
  [243241] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1314), 3,
      sym_file_descriptor,
      sym__concat,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1312), 24,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      aux_sym_concatenation_token1,
  [243276] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1326), 4,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1324), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [243311] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11389), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11475), 1,
      sym__concat,
    STATE(4286), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1267), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1265), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [243352] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11418), 1,
      sym__special_character,
    STATE(4394), 1,
      aux_sym__literal_repeat1,
    ACTIONS(5697), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5695), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_BQUOTE,
  [243391] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1354), 3,
      sym_file_descriptor,
      sym__concat,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1352), 24,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      aux_sym_concatenation_token1,
  [243426] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1346), 3,
      sym_file_descriptor,
      sym__concat,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1344), 24,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      aux_sym_concatenation_token1,
  [243461] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1358), 3,
      sym_file_descriptor,
      sym__concat,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1356), 24,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      aux_sym_concatenation_token1,
  [243496] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11301), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11477), 1,
      sym__concat,
    STATE(4266), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1267), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1265), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [243537] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11301), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11303), 1,
      sym__concat,
    STATE(4284), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(2096), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(2094), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [243578] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1314), 4,
      sym_file_descriptor,
      sym__concat,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1312), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      aux_sym_concatenation_token1,
      anon_sym_BQUOTE,
  [243613] = 13,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2721), 1,
      ts_builtin_sym_end,
    ACTIONS(5051), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11448), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(11483), 1,
      sym_file_descriptor,
    STATE(4912), 1,
      sym_herestring_redirect,
    ACTIONS(4253), 2,
      anon_sym_PIPE,
      anon_sym_PIPE_AMP,
    ACTIONS(4690), 2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
    ACTIONS(4692), 2,
      anon_sym_LT_LT,
      anon_sym_LT_LT_DASH,
    ACTIONS(11481), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(5049), 3,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
    STATE(4650), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(11479), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
  [243668] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1334), 5,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1332), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
      anon_sym_BQUOTE,
  [243703] = 13,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4257), 1,
      anon_sym_BQUOTE,
    ACTIONS(5379), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11410), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(11489), 1,
      sym_file_descriptor,
    STATE(4912), 1,
      sym_herestring_redirect,
    ACTIONS(4253), 2,
      anon_sym_PIPE,
      anon_sym_PIPE_AMP,
    ACTIONS(4692), 2,
      anon_sym_LT_LT,
      anon_sym_LT_LT_DASH,
    ACTIONS(5355), 2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
    ACTIONS(11487), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(5377), 3,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
    STATE(4697), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(11485), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
  [243758] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11491), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11493), 1,
      sym__concat,
    STATE(4471), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1263), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1261), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      sym__special_character,
  [243799] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1322), 5,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1320), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
      anon_sym_BQUOTE,
  [243834] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1294), 3,
      sym_file_descriptor,
      sym__concat,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1292), 24,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      aux_sym_concatenation_token1,
  [243869] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1338), 5,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1336), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
      anon_sym_BQUOTE,
  [243904] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11301), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11303), 1,
      sym__concat,
    STATE(4309), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5697), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5695), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [243945] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1330), 5,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1328), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
      anon_sym_BQUOTE,
  [243980] = 13,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4686), 1,
      ts_builtin_sym_end,
    ACTIONS(4694), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11448), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(11483), 1,
      sym_file_descriptor,
    STATE(4912), 1,
      sym_herestring_redirect,
    ACTIONS(4253), 2,
      anon_sym_PIPE,
      anon_sym_PIPE_AMP,
    ACTIONS(4690), 2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
    ACTIONS(4692), 2,
      anon_sym_LT_LT,
      anon_sym_LT_LT_DASH,
    ACTIONS(11481), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(4688), 3,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
    STATE(4650), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(11479), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
  [244035] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1350), 4,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1348), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [244070] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1322), 4,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1320), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [244105] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11495), 1,
      sym__special_character,
    STATE(4324), 1,
      aux_sym__literal_repeat1,
    ACTIONS(1364), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1362), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [244144] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1338), 4,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1336), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [244179] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1342), 4,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1340), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [244214] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1314), 4,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1312), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [244249] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1310), 3,
      sym_file_descriptor,
      sym__concat,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1308), 24,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      aux_sym_concatenation_token1,
  [244284] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1306), 3,
      sym_file_descriptor,
      sym__concat,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1304), 24,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      aux_sym_concatenation_token1,
  [244319] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1310), 3,
      sym_file_descriptor,
      sym__concat,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1308), 24,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      aux_sym_concatenation_token1,
  [244354] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11301), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11303), 1,
      sym__concat,
    STATE(4284), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(2074), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(2072), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [244395] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1354), 4,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1352), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [244430] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1342), 5,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1340), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
      anon_sym_BQUOTE,
  [244465] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1314), 5,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1312), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
      anon_sym_BQUOTE,
  [244500] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1298), 4,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1296), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [244535] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1302), 4,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1300), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [244570] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1253), 4,
      sym_file_descriptor,
      sym__concat,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1251), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      aux_sym_concatenation_token1,
      anon_sym_BQUOTE,
  [244605] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1326), 4,
      sym_file_descriptor,
      sym__concat,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1324), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      aux_sym_concatenation_token1,
      anon_sym_BQUOTE,
  [244640] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4272), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    STATE(4283), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(4270), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [244677] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1334), 4,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1332), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [244712] = 11,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4340), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11457), 1,
      sym_file_descriptor,
    ACTIONS(4251), 2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
    ACTIONS(4253), 2,
      anon_sym_PIPE,
      anon_sym_PIPE_AMP,
    ACTIONS(4255), 2,
      anon_sym_LT_LT,
      anon_sym_LT_LT_DASH,
    ACTIONS(4338), 2,
      anon_sym_SEMI,
      anon_sym_AMP,
    ACTIONS(11282), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    STATE(4283), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(2719), 4,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
    ACTIONS(11280), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
  [244763] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4253), 2,
      anon_sym_PIPE,
      anon_sym_PIPE_AMP,
    ACTIONS(4272), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    STATE(4283), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(4270), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [244802] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1294), 4,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1292), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [244837] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11498), 1,
      sym__concat,
    ACTIONS(6820), 5,
      anon_sym_COMMA,
      anon_sym_CARET,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_COLON,
    ACTIONS(6818), 21,
      sym__immediate_double_hash,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
      anon_sym_POUND,
      anon_sym_RBRACE3,
      anon_sym_AT,
      anon_sym_EQ2,
      anon_sym_COLON_EQ,
      anon_sym_DASH3,
      anon_sym_COLON_DASH,
      anon_sym_PLUS3,
      anon_sym_COLON_PLUS,
      anon_sym_QMARK2,
      anon_sym_COLON_QMARK,
      anon_sym_PERCENT_PERCENT,
      anon_sym_SLASH_SLASH,
      anon_sym_SLASH_POUND,
      anon_sym_SLASH_PERCENT,
      anon_sym_COMMA_COMMA,
      anon_sym_CARET_CARET,
  [244874] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11500), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11502), 1,
      sym__concat,
    STATE(4594), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1263), 3,
      sym_file_descriptor,
      sym_variable_name,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1261), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      sym__special_character,
      anon_sym_BQUOTE,
  [244915] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11284), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(11416), 1,
      sym_file_descriptor,
    ACTIONS(11506), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11282), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    STATE(4949), 2,
      sym_file_redirect,
      sym_herestring_redirect,
    ACTIONS(11280), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
    ACTIONS(11504), 12,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
  [244960] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11418), 1,
      sym__special_character,
    STATE(4394), 1,
      aux_sym__literal_repeat1,
    ACTIONS(4566), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4564), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_BQUOTE,
  [244999] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1338), 4,
      sym_file_descriptor,
      sym__concat,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1336), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      aux_sym_concatenation_token1,
      anon_sym_BQUOTE,
  [245034] = 11,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4259), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11457), 1,
      sym_file_descriptor,
    ACTIONS(4249), 2,
      anon_sym_SEMI,
      anon_sym_AMP,
    ACTIONS(4251), 2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
    ACTIONS(4253), 2,
      anon_sym_PIPE,
      anon_sym_PIPE_AMP,
    ACTIONS(4255), 2,
      anon_sym_LT_LT,
      anon_sym_LT_LT_DASH,
    ACTIONS(11282), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    STATE(4283), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(4257), 4,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
    ACTIONS(11280), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
  [245085] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11500), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11502), 1,
      sym__concat,
    STATE(4511), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1263), 4,
      sym_file_descriptor,
      sym_variable_name,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1261), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      sym__special_character,
  [245126] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11396), 1,
      sym__special_character,
    STATE(4324), 1,
      aux_sym__literal_repeat1,
    ACTIONS(5697), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5695), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [245165] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11301), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11303), 1,
      sym__concat,
    STATE(4309), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4566), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4564), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [245206] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11508), 1,
      anon_sym_LT_LT_LT,
    STATE(4912), 1,
      sym_herestring_redirect,
    ACTIONS(4348), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    STATE(4746), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(4253), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [245247] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1310), 4,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1308), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [245282] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1306), 4,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1304), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [245317] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1310), 4,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1308), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [245352] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1253), 5,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1251), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
      anon_sym_BQUOTE,
  [245387] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11510), 1,
      sym_variable_name,
    STATE(6772), 1,
      sym_subscript,
    ACTIONS(11329), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    STATE(4282), 2,
      sym_variable_assignment,
      aux_sym_variable_assignments_repeat1,
    ACTIONS(11327), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [245428] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11512), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11514), 1,
      sym__concat,
    STATE(4592), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1263), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1261), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      sym__special_character,
  [245469] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1358), 5,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1356), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
      anon_sym_BQUOTE,
  [245504] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11500), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11502), 1,
      sym__concat,
    STATE(4586), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1263), 3,
      sym_file_descriptor,
      sym_variable_name,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1261), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      sym__special_character,
  [245545] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11512), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11514), 1,
      sym__concat,
    STATE(4583), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1263), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1261), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      sym__special_character,
      anon_sym_BQUOTE,
  [245586] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11516), 1,
      sym__concat,
    ACTIONS(6829), 5,
      anon_sym_COMMA,
      anon_sym_CARET,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_COLON,
    ACTIONS(6827), 21,
      sym__immediate_double_hash,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
      anon_sym_POUND,
      anon_sym_RBRACE3,
      anon_sym_AT,
      anon_sym_EQ2,
      anon_sym_COLON_EQ,
      anon_sym_DASH3,
      anon_sym_COLON_DASH,
      anon_sym_PLUS3,
      anon_sym_COLON_PLUS,
      anon_sym_QMARK2,
      anon_sym_COLON_QMARK,
      anon_sym_PERCENT_PERCENT,
      anon_sym_SLASH_SLASH,
      anon_sym_SLASH_POUND,
      anon_sym_SLASH_PERCENT,
      anon_sym_COMMA_COMMA,
      anon_sym_CARET_CARET,
  [245623] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11389), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11391), 1,
      sym__concat,
    STATE(4304), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5697), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5695), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [245664] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1310), 5,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1308), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
      anon_sym_BQUOTE,
  [245699] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1346), 4,
      sym_file_descriptor,
      sym__concat,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1344), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      aux_sym_concatenation_token1,
      anon_sym_BQUOTE,
  [245734] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1358), 4,
      sym_file_descriptor,
      sym__concat,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1356), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      aux_sym_concatenation_token1,
      anon_sym_BQUOTE,
  [245769] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11518), 1,
      sym__special_character,
    STATE(4368), 1,
      aux_sym__literal_repeat1,
    ACTIONS(1364), 3,
      sym_file_descriptor,
      sym_variable_name,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1362), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [245808] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1298), 4,
      sym_file_descriptor,
      sym__concat,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1296), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      aux_sym_concatenation_token1,
      anon_sym_BQUOTE,
  [245843] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11389), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11391), 1,
      sym__concat,
    STATE(4297), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(2074), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(2072), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [245884] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11363), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11525), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(11527), 1,
      sym_file_descriptor,
    ACTIONS(11523), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    STATE(4279), 3,
      sym_file_redirect,
      sym_herestring_redirect,
      aux_sym_redirected_statement_repeat2,
    ACTIONS(11521), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
    ACTIONS(11357), 11,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
  [245929] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1354), 4,
      sym_file_descriptor,
      sym__concat,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1352), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      aux_sym_concatenation_token1,
      anon_sym_BQUOTE,
  [245964] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1346), 5,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1344), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
      anon_sym_BQUOTE,
  [245999] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1310), 5,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1308), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
      anon_sym_BQUOTE,
  [246034] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11301), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11303), 1,
      sym__concat,
    STATE(4284), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4554), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4552), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [246075] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11508), 1,
      anon_sym_LT_LT_LT,
    STATE(4912), 1,
      sym_herestring_redirect,
    ACTIONS(4253), 2,
      anon_sym_PIPE,
      anon_sym_PIPE_AMP,
    ACTIONS(4272), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    STATE(4746), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(4270), 18,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [246118] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1318), 5,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1316), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
      anon_sym_BQUOTE,
  [246153] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1354), 5,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1352), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
      anon_sym_BQUOTE,
  [246188] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1342), 4,
      sym_file_descriptor,
      sym__concat,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1340), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      aux_sym_concatenation_token1,
      anon_sym_BQUOTE,
  [246223] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11377), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11529), 1,
      sym__concat,
    STATE(4413), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1267), 3,
      sym_file_descriptor,
      sym_variable_name,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1265), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [246264] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11448), 1,
      anon_sym_LT_LT_LT,
    STATE(4912), 1,
      sym_herestring_redirect,
    ACTIONS(4348), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    STATE(4650), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(4253), 19,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [246305] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11396), 1,
      sym__special_character,
    STATE(4324), 1,
      aux_sym__literal_repeat1,
    ACTIONS(4469), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4467), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [246344] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11531), 1,
      sym__special_character,
    STATE(4368), 1,
      aux_sym__literal_repeat1,
    ACTIONS(5069), 3,
      sym_file_descriptor,
      sym_variable_name,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5067), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [246383] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1346), 4,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1344), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [246418] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1358), 4,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1356), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [246453] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1330), 4,
      sym_file_descriptor,
      sym__concat,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1328), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      aux_sym_concatenation_token1,
      anon_sym_BQUOTE,
  [246488] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11284), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(11416), 1,
      sym_file_descriptor,
    ACTIONS(11535), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11282), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    STATE(4889), 2,
      sym_file_redirect,
      sym_herestring_redirect,
    ACTIONS(11280), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
    ACTIONS(11533), 12,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
  [246533] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1330), 3,
      sym_file_descriptor,
      sym__concat,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1328), 24,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      aux_sym_concatenation_token1,
  [246568] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1350), 4,
      sym_file_descriptor,
      sym__concat,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1348), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      aux_sym_concatenation_token1,
      anon_sym_BQUOTE,
  [246603] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11377), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11379), 1,
      sym__concat,
    STATE(4380), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5069), 3,
      sym_file_descriptor,
      sym_variable_name,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5067), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [246644] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1322), 4,
      sym_file_descriptor,
      sym__concat,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1320), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      aux_sym_concatenation_token1,
      anon_sym_BQUOTE,
  [246679] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1334), 4,
      sym_file_descriptor,
      sym__concat,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1332), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      aux_sym_concatenation_token1,
      anon_sym_BQUOTE,
  [246714] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1350), 5,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1348), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
      anon_sym_BQUOTE,
  [246749] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11537), 1,
      sym__special_character,
    STATE(4394), 1,
      aux_sym__literal_repeat1,
    ACTIONS(1364), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1362), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_BQUOTE,
  [246788] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1318), 4,
      sym_file_descriptor,
      sym__concat,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1316), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      aux_sym_concatenation_token1,
      anon_sym_BQUOTE,
  [246823] = 13,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2719), 1,
      anon_sym_RPAREN,
    ACTIONS(4384), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11508), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(11544), 1,
      sym_file_descriptor,
    STATE(4912), 1,
      sym_herestring_redirect,
    ACTIONS(4253), 2,
      anon_sym_PIPE,
      anon_sym_PIPE_AMP,
    ACTIONS(4692), 2,
      anon_sym_LT_LT,
      anon_sym_LT_LT_DASH,
    ACTIONS(4809), 2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
    ACTIONS(11542), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(4378), 3,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
    STATE(4746), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(11540), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
  [246878] = 13,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2719), 1,
      anon_sym_BQUOTE,
    ACTIONS(5357), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11410), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(11489), 1,
      sym_file_descriptor,
    STATE(4912), 1,
      sym_herestring_redirect,
    ACTIONS(4253), 2,
      anon_sym_PIPE,
      anon_sym_PIPE_AMP,
    ACTIONS(4692), 2,
      anon_sym_LT_LT,
      anon_sym_LT_LT_DASH,
    ACTIONS(5355), 2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
    ACTIONS(11487), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(5353), 3,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
    STATE(4697), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(11485), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
  [246933] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1326), 5,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1324), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
      anon_sym_BQUOTE,
  [246968] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1318), 3,
      sym_file_descriptor,
      sym__concat,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1316), 24,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      aux_sym_concatenation_token1,
  [247003] = 13,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4257), 1,
      anon_sym_RPAREN,
    ACTIONS(4411), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11508), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(11544), 1,
      sym_file_descriptor,
    STATE(4912), 1,
      sym_herestring_redirect,
    ACTIONS(4253), 2,
      anon_sym_PIPE,
      anon_sym_PIPE_AMP,
    ACTIONS(4692), 2,
      anon_sym_LT_LT,
      anon_sym_LT_LT_DASH,
    ACTIONS(4809), 2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
    ACTIONS(11542), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(4409), 3,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
    STATE(4746), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(11540), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
  [247058] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11512), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11514), 1,
      sym__concat,
    STATE(4523), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1263), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1261), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      sym__special_character,
  [247099] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1306), 5,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1304), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
      anon_sym_BQUOTE,
  [247134] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1302), 4,
      sym_file_descriptor,
      sym__concat,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1300), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      aux_sym_concatenation_token1,
      anon_sym_BQUOTE,
  [247169] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1253), 3,
      sym_file_descriptor,
      sym__concat,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1251), 24,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      aux_sym_concatenation_token1,
  [247204] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1330), 4,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1328), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [247239] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11389), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11391), 1,
      sym__concat,
    STATE(4304), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5723), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5721), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [247280] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1298), 5,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1296), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
      anon_sym_BQUOTE,
  [247315] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1302), 5,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1300), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
      anon_sym_BQUOTE,
  [247350] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1326), 3,
      sym_file_descriptor,
      sym__concat,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1324), 24,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      aux_sym_concatenation_token1,
  [247385] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11377), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11379), 1,
      sym__concat,
    STATE(4414), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5375), 3,
      sym_file_descriptor,
      sym_variable_name,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5373), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [247426] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1306), 4,
      sym_file_descriptor,
      sym__concat,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1304), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      aux_sym_concatenation_token1,
      anon_sym_BQUOTE,
  [247461] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1318), 4,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1316), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [247496] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11546), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11549), 1,
      sym__concat,
    STATE(4413), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1253), 3,
      sym_file_descriptor,
      sym_variable_name,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1251), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [247537] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11377), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11552), 1,
      sym__concat,
    STATE(4413), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1288), 3,
      sym_file_descriptor,
      sym_variable_name,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1286), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [247578] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5142), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11558), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(11560), 1,
      sym_file_descriptor,
    STATE(4912), 1,
      sym_herestring_redirect,
    ACTIONS(4253), 2,
      anon_sym_PIPE,
      anon_sym_PIPE_AMP,
    ACTIONS(4692), 2,
      anon_sym_LT_LT,
      anon_sym_LT_LT_DASH,
    ACTIONS(5099), 2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
    ACTIONS(11556), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(5140), 3,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
    STATE(4986), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(11554), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
  [247630] = 11,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4411), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11562), 1,
      sym_file_descriptor,
    ACTIONS(4253), 2,
      anon_sym_PIPE,
      anon_sym_PIPE_AMP,
    ACTIONS(4380), 2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
    ACTIONS(4382), 2,
      anon_sym_LT_LT,
      anon_sym_LT_LT_DASH,
    ACTIONS(4409), 2,
      anon_sym_SEMI,
      anon_sym_AMP,
    ACTIONS(11335), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(4257), 3,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
    STATE(4455), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(11333), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
  [247680] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1302), 3,
      sym_file_descriptor,
      sym__concat,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1300), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [247714] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11512), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11514), 1,
      sym__concat,
    STATE(4592), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4469), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4467), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [247754] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1326), 4,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1324), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [247788] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11305), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(11535), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11564), 1,
      sym_file_descriptor,
    ACTIONS(11335), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    STATE(5043), 2,
      sym_file_redirect,
      sym_herestring_redirect,
    ACTIONS(11333), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
    ACTIONS(11533), 11,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
  [247832] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11566), 1,
      sym__special_character,
    STATE(4485), 1,
      aux_sym__literal_repeat1,
    ACTIONS(4566), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4564), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_BQUOTE,
  [247870] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11568), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11570), 1,
      sym__concat,
    STATE(4520), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1267), 3,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1265), 20,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [247910] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1263), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1261), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      sym__special_character,
      anon_sym_BQUOTE,
  [247944] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1322), 3,
      sym_file_descriptor,
      sym__concat,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1320), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      aux_sym_concatenation_token1,
  [247978] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1338), 3,
      sym_file_descriptor,
      sym__concat,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1336), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      aux_sym_concatenation_token1,
  [248012] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1342), 3,
      sym_file_descriptor,
      sym__concat,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1340), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      aux_sym_concatenation_token1,
  [248046] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1314), 3,
      sym_file_descriptor,
      sym__concat,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1312), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      aux_sym_concatenation_token1,
  [248080] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4272), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    STATE(4455), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(4270), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [248116] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1326), 4,
      sym_file_descriptor,
      sym__concat,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1324), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
      anon_sym_BQUOTE,
  [248150] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11491), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11493), 1,
      sym__concat,
    STATE(4468), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4554), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4552), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [248190] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11572), 1,
      sym__special_character,
    STATE(4533), 1,
      aux_sym__literal_repeat1,
    ACTIONS(5069), 3,
      sym_file_descriptor,
      sym_variable_name,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5067), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [248228] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1354), 3,
      sym_file_descriptor,
      sym__concat,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1352), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      aux_sym_concatenation_token1,
  [248262] = 11,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4384), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11562), 1,
      sym_file_descriptor,
    ACTIONS(4253), 2,
      anon_sym_PIPE,
      anon_sym_PIPE_AMP,
    ACTIONS(4378), 2,
      anon_sym_SEMI,
      anon_sym_AMP,
    ACTIONS(4380), 2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
    ACTIONS(4382), 2,
      anon_sym_LT_LT,
      anon_sym_LT_LT_DASH,
    ACTIONS(11335), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(2719), 3,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
    STATE(4455), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(11333), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
  [248312] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1322), 4,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1320), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [248346] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1338), 4,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1336), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [248380] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1342), 4,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1340), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [248414] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1314), 4,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1312), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [248448] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1298), 3,
      sym_file_descriptor,
      sym__concat,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1296), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      aux_sym_concatenation_token1,
  [248482] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11574), 1,
      sym_variable_name,
    STATE(6786), 1,
      sym_subscript,
    ACTIONS(11290), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    STATE(4439), 2,
      sym_variable_assignment,
      aux_sym_variable_assignments_repeat1,
    ACTIONS(11288), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [248522] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11512), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11514), 1,
      sym__concat,
    STATE(4591), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4372), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4370), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [248562] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1354), 4,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1352), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [248596] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1298), 3,
      sym_file_descriptor,
      sym__concat,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1296), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [248630] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1310), 3,
      sym_file_descriptor,
      sym__concat,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1308), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      aux_sym_concatenation_token1,
  [248664] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1306), 3,
      sym_file_descriptor,
      sym__concat,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1304), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      aux_sym_concatenation_token1,
  [248698] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1310), 3,
      sym_file_descriptor,
      sym__concat,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1308), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      aux_sym_concatenation_token1,
  [248732] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1350), 3,
      sym_file_descriptor,
      sym__concat,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1348), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      aux_sym_concatenation_token1,
  [248766] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1310), 4,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1308), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [248800] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1306), 4,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1304), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [248834] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1310), 4,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1308), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [248868] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11512), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11514), 1,
      sym__concat,
    STATE(4592), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4566), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4564), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [248908] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1318), 4,
      sym_file_descriptor,
      sym__concat,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1316), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
      anon_sym_BQUOTE,
  [248942] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11512), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11514), 1,
      sym__concat,
    STATE(4591), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4554), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4552), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [248982] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11577), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11580), 1,
      sym__concat,
    STATE(4453), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1253), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1251), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [249022] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1302), 3,
      sym_file_descriptor,
      sym__concat,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1300), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      aux_sym_concatenation_token1,
  [249056] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11455), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11562), 1,
      sym_file_descriptor,
    ACTIONS(4382), 2,
      anon_sym_LT_LT,
      anon_sym_LT_LT_DASH,
    ACTIONS(11335), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    STATE(4465), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(11333), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
    ACTIONS(11453), 9,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
  [249100] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1314), 4,
      sym_file_descriptor,
      sym__concat,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1312), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
      anon_sym_BQUOTE,
  [249134] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1350), 4,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1348), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [249168] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11568), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11583), 1,
      sym__concat,
    STATE(4422), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5395), 3,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5393), 20,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [249208] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11512), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11514), 1,
      sym__concat,
    STATE(4583), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4566), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4564), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_BQUOTE,
  [249248] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11568), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11583), 1,
      sym__concat,
    STATE(4558), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5391), 3,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5389), 20,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [249288] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1334), 3,
      sym_file_descriptor,
      sym__concat,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1332), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [249322] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11585), 1,
      sym__special_character,
    STATE(4483), 1,
      aux_sym__literal_repeat1,
    ACTIONS(4566), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4564), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [249360] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11512), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11514), 1,
      sym__concat,
    STATE(4584), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4372), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4370), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_BQUOTE,
  [249400] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1334), 3,
      sym_file_descriptor,
      sym__concat,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1332), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      aux_sym_concatenation_token1,
  [249434] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11431), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11596), 1,
      sym_file_descriptor,
    ACTIONS(11590), 2,
      anon_sym_LT_LT,
      anon_sym_LT_LT_DASH,
    ACTIONS(11593), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    STATE(4465), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(11587), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
    ACTIONS(11420), 9,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
  [249478] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11512), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11514), 1,
      sym__concat,
    STATE(4583), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4469), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4467), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_BQUOTE,
  [249518] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11512), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11514), 1,
      sym__concat,
    STATE(4523), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4469), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4467), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [249558] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11491), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11599), 1,
      sym__concat,
    STATE(4453), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1288), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1286), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [249598] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1346), 3,
      sym_file_descriptor,
      sym__concat,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1344), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [249632] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1358), 3,
      sym_file_descriptor,
      sym__concat,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1356), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [249666] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11491), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11601), 1,
      sym__concat,
    STATE(4453), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1267), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1265), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [249706] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11512), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11514), 1,
      sym__concat,
    STATE(4584), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(2074), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(2072), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_BQUOTE,
  [249746] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1334), 4,
      sym_file_descriptor,
      sym__concat,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1332), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
      anon_sym_BQUOTE,
  [249780] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1294), 3,
      sym_file_descriptor,
      sym__concat,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1292), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [249814] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1322), 4,
      sym_file_descriptor,
      sym__concat,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1320), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
      anon_sym_BQUOTE,
  [249848] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11512), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11514), 1,
      sym__concat,
    STATE(4583), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5697), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5695), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_BQUOTE,
  [249888] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11512), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11514), 1,
      sym__concat,
    STATE(4510), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4372), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4370), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [249928] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1298), 4,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1296), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [249962] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11512), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11514), 1,
      sym__concat,
    STATE(4584), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(2096), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(2094), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_BQUOTE,
  [250002] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11363), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11607), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(11609), 1,
      sym_file_descriptor,
    ACTIONS(11605), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    STATE(4517), 3,
      sym_file_redirect,
      sym_herestring_redirect,
      aux_sym_redirected_statement_repeat2,
    ACTIONS(11603), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
    ACTIONS(11357), 10,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
  [250046] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11512), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11514), 1,
      sym__concat,
    STATE(4592), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5697), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5695), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [250086] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1294), 3,
      sym_file_descriptor,
      sym__concat,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1292), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      aux_sym_concatenation_token1,
  [250120] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11611), 1,
      sym__special_character,
    STATE(4483), 1,
      aux_sym__literal_repeat1,
    ACTIONS(1364), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1362), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [250158] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11512), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11514), 1,
      sym__concat,
    STATE(4523), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5723), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5721), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [250198] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11614), 1,
      sym__special_character,
    STATE(4485), 1,
      aux_sym__literal_repeat1,
    ACTIONS(1364), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1362), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_BQUOTE,
  [250236] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11512), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11514), 1,
      sym__concat,
    STATE(4510), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(2096), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(2094), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [250276] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11500), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11502), 1,
      sym__concat,
    STATE(4781), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1263), 3,
      sym_file_descriptor,
      sym_variable_name,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1261), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      sym__special_character,
  [250316] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11512), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11514), 1,
      sym__concat,
    STATE(4583), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5723), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5721), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_BQUOTE,
  [250356] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5101), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11558), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(11560), 1,
      sym_file_descriptor,
    STATE(4912), 1,
      sym_herestring_redirect,
    ACTIONS(4253), 2,
      anon_sym_PIPE,
      anon_sym_PIPE_AMP,
    ACTIONS(4692), 2,
      anon_sym_LT_LT,
      anon_sym_LT_LT_DASH,
    ACTIONS(5099), 2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
    ACTIONS(11556), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(5097), 3,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
    STATE(4986), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(11554), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
  [250408] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1302), 4,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1300), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [250442] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11568), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11583), 1,
      sym__concat,
    STATE(4422), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1263), 3,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1261), 20,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [250482] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11617), 1,
      sym__special_character,
    STATE(4498), 1,
      aux_sym__literal_repeat1,
    ACTIONS(4469), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4467), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [250520] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1253), 3,
      sym_file_descriptor,
      sym__concat,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1251), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      aux_sym_concatenation_token1,
  [250554] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1326), 3,
      sym_file_descriptor,
      sym__concat,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1324), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      aux_sym_concatenation_token1,
  [250588] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11512), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11514), 1,
      sym__concat,
    STATE(4776), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1263), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1261), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      sym__special_character,
  [250628] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1322), 3,
      sym_file_descriptor,
      sym__concat,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1320), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [250662] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1338), 3,
      sym_file_descriptor,
      sym__concat,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1336), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [250696] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11619), 1,
      sym__special_character,
    STATE(4498), 1,
      aux_sym__literal_repeat1,
    ACTIONS(1364), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1362), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [250734] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1334), 4,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1332), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [250768] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11566), 1,
      sym__special_character,
    STATE(4485), 1,
      aux_sym__literal_repeat1,
    ACTIONS(4469), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4467), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_BQUOTE,
  [250806] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1342), 3,
      sym_file_descriptor,
      sym__concat,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1340), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [250840] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1314), 3,
      sym_file_descriptor,
      sym__concat,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1312), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [250874] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1263), 4,
      sym_file_descriptor,
      sym_variable_name,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1261), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      sym__special_character,
      anon_sym_BQUOTE,
  [250908] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11491), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11493), 1,
      sym__concat,
    STATE(4468), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(2074), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(2072), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [250948] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11500), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11622), 1,
      sym__concat,
    STATE(4244), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1288), 4,
      sym_file_descriptor,
      sym_variable_name,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1286), 19,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [250988] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11491), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11493), 1,
      sym__concat,
    STATE(4471), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5697), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5695), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [251028] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11585), 1,
      sym__special_character,
    STATE(4483), 1,
      aux_sym__literal_repeat1,
    ACTIONS(5723), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5721), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [251066] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1294), 4,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1292), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [251100] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(6829), 5,
      anon_sym_COMMA,
      anon_sym_CARET,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_COLON,
    ACTIONS(6827), 21,
      sym__immediate_double_hash,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
      anon_sym_POUND,
      anon_sym_RBRACE3,
      anon_sym_AT,
      anon_sym_EQ2,
      anon_sym_COLON_EQ,
      anon_sym_DASH3,
      anon_sym_COLON_DASH,
      anon_sym_PLUS3,
      anon_sym_COLON_PLUS,
      anon_sym_QMARK2,
      anon_sym_COLON_QMARK,
      anon_sym_PERCENT_PERCENT,
      anon_sym_SLASH_SLASH,
      anon_sym_SLASH_POUND,
      anon_sym_SLASH_PERCENT,
      anon_sym_COMMA_COMMA,
      anon_sym_CARET_CARET,
  [251134] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11512), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11624), 1,
      sym__concat,
    STATE(4253), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1288), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1286), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [251174] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11500), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11626), 1,
      sym__concat,
    STATE(4244), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1267), 4,
      sym_file_descriptor,
      sym_variable_name,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1265), 19,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [251214] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1350), 3,
      sym_file_descriptor,
      sym__concat,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1348), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [251248] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11512), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11514), 1,
      sym__concat,
    STATE(4584), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4554), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4552), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_BQUOTE,
  [251288] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11500), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11502), 1,
      sym__concat,
    STATE(4511), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5069), 4,
      sym_file_descriptor,
      sym_variable_name,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5067), 19,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [251328] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1338), 4,
      sym_file_descriptor,
      sym__concat,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1336), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
      anon_sym_BQUOTE,
  [251362] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1354), 3,
      sym_file_descriptor,
      sym__concat,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1352), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [251396] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11315), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11634), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(11637), 1,
      sym_file_descriptor,
    ACTIONS(11631), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    STATE(4517), 3,
      sym_file_redirect,
      sym_herestring_redirect,
      aux_sym_redirected_statement_repeat2,
    ACTIONS(11628), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
    ACTIONS(11307), 10,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
  [251440] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11512), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11514), 1,
      sym__concat,
    STATE(4510), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(2074), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(2072), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [251480] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11640), 1,
      sym_variable_name,
    STATE(6757), 1,
      sym_subscript,
    STATE(4519), 2,
      sym_variable_assignment,
      aux_sym_variable_assignments_repeat1,
    ACTIONS(11290), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11288), 19,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [251520] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11643), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11646), 1,
      sym__concat,
    STATE(4520), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1253), 3,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1251), 20,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [251560] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11566), 1,
      sym__special_character,
    STATE(4485), 1,
      aux_sym__literal_repeat1,
    ACTIONS(5723), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5721), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_BQUOTE,
  [251598] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11512), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11514), 1,
      sym__concat,
    STATE(4591), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(2096), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(2094), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [251638] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11512), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11649), 1,
      sym__concat,
    STATE(4253), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1267), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1265), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [251678] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11651), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11653), 1,
      sym__concat,
    STATE(4630), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1263), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1261), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      sym__special_character,
  [251718] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11500), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11502), 1,
      sym__concat,
    STATE(4505), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5375), 4,
      sym_file_descriptor,
      sym_variable_name,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5373), 19,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [251758] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1253), 4,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1251), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [251792] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11617), 1,
      sym__special_character,
    STATE(4498), 1,
      aux_sym__literal_repeat1,
    ACTIONS(4566), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4564), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [251830] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4253), 2,
      anon_sym_PIPE,
      anon_sym_PIPE_AMP,
    ACTIONS(4272), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    STATE(4455), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(4270), 19,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [251868] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11512), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11514), 1,
      sym__concat,
    STATE(4592), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5723), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5721), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [251908] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11661), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(11664), 1,
      sym_file_descriptor,
    ACTIONS(11315), 2,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11658), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    STATE(4530), 3,
      sym_file_redirect,
      sym_herestring_redirect,
      aux_sym_redirected_statement_repeat2,
    ACTIONS(11655), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
    ACTIONS(11307), 9,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
  [251952] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11500), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11502), 1,
      sym__concat,
    STATE(4585), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5375), 3,
      sym_file_descriptor,
      sym_variable_name,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5373), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [251992] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1253), 3,
      sym_file_descriptor,
      sym__concat,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1251), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [252026] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11667), 1,
      sym__special_character,
    STATE(4533), 1,
      aux_sym__literal_repeat1,
    ACTIONS(1364), 3,
      sym_file_descriptor,
      sym_variable_name,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1362), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [252064] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11670), 1,
      sym_variable_name,
    STATE(6786), 1,
      sym_subscript,
    ACTIONS(11329), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    STATE(4439), 2,
      sym_variable_assignment,
      aux_sym_variable_assignments_repeat1,
    ACTIONS(11327), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [252104] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11500), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11502), 1,
      sym__concat,
    STATE(4586), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5069), 3,
      sym_file_descriptor,
      sym_variable_name,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5067), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [252144] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1342), 4,
      sym_file_descriptor,
      sym__concat,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1340), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
      anon_sym_BQUOTE,
  [252178] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1346), 4,
      sym_file_descriptor,
      sym__concat,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1344), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
      anon_sym_BQUOTE,
  [252212] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11585), 1,
      sym__special_character,
    STATE(4483), 1,
      aux_sym__literal_repeat1,
    ACTIONS(4469), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4467), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [252250] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1326), 3,
      sym_file_descriptor,
      sym__concat,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1324), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [252284] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1358), 4,
      sym_file_descriptor,
      sym__concat,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1356), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
      anon_sym_BQUOTE,
  [252318] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(7166), 5,
      anon_sym_COMMA,
      anon_sym_CARET,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_COLON,
    ACTIONS(7164), 21,
      sym__immediate_double_hash,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
      anon_sym_POUND,
      anon_sym_RBRACE3,
      anon_sym_AT,
      anon_sym_EQ2,
      anon_sym_COLON_EQ,
      anon_sym_DASH3,
      anon_sym_COLON_DASH,
      anon_sym_PLUS3,
      anon_sym_COLON_PLUS,
      anon_sym_QMARK2,
      anon_sym_COLON_QMARK,
      anon_sym_PERCENT_PERCENT,
      anon_sym_SLASH_SLASH,
      anon_sym_SLASH_POUND,
      anon_sym_SLASH_PERCENT,
      anon_sym_COMMA_COMMA,
      anon_sym_CARET_CARET,
  [252352] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(7162), 5,
      anon_sym_COMMA,
      anon_sym_CARET,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_COLON,
    ACTIONS(7160), 21,
      sym__immediate_double_hash,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
      anon_sym_POUND,
      anon_sym_RBRACE3,
      anon_sym_AT,
      anon_sym_EQ2,
      anon_sym_COLON_EQ,
      anon_sym_DASH3,
      anon_sym_COLON_DASH,
      anon_sym_PLUS3,
      anon_sym_COLON_PLUS,
      anon_sym_QMARK2,
      anon_sym_COLON_QMARK,
      anon_sym_PERCENT_PERCENT,
      anon_sym_SLASH_SLASH,
      anon_sym_SLASH_POUND,
      anon_sym_SLASH_PERCENT,
      anon_sym_COMMA_COMMA,
      anon_sym_CARET_CARET,
  [252386] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1310), 4,
      sym_file_descriptor,
      sym__concat,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1308), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
      anon_sym_BQUOTE,
  [252420] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1306), 4,
      sym_file_descriptor,
      sym__concat,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1304), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
      anon_sym_BQUOTE,
  [252454] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1310), 4,
      sym_file_descriptor,
      sym__concat,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1308), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
      anon_sym_BQUOTE,
  [252488] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11566), 1,
      sym__special_character,
    STATE(4485), 1,
      aux_sym__literal_repeat1,
    ACTIONS(5697), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5695), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_BQUOTE,
  [252526] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11315), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11678), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(11681), 1,
      sym_file_descriptor,
    ACTIONS(11675), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    STATE(4547), 3,
      sym_file_redirect,
      sym_herestring_redirect,
      aux_sym_redirected_statement_repeat2,
    ACTIONS(11672), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
    ACTIONS(11307), 10,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
      anon_sym_BQUOTE,
  [252570] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1302), 4,
      sym_file_descriptor,
      sym__concat,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1300), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
      anon_sym_BQUOTE,
  [252604] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11305), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(11506), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11564), 1,
      sym_file_descriptor,
    ACTIONS(11335), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    STATE(5113), 2,
      sym_file_redirect,
      sym_herestring_redirect,
    ACTIONS(11333), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
    ACTIONS(11504), 11,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
  [252648] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11500), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11502), 1,
      sym__concat,
    STATE(4603), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5375), 3,
      sym_file_descriptor,
      sym_variable_name,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5373), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_BQUOTE,
  [252688] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5383), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11558), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(11560), 1,
      sym_file_descriptor,
    STATE(4912), 1,
      sym_herestring_redirect,
    ACTIONS(4253), 2,
      anon_sym_PIPE,
      anon_sym_PIPE_AMP,
    ACTIONS(4692), 2,
      anon_sym_LT_LT,
      anon_sym_LT_LT_DASH,
    ACTIONS(5099), 2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
    ACTIONS(11556), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(5381), 3,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
    STATE(4986), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(11554), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
  [252740] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1310), 3,
      sym_file_descriptor,
      sym__concat,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1308), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [252774] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1306), 3,
      sym_file_descriptor,
      sym__concat,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1304), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [252808] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11585), 1,
      sym__special_character,
    STATE(4483), 1,
      aux_sym__literal_repeat1,
    ACTIONS(5697), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5695), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [252846] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11491), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11493), 1,
      sym__concat,
    STATE(4468), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(2096), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(2094), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [252886] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1346), 4,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1344), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [252920] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1358), 4,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1356), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [252954] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11568), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11684), 1,
      sym__concat,
    STATE(4520), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1288), 3,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1286), 20,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [252994] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11500), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11502), 1,
      sym__concat,
    STATE(4594), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5069), 3,
      sym_file_descriptor,
      sym_variable_name,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5067), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_BQUOTE,
  [253034] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1310), 3,
      sym_file_descriptor,
      sym__concat,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1308), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [253068] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1330), 4,
      sym_file_descriptor,
      sym__concat,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1328), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
      anon_sym_BQUOTE,
  [253102] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11491), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11493), 1,
      sym__concat,
    STATE(4471), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5723), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5721), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [253142] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1263), 3,
      sym_file_descriptor,
      sym_variable_name,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1261), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      sym__special_character,
  [253176] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1354), 4,
      sym_file_descriptor,
      sym__concat,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1352), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
      anon_sym_BQUOTE,
  [253210] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11558), 1,
      anon_sym_LT_LT_LT,
    STATE(4912), 1,
      sym_herestring_redirect,
    ACTIONS(4348), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    STATE(4986), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(4253), 19,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [253250] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1253), 4,
      sym_file_descriptor,
      sym__concat,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1251), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
      anon_sym_BQUOTE,
  [253284] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5387), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11558), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(11560), 1,
      sym_file_descriptor,
    STATE(4912), 1,
      sym_herestring_redirect,
    ACTIONS(4253), 2,
      anon_sym_PIPE,
      anon_sym_PIPE_AMP,
    ACTIONS(4692), 2,
      anon_sym_LT_LT,
      anon_sym_LT_LT_DASH,
    ACTIONS(5099), 2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
    ACTIONS(11556), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(5385), 3,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
    STATE(4986), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(11554), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
  [253336] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11686), 1,
      sym_variable_name,
    STATE(6758), 1,
      sym_subscript,
    ACTIONS(11329), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    STATE(4581), 2,
      sym_variable_assignment,
      aux_sym_variable_assignments_repeat1,
    ACTIONS(11327), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_BQUOTE,
  [253376] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1298), 4,
      sym_file_descriptor,
      sym__concat,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1296), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
      anon_sym_BQUOTE,
  [253410] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11558), 1,
      anon_sym_LT_LT_LT,
    STATE(4912), 1,
      sym_herestring_redirect,
    ACTIONS(4253), 2,
      anon_sym_PIPE,
      anon_sym_PIPE_AMP,
    ACTIONS(4272), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    STATE(4986), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(4270), 17,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [253452] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11692), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(11694), 1,
      sym_file_descriptor,
    ACTIONS(11363), 2,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11690), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    STATE(4530), 3,
      sym_file_redirect,
      sym_herestring_redirect,
      aux_sym_redirected_statement_repeat2,
    ACTIONS(11688), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
    ACTIONS(11357), 9,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
  [253496] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1330), 3,
      sym_file_descriptor,
      sym__concat,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1328), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [253530] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11363), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11700), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(11702), 1,
      sym_file_descriptor,
    ACTIONS(11698), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    STATE(4547), 3,
      sym_file_redirect,
      sym_herestring_redirect,
      aux_sym_redirected_statement_repeat2,
    ACTIONS(11696), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
    ACTIONS(11357), 10,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
      anon_sym_BQUOTE,
  [253574] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11704), 1,
      sym_variable_name,
    STATE(6757), 1,
      sym_subscript,
    STATE(4519), 2,
      sym_variable_assignment,
      aux_sym_variable_assignments_repeat1,
    ACTIONS(11329), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11327), 19,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [253614] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1318), 3,
      sym_file_descriptor,
      sym__concat,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1316), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      aux_sym_concatenation_token1,
  [253648] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1330), 3,
      sym_file_descriptor,
      sym__concat,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1328), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      aux_sym_concatenation_token1,
  [253682] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1263), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1261), 24,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      sym__special_character,
  [253716] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11651), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11653), 1,
      sym__concat,
    STATE(4674), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1263), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1261), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      sym__special_character,
      anon_sym_BQUOTE,
  [253756] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11491), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11493), 1,
      sym__concat,
    STATE(4468), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4372), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4370), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [253796] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11305), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(11414), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11564), 1,
      sym_file_descriptor,
    ACTIONS(11335), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    STATE(5084), 2,
      sym_file_redirect,
      sym_herestring_redirect,
    ACTIONS(11333), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
    ACTIONS(11412), 11,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
  [253840] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11706), 1,
      sym_variable_name,
    STATE(6758), 1,
      sym_subscript,
    ACTIONS(11290), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    STATE(4581), 2,
      sym_variable_assignment,
      aux_sym_variable_assignments_repeat1,
    ACTIONS(11288), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_BQUOTE,
  [253880] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11512), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11514), 1,
      sym__concat,
    STATE(4591), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(2074), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(2072), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [253920] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11512), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11709), 1,
      sym__concat,
    STATE(4253), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1267), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1265), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_BQUOTE,
  [253960] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11512), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11711), 1,
      sym__concat,
    STATE(4253), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1288), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1286), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_BQUOTE,
  [254000] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11500), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11713), 1,
      sym__concat,
    STATE(4244), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1288), 3,
      sym_file_descriptor,
      sym_variable_name,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1286), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [254040] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11500), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11715), 1,
      sym__concat,
    STATE(4244), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1267), 3,
      sym_file_descriptor,
      sym_variable_name,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1265), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [254080] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1318), 4,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1316), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [254114] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1346), 3,
      sym_file_descriptor,
      sym__concat,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1344), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      aux_sym_concatenation_token1,
  [254148] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11491), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11493), 1,
      sym__concat,
    STATE(4471), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4469), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4467), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [254188] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1358), 3,
      sym_file_descriptor,
      sym__concat,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1356), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      aux_sym_concatenation_token1,
  [254222] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11512), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11717), 1,
      sym__concat,
    STATE(4253), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1288), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1286), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [254262] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11512), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11719), 1,
      sym__concat,
    STATE(4253), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1267), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1265), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [254302] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11651), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11653), 1,
      sym__concat,
    STATE(4634), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1263), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1261), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      sym__special_character,
  [254342] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11500), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11721), 1,
      sym__concat,
    STATE(4244), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1267), 3,
      sym_file_descriptor,
      sym_variable_name,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1265), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_BQUOTE,
  [254382] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1330), 4,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1328), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [254416] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11617), 1,
      sym__special_character,
    STATE(4498), 1,
      aux_sym__literal_repeat1,
    ACTIONS(5697), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5695), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [254454] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11617), 1,
      sym__special_character,
    STATE(4498), 1,
      aux_sym__literal_repeat1,
    ACTIONS(5723), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5721), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [254492] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1294), 4,
      sym_file_descriptor,
      sym__concat,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1292), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
      anon_sym_BQUOTE,
  [254526] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11512), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11514), 1,
      sym__concat,
    STATE(4523), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5697), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5695), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [254566] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1318), 3,
      sym_file_descriptor,
      sym__concat,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1316), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [254600] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11491), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11493), 1,
      sym__concat,
    STATE(4471), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4566), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4564), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [254640] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11512), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11514), 1,
      sym__concat,
    STATE(4523), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4566), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4564), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [254680] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11500), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11723), 1,
      sym__concat,
    STATE(4244), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1288), 3,
      sym_file_descriptor,
      sym_variable_name,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1286), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_BQUOTE,
  [254720] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11512), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11514), 1,
      sym__concat,
    STATE(4510), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4554), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4552), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [254760] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(6820), 5,
      anon_sym_COMMA,
      anon_sym_CARET,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_COLON,
    ACTIONS(6818), 21,
      sym__immediate_double_hash,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
      anon_sym_POUND,
      anon_sym_RBRACE3,
      anon_sym_AT,
      anon_sym_EQ2,
      anon_sym_COLON_EQ,
      anon_sym_DASH3,
      anon_sym_COLON_DASH,
      anon_sym_PLUS3,
      anon_sym_COLON_PLUS,
      anon_sym_QMARK2,
      anon_sym_COLON_QMARK,
      anon_sym_PERCENT_PERCENT,
      anon_sym_SLASH_SLASH,
      anon_sym_SLASH_POUND,
      anon_sym_SLASH_PERCENT,
      anon_sym_COMMA_COMMA,
      anon_sym_CARET_CARET,
  [254794] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1350), 4,
      sym_file_descriptor,
      sym__concat,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1348), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
      anon_sym_BQUOTE,
  [254828] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11727), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11725), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [254861] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4272), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    STATE(4650), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(4270), 19,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [254896] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11729), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11731), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_BQUOTE,
  [254929] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11733), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11735), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_BQUOTE,
  [254962] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11651), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11737), 1,
      sym__concat,
    STATE(4268), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1288), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1286), 19,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [255001] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11727), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11725), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_BQUOTE,
  [255034] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11733), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11735), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [255067] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11651), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11653), 1,
      sym__concat,
    STATE(4611), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(2074), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(2072), 19,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [255106] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11739), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11741), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_BQUOTE,
  [255139] = 5,
    ACTIONS(3), 1,
      sym_comment,
    STATE(5602), 1,
      aux_sym_pipeline_repeat1,
    ACTIONS(11745), 2,
      anon_sym_PIPE,
      anon_sym_PIPE_AMP,
    ACTIONS(11748), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11743), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [255176] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4554), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4552), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_BQUOTE,
  [255209] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1298), 3,
      sym_file_descriptor,
      sym__concat,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1296), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [255242] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11651), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11653), 1,
      sym__concat,
    STATE(4674), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4469), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4467), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_BQUOTE,
  [255281] = 10,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4259), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11457), 1,
      sym_file_descriptor,
    ACTIONS(4249), 2,
      anon_sym_SEMI,
      anon_sym_AMP,
    ACTIONS(4251), 2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
    ACTIONS(4255), 2,
      anon_sym_LT_LT,
      anon_sym_LT_LT_DASH,
    ACTIONS(11282), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    STATE(4283), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(4257), 4,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
    ACTIONS(11280), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
  [255328] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11651), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11653), 1,
      sym__concat,
    STATE(4634), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5723), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5721), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [255367] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11651), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11653), 1,
      sym__concat,
    STATE(4669), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4372), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4370), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_BQUOTE,
  [255406] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11750), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11752), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_BQUOTE,
  [255439] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11758), 1,
      anon_sym_RBRACE3,
    ACTIONS(11760), 1,
      anon_sym_BANG2,
    ACTIONS(11764), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(11766), 1,
      anon_sym_BQUOTE,
    ACTIONS(11768), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(11770), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(11772), 1,
      sym_variable_name,
    STATE(3532), 1,
      sym_subscript,
    STATE(6428), 1,
      aux_sym__expansion_body_repeat1,
    STATE(6472), 1,
      sym_command_substitution,
    STATE(6908), 1,
      sym__expansion_body,
    ACTIONS(11762), 2,
      anon_sym_POUND2,
      anon_sym_EQ2,
    ACTIONS(8050), 3,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
    ACTIONS(11754), 4,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_QMARK,
      anon_sym_AT2,
    ACTIONS(11756), 5,
      anon_sym_BANG,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_0,
      anon_sym__,
  [255498] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11448), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(11774), 1,
      sym_file_descriptor,
    ACTIONS(11414), 2,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11481), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    STATE(4868), 2,
      sym_file_redirect,
      sym_herestring_redirect,
    ACTIONS(11479), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
    ACTIONS(11412), 9,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
  [255541] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11651), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11653), 1,
      sym__concat,
    STATE(4674), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5697), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5695), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_BQUOTE,
  [255580] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11651), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11653), 1,
      sym__concat,
    STATE(4669), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(2096), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(2094), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_BQUOTE,
  [255619] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1310), 4,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1308), 21,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [255652] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11760), 1,
      anon_sym_BANG2,
    ACTIONS(11764), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(11766), 1,
      anon_sym_BQUOTE,
    ACTIONS(11768), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(11770), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(11772), 1,
      sym_variable_name,
    ACTIONS(11776), 1,
      anon_sym_RBRACE3,
    STATE(3532), 1,
      sym_subscript,
    STATE(6428), 1,
      aux_sym__expansion_body_repeat1,
    STATE(6472), 1,
      sym_command_substitution,
    STATE(7521), 1,
      sym__expansion_body,
    ACTIONS(11762), 2,
      anon_sym_POUND2,
      anon_sym_EQ2,
    ACTIONS(8050), 3,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
    ACTIONS(11754), 4,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_QMARK,
      anon_sym_AT2,
    ACTIONS(11756), 5,
      anon_sym_BANG,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_0,
      anon_sym__,
  [255711] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11651), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11778), 1,
      sym__concat,
    STATE(4268), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1267), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1265), 19,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [255750] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1306), 4,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1304), 21,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [255783] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11782), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11780), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [255816] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11651), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11653), 1,
      sym__concat,
    STATE(4674), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5723), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5721), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_BQUOTE,
  [255855] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11651), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11784), 1,
      sym__concat,
    STATE(4268), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1267), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1265), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [255894] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11651), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11786), 1,
      sym__concat,
    STATE(4268), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1288), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1286), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [255933] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5772), 4,
      sym_file_descriptor,
      sym_variable_name,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5770), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_BQUOTE,
  [255966] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1346), 3,
      sym_file_descriptor,
      sym__concat,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1344), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [255999] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1358), 3,
      sym_file_descriptor,
      sym__concat,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1356), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [256032] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5375), 3,
      sym_file_descriptor,
      sym_variable_name,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5373), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [256065] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11651), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11653), 1,
      sym__concat,
    STATE(4669), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(2074), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(2072), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_BQUOTE,
  [256104] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11760), 1,
      anon_sym_BANG2,
    ACTIONS(11764), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(11766), 1,
      anon_sym_BQUOTE,
    ACTIONS(11768), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(11770), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(11772), 1,
      sym_variable_name,
    ACTIONS(11788), 1,
      anon_sym_RBRACE3,
    STATE(3532), 1,
      sym_subscript,
    STATE(6428), 1,
      aux_sym__expansion_body_repeat1,
    STATE(6472), 1,
      sym_command_substitution,
    STATE(6810), 1,
      sym__expansion_body,
    ACTIONS(11762), 2,
      anon_sym_POUND2,
      anon_sym_EQ2,
    ACTIONS(8050), 3,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
    ACTIONS(11754), 4,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_QMARK,
      anon_sym_AT2,
    ACTIONS(11756), 5,
      anon_sym_BANG,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_0,
      anon_sym__,
  [256163] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11790), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11792), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_BQUOTE,
  [256196] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11729), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11731), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [256229] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11363), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11798), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(11800), 1,
      sym_file_descriptor,
    ACTIONS(11796), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    STATE(4665), 3,
      sym_file_redirect,
      sym_herestring_redirect,
      aux_sym_redirected_statement_repeat2,
    ACTIONS(11794), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
    ACTIONS(11357), 9,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
  [256272] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11802), 1,
      sym__special_character,
    STATE(4855), 1,
      aux_sym__literal_repeat1,
    ACTIONS(4566), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4564), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [256309] = 19,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(11804), 1,
      anon_sym_RPAREN_RPAREN,
    ACTIONS(11808), 1,
      anon_sym_LPAREN,
    ACTIONS(11810), 1,
      aux_sym__c_word_token1,
    ACTIONS(11812), 1,
      anon_sym_DOLLAR,
    ACTIONS(11814), 1,
      anon_sym_DQUOTE,
    ACTIONS(11816), 1,
      aux_sym_number_token1,
    ACTIONS(11818), 1,
      aux_sym_number_token2,
    ACTIONS(11820), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(11822), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(11824), 1,
      anon_sym_BQUOTE,
    ACTIONS(11826), 1,
      anon_sym_DOLLAR_BQUOTE,
    STATE(3285), 1,
      sym__c_postfix_expression,
    STATE(3286), 1,
      sym__c_binary_expression,
    STATE(3290), 1,
      sym__c_unary_expression,
    STATE(6702), 1,
      sym__c_expression,
    STATE(6768), 1,
      sym__c_variable_assignment,
    ACTIONS(11806), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    STATE(3292), 7,
      sym__c_expression_not_assignment,
      sym__c_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [256374] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11790), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11792), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [256407] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11830), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11828), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [256440] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11760), 1,
      anon_sym_BANG2,
    ACTIONS(11764), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(11766), 1,
      anon_sym_BQUOTE,
    ACTIONS(11768), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(11770), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(11772), 1,
      sym_variable_name,
    ACTIONS(11832), 1,
      anon_sym_RBRACE3,
    STATE(3532), 1,
      sym_subscript,
    STATE(6428), 1,
      aux_sym__expansion_body_repeat1,
    STATE(6472), 1,
      sym_command_substitution,
    STATE(6984), 1,
      sym__expansion_body,
    ACTIONS(11762), 2,
      anon_sym_POUND2,
      anon_sym_EQ2,
    ACTIONS(8050), 3,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
    ACTIONS(11754), 4,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_QMARK,
      anon_sym_AT2,
    ACTIONS(11756), 5,
      anon_sym_BANG,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_0,
      anon_sym__,
  [256499] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11834), 1,
      sym_file_descriptor,
    ACTIONS(4692), 2,
      anon_sym_LT_LT,
      anon_sym_LT_LT_DASH,
    ACTIONS(11455), 2,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11481), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    STATE(4734), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(11453), 7,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
    ACTIONS(11479), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
  [256542] = 19,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(11808), 1,
      anon_sym_LPAREN,
    ACTIONS(11810), 1,
      aux_sym__c_word_token1,
    ACTIONS(11812), 1,
      anon_sym_DOLLAR,
    ACTIONS(11814), 1,
      anon_sym_DQUOTE,
    ACTIONS(11816), 1,
      aux_sym_number_token1,
    ACTIONS(11818), 1,
      aux_sym_number_token2,
    ACTIONS(11820), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(11822), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(11824), 1,
      anon_sym_BQUOTE,
    ACTIONS(11826), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(11836), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(3285), 1,
      sym__c_postfix_expression,
    STATE(3286), 1,
      sym__c_binary_expression,
    STATE(3290), 1,
      sym__c_unary_expression,
    STATE(6665), 1,
      sym__c_expression,
    STATE(6768), 1,
      sym__c_variable_assignment,
    ACTIONS(11806), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    STATE(3292), 7,
      sym__c_expression_not_assignment,
      sym__c_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [256607] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11760), 1,
      anon_sym_BANG2,
    ACTIONS(11764), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(11766), 1,
      anon_sym_BQUOTE,
    ACTIONS(11768), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(11770), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(11772), 1,
      sym_variable_name,
    ACTIONS(11838), 1,
      anon_sym_RBRACE3,
    STATE(3532), 1,
      sym_subscript,
    STATE(6428), 1,
      aux_sym__expansion_body_repeat1,
    STATE(6472), 1,
      sym_command_substitution,
    STATE(7433), 1,
      sym__expansion_body,
    ACTIONS(11762), 2,
      anon_sym_POUND2,
      anon_sym_EQ2,
    ACTIONS(8050), 3,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
    ACTIONS(11754), 4,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_QMARK,
      anon_sym_AT2,
    ACTIONS(11756), 5,
      anon_sym_BANG,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_0,
      anon_sym__,
  [256666] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1310), 4,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1308), 21,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [256699] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11760), 1,
      anon_sym_BANG2,
    ACTIONS(11764), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(11766), 1,
      anon_sym_BQUOTE,
    ACTIONS(11768), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(11770), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(11772), 1,
      sym_variable_name,
    ACTIONS(11840), 1,
      anon_sym_RBRACE3,
    STATE(3532), 1,
      sym_subscript,
    STATE(6428), 1,
      aux_sym__expansion_body_repeat1,
    STATE(6472), 1,
      sym_command_substitution,
    STATE(7146), 1,
      sym__expansion_body,
    ACTIONS(11762), 2,
      anon_sym_POUND2,
      anon_sym_EQ2,
    ACTIONS(8050), 3,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
    ACTIONS(11754), 4,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_QMARK,
      anon_sym_AT2,
    ACTIONS(11756), 5,
      anon_sym_BANG,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_0,
      anon_sym__,
  [256758] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11842), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11844), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_BQUOTE,
  [256791] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1302), 3,
      sym_file_descriptor,
      sym__concat,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1300), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [256824] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11760), 1,
      anon_sym_BANG2,
    ACTIONS(11764), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(11766), 1,
      anon_sym_BQUOTE,
    ACTIONS(11768), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(11770), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(11772), 1,
      sym_variable_name,
    ACTIONS(11846), 1,
      anon_sym_RBRACE3,
    STATE(3532), 1,
      sym_subscript,
    STATE(6428), 1,
      aux_sym__expansion_body_repeat1,
    STATE(6472), 1,
      sym_command_substitution,
    STATE(7089), 1,
      sym__expansion_body,
    ACTIONS(11762), 2,
      anon_sym_POUND2,
      anon_sym_EQ2,
    ACTIONS(8050), 3,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
    ACTIONS(11754), 4,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_QMARK,
      anon_sym_AT2,
    ACTIONS(11756), 5,
      anon_sym_BANG,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_0,
      anon_sym__,
  [256883] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11760), 1,
      anon_sym_BANG2,
    ACTIONS(11764), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(11766), 1,
      anon_sym_BQUOTE,
    ACTIONS(11768), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(11770), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(11772), 1,
      sym_variable_name,
    ACTIONS(11848), 1,
      anon_sym_RBRACE3,
    STATE(3532), 1,
      sym_subscript,
    STATE(6428), 1,
      aux_sym__expansion_body_repeat1,
    STATE(6472), 1,
      sym_command_substitution,
    STATE(7105), 1,
      sym__expansion_body,
    ACTIONS(11762), 2,
      anon_sym_POUND2,
      anon_sym_EQ2,
    ACTIONS(8050), 3,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
    ACTIONS(11754), 4,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_QMARK,
      anon_sym_AT2,
    ACTIONS(11756), 5,
      anon_sym_BANG,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_0,
      anon_sym__,
  [256942] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11506), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11508), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(11850), 1,
      sym_file_descriptor,
    ACTIONS(11542), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    STATE(5017), 2,
      sym_file_redirect,
      sym_herestring_redirect,
    ACTIONS(11540), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
    ACTIONS(11504), 10,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
  [256985] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11760), 1,
      anon_sym_BANG2,
    ACTIONS(11764), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(11766), 1,
      anon_sym_BQUOTE,
    ACTIONS(11768), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(11770), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(11772), 1,
      sym_variable_name,
    ACTIONS(11852), 1,
      anon_sym_RBRACE3,
    STATE(3532), 1,
      sym_subscript,
    STATE(6428), 1,
      aux_sym__expansion_body_repeat1,
    STATE(6472), 1,
      sym_command_substitution,
    STATE(7261), 1,
      sym__expansion_body,
    ACTIONS(11762), 2,
      anon_sym_POUND2,
      anon_sym_EQ2,
    ACTIONS(8050), 3,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
    ACTIONS(11754), 4,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_QMARK,
      anon_sym_AT2,
    ACTIONS(11756), 5,
      anon_sym_BANG,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_0,
      anon_sym__,
  [257044] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11760), 1,
      anon_sym_BANG2,
    ACTIONS(11764), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(11766), 1,
      anon_sym_BQUOTE,
    ACTIONS(11768), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(11770), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(11772), 1,
      sym_variable_name,
    ACTIONS(11854), 1,
      anon_sym_RBRACE3,
    STATE(3532), 1,
      sym_subscript,
    STATE(6428), 1,
      aux_sym__expansion_body_repeat1,
    STATE(6472), 1,
      sym_command_substitution,
    STATE(7112), 1,
      sym__expansion_body,
    ACTIONS(11762), 2,
      anon_sym_POUND2,
      anon_sym_EQ2,
    ACTIONS(8050), 3,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
    ACTIONS(11754), 4,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_QMARK,
      anon_sym_AT2,
    ACTIONS(11756), 5,
      anon_sym_BANG,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_0,
      anon_sym__,
  [257103] = 11,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4257), 1,
      anon_sym_BQUOTE,
    ACTIONS(5379), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11856), 1,
      sym_file_descriptor,
    ACTIONS(4253), 2,
      anon_sym_PIPE,
      anon_sym_PIPE_AMP,
    ACTIONS(4692), 2,
      anon_sym_LT_LT,
      anon_sym_LT_LT_DASH,
    ACTIONS(5355), 2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
    ACTIONS(11487), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(5377), 3,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
    STATE(4697), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(11485), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
  [257152] = 11,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2721), 1,
      ts_builtin_sym_end,
    ACTIONS(5051), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11834), 1,
      sym_file_descriptor,
    ACTIONS(4253), 2,
      anon_sym_PIPE,
      anon_sym_PIPE_AMP,
    ACTIONS(4690), 2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
    ACTIONS(4692), 2,
      anon_sym_LT_LT,
      anon_sym_LT_LT_DASH,
    ACTIONS(11481), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(5049), 3,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
    STATE(4650), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(11479), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
  [257201] = 5,
    ACTIONS(3), 1,
      sym_comment,
    STATE(4664), 1,
      aux_sym_pipeline_repeat1,
    ACTIONS(11748), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11858), 2,
      anon_sym_PIPE,
      anon_sym_PIPE_AMP,
    ACTIONS(11743), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [257238] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11315), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11867), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(11870), 1,
      sym_file_descriptor,
    ACTIONS(11864), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    STATE(4665), 3,
      sym_file_redirect,
      sym_herestring_redirect,
      aux_sym_redirected_statement_repeat2,
    ACTIONS(11861), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
    ACTIONS(11307), 9,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
  [257281] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11782), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11780), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_BQUOTE,
  [257314] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11448), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(11774), 1,
      sym_file_descriptor,
    ACTIONS(11481), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(11535), 2,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    STATE(4877), 2,
      sym_file_redirect,
      sym_herestring_redirect,
    ACTIONS(11479), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
    ACTIONS(11533), 9,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
  [257357] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11760), 1,
      anon_sym_BANG2,
    ACTIONS(11764), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(11766), 1,
      anon_sym_BQUOTE,
    ACTIONS(11768), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(11770), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(11772), 1,
      sym_variable_name,
    ACTIONS(11873), 1,
      anon_sym_RBRACE3,
    STATE(3532), 1,
      sym_subscript,
    STATE(6428), 1,
      aux_sym__expansion_body_repeat1,
    STATE(6472), 1,
      sym_command_substitution,
    STATE(6860), 1,
      sym__expansion_body,
    ACTIONS(11762), 2,
      anon_sym_POUND2,
      anon_sym_EQ2,
    ACTIONS(8050), 3,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
    ACTIONS(11754), 4,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_QMARK,
      anon_sym_AT2,
    ACTIONS(11756), 5,
      anon_sym_BANG,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_0,
      anon_sym__,
  [257416] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11651), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11875), 1,
      sym__concat,
    STATE(4268), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1288), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1286), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_BQUOTE,
  [257455] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1302), 4,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1300), 21,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [257488] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11760), 1,
      anon_sym_BANG2,
    ACTIONS(11764), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(11766), 1,
      anon_sym_BQUOTE,
    ACTIONS(11768), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(11770), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(11772), 1,
      sym_variable_name,
    ACTIONS(11877), 1,
      anon_sym_RBRACE3,
    STATE(3532), 1,
      sym_subscript,
    STATE(6428), 1,
      aux_sym__expansion_body_repeat1,
    STATE(6472), 1,
      sym_command_substitution,
    STATE(7126), 1,
      sym__expansion_body,
    ACTIONS(11762), 2,
      anon_sym_POUND2,
      anon_sym_EQ2,
    ACTIONS(8050), 3,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
    ACTIONS(11754), 4,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_QMARK,
      anon_sym_AT2,
    ACTIONS(11756), 5,
      anon_sym_BANG,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_0,
      anon_sym__,
  [257547] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11760), 1,
      anon_sym_BANG2,
    ACTIONS(11764), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(11766), 1,
      anon_sym_BQUOTE,
    ACTIONS(11768), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(11770), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(11772), 1,
      sym_variable_name,
    ACTIONS(11879), 1,
      anon_sym_RBRACE3,
    STATE(3532), 1,
      sym_subscript,
    STATE(6428), 1,
      aux_sym__expansion_body_repeat1,
    STATE(6472), 1,
      sym_command_substitution,
    STATE(7135), 1,
      sym__expansion_body,
    ACTIONS(11762), 2,
      anon_sym_POUND2,
      anon_sym_EQ2,
    ACTIONS(8050), 3,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
    ACTIONS(11754), 4,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_QMARK,
      anon_sym_AT2,
    ACTIONS(11756), 5,
      anon_sym_BANG,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_0,
      anon_sym__,
  [257606] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11651), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11653), 1,
      sym__concat,
    STATE(4611), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4554), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4552), 19,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [257645] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11651), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11881), 1,
      sym__concat,
    STATE(4268), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1267), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1265), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_BQUOTE,
  [257684] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11760), 1,
      anon_sym_BANG2,
    ACTIONS(11764), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(11766), 1,
      anon_sym_BQUOTE,
    ACTIONS(11768), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(11770), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(11772), 1,
      sym_variable_name,
    ACTIONS(11883), 1,
      anon_sym_RBRACE3,
    STATE(3532), 1,
      sym_subscript,
    STATE(6428), 1,
      aux_sym__expansion_body_repeat1,
    STATE(6472), 1,
      sym_command_substitution,
    STATE(7141), 1,
      sym__expansion_body,
    ACTIONS(11762), 2,
      anon_sym_POUND2,
      anon_sym_EQ2,
    ACTIONS(8050), 3,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
    ACTIONS(11754), 4,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_QMARK,
      anon_sym_AT2,
    ACTIONS(11756), 5,
      anon_sym_BANG,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_0,
      anon_sym__,
  [257743] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11760), 1,
      anon_sym_BANG2,
    ACTIONS(11764), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(11766), 1,
      anon_sym_BQUOTE,
    ACTIONS(11768), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(11770), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(11772), 1,
      sym_variable_name,
    ACTIONS(11885), 1,
      anon_sym_RBRACE3,
    STATE(3532), 1,
      sym_subscript,
    STATE(6428), 1,
      aux_sym__expansion_body_repeat1,
    STATE(6472), 1,
      sym_command_substitution,
    STATE(7163), 1,
      sym__expansion_body,
    ACTIONS(11762), 2,
      anon_sym_POUND2,
      anon_sym_EQ2,
    ACTIONS(8050), 3,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
    ACTIONS(11754), 4,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_QMARK,
      anon_sym_AT2,
    ACTIONS(11756), 5,
      anon_sym_BANG,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_0,
      anon_sym__,
  [257802] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11760), 1,
      anon_sym_BANG2,
    ACTIONS(11764), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(11766), 1,
      anon_sym_BQUOTE,
    ACTIONS(11768), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(11770), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(11772), 1,
      sym_variable_name,
    ACTIONS(11887), 1,
      anon_sym_RBRACE3,
    STATE(3532), 1,
      sym_subscript,
    STATE(6428), 1,
      aux_sym__expansion_body_repeat1,
    STATE(6472), 1,
      sym_command_substitution,
    STATE(7170), 1,
      sym__expansion_body,
    ACTIONS(11762), 2,
      anon_sym_POUND2,
      anon_sym_EQ2,
    ACTIONS(8050), 3,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
    ACTIONS(11754), 4,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_QMARK,
      anon_sym_AT2,
    ACTIONS(11756), 5,
      anon_sym_BANG,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_0,
      anon_sym__,
  [257861] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11760), 1,
      anon_sym_BANG2,
    ACTIONS(11764), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(11766), 1,
      anon_sym_BQUOTE,
    ACTIONS(11768), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(11770), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(11772), 1,
      sym_variable_name,
    ACTIONS(11889), 1,
      anon_sym_RBRACE3,
    STATE(3532), 1,
      sym_subscript,
    STATE(6428), 1,
      aux_sym__expansion_body_repeat1,
    STATE(6472), 1,
      sym_command_substitution,
    STATE(7179), 1,
      sym__expansion_body,
    ACTIONS(11762), 2,
      anon_sym_POUND2,
      anon_sym_EQ2,
    ACTIONS(8050), 3,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
    ACTIONS(11754), 4,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_QMARK,
      anon_sym_AT2,
    ACTIONS(11756), 5,
      anon_sym_BANG,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_0,
      anon_sym__,
  [257920] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11760), 1,
      anon_sym_BANG2,
    ACTIONS(11764), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(11766), 1,
      anon_sym_BQUOTE,
    ACTIONS(11768), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(11770), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(11772), 1,
      sym_variable_name,
    ACTIONS(11891), 1,
      anon_sym_RBRACE3,
    STATE(3532), 1,
      sym_subscript,
    STATE(6428), 1,
      aux_sym__expansion_body_repeat1,
    STATE(6472), 1,
      sym_command_substitution,
    STATE(7187), 1,
      sym__expansion_body,
    ACTIONS(11762), 2,
      anon_sym_POUND2,
      anon_sym_EQ2,
    ACTIONS(8050), 3,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
    ACTIONS(11754), 4,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_QMARK,
      anon_sym_AT2,
    ACTIONS(11756), 5,
      anon_sym_BANG,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_0,
      anon_sym__,
  [257979] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11760), 1,
      anon_sym_BANG2,
    ACTIONS(11764), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(11766), 1,
      anon_sym_BQUOTE,
    ACTIONS(11768), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(11770), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(11772), 1,
      sym_variable_name,
    ACTIONS(11893), 1,
      anon_sym_RBRACE3,
    STATE(3532), 1,
      sym_subscript,
    STATE(6428), 1,
      aux_sym__expansion_body_repeat1,
    STATE(6472), 1,
      sym_command_substitution,
    STATE(7196), 1,
      sym__expansion_body,
    ACTIONS(11762), 2,
      anon_sym_POUND2,
      anon_sym_EQ2,
    ACTIONS(8050), 3,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
    ACTIONS(11754), 4,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_QMARK,
      anon_sym_AT2,
    ACTIONS(11756), 5,
      anon_sym_BANG,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_0,
      anon_sym__,
  [258038] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11760), 1,
      anon_sym_BANG2,
    ACTIONS(11764), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(11766), 1,
      anon_sym_BQUOTE,
    ACTIONS(11768), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(11770), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(11772), 1,
      sym_variable_name,
    ACTIONS(11895), 1,
      anon_sym_RBRACE3,
    STATE(3532), 1,
      sym_subscript,
    STATE(6428), 1,
      aux_sym__expansion_body_repeat1,
    STATE(6472), 1,
      sym_command_substitution,
    STATE(7207), 1,
      sym__expansion_body,
    ACTIONS(11762), 2,
      anon_sym_POUND2,
      anon_sym_EQ2,
    ACTIONS(8050), 3,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
    ACTIONS(11754), 4,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_QMARK,
      anon_sym_AT2,
    ACTIONS(11756), 5,
      anon_sym_BANG,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_0,
      anon_sym__,
  [258097] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11760), 1,
      anon_sym_BANG2,
    ACTIONS(11764), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(11766), 1,
      anon_sym_BQUOTE,
    ACTIONS(11768), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(11770), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(11772), 1,
      sym_variable_name,
    ACTIONS(11897), 1,
      anon_sym_RBRACE3,
    STATE(3532), 1,
      sym_subscript,
    STATE(6428), 1,
      aux_sym__expansion_body_repeat1,
    STATE(6472), 1,
      sym_command_substitution,
    STATE(7219), 1,
      sym__expansion_body,
    ACTIONS(11762), 2,
      anon_sym_POUND2,
      anon_sym_EQ2,
    ACTIONS(8050), 3,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
    ACTIONS(11754), 4,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_QMARK,
      anon_sym_AT2,
    ACTIONS(11756), 5,
      anon_sym_BANG,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_0,
      anon_sym__,
  [258156] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11651), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11653), 1,
      sym__concat,
    STATE(4634), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4469), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4467), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [258195] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11651), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11653), 1,
      sym__concat,
    STATE(4669), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4554), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4552), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_BQUOTE,
  [258234] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11414), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11508), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(11850), 1,
      sym_file_descriptor,
    ACTIONS(11542), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    STATE(4868), 2,
      sym_file_redirect,
      sym_herestring_redirect,
    ACTIONS(11540), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
    ACTIONS(11412), 10,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
  [258277] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11760), 1,
      anon_sym_BANG2,
    ACTIONS(11764), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(11766), 1,
      anon_sym_BQUOTE,
    ACTIONS(11768), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(11770), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(11772), 1,
      sym_variable_name,
    ACTIONS(11899), 1,
      anon_sym_RBRACE3,
    STATE(3532), 1,
      sym_subscript,
    STATE(6428), 1,
      aux_sym__expansion_body_repeat1,
    STATE(6472), 1,
      sym_command_substitution,
    STATE(7245), 1,
      sym__expansion_body,
    ACTIONS(11762), 2,
      anon_sym_POUND2,
      anon_sym_EQ2,
    ACTIONS(8050), 3,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
    ACTIONS(11754), 4,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_QMARK,
      anon_sym_AT2,
    ACTIONS(11756), 5,
      anon_sym_BANG,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_0,
      anon_sym__,
  [258336] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11651), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11653), 1,
      sym__concat,
    STATE(4635), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4372), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4370), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [258375] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5768), 4,
      sym_file_descriptor,
      sym_variable_name,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5766), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_BQUOTE,
  [258408] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11651), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11653), 1,
      sym__concat,
    STATE(4634), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5697), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5695), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [258447] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11760), 1,
      anon_sym_BANG2,
    ACTIONS(11764), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(11766), 1,
      anon_sym_BQUOTE,
    ACTIONS(11768), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(11770), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(11772), 1,
      sym_variable_name,
    ACTIONS(11901), 1,
      anon_sym_RBRACE3,
    STATE(3532), 1,
      sym_subscript,
    STATE(6428), 1,
      aux_sym__expansion_body_repeat1,
    STATE(6472), 1,
      sym_command_substitution,
    STATE(7257), 1,
      sym__expansion_body,
    ACTIONS(11762), 2,
      anon_sym_POUND2,
      anon_sym_EQ2,
    ACTIONS(8050), 3,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
    ACTIONS(11754), 4,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_QMARK,
      anon_sym_AT2,
    ACTIONS(11756), 5,
      anon_sym_BANG,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_0,
      anon_sym__,
  [258506] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11760), 1,
      anon_sym_BANG2,
    ACTIONS(11764), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(11766), 1,
      anon_sym_BQUOTE,
    ACTIONS(11768), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(11770), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(11772), 1,
      sym_variable_name,
    ACTIONS(11903), 1,
      anon_sym_RBRACE3,
    STATE(3532), 1,
      sym_subscript,
    STATE(6428), 1,
      aux_sym__expansion_body_repeat1,
    STATE(6472), 1,
      sym_command_substitution,
    STATE(6900), 1,
      sym__expansion_body,
    ACTIONS(11762), 2,
      anon_sym_POUND2,
      anon_sym_EQ2,
    ACTIONS(8050), 3,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
    ACTIONS(11754), 4,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_QMARK,
      anon_sym_AT2,
    ACTIONS(11756), 5,
      anon_sym_BANG,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_0,
      anon_sym__,
  [258565] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11651), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11653), 1,
      sym__concat,
    STATE(4635), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(2074), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(2072), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [258604] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11760), 1,
      anon_sym_BANG2,
    ACTIONS(11764), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(11766), 1,
      anon_sym_BQUOTE,
    ACTIONS(11768), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(11770), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(11772), 1,
      sym_variable_name,
    ACTIONS(11905), 1,
      anon_sym_RBRACE3,
    STATE(3532), 1,
      sym_subscript,
    STATE(6428), 1,
      aux_sym__expansion_body_repeat1,
    STATE(6472), 1,
      sym_command_substitution,
    STATE(7271), 1,
      sym__expansion_body,
    ACTIONS(11762), 2,
      anon_sym_POUND2,
      anon_sym_EQ2,
    ACTIONS(8050), 3,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
    ACTIONS(11754), 4,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_QMARK,
      anon_sym_AT2,
    ACTIONS(11756), 5,
      anon_sym_BANG,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_0,
      anon_sym__,
  [258663] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11760), 1,
      anon_sym_BANG2,
    ACTIONS(11764), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(11766), 1,
      anon_sym_BQUOTE,
    ACTIONS(11768), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(11770), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(11772), 1,
      sym_variable_name,
    ACTIONS(11907), 1,
      anon_sym_RBRACE3,
    STATE(3532), 1,
      sym_subscript,
    STATE(6428), 1,
      aux_sym__expansion_body_repeat1,
    STATE(6472), 1,
      sym_command_substitution,
    STATE(7082), 1,
      sym__expansion_body,
    ACTIONS(11762), 2,
      anon_sym_POUND2,
      anon_sym_EQ2,
    ACTIONS(8050), 3,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
    ACTIONS(11754), 4,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_QMARK,
      anon_sym_AT2,
    ACTIONS(11756), 5,
      anon_sym_BANG,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_0,
      anon_sym__,
  [258722] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11802), 1,
      sym__special_character,
    STATE(4855), 1,
      aux_sym__literal_repeat1,
    ACTIONS(5723), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5721), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [258759] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11760), 1,
      anon_sym_BANG2,
    ACTIONS(11764), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(11766), 1,
      anon_sym_BQUOTE,
    ACTIONS(11768), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(11770), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(11772), 1,
      sym_variable_name,
    ACTIONS(11909), 1,
      anon_sym_RBRACE3,
    STATE(3532), 1,
      sym_subscript,
    STATE(6428), 1,
      aux_sym__expansion_body_repeat1,
    STATE(6472), 1,
      sym_command_substitution,
    STATE(7284), 1,
      sym__expansion_body,
    ACTIONS(11762), 2,
      anon_sym_POUND2,
      anon_sym_EQ2,
    ACTIONS(8050), 3,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
    ACTIONS(11754), 4,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_QMARK,
      anon_sym_AT2,
    ACTIONS(11756), 5,
      anon_sym_BANG,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_0,
      anon_sym__,
  [258818] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11455), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11856), 1,
      sym_file_descriptor,
    ACTIONS(4692), 2,
      anon_sym_LT_LT,
      anon_sym_LT_LT_DASH,
    ACTIONS(11487), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    STATE(4725), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(11453), 8,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_BQUOTE,
    ACTIONS(11485), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
  [258861] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11651), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11653), 1,
      sym__concat,
    STATE(4634), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4566), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4564), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [258900] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11651), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11653), 1,
      sym__concat,
    STATE(4635), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4554), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4552), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [258939] = 19,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(11808), 1,
      anon_sym_LPAREN,
    ACTIONS(11810), 1,
      aux_sym__c_word_token1,
    ACTIONS(11812), 1,
      anon_sym_DOLLAR,
    ACTIONS(11814), 1,
      anon_sym_DQUOTE,
    ACTIONS(11816), 1,
      aux_sym_number_token1,
    ACTIONS(11818), 1,
      aux_sym_number_token2,
    ACTIONS(11820), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(11822), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(11824), 1,
      anon_sym_BQUOTE,
    ACTIONS(11826), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(11911), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(3285), 1,
      sym__c_postfix_expression,
    STATE(3286), 1,
      sym__c_binary_expression,
    STATE(3290), 1,
      sym__c_unary_expression,
    STATE(6618), 1,
      sym__c_expression,
    STATE(6768), 1,
      sym__c_variable_assignment,
    ACTIONS(11806), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    STATE(3292), 7,
      sym__c_expression_not_assignment,
      sym__c_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [259004] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1330), 3,
      sym_file_descriptor,
      sym__concat,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1328), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [259037] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11830), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11828), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_BQUOTE,
  [259070] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11760), 1,
      anon_sym_BANG2,
    ACTIONS(11764), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(11766), 1,
      anon_sym_BQUOTE,
    ACTIONS(11768), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(11770), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(11772), 1,
      sym_variable_name,
    ACTIONS(11913), 1,
      anon_sym_RBRACE3,
    STATE(3532), 1,
      sym_subscript,
    STATE(6428), 1,
      aux_sym__expansion_body_repeat1,
    STATE(6472), 1,
      sym_command_substitution,
    STATE(7296), 1,
      sym__expansion_body,
    ACTIONS(11762), 2,
      anon_sym_POUND2,
      anon_sym_EQ2,
    ACTIONS(8050), 3,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
    ACTIONS(11754), 4,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_QMARK,
      anon_sym_AT2,
    ACTIONS(11756), 5,
      anon_sym_BANG,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_0,
      anon_sym__,
  [259129] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11802), 1,
      sym__special_character,
    STATE(4855), 1,
      aux_sym__literal_repeat1,
    ACTIONS(4469), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4467), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [259166] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11651), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11653), 1,
      sym__concat,
    STATE(4611), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(2096), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(2094), 19,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [259205] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1318), 3,
      sym_file_descriptor,
      sym__concat,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1316), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [259238] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11760), 1,
      anon_sym_BANG2,
    ACTIONS(11764), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(11766), 1,
      anon_sym_BQUOTE,
    ACTIONS(11768), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(11770), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(11772), 1,
      sym_variable_name,
    ACTIONS(11915), 1,
      anon_sym_RBRACE3,
    STATE(3532), 1,
      sym_subscript,
    STATE(6428), 1,
      aux_sym__expansion_body_repeat1,
    STATE(6472), 1,
      sym_command_substitution,
    STATE(7317), 1,
      sym__expansion_body,
    ACTIONS(11762), 2,
      anon_sym_POUND2,
      anon_sym_EQ2,
    ACTIONS(8050), 3,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
    ACTIONS(11754), 4,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_QMARK,
      anon_sym_AT2,
    ACTIONS(11756), 5,
      anon_sym_BANG,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_0,
      anon_sym__,
  [259297] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1298), 4,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1296), 21,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [259330] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4554), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4552), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [259363] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11431), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11926), 1,
      sym_file_descriptor,
    ACTIONS(11920), 2,
      anon_sym_LT_LT,
      anon_sym_LT_LT_DASH,
    ACTIONS(11923), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    STATE(4710), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(11420), 8,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
    ACTIONS(11917), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
  [259406] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11651), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11653), 1,
      sym__concat,
    STATE(4630), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4566), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4564), 19,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [259445] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11931), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11929), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [259478] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11760), 1,
      anon_sym_BANG2,
    ACTIONS(11764), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(11766), 1,
      anon_sym_BQUOTE,
    ACTIONS(11768), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(11770), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(11772), 1,
      sym_variable_name,
    ACTIONS(11933), 1,
      anon_sym_RBRACE3,
    STATE(3532), 1,
      sym_subscript,
    STATE(6428), 1,
      aux_sym__expansion_body_repeat1,
    STATE(6472), 1,
      sym_command_substitution,
    STATE(6922), 1,
      sym__expansion_body,
    ACTIONS(11762), 2,
      anon_sym_POUND2,
      anon_sym_EQ2,
    ACTIONS(8050), 3,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
    ACTIONS(11754), 4,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_QMARK,
      anon_sym_AT2,
    ACTIONS(11756), 5,
      anon_sym_BANG,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_0,
      anon_sym__,
  [259537] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11760), 1,
      anon_sym_BANG2,
    ACTIONS(11764), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(11766), 1,
      anon_sym_BQUOTE,
    ACTIONS(11768), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(11770), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(11772), 1,
      sym_variable_name,
    ACTIONS(11935), 1,
      anon_sym_RBRACE3,
    STATE(3532), 1,
      sym_subscript,
    STATE(6428), 1,
      aux_sym__expansion_body_repeat1,
    STATE(6472), 1,
      sym_command_substitution,
    STATE(7373), 1,
      sym__expansion_body,
    ACTIONS(11762), 2,
      anon_sym_POUND2,
      anon_sym_EQ2,
    ACTIONS(8050), 3,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
    ACTIONS(11754), 4,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_QMARK,
      anon_sym_AT2,
    ACTIONS(11756), 5,
      anon_sym_BANG,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_0,
      anon_sym__,
  [259596] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11760), 1,
      anon_sym_BANG2,
    ACTIONS(11764), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(11766), 1,
      anon_sym_BQUOTE,
    ACTIONS(11768), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(11770), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(11772), 1,
      sym_variable_name,
    ACTIONS(11937), 1,
      anon_sym_RBRACE3,
    STATE(3532), 1,
      sym_subscript,
    STATE(6428), 1,
      aux_sym__expansion_body_repeat1,
    STATE(6472), 1,
      sym_command_substitution,
    STATE(7332), 1,
      sym__expansion_body,
    ACTIONS(11762), 2,
      anon_sym_POUND2,
      anon_sym_EQ2,
    ACTIONS(8050), 3,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
    ACTIONS(11754), 4,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_QMARK,
      anon_sym_AT2,
    ACTIONS(11756), 5,
      anon_sym_BANG,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_0,
      anon_sym__,
  [259655] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4272), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    STATE(4746), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(4270), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [259690] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4253), 2,
      anon_sym_PIPE,
      anon_sym_PIPE_AMP,
    ACTIONS(4272), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    STATE(4746), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(4270), 18,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [259727] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11760), 1,
      anon_sym_BANG2,
    ACTIONS(11764), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(11766), 1,
      anon_sym_BQUOTE,
    ACTIONS(11768), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(11770), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(11772), 1,
      sym_variable_name,
    ACTIONS(11939), 1,
      anon_sym_RBRACE3,
    STATE(3532), 1,
      sym_subscript,
    STATE(6428), 1,
      aux_sym__expansion_body_repeat1,
    STATE(6472), 1,
      sym_command_substitution,
    STATE(7345), 1,
      sym__expansion_body,
    ACTIONS(11762), 2,
      anon_sym_POUND2,
      anon_sym_EQ2,
    ACTIONS(8050), 3,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
    ACTIONS(11754), 4,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_QMARK,
      anon_sym_AT2,
    ACTIONS(11756), 5,
      anon_sym_BANG,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_0,
      anon_sym__,
  [259786] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11410), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(11535), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11941), 1,
      sym_file_descriptor,
    ACTIONS(11487), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    STATE(4877), 2,
      sym_file_redirect,
      sym_herestring_redirect,
    ACTIONS(11485), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
    ACTIONS(11533), 10,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
      anon_sym_BQUOTE,
  [259829] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11760), 1,
      anon_sym_BANG2,
    ACTIONS(11764), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(11766), 1,
      anon_sym_BQUOTE,
    ACTIONS(11768), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(11770), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(11772), 1,
      sym_variable_name,
    ACTIONS(11943), 1,
      anon_sym_RBRACE3,
    STATE(3532), 1,
      sym_subscript,
    STATE(6428), 1,
      aux_sym__expansion_body_repeat1,
    STATE(6472), 1,
      sym_command_substitution,
    STATE(6891), 1,
      sym__expansion_body,
    ACTIONS(11762), 2,
      anon_sym_POUND2,
      anon_sym_EQ2,
    ACTIONS(8050), 3,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
    ACTIONS(11754), 4,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_QMARK,
      anon_sym_AT2,
    ACTIONS(11756), 5,
      anon_sym_BANG,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_0,
      anon_sym__,
  [259888] = 19,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(11808), 1,
      anon_sym_LPAREN,
    ACTIONS(11810), 1,
      aux_sym__c_word_token1,
    ACTIONS(11812), 1,
      anon_sym_DOLLAR,
    ACTIONS(11814), 1,
      anon_sym_DQUOTE,
    ACTIONS(11816), 1,
      aux_sym_number_token1,
    ACTIONS(11818), 1,
      aux_sym_number_token2,
    ACTIONS(11820), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(11822), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(11824), 1,
      anon_sym_BQUOTE,
    ACTIONS(11826), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(11945), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(3285), 1,
      sym__c_postfix_expression,
    STATE(3286), 1,
      sym__c_binary_expression,
    STATE(3290), 1,
      sym__c_unary_expression,
    STATE(6658), 1,
      sym__c_expression,
    STATE(6768), 1,
      sym__c_variable_assignment,
    ACTIONS(11806), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    STATE(3292), 7,
      sym__c_expression_not_assignment,
      sym__c_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [259953] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4253), 2,
      anon_sym_PIPE,
      anon_sym_PIPE_AMP,
    ACTIONS(4272), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    STATE(4697), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(4270), 18,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_BQUOTE,
  [259990] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11508), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(11535), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11850), 1,
      sym_file_descriptor,
    ACTIONS(11542), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    STATE(4877), 2,
      sym_file_redirect,
      sym_herestring_redirect,
    ACTIONS(11540), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
    ACTIONS(11533), 10,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
  [260033] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4272), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    STATE(4697), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(4270), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_BQUOTE,
  [260068] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11431), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11953), 1,
      sym_file_descriptor,
    ACTIONS(11920), 2,
      anon_sym_LT_LT,
      anon_sym_LT_LT_DASH,
    ACTIONS(11950), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    STATE(4725), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(11420), 8,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_BQUOTE,
    ACTIONS(11947), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
  [260111] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1263), 3,
      sym_file_descriptor,
      sym_variable_name,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1261), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      sym__special_character,
  [260144] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11956), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11958), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_BQUOTE,
  [260177] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11760), 1,
      anon_sym_BANG2,
    ACTIONS(11764), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(11766), 1,
      anon_sym_BQUOTE,
    ACTIONS(11768), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(11770), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(11772), 1,
      sym_variable_name,
    ACTIONS(11960), 1,
      anon_sym_RBRACE3,
    STATE(3532), 1,
      sym_subscript,
    STATE(6428), 1,
      aux_sym__expansion_body_repeat1,
    STATE(6472), 1,
      sym_command_substitution,
    STATE(7601), 1,
      sym__expansion_body,
    ACTIONS(11762), 2,
      anon_sym_POUND2,
      anon_sym_EQ2,
    ACTIONS(8050), 3,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
    ACTIONS(11754), 4,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_QMARK,
      anon_sym_AT2,
    ACTIONS(11756), 5,
      anon_sym_BANG,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_0,
      anon_sym__,
  [260236] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11760), 1,
      anon_sym_BANG2,
    ACTIONS(11764), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(11766), 1,
      anon_sym_BQUOTE,
    ACTIONS(11768), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(11770), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(11772), 1,
      sym_variable_name,
    ACTIONS(11962), 1,
      anon_sym_RBRACE3,
    STATE(3532), 1,
      sym_subscript,
    STATE(6428), 1,
      aux_sym__expansion_body_repeat1,
    STATE(6472), 1,
      sym_command_substitution,
    STATE(7361), 1,
      sym__expansion_body,
    ACTIONS(11762), 2,
      anon_sym_POUND2,
      anon_sym_EQ2,
    ACTIONS(8050), 3,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
    ACTIONS(11754), 4,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_QMARK,
      anon_sym_AT2,
    ACTIONS(11756), 5,
      anon_sym_BANG,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_0,
      anon_sym__,
  [260295] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11760), 1,
      anon_sym_BANG2,
    ACTIONS(11764), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(11766), 1,
      anon_sym_BQUOTE,
    ACTIONS(11768), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(11770), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(11772), 1,
      sym_variable_name,
    ACTIONS(11964), 1,
      anon_sym_RBRACE3,
    STATE(3532), 1,
      sym_subscript,
    STATE(6428), 1,
      aux_sym__expansion_body_repeat1,
    STATE(6472), 1,
      sym_command_substitution,
    STATE(7046), 1,
      sym__expansion_body,
    ACTIONS(11762), 2,
      anon_sym_POUND2,
      anon_sym_EQ2,
    ACTIONS(8050), 3,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
    ACTIONS(11754), 4,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_QMARK,
      anon_sym_AT2,
    ACTIONS(11756), 5,
      anon_sym_BANG,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_0,
      anon_sym__,
  [260354] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11760), 1,
      anon_sym_BANG2,
    ACTIONS(11764), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(11766), 1,
      anon_sym_BQUOTE,
    ACTIONS(11768), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(11770), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(11772), 1,
      sym_variable_name,
    ACTIONS(11966), 1,
      anon_sym_RBRACE3,
    STATE(3532), 1,
      sym_subscript,
    STATE(6428), 1,
      aux_sym__expansion_body_repeat1,
    STATE(6472), 1,
      sym_command_substitution,
    STATE(7374), 1,
      sym__expansion_body,
    ACTIONS(11762), 2,
      anon_sym_POUND2,
      anon_sym_EQ2,
    ACTIONS(8050), 3,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
    ACTIONS(11754), 4,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_QMARK,
      anon_sym_AT2,
    ACTIONS(11756), 5,
      anon_sym_BANG,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_0,
      anon_sym__,
  [260413] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11760), 1,
      anon_sym_BANG2,
    ACTIONS(11764), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(11766), 1,
      anon_sym_BQUOTE,
    ACTIONS(11768), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(11770), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(11772), 1,
      sym_variable_name,
    ACTIONS(11968), 1,
      anon_sym_RBRACE3,
    STATE(3532), 1,
      sym_subscript,
    STATE(6428), 1,
      aux_sym__expansion_body_repeat1,
    STATE(6472), 1,
      sym_command_substitution,
    STATE(7162), 1,
      sym__expansion_body,
    ACTIONS(11762), 2,
      anon_sym_POUND2,
      anon_sym_EQ2,
    ACTIONS(8050), 3,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
    ACTIONS(11754), 4,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_QMARK,
      anon_sym_AT2,
    ACTIONS(11756), 5,
      anon_sym_BANG,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_0,
      anon_sym__,
  [260472] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11651), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11653), 1,
      sym__concat,
    STATE(4630), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5723), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5721), 19,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [260511] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11976), 1,
      sym_file_descriptor,
    ACTIONS(11431), 2,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11920), 2,
      anon_sym_LT_LT,
      anon_sym_LT_LT_DASH,
    ACTIONS(11973), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    STATE(4734), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(11420), 7,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
    ACTIONS(11970), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
  [260554] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1314), 4,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1312), 21,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [260587] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2074), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(2072), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_BQUOTE,
  [260620] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11760), 1,
      anon_sym_BANG2,
    ACTIONS(11764), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(11766), 1,
      anon_sym_BQUOTE,
    ACTIONS(11768), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(11770), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(11772), 1,
      sym_variable_name,
    ACTIONS(11979), 1,
      anon_sym_RBRACE3,
    STATE(3532), 1,
      sym_subscript,
    STATE(6428), 1,
      aux_sym__expansion_body_repeat1,
    STATE(6472), 1,
      sym_command_substitution,
    STATE(7386), 1,
      sym__expansion_body,
    ACTIONS(11762), 2,
      anon_sym_POUND2,
      anon_sym_EQ2,
    ACTIONS(8050), 3,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
    ACTIONS(11754), 4,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_QMARK,
      anon_sym_AT2,
    ACTIONS(11756), 5,
      anon_sym_BANG,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_0,
      anon_sym__,
  [260679] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11760), 1,
      anon_sym_BANG2,
    ACTIONS(11764), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(11766), 1,
      anon_sym_BQUOTE,
    ACTIONS(11768), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(11770), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(11772), 1,
      sym_variable_name,
    ACTIONS(11981), 1,
      anon_sym_RBRACE3,
    STATE(3532), 1,
      sym_subscript,
    STATE(6428), 1,
      aux_sym__expansion_body_repeat1,
    STATE(6472), 1,
      sym_command_substitution,
    STATE(7461), 1,
      sym__expansion_body,
    ACTIONS(11762), 2,
      anon_sym_POUND2,
      anon_sym_EQ2,
    ACTIONS(8050), 3,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
    ACTIONS(11754), 4,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_QMARK,
      anon_sym_AT2,
    ACTIONS(11756), 5,
      anon_sym_BANG,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_0,
      anon_sym__,
  [260738] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1354), 3,
      sym_file_descriptor,
      sym__concat,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1352), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [260771] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11760), 1,
      anon_sym_BANG2,
    ACTIONS(11764), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(11766), 1,
      anon_sym_BQUOTE,
    ACTIONS(11768), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(11770), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(11772), 1,
      sym_variable_name,
    ACTIONS(11983), 1,
      anon_sym_RBRACE3,
    STATE(3532), 1,
      sym_subscript,
    STATE(6428), 1,
      aux_sym__expansion_body_repeat1,
    STATE(6472), 1,
      sym_command_substitution,
    STATE(7395), 1,
      sym__expansion_body,
    ACTIONS(11762), 2,
      anon_sym_POUND2,
      anon_sym_EQ2,
    ACTIONS(8050), 3,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
    ACTIONS(11754), 4,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_QMARK,
      anon_sym_AT2,
    ACTIONS(11756), 5,
      anon_sym_BANG,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_0,
      anon_sym__,
  [260830] = 19,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(11808), 1,
      anon_sym_LPAREN,
    ACTIONS(11810), 1,
      aux_sym__c_word_token1,
    ACTIONS(11812), 1,
      anon_sym_DOLLAR,
    ACTIONS(11814), 1,
      anon_sym_DQUOTE,
    ACTIONS(11816), 1,
      aux_sym_number_token1,
    ACTIONS(11818), 1,
      aux_sym_number_token2,
    ACTIONS(11820), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(11822), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(11824), 1,
      anon_sym_BQUOTE,
    ACTIONS(11826), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(11985), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(3285), 1,
      sym__c_postfix_expression,
    STATE(3286), 1,
      sym__c_binary_expression,
    STATE(3290), 1,
      sym__c_unary_expression,
    STATE(6713), 1,
      sym__c_expression,
    STATE(6768), 1,
      sym__c_variable_assignment,
    ACTIONS(11806), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    STATE(3292), 7,
      sym__c_expression_not_assignment,
      sym__c_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [260895] = 19,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(11808), 1,
      anon_sym_LPAREN,
    ACTIONS(11810), 1,
      aux_sym__c_word_token1,
    ACTIONS(11812), 1,
      anon_sym_DOLLAR,
    ACTIONS(11814), 1,
      anon_sym_DQUOTE,
    ACTIONS(11816), 1,
      aux_sym_number_token1,
    ACTIONS(11818), 1,
      aux_sym_number_token2,
    ACTIONS(11820), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(11822), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(11824), 1,
      anon_sym_BQUOTE,
    ACTIONS(11826), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(11987), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(3285), 1,
      sym__c_postfix_expression,
    STATE(3286), 1,
      sym__c_binary_expression,
    STATE(3290), 1,
      sym__c_unary_expression,
    STATE(6678), 1,
      sym__c_expression,
    STATE(6768), 1,
      sym__c_variable_assignment,
    ACTIONS(11806), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    STATE(3292), 7,
      sym__c_expression_not_assignment,
      sym__c_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [260960] = 19,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(11808), 1,
      anon_sym_LPAREN,
    ACTIONS(11810), 1,
      aux_sym__c_word_token1,
    ACTIONS(11812), 1,
      anon_sym_DOLLAR,
    ACTIONS(11814), 1,
      anon_sym_DQUOTE,
    ACTIONS(11816), 1,
      aux_sym_number_token1,
    ACTIONS(11818), 1,
      aux_sym_number_token2,
    ACTIONS(11820), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(11822), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(11824), 1,
      anon_sym_BQUOTE,
    ACTIONS(11826), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(11989), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(3285), 1,
      sym__c_postfix_expression,
    STATE(3286), 1,
      sym__c_binary_expression,
    STATE(3290), 1,
      sym__c_unary_expression,
    STATE(6632), 1,
      sym__c_expression,
    STATE(6768), 1,
      sym__c_variable_assignment,
    ACTIONS(11806), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    STATE(3292), 7,
      sym__c_expression_not_assignment,
      sym__c_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [261025] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11760), 1,
      anon_sym_BANG2,
    ACTIONS(11764), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(11766), 1,
      anon_sym_BQUOTE,
    ACTIONS(11768), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(11770), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(11772), 1,
      sym_variable_name,
    ACTIONS(11991), 1,
      anon_sym_RBRACE3,
    STATE(3532), 1,
      sym_subscript,
    STATE(6428), 1,
      aux_sym__expansion_body_repeat1,
    STATE(6472), 1,
      sym_command_substitution,
    STATE(7412), 1,
      sym__expansion_body,
    ACTIONS(11762), 2,
      anon_sym_POUND2,
      anon_sym_EQ2,
    ACTIONS(8050), 3,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
    ACTIONS(11754), 4,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_QMARK,
      anon_sym_AT2,
    ACTIONS(11756), 5,
      anon_sym_BANG,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_0,
      anon_sym__,
  [261084] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4253), 2,
      anon_sym_PIPE,
      anon_sym_PIPE_AMP,
    ACTIONS(4272), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    STATE(4650), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(4270), 17,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [261121] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11455), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11993), 1,
      sym_file_descriptor,
    ACTIONS(4692), 2,
      anon_sym_LT_LT,
      anon_sym_LT_LT_DASH,
    ACTIONS(11542), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    STATE(4710), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(11453), 8,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
    ACTIONS(11540), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
  [261164] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11410), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(11414), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11941), 1,
      sym_file_descriptor,
    ACTIONS(11487), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    STATE(4868), 2,
      sym_file_redirect,
      sym_herestring_redirect,
    ACTIONS(11485), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
    ACTIONS(11412), 10,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
      anon_sym_BQUOTE,
  [261207] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11651), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11653), 1,
      sym__concat,
    STATE(4630), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4469), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4467), 19,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [261246] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11760), 1,
      anon_sym_BANG2,
    ACTIONS(11764), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(11766), 1,
      anon_sym_BQUOTE,
    ACTIONS(11768), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(11770), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(11772), 1,
      sym_variable_name,
    ACTIONS(11995), 1,
      anon_sym_RBRACE3,
    STATE(3532), 1,
      sym_subscript,
    STATE(6428), 1,
      aux_sym__expansion_body_repeat1,
    STATE(6472), 1,
      sym_command_substitution,
    STATE(7422), 1,
      sym__expansion_body,
    ACTIONS(11762), 2,
      anon_sym_POUND2,
      anon_sym_EQ2,
    ACTIONS(8050), 3,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
    ACTIONS(11754), 4,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_QMARK,
      anon_sym_AT2,
    ACTIONS(11756), 5,
      anon_sym_BANG,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_0,
      anon_sym__,
  [261305] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11512), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11514), 1,
      sym__concat,
    STATE(4777), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4554), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4552), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [261344] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11760), 1,
      anon_sym_BANG2,
    ACTIONS(11764), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(11766), 1,
      anon_sym_BQUOTE,
    ACTIONS(11768), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(11770), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(11772), 1,
      sym_variable_name,
    ACTIONS(11997), 1,
      anon_sym_RBRACE3,
    STATE(3532), 1,
      sym_subscript,
    STATE(6428), 1,
      aux_sym__expansion_body_repeat1,
    STATE(6472), 1,
      sym_command_substitution,
    STATE(7184), 1,
      sym__expansion_body,
    ACTIONS(11762), 2,
      anon_sym_POUND2,
      anon_sym_EQ2,
    ACTIONS(8050), 3,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
    ACTIONS(11754), 4,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_QMARK,
      anon_sym_AT2,
    ACTIONS(11756), 5,
      anon_sym_BANG,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_0,
      anon_sym__,
  [261403] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11760), 1,
      anon_sym_BANG2,
    ACTIONS(11764), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(11766), 1,
      anon_sym_BQUOTE,
    ACTIONS(11768), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(11770), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(11772), 1,
      sym_variable_name,
    ACTIONS(11999), 1,
      anon_sym_RBRACE3,
    STATE(3532), 1,
      sym_subscript,
    STATE(6428), 1,
      aux_sym__expansion_body_repeat1,
    STATE(6472), 1,
      sym_command_substitution,
    STATE(7437), 1,
      sym__expansion_body,
    ACTIONS(11762), 2,
      anon_sym_POUND2,
      anon_sym_EQ2,
    ACTIONS(8050), 3,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
    ACTIONS(11754), 4,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_QMARK,
      anon_sym_AT2,
    ACTIONS(11756), 5,
      anon_sym_BANG,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_0,
      anon_sym__,
  [261462] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1322), 3,
      sym_file_descriptor,
      sym__concat,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1320), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [261495] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11512), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11514), 1,
      sym__concat,
    STATE(4776), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4566), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4564), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [261534] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12003), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12001), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [261567] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11760), 1,
      anon_sym_BANG2,
    ACTIONS(11764), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(11766), 1,
      anon_sym_BQUOTE,
    ACTIONS(11768), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(11770), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(11772), 1,
      sym_variable_name,
    ACTIONS(12005), 1,
      anon_sym_RBRACE3,
    STATE(3532), 1,
      sym_subscript,
    STATE(6428), 1,
      aux_sym__expansion_body_repeat1,
    STATE(6472), 1,
      sym_command_substitution,
    STATE(7449), 1,
      sym__expansion_body,
    ACTIONS(11762), 2,
      anon_sym_POUND2,
      anon_sym_EQ2,
    ACTIONS(8050), 3,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
    ACTIONS(11754), 4,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_QMARK,
      anon_sym_AT2,
    ACTIONS(11756), 5,
      anon_sym_BANG,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_0,
      anon_sym__,
  [261626] = 19,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(11808), 1,
      anon_sym_LPAREN,
    ACTIONS(11810), 1,
      aux_sym__c_word_token1,
    ACTIONS(11812), 1,
      anon_sym_DOLLAR,
    ACTIONS(11814), 1,
      anon_sym_DQUOTE,
    ACTIONS(11816), 1,
      aux_sym_number_token1,
    ACTIONS(11818), 1,
      aux_sym_number_token2,
    ACTIONS(11820), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(11822), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(11824), 1,
      anon_sym_BQUOTE,
    ACTIONS(11826), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(12007), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(3285), 1,
      sym__c_postfix_expression,
    STATE(3286), 1,
      sym__c_binary_expression,
    STATE(3290), 1,
      sym__c_unary_expression,
    STATE(6675), 1,
      sym__c_expression,
    STATE(6768), 1,
      sym__c_variable_assignment,
    ACTIONS(11806), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    STATE(3292), 7,
      sym__c_expression_not_assignment,
      sym__c_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [261691] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11410), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(11506), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11941), 1,
      sym_file_descriptor,
    ACTIONS(11487), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    STATE(5017), 2,
      sym_file_redirect,
      sym_herestring_redirect,
    ACTIONS(11485), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
    ACTIONS(11504), 10,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
      anon_sym_BQUOTE,
  [261734] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5375), 4,
      sym_file_descriptor,
      sym_variable_name,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5373), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_BQUOTE,
  [261767] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11651), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11653), 1,
      sym__concat,
    STATE(4611), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4372), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4370), 19,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [261806] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11760), 1,
      anon_sym_BANG2,
    ACTIONS(11764), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(11766), 1,
      anon_sym_BQUOTE,
    ACTIONS(11768), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(11770), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(11772), 1,
      sym_variable_name,
    ACTIONS(12009), 1,
      anon_sym_RBRACE3,
    STATE(3532), 1,
      sym_subscript,
    STATE(6428), 1,
      aux_sym__expansion_body_repeat1,
    STATE(6472), 1,
      sym_command_substitution,
    STATE(7464), 1,
      sym__expansion_body,
    ACTIONS(11762), 2,
      anon_sym_POUND2,
      anon_sym_EQ2,
    ACTIONS(8050), 3,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
    ACTIONS(11754), 4,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_QMARK,
      anon_sym_AT2,
    ACTIONS(11756), 5,
      anon_sym_BANG,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_0,
      anon_sym__,
  [261865] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1338), 3,
      sym_file_descriptor,
      sym__concat,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1336), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [261898] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11760), 1,
      anon_sym_BANG2,
    ACTIONS(11764), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(11766), 1,
      anon_sym_BQUOTE,
    ACTIONS(11768), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(11770), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(11772), 1,
      sym_variable_name,
    ACTIONS(12011), 1,
      anon_sym_RBRACE3,
    STATE(3532), 1,
      sym_subscript,
    STATE(6428), 1,
      aux_sym__expansion_body_repeat1,
    STATE(6472), 1,
      sym_command_substitution,
    STATE(7255), 1,
      sym__expansion_body,
    ACTIONS(11762), 2,
      anon_sym_POUND2,
      anon_sym_EQ2,
    ACTIONS(8050), 3,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
    ACTIONS(11754), 4,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_QMARK,
      anon_sym_AT2,
    ACTIONS(11756), 5,
      anon_sym_BANG,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_0,
      anon_sym__,
  [261957] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11760), 1,
      anon_sym_BANG2,
    ACTIONS(11764), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(11766), 1,
      anon_sym_BQUOTE,
    ACTIONS(11768), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(11770), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(11772), 1,
      sym_variable_name,
    ACTIONS(12013), 1,
      anon_sym_RBRACE3,
    STATE(3532), 1,
      sym_subscript,
    STATE(6428), 1,
      aux_sym__expansion_body_repeat1,
    STATE(6472), 1,
      sym_command_substitution,
    STATE(7476), 1,
      sym__expansion_body,
    ACTIONS(11762), 2,
      anon_sym_POUND2,
      anon_sym_EQ2,
    ACTIONS(8050), 3,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
    ACTIONS(11754), 4,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_QMARK,
      anon_sym_AT2,
    ACTIONS(11756), 5,
      anon_sym_BANG,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_0,
      anon_sym__,
  [262016] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11760), 1,
      anon_sym_BANG2,
    ACTIONS(11764), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(11766), 1,
      anon_sym_BQUOTE,
    ACTIONS(11768), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(11770), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(11772), 1,
      sym_variable_name,
    ACTIONS(12015), 1,
      anon_sym_RBRACE3,
    STATE(3532), 1,
      sym_subscript,
    STATE(6428), 1,
      aux_sym__expansion_body_repeat1,
    STATE(6472), 1,
      sym_command_substitution,
    STATE(7331), 1,
      sym__expansion_body,
    ACTIONS(11762), 2,
      anon_sym_POUND2,
      anon_sym_EQ2,
    ACTIONS(8050), 3,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
    ACTIONS(11754), 4,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_QMARK,
      anon_sym_AT2,
    ACTIONS(11756), 5,
      anon_sym_BANG,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_0,
      anon_sym__,
  [262075] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1338), 4,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1336), 21,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [262108] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11802), 1,
      sym__special_character,
    STATE(4855), 1,
      aux_sym__literal_repeat1,
    ACTIONS(5697), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5695), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [262145] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11760), 1,
      anon_sym_BANG2,
    ACTIONS(11764), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(11766), 1,
      anon_sym_BQUOTE,
    ACTIONS(11768), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(11770), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(11772), 1,
      sym_variable_name,
    ACTIONS(12017), 1,
      anon_sym_RBRACE3,
    STATE(3532), 1,
      sym_subscript,
    STATE(6428), 1,
      aux_sym__expansion_body_repeat1,
    STATE(6472), 1,
      sym_command_substitution,
    STATE(7493), 1,
      sym__expansion_body,
    ACTIONS(11762), 2,
      anon_sym_POUND2,
      anon_sym_EQ2,
    ACTIONS(8050), 3,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
    ACTIONS(11754), 4,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_QMARK,
      anon_sym_AT2,
    ACTIONS(11756), 5,
      anon_sym_BANG,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_0,
      anon_sym__,
  [262204] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12003), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12001), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_BQUOTE,
  [262237] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1342), 3,
      sym_file_descriptor,
      sym__concat,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1340), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [262270] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11651), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11653), 1,
      sym__concat,
    STATE(4674), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4566), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4564), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_BQUOTE,
  [262309] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1263), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1261), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      sym__special_character,
  [262342] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11760), 1,
      anon_sym_BANG2,
    ACTIONS(11764), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(11766), 1,
      anon_sym_BQUOTE,
    ACTIONS(11768), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(11770), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(11772), 1,
      sym_variable_name,
    ACTIONS(12019), 1,
      anon_sym_RBRACE3,
    STATE(3532), 1,
      sym_subscript,
    STATE(6428), 1,
      aux_sym__expansion_body_repeat1,
    STATE(6472), 1,
      sym_command_substitution,
    STATE(7501), 1,
      sym__expansion_body,
    ACTIONS(11762), 2,
      anon_sym_POUND2,
      anon_sym_EQ2,
    ACTIONS(8050), 3,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
    ACTIONS(11754), 4,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_QMARK,
      anon_sym_AT2,
    ACTIONS(11756), 5,
      anon_sym_BANG,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_0,
      anon_sym__,
  [262401] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11651), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11653), 1,
      sym__concat,
    STATE(5031), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1263), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1261), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      sym__special_character,
  [262440] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11500), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11502), 1,
      sym__concat,
    STATE(4781), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5069), 3,
      sym_file_descriptor,
      sym_variable_name,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5067), 19,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [262479] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11512), 1,
      aux_sym_concatenation_token1,
    ACTIONS(12021), 1,
      sym__concat,
    STATE(4253), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1267), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1265), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [262518] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11512), 1,
      aux_sym_concatenation_token1,
    ACTIONS(12023), 1,
      sym__concat,
    STATE(4253), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1288), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1286), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [262557] = 11,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2719), 1,
      anon_sym_RPAREN,
    ACTIONS(4384), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11993), 1,
      sym_file_descriptor,
    ACTIONS(4253), 2,
      anon_sym_PIPE,
      anon_sym_PIPE_AMP,
    ACTIONS(4692), 2,
      anon_sym_LT_LT,
      anon_sym_LT_LT_DASH,
    ACTIONS(4809), 2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
    ACTIONS(11542), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(4378), 3,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
    STATE(4746), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(11540), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
  [262606] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11760), 1,
      anon_sym_BANG2,
    ACTIONS(11764), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(11766), 1,
      anon_sym_BQUOTE,
    ACTIONS(11768), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(11770), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(11772), 1,
      sym_variable_name,
    ACTIONS(12025), 1,
      anon_sym_RBRACE3,
    STATE(3532), 1,
      sym_subscript,
    STATE(6428), 1,
      aux_sym__expansion_body_repeat1,
    STATE(6472), 1,
      sym_command_substitution,
    STATE(7348), 1,
      sym__expansion_body,
    ACTIONS(11762), 2,
      anon_sym_POUND2,
      anon_sym_EQ2,
    ACTIONS(8050), 3,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
    ACTIONS(11754), 4,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_QMARK,
      anon_sym_AT2,
    ACTIONS(11756), 5,
      anon_sym_BANG,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_0,
      anon_sym__,
  [262665] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1314), 3,
      sym_file_descriptor,
      sym__concat,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1312), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [262698] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11500), 1,
      aux_sym_concatenation_token1,
    ACTIONS(12027), 1,
      sym__concat,
    STATE(4244), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1267), 3,
      sym_file_descriptor,
      sym_variable_name,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1265), 19,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [262737] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2096), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(2094), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_BQUOTE,
  [262770] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11500), 1,
      aux_sym_concatenation_token1,
    ACTIONS(12029), 1,
      sym__concat,
    STATE(4244), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1288), 3,
      sym_file_descriptor,
      sym_variable_name,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1286), 19,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [262809] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12031), 1,
      sym_variable_name,
    STATE(6727), 1,
      sym_subscript,
    ACTIONS(11290), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    STATE(4784), 2,
      sym_variable_assignment,
      aux_sym_variable_assignments_repeat1,
    ACTIONS(11288), 19,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [262848] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12036), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12034), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [262881] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11956), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11958), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [262914] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12038), 1,
      sym_variable_name,
    STATE(6727), 1,
      sym_subscript,
    ACTIONS(11329), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    STATE(4784), 2,
      sym_variable_assignment,
      aux_sym_variable_assignments_repeat1,
    ACTIONS(11327), 19,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [262953] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1263), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1261), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      sym__special_character,
  [262986] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2074), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(2072), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [263019] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1334), 3,
      sym_file_descriptor,
      sym__concat,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1332), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [263052] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11760), 1,
      anon_sym_BANG2,
    ACTIONS(11764), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(11766), 1,
      anon_sym_BQUOTE,
    ACTIONS(11768), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(11770), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(11772), 1,
      sym_variable_name,
    ACTIONS(12040), 1,
      anon_sym_RBRACE3,
    STATE(3532), 1,
      sym_subscript,
    STATE(6428), 1,
      aux_sym__expansion_body_repeat1,
    STATE(6472), 1,
      sym_command_substitution,
    STATE(7510), 1,
      sym__expansion_body,
    ACTIONS(11762), 2,
      anon_sym_POUND2,
      anon_sym_EQ2,
    ACTIONS(8050), 3,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
    ACTIONS(11754), 4,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_QMARK,
      anon_sym_AT2,
    ACTIONS(11756), 5,
      anon_sym_BANG,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_0,
      anon_sym__,
  [263111] = 10,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4340), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11457), 1,
      sym_file_descriptor,
    ACTIONS(4251), 2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
    ACTIONS(4255), 2,
      anon_sym_LT_LT,
      anon_sym_LT_LT_DASH,
    ACTIONS(4338), 2,
      anon_sym_SEMI,
      anon_sym_AMP,
    ACTIONS(11282), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    STATE(4283), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(2719), 4,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
    ACTIONS(11280), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
  [263158] = 11,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2719), 1,
      anon_sym_BQUOTE,
    ACTIONS(5357), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11856), 1,
      sym_file_descriptor,
    ACTIONS(4253), 2,
      anon_sym_PIPE,
      anon_sym_PIPE_AMP,
    ACTIONS(4692), 2,
      anon_sym_LT_LT,
      anon_sym_LT_LT_DASH,
    ACTIONS(5355), 2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
    ACTIONS(11487), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(5353), 3,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
    STATE(4697), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(11485), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
  [263207] = 5,
    ACTIONS(3), 1,
      sym_comment,
    STATE(5602), 1,
      aux_sym_pipeline_repeat1,
    ACTIONS(11745), 2,
      anon_sym_PIPE,
      anon_sym_PIPE_AMP,
    ACTIONS(11748), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11743), 19,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_BQUOTE,
  [263244] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11651), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11653), 1,
      sym__concat,
    STATE(4630), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5697), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5695), 19,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [263283] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1294), 3,
      sym_file_descriptor,
      sym__concat,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1292), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [263316] = 11,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4257), 1,
      anon_sym_RPAREN,
    ACTIONS(4411), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11993), 1,
      sym_file_descriptor,
    ACTIONS(4253), 2,
      anon_sym_PIPE,
      anon_sym_PIPE_AMP,
    ACTIONS(4692), 2,
      anon_sym_LT_LT,
      anon_sym_LT_LT_DASH,
    ACTIONS(4809), 2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
    ACTIONS(11542), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(4409), 3,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
    STATE(4746), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(11540), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
  [263365] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11760), 1,
      anon_sym_BANG2,
    ACTIONS(11764), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(11766), 1,
      anon_sym_BQUOTE,
    ACTIONS(11768), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(11770), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(11772), 1,
      sym_variable_name,
    ACTIONS(12042), 1,
      anon_sym_RBRACE3,
    STATE(3532), 1,
      sym_subscript,
    STATE(6428), 1,
      aux_sym__expansion_body_repeat1,
    STATE(6472), 1,
      sym_command_substitution,
    STATE(7128), 1,
      sym__expansion_body,
    ACTIONS(11762), 2,
      anon_sym_POUND2,
      anon_sym_EQ2,
    ACTIONS(8050), 3,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
    ACTIONS(11754), 4,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_QMARK,
      anon_sym_AT2,
    ACTIONS(11756), 5,
      anon_sym_BANG,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_0,
      anon_sym__,
  [263424] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11760), 1,
      anon_sym_BANG2,
    ACTIONS(11764), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(11766), 1,
      anon_sym_BQUOTE,
    ACTIONS(11768), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(11770), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(11772), 1,
      sym_variable_name,
    ACTIONS(12044), 1,
      anon_sym_RBRACE3,
    STATE(3532), 1,
      sym_subscript,
    STATE(6428), 1,
      aux_sym__expansion_body_repeat1,
    STATE(6472), 1,
      sym_command_substitution,
    STATE(6925), 1,
      sym__expansion_body,
    ACTIONS(11762), 2,
      anon_sym_POUND2,
      anon_sym_EQ2,
    ACTIONS(8050), 3,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
    ACTIONS(11754), 4,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_QMARK,
      anon_sym_AT2,
    ACTIONS(11756), 5,
      anon_sym_BANG,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_0,
      anon_sym__,
  [263483] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11760), 1,
      anon_sym_BANG2,
    ACTIONS(11764), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(11766), 1,
      anon_sym_BQUOTE,
    ACTIONS(11768), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(11770), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(11772), 1,
      sym_variable_name,
    ACTIONS(12046), 1,
      anon_sym_RBRACE3,
    STATE(3532), 1,
      sym_subscript,
    STATE(6428), 1,
      aux_sym__expansion_body_repeat1,
    STATE(6472), 1,
      sym_command_substitution,
    STATE(7533), 1,
      sym__expansion_body,
    ACTIONS(11762), 2,
      anon_sym_POUND2,
      anon_sym_EQ2,
    ACTIONS(8050), 3,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
    ACTIONS(11754), 4,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_QMARK,
      anon_sym_AT2,
    ACTIONS(11756), 5,
      anon_sym_BANG,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_0,
      anon_sym__,
  [263542] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1294), 4,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1292), 21,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [263575] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11500), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11502), 1,
      sym__concat,
    STATE(4783), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5375), 3,
      sym_file_descriptor,
      sym_variable_name,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5373), 19,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [263614] = 19,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(11808), 1,
      anon_sym_LPAREN,
    ACTIONS(11810), 1,
      aux_sym__c_word_token1,
    ACTIONS(11812), 1,
      anon_sym_DOLLAR,
    ACTIONS(11814), 1,
      anon_sym_DQUOTE,
    ACTIONS(11816), 1,
      aux_sym_number_token1,
    ACTIONS(11818), 1,
      aux_sym_number_token2,
    ACTIONS(11820), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(11822), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(11824), 1,
      anon_sym_BQUOTE,
    ACTIONS(11826), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(12048), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(3285), 1,
      sym__c_postfix_expression,
    STATE(3286), 1,
      sym__c_binary_expression,
    STATE(3290), 1,
      sym__c_unary_expression,
    STATE(6648), 1,
      sym__c_expression,
    STATE(6768), 1,
      sym__c_variable_assignment,
    ACTIONS(11806), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    STATE(3292), 7,
      sym__c_expression_not_assignment,
      sym__c_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [263679] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5768), 3,
      sym_file_descriptor,
      sym_variable_name,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5766), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [263712] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1326), 3,
      sym_file_descriptor,
      sym__concat,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1324), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [263745] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11512), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11514), 1,
      sym__concat,
    STATE(4777), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4372), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4370), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [263784] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1346), 4,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1344), 21,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [263817] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1350), 4,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1348), 21,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [263850] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11760), 1,
      anon_sym_BANG2,
    ACTIONS(11764), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(11766), 1,
      anon_sym_BQUOTE,
    ACTIONS(11768), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(11770), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(11772), 1,
      sym_variable_name,
    ACTIONS(12050), 1,
      anon_sym_RBRACE3,
    STATE(3532), 1,
      sym_subscript,
    STATE(6428), 1,
      aux_sym__expansion_body_repeat1,
    STATE(6472), 1,
      sym_command_substitution,
    STATE(6807), 1,
      sym__expansion_body,
    ACTIONS(11762), 2,
      anon_sym_POUND2,
      anon_sym_EQ2,
    ACTIONS(8050), 3,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
    ACTIONS(11754), 4,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_QMARK,
      anon_sym_AT2,
    ACTIONS(11756), 5,
      anon_sym_BANG,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_0,
      anon_sym__,
  [263909] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1253), 4,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1251), 21,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [263942] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1326), 4,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1324), 21,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [263975] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1358), 4,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1356), 21,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [264008] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11760), 1,
      anon_sym_BANG2,
    ACTIONS(11764), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(11766), 1,
      anon_sym_BQUOTE,
    ACTIONS(11768), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(11770), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(11772), 1,
      sym_variable_name,
    ACTIONS(12052), 1,
      anon_sym_RBRACE3,
    STATE(3532), 1,
      sym_subscript,
    STATE(6428), 1,
      aux_sym__expansion_body_repeat1,
    STATE(6472), 1,
      sym_command_substitution,
    STATE(7398), 1,
      sym__expansion_body,
    ACTIONS(11762), 2,
      anon_sym_POUND2,
      anon_sym_EQ2,
    ACTIONS(8050), 3,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
    ACTIONS(11754), 4,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_QMARK,
      anon_sym_AT2,
    ACTIONS(11756), 5,
      anon_sym_BANG,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_0,
      anon_sym__,
  [264067] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11760), 1,
      anon_sym_BANG2,
    ACTIONS(11764), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(11766), 1,
      anon_sym_BQUOTE,
    ACTIONS(11768), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(11770), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(11772), 1,
      sym_variable_name,
    ACTIONS(12054), 1,
      anon_sym_RBRACE3,
    STATE(3532), 1,
      sym_subscript,
    STATE(6428), 1,
      aux_sym__expansion_body_repeat1,
    STATE(6472), 1,
      sym_command_substitution,
    STATE(7113), 1,
      sym__expansion_body,
    ACTIONS(11762), 2,
      anon_sym_POUND2,
      anon_sym_EQ2,
    ACTIONS(8050), 3,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
    ACTIONS(11754), 4,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_QMARK,
      anon_sym_AT2,
    ACTIONS(11756), 5,
      anon_sym_BANG,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_0,
      anon_sym__,
  [264126] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11760), 1,
      anon_sym_BANG2,
    ACTIONS(11764), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(11766), 1,
      anon_sym_BQUOTE,
    ACTIONS(11768), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(11770), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(11772), 1,
      sym_variable_name,
    ACTIONS(12056), 1,
      anon_sym_RBRACE3,
    STATE(3532), 1,
      sym_subscript,
    STATE(6428), 1,
      aux_sym__expansion_body_repeat1,
    STATE(6472), 1,
      sym_command_substitution,
    STATE(7524), 1,
      sym__expansion_body,
    ACTIONS(11762), 2,
      anon_sym_POUND2,
      anon_sym_EQ2,
    ACTIONS(8050), 3,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
    ACTIONS(11754), 4,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_QMARK,
      anon_sym_AT2,
    ACTIONS(11756), 5,
      anon_sym_BANG,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_0,
      anon_sym__,
  [264185] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11512), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11514), 1,
      sym__concat,
    STATE(4776), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5723), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5721), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [264224] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11760), 1,
      anon_sym_BANG2,
    ACTIONS(11764), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(11766), 1,
      anon_sym_BQUOTE,
    ACTIONS(11768), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(11770), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(11772), 1,
      sym_variable_name,
    ACTIONS(12058), 1,
      anon_sym_RBRACE3,
    STATE(3532), 1,
      sym_subscript,
    STATE(6428), 1,
      aux_sym__expansion_body_repeat1,
    STATE(6472), 1,
      sym_command_substitution,
    STATE(7313), 1,
      sym__expansion_body,
    ACTIONS(11762), 2,
      anon_sym_POUND2,
      anon_sym_EQ2,
    ACTIONS(8050), 3,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
    ACTIONS(11754), 4,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_QMARK,
      anon_sym_AT2,
    ACTIONS(11756), 5,
      anon_sym_BANG,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_0,
      anon_sym__,
  [264283] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11750), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11752), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [264316] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11760), 1,
      anon_sym_BANG2,
    ACTIONS(11764), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(11766), 1,
      anon_sym_BQUOTE,
    ACTIONS(11768), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(11770), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(11772), 1,
      sym_variable_name,
    ACTIONS(12060), 1,
      anon_sym_RBRACE3,
    STATE(3532), 1,
      sym_subscript,
    STATE(6428), 1,
      aux_sym__expansion_body_repeat1,
    STATE(6472), 1,
      sym_command_substitution,
    STATE(7380), 1,
      sym__expansion_body,
    ACTIONS(11762), 2,
      anon_sym_POUND2,
      anon_sym_EQ2,
    ACTIONS(8050), 3,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
    ACTIONS(11754), 4,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_QMARK,
      anon_sym_AT2,
    ACTIONS(11756), 5,
      anon_sym_BANG,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_0,
      anon_sym__,
  [264375] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2096), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(2094), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [264408] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11842), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11844), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [264441] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11512), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11514), 1,
      sym__concat,
    STATE(4776), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4469), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4467), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [264480] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11760), 1,
      anon_sym_BANG2,
    ACTIONS(11764), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(11766), 1,
      anon_sym_BQUOTE,
    ACTIONS(11768), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(11770), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(11772), 1,
      sym_variable_name,
    ACTIONS(12062), 1,
      anon_sym_RBRACE3,
    STATE(3532), 1,
      sym_subscript,
    STATE(6428), 1,
      aux_sym__expansion_body_repeat1,
    STATE(6472), 1,
      sym_command_substitution,
    STATE(7298), 1,
      sym__expansion_body,
    ACTIONS(11762), 2,
      anon_sym_POUND2,
      anon_sym_EQ2,
    ACTIONS(8050), 3,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
    ACTIONS(11754), 4,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_QMARK,
      anon_sym_AT2,
    ACTIONS(11756), 5,
      anon_sym_BANG,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_0,
      anon_sym__,
  [264539] = 5,
    ACTIONS(3), 1,
      sym_comment,
    STATE(4664), 1,
      aux_sym_pipeline_repeat1,
    ACTIONS(12066), 2,
      anon_sym_PIPE,
      anon_sym_PIPE_AMP,
    ACTIONS(12068), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12064), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [264576] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1354), 4,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1352), 21,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [264609] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11760), 1,
      anon_sym_BANG2,
    ACTIONS(11764), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(11766), 1,
      anon_sym_BQUOTE,
    ACTIONS(11768), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(11770), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(11772), 1,
      sym_variable_name,
    ACTIONS(12070), 1,
      anon_sym_RBRACE3,
    STATE(3532), 1,
      sym_subscript,
    STATE(6428), 1,
      aux_sym__expansion_body_repeat1,
    STATE(6472), 1,
      sym_command_substitution,
    STATE(7197), 1,
      sym__expansion_body,
    ACTIONS(11762), 2,
      anon_sym_POUND2,
      anon_sym_EQ2,
    ACTIONS(8050), 3,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
    ACTIONS(11754), 4,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_QMARK,
      anon_sym_AT2,
    ACTIONS(11756), 5,
      anon_sym_BANG,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_0,
      anon_sym__,
  [264668] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4372), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4370), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [264701] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11651), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11653), 1,
      sym__concat,
    STATE(4635), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(2096), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(2094), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [264740] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11448), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(11774), 1,
      sym_file_descriptor,
    ACTIONS(11481), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(11506), 2,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    STATE(5017), 2,
      sym_file_redirect,
      sym_herestring_redirect,
    ACTIONS(11479), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
    ACTIONS(11504), 9,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
  [264783] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11760), 1,
      anon_sym_BANG2,
    ACTIONS(11764), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(11766), 1,
      anon_sym_BQUOTE,
    ACTIONS(11768), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(11770), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(11772), 1,
      sym_variable_name,
    ACTIONS(12072), 1,
      anon_sym_RBRACE3,
    STATE(3532), 1,
      sym_subscript,
    STATE(6428), 1,
      aux_sym__expansion_body_repeat1,
    STATE(6472), 1,
      sym_command_substitution,
    STATE(7308), 1,
      sym__expansion_body,
    ACTIONS(11762), 2,
      anon_sym_POUND2,
      anon_sym_EQ2,
    ACTIONS(8050), 3,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
    ACTIONS(11754), 4,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_QMARK,
      anon_sym_AT2,
    ACTIONS(11756), 5,
      anon_sym_BANG,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_0,
      anon_sym__,
  [264842] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1253), 3,
      sym_file_descriptor,
      sym__concat,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1251), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [264875] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11760), 1,
      anon_sym_BANG2,
    ACTIONS(11764), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(11766), 1,
      anon_sym_BQUOTE,
    ACTIONS(11768), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(11770), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(11772), 1,
      sym_variable_name,
    ACTIONS(12074), 1,
      anon_sym_RBRACE3,
    STATE(3532), 1,
      sym_subscript,
    STATE(6428), 1,
      aux_sym__expansion_body_repeat1,
    STATE(6472), 1,
      sym_command_substitution,
    STATE(7137), 1,
      sym__expansion_body,
    ACTIONS(11762), 2,
      anon_sym_POUND2,
      anon_sym_EQ2,
    ACTIONS(8050), 3,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
    ACTIONS(11754), 4,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_QMARK,
      anon_sym_AT2,
    ACTIONS(11756), 5,
      anon_sym_BANG,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_0,
      anon_sym__,
  [264934] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1334), 4,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1332), 21,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [264967] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11512), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11514), 1,
      sym__concat,
    STATE(4777), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(2096), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(2094), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [265006] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11512), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11514), 1,
      sym__concat,
    STATE(4776), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5697), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5695), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [265045] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11760), 1,
      anon_sym_BANG2,
    ACTIONS(11764), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(11766), 1,
      anon_sym_BQUOTE,
    ACTIONS(11768), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(11770), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(11772), 1,
      sym_variable_name,
    ACTIONS(12076), 1,
      anon_sym_RBRACE3,
    STATE(3532), 1,
      sym_subscript,
    STATE(6428), 1,
      aux_sym__expansion_body_repeat1,
    STATE(6472), 1,
      sym_command_substitution,
    STATE(6904), 1,
      sym__expansion_body,
    ACTIONS(11762), 2,
      anon_sym_POUND2,
      anon_sym_EQ2,
    ACTIONS(8050), 3,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
    ACTIONS(11754), 4,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_QMARK,
      anon_sym_AT2,
    ACTIONS(11756), 5,
      anon_sym_BANG,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_0,
      anon_sym__,
  [265104] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1318), 4,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1316), 21,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [265137] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11512), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11514), 1,
      sym__concat,
    STATE(4777), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(2074), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(2072), 20,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [265176] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11760), 1,
      anon_sym_BANG2,
    ACTIONS(11764), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(11766), 1,
      anon_sym_BQUOTE,
    ACTIONS(11768), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(11770), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(11772), 1,
      sym_variable_name,
    ACTIONS(12078), 1,
      anon_sym_RBRACE3,
    STATE(3532), 1,
      sym_subscript,
    STATE(6428), 1,
      aux_sym__expansion_body_repeat1,
    STATE(6472), 1,
      sym_command_substitution,
    STATE(6895), 1,
      sym__expansion_body,
    ACTIONS(11762), 2,
      anon_sym_POUND2,
      anon_sym_EQ2,
    ACTIONS(8050), 3,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
    ACTIONS(11754), 4,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_QMARK,
      anon_sym_AT2,
    ACTIONS(11756), 5,
      anon_sym_BANG,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_0,
      anon_sym__,
  [265235] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5772), 3,
      sym_file_descriptor,
      sym_variable_name,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5770), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [265268] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11760), 1,
      anon_sym_BANG2,
    ACTIONS(11764), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(11766), 1,
      anon_sym_BQUOTE,
    ACTIONS(11768), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(11770), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(11772), 1,
      sym_variable_name,
    ACTIONS(12080), 1,
      anon_sym_RBRACE3,
    STATE(3532), 1,
      sym_subscript,
    STATE(6428), 1,
      aux_sym__expansion_body_repeat1,
    STATE(6472), 1,
      sym_command_substitution,
    STATE(7377), 1,
      sym__expansion_body,
    ACTIONS(11762), 2,
      anon_sym_POUND2,
      anon_sym_EQ2,
    ACTIONS(8050), 3,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
    ACTIONS(11754), 4,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_QMARK,
      anon_sym_AT2,
    ACTIONS(11756), 5,
      anon_sym_BANG,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_0,
      anon_sym__,
  [265327] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1342), 4,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1340), 21,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [265360] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11760), 1,
      anon_sym_BANG2,
    ACTIONS(11764), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(11766), 1,
      anon_sym_BQUOTE,
    ACTIONS(11768), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(11770), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(11772), 1,
      sym_variable_name,
    ACTIONS(12082), 1,
      anon_sym_RBRACE3,
    STATE(3532), 1,
      sym_subscript,
    STATE(6428), 1,
      aux_sym__expansion_body_repeat1,
    STATE(6472), 1,
      sym_command_substitution,
    STATE(7260), 1,
      sym__expansion_body,
    ACTIONS(11762), 2,
      anon_sym_POUND2,
      anon_sym_EQ2,
    ACTIONS(8050), 3,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
    ACTIONS(11754), 4,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_QMARK,
      anon_sym_AT2,
    ACTIONS(11756), 5,
      anon_sym_BANG,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_0,
      anon_sym__,
  [265419] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11739), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11741), 23,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [265452] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11760), 1,
      anon_sym_BANG2,
    ACTIONS(11764), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(11766), 1,
      anon_sym_BQUOTE,
    ACTIONS(11768), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(11770), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(11772), 1,
      sym_variable_name,
    ACTIONS(12084), 1,
      anon_sym_RBRACE3,
    STATE(3532), 1,
      sym_subscript,
    STATE(6428), 1,
      aux_sym__expansion_body_repeat1,
    STATE(6472), 1,
      sym_command_substitution,
    STATE(7131), 1,
      sym__expansion_body,
    ACTIONS(11762), 2,
      anon_sym_POUND2,
      anon_sym_EQ2,
    ACTIONS(8050), 3,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
    ACTIONS(11754), 4,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_QMARK,
      anon_sym_AT2,
    ACTIONS(11756), 5,
      anon_sym_BANG,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_0,
      anon_sym__,
  [265511] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1263), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1261), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      sym__special_character,
      anon_sym_BQUOTE,
  [265544] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1322), 4,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1320), 21,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [265577] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11760), 1,
      anon_sym_BANG2,
    ACTIONS(11764), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(11766), 1,
      anon_sym_BQUOTE,
    ACTIONS(11768), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(11770), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(11772), 1,
      sym_variable_name,
    ACTIONS(12086), 1,
      anon_sym_RBRACE3,
    STATE(3532), 1,
      sym_subscript,
    STATE(6428), 1,
      aux_sym__expansion_body_repeat1,
    STATE(6472), 1,
      sym_command_substitution,
    STATE(6842), 1,
      sym__expansion_body,
    ACTIONS(11762), 2,
      anon_sym_POUND2,
      anon_sym_EQ2,
    ACTIONS(8050), 3,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
    ACTIONS(11754), 4,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_QMARK,
      anon_sym_AT2,
    ACTIONS(11756), 5,
      anon_sym_BANG,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_0,
      anon_sym__,
  [265636] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1310), 3,
      sym_file_descriptor,
      sym__concat,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1308), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [265669] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11760), 1,
      anon_sym_BANG2,
    ACTIONS(11764), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(11766), 1,
      anon_sym_BQUOTE,
    ACTIONS(11768), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(11770), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(11772), 1,
      sym_variable_name,
    ACTIONS(12088), 1,
      anon_sym_RBRACE3,
    STATE(3532), 1,
      sym_subscript,
    STATE(6428), 1,
      aux_sym__expansion_body_repeat1,
    STATE(6472), 1,
      sym_command_substitution,
    STATE(7104), 1,
      sym__expansion_body,
    ACTIONS(11762), 2,
      anon_sym_POUND2,
      anon_sym_EQ2,
    ACTIONS(8050), 3,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
    ACTIONS(11754), 4,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_QMARK,
      anon_sym_AT2,
    ACTIONS(11756), 5,
      anon_sym_BANG,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_0,
      anon_sym__,
  [265728] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11760), 1,
      anon_sym_BANG2,
    ACTIONS(11764), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(11766), 1,
      anon_sym_BQUOTE,
    ACTIONS(11768), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(11770), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(11772), 1,
      sym_variable_name,
    ACTIONS(12090), 1,
      anon_sym_RBRACE3,
    STATE(3532), 1,
      sym_subscript,
    STATE(6428), 1,
      aux_sym__expansion_body_repeat1,
    STATE(6472), 1,
      sym_command_substitution,
    STATE(6800), 1,
      sym__expansion_body,
    ACTIONS(11762), 2,
      anon_sym_POUND2,
      anon_sym_EQ2,
    ACTIONS(8050), 3,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
    ACTIONS(11754), 4,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_QMARK,
      anon_sym_AT2,
    ACTIONS(11756), 5,
      anon_sym_BANG,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_0,
      anon_sym__,
  [265787] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1306), 3,
      sym_file_descriptor,
      sym__concat,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1304), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [265820] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1310), 3,
      sym_file_descriptor,
      sym__concat,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1308), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [265853] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11760), 1,
      anon_sym_BANG2,
    ACTIONS(11764), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(11766), 1,
      anon_sym_BQUOTE,
    ACTIONS(11768), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(11770), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(11772), 1,
      sym_variable_name,
    ACTIONS(12092), 1,
      anon_sym_RBRACE3,
    STATE(3532), 1,
      sym_subscript,
    STATE(6428), 1,
      aux_sym__expansion_body_repeat1,
    STATE(6472), 1,
      sym_command_substitution,
    STATE(6832), 1,
      sym__expansion_body,
    ACTIONS(11762), 2,
      anon_sym_POUND2,
      anon_sym_EQ2,
    ACTIONS(8050), 3,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
    ACTIONS(11754), 4,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_QMARK,
      anon_sym_AT2,
    ACTIONS(11756), 5,
      anon_sym_BANG,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_0,
      anon_sym__,
  [265912] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12094), 1,
      sym__special_character,
    STATE(4855), 1,
      aux_sym__literal_repeat1,
    ACTIONS(1364), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1362), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [265949] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11931), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11929), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_BQUOTE,
  [265982] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11760), 1,
      anon_sym_BANG2,
    ACTIONS(11764), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(11766), 1,
      anon_sym_BQUOTE,
    ACTIONS(11768), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(11770), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(11772), 1,
      sym_variable_name,
    ACTIONS(12097), 1,
      anon_sym_RBRACE3,
    STATE(3532), 1,
      sym_subscript,
    STATE(6428), 1,
      aux_sym__expansion_body_repeat1,
    STATE(6472), 1,
      sym_command_substitution,
    STATE(6841), 1,
      sym__expansion_body,
    ACTIONS(11762), 2,
      anon_sym_POUND2,
      anon_sym_EQ2,
    ACTIONS(8050), 3,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
    ACTIONS(11754), 4,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_QMARK,
      anon_sym_AT2,
    ACTIONS(11756), 5,
      anon_sym_BANG,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_0,
      anon_sym__,
  [266041] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12036), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12034), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_BQUOTE,
  [266074] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1350), 3,
      sym_file_descriptor,
      sym__concat,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1348), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [266107] = 11,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4686), 1,
      ts_builtin_sym_end,
    ACTIONS(4694), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11834), 1,
      sym_file_descriptor,
    ACTIONS(4253), 2,
      anon_sym_PIPE,
      anon_sym_PIPE_AMP,
    ACTIONS(4690), 2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
    ACTIONS(4692), 2,
      anon_sym_LT_LT,
      anon_sym_LT_LT_DASH,
    ACTIONS(11481), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(4688), 3,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
    STATE(4650), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(11479), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
  [266156] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4372), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4370), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      anon_sym_BQUOTE,
  [266189] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1330), 4,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1328), 21,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [266222] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11760), 1,
      anon_sym_BANG2,
    ACTIONS(11764), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(11766), 1,
      anon_sym_BQUOTE,
    ACTIONS(11768), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(11770), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(11772), 1,
      sym_variable_name,
    ACTIONS(12099), 1,
      anon_sym_RBRACE3,
    STATE(3532), 1,
      sym_subscript,
    STATE(6428), 1,
      aux_sym__expansion_body_repeat1,
    STATE(6472), 1,
      sym_command_substitution,
    STATE(7051), 1,
      sym__expansion_body,
    ACTIONS(11762), 2,
      anon_sym_POUND2,
      anon_sym_EQ2,
    ACTIONS(8050), 3,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
    ACTIONS(11754), 4,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_QMARK,
      anon_sym_AT2,
    ACTIONS(11756), 5,
      anon_sym_BANG,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_0,
      anon_sym__,
  [266281] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11931), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11929), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [266313] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11830), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11828), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_BQUOTE,
  [266345] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12103), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12101), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [266377] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12107), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12105), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [266409] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12109), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12111), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_BQUOTE,
  [266441] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4554), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4552), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_BQUOTE,
  [266473] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5768), 3,
      sym_file_descriptor,
      sym_variable_name,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5766), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [266505] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12113), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12115), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_BQUOTE,
  [266537] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12119), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12117), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [266569] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12121), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12123), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_BQUOTE,
  [266601] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12125), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12127), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_BQUOTE,
  [266633] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11931), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11929), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_BQUOTE,
  [266665] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12129), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12131), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_BQUOTE,
  [266697] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12133), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12135), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_BQUOTE,
  [266729] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12139), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12137), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [266761] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2096), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(2094), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [266793] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12143), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12141), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [266825] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12119), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12117), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [266857] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12147), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12145), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [266889] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12151), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12149), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [266921] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12155), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12153), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [266953] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4554), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4552), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [266985] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12157), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12159), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_BQUOTE,
  [267017] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12157), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12159), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_BQUOTE,
  [267049] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4554), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4552), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [267081] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12133), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12135), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [267113] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12163), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12161), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [267145] = 18,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12167), 1,
      anon_sym_LPAREN,
    ACTIONS(12169), 1,
      aux_sym__c_word_token1,
    ACTIONS(12171), 1,
      anon_sym_DOLLAR,
    ACTIONS(12173), 1,
      anon_sym_DQUOTE,
    ACTIONS(12175), 1,
      aux_sym_number_token1,
    ACTIONS(12177), 1,
      aux_sym_number_token2,
    ACTIONS(12179), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(12181), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(12183), 1,
      anon_sym_BQUOTE,
    ACTIONS(12185), 1,
      anon_sym_DOLLAR_BQUOTE,
    STATE(3442), 1,
      sym__c_unary_expression,
    STATE(3443), 1,
      sym__c_binary_expression,
    STATE(3444), 1,
      sym__c_postfix_expression,
    STATE(6605), 1,
      sym__c_expression,
    STATE(6762), 1,
      sym__c_variable_assignment,
    ACTIONS(12165), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    STATE(3441), 7,
      sym__c_expression_not_assignment,
      sym__c_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [267207] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12187), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12189), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_BQUOTE,
  [267239] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4372), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4370), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [267271] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12121), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12123), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [267303] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12191), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12193), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_BQUOTE,
  [267335] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11651), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11653), 1,
      sym__concat,
    STATE(5033), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4554), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4552), 19,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [267373] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11651), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11653), 1,
      sym__concat,
    STATE(5031), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4566), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4564), 19,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [267411] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12195), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12197), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_BQUOTE,
  [267443] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12113), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12115), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [267475] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12199), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12201), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_BQUOTE,
  [267507] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11651), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11653), 1,
      sym__concat,
    STATE(5033), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(2074), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(2072), 19,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [267545] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11842), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11844), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [267577] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12155), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12153), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_BQUOTE,
  [267609] = 5,
    ACTIONS(3), 1,
      sym_comment,
    STATE(4971), 1,
      aux_sym_pipeline_repeat1,
    ACTIONS(12068), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12203), 2,
      anon_sym_PIPE,
      anon_sym_PIPE_AMP,
    ACTIONS(12064), 19,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [267645] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11739), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11741), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_BQUOTE,
  [267677] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11651), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11653), 1,
      sym__concat,
    STATE(5031), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5697), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5695), 19,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [267715] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2074), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(2072), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_BQUOTE,
  [267747] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12107), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12105), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [267779] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12151), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12149), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_BQUOTE,
  [267811] = 10,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4411), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11562), 1,
      sym_file_descriptor,
    ACTIONS(4380), 2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
    ACTIONS(4382), 2,
      anon_sym_LT_LT,
      anon_sym_LT_LT_DASH,
    ACTIONS(4409), 2,
      anon_sym_SEMI,
      anon_sym_AMP,
    ACTIONS(11335), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(4257), 3,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
    STATE(4455), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(11333), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
  [267857] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12205), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12207), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_BQUOTE,
  [267889] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12103), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12101), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_BQUOTE,
  [267921] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1263), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1261), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      sym__special_character,
  [267953] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2074), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(2072), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [267985] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12209), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12211), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_BQUOTE,
  [268017] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12215), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12213), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [268049] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12003), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12001), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [268081] = 18,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12167), 1,
      anon_sym_LPAREN,
    ACTIONS(12169), 1,
      aux_sym__c_word_token1,
    ACTIONS(12171), 1,
      anon_sym_DOLLAR,
    ACTIONS(12173), 1,
      anon_sym_DQUOTE,
    ACTIONS(12175), 1,
      aux_sym_number_token1,
    ACTIONS(12177), 1,
      aux_sym_number_token2,
    ACTIONS(12179), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(12181), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(12183), 1,
      anon_sym_BQUOTE,
    ACTIONS(12185), 1,
      anon_sym_DOLLAR_BQUOTE,
    STATE(3442), 1,
      sym__c_unary_expression,
    STATE(3443), 1,
      sym__c_binary_expression,
    STATE(3444), 1,
      sym__c_postfix_expression,
    STATE(6742), 1,
      sym__c_expression,
    STATE(6762), 1,
      sym__c_variable_assignment,
    ACTIONS(12165), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    STATE(3441), 7,
      sym__c_expression_not_assignment,
      sym__c_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [268143] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11651), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11653), 1,
      sym__concat,
    STATE(5031), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4469), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4467), 19,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [268181] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12199), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12201), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_BQUOTE,
  [268213] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11651), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11653), 1,
      sym__concat,
    STATE(5033), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(2096), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(2094), 19,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [268251] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12219), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12217), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [268283] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12199), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12201), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [268315] = 18,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12167), 1,
      anon_sym_LPAREN,
    ACTIONS(12169), 1,
      aux_sym__c_word_token1,
    ACTIONS(12171), 1,
      anon_sym_DOLLAR,
    ACTIONS(12173), 1,
      anon_sym_DQUOTE,
    ACTIONS(12175), 1,
      aux_sym_number_token1,
    ACTIONS(12177), 1,
      aux_sym_number_token2,
    ACTIONS(12179), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(12181), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(12183), 1,
      anon_sym_BQUOTE,
    ACTIONS(12185), 1,
      anon_sym_DOLLAR_BQUOTE,
    STATE(3442), 1,
      sym__c_unary_expression,
    STATE(3443), 1,
      sym__c_binary_expression,
    STATE(3444), 1,
      sym__c_postfix_expression,
    STATE(6728), 1,
      sym__c_expression,
    STATE(6762), 1,
      sym__c_variable_assignment,
    ACTIONS(12165), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    STATE(3441), 7,
      sym__c_expression_not_assignment,
      sym__c_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [268377] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5375), 3,
      sym_file_descriptor,
      sym_variable_name,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5373), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [268409] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12199), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12201), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [268441] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12215), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12213), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [268473] = 10,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5142), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12221), 1,
      sym_file_descriptor,
    ACTIONS(4253), 2,
      anon_sym_PIPE,
      anon_sym_PIPE_AMP,
    ACTIONS(4692), 2,
      anon_sym_LT_LT,
      anon_sym_LT_LT_DASH,
    ACTIONS(5099), 2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
    ACTIONS(11556), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(5140), 3,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
    STATE(4986), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(11554), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
  [268519] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12225), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12223), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [268551] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4372), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4370), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [268583] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12227), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12229), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_BQUOTE,
  [268615] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12209), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12211), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [268647] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12187), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12189), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [268679] = 18,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(11256), 1,
      anon_sym_DOLLAR,
    ACTIONS(11262), 1,
      aux_sym_number_token2,
    ACTIONS(12233), 1,
      anon_sym_LPAREN,
    ACTIONS(12235), 1,
      aux_sym__c_word_token1,
    ACTIONS(12237), 1,
      anon_sym_DQUOTE,
    ACTIONS(12239), 1,
      aux_sym_number_token1,
    ACTIONS(12241), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(12243), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(12245), 1,
      anon_sym_BQUOTE,
    ACTIONS(12247), 1,
      anon_sym_DOLLAR_BQUOTE,
    STATE(3113), 1,
      sym__c_postfix_expression,
    STATE(3116), 1,
      sym__c_binary_expression,
    STATE(3118), 1,
      sym__c_unary_expression,
    STATE(6493), 1,
      sym__c_variable_assignment,
    STATE(6531), 1,
      sym__c_expression,
    ACTIONS(12231), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    STATE(3119), 7,
      sym__c_expression_not_assignment,
      sym__c_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [268741] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12225), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12223), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_BQUOTE,
  [268773] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12225), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12223), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [268805] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12251), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12249), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [268837] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12227), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12229), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_BQUOTE,
  [268869] = 10,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4384), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11562), 1,
      sym_file_descriptor,
    ACTIONS(4378), 2,
      anon_sym_SEMI,
      anon_sym_AMP,
    ACTIONS(4380), 2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
    ACTIONS(4382), 2,
      anon_sym_LT_LT,
      anon_sym_LT_LT_DASH,
    ACTIONS(11335), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(2719), 3,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
    STATE(4455), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(11333), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
  [268915] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12253), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12255), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_BQUOTE,
  [268947] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11651), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11653), 1,
      sym__concat,
    STATE(5031), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5723), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5721), 19,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [268985] = 5,
    ACTIONS(3), 1,
      sym_comment,
    STATE(5602), 1,
      aux_sym_pipeline_repeat1,
    ACTIONS(11745), 2,
      anon_sym_PIPE,
      anon_sym_PIPE_AMP,
    ACTIONS(11748), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11743), 19,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [269021] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5932), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5930), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [269053] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12259), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12257), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [269085] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5932), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5930), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [269117] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11782), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11780), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [269149] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12259), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12257), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [269181] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12225), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12223), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_BQUOTE,
  [269213] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12263), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12261), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [269245] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11733), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11735), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [269277] = 10,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5387), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12221), 1,
      sym_file_descriptor,
    ACTIONS(4253), 2,
      anon_sym_PIPE,
      anon_sym_PIPE_AMP,
    ACTIONS(4692), 2,
      anon_sym_LT_LT,
      anon_sym_LT_LT_DASH,
    ACTIONS(5099), 2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
    ACTIONS(11556), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(5385), 3,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
    STATE(4986), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(11554), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
  [269323] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12265), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12267), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_BQUOTE,
  [269355] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11739), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11741), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [269387] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11727), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11725), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_BQUOTE,
  [269419] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11727), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11725), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [269451] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12157), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12159), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [269483] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12157), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12159), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [269515] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11729), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11731), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_BQUOTE,
  [269547] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11729), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11731), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [269579] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12269), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12271), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_BQUOTE,
  [269611] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12219), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12217), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_BQUOTE,
  [269643] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11956), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11958), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [269675] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12191), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12193), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [269707] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2096), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(2094), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [269739] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11790), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11792), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_BQUOTE,
  [269771] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12195), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12197), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [269803] = 18,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12167), 1,
      anon_sym_LPAREN,
    ACTIONS(12169), 1,
      aux_sym__c_word_token1,
    ACTIONS(12171), 1,
      anon_sym_DOLLAR,
    ACTIONS(12173), 1,
      anon_sym_DQUOTE,
    ACTIONS(12175), 1,
      aux_sym_number_token1,
    ACTIONS(12177), 1,
      aux_sym_number_token2,
    ACTIONS(12179), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(12181), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(12183), 1,
      anon_sym_BQUOTE,
    ACTIONS(12185), 1,
      anon_sym_DOLLAR_BQUOTE,
    STATE(3442), 1,
      sym__c_unary_expression,
    STATE(3443), 1,
      sym__c_binary_expression,
    STATE(3444), 1,
      sym__c_postfix_expression,
    STATE(6671), 1,
      sym__c_expression,
    STATE(6762), 1,
      sym__c_variable_assignment,
    ACTIONS(12165), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    STATE(3441), 7,
      sym__c_expression_not_assignment,
      sym__c_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [269865] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12163), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12161), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_BQUOTE,
  [269897] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12269), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12271), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [269929] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12119), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12117), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_BQUOTE,
  [269961] = 5,
    ACTIONS(3), 1,
      sym_comment,
    STATE(4971), 1,
      aux_sym_pipeline_repeat1,
    ACTIONS(11748), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12273), 2,
      anon_sym_PIPE,
      anon_sym_PIPE_AMP,
    ACTIONS(11743), 19,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [269997] = 18,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(11808), 1,
      anon_sym_LPAREN,
    ACTIONS(11810), 1,
      aux_sym__c_word_token1,
    ACTIONS(11812), 1,
      anon_sym_DOLLAR,
    ACTIONS(11814), 1,
      anon_sym_DQUOTE,
    ACTIONS(11816), 1,
      aux_sym_number_token1,
    ACTIONS(11818), 1,
      aux_sym_number_token2,
    ACTIONS(11820), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(11822), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(11824), 1,
      anon_sym_BQUOTE,
    ACTIONS(11826), 1,
      anon_sym_DOLLAR_BQUOTE,
    STATE(3285), 1,
      sym__c_postfix_expression,
    STATE(3286), 1,
      sym__c_binary_expression,
    STATE(3290), 1,
      sym__c_unary_expression,
    STATE(6741), 1,
      sym__c_expression,
    STATE(6768), 1,
      sym__c_variable_assignment,
    ACTIONS(11806), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    STATE(3292), 7,
      sym__c_expression_not_assignment,
      sym__c_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [270059] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12119), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12117), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_BQUOTE,
  [270091] = 18,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(11808), 1,
      anon_sym_LPAREN,
    ACTIONS(11810), 1,
      aux_sym__c_word_token1,
    ACTIONS(11812), 1,
      anon_sym_DOLLAR,
    ACTIONS(11814), 1,
      anon_sym_DQUOTE,
    ACTIONS(11816), 1,
      aux_sym_number_token1,
    ACTIONS(11818), 1,
      aux_sym_number_token2,
    ACTIONS(11820), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(11822), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(11824), 1,
      anon_sym_BQUOTE,
    ACTIONS(11826), 1,
      anon_sym_DOLLAR_BQUOTE,
    STATE(3285), 1,
      sym__c_postfix_expression,
    STATE(3286), 1,
      sym__c_binary_expression,
    STATE(3290), 1,
      sym__c_unary_expression,
    STATE(6738), 1,
      sym__c_expression,
    STATE(6768), 1,
      sym__c_variable_assignment,
    ACTIONS(11806), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    STATE(3292), 7,
      sym__c_expression_not_assignment,
      sym__c_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [270153] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12276), 1,
      sym__special_character,
    STATE(5021), 1,
      aux_sym__literal_repeat1,
    ACTIONS(5395), 3,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5393), 19,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [270189] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11739), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11741), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [270221] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12265), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12267), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [270253] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11790), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11792), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [270285] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12107), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12105), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_BQUOTE,
  [270317] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12129), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12131), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [270349] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12125), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12127), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [270381] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12036), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12034), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [270413] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12107), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12105), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_BQUOTE,
  [270445] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11830), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11828), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [270477] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12109), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12111), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [270509] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11455), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12221), 1,
      sym_file_descriptor,
    ACTIONS(4692), 2,
      anon_sym_LT_LT,
      anon_sym_LT_LT_DASH,
    ACTIONS(11556), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    STATE(4998), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(11453), 7,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
    ACTIONS(11554), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
  [270551] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12139), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12137), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_BQUOTE,
  [270583] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12143), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12141), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_BQUOTE,
  [270615] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12147), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12145), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_BQUOTE,
  [270647] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5772), 3,
      sym_file_descriptor,
      sym_variable_name,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5770), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [270679] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12205), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12207), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [270711] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11651), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11653), 1,
      sym__concat,
    STATE(5033), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4372), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4370), 19,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [270749] = 10,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5101), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12221), 1,
      sym_file_descriptor,
    ACTIONS(4253), 2,
      anon_sym_PIPE,
      anon_sym_PIPE_AMP,
    ACTIONS(4692), 2,
      anon_sym_LT_LT,
      anon_sym_LT_LT_DASH,
    ACTIONS(5099), 2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
    ACTIONS(11556), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(5097), 3,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
    STATE(4986), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(11554), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
  [270795] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11956), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11958), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_BQUOTE,
  [270827] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11535), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11558), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(12278), 1,
      sym_file_descriptor,
    ACTIONS(11556), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    STATE(4877), 2,
      sym_file_redirect,
      sym_herestring_redirect,
    ACTIONS(11554), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
    ACTIONS(11533), 9,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
  [270869] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4253), 2,
      anon_sym_PIPE,
      anon_sym_PIPE_AMP,
    ACTIONS(4272), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    STATE(4986), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(4270), 17,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [270905] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4272), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    STATE(4986), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(4270), 19,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [270939] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11431), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12286), 1,
      sym_file_descriptor,
    ACTIONS(11920), 2,
      anon_sym_LT_LT,
      anon_sym_LT_LT_DASH,
    ACTIONS(12283), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    STATE(4998), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(11420), 7,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
    ACTIONS(12280), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
  [270981] = 18,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(11256), 1,
      anon_sym_DOLLAR,
    ACTIONS(11262), 1,
      aux_sym_number_token2,
    ACTIONS(12233), 1,
      anon_sym_LPAREN,
    ACTIONS(12235), 1,
      aux_sym__c_word_token1,
    ACTIONS(12237), 1,
      anon_sym_DQUOTE,
    ACTIONS(12239), 1,
      aux_sym_number_token1,
    ACTIONS(12241), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(12243), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(12245), 1,
      anon_sym_BQUOTE,
    ACTIONS(12247), 1,
      anon_sym_DOLLAR_BQUOTE,
    STATE(3113), 1,
      sym__c_postfix_expression,
    STATE(3116), 1,
      sym__c_binary_expression,
    STATE(3118), 1,
      sym__c_unary_expression,
    STATE(6481), 1,
      sym__c_expression,
    STATE(6493), 1,
      sym__c_variable_assignment,
    ACTIONS(12231), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    STATE(3119), 7,
      sym__c_expression_not_assignment,
      sym__c_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [271043] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11414), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11558), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(12278), 1,
      sym_file_descriptor,
    ACTIONS(11556), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    STATE(4868), 2,
      sym_file_redirect,
      sym_herestring_redirect,
    ACTIONS(11554), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
    ACTIONS(11412), 9,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
  [271085] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11506), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11558), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(12278), 1,
      sym_file_descriptor,
    ACTIONS(11556), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    STATE(5017), 2,
      sym_file_redirect,
      sym_herestring_redirect,
    ACTIONS(11554), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
    ACTIONS(11504), 9,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
  [271127] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12259), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12257), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_BQUOTE,
  [271159] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12227), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12229), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [271191] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2074), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(2072), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [271223] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12215), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12213), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_BQUOTE,
  [271255] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12227), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12229), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [271287] = 10,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5383), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12221), 1,
      sym_file_descriptor,
    ACTIONS(4253), 2,
      anon_sym_PIPE,
      anon_sym_PIPE_AMP,
    ACTIONS(4692), 2,
      anon_sym_LT_LT,
      anon_sym_LT_LT_DASH,
    ACTIONS(5099), 2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
    ACTIONS(11556), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(5381), 3,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
    STATE(4986), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(11554), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
  [271333] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11931), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11929), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [271365] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12215), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12213), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_BQUOTE,
  [271397] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12036), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12034), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_BQUOTE,
  [271429] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12259), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12257), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_BQUOTE,
  [271461] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11956), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11958), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [271493] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12253), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12255), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [271525] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5932), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5930), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_BQUOTE,
  [271557] = 18,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12167), 1,
      anon_sym_LPAREN,
    ACTIONS(12169), 1,
      aux_sym__c_word_token1,
    ACTIONS(12171), 1,
      anon_sym_DOLLAR,
    ACTIONS(12173), 1,
      anon_sym_DQUOTE,
    ACTIONS(12175), 1,
      aux_sym_number_token1,
    ACTIONS(12177), 1,
      aux_sym_number_token2,
    ACTIONS(12179), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(12181), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(12183), 1,
      anon_sym_BQUOTE,
    ACTIONS(12185), 1,
      anon_sym_DOLLAR_BQUOTE,
    STATE(3442), 1,
      sym__c_unary_expression,
    STATE(3443), 1,
      sym__c_binary_expression,
    STATE(3444), 1,
      sym__c_postfix_expression,
    STATE(6551), 1,
      sym__c_expression,
    STATE(6762), 1,
      sym__c_variable_assignment,
    ACTIONS(12165), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    STATE(3441), 7,
      sym__c_expression_not_assignment,
      sym__c_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [271619] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12036), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12034), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [271651] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12263), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12261), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_BQUOTE,
  [271683] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11727), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11725), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [271715] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11729), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11731), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [271747] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11790), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11792), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [271779] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12289), 1,
      sym__special_character,
    STATE(5021), 1,
      aux_sym__literal_repeat1,
    ACTIONS(1364), 3,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1362), 19,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [271815] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12251), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12249), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_BQUOTE,
  [271847] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2096), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(2094), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_BQUOTE,
  [271879] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11750), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11752), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [271911] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11830), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11828), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [271943] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5932), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5930), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_BQUOTE,
  [271975] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12292), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12294), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_BQUOTE,
  [272007] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12298), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12296), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [272039] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4372), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4370), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_BQUOTE,
  [272071] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12298), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12296), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_BQUOTE,
  [272103] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11651), 1,
      aux_sym_concatenation_token1,
    ACTIONS(12300), 1,
      sym__concat,
    STATE(4268), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1267), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1265), 19,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [272141] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12292), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12294), 22,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_esac,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [272173] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11651), 1,
      aux_sym_concatenation_token1,
    ACTIONS(12302), 1,
      sym__concat,
    STATE(4268), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1288), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1286), 19,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [272211] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11568), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11583), 1,
      sym__concat,
    STATE(5057), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1263), 3,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1261), 17,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [272248] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12147), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12145), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [272279] = 5,
    ACTIONS(3), 1,
      sym_comment,
    STATE(5082), 1,
      aux_sym_pipeline_repeat1,
    ACTIONS(12068), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12304), 2,
      anon_sym_PIPE,
      anon_sym_PIPE_AMP,
    ACTIONS(12064), 18,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [272314] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12155), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12153), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [272345] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(5038), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(12306), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(1251), 6,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1253), 14,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_RPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [272380] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12151), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12149), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [272411] = 10,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2719), 1,
      anon_sym_BQUOTE,
    ACTIONS(5357), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11856), 1,
      sym_file_descriptor,
    ACTIONS(4692), 2,
      anon_sym_LT_LT,
      anon_sym_LT_LT_DASH,
    ACTIONS(5355), 2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
    ACTIONS(11487), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(5353), 3,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
    STATE(4697), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(11485), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
  [272456] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(5065), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(12309), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(5393), 6,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(5395), 14,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_RPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [272491] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11739), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11741), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [272522] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12133), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12135), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [272553] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1263), 3,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1261), 20,
      anon_sym_LPAREN_LPAREN,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [272584] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(5059), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(12309), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(5389), 6,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(5391), 14,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_RPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [272619] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5932), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5930), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [272650] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12143), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12141), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [272681] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11568), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11583), 1,
      sym__concat,
    STATE(5057), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(12313), 3,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12311), 17,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [272718] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11931), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11929), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [272749] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11956), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11958), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [272780] = 10,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2719), 1,
      anon_sym_RPAREN,
    ACTIONS(4384), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11993), 1,
      sym_file_descriptor,
    ACTIONS(4692), 2,
      anon_sym_LT_LT,
      anon_sym_LT_LT_DASH,
    ACTIONS(4809), 2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
    ACTIONS(11542), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(4378), 3,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
    STATE(4746), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(11540), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
  [272825] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12259), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12257), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [272856] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11790), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11792), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [272887] = 10,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4257), 1,
      anon_sym_RPAREN,
    ACTIONS(4411), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11993), 1,
      sym_file_descriptor,
    ACTIONS(4692), 2,
      anon_sym_LT_LT,
      anon_sym_LT_LT_DASH,
    ACTIONS(4809), 2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
    ACTIONS(11542), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(4409), 3,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
    STATE(4746), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(11540), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
  [272932] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11568), 1,
      aux_sym_concatenation_token1,
    ACTIONS(12315), 1,
      sym__concat,
    STATE(4520), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1288), 3,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1286), 17,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [272969] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12317), 1,
      anon_sym_LT_LT_LT,
    STATE(4912), 1,
      sym_herestring_redirect,
    ACTIONS(4253), 2,
      anon_sym_PIPE,
      anon_sym_PIPE_AMP,
    ACTIONS(4272), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    STATE(5366), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(4270), 14,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [273008] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11568), 1,
      aux_sym_concatenation_token1,
    ACTIONS(12319), 1,
      sym__concat,
    STATE(4520), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1267), 3,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1265), 17,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [273045] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12317), 1,
      anon_sym_LT_LT_LT,
    STATE(4912), 1,
      sym_herestring_redirect,
    ACTIONS(4348), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    STATE(5366), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(4253), 16,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [273082] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12309), 1,
      aux_sym_concatenation_token1,
    ACTIONS(12321), 1,
      sym__concat,
    STATE(5038), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1286), 6,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1288), 14,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_RPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [273119] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12139), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12137), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [273150] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12215), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12213), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [273181] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11568), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11583), 1,
      sym__concat,
    STATE(5055), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(12325), 3,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12323), 17,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [273218] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12215), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12213), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [273249] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(5175), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(12327), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(1261), 8,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      sym__special_character,
    ACTIONS(1263), 12,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [273284] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12309), 1,
      aux_sym_concatenation_token1,
    ACTIONS(12329), 1,
      sym__concat,
    STATE(5038), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1265), 6,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1267), 14,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_RPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [273321] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12209), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12211), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [273352] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12107), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12105), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [273383] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12107), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12105), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [273414] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12103), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12101), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [273445] = 10,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4257), 1,
      anon_sym_BQUOTE,
    ACTIONS(5379), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11856), 1,
      sym_file_descriptor,
    ACTIONS(4692), 2,
      anon_sym_LT_LT,
      anon_sym_LT_LT_DASH,
    ACTIONS(5355), 2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
    ACTIONS(11487), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(5377), 3,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
    STATE(4697), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(11485), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
  [273490] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5932), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5930), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [273521] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12119), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12117), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [273552] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12121), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12123), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [273583] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12036), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12034), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [273614] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12113), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12115), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [273645] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12129), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12131), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [273676] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12119), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12117), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [273707] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12125), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12127), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [273738] = 10,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4686), 1,
      ts_builtin_sym_end,
    ACTIONS(4694), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11834), 1,
      sym_file_descriptor,
    ACTIONS(4690), 2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
    ACTIONS(4692), 2,
      anon_sym_LT_LT,
      anon_sym_LT_LT_DASH,
    ACTIONS(11481), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(4688), 3,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
    STATE(4650), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(11479), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
  [273783] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12163), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12161), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [273814] = 5,
    ACTIONS(3), 1,
      sym_comment,
    STATE(5081), 1,
      aux_sym_pipeline_repeat1,
    ACTIONS(11748), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12331), 2,
      anon_sym_PIPE,
      anon_sym_PIPE_AMP,
    ACTIONS(11743), 18,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_BQUOTE,
  [273849] = 5,
    ACTIONS(3), 1,
      sym_comment,
    STATE(5082), 1,
      aux_sym_pipeline_repeat1,
    ACTIONS(11748), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12334), 2,
      anon_sym_PIPE,
      anon_sym_PIPE_AMP,
    ACTIONS(11743), 18,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RPAREN,
      anon_sym_SEMI_SEMI,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [273884] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12219), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12217), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [273915] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12109), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12111), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [273946] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12227), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12229), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [273977] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4554), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4552), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [274008] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12199), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12201), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [274039] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12227), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12229), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [274070] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12253), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12255), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [274101] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12191), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12193), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [274132] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11727), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11725), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [274163] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11729), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11731), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [274194] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12269), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12271), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [274225] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11830), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11828), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [274256] = 5,
    ACTIONS(3), 1,
      sym_comment,
    STATE(5115), 1,
      aux_sym_pipeline_repeat1,
    ACTIONS(12337), 2,
      anon_sym_PIPE,
      anon_sym_PIPE_AMP,
    ACTIONS(12068), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12064), 17,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [274291] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2096), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(2094), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [274322] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(5065), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(12309), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(1261), 6,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1263), 14,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_RPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [274357] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12199), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12201), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [274388] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12225), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12223), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [274419] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12225), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12223), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [274450] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12187), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12189), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [274481] = 5,
    ACTIONS(3), 1,
      sym_comment,
    STATE(5081), 1,
      aux_sym_pipeline_repeat1,
    ACTIONS(12068), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12339), 2,
      anon_sym_PIPE,
      anon_sym_PIPE_AMP,
    ACTIONS(12064), 18,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_BQUOTE,
  [274516] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12251), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12249), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [274547] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12195), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12197), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [274578] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12259), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12257), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [274609] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4372), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4370), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [274640] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12292), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12294), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [274671] = 10,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2721), 1,
      ts_builtin_sym_end,
    ACTIONS(5051), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11834), 1,
      sym_file_descriptor,
    ACTIONS(4690), 2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
    ACTIONS(4692), 2,
      anon_sym_LT_LT,
      anon_sym_LT_LT_DASH,
    ACTIONS(11481), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(5049), 3,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
    STATE(4650), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(11479), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
  [274716] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11512), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11514), 1,
      sym__concat,
    STATE(5207), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1263), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1261), 18,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      sym__special_character,
  [274753] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11500), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11502), 1,
      sym__concat,
    STATE(5209), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1263), 3,
      sym_file_descriptor,
      sym_variable_name,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1261), 17,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      sym__special_character,
  [274790] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12265), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12267), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [274821] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12157), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12159), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [274852] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12263), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12261), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [274883] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2074), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(2072), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [274914] = 5,
    ACTIONS(3), 1,
      sym_comment,
    STATE(5115), 1,
      aux_sym_pipeline_repeat1,
    ACTIONS(12341), 2,
      anon_sym_PIPE,
      anon_sym_PIPE_AMP,
    ACTIONS(11748), 3,
      sym_file_descriptor,
      ts_builtin_sym_end,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(11743), 17,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [274949] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12205), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12207), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [274980] = 11,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5944), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12317), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(12348), 1,
      sym_file_descriptor,
    STATE(4912), 1,
      sym_herestring_redirect,
    ACTIONS(4253), 2,
      anon_sym_PIPE,
      anon_sym_PIPE_AMP,
    ACTIONS(4692), 2,
      anon_sym_LT_LT,
      anon_sym_LT_LT_DASH,
    ACTIONS(5940), 2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
    ACTIONS(12346), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    STATE(5366), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(12344), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
  [275027] = 11,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5942), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12317), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(12348), 1,
      sym_file_descriptor,
    STATE(4912), 1,
      sym_herestring_redirect,
    ACTIONS(4253), 2,
      anon_sym_PIPE,
      anon_sym_PIPE_AMP,
    ACTIONS(4692), 2,
      anon_sym_LT_LT,
      anon_sym_LT_LT_DASH,
    ACTIONS(5940), 2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
    ACTIONS(12346), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    STATE(5366), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(12344), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
  [275074] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12298), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12296), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [275105] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12157), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12159), 21,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [275136] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12354), 1,
      anon_sym_esac,
    ACTIONS(12350), 6,
      anon_sym_LPAREN,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(12352), 15,
      sym_test_operator,
      sym_extglob_pattern,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [275168] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5101), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12221), 1,
      sym_file_descriptor,
    ACTIONS(4692), 2,
      anon_sym_LT_LT,
      anon_sym_LT_LT_DASH,
    ACTIONS(5099), 2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
    ACTIONS(11556), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(5097), 3,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
    STATE(4986), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(11554), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
  [275210] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1292), 6,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1294), 16,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_RPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [275240] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(5175), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(12327), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(4467), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(4469), 12,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [275274] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12360), 1,
      anon_sym_esac,
    ACTIONS(12356), 6,
      anon_sym_LPAREN,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(12358), 15,
      sym_test_operator,
      sym_extglob_pattern,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [275306] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12366), 1,
      anon_sym_esac,
    ACTIONS(12362), 6,
      anon_sym_LPAREN,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(12364), 15,
      sym_test_operator,
      sym_extglob_pattern,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [275338] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(5175), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(12327), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(4564), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(4566), 12,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [275372] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11512), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11514), 1,
      sym__concat,
    STATE(5208), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4554), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4552), 17,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [275408] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11512), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11514), 1,
      sym__concat,
    STATE(5207), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4566), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4564), 17,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [275444] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11512), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11514), 1,
      sym__concat,
    STATE(5208), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4372), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4370), 17,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [275480] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11512), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11514), 1,
      sym__concat,
    STATE(5207), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4469), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4467), 17,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [275516] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12372), 1,
      anon_sym_esac,
    ACTIONS(12368), 6,
      anon_sym_LPAREN,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(12370), 15,
      sym_test_operator,
      sym_extglob_pattern,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [275548] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11512), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11514), 1,
      sym__concat,
    STATE(5208), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(2074), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(2072), 17,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [275584] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11512), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11514), 1,
      sym__concat,
    STATE(5207), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5697), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5695), 17,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [275620] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11512), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11514), 1,
      sym__concat,
    STATE(5208), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(2096), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(2094), 17,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [275656] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11512), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11514), 1,
      sym__concat,
    STATE(5207), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5723), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5721), 17,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [275692] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11500), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11502), 1,
      sym__concat,
    STATE(5210), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5375), 3,
      sym_file_descriptor,
      sym_variable_name,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5373), 16,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [275728] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11500), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11502), 1,
      sym__concat,
    STATE(5209), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5069), 3,
      sym_file_descriptor,
      sym_variable_name,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5067), 16,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [275764] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12366), 1,
      anon_sym_esac,
    ACTIONS(12362), 6,
      anon_sym_LPAREN,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(12364), 15,
      sym_test_operator,
      sym_extglob_pattern,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [275796] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(5180), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(12327), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(4552), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(4554), 12,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [275830] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12354), 1,
      anon_sym_esac,
    ACTIONS(12350), 6,
      anon_sym_LPAREN,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(12352), 15,
      sym_test_operator,
      sym_extglob_pattern,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [275862] = 16,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(11808), 1,
      anon_sym_LPAREN,
    ACTIONS(11812), 1,
      anon_sym_DOLLAR,
    ACTIONS(11814), 1,
      anon_sym_DQUOTE,
    ACTIONS(11816), 1,
      aux_sym_number_token1,
    ACTIONS(11818), 1,
      aux_sym_number_token2,
    ACTIONS(11820), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(11822), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(11824), 1,
      anon_sym_BQUOTE,
    ACTIONS(11826), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(12374), 1,
      aux_sym__c_word_token1,
    STATE(3285), 1,
      sym__c_postfix_expression,
    STATE(3286), 1,
      sym__c_binary_expression,
    STATE(3290), 1,
      sym__c_unary_expression,
    ACTIONS(11806), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    STATE(3276), 7,
      sym__c_expression_not_assignment,
      sym__c_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [275918] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12380), 1,
      anon_sym_esac,
    ACTIONS(12376), 6,
      anon_sym_LPAREN,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(12378), 15,
      sym_test_operator,
      sym_extglob_pattern,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [275950] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12380), 1,
      anon_sym_esac,
    ACTIONS(12376), 6,
      anon_sym_LPAREN,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(12378), 15,
      sym_test_operator,
      sym_extglob_pattern,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [275982] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12372), 1,
      anon_sym_esac,
    ACTIONS(12368), 6,
      anon_sym_LPAREN,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(12370), 15,
      sym_test_operator,
      sym_extglob_pattern,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [276014] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12386), 1,
      anon_sym_esac,
    ACTIONS(12382), 6,
      anon_sym_LPAREN,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(12384), 15,
      sym_test_operator,
      sym_extglob_pattern,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [276046] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12388), 1,
      sym_variable_name,
    STATE(6746), 1,
      sym_subscript,
    ACTIONS(11290), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    STATE(5147), 2,
      sym_variable_assignment,
      aux_sym_variable_assignments_repeat1,
    ACTIONS(11288), 16,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [276082] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12391), 1,
      sym_variable_name,
    STATE(6746), 1,
      sym_subscript,
    ACTIONS(11329), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    STATE(5147), 2,
      sym_variable_assignment,
      aux_sym_variable_assignments_repeat1,
    ACTIONS(11327), 16,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [276118] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1316), 6,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1318), 16,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_RPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [276148] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1320), 6,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1322), 16,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_RPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [276178] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1328), 6,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1330), 16,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_RPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [276208] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1340), 6,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1342), 16,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_RPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [276238] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1352), 6,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1354), 16,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_RPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [276268] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1356), 6,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1358), 16,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_RPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [276298] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1332), 6,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1334), 16,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_RPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [276328] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1336), 6,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1338), 16,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_RPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [276358] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1312), 6,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1314), 16,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_RPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [276388] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1344), 6,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1346), 16,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_RPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [276418] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1308), 6,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1310), 16,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_RPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [276448] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1304), 6,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1306), 16,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_RPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [276478] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1308), 6,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1310), 16,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_RPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [276508] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12397), 1,
      anon_sym_esac,
    ACTIONS(12393), 6,
      anon_sym_LPAREN,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(12395), 15,
      sym_test_operator,
      sym_extglob_pattern,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [276540] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5387), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12221), 1,
      sym_file_descriptor,
    ACTIONS(4692), 2,
      anon_sym_LT_LT,
      anon_sym_LT_LT_DASH,
    ACTIONS(5099), 2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
    ACTIONS(11556), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(5385), 3,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
    STATE(4986), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(11554), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
  [276582] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12403), 1,
      anon_sym_esac,
    ACTIONS(12399), 6,
      anon_sym_LPAREN,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(12401), 15,
      sym_test_operator,
      sym_extglob_pattern,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [276614] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1300), 6,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1302), 16,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_RPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [276644] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12409), 1,
      anon_sym_esac,
    ACTIONS(12405), 6,
      anon_sym_LPAREN,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(12407), 15,
      sym_test_operator,
      sym_extglob_pattern,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [276676] = 9,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12417), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(12419), 1,
      sym_file_descriptor,
    ACTIONS(11357), 2,
      anon_sym_PIPE,
      anon_sym_LT_LT,
    ACTIONS(12415), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(12413), 3,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
    STATE(5205), 3,
      sym_file_redirect,
      sym_herestring_redirect,
      aux_sym_redirected_statement_repeat2,
    ACTIONS(11363), 5,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_LT_LT_DASH,
    ACTIONS(12411), 5,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
  [276718] = 16,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12167), 1,
      anon_sym_LPAREN,
    ACTIONS(12171), 1,
      anon_sym_DOLLAR,
    ACTIONS(12173), 1,
      anon_sym_DQUOTE,
    ACTIONS(12175), 1,
      aux_sym_number_token1,
    ACTIONS(12177), 1,
      aux_sym_number_token2,
    ACTIONS(12179), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(12181), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(12183), 1,
      anon_sym_BQUOTE,
    ACTIONS(12185), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(12421), 1,
      aux_sym__c_word_token1,
    STATE(3442), 1,
      sym__c_unary_expression,
    STATE(3443), 1,
      sym__c_binary_expression,
    STATE(3444), 1,
      sym__c_postfix_expression,
    ACTIONS(12165), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    STATE(3454), 7,
      sym__c_expression_not_assignment,
      sym__c_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [276774] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12427), 1,
      anon_sym_esac,
    ACTIONS(12423), 6,
      anon_sym_LPAREN,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(12425), 15,
      sym_test_operator,
      sym_extglob_pattern,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [276806] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(5175), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(12327), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(5721), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(5723), 12,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [276840] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1296), 6,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1298), 16,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_RPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [276870] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12433), 1,
      anon_sym_esac,
    ACTIONS(12429), 6,
      anon_sym_LPAREN,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(12431), 15,
      sym_test_operator,
      sym_extglob_pattern,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [276902] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(5180), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(12327), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(4370), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(4372), 12,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [276936] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(5322), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(12435), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(1261), 8,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      sym__special_character,
    ACTIONS(1263), 11,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [276970] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12327), 1,
      aux_sym_concatenation_token1,
    ACTIONS(12437), 1,
      sym__concat,
    STATE(5188), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1265), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1267), 12,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [277006] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11363), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12443), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(12445), 1,
      sym_file_descriptor,
    ACTIONS(12441), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    STATE(5181), 3,
      sym_file_redirect,
      sym_herestring_redirect,
      aux_sym_redirected_statement_repeat2,
    ACTIONS(11357), 6,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT_LT,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
    ACTIONS(12439), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
  [277046] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12451), 1,
      anon_sym_esac,
    ACTIONS(12447), 6,
      anon_sym_LPAREN,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(12449), 15,
      sym_test_operator,
      sym_extglob_pattern,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [277078] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(5180), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(12327), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(2072), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(2074), 12,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [277112] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(5175), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(12327), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(5695), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(5697), 12,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [277146] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12327), 1,
      aux_sym_concatenation_token1,
    ACTIONS(12453), 1,
      sym__concat,
    STATE(5188), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1286), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1288), 12,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [277182] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11315), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12461), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(12464), 1,
      sym_file_descriptor,
    ACTIONS(12458), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    STATE(5181), 3,
      sym_file_redirect,
      sym_herestring_redirect,
      aux_sym_redirected_statement_repeat2,
    ACTIONS(11307), 6,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT_LT,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
    ACTIONS(12455), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
  [277222] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(5180), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(12327), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(2094), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(2096), 12,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [277256] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12471), 1,
      anon_sym_esac,
    ACTIONS(12467), 6,
      anon_sym_LPAREN,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(12469), 15,
      sym_test_operator,
      sym_extglob_pattern,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [277288] = 8,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(4253), 1,
      anon_sym_PIPE,
    ACTIONS(4348), 1,
      anon_sym_PIPE_AMP,
    ACTIONS(12473), 1,
      anon_sym_LT_LT_LT,
    STATE(5679), 1,
      sym_herestring_redirect,
    STATE(5436), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(4270), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(4272), 9,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [277328] = 16,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12167), 1,
      anon_sym_LPAREN,
    ACTIONS(12171), 1,
      anon_sym_DOLLAR,
    ACTIONS(12173), 1,
      anon_sym_DQUOTE,
    ACTIONS(12175), 1,
      aux_sym_number_token1,
    ACTIONS(12177), 1,
      aux_sym_number_token2,
    ACTIONS(12179), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(12181), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(12183), 1,
      anon_sym_BQUOTE,
    ACTIONS(12185), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(12475), 1,
      aux_sym__c_word_token1,
    STATE(3442), 1,
      sym__c_unary_expression,
    STATE(3443), 1,
      sym__c_binary_expression,
    STATE(3444), 1,
      sym__c_postfix_expression,
    ACTIONS(12165), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    STATE(3405), 7,
      sym__c_expression_not_assignment,
      sym__c_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [277384] = 16,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12167), 1,
      anon_sym_LPAREN,
    ACTIONS(12171), 1,
      anon_sym_DOLLAR,
    ACTIONS(12173), 1,
      anon_sym_DQUOTE,
    ACTIONS(12175), 1,
      aux_sym_number_token1,
    ACTIONS(12177), 1,
      aux_sym_number_token2,
    ACTIONS(12179), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(12181), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(12183), 1,
      anon_sym_BQUOTE,
    ACTIONS(12185), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(12477), 1,
      aux_sym__c_word_token1,
    STATE(3442), 1,
      sym__c_unary_expression,
    STATE(3443), 1,
      sym__c_binary_expression,
    STATE(3444), 1,
      sym__c_postfix_expression,
    ACTIONS(12165), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    STATE(3404), 7,
      sym__c_expression_not_assignment,
      sym__c_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [277440] = 16,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12167), 1,
      anon_sym_LPAREN,
    ACTIONS(12171), 1,
      anon_sym_DOLLAR,
    ACTIONS(12173), 1,
      anon_sym_DQUOTE,
    ACTIONS(12175), 1,
      aux_sym_number_token1,
    ACTIONS(12177), 1,
      aux_sym_number_token2,
    ACTIONS(12179), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(12181), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(12183), 1,
      anon_sym_BQUOTE,
    ACTIONS(12185), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(12479), 1,
      aux_sym__c_word_token1,
    STATE(3442), 1,
      sym__c_unary_expression,
    STATE(3443), 1,
      sym__c_binary_expression,
    STATE(3444), 1,
      sym__c_postfix_expression,
    ACTIONS(12165), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    STATE(3403), 7,
      sym__c_expression_not_assignment,
      sym__c_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [277496] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(5188), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(12481), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(1251), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1253), 12,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [277530] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5383), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12221), 1,
      sym_file_descriptor,
    ACTIONS(4692), 2,
      anon_sym_LT_LT,
      anon_sym_LT_LT_DASH,
    ACTIONS(5099), 2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
    ACTIONS(11556), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(5381), 3,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
    STATE(4986), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(11554), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
  [277572] = 16,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12167), 1,
      anon_sym_LPAREN,
    ACTIONS(12171), 1,
      anon_sym_DOLLAR,
    ACTIONS(12173), 1,
      anon_sym_DQUOTE,
    ACTIONS(12175), 1,
      aux_sym_number_token1,
    ACTIONS(12177), 1,
      aux_sym_number_token2,
    ACTIONS(12179), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(12181), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(12183), 1,
      anon_sym_BQUOTE,
    ACTIONS(12185), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(12484), 1,
      aux_sym__c_word_token1,
    STATE(3442), 1,
      sym__c_unary_expression,
    STATE(3443), 1,
      sym__c_binary_expression,
    STATE(3444), 1,
      sym__c_postfix_expression,
    ACTIONS(12165), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    STATE(3402), 7,
      sym__c_expression_not_assignment,
      sym__c_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [277628] = 16,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12167), 1,
      anon_sym_LPAREN,
    ACTIONS(12171), 1,
      anon_sym_DOLLAR,
    ACTIONS(12173), 1,
      anon_sym_DQUOTE,
    ACTIONS(12175), 1,
      aux_sym_number_token1,
    ACTIONS(12177), 1,
      aux_sym_number_token2,
    ACTIONS(12179), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(12181), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(12183), 1,
      anon_sym_BQUOTE,
    ACTIONS(12185), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(12486), 1,
      aux_sym__c_word_token1,
    STATE(3442), 1,
      sym__c_unary_expression,
    STATE(3443), 1,
      sym__c_binary_expression,
    STATE(3444), 1,
      sym__c_postfix_expression,
    ACTIONS(12165), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    STATE(3399), 7,
      sym__c_expression_not_assignment,
      sym__c_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [277684] = 16,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12167), 1,
      anon_sym_LPAREN,
    ACTIONS(12171), 1,
      anon_sym_DOLLAR,
    ACTIONS(12173), 1,
      anon_sym_DQUOTE,
    ACTIONS(12175), 1,
      aux_sym_number_token1,
    ACTIONS(12177), 1,
      aux_sym_number_token2,
    ACTIONS(12179), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(12181), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(12183), 1,
      anon_sym_BQUOTE,
    ACTIONS(12185), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(12488), 1,
      aux_sym__c_word_token1,
    STATE(3442), 1,
      sym__c_unary_expression,
    STATE(3443), 1,
      sym__c_binary_expression,
    STATE(3444), 1,
      sym__c_postfix_expression,
    ACTIONS(12165), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    STATE(3395), 7,
      sym__c_expression_not_assignment,
      sym__c_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [277740] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12494), 1,
      anon_sym_esac,
    ACTIONS(12490), 6,
      anon_sym_LPAREN,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(12492), 15,
      sym_test_operator,
      sym_extglob_pattern,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [277772] = 16,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12167), 1,
      anon_sym_LPAREN,
    ACTIONS(12171), 1,
      anon_sym_DOLLAR,
    ACTIONS(12173), 1,
      anon_sym_DQUOTE,
    ACTIONS(12175), 1,
      aux_sym_number_token1,
    ACTIONS(12177), 1,
      aux_sym_number_token2,
    ACTIONS(12179), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(12181), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(12183), 1,
      anon_sym_BQUOTE,
    ACTIONS(12185), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(12496), 1,
      aux_sym__c_word_token1,
    STATE(3442), 1,
      sym__c_unary_expression,
    STATE(3443), 1,
      sym__c_binary_expression,
    STATE(3444), 1,
      sym__c_postfix_expression,
    ACTIONS(12165), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    STATE(3394), 7,
      sym__c_expression_not_assignment,
      sym__c_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [277828] = 13,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(4253), 1,
      anon_sym_PIPE,
    ACTIONS(4348), 1,
      anon_sym_PIPE_AMP,
    ACTIONS(5994), 1,
      anon_sym_LT_LT,
    ACTIONS(5996), 1,
      anon_sym_LT_LT_DASH,
    ACTIONS(12504), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(12506), 1,
      sym_file_descriptor,
    STATE(5556), 1,
      sym_herestring_redirect,
    ACTIONS(5992), 2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
    ACTIONS(12502), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(12500), 3,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
    STATE(5350), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(12498), 5,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
  [277878] = 16,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12167), 1,
      anon_sym_LPAREN,
    ACTIONS(12171), 1,
      anon_sym_DOLLAR,
    ACTIONS(12173), 1,
      anon_sym_DQUOTE,
    ACTIONS(12175), 1,
      aux_sym_number_token1,
    ACTIONS(12177), 1,
      aux_sym_number_token2,
    ACTIONS(12179), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(12181), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(12183), 1,
      anon_sym_BQUOTE,
    ACTIONS(12185), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(12508), 1,
      aux_sym__c_word_token1,
    STATE(3442), 1,
      sym__c_unary_expression,
    STATE(3443), 1,
      sym__c_binary_expression,
    STATE(3444), 1,
      sym__c_postfix_expression,
    ACTIONS(12165), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    STATE(3392), 7,
      sym__c_expression_not_assignment,
      sym__c_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [277934] = 16,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12167), 1,
      anon_sym_LPAREN,
    ACTIONS(12171), 1,
      anon_sym_DOLLAR,
    ACTIONS(12173), 1,
      anon_sym_DQUOTE,
    ACTIONS(12175), 1,
      aux_sym_number_token1,
    ACTIONS(12177), 1,
      aux_sym_number_token2,
    ACTIONS(12179), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(12181), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(12183), 1,
      anon_sym_BQUOTE,
    ACTIONS(12185), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(12510), 1,
      aux_sym__c_word_token1,
    STATE(3442), 1,
      sym__c_unary_expression,
    STATE(3443), 1,
      sym__c_binary_expression,
    STATE(3444), 1,
      sym__c_postfix_expression,
    ACTIONS(12165), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    STATE(3390), 7,
      sym__c_expression_not_assignment,
      sym__c_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [277990] = 16,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12167), 1,
      anon_sym_LPAREN,
    ACTIONS(12171), 1,
      anon_sym_DOLLAR,
    ACTIONS(12173), 1,
      anon_sym_DQUOTE,
    ACTIONS(12175), 1,
      aux_sym_number_token1,
    ACTIONS(12177), 1,
      aux_sym_number_token2,
    ACTIONS(12179), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(12181), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(12183), 1,
      anon_sym_BQUOTE,
    ACTIONS(12185), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(12512), 1,
      aux_sym__c_word_token1,
    STATE(3442), 1,
      sym__c_unary_expression,
    STATE(3443), 1,
      sym__c_binary_expression,
    STATE(3444), 1,
      sym__c_postfix_expression,
    ACTIONS(12165), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    STATE(3387), 7,
      sym__c_expression_not_assignment,
      sym__c_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [278046] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1251), 6,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1253), 16,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_RPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [278076] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1324), 6,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1326), 16,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_RPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [278106] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12471), 1,
      anon_sym_esac,
    ACTIONS(12467), 6,
      anon_sym_LPAREN,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(12469), 15,
      sym_test_operator,
      sym_extglob_pattern,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [278138] = 16,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12167), 1,
      anon_sym_LPAREN,
    ACTIONS(12171), 1,
      anon_sym_DOLLAR,
    ACTIONS(12173), 1,
      anon_sym_DQUOTE,
    ACTIONS(12175), 1,
      aux_sym_number_token1,
    ACTIONS(12177), 1,
      aux_sym_number_token2,
    ACTIONS(12179), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(12181), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(12183), 1,
      anon_sym_BQUOTE,
    ACTIONS(12185), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(12514), 1,
      aux_sym__c_word_token1,
    STATE(3442), 1,
      sym__c_unary_expression,
    STATE(3443), 1,
      sym__c_binary_expression,
    STATE(3444), 1,
      sym__c_postfix_expression,
    ACTIONS(12165), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    STATE(3384), 7,
      sym__c_expression_not_assignment,
      sym__c_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [278194] = 16,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12167), 1,
      anon_sym_LPAREN,
    ACTIONS(12171), 1,
      anon_sym_DOLLAR,
    ACTIONS(12173), 1,
      anon_sym_DQUOTE,
    ACTIONS(12175), 1,
      aux_sym_number_token1,
    ACTIONS(12177), 1,
      aux_sym_number_token2,
    ACTIONS(12179), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(12181), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(12183), 1,
      anon_sym_BQUOTE,
    ACTIONS(12185), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(12516), 1,
      aux_sym__c_word_token1,
    STATE(3442), 1,
      sym__c_unary_expression,
    STATE(3443), 1,
      sym__c_binary_expression,
    STATE(3444), 1,
      sym__c_postfix_expression,
    ACTIONS(12165), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    STATE(3381), 7,
      sym__c_expression_not_assignment,
      sym__c_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [278250] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1348), 6,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      sym_word,
    ACTIONS(1350), 16,
      sym__concat,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_RPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_concatenation_token1,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [278280] = 9,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12527), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(12530), 1,
      sym_file_descriptor,
    ACTIONS(11307), 2,
      anon_sym_PIPE,
      anon_sym_LT_LT,
    ACTIONS(12524), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(12521), 3,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
    STATE(5205), 3,
      sym_file_redirect,
      sym_herestring_redirect,
      aux_sym_redirected_statement_repeat2,
    ACTIONS(11315), 5,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_LT_LT_DASH,
    ACTIONS(12518), 5,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
  [278322] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11651), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11653), 1,
      sym__concat,
    STATE(5393), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1263), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1261), 17,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      sym__special_character,
  [278358] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11512), 1,
      aux_sym_concatenation_token1,
    ACTIONS(12533), 1,
      sym__concat,
    STATE(4253), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1267), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1265), 17,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [278394] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11512), 1,
      aux_sym_concatenation_token1,
    ACTIONS(12535), 1,
      sym__concat,
    STATE(4253), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1288), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1286), 17,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [278430] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11500), 1,
      aux_sym_concatenation_token1,
    ACTIONS(12537), 1,
      sym__concat,
    STATE(4244), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1267), 3,
      sym_file_descriptor,
      sym_variable_name,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1265), 16,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [278466] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11500), 1,
      aux_sym_concatenation_token1,
    ACTIONS(12539), 1,
      sym__concat,
    STATE(4244), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1288), 3,
      sym_file_descriptor,
      sym_variable_name,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1286), 16,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [278502] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(7793), 1,
      aux_sym_number_token1,
    ACTIONS(7795), 1,
      aux_sym_number_token2,
    ACTIONS(12543), 1,
      anon_sym_LPAREN,
    ACTIONS(12545), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12547), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(12549), 1,
      anon_sym_DOLLAR,
    ACTIONS(12551), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(12553), 1,
      anon_sym_RBRACE3,
    ACTIONS(12555), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(12557), 1,
      anon_sym_BQUOTE,
    ACTIONS(12559), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(12561), 1,
      aux_sym__simple_variable_name_token1,
    STATE(5460), 1,
      sym_simple_expansion,
    STATE(6264), 1,
      sym_number,
    STATE(6267), 1,
      sym_expansion,
    STATE(6268), 1,
      sym__expansion_max_length_binary_expression,
    STATE(6470), 1,
      sym__expansion_max_length_expression,
    STATE(7025), 1,
      sym_parenthesized_expression,
    STATE(7026), 1,
      sym_arithmetic_expansion,
    STATE(7027), 1,
      sym_command_substitution,
    ACTIONS(12541), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
  [278570] = 19,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(7793), 1,
      aux_sym_number_token1,
    ACTIONS(7795), 1,
      aux_sym_number_token2,
    ACTIONS(12543), 1,
      anon_sym_LPAREN,
    ACTIONS(12547), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(12549), 1,
      anon_sym_DOLLAR,
    ACTIONS(12551), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(12555), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(12557), 1,
      anon_sym_BQUOTE,
    ACTIONS(12559), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(12563), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12565), 1,
      anon_sym_RBRACE3,
    ACTIONS(12567), 1,
      aux_sym__simple_variable_name_token1,
    STATE(5458), 1,
      sym_simple_expansion,
    STATE(6256), 1,
      sym__expansion_max_length_binary_expression,
    STATE(6470), 1,
      sym__expansion_max_length_expression,
    ACTIONS(12541), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6344), 2,
      sym_number,
      sym_expansion,
    STATE(6977), 3,
      sym_parenthesized_expression,
      sym_arithmetic_expansion,
      sym_command_substitution,
  [278632] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12473), 1,
      anon_sym_LT_LT_LT,
    STATE(5679), 1,
      sym_herestring_redirect,
    STATE(5436), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(4253), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(4348), 10,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [278668] = 16,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(11808), 1,
      anon_sym_LPAREN,
    ACTIONS(11812), 1,
      anon_sym_DOLLAR,
    ACTIONS(11814), 1,
      anon_sym_DQUOTE,
    ACTIONS(11816), 1,
      aux_sym_number_token1,
    ACTIONS(11818), 1,
      aux_sym_number_token2,
    ACTIONS(11820), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(11822), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(11824), 1,
      anon_sym_BQUOTE,
    ACTIONS(11826), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(12569), 1,
      aux_sym__c_word_token1,
    STATE(3285), 1,
      sym__c_postfix_expression,
    STATE(3286), 1,
      sym__c_binary_expression,
    STATE(3290), 1,
      sym__c_unary_expression,
    ACTIONS(11806), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    STATE(3217), 7,
      sym__c_expression_not_assignment,
      sym__c_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [278724] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12360), 1,
      anon_sym_esac,
    ACTIONS(12356), 6,
      anon_sym_LPAREN,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(12358), 15,
      sym_test_operator,
      sym_extglob_pattern,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [278756] = 5,
    ACTIONS(3), 1,
      sym_comment,
    STATE(5216), 1,
      aux_sym_pipeline_repeat1,
    ACTIONS(11748), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12571), 2,
      anon_sym_PIPE,
      anon_sym_PIPE_AMP,
    ACTIONS(11743), 17,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [278790] = 16,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(11808), 1,
      anon_sym_LPAREN,
    ACTIONS(11812), 1,
      anon_sym_DOLLAR,
    ACTIONS(11814), 1,
      anon_sym_DQUOTE,
    ACTIONS(11816), 1,
      aux_sym_number_token1,
    ACTIONS(11818), 1,
      aux_sym_number_token2,
    ACTIONS(11820), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(11822), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(11824), 1,
      anon_sym_BQUOTE,
    ACTIONS(11826), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(12574), 1,
      aux_sym__c_word_token1,
    STATE(3285), 1,
      sym__c_postfix_expression,
    STATE(3286), 1,
      sym__c_binary_expression,
    STATE(3290), 1,
      sym__c_unary_expression,
    ACTIONS(11806), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    STATE(3218), 7,
      sym__c_expression_not_assignment,
      sym__c_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [278846] = 16,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(11808), 1,
      anon_sym_LPAREN,
    ACTIONS(11812), 1,
      anon_sym_DOLLAR,
    ACTIONS(11814), 1,
      anon_sym_DQUOTE,
    ACTIONS(11816), 1,
      aux_sym_number_token1,
    ACTIONS(11818), 1,
      aux_sym_number_token2,
    ACTIONS(11820), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(11822), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(11824), 1,
      anon_sym_BQUOTE,
    ACTIONS(11826), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(12576), 1,
      aux_sym__c_word_token1,
    STATE(3285), 1,
      sym__c_postfix_expression,
    STATE(3286), 1,
      sym__c_binary_expression,
    STATE(3290), 1,
      sym__c_unary_expression,
    ACTIONS(11806), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    STATE(3216), 7,
      sym__c_expression_not_assignment,
      sym__c_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [278902] = 16,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(11808), 1,
      anon_sym_LPAREN,
    ACTIONS(11812), 1,
      anon_sym_DOLLAR,
    ACTIONS(11814), 1,
      anon_sym_DQUOTE,
    ACTIONS(11816), 1,
      aux_sym_number_token1,
    ACTIONS(11818), 1,
      aux_sym_number_token2,
    ACTIONS(11820), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(11822), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(11824), 1,
      anon_sym_BQUOTE,
    ACTIONS(11826), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(12578), 1,
      aux_sym__c_word_token1,
    STATE(3285), 1,
      sym__c_postfix_expression,
    STATE(3286), 1,
      sym__c_binary_expression,
    STATE(3290), 1,
      sym__c_unary_expression,
    ACTIONS(11806), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    STATE(3214), 7,
      sym__c_expression_not_assignment,
      sym__c_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [278958] = 16,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(11808), 1,
      anon_sym_LPAREN,
    ACTIONS(11812), 1,
      anon_sym_DOLLAR,
    ACTIONS(11814), 1,
      anon_sym_DQUOTE,
    ACTIONS(11816), 1,
      aux_sym_number_token1,
    ACTIONS(11818), 1,
      aux_sym_number_token2,
    ACTIONS(11820), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(11822), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(11824), 1,
      anon_sym_BQUOTE,
    ACTIONS(11826), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(12580), 1,
      aux_sym__c_word_token1,
    STATE(3285), 1,
      sym__c_postfix_expression,
    STATE(3286), 1,
      sym__c_binary_expression,
    STATE(3290), 1,
      sym__c_unary_expression,
    ACTIONS(11806), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    STATE(3207), 7,
      sym__c_expression_not_assignment,
      sym__c_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [279014] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(5297), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(12582), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(1261), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1263), 12,
      sym_file_descriptor,
      sym_variable_name,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      sym__special_character,
  [279048] = 16,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(11808), 1,
      anon_sym_LPAREN,
    ACTIONS(11812), 1,
      anon_sym_DOLLAR,
    ACTIONS(11814), 1,
      anon_sym_DQUOTE,
    ACTIONS(11816), 1,
      aux_sym_number_token1,
    ACTIONS(11818), 1,
      aux_sym_number_token2,
    ACTIONS(11820), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(11822), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(11824), 1,
      anon_sym_BQUOTE,
    ACTIONS(11826), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(12584), 1,
      aux_sym__c_word_token1,
    STATE(3285), 1,
      sym__c_postfix_expression,
    STATE(3286), 1,
      sym__c_binary_expression,
    STATE(3290), 1,
      sym__c_unary_expression,
    ACTIONS(11806), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    STATE(3205), 7,
      sym__c_expression_not_assignment,
      sym__c_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [279104] = 16,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(11808), 1,
      anon_sym_LPAREN,
    ACTIONS(11812), 1,
      anon_sym_DOLLAR,
    ACTIONS(11814), 1,
      anon_sym_DQUOTE,
    ACTIONS(11816), 1,
      aux_sym_number_token1,
    ACTIONS(11818), 1,
      aux_sym_number_token2,
    ACTIONS(11820), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(11822), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(11824), 1,
      anon_sym_BQUOTE,
    ACTIONS(11826), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(12586), 1,
      aux_sym__c_word_token1,
    STATE(3285), 1,
      sym__c_postfix_expression,
    STATE(3286), 1,
      sym__c_binary_expression,
    STATE(3290), 1,
      sym__c_unary_expression,
    ACTIONS(11806), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    STATE(3202), 7,
      sym__c_expression_not_assignment,
      sym__c_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [279160] = 5,
    ACTIONS(3), 1,
      sym_comment,
    STATE(5216), 1,
      aux_sym_pipeline_repeat1,
    ACTIONS(12068), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12588), 2,
      anon_sym_PIPE,
      anon_sym_PIPE_AMP,
    ACTIONS(12064), 17,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI_SEMI,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [279194] = 16,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(11808), 1,
      anon_sym_LPAREN,
    ACTIONS(11812), 1,
      anon_sym_DOLLAR,
    ACTIONS(11814), 1,
      anon_sym_DQUOTE,
    ACTIONS(11816), 1,
      aux_sym_number_token1,
    ACTIONS(11818), 1,
      aux_sym_number_token2,
    ACTIONS(11820), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(11822), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(11824), 1,
      anon_sym_BQUOTE,
    ACTIONS(11826), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(12590), 1,
      aux_sym__c_word_token1,
    STATE(3285), 1,
      sym__c_postfix_expression,
    STATE(3286), 1,
      sym__c_binary_expression,
    STATE(3290), 1,
      sym__c_unary_expression,
    ACTIONS(11806), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    STATE(3201), 7,
      sym__c_expression_not_assignment,
      sym__c_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [279250] = 16,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(11256), 1,
      anon_sym_DOLLAR,
    ACTIONS(11262), 1,
      aux_sym_number_token2,
    ACTIONS(12233), 1,
      anon_sym_LPAREN,
    ACTIONS(12237), 1,
      anon_sym_DQUOTE,
    ACTIONS(12239), 1,
      aux_sym_number_token1,
    ACTIONS(12241), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(12243), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(12245), 1,
      anon_sym_BQUOTE,
    ACTIONS(12247), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(12592), 1,
      aux_sym__c_word_token1,
    STATE(3113), 1,
      sym__c_postfix_expression,
    STATE(3116), 1,
      sym__c_binary_expression,
    STATE(3118), 1,
      sym__c_unary_expression,
    ACTIONS(12231), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    STATE(3075), 7,
      sym__c_expression_not_assignment,
      sym__c_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [279306] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12494), 1,
      anon_sym_esac,
    ACTIONS(12490), 6,
      anon_sym_LPAREN,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(12492), 15,
      sym_test_operator,
      sym_extglob_pattern,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [279338] = 16,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(11808), 1,
      anon_sym_LPAREN,
    ACTIONS(11812), 1,
      anon_sym_DOLLAR,
    ACTIONS(11814), 1,
      anon_sym_DQUOTE,
    ACTIONS(11816), 1,
      aux_sym_number_token1,
    ACTIONS(11818), 1,
      aux_sym_number_token2,
    ACTIONS(11820), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(11822), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(11824), 1,
      anon_sym_BQUOTE,
    ACTIONS(11826), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(12594), 1,
      aux_sym__c_word_token1,
    STATE(3285), 1,
      sym__c_postfix_expression,
    STATE(3286), 1,
      sym__c_binary_expression,
    STATE(3290), 1,
      sym__c_unary_expression,
    ACTIONS(11806), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    STATE(3200), 7,
      sym__c_expression_not_assignment,
      sym__c_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [279394] = 16,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(11808), 1,
      anon_sym_LPAREN,
    ACTIONS(11812), 1,
      anon_sym_DOLLAR,
    ACTIONS(11814), 1,
      anon_sym_DQUOTE,
    ACTIONS(11816), 1,
      aux_sym_number_token1,
    ACTIONS(11818), 1,
      aux_sym_number_token2,
    ACTIONS(11820), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(11822), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(11824), 1,
      anon_sym_BQUOTE,
    ACTIONS(11826), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(12596), 1,
      aux_sym__c_word_token1,
    STATE(3285), 1,
      sym__c_postfix_expression,
    STATE(3286), 1,
      sym__c_binary_expression,
    STATE(3290), 1,
      sym__c_unary_expression,
    ACTIONS(11806), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    STATE(3199), 7,
      sym__c_expression_not_assignment,
      sym__c_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [279450] = 16,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(11808), 1,
      anon_sym_LPAREN,
    ACTIONS(11812), 1,
      anon_sym_DOLLAR,
    ACTIONS(11814), 1,
      anon_sym_DQUOTE,
    ACTIONS(11816), 1,
      aux_sym_number_token1,
    ACTIONS(11818), 1,
      aux_sym_number_token2,
    ACTIONS(11820), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(11822), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(11824), 1,
      anon_sym_BQUOTE,
    ACTIONS(11826), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(12598), 1,
      aux_sym__c_word_token1,
    STATE(3285), 1,
      sym__c_postfix_expression,
    STATE(3286), 1,
      sym__c_binary_expression,
    STATE(3290), 1,
      sym__c_unary_expression,
    ACTIONS(11806), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    STATE(3198), 7,
      sym__c_expression_not_assignment,
      sym__c_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [279506] = 16,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(11808), 1,
      anon_sym_LPAREN,
    ACTIONS(11812), 1,
      anon_sym_DOLLAR,
    ACTIONS(11814), 1,
      anon_sym_DQUOTE,
    ACTIONS(11816), 1,
      aux_sym_number_token1,
    ACTIONS(11818), 1,
      aux_sym_number_token2,
    ACTIONS(11820), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(11822), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(11824), 1,
      anon_sym_BQUOTE,
    ACTIONS(11826), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(12600), 1,
      aux_sym__c_word_token1,
    STATE(3285), 1,
      sym__c_postfix_expression,
    STATE(3286), 1,
      sym__c_binary_expression,
    STATE(3290), 1,
      sym__c_unary_expression,
    ACTIONS(11806), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    STATE(3196), 7,
      sym__c_expression_not_assignment,
      sym__c_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [279562] = 19,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(7793), 1,
      aux_sym_number_token1,
    ACTIONS(7795), 1,
      aux_sym_number_token2,
    ACTIONS(12543), 1,
      anon_sym_LPAREN,
    ACTIONS(12547), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(12549), 1,
      anon_sym_DOLLAR,
    ACTIONS(12551), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(12555), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(12557), 1,
      anon_sym_BQUOTE,
    ACTIONS(12559), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(12602), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12604), 1,
      anon_sym_RBRACE3,
    ACTIONS(12606), 1,
      aux_sym__simple_variable_name_token1,
    STATE(5429), 1,
      sym_simple_expansion,
    STATE(6266), 1,
      sym__expansion_max_length_binary_expression,
    STATE(6470), 1,
      sym__expansion_max_length_expression,
    ACTIONS(12541), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6260), 2,
      sym_number,
      sym_expansion,
    STATE(6798), 3,
      sym_parenthesized_expression,
      sym_arithmetic_expansion,
      sym_command_substitution,
  [279624] = 16,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(11256), 1,
      anon_sym_DOLLAR,
    ACTIONS(11262), 1,
      aux_sym_number_token2,
    ACTIONS(12233), 1,
      anon_sym_LPAREN,
    ACTIONS(12237), 1,
      anon_sym_DQUOTE,
    ACTIONS(12239), 1,
      aux_sym_number_token1,
    ACTIONS(12241), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(12243), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(12245), 1,
      anon_sym_BQUOTE,
    ACTIONS(12247), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(12608), 1,
      aux_sym__c_word_token1,
    STATE(3113), 1,
      sym__c_postfix_expression,
    STATE(3116), 1,
      sym__c_binary_expression,
    STATE(3118), 1,
      sym__c_unary_expression,
    ACTIONS(12231), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    STATE(2948), 7,
      sym__c_expression_not_assignment,
      sym__c_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [279680] = 16,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(11256), 1,
      anon_sym_DOLLAR,
    ACTIONS(11262), 1,
      aux_sym_number_token2,
    ACTIONS(12233), 1,
      anon_sym_LPAREN,
    ACTIONS(12237), 1,
      anon_sym_DQUOTE,
    ACTIONS(12239), 1,
      aux_sym_number_token1,
    ACTIONS(12241), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(12243), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(12245), 1,
      anon_sym_BQUOTE,
    ACTIONS(12247), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(12610), 1,
      aux_sym__c_word_token1,
    STATE(3113), 1,
      sym__c_postfix_expression,
    STATE(3116), 1,
      sym__c_binary_expression,
    STATE(3118), 1,
      sym__c_unary_expression,
    ACTIONS(12231), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    STATE(3005), 7,
      sym__c_expression_not_assignment,
      sym__c_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [279736] = 16,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(11256), 1,
      anon_sym_DOLLAR,
    ACTIONS(11262), 1,
      aux_sym_number_token2,
    ACTIONS(12233), 1,
      anon_sym_LPAREN,
    ACTIONS(12237), 1,
      anon_sym_DQUOTE,
    ACTIONS(12239), 1,
      aux_sym_number_token1,
    ACTIONS(12241), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(12243), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(12245), 1,
      anon_sym_BQUOTE,
    ACTIONS(12247), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(12612), 1,
      aux_sym__c_word_token1,
    STATE(3113), 1,
      sym__c_postfix_expression,
    STATE(3116), 1,
      sym__c_binary_expression,
    STATE(3118), 1,
      sym__c_unary_expression,
    ACTIONS(12231), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    STATE(2999), 7,
      sym__c_expression_not_assignment,
      sym__c_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [279792] = 16,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(11256), 1,
      anon_sym_DOLLAR,
    ACTIONS(11262), 1,
      aux_sym_number_token2,
    ACTIONS(12233), 1,
      anon_sym_LPAREN,
    ACTIONS(12237), 1,
      anon_sym_DQUOTE,
    ACTIONS(12239), 1,
      aux_sym_number_token1,
    ACTIONS(12241), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(12243), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(12245), 1,
      anon_sym_BQUOTE,
    ACTIONS(12247), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(12614), 1,
      aux_sym__c_word_token1,
    STATE(3113), 1,
      sym__c_postfix_expression,
    STATE(3116), 1,
      sym__c_binary_expression,
    STATE(3118), 1,
      sym__c_unary_expression,
    ACTIONS(12231), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    STATE(2993), 7,
      sym__c_expression_not_assignment,
      sym__c_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [279848] = 16,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(11256), 1,
      anon_sym_DOLLAR,
    ACTIONS(11262), 1,
      aux_sym_number_token2,
    ACTIONS(12233), 1,
      anon_sym_LPAREN,
    ACTIONS(12237), 1,
      anon_sym_DQUOTE,
    ACTIONS(12239), 1,
      aux_sym_number_token1,
    ACTIONS(12241), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(12243), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(12245), 1,
      anon_sym_BQUOTE,
    ACTIONS(12247), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(12616), 1,
      aux_sym__c_word_token1,
    STATE(3113), 1,
      sym__c_postfix_expression,
    STATE(3116), 1,
      sym__c_binary_expression,
    STATE(3118), 1,
      sym__c_unary_expression,
    ACTIONS(12231), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    STATE(2992), 7,
      sym__c_expression_not_assignment,
      sym__c_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [279904] = 16,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(11256), 1,
      anon_sym_DOLLAR,
    ACTIONS(11262), 1,
      aux_sym_number_token2,
    ACTIONS(12233), 1,
      anon_sym_LPAREN,
    ACTIONS(12237), 1,
      anon_sym_DQUOTE,
    ACTIONS(12239), 1,
      aux_sym_number_token1,
    ACTIONS(12241), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(12243), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(12245), 1,
      anon_sym_BQUOTE,
    ACTIONS(12247), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(12618), 1,
      aux_sym__c_word_token1,
    STATE(3113), 1,
      sym__c_postfix_expression,
    STATE(3116), 1,
      sym__c_binary_expression,
    STATE(3118), 1,
      sym__c_unary_expression,
    ACTIONS(12231), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    STATE(2991), 7,
      sym__c_expression_not_assignment,
      sym__c_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [279960] = 16,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(11256), 1,
      anon_sym_DOLLAR,
    ACTIONS(11262), 1,
      aux_sym_number_token2,
    ACTIONS(12233), 1,
      anon_sym_LPAREN,
    ACTIONS(12237), 1,
      anon_sym_DQUOTE,
    ACTIONS(12239), 1,
      aux_sym_number_token1,
    ACTIONS(12241), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(12243), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(12245), 1,
      anon_sym_BQUOTE,
    ACTIONS(12247), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(12620), 1,
      aux_sym__c_word_token1,
    STATE(3113), 1,
      sym__c_postfix_expression,
    STATE(3116), 1,
      sym__c_binary_expression,
    STATE(3118), 1,
      sym__c_unary_expression,
    ACTIONS(12231), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    STATE(3019), 7,
      sym__c_expression_not_assignment,
      sym__c_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [280016] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(5270), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(12622), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(1261), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1263), 12,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      sym__special_character,
  [280050] = 16,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(11256), 1,
      anon_sym_DOLLAR,
    ACTIONS(11262), 1,
      aux_sym_number_token2,
    ACTIONS(12233), 1,
      anon_sym_LPAREN,
    ACTIONS(12237), 1,
      anon_sym_DQUOTE,
    ACTIONS(12239), 1,
      aux_sym_number_token1,
    ACTIONS(12241), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(12243), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(12245), 1,
      anon_sym_BQUOTE,
    ACTIONS(12247), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(12624), 1,
      aux_sym__c_word_token1,
    STATE(3113), 1,
      sym__c_postfix_expression,
    STATE(3116), 1,
      sym__c_binary_expression,
    STATE(3118), 1,
      sym__c_unary_expression,
    ACTIONS(12231), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    STATE(2965), 7,
      sym__c_expression_not_assignment,
      sym__c_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [280106] = 16,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(11256), 1,
      anon_sym_DOLLAR,
    ACTIONS(11262), 1,
      aux_sym_number_token2,
    ACTIONS(12233), 1,
      anon_sym_LPAREN,
    ACTIONS(12237), 1,
      anon_sym_DQUOTE,
    ACTIONS(12239), 1,
      aux_sym_number_token1,
    ACTIONS(12241), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(12243), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(12245), 1,
      anon_sym_BQUOTE,
    ACTIONS(12247), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(12626), 1,
      aux_sym__c_word_token1,
    STATE(3113), 1,
      sym__c_postfix_expression,
    STATE(3116), 1,
      sym__c_binary_expression,
    STATE(3118), 1,
      sym__c_unary_expression,
    ACTIONS(12231), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    STATE(2956), 7,
      sym__c_expression_not_assignment,
      sym__c_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [280162] = 16,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(11256), 1,
      anon_sym_DOLLAR,
    ACTIONS(11262), 1,
      aux_sym_number_token2,
    ACTIONS(12233), 1,
      anon_sym_LPAREN,
    ACTIONS(12237), 1,
      anon_sym_DQUOTE,
    ACTIONS(12239), 1,
      aux_sym_number_token1,
    ACTIONS(12241), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(12243), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(12245), 1,
      anon_sym_BQUOTE,
    ACTIONS(12247), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(12628), 1,
      aux_sym__c_word_token1,
    STATE(3113), 1,
      sym__c_postfix_expression,
    STATE(3116), 1,
      sym__c_binary_expression,
    STATE(3118), 1,
      sym__c_unary_expression,
    ACTIONS(12231), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    STATE(2952), 7,
      sym__c_expression_not_assignment,
      sym__c_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [280218] = 16,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(11256), 1,
      anon_sym_DOLLAR,
    ACTIONS(11262), 1,
      aux_sym_number_token2,
    ACTIONS(12233), 1,
      anon_sym_LPAREN,
    ACTIONS(12237), 1,
      anon_sym_DQUOTE,
    ACTIONS(12239), 1,
      aux_sym_number_token1,
    ACTIONS(12241), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(12243), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(12245), 1,
      anon_sym_BQUOTE,
    ACTIONS(12247), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(12630), 1,
      aux_sym__c_word_token1,
    STATE(3113), 1,
      sym__c_postfix_expression,
    STATE(3116), 1,
      sym__c_binary_expression,
    STATE(3118), 1,
      sym__c_unary_expression,
    ACTIONS(12231), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    STATE(3036), 7,
      sym__c_expression_not_assignment,
      sym__c_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [280274] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12636), 1,
      anon_sym_esac,
    ACTIONS(12632), 6,
      anon_sym_LPAREN,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(12634), 15,
      sym_test_operator,
      sym_extglob_pattern,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [280306] = 16,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(11256), 1,
      anon_sym_DOLLAR,
    ACTIONS(11262), 1,
      aux_sym_number_token2,
    ACTIONS(12233), 1,
      anon_sym_LPAREN,
    ACTIONS(12237), 1,
      anon_sym_DQUOTE,
    ACTIONS(12239), 1,
      aux_sym_number_token1,
    ACTIONS(12241), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(12243), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(12245), 1,
      anon_sym_BQUOTE,
    ACTIONS(12247), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(12638), 1,
      aux_sym__c_word_token1,
    STATE(3113), 1,
      sym__c_postfix_expression,
    STATE(3116), 1,
      sym__c_binary_expression,
    STATE(3118), 1,
      sym__c_unary_expression,
    ACTIONS(12231), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    STATE(2947), 7,
      sym__c_expression_not_assignment,
      sym__c_parenthesized_expression,
      sym_string,
      sym_number,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [280362] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12644), 1,
      anon_sym_esac,
    ACTIONS(12640), 6,
      anon_sym_LPAREN,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(12642), 15,
      sym_test_operator,
      sym_extglob_pattern,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [280394] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5142), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12221), 1,
      sym_file_descriptor,
    ACTIONS(4692), 2,
      anon_sym_LT_LT,
      anon_sym_LT_LT_DASH,
    ACTIONS(5099), 2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
    ACTIONS(11556), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(5140), 3,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
    STATE(4986), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(11554), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
  [280436] = 15,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7789), 1,
      anon_sym_DQUOTE,
    ACTIONS(7797), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(7809), 1,
      sym_variable_name,
    ACTIONS(12648), 1,
      anon_sym_LPAREN,
    ACTIONS(12650), 1,
      anon_sym_DOLLAR,
    ACTIONS(12652), 1,
      anon_sym_RBRACE3,
    ACTIONS(12654), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(12656), 1,
      anon_sym_BQUOTE,
    ACTIONS(12658), 1,
      anon_sym_DOLLAR_BQUOTE,
    STATE(6631), 1,
      sym_process_substitution,
    STATE(7008), 1,
      sym__concatenation_in_expansion,
    ACTIONS(7240), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(12646), 4,
      sym__expansion_word,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(6498), 5,
      sym_string,
      sym_array,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [280490] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12644), 1,
      anon_sym_esac,
    ACTIONS(12640), 6,
      anon_sym_LPAREN,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(12642), 15,
      sym_test_operator,
      sym_extglob_pattern,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [280522] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12640), 6,
      anon_sym_LPAREN,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(12642), 15,
      sym_test_operator,
      sym_extglob_pattern,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [280551] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12490), 6,
      anon_sym_LPAREN,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(12492), 15,
      sym_test_operator,
      sym_extglob_pattern,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [280580] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(5390), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(12582), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(5373), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(5375), 11,
      sym_file_descriptor,
      sym_variable_name,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [280613] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12660), 6,
      anon_sym_LPAREN,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(12662), 15,
      sym_test_operator,
      sym_extglob_pattern,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [280642] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1336), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1338), 14,
      sym_file_descriptor,
      sym__concat,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      aux_sym_concatenation_token1,
  [280671] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(5297), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(12582), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(5067), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(5069), 11,
      sym_file_descriptor,
      sym_variable_name,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [280704] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12664), 6,
      anon_sym_LPAREN,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(12666), 15,
      sym_test_operator,
      sym_extglob_pattern,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [280733] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12668), 1,
      sym__special_character,
    STATE(5388), 1,
      aux_sym__literal_repeat1,
    ACTIONS(4467), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(4469), 12,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [280766] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12670), 6,
      anon_sym_LPAREN,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(12672), 15,
      sym_test_operator,
      sym_extglob_pattern,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [280795] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(7793), 1,
      aux_sym_number_token1,
    ACTIONS(7795), 1,
      aux_sym_number_token2,
    ACTIONS(12543), 1,
      anon_sym_LPAREN,
    ACTIONS(12547), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(12551), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(12555), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(12557), 1,
      anon_sym_BQUOTE,
    ACTIONS(12559), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(12674), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12676), 1,
      anon_sym_COLON,
    ACTIONS(12678), 1,
      anon_sym_RBRACE3,
    ACTIONS(12680), 1,
      aux_sym__simple_variable_name_token1,
    STATE(6218), 1,
      sym__expansion_max_length_binary_expression,
    STATE(6470), 1,
      sym__expansion_max_length_expression,
    ACTIONS(12541), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6208), 2,
      sym_number,
      sym_expansion,
    STATE(6780), 3,
      sym_parenthesized_expression,
      sym_arithmetic_expansion,
      sym_command_substitution,
  [280854] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12423), 6,
      anon_sym_LPAREN,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(12425), 15,
      sym_test_operator,
      sym_extglob_pattern,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [280883] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1308), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1310), 14,
      sym_file_descriptor,
      sym__concat,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      aux_sym_concatenation_token1,
  [280912] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1308), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1310), 14,
      sym_file_descriptor,
      sym__concat,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      aux_sym_concatenation_token1,
  [280941] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12682), 6,
      anon_sym_LPAREN,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(12684), 15,
      sym_test_operator,
      sym_extglob_pattern,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [280970] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12682), 6,
      anon_sym_LPAREN,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(12684), 15,
      sym_test_operator,
      sym_extglob_pattern,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [280999] = 12,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(4253), 1,
      anon_sym_PIPE,
    ACTIONS(4348), 1,
      anon_sym_PIPE_AMP,
    ACTIONS(5994), 1,
      anon_sym_LT_LT,
    ACTIONS(5996), 1,
      anon_sym_LT_LT_DASH,
    ACTIONS(7517), 1,
      anon_sym_RBRACK,
    ACTIONS(12686), 1,
      sym_file_descriptor,
    ACTIONS(5992), 2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
    ACTIONS(12502), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(12500), 3,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
    STATE(5350), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(12498), 5,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
  [281046] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(5471), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(12688), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(1261), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1263), 11,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      sym__special_character,
  [281079] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12690), 6,
      anon_sym_LPAREN,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(12692), 15,
      sym_test_operator,
      sym_extglob_pattern,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [281108] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1344), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1346), 14,
      sym_file_descriptor,
      sym__concat,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      aux_sym_concatenation_token1,
  [281137] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12622), 1,
      aux_sym_concatenation_token1,
    ACTIONS(12694), 1,
      sym__concat,
    STATE(5276), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1265), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1267), 11,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [281172] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12622), 1,
      aux_sym_concatenation_token1,
    ACTIONS(12696), 1,
      sym__concat,
    STATE(5276), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1286), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1288), 11,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [281207] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12690), 6,
      anon_sym_LPAREN,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(12692), 15,
      sym_test_operator,
      sym_extglob_pattern,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [281236] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12664), 6,
      anon_sym_LPAREN,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(12666), 15,
      sym_test_operator,
      sym_extglob_pattern,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [281265] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12698), 1,
      sym__special_character,
    STATE(5274), 1,
      aux_sym__literal_repeat1,
    ACTIONS(1362), 5,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(1364), 14,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_RPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [281298] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12701), 6,
      anon_sym_LPAREN,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(12703), 15,
      sym_test_operator,
      sym_extglob_pattern,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [281327] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(5276), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(12705), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(1251), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1253), 11,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [281360] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12708), 6,
      anon_sym_LPAREN,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(12710), 15,
      sym_test_operator,
      sym_extglob_pattern,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [281389] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(5328), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(12435), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(4552), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(4554), 11,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [281422] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1356), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1358), 14,
      sym_file_descriptor,
      sym__concat,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      aux_sym_concatenation_token1,
  [281451] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12664), 6,
      anon_sym_LPAREN,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(12666), 15,
      sym_test_operator,
      sym_extglob_pattern,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [281480] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12708), 6,
      anon_sym_LPAREN,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(12710), 15,
      sym_test_operator,
      sym_extglob_pattern,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [281509] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12682), 6,
      anon_sym_LPAREN,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(12684), 15,
      sym_test_operator,
      sym_extglob_pattern,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [281538] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12682), 6,
      anon_sym_LPAREN,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(12684), 15,
      sym_test_operator,
      sym_extglob_pattern,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [281567] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1304), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1306), 14,
      sym_file_descriptor,
      sym__concat,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      aux_sym_concatenation_token1,
  [281596] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12701), 6,
      anon_sym_LPAREN,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(12703), 15,
      sym_test_operator,
      sym_extglob_pattern,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [281625] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11651), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11653), 1,
      sym__concat,
    STATE(5392), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4554), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4552), 16,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [281660] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11651), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11653), 1,
      sym__concat,
    STATE(5393), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4566), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4564), 16,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [281695] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1352), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1354), 14,
      sym_file_descriptor,
      sym__concat,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      aux_sym_concatenation_token1,
  [281724] = 12,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(4253), 1,
      anon_sym_PIPE,
    ACTIONS(4348), 1,
      anon_sym_PIPE_AMP,
    ACTIONS(5994), 1,
      anon_sym_LT_LT,
    ACTIONS(5996), 1,
      anon_sym_LT_LT_DASH,
    ACTIONS(7561), 1,
      anon_sym_RBRACK,
    ACTIONS(12686), 1,
      sym_file_descriptor,
    ACTIONS(5992), 2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
    ACTIONS(12502), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(12500), 3,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
    STATE(5350), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(12498), 5,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
  [281771] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(5322), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(12435), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(4564), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(4566), 11,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [281804] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12701), 6,
      anon_sym_LPAREN,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(12703), 15,
      sym_test_operator,
      sym_extglob_pattern,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [281833] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12668), 1,
      sym__special_character,
    STATE(5388), 1,
      aux_sym__literal_repeat1,
    ACTIONS(4564), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(4566), 12,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [281866] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11651), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11653), 1,
      sym__concat,
    STATE(5392), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(2074), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(2072), 16,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [281901] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1340), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1342), 14,
      sym_file_descriptor,
      sym__concat,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      aux_sym_concatenation_token1,
  [281930] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12664), 6,
      anon_sym_LPAREN,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(12666), 15,
      sym_test_operator,
      sym_extglob_pattern,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [281959] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12690), 6,
      anon_sym_LPAREN,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(12692), 15,
      sym_test_operator,
      sym_extglob_pattern,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [281988] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12582), 1,
      aux_sym_concatenation_token1,
    ACTIONS(12712), 1,
      sym__concat,
    STATE(5374), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1265), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1267), 11,
      sym_file_descriptor,
      sym_variable_name,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [282023] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1320), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1322), 14,
      sym_file_descriptor,
      sym__concat,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      aux_sym_concatenation_token1,
  [282052] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12690), 6,
      anon_sym_LPAREN,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(12692), 15,
      sym_test_operator,
      sym_extglob_pattern,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [282081] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(5271), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(12622), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(4370), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(4372), 11,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [282114] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(5270), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(12622), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(4564), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(4566), 11,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [282147] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12405), 6,
      anon_sym_LPAREN,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(12407), 15,
      sym_test_operator,
      sym_extglob_pattern,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [282176] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(5271), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(12622), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(4552), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(4554), 11,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [282209] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12368), 6,
      anon_sym_LPAREN,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(12370), 15,
      sym_test_operator,
      sym_extglob_pattern,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [282238] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11651), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11653), 1,
      sym__concat,
    STATE(5393), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5697), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5695), 16,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [282273] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11651), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11653), 1,
      sym__concat,
    STATE(5392), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4372), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4370), 16,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [282308] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12399), 6,
      anon_sym_LPAREN,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(12401), 15,
      sym_test_operator,
      sym_extglob_pattern,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [282337] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12701), 6,
      anon_sym_LPAREN,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(12703), 15,
      sym_test_operator,
      sym_extglob_pattern,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [282366] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12714), 6,
      anon_sym_LPAREN,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(12716), 15,
      sym_test_operator,
      sym_extglob_pattern,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [282395] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12714), 6,
      anon_sym_LPAREN,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(12716), 15,
      sym_test_operator,
      sym_extglob_pattern,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [282424] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12393), 6,
      anon_sym_LPAREN,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(12395), 15,
      sym_test_operator,
      sym_extglob_pattern,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [282453] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11651), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11653), 1,
      sym__concat,
    STATE(5393), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4469), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4467), 16,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [282488] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12714), 6,
      anon_sym_LPAREN,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(12716), 15,
      sym_test_operator,
      sym_extglob_pattern,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [282517] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12714), 6,
      anon_sym_LPAREN,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(12716), 15,
      sym_test_operator,
      sym_extglob_pattern,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [282546] = 12,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(4253), 1,
      anon_sym_PIPE,
    ACTIONS(4348), 1,
      anon_sym_PIPE_AMP,
    ACTIONS(5994), 1,
      anon_sym_LT_LT,
    ACTIONS(5996), 1,
      anon_sym_LT_LT_DASH,
    ACTIONS(7463), 1,
      anon_sym_RBRACK,
    ACTIONS(12686), 1,
      sym_file_descriptor,
    ACTIONS(5992), 2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
    ACTIONS(12502), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(12500), 3,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
    STATE(5350), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(12498), 5,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
  [282593] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12718), 6,
      anon_sym_LPAREN,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(12720), 15,
      sym_test_operator,
      sym_extglob_pattern,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [282622] = 12,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(4253), 1,
      anon_sym_PIPE,
    ACTIONS(4348), 1,
      anon_sym_PIPE_AMP,
    ACTIONS(5994), 1,
      anon_sym_LT_LT,
    ACTIONS(5996), 1,
      anon_sym_LT_LT_DASH,
    ACTIONS(7557), 1,
      anon_sym_RBRACK,
    ACTIONS(12686), 1,
      sym_file_descriptor,
    ACTIONS(5992), 2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
    ACTIONS(12502), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(12500), 3,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
    STATE(5350), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(12498), 5,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
  [282669] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11651), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11653), 1,
      sym__concat,
    STATE(5393), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5723), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5721), 16,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [282704] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12382), 6,
      anon_sym_LPAREN,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(12384), 15,
      sym_test_operator,
      sym_extglob_pattern,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [282733] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12376), 6,
      anon_sym_LPAREN,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(12378), 15,
      sym_test_operator,
      sym_extglob_pattern,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [282762] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12722), 6,
      anon_sym_LPAREN,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(12724), 15,
      sym_test_operator,
      sym_extglob_pattern,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [282791] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12435), 1,
      aux_sym_concatenation_token1,
    ACTIONS(12726), 1,
      sym__concat,
    STATE(5349), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1265), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1267), 11,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [282826] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(5270), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(12622), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(4467), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(4469), 11,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [282859] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12350), 6,
      anon_sym_LPAREN,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(12352), 15,
      sym_test_operator,
      sym_extglob_pattern,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [282888] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12728), 6,
      anon_sym_LPAREN,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(12730), 15,
      sym_test_operator,
      sym_extglob_pattern,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [282917] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12376), 6,
      anon_sym_LPAREN,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(12378), 15,
      sym_test_operator,
      sym_extglob_pattern,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [282946] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1316), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1318), 14,
      sym_file_descriptor,
      sym__concat,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      aux_sym_concatenation_token1,
  [282975] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12435), 1,
      aux_sym_concatenation_token1,
    ACTIONS(12732), 1,
      sym__concat,
    STATE(5349), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1286), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1288), 11,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [283010] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12708), 6,
      anon_sym_LPAREN,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(12710), 15,
      sym_test_operator,
      sym_extglob_pattern,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [283039] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12708), 6,
      anon_sym_LPAREN,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(12710), 15,
      sym_test_operator,
      sym_extglob_pattern,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [283068] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(5328), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(12435), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(2072), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(2074), 11,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [283101] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12368), 6,
      anon_sym_LPAREN,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(12370), 15,
      sym_test_operator,
      sym_extglob_pattern,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [283130] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12356), 6,
      anon_sym_LPAREN,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(12358), 15,
      sym_test_operator,
      sym_extglob_pattern,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [283159] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12447), 6,
      anon_sym_LPAREN,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(12449), 15,
      sym_test_operator,
      sym_extglob_pattern,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [283188] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12490), 6,
      anon_sym_LPAREN,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(12492), 15,
      sym_test_operator,
      sym_extglob_pattern,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [283217] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12632), 6,
      anon_sym_LPAREN,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(12634), 15,
      sym_test_operator,
      sym_extglob_pattern,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [283246] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1312), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1314), 14,
      sym_file_descriptor,
      sym__concat,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      aux_sym_concatenation_token1,
  [283275] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12640), 6,
      anon_sym_LPAREN,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(12642), 15,
      sym_test_operator,
      sym_extglob_pattern,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [283304] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12467), 6,
      anon_sym_LPAREN,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(12469), 15,
      sym_test_operator,
      sym_extglob_pattern,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [283333] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12429), 6,
      anon_sym_LPAREN,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(12431), 15,
      sym_test_operator,
      sym_extglob_pattern,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [283362] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12734), 6,
      anon_sym_LPAREN,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(12736), 15,
      sym_test_operator,
      sym_extglob_pattern,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [283391] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12467), 6,
      anon_sym_LPAREN,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(12469), 15,
      sym_test_operator,
      sym_extglob_pattern,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [283420] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12356), 6,
      anon_sym_LPAREN,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(12358), 15,
      sym_test_operator,
      sym_extglob_pattern,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [283449] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12350), 6,
      anon_sym_LPAREN,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(12352), 15,
      sym_test_operator,
      sym_extglob_pattern,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [283478] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12362), 6,
      anon_sym_LPAREN,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(12364), 15,
      sym_test_operator,
      sym_extglob_pattern,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [283507] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12362), 6,
      anon_sym_LPAREN,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(12364), 15,
      sym_test_operator,
      sym_extglob_pattern,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [283536] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(5328), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(12435), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(4370), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(4372), 11,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [283569] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(5322), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(12435), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(4467), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(4469), 11,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [283602] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(5349), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(12738), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(1251), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1253), 11,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [283635] = 10,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(5994), 1,
      anon_sym_LT_LT,
    ACTIONS(5996), 1,
      anon_sym_LT_LT_DASH,
    ACTIONS(11453), 1,
      anon_sym_PIPE,
    ACTIONS(12686), 1,
      sym_file_descriptor,
    ACTIONS(12502), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(12500), 3,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
    STATE(5382), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(11455), 4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
    ACTIONS(12498), 5,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
  [283678] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12276), 1,
      sym__special_character,
    STATE(5021), 1,
      aux_sym__literal_repeat1,
    ACTIONS(12313), 3,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12311), 16,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [283711] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(5271), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(12622), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(2072), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(2074), 11,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [283744] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12741), 6,
      anon_sym_LPAREN,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(12743), 15,
      sym_test_operator,
      sym_extglob_pattern,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [283773] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12745), 6,
      anon_sym_LPAREN,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(12747), 15,
      sym_test_operator,
      sym_extglob_pattern,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [283802] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12734), 6,
      anon_sym_LPAREN,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(12736), 15,
      sym_test_operator,
      sym_extglob_pattern,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [283831] = 12,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(4253), 1,
      anon_sym_PIPE,
    ACTIONS(4348), 1,
      anon_sym_PIPE_AMP,
    ACTIONS(5994), 1,
      anon_sym_LT_LT,
    ACTIONS(5996), 1,
      anon_sym_LT_LT_DASH,
    ACTIONS(7551), 1,
      anon_sym_RBRACK,
    ACTIONS(12686), 1,
      sym_file_descriptor,
    ACTIONS(5992), 2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
    ACTIONS(12502), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(12500), 3,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
    STATE(5350), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(12498), 5,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
  [283878] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12668), 1,
      sym__special_character,
    STATE(5388), 1,
      aux_sym__literal_repeat1,
    ACTIONS(5695), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(5697), 12,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [283911] = 12,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(4253), 1,
      anon_sym_PIPE,
    ACTIONS(4348), 1,
      anon_sym_PIPE_AMP,
    ACTIONS(5994), 1,
      anon_sym_LT_LT,
    ACTIONS(5996), 1,
      anon_sym_LT_LT_DASH,
    ACTIONS(7374), 1,
      anon_sym_RBRACK,
    ACTIONS(12686), 1,
      sym_file_descriptor,
    ACTIONS(5992), 2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
    ACTIONS(12502), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(12500), 3,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
    STATE(5350), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(12498), 5,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
  [283958] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(5270), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(12622), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(5695), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(5697), 11,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [283991] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12749), 1,
      sym_variable_name,
    STATE(6737), 1,
      sym_subscript,
    STATE(5385), 2,
      sym_variable_assignment,
      aux_sym_variable_assignments_repeat1,
    ACTIONS(11327), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(11329), 10,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [284026] = 12,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(4253), 1,
      anon_sym_PIPE,
    ACTIONS(4348), 1,
      anon_sym_PIPE_AMP,
    ACTIONS(5994), 1,
      anon_sym_LT_LT,
    ACTIONS(5996), 1,
      anon_sym_LT_LT_DASH,
    ACTIONS(7563), 1,
      anon_sym_RBRACK,
    ACTIONS(12686), 1,
      sym_file_descriptor,
    ACTIONS(5992), 2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
    ACTIONS(12502), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(12500), 3,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
    STATE(5350), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(12498), 5,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
  [284073] = 9,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12757), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(12759), 1,
      sym_file_descriptor,
    ACTIONS(11357), 2,
      anon_sym_PIPE,
      anon_sym_LT_LT,
    ACTIONS(12755), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(12753), 3,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
    STATE(5363), 3,
      sym_file_redirect,
      sym_herestring_redirect,
      aux_sym_redirected_statement_repeat2,
    ACTIONS(11363), 4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
    ACTIONS(12751), 5,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
  [284114] = 9,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12770), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(12773), 1,
      sym_file_descriptor,
    ACTIONS(11307), 2,
      anon_sym_PIPE,
      anon_sym_LT_LT,
    ACTIONS(12767), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(12764), 3,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
    STATE(5363), 3,
      sym_file_redirect,
      sym_herestring_redirect,
      aux_sym_redirected_statement_repeat2,
    ACTIONS(11315), 4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
    ACTIONS(12761), 5,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
  [284155] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1348), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1350), 14,
      sym_file_descriptor,
      sym__concat,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      aux_sym_concatenation_token1,
  [284184] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12776), 1,
      sym__special_character,
    STATE(5274), 1,
      aux_sym__literal_repeat1,
    ACTIONS(5393), 5,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(5395), 14,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_RPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [284217] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11455), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12778), 1,
      sym_file_descriptor,
    ACTIONS(4692), 2,
      anon_sym_LT_LT,
      anon_sym_LT_LT_DASH,
    ACTIONS(12346), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    STATE(5370), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(11453), 4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_PIPE_AMP,
    ACTIONS(12344), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
  [284256] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11535), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12317), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(12780), 1,
      sym_file_descriptor,
    ACTIONS(12346), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    STATE(4877), 2,
      sym_file_redirect,
      sym_herestring_redirect,
    ACTIONS(11533), 6,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT_LT,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
    ACTIONS(12344), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
  [284295] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4253), 2,
      anon_sym_PIPE,
      anon_sym_PIPE_AMP,
    ACTIONS(4272), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    STATE(5366), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(4270), 14,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [284328] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4272), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    STATE(5366), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(4270), 16,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [284359] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11431), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12788), 1,
      sym_file_descriptor,
    ACTIONS(11920), 2,
      anon_sym_LT_LT,
      anon_sym_LT_LT_DASH,
    ACTIONS(12785), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    STATE(5370), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(11420), 4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_PIPE_AMP,
    ACTIONS(12782), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
  [284398] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(5322), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(12435), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(5721), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(5723), 11,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [284431] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11414), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12317), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(12780), 1,
      sym_file_descriptor,
    ACTIONS(12346), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    STATE(4868), 2,
      sym_file_redirect,
      sym_herestring_redirect,
    ACTIONS(11412), 6,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT_LT,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
    ACTIONS(12344), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
  [284470] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(5270), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(12622), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(5721), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(5723), 11,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [284503] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(5374), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(12791), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(1251), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1253), 11,
      sym_file_descriptor,
      sym_variable_name,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [284536] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11506), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12317), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(12780), 1,
      sym_file_descriptor,
    ACTIONS(12346), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    STATE(5017), 2,
      sym_file_redirect,
      sym_herestring_redirect,
    ACTIONS(11504), 6,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT_LT,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
    ACTIONS(12344), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
  [284575] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12668), 1,
      sym__special_character,
    STATE(5388), 1,
      aux_sym__literal_repeat1,
    ACTIONS(5721), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(5723), 12,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [284608] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(5322), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(12435), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(5695), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(5697), 11,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [284641] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(5271), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(12622), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(2094), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(2096), 11,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [284674] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(5328), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(12435), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(2094), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(2096), 11,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [284707] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1324), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1326), 14,
      sym_file_descriptor,
      sym__concat,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      aux_sym_concatenation_token1,
  [284736] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5942), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12778), 1,
      sym_file_descriptor,
    ACTIONS(4253), 2,
      anon_sym_PIPE,
      anon_sym_PIPE_AMP,
    ACTIONS(4692), 2,
      anon_sym_LT_LT,
      anon_sym_LT_LT_DASH,
    ACTIONS(5940), 2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
    ACTIONS(12346), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    STATE(5366), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(12344), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
  [284777] = 10,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(11420), 1,
      anon_sym_PIPE,
    ACTIONS(12797), 1,
      anon_sym_LT_LT,
    ACTIONS(12806), 1,
      anon_sym_LT_LT_DASH,
    ACTIONS(12809), 1,
      sym_file_descriptor,
    ACTIONS(12803), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(12800), 3,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
    STATE(5382), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(11431), 4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
    ACTIONS(12794), 5,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
  [284820] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5944), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12778), 1,
      sym_file_descriptor,
    ACTIONS(4253), 2,
      anon_sym_PIPE,
      anon_sym_PIPE_AMP,
    ACTIONS(4692), 2,
      anon_sym_LT_LT,
      anon_sym_LT_LT_DASH,
    ACTIONS(5940), 2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
    ACTIONS(12346), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    STATE(5366), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(12344), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
  [284861] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12670), 6,
      anon_sym_LPAREN,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(12672), 15,
      sym_test_operator,
      sym_extglob_pattern,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [284890] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12812), 1,
      sym_variable_name,
    STATE(6737), 1,
      sym_subscript,
    STATE(5385), 2,
      sym_variable_assignment,
      aux_sym_variable_assignments_repeat1,
    ACTIONS(11288), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(11290), 10,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [284925] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12734), 6,
      anon_sym_LPAREN,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(12736), 15,
      sym_test_operator,
      sym_extglob_pattern,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [284954] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1251), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1253), 14,
      sym_file_descriptor,
      sym__concat,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      aux_sym_concatenation_token1,
  [284983] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12815), 1,
      sym__special_character,
    STATE(5388), 1,
      aux_sym__literal_repeat1,
    ACTIONS(1362), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1364), 12,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [285016] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1296), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1298), 14,
      sym_file_descriptor,
      sym__concat,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      aux_sym_concatenation_token1,
  [285045] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12582), 1,
      aux_sym_concatenation_token1,
    ACTIONS(12818), 1,
      sym__concat,
    STATE(5374), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1286), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1288), 11,
      sym_file_descriptor,
      sym_variable_name,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [285080] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12820), 6,
      anon_sym_LPAREN,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(12822), 15,
      sym_test_operator,
      sym_extglob_pattern,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [285109] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11651), 1,
      aux_sym_concatenation_token1,
    ACTIONS(12824), 1,
      sym__concat,
    STATE(4268), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1288), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1286), 16,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [285144] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11651), 1,
      aux_sym_concatenation_token1,
    ACTIONS(12826), 1,
      sym__concat,
    STATE(4268), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1267), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1265), 16,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [285179] = 12,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(4253), 1,
      anon_sym_PIPE,
    ACTIONS(4348), 1,
      anon_sym_PIPE_AMP,
    ACTIONS(5994), 1,
      anon_sym_LT_LT,
    ACTIONS(5996), 1,
      anon_sym_LT_LT_DASH,
    ACTIONS(7543), 1,
      anon_sym_RBRACK,
    ACTIONS(12686), 1,
      sym_file_descriptor,
    ACTIONS(5992), 2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
    ACTIONS(12502), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(12500), 3,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
    STATE(5350), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(12498), 5,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
  [285226] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12734), 6,
      anon_sym_LPAREN,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(12736), 15,
      sym_test_operator,
      sym_extglob_pattern,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [285255] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12670), 6,
      anon_sym_LPAREN,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(12672), 15,
      sym_test_operator,
      sym_extglob_pattern,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [285284] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12828), 6,
      anon_sym_LPAREN,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(12830), 15,
      sym_test_operator,
      sym_extglob_pattern,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [285313] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11651), 1,
      aux_sym_concatenation_token1,
    ACTIONS(11653), 1,
      sym__concat,
    STATE(5392), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(2096), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(2094), 16,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [285348] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1328), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1330), 14,
      sym_file_descriptor,
      sym__concat,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      aux_sym_concatenation_token1,
  [285377] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1292), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1294), 14,
      sym_file_descriptor,
      sym__concat,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      aux_sym_concatenation_token1,
  [285406] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12670), 6,
      anon_sym_LPAREN,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(12672), 15,
      sym_test_operator,
      sym_extglob_pattern,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [285435] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1332), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1334), 14,
      sym_file_descriptor,
      sym__concat,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      aux_sym_concatenation_token1,
  [285464] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1300), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1302), 14,
      sym_file_descriptor,
      sym__concat,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      aux_sym_concatenation_token1,
  [285493] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1332), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1334), 13,
      sym_file_descriptor,
      sym__concat,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      aux_sym_concatenation_token1,
  [285521] = 9,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12473), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(12838), 1,
      sym_file_descriptor,
    ACTIONS(11504), 2,
      anon_sym_PIPE,
      anon_sym_LT_LT,
    ACTIONS(12836), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    STATE(5695), 2,
      sym_file_redirect,
      sym_herestring_redirect,
    ACTIONS(12834), 3,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
    ACTIONS(11506), 4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
    ACTIONS(12832), 5,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
  [285561] = 11,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(4253), 1,
      anon_sym_PIPE,
    ACTIONS(4348), 1,
      anon_sym_PIPE_AMP,
    ACTIONS(5994), 1,
      anon_sym_LT_LT,
    ACTIONS(5996), 1,
      anon_sym_LT_LT_DASH,
    ACTIONS(12686), 1,
      sym_file_descriptor,
    ACTIONS(5992), 2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
    ACTIONS(12502), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(12500), 3,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
    STATE(5350), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(12498), 5,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
  [285605] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12688), 1,
      aux_sym_concatenation_token1,
    ACTIONS(12840), 1,
      sym__concat,
    STATE(5411), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1286), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1288), 10,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [285639] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12842), 1,
      sym__special_character,
    STATE(5479), 1,
      aux_sym__literal_repeat1,
    ACTIONS(5067), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(5069), 11,
      sym_file_descriptor,
      sym_variable_name,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [285671] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1348), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1350), 13,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [285699] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1261), 8,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      sym__special_character,
    ACTIONS(1263), 12,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [285727] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(5411), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(12844), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(1251), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1253), 10,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [285759] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(5407), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(12688), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(4370), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(4372), 10,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [285791] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1324), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1326), 13,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [285819] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1292), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1294), 13,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [285847] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(5407), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(12688), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(4552), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(4554), 10,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [285879] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1261), 5,
      anon_sym_DOLLAR,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LPAREN,
      sym_word,
    ACTIONS(1263), 15,
      sym_test_operator,
      sym__brace_start,
      anon_sym_LPAREN_LPAREN,
      anon_sym_RPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
  [285907] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(5471), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(12688), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(5695), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(5697), 10,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [285939] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(5407), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(12688), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(2072), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(2074), 10,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [285971] = 4,
    ACTIONS(71), 1,
      sym_comment,
    STATE(5436), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(4270), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(4272), 10,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [286001] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(5471), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(12688), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(4467), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(4469), 10,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [286033] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(4253), 1,
      anon_sym_PIPE,
    ACTIONS(4348), 1,
      anon_sym_PIPE_AMP,
    STATE(5436), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(4270), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(4272), 9,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [286067] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1316), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1318), 13,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [286095] = 9,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12473), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(12838), 1,
      sym_file_descriptor,
    ACTIONS(11533), 2,
      anon_sym_PIPE,
      anon_sym_LT_LT,
    ACTIONS(12836), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    STATE(5672), 2,
      sym_file_redirect,
      sym_herestring_redirect,
    ACTIONS(12834), 3,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
    ACTIONS(11535), 4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
    ACTIONS(12832), 5,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
  [286135] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12847), 1,
      sym__special_character,
    STATE(5426), 1,
      aux_sym__literal_repeat1,
    ACTIONS(5721), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(5723), 11,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [286167] = 11,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7789), 1,
      anon_sym_DQUOTE,
    ACTIONS(7797), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(12648), 1,
      anon_sym_LPAREN,
    ACTIONS(12650), 1,
      anon_sym_DOLLAR,
    ACTIONS(12654), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(12656), 1,
      anon_sym_BQUOTE,
    ACTIONS(12658), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(7807), 2,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
    ACTIONS(12849), 5,
      sym_variable_name,
      sym__expansion_word,
      sym_raw_string,
      sym_ansi_c_string,
      sym_word,
    STATE(6487), 6,
      sym_string,
      sym_array,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      sym_process_substitution,
  [286211] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12851), 1,
      sym__special_character,
    STATE(5426), 1,
      aux_sym__literal_repeat1,
    ACTIONS(1362), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1364), 11,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [286243] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(5407), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(12688), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(2094), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(2096), 10,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [286275] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1251), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1253), 13,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [286303] = 17,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(7793), 1,
      aux_sym_number_token1,
    ACTIONS(7795), 1,
      aux_sym_number_token2,
    ACTIONS(12543), 1,
      anon_sym_LPAREN,
    ACTIONS(12547), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(12551), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(12555), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(12557), 1,
      anon_sym_BQUOTE,
    ACTIONS(12559), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(12854), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12856), 1,
      anon_sym_RBRACE3,
    ACTIONS(12858), 1,
      aux_sym__simple_variable_name_token1,
    STATE(6259), 1,
      sym__expansion_max_length_binary_expression,
    STATE(6470), 1,
      sym__expansion_max_length_expression,
    ACTIONS(12541), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6258), 2,
      sym_number,
      sym_expansion,
    STATE(7022), 3,
      sym_parenthesized_expression,
      sym_arithmetic_expansion,
      sym_command_substitution,
  [286359] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1340), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1342), 13,
      sym_file_descriptor,
      sym__concat,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      aux_sym_concatenation_token1,
  [286387] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1296), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1298), 13,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [286415] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1308), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1310), 13,
      sym_file_descriptor,
      sym__concat,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      aux_sym_concatenation_token1,
  [286443] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1304), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1306), 13,
      sym_file_descriptor,
      sym__concat,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      aux_sym_concatenation_token1,
  [286471] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1308), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1310), 13,
      sym_file_descriptor,
      sym__concat,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      aux_sym_concatenation_token1,
  [286499] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1328), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1330), 13,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [286527] = 10,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(11453), 1,
      anon_sym_PIPE,
    ACTIONS(12860), 1,
      anon_sym_LT_LT,
    ACTIONS(12862), 1,
      anon_sym_LT_LT_DASH,
    ACTIONS(12864), 1,
      sym_file_descriptor,
    ACTIONS(12836), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(11455), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_AMP,
    ACTIONS(12834), 3,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
    STATE(5465), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(12832), 5,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
  [286569] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12866), 1,
      sym__special_character,
    STATE(5469), 1,
      aux_sym__literal_repeat1,
    ACTIONS(5721), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(5723), 11,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [286601] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1324), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1326), 13,
      sym_file_descriptor,
      sym__concat,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      aux_sym_concatenation_token1,
  [286629] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1352), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1354), 13,
      sym_file_descriptor,
      sym__concat,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      aux_sym_concatenation_token1,
  [286657] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1251), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1253), 13,
      sym_file_descriptor,
      sym__concat,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      aux_sym_concatenation_token1,
  [286685] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1292), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1294), 13,
      sym_file_descriptor,
      sym__concat,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [286713] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1356), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1358), 13,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [286741] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1344), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1346), 13,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [286769] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1348), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1350), 13,
      sym_file_descriptor,
      sym__concat,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      aux_sym_concatenation_token1,
  [286797] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1312), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1314), 13,
      sym_file_descriptor,
      sym__concat,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      aux_sym_concatenation_token1,
  [286825] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1300), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1302), 13,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [286853] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12847), 1,
      sym__special_character,
    STATE(5426), 1,
      aux_sym__literal_repeat1,
    ACTIONS(4564), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(4566), 11,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [286885] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1332), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1334), 13,
      sym_file_descriptor,
      sym__concat,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [286913] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1336), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1338), 13,
      sym_file_descriptor,
      sym__concat,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      aux_sym_concatenation_token1,
  [286941] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1300), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1302), 13,
      sym_file_descriptor,
      sym__concat,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [286969] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1296), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1298), 13,
      sym_file_descriptor,
      sym__concat,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [286997] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1320), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1322), 13,
      sym_file_descriptor,
      sym__concat,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      aux_sym_concatenation_token1,
  [287025] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1296), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1298), 13,
      sym_file_descriptor,
      sym__concat,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      aux_sym_concatenation_token1,
  [287053] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1308), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1310), 13,
      sym_file_descriptor,
      sym__concat,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [287081] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1304), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1306), 13,
      sym_file_descriptor,
      sym__concat,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [287109] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12325), 3,
      sym_test_operator,
      sym__brace_start,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12323), 17,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      sym__special_character,
      anon_sym_DQUOTE,
      sym_raw_string,
      sym_ansi_c_string,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      anon_sym_LT_LPAREN,
      anon_sym_GT_LPAREN,
      sym_word,
  [287137] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1300), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1302), 13,
      sym_file_descriptor,
      sym__concat,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      aux_sym_concatenation_token1,
  [287165] = 17,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(7793), 1,
      aux_sym_number_token1,
    ACTIONS(7795), 1,
      aux_sym_number_token2,
    ACTIONS(12543), 1,
      anon_sym_LPAREN,
    ACTIONS(12547), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(12551), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(12555), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(12557), 1,
      anon_sym_BQUOTE,
    ACTIONS(12559), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(12868), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12870), 1,
      anon_sym_RBRACE3,
    ACTIONS(12872), 1,
      aux_sym__simple_variable_name_token1,
    STATE(6301), 1,
      sym__expansion_max_length_binary_expression,
    STATE(6470), 1,
      sym__expansion_max_length_expression,
    ACTIONS(12541), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6279), 2,
      sym_number,
      sym_expansion,
    STATE(6825), 3,
      sym_parenthesized_expression,
      sym_arithmetic_expansion,
      sym_command_substitution,
  [287221] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1308), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1310), 13,
      sym_file_descriptor,
      sym__concat,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [287249] = 20,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(7793), 1,
      aux_sym_number_token1,
    ACTIONS(7795), 1,
      aux_sym_number_token2,
    ACTIONS(12543), 1,
      anon_sym_LPAREN,
    ACTIONS(12547), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(12551), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(12555), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(12557), 1,
      anon_sym_BQUOTE,
    ACTIONS(12559), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(12874), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12876), 1,
      anon_sym_RBRACE3,
    ACTIONS(12878), 1,
      aux_sym__simple_variable_name_token1,
    STATE(6261), 1,
      sym__expansion_max_length_binary_expression,
    STATE(6287), 1,
      sym_number,
    STATE(6299), 1,
      sym_expansion,
    STATE(6470), 1,
      sym__expansion_max_length_expression,
    STATE(7251), 1,
      sym_parenthesized_expression,
    STATE(7302), 1,
      sym_arithmetic_expansion,
    STATE(7477), 1,
      sym_command_substitution,
    ACTIONS(12541), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
  [287311] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1251), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1253), 13,
      sym_file_descriptor,
      sym__concat,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [287339] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12866), 1,
      sym__special_character,
    STATE(5469), 1,
      aux_sym__literal_repeat1,
    ACTIONS(4467), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(4469), 11,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [287371] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1316), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1318), 13,
      sym_file_descriptor,
      sym__concat,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [287399] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1352), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1354), 13,
      sym_file_descriptor,
      sym__concat,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [287427] = 10,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(11420), 1,
      anon_sym_PIPE,
    ACTIONS(12883), 1,
      anon_sym_LT_LT,
    ACTIONS(12892), 1,
      anon_sym_LT_LT_DASH,
    ACTIONS(12895), 1,
      sym_file_descriptor,
    ACTIONS(12889), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(11431), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_AMP,
    ACTIONS(12886), 3,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
    STATE(5465), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(12880), 5,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
  [287469] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12866), 1,
      sym__special_character,
    STATE(5469), 1,
      aux_sym__literal_repeat1,
    ACTIONS(4564), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(4566), 11,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [287501] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1292), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1294), 13,
      sym_file_descriptor,
      sym__concat,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      aux_sym_concatenation_token1,
  [287529] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1320), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1322), 13,
      sym_file_descriptor,
      sym__concat,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [287557] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12898), 1,
      sym__special_character,
    STATE(5469), 1,
      aux_sym__literal_repeat1,
    ACTIONS(1362), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1364), 11,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [287589] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1328), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1330), 13,
      sym_file_descriptor,
      sym__concat,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [287617] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12688), 1,
      aux_sym_concatenation_token1,
    ACTIONS(12901), 1,
      sym__concat,
    STATE(5411), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1265), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1267), 10,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [287651] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1320), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1322), 13,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [287679] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1312), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1314), 13,
      sym_file_descriptor,
      sym__concat,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [287707] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(5471), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(12688), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(4564), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(4566), 10,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [287739] = 9,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12473), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(12838), 1,
      sym_file_descriptor,
    ACTIONS(11412), 2,
      anon_sym_PIPE,
      anon_sym_LT_LT,
    ACTIONS(12836), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    STATE(5681), 2,
      sym_file_redirect,
      sym_herestring_redirect,
    ACTIONS(12834), 3,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
    ACTIONS(11414), 4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_AMP,
      anon_sym_LT_LT_DASH,
    ACTIONS(12832), 5,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
  [287779] = 5,
    ACTIONS(71), 1,
      sym_comment,
    STATE(5471), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(12688), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(5721), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(5723), 10,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [287811] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1344), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1346), 13,
      sym_file_descriptor,
      sym__concat,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [287839] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1336), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1338), 13,
      sym_file_descriptor,
      sym__concat,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [287867] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12903), 1,
      sym__special_character,
    STATE(5479), 1,
      aux_sym__literal_repeat1,
    ACTIONS(1362), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1364), 11,
      sym_file_descriptor,
      sym_variable_name,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [287899] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1324), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1326), 13,
      sym_file_descriptor,
      sym__concat,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [287927] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1332), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1334), 13,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [287955] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1340), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1342), 13,
      sym_file_descriptor,
      sym__concat,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [287983] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1356), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1358), 13,
      sym_file_descriptor,
      sym__concat,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [288011] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1308), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1310), 13,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [288039] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1344), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1346), 13,
      sym_file_descriptor,
      sym__concat,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      aux_sym_concatenation_token1,
  [288067] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1356), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1358), 13,
      sym_file_descriptor,
      sym__concat,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      aux_sym_concatenation_token1,
  [288095] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1348), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1350), 13,
      sym_file_descriptor,
      sym__concat,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [288123] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1336), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1338), 13,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [288151] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1328), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1330), 13,
      sym_file_descriptor,
      sym__concat,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      aux_sym_concatenation_token1,
  [288179] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1340), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1342), 13,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [288207] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1312), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1314), 13,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [288235] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1316), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1318), 13,
      sym_file_descriptor,
      sym__concat,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      aux_sym_concatenation_token1,
  [288263] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1308), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1310), 13,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [288291] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12866), 1,
      sym__special_character,
    STATE(5469), 1,
      aux_sym__literal_repeat1,
    ACTIONS(5695), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(5697), 11,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [288323] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1304), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1306), 13,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [288351] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12847), 1,
      sym__special_character,
    STATE(5426), 1,
      aux_sym__literal_repeat1,
    ACTIONS(5695), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(5697), 11,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [288383] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1352), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1354), 13,
      sym_file_descriptor,
      sym__concat,
      sym_variable_name,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [288411] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12847), 1,
      sym__special_character,
    STATE(5426), 1,
      aux_sym__literal_repeat1,
    ACTIONS(4467), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(4469), 11,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [288443] = 5,
    ACTIONS(3), 1,
      sym_comment,
    STATE(5528), 1,
      aux_sym_pipeline_repeat1,
    ACTIONS(12068), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12906), 2,
      anon_sym_PIPE,
      anon_sym_PIPE_AMP,
    ACTIONS(12064), 14,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [288474] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12908), 1,
      sym__special_character,
    STATE(5520), 1,
      aux_sym__literal_repeat1,
    ACTIONS(4564), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(4566), 10,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [288505] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1356), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1358), 12,
      sym_file_descriptor,
      sym__concat,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [288532] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1344), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1346), 12,
      sym_file_descriptor,
      sym__concat,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [288559] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12912), 1,
      anon_sym_DQUOTE,
    ACTIONS(12916), 1,
      sym_variable_name,
    STATE(6206), 1,
      sym_string,
    ACTIONS(12914), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1239), 5,
      anon_sym_in,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12910), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [288594] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3727), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(3751), 1,
      sym_file_descriptor,
    ACTIONS(12918), 1,
      aux_sym_heredoc_redirect_token1,
    STATE(6998), 1,
      sym__heredoc_expression,
    ACTIONS(3717), 2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
    ACTIONS(3723), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    STATE(5609), 3,
      sym_file_redirect,
      sym_herestring_redirect,
      aux_sym_redirected_statement_repeat2,
    ACTIONS(3721), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
  [288633] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3727), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(3751), 1,
      sym_file_descriptor,
    ACTIONS(12920), 1,
      aux_sym_heredoc_redirect_token1,
    STATE(7030), 1,
      sym__heredoc_expression,
    ACTIONS(3717), 2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
    ACTIONS(3723), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    STATE(5609), 3,
      sym_file_redirect,
      sym_herestring_redirect,
      aux_sym_redirected_statement_repeat2,
    ACTIONS(3721), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
  [288672] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1261), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1263), 12,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
      sym__special_character,
  [288699] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1320), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1322), 12,
      sym_file_descriptor,
      sym__concat,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [288726] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1292), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1294), 12,
      sym_file_descriptor,
      sym__concat,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [288753] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1316), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1318), 12,
      sym_file_descriptor,
      sym__concat,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [288780] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1340), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1342), 12,
      sym_file_descriptor,
      sym__concat,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [288807] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1312), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1314), 12,
      sym_file_descriptor,
      sym__concat,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [288834] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1261), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1263), 12,
      sym_file_descriptor,
      sym_variable_name,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      sym__special_character,
  [288861] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1352), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1354), 12,
      sym_file_descriptor,
      sym__concat,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [288888] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12912), 1,
      anon_sym_DQUOTE,
    ACTIONS(12916), 1,
      sym_variable_name,
    STATE(6206), 1,
      sym_string,
    ACTIONS(12914), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1227), 5,
      anon_sym_in,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12910), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [288923] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3727), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(3751), 1,
      sym_file_descriptor,
    ACTIONS(12922), 1,
      aux_sym_heredoc_redirect_token1,
    STATE(6931), 1,
      sym__heredoc_expression,
    ACTIONS(3717), 2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
    ACTIONS(3723), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    STATE(5609), 3,
      sym_file_redirect,
      sym_herestring_redirect,
      aux_sym_redirected_statement_repeat2,
    ACTIONS(3721), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
  [288962] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3727), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(3751), 1,
      sym_file_descriptor,
    ACTIONS(12924), 1,
      aux_sym_heredoc_redirect_token1,
    STATE(6928), 1,
      sym__heredoc_expression,
    ACTIONS(3717), 2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
    ACTIONS(3723), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    STATE(5609), 3,
      sym_file_redirect,
      sym_herestring_redirect,
      aux_sym_redirected_statement_repeat2,
    ACTIONS(3721), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
  [289001] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3727), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(3751), 1,
      sym_file_descriptor,
    ACTIONS(12926), 1,
      aux_sym_heredoc_redirect_token1,
    STATE(6853), 1,
      sym__heredoc_expression,
    ACTIONS(3717), 2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
    ACTIONS(3723), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    STATE(5609), 3,
      sym_file_redirect,
      sym_herestring_redirect,
      aux_sym_redirected_statement_repeat2,
    ACTIONS(3721), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
  [289040] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12928), 1,
      aux_sym_concatenation_token1,
    ACTIONS(12930), 1,
      sym__concat,
    STATE(5577), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1263), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1261), 14,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      sym__special_character,
  [289073] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12908), 1,
      sym__special_character,
    STATE(5520), 1,
      aux_sym__literal_repeat1,
    ACTIONS(5721), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(5723), 10,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [289104] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12932), 1,
      sym__special_character,
    STATE(5520), 1,
      aux_sym__literal_repeat1,
    ACTIONS(1362), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1364), 10,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [289135] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1328), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1330), 12,
      sym_file_descriptor,
      sym__concat,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [289162] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1308), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1310), 12,
      sym_file_descriptor,
      sym__concat,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [289189] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1336), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1338), 12,
      sym_file_descriptor,
      sym__concat,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [289216] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1304), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1306), 12,
      sym_file_descriptor,
      sym__concat,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [289243] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1332), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1334), 12,
      sym_file_descriptor,
      sym__concat,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [289270] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(4552), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(4554), 12,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [289297] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1308), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1310), 12,
      sym_file_descriptor,
      sym__concat,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [289324] = 5,
    ACTIONS(3), 1,
      sym_comment,
    STATE(5528), 1,
      aux_sym_pipeline_repeat1,
    ACTIONS(11748), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12935), 2,
      anon_sym_PIPE,
      anon_sym_PIPE_AMP,
    ACTIONS(11743), 14,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [289355] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1261), 8,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      sym__special_character,
    ACTIONS(1263), 11,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [289382] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(2094), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(2096), 12,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [289409] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3727), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(3751), 1,
      sym_file_descriptor,
    ACTIONS(12938), 1,
      aux_sym_heredoc_redirect_token1,
    STATE(7028), 1,
      sym__heredoc_expression,
    ACTIONS(3717), 2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
    ACTIONS(3723), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    STATE(5609), 3,
      sym_file_redirect,
      sym_herestring_redirect,
      aux_sym_redirected_statement_repeat2,
    ACTIONS(3721), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
  [289448] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12908), 1,
      sym__special_character,
    STATE(5520), 1,
      aux_sym__literal_repeat1,
    ACTIONS(5695), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(5697), 10,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [289479] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5942), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12778), 1,
      sym_file_descriptor,
    ACTIONS(4692), 2,
      anon_sym_LT_LT,
      anon_sym_LT_LT_DASH,
    ACTIONS(5940), 2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
    ACTIONS(12346), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    STATE(5366), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(12344), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
  [289516] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1300), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1302), 12,
      sym_file_descriptor,
      sym__concat,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [289543] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5944), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(12778), 1,
      sym_file_descriptor,
    ACTIONS(4692), 2,
      anon_sym_LT_LT,
      anon_sym_LT_LT_DASH,
    ACTIONS(5940), 2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
    ACTIONS(12346), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    STATE(5366), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(12344), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
  [289580] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1296), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1298), 12,
      sym_file_descriptor,
      sym__concat,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [289607] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1348), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1350), 12,
      sym_file_descriptor,
      sym__concat,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [289634] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3727), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(3751), 1,
      sym_file_descriptor,
    ACTIONS(12940), 1,
      aux_sym_heredoc_redirect_token1,
    STATE(6868), 1,
      sym__heredoc_expression,
    ACTIONS(3717), 2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
    ACTIONS(3723), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    STATE(5609), 3,
      sym_file_redirect,
      sym_herestring_redirect,
      aux_sym_redirected_statement_repeat2,
    ACTIONS(3721), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
  [289673] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(2072), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(2074), 12,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [289700] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3727), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(3751), 1,
      sym_file_descriptor,
    ACTIONS(12942), 1,
      aux_sym_heredoc_redirect_token1,
    STATE(6866), 1,
      sym__heredoc_expression,
    ACTIONS(3717), 2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
    ACTIONS(3723), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    STATE(5609), 3,
      sym_file_redirect,
      sym_herestring_redirect,
      aux_sym_redirected_statement_repeat2,
    ACTIONS(3721), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
  [289739] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1251), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1253), 12,
      sym_file_descriptor,
      sym__concat,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [289766] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1324), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1326), 12,
      sym_file_descriptor,
      sym__concat,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      aux_sym_concatenation_token1,
  [289793] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(4370), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(4372), 12,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [289820] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12908), 1,
      sym__special_character,
    STATE(5520), 1,
      aux_sym__literal_repeat1,
    ACTIONS(4467), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(4469), 10,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [289851] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3727), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(3751), 1,
      sym_file_descriptor,
    ACTIONS(12944), 1,
      aux_sym_heredoc_redirect_token1,
    STATE(6835), 1,
      sym__heredoc_expression,
    ACTIONS(3717), 2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
    ACTIONS(3723), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    STATE(5609), 3,
      sym_file_redirect,
      sym_herestring_redirect,
      aux_sym_redirected_statement_repeat2,
    ACTIONS(3721), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
  [289890] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3727), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(3751), 1,
      sym_file_descriptor,
    ACTIONS(12946), 1,
      aux_sym_heredoc_redirect_token1,
    STATE(6836), 1,
      sym__heredoc_expression,
    ACTIONS(3717), 2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
    ACTIONS(3723), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    STATE(5609), 3,
      sym_file_redirect,
      sym_herestring_redirect,
      aux_sym_redirected_statement_repeat2,
    ACTIONS(3721), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
  [289929] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(11741), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(11739), 11,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [289955] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12928), 1,
      aux_sym_concatenation_token1,
    ACTIONS(12930), 1,
      sym__concat,
    STATE(5577), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4469), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4467), 13,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
  [289987] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(4552), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(4554), 11,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [290013] = 9,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(4692), 1,
      anon_sym_LT_LT,
    ACTIONS(11993), 1,
      sym_file_descriptor,
    ACTIONS(12954), 1,
      anon_sym_LT_LT_DASH,
    ACTIONS(12948), 2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
    ACTIONS(12952), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(12950), 3,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
    STATE(4746), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(11540), 5,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
  [290051] = 9,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12860), 1,
      anon_sym_LT_LT,
    ACTIONS(12862), 1,
      anon_sym_LT_LT_DASH,
    ACTIONS(12864), 1,
      sym_file_descriptor,
    ACTIONS(5992), 2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
    ACTIONS(12836), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(12834), 3,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
    STATE(5436), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(12832), 5,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
  [290089] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(5766), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(5768), 11,
      sym_file_descriptor,
      sym_variable_name,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [290115] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12928), 1,
      aux_sym_concatenation_token1,
    ACTIONS(12930), 1,
      sym__concat,
    STATE(5580), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(2074), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(2072), 13,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
  [290147] = 9,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(4692), 1,
      anon_sym_LT_LT,
    ACTIONS(12221), 1,
      sym_file_descriptor,
    ACTIONS(12954), 1,
      anon_sym_LT_LT_DASH,
    ACTIONS(12956), 2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
    ACTIONS(12960), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(12958), 3,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
    STATE(4986), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(11554), 5,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
  [290185] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(4370), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(4372), 11,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [290211] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12101), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(12103), 11,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [290237] = 9,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(4692), 1,
      anon_sym_LT_LT,
    ACTIONS(11834), 1,
      sym_file_descriptor,
    ACTIONS(12954), 1,
      anon_sym_LT_LT_DASH,
    ACTIONS(12962), 2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
    ACTIONS(12966), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(12964), 3,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
    STATE(4650), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(11479), 5,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
  [290275] = 9,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(4692), 1,
      anon_sym_LT_LT,
    ACTIONS(11856), 1,
      sym_file_descriptor,
    ACTIONS(12954), 1,
      anon_sym_LT_LT_DASH,
    ACTIONS(12968), 2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
    ACTIONS(12972), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(12970), 3,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
    STATE(4697), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(11485), 5,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
  [290313] = 9,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(4692), 1,
      anon_sym_LT_LT,
    ACTIONS(12778), 1,
      sym_file_descriptor,
    ACTIONS(12954), 1,
      anon_sym_LT_LT_DASH,
    ACTIONS(12974), 2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
    ACTIONS(12978), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(12976), 3,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
    STATE(5366), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(12344), 5,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
  [290351] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(11780), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(11782), 11,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [290377] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(11745), 1,
      anon_sym_PIPE,
    ACTIONS(12980), 1,
      anon_sym_PIPE_AMP,
    STATE(5602), 1,
      aux_sym_pipeline_repeat1,
    ACTIONS(11743), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(11748), 9,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [290409] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12928), 1,
      aux_sym_concatenation_token1,
    ACTIONS(12930), 1,
      sym__concat,
    STATE(5577), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5723), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5721), 13,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
  [290441] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(2072), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(2074), 11,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [290467] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(11752), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(11750), 11,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [290493] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12201), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(12199), 11,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [290519] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(5373), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(5375), 11,
      sym_file_descriptor,
      sym_variable_name,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [290545] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(11735), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(11733), 11,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [290571] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12189), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(12187), 11,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [290597] = 9,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(5994), 1,
      anon_sym_LT_LT,
    ACTIONS(5996), 1,
      anon_sym_LT_LT_DASH,
    ACTIONS(12686), 1,
      sym_file_descriptor,
    ACTIONS(5992), 2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
    ACTIONS(12502), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(12500), 3,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
    STATE(5350), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(12498), 5,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
  [290635] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12001), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(12003), 11,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [290661] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12145), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(12147), 11,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [290687] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12141), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(12143), 11,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [290713] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12137), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(12139), 11,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [290739] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12211), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(12209), 11,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [290765] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(2094), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(2096), 11,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [290791] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12105), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(12107), 11,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [290817] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12928), 1,
      aux_sym_concatenation_token1,
    ACTIONS(12983), 1,
      sym__concat,
    STATE(5583), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1267), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1265), 13,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
  [290849] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12117), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(12119), 11,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [290875] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12117), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(12119), 11,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [290901] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12928), 1,
      aux_sym_concatenation_token1,
    ACTIONS(12985), 1,
      sym__concat,
    STATE(5583), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1288), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1286), 13,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
  [290933] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12161), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(12163), 11,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [290959] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(11844), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(11842), 11,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [290985] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12987), 1,
      aux_sym_concatenation_token1,
    ACTIONS(12990), 1,
      sym__concat,
    STATE(5583), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1253), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1251), 13,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
  [291017] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12928), 1,
      aux_sym_concatenation_token1,
    ACTIONS(12930), 1,
      sym__concat,
    STATE(5577), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(5697), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5695), 13,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
  [291049] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12217), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(12219), 11,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [291075] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12223), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(12225), 11,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [291101] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(4370), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(4372), 11,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [291127] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(11725), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(11727), 11,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [291153] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12201), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(12199), 11,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [291179] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12223), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(12225), 11,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [291205] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(11731), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(11729), 11,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [291231] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(11792), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(11790), 11,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [291257] = 9,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(4382), 1,
      anon_sym_LT_LT,
    ACTIONS(11562), 1,
      sym_file_descriptor,
    ACTIONS(12999), 1,
      anon_sym_LT_LT_DASH,
    ACTIONS(12993), 2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
    ACTIONS(12997), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(12995), 3,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
    STATE(4455), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(11333), 5,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
  [291295] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(11828), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(11830), 11,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [291321] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12249), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(12251), 11,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [291347] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12257), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(12259), 11,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [291373] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12257), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(12259), 11,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [291399] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12159), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(12157), 11,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [291425] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12159), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(12157), 11,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [291451] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12105), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(12107), 11,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_RBRACK,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [291477] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(2072), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(2074), 11,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [291503] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(13001), 1,
      anon_sym_PIPE,
    ACTIONS(13003), 1,
      anon_sym_PIPE_AMP,
    STATE(5611), 1,
      aux_sym_pipeline_repeat1,
    ACTIONS(12064), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(12068), 9,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [291535] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(11958), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(11956), 11,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [291561] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(2094), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(2096), 11,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [291587] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12928), 1,
      aux_sym_concatenation_token1,
    ACTIONS(12930), 1,
      sym__concat,
    STATE(5580), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4554), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4552), 13,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
  [291619] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12928), 1,
      aux_sym_concatenation_token1,
    ACTIONS(12930), 1,
      sym__concat,
    STATE(5577), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4566), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4564), 13,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
  [291651] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1261), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(1263), 11,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      sym__special_character,
  [291677] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12928), 1,
      aux_sym_concatenation_token1,
    ACTIONS(12930), 1,
      sym__concat,
    STATE(5580), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(2096), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(2094), 13,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
  [291709] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11315), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(13011), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(13014), 1,
      sym_file_descriptor,
    ACTIONS(11307), 2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
    ACTIONS(13008), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    STATE(5609), 3,
      sym_file_redirect,
      sym_herestring_redirect,
      aux_sym_redirected_statement_repeat2,
    ACTIONS(13005), 8,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
  [291745] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(5770), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(5772), 11,
      sym_file_descriptor,
      sym_variable_name,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [291771] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(13017), 1,
      anon_sym_PIPE,
    ACTIONS(13020), 1,
      anon_sym_PIPE_AMP,
    STATE(5611), 1,
      aux_sym_pipeline_repeat1,
    ACTIONS(11743), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(11748), 9,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [291803] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(4552), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(4554), 11,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [291829] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12034), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(12036), 11,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [291855] = 9,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(4255), 1,
      anon_sym_LT_LT,
    ACTIONS(11457), 1,
      sym_file_descriptor,
    ACTIONS(13029), 1,
      anon_sym_LT_LT_DASH,
    ACTIONS(13023), 2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
    ACTIONS(13027), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(13025), 3,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
    STATE(4283), 3,
      sym_file_redirect,
      sym_heredoc_redirect,
      aux_sym_redirected_statement_repeat1,
    ACTIONS(11280), 5,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
  [291893] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(11929), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(11931), 11,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
      anon_sym_LT_LT_LT,
  [291919] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12928), 1,
      aux_sym_concatenation_token1,
    ACTIONS(12930), 1,
      sym__concat,
    STATE(5580), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(4372), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4370), 13,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
  [291951] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12201), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(12199), 10,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [291976] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1322), 3,
      sym_file_descriptor,
      sym__concat,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1320), 14,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      aux_sym_concatenation_token1,
  [292001] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(2094), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(2096), 10,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [292026] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(11958), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(11956), 10,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [292051] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12034), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(12036), 10,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [292076] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1298), 3,
      sym_file_descriptor,
      sym__concat,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1296), 14,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      aux_sym_concatenation_token1,
  [292101] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(5930), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(5932), 10,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [292126] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(5930), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(5932), 10,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [292151] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1318), 3,
      sym_file_descriptor,
      sym__concat,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1316), 14,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      aux_sym_concatenation_token1,
  [292176] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12159), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(12157), 10,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [292201] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12159), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(12157), 10,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [292226] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1330), 3,
      sym_file_descriptor,
      sym__concat,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1328), 14,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      aux_sym_concatenation_token1,
  [292251] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(4370), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(4372), 10,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [292276] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12257), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(12259), 10,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [292301] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12257), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(12259), 10,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [292326] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13031), 1,
      sym__special_character,
    STATE(5657), 1,
      aux_sym__literal_repeat1,
    ACTIONS(5697), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5695), 13,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
  [292355] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12249), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(12251), 10,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [292380] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12207), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(12205), 10,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [292405] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12223), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(12225), 10,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [292430] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12223), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(12225), 10,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [292455] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(11828), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(11830), 10,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [292480] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12201), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(12199), 10,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [292505] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12217), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(12219), 10,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [292530] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12161), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(12163), 10,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [292555] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12117), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(12119), 10,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [292580] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12117), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(12119), 10,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [292605] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13031), 1,
      sym__special_character,
    STATE(5657), 1,
      aux_sym__literal_repeat1,
    ACTIONS(4566), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4564), 13,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
  [292634] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(11792), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(11790), 10,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [292659] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12105), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(12107), 10,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [292684] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12105), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(12107), 10,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [292709] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(11731), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(11729), 10,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [292734] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(11725), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(11727), 10,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [292759] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12211), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(12209), 10,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [292784] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12137), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(12139), 10,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [292809] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13031), 1,
      sym__special_character,
    STATE(5657), 1,
      aux_sym__literal_repeat1,
    ACTIONS(4469), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4467), 13,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
  [292838] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(11929), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(11931), 10,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [292863] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12145), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(12147), 10,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [292888] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1358), 3,
      sym_file_descriptor,
      sym__concat,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1356), 14,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      aux_sym_concatenation_token1,
  [292913] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1346), 3,
      sym_file_descriptor,
      sym__concat,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1344), 14,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      aux_sym_concatenation_token1,
  [292938] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12267), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(12265), 10,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [292963] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13033), 1,
      sym__special_character,
    STATE(5657), 1,
      aux_sym__literal_repeat1,
    ACTIONS(1364), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1362), 13,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
  [292992] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12189), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(12187), 10,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [293017] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12141), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(12143), 10,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [293042] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1326), 3,
      sym_file_descriptor,
      sym__concat,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1324), 14,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      aux_sym_concatenation_token1,
  [293067] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12197), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(12195), 10,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [293092] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1338), 3,
      sym_file_descriptor,
      sym__concat,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1336), 14,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      aux_sym_concatenation_token1,
  [293117] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1342), 3,
      sym_file_descriptor,
      sym__concat,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1340), 14,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      aux_sym_concatenation_token1,
  [293142] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1314), 3,
      sym_file_descriptor,
      sym__concat,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1312), 14,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      aux_sym_concatenation_token1,
  [293167] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12153), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(12155), 10,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [293192] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1354), 3,
      sym_file_descriptor,
      sym__concat,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1352), 14,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      aux_sym_concatenation_token1,
  [293217] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12149), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(12151), 10,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [293242] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1310), 3,
      sym_file_descriptor,
      sym__concat,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1308), 14,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      aux_sym_concatenation_token1,
  [293267] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1306), 3,
      sym_file_descriptor,
      sym__concat,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1304), 14,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      aux_sym_concatenation_token1,
  [293292] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1310), 3,
      sym_file_descriptor,
      sym__concat,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1308), 14,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      aux_sym_concatenation_token1,
  [293317] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(4552), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(4554), 10,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [293342] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12135), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(12133), 10,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [293367] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12213), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(12215), 10,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [293392] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13031), 1,
      sym__special_character,
    STATE(5657), 1,
      aux_sym__literal_repeat1,
    ACTIONS(5723), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(5721), 13,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
  [293421] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12213), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(12215), 10,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [293446] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(11741), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(11739), 10,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [293471] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12131), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(12129), 10,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [293496] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12127), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(12125), 10,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [293521] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12101), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(12103), 10,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [293546] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1253), 3,
      sym_file_descriptor,
      sym__concat,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1251), 14,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      aux_sym_concatenation_token1,
  [293571] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12111), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(12109), 10,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [293596] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12271), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(12269), 10,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [293621] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1294), 3,
      sym_file_descriptor,
      sym__concat,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1292), 14,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      aux_sym_concatenation_token1,
  [293646] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12193), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(12191), 10,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [293671] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12123), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(12121), 10,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [293696] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12115), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(12113), 10,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [293721] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12229), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(12227), 10,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [293746] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1334), 3,
      sym_file_descriptor,
      sym__concat,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1332), 14,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      aux_sym_concatenation_token1,
  [293771] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12229), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(12227), 10,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [293796] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12255), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(12253), 10,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [293821] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12296), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(12298), 10,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [293846] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(2072), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(2074), 10,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [293871] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1302), 3,
      sym_file_descriptor,
      sym__concat,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1300), 14,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      aux_sym_concatenation_token1,
  [293896] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12294), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(12292), 10,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [293921] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12261), 7,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_LT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
    ACTIONS(12263), 10,
      sym_file_descriptor,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE_AMP,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_DASH,
  [293946] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1350), 3,
      sym_file_descriptor,
      sym__concat,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1348), 14,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      aux_sym_concatenation_token1,
  [293971] = 10,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11764), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(11766), 1,
      anon_sym_BQUOTE,
    ACTIONS(11768), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13040), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(13042), 1,
      sym_variable_name,
    STATE(3533), 1,
      sym_subscript,
    STATE(6448), 1,
      sym_command_substitution,
    ACTIONS(13038), 3,
      anon_sym_DOLLAR,
      anon_sym_0,
      anon_sym__,
    ACTIONS(13036), 6,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_POUND,
      anon_sym_AT2,
  [294009] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(6492), 1,
      anon_sym_DQUOTE,
    ACTIONS(13048), 1,
      sym_variable_name,
    STATE(6513), 1,
      sym_string,
    ACTIONS(1239), 2,
      anon_sym_PIPE,
      anon_sym_RPAREN,
    ACTIONS(13046), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13044), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [294041] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(6492), 1,
      anon_sym_DQUOTE,
    ACTIONS(13048), 1,
      sym_variable_name,
    STATE(6513), 1,
      sym_string,
    ACTIONS(1227), 2,
      anon_sym_PIPE,
      anon_sym_RPAREN,
    ACTIONS(13046), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13044), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [294073] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(11764), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(11766), 1,
      anon_sym_BQUOTE,
    ACTIONS(11768), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13054), 1,
      aux_sym__simple_variable_name_token1,
    ACTIONS(13056), 1,
      sym_variable_name,
    STATE(6448), 2,
      sym_subscript,
      sym_command_substitution,
    ACTIONS(13052), 3,
      anon_sym_DOLLAR,
      anon_sym_0,
      anon_sym__,
    ACTIONS(13050), 6,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_POUND,
      anon_sym_AT2,
  [294109] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1263), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1261), 14,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
      sym__special_character,
  [294133] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13062), 1,
      anon_sym_DOLLAR,
    ACTIONS(13064), 1,
      anon_sym_DQUOTE,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    STATE(5768), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [294174] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13076), 1,
      anon_sym_DOLLAR,
    ACTIONS(13078), 1,
      anon_sym_DQUOTE,
    STATE(5768), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [294215] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13080), 1,
      anon_sym_DOLLAR,
    ACTIONS(13082), 1,
      anon_sym_DQUOTE,
    STATE(5768), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [294256] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13084), 1,
      anon_sym_DOLLAR,
    ACTIONS(13086), 1,
      anon_sym_DQUOTE,
    STATE(5711), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [294297] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13088), 1,
      anon_sym_DOLLAR,
    ACTIONS(13090), 1,
      anon_sym_DQUOTE,
    STATE(5752), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [294338] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13092), 1,
      anon_sym_DOLLAR,
    ACTIONS(13094), 1,
      anon_sym_DQUOTE,
    STATE(5768), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [294379] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13096), 1,
      anon_sym_DOLLAR,
    ACTIONS(13098), 1,
      anon_sym_DQUOTE,
    STATE(5713), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [294420] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13100), 1,
      anon_sym_DOLLAR,
    ACTIONS(13102), 1,
      anon_sym_DQUOTE,
    STATE(5768), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [294461] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13104), 1,
      anon_sym_DOLLAR,
    ACTIONS(13106), 1,
      anon_sym_DQUOTE,
    STATE(5768), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [294502] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13108), 1,
      anon_sym_DOLLAR,
    ACTIONS(13110), 1,
      anon_sym_DQUOTE,
    STATE(5768), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [294543] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13112), 1,
      anon_sym_DOLLAR,
    ACTIONS(13114), 1,
      anon_sym_DQUOTE,
    STATE(5768), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [294584] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13116), 1,
      anon_sym_DOLLAR,
    ACTIONS(13118), 1,
      anon_sym_DQUOTE,
    STATE(5768), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [294625] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13120), 1,
      anon_sym_DOLLAR,
    ACTIONS(13122), 1,
      anon_sym_DQUOTE,
    STATE(5768), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [294666] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13124), 1,
      anon_sym_DOLLAR,
    ACTIONS(13126), 1,
      anon_sym_DQUOTE,
    STATE(5768), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [294707] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13128), 1,
      anon_sym_DOLLAR,
    ACTIONS(13130), 1,
      anon_sym_DQUOTE,
    STATE(5719), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [294748] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13132), 1,
      anon_sym_DOLLAR,
    ACTIONS(13134), 1,
      anon_sym_DQUOTE,
    STATE(5768), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [294789] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13136), 1,
      anon_sym_DOLLAR,
    ACTIONS(13138), 1,
      anon_sym_DQUOTE,
    STATE(5801), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [294830] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13140), 1,
      anon_sym_DOLLAR,
    ACTIONS(13142), 1,
      anon_sym_DQUOTE,
    STATE(5768), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [294871] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13144), 1,
      anon_sym_DOLLAR,
    ACTIONS(13146), 1,
      anon_sym_DQUOTE,
    STATE(5712), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [294912] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13148), 1,
      anon_sym_DOLLAR,
    ACTIONS(13150), 1,
      anon_sym_DQUOTE,
    STATE(5766), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [294953] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13152), 1,
      anon_sym_DOLLAR,
    ACTIONS(13154), 1,
      anon_sym_DQUOTE,
    STATE(5704), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [294994] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13156), 1,
      anon_sym_DOLLAR,
    ACTIONS(13158), 1,
      anon_sym_DQUOTE,
    STATE(5710), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [295035] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13160), 1,
      anon_sym_DOLLAR,
    ACTIONS(13162), 1,
      anon_sym_DQUOTE,
    STATE(5707), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [295076] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13164), 1,
      anon_sym_DOLLAR,
    ACTIONS(13166), 1,
      anon_sym_DQUOTE,
    STATE(5731), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [295117] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13168), 1,
      anon_sym_DOLLAR,
    ACTIONS(13170), 1,
      anon_sym_DQUOTE,
    STATE(5768), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [295158] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13172), 1,
      anon_sym_DOLLAR,
    ACTIONS(13174), 1,
      anon_sym_DQUOTE,
    STATE(5709), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [295199] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13176), 1,
      anon_sym_DOLLAR,
    ACTIONS(13178), 1,
      anon_sym_DQUOTE,
    STATE(5768), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [295240] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13180), 1,
      anon_sym_DOLLAR,
    ACTIONS(13182), 1,
      anon_sym_DQUOTE,
    STATE(5728), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [295281] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13184), 1,
      anon_sym_DOLLAR,
    ACTIONS(13186), 1,
      anon_sym_DQUOTE,
    STATE(5740), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [295322] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13188), 1,
      anon_sym_DOLLAR,
    ACTIONS(13190), 1,
      anon_sym_DQUOTE,
    STATE(5768), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [295363] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13192), 1,
      anon_sym_DOLLAR,
    ACTIONS(13194), 1,
      anon_sym_DQUOTE,
    STATE(5768), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [295404] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13196), 1,
      anon_sym_DOLLAR,
    ACTIONS(13198), 1,
      anon_sym_DQUOTE,
    STATE(5726), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [295445] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13200), 1,
      anon_sym_DOLLAR,
    ACTIONS(13202), 1,
      anon_sym_DQUOTE,
    STATE(5702), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [295486] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13204), 1,
      anon_sym_DOLLAR,
    ACTIONS(13206), 1,
      anon_sym_DQUOTE,
    STATE(5848), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [295527] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13208), 1,
      anon_sym_DOLLAR,
    ACTIONS(13210), 1,
      anon_sym_DQUOTE,
    STATE(5732), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [295568] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13212), 1,
      anon_sym_DOLLAR,
    ACTIONS(13214), 1,
      anon_sym_DQUOTE,
    STATE(5768), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [295609] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2096), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(2094), 13,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
  [295632] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13216), 1,
      anon_sym_DOLLAR,
    ACTIONS(13218), 1,
      anon_sym_DQUOTE,
    STATE(5768), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [295673] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13220), 1,
      anon_sym_DOLLAR,
    ACTIONS(13222), 1,
      anon_sym_DQUOTE,
    STATE(5768), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [295714] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13224), 1,
      anon_sym_DOLLAR,
    ACTIONS(13226), 1,
      anon_sym_DQUOTE,
    STATE(5739), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [295755] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13228), 1,
      anon_sym_DOLLAR,
    ACTIONS(13230), 1,
      anon_sym_DQUOTE,
    STATE(5768), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [295796] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13232), 1,
      anon_sym_DOLLAR,
    ACTIONS(13234), 1,
      anon_sym_DQUOTE,
    STATE(5749), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [295837] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4372), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4370), 13,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
  [295860] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13236), 1,
      anon_sym_DOLLAR,
    ACTIONS(13238), 1,
      anon_sym_DQUOTE,
    STATE(5768), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [295901] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13240), 1,
      anon_sym_DOLLAR,
    ACTIONS(13242), 1,
      anon_sym_DQUOTE,
    STATE(5737), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [295942] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4554), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(4552), 13,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
  [295965] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13244), 1,
      anon_sym_DOLLAR,
    ACTIONS(13246), 1,
      anon_sym_DQUOTE,
    STATE(5768), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [296006] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13248), 1,
      anon_sym_DOLLAR,
    ACTIONS(13250), 1,
      anon_sym_DQUOTE,
    STATE(5768), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [296047] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13252), 1,
      anon_sym_DOLLAR,
    ACTIONS(13254), 1,
      anon_sym_DQUOTE,
    STATE(5745), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [296088] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13256), 1,
      anon_sym_DOLLAR,
    ACTIONS(13258), 1,
      anon_sym_DQUOTE,
    STATE(5768), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [296129] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13260), 1,
      anon_sym_DOLLAR,
    ACTIONS(13262), 1,
      anon_sym_DQUOTE,
    STATE(5768), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [296170] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13264), 1,
      anon_sym_DOLLAR,
    ACTIONS(13266), 1,
      anon_sym_DQUOTE,
    STATE(5760), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [296211] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13268), 1,
      anon_sym_DOLLAR,
    ACTIONS(13270), 1,
      anon_sym_DQUOTE,
    STATE(5759), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [296252] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13272), 1,
      anon_sym_DOLLAR,
    ACTIONS(13274), 1,
      anon_sym_DQUOTE,
    STATE(5864), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [296293] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13276), 1,
      anon_sym_DOLLAR,
    ACTIONS(13278), 1,
      anon_sym_DQUOTE,
    STATE(5768), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [296334] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13280), 1,
      anon_sym_DOLLAR,
    ACTIONS(13282), 1,
      anon_sym_DQUOTE,
    STATE(5715), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [296375] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13284), 1,
      anon_sym_DOLLAR,
    ACTIONS(13286), 1,
      anon_sym_DQUOTE,
    STATE(5742), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [296416] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13288), 1,
      anon_sym_DOLLAR,
    ACTIONS(13290), 1,
      anon_sym_DQUOTE,
    STATE(5768), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [296457] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13292), 1,
      anon_sym_DOLLAR,
    ACTIONS(13294), 1,
      anon_sym_DQUOTE,
    STATE(5768), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [296498] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13296), 1,
      anon_sym_DOLLAR,
    ACTIONS(13298), 1,
      anon_sym_DQUOTE,
    STATE(5767), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [296539] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13300), 1,
      anon_sym_DOLLAR,
    ACTIONS(13302), 1,
      anon_sym_DQUOTE,
    STATE(5756), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [296580] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13304), 1,
      anon_sym_DOLLAR,
    ACTIONS(13306), 1,
      anon_sym_DQUOTE,
    STATE(5768), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [296621] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13308), 1,
      anon_sym_DOLLAR,
    ACTIONS(13310), 1,
      anon_sym_DQUOTE,
    STATE(5763), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [296662] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13312), 1,
      anon_sym_DOLLAR,
    ACTIONS(13314), 1,
      anon_sym_DQUOTE,
    STATE(5768), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [296703] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13316), 1,
      anon_sym_DOLLAR,
    ACTIONS(13318), 1,
      anon_sym_DQUOTE,
    STATE(5768), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [296744] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13320), 1,
      anon_sym_DOLLAR,
    ACTIONS(13322), 1,
      anon_sym_DQUOTE,
    STATE(5768), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [296785] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13327), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13330), 1,
      anon_sym_DOLLAR,
    ACTIONS(13333), 1,
      anon_sym_DQUOTE,
    ACTIONS(13335), 1,
      sym_string_content,
    ACTIONS(13338), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13341), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13344), 1,
      anon_sym_BQUOTE,
    ACTIONS(13347), 1,
      anon_sym_DOLLAR_BQUOTE,
    STATE(5768), 1,
      aux_sym_string_repeat1,
    ACTIONS(13324), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [296826] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13350), 1,
      anon_sym_DOLLAR,
    ACTIONS(13352), 1,
      anon_sym_DQUOTE,
    STATE(5770), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [296867] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13354), 1,
      anon_sym_DOLLAR,
    ACTIONS(13356), 1,
      anon_sym_DQUOTE,
    STATE(5768), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [296908] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13358), 1,
      anon_sym_DOLLAR,
    ACTIONS(13360), 1,
      anon_sym_DQUOTE,
    STATE(5768), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [296949] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13362), 1,
      anon_sym_DOLLAR,
    ACTIONS(13364), 1,
      anon_sym_DQUOTE,
    STATE(5813), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [296990] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13366), 1,
      anon_sym_DOLLAR,
    ACTIONS(13368), 1,
      anon_sym_DQUOTE,
    STATE(5771), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [297031] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13370), 1,
      anon_sym_DOLLAR,
    ACTIONS(13372), 1,
      anon_sym_DQUOTE,
    STATE(5768), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [297072] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13374), 1,
      anon_sym_DOLLAR,
    ACTIONS(13376), 1,
      anon_sym_DQUOTE,
    STATE(5774), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [297113] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13378), 1,
      anon_sym_DOLLAR,
    ACTIONS(13380), 1,
      anon_sym_DQUOTE,
    STATE(5768), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [297154] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13382), 1,
      anon_sym_DOLLAR,
    ACTIONS(13384), 1,
      anon_sym_DQUOTE,
    STATE(5776), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [297195] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13386), 1,
      anon_sym_DOLLAR,
    ACTIONS(13388), 1,
      anon_sym_DQUOTE,
    STATE(5768), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [297236] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13390), 1,
      anon_sym_DOLLAR,
    ACTIONS(13392), 1,
      anon_sym_DQUOTE,
    STATE(5785), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [297277] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13394), 1,
      anon_sym_DOLLAR,
    ACTIONS(13396), 1,
      anon_sym_DQUOTE,
    STATE(5795), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [297318] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13398), 1,
      anon_sym_DOLLAR,
    ACTIONS(13400), 1,
      anon_sym_DQUOTE,
    STATE(5768), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [297359] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13402), 1,
      anon_sym_DOLLAR,
    ACTIONS(13404), 1,
      anon_sym_DQUOTE,
    STATE(5778), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [297400] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13406), 1,
      anon_sym_DOLLAR,
    ACTIONS(13408), 1,
      anon_sym_DQUOTE,
    STATE(5714), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [297441] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13410), 1,
      anon_sym_DOLLAR,
    ACTIONS(13412), 1,
      anon_sym_DQUOTE,
    STATE(5703), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [297482] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13414), 1,
      anon_sym_DOLLAR,
    ACTIONS(13416), 1,
      anon_sym_DQUOTE,
    STATE(5768), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [297523] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13418), 1,
      anon_sym_DOLLAR,
    ACTIONS(13420), 1,
      anon_sym_DQUOTE,
    STATE(5768), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [297564] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13422), 1,
      anon_sym_DOLLAR,
    ACTIONS(13424), 1,
      anon_sym_DQUOTE,
    STATE(5748), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [297605] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13426), 1,
      anon_sym_DOLLAR,
    ACTIONS(13428), 1,
      anon_sym_DQUOTE,
    STATE(5786), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [297646] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2074), 2,
      sym_file_descriptor,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(2072), 13,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
      anon_sym_GT_PIPE,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
      anon_sym_LT_LT_LT,
  [297669] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13430), 1,
      anon_sym_DOLLAR,
    ACTIONS(13432), 1,
      anon_sym_DQUOTE,
    STATE(5768), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [297710] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13434), 1,
      anon_sym_DOLLAR,
    ACTIONS(13436), 1,
      anon_sym_DQUOTE,
    STATE(5790), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [297751] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13438), 1,
      anon_sym_DOLLAR,
    ACTIONS(13440), 1,
      anon_sym_DQUOTE,
    STATE(5768), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [297792] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13442), 1,
      anon_sym_DOLLAR,
    ACTIONS(13444), 1,
      anon_sym_DQUOTE,
    STATE(5792), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [297833] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13446), 1,
      anon_sym_DOLLAR,
    ACTIONS(13448), 1,
      anon_sym_DQUOTE,
    STATE(5768), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [297874] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13450), 1,
      anon_sym_DOLLAR,
    ACTIONS(13452), 1,
      anon_sym_DQUOTE,
    STATE(5768), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [297915] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13454), 1,
      anon_sym_DOLLAR,
    ACTIONS(13456), 1,
      anon_sym_DQUOTE,
    STATE(5794), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [297956] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13458), 1,
      anon_sym_DOLLAR,
    ACTIONS(13460), 1,
      anon_sym_DQUOTE,
    STATE(5803), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [297997] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13462), 1,
      anon_sym_DOLLAR,
    ACTIONS(13464), 1,
      anon_sym_DQUOTE,
    STATE(5768), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [298038] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13466), 1,
      anon_sym_DOLLAR,
    ACTIONS(13468), 1,
      anon_sym_DQUOTE,
    STATE(5798), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [298079] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13470), 1,
      anon_sym_DOLLAR,
    ACTIONS(13472), 1,
      anon_sym_DQUOTE,
    STATE(5751), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [298120] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13474), 1,
      anon_sym_DOLLAR,
    ACTIONS(13476), 1,
      anon_sym_DQUOTE,
    STATE(5768), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [298161] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13478), 1,
      anon_sym_DOLLAR,
    ACTIONS(13480), 1,
      anon_sym_DQUOTE,
    STATE(5768), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [298202] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13482), 1,
      anon_sym_DOLLAR,
    ACTIONS(13484), 1,
      anon_sym_DQUOTE,
    STATE(5768), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [298243] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13486), 1,
      anon_sym_DOLLAR,
    ACTIONS(13488), 1,
      anon_sym_DQUOTE,
    STATE(5802), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [298284] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13490), 1,
      anon_sym_DOLLAR,
    ACTIONS(13492), 1,
      anon_sym_DQUOTE,
    STATE(5768), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [298325] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13494), 1,
      anon_sym_DOLLAR,
    ACTIONS(13496), 1,
      anon_sym_DQUOTE,
    STATE(5765), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [298366] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13498), 1,
      anon_sym_DOLLAR,
    ACTIONS(13500), 1,
      anon_sym_DQUOTE,
    STATE(5805), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [298407] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13502), 1,
      anon_sym_DOLLAR,
    ACTIONS(13504), 1,
      anon_sym_DQUOTE,
    STATE(5768), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [298448] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13506), 1,
      anon_sym_DOLLAR,
    ACTIONS(13508), 1,
      anon_sym_DQUOTE,
    STATE(5808), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [298489] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13510), 1,
      anon_sym_DOLLAR,
    ACTIONS(13512), 1,
      anon_sym_DQUOTE,
    STATE(5768), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [298530] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13514), 1,
      anon_sym_DOLLAR,
    ACTIONS(13516), 1,
      anon_sym_DQUOTE,
    STATE(5810), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [298571] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13518), 1,
      anon_sym_DOLLAR,
    ACTIONS(13520), 1,
      anon_sym_DQUOTE,
    STATE(5768), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [298612] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13522), 1,
      anon_sym_DOLLAR,
    ACTIONS(13524), 1,
      anon_sym_DQUOTE,
    STATE(5768), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [298653] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13526), 1,
      anon_sym_DOLLAR,
    ACTIONS(13528), 1,
      anon_sym_DQUOTE,
    STATE(5812), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [298694] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13530), 1,
      anon_sym_DOLLAR,
    ACTIONS(13532), 1,
      anon_sym_DQUOTE,
    STATE(5821), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [298735] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13534), 1,
      anon_sym_DOLLAR,
    ACTIONS(13536), 1,
      anon_sym_DQUOTE,
    STATE(5768), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [298776] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13538), 1,
      anon_sym_DOLLAR,
    ACTIONS(13540), 1,
      anon_sym_DQUOTE,
    STATE(5816), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [298817] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13542), 1,
      anon_sym_DOLLAR,
    ACTIONS(13544), 1,
      anon_sym_DQUOTE,
    STATE(5768), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [298858] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13546), 1,
      anon_sym_DOLLAR,
    ACTIONS(13548), 1,
      anon_sym_DQUOTE,
    STATE(5818), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [298899] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13550), 1,
      anon_sym_DOLLAR,
    ACTIONS(13552), 1,
      anon_sym_DQUOTE,
    STATE(5768), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [298940] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13554), 1,
      anon_sym_DOLLAR,
    ACTIONS(13556), 1,
      anon_sym_DQUOTE,
    STATE(5768), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [298981] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13558), 1,
      anon_sym_DOLLAR,
    ACTIONS(13560), 1,
      anon_sym_DQUOTE,
    STATE(5829), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [299022] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13562), 1,
      anon_sym_DOLLAR,
    ACTIONS(13564), 1,
      anon_sym_DQUOTE,
    STATE(5820), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [299063] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13566), 1,
      anon_sym_DOLLAR,
    ACTIONS(13568), 1,
      anon_sym_DQUOTE,
    STATE(5768), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [299104] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13570), 1,
      anon_sym_DOLLAR,
    ACTIONS(13572), 1,
      anon_sym_DQUOTE,
    STATE(5824), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [299145] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13574), 1,
      anon_sym_DOLLAR,
    ACTIONS(13576), 1,
      anon_sym_DQUOTE,
    STATE(5768), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [299186] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13578), 1,
      anon_sym_DOLLAR,
    ACTIONS(13580), 1,
      anon_sym_DQUOTE,
    STATE(5826), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [299227] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13582), 1,
      anon_sym_DOLLAR,
    ACTIONS(13584), 1,
      anon_sym_DQUOTE,
    STATE(5768), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [299268] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13586), 1,
      anon_sym_DOLLAR,
    ACTIONS(13588), 1,
      anon_sym_DQUOTE,
    STATE(5768), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [299309] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13590), 1,
      anon_sym_DOLLAR,
    ACTIONS(13592), 1,
      anon_sym_DQUOTE,
    STATE(5828), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [299350] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13594), 1,
      anon_sym_DOLLAR,
    ACTIONS(13596), 1,
      anon_sym_DQUOTE,
    STATE(5768), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [299391] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13598), 1,
      anon_sym_DOLLAR,
    ACTIONS(13600), 1,
      anon_sym_DQUOTE,
    STATE(5831), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [299432] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13602), 1,
      anon_sym_DOLLAR,
    ACTIONS(13604), 1,
      anon_sym_DQUOTE,
    STATE(5839), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [299473] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13606), 1,
      anon_sym_DOLLAR,
    ACTIONS(13608), 1,
      anon_sym_DQUOTE,
    STATE(5768), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [299514] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13610), 1,
      anon_sym_DOLLAR,
    ACTIONS(13612), 1,
      anon_sym_DQUOTE,
    STATE(5834), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [299555] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13614), 1,
      anon_sym_DOLLAR,
    ACTIONS(13616), 1,
      anon_sym_DQUOTE,
    STATE(5768), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [299596] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13618), 1,
      anon_sym_DOLLAR,
    ACTIONS(13620), 1,
      anon_sym_DQUOTE,
    STATE(5836), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [299637] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13622), 1,
      anon_sym_DOLLAR,
    ACTIONS(13624), 1,
      anon_sym_DQUOTE,
    STATE(5768), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [299678] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13626), 1,
      anon_sym_DOLLAR,
    ACTIONS(13628), 1,
      anon_sym_DQUOTE,
    STATE(5768), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [299719] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13630), 1,
      anon_sym_DOLLAR,
    ACTIONS(13632), 1,
      anon_sym_DQUOTE,
    STATE(5838), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [299760] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13634), 1,
      anon_sym_DOLLAR,
    ACTIONS(13636), 1,
      anon_sym_DQUOTE,
    STATE(5768), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [299801] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13638), 1,
      anon_sym_DOLLAR,
    ACTIONS(13640), 1,
      anon_sym_DQUOTE,
    STATE(5841), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [299842] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13642), 1,
      anon_sym_DOLLAR,
    ACTIONS(13644), 1,
      anon_sym_DQUOTE,
    STATE(5768), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [299883] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13646), 1,
      anon_sym_DOLLAR,
    ACTIONS(13648), 1,
      anon_sym_DQUOTE,
    STATE(5843), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [299924] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13650), 1,
      anon_sym_DOLLAR,
    ACTIONS(13652), 1,
      anon_sym_DQUOTE,
    STATE(5768), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [299965] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13654), 1,
      anon_sym_DOLLAR,
    ACTIONS(13656), 1,
      anon_sym_DQUOTE,
    STATE(5845), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [300006] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13658), 1,
      anon_sym_DOLLAR,
    ACTIONS(13660), 1,
      anon_sym_DQUOTE,
    STATE(5768), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [300047] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13662), 1,
      anon_sym_DOLLAR,
    ACTIONS(13664), 1,
      anon_sym_DQUOTE,
    STATE(5768), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [300088] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13666), 1,
      anon_sym_DOLLAR,
    ACTIONS(13668), 1,
      anon_sym_DQUOTE,
    STATE(5781), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [300129] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13670), 1,
      anon_sym_DOLLAR,
    ACTIONS(13672), 1,
      anon_sym_DQUOTE,
    STATE(5847), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [300170] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13674), 1,
      anon_sym_DOLLAR,
    ACTIONS(13676), 1,
      anon_sym_DQUOTE,
    STATE(5857), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [300211] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13678), 1,
      anon_sym_DOLLAR,
    ACTIONS(13680), 1,
      anon_sym_DQUOTE,
    STATE(5717), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [300252] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13682), 1,
      anon_sym_DOLLAR,
    ACTIONS(13684), 1,
      anon_sym_DQUOTE,
    STATE(5768), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [300293] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13686), 1,
      anon_sym_DOLLAR,
    ACTIONS(13688), 1,
      anon_sym_DQUOTE,
    STATE(5853), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [300334] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13690), 1,
      anon_sym_DOLLAR,
    ACTIONS(13692), 1,
      anon_sym_DQUOTE,
    STATE(5768), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [300375] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13694), 1,
      anon_sym_DOLLAR,
    ACTIONS(13696), 1,
      anon_sym_DQUOTE,
    STATE(5855), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [300416] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13698), 1,
      anon_sym_DOLLAR,
    ACTIONS(13700), 1,
      anon_sym_DQUOTE,
    STATE(5768), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [300457] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13702), 1,
      anon_sym_DOLLAR,
    ACTIONS(13704), 1,
      anon_sym_DQUOTE,
    STATE(5860), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [300498] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13706), 1,
      anon_sym_DOLLAR,
    ACTIONS(13708), 1,
      anon_sym_DQUOTE,
    STATE(5768), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [300539] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13710), 1,
      anon_sym_DOLLAR,
    ACTIONS(13712), 1,
      anon_sym_DQUOTE,
    STATE(5768), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [300580] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13714), 1,
      anon_sym_DOLLAR,
    ACTIONS(13716), 1,
      anon_sym_DQUOTE,
    STATE(5859), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [300621] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13718), 1,
      anon_sym_DOLLAR,
    ACTIONS(13720), 1,
      anon_sym_DQUOTE,
    STATE(5768), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [300662] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13722), 1,
      anon_sym_DOLLAR,
    ACTIONS(13724), 1,
      anon_sym_DQUOTE,
    STATE(5862), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [300703] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13060), 1,
      anon_sym_DOLLAR_LBRACK,
    ACTIONS(13066), 1,
      sym_string_content,
    ACTIONS(13068), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(13070), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(13072), 1,
      anon_sym_BQUOTE,
    ACTIONS(13074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(13726), 1,
      anon_sym_DOLLAR,
    ACTIONS(13728), 1,
      anon_sym_DQUOTE,
    STATE(5768), 1,
      aux_sym_string_repeat1,
    ACTIONS(13058), 2,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
    STATE(6127), 4,
      sym_arithmetic_expansion,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
  [300744] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3775), 1,
      anon_sym_DQUOTE,
    ACTIONS(3779), 1,
      sym_variable_name,
    STATE(2093), 1,
      sym_string,
    ACTIONS(3777), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(3773), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [300772] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13170), 1,
      anon_sym_DQUOTE,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [300800] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13118), 1,
      anon_sym_DQUOTE,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [300828] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13246), 1,
      anon_sym_DQUOTE,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [300856] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13142), 1,
      anon_sym_DQUOTE,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [300884] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8556), 1,
      anon_sym_DQUOTE,
    ACTIONS(8560), 1,
      sym_variable_name,
    STATE(5364), 1,
      sym_string,
    ACTIONS(8558), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(8554), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [300912] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13738), 1,
      anon_sym_DQUOTE,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [300940] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13740), 1,
      anon_sym_DQUOTE,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [300968] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13742), 1,
      anon_sym_DQUOTE,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [300996] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13744), 1,
      anon_sym_DQUOTE,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [301024] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(7673), 1,
      anon_sym_DQUOTE,
    ACTIONS(7677), 1,
      sym_variable_name,
    STATE(4393), 1,
      sym_string,
    ACTIONS(7675), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(7671), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [301052] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4857), 1,
      anon_sym_DQUOTE,
    ACTIONS(4861), 1,
      sym_variable_name,
    STATE(2638), 1,
      sym_string,
    ACTIONS(4859), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(4855), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [301080] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13746), 1,
      anon_sym_DQUOTE,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [301108] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13114), 1,
      anon_sym_DQUOTE,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [301136] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5061), 1,
      anon_sym_DQUOTE,
    ACTIONS(5065), 1,
      sym_variable_name,
    STATE(2712), 1,
      sym_string,
    ACTIONS(5063), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(5059), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [301164] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13134), 1,
      anon_sym_DQUOTE,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [301192] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13708), 1,
      anon_sym_DQUOTE,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [301220] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13190), 1,
      anon_sym_DQUOTE,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [301248] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4558), 1,
      anon_sym_DQUOTE,
    ACTIONS(4562), 1,
      sym_variable_name,
    STATE(2447), 1,
      sym_string,
    ACTIONS(4560), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(4556), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [301276] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1591), 1,
      anon_sym_DQUOTE,
    ACTIONS(2155), 1,
      sym_variable_name,
    STATE(1282), 1,
      sym_string,
    ACTIONS(2153), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(2151), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [301304] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13748), 1,
      anon_sym_DQUOTE,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [301332] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13318), 1,
      anon_sym_DQUOTE,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [301360] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13750), 1,
      anon_sym_DQUOTE,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [301388] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13752), 1,
      anon_sym_DQUOTE,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [301416] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13754), 1,
      anon_sym_DQUOTE,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [301444] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13314), 1,
      anon_sym_DQUOTE,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [301472] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13222), 1,
      anon_sym_DQUOTE,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [301500] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8714), 1,
      anon_sym_DQUOTE,
    ACTIONS(8718), 1,
      sym_variable_name,
    STATE(5444), 1,
      sym_string,
    ACTIONS(8716), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(8712), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [301528] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13756), 1,
      anon_sym_DQUOTE,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [301556] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5152), 1,
      anon_sym_DQUOTE,
    ACTIONS(5156), 1,
      sym_variable_name,
    STATE(2566), 1,
      sym_string,
    ACTIONS(5154), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(5150), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [301584] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8868), 1,
      anon_sym_DQUOTE,
    ACTIONS(8872), 1,
      sym_variable_name,
    STATE(5409), 1,
      sym_string,
    ACTIONS(8870), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(8866), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [301612] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13758), 1,
      anon_sym_DQUOTE,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [301640] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13712), 1,
      anon_sym_DQUOTE,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [301668] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13250), 1,
      anon_sym_DQUOTE,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [301696] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13258), 1,
      anon_sym_DQUOTE,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [301724] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5086), 1,
      anon_sym_DQUOTE,
    ACTIONS(5090), 1,
      sym_variable_name,
    STATE(2442), 1,
      sym_string,
    ACTIONS(5088), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(5084), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [301752] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13760), 1,
      anon_sym_DQUOTE,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [301780] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13178), 1,
      anon_sym_DQUOTE,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [301808] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8623), 1,
      anon_sym_DQUOTE,
    ACTIONS(8627), 1,
      sym_variable_name,
    STATE(5204), 1,
      sym_string,
    ACTIONS(8625), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(8621), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [301836] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13762), 1,
      anon_sym_DQUOTE,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [301864] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13764), 1,
      anon_sym_DQUOTE,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [301892] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13290), 1,
      anon_sym_DQUOTE,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [301920] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1231), 1,
      anon_sym_DQUOTE,
    ACTIONS(1237), 1,
      sym_variable_name,
    STATE(489), 1,
      sym_string,
    ACTIONS(1233), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1229), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [301948] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2428), 1,
      anon_sym_DQUOTE,
    ACTIONS(2432), 1,
      sym_variable_name,
    STATE(1204), 1,
      sym_string,
    ACTIONS(2430), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(2426), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [301976] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13766), 1,
      anon_sym_DQUOTE,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [302004] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13768), 1,
      anon_sym_DQUOTE,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [302032] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13770), 1,
      anon_sym_DQUOTE,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [302060] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13322), 1,
      anon_sym_DQUOTE,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [302088] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13728), 1,
      anon_sym_DQUOTE,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [302116] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13194), 1,
      anon_sym_DQUOTE,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [302144] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1245), 1,
      anon_sym_DQUOTE,
    ACTIONS(1249), 1,
      sym_variable_name,
    STATE(503), 1,
      sym_string,
    ACTIONS(1247), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1243), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [302172] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13772), 1,
      anon_sym_DQUOTE,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [302200] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13294), 1,
      anon_sym_DQUOTE,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [302228] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13774), 1,
      anon_sym_DQUOTE,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [302256] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(8694), 1,
      anon_sym_DQUOTE,
    ACTIONS(8698), 1,
      sym_variable_name,
    STATE(5487), 1,
      sym_string,
    ACTIONS(8696), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(8692), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [302284] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13778), 1,
      anon_sym_DQUOTE,
    ACTIONS(13782), 1,
      sym_variable_name,
    STATE(6761), 1,
      sym_string,
    ACTIONS(13780), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13776), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [302312] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13784), 1,
      anon_sym_DQUOTE,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [302340] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13786), 1,
      anon_sym_DQUOTE,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [302368] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4397), 1,
      anon_sym_DQUOTE,
    ACTIONS(4401), 1,
      sym_variable_name,
    STATE(2412), 1,
      sym_string,
    ACTIONS(4399), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(4395), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [302396] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1765), 1,
      anon_sym_DQUOTE,
    ACTIONS(2959), 1,
      sym_variable_name,
    STATE(1213), 1,
      sym_string,
    ACTIONS(2957), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(2955), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [302424] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(12912), 1,
      anon_sym_DQUOTE,
    ACTIONS(12916), 1,
      sym_variable_name,
    STATE(6206), 1,
      sym_string,
    ACTIONS(12914), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(12910), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [302452] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3701), 1,
      anon_sym_DQUOTE,
    ACTIONS(3705), 1,
      sym_variable_name,
    STATE(2025), 1,
      sym_string,
    ACTIONS(3703), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(3699), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [302480] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13356), 1,
      anon_sym_DQUOTE,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [302508] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13788), 1,
      anon_sym_DQUOTE,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [302536] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13720), 1,
      anon_sym_DQUOTE,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [302564] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3423), 1,
      anon_sym_DQUOTE,
    ACTIONS(3427), 1,
      sym_variable_name,
    STATE(1686), 1,
      sym_string,
    ACTIONS(3425), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(3421), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [302592] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13790), 1,
      anon_sym_DQUOTE,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [302620] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13218), 1,
      anon_sym_DQUOTE,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [302648] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13792), 1,
      anon_sym_DQUOTE,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [302676] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13230), 1,
      anon_sym_DQUOTE,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [302704] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4663), 1,
      anon_sym_DQUOTE,
    ACTIONS(4667), 1,
      sym_variable_name,
    STATE(2716), 1,
      sym_string,
    ACTIONS(4665), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(4661), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [302732] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4797), 1,
      anon_sym_DQUOTE,
    ACTIONS(4801), 1,
      sym_variable_name,
    STATE(2585), 1,
      sym_string,
    ACTIONS(4799), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(4795), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [302760] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13794), 1,
      anon_sym_DQUOTE,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [302788] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13664), 1,
      anon_sym_DQUOTE,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [302816] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13360), 1,
      anon_sym_DQUOTE,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [302844] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2307), 1,
      anon_sym_DQUOTE,
    ACTIONS(2947), 1,
      sym_variable_name,
    STATE(1432), 1,
      sym_string,
    ACTIONS(2945), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(2943), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [302872] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13796), 1,
      anon_sym_DQUOTE,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [302900] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13798), 1,
      anon_sym_DQUOTE,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [302928] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13372), 1,
      anon_sym_DQUOTE,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [302956] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13238), 1,
      anon_sym_DQUOTE,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [302984] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13800), 1,
      anon_sym_DQUOTE,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [303012] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13692), 1,
      anon_sym_DQUOTE,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [303040] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13700), 1,
      anon_sym_DQUOTE,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [303068] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13802), 1,
      anon_sym_DQUOTE,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [303096] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1819), 1,
      anon_sym_DQUOTE,
    ACTIONS(1823), 1,
      sym_variable_name,
    STATE(1125), 1,
      sym_string,
    ACTIONS(1821), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1817), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [303124] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4867), 1,
      anon_sym_DQUOTE,
    ACTIONS(4871), 1,
      sym_variable_name,
    STATE(2525), 1,
      sym_string,
    ACTIONS(4869), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(4865), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [303152] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13380), 1,
      anon_sym_DQUOTE,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [303180] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5614), 1,
      anon_sym_DQUOTE,
    ACTIONS(5719), 1,
      sym_variable_name,
    STATE(3393), 1,
      sym_string,
    ACTIONS(5717), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(5715), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [303208] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13804), 1,
      anon_sym_DQUOTE,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [303236] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13806), 1,
      anon_sym_DQUOTE,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [303264] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13808), 1,
      anon_sym_DQUOTE,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [303292] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4084), 1,
      anon_sym_DQUOTE,
    ACTIONS(7613), 1,
      sym_variable_name,
    STATE(4606), 1,
      sym_string,
    ACTIONS(7611), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(7609), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [303320] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13388), 1,
      anon_sym_DQUOTE,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [303348] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1418), 1,
      anon_sym_DQUOTE,
    ACTIONS(1807), 1,
      sym_variable_name,
    STATE(922), 1,
      sym_string,
    ACTIONS(1805), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1803), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [303376] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13400), 1,
      anon_sym_DQUOTE,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [303404] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13810), 1,
      anon_sym_DQUOTE,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [303432] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13064), 1,
      anon_sym_DQUOTE,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [303460] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13262), 1,
      anon_sym_DQUOTE,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [303488] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13278), 1,
      anon_sym_DQUOTE,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [303516] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13684), 1,
      anon_sym_DQUOTE,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [303544] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13416), 1,
      anon_sym_DQUOTE,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [303572] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13078), 1,
      anon_sym_DQUOTE,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [303600] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13812), 1,
      anon_sym_DQUOTE,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [303628] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1811), 1,
      anon_sym_DQUOTE,
    ACTIONS(1815), 1,
      sym_variable_name,
    STATE(1061), 1,
      sym_string,
    ACTIONS(1813), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1809), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [303656] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(6492), 1,
      anon_sym_DQUOTE,
    ACTIONS(13048), 1,
      sym_variable_name,
    STATE(6513), 1,
      sym_string,
    ACTIONS(13046), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13044), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [303684] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13814), 1,
      anon_sym_DQUOTE,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [303712] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(6210), 1,
      anon_sym_DQUOTE,
    ACTIONS(6214), 1,
      sym_variable_name,
    STATE(3525), 1,
      sym_string,
    ACTIONS(6212), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(6208), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [303740] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13420), 1,
      anon_sym_DQUOTE,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [303768] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13816), 1,
      anon_sym_DQUOTE,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [303796] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(262), 1,
      anon_sym_DQUOTE,
    ACTIONS(5162), 1,
      sym_variable_name,
    STATE(2734), 1,
      sym_string,
    ACTIONS(5160), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(5158), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [303824] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13818), 1,
      anon_sym_DQUOTE,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [303852] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2608), 1,
      anon_sym_DQUOTE,
    ACTIONS(2778), 1,
      sym_variable_name,
    STATE(1525), 1,
      sym_string,
    ACTIONS(2776), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(2774), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [303880] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13820), 1,
      anon_sym_DQUOTE,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [303908] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13822), 1,
      anon_sym_DQUOTE,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [303936] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(9150), 1,
      anon_sym_DQUOTE,
    ACTIONS(9154), 1,
      sym_variable_name,
    STATE(5537), 1,
      sym_string,
    ACTIONS(9152), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(9148), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [303964] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13432), 1,
      anon_sym_DQUOTE,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [303992] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3791), 1,
      anon_sym_DQUOTE,
    ACTIONS(7444), 1,
      sym_variable_name,
    STATE(4389), 1,
      sym_string,
    ACTIONS(7442), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(7440), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [304020] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13660), 1,
      anon_sym_DQUOTE,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [304048] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13824), 1,
      anon_sym_DQUOTE,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [304076] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13306), 1,
      anon_sym_DQUOTE,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [304104] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13826), 1,
      anon_sym_DQUOTE,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [304132] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13440), 1,
      anon_sym_DQUOTE,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [304160] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13452), 1,
      anon_sym_DQUOTE,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [304188] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13828), 1,
      anon_sym_DQUOTE,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [304216] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13830), 1,
      anon_sym_DQUOTE,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [304244] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13448), 1,
      anon_sym_DQUOTE,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [304272] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13832), 1,
      anon_sym_DQUOTE,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [304300] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5874), 1,
      anon_sym_DQUOTE,
    ACTIONS(11244), 1,
      sym_variable_name,
    STATE(5696), 1,
      sym_string,
    ACTIONS(11242), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(11240), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [304328] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13214), 1,
      anon_sym_DQUOTE,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [304356] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3371), 1,
      anon_sym_DQUOTE,
    ACTIONS(7153), 1,
      sym_variable_name,
    STATE(4512), 1,
      sym_string,
    ACTIONS(7151), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(7149), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [304384] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13834), 1,
      anon_sym_DQUOTE,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [304412] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13836), 1,
      anon_sym_DQUOTE,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [304440] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1555), 1,
      anon_sym_DQUOTE,
    ACTIONS(2070), 1,
      sym_variable_name,
    STATE(1176), 1,
      sym_string,
    ACTIONS(2068), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(2066), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [304468] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13464), 1,
      anon_sym_DQUOTE,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [304496] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13484), 1,
      anon_sym_DQUOTE,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [304524] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13652), 1,
      anon_sym_DQUOTE,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [304552] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13838), 1,
      anon_sym_DQUOTE,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [304580] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13840), 1,
      anon_sym_DQUOTE,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [304608] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13842), 1,
      anon_sym_DQUOTE,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [304636] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13844), 1,
      anon_sym_DQUOTE,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [304664] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3735), 1,
      anon_sym_DQUOTE,
    ACTIONS(7757), 1,
      sym_variable_name,
    STATE(4808), 1,
      sym_string,
    ACTIONS(7755), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(7753), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [304692] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13480), 1,
      anon_sym_DQUOTE,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [304720] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3439), 1,
      anon_sym_DQUOTE,
    ACTIONS(3653), 1,
      sym_variable_name,
    STATE(1878), 1,
      sym_string,
    ACTIONS(3651), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(3649), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [304748] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13846), 1,
      anon_sym_DQUOTE,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [304776] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13848), 1,
      anon_sym_DQUOTE,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [304804] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13492), 1,
      anon_sym_DQUOTE,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [304832] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13850), 1,
      anon_sym_DQUOTE,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [304860] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2339), 1,
      anon_sym_DQUOTE,
    ACTIONS(2492), 1,
      sym_variable_name,
    STATE(1463), 1,
      sym_string,
    ACTIONS(2490), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(2488), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [304888] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13852), 1,
      anon_sym_DQUOTE,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [304916] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4115), 1,
      anon_sym_DQUOTE,
    ACTIONS(4119), 1,
      sym_variable_name,
    STATE(2334), 1,
      sym_string,
    ACTIONS(4117), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(4113), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [304944] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13504), 1,
      anon_sym_DQUOTE,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [304972] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13854), 1,
      anon_sym_DQUOTE,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [305000] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13856), 1,
      anon_sym_DQUOTE,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [305028] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13644), 1,
      anon_sym_DQUOTE,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [305056] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13512), 1,
      anon_sym_DQUOTE,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [305084] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13524), 1,
      anon_sym_DQUOTE,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [305112] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13858), 1,
      anon_sym_DQUOTE,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [305140] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13860), 1,
      anon_sym_DQUOTE,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [305168] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13520), 1,
      anon_sym_DQUOTE,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [305196] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1619), 1,
      anon_sym_DQUOTE,
    ACTIONS(1623), 1,
      sym_variable_name,
    STATE(868), 1,
      sym_string,
    ACTIONS(1621), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1617), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [305224] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13476), 1,
      anon_sym_DQUOTE,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [305252] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13106), 1,
      anon_sym_DQUOTE,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [305280] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13862), 1,
      anon_sym_DQUOTE,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [305308] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(7358), 1,
      anon_sym_DQUOTE,
    ACTIONS(7362), 1,
      sym_variable_name,
    STATE(4457), 1,
      sym_string,
    ACTIONS(7360), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(7356), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [305336] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13864), 1,
      anon_sym_DQUOTE,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [305364] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2867), 1,
      anon_sym_DQUOTE,
    ACTIONS(6852), 1,
      sym_variable_name,
    STATE(4263), 1,
      sym_string,
    ACTIONS(6850), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(6848), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [305392] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13866), 1,
      anon_sym_DQUOTE,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [305420] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13536), 1,
      anon_sym_DQUOTE,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [305448] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13868), 1,
      anon_sym_DQUOTE,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [305476] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2042), 1,
      anon_sym_DQUOTE,
    ACTIONS(3647), 1,
      sym_variable_name,
    STATE(1547), 1,
      sym_string,
    ACTIONS(3645), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(3643), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [305504] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3030), 1,
      anon_sym_DQUOTE,
    ACTIONS(7135), 1,
      sym_variable_name,
    STATE(4446), 1,
      sym_string,
    ACTIONS(7133), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(7131), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [305532] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13636), 1,
      anon_sym_DQUOTE,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [305560] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13870), 1,
      anon_sym_DQUOTE,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [305588] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3282), 1,
      anon_sym_DQUOTE,
    ACTIONS(3286), 1,
      sym_variable_name,
    STATE(1628), 1,
      sym_string,
    ACTIONS(3284), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(3280), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [305616] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3115), 1,
      anon_sym_DQUOTE,
    ACTIONS(3119), 1,
      sym_variable_name,
    STATE(1864), 1,
      sym_string,
    ACTIONS(3117), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(3113), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [305644] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13872), 1,
      anon_sym_DQUOTE,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [305672] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13544), 1,
      anon_sym_DQUOTE,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [305700] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13874), 1,
      anon_sym_DQUOTE,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [305728] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13556), 1,
      anon_sym_DQUOTE,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [305756] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2578), 1,
      anon_sym_DQUOTE,
    ACTIONS(2953), 1,
      sym_variable_name,
    STATE(1585), 1,
      sym_string,
    ACTIONS(2951), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(2949), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [305784] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13876), 1,
      anon_sym_DQUOTE,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [305812] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13878), 1,
      anon_sym_DQUOTE,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [305840] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3189), 1,
      anon_sym_DQUOTE,
    ACTIONS(4058), 1,
      sym_variable_name,
    STATE(1737), 1,
      sym_string,
    ACTIONS(4056), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(4054), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [305868] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1791), 1,
      anon_sym_DQUOTE,
    ACTIONS(1795), 1,
      sym_variable_name,
    STATE(998), 1,
      sym_string,
    ACTIONS(1793), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1789), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [305896] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13880), 1,
      anon_sym_DQUOTE,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [305924] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13552), 1,
      anon_sym_DQUOTE,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [305952] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3213), 1,
      anon_sym_DQUOTE,
    ACTIONS(3217), 1,
      sym_variable_name,
    STATE(1919), 1,
      sym_string,
    ACTIONS(3215), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(3211), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [305980] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13882), 1,
      anon_sym_DQUOTE,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [306008] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3679), 1,
      anon_sym_DQUOTE,
    ACTIONS(7352), 1,
      sym_variable_name,
    STATE(4859), 1,
      sym_string,
    ACTIONS(7350), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(7348), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [306036] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13624), 1,
      anon_sym_DQUOTE,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [306064] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13884), 1,
      anon_sym_DQUOTE,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [306092] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13122), 1,
      anon_sym_DQUOTE,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [306120] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13110), 1,
      anon_sym_DQUOTE,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [306148] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13568), 1,
      anon_sym_DQUOTE,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [306176] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13126), 1,
      anon_sym_DQUOTE,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [306204] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3580), 1,
      anon_sym_DQUOTE,
    ACTIONS(3584), 1,
      sym_variable_name,
    STATE(2121), 1,
      sym_string,
    ACTIONS(3582), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(3578), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [306232] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3274), 1,
      anon_sym_DQUOTE,
    ACTIONS(3278), 1,
      sym_variable_name,
    STATE(1825), 1,
      sym_string,
    ACTIONS(3276), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(3272), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [306260] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13886), 1,
      anon_sym_DQUOTE,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [306288] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13888), 1,
      anon_sym_DQUOTE,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [306316] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13082), 1,
      anon_sym_DQUOTE,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [306344] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13890), 1,
      anon_sym_DQUOTE,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [306372] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(7143), 1,
      anon_sym_DQUOTE,
    ACTIONS(7147), 1,
      sym_variable_name,
    STATE(4322), 1,
      sym_string,
    ACTIONS(7145), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(7141), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [306400] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13628), 1,
      anon_sym_DQUOTE,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [306428] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13576), 1,
      anon_sym_DQUOTE,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [306456] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13588), 1,
      anon_sym_DQUOTE,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [306484] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1458), 1,
      anon_sym_DQUOTE,
    ACTIONS(1973), 1,
      sym_variable_name,
    STATE(1130), 1,
      sym_string,
    ACTIONS(1971), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(1969), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [306512] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2267), 1,
      anon_sym_DQUOTE,
    ACTIONS(3829), 1,
      sym_variable_name,
    STATE(1595), 1,
      sym_string,
    ACTIONS(3827), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(3825), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [306540] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13616), 1,
      anon_sym_DQUOTE,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [306568] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13892), 1,
      anon_sym_DQUOTE,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [306596] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13894), 1,
      anon_sym_DQUOTE,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [306624] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13102), 1,
      anon_sym_DQUOTE,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [306652] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13584), 1,
      anon_sym_DQUOTE,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [306680] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13896), 1,
      anon_sym_DQUOTE,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [306708] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2790), 1,
      anon_sym_DQUOTE,
    ACTIONS(3223), 1,
      sym_variable_name,
    STATE(1600), 1,
      sym_string,
    ACTIONS(3221), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(3219), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [306736] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13094), 1,
      anon_sym_DQUOTE,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [306764] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13898), 1,
      anon_sym_DQUOTE,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [306792] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13900), 1,
      anon_sym_DQUOTE,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [306820] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5361), 1,
      anon_sym_DQUOTE,
    ACTIONS(5365), 1,
      sym_variable_name,
    STATE(2850), 1,
      sym_string,
    ACTIONS(5363), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(5359), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [306848] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13596), 1,
      anon_sym_DQUOTE,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [306876] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2965), 1,
      anon_sym_DQUOTE,
    ACTIONS(2969), 1,
      sym_variable_name,
    STATE(1454), 1,
      sym_string,
    ACTIONS(2967), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(2963), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [306904] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1835), 1,
      anon_sym_DQUOTE,
    ACTIONS(3125), 1,
      sym_variable_name,
    STATE(1416), 1,
      sym_string,
    ACTIONS(3123), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(3121), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [306932] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13902), 1,
      anon_sym_DQUOTE,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [306960] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13608), 1,
      anon_sym_DQUOTE,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [306988] = 7,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(13906), 1,
      anon_sym_LT_LT,
    ACTIONS(13912), 1,
      anon_sym_LT_LT_DASH,
    ACTIONS(13914), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(13910), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(13908), 3,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
    ACTIONS(13904), 5,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
  [307017] = 7,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(13918), 1,
      anon_sym_LT_LT,
    ACTIONS(13924), 1,
      anon_sym_LT_LT_DASH,
    ACTIONS(13926), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(13922), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(13920), 3,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
    ACTIONS(13916), 5,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
  [307046] = 7,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(13918), 1,
      anon_sym_LT_LT,
    ACTIONS(13924), 1,
      anon_sym_LT_LT_DASH,
    ACTIONS(13934), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(13932), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(13930), 3,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
    ACTIONS(13928), 5,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
  [307075] = 7,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(13918), 1,
      anon_sym_LT_LT,
    ACTIONS(13924), 1,
      anon_sym_LT_LT_DASH,
    ACTIONS(13942), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(13940), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(13938), 3,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
    ACTIONS(13936), 5,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
  [307104] = 7,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(13946), 1,
      anon_sym_LT_LT,
    ACTIONS(13952), 1,
      anon_sym_LT_LT_DASH,
    ACTIONS(13954), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(13950), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(13948), 3,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
    ACTIONS(13944), 5,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
  [307133] = 7,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(13918), 1,
      anon_sym_LT_LT,
    ACTIONS(13924), 1,
      anon_sym_LT_LT_DASH,
    ACTIONS(13962), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(13960), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(13958), 3,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
    ACTIONS(13956), 5,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
  [307162] = 7,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(13918), 1,
      anon_sym_LT_LT,
    ACTIONS(13924), 1,
      anon_sym_LT_LT_DASH,
    ACTIONS(13970), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(13968), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(13966), 3,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
    ACTIONS(13964), 5,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
  [307191] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1346), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1344), 12,
      anon_sym_LPAREN_LPAREN,
      anon_sym_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_RBRACE3,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      aux_sym__simple_variable_name_token1,
  [307212] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1358), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1356), 12,
      anon_sym_LPAREN_LPAREN,
      anon_sym_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      aux_sym_number_token1,
      aux_sym_number_token2,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_RBRACE3,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
      aux_sym__simple_variable_name_token1,
  [307233] = 7,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(13974), 1,
      anon_sym_LT_LT,
    ACTIONS(13980), 1,
      anon_sym_LT_LT_DASH,
    ACTIONS(13982), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(13978), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(13976), 3,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
    ACTIONS(13972), 5,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
  [307262] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13732), 1,
      sym_string_content,
    ACTIONS(13736), 1,
      sym_variable_name,
    ACTIONS(13734), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13730), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [307287] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13988), 1,
      sym_variable_name,
    ACTIONS(13986), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13984), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [307309] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(13974), 1,
      anon_sym_LT_LT,
    ACTIONS(13980), 1,
      anon_sym_LT_LT_DASH,
    ACTIONS(13978), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(13976), 3,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
    ACTIONS(13972), 5,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
  [307335] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13994), 1,
      sym_variable_name,
    ACTIONS(13992), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13990), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [307357] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(14000), 1,
      sym_variable_name,
    ACTIONS(13998), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13996), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [307379] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(14006), 1,
      sym_variable_name,
    ACTIONS(14004), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(14002), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [307401] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(13906), 1,
      anon_sym_LT_LT,
    ACTIONS(13912), 1,
      anon_sym_LT_LT_DASH,
    ACTIONS(13910), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(13908), 3,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
    ACTIONS(13904), 5,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
  [307427] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14010), 1,
      anon_sym_LT_LT,
    ACTIONS(14016), 1,
      anon_sym_LT_LT_DASH,
    ACTIONS(14014), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(14012), 3,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
    ACTIONS(14008), 5,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
  [307453] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(13918), 1,
      anon_sym_LT_LT,
    ACTIONS(13924), 1,
      anon_sym_LT_LT_DASH,
    ACTIONS(13932), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(13930), 3,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
    ACTIONS(13928), 5,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
  [307479] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(13918), 1,
      anon_sym_LT_LT,
    ACTIONS(13924), 1,
      anon_sym_LT_LT_DASH,
    ACTIONS(13960), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(13958), 3,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
    ACTIONS(13956), 5,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
  [307505] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(14022), 1,
      sym_variable_name,
    ACTIONS(14020), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(14018), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [307527] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(13918), 1,
      anon_sym_LT_LT,
    ACTIONS(13924), 1,
      anon_sym_LT_LT_DASH,
    ACTIONS(13940), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(13938), 3,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
    ACTIONS(13936), 5,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
  [307553] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(14028), 1,
      sym_variable_name,
    ACTIONS(14026), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(14024), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [307575] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(14034), 1,
      sym_variable_name,
    ACTIONS(14032), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(14030), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [307597] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13782), 1,
      sym_variable_name,
    ACTIONS(13780), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(13776), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [307619] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(13918), 1,
      anon_sym_LT_LT,
    ACTIONS(13924), 1,
      anon_sym_LT_LT_DASH,
    ACTIONS(13922), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(13920), 3,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
    ACTIONS(13916), 5,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
  [307645] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(14040), 1,
      sym_variable_name,
    ACTIONS(14038), 2,
      aux_sym__simple_variable_name_token1,
      aux_sym__multiline_variable_name_token1,
    ACTIONS(14036), 9,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
      anon_sym_POUND,
      anon_sym_AT2,
      anon_sym_0,
      anon_sym__,
  [307667] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(13918), 1,
      anon_sym_LT_LT,
    ACTIONS(13924), 1,
      anon_sym_LT_LT_DASH,
    ACTIONS(13968), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(13966), 3,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
    ACTIONS(13964), 5,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
  [307693] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(13946), 1,
      anon_sym_LT_LT,
    ACTIONS(13952), 1,
      anon_sym_LT_LT_DASH,
    ACTIONS(13950), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(13948), 3,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
    ACTIONS(13944), 5,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
  [307719] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14048), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(14046), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(14044), 3,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
    ACTIONS(14042), 5,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
  [307742] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14056), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(14054), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(14052), 3,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
    ACTIONS(14050), 5,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
  [307765] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14064), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(14062), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(14060), 3,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
    ACTIONS(14058), 5,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
  [307788] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14072), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(14070), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(14068), 3,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
    ACTIONS(14066), 5,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
  [307811] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14074), 1,
      anon_sym_RBRACE3,
    ACTIONS(14076), 10,
      anon_sym_U,
      anon_sym_u,
      anon_sym_L,
      anon_sym_Q,
      anon_sym_E,
      anon_sym_P,
      anon_sym_A,
      anon_sym_K,
      anon_sym_a,
      anon_sym_k,
  [307830] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1314), 1,
      sym__concat,
    ACTIONS(1312), 10,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      anon_sym_DQUOTE,
      sym_string_content,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
  [307849] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1310), 1,
      sym__concat,
    ACTIONS(1308), 10,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      anon_sym_DQUOTE,
      sym_string_content,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
  [307868] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1338), 1,
      sym__concat,
    ACTIONS(1336), 10,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      anon_sym_DQUOTE,
      sym_string_content,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
  [307887] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(14078), 1,
      sym__concat,
    ACTIONS(13333), 10,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      anon_sym_DQUOTE,
      sym_string_content,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
  [307906] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(13962), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(13960), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(13958), 3,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
    ACTIONS(13956), 5,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
  [307929] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(14082), 1,
      sym__concat,
    ACTIONS(14080), 10,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      anon_sym_DQUOTE,
      sym_string_content,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
  [307948] = 9,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14084), 1,
      anon_sym_DOLLAR,
    ACTIONS(14086), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(14088), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(14090), 1,
      anon_sym_BQUOTE,
    ACTIONS(14092), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(14094), 1,
      sym_heredoc_content,
    ACTIONS(14096), 1,
      sym_heredoc_end,
    STATE(6131), 4,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      aux_sym_heredoc_body_repeat1,
  [307979] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(13926), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(13922), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(13920), 3,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
    ACTIONS(13916), 5,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
  [308002] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(13954), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(13950), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(13948), 3,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
    ACTIONS(13944), 5,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
  [308025] = 9,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14084), 1,
      anon_sym_DOLLAR,
    ACTIONS(14086), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(14088), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(14090), 1,
      anon_sym_BQUOTE,
    ACTIONS(14092), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(14098), 1,
      sym_heredoc_content,
    ACTIONS(14100), 1,
      sym_heredoc_end,
    STATE(6137), 4,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      aux_sym_heredoc_body_repeat1,
  [308056] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14108), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(14106), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(14104), 3,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
    ACTIONS(14102), 5,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
  [308079] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14116), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(14114), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(14112), 3,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
    ACTIONS(14110), 5,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
  [308102] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(13942), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(13940), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(13938), 3,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
    ACTIONS(13936), 5,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
  [308125] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14124), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(14122), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(14120), 3,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
    ACTIONS(14118), 5,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
  [308148] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14132), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(14130), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(14128), 3,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
    ACTIONS(14126), 5,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
  [308171] = 9,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14134), 1,
      anon_sym_DOLLAR,
    ACTIONS(14137), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(14140), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(14143), 1,
      anon_sym_BQUOTE,
    ACTIONS(14146), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(14149), 1,
      sym_heredoc_content,
    ACTIONS(14152), 1,
      sym_heredoc_end,
    STATE(6137), 4,
      sym_simple_expansion,
      sym_expansion,
      sym_command_substitution,
      aux_sym_heredoc_body_repeat1,
  [308202] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1294), 1,
      sym__concat,
    ACTIONS(1292), 10,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      anon_sym_DQUOTE,
      sym_string_content,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
  [308221] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14160), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(14158), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(14156), 3,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
    ACTIONS(14154), 5,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
  [308244] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1358), 1,
      sym__concat,
    ACTIONS(1356), 10,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      anon_sym_DQUOTE,
      sym_string_content,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
  [308263] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1334), 1,
      sym__concat,
    ACTIONS(1332), 10,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      anon_sym_DQUOTE,
      sym_string_content,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
  [308282] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14168), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(14166), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(14164), 3,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
    ACTIONS(14162), 5,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
  [308305] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1346), 1,
      sym__concat,
    ACTIONS(1344), 10,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      anon_sym_DQUOTE,
      sym_string_content,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
  [308324] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14176), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(14174), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(14172), 3,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
    ACTIONS(14170), 5,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
  [308347] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14184), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(14182), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(14180), 3,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
    ACTIONS(14178), 5,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
  [308370] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(13982), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(13978), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(13976), 3,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
    ACTIONS(13972), 5,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
  [308393] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1306), 1,
      sym__concat,
    ACTIONS(1304), 10,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      anon_sym_DQUOTE,
      sym_string_content,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
  [308412] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14192), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(14190), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(14188), 3,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
    ACTIONS(14186), 5,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
  [308435] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14200), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(14198), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(14196), 3,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
    ACTIONS(14194), 5,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
  [308458] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14202), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(14014), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(14012), 3,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
    ACTIONS(14008), 5,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
  [308481] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14210), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(14208), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(14206), 3,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
    ACTIONS(14204), 5,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
  [308504] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14218), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(14216), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(14214), 3,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
    ACTIONS(14212), 5,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
  [308527] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(13970), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(13968), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(13966), 3,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
    ACTIONS(13964), 5,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
  [308550] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1310), 1,
      sym__concat,
    ACTIONS(1308), 10,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      anon_sym_DQUOTE,
      sym_string_content,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
  [308569] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14226), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(14224), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(14222), 3,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
    ACTIONS(14220), 5,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
  [308592] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14234), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(14232), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(14230), 3,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
    ACTIONS(14228), 5,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
  [308615] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14242), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(14240), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(14238), 3,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
    ACTIONS(14236), 5,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
  [308638] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(13934), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(13932), 2,
      anon_sym_LT_AMP_DASH,
      anon_sym_GT_AMP_DASH,
    ACTIONS(13930), 3,
      anon_sym_GT_GT,
      anon_sym_AMP_GT_GT,
      anon_sym_GT_PIPE,
    ACTIONS(13928), 5,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_GT,
      anon_sym_LT_AMP,
      anon_sym_GT_AMP,
  [308661] = 8,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14244), 1,
      anon_sym_LPAREN_LPAREN,
    ACTIONS(14246), 1,
      anon_sym_LPAREN,
    ACTIONS(14248), 1,
      anon_sym_if,
    ACTIONS(14250), 1,
      anon_sym_LBRACE,
    ACTIONS(14252), 1,
      anon_sym_LBRACK,
    ACTIONS(14254), 1,
      anon_sym_LBRACK_LBRACK,
    STATE(4580), 4,
      sym_if_statement,
      sym_compound_statement,
      sym_subshell,
      sym_test_command,
  [308689] = 8,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14256), 1,
      anon_sym_LPAREN_LPAREN,
    ACTIONS(14258), 1,
      anon_sym_LPAREN,
    ACTIONS(14260), 1,
      anon_sym_if,
    ACTIONS(14262), 1,
      anon_sym_LBRACE,
    ACTIONS(14264), 1,
      anon_sym_LBRACK,
    ACTIONS(14266), 1,
      anon_sym_LBRACK_LBRACK,
    STATE(4387), 4,
      sym_if_statement,
      sym_compound_statement,
      sym_subshell,
      sym_test_command,
  [308717] = 8,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14268), 1,
      anon_sym_LPAREN_LPAREN,
    ACTIONS(14270), 1,
      anon_sym_LPAREN,
    ACTIONS(14272), 1,
      anon_sym_if,
    ACTIONS(14274), 1,
      anon_sym_LBRACE,
    ACTIONS(14276), 1,
      anon_sym_LBRACK,
    ACTIONS(14278), 1,
      anon_sym_LBRACK_LBRACK,
    STATE(5001), 4,
      sym_if_statement,
      sym_compound_statement,
      sym_subshell,
      sym_test_command,
  [308745] = 8,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14268), 1,
      anon_sym_LPAREN_LPAREN,
    ACTIONS(14270), 1,
      anon_sym_LPAREN,
    ACTIONS(14272), 1,
      anon_sym_if,
    ACTIONS(14274), 1,
      anon_sym_LBRACE,
    ACTIONS(14276), 1,
      anon_sym_LBRACK,
    ACTIONS(14278), 1,
      anon_sym_LBRACK_LBRACK,
    STATE(4659), 4,
      sym_if_statement,
      sym_compound_statement,
      sym_subshell,
      sym_test_command,
  [308773] = 8,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14244), 1,
      anon_sym_LPAREN_LPAREN,
    ACTIONS(14246), 1,
      anon_sym_LPAREN,
    ACTIONS(14248), 1,
      anon_sym_if,
    ACTIONS(14250), 1,
      anon_sym_LBRACE,
    ACTIONS(14252), 1,
      anon_sym_LBRACK,
    ACTIONS(14254), 1,
      anon_sym_LBRACK_LBRACK,
    STATE(4549), 4,
      sym_if_statement,
      sym_compound_statement,
      sym_subshell,
      sym_test_command,
  [308801] = 8,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14280), 1,
      anon_sym_DQUOTE,
    ACTIONS(14282), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(14284), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(14286), 1,
      anon_sym_BQUOTE,
    ACTIONS(14288), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(5433), 2,
      anon_sym_PIPE,
      anon_sym_RPAREN,
    STATE(6677), 3,
      sym_string,
      sym_expansion,
      sym_command_substitution,
  [308829] = 8,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14290), 1,
      anon_sym_LPAREN_LPAREN,
    ACTIONS(14292), 1,
      anon_sym_LPAREN,
    ACTIONS(14294), 1,
      anon_sym_if,
    ACTIONS(14296), 1,
      anon_sym_LBRACE,
    ACTIONS(14298), 1,
      anon_sym_LBRACK,
    ACTIONS(14300), 1,
      anon_sym_LBRACK_LBRACK,
    STATE(5405), 4,
      sym_if_statement,
      sym_compound_statement,
      sym_subshell,
      sym_test_command,
  [308857] = 8,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14268), 1,
      anon_sym_LPAREN_LPAREN,
    ACTIONS(14270), 1,
      anon_sym_LPAREN,
    ACTIONS(14272), 1,
      anon_sym_if,
    ACTIONS(14274), 1,
      anon_sym_LBRACE,
    ACTIONS(14276), 1,
      anon_sym_LBRACK,
    ACTIONS(14278), 1,
      anon_sym_LBRACK_LBRACK,
    STATE(4829), 4,
      sym_if_statement,
      sym_compound_statement,
      sym_subshell,
      sym_test_command,
  [308885] = 8,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14244), 1,
      anon_sym_LPAREN_LPAREN,
    ACTIONS(14248), 1,
      anon_sym_if,
    ACTIONS(14250), 1,
      anon_sym_LBRACE,
    ACTIONS(14252), 1,
      anon_sym_LBRACK,
    ACTIONS(14254), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(14302), 1,
      anon_sym_LPAREN,
    STATE(4420), 4,
      sym_if_statement,
      sym_compound_statement,
      sym_subshell,
      sym_test_command,
  [308913] = 8,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14256), 1,
      anon_sym_LPAREN_LPAREN,
    ACTIONS(14260), 1,
      anon_sym_if,
    ACTIONS(14262), 1,
      anon_sym_LBRACE,
    ACTIONS(14264), 1,
      anon_sym_LBRACK,
    ACTIONS(14266), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(14304), 1,
      anon_sym_LPAREN,
    STATE(4273), 4,
      sym_if_statement,
      sym_compound_statement,
      sym_subshell,
      sym_test_command,
  [308941] = 8,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14268), 1,
      anon_sym_LPAREN_LPAREN,
    ACTIONS(14270), 1,
      anon_sym_LPAREN,
    ACTIONS(14272), 1,
      anon_sym_if,
    ACTIONS(14274), 1,
      anon_sym_LBRACE,
    ACTIONS(14276), 1,
      anon_sym_LBRACK,
    ACTIONS(14278), 1,
      anon_sym_LBRACK_LBRACK,
    STATE(5000), 4,
      sym_if_statement,
      sym_compound_statement,
      sym_subshell,
      sym_test_command,
  [308969] = 8,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14256), 1,
      anon_sym_LPAREN_LPAREN,
    ACTIONS(14260), 1,
      anon_sym_if,
    ACTIONS(14262), 1,
      anon_sym_LBRACE,
    ACTIONS(14264), 1,
      anon_sym_LBRACK,
    ACTIONS(14266), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(14304), 1,
      anon_sym_LPAREN,
    STATE(4346), 4,
      sym_if_statement,
      sym_compound_statement,
      sym_subshell,
      sym_test_command,
  [308997] = 8,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14268), 1,
      anon_sym_LPAREN_LPAREN,
    ACTIONS(14272), 1,
      anon_sym_if,
    ACTIONS(14274), 1,
      anon_sym_LBRACE,
    ACTIONS(14276), 1,
      anon_sym_LBRACK,
    ACTIONS(14278), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(14306), 1,
      anon_sym_LPAREN,
    STATE(4723), 4,
      sym_if_statement,
      sym_compound_statement,
      sym_subshell,
      sym_test_command,
  [309025] = 8,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14268), 1,
      anon_sym_LPAREN_LPAREN,
    ACTIONS(14272), 1,
      anon_sym_if,
    ACTIONS(14274), 1,
      anon_sym_LBRACE,
    ACTIONS(14276), 1,
      anon_sym_LBRACK,
    ACTIONS(14278), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(14308), 1,
      anon_sym_LPAREN,
    STATE(5367), 4,
      sym_if_statement,
      sym_compound_statement,
      sym_subshell,
      sym_test_command,
  [309053] = 8,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14268), 1,
      anon_sym_LPAREN_LPAREN,
    ACTIONS(14270), 1,
      anon_sym_LPAREN,
    ACTIONS(14272), 1,
      anon_sym_if,
    ACTIONS(14274), 1,
      anon_sym_LBRACE,
    ACTIONS(14276), 1,
      anon_sym_LBRACK,
    ACTIONS(14278), 1,
      anon_sym_LBRACK_LBRACK,
    STATE(4625), 4,
      sym_if_statement,
      sym_compound_statement,
      sym_subshell,
      sym_test_command,
  [309081] = 8,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14268), 1,
      anon_sym_LPAREN_LPAREN,
    ACTIONS(14272), 1,
      anon_sym_if,
    ACTIONS(14274), 1,
      anon_sym_LBRACE,
    ACTIONS(14276), 1,
      anon_sym_LBRACK,
    ACTIONS(14278), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(14310), 1,
      anon_sym_LPAREN,
    STATE(4719), 4,
      sym_if_statement,
      sym_compound_statement,
      sym_subshell,
      sym_test_command,
  [309109] = 8,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14290), 1,
      anon_sym_LPAREN_LPAREN,
    ACTIONS(14294), 1,
      anon_sym_if,
    ACTIONS(14296), 1,
      anon_sym_LBRACE,
    ACTIONS(14298), 1,
      anon_sym_LBRACK,
    ACTIONS(14300), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(14312), 1,
      anon_sym_LPAREN,
    STATE(5423), 4,
      sym_if_statement,
      sym_compound_statement,
      sym_subshell,
      sym_test_command,
  [309137] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14076), 10,
      anon_sym_U,
      anon_sym_u,
      anon_sym_L,
      anon_sym_Q,
      anon_sym_E,
      anon_sym_P,
      anon_sym_A,
      anon_sym_K,
      anon_sym_a,
      anon_sym_k,
  [309153] = 8,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14268), 1,
      anon_sym_LPAREN_LPAREN,
    ACTIONS(14270), 1,
      anon_sym_LPAREN,
    ACTIONS(14272), 1,
      anon_sym_if,
    ACTIONS(14274), 1,
      anon_sym_LBRACE,
    ACTIONS(14276), 1,
      anon_sym_LBRACK,
    ACTIONS(14278), 1,
      anon_sym_LBRACK_LBRACK,
    STATE(5372), 4,
      sym_if_statement,
      sym_compound_statement,
      sym_subshell,
      sym_test_command,
  [309181] = 8,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14268), 1,
      anon_sym_LPAREN_LPAREN,
    ACTIONS(14270), 1,
      anon_sym_LPAREN,
    ACTIONS(14272), 1,
      anon_sym_if,
    ACTIONS(14274), 1,
      anon_sym_LBRACE,
    ACTIONS(14276), 1,
      anon_sym_LBRACK,
    ACTIONS(14278), 1,
      anon_sym_LBRACK_LBRACK,
    STATE(5375), 4,
      sym_if_statement,
      sym_compound_statement,
      sym_subshell,
      sym_test_command,
  [309209] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(14314), 10,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      anon_sym_DQUOTE,
      sym_string_content,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
  [309225] = 8,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14290), 1,
      anon_sym_LPAREN_LPAREN,
    ACTIONS(14292), 1,
      anon_sym_LPAREN,
    ACTIONS(14294), 1,
      anon_sym_if,
    ACTIONS(14296), 1,
      anon_sym_LBRACE,
    ACTIONS(14298), 1,
      anon_sym_LBRACK,
    ACTIONS(14300), 1,
      anon_sym_LBRACK_LBRACK,
    STATE(5475), 4,
      sym_if_statement,
      sym_compound_statement,
      sym_subshell,
      sym_test_command,
  [309253] = 8,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14268), 1,
      anon_sym_LPAREN_LPAREN,
    ACTIONS(14270), 1,
      anon_sym_LPAREN,
    ACTIONS(14272), 1,
      anon_sym_if,
    ACTIONS(14274), 1,
      anon_sym_LBRACE,
    ACTIONS(14276), 1,
      anon_sym_LBRACK,
    ACTIONS(14278), 1,
      anon_sym_LBRACK_LBRACK,
    STATE(4747), 4,
      sym_if_statement,
      sym_compound_statement,
      sym_subshell,
      sym_test_command,
  [309281] = 8,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14268), 1,
      anon_sym_LPAREN_LPAREN,
    ACTIONS(14270), 1,
      anon_sym_LPAREN,
    ACTIONS(14272), 1,
      anon_sym_if,
    ACTIONS(14274), 1,
      anon_sym_LBRACE,
    ACTIONS(14276), 1,
      anon_sym_LBRACK,
    ACTIONS(14278), 1,
      anon_sym_LBRACK_LBRACK,
    STATE(4758), 4,
      sym_if_statement,
      sym_compound_statement,
      sym_subshell,
      sym_test_command,
  [309309] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(13333), 10,
      anon_sym_LPAREN_LPAREN,
      anon_sym_DOLLAR_LPAREN_LPAREN,
      anon_sym_DOLLAR_LBRACK,
      anon_sym_DOLLAR,
      anon_sym_DQUOTE,
      sym_string_content,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
  [309325] = 8,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14268), 1,
      anon_sym_LPAREN_LPAREN,
    ACTIONS(14270), 1,
      anon_sym_LPAREN,
    ACTIONS(14272), 1,
      anon_sym_if,
    ACTIONS(14274), 1,
      anon_sym_LBRACE,
    ACTIONS(14276), 1,
      anon_sym_LBRACK,
    ACTIONS(14278), 1,
      anon_sym_LBRACK_LBRACK,
    STATE(4685), 4,
      sym_if_statement,
      sym_compound_statement,
      sym_subshell,
      sym_test_command,
  [309353] = 8,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14268), 1,
      anon_sym_LPAREN_LPAREN,
    ACTIONS(14272), 1,
      anon_sym_if,
    ACTIONS(14274), 1,
      anon_sym_LBRACE,
    ACTIONS(14276), 1,
      anon_sym_LBRACK,
    ACTIONS(14278), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(14316), 1,
      anon_sym_LPAREN,
    STATE(4667), 4,
      sym_if_statement,
      sym_compound_statement,
      sym_subshell,
      sym_test_command,
  [309381] = 8,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14268), 1,
      anon_sym_LPAREN_LPAREN,
    ACTIONS(14272), 1,
      anon_sym_if,
    ACTIONS(14274), 1,
      anon_sym_LBRACE,
    ACTIONS(14276), 1,
      anon_sym_LBRACK,
    ACTIONS(14278), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(14318), 1,
      anon_sym_LPAREN,
    STATE(4995), 4,
      sym_if_statement,
      sym_compound_statement,
      sym_subshell,
      sym_test_command,
  [309409] = 10,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14320), 1,
      anon_sym_SLASH,
    ACTIONS(14322), 1,
      anon_sym_DQUOTE,
    ACTIONS(14324), 1,
      anon_sym_RBRACE3,
    ACTIONS(14326), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(14328), 1,
      anon_sym_BQUOTE,
    ACTIONS(14330), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(14332), 1,
      sym__regex_no_slash,
    STATE(6582), 1,
      sym_string,
    STATE(6749), 1,
      sym_command_substitution,
  [309440] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1263), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(14334), 1,
      aux_sym_concatenation_token1,
    ACTIONS(14336), 1,
      sym__concat,
    STATE(6201), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1261), 5,
      anon_sym_in,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
      sym__special_character,
  [309463] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(14334), 1,
      aux_sym_concatenation_token1,
    ACTIONS(14336), 1,
      sym__concat,
    ACTIONS(14338), 1,
      anon_sym_in,
    ACTIONS(14342), 1,
      aux_sym_heredoc_redirect_token1,
    STATE(6201), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(14340), 3,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
  [309487] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(14344), 1,
      anon_sym_RPAREN,
    ACTIONS(14346), 1,
      anon_sym_DQUOTE,
    ACTIONS(14348), 1,
      sym_raw_string,
    ACTIONS(14350), 1,
      anon_sym_RBRACE3,
    ACTIONS(14352), 1,
      aux_sym__expansion_regex_token1,
    ACTIONS(14354), 1,
      sym_regex,
    STATE(6194), 2,
      sym_string,
      aux_sym__expansion_regex_repeat1,
  [309513] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(14334), 1,
      aux_sym_concatenation_token1,
    ACTIONS(14336), 1,
      sym__concat,
    ACTIONS(14356), 1,
      anon_sym_in,
    ACTIONS(14360), 1,
      aux_sym_heredoc_redirect_token1,
    STATE(6201), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(14358), 3,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
  [309537] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(14334), 1,
      aux_sym_concatenation_token1,
    ACTIONS(14336), 1,
      sym__concat,
    ACTIONS(14362), 1,
      anon_sym_in,
    ACTIONS(14366), 1,
      aux_sym_heredoc_redirect_token1,
    STATE(6193), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(14364), 3,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
  [309561] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1288), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(14334), 1,
      aux_sym_concatenation_token1,
    ACTIONS(14368), 1,
      sym__concat,
    STATE(6204), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1286), 4,
      anon_sym_in,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
  [309583] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(14370), 1,
      anon_sym_RPAREN,
    ACTIONS(14373), 1,
      anon_sym_DQUOTE,
    ACTIONS(14376), 1,
      sym_raw_string,
    ACTIONS(14379), 1,
      anon_sym_RBRACE3,
    ACTIONS(14381), 1,
      aux_sym__expansion_regex_token1,
    ACTIONS(14384), 1,
      sym_regex,
    STATE(6194), 2,
      sym_string,
      aux_sym__expansion_regex_repeat1,
  [309609] = 7,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7793), 1,
      aux_sym_number_token1,
    ACTIONS(7795), 1,
      aux_sym_number_token2,
    ACTIONS(14387), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(14389), 1,
      aux_sym__simple_variable_name_token1,
    STATE(6227), 1,
      sym__expansion_max_length_binary_expression,
    STATE(6215), 3,
      sym_number,
      sym_expansion,
      sym__expansion_max_length_expression,
  [309633] = 7,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7793), 1,
      aux_sym_number_token1,
    ACTIONS(7795), 1,
      aux_sym_number_token2,
    ACTIONS(14387), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(14391), 1,
      aux_sym__simple_variable_name_token1,
    STATE(6227), 1,
      sym__expansion_max_length_binary_expression,
    STATE(6212), 3,
      sym_number,
      sym_expansion,
      sym__expansion_max_length_expression,
  [309657] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(14344), 1,
      anon_sym_RPAREN,
    ACTIONS(14346), 1,
      anon_sym_DQUOTE,
    ACTIONS(14352), 1,
      aux_sym__expansion_regex_token1,
    ACTIONS(14393), 1,
      sym_raw_string,
    ACTIONS(14395), 1,
      anon_sym_RBRACE3,
    ACTIONS(14397), 1,
      sym_regex,
    STATE(6190), 2,
      sym_string,
      aux_sym__expansion_regex_repeat1,
  [309683] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(14334), 1,
      aux_sym_concatenation_token1,
    ACTIONS(14336), 1,
      sym__concat,
    ACTIONS(14399), 1,
      anon_sym_in,
    ACTIONS(14403), 1,
      aux_sym_heredoc_redirect_token1,
    STATE(6193), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(14401), 3,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
  [309707] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(14334), 1,
      aux_sym_concatenation_token1,
    ACTIONS(14336), 1,
      sym__concat,
    ACTIONS(14405), 1,
      anon_sym_in,
    ACTIONS(14409), 1,
      aux_sym_heredoc_redirect_token1,
    STATE(6193), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(14407), 3,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
  [309731] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(14334), 1,
      aux_sym_concatenation_token1,
    ACTIONS(14336), 1,
      sym__concat,
    ACTIONS(14411), 1,
      anon_sym_in,
    ACTIONS(14415), 1,
      aux_sym_heredoc_redirect_token1,
    STATE(6201), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(14413), 3,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
  [309755] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1267), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(14334), 1,
      aux_sym_concatenation_token1,
    ACTIONS(14417), 1,
      sym__concat,
    STATE(6204), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1265), 4,
      anon_sym_in,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
  [309777] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(14334), 1,
      aux_sym_concatenation_token1,
    ACTIONS(14336), 1,
      sym__concat,
    ACTIONS(14419), 1,
      anon_sym_in,
    ACTIONS(14423), 1,
      aux_sym_heredoc_redirect_token1,
    STATE(6193), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(14421), 3,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
  [309801] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(14334), 1,
      aux_sym_concatenation_token1,
    ACTIONS(14336), 1,
      sym__concat,
    ACTIONS(14425), 1,
      anon_sym_in,
    ACTIONS(14429), 1,
      aux_sym_heredoc_redirect_token1,
    STATE(6201), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(14427), 3,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
  [309825] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1253), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(14431), 1,
      aux_sym_concatenation_token1,
    ACTIONS(14434), 1,
      sym__concat,
    STATE(6204), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1251), 4,
      anon_sym_in,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
  [309847] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1326), 2,
      sym__concat,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1324), 5,
      anon_sym_in,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
      aux_sym_concatenation_token1,
  [309862] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1350), 2,
      sym__concat,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1348), 5,
      anon_sym_in,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
      aux_sym_concatenation_token1,
  [309877] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1294), 2,
      sym__concat,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1292), 5,
      anon_sym_in,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
      aux_sym_concatenation_token1,
  [309892] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14439), 1,
      anon_sym_COLON,
    ACTIONS(14441), 1,
      anon_sym_RBRACE3,
    ACTIONS(14437), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
  [309909] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1322), 2,
      sym__concat,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1320), 5,
      anon_sym_in,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
      aux_sym_concatenation_token1,
  [309924] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1302), 2,
      sym__concat,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1300), 5,
      anon_sym_in,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
      aux_sym_concatenation_token1,
  [309939] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1356), 1,
      anon_sym_DOLLAR,
    ACTIONS(1358), 6,
      sym_heredoc_content,
      sym_heredoc_end,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
  [309954] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14445), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(14443), 4,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_COLON,
      anon_sym_RBRACE3,
  [309969] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(14425), 1,
      anon_sym_in,
    ACTIONS(14429), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(14447), 1,
      sym__special_character,
    STATE(6240), 1,
      aux_sym__literal_repeat1,
    ACTIONS(14427), 3,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
  [309990] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1346), 2,
      sym__concat,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1344), 5,
      anon_sym_in,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
      aux_sym_concatenation_token1,
  [310005] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14443), 7,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_COLON,
      anon_sym_RBRACE3,
  [310018] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1338), 7,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_COLON,
      anon_sym_RBRACE3,
  [310031] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1338), 2,
      sym__concat,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1336), 5,
      anon_sym_in,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
      aux_sym_concatenation_token1,
  [310046] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14451), 1,
      anon_sym_COLON,
    ACTIONS(14453), 1,
      anon_sym_RBRACE3,
    ACTIONS(14449), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
  [310063] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(14356), 1,
      anon_sym_in,
    ACTIONS(14360), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(14447), 1,
      sym__special_character,
    STATE(6240), 1,
      aux_sym__literal_repeat1,
    ACTIONS(14358), 3,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
  [310084] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(14411), 1,
      anon_sym_in,
    ACTIONS(14415), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(14447), 1,
      sym__special_character,
    STATE(6240), 1,
      aux_sym__literal_repeat1,
    ACTIONS(14413), 3,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
  [310105] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1314), 2,
      sym__concat,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1312), 5,
      anon_sym_in,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
      aux_sym_concatenation_token1,
  [310120] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1358), 2,
      sym__concat,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1356), 5,
      anon_sym_in,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
      aux_sym_concatenation_token1,
  [310135] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(14338), 1,
      anon_sym_in,
    ACTIONS(14342), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(14447), 1,
      sym__special_character,
    STATE(6240), 1,
      aux_sym__literal_repeat1,
    ACTIONS(14340), 3,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
  [310156] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1308), 1,
      anon_sym_DOLLAR,
    ACTIONS(1310), 6,
      sym_heredoc_content,
      sym_heredoc_end,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
  [310171] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1253), 2,
      sym__concat,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1251), 5,
      anon_sym_in,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
      aux_sym_concatenation_token1,
  [310186] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1304), 1,
      anon_sym_DOLLAR,
    ACTIONS(1306), 6,
      sym_heredoc_content,
      sym_heredoc_end,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
  [310201] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14449), 7,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_COLON,
      anon_sym_RBRACE3,
  [310214] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1330), 7,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_COLON,
      anon_sym_RBRACE3,
  [310227] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1318), 7,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_COLON,
      anon_sym_RBRACE3,
  [310240] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1298), 2,
      sym__concat,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1296), 5,
      anon_sym_in,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
      aux_sym_concatenation_token1,
  [310255] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1334), 2,
      sym__concat,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1332), 5,
      anon_sym_in,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
      aux_sym_concatenation_token1,
  [310270] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1310), 2,
      sym__concat,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1308), 5,
      anon_sym_in,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
      aux_sym_concatenation_token1,
  [310285] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1308), 1,
      anon_sym_DOLLAR,
    ACTIONS(1310), 6,
      sym_heredoc_content,
      sym_heredoc_end,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
  [310300] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1314), 7,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_COLON,
      anon_sym_RBRACE3,
  [310313] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1344), 1,
      anon_sym_DOLLAR,
    ACTIONS(1346), 6,
      sym_heredoc_content,
      sym_heredoc_end,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
  [310328] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1330), 2,
      sym__concat,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1328), 5,
      anon_sym_in,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
      aux_sym_concatenation_token1,
  [310343] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1312), 1,
      anon_sym_DOLLAR,
    ACTIONS(1314), 6,
      sym_heredoc_content,
      sym_heredoc_end,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
  [310358] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1318), 2,
      sym__concat,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1316), 5,
      anon_sym_in,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
      aux_sym_concatenation_token1,
  [310373] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1336), 1,
      anon_sym_DOLLAR,
    ACTIONS(1338), 6,
      sym_heredoc_content,
      sym_heredoc_end,
      anon_sym_DOLLAR_LBRACE,
      anon_sym_DOLLAR_LPAREN,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_BQUOTE,
  [310388] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1364), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(14455), 1,
      sym__special_character,
    STATE(6240), 1,
      aux_sym__literal_repeat1,
    ACTIONS(1362), 4,
      anon_sym_in,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
  [310407] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1310), 2,
      sym__concat,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1308), 5,
      anon_sym_in,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
      aux_sym_concatenation_token1,
  [310422] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1306), 2,
      sym__concat,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1304), 5,
      anon_sym_in,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
      aux_sym_concatenation_token1,
  [310437] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1342), 2,
      sym__concat,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1340), 5,
      anon_sym_in,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
      aux_sym_concatenation_token1,
  [310452] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1354), 2,
      sym__concat,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1352), 5,
      anon_sym_in,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
      aux_sym_concatenation_token1,
  [310467] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1310), 7,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_COLON,
      anon_sym_RBRACE3,
  [310480] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1306), 7,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_COLON,
      anon_sym_RBRACE3,
  [310493] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1310), 7,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_COLON,
      anon_sym_RBRACE3,
  [310506] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(10142), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(10144), 1,
      anon_sym_BQUOTE,
    ACTIONS(10146), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(14458), 1,
      anon_sym_DOLLAR_LPAREN,
    STATE(4576), 2,
      sym_expansion,
      sym_command_substitution,
  [310526] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9382), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9384), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9386), 1,
      anon_sym_BQUOTE,
    ACTIONS(9388), 1,
      anon_sym_DOLLAR_BQUOTE,
    STATE(3215), 2,
      sym_expansion,
      sym_command_substitution,
  [310546] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14326), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(14328), 1,
      anon_sym_BQUOTE,
    ACTIONS(14330), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(14387), 1,
      anon_sym_DOLLAR_LBRACE,
    STATE(6228), 2,
      sym_expansion,
      sym_command_substitution,
  [310566] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(10512), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(10514), 1,
      anon_sym_BQUOTE,
    ACTIONS(10516), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(14460), 1,
      anon_sym_DOLLAR_LPAREN,
    STATE(1503), 2,
      sym_expansion,
      sym_command_substitution,
  [310586] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9054), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9056), 1,
      anon_sym_BQUOTE,
    ACTIONS(9058), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(14462), 1,
      anon_sym_DOLLAR_LPAREN,
    STATE(1781), 2,
      sym_expansion,
      sym_command_substitution,
  [310606] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9772), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9776), 1,
      anon_sym_BQUOTE,
    ACTIONS(9778), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(14464), 1,
      anon_sym_DOLLAR_LPAREN,
    STATE(6236), 2,
      sym_expansion,
      sym_command_substitution,
  [310626] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(8886), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8888), 1,
      anon_sym_BQUOTE,
    ACTIONS(8890), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(14466), 1,
      anon_sym_DOLLAR_LPAREN,
    STATE(1157), 2,
      sym_expansion,
      sym_command_substitution,
  [310646] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(8116), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8120), 1,
      anon_sym_BQUOTE,
    ACTIONS(8122), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(14468), 1,
      anon_sym_DOLLAR_LPAREN,
    STATE(1233), 2,
      sym_expansion,
      sym_command_substitution,
  [310666] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14470), 1,
      anon_sym_RBRACE3,
    ACTIONS(14449), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
  [310680] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9686), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9688), 1,
      anon_sym_BQUOTE,
    ACTIONS(9690), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(14472), 1,
      anon_sym_DOLLAR_LPAREN,
    STATE(1793), 2,
      sym_expansion,
      sym_command_substitution,
  [310700] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14474), 1,
      anon_sym_RBRACE3,
    ACTIONS(14437), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
  [310714] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14476), 1,
      anon_sym_RBRACE3,
    ACTIONS(14449), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
  [310728] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14478), 1,
      anon_sym_RBRACE3,
    ACTIONS(14437), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
  [310742] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14480), 1,
      anon_sym_RBRACE3,
    ACTIONS(14449), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
  [310756] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14482), 1,
      anon_sym_RBRACE3,
    ACTIONS(14437), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
  [310770] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(10070), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(10072), 1,
      anon_sym_BQUOTE,
    ACTIONS(10074), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(14484), 1,
      anon_sym_DOLLAR_LPAREN,
    STATE(4701), 2,
      sym_expansion,
      sym_command_substitution,
  [310790] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14482), 1,
      anon_sym_RBRACE3,
    ACTIONS(14437), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
  [310804] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14486), 1,
      anon_sym_fi,
    ACTIONS(14488), 1,
      anon_sym_elif,
    ACTIONS(14490), 1,
      anon_sym_else,
    STATE(7320), 1,
      sym_else_clause,
    STATE(6400), 2,
      sym_elif_clause,
      aux_sym_if_statement_repeat1,
  [310824] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14492), 1,
      anon_sym_RBRACE3,
    ACTIONS(14449), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
  [310838] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14482), 1,
      anon_sym_RBRACE3,
    ACTIONS(14437), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
  [310852] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14494), 1,
      anon_sym_RBRACE3,
    ACTIONS(14449), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
  [310866] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(465), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(469), 1,
      anon_sym_BQUOTE,
    ACTIONS(471), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(14496), 1,
      anon_sym_DOLLAR_LPAREN,
    STATE(887), 2,
      sym_expansion,
      sym_command_substitution,
  [310886] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(14500), 1,
      anon_sym_COMMA,
    ACTIONS(14502), 1,
      aux_sym_heredoc_redirect_token1,
    STATE(4651), 1,
      sym__c_terminator,
    STATE(6334), 1,
      aux_sym__for_body_repeat1,
    ACTIONS(14498), 2,
      anon_sym_SEMI,
      anon_sym_AMP,
  [310906] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1085), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1087), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3598), 1,
      anon_sym_BQUOTE,
    ACTIONS(14504), 1,
      anon_sym_DOLLAR_LPAREN,
    STATE(2586), 2,
      sym_expansion,
      sym_command_substitution,
  [310926] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(8970), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8972), 1,
      anon_sym_BQUOTE,
    ACTIONS(8974), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(14506), 1,
      anon_sym_DOLLAR_LPAREN,
    STATE(1530), 2,
      sym_expansion,
      sym_command_substitution,
  [310946] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1157), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1163), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(3060), 1,
      anon_sym_BQUOTE,
    ACTIONS(14508), 1,
      anon_sym_DOLLAR_LPAREN,
    STATE(2276), 2,
      sym_expansion,
      sym_command_substitution,
  [310966] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9072), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9074), 1,
      anon_sym_BQUOTE,
    ACTIONS(9076), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(14510), 1,
      anon_sym_DOLLAR_LPAREN,
    STATE(1617), 2,
      sym_expansion,
      sym_command_substitution,
  [310986] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14512), 1,
      anon_sym_RBRACE3,
    ACTIONS(14437), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
  [311000] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9320), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9322), 1,
      anon_sym_BQUOTE,
    ACTIONS(9324), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(14514), 1,
      anon_sym_DOLLAR_LPAREN,
    STATE(4561), 2,
      sym_expansion,
      sym_command_substitution,
  [311020] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(4834), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(4838), 1,
      anon_sym_BQUOTE,
    ACTIONS(4840), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(14516), 1,
      anon_sym_DOLLAR_LPAREN,
    STATE(2663), 2,
      sym_expansion,
      sym_command_substitution,
  [311040] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9916), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9918), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9920), 1,
      anon_sym_BQUOTE,
    ACTIONS(9922), 1,
      anon_sym_DOLLAR_BQUOTE,
    STATE(3295), 2,
      sym_expansion,
      sym_command_substitution,
  [311060] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14518), 1,
      anon_sym_RBRACE3,
    ACTIONS(14437), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
  [311074] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(14500), 1,
      anon_sym_COMMA,
    ACTIONS(14522), 1,
      aux_sym_heredoc_redirect_token1,
    STATE(4646), 1,
      sym__c_terminator,
    STATE(6463), 1,
      aux_sym__for_body_repeat1,
    ACTIONS(14520), 2,
      anon_sym_SEMI,
      anon_sym_AMP,
  [311094] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7972), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(7976), 1,
      anon_sym_BQUOTE,
    ACTIONS(7978), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(14524), 1,
      anon_sym_DOLLAR_LPAREN,
    STATE(4320), 2,
      sym_expansion,
      sym_command_substitution,
  [311114] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(8434), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8436), 1,
      anon_sym_BQUOTE,
    ACTIONS(8438), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(14526), 1,
      anon_sym_DOLLAR_LPAREN,
    STATE(1236), 2,
      sym_expansion,
      sym_command_substitution,
  [311134] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(4961), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(4965), 1,
      anon_sym_BQUOTE,
    ACTIONS(4967), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(14528), 1,
      anon_sym_DOLLAR_LPAREN,
    STATE(2606), 2,
      sym_expansion,
      sym_command_substitution,
  [311154] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9810), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9812), 1,
      anon_sym_BQUOTE,
    ACTIONS(9814), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(14530), 1,
      anon_sym_DOLLAR_LPAREN,
    STATE(4572), 2,
      sym_expansion,
      sym_command_substitution,
  [311174] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(8240), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8242), 1,
      anon_sym_BQUOTE,
    ACTIONS(8244), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(14532), 1,
      anon_sym_DOLLAR_LPAREN,
    STATE(1472), 2,
      sym_expansion,
      sym_command_substitution,
  [311194] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(14500), 1,
      anon_sym_COMMA,
    ACTIONS(14536), 1,
      aux_sym_heredoc_redirect_token1,
    STATE(4700), 1,
      sym__c_terminator,
    STATE(6463), 1,
      aux_sym__for_body_repeat1,
    ACTIONS(14534), 2,
      anon_sym_SEMI,
      anon_sym_AMP,
  [311214] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14512), 1,
      anon_sym_RBRACE3,
    ACTIONS(14437), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
  [311228] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9644), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9646), 1,
      anon_sym_BQUOTE,
    ACTIONS(9648), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(14538), 1,
      anon_sym_DOLLAR_LPAREN,
    STATE(4386), 2,
      sym_expansion,
      sym_command_substitution,
  [311248] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14488), 1,
      anon_sym_elif,
    ACTIONS(14490), 1,
      anon_sym_else,
    ACTIONS(14540), 1,
      anon_sym_fi,
    STATE(7480), 1,
      sym_else_clause,
    STATE(6400), 2,
      sym_elif_clause,
      aux_sym_if_statement_repeat1,
  [311268] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(125), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(129), 1,
      anon_sym_BQUOTE,
    ACTIONS(131), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(14542), 1,
      anon_sym_DOLLAR_LPAREN,
    STATE(484), 2,
      sym_expansion,
      sym_command_substitution,
  [311288] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7234), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(7236), 1,
      anon_sym_BQUOTE,
    ACTIONS(7238), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(14544), 1,
      anon_sym_DOLLAR_LPAREN,
    STATE(6519), 2,
      sym_expansion,
      sym_command_substitution,
  [311308] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(8192), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8196), 1,
      anon_sym_BQUOTE,
    ACTIONS(8198), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(14546), 1,
      anon_sym_DOLLAR_LPAREN,
    STATE(1668), 2,
      sym_expansion,
      sym_command_substitution,
  [311328] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1354), 2,
      sym_regex,
      aux_sym__expansion_regex_token1,
    ACTIONS(1352), 4,
      anon_sym_RPAREN,
      anon_sym_DQUOTE,
      sym_raw_string,
      anon_sym_RBRACE3,
  [311342] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(5275), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(5279), 1,
      anon_sym_BQUOTE,
    ACTIONS(5281), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(14548), 1,
      anon_sym_DOLLAR_LPAREN,
    STATE(5489), 2,
      sym_expansion,
      sym_command_substitution,
  [311362] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(4781), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(4785), 1,
      anon_sym_BQUOTE,
    ACTIONS(4787), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(14550), 1,
      anon_sym_DOLLAR_LPAREN,
    STATE(5399), 2,
      sym_expansion,
      sym_command_substitution,
  [311382] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(8154), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8158), 1,
      anon_sym_BQUOTE,
    ACTIONS(8160), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(14552), 1,
      anon_sym_DOLLAR_LPAREN,
    STATE(4595), 2,
      sym_expansion,
      sym_command_substitution,
  [311402] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(10944), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(10946), 1,
      anon_sym_BQUOTE,
    ACTIONS(10948), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(14554), 1,
      anon_sym_DOLLAR_LPAREN,
    STATE(1875), 2,
      sym_expansion,
      sym_command_substitution,
  [311422] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(4165), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(4169), 1,
      anon_sym_BQUOTE,
    ACTIONS(4171), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(14556), 1,
      anon_sym_DOLLAR_LPAREN,
    STATE(2444), 2,
      sym_expansion,
      sym_command_substitution,
  [311442] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14512), 1,
      anon_sym_RBRACE3,
    ACTIONS(14437), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
  [311456] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(14500), 1,
      anon_sym_COMMA,
    ACTIONS(14560), 1,
      aux_sym_heredoc_redirect_token1,
    STATE(4742), 1,
      sym__c_terminator,
    STATE(6280), 1,
      aux_sym__for_body_repeat1,
    ACTIONS(14558), 2,
      anon_sym_SEMI,
      anon_sym_AMP,
  [311476] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14562), 1,
      anon_sym_RBRACE3,
    ACTIONS(14449), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
  [311490] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14564), 1,
      anon_sym_PIPE,
    ACTIONS(14566), 1,
      anon_sym_RPAREN,
    STATE(6439), 1,
      aux_sym_concatenation_repeat1,
    STATE(6570), 1,
      aux_sym_case_item_repeat1,
    ACTIONS(14568), 2,
      sym__concat,
      aux_sym_concatenation_token1,
  [311510] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14564), 1,
      anon_sym_PIPE,
    ACTIONS(14570), 1,
      anon_sym_RPAREN,
    STATE(6439), 1,
      aux_sym_concatenation_repeat1,
    STATE(6710), 1,
      aux_sym_case_item_repeat1,
    ACTIONS(14568), 2,
      sym__concat,
      aux_sym_concatenation_token1,
  [311530] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1263), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(1261), 5,
      anon_sym_in,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
      sym__special_character,
  [311544] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9018), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9020), 1,
      anon_sym_BQUOTE,
    ACTIONS(9022), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(14572), 1,
      anon_sym_DOLLAR_LPAREN,
    STATE(1384), 2,
      sym_expansion,
      sym_command_substitution,
  [311564] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1342), 2,
      sym_regex,
      aux_sym__expansion_regex_token1,
    ACTIONS(1340), 4,
      anon_sym_RPAREN,
      anon_sym_DQUOTE,
      sym_raw_string,
      anon_sym_RBRACE3,
  [311578] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(379), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(385), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(5809), 1,
      anon_sym_BQUOTE,
    ACTIONS(14574), 1,
      anon_sym_DOLLAR_LPAREN,
    STATE(2697), 2,
      sym_expansion,
      sym_command_substitution,
  [311598] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(205), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(209), 1,
      anon_sym_BQUOTE,
    ACTIONS(211), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(14576), 1,
      anon_sym_DOLLAR_LPAREN,
    STATE(505), 2,
      sym_expansion,
      sym_command_substitution,
  [311618] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9248), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9250), 1,
      anon_sym_BQUOTE,
    ACTIONS(9252), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(14578), 1,
      anon_sym_DOLLAR_LPAREN,
    STATE(1248), 2,
      sym_expansion,
      sym_command_substitution,
  [311638] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14564), 1,
      anon_sym_PIPE,
    ACTIONS(14580), 1,
      anon_sym_RPAREN,
    STATE(6373), 1,
      aux_sym_concatenation_repeat1,
    STATE(6707), 1,
      aux_sym_case_item_repeat1,
    ACTIONS(14568), 2,
      sym__concat,
      aux_sym_concatenation_token1,
  [311658] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(8282), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8286), 1,
      anon_sym_BQUOTE,
    ACTIONS(8288), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(14582), 1,
      anon_sym_DOLLAR_LPAREN,
    STATE(1449), 2,
      sym_expansion,
      sym_command_substitution,
  [311678] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14564), 1,
      anon_sym_PIPE,
    ACTIONS(14584), 1,
      anon_sym_RPAREN,
    STATE(6373), 1,
      aux_sym_concatenation_repeat1,
    STATE(6590), 1,
      aux_sym_case_item_repeat1,
    ACTIONS(14568), 2,
      sym__concat,
      aux_sym_concatenation_token1,
  [311698] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(8526), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8530), 1,
      anon_sym_BQUOTE,
    ACTIONS(8532), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(14586), 1,
      anon_sym_DOLLAR_LPAREN,
    STATE(2510), 2,
      sym_expansion,
      sym_command_substitution,
  [311718] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14588), 1,
      anon_sym_SLASH,
    ACTIONS(14592), 1,
      anon_sym_RBRACE3,
    ACTIONS(14594), 1,
      sym__expansion_word,
    STATE(6475), 1,
      aux_sym__concatenation_in_expansion_repeat1,
    ACTIONS(14590), 2,
      sym__concat,
      aux_sym_concatenation_token1,
  [311738] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(8952), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8954), 1,
      anon_sym_BQUOTE,
    ACTIONS(8956), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(14596), 1,
      anon_sym_DOLLAR_LPAREN,
    STATE(3175), 2,
      sym_expansion,
      sym_command_substitution,
  [311758] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9132), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9136), 1,
      anon_sym_BQUOTE,
    ACTIONS(9138), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(14598), 1,
      anon_sym_DOLLAR_LPAREN,
    STATE(2399), 2,
      sym_expansion,
      sym_command_substitution,
  [311778] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1322), 2,
      sym_regex,
      aux_sym__expansion_regex_token1,
    ACTIONS(1320), 4,
      anon_sym_RPAREN,
      anon_sym_DQUOTE,
      sym_raw_string,
      anon_sym_RBRACE3,
  [311792] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14564), 1,
      anon_sym_PIPE,
    ACTIONS(14600), 1,
      anon_sym_RPAREN,
    STATE(6373), 1,
      aux_sym_concatenation_repeat1,
    STATE(6652), 1,
      aux_sym_case_item_repeat1,
    ACTIONS(14568), 2,
      sym__concat,
      aux_sym_concatenation_token1,
  [311812] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(14500), 1,
      anon_sym_COMMA,
    ACTIONS(14604), 1,
      aux_sym_heredoc_redirect_token1,
    STATE(4236), 1,
      sym__c_terminator,
    STATE(6463), 1,
      aux_sym__for_body_repeat1,
    ACTIONS(14602), 2,
      anon_sym_SEMI,
      anon_sym_AMP,
  [311832] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(4647), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(4651), 1,
      anon_sym_BQUOTE,
    ACTIONS(4653), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(14606), 1,
      anon_sym_DOLLAR_LPAREN,
    STATE(2583), 2,
      sym_expansion,
      sym_command_substitution,
  [311852] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(8352), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8354), 1,
      anon_sym_BQUOTE,
    ACTIONS(8356), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(14608), 1,
      anon_sym_DOLLAR_LPAREN,
    STATE(1538), 2,
      sym_expansion,
      sym_command_substitution,
  [311872] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(5329), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(5333), 1,
      anon_sym_BQUOTE,
    ACTIONS(5335), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(14610), 1,
      anon_sym_DOLLAR_LPAREN,
    STATE(5470), 2,
      sym_expansion,
      sym_command_substitution,
  [311892] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(11820), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(11822), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(11824), 1,
      anon_sym_BQUOTE,
    ACTIONS(11826), 1,
      anon_sym_DOLLAR_BQUOTE,
    STATE(3245), 2,
      sym_expansion,
      sym_command_substitution,
  [311912] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9996), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9998), 1,
      anon_sym_BQUOTE,
    ACTIONS(10000), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(14612), 1,
      anon_sym_DOLLAR_LPAREN,
    STATE(1479), 2,
      sym_expansion,
      sym_command_substitution,
  [311932] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(8320), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8324), 1,
      anon_sym_BQUOTE,
    ACTIONS(8326), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(14614), 1,
      anon_sym_DOLLAR_LPAREN,
    STATE(5435), 2,
      sym_expansion,
      sym_command_substitution,
  [311952] = 4,
    ACTIONS(71), 1,
      sym_comment,
    STATE(6373), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(14568), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(1263), 3,
      anon_sym_PIPE,
      anon_sym_RPAREN,
      sym__special_character,
  [311968] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14488), 1,
      anon_sym_elif,
    ACTIONS(14490), 1,
      anon_sym_else,
    ACTIONS(14616), 1,
      anon_sym_fi,
    STATE(7012), 1,
      sym_else_clause,
    STATE(6400), 2,
      sym_elif_clause,
      aux_sym_if_statement_repeat1,
  [311988] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14564), 1,
      anon_sym_PIPE,
    ACTIONS(14618), 1,
      anon_sym_RPAREN,
    STATE(6439), 1,
      aux_sym_concatenation_repeat1,
    STATE(6711), 1,
      aux_sym_case_item_repeat1,
    ACTIONS(14568), 2,
      sym__concat,
      aux_sym_concatenation_token1,
  [312008] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(8386), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8390), 1,
      anon_sym_BQUOTE,
    ACTIONS(8392), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(14620), 1,
      anon_sym_DOLLAR_LPAREN,
    STATE(4405), 2,
      sym_expansion,
      sym_command_substitution,
  [312028] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(14500), 1,
      anon_sym_COMMA,
    ACTIONS(14624), 1,
      aux_sym_heredoc_redirect_token1,
    STATE(4803), 1,
      sym__c_terminator,
    STATE(6286), 1,
      aux_sym__for_body_repeat1,
    ACTIONS(14622), 2,
      anon_sym_SEMI,
      anon_sym_AMP,
  [312048] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14488), 1,
      anon_sym_elif,
    ACTIONS(14490), 1,
      anon_sym_else,
    ACTIONS(14626), 1,
      anon_sym_fi,
    STATE(6921), 1,
      sym_else_clause,
    STATE(6400), 2,
      sym_elif_clause,
      aux_sym_if_statement_repeat1,
  [312068] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(8010), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8014), 1,
      anon_sym_BQUOTE,
    ACTIONS(8016), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(14628), 1,
      anon_sym_DOLLAR_LPAREN,
    STATE(3529), 2,
      sym_expansion,
      sym_command_substitution,
  [312088] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(8858), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8860), 1,
      anon_sym_BQUOTE,
    ACTIONS(8862), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(14630), 1,
      anon_sym_DOLLAR_LPAREN,
    STATE(2111), 2,
      sym_expansion,
      sym_command_substitution,
  [312108] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(14500), 1,
      anon_sym_COMMA,
    ACTIONS(14634), 1,
      aux_sym_heredoc_redirect_token1,
    STATE(4741), 1,
      sym__c_terminator,
    STATE(6463), 1,
      aux_sym__for_body_repeat1,
    ACTIONS(14632), 2,
      anon_sym_SEMI,
      anon_sym_AMP,
  [312128] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1057), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(1061), 1,
      anon_sym_BQUOTE,
    ACTIONS(1063), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(14636), 1,
      anon_sym_DOLLAR_LPAREN,
    STATE(1971), 2,
      sym_expansion,
      sym_command_substitution,
  [312148] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(8585), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8589), 1,
      anon_sym_BQUOTE,
    ACTIONS(8591), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(14638), 1,
      anon_sym_DOLLAR_LPAREN,
    STATE(5151), 2,
      sym_expansion,
      sym_command_substitution,
  [312168] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14488), 1,
      anon_sym_elif,
    ACTIONS(14490), 1,
      anon_sym_else,
    ACTIONS(14640), 1,
      anon_sym_fi,
    STATE(7004), 1,
      sym_else_clause,
    STATE(6400), 2,
      sym_elif_clause,
      aux_sym_if_statement_repeat1,
  [312188] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14564), 1,
      anon_sym_PIPE,
    ACTIONS(14642), 1,
      anon_sym_RPAREN,
    STATE(6439), 1,
      aux_sym_concatenation_repeat1,
    STATE(6597), 1,
      aux_sym_case_item_repeat1,
    ACTIONS(14568), 2,
      sym__concat,
      aux_sym_concatenation_token1,
  [312208] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14488), 1,
      anon_sym_elif,
    ACTIONS(14490), 1,
      anon_sym_else,
    ACTIONS(14644), 1,
      anon_sym_fi,
    STATE(6862), 1,
      sym_else_clause,
    STATE(6400), 2,
      sym_elif_clause,
      aux_sym_if_statement_repeat1,
  [312228] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14488), 1,
      anon_sym_elif,
    ACTIONS(14490), 1,
      anon_sym_else,
    ACTIONS(14646), 1,
      anon_sym_fi,
    STATE(7416), 1,
      sym_else_clause,
    STATE(6400), 2,
      sym_elif_clause,
      aux_sym_if_statement_repeat1,
  [312248] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9948), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9950), 1,
      anon_sym_BQUOTE,
    ACTIONS(9952), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(14648), 1,
      anon_sym_DOLLAR_LPAREN,
    STATE(4388), 2,
      sym_expansion,
      sym_command_substitution,
  [312268] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(2650), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(2654), 1,
      anon_sym_BQUOTE,
    ACTIONS(2656), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(14650), 1,
      anon_sym_DOLLAR_LPAREN,
    STATE(2333), 2,
      sym_expansion,
      sym_command_substitution,
  [312288] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(8074), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8076), 1,
      anon_sym_BQUOTE,
    ACTIONS(8078), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(14652), 1,
      anon_sym_DOLLAR_LPAREN,
    STATE(1689), 2,
      sym_expansion,
      sym_command_substitution,
  [312308] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14654), 1,
      anon_sym_RBRACE3,
    ACTIONS(14437), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
  [312322] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(5419), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(5423), 1,
      anon_sym_BQUOTE,
    ACTIONS(5425), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(14656), 1,
      anon_sym_DOLLAR_LPAREN,
    STATE(5521), 2,
      sym_expansion,
      sym_command_substitution,
  [312342] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14658), 1,
      anon_sym_SLASH,
    ACTIONS(14660), 1,
      anon_sym_RBRACE3,
    ACTIONS(14662), 1,
      sym__expansion_word,
    STATE(6475), 1,
      aux_sym__concatenation_in_expansion_repeat1,
    ACTIONS(14590), 2,
      sym__concat,
      aux_sym_concatenation_token1,
  [312362] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9296), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9298), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(9300), 1,
      anon_sym_BQUOTE,
    ACTIONS(9302), 1,
      anon_sym_DOLLAR_BQUOTE,
    STATE(2852), 2,
      sym_expansion,
      sym_command_substitution,
  [312382] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9000), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9002), 1,
      anon_sym_BQUOTE,
    ACTIONS(9004), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(14664), 1,
      anon_sym_DOLLAR_LPAREN,
    STATE(2257), 2,
      sym_expansion,
      sym_command_substitution,
  [312402] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(4516), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(4520), 1,
      anon_sym_BQUOTE,
    ACTIONS(4522), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(14666), 1,
      anon_sym_DOLLAR_LPAREN,
    STATE(2576), 2,
      sym_expansion,
      sym_command_substitution,
  [312422] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9206), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9208), 1,
      anon_sym_BQUOTE,
    ACTIONS(9210), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(14668), 1,
      anon_sym_DOLLAR_LPAREN,
    STATE(1300), 2,
      sym_expansion,
      sym_command_substitution,
  [312442] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(8914), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8916), 1,
      anon_sym_BQUOTE,
    ACTIONS(8918), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(14670), 1,
      anon_sym_DOLLAR_LPAREN,
    STATE(1888), 2,
      sym_expansion,
      sym_command_substitution,
  [312462] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(8828), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8830), 1,
      anon_sym_BQUOTE,
    ACTIONS(8832), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(14672), 1,
      anon_sym_DOLLAR_LPAREN,
    STATE(4862), 2,
      sym_expansion,
      sym_command_substitution,
  [312482] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14674), 1,
      anon_sym_SLASH,
    ACTIONS(14676), 1,
      anon_sym_RBRACE3,
    ACTIONS(14678), 1,
      sym__expansion_word,
    STATE(6475), 1,
      aux_sym__concatenation_in_expansion_repeat1,
    ACTIONS(14590), 2,
      sym__concat,
      aux_sym_concatenation_token1,
  [312502] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9102), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9104), 1,
      anon_sym_BQUOTE,
    ACTIONS(9106), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(14680), 1,
      anon_sym_DOLLAR_LPAREN,
    STATE(984), 2,
      sym_expansion,
      sym_command_substitution,
  [312522] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(721), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(725), 1,
      anon_sym_BQUOTE,
    ACTIONS(727), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(14682), 1,
      anon_sym_DOLLAR_LPAREN,
    STATE(964), 2,
      sym_expansion,
      sym_command_substitution,
  [312542] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12241), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(12243), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(12245), 1,
      anon_sym_BQUOTE,
    ACTIONS(12247), 1,
      anon_sym_DOLLAR_BQUOTE,
    STATE(3021), 2,
      sym_expansion,
      sym_command_substitution,
  [312562] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(8934), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8936), 1,
      anon_sym_BQUOTE,
    ACTIONS(8938), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(14684), 1,
      anon_sym_DOLLAR_LPAREN,
    STATE(1783), 2,
      sym_expansion,
      sym_command_substitution,
  [312582] = 6,
    ACTIONS(61), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(65), 1,
      anon_sym_BQUOTE,
    ACTIONS(67), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14686), 1,
      anon_sym_DOLLAR_LPAREN,
    STATE(1113), 2,
      sym_expansion,
      sym_command_substitution,
  [312602] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(328), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(332), 1,
      anon_sym_BQUOTE,
    ACTIONS(334), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(14688), 1,
      anon_sym_DOLLAR_LPAREN,
    STATE(1058), 2,
      sym_expansion,
      sym_command_substitution,
  [312622] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12179), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(12181), 1,
      anon_sym_DOLLAR_LPAREN,
    ACTIONS(12183), 1,
      anon_sym_BQUOTE,
    ACTIONS(12185), 1,
      anon_sym_DOLLAR_BQUOTE,
    STATE(3282), 2,
      sym_expansion,
      sym_command_substitution,
  [312642] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(14692), 2,
      sym_regex,
      aux_sym__expansion_regex_token1,
    ACTIONS(14690), 4,
      anon_sym_RPAREN,
      anon_sym_DQUOTE,
      sym_raw_string,
      anon_sym_RBRACE3,
  [312656] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9554), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(9556), 1,
      anon_sym_BQUOTE,
    ACTIONS(9558), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(14694), 1,
      anon_sym_DOLLAR_LPAREN,
    STATE(5628), 2,
      sym_expansion,
      sym_command_substitution,
  [312676] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14488), 1,
      anon_sym_elif,
    ACTIONS(14490), 1,
      anon_sym_else,
    ACTIONS(14696), 1,
      anon_sym_fi,
    STATE(7417), 1,
      sym_else_clause,
    STATE(6400), 2,
      sym_elif_clause,
      aux_sym_if_statement_repeat1,
  [312696] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14564), 1,
      anon_sym_PIPE,
    ACTIONS(14698), 1,
      anon_sym_RPAREN,
    STATE(6373), 1,
      aux_sym_concatenation_repeat1,
    STATE(6598), 1,
      aux_sym_case_item_repeat1,
    ACTIONS(14568), 2,
      sym__concat,
      aux_sym_concatenation_token1,
  [312716] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(14500), 1,
      anon_sym_COMMA,
    ACTIONS(14702), 1,
      aux_sym_heredoc_redirect_token1,
    STATE(4241), 1,
      sym__c_terminator,
    STATE(6319), 1,
      aux_sym__for_body_repeat1,
    ACTIONS(14700), 2,
      anon_sym_SEMI,
      anon_sym_AMP,
  [312736] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(5126), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(5130), 1,
      anon_sym_BQUOTE,
    ACTIONS(5132), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(14704), 1,
      anon_sym_DOLLAR_LPAREN,
    STATE(2870), 2,
      sym_expansion,
      sym_command_substitution,
  [312756] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(8488), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(8492), 1,
      anon_sym_BQUOTE,
    ACTIONS(8494), 1,
      anon_sym_DOLLAR_BQUOTE,
    ACTIONS(14706), 1,
      anon_sym_DOLLAR_LPAREN,
    STATE(1683), 2,
      sym_expansion,
      sym_command_substitution,
  [312776] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14708), 1,
      sym_simple_heredoc_body,
    ACTIONS(14710), 1,
      sym__heredoc_body_beginning,
    STATE(5099), 1,
      sym__heredoc_body,
    STATE(5100), 1,
      sym__simple_heredoc_body,
    STATE(6796), 1,
      sym_heredoc_body,
  [312795] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14710), 1,
      sym__heredoc_body_beginning,
    ACTIONS(14712), 1,
      sym_simple_heredoc_body,
    STATE(5630), 1,
      sym__simple_heredoc_body,
    STATE(5631), 1,
      sym__heredoc_body,
    STATE(7466), 1,
      sym_heredoc_body,
  [312814] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14714), 1,
      sym__concat,
    ACTIONS(6836), 4,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
      anon_sym_RBRACE3,
  [312827] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14710), 1,
      sym__heredoc_body_beginning,
    ACTIONS(14712), 1,
      sym_simple_heredoc_body,
    STATE(5641), 1,
      sym__simple_heredoc_body,
    STATE(5642), 1,
      sym__heredoc_body,
    STATE(7466), 1,
      sym_heredoc_body,
  [312846] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14716), 1,
      anon_sym_SLASH,
    ACTIONS(14718), 1,
      anon_sym_RBRACE3,
    STATE(6475), 1,
      aux_sym__concatenation_in_expansion_repeat1,
    ACTIONS(14590), 2,
      sym__concat,
      aux_sym_concatenation_token1,
  [312863] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14568), 1,
      aux_sym_concatenation_token1,
    ACTIONS(14720), 1,
      sym__concat,
    STATE(6419), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1267), 2,
      anon_sym_PIPE,
      anon_sym_RPAREN,
  [312880] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(14419), 1,
      anon_sym_in,
    ACTIONS(14423), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(14421), 3,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
  [312895] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14710), 1,
      sym__heredoc_body_beginning,
    ACTIONS(14722), 1,
      sym_simple_heredoc_body,
    STATE(6845), 1,
      sym_heredoc_body,
    STATE(4892), 2,
      sym__heredoc_body,
      sym__simple_heredoc_body,
  [312912] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(14399), 1,
      anon_sym_in,
    ACTIONS(14403), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(14401), 3,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
  [312927] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14710), 1,
      sym__heredoc_body_beginning,
    ACTIONS(14722), 1,
      sym_simple_heredoc_body,
    STATE(4979), 1,
      sym__simple_heredoc_body,
    STATE(4983), 1,
      sym__heredoc_body,
    STATE(6845), 1,
      sym_heredoc_body,
  [312946] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14710), 1,
      sym__heredoc_body_beginning,
    ACTIONS(14722), 1,
      sym_simple_heredoc_body,
    STATE(4970), 1,
      sym__simple_heredoc_body,
    STATE(4973), 1,
      sym__heredoc_body,
    STATE(6845), 1,
      sym_heredoc_body,
  [312965] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14674), 1,
      anon_sym_SLASH,
    ACTIONS(14676), 1,
      anon_sym_RBRACE3,
    STATE(6475), 1,
      aux_sym__concatenation_in_expansion_repeat1,
    ACTIONS(14590), 2,
      sym__concat,
      aux_sym_concatenation_token1,
  [312982] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14710), 1,
      sym__heredoc_body_beginning,
    ACTIONS(14724), 1,
      sym_simple_heredoc_body,
    STATE(7519), 1,
      sym_heredoc_body,
    STATE(4890), 2,
      sym__heredoc_body,
      sym__simple_heredoc_body,
  [312999] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(14726), 1,
      anon_sym_in,
    ACTIONS(14730), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(14728), 3,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
  [313014] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14710), 1,
      sym__heredoc_body_beginning,
    ACTIONS(14724), 1,
      sym_simple_heredoc_body,
    STATE(7519), 1,
      sym_heredoc_body,
    STATE(4937), 2,
      sym__heredoc_body,
      sym__simple_heredoc_body,
  [313031] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14708), 1,
      sym_simple_heredoc_body,
    ACTIONS(14710), 1,
      sym__heredoc_body_beginning,
    STATE(5112), 1,
      sym__simple_heredoc_body,
    STATE(5120), 1,
      sym__heredoc_body,
    STATE(6796), 1,
      sym_heredoc_body,
  [313050] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14710), 1,
      sym__heredoc_body_beginning,
    ACTIONS(14722), 1,
      sym_simple_heredoc_body,
    STATE(6845), 1,
      sym_heredoc_body,
    STATE(4968), 2,
      sym__heredoc_body,
      sym__simple_heredoc_body,
  [313067] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14710), 1,
      sym__heredoc_body_beginning,
    ACTIONS(14712), 1,
      sym_simple_heredoc_body,
    STATE(5645), 1,
      sym__simple_heredoc_body,
    STATE(5646), 1,
      sym__heredoc_body,
    STATE(7466), 1,
      sym_heredoc_body,
  [313086] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(689), 1,
      anon_sym_LBRACE,
    ACTIONS(14732), 1,
      anon_sym_SEMI,
    ACTIONS(14734), 1,
      anon_sym_do,
    STATE(5061), 1,
      sym_do_group,
    STATE(5063), 1,
      sym_compound_statement,
  [313105] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14710), 1,
      sym__heredoc_body_beginning,
    ACTIONS(14724), 1,
      sym_simple_heredoc_body,
    STATE(4944), 1,
      sym__heredoc_body,
    STATE(4947), 1,
      sym__simple_heredoc_body,
    STATE(7519), 1,
      sym_heredoc_body,
  [313124] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14736), 1,
      sym__concat,
    ACTIONS(6842), 4,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
      anon_sym_RBRACE3,
  [313137] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14738), 1,
      anon_sym_RBRACE3,
    STATE(6474), 1,
      aux_sym__expansion_body_repeat1,
    ACTIONS(8050), 3,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
  [313152] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14738), 1,
      anon_sym_RBRACE3,
    STATE(6474), 1,
      aux_sym__expansion_body_repeat1,
    ACTIONS(8050), 3,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
  [313167] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14710), 1,
      sym__heredoc_body_beginning,
    ACTIONS(14712), 1,
      sym_simple_heredoc_body,
    STATE(7466), 1,
      sym_heredoc_body,
    STATE(5649), 2,
      sym__heredoc_body,
      sym__simple_heredoc_body,
  [313184] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14710), 1,
      sym__heredoc_body_beginning,
    ACTIONS(14724), 1,
      sym_simple_heredoc_body,
    STATE(4872), 1,
      sym__heredoc_body,
    STATE(4881), 1,
      sym__simple_heredoc_body,
    STATE(7519), 1,
      sym_heredoc_body,
  [313203] = 4,
    ACTIONS(71), 1,
      sym_comment,
    STATE(6373), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(14568), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(14740), 2,
      anon_sym_PIPE,
      anon_sym_RPAREN,
  [313218] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14738), 1,
      anon_sym_RBRACE3,
    STATE(6474), 1,
      aux_sym__expansion_body_repeat1,
    ACTIONS(8050), 3,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
  [313233] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14708), 1,
      sym_simple_heredoc_body,
    ACTIONS(14710), 1,
      sym__heredoc_body_beginning,
    STATE(6796), 1,
      sym_heredoc_body,
    STATE(5101), 2,
      sym__heredoc_body,
      sym__simple_heredoc_body,
  [313250] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(165), 1,
      anon_sym_LBRACE,
    ACTIONS(14742), 1,
      anon_sym_SEMI,
    ACTIONS(14744), 1,
      anon_sym_do,
    STATE(5673), 1,
      sym_do_group,
    STATE(5675), 1,
      sym_compound_statement,
  [313269] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14710), 1,
      sym__heredoc_body_beginning,
    ACTIONS(14724), 1,
      sym_simple_heredoc_body,
    STATE(7519), 1,
      sym_heredoc_body,
    STATE(4882), 2,
      sym__heredoc_body,
      sym__simple_heredoc_body,
  [313286] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14708), 1,
      sym_simple_heredoc_body,
    ACTIONS(14710), 1,
      sym__heredoc_body_beginning,
    STATE(5052), 1,
      sym__simple_heredoc_body,
    STATE(5105), 1,
      sym__heredoc_body,
    STATE(6796), 1,
      sym_heredoc_body,
  [313305] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14710), 1,
      sym__heredoc_body_beginning,
    ACTIONS(14712), 1,
      sym_simple_heredoc_body,
    STATE(7466), 1,
      sym_heredoc_body,
    STATE(5633), 2,
      sym__heredoc_body,
      sym__simple_heredoc_body,
  [313322] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14748), 1,
      anon_sym_elif,
    ACTIONS(14746), 2,
      anon_sym_fi,
      anon_sym_else,
    STATE(6400), 2,
      sym_elif_clause,
      aux_sym_if_statement_repeat1,
  [313337] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14710), 1,
      sym__heredoc_body_beginning,
    ACTIONS(14724), 1,
      sym_simple_heredoc_body,
    STATE(4929), 1,
      sym__heredoc_body,
    STATE(4936), 1,
      sym__simple_heredoc_body,
    STATE(7519), 1,
      sym_heredoc_body,
  [313356] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14751), 1,
      anon_sym_RBRACE3,
    STATE(6474), 1,
      aux_sym__expansion_body_repeat1,
    ACTIONS(8050), 3,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
  [313371] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14753), 1,
      sym__concat,
    ACTIONS(6827), 4,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
      anon_sym_RBRACE3,
  [313384] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14708), 1,
      sym_simple_heredoc_body,
    ACTIONS(14710), 1,
      sym__heredoc_body_beginning,
    STATE(6796), 1,
      sym_heredoc_body,
    STATE(5103), 2,
      sym__heredoc_body,
      sym__simple_heredoc_body,
  [313401] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14564), 1,
      anon_sym_PIPE,
    ACTIONS(14698), 1,
      anon_sym_RPAREN,
    ACTIONS(14755), 1,
      sym__special_character,
    STATE(6545), 1,
      aux_sym__literal_repeat1,
    STATE(6595), 1,
      aux_sym_case_item_repeat1,
  [313420] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14710), 1,
      sym__heredoc_body_beginning,
    ACTIONS(14757), 1,
      sym_simple_heredoc_body,
    STATE(7011), 1,
      sym_heredoc_body,
    STATE(5568), 2,
      sym__heredoc_body,
      sym__simple_heredoc_body,
  [313437] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14759), 1,
      sym__concat,
    ACTIONS(6818), 4,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
      anon_sym_RBRACE3,
  [313450] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(14761), 1,
      anon_sym_in,
    ACTIONS(14765), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(14763), 3,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
  [313465] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14564), 1,
      anon_sym_PIPE,
    ACTIONS(14600), 1,
      anon_sym_RPAREN,
    ACTIONS(14755), 1,
      sym__special_character,
    STATE(6545), 1,
      aux_sym__literal_repeat1,
    STATE(6627), 1,
      aux_sym_case_item_repeat1,
  [313484] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14767), 1,
      anon_sym_SLASH,
    ACTIONS(14769), 1,
      anon_sym_RBRACE3,
    STATE(6475), 1,
      aux_sym__concatenation_in_expansion_repeat1,
    ACTIONS(14590), 2,
      sym__concat,
      aux_sym_concatenation_token1,
  [313501] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14710), 1,
      sym__heredoc_body_beginning,
    ACTIONS(14712), 1,
      sym_simple_heredoc_body,
    STATE(7466), 1,
      sym_heredoc_body,
    STATE(5653), 2,
      sym__heredoc_body,
      sym__simple_heredoc_body,
  [313518] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14708), 1,
      sym_simple_heredoc_body,
    ACTIONS(14710), 1,
      sym__heredoc_body_beginning,
    STATE(5067), 1,
      sym__heredoc_body,
    STATE(5068), 1,
      sym__simple_heredoc_body,
    STATE(6796), 1,
      sym_heredoc_body,
  [313537] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(14405), 1,
      anon_sym_in,
    ACTIONS(14409), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(14407), 3,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
  [313552] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14710), 1,
      sym__heredoc_body_beginning,
    ACTIONS(14724), 1,
      sym_simple_heredoc_body,
    STATE(7519), 1,
      sym_heredoc_body,
    STATE(4922), 2,
      sym__heredoc_body,
      sym__simple_heredoc_body,
  [313569] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14564), 1,
      anon_sym_PIPE,
    ACTIONS(14584), 1,
      anon_sym_RPAREN,
    ACTIONS(14755), 1,
      sym__special_character,
    STATE(6545), 1,
      aux_sym__literal_repeat1,
    STATE(6703), 1,
      aux_sym_case_item_repeat1,
  [313588] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14710), 1,
      sym__heredoc_body_beginning,
    ACTIONS(14757), 1,
      sym_simple_heredoc_body,
    STATE(7011), 1,
      sym_heredoc_body,
    STATE(5571), 2,
      sym__heredoc_body,
      sym__simple_heredoc_body,
  [313605] = 6,
    ACTIONS(29), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14771), 1,
      anon_sym_SEMI,
    ACTIONS(14773), 1,
      anon_sym_do,
    STATE(5005), 1,
      sym_do_group,
    STATE(5009), 1,
      sym_compound_statement,
  [313624] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14710), 1,
      sym__heredoc_body_beginning,
    ACTIONS(14724), 1,
      sym_simple_heredoc_body,
    STATE(7519), 1,
      sym_heredoc_body,
    STATE(4932), 2,
      sym__heredoc_body,
      sym__simple_heredoc_body,
  [313641] = 4,
    ACTIONS(71), 1,
      sym_comment,
    STATE(6419), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1253), 2,
      anon_sym_PIPE,
      anon_sym_RPAREN,
    ACTIONS(14775), 2,
      sym__concat,
      aux_sym_concatenation_token1,
  [313656] = 4,
    ACTIONS(71), 1,
      sym_comment,
    STATE(6439), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(14568), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(14778), 2,
      anon_sym_PIPE,
      anon_sym_RPAREN,
  [313671] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14710), 1,
      sym__heredoc_body_beginning,
    ACTIONS(14712), 1,
      sym_simple_heredoc_body,
    STATE(5635), 1,
      sym__simple_heredoc_body,
    STATE(5636), 1,
      sym__heredoc_body,
    STATE(7466), 1,
      sym_heredoc_body,
  [313690] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14710), 1,
      sym__heredoc_body_beginning,
    ACTIONS(14722), 1,
      sym_simple_heredoc_body,
    STATE(5002), 1,
      sym__heredoc_body,
    STATE(5011), 1,
      sym__simple_heredoc_body,
    STATE(6845), 1,
      sym_heredoc_body,
  [313709] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14710), 1,
      sym__heredoc_body_beginning,
    ACTIONS(14722), 1,
      sym_simple_heredoc_body,
    STATE(6845), 1,
      sym_heredoc_body,
    STATE(4915), 2,
      sym__heredoc_body,
      sym__simple_heredoc_body,
  [313726] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14751), 1,
      anon_sym_RBRACE3,
    STATE(6474), 1,
      aux_sym__expansion_body_repeat1,
    ACTIONS(8050), 3,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
  [313741] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14780), 1,
      anon_sym_SLASH,
    ACTIONS(14782), 1,
      anon_sym_RBRACE3,
    STATE(6475), 1,
      aux_sym__concatenation_in_expansion_repeat1,
    ACTIONS(14590), 2,
      sym__concat,
      aux_sym_concatenation_token1,
  [313758] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14564), 1,
      anon_sym_PIPE,
    ACTIONS(14580), 1,
      anon_sym_RPAREN,
    ACTIONS(14755), 1,
      sym__special_character,
    STATE(6545), 1,
      aux_sym__literal_repeat1,
    STATE(6694), 1,
      aux_sym_case_item_repeat1,
  [313777] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14658), 1,
      anon_sym_SLASH,
    ACTIONS(14660), 1,
      anon_sym_RBRACE3,
    STATE(6475), 1,
      aux_sym__concatenation_in_expansion_repeat1,
    ACTIONS(14590), 2,
      sym__concat,
      aux_sym_concatenation_token1,
  [313794] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14784), 1,
      anon_sym_RBRACE3,
    STATE(6474), 1,
      aux_sym__expansion_body_repeat1,
    ACTIONS(8050), 3,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
  [313809] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14710), 1,
      sym__heredoc_body_beginning,
    ACTIONS(14757), 1,
      sym_simple_heredoc_body,
    STATE(7011), 1,
      sym_heredoc_body,
    STATE(5574), 2,
      sym__heredoc_body,
      sym__simple_heredoc_body,
  [313826] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14710), 1,
      sym__heredoc_body_beginning,
    ACTIONS(14724), 1,
      sym_simple_heredoc_body,
    STATE(4923), 1,
      sym__heredoc_body,
    STATE(4926), 1,
      sym__simple_heredoc_body,
    STATE(7519), 1,
      sym_heredoc_body,
  [313845] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14708), 1,
      sym_simple_heredoc_body,
    ACTIONS(14710), 1,
      sym__heredoc_body_beginning,
    STATE(5087), 1,
      sym__heredoc_body,
    STATE(5098), 1,
      sym__simple_heredoc_body,
    STATE(6796), 1,
      sym_heredoc_body,
  [313864] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14710), 1,
      sym__heredoc_body_beginning,
    ACTIONS(14757), 1,
      sym_simple_heredoc_body,
    STATE(5576), 1,
      sym__simple_heredoc_body,
    STATE(5600), 1,
      sym__heredoc_body,
    STATE(7011), 1,
      sym_heredoc_body,
  [313883] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14708), 1,
      sym_simple_heredoc_body,
    ACTIONS(14710), 1,
      sym__heredoc_body_beginning,
    STATE(5072), 1,
      sym__heredoc_body,
    STATE(5077), 1,
      sym__simple_heredoc_body,
    STATE(6796), 1,
      sym_heredoc_body,
  [313902] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14708), 1,
      sym_simple_heredoc_body,
    ACTIONS(14710), 1,
      sym__heredoc_body_beginning,
    STATE(6796), 1,
      sym_heredoc_body,
    STATE(5083), 2,
      sym__heredoc_body,
      sym__simple_heredoc_body,
  [313919] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1310), 5,
      sym__concat,
      sym__expansion_word,
      anon_sym_SLASH,
      aux_sym_concatenation_token1,
      anon_sym_RBRACE3,
  [313930] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1306), 5,
      sym__concat,
      sym__expansion_word,
      anon_sym_SLASH,
      aux_sym_concatenation_token1,
      anon_sym_RBRACE3,
  [313941] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14710), 1,
      sym__heredoc_body_beginning,
    ACTIONS(14757), 1,
      sym_simple_heredoc_body,
    STATE(5578), 1,
      sym__heredoc_body,
    STATE(5579), 1,
      sym__simple_heredoc_body,
    STATE(7011), 1,
      sym_heredoc_body,
  [313960] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1310), 5,
      sym__concat,
      sym__expansion_word,
      anon_sym_SLASH,
      aux_sym_concatenation_token1,
      anon_sym_RBRACE3,
  [313971] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14568), 1,
      aux_sym_concatenation_token1,
    ACTIONS(14786), 1,
      sym__concat,
    STATE(6419), 1,
      aux_sym_concatenation_repeat1,
    ACTIONS(1288), 2,
      anon_sym_PIPE,
      anon_sym_RPAREN,
  [313988] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14710), 1,
      sym__heredoc_body_beginning,
    ACTIONS(14724), 1,
      sym_simple_heredoc_body,
    STATE(4867), 1,
      sym__heredoc_body,
    STATE(4908), 1,
      sym__simple_heredoc_body,
    STATE(7519), 1,
      sym_heredoc_body,
  [314007] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(433), 1,
      anon_sym_LBRACE,
    ACTIONS(14788), 1,
      anon_sym_SEMI,
    ACTIONS(14790), 1,
      anon_sym_do,
    STATE(4916), 1,
      sym_do_group,
    STATE(4927), 1,
      sym_compound_statement,
  [314026] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14710), 1,
      sym__heredoc_body_beginning,
    ACTIONS(14722), 1,
      sym_simple_heredoc_body,
    STATE(6845), 1,
      sym_heredoc_body,
    STATE(4989), 2,
      sym__heredoc_body,
      sym__simple_heredoc_body,
  [314043] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(14792), 1,
      anon_sym_in,
    ACTIONS(14796), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(14794), 3,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
  [314058] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14710), 1,
      sym__heredoc_body_beginning,
    ACTIONS(14724), 1,
      sym_simple_heredoc_body,
    STATE(4956), 1,
      sym__heredoc_body,
    STATE(4957), 1,
      sym__simple_heredoc_body,
    STATE(7519), 1,
      sym_heredoc_body,
  [314077] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14708), 1,
      sym_simple_heredoc_body,
    ACTIONS(14710), 1,
      sym__heredoc_body_beginning,
    STATE(6796), 1,
      sym_heredoc_body,
    STATE(5080), 2,
      sym__heredoc_body,
      sym__simple_heredoc_body,
  [314094] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14710), 1,
      sym__heredoc_body_beginning,
    ACTIONS(14712), 1,
      sym_simple_heredoc_body,
    STATE(5626), 1,
      sym__simple_heredoc_body,
    STATE(5627), 1,
      sym__heredoc_body,
    STATE(7466), 1,
      sym_heredoc_body,
  [314113] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14710), 1,
      sym__heredoc_body_beginning,
    ACTIONS(14724), 1,
      sym_simple_heredoc_body,
    STATE(7519), 1,
      sym_heredoc_body,
    STATE(4933), 2,
      sym__heredoc_body,
      sym__simple_heredoc_body,
  [314130] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(8038), 1,
      anon_sym_RBRACE3,
    STATE(6389), 1,
      aux_sym__expansion_body_repeat1,
    ACTIONS(8050), 3,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
  [314145] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14588), 1,
      anon_sym_SLASH,
    ACTIONS(14592), 1,
      anon_sym_RBRACE3,
    STATE(6475), 1,
      aux_sym__concatenation_in_expansion_repeat1,
    ACTIONS(14590), 2,
      sym__concat,
      aux_sym_concatenation_token1,
  [314162] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14710), 1,
      sym__heredoc_body_beginning,
    ACTIONS(14712), 1,
      sym_simple_heredoc_body,
    STATE(7466), 1,
      sym_heredoc_body,
    STATE(5639), 2,
      sym__heredoc_body,
      sym__simple_heredoc_body,
  [314179] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14710), 1,
      sym__heredoc_body_beginning,
    ACTIONS(14712), 1,
      sym_simple_heredoc_body,
    STATE(7466), 1,
      sym_heredoc_body,
    STATE(5640), 2,
      sym__heredoc_body,
      sym__simple_heredoc_body,
  [314196] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14708), 1,
      sym_simple_heredoc_body,
    ACTIONS(14710), 1,
      sym__heredoc_body_beginning,
    STATE(6796), 1,
      sym_heredoc_body,
    STATE(5035), 2,
      sym__heredoc_body,
      sym__simple_heredoc_body,
  [314213] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14710), 1,
      sym__heredoc_body_beginning,
    ACTIONS(14757), 1,
      sym_simple_heredoc_body,
    STATE(7011), 1,
      sym_heredoc_body,
    STATE(5581), 2,
      sym__heredoc_body,
      sym__simple_heredoc_body,
  [314230] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14710), 1,
      sym__heredoc_body_beginning,
    ACTIONS(14757), 1,
      sym_simple_heredoc_body,
    STATE(7011), 1,
      sym_heredoc_body,
    STATE(5585), 2,
      sym__heredoc_body,
      sym__simple_heredoc_body,
  [314247] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14710), 1,
      sym__heredoc_body_beginning,
    ACTIONS(14722), 1,
      sym_simple_heredoc_body,
    STATE(6845), 1,
      sym_heredoc_body,
    STATE(5022), 2,
      sym__heredoc_body,
      sym__simple_heredoc_body,
  [314264] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14710), 1,
      sym__heredoc_body_beginning,
    ACTIONS(14722), 1,
      sym_simple_heredoc_body,
    STATE(4935), 1,
      sym__heredoc_body,
    STATE(4948), 1,
      sym__simple_heredoc_body,
    STATE(6845), 1,
      sym_heredoc_body,
  [314283] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14710), 1,
      sym__heredoc_body_beginning,
    ACTIONS(14722), 1,
      sym_simple_heredoc_body,
    STATE(4900), 1,
      sym__heredoc_body,
    STATE(4920), 1,
      sym__simple_heredoc_body,
    STATE(6845), 1,
      sym_heredoc_body,
  [314302] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14710), 1,
      sym__heredoc_body_beginning,
    ACTIONS(14722), 1,
      sym_simple_heredoc_body,
    STATE(6845), 1,
      sym_heredoc_body,
    STATE(4961), 2,
      sym__heredoc_body,
      sym__simple_heredoc_body,
  [314319] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14710), 1,
      sym__heredoc_body_beginning,
    ACTIONS(14757), 1,
      sym_simple_heredoc_body,
    STATE(5565), 1,
      sym__heredoc_body,
    STATE(5589), 1,
      sym__simple_heredoc_body,
    STATE(7011), 1,
      sym_heredoc_body,
  [314338] = 4,
    ACTIONS(71), 1,
      sym_comment,
    STATE(6460), 1,
      aux_sym__concatenation_in_expansion_repeat1,
    ACTIONS(14798), 2,
      anon_sym_SLASH,
      anon_sym_RBRACE3,
    ACTIONS(14800), 2,
      sym__concat,
      aux_sym_concatenation_token1,
  [314353] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14710), 1,
      sym__heredoc_body_beginning,
    ACTIONS(14757), 1,
      sym_simple_heredoc_body,
    STATE(5586), 1,
      sym__simple_heredoc_body,
    STATE(5590), 1,
      sym__heredoc_body,
    STATE(7011), 1,
      sym_heredoc_body,
  [314372] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14710), 1,
      sym__heredoc_body_beginning,
    ACTIONS(14712), 1,
      sym_simple_heredoc_body,
    STATE(5617), 1,
      sym__simple_heredoc_body,
    STATE(5638), 1,
      sym__heredoc_body,
    STATE(7466), 1,
      sym_heredoc_body,
  [314391] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(14805), 1,
      anon_sym_COMMA,
    ACTIONS(14808), 1,
      aux_sym_heredoc_redirect_token1,
    STATE(6463), 1,
      aux_sym__for_body_repeat1,
    ACTIONS(14803), 2,
      anon_sym_SEMI,
      anon_sym_AMP,
  [314408] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14710), 1,
      sym__heredoc_body_beginning,
    ACTIONS(14722), 1,
      sym_simple_heredoc_body,
    STATE(4886), 1,
      sym__heredoc_body,
    STATE(4887), 1,
      sym__simple_heredoc_body,
    STATE(6845), 1,
      sym_heredoc_body,
  [314427] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14810), 1,
      anon_sym_RBRACE3,
    STATE(6474), 1,
      aux_sym__expansion_body_repeat1,
    ACTIONS(8050), 3,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
  [314442] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14710), 1,
      sym__heredoc_body_beginning,
    ACTIONS(14712), 1,
      sym_simple_heredoc_body,
    STATE(7466), 1,
      sym_heredoc_body,
    STATE(5658), 2,
      sym__heredoc_body,
      sym__simple_heredoc_body,
  [314459] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14710), 1,
      sym__heredoc_body_beginning,
    ACTIONS(14757), 1,
      sym_simple_heredoc_body,
    STATE(7011), 1,
      sym_heredoc_body,
    STATE(5595), 2,
      sym__heredoc_body,
      sym__simple_heredoc_body,
  [314476] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14710), 1,
      sym__heredoc_body_beginning,
    ACTIONS(14757), 1,
      sym_simple_heredoc_body,
    STATE(5596), 1,
      sym__heredoc_body,
    STATE(5597), 1,
      sym__simple_heredoc_body,
    STATE(7011), 1,
      sym_heredoc_body,
  [314495] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(14362), 1,
      anon_sym_in,
    ACTIONS(14366), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(14364), 3,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
  [314510] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14812), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(14445), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
  [314523] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(14814), 1,
      anon_sym_in,
    ACTIONS(14818), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(14816), 3,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_SEMI_SEMI,
  [314538] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(8548), 1,
      anon_sym_RBRACE3,
    STATE(6476), 1,
      aux_sym__expansion_body_repeat1,
    ACTIONS(8050), 3,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
  [314553] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14708), 1,
      sym_simple_heredoc_body,
    ACTIONS(14710), 1,
      sym__heredoc_body_beginning,
    STATE(6796), 1,
      sym_heredoc_body,
    STATE(5066), 2,
      sym__heredoc_body,
      sym__simple_heredoc_body,
  [314570] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14820), 1,
      anon_sym_RBRACE3,
    STATE(6474), 1,
      aux_sym__expansion_body_repeat1,
    ACTIONS(14822), 3,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
  [314585] = 4,
    ACTIONS(71), 1,
      sym_comment,
    STATE(6460), 1,
      aux_sym__concatenation_in_expansion_repeat1,
    ACTIONS(14590), 2,
      sym__concat,
      aux_sym_concatenation_token1,
    ACTIONS(14825), 2,
      anon_sym_SLASH,
      anon_sym_RBRACE3,
  [314600] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14827), 1,
      anon_sym_RBRACE3,
    STATE(6474), 1,
      aux_sym__expansion_body_repeat1,
    ACTIONS(8050), 3,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
  [314615] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14827), 1,
      anon_sym_RBRACE3,
    STATE(6474), 1,
      aux_sym__expansion_body_repeat1,
    ACTIONS(8050), 3,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
  [314630] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(8550), 1,
      anon_sym_RBRACE3,
    STATE(6424), 1,
      aux_sym__expansion_body_repeat1,
    ACTIONS(8050), 3,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
  [314645] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14710), 1,
      sym__heredoc_body_beginning,
    ACTIONS(14757), 1,
      sym_simple_heredoc_body,
    STATE(5598), 1,
      sym__heredoc_body,
    STATE(5599), 1,
      sym__simple_heredoc_body,
    STATE(7011), 1,
      sym_heredoc_body,
  [314664] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1338), 4,
      sym__concat,
      anon_sym_SLASH,
      aux_sym_concatenation_token1,
      anon_sym_RBRACE3,
  [314674] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(14808), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(14803), 3,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_AMP,
  [314686] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7787), 1,
      sym__special_character,
    ACTIONS(14780), 1,
      anon_sym_SLASH,
    ACTIONS(14782), 1,
      anon_sym_RBRACE3,
    STATE(6500), 1,
      aux_sym__literal_repeat1,
  [314702] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1310), 4,
      sym__concat,
      anon_sym_PIPE,
      anon_sym_RPAREN,
      aux_sym_concatenation_token1,
  [314712] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1306), 4,
      sym__concat,
      anon_sym_PIPE,
      anon_sym_RPAREN,
      aux_sym_concatenation_token1,
  [314722] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1310), 4,
      sym__concat,
      anon_sym_PIPE,
      anon_sym_RPAREN,
      aux_sym_concatenation_token1,
  [314732] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1326), 4,
      sym__concat,
      anon_sym_PIPE,
      anon_sym_RPAREN,
      aux_sym_concatenation_token1,
  [314742] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14798), 4,
      sym__concat,
      anon_sym_SLASH,
      aux_sym_concatenation_token1,
      anon_sym_RBRACE3,
  [314752] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7787), 1,
      sym__special_character,
    ACTIONS(14767), 1,
      anon_sym_SLASH,
    ACTIONS(14769), 1,
      anon_sym_RBRACE3,
    STATE(6500), 1,
      aux_sym__literal_repeat1,
  [314768] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14829), 4,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
      anon_sym_RBRACE3,
  [314778] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1298), 4,
      sym__concat,
      anon_sym_SLASH,
      aux_sym_concatenation_token1,
      anon_sym_RBRACE3,
  [314788] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14831), 1,
      anon_sym_esac,
    ACTIONS(14833), 1,
      anon_sym_SEMI_SEMI,
    ACTIONS(14835), 1,
      anon_sym_SEMI_AMP,
    ACTIONS(14837), 1,
      anon_sym_SEMI_SEMI_AMP,
  [314804] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14839), 1,
      anon_sym_esac,
    ACTIONS(14841), 1,
      anon_sym_SEMI_SEMI,
    ACTIONS(14843), 1,
      anon_sym_SEMI_AMP,
    ACTIONS(14845), 1,
      anon_sym_SEMI_SEMI_AMP,
  [314820] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(14849), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(14847), 3,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_AMP,
  [314832] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1253), 4,
      sym__concat,
      anon_sym_PIPE,
      anon_sym_RPAREN,
      aux_sym_concatenation_token1,
  [314842] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(689), 1,
      anon_sym_LBRACE,
    ACTIONS(14734), 1,
      anon_sym_do,
    STATE(5085), 1,
      sym_do_group,
    STATE(5088), 1,
      sym_compound_statement,
  [314858] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14831), 1,
      anon_sym_esac,
    ACTIONS(14851), 1,
      anon_sym_SEMI_SEMI,
    ACTIONS(14853), 1,
      anon_sym_SEMI_AMP,
    ACTIONS(14855), 1,
      anon_sym_SEMI_SEMI_AMP,
  [314874] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1314), 4,
      sym__concat,
      anon_sym_PIPE,
      anon_sym_RPAREN,
      aux_sym_concatenation_token1,
  [314884] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14857), 1,
      anon_sym_RBRACE3,
    STATE(6475), 1,
      aux_sym__concatenation_in_expansion_repeat1,
    ACTIONS(14590), 2,
      sym__concat,
      aux_sym_concatenation_token1,
  [314898] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1310), 4,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
      anon_sym_RBRACE3,
  [314908] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14859), 1,
      sym__special_character,
    STATE(6500), 1,
      aux_sym__literal_repeat1,
    ACTIONS(1364), 2,
      anon_sym_SLASH,
      anon_sym_RBRACE3,
  [314922] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14862), 1,
      anon_sym_esac,
    ACTIONS(14864), 1,
      anon_sym_SEMI_SEMI,
    ACTIONS(14866), 1,
      anon_sym_SEMI_AMP,
    ACTIONS(14868), 1,
      anon_sym_SEMI_SEMI_AMP,
  [314938] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1338), 4,
      sym__concat,
      anon_sym_PIPE,
      anon_sym_RPAREN,
      aux_sym_concatenation_token1,
  [314948] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1346), 4,
      sym__concat,
      anon_sym_PIPE,
      anon_sym_RPAREN,
      aux_sym_concatenation_token1,
  [314958] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1358), 4,
      sym__concat,
      anon_sym_PIPE,
      anon_sym_RPAREN,
      aux_sym_concatenation_token1,
  [314968] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(5772), 4,
      sym__concat,
      anon_sym_SLASH,
      aux_sym_concatenation_token1,
      anon_sym_RBRACE3,
  [314978] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1358), 4,
      sym__concat,
      anon_sym_SLASH,
      aux_sym_concatenation_token1,
      anon_sym_RBRACE3,
  [314988] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1346), 4,
      sym__concat,
      anon_sym_SLASH,
      aux_sym_concatenation_token1,
      anon_sym_RBRACE3,
  [314998] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(165), 1,
      anon_sym_LBRACE,
    ACTIONS(14744), 1,
      anon_sym_do,
    STATE(5687), 1,
      sym_do_group,
    STATE(5689), 1,
      sym_compound_statement,
  [315014] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1354), 4,
      sym__concat,
      anon_sym_PIPE,
      anon_sym_RPAREN,
      aux_sym_concatenation_token1,
  [315024] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1342), 4,
      sym__concat,
      anon_sym_PIPE,
      anon_sym_RPAREN,
      aux_sym_concatenation_token1,
  [315034] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14839), 1,
      anon_sym_esac,
    ACTIONS(14870), 1,
      anon_sym_SEMI_SEMI,
    ACTIONS(14872), 1,
      anon_sym_SEMI_AMP,
    ACTIONS(14874), 1,
      anon_sym_SEMI_SEMI_AMP,
  [315050] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(433), 1,
      anon_sym_LBRACE,
    ACTIONS(14790), 1,
      anon_sym_do,
    STATE(5003), 1,
      sym_do_group,
    STATE(5006), 1,
      sym_compound_statement,
  [315066] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1350), 4,
      sym__concat,
      anon_sym_PIPE,
      anon_sym_RPAREN,
      aux_sym_concatenation_token1,
  [315076] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1322), 4,
      sym__concat,
      anon_sym_SLASH,
      aux_sym_concatenation_token1,
      anon_sym_RBRACE3,
  [315086] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1342), 4,
      sym__concat,
      anon_sym_SLASH,
      aux_sym_concatenation_token1,
      anon_sym_RBRACE3,
  [315096] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1314), 4,
      sym__concat,
      anon_sym_SLASH,
      aux_sym_concatenation_token1,
      anon_sym_RBRACE3,
  [315106] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(5768), 4,
      sym__concat,
      anon_sym_SLASH,
      aux_sym_concatenation_token1,
      anon_sym_RBRACE3,
  [315116] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1354), 4,
      sym__concat,
      anon_sym_SLASH,
      aux_sym_concatenation_token1,
      anon_sym_RBRACE3,
  [315126] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1330), 4,
      sym__concat,
      anon_sym_PIPE,
      anon_sym_RPAREN,
      aux_sym_concatenation_token1,
  [315136] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1322), 4,
      sym__concat,
      anon_sym_PIPE,
      anon_sym_RPAREN,
      aux_sym_concatenation_token1,
  [315146] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14862), 1,
      anon_sym_esac,
    ACTIONS(14876), 1,
      anon_sym_SEMI_SEMI,
    ACTIONS(14878), 1,
      anon_sym_SEMI_AMP,
    ACTIONS(14880), 1,
      anon_sym_SEMI_SEMI_AMP,
  [315162] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1298), 4,
      sym__concat,
      anon_sym_PIPE,
      anon_sym_RPAREN,
      aux_sym_concatenation_token1,
  [315172] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1318), 4,
      sym__concat,
      anon_sym_PIPE,
      anon_sym_RPAREN,
      aux_sym_concatenation_token1,
  [315182] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14882), 1,
      anon_sym_esac,
    ACTIONS(14884), 1,
      anon_sym_SEMI_SEMI,
    ACTIONS(14886), 1,
      anon_sym_SEMI_AMP,
    ACTIONS(14888), 1,
      anon_sym_SEMI_SEMI_AMP,
  [315198] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1310), 4,
      sym__concat,
      anon_sym_SLASH,
      aux_sym_concatenation_token1,
      anon_sym_RBRACE3,
  [315208] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1334), 4,
      sym__concat,
      anon_sym_PIPE,
      anon_sym_RPAREN,
      aux_sym_concatenation_token1,
  [315218] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7160), 4,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
      anon_sym_RBRACE3,
  [315228] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6818), 4,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
      anon_sym_RBRACE3,
  [315238] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7164), 4,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
      anon_sym_RBRACE3,
  [315248] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7787), 1,
      sym__special_character,
    ACTIONS(14716), 1,
      anon_sym_SLASH,
    ACTIONS(14718), 1,
      anon_sym_RBRACE3,
    STATE(6500), 1,
      aux_sym__literal_repeat1,
  [315264] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(14892), 1,
      aux_sym_heredoc_redirect_token1,
    ACTIONS(14890), 3,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_AMP,
  [315276] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1306), 4,
      sym__concat,
      anon_sym_SLASH,
      aux_sym_concatenation_token1,
      anon_sym_RBRACE3,
  [315286] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14894), 1,
      anon_sym_esac,
    ACTIONS(14896), 1,
      anon_sym_SEMI_SEMI,
    ACTIONS(14898), 2,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
  [315300] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1302), 4,
      sym__concat,
      anon_sym_PIPE,
      anon_sym_RPAREN,
      aux_sym_concatenation_token1,
  [315310] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1294), 4,
      sym__concat,
      anon_sym_PIPE,
      anon_sym_RPAREN,
      aux_sym_concatenation_token1,
  [315320] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14900), 1,
      anon_sym_esac,
    ACTIONS(14902), 1,
      anon_sym_SEMI_SEMI,
    ACTIONS(14904), 2,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
  [315334] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1310), 4,
      sym__concat,
      anon_sym_SLASH,
      aux_sym_concatenation_token1,
      anon_sym_RBRACE3,
  [315344] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1306), 4,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
      anon_sym_RBRACE3,
  [315354] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14906), 1,
      anon_sym_esac,
    ACTIONS(14908), 1,
      anon_sym_SEMI_SEMI,
    ACTIONS(14910), 2,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
  [315368] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1310), 4,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
      anon_sym_RBRACE3,
  [315378] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14912), 1,
      anon_sym_esac,
    ACTIONS(14914), 1,
      anon_sym_SEMI_SEMI,
    ACTIONS(14916), 2,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
  [315392] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14882), 1,
      anon_sym_esac,
    ACTIONS(14918), 1,
      anon_sym_SEMI_SEMI,
    ACTIONS(14920), 1,
      anon_sym_SEMI_AMP,
    ACTIONS(14922), 1,
      anon_sym_SEMI_SEMI_AMP,
  [315408] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14755), 1,
      sym__special_character,
    STATE(6545), 1,
      aux_sym__literal_repeat1,
    ACTIONS(14740), 2,
      anon_sym_PIPE,
      anon_sym_RPAREN,
  [315422] = 5,
    ACTIONS(29), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14773), 1,
      anon_sym_do,
    STATE(4931), 1,
      sym_do_group,
    STATE(4938), 1,
      sym_compound_statement,
  [315438] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14924), 1,
      sym__special_character,
    STATE(6545), 1,
      aux_sym__literal_repeat1,
    ACTIONS(1364), 2,
      anon_sym_PIPE,
      anon_sym_RPAREN,
  [315452] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6827), 4,
      sym__external_expansion_sym_hash,
      sym__external_expansion_sym_bang,
      sym__external_expansion_sym_equal,
      anon_sym_RBRACE3,
  [315462] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14929), 1,
      anon_sym_LBRACK,
    ACTIONS(14927), 2,
      anon_sym_EQ,
      anon_sym_PLUS_EQ,
  [315473] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(14931), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6553), 1,
      aux_sym_arithmetic_expansion_repeat1,
  [315486] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14929), 1,
      anon_sym_LBRACK,
    ACTIONS(14933), 2,
      anon_sym_EQ,
      anon_sym_PLUS_EQ,
  [315497] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14808), 1,
      anon_sym_RPAREN_RPAREN,
    ACTIONS(14935), 1,
      anon_sym_COMMA,
    STATE(6550), 1,
      aux_sym__for_body_repeat1,
  [315510] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14938), 1,
      anon_sym_COMMA,
    ACTIONS(14940), 1,
      anon_sym_RPAREN,
    STATE(6640), 1,
      aux_sym__for_body_repeat1,
  [315523] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(14942), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6553), 1,
      aux_sym_arithmetic_expansion_repeat1,
  [315536] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7354), 1,
      anon_sym_RPAREN_RPAREN,
    ACTIONS(14944), 1,
      anon_sym_COMMA,
    STATE(6553), 1,
      aux_sym_arithmetic_expansion_repeat1,
  [315549] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1310), 3,
      sym_extglob_pattern,
      anon_sym_PIPE,
      anon_sym_RPAREN,
  [315558] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1306), 3,
      sym_extglob_pattern,
      anon_sym_PIPE,
      anon_sym_RPAREN,
  [315567] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1310), 3,
      sym_extglob_pattern,
      anon_sym_PIPE,
      anon_sym_RPAREN,
  [315576] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1354), 3,
      sym_extglob_pattern,
      anon_sym_PIPE,
      anon_sym_RPAREN,
  [315585] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12339), 1,
      anon_sym_PIPE,
    ACTIONS(14947), 1,
      anon_sym_PIPE_AMP,
    STATE(5102), 1,
      aux_sym_pipeline_repeat1,
  [315598] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(14949), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6553), 1,
      aux_sym_arithmetic_expansion_repeat1,
  [315611] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14929), 1,
      anon_sym_LBRACK,
    ACTIONS(14951), 2,
      anon_sym_EQ,
      anon_sym_PLUS_EQ,
  [315622] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1334), 3,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_RBRACE3,
  [315631] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(14953), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6553), 1,
      aux_sym_arithmetic_expansion_repeat1,
  [315644] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1314), 3,
      sym_extglob_pattern,
      anon_sym_PIPE,
      anon_sym_RPAREN,
  [315653] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1342), 3,
      sym_extglob_pattern,
      anon_sym_PIPE,
      anon_sym_RPAREN,
  [315662] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(14955), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6553), 1,
      aux_sym_arithmetic_expansion_repeat1,
  [315675] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12906), 1,
      anon_sym_PIPE,
    ACTIONS(14957), 1,
      anon_sym_PIPE_AMP,
    STATE(5499), 1,
      aux_sym_pipeline_repeat1,
  [315688] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14564), 1,
      anon_sym_PIPE,
    ACTIONS(14959), 1,
      anon_sym_RPAREN,
    STATE(6599), 1,
      aux_sym_case_item_repeat1,
  [315701] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(13001), 1,
      anon_sym_PIPE,
    ACTIONS(13003), 1,
      anon_sym_PIPE_AMP,
    STATE(5602), 1,
      aux_sym_pipeline_repeat1,
  [315714] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14564), 1,
      anon_sym_PIPE,
    ACTIONS(14566), 1,
      anon_sym_RPAREN,
    STATE(6567), 1,
      aux_sym_case_item_repeat1,
  [315727] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14564), 1,
      anon_sym_PIPE,
    ACTIONS(14961), 1,
      anon_sym_RPAREN,
    STATE(6599), 1,
      aux_sym_case_item_repeat1,
  [315740] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1338), 3,
      sym_extglob_pattern,
      anon_sym_PIPE,
      anon_sym_RPAREN,
  [315749] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12337), 1,
      anon_sym_PIPE,
    ACTIONS(14963), 1,
      anon_sym_PIPE_AMP,
    STATE(5095), 1,
      aux_sym_pipeline_repeat1,
  [315762] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1322), 3,
      sym_extglob_pattern,
      anon_sym_PIPE,
      anon_sym_RPAREN,
  [315771] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(14965), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6553), 1,
      aux_sym_arithmetic_expansion_repeat1,
  [315784] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14929), 1,
      anon_sym_LBRACK,
    ACTIONS(14967), 2,
      anon_sym_EQ,
      anon_sym_PLUS_EQ,
  [315795] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(14969), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6553), 1,
      aux_sym_arithmetic_expansion_repeat1,
  [315808] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(14971), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6553), 1,
      aux_sym_arithmetic_expansion_repeat1,
  [315821] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14929), 1,
      anon_sym_LBRACK,
    ACTIONS(14973), 2,
      anon_sym_EQ,
      anon_sym_PLUS_EQ,
  [315832] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(14975), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6553), 1,
      aux_sym_arithmetic_expansion_repeat1,
  [315845] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(14977), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6553), 1,
      aux_sym_arithmetic_expansion_repeat1,
  [315858] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1263), 3,
      anon_sym_PIPE,
      anon_sym_RPAREN,
      sym__special_character,
  [315867] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14979), 1,
      anon_sym_SLASH,
    ACTIONS(14981), 1,
      anon_sym_RBRACE3,
    ACTIONS(14983), 1,
      sym__regex_no_slash,
  [315880] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14985), 1,
      sym__concat,
    ACTIONS(6827), 2,
      anon_sym_EQ,
      anon_sym_PLUS_EQ,
  [315891] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12066), 1,
      anon_sym_PIPE,
    ACTIONS(14987), 1,
      anon_sym_PIPE_AMP,
    STATE(4824), 1,
      aux_sym_pipeline_repeat1,
  [315904] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14989), 1,
      sym__concat,
    ACTIONS(6818), 2,
      anon_sym_EQ,
      anon_sym_PLUS_EQ,
  [315915] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(14991), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6553), 1,
      aux_sym_arithmetic_expansion_repeat1,
  [315928] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14993), 1,
      anon_sym_RPAREN_RPAREN,
    ACTIONS(14995), 1,
      anon_sym_COMMA,
    STATE(6550), 1,
      aux_sym__for_body_repeat1,
  [315941] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(14997), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6553), 1,
      aux_sym_arithmetic_expansion_repeat1,
  [315954] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14938), 1,
      anon_sym_COMMA,
    ACTIONS(14999), 1,
      anon_sym_RPAREN,
    STATE(6679), 1,
      aux_sym__for_body_repeat1,
  [315967] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14564), 1,
      anon_sym_PIPE,
    ACTIONS(15001), 1,
      anon_sym_RPAREN,
    STATE(6599), 1,
      aux_sym_case_item_repeat1,
  [315980] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14564), 1,
      anon_sym_PIPE,
    ACTIONS(14618), 1,
      anon_sym_RPAREN,
    STATE(6676), 1,
      aux_sym_case_item_repeat1,
  [315993] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(15003), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6553), 1,
      aux_sym_arithmetic_expansion_repeat1,
  [316006] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(15005), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6553), 1,
      aux_sym_arithmetic_expansion_repeat1,
  [316019] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(15007), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6553), 1,
      aux_sym_arithmetic_expansion_repeat1,
  [316032] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14564), 1,
      anon_sym_PIPE,
    ACTIONS(15009), 1,
      anon_sym_RPAREN,
    STATE(6599), 1,
      aux_sym_case_item_repeat1,
  [316045] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14564), 1,
      anon_sym_PIPE,
    ACTIONS(15011), 1,
      anon_sym_RPAREN,
    STATE(6599), 1,
      aux_sym_case_item_repeat1,
  [316058] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14564), 1,
      anon_sym_PIPE,
    ACTIONS(15013), 1,
      anon_sym_RPAREN,
    STATE(6599), 1,
      aux_sym_case_item_repeat1,
  [316071] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14564), 1,
      anon_sym_PIPE,
    ACTIONS(15015), 1,
      anon_sym_RPAREN,
    STATE(6599), 1,
      aux_sym_case_item_repeat1,
  [316084] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15017), 1,
      anon_sym_PIPE,
    ACTIONS(15020), 1,
      anon_sym_RPAREN,
    STATE(6599), 1,
      aux_sym_case_item_repeat1,
  [316097] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(15022), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6553), 1,
      aux_sym_arithmetic_expansion_repeat1,
  [316110] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(15024), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6553), 1,
      aux_sym_arithmetic_expansion_repeat1,
  [316123] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(15026), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6553), 1,
      aux_sym_arithmetic_expansion_repeat1,
  [316136] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14929), 1,
      anon_sym_LBRACK,
    ACTIONS(15028), 2,
      anon_sym_EQ,
      anon_sym_PLUS_EQ,
  [316147] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14929), 1,
      anon_sym_LBRACK,
    ACTIONS(15030), 2,
      anon_sym_EQ,
      anon_sym_PLUS_EQ,
  [316158] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14938), 1,
      anon_sym_COMMA,
    ACTIONS(15032), 1,
      anon_sym_RPAREN,
    STATE(6589), 1,
      aux_sym__for_body_repeat1,
  [316171] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(15034), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6553), 1,
      aux_sym_arithmetic_expansion_repeat1,
  [316184] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(15036), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6553), 1,
      aux_sym_arithmetic_expansion_repeat1,
  [316197] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(15038), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6553), 1,
      aux_sym_arithmetic_expansion_repeat1,
  [316210] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15040), 1,
      sym__concat,
    ACTIONS(6836), 2,
      anon_sym_EQ,
      anon_sym_PLUS_EQ,
  [316221] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15042), 1,
      sym__concat,
    ACTIONS(6842), 2,
      anon_sym_EQ,
      anon_sym_PLUS_EQ,
  [316232] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15044), 1,
      anon_sym_SEMI_SEMI,
    ACTIONS(14910), 2,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
  [316243] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(15046), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6553), 1,
      aux_sym_arithmetic_expansion_repeat1,
  [316256] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(15048), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6553), 1,
      aux_sym_arithmetic_expansion_repeat1,
  [316269] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12304), 1,
      anon_sym_PIPE,
    ACTIONS(15050), 1,
      anon_sym_PIPE_AMP,
    STATE(5036), 1,
      aux_sym_pipeline_repeat1,
  [316282] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1294), 3,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_RBRACE3,
  [316291] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14929), 1,
      anon_sym_LBRACK,
    ACTIONS(15052), 2,
      anon_sym_EQ,
      anon_sym_PLUS_EQ,
  [316302] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(15054), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6553), 1,
      aux_sym_arithmetic_expansion_repeat1,
  [316315] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14995), 1,
      anon_sym_COMMA,
    ACTIONS(15056), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6681), 1,
      aux_sym__for_body_repeat1,
  [316328] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14929), 1,
      anon_sym_LBRACK,
    ACTIONS(15058), 2,
      anon_sym_EQ,
      anon_sym_PLUS_EQ,
  [316339] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14843), 1,
      anon_sym_SEMI_AMP,
    ACTIONS(14845), 1,
      anon_sym_SEMI_SEMI_AMP,
    ACTIONS(15060), 1,
      anon_sym_SEMI_SEMI,
  [316352] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(15062), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6553), 1,
      aux_sym_arithmetic_expansion_repeat1,
  [316365] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(15064), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6553), 1,
      aux_sym_arithmetic_expansion_repeat1,
  [316378] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(15066), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6553), 1,
      aux_sym_arithmetic_expansion_repeat1,
  [316391] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(15068), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6553), 1,
      aux_sym_arithmetic_expansion_repeat1,
  [316404] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(15070), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6553), 1,
      aux_sym_arithmetic_expansion_repeat1,
  [316417] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14995), 1,
      anon_sym_COMMA,
    ACTIONS(15072), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6550), 1,
      aux_sym__for_body_repeat1,
  [316430] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14564), 1,
      anon_sym_PIPE,
    ACTIONS(15074), 1,
      anon_sym_RPAREN,
    STATE(6599), 1,
      aux_sym_case_item_repeat1,
  [316443] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14929), 1,
      anon_sym_LBRACK,
    ACTIONS(15076), 2,
      anon_sym_EQ,
      anon_sym_PLUS_EQ,
  [316454] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(15078), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6553), 1,
      aux_sym_arithmetic_expansion_repeat1,
  [316467] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(15080), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6553), 1,
      aux_sym_arithmetic_expansion_repeat1,
  [316480] = 3,
    ACTIONS(71), 1,
      sym_comment,
    STATE(6475), 1,
      aux_sym__concatenation_in_expansion_repeat1,
    ACTIONS(14590), 2,
      sym__concat,
      aux_sym_concatenation_token1,
  [316491] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14995), 1,
      anon_sym_COMMA,
    ACTIONS(15082), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6587), 1,
      aux_sym__for_body_repeat1,
  [316504] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(15084), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6553), 1,
      aux_sym_arithmetic_expansion_repeat1,
  [316517] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14929), 1,
      anon_sym_LBRACK,
    ACTIONS(15086), 2,
      anon_sym_EQ,
      anon_sym_PLUS_EQ,
  [316528] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1354), 3,
      sym__regex_no_slash,
      anon_sym_SLASH,
      anon_sym_RBRACE3,
  [316537] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14929), 1,
      anon_sym_LBRACK,
    ACTIONS(15088), 2,
      anon_sym_EQ,
      anon_sym_PLUS_EQ,
  [316548] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14886), 1,
      anon_sym_SEMI_AMP,
    ACTIONS(14888), 1,
      anon_sym_SEMI_SEMI_AMP,
    ACTIONS(15090), 1,
      anon_sym_SEMI_SEMI,
  [316561] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14929), 1,
      anon_sym_LBRACK,
    ACTIONS(15092), 2,
      anon_sym_EQ,
      anon_sym_PLUS_EQ,
  [316572] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(15094), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6553), 1,
      aux_sym_arithmetic_expansion_repeat1,
  [316585] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14938), 1,
      anon_sym_COMMA,
    ACTIONS(15096), 1,
      anon_sym_RPAREN,
    STATE(6679), 1,
      aux_sym__for_body_repeat1,
  [316598] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14929), 1,
      anon_sym_LBRACK,
    ACTIONS(15098), 2,
      anon_sym_EQ,
      anon_sym_PLUS_EQ,
  [316609] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(15100), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6553), 1,
      aux_sym_arithmetic_expansion_repeat1,
  [316622] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(15102), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6553), 1,
      aux_sym_arithmetic_expansion_repeat1,
  [316635] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(15104), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6553), 1,
      aux_sym_arithmetic_expansion_repeat1,
  [316648] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1342), 3,
      sym__regex_no_slash,
      anon_sym_SLASH,
      anon_sym_RBRACE3,
  [316657] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14995), 1,
      anon_sym_COMMA,
    ACTIONS(15106), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6550), 1,
      aux_sym__for_body_repeat1,
  [316670] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(15108), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6553), 1,
      aux_sym_arithmetic_expansion_repeat1,
  [316683] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14995), 1,
      anon_sym_COMMA,
    ACTIONS(15110), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6626), 1,
      aux_sym__for_body_repeat1,
  [316696] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15112), 1,
      anon_sym_SEMI_SEMI,
    ACTIONS(14916), 2,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
  [316707] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(15114), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6553), 1,
      aux_sym_arithmetic_expansion_repeat1,
  [316720] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1261), 1,
      sym__special_character,
    ACTIONS(1263), 2,
      anon_sym_SLASH,
      anon_sym_RBRACE3,
  [316731] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14564), 1,
      anon_sym_PIPE,
    ACTIONS(15116), 1,
      anon_sym_RPAREN,
    STATE(6599), 1,
      aux_sym_case_item_repeat1,
  [316744] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(15118), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6553), 1,
      aux_sym_arithmetic_expansion_repeat1,
  [316757] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14995), 1,
      anon_sym_COMMA,
    ACTIONS(15120), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6550), 1,
      aux_sym__for_body_repeat1,
  [316770] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(15122), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6553), 1,
      aux_sym_arithmetic_expansion_repeat1,
  [316783] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(15124), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6553), 1,
      aux_sym_arithmetic_expansion_repeat1,
  [316796] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(15126), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6553), 1,
      aux_sym_arithmetic_expansion_repeat1,
  [316809] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14995), 1,
      anon_sym_COMMA,
    ACTIONS(15128), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6654), 1,
      aux_sym__for_body_repeat1,
  [316822] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14938), 1,
      anon_sym_COMMA,
    ACTIONS(15130), 1,
      anon_sym_RPAREN,
    STATE(6679), 1,
      aux_sym__for_body_repeat1,
  [316835] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(15132), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6553), 1,
      aux_sym_arithmetic_expansion_repeat1,
  [316848] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(15134), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6553), 1,
      aux_sym_arithmetic_expansion_repeat1,
  [316861] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14929), 1,
      anon_sym_LBRACK,
    ACTIONS(15136), 2,
      anon_sym_EQ,
      anon_sym_PLUS_EQ,
  [316872] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1322), 3,
      sym__regex_no_slash,
      anon_sym_SLASH,
      anon_sym_RBRACE3,
  [316881] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14929), 1,
      anon_sym_LBRACK,
    ACTIONS(15138), 2,
      anon_sym_EQ,
      anon_sym_PLUS_EQ,
  [316892] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14995), 1,
      anon_sym_COMMA,
    ACTIONS(15140), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6706), 1,
      aux_sym__for_body_repeat1,
  [316905] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15142), 1,
      anon_sym_SEMI_SEMI,
    ACTIONS(14904), 2,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
  [316916] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14872), 1,
      anon_sym_SEMI_AMP,
    ACTIONS(14874), 1,
      anon_sym_SEMI_SEMI_AMP,
    ACTIONS(15144), 1,
      anon_sym_SEMI_SEMI,
  [316929] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14929), 1,
      anon_sym_LBRACK,
    ACTIONS(15146), 2,
      anon_sym_EQ,
      anon_sym_PLUS_EQ,
  [316940] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(15148), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6553), 1,
      aux_sym_arithmetic_expansion_repeat1,
  [316953] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14995), 1,
      anon_sym_COMMA,
    ACTIONS(15150), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6550), 1,
      aux_sym__for_body_repeat1,
  [316966] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14938), 1,
      anon_sym_COMMA,
    ACTIONS(15152), 1,
      anon_sym_RPAREN,
    STATE(6659), 1,
      aux_sym__for_body_repeat1,
  [316979] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15154), 1,
      anon_sym_SEMI_SEMI,
    ACTIONS(14898), 2,
      anon_sym_SEMI_AMP,
      anon_sym_SEMI_SEMI_AMP,
  [316990] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14929), 1,
      anon_sym_LBRACK,
    ACTIONS(15156), 2,
      anon_sym_EQ,
      anon_sym_PLUS_EQ,
  [317001] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12203), 1,
      anon_sym_PIPE,
    ACTIONS(15158), 1,
      anon_sym_PIPE_AMP,
    STATE(4904), 1,
      aux_sym_pipeline_repeat1,
  [317014] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14995), 1,
      anon_sym_COMMA,
    ACTIONS(15160), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6646), 1,
      aux_sym__for_body_repeat1,
  [317027] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14564), 1,
      anon_sym_PIPE,
    ACTIONS(15162), 1,
      anon_sym_RPAREN,
    STATE(6599), 1,
      aux_sym_case_item_repeat1,
  [317040] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15164), 1,
      sym_extglob_pattern,
    ACTIONS(6749), 2,
      anon_sym_PIPE,
      anon_sym_RPAREN,
  [317051] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14995), 1,
      anon_sym_COMMA,
    ACTIONS(15166), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6714), 1,
      aux_sym__for_body_repeat1,
  [317064] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14808), 1,
      anon_sym_RPAREN,
    ACTIONS(15168), 1,
      anon_sym_COMMA,
    STATE(6679), 1,
      aux_sym__for_body_repeat1,
  [317077] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(15171), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6553), 1,
      aux_sym_arithmetic_expansion_repeat1,
  [317090] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14995), 1,
      anon_sym_COMMA,
    ACTIONS(15173), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6550), 1,
      aux_sym__for_body_repeat1,
  [317103] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(15175), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6553), 1,
      aux_sym_arithmetic_expansion_repeat1,
  [317116] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(15177), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6553), 1,
      aux_sym_arithmetic_expansion_repeat1,
  [317129] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(15179), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6553), 1,
      aux_sym_arithmetic_expansion_repeat1,
  [317142] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(15181), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6553), 1,
      aux_sym_arithmetic_expansion_repeat1,
  [317155] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14929), 1,
      anon_sym_LBRACK,
    ACTIONS(15183), 2,
      anon_sym_EQ,
      anon_sym_PLUS_EQ,
  [317166] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(15185), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6553), 1,
      aux_sym_arithmetic_expansion_repeat1,
  [317179] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(15187), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6553), 1,
      aux_sym_arithmetic_expansion_repeat1,
  [317192] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14995), 1,
      anon_sym_COMMA,
    ACTIONS(15189), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6550), 1,
      aux_sym__for_body_repeat1,
  [317205] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14920), 1,
      anon_sym_SEMI_AMP,
    ACTIONS(14922), 1,
      anon_sym_SEMI_SEMI_AMP,
    ACTIONS(15191), 1,
      anon_sym_SEMI_SEMI,
  [317218] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(15193), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6553), 1,
      aux_sym_arithmetic_expansion_repeat1,
  [317231] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(15195), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6553), 1,
      aux_sym_arithmetic_expansion_repeat1,
  [317244] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(15197), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6553), 1,
      aux_sym_arithmetic_expansion_repeat1,
  [317257] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14564), 1,
      anon_sym_PIPE,
    ACTIONS(15199), 1,
      anon_sym_RPAREN,
    STATE(6599), 1,
      aux_sym_case_item_repeat1,
  [317270] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14564), 1,
      anon_sym_PIPE,
    ACTIONS(14570), 1,
      anon_sym_RPAREN,
    STATE(6709), 1,
      aux_sym_case_item_repeat1,
  [317283] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14929), 1,
      anon_sym_LBRACK,
    ACTIONS(15201), 2,
      anon_sym_EQ,
      anon_sym_PLUS_EQ,
  [317294] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(15203), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6553), 1,
      aux_sym_arithmetic_expansion_repeat1,
  [317307] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(15205), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6553), 1,
      aux_sym_arithmetic_expansion_repeat1,
  [317320] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(15207), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6553), 1,
      aux_sym_arithmetic_expansion_repeat1,
  [317333] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14835), 1,
      anon_sym_SEMI_AMP,
    ACTIONS(14837), 1,
      anon_sym_SEMI_SEMI_AMP,
    ACTIONS(15209), 1,
      anon_sym_SEMI_SEMI,
  [317346] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(15211), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6553), 1,
      aux_sym_arithmetic_expansion_repeat1,
  [317359] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14995), 1,
      anon_sym_COMMA,
    ACTIONS(15213), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6670), 1,
      aux_sym__for_body_repeat1,
  [317372] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14564), 1,
      anon_sym_PIPE,
    ACTIONS(15215), 1,
      anon_sym_RPAREN,
    STATE(6599), 1,
      aux_sym_case_item_repeat1,
  [317385] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14929), 1,
      anon_sym_LBRACK,
    ACTIONS(15217), 2,
      anon_sym_EQ,
      anon_sym_PLUS_EQ,
  [317396] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14929), 1,
      anon_sym_LBRACK,
    ACTIONS(15219), 2,
      anon_sym_EQ,
      anon_sym_PLUS_EQ,
  [317407] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14995), 1,
      anon_sym_COMMA,
    ACTIONS(15221), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6550), 1,
      aux_sym__for_body_repeat1,
  [317420] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14564), 1,
      anon_sym_PIPE,
    ACTIONS(15223), 1,
      anon_sym_RPAREN,
    STATE(6599), 1,
      aux_sym_case_item_repeat1,
  [317433] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(15225), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6553), 1,
      aux_sym_arithmetic_expansion_repeat1,
  [317446] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14564), 1,
      anon_sym_PIPE,
    ACTIONS(15227), 1,
      anon_sym_RPAREN,
    STATE(6599), 1,
      aux_sym_case_item_repeat1,
  [317459] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14564), 1,
      anon_sym_PIPE,
    ACTIONS(15229), 1,
      anon_sym_RPAREN,
    STATE(6599), 1,
      aux_sym_case_item_repeat1,
  [317472] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14564), 1,
      anon_sym_PIPE,
    ACTIONS(15231), 1,
      anon_sym_RPAREN,
    STATE(6599), 1,
      aux_sym_case_item_repeat1,
  [317485] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(15233), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6553), 1,
      aux_sym_arithmetic_expansion_repeat1,
  [317498] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14995), 1,
      anon_sym_COMMA,
    ACTIONS(15235), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6689), 1,
      aux_sym__for_body_repeat1,
  [317511] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14995), 1,
      anon_sym_COMMA,
    ACTIONS(15237), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6550), 1,
      aux_sym__for_body_repeat1,
  [317524] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(15239), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6553), 1,
      aux_sym_arithmetic_expansion_repeat1,
  [317537] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14878), 1,
      anon_sym_SEMI_AMP,
    ACTIONS(14880), 1,
      anon_sym_SEMI_SEMI_AMP,
    ACTIONS(15241), 1,
      anon_sym_SEMI_SEMI,
  [317550] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14929), 1,
      anon_sym_LBRACK,
    ACTIONS(15243), 2,
      anon_sym_EQ,
      anon_sym_PLUS_EQ,
  [317561] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(12588), 1,
      anon_sym_PIPE,
    ACTIONS(15245), 1,
      anon_sym_PIPE_AMP,
    STATE(5224), 1,
      aux_sym_pipeline_repeat1,
  [317574] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14866), 1,
      anon_sym_SEMI_AMP,
    ACTIONS(14868), 1,
      anon_sym_SEMI_SEMI_AMP,
    ACTIONS(15247), 1,
      anon_sym_SEMI_SEMI,
  [317587] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(15249), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6553), 1,
      aux_sym_arithmetic_expansion_repeat1,
  [317600] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(15251), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6553), 1,
      aux_sym_arithmetic_expansion_repeat1,
  [317613] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14853), 1,
      anon_sym_SEMI_AMP,
    ACTIONS(14855), 1,
      anon_sym_SEMI_SEMI_AMP,
    ACTIONS(15253), 1,
      anon_sym_SEMI_SEMI,
  [317626] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(15255), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6553), 1,
      aux_sym_arithmetic_expansion_repeat1,
  [317639] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(15257), 1,
      anon_sym_RPAREN_RPAREN,
    STATE(6553), 1,
      aux_sym_arithmetic_expansion_repeat1,
  [317652] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14564), 1,
      anon_sym_PIPE,
    ACTIONS(14642), 1,
      anon_sym_RPAREN,
    STATE(6596), 1,
      aux_sym_case_item_repeat1,
  [317665] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15136), 2,
      anon_sym_EQ,
      anon_sym_PLUS_EQ,
  [317673] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15030), 2,
      anon_sym_EQ,
      anon_sym_PLUS_EQ,
  [317681] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14892), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
  [317689] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15259), 1,
      anon_sym_LPAREN_LPAREN,
    ACTIONS(15261), 1,
      aux_sym__simple_variable_name_token1,
  [317699] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14790), 1,
      anon_sym_do,
    STATE(4966), 1,
      sym_do_group,
  [317709] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15217), 2,
      anon_sym_EQ,
      anon_sym_PLUS_EQ,
  [317717] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14773), 1,
      anon_sym_do,
    STATE(4898), 1,
      sym_do_group,
  [317727] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7160), 2,
      anon_sym_EQ,
      anon_sym_PLUS_EQ,
  [317735] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(7164), 2,
      anon_sym_EQ,
      anon_sym_PLUS_EQ,
  [317743] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6799), 2,
      anon_sym_COLON,
      anon_sym_RBRACE3,
  [317751] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14927), 2,
      anon_sym_EQ,
      anon_sym_PLUS_EQ,
  [317759] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15138), 2,
      anon_sym_EQ,
      anon_sym_PLUS_EQ,
  [317767] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14892), 2,
      anon_sym_RPAREN_RPAREN,
      anon_sym_COMMA,
  [317775] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15201), 2,
      anon_sym_EQ,
      anon_sym_PLUS_EQ,
  [317783] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1302), 2,
      anon_sym_SLASH,
      anon_sym_RBRACE3,
  [317791] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14808), 2,
      anon_sym_RPAREN_RPAREN,
      anon_sym_COMMA,
  [317799] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14808), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
  [317807] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14734), 1,
      anon_sym_do,
    STATE(5089), 1,
      sym_do_group,
  [317817] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14773), 1,
      anon_sym_do,
    STATE(4940), 1,
      sym_do_group,
  [317827] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14734), 1,
      anon_sym_do,
    STATE(5104), 1,
      sym_do_group,
  [317837] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15183), 2,
      anon_sym_EQ,
      anon_sym_PLUS_EQ,
  [317845] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15263), 1,
      anon_sym_SLASH,
    ACTIONS(15265), 1,
      anon_sym_RBRACE3,
  [317855] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14767), 1,
      anon_sym_SLASH,
    ACTIONS(14769), 1,
      anon_sym_RBRACE3,
  [317865] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14979), 1,
      anon_sym_SLASH,
    ACTIONS(14981), 1,
      anon_sym_RBRACE3,
  [317875] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15156), 2,
      anon_sym_EQ,
      anon_sym_PLUS_EQ,
  [317883] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15088), 2,
      anon_sym_EQ,
      anon_sym_PLUS_EQ,
  [317891] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15267), 1,
      anon_sym_LPAREN_LPAREN,
    ACTIONS(15269), 1,
      aux_sym__simple_variable_name_token1,
  [317901] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1354), 2,
      anon_sym_SLASH,
      anon_sym_RBRACE3,
  [317909] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14933), 2,
      anon_sym_EQ,
      anon_sym_PLUS_EQ,
  [317917] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15271), 1,
      anon_sym_RBRACE3,
    ACTIONS(15273), 1,
      sym_regex,
  [317927] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15092), 2,
      anon_sym_EQ,
      anon_sym_PLUS_EQ,
  [317935] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15146), 2,
      anon_sym_EQ,
      anon_sym_PLUS_EQ,
  [317943] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15219), 2,
      anon_sym_EQ,
      anon_sym_PLUS_EQ,
  [317951] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15243), 2,
      anon_sym_EQ,
      anon_sym_PLUS_EQ,
  [317959] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15098), 2,
      anon_sym_EQ,
      anon_sym_PLUS_EQ,
  [317967] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1350), 2,
      anon_sym_SLASH,
      anon_sym_RBRACE3,
  [317975] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14849), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
  [317983] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14790), 1,
      anon_sym_do,
    STATE(5013), 1,
      sym_do_group,
  [317993] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14780), 1,
      anon_sym_SLASH,
    ACTIONS(14782), 1,
      anon_sym_RBRACE3,
  [318003] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1322), 2,
      anon_sym_SLASH,
      anon_sym_RBRACE3,
  [318011] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6862), 2,
      anon_sym_PIPE,
      anon_sym_RPAREN,
  [318019] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1342), 2,
      anon_sym_SLASH,
      anon_sym_RBRACE3,
  [318027] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14849), 2,
      anon_sym_RPAREN_RPAREN,
      anon_sym_COMMA,
  [318035] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15076), 2,
      anon_sym_EQ,
      anon_sym_PLUS_EQ,
  [318043] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14744), 1,
      anon_sym_do,
    STATE(5690), 1,
      sym_do_group,
  [318053] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15086), 2,
      anon_sym_EQ,
      anon_sym_PLUS_EQ,
  [318061] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15058), 2,
      anon_sym_EQ,
      anon_sym_PLUS_EQ,
  [318069] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14951), 2,
      anon_sym_EQ,
      anon_sym_PLUS_EQ,
  [318077] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14967), 2,
      anon_sym_EQ,
      anon_sym_PLUS_EQ,
  [318085] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14778), 2,
      anon_sym_PIPE,
      anon_sym_RPAREN,
  [318093] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6818), 2,
      anon_sym_EQ,
      anon_sym_PLUS_EQ,
  [318101] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15275), 1,
      anon_sym_LPAREN_LPAREN,
    ACTIONS(15277), 1,
      aux_sym__simple_variable_name_token1,
  [318111] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(6827), 2,
      anon_sym_EQ,
      anon_sym_PLUS_EQ,
  [318119] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15279), 1,
      anon_sym_SLASH,
    ACTIONS(15281), 1,
      anon_sym_RBRACE3,
  [318129] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14439), 1,
      anon_sym_COLON,
    ACTIONS(14441), 1,
      anon_sym_RBRACE3,
  [318139] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15028), 2,
      anon_sym_EQ,
      anon_sym_PLUS_EQ,
  [318147] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14658), 1,
      anon_sym_SLASH,
    ACTIONS(14660), 1,
      anon_sym_RBRACE3,
  [318157] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14744), 1,
      anon_sym_do,
    STATE(5661), 1,
      sym_do_group,
  [318167] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14716), 1,
      anon_sym_SLASH,
    ACTIONS(14718), 1,
      anon_sym_RBRACE3,
  [318177] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15283), 1,
      anon_sym_LPAREN_LPAREN,
    ACTIONS(15285), 1,
      aux_sym__simple_variable_name_token1,
  [318187] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14973), 2,
      anon_sym_EQ,
      anon_sym_PLUS_EQ,
  [318195] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14588), 1,
      anon_sym_SLASH,
    ACTIONS(14592), 1,
      anon_sym_RBRACE3,
  [318205] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14674), 1,
      anon_sym_SLASH,
    ACTIONS(14676), 1,
      anon_sym_RBRACE3,
  [318215] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15287), 1,
      anon_sym_SLASH,
    ACTIONS(15289), 1,
      anon_sym_RBRACE3,
  [318225] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15052), 2,
      anon_sym_EQ,
      anon_sym_PLUS_EQ,
  [318233] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15291), 1,
      anon_sym_SLASH,
    ACTIONS(15293), 1,
      anon_sym_RBRACE3,
  [318243] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(15295), 1,
      aux_sym_heredoc_redirect_token1,
  [318250] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15297), 1,
      aux_sym_brace_expression_token1,
  [318257] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15299), 1,
      anon_sym_BQUOTE,
  [318264] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15301), 1,
      anon_sym_BQUOTE,
  [318271] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15303), 1,
      sym_heredoc_end,
  [318278] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15305), 1,
      aux_sym_brace_expression_token1,
  [318285] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14478), 1,
      anon_sym_RBRACE3,
  [318292] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15307), 1,
      anon_sym_RPAREN,
  [318299] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15309), 1,
      anon_sym_RBRACE3,
  [318306] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15311), 1,
      anon_sym_RPAREN,
  [318313] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15313), 1,
      aux_sym_brace_expression_token1,
  [318320] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15315), 1,
      aux_sym_brace_expression_token1,
  [318327] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15317), 1,
      anon_sym_RPAREN,
  [318334] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15319), 1,
      anon_sym_LT_LT_LT,
  [318341] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(15321), 1,
      aux_sym_heredoc_redirect_token1,
  [318348] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15323), 1,
      anon_sym_RBRACE3,
  [318355] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15307), 1,
      anon_sym_BQUOTE,
  [318362] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15325), 1,
      anon_sym_RPAREN,
  [318369] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15327), 1,
      anon_sym_RBRACE3,
  [318376] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15329), 1,
      aux_sym_brace_expression_token1,
  [318383] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15331), 1,
      aux_sym_brace_expression_token1,
  [318390] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15333), 1,
      anon_sym_LT_LT_LT,
  [318397] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15335), 1,
      anon_sym_RBRACE2,
  [318404] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15337), 1,
      anon_sym_RPAREN,
  [318411] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15339), 1,
      anon_sym_BQUOTE,
  [318418] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15341), 1,
      anon_sym_RBRACE2,
  [318425] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15343), 1,
      anon_sym_RPAREN,
  [318432] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15337), 1,
      anon_sym_BQUOTE,
  [318439] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15345), 1,
      aux_sym_brace_expression_token1,
  [318446] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15347), 1,
      anon_sym_RPAREN,
  [318453] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15349), 1,
      anon_sym_BQUOTE,
  [318460] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15351), 1,
      anon_sym_BQUOTE,
  [318467] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15353), 1,
      anon_sym_esac,
  [318474] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14518), 1,
      anon_sym_RBRACE3,
  [318481] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15355), 1,
      sym_heredoc_end,
  [318488] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15357), 1,
      anon_sym_BQUOTE,
  [318495] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15359), 1,
      anon_sym_RBRACE2,
  [318502] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15361), 1,
      aux_sym_brace_expression_token1,
  [318509] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15363), 1,
      anon_sym_RBRACE2,
  [318516] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15349), 1,
      anon_sym_RPAREN,
  [318523] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15365), 1,
      anon_sym_RBRACE3,
  [318530] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(8838), 1,
      anon_sym_RBRACK,
  [318537] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(8846), 1,
      anon_sym_RBRACK,
  [318544] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(15367), 1,
      aux_sym_heredoc_redirect_token1,
  [318551] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(15369), 1,
      aux_sym_heredoc_redirect_token1,
  [318558] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(15371), 1,
      aux_sym_heredoc_redirect_token1,
  [318565] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15373), 1,
      aux_sym_brace_expression_token1,
  [318572] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(15375), 1,
      aux_sym_heredoc_redirect_token1,
  [318579] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15377), 1,
      anon_sym_BQUOTE,
  [318586] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15379), 1,
      anon_sym_RBRACE3,
  [318593] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15381), 1,
      anon_sym_RBRACE3,
  [318600] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(15383), 1,
      aux_sym_heredoc_redirect_token1,
  [318607] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(15385), 1,
      aux_sym_heredoc_redirect_token1,
  [318614] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15387), 1,
      sym_heredoc_end,
  [318621] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15389), 1,
      aux_sym_brace_expression_token1,
  [318628] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15391), 1,
      aux_sym_brace_expression_token1,
  [318635] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15393), 1,
      anon_sym_LT_LT_LT,
  [318642] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15395), 1,
      anon_sym_RPAREN,
  [318649] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15397), 1,
      anon_sym_BQUOTE,
  [318656] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15399), 1,
      anon_sym_esac,
  [318663] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15401), 1,
      anon_sym_RBRACE2,
  [318670] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(15403), 1,
      aux_sym_heredoc_redirect_token1,
  [318677] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15405), 1,
      anon_sym_BQUOTE,
  [318684] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15395), 1,
      anon_sym_BQUOTE,
  [318691] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15407), 1,
      aux_sym_brace_expression_token1,
  [318698] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15409), 1,
      anon_sym_RPAREN,
  [318705] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15411), 1,
      anon_sym_RBRACE2,
  [318712] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15413), 1,
      anon_sym_RPAREN,
  [318719] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15415), 1,
      anon_sym_RBRACE3,
  [318726] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15417), 1,
      anon_sym_RPAREN,
  [318733] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14626), 1,
      anon_sym_fi,
  [318740] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15419), 1,
      anon_sym_BQUOTE,
  [318747] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15417), 1,
      anon_sym_BQUOTE,
  [318754] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15421), 1,
      aux_sym_brace_expression_token1,
  [318761] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(15423), 1,
      aux_sym_heredoc_redirect_token1,
  [318768] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15425), 1,
      anon_sym_RPAREN,
  [318775] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(15427), 1,
      aux_sym_heredoc_redirect_token1,
  [318782] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15429), 1,
      anon_sym_RBRACE2,
  [318789] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(15431), 1,
      aux_sym_heredoc_redirect_token1,
  [318796] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(15433), 1,
      aux_sym_heredoc_redirect_token1,
  [318803] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15435), 1,
      anon_sym_RPAREN,
  [318810] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(15437), 1,
      aux_sym_heredoc_redirect_token1,
  [318817] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15439), 1,
      aux_sym_brace_expression_token1,
  [318824] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(15441), 1,
      aux_sym_heredoc_redirect_token1,
  [318831] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14827), 1,
      anon_sym_RBRACE3,
  [318838] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15443), 1,
      anon_sym_BQUOTE,
  [318845] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15445), 1,
      anon_sym_BQUOTE,
  [318852] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15443), 1,
      anon_sym_RPAREN,
  [318859] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15447), 1,
      anon_sym_BQUOTE,
  [318866] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15449), 1,
      anon_sym_BQUOTE,
  [318873] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15451), 1,
      anon_sym_in,
  [318880] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15453), 1,
      aux_sym_brace_expression_token1,
  [318887] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15455), 1,
      anon_sym_in,
  [318894] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15457), 1,
      anon_sym_RBRACE2,
  [318901] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15459), 1,
      anon_sym_RPAREN_RPAREN,
  [318908] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15461), 1,
      aux_sym_brace_expression_token1,
  [318915] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15447), 1,
      anon_sym_RPAREN,
  [318922] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15463), 1,
      anon_sym_LT_LT_LT,
  [318929] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15465), 1,
      anon_sym_RPAREN,
  [318936] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15467), 1,
      anon_sym_RBRACE3,
  [318943] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15469), 1,
      aux_sym_brace_expression_token1,
  [318950] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15471), 1,
      anon_sym_RPAREN,
  [318957] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15473), 1,
      sym_word,
  [318964] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15475), 1,
      anon_sym_RBRACE3,
  [318971] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15477), 1,
      anon_sym_RPAREN,
  [318978] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15479), 1,
      anon_sym_BQUOTE,
  [318985] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15481), 1,
      anon_sym_esac,
  [318992] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15483), 1,
      anon_sym_BQUOTE,
  [318999] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15485), 1,
      anon_sym_RBRACE3,
  [319006] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15487), 1,
      aux_sym_brace_expression_token1,
  [319013] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15489), 1,
      anon_sym_RPAREN,
  [319020] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15491), 1,
      anon_sym_BQUOTE,
  [319027] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15493), 1,
      anon_sym_RBRACE3,
  [319034] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15495), 1,
      anon_sym_BQUOTE,
  [319041] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15483), 1,
      anon_sym_RPAREN,
  [319048] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15489), 1,
      anon_sym_BQUOTE,
  [319055] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15497), 1,
      anon_sym_RBRACE3,
  [319062] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15499), 1,
      anon_sym_RPAREN,
  [319069] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15501), 1,
      aux_sym_brace_expression_token1,
  [319076] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15325), 1,
      anon_sym_BQUOTE,
  [319083] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15503), 1,
      anon_sym_RBRACE2,
  [319090] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15505), 1,
      anon_sym_RBRACK_RBRACK,
  [319097] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15507), 1,
      anon_sym_RPAREN,
  [319104] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15509), 1,
      anon_sym_RPAREN,
  [319111] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15511), 1,
      anon_sym_RBRACE2,
  [319118] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15513), 1,
      anon_sym_RPAREN,
  [319125] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15301), 1,
      anon_sym_RPAREN,
  [319132] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15515), 1,
      aux_sym_brace_expression_token1,
  [319139] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15517), 1,
      anon_sym_RBRACK_RBRACK,
  [319146] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15519), 1,
      anon_sym_fi,
  [319153] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15521), 1,
      anon_sym_RBRACE3,
  [319160] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15523), 1,
      anon_sym_esac,
  [319167] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15525), 1,
      anon_sym_RPAREN,
  [319174] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15527), 1,
      anon_sym_RBRACE3,
  [319181] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15529), 1,
      aux_sym_brace_expression_token1,
  [319188] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15531), 1,
      anon_sym_RPAREN,
  [319195] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(15533), 1,
      aux_sym_heredoc_redirect_token1,
  [319202] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15535), 1,
      anon_sym_BQUOTE,
  [319209] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15537), 1,
      anon_sym_BQUOTE,
  [319216] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(15539), 1,
      aux_sym_heredoc_redirect_token1,
  [319223] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15525), 1,
      anon_sym_BQUOTE,
  [319230] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15541), 1,
      aux_sym_brace_expression_token1,
  [319237] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15531), 1,
      anon_sym_BQUOTE,
  [319244] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(15543), 1,
      aux_sym_heredoc_redirect_token1,
  [319251] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(15545), 1,
      aux_sym_heredoc_redirect_token1,
  [319258] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15547), 1,
      anon_sym_RPAREN,
  [319265] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15549), 1,
      anon_sym_RPAREN,
  [319272] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(15551), 1,
      aux_sym_heredoc_redirect_token1,
  [319279] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15553), 1,
      aux_sym_brace_expression_token1,
  [319286] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(15555), 1,
      aux_sym_heredoc_redirect_token1,
  [319293] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15557), 1,
      anon_sym_esac,
  [319300] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15559), 1,
      anon_sym_BQUOTE,
  [319307] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(15561), 1,
      aux_sym_heredoc_redirect_token1,
  [319314] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(15563), 1,
      aux_sym_heredoc_redirect_token1,
  [319321] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15565), 1,
      anon_sym_RBRACE3,
  [319328] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15567), 1,
      aux_sym_brace_expression_token1,
  [319335] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15569), 1,
      anon_sym_in,
  [319342] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15571), 1,
      anon_sym_RBRACE2,
  [319349] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15573), 1,
      anon_sym_in,
  [319356] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15397), 1,
      anon_sym_RPAREN,
  [319363] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15575), 1,
      anon_sym_RPAREN_RPAREN,
  [319370] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15577), 1,
      aux_sym_brace_expression_token1,
  [319377] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15579), 1,
      aux_sym_brace_expression_token1,
  [319384] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15581), 1,
      anon_sym_LT_LT_LT,
  [319391] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15583), 1,
      anon_sym_RPAREN,
  [319398] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15585), 1,
      anon_sym_RBRACE3,
  [319405] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15587), 1,
      anon_sym_RBRACE3,
  [319412] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14738), 1,
      anon_sym_RBRACE3,
  [319419] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14738), 1,
      anon_sym_RBRACE3,
  [319426] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15589), 1,
      aux_sym_brace_expression_token1,
  [319433] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9218), 1,
      anon_sym_RBRACK,
  [319440] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15591), 1,
      anon_sym_RPAREN,
  [319447] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14738), 1,
      anon_sym_RBRACE3,
  [319454] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14738), 1,
      anon_sym_RBRACE3,
  [319461] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14738), 1,
      anon_sym_RBRACE3,
  [319468] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14738), 1,
      anon_sym_RBRACE3,
  [319475] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15593), 1,
      aux_sym_brace_expression_token1,
  [319482] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9220), 1,
      anon_sym_RBRACK,
  [319489] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15595), 1,
      sym_word,
  [319496] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15597), 1,
      anon_sym_RBRACE2,
  [319503] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14738), 1,
      anon_sym_RBRACE3,
  [319510] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14738), 1,
      anon_sym_RBRACE3,
  [319517] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14738), 1,
      anon_sym_RBRACE3,
  [319524] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15599), 1,
      aux_sym_brace_expression_token1,
  [319531] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15601), 1,
      anon_sym_BQUOTE,
  [319538] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14654), 1,
      anon_sym_RBRACE3,
  [319545] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14738), 1,
      anon_sym_RBRACE3,
  [319552] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14738), 1,
      anon_sym_RBRACE3,
  [319559] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14738), 1,
      anon_sym_RBRACE3,
  [319566] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14640), 1,
      anon_sym_fi,
  [319573] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15603), 1,
      aux_sym_brace_expression_token1,
  [319580] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15377), 1,
      anon_sym_RPAREN,
  [319587] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15605), 1,
      anon_sym_RBRACE3,
  [319594] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14751), 1,
      anon_sym_RBRACE3,
  [319601] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14751), 1,
      anon_sym_RBRACE3,
  [319608] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14751), 1,
      anon_sym_RBRACE3,
  [319615] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14751), 1,
      anon_sym_RBRACE3,
  [319622] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15607), 1,
      aux_sym_brace_expression_token1,
  [319629] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9080), 1,
      anon_sym_RBRACK,
  [319636] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15609), 1,
      anon_sym_esac,
  [319643] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14751), 1,
      anon_sym_RBRACE3,
  [319650] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14751), 1,
      anon_sym_RBRACE3,
  [319657] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14738), 1,
      anon_sym_RBRACE3,
  [319664] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14738), 1,
      anon_sym_RBRACE3,
  [319671] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15611), 1,
      aux_sym_brace_expression_token1,
  [319678] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15613), 1,
      anon_sym_esac,
  [319685] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(15615), 1,
      aux_sym_heredoc_redirect_token1,
  [319692] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14738), 1,
      anon_sym_RBRACE3,
  [319699] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14738), 1,
      anon_sym_RBRACE3,
  [319706] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14738), 1,
      anon_sym_RBRACE3,
  [319713] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14738), 1,
      anon_sym_RBRACE3,
  [319720] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15617), 1,
      aux_sym_brace_expression_token1,
  [319727] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14616), 1,
      anon_sym_fi,
  [319734] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15619), 1,
      anon_sym_esac,
  [319741] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14074), 1,
      anon_sym_RBRACE3,
  [319748] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15621), 1,
      anon_sym_esac,
  [319755] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14857), 1,
      anon_sym_RBRACE3,
  [319762] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15623), 1,
      sym_heredoc_end,
  [319769] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15625), 1,
      aux_sym_brace_expression_token1,
  [319776] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15627), 1,
      sym_heredoc_end,
  [319783] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15629), 1,
      anon_sym_fi,
  [319790] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15631), 1,
      anon_sym_esac,
  [319797] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15633), 1,
      anon_sym_RBRACE3,
  [319804] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15635), 1,
      anon_sym_RBRACE2,
  [319811] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15637), 1,
      anon_sym_esac,
  [319818] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15639), 1,
      aux_sym_brace_expression_token1,
  [319825] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9082), 1,
      anon_sym_RBRACK,
  [319832] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15641), 1,
      anon_sym_RBRACE2,
  [319839] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15643), 1,
      anon_sym_esac,
  [319846] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15645), 1,
      anon_sym_esac,
  [319853] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14474), 1,
      anon_sym_RBRACE3,
  [319860] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14482), 1,
      anon_sym_RBRACE3,
  [319867] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15647), 1,
      aux_sym_brace_expression_token1,
  [319874] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14482), 1,
      anon_sym_RBRACE3,
  [319881] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14482), 1,
      anon_sym_RBRACE3,
  [319888] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14482), 1,
      anon_sym_RBRACE3,
  [319895] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(15649), 1,
      aux_sym_heredoc_redirect_token1,
  [319902] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15651), 1,
      anon_sym_RBRACE3,
  [319909] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(15653), 1,
      aux_sym_heredoc_redirect_token1,
  [319916] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15655), 1,
      aux_sym_brace_expression_token1,
  [319923] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15657), 1,
      anon_sym_RPAREN,
  [319930] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15659), 1,
      anon_sym_BQUOTE,
  [319937] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15661), 1,
      anon_sym_BQUOTE,
  [319944] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15663), 1,
      anon_sym_esac,
  [319951] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(15665), 1,
      aux_sym_heredoc_redirect_token1,
  [319958] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15667), 1,
      anon_sym_RBRACK_RBRACK,
  [319965] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15669), 1,
      aux_sym_brace_expression_token1,
  [319972] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(15671), 1,
      aux_sym_heredoc_redirect_token1,
  [319979] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15673), 1,
      anon_sym_BQUOTE,
  [319986] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(15675), 1,
      aux_sym_heredoc_redirect_token1,
  [319993] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(15677), 1,
      aux_sym_heredoc_redirect_token1,
  [320000] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15679), 1,
      anon_sym_BQUOTE,
  [320007] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15681), 1,
      anon_sym_esac,
  [320014] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15683), 1,
      aux_sym_brace_expression_token1,
  [320021] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15685), 1,
      anon_sym_RBRACE3,
  [320028] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15661), 1,
      anon_sym_RPAREN,
  [320035] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15687), 1,
      anon_sym_RPAREN,
  [320042] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15689), 1,
      anon_sym_esac,
  [320049] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15691), 1,
      anon_sym_BQUOTE,
  [320056] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15693), 1,
      anon_sym_RBRACE3,
  [320063] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15695), 1,
      aux_sym_brace_expression_token1,
  [320070] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15697), 1,
      anon_sym_RPAREN,
  [320077] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15699), 1,
      anon_sym_RBRACE2,
  [320084] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15701), 1,
      anon_sym_in,
  [320091] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15703), 1,
      anon_sym_esac,
  [320098] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15705), 1,
      anon_sym_in,
  [320105] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15687), 1,
      anon_sym_BQUOTE,
  [320112] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15707), 1,
      aux_sym_brace_expression_token1,
  [320119] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15709), 1,
      anon_sym_RPAREN,
  [320126] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15711), 1,
      anon_sym_RPAREN_RPAREN,
  [320133] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15713), 1,
      aux_sym_brace_expression_token1,
  [320140] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15715), 1,
      anon_sym_LT_LT_LT,
  [320147] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15717), 1,
      anon_sym_RPAREN,
  [320154] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15719), 1,
      anon_sym_RBRACE3,
  [320161] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15721), 1,
      aux_sym_brace_expression_token1,
  [320168] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15723), 1,
      anon_sym_RPAREN,
  [320175] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15725), 1,
      sym_word,
  [320182] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14644), 1,
      anon_sym_fi,
  [320189] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15727), 1,
      anon_sym_BQUOTE,
  [320196] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15729), 1,
      anon_sym_RBRACE2,
  [320203] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15731), 1,
      anon_sym_BQUOTE,
  [320210] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15733), 1,
      aux_sym_brace_expression_token1,
  [320217] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15735), 1,
      anon_sym_RBRACE2,
  [320224] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15737), 1,
      anon_sym_BQUOTE,
  [320231] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15739), 1,
      anon_sym_RBRACE2,
  [320238] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15741), 1,
      anon_sym_BQUOTE,
  [320245] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15731), 1,
      anon_sym_RPAREN,
  [320252] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15743), 1,
      anon_sym_BQUOTE,
  [320259] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15745), 1,
      aux_sym_brace_expression_token1,
  [320266] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15727), 1,
      anon_sym_RPAREN,
  [320273] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15747), 1,
      anon_sym_RBRACE3,
  [320280] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15749), 1,
      anon_sym_BQUOTE,
  [320287] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15743), 1,
      anon_sym_RPAREN,
  [320294] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15751), 1,
      anon_sym_BQUOTE,
  [320301] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15753), 1,
      anon_sym_BQUOTE,
  [320308] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15755), 1,
      aux_sym_brace_expression_token1,
  [320315] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15751), 1,
      anon_sym_RPAREN,
  [320322] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15757), 1,
      anon_sym_RBRACE3,
  [320329] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15759), 1,
      anon_sym_RBRACE2,
  [320336] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15761), 1,
      anon_sym_BQUOTE,
  [320343] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15763), 1,
      anon_sym_RPAREN,
  [320350] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15765), 1,
      anon_sym_BQUOTE,
  [320357] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15767), 1,
      aux_sym_brace_expression_token1,
  [320364] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15769), 1,
      anon_sym_BQUOTE,
  [320371] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15771), 1,
      anon_sym_BQUOTE,
  [320378] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15769), 1,
      anon_sym_RPAREN,
  [320385] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15773), 1,
      anon_sym_RPAREN,
  [320392] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15775), 1,
      anon_sym_BQUOTE,
  [320399] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15777), 1,
      anon_sym_BQUOTE,
  [320406] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15779), 1,
      aux_sym_brace_expression_token1,
  [320413] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14512), 1,
      anon_sym_RBRACE3,
  [320420] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15775), 1,
      anon_sym_RPAREN,
  [320427] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15781), 1,
      anon_sym_RBRACE3,
  [320434] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15783), 1,
      anon_sym_RBRACE3,
  [320441] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15785), 1,
      anon_sym_RBRACE2,
  [320448] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15787), 1,
      anon_sym_BQUOTE,
  [320455] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15789), 1,
      aux_sym_brace_expression_token1,
  [320462] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15791), 1,
      anon_sym_BQUOTE,
  [320469] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15793), 1,
      anon_sym_BQUOTE,
  [320476] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15787), 1,
      anon_sym_RPAREN,
  [320483] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15795), 1,
      anon_sym_RBRACE3,
  [320490] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15797), 1,
      anon_sym_RBRACE3,
  [320497] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14646), 1,
      anon_sym_fi,
  [320504] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15799), 1,
      aux_sym_brace_expression_token1,
  [320511] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15801), 1,
      anon_sym_RBRACE2,
  [320518] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15803), 1,
      anon_sym_RPAREN,
  [320525] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15805), 1,
      anon_sym_BQUOTE,
  [320532] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15807), 1,
      anon_sym_BQUOTE,
  [320539] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15809), 1,
      anon_sym_RPAREN,
  [320546] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15811), 1,
      anon_sym_RBRACK_RBRACK,
  [320553] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15813), 1,
      aux_sym_brace_expression_token1,
  [320560] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15815), 1,
      anon_sym_BQUOTE,
  [320567] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15817), 1,
      anon_sym_BQUOTE,
  [320574] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15805), 1,
      anon_sym_RPAREN,
  [320581] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15819), 1,
      anon_sym_RBRACE3,
  [320588] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15815), 1,
      anon_sym_RPAREN,
  [320595] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15821), 1,
      anon_sym_RBRACE3,
  [320602] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15823), 1,
      aux_sym_brace_expression_token1,
  [320609] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15825), 1,
      anon_sym_BQUOTE,
  [320616] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15827), 1,
      anon_sym_RBRACE3,
  [320623] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15829), 1,
      anon_sym_BQUOTE,
  [320630] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15825), 1,
      anon_sym_RPAREN,
  [320637] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15831), 1,
      anon_sym_esac,
  [320644] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15833), 1,
      anon_sym_RBRACE3,
  [320651] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15835), 1,
      aux_sym_brace_expression_token1,
  [320658] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15837), 1,
      anon_sym_RBRACE3,
  [320665] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15839), 1,
      anon_sym_BQUOTE,
  [320672] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15841), 1,
      anon_sym_BQUOTE,
  [320679] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15839), 1,
      anon_sym_RPAREN,
  [320686] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15843), 1,
      anon_sym_RBRACE3,
  [320693] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15845), 1,
      sym_heredoc_end,
  [320700] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15847), 1,
      aux_sym_brace_expression_token1,
  [320707] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15761), 1,
      anon_sym_RPAREN,
  [320714] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15849), 1,
      anon_sym_RBRACE2,
  [320721] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15851), 1,
      anon_sym_RBRACE3,
  [320728] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15659), 1,
      anon_sym_RPAREN,
  [320735] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15853), 1,
      anon_sym_RPAREN,
  [320742] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15855), 1,
      anon_sym_RBRACE2,
  [320749] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15857), 1,
      aux_sym_brace_expression_token1,
  [320756] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15859), 1,
      anon_sym_BQUOTE,
  [320763] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15861), 1,
      anon_sym_BQUOTE,
  [320770] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15863), 1,
      anon_sym_BQUOTE,
  [320777] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15865), 1,
      anon_sym_RPAREN,
  [320784] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15867), 1,
      anon_sym_BQUOTE,
  [320791] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15869), 1,
      anon_sym_BQUOTE,
  [320798] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15871), 1,
      aux_sym_brace_expression_token1,
  [320805] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15853), 1,
      anon_sym_BQUOTE,
  [320812] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15873), 1,
      anon_sym_RBRACK_RBRACK,
  [320819] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15859), 1,
      anon_sym_RPAREN,
  [320826] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15867), 1,
      anon_sym_RPAREN,
  [320833] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15875), 1,
      anon_sym_RBRACE3,
  [320840] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15877), 1,
      anon_sym_RBRACE3,
  [320847] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15879), 1,
      aux_sym_brace_expression_token1,
  [320854] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15881), 1,
      anon_sym_RPAREN,
  [320861] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15883), 1,
      anon_sym_BQUOTE,
  [320868] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15885), 1,
      anon_sym_RPAREN,
  [320875] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15887), 1,
      anon_sym_BQUOTE,
  [320882] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15883), 1,
      anon_sym_RPAREN,
  [320889] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15889), 1,
      anon_sym_RBRACE3,
  [320896] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15891), 1,
      aux_sym_brace_expression_token1,
  [320903] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15893), 1,
      anon_sym_RBRACE2,
  [320910] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15895), 1,
      anon_sym_BQUOTE,
  [320917] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15897), 1,
      anon_sym_BQUOTE,
  [320924] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15895), 1,
      anon_sym_RPAREN,
  [320931] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15899), 1,
      anon_sym_RBRACE2,
  [320938] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9084), 1,
      anon_sym_RBRACK,
  [320945] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15901), 1,
      aux_sym_brace_expression_token1,
  [320952] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15903), 1,
      anon_sym_RBRACE3,
  [320959] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15905), 1,
      anon_sym_BQUOTE,
  [320966] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9146), 1,
      anon_sym_RBRACK,
  [320973] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15907), 1,
      anon_sym_BQUOTE,
  [320980] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15909), 1,
      anon_sym_RPAREN,
  [320987] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15911), 1,
      anon_sym_RBRACE3,
  [320994] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15913), 1,
      aux_sym_brace_expression_token1,
  [321001] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15905), 1,
      anon_sym_RPAREN,
  [321008] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15915), 1,
      anon_sym_RBRACE3,
  [321015] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15917), 1,
      anon_sym_RBRACK_RBRACK,
  [321022] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15919), 1,
      anon_sym_BQUOTE,
  [321029] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15921), 1,
      anon_sym_BQUOTE,
  [321036] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15923), 1,
      anon_sym_RBRACE2,
  [321043] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15925), 1,
      aux_sym_brace_expression_token1,
  [321050] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9028), 1,
      anon_sym_RBRACK,
  [321057] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15927), 1,
      anon_sym_RPAREN,
  [321064] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15919), 1,
      anon_sym_RPAREN,
  [321071] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15929), 1,
      anon_sym_RBRACE3,
  [321078] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15931), 1,
      anon_sym_RBRACE3,
  [321085] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9026), 1,
      anon_sym_RBRACK,
  [321092] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15933), 1,
      aux_sym_brace_expression_token1,
  [321099] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15935), 1,
      anon_sym_RPAREN,
  [321106] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15937), 1,
      anon_sym_BQUOTE,
  [321113] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15939), 1,
      anon_sym_BQUOTE,
  [321120] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15941), 1,
      anon_sym_BQUOTE,
  [321127] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15937), 1,
      anon_sym_RPAREN,
  [321134] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15943), 1,
      sym_heredoc_start,
  [321141] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15945), 1,
      aux_sym_brace_expression_token1,
  [321148] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15947), 1,
      anon_sym_RBRACE3,
  [321155] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15949), 1,
      anon_sym_BQUOTE,
  [321162] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15951), 1,
      anon_sym_BQUOTE,
  [321169] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15953), 1,
      anon_sym_RPAREN,
  [321176] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15935), 1,
      anon_sym_BQUOTE,
  [321183] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15955), 1,
      anon_sym_BQUOTE,
  [321190] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15957), 1,
      aux_sym_brace_expression_token1,
  [321197] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15959), 1,
      anon_sym_BQUOTE,
  [321204] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15961), 1,
      anon_sym_BQUOTE,
  [321211] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15927), 1,
      anon_sym_BQUOTE,
  [321218] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15955), 1,
      anon_sym_RPAREN,
  [321225] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14827), 1,
      anon_sym_RBRACE3,
  [321232] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15963), 1,
      anon_sym_RBRACE3,
  [321239] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14827), 1,
      anon_sym_RBRACE3,
  [321246] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14827), 1,
      anon_sym_RBRACE3,
  [321253] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14827), 1,
      anon_sym_RBRACE3,
  [321260] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14827), 1,
      anon_sym_RBRACE3,
  [321267] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15965), 1,
      anon_sym_RPAREN,
  [321274] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15967), 1,
      anon_sym_LBRACK,
  [321281] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15969), 1,
      anon_sym_RBRACE2,
  [321288] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15971), 1,
      anon_sym_BQUOTE,
  [321295] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15973), 1,
      anon_sym_BQUOTE,
  [321302] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15975), 1,
      anon_sym_BQUOTE,
  [321309] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15977), 1,
      anon_sym_RPAREN,
  [321316] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15973), 1,
      anon_sym_RPAREN,
  [321323] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14810), 1,
      anon_sym_RBRACE3,
  [321330] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14810), 1,
      anon_sym_RBRACE3,
  [321337] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14810), 1,
      anon_sym_RBRACE3,
  [321344] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14810), 1,
      anon_sym_RBRACE3,
  [321351] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14810), 1,
      anon_sym_RBRACE3,
  [321358] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14810), 1,
      anon_sym_RBRACE3,
  [321365] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14827), 1,
      anon_sym_RBRACE3,
  [321372] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14827), 1,
      anon_sym_RBRACE3,
  [321379] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14827), 1,
      anon_sym_RBRACE3,
  [321386] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14827), 1,
      anon_sym_RBRACE3,
  [321393] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14827), 1,
      anon_sym_RBRACE3,
  [321400] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14827), 1,
      anon_sym_RBRACE3,
  [321407] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15979), 1,
      anon_sym_RBRACE2,
  [321414] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15981), 1,
      anon_sym_RBRACE3,
  [321421] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15961), 1,
      anon_sym_RPAREN,
  [321428] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15983), 1,
      anon_sym_RBRACE2,
  [321435] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15985), 1,
      anon_sym_RBRACE2,
  [321442] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15987), 1,
      anon_sym_RPAREN,
  [321449] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15989), 1,
      anon_sym_BQUOTE,
  [321456] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14512), 1,
      anon_sym_RBRACE3,
  [321463] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15991), 1,
      anon_sym_BQUOTE,
  [321470] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15993), 1,
      anon_sym_BQUOTE,
  [321477] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15991), 1,
      anon_sym_RPAREN,
  [321484] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15995), 1,
      anon_sym_RBRACE3,
  [321491] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15765), 1,
      anon_sym_RPAREN,
  [321498] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15997), 1,
      anon_sym_RBRACE3,
  [321505] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15999), 1,
      anon_sym_in,
  [321512] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16001), 1,
      anon_sym_in,
  [321519] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16003), 1,
      anon_sym_RBRACE3,
  [321526] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16005), 1,
      anon_sym_RBRACE3,
  [321533] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16007), 1,
      anon_sym_RPAREN,
  [321540] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16009), 1,
      anon_sym_RBRACE2,
  [321547] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16011), 1,
      anon_sym_RPAREN,
  [321554] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15269), 1,
      aux_sym__simple_variable_name_token1,
  [321561] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16013), 1,
      anon_sym_BQUOTE,
  [321568] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16015), 1,
      anon_sym_BQUOTE,
  [321575] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16017), 1,
      sym_word,
  [321582] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16019), 1,
      anon_sym_DOT_DOT,
  [321589] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16013), 1,
      anon_sym_RPAREN,
  [321596] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16021), 1,
      anon_sym_RBRACE3,
  [321603] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16023), 1,
      anon_sym_RBRACE3,
  [321610] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16025), 1,
      anon_sym_RPAREN,
  [321617] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16027), 1,
      sym_heredoc_start,
  [321624] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16029), 1,
      anon_sym_esac,
  [321631] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16031), 1,
      anon_sym_RBRACE2,
  [321638] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16033), 1,
      anon_sym_BQUOTE,
  [321645] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16035), 1,
      anon_sym_RBRACE2,
  [321652] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16037), 1,
      anon_sym_RPAREN,
  [321659] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16039), 1,
      anon_sym_BQUOTE,
  [321666] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16041), 1,
      anon_sym_RBRACE2,
  [321673] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16043), 1,
      anon_sym_BQUOTE,
  [321680] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16039), 1,
      anon_sym_RPAREN,
  [321687] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16045), 1,
      anon_sym_RBRACE3,
  [321694] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16047), 1,
      anon_sym_RBRACE3,
  [321701] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16049), 1,
      anon_sym_RPAREN,
  [321708] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16051), 1,
      anon_sym_RBRACE2,
  [321715] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16053), 1,
      anon_sym_RPAREN,
  [321722] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16055), 1,
      anon_sym_BQUOTE,
  [321729] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16057), 1,
      anon_sym_BQUOTE,
  [321736] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16059), 1,
      anon_sym_esac,
  [321743] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16061), 1,
      sym_heredoc_start,
  [321750] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16063), 1,
      anon_sym_RPAREN,
  [321757] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16065), 1,
      anon_sym_RPAREN,
  [321764] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16055), 1,
      anon_sym_RPAREN,
  [321771] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16067), 1,
      anon_sym_RBRACE3,
  [321778] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16069), 1,
      anon_sym_RBRACE2,
  [321785] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16071), 1,
      anon_sym_RBRACE3,
  [321792] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16073), 1,
      anon_sym_RPAREN,
  [321799] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16075), 1,
      anon_sym_BQUOTE,
  [321806] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16077), 1,
      anon_sym_BQUOTE,
  [321813] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14512), 1,
      anon_sym_RBRACE3,
  [321820] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16079), 1,
      anon_sym_RBRACE3,
  [321827] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16081), 1,
      anon_sym_RPAREN,
  [321834] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16083), 1,
      anon_sym_BQUOTE,
  [321841] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16075), 1,
      anon_sym_RPAREN,
  [321848] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16085), 1,
      anon_sym_BQUOTE,
  [321855] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16087), 1,
      anon_sym_RBRACE3,
  [321862] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16089), 1,
      anon_sym_RBRACE2,
  [321869] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16091), 1,
      anon_sym_RPAREN,
  [321876] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16093), 1,
      anon_sym_BQUOTE,
  [321883] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16083), 1,
      anon_sym_RPAREN,
  [321890] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16095), 1,
      anon_sym_RBRACE3,
  [321897] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16097), 1,
      anon_sym_BQUOTE,
  [321904] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16099), 1,
      anon_sym_esac,
  [321911] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16093), 1,
      anon_sym_RPAREN,
  [321918] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16101), 1,
      anon_sym_RBRACE3,
  [321925] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16103), 1,
      anon_sym_RBRACK_RBRACK,
  [321932] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16105), 1,
      anon_sym_BQUOTE,
  [321939] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16107), 1,
      anon_sym_fi,
  [321946] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16007), 1,
      anon_sym_BQUOTE,
  [321953] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16109), 1,
      anon_sym_RBRACE2,
  [321960] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16073), 1,
      anon_sym_BQUOTE,
  [321967] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16111), 1,
      anon_sym_RBRACE2,
  [321974] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16113), 1,
      anon_sym_RPAREN,
  [321981] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16115), 1,
      anon_sym_RPAREN_RPAREN,
  [321988] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16117), 1,
      anon_sym_BQUOTE,
  [321995] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16119), 1,
      anon_sym_BQUOTE,
  [322002] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16117), 1,
      anon_sym_RPAREN,
  [322009] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16121), 1,
      aux_sym_brace_expression_token1,
  [322016] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16123), 1,
      anon_sym_RBRACE3,
  [322023] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16125), 1,
      anon_sym_RBRACE3,
  [322030] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16127), 1,
      anon_sym_RPAREN,
  [322037] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16129), 1,
      anon_sym_RPAREN,
  [322044] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16131), 1,
      anon_sym_BQUOTE,
  [322051] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16133), 1,
      anon_sym_RPAREN,
  [322058] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16135), 1,
      anon_sym_RBRACE2,
  [322065] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16137), 1,
      anon_sym_RPAREN,
  [322072] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16139), 1,
      anon_sym_BQUOTE,
  [322079] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16141), 1,
      anon_sym_BQUOTE,
  [322086] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16139), 1,
      anon_sym_RPAREN,
  [322093] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16143), 1,
      anon_sym_BQUOTE,
  [322100] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16145), 1,
      anon_sym_RBRACE2,
  [322107] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16147), 1,
      anon_sym_LT_LT_LT,
  [322114] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16149), 1,
      anon_sym_RBRACE3,
  [322121] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16151), 1,
      anon_sym_RBRACE2,
  [322128] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16133), 1,
      anon_sym_BQUOTE,
  [322135] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16153), 1,
      anon_sym_RBRACE3,
  [322142] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16155), 1,
      sym_heredoc_start,
  [322149] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16157), 1,
      anon_sym_RPAREN,
  [322156] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16159), 1,
      anon_sym_RBRACE2,
  [322163] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16161), 1,
      anon_sym_RPAREN,
  [322170] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16163), 1,
      anon_sym_RPAREN,
  [322177] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16165), 1,
      anon_sym_BQUOTE,
  [322184] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16167), 1,
      anon_sym_BQUOTE,
  [322191] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16165), 1,
      anon_sym_RPAREN,
  [322198] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16169), 1,
      anon_sym_RPAREN,
  [322205] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16171), 1,
      anon_sym_BQUOTE,
  [322212] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16173), 1,
      anon_sym_BQUOTE,
  [322219] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16171), 1,
      anon_sym_RPAREN,
  [322226] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16175), 1,
      anon_sym_RBRACE3,
  [322233] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9158), 1,
      anon_sym_RBRACK,
  [322240] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(9164), 1,
      anon_sym_RBRACK,
  [322247] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16177), 1,
      anon_sym_RPAREN,
  [322254] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16179), 1,
      anon_sym_RBRACE2,
  [322261] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16181), 1,
      anon_sym_RPAREN,
  [322268] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16183), 1,
      anon_sym_BQUOTE,
  [322275] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16185), 1,
      anon_sym_BQUOTE,
  [322282] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16187), 1,
      anon_sym_RPAREN,
  [322289] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16189), 1,
      anon_sym_BQUOTE,
  [322296] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16191), 1,
      anon_sym_BQUOTE,
  [322303] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16183), 1,
      anon_sym_RPAREN,
  [322310] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16193), 1,
      anon_sym_RBRACE3,
  [322317] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16195), 1,
      anon_sym_RBRACE3,
  [322324] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16189), 1,
      anon_sym_RPAREN,
  [322331] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16197), 1,
      anon_sym_RBRACE2,
  [322338] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16199), 1,
      anon_sym_RBRACE3,
  [322345] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16201), 1,
      anon_sym_RPAREN,
  [322352] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16131), 1,
      anon_sym_RPAREN,
  [322359] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16203), 1,
      anon_sym_RBRACE3,
  [322366] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16205), 1,
      anon_sym_RBRACE2,
  [322373] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16207), 1,
      anon_sym_RPAREN,
  [322380] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16209), 1,
      anon_sym_BQUOTE,
  [322387] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16211), 1,
      anon_sym_BQUOTE,
  [322394] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16209), 1,
      anon_sym_RPAREN,
  [322401] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16213), 1,
      anon_sym_RBRACE3,
  [322408] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16215), 1,
      anon_sym_RBRACE2,
  [322415] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16217), 1,
      anon_sym_RPAREN,
  [322422] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15261), 1,
      aux_sym__simple_variable_name_token1,
  [322429] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16219), 1,
      anon_sym_BQUOTE,
  [322436] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16221), 1,
      anon_sym_BQUOTE,
  [322443] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16223), 1,
      sym_word,
  [322450] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16225), 1,
      anon_sym_DOT_DOT,
  [322457] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16219), 1,
      anon_sym_RPAREN,
  [322464] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16227), 1,
      anon_sym_RBRACE3,
  [322471] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16229), 1,
      anon_sym_RPAREN,
  [322478] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16231), 1,
      sym_heredoc_start,
  [322485] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16233), 1,
      anon_sym_RBRACE3,
  [322492] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16235), 1,
      anon_sym_BQUOTE,
  [322499] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16201), 1,
      anon_sym_BQUOTE,
  [322506] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16237), 1,
      sym_heredoc_start,
  [322513] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16239), 1,
      anon_sym_esac,
  [322520] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16241), 1,
      anon_sym_RPAREN,
  [322527] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16243), 1,
      anon_sym_RBRACE2,
  [322534] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16245), 1,
      anon_sym_RPAREN,
  [322541] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16247), 1,
      anon_sym_RBRACE2,
  [322548] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16249), 1,
      anon_sym_RPAREN,
  [322555] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16251), 1,
      anon_sym_RBRACK_RBRACK,
  [322562] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16253), 1,
      anon_sym_BQUOTE,
  [322569] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16255), 1,
      anon_sym_BQUOTE,
  [322576] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16253), 1,
      anon_sym_RPAREN,
  [322583] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16257), 1,
      anon_sym_RBRACE3,
  [322590] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16259), 1,
      anon_sym_esac,
  [322597] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16261), 1,
      anon_sym_BQUOTE,
  [322604] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16263), 1,
      anon_sym_RBRACE2,
  [322611] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14486), 1,
      anon_sym_fi,
  [322618] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14540), 1,
      anon_sym_fi,
  [322625] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16265), 1,
      anon_sym_RPAREN,
  [322632] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16267), 1,
      anon_sym_BQUOTE,
  [322639] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16269), 1,
      anon_sym_BQUOTE,
  [322646] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16267), 1,
      anon_sym_RPAREN,
  [322653] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16271), 1,
      anon_sym_RBRACE3,
  [322660] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16241), 1,
      anon_sym_BQUOTE,
  [322667] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16273), 1,
      anon_sym_RPAREN,
  [322674] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16275), 1,
      anon_sym_RBRACE2,
  [322681] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16277), 1,
      anon_sym_RPAREN,
  [322688] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16279), 1,
      anon_sym_RPAREN,
  [322695] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16281), 1,
      anon_sym_RPAREN,
  [322702] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16283), 1,
      anon_sym_BQUOTE,
  [322709] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16285), 1,
      anon_sym_BQUOTE,
  [322716] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16287), 1,
      anon_sym_BQUOTE,
  [322723] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16283), 1,
      anon_sym_RPAREN,
  [322730] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16289), 1,
      anon_sym_RBRACE3,
  [322737] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16291), 1,
      anon_sym_BQUOTE,
  [322744] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16293), 1,
      anon_sym_RBRACE2,
  [322751] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16287), 1,
      anon_sym_RPAREN,
  [322758] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16295), 1,
      anon_sym_RBRACE3,
  [322765] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16297), 1,
      anon_sym_RPAREN,
  [322772] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16299), 1,
      anon_sym_DOT_DOT,
  [322779] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16301), 1,
      anon_sym_esac,
  [322786] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16303), 1,
      anon_sym_RPAREN,
  [322793] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16305), 1,
      anon_sym_RBRACE2,
  [322800] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16307), 1,
      anon_sym_RBRACE2,
  [322807] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16309), 1,
      anon_sym_RBRACK_RBRACK,
  [322814] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16311), 1,
      anon_sym_RPAREN,
  [322821] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16313), 1,
      anon_sym_BQUOTE,
  [322828] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16315), 1,
      anon_sym_BQUOTE,
  [322835] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16313), 1,
      anon_sym_RPAREN,
  [322842] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16317), 1,
      anon_sym_RBRACE3,
  [322849] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16319), 1,
      anon_sym_RBRACE2,
  [322856] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16321), 1,
      anon_sym_esac,
  [322863] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16323), 1,
      anon_sym_RBRACE2,
  [322870] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16325), 1,
      anon_sym_RBRACE2,
  [322877] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16327), 1,
      anon_sym_RPAREN,
  [322884] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16329), 1,
      anon_sym_BQUOTE,
  [322891] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16331), 1,
      anon_sym_RPAREN,
  [322898] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16333), 1,
      anon_sym_BQUOTE,
  [322905] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16335), 1,
      anon_sym_BQUOTE,
  [322912] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16337), 1,
      anon_sym_BQUOTE,
  [322919] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16333), 1,
      anon_sym_RPAREN,
  [322926] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16339), 1,
      anon_sym_RBRACE3,
  [322933] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16341), 1,
      anon_sym_RPAREN,
  [322940] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16329), 1,
      anon_sym_RPAREN,
  [322947] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16343), 1,
      anon_sym_RBRACE3,
  [322954] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16345), 1,
      sym_heredoc_end,
  [322961] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16347), 1,
      sym_heredoc_end,
  [322968] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16349), 1,
      anon_sym_BQUOTE,
  [322975] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16351), 1,
      anon_sym_RBRACE2,
  [322982] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16353), 1,
      ts_builtin_sym_end,
  [322989] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16355), 1,
      ts_builtin_sym_end,
  [322996] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16357), 1,
      anon_sym_RPAREN,
  [323003] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16359), 1,
      anon_sym_BQUOTE,
  [323010] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16361), 1,
      anon_sym_BQUOTE,
  [323017] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16363), 1,
      anon_sym_BQUOTE,
  [323024] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16359), 1,
      anon_sym_RPAREN,
  [323031] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16365), 1,
      anon_sym_RBRACE3,
  [323038] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14512), 1,
      anon_sym_RBRACE3,
  [323045] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(14696), 1,
      anon_sym_fi,
  [323052] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15285), 1,
      aux_sym__simple_variable_name_token1,
  [323059] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16367), 1,
      anon_sym_fi,
  [323066] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16369), 1,
      anon_sym_RBRACE2,
  [323073] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16371), 1,
      sym_word,
  [323080] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16373), 1,
      anon_sym_DOT_DOT,
  [323087] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16375), 1,
      anon_sym_RPAREN,
  [323094] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16377), 1,
      anon_sym_BQUOTE,
  [323101] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16379), 1,
      anon_sym_RPAREN,
  [323108] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16381), 1,
      sym_heredoc_start,
  [323115] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16383), 1,
      anon_sym_BQUOTE,
  [323122] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16385), 1,
      anon_sym_esac,
  [323129] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16387), 1,
      anon_sym_RBRACE2,
  [323136] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16389), 1,
      sym_heredoc_start,
  [323143] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16377), 1,
      anon_sym_RPAREN,
  [323150] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16391), 1,
      anon_sym_RBRACE3,
  [323157] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16393), 1,
      anon_sym_esac,
  [323164] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16395), 1,
      anon_sym_esac,
  [323171] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16397), 1,
      anon_sym_RBRACE2,
  [323178] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16399), 1,
      anon_sym_RPAREN,
  [323185] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16401), 1,
      anon_sym_BQUOTE,
  [323192] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16403), 1,
      anon_sym_BQUOTE,
  [323199] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16401), 1,
      anon_sym_RPAREN,
  [323206] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16405), 1,
      anon_sym_RBRACE3,
  [323213] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16407), 1,
      anon_sym_RPAREN,
  [323220] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16409), 1,
      anon_sym_esac,
  [323227] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16411), 1,
      anon_sym_esac,
  [323234] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16413), 1,
      anon_sym_RBRACE2,
  [323241] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16415), 1,
      anon_sym_RPAREN,
  [323248] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16417), 1,
      anon_sym_BQUOTE,
  [323255] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16419), 1,
      anon_sym_BQUOTE,
  [323262] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16417), 1,
      anon_sym_RPAREN,
  [323269] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16421), 1,
      anon_sym_RBRACE3,
  [323276] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16423), 1,
      sym_word,
  [323283] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16425), 1,
      anon_sym_DOT_DOT,
  [323290] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16427), 1,
      anon_sym_RPAREN,
  [323297] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16429), 1,
      sym_heredoc_start,
  [323304] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16431), 1,
      anon_sym_BQUOTE,
  [323311] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16433), 1,
      sym_word,
  [323318] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16435), 1,
      sym_heredoc_start,
  [323325] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16437), 1,
      anon_sym_BQUOTE,
  [323332] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16439), 1,
      sym_heredoc_end,
  [323339] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16431), 1,
      anon_sym_RPAREN,
  [323346] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16441), 1,
      anon_sym_RBRACE3,
  [323353] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16349), 1,
      anon_sym_RPAREN,
  [323360] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16443), 1,
      sym_heredoc_end,
  [323367] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16445), 1,
      anon_sym_RBRACE3,
  [323374] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16447), 1,
      anon_sym_RBRACE2,
  [323381] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16449), 1,
      anon_sym_RPAREN,
  [323388] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16451), 1,
      anon_sym_BQUOTE,
  [323395] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16453), 1,
      anon_sym_esac,
  [323402] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16455), 1,
      anon_sym_BQUOTE,
  [323409] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16457), 1,
      anon_sym_RBRACE3,
  [323416] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16451), 1,
      anon_sym_RPAREN,
  [323423] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16459), 1,
      anon_sym_RBRACE2,
  [323430] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16461), 1,
      anon_sym_RBRACE3,
  [323437] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16463), 1,
      anon_sym_RBRACE2,
  [323444] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16465), 1,
      anon_sym_RPAREN,
  [323451] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16467), 1,
      anon_sym_BQUOTE,
  [323458] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16469), 1,
      anon_sym_DOT_DOT,
  [323465] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16471), 1,
      anon_sym_DOT_DOT,
  [323472] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16473), 1,
      anon_sym_DOT_DOT,
  [323479] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16475), 1,
      anon_sym_DOT_DOT,
  [323486] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16477), 1,
      anon_sym_DOT_DOT,
  [323493] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16479), 1,
      anon_sym_DOT_DOT,
  [323500] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16481), 1,
      anon_sym_DOT_DOT,
  [323507] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16483), 1,
      anon_sym_DOT_DOT,
  [323514] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16485), 1,
      anon_sym_DOT_DOT,
  [323521] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16487), 1,
      anon_sym_DOT_DOT,
  [323528] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16489), 1,
      anon_sym_DOT_DOT,
  [323535] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16491), 1,
      anon_sym_DOT_DOT,
  [323542] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16493), 1,
      anon_sym_DOT_DOT,
  [323549] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16495), 1,
      anon_sym_DOT_DOT,
  [323556] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16497), 1,
      anon_sym_DOT_DOT,
  [323563] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16499), 1,
      anon_sym_DOT_DOT,
  [323570] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16501), 1,
      anon_sym_DOT_DOT,
  [323577] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16503), 1,
      anon_sym_DOT_DOT,
  [323584] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16505), 1,
      anon_sym_DOT_DOT,
  [323591] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16507), 1,
      anon_sym_DOT_DOT,
  [323598] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16509), 1,
      anon_sym_DOT_DOT,
  [323605] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16511), 1,
      anon_sym_DOT_DOT,
  [323612] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16513), 1,
      anon_sym_DOT_DOT,
  [323619] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16515), 1,
      anon_sym_DOT_DOT,
  [323626] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16517), 1,
      anon_sym_DOT_DOT,
  [323633] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16519), 1,
      anon_sym_DOT_DOT,
  [323640] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16521), 1,
      anon_sym_DOT_DOT,
  [323647] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16523), 1,
      anon_sym_DOT_DOT,
  [323654] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16525), 1,
      anon_sym_DOT_DOT,
  [323661] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16527), 1,
      anon_sym_DOT_DOT,
  [323668] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16529), 1,
      anon_sym_DOT_DOT,
  [323675] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16531), 1,
      anon_sym_DOT_DOT,
  [323682] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16533), 1,
      anon_sym_DOT_DOT,
  [323689] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16535), 1,
      anon_sym_DOT_DOT,
  [323696] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16537), 1,
      anon_sym_DOT_DOT,
  [323703] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16539), 1,
      anon_sym_DOT_DOT,
  [323710] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16541), 1,
      anon_sym_DOT_DOT,
  [323717] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16543), 1,
      anon_sym_DOT_DOT,
  [323724] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16545), 1,
      anon_sym_DOT_DOT,
  [323731] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16547), 1,
      anon_sym_DOT_DOT,
  [323738] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16549), 1,
      anon_sym_DOT_DOT,
  [323745] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16551), 1,
      anon_sym_DOT_DOT,
  [323752] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16553), 1,
      anon_sym_DOT_DOT,
  [323759] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16555), 1,
      anon_sym_DOT_DOT,
  [323766] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16557), 1,
      anon_sym_DOT_DOT,
  [323773] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16559), 1,
      anon_sym_DOT_DOT,
  [323780] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16561), 1,
      anon_sym_DOT_DOT,
  [323787] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16563), 1,
      anon_sym_DOT_DOT,
  [323794] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16565), 1,
      anon_sym_DOT_DOT,
  [323801] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16567), 1,
      anon_sym_DOT_DOT,
  [323808] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16569), 1,
      anon_sym_DOT_DOT,
  [323815] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16571), 1,
      anon_sym_DOT_DOT,
  [323822] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16573), 1,
      anon_sym_DOT_DOT,
  [323829] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16575), 1,
      anon_sym_DOT_DOT,
  [323836] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16577), 1,
      anon_sym_DOT_DOT,
  [323843] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16579), 1,
      anon_sym_DOT_DOT,
  [323850] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16581), 1,
      anon_sym_DOT_DOT,
  [323857] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16583), 1,
      anon_sym_DOT_DOT,
  [323864] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16585), 1,
      anon_sym_DOT_DOT,
  [323871] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16587), 1,
      anon_sym_DOT_DOT,
  [323878] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16589), 1,
      aux_sym_brace_expression_token1,
  [323885] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16591), 1,
      anon_sym_BQUOTE,
  [323892] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16467), 1,
      anon_sym_RPAREN,
  [323899] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16593), 1,
      anon_sym_RPAREN,
  [323906] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16595), 1,
      anon_sym_RBRACE3,
  [323913] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16593), 1,
      anon_sym_BQUOTE,
  [323920] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16597), 1,
      anon_sym_esac,
  [323927] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(15277), 1,
      aux_sym__simple_variable_name_token1,
  [323934] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16599), 1,
      aux_sym_brace_expression_token1,
  [323941] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16601), 1,
      anon_sym_BQUOTE,
  [323948] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16603), 1,
      aux_sym_brace_expression_token1,
  [323955] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16605), 1,
      aux_sym_brace_expression_token1,
  [323962] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16607), 1,
      aux_sym_brace_expression_token1,
  [323969] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16609), 1,
      aux_sym_brace_expression_token1,
  [323976] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16611), 1,
      aux_sym_brace_expression_token1,
  [323983] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16613), 1,
      aux_sym_brace_expression_token1,
  [323990] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16615), 1,
      aux_sym_brace_expression_token1,
  [323997] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16617), 1,
      aux_sym_brace_expression_token1,
  [324004] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16619), 1,
      aux_sym_brace_expression_token1,
  [324011] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16621), 1,
      aux_sym_brace_expression_token1,
  [324018] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16623), 1,
      aux_sym_brace_expression_token1,
  [324025] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16625), 1,
      aux_sym_brace_expression_token1,
  [324032] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16627), 1,
      aux_sym_brace_expression_token1,
  [324039] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16629), 1,
      aux_sym_brace_expression_token1,
  [324046] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16631), 1,
      aux_sym_brace_expression_token1,
  [324053] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16633), 1,
      aux_sym_brace_expression_token1,
  [324060] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16635), 1,
      aux_sym_brace_expression_token1,
  [324067] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16637), 1,
      aux_sym_brace_expression_token1,
  [324074] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16639), 1,
      aux_sym_brace_expression_token1,
  [324081] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16641), 1,
      aux_sym_brace_expression_token1,
  [324088] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16643), 1,
      aux_sym_brace_expression_token1,
  [324095] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16645), 1,
      aux_sym_brace_expression_token1,
  [324102] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16647), 1,
      aux_sym_brace_expression_token1,
  [324109] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16649), 1,
      aux_sym_brace_expression_token1,
  [324116] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16651), 1,
      aux_sym_brace_expression_token1,
  [324123] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16653), 1,
      aux_sym_brace_expression_token1,
  [324130] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16655), 1,
      aux_sym_brace_expression_token1,
  [324137] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16657), 1,
      aux_sym_brace_expression_token1,
  [324144] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16659), 1,
      aux_sym_brace_expression_token1,
  [324151] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16661), 1,
      aux_sym_brace_expression_token1,
  [324158] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16663), 1,
      aux_sym_brace_expression_token1,
  [324165] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16665), 1,
      aux_sym_brace_expression_token1,
  [324172] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16667), 1,
      aux_sym_brace_expression_token1,
  [324179] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16669), 1,
      aux_sym_brace_expression_token1,
  [324186] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16671), 1,
      aux_sym_brace_expression_token1,
  [324193] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16673), 1,
      aux_sym_brace_expression_token1,
  [324200] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16675), 1,
      aux_sym_brace_expression_token1,
  [324207] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16677), 1,
      aux_sym_brace_expression_token1,
  [324214] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16679), 1,
      aux_sym_brace_expression_token1,
  [324221] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16681), 1,
      aux_sym_brace_expression_token1,
  [324228] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16683), 1,
      aux_sym_brace_expression_token1,
  [324235] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16685), 1,
      aux_sym_brace_expression_token1,
  [324242] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16687), 1,
      aux_sym_brace_expression_token1,
  [324249] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16689), 1,
      aux_sym_brace_expression_token1,
  [324256] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16691), 1,
      aux_sym_brace_expression_token1,
  [324263] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16693), 1,
      aux_sym_brace_expression_token1,
  [324270] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16695), 1,
      aux_sym_brace_expression_token1,
  [324277] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16697), 1,
      aux_sym_brace_expression_token1,
  [324284] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16699), 1,
      aux_sym_brace_expression_token1,
  [324291] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16701), 1,
      aux_sym_brace_expression_token1,
  [324298] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16703), 1,
      aux_sym_brace_expression_token1,
  [324305] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16705), 1,
      aux_sym_brace_expression_token1,
  [324312] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16707), 1,
      aux_sym_brace_expression_token1,
  [324319] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16709), 1,
      aux_sym_brace_expression_token1,
  [324326] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16711), 1,
      aux_sym_brace_expression_token1,
  [324333] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16713), 1,
      aux_sym_brace_expression_token1,
  [324340] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16715), 1,
      aux_sym_brace_expression_token1,
  [324347] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16717), 1,
      aux_sym_brace_expression_token1,
  [324354] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16719), 1,
      aux_sym_brace_expression_token1,
  [324361] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16721), 1,
      aux_sym_brace_expression_token1,
  [324368] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16723), 1,
      aux_sym_brace_expression_token1,
  [324375] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16725), 1,
      aux_sym_brace_expression_token1,
  [324382] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16727), 1,
      aux_sym_brace_expression_token1,
  [324389] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(16729), 1,
      anon_sym_RBRACE3,
};

static const uint32_t ts_small_parse_table_map[] = {
  [SMALL_STATE(528)] = 0,
  [SMALL_STATE(529)] = 120,
  [SMALL_STATE(530)] = 240,
  [SMALL_STATE(531)] = 359,
  [SMALL_STATE(532)] = 478,
  [SMALL_STATE(533)] = 594,
  [SMALL_STATE(534)] = 712,
  [SMALL_STATE(535)] = 830,
  [SMALL_STATE(536)] = 944,
  [SMALL_STATE(537)] = 1062,
  [SMALL_STATE(538)] = 1180,
  [SMALL_STATE(539)] = 1258,
  [SMALL_STATE(540)] = 1336,
  [SMALL_STATE(541)] = 1450,
  [SMALL_STATE(542)] = 1565,
  [SMALL_STATE(543)] = 1678,
  [SMALL_STATE(544)] = 1793,
  [SMALL_STATE(545)] = 1910,
  [SMALL_STATE(546)] = 2027,
  [SMALL_STATE(547)] = 2136,
  [SMALL_STATE(548)] = 2249,
  [SMALL_STATE(549)] = 2358,
  [SMALL_STATE(550)] = 2435,
  [SMALL_STATE(551)] = 2512,
  [SMALL_STATE(552)] = 2621,
  [SMALL_STATE(553)] = 2698,
  [SMALL_STATE(554)] = 2775,
  [SMALL_STATE(555)] = 2890,
  [SMALL_STATE(556)] = 2966,
  [SMALL_STATE(557)] = 3042,
  [SMALL_STATE(558)] = 3154,
  [SMALL_STATE(559)] = 3230,
  [SMALL_STATE(560)] = 3338,
  [SMALL_STATE(561)] = 3450,
  [SMALL_STATE(562)] = 3562,
  [SMALL_STATE(563)] = 3676,
  [SMALL_STATE(564)] = 3784,
  [SMALL_STATE(565)] = 3860,
  [SMALL_STATE(566)] = 3968,
  [SMALL_STATE(567)] = 4080,
  [SMALL_STATE(568)] = 4156,
  [SMALL_STATE(569)] = 4232,
  [SMALL_STATE(570)] = 4346,
  [SMALL_STATE(571)] = 4453,
  [SMALL_STATE(572)] = 4564,
  [SMALL_STATE(573)] = 4639,
  [SMALL_STATE(574)] = 4710,
  [SMALL_STATE(575)] = 4823,
  [SMALL_STATE(576)] = 4894,
  [SMALL_STATE(577)] = 5001,
  [SMALL_STATE(578)] = 5108,
  [SMALL_STATE(579)] = 5183,
  [SMALL_STATE(580)] = 5258,
  [SMALL_STATE(581)] = 5333,
  [SMALL_STATE(582)] = 5440,
  [SMALL_STATE(583)] = 5551,
  [SMALL_STATE(584)] = 5652,
  [SMALL_STATE(585)] = 5761,
  [SMALL_STATE(586)] = 5868,
  [SMALL_STATE(587)] = 5975,
  [SMALL_STATE(588)] = 6084,
  [SMALL_STATE(589)] = 6159,
  [SMALL_STATE(590)] = 6234,
  [SMALL_STATE(591)] = 6336,
  [SMALL_STATE(592)] = 6436,
  [SMALL_STATE(593)] = 6540,
  [SMALL_STATE(594)] = 6640,
  [SMALL_STATE(595)] = 6744,
  [SMALL_STATE(596)] = 6844,
  [SMALL_STATE(597)] = 6958,
  [SMALL_STATE(598)] = 7032,
  [SMALL_STATE(599)] = 7134,
  [SMALL_STATE(600)] = 7198,
  [SMALL_STATE(601)] = 7312,
  [SMALL_STATE(602)] = 7382,
  [SMALL_STATE(603)] = 7488,
  [SMALL_STATE(604)] = 7562,
  [SMALL_STATE(605)] = 7632,
  [SMALL_STATE(606)] = 7706,
  [SMALL_STATE(607)] = 7780,
  [SMALL_STATE(608)] = 7850,
  [SMALL_STATE(609)] = 7924,
  [SMALL_STATE(610)] = 8024,
  [SMALL_STATE(611)] = 8130,
  [SMALL_STATE(612)] = 8204,
  [SMALL_STATE(613)] = 8310,
  [SMALL_STATE(614)] = 8412,
  [SMALL_STATE(615)] = 8482,
  [SMALL_STATE(616)] = 8581,
  [SMALL_STATE(617)] = 8680,
  [SMALL_STATE(618)] = 8795,
  [SMALL_STATE(619)] = 8926,
  [SMALL_STATE(620)] = 9057,
  [SMALL_STATE(621)] = 9172,
  [SMALL_STATE(622)] = 9271,
  [SMALL_STATE(623)] = 9338,
  [SMALL_STATE(624)] = 9469,
  [SMALL_STATE(625)] = 9536,
  [SMALL_STATE(626)] = 9603,
  [SMALL_STATE(627)] = 9734,
  [SMALL_STATE(628)] = 9833,
  [SMALL_STATE(629)] = 9906,
  [SMALL_STATE(630)] = 10007,
  [SMALL_STATE(631)] = 10106,
  [SMALL_STATE(632)] = 10205,
  [SMALL_STATE(633)] = 10306,
  [SMALL_STATE(634)] = 10407,
  [SMALL_STATE(635)] = 10480,
  [SMALL_STATE(636)] = 10579,
  [SMALL_STATE(637)] = 10710,
  [SMALL_STATE(638)] = 10783,
  [SMALL_STATE(639)] = 10856,
  [SMALL_STATE(640)] = 10929,
  [SMALL_STATE(641)] = 11002,
  [SMALL_STATE(642)] = 11075,
  [SMALL_STATE(643)] = 11144,
  [SMALL_STATE(644)] = 11217,
  [SMALL_STATE(645)] = 11348,
  [SMALL_STATE(646)] = 11417,
  [SMALL_STATE(647)] = 11486,
  [SMALL_STATE(648)] = 11559,
  [SMALL_STATE(649)] = 11628,
  [SMALL_STATE(650)] = 11697,
  [SMALL_STATE(651)] = 11828,
  [SMALL_STATE(652)] = 11927,
  [SMALL_STATE(653)] = 12000,
  [SMALL_STATE(654)] = 12099,
  [SMALL_STATE(655)] = 12168,
  [SMALL_STATE(656)] = 12235,
  [SMALL_STATE(657)] = 12366,
  [SMALL_STATE(658)] = 12465,
  [SMALL_STATE(659)] = 12563,
  [SMALL_STATE(660)] = 12691,
  [SMALL_STATE(661)] = 12789,
  [SMALL_STATE(662)] = 12917,
  [SMALL_STATE(663)] = 12989,
  [SMALL_STATE(664)] = 13061,
  [SMALL_STATE(665)] = 13161,
  [SMALL_STATE(666)] = 13259,
  [SMALL_STATE(667)] = 13331,
  [SMALL_STATE(668)] = 13431,
  [SMALL_STATE(669)] = 13503,
  [SMALL_STATE(670)] = 13575,
  [SMALL_STATE(671)] = 13647,
  [SMALL_STATE(672)] = 13775,
  [SMALL_STATE(673)] = 13873,
  [SMALL_STATE(674)] = 13941,
  [SMALL_STATE(675)] = 14013,
  [SMALL_STATE(676)] = 14121,
  [SMALL_STATE(677)] = 14193,
  [SMALL_STATE(678)] = 14265,
  [SMALL_STATE(679)] = 14363,
  [SMALL_STATE(680)] = 14435,
  [SMALL_STATE(681)] = 14533,
  [SMALL_STATE(682)] = 14631,
  [SMALL_STATE(683)] = 14699,
  [SMALL_STATE(684)] = 14771,
  [SMALL_STATE(685)] = 14869,
  [SMALL_STATE(686)] = 14997,
  [SMALL_STATE(687)] = 15065,
  [SMALL_STATE(688)] = 15137,
  [SMALL_STATE(689)] = 15245,
  [SMALL_STATE(690)] = 15313,
  [SMALL_STATE(691)] = 15413,
  [SMALL_STATE(692)] = 15511,
  [SMALL_STATE(693)] = 15583,
  [SMALL_STATE(694)] = 15683,
  [SMALL_STATE(695)] = 15751,
  [SMALL_STATE(696)] = 15851,
  [SMALL_STATE(697)] = 15923,
  [SMALL_STATE(698)] = 16051,
  [SMALL_STATE(699)] = 16179,
  [SMALL_STATE(700)] = 16251,
  [SMALL_STATE(701)] = 16323,
  [SMALL_STATE(702)] = 16451,
  [SMALL_STATE(703)] = 16549,
  [SMALL_STATE(704)] = 16659,
  [SMALL_STATE(705)] = 16757,
  [SMALL_STATE(706)] = 16855,
  [SMALL_STATE(707)] = 16953,
  [SMALL_STATE(708)] = 17051,
  [SMALL_STATE(709)] = 17119,
  [SMALL_STATE(710)] = 17219,
  [SMALL_STATE(711)] = 17290,
  [SMALL_STATE(712)] = 17415,
  [SMALL_STATE(713)] = 17540,
  [SMALL_STATE(714)] = 17665,
  [SMALL_STATE(715)] = 17762,
  [SMALL_STATE(716)] = 17859,
  [SMALL_STATE(717)] = 17930,
  [SMALL_STATE(718)] = 18001,
  [SMALL_STATE(719)] = 18100,
  [SMALL_STATE(720)] = 18197,
  [SMALL_STATE(721)] = 18294,
  [SMALL_STATE(722)] = 18365,
  [SMALL_STATE(723)] = 18462,
  [SMALL_STATE(724)] = 18533,
  [SMALL_STATE(725)] = 18630,
  [SMALL_STATE(726)] = 18701,
  [SMALL_STATE(727)] = 18798,
  [SMALL_STATE(728)] = 18907,
  [SMALL_STATE(729)] = 19022,
  [SMALL_STATE(730)] = 19137,
  [SMALL_STATE(731)] = 19262,
  [SMALL_STATE(732)] = 19361,
  [SMALL_STATE(733)] = 19432,
  [SMALL_STATE(734)] = 19529,
  [SMALL_STATE(735)] = 19654,
  [SMALL_STATE(736)] = 19725,
  [SMALL_STATE(737)] = 19840,
  [SMALL_STATE(738)] = 19911,
  [SMALL_STATE(739)] = 20008,
  [SMALL_STATE(740)] = 20105,
  [SMALL_STATE(741)] = 20176,
  [SMALL_STATE(742)] = 20273,
  [SMALL_STATE(743)] = 20370,
  [SMALL_STATE(744)] = 20469,
  [SMALL_STATE(745)] = 20584,
  [SMALL_STATE(746)] = 20699,
  [SMALL_STATE(747)] = 20802,
  [SMALL_STATE(748)] = 20899,
  [SMALL_STATE(749)] = 21002,
  [SMALL_STATE(750)] = 21073,
  [SMALL_STATE(751)] = 21144,
  [SMALL_STATE(752)] = 21255,
  [SMALL_STATE(753)] = 21364,
  [SMALL_STATE(754)] = 21479,
  [SMALL_STATE(755)] = 21574,
  [SMALL_STATE(756)] = 21669,
  [SMALL_STATE(757)] = 21740,
  [SMALL_STATE(758)] = 21837,
  [SMALL_STATE(759)] = 21952,
  [SMALL_STATE(760)] = 22019,
  [SMALL_STATE(761)] = 22144,
  [SMALL_STATE(762)] = 22241,
  [SMALL_STATE(763)] = 22366,
  [SMALL_STATE(764)] = 22491,
  [SMALL_STATE(765)] = 22588,
  [SMALL_STATE(766)] = 22659,
  [SMALL_STATE(767)] = 22730,
  [SMALL_STATE(768)] = 22801,
  [SMALL_STATE(769)] = 22872,
  [SMALL_STATE(770)] = 22987,
  [SMALL_STATE(771)] = 23102,
  [SMALL_STATE(772)] = 23217,
  [SMALL_STATE(773)] = 23284,
  [SMALL_STATE(774)] = 23381,
  [SMALL_STATE(775)] = 23452,
  [SMALL_STATE(776)] = 23523,
  [SMALL_STATE(777)] = 23594,
  [SMALL_STATE(778)] = 23697,
  [SMALL_STATE(779)] = 23791,
  [SMALL_STATE(780)] = 23887,
  [SMALL_STATE(781)] = 23947,
  [SMALL_STATE(782)] = 24043,
  [SMALL_STATE(783)] = 24113,
  [SMALL_STATE(784)] = 24209,
  [SMALL_STATE(785)] = 24279,
  [SMALL_STATE(786)] = 24373,
  [SMALL_STATE(787)] = 24467,
  [SMALL_STATE(788)] = 24537,
  [SMALL_STATE(789)] = 24633,
  [SMALL_STATE(790)] = 24729,
  [SMALL_STATE(791)] = 24825,
  [SMALL_STATE(792)] = 24919,
  [SMALL_STATE(793)] = 25015,
  [SMALL_STATE(794)] = 25085,
  [SMALL_STATE(795)] = 25147,
  [SMALL_STATE(796)] = 25217,
  [SMALL_STATE(797)] = 25287,
  [SMALL_STATE(798)] = 25391,
  [SMALL_STATE(799)] = 25487,
  [SMALL_STATE(800)] = 25591,
  [SMALL_STATE(801)] = 25653,
  [SMALL_STATE(802)] = 25723,
  [SMALL_STATE(803)] = 25783,
  [SMALL_STATE(804)] = 25887,
  [SMALL_STATE(805)] = 25983,
  [SMALL_STATE(806)] = 26053,
  [SMALL_STATE(807)] = 26118,
  [SMALL_STATE(808)] = 26197,
  [SMALL_STATE(809)] = 26262,
  [SMALL_STATE(810)] = 26335,
  [SMALL_STATE(811)] = 26396,
  [SMALL_STATE(812)] = 26491,
  [SMALL_STATE(813)] = 26552,
  [SMALL_STATE(814)] = 26619,
  [SMALL_STATE(815)] = 26684,
  [SMALL_STATE(816)] = 26749,
  [SMALL_STATE(817)] = 26814,
  [SMALL_STATE(818)] = 26911,
  [SMALL_STATE(819)] = 26990,
  [SMALL_STATE(820)] = 27055,
  [SMALL_STATE(821)] = 27120,
  [SMALL_STATE(822)] = 27185,
  [SMALL_STATE(823)] = 27256,
  [SMALL_STATE(824)] = 27321,
  [SMALL_STATE(825)] = 27386,
  [SMALL_STATE(826)] = 27481,
  [SMALL_STATE(827)] = 27576,
  [SMALL_STATE(828)] = 27635,
  [SMALL_STATE(829)] = 27728,
  [SMALL_STATE(830)] = 27821,
  [SMALL_STATE(831)] = 27885,
  [SMALL_STATE(832)] = 27943,
  [SMALL_STATE(833)] = 28007,
  [SMALL_STATE(834)] = 28071,
  [SMALL_STATE(835)] = 28135,
  [SMALL_STATE(836)] = 28213,
  [SMALL_STATE(837)] = 28277,
  [SMALL_STATE(838)] = 28341,
  [SMALL_STATE(839)] = 28405,
  [SMALL_STATE(840)] = 28477,
  [SMALL_STATE(841)] = 28545,
  [SMALL_STATE(842)] = 28609,
  [SMALL_STATE(843)] = 28677,
  [SMALL_STATE(844)] = 28735,
  [SMALL_STATE(845)] = 28793,
  [SMALL_STATE(846)] = 28851,
  [SMALL_STATE(847)] = 28909,
  [SMALL_STATE(848)] = 28979,
  [SMALL_STATE(849)] = 29037,
  [SMALL_STATE(850)] = 29115,
  [SMALL_STATE(851)] = 29179,
  [SMALL_STATE(852)] = 29275,
  [SMALL_STATE(853)] = 29339,
  [SMALL_STATE(854)] = 29403,
  [SMALL_STATE(855)] = 29467,
  [SMALL_STATE(856)] = 29525,
  [SMALL_STATE(857)] = 29583,
  [SMALL_STATE(858)] = 29641,
  [SMALL_STATE(859)] = 29705,
  [SMALL_STATE(860)] = 29769,
  [SMALL_STATE(861)] = 29863,
  [SMALL_STATE(862)] = 29927,
  [SMALL_STATE(863)] = 30021,
  [SMALL_STATE(864)] = 30115,
  [SMALL_STATE(865)] = 30173,
  [SMALL_STATE(866)] = 30269,
  [SMALL_STATE(867)] = 30327,
  [SMALL_STATE(868)] = 30423,
  [SMALL_STATE(869)] = 30481,
  [SMALL_STATE(870)] = 30539,
  [SMALL_STATE(871)] = 30603,
  [SMALL_STATE(872)] = 30661,
  [SMALL_STATE(873)] = 30719,
  [SMALL_STATE(874)] = 30787,
  [SMALL_STATE(875)] = 30851,
  [SMALL_STATE(876)] = 30947,
  [SMALL_STATE(877)] = 31011,
  [SMALL_STATE(878)] = 31079,
  [SMALL_STATE(879)] = 31175,
  [SMALL_STATE(880)] = 31271,
  [SMALL_STATE(881)] = 31335,
  [SMALL_STATE(882)] = 31403,
  [SMALL_STATE(883)] = 31467,
  [SMALL_STATE(884)] = 31535,
  [SMALL_STATE(885)] = 31593,
  [SMALL_STATE(886)] = 31657,
  [SMALL_STATE(887)] = 31723,
  [SMALL_STATE(888)] = 31781,
  [SMALL_STATE(889)] = 31839,
  [SMALL_STATE(890)] = 31897,
  [SMALL_STATE(891)] = 31965,
  [SMALL_STATE(892)] = 32029,
  [SMALL_STATE(893)] = 32097,
  [SMALL_STATE(894)] = 32161,
  [SMALL_STATE(895)] = 32225,
  [SMALL_STATE(896)] = 32321,
  [SMALL_STATE(897)] = 32388,
  [SMALL_STATE(898)] = 32483,
  [SMALL_STATE(899)] = 32540,
  [SMALL_STATE(900)] = 32607,
  [SMALL_STATE(901)] = 32664,
  [SMALL_STATE(902)] = 32721,
  [SMALL_STATE(903)] = 32788,
  [SMALL_STATE(904)] = 32883,
  [SMALL_STATE(905)] = 32950,
  [SMALL_STATE(906)] = 33013,
  [SMALL_STATE(907)] = 33080,
  [SMALL_STATE(908)] = 33145,
  [SMALL_STATE(909)] = 33238,
  [SMALL_STATE(910)] = 33331,
  [SMALL_STATE(911)] = 33388,
  [SMALL_STATE(912)] = 33455,
  [SMALL_STATE(913)] = 33512,
  [SMALL_STATE(914)] = 33589,
  [SMALL_STATE(915)] = 33646,
  [SMALL_STATE(916)] = 33703,
  [SMALL_STATE(917)] = 33760,
  [SMALL_STATE(918)] = 33817,
  [SMALL_STATE(919)] = 33874,
  [SMALL_STATE(920)] = 33931,
  [SMALL_STATE(921)] = 33994,
  [SMALL_STATE(922)] = 34051,
  [SMALL_STATE(923)] = 34108,
  [SMALL_STATE(924)] = 34165,
  [SMALL_STATE(925)] = 34222,
  [SMALL_STATE(926)] = 34333,
  [SMALL_STATE(927)] = 34394,
  [SMALL_STATE(928)] = 34451,
  [SMALL_STATE(929)] = 34508,
  [SMALL_STATE(930)] = 34565,
  [SMALL_STATE(931)] = 34622,
  [SMALL_STATE(932)] = 34679,
  [SMALL_STATE(933)] = 34742,
  [SMALL_STATE(934)] = 34853,
  [SMALL_STATE(935)] = 34910,
  [SMALL_STATE(936)] = 34973,
  [SMALL_STATE(937)] = 35068,
  [SMALL_STATE(938)] = 35179,
  [SMALL_STATE(939)] = 35242,
  [SMALL_STATE(940)] = 35337,
  [SMALL_STATE(941)] = 35402,
  [SMALL_STATE(942)] = 35469,
  [SMALL_STATE(943)] = 35536,
  [SMALL_STATE(944)] = 35603,
  [SMALL_STATE(945)] = 35698,
  [SMALL_STATE(946)] = 35775,
  [SMALL_STATE(947)] = 35842,
  [SMALL_STATE(948)] = 35905,
  [SMALL_STATE(949)] = 35976,
  [SMALL_STATE(950)] = 36039,
  [SMALL_STATE(951)] = 36136,
  [SMALL_STATE(952)] = 36247,
  [SMALL_STATE(953)] = 36308,
  [SMALL_STATE(954)] = 36375,
  [SMALL_STATE(955)] = 36442,
  [SMALL_STATE(956)] = 36553,
  [SMALL_STATE(957)] = 36620,
  [SMALL_STATE(958)] = 36683,
  [SMALL_STATE(959)] = 36794,
  [SMALL_STATE(960)] = 36857,
  [SMALL_STATE(961)] = 36924,
  [SMALL_STATE(962)] = 36993,
  [SMALL_STATE(963)] = 37090,
  [SMALL_STATE(964)] = 37147,
  [SMALL_STATE(965)] = 37204,
  [SMALL_STATE(966)] = 37267,
  [SMALL_STATE(967)] = 37330,
  [SMALL_STATE(968)] = 37425,
  [SMALL_STATE(969)] = 37482,
  [SMALL_STATE(970)] = 37545,
  [SMALL_STATE(971)] = 37608,
  [SMALL_STATE(972)] = 37703,
  [SMALL_STATE(973)] = 37796,
  [SMALL_STATE(974)] = 37889,
  [SMALL_STATE(975)] = 37952,
  [SMALL_STATE(976)] = 38009,
  [SMALL_STATE(977)] = 38102,
  [SMALL_STATE(978)] = 38159,
  [SMALL_STATE(979)] = 38222,
  [SMALL_STATE(980)] = 38319,
  [SMALL_STATE(981)] = 38376,
  [SMALL_STATE(982)] = 38471,
  [SMALL_STATE(983)] = 38528,
  [SMALL_STATE(984)] = 38591,
  [SMALL_STATE(985)] = 38648,
  [SMALL_STATE(986)] = 38711,
  [SMALL_STATE(987)] = 38780,
  [SMALL_STATE(988)] = 38837,
  [SMALL_STATE(989)] = 38894,
  [SMALL_STATE(990)] = 39005,
  [SMALL_STATE(991)] = 39062,
  [SMALL_STATE(992)] = 39139,
  [SMALL_STATE(993)] = 39202,
  [SMALL_STATE(994)] = 39265,
  [SMALL_STATE(995)] = 39328,
  [SMALL_STATE(996)] = 39391,
  [SMALL_STATE(997)] = 39454,
  [SMALL_STATE(998)] = 39565,
  [SMALL_STATE(999)] = 39622,
  [SMALL_STATE(1000)] = 39679,
  [SMALL_STATE(1001)] = 39742,
  [SMALL_STATE(1002)] = 39805,
  [SMALL_STATE(1003)] = 39882,
  [SMALL_STATE(1004)] = 39945,
  [SMALL_STATE(1005)] = 40002,
  [SMALL_STATE(1006)] = 40059,
  [SMALL_STATE(1007)] = 40130,
  [SMALL_STATE(1008)] = 40193,
  [SMALL_STATE(1009)] = 40250,
  [SMALL_STATE(1010)] = 40307,
  [SMALL_STATE(1011)] = 40373,
  [SMALL_STATE(1012)] = 40435,
  [SMALL_STATE(1013)] = 40499,
  [SMALL_STATE(1014)] = 40555,
  [SMALL_STATE(1015)] = 40615,
  [SMALL_STATE(1016)] = 40671,
  [SMALL_STATE(1017)] = 40727,
  [SMALL_STATE(1018)] = 40783,
  [SMALL_STATE(1019)] = 40839,
  [SMALL_STATE(1020)] = 40895,
  [SMALL_STATE(1021)] = 40951,
  [SMALL_STATE(1022)] = 41013,
  [SMALL_STATE(1023)] = 41069,
  [SMALL_STATE(1024)] = 41131,
  [SMALL_STATE(1025)] = 41197,
  [SMALL_STATE(1026)] = 41263,
  [SMALL_STATE(1027)] = 41323,
  [SMALL_STATE(1028)] = 41379,
  [SMALL_STATE(1029)] = 41447,
  [SMALL_STATE(1030)] = 41503,
  [SMALL_STATE(1031)] = 41577,
  [SMALL_STATE(1032)] = 41633,
  [SMALL_STATE(1033)] = 41695,
  [SMALL_STATE(1034)] = 41751,
  [SMALL_STATE(1035)] = 41807,
  [SMALL_STATE(1036)] = 41901,
  [SMALL_STATE(1037)] = 41975,
  [SMALL_STATE(1038)] = 42031,
  [SMALL_STATE(1039)] = 42087,
  [SMALL_STATE(1040)] = 42149,
  [SMALL_STATE(1041)] = 42211,
  [SMALL_STATE(1042)] = 42277,
  [SMALL_STATE(1043)] = 42339,
  [SMALL_STATE(1044)] = 42395,
  [SMALL_STATE(1045)] = 42451,
  [SMALL_STATE(1046)] = 42511,
  [SMALL_STATE(1047)] = 42567,
  [SMALL_STATE(1048)] = 42623,
  [SMALL_STATE(1049)] = 42689,
  [SMALL_STATE(1050)] = 42745,
  [SMALL_STATE(1051)] = 42811,
  [SMALL_STATE(1052)] = 42877,
  [SMALL_STATE(1053)] = 42933,
  [SMALL_STATE(1054)] = 42999,
  [SMALL_STATE(1055)] = 43061,
  [SMALL_STATE(1056)] = 43123,
  [SMALL_STATE(1057)] = 43217,
  [SMALL_STATE(1058)] = 43273,
  [SMALL_STATE(1059)] = 43329,
  [SMALL_STATE(1060)] = 43399,
  [SMALL_STATE(1061)] = 43455,
  [SMALL_STATE(1062)] = 43511,
  [SMALL_STATE(1063)] = 43567,
  [SMALL_STATE(1064)] = 43623,
  [SMALL_STATE(1065)] = 43679,
  [SMALL_STATE(1066)] = 43741,
  [SMALL_STATE(1067)] = 43803,
  [SMALL_STATE(1068)] = 43865,
  [SMALL_STATE(1069)] = 43927,
  [SMALL_STATE(1070)] = 43989,
  [SMALL_STATE(1071)] = 44055,
  [SMALL_STATE(1072)] = 44121,
  [SMALL_STATE(1073)] = 44183,
  [SMALL_STATE(1074)] = 44245,
  [SMALL_STATE(1075)] = 44307,
  [SMALL_STATE(1076)] = 44401,
  [SMALL_STATE(1077)] = 44463,
  [SMALL_STATE(1078)] = 44525,
  [SMALL_STATE(1079)] = 44587,
  [SMALL_STATE(1080)] = 44649,
  [SMALL_STATE(1081)] = 44713,
  [SMALL_STATE(1082)] = 44775,
  [SMALL_STATE(1083)] = 44831,
  [SMALL_STATE(1084)] = 44887,
  [SMALL_STATE(1085)] = 44943,
  [SMALL_STATE(1086)] = 44999,
  [SMALL_STATE(1087)] = 45061,
  [SMALL_STATE(1088)] = 45123,
  [SMALL_STATE(1089)] = 45185,
  [SMALL_STATE(1090)] = 45247,
  [SMALL_STATE(1091)] = 45341,
  [SMALL_STATE(1092)] = 45403,
  [SMALL_STATE(1093)] = 45465,
  [SMALL_STATE(1094)] = 45527,
  [SMALL_STATE(1095)] = 45589,
  [SMALL_STATE(1096)] = 45649,
  [SMALL_STATE(1097)] = 45719,
  [SMALL_STATE(1098)] = 45779,
  [SMALL_STATE(1099)] = 45847,
  [SMALL_STATE(1100)] = 45903,
  [SMALL_STATE(1101)] = 45965,
  [SMALL_STATE(1102)] = 46021,
  [SMALL_STATE(1103)] = 46115,
  [SMALL_STATE(1104)] = 46171,
  [SMALL_STATE(1105)] = 46227,
  [SMALL_STATE(1106)] = 46289,
  [SMALL_STATE(1107)] = 46351,
  [SMALL_STATE(1108)] = 46413,
  [SMALL_STATE(1109)] = 46475,
  [SMALL_STATE(1110)] = 46537,
  [SMALL_STATE(1111)] = 46629,
  [SMALL_STATE(1112)] = 46721,
  [SMALL_STATE(1113)] = 46777,
  [SMALL_STATE(1114)] = 46833,
  [SMALL_STATE(1115)] = 46899,
  [SMALL_STATE(1116)] = 46965,
  [SMALL_STATE(1117)] = 47021,
  [SMALL_STATE(1118)] = 47115,
  [SMALL_STATE(1119)] = 47209,
  [SMALL_STATE(1120)] = 47265,
  [SMALL_STATE(1121)] = 47321,
  [SMALL_STATE(1122)] = 47383,
  [SMALL_STATE(1123)] = 47445,
  [SMALL_STATE(1124)] = 47507,
  [SMALL_STATE(1125)] = 47569,
  [SMALL_STATE(1126)] = 47625,
  [SMALL_STATE(1127)] = 47681,
  [SMALL_STATE(1128)] = 47743,
  [SMALL_STATE(1129)] = 47799,
  [SMALL_STATE(1130)] = 47855,
  [SMALL_STATE(1131)] = 47911,
  [SMALL_STATE(1132)] = 47973,
  [SMALL_STATE(1133)] = 48049,
  [SMALL_STATE(1134)] = 48105,
  [SMALL_STATE(1135)] = 48171,
  [SMALL_STATE(1136)] = 48237,
  [SMALL_STATE(1137)] = 48293,
  [SMALL_STATE(1138)] = 48349,
  [SMALL_STATE(1139)] = 48405,
  [SMALL_STATE(1140)] = 48461,
  [SMALL_STATE(1141)] = 48523,
  [SMALL_STATE(1142)] = 48589,
  [SMALL_STATE(1143)] = 48645,
  [SMALL_STATE(1144)] = 48701,
  [SMALL_STATE(1145)] = 48763,
  [SMALL_STATE(1146)] = 48819,
  [SMALL_STATE(1147)] = 48879,
  [SMALL_STATE(1148)] = 48941,
  [SMALL_STATE(1149)] = 49003,
  [SMALL_STATE(1150)] = 49063,
  [SMALL_STATE(1151)] = 49139,
  [SMALL_STATE(1152)] = 49213,
  [SMALL_STATE(1153)] = 49269,
  [SMALL_STATE(1154)] = 49343,
  [SMALL_STATE(1155)] = 49409,
  [SMALL_STATE(1156)] = 49465,
  [SMALL_STATE(1157)] = 49527,
  [SMALL_STATE(1158)] = 49583,
  [SMALL_STATE(1159)] = 49645,
  [SMALL_STATE(1160)] = 49701,
  [SMALL_STATE(1161)] = 49763,
  [SMALL_STATE(1162)] = 49819,
  [SMALL_STATE(1163)] = 49881,
  [SMALL_STATE(1164)] = 49943,
  [SMALL_STATE(1165)] = 50008,
  [SMALL_STATE(1166)] = 50063,
  [SMALL_STATE(1167)] = 50124,
  [SMALL_STATE(1168)] = 50185,
  [SMALL_STATE(1169)] = 50240,
  [SMALL_STATE(1170)] = 50295,
  [SMALL_STATE(1171)] = 50356,
  [SMALL_STATE(1172)] = 50417,
  [SMALL_STATE(1173)] = 50510,
  [SMALL_STATE(1174)] = 50577,
  [SMALL_STATE(1175)] = 50638,
  [SMALL_STATE(1176)] = 50697,
  [SMALL_STATE(1177)] = 50752,
  [SMALL_STATE(1178)] = 50813,
  [SMALL_STATE(1179)] = 50906,
  [SMALL_STATE(1180)] = 50965,
  [SMALL_STATE(1181)] = 51024,
  [SMALL_STATE(1182)] = 51079,
  [SMALL_STATE(1183)] = 51134,
  [SMALL_STATE(1184)] = 51189,
  [SMALL_STATE(1185)] = 51250,
  [SMALL_STATE(1186)] = 51305,
  [SMALL_STATE(1187)] = 51360,
  [SMALL_STATE(1188)] = 51415,
  [SMALL_STATE(1189)] = 51470,
  [SMALL_STATE(1190)] = 51525,
  [SMALL_STATE(1191)] = 51586,
  [SMALL_STATE(1192)] = 51647,
  [SMALL_STATE(1193)] = 51708,
  [SMALL_STATE(1194)] = 51769,
  [SMALL_STATE(1195)] = 51824,
  [SMALL_STATE(1196)] = 51879,
  [SMALL_STATE(1197)] = 51934,
  [SMALL_STATE(1198)] = 51999,
  [SMALL_STATE(1199)] = 52054,
  [SMALL_STATE(1200)] = 52109,
  [SMALL_STATE(1201)] = 52164,
  [SMALL_STATE(1202)] = 52223,
  [SMALL_STATE(1203)] = 52282,
  [SMALL_STATE(1204)] = 52343,
  [SMALL_STATE(1205)] = 52398,
  [SMALL_STATE(1206)] = 52453,
  [SMALL_STATE(1207)] = 52508,
  [SMALL_STATE(1208)] = 52563,
  [SMALL_STATE(1209)] = 52618,
  [SMALL_STATE(1210)] = 52677,
  [SMALL_STATE(1211)] = 52732,
  [SMALL_STATE(1212)] = 52793,
  [SMALL_STATE(1213)] = 52852,
  [SMALL_STATE(1214)] = 52907,
  [SMALL_STATE(1215)] = 52968,
  [SMALL_STATE(1216)] = 53029,
  [SMALL_STATE(1217)] = 53084,
  [SMALL_STATE(1218)] = 53145,
  [SMALL_STATE(1219)] = 53200,
  [SMALL_STATE(1220)] = 53261,
  [SMALL_STATE(1221)] = 53316,
  [SMALL_STATE(1222)] = 53371,
  [SMALL_STATE(1223)] = 53426,
  [SMALL_STATE(1224)] = 53481,
  [SMALL_STATE(1225)] = 53536,
  [SMALL_STATE(1226)] = 53591,
  [SMALL_STATE(1227)] = 53652,
  [SMALL_STATE(1228)] = 53713,
  [SMALL_STATE(1229)] = 53774,
  [SMALL_STATE(1230)] = 53829,
  [SMALL_STATE(1231)] = 53890,
  [SMALL_STATE(1232)] = 53945,
  [SMALL_STATE(1233)] = 54000,
  [SMALL_STATE(1234)] = 54055,
  [SMALL_STATE(1235)] = 54116,
  [SMALL_STATE(1236)] = 54171,
  [SMALL_STATE(1237)] = 54226,
  [SMALL_STATE(1238)] = 54281,
  [SMALL_STATE(1239)] = 54336,
  [SMALL_STATE(1240)] = 54391,
  [SMALL_STATE(1241)] = 54446,
  [SMALL_STATE(1242)] = 54501,
  [SMALL_STATE(1243)] = 54556,
  [SMALL_STATE(1244)] = 54617,
  [SMALL_STATE(1245)] = 54678,
  [SMALL_STATE(1246)] = 54739,
  [SMALL_STATE(1247)] = 54794,
  [SMALL_STATE(1248)] = 54849,
  [SMALL_STATE(1249)] = 54904,
  [SMALL_STATE(1250)] = 54965,
  [SMALL_STATE(1251)] = 55026,
  [SMALL_STATE(1252)] = 55087,
  [SMALL_STATE(1253)] = 55142,
  [SMALL_STATE(1254)] = 55197,
  [SMALL_STATE(1255)] = 55252,
  [SMALL_STATE(1256)] = 55307,
  [SMALL_STATE(1257)] = 55362,
  [SMALL_STATE(1258)] = 55423,
  [SMALL_STATE(1259)] = 55484,
  [SMALL_STATE(1260)] = 55539,
  [SMALL_STATE(1261)] = 55594,
  [SMALL_STATE(1262)] = 55655,
  [SMALL_STATE(1263)] = 55716,
  [SMALL_STATE(1264)] = 55777,
  [SMALL_STATE(1265)] = 55838,
  [SMALL_STATE(1266)] = 55899,
  [SMALL_STATE(1267)] = 55960,
  [SMALL_STATE(1268)] = 56019,
  [SMALL_STATE(1269)] = 56074,
  [SMALL_STATE(1270)] = 56135,
  [SMALL_STATE(1271)] = 56196,
  [SMALL_STATE(1272)] = 56257,
  [SMALL_STATE(1273)] = 56312,
  [SMALL_STATE(1274)] = 56373,
  [SMALL_STATE(1275)] = 56434,
  [SMALL_STATE(1276)] = 56495,
  [SMALL_STATE(1277)] = 56556,
  [SMALL_STATE(1278)] = 56611,
  [SMALL_STATE(1279)] = 56666,
  [SMALL_STATE(1280)] = 56721,
  [SMALL_STATE(1281)] = 56776,
  [SMALL_STATE(1282)] = 56831,
  [SMALL_STATE(1283)] = 56886,
  [SMALL_STATE(1284)] = 56941,
  [SMALL_STATE(1285)] = 56996,
  [SMALL_STATE(1286)] = 57051,
  [SMALL_STATE(1287)] = 57106,
  [SMALL_STATE(1288)] = 57161,
  [SMALL_STATE(1289)] = 57216,
  [SMALL_STATE(1290)] = 57271,
  [SMALL_STATE(1291)] = 57332,
  [SMALL_STATE(1292)] = 57393,
  [SMALL_STATE(1293)] = 57448,
  [SMALL_STATE(1294)] = 57503,
  [SMALL_STATE(1295)] = 57558,
  [SMALL_STATE(1296)] = 57617,
  [SMALL_STATE(1297)] = 57672,
  [SMALL_STATE(1298)] = 57727,
  [SMALL_STATE(1299)] = 57782,
  [SMALL_STATE(1300)] = 57837,
  [SMALL_STATE(1301)] = 57892,
  [SMALL_STATE(1302)] = 57947,
  [SMALL_STATE(1303)] = 58002,
  [SMALL_STATE(1304)] = 58057,
  [SMALL_STATE(1305)] = 58112,
  [SMALL_STATE(1306)] = 58167,
  [SMALL_STATE(1307)] = 58222,
  [SMALL_STATE(1308)] = 58282,
  [SMALL_STATE(1309)] = 58342,
  [SMALL_STATE(1310)] = 58396,
  [SMALL_STATE(1311)] = 58450,
  [SMALL_STATE(1312)] = 58510,
  [SMALL_STATE(1313)] = 58576,
  [SMALL_STATE(1314)] = 58634,
  [SMALL_STATE(1315)] = 58724,
  [SMALL_STATE(1316)] = 58778,
  [SMALL_STATE(1317)] = 58832,
  [SMALL_STATE(1318)] = 58922,
  [SMALL_STATE(1319)] = 58982,
  [SMALL_STATE(1320)] = 59036,
  [SMALL_STATE(1321)] = 59096,
  [SMALL_STATE(1322)] = 59154,
  [SMALL_STATE(1323)] = 59214,
  [SMALL_STATE(1324)] = 59268,
  [SMALL_STATE(1325)] = 59322,
  [SMALL_STATE(1326)] = 59382,
  [SMALL_STATE(1327)] = 59442,
  [SMALL_STATE(1328)] = 59496,
  [SMALL_STATE(1329)] = 59550,
  [SMALL_STATE(1330)] = 59604,
  [SMALL_STATE(1331)] = 59664,
  [SMALL_STATE(1332)] = 59724,
  [SMALL_STATE(1333)] = 59814,
  [SMALL_STATE(1334)] = 59874,
  [SMALL_STATE(1335)] = 59928,
  [SMALL_STATE(1336)] = 59986,
  [SMALL_STATE(1337)] = 60046,
  [SMALL_STATE(1338)] = 60106,
  [SMALL_STATE(1339)] = 60160,
  [SMALL_STATE(1340)] = 60220,
  [SMALL_STATE(1341)] = 60274,
  [SMALL_STATE(1342)] = 60334,
  [SMALL_STATE(1343)] = 60394,
  [SMALL_STATE(1344)] = 60454,
  [SMALL_STATE(1345)] = 60514,
  [SMALL_STATE(1346)] = 60568,
  [SMALL_STATE(1347)] = 60622,
  [SMALL_STATE(1348)] = 60682,
  [SMALL_STATE(1349)] = 60736,
  [SMALL_STATE(1350)] = 60794,
  [SMALL_STATE(1351)] = 60848,
  [SMALL_STATE(1352)] = 60902,
  [SMALL_STATE(1353)] = 60956,
  [SMALL_STATE(1354)] = 61010,
  [SMALL_STATE(1355)] = 61070,
  [SMALL_STATE(1356)] = 61130,
  [SMALL_STATE(1357)] = 61190,
  [SMALL_STATE(1358)] = 61250,
  [SMALL_STATE(1359)] = 61308,
  [SMALL_STATE(1360)] = 61362,
  [SMALL_STATE(1361)] = 61422,
  [SMALL_STATE(1362)] = 61480,
  [SMALL_STATE(1363)] = 61540,
  [SMALL_STATE(1364)] = 61600,
  [SMALL_STATE(1365)] = 61654,
  [SMALL_STATE(1366)] = 61708,
  [SMALL_STATE(1367)] = 61762,
  [SMALL_STATE(1368)] = 61822,
  [SMALL_STATE(1369)] = 61880,
  [SMALL_STATE(1370)] = 61940,
  [SMALL_STATE(1371)] = 62000,
  [SMALL_STATE(1372)] = 62060,
  [SMALL_STATE(1373)] = 62120,
  [SMALL_STATE(1374)] = 62180,
  [SMALL_STATE(1375)] = 62238,
  [SMALL_STATE(1376)] = 62292,
  [SMALL_STATE(1377)] = 62346,
  [SMALL_STATE(1378)] = 62404,
  [SMALL_STATE(1379)] = 62458,
  [SMALL_STATE(1380)] = 62516,
  [SMALL_STATE(1381)] = 62574,
  [SMALL_STATE(1382)] = 62634,
  [SMALL_STATE(1383)] = 62692,
  [SMALL_STATE(1384)] = 62752,
  [SMALL_STATE(1385)] = 62806,
  [SMALL_STATE(1386)] = 62860,
  [SMALL_STATE(1387)] = 62914,
  [SMALL_STATE(1388)] = 62968,
  [SMALL_STATE(1389)] = 63028,
  [SMALL_STATE(1390)] = 63082,
  [SMALL_STATE(1391)] = 63136,
  [SMALL_STATE(1392)] = 63196,
  [SMALL_STATE(1393)] = 63256,
  [SMALL_STATE(1394)] = 63310,
  [SMALL_STATE(1395)] = 63370,
  [SMALL_STATE(1396)] = 63430,
  [SMALL_STATE(1397)] = 63490,
  [SMALL_STATE(1398)] = 63550,
  [SMALL_STATE(1399)] = 63610,
  [SMALL_STATE(1400)] = 63670,
  [SMALL_STATE(1401)] = 63724,
  [SMALL_STATE(1402)] = 63782,
  [SMALL_STATE(1403)] = 63836,
  [SMALL_STATE(1404)] = 63890,
  [SMALL_STATE(1405)] = 63944,
  [SMALL_STATE(1406)] = 64004,
  [SMALL_STATE(1407)] = 64058,
  [SMALL_STATE(1408)] = 64118,
  [SMALL_STATE(1409)] = 64172,
  [SMALL_STATE(1410)] = 64226,
  [SMALL_STATE(1411)] = 64280,
  [SMALL_STATE(1412)] = 64334,
  [SMALL_STATE(1413)] = 64394,
  [SMALL_STATE(1414)] = 64448,
  [SMALL_STATE(1415)] = 64502,
  [SMALL_STATE(1416)] = 64562,
  [SMALL_STATE(1417)] = 64616,
  [SMALL_STATE(1418)] = 64670,
  [SMALL_STATE(1419)] = 64724,
  [SMALL_STATE(1420)] = 64778,
  [SMALL_STATE(1421)] = 64838,
  [SMALL_STATE(1422)] = 64898,
  [SMALL_STATE(1423)] = 64952,
  [SMALL_STATE(1424)] = 65006,
  [SMALL_STATE(1425)] = 65060,
  [SMALL_STATE(1426)] = 65124,
  [SMALL_STATE(1427)] = 65188,
  [SMALL_STATE(1428)] = 65242,
  [SMALL_STATE(1429)] = 65296,
  [SMALL_STATE(1430)] = 65354,
  [SMALL_STATE(1431)] = 65408,
  [SMALL_STATE(1432)] = 65466,
  [SMALL_STATE(1433)] = 65520,
  [SMALL_STATE(1434)] = 65578,
  [SMALL_STATE(1435)] = 65632,
  [SMALL_STATE(1436)] = 65686,
  [SMALL_STATE(1437)] = 65744,
  [SMALL_STATE(1438)] = 65798,
  [SMALL_STATE(1439)] = 65852,
  [SMALL_STATE(1440)] = 65906,
  [SMALL_STATE(1441)] = 65960,
  [SMALL_STATE(1442)] = 66020,
  [SMALL_STATE(1443)] = 66074,
  [SMALL_STATE(1444)] = 66134,
  [SMALL_STATE(1445)] = 66194,
  [SMALL_STATE(1446)] = 66254,
  [SMALL_STATE(1447)] = 66314,
  [SMALL_STATE(1448)] = 66374,
  [SMALL_STATE(1449)] = 66432,
  [SMALL_STATE(1450)] = 66486,
  [SMALL_STATE(1451)] = 66540,
  [SMALL_STATE(1452)] = 66600,
  [SMALL_STATE(1453)] = 66654,
  [SMALL_STATE(1454)] = 66708,
  [SMALL_STATE(1455)] = 66762,
  [SMALL_STATE(1456)] = 66816,
  [SMALL_STATE(1457)] = 66870,
  [SMALL_STATE(1458)] = 66930,
  [SMALL_STATE(1459)] = 66984,
  [SMALL_STATE(1460)] = 67038,
  [SMALL_STATE(1461)] = 67098,
  [SMALL_STATE(1462)] = 67152,
  [SMALL_STATE(1463)] = 67206,
  [SMALL_STATE(1464)] = 67260,
  [SMALL_STATE(1465)] = 67314,
  [SMALL_STATE(1466)] = 67368,
  [SMALL_STATE(1467)] = 67422,
  [SMALL_STATE(1468)] = 67482,
  [SMALL_STATE(1469)] = 67542,
  [SMALL_STATE(1470)] = 67596,
  [SMALL_STATE(1471)] = 67656,
  [SMALL_STATE(1472)] = 67710,
  [SMALL_STATE(1473)] = 67764,
  [SMALL_STATE(1474)] = 67818,
  [SMALL_STATE(1475)] = 67876,
  [SMALL_STATE(1476)] = 67936,
  [SMALL_STATE(1477)] = 67990,
  [SMALL_STATE(1478)] = 68044,
  [SMALL_STATE(1479)] = 68098,
  [SMALL_STATE(1480)] = 68152,
  [SMALL_STATE(1481)] = 68210,
  [SMALL_STATE(1482)] = 68269,
  [SMALL_STATE(1483)] = 68322,
  [SMALL_STATE(1484)] = 68375,
  [SMALL_STATE(1485)] = 68428,
  [SMALL_STATE(1486)] = 68481,
  [SMALL_STATE(1487)] = 68534,
  [SMALL_STATE(1488)] = 68587,
  [SMALL_STATE(1489)] = 68640,
  [SMALL_STATE(1490)] = 68693,
  [SMALL_STATE(1491)] = 68746,
  [SMALL_STATE(1492)] = 68799,
  [SMALL_STATE(1493)] = 68852,
  [SMALL_STATE(1494)] = 68905,
  [SMALL_STATE(1495)] = 68958,
  [SMALL_STATE(1496)] = 69011,
  [SMALL_STATE(1497)] = 69064,
  [SMALL_STATE(1498)] = 69117,
  [SMALL_STATE(1499)] = 69170,
  [SMALL_STATE(1500)] = 69223,
  [SMALL_STATE(1501)] = 69276,
  [SMALL_STATE(1502)] = 69329,
  [SMALL_STATE(1503)] = 69382,
  [SMALL_STATE(1504)] = 69435,
  [SMALL_STATE(1505)] = 69488,
  [SMALL_STATE(1506)] = 69541,
  [SMALL_STATE(1507)] = 69594,
  [SMALL_STATE(1508)] = 69647,
  [SMALL_STATE(1509)] = 69700,
  [SMALL_STATE(1510)] = 69757,
  [SMALL_STATE(1511)] = 69810,
  [SMALL_STATE(1512)] = 69863,
  [SMALL_STATE(1513)] = 69922,
  [SMALL_STATE(1514)] = 69975,
  [SMALL_STATE(1515)] = 70028,
  [SMALL_STATE(1516)] = 70081,
  [SMALL_STATE(1517)] = 70134,
  [SMALL_STATE(1518)] = 70187,
  [SMALL_STATE(1519)] = 70240,
  [SMALL_STATE(1520)] = 70293,
  [SMALL_STATE(1521)] = 70346,
  [SMALL_STATE(1522)] = 70399,
  [SMALL_STATE(1523)] = 70456,
  [SMALL_STATE(1524)] = 70515,
  [SMALL_STATE(1525)] = 70574,
  [SMALL_STATE(1526)] = 70627,
  [SMALL_STATE(1527)] = 70686,
  [SMALL_STATE(1528)] = 70739,
  [SMALL_STATE(1529)] = 70798,
  [SMALL_STATE(1530)] = 70857,
  [SMALL_STATE(1531)] = 70910,
  [SMALL_STATE(1532)] = 70963,
  [SMALL_STATE(1533)] = 71016,
  [SMALL_STATE(1534)] = 71069,
  [SMALL_STATE(1535)] = 71122,
  [SMALL_STATE(1536)] = 71175,
  [SMALL_STATE(1537)] = 71234,
  [SMALL_STATE(1538)] = 71287,
  [SMALL_STATE(1539)] = 71340,
  [SMALL_STATE(1540)] = 71397,
  [SMALL_STATE(1541)] = 71450,
  [SMALL_STATE(1542)] = 71503,
  [SMALL_STATE(1543)] = 71556,
  [SMALL_STATE(1544)] = 71609,
  [SMALL_STATE(1545)] = 71662,
  [SMALL_STATE(1546)] = 71715,
  [SMALL_STATE(1547)] = 71768,
  [SMALL_STATE(1548)] = 71821,
  [SMALL_STATE(1549)] = 71874,
  [SMALL_STATE(1550)] = 71927,
  [SMALL_STATE(1551)] = 71980,
  [SMALL_STATE(1552)] = 72033,
  [SMALL_STATE(1553)] = 72086,
  [SMALL_STATE(1554)] = 72139,
  [SMALL_STATE(1555)] = 72192,
  [SMALL_STATE(1556)] = 72245,
  [SMALL_STATE(1557)] = 72298,
  [SMALL_STATE(1558)] = 72357,
  [SMALL_STATE(1559)] = 72416,
  [SMALL_STATE(1560)] = 72475,
  [SMALL_STATE(1561)] = 72534,
  [SMALL_STATE(1562)] = 72587,
  [SMALL_STATE(1563)] = 72654,
  [SMALL_STATE(1564)] = 72707,
  [SMALL_STATE(1565)] = 72760,
  [SMALL_STATE(1566)] = 72813,
  [SMALL_STATE(1567)] = 72866,
  [SMALL_STATE(1568)] = 72923,
  [SMALL_STATE(1569)] = 72976,
  [SMALL_STATE(1570)] = 73029,
  [SMALL_STATE(1571)] = 73082,
  [SMALL_STATE(1572)] = 73135,
  [SMALL_STATE(1573)] = 73194,
  [SMALL_STATE(1574)] = 73247,
  [SMALL_STATE(1575)] = 73300,
  [SMALL_STATE(1576)] = 73359,
  [SMALL_STATE(1577)] = 73462,
  [SMALL_STATE(1578)] = 73515,
  [SMALL_STATE(1579)] = 73568,
  [SMALL_STATE(1580)] = 73621,
  [SMALL_STATE(1581)] = 73674,
  [SMALL_STATE(1582)] = 73727,
  [SMALL_STATE(1583)] = 73786,
  [SMALL_STATE(1584)] = 73839,
  [SMALL_STATE(1585)] = 73892,
  [SMALL_STATE(1586)] = 73945,
  [SMALL_STATE(1587)] = 73998,
  [SMALL_STATE(1588)] = 74051,
  [SMALL_STATE(1589)] = 74104,
  [SMALL_STATE(1590)] = 74157,
  [SMALL_STATE(1591)] = 74210,
  [SMALL_STATE(1592)] = 74263,
  [SMALL_STATE(1593)] = 74316,
  [SMALL_STATE(1594)] = 74369,
  [SMALL_STATE(1595)] = 74422,
  [SMALL_STATE(1596)] = 74475,
  [SMALL_STATE(1597)] = 74534,
  [SMALL_STATE(1598)] = 74593,
  [SMALL_STATE(1599)] = 74646,
  [SMALL_STATE(1600)] = 74705,
  [SMALL_STATE(1601)] = 74758,
  [SMALL_STATE(1602)] = 74817,
  [SMALL_STATE(1603)] = 74876,
  [SMALL_STATE(1604)] = 74941,
  [SMALL_STATE(1605)] = 74998,
  [SMALL_STATE(1606)] = 75051,
  [SMALL_STATE(1607)] = 75108,
  [SMALL_STATE(1608)] = 75161,
  [SMALL_STATE(1609)] = 75218,
  [SMALL_STATE(1610)] = 75271,
  [SMALL_STATE(1611)] = 75324,
  [SMALL_STATE(1612)] = 75377,
  [SMALL_STATE(1613)] = 75430,
  [SMALL_STATE(1614)] = 75483,
  [SMALL_STATE(1615)] = 75536,
  [SMALL_STATE(1616)] = 75639,
  [SMALL_STATE(1617)] = 75692,
  [SMALL_STATE(1618)] = 75745,
  [SMALL_STATE(1619)] = 75802,
  [SMALL_STATE(1620)] = 75855,
  [SMALL_STATE(1621)] = 75958,
  [SMALL_STATE(1622)] = 76011,
  [SMALL_STATE(1623)] = 76064,
  [SMALL_STATE(1624)] = 76117,
  [SMALL_STATE(1625)] = 76170,
  [SMALL_STATE(1626)] = 76223,
  [SMALL_STATE(1627)] = 76276,
  [SMALL_STATE(1628)] = 76341,
  [SMALL_STATE(1629)] = 76394,
  [SMALL_STATE(1630)] = 76447,
  [SMALL_STATE(1631)] = 76506,
  [SMALL_STATE(1632)] = 76563,
  [SMALL_STATE(1633)] = 76622,
  [SMALL_STATE(1634)] = 76675,
  [SMALL_STATE(1635)] = 76736,
  [SMALL_STATE(1636)] = 76797,
  [SMALL_STATE(1637)] = 76900,
  [SMALL_STATE(1638)] = 76959,
  [SMALL_STATE(1639)] = 77018,
  [SMALL_STATE(1640)] = 77071,
  [SMALL_STATE(1641)] = 77130,
  [SMALL_STATE(1642)] = 77183,
  [SMALL_STATE(1643)] = 77236,
  [SMALL_STATE(1644)] = 77293,
  [SMALL_STATE(1645)] = 77346,
  [SMALL_STATE(1646)] = 77405,
  [SMALL_STATE(1647)] = 77462,
  [SMALL_STATE(1648)] = 77521,
  [SMALL_STATE(1649)] = 77586,
  [SMALL_STATE(1650)] = 77645,
  [SMALL_STATE(1651)] = 77704,
  [SMALL_STATE(1652)] = 77757,
  [SMALL_STATE(1653)] = 77810,
  [SMALL_STATE(1654)] = 77869,
  [SMALL_STATE(1655)] = 77958,
  [SMALL_STATE(1656)] = 78011,
  [SMALL_STATE(1657)] = 78064,
  [SMALL_STATE(1658)] = 78117,
  [SMALL_STATE(1659)] = 78174,
  [SMALL_STATE(1660)] = 78233,
  [SMALL_STATE(1661)] = 78292,
  [SMALL_STATE(1662)] = 78351,
  [SMALL_STATE(1663)] = 78410,
  [SMALL_STATE(1664)] = 78469,
  [SMALL_STATE(1665)] = 78528,
  [SMALL_STATE(1666)] = 78587,
  [SMALL_STATE(1667)] = 78640,
  [SMALL_STATE(1668)] = 78693,
  [SMALL_STATE(1669)] = 78746,
  [SMALL_STATE(1670)] = 78799,
  [SMALL_STATE(1671)] = 78888,
  [SMALL_STATE(1672)] = 78941,
  [SMALL_STATE(1673)] = 79000,
  [SMALL_STATE(1674)] = 79053,
  [SMALL_STATE(1675)] = 79106,
  [SMALL_STATE(1676)] = 79165,
  [SMALL_STATE(1677)] = 79218,
  [SMALL_STATE(1678)] = 79271,
  [SMALL_STATE(1679)] = 79324,
  [SMALL_STATE(1680)] = 79427,
  [SMALL_STATE(1681)] = 79480,
  [SMALL_STATE(1682)] = 79539,
  [SMALL_STATE(1683)] = 79592,
  [SMALL_STATE(1684)] = 79645,
  [SMALL_STATE(1685)] = 79698,
  [SMALL_STATE(1686)] = 79801,
  [SMALL_STATE(1687)] = 79854,
  [SMALL_STATE(1688)] = 79907,
  [SMALL_STATE(1689)] = 79966,
  [SMALL_STATE(1690)] = 80019,
  [SMALL_STATE(1691)] = 80072,
  [SMALL_STATE(1692)] = 80125,
  [SMALL_STATE(1693)] = 80178,
  [SMALL_STATE(1694)] = 80231,
  [SMALL_STATE(1695)] = 80290,
  [SMALL_STATE(1696)] = 80349,
  [SMALL_STATE(1697)] = 80408,
  [SMALL_STATE(1698)] = 80461,
  [SMALL_STATE(1699)] = 80564,
  [SMALL_STATE(1700)] = 80621,
  [SMALL_STATE(1701)] = 80674,
  [SMALL_STATE(1702)] = 80727,
  [SMALL_STATE(1703)] = 80780,
  [SMALL_STATE(1704)] = 80837,
  [SMALL_STATE(1705)] = 80896,
  [SMALL_STATE(1706)] = 80955,
  [SMALL_STATE(1707)] = 81024,
  [SMALL_STATE(1708)] = 81093,
  [SMALL_STATE(1709)] = 81152,
  [SMALL_STATE(1710)] = 81205,
  [SMALL_STATE(1711)] = 81264,
  [SMALL_STATE(1712)] = 81323,
  [SMALL_STATE(1713)] = 81376,
  [SMALL_STATE(1714)] = 81429,
  [SMALL_STATE(1715)] = 81488,
  [SMALL_STATE(1716)] = 81547,
  [SMALL_STATE(1717)] = 81600,
  [SMALL_STATE(1718)] = 81659,
  [SMALL_STATE(1719)] = 81712,
  [SMALL_STATE(1720)] = 81773,
  [SMALL_STATE(1721)] = 81830,
  [SMALL_STATE(1722)] = 81889,
  [SMALL_STATE(1723)] = 81948,
  [SMALL_STATE(1724)] = 82001,
  [SMALL_STATE(1725)] = 82054,
  [SMALL_STATE(1726)] = 82113,
  [SMALL_STATE(1727)] = 82165,
  [SMALL_STATE(1728)] = 82223,
  [SMALL_STATE(1729)] = 82275,
  [SMALL_STATE(1730)] = 82327,
  [SMALL_STATE(1731)] = 82427,
  [SMALL_STATE(1732)] = 82479,
  [SMALL_STATE(1733)] = 82531,
  [SMALL_STATE(1734)] = 82583,
  [SMALL_STATE(1735)] = 82635,
  [SMALL_STATE(1736)] = 82687,
  [SMALL_STATE(1737)] = 82743,
  [SMALL_STATE(1738)] = 82795,
  [SMALL_STATE(1739)] = 82851,
  [SMALL_STATE(1740)] = 82907,
  [SMALL_STATE(1741)] = 82963,
  [SMALL_STATE(1742)] = 83019,
  [SMALL_STATE(1743)] = 83075,
  [SMALL_STATE(1744)] = 83131,
  [SMALL_STATE(1745)] = 83187,
  [SMALL_STATE(1746)] = 83245,
  [SMALL_STATE(1747)] = 83297,
  [SMALL_STATE(1748)] = 83349,
  [SMALL_STATE(1749)] = 83401,
  [SMALL_STATE(1750)] = 83501,
  [SMALL_STATE(1751)] = 83559,
  [SMALL_STATE(1752)] = 83615,
  [SMALL_STATE(1753)] = 83673,
  [SMALL_STATE(1754)] = 83725,
  [SMALL_STATE(1755)] = 83777,
  [SMALL_STATE(1756)] = 83829,
  [SMALL_STATE(1757)] = 83881,
  [SMALL_STATE(1758)] = 83981,
  [SMALL_STATE(1759)] = 84039,
  [SMALL_STATE(1760)] = 84091,
  [SMALL_STATE(1761)] = 84143,
  [SMALL_STATE(1762)] = 84195,
  [SMALL_STATE(1763)] = 84247,
  [SMALL_STATE(1764)] = 84299,
  [SMALL_STATE(1765)] = 84351,
  [SMALL_STATE(1766)] = 84403,
  [SMALL_STATE(1767)] = 84455,
  [SMALL_STATE(1768)] = 84507,
  [SMALL_STATE(1769)] = 84559,
  [SMALL_STATE(1770)] = 84611,
  [SMALL_STATE(1771)] = 84663,
  [SMALL_STATE(1772)] = 84715,
  [SMALL_STATE(1773)] = 84767,
  [SMALL_STATE(1774)] = 84819,
  [SMALL_STATE(1775)] = 84871,
  [SMALL_STATE(1776)] = 84923,
  [SMALL_STATE(1777)] = 84975,
  [SMALL_STATE(1778)] = 85027,
  [SMALL_STATE(1779)] = 85127,
  [SMALL_STATE(1780)] = 85185,
  [SMALL_STATE(1781)] = 85243,
  [SMALL_STATE(1782)] = 85295,
  [SMALL_STATE(1783)] = 85347,
  [SMALL_STATE(1784)] = 85399,
  [SMALL_STATE(1785)] = 85457,
  [SMALL_STATE(1786)] = 85509,
  [SMALL_STATE(1787)] = 85567,
  [SMALL_STATE(1788)] = 85625,
  [SMALL_STATE(1789)] = 85677,
  [SMALL_STATE(1790)] = 85733,
  [SMALL_STATE(1791)] = 85785,
  [SMALL_STATE(1792)] = 85885,
  [SMALL_STATE(1793)] = 85937,
  [SMALL_STATE(1794)] = 85989,
  [SMALL_STATE(1795)] = 86047,
  [SMALL_STATE(1796)] = 86105,
  [SMALL_STATE(1797)] = 86157,
  [SMALL_STATE(1798)] = 86213,
  [SMALL_STATE(1799)] = 86265,
  [SMALL_STATE(1800)] = 86321,
  [SMALL_STATE(1801)] = 86379,
  [SMALL_STATE(1802)] = 86435,
  [SMALL_STATE(1803)] = 86487,
  [SMALL_STATE(1804)] = 86543,
  [SMALL_STATE(1805)] = 86595,
  [SMALL_STATE(1806)] = 86647,
  [SMALL_STATE(1807)] = 86747,
  [SMALL_STATE(1808)] = 86799,
  [SMALL_STATE(1809)] = 86857,
  [SMALL_STATE(1810)] = 86909,
  [SMALL_STATE(1811)] = 86961,
  [SMALL_STATE(1812)] = 87031,
  [SMALL_STATE(1813)] = 87083,
  [SMALL_STATE(1814)] = 87135,
  [SMALL_STATE(1815)] = 87191,
  [SMALL_STATE(1816)] = 87243,
  [SMALL_STATE(1817)] = 87299,
  [SMALL_STATE(1818)] = 87351,
  [SMALL_STATE(1819)] = 87403,
  [SMALL_STATE(1820)] = 87455,
  [SMALL_STATE(1821)] = 87507,
  [SMALL_STATE(1822)] = 87565,
  [SMALL_STATE(1823)] = 87623,
  [SMALL_STATE(1824)] = 87681,
  [SMALL_STATE(1825)] = 87781,
  [SMALL_STATE(1826)] = 87833,
  [SMALL_STATE(1827)] = 87889,
  [SMALL_STATE(1828)] = 87941,
  [SMALL_STATE(1829)] = 87997,
  [SMALL_STATE(1830)] = 88053,
  [SMALL_STATE(1831)] = 88153,
  [SMALL_STATE(1832)] = 88211,
  [SMALL_STATE(1833)] = 88269,
  [SMALL_STATE(1834)] = 88321,
  [SMALL_STATE(1835)] = 88373,
  [SMALL_STATE(1836)] = 88425,
  [SMALL_STATE(1837)] = 88477,
  [SMALL_STATE(1838)] = 88529,
  [SMALL_STATE(1839)] = 88581,
  [SMALL_STATE(1840)] = 88639,
  [SMALL_STATE(1841)] = 88691,
  [SMALL_STATE(1842)] = 88743,
  [SMALL_STATE(1843)] = 88843,
  [SMALL_STATE(1844)] = 88901,
  [SMALL_STATE(1845)] = 88953,
  [SMALL_STATE(1846)] = 89011,
  [SMALL_STATE(1847)] = 89069,
  [SMALL_STATE(1848)] = 89121,
  [SMALL_STATE(1849)] = 89173,
  [SMALL_STATE(1850)] = 89225,
  [SMALL_STATE(1851)] = 89325,
  [SMALL_STATE(1852)] = 89377,
  [SMALL_STATE(1853)] = 89429,
  [SMALL_STATE(1854)] = 89481,
  [SMALL_STATE(1855)] = 89533,
  [SMALL_STATE(1856)] = 89585,
  [SMALL_STATE(1857)] = 89637,
  [SMALL_STATE(1858)] = 89689,
  [SMALL_STATE(1859)] = 89741,
  [SMALL_STATE(1860)] = 89793,
  [SMALL_STATE(1861)] = 89849,
  [SMALL_STATE(1862)] = 89901,
  [SMALL_STATE(1863)] = 89953,
  [SMALL_STATE(1864)] = 90005,
  [SMALL_STATE(1865)] = 90057,
  [SMALL_STATE(1866)] = 90109,
  [SMALL_STATE(1867)] = 90161,
  [SMALL_STATE(1868)] = 90261,
  [SMALL_STATE(1869)] = 90351,
  [SMALL_STATE(1870)] = 90451,
  [SMALL_STATE(1871)] = 90503,
  [SMALL_STATE(1872)] = 90603,
  [SMALL_STATE(1873)] = 90655,
  [SMALL_STATE(1874)] = 90711,
  [SMALL_STATE(1875)] = 90763,
  [SMALL_STATE(1876)] = 90815,
  [SMALL_STATE(1877)] = 90867,
  [SMALL_STATE(1878)] = 90919,
  [SMALL_STATE(1879)] = 90971,
  [SMALL_STATE(1880)] = 91029,
  [SMALL_STATE(1881)] = 91081,
  [SMALL_STATE(1882)] = 91133,
  [SMALL_STATE(1883)] = 91201,
  [SMALL_STATE(1884)] = 91253,
  [SMALL_STATE(1885)] = 91353,
  [SMALL_STATE(1886)] = 91405,
  [SMALL_STATE(1887)] = 91461,
  [SMALL_STATE(1888)] = 91517,
  [SMALL_STATE(1889)] = 91569,
  [SMALL_STATE(1890)] = 91627,
  [SMALL_STATE(1891)] = 91679,
  [SMALL_STATE(1892)] = 91731,
  [SMALL_STATE(1893)] = 91791,
  [SMALL_STATE(1894)] = 91851,
  [SMALL_STATE(1895)] = 91909,
  [SMALL_STATE(1896)] = 91965,
  [SMALL_STATE(1897)] = 92017,
  [SMALL_STATE(1898)] = 92075,
  [SMALL_STATE(1899)] = 92131,
  [SMALL_STATE(1900)] = 92187,
  [SMALL_STATE(1901)] = 92243,
  [SMALL_STATE(1902)] = 92301,
  [SMALL_STATE(1903)] = 92359,
  [SMALL_STATE(1904)] = 92417,
  [SMALL_STATE(1905)] = 92473,
  [SMALL_STATE(1906)] = 92573,
  [SMALL_STATE(1907)] = 92673,
  [SMALL_STATE(1908)] = 92737,
  [SMALL_STATE(1909)] = 92795,
  [SMALL_STATE(1910)] = 92851,
  [SMALL_STATE(1911)] = 92909,
  [SMALL_STATE(1912)] = 92967,
  [SMALL_STATE(1913)] = 93023,
  [SMALL_STATE(1914)] = 93123,
  [SMALL_STATE(1915)] = 93181,
  [SMALL_STATE(1916)] = 93239,
  [SMALL_STATE(1917)] = 93297,
  [SMALL_STATE(1918)] = 93349,
  [SMALL_STATE(1919)] = 93407,
  [SMALL_STATE(1920)] = 93459,
  [SMALL_STATE(1921)] = 93511,
  [SMALL_STATE(1922)] = 93563,
  [SMALL_STATE(1923)] = 93621,
  [SMALL_STATE(1924)] = 93673,
  [SMALL_STATE(1925)] = 93725,
  [SMALL_STATE(1926)] = 93825,
  [SMALL_STATE(1927)] = 93925,
  [SMALL_STATE(1928)] = 94025,
  [SMALL_STATE(1929)] = 94077,
  [SMALL_STATE(1930)] = 94133,
  [SMALL_STATE(1931)] = 94194,
  [SMALL_STATE(1932)] = 94255,
  [SMALL_STATE(1933)] = 94352,
  [SMALL_STATE(1934)] = 94449,
  [SMALL_STATE(1935)] = 94546,
  [SMALL_STATE(1936)] = 94643,
  [SMALL_STATE(1937)] = 94740,
  [SMALL_STATE(1938)] = 94837,
  [SMALL_STATE(1939)] = 94934,
  [SMALL_STATE(1940)] = 94995,
  [SMALL_STATE(1941)] = 95092,
  [SMALL_STATE(1942)] = 95189,
  [SMALL_STATE(1943)] = 95286,
  [SMALL_STATE(1944)] = 95383,
  [SMALL_STATE(1945)] = 95480,
  [SMALL_STATE(1946)] = 95531,
  [SMALL_STATE(1947)] = 95582,
  [SMALL_STATE(1948)] = 95633,
  [SMALL_STATE(1949)] = 95730,
  [SMALL_STATE(1950)] = 95827,
  [SMALL_STATE(1951)] = 95924,
  [SMALL_STATE(1952)] = 95975,
  [SMALL_STATE(1953)] = 96026,
  [SMALL_STATE(1954)] = 96077,
  [SMALL_STATE(1955)] = 96128,
  [SMALL_STATE(1956)] = 96179,
  [SMALL_STATE(1957)] = 96230,
  [SMALL_STATE(1958)] = 96281,
  [SMALL_STATE(1959)] = 96332,
  [SMALL_STATE(1960)] = 96383,
  [SMALL_STATE(1961)] = 96434,
  [SMALL_STATE(1962)] = 96485,
  [SMALL_STATE(1963)] = 96536,
  [SMALL_STATE(1964)] = 96633,
  [SMALL_STATE(1965)] = 96688,
  [SMALL_STATE(1966)] = 96739,
  [SMALL_STATE(1967)] = 96794,
  [SMALL_STATE(1968)] = 96891,
  [SMALL_STATE(1969)] = 96948,
  [SMALL_STATE(1970)] = 97045,
  [SMALL_STATE(1971)] = 97142,
  [SMALL_STATE(1972)] = 97193,
  [SMALL_STATE(1973)] = 97250,
  [SMALL_STATE(1974)] = 97301,
  [SMALL_STATE(1975)] = 97356,
  [SMALL_STATE(1976)] = 97453,
  [SMALL_STATE(1977)] = 97550,
  [SMALL_STATE(1978)] = 97601,
  [SMALL_STATE(1979)] = 97698,
  [SMALL_STATE(1980)] = 97749,
  [SMALL_STATE(1981)] = 97846,
  [SMALL_STATE(1982)] = 97943,
  [SMALL_STATE(1983)] = 98040,
  [SMALL_STATE(1984)] = 98137,
  [SMALL_STATE(1985)] = 98234,
  [SMALL_STATE(1986)] = 98331,
  [SMALL_STATE(1987)] = 98382,
  [SMALL_STATE(1988)] = 98479,
  [SMALL_STATE(1989)] = 98576,
  [SMALL_STATE(1990)] = 98673,
  [SMALL_STATE(1991)] = 98770,
  [SMALL_STATE(1992)] = 98867,
  [SMALL_STATE(1993)] = 98918,
  [SMALL_STATE(1994)] = 98969,
  [SMALL_STATE(1995)] = 99020,
  [SMALL_STATE(1996)] = 99071,
  [SMALL_STATE(1997)] = 99122,
  [SMALL_STATE(1998)] = 99173,
  [SMALL_STATE(1999)] = 99270,
  [SMALL_STATE(2000)] = 99367,
  [SMALL_STATE(2001)] = 99464,
  [SMALL_STATE(2002)] = 99561,
  [SMALL_STATE(2003)] = 99612,
  [SMALL_STATE(2004)] = 99709,
  [SMALL_STATE(2005)] = 99760,
  [SMALL_STATE(2006)] = 99811,
  [SMALL_STATE(2007)] = 99862,
  [SMALL_STATE(2008)] = 99959,
  [SMALL_STATE(2009)] = 100056,
  [SMALL_STATE(2010)] = 100153,
  [SMALL_STATE(2011)] = 100250,
  [SMALL_STATE(2012)] = 100347,
  [SMALL_STATE(2013)] = 100398,
  [SMALL_STATE(2014)] = 100495,
  [SMALL_STATE(2015)] = 100592,
  [SMALL_STATE(2016)] = 100643,
  [SMALL_STATE(2017)] = 100740,
  [SMALL_STATE(2018)] = 100799,
  [SMALL_STATE(2019)] = 100896,
  [SMALL_STATE(2020)] = 100955,
  [SMALL_STATE(2021)] = 101006,
  [SMALL_STATE(2022)] = 101103,
  [SMALL_STATE(2023)] = 101200,
  [SMALL_STATE(2024)] = 101297,
  [SMALL_STATE(2025)] = 101352,
  [SMALL_STATE(2026)] = 101403,
  [SMALL_STATE(2027)] = 101458,
  [SMALL_STATE(2028)] = 101555,
  [SMALL_STATE(2029)] = 101652,
  [SMALL_STATE(2030)] = 101749,
  [SMALL_STATE(2031)] = 101846,
  [SMALL_STATE(2032)] = 101943,
  [SMALL_STATE(2033)] = 101994,
  [SMALL_STATE(2034)] = 102049,
  [SMALL_STATE(2035)] = 102146,
  [SMALL_STATE(2036)] = 102243,
  [SMALL_STATE(2037)] = 102340,
  [SMALL_STATE(2038)] = 102391,
  [SMALL_STATE(2039)] = 102488,
  [SMALL_STATE(2040)] = 102585,
  [SMALL_STATE(2041)] = 102682,
  [SMALL_STATE(2042)] = 102779,
  [SMALL_STATE(2043)] = 102876,
  [SMALL_STATE(2044)] = 102973,
  [SMALL_STATE(2045)] = 103070,
  [SMALL_STATE(2046)] = 103167,
  [SMALL_STATE(2047)] = 103218,
  [SMALL_STATE(2048)] = 103315,
  [SMALL_STATE(2049)] = 103412,
  [SMALL_STATE(2050)] = 103509,
  [SMALL_STATE(2051)] = 103606,
  [SMALL_STATE(2052)] = 103703,
  [SMALL_STATE(2053)] = 103800,
  [SMALL_STATE(2054)] = 103897,
  [SMALL_STATE(2055)] = 103994,
  [SMALL_STATE(2056)] = 104053,
  [SMALL_STATE(2057)] = 104150,
  [SMALL_STATE(2058)] = 104201,
  [SMALL_STATE(2059)] = 104252,
  [SMALL_STATE(2060)] = 104349,
  [SMALL_STATE(2061)] = 104410,
  [SMALL_STATE(2062)] = 104509,
  [SMALL_STATE(2063)] = 104606,
  [SMALL_STATE(2064)] = 104703,
  [SMALL_STATE(2065)] = 104800,
  [SMALL_STATE(2066)] = 104851,
  [SMALL_STATE(2067)] = 104948,
  [SMALL_STATE(2068)] = 105005,
  [SMALL_STATE(2069)] = 105062,
  [SMALL_STATE(2070)] = 105159,
  [SMALL_STATE(2071)] = 105220,
  [SMALL_STATE(2072)] = 105271,
  [SMALL_STATE(2073)] = 105368,
  [SMALL_STATE(2074)] = 105419,
  [SMALL_STATE(2075)] = 105476,
  [SMALL_STATE(2076)] = 105527,
  [SMALL_STATE(2077)] = 105588,
  [SMALL_STATE(2078)] = 105639,
  [SMALL_STATE(2079)] = 105700,
  [SMALL_STATE(2080)] = 105761,
  [SMALL_STATE(2081)] = 105812,
  [SMALL_STATE(2082)] = 105863,
  [SMALL_STATE(2083)] = 105924,
  [SMALL_STATE(2084)] = 105985,
  [SMALL_STATE(2085)] = 106046,
  [SMALL_STATE(2086)] = 106107,
  [SMALL_STATE(2087)] = 106158,
  [SMALL_STATE(2088)] = 106255,
  [SMALL_STATE(2089)] = 106316,
  [SMALL_STATE(2090)] = 106413,
  [SMALL_STATE(2091)] = 106512,
  [SMALL_STATE(2092)] = 106563,
  [SMALL_STATE(2093)] = 106660,
  [SMALL_STATE(2094)] = 106711,
  [SMALL_STATE(2095)] = 106762,
  [SMALL_STATE(2096)] = 106813,
  [SMALL_STATE(2097)] = 106874,
  [SMALL_STATE(2098)] = 106935,
  [SMALL_STATE(2099)] = 106996,
  [SMALL_STATE(2100)] = 107057,
  [SMALL_STATE(2101)] = 107108,
  [SMALL_STATE(2102)] = 107169,
  [SMALL_STATE(2103)] = 107230,
  [SMALL_STATE(2104)] = 107281,
  [SMALL_STATE(2105)] = 107342,
  [SMALL_STATE(2106)] = 107393,
  [SMALL_STATE(2107)] = 107444,
  [SMALL_STATE(2108)] = 107495,
  [SMALL_STATE(2109)] = 107546,
  [SMALL_STATE(2110)] = 107597,
  [SMALL_STATE(2111)] = 107658,
  [SMALL_STATE(2112)] = 107709,
  [SMALL_STATE(2113)] = 107770,
  [SMALL_STATE(2114)] = 107831,
  [SMALL_STATE(2115)] = 107882,
  [SMALL_STATE(2116)] = 107943,
  [SMALL_STATE(2117)] = 107994,
  [SMALL_STATE(2118)] = 108045,
  [SMALL_STATE(2119)] = 108100,
  [SMALL_STATE(2120)] = 108197,
  [SMALL_STATE(2121)] = 108258,
  [SMALL_STATE(2122)] = 108309,
  [SMALL_STATE(2123)] = 108364,
  [SMALL_STATE(2124)] = 108421,
  [SMALL_STATE(2125)] = 108476,
  [SMALL_STATE(2126)] = 108531,
  [SMALL_STATE(2127)] = 108586,
  [SMALL_STATE(2128)] = 108641,
  [SMALL_STATE(2129)] = 108702,
  [SMALL_STATE(2130)] = 108799,
  [SMALL_STATE(2131)] = 108896,
  [SMALL_STATE(2132)] = 108993,
  [SMALL_STATE(2133)] = 109090,
  [SMALL_STATE(2134)] = 109151,
  [SMALL_STATE(2135)] = 109248,
  [SMALL_STATE(2136)] = 109305,
  [SMALL_STATE(2137)] = 109366,
  [SMALL_STATE(2138)] = 109463,
  [SMALL_STATE(2139)] = 109520,
  [SMALL_STATE(2140)] = 109617,
  [SMALL_STATE(2141)] = 109678,
  [SMALL_STATE(2142)] = 109739,
  [SMALL_STATE(2143)] = 109800,
  [SMALL_STATE(2144)] = 109861,
  [SMALL_STATE(2145)] = 109922,
  [SMALL_STATE(2146)] = 110019,
  [SMALL_STATE(2147)] = 110080,
  [SMALL_STATE(2148)] = 110135,
  [SMALL_STATE(2149)] = 110196,
  [SMALL_STATE(2150)] = 110257,
  [SMALL_STATE(2151)] = 110318,
  [SMALL_STATE(2152)] = 110379,
  [SMALL_STATE(2153)] = 110436,
  [SMALL_STATE(2154)] = 110497,
  [SMALL_STATE(2155)] = 110552,
  [SMALL_STATE(2156)] = 110607,
  [SMALL_STATE(2157)] = 110666,
  [SMALL_STATE(2158)] = 110727,
  [SMALL_STATE(2159)] = 110778,
  [SMALL_STATE(2160)] = 110839,
  [SMALL_STATE(2161)] = 110890,
  [SMALL_STATE(2162)] = 110945,
  [SMALL_STATE(2163)] = 110996,
  [SMALL_STATE(2164)] = 111057,
  [SMALL_STATE(2165)] = 111118,
  [SMALL_STATE(2166)] = 111179,
  [SMALL_STATE(2167)] = 111278,
  [SMALL_STATE(2168)] = 111339,
  [SMALL_STATE(2169)] = 111400,
  [SMALL_STATE(2170)] = 111461,
  [SMALL_STATE(2171)] = 111522,
  [SMALL_STATE(2172)] = 111583,
  [SMALL_STATE(2173)] = 111644,
  [SMALL_STATE(2174)] = 111695,
  [SMALL_STATE(2175)] = 111792,
  [SMALL_STATE(2176)] = 111853,
  [SMALL_STATE(2177)] = 111914,
  [SMALL_STATE(2178)] = 111975,
  [SMALL_STATE(2179)] = 112036,
  [SMALL_STATE(2180)] = 112097,
  [SMALL_STATE(2181)] = 112158,
  [SMALL_STATE(2182)] = 112219,
  [SMALL_STATE(2183)] = 112270,
  [SMALL_STATE(2184)] = 112331,
  [SMALL_STATE(2185)] = 112428,
  [SMALL_STATE(2186)] = 112489,
  [SMALL_STATE(2187)] = 112550,
  [SMALL_STATE(2188)] = 112611,
  [SMALL_STATE(2189)] = 112708,
  [SMALL_STATE(2190)] = 112769,
  [SMALL_STATE(2191)] = 112824,
  [SMALL_STATE(2192)] = 112885,
  [SMALL_STATE(2193)] = 112982,
  [SMALL_STATE(2194)] = 113043,
  [SMALL_STATE(2195)] = 113104,
  [SMALL_STATE(2196)] = 113165,
  [SMALL_STATE(2197)] = 113226,
  [SMALL_STATE(2198)] = 113323,
  [SMALL_STATE(2199)] = 113420,
  [SMALL_STATE(2200)] = 113475,
  [SMALL_STATE(2201)] = 113536,
  [SMALL_STATE(2202)] = 113587,
  [SMALL_STATE(2203)] = 113684,
  [SMALL_STATE(2204)] = 113739,
  [SMALL_STATE(2205)] = 113794,
  [SMALL_STATE(2206)] = 113891,
  [SMALL_STATE(2207)] = 113952,
  [SMALL_STATE(2208)] = 114013,
  [SMALL_STATE(2209)] = 114110,
  [SMALL_STATE(2210)] = 114207,
  [SMALL_STATE(2211)] = 114258,
  [SMALL_STATE(2212)] = 114313,
  [SMALL_STATE(2213)] = 114374,
  [SMALL_STATE(2214)] = 114471,
  [SMALL_STATE(2215)] = 114522,
  [SMALL_STATE(2216)] = 114573,
  [SMALL_STATE(2217)] = 114624,
  [SMALL_STATE(2218)] = 114685,
  [SMALL_STATE(2219)] = 114736,
  [SMALL_STATE(2220)] = 114797,
  [SMALL_STATE(2221)] = 114858,
  [SMALL_STATE(2222)] = 114919,
  [SMALL_STATE(2223)] = 114970,
  [SMALL_STATE(2224)] = 115031,
  [SMALL_STATE(2225)] = 115082,
  [SMALL_STATE(2226)] = 115179,
  [SMALL_STATE(2227)] = 115276,
  [SMALL_STATE(2228)] = 115327,
  [SMALL_STATE(2229)] = 115388,
  [SMALL_STATE(2230)] = 115485,
  [SMALL_STATE(2231)] = 115582,
  [SMALL_STATE(2232)] = 115679,
  [SMALL_STATE(2233)] = 115776,
  [SMALL_STATE(2234)] = 115873,
  [SMALL_STATE(2235)] = 115970,
  [SMALL_STATE(2236)] = 116067,
  [SMALL_STATE(2237)] = 116128,
  [SMALL_STATE(2238)] = 116179,
  [SMALL_STATE(2239)] = 116276,
  [SMALL_STATE(2240)] = 116373,
  [SMALL_STATE(2241)] = 116470,
  [SMALL_STATE(2242)] = 116569,
  [SMALL_STATE(2243)] = 116666,
  [SMALL_STATE(2244)] = 116763,
  [SMALL_STATE(2245)] = 116862,
  [SMALL_STATE(2246)] = 116959,
  [SMALL_STATE(2247)] = 117056,
  [SMALL_STATE(2248)] = 117107,
  [SMALL_STATE(2249)] = 117204,
  [SMALL_STATE(2250)] = 117301,
  [SMALL_STATE(2251)] = 117398,
  [SMALL_STATE(2252)] = 117497,
  [SMALL_STATE(2253)] = 117594,
  [SMALL_STATE(2254)] = 117645,
  [SMALL_STATE(2255)] = 117700,
  [SMALL_STATE(2256)] = 117761,
  [SMALL_STATE(2257)] = 117858,
  [SMALL_STATE(2258)] = 117909,
  [SMALL_STATE(2259)] = 118006,
  [SMALL_STATE(2260)] = 118067,
  [SMALL_STATE(2261)] = 118118,
  [SMALL_STATE(2262)] = 118215,
  [SMALL_STATE(2263)] = 118276,
  [SMALL_STATE(2264)] = 118327,
  [SMALL_STATE(2265)] = 118388,
  [SMALL_STATE(2266)] = 118449,
  [SMALL_STATE(2267)] = 118546,
  [SMALL_STATE(2268)] = 118643,
  [SMALL_STATE(2269)] = 118693,
  [SMALL_STATE(2270)] = 118743,
  [SMALL_STATE(2271)] = 118797,
  [SMALL_STATE(2272)] = 118851,
  [SMALL_STATE(2273)] = 118909,
  [SMALL_STATE(2274)] = 118967,
  [SMALL_STATE(2275)] = 119017,
  [SMALL_STATE(2276)] = 119067,
  [SMALL_STATE(2277)] = 119117,
  [SMALL_STATE(2278)] = 119167,
  [SMALL_STATE(2279)] = 119221,
  [SMALL_STATE(2280)] = 119277,
  [SMALL_STATE(2281)] = 119327,
  [SMALL_STATE(2282)] = 119381,
  [SMALL_STATE(2283)] = 119437,
  [SMALL_STATE(2284)] = 119493,
  [SMALL_STATE(2285)] = 119547,
  [SMALL_STATE(2286)] = 119603,
  [SMALL_STATE(2287)] = 119657,
  [SMALL_STATE(2288)] = 119707,
  [SMALL_STATE(2289)] = 119763,
  [SMALL_STATE(2290)] = 119813,
  [SMALL_STATE(2291)] = 119869,
  [SMALL_STATE(2292)] = 119925,
  [SMALL_STATE(2293)] = 119975,
  [SMALL_STATE(2294)] = 120031,
  [SMALL_STATE(2295)] = 120081,
  [SMALL_STATE(2296)] = 120131,
  [SMALL_STATE(2297)] = 120185,
  [SMALL_STATE(2298)] = 120241,
  [SMALL_STATE(2299)] = 120295,
  [SMALL_STATE(2300)] = 120345,
  [SMALL_STATE(2301)] = 120399,
  [SMALL_STATE(2302)] = 120449,
  [SMALL_STATE(2303)] = 120505,
  [SMALL_STATE(2304)] = 120559,
  [SMALL_STATE(2305)] = 120609,
  [SMALL_STATE(2306)] = 120659,
  [SMALL_STATE(2307)] = 120709,
  [SMALL_STATE(2308)] = 120759,
  [SMALL_STATE(2309)] = 120813,
  [SMALL_STATE(2310)] = 120867,
  [SMALL_STATE(2311)] = 120917,
  [SMALL_STATE(2312)] = 120971,
  [SMALL_STATE(2313)] = 121025,
  [SMALL_STATE(2314)] = 121075,
  [SMALL_STATE(2315)] = 121129,
  [SMALL_STATE(2316)] = 121179,
  [SMALL_STATE(2317)] = 121233,
  [SMALL_STATE(2318)] = 121283,
  [SMALL_STATE(2319)] = 121333,
  [SMALL_STATE(2320)] = 121383,
  [SMALL_STATE(2321)] = 121433,
  [SMALL_STATE(2322)] = 121483,
  [SMALL_STATE(2323)] = 121533,
  [SMALL_STATE(2324)] = 121589,
  [SMALL_STATE(2325)] = 121645,
  [SMALL_STATE(2326)] = 121701,
  [SMALL_STATE(2327)] = 121751,
  [SMALL_STATE(2328)] = 121807,
  [SMALL_STATE(2329)] = 121863,
  [SMALL_STATE(2330)] = 121919,
  [SMALL_STATE(2331)] = 121975,
  [SMALL_STATE(2332)] = 122031,
  [SMALL_STATE(2333)] = 122085,
  [SMALL_STATE(2334)] = 122135,
  [SMALL_STATE(2335)] = 122185,
  [SMALL_STATE(2336)] = 122235,
  [SMALL_STATE(2337)] = 122291,
  [SMALL_STATE(2338)] = 122347,
  [SMALL_STATE(2339)] = 122397,
  [SMALL_STATE(2340)] = 122447,
  [SMALL_STATE(2341)] = 122503,
  [SMALL_STATE(2342)] = 122553,
  [SMALL_STATE(2343)] = 122609,
  [SMALL_STATE(2344)] = 122659,
  [SMALL_STATE(2345)] = 122709,
  [SMALL_STATE(2346)] = 122763,
  [SMALL_STATE(2347)] = 122812,
  [SMALL_STATE(2348)] = 122865,
  [SMALL_STATE(2349)] = 122914,
  [SMALL_STATE(2350)] = 123009,
  [SMALL_STATE(2351)] = 123064,
  [SMALL_STATE(2352)] = 123113,
  [SMALL_STATE(2353)] = 123166,
  [SMALL_STATE(2354)] = 123261,
  [SMALL_STATE(2355)] = 123310,
  [SMALL_STATE(2356)] = 123359,
  [SMALL_STATE(2357)] = 123412,
  [SMALL_STATE(2358)] = 123465,
  [SMALL_STATE(2359)] = 123560,
  [SMALL_STATE(2360)] = 123613,
  [SMALL_STATE(2361)] = 123708,
  [SMALL_STATE(2362)] = 123803,
  [SMALL_STATE(2363)] = 123852,
  [SMALL_STATE(2364)] = 123901,
  [SMALL_STATE(2365)] = 123996,
  [SMALL_STATE(2366)] = 124051,
  [SMALL_STATE(2367)] = 124104,
  [SMALL_STATE(2368)] = 124159,
  [SMALL_STATE(2369)] = 124208,
  [SMALL_STATE(2370)] = 124261,
  [SMALL_STATE(2371)] = 124314,
  [SMALL_STATE(2372)] = 124367,
  [SMALL_STATE(2373)] = 124462,
  [SMALL_STATE(2374)] = 124515,
  [SMALL_STATE(2375)] = 124568,
  [SMALL_STATE(2376)] = 124617,
  [SMALL_STATE(2377)] = 124712,
  [SMALL_STATE(2378)] = 124807,
  [SMALL_STATE(2379)] = 124902,
  [SMALL_STATE(2380)] = 124997,
  [SMALL_STATE(2381)] = 125046,
  [SMALL_STATE(2382)] = 125095,
  [SMALL_STATE(2383)] = 125148,
  [SMALL_STATE(2384)] = 125197,
  [SMALL_STATE(2385)] = 125246,
  [SMALL_STATE(2386)] = 125301,
  [SMALL_STATE(2387)] = 125356,
  [SMALL_STATE(2388)] = 125411,
  [SMALL_STATE(2389)] = 125460,
  [SMALL_STATE(2390)] = 125515,
  [SMALL_STATE(2391)] = 125564,
  [SMALL_STATE(2392)] = 125619,
  [SMALL_STATE(2393)] = 125668,
  [SMALL_STATE(2394)] = 125717,
  [SMALL_STATE(2395)] = 125770,
  [SMALL_STATE(2396)] = 125823,
  [SMALL_STATE(2397)] = 125876,
  [SMALL_STATE(2398)] = 125931,
  [SMALL_STATE(2399)] = 125980,
  [SMALL_STATE(2400)] = 126029,
  [SMALL_STATE(2401)] = 126078,
  [SMALL_STATE(2402)] = 126127,
  [SMALL_STATE(2403)] = 126176,
  [SMALL_STATE(2404)] = 126231,
  [SMALL_STATE(2405)] = 126286,
  [SMALL_STATE(2406)] = 126339,
  [SMALL_STATE(2407)] = 126392,
  [SMALL_STATE(2408)] = 126447,
  [SMALL_STATE(2409)] = 126500,
  [SMALL_STATE(2410)] = 126553,
  [SMALL_STATE(2411)] = 126648,
  [SMALL_STATE(2412)] = 126743,
  [SMALL_STATE(2413)] = 126792,
  [SMALL_STATE(2414)] = 126845,
  [SMALL_STATE(2415)] = 126940,
  [SMALL_STATE(2416)] = 126993,
  [SMALL_STATE(2417)] = 127088,
  [SMALL_STATE(2418)] = 127141,
  [SMALL_STATE(2419)] = 127190,
  [SMALL_STATE(2420)] = 127243,
  [SMALL_STATE(2421)] = 127292,
  [SMALL_STATE(2422)] = 127341,
  [SMALL_STATE(2423)] = 127396,
  [SMALL_STATE(2424)] = 127445,
  [SMALL_STATE(2425)] = 127500,
  [SMALL_STATE(2426)] = 127553,
  [SMALL_STATE(2427)] = 127608,
  [SMALL_STATE(2428)] = 127661,
  [SMALL_STATE(2429)] = 127756,
  [SMALL_STATE(2430)] = 127811,
  [SMALL_STATE(2431)] = 127860,
  [SMALL_STATE(2432)] = 127915,
  [SMALL_STATE(2433)] = 127968,
  [SMALL_STATE(2434)] = 128021,
  [SMALL_STATE(2435)] = 128074,
  [SMALL_STATE(2436)] = 128127,
  [SMALL_STATE(2437)] = 128180,
  [SMALL_STATE(2438)] = 128233,
  [SMALL_STATE(2439)] = 128286,
  [SMALL_STATE(2440)] = 128335,
  [SMALL_STATE(2441)] = 128388,
  [SMALL_STATE(2442)] = 128443,
  [SMALL_STATE(2443)] = 128492,
  [SMALL_STATE(2444)] = 128540,
  [SMALL_STATE(2445)] = 128588,
  [SMALL_STATE(2446)] = 128644,
  [SMALL_STATE(2447)] = 128692,
  [SMALL_STATE(2448)] = 128740,
  [SMALL_STATE(2449)] = 128788,
  [SMALL_STATE(2450)] = 128836,
  [SMALL_STATE(2451)] = 128884,
  [SMALL_STATE(2452)] = 128936,
  [SMALL_STATE(2453)] = 128984,
  [SMALL_STATE(2454)] = 129036,
  [SMALL_STATE(2455)] = 129084,
  [SMALL_STATE(2456)] = 129140,
  [SMALL_STATE(2457)] = 129192,
  [SMALL_STATE(2458)] = 129244,
  [SMALL_STATE(2459)] = 129296,
  [SMALL_STATE(2460)] = 129350,
  [SMALL_STATE(2461)] = 129402,
  [SMALL_STATE(2462)] = 129458,
  [SMALL_STATE(2463)] = 129506,
  [SMALL_STATE(2464)] = 129562,
  [SMALL_STATE(2465)] = 129614,
  [SMALL_STATE(2466)] = 129662,
  [SMALL_STATE(2467)] = 129716,
  [SMALL_STATE(2468)] = 129764,
  [SMALL_STATE(2469)] = 129812,
  [SMALL_STATE(2470)] = 129860,
  [SMALL_STATE(2471)] = 129914,
  [SMALL_STATE(2472)] = 129968,
  [SMALL_STATE(2473)] = 130020,
  [SMALL_STATE(2474)] = 130074,
  [SMALL_STATE(2475)] = 130126,
  [SMALL_STATE(2476)] = 130180,
  [SMALL_STATE(2477)] = 130232,
  [SMALL_STATE(2478)] = 130280,
  [SMALL_STATE(2479)] = 130328,
  [SMALL_STATE(2480)] = 130376,
  [SMALL_STATE(2481)] = 130430,
  [SMALL_STATE(2482)] = 130478,
  [SMALL_STATE(2483)] = 130526,
  [SMALL_STATE(2484)] = 130578,
  [SMALL_STATE(2485)] = 130630,
  [SMALL_STATE(2486)] = 130678,
  [SMALL_STATE(2487)] = 130730,
  [SMALL_STATE(2488)] = 130782,
  [SMALL_STATE(2489)] = 130836,
  [SMALL_STATE(2490)] = 130888,
  [SMALL_STATE(2491)] = 130940,
  [SMALL_STATE(2492)] = 130994,
  [SMALL_STATE(2493)] = 131042,
  [SMALL_STATE(2494)] = 131094,
  [SMALL_STATE(2495)] = 131148,
  [SMALL_STATE(2496)] = 131202,
  [SMALL_STATE(2497)] = 131250,
  [SMALL_STATE(2498)] = 131306,
  [SMALL_STATE(2499)] = 131354,
  [SMALL_STATE(2500)] = 131402,
  [SMALL_STATE(2501)] = 131454,
  [SMALL_STATE(2502)] = 131510,
  [SMALL_STATE(2503)] = 131558,
  [SMALL_STATE(2504)] = 131610,
  [SMALL_STATE(2505)] = 131658,
  [SMALL_STATE(2506)] = 131706,
  [SMALL_STATE(2507)] = 131760,
  [SMALL_STATE(2508)] = 131808,
  [SMALL_STATE(2509)] = 131856,
  [SMALL_STATE(2510)] = 131910,
  [SMALL_STATE(2511)] = 131958,
  [SMALL_STATE(2512)] = 132010,
  [SMALL_STATE(2513)] = 132058,
  [SMALL_STATE(2514)] = 132106,
  [SMALL_STATE(2515)] = 132154,
  [SMALL_STATE(2516)] = 132202,
  [SMALL_STATE(2517)] = 132250,
  [SMALL_STATE(2518)] = 132298,
  [SMALL_STATE(2519)] = 132346,
  [SMALL_STATE(2520)] = 132394,
  [SMALL_STATE(2521)] = 132442,
  [SMALL_STATE(2522)] = 132494,
  [SMALL_STATE(2523)] = 132542,
  [SMALL_STATE(2524)] = 132590,
  [SMALL_STATE(2525)] = 132638,
  [SMALL_STATE(2526)] = 132686,
  [SMALL_STATE(2527)] = 132734,
  [SMALL_STATE(2528)] = 132782,
  [SMALL_STATE(2529)] = 132830,
  [SMALL_STATE(2530)] = 132878,
  [SMALL_STATE(2531)] = 132926,
  [SMALL_STATE(2532)] = 132974,
  [SMALL_STATE(2533)] = 133026,
  [SMALL_STATE(2534)] = 133080,
  [SMALL_STATE(2535)] = 133128,
  [SMALL_STATE(2536)] = 133176,
  [SMALL_STATE(2537)] = 133224,
  [SMALL_STATE(2538)] = 133276,
  [SMALL_STATE(2539)] = 133328,
  [SMALL_STATE(2540)] = 133380,
  [SMALL_STATE(2541)] = 133434,
  [SMALL_STATE(2542)] = 133486,
  [SMALL_STATE(2543)] = 133534,
  [SMALL_STATE(2544)] = 133588,
  [SMALL_STATE(2545)] = 133636,
  [SMALL_STATE(2546)] = 133684,
  [SMALL_STATE(2547)] = 133740,
  [SMALL_STATE(2548)] = 133796,
  [SMALL_STATE(2549)] = 133848,
  [SMALL_STATE(2550)] = 133904,
  [SMALL_STATE(2551)] = 133958,
  [SMALL_STATE(2552)] = 134012,
  [SMALL_STATE(2553)] = 134068,
  [SMALL_STATE(2554)] = 134122,
  [SMALL_STATE(2555)] = 134170,
  [SMALL_STATE(2556)] = 134224,
  [SMALL_STATE(2557)] = 134272,
  [SMALL_STATE(2558)] = 134324,
  [SMALL_STATE(2559)] = 134372,
  [SMALL_STATE(2560)] = 134422,
  [SMALL_STATE(2561)] = 134470,
  [SMALL_STATE(2562)] = 134526,
  [SMALL_STATE(2563)] = 134574,
  [SMALL_STATE(2564)] = 134630,
  [SMALL_STATE(2565)] = 134684,
  [SMALL_STATE(2566)] = 134732,
  [SMALL_STATE(2567)] = 134780,
  [SMALL_STATE(2568)] = 134828,
  [SMALL_STATE(2569)] = 134880,
  [SMALL_STATE(2570)] = 134932,
  [SMALL_STATE(2571)] = 134980,
  [SMALL_STATE(2572)] = 135028,
  [SMALL_STATE(2573)] = 135082,
  [SMALL_STATE(2574)] = 135134,
  [SMALL_STATE(2575)] = 135182,
  [SMALL_STATE(2576)] = 135236,
  [SMALL_STATE(2577)] = 135284,
  [SMALL_STATE(2578)] = 135332,
  [SMALL_STATE(2579)] = 135386,
  [SMALL_STATE(2580)] = 135434,
  [SMALL_STATE(2581)] = 135482,
  [SMALL_STATE(2582)] = 135530,
  [SMALL_STATE(2583)] = 135584,
  [SMALL_STATE(2584)] = 135631,
  [SMALL_STATE(2585)] = 135678,
  [SMALL_STATE(2586)] = 135725,
  [SMALL_STATE(2587)] = 135772,
  [SMALL_STATE(2588)] = 135819,
  [SMALL_STATE(2589)] = 135870,
  [SMALL_STATE(2590)] = 135919,
  [SMALL_STATE(2591)] = 135966,
  [SMALL_STATE(2592)] = 136013,
  [SMALL_STATE(2593)] = 136060,
  [SMALL_STATE(2594)] = 136107,
  [SMALL_STATE(2595)] = 136154,
  [SMALL_STATE(2596)] = 136205,
  [SMALL_STATE(2597)] = 136252,
  [SMALL_STATE(2598)] = 136299,
  [SMALL_STATE(2599)] = 136346,
  [SMALL_STATE(2600)] = 136393,
  [SMALL_STATE(2601)] = 136440,
  [SMALL_STATE(2602)] = 136489,
  [SMALL_STATE(2603)] = 136540,
  [SMALL_STATE(2604)] = 136587,
  [SMALL_STATE(2605)] = 136634,
  [SMALL_STATE(2606)] = 136681,
  [SMALL_STATE(2607)] = 136728,
  [SMALL_STATE(2608)] = 136775,
  [SMALL_STATE(2609)] = 136822,
  [SMALL_STATE(2610)] = 136869,
  [SMALL_STATE(2611)] = 136916,
  [SMALL_STATE(2612)] = 136963,
  [SMALL_STATE(2613)] = 137014,
  [SMALL_STATE(2614)] = 137065,
  [SMALL_STATE(2615)] = 137112,
  [SMALL_STATE(2616)] = 137159,
  [SMALL_STATE(2617)] = 137206,
  [SMALL_STATE(2618)] = 137253,
  [SMALL_STATE(2619)] = 137306,
  [SMALL_STATE(2620)] = 137359,
  [SMALL_STATE(2621)] = 137410,
  [SMALL_STATE(2622)] = 137459,
  [SMALL_STATE(2623)] = 137506,
  [SMALL_STATE(2624)] = 137553,
  [SMALL_STATE(2625)] = 137600,
  [SMALL_STATE(2626)] = 137647,
  [SMALL_STATE(2627)] = 137694,
  [SMALL_STATE(2628)] = 137741,
  [SMALL_STATE(2629)] = 137788,
  [SMALL_STATE(2630)] = 137835,
  [SMALL_STATE(2631)] = 137882,
  [SMALL_STATE(2632)] = 137929,
  [SMALL_STATE(2633)] = 137980,
  [SMALL_STATE(2634)] = 138027,
  [SMALL_STATE(2635)] = 138074,
  [SMALL_STATE(2636)] = 138121,
  [SMALL_STATE(2637)] = 138168,
  [SMALL_STATE(2638)] = 138215,
  [SMALL_STATE(2639)] = 138262,
  [SMALL_STATE(2640)] = 138309,
  [SMALL_STATE(2641)] = 138356,
  [SMALL_STATE(2642)] = 138407,
  [SMALL_STATE(2643)] = 138454,
  [SMALL_STATE(2644)] = 138501,
  [SMALL_STATE(2645)] = 138548,
  [SMALL_STATE(2646)] = 138595,
  [SMALL_STATE(2647)] = 138642,
  [SMALL_STATE(2648)] = 138689,
  [SMALL_STATE(2649)] = 138736,
  [SMALL_STATE(2650)] = 138783,
  [SMALL_STATE(2651)] = 138830,
  [SMALL_STATE(2652)] = 138877,
  [SMALL_STATE(2653)] = 138924,
  [SMALL_STATE(2654)] = 138971,
  [SMALL_STATE(2655)] = 139018,
  [SMALL_STATE(2656)] = 139065,
  [SMALL_STATE(2657)] = 139112,
  [SMALL_STATE(2658)] = 139159,
  [SMALL_STATE(2659)] = 139206,
  [SMALL_STATE(2660)] = 139253,
  [SMALL_STATE(2661)] = 139300,
  [SMALL_STATE(2662)] = 139347,
  [SMALL_STATE(2663)] = 139394,
  [SMALL_STATE(2664)] = 139441,
  [SMALL_STATE(2665)] = 139488,
  [SMALL_STATE(2666)] = 139535,
  [SMALL_STATE(2667)] = 139582,
  [SMALL_STATE(2668)] = 139629,
  [SMALL_STATE(2669)] = 139676,
  [SMALL_STATE(2670)] = 139723,
  [SMALL_STATE(2671)] = 139770,
  [SMALL_STATE(2672)] = 139821,
  [SMALL_STATE(2673)] = 139868,
  [SMALL_STATE(2674)] = 139915,
  [SMALL_STATE(2675)] = 139962,
  [SMALL_STATE(2676)] = 140009,
  [SMALL_STATE(2677)] = 140056,
  [SMALL_STATE(2678)] = 140103,
  [SMALL_STATE(2679)] = 140154,
  [SMALL_STATE(2680)] = 140201,
  [SMALL_STATE(2681)] = 140248,
  [SMALL_STATE(2682)] = 140295,
  [SMALL_STATE(2683)] = 140342,
  [SMALL_STATE(2684)] = 140389,
  [SMALL_STATE(2685)] = 140436,
  [SMALL_STATE(2686)] = 140483,
  [SMALL_STATE(2687)] = 140530,
  [SMALL_STATE(2688)] = 140577,
  [SMALL_STATE(2689)] = 140628,
  [SMALL_STATE(2690)] = 140679,
  [SMALL_STATE(2691)] = 140730,
  [SMALL_STATE(2692)] = 140785,
  [SMALL_STATE(2693)] = 140836,
  [SMALL_STATE(2694)] = 140891,
  [SMALL_STATE(2695)] = 140940,
  [SMALL_STATE(2696)] = 140991,
  [SMALL_STATE(2697)] = 141040,
  [SMALL_STATE(2698)] = 141087,
  [SMALL_STATE(2699)] = 141134,
  [SMALL_STATE(2700)] = 141185,
  [SMALL_STATE(2701)] = 141232,
  [SMALL_STATE(2702)] = 141279,
  [SMALL_STATE(2703)] = 141328,
  [SMALL_STATE(2704)] = 141379,
  [SMALL_STATE(2705)] = 141428,
  [SMALL_STATE(2706)] = 141485,
  [SMALL_STATE(2707)] = 141542,
  [SMALL_STATE(2708)] = 141589,
  [SMALL_STATE(2709)] = 141636,
  [SMALL_STATE(2710)] = 141683,
  [SMALL_STATE(2711)] = 141730,
  [SMALL_STATE(2712)] = 141777,
  [SMALL_STATE(2713)] = 141824,
  [SMALL_STATE(2714)] = 141871,
  [SMALL_STATE(2715)] = 141922,
  [SMALL_STATE(2716)] = 141973,
  [SMALL_STATE(2717)] = 142020,
  [SMALL_STATE(2718)] = 142071,
  [SMALL_STATE(2719)] = 142122,
  [SMALL_STATE(2720)] = 142169,
  [SMALL_STATE(2721)] = 142222,
  [SMALL_STATE(2722)] = 142269,
  [SMALL_STATE(2723)] = 142316,
  [SMALL_STATE(2724)] = 142369,
  [SMALL_STATE(2725)] = 142416,
  [SMALL_STATE(2726)] = 142463,
  [SMALL_STATE(2727)] = 142514,
  [SMALL_STATE(2728)] = 142565,
  [SMALL_STATE(2729)] = 142612,
  [SMALL_STATE(2730)] = 142659,
  [SMALL_STATE(2731)] = 142706,
  [SMALL_STATE(2732)] = 142753,
  [SMALL_STATE(2733)] = 142800,
  [SMALL_STATE(2734)] = 142847,
  [SMALL_STATE(2735)] = 142894,
  [SMALL_STATE(2736)] = 142941,
  [SMALL_STATE(2737)] = 142988,
  [SMALL_STATE(2738)] = 143034,
  [SMALL_STATE(2739)] = 143090,
  [SMALL_STATE(2740)] = 143172,
  [SMALL_STATE(2741)] = 143254,
  [SMALL_STATE(2742)] = 143336,
  [SMALL_STATE(2743)] = 143418,
  [SMALL_STATE(2744)] = 143500,
  [SMALL_STATE(2745)] = 143582,
  [SMALL_STATE(2746)] = 143664,
  [SMALL_STATE(2747)] = 143746,
  [SMALL_STATE(2748)] = 143828,
  [SMALL_STATE(2749)] = 143910,
  [SMALL_STATE(2750)] = 143992,
  [SMALL_STATE(2751)] = 144074,
  [SMALL_STATE(2752)] = 144156,
  [SMALL_STATE(2753)] = 144238,
  [SMALL_STATE(2754)] = 144320,
  [SMALL_STATE(2755)] = 144402,
  [SMALL_STATE(2756)] = 144484,
  [SMALL_STATE(2757)] = 144566,
  [SMALL_STATE(2758)] = 144648,
  [SMALL_STATE(2759)] = 144730,
  [SMALL_STATE(2760)] = 144812,
  [SMALL_STATE(2761)] = 144894,
  [SMALL_STATE(2762)] = 144976,
  [SMALL_STATE(2763)] = 145058,
  [SMALL_STATE(2764)] = 145140,
  [SMALL_STATE(2765)] = 145222,
  [SMALL_STATE(2766)] = 145304,
  [SMALL_STATE(2767)] = 145386,
  [SMALL_STATE(2768)] = 145468,
  [SMALL_STATE(2769)] = 145550,
  [SMALL_STATE(2770)] = 145632,
  [SMALL_STATE(2771)] = 145714,
  [SMALL_STATE(2772)] = 145796,
  [SMALL_STATE(2773)] = 145878,
  [SMALL_STATE(2774)] = 145960,
  [SMALL_STATE(2775)] = 146042,
  [SMALL_STATE(2776)] = 146124,
  [SMALL_STATE(2777)] = 146174,
  [SMALL_STATE(2778)] = 146256,
  [SMALL_STATE(2779)] = 146338,
  [SMALL_STATE(2780)] = 146420,
  [SMALL_STATE(2781)] = 146474,
  [SMALL_STATE(2782)] = 146556,
  [SMALL_STATE(2783)] = 146638,
  [SMALL_STATE(2784)] = 146720,
  [SMALL_STATE(2785)] = 146802,
  [SMALL_STATE(2786)] = 146884,
  [SMALL_STATE(2787)] = 146966,
  [SMALL_STATE(2788)] = 147048,
  [SMALL_STATE(2789)] = 147130,
  [SMALL_STATE(2790)] = 147212,
  [SMALL_STATE(2791)] = 147294,
  [SMALL_STATE(2792)] = 147376,
  [SMALL_STATE(2793)] = 147458,
  [SMALL_STATE(2794)] = 147540,
  [SMALL_STATE(2795)] = 147622,
  [SMALL_STATE(2796)] = 147704,
  [SMALL_STATE(2797)] = 147750,
  [SMALL_STATE(2798)] = 147832,
  [SMALL_STATE(2799)] = 147878,
  [SMALL_STATE(2800)] = 147960,
  [SMALL_STATE(2801)] = 148042,
  [SMALL_STATE(2802)] = 148088,
  [SMALL_STATE(2803)] = 148158,
  [SMALL_STATE(2804)] = 148204,
  [SMALL_STATE(2805)] = 148286,
  [SMALL_STATE(2806)] = 148332,
  [SMALL_STATE(2807)] = 148382,
  [SMALL_STATE(2808)] = 148428,
  [SMALL_STATE(2809)] = 148478,
  [SMALL_STATE(2810)] = 148524,
  [SMALL_STATE(2811)] = 148574,
  [SMALL_STATE(2812)] = 148656,
  [SMALL_STATE(2813)] = 148706,
  [SMALL_STATE(2814)] = 148758,
  [SMALL_STATE(2815)] = 148812,
  [SMALL_STATE(2816)] = 148862,
  [SMALL_STATE(2817)] = 148920,
  [SMALL_STATE(2818)] = 148966,
  [SMALL_STATE(2819)] = 149012,
  [SMALL_STATE(2820)] = 149058,
  [SMALL_STATE(2821)] = 149140,
  [SMALL_STATE(2822)] = 149186,
  [SMALL_STATE(2823)] = 149232,
  [SMALL_STATE(2824)] = 149278,
  [SMALL_STATE(2825)] = 149324,
  [SMALL_STATE(2826)] = 149370,
  [SMALL_STATE(2827)] = 149416,
  [SMALL_STATE(2828)] = 149478,
  [SMALL_STATE(2829)] = 149542,
  [SMALL_STATE(2830)] = 149608,
  [SMALL_STATE(2831)] = 149676,
  [SMALL_STATE(2832)] = 149748,
  [SMALL_STATE(2833)] = 149826,
  [SMALL_STATE(2834)] = 149876,
  [SMALL_STATE(2835)] = 149926,
  [SMALL_STATE(2836)] = 150002,
  [SMALL_STATE(2837)] = 150084,
  [SMALL_STATE(2838)] = 150140,
  [SMALL_STATE(2839)] = 150188,
  [SMALL_STATE(2840)] = 150234,
  [SMALL_STATE(2841)] = 150290,
  [SMALL_STATE(2842)] = 150336,
  [SMALL_STATE(2843)] = 150382,
  [SMALL_STATE(2844)] = 150430,
  [SMALL_STATE(2845)] = 150512,
  [SMALL_STATE(2846)] = 150584,
  [SMALL_STATE(2847)] = 150630,
  [SMALL_STATE(2848)] = 150676,
  [SMALL_STATE(2849)] = 150732,
  [SMALL_STATE(2850)] = 150778,
  [SMALL_STATE(2851)] = 150824,
  [SMALL_STATE(2852)] = 150874,
  [SMALL_STATE(2853)] = 150920,
  [SMALL_STATE(2854)] = 150966,
  [SMALL_STATE(2855)] = 151012,
  [SMALL_STATE(2856)] = 151058,
  [SMALL_STATE(2857)] = 151104,
  [SMALL_STATE(2858)] = 151150,
  [SMALL_STATE(2859)] = 151232,
  [SMALL_STATE(2860)] = 151278,
  [SMALL_STATE(2861)] = 151324,
  [SMALL_STATE(2862)] = 151370,
  [SMALL_STATE(2863)] = 151418,
  [SMALL_STATE(2864)] = 151464,
  [SMALL_STATE(2865)] = 151510,
  [SMALL_STATE(2866)] = 151558,
  [SMALL_STATE(2867)] = 151604,
  [SMALL_STATE(2868)] = 151654,
  [SMALL_STATE(2869)] = 151728,
  [SMALL_STATE(2870)] = 151774,
  [SMALL_STATE(2871)] = 151820,
  [SMALL_STATE(2872)] = 151866,
  [SMALL_STATE(2873)] = 151912,
  [SMALL_STATE(2874)] = 151958,
  [SMALL_STATE(2875)] = 152008,
  [SMALL_STATE(2876)] = 152054,
  [SMALL_STATE(2877)] = 152100,
  [SMALL_STATE(2878)] = 152146,
  [SMALL_STATE(2879)] = 152192,
  [SMALL_STATE(2880)] = 152238,
  [SMALL_STATE(2881)] = 152320,
  [SMALL_STATE(2882)] = 152366,
  [SMALL_STATE(2883)] = 152412,
  [SMALL_STATE(2884)] = 152458,
  [SMALL_STATE(2885)] = 152504,
  [SMALL_STATE(2886)] = 152550,
  [SMALL_STATE(2887)] = 152604,
  [SMALL_STATE(2888)] = 152660,
  [SMALL_STATE(2889)] = 152716,
  [SMALL_STATE(2890)] = 152762,
  [SMALL_STATE(2891)] = 152836,
  [SMALL_STATE(2892)] = 152882,
  [SMALL_STATE(2893)] = 152928,
  [SMALL_STATE(2894)] = 152974,
  [SMALL_STATE(2895)] = 153050,
  [SMALL_STATE(2896)] = 153106,
  [SMALL_STATE(2897)] = 153162,
  [SMALL_STATE(2898)] = 153208,
  [SMALL_STATE(2899)] = 153278,
  [SMALL_STATE(2900)] = 153346,
  [SMALL_STATE(2901)] = 153392,
  [SMALL_STATE(2902)] = 153438,
  [SMALL_STATE(2903)] = 153494,
  [SMALL_STATE(2904)] = 153550,
  [SMALL_STATE(2905)] = 153596,
  [SMALL_STATE(2906)] = 153642,
  [SMALL_STATE(2907)] = 153688,
  [SMALL_STATE(2908)] = 153734,
  [SMALL_STATE(2909)] = 153780,
  [SMALL_STATE(2910)] = 153826,
  [SMALL_STATE(2911)] = 153872,
  [SMALL_STATE(2912)] = 153918,
  [SMALL_STATE(2913)] = 153968,
  [SMALL_STATE(2914)] = 154014,
  [SMALL_STATE(2915)] = 154064,
  [SMALL_STATE(2916)] = 154130,
  [SMALL_STATE(2917)] = 154194,
  [SMALL_STATE(2918)] = 154256,
  [SMALL_STATE(2919)] = 154302,
  [SMALL_STATE(2920)] = 154348,
  [SMALL_STATE(2921)] = 154394,
  [SMALL_STATE(2922)] = 154454,
  [SMALL_STATE(2923)] = 154510,
  [SMALL_STATE(2924)] = 154564,
  [SMALL_STATE(2925)] = 154620,
  [SMALL_STATE(2926)] = 154668,
  [SMALL_STATE(2927)] = 154714,
  [SMALL_STATE(2928)] = 154760,
  [SMALL_STATE(2929)] = 154806,
  [SMALL_STATE(2930)] = 154852,
  [SMALL_STATE(2931)] = 154898,
  [SMALL_STATE(2932)] = 154944,
  [SMALL_STATE(2933)] = 155026,
  [SMALL_STATE(2934)] = 155084,
  [SMALL_STATE(2935)] = 155130,
  [SMALL_STATE(2936)] = 155176,
  [SMALL_STATE(2937)] = 155228,
  [SMALL_STATE(2938)] = 155278,
  [SMALL_STATE(2939)] = 155326,
  [SMALL_STATE(2940)] = 155376,
  [SMALL_STATE(2941)] = 155422,
  [SMALL_STATE(2942)] = 155472,
  [SMALL_STATE(2943)] = 155551,
  [SMALL_STATE(2944)] = 155614,
  [SMALL_STATE(2945)] = 155703,
  [SMALL_STATE(2946)] = 155750,
  [SMALL_STATE(2947)] = 155795,
  [SMALL_STATE(2948)] = 155864,
  [SMALL_STATE(2949)] = 155929,
  [SMALL_STATE(2950)] = 155976,
  [SMALL_STATE(2951)] = 156021,
  [SMALL_STATE(2952)] = 156094,
  [SMALL_STATE(2953)] = 156157,
  [SMALL_STATE(2954)] = 156202,
  [SMALL_STATE(2955)] = 156247,
  [SMALL_STATE(2956)] = 156320,
  [SMALL_STATE(2957)] = 156381,
  [SMALL_STATE(2958)] = 156430,
  [SMALL_STATE(2959)] = 156479,
  [SMALL_STATE(2960)] = 156530,
  [SMALL_STATE(2961)] = 156583,
  [SMALL_STATE(2962)] = 156640,
  [SMALL_STATE(2963)] = 156695,
  [SMALL_STATE(2964)] = 156756,
  [SMALL_STATE(2965)] = 156819,
  [SMALL_STATE(2966)] = 156878,
  [SMALL_STATE(2967)] = 156943,
  [SMALL_STATE(2968)] = 156992,
  [SMALL_STATE(2969)] = 157041,
  [SMALL_STATE(2970)] = 157092,
  [SMALL_STATE(2971)] = 157145,
  [SMALL_STATE(2972)] = 157202,
  [SMALL_STATE(2973)] = 157263,
  [SMALL_STATE(2974)] = 157326,
  [SMALL_STATE(2975)] = 157391,
  [SMALL_STATE(2976)] = 157458,
  [SMALL_STATE(2977)] = 157527,
  [SMALL_STATE(2978)] = 157598,
  [SMALL_STATE(2979)] = 157675,
  [SMALL_STATE(2980)] = 157750,
  [SMALL_STATE(2981)] = 157795,
  [SMALL_STATE(2982)] = 157862,
  [SMALL_STATE(2983)] = 157951,
  [SMALL_STATE(2984)] = 157998,
  [SMALL_STATE(2985)] = 158045,
  [SMALL_STATE(2986)] = 158114,
  [SMALL_STATE(2987)] = 158185,
  [SMALL_STATE(2988)] = 158236,
  [SMALL_STATE(2989)] = 158313,
  [SMALL_STATE(2990)] = 158388,
  [SMALL_STATE(2991)] = 158433,
  [SMALL_STATE(2992)] = 158488,
  [SMALL_STATE(2993)] = 158541,
  [SMALL_STATE(2994)] = 158592,
  [SMALL_STATE(2995)] = 158647,
  [SMALL_STATE(2996)] = 158722,
  [SMALL_STATE(2997)] = 158799,
  [SMALL_STATE(2998)] = 158844,
  [SMALL_STATE(2999)] = 158889,
  [SMALL_STATE(3000)] = 158938,
  [SMALL_STATE(3001)] = 159007,
  [SMALL_STATE(3002)] = 159074,
  [SMALL_STATE(3003)] = 159139,
  [SMALL_STATE(3004)] = 159202,
  [SMALL_STATE(3005)] = 159263,
  [SMALL_STATE(3006)] = 159312,
  [SMALL_STATE(3007)] = 159367,
  [SMALL_STATE(3008)] = 159424,
  [SMALL_STATE(3009)] = 159477,
  [SMALL_STATE(3010)] = 159524,
  [SMALL_STATE(3011)] = 159575,
  [SMALL_STATE(3012)] = 159630,
  [SMALL_STATE(3013)] = 159687,
  [SMALL_STATE(3014)] = 159736,
  [SMALL_STATE(3015)] = 159785,
  [SMALL_STATE(3016)] = 159840,
  [SMALL_STATE(3017)] = 159895,
  [SMALL_STATE(3018)] = 159972,
  [SMALL_STATE(3019)] = 160027,
  [SMALL_STATE(3020)] = 160084,
  [SMALL_STATE(3021)] = 160129,
  [SMALL_STATE(3022)] = 160174,
  [SMALL_STATE(3023)] = 160253,
  [SMALL_STATE(3024)] = 160302,
  [SMALL_STATE(3025)] = 160347,
  [SMALL_STATE(3026)] = 160392,
  [SMALL_STATE(3027)] = 160437,
  [SMALL_STATE(3028)] = 160492,
  [SMALL_STATE(3029)] = 160539,
  [SMALL_STATE(3030)] = 160586,
  [SMALL_STATE(3031)] = 160635,
  [SMALL_STATE(3032)] = 160682,
  [SMALL_STATE(3033)] = 160771,
  [SMALL_STATE(3034)] = 160850,
  [SMALL_STATE(3035)] = 160921,
  [SMALL_STATE(3036)] = 160968,
  [SMALL_STATE(3037)] = 161035,
  [SMALL_STATE(3038)] = 161114,
  [SMALL_STATE(3039)] = 161203,
  [SMALL_STATE(3040)] = 161256,
  [SMALL_STATE(3041)] = 161335,
  [SMALL_STATE(3042)] = 161392,
  [SMALL_STATE(3043)] = 161471,
  [SMALL_STATE(3044)] = 161526,
  [SMALL_STATE(3045)] = 161581,
  [SMALL_STATE(3046)] = 161626,
  [SMALL_STATE(3047)] = 161683,
  [SMALL_STATE(3048)] = 161738,
  [SMALL_STATE(3049)] = 161793,
  [SMALL_STATE(3050)] = 161838,
  [SMALL_STATE(3051)] = 161911,
  [SMALL_STATE(3052)] = 161972,
  [SMALL_STATE(3053)] = 162051,
  [SMALL_STATE(3054)] = 162096,
  [SMALL_STATE(3055)] = 162151,
  [SMALL_STATE(3056)] = 162196,
  [SMALL_STATE(3057)] = 162241,
  [SMALL_STATE(3058)] = 162330,
  [SMALL_STATE(3059)] = 162375,
  [SMALL_STATE(3060)] = 162440,
  [SMALL_STATE(3061)] = 162513,
  [SMALL_STATE(3062)] = 162580,
  [SMALL_STATE(3063)] = 162649,
  [SMALL_STATE(3064)] = 162720,
  [SMALL_STATE(3065)] = 162797,
  [SMALL_STATE(3066)] = 162846,
  [SMALL_STATE(3067)] = 162935,
  [SMALL_STATE(3068)] = 163010,
  [SMALL_STATE(3069)] = 163065,
  [SMALL_STATE(3070)] = 163112,
  [SMALL_STATE(3071)] = 163201,
  [SMALL_STATE(3072)] = 163246,
  [SMALL_STATE(3073)] = 163293,
  [SMALL_STATE(3074)] = 163372,
  [SMALL_STATE(3075)] = 163417,
  [SMALL_STATE(3076)] = 163464,
  [SMALL_STATE(3077)] = 163509,
  [SMALL_STATE(3078)] = 163558,
  [SMALL_STATE(3079)] = 163607,
  [SMALL_STATE(3080)] = 163658,
  [SMALL_STATE(3081)] = 163711,
  [SMALL_STATE(3082)] = 163768,
  [SMALL_STATE(3083)] = 163829,
  [SMALL_STATE(3084)] = 163892,
  [SMALL_STATE(3085)] = 163957,
  [SMALL_STATE(3086)] = 164036,
  [SMALL_STATE(3087)] = 164103,
  [SMALL_STATE(3088)] = 164182,
  [SMALL_STATE(3089)] = 164251,
  [SMALL_STATE(3090)] = 164308,
  [SMALL_STATE(3091)] = 164379,
  [SMALL_STATE(3092)] = 164458,
  [SMALL_STATE(3093)] = 164509,
  [SMALL_STATE(3094)] = 164560,
  [SMALL_STATE(3095)] = 164637,
  [SMALL_STATE(3096)] = 164712,
  [SMALL_STATE(3097)] = 164767,
  [SMALL_STATE(3098)] = 164822,
  [SMALL_STATE(3099)] = 164879,
  [SMALL_STATE(3100)] = 164926,
  [SMALL_STATE(3101)] = 164981,
  [SMALL_STATE(3102)] = 165036,
  [SMALL_STATE(3103)] = 165091,
  [SMALL_STATE(3104)] = 165138,
  [SMALL_STATE(3105)] = 165217,
  [SMALL_STATE(3106)] = 165306,
  [SMALL_STATE(3107)] = 165395,
  [SMALL_STATE(3108)] = 165450,
  [SMALL_STATE(3109)] = 165505,
  [SMALL_STATE(3110)] = 165594,
  [SMALL_STATE(3111)] = 165645,
  [SMALL_STATE(3112)] = 165724,
  [SMALL_STATE(3113)] = 165781,
  [SMALL_STATE(3114)] = 165826,
  [SMALL_STATE(3115)] = 165875,
  [SMALL_STATE(3116)] = 165926,
  [SMALL_STATE(3117)] = 165971,
  [SMALL_STATE(3118)] = 166022,
  [SMALL_STATE(3119)] = 166067,
  [SMALL_STATE(3120)] = 166138,
  [SMALL_STATE(3121)] = 166189,
  [SMALL_STATE(3122)] = 166234,
  [SMALL_STATE(3123)] = 166323,
  [SMALL_STATE(3124)] = 166402,
  [SMALL_STATE(3125)] = 166481,
  [SMALL_STATE(3126)] = 166560,
  [SMALL_STATE(3127)] = 166639,
  [SMALL_STATE(3128)] = 166702,
  [SMALL_STATE(3129)] = 166781,
  [SMALL_STATE(3130)] = 166860,
  [SMALL_STATE(3131)] = 166907,
  [SMALL_STATE(3132)] = 166986,
  [SMALL_STATE(3133)] = 167041,
  [SMALL_STATE(3134)] = 167130,
  [SMALL_STATE(3135)] = 167185,
  [SMALL_STATE(3136)] = 167274,
  [SMALL_STATE(3137)] = 167363,
  [SMALL_STATE(3138)] = 167452,
  [SMALL_STATE(3139)] = 167497,
  [SMALL_STATE(3140)] = 167576,
  [SMALL_STATE(3141)] = 167655,
  [SMALL_STATE(3142)] = 167734,
  [SMALL_STATE(3143)] = 167813,
  [SMALL_STATE(3144)] = 167892,
  [SMALL_STATE(3145)] = 167971,
  [SMALL_STATE(3146)] = 168050,
  [SMALL_STATE(3147)] = 168095,
  [SMALL_STATE(3148)] = 168174,
  [SMALL_STATE(3149)] = 168253,
  [SMALL_STATE(3150)] = 168298,
  [SMALL_STATE(3151)] = 168377,
  [SMALL_STATE(3152)] = 168450,
  [SMALL_STATE(3153)] = 168529,
  [SMALL_STATE(3154)] = 168608,
  [SMALL_STATE(3155)] = 168687,
  [SMALL_STATE(3156)] = 168766,
  [SMALL_STATE(3157)] = 168845,
  [SMALL_STATE(3158)] = 168890,
  [SMALL_STATE(3159)] = 168969,
  [SMALL_STATE(3160)] = 169014,
  [SMALL_STATE(3161)] = 169059,
  [SMALL_STATE(3162)] = 169104,
  [SMALL_STATE(3163)] = 169159,
  [SMALL_STATE(3164)] = 169214,
  [SMALL_STATE(3165)] = 169303,
  [SMALL_STATE(3166)] = 169379,
  [SMALL_STATE(3167)] = 169455,
  [SMALL_STATE(3168)] = 169499,
  [SMALL_STATE(3169)] = 169543,
  [SMALL_STATE(3170)] = 169587,
  [SMALL_STATE(3171)] = 169631,
  [SMALL_STATE(3172)] = 169675,
  [SMALL_STATE(3173)] = 169719,
  [SMALL_STATE(3174)] = 169763,
  [SMALL_STATE(3175)] = 169807,
  [SMALL_STATE(3176)] = 169851,
  [SMALL_STATE(3177)] = 169895,
  [SMALL_STATE(3178)] = 169939,
  [SMALL_STATE(3179)] = 169983,
  [SMALL_STATE(3180)] = 170027,
  [SMALL_STATE(3181)] = 170071,
  [SMALL_STATE(3182)] = 170115,
  [SMALL_STATE(3183)] = 170191,
  [SMALL_STATE(3184)] = 170245,
  [SMALL_STATE(3185)] = 170289,
  [SMALL_STATE(3186)] = 170365,
  [SMALL_STATE(3187)] = 170419,
  [SMALL_STATE(3188)] = 170473,
  [SMALL_STATE(3189)] = 170543,
  [SMALL_STATE(3190)] = 170587,
  [SMALL_STATE(3191)] = 170631,
  [SMALL_STATE(3192)] = 170675,
  [SMALL_STATE(3193)] = 170751,
  [SMALL_STATE(3194)] = 170805,
  [SMALL_STATE(3195)] = 170849,
  [SMALL_STATE(3196)] = 170925,
  [SMALL_STATE(3197)] = 170973,
  [SMALL_STATE(3198)] = 171029,
  [SMALL_STATE(3199)] = 171077,
  [SMALL_STATE(3200)] = 171127,
  [SMALL_STATE(3201)] = 171179,
  [SMALL_STATE(3202)] = 171233,
  [SMALL_STATE(3203)] = 171291,
  [SMALL_STATE(3204)] = 171335,
  [SMALL_STATE(3205)] = 171411,
  [SMALL_STATE(3206)] = 171471,
  [SMALL_STATE(3207)] = 171515,
  [SMALL_STATE(3208)] = 171577,
  [SMALL_STATE(3209)] = 171621,
  [SMALL_STATE(3210)] = 171697,
  [SMALL_STATE(3211)] = 171741,
  [SMALL_STATE(3212)] = 171785,
  [SMALL_STATE(3213)] = 171861,
  [SMALL_STATE(3214)] = 171905,
  [SMALL_STATE(3215)] = 171969,
  [SMALL_STATE(3216)] = 172013,
  [SMALL_STATE(3217)] = 172077,
  [SMALL_STATE(3218)] = 172143,
  [SMALL_STATE(3219)] = 172211,
  [SMALL_STATE(3220)] = 172255,
  [SMALL_STATE(3221)] = 172299,
  [SMALL_STATE(3222)] = 172343,
  [SMALL_STATE(3223)] = 172387,
  [SMALL_STATE(3224)] = 172431,
  [SMALL_STATE(3225)] = 172507,
  [SMALL_STATE(3226)] = 172551,
  [SMALL_STATE(3227)] = 172595,
  [SMALL_STATE(3228)] = 172649,
  [SMALL_STATE(3229)] = 172693,
  [SMALL_STATE(3230)] = 172737,
  [SMALL_STATE(3231)] = 172791,
  [SMALL_STATE(3232)] = 172835,
  [SMALL_STATE(3233)] = 172879,
  [SMALL_STATE(3234)] = 172965,
  [SMALL_STATE(3235)] = 173009,
  [SMALL_STATE(3236)] = 173085,
  [SMALL_STATE(3237)] = 173171,
  [SMALL_STATE(3238)] = 173225,
  [SMALL_STATE(3239)] = 173279,
  [SMALL_STATE(3240)] = 173355,
  [SMALL_STATE(3241)] = 173409,
  [SMALL_STATE(3242)] = 173453,
  [SMALL_STATE(3243)] = 173499,
  [SMALL_STATE(3244)] = 173553,
  [SMALL_STATE(3245)] = 173629,
  [SMALL_STATE(3246)] = 173673,
  [SMALL_STATE(3247)] = 173721,
  [SMALL_STATE(3248)] = 173771,
  [SMALL_STATE(3249)] = 173823,
  [SMALL_STATE(3250)] = 173877,
  [SMALL_STATE(3251)] = 173921,
  [SMALL_STATE(3252)] = 173965,
  [SMALL_STATE(3253)] = 174041,
  [SMALL_STATE(3254)] = 174085,
  [SMALL_STATE(3255)] = 174129,
  [SMALL_STATE(3256)] = 174173,
  [SMALL_STATE(3257)] = 174217,
  [SMALL_STATE(3258)] = 174293,
  [SMALL_STATE(3259)] = 174351,
  [SMALL_STATE(3260)] = 174411,
  [SMALL_STATE(3261)] = 174487,
  [SMALL_STATE(3262)] = 174549,
  [SMALL_STATE(3263)] = 174613,
  [SMALL_STATE(3264)] = 174679,
  [SMALL_STATE(3265)] = 174747,
  [SMALL_STATE(3266)] = 174821,
  [SMALL_STATE(3267)] = 174893,
  [SMALL_STATE(3268)] = 174937,
  [SMALL_STATE(3269)] = 175013,
  [SMALL_STATE(3270)] = 175067,
  [SMALL_STATE(3271)] = 175121,
  [SMALL_STATE(3272)] = 175165,
  [SMALL_STATE(3273)] = 175209,
  [SMALL_STATE(3274)] = 175285,
  [SMALL_STATE(3275)] = 175361,
  [SMALL_STATE(3276)] = 175405,
  [SMALL_STATE(3277)] = 175451,
  [SMALL_STATE(3278)] = 175495,
  [SMALL_STATE(3279)] = 175549,
  [SMALL_STATE(3280)] = 175593,
  [SMALL_STATE(3281)] = 175637,
  [SMALL_STATE(3282)] = 175683,
  [SMALL_STATE(3283)] = 175727,
  [SMALL_STATE(3284)] = 175781,
  [SMALL_STATE(3285)] = 175857,
  [SMALL_STATE(3286)] = 175901,
  [SMALL_STATE(3287)] = 175945,
  [SMALL_STATE(3288)] = 175989,
  [SMALL_STATE(3289)] = 176033,
  [SMALL_STATE(3290)] = 176109,
  [SMALL_STATE(3291)] = 176153,
  [SMALL_STATE(3292)] = 176197,
  [SMALL_STATE(3293)] = 176267,
  [SMALL_STATE(3294)] = 176343,
  [SMALL_STATE(3295)] = 176387,
  [SMALL_STATE(3296)] = 176431,
  [SMALL_STATE(3297)] = 176507,
  [SMALL_STATE(3298)] = 176551,
  [SMALL_STATE(3299)] = 176605,
  [SMALL_STATE(3300)] = 176681,
  [SMALL_STATE(3301)] = 176725,
  [SMALL_STATE(3302)] = 176801,
  [SMALL_STATE(3303)] = 176877,
  [SMALL_STATE(3304)] = 176953,
  [SMALL_STATE(3305)] = 177029,
  [SMALL_STATE(3306)] = 177083,
  [SMALL_STATE(3307)] = 177127,
  [SMALL_STATE(3308)] = 177171,
  [SMALL_STATE(3309)] = 177247,
  [SMALL_STATE(3310)] = 177323,
  [SMALL_STATE(3311)] = 177409,
  [SMALL_STATE(3312)] = 177453,
  [SMALL_STATE(3313)] = 177497,
  [SMALL_STATE(3314)] = 177551,
  [SMALL_STATE(3315)] = 177625,
  [SMALL_STATE(3316)] = 177701,
  [SMALL_STATE(3317)] = 177777,
  [SMALL_STATE(3318)] = 177853,
  [SMALL_STATE(3319)] = 177929,
  [SMALL_STATE(3320)] = 177999,
  [SMALL_STATE(3321)] = 178075,
  [SMALL_STATE(3322)] = 178151,
  [SMALL_STATE(3323)] = 178195,
  [SMALL_STATE(3324)] = 178263,
  [SMALL_STATE(3325)] = 178339,
  [SMALL_STATE(3326)] = 178415,
  [SMALL_STATE(3327)] = 178491,
  [SMALL_STATE(3328)] = 178557,
  [SMALL_STATE(3329)] = 178633,
  [SMALL_STATE(3330)] = 178709,
  [SMALL_STATE(3331)] = 178773,
  [SMALL_STATE(3332)] = 178817,
  [SMALL_STATE(3333)] = 178879,
  [SMALL_STATE(3334)] = 178939,
  [SMALL_STATE(3335)] = 179015,
  [SMALL_STATE(3336)] = 179071,
  [SMALL_STATE(3337)] = 179147,
  [SMALL_STATE(3338)] = 179223,
  [SMALL_STATE(3339)] = 179299,
  [SMALL_STATE(3340)] = 179375,
  [SMALL_STATE(3341)] = 179451,
  [SMALL_STATE(3342)] = 179527,
  [SMALL_STATE(3343)] = 179603,
  [SMALL_STATE(3344)] = 179679,
  [SMALL_STATE(3345)] = 179755,
  [SMALL_STATE(3346)] = 179807,
  [SMALL_STATE(3347)] = 179883,
  [SMALL_STATE(3348)] = 179933,
  [SMALL_STATE(3349)] = 180009,
  [SMALL_STATE(3350)] = 180085,
  [SMALL_STATE(3351)] = 180161,
  [SMALL_STATE(3352)] = 180209,
  [SMALL_STATE(3353)] = 180253,
  [SMALL_STATE(3354)] = 180297,
  [SMALL_STATE(3355)] = 180341,
  [SMALL_STATE(3356)] = 180417,
  [SMALL_STATE(3357)] = 180465,
  [SMALL_STATE(3358)] = 180541,
  [SMALL_STATE(3359)] = 180613,
  [SMALL_STATE(3360)] = 180657,
  [SMALL_STATE(3361)] = 180701,
  [SMALL_STATE(3362)] = 180745,
  [SMALL_STATE(3363)] = 180821,
  [SMALL_STATE(3364)] = 180897,
  [SMALL_STATE(3365)] = 180973,
  [SMALL_STATE(3366)] = 181049,
  [SMALL_STATE(3367)] = 181125,
  [SMALL_STATE(3368)] = 181169,
  [SMALL_STATE(3369)] = 181245,
  [SMALL_STATE(3370)] = 181321,
  [SMALL_STATE(3371)] = 181397,
  [SMALL_STATE(3372)] = 181473,
  [SMALL_STATE(3373)] = 181549,
  [SMALL_STATE(3374)] = 181625,
  [SMALL_STATE(3375)] = 181701,
  [SMALL_STATE(3376)] = 181745,
  [SMALL_STATE(3377)] = 181789,
  [SMALL_STATE(3378)] = 181875,
  [SMALL_STATE(3379)] = 181961,
  [SMALL_STATE(3380)] = 182031,
  [SMALL_STATE(3381)] = 182075,
  [SMALL_STATE(3382)] = 182123,
  [SMALL_STATE(3383)] = 182209,
  [SMALL_STATE(3384)] = 182295,
  [SMALL_STATE(3385)] = 182343,
  [SMALL_STATE(3386)] = 182429,
  [SMALL_STATE(3387)] = 182515,
  [SMALL_STATE(3388)] = 182565,
  [SMALL_STATE(3389)] = 182651,
  [SMALL_STATE(3390)] = 182737,
  [SMALL_STATE(3391)] = 182789,
  [SMALL_STATE(3392)] = 182833,
  [SMALL_STATE(3393)] = 182887,
  [SMALL_STATE(3394)] = 182931,
  [SMALL_STATE(3395)] = 182989,
  [SMALL_STATE(3396)] = 183049,
  [SMALL_STATE(3397)] = 183093,
  [SMALL_STATE(3398)] = 183169,
  [SMALL_STATE(3399)] = 183213,
  [SMALL_STATE(3400)] = 183275,
  [SMALL_STATE(3401)] = 183319,
  [SMALL_STATE(3402)] = 183395,
  [SMALL_STATE(3403)] = 183459,
  [SMALL_STATE(3404)] = 183523,
  [SMALL_STATE(3405)] = 183589,
  [SMALL_STATE(3406)] = 183657,
  [SMALL_STATE(3407)] = 183701,
  [SMALL_STATE(3408)] = 183787,
  [SMALL_STATE(3409)] = 183863,
  [SMALL_STATE(3410)] = 183909,
  [SMALL_STATE(3411)] = 183957,
  [SMALL_STATE(3412)] = 184043,
  [SMALL_STATE(3413)] = 184093,
  [SMALL_STATE(3414)] = 184145,
  [SMALL_STATE(3415)] = 184199,
  [SMALL_STATE(3416)] = 184243,
  [SMALL_STATE(3417)] = 184329,
  [SMALL_STATE(3418)] = 184383,
  [SMALL_STATE(3419)] = 184427,
  [SMALL_STATE(3420)] = 184513,
  [SMALL_STATE(3421)] = 184599,
  [SMALL_STATE(3422)] = 184653,
  [SMALL_STATE(3423)] = 184739,
  [SMALL_STATE(3424)] = 184825,
  [SMALL_STATE(3425)] = 184911,
  [SMALL_STATE(3426)] = 184969,
  [SMALL_STATE(3427)] = 185045,
  [SMALL_STATE(3428)] = 185089,
  [SMALL_STATE(3429)] = 185149,
  [SMALL_STATE(3430)] = 185211,
  [SMALL_STATE(3431)] = 185275,
  [SMALL_STATE(3432)] = 185341,
  [SMALL_STATE(3433)] = 185409,
  [SMALL_STATE(3434)] = 185483,
  [SMALL_STATE(3435)] = 185555,
  [SMALL_STATE(3436)] = 185631,
  [SMALL_STATE(3437)] = 185675,
  [SMALL_STATE(3438)] = 185761,
  [SMALL_STATE(3439)] = 185847,
  [SMALL_STATE(3440)] = 185891,
  [SMALL_STATE(3441)] = 185977,
  [SMALL_STATE(3442)] = 186047,
  [SMALL_STATE(3443)] = 186091,
  [SMALL_STATE(3444)] = 186135,
  [SMALL_STATE(3445)] = 186179,
  [SMALL_STATE(3446)] = 186223,
  [SMALL_STATE(3447)] = 186277,
  [SMALL_STATE(3448)] = 186323,
  [SMALL_STATE(3449)] = 186367,
  [SMALL_STATE(3450)] = 186421,
  [SMALL_STATE(3451)] = 186465,
  [SMALL_STATE(3452)] = 186551,
  [SMALL_STATE(3453)] = 186637,
  [SMALL_STATE(3454)] = 186681,
  [SMALL_STATE(3455)] = 186727,
  [SMALL_STATE(3456)] = 186771,
  [SMALL_STATE(3457)] = 186818,
  [SMALL_STATE(3458)] = 186871,
  [SMALL_STATE(3459)] = 186950,
  [SMALL_STATE(3460)] = 186999,
  [SMALL_STATE(3461)] = 187078,
  [SMALL_STATE(3462)] = 187127,
  [SMALL_STATE(3463)] = 187214,
  [SMALL_STATE(3464)] = 187267,
  [SMALL_STATE(3465)] = 187320,
  [SMALL_STATE(3466)] = 187367,
  [SMALL_STATE(3467)] = 187414,
  [SMALL_STATE(3468)] = 187461,
  [SMALL_STATE(3469)] = 187508,
  [SMALL_STATE(3470)] = 187555,
  [SMALL_STATE(3471)] = 187642,
  [SMALL_STATE(3472)] = 187689,
  [SMALL_STATE(3473)] = 187736,
  [SMALL_STATE(3474)] = 187783,
  [SMALL_STATE(3475)] = 187836,
  [SMALL_STATE(3476)] = 187915,
  [SMALL_STATE(3477)] = 187962,
  [SMALL_STATE(3478)] = 188045,
  [SMALL_STATE(3479)] = 188124,
  [SMALL_STATE(3480)] = 188171,
  [SMALL_STATE(3481)] = 188218,
  [SMALL_STATE(3482)] = 188265,
  [SMALL_STATE(3483)] = 188318,
  [SMALL_STATE(3484)] = 188365,
  [SMALL_STATE(3485)] = 188444,
  [SMALL_STATE(3486)] = 188497,
  [SMALL_STATE(3487)] = 188584,
  [SMALL_STATE(3488)] = 188666,
  [SMALL_STATE(3489)] = 188708,
  [SMALL_STATE(3490)] = 188790,
  [SMALL_STATE(3491)] = 188832,
  [SMALL_STATE(3492)] = 188908,
  [SMALL_STATE(3493)] = 188950,
  [SMALL_STATE(3494)] = 189032,
  [SMALL_STATE(3495)] = 189074,
  [SMALL_STATE(3496)] = 189156,
  [SMALL_STATE(3497)] = 189238,
  [SMALL_STATE(3498)] = 189320,
  [SMALL_STATE(3499)] = 189362,
  [SMALL_STATE(3500)] = 189444,
  [SMALL_STATE(3501)] = 189526,
  [SMALL_STATE(3502)] = 189568,
  [SMALL_STATE(3503)] = 189650,
  [SMALL_STATE(3504)] = 189732,
  [SMALL_STATE(3505)] = 189774,
  [SMALL_STATE(3506)] = 189816,
  [SMALL_STATE(3507)] = 189858,
  [SMALL_STATE(3508)] = 189940,
  [SMALL_STATE(3509)] = 189982,
  [SMALL_STATE(3510)] = 190064,
  [SMALL_STATE(3511)] = 190106,
  [SMALL_STATE(3512)] = 190148,
  [SMALL_STATE(3513)] = 190190,
  [SMALL_STATE(3514)] = 190232,
  [SMALL_STATE(3515)] = 190314,
  [SMALL_STATE(3516)] = 190356,
  [SMALL_STATE(3517)] = 190398,
  [SMALL_STATE(3518)] = 190480,
  [SMALL_STATE(3519)] = 190562,
  [SMALL_STATE(3520)] = 190644,
  [SMALL_STATE(3521)] = 190726,
  [SMALL_STATE(3522)] = 190808,
  [SMALL_STATE(3523)] = 190850,
  [SMALL_STATE(3524)] = 190932,
  [SMALL_STATE(3525)] = 191014,
  [SMALL_STATE(3526)] = 191056,
  [SMALL_STATE(3527)] = 191138,
  [SMALL_STATE(3528)] = 191220,
  [SMALL_STATE(3529)] = 191262,
  [SMALL_STATE(3530)] = 191304,
  [SMALL_STATE(3531)] = 191386,
  [SMALL_STATE(3532)] = 191459,
  [SMALL_STATE(3533)] = 191532,
  [SMALL_STATE(3534)] = 191605,
  [SMALL_STATE(3535)] = 191678,
  [SMALL_STATE(3536)] = 191729,
  [SMALL_STATE(3537)] = 191774,
  [SMALL_STATE(3538)] = 191825,
  [SMALL_STATE(3539)] = 191876,
  [SMALL_STATE(3540)] = 191921,
  [SMALL_STATE(3541)] = 191972,
  [SMALL_STATE(3542)] = 192017,
  [SMALL_STATE(3543)] = 192062,
  [SMALL_STATE(3544)] = 192107,
  [SMALL_STATE(3545)] = 192152,
  [SMALL_STATE(3546)] = 192197,
  [SMALL_STATE(3547)] = 192273,
  [SMALL_STATE(3548)] = 192313,
  [SMALL_STATE(3549)] = 192389,
  [SMALL_STATE(3550)] = 192465,
  [SMALL_STATE(3551)] = 192541,
  [SMALL_STATE(3552)] = 192617,
  [SMALL_STATE(3553)] = 192657,
  [SMALL_STATE(3554)] = 192733,
  [SMALL_STATE(3555)] = 192809,
  [SMALL_STATE(3556)] = 192885,
  [SMALL_STATE(3557)] = 192925,
  [SMALL_STATE(3558)] = 193001,
  [SMALL_STATE(3559)] = 193077,
  [SMALL_STATE(3560)] = 193127,
  [SMALL_STATE(3561)] = 193177,
  [SMALL_STATE(3562)] = 193253,
  [SMALL_STATE(3563)] = 193303,
  [SMALL_STATE(3564)] = 193379,
  [SMALL_STATE(3565)] = 193455,
  [SMALL_STATE(3566)] = 193531,
  [SMALL_STATE(3567)] = 193607,
  [SMALL_STATE(3568)] = 193683,
  [SMALL_STATE(3569)] = 193759,
  [SMALL_STATE(3570)] = 193799,
  [SMALL_STATE(3571)] = 193875,
  [SMALL_STATE(3572)] = 193951,
  [SMALL_STATE(3573)] = 194001,
  [SMALL_STATE(3574)] = 194041,
  [SMALL_STATE(3575)] = 194117,
  [SMALL_STATE(3576)] = 194193,
  [SMALL_STATE(3577)] = 194269,
  [SMALL_STATE(3578)] = 194345,
  [SMALL_STATE(3579)] = 194385,
  [SMALL_STATE(3580)] = 194435,
  [SMALL_STATE(3581)] = 194485,
  [SMALL_STATE(3582)] = 194525,
  [SMALL_STATE(3583)] = 194575,
  [SMALL_STATE(3584)] = 194651,
  [SMALL_STATE(3585)] = 194701,
  [SMALL_STATE(3586)] = 194777,
  [SMALL_STATE(3587)] = 194853,
  [SMALL_STATE(3588)] = 194929,
  [SMALL_STATE(3589)] = 195005,
  [SMALL_STATE(3590)] = 195081,
  [SMALL_STATE(3591)] = 195157,
  [SMALL_STATE(3592)] = 195233,
  [SMALL_STATE(3593)] = 195309,
  [SMALL_STATE(3594)] = 195385,
  [SMALL_STATE(3595)] = 195461,
  [SMALL_STATE(3596)] = 195537,
  [SMALL_STATE(3597)] = 195613,
  [SMALL_STATE(3598)] = 195653,
  [SMALL_STATE(3599)] = 195703,
  [SMALL_STATE(3600)] = 195779,
  [SMALL_STATE(3601)] = 195829,
  [SMALL_STATE(3602)] = 195869,
  [SMALL_STATE(3603)] = 195945,
  [SMALL_STATE(3604)] = 196021,
  [SMALL_STATE(3605)] = 196097,
  [SMALL_STATE(3606)] = 196173,
  [SMALL_STATE(3607)] = 196249,
  [SMALL_STATE(3608)] = 196299,
  [SMALL_STATE(3609)] = 196375,
  [SMALL_STATE(3610)] = 196451,
  [SMALL_STATE(3611)] = 196527,
  [SMALL_STATE(3612)] = 196603,
  [SMALL_STATE(3613)] = 196653,
  [SMALL_STATE(3614)] = 196729,
  [SMALL_STATE(3615)] = 196802,
  [SMALL_STATE(3616)] = 196873,
  [SMALL_STATE(3617)] = 196944,
  [SMALL_STATE(3618)] = 197017,
  [SMALL_STATE(3619)] = 197090,
  [SMALL_STATE(3620)] = 197139,
  [SMALL_STATE(3621)] = 197214,
  [SMALL_STATE(3622)] = 197287,
  [SMALL_STATE(3623)] = 197360,
  [SMALL_STATE(3624)] = 197433,
  [SMALL_STATE(3625)] = 197506,
  [SMALL_STATE(3626)] = 197579,
  [SMALL_STATE(3627)] = 197652,
  [SMALL_STATE(3628)] = 197725,
  [SMALL_STATE(3629)] = 197796,
  [SMALL_STATE(3630)] = 197869,
  [SMALL_STATE(3631)] = 197942,
  [SMALL_STATE(3632)] = 198015,
  [SMALL_STATE(3633)] = 198088,
  [SMALL_STATE(3634)] = 198161,
  [SMALL_STATE(3635)] = 198234,
  [SMALL_STATE(3636)] = 198307,
  [SMALL_STATE(3637)] = 198356,
  [SMALL_STATE(3638)] = 198429,
  [SMALL_STATE(3639)] = 198502,
  [SMALL_STATE(3640)] = 198575,
  [SMALL_STATE(3641)] = 198648,
  [SMALL_STATE(3642)] = 198721,
  [SMALL_STATE(3643)] = 198794,
  [SMALL_STATE(3644)] = 198865,
  [SMALL_STATE(3645)] = 198936,
  [SMALL_STATE(3646)] = 199009,
  [SMALL_STATE(3647)] = 199082,
  [SMALL_STATE(3648)] = 199155,
  [SMALL_STATE(3649)] = 199228,
  [SMALL_STATE(3650)] = 199301,
  [SMALL_STATE(3651)] = 199374,
  [SMALL_STATE(3652)] = 199447,
  [SMALL_STATE(3653)] = 199520,
  [SMALL_STATE(3654)] = 199593,
  [SMALL_STATE(3655)] = 199666,
  [SMALL_STATE(3656)] = 199739,
  [SMALL_STATE(3657)] = 199810,
  [SMALL_STATE(3658)] = 199881,
  [SMALL_STATE(3659)] = 199952,
  [SMALL_STATE(3660)] = 200025,
  [SMALL_STATE(3661)] = 200100,
  [SMALL_STATE(3662)] = 200173,
  [SMALL_STATE(3663)] = 200246,
  [SMALL_STATE(3664)] = 200317,
  [SMALL_STATE(3665)] = 200366,
  [SMALL_STATE(3666)] = 200439,
  [SMALL_STATE(3667)] = 200488,
  [SMALL_STATE(3668)] = 200561,
  [SMALL_STATE(3669)] = 200632,
  [SMALL_STATE(3670)] = 200705,
  [SMALL_STATE(3671)] = 200778,
  [SMALL_STATE(3672)] = 200849,
  [SMALL_STATE(3673)] = 200924,
  [SMALL_STATE(3674)] = 200997,
  [SMALL_STATE(3675)] = 201070,
  [SMALL_STATE(3676)] = 201143,
  [SMALL_STATE(3677)] = 201216,
  [SMALL_STATE(3678)] = 201289,
  [SMALL_STATE(3679)] = 201362,
  [SMALL_STATE(3680)] = 201435,
  [SMALL_STATE(3681)] = 201508,
  [SMALL_STATE(3682)] = 201581,
  [SMALL_STATE(3683)] = 201656,
  [SMALL_STATE(3684)] = 201727,
  [SMALL_STATE(3685)] = 201800,
  [SMALL_STATE(3686)] = 201871,
  [SMALL_STATE(3687)] = 201946,
  [SMALL_STATE(3688)] = 202019,
  [SMALL_STATE(3689)] = 202092,
  [SMALL_STATE(3690)] = 202165,
  [SMALL_STATE(3691)] = 202238,
  [SMALL_STATE(3692)] = 202311,
  [SMALL_STATE(3693)] = 202384,
  [SMALL_STATE(3694)] = 202457,
  [SMALL_STATE(3695)] = 202530,
  [SMALL_STATE(3696)] = 202605,
  [SMALL_STATE(3697)] = 202678,
  [SMALL_STATE(3698)] = 202753,
  [SMALL_STATE(3699)] = 202821,
  [SMALL_STATE(3700)] = 202895,
  [SMALL_STATE(3701)] = 202963,
  [SMALL_STATE(3702)] = 203031,
  [SMALL_STATE(3703)] = 203099,
  [SMALL_STATE(3704)] = 203171,
  [SMALL_STATE(3705)] = 203239,
  [SMALL_STATE(3706)] = 203313,
  [SMALL_STATE(3707)] = 203385,
  [SMALL_STATE(3708)] = 203453,
  [SMALL_STATE(3709)] = 203527,
  [SMALL_STATE(3710)] = 203601,
  [SMALL_STATE(3711)] = 203675,
  [SMALL_STATE(3712)] = 203743,
  [SMALL_STATE(3713)] = 203811,
  [SMALL_STATE(3714)] = 203885,
  [SMALL_STATE(3715)] = 203959,
  [SMALL_STATE(3716)] = 204031,
  [SMALL_STATE(3717)] = 204103,
  [SMALL_STATE(3718)] = 204171,
  [SMALL_STATE(3719)] = 204243,
  [SMALL_STATE(3720)] = 204311,
  [SMALL_STATE(3721)] = 204383,
  [SMALL_STATE(3722)] = 204455,
  [SMALL_STATE(3723)] = 204529,
  [SMALL_STATE(3724)] = 204597,
  [SMALL_STATE(3725)] = 204671,
  [SMALL_STATE(3726)] = 204743,
  [SMALL_STATE(3727)] = 204817,
  [SMALL_STATE(3728)] = 204885,
  [SMALL_STATE(3729)] = 204959,
  [SMALL_STATE(3730)] = 205033,
  [SMALL_STATE(3731)] = 205107,
  [SMALL_STATE(3732)] = 205181,
  [SMALL_STATE(3733)] = 205249,
  [SMALL_STATE(3734)] = 205317,
  [SMALL_STATE(3735)] = 205389,
  [SMALL_STATE(3736)] = 205457,
  [SMALL_STATE(3737)] = 205529,
  [SMALL_STATE(3738)] = 205597,
  [SMALL_STATE(3739)] = 205665,
  [SMALL_STATE(3740)] = 205737,
  [SMALL_STATE(3741)] = 205809,
  [SMALL_STATE(3742)] = 205883,
  [SMALL_STATE(3743)] = 205957,
  [SMALL_STATE(3744)] = 206031,
  [SMALL_STATE(3745)] = 206105,
  [SMALL_STATE(3746)] = 206173,
  [SMALL_STATE(3747)] = 206241,
  [SMALL_STATE(3748)] = 206309,
  [SMALL_STATE(3749)] = 206383,
  [SMALL_STATE(3750)] = 206457,
  [SMALL_STATE(3751)] = 206525,
  [SMALL_STATE(3752)] = 206599,
  [SMALL_STATE(3753)] = 206667,
  [SMALL_STATE(3754)] = 206741,
  [SMALL_STATE(3755)] = 206815,
  [SMALL_STATE(3756)] = 206889,
  [SMALL_STATE(3757)] = 206957,
  [SMALL_STATE(3758)] = 207025,
  [SMALL_STATE(3759)] = 207093,
  [SMALL_STATE(3760)] = 207167,
  [SMALL_STATE(3761)] = 207235,
  [SMALL_STATE(3762)] = 207307,
  [SMALL_STATE(3763)] = 207379,
  [SMALL_STATE(3764)] = 207451,
  [SMALL_STATE(3765)] = 207525,
  [SMALL_STATE(3766)] = 207597,
  [SMALL_STATE(3767)] = 207665,
  [SMALL_STATE(3768)] = 207733,
  [SMALL_STATE(3769)] = 207801,
  [SMALL_STATE(3770)] = 207873,
  [SMALL_STATE(3771)] = 207941,
  [SMALL_STATE(3772)] = 208009,
  [SMALL_STATE(3773)] = 208081,
  [SMALL_STATE(3774)] = 208149,
  [SMALL_STATE(3775)] = 208217,
  [SMALL_STATE(3776)] = 208289,
  [SMALL_STATE(3777)] = 208357,
  [SMALL_STATE(3778)] = 208429,
  [SMALL_STATE(3779)] = 208497,
  [SMALL_STATE(3780)] = 208569,
  [SMALL_STATE(3781)] = 208641,
  [SMALL_STATE(3782)] = 208709,
  [SMALL_STATE(3783)] = 208781,
  [SMALL_STATE(3784)] = 208855,
  [SMALL_STATE(3785)] = 208929,
  [SMALL_STATE(3786)] = 208997,
  [SMALL_STATE(3787)] = 209071,
  [SMALL_STATE(3788)] = 209143,
  [SMALL_STATE(3789)] = 209211,
  [SMALL_STATE(3790)] = 209285,
  [SMALL_STATE(3791)] = 209353,
  [SMALL_STATE(3792)] = 209421,
  [SMALL_STATE(3793)] = 209489,
  [SMALL_STATE(3794)] = 209563,
  [SMALL_STATE(3795)] = 209631,
  [SMALL_STATE(3796)] = 209699,
  [SMALL_STATE(3797)] = 209767,
  [SMALL_STATE(3798)] = 209841,
  [SMALL_STATE(3799)] = 209909,
  [SMALL_STATE(3800)] = 209981,
  [SMALL_STATE(3801)] = 210055,
  [SMALL_STATE(3802)] = 210127,
  [SMALL_STATE(3803)] = 210195,
  [SMALL_STATE(3804)] = 210263,
  [SMALL_STATE(3805)] = 210337,
  [SMALL_STATE(3806)] = 210409,
  [SMALL_STATE(3807)] = 210477,
  [SMALL_STATE(3808)] = 210549,
  [SMALL_STATE(3809)] = 210621,
  [SMALL_STATE(3810)] = 210695,
  [SMALL_STATE(3811)] = 210763,
  [SMALL_STATE(3812)] = 210831,
  [SMALL_STATE(3813)] = 210905,
  [SMALL_STATE(3814)] = 210973,
  [SMALL_STATE(3815)] = 211041,
  [SMALL_STATE(3816)] = 211109,
  [SMALL_STATE(3817)] = 211183,
  [SMALL_STATE(3818)] = 211257,
  [SMALL_STATE(3819)] = 211331,
  [SMALL_STATE(3820)] = 211399,
  [SMALL_STATE(3821)] = 211473,
  [SMALL_STATE(3822)] = 211545,
  [SMALL_STATE(3823)] = 211617,
  [SMALL_STATE(3824)] = 211685,
  [SMALL_STATE(3825)] = 211759,
  [SMALL_STATE(3826)] = 211833,
  [SMALL_STATE(3827)] = 211901,
  [SMALL_STATE(3828)] = 211973,
  [SMALL_STATE(3829)] = 212045,
  [SMALL_STATE(3830)] = 212113,
  [SMALL_STATE(3831)] = 212187,
  [SMALL_STATE(3832)] = 212261,
  [SMALL_STATE(3833)] = 212333,
  [SMALL_STATE(3834)] = 212405,
  [SMALL_STATE(3835)] = 212473,
  [SMALL_STATE(3836)] = 212547,
  [SMALL_STATE(3837)] = 212619,
  [SMALL_STATE(3838)] = 212687,
  [SMALL_STATE(3839)] = 212759,
  [SMALL_STATE(3840)] = 212827,
  [SMALL_STATE(3841)] = 212899,
  [SMALL_STATE(3842)] = 212967,
  [SMALL_STATE(3843)] = 213041,
  [SMALL_STATE(3844)] = 213109,
  [SMALL_STATE(3845)] = 213177,
  [SMALL_STATE(3846)] = 213245,
  [SMALL_STATE(3847)] = 213313,
  [SMALL_STATE(3848)] = 213381,
  [SMALL_STATE(3849)] = 213449,
  [SMALL_STATE(3850)] = 213523,
  [SMALL_STATE(3851)] = 213597,
  [SMALL_STATE(3852)] = 213665,
  [SMALL_STATE(3853)] = 213733,
  [SMALL_STATE(3854)] = 213801,
  [SMALL_STATE(3855)] = 213869,
  [SMALL_STATE(3856)] = 213943,
  [SMALL_STATE(3857)] = 214011,
  [SMALL_STATE(3858)] = 214085,
  [SMALL_STATE(3859)] = 214153,
  [SMALL_STATE(3860)] = 214225,
  [SMALL_STATE(3861)] = 214293,
  [SMALL_STATE(3862)] = 214361,
  [SMALL_STATE(3863)] = 214435,
  [SMALL_STATE(3864)] = 214509,
  [SMALL_STATE(3865)] = 214577,
  [SMALL_STATE(3866)] = 214645,
  [SMALL_STATE(3867)] = 214713,
  [SMALL_STATE(3868)] = 214781,
  [SMALL_STATE(3869)] = 214849,
  [SMALL_STATE(3870)] = 214917,
  [SMALL_STATE(3871)] = 214985,
  [SMALL_STATE(3872)] = 215053,
  [SMALL_STATE(3873)] = 215121,
  [SMALL_STATE(3874)] = 215189,
  [SMALL_STATE(3875)] = 215263,
  [SMALL_STATE(3876)] = 215337,
  [SMALL_STATE(3877)] = 215411,
  [SMALL_STATE(3878)] = 215479,
  [SMALL_STATE(3879)] = 215547,
  [SMALL_STATE(3880)] = 215621,
  [SMALL_STATE(3881)] = 215689,
  [SMALL_STATE(3882)] = 215757,
  [SMALL_STATE(3883)] = 215825,
  [SMALL_STATE(3884)] = 215893,
  [SMALL_STATE(3885)] = 215961,
  [SMALL_STATE(3886)] = 216029,
  [SMALL_STATE(3887)] = 216097,
  [SMALL_STATE(3888)] = 216165,
  [SMALL_STATE(3889)] = 216239,
  [SMALL_STATE(3890)] = 216307,
  [SMALL_STATE(3891)] = 216375,
  [SMALL_STATE(3892)] = 216443,
  [SMALL_STATE(3893)] = 216517,
  [SMALL_STATE(3894)] = 216591,
  [SMALL_STATE(3895)] = 216659,
  [SMALL_STATE(3896)] = 216727,
  [SMALL_STATE(3897)] = 216795,
  [SMALL_STATE(3898)] = 216863,
  [SMALL_STATE(3899)] = 216935,
  [SMALL_STATE(3900)] = 217003,
  [SMALL_STATE(3901)] = 217071,
  [SMALL_STATE(3902)] = 217139,
  [SMALL_STATE(3903)] = 217211,
  [SMALL_STATE(3904)] = 217279,
  [SMALL_STATE(3905)] = 217347,
  [SMALL_STATE(3906)] = 217415,
  [SMALL_STATE(3907)] = 217483,
  [SMALL_STATE(3908)] = 217555,
  [SMALL_STATE(3909)] = 217623,
  [SMALL_STATE(3910)] = 217691,
  [SMALL_STATE(3911)] = 217759,
  [SMALL_STATE(3912)] = 217827,
  [SMALL_STATE(3913)] = 217895,
  [SMALL_STATE(3914)] = 217963,
  [SMALL_STATE(3915)] = 218031,
  [SMALL_STATE(3916)] = 218105,
  [SMALL_STATE(3917)] = 218179,
  [SMALL_STATE(3918)] = 218247,
  [SMALL_STATE(3919)] = 218315,
  [SMALL_STATE(3920)] = 218389,
  [SMALL_STATE(3921)] = 218463,
  [SMALL_STATE(3922)] = 218531,
  [SMALL_STATE(3923)] = 218599,
  [SMALL_STATE(3924)] = 218667,
  [SMALL_STATE(3925)] = 218735,
  [SMALL_STATE(3926)] = 218803,
  [SMALL_STATE(3927)] = 218871,
  [SMALL_STATE(3928)] = 218939,
  [SMALL_STATE(3929)] = 219007,
  [SMALL_STATE(3930)] = 219075,
  [SMALL_STATE(3931)] = 219143,
  [SMALL_STATE(3932)] = 219211,
  [SMALL_STATE(3933)] = 219285,
  [SMALL_STATE(3934)] = 219359,
  [SMALL_STATE(3935)] = 219433,
  [SMALL_STATE(3936)] = 219501,
  [SMALL_STATE(3937)] = 219569,
  [SMALL_STATE(3938)] = 219637,
  [SMALL_STATE(3939)] = 219705,
  [SMALL_STATE(3940)] = 219773,
  [SMALL_STATE(3941)] = 219841,
  [SMALL_STATE(3942)] = 219909,
  [SMALL_STATE(3943)] = 219977,
  [SMALL_STATE(3944)] = 220045,
  [SMALL_STATE(3945)] = 220119,
  [SMALL_STATE(3946)] = 220193,
  [SMALL_STATE(3947)] = 220261,
  [SMALL_STATE(3948)] = 220329,
  [SMALL_STATE(3949)] = 220397,
  [SMALL_STATE(3950)] = 220465,
  [SMALL_STATE(3951)] = 220533,
  [SMALL_STATE(3952)] = 220601,
  [SMALL_STATE(3953)] = 220669,
  [SMALL_STATE(3954)] = 220737,
  [SMALL_STATE(3955)] = 220805,
  [SMALL_STATE(3956)] = 220873,
  [SMALL_STATE(3957)] = 220941,
  [SMALL_STATE(3958)] = 221009,
  [SMALL_STATE(3959)] = 221077,
  [SMALL_STATE(3960)] = 221145,
  [SMALL_STATE(3961)] = 221219,
  [SMALL_STATE(3962)] = 221287,
  [SMALL_STATE(3963)] = 221355,
  [SMALL_STATE(3964)] = 221427,
  [SMALL_STATE(3965)] = 221495,
  [SMALL_STATE(3966)] = 221567,
  [SMALL_STATE(3967)] = 221641,
  [SMALL_STATE(3968)] = 221715,
  [SMALL_STATE(3969)] = 221783,
  [SMALL_STATE(3970)] = 221851,
  [SMALL_STATE(3971)] = 221925,
  [SMALL_STATE(3972)] = 221993,
  [SMALL_STATE(3973)] = 222061,
  [SMALL_STATE(3974)] = 222135,
  [SMALL_STATE(3975)] = 222209,
  [SMALL_STATE(3976)] = 222277,
  [SMALL_STATE(3977)] = 222345,
  [SMALL_STATE(3978)] = 222419,
  [SMALL_STATE(3979)] = 222493,
  [SMALL_STATE(3980)] = 222561,
  [SMALL_STATE(3981)] = 222629,
  [SMALL_STATE(3982)] = 222703,
  [SMALL_STATE(3983)] = 222777,
  [SMALL_STATE(3984)] = 222851,
  [SMALL_STATE(3985)] = 222919,
  [SMALL_STATE(3986)] = 222993,
  [SMALL_STATE(3987)] = 223067,
  [SMALL_STATE(3988)] = 223141,
  [SMALL_STATE(3989)] = 223209,
  [SMALL_STATE(3990)] = 223277,
  [SMALL_STATE(3991)] = 223345,
  [SMALL_STATE(3992)] = 223413,
  [SMALL_STATE(3993)] = 223481,
  [SMALL_STATE(3994)] = 223549,
  [SMALL_STATE(3995)] = 223623,
  [SMALL_STATE(3996)] = 223697,
  [SMALL_STATE(3997)] = 223765,
  [SMALL_STATE(3998)] = 223839,
  [SMALL_STATE(3999)] = 223907,
  [SMALL_STATE(4000)] = 223975,
  [SMALL_STATE(4001)] = 224049,
  [SMALL_STATE(4002)] = 224123,
  [SMALL_STATE(4003)] = 224191,
  [SMALL_STATE(4004)] = 224259,
  [SMALL_STATE(4005)] = 224327,
  [SMALL_STATE(4006)] = 224401,
  [SMALL_STATE(4007)] = 224475,
  [SMALL_STATE(4008)] = 224543,
  [SMALL_STATE(4009)] = 224617,
  [SMALL_STATE(4010)] = 224685,
  [SMALL_STATE(4011)] = 224753,
  [SMALL_STATE(4012)] = 224821,
  [SMALL_STATE(4013)] = 224895,
  [SMALL_STATE(4014)] = 224963,
  [SMALL_STATE(4015)] = 225031,
  [SMALL_STATE(4016)] = 225099,
  [SMALL_STATE(4017)] = 225167,
  [SMALL_STATE(4018)] = 225235,
  [SMALL_STATE(4019)] = 225307,
  [SMALL_STATE(4020)] = 225375,
  [SMALL_STATE(4021)] = 225443,
  [SMALL_STATE(4022)] = 225517,
  [SMALL_STATE(4023)] = 225585,
  [SMALL_STATE(4024)] = 225653,
  [SMALL_STATE(4025)] = 225725,
  [SMALL_STATE(4026)] = 225799,
  [SMALL_STATE(4027)] = 225867,
  [SMALL_STATE(4028)] = 225935,
  [SMALL_STATE(4029)] = 226003,
  [SMALL_STATE(4030)] = 226071,
  [SMALL_STATE(4031)] = 226143,
  [SMALL_STATE(4032)] = 226215,
  [SMALL_STATE(4033)] = 226289,
  [SMALL_STATE(4034)] = 226363,
  [SMALL_STATE(4035)] = 226431,
  [SMALL_STATE(4036)] = 226505,
  [SMALL_STATE(4037)] = 226577,
  [SMALL_STATE(4038)] = 226651,
  [SMALL_STATE(4039)] = 226719,
  [SMALL_STATE(4040)] = 226791,
  [SMALL_STATE(4041)] = 226865,
  [SMALL_STATE(4042)] = 226939,
  [SMALL_STATE(4043)] = 227007,
  [SMALL_STATE(4044)] = 227081,
  [SMALL_STATE(4045)] = 227155,
  [SMALL_STATE(4046)] = 227227,
  [SMALL_STATE(4047)] = 227295,
  [SMALL_STATE(4048)] = 227369,
  [SMALL_STATE(4049)] = 227443,
  [SMALL_STATE(4050)] = 227517,
  [SMALL_STATE(4051)] = 227589,
  [SMALL_STATE(4052)] = 227663,
  [SMALL_STATE(4053)] = 227735,
  [SMALL_STATE(4054)] = 227807,
  [SMALL_STATE(4055)] = 227881,
  [SMALL_STATE(4056)] = 227949,
  [SMALL_STATE(4057)] = 228023,
  [SMALL_STATE(4058)] = 228097,
  [SMALL_STATE(4059)] = 228171,
  [SMALL_STATE(4060)] = 228239,
  [SMALL_STATE(4061)] = 228313,
  [SMALL_STATE(4062)] = 228387,
  [SMALL_STATE(4063)] = 228461,
  [SMALL_STATE(4064)] = 228535,
  [SMALL_STATE(4065)] = 228609,
  [SMALL_STATE(4066)] = 228677,
  [SMALL_STATE(4067)] = 228751,
  [SMALL_STATE(4068)] = 228825,
  [SMALL_STATE(4069)] = 228899,
  [SMALL_STATE(4070)] = 228973,
  [SMALL_STATE(4071)] = 229047,
  [SMALL_STATE(4072)] = 229115,
  [SMALL_STATE(4073)] = 229189,
  [SMALL_STATE(4074)] = 229263,
  [SMALL_STATE(4075)] = 229337,
  [SMALL_STATE(4076)] = 229405,
  [SMALL_STATE(4077)] = 229479,
  [SMALL_STATE(4078)] = 229547,
  [SMALL_STATE(4079)] = 229619,
  [SMALL_STATE(4080)] = 229693,
  [SMALL_STATE(4081)] = 229761,
  [SMALL_STATE(4082)] = 229829,
  [SMALL_STATE(4083)] = 229903,
  [SMALL_STATE(4084)] = 229977,
  [SMALL_STATE(4085)] = 230051,
  [SMALL_STATE(4086)] = 230125,
  [SMALL_STATE(4087)] = 230193,
  [SMALL_STATE(4088)] = 230261,
  [SMALL_STATE(4089)] = 230329,
  [SMALL_STATE(4090)] = 230397,
  [SMALL_STATE(4091)] = 230471,
  [SMALL_STATE(4092)] = 230545,
  [SMALL_STATE(4093)] = 230613,
  [SMALL_STATE(4094)] = 230687,
  [SMALL_STATE(4095)] = 230755,
  [SMALL_STATE(4096)] = 230823,
  [SMALL_STATE(4097)] = 230897,
  [SMALL_STATE(4098)] = 230965,
  [SMALL_STATE(4099)] = 231039,
  [SMALL_STATE(4100)] = 231107,
  [SMALL_STATE(4101)] = 231175,
  [SMALL_STATE(4102)] = 231249,
  [SMALL_STATE(4103)] = 231323,
  [SMALL_STATE(4104)] = 231397,
  [SMALL_STATE(4105)] = 231465,
  [SMALL_STATE(4106)] = 231533,
  [SMALL_STATE(4107)] = 231601,
  [SMALL_STATE(4108)] = 231673,
  [SMALL_STATE(4109)] = 231747,
  [SMALL_STATE(4110)] = 231819,
  [SMALL_STATE(4111)] = 231893,
  [SMALL_STATE(4112)] = 231965,
  [SMALL_STATE(4113)] = 232039,
  [SMALL_STATE(4114)] = 232107,
  [SMALL_STATE(4115)] = 232181,
  [SMALL_STATE(4116)] = 232253,
  [SMALL_STATE(4117)] = 232321,
  [SMALL_STATE(4118)] = 232389,
  [SMALL_STATE(4119)] = 232461,
  [SMALL_STATE(4120)] = 232533,
  [SMALL_STATE(4121)] = 232607,
  [SMALL_STATE(4122)] = 232681,
  [SMALL_STATE(4123)] = 232753,
  [SMALL_STATE(4124)] = 232827,
  [SMALL_STATE(4125)] = 232901,
  [SMALL_STATE(4126)] = 232973,
  [SMALL_STATE(4127)] = 233045,
  [SMALL_STATE(4128)] = 233113,
  [SMALL_STATE(4129)] = 233181,
  [SMALL_STATE(4130)] = 233249,
  [SMALL_STATE(4131)] = 233323,
  [SMALL_STATE(4132)] = 233397,
  [SMALL_STATE(4133)] = 233471,
  [SMALL_STATE(4134)] = 233539,
  [SMALL_STATE(4135)] = 233613,
  [SMALL_STATE(4136)] = 233687,
  [SMALL_STATE(4137)] = 233761,
  [SMALL_STATE(4138)] = 233835,
  [SMALL_STATE(4139)] = 233907,
  [SMALL_STATE(4140)] = 233975,
  [SMALL_STATE(4141)] = 234043,
  [SMALL_STATE(4142)] = 234111,
  [SMALL_STATE(4143)] = 234185,
  [SMALL_STATE(4144)] = 234259,
  [SMALL_STATE(4145)] = 234333,
  [SMALL_STATE(4146)] = 234407,
  [SMALL_STATE(4147)] = 234481,
  [SMALL_STATE(4148)] = 234555,
  [SMALL_STATE(4149)] = 234629,
  [SMALL_STATE(4150)] = 234703,
  [SMALL_STATE(4151)] = 234777,
  [SMALL_STATE(4152)] = 234851,
  [SMALL_STATE(4153)] = 234923,
  [SMALL_STATE(4154)] = 234991,
  [SMALL_STATE(4155)] = 235065,
  [SMALL_STATE(4156)] = 235139,
  [SMALL_STATE(4157)] = 235207,
  [SMALL_STATE(4158)] = 235275,
  [SMALL_STATE(4159)] = 235343,
  [SMALL_STATE(4160)] = 235411,
  [SMALL_STATE(4161)] = 235479,
  [SMALL_STATE(4162)] = 235547,
  [SMALL_STATE(4163)] = 235615,
  [SMALL_STATE(4164)] = 235683,
  [SMALL_STATE(4165)] = 235757,
  [SMALL_STATE(4166)] = 235829,
  [SMALL_STATE(4167)] = 235903,
  [SMALL_STATE(4168)] = 235975,
  [SMALL_STATE(4169)] = 236043,
  [SMALL_STATE(4170)] = 236117,
  [SMALL_STATE(4171)] = 236189,
  [SMALL_STATE(4172)] = 236257,
  [SMALL_STATE(4173)] = 236325,
  [SMALL_STATE(4174)] = 236393,
  [SMALL_STATE(4175)] = 236465,
  [SMALL_STATE(4176)] = 236539,
  [SMALL_STATE(4177)] = 236613,
  [SMALL_STATE(4178)] = 236681,
  [SMALL_STATE(4179)] = 236749,
  [SMALL_STATE(4180)] = 236817,
  [SMALL_STATE(4181)] = 236885,
  [SMALL_STATE(4182)] = 236953,
  [SMALL_STATE(4183)] = 237025,
  [SMALL_STATE(4184)] = 237099,
  [SMALL_STATE(4185)] = 237173,
  [SMALL_STATE(4186)] = 237247,
  [SMALL_STATE(4187)] = 237319,
  [SMALL_STATE(4188)] = 237387,
  [SMALL_STATE(4189)] = 237459,
  [SMALL_STATE(4190)] = 237531,
  [SMALL_STATE(4191)] = 237605,
  [SMALL_STATE(4192)] = 237673,
  [SMALL_STATE(4193)] = 237747,
  [SMALL_STATE(4194)] = 237821,
  [SMALL_STATE(4195)] = 237895,
  [SMALL_STATE(4196)] = 237969,
  [SMALL_STATE(4197)] = 238043,
  [SMALL_STATE(4198)] = 238111,
  [SMALL_STATE(4199)] = 238179,
  [SMALL_STATE(4200)] = 238247,
  [SMALL_STATE(4201)] = 238321,
  [SMALL_STATE(4202)] = 238395,
  [SMALL_STATE(4203)] = 238469,
  [SMALL_STATE(4204)] = 238537,
  [SMALL_STATE(4205)] = 238611,
  [SMALL_STATE(4206)] = 238685,
  [SMALL_STATE(4207)] = 238759,
  [SMALL_STATE(4208)] = 238833,
  [SMALL_STATE(4209)] = 238907,
  [SMALL_STATE(4210)] = 238981,
  [SMALL_STATE(4211)] = 239055,
  [SMALL_STATE(4212)] = 239129,
  [SMALL_STATE(4213)] = 239203,
  [SMALL_STATE(4214)] = 239271,
  [SMALL_STATE(4215)] = 239339,
  [SMALL_STATE(4216)] = 239413,
  [SMALL_STATE(4217)] = 239460,
  [SMALL_STATE(4218)] = 239535,
  [SMALL_STATE(4219)] = 239578,
  [SMALL_STATE(4220)] = 239653,
  [SMALL_STATE(4221)] = 239728,
  [SMALL_STATE(4222)] = 239793,
  [SMALL_STATE(4223)] = 239836,
  [SMALL_STATE(4224)] = 239893,
  [SMALL_STATE(4225)] = 239938,
  [SMALL_STATE(4226)] = 239995,
  [SMALL_STATE(4227)] = 240070,
  [SMALL_STATE(4228)] = 240113,
  [SMALL_STATE(4229)] = 240160,
  [SMALL_STATE(4230)] = 240202,
  [SMALL_STATE(4231)] = 240244,
  [SMALL_STATE(4232)] = 240286,
  [SMALL_STATE(4233)] = 240328,
  [SMALL_STATE(4234)] = 240370,
  [SMALL_STATE(4235)] = 240412,
  [SMALL_STATE(4236)] = 240458,
  [SMALL_STATE(4237)] = 240530,
  [SMALL_STATE(4238)] = 240572,
  [SMALL_STATE(4239)] = 240614,
  [SMALL_STATE(4240)] = 240670,
  [SMALL_STATE(4241)] = 240712,
  [SMALL_STATE(4242)] = 240784,
  [SMALL_STATE(4243)] = 240826,
  [SMALL_STATE(4244)] = 240898,
  [SMALL_STATE(4245)] = 240940,
  [SMALL_STATE(4246)] = 240982,
  [SMALL_STATE(4247)] = 241024,
  [SMALL_STATE(4248)] = 241066,
  [SMALL_STATE(4249)] = 241108,
  [SMALL_STATE(4250)] = 241164,
  [SMALL_STATE(4251)] = 241210,
  [SMALL_STATE(4252)] = 241252,
  [SMALL_STATE(4253)] = 241294,
  [SMALL_STATE(4254)] = 241336,
  [SMALL_STATE(4255)] = 241380,
  [SMALL_STATE(4256)] = 241422,
  [SMALL_STATE(4257)] = 241464,
  [SMALL_STATE(4258)] = 241506,
  [SMALL_STATE(4259)] = 241548,
  [SMALL_STATE(4260)] = 241590,
  [SMALL_STATE(4261)] = 241632,
  [SMALL_STATE(4262)] = 241671,
  [SMALL_STATE(4263)] = 241710,
  [SMALL_STATE(4264)] = 241745,
  [SMALL_STATE(4265)] = 241780,
  [SMALL_STATE(4266)] = 241815,
  [SMALL_STATE(4267)] = 241856,
  [SMALL_STATE(4268)] = 241891,
  [SMALL_STATE(4269)] = 241932,
  [SMALL_STATE(4270)] = 241973,
  [SMALL_STATE(4271)] = 242014,
  [SMALL_STATE(4272)] = 242055,
  [SMALL_STATE(4273)] = 242096,
  [SMALL_STATE(4274)] = 242141,
  [SMALL_STATE(4275)] = 242182,
  [SMALL_STATE(4276)] = 242217,
  [SMALL_STATE(4277)] = 242258,
  [SMALL_STATE(4278)] = 242297,
  [SMALL_STATE(4279)] = 242342,
  [SMALL_STATE(4280)] = 242387,
  [SMALL_STATE(4281)] = 242422,
  [SMALL_STATE(4282)] = 242465,
  [SMALL_STATE(4283)] = 242506,
  [SMALL_STATE(4284)] = 242551,
  [SMALL_STATE(4285)] = 242592,
  [SMALL_STATE(4286)] = 242627,
  [SMALL_STATE(4287)] = 242668,
  [SMALL_STATE(4288)] = 242707,
  [SMALL_STATE(4289)] = 242742,
  [SMALL_STATE(4290)] = 242785,
  [SMALL_STATE(4291)] = 242824,
  [SMALL_STATE(4292)] = 242861,
  [SMALL_STATE(4293)] = 242902,
  [SMALL_STATE(4294)] = 242939,
  [SMALL_STATE(4295)] = 242978,
  [SMALL_STATE(4296)] = 243013,
  [SMALL_STATE(4297)] = 243054,
  [SMALL_STATE(4298)] = 243095,
  [SMALL_STATE(4299)] = 243130,
  [SMALL_STATE(4300)] = 243171,
  [SMALL_STATE(4301)] = 243206,
  [SMALL_STATE(4302)] = 243241,
  [SMALL_STATE(4303)] = 243276,
  [SMALL_STATE(4304)] = 243311,
  [SMALL_STATE(4305)] = 243352,
  [SMALL_STATE(4306)] = 243391,
  [SMALL_STATE(4307)] = 243426,
  [SMALL_STATE(4308)] = 243461,
  [SMALL_STATE(4309)] = 243496,
  [SMALL_STATE(4310)] = 243537,
  [SMALL_STATE(4311)] = 243578,
  [SMALL_STATE(4312)] = 243613,
  [SMALL_STATE(4313)] = 243668,
  [SMALL_STATE(4314)] = 243703,
  [SMALL_STATE(4315)] = 243758,
  [SMALL_STATE(4316)] = 243799,
  [SMALL_STATE(4317)] = 243834,
  [SMALL_STATE(4318)] = 243869,
  [SMALL_STATE(4319)] = 243904,
  [SMALL_STATE(4320)] = 243945,
  [SMALL_STATE(4321)] = 243980,
  [SMALL_STATE(4322)] = 244035,
  [SMALL_STATE(4323)] = 244070,
  [SMALL_STATE(4324)] = 244105,
  [SMALL_STATE(4325)] = 244144,
  [SMALL_STATE(4326)] = 244179,
  [SMALL_STATE(4327)] = 244214,
  [SMALL_STATE(4328)] = 244249,
  [SMALL_STATE(4329)] = 244284,
  [SMALL_STATE(4330)] = 244319,
  [SMALL_STATE(4331)] = 244354,
  [SMALL_STATE(4332)] = 244395,
  [SMALL_STATE(4333)] = 244430,
  [SMALL_STATE(4334)] = 244465,
  [SMALL_STATE(4335)] = 244500,
  [SMALL_STATE(4336)] = 244535,
  [SMALL_STATE(4337)] = 244570,
  [SMALL_STATE(4338)] = 244605,
  [SMALL_STATE(4339)] = 244640,
  [SMALL_STATE(4340)] = 244677,
  [SMALL_STATE(4341)] = 244712,
  [SMALL_STATE(4342)] = 244763,
  [SMALL_STATE(4343)] = 244802,
  [SMALL_STATE(4344)] = 244837,
  [SMALL_STATE(4345)] = 244874,
  [SMALL_STATE(4346)] = 244915,
  [SMALL_STATE(4347)] = 244960,
  [SMALL_STATE(4348)] = 244999,
  [SMALL_STATE(4349)] = 245034,
  [SMALL_STATE(4350)] = 245085,
  [SMALL_STATE(4351)] = 245126,
  [SMALL_STATE(4352)] = 245165,
  [SMALL_STATE(4353)] = 245206,
  [SMALL_STATE(4354)] = 245247,
  [SMALL_STATE(4355)] = 245282,
  [SMALL_STATE(4356)] = 245317,
  [SMALL_STATE(4357)] = 245352,
  [SMALL_STATE(4358)] = 245387,
  [SMALL_STATE(4359)] = 245428,
  [SMALL_STATE(4360)] = 245469,
  [SMALL_STATE(4361)] = 245504,
  [SMALL_STATE(4362)] = 245545,
  [SMALL_STATE(4363)] = 245586,
  [SMALL_STATE(4364)] = 245623,
  [SMALL_STATE(4365)] = 245664,
  [SMALL_STATE(4366)] = 245699,
  [SMALL_STATE(4367)] = 245734,
  [SMALL_STATE(4368)] = 245769,
  [SMALL_STATE(4369)] = 245808,
  [SMALL_STATE(4370)] = 245843,
  [SMALL_STATE(4371)] = 245884,
  [SMALL_STATE(4372)] = 245929,
  [SMALL_STATE(4373)] = 245964,
  [SMALL_STATE(4374)] = 245999,
  [SMALL_STATE(4375)] = 246034,
  [SMALL_STATE(4376)] = 246075,
  [SMALL_STATE(4377)] = 246118,
  [SMALL_STATE(4378)] = 246153,
  [SMALL_STATE(4379)] = 246188,
  [SMALL_STATE(4380)] = 246223,
  [SMALL_STATE(4381)] = 246264,
  [SMALL_STATE(4382)] = 246305,
  [SMALL_STATE(4383)] = 246344,
  [SMALL_STATE(4384)] = 246383,
  [SMALL_STATE(4385)] = 246418,
  [SMALL_STATE(4386)] = 246453,
  [SMALL_STATE(4387)] = 246488,
  [SMALL_STATE(4388)] = 246533,
  [SMALL_STATE(4389)] = 246568,
  [SMALL_STATE(4390)] = 246603,
  [SMALL_STATE(4391)] = 246644,
  [SMALL_STATE(4392)] = 246679,
  [SMALL_STATE(4393)] = 246714,
  [SMALL_STATE(4394)] = 246749,
  [SMALL_STATE(4395)] = 246788,
  [SMALL_STATE(4396)] = 246823,
  [SMALL_STATE(4397)] = 246878,
  [SMALL_STATE(4398)] = 246933,
  [SMALL_STATE(4399)] = 246968,
  [SMALL_STATE(4400)] = 247003,
  [SMALL_STATE(4401)] = 247058,
  [SMALL_STATE(4402)] = 247099,
  [SMALL_STATE(4403)] = 247134,
  [SMALL_STATE(4404)] = 247169,
  [SMALL_STATE(4405)] = 247204,
  [SMALL_STATE(4406)] = 247239,
  [SMALL_STATE(4407)] = 247280,
  [SMALL_STATE(4408)] = 247315,
  [SMALL_STATE(4409)] = 247350,
  [SMALL_STATE(4410)] = 247385,
  [SMALL_STATE(4411)] = 247426,
  [SMALL_STATE(4412)] = 247461,
  [SMALL_STATE(4413)] = 247496,
  [SMALL_STATE(4414)] = 247537,
  [SMALL_STATE(4415)] = 247578,
  [SMALL_STATE(4416)] = 247630,
  [SMALL_STATE(4417)] = 247680,
  [SMALL_STATE(4418)] = 247714,
  [SMALL_STATE(4419)] = 247754,
  [SMALL_STATE(4420)] = 247788,
  [SMALL_STATE(4421)] = 247832,
  [SMALL_STATE(4422)] = 247870,
  [SMALL_STATE(4423)] = 247910,
  [SMALL_STATE(4424)] = 247944,
  [SMALL_STATE(4425)] = 247978,
  [SMALL_STATE(4426)] = 248012,
  [SMALL_STATE(4427)] = 248046,
  [SMALL_STATE(4428)] = 248080,
  [SMALL_STATE(4429)] = 248116,
  [SMALL_STATE(4430)] = 248150,
  [SMALL_STATE(4431)] = 248190,
  [SMALL_STATE(4432)] = 248228,
  [SMALL_STATE(4433)] = 248262,
  [SMALL_STATE(4434)] = 248312,
  [SMALL_STATE(4435)] = 248346,
  [SMALL_STATE(4436)] = 248380,
  [SMALL_STATE(4437)] = 248414,
  [SMALL_STATE(4438)] = 248448,
  [SMALL_STATE(4439)] = 248482,
  [SMALL_STATE(4440)] = 248522,
  [SMALL_STATE(4441)] = 248562,
  [SMALL_STATE(4442)] = 248596,
  [SMALL_STATE(4443)] = 248630,
  [SMALL_STATE(4444)] = 248664,
  [SMALL_STATE(4445)] = 248698,
  [SMALL_STATE(4446)] = 248732,
  [SMALL_STATE(4447)] = 248766,
  [SMALL_STATE(4448)] = 248800,
  [SMALL_STATE(4449)] = 248834,
  [SMALL_STATE(4450)] = 248868,
  [SMALL_STATE(4451)] = 248908,
  [SMALL_STATE(4452)] = 248942,
  [SMALL_STATE(4453)] = 248982,
  [SMALL_STATE(4454)] = 249022,
  [SMALL_STATE(4455)] = 249056,
  [SMALL_STATE(4456)] = 249100,
  [SMALL_STATE(4457)] = 249134,
  [SMALL_STATE(4458)] = 249168,
  [SMALL_STATE(4459)] = 249208,
  [SMALL_STATE(4460)] = 249248,
  [SMALL_STATE(4461)] = 249288,
  [SMALL_STATE(4462)] = 249322,
  [SMALL_STATE(4463)] = 249360,
  [SMALL_STATE(4464)] = 249400,
  [SMALL_STATE(4465)] = 249434,
  [SMALL_STATE(4466)] = 249478,
  [SMALL_STATE(4467)] = 249518,
  [SMALL_STATE(4468)] = 249558,
  [SMALL_STATE(4469)] = 249598,
  [SMALL_STATE(4470)] = 249632,
  [SMALL_STATE(4471)] = 249666,
  [SMALL_STATE(4472)] = 249706,
  [SMALL_STATE(4473)] = 249746,
  [SMALL_STATE(4474)] = 249780,
  [SMALL_STATE(4475)] = 249814,
  [SMALL_STATE(4476)] = 249848,
  [SMALL_STATE(4477)] = 249888,
  [SMALL_STATE(4478)] = 249928,
  [SMALL_STATE(4479)] = 249962,
  [SMALL_STATE(4480)] = 250002,
  [SMALL_STATE(4481)] = 250046,
  [SMALL_STATE(4482)] = 250086,
  [SMALL_STATE(4483)] = 250120,
  [SMALL_STATE(4484)] = 250158,
  [SMALL_STATE(4485)] = 250198,
  [SMALL_STATE(4486)] = 250236,
  [SMALL_STATE(4487)] = 250276,
  [SMALL_STATE(4488)] = 250316,
  [SMALL_STATE(4489)] = 250356,
  [SMALL_STATE(4490)] = 250408,
  [SMALL_STATE(4491)] = 250442,
  [SMALL_STATE(4492)] = 250482,
  [SMALL_STATE(4493)] = 250520,
  [SMALL_STATE(4494)] = 250554,
  [SMALL_STATE(4495)] = 250588,
  [SMALL_STATE(4496)] = 250628,
  [SMALL_STATE(4497)] = 250662,
  [SMALL_STATE(4498)] = 250696,
  [SMALL_STATE(4499)] = 250734,
  [SMALL_STATE(4500)] = 250768,
  [SMALL_STATE(4501)] = 250806,
  [SMALL_STATE(4502)] = 250840,
  [SMALL_STATE(4503)] = 250874,
  [SMALL_STATE(4504)] = 250908,
  [SMALL_STATE(4505)] = 250948,
  [SMALL_STATE(4506)] = 250988,
  [SMALL_STATE(4507)] = 251028,
  [SMALL_STATE(4508)] = 251066,
  [SMALL_STATE(4509)] = 251100,
  [SMALL_STATE(4510)] = 251134,
  [SMALL_STATE(4511)] = 251174,
  [SMALL_STATE(4512)] = 251214,
  [SMALL_STATE(4513)] = 251248,
  [SMALL_STATE(4514)] = 251288,
  [SMALL_STATE(4515)] = 251328,
  [SMALL_STATE(4516)] = 251362,
  [SMALL_STATE(4517)] = 251396,
  [SMALL_STATE(4518)] = 251440,
  [SMALL_STATE(4519)] = 251480,
  [SMALL_STATE(4520)] = 251520,
  [SMALL_STATE(4521)] = 251560,
  [SMALL_STATE(4522)] = 251598,
  [SMALL_STATE(4523)] = 251638,
  [SMALL_STATE(4524)] = 251678,
  [SMALL_STATE(4525)] = 251718,
  [SMALL_STATE(4526)] = 251758,
  [SMALL_STATE(4527)] = 251792,
  [SMALL_STATE(4528)] = 251830,
  [SMALL_STATE(4529)] = 251868,
  [SMALL_STATE(4530)] = 251908,
  [SMALL_STATE(4531)] = 251952,
  [SMALL_STATE(4532)] = 251992,
  [SMALL_STATE(4533)] = 252026,
  [SMALL_STATE(4534)] = 252064,
  [SMALL_STATE(4535)] = 252104,
  [SMALL_STATE(4536)] = 252144,
  [SMALL_STATE(4537)] = 252178,
  [SMALL_STATE(4538)] = 252212,
  [SMALL_STATE(4539)] = 252250,
  [SMALL_STATE(4540)] = 252284,
  [SMALL_STATE(4541)] = 252318,
  [SMALL_STATE(4542)] = 252352,
  [SMALL_STATE(4543)] = 252386,
  [SMALL_STATE(4544)] = 252420,
  [SMALL_STATE(4545)] = 252454,
  [SMALL_STATE(4546)] = 252488,
  [SMALL_STATE(4547)] = 252526,
  [SMALL_STATE(4548)] = 252570,
  [SMALL_STATE(4549)] = 252604,
  [SMALL_STATE(4550)] = 252648,
  [SMALL_STATE(4551)] = 252688,
  [SMALL_STATE(4552)] = 252740,
  [SMALL_STATE(4553)] = 252774,
  [SMALL_STATE(4554)] = 252808,
  [SMALL_STATE(4555)] = 252846,
  [SMALL_STATE(4556)] = 252886,
  [SMALL_STATE(4557)] = 252920,
  [SMALL_STATE(4558)] = 252954,
  [SMALL_STATE(4559)] = 252994,
  [SMALL_STATE(4560)] = 253034,
  [SMALL_STATE(4561)] = 253068,
  [SMALL_STATE(4562)] = 253102,
  [SMALL_STATE(4563)] = 253142,
  [SMALL_STATE(4564)] = 253176,
  [SMALL_STATE(4565)] = 253210,
  [SMALL_STATE(4566)] = 253250,
  [SMALL_STATE(4567)] = 253284,
  [SMALL_STATE(4568)] = 253336,
  [SMALL_STATE(4569)] = 253376,
  [SMALL_STATE(4570)] = 253410,
  [SMALL_STATE(4571)] = 253452,
  [SMALL_STATE(4572)] = 253496,
  [SMALL_STATE(4573)] = 253530,
  [SMALL_STATE(4574)] = 253574,
  [SMALL_STATE(4575)] = 253614,
  [SMALL_STATE(4576)] = 253648,
  [SMALL_STATE(4577)] = 253682,
  [SMALL_STATE(4578)] = 253716,
  [SMALL_STATE(4579)] = 253756,
  [SMALL_STATE(4580)] = 253796,
  [SMALL_STATE(4581)] = 253840,
  [SMALL_STATE(4582)] = 253880,
  [SMALL_STATE(4583)] = 253920,
  [SMALL_STATE(4584)] = 253960,
  [SMALL_STATE(4585)] = 254000,
  [SMALL_STATE(4586)] = 254040,
  [SMALL_STATE(4587)] = 254080,
  [SMALL_STATE(4588)] = 254114,
  [SMALL_STATE(4589)] = 254148,
  [SMALL_STATE(4590)] = 254188,
  [SMALL_STATE(4591)] = 254222,
  [SMALL_STATE(4592)] = 254262,
  [SMALL_STATE(4593)] = 254302,
  [SMALL_STATE(4594)] = 254342,
  [SMALL_STATE(4595)] = 254382,
  [SMALL_STATE(4596)] = 254416,
  [SMALL_STATE(4597)] = 254454,
  [SMALL_STATE(4598)] = 254492,
  [SMALL_STATE(4599)] = 254526,
  [SMALL_STATE(4600)] = 254566,
  [SMALL_STATE(4601)] = 254600,
  [SMALL_STATE(4602)] = 254640,
  [SMALL_STATE(4603)] = 254680,
  [SMALL_STATE(4604)] = 254720,
  [SMALL_STATE(4605)] = 254760,
  [SMALL_STATE(4606)] = 254794,
  [SMALL_STATE(4607)] = 254828,
  [SMALL_STATE(4608)] = 254861,
  [SMALL_STATE(4609)] = 254896,
  [SMALL_STATE(4610)] = 254929,
  [SMALL_STATE(4611)] = 254962,
  [SMALL_STATE(4612)] = 255001,
  [SMALL_STATE(4613)] = 255034,
  [SMALL_STATE(4614)] = 255067,
  [SMALL_STATE(4615)] = 255106,
  [SMALL_STATE(4616)] = 255139,
  [SMALL_STATE(4617)] = 255176,
  [SMALL_STATE(4618)] = 255209,
  [SMALL_STATE(4619)] = 255242,
  [SMALL_STATE(4620)] = 255281,
  [SMALL_STATE(4621)] = 255328,
  [SMALL_STATE(4622)] = 255367,
  [SMALL_STATE(4623)] = 255406,
  [SMALL_STATE(4624)] = 255439,
  [SMALL_STATE(4625)] = 255498,
  [SMALL_STATE(4626)] = 255541,
  [SMALL_STATE(4627)] = 255580,
  [SMALL_STATE(4628)] = 255619,
  [SMALL_STATE(4629)] = 255652,
  [SMALL_STATE(4630)] = 255711,
  [SMALL_STATE(4631)] = 255750,
  [SMALL_STATE(4632)] = 255783,
  [SMALL_STATE(4633)] = 255816,
  [SMALL_STATE(4634)] = 255855,
  [SMALL_STATE(4635)] = 255894,
  [SMALL_STATE(4636)] = 255933,
  [SMALL_STATE(4637)] = 255966,
  [SMALL_STATE(4638)] = 255999,
  [SMALL_STATE(4639)] = 256032,
  [SMALL_STATE(4640)] = 256065,
  [SMALL_STATE(4641)] = 256104,
  [SMALL_STATE(4642)] = 256163,
  [SMALL_STATE(4643)] = 256196,
  [SMALL_STATE(4644)] = 256229,
  [SMALL_STATE(4645)] = 256272,
  [SMALL_STATE(4646)] = 256309,
  [SMALL_STATE(4647)] = 256374,
  [SMALL_STATE(4648)] = 256407,
  [SMALL_STATE(4649)] = 256440,
  [SMALL_STATE(4650)] = 256499,
  [SMALL_STATE(4651)] = 256542,
  [SMALL_STATE(4652)] = 256607,
  [SMALL_STATE(4653)] = 256666,
  [SMALL_STATE(4654)] = 256699,
  [SMALL_STATE(4655)] = 256758,
  [SMALL_STATE(4656)] = 256791,
  [SMALL_STATE(4657)] = 256824,
  [SMALL_STATE(4658)] = 256883,
  [SMALL_STATE(4659)] = 256942,
  [SMALL_STATE(4660)] = 256985,
  [SMALL_STATE(4661)] = 257044,
  [SMALL_STATE(4662)] = 257103,
  [SMALL_STATE(4663)] = 257152,
  [SMALL_STATE(4664)] = 257201,
  [SMALL_STATE(4665)] = 257238,
  [SMALL_STATE(4666)] = 257281,
  [SMALL_STATE(4667)] = 257314,
  [SMALL_STATE(4668)] = 257357,
  [SMALL_STATE(4669)] = 257416,
  [SMALL_STATE(4670)] = 257455,
  [SMALL_STATE(4671)] = 257488,
  [SMALL_STATE(4672)] = 257547,
  [SMALL_STATE(4673)] = 257606,
  [SMALL_STATE(4674)] = 257645,
  [SMALL_STATE(4675)] = 257684,
  [SMALL_STATE(4676)] = 257743,
  [SMALL_STATE(4677)] = 257802,
  [SMALL_STATE(4678)] = 257861,
  [SMALL_STATE(4679)] = 257920,
  [SMALL_STATE(4680)] = 257979,
  [SMALL_STATE(4681)] = 258038,
  [SMALL_STATE(4682)] = 258097,
  [SMALL_STATE(4683)] = 258156,
  [SMALL_STATE(4684)] = 258195,
  [SMALL_STATE(4685)] = 258234,
  [SMALL_STATE(4686)] = 258277,
  [SMALL_STATE(4687)] = 258336,
  [SMALL_STATE(4688)] = 258375,
  [SMALL_STATE(4689)] = 258408,
  [SMALL_STATE(4690)] = 258447,
  [SMALL_STATE(4691)] = 258506,
  [SMALL_STATE(4692)] = 258565,
  [SMALL_STATE(4693)] = 258604,
  [SMALL_STATE(4694)] = 258663,
  [SMALL_STATE(4695)] = 258722,
  [SMALL_STATE(4696)] = 258759,
  [SMALL_STATE(4697)] = 258818,
  [SMALL_STATE(4698)] = 258861,
  [SMALL_STATE(4699)] = 258900,
  [SMALL_STATE(4700)] = 258939,
  [SMALL_STATE(4701)] = 259004,
  [SMALL_STATE(4702)] = 259037,
  [SMALL_STATE(4703)] = 259070,
  [SMALL_STATE(4704)] = 259129,
  [SMALL_STATE(4705)] = 259166,
  [SMALL_STATE(4706)] = 259205,
  [SMALL_STATE(4707)] = 259238,
  [SMALL_STATE(4708)] = 259297,
  [SMALL_STATE(4709)] = 259330,
  [SMALL_STATE(4710)] = 259363,
  [SMALL_STATE(4711)] = 259406,
  [SMALL_STATE(4712)] = 259445,
  [SMALL_STATE(4713)] = 259478,
  [SMALL_STATE(4714)] = 259537,
  [SMALL_STATE(4715)] = 259596,
  [SMALL_STATE(4716)] = 259655,
  [SMALL_STATE(4717)] = 259690,
  [SMALL_STATE(4718)] = 259727,
  [SMALL_STATE(4719)] = 259786,
  [SMALL_STATE(4720)] = 259829,
  [SMALL_STATE(4721)] = 259888,
  [SMALL_STATE(4722)] = 259953,
  [SMALL_STATE(4723)] = 259990,
  [SMALL_STATE(4724)] = 260033,
  [SMALL_STATE(4725)] = 260068,
  [SMALL_STATE(4726)] = 260111,
  [SMALL_STATE(4727)] = 260144,
  [SMALL_STATE(4728)] = 260177,
  [SMALL_STATE(4729)] = 260236,
  [SMALL_STATE(4730)] = 260295,
  [SMALL_STATE(4731)] = 260354,
  [SMALL_STATE(4732)] = 260413,
  [SMALL_STATE(4733)] = 260472,
  [SMALL_STATE(4734)] = 260511,
  [SMALL_STATE(4735)] = 260554,
  [SMALL_STATE(4736)] = 260587,
  [SMALL_STATE(4737)] = 260620,
  [SMALL_STATE(4738)] = 260679,
  [SMALL_STATE(4739)] = 260738,
  [SMALL_STATE(4740)] = 260771,
  [SMALL_STATE(4741)] = 260830,
  [SMALL_STATE(4742)] = 260895,
  [SMALL_STATE(4743)] = 260960,
  [SMALL_STATE(4744)] = 261025,
  [SMALL_STATE(4745)] = 261084,
  [SMALL_STATE(4746)] = 261121,
  [SMALL_STATE(4747)] = 261164,
  [SMALL_STATE(4748)] = 261207,
  [SMALL_STATE(4749)] = 261246,
  [SMALL_STATE(4750)] = 261305,
  [SMALL_STATE(4751)] = 261344,
  [SMALL_STATE(4752)] = 261403,
  [SMALL_STATE(4753)] = 261462,
  [SMALL_STATE(4754)] = 261495,
  [SMALL_STATE(4755)] = 261534,
  [SMALL_STATE(4756)] = 261567,
  [SMALL_STATE(4757)] = 261626,
  [SMALL_STATE(4758)] = 261691,
  [SMALL_STATE(4759)] = 261734,
  [SMALL_STATE(4760)] = 261767,
  [SMALL_STATE(4761)] = 261806,
  [SMALL_STATE(4762)] = 261865,
  [SMALL_STATE(4763)] = 261898,
  [SMALL_STATE(4764)] = 261957,
  [SMALL_STATE(4765)] = 262016,
  [SMALL_STATE(4766)] = 262075,
  [SMALL_STATE(4767)] = 262108,
  [SMALL_STATE(4768)] = 262145,
  [SMALL_STATE(4769)] = 262204,
  [SMALL_STATE(4770)] = 262237,
  [SMALL_STATE(4771)] = 262270,
  [SMALL_STATE(4772)] = 262309,
  [SMALL_STATE(4773)] = 262342,
  [SMALL_STATE(4774)] = 262401,
  [SMALL_STATE(4775)] = 262440,
  [SMALL_STATE(4776)] = 262479,
  [SMALL_STATE(4777)] = 262518,
  [SMALL_STATE(4778)] = 262557,
  [SMALL_STATE(4779)] = 262606,
  [SMALL_STATE(4780)] = 262665,
  [SMALL_STATE(4781)] = 262698,
  [SMALL_STATE(4782)] = 262737,
  [SMALL_STATE(4783)] = 262770,
  [SMALL_STATE(4784)] = 262809,
  [SMALL_STATE(4785)] = 262848,
  [SMALL_STATE(4786)] = 262881,
  [SMALL_STATE(4787)] = 262914,
  [SMALL_STATE(4788)] = 262953,
  [SMALL_STATE(4789)] = 262986,
  [SMALL_STATE(4790)] = 263019,
  [SMALL_STATE(4791)] = 263052,
  [SMALL_STATE(4792)] = 263111,
  [SMALL_STATE(4793)] = 263158,
  [SMALL_STATE(4794)] = 263207,
  [SMALL_STATE(4795)] = 263244,
  [SMALL_STATE(4796)] = 263283,
  [SMALL_STATE(4797)] = 263316,
  [SMALL_STATE(4798)] = 263365,
  [SMALL_STATE(4799)] = 263424,
  [SMALL_STATE(4800)] = 263483,
  [SMALL_STATE(4801)] = 263542,
  [SMALL_STATE(4802)] = 263575,
  [SMALL_STATE(4803)] = 263614,
  [SMALL_STATE(4804)] = 263679,
  [SMALL_STATE(4805)] = 263712,
  [SMALL_STATE(4806)] = 263745,
  [SMALL_STATE(4807)] = 263784,
  [SMALL_STATE(4808)] = 263817,
  [SMALL_STATE(4809)] = 263850,
  [SMALL_STATE(4810)] = 263909,
  [SMALL_STATE(4811)] = 263942,
  [SMALL_STATE(4812)] = 263975,
  [SMALL_STATE(4813)] = 264008,
  [SMALL_STATE(4814)] = 264067,
  [SMALL_STATE(4815)] = 264126,
  [SMALL_STATE(4816)] = 264185,
  [SMALL_STATE(4817)] = 264224,
  [SMALL_STATE(4818)] = 264283,
  [SMALL_STATE(4819)] = 264316,
  [SMALL_STATE(4820)] = 264375,
  [SMALL_STATE(4821)] = 264408,
  [SMALL_STATE(4822)] = 264441,
  [SMALL_STATE(4823)] = 264480,
  [SMALL_STATE(4824)] = 264539,
  [SMALL_STATE(4825)] = 264576,
  [SMALL_STATE(4826)] = 264609,
  [SMALL_STATE(4827)] = 264668,
  [SMALL_STATE(4828)] = 264701,
  [SMALL_STATE(4829)] = 264740,
  [SMALL_STATE(4830)] = 264783,
  [SMALL_STATE(4831)] = 264842,
  [SMALL_STATE(4832)] = 264875,
  [SMALL_STATE(4833)] = 264934,
  [SMALL_STATE(4834)] = 264967,
  [SMALL_STATE(4835)] = 265006,
  [SMALL_STATE(4836)] = 265045,
  [SMALL_STATE(4837)] = 265104,
  [SMALL_STATE(4838)] = 265137,
  [SMALL_STATE(4839)] = 265176,
  [SMALL_STATE(4840)] = 265235,
  [SMALL_STATE(4841)] = 265268,
  [SMALL_STATE(4842)] = 265327,
  [SMALL_STATE(4843)] = 265360,
  [SMALL_STATE(4844)] = 265419,
  [SMALL_STATE(4845)] = 265452,
  [SMALL_STATE(4846)] = 265511,
  [SMALL_STATE(4847)] = 265544,
  [SMALL_STATE(4848)] = 265577,
  [SMALL_STATE(4849)] = 265636,
  [SMALL_STATE(4850)] = 265669,
  [SMALL_STATE(4851)] = 265728,
  [SMALL_STATE(4852)] = 265787,
  [SMALL_STATE(4853)] = 265820,
  [SMALL_STATE(4854)] = 265853,
  [SMALL_STATE(4855)] = 265912,
  [SMALL_STATE(4856)] = 265949,
  [SMALL_STATE(4857)] = 265982,
  [SMALL_STATE(4858)] = 266041,
  [SMALL_STATE(4859)] = 266074,
  [SMALL_STATE(4860)] = 266107,
  [SMALL_STATE(4861)] = 266156,
  [SMALL_STATE(4862)] = 266189,
  [SMALL_STATE(4863)] = 266222,
  [SMALL_STATE(4864)] = 266281,
  [SMALL_STATE(4865)] = 266313,
  [SMALL_STATE(4866)] = 266345,
  [SMALL_STATE(4867)] = 266377,
  [SMALL_STATE(4868)] = 266409,
  [SMALL_STATE(4869)] = 266441,
  [SMALL_STATE(4870)] = 266473,
  [SMALL_STATE(4871)] = 266505,
  [SMALL_STATE(4872)] = 266537,
  [SMALL_STATE(4873)] = 266569,
  [SMALL_STATE(4874)] = 266601,
  [SMALL_STATE(4875)] = 266633,
  [SMALL_STATE(4876)] = 266665,
  [SMALL_STATE(4877)] = 266697,
  [SMALL_STATE(4878)] = 266729,
  [SMALL_STATE(4879)] = 266761,
  [SMALL_STATE(4880)] = 266793,
  [SMALL_STATE(4881)] = 266825,
  [SMALL_STATE(4882)] = 266857,
  [SMALL_STATE(4883)] = 266889,
  [SMALL_STATE(4884)] = 266921,
  [SMALL_STATE(4885)] = 266953,
  [SMALL_STATE(4886)] = 266985,
  [SMALL_STATE(4887)] = 267017,
  [SMALL_STATE(4888)] = 267049,
  [SMALL_STATE(4889)] = 267081,
  [SMALL_STATE(4890)] = 267113,
  [SMALL_STATE(4891)] = 267145,
  [SMALL_STATE(4892)] = 267207,
  [SMALL_STATE(4893)] = 267239,
  [SMALL_STATE(4894)] = 267271,
  [SMALL_STATE(4895)] = 267303,
  [SMALL_STATE(4896)] = 267335,
  [SMALL_STATE(4897)] = 267373,
  [SMALL_STATE(4898)] = 267411,
  [SMALL_STATE(4899)] = 267443,
  [SMALL_STATE(4900)] = 267475,
  [SMALL_STATE(4901)] = 267507,
  [SMALL_STATE(4902)] = 267545,
  [SMALL_STATE(4903)] = 267577,
  [SMALL_STATE(4904)] = 267609,
  [SMALL_STATE(4905)] = 267645,
  [SMALL_STATE(4906)] = 267677,
  [SMALL_STATE(4907)] = 267715,
  [SMALL_STATE(4908)] = 267747,
  [SMALL_STATE(4909)] = 267779,
  [SMALL_STATE(4910)] = 267811,
  [SMALL_STATE(4911)] = 267857,
  [SMALL_STATE(4912)] = 267889,
  [SMALL_STATE(4913)] = 267921,
  [SMALL_STATE(4914)] = 267953,
  [SMALL_STATE(4915)] = 267985,
  [SMALL_STATE(4916)] = 268017,
  [SMALL_STATE(4917)] = 268049,
  [SMALL_STATE(4918)] = 268081,
  [SMALL_STATE(4919)] = 268143,
  [SMALL_STATE(4920)] = 268181,
  [SMALL_STATE(4921)] = 268213,
  [SMALL_STATE(4922)] = 268251,
  [SMALL_STATE(4923)] = 268283,
  [SMALL_STATE(4924)] = 268315,
  [SMALL_STATE(4925)] = 268377,
  [SMALL_STATE(4926)] = 268409,
  [SMALL_STATE(4927)] = 268441,
  [SMALL_STATE(4928)] = 268473,
  [SMALL_STATE(4929)] = 268519,
  [SMALL_STATE(4930)] = 268551,
  [SMALL_STATE(4931)] = 268583,
  [SMALL_STATE(4932)] = 268615,
  [SMALL_STATE(4933)] = 268647,
  [SMALL_STATE(4934)] = 268679,
  [SMALL_STATE(4935)] = 268741,
  [SMALL_STATE(4936)] = 268773,
  [SMALL_STATE(4937)] = 268805,
  [SMALL_STATE(4938)] = 268837,
  [SMALL_STATE(4939)] = 268869,
  [SMALL_STATE(4940)] = 268915,
  [SMALL_STATE(4941)] = 268947,
  [SMALL_STATE(4942)] = 268985,
  [SMALL_STATE(4943)] = 269021,
  [SMALL_STATE(4944)] = 269053,
  [SMALL_STATE(4945)] = 269085,
  [SMALL_STATE(4946)] = 269117,
  [SMALL_STATE(4947)] = 269149,
  [SMALL_STATE(4948)] = 269181,
  [SMALL_STATE(4949)] = 269213,
  [SMALL_STATE(4950)] = 269245,
  [SMALL_STATE(4951)] = 269277,
  [SMALL_STATE(4952)] = 269323,
  [SMALL_STATE(4953)] = 269355,
  [SMALL_STATE(4954)] = 269387,
  [SMALL_STATE(4955)] = 269419,
  [SMALL_STATE(4956)] = 269451,
  [SMALL_STATE(4957)] = 269483,
  [SMALL_STATE(4958)] = 269515,
  [SMALL_STATE(4959)] = 269547,
  [SMALL_STATE(4960)] = 269579,
  [SMALL_STATE(4961)] = 269611,
  [SMALL_STATE(4962)] = 269643,
  [SMALL_STATE(4963)] = 269675,
  [SMALL_STATE(4964)] = 269707,
  [SMALL_STATE(4965)] = 269739,
  [SMALL_STATE(4966)] = 269771,
  [SMALL_STATE(4967)] = 269803,
  [SMALL_STATE(4968)] = 269865,
  [SMALL_STATE(4969)] = 269897,
  [SMALL_STATE(4970)] = 269929,
  [SMALL_STATE(4971)] = 269961,
  [SMALL_STATE(4972)] = 269997,
  [SMALL_STATE(4973)] = 270059,
  [SMALL_STATE(4974)] = 270091,
  [SMALL_STATE(4975)] = 270153,
  [SMALL_STATE(4976)] = 270189,
  [SMALL_STATE(4977)] = 270221,
  [SMALL_STATE(4978)] = 270253,
  [SMALL_STATE(4979)] = 270285,
  [SMALL_STATE(4980)] = 270317,
  [SMALL_STATE(4981)] = 270349,
  [SMALL_STATE(4982)] = 270381,
  [SMALL_STATE(4983)] = 270413,
  [SMALL_STATE(4984)] = 270445,
  [SMALL_STATE(4985)] = 270477,
  [SMALL_STATE(4986)] = 270509,
  [SMALL_STATE(4987)] = 270551,
  [SMALL_STATE(4988)] = 270583,
  [SMALL_STATE(4989)] = 270615,
  [SMALL_STATE(4990)] = 270647,
  [SMALL_STATE(4991)] = 270679,
  [SMALL_STATE(4992)] = 270711,
  [SMALL_STATE(4993)] = 270749,
  [SMALL_STATE(4994)] = 270795,
  [SMALL_STATE(4995)] = 270827,
  [SMALL_STATE(4996)] = 270869,
  [SMALL_STATE(4997)] = 270905,
  [SMALL_STATE(4998)] = 270939,
  [SMALL_STATE(4999)] = 270981,
  [SMALL_STATE(5000)] = 271043,
  [SMALL_STATE(5001)] = 271085,
  [SMALL_STATE(5002)] = 271127,
  [SMALL_STATE(5003)] = 271159,
  [SMALL_STATE(5004)] = 271191,
  [SMALL_STATE(5005)] = 271223,
  [SMALL_STATE(5006)] = 271255,
  [SMALL_STATE(5007)] = 271287,
  [SMALL_STATE(5008)] = 271333,
  [SMALL_STATE(5009)] = 271365,
  [SMALL_STATE(5010)] = 271397,
  [SMALL_STATE(5011)] = 271429,
  [SMALL_STATE(5012)] = 271461,
  [SMALL_STATE(5013)] = 271493,
  [SMALL_STATE(5014)] = 271525,
  [SMALL_STATE(5015)] = 271557,
  [SMALL_STATE(5016)] = 271619,
  [SMALL_STATE(5017)] = 271651,
  [SMALL_STATE(5018)] = 271683,
  [SMALL_STATE(5019)] = 271715,
  [SMALL_STATE(5020)] = 271747,
  [SMALL_STATE(5021)] = 271779,
  [SMALL_STATE(5022)] = 271815,
  [SMALL_STATE(5023)] = 271847,
  [SMALL_STATE(5024)] = 271879,
  [SMALL_STATE(5025)] = 271911,
  [SMALL_STATE(5026)] = 271943,
  [SMALL_STATE(5027)] = 271975,
  [SMALL_STATE(5028)] = 272007,
  [SMALL_STATE(5029)] = 272039,
  [SMALL_STATE(5030)] = 272071,
  [SMALL_STATE(5031)] = 272103,
  [SMALL_STATE(5032)] = 272141,
  [SMALL_STATE(5033)] = 272173,
  [SMALL_STATE(5034)] = 272211,
  [SMALL_STATE(5035)] = 272248,
  [SMALL_STATE(5036)] = 272279,
  [SMALL_STATE(5037)] = 272314,
  [SMALL_STATE(5038)] = 272345,
  [SMALL_STATE(5039)] = 272380,
  [SMALL_STATE(5040)] = 272411,
  [SMALL_STATE(5041)] = 272456,
  [SMALL_STATE(5042)] = 272491,
  [SMALL_STATE(5043)] = 272522,
  [SMALL_STATE(5044)] = 272553,
  [SMALL_STATE(5045)] = 272584,
  [SMALL_STATE(5046)] = 272619,
  [SMALL_STATE(5047)] = 272650,
  [SMALL_STATE(5048)] = 272681,
  [SMALL_STATE(5049)] = 272718,
  [SMALL_STATE(5050)] = 272749,
  [SMALL_STATE(5051)] = 272780,
  [SMALL_STATE(5052)] = 272825,
  [SMALL_STATE(5053)] = 272856,
  [SMALL_STATE(5054)] = 272887,
  [SMALL_STATE(5055)] = 272932,
  [SMALL_STATE(5056)] = 272969,
  [SMALL_STATE(5057)] = 273008,
  [SMALL_STATE(5058)] = 273045,
  [SMALL_STATE(5059)] = 273082,
  [SMALL_STATE(5060)] = 273119,
  [SMALL_STATE(5061)] = 273150,
  [SMALL_STATE(5062)] = 273181,
  [SMALL_STATE(5063)] = 273218,
  [SMALL_STATE(5064)] = 273249,
  [SMALL_STATE(5065)] = 273284,
  [SMALL_STATE(5066)] = 273321,
  [SMALL_STATE(5067)] = 273352,
  [SMALL_STATE(5068)] = 273383,
  [SMALL_STATE(5069)] = 273414,
  [SMALL_STATE(5070)] = 273445,
  [SMALL_STATE(5071)] = 273490,
  [SMALL_STATE(5072)] = 273521,
  [SMALL_STATE(5073)] = 273552,
  [SMALL_STATE(5074)] = 273583,
  [SMALL_STATE(5075)] = 273614,
  [SMALL_STATE(5076)] = 273645,
  [SMALL_STATE(5077)] = 273676,
  [SMALL_STATE(5078)] = 273707,
  [SMALL_STATE(5079)] = 273738,
  [SMALL_STATE(5080)] = 273783,
  [SMALL_STATE(5081)] = 273814,
  [SMALL_STATE(5082)] = 273849,
  [SMALL_STATE(5083)] = 273884,
  [SMALL_STATE(5084)] = 273915,
  [SMALL_STATE(5085)] = 273946,
  [SMALL_STATE(5086)] = 273977,
  [SMALL_STATE(5087)] = 274008,
  [SMALL_STATE(5088)] = 274039,
  [SMALL_STATE(5089)] = 274070,
  [SMALL_STATE(5090)] = 274101,
  [SMALL_STATE(5091)] = 274132,
  [SMALL_STATE(5092)] = 274163,
  [SMALL_STATE(5093)] = 274194,
  [SMALL_STATE(5094)] = 274225,
  [SMALL_STATE(5095)] = 274256,
  [SMALL_STATE(5096)] = 274291,
  [SMALL_STATE(5097)] = 274322,
  [SMALL_STATE(5098)] = 274357,
  [SMALL_STATE(5099)] = 274388,
  [SMALL_STATE(5100)] = 274419,
  [SMALL_STATE(5101)] = 274450,
  [SMALL_STATE(5102)] = 274481,
  [SMALL_STATE(5103)] = 274516,
  [SMALL_STATE(5104)] = 274547,
  [SMALL_STATE(5105)] = 274578,
  [SMALL_STATE(5106)] = 274609,
  [SMALL_STATE(5107)] = 274640,
  [SMALL_STATE(5108)] = 274671,
  [SMALL_STATE(5109)] = 274716,
  [SMALL_STATE(5110)] = 274753,
  [SMALL_STATE(5111)] = 274790,
  [SMALL_STATE(5112)] = 274821,
  [SMALL_STATE(5113)] = 274852,
  [SMALL_STATE(5114)] = 274883,
  [SMALL_STATE(5115)] = 274914,
  [SMALL_STATE(5116)] = 274949,
  [SMALL_STATE(5117)] = 274980,
  [SMALL_STATE(5118)] = 275027,
  [SMALL_STATE(5119)] = 275074,
  [SMALL_STATE(5120)] = 275105,
  [SMALL_STATE(5121)] = 275136,
  [SMALL_STATE(5122)] = 275168,
  [SMALL_STATE(5123)] = 275210,
  [SMALL_STATE(5124)] = 275240,
  [SMALL_STATE(5125)] = 275274,
  [SMALL_STATE(5126)] = 275306,
  [SMALL_STATE(5127)] = 275338,
  [SMALL_STATE(5128)] = 275372,
  [SMALL_STATE(5129)] = 275408,
  [SMALL_STATE(5130)] = 275444,
  [SMALL_STATE(5131)] = 275480,
  [SMALL_STATE(5132)] = 275516,
  [SMALL_STATE(5133)] = 275548,
  [SMALL_STATE(5134)] = 275584,
  [SMALL_STATE(5135)] = 275620,
  [SMALL_STATE(5136)] = 275656,
  [SMALL_STATE(5137)] = 275692,
  [SMALL_STATE(5138)] = 275728,
  [SMALL_STATE(5139)] = 275764,
  [SMALL_STATE(5140)] = 275796,
  [SMALL_STATE(5141)] = 275830,
  [SMALL_STATE(5142)] = 275862,
  [SMALL_STATE(5143)] = 275918,
  [SMALL_STATE(5144)] = 275950,
  [SMALL_STATE(5145)] = 275982,
  [SMALL_STATE(5146)] = 276014,
  [SMALL_STATE(5147)] = 276046,
  [SMALL_STATE(5148)] = 276082,
  [SMALL_STATE(5149)] = 276118,
  [SMALL_STATE(5150)] = 276148,
  [SMALL_STATE(5151)] = 276178,
  [SMALL_STATE(5152)] = 276208,
  [SMALL_STATE(5153)] = 276238,
  [SMALL_STATE(5154)] = 276268,
  [SMALL_STATE(5155)] = 276298,
  [SMALL_STATE(5156)] = 276328,
  [SMALL_STATE(5157)] = 276358,
  [SMALL_STATE(5158)] = 276388,
  [SMALL_STATE(5159)] = 276418,
  [SMALL_STATE(5160)] = 276448,
  [SMALL_STATE(5161)] = 276478,
  [SMALL_STATE(5162)] = 276508,
  [SMALL_STATE(5163)] = 276540,
  [SMALL_STATE(5164)] = 276582,
  [SMALL_STATE(5165)] = 276614,
  [SMALL_STATE(5166)] = 276644,
  [SMALL_STATE(5167)] = 276676,
  [SMALL_STATE(5168)] = 276718,
  [SMALL_STATE(5169)] = 276774,
  [SMALL_STATE(5170)] = 276806,
  [SMALL_STATE(5171)] = 276840,
  [SMALL_STATE(5172)] = 276870,
  [SMALL_STATE(5173)] = 276902,
  [SMALL_STATE(5174)] = 276936,
  [SMALL_STATE(5175)] = 276970,
  [SMALL_STATE(5176)] = 277006,
  [SMALL_STATE(5177)] = 277046,
  [SMALL_STATE(5178)] = 277078,
  [SMALL_STATE(5179)] = 277112,
  [SMALL_STATE(5180)] = 277146,
  [SMALL_STATE(5181)] = 277182,
  [SMALL_STATE(5182)] = 277222,
  [SMALL_STATE(5183)] = 277256,
  [SMALL_STATE(5184)] = 277288,
  [SMALL_STATE(5185)] = 277328,
  [SMALL_STATE(5186)] = 277384,
  [SMALL_STATE(5187)] = 277440,
  [SMALL_STATE(5188)] = 277496,
  [SMALL_STATE(5189)] = 277530,
  [SMALL_STATE(5190)] = 277572,
  [SMALL_STATE(5191)] = 277628,
  [SMALL_STATE(5192)] = 277684,
  [SMALL_STATE(5193)] = 277740,
  [SMALL_STATE(5194)] = 277772,
  [SMALL_STATE(5195)] = 277828,
  [SMALL_STATE(5196)] = 277878,
  [SMALL_STATE(5197)] = 277934,
  [SMALL_STATE(5198)] = 277990,
  [SMALL_STATE(5199)] = 278046,
  [SMALL_STATE(5200)] = 278076,
  [SMALL_STATE(5201)] = 278106,
  [SMALL_STATE(5202)] = 278138,
  [SMALL_STATE(5203)] = 278194,
  [SMALL_STATE(5204)] = 278250,
  [SMALL_STATE(5205)] = 278280,
  [SMALL_STATE(5206)] = 278322,
  [SMALL_STATE(5207)] = 278358,
  [SMALL_STATE(5208)] = 278394,
  [SMALL_STATE(5209)] = 278430,
  [SMALL_STATE(5210)] = 278466,
  [SMALL_STATE(5211)] = 278502,
  [SMALL_STATE(5212)] = 278570,
  [SMALL_STATE(5213)] = 278632,
  [SMALL_STATE(5214)] = 278668,
  [SMALL_STATE(5215)] = 278724,
  [SMALL_STATE(5216)] = 278756,
  [SMALL_STATE(5217)] = 278790,
  [SMALL_STATE(5218)] = 278846,
  [SMALL_STATE(5219)] = 278902,
  [SMALL_STATE(5220)] = 278958,
  [SMALL_STATE(5221)] = 279014,
  [SMALL_STATE(5222)] = 279048,
  [SMALL_STATE(5223)] = 279104,
  [SMALL_STATE(5224)] = 279160,
  [SMALL_STATE(5225)] = 279194,
  [SMALL_STATE(5226)] = 279250,
  [SMALL_STATE(5227)] = 279306,
  [SMALL_STATE(5228)] = 279338,
  [SMALL_STATE(5229)] = 279394,
  [SMALL_STATE(5230)] = 279450,
  [SMALL_STATE(5231)] = 279506,
  [SMALL_STATE(5232)] = 279562,
  [SMALL_STATE(5233)] = 279624,
  [SMALL_STATE(5234)] = 279680,
  [SMALL_STATE(5235)] = 279736,
  [SMALL_STATE(5236)] = 279792,
  [SMALL_STATE(5237)] = 279848,
  [SMALL_STATE(5238)] = 279904,
  [SMALL_STATE(5239)] = 279960,
  [SMALL_STATE(5240)] = 280016,
  [SMALL_STATE(5241)] = 280050,
  [SMALL_STATE(5242)] = 280106,
  [SMALL_STATE(5243)] = 280162,
  [SMALL_STATE(5244)] = 280218,
  [SMALL_STATE(5245)] = 280274,
  [SMALL_STATE(5246)] = 280306,
  [SMALL_STATE(5247)] = 280362,
  [SMALL_STATE(5248)] = 280394,
  [SMALL_STATE(5249)] = 280436,
  [SMALL_STATE(5250)] = 280490,
  [SMALL_STATE(5251)] = 280522,
  [SMALL_STATE(5252)] = 280551,
  [SMALL_STATE(5253)] = 280580,
  [SMALL_STATE(5254)] = 280613,
  [SMALL_STATE(5255)] = 280642,
  [SMALL_STATE(5256)] = 280671,
  [SMALL_STATE(5257)] = 280704,
  [SMALL_STATE(5258)] = 280733,
  [SMALL_STATE(5259)] = 280766,
  [SMALL_STATE(5260)] = 280795,
  [SMALL_STATE(5261)] = 280854,
  [SMALL_STATE(5262)] = 280883,
  [SMALL_STATE(5263)] = 280912,
  [SMALL_STATE(5264)] = 280941,
  [SMALL_STATE(5265)] = 280970,
  [SMALL_STATE(5266)] = 280999,
  [SMALL_STATE(5267)] = 281046,
  [SMALL_STATE(5268)] = 281079,
  [SMALL_STATE(5269)] = 281108,
  [SMALL_STATE(5270)] = 281137,
  [SMALL_STATE(5271)] = 281172,
  [SMALL_STATE(5272)] = 281207,
  [SMALL_STATE(5273)] = 281236,
  [SMALL_STATE(5274)] = 281265,
  [SMALL_STATE(5275)] = 281298,
  [SMALL_STATE(5276)] = 281327,
  [SMALL_STATE(5277)] = 281360,
  [SMALL_STATE(5278)] = 281389,
  [SMALL_STATE(5279)] = 281422,
  [SMALL_STATE(5280)] = 281451,
  [SMALL_STATE(5281)] = 281480,
  [SMALL_STATE(5282)] = 281509,
  [SMALL_STATE(5283)] = 281538,
  [SMALL_STATE(5284)] = 281567,
  [SMALL_STATE(5285)] = 281596,
  [SMALL_STATE(5286)] = 281625,
  [SMALL_STATE(5287)] = 281660,
  [SMALL_STATE(5288)] = 281695,
  [SMALL_STATE(5289)] = 281724,
  [SMALL_STATE(5290)] = 281771,
  [SMALL_STATE(5291)] = 281804,
  [SMALL_STATE(5292)] = 281833,
  [SMALL_STATE(5293)] = 281866,
  [SMALL_STATE(5294)] = 281901,
  [SMALL_STATE(5295)] = 281930,
  [SMALL_STATE(5296)] = 281959,
  [SMALL_STATE(5297)] = 281988,
  [SMALL_STATE(5298)] = 282023,
  [SMALL_STATE(5299)] = 282052,
  [SMALL_STATE(5300)] = 282081,
  [SMALL_STATE(5301)] = 282114,
  [SMALL_STATE(5302)] = 282147,
  [SMALL_STATE(5303)] = 282176,
  [SMALL_STATE(5304)] = 282209,
  [SMALL_STATE(5305)] = 282238,
  [SMALL_STATE(5306)] = 282273,
  [SMALL_STATE(5307)] = 282308,
  [SMALL_STATE(5308)] = 282337,
  [SMALL_STATE(5309)] = 282366,
  [SMALL_STATE(5310)] = 282395,
  [SMALL_STATE(5311)] = 282424,
  [SMALL_STATE(5312)] = 282453,
  [SMALL_STATE(5313)] = 282488,
  [SMALL_STATE(5314)] = 282517,
  [SMALL_STATE(5315)] = 282546,
  [SMALL_STATE(5316)] = 282593,
  [SMALL_STATE(5317)] = 282622,
  [SMALL_STATE(5318)] = 282669,
  [SMALL_STATE(5319)] = 282704,
  [SMALL_STATE(5320)] = 282733,
  [SMALL_STATE(5321)] = 282762,
  [SMALL_STATE(5322)] = 282791,
  [SMALL_STATE(5323)] = 282826,
  [SMALL_STATE(5324)] = 282859,
  [SMALL_STATE(5325)] = 282888,
  [SMALL_STATE(5326)] = 282917,
  [SMALL_STATE(5327)] = 282946,
  [SMALL_STATE(5328)] = 282975,
  [SMALL_STATE(5329)] = 283010,
  [SMALL_STATE(5330)] = 283039,
  [SMALL_STATE(5331)] = 283068,
  [SMALL_STATE(5332)] = 283101,
  [SMALL_STATE(5333)] = 283130,
  [SMALL_STATE(5334)] = 283159,
  [SMALL_STATE(5335)] = 283188,
  [SMALL_STATE(5336)] = 283217,
  [SMALL_STATE(5337)] = 283246,
  [SMALL_STATE(5338)] = 283275,
  [SMALL_STATE(5339)] = 283304,
  [SMALL_STATE(5340)] = 283333,
  [SMALL_STATE(5341)] = 283362,
  [SMALL_STATE(5342)] = 283391,
  [SMALL_STATE(5343)] = 283420,
  [SMALL_STATE(5344)] = 283449,
  [SMALL_STATE(5345)] = 283478,
  [SMALL_STATE(5346)] = 283507,
  [SMALL_STATE(5347)] = 283536,
  [SMALL_STATE(5348)] = 283569,
  [SMALL_STATE(5349)] = 283602,
  [SMALL_STATE(5350)] = 283635,
  [SMALL_STATE(5351)] = 283678,
  [SMALL_STATE(5352)] = 283711,
  [SMALL_STATE(5353)] = 283744,
  [SMALL_STATE(5354)] = 283773,
  [SMALL_STATE(5355)] = 283802,
  [SMALL_STATE(5356)] = 283831,
  [SMALL_STATE(5357)] = 283878,
  [SMALL_STATE(5358)] = 283911,
  [SMALL_STATE(5359)] = 283958,
  [SMALL_STATE(5360)] = 283991,
  [SMALL_STATE(5361)] = 284026,
  [SMALL_STATE(5362)] = 284073,
  [SMALL_STATE(5363)] = 284114,
  [SMALL_STATE(5364)] = 284155,
  [SMALL_STATE(5365)] = 284184,
  [SMALL_STATE(5366)] = 284217,
  [SMALL_STATE(5367)] = 284256,
  [SMALL_STATE(5368)] = 284295,
  [SMALL_STATE(5369)] = 284328,
  [SMALL_STATE(5370)] = 284359,
  [SMALL_STATE(5371)] = 284398,
  [SMALL_STATE(5372)] = 284431,
  [SMALL_STATE(5373)] = 284470,
  [SMALL_STATE(5374)] = 284503,
  [SMALL_STATE(5375)] = 284536,
  [SMALL_STATE(5376)] = 284575,
  [SMALL_STATE(5377)] = 284608,
  [SMALL_STATE(5378)] = 284641,
  [SMALL_STATE(5379)] = 284674,
  [SMALL_STATE(5380)] = 284707,
  [SMALL_STATE(5381)] = 284736,
  [SMALL_STATE(5382)] = 284777,
  [SMALL_STATE(5383)] = 284820,
  [SMALL_STATE(5384)] = 284861,
  [SMALL_STATE(5385)] = 284890,
  [SMALL_STATE(5386)] = 284925,
  [SMALL_STATE(5387)] = 284954,
  [SMALL_STATE(5388)] = 284983,
  [SMALL_STATE(5389)] = 285016,
  [SMALL_STATE(5390)] = 285045,
  [SMALL_STATE(5391)] = 285080,
  [SMALL_STATE(5392)] = 285109,
  [SMALL_STATE(5393)] = 285144,
  [SMALL_STATE(5394)] = 285179,
  [SMALL_STATE(5395)] = 285226,
  [SMALL_STATE(5396)] = 285255,
  [SMALL_STATE(5397)] = 285284,
  [SMALL_STATE(5398)] = 285313,
  [SMALL_STATE(5399)] = 285348,
  [SMALL_STATE(5400)] = 285377,
  [SMALL_STATE(5401)] = 285406,
  [SMALL_STATE(5402)] = 285435,
  [SMALL_STATE(5403)] = 285464,
  [SMALL_STATE(5404)] = 285493,
  [SMALL_STATE(5405)] = 285521,
  [SMALL_STATE(5406)] = 285561,
  [SMALL_STATE(5407)] = 285605,
  [SMALL_STATE(5408)] = 285639,
  [SMALL_STATE(5409)] = 285671,
  [SMALL_STATE(5410)] = 285699,
  [SMALL_STATE(5411)] = 285727,
  [SMALL_STATE(5412)] = 285759,
  [SMALL_STATE(5413)] = 285791,
  [SMALL_STATE(5414)] = 285819,
  [SMALL_STATE(5415)] = 285847,
  [SMALL_STATE(5416)] = 285879,
  [SMALL_STATE(5417)] = 285907,
  [SMALL_STATE(5418)] = 285939,
  [SMALL_STATE(5419)] = 285971,
  [SMALL_STATE(5420)] = 286001,
  [SMALL_STATE(5421)] = 286033,
  [SMALL_STATE(5422)] = 286067,
  [SMALL_STATE(5423)] = 286095,
  [SMALL_STATE(5424)] = 286135,
  [SMALL_STATE(5425)] = 286167,
  [SMALL_STATE(5426)] = 286211,
  [SMALL_STATE(5427)] = 286243,
  [SMALL_STATE(5428)] = 286275,
  [SMALL_STATE(5429)] = 286303,
  [SMALL_STATE(5430)] = 286359,
  [SMALL_STATE(5431)] = 286387,
  [SMALL_STATE(5432)] = 286415,
  [SMALL_STATE(5433)] = 286443,
  [SMALL_STATE(5434)] = 286471,
  [SMALL_STATE(5435)] = 286499,
  [SMALL_STATE(5436)] = 286527,
  [SMALL_STATE(5437)] = 286569,
  [SMALL_STATE(5438)] = 286601,
  [SMALL_STATE(5439)] = 286629,
  [SMALL_STATE(5440)] = 286657,
  [SMALL_STATE(5441)] = 286685,
  [SMALL_STATE(5442)] = 286713,
  [SMALL_STATE(5443)] = 286741,
  [SMALL_STATE(5444)] = 286769,
  [SMALL_STATE(5445)] = 286797,
  [SMALL_STATE(5446)] = 286825,
  [SMALL_STATE(5447)] = 286853,
  [SMALL_STATE(5448)] = 286885,
  [SMALL_STATE(5449)] = 286913,
  [SMALL_STATE(5450)] = 286941,
  [SMALL_STATE(5451)] = 286969,
  [SMALL_STATE(5452)] = 286997,
  [SMALL_STATE(5453)] = 287025,
  [SMALL_STATE(5454)] = 287053,
  [SMALL_STATE(5455)] = 287081,
  [SMALL_STATE(5456)] = 287109,
  [SMALL_STATE(5457)] = 287137,
  [SMALL_STATE(5458)] = 287165,
  [SMALL_STATE(5459)] = 287221,
  [SMALL_STATE(5460)] = 287249,
  [SMALL_STATE(5461)] = 287311,
  [SMALL_STATE(5462)] = 287339,
  [SMALL_STATE(5463)] = 287371,
  [SMALL_STATE(5464)] = 287399,
  [SMALL_STATE(5465)] = 287427,
  [SMALL_STATE(5466)] = 287469,
  [SMALL_STATE(5467)] = 287501,
  [SMALL_STATE(5468)] = 287529,
  [SMALL_STATE(5469)] = 287557,
  [SMALL_STATE(5470)] = 287589,
  [SMALL_STATE(5471)] = 287617,
  [SMALL_STATE(5472)] = 287651,
  [SMALL_STATE(5473)] = 287679,
  [SMALL_STATE(5474)] = 287707,
  [SMALL_STATE(5475)] = 287739,
  [SMALL_STATE(5476)] = 287779,
  [SMALL_STATE(5477)] = 287811,
  [SMALL_STATE(5478)] = 287839,
  [SMALL_STATE(5479)] = 287867,
  [SMALL_STATE(5480)] = 287899,
  [SMALL_STATE(5481)] = 287927,
  [SMALL_STATE(5482)] = 287955,
  [SMALL_STATE(5483)] = 287983,
  [SMALL_STATE(5484)] = 288011,
  [SMALL_STATE(5485)] = 288039,
  [SMALL_STATE(5486)] = 288067,
  [SMALL_STATE(5487)] = 288095,
  [SMALL_STATE(5488)] = 288123,
  [SMALL_STATE(5489)] = 288151,
  [SMALL_STATE(5490)] = 288179,
  [SMALL_STATE(5491)] = 288207,
  [SMALL_STATE(5492)] = 288235,
  [SMALL_STATE(5493)] = 288263,
  [SMALL_STATE(5494)] = 288291,
  [SMALL_STATE(5495)] = 288323,
  [SMALL_STATE(5496)] = 288351,
  [SMALL_STATE(5497)] = 288383,
  [SMALL_STATE(5498)] = 288411,
  [SMALL_STATE(5499)] = 288443,
  [SMALL_STATE(5500)] = 288474,
  [SMALL_STATE(5501)] = 288505,
  [SMALL_STATE(5502)] = 288532,
  [SMALL_STATE(5503)] = 288559,
  [SMALL_STATE(5504)] = 288594,
  [SMALL_STATE(5505)] = 288633,
  [SMALL_STATE(5506)] = 288672,
  [SMALL_STATE(5507)] = 288699,
  [SMALL_STATE(5508)] = 288726,
  [SMALL_STATE(5509)] = 288753,
  [SMALL_STATE(5510)] = 288780,
  [SMALL_STATE(5511)] = 288807,
  [SMALL_STATE(5512)] = 288834,
  [SMALL_STATE(5513)] = 288861,
  [SMALL_STATE(5514)] = 288888,
  [SMALL_STATE(5515)] = 288923,
  [SMALL_STATE(5516)] = 288962,
  [SMALL_STATE(5517)] = 289001,
  [SMALL_STATE(5518)] = 289040,
  [SMALL_STATE(5519)] = 289073,
  [SMALL_STATE(5520)] = 289104,
  [SMALL_STATE(5521)] = 289135,
  [SMALL_STATE(5522)] = 289162,
  [SMALL_STATE(5523)] = 289189,
  [SMALL_STATE(5524)] = 289216,
  [SMALL_STATE(5525)] = 289243,
  [SMALL_STATE(5526)] = 289270,
  [SMALL_STATE(5527)] = 289297,
  [SMALL_STATE(5528)] = 289324,
  [SMALL_STATE(5529)] = 289355,
  [SMALL_STATE(5530)] = 289382,
  [SMALL_STATE(5531)] = 289409,
  [SMALL_STATE(5532)] = 289448,
  [SMALL_STATE(5533)] = 289479,
  [SMALL_STATE(5534)] = 289516,
  [SMALL_STATE(5535)] = 289543,
  [SMALL_STATE(5536)] = 289580,
  [SMALL_STATE(5537)] = 289607,
  [SMALL_STATE(5538)] = 289634,
  [SMALL_STATE(5539)] = 289673,
  [SMALL_STATE(5540)] = 289700,
  [SMALL_STATE(5541)] = 289739,
  [SMALL_STATE(5542)] = 289766,
  [SMALL_STATE(5543)] = 289793,
  [SMALL_STATE(5544)] = 289820,
  [SMALL_STATE(5545)] = 289851,
  [SMALL_STATE(5546)] = 289890,
  [SMALL_STATE(5547)] = 289929,
  [SMALL_STATE(5548)] = 289955,
  [SMALL_STATE(5549)] = 289987,
  [SMALL_STATE(5550)] = 290013,
  [SMALL_STATE(5551)] = 290051,
  [SMALL_STATE(5552)] = 290089,
  [SMALL_STATE(5553)] = 290115,
  [SMALL_STATE(5554)] = 290147,
  [SMALL_STATE(5555)] = 290185,
  [SMALL_STATE(5556)] = 290211,
  [SMALL_STATE(5557)] = 290237,
  [SMALL_STATE(5558)] = 290275,
  [SMALL_STATE(5559)] = 290313,
  [SMALL_STATE(5560)] = 290351,
  [SMALL_STATE(5561)] = 290377,
  [SMALL_STATE(5562)] = 290409,
  [SMALL_STATE(5563)] = 290441,
  [SMALL_STATE(5564)] = 290467,
  [SMALL_STATE(5565)] = 290493,
  [SMALL_STATE(5566)] = 290519,
  [SMALL_STATE(5567)] = 290545,
  [SMALL_STATE(5568)] = 290571,
  [SMALL_STATE(5569)] = 290597,
  [SMALL_STATE(5570)] = 290635,
  [SMALL_STATE(5571)] = 290661,
  [SMALL_STATE(5572)] = 290687,
  [SMALL_STATE(5573)] = 290713,
  [SMALL_STATE(5574)] = 290739,
  [SMALL_STATE(5575)] = 290765,
  [SMALL_STATE(5576)] = 290791,
  [SMALL_STATE(5577)] = 290817,
  [SMALL_STATE(5578)] = 290849,
  [SMALL_STATE(5579)] = 290875,
  [SMALL_STATE(5580)] = 290901,
  [SMALL_STATE(5581)] = 290933,
  [SMALL_STATE(5582)] = 290959,
  [SMALL_STATE(5583)] = 290985,
  [SMALL_STATE(5584)] = 291017,
  [SMALL_STATE(5585)] = 291049,
  [SMALL_STATE(5586)] = 291075,
  [SMALL_STATE(5587)] = 291101,
  [SMALL_STATE(5588)] = 291127,
  [SMALL_STATE(5589)] = 291153,
  [SMALL_STATE(5590)] = 291179,
  [SMALL_STATE(5591)] = 291205,
  [SMALL_STATE(5592)] = 291231,
  [SMALL_STATE(5593)] = 291257,
  [SMALL_STATE(5594)] = 291295,
  [SMALL_STATE(5595)] = 291321,
  [SMALL_STATE(5596)] = 291347,
  [SMALL_STATE(5597)] = 291373,
  [SMALL_STATE(5598)] = 291399,
  [SMALL_STATE(5599)] = 291425,
  [SMALL_STATE(5600)] = 291451,
  [SMALL_STATE(5601)] = 291477,
  [SMALL_STATE(5602)] = 291503,
  [SMALL_STATE(5603)] = 291535,
  [SMALL_STATE(5604)] = 291561,
  [SMALL_STATE(5605)] = 291587,
  [SMALL_STATE(5606)] = 291619,
  [SMALL_STATE(5607)] = 291651,
  [SMALL_STATE(5608)] = 291677,
  [SMALL_STATE(5609)] = 291709,
  [SMALL_STATE(5610)] = 291745,
  [SMALL_STATE(5611)] = 291771,
  [SMALL_STATE(5612)] = 291803,
  [SMALL_STATE(5613)] = 291829,
  [SMALL_STATE(5614)] = 291855,
  [SMALL_STATE(5615)] = 291893,
  [SMALL_STATE(5616)] = 291919,
  [SMALL_STATE(5617)] = 291951,
  [SMALL_STATE(5618)] = 291976,
  [SMALL_STATE(5619)] = 292001,
  [SMALL_STATE(5620)] = 292026,
  [SMALL_STATE(5621)] = 292051,
  [SMALL_STATE(5622)] = 292076,
  [SMALL_STATE(5623)] = 292101,
  [SMALL_STATE(5624)] = 292126,
  [SMALL_STATE(5625)] = 292151,
  [SMALL_STATE(5626)] = 292176,
  [SMALL_STATE(5627)] = 292201,
  [SMALL_STATE(5628)] = 292226,
  [SMALL_STATE(5629)] = 292251,
  [SMALL_STATE(5630)] = 292276,
  [SMALL_STATE(5631)] = 292301,
  [SMALL_STATE(5632)] = 292326,
  [SMALL_STATE(5633)] = 292355,
  [SMALL_STATE(5634)] = 292380,
  [SMALL_STATE(5635)] = 292405,
  [SMALL_STATE(5636)] = 292430,
  [SMALL_STATE(5637)] = 292455,
  [SMALL_STATE(5638)] = 292480,
  [SMALL_STATE(5639)] = 292505,
  [SMALL_STATE(5640)] = 292530,
  [SMALL_STATE(5641)] = 292555,
  [SMALL_STATE(5642)] = 292580,
  [SMALL_STATE(5643)] = 292605,
  [SMALL_STATE(5644)] = 292634,
  [SMALL_STATE(5645)] = 292659,
  [SMALL_STATE(5646)] = 292684,
  [SMALL_STATE(5647)] = 292709,
  [SMALL_STATE(5648)] = 292734,
  [SMALL_STATE(5649)] = 292759,
  [SMALL_STATE(5650)] = 292784,
  [SMALL_STATE(5651)] = 292809,
  [SMALL_STATE(5652)] = 292838,
  [SMALL_STATE(5653)] = 292863,
  [SMALL_STATE(5654)] = 292888,
  [SMALL_STATE(5655)] = 292913,
  [SMALL_STATE(5656)] = 292938,
  [SMALL_STATE(5657)] = 292963,
  [SMALL_STATE(5658)] = 292992,
  [SMALL_STATE(5659)] = 293017,
  [SMALL_STATE(5660)] = 293042,
  [SMALL_STATE(5661)] = 293067,
  [SMALL_STATE(5662)] = 293092,
  [SMALL_STATE(5663)] = 293117,
  [SMALL_STATE(5664)] = 293142,
  [SMALL_STATE(5665)] = 293167,
  [SMALL_STATE(5666)] = 293192,
  [SMALL_STATE(5667)] = 293217,
  [SMALL_STATE(5668)] = 293242,
  [SMALL_STATE(5669)] = 293267,
  [SMALL_STATE(5670)] = 293292,
  [SMALL_STATE(5671)] = 293317,
  [SMALL_STATE(5672)] = 293342,
  [SMALL_STATE(5673)] = 293367,
  [SMALL_STATE(5674)] = 293392,
  [SMALL_STATE(5675)] = 293421,
  [SMALL_STATE(5676)] = 293446,
  [SMALL_STATE(5677)] = 293471,
  [SMALL_STATE(5678)] = 293496,
  [SMALL_STATE(5679)] = 293521,
  [SMALL_STATE(5680)] = 293546,
  [SMALL_STATE(5681)] = 293571,
  [SMALL_STATE(5682)] = 293596,
  [SMALL_STATE(5683)] = 293621,
  [SMALL_STATE(5684)] = 293646,
  [SMALL_STATE(5685)] = 293671,
  [SMALL_STATE(5686)] = 293696,
  [SMALL_STATE(5687)] = 293721,
  [SMALL_STATE(5688)] = 293746,
  [SMALL_STATE(5689)] = 293771,
  [SMALL_STATE(5690)] = 293796,
  [SMALL_STATE(5691)] = 293821,
  [SMALL_STATE(5692)] = 293846,
  [SMALL_STATE(5693)] = 293871,
  [SMALL_STATE(5694)] = 293896,
  [SMALL_STATE(5695)] = 293921,
  [SMALL_STATE(5696)] = 293946,
  [SMALL_STATE(5697)] = 293971,
  [SMALL_STATE(5698)] = 294009,
  [SMALL_STATE(5699)] = 294041,
  [SMALL_STATE(5700)] = 294073,
  [SMALL_STATE(5701)] = 294109,
  [SMALL_STATE(5702)] = 294133,
  [SMALL_STATE(5703)] = 294174,
  [SMALL_STATE(5704)] = 294215,
  [SMALL_STATE(5705)] = 294256,
  [SMALL_STATE(5706)] = 294297,
  [SMALL_STATE(5707)] = 294338,
  [SMALL_STATE(5708)] = 294379,
  [SMALL_STATE(5709)] = 294420,
  [SMALL_STATE(5710)] = 294461,
  [SMALL_STATE(5711)] = 294502,
  [SMALL_STATE(5712)] = 294543,
  [SMALL_STATE(5713)] = 294584,
  [SMALL_STATE(5714)] = 294625,
  [SMALL_STATE(5715)] = 294666,
  [SMALL_STATE(5716)] = 294707,
  [SMALL_STATE(5717)] = 294748,
  [SMALL_STATE(5718)] = 294789,
  [SMALL_STATE(5719)] = 294830,
  [SMALL_STATE(5720)] = 294871,
  [SMALL_STATE(5721)] = 294912,
  [SMALL_STATE(5722)] = 294953,
  [SMALL_STATE(5723)] = 294994,
  [SMALL_STATE(5724)] = 295035,
  [SMALL_STATE(5725)] = 295076,
  [SMALL_STATE(5726)] = 295117,
  [SMALL_STATE(5727)] = 295158,
  [SMALL_STATE(5728)] = 295199,
  [SMALL_STATE(5729)] = 295240,
  [SMALL_STATE(5730)] = 295281,
  [SMALL_STATE(5731)] = 295322,
  [SMALL_STATE(5732)] = 295363,
  [SMALL_STATE(5733)] = 295404,
  [SMALL_STATE(5734)] = 295445,
  [SMALL_STATE(5735)] = 295486,
  [SMALL_STATE(5736)] = 295527,
  [SMALL_STATE(5737)] = 295568,
  [SMALL_STATE(5738)] = 295609,
  [SMALL_STATE(5739)] = 295632,
  [SMALL_STATE(5740)] = 295673,
  [SMALL_STATE(5741)] = 295714,
  [SMALL_STATE(5742)] = 295755,
  [SMALL_STATE(5743)] = 295796,
  [SMALL_STATE(5744)] = 295837,
  [SMALL_STATE(5745)] = 295860,
  [SMALL_STATE(5746)] = 295901,
  [SMALL_STATE(5747)] = 295942,
  [SMALL_STATE(5748)] = 295965,
  [SMALL_STATE(5749)] = 296006,
  [SMALL_STATE(5750)] = 296047,
  [SMALL_STATE(5751)] = 296088,
  [SMALL_STATE(5752)] = 296129,
  [SMALL_STATE(5753)] = 296170,
  [SMALL_STATE(5754)] = 296211,
  [SMALL_STATE(5755)] = 296252,
  [SMALL_STATE(5756)] = 296293,
  [SMALL_STATE(5757)] = 296334,
  [SMALL_STATE(5758)] = 296375,
  [SMALL_STATE(5759)] = 296416,
  [SMALL_STATE(5760)] = 296457,
  [SMALL_STATE(5761)] = 296498,
  [SMALL_STATE(5762)] = 296539,
  [SMALL_STATE(5763)] = 296580,
  [SMALL_STATE(5764)] = 296621,
  [SMALL_STATE(5765)] = 296662,
  [SMALL_STATE(5766)] = 296703,
  [SMALL_STATE(5767)] = 296744,
  [SMALL_STATE(5768)] = 296785,
  [SMALL_STATE(5769)] = 296826,
  [SMALL_STATE(5770)] = 296867,
  [SMALL_STATE(5771)] = 296908,
  [SMALL_STATE(5772)] = 296949,
  [SMALL_STATE(5773)] = 296990,
  [SMALL_STATE(5774)] = 297031,
  [SMALL_STATE(5775)] = 297072,
  [SMALL_STATE(5776)] = 297113,
  [SMALL_STATE(5777)] = 297154,
  [SMALL_STATE(5778)] = 297195,
  [SMALL_STATE(5779)] = 297236,
  [SMALL_STATE(5780)] = 297277,
  [SMALL_STATE(5781)] = 297318,
  [SMALL_STATE(5782)] = 297359,
  [SMALL_STATE(5783)] = 297400,
  [SMALL_STATE(5784)] = 297441,
  [SMALL_STATE(5785)] = 297482,
  [SMALL_STATE(5786)] = 297523,
  [SMALL_STATE(5787)] = 297564,
  [SMALL_STATE(5788)] = 297605,
  [SMALL_STATE(5789)] = 297646,
  [SMALL_STATE(5790)] = 297669,
  [SMALL_STATE(5791)] = 297710,
  [SMALL_STATE(5792)] = 297751,
  [SMALL_STATE(5793)] = 297792,
  [SMALL_STATE(5794)] = 297833,
  [SMALL_STATE(5795)] = 297874,
  [SMALL_STATE(5796)] = 297915,
  [SMALL_STATE(5797)] = 297956,
  [SMALL_STATE(5798)] = 297997,
  [SMALL_STATE(5799)] = 298038,
  [SMALL_STATE(5800)] = 298079,
  [SMALL_STATE(5801)] = 298120,
  [SMALL_STATE(5802)] = 298161,
  [SMALL_STATE(5803)] = 298202,
  [SMALL_STATE(5804)] = 298243,
  [SMALL_STATE(5805)] = 298284,
  [SMALL_STATE(5806)] = 298325,
  [SMALL_STATE(5807)] = 298366,
  [SMALL_STATE(5808)] = 298407,
  [SMALL_STATE(5809)] = 298448,
  [SMALL_STATE(5810)] = 298489,
  [SMALL_STATE(5811)] = 298530,
  [SMALL_STATE(5812)] = 298571,
  [SMALL_STATE(5813)] = 298612,
  [SMALL_STATE(5814)] = 298653,
  [SMALL_STATE(5815)] = 298694,
  [SMALL_STATE(5816)] = 298735,
  [SMALL_STATE(5817)] = 298776,
  [SMALL_STATE(5818)] = 298817,
  [SMALL_STATE(5819)] = 298858,
  [SMALL_STATE(5820)] = 298899,
  [SMALL_STATE(5821)] = 298940,
  [SMALL_STATE(5822)] = 298981,
  [SMALL_STATE(5823)] = 299022,
  [SMALL_STATE(5824)] = 299063,
  [SMALL_STATE(5825)] = 299104,
  [SMALL_STATE(5826)] = 299145,
  [SMALL_STATE(5827)] = 299186,
  [SMALL_STATE(5828)] = 299227,
  [SMALL_STATE(5829)] = 299268,
  [SMALL_STATE(5830)] = 299309,
  [SMALL_STATE(5831)] = 299350,
  [SMALL_STATE(5832)] = 299391,
  [SMALL_STATE(5833)] = 299432,
  [SMALL_STATE(5834)] = 299473,
  [SMALL_STATE(5835)] = 299514,
  [SMALL_STATE(5836)] = 299555,
  [SMALL_STATE(5837)] = 299596,
  [SMALL_STATE(5838)] = 299637,
  [SMALL_STATE(5839)] = 299678,
  [SMALL_STATE(5840)] = 299719,
  [SMALL_STATE(5841)] = 299760,
  [SMALL_STATE(5842)] = 299801,
  [SMALL_STATE(5843)] = 299842,
  [SMALL_STATE(5844)] = 299883,
  [SMALL_STATE(5845)] = 299924,
  [SMALL_STATE(5846)] = 299965,
  [SMALL_STATE(5847)] = 300006,
  [SMALL_STATE(5848)] = 300047,
  [SMALL_STATE(5849)] = 300088,
  [SMALL_STATE(5850)] = 300129,
  [SMALL_STATE(5851)] = 300170,
  [SMALL_STATE(5852)] = 300211,
  [SMALL_STATE(5853)] = 300252,
  [SMALL_STATE(5854)] = 300293,
  [SMALL_STATE(5855)] = 300334,
  [SMALL_STATE(5856)] = 300375,
  [SMALL_STATE(5857)] = 300416,
  [SMALL_STATE(5858)] = 300457,
  [SMALL_STATE(5859)] = 300498,
  [SMALL_STATE(5860)] = 300539,
  [SMALL_STATE(5861)] = 300580,
  [SMALL_STATE(5862)] = 300621,
  [SMALL_STATE(5863)] = 300662,
  [SMALL_STATE(5864)] = 300703,
  [SMALL_STATE(5865)] = 300744,
  [SMALL_STATE(5866)] = 300772,
  [SMALL_STATE(5867)] = 300800,
  [SMALL_STATE(5868)] = 300828,
  [SMALL_STATE(5869)] = 300856,
  [SMALL_STATE(5870)] = 300884,
  [SMALL_STATE(5871)] = 300912,
  [SMALL_STATE(5872)] = 300940,
  [SMALL_STATE(5873)] = 300968,
  [SMALL_STATE(5874)] = 300996,
  [SMALL_STATE(5875)] = 301024,
  [SMALL_STATE(5876)] = 301052,
  [SMALL_STATE(5877)] = 301080,
  [SMALL_STATE(5878)] = 301108,
  [SMALL_STATE(5879)] = 301136,
  [SMALL_STATE(5880)] = 301164,
  [SMALL_STATE(5881)] = 301192,
  [SMALL_STATE(5882)] = 301220,
  [SMALL_STATE(5883)] = 301248,
  [SMALL_STATE(5884)] = 301276,
  [SMALL_STATE(5885)] = 301304,
  [SMALL_STATE(5886)] = 301332,
  [SMALL_STATE(5887)] = 301360,
  [SMALL_STATE(5888)] = 301388,
  [SMALL_STATE(5889)] = 301416,
  [SMALL_STATE(5890)] = 301444,
  [SMALL_STATE(5891)] = 301472,
  [SMALL_STATE(5892)] = 301500,
  [SMALL_STATE(5893)] = 301528,
  [SMALL_STATE(5894)] = 301556,
  [SMALL_STATE(5895)] = 301584,
  [SMALL_STATE(5896)] = 301612,
  [SMALL_STATE(5897)] = 301640,
  [SMALL_STATE(5898)] = 301668,
  [SMALL_STATE(5899)] = 301696,
  [SMALL_STATE(5900)] = 301724,
  [SMALL_STATE(5901)] = 301752,
  [SMALL_STATE(5902)] = 301780,
  [SMALL_STATE(5903)] = 301808,
  [SMALL_STATE(5904)] = 301836,
  [SMALL_STATE(5905)] = 301864,
  [SMALL_STATE(5906)] = 301892,
  [SMALL_STATE(5907)] = 301920,
  [SMALL_STATE(5908)] = 301948,
  [SMALL_STATE(5909)] = 301976,
  [SMALL_STATE(5910)] = 302004,
  [SMALL_STATE(5911)] = 302032,
  [SMALL_STATE(5912)] = 302060,
  [SMALL_STATE(5913)] = 302088,
  [SMALL_STATE(5914)] = 302116,
  [SMALL_STATE(5915)] = 302144,
  [SMALL_STATE(5916)] = 302172,
  [SMALL_STATE(5917)] = 302200,
  [SMALL_STATE(5918)] = 302228,
  [SMALL_STATE(5919)] = 302256,
  [SMALL_STATE(5920)] = 302284,
  [SMALL_STATE(5921)] = 302312,
  [SMALL_STATE(5922)] = 302340,
  [SMALL_STATE(5923)] = 302368,
  [SMALL_STATE(5924)] = 302396,
  [SMALL_STATE(5925)] = 302424,
  [SMALL_STATE(5926)] = 302452,
  [SMALL_STATE(5927)] = 302480,
  [SMALL_STATE(5928)] = 302508,
  [SMALL_STATE(5929)] = 302536,
  [SMALL_STATE(5930)] = 302564,
  [SMALL_STATE(5931)] = 302592,
  [SMALL_STATE(5932)] = 302620,
  [SMALL_STATE(5933)] = 302648,
  [SMALL_STATE(5934)] = 302676,
  [SMALL_STATE(5935)] = 302704,
  [SMALL_STATE(5936)] = 302732,
  [SMALL_STATE(5937)] = 302760,
  [SMALL_STATE(5938)] = 302788,
  [SMALL_STATE(5939)] = 302816,
  [SMALL_STATE(5940)] = 302844,
  [SMALL_STATE(5941)] = 302872,
  [SMALL_STATE(5942)] = 302900,
  [SMALL_STATE(5943)] = 302928,
  [SMALL_STATE(5944)] = 302956,
  [SMALL_STATE(5945)] = 302984,
  [SMALL_STATE(5946)] = 303012,
  [SMALL_STATE(5947)] = 303040,
  [SMALL_STATE(5948)] = 303068,
  [SMALL_STATE(5949)] = 303096,
  [SMALL_STATE(5950)] = 303124,
  [SMALL_STATE(5951)] = 303152,
  [SMALL_STATE(5952)] = 303180,
  [SMALL_STATE(5953)] = 303208,
  [SMALL_STATE(5954)] = 303236,
  [SMALL_STATE(5955)] = 303264,
  [SMALL_STATE(5956)] = 303292,
  [SMALL_STATE(5957)] = 303320,
  [SMALL_STATE(5958)] = 303348,
  [SMALL_STATE(5959)] = 303376,
  [SMALL_STATE(5960)] = 303404,
  [SMALL_STATE(5961)] = 303432,
  [SMALL_STATE(5962)] = 303460,
  [SMALL_STATE(5963)] = 303488,
  [SMALL_STATE(5964)] = 303516,
  [SMALL_STATE(5965)] = 303544,
  [SMALL_STATE(5966)] = 303572,
  [SMALL_STATE(5967)] = 303600,
  [SMALL_STATE(5968)] = 303628,
  [SMALL_STATE(5969)] = 303656,
  [SMALL_STATE(5970)] = 303684,
  [SMALL_STATE(5971)] = 303712,
  [SMALL_STATE(5972)] = 303740,
  [SMALL_STATE(5973)] = 303768,
  [SMALL_STATE(5974)] = 303796,
  [SMALL_STATE(5975)] = 303824,
  [SMALL_STATE(5976)] = 303852,
  [SMALL_STATE(5977)] = 303880,
  [SMALL_STATE(5978)] = 303908,
  [SMALL_STATE(5979)] = 303936,
  [SMALL_STATE(5980)] = 303964,
  [SMALL_STATE(5981)] = 303992,
  [SMALL_STATE(5982)] = 304020,
  [SMALL_STATE(5983)] = 304048,
  [SMALL_STATE(5984)] = 304076,
  [SMALL_STATE(5985)] = 304104,
  [SMALL_STATE(5986)] = 304132,
  [SMALL_STATE(5987)] = 304160,
  [SMALL_STATE(5988)] = 304188,
  [SMALL_STATE(5989)] = 304216,
  [SMALL_STATE(5990)] = 304244,
  [SMALL_STATE(5991)] = 304272,
  [SMALL_STATE(5992)] = 304300,
  [SMALL_STATE(5993)] = 304328,
  [SMALL_STATE(5994)] = 304356,
  [SMALL_STATE(5995)] = 304384,
  [SMALL_STATE(5996)] = 304412,
  [SMALL_STATE(5997)] = 304440,
  [SMALL_STATE(5998)] = 304468,
  [SMALL_STATE(5999)] = 304496,
  [SMALL_STATE(6000)] = 304524,
  [SMALL_STATE(6001)] = 304552,
  [SMALL_STATE(6002)] = 304580,
  [SMALL_STATE(6003)] = 304608,
  [SMALL_STATE(6004)] = 304636,
  [SMALL_STATE(6005)] = 304664,
  [SMALL_STATE(6006)] = 304692,
  [SMALL_STATE(6007)] = 304720,
  [SMALL_STATE(6008)] = 304748,
  [SMALL_STATE(6009)] = 304776,
  [SMALL_STATE(6010)] = 304804,
  [SMALL_STATE(6011)] = 304832,
  [SMALL_STATE(6012)] = 304860,
  [SMALL_STATE(6013)] = 304888,
  [SMALL_STATE(6014)] = 304916,
  [SMALL_STATE(6015)] = 304944,
  [SMALL_STATE(6016)] = 304972,
  [SMALL_STATE(6017)] = 305000,
  [SMALL_STATE(6018)] = 305028,
  [SMALL_STATE(6019)] = 305056,
  [SMALL_STATE(6020)] = 305084,
  [SMALL_STATE(6021)] = 305112,
  [SMALL_STATE(6022)] = 305140,
  [SMALL_STATE(6023)] = 305168,
  [SMALL_STATE(6024)] = 305196,
  [SMALL_STATE(6025)] = 305224,
  [SMALL_STATE(6026)] = 305252,
  [SMALL_STATE(6027)] = 305280,
  [SMALL_STATE(6028)] = 305308,
  [SMALL_STATE(6029)] = 305336,
  [SMALL_STATE(6030)] = 305364,
  [SMALL_STATE(6031)] = 305392,
  [SMALL_STATE(6032)] = 305420,
  [SMALL_STATE(6033)] = 305448,
  [SMALL_STATE(6034)] = 305476,
  [SMALL_STATE(6035)] = 305504,
  [SMALL_STATE(6036)] = 305532,
  [SMALL_STATE(6037)] = 305560,
  [SMALL_STATE(6038)] = 305588,
  [SMALL_STATE(6039)] = 305616,
  [SMALL_STATE(6040)] = 305644,
  [SMALL_STATE(6041)] = 305672,
  [SMALL_STATE(6042)] = 305700,
  [SMALL_STATE(6043)] = 305728,
  [SMALL_STATE(6044)] = 305756,
  [SMALL_STATE(6045)] = 305784,
  [SMALL_STATE(6046)] = 305812,
  [SMALL_STATE(6047)] = 305840,
  [SMALL_STATE(6048)] = 305868,
  [SMALL_STATE(6049)] = 305896,
  [SMALL_STATE(6050)] = 305924,
  [SMALL_STATE(6051)] = 305952,
  [SMALL_STATE(6052)] = 305980,
  [SMALL_STATE(6053)] = 306008,
  [SMALL_STATE(6054)] = 306036,
  [SMALL_STATE(6055)] = 306064,
  [SMALL_STATE(6056)] = 306092,
  [SMALL_STATE(6057)] = 306120,
  [SMALL_STATE(6058)] = 306148,
  [SMALL_STATE(6059)] = 306176,
  [SMALL_STATE(6060)] = 306204,
  [SMALL_STATE(6061)] = 306232,
  [SMALL_STATE(6062)] = 306260,
  [SMALL_STATE(6063)] = 306288,
  [SMALL_STATE(6064)] = 306316,
  [SMALL_STATE(6065)] = 306344,
  [SMALL_STATE(6066)] = 306372,
  [SMALL_STATE(6067)] = 306400,
  [SMALL_STATE(6068)] = 306428,
  [SMALL_STATE(6069)] = 306456,
  [SMALL_STATE(6070)] = 306484,
  [SMALL_STATE(6071)] = 306512,
  [SMALL_STATE(6072)] = 306540,
  [SMALL_STATE(6073)] = 306568,
  [SMALL_STATE(6074)] = 306596,
  [SMALL_STATE(6075)] = 306624,
  [SMALL_STATE(6076)] = 306652,
  [SMALL_STATE(6077)] = 306680,
  [SMALL_STATE(6078)] = 306708,
  [SMALL_STATE(6079)] = 306736,
  [SMALL_STATE(6080)] = 306764,
  [SMALL_STATE(6081)] = 306792,
  [SMALL_STATE(6082)] = 306820,
  [SMALL_STATE(6083)] = 306848,
  [SMALL_STATE(6084)] = 306876,
  [SMALL_STATE(6085)] = 306904,
  [SMALL_STATE(6086)] = 306932,
  [SMALL_STATE(6087)] = 306960,
  [SMALL_STATE(6088)] = 306988,
  [SMALL_STATE(6089)] = 307017,
  [SMALL_STATE(6090)] = 307046,
  [SMALL_STATE(6091)] = 307075,
  [SMALL_STATE(6092)] = 307104,
  [SMALL_STATE(6093)] = 307133,
  [SMALL_STATE(6094)] = 307162,
  [SMALL_STATE(6095)] = 307191,
  [SMALL_STATE(6096)] = 307212,
  [SMALL_STATE(6097)] = 307233,
  [SMALL_STATE(6098)] = 307262,
  [SMALL_STATE(6099)] = 307287,
  [SMALL_STATE(6100)] = 307309,
  [SMALL_STATE(6101)] = 307335,
  [SMALL_STATE(6102)] = 307357,
  [SMALL_STATE(6103)] = 307379,
  [SMALL_STATE(6104)] = 307401,
  [SMALL_STATE(6105)] = 307427,
  [SMALL_STATE(6106)] = 307453,
  [SMALL_STATE(6107)] = 307479,
  [SMALL_STATE(6108)] = 307505,
  [SMALL_STATE(6109)] = 307527,
  [SMALL_STATE(6110)] = 307553,
  [SMALL_STATE(6111)] = 307575,
  [SMALL_STATE(6112)] = 307597,
  [SMALL_STATE(6113)] = 307619,
  [SMALL_STATE(6114)] = 307645,
  [SMALL_STATE(6115)] = 307667,
  [SMALL_STATE(6116)] = 307693,
  [SMALL_STATE(6117)] = 307719,
  [SMALL_STATE(6118)] = 307742,
  [SMALL_STATE(6119)] = 307765,
  [SMALL_STATE(6120)] = 307788,
  [SMALL_STATE(6121)] = 307811,
  [SMALL_STATE(6122)] = 307830,
  [SMALL_STATE(6123)] = 307849,
  [SMALL_STATE(6124)] = 307868,
  [SMALL_STATE(6125)] = 307887,
  [SMALL_STATE(6126)] = 307906,
  [SMALL_STATE(6127)] = 307929,
  [SMALL_STATE(6128)] = 307948,
  [SMALL_STATE(6129)] = 307979,
  [SMALL_STATE(6130)] = 308002,
  [SMALL_STATE(6131)] = 308025,
  [SMALL_STATE(6132)] = 308056,
  [SMALL_STATE(6133)] = 308079,
  [SMALL_STATE(6134)] = 308102,
  [SMALL_STATE(6135)] = 308125,
  [SMALL_STATE(6136)] = 308148,
  [SMALL_STATE(6137)] = 308171,
  [SMALL_STATE(6138)] = 308202,
  [SMALL_STATE(6139)] = 308221,
  [SMALL_STATE(6140)] = 308244,
  [SMALL_STATE(6141)] = 308263,
  [SMALL_STATE(6142)] = 308282,
  [SMALL_STATE(6143)] = 308305,
  [SMALL_STATE(6144)] = 308324,
  [SMALL_STATE(6145)] = 308347,
  [SMALL_STATE(6146)] = 308370,
  [SMALL_STATE(6147)] = 308393,
  [SMALL_STATE(6148)] = 308412,
  [SMALL_STATE(6149)] = 308435,
  [SMALL_STATE(6150)] = 308458,
  [SMALL_STATE(6151)] = 308481,
  [SMALL_STATE(6152)] = 308504,
  [SMALL_STATE(6153)] = 308527,
  [SMALL_STATE(6154)] = 308550,
  [SMALL_STATE(6155)] = 308569,
  [SMALL_STATE(6156)] = 308592,
  [SMALL_STATE(6157)] = 308615,
  [SMALL_STATE(6158)] = 308638,
  [SMALL_STATE(6159)] = 308661,
  [SMALL_STATE(6160)] = 308689,
  [SMALL_STATE(6161)] = 308717,
  [SMALL_STATE(6162)] = 308745,
  [SMALL_STATE(6163)] = 308773,
  [SMALL_STATE(6164)] = 308801,
  [SMALL_STATE(6165)] = 308829,
  [SMALL_STATE(6166)] = 308857,
  [SMALL_STATE(6167)] = 308885,
  [SMALL_STATE(6168)] = 308913,
  [SMALL_STATE(6169)] = 308941,
  [SMALL_STATE(6170)] = 308969,
  [SMALL_STATE(6171)] = 308997,
  [SMALL_STATE(6172)] = 309025,
  [SMALL_STATE(6173)] = 309053,
  [SMALL_STATE(6174)] = 309081,
  [SMALL_STATE(6175)] = 309109,
  [SMALL_STATE(6176)] = 309137,
  [SMALL_STATE(6177)] = 309153,
  [SMALL_STATE(6178)] = 309181,
  [SMALL_STATE(6179)] = 309209,
  [SMALL_STATE(6180)] = 309225,
  [SMALL_STATE(6181)] = 309253,
  [SMALL_STATE(6182)] = 309281,
  [SMALL_STATE(6183)] = 309309,
  [SMALL_STATE(6184)] = 309325,
  [SMALL_STATE(6185)] = 309353,
  [SMALL_STATE(6186)] = 309381,
  [SMALL_STATE(6187)] = 309409,
  [SMALL_STATE(6188)] = 309440,
  [SMALL_STATE(6189)] = 309463,
  [SMALL_STATE(6190)] = 309487,
  [SMALL_STATE(6191)] = 309513,
  [SMALL_STATE(6192)] = 309537,
  [SMALL_STATE(6193)] = 309561,
  [SMALL_STATE(6194)] = 309583,
  [SMALL_STATE(6195)] = 309609,
  [SMALL_STATE(6196)] = 309633,
  [SMALL_STATE(6197)] = 309657,
  [SMALL_STATE(6198)] = 309683,
  [SMALL_STATE(6199)] = 309707,
  [SMALL_STATE(6200)] = 309731,
  [SMALL_STATE(6201)] = 309755,
  [SMALL_STATE(6202)] = 309777,
  [SMALL_STATE(6203)] = 309801,
  [SMALL_STATE(6204)] = 309825,
  [SMALL_STATE(6205)] = 309847,
  [SMALL_STATE(6206)] = 309862,
  [SMALL_STATE(6207)] = 309877,
  [SMALL_STATE(6208)] = 309892,
  [SMALL_STATE(6209)] = 309909,
  [SMALL_STATE(6210)] = 309924,
  [SMALL_STATE(6211)] = 309939,
  [SMALL_STATE(6212)] = 309954,
  [SMALL_STATE(6213)] = 309969,
  [SMALL_STATE(6214)] = 309990,
  [SMALL_STATE(6215)] = 310005,
  [SMALL_STATE(6216)] = 310018,
  [SMALL_STATE(6217)] = 310031,
  [SMALL_STATE(6218)] = 310046,
  [SMALL_STATE(6219)] = 310063,
  [SMALL_STATE(6220)] = 310084,
  [SMALL_STATE(6221)] = 310105,
  [SMALL_STATE(6222)] = 310120,
  [SMALL_STATE(6223)] = 310135,
  [SMALL_STATE(6224)] = 310156,
  [SMALL_STATE(6225)] = 310171,
  [SMALL_STATE(6226)] = 310186,
  [SMALL_STATE(6227)] = 310201,
  [SMALL_STATE(6228)] = 310214,
  [SMALL_STATE(6229)] = 310227,
  [SMALL_STATE(6230)] = 310240,
  [SMALL_STATE(6231)] = 310255,
  [SMALL_STATE(6232)] = 310270,
  [SMALL_STATE(6233)] = 310285,
  [SMALL_STATE(6234)] = 310300,
  [SMALL_STATE(6235)] = 310313,
  [SMALL_STATE(6236)] = 310328,
  [SMALL_STATE(6237)] = 310343,
  [SMALL_STATE(6238)] = 310358,
  [SMALL_STATE(6239)] = 310373,
  [SMALL_STATE(6240)] = 310388,
  [SMALL_STATE(6241)] = 310407,
  [SMALL_STATE(6242)] = 310422,
  [SMALL_STATE(6243)] = 310437,
  [SMALL_STATE(6244)] = 310452,
  [SMALL_STATE(6245)] = 310467,
  [SMALL_STATE(6246)] = 310480,
  [SMALL_STATE(6247)] = 310493,
  [SMALL_STATE(6248)] = 310506,
  [SMALL_STATE(6249)] = 310526,
  [SMALL_STATE(6250)] = 310546,
  [SMALL_STATE(6251)] = 310566,
  [SMALL_STATE(6252)] = 310586,
  [SMALL_STATE(6253)] = 310606,
  [SMALL_STATE(6254)] = 310626,
  [SMALL_STATE(6255)] = 310646,
  [SMALL_STATE(6256)] = 310666,
  [SMALL_STATE(6257)] = 310680,
  [SMALL_STATE(6258)] = 310700,
  [SMALL_STATE(6259)] = 310714,
  [SMALL_STATE(6260)] = 310728,
  [SMALL_STATE(6261)] = 310742,
  [SMALL_STATE(6262)] = 310756,
  [SMALL_STATE(6263)] = 310770,
  [SMALL_STATE(6264)] = 310790,
  [SMALL_STATE(6265)] = 310804,
  [SMALL_STATE(6266)] = 310824,
  [SMALL_STATE(6267)] = 310838,
  [SMALL_STATE(6268)] = 310852,
  [SMALL_STATE(6269)] = 310866,
  [SMALL_STATE(6270)] = 310886,
  [SMALL_STATE(6271)] = 310906,
  [SMALL_STATE(6272)] = 310926,
  [SMALL_STATE(6273)] = 310946,
  [SMALL_STATE(6274)] = 310966,
  [SMALL_STATE(6275)] = 310986,
  [SMALL_STATE(6276)] = 311000,
  [SMALL_STATE(6277)] = 311020,
  [SMALL_STATE(6278)] = 311040,
  [SMALL_STATE(6279)] = 311060,
  [SMALL_STATE(6280)] = 311074,
  [SMALL_STATE(6281)] = 311094,
  [SMALL_STATE(6282)] = 311114,
  [SMALL_STATE(6283)] = 311134,
  [SMALL_STATE(6284)] = 311154,
  [SMALL_STATE(6285)] = 311174,
  [SMALL_STATE(6286)] = 311194,
  [SMALL_STATE(6287)] = 311214,
  [SMALL_STATE(6288)] = 311228,
  [SMALL_STATE(6289)] = 311248,
  [SMALL_STATE(6290)] = 311268,
  [SMALL_STATE(6291)] = 311288,
  [SMALL_STATE(6292)] = 311308,
  [SMALL_STATE(6293)] = 311328,
  [SMALL_STATE(6294)] = 311342,
  [SMALL_STATE(6295)] = 311362,
  [SMALL_STATE(6296)] = 311382,
  [SMALL_STATE(6297)] = 311402,
  [SMALL_STATE(6298)] = 311422,
  [SMALL_STATE(6299)] = 311442,
  [SMALL_STATE(6300)] = 311456,
  [SMALL_STATE(6301)] = 311476,
  [SMALL_STATE(6302)] = 311490,
  [SMALL_STATE(6303)] = 311510,
  [SMALL_STATE(6304)] = 311530,
  [SMALL_STATE(6305)] = 311544,
  [SMALL_STATE(6306)] = 311564,
  [SMALL_STATE(6307)] = 311578,
  [SMALL_STATE(6308)] = 311598,
  [SMALL_STATE(6309)] = 311618,
  [SMALL_STATE(6310)] = 311638,
  [SMALL_STATE(6311)] = 311658,
  [SMALL_STATE(6312)] = 311678,
  [SMALL_STATE(6313)] = 311698,
  [SMALL_STATE(6314)] = 311718,
  [SMALL_STATE(6315)] = 311738,
  [SMALL_STATE(6316)] = 311758,
  [SMALL_STATE(6317)] = 311778,
  [SMALL_STATE(6318)] = 311792,
  [SMALL_STATE(6319)] = 311812,
  [SMALL_STATE(6320)] = 311832,
  [SMALL_STATE(6321)] = 311852,
  [SMALL_STATE(6322)] = 311872,
  [SMALL_STATE(6323)] = 311892,
  [SMALL_STATE(6324)] = 311912,
  [SMALL_STATE(6325)] = 311932,
  [SMALL_STATE(6326)] = 311952,
  [SMALL_STATE(6327)] = 311968,
  [SMALL_STATE(6328)] = 311988,
  [SMALL_STATE(6329)] = 312008,
  [SMALL_STATE(6330)] = 312028,
  [SMALL_STATE(6331)] = 312048,
  [SMALL_STATE(6332)] = 312068,
  [SMALL_STATE(6333)] = 312088,
  [SMALL_STATE(6334)] = 312108,
  [SMALL_STATE(6335)] = 312128,
  [SMALL_STATE(6336)] = 312148,
  [SMALL_STATE(6337)] = 312168,
  [SMALL_STATE(6338)] = 312188,
  [SMALL_STATE(6339)] = 312208,
  [SMALL_STATE(6340)] = 312228,
  [SMALL_STATE(6341)] = 312248,
  [SMALL_STATE(6342)] = 312268,
  [SMALL_STATE(6343)] = 312288,
  [SMALL_STATE(6344)] = 312308,
  [SMALL_STATE(6345)] = 312322,
  [SMALL_STATE(6346)] = 312342,
  [SMALL_STATE(6347)] = 312362,
  [SMALL_STATE(6348)] = 312382,
  [SMALL_STATE(6349)] = 312402,
  [SMALL_STATE(6350)] = 312422,
  [SMALL_STATE(6351)] = 312442,
  [SMALL_STATE(6352)] = 312462,
  [SMALL_STATE(6353)] = 312482,
  [SMALL_STATE(6354)] = 312502,
  [SMALL_STATE(6355)] = 312522,
  [SMALL_STATE(6356)] = 312542,
  [SMALL_STATE(6357)] = 312562,
  [SMALL_STATE(6358)] = 312582,
  [SMALL_STATE(6359)] = 312602,
  [SMALL_STATE(6360)] = 312622,
  [SMALL_STATE(6361)] = 312642,
  [SMALL_STATE(6362)] = 312656,
  [SMALL_STATE(6363)] = 312676,
  [SMALL_STATE(6364)] = 312696,
  [SMALL_STATE(6365)] = 312716,
  [SMALL_STATE(6366)] = 312736,
  [SMALL_STATE(6367)] = 312756,
  [SMALL_STATE(6368)] = 312776,
  [SMALL_STATE(6369)] = 312795,
  [SMALL_STATE(6370)] = 312814,
  [SMALL_STATE(6371)] = 312827,
  [SMALL_STATE(6372)] = 312846,
  [SMALL_STATE(6373)] = 312863,
  [SMALL_STATE(6374)] = 312880,
  [SMALL_STATE(6375)] = 312895,
  [SMALL_STATE(6376)] = 312912,
  [SMALL_STATE(6377)] = 312927,
  [SMALL_STATE(6378)] = 312946,
  [SMALL_STATE(6379)] = 312965,
  [SMALL_STATE(6380)] = 312982,
  [SMALL_STATE(6381)] = 312999,
  [SMALL_STATE(6382)] = 313014,
  [SMALL_STATE(6383)] = 313031,
  [SMALL_STATE(6384)] = 313050,
  [SMALL_STATE(6385)] = 313067,
  [SMALL_STATE(6386)] = 313086,
  [SMALL_STATE(6387)] = 313105,
  [SMALL_STATE(6388)] = 313124,
  [SMALL_STATE(6389)] = 313137,
  [SMALL_STATE(6390)] = 313152,
  [SMALL_STATE(6391)] = 313167,
  [SMALL_STATE(6392)] = 313184,
  [SMALL_STATE(6393)] = 313203,
  [SMALL_STATE(6394)] = 313218,
  [SMALL_STATE(6395)] = 313233,
  [SMALL_STATE(6396)] = 313250,
  [SMALL_STATE(6397)] = 313269,
  [SMALL_STATE(6398)] = 313286,
  [SMALL_STATE(6399)] = 313305,
  [SMALL_STATE(6400)] = 313322,
  [SMALL_STATE(6401)] = 313337,
  [SMALL_STATE(6402)] = 313356,
  [SMALL_STATE(6403)] = 313371,
  [SMALL_STATE(6404)] = 313384,
  [SMALL_STATE(6405)] = 313401,
  [SMALL_STATE(6406)] = 313420,
  [SMALL_STATE(6407)] = 313437,
  [SMALL_STATE(6408)] = 313450,
  [SMALL_STATE(6409)] = 313465,
  [SMALL_STATE(6410)] = 313484,
  [SMALL_STATE(6411)] = 313501,
  [SMALL_STATE(6412)] = 313518,
  [SMALL_STATE(6413)] = 313537,
  [SMALL_STATE(6414)] = 313552,
  [SMALL_STATE(6415)] = 313569,
  [SMALL_STATE(6416)] = 313588,
  [SMALL_STATE(6417)] = 313605,
  [SMALL_STATE(6418)] = 313624,
  [SMALL_STATE(6419)] = 313641,
  [SMALL_STATE(6420)] = 313656,
  [SMALL_STATE(6421)] = 313671,
  [SMALL_STATE(6422)] = 313690,
  [SMALL_STATE(6423)] = 313709,
  [SMALL_STATE(6424)] = 313726,
  [SMALL_STATE(6425)] = 313741,
  [SMALL_STATE(6426)] = 313758,
  [SMALL_STATE(6427)] = 313777,
  [SMALL_STATE(6428)] = 313794,
  [SMALL_STATE(6429)] = 313809,
  [SMALL_STATE(6430)] = 313826,
  [SMALL_STATE(6431)] = 313845,
  [SMALL_STATE(6432)] = 313864,
  [SMALL_STATE(6433)] = 313883,
  [SMALL_STATE(6434)] = 313902,
  [SMALL_STATE(6435)] = 313919,
  [SMALL_STATE(6436)] = 313930,
  [SMALL_STATE(6437)] = 313941,
  [SMALL_STATE(6438)] = 313960,
  [SMALL_STATE(6439)] = 313971,
  [SMALL_STATE(6440)] = 313988,
  [SMALL_STATE(6441)] = 314007,
  [SMALL_STATE(6442)] = 314026,
  [SMALL_STATE(6443)] = 314043,
  [SMALL_STATE(6444)] = 314058,
  [SMALL_STATE(6445)] = 314077,
  [SMALL_STATE(6446)] = 314094,
  [SMALL_STATE(6447)] = 314113,
  [SMALL_STATE(6448)] = 314130,
  [SMALL_STATE(6449)] = 314145,
  [SMALL_STATE(6450)] = 314162,
  [SMALL_STATE(6451)] = 314179,
  [SMALL_STATE(6452)] = 314196,
  [SMALL_STATE(6453)] = 314213,
  [SMALL_STATE(6454)] = 314230,
  [SMALL_STATE(6455)] = 314247,
  [SMALL_STATE(6456)] = 314264,
  [SMALL_STATE(6457)] = 314283,
  [SMALL_STATE(6458)] = 314302,
  [SMALL_STATE(6459)] = 314319,
  [SMALL_STATE(6460)] = 314338,
  [SMALL_STATE(6461)] = 314353,
  [SMALL_STATE(6462)] = 314372,
  [SMALL_STATE(6463)] = 314391,
  [SMALL_STATE(6464)] = 314408,
  [SMALL_STATE(6465)] = 314427,
  [SMALL_STATE(6466)] = 314442,
  [SMALL_STATE(6467)] = 314459,
  [SMALL_STATE(6468)] = 314476,
  [SMALL_STATE(6469)] = 314495,
  [SMALL_STATE(6470)] = 314510,
  [SMALL_STATE(6471)] = 314523,
  [SMALL_STATE(6472)] = 314538,
  [SMALL_STATE(6473)] = 314553,
  [SMALL_STATE(6474)] = 314570,
  [SMALL_STATE(6475)] = 314585,
  [SMALL_STATE(6476)] = 314600,
  [SMALL_STATE(6477)] = 314615,
  [SMALL_STATE(6478)] = 314630,
  [SMALL_STATE(6479)] = 314645,
  [SMALL_STATE(6480)] = 314664,
  [SMALL_STATE(6481)] = 314674,
  [SMALL_STATE(6482)] = 314686,
  [SMALL_STATE(6483)] = 314702,
  [SMALL_STATE(6484)] = 314712,
  [SMALL_STATE(6485)] = 314722,
  [SMALL_STATE(6486)] = 314732,
  [SMALL_STATE(6487)] = 314742,
  [SMALL_STATE(6488)] = 314752,
  [SMALL_STATE(6489)] = 314768,
  [SMALL_STATE(6490)] = 314778,
  [SMALL_STATE(6491)] = 314788,
  [SMALL_STATE(6492)] = 314804,
  [SMALL_STATE(6493)] = 314820,
  [SMALL_STATE(6494)] = 314832,
  [SMALL_STATE(6495)] = 314842,
  [SMALL_STATE(6496)] = 314858,
  [SMALL_STATE(6497)] = 314874,
  [SMALL_STATE(6498)] = 314884,
  [SMALL_STATE(6499)] = 314898,
  [SMALL_STATE(6500)] = 314908,
  [SMALL_STATE(6501)] = 314922,
  [SMALL_STATE(6502)] = 314938,
  [SMALL_STATE(6503)] = 314948,
  [SMALL_STATE(6504)] = 314958,
  [SMALL_STATE(6505)] = 314968,
  [SMALL_STATE(6506)] = 314978,
  [SMALL_STATE(6507)] = 314988,
  [SMALL_STATE(6508)] = 314998,
  [SMALL_STATE(6509)] = 315014,
  [SMALL_STATE(6510)] = 315024,
  [SMALL_STATE(6511)] = 315034,
  [SMALL_STATE(6512)] = 315050,
  [SMALL_STATE(6513)] = 315066,
  [SMALL_STATE(6514)] = 315076,
  [SMALL_STATE(6515)] = 315086,
  [SMALL_STATE(6516)] = 315096,
  [SMALL_STATE(6517)] = 315106,
  [SMALL_STATE(6518)] = 315116,
  [SMALL_STATE(6519)] = 315126,
  [SMALL_STATE(6520)] = 315136,
  [SMALL_STATE(6521)] = 315146,
  [SMALL_STATE(6522)] = 315162,
  [SMALL_STATE(6523)] = 315172,
  [SMALL_STATE(6524)] = 315182,
  [SMALL_STATE(6525)] = 315198,
  [SMALL_STATE(6526)] = 315208,
  [SMALL_STATE(6527)] = 315218,
  [SMALL_STATE(6528)] = 315228,
  [SMALL_STATE(6529)] = 315238,
  [SMALL_STATE(6530)] = 315248,
  [SMALL_STATE(6531)] = 315264,
  [SMALL_STATE(6532)] = 315276,
  [SMALL_STATE(6533)] = 315286,
  [SMALL_STATE(6534)] = 315300,
  [SMALL_STATE(6535)] = 315310,
  [SMALL_STATE(6536)] = 315320,
  [SMALL_STATE(6537)] = 315334,
  [SMALL_STATE(6538)] = 315344,
  [SMALL_STATE(6539)] = 315354,
  [SMALL_STATE(6540)] = 315368,
  [SMALL_STATE(6541)] = 315378,
  [SMALL_STATE(6542)] = 315392,
  [SMALL_STATE(6543)] = 315408,
  [SMALL_STATE(6544)] = 315422,
  [SMALL_STATE(6545)] = 315438,
  [SMALL_STATE(6546)] = 315452,
  [SMALL_STATE(6547)] = 315462,
  [SMALL_STATE(6548)] = 315473,
  [SMALL_STATE(6549)] = 315486,
  [SMALL_STATE(6550)] = 315497,
  [SMALL_STATE(6551)] = 315510,
  [SMALL_STATE(6552)] = 315523,
  [SMALL_STATE(6553)] = 315536,
  [SMALL_STATE(6554)] = 315549,
  [SMALL_STATE(6555)] = 315558,
  [SMALL_STATE(6556)] = 315567,
  [SMALL_STATE(6557)] = 315576,
  [SMALL_STATE(6558)] = 315585,
  [SMALL_STATE(6559)] = 315598,
  [SMALL_STATE(6560)] = 315611,
  [SMALL_STATE(6561)] = 315622,
  [SMALL_STATE(6562)] = 315631,
  [SMALL_STATE(6563)] = 315644,
  [SMALL_STATE(6564)] = 315653,
  [SMALL_STATE(6565)] = 315662,
  [SMALL_STATE(6566)] = 315675,
  [SMALL_STATE(6567)] = 315688,
  [SMALL_STATE(6568)] = 315701,
  [SMALL_STATE(6569)] = 315714,
  [SMALL_STATE(6570)] = 315727,
  [SMALL_STATE(6571)] = 315740,
  [SMALL_STATE(6572)] = 315749,
  [SMALL_STATE(6573)] = 315762,
  [SMALL_STATE(6574)] = 315771,
  [SMALL_STATE(6575)] = 315784,
  [SMALL_STATE(6576)] = 315795,
  [SMALL_STATE(6577)] = 315808,
  [SMALL_STATE(6578)] = 315821,
  [SMALL_STATE(6579)] = 315832,
  [SMALL_STATE(6580)] = 315845,
  [SMALL_STATE(6581)] = 315858,
  [SMALL_STATE(6582)] = 315867,
  [SMALL_STATE(6583)] = 315880,
  [SMALL_STATE(6584)] = 315891,
  [SMALL_STATE(6585)] = 315904,
  [SMALL_STATE(6586)] = 315915,
  [SMALL_STATE(6587)] = 315928,
  [SMALL_STATE(6588)] = 315941,
  [SMALL_STATE(6589)] = 315954,
  [SMALL_STATE(6590)] = 315967,
  [SMALL_STATE(6591)] = 315980,
  [SMALL_STATE(6592)] = 315993,
  [SMALL_STATE(6593)] = 316006,
  [SMALL_STATE(6594)] = 316019,
  [SMALL_STATE(6595)] = 316032,
  [SMALL_STATE(6596)] = 316045,
  [SMALL_STATE(6597)] = 316058,
  [SMALL_STATE(6598)] = 316071,
  [SMALL_STATE(6599)] = 316084,
  [SMALL_STATE(6600)] = 316097,
  [SMALL_STATE(6601)] = 316110,
  [SMALL_STATE(6602)] = 316123,
  [SMALL_STATE(6603)] = 316136,
  [SMALL_STATE(6604)] = 316147,
  [SMALL_STATE(6605)] = 316158,
  [SMALL_STATE(6606)] = 316171,
  [SMALL_STATE(6607)] = 316184,
  [SMALL_STATE(6608)] = 316197,
  [SMALL_STATE(6609)] = 316210,
  [SMALL_STATE(6610)] = 316221,
  [SMALL_STATE(6611)] = 316232,
  [SMALL_STATE(6612)] = 316243,
  [SMALL_STATE(6613)] = 316256,
  [SMALL_STATE(6614)] = 316269,
  [SMALL_STATE(6615)] = 316282,
  [SMALL_STATE(6616)] = 316291,
  [SMALL_STATE(6617)] = 316302,
  [SMALL_STATE(6618)] = 316315,
  [SMALL_STATE(6619)] = 316328,
  [SMALL_STATE(6620)] = 316339,
  [SMALL_STATE(6621)] = 316352,
  [SMALL_STATE(6622)] = 316365,
  [SMALL_STATE(6623)] = 316378,
  [SMALL_STATE(6624)] = 316391,
  [SMALL_STATE(6625)] = 316404,
  [SMALL_STATE(6626)] = 316417,
  [SMALL_STATE(6627)] = 316430,
  [SMALL_STATE(6628)] = 316443,
  [SMALL_STATE(6629)] = 316454,
  [SMALL_STATE(6630)] = 316467,
  [SMALL_STATE(6631)] = 316480,
  [SMALL_STATE(6632)] = 316491,
  [SMALL_STATE(6633)] = 316504,
  [SMALL_STATE(6634)] = 316517,
  [SMALL_STATE(6635)] = 316528,
  [SMALL_STATE(6636)] = 316537,
  [SMALL_STATE(6637)] = 316548,
  [SMALL_STATE(6638)] = 316561,
  [SMALL_STATE(6639)] = 316572,
  [SMALL_STATE(6640)] = 316585,
  [SMALL_STATE(6641)] = 316598,
  [SMALL_STATE(6642)] = 316609,
  [SMALL_STATE(6643)] = 316622,
  [SMALL_STATE(6644)] = 316635,
  [SMALL_STATE(6645)] = 316648,
  [SMALL_STATE(6646)] = 316657,
  [SMALL_STATE(6647)] = 316670,
  [SMALL_STATE(6648)] = 316683,
  [SMALL_STATE(6649)] = 316696,
  [SMALL_STATE(6650)] = 316707,
  [SMALL_STATE(6651)] = 316720,
  [SMALL_STATE(6652)] = 316731,
  [SMALL_STATE(6653)] = 316744,
  [SMALL_STATE(6654)] = 316757,
  [SMALL_STATE(6655)] = 316770,
  [SMALL_STATE(6656)] = 316783,
  [SMALL_STATE(6657)] = 316796,
  [SMALL_STATE(6658)] = 316809,
  [SMALL_STATE(6659)] = 316822,
  [SMALL_STATE(6660)] = 316835,
  [SMALL_STATE(6661)] = 316848,
  [SMALL_STATE(6662)] = 316861,
  [SMALL_STATE(6663)] = 316872,
  [SMALL_STATE(6664)] = 316881,
  [SMALL_STATE(6665)] = 316892,
  [SMALL_STATE(6666)] = 316905,
  [SMALL_STATE(6667)] = 316916,
  [SMALL_STATE(6668)] = 316929,
  [SMALL_STATE(6669)] = 316940,
  [SMALL_STATE(6670)] = 316953,
  [SMALL_STATE(6671)] = 316966,
  [SMALL_STATE(6672)] = 316979,
  [SMALL_STATE(6673)] = 316990,
  [SMALL_STATE(6674)] = 317001,
  [SMALL_STATE(6675)] = 317014,
  [SMALL_STATE(6676)] = 317027,
  [SMALL_STATE(6677)] = 317040,
  [SMALL_STATE(6678)] = 317051,
  [SMALL_STATE(6679)] = 317064,
  [SMALL_STATE(6680)] = 317077,
  [SMALL_STATE(6681)] = 317090,
  [SMALL_STATE(6682)] = 317103,
  [SMALL_STATE(6683)] = 317116,
  [SMALL_STATE(6684)] = 317129,
  [SMALL_STATE(6685)] = 317142,
  [SMALL_STATE(6686)] = 317155,
  [SMALL_STATE(6687)] = 317166,
  [SMALL_STATE(6688)] = 317179,
  [SMALL_STATE(6689)] = 317192,
  [SMALL_STATE(6690)] = 317205,
  [SMALL_STATE(6691)] = 317218,
  [SMALL_STATE(6692)] = 317231,
  [SMALL_STATE(6693)] = 317244,
  [SMALL_STATE(6694)] = 317257,
  [SMALL_STATE(6695)] = 317270,
  [SMALL_STATE(6696)] = 317283,
  [SMALL_STATE(6697)] = 317294,
  [SMALL_STATE(6698)] = 317307,
  [SMALL_STATE(6699)] = 317320,
  [SMALL_STATE(6700)] = 317333,
  [SMALL_STATE(6701)] = 317346,
  [SMALL_STATE(6702)] = 317359,
  [SMALL_STATE(6703)] = 317372,
  [SMALL_STATE(6704)] = 317385,
  [SMALL_STATE(6705)] = 317396,
  [SMALL_STATE(6706)] = 317407,
  [SMALL_STATE(6707)] = 317420,
  [SMALL_STATE(6708)] = 317433,
  [SMALL_STATE(6709)] = 317446,
  [SMALL_STATE(6710)] = 317459,
  [SMALL_STATE(6711)] = 317472,
  [SMALL_STATE(6712)] = 317485,
  [SMALL_STATE(6713)] = 317498,
  [SMALL_STATE(6714)] = 317511,
  [SMALL_STATE(6715)] = 317524,
  [SMALL_STATE(6716)] = 317537,
  [SMALL_STATE(6717)] = 317550,
  [SMALL_STATE(6718)] = 317561,
  [SMALL_STATE(6719)] = 317574,
  [SMALL_STATE(6720)] = 317587,
  [SMALL_STATE(6721)] = 317600,
  [SMALL_STATE(6722)] = 317613,
  [SMALL_STATE(6723)] = 317626,
  [SMALL_STATE(6724)] = 317639,
  [SMALL_STATE(6725)] = 317652,
  [SMALL_STATE(6726)] = 317665,
  [SMALL_STATE(6727)] = 317673,
  [SMALL_STATE(6728)] = 317681,
  [SMALL_STATE(6729)] = 317689,
  [SMALL_STATE(6730)] = 317699,
  [SMALL_STATE(6731)] = 317709,
  [SMALL_STATE(6732)] = 317717,
  [SMALL_STATE(6733)] = 317727,
  [SMALL_STATE(6734)] = 317735,
  [SMALL_STATE(6735)] = 317743,
  [SMALL_STATE(6736)] = 317751,
  [SMALL_STATE(6737)] = 317759,
  [SMALL_STATE(6738)] = 317767,
  [SMALL_STATE(6739)] = 317775,
  [SMALL_STATE(6740)] = 317783,
  [SMALL_STATE(6741)] = 317791,
  [SMALL_STATE(6742)] = 317799,
  [SMALL_STATE(6743)] = 317807,
  [SMALL_STATE(6744)] = 317817,
  [SMALL_STATE(6745)] = 317827,
  [SMALL_STATE(6746)] = 317837,
  [SMALL_STATE(6747)] = 317845,
  [SMALL_STATE(6748)] = 317855,
  [SMALL_STATE(6749)] = 317865,
  [SMALL_STATE(6750)] = 317875,
  [SMALL_STATE(6751)] = 317883,
  [SMALL_STATE(6752)] = 317891,
  [SMALL_STATE(6753)] = 317901,
  [SMALL_STATE(6754)] = 317909,
  [SMALL_STATE(6755)] = 317917,
  [SMALL_STATE(6756)] = 317927,
  [SMALL_STATE(6757)] = 317935,
  [SMALL_STATE(6758)] = 317943,
  [SMALL_STATE(6759)] = 317951,
  [SMALL_STATE(6760)] = 317959,
  [SMALL_STATE(6761)] = 317967,
  [SMALL_STATE(6762)] = 317975,
  [SMALL_STATE(6763)] = 317983,
  [SMALL_STATE(6764)] = 317993,
  [SMALL_STATE(6765)] = 318003,
  [SMALL_STATE(6766)] = 318011,
  [SMALL_STATE(6767)] = 318019,
  [SMALL_STATE(6768)] = 318027,
  [SMALL_STATE(6769)] = 318035,
  [SMALL_STATE(6770)] = 318043,
  [SMALL_STATE(6771)] = 318053,
  [SMALL_STATE(6772)] = 318061,
  [SMALL_STATE(6773)] = 318069,
  [SMALL_STATE(6774)] = 318077,
  [SMALL_STATE(6775)] = 318085,
  [SMALL_STATE(6776)] = 318093,
  [SMALL_STATE(6777)] = 318101,
  [SMALL_STATE(6778)] = 318111,
  [SMALL_STATE(6779)] = 318119,
  [SMALL_STATE(6780)] = 318129,
  [SMALL_STATE(6781)] = 318139,
  [SMALL_STATE(6782)] = 318147,
  [SMALL_STATE(6783)] = 318157,
  [SMALL_STATE(6784)] = 318167,
  [SMALL_STATE(6785)] = 318177,
  [SMALL_STATE(6786)] = 318187,
  [SMALL_STATE(6787)] = 318195,
  [SMALL_STATE(6788)] = 318205,
  [SMALL_STATE(6789)] = 318215,
  [SMALL_STATE(6790)] = 318225,
  [SMALL_STATE(6791)] = 318233,
  [SMALL_STATE(6792)] = 318243,
  [SMALL_STATE(6793)] = 318250,
  [SMALL_STATE(6794)] = 318257,
  [SMALL_STATE(6795)] = 318264,
  [SMALL_STATE(6796)] = 318271,
  [SMALL_STATE(6797)] = 318278,
  [SMALL_STATE(6798)] = 318285,
  [SMALL_STATE(6799)] = 318292,
  [SMALL_STATE(6800)] = 318299,
  [SMALL_STATE(6801)] = 318306,
  [SMALL_STATE(6802)] = 318313,
  [SMALL_STATE(6803)] = 318320,
  [SMALL_STATE(6804)] = 318327,
  [SMALL_STATE(6805)] = 318334,
  [SMALL_STATE(6806)] = 318341,
  [SMALL_STATE(6807)] = 318348,
  [SMALL_STATE(6808)] = 318355,
  [SMALL_STATE(6809)] = 318362,
  [SMALL_STATE(6810)] = 318369,
  [SMALL_STATE(6811)] = 318376,
  [SMALL_STATE(6812)] = 318383,
  [SMALL_STATE(6813)] = 318390,
  [SMALL_STATE(6814)] = 318397,
  [SMALL_STATE(6815)] = 318404,
  [SMALL_STATE(6816)] = 318411,
  [SMALL_STATE(6817)] = 318418,
  [SMALL_STATE(6818)] = 318425,
  [SMALL_STATE(6819)] = 318432,
  [SMALL_STATE(6820)] = 318439,
  [SMALL_STATE(6821)] = 318446,
  [SMALL_STATE(6822)] = 318453,
  [SMALL_STATE(6823)] = 318460,
  [SMALL_STATE(6824)] = 318467,
  [SMALL_STATE(6825)] = 318474,
  [SMALL_STATE(6826)] = 318481,
  [SMALL_STATE(6827)] = 318488,
  [SMALL_STATE(6828)] = 318495,
  [SMALL_STATE(6829)] = 318502,
  [SMALL_STATE(6830)] = 318509,
  [SMALL_STATE(6831)] = 318516,
  [SMALL_STATE(6832)] = 318523,
  [SMALL_STATE(6833)] = 318530,
  [SMALL_STATE(6834)] = 318537,
  [SMALL_STATE(6835)] = 318544,
  [SMALL_STATE(6836)] = 318551,
  [SMALL_STATE(6837)] = 318558,
  [SMALL_STATE(6838)] = 318565,
  [SMALL_STATE(6839)] = 318572,
  [SMALL_STATE(6840)] = 318579,
  [SMALL_STATE(6841)] = 318586,
  [SMALL_STATE(6842)] = 318593,
  [SMALL_STATE(6843)] = 318600,
  [SMALL_STATE(6844)] = 318607,
  [SMALL_STATE(6845)] = 318614,
  [SMALL_STATE(6846)] = 318621,
  [SMALL_STATE(6847)] = 318628,
  [SMALL_STATE(6848)] = 318635,
  [SMALL_STATE(6849)] = 318642,
  [SMALL_STATE(6850)] = 318649,
  [SMALL_STATE(6851)] = 318656,
  [SMALL_STATE(6852)] = 318663,
  [SMALL_STATE(6853)] = 318670,
  [SMALL_STATE(6854)] = 318677,
  [SMALL_STATE(6855)] = 318684,
  [SMALL_STATE(6856)] = 318691,
  [SMALL_STATE(6857)] = 318698,
  [SMALL_STATE(6858)] = 318705,
  [SMALL_STATE(6859)] = 318712,
  [SMALL_STATE(6860)] = 318719,
  [SMALL_STATE(6861)] = 318726,
  [SMALL_STATE(6862)] = 318733,
  [SMALL_STATE(6863)] = 318740,
  [SMALL_STATE(6864)] = 318747,
  [SMALL_STATE(6865)] = 318754,
  [SMALL_STATE(6866)] = 318761,
  [SMALL_STATE(6867)] = 318768,
  [SMALL_STATE(6868)] = 318775,
  [SMALL_STATE(6869)] = 318782,
  [SMALL_STATE(6870)] = 318789,
  [SMALL_STATE(6871)] = 318796,
  [SMALL_STATE(6872)] = 318803,
  [SMALL_STATE(6873)] = 318810,
  [SMALL_STATE(6874)] = 318817,
  [SMALL_STATE(6875)] = 318824,
  [SMALL_STATE(6876)] = 318831,
  [SMALL_STATE(6877)] = 318838,
  [SMALL_STATE(6878)] = 318845,
  [SMALL_STATE(6879)] = 318852,
  [SMALL_STATE(6880)] = 318859,
  [SMALL_STATE(6881)] = 318866,
  [SMALL_STATE(6882)] = 318873,
  [SMALL_STATE(6883)] = 318880,
  [SMALL_STATE(6884)] = 318887,
  [SMALL_STATE(6885)] = 318894,
  [SMALL_STATE(6886)] = 318901,
  [SMALL_STATE(6887)] = 318908,
  [SMALL_STATE(6888)] = 318915,
  [SMALL_STATE(6889)] = 318922,
  [SMALL_STATE(6890)] = 318929,
  [SMALL_STATE(6891)] = 318936,
  [SMALL_STATE(6892)] = 318943,
  [SMALL_STATE(6893)] = 318950,
  [SMALL_STATE(6894)] = 318957,
  [SMALL_STATE(6895)] = 318964,
  [SMALL_STATE(6896)] = 318971,
  [SMALL_STATE(6897)] = 318978,
  [SMALL_STATE(6898)] = 318985,
  [SMALL_STATE(6899)] = 318992,
  [SMALL_STATE(6900)] = 318999,
  [SMALL_STATE(6901)] = 319006,
  [SMALL_STATE(6902)] = 319013,
  [SMALL_STATE(6903)] = 319020,
  [SMALL_STATE(6904)] = 319027,
  [SMALL_STATE(6905)] = 319034,
  [SMALL_STATE(6906)] = 319041,
  [SMALL_STATE(6907)] = 319048,
  [SMALL_STATE(6908)] = 319055,
  [SMALL_STATE(6909)] = 319062,
  [SMALL_STATE(6910)] = 319069,
  [SMALL_STATE(6911)] = 319076,
  [SMALL_STATE(6912)] = 319083,
  [SMALL_STATE(6913)] = 319090,
  [SMALL_STATE(6914)] = 319097,
  [SMALL_STATE(6915)] = 319104,
  [SMALL_STATE(6916)] = 319111,
  [SMALL_STATE(6917)] = 319118,
  [SMALL_STATE(6918)] = 319125,
  [SMALL_STATE(6919)] = 319132,
  [SMALL_STATE(6920)] = 319139,
  [SMALL_STATE(6921)] = 319146,
  [SMALL_STATE(6922)] = 319153,
  [SMALL_STATE(6923)] = 319160,
  [SMALL_STATE(6924)] = 319167,
  [SMALL_STATE(6925)] = 319174,
  [SMALL_STATE(6926)] = 319181,
  [SMALL_STATE(6927)] = 319188,
  [SMALL_STATE(6928)] = 319195,
  [SMALL_STATE(6929)] = 319202,
  [SMALL_STATE(6930)] = 319209,
  [SMALL_STATE(6931)] = 319216,
  [SMALL_STATE(6932)] = 319223,
  [SMALL_STATE(6933)] = 319230,
  [SMALL_STATE(6934)] = 319237,
  [SMALL_STATE(6935)] = 319244,
  [SMALL_STATE(6936)] = 319251,
  [SMALL_STATE(6937)] = 319258,
  [SMALL_STATE(6938)] = 319265,
  [SMALL_STATE(6939)] = 319272,
  [SMALL_STATE(6940)] = 319279,
  [SMALL_STATE(6941)] = 319286,
  [SMALL_STATE(6942)] = 319293,
  [SMALL_STATE(6943)] = 319300,
  [SMALL_STATE(6944)] = 319307,
  [SMALL_STATE(6945)] = 319314,
  [SMALL_STATE(6946)] = 319321,
  [SMALL_STATE(6947)] = 319328,
  [SMALL_STATE(6948)] = 319335,
  [SMALL_STATE(6949)] = 319342,
  [SMALL_STATE(6950)] = 319349,
  [SMALL_STATE(6951)] = 319356,
  [SMALL_STATE(6952)] = 319363,
  [SMALL_STATE(6953)] = 319370,
  [SMALL_STATE(6954)] = 319377,
  [SMALL_STATE(6955)] = 319384,
  [SMALL_STATE(6956)] = 319391,
  [SMALL_STATE(6957)] = 319398,
  [SMALL_STATE(6958)] = 319405,
  [SMALL_STATE(6959)] = 319412,
  [SMALL_STATE(6960)] = 319419,
  [SMALL_STATE(6961)] = 319426,
  [SMALL_STATE(6962)] = 319433,
  [SMALL_STATE(6963)] = 319440,
  [SMALL_STATE(6964)] = 319447,
  [SMALL_STATE(6965)] = 319454,
  [SMALL_STATE(6966)] = 319461,
  [SMALL_STATE(6967)] = 319468,
  [SMALL_STATE(6968)] = 319475,
  [SMALL_STATE(6969)] = 319482,
  [SMALL_STATE(6970)] = 319489,
  [SMALL_STATE(6971)] = 319496,
  [SMALL_STATE(6972)] = 319503,
  [SMALL_STATE(6973)] = 319510,
  [SMALL_STATE(6974)] = 319517,
  [SMALL_STATE(6975)] = 319524,
  [SMALL_STATE(6976)] = 319531,
  [SMALL_STATE(6977)] = 319538,
  [SMALL_STATE(6978)] = 319545,
  [SMALL_STATE(6979)] = 319552,
  [SMALL_STATE(6980)] = 319559,
  [SMALL_STATE(6981)] = 319566,
  [SMALL_STATE(6982)] = 319573,
  [SMALL_STATE(6983)] = 319580,
  [SMALL_STATE(6984)] = 319587,
  [SMALL_STATE(6985)] = 319594,
  [SMALL_STATE(6986)] = 319601,
  [SMALL_STATE(6987)] = 319608,
  [SMALL_STATE(6988)] = 319615,
  [SMALL_STATE(6989)] = 319622,
  [SMALL_STATE(6990)] = 319629,
  [SMALL_STATE(6991)] = 319636,
  [SMALL_STATE(6992)] = 319643,
  [SMALL_STATE(6993)] = 319650,
  [SMALL_STATE(6994)] = 319657,
  [SMALL_STATE(6995)] = 319664,
  [SMALL_STATE(6996)] = 319671,
  [SMALL_STATE(6997)] = 319678,
  [SMALL_STATE(6998)] = 319685,
  [SMALL_STATE(6999)] = 319692,
  [SMALL_STATE(7000)] = 319699,
  [SMALL_STATE(7001)] = 319706,
  [SMALL_STATE(7002)] = 319713,
  [SMALL_STATE(7003)] = 319720,
  [SMALL_STATE(7004)] = 319727,
  [SMALL_STATE(7005)] = 319734,
  [SMALL_STATE(7006)] = 319741,
  [SMALL_STATE(7007)] = 319748,
  [SMALL_STATE(7008)] = 319755,
  [SMALL_STATE(7009)] = 319762,
  [SMALL_STATE(7010)] = 319769,
  [SMALL_STATE(7011)] = 319776,
  [SMALL_STATE(7012)] = 319783,
  [SMALL_STATE(7013)] = 319790,
  [SMALL_STATE(7014)] = 319797,
  [SMALL_STATE(7015)] = 319804,
  [SMALL_STATE(7016)] = 319811,
  [SMALL_STATE(7017)] = 319818,
  [SMALL_STATE(7018)] = 319825,
  [SMALL_STATE(7019)] = 319832,
  [SMALL_STATE(7020)] = 319839,
  [SMALL_STATE(7021)] = 319846,
  [SMALL_STATE(7022)] = 319853,
  [SMALL_STATE(7023)] = 319860,
  [SMALL_STATE(7024)] = 319867,
  [SMALL_STATE(7025)] = 319874,
  [SMALL_STATE(7026)] = 319881,
  [SMALL_STATE(7027)] = 319888,
  [SMALL_STATE(7028)] = 319895,
  [SMALL_STATE(7029)] = 319902,
  [SMALL_STATE(7030)] = 319909,
  [SMALL_STATE(7031)] = 319916,
  [SMALL_STATE(7032)] = 319923,
  [SMALL_STATE(7033)] = 319930,
  [SMALL_STATE(7034)] = 319937,
  [SMALL_STATE(7035)] = 319944,
  [SMALL_STATE(7036)] = 319951,
  [SMALL_STATE(7037)] = 319958,
  [SMALL_STATE(7038)] = 319965,
  [SMALL_STATE(7039)] = 319972,
  [SMALL_STATE(7040)] = 319979,
  [SMALL_STATE(7041)] = 319986,
  [SMALL_STATE(7042)] = 319993,
  [SMALL_STATE(7043)] = 320000,
  [SMALL_STATE(7044)] = 320007,
  [SMALL_STATE(7045)] = 320014,
  [SMALL_STATE(7046)] = 320021,
  [SMALL_STATE(7047)] = 320028,
  [SMALL_STATE(7048)] = 320035,
  [SMALL_STATE(7049)] = 320042,
  [SMALL_STATE(7050)] = 320049,
  [SMALL_STATE(7051)] = 320056,
  [SMALL_STATE(7052)] = 320063,
  [SMALL_STATE(7053)] = 320070,
  [SMALL_STATE(7054)] = 320077,
  [SMALL_STATE(7055)] = 320084,
  [SMALL_STATE(7056)] = 320091,
  [SMALL_STATE(7057)] = 320098,
  [SMALL_STATE(7058)] = 320105,
  [SMALL_STATE(7059)] = 320112,
  [SMALL_STATE(7060)] = 320119,
  [SMALL_STATE(7061)] = 320126,
  [SMALL_STATE(7062)] = 320133,
  [SMALL_STATE(7063)] = 320140,
  [SMALL_STATE(7064)] = 320147,
  [SMALL_STATE(7065)] = 320154,
  [SMALL_STATE(7066)] = 320161,
  [SMALL_STATE(7067)] = 320168,
  [SMALL_STATE(7068)] = 320175,
  [SMALL_STATE(7069)] = 320182,
  [SMALL_STATE(7070)] = 320189,
  [SMALL_STATE(7071)] = 320196,
  [SMALL_STATE(7072)] = 320203,
  [SMALL_STATE(7073)] = 320210,
  [SMALL_STATE(7074)] = 320217,
  [SMALL_STATE(7075)] = 320224,
  [SMALL_STATE(7076)] = 320231,
  [SMALL_STATE(7077)] = 320238,
  [SMALL_STATE(7078)] = 320245,
  [SMALL_STATE(7079)] = 320252,
  [SMALL_STATE(7080)] = 320259,
  [SMALL_STATE(7081)] = 320266,
  [SMALL_STATE(7082)] = 320273,
  [SMALL_STATE(7083)] = 320280,
  [SMALL_STATE(7084)] = 320287,
  [SMALL_STATE(7085)] = 320294,
  [SMALL_STATE(7086)] = 320301,
  [SMALL_STATE(7087)] = 320308,
  [SMALL_STATE(7088)] = 320315,
  [SMALL_STATE(7089)] = 320322,
  [SMALL_STATE(7090)] = 320329,
  [SMALL_STATE(7091)] = 320336,
  [SMALL_STATE(7092)] = 320343,
  [SMALL_STATE(7093)] = 320350,
  [SMALL_STATE(7094)] = 320357,
  [SMALL_STATE(7095)] = 320364,
  [SMALL_STATE(7096)] = 320371,
  [SMALL_STATE(7097)] = 320378,
  [SMALL_STATE(7098)] = 320385,
  [SMALL_STATE(7099)] = 320392,
  [SMALL_STATE(7100)] = 320399,
  [SMALL_STATE(7101)] = 320406,
  [SMALL_STATE(7102)] = 320413,
  [SMALL_STATE(7103)] = 320420,
  [SMALL_STATE(7104)] = 320427,
  [SMALL_STATE(7105)] = 320434,
  [SMALL_STATE(7106)] = 320441,
  [SMALL_STATE(7107)] = 320448,
  [SMALL_STATE(7108)] = 320455,
  [SMALL_STATE(7109)] = 320462,
  [SMALL_STATE(7110)] = 320469,
  [SMALL_STATE(7111)] = 320476,
  [SMALL_STATE(7112)] = 320483,
  [SMALL_STATE(7113)] = 320490,
  [SMALL_STATE(7114)] = 320497,
  [SMALL_STATE(7115)] = 320504,
  [SMALL_STATE(7116)] = 320511,
  [SMALL_STATE(7117)] = 320518,
  [SMALL_STATE(7118)] = 320525,
  [SMALL_STATE(7119)] = 320532,
  [SMALL_STATE(7120)] = 320539,
  [SMALL_STATE(7121)] = 320546,
  [SMALL_STATE(7122)] = 320553,
  [SMALL_STATE(7123)] = 320560,
  [SMALL_STATE(7124)] = 320567,
  [SMALL_STATE(7125)] = 320574,
  [SMALL_STATE(7126)] = 320581,
  [SMALL_STATE(7127)] = 320588,
  [SMALL_STATE(7128)] = 320595,
  [SMALL_STATE(7129)] = 320602,
  [SMALL_STATE(7130)] = 320609,
  [SMALL_STATE(7131)] = 320616,
  [SMALL_STATE(7132)] = 320623,
  [SMALL_STATE(7133)] = 320630,
  [SMALL_STATE(7134)] = 320637,
  [SMALL_STATE(7135)] = 320644,
  [SMALL_STATE(7136)] = 320651,
  [SMALL_STATE(7137)] = 320658,
  [SMALL_STATE(7138)] = 320665,
  [SMALL_STATE(7139)] = 320672,
  [SMALL_STATE(7140)] = 320679,
  [SMALL_STATE(7141)] = 320686,
  [SMALL_STATE(7142)] = 320693,
  [SMALL_STATE(7143)] = 320700,
  [SMALL_STATE(7144)] = 320707,
  [SMALL_STATE(7145)] = 320714,
  [SMALL_STATE(7146)] = 320721,
  [SMALL_STATE(7147)] = 320728,
  [SMALL_STATE(7148)] = 320735,
  [SMALL_STATE(7149)] = 320742,
  [SMALL_STATE(7150)] = 320749,
  [SMALL_STATE(7151)] = 320756,
  [SMALL_STATE(7152)] = 320763,
  [SMALL_STATE(7153)] = 320770,
  [SMALL_STATE(7154)] = 320777,
  [SMALL_STATE(7155)] = 320784,
  [SMALL_STATE(7156)] = 320791,
  [SMALL_STATE(7157)] = 320798,
  [SMALL_STATE(7158)] = 320805,
  [SMALL_STATE(7159)] = 320812,
  [SMALL_STATE(7160)] = 320819,
  [SMALL_STATE(7161)] = 320826,
  [SMALL_STATE(7162)] = 320833,
  [SMALL_STATE(7163)] = 320840,
  [SMALL_STATE(7164)] = 320847,
  [SMALL_STATE(7165)] = 320854,
  [SMALL_STATE(7166)] = 320861,
  [SMALL_STATE(7167)] = 320868,
  [SMALL_STATE(7168)] = 320875,
  [SMALL_STATE(7169)] = 320882,
  [SMALL_STATE(7170)] = 320889,
  [SMALL_STATE(7171)] = 320896,
  [SMALL_STATE(7172)] = 320903,
  [SMALL_STATE(7173)] = 320910,
  [SMALL_STATE(7174)] = 320917,
  [SMALL_STATE(7175)] = 320924,
  [SMALL_STATE(7176)] = 320931,
  [SMALL_STATE(7177)] = 320938,
  [SMALL_STATE(7178)] = 320945,
  [SMALL_STATE(7179)] = 320952,
  [SMALL_STATE(7180)] = 320959,
  [SMALL_STATE(7181)] = 320966,
  [SMALL_STATE(7182)] = 320973,
  [SMALL_STATE(7183)] = 320980,
  [SMALL_STATE(7184)] = 320987,
  [SMALL_STATE(7185)] = 320994,
  [SMALL_STATE(7186)] = 321001,
  [SMALL_STATE(7187)] = 321008,
  [SMALL_STATE(7188)] = 321015,
  [SMALL_STATE(7189)] = 321022,
  [SMALL_STATE(7190)] = 321029,
  [SMALL_STATE(7191)] = 321036,
  [SMALL_STATE(7192)] = 321043,
  [SMALL_STATE(7193)] = 321050,
  [SMALL_STATE(7194)] = 321057,
  [SMALL_STATE(7195)] = 321064,
  [SMALL_STATE(7196)] = 321071,
  [SMALL_STATE(7197)] = 321078,
  [SMALL_STATE(7198)] = 321085,
  [SMALL_STATE(7199)] = 321092,
  [SMALL_STATE(7200)] = 321099,
  [SMALL_STATE(7201)] = 321106,
  [SMALL_STATE(7202)] = 321113,
  [SMALL_STATE(7203)] = 321120,
  [SMALL_STATE(7204)] = 321127,
  [SMALL_STATE(7205)] = 321134,
  [SMALL_STATE(7206)] = 321141,
  [SMALL_STATE(7207)] = 321148,
  [SMALL_STATE(7208)] = 321155,
  [SMALL_STATE(7209)] = 321162,
  [SMALL_STATE(7210)] = 321169,
  [SMALL_STATE(7211)] = 321176,
  [SMALL_STATE(7212)] = 321183,
  [SMALL_STATE(7213)] = 321190,
  [SMALL_STATE(7214)] = 321197,
  [SMALL_STATE(7215)] = 321204,
  [SMALL_STATE(7216)] = 321211,
  [SMALL_STATE(7217)] = 321218,
  [SMALL_STATE(7218)] = 321225,
  [SMALL_STATE(7219)] = 321232,
  [SMALL_STATE(7220)] = 321239,
  [SMALL_STATE(7221)] = 321246,
  [SMALL_STATE(7222)] = 321253,
  [SMALL_STATE(7223)] = 321260,
  [SMALL_STATE(7224)] = 321267,
  [SMALL_STATE(7225)] = 321274,
  [SMALL_STATE(7226)] = 321281,
  [SMALL_STATE(7227)] = 321288,
  [SMALL_STATE(7228)] = 321295,
  [SMALL_STATE(7229)] = 321302,
  [SMALL_STATE(7230)] = 321309,
  [SMALL_STATE(7231)] = 321316,
  [SMALL_STATE(7232)] = 321323,
  [SMALL_STATE(7233)] = 321330,
  [SMALL_STATE(7234)] = 321337,
  [SMALL_STATE(7235)] = 321344,
  [SMALL_STATE(7236)] = 321351,
  [SMALL_STATE(7237)] = 321358,
  [SMALL_STATE(7238)] = 321365,
  [SMALL_STATE(7239)] = 321372,
  [SMALL_STATE(7240)] = 321379,
  [SMALL_STATE(7241)] = 321386,
  [SMALL_STATE(7242)] = 321393,
  [SMALL_STATE(7243)] = 321400,
  [SMALL_STATE(7244)] = 321407,
  [SMALL_STATE(7245)] = 321414,
  [SMALL_STATE(7246)] = 321421,
  [SMALL_STATE(7247)] = 321428,
  [SMALL_STATE(7248)] = 321435,
  [SMALL_STATE(7249)] = 321442,
  [SMALL_STATE(7250)] = 321449,
  [SMALL_STATE(7251)] = 321456,
  [SMALL_STATE(7252)] = 321463,
  [SMALL_STATE(7253)] = 321470,
  [SMALL_STATE(7254)] = 321477,
  [SMALL_STATE(7255)] = 321484,
  [SMALL_STATE(7256)] = 321491,
  [SMALL_STATE(7257)] = 321498,
  [SMALL_STATE(7258)] = 321505,
  [SMALL_STATE(7259)] = 321512,
  [SMALL_STATE(7260)] = 321519,
  [SMALL_STATE(7261)] = 321526,
  [SMALL_STATE(7262)] = 321533,
  [SMALL_STATE(7263)] = 321540,
  [SMALL_STATE(7264)] = 321547,
  [SMALL_STATE(7265)] = 321554,
  [SMALL_STATE(7266)] = 321561,
  [SMALL_STATE(7267)] = 321568,
  [SMALL_STATE(7268)] = 321575,
  [SMALL_STATE(7269)] = 321582,
  [SMALL_STATE(7270)] = 321589,
  [SMALL_STATE(7271)] = 321596,
  [SMALL_STATE(7272)] = 321603,
  [SMALL_STATE(7273)] = 321610,
  [SMALL_STATE(7274)] = 321617,
  [SMALL_STATE(7275)] = 321624,
  [SMALL_STATE(7276)] = 321631,
  [SMALL_STATE(7277)] = 321638,
  [SMALL_STATE(7278)] = 321645,
  [SMALL_STATE(7279)] = 321652,
  [SMALL_STATE(7280)] = 321659,
  [SMALL_STATE(7281)] = 321666,
  [SMALL_STATE(7282)] = 321673,
  [SMALL_STATE(7283)] = 321680,
  [SMALL_STATE(7284)] = 321687,
  [SMALL_STATE(7285)] = 321694,
  [SMALL_STATE(7286)] = 321701,
  [SMALL_STATE(7287)] = 321708,
  [SMALL_STATE(7288)] = 321715,
  [SMALL_STATE(7289)] = 321722,
  [SMALL_STATE(7290)] = 321729,
  [SMALL_STATE(7291)] = 321736,
  [SMALL_STATE(7292)] = 321743,
  [SMALL_STATE(7293)] = 321750,
  [SMALL_STATE(7294)] = 321757,
  [SMALL_STATE(7295)] = 321764,
  [SMALL_STATE(7296)] = 321771,
  [SMALL_STATE(7297)] = 321778,
  [SMALL_STATE(7298)] = 321785,
  [SMALL_STATE(7299)] = 321792,
  [SMALL_STATE(7300)] = 321799,
  [SMALL_STATE(7301)] = 321806,
  [SMALL_STATE(7302)] = 321813,
  [SMALL_STATE(7303)] = 321820,
  [SMALL_STATE(7304)] = 321827,
  [SMALL_STATE(7305)] = 321834,
  [SMALL_STATE(7306)] = 321841,
  [SMALL_STATE(7307)] = 321848,
  [SMALL_STATE(7308)] = 321855,
  [SMALL_STATE(7309)] = 321862,
  [SMALL_STATE(7310)] = 321869,
  [SMALL_STATE(7311)] = 321876,
  [SMALL_STATE(7312)] = 321883,
  [SMALL_STATE(7313)] = 321890,
  [SMALL_STATE(7314)] = 321897,
  [SMALL_STATE(7315)] = 321904,
  [SMALL_STATE(7316)] = 321911,
  [SMALL_STATE(7317)] = 321918,
  [SMALL_STATE(7318)] = 321925,
  [SMALL_STATE(7319)] = 321932,
  [SMALL_STATE(7320)] = 321939,
  [SMALL_STATE(7321)] = 321946,
  [SMALL_STATE(7322)] = 321953,
  [SMALL_STATE(7323)] = 321960,
  [SMALL_STATE(7324)] = 321967,
  [SMALL_STATE(7325)] = 321974,
  [SMALL_STATE(7326)] = 321981,
  [SMALL_STATE(7327)] = 321988,
  [SMALL_STATE(7328)] = 321995,
  [SMALL_STATE(7329)] = 322002,
  [SMALL_STATE(7330)] = 322009,
  [SMALL_STATE(7331)] = 322016,
  [SMALL_STATE(7332)] = 322023,
  [SMALL_STATE(7333)] = 322030,
  [SMALL_STATE(7334)] = 322037,
  [SMALL_STATE(7335)] = 322044,
  [SMALL_STATE(7336)] = 322051,
  [SMALL_STATE(7337)] = 322058,
  [SMALL_STATE(7338)] = 322065,
  [SMALL_STATE(7339)] = 322072,
  [SMALL_STATE(7340)] = 322079,
  [SMALL_STATE(7341)] = 322086,
  [SMALL_STATE(7342)] = 322093,
  [SMALL_STATE(7343)] = 322100,
  [SMALL_STATE(7344)] = 322107,
  [SMALL_STATE(7345)] = 322114,
  [SMALL_STATE(7346)] = 322121,
  [SMALL_STATE(7347)] = 322128,
  [SMALL_STATE(7348)] = 322135,
  [SMALL_STATE(7349)] = 322142,
  [SMALL_STATE(7350)] = 322149,
  [SMALL_STATE(7351)] = 322156,
  [SMALL_STATE(7352)] = 322163,
  [SMALL_STATE(7353)] = 322170,
  [SMALL_STATE(7354)] = 322177,
  [SMALL_STATE(7355)] = 322184,
  [SMALL_STATE(7356)] = 322191,
  [SMALL_STATE(7357)] = 322198,
  [SMALL_STATE(7358)] = 322205,
  [SMALL_STATE(7359)] = 322212,
  [SMALL_STATE(7360)] = 322219,
  [SMALL_STATE(7361)] = 322226,
  [SMALL_STATE(7362)] = 322233,
  [SMALL_STATE(7363)] = 322240,
  [SMALL_STATE(7364)] = 322247,
  [SMALL_STATE(7365)] = 322254,
  [SMALL_STATE(7366)] = 322261,
  [SMALL_STATE(7367)] = 322268,
  [SMALL_STATE(7368)] = 322275,
  [SMALL_STATE(7369)] = 322282,
  [SMALL_STATE(7370)] = 322289,
  [SMALL_STATE(7371)] = 322296,
  [SMALL_STATE(7372)] = 322303,
  [SMALL_STATE(7373)] = 322310,
  [SMALL_STATE(7374)] = 322317,
  [SMALL_STATE(7375)] = 322324,
  [SMALL_STATE(7376)] = 322331,
  [SMALL_STATE(7377)] = 322338,
  [SMALL_STATE(7378)] = 322345,
  [SMALL_STATE(7379)] = 322352,
  [SMALL_STATE(7380)] = 322359,
  [SMALL_STATE(7381)] = 322366,
  [SMALL_STATE(7382)] = 322373,
  [SMALL_STATE(7383)] = 322380,
  [SMALL_STATE(7384)] = 322387,
  [SMALL_STATE(7385)] = 322394,
  [SMALL_STATE(7386)] = 322401,
  [SMALL_STATE(7387)] = 322408,
  [SMALL_STATE(7388)] = 322415,
  [SMALL_STATE(7389)] = 322422,
  [SMALL_STATE(7390)] = 322429,
  [SMALL_STATE(7391)] = 322436,
  [SMALL_STATE(7392)] = 322443,
  [SMALL_STATE(7393)] = 322450,
  [SMALL_STATE(7394)] = 322457,
  [SMALL_STATE(7395)] = 322464,
  [SMALL_STATE(7396)] = 322471,
  [SMALL_STATE(7397)] = 322478,
  [SMALL_STATE(7398)] = 322485,
  [SMALL_STATE(7399)] = 322492,
  [SMALL_STATE(7400)] = 322499,
  [SMALL_STATE(7401)] = 322506,
  [SMALL_STATE(7402)] = 322513,
  [SMALL_STATE(7403)] = 322520,
  [SMALL_STATE(7404)] = 322527,
  [SMALL_STATE(7405)] = 322534,
  [SMALL_STATE(7406)] = 322541,
  [SMALL_STATE(7407)] = 322548,
  [SMALL_STATE(7408)] = 322555,
  [SMALL_STATE(7409)] = 322562,
  [SMALL_STATE(7410)] = 322569,
  [SMALL_STATE(7411)] = 322576,
  [SMALL_STATE(7412)] = 322583,
  [SMALL_STATE(7413)] = 322590,
  [SMALL_STATE(7414)] = 322597,
  [SMALL_STATE(7415)] = 322604,
  [SMALL_STATE(7416)] = 322611,
  [SMALL_STATE(7417)] = 322618,
  [SMALL_STATE(7418)] = 322625,
  [SMALL_STATE(7419)] = 322632,
  [SMALL_STATE(7420)] = 322639,
  [SMALL_STATE(7421)] = 322646,
  [SMALL_STATE(7422)] = 322653,
  [SMALL_STATE(7423)] = 322660,
  [SMALL_STATE(7424)] = 322667,
  [SMALL_STATE(7425)] = 322674,
  [SMALL_STATE(7426)] = 322681,
  [SMALL_STATE(7427)] = 322688,
  [SMALL_STATE(7428)] = 322695,
  [SMALL_STATE(7429)] = 322702,
  [SMALL_STATE(7430)] = 322709,
  [SMALL_STATE(7431)] = 322716,
  [SMALL_STATE(7432)] = 322723,
  [SMALL_STATE(7433)] = 322730,
  [SMALL_STATE(7434)] = 322737,
  [SMALL_STATE(7435)] = 322744,
  [SMALL_STATE(7436)] = 322751,
  [SMALL_STATE(7437)] = 322758,
  [SMALL_STATE(7438)] = 322765,
  [SMALL_STATE(7439)] = 322772,
  [SMALL_STATE(7440)] = 322779,
  [SMALL_STATE(7441)] = 322786,
  [SMALL_STATE(7442)] = 322793,
  [SMALL_STATE(7443)] = 322800,
  [SMALL_STATE(7444)] = 322807,
  [SMALL_STATE(7445)] = 322814,
  [SMALL_STATE(7446)] = 322821,
  [SMALL_STATE(7447)] = 322828,
  [SMALL_STATE(7448)] = 322835,
  [SMALL_STATE(7449)] = 322842,
  [SMALL_STATE(7450)] = 322849,
  [SMALL_STATE(7451)] = 322856,
  [SMALL_STATE(7452)] = 322863,
  [SMALL_STATE(7453)] = 322870,
  [SMALL_STATE(7454)] = 322877,
  [SMALL_STATE(7455)] = 322884,
  [SMALL_STATE(7456)] = 322891,
  [SMALL_STATE(7457)] = 322898,
  [SMALL_STATE(7458)] = 322905,
  [SMALL_STATE(7459)] = 322912,
  [SMALL_STATE(7460)] = 322919,
  [SMALL_STATE(7461)] = 322926,
  [SMALL_STATE(7462)] = 322933,
  [SMALL_STATE(7463)] = 322940,
  [SMALL_STATE(7464)] = 322947,
  [SMALL_STATE(7465)] = 322954,
  [SMALL_STATE(7466)] = 322961,
  [SMALL_STATE(7467)] = 322968,
  [SMALL_STATE(7468)] = 322975,
  [SMALL_STATE(7469)] = 322982,
  [SMALL_STATE(7470)] = 322989,
  [SMALL_STATE(7471)] = 322996,
  [SMALL_STATE(7472)] = 323003,
  [SMALL_STATE(7473)] = 323010,
  [SMALL_STATE(7474)] = 323017,
  [SMALL_STATE(7475)] = 323024,
  [SMALL_STATE(7476)] = 323031,
  [SMALL_STATE(7477)] = 323038,
  [SMALL_STATE(7478)] = 323045,
  [SMALL_STATE(7479)] = 323052,
  [SMALL_STATE(7480)] = 323059,
  [SMALL_STATE(7481)] = 323066,
  [SMALL_STATE(7482)] = 323073,
  [SMALL_STATE(7483)] = 323080,
  [SMALL_STATE(7484)] = 323087,
  [SMALL_STATE(7485)] = 323094,
  [SMALL_STATE(7486)] = 323101,
  [SMALL_STATE(7487)] = 323108,
  [SMALL_STATE(7488)] = 323115,
  [SMALL_STATE(7489)] = 323122,
  [SMALL_STATE(7490)] = 323129,
  [SMALL_STATE(7491)] = 323136,
  [SMALL_STATE(7492)] = 323143,
  [SMALL_STATE(7493)] = 323150,
  [SMALL_STATE(7494)] = 323157,
  [SMALL_STATE(7495)] = 323164,
  [SMALL_STATE(7496)] = 323171,
  [SMALL_STATE(7497)] = 323178,
  [SMALL_STATE(7498)] = 323185,
  [SMALL_STATE(7499)] = 323192,
  [SMALL_STATE(7500)] = 323199,
  [SMALL_STATE(7501)] = 323206,
  [SMALL_STATE(7502)] = 323213,
  [SMALL_STATE(7503)] = 323220,
  [SMALL_STATE(7504)] = 323227,
  [SMALL_STATE(7505)] = 323234,
  [SMALL_STATE(7506)] = 323241,
  [SMALL_STATE(7507)] = 323248,
  [SMALL_STATE(7508)] = 323255,
  [SMALL_STATE(7509)] = 323262,
  [SMALL_STATE(7510)] = 323269,
  [SMALL_STATE(7511)] = 323276,
  [SMALL_STATE(7512)] = 323283,
  [SMALL_STATE(7513)] = 323290,
  [SMALL_STATE(7514)] = 323297,
  [SMALL_STATE(7515)] = 323304,
  [SMALL_STATE(7516)] = 323311,
  [SMALL_STATE(7517)] = 323318,
  [SMALL_STATE(7518)] = 323325,
  [SMALL_STATE(7519)] = 323332,
  [SMALL_STATE(7520)] = 323339,
  [SMALL_STATE(7521)] = 323346,
  [SMALL_STATE(7522)] = 323353,
  [SMALL_STATE(7523)] = 323360,
  [SMALL_STATE(7524)] = 323367,
  [SMALL_STATE(7525)] = 323374,
  [SMALL_STATE(7526)] = 323381,
  [SMALL_STATE(7527)] = 323388,
  [SMALL_STATE(7528)] = 323395,
  [SMALL_STATE(7529)] = 323402,
  [SMALL_STATE(7530)] = 323409,
  [SMALL_STATE(7531)] = 323416,
  [SMALL_STATE(7532)] = 323423,
  [SMALL_STATE(7533)] = 323430,
  [SMALL_STATE(7534)] = 323437,
  [SMALL_STATE(7535)] = 323444,
  [SMALL_STATE(7536)] = 323451,
  [SMALL_STATE(7537)] = 323458,
  [SMALL_STATE(7538)] = 323465,
  [SMALL_STATE(7539)] = 323472,
  [SMALL_STATE(7540)] = 323479,
  [SMALL_STATE(7541)] = 323486,
  [SMALL_STATE(7542)] = 323493,
  [SMALL_STATE(7543)] = 323500,
  [SMALL_STATE(7544)] = 323507,
  [SMALL_STATE(7545)] = 323514,
  [SMALL_STATE(7546)] = 323521,
  [SMALL_STATE(7547)] = 323528,
  [SMALL_STATE(7548)] = 323535,
  [SMALL_STATE(7549)] = 323542,
  [SMALL_STATE(7550)] = 323549,
  [SMALL_STATE(7551)] = 323556,
  [SMALL_STATE(7552)] = 323563,
  [SMALL_STATE(7553)] = 323570,
  [SMALL_STATE(7554)] = 323577,
  [SMALL_STATE(7555)] = 323584,
  [SMALL_STATE(7556)] = 323591,
  [SMALL_STATE(7557)] = 323598,
  [SMALL_STATE(7558)] = 323605,
  [SMALL_STATE(7559)] = 323612,
  [SMALL_STATE(7560)] = 323619,
  [SMALL_STATE(7561)] = 323626,
  [SMALL_STATE(7562)] = 323633,
  [SMALL_STATE(7563)] = 323640,
  [SMALL_STATE(7564)] = 323647,
  [SMALL_STATE(7565)] = 323654,
  [SMALL_STATE(7566)] = 323661,
  [SMALL_STATE(7567)] = 323668,
  [SMALL_STATE(7568)] = 323675,
  [SMALL_STATE(7569)] = 323682,
  [SMALL_STATE(7570)] = 323689,
  [SMALL_STATE(7571)] = 323696,
  [SMALL_STATE(7572)] = 323703,
  [SMALL_STATE(7573)] = 323710,
  [SMALL_STATE(7574)] = 323717,
  [SMALL_STATE(7575)] = 323724,
  [SMALL_STATE(7576)] = 323731,
  [SMALL_STATE(7577)] = 323738,
  [SMALL_STATE(7578)] = 323745,
  [SMALL_STATE(7579)] = 323752,
  [SMALL_STATE(7580)] = 323759,
  [SMALL_STATE(7581)] = 323766,
  [SMALL_STATE(7582)] = 323773,
  [SMALL_STATE(7583)] = 323780,
  [SMALL_STATE(7584)] = 323787,
  [SMALL_STATE(7585)] = 323794,
  [SMALL_STATE(7586)] = 323801,
  [SMALL_STATE(7587)] = 323808,
  [SMALL_STATE(7588)] = 323815,
  [SMALL_STATE(7589)] = 323822,
  [SMALL_STATE(7590)] = 323829,
  [SMALL_STATE(7591)] = 323836,
  [SMALL_STATE(7592)] = 323843,
  [SMALL_STATE(7593)] = 323850,
  [SMALL_STATE(7594)] = 323857,
  [SMALL_STATE(7595)] = 323864,
  [SMALL_STATE(7596)] = 323871,
  [SMALL_STATE(7597)] = 323878,
  [SMALL_STATE(7598)] = 323885,
  [SMALL_STATE(7599)] = 323892,
  [SMALL_STATE(7600)] = 323899,
  [SMALL_STATE(7601)] = 323906,
  [SMALL_STATE(7602)] = 323913,
  [SMALL_STATE(7603)] = 323920,
  [SMALL_STATE(7604)] = 323927,
  [SMALL_STATE(7605)] = 323934,
  [SMALL_STATE(7606)] = 323941,
  [SMALL_STATE(7607)] = 323948,
  [SMALL_STATE(7608)] = 323955,
  [SMALL_STATE(7609)] = 323962,
  [SMALL_STATE(7610)] = 323969,
  [SMALL_STATE(7611)] = 323976,
  [SMALL_STATE(7612)] = 323983,
  [SMALL_STATE(7613)] = 323990,
  [SMALL_STATE(7614)] = 323997,
  [SMALL_STATE(7615)] = 324004,
  [SMALL_STATE(7616)] = 324011,
  [SMALL_STATE(7617)] = 324018,
  [SMALL_STATE(7618)] = 324025,
  [SMALL_STATE(7619)] = 324032,
  [SMALL_STATE(7620)] = 324039,
  [SMALL_STATE(7621)] = 324046,
  [SMALL_STATE(7622)] = 324053,
  [SMALL_STATE(7623)] = 324060,
  [SMALL_STATE(7624)] = 324067,
  [SMALL_STATE(7625)] = 324074,
  [SMALL_STATE(7626)] = 324081,
  [SMALL_STATE(7627)] = 324088,
  [SMALL_STATE(7628)] = 324095,
  [SMALL_STATE(7629)] = 324102,
  [SMALL_STATE(7630)] = 324109,
  [SMALL_STATE(7631)] = 324116,
  [SMALL_STATE(7632)] = 324123,
  [SMALL_STATE(7633)] = 324130,
  [SMALL_STATE(7634)] = 324137,
  [SMALL_STATE(7635)] = 324144,
  [SMALL_STATE(7636)] = 324151,
  [SMALL_STATE(7637)] = 324158,
  [SMALL_STATE(7638)] = 324165,
  [SMALL_STATE(7639)] = 324172,
  [SMALL_STATE(7640)] = 324179,
  [SMALL_STATE(7641)] = 324186,
  [SMALL_STATE(7642)] = 324193,
  [SMALL_STATE(7643)] = 324200,
  [SMALL_STATE(7644)] = 324207,
  [SMALL_STATE(7645)] = 324214,
  [SMALL_STATE(7646)] = 324221,
  [SMALL_STATE(7647)] = 324228,
  [SMALL_STATE(7648)] = 324235,
  [SMALL_STATE(7649)] = 324242,
  [SMALL_STATE(7650)] = 324249,
  [SMALL_STATE(7651)] = 324256,
  [SMALL_STATE(7652)] = 324263,
  [SMALL_STATE(7653)] = 324270,
  [SMALL_STATE(7654)] = 324277,
  [SMALL_STATE(7655)] = 324284,
  [SMALL_STATE(7656)] = 324291,
  [SMALL_STATE(7657)] = 324298,
  [SMALL_STATE(7658)] = 324305,
  [SMALL_STATE(7659)] = 324312,
  [SMALL_STATE(7660)] = 324319,
  [SMALL_STATE(7661)] = 324326,
  [SMALL_STATE(7662)] = 324333,
  [SMALL_STATE(7663)] = 324340,
  [SMALL_STATE(7664)] = 324347,
  [SMALL_STATE(7665)] = 324354,
  [SMALL_STATE(7666)] = 324361,
  [SMALL_STATE(7667)] = 324368,
  [SMALL_STATE(7668)] = 324375,
  [SMALL_STATE(7669)] = 324382,
  [SMALL_STATE(7670)] = 324389,
};

static const TSParseActionEntry ts_parse_actions[] = {
  [0] = {.entry = {.count = 0, .reusable = false}},
  [1] = {.entry = {.count = 1, .reusable = false}}, RECOVER(),
  [3] = {.entry = {.count = 1, .reusable = false}}, SHIFT_EXTRA(),
  [5] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_program, 0),
  [7] = {.entry = {.count = 1, .reusable = false}}, SHIFT(907),
  [9] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6777),
  [11] = {.entry = {.count = 1, .reusable = false}}, SHIFT(7604),
  [13] = {.entry = {.count = 1, .reusable = true}}, SHIFT(661),
  [15] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3690),
  [17] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3690),
  [19] = {.entry = {.count = 1, .reusable = false}}, SHIFT(240),
  [21] = {.entry = {.count = 1, .reusable = false}}, SHIFT(432),
  [23] = {.entry = {.count = 1, .reusable = false}}, SHIFT(428),
  [25] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4078),
  [27] = {.entry = {.count = 1, .reusable = false}}, SHIFT(7516),
  [29] = {.entry = {.count = 1, .reusable = true}}, SHIFT(154),
  [31] = {.entry = {.count = 1, .reusable = false}}, SHIFT(618),
  [33] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4),
  [35] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1730),
  [37] = {.entry = {.count = 1, .reusable = false}}, SHIFT(576),
  [39] = {.entry = {.count = 1, .reusable = false}}, SHIFT(690),
  [41] = {.entry = {.count = 1, .reusable = true}}, SHIFT(682),
  [43] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3808),
  [45] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3830),
  [47] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3831),
  [49] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5949),
  [51] = {.entry = {.count = 1, .reusable = false}}, SHIFT(966),
  [53] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5757),
  [55] = {.entry = {.count = 1, .reusable = true}}, SHIFT(957),
  [57] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1104),
  [59] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6358),
  [61] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4714),
  [63] = {.entry = {.count = 1, .reusable = false}}, SHIFT(64),
  [65] = {.entry = {.count = 1, .reusable = true}}, SHIFT(65),
  [67] = {.entry = {.count = 1, .reusable = true}}, SHIFT(66),
  [69] = {.entry = {.count = 1, .reusable = true}}, SHIFT(67),
  [71] = {.entry = {.count = 1, .reusable = true}}, SHIFT_EXTRA(),
  [73] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6144),
  [75] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6575),
  [77] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1007),
  [79] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7607),
  [81] = {.entry = {.count = 1, .reusable = false}}, SHIFT(471),
  [83] = {.entry = {.count = 1, .reusable = true}}, SHIFT(697),
  [85] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3637),
  [87] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3637),
  [89] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2),
  [91] = {.entry = {.count = 1, .reusable = false}}, SHIFT(7268),
  [93] = {.entry = {.count = 1, .reusable = false}}, SHIFT(526),
  [95] = {.entry = {.count = 1, .reusable = false}}, SHIFT(586),
  [97] = {.entry = {.count = 1, .reusable = false}}, SHIFT(693),
  [99] = {.entry = {.count = 1, .reusable = true}}, SHIFT(694),
  [101] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4115),
  [103] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2087),
  [105] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2089),
  [107] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2089),
  [109] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4136),
  [111] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4135),
  [113] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5907),
  [115] = {.entry = {.count = 1, .reusable = false}}, SHIFT(469),
  [117] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5800),
  [119] = {.entry = {.count = 1, .reusable = true}}, SHIFT(472),
  [121] = {.entry = {.count = 1, .reusable = false}}, SHIFT(480),
  [123] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6290),
  [125] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4713),
  [127] = {.entry = {.count = 1, .reusable = false}}, SHIFT(266),
  [129] = {.entry = {.count = 1, .reusable = true}}, SHIFT(267),
  [131] = {.entry = {.count = 1, .reusable = true}}, SHIFT(271),
  [133] = {.entry = {.count = 1, .reusable = true}}, SHIFT(242),
  [135] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6119),
  [137] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6549),
  [139] = {.entry = {.count = 1, .reusable = true}}, SHIFT(12),
  [141] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7622),
  [143] = {.entry = {.count = 1, .reusable = false}}, SHIFT(497),
  [145] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6752),
  [147] = {.entry = {.count = 1, .reusable = false}}, SHIFT(7265),
  [149] = {.entry = {.count = 1, .reusable = true}}, SHIFT(685),
  [151] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3665),
  [153] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3665),
  [155] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3),
  [157] = {.entry = {.count = 1, .reusable = false}}, SHIFT(420),
  [159] = {.entry = {.count = 1, .reusable = false}}, SHIFT(425),
  [161] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4018),
  [163] = {.entry = {.count = 1, .reusable = false}}, SHIFT(7068),
  [165] = {.entry = {.count = 1, .reusable = true}}, SHIFT(284),
  [167] = {.entry = {.count = 1, .reusable = false}}, SHIFT(527),
  [169] = {.entry = {.count = 1, .reusable = false}}, SHIFT(8),
  [171] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4994),
  [173] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1905),
  [175] = {.entry = {.count = 1, .reusable = false}}, SHIFT(799),
  [177] = {.entry = {.count = 1, .reusable = false}}, SHIFT(950),
  [179] = {.entry = {.count = 1, .reusable = true}}, SHIFT(830),
  [181] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3902),
  [183] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2129),
  [185] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2130),
  [187] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2130),
  [189] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4190),
  [191] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4194),
  [193] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5915),
  [195] = {.entry = {.count = 1, .reusable = false}}, SHIFT(500),
  [197] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5734),
  [199] = {.entry = {.count = 1, .reusable = true}}, SHIFT(496),
  [201] = {.entry = {.count = 1, .reusable = false}}, SHIFT(510),
  [203] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6308),
  [205] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4654),
  [207] = {.entry = {.count = 1, .reusable = false}}, SHIFT(166),
  [209] = {.entry = {.count = 1, .reusable = true}}, SHIFT(167),
  [211] = {.entry = {.count = 1, .reusable = true}}, SHIFT(170),
  [213] = {.entry = {.count = 1, .reusable = true}}, SHIFT(171),
  [215] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6157),
  [217] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6638),
  [219] = {.entry = {.count = 1, .reusable = true}}, SHIFT(16),
  [221] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7610),
  [223] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5603),
  [225] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4786),
  [227] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4962),
  [229] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5620),
  [231] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5012),
  [233] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4727),
  [235] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5050),
  [237] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2500),
  [239] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4210),
  [241] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_command_name, 1, .production_id = 1),
  [243] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__expression, 1, .production_id = 1),
  [245] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_command_name, 1, .production_id = 1), REDUCE(sym__expression, 1, .production_id = 1),
  [248] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2202),
  [250] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2036),
  [252] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2087),
  [254] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4208),
  [256] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3702),
  [258] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5974),
  [260] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2433),
  [262] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5718),
  [264] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2631),
  [266] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6271),
  [268] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4732),
  [270] = {.entry = {.count = 1, .reusable = false}}, SHIFT(224),
  [272] = {.entry = {.count = 1, .reusable = false}}, SHIFT(225),
  [274] = {.entry = {.count = 1, .reusable = false}}, SHIFT(227),
  [276] = {.entry = {.count = 1, .reusable = false}}, SHIFT(200),
  [278] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_command_name, 1, .production_id = 1),
  [280] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3702),
  [282] = {.entry = {.count = 1, .reusable = true}}, SHIFT(451),
  [284] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7615),
  [286] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1080),
  [288] = {.entry = {.count = 1, .reusable = true}}, SHIFT(659),
  [290] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3678),
  [292] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3678),
  [294] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4613),
  [296] = {.entry = {.count = 1, .reusable = false}}, SHIFT(418),
  [298] = {.entry = {.count = 1, .reusable = false}}, SHIFT(226),
  [300] = {.entry = {.count = 1, .reusable = false}}, SHIFT(7392),
  [302] = {.entry = {.count = 1, .reusable = false}}, SHIFT(623),
  [304] = {.entry = {.count = 1, .reusable = false}}, SHIFT(602),
  [306] = {.entry = {.count = 1, .reusable = false}}, SHIFT(718),
  [308] = {.entry = {.count = 1, .reusable = true}}, SHIFT(759),
  [310] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4030),
  [312] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4154),
  [314] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4155),
  [316] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5968),
  [318] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1105),
  [320] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5849),
  [322] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1078),
  [324] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1085),
  [326] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6359),
  [328] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4814),
  [330] = {.entry = {.count = 1, .reusable = false}}, SHIFT(148),
  [332] = {.entry = {.count = 1, .reusable = true}}, SHIFT(150),
  [334] = {.entry = {.count = 1, .reusable = true}}, SHIFT(151),
  [336] = {.entry = {.count = 1, .reusable = true}}, SHIFT(152),
  [338] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6148),
  [340] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6616),
  [342] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1077),
  [344] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7608),
  [346] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5567),
  [348] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4950),
  [350] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2521),
  [352] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3699),
  [354] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__expression, 1, .production_id = 1),
  [356] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_command_name, 1, .production_id = 1), REDUCE(sym__expression, 1, .production_id = 1),
  [359] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2261),
  [361] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2001),
  [363] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4215),
  [365] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3921),
  [367] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5935),
  [369] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2382),
  [371] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5724),
  [373] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2521),
  [375] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2679),
  [377] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6307),
  [379] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4817),
  [381] = {.entry = {.count = 1, .reusable = false}}, SHIFT(215),
  [383] = {.entry = {.count = 1, .reusable = false}}, SHIFT(216),
  [385] = {.entry = {.count = 1, .reusable = true}}, SHIFT(218),
  [387] = {.entry = {.count = 1, .reusable = true}}, SHIFT(194),
  [389] = {.entry = {.count = 1, .reusable = true}}, SHIFT(454),
  [391] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7614),
  [393] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5024),
  [395] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4818),
  [397] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4623),
  [399] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4610),
  [401] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5564),
  [403] = {.entry = {.count = 1, .reusable = false}}, SHIFT(813),
  [405] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6729),
  [407] = {.entry = {.count = 1, .reusable = false}}, SHIFT(7389),
  [409] = {.entry = {.count = 1, .reusable = true}}, SHIFT(671),
  [411] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3622),
  [413] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3622),
  [415] = {.entry = {.count = 1, .reusable = false}}, SHIFT(155),
  [417] = {.entry = {.count = 1, .reusable = false}}, SHIFT(430),
  [419] = {.entry = {.count = 1, .reusable = false}}, SHIFT(422),
  [421] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4122),
  [423] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_last_case_item, 4, .production_id = 112),
  [425] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5247),
  [427] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5280),
  [429] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5295),
  [431] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6970),
  [433] = {.entry = {.count = 1, .reusable = true}}, SHIFT(81),
  [435] = {.entry = {.count = 1, .reusable = false}}, SHIFT(656),
  [437] = {.entry = {.count = 1, .reusable = false}}, SHIFT(9),
  [439] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1830),
  [441] = {.entry = {.count = 1, .reusable = false}}, SHIFT(548),
  [443] = {.entry = {.count = 1, .reusable = false}}, SHIFT(590),
  [445] = {.entry = {.count = 1, .reusable = true}}, SHIFT(604),
  [447] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3775),
  [449] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3857),
  [451] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3855),
  [453] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6024),
  [455] = {.entry = {.count = 1, .reusable = false}}, SHIFT(824),
  [457] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5840),
  [459] = {.entry = {.count = 1, .reusable = true}}, SHIFT(815),
  [461] = {.entry = {.count = 1, .reusable = false}}, SHIFT(884),
  [463] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6269),
  [465] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4740),
  [467] = {.entry = {.count = 1, .reusable = false}}, SHIFT(388),
  [469] = {.entry = {.count = 1, .reusable = true}}, SHIFT(387),
  [471] = {.entry = {.count = 1, .reusable = true}}, SHIFT(386),
  [473] = {.entry = {.count = 1, .reusable = true}}, SHIFT(195),
  [475] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6156),
  [477] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6547),
  [479] = {.entry = {.count = 1, .reusable = true}}, SHIFT(819),
  [481] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7646),
  [483] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_last_case_item, 3, .production_id = 46),
  [485] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5162),
  [487] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5321),
  [489] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_last_case_item, 3, .production_id = 83),
  [491] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5143),
  [493] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5314),
  [495] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5313),
  [497] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5144),
  [499] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5310),
  [501] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5309),
  [503] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_last_case_item, 3, .production_id = 79),
  [505] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5132),
  [507] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5281),
  [509] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5277),
  [511] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_last_case_item, 4, .production_id = 109),
  [513] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5227),
  [515] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5259),
  [517] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5384),
  [519] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5193),
  [521] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5396),
  [523] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5401),
  [525] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5145),
  [527] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5330),
  [529] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5329),
  [531] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_last_case_item, 2, .production_id = 60),
  [533] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5169),
  [535] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5353),
  [537] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_last_case_item, 3, .production_id = 45),
  [539] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5164),
  [541] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5325),
  [543] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_last_case_item, 2, .production_id = 59),
  [545] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5172),
  [547] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5354),
  [549] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5250),
  [551] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5257),
  [553] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5273),
  [555] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__terminated_statement, 2), SHIFT_REPEAT(1080),
  [558] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__terminated_statement, 2), SHIFT_REPEAT(6777),
  [561] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__terminated_statement, 2), SHIFT_REPEAT(7604),
  [564] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__terminated_statement, 2), SHIFT_REPEAT(659),
  [567] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__terminated_statement, 2), SHIFT_REPEAT(3678),
  [570] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__terminated_statement, 2), SHIFT_REPEAT(3678),
  [573] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__terminated_statement, 2), SHIFT_REPEAT(240),
  [576] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__terminated_statement, 2), SHIFT_REPEAT(432),
  [579] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__terminated_statement, 2),
  [581] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__terminated_statement, 2), SHIFT_REPEAT(428),
  [584] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__terminated_statement, 2), SHIFT_REPEAT(4078),
  [587] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__terminated_statement, 2), SHIFT_REPEAT(7392),
  [590] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__terminated_statement, 2), SHIFT_REPEAT(154),
  [593] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__terminated_statement, 2), SHIFT_REPEAT(623),
  [596] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__terminated_statement, 2), SHIFT_REPEAT(4),
  [599] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__terminated_statement, 2), SHIFT_REPEAT(1730),
  [602] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__terminated_statement, 2), SHIFT_REPEAT(602),
  [605] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__terminated_statement, 2), SHIFT_REPEAT(718),
  [608] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__terminated_statement, 2), SHIFT_REPEAT(759),
  [611] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__terminated_statement, 2), SHIFT_REPEAT(4030),
  [614] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__terminated_statement, 2), SHIFT_REPEAT(4154),
  [617] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__terminated_statement, 2), SHIFT_REPEAT(4155),
  [620] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__terminated_statement, 2), SHIFT_REPEAT(5968),
  [623] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__terminated_statement, 2), SHIFT_REPEAT(1105),
  [626] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__terminated_statement, 2), SHIFT_REPEAT(5849),
  [629] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__terminated_statement, 2), SHIFT_REPEAT(1078),
  [632] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__terminated_statement, 2), SHIFT_REPEAT(1085),
  [635] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__terminated_statement, 2), SHIFT_REPEAT(6359),
  [638] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__terminated_statement, 2), SHIFT_REPEAT(4814),
  [641] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__terminated_statement, 2), SHIFT_REPEAT(148),
  [644] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__terminated_statement, 2), SHIFT_REPEAT(150),
  [647] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__terminated_statement, 2), SHIFT_REPEAT(151),
  [650] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__terminated_statement, 2), SHIFT_REPEAT(152),
  [653] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__terminated_statement, 2), SHIFT_REPEAT(6148),
  [656] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__terminated_statement, 2), SHIFT_REPEAT(6616),
  [659] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__terminated_statement, 2), SHIFT_REPEAT(1077),
  [662] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__terminated_statement, 2), SHIFT_REPEAT(7608),
  [665] = {.entry = {.count = 1, .reusable = false}}, SHIFT(886),
  [667] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6785),
  [669] = {.entry = {.count = 1, .reusable = false}}, SHIFT(7479),
  [671] = {.entry = {.count = 1, .reusable = true}}, SHIFT(701),
  [673] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3651),
  [675] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3651),
  [677] = {.entry = {.count = 1, .reusable = false}}, SHIFT(163),
  [679] = {.entry = {.count = 1, .reusable = false}}, SHIFT(424),
  [681] = {.entry = {.count = 1, .reusable = false}}, SHIFT(421),
  [683] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3827),
  [685] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5261),
  [687] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6894),
  [689] = {.entry = {.count = 1, .reusable = true}}, SHIFT(92),
  [691] = {.entry = {.count = 1, .reusable = false}}, SHIFT(650),
  [693] = {.entry = {.count = 1, .reusable = false}}, SHIFT(11),
  [695] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1869),
  [697] = {.entry = {.count = 1, .reusable = false}}, SHIFT(563),
  [699] = {.entry = {.count = 1, .reusable = false}}, SHIFT(632),
  [701] = {.entry = {.count = 1, .reusable = true}}, SHIFT(648),
  [703] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3821),
  [705] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3710),
  [707] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3709),
  [709] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6048),
  [711] = {.entry = {.count = 1, .reusable = false}}, SHIFT(850),
  [713] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5804),
  [715] = {.entry = {.count = 1, .reusable = true}}, SHIFT(852),
  [717] = {.entry = {.count = 1, .reusable = false}}, SHIFT(968),
  [719] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6355),
  [721] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4680),
  [723] = {.entry = {.count = 1, .reusable = false}}, SHIFT(320),
  [725] = {.entry = {.count = 1, .reusable = true}}, SHIFT(319),
  [727] = {.entry = {.count = 1, .reusable = true}}, SHIFT(318),
  [729] = {.entry = {.count = 1, .reusable = true}}, SHIFT(393),
  [731] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6142),
  [733] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6641),
  [735] = {.entry = {.count = 1, .reusable = true}}, SHIFT(853),
  [737] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7659),
  [739] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5326),
  [741] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5252),
  [743] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5338),
  [745] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5251),
  [747] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5335),
  [749] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5311),
  [751] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5340),
  [753] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5320),
  [755] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5332),
  [757] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5307),
  [759] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5304),
  [761] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_elif_clause, 4),
  [763] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_elif_clause, 3),
  [765] = {.entry = {.count = 1, .reusable = false}}, SHIFT(940),
  [767] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6166),
  [769] = {.entry = {.count = 1, .reusable = false}}, SHIFT(636),
  [771] = {.entry = {.count = 1, .reusable = false}}, SHIFT(920),
  [773] = {.entry = {.count = 1, .reusable = true}}, SHIFT(938),
  [775] = {.entry = {.count = 1, .reusable = true}}, SHIFT(932),
  [777] = {.entry = {.count = 1, .reusable = false}}, SHIFT(94),
  [779] = {.entry = {.count = 1, .reusable = false}}, SHIFT(84),
  [781] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6162),
  [783] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6170),
  [785] = {.entry = {.count = 1, .reusable = false}}, SHIFT(280),
  [787] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6163),
  [789] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6178),
  [791] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6161),
  [793] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6165),
  [795] = {.entry = {.count = 1, .reusable = false}}, SHIFT(70),
  [797] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6182),
  [799] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1012),
  [801] = {.entry = {.count = 1, .reusable = false}}, SHIFT(7482),
  [803] = {.entry = {.count = 1, .reusable = false}}, SHIFT(626),
  [805] = {.entry = {.count = 1, .reusable = false}}, SHIFT(592),
  [807] = {.entry = {.count = 1, .reusable = false}}, SHIFT(761),
  [809] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__terminated_statement, 2),
  [811] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4858),
  [813] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4609),
  [815] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5547),
  [817] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5615),
  [819] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5613),
  [821] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5591),
  [823] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5676),
  [825] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5652),
  [827] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5016),
  [829] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4959),
  [831] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4844),
  [833] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4712),
  [835] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4785),
  [837] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4643),
  [839] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4976),
  [841] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5008),
  [843] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5074),
  [845] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5092),
  [847] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4953),
  [849] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4864),
  [851] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4982),
  [853] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5019),
  [855] = {.entry = {.count = 1, .reusable = false}}, SHIFT(19),
  [857] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5042),
  [859] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5049),
  [861] = {.entry = {.count = 1, .reusable = false}}, SHIFT(21),
  [863] = {.entry = {.count = 1, .reusable = false}}, SHIFT(18),
  [865] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4875),
  [867] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4905),
  [869] = {.entry = {.count = 1, .reusable = false}}, SHIFT(17),
  [871] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5010),
  [873] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_else_clause, 1),
  [875] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_else_clause, 2),
  [877] = {.entry = {.count = 1, .reusable = false}}, SHIFT(48),
  [879] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4856),
  [881] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4615),
  [883] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4958),
  [885] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5621),
  [887] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5647),
  [889] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__statements_repeat1, 2), SHIFT_REPEAT(1080),
  [892] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__statements_repeat1, 2), SHIFT_REPEAT(6777),
  [895] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__statements_repeat1, 2), SHIFT_REPEAT(7604),
  [898] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__statements_repeat1, 2), SHIFT_REPEAT(659),
  [901] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__statements_repeat1, 2), SHIFT_REPEAT(3678),
  [904] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__statements_repeat1, 2), SHIFT_REPEAT(3678),
  [907] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__statements_repeat1, 2), SHIFT_REPEAT(240),
  [910] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__statements_repeat1, 2), SHIFT_REPEAT(432),
  [913] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__statements_repeat1, 2), SHIFT_REPEAT(428),
  [916] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__statements_repeat1, 2), SHIFT_REPEAT(4078),
  [919] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__statements_repeat1, 2), SHIFT_REPEAT(7392),
  [922] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__statements_repeat1, 2), SHIFT_REPEAT(154),
  [925] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__statements_repeat1, 2), SHIFT_REPEAT(623),
  [928] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__statements_repeat1, 2), SHIFT_REPEAT(4),
  [931] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__statements_repeat1, 2), SHIFT_REPEAT(1730),
  [934] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__statements_repeat1, 2), SHIFT_REPEAT(602),
  [937] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__statements_repeat1, 2), SHIFT_REPEAT(718),
  [940] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__statements_repeat1, 2), SHIFT_REPEAT(759),
  [943] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__statements_repeat1, 2), SHIFT_REPEAT(4030),
  [946] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__statements_repeat1, 2), SHIFT_REPEAT(4154),
  [949] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__statements_repeat1, 2), SHIFT_REPEAT(4155),
  [952] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__statements_repeat1, 2), SHIFT_REPEAT(5968),
  [955] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__statements_repeat1, 2), SHIFT_REPEAT(1105),
  [958] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__statements_repeat1, 2), SHIFT_REPEAT(5849),
  [961] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__statements_repeat1, 2), SHIFT_REPEAT(1078),
  [964] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__statements_repeat1, 2), SHIFT_REPEAT(1085),
  [967] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__statements_repeat1, 2), SHIFT_REPEAT(6359),
  [970] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__statements_repeat1, 2), SHIFT_REPEAT(4814),
  [973] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__statements_repeat1, 2), SHIFT_REPEAT(148),
  [976] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__statements_repeat1, 2), SHIFT_REPEAT(150),
  [979] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__statements_repeat1, 2), SHIFT_REPEAT(151),
  [982] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__statements_repeat1, 2), SHIFT_REPEAT(152),
  [985] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__statements_repeat1, 2), SHIFT_REPEAT(6148),
  [988] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__statements_repeat1, 2), SHIFT_REPEAT(6616),
  [991] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__statements_repeat1, 2), SHIFT_REPEAT(1077),
  [994] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__statements_repeat1, 2), SHIFT_REPEAT(7608),
  [997] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1719),
  [999] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3688),
  [1001] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3688),
  [1003] = {.entry = {.count = 1, .reusable = false}}, SHIFT(7511),
  [1005] = {.entry = {.count = 1, .reusable = false}}, SHIFT(644),
  [1007] = {.entry = {.count = 1, .reusable = false}}, SHIFT(748),
  [1009] = {.entry = {.count = 1, .reusable = false}}, SHIFT(878),
  [1011] = {.entry = {.count = 1, .reusable = true}}, SHIFT(836),
  [1013] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4107),
  [1015] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1572),
  [1017] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1717),
  [1019] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6149),
  [1021] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6717),
  [1023] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1653),
  [1025] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1800),
  [1027] = {.entry = {.count = 1, .reusable = true}}, SHIFT(698),
  [1029] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3640),
  [1031] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3640),
  [1033] = {.entry = {.count = 1, .reusable = false}}, SHIFT(112),
  [1035] = {.entry = {.count = 1, .reusable = false}}, SHIFT(619),
  [1037] = {.entry = {.count = 1, .reusable = true}}, SHIFT(905),
  [1039] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3734),
  [1041] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4076),
  [1043] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4074),
  [1045] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5865),
  [1047] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1912),
  [1049] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5822),
  [1051] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1799),
  [1053] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1977),
  [1055] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6335),
  [1057] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4798),
  [1059] = {.entry = {.count = 1, .reusable = false}}, SHIFT(351),
  [1061] = {.entry = {.count = 1, .reusable = true}}, SHIFT(353),
  [1063] = {.entry = {.count = 1, .reusable = true}}, SHIFT(354),
  [1065] = {.entry = {.count = 1, .reusable = true}}, SHIFT(265),
  [1067] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6139),
  [1069] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1797),
  [1071] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7626),
  [1073] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4210),
  [1075] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4208),
  [1077] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3929),
  [1079] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2433),
  [1081] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5718),
  [1083] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2500),
  [1085] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4732),
  [1087] = {.entry = {.count = 1, .reusable = true}}, SHIFT(227),
  [1089] = {.entry = {.count = 1, .reusable = true}}, SHIFT(200),
  [1091] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2451),
  [1093] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2041),
  [1095] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1967),
  [1097] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2704),
  [1099] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1950),
  [1101] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1949),
  [1103] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1949),
  [1105] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4016),
  [1107] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2374),
  [1109] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2451),
  [1111] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3685),
  [1113] = {.entry = {.count = 1, .reusable = true}}, SHIFT(456),
  [1115] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6610),
  [1117] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3658),
  [1119] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6388),
  [1121] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3643),
  [1123] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3103),
  [1125] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3615),
  [1127] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2573),
  [1129] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3966),
  [1131] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2139),
  [1133] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1948),
  [1135] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2056),
  [1137] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2049),
  [1139] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2049),
  [1141] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3967),
  [1143] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4028),
  [1145] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5900),
  [1147] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2409),
  [1149] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5753),
  [1151] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2573),
  [1153] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2292),
  [1155] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6273),
  [1157] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4799),
  [1159] = {.entry = {.count = 1, .reusable = false}}, SHIFT(114),
  [1161] = {.entry = {.count = 1, .reusable = false}}, SHIFT(117),
  [1163] = {.entry = {.count = 1, .reusable = true}}, SHIFT(118),
  [1165] = {.entry = {.count = 1, .reusable = true}}, SHIFT(120),
  [1167] = {.entry = {.count = 1, .reusable = true}}, SHIFT(458),
  [1169] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7597),
  [1171] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2456),
  [1173] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2137),
  [1175] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2131),
  [1177] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2132),
  [1179] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2132),
  [1181] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2408),
  [1183] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2456),
  [1185] = {.entry = {.count = 1, .reusable = true}}, SHIFT(459),
  [1187] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2503),
  [1189] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2198),
  [1191] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1970),
  [1193] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1969),
  [1195] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1969),
  [1197] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2347),
  [1199] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2503),
  [1201] = {.entry = {.count = 1, .reusable = true}}, SHIFT(460),
  [1203] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4291),
  [1205] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3657),
  [1207] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2983),
  [1209] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3671),
  [1211] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2588),
  [1213] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2069),
  [1215] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2184),
  [1217] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2188),
  [1219] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2188),
  [1221] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2537),
  [1223] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2588),
  [1225] = {.entry = {.count = 1, .reusable = true}}, SHIFT(463),
  [1227] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_concatenation, 4),
  [1229] = {.entry = {.count = 1, .reusable = false}}, SHIFT(488),
  [1231] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5800),
  [1233] = {.entry = {.count = 1, .reusable = false}}, SHIFT(492),
  [1235] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_concatenation, 4),
  [1237] = {.entry = {.count = 1, .reusable = true}}, SHIFT(492),
  [1239] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_concatenation, 4, .production_id = 1),
  [1241] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_concatenation, 4, .production_id = 1),
  [1243] = {.entry = {.count = 1, .reusable = false}}, SHIFT(506),
  [1245] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5734),
  [1247] = {.entry = {.count = 1, .reusable = false}}, SHIFT(507),
  [1249] = {.entry = {.count = 1, .reusable = true}}, SHIFT(507),
  [1251] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_concatenation_repeat1, 2),
  [1253] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_concatenation_repeat1, 2),
  [1255] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_concatenation_repeat1, 2), SHIFT_REPEAT(3702),
  [1258] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_concatenation_repeat1, 2), SHIFT_REPEAT(3702),
  [1261] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__literal_repeat1, 1),
  [1263] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__literal_repeat1, 1),
  [1265] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_concatenation, 2, .production_id = 1),
  [1267] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_concatenation, 2, .production_id = 1),
  [1269] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3771),
  [1271] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_command_name, 1),
  [1273] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__expression, 1),
  [1275] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_command_name, 1), REDUCE(sym__expression, 1),
  [1278] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_command_name, 1), SHIFT(7273),
  [1281] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_command_name, 1),
  [1283] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_command_name, 1), REDUCE(sym__expression, 1),
  [1286] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_concatenation, 2),
  [1288] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_concatenation, 2),
  [1290] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3813),
  [1292] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_arithmetic_expansion, 4),
  [1294] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_arithmetic_expansion, 4),
  [1296] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_process_substitution, 3),
  [1298] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_process_substitution, 3),
  [1300] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_brace_expression, 5),
  [1302] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_brace_expression, 5),
  [1304] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_command_substitution, 3, .production_id = 35),
  [1306] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_command_substitution, 3, .production_id = 35),
  [1308] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_command_substitution, 3),
  [1310] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_command_substitution, 3),
  [1312] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_expansion, 3, .production_id = 31),
  [1314] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_expansion, 3, .production_id = 31),
  [1316] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_number, 1),
  [1318] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_number, 1),
  [1320] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_string, 2),
  [1322] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_string, 2),
  [1324] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_concatenation_repeat1, 2, .production_id = 10),
  [1326] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_concatenation_repeat1, 2, .production_id = 10),
  [1328] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_number, 2),
  [1330] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_number, 2),
  [1332] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_arithmetic_expansion, 3),
  [1334] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_arithmetic_expansion, 3),
  [1336] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_expansion, 2),
  [1338] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_expansion, 2),
  [1340] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_string, 3),
  [1342] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_string, 3),
  [1344] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_simple_expansion, 2, .production_id = 11),
  [1346] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_simple_expansion, 2, .production_id = 11),
  [1348] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_translated_string, 2),
  [1350] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_translated_string, 2),
  [1352] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_string, 4),
  [1354] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_string, 4),
  [1356] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_simple_expansion, 2),
  [1358] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_simple_expansion, 2),
  [1360] = {.entry = {.count = 1, .reusable = false}}, SHIFT(495),
  [1362] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__literal_repeat1, 2),
  [1364] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__literal_repeat1, 2),
  [1366] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__literal_repeat1, 2), SHIFT_REPEAT(495),
  [1369] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__expression, 1),
  [1371] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_command_name, 1), SHIFT(7064),
  [1374] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3991),
  [1376] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3969),
  [1378] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_concatenation_repeat1, 2), SHIFT_REPEAT(3921),
  [1381] = {.entry = {.count = 1, .reusable = false}}, SHIFT(525),
  [1383] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__literal_repeat1, 2), SHIFT_REPEAT(525),
  [1386] = {.entry = {.count = 1, .reusable = false}}, SHIFT(472),
  [1388] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3677),
  [1390] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3677),
  [1392] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2138),
  [1394] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3965),
  [1396] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6132),
  [1398] = {.entry = {.count = 1, .reusable = false}}, SHIFT(496),
  [1400] = {.entry = {.count = 1, .reusable = false}}, SHIFT(874),
  [1402] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3749),
  [1404] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_command, 1, .production_id = 2),
  [1406] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3660),
  [1408] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_command, 1, .production_id = 2),
  [1410] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3799),
  [1412] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3748),
  [1414] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5958),
  [1416] = {.entry = {.count = 1, .reusable = false}}, SHIFT(834),
  [1418] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5817),
  [1420] = {.entry = {.count = 1, .reusable = false}}, SHIFT(987),
  [1422] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6354),
  [1424] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4693),
  [1426] = {.entry = {.count = 1, .reusable = false}}, SHIFT(335),
  [1428] = {.entry = {.count = 1, .reusable = false}}, SHIFT(334),
  [1430] = {.entry = {.count = 1, .reusable = false}}, SHIFT(333),
  [1432] = {.entry = {.count = 1, .reusable = false}}, SHIFT(414),
  [1434] = {.entry = {.count = 1, .reusable = true}}, SHIFT(876),
  [1436] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1182),
  [1438] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7654),
  [1440] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_command, 2, .production_id = 19),
  [1442] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_command, 2, .production_id = 19),
  [1444] = {.entry = {.count = 1, .reusable = false}}, SHIFT(985),
  [1446] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3809),
  [1448] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3620),
  [1450] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3832),
  [1452] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3812),
  [1454] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6070),
  [1456] = {.entry = {.count = 1, .reusable = false}}, SHIFT(969),
  [1458] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5788),
  [1460] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1155),
  [1462] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6254),
  [1464] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4675),
  [1466] = {.entry = {.count = 1, .reusable = false}}, SHIFT(305),
  [1468] = {.entry = {.count = 1, .reusable = false}}, SHIFT(304),
  [1470] = {.entry = {.count = 1, .reusable = false}}, SHIFT(303),
  [1472] = {.entry = {.count = 1, .reusable = false}}, SHIFT(372),
  [1474] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1003),
  [1476] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1328),
  [1478] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7664),
  [1480] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_repeat2, 2, .production_id = 38), SHIFT_REPEAT(874),
  [1483] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_repeat2, 2, .production_id = 38), SHIFT_REPEAT(3749),
  [1486] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_command_repeat2, 2, .production_id = 38),
  [1488] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_repeat2, 2, .production_id = 38), SHIFT_REPEAT(3660),
  [1491] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_command_repeat2, 2, .production_id = 38),
  [1493] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_repeat2, 2, .production_id = 38), SHIFT_REPEAT(3799),
  [1496] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_repeat2, 2, .production_id = 38), SHIFT_REPEAT(3748),
  [1499] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_repeat2, 2, .production_id = 38), SHIFT_REPEAT(5958),
  [1502] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_repeat2, 2, .production_id = 38), SHIFT_REPEAT(834),
  [1505] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_repeat2, 2, .production_id = 38), SHIFT_REPEAT(5817),
  [1508] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_repeat2, 2, .production_id = 38), SHIFT_REPEAT(987),
  [1511] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_repeat2, 2, .production_id = 38), SHIFT_REPEAT(6354),
  [1514] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_repeat2, 2, .production_id = 38), SHIFT_REPEAT(4693),
  [1517] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_repeat2, 2, .production_id = 38), SHIFT_REPEAT(335),
  [1520] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_repeat2, 2, .production_id = 38), SHIFT_REPEAT(334),
  [1523] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_repeat2, 2, .production_id = 38), SHIFT_REPEAT(333),
  [1526] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_repeat2, 2, .production_id = 38), SHIFT_REPEAT(414),
  [1529] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_command_repeat2, 2, .production_id = 38), SHIFT_REPEAT(6813),
  [1532] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_command_repeat2, 2, .production_id = 38), SHIFT_REPEAT(876),
  [1535] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_command_repeat2, 2, .production_id = 38), SHIFT_REPEAT(1182),
  [1538] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_command_repeat2, 2, .production_id = 38), SHIFT_REPEAT(7654),
  [1541] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1023),
  [1543] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4184),
  [1545] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3686),
  [1547] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3725),
  [1549] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4183),
  [1551] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5997),
  [1553] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1091),
  [1555] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5729),
  [1557] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1268),
  [1559] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6350),
  [1561] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4641),
  [1563] = {.entry = {.count = 1, .reusable = false}}, SHIFT(249),
  [1565] = {.entry = {.count = 1, .reusable = false}}, SHIFT(251),
  [1567] = {.entry = {.count = 1, .reusable = false}}, SHIFT(253),
  [1569] = {.entry = {.count = 1, .reusable = false}}, SHIFT(223),
  [1571] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1021),
  [1573] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1716),
  [1575] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7619),
  [1577] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1093),
  [1579] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4211),
  [1581] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3695),
  [1583] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3787),
  [1585] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4212),
  [1587] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5884),
  [1589] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1144),
  [1591] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5746),
  [1593] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1272),
  [1595] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6309),
  [1597] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4841),
  [1599] = {.entry = {.count = 1, .reusable = false}}, SHIFT(205),
  [1601] = {.entry = {.count = 1, .reusable = false}}, SHIFT(206),
  [1603] = {.entry = {.count = 1, .reusable = false}}, SHIFT(207),
  [1605] = {.entry = {.count = 1, .reusable = false}}, SHIFT(189),
  [1607] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1094),
  [1609] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1521),
  [1611] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7613),
  [1613] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_command, 3, .production_id = 39),
  [1615] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_command, 3, .production_id = 39),
  [1617] = {.entry = {.count = 1, .reusable = false}}, SHIFT(888),
  [1619] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5840),
  [1621] = {.entry = {.count = 1, .reusable = false}}, SHIFT(889),
  [1623] = {.entry = {.count = 1, .reusable = true}}, SHIFT(889),
  [1625] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_command, 2, .production_id = 18),
  [1627] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_command, 2, .production_id = 18),
  [1629] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_repeat2, 2, .production_id = 38), SHIFT_REPEAT(985),
  [1632] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_repeat2, 2, .production_id = 38), SHIFT_REPEAT(3809),
  [1635] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_repeat2, 2, .production_id = 38), SHIFT_REPEAT(3620),
  [1638] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_repeat2, 2, .production_id = 38), SHIFT_REPEAT(3832),
  [1641] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_repeat2, 2, .production_id = 38), SHIFT_REPEAT(3812),
  [1644] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_repeat2, 2, .production_id = 38), SHIFT_REPEAT(6070),
  [1647] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_repeat2, 2, .production_id = 38), SHIFT_REPEAT(969),
  [1650] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_repeat2, 2, .production_id = 38), SHIFT_REPEAT(5788),
  [1653] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_repeat2, 2, .production_id = 38), SHIFT_REPEAT(1155),
  [1656] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_repeat2, 2, .production_id = 38), SHIFT_REPEAT(6254),
  [1659] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_repeat2, 2, .production_id = 38), SHIFT_REPEAT(4675),
  [1662] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_repeat2, 2, .production_id = 38), SHIFT_REPEAT(305),
  [1665] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_repeat2, 2, .production_id = 38), SHIFT_REPEAT(304),
  [1668] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_repeat2, 2, .production_id = 38), SHIFT_REPEAT(303),
  [1671] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_repeat2, 2, .production_id = 38), SHIFT_REPEAT(372),
  [1674] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_command_repeat2, 2, .production_id = 38), SHIFT_REPEAT(6805),
  [1677] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_command_repeat2, 2, .production_id = 38), SHIFT_REPEAT(1003),
  [1680] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_command_repeat2, 2, .production_id = 38), SHIFT_REPEAT(1328),
  [1683] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_command_repeat2, 2, .production_id = 38), SHIFT_REPEAT(7664),
  [1686] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1219),
  [1688] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3682),
  [1690] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4052),
  [1692] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1174),
  [1694] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1203),
  [1696] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_command_repeat1, 2), SHIFT_REPEAT(1067),
  [1699] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_command_repeat1, 2), SHIFT_REPEAT(3755),
  [1702] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_declaration_command_repeat1, 2),
  [1704] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_declaration_command_repeat1, 2),
  [1706] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_command_repeat1, 2), SHIFT_REPEAT(3754),
  [1709] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_command_repeat1, 2), SHIFT_REPEAT(5924),
  [1712] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_command_repeat1, 2), SHIFT_REPEAT(1092),
  [1715] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_command_repeat1, 2), SHIFT_REPEAT(5823),
  [1718] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_command_repeat1, 2), SHIFT_REPEAT(1235),
  [1721] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_command_repeat1, 2), SHIFT_REPEAT(6282),
  [1724] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_command_repeat1, 2), SHIFT_REPEAT(4703),
  [1727] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_command_repeat1, 2), SHIFT_REPEAT(348),
  [1730] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_command_repeat1, 2), SHIFT_REPEAT(347),
  [1733] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_command_repeat1, 2), SHIFT_REPEAT(346),
  [1736] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_command_repeat1, 2), SHIFT_REPEAT(285),
  [1739] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_command_repeat1, 2), SHIFT_REPEAT(546),
  [1742] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_declaration_command_repeat1, 2), SHIFT_REPEAT(6673),
  [1745] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_declaration_command_repeat1, 2), SHIFT_REPEAT(1068),
  [1748] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_declaration_command_repeat1, 2), SHIFT_REPEAT(7652),
  [1751] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1067),
  [1753] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3755),
  [1755] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_declaration_command, 1),
  [1757] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_declaration_command, 1),
  [1759] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3754),
  [1761] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5924),
  [1763] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1092),
  [1765] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5823),
  [1767] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1235),
  [1769] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6282),
  [1771] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4703),
  [1773] = {.entry = {.count = 1, .reusable = false}}, SHIFT(348),
  [1775] = {.entry = {.count = 1, .reusable = false}}, SHIFT(347),
  [1777] = {.entry = {.count = 1, .reusable = false}}, SHIFT(346),
  [1779] = {.entry = {.count = 1, .reusable = false}}, SHIFT(285),
  [1781] = {.entry = {.count = 1, .reusable = false}}, SHIFT(551),
  [1783] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6673),
  [1785] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1068),
  [1787] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7652),
  [1789] = {.entry = {.count = 1, .reusable = false}}, SHIFT(931),
  [1791] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5804),
  [1793] = {.entry = {.count = 1, .reusable = false}}, SHIFT(934),
  [1795] = {.entry = {.count = 1, .reusable = true}}, SHIFT(934),
  [1797] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_declaration_command, 2),
  [1799] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_declaration_command, 2),
  [1801] = {.entry = {.count = 1, .reusable = false}}, SHIFT(546),
  [1803] = {.entry = {.count = 1, .reusable = false}}, SHIFT(975),
  [1805] = {.entry = {.count = 1, .reusable = false}}, SHIFT(977),
  [1807] = {.entry = {.count = 1, .reusable = true}}, SHIFT(977),
  [1809] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1062),
  [1811] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5849),
  [1813] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1063),
  [1815] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1063),
  [1817] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1126),
  [1819] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5757),
  [1821] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1128),
  [1823] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1128),
  [1825] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1171),
  [1827] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3741),
  [1829] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3742),
  [1831] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6085),
  [1833] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1266),
  [1835] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5793),
  [1837] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1469),
  [1839] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6285),
  [1841] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4677),
  [1843] = {.entry = {.count = 1, .reusable = false}}, SHIFT(311),
  [1845] = {.entry = {.count = 1, .reusable = false}}, SHIFT(310),
  [1847] = {.entry = {.count = 1, .reusable = false}}, SHIFT(309),
  [1849] = {.entry = {.count = 1, .reusable = false}}, SHIFT(381),
  [1851] = {.entry = {.count = 1, .reusable = false}}, SHIFT(565),
  [1853] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6628),
  [1855] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1170),
  [1857] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7662),
  [1859] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_repeat2, 2, .production_id = 38), SHIFT_REPEAT(1023),
  [1862] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_repeat2, 2, .production_id = 38), SHIFT_REPEAT(4184),
  [1865] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_repeat2, 2, .production_id = 38), SHIFT_REPEAT(3686),
  [1868] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_repeat2, 2, .production_id = 38), SHIFT_REPEAT(3725),
  [1871] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_repeat2, 2, .production_id = 38), SHIFT_REPEAT(4183),
  [1874] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_repeat2, 2, .production_id = 38), SHIFT_REPEAT(5997),
  [1877] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_repeat2, 2, .production_id = 38), SHIFT_REPEAT(1091),
  [1880] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_repeat2, 2, .production_id = 38), SHIFT_REPEAT(5729),
  [1883] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_repeat2, 2, .production_id = 38), SHIFT_REPEAT(1268),
  [1886] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_repeat2, 2, .production_id = 38), SHIFT_REPEAT(6350),
  [1889] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_repeat2, 2, .production_id = 38), SHIFT_REPEAT(4641),
  [1892] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_repeat2, 2, .production_id = 38), SHIFT_REPEAT(249),
  [1895] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_repeat2, 2, .production_id = 38), SHIFT_REPEAT(251),
  [1898] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_repeat2, 2, .production_id = 38), SHIFT_REPEAT(253),
  [1901] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_repeat2, 2, .production_id = 38), SHIFT_REPEAT(223),
  [1904] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_command_repeat2, 2, .production_id = 38), SHIFT_REPEAT(7063),
  [1907] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_command_repeat2, 2, .production_id = 38), SHIFT_REPEAT(1021),
  [1910] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_command_repeat2, 2, .production_id = 38), SHIFT_REPEAT(1716),
  [1913] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_command_repeat2, 2, .production_id = 38), SHIFT_REPEAT(7619),
  [1916] = {.entry = {.count = 1, .reusable = false}}, SHIFT(559),
  [1918] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_command_repeat1, 2), SHIFT_REPEAT(1171),
  [1921] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_command_repeat1, 2), SHIFT_REPEAT(3741),
  [1924] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_command_repeat1, 2), SHIFT_REPEAT(3742),
  [1927] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_command_repeat1, 2), SHIFT_REPEAT(6085),
  [1930] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_command_repeat1, 2), SHIFT_REPEAT(1266),
  [1933] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_command_repeat1, 2), SHIFT_REPEAT(5793),
  [1936] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_command_repeat1, 2), SHIFT_REPEAT(1469),
  [1939] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_command_repeat1, 2), SHIFT_REPEAT(6285),
  [1942] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_command_repeat1, 2), SHIFT_REPEAT(4677),
  [1945] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_command_repeat1, 2), SHIFT_REPEAT(311),
  [1948] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_command_repeat1, 2), SHIFT_REPEAT(310),
  [1951] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_command_repeat1, 2), SHIFT_REPEAT(309),
  [1954] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_command_repeat1, 2), SHIFT_REPEAT(381),
  [1957] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_command_repeat1, 2), SHIFT_REPEAT(565),
  [1960] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_declaration_command_repeat1, 2), SHIFT_REPEAT(6628),
  [1963] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_declaration_command_repeat1, 2), SHIFT_REPEAT(1170),
  [1966] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_declaration_command_repeat1, 2), SHIFT_REPEAT(7662),
  [1969] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1159),
  [1971] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1161),
  [1973] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1161),
  [1975] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_repeat2, 2, .production_id = 38), SHIFT_REPEAT(1093),
  [1978] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_repeat2, 2, .production_id = 38), SHIFT_REPEAT(4211),
  [1981] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_repeat2, 2, .production_id = 38), SHIFT_REPEAT(3695),
  [1984] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_repeat2, 2, .production_id = 38), SHIFT_REPEAT(3787),
  [1987] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_repeat2, 2, .production_id = 38), SHIFT_REPEAT(4212),
  [1990] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_repeat2, 2, .production_id = 38), SHIFT_REPEAT(5884),
  [1993] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_repeat2, 2, .production_id = 38), SHIFT_REPEAT(1144),
  [1996] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_repeat2, 2, .production_id = 38), SHIFT_REPEAT(5746),
  [1999] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_repeat2, 2, .production_id = 38), SHIFT_REPEAT(1272),
  [2002] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_repeat2, 2, .production_id = 38), SHIFT_REPEAT(6309),
  [2005] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_repeat2, 2, .production_id = 38), SHIFT_REPEAT(4841),
  [2008] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_repeat2, 2, .production_id = 38), SHIFT_REPEAT(205),
  [2011] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_repeat2, 2, .production_id = 38), SHIFT_REPEAT(206),
  [2014] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_repeat2, 2, .production_id = 38), SHIFT_REPEAT(207),
  [2017] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_repeat2, 2, .production_id = 38), SHIFT_REPEAT(189),
  [2020] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_command_repeat2, 2, .production_id = 38), SHIFT_REPEAT(7344),
  [2023] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_command_repeat2, 2, .production_id = 38), SHIFT_REPEAT(1094),
  [2026] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_command_repeat2, 2, .production_id = 38), SHIFT_REPEAT(1521),
  [2029] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_command_repeat2, 2, .production_id = 38), SHIFT_REPEAT(7613),
  [2032] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1322),
  [2034] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4200),
  [2036] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4204),
  [2038] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6034),
  [2040] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1460),
  [2042] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5722),
  [2044] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1580),
  [2046] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6321),
  [2048] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4765),
  [2050] = {.entry = {.count = 1, .reusable = false}}, SHIFT(174),
  [2052] = {.entry = {.count = 1, .reusable = false}}, SHIFT(175),
  [2054] = {.entry = {.count = 1, .reusable = false}}, SHIFT(176),
  [2056] = {.entry = {.count = 1, .reusable = false}}, SHIFT(177),
  [2058] = {.entry = {.count = 1, .reusable = false}}, SHIFT(577),
  [2060] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6704),
  [2062] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1318),
  [2064] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7611),
  [2066] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1287),
  [2068] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1289),
  [2070] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1289),
  [2072] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_file_redirect, 3, .production_id = 21),
  [2074] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_file_redirect, 3, .production_id = 21),
  [2076] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_repeat2, 2, .production_id = 38), SHIFT_REPEAT(1219),
  [2079] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_repeat2, 2, .production_id = 38), SHIFT_REPEAT(3682),
  [2082] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_repeat2, 2, .production_id = 38), SHIFT_REPEAT(4052),
  [2085] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_repeat2, 2, .production_id = 38), SHIFT_REPEAT(1174),
  [2088] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_command_repeat2, 2, .production_id = 38), SHIFT_REPEAT(6955),
  [2091] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_command_repeat2, 2, .production_id = 38), SHIFT_REPEAT(1203),
  [2094] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_file_redirect, 2, .production_id = 8),
  [2096] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_file_redirect, 2, .production_id = 8),
  [2098] = {.entry = {.count = 1, .reusable = false}}, SHIFT(570),
  [2100] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_command_repeat1, 2), SHIFT_REPEAT(1322),
  [2103] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_command_repeat1, 2), SHIFT_REPEAT(4200),
  [2106] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_command_repeat1, 2), SHIFT_REPEAT(4204),
  [2109] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_command_repeat1, 2), SHIFT_REPEAT(6034),
  [2112] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_command_repeat1, 2), SHIFT_REPEAT(1460),
  [2115] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_command_repeat1, 2), SHIFT_REPEAT(5722),
  [2118] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_command_repeat1, 2), SHIFT_REPEAT(1580),
  [2121] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_command_repeat1, 2), SHIFT_REPEAT(6321),
  [2124] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_command_repeat1, 2), SHIFT_REPEAT(4765),
  [2127] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_command_repeat1, 2), SHIFT_REPEAT(174),
  [2130] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_command_repeat1, 2), SHIFT_REPEAT(175),
  [2133] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_command_repeat1, 2), SHIFT_REPEAT(176),
  [2136] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_command_repeat1, 2), SHIFT_REPEAT(177),
  [2139] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_command_repeat1, 2), SHIFT_REPEAT(577),
  [2142] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_declaration_command_repeat1, 2), SHIFT_REPEAT(6704),
  [2145] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_declaration_command_repeat1, 2), SHIFT_REPEAT(1318),
  [2148] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_declaration_command_repeat1, 2), SHIFT_REPEAT(7611),
  [2151] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1194),
  [2153] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1195),
  [2155] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1195),
  [2157] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_command_repeat1, 2), SHIFT_REPEAT(1360),
  [2160] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_command_repeat1, 2), SHIFT_REPEAT(4196),
  [2163] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_command_repeat1, 2), SHIFT_REPEAT(4195),
  [2166] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_command_repeat1, 2), SHIFT_REPEAT(6071),
  [2169] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_command_repeat1, 2), SHIFT_REPEAT(1356),
  [2172] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_command_repeat1, 2), SHIFT_REPEAT(5758),
  [2175] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_command_repeat1, 2), SHIFT_REPEAT(1571),
  [2178] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_command_repeat1, 2), SHIFT_REPEAT(6343),
  [2181] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_command_repeat1, 2), SHIFT_REPEAT(4649),
  [2184] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_command_repeat1, 2), SHIFT_REPEAT(236),
  [2187] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_command_repeat1, 2), SHIFT_REPEAT(238),
  [2190] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_command_repeat1, 2), SHIFT_REPEAT(241),
  [2193] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_command_repeat1, 2), SHIFT_REPEAT(214),
  [2196] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_command_repeat1, 2), SHIFT_REPEAT(581),
  [2199] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_declaration_command_repeat1, 2), SHIFT_REPEAT(6603),
  [2202] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_declaration_command_repeat1, 2), SHIFT_REPEAT(1362),
  [2205] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_declaration_command_repeat1, 2), SHIFT_REPEAT(7617),
  [2208] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(1158),
  [2211] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(3764),
  [2214] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2),
  [2216] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2),
  [2218] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(3759),
  [2221] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(5908),
  [2224] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(1089),
  [2227] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(5825),
  [2230] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(1232),
  [2233] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(6255),
  [2236] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(4707),
  [2239] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(360),
  [2242] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(358),
  [2245] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(352),
  [2248] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(273),
  [2251] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(1163),
  [2254] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(7651),
  [2257] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1360),
  [2259] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4196),
  [2261] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4195),
  [2263] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6071),
  [2265] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1356),
  [2267] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5758),
  [2269] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1571),
  [2271] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6343),
  [2273] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4649),
  [2275] = {.entry = {.count = 1, .reusable = false}}, SHIFT(236),
  [2277] = {.entry = {.count = 1, .reusable = false}}, SHIFT(238),
  [2279] = {.entry = {.count = 1, .reusable = false}}, SHIFT(241),
  [2281] = {.entry = {.count = 1, .reusable = false}}, SHIFT(214),
  [2283] = {.entry = {.count = 1, .reusable = false}}, SHIFT(581),
  [2285] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6603),
  [2287] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1362),
  [2289] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7617),
  [2291] = {.entry = {.count = 1, .reusable = false}}, SHIFT(585),
  [2293] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1190),
  [2295] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3753),
  [2297] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_unset_command, 1),
  [2299] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_unset_command, 1),
  [2301] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3751),
  [2303] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5940),
  [2305] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1226),
  [2307] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5819),
  [2309] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1478),
  [2311] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6324),
  [2313] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4696),
  [2315] = {.entry = {.count = 1, .reusable = false}}, SHIFT(344),
  [2317] = {.entry = {.count = 1, .reusable = false}}, SHIFT(343),
  [2319] = {.entry = {.count = 1, .reusable = false}}, SHIFT(338),
  [2321] = {.entry = {.count = 1, .reusable = false}}, SHIFT(357),
  [2323] = {.entry = {.count = 1, .reusable = false}}, SHIFT(613),
  [2325] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1191),
  [2327] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7653),
  [2329] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1290),
  [2331] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3724),
  [2333] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3722),
  [2335] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6012),
  [2337] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1234),
  [2339] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5809),
  [2341] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1406),
  [2343] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6305),
  [2345] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4682),
  [2347] = {.entry = {.count = 1, .reusable = false}}, SHIFT(326),
  [2349] = {.entry = {.count = 1, .reusable = false}}, SHIFT(325),
  [2351] = {.entry = {.count = 1, .reusable = false}}, SHIFT(324),
  [2353] = {.entry = {.count = 1, .reusable = false}}, SHIFT(401),
  [2355] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1291),
  [2357] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7657),
  [2359] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1602),
  [2361] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1665),
  [2363] = {.entry = {.count = 1, .reusable = false}}, SHIFT(594),
  [2365] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6560),
  [2367] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1481),
  [2369] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(1290),
  [2372] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(3724),
  [2375] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(3722),
  [2378] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(6012),
  [2381] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(1234),
  [2384] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(5809),
  [2387] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(1406),
  [2390] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(6305),
  [2393] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(4682),
  [2396] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(326),
  [2399] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(325),
  [2402] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(324),
  [2405] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(401),
  [2408] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(1291),
  [2411] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(7657),
  [2414] = {.entry = {.count = 1, .reusable = false}}, SHIFT(612),
  [2416] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1794),
  [2418] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3697),
  [2420] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4138),
  [2422] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1823),
  [2424] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1795),
  [2426] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1247),
  [2428] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5825),
  [2430] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1246),
  [2432] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1246),
  [2434] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_unset_command_repeat1, 2), SHIFT_REPEAT(1190),
  [2437] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_unset_command_repeat1, 2), SHIFT_REPEAT(3753),
  [2440] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_unset_command_repeat1, 2),
  [2442] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_unset_command_repeat1, 2),
  [2444] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_unset_command_repeat1, 2), SHIFT_REPEAT(3751),
  [2447] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_unset_command_repeat1, 2), SHIFT_REPEAT(5940),
  [2450] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_unset_command_repeat1, 2), SHIFT_REPEAT(1226),
  [2453] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_unset_command_repeat1, 2), SHIFT_REPEAT(5819),
  [2456] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_unset_command_repeat1, 2), SHIFT_REPEAT(1478),
  [2459] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_unset_command_repeat1, 2), SHIFT_REPEAT(6324),
  [2462] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_unset_command_repeat1, 2), SHIFT_REPEAT(4696),
  [2465] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_unset_command_repeat1, 2), SHIFT_REPEAT(344),
  [2468] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_unset_command_repeat1, 2), SHIFT_REPEAT(343),
  [2471] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_unset_command_repeat1, 2), SHIFT_REPEAT(338),
  [2474] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_unset_command_repeat1, 2), SHIFT_REPEAT(357),
  [2477] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_unset_command_repeat1, 2), SHIFT_REPEAT(598),
  [2480] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_unset_command_repeat1, 2), SHIFT_REPEAT(1191),
  [2483] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_unset_command_repeat1, 2), SHIFT_REPEAT(7653),
  [2486] = {.entry = {.count = 1, .reusable = false}}, SHIFT(610),
  [2488] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1309),
  [2490] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1310),
  [2492] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1310),
  [2494] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_file_redirect, 1),
  [2496] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_file_redirect, 1),
  [2498] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_file_redirect, 2, .production_id = 4),
  [2500] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_file_redirect, 2, .production_id = 4),
  [2502] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(1269),
  [2505] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(3729),
  [2508] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(3730),
  [2511] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(6084),
  [2514] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(1264),
  [2517] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(5796),
  [2520] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(1393),
  [2523] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(6311),
  [2526] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(4678),
  [2529] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(314),
  [2532] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(313),
  [2535] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(312),
  [2538] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(385),
  [2541] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(1270),
  [2544] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(7661),
  [2547] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_command_repeat1, 2), SHIFT_REPEAT(1602),
  [2550] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_command_repeat1, 2), SHIFT_REPEAT(1665),
  [2553] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_command_repeat1, 2), SHIFT_REPEAT(612),
  [2556] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_declaration_command_repeat1, 2), SHIFT_REPEAT(6560),
  [2559] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_declaration_command_repeat1, 2), SHIFT_REPEAT(1481),
  [2562] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_unset_command, 2),
  [2564] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_unset_command, 2),
  [2566] = {.entry = {.count = 1, .reusable = false}}, SHIFT(598),
  [2568] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1394),
  [2570] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3849),
  [2572] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3850),
  [2574] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6044),
  [2576] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1468),
  [2578] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5777),
  [2580] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1616),
  [2582] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6274),
  [2584] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4661),
  [2586] = {.entry = {.count = 1, .reusable = false}}, SHIFT(292),
  [2588] = {.entry = {.count = 1, .reusable = false}}, SHIFT(291),
  [2590] = {.entry = {.count = 1, .reusable = false}}, SHIFT(288),
  [2592] = {.entry = {.count = 1, .reusable = false}}, SHIFT(349),
  [2594] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1307),
  [2596] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7667),
  [2598] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1381),
  [2600] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3744),
  [2602] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3743),
  [2604] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5976),
  [2606] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1308),
  [2608] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5814),
  [2610] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1515),
  [2612] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6272),
  [2614] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4690),
  [2616] = {.entry = {.count = 1, .reusable = false}}, SHIFT(332),
  [2618] = {.entry = {.count = 1, .reusable = false}}, SHIFT(331),
  [2620] = {.entry = {.count = 1, .reusable = false}}, SHIFT(330),
  [2622] = {.entry = {.count = 1, .reusable = false}}, SHIFT(409),
  [2624] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1388),
  [2626] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7655),
  [2628] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2211),
  [2630] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4082),
  [2632] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3672),
  [2634] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3715),
  [2636] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4079),
  [2638] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6014),
  [2640] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2033),
  [2642] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5780),
  [2644] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2211),
  [2646] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2268),
  [2648] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6342),
  [2650] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4779),
  [2652] = {.entry = {.count = 1, .reusable = false}}, SHIFT(340),
  [2654] = {.entry = {.count = 1, .reusable = true}}, SHIFT(341),
  [2656] = {.entry = {.count = 1, .reusable = true}}, SHIFT(342),
  [2658] = {.entry = {.count = 1, .reusable = true}}, SHIFT(258),
  [2660] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2203),
  [2662] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2558),
  [2664] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7625),
  [2666] = {.entry = {.count = 1, .reusable = false}}, SHIFT(957),
  [2668] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1799),
  [2670] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(1394),
  [2673] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(3849),
  [2676] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(3850),
  [2679] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(6044),
  [2682] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(1468),
  [2685] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(5777),
  [2688] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(1616),
  [2691] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(6274),
  [2694] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(4661),
  [2697] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(292),
  [2700] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(291),
  [2703] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(288),
  [2706] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(349),
  [2709] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(1307),
  [2712] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(7667),
  [2715] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__statements_repeat1, 2),
  [2717] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__statements_repeat1, 2),
  [2719] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__statements, 2),
  [2721] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__statements, 2),
  [2723] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1078),
  [2725] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__statements, 3),
  [2727] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__statements, 3),
  [2729] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(1370),
  [2732] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(4202),
  [2735] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(4201),
  [2738] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(6038),
  [2741] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(1445),
  [2744] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(5764),
  [2747] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(1702),
  [2750] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(6292),
  [2753] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(4694),
  [2756] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(229),
  [2759] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(230),
  [2762] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(231),
  [2765] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(204),
  [2768] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(1372),
  [2771] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(7616),
  [2774] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1642),
  [2776] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1641),
  [2778] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1641),
  [2780] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1326),
  [2782] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3786),
  [2784] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3789),
  [2786] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6078),
  [2788] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1369),
  [2790] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5791),
  [2792] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1496),
  [2794] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6251),
  [2796] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4676),
  [2798] = {.entry = {.count = 1, .reusable = false}}, SHIFT(308),
  [2800] = {.entry = {.count = 1, .reusable = false}}, SHIFT(307),
  [2802] = {.entry = {.count = 1, .reusable = false}}, SHIFT(306),
  [2804] = {.entry = {.count = 1, .reusable = false}}, SHIFT(377),
  [2806] = {.entry = {.count = 1, .reusable = false}}, SHIFT(633),
  [2808] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1325),
  [2810] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7663),
  [2812] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(1381),
  [2815] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(3744),
  [2818] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(3743),
  [2821] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(5976),
  [2824] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(1308),
  [2827] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(5814),
  [2830] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(1515),
  [2833] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(6272),
  [2836] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(4690),
  [2839] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(332),
  [2842] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(331),
  [2845] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(330),
  [2848] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(409),
  [2851] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(1388),
  [2854] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(7655),
  [2857] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4240),
  [2859] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3714),
  [2861] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3713),
  [2863] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6030),
  [2865] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4218),
  [2867] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5807),
  [2869] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4399),
  [2871] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6341),
  [2873] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4681),
  [2875] = {.entry = {.count = 1, .reusable = false}}, SHIFT(323),
  [2877] = {.entry = {.count = 1, .reusable = false}}, SHIFT(322),
  [2879] = {.entry = {.count = 1, .reusable = false}}, SHIFT(321),
  [2881] = {.entry = {.count = 1, .reusable = false}}, SHIFT(397),
  [2883] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4252),
  [2885] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7658),
  [2887] = {.entry = {.count = 1, .reusable = false}}, SHIFT(629),
  [2889] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_unset_command_repeat1, 2), SHIFT_REPEAT(1326),
  [2892] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_unset_command_repeat1, 2), SHIFT_REPEAT(3786),
  [2895] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_unset_command_repeat1, 2), SHIFT_REPEAT(3789),
  [2898] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_unset_command_repeat1, 2), SHIFT_REPEAT(6078),
  [2901] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_unset_command_repeat1, 2), SHIFT_REPEAT(1369),
  [2904] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_unset_command_repeat1, 2), SHIFT_REPEAT(5791),
  [2907] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_unset_command_repeat1, 2), SHIFT_REPEAT(1496),
  [2910] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_unset_command_repeat1, 2), SHIFT_REPEAT(6251),
  [2913] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_unset_command_repeat1, 2), SHIFT_REPEAT(4676),
  [2916] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_unset_command_repeat1, 2), SHIFT_REPEAT(308),
  [2919] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_unset_command_repeat1, 2), SHIFT_REPEAT(307),
  [2922] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_unset_command_repeat1, 2), SHIFT_REPEAT(306),
  [2925] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_unset_command_repeat1, 2), SHIFT_REPEAT(377),
  [2928] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_unset_command_repeat1, 2), SHIFT_REPEAT(633),
  [2931] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_unset_command_repeat1, 2), SHIFT_REPEAT(1325),
  [2934] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_unset_command_repeat1, 2), SHIFT_REPEAT(7663),
  [2937] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4247),
  [2939] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4231),
  [2941] = {.entry = {.count = 1, .reusable = false}}, SHIFT(938),
  [2943] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1417),
  [2945] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1418),
  [2947] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1418),
  [2949] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1709),
  [2951] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1701),
  [2953] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1701),
  [2955] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1253),
  [2957] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1252),
  [2959] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1252),
  [2961] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1717),
  [2963] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1427),
  [2965] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5796),
  [2967] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1428),
  [2969] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1428),
  [2971] = {.entry = {.count = 1, .reusable = false}}, SHIFT(852),
  [2973] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(1399),
  [2976] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(4062),
  [2979] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(4063),
  [2982] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(5930),
  [2985] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(1405),
  [2988] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(5772),
  [2991] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(1605),
  [2994] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(6367),
  [2997] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(4660),
  [3000] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(128),
  [3003] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(129),
  [3006] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(130),
  [3009] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(132),
  [3012] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(1407),
  [3015] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(7605),
  [3018] = {.entry = {.count = 1, .reusable = false}}, SHIFT(815),
  [3020] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4310),
  [3022] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3888),
  [3024] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3893),
  [3026] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6035),
  [3028] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4233),
  [3030] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5775),
  [3032] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4575),
  [3034] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6248),
  [3036] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4658),
  [3038] = {.entry = {.count = 1, .reusable = false}}, SHIFT(286),
  [3040] = {.entry = {.count = 1, .reusable = false}}, SHIFT(106),
  [3042] = {.entry = {.count = 1, .reusable = false}}, SHIFT(279),
  [3044] = {.entry = {.count = 1, .reusable = false}}, SHIFT(345),
  [3046] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4299),
  [3048] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7668),
  [3050] = {.entry = {.count = 1, .reusable = false}}, SHIFT(760),
  [3052] = {.entry = {.count = 1, .reusable = false}}, SHIFT(711),
  [3054] = {.entry = {.count = 1, .reusable = true}}, SHIFT(762),
  [3056] = {.entry = {.count = 1, .reusable = false}}, SHIFT(763),
  [3058] = {.entry = {.count = 1, .reusable = true}}, SHIFT(763),
  [3060] = {.entry = {.count = 1, .reusable = true}}, SHIFT(117),
  [3062] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2836),
  [3064] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2589),
  [3066] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(1526),
  [3069] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(3817),
  [3072] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(3820),
  [3075] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(6061),
  [3078] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(1725),
  [3081] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(5784),
  [3084] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(1782),
  [3087] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(6252),
  [3090] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(4672),
  [3093] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(302),
  [3096] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(301),
  [3099] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(300),
  [3102] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(368),
  [3105] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(1524),
  [3108] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(7665),
  [3111] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2799),
  [3113] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1769),
  [3115] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5720),
  [3117] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1771),
  [3119] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1771),
  [3121] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1414),
  [3123] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1419),
  [3125] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1419),
  [3127] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_unset_command_repeat1, 2), SHIFT_REPEAT(1681),
  [3130] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_unset_command_repeat1, 2), SHIFT_REPEAT(4205),
  [3133] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_unset_command_repeat1, 2), SHIFT_REPEAT(4206),
  [3136] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_unset_command_repeat1, 2), SHIFT_REPEAT(6007),
  [3139] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_unset_command_repeat1, 2), SHIFT_REPEAT(1597),
  [3142] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_unset_command_repeat1, 2), SHIFT_REPEAT(5735),
  [3145] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_unset_command_repeat1, 2), SHIFT_REPEAT(1726),
  [3148] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_unset_command_repeat1, 2), SHIFT_REPEAT(6297),
  [3151] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_unset_command_repeat1, 2), SHIFT_REPEAT(4728),
  [3154] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_unset_command_repeat1, 2), SHIFT_REPEAT(180),
  [3157] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_unset_command_repeat1, 2), SHIFT_REPEAT(181),
  [3160] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_unset_command_repeat1, 2), SHIFT_REPEAT(184),
  [3163] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_unset_command_repeat1, 2), SHIFT_REPEAT(185),
  [3166] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_unset_command_repeat1, 2), SHIFT_REPEAT(664),
  [3169] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_unset_command_repeat1, 2), SHIFT_REPEAT(1696),
  [3172] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_unset_command_repeat1, 2), SHIFT_REPEAT(7612),
  [3175] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4331),
  [3177] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4319),
  [3179] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1650),
  [3181] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4193),
  [3183] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4192),
  [3185] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6047),
  [3187] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1705),
  [3189] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5736),
  [3191] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1788),
  [3193] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6257),
  [3195] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4624),
  [3197] = {.entry = {.count = 1, .reusable = false}}, SHIFT(243),
  [3199] = {.entry = {.count = 1, .reusable = false}}, SHIFT(244),
  [3201] = {.entry = {.count = 1, .reusable = false}}, SHIFT(245),
  [3203] = {.entry = {.count = 1, .reusable = false}}, SHIFT(219),
  [3205] = {.entry = {.count = 1, .reusable = false}}, SHIFT(709),
  [3207] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1647),
  [3209] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7618),
  [3211] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1865),
  [3213] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5783),
  [3215] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1866),
  [3217] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1866),
  [3219] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1610),
  [3221] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1609),
  [3223] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1609),
  [3225] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2756),
  [3227] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(1711),
  [3230] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(3995),
  [3233] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(3994),
  [3236] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(6039),
  [3239] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(1710),
  [3242] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(5720),
  [3245] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(1790),
  [3248] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(6357),
  [3251] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(4851),
  [3254] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(124),
  [3257] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(126),
  [3260] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(127),
  [3263] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(375),
  [3266] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(1708),
  [3269] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(7634),
  [3272] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1773),
  [3274] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5784),
  [3276] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1775),
  [3278] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1775),
  [3280] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1700),
  [3282] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5764),
  [3284] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1586),
  [3286] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1586),
  [3288] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(1630),
  [3291] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(4047),
  [3294] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(4037),
  [3297] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(6051),
  [3300] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(1675),
  [3303] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(5783),
  [3306] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(1896),
  [3309] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(6351),
  [3312] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(4843),
  [3315] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(69),
  [3318] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(73),
  [3321] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(75),
  [3324] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(339),
  [3327] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(1632),
  [3330] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(7629),
  [3333] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1526),
  [3335] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3817),
  [3337] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3820),
  [3339] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6061),
  [3341] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1725),
  [3343] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1782),
  [3345] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6252),
  [3347] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4672),
  [3349] = {.entry = {.count = 1, .reusable = false}}, SHIFT(302),
  [3351] = {.entry = {.count = 1, .reusable = false}}, SHIFT(301),
  [3353] = {.entry = {.count = 1, .reusable = false}}, SHIFT(300),
  [3355] = {.entry = {.count = 1, .reusable = false}}, SHIFT(368),
  [3357] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1524),
  [3359] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7665),
  [3361] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4271),
  [3363] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3728),
  [3365] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3726),
  [3367] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5994),
  [3369] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4259),
  [3371] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5811),
  [3373] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4600),
  [3375] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6284),
  [3377] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4686),
  [3379] = {.entry = {.count = 1, .reusable = false}}, SHIFT(329),
  [3381] = {.entry = {.count = 1, .reusable = false}}, SHIFT(328),
  [3383] = {.entry = {.count = 1, .reusable = false}}, SHIFT(327),
  [3385] = {.entry = {.count = 1, .reusable = false}}, SHIFT(405),
  [3387] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4406),
  [3389] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7656),
  [3391] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1711),
  [3393] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3995),
  [3395] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3994),
  [3397] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6039),
  [3399] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1710),
  [3401] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1790),
  [3403] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6357),
  [3405] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4851),
  [3407] = {.entry = {.count = 1, .reusable = false}}, SHIFT(124),
  [3409] = {.entry = {.count = 1, .reusable = false}}, SHIFT(126),
  [3411] = {.entry = {.count = 1, .reusable = false}}, SHIFT(127),
  [3413] = {.entry = {.count = 1, .reusable = false}}, SHIFT(375),
  [3415] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1708),
  [3417] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7634),
  [3419] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2811),
  [3421] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1482),
  [3423] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5772),
  [3425] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1690),
  [3427] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1690),
  [3429] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1681),
  [3431] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4205),
  [3433] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4206),
  [3435] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6007),
  [3437] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1597),
  [3439] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5735),
  [3441] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1726),
  [3443] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6297),
  [3445] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4728),
  [3447] = {.entry = {.count = 1, .reusable = false}}, SHIFT(180),
  [3449] = {.entry = {.count = 1, .reusable = false}}, SHIFT(181),
  [3451] = {.entry = {.count = 1, .reusable = false}}, SHIFT(184),
  [3453] = {.entry = {.count = 1, .reusable = false}}, SHIFT(185),
  [3455] = {.entry = {.count = 1, .reusable = false}}, SHIFT(695),
  [3457] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1696),
  [3459] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7612),
  [3461] = {.entry = {.count = 1, .reusable = false}}, SHIFT(667),
  [3463] = {.entry = {.count = 1, .reusable = false}}, SHIFT(664),
  [3465] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2782),
  [3467] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2777),
  [3469] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2743),
  [3471] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1630),
  [3473] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4047),
  [3475] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4037),
  [3477] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6051),
  [3479] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1675),
  [3481] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1896),
  [3483] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6351),
  [3485] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4843),
  [3487] = {.entry = {.count = 1, .reusable = false}}, SHIFT(69),
  [3489] = {.entry = {.count = 1, .reusable = false}}, SHIFT(73),
  [3491] = {.entry = {.count = 1, .reusable = false}}, SHIFT(75),
  [3493] = {.entry = {.count = 1, .reusable = false}}, SHIFT(339),
  [3495] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1632),
  [3497] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7629),
  [3499] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_repeat2, 2, .production_id = 38), SHIFT_REPEAT(1794),
  [3502] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_repeat2, 2, .production_id = 38), SHIFT_REPEAT(3697),
  [3505] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_repeat2, 2, .production_id = 38), SHIFT_REPEAT(4138),
  [3508] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_repeat2, 2, .production_id = 38), SHIFT_REPEAT(1823),
  [3511] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_command_repeat2, 2, .production_id = 38), SHIFT_REPEAT(6848),
  [3514] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_command_repeat2, 2, .production_id = 38), SHIFT_REPEAT(1795),
  [3517] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(1688),
  [3520] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(1662),
  [3523] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(1575),
  [3526] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4370),
  [3528] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4364),
  [3530] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_unset_command_repeat1, 2), SHIFT_REPEAT(1650),
  [3533] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_unset_command_repeat1, 2), SHIFT_REPEAT(4193),
  [3536] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_unset_command_repeat1, 2), SHIFT_REPEAT(4192),
  [3539] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_unset_command_repeat1, 2), SHIFT_REPEAT(6047),
  [3542] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_unset_command_repeat1, 2), SHIFT_REPEAT(1705),
  [3545] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_unset_command_repeat1, 2), SHIFT_REPEAT(5736),
  [3548] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_unset_command_repeat1, 2), SHIFT_REPEAT(1788),
  [3551] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_unset_command_repeat1, 2), SHIFT_REPEAT(6257),
  [3554] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_unset_command_repeat1, 2), SHIFT_REPEAT(4624),
  [3557] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_unset_command_repeat1, 2), SHIFT_REPEAT(243),
  [3560] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_unset_command_repeat1, 2), SHIFT_REPEAT(244),
  [3563] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_unset_command_repeat1, 2), SHIFT_REPEAT(245),
  [3566] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_unset_command_repeat1, 2), SHIFT_REPEAT(219),
  [3569] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_unset_command_repeat1, 2), SHIFT_REPEAT(709),
  [3572] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_unset_command_repeat1, 2), SHIFT_REPEAT(1647),
  [3575] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_unset_command_repeat1, 2), SHIFT_REPEAT(7618),
  [3578] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2081),
  [3580] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5733),
  [3582] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2086),
  [3584] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2086),
  [3586] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2840),
  [3588] = {.entry = {.count = 1, .reusable = false}}, SHIFT(734),
  [3590] = {.entry = {.count = 1, .reusable = false}}, SHIFT(713),
  [3592] = {.entry = {.count = 1, .reusable = true}}, SHIFT(712),
  [3594] = {.entry = {.count = 1, .reusable = false}}, SHIFT(730),
  [3596] = {.entry = {.count = 1, .reusable = true}}, SHIFT(730),
  [3598] = {.entry = {.count = 1, .reusable = true}}, SHIFT(225),
  [3600] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3281),
  [3602] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2945),
  [3604] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3283),
  [3606] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1839),
  [3608] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4000),
  [3610] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3997),
  [3612] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6060),
  [3614] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1832),
  [3616] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2116),
  [3618] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6333),
  [3620] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4854),
  [3622] = {.entry = {.count = 1, .reusable = false}}, SHIFT(116),
  [3624] = {.entry = {.count = 1, .reusable = false}}, SHIFT(119),
  [3626] = {.entry = {.count = 1, .reusable = false}}, SHIFT(121),
  [3628] = {.entry = {.count = 1, .reusable = false}}, SHIFT(182),
  [3630] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1843),
  [3632] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7633),
  [3634] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(1918),
  [3637] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(1786),
  [3640] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(1922),
  [3643] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1548),
  [3645] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1549),
  [3647] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1549),
  [3649] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1880),
  [3651] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1881),
  [3653] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1881),
  [3655] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1902),
  [3657] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1915),
  [3659] = {.entry = {.count = 1, .reusable = false}}, SHIFT(743),
  [3661] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1903),
  [3663] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1918),
  [3665] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1786),
  [3667] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1922),
  [3669] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4504),
  [3671] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3824),
  [3673] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3825),
  [3675] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6053),
  [3677] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4315),
  [3679] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5782),
  [3681] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4706),
  [3683] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6263),
  [3685] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4671),
  [3687] = {.entry = {.count = 1, .reusable = false}}, SHIFT(299),
  [3689] = {.entry = {.count = 1, .reusable = false}}, SHIFT(298),
  [3691] = {.entry = {.count = 1, .reusable = false}}, SHIFT(293),
  [3693] = {.entry = {.count = 1, .reusable = false}}, SHIFT(361),
  [3695] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4506),
  [3697] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7666),
  [3699] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2095),
  [3701] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5721),
  [3703] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2091),
  [3705] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2091),
  [3707] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4555),
  [3709] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4562),
  [3711] = {.entry = {.count = 1, .reusable = false}}, SHIFT(731),
  [3713] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5062),
  [3715] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4006),
  [3717] = {.entry = {.count = 1, .reusable = false}}, SHIFT(434),
  [3719] = {.entry = {.count = 1, .reusable = false}}, SHIFT(440),
  [3721] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3634),
  [3723] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1654),
  [3725] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6416),
  [3727] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3739),
  [3729] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4005),
  [3731] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6005),
  [3733] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5034),
  [3735] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5741),
  [3737] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4837),
  [3739] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6352),
  [3741] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4857),
  [3743] = {.entry = {.count = 1, .reusable = false}}, SHIFT(110),
  [3745] = {.entry = {.count = 1, .reusable = false}}, SHIFT(111),
  [3747] = {.entry = {.count = 1, .reusable = false}}, SHIFT(113),
  [3749] = {.entry = {.count = 1, .reusable = false}}, SHIFT(366),
  [3751] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6155),
  [3753] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5048),
  [3755] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7632),
  [3757] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6406),
  [3759] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3278),
  [3761] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_unset_command_repeat1, 2), SHIFT_REPEAT(1902),
  [3764] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_unset_command_repeat1, 2), SHIFT_REPEAT(1915),
  [3767] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_unset_command_repeat1, 2), SHIFT_REPEAT(731),
  [3770] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_unset_command_repeat1, 2), SHIFT_REPEAT(1903),
  [3773] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1956),
  [3775] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5822),
  [3777] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1960),
  [3779] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1960),
  [3781] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4522),
  [3783] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4025),
  [3785] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4021),
  [3787] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5981),
  [3789] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4359),
  [3791] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5723),
  [3793] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4395),
  [3795] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6288),
  [3797] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4845),
  [3799] = {.entry = {.count = 1, .reusable = false}}, SHIFT(79),
  [3801] = {.entry = {.count = 1, .reusable = false}}, SHIFT(80),
  [3803] = {.entry = {.count = 1, .reusable = false}}, SHIFT(83),
  [3805] = {.entry = {.count = 1, .reusable = false}}, SHIFT(350),
  [3807] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4529),
  [3809] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7630),
  [3811] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3365),
  [3813] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6442),
  [3815] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4518),
  [3817] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4401),
  [3819] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4599),
  [3821] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4486),
  [3823] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4484),
  [3825] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1484),
  [3827] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1492),
  [3829] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1492),
  [3831] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(1839),
  [3834] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(4000),
  [3837] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(3997),
  [3840] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(6060),
  [3843] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(1832),
  [3846] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(5733),
  [3849] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(2116),
  [3852] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(6333),
  [3855] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(4854),
  [3858] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(116),
  [3861] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(119),
  [3864] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(121),
  [3867] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(182),
  [3870] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(1843),
  [3873] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(7633),
  [3876] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1846),
  [3878] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4061),
  [3880] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4054),
  [3882] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5926),
  [3884] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1916),
  [3886] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1946),
  [3888] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6348),
  [3890] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4819),
  [3892] = {.entry = {.count = 1, .reusable = false}}, SHIFT(374),
  [3894] = {.entry = {.count = 1, .reusable = false}}, SHIFT(410),
  [3896] = {.entry = {.count = 1, .reusable = false}}, SHIFT(247),
  [3898] = {.entry = {.count = 1, .reusable = false}}, SHIFT(272),
  [3900] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1845),
  [3902] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7627),
  [3904] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6466),
  [3906] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6411),
  [3908] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_command_repeat1, 2), SHIFT_REPEAT(2330),
  [3911] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_command_repeat1, 2), SHIFT_REPEAT(2342),
  [3914] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_command_repeat1, 2), SHIFT_REPEAT(746),
  [3917] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_declaration_command_repeat1, 2), SHIFT_REPEAT(6696),
  [3920] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_declaration_command_repeat1, 2), SHIFT_REPEAT(2331),
  [3923] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(1846),
  [3926] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(4061),
  [3929] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(4054),
  [3932] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(5926),
  [3935] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(1916),
  [3938] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(5721),
  [3941] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(1946),
  [3944] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(6348),
  [3947] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(4819),
  [3950] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(374),
  [3953] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(410),
  [3956] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(247),
  [3959] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(272),
  [3962] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(1845),
  [3965] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(7627),
  [3968] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2330),
  [3970] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2342),
  [3972] = {.entry = {.count = 1, .reusable = false}}, SHIFT(777),
  [3974] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6696),
  [3976] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2331),
  [3978] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_repeat2, 2, .production_id = 38), SHIFT_REPEAT(2211),
  [3981] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_command_repeat2, 2, .production_id = 38), SHIFT_REPEAT(4082),
  [3984] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_repeat2, 2, .production_id = 38), SHIFT_REPEAT(3672),
  [3987] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_command_repeat2, 2, .production_id = 38), SHIFT_REPEAT(3715),
  [3990] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_command_repeat2, 2, .production_id = 38), SHIFT_REPEAT(4079),
  [3993] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_repeat2, 2, .production_id = 38), SHIFT_REPEAT(6014),
  [3996] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_command_repeat2, 2, .production_id = 38), SHIFT_REPEAT(2033),
  [3999] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_command_repeat2, 2, .production_id = 38), SHIFT_REPEAT(5780),
  [4002] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_command_repeat2, 2, .production_id = 38), SHIFT_REPEAT(2211),
  [4005] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_repeat2, 2, .production_id = 38), SHIFT_REPEAT(2268),
  [4008] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_repeat2, 2, .production_id = 38), SHIFT_REPEAT(6342),
  [4011] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_command_repeat2, 2, .production_id = 38), SHIFT_REPEAT(4779),
  [4014] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_repeat2, 2, .production_id = 38), SHIFT_REPEAT(340),
  [4017] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_command_repeat2, 2, .production_id = 38), SHIFT_REPEAT(341),
  [4020] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_command_repeat2, 2, .production_id = 38), SHIFT_REPEAT(342),
  [4023] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_command_repeat2, 2, .production_id = 38), SHIFT_REPEAT(258),
  [4026] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_command_repeat2, 2, .production_id = 38), SHIFT_REPEAT(6889),
  [4029] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_command_repeat2, 2, .production_id = 38), SHIFT_REPEAT(2203),
  [4032] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_command_repeat2, 2, .production_id = 38), SHIFT_REPEAT(2558),
  [4035] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_command_repeat2, 2, .production_id = 38), SHIFT_REPEAT(7625),
  [4038] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6452),
  [4040] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6395),
  [4042] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3368),
  [4044] = {.entry = {.count = 1, .reusable = false}}, SHIFT(726),
  [4046] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2843),
  [4048] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2848),
  [4050] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4582),
  [4052] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4481),
  [4054] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1805),
  [4056] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1804),
  [4058] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1804),
  [4060] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6375),
  [4062] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6397),
  [4064] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6447),
  [4066] = {.entry = {.count = 1, .reusable = false}}, SHIFT(746),
  [4068] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4472),
  [4070] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4362),
  [4072] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4476),
  [4074] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4828),
  [4076] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4049),
  [4078] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4048),
  [4080] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5956),
  [4082] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4593),
  [4084] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5727),
  [4086] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4451),
  [4088] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6276),
  [4090] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4830),
  [4092] = {.entry = {.count = 1, .reusable = false}}, SHIFT(165),
  [4094] = {.entry = {.count = 1, .reusable = false}}, SHIFT(149),
  [4096] = {.entry = {.count = 1, .reusable = false}}, SHIFT(146),
  [4098] = {.entry = {.count = 1, .reusable = false}}, SHIFT(294),
  [4100] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4621),
  [4102] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7628),
  [4104] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(1968),
  [4107] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(2074),
  [4110] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(1972),
  [4113] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2320),
  [4115] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5780),
  [4117] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2326),
  [4119] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2326),
  [4121] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4705),
  [4123] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4524),
  [4125] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4733),
  [4127] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4479),
  [4129] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4488),
  [4131] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1968),
  [4133] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2074),
  [4135] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1972),
  [4137] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4834),
  [4139] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4495),
  [4141] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4816),
  [4143] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4838),
  [4145] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4835),
  [4147] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2419),
  [4149] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4114),
  [4151] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4112),
  [4153] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5894),
  [4155] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2373),
  [4157] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5754),
  [4159] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2419),
  [4161] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2469),
  [4163] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6298),
  [4165] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4730),
  [4167] = {.entry = {.count = 1, .reusable = false}}, SHIFT(278),
  [4169] = {.entry = {.count = 1, .reusable = true}}, SHIFT(136),
  [4171] = {.entry = {.count = 1, .reusable = true}}, SHIFT(289),
  [4173] = {.entry = {.count = 1, .reusable = true}}, SHIFT(248),
  [4175] = {.entry = {.count = 1, .reusable = false}}, SHIFT(803),
  [4177] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6636),
  [4179] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2395),
  [4181] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7623),
  [4183] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4692),
  [4185] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4689),
  [4187] = {.entry = {.count = 1, .reusable = false}}, SHIFT(797),
  [4189] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_command_repeat1, 2), SHIFT_REPEAT(2419),
  [4192] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_declaration_command_repeat1, 2), SHIFT_REPEAT(4114),
  [4195] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_declaration_command_repeat1, 2), SHIFT_REPEAT(4112),
  [4198] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_command_repeat1, 2), SHIFT_REPEAT(5894),
  [4201] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_declaration_command_repeat1, 2), SHIFT_REPEAT(2373),
  [4204] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_declaration_command_repeat1, 2), SHIFT_REPEAT(5754),
  [4207] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_declaration_command_repeat1, 2), SHIFT_REPEAT(2419),
  [4210] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_command_repeat1, 2), SHIFT_REPEAT(2469),
  [4213] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_command_repeat1, 2), SHIFT_REPEAT(6298),
  [4216] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_declaration_command_repeat1, 2), SHIFT_REPEAT(4730),
  [4219] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_command_repeat1, 2), SHIFT_REPEAT(278),
  [4222] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_declaration_command_repeat1, 2), SHIFT_REPEAT(136),
  [4225] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_declaration_command_repeat1, 2), SHIFT_REPEAT(289),
  [4228] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_declaration_command_repeat1, 2), SHIFT_REPEAT(248),
  [4231] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_command_repeat1, 2), SHIFT_REPEAT(803),
  [4234] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_declaration_command_repeat1, 2), SHIFT_REPEAT(6636),
  [4237] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_declaration_command_repeat1, 2), SHIFT_REPEAT(2395),
  [4240] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_declaration_command_repeat1, 2), SHIFT_REPEAT(7623),
  [4243] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4614),
  [4245] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4795),
  [4247] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_command_repeat1, 1),
  [4249] = {.entry = {.count = 1, .reusable = false}}, SHIFT(622),
  [4251] = {.entry = {.count = 1, .reusable = false}}, SHIFT(450),
  [4253] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__statement_not_pipeline, 1),
  [4255] = {.entry = {.count = 1, .reusable = false}}, SHIFT(7487),
  [4257] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__statements, 1),
  [4259] = {.entry = {.count = 1, .reusable = true}}, SHIFT(622),
  [4261] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_command_repeat1, 1),
  [4263] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_command_repeat1, 1), SHIFT(6662),
  [4266] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3881),
  [4268] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3939),
  [4270] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_list, 3),
  [4272] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_list, 3),
  [4274] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(2337),
  [4277] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(2302),
  [4280] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(2340),
  [4283] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_command_name, 1), SHIFT(6956),
  [4286] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3881),
  [4288] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3924),
  [4290] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(2271),
  [4293] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(4164),
  [4296] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(4144),
  [4299] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(5923),
  [4302] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(2298),
  [4305] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(5730),
  [4308] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(2271),
  [4311] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(2420),
  [4314] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(6316),
  [4317] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(4691),
  [4320] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(261),
  [4323] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(262),
  [4326] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(263),
  [4329] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(235),
  [4332] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(2309),
  [4335] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(7621),
  [4338] = {.entry = {.count = 1, .reusable = false}}, SHIFT(655),
  [4340] = {.entry = {.count = 1, .reusable = true}}, SHIFT(655),
  [4342] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_concatenation_repeat1, 2), SHIFT_REPEAT(3881),
  [4345] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_concatenation_repeat1, 2), SHIFT_REPEAT(3881),
  [4348] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__statement_not_pipeline, 1),
  [4350] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4921),
  [4352] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4774),
  [4354] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4941),
  [4356] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4901),
  [4358] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4906),
  [4360] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4640),
  [4362] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4578),
  [4364] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4626),
  [4366] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4627),
  [4368] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4633),
  [4370] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_herestring_redirect, 2),
  [4372] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_herestring_redirect, 2),
  [4374] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3896),
  [4376] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3896),
  [4378] = {.entry = {.count = 1, .reusable = false}}, SHIFT(625),
  [4380] = {.entry = {.count = 1, .reusable = false}}, SHIFT(436),
  [4382] = {.entry = {.count = 1, .reusable = false}}, SHIFT(7514),
  [4384] = {.entry = {.count = 1, .reusable = true}}, SHIFT(625),
  [4386] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_command_repeat1, 1), SHIFT(6619),
  [4389] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3940),
  [4391] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3984),
  [4393] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3988),
  [4395] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2362),
  [4397] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5730),
  [4399] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2363),
  [4401] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2363),
  [4403] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_concatenation_repeat1, 2), SHIFT_REPEAT(3940),
  [4406] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_concatenation_repeat1, 2), SHIFT_REPEAT(3940),
  [4409] = {.entry = {.count = 1, .reusable = false}}, SHIFT(624),
  [4411] = {.entry = {.count = 1, .reusable = true}}, SHIFT(624),
  [4413] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3940),
  [4415] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(2369),
  [4418] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(3983),
  [4421] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(3982),
  [4424] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(5950),
  [4427] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(2396),
  [4430] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(5725),
  [4433] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(2369),
  [4436] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(2507),
  [4439] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(6313),
  [4442] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(4839),
  [4445] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(140),
  [4448] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(141),
  [4451] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(142),
  [4454] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(88),
  [4457] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(2371),
  [4460] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(7636),
  [4463] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_command_repeat2, 2, .production_id = 37),
  [4465] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_command_repeat2, 2, .production_id = 37),
  [4467] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_herestring_redirect, 2, .production_id = 10),
  [4469] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_herestring_redirect, 2, .production_id = 10),
  [4471] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(2422),
  [4474] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(2389),
  [4477] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(2431),
  [4480] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_command_repeat2, 2, .production_id = 36),
  [4482] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_command_repeat2, 2, .production_id = 36),
  [4484] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2422),
  [4486] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2389),
  [4488] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2431),
  [4490] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2367),
  [4492] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2386),
  [4494] = {.entry = {.count = 1, .reusable = false}}, SHIFT(875),
  [4496] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2365),
  [4498] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2438),
  [4500] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3945),
  [4502] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3944),
  [4504] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5883),
  [4506] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2356),
  [4508] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5779),
  [4510] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2438),
  [4512] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2560),
  [4514] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6349),
  [4516] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4826),
  [4518] = {.entry = {.count = 1, .reusable = false}}, SHIFT(178),
  [4520] = {.entry = {.count = 1, .reusable = true}}, SHIFT(183),
  [4522] = {.entry = {.count = 1, .reusable = true}}, SHIFT(191),
  [4524] = {.entry = {.count = 1, .reusable = true}}, SHIFT(115),
  [4526] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2437),
  [4528] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7639),
  [4530] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_command_repeat2, 1, .production_id = 17),
  [4532] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_command_repeat2, 1, .production_id = 17),
  [4534] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_unset_command_repeat1, 2), SHIFT_REPEAT(2367),
  [4537] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_unset_command_repeat1, 2), SHIFT_REPEAT(2386),
  [4540] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_unset_command_repeat1, 2), SHIFT_REPEAT(875),
  [4543] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_unset_command_repeat1, 2), SHIFT_REPEAT(2365),
  [4546] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_command_repeat2, 1, .production_id = 16),
  [4548] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_command_repeat2, 1, .production_id = 16),
  [4550] = {.entry = {.count = 1, .reusable = false}}, SHIFT(865),
  [4552] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_herestring_redirect, 3, .production_id = 4),
  [4554] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_herestring_redirect, 3, .production_id = 4),
  [4556] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2567),
  [4558] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5779),
  [4560] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2536),
  [4562] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2536),
  [4564] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_herestring_redirect, 3, .production_id = 23),
  [4566] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_herestring_redirect, 3, .production_id = 23),
  [4568] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_command_name, 1), SHIFT(6890),
  [4571] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3839),
  [4573] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3843),
  [4575] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_concatenation_repeat1, 2), SHIFT_REPEAT(3896),
  [4578] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_concatenation_repeat1, 2), SHIFT_REPEAT(3896),
  [4581] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(2438),
  [4584] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(3945),
  [4587] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(3944),
  [4590] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(5883),
  [4593] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(2356),
  [4596] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(5779),
  [4599] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(2438),
  [4602] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(2560),
  [4605] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(6349),
  [4608] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(4826),
  [4611] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(178),
  [4614] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(183),
  [4617] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(191),
  [4620] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(115),
  [4623] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(2437),
  [4626] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(7639),
  [4629] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2487),
  [4631] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3797),
  [4633] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3793),
  [4635] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5876),
  [4637] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2458),
  [4639] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5832),
  [4641] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2487),
  [4643] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2651),
  [4645] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6320),
  [4647] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4729),
  [4649] = {.entry = {.count = 1, .reusable = false}}, SHIFT(376),
  [4651] = {.entry = {.count = 1, .reusable = true}}, SHIFT(61),
  [4653] = {.entry = {.count = 1, .reusable = true}}, SHIFT(373),
  [4655] = {.entry = {.count = 1, .reusable = true}}, SHIFT(250),
  [4657] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2493),
  [4659] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7649),
  [4661] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2664),
  [4663] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5724),
  [4665] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2665),
  [4667] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2665),
  [4669] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_command_name, 1), SHIFT(7352),
  [4672] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4027),
  [4674] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4027),
  [4676] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5135),
  [4678] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5109),
  [4680] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5136),
  [4682] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5133),
  [4684] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5134),
  [4686] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__statements, 1),
  [4688] = {.entry = {.count = 1, .reusable = false}}, SHIFT(800),
  [4690] = {.entry = {.count = 1, .reusable = false}}, SHIFT(433),
  [4692] = {.entry = {.count = 1, .reusable = false}}, SHIFT(7349),
  [4694] = {.entry = {.count = 1, .reusable = true}}, SHIFT(800),
  [4696] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_command_repeat1, 1), SHIFT(6668),
  [4699] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3707),
  [4701] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3707),
  [4703] = {.entry = {.count = 1, .reusable = true}}, SHIFT(824),
  [4705] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6634),
  [4707] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1052),
  [4709] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1572),
  [4711] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4080),
  [4713] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(2487),
  [4716] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(3797),
  [4719] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(3793),
  [4722] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(5876),
  [4725] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(2458),
  [4728] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(5832),
  [4731] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(2487),
  [4734] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(2651),
  [4737] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(6320),
  [4740] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(4729),
  [4743] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(376),
  [4746] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(61),
  [4749] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(373),
  [4752] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(250),
  [4755] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(2493),
  [4758] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(7649),
  [4761] = {.entry = {.count = 1, .reusable = true}}, SHIFT(966),
  [4763] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5178),
  [4765] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3934),
  [4767] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3933),
  [4769] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5870),
  [4771] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5064),
  [4773] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5797),
  [4775] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5178),
  [4777] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5327),
  [4779] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6295),
  [4781] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4823),
  [4783] = {.entry = {.count = 1, .reusable = false}}, SHIFT(199),
  [4785] = {.entry = {.count = 1, .reusable = true}}, SHIFT(208),
  [4787] = {.entry = {.count = 1, .reusable = true}}, SHIFT(209),
  [4789] = {.entry = {.count = 1, .reusable = true}}, SHIFT(123),
  [4791] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5179),
  [4793] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7640),
  [4795] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2610),
  [4797] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5743),
  [4799] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2609),
  [4801] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2609),
  [4803] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5182),
  [4805] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5182),
  [4807] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5170),
  [4809] = {.entry = {.count = 1, .reusable = false}}, SHIFT(441),
  [4811] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_command_repeat1, 1), SHIFT(6578),
  [4814] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4153),
  [4816] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2539),
  [4818] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4103),
  [4820] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4101),
  [4822] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5879),
  [4824] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2453),
  [4826] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5769),
  [4828] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2539),
  [4830] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2667),
  [4832] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6277),
  [4834] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4751),
  [4836] = {.entry = {.count = 1, .reusable = false}}, SHIFT(296),
  [4838] = {.entry = {.count = 1, .reusable = true}}, SHIFT(297),
  [4840] = {.entry = {.count = 1, .reusable = true}}, SHIFT(336),
  [4842] = {.entry = {.count = 1, .reusable = true}}, SHIFT(254),
  [4844] = {.entry = {.count = 1, .reusable = false}}, SHIFT(962),
  [4846] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2538),
  [4848] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7624),
  [4850] = {.entry = {.count = 1, .reusable = true}}, SHIFT(920),
  [4852] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__literal_repeat1, 2), SHIFT_REPEAT(1052),
  [4855] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2655),
  [4857] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5832),
  [4859] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2653),
  [4861] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2653),
  [4863] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1912),
  [4865] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2516),
  [4867] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5725),
  [4869] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2515),
  [4871] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2515),
  [4873] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1105),
  [4875] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4179),
  [4877] = {.entry = {.count = 1, .reusable = false}}, SHIFT(979),
  [4879] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_concatenation_repeat1, 2), SHIFT_REPEAT(3707),
  [4882] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_concatenation_repeat1, 2), SHIFT_REPEAT(3707),
  [4885] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(2489),
  [4888] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(3978),
  [4891] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(3977),
  [4894] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(5936),
  [4897] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(2460),
  [4900] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(5743),
  [4903] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(2489),
  [4906] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(2604),
  [4909] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(6283),
  [4912] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(4836),
  [4915] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(147),
  [4918] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(156),
  [4921] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(157),
  [4924] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(100),
  [4927] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(2490),
  [4930] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(7637),
  [4933] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3947),
  [4935] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3947),
  [4937] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_concatenation_repeat1, 2), SHIFT_REPEAT(3947),
  [4940] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_concatenation_repeat1, 2), SHIFT_REPEAT(3947),
  [4943] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2489),
  [4945] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3978),
  [4947] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3977),
  [4949] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5936),
  [4951] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2460),
  [4953] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5743),
  [4955] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2489),
  [4957] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2604),
  [4959] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6283),
  [4961] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4836),
  [4963] = {.entry = {.count = 1, .reusable = false}}, SHIFT(147),
  [4965] = {.entry = {.count = 1, .reusable = true}}, SHIFT(156),
  [4967] = {.entry = {.count = 1, .reusable = true}}, SHIFT(157),
  [4969] = {.entry = {.count = 1, .reusable = true}}, SHIFT(100),
  [4971] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2490),
  [4973] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7637),
  [4975] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2572),
  [4977] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2480),
  [4979] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2578),
  [4981] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4133),
  [4983] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(2572),
  [4986] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(2480),
  [4989] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(2578),
  [4992] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4129),
  [4994] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_unset_command_repeat1, 2), SHIFT_REPEAT(2539),
  [4997] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_unset_command_repeat1, 2), SHIFT_REPEAT(4103),
  [5000] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_unset_command_repeat1, 2), SHIFT_REPEAT(4101),
  [5003] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_unset_command_repeat1, 2), SHIFT_REPEAT(5879),
  [5006] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_unset_command_repeat1, 2), SHIFT_REPEAT(2453),
  [5009] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_unset_command_repeat1, 2), SHIFT_REPEAT(5769),
  [5012] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_unset_command_repeat1, 2), SHIFT_REPEAT(2539),
  [5015] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_unset_command_repeat1, 2), SHIFT_REPEAT(2667),
  [5018] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_unset_command_repeat1, 2), SHIFT_REPEAT(6277),
  [5021] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_unset_command_repeat1, 2), SHIFT_REPEAT(4751),
  [5024] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_unset_command_repeat1, 2), SHIFT_REPEAT(296),
  [5027] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_unset_command_repeat1, 2), SHIFT_REPEAT(297),
  [5030] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_unset_command_repeat1, 2), SHIFT_REPEAT(336),
  [5033] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_unset_command_repeat1, 2), SHIFT_REPEAT(254),
  [5036] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_unset_command_repeat1, 2), SHIFT_REPEAT(979),
  [5039] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_unset_command_repeat1, 2), SHIFT_REPEAT(2538),
  [5042] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_unset_command_repeat1, 2), SHIFT_REPEAT(7624),
  [5045] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3936),
  [5047] = {.entry = {.count = 1, .reusable = true}}, SHIFT(850),
  [5049] = {.entry = {.count = 1, .reusable = false}}, SHIFT(794),
  [5051] = {.entry = {.count = 1, .reusable = true}}, SHIFT(794),
  [5053] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_concatenation_repeat1, 2), SHIFT_REPEAT(4027),
  [5056] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_concatenation_repeat1, 2), SHIFT_REPEAT(4027),
  [5059] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2646),
  [5061] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5769),
  [5063] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2649),
  [5065] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2649),
  [5067] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_variable_assignment, 3, .production_id = 25),
  [5069] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_variable_assignment, 3, .production_id = 25),
  [5071] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3894),
  [5073] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3894),
  [5075] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_command_name, 1), SHIFT(7486),
  [5078] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1229),
  [5080] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3733),
  [5082] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3733),
  [5084] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2304),
  [5086] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5753),
  [5088] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2307),
  [5090] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2307),
  [5092] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1260),
  [5094] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_command_repeat1, 1), SHIFT(6705),
  [5097] = {.entry = {.count = 1, .reusable = false}}, SHIFT(780),
  [5099] = {.entry = {.count = 1, .reusable = false}}, SHIFT(444),
  [5101] = {.entry = {.count = 1, .reusable = true}}, SHIFT(780),
  [5103] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_command_repeat1, 1), SHIFT(6604),
  [5106] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3745),
  [5108] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2613),
  [5110] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3835),
  [5112] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3818),
  [5114] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6082),
  [5116] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2632),
  [5118] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5837),
  [5120] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2613),
  [5122] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2869),
  [5124] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6366),
  [5126] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4737),
  [5128] = {.entry = {.count = 1, .reusable = false}}, SHIFT(384),
  [5130] = {.entry = {.count = 1, .reusable = true}}, SHIFT(383),
  [5132] = {.entry = {.count = 1, .reusable = true}}, SHIFT(382),
  [5134] = {.entry = {.count = 1, .reusable = true}}, SHIFT(210),
  [5136] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2612),
  [5138] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7647),
  [5140] = {.entry = {.count = 1, .reusable = false}}, SHIFT(827),
  [5142] = {.entry = {.count = 1, .reusable = true}}, SHIFT(827),
  [5144] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4034),
  [5146] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4034),
  [5148] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3750),
  [5150] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2468),
  [5152] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5754),
  [5154] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2467),
  [5156] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2467),
  [5158] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2625),
  [5160] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2624),
  [5162] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2624),
  [5164] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4019),
  [5166] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4020),
  [5168] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_concatenation_repeat1, 2), SHIFT_REPEAT(3733),
  [5171] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_concatenation_repeat1, 2), SHIFT_REPEAT(3733),
  [5174] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_declaration_command_repeat1, 1),
  [5176] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_declaration_command_repeat1, 1),
  [5178] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_declaration_command_repeat1, 1, .production_id = 1),
  [5180] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_declaration_command_repeat1, 1, .production_id = 1),
  [5182] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3891),
  [5184] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3752),
  [5186] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3756),
  [5188] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(2613),
  [5191] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(3835),
  [5194] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(3818),
  [5197] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(6082),
  [5200] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(2632),
  [5203] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(5837),
  [5206] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(2613),
  [5209] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(2869),
  [5212] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(6366),
  [5215] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(4737),
  [5218] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(384),
  [5221] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(383),
  [5224] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(382),
  [5227] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(210),
  [5230] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(2612),
  [5233] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(7647),
  [5236] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_concatenation_repeat1, 2), SHIFT_REPEAT(3891),
  [5239] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_concatenation_repeat1, 2), SHIFT_REPEAT(3891),
  [5242] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_command_name, 1), SHIFT(7396),
  [5245] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3891),
  [5247] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3781),
  [5249] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3785),
  [5251] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_concatenation_repeat1, 2), SHIFT_REPEAT(3894),
  [5254] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_concatenation_repeat1, 2), SHIFT_REPEAT(3894),
  [5257] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5378),
  [5259] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3784),
  [5261] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3783),
  [5263] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5892),
  [5265] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5240),
  [5267] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5830),
  [5269] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5378),
  [5271] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5492),
  [5273] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6294),
  [5275] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4718),
  [5277] = {.entry = {.count = 1, .reusable = false}}, SHIFT(371),
  [5279] = {.entry = {.count = 1, .reusable = true}}, SHIFT(370),
  [5281] = {.entry = {.count = 1, .reusable = true}}, SHIFT(369),
  [5283] = {.entry = {.count = 1, .reusable = true}}, SHIFT(264),
  [5285] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5373),
  [5287] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7650),
  [5289] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__literal_repeat1, 2), SHIFT_REPEAT(1229),
  [5292] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__literal_repeat1, 2), SHIFT_REPEAT(1260),
  [5295] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5352),
  [5297] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5352),
  [5299] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5359),
  [5301] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5293),
  [5303] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5206),
  [5305] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5305),
  [5307] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5398),
  [5309] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5318),
  [5311] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5379),
  [5313] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3974),
  [5315] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3973),
  [5317] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5919),
  [5319] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5174),
  [5321] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5761),
  [5323] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5379),
  [5325] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5463),
  [5327] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6322),
  [5329] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4832),
  [5331] = {.entry = {.count = 1, .reusable = false}}, SHIFT(164),
  [5333] = {.entry = {.count = 1, .reusable = true}}, SHIFT(168),
  [5335] = {.entry = {.count = 1, .reusable = true}}, SHIFT(169),
  [5337] = {.entry = {.count = 1, .reusable = true}}, SHIFT(109),
  [5339] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5371),
  [5341] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7638),
  [5343] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5331),
  [5345] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5331),
  [5347] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5377),
  [5349] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3867),
  [5351] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3834),
  [5353] = {.entry = {.count = 1, .reusable = false}}, SHIFT(810),
  [5355] = {.entry = {.count = 1, .reusable = false}}, SHIFT(449),
  [5357] = {.entry = {.count = 1, .reusable = true}}, SHIFT(810),
  [5359] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2879),
  [5361] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5837),
  [5363] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2877),
  [5365] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2877),
  [5367] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_concatenation_repeat1, 2), SHIFT_REPEAT(4034),
  [5370] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_concatenation_repeat1, 2), SHIFT_REPEAT(4034),
  [5373] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_variable_assignment, 3, .production_id = 24),
  [5375] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_variable_assignment, 3, .production_id = 24),
  [5377] = {.entry = {.count = 1, .reusable = false}}, SHIFT(812),
  [5379] = {.entry = {.count = 1, .reusable = true}}, SHIFT(812),
  [5381] = {.entry = {.count = 1, .reusable = false}}, SHIFT(599),
  [5383] = {.entry = {.count = 1, .reusable = true}}, SHIFT(599),
  [5385] = {.entry = {.count = 1, .reusable = false}}, SHIFT(802),
  [5387] = {.entry = {.count = 1, .reusable = true}}, SHIFT(802),
  [5389] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 1),
  [5391] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 1),
  [5393] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 1, .production_id = 1),
  [5395] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 1, .production_id = 1),
  [5397] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3943),
  [5399] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3943),
  [5401] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5418),
  [5403] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3816),
  [5405] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3804),
  [5407] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5979),
  [5409] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5267),
  [5411] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5835),
  [5413] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5418),
  [5415] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5509),
  [5417] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6345),
  [5419] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4731),
  [5421] = {.entry = {.count = 1, .reusable = false}}, SHIFT(380),
  [5423] = {.entry = {.count = 1, .reusable = true}}, SHIFT(379),
  [5425] = {.entry = {.count = 1, .reusable = true}}, SHIFT(378),
  [5427] = {.entry = {.count = 1, .reusable = true}}, SHIFT(217),
  [5429] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5417),
  [5431] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7648),
  [5433] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__extglob_blob, 1),
  [5435] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__extglob_blob, 1),
  [5437] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5755),
  [5439] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4773),
  [5441] = {.entry = {.count = 1, .reusable = true}}, SHIFT(274),
  [5443] = {.entry = {.count = 1, .reusable = true}}, SHIFT(281),
  [5445] = {.entry = {.count = 1, .reusable = true}}, SHIFT(283),
  [5447] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1338),
  [5449] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3698),
  [5451] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5427),
  [5453] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5427),
  [5455] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5476),
  [5457] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__literal_repeat1, 2), SHIFT_REPEAT(1351),
  [5460] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_command_repeat2, 1, .production_id = 3),
  [5462] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_command_repeat2, 1, .production_id = 3),
  [5464] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3942),
  [5466] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3942),
  [5468] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_unset_command_repeat1, 1),
  [5470] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_unset_command_repeat1, 1),
  [5472] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3895),
  [5474] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3895),
  [5476] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_unset_command_repeat1, 1, .production_id = 1),
  [5478] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_unset_command_repeat1, 1, .production_id = 1),
  [5480] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__literal_repeat1, 2), SHIFT_REPEAT(1404),
  [5483] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3810),
  [5485] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1404),
  [5487] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3814),
  [5489] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4002),
  [5491] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_concatenation_repeat1, 2), SHIFT_REPEAT(3895),
  [5494] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_concatenation_repeat1, 2), SHIFT_REPEAT(3895),
  [5497] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3905),
  [5499] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3926),
  [5501] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3930),
  [5503] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_concatenation_repeat1, 2), SHIFT_REPEAT(3905),
  [5506] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_concatenation_repeat1, 2), SHIFT_REPEAT(3905),
  [5509] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3905),
  [5511] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4042),
  [5513] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4046),
  [5515] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_concatenation_repeat1, 2), SHIFT_REPEAT(3942),
  [5518] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_concatenation_repeat1, 2), SHIFT_REPEAT(3942),
  [5521] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4071),
  [5523] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4075),
  [5525] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_concatenation_repeat1, 2), SHIFT_REPEAT(3943),
  [5528] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_concatenation_repeat1, 2), SHIFT_REPEAT(3943),
  [5531] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1351),
  [5533] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__literal_repeat1, 2), SHIFT_REPEAT(1338),
  [5536] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3951),
  [5538] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3951),
  [5540] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3899),
  [5542] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3899),
  [5544] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4214),
  [5546] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5844),
  [5548] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4749),
  [5550] = {.entry = {.count = 1, .reusable = true}}, SHIFT(396),
  [5552] = {.entry = {.count = 1, .reusable = true}}, SHIFT(395),
  [5554] = {.entry = {.count = 1, .reusable = true}}, SHIFT(394),
  [5556] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__literal_repeat1, 2), SHIFT_REPEAT(1724),
  [5559] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(3093),
  [5562] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(3875),
  [5565] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(3874),
  [5568] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(5952),
  [5571] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(3110),
  [5574] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(5850),
  [5577] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(3184),
  [5580] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(6315),
  [5583] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(4756),
  [5586] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(404),
  [5589] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(403),
  [5592] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(402),
  [5595] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(162),
  [5598] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(3092),
  [5601] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(7644),
  [5604] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3093),
  [5606] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3875),
  [5608] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3874),
  [5610] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5952),
  [5612] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3110),
  [5614] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5850),
  [5616] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3184),
  [5618] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6315),
  [5620] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4756),
  [5622] = {.entry = {.count = 1, .reusable = false}}, SHIFT(404),
  [5624] = {.entry = {.count = 1, .reusable = false}}, SHIFT(403),
  [5626] = {.entry = {.count = 1, .reusable = false}}, SHIFT(402),
  [5628] = {.entry = {.count = 1, .reusable = false}}, SHIFT(162),
  [5630] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3092),
  [5632] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7644),
  [5634] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4065),
  [5636] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4065),
  [5638] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3853),
  [5640] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3853),
  [5642] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1656),
  [5644] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3946),
  [5646] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3946),
  [5648] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4168),
  [5650] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4023),
  [5652] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1724),
  [5654] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4168),
  [5656] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_concatenation_repeat1, 2), SHIFT_REPEAT(3951),
  [5659] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_concatenation_repeat1, 2), SHIFT_REPEAT(3951),
  [5662] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1551),
  [5664] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3798),
  [5666] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3798),
  [5668] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4086),
  [5670] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__literal_repeat1, 2), SHIFT_REPEAT(1656),
  [5673] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_concatenation_repeat1, 2), SHIFT_REPEAT(3853),
  [5676] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_concatenation_repeat1, 2), SHIFT_REPEAT(3853),
  [5679] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4178),
  [5681] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1535),
  [5683] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_concatenation_repeat1, 2), SHIFT_REPEAT(3798),
  [5686] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_concatenation_repeat1, 2), SHIFT_REPEAT(3798),
  [5689] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_concatenation_repeat1, 2), SHIFT_REPEAT(4168),
  [5692] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_concatenation_repeat1, 2), SHIFT_REPEAT(4168),
  [5695] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_file_redirect, 3, .production_id = 22),
  [5697] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_file_redirect, 3, .production_id = 22),
  [5699] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3746),
  [5701] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4163),
  [5703] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_concatenation_repeat1, 2), SHIFT_REPEAT(3946),
  [5706] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_concatenation_repeat1, 2), SHIFT_REPEAT(3946),
  [5709] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4104),
  [5711] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4100),
  [5713] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3770),
  [5715] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3181),
  [5717] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3177),
  [5719] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3177),
  [5721] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_file_redirect, 2, .production_id = 9),
  [5723] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_file_redirect, 2, .production_id = 9),
  [5725] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__literal_repeat1, 2), SHIFT_REPEAT(1551),
  [5728] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3868),
  [5730] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4161),
  [5732] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3872),
  [5734] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3747),
  [5736] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__literal_repeat1, 2), SHIFT_REPEAT(1535),
  [5739] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_concatenation_repeat1, 2), SHIFT_REPEAT(3899),
  [5742] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_concatenation_repeat1, 2), SHIFT_REPEAT(3899),
  [5745] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4213),
  [5747] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_concatenation_repeat1, 2), SHIFT_REPEAT(4065),
  [5750] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_concatenation_repeat1, 2), SHIFT_REPEAT(4065),
  [5753] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1844),
  [5755] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__literal_repeat1, 2), SHIFT_REPEAT(1890),
  [5758] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3948),
  [5760] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3948),
  [5762] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4127),
  [5764] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4106),
  [5766] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_array, 3),
  [5768] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_array, 3),
  [5770] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_array, 2),
  [5772] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_array, 2),
  [5774] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4141),
  [5776] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_command_repeat1, 1), SHIFT(6686),
  [5779] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2347),
  [5781] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1173),
  [5783] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_concatenation_repeat1, 2), SHIFT_REPEAT(4127),
  [5786] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_concatenation_repeat1, 2), SHIFT_REPEAT(4127),
  [5789] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4127),
  [5791] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5854),
  [5793] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4761),
  [5795] = {.entry = {.count = 1, .reusable = true}}, SHIFT(413),
  [5797] = {.entry = {.count = 1, .reusable = true}}, SHIFT(412),
  [5799] = {.entry = {.count = 1, .reusable = true}}, SHIFT(411),
  [5801] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__literal_repeat1, 2), SHIFT_REPEAT(1885),
  [5804] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__literal_repeat1, 2), SHIFT_REPEAT(1844),
  [5807] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2382),
  [5809] = {.entry = {.count = 1, .reusable = true}}, SHIFT(216),
  [5811] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1603),
  [5813] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3788),
  [5815] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3788),
  [5817] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1890),
  [5819] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_command_repeat1, 1, .production_id = 3),
  [5821] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat2, 1),
  [5823] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat2, 1), REDUCE(aux_sym_command_repeat1, 1, .production_id = 3),
  [5826] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_redirected_statement_repeat2, 1),
  [5828] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_redirected_statement_repeat2, 1), REDUCE(aux_sym_command_repeat1, 1, .production_id = 3),
  [5831] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_command_repeat1, 1, .production_id = 3),
  [5833] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3906),
  [5835] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3914),
  [5837] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4158),
  [5839] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4162),
  [5841] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__literal_repeat1, 2), SHIFT_REPEAT(1835),
  [5844] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3766),
  [5846] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3766),
  [5848] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5842),
  [5850] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4744),
  [5852] = {.entry = {.count = 1, .reusable = true}}, SHIFT(392),
  [5854] = {.entry = {.count = 1, .reusable = true}}, SHIFT(391),
  [5856] = {.entry = {.count = 1, .reusable = true}}, SHIFT(390),
  [5858] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_concatenation_repeat1, 2), SHIFT_REPEAT(3788),
  [5861] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_concatenation_repeat1, 2), SHIFT_REPEAT(3788),
  [5864] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5608),
  [5866] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3863),
  [5868] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3862),
  [5870] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5992),
  [5872] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5518),
  [5874] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5846),
  [5876] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5625),
  [5878] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6362),
  [5880] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4752),
  [5882] = {.entry = {.count = 1, .reusable = false}}, SHIFT(400),
  [5884] = {.entry = {.count = 1, .reusable = false}}, SHIFT(399),
  [5886] = {.entry = {.count = 1, .reusable = false}}, SHIFT(398),
  [5888] = {.entry = {.count = 1, .reusable = false}}, SHIFT(172),
  [5890] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5562),
  [5892] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7645),
  [5894] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1885),
  [5896] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_concatenation_repeat1, 2), SHIFT_REPEAT(3948),
  [5899] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_concatenation_repeat1, 2), SHIFT_REPEAT(3948),
  [5902] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4011),
  [5904] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4010),
  [5906] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4009),
  [5908] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4007),
  [5910] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5553),
  [5912] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5584),
  [5914] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4094),
  [5916] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2374),
  [5918] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1312),
  [5920] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3837),
  [5922] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3767),
  [5924] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3773),
  [5926] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1648),
  [5928] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1835),
  [5930] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_negated_command, 2),
  [5932] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_negated_command, 2),
  [5934] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_concatenation_repeat1, 2), SHIFT_REPEAT(3837),
  [5937] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_concatenation_repeat1, 2), SHIFT_REPEAT(3837),
  [5940] = {.entry = {.count = 1, .reusable = false}}, SHIFT(445),
  [5942] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__heredoc_expression, 2, .production_id = 66),
  [5944] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__heredoc_pipeline, 2),
  [5946] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3837),
  [5948] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4038),
  [5950] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3935),
  [5952] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_command_name, 1), SHIFT(7513),
  [5955] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3931),
  [5957] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_concatenation_repeat1, 2), SHIFT_REPEAT(3766),
  [5960] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_concatenation_repeat1, 2), SHIFT_REPEAT(3766),
  [5963] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4003),
  [5965] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__literal_repeat1, 2), SHIFT_REPEAT(1951),
  [5968] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1994),
  [5970] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1951),
  [5972] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2953),
  [5974] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__literal_repeat1, 2), SHIFT_REPEAT(1994),
  [5977] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2889),
  [5979] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3841),
  [5981] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3861),
  [5983] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__literal_repeat1, 2), SHIFT_REPEAT(2094),
  [5986] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3768),
  [5988] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2237),
  [5990] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2160),
  [5992] = {.entry = {.count = 1, .reusable = true}}, SHIFT(447),
  [5994] = {.entry = {.count = 1, .reusable = false}}, SHIFT(7274),
  [5996] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7274),
  [5998] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_command_repeat1, 1), SHIFT(6664),
  [6001] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__literal_repeat1, 2), SHIFT_REPEAT(2182),
  [6004] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2215),
  [6006] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3917),
  [6008] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3913),
  [6010] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__literal_repeat1, 2), SHIFT_REPEAT(2237),
  [6013] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3829),
  [6015] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3829),
  [6017] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3774),
  [6019] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3774),
  [6021] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3146),
  [6023] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2182),
  [6025] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2659),
  [6027] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(3480),
  [6030] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(4169),
  [6033] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(4166),
  [6036] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(5971),
  [6039] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(3476),
  [6042] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(5716),
  [6045] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(3480),
  [6048] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(3516),
  [6051] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(6332),
  [6054] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(4668),
  [6057] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(255),
  [6060] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(256),
  [6063] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(257),
  [6066] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(228),
  [6069] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(3479),
  [6072] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(7620),
  [6075] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__literal_repeat1, 2), SHIFT_REPEAT(2215),
  [6078] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_concatenation_repeat1, 2), SHIFT_REPEAT(3829),
  [6081] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_concatenation_repeat1, 2), SHIFT_REPEAT(3829),
  [6084] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__literal_repeat1, 2), SHIFT_REPEAT(2160),
  [6087] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3737),
  [6089] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3735),
  [6091] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3990),
  [6093] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2094),
  [6095] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3971),
  [6097] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_concatenation_repeat1, 2), SHIFT_REPEAT(3768),
  [6100] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3886),
  [6102] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3854),
  [6104] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_concatenation_repeat1, 2), SHIFT_REPEAT(3774),
  [6107] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_concatenation_repeat1, 2), SHIFT_REPEAT(3774),
  [6110] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4004),
  [6112] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7037),
  [6114] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat2, 1), SHIFT(5495),
  [6117] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat2, 1), SHIFT(2597),
  [6120] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat2, 1), SHIFT(2317),
  [6123] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__literal_repeat1, 2), SHIFT_REPEAT(2310),
  [6126] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2310),
  [6128] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2306),
  [6130] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__literal_repeat1, 2), SHIFT_REPEAT(2306),
  [6133] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3758),
  [6135] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat2, 1), SHIFT(1550),
  [6138] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2563),
  [6140] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2563),
  [6142] = {.entry = {.count = 1, .reusable = true}}, SHIFT(452),
  [6144] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3803),
  [6146] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3819),
  [6148] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat2, 1), SHIFT(2882),
  [6151] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat2, 1), SHIFT(4448),
  [6154] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat2, 1), SHIFT(4444),
  [6157] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat2, 1), SHIFT(1592),
  [6160] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat2, 1), SHIFT(4852),
  [6163] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat2, 1), SHIFT(1761),
  [6166] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat2, 1), SHIFT(1101),
  [6169] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat2, 1), SHIFT(1591),
  [6172] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat2, 1), SHIFT(1353),
  [6175] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2461),
  [6177] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2461),
  [6179] = {.entry = {.count = 1, .reusable = true}}, SHIFT(453),
  [6181] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat2, 1), SHIFT(1324),
  [6184] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat2, 1), SHIFT(4355),
  [6187] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat2, 1), SHIFT(928),
  [6190] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat2, 1), SHIFT(4329),
  [6193] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat2, 1), SHIFT(1365),
  [6196] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat2, 1), SHIFT(4553),
  [6199] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat2, 1), SHIFT(1579),
  [6202] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat2, 1), SHIFT(900),
  [6205] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat2, 1), SHIFT(1438),
  [6208] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3494),
  [6210] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5716),
  [6212] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3498),
  [6214] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3498),
  [6216] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__literal_repeat1, 2), SHIFT_REPEAT(2305),
  [6219] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat2, 1), SHIFT(1018),
  [6222] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__literal_repeat1, 2), SHIFT_REPEAT(2338),
  [6225] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_concatenation_repeat1, 2), SHIFT_REPEAT(3758),
  [6228] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat2, 1), SHIFT(1241),
  [6231] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat2, 1), SHIFT(3505),
  [6234] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4089),
  [6236] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat2, 1), SHIFT(6226),
  [6239] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat2, 1), SHIFT(5433),
  [6242] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat2, 1), SHIFT(1255),
  [6245] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat2, 1), SHIFT(5524),
  [6248] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat2, 1), SHIFT(2919),
  [6251] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat2, 1), SHIFT(856),
  [6254] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat2, 1), SHIFT(2909),
  [6257] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2305),
  [6259] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat2, 1), SHIFT(2669),
  [6262] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat2, 1), SHIFT(5669),
  [6265] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat2, 1), SHIFT(3253),
  [6268] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat2, 1), SHIFT(6436),
  [6271] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4140),
  [6273] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat2, 1), SHIFT(2527),
  [6276] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat2, 1), SHIFT(6555),
  [6279] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat2, 1), SHIFT(2530),
  [6282] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2338),
  [6284] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat2, 1), SHIFT(3172),
  [6287] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat2, 1), SHIFT(6242),
  [6290] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat2, 1), SHIFT(4402),
  [6293] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2497),
  [6295] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2497),
  [6297] = {.entry = {.count = 1, .reusable = true}}, SHIFT(461),
  [6299] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat2, 1), SHIFT(6532),
  [6302] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat2, 1), SHIFT(6246),
  [6305] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat2, 1), SHIFT(5284),
  [6308] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat2, 1), SHIFT(2448),
  [6311] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat2, 1), SHIFT(5455),
  [6314] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat2, 1), SHIFT(2616),
  [6317] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat2, 1), SHIFT(2542),
  [6320] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat2, 1), SHIFT(6484),
  [6323] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat2, 1), SHIFT(1755),
  [6326] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat2, 1), SHIFT(2006),
  [6329] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat2, 1), SHIFT(4631),
  [6332] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat2, 1), SHIFT(3353),
  [6335] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat2, 1), SHIFT(5160),
  [6338] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat2, 1), SHIFT(4411),
  [6341] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat2, 1), SHIFT(518),
  [6344] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat2, 1), SHIFT(1818),
  [6347] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat2, 1), SHIFT(4544),
  [6350] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat2, 1), SHIFT(2216),
  [6353] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2280),
  [6355] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat2, 1), SHIFT(1044),
  [6358] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat2, 1), SHIFT(3179),
  [6361] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat2, 1), SHIFT(2065),
  [6364] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat2, 1), SHIFT(2295),
  [6367] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat2, 1), SHIFT(2628),
  [6370] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat2, 1), SHIFT(1488),
  [6373] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat2, 1), SHIFT(477),
  [6376] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat2, 1), SHIFT(2423),
  [6379] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat2, 1), SHIFT(2686),
  [6382] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat2, 1), SHIFT(1238),
  [6385] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat2, 1), SHIFT(1837),
  [6388] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat2, 1), SHIFT(1693),
  [6391] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat2, 1), SHIFT(1565),
  [6394] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat2, 1), SHIFT(2709),
  [6397] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat2, 1), SHIFT(1854),
  [6400] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat2, 1), SHIFT(2822),
  [6403] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2546),
  [6405] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2546),
  [6407] = {.entry = {.count = 1, .reusable = true}}, SHIFT(457),
  [6409] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2549),
  [6411] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2549),
  [6413] = {.entry = {.count = 1, .reusable = true}}, SHIFT(455),
  [6415] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2455),
  [6417] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2455),
  [6419] = {.entry = {.count = 1, .reusable = true}}, SHIFT(462),
  [6421] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__literal_repeat1, 2), SHIFT_REPEAT(2280),
  [6424] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat2, 1), SHIFT(3025),
  [6427] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat2, 1), SHIFT(1185),
  [6430] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat2, 1), SHIFT(3191),
  [6433] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat2, 1), SHIFT(6538),
  [6436] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat2, 1), SHIFT(6147),
  [6439] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2418),
  [6441] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3717),
  [6443] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2381),
  [6445] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3826),
  [6447] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3928),
  [6449] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3887),
  [6451] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_concatenation_repeat1, 2), SHIFT_REPEAT(3717),
  [6454] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3927),
  [6456] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__literal_repeat1, 2), SHIFT_REPEAT(2418),
  [6459] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_concatenation_repeat1, 2), SHIFT_REPEAT(4028),
  [6462] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__literal_repeat1, 2), SHIFT_REPEAT(2381),
  [6465] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3925),
  [6467] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2351),
  [6469] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3923),
  [6471] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__literal_repeat1, 2), SHIFT_REPEAT(2351),
  [6474] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6303),
  [6476] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3986),
  [6478] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3038),
  [6480] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3555),
  [6482] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4903),
  [6484] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3038),
  [6486] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3985),
  [6488] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5969),
  [6490] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6326),
  [6492] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5708),
  [6494] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6523),
  [6496] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6291),
  [6498] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4848),
  [6500] = {.entry = {.count = 1, .reusable = false}}, SHIFT(134),
  [6502] = {.entry = {.count = 1, .reusable = false}}, SHIFT(135),
  [6504] = {.entry = {.count = 1, .reusable = false}}, SHIFT(138),
  [6506] = {.entry = {.count = 1, .reusable = false}}, SHIFT(77),
  [6508] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6310),
  [6510] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6164),
  [6512] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7635),
  [6514] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3864),
  [6516] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4026),
  [6518] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2492),
  [6520] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2944),
  [6522] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4909),
  [6524] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2944),
  [6526] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3869),
  [6528] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3066),
  [6530] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5677),
  [6532] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3066),
  [6534] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_concatenation_repeat1, 2), SHIFT_REPEAT(3864),
  [6537] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3135),
  [6539] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5037),
  [6541] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3135),
  [6543] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3133),
  [6545] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5039),
  [6547] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3133),
  [6549] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3070),
  [6551] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5678),
  [6553] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3070),
  [6555] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__literal_repeat1, 2), SHIFT_REPEAT(2492),
  [6558] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3164),
  [6560] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4876),
  [6562] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3164),
  [6564] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3732),
  [6566] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3057),
  [6568] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5665),
  [6570] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3057),
  [6572] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3137),
  [6574] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5078),
  [6576] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3137),
  [6578] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3032),
  [6580] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4874),
  [6582] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3032),
  [6584] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2982),
  [6586] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5667),
  [6588] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2982),
  [6590] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3922),
  [6592] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__arithmetic_expression, 1),
  [6594] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__arithmetic_expression, 1),
  [6596] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3975),
  [6598] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__arithmetic_literal, 1),
  [6600] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__arithmetic_literal, 1),
  [6602] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3918),
  [6604] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3976),
  [6606] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3956),
  [6608] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3122),
  [6610] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4981),
  [6612] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3122),
  [6614] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3109),
  [6616] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4980),
  [6618] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3109),
  [6620] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3106),
  [6622] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4884),
  [6624] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3106),
  [6626] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3105),
  [6628] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4883),
  [6630] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3105),
  [6632] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4113),
  [6634] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_concatenation_repeat1, 2), SHIFT_REPEAT(3869),
  [6637] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3961),
  [6639] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3136),
  [6641] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5076),
  [6643] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3136),
  [6645] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3949),
  [6647] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_concatenation_repeat1, 2), SHIFT_REPEAT(3732),
  [6650] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4117),
  [6652] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2584),
  [6654] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7177),
  [6656] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3738),
  [6658] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2984),
  [6660] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3668),
  [6662] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2725),
  [6664] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3885),
  [6666] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3938),
  [6668] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3865),
  [6670] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6609),
  [6672] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3663),
  [6674] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7363),
  [6676] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_concatenation_repeat1, 2), SHIFT_REPEAT(3929),
  [6679] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4029),
  [6681] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3980),
  [6683] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3979),
  [6685] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4022),
  [6687] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__literal_repeat1, 2), SHIFT_REPEAT(2584),
  [6690] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3957),
  [6692] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_concatenation_repeat1, 2), SHIFT_REPEAT(3885),
  [6695] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2713),
  [6697] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_concatenation_repeat1, 2), SHIFT_REPEAT(3865),
  [6700] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3962),
  [6702] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3954),
  [6704] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3958),
  [6706] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4116),
  [6708] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4293),
  [6710] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3656),
  [6712] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7018),
  [6714] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_concatenation_repeat1, 2), SHIFT_REPEAT(4016),
  [6717] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3757),
  [6719] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3760),
  [6721] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3701),
  [6723] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3099),
  [6725] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3616),
  [6727] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6833),
  [6729] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_concatenation_repeat1, 2), SHIFT_REPEAT(3738),
  [6732] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6370),
  [6734] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3644),
  [6736] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4055),
  [6738] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4014),
  [6740] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7198),
  [6742] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4177),
  [6744] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4015),
  [6746] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__literal_repeat1, 2), SHIFT_REPEAT(2713),
  [6749] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__extglob_blob, 2),
  [6751] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__extglob_blob, 2),
  [6753] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2724),
  [6755] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6969),
  [6757] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2702),
  [6759] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3683),
  [6761] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4128),
  [6763] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__literal_repeat1, 2), SHIFT_REPEAT(2725),
  [6766] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4156),
  [6768] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3937),
  [6770] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2061),
  [6772] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__literal_repeat1, 2), SHIFT_REPEAT(2798),
  [6775] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2805),
  [6777] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__literal_repeat1, 2), SHIFT_REPEAT(2819),
  [6780] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3882),
  [6782] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3871),
  [6784] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3852),
  [6786] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_concatenation_repeat1, 2), SHIFT_REPEAT(3882),
  [6789] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_unary_expression, 2, .production_id = 12),
  [6791] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_unary_expression, 2, .production_id = 12),
  [6793] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2627),
  [6795] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_postfix_expression, 2, .production_id = 27),
  [6797] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_postfix_expression, 2, .production_id = 27),
  [6799] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parenthesized_expression, 3),
  [6801] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parenthesized_expression, 3),
  [6803] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2796),
  [6805] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_binary_expression, 3, .production_id = 43),
  [6807] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_binary_expression, 3, .production_id = 43),
  [6809] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2798),
  [6811] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2935),
  [6813] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2927),
  [6815] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__literal_repeat1, 2), SHIFT_REPEAT(2935),
  [6818] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_subscript, 5, .production_id = 41),
  [6820] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_subscript, 5, .production_id = 41),
  [6822] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2931),
  [6824] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__literal_repeat1, 2), SHIFT_REPEAT(2927),
  [6827] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_subscript, 5, .production_id = 40),
  [6829] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_subscript, 5, .production_id = 40),
  [6831] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2930),
  [6833] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__literal_repeat1, 2), SHIFT_REPEAT(2928),
  [6836] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_subscript, 4, .production_id = 41),
  [6838] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_subscript, 4, .production_id = 41),
  [6840] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2929),
  [6842] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_subscript, 4, .production_id = 40),
  [6844] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_subscript, 4, .production_id = 40),
  [6846] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2926),
  [6848] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4307),
  [6850] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4308),
  [6852] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4308),
  [6854] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2819),
  [6856] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2928),
  [6858] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3856),
  [6860] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3858),
  [6862] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__extglob_blob, 3),
  [6864] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__extglob_blob, 3),
  [6866] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__literal_repeat1, 2), SHIFT_REPEAT(2796),
  [6869] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2839),
  [6871] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2054),
  [6873] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1963),
  [6875] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2062),
  [6877] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2063),
  [6879] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1666),
  [6881] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4108),
  [6883] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3731),
  [6885] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2849),
  [6887] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4098),
  [6889] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3970),
  [6891] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4093),
  [6893] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4091),
  [6895] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4090),
  [6897] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4084),
  [6899] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4073),
  [6901] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4072),
  [6903] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4072),
  [6905] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4070),
  [6907] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4068),
  [6909] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4064),
  [6911] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4001),
  [6913] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3731),
  [6915] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3842),
  [6917] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1378),
  [6919] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1411),
  [6921] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4340),
  [6923] = {.entry = {.count = 1, .reusable = true}}, SHIFT(982),
  [6925] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4295),
  [6927] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1476),
  [6929] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4461),
  [6931] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1506),
  [6933] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1008),
  [6935] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1452),
  [6937] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1224),
  [6939] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1218),
  [6941] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5404),
  [6943] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2643),
  [6945] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5525),
  [6947] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2859),
  [6949] = {.entry = {.count = 1, .reusable = true}}, SHIFT(871),
  [6951] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5688),
  [6953] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3331),
  [6955] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5481),
  [6957] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4313),
  [6959] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6561),
  [6961] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5402),
  [6963] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2482),
  [6965] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5448),
  [6967] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2599),
  [6969] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2481),
  [6971] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6526),
  [6973] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1820),
  [6975] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2263),
  [6977] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4833),
  [6979] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5155),
  [6981] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4392),
  [6983] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1923),
  [6985] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4473),
  [6987] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1986),
  [6989] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3149),
  [6991] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2002),
  [6993] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2269),
  [6995] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2685),
  [6997] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2545),
  [6999] = {.entry = {.count = 1, .reusable = true}}, SHIFT(485),
  [7001] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2375),
  [7003] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3515),
  [7005] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1200),
  [7007] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1746),
  [7009] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1723),
  [7011] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1633),
  [7013] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2681),
  [7015] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2736),
  [7017] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1283),
  [7019] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1807),
  [7021] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4790),
  [7023] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1497),
  [7025] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6141),
  [7027] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4464),
  [7029] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1064),
  [7031] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1858),
  [7033] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__arithmetic_expression, 1, .production_id = 5),
  [7035] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__arithmetic_expression, 1, .production_id = 5),
  [7037] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2027),
  [7039] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2030),
  [7041] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2031),
  [7043] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1685),
  [7045] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2034),
  [7047] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2034),
  [7049] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2007),
  [7051] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__arithmetic_expression, 1, .production_id = 6),
  [7053] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__arithmetic_expression, 1, .production_id = 6),
  [7055] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1490),
  [7057] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__literal_repeat1, 2), SHIFT_REPEAT(3149),
  [7060] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7193),
  [7062] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__arithmetic_expression, 1, .production_id = 7),
  [7064] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__arithmetic_expression, 1, .production_id = 7),
  [7066] = {.entry = {.count = 1, .reusable = true}}, SHIFT(514),
  [7068] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6834),
  [7070] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6231),
  [7072] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2022),
  [7074] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2014),
  [7076] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2021),
  [7078] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1778),
  [7080] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2266),
  [7082] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6990),
  [7084] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7181),
  [7086] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1027),
  [7088] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__c_expression_not_assignment, 1),
  [7090] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4934),
  [7092] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__c_expression_not_assignment, 1),
  [7094] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__arithmetic_unary_expression, 2, .production_id = 12),
  [7096] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__arithmetic_unary_expression, 2, .production_id = 12),
  [7098] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4499),
  [7100] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__arithmetic_ternary_expression, 5, .production_id = 76),
  [7102] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__arithmetic_ternary_expression, 5, .production_id = 76),
  [7104] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__arithmetic_postfix_expression, 2, .production_id = 27),
  [7106] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__arithmetic_postfix_expression, 2, .production_id = 27),
  [7108] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7362),
  [7110] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1570),
  [7112] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__arithmetic_parenthesized_expression, 3),
  [7114] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__arithmetic_parenthesized_expression, 3),
  [7116] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3058),
  [7118] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3049),
  [7120] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__literal_repeat1, 2), SHIFT_REPEAT(3055),
  [7123] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_ternary_expression, 5, .production_id = 76),
  [7125] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_ternary_expression, 5, .production_id = 76),
  [7127] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6962),
  [7129] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2355),
  [7131] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4588),
  [7133] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4590),
  [7135] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4590),
  [7137] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__arithmetic_binary_expression, 3, .production_id = 43),
  [7139] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__arithmetic_binary_expression, 3, .production_id = 43),
  [7141] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4384),
  [7143] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5799),
  [7145] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4385),
  [7147] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4385),
  [7149] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4469),
  [7151] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4470),
  [7153] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4470),
  [7155] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__literal_repeat1, 2), SHIFT_REPEAT(2980),
  [7158] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3055),
  [7160] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_subscript, 6, .production_id = 40),
  [7162] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_subscript, 6, .production_id = 40),
  [7164] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_subscript, 6, .production_id = 41),
  [7166] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_subscript, 6, .production_id = 41),
  [7168] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1143),
  [7170] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2980),
  [7172] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1932),
  [7174] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1933),
  [7176] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1934),
  [7178] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1935),
  [7180] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1936),
  [7182] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1937),
  [7184] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1938),
  [7186] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1679),
  [7188] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1940),
  [7190] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1940),
  [7192] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1941),
  [7194] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1942),
  [7196] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1943),
  [7198] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1944),
  [7200] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1927),
  [7202] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2235),
  [7204] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2092),
  [7206] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2003),
  [7208] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1620),
  [7210] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2040),
  [7212] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2040),
  [7214] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2039),
  [7216] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2038),
  [7218] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2029),
  [7220] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2028),
  [7222] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2053),
  [7224] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3986),
  [7226] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3985),
  [7228] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6326),
  [7230] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5708),
  [7232] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6303),
  [7234] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4848),
  [7236] = {.entry = {.count = 1, .reusable = true}}, SHIFT(135),
  [7238] = {.entry = {.count = 1, .reusable = true}}, SHIFT(138),
  [7240] = {.entry = {.count = 1, .reusable = true}}, SHIFT(77),
  [7242] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2241),
  [7244] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__c_parenthesized_expression, 3),
  [7246] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__c_parenthesized_expression, 3),
  [7248] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__c_binary_expression, 3, .production_id = 43),
  [7250] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3074),
  [7252] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5244),
  [7254] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5233),
  [7256] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5243),
  [7258] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5242),
  [7260] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5241),
  [7262] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5239),
  [7264] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5238),
  [7266] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5237),
  [7268] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5236),
  [7270] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5235),
  [7272] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5234),
  [7274] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__c_binary_expression, 3, .production_id = 43),
  [7276] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4974),
  [7278] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2998),
  [7280] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1980),
  [7282] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1981),
  [7284] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1983),
  [7286] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1984),
  [7288] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1985),
  [7290] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1615),
  [7292] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1987),
  [7294] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1987),
  [7296] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1988),
  [7298] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1989),
  [7300] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1990),
  [7302] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1991),
  [7304] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1975),
  [7306] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3232),
  [7308] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3231),
  [7310] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1976),
  [7312] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1749),
  [7314] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2035),
  [7316] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2232),
  [7318] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2229),
  [7320] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2226),
  [7322] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2225),
  [7324] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2250),
  [7326] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2248),
  [7328] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2239),
  [7330] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2238),
  [7332] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2197),
  [7334] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1576),
  [7336] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2233),
  [7338] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2233),
  [7340] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2267),
  [7342] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2258),
  [7344] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2192),
  [7346] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1867),
  [7348] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4637),
  [7350] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4638),
  [7352] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4638),
  [7354] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_arithmetic_expansion_repeat1, 2),
  [7356] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4556),
  [7358] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5773),
  [7360] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4557),
  [7362] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4557),
  [7364] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2213),
  [7366] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3194),
  [7368] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3190),
  [7370] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4924),
  [7372] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5027),
  [7374] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5637),
  [7376] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2051),
  [7378] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2048),
  [7380] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2047),
  [7382] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2045),
  [7384] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2044),
  [7386] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2043),
  [7388] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2042),
  [7390] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1925),
  [7392] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2023),
  [7394] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2251),
  [7396] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2066),
  [7398] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1926),
  [7400] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2252),
  [7402] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2256),
  [7404] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2954),
  [7406] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2072),
  [7408] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2208),
  [7410] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2209),
  [7412] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2230),
  [7414] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2231),
  [7416] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2240),
  [7418] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1698),
  [7420] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2242),
  [7422] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2242),
  [7424] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2205),
  [7426] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2245),
  [7428] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2246),
  [7430] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2249),
  [7432] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2990),
  [7434] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1850),
  [7436] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2234),
  [7438] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2174),
  [7440] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4366),
  [7442] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4367),
  [7444] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4367),
  [7446] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1978),
  [7448] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__c_parenthesized_expression, 4),
  [7450] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__c_parenthesized_expression, 4),
  [7452] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__literal_repeat1, 2), SHIFT_REPEAT(3206),
  [7455] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5691),
  [7457] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3168),
  [7459] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5694),
  [7461] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3169),
  [7463] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4702),
  [7465] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__c_postfix_expression, 2, .production_id = 27),
  [7467] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__c_postfix_expression, 2, .production_id = 27),
  [7469] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__c_unary_expression, 2, .production_id = 12),
  [7471] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__c_unary_expression, 2, .production_id = 12),
  [7473] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2900),
  [7475] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3878),
  [7477] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3878),
  [7479] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3208),
  [7481] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3219),
  [7483] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2145),
  [7485] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5028),
  [7487] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2636),
  [7489] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__c_expression_not_assignment, 1, .production_id = 7),
  [7491] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__c_expression_not_assignment, 1, .production_id = 7),
  [7493] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3206),
  [7495] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_concatenation_repeat1, 2), SHIFT_REPEAT(3878),
  [7498] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_concatenation_repeat1, 2), SHIFT_REPEAT(3878),
  [7501] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__c_expression_not_assignment, 1, .production_id = 5),
  [7503] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__c_expression_not_assignment, 1, .production_id = 5),
  [7505] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4097),
  [7507] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__c_expression, 1),
  [7509] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5246),
  [7511] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__c_expression, 1),
  [7513] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4139),
  [7515] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5032),
  [7517] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5594),
  [7519] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3121),
  [7521] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_command_repeat1, 2, .production_id = 20),
  [7523] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_command_repeat1, 2, .production_id = 20),
  [7525] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_repeat1, 2, .production_id = 20), SHIFT_REPEAT(3677),
  [7528] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_command_repeat1, 2, .production_id = 20), SHIFT_REPEAT(3677),
  [7531] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_command_repeat1, 2, .production_id = 20), SHIFT_REPEAT(2138),
  [7534] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_command_repeat1, 2, .production_id = 20), SHIFT_REPEAT(3965),
  [7537] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_command_repeat1, 2, .production_id = 20), SHIFT_REPEAT(6132),
  [7540] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_command_repeat1, 2, .production_id = 20), SHIFT_REPEAT(6634),
  [7543] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4865),
  [7545] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2134),
  [7547] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5119),
  [7549] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5107),
  [7551] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4984),
  [7553] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6735),
  [7555] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2119),
  [7557] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4648),
  [7559] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1982),
  [7561] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5094),
  [7563] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5025),
  [7565] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5030),
  [7567] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4130),
  [7569] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3450),
  [7571] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4131),
  [7573] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4132),
  [7575] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4134),
  [7577] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4142),
  [7579] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4143),
  [7581] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4145),
  [7583] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4146),
  [7585] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4147),
  [7587] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4147),
  [7589] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4148),
  [7591] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4149),
  [7593] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4150),
  [7595] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4151),
  [7597] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4130),
  [7599] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3987),
  [7601] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2016),
  [7603] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2018),
  [7605] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2050),
  [7607] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2052),
  [7609] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4537),
  [7611] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4540),
  [7613] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4540),
  [7615] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3277),
  [7617] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4057),
  [7619] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4056),
  [7621] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3960),
  [7623] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4051),
  [7625] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4044),
  [7627] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4043),
  [7629] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4041),
  [7631] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4041),
  [7633] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4040),
  [7635] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4035),
  [7637] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4033),
  [7639] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4032),
  [7641] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4137),
  [7643] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3272),
  [7645] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5231),
  [7647] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1998),
  [7649] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5230),
  [7651] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5229),
  [7653] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5228),
  [7655] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5225),
  [7657] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5225),
  [7659] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5223),
  [7661] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5222),
  [7663] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5220),
  [7665] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5219),
  [7667] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5218),
  [7669] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5214),
  [7671] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4373),
  [7673] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5861),
  [7675] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4360),
  [7677] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4360),
  [7679] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4060),
  [7681] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4060),
  [7683] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4067),
  [7685] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5217),
  [7687] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2000),
  [7689] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2008),
  [7691] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2009),
  [7693] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2010),
  [7695] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2011),
  [7697] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1636),
  [7699] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2013),
  [7701] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2013),
  [7703] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2059),
  [7705] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1999),
  [7707] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1906),
  [7709] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3800),
  [7711] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4058),
  [7713] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3271),
  [7715] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2860),
  [7717] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3455),
  [7719] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5203),
  [7721] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5202),
  [7723] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5198),
  [7725] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5197),
  [7727] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5196),
  [7729] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5196),
  [7731] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5194),
  [7733] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2064),
  [7735] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5192),
  [7737] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4185),
  [7739] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5191),
  [7741] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5190),
  [7743] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5187),
  [7745] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5186),
  [7747] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3436),
  [7749] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5185),
  [7751] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3704),
  [7753] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4807),
  [7755] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4812),
  [7757] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4812),
  [7759] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4460),
  [7761] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6744),
  [7763] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6744),
  [7765] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4491),
  [7767] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4458),
  [7769] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4092),
  [7771] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6743),
  [7773] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6743),
  [7775] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4105),
  [7777] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6427),
  [7779] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3932),
  [7781] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3602),
  [7783] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3916),
  [7785] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5920),
  [7787] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6651),
  [7789] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5851),
  [7791] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6427),
  [7793] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6229),
  [7795] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6250),
  [7797] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4809),
  [7799] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__expansion_regex_replacement, 2, .production_id = 49),
  [7801] = {.entry = {.count = 1, .reusable = false}}, SHIFT(408),
  [7803] = {.entry = {.count = 1, .reusable = true}}, SHIFT(407),
  [7805] = {.entry = {.count = 1, .reusable = true}}, SHIFT(406),
  [7807] = {.entry = {.count = 1, .reusable = true}}, SHIFT(144),
  [7809] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6631),
  [7811] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6764),
  [7813] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6425),
  [7815] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7643),
  [7817] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_concatenation_repeat1, 2), SHIFT_REPEAT(3704),
  [7820] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3511),
  [7822] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6379),
  [7824] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6379),
  [7826] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__expansion_regex_replacement, 3, .production_id = 51),
  [7828] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6784),
  [7830] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6372),
  [7832] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__literal_repeat1, 2), SHIFT_REPEAT(3511),
  [7835] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6763),
  [7837] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6763),
  [7839] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2), SHIFT_REPEAT(6328),
  [7842] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_case_statement_repeat1, 2), SHIFT_REPEAT(3986),
  [7845] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2), SHIFT_REPEAT(3609),
  [7848] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_case_statement_repeat1, 2), SHIFT_REPEAT(3985),
  [7851] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2), SHIFT_REPEAT(5969),
  [7854] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_case_statement_repeat1, 2), SHIFT_REPEAT(6326),
  [7857] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_case_statement_repeat1, 2), SHIFT_REPEAT(5708),
  [7860] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_case_statement_repeat1, 2), SHIFT_REPEAT(6328),
  [7863] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2), SHIFT_REPEAT(6523),
  [7866] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2), SHIFT_REPEAT(6291),
  [7869] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_case_statement_repeat1, 2), SHIFT_REPEAT(4848),
  [7872] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2), SHIFT_REPEAT(134),
  [7875] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_case_statement_repeat1, 2), SHIFT_REPEAT(135),
  [7878] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_case_statement_repeat1, 2), SHIFT_REPEAT(138),
  [7881] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_case_statement_repeat1, 2), SHIFT_REPEAT(77),
  [7884] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_case_statement_repeat1, 2), SHIFT_REPEAT(6318),
  [7887] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_case_statement_repeat1, 2), SHIFT_REPEAT(6164),
  [7890] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_case_statement_repeat1, 2), SHIFT_REPEAT(7635),
  [7893] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(4460),
  [7896] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(4006),
  [7899] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(4005),
  [7902] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(6005),
  [7905] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(4491),
  [7908] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(5741),
  [7911] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(4837),
  [7914] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(6352),
  [7917] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(4857),
  [7920] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(110),
  [7923] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(111),
  [7926] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(113),
  [7929] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(366),
  [7932] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(4458),
  [7935] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(7632),
  [7938] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6770),
  [7940] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6770),
  [7942] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6449),
  [7944] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6449),
  [7946] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__expansion_regex_replacement, 4, .production_id = 89),
  [7948] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6748),
  [7950] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6410),
  [7952] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5137),
  [7954] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3915),
  [7956] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3593),
  [7958] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3892),
  [7960] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5875),
  [7962] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5110),
  [7964] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5861),
  [7966] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5137),
  [7968] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4377),
  [7970] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6281),
  [7972] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4800),
  [7974] = {.entry = {.count = 1, .reusable = false}}, SHIFT(252),
  [7976] = {.entry = {.count = 1, .reusable = true}}, SHIFT(259),
  [7978] = {.entry = {.count = 1, .reusable = true}}, SHIFT(260),
  [7980] = {.entry = {.count = 1, .reusable = true}}, SHIFT(131),
  [7982] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4759),
  [7984] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4759),
  [7986] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5138),
  [7988] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7641),
  [7990] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3466),
  [7992] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4169),
  [7994] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3551),
  [7996] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4166),
  [7998] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5971),
  [8000] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3476),
  [8002] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5716),
  [8004] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3466),
  [8006] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3516),
  [8008] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6332),
  [8010] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4668),
  [8012] = {.entry = {.count = 1, .reusable = false}}, SHIFT(255),
  [8014] = {.entry = {.count = 1, .reusable = true}}, SHIFT(256),
  [8016] = {.entry = {.count = 1, .reusable = true}}, SHIFT(257),
  [8018] = {.entry = {.count = 1, .reusable = true}}, SHIFT(228),
  [8020] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3601),
  [8022] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3601),
  [8024] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3465),
  [8026] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7620),
  [8028] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6755),
  [8030] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6187),
  [8032] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6197),
  [8034] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5260),
  [8036] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6197),
  [8038] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__expansion_body, 2, .production_id = 12),
  [8040] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6121),
  [8042] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7006),
  [8044] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5249),
  [8046] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6187),
  [8048] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6755),
  [8050] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6489),
  [8052] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4802),
  [8054] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4487),
  [8056] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4802),
  [8058] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4775),
  [8060] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2324),
  [8062] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4196),
  [8064] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3592),
  [8066] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4195),
  [8068] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2342),
  [8070] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5758),
  [8072] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2324),
  [8074] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4649),
  [8076] = {.entry = {.count = 1, .reusable = true}}, SHIFT(238),
  [8078] = {.entry = {.count = 1, .reusable = true}}, SHIFT(241),
  [8080] = {.entry = {.count = 1, .reusable = true}}, SHIFT(214),
  [8082] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2158),
  [8084] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2158),
  [8086] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2323),
  [8088] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4550),
  [8090] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4345),
  [8092] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4550),
  [8094] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4559),
  [8096] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1148),
  [8098] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3764),
  [8100] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3589),
  [8102] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3759),
  [8104] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5908),
  [8106] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1089),
  [8108] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5825),
  [8110] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1148),
  [8112] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1232),
  [8114] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6255),
  [8116] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4707),
  [8118] = {.entry = {.count = 1, .reusable = false}}, SHIFT(360),
  [8120] = {.entry = {.count = 1, .reusable = true}}, SHIFT(358),
  [8122] = {.entry = {.count = 1, .reusable = true}}, SHIFT(352),
  [8124] = {.entry = {.count = 1, .reusable = true}}, SHIFT(273),
  [8126] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1667),
  [8128] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1667),
  [8130] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1147),
  [8132] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7651),
  [8134] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4410),
  [8136] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3919),
  [8138] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3611),
  [8140] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3920),
  [8142] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6028),
  [8144] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4255),
  [8146] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5773),
  [8148] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4410),
  [8150] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4587),
  [8152] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6296),
  [8154] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4657),
  [8156] = {.entry = {.count = 1, .reusable = false}}, SHIFT(277),
  [8158] = {.entry = {.count = 1, .reusable = true}}, SHIFT(276),
  [8160] = {.entry = {.count = 1, .reusable = true}}, SHIFT(275),
  [8162] = {.entry = {.count = 1, .reusable = true}}, SHIFT(337),
  [8164] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4925),
  [8166] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4925),
  [8168] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4390),
  [8170] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7669),
  [8172] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2279),
  [8174] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4202),
  [8176] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3575),
  [8178] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4201),
  [8180] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6038),
  [8182] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2302),
  [8184] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5764),
  [8186] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2279),
  [8188] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1702),
  [8190] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6292),
  [8192] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4694),
  [8194] = {.entry = {.count = 1, .reusable = false}}, SHIFT(229),
  [8196] = {.entry = {.count = 1, .reusable = true}}, SHIFT(230),
  [8198] = {.entry = {.count = 1, .reusable = true}}, SHIFT(231),
  [8200] = {.entry = {.count = 1, .reusable = true}}, SHIFT(204),
  [8202] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1959),
  [8204] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1959),
  [8206] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2282),
  [8208] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7616),
  [8210] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4525),
  [8212] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4350),
  [8214] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4525),
  [8216] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4514),
  [8218] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1355),
  [8220] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1445),
  [8222] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1355),
  [8224] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1354),
  [8226] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1166),
  [8228] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3741),
  [8230] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3558),
  [8232] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3742),
  [8234] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1266),
  [8236] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5793),
  [8238] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1166),
  [8240] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4677),
  [8242] = {.entry = {.count = 1, .reusable = true}}, SHIFT(310),
  [8244] = {.entry = {.count = 1, .reusable = true}}, SHIFT(309),
  [8246] = {.entry = {.count = 1, .reusable = true}}, SHIFT(381),
  [8248] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1735),
  [8250] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1735),
  [8252] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1167),
  [8254] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1470),
  [8256] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1356),
  [8258] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1470),
  [8260] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1467),
  [8262] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1192),
  [8264] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3729),
  [8266] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3586),
  [8268] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3730),
  [8270] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6084),
  [8272] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1264),
  [8274] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5796),
  [8276] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1192),
  [8278] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1393),
  [8280] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6311),
  [8282] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4678),
  [8284] = {.entry = {.count = 1, .reusable = false}}, SHIFT(314),
  [8286] = {.entry = {.count = 1, .reusable = true}}, SHIFT(313),
  [8288] = {.entry = {.count = 1, .reusable = true}}, SHIFT(312),
  [8290] = {.entry = {.count = 1, .reusable = true}}, SHIFT(385),
  [8292] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1747),
  [8294] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1747),
  [8296] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1193),
  [8298] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7661),
  [8300] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5253),
  [8302] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3879),
  [8304] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3594),
  [8306] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3876),
  [8308] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5895),
  [8310] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5221),
  [8312] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5856),
  [8314] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5253),
  [8316] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5422),
  [8318] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6325),
  [8320] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4764),
  [8322] = {.entry = {.count = 1, .reusable = false}}, SHIFT(417),
  [8324] = {.entry = {.count = 1, .reusable = true}}, SHIFT(416),
  [8326] = {.entry = {.count = 1, .reusable = true}}, SHIFT(415),
  [8328] = {.entry = {.count = 1, .reusable = true}}, SHIFT(139),
  [8330] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5566),
  [8332] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5566),
  [8334] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5256),
  [8336] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7642),
  [8338] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1337),
  [8340] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4200),
  [8342] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3563),
  [8344] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4204),
  [8346] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1460),
  [8348] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5722),
  [8350] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1337),
  [8352] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4765),
  [8354] = {.entry = {.count = 1, .reusable = true}}, SHIFT(175),
  [8356] = {.entry = {.count = 1, .reusable = true}}, SHIFT(176),
  [8358] = {.entry = {.count = 1, .reusable = true}}, SHIFT(177),
  [8360] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1947),
  [8362] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1947),
  [8364] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1341),
  [8366] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4238),
  [8368] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3708),
  [8370] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3566),
  [8372] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3705),
  [8374] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6066),
  [8376] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4222),
  [8378] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5799),
  [8380] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4238),
  [8382] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4412),
  [8384] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6329),
  [8386] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4679),
  [8388] = {.entry = {.count = 1, .reusable = false}}, SHIFT(317),
  [8390] = {.entry = {.count = 1, .reusable = true}}, SHIFT(316),
  [8392] = {.entry = {.count = 1, .reusable = true}}, SHIFT(315),
  [8394] = {.entry = {.count = 1, .reusable = true}}, SHIFT(389),
  [8396] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4639),
  [8398] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4639),
  [8400] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4229),
  [8402] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7660),
  [8404] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1558),
  [8406] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1665),
  [8408] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1558),
  [8410] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1536),
  [8412] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4531),
  [8414] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4361),
  [8416] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4531),
  [8418] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4535),
  [8420] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1156),
  [8422] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3755),
  [8424] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3553),
  [8426] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3754),
  [8428] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1092),
  [8430] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5823),
  [8432] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1156),
  [8434] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4703),
  [8436] = {.entry = {.count = 1, .reusable = true}}, SHIFT(347),
  [8438] = {.entry = {.count = 1, .reusable = true}}, SHIFT(346),
  [8440] = {.entry = {.count = 1, .reusable = true}}, SHIFT(285),
  [8442] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1498),
  [8444] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1498),
  [8446] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1011),
  [8448] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1512),
  [8450] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1662),
  [8452] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1512),
  [8454] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1523),
  [8456] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2406),
  [8458] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3599),
  [8460] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2406),
  [8462] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2878),
  [8464] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2878),
  [8466] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2405),
  [8468] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1391),
  [8470] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4062),
  [8472] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3557),
  [8474] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4063),
  [8476] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5930),
  [8478] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1405),
  [8480] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5772),
  [8482] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1391),
  [8484] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1605),
  [8486] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6367),
  [8488] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4660),
  [8490] = {.entry = {.count = 1, .reusable = false}}, SHIFT(128),
  [8492] = {.entry = {.count = 1, .reusable = true}}, SHIFT(129),
  [8494] = {.entry = {.count = 1, .reusable = true}}, SHIFT(130),
  [8496] = {.entry = {.count = 1, .reusable = true}}, SHIFT(132),
  [8498] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2103),
  [8500] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2103),
  [8502] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1392),
  [8504] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7605),
  [8506] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2440),
  [8508] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3983),
  [8510] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3548),
  [8512] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3982),
  [8514] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5950),
  [8516] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2396),
  [8518] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5725),
  [8520] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2440),
  [8522] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2507),
  [8524] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6313),
  [8526] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4839),
  [8528] = {.entry = {.count = 1, .reusable = false}}, SHIFT(140),
  [8530] = {.entry = {.count = 1, .reusable = true}}, SHIFT(141),
  [8532] = {.entry = {.count = 1, .reusable = true}}, SHIFT(142),
  [8534] = {.entry = {.count = 1, .reusable = true}}, SHIFT(88),
  [8536] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2934),
  [8538] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2934),
  [8540] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2434),
  [8542] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7636),
  [8544] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__expansion_body, 1, .production_id = 13),
  [8546] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6176),
  [8548] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__expansion_body, 1),
  [8550] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__expansion_body, 2, .production_id = 33),
  [8552] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3552),
  [8554] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5269),
  [8556] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5797),
  [8558] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5279),
  [8560] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5279),
  [8562] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__literal_repeat1, 2), SHIFT_REPEAT(3552),
  [8565] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5045),
  [8567] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4012),
  [8569] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2173),
  [8571] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4008),
  [8573] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5903),
  [8575] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5097),
  [8577] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5762),
  [8579] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5045),
  [8581] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5149),
  [8583] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6336),
  [8585] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4850),
  [8587] = {.entry = {.count = 1, .reusable = false}}, SHIFT(89),
  [8589] = {.entry = {.count = 1, .reusable = true}}, SHIFT(97),
  [8591] = {.entry = {.count = 1, .reusable = true}}, SHIFT(98),
  [8593] = {.entry = {.count = 1, .reusable = true}}, SHIFT(355),
  [8595] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5041),
  [8597] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7631),
  [8599] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2913),
  [8601] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2884),
  [8603] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3578),
  [8605] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3597),
  [8607] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1574),
  [8609] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1542),
  [8611] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6338),
  [8613] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6338),
  [8615] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6364),
  [8617] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2106),
  [8619] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1748),
  [8621] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5158),
  [8623] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5762),
  [8625] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5154),
  [8627] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5154),
  [8629] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2071),
  [8631] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1961),
  [8633] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2015),
  [8635] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6368),
  [8637] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4840),
  [8639] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6433),
  [8641] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4804),
  [8643] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6461),
  [8645] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__heredoc_command, 2, .production_id = 67), SHIFT_REPEAT(5062),
  [8648] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__heredoc_command, 2, .production_id = 67), SHIFT_REPEAT(4006),
  [8651] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__heredoc_command, 2, .production_id = 67),
  [8653] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__heredoc_command, 2, .production_id = 67), SHIFT_REPEAT(4005),
  [8656] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__heredoc_command, 2, .production_id = 67), SHIFT_REPEAT(6005),
  [8659] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__heredoc_command, 2, .production_id = 67), SHIFT_REPEAT(5034),
  [8662] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__heredoc_command, 2, .production_id = 67), SHIFT_REPEAT(5741),
  [8665] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__heredoc_command, 2, .production_id = 67), SHIFT_REPEAT(4837),
  [8668] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__heredoc_command, 2, .production_id = 67), SHIFT_REPEAT(6352),
  [8671] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__heredoc_command, 2, .production_id = 67), SHIFT_REPEAT(4857),
  [8674] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__heredoc_command, 2, .production_id = 67), SHIFT_REPEAT(110),
  [8677] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__heredoc_command, 2, .production_id = 67), SHIFT_REPEAT(111),
  [8680] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__heredoc_command, 2, .production_id = 67), SHIFT_REPEAT(113),
  [8683] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__heredoc_command, 2, .production_id = 67), SHIFT_REPEAT(366),
  [8686] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__heredoc_command, 2, .production_id = 67), SHIFT_REPEAT(5048),
  [8689] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__heredoc_command, 2, .production_id = 67), SHIFT_REPEAT(7632),
  [8692] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5477),
  [8694] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5761),
  [8696] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5483),
  [8698] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5483),
  [8700] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6420),
  [8702] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6420),
  [8704] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6393),
  [8706] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2020),
  [8708] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2817),
  [8710] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1561),
  [8712] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5485),
  [8714] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5830),
  [8716] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5486),
  [8718] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5486),
  [8720] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6371),
  [8722] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6421),
  [8724] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1731),
  [8726] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4688),
  [8728] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5552),
  [8730] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1556),
  [8732] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1729),
  [8734] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6392),
  [8736] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2247),
  [8738] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4636),
  [8740] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5610),
  [8742] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2253),
  [8744] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6378),
  [8746] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2807),
  [8748] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6505),
  [8750] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6456),
  [8752] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(5045),
  [8755] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(4012),
  [8758] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(4008),
  [8761] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(5903),
  [8764] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(5097),
  [8767] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(5762),
  [8770] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(5045),
  [8773] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(5149),
  [8776] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(6336),
  [8779] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(4850),
  [8782] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(89),
  [8785] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(97),
  [8788] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(98),
  [8791] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(355),
  [8794] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(5041),
  [8797] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_statement_repeat1, 2), SHIFT_REPEAT(7631),
  [8800] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6437),
  [8802] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6517),
  [8804] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6401),
  [8806] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6302),
  [8808] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6302),
  [8810] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6312),
  [8812] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4870),
  [8814] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4990),
  [8816] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1759),
  [8818] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4006),
  [8820] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4005),
  [8822] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4491),
  [8824] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5741),
  [8826] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4460),
  [8828] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4857),
  [8830] = {.entry = {.count = 1, .reusable = true}}, SHIFT(111),
  [8832] = {.entry = {.count = 1, .reusable = true}}, SHIFT(113),
  [8834] = {.entry = {.count = 1, .reusable = true}}, SHIFT(366),
  [8836] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2730),
  [8838] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3072),
  [8840] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2729),
  [8842] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2730),
  [8844] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2729),
  [8846] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3069),
  [8848] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4000),
  [8850] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3997),
  [8852] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2480),
  [8854] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5733),
  [8856] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2572),
  [8858] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4854),
  [8860] = {.entry = {.count = 1, .reusable = true}}, SHIFT(119),
  [8862] = {.entry = {.count = 1, .reusable = true}}, SHIFT(121),
  [8864] = {.entry = {.count = 1, .reusable = true}}, SHIFT(182),
  [8866] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5443),
  [8868] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5856),
  [8870] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5442),
  [8872] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5442),
  [8874] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1000),
  [8876] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3809),
  [8878] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3812),
  [8880] = {.entry = {.count = 1, .reusable = true}}, SHIFT(969),
  [8882] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5788),
  [8884] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1000),
  [8886] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4675),
  [8888] = {.entry = {.count = 1, .reusable = true}}, SHIFT(304),
  [8890] = {.entry = {.count = 1, .reusable = true}}, SHIFT(303),
  [8892] = {.entry = {.count = 1, .reusable = true}}, SHIFT(372),
  [8894] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1001),
  [8896] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1346),
  [8898] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1158),
  [8900] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1158),
  [8902] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1163),
  [8904] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4047),
  [8906] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4037),
  [8908] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1675),
  [8910] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5783),
  [8912] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1630),
  [8914] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4843),
  [8916] = {.entry = {.count = 1, .reusable = true}}, SHIFT(73),
  [8918] = {.entry = {.count = 1, .reusable = true}}, SHIFT(75),
  [8920] = {.entry = {.count = 1, .reusable = true}}, SHIFT(339),
  [8922] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2166),
  [8924] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3995),
  [8926] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3994),
  [8928] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2389),
  [8930] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5720),
  [8932] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2422),
  [8934] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4851),
  [8936] = {.entry = {.count = 1, .reusable = true}}, SHIFT(126),
  [8938] = {.entry = {.count = 1, .reusable = true}}, SHIFT(127),
  [8940] = {.entry = {.count = 1, .reusable = true}}, SHIFT(375),
  [8942] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3875),
  [8944] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3874),
  [8946] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3110),
  [8948] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5850),
  [8950] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3093),
  [8952] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4756),
  [8954] = {.entry = {.count = 1, .reusable = true}}, SHIFT(403),
  [8956] = {.entry = {.count = 1, .reusable = true}}, SHIFT(402),
  [8958] = {.entry = {.count = 1, .reusable = true}}, SHIFT(162),
  [8960] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3744),
  [8962] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3743),
  [8964] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1308),
  [8966] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5814),
  [8968] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1381),
  [8970] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4690),
  [8972] = {.entry = {.count = 1, .reusable = true}}, SHIFT(331),
  [8974] = {.entry = {.count = 1, .reusable = true}}, SHIFT(330),
  [8976] = {.entry = {.count = 1, .reusable = true}}, SHIFT(409),
  [8978] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1370),
  [8980] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1370),
  [8982] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1372),
  [8984] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2369),
  [8986] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2369),
  [8988] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2371),
  [8990] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4061),
  [8992] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4054),
  [8994] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1916),
  [8996] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5721),
  [8998] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1846),
  [9000] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4819),
  [9002] = {.entry = {.count = 1, .reusable = true}}, SHIFT(410),
  [9004] = {.entry = {.count = 1, .reusable = true}}, SHIFT(247),
  [9006] = {.entry = {.count = 1, .reusable = true}}, SHIFT(272),
  [9008] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3724),
  [9010] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3722),
  [9012] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1234),
  [9014] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5809),
  [9016] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1290),
  [9018] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4682),
  [9020] = {.entry = {.count = 1, .reusable = true}}, SHIFT(325),
  [9022] = {.entry = {.count = 1, .reusable = true}}, SHIFT(324),
  [9024] = {.entry = {.count = 1, .reusable = true}}, SHIFT(401),
  [9026] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6403),
  [9028] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6407),
  [9030] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1832),
  [9032] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1839),
  [9034] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1786),
  [9036] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1918),
  [9038] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1269),
  [9040] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1269),
  [9042] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1270),
  [9044] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3817),
  [9046] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3820),
  [9048] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1725),
  [9050] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5784),
  [9052] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1526),
  [9054] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4672),
  [9056] = {.entry = {.count = 1, .reusable = true}}, SHIFT(301),
  [9058] = {.entry = {.count = 1, .reusable = true}}, SHIFT(300),
  [9060] = {.entry = {.count = 1, .reusable = true}}, SHIFT(368),
  [9062] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3849),
  [9064] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3850),
  [9066] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1468),
  [9068] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5777),
  [9070] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1394),
  [9072] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4661),
  [9074] = {.entry = {.count = 1, .reusable = true}}, SHIFT(291),
  [9076] = {.entry = {.count = 1, .reusable = true}}, SHIFT(288),
  [9078] = {.entry = {.count = 1, .reusable = true}}, SHIFT(349),
  [9080] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4344),
  [9082] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4363),
  [9084] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6583),
  [9086] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2074),
  [9088] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1968),
  [9090] = {.entry = {.count = 1, .reusable = false}}, SHIFT(854),
  [9092] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3749),
  [9094] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3748),
  [9096] = {.entry = {.count = 1, .reusable = true}}, SHIFT(834),
  [9098] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5817),
  [9100] = {.entry = {.count = 1, .reusable = true}}, SHIFT(854),
  [9102] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4693),
  [9104] = {.entry = {.count = 1, .reusable = true}}, SHIFT(334),
  [9106] = {.entry = {.count = 1, .reusable = true}}, SHIFT(333),
  [9108] = {.entry = {.count = 1, .reusable = true}}, SHIFT(414),
  [9110] = {.entry = {.count = 1, .reusable = true}}, SHIFT(861),
  [9112] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1168),
  [9114] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2271),
  [9116] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4164),
  [9118] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4144),
  [9120] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5923),
  [9122] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2298),
  [9124] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5730),
  [9126] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2271),
  [9128] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2420),
  [9130] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6316),
  [9132] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4691),
  [9134] = {.entry = {.count = 1, .reusable = false}}, SHIFT(261),
  [9136] = {.entry = {.count = 1, .reusable = true}}, SHIFT(262),
  [9138] = {.entry = {.count = 1, .reusable = true}}, SHIFT(263),
  [9140] = {.entry = {.count = 1, .reusable = true}}, SHIFT(235),
  [9142] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2309),
  [9144] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7621),
  [9146] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6585),
  [9148] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5502),
  [9150] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5835),
  [9152] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5501),
  [9154] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5501),
  [9156] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2356),
  [9158] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3029),
  [9160] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1710),
  [9162] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1711),
  [9164] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3028),
  [9166] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2199),
  [9168] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2199),
  [9170] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2204),
  [9172] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2556),
  [9174] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1399),
  [9176] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1399),
  [9178] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1407),
  [9180] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3480),
  [9182] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3480),
  [9184] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3479),
  [9186] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1688),
  [9188] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1688),
  [9190] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1575),
  [9192] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2460),
  [9194] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1244),
  [9196] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4184),
  [9198] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4183),
  [9200] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1174),
  [9202] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5729),
  [9204] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1244),
  [9206] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4641),
  [9208] = {.entry = {.count = 1, .reusable = true}}, SHIFT(251),
  [9210] = {.entry = {.count = 1, .reusable = true}}, SHIFT(253),
  [9212] = {.entry = {.count = 1, .reusable = true}}, SHIFT(223),
  [9214] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1243),
  [9216] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1611),
  [9218] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2694),
  [9220] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2696),
  [9222] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1124),
  [9224] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1091),
  [9226] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1124),
  [9228] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1122),
  [9230] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2337),
  [9232] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2337),
  [9234] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2340),
  [9236] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1039),
  [9238] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4211),
  [9240] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4212),
  [9242] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1144),
  [9244] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5746),
  [9246] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1039),
  [9248] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4841),
  [9250] = {.entry = {.count = 1, .reusable = true}}, SHIFT(206),
  [9252] = {.entry = {.count = 1, .reusable = true}}, SHIFT(207),
  [9254] = {.entry = {.count = 1, .reusable = true}}, SHIFT(189),
  [9256] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1042),
  [9258] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1673),
  [9260] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1784),
  [9262] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1823),
  [9264] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1784),
  [9266] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1787),
  [9268] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1189),
  [9270] = {.entry = {.count = 1, .reusable = false}}, SHIFT(608),
  [9272] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1188),
  [9274] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1189),
  [9276] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1188),
  [9278] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4124),
  [9280] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4096),
  [9282] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4110),
  [9284] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4102),
  [9286] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4102),
  [9288] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6101),
  [9290] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5806),
  [9292] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2866),
  [9294] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6347),
  [9296] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4629),
  [9298] = {.entry = {.count = 1, .reusable = true}}, SHIFT(186),
  [9300] = {.entry = {.count = 1, .reusable = true}}, SHIFT(187),
  [9302] = {.entry = {.count = 1, .reusable = true}}, SHIFT(188),
  [9304] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2790),
  [9306] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4566),
  [9308] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4049),
  [9310] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4048),
  [9312] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3243),
  [9314] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4429),
  [9316] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5727),
  [9318] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4566),
  [9320] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4830),
  [9322] = {.entry = {.count = 1, .reusable = true}}, SHIFT(149),
  [9324] = {.entry = {.count = 1, .reusable = true}}, SHIFT(146),
  [9326] = {.entry = {.count = 1, .reusable = true}}, SHIFT(294),
  [9328] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4429),
  [9330] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2634),
  [9332] = {.entry = {.count = 1, .reusable = false}}, SHIFT(953),
  [9334] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2633),
  [9336] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2634),
  [9338] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2633),
  [9340] = {.entry = {.count = 1, .reusable = false}}, SHIFT(482),
  [9342] = {.entry = {.count = 1, .reusable = true}}, SHIFT(483),
  [9344] = {.entry = {.count = 1, .reusable = true}}, SHIFT(482),
  [9346] = {.entry = {.count = 1, .reusable = false}}, SHIFT(483),
  [9348] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5347),
  [9350] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5174),
  [9352] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5347),
  [9354] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5348),
  [9356] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3510),
  [9358] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3508),
  [9360] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3510),
  [9362] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3508),
  [9364] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3981),
  [9366] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4069),
  [9368] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4083),
  [9370] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4085),
  [9372] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4085),
  [9374] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6099),
  [9376] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5706),
  [9378] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3267),
  [9380] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6249),
  [9382] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4738),
  [9384] = {.entry = {.count = 1, .reusable = true}}, SHIFT(190),
  [9386] = {.entry = {.count = 1, .reusable = true}}, SHIFT(192),
  [9388] = {.entry = {.count = 1, .reusable = true}}, SHIFT(193),
  [9390] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3357),
  [9392] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3035),
  [9394] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2125),
  [9396] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2125),
  [9398] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2126),
  [9400] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1037),
  [9402] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1038),
  [9404] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1037),
  [9406] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1038),
  [9408] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2742),
  [9410] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3355),
  [9412] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4810),
  [9414] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3457),
  [9416] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4811),
  [9418] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4810),
  [9420] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4811),
  [9422] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3474),
  [9424] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3349),
  [9426] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2744),
  [9428] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2154),
  [9430] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2154),
  [9432] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2155),
  [9434] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5173),
  [9436] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5064),
  [9438] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5173),
  [9440] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5124),
  [9442] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2401),
  [9444] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2402),
  [9446] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2401),
  [9448] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2402),
  [9450] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5278),
  [9452] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5278),
  [9454] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5290),
  [9456] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3240),
  [9458] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5140),
  [9460] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5140),
  [9462] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5127),
  [9464] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1109),
  [9466] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1109),
  [9468] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1108),
  [9470] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3348),
  [9472] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5440),
  [9474] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3580),
  [9476] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5438),
  [9478] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5440),
  [9480] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5438),
  [9482] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2745),
  [9484] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1107),
  [9486] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1107),
  [9488] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1106),
  [9490] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3346),
  [9492] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3582),
  [9494] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2746),
  [9496] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2741),
  [9498] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3362),
  [9500] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2890),
  [9502] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2580),
  [9504] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2579),
  [9506] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2580),
  [9508] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2579),
  [9510] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2436),
  [9512] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2436),
  [9514] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2432),
  [9516] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2210),
  [9518] = {.entry = {.count = 1, .reusable = false}}, SHIFT(774),
  [9520] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2162),
  [9522] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2210),
  [9524] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2162),
  [9526] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2413),
  [9528] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2413),
  [9530] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2415),
  [9532] = {.entry = {.count = 1, .reusable = false}}, SHIFT(710),
  [9534] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2707),
  [9536] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2711),
  [9538] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2707),
  [9540] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2711),
  [9542] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5616),
  [9544] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3863),
  [9546] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3862),
  [9548] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5518),
  [9550] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5846),
  [9552] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5616),
  [9554] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4752),
  [9556] = {.entry = {.count = 1, .reusable = true}}, SHIFT(399),
  [9558] = {.entry = {.count = 1, .reusable = true}}, SHIFT(398),
  [9560] = {.entry = {.count = 1, .reusable = true}}, SHIFT(172),
  [9562] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5548),
  [9564] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5605),
  [9566] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5605),
  [9568] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5606),
  [9570] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2740),
  [9572] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3363),
  [9574] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3344),
  [9576] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2747),
  [9578] = {.entry = {.count = 1, .reusable = false}}, SHIFT(589),
  [9580] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1623),
  [9582] = {.entry = {.count = 1, .reusable = false}}, SHIFT(740),
  [9584] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1619),
  [9586] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1623),
  [9588] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1619),
  [9590] = {.entry = {.count = 1, .reusable = false}}, SHIFT(756),
  [9592] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3343),
  [9594] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2748),
  [9596] = {.entry = {.count = 1, .reusable = false}}, SHIFT(572),
  [9598] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3342),
  [9600] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1199),
  [9602] = {.entry = {.count = 1, .reusable = false}}, SHIFT(611),
  [9604] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1198),
  [9606] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1199),
  [9608] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1198),
  [9610] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2749),
  [9612] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3341),
  [9614] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2750),
  [9616] = {.entry = {.count = 1, .reusable = false}}, SHIFT(597),
  [9618] = {.entry = {.count = 1, .reusable = false}}, SHIFT(904),
  [9620] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2344),
  [9622] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2341),
  [9624] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2344),
  [9626] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2341),
  [9628] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3340),
  [9630] = {.entry = {.count = 1, .reusable = false}}, SHIFT(902),
  [9632] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4604),
  [9634] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4025),
  [9636] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4021),
  [9638] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4401),
  [9640] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5723),
  [9642] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4604),
  [9644] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4845),
  [9646] = {.entry = {.count = 1, .reusable = true}}, SHIFT(80),
  [9648] = {.entry = {.count = 1, .reusable = true}}, SHIFT(83),
  [9650] = {.entry = {.count = 1, .reusable = true}}, SHIFT(350),
  [9652] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4602),
  [9654] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5412),
  [9656] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5412),
  [9658] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5420),
  [9660] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4477),
  [9662] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4477),
  [9664] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4467),
  [9666] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2751),
  [9668] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5415),
  [9670] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5415),
  [9672] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5474),
  [9674] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1733),
  [9676] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4193),
  [9678] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4192),
  [9680] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1734),
  [9682] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5736),
  [9684] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1733),
  [9686] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4624),
  [9688] = {.entry = {.count = 1, .reusable = true}}, SHIFT(244),
  [9690] = {.entry = {.count = 1, .reusable = true}}, SHIFT(245),
  [9692] = {.entry = {.count = 1, .reusable = true}}, SHIFT(219),
  [9694] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1734),
  [9696] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1851),
  [9698] = {.entry = {.count = 1, .reusable = false}}, SHIFT(668),
  [9700] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1861),
  [9702] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1851),
  [9704] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1861),
  [9706] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2075),
  [9708] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2077),
  [9710] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2075),
  [9712] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2077),
  [9714] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5300),
  [9716] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5300),
  [9718] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5323),
  [9720] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1590),
  [9722] = {.entry = {.count = 1, .reusable = false}}, SHIFT(692),
  [9724] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1621),
  [9726] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1590),
  [9728] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1621),
  [9730] = {.entry = {.count = 1, .reusable = false}}, SHIFT(465),
  [9732] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5303),
  [9734] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5303),
  [9736] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5301),
  [9738] = {.entry = {.count = 1, .reusable = false}}, SHIFT(662),
  [9740] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2004),
  [9742] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2005),
  [9744] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2004),
  [9746] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2005),
  [9748] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1127),
  [9750] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1127),
  [9752] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1121),
  [9754] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6225),
  [9756] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4175),
  [9758] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4176),
  [9760] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5925),
  [9762] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6205),
  [9764] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5787),
  [9766] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6225),
  [9768] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6238),
  [9770] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6253),
  [9772] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4720),
  [9774] = {.entry = {.count = 1, .reusable = false}}, SHIFT(158),
  [9776] = {.entry = {.count = 1, .reusable = true}}, SHIFT(159),
  [9778] = {.entry = {.count = 1, .reusable = true}}, SHIFT(160),
  [9780] = {.entry = {.count = 1, .reusable = true}}, SHIFT(161),
  [9782] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6205),
  [9784] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7609),
  [9786] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1162),
  [9788] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1162),
  [9790] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1160),
  [9792] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1076),
  [9794] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1076),
  [9796] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1072),
  [9798] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4292),
  [9800] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3728),
  [9802] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3726),
  [9804] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4259),
  [9806] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5811),
  [9808] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4292),
  [9810] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4686),
  [9812] = {.entry = {.count = 1, .reusable = true}}, SHIFT(328),
  [9814] = {.entry = {.count = 1, .reusable = true}}, SHIFT(327),
  [9816] = {.entry = {.count = 1, .reusable = true}}, SHIFT(405),
  [9818] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4296),
  [9820] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1207),
  [9822] = {.entry = {.count = 1, .reusable = false}}, SHIFT(643),
  [9824] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1206),
  [9826] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1207),
  [9828] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1206),
  [9830] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4270),
  [9832] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4270),
  [9834] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4272),
  [9836] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3339),
  [9838] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2752),
  [9840] = {.entry = {.count = 1, .reusable = false}}, SHIFT(641),
  [9842] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2739),
  [9844] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1069),
  [9846] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1069),
  [9848] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1066),
  [9850] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1921),
  [9852] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1924),
  [9854] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1921),
  [9856] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1924),
  [9858] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3364),
  [9860] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4357),
  [9862] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3579),
  [9864] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4398),
  [9866] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4357),
  [9868] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4398),
  [9870] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3607),
  [9872] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4337),
  [9874] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3535),
  [9876] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4338),
  [9878] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4337),
  [9880] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4338),
  [9882] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3338),
  [9884] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3540),
  [9886] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3612),
  [9888] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3600),
  [9890] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2753),
  [9892] = {.entry = {.count = 1, .reusable = false}}, SHIFT(832),
  [9894] = {.entry = {.count = 1, .reusable = true}}, SHIFT(832),
  [9896] = {.entry = {.count = 1, .reusable = true}}, SHIFT(859),
  [9898] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4066),
  [9900] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4123),
  [9902] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4121),
  [9904] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4120),
  [9906] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4120),
  [9908] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6108),
  [9910] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5852),
  [9912] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3287),
  [9914] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6278),
  [9916] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4815),
  [9918] = {.entry = {.count = 1, .reusable = true}}, SHIFT(359),
  [9920] = {.entry = {.count = 1, .reusable = true}}, SHIFT(363),
  [9922] = {.entry = {.count = 1, .reusable = true}}, SHIFT(365),
  [9924] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3188),
  [9926] = {.entry = {.count = 1, .reusable = false}}, SHIFT(880),
  [9928] = {.entry = {.count = 1, .reusable = true}}, SHIFT(880),
  [9930] = {.entry = {.count = 1, .reusable = true}}, SHIFT(885),
  [9932] = {.entry = {.count = 1, .reusable = false}}, SHIFT(793),
  [9934] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3337),
  [9936] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4242),
  [9938] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3714),
  [9940] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3713),
  [9942] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4218),
  [9944] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5807),
  [9946] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4242),
  [9948] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4681),
  [9950] = {.entry = {.count = 1, .reusable = true}}, SHIFT(322),
  [9952] = {.entry = {.count = 1, .reusable = true}}, SHIFT(321),
  [9954] = {.entry = {.count = 1, .reusable = true}}, SHIFT(397),
  [9956] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4246),
  [9958] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5541),
  [9960] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3666),
  [9962] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5542),
  [9964] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5541),
  [9966] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5542),
  [9968] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4260),
  [9970] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4260),
  [9972] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4257),
  [9974] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1342),
  [9976] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1342),
  [9978] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1339),
  [9980] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2932),
  [9982] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1423),
  [9984] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3753),
  [9986] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3751),
  [9988] = {.entry = {.count = 1, .reusable = false}}, SHIFT(638),
  [9990] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1422),
  [9992] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5819),
  [9994] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1423),
  [9996] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4696),
  [9998] = {.entry = {.count = 1, .reusable = true}}, SHIFT(343),
  [10000] = {.entry = {.count = 1, .reusable = true}}, SHIFT(338),
  [10002] = {.entry = {.count = 1, .reusable = true}}, SHIFT(357),
  [10004] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1422),
  [10006] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3664),
  [10008] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3366),
  [10010] = {.entry = {.count = 1, .reusable = false}}, SHIFT(464),
  [10012] = {.entry = {.count = 1, .reusable = false}}, SHIFT(637),
  [10014] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5199),
  [10016] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5200),
  [10018] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5199),
  [10020] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5200),
  [10022] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2754),
  [10024] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2792),
  [10026] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3336),
  [10028] = {.entry = {.count = 1, .reusable = false}}, SHIFT(796),
  [10030] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3369),
  [10032] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1263),
  [10034] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1263),
  [10036] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1265),
  [10038] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1258),
  [10040] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1258),
  [10042] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1261),
  [10044] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2793),
  [10046] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3370),
  [10048] = {.entry = {.count = 1, .reusable = false}}, SHIFT(842),
  [10050] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6192),
  [10052] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6188),
  [10054] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6192),
  [10056] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6191),
  [10058] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4430),
  [10060] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3824),
  [10062] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3825),
  [10064] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4315),
  [10066] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5782),
  [10068] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4430),
  [10070] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4671),
  [10072] = {.entry = {.count = 1, .reusable = true}}, SHIFT(298),
  [10074] = {.entry = {.count = 1, .reusable = true}}, SHIFT(293),
  [10076] = {.entry = {.count = 1, .reusable = true}}, SHIFT(361),
  [10078] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4601),
  [10080] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3435),
  [10082] = {.entry = {.count = 1, .reusable = false}}, SHIFT(993),
  [10084] = {.entry = {.count = 1, .reusable = true}}, SHIFT(993),
  [10086] = {.entry = {.count = 1, .reusable = true}}, SHIFT(992),
  [10088] = {.entry = {.count = 1, .reusable = false}}, SHIFT(995),
  [10090] = {.entry = {.count = 1, .reusable = true}}, SHIFT(995),
  [10092] = {.entry = {.count = 1, .reusable = true}}, SHIFT(994),
  [10094] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1293),
  [10096] = {.entry = {.count = 1, .reusable = false}}, SHIFT(588),
  [10098] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1292),
  [10100] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1293),
  [10102] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1292),
  [10104] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2755),
  [10106] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4673),
  [10108] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4524),
  [10110] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4673),
  [10112] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4711),
  [10114] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4760),
  [10116] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4760),
  [10118] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4748),
  [10120] = {.entry = {.count = 1, .reusable = false}}, SHIFT(910),
  [10122] = {.entry = {.count = 1, .reusable = false}}, SHIFT(552),
  [10124] = {.entry = {.count = 1, .reusable = true}}, SHIFT(912),
  [10126] = {.entry = {.count = 1, .reusable = true}}, SHIFT(910),
  [10128] = {.entry = {.count = 1, .reusable = false}}, SHIFT(912),
  [10130] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4274),
  [10132] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3888),
  [10134] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3893),
  [10136] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4233),
  [10138] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5775),
  [10140] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4274),
  [10142] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4658),
  [10144] = {.entry = {.count = 1, .reusable = true}}, SHIFT(106),
  [10146] = {.entry = {.count = 1, .reusable = true}}, SHIFT(279),
  [10148] = {.entry = {.count = 1, .reusable = true}}, SHIFT(345),
  [10150] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4276),
  [10152] = {.entry = {.count = 1, .reusable = false}}, SHIFT(750),
  [10154] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3195),
  [10156] = {.entry = {.count = 1, .reusable = false}}, SHIFT(553),
  [10158] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3270),
  [10160] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3269),
  [10162] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3044),
  [10164] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3047),
  [10166] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3187),
  [10168] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2794),
  [10170] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3371),
  [10172] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3186),
  [10174] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2842),
  [10176] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1141),
  [10178] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2841),
  [10180] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2842),
  [10182] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2841),
  [10184] = {.entry = {.count = 1, .reusable = false}}, SHIFT(723),
  [10186] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3334),
  [10188] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2392),
  [10190] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1197),
  [10192] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2393),
  [10194] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2392),
  [10196] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2393),
  [10198] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1164),
  [10200] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4375),
  [10202] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4375),
  [10204] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4352),
  [10206] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6494),
  [10208] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6486),
  [10210] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6494),
  [10212] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6486),
  [10214] = {.entry = {.count = 1, .reusable = false}}, SHIFT(749),
  [10216] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3329),
  [10218] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2757),
  [10220] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2514),
  [10222] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2518),
  [10224] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2514),
  [10226] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2518),
  [10228] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2590),
  [10230] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2587),
  [10232] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2590),
  [10234] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2587),
  [10236] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5461),
  [10238] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5480),
  [10240] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5461),
  [10242] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5480),
  [10244] = {.entry = {.count = 1, .reusable = false}}, SHIFT(578),
  [10246] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1532),
  [10248] = {.entry = {.count = 1, .reusable = false}}, SHIFT(628),
  [10250] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1533),
  [10252] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1532),
  [10254] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1533),
  [10256] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2450),
  [10258] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2449),
  [10260] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2450),
  [10262] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2449),
  [10264] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5387),
  [10266] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5380),
  [10268] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5387),
  [10270] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5380),
  [10272] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1135),
  [10274] = {.entry = {.count = 1, .reusable = false}}, SHIFT(634),
  [10276] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3328),
  [10278] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2758),
  [10280] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3326),
  [10282] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5428),
  [10284] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5413),
  [10286] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5428),
  [10288] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5413),
  [10290] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3415),
  [10292] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3418),
  [10294] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3415),
  [10296] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3418),
  [10298] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2759),
  [10300] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5680),
  [10302] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5660),
  [10304] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5680),
  [10306] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5660),
  [10308] = {.entry = {.count = 1, .reusable = false}}, SHIFT(866),
  [10310] = {.entry = {.count = 1, .reusable = true}}, SHIFT(864),
  [10312] = {.entry = {.count = 1, .reusable = true}}, SHIFT(866),
  [10314] = {.entry = {.count = 1, .reusable = false}}, SHIFT(864),
  [10316] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5698),
  [10318] = {.entry = {.count = 1, .reusable = false}}, SHIFT(725),
  [10320] = {.entry = {.count = 1, .reusable = false}}, SHIFT(840),
  [10322] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2797),
  [10324] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5699),
  [10326] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3324),
  [10328] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3372),
  [10330] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4532),
  [10332] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2903),
  [10334] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4539),
  [10336] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4532),
  [10338] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4539),
  [10340] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2278),
  [10342] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2278),
  [10344] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2284),
  [10346] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2902),
  [10348] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2314),
  [10350] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2314),
  [10352] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2308),
  [10354] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3464),
  [10356] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3463),
  [10358] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1462),
  [10360] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1461),
  [10362] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1462),
  [10364] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1461),
  [10366] = {.entry = {.count = 1, .reusable = false}}, SHIFT(669),
  [10368] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4579),
  [10370] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4579),
  [10372] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4589),
  [10374] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3298),
  [10376] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3313),
  [10378] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3485),
  [10380] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3482),
  [10382] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4404),
  [10384] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4409),
  [10386] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4404),
  [10388] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4409),
  [10390] = {.entry = {.count = 1, .reusable = false}}, SHIFT(767),
  [10392] = {.entry = {.count = 1, .reusable = false}}, SHIFT(683),
  [10394] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2760),
  [10396] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3321),
  [10398] = {.entry = {.count = 1, .reusable = false}}, SHIFT(768),
  [10400] = {.entry = {.count = 1, .reusable = false}}, SHIFT(911),
  [10402] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2844),
  [10404] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3374),
  [10406] = {.entry = {.count = 1, .reusable = false}}, SHIFT(513),
  [10408] = {.entry = {.count = 1, .reusable = true}}, SHIFT(512),
  [10410] = {.entry = {.count = 1, .reusable = true}}, SHIFT(513),
  [10412] = {.entry = {.count = 1, .reusable = false}}, SHIFT(512),
  [10414] = {.entry = {.count = 1, .reusable = false}}, SHIFT(906),
  [10416] = {.entry = {.count = 1, .reusable = false}}, SHIFT(896),
  [10418] = {.entry = {.count = 1, .reusable = false}}, SHIFT(539),
  [10420] = {.entry = {.count = 1, .reusable = false}}, SHIFT(899),
  [10422] = {.entry = {.count = 1, .reusable = false}}, SHIFT(603),
  [10424] = {.entry = {.count = 1, .reusable = false}}, SHIFT(892),
  [10426] = {.entry = {.count = 1, .reusable = false}}, SHIFT(890),
  [10428] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2732),
  [10430] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2733),
  [10432] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2732),
  [10434] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2733),
  [10436] = {.entry = {.count = 1, .reusable = false}}, SHIFT(605),
  [10438] = {.entry = {.count = 1, .reusable = false}}, SHIFT(699),
  [10440] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2761),
  [10442] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3320),
  [10444] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2762),
  [10446] = {.entry = {.count = 1, .reusable = false}}, SHIFT(696),
  [10448] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1083),
  [10450] = {.entry = {.count = 1, .reusable = false}}, SHIFT(558),
  [10452] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1084),
  [10454] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1083),
  [10456] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1084),
  [10458] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1024),
  [10460] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1025),
  [10462] = {.entry = {.count = 1, .reusable = false}}, SHIFT(538),
  [10464] = {.entry = {.count = 1, .reusable = false}}, SHIFT(999),
  [10466] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1004),
  [10468] = {.entry = {.count = 1, .reusable = true}}, SHIFT(999),
  [10470] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1004),
  [10472] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4285),
  [10474] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4303),
  [10476] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4285),
  [10478] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4303),
  [10480] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1465),
  [10482] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1466),
  [10484] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1465),
  [10486] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1466),
  [10488] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1440),
  [10490] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1442),
  [10492] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1440),
  [10494] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1442),
  [10496] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3317),
  [10498] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2763),
  [10500] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1569),
  [10502] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3786),
  [10504] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3789),
  [10506] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1568),
  [10508] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5791),
  [10510] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1569),
  [10512] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4676),
  [10514] = {.entry = {.count = 1, .reusable = true}}, SHIFT(307),
  [10516] = {.entry = {.count = 1, .reusable = true}}, SHIFT(306),
  [10518] = {.entry = {.count = 1, .reusable = true}}, SHIFT(377),
  [10520] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1568),
  [10522] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1120),
  [10524] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1119),
  [10526] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1120),
  [10528] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1119),
  [10530] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1827),
  [10532] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1834),
  [10534] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1827),
  [10536] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1834),
  [10538] = {.entry = {.count = 1, .reusable = false}}, SHIFT(960),
  [10540] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4831),
  [10542] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4805),
  [10544] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4831),
  [10546] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4805),
  [10548] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1669),
  [10550] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1678),
  [10552] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1669),
  [10554] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1678),
  [10556] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4493),
  [10558] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4494),
  [10560] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4493),
  [10562] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4494),
  [10564] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4526),
  [10566] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4419),
  [10568] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4526),
  [10570] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4419),
  [10572] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1053),
  [10574] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2705),
  [10576] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1050),
  [10578] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1051),
  [10580] = {.entry = {.count = 1, .reusable = false}}, SHIFT(946),
  [10582] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2706),
  [10584] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3262),
  [10586] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1048),
  [10588] = {.entry = {.count = 1, .reusable = false}}, SHIFT(943),
  [10590] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3467),
  [10592] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3467),
  [10594] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3472),
  [10596] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3163),
  [10598] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3456),
  [10600] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3456),
  [10602] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3483),
  [10604] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2880),
  [10606] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3426),
  [10608] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3238),
  [10610] = {.entry = {.count = 1, .reusable = false}}, SHIFT(467),
  [10612] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2898),
  [10614] = {.entry = {.count = 1, .reusable = false}}, SHIFT(732),
  [10616] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3237),
  [10618] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3316),
  [10620] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2764),
  [10622] = {.entry = {.count = 1, .reusable = false}}, SHIFT(873),
  [10624] = {.entry = {.count = 1, .reusable = false}}, SHIFT(877),
  [10626] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3315),
  [10628] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2765),
  [10630] = {.entry = {.count = 1, .reusable = false}}, SHIFT(942),
  [10632] = {.entry = {.count = 1, .reusable = false}}, SHIFT(941),
  [10634] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3408),
  [10636] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3166),
  [10638] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2766),
  [10640] = {.entry = {.count = 1, .reusable = false}}, SHIFT(549),
  [10642] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3309),
  [10644] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2767),
  [10646] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3401),
  [10648] = {.entry = {.count = 1, .reusable = false}}, SHIFT(550),
  [10650] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3584),
  [10652] = {.entry = {.count = 1, .reusable = false}}, SHIFT(737),
  [10654] = {.entry = {.count = 1, .reusable = false}}, SHIFT(466),
  [10656] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3562),
  [10658] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3132),
  [10660] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3308),
  [10662] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2768),
  [10664] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3134),
  [10666] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3304),
  [10668] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3417),
  [10670] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3414),
  [10672] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2769),
  [10674] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2938),
  [10676] = {.entry = {.count = 1, .reusable = false}}, SHIFT(606),
  [10678] = {.entry = {.count = 1, .reusable = false}}, SHIFT(784),
  [10680] = {.entry = {.count = 1, .reusable = false}}, SHIFT(787),
  [10682] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3303),
  [10684] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2770),
  [10686] = {.entry = {.count = 1, .reusable = false}}, SHIFT(795),
  [10688] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3302),
  [10690] = {.entry = {.count = 1, .reusable = false}}, SHIFT(801),
  [10692] = {.entry = {.count = 1, .reusable = false}}, SHIFT(776),
  [10694] = {.entry = {.count = 1, .reusable = false}}, SHIFT(775),
  [10696] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2771),
  [10698] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2896),
  [10700] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1070),
  [10702] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1071),
  [10704] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2895),
  [10706] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6199),
  [10708] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6199),
  [10710] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6203),
  [10712] = {.entry = {.count = 1, .reusable = false}}, SHIFT(580),
  [10714] = {.entry = {.count = 1, .reusable = false}}, SHIFT(579),
  [10716] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3301),
  [10718] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1115),
  [10720] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1625),
  [10722] = {.entry = {.count = 1, .reusable = false}}, SHIFT(687),
  [10724] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1626),
  [10726] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1625),
  [10728] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1626),
  [10730] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1395),
  [10732] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1395),
  [10734] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1396),
  [10736] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2772),
  [10738] = {.entry = {.count = 1, .reusable = false}}, SHIFT(956),
  [10740] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1114),
  [10742] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1596),
  [10744] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1596),
  [10746] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1599),
  [10748] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1529),
  [10750] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1529),
  [10752] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1559),
  [10754] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3242),
  [10756] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3246),
  [10758] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3247),
  [10760] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4687),
  [10762] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4593),
  [10764] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4687),
  [10766] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4683),
  [10768] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3299),
  [10770] = {.entry = {.count = 1, .reusable = false}}, SHIFT(766),
  [10772] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4699),
  [10774] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4699),
  [10776] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4698),
  [10778] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3248),
  [10780] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3249),
  [10782] = {.entry = {.count = 1, .reusable = false}}, SHIFT(647),
  [10784] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3258),
  [10786] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3259),
  [10788] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4440),
  [10790] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4359),
  [10792] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4440),
  [10794] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4418),
  [10796] = {.entry = {.count = 1, .reusable = false}}, SHIFT(652),
  [10798] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2773),
  [10800] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3296),
  [10802] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2774),
  [10804] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4452),
  [10806] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4452),
  [10808] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4450),
  [10810] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3261),
  [10812] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1273),
  [10814] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1273),
  [10816] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1271),
  [10818] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1276),
  [10820] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1276),
  [10822] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1274),
  [10824] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3293),
  [10826] = {.entry = {.count = 1, .reusable = false}}, SHIFT(954),
  [10828] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3263),
  [10830] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3264),
  [10832] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3265),
  [10834] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3598),
  [10836] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3266),
  [10838] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2775),
  [10840] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2858),
  [10842] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3373),
  [10844] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2937),
  [10846] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1493),
  [10848] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1494),
  [10850] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1493),
  [10852] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1494),
  [10854] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3350),
  [10856] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2936),
  [10858] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3446),
  [10860] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2923),
  [10862] = {.entry = {.count = 1, .reusable = false}}, SHIFT(666),
  [10864] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2922),
  [10866] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2921),
  [10868] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3289),
  [10870] = {.entry = {.count = 1, .reusable = false}}, SHIFT(663),
  [10872] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3572),
  [10874] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6202),
  [10876] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6202),
  [10878] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6189),
  [10880] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3284),
  [10882] = {.entry = {.count = 1, .reusable = false}}, SHIFT(564),
  [10884] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3559),
  [10886] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2778),
  [10888] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3447),
  [10890] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2917),
  [10892] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3449),
  [10894] = {.entry = {.count = 1, .reusable = false}}, SHIFT(700),
  [10896] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5503),
  [10898] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4216),
  [10900] = {.entry = {.count = 1, .reusable = false}}, SHIFT(782),
  [10902] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2916),
  [10904] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2915),
  [10906] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2115),
  [10908] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2899),
  [10910] = {.entry = {.count = 1, .reusable = false}}, SHIFT(765),
  [10912] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4228),
  [10914] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1425),
  [10916] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2894),
  [10918] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3560),
  [10920] = {.entry = {.count = 1, .reusable = false}}, SHIFT(670),
  [10922] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3273),
  [10924] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2779),
  [10926] = {.entry = {.count = 1, .reusable = false}}, SHIFT(674),
  [10928] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2113),
  [10930] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1859),
  [10932] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4205),
  [10934] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4206),
  [10936] = {.entry = {.count = 1, .reusable = false}}, SHIFT(721),
  [10938] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1863),
  [10940] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5735),
  [10942] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1859),
  [10944] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4728),
  [10946] = {.entry = {.count = 1, .reusable = true}}, SHIFT(181),
  [10948] = {.entry = {.count = 1, .reusable = true}}, SHIFT(184),
  [10950] = {.entry = {.count = 1, .reusable = true}}, SHIFT(185),
  [10952] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1863),
  [10954] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2325),
  [10956] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2325),
  [10958] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2327),
  [10960] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3017),
  [10962] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1344),
  [10964] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1344),
  [10966] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1343),
  [10968] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2283),
  [10970] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2283),
  [10972] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2288),
  [10974] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3268),
  [10976] = {.entry = {.count = 1, .reusable = false}}, SHIFT(881),
  [10978] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2781),
  [10980] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1330),
  [10982] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1330),
  [10984] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1331),
  [10986] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1134),
  [10988] = {.entry = {.count = 1, .reusable = false}}, SHIFT(883),
  [10990] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4992),
  [10992] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4774),
  [10994] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4992),
  [10996] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4919),
  [10998] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4896),
  [11000] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4896),
  [11002] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4897),
  [11004] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6198),
  [11006] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6198),
  [11008] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6200),
  [11010] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4806),
  [11012] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4495),
  [11014] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4806),
  [11016] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4822),
  [11018] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4750),
  [11020] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4750),
  [11022] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4754),
  [11024] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1010),
  [11026] = {.entry = {.count = 1, .reusable = false}}, SHIFT(568),
  [11028] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3434),
  [11030] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3433),
  [11032] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3432),
  [11034] = {.entry = {.count = 1, .reusable = false}}, SHIFT(567),
  [11036] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3431),
  [11038] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3260),
  [11040] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2845),
  [11042] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1752),
  [11044] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1752),
  [11046] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1758),
  [11048] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1426),
  [11050] = {.entry = {.count = 1, .reusable = false}}, SHIFT(805),
  [11052] = {.entry = {.count = 1, .reusable = false}}, SHIFT(717),
  [11054] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3430),
  [11056] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3429),
  [11058] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3257),
  [11060] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3428),
  [11062] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3425),
  [11064] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3421),
  [11066] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3413),
  [11068] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3412),
  [11070] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3410),
  [11072] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3409),
  [11074] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1745),
  [11076] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1745),
  [11078] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1750),
  [11080] = {.entry = {.count = 1, .reusable = false}}, SHIFT(555),
  [11082] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3325),
  [11084] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1154),
  [11086] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3537),
  [11088] = {.entry = {.count = 1, .reusable = false}}, SHIFT(676),
  [11090] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3538),
  [11092] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5514),
  [11094] = {.entry = {.count = 1, .reusable = false}}, SHIFT(716),
  [11096] = {.entry = {.count = 1, .reusable = false}}, SHIFT(679),
  [11098] = {.entry = {.count = 1, .reusable = false}}, SHIFT(735),
  [11100] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2783),
  [11102] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4622),
  [11104] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4578),
  [11106] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4622),
  [11108] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4619),
  [11110] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3252),
  [11112] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4684),
  [11114] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4684),
  [11116] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4771),
  [11118] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2784),
  [11120] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4463),
  [11122] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4362),
  [11124] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4463),
  [11126] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4466),
  [11128] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3162),
  [11130] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3227),
  [11132] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3230),
  [11134] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4513),
  [11136] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4513),
  [11138] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4459),
  [11140] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2820),
  [11142] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3274),
  [11144] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1041),
  [11146] = {.entry = {.count = 1, .reusable = false}}, SHIFT(677),
  [11148] = {.entry = {.count = 1, .reusable = false}}, SHIFT(556),
  [11150] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3027),
  [11152] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3018),
  [11154] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5306),
  [11156] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5206),
  [11158] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5306),
  [11160] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5312),
  [11162] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3244),
  [11164] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2785),
  [11166] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3379),
  [11168] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5286),
  [11170] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5286),
  [11172] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5287),
  [11174] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3015),
  [11176] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5130),
  [11178] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5109),
  [11180] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5130),
  [11182] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5131),
  [11184] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5128),
  [11186] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5128),
  [11188] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5129),
  [11190] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3016),
  [11192] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3235),
  [11194] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2786),
  [11196] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3239),
  [11198] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3224),
  [11200] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2787),
  [11202] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2887),
  [11204] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2888),
  [11206] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3619),
  [11208] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2804),
  [11210] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3212),
  [11212] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2788),
  [11214] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3636),
  [11216] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3209),
  [11218] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2800),
  [11220] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3185),
  [11222] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2795),
  [11224] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3204),
  [11226] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3165),
  [11228] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2789),
  [11230] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2791),
  [11232] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3182),
  [11234] = {.entry = {.count = 1, .reusable = false}}, SHIFT(639),
  [11236] = {.entry = {.count = 1, .reusable = false}}, SHIFT(640),
  [11238] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3192),
  [11240] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5655),
  [11242] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5654),
  [11244] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5654),
  [11246] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4243),
  [11248] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5226),
  [11250] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4891),
  [11252] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2838),
  [11254] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4243),
  [11256] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6103),
  [11258] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5705),
  [11260] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2997),
  [11262] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6356),
  [11264] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4763),
  [11266] = {.entry = {.count = 1, .reusable = false}}, SHIFT(220),
  [11268] = {.entry = {.count = 1, .reusable = false}}, SHIFT(221),
  [11270] = {.entry = {.count = 1, .reusable = false}}, SHIFT(222),
  [11272] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3912),
  [11274] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3912),
  [11276] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3941),
  [11278] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3941),
  [11280] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3635),
  [11282] = {.entry = {.count = 1, .reusable = false}}, SHIFT(681),
  [11284] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3780),
  [11286] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6097),
  [11288] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_variable_assignments_repeat1, 2),
  [11290] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_variable_assignments_repeat1, 2),
  [11292] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_variable_assignments_repeat1, 2), SHIFT_REPEAT(6662),
  [11295] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_concatenation_repeat1, 2), SHIFT_REPEAT(3912),
  [11298] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_concatenation_repeat1, 2), SHIFT_REPEAT(3912),
  [11301] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3952),
  [11303] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3952),
  [11305] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3907),
  [11307] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat2, 2),
  [11309] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat2, 2), SHIFT_REPEAT(3642),
  [11312] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat2, 2), SHIFT_REPEAT(631),
  [11315] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_redirected_statement_repeat2, 2),
  [11317] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat2, 2), SHIFT_REPEAT(3805),
  [11320] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_redirected_statement_repeat2, 2), SHIFT_REPEAT(6151),
  [11323] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4743),
  [11325] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4743),
  [11327] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_variable_assignments, 2),
  [11329] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_variable_assignments, 2),
  [11331] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6662),
  [11333] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3653),
  [11335] = {.entry = {.count = 1, .reusable = false}}, SHIFT(724),
  [11337] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6092),
  [11339] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4721),
  [11341] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4721),
  [11343] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4757),
  [11345] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4757),
  [11347] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_concatenation_repeat1, 2), SHIFT_REPEAT(3873),
  [11350] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_concatenation_repeat1, 2), SHIFT_REPEAT(3873),
  [11353] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3959),
  [11355] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3955),
  [11357] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_redirected_statement, 1, .dynamic_precedence = -1, .production_id = 3),
  [11359] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3642),
  [11361] = {.entry = {.count = 1, .reusable = false}}, SHIFT(631),
  [11363] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_redirected_statement, 1, .dynamic_precedence = -1, .production_id = 3),
  [11365] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3805),
  [11367] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6151),
  [11369] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4013),
  [11371] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_concatenation_repeat1, 2), SHIFT_REPEAT(3802),
  [11374] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_concatenation_repeat1, 2), SHIFT_REPEAT(3802),
  [11377] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3953),
  [11379] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3953),
  [11381] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_concatenation_repeat1, 2), SHIFT_REPEAT(3941),
  [11384] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_concatenation_repeat1, 2), SHIFT_REPEAT(3941),
  [11387] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4017),
  [11389] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3900),
  [11391] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3900),
  [11393] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__literal_repeat1, 2), SHIFT_REPEAT(4503),
  [11396] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4577),
  [11398] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_concatenation_repeat1, 2), SHIFT_REPEAT(3952),
  [11401] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_concatenation_repeat1, 2), SHIFT_REPEAT(3952),
  [11404] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_concatenation_repeat1, 2), SHIFT_REPEAT(3778),
  [11407] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_concatenation_repeat1, 2), SHIFT_REPEAT(3778),
  [11410] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4165),
  [11412] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_definition, 4, .production_id = 54),
  [11414] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_definition, 4, .production_id = 54),
  [11416] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6146),
  [11418] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4423),
  [11420] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat1, 2),
  [11422] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat1, 2), SHIFT_REPEAT(3635),
  [11425] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat1, 2), SHIFT_REPEAT(7487),
  [11428] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat1, 2), SHIFT_REPEAT(681),
  [11431] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_redirected_statement_repeat1, 2),
  [11433] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_redirected_statement_repeat1, 2), SHIFT_REPEAT(6100),
  [11436] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat2, 2), SHIFT_REPEAT(3654),
  [11439] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat2, 2), SHIFT_REPEAT(658),
  [11442] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat2, 2), SHIFT_REPEAT(3840),
  [11445] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_redirected_statement_repeat2, 2), SHIFT_REPEAT(6133),
  [11448] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3838),
  [11450] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_variable_assignments_repeat1, 2), SHIFT_REPEAT(6619),
  [11453] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_redirected_statement, 2, .dynamic_precedence = -1, .production_id = 14),
  [11455] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_redirected_statement, 2, .dynamic_precedence = -1, .production_id = 14),
  [11457] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6100),
  [11459] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4197),
  [11461] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_concatenation_repeat1, 2), SHIFT_REPEAT(3900),
  [11464] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_concatenation_repeat1, 2), SHIFT_REPEAT(3900),
  [11467] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4503),
  [11469] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4509),
  [11471] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4605),
  [11473] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3901),
  [11475] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3897),
  [11477] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4198),
  [11479] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3645),
  [11481] = {.entry = {.count = 1, .reusable = false}}, SHIFT(783),
  [11483] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6090),
  [11485] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3661),
  [11487] = {.entry = {.count = 1, .reusable = false}}, SHIFT(829),
  [11489] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6094),
  [11491] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3950),
  [11493] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3950),
  [11495] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__literal_repeat1, 2), SHIFT_REPEAT(4577),
  [11498] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4541),
  [11500] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3873),
  [11502] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3873),
  [11504] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_definition, 5, .production_id = 61),
  [11506] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_definition, 5, .production_id = 61),
  [11508] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4036),
  [11510] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6619),
  [11512] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3802),
  [11514] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3802),
  [11516] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4542),
  [11518] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__literal_repeat1, 2), SHIFT_REPEAT(4563),
  [11521] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3654),
  [11523] = {.entry = {.count = 1, .reusable = false}}, SHIFT(658),
  [11525] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3840),
  [11527] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6133),
  [11529] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4181),
  [11531] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4563),
  [11533] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_definition, 3, .production_id = 29),
  [11535] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_definition, 3, .production_id = 29),
  [11537] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__literal_repeat1, 2), SHIFT_REPEAT(4423),
  [11540] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3647),
  [11542] = {.entry = {.count = 1, .reusable = false}}, SHIFT(779),
  [11544] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6093),
  [11546] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_concatenation_repeat1, 2), SHIFT_REPEAT(3953),
  [11549] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_concatenation_repeat1, 2), SHIFT_REPEAT(3953),
  [11552] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4180),
  [11554] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3696),
  [11556] = {.entry = {.count = 1, .reusable = false}}, SHIFT(825),
  [11558] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4118),
  [11560] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6089),
  [11562] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6116),
  [11564] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6130),
  [11566] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4846),
  [11568] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3823),
  [11570] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3711),
  [11572] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4726),
  [11574] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_variable_assignments_repeat1, 2), SHIFT_REPEAT(6578),
  [11577] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_concatenation_repeat1, 2), SHIFT_REPEAT(3950),
  [11580] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_concatenation_repeat1, 2), SHIFT_REPEAT(3950),
  [11583] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3823),
  [11585] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4788),
  [11587] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat1, 2), SHIFT_REPEAT(3653),
  [11590] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat1, 2), SHIFT_REPEAT(7514),
  [11593] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat1, 2), SHIFT_REPEAT(724),
  [11596] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_redirected_statement_repeat1, 2), SHIFT_REPEAT(6116),
  [11599] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4191),
  [11601] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4187),
  [11603] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3670),
  [11605] = {.entry = {.count = 1, .reusable = false}}, SHIFT(733),
  [11607] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4045),
  [11609] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6152),
  [11611] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__literal_repeat1, 2), SHIFT_REPEAT(4788),
  [11614] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__literal_repeat1, 2), SHIFT_REPEAT(4846),
  [11617] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4772),
  [11619] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__literal_repeat1, 2), SHIFT_REPEAT(4772),
  [11622] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4173),
  [11624] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4171),
  [11626] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4172),
  [11628] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat2, 2), SHIFT_REPEAT(3670),
  [11631] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat2, 2), SHIFT_REPEAT(733),
  [11634] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat2, 2), SHIFT_REPEAT(4045),
  [11637] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_redirected_statement_repeat2, 2), SHIFT_REPEAT(6152),
  [11640] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_variable_assignments_repeat1, 2), SHIFT_REPEAT(6668),
  [11643] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_concatenation_repeat1, 2), SHIFT_REPEAT(3823),
  [11646] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_concatenation_repeat1, 2), SHIFT_REPEAT(3823),
  [11649] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3964),
  [11651] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3778),
  [11653] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3778),
  [11655] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat2, 2), SHIFT_REPEAT(3625),
  [11658] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat2, 2), SHIFT_REPEAT(739),
  [11661] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat2, 2), SHIFT_REPEAT(3763),
  [11664] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_redirected_statement_repeat2, 2), SHIFT_REPEAT(6145),
  [11667] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__literal_repeat1, 2), SHIFT_REPEAT(4726),
  [11670] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6578),
  [11672] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat2, 2), SHIFT_REPEAT(3649),
  [11675] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat2, 2), SHIFT_REPEAT(785),
  [11678] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat2, 2), SHIFT_REPEAT(4170),
  [11681] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_redirected_statement_repeat2, 2), SHIFT_REPEAT(6135),
  [11684] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3712),
  [11686] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6705),
  [11688] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3625),
  [11690] = {.entry = {.count = 1, .reusable = false}}, SHIFT(739),
  [11692] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3763),
  [11694] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6145),
  [11696] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3649),
  [11698] = {.entry = {.count = 1, .reusable = false}}, SHIFT(785),
  [11700] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4170),
  [11702] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6135),
  [11704] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6668),
  [11706] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_variable_assignments_repeat1, 2), SHIFT_REPEAT(6705),
  [11709] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3847),
  [11711] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3846),
  [11713] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3968),
  [11715] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3972),
  [11717] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3993),
  [11719] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3996),
  [11721] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3845),
  [11723] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3844),
  [11725] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_test_command, 3, .production_id = 30),
  [11727] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_test_command, 3, .production_id = 30),
  [11729] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_compound_statement, 3),
  [11731] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_compound_statement, 3),
  [11733] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_if_statement, 5, .production_id = 44),
  [11735] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_if_statement, 5, .production_id = 44),
  [11737] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3700),
  [11739] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_do_group, 2),
  [11741] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_do_group, 2),
  [11743] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_pipeline_repeat1, 2),
  [11745] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_pipeline_repeat1, 2), SHIFT(439),
  [11748] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_pipeline_repeat1, 2),
  [11750] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_if_statement, 4, .production_id = 44),
  [11752] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_if_statement, 4, .production_id = 44),
  [11754] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3531),
  [11756] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3531),
  [11758] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1810),
  [11760] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5697),
  [11762] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5700),
  [11764] = {.entry = {.count = 1, .reusable = true}}, SHIFT(201),
  [11766] = {.entry = {.count = 1, .reusable = true}}, SHIFT(202),
  [11768] = {.entry = {.count = 1, .reusable = true}}, SHIFT(203),
  [11770] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3532),
  [11772] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4221),
  [11774] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6158),
  [11776] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2847),
  [11778] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3719),
  [11780] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_while_statement, 3, .production_id = 28),
  [11782] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_while_statement, 3, .production_id = 28),
  [11784] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3999),
  [11786] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3998),
  [11788] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1285),
  [11790] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_subshell, 3),
  [11792] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_subshell, 3),
  [11794] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3684),
  [11796] = {.entry = {.count = 1, .reusable = false}}, SHIFT(790),
  [11798] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4125),
  [11800] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6136),
  [11802] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4913),
  [11804] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__for_body, 5, .production_id = 97),
  [11806] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5142),
  [11808] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5015),
  [11810] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2949),
  [11812] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6114),
  [11814] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5858),
  [11816] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3254),
  [11818] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6323),
  [11820] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4791),
  [11822] = {.entry = {.count = 1, .reusable = true}}, SHIFT(268),
  [11824] = {.entry = {.count = 1, .reusable = true}}, SHIFT(269),
  [11826] = {.entry = {.count = 1, .reusable = true}}, SHIFT(270),
  [11828] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_test_command, 3),
  [11830] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_test_command, 3),
  [11832] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1499),
  [11834] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6106),
  [11836] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__for_body, 5, .production_id = 99),
  [11838] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6124),
  [11840] = {.entry = {.count = 1, .reusable = true}}, SHIFT(511),
  [11842] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_if_statement, 7, .production_id = 44),
  [11844] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_if_statement, 7, .production_id = 44),
  [11846] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4435),
  [11848] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4425),
  [11850] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6126),
  [11852] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1682),
  [11854] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1510),
  [11856] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6115),
  [11858] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_pipeline_repeat1, 2), SHIFT_REPEAT(438),
  [11861] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat2, 2), SHIFT_REPEAT(3684),
  [11864] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat2, 2), SHIFT_REPEAT(790),
  [11867] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat2, 2), SHIFT_REPEAT(4125),
  [11870] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_redirected_statement_repeat2, 2), SHIFT_REPEAT(6136),
  [11873] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3492),
  [11875] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3848),
  [11877] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4762),
  [11879] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1770),
  [11881] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3851),
  [11883] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1138),
  [11885] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1657),
  [11887] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1316),
  [11889] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1348),
  [11891] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4325),
  [11893] = {.entry = {.count = 1, .reusable = true}}, SHIFT(915),
  [11895] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4300),
  [11897] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1403),
  [11899] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4497),
  [11901] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1507),
  [11903] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2348),
  [11905] = {.entry = {.count = 1, .reusable = true}}, SHIFT(924),
  [11907] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1687),
  [11909] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1450),
  [11911] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__for_body, 4, .production_id = 73),
  [11913] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1297),
  [11915] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1302),
  [11917] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat1, 2), SHIFT_REPEAT(3647),
  [11920] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat1, 2), SHIFT_REPEAT(7349),
  [11923] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat1, 2), SHIFT_REPEAT(779),
  [11926] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_redirected_statement_repeat1, 2), SHIFT_REPEAT(6107),
  [11929] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_do_group, 3),
  [11931] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_do_group, 3),
  [11933] = {.entry = {.count = 1, .reusable = true}}, SHIFT(486),
  [11935] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1112),
  [11937] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6239),
  [11939] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5449),
  [11941] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6153),
  [11943] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6217),
  [11945] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__for_body, 3, .production_id = 57),
  [11947] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat1, 2), SHIFT_REPEAT(3661),
  [11950] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat1, 2), SHIFT_REPEAT(829),
  [11953] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_redirected_statement_repeat1, 2), SHIFT_REPEAT(6115),
  [11956] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_test_command, 2),
  [11958] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_test_command, 2),
  [11960] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1874),
  [11962] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2674),
  [11964] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2477),
  [11966] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5523),
  [11968] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2630),
  [11970] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat1, 2), SHIFT_REPEAT(3645),
  [11973] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat1, 2), SHIFT_REPEAT(783),
  [11976] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_redirected_statement_repeat1, 2), SHIFT_REPEAT(6106),
  [11979] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2904),
  [11981] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3211),
  [11983] = {.entry = {.count = 1, .reusable = true}}, SHIFT(844),
  [11985] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__for_body, 6, .production_id = 131),
  [11987] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__for_body, 4, .production_id = 69),
  [11989] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__for_body, 4, .production_id = 71),
  [11991] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2737),
  [11993] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6107),
  [11995] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2657),
  [11997] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2645),
  [11999] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5662),
  [12001] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_if_statement, 6, .production_id = 44),
  [12003] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_if_statement, 6, .production_id = 44),
  [12005] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3221),
  [12007] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__for_body, 2),
  [12009] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2872),
  [12011] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3161),
  [12013] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5488),
  [12015] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1537),
  [12017] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6571),
  [12019] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2505),
  [12021] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3909),
  [12023] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3908),
  [12025] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2318),
  [12027] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3904),
  [12029] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3903),
  [12031] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_variable_assignments_repeat1, 2), SHIFT_REPEAT(6604),
  [12034] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_compound_statement, 2),
  [12036] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_compound_statement, 2),
  [12038] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6604),
  [12040] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3174),
  [12042] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1955),
  [12044] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2277),
  [12046] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4318),
  [12048] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__for_body, 3, .production_id = 44),
  [12050] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6480),
  [12052] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6216),
  [12054] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1057),
  [12056] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3439),
  [12058] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2661),
  [12060] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2100),
  [12062] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5255),
  [12064] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_pipeline, 2),
  [12066] = {.entry = {.count = 1, .reusable = false}}, SHIFT(438),
  [12068] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_pipeline, 2),
  [12070] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2523),
  [12072] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4515),
  [12074] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5478),
  [12076] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2611),
  [12078] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2517),
  [12080] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1186),
  [12082] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1862),
  [12084] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4348),
  [12086] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6502),
  [12088] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5156),
  [12090] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1768),
  [12092] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2073),
  [12094] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__literal_repeat1, 2), SHIFT_REPEAT(4913),
  [12097] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4766),
  [12099] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3312),
  [12101] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_redirected_statement, 2, .dynamic_precedence = -1, .production_id = 15),
  [12103] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_redirected_statement, 2, .dynamic_precedence = -1, .production_id = 15),
  [12105] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_heredoc_redirect, 5, .production_id = 93),
  [12107] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_heredoc_redirect, 5, .production_id = 93),
  [12109] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_definition, 5, .production_id = 65),
  [12111] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_definition, 5, .production_id = 65),
  [12113] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_case_statement, 8, .production_id = 45),
  [12115] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_case_statement, 8, .production_id = 45),
  [12117] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_heredoc_redirect, 5, .production_id = 94),
  [12119] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_heredoc_redirect, 5, .production_id = 94),
  [12121] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_case_statement, 8, .production_id = 46),
  [12123] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_case_statement, 8, .production_id = 46),
  [12125] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_case_statement, 5, .production_id = 46),
  [12127] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_case_statement, 5, .production_id = 46),
  [12129] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_case_statement, 5, .production_id = 45),
  [12131] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_case_statement, 5, .production_id = 45),
  [12133] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_definition, 4, .production_id = 47),
  [12135] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_definition, 4, .production_id = 47),
  [12137] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__heredoc_body, 2),
  [12139] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__heredoc_body, 2),
  [12141] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__simple_heredoc_body, 2),
  [12143] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__simple_heredoc_body, 2),
  [12145] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_heredoc_redirect, 5, .production_id = 4),
  [12147] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_heredoc_redirect, 5, .production_id = 4),
  [12149] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_case_statement, 4, .production_id = 45),
  [12151] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_case_statement, 4, .production_id = 45),
  [12153] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_case_statement, 4, .production_id = 46),
  [12155] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_case_statement, 4, .production_id = 46),
  [12157] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_heredoc_redirect, 7, .production_id = 149),
  [12159] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_heredoc_redirect, 7, .production_id = 149),
  [12161] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_heredoc_redirect, 5, .production_id = 95),
  [12163] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_heredoc_redirect, 5, .production_id = 95),
  [12165] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5168),
  [12167] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4967),
  [12169] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3031),
  [12171] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6111),
  [12173] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5750),
  [12175] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3275),
  [12177] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6360),
  [12179] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4863),
  [12181] = {.entry = {.count = 1, .reusable = true}}, SHIFT(101),
  [12183] = {.entry = {.count = 1, .reusable = true}}, SHIFT(107),
  [12185] = {.entry = {.count = 1, .reusable = true}}, SHIFT(108),
  [12187] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_heredoc_redirect, 4),
  [12189] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_heredoc_redirect, 4),
  [12191] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_case_statement, 7, .production_id = 46),
  [12193] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_case_statement, 7, .production_id = 46),
  [12195] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_statement, 4, .production_id = 42),
  [12197] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_statement, 4, .production_id = 42),
  [12199] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_heredoc_redirect, 6, .production_id = 124),
  [12201] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_heredoc_redirect, 6, .production_id = 124),
  [12203] = {.entry = {.count = 1, .reusable = false}}, SHIFT(435),
  [12205] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_command, 2, .production_id = 2),
  [12207] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_command, 2, .production_id = 2),
  [12209] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_heredoc_redirect, 5),
  [12211] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_heredoc_redirect, 5),
  [12213] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_c_style_for_statement, 5, .production_id = 56),
  [12215] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_c_style_for_statement, 5, .production_id = 56),
  [12217] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_heredoc_redirect, 6, .production_id = 4),
  [12219] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_heredoc_redirect, 6, .production_id = 4),
  [12221] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6113),
  [12223] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_heredoc_redirect, 6, .production_id = 125),
  [12225] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_heredoc_redirect, 6, .production_id = 125),
  [12227] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_c_style_for_statement, 6, .production_id = 68),
  [12229] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_c_style_for_statement, 6, .production_id = 68),
  [12231] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5226),
  [12233] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4891),
  [12235] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2838),
  [12237] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5705),
  [12239] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2997),
  [12241] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4763),
  [12243] = {.entry = {.count = 1, .reusable = true}}, SHIFT(220),
  [12245] = {.entry = {.count = 1, .reusable = true}}, SHIFT(221),
  [12247] = {.entry = {.count = 1, .reusable = true}}, SHIFT(222),
  [12249] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_heredoc_redirect, 6, .production_id = 126),
  [12251] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_heredoc_redirect, 6, .production_id = 126),
  [12253] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_statement, 6, .production_id = 75),
  [12255] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_statement, 6, .production_id = 75),
  [12257] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_heredoc_redirect, 6, .production_id = 127),
  [12259] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_heredoc_redirect, 6, .production_id = 127),
  [12261] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_definition, 6, .production_id = 84),
  [12263] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_definition, 6, .production_id = 84),
  [12265] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_command, 3, .production_id = 19),
  [12267] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_command, 3, .production_id = 19),
  [12269] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_case_statement, 7, .production_id = 45),
  [12271] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_case_statement, 7, .production_id = 45),
  [12273] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_pipeline_repeat1, 2), SHIFT_REPEAT(435),
  [12276] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5044),
  [12278] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6129),
  [12280] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat1, 2), SHIFT_REPEAT(3696),
  [12283] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat1, 2), SHIFT_REPEAT(825),
  [12286] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_redirected_statement_repeat1, 2), SHIFT_REPEAT(6113),
  [12289] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__literal_repeat1, 2), SHIFT_REPEAT(5044),
  [12292] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_case_statement, 6, .production_id = 46),
  [12294] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_case_statement, 6, .production_id = 46),
  [12296] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_case_statement, 6, .production_id = 45),
  [12298] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_case_statement, 6, .production_id = 45),
  [12300] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3911),
  [12302] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3910),
  [12304] = {.entry = {.count = 1, .reusable = false}}, SHIFT(442),
  [12306] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_concatenation_repeat1, 2), SHIFT_REPEAT(3815),
  [12309] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3815),
  [12311] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__heredoc_command, 1, .production_id = 16),
  [12313] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__heredoc_command, 1, .production_id = 16),
  [12315] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3989),
  [12317] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4182),
  [12319] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3992),
  [12321] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4081),
  [12323] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__heredoc_command, 1, .production_id = 17),
  [12325] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__heredoc_command, 1, .production_id = 17),
  [12327] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3870),
  [12329] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4099),
  [12331] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_pipeline_repeat1, 2), SHIFT_REPEAT(437),
  [12334] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_pipeline_repeat1, 2), SHIFT_REPEAT(442),
  [12337] = {.entry = {.count = 1, .reusable = false}}, SHIFT(448),
  [12339] = {.entry = {.count = 1, .reusable = false}}, SHIFT(437),
  [12341] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_pipeline_repeat1, 2), SHIFT_REPEAT(448),
  [12344] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3618),
  [12346] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1111),
  [12348] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6091),
  [12350] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_case_item, 6, .production_id = 153),
  [12352] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_case_item, 6, .production_id = 153),
  [12354] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_last_case_item, 6, .production_id = 109),
  [12356] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_case_item, 5, .production_id = 134),
  [12358] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_case_item, 5, .production_id = 134),
  [12360] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_last_case_item, 5, .production_id = 79),
  [12362] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_case_item, 6, .production_id = 155),
  [12364] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_case_item, 6, .production_id = 155),
  [12366] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_last_case_item, 6, .production_id = 112),
  [12368] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_case_item, 4, .production_id = 105),
  [12370] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_case_item, 4, .production_id = 105),
  [12372] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_last_case_item, 4, .production_id = 79),
  [12374] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3276),
  [12376] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_case_item, 4, .production_id = 115),
  [12378] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_case_item, 4, .production_id = 115),
  [12380] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_last_case_item, 4, .production_id = 83),
  [12382] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_case_item, 4, .production_id = 113),
  [12384] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_case_item, 4, .production_id = 113),
  [12386] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_last_case_item, 4, .production_id = 60),
  [12388] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_variable_assignments_repeat1, 2), SHIFT_REPEAT(6686),
  [12391] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6686),
  [12393] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_case_item, 4, .production_id = 110),
  [12395] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_case_item, 4, .production_id = 110),
  [12397] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_last_case_item, 4, .production_id = 46),
  [12399] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_case_item, 4, .production_id = 107),
  [12401] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_case_item, 4, .production_id = 107),
  [12403] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_last_case_item, 4, .production_id = 45),
  [12405] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_case_item, 4, .production_id = 103),
  [12407] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_case_item, 4, .production_id = 103),
  [12409] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_last_case_item, 4, .production_id = 59),
  [12411] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3669),
  [12413] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3669),
  [12415] = {.entry = {.count = 1, .reusable = true}}, SHIFT(944),
  [12417] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3716),
  [12419] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6118),
  [12421] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3454),
  [12423] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_case_item, 3, .production_id = 81),
  [12425] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_case_item, 3, .production_id = 81),
  [12427] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_last_case_item, 3, .production_id = 60),
  [12429] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_case_item, 3, .production_id = 77),
  [12431] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_case_item, 3, .production_id = 77),
  [12433] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_last_case_item, 3, .production_id = 59),
  [12435] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3866),
  [12437] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4157),
  [12439] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3633),
  [12441] = {.entry = {.count = 1, .reusable = false}}, SHIFT(908),
  [12443] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4188),
  [12445] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6120),
  [12447] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_case_item, 5, .production_id = 136),
  [12449] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_case_item, 5, .production_id = 136),
  [12451] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_last_case_item, 5, .production_id = 45),
  [12453] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4159),
  [12455] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat2, 2), SHIFT_REPEAT(3633),
  [12458] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat2, 2), SHIFT_REPEAT(908),
  [12461] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat2, 2), SHIFT_REPEAT(4188),
  [12464] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_redirected_statement_repeat2, 2), SHIFT_REPEAT(6120),
  [12467] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_case_item, 5, .production_id = 144),
  [12469] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_case_item, 5, .production_id = 144),
  [12471] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_last_case_item, 5, .production_id = 83),
  [12473] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3762),
  [12475] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3405),
  [12477] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3404),
  [12479] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3403),
  [12481] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_concatenation_repeat1, 2), SHIFT_REPEAT(3870),
  [12484] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3402),
  [12486] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3399),
  [12488] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3395),
  [12490] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_case_item, 5, .production_id = 138),
  [12492] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_case_item, 5, .production_id = 138),
  [12494] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_last_case_item, 5, .production_id = 109),
  [12496] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3394),
  [12498] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3689),
  [12500] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3689),
  [12502] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1117),
  [12504] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3703),
  [12506] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6088),
  [12508] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3392),
  [12510] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3390),
  [12512] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3387),
  [12514] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3384),
  [12516] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3381),
  [12518] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat2, 2), SHIFT_REPEAT(3669),
  [12521] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_redirected_statement_repeat2, 2), SHIFT_REPEAT(3669),
  [12524] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_redirected_statement_repeat2, 2), SHIFT_REPEAT(944),
  [12527] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_redirected_statement_repeat2, 2), SHIFT_REPEAT(3716),
  [12530] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_redirected_statement_repeat2, 2), SHIFT_REPEAT(6118),
  [12533] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3794),
  [12535] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3792),
  [12537] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3791),
  [12539] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3790),
  [12541] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3932),
  [12543] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2243),
  [12545] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7023),
  [12547] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3916),
  [12549] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6102),
  [12551] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4813),
  [12553] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__expansion_max_length, 3, .production_id = 64),
  [12555] = {.entry = {.count = 1, .reusable = false}}, SHIFT(211),
  [12557] = {.entry = {.count = 1, .reusable = false}}, SHIFT(212),
  [12559] = {.entry = {.count = 1, .reusable = false}}, SHIFT(213),
  [12561] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6262),
  [12563] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6977),
  [12565] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__expansion_max_length, 3, .production_id = 51),
  [12567] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6344),
  [12569] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3217),
  [12571] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_pipeline_repeat1, 2), SHIFT_REPEAT(443),
  [12574] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3218),
  [12576] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3216),
  [12578] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3214),
  [12580] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3207),
  [12582] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3877),
  [12584] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3205),
  [12586] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3202),
  [12588] = {.entry = {.count = 1, .reusable = false}}, SHIFT(443),
  [12590] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3201),
  [12592] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3075),
  [12594] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3200),
  [12596] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3199),
  [12598] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3198),
  [12600] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3196),
  [12602] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6798),
  [12604] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__expansion_max_length, 2, .production_id = 49),
  [12606] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6260),
  [12608] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2948),
  [12610] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3005),
  [12612] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2999),
  [12614] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2993),
  [12616] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2992),
  [12618] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2991),
  [12620] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3019),
  [12622] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3890),
  [12624] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2965),
  [12626] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2956),
  [12628] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2952),
  [12630] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3036),
  [12632] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_case_item, 5, .production_id = 140),
  [12634] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_case_item, 5, .production_id = 140),
  [12636] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_last_case_item, 5, .production_id = 46),
  [12638] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2947),
  [12640] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_case_item, 5, .production_id = 142),
  [12642] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_case_item, 5, .production_id = 142),
  [12644] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_last_case_item, 5, .production_id = 112),
  [12646] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6498),
  [12648] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3602),
  [12650] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6112),
  [12652] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__expansion_expression, 1, .production_id = 12),
  [12654] = {.entry = {.count = 1, .reusable = true}}, SHIFT(233),
  [12656] = {.entry = {.count = 1, .reusable = true}}, SHIFT(239),
  [12658] = {.entry = {.count = 1, .reusable = true}}, SHIFT(246),
  [12660] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_case_item, 5, .production_id = 141),
  [12662] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_case_item, 5, .production_id = 141),
  [12664] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_case_item, 5, .production_id = 143),
  [12666] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_case_item, 5, .production_id = 143),
  [12668] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5410),
  [12670] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_case_item, 5, .production_id = 139),
  [12672] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_case_item, 5, .production_id = 139),
  [12674] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6780),
  [12676] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5232),
  [12678] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__expansion_max_length, 1, .production_id = 12),
  [12680] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6208),
  [12682] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_case_item, 5, .production_id = 135),
  [12684] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_case_item, 5, .production_id = 135),
  [12686] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6104),
  [12688] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3884),
  [12690] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_case_item, 6, .production_id = 154),
  [12692] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_case_item, 6, .production_id = 154),
  [12694] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3723),
  [12696] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3727),
  [12698] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__literal_repeat1, 2), SHIFT_REPEAT(5416),
  [12701] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_case_item, 6, .production_id = 156),
  [12703] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_case_item, 6, .production_id = 156),
  [12705] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_concatenation_repeat1, 2), SHIFT_REPEAT(3890),
  [12708] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_case_item, 4, .production_id = 106),
  [12710] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_case_item, 4, .production_id = 106),
  [12712] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4199),
  [12714] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_case_item, 4, .production_id = 116),
  [12716] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_case_item, 4, .production_id = 116),
  [12718] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_case_item, 4, .production_id = 114),
  [12720] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_case_item, 4, .production_id = 114),
  [12722] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_case_item, 4, .production_id = 111),
  [12724] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_case_item, 4, .production_id = 111),
  [12726] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4059),
  [12728] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_case_item, 4, .production_id = 108),
  [12730] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_case_item, 4, .production_id = 108),
  [12732] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4077),
  [12734] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_case_item, 5, .production_id = 145),
  [12736] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_case_item, 5, .production_id = 145),
  [12738] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_concatenation_repeat1, 2), SHIFT_REPEAT(3866),
  [12741] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_case_item, 3, .production_id = 82),
  [12743] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_case_item, 3, .production_id = 82),
  [12745] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_case_item, 3, .production_id = 78),
  [12747] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_case_item, 3, .production_id = 78),
  [12749] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6664),
  [12751] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3624),
  [12753] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3624),
  [12755] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1090),
  [12757] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3769),
  [12759] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6117),
  [12761] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat2, 2), SHIFT_REPEAT(3624),
  [12764] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_redirected_statement_repeat2, 2), SHIFT_REPEAT(3624),
  [12767] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_redirected_statement_repeat2, 2), SHIFT_REPEAT(1090),
  [12770] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_redirected_statement_repeat2, 2), SHIFT_REPEAT(3769),
  [12773] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_redirected_statement_repeat2, 2), SHIFT_REPEAT(6117),
  [12776] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5416),
  [12778] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6109),
  [12780] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6134),
  [12782] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat1, 2), SHIFT_REPEAT(3618),
  [12785] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat1, 2), SHIFT_REPEAT(1111),
  [12788] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_redirected_statement_repeat1, 2), SHIFT_REPEAT(6109),
  [12791] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_concatenation_repeat1, 2), SHIFT_REPEAT(3877),
  [12794] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat1, 2), SHIFT_REPEAT(3689),
  [12797] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat1, 2), SHIFT_REPEAT(7274),
  [12800] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_redirected_statement_repeat1, 2), SHIFT_REPEAT(3689),
  [12803] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_redirected_statement_repeat1, 2), SHIFT_REPEAT(1117),
  [12806] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_redirected_statement_repeat1, 2), SHIFT_REPEAT(7274),
  [12809] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_redirected_statement_repeat1, 2), SHIFT_REPEAT(6104),
  [12812] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_variable_assignments_repeat1, 2), SHIFT_REPEAT(6664),
  [12815] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__literal_repeat1, 2), SHIFT_REPEAT(5410),
  [12818] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4203),
  [12820] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_case_item, 5, .production_id = 137),
  [12822] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_case_item, 5, .production_id = 137),
  [12824] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3795),
  [12826] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3796),
  [12828] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_case_item, 4, .production_id = 104),
  [12830] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_case_item, 4, .production_id = 104),
  [12832] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3629),
  [12834] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3629),
  [12836] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1178),
  [12838] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6150),
  [12840] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3806),
  [12842] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5512),
  [12844] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_concatenation_repeat1, 2), SHIFT_REPEAT(3884),
  [12847] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5529),
  [12849] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6487),
  [12851] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__literal_repeat1, 2), SHIFT_REPEAT(5529),
  [12854] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7022),
  [12856] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__expansion_max_length, 3, .production_id = 49),
  [12858] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6258),
  [12860] = {.entry = {.count = 1, .reusable = false}}, SHIFT(7397),
  [12862] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7397),
  [12864] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6105),
  [12866] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5506),
  [12868] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6825),
  [12870] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__expansion_max_length, 4, .production_id = 51),
  [12872] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6279),
  [12874] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7102),
  [12876] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__expansion_max_length, 4, .production_id = 64),
  [12878] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6275),
  [12880] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat1, 2), SHIFT_REPEAT(3629),
  [12883] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat1, 2), SHIFT_REPEAT(7397),
  [12886] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_redirected_statement_repeat1, 2), SHIFT_REPEAT(3629),
  [12889] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_redirected_statement_repeat1, 2), SHIFT_REPEAT(1178),
  [12892] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_redirected_statement_repeat1, 2), SHIFT_REPEAT(7397),
  [12895] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_redirected_statement_repeat1, 2), SHIFT_REPEAT(6105),
  [12898] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__literal_repeat1, 2), SHIFT_REPEAT(5506),
  [12901] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3811),
  [12903] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__literal_repeat1, 2), SHIFT_REPEAT(5512),
  [12906] = {.entry = {.count = 1, .reusable = false}}, SHIFT(446),
  [12908] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5607),
  [12910] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6214),
  [12912] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5787),
  [12914] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6222),
  [12916] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6222),
  [12918] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6453),
  [12920] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6467),
  [12922] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6451),
  [12924] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6399),
  [12926] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6384),
  [12928] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3880),
  [12930] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3880),
  [12932] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__literal_repeat1, 2), SHIFT_REPEAT(5607),
  [12935] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_pipeline_repeat1, 2), SHIFT_REPEAT(446),
  [12938] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6455),
  [12940] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6380),
  [12942] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6382),
  [12944] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6404),
  [12946] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6445),
  [12948] = {.entry = {.count = 1, .reusable = true}}, SHIFT(441),
  [12950] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3647),
  [12952] = {.entry = {.count = 1, .reusable = true}}, SHIFT(779),
  [12954] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7349),
  [12956] = {.entry = {.count = 1, .reusable = true}}, SHIFT(444),
  [12958] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3696),
  [12960] = {.entry = {.count = 1, .reusable = true}}, SHIFT(825),
  [12962] = {.entry = {.count = 1, .reusable = true}}, SHIFT(433),
  [12964] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3645),
  [12966] = {.entry = {.count = 1, .reusable = true}}, SHIFT(783),
  [12968] = {.entry = {.count = 1, .reusable = true}}, SHIFT(449),
  [12970] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3661),
  [12972] = {.entry = {.count = 1, .reusable = true}}, SHIFT(829),
  [12974] = {.entry = {.count = 1, .reusable = true}}, SHIFT(445),
  [12976] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3618),
  [12978] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1111),
  [12980] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_pipeline_repeat1, 2), SHIFT(439),
  [12983] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4095),
  [12985] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4088),
  [12987] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_concatenation_repeat1, 2), SHIFT_REPEAT(3880),
  [12990] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_concatenation_repeat1, 2), SHIFT_REPEAT(3880),
  [12993] = {.entry = {.count = 1, .reusable = true}}, SHIFT(436),
  [12995] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3653),
  [12997] = {.entry = {.count = 1, .reusable = true}}, SHIFT(724),
  [12999] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7514),
  [13001] = {.entry = {.count = 1, .reusable = false}}, SHIFT(439),
  [13003] = {.entry = {.count = 1, .reusable = true}}, SHIFT(439),
  [13005] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat2, 2), SHIFT_REPEAT(3634),
  [13008] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat2, 2), SHIFT_REPEAT(1654),
  [13011] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_redirected_statement_repeat2, 2), SHIFT_REPEAT(3739),
  [13014] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_redirected_statement_repeat2, 2), SHIFT_REPEAT(6155),
  [13017] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_pipeline_repeat1, 2), SHIFT_REPEAT(439),
  [13020] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_pipeline_repeat1, 2), SHIFT_REPEAT(439),
  [13023] = {.entry = {.count = 1, .reusable = true}}, SHIFT(450),
  [13025] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3635),
  [13027] = {.entry = {.count = 1, .reusable = true}}, SHIFT(681),
  [13029] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7487),
  [13031] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5701),
  [13033] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__literal_repeat1, 2), SHIFT_REPEAT(5701),
  [13036] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3534),
  [13038] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3534),
  [13040] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3491),
  [13042] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3628),
  [13044] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6503),
  [13046] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6504),
  [13048] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6504),
  [13050] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6478),
  [13052] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6478),
  [13054] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6448),
  [13056] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7225),
  [13058] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4207),
  [13060] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4209),
  [13062] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6002),
  [13064] = {.entry = {.count = 1, .reusable = false}}, SHIFT(521),
  [13066] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6127),
  [13068] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4652),
  [13070] = {.entry = {.count = 1, .reusable = false}}, SHIFT(196),
  [13072] = {.entry = {.count = 1, .reusable = false}}, SHIFT(197),
  [13074] = {.entry = {.count = 1, .reusable = false}}, SHIFT(198),
  [13076] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5960),
  [13078] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1765),
  [13080] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6049),
  [13082] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1483),
  [13084] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6057),
  [13086] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3020),
  [13088] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5962),
  [13090] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3279),
  [13092] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6042),
  [13094] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2680),
  [13096] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5867),
  [13098] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6520),
  [13100] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6065),
  [13102] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4536),
  [13104] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5978),
  [13106] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4379),
  [13108] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6052),
  [13110] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3045),
  [13112] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5872),
  [13114] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1777),
  [13116] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5871),
  [13118] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6510),
  [13120] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6040),
  [13122] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1877),
  [13124] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6022),
  [13126] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1049),
  [13128] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5869),
  [13130] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3488),
  [13132] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5948),
  [13134] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3427),
  [13136] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6025),
  [13138] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2728),
  [13140] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5877),
  [13142] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3528),
  [13144] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5878),
  [13146] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1785),
  [13148] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5886),
  [13150] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1945),
  [13152] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6064),
  [13154] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1540),
  [13156] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6026),
  [13158] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4391),
  [13160] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6079),
  [13162] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2698),
  [13164] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5882),
  [13166] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2508),
  [13168] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5885),
  [13170] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2109),
  [13172] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6075),
  [13174] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4475),
  [13176] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5896),
  [13178] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1304),
  [13180] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5902),
  [13182] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1284),
  [13184] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5891),
  [13186] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2400),
  [13188] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5888),
  [13190] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2512),
  [13192] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5910),
  [13194] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1796),
  [13196] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5866),
  [13198] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2114),
  [13200] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5961),
  [13202] = {.entry = {.count = 1, .reusable = false}}, SHIFT(509),
  [13204] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5938),
  [13206] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1876),
  [13208] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5914),
  [13210] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1792),
  [13212] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6013),
  [13214] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1231),
  [13216] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5922),
  [13218] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4842),
  [13220] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5893),
  [13222] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2390),
  [13224] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5932),
  [13226] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4847),
  [13228] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5928),
  [13230] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1589),
  [13232] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5898),
  [13234] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2605),
  [13236] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5911),
  [13238] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3291),
  [13240] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5993),
  [13242] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1259),
  [13244] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5887),
  [13246] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6243),
  [13248] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5904),
  [13250] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2607),
  [13252] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5944),
  [13254] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3280),
  [13256] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5901),
  [13258] = {.entry = {.count = 1, .reusable = false}}, SHIFT(487),
  [13260] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5970),
  [13262] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3203),
  [13264] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5917),
  [13266] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2274),
  [13268] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5906),
  [13270] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2454),
  [13272] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5913),
  [13274] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2485),
  [13276] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5954),
  [13278] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5152),
  [13280] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6059),
  [13282] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1116),
  [13284] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5934),
  [13286] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1691),
  [13288] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5916),
  [13290] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2452),
  [13292] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5909),
  [13294] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2322),
  [13296] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5912),
  [13298] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5468),
  [13300] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5963),
  [13302] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5150),
  [13304] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5977),
  [13306] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1614),
  [13308] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5984),
  [13310] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1680),
  [13312] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5918),
  [13314] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2826),
  [13316] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5988),
  [13318] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2032),
  [13320] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5942),
  [13322] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5482),
  [13324] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_string_repeat1, 2), SHIFT_REPEAT(4207),
  [13327] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_string_repeat1, 2), SHIFT_REPEAT(4209),
  [13330] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_string_repeat1, 2), SHIFT_REPEAT(6098),
  [13333] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_string_repeat1, 2),
  [13335] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_string_repeat1, 2), SHIFT_REPEAT(6127),
  [13338] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_string_repeat1, 2), SHIFT_REPEAT(4652),
  [13341] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_string_repeat1, 2), SHIFT_REPEAT(196),
  [13344] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_string_repeat1, 2), SHIFT_REPEAT(197),
  [13347] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_string_repeat1, 2), SHIFT_REPEAT(198),
  [13350] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5927),
  [13352] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2666),
  [13354] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5931),
  [13356] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2652),
  [13358] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5933),
  [13360] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4436),
  [13362] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6020),
  [13364] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1684),
  [13366] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5939),
  [13368] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4434),
  [13370] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5941),
  [13372] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4426),
  [13374] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5943),
  [13376] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4424),
  [13378] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5945),
  [13380] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1518),
  [13382] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5951),
  [13384] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1676),
  [13386] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5953),
  [13388] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4770),
  [13390] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5965),
  [13392] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2571),
  [13394] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5987),
  [13396] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2343),
  [13398] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6003),
  [13400] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1022),
  [13402] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5957),
  [13404] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4753),
  [13406] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6056),
  [13408] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1891),
  [13410] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5966),
  [13412] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1772),
  [13414] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5985),
  [13416] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2577),
  [13418] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5967),
  [13420] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1137),
  [13422] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5868),
  [13424] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6209),
  [13426] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5972),
  [13428] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1139),
  [13430] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5975),
  [13432] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1718),
  [13434] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5980),
  [13436] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1655),
  [13438] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5983),
  [13440] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1315),
  [13442] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5986),
  [13444] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1319),
  [13446] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5989),
  [13448] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1345),
  [13450] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6031),
  [13452] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2339),
  [13454] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5990),
  [13456] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1352),
  [13458] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5999),
  [13460] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5298),
  [13462] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5996),
  [13464] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4326),
  [13466] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5998),
  [13468] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4323),
  [13470] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5899),
  [13472] = {.entry = {.count = 1, .reusable = false}}, SHIFT(481),
  [13474] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6004),
  [13476] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2593),
  [13478] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6001),
  [13480] = {.entry = {.count = 1, .reusable = false}}, SHIFT(916),
  [13482] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6017),
  [13484] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5294),
  [13486] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6006),
  [13488] = {.entry = {.count = 1, .reusable = false}}, SHIFT(914),
  [13490] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6008),
  [13492] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4301),
  [13494] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5890),
  [13496] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2853),
  [13498] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6010),
  [13500] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4298),
  [13502] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6011),
  [13504] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1402),
  [13506] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6015),
  [13508] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1413),
  [13510] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6016),
  [13512] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4501),
  [13514] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6019),
  [13516] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4496),
  [13518] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6021),
  [13520] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1513),
  [13522] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6077),
  [13524] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1555),
  [13526] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6023),
  [13528] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1501),
  [13530] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6043),
  [13532] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6317),
  [13534] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6029),
  [13536] = {.entry = {.count = 1, .reusable = false}}, SHIFT(919),
  [13538] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6032),
  [13540] = {.entry = {.count = 1, .reusable = false}}, SHIFT(930),
  [13542] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6037),
  [13544] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1453),
  [13546] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6041),
  [13548] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1439),
  [13550] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6046),
  [13552] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1288),
  [13554] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6074),
  [13556] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6306),
  [13558] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6069),
  [13560] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1973),
  [13562] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6050),
  [13564] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1299),
  [13566] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6055),
  [13568] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1165),
  [13570] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6058),
  [13572] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1305),
  [13574] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6062),
  [13576] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6767),
  [13578] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6068),
  [13580] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6765),
  [13582] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6073),
  [13584] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5430),
  [13586] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6033),
  [13588] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1965),
  [13590] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6076),
  [13592] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5452),
  [13594] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6080),
  [13596] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2676),
  [13598] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6083),
  [13600] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2672),
  [13602] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6067),
  [13604] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6663),
  [13606] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6086),
  [13608] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5510),
  [13610] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6087),
  [13612] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5507),
  [13614] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6081),
  [13616] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2906),
  [13618] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6072),
  [13620] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2901),
  [13622] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6063),
  [13624] = {.entry = {.count = 1, .reusable = false}}, SHIFT(845),
  [13626] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5995),
  [13628] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6645),
  [13630] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6054),
  [13632] = {.entry = {.count = 1, .reusable = false}}, SHIFT(843),
  [13634] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6045),
  [13636] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2891),
  [13638] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6036),
  [13640] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2885),
  [13642] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6027),
  [13644] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2658),
  [13646] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6018),
  [13648] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2656),
  [13650] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6009),
  [13652] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5663),
  [13654] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6000),
  [13656] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5618),
  [13658] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5991),
  [13660] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3225),
  [13662] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5874),
  [13664] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1857),
  [13666] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5959),
  [13668] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1060),
  [13670] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5982),
  [13672] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3220),
  [13674] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5947),
  [13676] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6514),
  [13678] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5880),
  [13680] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3294),
  [13682] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5973),
  [13684] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2873),
  [13686] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5964),
  [13688] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2871),
  [13690] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5955),
  [13692] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5490),
  [13694] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5946),
  [13696] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5472),
  [13698] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5873),
  [13700] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6515),
  [13702] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5897),
  [13704] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3176),
  [13706] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5889),
  [13708] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4333),
  [13710] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5905),
  [13712] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3167),
  [13714] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5881),
  [13716] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4316),
  [13718] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5937),
  [13720] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6564),
  [13722] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5929),
  [13724] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6573),
  [13726] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5921),
  [13728] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2519),
  [13730] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6143),
  [13732] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6125),
  [13734] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6140),
  [13736] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6140),
  [13738] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6509),
  [13740] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1774),
  [13742] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6518),
  [13744] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1848),
  [13746] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3522),
  [13748] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2108),
  [13750] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6244),
  [13752] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2513),
  [13754] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4378),
  [13756] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2380),
  [13758] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1294),
  [13760] = {.entry = {.count = 1, .reusable = false}}, SHIFT(490),
  [13762] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2608),
  [13764] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3297),
  [13766] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2275),
  [13768] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1798),
  [13770] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3300),
  [13772] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2465),
  [13774] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2818),
  [13776] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6507),
  [13778] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5827),
  [13780] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6506),
  [13782] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6506),
  [13784] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2524),
  [13786] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4825),
  [13788] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1563),
  [13790] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2650),
  [13792] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4441),
  [13794] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6557),
  [13796] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4432),
  [13798] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5464),
  [13800] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1577),
  [13802] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3453),
  [13804] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4739),
  [13806] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5153),
  [13808] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5497),
  [13810] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1763),
  [13812] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1129),
  [13814] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3288),
  [13816] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2876),
  [13818] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1677),
  [13820] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1489),
  [13822] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4372),
  [13824] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1385),
  [13826] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2581),
  [13828] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2080),
  [13830] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1327),
  [13832] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3229),
  [13834] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6635),
  [13836] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4332),
  [13838] = {.entry = {.count = 1, .reusable = false}}, SHIFT(923),
  [13840] = {.entry = {.count = 1, .reusable = false}}, SHIFT(504),
  [13842] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1152),
  [13844] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2603),
  [13846] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4306),
  [13848] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5666),
  [13850] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1464),
  [13852] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1223),
  [13854] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4516),
  [13856] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5288),
  [13858] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1546),
  [13860] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1031),
  [13862] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2662),
  [13864] = {.entry = {.count = 1, .reusable = false}}, SHIFT(917),
  [13866] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2335),
  [13868] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1962),
  [13870] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1458),
  [13872] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1870),
  [13874] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2673),
  [13876] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2893),
  [13878] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1278),
  [13880] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1502),
  [13882] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3076),
  [13884] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1298),
  [13886] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6753),
  [13888] = {.entry = {.count = 1, .reusable = false}}, SHIFT(848),
  [13890] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4564),
  [13892] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5439),
  [13894] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6293),
  [13896] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1511),
  [13898] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2682),
  [13900] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2911),
  [13902] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5513),
  [13904] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3680),
  [13906] = {.entry = {.count = 1, .reusable = false}}, SHIFT(7292),
  [13908] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3680),
  [13910] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1118),
  [13912] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7292),
  [13914] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3718),
  [13916] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3681),
  [13918] = {.entry = {.count = 1, .reusable = false}}, SHIFT(7205),
  [13920] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3681),
  [13922] = {.entry = {.count = 1, .reusable = true}}, SHIFT(826),
  [13924] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7205),
  [13926] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4119),
  [13928] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3641),
  [13930] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3641),
  [13932] = {.entry = {.count = 1, .reusable = true}}, SHIFT(804),
  [13934] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3836),
  [13936] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3631),
  [13938] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3631),
  [13940] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1110),
  [13942] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4186),
  [13944] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3687),
  [13946] = {.entry = {.count = 1, .reusable = false}}, SHIFT(7517),
  [13948] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3687),
  [13950] = {.entry = {.count = 1, .reusable = true}}, SHIFT(722),
  [13952] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7517),
  [13954] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3828),
  [13956] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3650),
  [13958] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3650),
  [13960] = {.entry = {.count = 1, .reusable = true}}, SHIFT(798),
  [13962] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4039),
  [13964] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3659),
  [13966] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3659),
  [13968] = {.entry = {.count = 1, .reusable = true}}, SHIFT(828),
  [13970] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4167),
  [13972] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3638),
  [13974] = {.entry = {.count = 1, .reusable = false}}, SHIFT(7491),
  [13976] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3638),
  [13978] = {.entry = {.count = 1, .reusable = true}}, SHIFT(706),
  [13980] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7491),
  [13982] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3782),
  [13984] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3223),
  [13986] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3241),
  [13988] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3241),
  [13990] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2855),
  [13992] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2856),
  [13994] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2856),
  [13996] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6095),
  [13998] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6096),
  [14000] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6096),
  [14002] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3159),
  [14004] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3160),
  [14006] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3160),
  [14008] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3626),
  [14010] = {.entry = {.count = 1, .reusable = false}}, SHIFT(7401),
  [14012] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3626),
  [14014] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1172),
  [14016] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7401),
  [14018] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3445),
  [14020] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3448),
  [14022] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3448),
  [14024] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6235),
  [14026] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6211),
  [14028] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6211),
  [14030] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3307),
  [14032] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3306),
  [14034] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3306),
  [14036] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3210),
  [14038] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3213),
  [14040] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3213),
  [14042] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3623),
  [14044] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3623),
  [14046] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1102),
  [14048] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3772),
  [14050] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3667),
  [14052] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3667),
  [14054] = {.entry = {.count = 1, .reusable = true}}, SHIFT(939),
  [14056] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3720),
  [14058] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3693),
  [14060] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3693),
  [14062] = {.entry = {.count = 1, .reusable = true}}, SHIFT(708),
  [14064] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4109),
  [14066] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3630),
  [14068] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3630),
  [14070] = {.entry = {.count = 1, .reusable = true}}, SHIFT(909),
  [14072] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4189),
  [14074] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__expansion_body, 3, .production_id = 51),
  [14076] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7014),
  [14078] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6179),
  [14080] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_string_repeat1, 1),
  [14082] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6183),
  [14084] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6110),
  [14086] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4715),
  [14088] = {.entry = {.count = 1, .reusable = true}}, SHIFT(367),
  [14090] = {.entry = {.count = 1, .reusable = true}}, SHIFT(364),
  [14092] = {.entry = {.count = 1, .reusable = true}}, SHIFT(362),
  [14094] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6131),
  [14096] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_heredoc_body, 1),
  [14098] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6137),
  [14100] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_heredoc_body, 2),
  [14102] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3675),
  [14104] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3675),
  [14106] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2123),
  [14108] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3963),
  [14110] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3655),
  [14112] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3655),
  [14114] = {.entry = {.count = 1, .reusable = true}}, SHIFT(665),
  [14116] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3859),
  [14118] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3648),
  [14120] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3648),
  [14122] = {.entry = {.count = 1, .reusable = true}}, SHIFT(778),
  [14124] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4174),
  [14126] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3694),
  [14128] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3694),
  [14130] = {.entry = {.count = 1, .reusable = true}}, SHIFT(792),
  [14132] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4126),
  [14134] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_heredoc_body_repeat1, 2), SHIFT_REPEAT(6110),
  [14137] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_heredoc_body_repeat1, 2), SHIFT_REPEAT(4715),
  [14140] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_heredoc_body_repeat1, 2), SHIFT_REPEAT(367),
  [14143] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_heredoc_body_repeat1, 2), SHIFT_REPEAT(364),
  [14146] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_heredoc_body_repeat1, 2), SHIFT_REPEAT(362),
  [14149] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_heredoc_body_repeat1, 2), SHIFT_REPEAT(6137),
  [14152] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_heredoc_body_repeat1, 2),
  [14154] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3639),
  [14156] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3639),
  [14158] = {.entry = {.count = 1, .reusable = true}}, SHIFT(949),
  [14160] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3736),
  [14162] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3652),
  [14164] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3652),
  [14166] = {.entry = {.count = 1, .reusable = true}}, SHIFT(645),
  [14168] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3822),
  [14170] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3673),
  [14172] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3673),
  [14174] = {.entry = {.count = 1, .reusable = true}}, SHIFT(689),
  [14176] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4024),
  [14178] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3627),
  [14180] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3627),
  [14182] = {.entry = {.count = 1, .reusable = true}}, SHIFT(738),
  [14184] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3761),
  [14186] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3679),
  [14188] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3679),
  [14190] = {.entry = {.count = 1, .reusable = true}}, SHIFT(772),
  [14192] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4031),
  [14194] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3691),
  [14196] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3691),
  [14198] = {.entry = {.count = 1, .reusable = true}}, SHIFT(833),
  [14200] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4111),
  [14202] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3765),
  [14204] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3646),
  [14206] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3646),
  [14208] = {.entry = {.count = 1, .reusable = true}}, SHIFT(635),
  [14210] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3807),
  [14212] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3674),
  [14214] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3674),
  [14216] = {.entry = {.count = 1, .reusable = true}}, SHIFT(764),
  [14218] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4050),
  [14220] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3632),
  [14222] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3632),
  [14224] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1670),
  [14226] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3740),
  [14228] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3621),
  [14230] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3621),
  [14232] = {.entry = {.count = 1, .reusable = true}}, SHIFT(607),
  [14234] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3777),
  [14236] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3662),
  [14238] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3662),
  [14240] = {.entry = {.count = 1, .reusable = true}}, SHIFT(882),
  [14242] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3898),
  [14244] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1806),
  [14246] = {.entry = {.count = 1, .reusable = false}}, SHIFT(179),
  [14248] = {.entry = {.count = 1, .reusable = true}}, SHIFT(421),
  [14250] = {.entry = {.count = 1, .reusable = true}}, SHIFT(96),
  [14252] = {.entry = {.count = 1, .reusable = false}}, SHIFT(7),
  [14254] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1871),
  [14256] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1791),
  [14258] = {.entry = {.count = 1, .reusable = false}}, SHIFT(53),
  [14260] = {.entry = {.count = 1, .reusable = true}}, SHIFT(422),
  [14262] = {.entry = {.count = 1, .reusable = true}}, SHIFT(86),
  [14264] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6),
  [14266] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1842),
  [14268] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1913),
  [14270] = {.entry = {.count = 1, .reusable = false}}, SHIFT(125),
  [14272] = {.entry = {.count = 1, .reusable = true}}, SHIFT(428),
  [14274] = {.entry = {.count = 1, .reusable = true}}, SHIFT(63),
  [14276] = {.entry = {.count = 1, .reusable = false}}, SHIFT(10),
  [14278] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1884),
  [14280] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5863),
  [14282] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4768),
  [14284] = {.entry = {.count = 1, .reusable = true}}, SHIFT(287),
  [14286] = {.entry = {.count = 1, .reusable = true}}, SHIFT(295),
  [14288] = {.entry = {.count = 1, .reusable = true}}, SHIFT(356),
  [14290] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1757),
  [14292] = {.entry = {.count = 1, .reusable = false}}, SHIFT(145),
  [14294] = {.entry = {.count = 1, .reusable = true}}, SHIFT(425),
  [14296] = {.entry = {.count = 1, .reusable = true}}, SHIFT(72),
  [14298] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5),
  [14300] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1824),
  [14302] = {.entry = {.count = 1, .reusable = false}}, SHIFT(55),
  [14304] = {.entry = {.count = 1, .reusable = false}}, SHIFT(173),
  [14306] = {.entry = {.count = 1, .reusable = false}}, SHIFT(52),
  [14308] = {.entry = {.count = 1, .reusable = false}}, SHIFT(56),
  [14310] = {.entry = {.count = 1, .reusable = false}}, SHIFT(60),
  [14312] = {.entry = {.count = 1, .reusable = false}}, SHIFT(58),
  [14314] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_string_repeat1, 3),
  [14316] = {.entry = {.count = 1, .reusable = false}}, SHIFT(49),
  [14318] = {.entry = {.count = 1, .reusable = false}}, SHIFT(57),
  [14320] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3462),
  [14322] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5833),
  [14324] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__expansion_regex_replacement, 1, .production_id = 12),
  [14326] = {.entry = {.count = 1, .reusable = true}}, SHIFT(211),
  [14328] = {.entry = {.count = 1, .reusable = true}}, SHIFT(212),
  [14330] = {.entry = {.count = 1, .reusable = true}}, SHIFT(213),
  [14332] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6749),
  [14334] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3776),
  [14336] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3776),
  [14338] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2353),
  [14340] = {.entry = {.count = 1, .reusable = false}}, SHIFT(7259),
  [14342] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7259),
  [14344] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6361),
  [14346] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5815),
  [14348] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6194),
  [14350] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__expansion_regex, 2, .production_id = 12),
  [14352] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6361),
  [14354] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6194),
  [14356] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2361),
  [14358] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6884),
  [14360] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6884),
  [14362] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2360),
  [14364] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6882),
  [14366] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6882),
  [14368] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4160),
  [14370] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__expansion_regex_repeat1, 2), SHIFT_REPEAT(6361),
  [14373] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__expansion_regex_repeat1, 2), SHIFT_REPEAT(5815),
  [14376] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__expansion_regex_repeat1, 2), SHIFT_REPEAT(6194),
  [14379] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__expansion_regex_repeat1, 2),
  [14381] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__expansion_regex_repeat1, 2), SHIFT_REPEAT(6361),
  [14384] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__expansion_regex_repeat1, 2), SHIFT_REPEAT(6194),
  [14387] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4813),
  [14389] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6215),
  [14391] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6212),
  [14393] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6190),
  [14395] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__expansion_regex, 1, .production_id = 12),
  [14397] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6190),
  [14399] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2414),
  [14401] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6948),
  [14403] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6948),
  [14405] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2376),
  [14407] = {.entry = {.count = 1, .reusable = false}}, SHIFT(7055),
  [14409] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7055),
  [14411] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2416),
  [14413] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6950),
  [14415] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6950),
  [14417] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4087),
  [14419] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2349),
  [14421] = {.entry = {.count = 1, .reusable = false}}, SHIFT(7258),
  [14423] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7258),
  [14425] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2379),
  [14427] = {.entry = {.count = 1, .reusable = false}}, SHIFT(7057),
  [14429] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7057),
  [14431] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_concatenation_repeat1, 2), SHIFT_REPEAT(3776),
  [14434] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_concatenation_repeat1, 2), SHIFT_REPEAT(3776),
  [14437] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__expansion_max_length_expression, 1),
  [14439] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5212),
  [14441] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__expansion_max_length, 2, .production_id = 12),
  [14443] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__expansion_max_length_binary_expression, 3, .production_id = 27),
  [14445] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6195),
  [14447] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6304),
  [14449] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__expansion_max_length_expression, 1, .production_id = 7),
  [14451] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5211),
  [14453] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__expansion_max_length, 2, .production_id = 50),
  [14455] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__literal_repeat1, 2), SHIFT_REPEAT(6304),
  [14458] = {.entry = {.count = 1, .reusable = true}}, SHIFT(286),
  [14460] = {.entry = {.count = 1, .reusable = true}}, SHIFT(308),
  [14462] = {.entry = {.count = 1, .reusable = true}}, SHIFT(302),
  [14464] = {.entry = {.count = 1, .reusable = true}}, SHIFT(158),
  [14466] = {.entry = {.count = 1, .reusable = true}}, SHIFT(305),
  [14468] = {.entry = {.count = 1, .reusable = true}}, SHIFT(360),
  [14470] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__expansion_max_length, 4, .production_id = 90),
  [14472] = {.entry = {.count = 1, .reusable = true}}, SHIFT(243),
  [14474] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__expansion_max_length, 4, .production_id = 49),
  [14476] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__expansion_max_length, 4, .production_id = 91),
  [14478] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__expansion_max_length, 3, .production_id = 49),
  [14480] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__expansion_max_length, 5, .production_id = 123),
  [14482] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__expansion_max_length, 4, .production_id = 64),
  [14484] = {.entry = {.count = 1, .reusable = true}}, SHIFT(299),
  [14486] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4917),
  [14488] = {.entry = {.count = 1, .reusable = true}}, SHIFT(418),
  [14490] = {.entry = {.count = 1, .reusable = true}}, SHIFT(226),
  [14492] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__expansion_max_length, 3, .production_id = 63),
  [14494] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__expansion_max_length, 4, .production_id = 92),
  [14496] = {.entry = {.count = 1, .reusable = true}}, SHIFT(388),
  [14498] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4651),
  [14500] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4999),
  [14502] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4651),
  [14504] = {.entry = {.count = 1, .reusable = true}}, SHIFT(224),
  [14506] = {.entry = {.count = 1, .reusable = true}}, SHIFT(332),
  [14508] = {.entry = {.count = 1, .reusable = true}}, SHIFT(114),
  [14510] = {.entry = {.count = 1, .reusable = true}}, SHIFT(292),
  [14512] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__expansion_max_length, 5, .production_id = 64),
  [14514] = {.entry = {.count = 1, .reusable = true}}, SHIFT(165),
  [14516] = {.entry = {.count = 1, .reusable = true}}, SHIFT(296),
  [14518] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__expansion_max_length, 5, .production_id = 51),
  [14520] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4646),
  [14522] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4646),
  [14524] = {.entry = {.count = 1, .reusable = true}}, SHIFT(252),
  [14526] = {.entry = {.count = 1, .reusable = true}}, SHIFT(348),
  [14528] = {.entry = {.count = 1, .reusable = true}}, SHIFT(147),
  [14530] = {.entry = {.count = 1, .reusable = true}}, SHIFT(329),
  [14532] = {.entry = {.count = 1, .reusable = true}}, SHIFT(311),
  [14534] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4700),
  [14536] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4700),
  [14538] = {.entry = {.count = 1, .reusable = true}}, SHIFT(79),
  [14540] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4755),
  [14542] = {.entry = {.count = 1, .reusable = true}}, SHIFT(266),
  [14544] = {.entry = {.count = 1, .reusable = true}}, SHIFT(134),
  [14546] = {.entry = {.count = 1, .reusable = true}}, SHIFT(229),
  [14548] = {.entry = {.count = 1, .reusable = true}}, SHIFT(371),
  [14550] = {.entry = {.count = 1, .reusable = true}}, SHIFT(199),
  [14552] = {.entry = {.count = 1, .reusable = true}}, SHIFT(277),
  [14554] = {.entry = {.count = 1, .reusable = true}}, SHIFT(180),
  [14556] = {.entry = {.count = 1, .reusable = true}}, SHIFT(278),
  [14558] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4742),
  [14560] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4742),
  [14562] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__expansion_max_length, 5, .production_id = 122),
  [14564] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3574),
  [14566] = {.entry = {.count = 1, .reusable = true}}, SHIFT(41),
  [14568] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3860),
  [14570] = {.entry = {.count = 1, .reusable = true}}, SHIFT(30),
  [14572] = {.entry = {.count = 1, .reusable = true}}, SHIFT(326),
  [14574] = {.entry = {.count = 1, .reusable = true}}, SHIFT(215),
  [14576] = {.entry = {.count = 1, .reusable = true}}, SHIFT(166),
  [14578] = {.entry = {.count = 1, .reusable = true}}, SHIFT(205),
  [14580] = {.entry = {.count = 1, .reusable = true}}, SHIFT(32),
  [14582] = {.entry = {.count = 1, .reusable = true}}, SHIFT(314),
  [14584] = {.entry = {.count = 1, .reusable = true}}, SHIFT(45),
  [14586] = {.entry = {.count = 1, .reusable = true}}, SHIFT(140),
  [14588] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7285),
  [14590] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5425),
  [14592] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__expansion_regex_replacement, 5, .production_id = 89),
  [14594] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6779),
  [14596] = {.entry = {.count = 1, .reusable = true}}, SHIFT(404),
  [14598] = {.entry = {.count = 1, .reusable = true}}, SHIFT(261),
  [14600] = {.entry = {.count = 1, .reusable = true}}, SHIFT(42),
  [14602] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4236),
  [14604] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4236),
  [14606] = {.entry = {.count = 1, .reusable = true}}, SHIFT(376),
  [14608] = {.entry = {.count = 1, .reusable = true}}, SHIFT(174),
  [14610] = {.entry = {.count = 1, .reusable = true}}, SHIFT(164),
  [14612] = {.entry = {.count = 1, .reusable = true}}, SHIFT(344),
  [14614] = {.entry = {.count = 1, .reusable = true}}, SHIFT(417),
  [14616] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5570),
  [14618] = {.entry = {.count = 1, .reusable = true}}, SHIFT(35),
  [14620] = {.entry = {.count = 1, .reusable = true}}, SHIFT(317),
  [14622] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4803),
  [14624] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4803),
  [14626] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4769),
  [14628] = {.entry = {.count = 1, .reusable = true}}, SHIFT(255),
  [14630] = {.entry = {.count = 1, .reusable = true}}, SHIFT(116),
  [14632] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4741),
  [14634] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4741),
  [14636] = {.entry = {.count = 1, .reusable = true}}, SHIFT(351),
  [14638] = {.entry = {.count = 1, .reusable = true}}, SHIFT(89),
  [14640] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5567),
  [14642] = {.entry = {.count = 1, .reusable = true}}, SHIFT(23),
  [14644] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4610),
  [14646] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4950),
  [14648] = {.entry = {.count = 1, .reusable = true}}, SHIFT(323),
  [14650] = {.entry = {.count = 1, .reusable = true}}, SHIFT(340),
  [14652] = {.entry = {.count = 1, .reusable = true}}, SHIFT(236),
  [14654] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__expansion_max_length, 4, .production_id = 51),
  [14656] = {.entry = {.count = 1, .reusable = true}}, SHIFT(380),
  [14658] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6958),
  [14660] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__expansion_regex_replacement, 3, .production_id = 49),
  [14662] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6791),
  [14664] = {.entry = {.count = 1, .reusable = true}}, SHIFT(374),
  [14666] = {.entry = {.count = 1, .reusable = true}}, SHIFT(178),
  [14668] = {.entry = {.count = 1, .reusable = true}}, SHIFT(249),
  [14670] = {.entry = {.count = 1, .reusable = true}}, SHIFT(69),
  [14672] = {.entry = {.count = 1, .reusable = true}}, SHIFT(110),
  [14674] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7670),
  [14676] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__expansion_regex_replacement, 4, .production_id = 51),
  [14678] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6747),
  [14680] = {.entry = {.count = 1, .reusable = true}}, SHIFT(335),
  [14682] = {.entry = {.count = 1, .reusable = true}}, SHIFT(320),
  [14684] = {.entry = {.count = 1, .reusable = true}}, SHIFT(124),
  [14686] = {.entry = {.count = 1, .reusable = true}}, SHIFT(64),
  [14688] = {.entry = {.count = 1, .reusable = true}}, SHIFT(148),
  [14690] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__expansion_regex_repeat1, 1, .production_id = 48),
  [14692] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__expansion_regex_repeat1, 1, .production_id = 48),
  [14694] = {.entry = {.count = 1, .reusable = true}}, SHIFT(400),
  [14696] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4613),
  [14698] = {.entry = {.count = 1, .reusable = true}}, SHIFT(31),
  [14700] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4241),
  [14702] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4241),
  [14704] = {.entry = {.count = 1, .reusable = true}}, SHIFT(384),
  [14706] = {.entry = {.count = 1, .reusable = true}}, SHIFT(128),
  [14708] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7142),
  [14710] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6128),
  [14712] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7465),
  [14714] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6528),
  [14716] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7530),
  [14718] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__expansion_regex_replacement, 4, .production_id = 85),
  [14720] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3883),
  [14722] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6826),
  [14724] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7523),
  [14726] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3692),
  [14728] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6730),
  [14730] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6730),
  [14732] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6495),
  [14734] = {.entry = {.count = 1, .reusable = true}}, SHIFT(103),
  [14736] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6546),
  [14738] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__expansion_body, 3, .production_id = 52),
  [14740] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_case_item_repeat1, 2, .production_id = 45),
  [14742] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6508),
  [14744] = {.entry = {.count = 1, .reusable = true}}, SHIFT(76),
  [14746] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_if_statement_repeat1, 2),
  [14748] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_if_statement_repeat1, 2), SHIFT_REPEAT(418),
  [14751] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__expansion_body, 3, .production_id = 53),
  [14753] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6527),
  [14755] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6581),
  [14757] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7009),
  [14759] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6529),
  [14761] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3614),
  [14763] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6783),
  [14765] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6783),
  [14767] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7272),
  [14769] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__expansion_regex_replacement, 5, .production_id = 121),
  [14771] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6544),
  [14773] = {.entry = {.count = 1, .reusable = true}}, SHIFT(143),
  [14775] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_concatenation_repeat1, 2), SHIFT_REPEAT(3860),
  [14778] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_case_item_repeat1, 2, .production_id = 46),
  [14780] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6957),
  [14782] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__expansion_regex_replacement, 3, .production_id = 62),
  [14784] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__expansion_body, 1, .production_id = 7),
  [14786] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3889),
  [14788] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6512),
  [14790] = {.entry = {.count = 1, .reusable = true}}, SHIFT(90),
  [14792] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3676),
  [14794] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6732),
  [14796] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6732),
  [14798] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__concatenation_in_expansion_repeat1, 2),
  [14800] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__concatenation_in_expansion_repeat1, 2), SHIFT_REPEAT(5425),
  [14803] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__for_body_repeat1, 2),
  [14805] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__for_body_repeat1, 2), SHIFT_REPEAT(4999),
  [14808] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__for_body_repeat1, 2),
  [14810] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__expansion_body, 2, .production_id = 32),
  [14812] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6196),
  [14814] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3617),
  [14816] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6745),
  [14818] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6745),
  [14820] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__expansion_body_repeat1, 2, .production_id = 34),
  [14822] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__expansion_body_repeat1, 2, .production_id = 34), SHIFT_REPEAT(6489),
  [14825] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__concatenation_in_expansion, 2),
  [14827] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__expansion_body, 2, .production_id = 31),
  [14829] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__expansion_body_repeat1, 1, .production_id = 12),
  [14831] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_last_case_item, 5, .production_id = 109),
  [14833] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5121),
  [14835] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5268),
  [14837] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5272),
  [14839] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_last_case_item, 4, .production_id = 83),
  [14841] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5183),
  [14843] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5355),
  [14845] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5341),
  [14847] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__c_expression, 1, .production_id = 26),
  [14849] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__c_expression, 1, .production_id = 26),
  [14851] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5141),
  [14853] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5296),
  [14855] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5299),
  [14857] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__expansion_expression, 2, .production_id = 12),
  [14859] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__literal_repeat1, 2), SHIFT_REPEAT(6651),
  [14862] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_last_case_item, 5, .production_id = 112),
  [14864] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5139),
  [14866] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5291),
  [14868] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5308),
  [14870] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5201),
  [14872] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5395),
  [14874] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5386),
  [14876] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5126),
  [14878] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5275),
  [14880] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5285),
  [14882] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_last_case_item, 4, .production_id = 79),
  [14884] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5125),
  [14886] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5264),
  [14888] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5265),
  [14890] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__c_variable_assignment, 3, .production_id = 55),
  [14892] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__c_variable_assignment, 3, .production_id = 55),
  [14894] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_last_case_item, 3, .production_id = 59),
  [14896] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5166),
  [14898] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5397),
  [14900] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_last_case_item, 4, .production_id = 45),
  [14902] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5177),
  [14904] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5391),
  [14906] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_last_case_item, 4, .production_id = 46),
  [14908] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5245),
  [14910] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5254),
  [14912] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_last_case_item, 3, .production_id = 60),
  [14914] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5146),
  [14916] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5316),
  [14918] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5215),
  [14920] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5283),
  [14922] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5282),
  [14924] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__literal_repeat1, 2), SHIFT_REPEAT(6581),
  [14927] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3497),
  [14929] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2090),
  [14931] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4796),
  [14933] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3503),
  [14935] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__for_body_repeat1, 2), SHIFT_REPEAT(4972),
  [14938] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4918),
  [14940] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3228),
  [14942] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3513),
  [14944] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_arithmetic_expansion_repeat1, 2), SHIFT_REPEAT(4108),
  [14947] = {.entry = {.count = 1, .reusable = true}}, SHIFT(437),
  [14949] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1221),
  [14951] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3520),
  [14953] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4288),
  [14955] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6535),
  [14957] = {.entry = {.count = 1, .reusable = true}}, SHIFT(446),
  [14959] = {.entry = {.count = 1, .reusable = true}}, SHIFT(38),
  [14961] = {.entry = {.count = 1, .reusable = true}}, SHIFT(39),
  [14963] = {.entry = {.count = 1, .reusable = true}}, SHIFT(448),
  [14965] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2439),
  [14967] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3527),
  [14969] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2289),
  [14971] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1225),
  [14973] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3521),
  [14975] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1815),
  [14977] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1485),
  [14979] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3470),
  [14981] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__expansion_regex_replacement, 2, .production_id = 12),
  [14983] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6789),
  [14985] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6733),
  [14987] = {.entry = {.count = 1, .reusable = true}}, SHIFT(438),
  [14989] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6734),
  [14991] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2479),
  [14993] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__for_body, 6, .production_id = 132),
  [14995] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4972),
  [14997] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1279),
  [14999] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3056),
  [15001] = {.entry = {.count = 1, .reusable = true}}, SHIFT(40),
  [15003] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1455),
  [15005] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1753),
  [15007] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4280),
  [15009] = {.entry = {.count = 1, .reusable = true}}, SHIFT(27),
  [15011] = {.entry = {.count = 1, .reusable = true}}, SHIFT(22),
  [15013] = {.entry = {.count = 1, .reusable = true}}, SHIFT(33),
  [15015] = {.entry = {.count = 1, .reusable = true}}, SHIFT(28),
  [15017] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_case_item_repeat1, 2, .production_id = 80), SHIFT_REPEAT(3574),
  [15020] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_case_item_repeat1, 2, .production_id = 80),
  [15022] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5414),
  [15024] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2260),
  [15026] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4343),
  [15028] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3509),
  [15030] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3493),
  [15032] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2946),
  [15034] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2644),
  [15036] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6615),
  [15038] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1005),
  [15040] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6776),
  [15042] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6778),
  [15044] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5336),
  [15046] = {.entry = {.count = 1, .reusable = true}}, SHIFT(872),
  [15048] = {.entry = {.count = 1, .reusable = true}}, SHIFT(515),
  [15050] = {.entry = {.count = 1, .reusable = true}}, SHIFT(442),
  [15052] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3524),
  [15054] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1598),
  [15056] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__for_body, 5, .production_id = 102),
  [15058] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3499),
  [15060] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5342),
  [15062] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2683),
  [15064] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2675),
  [15066] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1504),
  [15068] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1995),
  [15070] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1409),
  [15072] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__for_body, 5, .production_id = 101),
  [15074] = {.entry = {.count = 1, .reusable = true}}, SHIFT(44),
  [15076] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3507),
  [15078] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2721),
  [15080] = {.entry = {.count = 1, .reusable = true}}, SHIFT(980),
  [15082] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__for_body, 5, .production_id = 100),
  [15084] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2504),
  [15086] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3489),
  [15088] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3526),
  [15090] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5343),
  [15092] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3530),
  [15094] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3311),
  [15096] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3189),
  [15098] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3514),
  [15100] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4508),
  [15102] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1849),
  [15104] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2398),
  [15106] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__for_body, 4, .production_id = 74),
  [15108] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1979),
  [15110] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__for_body, 4, .production_id = 72),
  [15112] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5319),
  [15114] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4482),
  [15116] = {.entry = {.count = 1, .reusable = true}}, SHIFT(46),
  [15118] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2861),
  [15120] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__for_body, 5, .production_id = 98),
  [15122] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4474),
  [15124] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6138),
  [15126] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1033),
  [15128] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__for_body, 4, .production_id = 70),
  [15130] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3380),
  [15132] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5441),
  [15134] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1015),
  [15136] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3519),
  [15138] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3517),
  [15140] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__for_body, 6, .production_id = 130),
  [15142] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5334),
  [15144] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5339),
  [15146] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3502),
  [15148] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4598),
  [15150] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__for_body, 7, .production_id = 150),
  [15152] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3406),
  [15154] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5302),
  [15156] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3523),
  [15158] = {.entry = {.count = 1, .reusable = true}}, SHIFT(435),
  [15160] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__for_body, 3, .production_id = 58),
  [15162] = {.entry = {.count = 1, .reusable = true}}, SHIFT(36),
  [15164] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6766),
  [15166] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__for_body, 5, .production_id = 96),
  [15168] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__for_body_repeat1, 2), SHIFT_REPEAT(4918),
  [15171] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1216),
  [15173] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__for_body, 6, .production_id = 133),
  [15175] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1477),
  [15177] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4801),
  [15179] = {.entry = {.count = 1, .reusable = true}}, SHIFT(474),
  [15181] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2544),
  [15183] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3487),
  [15185] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1375),
  [15187] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1802),
  [15189] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__for_body, 8, .production_id = 158),
  [15191] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5333),
  [15193] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1651),
  [15195] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1500),
  [15197] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2600),
  [15199] = {.entry = {.count = 1, .reusable = true}}, SHIFT(26),
  [15201] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3495),
  [15203] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5508),
  [15205] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6207),
  [15207] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1145),
  [15209] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5344),
  [15211] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5400),
  [15213] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__for_body, 6, .production_id = 129),
  [15215] = {.entry = {.count = 1, .reusable = true}}, SHIFT(37),
  [15217] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3518),
  [15219] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3496),
  [15221] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__for_body, 7, .production_id = 151),
  [15223] = {.entry = {.count = 1, .reusable = true}}, SHIFT(29),
  [15225] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1917),
  [15227] = {.entry = {.count = 1, .reusable = true}}, SHIFT(25),
  [15229] = {.entry = {.count = 1, .reusable = true}}, SHIFT(24),
  [15231] = {.entry = {.count = 1, .reusable = true}}, SHIFT(43),
  [15233] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5467),
  [15235] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__for_body, 7, .production_id = 152),
  [15237] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__for_body, 6, .production_id = 128),
  [15239] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5683),
  [15241] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5345),
  [15243] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3500),
  [15245] = {.entry = {.count = 1, .reusable = true}}, SHIFT(443),
  [15247] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5346),
  [15249] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4317),
  [15251] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1671),
  [15253] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5324),
  [15255] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5123),
  [15257] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1505),
  [15259] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4219),
  [15261] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6381),
  [15263] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6946),
  [15265] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__expansion_regex_replacement, 5, .production_id = 119),
  [15267] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4217),
  [15269] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6408),
  [15271] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__expansion_regex_removal, 1, .production_id = 12),
  [15273] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7029),
  [15275] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4226),
  [15277] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6443),
  [15279] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7065),
  [15281] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__expansion_regex_replacement, 6, .production_id = 89),
  [15283] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4220),
  [15285] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6471),
  [15287] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3486),
  [15289] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__expansion_regex_replacement, 3, .production_id = 12),
  [15291] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7303),
  [15293] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__expansion_regex_replacement, 4, .production_id = 88),
  [15295] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6457),
  [15297] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7490),
  [15299] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1996),
  [15301] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2299),
  [15303] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5060),
  [15305] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7532),
  [15307] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2012),
  [15309] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1766),
  [15311] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5389),
  [15313] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7453),
  [15315] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7404),
  [15317] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4369),
  [15319] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3833),
  [15321] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6458),
  [15323] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6516),
  [15325] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6483),
  [15327] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1280),
  [15329] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7415),
  [15331] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7244),
  [15333] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3801),
  [15335] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2478),
  [15337] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1836),
  [15339] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1838),
  [15341] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5403),
  [15343] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2462),
  [15345] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7343),
  [15347] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2574),
  [15349] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4653),
  [15351] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4628),
  [15353] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5027),
  [15355] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4988),
  [15357] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2287),
  [15359] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2565),
  [15361] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7297),
  [15363] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6210),
  [15365] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2058),
  [15367] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6383),
  [15369] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6398),
  [15371] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6431),
  [15373] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7191),
  [15375] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6434),
  [15377] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1564),
  [15379] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4735),
  [15381] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6497),
  [15383] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6412),
  [15385] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6473),
  [15387] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4987),
  [15389] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6830),
  [15391] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7149),
  [15393] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4152),
  [15395] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1756),
  [15397] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1019),
  [15399] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5030),
  [15401] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5450),
  [15403] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6422),
  [15405] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1754),
  [15407] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7054),
  [15409] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4407),
  [15411] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4408),
  [15413] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5451),
  [15415] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3490),
  [15417] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1239),
  [15419] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1237),
  [15421] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6949),
  [15423] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6444),
  [15425] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2701),
  [15427] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6387),
  [15429] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2700),
  [15431] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6430),
  [15433] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6414),
  [15435] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6230),
  [15437] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6440),
  [15439] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6885),
  [15441] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6418),
  [15443] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6232),
  [15445] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6241),
  [15447] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3352),
  [15449] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3354),
  [15451] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2377),
  [15453] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6828),
  [15455] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2428),
  [15457] = {.entry = {.count = 1, .reusable = true}}, SHIFT(476),
  [15459] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6386),
  [15461] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7074),
  [15463] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3706),
  [15465] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6159),
  [15467] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6221),
  [15469] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6869),
  [15471] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1016),
  [15473] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6167),
  [15475] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2520),
  [15477] = {.entry = {.count = 1, .reusable = true}}, SHIFT(475),
  [15479] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6485),
  [15481] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4873),
  [15483] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1697),
  [15485] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2346),
  [15487] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6912),
  [15489] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3506),
  [15491] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1692),
  [15493] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2614),
  [15495] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3504),
  [15497] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1813),
  [15499] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2301),
  [15501] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7015),
  [15503] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2294),
  [15505] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4955),
  [15507] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5644),
  [15509] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5431),
  [15511] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5446),
  [15513] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4978),
  [15515] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7071),
  [15517] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5648),
  [15519] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4655),
  [15521] = {.entry = {.count = 1, .reusable = true}}, SHIFT(479),
  [15523] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4960),
  [15525] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2430),
  [15527] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2321),
  [15529] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7276),
  [15531] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2319),
  [15533] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6446),
  [15535] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2421),
  [15537] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2315),
  [15539] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6369),
  [15541] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7435),
  [15543] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6462),
  [15545] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6450),
  [15547] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2383),
  [15549] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2057),
  [15551] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6385),
  [15553] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7076),
  [15555] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6391),
  [15557] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4895),
  [15559] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1017),
  [15561] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6377),
  [15563] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6423),
  [15565] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__expansion_regex_replacement, 6, .production_id = 146),
  [15567] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7443),
  [15569] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2410),
  [15571] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2384),
  [15573] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2411),
  [15575] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6441),
  [15577] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7376),
  [15579] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7322),
  [15581] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4053),
  [15583] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6168),
  [15585] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__expansion_regex_replacement, 4, .production_id = 86),
  [15587] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__expansion_regex_replacement, 4, .production_id = 87),
  [15589] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7281),
  [15591] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2388),
  [15593] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7145),
  [15595] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6160),
  [15597] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2354),
  [15599] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7116),
  [15601] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1566),
  [15603] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7090),
  [15605] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1531),
  [15607] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7019),
  [15609] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5677),
  [15611] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6852),
  [15613] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5678),
  [15615] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6468),
  [15617] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6814),
  [15619] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5691),
  [15621] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5694),
  [15623] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5572),
  [15625] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6817),
  [15627] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5573),
  [15629] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5582),
  [15631] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5682),
  [15633] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__expansion_operator, 2, .production_id = 49),
  [15635] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2046),
  [15637] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5684),
  [15639] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6858),
  [15641] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2592),
  [15643] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5686),
  [15645] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5685),
  [15647] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6916),
  [15649] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6464),
  [15651] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__expansion_regex_removal, 2, .production_id = 12),
  [15653] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6479),
  [15655] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7106),
  [15657] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2591),
  [15659] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2535),
  [15661] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5159),
  [15663] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5073),
  [15665] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6459),
  [15667] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__test_command_binary_expression, 3, .production_id = 43),
  [15669] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7172),
  [15671] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6454),
  [15673] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2554),
  [15675] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6432),
  [15677] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6429),
  [15679] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5161),
  [15681] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4874),
  [15683] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7247),
  [15685] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2499),
  [15687] = {.entry = {.count = 1, .reusable = true}}, SHIFT(478),
  [15689] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5075),
  [15691] = {.entry = {.count = 1, .reusable = true}}, SHIFT(491),
  [15693] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3322),
  [15695] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7287),
  [15697] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4647),
  [15699] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3501),
  [15701] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2364),
  [15703] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4876),
  [15705] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2358),
  [15707] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7450),
  [15709] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1997),
  [15711] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6396),
  [15713] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6971),
  [15715] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3721),
  [15717] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6180),
  [15719] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__expansion_regex_replacement, 7, .production_id = 157),
  [15721] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7534),
  [15723] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3512),
  [15725] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6175),
  [15727] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2708),
  [15729] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1993),
  [15731] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4447),
  [15733] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7525),
  [15735] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1013),
  [15737] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2710),
  [15739] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4403),
  [15741] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4449),
  [15743] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4443),
  [15745] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7505),
  [15747] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1583),
  [15749] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4445),
  [15751] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1584),
  [15753] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1594),
  [15755] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7496),
  [15757] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4437),
  [15759] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2498),
  [15761] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2615),
  [15763] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6490),
  [15765] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4543),
  [15767] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7481),
  [15769] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4853),
  [15771] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4849),
  [15773] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2502),
  [15775] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4275),
  [15777] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4264),
  [15779] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7468),
  [15781] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5157),
  [15783] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4427),
  [15785] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6740),
  [15787] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1762),
  [15789] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7452),
  [15791] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1760),
  [15793] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2598),
  [15795] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1527),
  [15797] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1020),
  [15799] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7442),
  [15801] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6534),
  [15803] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4642),
  [15805] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1103),
  [15807] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1099),
  [15809] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6522),
  [15811] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4607),
  [15813] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7425),
  [15815] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1819),
  [15817] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1817),
  [15819] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4780),
  [15821] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1954),
  [15823] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7346),
  [15825] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1593),
  [15827] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4311),
  [15829] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1588),
  [15831] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5076),
  [15833] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1764),
  [15835] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7406),
  [15837] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5473),
  [15839] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1350),
  [15841] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1359),
  [15843] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1136),
  [15845] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5047),
  [15847] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7387),
  [15849] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1833),
  [15851] = {.entry = {.count = 1, .reusable = true}}, SHIFT(520),
  [15853] = {.entry = {.count = 1, .reusable = true}}, SHIFT(519),
  [15855] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1205),
  [15857] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7381),
  [15859] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1323),
  [15861] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1340),
  [15863] = {.entry = {.count = 1, .reusable = true}}, SHIFT(517),
  [15865] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1196),
  [15867] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3026),
  [15869] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3024),
  [15871] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7365),
  [15873] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4612),
  [15875] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2654),
  [15877] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1713),
  [15879] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7351),
  [15881] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3376),
  [15883] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4354),
  [15885] = {.entry = {.count = 1, .reusable = true}}, SHIFT(516),
  [15887] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4356),
  [15889] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1389),
  [15891] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7337),
  [15893] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3360),
  [15895] = {.entry = {.count = 1, .reusable = true}}, SHIFT(927),
  [15897] = {.entry = {.count = 1, .reusable = true}}, SHIFT(929),
  [15899] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1029),
  [15901] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7324),
  [15903] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1334),
  [15905] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4328),
  [15907] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4330),
  [15909] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1840),
  [15911] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2642),
  [15913] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7309),
  [15915] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4327),
  [15917] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5588),
  [15919] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1366),
  [15921] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1364),
  [15923] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1728),
  [15925] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7226),
  [15927] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2526),
  [15929] = {.entry = {.count = 1, .reusable = true}}, SHIFT(921),
  [15931] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2496),
  [15933] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7278),
  [15935] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5459),
  [15937] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4552),
  [15939] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4560),
  [15941] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2528),
  [15943] = {.entry = {.count = 1, .reusable = true}}, SHIFT(736),
  [15945] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7263),
  [15947] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4302),
  [15949] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2617),
  [15951] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5454),
  [15953] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1732),
  [15955] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1578),
  [15957] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7248),
  [15959] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1581),
  [15961] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2596),
  [15963] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1400),
  [15965] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5622),
  [15967] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2244),
  [15969] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4656),
  [15971] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3180),
  [15973] = {.entry = {.count = 1, .reusable = true}}, SHIFT(898),
  [15975] = {.entry = {.count = 1, .reusable = true}}, SHIFT(901),
  [15977] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4569),
  [15979] = {.entry = {.count = 1, .reusable = true}}, SHIFT(508),
  [15981] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4502),
  [15983] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5693),
  [15985] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4490),
  [15987] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4478),
  [15989] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4545),
  [15991] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1459),
  [15993] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1435),
  [15995] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3157),
  [15997] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1514),
  [15999] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2378),
  [16001] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2372),
  [16003] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1841),
  [16005] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1554),
  [16007] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1553),
  [16009] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4454),
  [16011] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4438),
  [16013] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1242),
  [16015] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1240),
  [16017] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6171),
  [16019] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7062),
  [16021] = {.entry = {.count = 1, .reusable = true}}, SHIFT(918),
  [16023] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__expansion_regex_replacement, 6, .production_id = 147),
  [16025] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6184),
  [16027] = {.entry = {.count = 1, .reusable = true}}, SHIFT(729),
  [16029] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5078),
  [16031] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4548),
  [16033] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1545),
  [16035] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1674),
  [16037] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1516),
  [16039] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1256),
  [16041] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2224),
  [16043] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1254),
  [16045] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1456),
  [16047] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__expansion_regex_replacement, 6, .production_id = 148),
  [16049] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1544),
  [16051] = {.entry = {.count = 1, .reusable = true}}, SHIFT(831),
  [16053] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4618),
  [16055] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6224),
  [16057] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6233),
  [16059] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5093),
  [16061] = {.entry = {.count = 1, .reusable = true}}, SHIFT(728),
  [16063] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5053),
  [16065] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2222),
  [16067] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1286),
  [16069] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1652),
  [16071] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5337),
  [16073] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2446),
  [16075] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2214),
  [16077] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2218),
  [16079] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__expansion_regex_replacement, 5, .production_id = 120),
  [16081] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1644),
  [16083] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1187),
  [16085] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1183),
  [16087] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4456),
  [16089] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1809),
  [16091] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1812),
  [16093] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5434),
  [16095] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2647),
  [16097] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5432),
  [16099] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5090),
  [16101] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1306),
  [16103] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5091),
  [16105] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2570),
  [16107] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4902),
  [16109] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4670),
  [16111] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1142),
  [16113] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1133),
  [16115] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6417),
  [16117] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2684),
  [16119] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2687),
  [16121] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7176),
  [16123] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1486),
  [16125] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6237),
  [16127] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4708),
  [16129] = {.entry = {.count = 1, .reusable = true}}, SHIFT(869),
  [16131] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3178),
  [16133] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1487),
  [16135] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1624),
  [16137] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1612),
  [16139] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5522),
  [16141] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5527),
  [16143] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1491),
  [16145] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1541),
  [16147] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3779),
  [16149] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5445),
  [16151] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1473),
  [16153] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2313),
  [16155] = {.entry = {.count = 1, .reusable = true}}, SHIFT(769),
  [16157] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5592),
  [16159] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1408),
  [16161] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6173),
  [16163] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1034),
  [16165] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1046),
  [16167] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1043),
  [16169] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1410),
  [16171] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2918),
  [16173] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2920),
  [16175] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2677),
  [16177] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1495),
  [16179] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1424),
  [16181] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1430),
  [16183] = {.entry = {.count = 1, .reusable = true}}, SHIFT(855),
  [16185] = {.entry = {.count = 1, .reusable = true}}, SHIFT(857),
  [16187] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1587),
  [16189] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6499),
  [16191] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6540),
  [16193] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1047),
  [16195] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5511),
  [16197] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1519),
  [16199] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1277),
  [16201] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2629),
  [16203] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2105),
  [16205] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4336),
  [16207] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4335),
  [16209] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2908),
  [16211] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2910),
  [16213] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2907),
  [16215] = {.entry = {.count = 1, .reusable = true}}, SHIFT(988),
  [16217] = {.entry = {.count = 1, .reusable = true}}, SHIFT(990),
  [16219] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2668),
  [16221] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2670),
  [16223] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6186),
  [16225] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6953),
  [16227] = {.entry = {.count = 1, .reusable = true}}, SHIFT(846),
  [16229] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6169),
  [16231] = {.entry = {.count = 1, .reusable = true}}, SHIFT(744),
  [16233] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6234),
  [16235] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2626),
  [16237] = {.entry = {.count = 1, .reusable = true}}, SHIFT(745),
  [16239] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4980),
  [16241] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5262),
  [16243] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1508),
  [16245] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1928),
  [16247] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4267),
  [16249] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4265),
  [16251] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4954),
  [16253] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5668),
  [16255] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5670),
  [16257] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2892),
  [16259] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4981),
  [16261] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5263),
  [16263] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2719),
  [16265] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1471),
  [16267] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3250),
  [16269] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3256),
  [16271] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2660),
  [16273] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2854),
  [16275] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4417),
  [16277] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2722),
  [16279] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4965),
  [16281] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4442),
  [16283] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6123),
  [16285] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6154),
  [16287] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6435),
  [16289] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6122),
  [16291] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6438),
  [16293] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1920),
  [16295] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5664),
  [16297] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5536),
  [16299] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7330),
  [16301] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5028),
  [16303] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5020),
  [16305] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1517),
  [16307] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5165),
  [16309] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5018),
  [16311] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1520),
  [16313] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2881),
  [16315] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2883),
  [16317] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3226),
  [16319] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2857),
  [16321] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5032),
  [16323] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1009),
  [16325] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2735),
  [16327] = {.entry = {.count = 1, .reusable = true}}, SHIFT(963),
  [16329] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5484),
  [16331] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2731),
  [16333] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3234),
  [16335] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3251),
  [16337] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5493),
  [16339] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3222),
  [16341] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5171),
  [16343] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2875),
  [16345] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5659),
  [16347] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5650),
  [16349] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2037),
  [16351] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1437),
  [16353] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_program, 1),
  [16355] = {.entry = {.count = 1, .reusable = true}},  ACCEPT_INPUT(),
  [16357] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1434),
  [16359] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6556),
  [16361] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1992),
  [16363] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6554),
  [16365] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5491),
  [16367] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4821),
  [16369] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1222),
  [16371] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6174),
  [16373] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6887),
  [16375] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1220),
  [16377] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2529),
  [16379] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6181),
  [16381] = {.entry = {.count = 1, .reusable = true}}, SHIFT(771),
  [16383] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2531),
  [16385] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4871),
  [16387] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1303),
  [16389] = {.entry = {.count = 1, .reusable = true}}, SHIFT(770),
  [16391] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6563),
  [16393] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4969),
  [16395] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4963),
  [16397] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1210),
  [16399] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1208),
  [16401] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3171),
  [16403] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3173),
  [16405] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2443),
  [16407] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1296),
  [16409] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4899),
  [16411] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4894),
  [16413] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5457),
  [16415] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5453),
  [16417] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4365),
  [16419] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4374),
  [16421] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3170),
  [16423] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6172),
  [16425] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6846),
  [16427] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6177),
  [16429] = {.entry = {.count = 1, .reusable = true}}, SHIFT(758),
  [16431] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2823),
  [16433] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6185),
  [16435] = {.entry = {.count = 1, .reusable = true}}, SHIFT(753),
  [16437] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2821),
  [16439] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4878),
  [16441] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2824),
  [16443] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4880),
  [16445] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3255),
  [16447] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2640),
  [16449] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2639),
  [16451] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6537),
  [16453] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5119),
  [16455] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6525),
  [16457] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__expansion_regex_replacement, 5, .production_id = 117),
  [16459] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1847),
  [16461] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4334),
  [16463] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5534),
  [16465] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1852),
  [16467] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1855),
  [16469] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6812),
  [16471] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6803),
  [16473] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6797),
  [16475] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6793),
  [16477] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6802),
  [16479] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6811),
  [16481] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6820),
  [16483] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6829),
  [16485] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6838),
  [16487] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6847),
  [16489] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6856),
  [16491] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6865),
  [16493] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6874),
  [16495] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6883),
  [16497] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6892),
  [16499] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6901),
  [16501] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6910),
  [16503] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6919),
  [16505] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6926),
  [16507] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6933),
  [16509] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6940),
  [16511] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6947),
  [16513] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6954),
  [16515] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6961),
  [16517] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6968),
  [16519] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6975),
  [16521] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6982),
  [16523] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6989),
  [16525] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6996),
  [16527] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7003),
  [16529] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7010),
  [16531] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7017),
  [16533] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7024),
  [16535] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7031),
  [16537] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7038),
  [16539] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7045),
  [16541] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7052),
  [16543] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7059),
  [16545] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7066),
  [16547] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7073),
  [16549] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7080),
  [16551] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7087),
  [16553] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7094),
  [16555] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7101),
  [16557] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7108),
  [16559] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7115),
  [16561] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7122),
  [16563] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7129),
  [16565] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7136),
  [16567] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7143),
  [16569] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7150),
  [16571] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7157),
  [16573] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7164),
  [16575] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7171),
  [16577] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7178),
  [16579] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7185),
  [16581] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7192),
  [16583] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7199),
  [16585] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7206),
  [16587] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7213),
  [16589] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7269),
  [16591] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1853),
  [16593] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6245),
  [16595] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1856),
  [16597] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5107),
  [16599] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7393),
  [16601] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6247),
  [16603] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7439),
  [16605] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7483),
  [16607] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7512),
  [16609] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7537),
  [16611] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7538),
  [16613] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7539),
  [16615] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7540),
  [16617] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7541),
  [16619] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7542),
  [16621] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7543),
  [16623] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7544),
  [16625] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7545),
  [16627] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7546),
  [16629] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7547),
  [16631] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7548),
  [16633] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7549),
  [16635] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7550),
  [16637] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7551),
  [16639] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7552),
  [16641] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7553),
  [16643] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7554),
  [16645] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7555),
  [16647] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7556),
  [16649] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7557),
  [16651] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7558),
  [16653] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7559),
  [16655] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7560),
  [16657] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7561),
  [16659] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7562),
  [16661] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7563),
  [16663] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7564),
  [16665] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7565),
  [16667] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7566),
  [16669] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7567),
  [16671] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7568),
  [16673] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7569),
  [16675] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7570),
  [16677] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7571),
  [16679] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7572),
  [16681] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7573),
  [16683] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7574),
  [16685] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7575),
  [16687] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7576),
  [16689] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7577),
  [16691] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7578),
  [16693] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7579),
  [16695] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7580),
  [16697] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7581),
  [16699] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7582),
  [16701] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7583),
  [16703] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7584),
  [16705] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7585),
  [16707] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7586),
  [16709] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7587),
  [16711] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7588),
  [16713] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7589),
  [16715] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7590),
  [16717] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7591),
  [16719] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7592),
  [16721] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7593),
  [16723] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7594),
  [16725] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7595),
  [16727] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7596),
  [16729] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__expansion_regex_replacement, 5, .production_id = 118),
};

enum ts_external_scanner_symbol_identifiers {
  ts_external_token_heredoc_start = 0,
  ts_external_token_simple_heredoc_body = 1,
  ts_external_token__heredoc_body_beginning = 2,
  ts_external_token_heredoc_content = 3,
  ts_external_token_heredoc_end = 4,
  ts_external_token_file_descriptor = 5,
  ts_external_token__empty_value = 6,
  ts_external_token__concat = 7,
  ts_external_token_variable_name = 8,
  ts_external_token_test_operator = 9,
  ts_external_token_regex = 10,
  ts_external_token__regex_no_slash = 11,
  ts_external_token__regex_no_space = 12,
  ts_external_token__expansion_word = 13,
  ts_external_token_extglob_pattern = 14,
  ts_external_token__bare_dollar = 15,
  ts_external_token__brace_start = 16,
  ts_external_token__immediate_double_hash = 17,
  ts_external_token__external_expansion_sym_hash = 18,
  ts_external_token__external_expansion_sym_bang = 19,
  ts_external_token__external_expansion_sym_equal = 20,
  ts_external_token_RBRACE = 21,
  ts_external_token_RBRACK = 22,
  ts_external_token_LT_LT = 23,
  ts_external_token_LT_LT_DASH = 24,
  ts_external_token_heredoc_redirect_token1 = 25,
  ts_external_token_LPAREN = 26,
  ts_external_token_esac = 27,
  ts_external_token___error_recovery = 28,
};

static const TSSymbol ts_external_scanner_symbol_map[EXTERNAL_TOKEN_COUNT] = {
  [ts_external_token_heredoc_start] = sym_heredoc_start,
  [ts_external_token_simple_heredoc_body] = sym_simple_heredoc_body,
  [ts_external_token__heredoc_body_beginning] = sym__heredoc_body_beginning,
  [ts_external_token_heredoc_content] = sym_heredoc_content,
  [ts_external_token_heredoc_end] = sym_heredoc_end,
  [ts_external_token_file_descriptor] = sym_file_descriptor,
  [ts_external_token__empty_value] = sym__empty_value,
  [ts_external_token__concat] = sym__concat,
  [ts_external_token_variable_name] = sym_variable_name,
  [ts_external_token_test_operator] = sym_test_operator,
  [ts_external_token_regex] = sym_regex,
  [ts_external_token__regex_no_slash] = sym__regex_no_slash,
  [ts_external_token__regex_no_space] = sym__regex_no_space,
  [ts_external_token__expansion_word] = sym__expansion_word,
  [ts_external_token_extglob_pattern] = sym_extglob_pattern,
  [ts_external_token__bare_dollar] = sym__bare_dollar,
  [ts_external_token__brace_start] = sym__brace_start,
  [ts_external_token__immediate_double_hash] = sym__immediate_double_hash,
  [ts_external_token__external_expansion_sym_hash] = sym__external_expansion_sym_hash,
  [ts_external_token__external_expansion_sym_bang] = sym__external_expansion_sym_bang,
  [ts_external_token__external_expansion_sym_equal] = sym__external_expansion_sym_equal,
  [ts_external_token_RBRACE] = anon_sym_RBRACE3,
  [ts_external_token_RBRACK] = anon_sym_RBRACK,
  [ts_external_token_LT_LT] = anon_sym_LT_LT,
  [ts_external_token_LT_LT_DASH] = anon_sym_LT_LT_DASH,
  [ts_external_token_heredoc_redirect_token1] = aux_sym_heredoc_redirect_token1,
  [ts_external_token_LPAREN] = anon_sym_LPAREN,
  [ts_external_token_esac] = anon_sym_esac,
  [ts_external_token___error_recovery] = sym___error_recovery,
};

static const bool ts_external_scanner_states[131][EXTERNAL_TOKEN_COUNT] = {
  [1] = {
    [ts_external_token_heredoc_start] = true,
    [ts_external_token_simple_heredoc_body] = true,
    [ts_external_token__heredoc_body_beginning] = true,
    [ts_external_token_heredoc_content] = true,
    [ts_external_token_heredoc_end] = true,
    [ts_external_token_file_descriptor] = true,
    [ts_external_token__empty_value] = true,
    [ts_external_token__concat] = true,
    [ts_external_token_variable_name] = true,
    [ts_external_token_test_operator] = true,
    [ts_external_token_regex] = true,
    [ts_external_token__regex_no_slash] = true,
    [ts_external_token__regex_no_space] = true,
    [ts_external_token__expansion_word] = true,
    [ts_external_token_extglob_pattern] = true,
    [ts_external_token__bare_dollar] = true,
    [ts_external_token__brace_start] = true,
    [ts_external_token__immediate_double_hash] = true,
    [ts_external_token__external_expansion_sym_hash] = true,
    [ts_external_token__external_expansion_sym_bang] = true,
    [ts_external_token__external_expansion_sym_equal] = true,
    [ts_external_token_RBRACE] = true,
    [ts_external_token_RBRACK] = true,
    [ts_external_token_LT_LT] = true,
    [ts_external_token_LT_LT_DASH] = true,
    [ts_external_token_LPAREN] = true,
    [ts_external_token_esac] = true,
    [ts_external_token___error_recovery] = true,
  },
  [2] = {
    [ts_external_token_file_descriptor] = true,
    [ts_external_token_variable_name] = true,
    [ts_external_token_test_operator] = true,
    [ts_external_token__brace_start] = true,
    [ts_external_token_LPAREN] = true,
  },
  [3] = {
    [ts_external_token_file_descriptor] = true,
    [ts_external_token_variable_name] = true,
    [ts_external_token_test_operator] = true,
    [ts_external_token__brace_start] = true,
    [ts_external_token_RBRACK] = true,
    [ts_external_token_LPAREN] = true,
  },
  [4] = {
    [ts_external_token_file_descriptor] = true,
    [ts_external_token__concat] = true,
    [ts_external_token_test_operator] = true,
    [ts_external_token__bare_dollar] = true,
    [ts_external_token__brace_start] = true,
    [ts_external_token_LT_LT] = true,
    [ts_external_token_LT_LT_DASH] = true,
    [ts_external_token_heredoc_redirect_token1] = true,
    [ts_external_token_LPAREN] = true,
  },
  [5] = {
    [ts_external_token_file_descriptor] = true,
    [ts_external_token__concat] = true,
    [ts_external_token_test_operator] = true,
    [ts_external_token__bare_dollar] = true,
    [ts_external_token__brace_start] = true,
    [ts_external_token_RBRACK] = true,
    [ts_external_token_LT_LT] = true,
    [ts_external_token_LT_LT_DASH] = true,
    [ts_external_token_LPAREN] = true,
  },
  [6] = {
    [ts_external_token_file_descriptor] = true,
    [ts_external_token_variable_name] = true,
    [ts_external_token_test_operator] = true,
    [ts_external_token__brace_start] = true,
    [ts_external_token_LPAREN] = true,
    [ts_external_token_esac] = true,
  },
  [7] = {
    [ts_external_token__concat] = true,
    [ts_external_token_test_operator] = true,
    [ts_external_token__brace_start] = true,
    [ts_external_token_LT_LT] = true,
    [ts_external_token_LPAREN] = true,
  },
  [8] = {
    [ts_external_token__concat] = true,
    [ts_external_token_test_operator] = true,
    [ts_external_token__brace_start] = true,
    [ts_external_token_RBRACK] = true,
    [ts_external_token_LT_LT] = true,
    [ts_external_token_LPAREN] = true,
  },
  [9] = {
    [ts_external_token_file_descriptor] = true,
    [ts_external_token_variable_name] = true,
    [ts_external_token_test_operator] = true,
    [ts_external_token__bare_dollar] = true,
    [ts_external_token__brace_start] = true,
    [ts_external_token_LT_LT] = true,
    [ts_external_token_LT_LT_DASH] = true,
    [ts_external_token_heredoc_redirect_token1] = true,
    [ts_external_token_LPAREN] = true,
  },
  [10] = {
    [ts_external_token_file_descriptor] = true,
    [ts_external_token_variable_name] = true,
    [ts_external_token_test_operator] = true,
    [ts_external_token__bare_dollar] = true,
    [ts_external_token__brace_start] = true,
    [ts_external_token_RBRACK] = true,
    [ts_external_token_LT_LT] = true,
    [ts_external_token_LT_LT_DASH] = true,
    [ts_external_token_LPAREN] = true,
  },
  [11] = {
    [ts_external_token_file_descriptor] = true,
    [ts_external_token_test_operator] = true,
    [ts_external_token__bare_dollar] = true,
    [ts_external_token__brace_start] = true,
    [ts_external_token_LT_LT] = true,
    [ts_external_token_LT_LT_DASH] = true,
    [ts_external_token_heredoc_redirect_token1] = true,
    [ts_external_token_LPAREN] = true,
  },
  [12] = {
    [ts_external_token_file_descriptor] = true,
    [ts_external_token_test_operator] = true,
    [ts_external_token__bare_dollar] = true,
    [ts_external_token__brace_start] = true,
    [ts_external_token_RBRACK] = true,
    [ts_external_token_LT_LT] = true,
    [ts_external_token_LT_LT_DASH] = true,
    [ts_external_token_LPAREN] = true,
  },
  [13] = {
    [ts_external_token_file_descriptor] = true,
    [ts_external_token_test_operator] = true,
    [ts_external_token__bare_dollar] = true,
    [ts_external_token__brace_start] = true,
    [ts_external_token_LT_LT] = true,
    [ts_external_token_LT_LT_DASH] = true,
    [ts_external_token_heredoc_redirect_token1] = true,
    [ts_external_token_LPAREN] = true,
    [ts_external_token_esac] = true,
  },
  [14] = {
    [ts_external_token_file_descriptor] = true,
    [ts_external_token_test_operator] = true,
    [ts_external_token__bare_dollar] = true,
    [ts_external_token__brace_start] = true,
    [ts_external_token_LT_LT] = true,
    [ts_external_token_LT_LT_DASH] = true,
    [ts_external_token_heredoc_redirect_token1] = true,
    [ts_external_token_esac] = true,
  },
  [15] = {
    [ts_external_token_file_descriptor] = true,
    [ts_external_token_variable_name] = true,
    [ts_external_token_test_operator] = true,
    [ts_external_token__bare_dollar] = true,
    [ts_external_token__brace_start] = true,
    [ts_external_token_LT_LT] = true,
    [ts_external_token_LT_LT_DASH] = true,
    [ts_external_token_heredoc_redirect_token1] = true,
    [ts_external_token_LPAREN] = true,
    [ts_external_token_esac] = true,
  },
  [16] = {
    [ts_external_token_file_descriptor] = true,
    [ts_external_token_test_operator] = true,
    [ts_external_token__bare_dollar] = true,
    [ts_external_token__brace_start] = true,
    [ts_external_token_LT_LT] = true,
    [ts_external_token_LT_LT_DASH] = true,
    [ts_external_token_heredoc_redirect_token1] = true,
  },
  [17] = {
    [ts_external_token_file_descriptor] = true,
    [ts_external_token_variable_name] = true,
    [ts_external_token_test_operator] = true,
    [ts_external_token__brace_start] = true,
    [ts_external_token_LT_LT] = true,
    [ts_external_token_LT_LT_DASH] = true,
    [ts_external_token_heredoc_redirect_token1] = true,
    [ts_external_token_esac] = true,
  },
  [18] = {
    [ts_external_token_file_descriptor] = true,
    [ts_external_token_variable_name] = true,
    [ts_external_token_test_operator] = true,
    [ts_external_token__bare_dollar] = true,
    [ts_external_token__brace_start] = true,
    [ts_external_token_LT_LT] = true,
    [ts_external_token_LT_LT_DASH] = true,
    [ts_external_token_heredoc_redirect_token1] = true,
    [ts_external_token_esac] = true,
  },
  [19] = {
    [ts_external_token_file_descriptor] = true,
    [ts_external_token_variable_name] = true,
    [ts_external_token_test_operator] = true,
    [ts_external_token__brace_start] = true,
    [ts_external_token_LT_LT] = true,
    [ts_external_token_LT_LT_DASH] = true,
    [ts_external_token_heredoc_redirect_token1] = true,
  },
  [20] = {
    [ts_external_token_file_descriptor] = true,
    [ts_external_token_variable_name] = true,
    [ts_external_token_test_operator] = true,
    [ts_external_token__bare_dollar] = true,
    [ts_external_token__brace_start] = true,
    [ts_external_token_LT_LT] = true,
    [ts_external_token_LT_LT_DASH] = true,
    [ts_external_token_heredoc_redirect_token1] = true,
  },
  [21] = {
    [ts_external_token_file_descriptor] = true,
    [ts_external_token_test_operator] = true,
    [ts_external_token__brace_start] = true,
    [ts_external_token_LT_LT] = true,
    [ts_external_token_LT_LT_DASH] = true,
    [ts_external_token_heredoc_redirect_token1] = true,
    [ts_external_token_esac] = true,
  },
  [22] = {
    [ts_external_token_file_descriptor] = true,
    [ts_external_token_test_operator] = true,
    [ts_external_token__brace_start] = true,
    [ts_external_token_LT_LT] = true,
    [ts_external_token_LT_LT_DASH] = true,
    [ts_external_token_heredoc_redirect_token1] = true,
  },
  [23] = {
    [ts_external_token_file_descriptor] = true,
    [ts_external_token_test_operator] = true,
    [ts_external_token__bare_dollar] = true,
    [ts_external_token__brace_start] = true,
    [ts_external_token_LT_LT] = true,
    [ts_external_token_LT_LT_DASH] = true,
    [ts_external_token_LPAREN] = true,
  },
  [24] = {
    [ts_external_token_variable_name] = true,
    [ts_external_token_test_operator] = true,
    [ts_external_token__brace_start] = true,
    [ts_external_token_LPAREN] = true,
  },
  [25] = {
    [ts_external_token_file_descriptor] = true,
    [ts_external_token_test_operator] = true,
    [ts_external_token__bare_dollar] = true,
    [ts_external_token__brace_start] = true,
    [ts_external_token_LT_LT] = true,
    [ts_external_token_LT_LT_DASH] = true,
  },
  [26] = {
    [ts_external_token_file_descriptor] = true,
    [ts_external_token_test_operator] = true,
    [ts_external_token__brace_start] = true,
    [ts_external_token_heredoc_redirect_token1] = true,
  },
  [27] = {
    [ts_external_token_file_descriptor] = true,
    [ts_external_token_variable_name] = true,
    [ts_external_token_test_operator] = true,
    [ts_external_token__bare_dollar] = true,
    [ts_external_token__brace_start] = true,
    [ts_external_token_LT_LT] = true,
    [ts_external_token_LT_LT_DASH] = true,
    [ts_external_token_LPAREN] = true,
  },
  [28] = {
    [ts_external_token_file_descriptor] = true,
    [ts_external_token_variable_name] = true,
    [ts_external_token_test_operator] = true,
    [ts_external_token__bare_dollar] = true,
    [ts_external_token__brace_start] = true,
    [ts_external_token_LT_LT] = true,
    [ts_external_token_LT_LT_DASH] = true,
  },
  [29] = {
    [ts_external_token_file_descriptor] = true,
    [ts_external_token_variable_name] = true,
    [ts_external_token_test_operator] = true,
    [ts_external_token__brace_start] = true,
    [ts_external_token_LT_LT] = true,
    [ts_external_token_LT_LT_DASH] = true,
  },
  [30] = {
    [ts_external_token_file_descriptor] = true,
    [ts_external_token__concat] = true,
    [ts_external_token_test_operator] = true,
    [ts_external_token__bare_dollar] = true,
    [ts_external_token__brace_start] = true,
    [ts_external_token_LT_LT] = true,
    [ts_external_token_LT_LT_DASH] = true,
    [ts_external_token_heredoc_redirect_token1] = true,
    [ts_external_token_LPAREN] = true,
    [ts_external_token_esac] = true,
  },
  [31] = {
    [ts_external_token_file_descriptor] = true,
    [ts_external_token_variable_name] = true,
    [ts_external_token_test_operator] = true,
    [ts_external_token__brace_start] = true,
    [ts_external_token_RBRACK] = true,
    [ts_external_token_LT_LT] = true,
    [ts_external_token_LT_LT_DASH] = true,
  },
  [32] = {
    [ts_external_token_file_descriptor] = true,
    [ts_external_token__concat] = true,
    [ts_external_token_test_operator] = true,
    [ts_external_token__bare_dollar] = true,
    [ts_external_token__brace_start] = true,
    [ts_external_token_LT_LT] = true,
    [ts_external_token_LT_LT_DASH] = true,
    [ts_external_token_heredoc_redirect_token1] = true,
    [ts_external_token_esac] = true,
  },
  [33] = {
    [ts_external_token_file_descriptor] = true,
    [ts_external_token_test_operator] = true,
    [ts_external_token__brace_start] = true,
    [ts_external_token_RBRACK] = true,
    [ts_external_token_LT_LT] = true,
    [ts_external_token_LT_LT_DASH] = true,
  },
  [34] = {
    [ts_external_token_file_descriptor] = true,
    [ts_external_token_test_operator] = true,
    [ts_external_token__brace_start] = true,
    [ts_external_token_LT_LT] = true,
    [ts_external_token_LT_LT_DASH] = true,
  },
  [35] = {
    [ts_external_token__concat] = true,
    [ts_external_token_variable_name] = true,
    [ts_external_token_test_operator] = true,
    [ts_external_token_RBRACK] = true,
    [ts_external_token_LT_LT] = true,
  },
  [36] = {
    [ts_external_token_file_descriptor] = true,
    [ts_external_token_variable_name] = true,
    [ts_external_token_test_operator] = true,
    [ts_external_token__brace_start] = true,
  },
  [37] = {
    [ts_external_token_file_descriptor] = true,
    [ts_external_token__concat] = true,
    [ts_external_token_test_operator] = true,
    [ts_external_token__bare_dollar] = true,
    [ts_external_token__brace_start] = true,
    [ts_external_token_LT_LT] = true,
    [ts_external_token_LT_LT_DASH] = true,
    [ts_external_token_heredoc_redirect_token1] = true,
  },
  [38] = {
    [ts_external_token_file_descriptor] = true,
    [ts_external_token__concat] = true,
    [ts_external_token_variable_name] = true,
    [ts_external_token_test_operator] = true,
    [ts_external_token__brace_start] = true,
    [ts_external_token_LT_LT] = true,
    [ts_external_token_LT_LT_DASH] = true,
    [ts_external_token_heredoc_redirect_token1] = true,
    [ts_external_token_esac] = true,
  },
  [39] = {
    [ts_external_token_variable_name] = true,
    [ts_external_token_test_operator] = true,
    [ts_external_token_LT_LT] = true,
  },
  [40] = {
    [ts_external_token_variable_name] = true,
    [ts_external_token_test_operator] = true,
    [ts_external_token_RBRACK] = true,
    [ts_external_token_LT_LT] = true,
  },
  [41] = {
    [ts_external_token_file_descriptor] = true,
    [ts_external_token__concat] = true,
    [ts_external_token_variable_name] = true,
    [ts_external_token_test_operator] = true,
    [ts_external_token__brace_start] = true,
    [ts_external_token_LT_LT] = true,
    [ts_external_token_LT_LT_DASH] = true,
    [ts_external_token_heredoc_redirect_token1] = true,
  },
  [42] = {
    [ts_external_token_test_operator] = true,
    [ts_external_token_LT_LT] = true,
  },
  [43] = {
    [ts_external_token_file_descriptor] = true,
    [ts_external_token__concat] = true,
    [ts_external_token_test_operator] = true,
    [ts_external_token__brace_start] = true,
    [ts_external_token_LT_LT] = true,
    [ts_external_token_LT_LT_DASH] = true,
    [ts_external_token_heredoc_redirect_token1] = true,
    [ts_external_token_esac] = true,
  },
  [44] = {
    [ts_external_token_file_descriptor] = true,
    [ts_external_token__concat] = true,
    [ts_external_token_test_operator] = true,
    [ts_external_token__brace_start] = true,
    [ts_external_token_LT_LT] = true,
    [ts_external_token_LT_LT_DASH] = true,
    [ts_external_token_heredoc_redirect_token1] = true,
  },
  [45] = {
    [ts_external_token__concat] = true,
    [ts_external_token_test_operator] = true,
    [ts_external_token_RBRACK] = true,
    [ts_external_token_LT_LT] = true,
  },
  [46] = {
    [ts_external_token_file_descriptor] = true,
    [ts_external_token_variable_name] = true,
    [ts_external_token_test_operator] = true,
    [ts_external_token__brace_start] = true,
    [ts_external_token_heredoc_redirect_token1] = true,
  },
  [47] = {
    [ts_external_token_test_operator] = true,
    [ts_external_token_extglob_pattern] = true,
    [ts_external_token__brace_start] = true,
    [ts_external_token_LPAREN] = true,
  },
  [48] = {
    [ts_external_token_test_operator] = true,
    [ts_external_token_RBRACK] = true,
    [ts_external_token_LT_LT] = true,
  },
  [49] = {
    [ts_external_token_test_operator] = true,
    [ts_external_token__brace_start] = true,
    [ts_external_token_LPAREN] = true,
  },
  [50] = {
    [ts_external_token_test_operator] = true,
    [ts_external_token__regex_no_space] = true,
    [ts_external_token__brace_start] = true,
    [ts_external_token_LPAREN] = true,
  },
  [51] = {
    [ts_external_token_file_descriptor] = true,
    [ts_external_token__concat] = true,
    [ts_external_token_test_operator] = true,
    [ts_external_token__bare_dollar] = true,
    [ts_external_token__brace_start] = true,
    [ts_external_token_LT_LT] = true,
    [ts_external_token_LT_LT_DASH] = true,
    [ts_external_token_LPAREN] = true,
  },
  [52] = {
    [ts_external_token_file_descriptor] = true,
    [ts_external_token__concat] = true,
    [ts_external_token_test_operator] = true,
    [ts_external_token__bare_dollar] = true,
    [ts_external_token__brace_start] = true,
    [ts_external_token_LT_LT] = true,
    [ts_external_token_LT_LT_DASH] = true,
  },
  [53] = {
    [ts_external_token_file_descriptor] = true,
    [ts_external_token__concat] = true,
    [ts_external_token_variable_name] = true,
    [ts_external_token_test_operator] = true,
    [ts_external_token__brace_start] = true,
    [ts_external_token_RBRACK] = true,
    [ts_external_token_LT_LT] = true,
    [ts_external_token_LT_LT_DASH] = true,
  },
  [54] = {
    [ts_external_token__concat] = true,
    [ts_external_token_test_operator] = true,
    [ts_external_token_LT_LT] = true,
  },
  [55] = {
    [ts_external_token_test_operator] = true,
    [ts_external_token_extglob_pattern] = true,
    [ts_external_token__brace_start] = true,
    [ts_external_token_heredoc_redirect_token1] = true,
    [ts_external_token_LPAREN] = true,
    [ts_external_token_esac] = true,
  },
  [56] = {
    [ts_external_token_file_descriptor] = true,
    [ts_external_token__concat] = true,
    [ts_external_token_variable_name] = true,
    [ts_external_token_test_operator] = true,
    [ts_external_token__brace_start] = true,
    [ts_external_token_LT_LT] = true,
    [ts_external_token_LT_LT_DASH] = true,
  },
  [57] = {
    [ts_external_token_file_descriptor] = true,
    [ts_external_token__concat] = true,
    [ts_external_token_test_operator] = true,
    [ts_external_token__brace_start] = true,
    [ts_external_token_RBRACK] = true,
    [ts_external_token_LT_LT] = true,
    [ts_external_token_LT_LT_DASH] = true,
  },
  [58] = {
    [ts_external_token_test_operator] = true,
    [ts_external_token_extglob_pattern] = true,
    [ts_external_token_LT_LT] = true,
  },
  [59] = {
    [ts_external_token_file_descriptor] = true,
    [ts_external_token__concat] = true,
    [ts_external_token_test_operator] = true,
    [ts_external_token__brace_start] = true,
    [ts_external_token_LT_LT] = true,
    [ts_external_token_LT_LT_DASH] = true,
  },
  [60] = {
    [ts_external_token_LT_LT] = true,
  },
  [61] = {
    [ts_external_token__concat] = true,
    [ts_external_token_test_operator] = true,
    [ts_external_token_extglob_pattern] = true,
    [ts_external_token_RBRACK] = true,
    [ts_external_token_LT_LT] = true,
  },
  [62] = {
    [ts_external_token__concat] = true,
    [ts_external_token_LT_LT] = true,
  },
  [63] = {
    [ts_external_token_file_descriptor] = true,
    [ts_external_token_variable_name] = true,
    [ts_external_token_LT_LT] = true,
    [ts_external_token_LT_LT_DASH] = true,
    [ts_external_token_heredoc_redirect_token1] = true,
    [ts_external_token_esac] = true,
  },
  [64] = {
    [ts_external_token_LT_LT] = true,
    [ts_external_token_heredoc_redirect_token1] = true,
  },
  [65] = {
    [ts_external_token_test_operator] = true,
    [ts_external_token_extglob_pattern] = true,
    [ts_external_token_RBRACK] = true,
    [ts_external_token_LT_LT] = true,
  },
  [66] = {
    [ts_external_token_file_descriptor] = true,
    [ts_external_token_variable_name] = true,
    [ts_external_token_LT_LT] = true,
    [ts_external_token_LT_LT_DASH] = true,
    [ts_external_token_heredoc_redirect_token1] = true,
  },
  [67] = {
    [ts_external_token_test_operator] = true,
    [ts_external_token_extglob_pattern] = true,
    [ts_external_token__brace_start] = true,
    [ts_external_token_LPAREN] = true,
    [ts_external_token_esac] = true,
  },
  [68] = {
    [ts_external_token__concat] = true,
    [ts_external_token_RBRACK] = true,
    [ts_external_token_LT_LT] = true,
  },
  [69] = {
    [ts_external_token_RBRACK] = true,
    [ts_external_token_LT_LT] = true,
  },
  [70] = {
    [ts_external_token_file_descriptor] = true,
    [ts_external_token__concat] = true,
    [ts_external_token_test_operator] = true,
    [ts_external_token__brace_start] = true,
    [ts_external_token_heredoc_redirect_token1] = true,
  },
  [71] = {
    [ts_external_token_file_descriptor] = true,
    [ts_external_token__concat] = true,
    [ts_external_token_variable_name] = true,
    [ts_external_token_test_operator] = true,
    [ts_external_token__brace_start] = true,
  },
  [72] = {
    [ts_external_token_variable_name] = true,
    [ts_external_token_test_operator] = true,
    [ts_external_token__brace_start] = true,
    [ts_external_token_heredoc_redirect_token1] = true,
  },
  [73] = {
    [ts_external_token_test_operator] = true,
    [ts_external_token__brace_start] = true,
    [ts_external_token_heredoc_redirect_token1] = true,
  },
  [74] = {
    [ts_external_token_variable_name] = true,
    [ts_external_token_test_operator] = true,
    [ts_external_token__expansion_word] = true,
    [ts_external_token__brace_start] = true,
    [ts_external_token_RBRACE] = true,
    [ts_external_token_LPAREN] = true,
  },
  [75] = {
    [ts_external_token__empty_value] = true,
    [ts_external_token_test_operator] = true,
    [ts_external_token__brace_start] = true,
    [ts_external_token_LPAREN] = true,
  },
  [76] = {
    [ts_external_token__immediate_double_hash] = true,
    [ts_external_token__external_expansion_sym_hash] = true,
    [ts_external_token__external_expansion_sym_bang] = true,
    [ts_external_token__external_expansion_sym_equal] = true,
    [ts_external_token_RBRACE] = true,
  },
  [77] = {
    [ts_external_token_file_descriptor] = true,
    [ts_external_token_variable_name] = true,
    [ts_external_token_RBRACK] = true,
    [ts_external_token_LT_LT] = true,
    [ts_external_token_LT_LT_DASH] = true,
  },
  [78] = {
    [ts_external_token_test_operator] = true,
    [ts_external_token__brace_start] = true,
  },
  [79] = {
    [ts_external_token_test_operator] = true,
    [ts_external_token_extglob_pattern] = true,
    [ts_external_token__brace_start] = true,
  },
  [80] = {
    [ts_external_token_variable_name] = true,
    [ts_external_token_test_operator] = true,
    [ts_external_token__brace_start] = true,
  },
  [81] = {
    [ts_external_token_file_descriptor] = true,
    [ts_external_token_variable_name] = true,
    [ts_external_token_LT_LT] = true,
    [ts_external_token_LT_LT_DASH] = true,
  },
  [82] = {
    [ts_external_token_test_operator] = true,
    [ts_external_token__bare_dollar] = true,
    [ts_external_token__brace_start] = true,
    [ts_external_token_RBRACK] = true,
  },
  [83] = {
    [ts_external_token_test_operator] = true,
    [ts_external_token_regex] = true,
    [ts_external_token__brace_start] = true,
  },
  [84] = {
    [ts_external_token__immediate_double_hash] = true,
    [ts_external_token_RBRACE] = true,
  },
  [85] = {
    [ts_external_token_test_operator] = true,
    [ts_external_token__bare_dollar] = true,
    [ts_external_token__brace_start] = true,
  },
  [86] = {
    [ts_external_token_variable_name] = true,
    [ts_external_token_LPAREN] = true,
  },
  [87] = {
    [ts_external_token_file_descriptor] = true,
    [ts_external_token_variable_name] = true,
    [ts_external_token_heredoc_redirect_token1] = true,
  },
  [88] = {
    [ts_external_token_heredoc_redirect_token1] = true,
    [ts_external_token_LPAREN] = true,
  },
  [89] = {
    [ts_external_token_file_descriptor] = true,
    [ts_external_token__concat] = true,
    [ts_external_token_LT_LT] = true,
    [ts_external_token_LT_LT_DASH] = true,
    [ts_external_token_heredoc_redirect_token1] = true,
    [ts_external_token_esac] = true,
  },
  [90] = {
    [ts_external_token__immediate_double_hash] = true,
  },
  [91] = {
    [ts_external_token_file_descriptor] = true,
    [ts_external_token__concat] = true,
    [ts_external_token_variable_name] = true,
    [ts_external_token_LT_LT] = true,
    [ts_external_token_LT_LT_DASH] = true,
    [ts_external_token_heredoc_redirect_token1] = true,
    [ts_external_token_esac] = true,
  },
  [92] = {
    [ts_external_token_file_descriptor] = true,
    [ts_external_token_LT_LT] = true,
    [ts_external_token_LT_LT_DASH] = true,
    [ts_external_token_heredoc_redirect_token1] = true,
    [ts_external_token_esac] = true,
  },
  [93] = {
    [ts_external_token_file_descriptor] = true,
    [ts_external_token__concat] = true,
    [ts_external_token_LT_LT] = true,
    [ts_external_token_LT_LT_DASH] = true,
    [ts_external_token_heredoc_redirect_token1] = true,
  },
  [94] = {
    [ts_external_token_file_descriptor] = true,
    [ts_external_token_LT_LT] = true,
    [ts_external_token_LT_LT_DASH] = true,
    [ts_external_token_heredoc_redirect_token1] = true,
  },
  [95] = {
    [ts_external_token_file_descriptor] = true,
    [ts_external_token__concat] = true,
    [ts_external_token_variable_name] = true,
    [ts_external_token_LT_LT] = true,
    [ts_external_token_LT_LT_DASH] = true,
    [ts_external_token_heredoc_redirect_token1] = true,
  },
  [96] = {
    [ts_external_token__concat] = true,
    [ts_external_token__immediate_double_hash] = true,
    [ts_external_token__external_expansion_sym_hash] = true,
    [ts_external_token__external_expansion_sym_bang] = true,
    [ts_external_token__external_expansion_sym_equal] = true,
    [ts_external_token_RBRACE] = true,
  },
  [97] = {
    [ts_external_token__concat] = true,
    [ts_external_token_test_operator] = true,
    [ts_external_token__brace_start] = true,
    [ts_external_token_heredoc_redirect_token1] = true,
  },
  [98] = {
    [ts_external_token_variable_name] = true,
    [ts_external_token__external_expansion_sym_hash] = true,
    [ts_external_token__external_expansion_sym_bang] = true,
    [ts_external_token__external_expansion_sym_equal] = true,
    [ts_external_token_RBRACE] = true,
  },
  [99] = {
    [ts_external_token_LPAREN] = true,
  },
  [100] = {
    [ts_external_token__concat] = true,
    [ts_external_token_test_operator] = true,
    [ts_external_token__brace_start] = true,
  },
  [101] = {
    [ts_external_token_file_descriptor] = true,
    [ts_external_token__concat] = true,
    [ts_external_token_RBRACK] = true,
    [ts_external_token_LT_LT] = true,
    [ts_external_token_LT_LT_DASH] = true,
  },
  [102] = {
    [ts_external_token_file_descriptor] = true,
    [ts_external_token_RBRACK] = true,
    [ts_external_token_LT_LT] = true,
    [ts_external_token_LT_LT_DASH] = true,
  },
  [103] = {
    [ts_external_token_file_descriptor] = true,
    [ts_external_token_LT_LT] = true,
    [ts_external_token_LT_LT_DASH] = true,
  },
  [104] = {
    [ts_external_token_RBRACE] = true,
    [ts_external_token_heredoc_redirect_token1] = true,
    [ts_external_token_LPAREN] = true,
  },
  [105] = {
    [ts_external_token_file_descriptor] = true,
    [ts_external_token__concat] = true,
    [ts_external_token_variable_name] = true,
    [ts_external_token_LT_LT] = true,
    [ts_external_token_LT_LT_DASH] = true,
  },
  [106] = {
    [ts_external_token_file_descriptor] = true,
    [ts_external_token__concat] = true,
    [ts_external_token_LT_LT] = true,
    [ts_external_token_LT_LT_DASH] = true,
  },
  [107] = {
    [ts_external_token_variable_name] = true,
    [ts_external_token__expansion_word] = true,
    [ts_external_token_RBRACE] = true,
    [ts_external_token_LPAREN] = true,
  },
  [108] = {
    [ts_external_token_variable_name] = true,
    [ts_external_token__expansion_word] = true,
    [ts_external_token_LPAREN] = true,
  },
  [109] = {
    [ts_external_token_variable_name] = true,
    [ts_external_token_heredoc_redirect_token1] = true,
  },
  [110] = {
    [ts_external_token_file_descriptor] = true,
    [ts_external_token_heredoc_redirect_token1] = true,
  },
  [111] = {
    [ts_external_token_file_descriptor] = true,
    [ts_external_token__concat] = true,
    [ts_external_token_heredoc_redirect_token1] = true,
  },
  [112] = {
    [ts_external_token_variable_name] = true,
  },
  [113] = {
    [ts_external_token_LT_LT] = true,
    [ts_external_token_LT_LT_DASH] = true,
  },
  [114] = {
    [ts_external_token_RBRACE] = true,
  },
  [115] = {
    [ts_external_token__concat] = true,
  },
  [116] = {
    [ts_external_token_heredoc_content] = true,
    [ts_external_token_heredoc_end] = true,
  },
  [117] = {
    [ts_external_token__regex_no_slash] = true,
    [ts_external_token_RBRACE] = true,
  },
  [118] = {
    [ts_external_token__concat] = true,
    [ts_external_token_heredoc_redirect_token1] = true,
  },
  [119] = {
    [ts_external_token_regex] = true,
    [ts_external_token_RBRACE] = true,
  },
  [120] = {
    [ts_external_token_heredoc_redirect_token1] = true,
  },
  [121] = {
    [ts_external_token__concat] = true,
    [ts_external_token__expansion_word] = true,
    [ts_external_token_RBRACE] = true,
  },
  [122] = {
    [ts_external_token_simple_heredoc_body] = true,
    [ts_external_token__heredoc_body_beginning] = true,
  },
  [123] = {
    [ts_external_token__concat] = true,
    [ts_external_token__external_expansion_sym_hash] = true,
    [ts_external_token__external_expansion_sym_bang] = true,
    [ts_external_token__external_expansion_sym_equal] = true,
    [ts_external_token_RBRACE] = true,
  },
  [124] = {
    [ts_external_token__concat] = true,
    [ts_external_token_RBRACE] = true,
  },
  [125] = {
    [ts_external_token__external_expansion_sym_hash] = true,
    [ts_external_token__external_expansion_sym_bang] = true,
    [ts_external_token__external_expansion_sym_equal] = true,
    [ts_external_token_RBRACE] = true,
  },
  [126] = {
    [ts_external_token_esac] = true,
  },
  [127] = {
    [ts_external_token_extglob_pattern] = true,
  },
  [128] = {
    [ts_external_token_heredoc_end] = true,
  },
  [129] = {
    [ts_external_token_RBRACK] = true,
  },
  [130] = {
    [ts_external_token_heredoc_start] = true,
  },
};

#ifdef __cplusplus
extern "C" {
#endif
void *tree_sitter_bash_external_scanner_create(void);
void tree_sitter_bash_external_scanner_destroy(void *);
bool tree_sitter_bash_external_scanner_scan(void *, TSLexer *, const bool *);
unsigned tree_sitter_bash_external_scanner_serialize(void *, char *);
void tree_sitter_bash_external_scanner_deserialize(void *, const char *, unsigned);

#ifdef _WIN32
#define TS_PUBLIC __declspec(dllexport)
#else
#define TS_PUBLIC __attribute__((visibility("default")))
#endif

TS_PUBLIC const TSLanguage *tree_sitter_bash() {
  static const TSLanguage language = {
    .version = LANGUAGE_VERSION,
    .symbol_count = SYMBOL_COUNT,
    .alias_count = ALIAS_COUNT,
    .token_count = TOKEN_COUNT,
    .external_token_count = EXTERNAL_TOKEN_COUNT,
    .state_count = STATE_COUNT,
    .large_state_count = LARGE_STATE_COUNT,
    .production_id_count = PRODUCTION_ID_COUNT,
    .field_count = FIELD_COUNT,
    .max_alias_sequence_length = MAX_ALIAS_SEQUENCE_LENGTH,
    .parse_table = &ts_parse_table[0][0],
    .small_parse_table = ts_small_parse_table,
    .small_parse_table_map = ts_small_parse_table_map,
    .parse_actions = ts_parse_actions,
    .symbol_names = ts_symbol_names,
    .field_names = ts_field_names,
    .field_map_slices = ts_field_map_slices,
    .field_map_entries = ts_field_map_entries,
    .symbol_metadata = ts_symbol_metadata,
    .public_symbol_map = ts_symbol_map,
    .alias_map = ts_non_terminal_alias_map,
    .alias_sequences = &ts_alias_sequences[0][0],
    .lex_modes = ts_lex_modes,
    .lex_fn = ts_lex,
    .keyword_lex_fn = ts_lex_keywords,
    .keyword_capture_token = sym_word,
    .external_scanner = {
      &ts_external_scanner_states[0][0],
      ts_external_scanner_symbol_map,
      tree_sitter_bash_external_scanner_create,
      tree_sitter_bash_external_scanner_destroy,
      tree_sitter_bash_external_scanner_scan,
      tree_sitter_bash_external_scanner_serialize,
      tree_sitter_bash_external_scanner_deserialize,
    },
    .primary_state_ids = ts_primary_state_ids,
  };
  return &language;
}
#ifdef __cplusplus
}
#endif
